# Table of Contents

1. [Routing](#routing)
2. [Django applications](#django-applications)
      - [apps](#apps)
      - [datasets](#datasets)
      - [jobs](#jobs)
      - [rbac](#rbac)
      - [reports](#reports)
      - [searches](#searches)
      - [sql](#sql)
      - [workspace](#workspace)
3. [Other modules](#other-modules)
      - [dataconn](#dataconn)
      - [vizlib](#vizlib)
      - [sqlrun](#sqlrun)
4. [Testing](#testing)
      - [CLI](#cli)
      - [PyCharm](#pycharm)
      - [Visual Studio Code](#visual-studio-code)

## Routing

Note that the base arcwebbase/urls.py file maps URLs to the apps below. This mapping is:

```python
urlpatterns = [
  re_path(r'^$', views.home, name='root-url'),

  re_path(r'^stats_data$', views.stats_data, name='stats-data-url'),

  re_path(r'^actions/', include('actions.urls')),
  re_path(r'^api/', include('api.urls')),
  re_path(r'^searches/', include('searches.urls')),
  re_path(r'^apps/', include('apps.urls')),
  re_path(r'^reports/', include('reports.urls')),
  re_path(r'^datasets/', include('datasets.urls')),
  re_path(r'^datasources/', include('datasources.urls')),
  re_path(r'^roles/', include('rbac.urls')),
  re_path(r'^sqlrun/', include('sqlrun.urls')),
  re_path(r'^trustedauth/', include('trustedauth.urls')),
  re_path(r'^jobs/', include('jobs.urls')),
  re_path(r'^migration/', include('migration.urls')),
  re_path(r'^workspace/', include('workspace.urls')),
  re_path(r'^sql/', include('sql.urls')),
]
```

## Django applications

The following django app directories are linked from here

### apps

- Data models
  1. UserProfile: store per-user information
  2. APIKey: store api keys used for access to the vizapps API

- Views
  1. all the apps/ URLs get served from here
  2. most of the main pages and admin pages start here
     - example: home page, visuals page, dashboard view/edit pages, app view/edit pages
  3. most django template files are stored within arcweb/apps/templates/, even for pages
    not served by apps/ URLs

### datasets

- Data models
  1. DataConnection: store information about data connections that this vizapps instance connects to
  2. DataSet: store information about datasets modeled in vizapps
  3. Segment: dataset segments are stored here
  4. Event: dataset events are stored here
  5. FilterAssociation: dataset filter associations are stored here

- Views
  1. all the datasets/ URLs get served from here
  2. mainly data connection and dataset administration pages

### jobs

- Data models
  1. ScheduleGroup. Track a particular schedule (e.g. Weekly or Daily)
  2. JobSchedule
  3. JobTarget
  4. JobLog
  5. JobContent

- Views
  1. all the jobs/ URLs get served from here
     - used for job scheduling and status

### rbac

- Data models
  1. Role: store roles (like Analyst, Database admin, etc)

- Views
  1. all the rbac/ URLs get served from here
     - used for RBAC role administration

### reports

- Data models
  1. StaticAsset. store static assets (.js files, .jpg files) that are uploaded to viz apps
  2. CustomCSS. store custom css/js extensions
  3. ColorPalette. store custom color palettes
  4. ParameterSet. store a set of parameeters that can be referenced via a short url reference
  5. AppSetting. site-wide setting key/value pairs referenced in viz apps via 'Site Settings'
  6. AppGroup. store appgroup definitions
  7. UserAppGroupMembership. user -> appgroup pinned mapping
  8. Report. store visual information for a single visual
  9. ReportTemplate. store visual templates. I believe this is an unused feature now
  10. ReportImage. visual snapshot
  11. ReportUser. user -> report pinned mapping
  12. ReportUsage. track report usage over time for stats
  13. AppGroupAppMembership. map dashboard to appgroup mapping

### searches

- Data models
  1. Search. store historical searches
  2. Bookmark. store bookmarked searches

### sql

- Data models
  1. Bookmark. store bookmarkes queries
  2. Query. store query history in the 'Direct Access' screen

### workspace

- Data models
  1. Workspace: store information about workspaces

## Other modules

### dataconn

All the code related to connecting to backend data connections is in this direcotry. There is a file for each connection. For example, the connector for postgres is in dotaconn_pg.py

Each connection implements 2 interfaces, both defined in dataconn_dataengine.py

ConnProperties is a set of properties that modify the behavior of helper functions that work with multiple connectors. Some of these properties are about the specific quote characters in each data connection and how to escape them, others are related to SQL variants that are supported by the connectors. Some example properties are below

```python
    self.quote = "`"
    self.quote_end = "`"
    self.quote_alias = "`"
    self.quote_escape_character = "\\"
    self.quote_string_constant = '"'
    self.supports_partitions = True
    self.supports_async = True
    self.supports_backticks = True
    self.supports_reuse = True
    self.supports_order_outside_union = True
    self.supports_limit_outside_union = True
```

DataEngine is a set of interfaces that need to be supported by each connection, along with some helper implementations. The most obvious example here is the interface used to run queries against the connection, but there are many other functions needed as well because of the complexities of working with a variety of types of connections.

```python
  def sync_runquery(self, query, build_request=None, disable_codegen=True,
                    username=None, sample_pct=None, comment='',
                    stream=False, qcntl=None):
```

### vizlib

vizlib/ has a set of helper libraries that generate SQL statements for different types of backend connections. Queries from the frontend come to the python layer as dataset requests. Dataset requests have a dataset id, a set of dimensions and aggregates, and other properties that try to define the data that needs to be fetched from the backend in a backend-neutral specification. vizlib/ files convert this spec into an actual SQL command for use with the specific backend that the dataset comes from.

### sqlrun

sqlrun/ has a set of helper libraries that help schedule the execution of queries against the different data connections that viz apps is connected to. It translates the long-running queries that run against backends into fast-response operations that respond back to the UI with an async id in many cases that can be used to poll for results. It also supports interfaces where a set of dataset requests can be batched together - this is used to speed up messaging between the client and server when a multi-visual dashboard is being loaded.

## Testing

### CLI

```scons pytests```

### PyCharm

1. Run > Edit Configurations
2. Add a new test configuration
    - Target: vizlib.tests
    - Environmental variables: DJANGO_SETTINGS_MODULES:arcwebbase.settings_tests
3. Run configuration

### Visual Studio Code

1. Go to "Run and Debug"
2. On the top bar click on the dropdown
3. Select add configuration
4. Add the following object in tho configurations array

```json
{
  "name": "Python: test",
  "type": "python",
  "request": "launch",
  "program": "${workspaceFolder}/arcweb/manage.py",
  "args": [
    "test",
    // Uncomment to change port: 8000 is the default
    "8000",
    // Run just certain tests
    // "`echo -n ${relativeFileDirname} | tr \/ .`.${fileBasenameNoExtension}"

  ],
  "env": {
    "PYTHONPATH": "~/.arc",
    "DJANGO_SETTINGS_MODULE": "arcwebbase.settings_tests",
    // Settings for downloading PNG/PDF
    "LC_ALL": "en_US.UTF-8",
    "LANG": "en_US.UTF-8",
    "CHROMIUM_PATH": "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
  },
  "django": true
}
```
