# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

# Provide a standard web interface for actions

from django.http import HttpResponse
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_http_methods

from jobs.views_encryption import decrypt_secret

import reports.views_settings

import logging
import json

_WEBHOOK_POOL = None

LOGGER = logging.getLogger("arcweb")

try:
  import urllib3
  _WEBHOOK_POOL = urllib3.PoolManager()
except Exception:
  LOGGER.info("Did not find urllib3 for webhooks")

JIRA_LIB_AVAILABLE = False

try:
  from jira import JIRA
  JIRA_LIB_AVAILABLE = True
except Exception:
  LOGGER.info('JIRA integration not possible because of missing libraries')
  pass

@login_required
@require_http_methods(["POST"])
def webhookinvoke(request):
  url = request.POST.get("url", "")
  request_type = request.POST.get("request_type", "")
  headers = request.POST.get("headers", None)
  parameters = request.POST.get("parameters", "{}")

  webhook_whitelist = reports.views_settings.setting_get_single("webhook_whitelist")
  valid_urls = webhook_whitelist.split("\n")
  if not url in valid_urls:
    return HttpResponse(
        "URL specified is not in webhook settings whitelist. Please contact your system administrator.",
        status=500)

  if not request_type in ["GET", "POST", "PUT"]:
    return HttpResponse(
        "Request type is invalid",
        status=500)

  LOGGER.info("Invoking webhook. Type (%s), URL (%s), Headers (%s), Data (%s)",
      request_type, url, str(headers), str(parameters))

  response = _WEBHOOK_POOL.request(
      request_type,
      url,
      headers=headers,
      body=json.dumps(parameters))

  status = response.status

  LOGGER.info("Done invoking webhook. Type (%s), URL (%s). Response status (%s).",
      request_type, url, status)

  if status != 200:
    LOGGER.info("webhook call failed: %s %s", status, str(response))
    return HttpResponse(
        "Failed. Status %s" % (status),
        status=500)

  return HttpResponse("success", content_type="text/plain")

@login_required
@require_http_methods(["GET"])
def webhookwhitelist(request):
  data = {}
  settings = ["webhook_whitelist"]
  for s in settings:
    data[s] = reports.views_settings.setting_get_single(s)

  return HttpResponse(json.dumps(data), content_type="application/json")

def __jiraChecks():
  """check if JIRA functionality is available and enabled"""

  if not JIRA_LIB_AVAILABLE:
    return HttpResponse(
        'JIRA integration is not available because of missing dependencies',
        status=500)

  enabled = reports.views_settings.setting_get_single('jira_enabled')
  if not enabled:
    return HttpResponse('JIRA integration is not enabled',
        status=500)

  return None

def __jiraLoggedIn():
  server = reports.views_settings.setting_get_single('jira_server')
  user = reports.views_settings.setting_get_single('jira_user')
  password = reports.views_settings.setting_get_single('jira_password')
  password = decrypt_secret(**json.loads(password))
  jira = JIRA(basic_auth=(user, password), server=server)

  return jira

@login_required
@require_http_methods(["GET"])
def jiraDefaults(request):
  err = __jiraChecks()
  if err:
    return err

  data = {}
  settings = ["jira_newissue_defaults"]
  for s in settings:
    data[s] = reports.views_settings.setting_get_single(s)

  return HttpResponse(json.dumps(data), content_type="application/json")

@login_required
@require_http_methods(["GET"])
def jiraProperties(request):
  err = __jiraChecks()
  if err:
    return err

  jira = __jiraLoggedIn()

  out = {
    "projects":
        [{
          "key": project.key,
          "name": project.name
         } for project in jira.projects()],

    "priorities":
        [{
          "id": priority.id,
          "name": priority.name
         } for priority in jira.priorities()]
  }

  return HttpResponse(json.dumps(out), content_type="application/json")

@login_required
@require_http_methods(["POST"])
def jiraCreate(request):
  """create a JIRA ticket"""

  err = __jiraChecks()
  if err:
    return err

  summary = request.POST.get("summary", "")
  description = request.POST.get("description", "")
  project = request.POST.get("project", "")
  priority = request.POST.get("priority", "")
  LOGGER.info('User %s creating JIRA issue', request.user.username)
  description += "\nUser: " + request.user.username
  other = request.POST.get("other", "")

  jira_params = dict()
  if other:
    jira_params = json.loads(other)

  if len(jira_params) == 0:
    jira_newissue_defaults = reports.views_settings.setting_get_single('jira_newissue_defaults')
    if jira_newissue_defaults:
      jira_params = json.loads(jira_newissue_defaults)

  jira_params['summary'] = summary
  jira_params['description'] = description
  if project:
    jira_params['project'] = project
  if priority:
    jira_params['priority'] = {
      "id": priority
    }

  jira = __jiraLoggedIn()

  try:
    newIssue = jira.create_issue(**jira_params)
  except Exception as e:
    error_message = 'Failed to create JIRA: %s' % str(e)
    LOGGER.exception(error_message)
    return HttpResponse(error_message, status=500)

  out = {}

  out['status'] = 'success'
  out['id'] = newIssue.key

  LOGGER.info('User %s created JIRA issue %s', request.user.username, newIssue.key)

  return HttpResponse(json.dumps(out), content_type="application/json")
