# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import django.test
from django.contrib.auth.models import User
import gc
import json

from arcwebbase import gcmonitor
from apps.tests.userrole import base
from requests.adapters import HTTPAdapter


class AdminApiTestCase(django.test.TestCase):

    def setUp(self):

        self.user = User.objects.create_user(
            username='testuser',
            email='testuser@example.com',
            password='testpassword'
        )

        self.apikeyRequestData = {
            'username': 'testuser',
            'password': 'testpassword',
            'active': 'true',
            'expires': ''
        }

        self.perm_editers = base.make_role("perm_editers", ["sys_editperm"])
        self.perm_editers.add_users([self.user.username])
        self.perm_editers.save()

        response = self.client.post('/apps/demoapikeys', data=self.apikeyRequestData)
        self.secret_apikey = json.loads(response.content)['secret_apikey']

        self.headers = {
         'HTTP_AUTHORIZATION': f'apikey {self.secret_apikey}',
       }

    def tearDown(self):
        self.user.delete()

    def test_edit_userprofile_with_valid_credentials_empty_data(self):

        response = self.client.post(f'/adminapi/edituserprofile/{self.user.username}', data={}, **self.headers)
        result = json.loads(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertIn('should_change_password', result)
        self.assertIn('is_password_editable', result)
        self.assertIn('jwt_token', result)
        self.assertIn('user_data', result)
        self.assertIn('authentication_source', result)

    def test_edit_userprofile_with_invalid_credentials(self):

        headers = {
         'HTTP_AUTHORIZATION': 'apikey thisisanoutmostinvalidapikey',
       }

        response = self.client.get(f'/adminapi/edituserprofile/{self.user.username}', data={}, **headers)
        self.assertEqual(response.content.decode(), 'Unauthenticated user')
        self.assertEqual(response.status_code, 401)

    def test_edit_userprofile_disable_new_password_popup(self):

        data = {
            'should_change_password': '0'
        }

        response = self.client.post(f'/adminapi/edituserprofile/{self.user.username}', data=data, **self.headers)
        result = json.loads(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(result['should_change_password'], False)


    def test_edit_userprofile_force_new_password_popup(self):

        data = {
            'should_change_password': '1'
        }

        response = self.client.post(f'/adminapi/edituserprofile/{self.user.username}', data=data, **self.headers)
        result = json.loads(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(result['should_change_password'], True)

    def test_edit_userprofile_re_enable_new_password_popup(self):
        # Double checking to make sure it not works only because of the initial state of `should_change_password`
        data = {
            'should_change_password': '1'
        }

        response = self.client.post(f'/adminapi/edituserprofile/{self.user.username}', data=data, **self.headers)
        result = json.loads(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(result['should_change_password'], True)

    def test_edit_userprofile_without_valid_username(self):

        data = {
            'should_change_password': '1'
        }

        invalid_name = 'no_gin_tonic_no_party'

        response = self.client.post(f'/adminapi/edituserprofile/{invalid_name}', data=data, **self.headers)
        self.assertEqual(response.content.decode(), 'username not found')
        self.assertEqual(response.status_code, 500)


    def test_edit_userprofile_with_bad_request_type(self):

        response = self.client.get(f'/adminapi/edituserprofile/{self.user.username}', data={}, **self.headers)
        self.assertEqual(response.content.decode(), 'Invalid request type')
        self.assertEqual(response.status_code, 401)

class TestAdminapiLoglevel(django.test.TestCase):
    def setUp(self):
        AdminApiTestCase.setUp(self)

        # changing the log level requires viewlogs permission:
        self.perm_editers.add_priv('system', ['sys_viewlogs'])

    def test_get_no_arg(self):
        response = self.client.get('/adminapi/loglevel', **self.headers)
        result = json.loads(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(result['logger'], 'root')

    def test_invalid_body(self):
        result = self.client.post('/adminapi/loglevel/a.b', content_type='application/json', **self.headers)
        self.assertEquals(result.status_code, 400)
        self.assertEquals(result.json()['error'], 'Desired level must be specified in the request body.')

    def test_invalid_level(self):
        result = self.client.post('/adminapi/loglevel/a.b', data={'level': 'NO_SUCH_LEVEL'}, content_type='application/json', **self.headers)
        self.assertEquals(result.status_code, 400)
        self.assertEqual(result.json()['error'],'level NO_SUCH_LEVEL not recognized')

    def test_set_get(self):

        original_level = self.client.get('/adminapi/loglevel/a.b', **self.headers).json()['level']
        self.assertIsNotNone(original_level)

        result = self.client.post('/adminapi/loglevel/a.b', data={'level': 'FATAL'}, content_type='application/json', **self.headers).json()
        self.assertNotEqual(result['level'], original_level)
        self.assertEquals(result, {'logger':'a.b', 'level': 'CRITICAL'}) # CRITICAL is synonym to FATAL

        result = self.client.get('/adminapi/loglevel/a.b.c', **self.headers).json()
        self.assertEquals(result, {'logger':'a.b.c', 'level': 'CRITICAL'})

class TestAdminapiGCStats(django.test.TestCase):
    def setUp(self):
        AdminApiTestCase.setUp(self)

        # changing the garbage collector debug settings requires viewlogs permission:
        self.perm_editers.add_priv('system', ['sys_viewlogs'])
        self.gc_debug = gc.get_debug()

    def tearDown(self):
        gc.set_debug(self.gc_debug)
        AdminApiTestCase.tearDown(self)

    def test_authorization(self):
        self.perm_editers.delete_users([self.user.username])

        # retrieval does not need special permissions
        response = self.client.get('/adminapi/gcstats', **self.headers)
        result = json.loads(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(result['debug_flags'], self.gc_debug)

        # posting should fail
        result = self.client.post('/adminapi/gcstats', data={}, content_type='application/json', **self.headers)
        self.assertEqual(result.status_code, 500)

    def test_get_no_arg(self):
        response = self.client.get('/adminapi/gcstats', **self.headers)
        result = json.loads(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(result['debug_flags'], self.gc_debug)

    def test_set_get(self):

        # post with empty data sets debugging off
        result = self.client.post('/adminapi/gcstats', data={}, content_type='application/json', **self.headers).json()
        self.assertEqual(result['debug_flags'], 0)

        result = self.client.post('/adminapi/gcstats', data={'debug_flags': 'DEBUG_STATS'}, content_type='application/json', **self.headers).json()
        self.assertEqual(result['debug_flags'], gc.DEBUG_STATS)

        result = self.client.get('/adminapi/gcstats', **self.headers).json()
        self.assertEqual(result['debug_flags'], gc.DEBUG_STATS)

        # post with invalid flag data sets debugging off
        result = self.client.post('/adminapi/gcstats', data={'debug_flags': 'None'}, content_type='application/json', **self.headers).json()
        self.assertEqual(result['debug_flags'], 0)

        result = self.client.get('/adminapi/gcstats', **self.headers).json()
        self.assertEqual(result['debug_flags'], 0)

class TestAdminapiGCMonitor(django.test.TestCase):
    def setUp(self):
        AdminApiTestCase.setUp(self)

        # enabling garbage collector monitoring requires viewlogs permission:
        self.perm_editers.add_priv('system', ['sys_viewlogs'])
        gcmonitor.INSTANCE.unregister()

    def tearDown(self):
        gcmonitor.INSTANCE.unregister()
        AdminApiTestCase.tearDown(self)

    def test_authorization(self):
        self.perm_editers.delete_users([self.user.username])

        # retrieval does not need special permissions
        response = self.client.get('/adminapi/gcmonitor', **self.headers)
        result = json.loads(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertFalse(result['enabled'])

        # posting should fail
        result = self.client.post('/adminapi/gcmonitor', data={}, content_type='application/json', **self.headers)
        # RBAC.checkPerm throws an error, 500 is returned instead of 401:
        self.assertEqual(result.status_code, 500)

    def test_set_get(self):

        # post with empty data is invalid
        result = self.client.post('/adminapi/gcmonitor', data={}, content_type='application/json', **self.headers)
        self.assertEqual(result.status_code, 400)

        result = self.client.post('/adminapi/gcmonitor', data={'enabled': True}, content_type='application/json', **self.headers).json()
        self.assertTrue(result['enabled'])

        result = self.client.get('/adminapi/gcmonitor', **self.headers).json()
        self.assertTrue(result['enabled'])
        self.assertTrue(gcmonitor.INSTANCE.enabled())

        result = self.client.post('/adminapi/gcmonitor', data={'enabled': False}, content_type='application/json', **self.headers).json()
        self.assertFalse(result['enabled'])

        result = self.client.get('/adminapi/gcmonitor', **self.headers).json()
        self.assertFalse(result['enabled'])
        self.assertFalse(gcmonitor.INSTANCE.enabled())


