# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import django.test

from apps.tests.userrole import base
from adminapi.tests import tests_views_adminapi

from unittest import mock

function_body = mock.Mock()
method_body = mock.Mock()


def function_to_profile():
  function_body()


class ClassToProfile:
  def method_to_profile(self):
    method_body()


class TestAdminapiCprofile(django.test.TestCase):
  def setUp(self):
    tests_views_adminapi.AdminApiTestCase.setUp(self)

    # changing the log level requires viewlogs permission:
    self.perm_editers.add_priv('system', ['sys_viewlogs'])
    function_body.reset_mock()
    method_body.reset_mock()

  def cprofile_get(self, data={}):
    # headers contain API keys created during the setup
    # response = self.client.request(method='GET', path='/adminapi/cprofile', **self.headers)

    response = self.client.get('/adminapi/cprofile', data=data, **self.headers)
    return response

  def cprofile_post(self, data: dict = {}):
    # headers contain API keys created during the setup
    # response = self.client.request(method='GET', path='/adminapi/cprofile', **self.headers)
    query = ''
    if (len(data)):
      values = [f'{key}={value}' for key, value in data.items()]
      query = '&'.join(values)
      query = '?' + query
    response = self.client.post(
      '/adminapi/cprofile' + query, data=data, **self.headers)
    return response

  def test_authorization(self):
    # create a user with no permissions:
    user1 = base.make_user('user1')
    user1_client = base.create_client(user1)
    response = user1_client.get('/adminapi/cprofile')
    self.assertEqual(response.status_code, 500)

    response = user1_client.post('/adminapi/cprofile')
    self.assertEqual(response.status_code, 500)

    response = self.cprofile_get()
    self.assertEqual(response.status_code, 200)

  @mock.patch('adminapi.views_cprofile.LOG')
  def test_function_profiling(self, LOG: mock.Mock):

    function_test = {
        'module': __name__,
        'func': function_to_profile.__name__
    }

    response = self.cprofile_get(function_test).json()
    self.assertFalse(response['profiler_on'])
    self.assertTrue('function_to_profile' in response['target_function'])

    response = self.cprofile_post(function_test).json()
    self.assertTrue(response['profiler_on'])
    self.assertTrue('function_to_profile' in response['target_function'])

    response = self.cprofile_get(function_test).json()
    self.assertTrue(response['profiler_on'])

    self.assertEqual(LOG.info.call_count, 0)
    function_to_profile()
    # we issue 3 logs:
    self.assertEqual(LOG.info.call_count, 3)
    log_message = LOG.info.call_args_list[2].args[0]
    self.assertTrue('Ordered by: cumulative time' in log_message)

    # by default we only print concise information. Module paths, callees+callers information is missing
    self.assertFalse(
      'arcweb/adminapi/tests/tests_views_cprofile.py' in log_message)
    # callees:
    self.assertFalse('called...' in log_message)
    # callees:
    self.assertFalse('called...' in log_message)
    # callers:
    self.assertFalse('was called by...' in log_message)

    # switch profiling off by POSTing again:
    response = self.cprofile_post(function_test).json()
    self.assertFalse(response['profiler_on'])

    LOG.info.reset_mock()
    # invoke the function again
    function_to_profile()
    # no logs this time:
    self.assertEqual(LOG.info.call_count, 0)
    self.assertEqual(function_body.call_count, 2)

  @mock.patch('adminapi.views_cprofile.LOG')
  def test_method_profiling(self, LOG: mock.Mock):

    function_test = {
        'module': __name__,
        'owner': ClassToProfile.__name__,
        'func': ClassToProfile.method_to_profile.__name__
    }

    response = self.cprofile_get(function_test).json()
    self.assertFalse(response['profiler_on'])
    self.assertTrue('method_to_profile' in response['target_function'])

    response = self.cprofile_post(function_test).json()
    self.assertTrue(response['profiler_on'])
    self.assertTrue('method_to_profile' in response['target_function'])

    response = self.cprofile_get(function_test).json()
    self.assertTrue(response['profiler_on'])

    self.assertEqual(LOG.info.call_count, 0)
    ClassToProfile().method_to_profile()
    # we issue 3 logs:
    self.assertEqual(LOG.info.call_count, 3)
    log_message = LOG.info.call_args_list[2].args[0]
    self.assertTrue('Ordered by: cumulative time' in log_message)

    # switch profiling off by POSTing again:
    response = self.cprofile_post(function_test).json()
    self.assertFalse(response['profiler_on'])

    LOG.info.reset_mock()
    # invoke the method again
    ClassToProfile().method_to_profile()
    # no logs this time:
    self.assertEqual(LOG.info.call_count, 0)
    self.assertEqual(method_body.call_count, 2)

  @mock.patch('adminapi.views_cprofile.LOG')
  def test_profiling_options(self, LOG: mock.Mock):

    function_test = {
        'module': __name__,
        'func': function_to_profile.__name__,
        'strip_dirnames': 'no',
        'print_callees': 'yes',
        'print_callers': 'yes'
    }

    response = self.cprofile_post(function_test).json()
    self.assertTrue(response['profiler_on'])
    self.assertTrue('function_to_profile' in response['target_function'])

    self.assertEqual(LOG.info.call_count, 0)
    function_to_profile()
    # we issue 3 logs:
    self.assertEqual(LOG.info.call_count, 3)

    log_message = LOG.info.call_args_list[2].args[0]
    self.assertTrue('Ordered by: cumulative time' in log_message)
    # dirnames:
    self.assertTrue(
      'arcweb/adminapi/tests/tests_views_cprofile.py' in log_message)
    # callees:
    self.assertTrue('called...' in log_message)
    # callees:
    self.assertTrue('called...' in log_message)
    # callers:
    self.assertTrue('was called by...' in log_message)

    # switch profiling off by POSTing again:
    response = self.cprofile_post(function_test).json()
    self.assertFalse(response['profiler_on'])

    LOG.info.reset_mock()
    # invoke the function again
    function_to_profile()
    # no logs this time:
    self.assertEqual(LOG.info.call_count, 0)
