# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Provide a standard web interface for admin access to Arcadia elements.

Format: .../adminapi/<object type>[/<id>][?detail={0,1}]

Http method indicates operation:

  GET - Fetch one or all objects
  DELETE - Delete one object
  POST - Update/Create one object

Returned data content for GET/POST is always a JSON formatted list of objects.

Input format for POST is a JSON struct containing a list of objects.  Currently updates
must be done on a single object so the list must be exactly one element long.

All objects have an 'id' field -- a create operation is done using a POST request
where the id field is not present.  The returned data will contain the object as
created, including the assigned id field.

Note that not all object fields can be updated using this interface--for example
the dataset object contains a 'created' timestamp which is managed by the system
and not changeable in this interface.

In the Django mode, some fields are stored as JSON strings.  This interface hides that
from the client.

This interface purposely layers on top of the same interfaces that our arcviz UI uses
to access the back end objects.  This ensure the same validation and permission
checking is performed as in the UI usage.

"""
import json
import gc
import logging
import codecs
from functools import wraps
import asyncio
import urllib.parse


from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse, QueryDict, HttpRequest, JsonResponse
from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.auth.decorators import login_required
from django.contrib import auth
from django.shortcuts import render
from django.urls import reverse

from vizlib.misc import check_login_or_use_apikey
from arcwebbase import gcmonitor
from reports import views_settings
from reports.views import report_internal, create_internal
from vizlib.dataset import column_is_measure
from datasets.views_connections import dataconnection_internal
from datasets.views import segment_internal, filterassociation_internal, _auditDataset, _dataset_create_internal, ActionType, dataset_internal, dataset_get_internal
from dataconn.dataconn import get_dataconn
from apps.views import users_internal, groups_internal
from rbac.permcache import RBAC
from rbac.views import roles_internal
from workspace.views import workspace_internal
from reports.models import Report
from datasets.models import DataSet
from util.json_utils import json_dumps
from importlib import import_module


LOGGER = logging.getLogger("arcweb")
SessionStore = import_module(settings.SESSION_ENGINE).SessionStore

def login_or_apikey_required(view_func):
    @wraps(view_func)
    def _wrapped_view(request, *args, **kwargs):
        err = check_login_or_use_apikey(request)
        if err:
            return err  # Return the error response if authentication fails
        return view_func(request, *args, **kwargs)  # Proceed to the view if authenticated
    return _wrapped_view

# Decode a set of b64 encoded query filter options
# No additional filtering if id is set
# 'name' option is optionally translated
def decode_request_filter(request, id=None, namexlat=None):
  byname = False
  if id is not None:
    return (None, byname)
  filter = request.GET.get('filter', {})
  if filter:
    filter = json.loads(codecs.decode(bytes(filter, 'ascii'), 'base64').decode('ascii'))
  name = request.GET.get('name', None)
  if name:
    byname = True
    filter[namexlat or 'name'] = name
  return (filter, byname)

# Check if got an empty response
def empty_response(id, byname, rsp):
  return ((id or byname) and (rsp.content == b'[]'))

#
# Reformat a response to include only selected keys and also expand some strings into json
# for easier human consumption.
#
def reformat_items(input, keys, keys_json):
  items = []
  for r in input:
    newr = {}
    for key in keys:
      val = r.get(key)
      if val and key in keys_json:
        val = json.loads(val)
      newr[key] = val
    items.append(newr)
  return items

# Generate a response from reformatted list of items
def reformat_rsp(rsp, keys, keys_json):
  if rsp.status_code != 200:
    return rsp
  items = json.loads(rsp.content)
  items = reformat_items(items, keys, keys_json)
  return HttpResponse(json.dumps(items), content_type="application/json")

# Helper function to return error response
def error(msg='Error'):
  return HttpResponse(msg, status=500)

# Ensure ID in data matches that in the URL
def checkid(data, id):
  if id is not None:
    if 'id' in data and str(data['id']) != str(id):
      return 'ID in data does not match URL ID'
    data['id'] = int(id)
  else:
    if 'id' in data:
      return 'ID in data not specified in URL'

  return None

# Merge input data with existing structure
# Plus handle conversion to json
def merge_data(keys, keys_json, old_data, new_data):
  for key in keys:
    val = new_data.get(key)
    if val is not None:
      if key in keys_json and val != '':
        val = json.dumps(val)
      old_data[key] = val

#
# Support an api for app management
#
@csrf_exempt
def aa_visuals(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  # Return only api supported fields
  keys =  ['id', 'title', 'type', 'description',
           'created', 'created_by', 'updated', 'updated_by',
           'dataset_id', 'dataset_name', 'dataconn_name', 'workspace_id', 'workspace_name',
           'perm', 'shared_visual_dashboards']
  keys_detail = ['url', 'img', 'baseImg', 'data']
  keys_json = ['data', 'shared_visual_dashboards']

  if request.method == 'DELETE':
    if id is None:
      return error('Visual ID not specified')
    return report_internal(request, id)

  elif request.method == 'GET':
    detail = request.GET.get('detail')
    filter, byname = decode_request_filter(request, id, 'report_name')
    rsp = report_internal(request, id, filter=filter)
    if rsp.status_code != 200:
      return rsp

    if empty_response(id, byname, rsp):
      return error('Visual ID %s not found' % id)

    if request.GET.get('detail'):
      keys.extend(keys_detail)
    return reformat_rsp(rsp, keys, keys_json)

  elif request.method == 'POST':
    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = []
    if len(data) != 1:
      return error('Invalid data format')

    new_data = data[0]
    keys.extend(keys_detail)

    if not id:
      req = HttpRequest()
      req.user = request.user
      req.method = 'POST'
      req.POST = QueryDict('')

      rsp = create_internal(req, new_data.get('dataset_id'))
      if rsp.status_code != 200:
        return rsp

      id = json.loads(rsp.content)[0].get('id')
      new_data['id'] = id
      old_data = {'title':'', 'type':''}

    else:
      msg = checkid(new_data, id)
      if msg:
        return error(msg)

      req = HttpRequest()
      req.user = request.user
      req.method = 'GET'
      rsp = report_internal(req, id)
      if rsp.status_code != 200:
        return error('Unable to fetch existing visual state')
      old_data = json.loads(rsp.content)[0]

    # report title needs to be updated via report data.
    # it also take precedence over any value in the incomming report data
    dashboard = old_data['type'] == 'dashboard'
    if 'title' in new_data and new_data['title'] != old_data['title']:
      if 'data' in new_data:
        if dashboard:
          new_data['data']['report_title'] = new_data['title']
        elif 'report_data' in new_data['data']:
          new_data['data']['report_data']['report_title'] = new_data['title']
      else:
        temp = json.loads(old_data['data'])
        if dashboard:
          temp['report_title'] = new_data['title']
        else:
          temp['report_data']['report_title'] = new_data['title']
        old_data['data'] = json.dumps(temp)

    # Now merge in updated fields with existing data
    merge_data(keys, keys_json, old_data, new_data)

    # Handle some renaming requirements for update
    for key in ['type', 'description', 'data']:
      old_data['report_' + key] = old_data.pop(key)

    request.POST = request.POST.copy()
    request.POST.update(old_data)
    rsp = report_internal(request, id)
    if rsp.status_code != 200:
      return rsp

    req = HttpRequest()
    req.user = request.user
    req.method = 'GET'
    rsp = report_internal(req, id)
    if rsp.status_code != 200:
      return rsp
    return reformat_rsp(rsp, keys, keys_json)

  else:
    return HttpResponse('Invalid request type', status=401)

#
# Dataset admin api
#
@csrf_exempt
def aa_datasets(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  keys = ['id', 'name', 'description', 'created', 'created_by', 'type', 'detail', 'dc_id', 'dc_name', 'lvname']
  keys_detail = ['perms', 'settings', 'info', 'lvcolumns']
  keys_json = ['settings', 'info']

  if request.method == 'DELETE':
    if id is None:
      return error('Dataset ID not specified')
    return dataset_internal(request, id)

  elif request.method == 'GET':
    filter, byname = decode_request_filter(request, id, 'dataset_name')
    rsp = dataset_internal(request, id, filter=filter)
    if rsp.status_code != 200:
      return rsp

    if request.GET.get('detail'):
      keys.extend(keys_detail)
      # if data_page=true then dataset_get_internal won't return dataset_info,
      # remove info from the result because it's value is null
      if request.GET.get('data_page', None) == 'true':
        keys.remove('info')

    if empty_response(id, byname, rsp):
      return error('Dataset not found')

    return reformat_rsp(rsp, keys, keys_json)

  elif request.method == 'POST':
    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = []
    if len(data) != 1:
      return error('Invalid data format')

    new_data = data[0]
    keys.extend(keys_detail)

    if not id:
      old_data = {}

    else:
      msg = checkid(new_data, id)
      if msg:
        return error(msg)

      req = HttpRequest()
      req.user = request.user
      req.method = 'GET'
      rsp = dataset_internal(req, id)
      if rsp.status_code != 200:
        return error('Unable to fetch existing dataset state')
      old_data = json.loads(rsp.content)[0]

    merge_data(keys, keys_json, old_data, new_data)

    # This is a little funky in that the internal api dataset get and put formats are so different
    data = {}
    data['dataconnection_id'] = old_data.get('dc_id')
    data['description'] = old_data.get('description')
    # fixup field names that need "dataset_" for the put action
    for f in ['name', 'type', 'info', 'detail', 'lvname', 'settings']:
      data['dataset_' + f] = old_data.get(f)

    if id:
      request.method = 'PUT'
      rsp = dataset_internal(request, dataset_id=id, input=data)
    else:
      request.POST = QueryDict('', mutable=True)
      request.POST.update(data)
      rsp = dataset_internal(request)
      if getattr(settings, "ADMINAPI_RETURN_EXISTING_DATASET", False) and \
          hasattr(rsp, '_duplicate_dataset'):
        # Duplicate Dataset exists so return that Dataset information
        request.method = "GET"
        id = rsp._duplicate_dataset.id
        rsp = dataset_internal(request, id, filter=None)
        if rsp.status_code != 200:
          return rsp
        keys.extend(keys_detail)
        return reformat_rsp(rsp, keys, keys_json)

    return reformat_rsp(rsp, keys, keys_json)

  else:
    return HttpResponse('Invalid request type', status=401)

#
# Dataset admin api
#
@csrf_exempt
def aa_connections(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  keys = ['id', 'name', 'type']
  keys_detail = ['info', 'created_by', 'created', 'updated_by', 'updated', 'password']
  keys_json = ['info']

  if request.method == 'DELETE':
    if id is None:
      return error('Connection ID not specified')
    return dataconnection_internal(request, id)

  elif request.method == 'GET':
    filter, byname = decode_request_filter(request, id, 'dataconnection_name')
    rsp = dataconnection_internal(request, id, filter=filter)
    if rsp.status_code != 200:
      return rsp

    if request.GET.get('detail'):
      keys.extend(keys_detail)

    if empty_response(id, byname, rsp):
      return error('Connection not found')

    return reformat_rsp(rsp, keys, keys_json)

  elif request.method == 'POST':
    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = []
    if len(data) != 1:
      return error('Invalid data format')

    new_data = data[0]
    keys.extend(keys_detail)

    if not id:
      old_data = {}

    else:
      msg = checkid(new_data, id)
      if msg:
        return error(msg)

      req = HttpRequest()
      req.user = request.user
      req.method = 'GET'
      rsp = dataconnection_internal(req, id)
      if rsp.status_code != 200:
        return error('Unable to fetch existing connection state')
      old_data = json.loads(rsp.content)[0]

    merge_data(keys, keys_json, old_data, new_data)

    # Same funky difference in get/put data format for connections and datasets
    data = {}

    # fixup field names that need "dataconnection_" for the put action
    for f in ['name', 'type', 'password', 'info']:
      data['dataconnection_' + f] = old_data.get(f)

    if new_data.get('validate'):
      data['do_validate'] = 'true'

    if id:
      request.method = 'PUT'
      request.POST = None
      rsp = dataconnection_internal(request, id, input=data)
      if rsp.status_code != 200:
        return rsp
    else:
      request.POST = QueryDict('', mutable=True)
      request.POST.update(data)
      rsp = dataconnection_internal(request, id)
      if rsp.status_code != 200:
        return rsp
      id = json.loads(rsp.content).get('id', None)

    if new_data.get('validate') and not id:
      keys.remove('id')
      items = reformat_items([old_data], keys, keys_json)
      return HttpResponse(json.dumps(items), content_type="application/json")

    req = HttpRequest()
    req.user = request.user
    req.method = 'GET'
    rsp = dataconnection_internal(req, id)
    if rsp.status_code != 200:
      return rsp
    return reformat_rsp(rsp, keys, keys_json)

  else:
    return HttpResponse('Invalid request type', status=401)

#
# Segment admin api
#
@csrf_exempt
def aa_segments(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  keys = ['id', 'dataset_id', 'name']
  keys_detail = ['created', 'created_by', 'updated', 'updated_by', 'data']
  keys_json = ['data']

  # Kind of a hack to simplify data element inside the segment
  def filter_data(rsp, keys, keys_json):
    items = json.loads(rsp.content)
    if 'data' in keys:
      for i in items:
        data = json.loads(i.get('data'))
        for key in ['id', 'name', 'dataset_id']:
          if key in data:
            del data[key]
        i['data'] = json.dumps(data)
    items = reformat_items(items, keys, keys_json)
    return HttpResponse(json.dumps(items), content_type="application/json")

  if request.method == 'DELETE':
    if id is None:
      return error('Segment ID not specified')
    return segment_internal(request, id)

  elif request.method == 'GET':
    filter, byname = decode_request_filter(request, id, 'segment_name')
    rsp = segment_internal(request, id, filter=filter)
    if rsp.status_code != 200:
      return rsp

    if request.GET.get('detail'):
      keys.extend(keys_detail)

    if empty_response(id, byname, rsp):
      return error('Segment not found')

    return filter_data(rsp, keys, keys_json)

  elif request.method == 'POST':
    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = []
    if len(data) != 1:
      return error('Invalid data format')

    new_data = data[0]
    keys.extend(keys_detail)

    if not id:
      old_data = {}

    else:
      msg = checkid(new_data, id)
      if msg:
        return error(msg)

      req = HttpRequest()
      req.user = request.user
      req.method = 'GET'
      rsp = segment_internal(req, id)
      if rsp.status_code != 200:
        return error('Unable to fetch existing segment state')
      old_data = json.loads(rsp.content)[0]

    merge_data(keys, keys_json, old_data, new_data)

    # Segment name is stored both in the object and in the 'data' json element
    # but we hide that from the user
    dj = json.loads(old_data.get('data'))
    dj['name'] = old_data.get('name')
    old_data['data'] = json.dumps(dj)

    if id:
      request.method = 'PUT'
      rsp = segment_internal(request, segment_id=id, input=old_data)
    else:
      request.POST = QueryDict('', mutable=True)
      request.POST.update(old_data)
      rsp = segment_internal(request)
    return filter_data(rsp, keys, keys_json)

  else:
    return HttpResponse('Invalid request type', status=401)

#
# FilterAssociation admin api
#
@csrf_exempt
def aa_filterassociations(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  keys = ['id', 'dataset_id', 'name']
  keys_detail = ['created', 'created_by', 'updated', 'updated_by', 'data', 'users', 'groups']
  keys_json = ['data']

  if request.method == 'DELETE':
    if id is None:
      return error('Filter Association ID not specified')
    return filterassociation_internal(request, id)

  elif request.method == 'GET':
    filter, byname = decode_request_filter(request, id, 'filterassociation_name')
    detail = request.GET.get('detail')
    rsp = filterassociation_internal(request, id, detail=detail, filter=filter)
    if rsp.status_code != 200:
      return rsp

    if detail:
      keys.extend(keys_detail)

    if empty_response(id, byname, rsp):
      return error('Filter Association not found')

    return reformat_rsp(rsp, keys, keys_json)

  elif request.method == 'POST':
    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = []
    if len(data) != 1:
      return error('Invalid data format')

    new_data = data[0]
    keys.extend(keys_detail)

    if not id:
      old_data = {}

    else:
      msg = checkid(new_data, id)
      if msg:
        return error(msg)

      req = HttpRequest()
      req.user = request.user
      req.method = 'GET'
      rsp = filterassociation_internal(req, id)
      if rsp.status_code != 200:
        return error('Unable to fetch existing filter association state')
      old_data = json.loads(rsp.content)[0]

    merge_data(keys, keys_json, old_data, new_data)

    data = {}
    for key in ['data', 'dataset_id', 'name', 'groups', 'users']:
      val = old_data.get(key)
      if val is not None:
        if key in ['users', 'groups']:
          val = json.dumps(val)
        data[key] = val

    if id:
      request.method = 'PUT'
      rsp = filterassociation_internal(request, filterassociation_id=id, input=data)
    else:
      request.POST = QueryDict('', mutable=True)
      request.POST.update(data)
      rsp = filterassociation_internal(request)
    return reformat_rsp(rsp, keys, keys_json)

  else:
    return HttpResponse('Invalid request type', status=401)

#
# Users admin api
#
@csrf_exempt
def aa_users(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  # Return only api supported fields
  keys = ['id', 'username', 'is_superuser']
  keys_detail = ['is_active', 'date_joined', 'last_login', 'first_name', 'last_name',
                 'groups', 'roles', 'profile']
  keys_json = ['profile']

  # Hack here to convert ID into username for internal access
  urlid = id
  if id is not None:
    try:
      user = User.objects.get(id=int(id))
    except User.DoesNotExist:
      return error('User not found')
    id = user.username

  if request.method == 'DELETE':
    if id is None:
      return error('User ID not specified')
    return users_internal(request, id)

  elif request.method == 'GET':
    request.GET = request.GET.copy()
    if request.GET.get('detail'):
      keys.extend(keys_detail)
      request.GET.update({'summary': False})
    else:
      request.GET.update({'summary': True})

    filter, byname = decode_request_filter(request, id, 'username')
    rsp = users_internal(request, id, filter=filter)
    if rsp.status_code != 200:
      return rsp

    if empty_response(urlid, byname, rsp):
      return error('User not found')

    return reformat_rsp(rsp, keys, keys_json)

  elif request.method == 'POST':

    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = []
    if len(data) != 1:
      return error('Invalid data format')

    # The create/update format is different than the GET format
    new_data = data[0]
    keys.extend(keys_detail)

    parms = {}
    if not id:
      if 'password' in new_data:
        parms['password'] = new_data['password']

    else:
      msg = checkid(new_data, urlid)
      if msg:
        return error(msg)

      if 'password' in new_data:
        parms['new_password'] = new_data['password']
      if 'old_password' in new_data:
        parms['old_password'] = new_data['old_password']

    for key in ['id', 'is_active', 'is_superuser', 'username', 'first_name', 'last_name']:
      if key in new_data:
        if key in ['is_active', 'is_superuser']:
          parms[key] = "false" if not new_data[key] else "true"
        else:
          parms[key] = new_data[key]

    for key in ['groups', 'roles']:
      if key in new_data:
        parms[key] = json.dumps([d.get('id') for d in new_data[key]])

    for key in ['profile']:
      if key in new_data:
        parms[key] = json.dumps(new_data[key])

    if not id:
      request.POST = QueryDict('', mutable=True)
      request.POST.update(parms)
      rsp = users_internal(request, new_data.get('username'))
    else:
      request.method = 'PUT'
      rsp = users_internal(request, id, input=parms)

    if rsp.status_code != 200:
      return rsp

    # Return full user info after update/create
    id = json.loads(rsp.content)[0].get('username')
    request.GET = QueryDict('', mutable=True)
    request.GET.update({'summary': False})
    request.method = 'GET'
    rsp = users_internal(request, id)
    if rsp.status_code != 200:
      return rsp
    return reformat_rsp(rsp, keys, keys_json)

  else:
    return HttpResponse('Invalid request type', status=401)

#
# Groups admin api
#
@csrf_exempt
def aa_groups(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  # Return only api supported fields
  keys = ['id', 'name']
  keys_detail = ['users', 'roles']
  keys_json = []

  # Kind of a hack to simplify user info returned inside the group data
  def filter_users(rsp, keys, keys_json):
    items = json.loads(rsp.content)
    for i in items:
      for u in i.get('users', []):
        for key in list(u.keys()):
          if key not in ['id', 'username']:
            del u[key]
    items = reformat_items(items, keys, keys_json)
    return HttpResponse(json.dumps(items), content_type="application/json")

  if request.method == 'DELETE':
    if id is None:
      return error('Group ID not specified')
    return groups_internal(request, id)

  elif request.method == 'GET':
    filter, byname = decode_request_filter(request, id)
    rsp = groups_internal(request, id, filter=filter)
    if rsp.status_code != 200:
      return rsp
    if request.GET.get('detail'):
      keys.extend(keys_detail)
    if empty_response(id, byname, rsp):
      return error('Group not found')
    return filter_users(rsp, keys, keys_json)

  elif request.method == 'POST':

    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = []
    if len(data) != 1:
      return error('Invalid data format')

    # The create/update format is different than the GET format
    new_data = data[0]
    keys.extend(keys_detail)

    parms = {}
    if not id:
      pass

    else:
      msg = checkid(new_data, id)
      if msg:
        return error(msg)

    for key in ['id', 'name']:
      if key in new_data:
        parms[key] = new_data[key]

    for key in ['users', 'roles']:
      if key in new_data:
        parms[key] = json.dumps([d.get('id') for d in new_data[key]])

    if not id:
      request.POST = QueryDict('', mutable=True)
      request.POST.update(parms)
      rsp = groups_internal(request, new_data.get('username'))
    else:
      request.method = 'PUT'
      rsp = groups_internal(request, id, input=parms)

    if rsp.status_code != 200:
      return rsp

    # Return full info after update/create
    id = json.loads(rsp.content)[0].get('id')
    request.GET = QueryDict('', mutable=True)
    request.method = 'GET'
    rsp = groups_internal(request, id)
    if rsp.status_code != 200:
      return rsp
    return filter_users(rsp, keys, keys_json)

  else:
    return HttpResponse('Invalid request type', status=401)


#
# Roles admin api
#
@csrf_exempt
def aa_roles(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  # Return only api supported fields
  keys = ['id', 'name', 'desc', 'require_all_groups', 'perms', 'users', 'groups']
  keys_detail = ['privs']
  keys_json = []

  if request.method == 'DELETE':
    if id is None:
      return error('Role ID not specified')
    return roles_internal(request, id)

  elif request.method == 'GET':
    filter, byname = decode_request_filter(request, id)
    rsp = roles_internal(request, id, filter=filter)
    if rsp.status_code != 200:
      return rsp
    if empty_response(id, byname, rsp):
      return error('Role not found')
    if request.GET.get('detail'):
      keys.extend(keys_detail)
    return reformat_rsp(rsp, keys, keys_json)

  elif request.method == 'POST':

    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = []
    if len(data) != 1:
      return error('Invalid data format')

    new_data = data[0]
    keys.extend(keys_detail)

    if not id:
      old_data = {}

    else:
      msg = checkid(new_data, id)
      if msg:
        return error(msg)

      req = HttpRequest()
      req.user = request.user
      req.method = 'GET'
      rsp = roles_internal(req, id)
      if rsp.status_code != 200:
        return error('Unable to fetch existing role state')
      old_data = json.loads(rsp.content)[0]

    merge_data(keys, keys_json, old_data, new_data)

    request.POST = request.POST.copy()
    request.POST.update({'data' : json.dumps([old_data])})
    rsp = roles_internal(request, id)
    if rsp.status_code != 200:
      return rsp
    items = json.loads(rsp.content).get('data', [])
    items = reformat_items(items, keys, keys_json)
    return HttpResponse(json.dumps(items), content_type="application/json")

  else:
    return HttpResponse('Invalid request type', status=401)


#
# Workspaces admin api
#
@csrf_exempt
def aa_workspaces(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  # Return only api supported fields
  keys = ['id', 'name', 'desc']
  keys_detail = ['editable', 'private_user_id', 'acl', 'perms']
  keys_json = []

  if request.method == 'DELETE':
    if id is None:
      return error('Workspace ID not specified')
    return workspace_internal(request, id)

  elif request.method == 'GET':
    filter, byname = decode_request_filter(request, id)
    rsp = workspace_internal(request, id, filter=filter)
    if rsp.status_code != 200:
      return rsp
    if empty_response(id, byname, rsp):
      return error('Workspace not found')
    if request.GET.get('detail'):
      keys.extend(keys_detail)
    return reformat_rsp(rsp, keys, keys_json)

  elif request.method == 'POST':

    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = []
    if len(data) != 1:
      return error('Invalid data format')

    new_data = data[0]
    keys.extend(keys_detail)

    if not id:
      old_data = {}

    else:
      msg = checkid(new_data, id)
      if msg:
        return error(msg)

      req = HttpRequest()
      req.user = request.user
      req.method = 'GET'
      rsp = workspace_internal(req, id)
      if rsp.status_code != 200:
        return error('Unable to fetch existing workspace state')
      old_data = json.loads(rsp.content)[0]

    merge_data(keys, keys_json, old_data, new_data)

    request.POST = request.POST.copy()
    request.POST.update({'data' : json.dumps([old_data])})
    rsp = workspace_internal(request, id)
    if rsp.status_code != 200:
      return rsp
    items = json.loads(rsp.content).get('data', [])
    items = reformat_items(items, keys, keys_json)
    return HttpResponse(json.dumps(items), content_type="application/json")

  else:
    return HttpResponse('Invalid request type', status=401)

@csrf_exempt
def edit_userprofile(request, username=None):
  err = check_login_or_use_apikey(request)
  if err:
    return err

  if username is None:
    return error('Username not specified')

  if request.method == 'POST':
    user = User.objects.filter(username=username).first()
    if user is None:
      return HttpResponse("username not found", status=500)
    userprofile = user.userprofile

    request_data = request.POST

    userprofile.should_change_password = request_data.get('should_change_password', userprofile.should_change_password)
    userprofile.is_password_editable = request_data.get('is_password_editable', userprofile.is_password_editable )
    userprofile.jwt_token = request_data.get('jwt_token', userprofile.jwt_token )
    userprofile.user_data = request_data.get('user_data', userprofile.user_data )
    userprofile.authentication_source = request_data.get('authentication_source', userprofile.authentication_source)

    userprofile.save()
    return JsonResponse({
      'should_change_password': userprofile.should_change_password,
      'is_password_editable': userprofile.is_password_editable,
      'jwt_token': userprofile.jwt_token,
      'user_data': userprofile.user_data,
      'authentication_source': userprofile.authentication_source
      })

  else:
    return HttpResponse('Invalid request type', status=401)


@csrf_exempt
def admin_gcmonitor(request):
  err = check_login_or_use_apikey(request)
  if err:
    return err

  if request.method == 'GET':
    response = {'enabled': gcmonitor.INSTANCE.enabled()}
    return JsonResponse(response)

  RBAC.checkPerm(request.user, 'sys_viewlogs')
  body = getattr(request, 'body', '{}')
  req_body: dict = json.loads(body)
  enabled = req_body.get('enabled')
  if enabled is None:
    return JsonResponse({'error': 'enabled parameter is required'}, status=400)

  if enabled:
    gcmonitor.INSTANCE.register()
  else:
    gcmonitor.INSTANCE.unregister()

  return JsonResponse({'enabled': gcmonitor.INSTANCE.enabled()})

@csrf_exempt
def admin_gcstats(request):
  err = check_login_or_use_apikey(request)
  if err:
    return err

  if request.method == 'GET':
    response = {'debug_flags': gc.get_debug()}
    return JsonResponse(response)

  RBAC.checkPerm(request.user, 'sys_viewlogs')
  body = getattr(request, 'body', '{}')
  req_body: dict = json.loads(body)
  debug_flags_str = req_body.get('debug_flags', 'None')

  # 0 means no debuging
  debug_flags = getattr(gc, debug_flags_str, 0)
  gc.set_debug(debug_flags)

  return JsonResponse({'debug_flags': gc.get_debug()})

@csrf_exempt
def admin_loglevel(request, logger_name='root'):
  err = check_login_or_use_apikey(request)
  if err:
    return err

  if request.method == 'GET':
    return _getLoglevelResponse(logger_name)

  RBAC.checkPerm(request.user, 'sys_viewlogs')

  body = getattr(request, 'body', '{}')
  req_body = json.loads(body)
  if 'level' not in req_body:
    return _json_error('Desired level must be specified in the request body.')

  level_str = req_body['level']

  # fun fact: getLevelName returns an int if the parameter is a string matching a level name
  level_int = logging.getLevelName(level_str)
  if not type(level_int) == int:
    return _json_error(f'level {level_str} not recognized')

  logging.getLogger(logger_name).setLevel(level_int)

  return _getLoglevelResponse(logger_name)

def _json_error(message: str, status=400):
  return JsonResponse({'error': message}, status=status)

def _getLoglevelResponse(logger_name: str):
  log_level = logging.getLogger(logger_name).getEffectiveLevel()
  return JsonResponse({
    'logger': logger_name,
    'level': logging.getLevelName(log_level)
  })

@csrf_exempt
@login_or_apikey_required
def create_dataset_from_sql(request):
  connection_id = request.POST.get('connection_id')
  title = request.POST.get('title')
  sql = request.POST.get('sql')

  if connection_id == None:
    return JsonResponse({ "error": "connection_id is missing" }, status=400)

  if title == None:
    return JsonResponse({ "error": "title is missing" }, status=400)

  if sql == None:
    return JsonResponse({ "error": "sql is missing" }, status=400)

  columns = _get_columns_based_on_sql_query(sql=sql, connection_id=connection_id, username=request.user.username)

  dataset_info = json.dumps([
    {
      "tablename": "text.sql",
      "select": sql,
      "columns": columns
    }
  ])

  new_dataset = _dataset_create_internal(
    dataconnection_id=connection_id,
    dataset_name=title,
    create_user=request.user,
    user=request.user,
    dataset_info=dataset_info,
    dataset_settings="{\"HAS_TITLE_CASED_NAMES\":false}",
    dataset_type="singletable",
    dataset_lvinfo="",
    dataset_detail='Created from SQL',
    dataset_description="",
    dashboards=[],
    version_name=None,
    version_group_id=None,
    is_active_version=1,
    request=request
  )

  _auditDataset(new_dataset, ActionType.CREATE, request)

  return JsonResponse({ "id": new_dataset.id })

def _get_columns_based_on_sql_query(sql, connection_id, username):
  connection = get_dataconn(connection_id)
  connection_type = connection.connection_type

  # This is the same logis ac in arc_dataset_sql.js
  query = sql
  if connection_type == 'oracle':
    query = f"select * from ({query}) t1 where rownum <= 1"
  elif (connection_type == 'mssqlserver'):
    query = f"select top(1) * from ({query}) t1"
  elif (connection_type == 'sqlite'):
    query = f"select * from ({query}) t1 limit 1"
  else:
    query = f"select * from ({query}) t1 limit 0"

  result = connection.log_sync_runquery(query, username)
  column_names = result.get('colnames')
  column_types = result.get('coltypes')

  columns = []
  for index, column_name in enumerate(column_names):
    column_type = column_types[index]
    columns.append({
      "name": column_name,
      "type": column_type,
      "alias": column_name,
      'isdim': not column_is_measure(column_type)
    })

  return columns

@csrf_exempt
@login_or_apikey_required
def generate_visual(request):
  if request.method == 'POST':
    dataset_id = request.POST.get('dataset_id')
    columns = json.loads(request.POST.get('columns', '[]'))
    filters = json.loads(request.POST.get('filters', '[]'))
    visual_type = request.POST.get('type', None)
    visual_title = request.POST.get('title', '')

    if dataset_id == None:
      return JsonResponse({ "error": "dataset_id is missing" }, status=400)

    if len(columns) == 0:
      return JsonResponse({ "error": "columns are missing" }, status=400)

    datasets = json.loads(
      dataset_get_internal(request, dataset_id=dataset_id).content
    )

    if len(datasets) == 0:
      return JsonResponse({ "error": f"Dataset doesn't exist: id { dataset_id }" }, status=400)

    dataset_json = datasets[0]

    dataset_info = json.loads(dataset_json["info"])
    dataset_column_info_by_alias = {}
    for table in dataset_info:
      for dataset_column in table["columns"]:
        dataset_column_info_by_alias[dataset_column['alias']] = {
          "name": dataset_column["name"] or dataset_column["basecol"],
          "type": dataset_column["type"],
          "isdim": dataset_column["isdim"],
        }

    def get_shelf_by_column(column, dataset_column):
      shelf = {
        "dataset_colname": dataset_column["name"],
        "dataset_coltype": dataset_column["type"],
      }

      aggregate_function = column.get("aggregate_function")
      display_name = column.get("display_name", dataset_column["name"])

      if aggregate_function:
        shelf["custom_expr"] = f"{aggregate_function.upper()}([{dataset_column['name']}])"
      else:
        shelf["custom_expr"] = f"[{dataset_column['name']}]"

      if display_name:
        shelf["custom_expr"] += f' as "{display_name}"'

      if column.get('limit'):
        shelf["order_limit"] = column["limit"]

      if column.get('order'):
        shelf["order"] = {
          "ascending": column["order"] == 'asc'
        }

      return shelf

    aggregates = []
    dimensions = []
    for column in columns:
      dataset_column = dataset_column_info_by_alias[column["column_name"]]
      shelf = get_shelf_by_column(column, dataset_column)
      is_dimension = dataset_column["isdim"]
      aggregate_function = column.get("aggregate_function")

      if not is_dimension or aggregate_function:
        aggregates.append(shelf)
      else:
        dimensions.append(shelf)

    shelf_data = {}
    if visual_type:
      shelf_data["type"] = visual_type

    shelf_data["aggregates"] = aggregates
    shelf_data["dimensions"] = dimensions

    def get_shelf_by_filter(filter, dataset_column):
      shelf = {
        "dataset_colname": dataset_column["name"],
        "dataset_coltype": dataset_column["type"],
      }

      if filter["type"].upper() == 'IN':
        shelf["custom_expr"] = f"[{dataset_column['name']}] in ({filter['value']})"
      elif filter["type"] in ("<", "<=", ">", ">="):
        shelf["custom_expr"] = f"[{dataset_column['name']}] {filter['type']} {filter['value']}"
      else:
        raise Exception(f'Filter type is not supported: {filter["type"]}')

      return shelf

    filter_shelves = []
    for filter in filters:
      dataset_column = dataset_column_info_by_alias[filter["column_name"]]
      shelf = get_shelf_by_filter(filter, dataset_column)
      filter_shelves.append(shelf)
    shelf_data["filters"] = filter_shelves

    session = SessionStore()
    session.set_expiry(settings.SNAPSHOT_TIMEOUT)
    session[auth.BACKEND_SESSION_KEY] = \
        'django.contrib.auth.backends.ModelBackend'
    session[auth.SESSION_KEY] = request.user.id
    session[auth.HASH_SESSION_KEY] = request.user.get_session_auth_hash()
    session['CSRF_COOKIE'] = request.user.get_session_auth_hash()
    session.create()

    ip = views_settings.setting_get_single("site_thumbnails_ip")
    if not ip:
      # TODO: different error message
      raise Exception('Thumbnails IP is not set in Site Settings')

    domain = urllib.parse.urlparse(ip).hostname

    try:
      report_json = asyncio.run(
        generate_visual_via_ui(
          shelf_data,
          dataset_json,
          ip,
          domain,
          session
        )
      )
    except Exception as error:
      return JsonResponse({ "error": str(error) }, status=500)

    dataset = DataSet.objects.get(id=dataset_id)

    report_json["report_data"]['report_title'] = visual_title

    report = Report(
        report_data=json_dumps(report_json),
        report_type=report_json.get('report_type'),
        dataset=dataset,
        shared_visual_dashboards='[]'
      )

    report.save()

    return JsonResponse({ "id": report.id })
  else:
    return JsonResponse({'error': 'Only POST request is accepted'}, status=400)

async def generate_visual_via_ui(shelf_data, dataset_json, ip, domain, session):
      import pyppeteer
      launch_options = {
        'args': [
          '--no-sandbox',
          '--disable-setuid-sandbox',
          '--disable-gpu',
          '--hide-scrollbars',
          '--disable-web-security',
          '--disable-dev-shm-usage',
        ],
        'headless': True,
        'handleSIGINT': False,
        'handleSIGTERM': False,
        'handleSIGHUP': False,
      }

      PDF_OPTIONS = getattr(settings, 'PDF_OPTIONS', {})
      CHROMIUM_PATH = PDF_OPTIONS.get('CHROMIUM_PATH', None)
      if CHROMIUM_PATH != "*":
        launch_options['executablePath'] = CHROMIUM_PATH

      browser = await pyppeteer.launch(launch_options)
      page = await browser.newPage()

      @page.on('console')
      def patcher(msg):
        level = logging.ERROR if msg.type == 'error' else logging.DEBUG

        def get_info(arg):
          info = ''
          if 'description' in arg._remoteObject:
            info = arg._remoteObject['description']
          elif 'value' in arg._remoteObject:
            info = arg._remoteObject['value']
          return info

        info = [get_info(arg) for arg in msg.args]
        msg_text = '\n'.join(info).strip()

        if not msg_text:
          # no useful info could be extract from the arguments
          # use the message text:
          msg_text = msg.text

        LOGGER.log(level, 'Javascript console log while generating visual:\n%s: %s' % (msg.type, msg_text))

      await page.setCookie({"name": settings.SESSION_COOKIE_NAME, "value": session.session_key, "domain": domain, "path": "/"})
      await page.setCookie({"name": settings.CSRF_COOKIE_NAME, "value": session.get('CSRF_COOKIE', None), "domain": domain, "path": "/"})

      await page.goto(f'{ip}{reverse("render_skeleton")}')
      await page.waitForSelector('#cdv-script-skeleton')

      report_json = await page.evaluate('window.arc.showme.api', shelf_data, dataset_json)

      await browser.close()

      return report_json

@login_required
def render_skeleton(request):
  return render(request, 'script_skeleton.html')
