# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from vizlib.misc import check_login_or_use_apikey
from django.http import HttpRequest, JsonResponse
from django.views.decorators.csrf import csrf_exempt
from rbac.permcache import RBAC

import cProfile
import importlib
import io
import logging
import pstats


LOG = logging.getLogger('arcweb.cprofile')


@csrf_exempt
def toggle_cprofile(request: HttpRequest):
  err = check_login_or_use_apikey(request)
  if err:
    return err

  RBAC.checkPerm(request.user, 'sys_viewlogs')

  module_name = request.GET.get('module', 'sqlrun.views_jsonselect')
  owner_name = request.GET.get('owner')
  function_name = request.GET.get('func', 'jsonselect_parallel')
  strip_dirnames = request.GET.get('strip_dirnames', 'yes') == 'yes'
  print_callees = request.GET.get('print_callees', 'no') == 'yes'
  print_callers = request.GET.get('print_callers', 'no') == 'yes'

  module = importlib.import_module(module_name)
  if owner_name:
    owner = getattr(module, owner_name)
  else:
    owner = module

  function = getattr(owner, function_name)

  already_profiled = hasattr(function, 'restore_profile')
  if request.method == 'POST' and already_profiled:
    function.restore_profile()
  elif request.method == 'POST' and not already_profiled:
    wrap_profile(owner, function_name, strip_dirnames,
                 print_callees, print_callers)

  target_function_str = f'{str(owner)}.{function_name}'
  if module != owner:
    target_function_str = f'{str(module)}.{target_function_str}'

  response = {
    'profiler_on': hasattr(getattr(owner, function_name), 'restore_profile'),
    'target_function': target_function_str
  }
  return JsonResponse(response)


def wrap_profile(owner: object, function_name: str, strip_dirnames=True, print_callees=False, print_callers=False):
  original_function = getattr(owner, function_name)

  def profiler_func(*args, **kwargs):
    info_str = f'RUNNING PROFILE for {str(owner)}.{function_name}'
    LOG.info(info_str)

    with cProfile.Profile() as pr:
      __return_value = original_function(*args, **kwargs)

    buffer = io.StringIO()
    stats = pstats.Stats(pr, stream=buffer)

    if strip_dirnames:
      stats = stats.strip_dirs()  # this is the default behavior of cProfile.run()

    stats = stats.sort_stats(pstats.SortKey.CUMULATIVE)
    stats.print_stats()

    if print_callees:
      stats.print_callees()
    if print_callers:
      stats.print_callers()

    LOG.info(f'DONE {info_str}')
    LOG.info(buffer.getvalue())

    return __return_value

  def restore_profile():
    setattr(owner, function_name, original_function)
  profiler_func.restore_profile = restore_profile

  setattr(owner, function_name, profiler_func)
