# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import os
import string

product_name = "Data Discovery and Visualization" if os.getenv('CDV_SUB_MODE') == 'MLExplore' else "Cloudera Data Visualization"

documentation_protocol = 'https://'
documentation_base_url = "docs.cloudera.com/data-visualization/8"

release_notes_url = f'{documentation_protocol}{documentation_base_url}/release-notes/topics/viz-pc-whats-new.html'

default_documentation_url = f'{documentation_protocol}{documentation_base_url}/index.html'
default_documentation_label = 'Documentation'

default_support_web_page_label = 'Cloudera Support Hub'
default_support_web_page_link = 'https://my.cloudera.com/support.html'
default_support_email = 'support@cloudera.com'

# Shared code template
shared_code_template = string.Template("""
# This function creates and returns the request and response transformers
def create_transformers():
    def request_transformer(headers: dict, payload: dict):
        # Build new headers using the SECRET from the input field
        new_headers = {
            'api-key': headers['secret'],
            'Content-Type': 'application/json',
        }

        # Extract the input data from the payload
        input_data = payload['input']

        # Build the new body with the input data
        new_body = {
            $request_body_content
        }

        # Return the new headers and body
        return new_headers, new_body

    def response_transformer(body: dict):
        # Transform the response body from the service
        new_body = {
            $response_body_content
        }
        return new_body

    return request_transformer, response_transformer

# Assign the request and response transformer functions
[request_transformer, response_transformer] = create_transformers()

REQUEST_TRANSFORMER = request_transformer
RESPONSE_TRANSFORMER = response_transformer
""")

# Embeddings-specific request and response content
embeddings_request_body_content = "'input': input_data"
embeddings_response_body_content = "'data': [{'embedding': body['embedding']}]"

default_embeddings_engine_settings = shared_code_template.substitute(
    request_body_content=embeddings_request_body_content,
    response_body_content=embeddings_response_body_content
)

# Completions-specific request and response content
completions_request_body_content = "'messages': input_data"
completions_response_body_content = "'choices': [{'message': {'content': body['outputText']}}]"

default_completion_engine_settings = shared_code_template.substitute(
    request_body_content=completions_request_body_content,
    response_body_content=completions_response_body_content
)

default_bedrock_embedding_transformers = """
# This function creates and returns the request and response transformers
def create_transformers():
    def request_transformer(headers: dict, body: dict):
        return headers, {
            'someKey': body['inputText']
        }

    def response_transformer(body: dict):
        return {
            'data': [body]
        }

    return request_transformer, response_transformer

# Assign the request and response transformer functions
[request_transformer, response_transformer] = create_transformers()

REQUEST_TRANSFORMER = request_transformer
RESPONSE_TRANSFORMER = response_transformer
"""

default_bedrock_completion_transformers = """
def request_transformer(headers: dict, body: dict):
    messages = body.get("inputText", [])
    system_prompt = ""
    filtered_messages = []

    # Remove system messages and save their content.
    for m in messages:
        if m["role"] == "system":
            system_prompt += (m["content"] + "\\n")
        else:
            filtered_messages.append({
                "role": m["role"],
                "content": [{"type": "text", "text": m["content"]}]
            })

    # Ensure the first message is from a user.
    # If the very first message is not from a user, move the first user message to the beginning.
    first_user_index = next((i for i, msg in enumerate(filtered_messages) if msg["role"] == "user"), None)
    if first_user_index is None:
        raise ValueError("No user message found; the first message must be from a user.")
    if first_user_index != 0:
        # Remove messages before the first user message.
        filtered_messages = filtered_messages[first_user_index:]

    new_body = {
        "max_tokens": 1024,
        "anthropic_version": "bedrock-2023-05-31",
        "messages": filtered_messages
    }
    if system_prompt.strip():
        new_body["system"] = system_prompt.strip()

    return headers, new_body

def response_transformer(body: dict):
    response = 'NO RESPONSE'
    results = body.get('content')
    if results:
        response = results[0]['text']
    return {
        'choices': [
            {
            'message': {
                'content': response
            }
            }
        ]
    }

def get_anthropic_chunk_text (chunk_data):
    return chunk_data.get('delta', {}).get('text', '')

REQUEST_TRANSFORMER = request_transformer
RESPONSE_TRANSFORMER = response_transformer
CUSTOM_GET_CHUNK_TEXT = get_anthropic_chunk_text
"""
