#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import sys
import os
import signal
import time
import errno
from django.core.management.base import BaseCommand
from django.conf import settings
from arcwebbase.initialize import arcviz_initialize
import logging

import ssl
from cheroot.ssl.builtin import BuiltinSSLAdapter

LOGGER = logging.getLogger('arcweb')

CPSERVER_HELP = r"""
  Run this project in a CherryPy webserver. To do this, CherryPy from
  http://www.cherrypy.org/ is required.

   runcpserver [options] [cpserver settings] [stop]

Optional CherryPy server settings: (setting=value)
  host=HOSTNAME         hostname to listen on
                        Defaults to localhost
  port=PORTNUM          port to listen on
                        Defaults to 8088
  server_name=STRING    CherryPy's SERVER_NAME environ entry
                        Defaults to localhost
  daemonize=0|1         whether to detach from terminal
                        Defaults to 0 (False)
  pidfile=FILE          write the spawned process-id to this file
  workdir=DIRECTORY     change to this directory when daemonizing
  minthreads=NUMBER     Number of threads for server to start with
  maxthreads=NUMBER     Max mumber of threads for server to use
  ssl_certificate=FILE  SSL certificate file
  ssl_private_key=FILE  SSL private key file
  ssl_certificate_chain=FILE  SSL CA certificate chain file
  ssl_private_key_pass=STIRNG  SSL private key file passphrase
  server_user=STRING    user to run daemonized process
                        Defaults to www-data
  server_group=STRING   group to daemonized process
                        Defaults to www-data
  shutdown_timeout=NUMBER      grace period for allowing current requests to be processed (default 60)
  request_queue_size=NUMBER    maximum number of queued connections (default 5)

Examples:
  Run a "standard" CherryPy server
    $ manage.py runcpserver

  Run a CherryPy server on port 80
    $ manage.py runcpserver port=80

  Run a CherryPy server as a daemon and write the spawned PID in a file
    $ manage.py runcpserver daemonize=true pidfile=/var/run/django-cpserver.pid

"""

CPSERVER_OPTIONS = {
  'host': 'localhost',
  'port': 8000,
  'server_name': 'localhost',
  'minthreads': 10,
  'maxthreads': 30,
  'daemonize': 0,
  'workdir': None,
  'pidfile': None,
  'server_user': 'www-data',
  'server_group': 'www-data',
  'ssl_certificate': None,
  'ssl_certificate_chain': None,
  'ssl_private_key': None,
  'ssl_private_key_pass': None,
  'verbose': 1,
  'shutdown_timeout': 60,
  'request_queue_size': 5,
}

SERVER = None

def debug(msg: str):
  """
  Write to stderr that is unbuffered, so we can see the messages immediattely
  """
  sys.stderr.write(f'{msg}\n')

# Define a custom BuiltinSSLAdapter class to load the certchain with key password
class BuiltinSSLAdapterPassword(BuiltinSSLAdapter):
  def __init__(
    self, certificate, private_key, certificate_chain=None,
        ciphers=None, password=None):
    """Set up context in addition to base class properties if available."""
    if ssl is None:
      raise ImportError('You must install the ssl module to use HTTPS.')

    # CherryPy does not support password-protected keys: https://github.com/cherrypy/cherrypy/issues/1583
    # Workaround: override SSLContext.load_cert_chain so that it uses the provided password
    PASSWORD = password

    context = ssl.create_default_context(
        purpose=ssl.Purpose.CLIENT_AUTH,
        cafile=certificate_chain
    )

    ori_load_cert_chain = context.__class__.load_cert_chain
    debug(f'original load_cert_chain: {ori_load_cert_chain}')
    owns_load_cert_chain = 'load_cert_chain' in dir(context.__class__)

    def load_cert_chain(self, certificate, private_key, password=None):
      debug('running overridden load_cert_chain that uses PASSWORD')
      return ori_load_cert_chain(self, certificate, private_key, PASSWORD)

    try:
      context.__class__.load_cert_chain = load_cert_chain
      super(BuiltinSSLAdapterPassword, self).__init__(
        certificate, private_key, certificate_chain, ciphers)
    finally:
      if owns_load_cert_chain:
        debug('restoring original load_cert_chain')
        context.__class__.load_cert_chain = ori_load_cert_chain
      else:
        debug('removing load_cert_chain override')
        # the method is inherited, so just remove the override:
        del context.__class__.load_cert_chain

      debug(f'load_cert_chain after restoring: {context.__class__.load_cert_chain}')

    super(BuiltinSSLAdapter, self).__init__(
      certificate, private_key, certificate_chain, ciphers)
    self.context = ssl.create_default_context(
        purpose=ssl.Purpose.CLIENT_AUTH,
        cafile=certificate_chain
    )
    self.context.load_cert_chain(certificate, private_key, password)
    if self.ciphers is not None:
      self.context.set_ciphers(ciphers)

class Command(BaseCommand):
  help = "CherryPy Server for project. Requires Cheroot."
  args = "[various KEY=val options, use `runcpserver help` for help]"

  def add_arguments(self, parser):
    parser.add_argument(
      '--host',
      action='store',
      dest='host',
      default='127.0.0.1',
      help='IP address of server',
    )
    parser.add_argument(
      '--port',
      action='store',
      dest='port',
      default=8000,
      help='Port',
    )
    parser.add_argument(
      '--ssl_certificate',
      action='store',
      dest='ssl_certificate',
      default=None,
      help='ssl certificate',
    )
    parser.add_argument(
      '--ssl_certificate_chain',
      action='store',
      dest='ssl_certificate_chain',
      default=None,
      help='ssl certificate chain',
    )
    parser.add_argument(
      '--ssl_private_key',
      action='store',
      dest='ssl_private_key',
      default=None,
      help='ssl private key',
    )
    parser.add_argument(
      '--ssl_private_key_pass',
      action='store',
      dest='ssl_private_key_pass',
      default=None,
      help='ssl private key password',
    )
    parser.add_argument(
      '--daemonize',
      action='store_true',
      dest='daemonize',
      default=False,
      help='Run as unix daemon',
    )
    parser.add_argument(
      '--pidfile',
      action='store',
      dest='pidfile',
      default=None,
      help='pid file',
    )
    parser.add_argument(
      '--workdir',
      action='store',
      dest='workdir',
      default=None,
      help='working directory',
    )
    parser.add_argument(
      '--server_user',
      action='store',
      dest='server_user',
      default='arcadia',
      help='server user',
    )
    parser.add_argument(
      '--server_group',
      action='store',
      dest='server_group',
      default='arcadia',
      help='server group',
    )
    parser.add_argument(
      '--minthreads',
      action='store',
      dest='minthreads',
      default=10,
      help='min threads',
    )
    parser.add_argument(
      '--maxthreads',
      action='store',
      dest='maxthreads',
      default=30,
      help='max threads',
    )
    parser.add_argument(
      '--request_queue_size',
      action='store',
      dest='request_queue_size',
      default=5,
      help='request queue size',
    )
    parser.add_argument(
      '--shutdown_timeout',
      action='store',
      dest='shutdown_timeout',
      default=60,
      help='shutdown timeout',
    )
    parser.add_argument(
      '--verbose',
      action='store',
      dest='verbose',
      default=1,
      help='verbose',
    )

  def handle(self, *args, **options):
    from django.conf import settings
    from django.utils import translation
    # Activate the current language, because it won't get activated later.
    try:
      translation.activate(settings.LANGUAGE_CODE)
    except AttributeError:
      pass
    runcpserver(*args, **options)

  def usage(self, subcommand):
    return CPSERVER_HELP


def change_uid_gid(uid, gid=None):
  """Try to change UID and GID to the provided values.
  UID and GID are given as names like 'nobody' not integer.

  Src: http://mail.mems-exchange.org/durusmail/quixote-users/4940/1/
  """
  if not os.geteuid() == 0:
    # Do not try to change the gid/uid if not root.
    return
  (uid, gid) = get_uid_gid(uid, gid)
  os.setgid(gid)
  os.setuid(uid)


def get_uid_gid(uid, gid=None):
  """Try to change UID and GID to the provided values.
  UID and GID are given as names like 'nobody' not integer.

  Src: http://mail.mems-exchange.org/durusmail/quixote-users/4940/1/
  """
  import pwd
  import grp
  uid, default_grp = pwd.getpwnam(uid)[2:4]
  if gid is None:
    gid = default_grp
  else:
    try:
      gid = grp.getgrnam(gid)[2]
    except KeyError:
      gid = default_grp
  return (uid, gid)


def poll_process(pid):
  """
  Poll for process with given pid up to 10 times waiting .25 seconds in between each poll.
  Returns False if the process no longer exists otherwise, True.
  """
  for n in range(10):
    time.sleep(0.25)
    try:
      # poll the process state
      os.kill(pid, 0)
    except OSError as e:
      if e[0] == errno.ESRCH:
        # process has died
        return False
      else:
        raise  # TODO
  return True


def stop_server(pidfile):
  """
  Stop process whose pid was written to supplied pidfile.
  First try SIGTERM and if it fails, SIGKILL. If process is still running, an exception is raised.
  """
  if SERVER:
    SERVER.stop()

  if os.path.exists(pidfile):
    pid = int(open(pidfile).read())
    try:
      os.kill(pid, signal.SIGTERM)
    except OSError:  # process does not exist
      os.remove(pidfile)
      return
    if poll_process(pid):
      # process didn't exit cleanly, make one last effort to kill it
      try:
        os.kill(pid, signal.SIGKILL)
      except OSError:
        print("Process {0} did not stop.".format(pid))
        raise
    os.remove(pidfile)


def start_server(options):
  """
  Start CherryPy server
  """

  global SERVER

  message = 'starting server with options %s' % options

  print(message) # original message goes to stdout - might show up late because of buffering
  debug(message) # should appear quickly on stderr
  LOGGER.info(message) # logs are delayed because logging is configured in initialize() - but at least they go to a file

  if options['daemonize'] and options['server_user'] and \
      options['server_group']:
    # ensure the that the daemon runs as specified user
    change_uid_gid(options['server_user'], options['server_group'])

  from cheroot.wsgi import Server
  from django.core.handlers.wsgi import WSGIHandler
  SERVER = Server(
    (options['host'], int(options['port'])),
    WSGIHandler(),
    numthreads=int(options['minthreads']),
    max=int(options['maxthreads']),
    server_name=options['server_name'],
    shutdown_timeout=int(options['shutdown_timeout']),
    request_queue_size=int(options['request_queue_size'])
  )

  # Configure TLS/SSL for https
  if options['ssl_certificate'] and options['ssl_private_key']:
    import ssl

    #Create ssl_adapter with modified BuiltinSSLAdapter class which supports password for private key
    ssl_adapter = BuiltinSSLAdapterPassword(
      certificate=options['ssl_certificate'],
      private_key=options['ssl_private_key'],
      certificate_chain=options['ssl_certificate_chain'],
      password=options['ssl_private_key_pass'])

    context = ssl_adapter.context

    #Disable TLS v1.0
    context.options |= ssl.OP_NO_TLSv1

    ciphers = getattr(settings, "HTTPS_CIPHERS",
                      'ECDH+AESGCM:DH+AESGCM:ECDH+AES256:DH+AES256:ECDH+AES128:DH+AES:RSA+AESGCM:RSA+AES:!aNULL:!MD5:!DSS')
    if ciphers is not None:
      context.set_ciphers(ciphers)

    # Disable SSL only TLS is secure
    ssl_options = getattr(settings, "HTTPS_OPTIONS", None)
    if ssl_options:
      context.options = ssl_options

    SERVER.ssl_adapter = ssl_adapter

  # Initialize Arcviz before starting server and AFTER we may have daemonized
  arcviz_initialize()

  try:
    SERVER.start()
    # from django.utils import autoreload
    # autoreload.main(SERVER.start)
  except KeyboardInterrupt:
    SERVER.stop()

def runcpserver(*args, **options):

  # Get the options
  defaultoptions = CPSERVER_OPTIONS.copy()
  defaultoptions.update(**options)
  options = defaultoptions

  if "help" in args:
    print(CPSERVER_HELP)
    return

  if "stop" in args:
    stop_server(options['pidfile'])
    return True

  if options['daemonize']:
    if not options['pidfile']:
      options['pidfile'] = '/var/run/arcviz_%s.pid' % options['port']
    stop_server(options['pidfile'])

    from ._daemonize import become_daemon
    if options['workdir']:
      become_daemon(our_home_dir=options['workdir'])
    else:
      become_daemon()

    fp = open(options['pidfile'], 'w')
    fp.write("%d\n" % os.getpid())
    fp.close()

  # Start the webserver
  start_server(options)


if __name__ == '__main__':
  runcpserver(sys.argv[1:])
