# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.db import models
from django.conf import settings
from django.db.models.signals import post_save, post_delete
from django.contrib.auth.signals import user_logged_in, user_logged_out
from django.dispatch import receiver
from django.contrib.auth.models import User
from django.db import connection, transaction
from django.utils.crypto import get_random_string
from django.utils import timezone
from django.contrib.auth.hashers import make_password, check_password
from hashlib import md5

try:
  from arcwebbase.backends import PAMUser
except:
  PAMUser = None

from arcwebbase.validator import validateNoHtml, validateStringNoHtml

import datetime, json, threading


LAST_VIEWED_HISTORY_LENGTH = 15

class UserProfile(models.Model):

  # Define the top level profile keys that are accessible to the UI
  PROFILE_KEYS_VIEW = ['last_viewed']
  PROFILE_KEYS_PERM = []
  if getattr(settings, 'PROFILE_PROXY_USERNAME', False):
    PROFILE_KEYS_PERM.append('proxy_username');
  PROFILE_KEYS_EDIT = ['homeapp', 'new_release_modal_seen', 'incompatible_browser_warning_dismissed'] + PROFILE_KEYS_PERM

  def save(self, *args, **kwargs):
    validateNoHtml(self)

    super(UserProfile, self).save(*args, **kwargs)

  user = models.OneToOneField(User, on_delete=models.CASCADE)

  # json structure with user_data
  user_data = models.TextField(max_length=32*1024, default="{}",
      validators=[validateStringNoHtml])

  create_time = models.DateTimeField(auto_now_add=True, null=True,blank=True)

  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)

  AUTH_DEFAULT = "D"
  AUTH_SAML = "S"
  AUTH_LDAP = "L"
  AUTH_REMOTE = "R"

  AUTHENTICATION_SOURCES = (
    (AUTH_DEFAULT, "Default"),
    (AUTH_SAML, "SAML"),
    (AUTH_LDAP, "LDAP"),
    (AUTH_REMOTE, "Remote"),
  )

  authentication_source = models.CharField(
    max_length=1,
    choices=AUTHENTICATION_SOURCES,
    default=AUTH_DEFAULT,
  )

  is_password_editable = models.BooleanField(default=True)

  should_change_password = models.BooleanField(default=False)

  jwt_token = models.TextField(default='', blank=True)

  # keep track of last viewed reports by different users. This information
  # will be written into database periodically by bulk update thread.
  last_viewed = {}
  lock = threading.Lock()

  @classmethod
  def update_last_viewed(cls, username, report_id):
    report_id = str(report_id)
    with cls.lock:
      if username not in cls.last_viewed:
        cls.last_viewed[username] = [report_id]
      else:
        cls.last_viewed[username].append(report_id)

  @classmethod
  def commit_changes(cls):
    if cls.last_viewed == {}:
      return

    cls._commit()

  @classmethod
  @transaction.atomic
  def _commit(cls):
    def get_unique_ordered_list(report_ids):
      l = []
      for i in range(len(report_ids)-1, -1, -1):
        if report_ids[i] not in l:
          l.append(report_ids[i])
          if len(l) > LAST_VIEWED_HISTORY_LENGTH:
            break
      l.reverse()
      return l

    with cls.lock:
      last_viewed = cls.last_viewed
      cls.last_viewed = {}

    user_profiles = UserProfile.objects.filter(
            user__username__in=last_viewed.keys()).select_related('user')
    for p in user_profiles:
      username = p.user.username
      if username not in last_viewed:
        # if there is nothing to be updated for the user
        continue

      user_data = p.user_data_obj
      if 'last_viewed' not in user_data:
        user_data['last_viewed'] = last_viewed[username]
      else:
        user_data['last_viewed'] = user_data['last_viewed'] + last_viewed[username]

      user_data['last_viewed'] = get_unique_ordered_list(user_data['last_viewed'])
      p.user_data_obj = user_data
      p.save()

  @property
  def user_data_obj(self):
    return json.loads(self.user_data)

  @user_data_obj.setter
  def user_data_obj(self, data):
    self.user_data=json.dumps(data)

  # Get a profile key
  def getkey(self, key, dfl=None):
    return self.user_data_obj.get(key, dfl)

  # Set a profile key
  def setkey(self, key, val):
    user_data = self.user_data_obj
    user_data[key] = val
    self.user_data_obj = user_data
    self.save()

def conditional_decorator(condition, decorator):
    return decorator if condition else lambda x: x

@conditional_decorator(PAMUser is not None, receiver(post_save, sender=PAMUser))
@receiver(post_save, sender=User)
def user_post_save(sender, instance, **kwargs):
  if isinstance(instance, User):
    # note: in the initial syncdb the UserProfile table does not exist
    # so creating the superuser account will fail.  So skip creating a
    # UserProfile if the table does not exist.  "migrate" will create the
    # missing row.
    if UserProfile._meta.db_table in connection.introspection.table_names():
      if not UserProfile.objects.filter(user_id=instance.id).exists():
        instance.userprofile = UserProfile()
        instance.userprofile.save()

# Save last_login in userprofile to show on homepage
@receiver(user_logged_in)
def update_last_login(request, user, *args, **kwargs):
  from django.contrib.auth.models import Group

  user_data = user.userprofile.user_data_obj
  if user_data.get('last_login') is None:
    # This is the first time this user has logged in
    # Add the user to the configured default groups
    from reports.views_settings import setting_get_single
    default_groups_str = setting_get_single('default_group_membership_for_new_users')
    if default_groups_str is not None:
      default_groups = default_groups_str.split(",")
      for default_group in default_groups:
        groups = Group.objects.filter(name=default_group)
        if len(groups) == 1:
          group = groups[0]
          group.user_set.add(user.id)
          group.save()

  last_login = str(user.last_login) if user.last_login else 'N/A'
  user_data['last_login'] = user_data.get('current_login', last_login)
  user_data['current_login'] = last_login
  user.userprofile.user_data_obj = user_data
  request.session['login_time'] = user_data['current_login']
  user.userprofile.save()


@receiver(user_logged_out)
def update_session_duration(request, user, *args, **kwargs):
  login_time_str = request.session.get('login_time')

  if login_time_str and login_time_str != 'N/A':
    login_time = timezone.datetime.strptime(login_time_str, '%Y-%m-%d %H:%M:%S.%f%z')

    current_time = timezone.now()
    session_duration = current_time - login_time
    from usage.events_tracker import user_session_loader
    user_session_loader(user.id, session_duration)

# Get a profile key for a given username
def profile_getkey_for_username(username, key, dfl=None):
  try:
    user = User.objects.get(username=username)
  except User.DoesNotExist:
    return dfl
  return user.userprofile.getkey(key, dfl=dfl)

def _generate_apikey():
  return get_random_string(getattr(settings,'APIKEYSIZE', 48))

def _default_expire_time():
  return datetime.datetime.now() + datetime.timedelta(days=getattr(settings,'APIKEYDEFAULTDAYS', 365))

def lookup_apikey(raw_apikey):
  apikey_hash = md5(raw_apikey.encode()).hexdigest()
  try:
    return APIKey.objects.get(apikey_hash=apikey_hash)
  except:
    return None


class APIKey(models.Model):

  apikey = models.CharField(max_length=128 , unique=True, null=False, blank=False)
  apikey_hash = models.CharField(max_length=64 , unique=True, null=False, blank=False)
  user = models.ForeignKey(User, on_delete=models.CASCADE)
  active = models.BooleanField()
  # Expire time
  expire_time = models.DateTimeField(default=_default_expire_time, null=True, blank=True)

  # Create time
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  # Update time
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(APIKey, self).save(*args, **kwargs)

  def set_apikey(self, raw_apikey):
    self.apikey = make_password(raw_apikey)
    self.apikey_hash = md5(raw_apikey.encode()).hexdigest()

  def check_apikey(self, raw_apikey):
    return check_password(raw_apikey, self.apikey)

  def valid(self, raw_apikey):
    if not check_password(raw_apikey, self.apikey):
      return False
    if self.expire_time is None:
      return self.active
    return self.active and self.expire_time > timezone.now()

  @property
  def expire_time_str(self):
    return str(self.expire_time)

  @property
  def create_time_str(self):
    return str(self.create_time)

  @property
  def update_time_str(self):
    return str(self.update_time)

  @property
  def authenticate(self):
    if self.valid:
      return self.user
    return None

  @property
  def as_dict(self):
    return {
      'id': self.id,
      'username': self.user.username,
      'apikey': self.apikey_hash,
      'secret_apikey': None,
      'active': self.active,
      'expires': 'None' if self.expire_time is None else self.expire_time.isoformat(),
      'create_time': 'None' if self.create_time is None else self.create_time.isoformat(),
      'update_time': 'None' if self.update_time is None else self.update_time.isoformat(),
    }

  @property
  def as_json(self):
    return json.dumps(self.as_dict)

  class Meta:
    ordering = [ "user" ]

