# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Models for URL shortener
"""

from django.db import models, transaction
from django.contrib.auth.models import User
from threading import Lock
import uuid
import datetime

from reports.models import Report, AppGroup

class ShortUrl(models.Model):

  creator = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)

  create_time = models.DateTimeField(auto_now_add=True)

  update_time = models.DateTimeField(auto_now=True)

  url_data = models.JSONField()

  report = models.ForeignKey(Report, on_delete=models.SET_NULL, null=True)

  app = models.ForeignKey(AppGroup, on_delete=models.SET_NULL, null=True)

  slug = models.CharField(max_length=36, unique=True, default=uuid.uuid4, db_index=True)

  last_viewed = models.DateTimeField(null=True)

  times_followed = models.PositiveIntegerField(default=0)
  usage_count = {}
  lock = Lock()

  @classmethod
  def update_usage_count_and_last_viewed(cls, id):
    with cls.lock:
      if id in cls.usage_count:
        cls.usage_count[id]['count'] += 1
        cls.usage_count[id]['last_viewed'] = datetime.datetime.now()
      else:
        cls.usage_count[id] = {'count': 1, 'last_viewed': datetime.datetime.now()}

  @classmethod
  def commit_changes(cls):
    if cls.usage_count:
      cls._commit()

  @classmethod
  @transaction.atomic
  def _commit(cls):
    with cls.lock:
      usage_count = cls.usage_count
      cls.usage_count = {}

    short_urls_to_update = ShortUrl.objects.filter(id__in=usage_count.keys())

    for shortener in short_urls_to_update:
      try:
        count_and_last_viewed = usage_count[shortener.id]
        shortener.times_followed = shortener.times_followed + count_and_last_viewed['count']
        shortener.last_viewed = count_and_last_viewed['last_viewed']
        shortener.save()
      except ShortUrl.DoesNotExist:
        continue
  class Meta:

    ordering = ["-update_time"]

