# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from django.test import TestCase
from reports.models import Report
import json

def get_broken_dashboard_ids():
  dashboards = Report.objects.filter(
    report_type='dashboard').only('id', 'report_data')
  broken = list[int]()
  for dashboard in dashboards:
    visuals, filters = dashboard.get_visuals_in_dashboard()

    # visual_refs = set(visuals+filters)
    # filters are ignored by the pre-delete checks
    visual_refs = set(visuals)

    found_ids = set(Report.objects.filter(
      id__in=list(visual_refs)).values_list('id', flat=True))
    if visual_refs != found_ids:
      dataset_ids = set(Report.objects.filter(
        id__in=found_ids).values_list('dataset_id', flat=True))
      print(
        f'dashboard {dashboard.id} built on datasets {list(dataset_ids)} has broken references: {list(visual_refs-found_ids)}')
      broken.append(dashboard.id)

  return broken


def get_visual_id(visualWidget: dict):
  return int(visualWidget.get('id').rpartition("-")[-1])


def fix_broken_dashboards():
  dashboards = Report.objects.filter(
    report_type='dashboard').only('id', 'report_data')
  broken = list[int]()
  for dashboard in dashboards:
    visuals, filters = dashboard.get_visuals_in_dashboard()

    # visual_refs = set(visuals+filters)
    # filters are ignored by the pre-delete checks
    visual_refs = set(visuals)

    found_ids = set(Report.objects.filter(
      id__in=list(visual_refs)).values_list('id', flat=True))
    if visual_refs != found_ids:

      print(f'fixing dashboard {dashboard.id}')
      report_data = json.loads(dashboard.report_data)
      sheets = report_data.get('dashboard_sheets', [])
      for sheet in sheets:
        sheet_visuals = sheet.get('visual_widgets', [])
        existing_sheet_visuals = [
          visual for visual in sheet_visuals if get_visual_id(visual) in found_ids]

        if len(sheet_visuals) != len(existing_sheet_visuals):
          # update sheet information with the list of existing visuals
          sheet['visual_widgets'] = existing_sheet_visuals

      dashboard.report_data = json.dumps(report_data)
      dashboard.save()
      broken.append(dashboard.id)

  return broken


class TestDashboardFixing(TestCase):
  def _create_visual(self, report_type: str, report_data="{}"):
    r = Report(report_type=report_type, report_data=report_data)
    r.save()
    return r

  def _create_dashboard(self, dashboard_data: dict):
    return self._create_visual(report_type="dashboard", report_data=json.dumps(dashboard_data))

  def _visual_list(self, visual_ids: list[int]):
    return [{'id': f'uri-1-widget-{visual_id}'} for visual_id in visual_ids]

  def test_fixing(self):
    visual_1 = self._create_visual(report_type="table")

    dashboard_data = {
      'dashboard_sheets': [
        {
          'sheet_id': 1,
          'visual_widgets': self._visual_list([visual_1.id]),
          'control_widgets': []
        }
      ],
      'global_control_widgets': [],
    }

    db_1 = self._create_dashboard(dashboard_data)
    self.assertEqual(len(get_broken_dashboard_ids()), 0)

    # control widgets are not checked:
    missing_id = 12_345
    self.assertEquals(Report.objects.filter(id=missing_id).count(), 0)

    # filters are ignored by the check:
    dashboard_data['dashboard_sheets'][0]['control_widgets'].append(missing_id)
    dashboard_data['global_control_widgets'].append(missing_id)
    db_1.report_data = json.dumps(dashboard_data)
    db_1.save()
    self.assertEqual(len(get_broken_dashboard_ids()), 0)

    dashboard_data['dashboard_sheets'][0]['visual_widgets'] = self._visual_list([
                                                                                missing_id, visual_1.id])
    print(dashboard_data['dashboard_sheets'][0])
    db_1.report_data = json.dumps(dashboard_data)
    db_1.save()
    self.assertEqual(len(get_broken_dashboard_ids()), 1)

    self.assertEqual(len(fix_broken_dashboards()), 1)
    self.assertEqual(len(get_broken_dashboard_ids()), 0)
    db_1.refresh_from_db()

    self.assertEqual(db_1.get_visuals_in_dashboard()[0], [visual_1.id])
