# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.http import HttpRequest
from django.test import TestCase
from django.conf import settings

import apps
from unittest import mock

from rbac.permdefs import PermissionException
from util.ff_flipper import FeatureFlipper

class MockDataConnProperties(object):
  def __init__(self, test):
    self.test = test
    self.supports_limit = True
    self.supports_limit_0 = True

class MockDataConn(object):
  def __init__(self, test, id, dataconnectionName):
    self.id = id
    self.dataconnectionName = dataconnectionName
    self.test = test

  def properties(self):
    return MockDataConnProperties(self.test)

  def table_columns(self, tablename, username=None):
    return self.test.table_columns(tablename, username)

  def sync_runquery(self, cmd, username=None):
    return self.test.sync_runquery(cmd, username)

class MockDataConnection(object):
  def __init__(self, test, id, dataconnectionName):
    self.id = id
    self.dataconnectionName = dataconnectionName
    self.test = test

  def checkPerm(self, user, perm):
   self.test.mockCheckPerm(user, perm)
class MockDataset(object):
  def __init__(self, id, dataset_detail):
    self.id = id
    self.dataset_detail = dataset_detail

class MockUser(object):
  def __init__(self, username):
    self.username = username

class Test_DatasetCreate(TestCase):
  def setUp(self):
    FeatureFlipper().set_flipper('enable_direct_access', True, 'SETTINGS')
    self.mockCheckPermRaiseException = False

  def __init__(self, *args, **kwargs):
    super(Test_DatasetCreate, self).__init__(*args, **kwargs)
    self.user = MockUser('user3')
    self.dataconnection_name = "samples"
    self.dataconnection_id = 7
    self.query = None
    self.fq_tablename = "main.cereals"

  @mock.patch('apps.views.LOGGER')
  def test_disabled_autocreate(self, mock_logger):
    # GIVEN
    settings.ENABLE_DATASET_AUTOCREATE = False
    FeatureFlipper().set_flipper('enable_direct_access', False, 'SETTINGS')

    # WHEN
    dsId = apps.views.getOrCreateDatasetFromQuery(self.user, self.dataconnection_name,
        self.query, self.fq_tablename)

    # THEN
    self.assertEqual(dsId, None)
    self.assertEquals(str(mock_logger.method_calls[-1]),
        "call.warn('autocreate not enabled')")

  @mock.patch('apps.views.LOGGER')
  @mock.patch('apps.views.fetchDataConnectionByName')
  @mock.patch('apps.views.fetchDatasetsByDataConnection')
  def test_no_permission_autocreate_warn(self,mock_fetchDatasetsByDataConnection, mock_fetchDataConnection, mock_logger):
    # GIVEN
    mock_fetchDatasetsByDataConnection.side_effect = self.mock_fetchDatasetsByDataConnection
    mock_fetchDataConnection.side_effect = self.mock_fetchDataConnection

    self.mockCheckPermRaiseException = True

    # WHEN
    dsId = apps.views.getOrCreateDatasetFromQuery(self.user, self.dataconnection_name,
        self.query, self.fq_tablename)

    # THEN
    self.assertEqual(dsId, None)
    self.assertEquals(str(mock_logger.method_calls[-1]),
        "call.warn('User has no permission to get or create dataset')")

  @mock.patch('apps.views.LOGGER')
  def test_directaccess_not_enabled(self, mock_logger):
    # WHEN
    FeatureFlipper().set_flipper('enable_direct_access', False, 'SETTINGS')
    dsId = apps.views.getOrCreateDatasetFromQuery(self.user, self.dataconnection_name,
        self.query, self.fq_tablename)

    # THEN
    self.assertEqual(dsId, None)
    self.assertEquals(str(mock_logger.method_calls[-1]),
        "call.warn('autocreate not enabled')")

  @mock.patch('apps.views.LOGGER')
  def test_no_query_or_table(self, mock_logger):
    dsId = apps.views.getOrCreateDatasetFromQuery(self.user, self.dataconnection_name,
        None, None)
    self.assertEqual(dsId, None)
    self.assertEquals(str(mock_logger.method_calls[-1]),
        "call.warn('No query or table specified')")

  @mock.patch('apps.views.fetchDataConnectionByName')
  @mock.patch('apps.views.fetchDatasetsByDataConnection')
  def test_create_by_dataconnection_name(self, mock_fetchDatasetsByDataConnection, mock_fetchDataConnection):
    mock_fetchDatasetsByDataConnection.side_effect = self.mock_fetchDatasetsByDataConnection
    mock_fetchDataConnection.side_effect = self.mock_fetchDataConnection
    dsId = apps.views.getOrCreateDatasetFromQuery(self.user, self.dataconnection_name,
        self.query, "main.census_pop")
    self.assertEqual(dsId, 3)

    dsId = apps.views.getOrCreateDatasetFromQuery(self.user, self.dataconnection_name,
        self.query, "main.cereals")
    self.assertEqual(dsId, 4)

  def mock_get_dataconn(self, id):
        return MockDataConn(self, id, 'dc11')

  def mock_dataconnection_obj(self, get):
    self.get = get
    return self

  @mock.patch('apps.views.get_dataconn')
  @mock.patch('datasets.models.DataConnection.objects')
  @mock.patch('apps.views.fetchDatasetsByDataConnection')
  def test_create_by_dataconnection_id(self, mock_fetchDatasetsByDataConnection, mock_dataconnection_obj, mock_get_dataconn):
    mock_get_dataconn.side_effect = self.mock_get_dataconn
    mock_dataconnection_obj.side_effect = self.mock_dataconnection_obj
    mock_dataconnection_obj.get.return_value = MockDataConnection(self, 7, "connection_name")
    mock_fetchDatasetsByDataConnection.side_effect = self.mock_fetchDatasetsByDataConnection
    dsId = apps.views.getOrCreateDatasetFromQuery(self.user, 8,
        self.query, "main.iris")
    self.assertEqual(dsId, 5)

  def mockCheckPerm(self, user, perm):
    if self.mockCheckPermRaiseException:
      raise PermissionException(user, 'mocked behaviour', 'dc_explore')
    else:
      self.assertEqual(user.username, 'user3')
      self.assertEqual(perm, 'dc_explore')

  def mock_fetchDatasetsByDataConnection(self, dataconnectionName):
    return [MockDataset(3, "main.census_pop"), MockDataset(4, "main.cereals"), MockDataset(5, "main.iris")]

  def mock_fetchDataConnection(self, dataconnectionName):
    self.request_dataconnectionName = dataconnectionName
    return [MockDataConnection(self, 7, dataconnectionName)]

  @mock.patch('apps.views.fetchDataConnectionByName')
  @mock.patch('apps.views.fetchDatasetsByDataConnection')
  @mock.patch('apps.views.get_dataconn')
  @mock.patch('apps.views.datasets.views.dataset_create_or_clone')
  def base_create_from_table(self, mock_dataset_create_or_clone, mock_getdataconn,
      mock_fetchDatasetsByDataConnection, mock_fetchDataConnection):

    mock_fetchDatasetsByDataConnection.side_effect = self.mock_fetchDatasetsByDataConnection
    mock_fetchDataConnection.side_effect = self.mock_fetchDataConnection
    mock_dataset_create_or_clone.side_effect = self.mock_dataset_create_or_clone_table
    mock_getdataconn.side_effect = self.mock_get_dataconn

    return apps.views.getOrCreateDatasetFromQuery(self.user, self.dataconnection_name,
        self.query, self.fq_tablename)

  def test_create_from_table(self):
    self.fq_tablename = "main.abc"
    dsId = self.base_create_from_table()
    self.assertEqual(dsId, 3)
    self.assertEqual(self.request_dataconnectionName, "samples")

  def test_create_from_table_default_dataconn(self):
    self.fq_tablename = "main.abc"
    self.dataconnection_name = None

    previous = settings.DATASET_AUTOCREATE_DATACONNECTION
    try:
      settings.DATASET_AUTOCREATE_DATACONNECTION = "override"
      dsId = self.base_create_from_table()
      self.assertEqual(dsId, 3)
      self.assertEqual(self.request_dataconnectionName, "override")
    finally:
      settings.DATASET_AUTOCREATE_DATACONNECTION = previous

  @mock.patch('apps.views.LOGGER')
  def test_create_from_table_invalidname(self, mock_logger):
    self.fq_tablename = "invalid_table"
    dsId = self.base_create_from_table()
    self.assertEqual(dsId, None)
    self.assertEquals(str(mock_logger.method_calls[-1]),
        "call.warn('Invalid table specified %s', 'invalid_table')")

  @mock.patch('apps.views.LOGGER')
  def test_create_from_table_invalidquery(self, mock_logger):
    self.fq_tablename = None
    self.query = "insert into abc"
    dsId = self.base_create_from_table()
    self.assertEqual(dsId, None)
    self.assertEquals(str(mock_logger.method_calls[-1]),
        "call.warn('Invalid query specified %s', 'insert into abc')")

  def table_columns(self, tablename, username):
    self.assertEquals(username, 'user3')
    return {
      'colnames': ['col1', 'col2'],
      'coltypes': ['STRING', 'REAL'],
    }

  def sync_runquery(self, cmd, username):
    self.assertEquals(username, 'user3')
    self.assertEquals(cmd, 'select * from (select * from main.abc) cdv_t1 limit 0')
    return {
      'colnames': ['col1', 'col2'],
      'coltypes': ['STRING', 'REAL'],
    }

  def mock_dataset_create_or_clone_table(self, data, request: HttpRequest):
    user = request.user
    self.assertEquals(user.username, 'user3')
    self.assertEquals(data, {
      'dataconnection_id': 7,
      'dataset_type': 'singletable',
      'dataset_settings': '{}',
      'dataset_lvname': '',
      'dataset_detail': 'main.abc',
      'description': 'Auto-created for SQL apps',
      'dataset_name': 'main.abc',
      'dataset_info': '[{"tablename": "main.abc", "columns": [{"name": "col1", "alias": "col1", "type": "STRING", "isdim": true}, {"name": "col2", "alias": "col2", "type": "REAL", "isdim": false}]}]'
    })

    return 3

  @mock.patch('apps.views.fetchDataConnectionByName')
  @mock.patch('apps.views.fetchDatasetsByDataConnection')
  @mock.patch('apps.views.get_dataconn')
  @mock.patch('apps.views.datasets.views.dataset_create_or_clone')
  def test_create_from_query(self, mock_dataset_create_or_clone, mock_getdataconn,
      mock_fetchDatasetsByDataConnection, mock_fetchDataConnection):

    mock_fetchDatasetsByDataConnection.side_effect = self.mock_fetchDatasetsByDataConnection
    mock_fetchDataConnection.side_effect = self.mock_fetchDataConnection
    mock_dataset_create_or_clone.side_effect = self.mock_dataset_create_or_clone_query
    mock_getdataconn.side_effect = self.mock_get_dataconn

    dsId = apps.views.getOrCreateDatasetFromQuery(self.user, self.dataconnection_name,
        "select * from main.abc", None)
    self.assertEqual(dsId, 3)

  def mock_dataset_create_or_clone_query(self, data, request: HttpRequest):
    user = request.user
    self.assertEquals(user.username, 'user3')
    expected_result = {
      'dataconnection_id': 7,
      'dataset_type': 'singletable',
      'dataset_settings': '{}',
      'dataset_lvname': '',
      'dataset_detail': 'Created from SQL',
      'description': 'Auto-created for SQL apps',
      'dataset_name': 'autogen 79d7d42e-4df2-4109-a4c9-d16fabb73136',
      'dataset_info': '[{"tablename": "text.sql", "select": "select * from main.abc", "columns": [{"name": "col1", "alias": "col1", "type": "STRING", "isdim": true}, {"name": "col2", "alias": "col2", "type": "REAL", "isdim": false}]}]'
    }

    expected_result['dataset_name'] = data['dataset_name']
    self.assertEquals(data, expected_result)

    return 3

