# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.test import TestCase, override_settings
from django.test.client import Client
from django.http import HttpResponse
import json
import os
import re
import random

from apps.tests.userrole import base
from dataconn.ext.tests.test_main import MyExtension
from datasets.models import DataConnection
from util.ff_flipper import FeatureFlipper
from vizlib.tests import test_sqlrun
from vizlib.tests.util import run_with_db_query_check, get_db_queries


def make_dataconnection_sqlite(file_path: str, name='sqlite'):
  info = json.dumps({
      'PARAMS': {
          'FILE': file_path
      }
  })
  return DataConnection.objects.create(
      dataconnection_name=name, dataconnection_info=info, dataconnection_type='sqlite'
  )


class TestViews(TestCase):

  def setUp(self):
    self.sqlite_file = 'tests_views.sqlite'
    self.user1 = base.make_user('user1')
    self.client1 = base.create_client(self.user1)

  def tearDown(self):
    if os.path.exists(self.sqlite_file):
      print(f'removing {self.sqlite_file}')
      os.remove(self.sqlite_file)

  def test_dashboard_creation_from_sql(self):
    dc = make_dataconnection_sqlite(file_path=self.sqlite_file)

    user_admin = test_sqlrun.Common.make_user('admin', is_superuser=True)
    client_admin = Client()
    client_admin.login(
        username=user_admin.username, password=user_admin.username
    )

    query_params = {
        'report_type': 'dashboard',
        'dataconnection_id': dc.id,
        'query': 'select "Hello from sqlite"'
    }
    response = client_admin.get('/apps/builder', query_params)
    self.assertEquals(response.status_code, 200)

  def test_home(self):
    response = self.client1.get('/apps/home')
    self.assertEqual(response.status_code, 200)
    self.assertFalse(response.context['mobile'])

  def test_home_mobile(self):
    response = self.client1.get('/apps/home', {'mobile': 'true'})
    self.assertEqual(response.status_code, 200)
    self.assertTrue(response.context['mobile'])

HTTP_DATACONN_ID_SETTER = '\.http_dataconnection_id\(([^\)]*)\)'


class TestViewTemplates(TestCase):

  def setUp(self):
    self.FF = FeatureFlipper()

    # store original settings
    self.site_enable_search = self.FF.get_flipper_value('site_enable_search')

  def tearDown(self):
    # restore settings
    self.FF.set_flipper('site_enable_search', self.site_enable_search, 'DB')

  def _check(self, response: HttpResponse, expected_ids: list[str]):
    self.assertEquals(response.status_code, 200)

    response_string = response.content.decode()
    http_dataconn_ids = re.findall(HTTP_DATACONN_ID_SETTER, response_string)
    self.assertEquals(http_dataconn_ids, expected_ids)

  def test_login_template(self):
    anonymous_client = Client()

    response = anonymous_client.get('/apps/login')
    # no globals are set if search is not enabled
    self._check(response, [])

    self.FF.set_flipper('site_enable_search', True, 'DB')
    response = anonymous_client.get('/apps/login')
    # login page uses a default value of 0
    self._check(response, ['0'])

  def test_apps_template(self):
    user1 = base.make_user('user1')
    client = base.create_client(user1)

    response = client.get('/apps/apps')
    # no globals are set if search is not enabled
    self._check(response, [])

    self.FF.set_flipper('site_enable_search', True, 'DB')
    response = client.get('/apps/apps')
    # non-login pages use -10
    self._check(response, ['-10'])

class TestViewsData(TestCase):

  @override_settings(DATACONN_EXTENSIONS=[MyExtension()])
  def test_dc_extensions_on_data_page(self):
    admin = base.make_user('admin1', is_superuser=True)
    admin_client = base.create_client(admin)

    response = admin_client.get('/apps/data')
    self.assertEquals(response.status_code, 200)

    dc_extensions = response.context['dc_ext_list']

    # extract default extensions:
    extension_names = [json.loads(ext)['name'] for ext in dc_extensions]
    self.assertListEqual(extension_names, ['my-conn'])

class TestGroups(TestCase):
  def setUp(self):
    self.admin1 = base.make_user('admin1', is_superuser=True)
    self.admin1_client = base.create_client(self.admin1)

  def get_query_count(self):
    """
    We omit RBAC queries because of the undeterministic caching behavior
    """
    return len(get_db_queries(exclude_rbac_role_loads=True))

  def test_groups_get(self):

    response, _ = run_with_db_query_check(lambda: self.admin1_client.get('/apps/groups_api'))
    self.assertTrue(response.status_code, 200)
    groups = response.json()
    # no groups yet
    self.assertEquals(groups, [])

    # create a group with one user
    group_1 = base.make_group('group_1')
    self.admin1.groups.add(group_1)
    self.admin1.save()

    response, _ = run_with_db_query_check(lambda: self.admin1_client.get('/apps/groups_api'))
    query_count_2 = self.get_query_count()
    groups = response.json()
    self.assertEquals(len(groups), 1)
    self.assertEquals(groups[0]['name'], 'group_1')

    user1 = base.make_user('user1', False)
    user1.groups.add(group_1)
    response, _ = run_with_db_query_check(lambda: self.admin1_client.get('/apps/groups_api'))
    query_count_3 = self.get_query_count()
    self.assertEqual(query_count_2, query_count_3, 'equal DB queries')

    # create another group with one user
    group_2 = base.make_group('group_2')
    self.admin1.groups.add(group_2)
    self.admin1.save()

    response, _ = run_with_db_query_check(lambda: self.admin1_client.get('/apps/groups_api'))
    query_count_4 = self.get_query_count()
    groups = response.json()
    self.assertEquals(len(groups), 2)
    self.assertEqual(query_count_4, query_count_3, 'equal DB queries')

  def create_larger_dataset(self):
    # Create 5 users
    users = [base.make_user(f'user{i}', False) for i in range(0, 5)]

    # Create 11 groups
    groups = [base.make_group(f'group_{i}') for i in range(0, 11)]

    # Assign each user randomly to one of the 11 groups
    for user in users:
      random_group = random.choice(groups)
      user.groups.add(random_group)

    return users, groups
  def test_groups_pagination(self):
    # Create a dataset with 5 users and 11 groups
    users, groups = self.create_larger_dataset()

    # Trigger the DataTables AJAX request (page 1)
    response1, _ = run_with_db_query_check(
      lambda: self.admin1_client.get('/apps/groups_api', {
        'draw': 1,
        'start': 0,
        'length': 10,
      }))
    query_count_1 = self.get_query_count()
    self.assertEquals(query_count_1, 7)

    groups_page1 = response1.json()
    self.assertEquals(len(groups_page1['groups']), 10)

    self.assertEquals(groups_page1['numberOfGroups'], 11)

    # Trigger the DataTables AJAX request (page 2)
    response2, _ = run_with_db_query_check(
      lambda: self.admin1_client.get('/apps/groups_api', {
        'draw': 2,
        'start': 10,
        'length': 10,
      }))

    groups_page2 = response2.json()
    self.assertEquals(len(groups_page2['groups']), 1)

    self.assertEquals(groups_page2['numberOfGroups'], 11)

    # search on the groups tab
    response3, _ = run_with_db_query_check(
      lambda: self.admin1_client.get('/apps/groups_api', {
        'draw': 1,
        'start': 0,
        'length': 10,
        'search[value]': 'group_7'
      }))

    search_result_page = response3.json()

    self.assertEquals(len(search_result_page['groups']), 1)

    self.assertEquals(search_result_page['numberOfFiltered'], 1)
    self.assertEquals(search_result_page['numberOfGroups'], 11)

    # search in the user edit modal
    response4, _ = run_with_db_query_check(
      lambda: self.admin1_client.get('/apps/groups_api', {
        'search[value]': 'group_7'
      }))

    search_result_in_user_modal = response4.json()

    self.assertTrue(type(search_result_in_user_modal) == list)
    self.assertEquals(len(search_result_in_user_modal), 1)
