# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import django.test
from django.contrib.auth.models import User
import json
from apps.tests.userrole import base


class DemoApiKeysTestCase(django.test.TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            username='testuser',
            email='testuser@example.com',
            password='testpassword'
        )

        self.requestData = {
            'username': 'testuser',
            'password': 'testpassword',
            'active': 'true',
            'expires': '2024-04-19T00:00:00+02:00'
        }

        self.perm_editers = base.make_role("perm_editers", ["sys_editperm"])
        self.perm_editers.add_users([self.user.username])
        self.perm_editers.save()
        self.client = base.create_client(self.user)

    def tearDown(self):
        self.user.delete()

    def test_demo_apikeys_with_valid_credentials(self):

        response = self.client.post('/apps/demoapikeys', data=self.requestData)

        result = json.loads(response.content)
        hasApiKey = 'apikey' in result

        self.assertEqual(response.status_code, 200)
        self.assertEqual(result['expires'], self.requestData['expires'])
        self.assertTrue(hasApiKey, msg='response not contains apikey')
        self.assertEqual(len(result['apikey']), 32)

    def test_demo_apikeys_with_unlimited_expiry_date (self):
        data = self.requestData
        data['expires'] = ''

        response = self.client.post('/apps/demoapikeys', data=data)
        result = json.loads(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(result['expires'], 'None')

    def test_demo_apikeys_with_invalid_password(self):
        data = self.requestData
        data['password'] = 'wrongpassword'

        response = self.client.post('/apps/demoapikeys', data=data)

        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.content, b'Invalid authentication')

    def test_demo_apikeys_with_invalid_username(self):
        data = self.requestData
        data['username'] = 'wronguser'

        response = self.client.post('/apps/demoapikeys', data=data)

        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.content, b'Invalid authentication')


    def test_demo_apikeys_without_required_permission(self):
        self.perm_editers.delete_users([self.user.username])
        self.perm_editers.save()

        response = self.client.post('/apps/demoapikeys', data=self.requestData)

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.content, b'You do not have the necessary permission on this system')
