# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from apps.tests import tests_views_visuals_page
from apps.views_home_page import COLLECT_RELATED_USING_PREFETCH, get_home_page_info
from datasets.models import DataConnection
from django.http import HttpRequest
from django.test import TestCase, Client, RequestFactory
from django.contrib.auth.models import User
from extracts.tests.tests_extract_get import create_role_with_permissions
from instant import instant
from jobs.tests.test_jobs_views import grant_all_dataconn_permissions
from sql.models import Query
from reports.models import AppGroup, ReportUsage
from vizlib.tests.util import run_with_db_query_check, get_db_queries
from vizlib.tests.test_report_operations import make_user

import json

# reports need 3 extra queries to get their related data when using prefetch
EXTRA_QUERIES = 3 if COLLECT_RELATED_USING_PREFETCH else 0

NUMBER_OF_SAMPLE_DATACONNECTIONS=1
NUMBER_OF_SAMPLE_DATASETS=11
NUMBER_OF_SAMPLE_DASHBOARDS=16
NUMBER_OF_SAMPLE_APPS=2

def grant_dataconn_creation_permission(user: User):
  create_role_with_permissions(user, [{'ptype':'system','perms':['sys_editconn']}])

class TestViewsHomePageBase(TestCase):
  def setUp(self):
    self.common = tests_views_visuals_page.TestReportCountPerGroup()
    self.common.common_setup()
    self.common.util.create_dataconnection(self.common.admin_client, connection_name='DC 2 created by admin')

    self.simulate_dashboard_usage()
    self.simulate_queries()

  def get_data(self, url: str):
    return self.get(url, data_only=True)[0]

  def get(self, url: str, check_dataset_info=True, expected_query_count=None, report_queries=0, view_function: callable = None, data_only=False) -> tuple[dict, int]:

    def action():
      # support testing view functions:
      if view_function:
        request = self._get_request(url)
        response = view_function(request)
        response.json = lambda: json.loads(response.content.decode())
        return response
      else:
        client = self._get_client()
        return client.get(url)

    # this is when we want to compare response contents are not interested in the DB queries
    if data_only:
      response = action()
      self.assertEquals(response.status_code, 200)
      return response.json(), None

    response, query_count = run_with_db_query_check(
      action,
      asserter=self,
      # when using prefetch, the number of extra queries is proportional with the Report queries.
      # some endpoints do not return reports (only summary data), others issue one Report query that needs related information
      expected_query_count=expected_query_count + EXTRA_QUERIES*report_queries if expected_query_count is not None else None)
    self.assertEquals(response.status_code, 200)

    if check_dataset_info:
      # make sure we do not retrieve dataset_info:
      query_text = ';\n'.join(get_db_queries())
      self.assertFalse('dataset_info' in query_text,
                       'dataset_info is retrieved')

    return response.json(), query_count

  def get_dashboard_ids(self) -> list[int]:
    # use the web endpoint so that we get the dashboards accessible
    # to the current user:
    client = self._get_client()
    response: list[dict] = client.get('/reports/report').json()
    return [d['id'] for d in response]

  def get_apps(self):
    return  AppGroup.objects.filter(group_type__in=['default', 'sample-app'],group_parent_id=None).all()

  def simulate_queries(self):
    Query.objects.create(create_user=self.common.admin.username, dataconnection_id=self.common.dc_id1)
    Query.objects.create(create_user=self.common.user1.username, dataconnection_id=self.common.dc_id1)

  def simulate_dashboard_usage(self):
    for d_id in self.get_dashboard_ids()[:2]:
      ReportUsage.objects.create(report_id=d_id, report_or_appgroup_id=d_id, count=1)

  def simulate_apps_usage(self):
    for a in self.get_apps()[:2]:
      ReportUsage.objects.create(report_or_appgroup_id=-a.id, count=1)

  def createSamples(self):
    make_user(instant.ADMIN_USERNAME, is_superuser=True)
    instant.createSamples()

    # enable user access to the samples
    samples_dc_id = DataConnection.objects.filter(dataconnection_name='samples')[0].id
    tests_views_visuals_page.grant_read_access_on_datasets(self.common.user1, dc_id=samples_dc_id)

    # currently there is a single app in the samples.
    # create one more so that we notice if query count depends on app count:
    AppGroup.objects.create(group_type='default')

    self.simulate_apps_usage()

  def _get_client(self) -> Client:
    raise NotImplementedError()

  def _get_request(self, url: str) -> HttpRequest:
    raise NotImplementedError()

class TestAdmin(TestViewsHomePageBase):

  def _get_client(self):
    return self.common.admin_client

  def _get_request(self, url: str) -> HttpRequest:
    request = RequestFactory().get(url)
    request.user = self.common.admin
    return request

  def test_homepage_statistics(self):
    response, _ = self.get(
      '/apps/homepage_statistics',
      expected_query_count=8)
    self.assertEquals(response['dashboards'], 6)
    self.assertEquals(response['datasets'], 2)
    self.assertEquals(response['apps'], 0)
    self.assertEquals(response['views'], 2)
    self.assertEquals(response['dataConnections'], 2)
    # only the queries created by the current user:
    self.assertEquals(response['queries'], 1)

    self.createSamples()

    response, _ = self.get(
      '/apps/homepage_statistics',
      expected_query_count=8)
    self.assertEquals(response['dashboards'], 6 + NUMBER_OF_SAMPLE_DASHBOARDS)
    self.assertEquals(response['datasets'], 2 + NUMBER_OF_SAMPLE_DATASETS)
    self.assertEquals(response['apps'], 0 + NUMBER_OF_SAMPLE_APPS)
    # does not contain app views, only dashboards:
    self.assertEquals(response['views'], 2 + 0)
    self.assertEquals(response['dataConnections'], 2 + NUMBER_OF_SAMPLE_DATACONNECTIONS)
    self.assertEquals(response['queries'], 1)

  def test_homepage_last_viewed_reports_by_user(self):
    user_profile = self.common.admin.userprofile
    user_profile.user_data_obj={'last_viewed': self.get_dashboard_ids()[:2]}
    user_profile.save()
    user_profile.refresh_from_db()

    response, _ = self.get(
      '/apps/homepage_last_viewed_reports_by_user',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=6)
    self.assertEqual(len(response['data']),2)

    self.createSamples()

    # simulate app viewing:
    apps = [-a.id for a in self.get_apps()[:2]]
    self.assertEqual(len(apps), 2)
    view_data = user_profile.user_data_obj
    view_data['last_viewed'] += apps
    user_profile.user_data_obj = view_data
    user_profile.save()
    user_profile.refresh_from_db()

    response, _ = self.get(
      '/apps/homepage_last_viewed_reports_by_user',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=6+1) # one extra query because of apps being viewed
    self.assertEqual(len(response['data']), 2 + 2)

  def test_homepage_most_popular_reports(self):
    response, _ = self.get(
      '/apps/homepage_most_popular_reports',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=8)
    self.assertEqual(len(response['data']), 2)

    self.createSamples()

    response, _ = self.get(
      '/apps/homepage_most_popular_reports',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=8+1) # 1 extra query because of appgroups
    self.assertEqual(len(response['data']), 2 + 2)

  def test_homepage_number_of_recently_created_resources(self):
    response, _ = self.get(
      '/apps/homepage_number_of_recently_created_resources',
      expected_query_count=5)
    self.assertEqual(response['datasets'], 2)
    self.assertEqual(response['dashboards'], 6)
    self.assertEqual(response['apps'], 0)

    self.createSamples()
    response, _ = self.get(
      '/apps/homepage_number_of_recently_created_resources',
      expected_query_count=5)
    self.assertEqual(response['datasets'], 2 + NUMBER_OF_SAMPLE_DATASETS)
    self.assertEqual(response['dashboards'], 6 + NUMBER_OF_SAMPLE_DASHBOARDS)
    self.assertEqual(response['apps'], 0 + NUMBER_OF_SAMPLE_APPS)

  def test_homepage_recently_created_dashboards_by_user(self):
    response, _ = self.get(
      '/apps/homepage_recently_created_dashboards_by_user',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=5
      )
    self.assertEqual(len(response['data']), 4) # admin created only 4 dashboards

    self.createSamples()
    response, _ = self.get(
      '/apps/homepage_recently_created_dashboards_by_user',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=5
      )
    self.assertEqual(len(response['data']), 4 + 0) # samples are imported by vizapps_admin, not admin

  def test_homepage_samples_dashboards(self):
    # there are no samples, but must not fail
    response, _ = self.get(
      '/apps/homepage_samples',
      expected_query_count=4)
    self.assertEqual(len(response['data']), 0)

    self.createSamples()
    response, _ = self.get(
      '/apps/homepage_samples',
      check_dataset_info=False,
      report_queries=2,
      expected_query_count=10) # TODO: this seems a bit high, there might be extra queries
    self.assertEqual(len(response['data']), 11) # limited by the tile count (=10)


  def test_homepage_favorite_dashboards(self):

    # must invoke it twice as on first invocation it inserts data to the DB:
    self.common.admin_client.get('/apps/homepage_favorite_dashboards')

    response, _ = self.get(
      '/apps/homepage_favorite_dashboards',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=7)
    self.assertEqual(len(response['data']), 4) # automatic favorites creation makes all dashboards starred

  def test_homepage_datasets(self):
    response, _ = self.get(
      '/apps/homepage_datasets?limit=500',
      check_dataset_info=False,
      expected_query_count=3)
    self.assertEqual(len(response['datasets']),2)

  def test_get_home_page_info(self):
    # create samples so we have more data:
    self.createSamples()

    # the URL does not matter as we are invoking the view function directly
    # it might be important if it had query parameters, for example
    url = '/does/not/matter/home_page_info'

    # at the time of writing this needs 44 queries:
    response, _ = self.get(url, view_function=get_home_page_info, check_dataset_info=False)
    self.assertEqual(response['statistics'], self.get_data('/apps/homepage_statistics'))
    self.assertEqual(response['number_of_recently_created_resources'], self.get_data('/apps/homepage_number_of_recently_created_resources'))
    self.assertEqual(response['last_viewed_reports_by_user'], self.get_data('/apps/homepage_last_viewed_reports_by_user')['data'])
    self.assertEqual(response['most_popular_reports'], self.get_data('/apps/homepage_most_popular_reports')['data'])
    self.assertEqual(response['samples'], self.get_data('/apps/homepage_samples')['data'])
    self.assertEqual(response['favorite_dashboards'], self.get_data('/apps/homepage_favorite_dashboards')['data'])

class TestUser(TestViewsHomePageBase):

  def _get_client(self):
    return self.common.user1_client

  def _get_request(self, url: str) -> HttpRequest:
    request = RequestFactory().get(url)
    request.user = self.common.user1
    return request

  def test_homepage_statistics(self):
    response, _ = self.get(
      '/apps/homepage_statistics',
      expected_query_count=9 + 3) # user group + roles + accessible datasets + workspaces
    self.assertEquals(response['dashboards'], 2 + 1) # 2 created by the user, 1 public created by Admin
    self.assertEquals(response['datasets'], 1)
    self.assertEquals(response['apps'], 0)
    self.assertEquals(response['views'], 2)
    self.assertEquals(response['dataConnections'], 1)
    # only the queries created by the current user:
    self.assertEquals(response['queries'], 1)

    self.createSamples()

    response, _ = self.get(
      '/apps/homepage_statistics',
      expected_query_count=12)
    self.assertEquals(response['dashboards'], 3 + NUMBER_OF_SAMPLE_DASHBOARDS)
    self.assertEquals(response['datasets'], 1 + NUMBER_OF_SAMPLE_DATASETS)
    self.assertEquals(response['apps'], 0 + NUMBER_OF_SAMPLE_APPS)
    # does not contain app views, only dashboards:
    self.assertEquals(response['views'], 2 + 0)
    self.assertEquals(response['dataConnections'], 1 + NUMBER_OF_SAMPLE_DATACONNECTIONS)
    self.assertEquals(response['queries'], 1)

    # test dataconnection access: self-created data connections should show up
    grant_dataconn_creation_permission(self.common.user1)

    # test that granting sys_editconn did not make DC2 visible for user1
    response, _ = self.get(
      '/apps/homepage_statistics',
      expected_query_count=12)
    self.assertEquals(response['dataConnections'], 1 + 1)

    # the self created dataconnection should show up:
    self.common.util.create_dataconnection(self.common.user1_client, connection_name='DC for user1')
    response, _ = self.get(
      '/apps/homepage_statistics',
      expected_query_count=12)
    self.assertEquals(response['dataConnections'], 1 + 1 + 1)

  def test_homepage_last_viewed_reports_by_user(self):
    user_profile = self.common.user1.userprofile
    user_profile.user_data_obj={'last_viewed':self.get_dashboard_ids()[:2]}
    user_profile.save()
    user_profile.refresh_from_db()

    response, _ = self.get(
      '/apps/homepage_last_viewed_reports_by_user',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=6 + 4) # a non admin has 4 extra queries
    self.assertEqual(len(response['data']),2)

    self.createSamples()

    # simulate app viewing:
    apps = [-a.id for a in self.get_apps()[:2]]
    self.assertEqual(len(apps), 2)
    view_data = user_profile.user_data_obj
    view_data['last_viewed'] += apps
    user_profile.user_data_obj = view_data
    user_profile.save()
    user_profile.refresh_from_db()

    response, _ = self.get(
      '/apps/homepage_last_viewed_reports_by_user',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=6+1+4) # one extra query because of apps being viewed
    self.assertEqual(len(response['data']), 2 + 2)

  def test_homepage_most_popular_reports(self):
    response, _ = self.get(
      '/apps/homepage_most_popular_reports',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=9 + 3)
    self.assertEqual(len(response['data']), 2)

    self.createSamples()

    response, _ = self.get(
      '/apps/homepage_most_popular_reports',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=9+1+3) # 1 extra query because of appgroups
    self.assertEqual(len(response['data']), 2 + 2)

  def test_homepage_number_of_recently_created_resources(self):
    response, _ = self.get(
      '/apps/homepage_number_of_recently_created_resources',
      expected_query_count=6 + 3)
    self.assertEqual(response['datasets'], 1)
    self.assertEqual(response['dashboards'], 2 + 1) # the public dashboard created by admin is also visible
    self.assertEqual(response['apps'], 0)

    self.createSamples()
    response, _ = self.get(
      '/apps/homepage_number_of_recently_created_resources',
      expected_query_count=6 + 3)
    self.assertEqual(response['datasets'], 1 + NUMBER_OF_SAMPLE_DATASETS)
    self.assertEqual(response['dashboards'], 2 + 1+ NUMBER_OF_SAMPLE_DASHBOARDS)
    self.assertEqual(response['apps'], 0 + NUMBER_OF_SAMPLE_APPS)

  def test_homepage_recently_created_dashboards_by_user(self):
    response, _ = self.get(
      '/apps/homepage_recently_created_dashboards_by_user',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=6 + 3
      )
    self.assertEqual(len(response['data']), 2) # 2 dashboards were created by user

    self.createSamples()
    response, _ = self.get(
      '/apps/homepage_recently_created_dashboards_by_user',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=6 + 3
      )
    self.assertEqual(len(response['data']), 2 + 0) # samples are imported by vizapps_admin, not admin

  def test_homepage_samples_dashboards(self):
    # there are no samples, but must not fail
    response, _ = self.get(
      '/apps/homepage_samples',
      expected_query_count=4 + 3 + 1)
    self.assertEqual(len(response['data']), 0)

    self.createSamples()
    response, _ = self.get(
      '/apps/homepage_samples',
      check_dataset_info=False,
      report_queries=2,
      expected_query_count=10 + 3 + 1) # TODO: this seems a bit high, there might be extra queries
    self.assertEqual(len(response['data']), 11) # limited by the tile count (=10)


  def test_homepage_favorite_dashboards(self):

    # must invoke it twice as on first invocation it inserts data to the DB:
    self.common.user1_client.get('/apps/homepage_favorite_dashboards')

    response, _ = self.get(
      '/apps/homepage_favorite_dashboards',
      check_dataset_info=False,
      report_queries=1,
      expected_query_count=8 + 3)
    self.assertEqual(len(response['data']), 2) # automatic favorites creation makes all dashboards starred

  def test_homepage_datasets(self):
    # the existing datasets were created by admin
    # create one as user so that we have some data:
    grant_all_dataconn_permissions(self.common.user1, self.common.dc_id1)
    self.common.util.create_dataset(self._get_client(), self.common.dc_id1, 'Dataset created by user')

    response, _ = self.get(
      '/apps/homepage_datasets?limit=500',
      check_dataset_info=False,
      expected_query_count=3 + 0) # these are the datasets created by the current user, no permission check is made
    self.assertEqual(len(response['datasets']),1)

  def test_get_home_page_info(self):
    # create samples so we have more data:
    self.createSamples()

    # the URL does not matter as we are invoking the view function directly
    # it might be important if it had query parameters, for example
    url = '/does/not/matter/home_page_info'

    # at the time of writing this needs 48 queries:
    response, _ = self.get(url, view_function=get_home_page_info, check_dataset_info=False)
    self.assertEqual(response['statistics'], self.get_data('/apps/homepage_statistics'))
    self.assertEqual(response['number_of_recently_created_resources'], self.get_data('/apps/homepage_number_of_recently_created_resources'))
    self.assertEqual(response['last_viewed_reports_by_user'], self.get_data('/apps/homepage_last_viewed_reports_by_user')['data'])
    self.assertEqual(response['most_popular_reports'], self.get_data('/apps/homepage_most_popular_reports')['data'])
    self.assertEqual(response['samples'], self.get_data('/apps/homepage_samples')['data'])
    self.assertEqual(response['favorite_dashboards'], self.get_data('/apps/homepage_favorite_dashboards')['data'])

  """
  from django.db import connection
  c=connection.cursor()
  c.execute('select count(*) from reports_reportusage ru join reports_report r on ru.report_or_appgroup_id=r.id join datasets_dataset ds on r.dataset_id=ds.id join workspace_workspace ws on r.workspace_id=ws.id where ru.report_or_appgroup_id>=0').fetchall()

  c.execute('select count(*),sum(ru.count) from reports_reportusage ru join reports_appgroup ag on ru.report_or_appgroup_id=-ag.id join workspace_workspace ws on ag.workspace_id=ws.id where
ru.report_or_appgroup_id<0').fetchall()
  """