# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from unittest.mock import AsyncMock, MagicMock, Mock, patch
from django.conf import settings
from django.test import TestCase, TransactionTestCase
from django.contrib.auth.models import User
from django.http import HttpRequest

import os
import tempfile
import time
import json
import tempfile
import xlrd

import apps.views_snapshot as vs
from apps.tests.userrole import base
from reports.models import Report, AppSetting
from util.request import create_request
from vizlib import misc
from extracts.tests import tests_rewrites
from vizlib.tests import test_report_operations as test_utils, test_sqlrun, test_report_operations
from apps.tests.userrole import base
from datetime import datetime
from django.core.cache import cache
from reports.views_settings import setting_get_single, setting_update_single
from sqlrun import views
from unittest import mock, skip
from vizlib.dsreq import DsreqCreator

helper = test_utils.TestDashboardOperations()


def poll_until_ready(client, url, timeout=10, interval=1):
  start_time = time.time()
  while time.time() - start_time < timeout:
    response = client.get(url)
    if response.content.decode('utf-8') == "ready":
      return response
    time.sleep(interval)
  raise TimeoutError(f"'ready' not found in response body within {timeout} seconds")

class SnapshotTest(TestCase):
    def setUp(self):
        self.outDir = tempfile.TemporaryDirectory()

        # enable snapshotting
        self.OLD_PDF_OPTIONS = vs.PDF_OPTIONS
        vs.PDF_OPTIONS = {'ENABLE_CHROMIUM_PDF': True}
        vs.init()
        vs.CHROMIUM_PDF_AVAILABLE = True

        self.create_app()
        self.create_user()
        self.add_required_settings()

        setattr(vs, '_SNAPSHOT_RUNNER', None)
        setattr(vs, '_ADHOC_SNAPSHOT_RUNNER', None)

        # mock PNG content reading:
        self.get_thumbnail_bytes_patch = patch.object(vs, 'get_thumbnail_bytes')
        get_thumbnail_bytes_mock = self.get_thumbnail_bytes_patch.start()
        get_thumbnail_bytes_mock.return_value = 'mocked thumbnail bytes'.encode()

        # spy on temp folder creation
        self._saferemove_snapshotdir_mock = MagicMock(wraps=vs._saferemove_snapshotdir)
        self._saferemove_snapshotdir_patch = patch.object(
            vs,
            '_saferemove_snapshotdir',
            new=self._saferemove_snapshotdir_mock)
        self._saferemove_snapshotdir_patch.start()


    def tearDown(self):
        self._saferemove_snapshotdir_patch.stop()
        self.get_thumbnail_bytes_patch.stop()

        setattr(vs, '_SNAPSHOT_RUNNER', None)
        setattr(vs, '_ADHOC_SNAPSHOT_RUNNER', None)

        vs.CHROMIUM_PDF_AVAILABLE = False
        vs.PDF_OPTIONS = self.OLD_PDF_OPTIONS

        self.outDir.cleanup()

    def add_required_settings(self):
        settings = AppSetting()
        settings.setting_key = "site_thumbnails_ip"
        settings.setting_value = "http://localhost:8100"
        settings.save()

    def create_app(self):
        app = Report(
            report_name="app1",
            report_description="report desc",
            report_type="report type",
            report_data="{}",
            create_user="user1",
            update_user="user2",
        )
        app.save()
        self.app = app

    def create_user(self):
        user = User()
        user.save()
        self.user = user

        request = HttpRequest()
        request.user = user
        self.request = request

    def prepare_snapshot_runner(self, is_thumbnail=False):
        snapshot_runner = vs._SnapshotRunner_chromium(
            appId=self.app.id,
            request=self.request,
            config={
                "outDir": self.outDir.name,
                "saveType": "screenshot" if not is_thumbnail else "thumbnail"
            },
        )

        snapshot_runner.create_browser = AsyncMock()

        browser = MagicMock()
        browser.close = AsyncMock()
        browser.process.pid = 123
        snapshot_runner.create_browser.return_value = browser

        snapshot_runner.perform_snapshotting = AsyncMock()

        self.snapshot_runner = snapshot_runner
        self.browser = browser

        # mock manager to enable invocation order check
        mock_manager = MagicMock()
        mock_manager.attach_mock(browser, 'browser')
        mock_manager.attach_mock(self._saferemove_snapshotdir_mock, '_saferemove_snapshotdir')
        self.mock_manager = mock_manager

    def perform_assertions(self):
        browser = self.browser

        browser.close.assert_called_once()
        browser.process.terminate.assert_called_once()
        browser.process.wait.assert_called_once()

        self.assertIsNone(self.snapshot_runner.browser)

        self.assertEqual((1, 1), self.snapshot_runner.progress())

    def assert_saferemove_snapshotdir_called_last(self):
        call_list = [str(call) for call in self.mock_manager.mock_calls]

        remove_call = [call for call in call_list if '_saferemove_snapshotdir' in call]
        self.assertEqual(len(remove_call), 1)

        remove_call_index = call_list.index(remove_call[0])
        self.assertEquals(remove_call_index, len(call_list)-1, 'snapshot dir remove call should be the last')

    def test_browser_closing_normally(self):
        self.prepare_snapshot_runner()

        self.snapshot_runner.run()
        self.perform_assertions()

        # snapshots are fetched later, the snapshot dir must be preserved
        self._saferemove_snapshotdir_mock.assert_not_called()

    def test_browser_closing_normally_for_thumbnail(self):
        self.prepare_snapshot_runner(is_thumbnail=True)

        self.snapshot_runner.run()
        self.perform_assertions()

        # thumbnails are saved to the DB on creation, the snapshot dir needs to be removed
        self._saferemove_snapshotdir_mock.assert_called_once()
        self.assert_saferemove_snapshotdir_called_last()

    def test_browser_closing_after_error(self):
        self.prepare_snapshot_runner()

        self.snapshot_runner.perform_snapshotting.side_effect = Exception(
            "mocked snapshotting error"
        )
        self.snapshot_runner.run()
        self.perform_assertions()

        # browser errors for snapshots are read from snaphotdir/shapshot.error
        # so the directory must not be deleted
        self._saferemove_snapshotdir_mock.assert_not_called()


    def test_browser_closing_after_error_for_thumbnail(self):
        self.prepare_snapshot_runner(is_thumbnail=True)

        self.snapshot_runner.perform_snapshotting.side_effect = Exception(
            "mocked snapshotting error"
        )
        self.snapshot_runner.run()
        self.perform_assertions()

        # we always remove the snapshotdir for thumbnails, even on error
        # noone will read the error file
        self._saferemove_snapshotdir_mock.assert_called_once()
        self.assert_saferemove_snapshotdir_called_last()

    def test_error_while_closing_browser(self):
        self.prepare_snapshot_runner()
        self.browser.close.side_effect = Exception("Unable to close - mocked error")
        self.snapshot_runner.run()
        self.perform_assertions()

        # if we ran in an error only during close, then we preserve the snapshot dir
        # (the thumbnail was created, so it can be fetched)
        self._saferemove_snapshotdir_mock.assert_not_called()

    def test_error_while_closing_browser_for_thumbnail(self):
        self.prepare_snapshot_runner(is_thumbnail=True)
        self.browser.close.side_effect = Exception("Unable to close - mocked error")
        self.snapshot_runner.run()
        self.perform_assertions()

        # for thumbnails we always delete the snapshot dir, no matter if success or error:
        self._saferemove_snapshotdir_mock.assert_called_once()
        self.assert_saferemove_snapshotdir_called_last()

    def _setup_runner_mocks(self, AsyncRunnerMock: MagicMock):
        """
        Create mocks for the two runners used for snapshotting.
        The runners are identified by their name
        """

        snapshot_mock = MagicMock()
        adhoc_snapshot_mock = MagicMock()
        csv_xslx_mock = MagicMock()

        def runner_factory(threadcount, name):
            if name == 'snapshot':
                return snapshot_mock
            elif name == 'adhoc.snapshot':
                return adhoc_snapshot_mock
            if name == 'csv_xlsx':
                return csv_xslx_mock
            else:
                raise Exception(f'unknown runner {name}')

        AsyncRunnerMock.side_effect = runner_factory
        return snapshot_mock, adhoc_snapshot_mock, csv_xslx_mock

    @patch.object(vs.vizlib.workq,'AsyncRunner')
    def test_ad_hoc_download(self, AsyncRunnerMock: MagicMock):
        admin1 = base.make_user('admin1', is_superuser=True)
        client1 = base.create_client(admin1)

        snapshot_mock, adhoc_snapshot_mock, csv_xslx_mock = self._setup_runner_mocks(AsyncRunnerMock)
        body_params = {
            'saveType': 'screenshot',
        }
        client1.post(f'/apps/snapshot/{self.app.id}', data=body_params)

        # three runner instances were created:
        self.assertEquals(AsyncRunnerMock.call_count, 3)

        # the first was not used:
        self.assertEquals(snapshot_mock.queue_untracked.call_count, 0)

        # the snapshot task was submitted to the second AsyncRunner
        self.assertEquals(adhoc_snapshot_mock.queue_untracked.call_count, 1)

    @patch.object(vs.vizlib.workq,'AsyncRunner')
    def test_thumbnail_creation(self, AsyncRunnerMock: MagicMock):
        admin1 = base.make_user('admin1', is_superuser=True)
        client1 = base.create_client(admin1)

        dc_id = helper.create_dataconnection(client1)
        ds_id = helper.create_dataset(client1, dc_id, 'test_dataset')
        visual_id = helper.create_visual(client1, ds_id)

        # creating a dashboard will create a thumbnail, too:
        snapshot_mock, adhoc_snapshot_mock, csv_xslx_mock = self._setup_runner_mocks(AsyncRunnerMock)
        helper.create_dashboard(client1, ds_id, [visual_id])

        # three runner instances were created:
        self.assertEquals(AsyncRunnerMock.call_count, 3)

        # the thumbnail task was submitted to the thumbnails worker:
        self.assertEquals(snapshot_mock.queue_untracked.call_count, 1)

        # the other worker was not used
        self.assertEquals(adhoc_snapshot_mock.queue_untracked.call_count, 0)

    @patch.object(vs,'logAction')
    def test_audit_logging(self, logger_mock: Mock):
        user1 = base.make_user('user1')
        req = create_request(user1, 'ip1')

        runner = vs.create_snapshotrunner(self.app.id, request=req, config={'uploadFormat': 'png'})
        runner.run_callback()

        logger_mock.assert_called_once()
        log_request = logger_mock.call_args.kwargs['request']
        self.assertEquals(log_request.user.username, 'user1')
        self.assertEquals(misc.get_request_source_ip(log_request), 'ip1')

        # test with explicitly provided user:
        user2 = base.make_user('user2')
        logger_mock.reset_mock()
        runner = vs.create_snapshotrunner(self.app.id, request=req, user=user2, config={'uploadFormat': 'png'})
        runner.run_callback()

        logger_mock.assert_called_once()
        log_request = logger_mock.call_args.kwargs['request']
        self.assertEquals(log_request.user.username, 'user2')
        self.assertEquals(misc.get_request_source_ip(log_request), 'ip1')



class TestThumbnailCreation(TestCase):
    def setUp(self):
        self.file_name = "test-image.png"

        img =vs.Image.new(mode = "RGB", size = (400, 300), color = (255, 0, 0) )
        img.save(self.file_name)

    def tearDown(self):
        os.remove(self.file_name)

    def test(self):
        thumbnail_bytes = vs.get_thumbnail_bytes(self.file_name)
        self.assertGreater(len(thumbnail_bytes),0)

class DownloadBase(TransactionTestCase):

  def setUp(self):
    super().setUp()
    self.ts_1 = '2024-11-20 16:30:00.123456'
    self.date_1 = '2024-11-30'
    self.mock_column_names = ['row_num', 'msg', 'ts_1', 'date_1']

    self.csv_excel_cache_result = setting_get_single('csv_excel_cache_result')
    self.visual_max_dataresults = setting_get_single('visual_max_dataresults')

  def tearDown(self):
    setting_update_single('csv_excel_cache_result', self.csv_excel_cache_result, False)
    setting_update_single('visual_max_dataresults', self.visual_max_dataresults, False)
    super().tearDown()

  @classmethod
  def tearDownClass(cls):
    # recreate the test DB because with TransactionTestCase all tables get truncated after each test case:
    from django.db import connection
    connection._connections['default'].creation.create_test_db(
      autoclobber=True)

    # clear the cache as it can disturb other tests
    # (visual_maxrows gets cached)
    from django.core.cache import cache
    cache.clear()

  def configure_mock_connection(self, conn: mock.Mock, is_iter: bool = True):
    mock_column_types = ['INT', 'STRING', 'TIMESTAMP', 'TIMESTAMP']

    mock_data=[
      [1, 'A', self.ts_1, None],
      [2, None, None, None],
      [None, 'C', None, self.date_1],
    ]
    if is_iter:
      return_value = iter([mock_column_types, self.mock_column_names] + mock_data)
    else:
      return_value = {'colnames': self.mock_column_names, 'coltypes': mock_column_types, 'rows': mock_data, 'info': ['Query:']} 
    conn.execute.return_value = return_value

  def assert_cached_info(self, cached_info: dict):

    self.assertIsNotNone(cached_info)

    self.assertListEqual(cached_info['colnames'], ['row_num', 'msg', 'ts_1', 'date_1'])
    self.assertListEqual(cached_info['coltypes'], ['INT', 'STRING', 'TIMESTAMP', 'TIMESTAMP'])

    expected_cached_rows=[
      f'[1,"A","{self.ts_1}",null]',
      f'[2,null,null,null]',
      f'[null,"C",null,"{self.date_1}"]'
    ]
    # the cached value is a JSON-serialized list of lists:
    expected_cached_rows_string = f'[{",".join(expected_cached_rows)}]'
    self.assertEquals(cached_info['rows'],expected_cached_rows_string)

class CsvDownloadBase(DownloadBase):
  __test__ = False

  def setUp(self):
    super().setUp()
    self.start_mocks()

  def tearDown(self):
    self.stop_mocks()
    super().tearDown()

  def create_mock(self, owner: object, property: str):
    patch =  mock.patch.object(owner, property)
    self.patches.append(patch)
    return patch.start()

  def start_mocks(self):
    pass

  def stop_mocks(self):
    for patch in self.patches:
      patch.stop()

  def assert_csv_content(self, actual_rows: list[str], has_format: bool):
    expected_rows = [
      'row_num,msg,ts_1,date_1',
      '1,A,2024-11-20 16:30:00.123456,',
      '2,,,',
    ]

    if has_format:
      # if there is a format settings list (even if it consists of nulls only)
      # dates are parsed as datetimes an are formatted as follows:
      str(datetime.fromisoformat(self.date_1))

      expected_rows.append(',C,,2024-11-30 00:00:00')
    else:
      expected_rows.append(',C,,2024-11-30')


    self.assertListEqual(actual_rows, expected_rows)

  def base_test_download_through_polling(self, should_use_format_settings: bool = True):
    """
    Instead of invoking _csvStream directly this method uses an HTTP POST and GET to generate a CSV
    """
    vs.PDF_OPTIONS = {'ENABLE_CHROMIUM_PDF': True}
    vs.init()
    vs.CHROMIUM_PDF_AVAILABLE = True
    admin = base.make_user('admin', is_superuser=True)
    admin_client = base.create_client(admin)
    util = test_report_operations.TestDashboardOperations()

    dc_id = util.create_dataconnection(admin_client)
    ds = util.create_dataset(admin_client, dc_id, 'ds_1')

    visual = util.create_visual(admin_client, ds)


    dsreq = DsreqCreator() \
      .datasetId(ds) \
      .getdsreq()


    format_settings = [None]*4
    filename = 'test.csv'
    params = {}
    params['query'] = '/* , , , , , , , , , , ,  */ select *'
    params['csvdownload'] = 'true'
    params['csvfilename'] = filename
    params['async'] = 'false'
    params['language_code'] = 'en_US'
    params['dataconnection_id'] = dc_id
    params['dsreq'] = json.dumps(dsreq)
    params['formattingSettings'] = json.dumps(format_settings)
    params['uploadFormat'] = 'csv'
    rsp = admin_client.post(f'/apps/snapshot/{visual}', data=params)
    assert(rsp.status_code == 200)

    snapshot_id = json.loads(rsp.content)['snapshotId']
    url = f'/apps/snapshot_ready/{snapshot_id}'
    try:
      response = poll_until_ready(admin_client, url, timeout=30, interval=1)
      self.assertIn('ready', response.content.decode('utf-8'))
    except TimeoutError:
      self.fail('Snapshot was not ready within the expected time')

    file_url = f'/apps/snapshot_fetch/{snapshot_id}?title={filename[:-4]}'
    file_response = admin_client.get(file_url)
    assert(file_response.status_code == 200)

    result_csv = tempfile.NamedTemporaryFile(dir='/tmp').name
    test_sqlrun.save_response_to_file(file_response, result_csv)

    with open(result_csv, mode='rt', encoding='utf-8') as f:
      csv_lines = f.readlines()

    # strip newlines:
    csv_lines = [line.strip() for line in csv_lines]
    has_format = format_settings is not None and should_use_format_settings
    self.assert_csv_content(csv_lines, has_format)

class TestCsvDownloadStreaming(CsvDownloadBase):
  def start_mocks(self):
    self.patches = list[mock._patch]()
    css_mock = self.create_mock(views,'connectionSupportsStreaming')
    css_mock.return_value = True

    self.get_dataconn_mock = self.create_mock(views,'get_dataconn')
    self.configure_mock_connection(self.get_dataconn_mock.return_value)

  def test_download_through_polling(self):
    self.base_test_download_through_polling()

class TestCsvDownloadWithoutStreaming(CsvDownloadBase):
  def start_mocks(self):
    self.patches = list[mock._patch]()
    css_mock = self.create_mock(views,'connectionSupportsStreaming')
    css_mock.return_value = False

    self.get_dataconn_mock = self.create_mock(views,'get_dataconn')
    self.configure_mock_connection(self.get_dataconn_mock.return_value, is_iter=False)

  def test_download_through_polling(self):
    self.base_test_download_through_polling(should_use_format_settings=False)

def create_helper():
  helper = tests_rewrites.Test_Rewrites()
  helper.supports_dataextract_source = True
  helper.supports_partitions = True
  helper.supports_partitions_by_id = {}
  helper.supports_dataextract_target = True

  return helper

def get_xlsx_content(filename: str):
  book = xlrd.open_workbook(filename)
  sheet = book.sheet_by_index(0)

  def cell_value(cell: xlrd.sheet.Cell):
    if cell.ctype == xlrd.XL_CELL_EMPTY:
      return None
    if cell.ctype == xlrd.XL_CELL_DATE:
      return xlrd.xldate.xldate_as_datetime(cell.value, book.datemode)

    return cell.value

  for row in sheet.get_rows():
    yield [cell_value(cell) for cell in row]

class XlsxDownloadBase(DownloadBase):
  __test__ = False

  def setUp(self):
    super().setUp()
    self.start_mocks()

  def tearDown(self):
    self.stop_mocks()
    super().tearDown()

  def create_mock(self, owner: object, property: str):
    patch =  mock.patch.object(owner, property)
    self.patches.append(patch)
    return patch.start()

  def start_mocks(self):
    pass

  def stop_mocks(self):
    for patch in self.patches:
      patch.stop()

  def assert_xlsx_content(self, actual_content: list[list[object]]):
    # 1. excel stores dates as floats. On request, xlrd can convert them to datetimes.
    # 2. datetime precision is milliseconds (microseconds are lost)
    # 3. numbers are stored as floats (there is no int type)
    expected_content = [
      self.mock_column_names,
      [1.0, 'A', datetime.fromisoformat('2024-11-20 16:30:00.123000'), None],
      [2.0, None, None, None],
      [None, 'C', None, datetime.fromisoformat(self.date_1)],
    ]
    self.assertListEqual(actual_content, expected_content)

  def do_test_download(self):
    """
    Instead of invoking _csvStream directly this method uses an HTTP POST and GET to generate a CSV
    """
    setting_update_single('csv_excel_cache_result', 'false', False)
    vs.PDF_OPTIONS = {'ENABLE_CHROMIUM_PDF': True}
    vs.init()
    vs.CHROMIUM_PDF_AVAILABLE = True
    admin = base.make_user('admin', is_superuser=True)
    admin_client = base.create_client(admin)
    util = test_report_operations.TestDashboardOperations()

    dc_id = util.create_dataconnection(admin_client)
    ds = util.create_dataset(admin_client, dc_id, 'ds_1')

    visual = util.create_visual(admin_client, ds)

    dsreq = DsreqCreator() \
      .datasetId(ds) \
      .getdsreq()


    format_settings = [None]*4
    params = {}
    filename = 'test.xlsx'
    params['query'] = '/* , , , , , , , , , , ,  */ select *'
    params['xlsxdownload'] = 'true'
    params['xlsxfilename'] = filename
    params['async'] = 'false'
    params['language_code'] = 'en_US'
    params['dsreq'] = json.dumps(dsreq)
    params['formattingSettings'] = json.dumps(format_settings)
    params['uploadFormat'] = 'xlsx'
    params['dataconnection_id'] = dc_id
    rsp = admin_client.post(f'/apps/snapshot/{visual}', data=params)
    assert(rsp.status_code == 200)

    snapshot_id = json.loads(rsp.content)['snapshotId']
    url = f'/apps/snapshot_ready/{snapshot_id}'
    try:
      response = poll_until_ready(admin_client, url, timeout=60, interval=1)
      self.assertIn('ready', response.content.decode('utf-8'))
    except TimeoutError:
      self.fail('Snapshot was not ready within the expected time')

    file_url = f'/apps/snapshot_fetch/{snapshot_id}?title={filename[:-5]}'
    file_response = admin_client.get(file_url)
    assert(file_response.status_code == 200)

    result_file = tempfile.NamedTemporaryFile(dir='/tmp').name
    test_sqlrun.save_response_to_file(file_response, result_file)

    lines = list(get_xlsx_content(result_file))
    self.assert_xlsx_content(lines)

class TestXlsxDownloadStreaming(XlsxDownloadBase):
  def start_mocks(self):
    self.patches = list[mock._patch]()
    css_mock = self.create_mock(views,'connectionSupportsStreaming')
    css_mock.return_value = True

    self.get_dataconn_mock = self.create_mock(views,'get_dataconn')
    self.configure_mock_connection(self.get_dataconn_mock.return_value)

  def test_download_default_caching(self):
    self.do_test_download()

class TestXlsxDownloadWithoutStreaming(XlsxDownloadBase):
  def start_mocks(self):
    self.patches = list[mock._patch]()
    css_mock = self.create_mock(views,'connectionSupportsStreaming')
    css_mock.return_value = False

    self.get_dataconn_mock = self.create_mock(views,'get_dataconn')
    self.configure_mock_connection(self.get_dataconn_mock.return_value, is_iter=False)

  def test_download_default_caching(self):
    self.do_test_download()
