# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from django.contrib.auth.models import User
from django.test import TestCase
from django.test import Client
from datasets.models import DataSet, Extract
from rbac.models import Role
from vizlib.tests import test_report_operations
from vizlib.tests.test_report_operations import make_user, create_client
from vizlib.tests.util import run_with_db_query_check, get_db_queries
from workspace.models import Workspace


import json

def grant_read_access_on_datasets(user: User, dc_id: int):
  # -1 means all datasets:
  grant_read_access_on_dataset(user, dc_id, -1)

def grant_read_access_on_dataset(user: User, dc_id: int, ds_id: int):
  role = Role(name=f'read_access_on_{ds_id}')
  privs = [
    # no system permissions:
    {'ptype': 'system', 'perms': []},
    # full rights on the dataset
    {
      'ptype': 'dataset',
      'dcid': str(dc_id),
      'dslist': [str(ds_id)],
      'perms': ['ds_appview'],
    },
  ]
  role.privileges = json.dumps(privs)

  # add_users automatically saves:
  role.add_users([user.username])

def grant_full_access_on_dataset(user: User, dc_id: int, ds_id: int):
  role = Role(name=f'full_access_on_{ds_id}')
  privs = [
    # no system permissions:
    {'ptype': 'system', 'perms': []},
    # full rights on the dataset
    {
      'ptype': 'dataset',
      'dcid': str(dc_id),
      'dslist': [str(ds_id)],
      'perms': ['ds_manage', 'ds_appview', 'ds_appedit'],
    },
  ]
  role.privileges = json.dumps(privs)

  # add_users automatically saves:
  role.add_users([user.username])


class TestReportCountPerGroup(TestCase):
  def setUp(self):
    self.common_setup()

  def common_setup(self):
    util = test_report_operations.TestDashboardOperations()
    self.util = util

    # create admin
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    self.admin = admin
    self.admin_client = admin_client

    # create user1
    user1 = make_user('user1', is_superuser=False)
    user1_client = create_client(user1)
    self.user1 = user1
    self.user1_client = user1_client

    dc_id1 = util.create_dataconnection(admin_client)
    self.dc_id1 = dc_id1

    ds1_id = util.create_dataset(admin_client, dc_id1, 'ds1')
    self.ds1_id = ds1_id
    # convert this dataset to multitable:
    ds1 = DataSet.objects.get(id=ds1_id)
    ds1.dataset_type = 'multitable'
    ds1.save()

    ds2_id = util.create_dataset(admin_client, dc_id1, 'ds2')
    self.ds2_id = ds2_id

    grant_full_access_on_dataset(user1, dc_id1, ds1_id)

    self.create_dashboard(admin_client, ds1_id)
    self.create_dashboard(admin_client, ds1_id, public=True)

    self.create_dashboard(admin_client, ds2_id)
    self.create_dashboard(admin_client, ds2_id, public=True)

    self.create_dashboard(user1_client, ds1_id)
    self.create_dashboard(user1_client, ds1_id, public=True)

    extract_1 = Extract.objects.create(
      srcDataset_id=ds1_id, tgtDataconnection_id=dc_id1, extract_config={}, extract_runstate={})
    extract_2 = Extract.objects.create(
      srcDataset_id=ds2_id, tgtDataconnection_id=dc_id1, extract_config={}, extract_runstate={})

  def create_dashboard(self, client: Client, ds_id: int, public=False):
    visual_id = self.util.create_visual(client, ds_id)

    dashboard_id = self.util.create_dashboard(
      client, dataset_id=ds_id, visuals=[visual_id]
    )

    if public:
      self.util.move_to_public_workspace(client, dashboard_id=dashboard_id)

  def get_workspace_map(self):
    """
    Returns a workspace name to workspace id map.
    Private workspace keys are created as "Private_<username>"
    """
    result = dict[str, str]()
    for ws in Workspace.objects.select_related('private_user').all():
      key = ws.name
      if ws.private_user:
        key += f'_{ws.private_user.username}'
      result[key] = str(ws.id)

    return result

  def test_workspace_count_admin(self):
    workspaces = self.get_workspace_map()

    # first run is terrible, it creates reports_appgroupappmembership entries, query count is 33
    self.admin_client.get('/apps/get_report_count_per_group')

    response, _ = run_with_db_query_check(
      lambda: self.admin_client.get('/apps/get_report_count_per_group'),
      asserter=self,
      expected_query_count=8
      )
    self.assertEquals(response.status_code, 200)
    response_json = response.json()

    # admin created 4 dashboards
    self.assertEqual(4, response_json['favorites_report_count'])
    ws_counts = response_json['workspace_id_to_report_count']
    self.assertEqual(len(ws_counts.keys()), 3)  # access to all workspaces
    self.assertEqual(int(ws_counts[workspaces['Public']]), 3)
    self.assertEqual(int(ws_counts[workspaces['Private_admin']]), 2)
    self.assertEqual(int(ws_counts[workspaces['Private_user1']]), 1)

    queries = ';\n'.join(get_db_queries())
    # make sure we do not download dataset_info
    self.assertFalse('dataset_info' in queries, 'dataset_info is retrieved')

  def test_workspace_count_user(self):
    workspaces = self.get_workspace_map()

    # first run is terrible, it creates reports_appgroupappmembership entries, query count is 31
    self.user1_client.get('/apps/get_report_count_per_group')

    response, _ = run_with_db_query_check(
      lambda: self.user1_client.get('/apps/get_report_count_per_group'),
      asserter=self,
      expected_query_count=9 + 3,
      )
    self.assertEquals(response.status_code, 200)
    response_json = response.json()

    # user1 created 2 dashboards
    self.assertEqual(2, response_json['favorites_report_count'])
    ws_counts = response_json['workspace_id_to_report_count']
    self.assertEqual(len(ws_counts.keys()), 2)  # access to 2 workspaces
    # one dashboard is missing as it is on ds_2
    self.assertEqual(int(ws_counts[workspaces['Public']]), 2)
    self.assertEqual(int(ws_counts[workspaces['Private_user1']]), 1)

    queries = ';\n'.join(get_db_queries())
    # make sure we do not download dataset_info
    self.assertFalse('dataset_info' in queries, 'dataset_info is retrieved')
