# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from urllib.parse import urlencode
from django.test import Client
from parameterized import parameterized
from django.contrib.auth.models import User

from apps.tests.userrole.base import (
    TestBase,
    make_group,
    make_role,
    make_user,
    new_group_payload,
    new_user_payload,
    role_payload,
    update_user_payload,
)
from apps.tests.userrole.permission_matrix import create_test_parameters
from rbac.models import Group, Role


def msg_users(action, target_user_is_admin, acting_user):
    return (
        f"{action} {'superuser' if target_user_is_admin else 'user'} by {acting_user}"
    )


def msg_users_update(detail, target_user_is_admin, acting_user):
    return f"update {'superuser' if target_user_is_admin else 'user'} {detail} by {acting_user}"


def msg_groups(action, acting_user):
    return f"{action} groups by {acting_user}"


def msg_roles(action, acting_user):
    return f"{action} roles by {acting_user}"


class TestPermissionMatrix(TestBase):
    @parameterized.expand(create_test_parameters())
    def test(self, test_method_name, params):
        method = getattr(self, test_method_name)
        if method:
            method(*params)
        else:
            self.fail(f"no test method {test_method_name}")

    def _test_users_create(self, is_admin: bool, user: str, expected_status: int):
        client = self.clients[user]
        new_user = f"user_created_by_{user}"
        user_data = new_user_payload(new_user, is_admin=is_admin)

        response = client.post(f"/apps/users_api/{new_user}", user_data)

        msg = msg_users("create", is_admin, user)
        self.assertEquals(response.status_code, expected_status, msg)
        if expected_status == 200:
            self.assertIsNotNone(User.objects.get(username=new_user), msg)

    def get_target_user(self, is_admin: bool):
        return make_user("test_user", is_superuser=is_admin)

    def _put(self, client: Client, user: User, data):
        return client.put(
            f"/apps/users_api/{user.username}",
            urlencode(data),
            content_type="application/x-www-form-urlencoded",
        )

    def _update_text_attribute(
        self, attr: str, is_admin: bool, user: str, expected_status: int
    ):
        client = self.clients[user]

        target_user = self.get_target_user(is_admin=is_admin)
        data = update_user_payload(target_user)
        new_atrr = f"{attr}_udpated_by_{user}"
        data[attr] = new_atrr

        response = self._put(client, target_user, data)

        msg = msg_users_update(attr, is_admin, user)
        self.assertEquals(response.status_code, expected_status, msg)
        if expected_status == 200:
            target_user.refresh_from_db()
            self.assertEqual(getattr(target_user, attr), new_atrr, msg)

    def _update_bool_attribute(
        self, attr: str, is_admin: bool, user: str, expected_status: int
    ):
        client = self.clients[user]

        target_user = self.get_target_user(is_admin=is_admin)
        data = update_user_payload(target_user)
        new_attr = not getattr(target_user, attr)
        data[attr] = "true" if new_attr else "false"

        response = self._put(client, target_user, data)

        msg = msg_users_update(attr, is_admin, user)
        self.assertEquals(response.status_code, expected_status, msg)
        if expected_status == 200:
            target_user.refresh_from_db()
            self.assertEqual(getattr(target_user, attr), new_attr, msg)

    def _test_users_update_username(
        self, is_admin: bool, user: str, expected_status: int
    ):
        self._update_text_attribute("username", is_admin, user, expected_status)

    def _test_users_update_first_name(
        self, is_admin: bool, user: str, expected_status: int
    ):
        self._update_text_attribute("first_name", is_admin, user, expected_status)

    def _test_users_update_last_name(
        self, is_admin: bool, user: str, expected_status: int
    ):
        self._update_text_attribute("last_name", is_admin, user, expected_status)

    def _test_users_update_is_superuser(
        self, is_admin: bool, user: str, expected_status: int
    ):
        self._update_bool_attribute("is_superuser", is_admin, user, expected_status)

    def _test_users_update_groups(
        self, is_admin: bool, user: str, expected_status: int
    ):
        client = self.clients[user]

        target_user = self.get_target_user(is_admin=is_admin)
        data = update_user_payload(target_user)

        group = make_group(f"test_group_to_use_by_{user}")
        data["groups"] = [group.id]

        response = self._put(client, target_user, data)

        msg = msg_users_update("groups", is_admin, user)
        self.assertEquals(response.status_code, expected_status, msg)
        if expected_status == 200:
            target_user.refresh_from_db()
            actual_groups = [g.id for g in target_user.groups.all()]
            self.assertEqual(
                actual_groups,
                [group.id],
                msg,
            )

    def _test_users_update_roles(self, is_admin: bool, user: str, expected_status: int):
        client = self.clients[user]

        target_user = self.get_target_user(is_admin=is_admin)
        data = update_user_payload(target_user)

        role = make_role(f"test_group_to_use_by_{user}", [])
        data["roles"] = [role.id]

        response = self._put(client, target_user, data)

        msg = msg_users_update("roles", is_admin, user)
        self.assertEquals(response.status_code, expected_status, msg)
        if expected_status == 200:
            role.refresh_from_db()
            self.assertEqual(
                role.userSet(),
                {
                    target_user.username,
                },
                msg,
            )

    def _test_users_update_is_active(
        self, is_admin: bool, user: str, expected_status: int
    ):
        self._update_bool_attribute("is_active", is_admin, user, expected_status)

    def _test_users_update_password(
        self, is_admin: bool, user: str, expected_status: int, old_password=None
    ):
        client = self.clients[user]

        if old_password:
            # scenario of updating own password:
            target_user = self.users[user]
            target_user_info = "self"
        else:
            target_user = self.get_target_user(is_admin=is_admin)
            target_user_info = f"is_admin={is_admin}"

        data = update_user_payload(target_user)
        data["new_password"] = "new_password"
        if old_password:
            data["old_password"] = old_password
            data["confirm_password"] = "new_password"

        old_password = target_user.password

        response = self._put(client, target_user, data)

        msg = msg_users_update("password", is_admin, user)
        if old_password:
            msg = msg_users_update("own password", is_admin, user)

        self.assertEquals(
            response.status_code,
            expected_status,
            msg,
        )
        if expected_status == 200:
            target_user.refresh_from_db()
            self.assertNotEqual(
                target_user.password,
                old_password,
                msg,
            )

    @parameterized.expand(
        [
            ["normal", 200],
            ["manager", 200],
            ["admin", 200],
        ]
    )
    def test_users_update_own_password(self, acting_user_type, expected_status):
        acting_user = self.users[acting_user_type]
        old_password = acting_user.username  # users have their username as password
        is_admin = False  # does not matter when we change the acting user's password
        # but needed for the parameterized _test_users_update_password method
        self._test_users_update_password(
            is_admin, acting_user_type, expected_status, old_password
        )

    def _test_users_delete(self, is_admin: bool, user: str, expected_status: int):
        test_user = make_user("test_normal_user", is_superuser=is_admin)
        client = self.clients[user]
        response = client.delete(f"/apps/users_api/{test_user.username}")

        msg = msg_users("delete", is_admin, user)
        self.assertEquals(response.status_code, expected_status, msg)

        if response.status_code == 200:
            count = len(User.objects.filter(id=test_user.id).all())
            self.assertEquals(count, 0, msg)

    def _test_groups_create(self, user: str, expected_status: int):
        client = self.clients[user]
        group_name = f"group_created_by_{user}"
        group_data = new_group_payload(group_name)

        response = client.post(f"/apps/groups_api", group_data)

        msg = msg_groups("create", user)
        self.assertEquals(response.status_code, expected_status, msg)

        if response.status_code == 200:
            group = Group.objects.get(name=group_name)
            self.assertIsNotNone(group, msg)

    def _test_groups_update(self, user: str, expected_status: int):
        client = self.clients[user]
        group_to_update = make_group("test_group")
        new_group_name = f"group_updated_by_{user}"
        payload = new_group_payload(new_group_name)

        response = client.put(
            f"/apps/groups_api/{group_to_update.id}",
            urlencode(payload),
            content_type="application/x-www-form-urlencoded",
        )

        msg = msg_groups("update", user)
        self.assertEquals(response.status_code, expected_status, msg)

        if response.status_code == 200:
            group_to_update.refresh_from_db()
            self.assertEquals(group_to_update.name, new_group_name, msg)

    def _test_groups_delete(self, user: str, expected_status: int):
        client = self.clients[user]

        group_to_delete = make_group("test_group")

        response = client.delete(f"/apps/groups_api/{group_to_delete.id}")

        msg = msg_groups("delete", user)
        self.assertEquals(response.status_code, expected_status, msg)

        if response.status_code == 200:
            role_count = len(Group.objects.filter(id=group_to_delete.id).all())
            self.assertEquals(role_count, 0, msg)

    def _test_roles_create(self, user: str, expected_status: int):
        client = self.clients[user]
        role_name = f"role_created_by_{user}"
        role_data = role_payload(role_name)

        response = client.post(f"/roles/roles", role_data)

        msg = msg_roles("create", user)
        self.assertEquals(response.status_code, expected_status, msg)

        if response.status_code == 200:
            role = Role.objects.get(name=role_name)
            self.assertIsNotNone(role, msg)

    def _test_roles_update(self, user: str, expected_status: int):
        client = self.clients[user]

        role_to_update = make_role("role_to_update", [])
        new_role_name = f"role_updated_by_{user}"

        payload = role_payload(new_role_name, id=role_to_update.id)

        response = client.post(f"/roles/roles", payload)

        msg = msg_roles("update", user)
        self.assertEquals(response.status_code, expected_status, msg)

        if response.status_code == 200:
            role_to_update.refresh_from_db()
            self.assertEquals(role_to_update.name, new_role_name, msg)

    def _test_roles_delete(self, user: str, expected_status: int):
        role_to_delete = make_role("role_to_delete", [])

        response = self.clients[user].delete(f"/roles/roles/{role_to_delete.id}")

        msg = msg_roles("delete", user)
        self.assertEquals(response.status_code, expected_status, msg)

        if response.status_code == 200:
            role_count = len(Role.objects.filter(id=role_to_delete.id).all())
            self.assertEquals(role_count, 0, msg)
