# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from django.contrib.auth.models import AnonymousUser, User
import sys, json, re

from apps.config import product_name, release_notes_url, documentation_base_url
import reports

from rbac.permdefs import PERMDEFS, PERMLISTS
from reports.views_settings import use_derived_data, setting_get_globals
from django.conf import settings
from rbac.permcache import RBAC
from django.template.context_processors import csrf
from reports.models import AppSetting
from dataconn import dataconn_http
from util.ff_flipper import FeatureFlipper
from datasets.models import DataConnection


def get_should_change_password(user: User):
  if user.username == settings.AUTOLOGIN_USERNAME:
    return False

  is_anonymous_user = isinstance(user, AnonymousUser)

  if is_anonymous_user:
    return False

  return user.userprofile.should_change_password

WHATSNEWVER = '8.0.1'

def get_version():
    try:
      from version import version
      v = version.get_version_info()
      return [v["Arcviz Version"], v["Raw Timestamp of last commit"], v["SHA of last commit"]]
    except:
      return ["unknown", "unknown", "unknown"]

def _instant_flags(request):
  v = get_version()
  c = {
    'version': v[0].split('_')[0],
    'trimmed_version': v[0],
    'timestamp': v[1],
    'platform': sys.platform,
  }
  userid = reports.views_settings.registered_userid()
  if userid:
    c['userid'] = userid + "." + str(request.user.id)
    c['instantid'] = userid

  c["product_name"] = product_name
  c["full_version"] = str(c['version']) + "." + str(c['timestamp'])
  c['release_notes_url'] = release_notes_url
  c['documentation_base_url'] = documentation_base_url
  return c

# The 'styles=off' URL parameter turns off custom CSS and javascript
# This is expected to be very useful for debugging customer issues
def _styleflags(request):
  s = 'on'
  if request.GET:
    s = request.GET.get('styles', 'on').lower()
    if s == "false":
      s = "off"
    if s not in ['on', 'off']:
      s = 'on'

  c = {
    'styles': s
  }

  # Default is enabled
  enable_explore_interfaces = reports.views_settings.setting_get_single(
      'site_enable_explore_interfaces')
  if enable_explore_interfaces != "false":
    enable_explore_interfaces = "true"
  c['enable_explore_interfaces'] = enable_explore_interfaces

  c['max_trellis_displayed'] =\
    reports.views_settings.setting_get_single('site_max_trellis_displayed')

  # Default is disabled
  enable_visual_transitions = reports.views_settings.setting_get_single(
      'site_enable_visual_transitions')
  if enable_visual_transitions != "true":
    enable_visual_transitions = "false"
  c['enable_visual_transitions'] = enable_visual_transitions

  enable_styles = reports.views_settings.setting_get_single('site_enable_custom_styles')
  if enable_styles != "false":
    enable_styles = "true"
  c['enable_styles'] = enable_styles

  if getattr(settings,"ENABLE_SCRIPT_PARAMETERS", False):
    c['enable_script_parameters'] = 'true'
  else:
    c['enable_script_parameters'] = 'false'

  if getattr(settings,"ENABLE_TOOLTIP_HIGHLIGHTS", False):
    c['enable_tooltip_highlights'] = 'true'
  else:
    c['enable_tooltip_highlights'] = 'false'

  return c

def _thumbnail_flags():
  c = {}

  enable_thumbnails = reports.views_settings.setting_get_single(
    'site_thumbnails_enable')
  if enable_thumbnails != 'true':
    enable_thumbnails = 'false'
  c['enable_thumbnails'] = enable_thumbnails

  return c

# Return the permission definitions along with the requesting
# user's system level permissions
def _permstuff(request):
  c = {}
  c['everyone_group'] = settings.EVERYONE_GROUP
  c['permdefs'] = json.dumps(PERMDEFS)
  c['permlists'] = json.dumps(PERMLISTS)
  c['sysperms'] = json.dumps(list(RBAC.get_user_obj_perms(request.user)))
  return c

# Check if request is using trustedauth in order to disable license banner
def _trustedauth_flags(request):
  return {"trustedauth": request.path.startswith("/arc/trustedauth")}

def _is_mobile_browser(request):
  # Mobile browser detection using HTTP_USER_AGENT
  # From: https://djangosnippets.org/snippets/2001/
  is_mobile = False

  if 'HTTP_USER_AGENT' in request.META:
    user_agent = request.META['HTTP_USER_AGENT']

    # Test common mobile values.
    pattern = "(up.browser|up.link|mmp|symbian|smartphone|midp|wap|phone|windows ce|pda|mobile|mini|palm|netfront)"
    prog = re.compile(pattern, re.IGNORECASE)
    match = prog.search(user_agent)

    if match:
      is_mobile = True
    else:
      # Nokia like test for WAP browsers.
      # http://www.developershome.com/wap/xhtmlmp/xhtml_mp_tutorial.asp?page=mimeTypesFileExtension
      if 'HTTP_ACCEPT' in request.META:
        http_accept = request.META['HTTP_ACCEPT']
        pattern = "application/vnd\.wap\.xhtml\+xml"
        prog = re.compile(pattern, re.IGNORECASE)
        match = prog.search(http_accept)
        if match:
          is_mobile = True

    if not is_mobile:
      # Now we test the user_agent from a big list.
      user_agents_test = ("w3c ", "acs-", "alav", "alca", "amoi", "audi",
                          "avan", "benq", "bird", "blac", "blaz", "brew",
                          "cell", "cldc", "cmd-", "dang", "doco", "eric",
                          "hipt", "inno", "ipaq", "java", "jigs", "kddi",
                          "keji", "leno", "lg-c", "lg-d", "lg-g", "lge-",
                          "maui", "maxo", "midp", "mits", "mmef", "mobi",
                          "mot-", "moto", "mwbp", "nec-", "newt", "noki",
                          "xda", "palm", "pana", "pant", "phil", "play",
                          "port", "prox", "qwap", "sage", "sams", "sany",
                          "sch-", "sec-", "send", "seri", "sgh-", "shar",
                          "sie-", "siem", "smal", "smar", "sony", "sph-",
                          "symb", "t-mo", "teli", "tim-", "tosh", "tsm-",
                          "upg1", "upsi", "vk-v", "voda", "wap-", "wapa",
                          "wapi", "wapp", "wapr", "webc", "winw", "winw",
                          "xda-",)

      test = user_agent[0:4].lower()
      if test in user_agents_test:
        is_mobile = True
  return is_mobile

def _getMobile(request):
  # FIXME is there any performance penalty by calling this function inside _appflags
  mobilearg = request.GET.get("mobile")
  if mobilearg == "true":
    return True
  if mobilearg == "false":
    return False

  return _is_mobile_browser(request)

def can_user_explore_any_dataconnections(request):
  user_can_explore_any_dataconnections = False
  if FeatureFlipper().get_flipper_value('enable_direct_access') and request.user is not None:
    for dc in DataConnection.objects.all():
      if 'dc_explore' in dc.getPerms(request.user):
        user_can_explore_any_dataconnections = True
        break
  return user_can_explore_any_dataconnections


# App-rendering flags. Used for all templates where apps or visuals are being
# rendered.
def _appflags(request):
  from dataconn.ext.main import ui_configs
  c = {}
  c.update(_styleflags(request))
  c.update(_instant_flags(request))
  c.update(_permstuff(request))
  c.update(_thumbnail_flags())
  c.update(_trustedauth_flags(request))
  c.update(setting_get_globals())
  c.update(csrf(request))
  c['is_dashboard'] = False
  c['send_message_params'] = False
  c['embed'] = False
  c['on_homepage'] = False
  c['email_addresses'] = False
  c['whats_new_version'] = WHATSNEWVER

  c['disable_homepage_override'] = False
  try:
    c['disable_homepage_override'] = AppSetting.objects.get(setting_key='disable_homepage_override').setting_value == 'True'
  except:
    pass

  c['mobile'] = _getMobile(request)
  c['auto_refresh'] = False
  c['use_derived_data'] = use_derived_data(request.user.is_superuser)

  c['show_release_modal'] = False
  c['http_dataconnection_id'] = dataconn_http.INSTANCE.id
  c['user_can_explore_any_dataconnections'] = can_user_explore_any_dataconnections(request)
  c['incompatible_browser_warning_dismissed'] = False
  if request.user:
    userdata = json.loads(request.user.userprofile.user_data)
    c['new_release_modal_seen'] = userdata.get('new_release_modal_seen', None) == WHATSNEWVER
    c['incompatible_browser_warning_dismissed'] = userdata.get('incompatible_browser_warning_dismissed', False)

  c['dc_ext_list'] = ui_configs()
  return c
