# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
views for apps
"""

import os, sys, json, re, urllib.request, urllib.parse, urllib.error, logging, errno

from django.http import HttpResponse, HttpResponseRedirect, QueryDict, JsonResponse, HttpRequest
from django.template import loader
from django.contrib.auth import authenticate, login, logout, update_session_auth_hash
from django.template.context_processors import csrf
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.contrib.auth.models import Group
from django.core.paginator import Paginator

from django.conf import settings
from django.shortcuts import redirect, render
from django.core.cache import cache
from django.views.defaults import page_not_found

from django.db.models import Q
from urllib.parse import parse_qs
from util.request import create_request

import reports
from datasets.models import DataSet, DataConnection
import datasets.views
from apps.models import UserProfile
from apps import views_visuals_page
from util.ff_flipper import FeatureFlipper
from vizlib.querylog import logAction, ComponentType, ActionType, auditLogExists
from vizlib.django_sql import GroupConcat, GroupConcatEnabled
from util.json_utils import json_dumps
from vizlib.accel import get_existing_tables

from datasources.forms import DataSourceForm
from datasources.dataimport import DataImport

from migration.form import ImportUploadFileForm

from rbac.permdefs import PermissionException, PERMDEFS, PERMLISTS
from rbac.permcache import RBAC

from reports.views_settings import use_derived_data, setting_get_globals, get_app_name
from reports.views_appgroups import _auditAppGroup

from reports.models import AppGroup, AppSetting, AppGroupAppMembership, Report, ENABLE_URL_NAMES
from reports.views import report_get_internal

from arcwebbase.validator import validateStringNoHtml, validateUsernameCharacters, validatePassword

from operator import itemgetter

from jobs.models import _SCHEDULER_ENABLED, EmailJobContent, CustomEmailTemplate

from jobs.views_schedulegroups import schedule_group_get

from jobs.views_encryption import encrypt_secret

from dataconn import dataconn_http
from dataconn.connections_types import ALL_CONNECTION_TYPES
from dataconn.dataconn import get_dataconn
from dataconn.data_api_cmds import SPECIAL_CMDS

from django.views import View
from django.utils.decorators import method_decorator

from apps.utils import get_should_change_password, WHATSNEWVER, _appflags, _getMobile

if settings.AXES_ENABLED:
  from axes.decorators import axes_dispatch
else:
  def axes_dispatch(inner):
    return inner

from django.contrib.auth.signals import user_logged_in, user_login_failed
from .views_snapshot import _PIL_ENABLED

from .views_static_model_helper import load_static_dashboard

import uuid

LOGGER = logging.getLogger('arcweb')

#
# Implements the AUTOLOGIN logic as a decorator function
#
# In some cases, the viz platform is deployed in a mode where you want to
# give unauthenticated users the ability to interact with the platform.
# In these situations, you create a user with limited privileges, and then
# login that user automatically when certain entry points are accessed
#
def autologin(f):
  def w(*args, **kwargs):
    request = args[0]

    if not request.user.is_authenticated and settings.AUTOLOGIN_ENABLED:
      try:
        username = settings.AUTOLOGIN_USERNAME
        password = settings.AUTOLOGIN_PASSWORD
        user = authenticate(username=username, password=password)
        if user is not None:
          if user.is_active:
            login(request, user)
            logged_in = True
        else:
          LOGGER.error('AUTOLOGIN failed')
      except Exception as e:
        LOGGER.exception(e)
        pass

    return f(*args, **kwargs)

  return w

def _auditUser(user, action, request=None, username=None):
  logAction(action=action, component_type=ComponentType.USER,
      component_id=user.id, request=request, username=username)

@method_decorator(axes_dispatch, name='dispatch')
class LogIn(View):
  def get(self, request):
    template = loader.get_template('login.html')
    c = {}
    c['is_registered'] = reports.views_settings.is_registered()
    next = request.GET.get('next', '')
    querystring = urllib.parse.urlencode({k: v for k, v in request.GET.items() if k != 'next'})
    c['next'] = urllib.parse.urlunparse(['','',next,'', querystring, ''])
    c['use_logout_page'] = False
    c['failure_limit'] = ''
    c['username'] = ''
    c['querystring'] = ''
    c['error_msg'] = ''
    c.update(csrf(request))
    c['login_instructions'] = reports.views_settings.setting_get_single('login_instructions')
    return HttpResponse(template.render(c, request))

  def post(self, request):
    """Login form"""
    username = request.POST.get('username', None)
    password = request.POST.get('password', None)
    querystring = request.GET.copy()
    querystring.pop('next', None)
    template = loader.get_template('login.html')
    c = {}
    c['is_registered'] = reports.views_settings.is_registered()
    c['next'] = request.POST.get('next', '')
    c['username'] = username
    c['use_logout_page'] = False
    c['querystring'] = querystring.urlencode()
    c['error_msg'] = ''
    c['failure_limit'] = ''
    c.update(csrf(request))
    if username is None or password is None:
      c['error_msg'] = ''
      return HttpResponse(template.render(c, request))

    #check if the user is disabled
    user = User.objects.filter(username=username).first()
    if user is not None and not user.is_active:
      c['error_msg'] = 'Your account is disabled.'
      return HttpResponse(template.render(c, request))

    user = authenticate(request=request, username=username, password=password)
    if user is not None:
      login(request, user)
      user_logged_in.send(
        sender=User,
        request=request,
        user=user,
      )
      next_url = request.POST.get('next', '')
      if next_url and re.search(r'^/[^/].*', next_url):
        # if next_url is not empty and contains only one slash. Having
        # two slashes could be used to redirect the client to a third party domain
        # which is considered a security risk
        return HttpResponseRedirect(next_url)
      return HttpResponseRedirect(getattr(settings, 'HOME_URL',
                                          "%s/apps/" % settings.ARC_BASEURL))
    else:
      user_login_failed.send(
        sender=User,
        request=request,
        credentials={
          'username': username
        },
      )
      c['error_msg'] = 'Invalid login'
      return HttpResponse(template.render(c, request))

def dologout(request):
  """Logout"""

  imported_data = request.session.get('arc_import_data', None)
  if imported_data:
    # clean up any local files or data related to an import on logout
    ds = DataImport(imported_data['datasource_uploaded_file'])
    ds.remove_uploaded_file()

  if getattr(settings,"PRE_LOGOUT_REDIRECT_URL", False):
    return HttpResponseRedirect(settings.PRE_LOGOUT_REDIRECT_URL)

  logout(request)

  if getattr(settings,"LOGOUT_REDIRECT_URL", False):
    return HttpResponseRedirect(settings.LOGOUT_REDIRECT_URL)
  elif getattr(settings,"USE_LOGOUT_PAGE", False):
    template = loader.get_template('login.html')
    c = {}
    c['home_url'] = getattr(settings,'HOME_URL',
                                  "%s/apps/" % settings.ARC_BASEURL)
    c['use_logout_page'] = True
    return HttpResponse(template.render(c, request))
  else:
    base_url = "%s/apps/login" % settings.ARC_BASEURL
    if hasattr(request, 'GET'):
      next_url = request.GET.get('next')
      if next_url:
        base_url = f'{base_url}?next={next_url}'

    return HttpResponseRedirect(base_url)

def shutdown(request, shutdown_key=None):
  """shutdown"""
  if os.environ.get("SHUTDOWN_KEY",False) == shutdown_key:
    # Need to use an OS exit because sys.exit doesn't work in a thread
    os._exit(0)
  return HttpResponse("")

def get_app(request):
  try:
    app_id = json.loads(request.user.userprofile.user_data).get('homeapp', None)

    default_app = reports.views_settings.setting_get_single('default_app')

    # Remove app ids if they don't exist
    if not app_id or (app_id != -1 and not Report.objects.filter(id = app_id).exists()):
      app_id = None
      app_name = None
      app_type = None
    else:
      app_data = get_app_name(app_id)
      app_name = app_data[1]
      app_type = app_data[2]

    if not default_app or not Report.objects.filter(id = default_app).exists():
      default_app = None
      default_app_name = None
      default_app_type = None
    else:
      default_app_data = get_app_name(default_app)
      default_app_name = default_app_data[1]
      default_app_type = default_app_data[2]

    return {
      "app_id": app_id,
      "app_name": app_name,
      "app_type": app_type,
      "default_app_id": default_app,
      "default_app_name": default_app_name,
      "default_app_type": default_app_type,
    }
  except:
    return {
      "app_id": None,
      "default_app_id": None,
      "app_name": None,
      "app_type": None,
      "default_app_name": None,
      "default_app_type": None,
    }

@autologin
@login_required
def home(request: HttpRequest):
  """home"""
  should_change_password = get_should_change_password(request.user)
  if should_change_password:
    return render(request, 'change_password.html')

  mobile = _getMobile(request)
  if mobile:
    return views_visuals_page.page(request)

  c = _appflags(request)

  app_info = get_app(request)
  home_app_id = app_info["default_app_id"] if c['disable_homepage_override'] else (app_info["app_id"] or app_info["default_app_id"])
  c['on_homepage'] = True
  user_data = json.loads(request.user.userprofile.user_data)
  c['incompatible_browser_warning_dismissed'] = user_data.get('incompatible_browser_warning_dismissed')

  if not home_app_id or home_app_id == -1:
    last_version_seen = user_data.get('new_release_modal_seen')
    modal_seen = last_version_seen == WHATSNEWVER
    c['new_release_modal_seen'] = modal_seen
    c['show_release_modal'] = not modal_seen and not FeatureFlipper().get_flipper_value('enable_explore_style')

    template = loader.get_template('apps_home.html')

  else:
    c['homeapp'] = home_app_id
    template = loader.get_template('apps_home_custom.html')

  return HttpResponse(template.render(c, request))

#
# Get from a visual to the dashboards that contain it. Used in cssusage() below
#   in order to show dashboards that contain visuals that contain custom styles
#
class _VisualTracker(object):
  def __init__(self):
    self.__visualToDashboard = {}
    self.__cssIdToAppgroup = {}

  def getDashboards(self, ids):
    dashboardIds = set()
    for id in ids:
      visualToDashboard = self.__visualToDashboard.get(id, set())
      # in case the id does not belong to a visual, but a dashboard lets pair the dashboard with None
      if not visualToDashboard:
        dashboardIds.add((id, None));
      else:
        dashboardIds |= visualToDashboard

    return dashboardIds

  def trackAppgroups(self):
    agl = AppGroup.objects.all()

    for ag in agl:
      try:
        if ag.group_parent:
          continue

        if not ag.group_options:
          continue

        ag_settings = json.loads(ag.group_options).get('user_settings', {})
        styles = json.loads(ag_settings.get('css_styles', '[]'))
        for style in styles:
          cssId = style['cssid']
          if not cssId in self.__cssIdToAppgroup:
            self.__cssIdToAppgroup[cssId] = []

          self.__cssIdToAppgroup[cssId].append(ag.id)

      except:
        pass

    return self.__cssIdToAppgroup

  def trackVisuals(self, ids):
    dashboardList = Report.objects.filter(report_type='dashboard')
    for dashboard in dashboardList:
      visuals = set()

      try:
        (_visuals, _pickers) = dashboard.get_visuals_in_dashboard()
        visuals.update(_visuals)
        visuals.update(_pickers)
      except Exception as e:
        pass

      idsToTrack = ids & visuals
      if idsToTrack:
        for id in idsToTrack:
          if not self.__visualToDashboard.get(id):
            self.__visualToDashboard[id] = set()
          self.__visualToDashboard[id].add(tuple([dashboard.id, id]))

@login_required
def cssusage(request):
  """Return a list of CSS styles <--> apps mapping"""

  # The css styles used by a report are stored within a json object
  # This means we cannot use any of django's query interfaces to get
  # a list of reports to css styles easily. Instead, we fetch all the
  # reports, parse the json, and retrieve any stored styles

  css_report_list = Report.objects         \
                           .filter(has_css_styles=True)   \
                           .select_related('dataset')

  stylesToUnlinkedVisuals = {}
  stylesToDashboards = {}
  for report in css_report_list:
    css_styles = None
    try:
      userPerm = report.dataset.getPerms(request.user)
      if 'ds_appview' in userPerm:
        report_data = json.loads(report.report_data)
        if report.report_type == 'dashboard':
          css_styles = json.loads(
              report_data['user_settings']['css_styles'])
        else:
          css_styles = json.loads(
              report_data['report_data']['user_settings']['css_styles'])
    except:
      pass

    if not css_styles:
      continue

    report_id = report.id
    for style in css_styles:
      if type(style) == type({}):
        cssId = style['cssid']
      else:
        cssId = style

      if not cssId in stylesToDashboards:
        stylesToDashboards[cssId] = set()
        stylesToUnlinkedVisuals[cssId] = set()

      stylesToDashboards[cssId].add(report_id)

      if report.report_type == "dashboard":
        continue

      if report.shared_visual_dashboards:
        continue

      # This is a visual within a single dashboard - mark is as such
      stylesToUnlinkedVisuals[cssId].add(report_id)

  dashboardAndVisualIds = set()
  for ids in stylesToDashboards.values():
    dashboardAndVisualIds.update(ids)

  tracker = _VisualTracker()
  tracker.trackVisuals(dashboardAndVisualIds)
  stylesForAppgroups = tracker.trackAppgroups()

  for cssId in stylesToDashboards:
    s = set()
    s |= tracker.getDashboards(stylesToDashboards[cssId])

    # Do not return linked visuals - they should only be seen in the context of
    # the dashboards they are in
    s -= stylesToUnlinkedVisuals[cssId]

    dashboardAndVisualIds = list(s)

    # DashboardAndVisualIds is a tuple collection {(dashIdx, visualIdx), ..., (dashIdx, None)}.
    # Sort can not handle None value, for comparison it should be convert to a number
    def sort_paired_tuples_with_None(item):
      if not item[1]:
        return (item[0], -1)
      return item
    stylesToDashboards[cssId] = sorted(dashboardAndVisualIds,key=sort_paired_tuples_with_None)
    stylesToUnlinkedVisuals[cssId] = list(stylesToUnlinkedVisuals[cssId])
    stylesToUnlinkedVisuals[cssId].sort()

  return HttpResponse(
      json.dumps({
        'stylesToDashboards': stylesToDashboards,
        'stylesToUnlinkedVisuals': stylesToUnlinkedVisuals,
        'stylesForAppgroups': stylesForAppgroups
      }), content_type="application/json")

@login_required
def colorpaletteUsage(request):
  """Return a list of Color Palettes <--> apps mapping"""
  # Same reasoning as of cssusage

  datal = Report.objects.values('id', 'report_type', 'report_data')
  outl = {}
  for data in datal:
    cp_style = None
    try:
      report_data = json.loads(data['report_data'])
      cp_style = report_data['report_data']['settings']['colorSettings']
    except:
      pass
    if cp_style:
      report_id = data['id']
      try:
        cp_style_id = cp_style['id']
      except KeyError:
        cp_style_id = None
      if not cp_style_id in outl:
        outl[cp_style_id] = []
      outl[cp_style_id].append(report_id)
  return HttpResponse(json.dumps(outl), content_type="application/json")

# Given a request, fetch and parse the 'auto_refresh' parameter,
# which should represent the number of seconds after which we refresh
# the display
def _get_auto_refresh(request):
  auto_refresh = request.GET.get("auto_refresh")
  auto_refresh_minimum = 10
  if auto_refresh:
    try:
      auto_refresh = int(auto_refresh)
      if auto_refresh < auto_refresh_minimum:
        # Minimum refresh in seconds
        auto_refresh = auto_refresh_minimum
      return auto_refresh
    except Exception:
      return None
  return None

def get_all_app_scope_filter_ids(app_datal):
  globalFilters = []
  for a in app_datal:
    j = json.loads(a[1])
    globalFilters.extend([int(x) for x in j.get('global_control_widgets', [])])
  return globalFilters


@autologin
@login_required
def appgroup(request, appgroup_id=None, appgroup_name=None):
  """Render appgroup in display only mode"""
  # Allow load by name
  if appgroup_name:
    if not ENABLE_URL_NAMES:
      return page_not_found(request, None)

    idlist = AppGroup.objects.filter(group_url_name=appgroup_name).values_list('id', flat=True)

    if len(idlist) == 0:
      return HttpResponse('App not found', status=500)
    if len(idlist) > 1:
      return HttpResponse('Multiple IDs found: %s' % idlist, status=500)

    appgroup_id = idlist[0]

  # TODO: Do we need SSO workarounds like in app() below?

  appgroup_id = int(appgroup_id)
  agl = AppGroup.objects \
      .filter(Q(id=appgroup_id) | Q(group_parent__id=appgroup_id)) \
      .select_related('workspace')

  mainAppgroup = None
  appgroup_ids = []
  dl = []
  appgroup_custom_styles = []
  appgroup_inline_css = ""
  for ag in agl:
    if ag.id == appgroup_id and ag.group_parent:
      raise Exception("subgroups cannot be viewed independently");

    if ag.id == appgroup_id:
      mainAppgroup = ag

    if not request.user.is_superuser and 'sys_misc' not in RBAC.get_user_obj_perms(request.user):
      ag.workspace.checkPerm(request.user, 'ws_appview')
    appgroup_ids.append(ag.id)
    order = ag.group_subgroup_order
    if order is None:
      order= 0
    d = {
      'group_id': ag.id,
      'name': str(ag.group_name),
      'subgroup_order': order,
      'options': str(ag.group_options)
    }
    dl.append(d)

    if ag.id == appgroup_id:
      ag_settings = ag.group_options
      if ag_settings:
        ag_settings = json.loads(ag_settings).get('user_settings', {})
        appgroup_custom_styles = json.loads(ag_settings.get('css_styles', '[]'))
        appgroup_inline_css = ag_settings.get('inline_css', '')

  dl.sort(key=itemgetter('subgroup_order'))

  app_groupmembership = []
  app_ids = []
  appsl = AppGroupAppMembership.objects \
      .filter(appgroup_group__id__in=appgroup_ids)
  for app in appsl:
    app_ids.append(app.appgroup_app_id)
    app_groupmembership.append([app.appgroup_app_id, app.appgroup_group_id,
        app.appgroup_order])

  app_datal = Report.objects \
                    .filter(id__in=app_ids) \
                    .select_related('workspace', 'dataset')
  app_titles = []
  for a in app_datal:
    has_access_to_workspace = request.user.is_superuser or 'sys_misc' in RBAC.get_user_obj_perms(request.user) or 'ws_appview' in a.workspace.getPerms(request.user)
    has_access_to_dataset = has_access_to_workspace or 'ds_appview'
    if not has_access_to_dataset not in a.dataset.getPerms(request.user):
      id_to_delete = a.id
      app_groupmembership = [report for report in app_groupmembership if report[0] != id_to_delete]
      continue

    j = json.loads(a.report_data)
    if 'dashboard_widgets' in j:
      rtype = "app"
      rtitle = j.get('report_title', 'Untitled')
    else:
      rd = j.get('report_data', {})
      rtype = j.get('report_type')
      rtitle = rd.get('report_title', 'Untitled')
    app_titles.append([a.id, rtitle, rtype])

  globalFilters = get_all_app_scope_filter_ids(app_datal.values_list('id', 'report_data'))

  _auditAppGroup(mainAppgroup, ActionType.READ, request=request)

  c = _appflags(request)
  c.update(createDashboardParams(request))
  c['appgroup'] = json.dumps(dl)
  c['appgroup_membership'] = json.dumps(app_groupmembership)
  c['apptitles'] = json.dumps(app_titles)
  c['globalFilters'] = json.dumps(globalFilters)
  c['appgroup_id'] = appgroup_id
  c['sheet_id'] = request.GET.get('sheet')
  c['is_dashboard'] = True
  c['appgroup_styles'] = [style['cssid'] for style in appgroup_custom_styles]
  c['appgroup_inline_css'] = appgroup_inline_css
  c['dataset_id'] = ''
  c['initialapp'] = request.GET.get('initialapp', '')
  template = loader.get_template('appgroup.html')

  return HttpResponse(template.render(c, request))

def parseFlags(c, request):
  mobile = _getMobile(request)
  c['mobile'] = mobile

  c['embed'] = False
  c['send_message_params'] = False

  send_message_params = False

  embed = request.GET.get("embed")
  if embed == "true":
    embed = True
  else:
    embed = False

  if embed:
    send_message_params = False
    send_message_params = request.GET.get("send_message_params")
    if send_message_params == "true":
      send_message_params = True
    else:
      send_message_params = False

  c['embed'] = embed
  c['send_message_params'] = send_message_params

@autologin
@login_required
# pylint: disable=R0912
def app(request, report_id='', dbname=''):
  """Render app in display only mode"""
  if not request.user.is_authenticated:
    logged_in = False
    if request.GET.get("autologin") == "true" or settings.AUTOLOGIN_ENABLED:
      try:
        username = settings.AUTOLOGIN_USERNAME
        password = settings.AUTOLOGIN_PASSWORD
        user = authenticate(username=username, password=password)
        if user is not None:
          if user.is_active:
            login(request, user)
            logged_in = True
      except Exception:
        pass
    if not logged_in:
      querystring  = request.GET.copy()
      querystring['next'] = request.path
      return redirect('%s?%s' % (settings.LOGIN_URL, querystring.urlencode()))

  assets = None
  temp_dashboard_name = request.GET.get('temp_dashboard')
  if temp_dashboard_name:
    assets = load_static_dashboard(temp_dashboard_name)

  else:
    # Allow load by name
    if dbname:
      if not ENABLE_URL_NAMES:
        return page_not_found(request, None)

      idlist = Report.objects.filter(report_type='dashboard',
                                    dashboard_url_name=dbname).values_list('id', flat=True)

      if len(idlist) == 0:
        return HttpResponse('Dashboard not found', status=500)
      if len(idlist) > 1:
        return HttpResponse('Multiple IDs found: %s' % idlist, status=500)

      report_id = idlist[0]

    if report_id != '':
      try:
        assets = report_get_internal(request, report_id, dataset_id=None, get_assets=True)
        if type(assets) == HttpResponse:
          assets = None
      except:
        LOGGER.exception('Error loading report %s' % report_id)
        assets = None

  c = _appflags(request)
  c.update(createDashboardParams(request))
  parseFlags(c, request)
  c['report_id'] = report_id

  app_datal = {}
  if (assets and assets[0]['report_type'] == 'dashboard'):
    c['dashboard_id'] = report_id
    c['sheet_id'] = request.GET.get('sheet')
    c['is_dashboard'] = True
    c['preload_assets'] = json.dumps({"/arc/reports/report/%s?assets=true" % report_id: assets})
    c['is_appgroup'] = False
    c['enable_dataconn_arcengine'] = settings.ENABLE_DATACONN_ARCENGINE
    template = loader.get_template('dashboard_detail.html')
    app_datal = json.loads(assets[0]['data'])

  else:
    template = loader.get_template('app_detail.html')
    c['sheet_id'] = ''

  c['is_edit'] = False
  if assets:
    c['dataset_id'] = assets[0]['dataset_id']

  c['dashboard_save_template_enabled'] = \
    getattr(settings, 'DASHBOARD_TEMPLATE', False) and \
    getattr(settings, 'DASHBOARD_SAVE_AS_TEMPLATE', False)

  c['globalFilters'] = json.dumps(app_datal.get('global_control_widgets', []))

  return HttpResponse(template.render(c, request))


@login_required
def appnew(request):
  """new visual """
  dataset_id = request.GET.get("dataset_id")
  visual_type = request.GET.get("visual_type", "table")
  template = loader.get_template("app_detail.html")

  if not dataset_id:
    return HttpResponse('Invalid dataset id', status=500)

  try:
    d = DataSet.objects.get(id=dataset_id)
  except DataSet.DoesNotExist:
    return HttpResponse("Dataset does not exist", status=500)
  except:
    return HttpResponse("Invalid dataset ID", status=500)

  d.checkPerm(request.user, 'ds_appedit')
  c = _appflags(request)
  c.update({
      # -2 is a special report id indicating that none is allocated yet
      "report_id": -2,
      "is_edit": True,
      "dataset_id": dataset_id,
      "visual_type": visual_type,
      "email_addresses": json.dumps(EmailJobContent.all_addresses(request.user)),
  })
  c['disable_v1_bars'] = getattr(settings, 'DISABLE_V1_BARS', True)
  c['sheet_id'] = ''
  return HttpResponse(template.render(c, request))

# See comments in reports/models.py: ParameterSet for the expected format for stored
# parameter sets
def getParamsFromSet(paramset_name):
  paramsetl = reports.models.ParameterSet.objects.filter(
      parameterset_name=paramset_name)
  if len(paramsetl) < 1:
    # No match - return an empty set of parameters
    return {}

  paramset = paramsetl[0]
  parsedParams = {}
  queryParams = parse_qs(paramset.parameterset_value)
  for k, vl in queryParams.items():
    for v in vl:
      parsedParams[k] = str(v)
  return parsedParams

def _get_param_context(request, is_dashboard=False, param_dict=None):
  paramsl = []
  params = {}
  parameters = {}
  dssw = {}
  gflags = {}
  dataset_scope = []
  dsswparam = 'arcparam.dssw.'
  gflag_name = 'gflag.'
  param_name = 'param.'
  parameter_name = 'config.'
  paramset_name = 'arcparam.paramset.'
  dataset_scope_name = 'scope'

  if param_dict is None:
    param_dict = request.GET
  for k, v in param_dict.items():
    k = urllib.parse.quote(k.encode("utf-8"))
    v = urllib.parse.quote(v.encode("utf-8"))

    if k.startswith(param_name):
      params[k[len(param_name):]] = v
      paramsl.append("%s=%s" % (k, v))
    elif k.startswith(gflag_name):
      gflags[k[len(gflag_name):]] = v
    elif is_dashboard and k.startswith(parameter_name):
      parameters[k[len(parameter_name):]] = v
      paramsl.append("%s=%s" % (k, v))
    elif is_dashboard and k.startswith(dataset_scope_name):
      dataset_scope = v
      paramsl.append("%s=%s" % (k, v))
    elif k.startswith(paramset_name):
      storedParams = getParamsFromSet(k[len(paramset_name):])
      for sk, sv in storedParams.items():
        if sk.startswith(param_name):
          params[sk[len(param_name):]] = sv
        elif is_dashboard and sk.startswith(dsswparam):
          try:
            dssw[int(sk[len(dsswparam):])] = int(sv)
          except Exception as e:
            pass
      paramsl.append("%s=%s" % (k, v))
    elif is_dashboard and k.startswith(dsswparam):
      try:
        dssw[int(k[len(dsswparam):])] = int(v)
      except Exception as e:
        pass

  if paramsl:
    params_str = "?" + "&".join(paramsl)
  else:
    params_str = ""

  c = {
    # pylint: disable=E1101
    'params': params,
    'parameters': parameters,
    'params_str': params_str,
    'dataset_scope': dataset_scope,
    'gflags': gflags,
  }

  if is_dashboard:
    c['dssw'] = dssw

  return c

def _get_context(request, is_dashboard=False):
  c = _get_param_context(request, is_dashboard=is_dashboard)

  c['email_addresses'] = json.dumps(EmailJobContent.all_addresses(request.user))
  c['schedule_groups'] = schedule_group_get()
  c['email_templates'] = json.loads(CustomEmailTemplate.email_templates('id', 'template_name', 'template_description', 'template_default_message'))
  c['scheduler_enabled'] = _SCHEDULER_ENABLED
  c['auto_refresh'] = False

  return c

def createDashboardParams(request):
  c = _get_context(request, is_dashboard=True)

  auto_refresh_seconds = _get_auto_refresh(request)
  if auto_refresh_seconds:
    c['auto_refresh'] = auto_refresh_seconds

  return c

@login_required
def dashboardedit(request, report_id=''):
  """Render new dashboard html to create a dashboard or load in edit mode"""
  template = loader.get_template('dashboard_edit.html')
  dashboard_id = report_id

  c = _appflags(request)
  c.update(createDashboardParams(request))
  c['dashboard_id'] = dashboard_id
  c['report_id'] = dashboard_id
  c['sheet_id'] = request.GET.get('sheet')
  c['is_dashboard'] = True
  c['is_appgroup'] = False
  c['enable_dataconn_arcengine'] = settings.ENABLE_DATACONN_ARCENGINE
  c['dashboard_save_template_enabled'] = \
    getattr(settings, 'DASHBOARD_TEMPLATE', False) and \
    getattr(settings, 'DASHBOARD_SAVE_AS_TEMPLATE', False)

  dataset_id = request.GET.get("dataset_id")
  if not dataset_id:
    dataset_id = getOrCreateDatasetFromQuery(request.user,
        request.GET.get("dataconnection_name", request.GET.get("dataconnection_id", None)),
        request.GET.get("query", None),
        request.GET.get("table", None), request=request)

  if dataset_id != None:
    try:
      d = DataSet.objects.get(id=dataset_id)
      d.checkPerm(request.user, 'ds_appedit')
      visual_type = request.GET.get("visual_type", "table")
      if c['report_id'] == "":
        c['report_id'] = -2
      c['is_edit'] = True
      c['dataset_id'] = dataset_id
      c['visual_type'] = visual_type
    except:
      LOGGER.warn("Invalid dataset id. Displaying empty dashboard")
  else:
    c['dataset_id'] = ""
    c['visual_type'] = False

  if not report_id:
    report_id = -2
  app_datal = Report.objects \
                    .values_list('id', 'report_data') \
                    .filter(id=report_id)

  c['globalFilters'] = json.dumps(get_all_app_scope_filter_ids(app_datal))

  return HttpResponse(template.render(c, request))

@login_required
def admin_arcvizsettings(request):
  """return arcviz setting as json for debugging purposes"""

  RBAC.checkPerm(request.user, 'sys_viewlogs')

  if reports.views_settings.setting_get_single('inspect_settings_enabled') != 'true':
    raise Exception('inspection of settings not enabled')

  name = request.GET.get('name', '')

  hiddenSettings = set(getattr(settings, 'HIDE_SETTINGS_FROM_DISPLAY', ['DATABASES', 'SECRET_KEY']))
  data = {}

  if name not in hiddenSettings and not name.startswith('_'):
    s = getattr(settings, name, None)
    if s is not None:
      data[name] = s

  return HttpResponse(json.dumps(data), content_type="application/json")

@login_required
def admin_users(request):
  """manage users"""
  RBAC.checkPerm(request.user, 'sys_editperm')

  c = _appflags(request)

  # Default to don't fetch user detail when LDAP is configured
  detail = getattr(settings, 'USER_ADMIN_FETCH_DETAIL', None)
  if detail is None:
    detail = not ('django_auth_ldap.backend.LDAPBackend' in settings.AUTHENTICATION_BACKENDS or \
                  'arcwebbase.backends.LdapBackend' in settings.AUTHENTICATION_BACKENDS or \
                  'arcweb.arcwebbase.basebackends.VizBaseLDAPBackend' in settings.AUTHENTICATION_BACKENDS)
  c['fetch_detail'] = int(detail)

  template = loader.get_template('apps_users.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_role_detail(request, role_id=None):
  """manage roles"""
  RBAC.checkPerm(request.user, 'sys_editperm')

  c = _appflags(request)
  c.update({
    'role_id':'undefined' if role_id is None else role_id,
    'roles_href':"../apps/roles" if role_id is None else "../roles",
  })
  return render(request, 'admin_role_detail.html', c)

@login_required
def admin_accel(request):
  """Start of app acceleration"""
  c = _appflags(request)
  c.update({
    'scheduler_enabled': _SCHEDULER_ENABLED,
  })

  appids = ''
  timestamp = ''
  for k, v in request.GET.items():
    if k == 'appids':
      appids = v
    elif k == 'timestamp':
      timestamp = v

  schedule_groups = schedule_group_get();

  c.update({'appids':appids, 'schedule_groups': schedule_groups, 'timestamp': timestamp})

  key = 'accel-queries-%s-%s' % (request.session.session_key, timestamp)
  data = cache.get(key)
  if data:
    queries = data.get('queries')
    dcId = data.get('dcId')
    if queries:
      context_updates = {'queries':queries, 'dcId': dcId, 'isQueryContext': True}

    if data.get('fromDirect'):
      context_updates['fromDirect'] = True

    if context_updates:
      c.update(context_updates)

  template = loader.get_template('admin_accel.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_accel_query(request, dc_id=None):

  if request.method != 'GET':
    return HttpResponse('Invalid request', status=400)

  try:
    dataconn = DataConnection.objects.get(id=dc_id)
  except:
    return HttpResponse('Invalid data connection.', status=400)

  dataconn.checkPerm(request.user, 'dc_aviews')

  if dataconn.dataconnection_type != 'arcengine':
    return HttpResponse('Invalid data connection type', status=400)

  session_key = request.session.session_key
  timestamp = request.GET.get('timestamp', '')
  state_key = 'table-state-%s-%s' % (session_key, timestamp)
  state = cache.get(state_key, '')

  # Determine which datasets can have their visuals and dashboards shown as links based on user perms
  perms = {}
  for d in DataSet.objects.filter(dataconnection__id=dc_id):
    perms[d.id] = 'ds_appview' in d.getPerms(request.user)

  c = _appflags(request)
  c.update({'dc_id': dc_id, 'state': state, 'perms': json.dumps(perms)})
  template = loader.get_template('admin_query_accel.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_roles(request):
  """manage roles"""
  RBAC.checkPerm(request.user, 'sys_editperm')

  c = _appflags(request)

  template = loader.get_template('admin_roles.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_aliases(request):
  """manage URL aliases"""

  c = _appflags(request)

  return render(request, 'admin_aliases.html', c)

@login_required
def admin_activity(request):
  """view activity logs"""
  RBAC.checkPerm(request.user, 'sys_viewlogs')

  c = _appflags(request)
  c['auditLogExists'] = auditLogExists()
  c['arcviz_log_ui_enable'] = getattr(settings, 'ARCVIZ_LOG_UI_ENABLE', False)
  c['arcengine_log_ui_enable'] = getattr(settings, 'ARCENGINE_LOG_UI_ENABLE', False)
  c['all_connection_types'] = ALL_CONNECTION_TYPES

  template = loader.get_template('admin_activity.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_apidemo(request):
  """admin API demo page"""
  c = _appflags(request)
  c['apidemo_list'] = json.dumps(getattr(settings, 'ADMIN_API_DEMO_LIST', []))
  c['auditLogExists'] = auditLogExists()
  c['demofilters'] = getattr(settings, 'ADMIN_API_DEMO_FILTERS', False)

  template = loader.get_template('admin_apidemo.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_settings(request):
  """view settings"""
  RBAC.checkPerm(request.user, 'sys_settings')

  c = _appflags(request)
  c.update({
    'email_encryption_supported': bool(encrypt_secret('test_secret')),
    'thumbnail_pil_supported': _PIL_ENABLED
  })

  template = loader.get_template('admin_settings.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_restart(request):
  if not request.user.is_superuser and 'sys_misc' not in RBAC.get_user_obj_perms(request.user):
    return HttpResponse("Only admins can view this page", status=401)

  c = _appflags(request)
  template = loader.get_template('admin_restart.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_stop_work(request):
  if not request.user.is_superuser and 'sys_misc' not in RBAC.get_user_obj_perms(request.user):
    return HttpResponse("Only admins can perform this action", status=401)

  if not request.method == 'POST':
    raise TypeError('Invalid call')

  os._exit(errno.EAGAIN)

@login_required
def admin_staticassets(request, assetid=None):
  """manage static assets"""
  RBAC.checkPerm(request.user, 'sys_styles')

  c = _appflags(request)
  c['limit_in_mb'] = int(getattr(settings, 'DATA_UPLOAD_MAX_MEMORY_SIZE')/1000000)

  if assetid is None:
    template = loader.get_template('apps_staticassets.html')
    return HttpResponse(template.render(c, request))

  c['assetid'] = assetid
  template = loader.get_template('apps_staticasset_form.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_customstyles(request, styleid=None):
  """manage custom styles"""
  RBAC.checkPerm(request.user, 'sys_styles')

  c = _appflags(request)

  if styleid is None:
    template = loader.get_template('apps_customstyles.html')
    return HttpResponse(template.render(c, request))

  c['styleid'] = styleid
  template = loader.get_template('apps_customstyle_form.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_customcolors(request, colorid=None):
  """manage custom colors"""
  RBAC.checkPerm(request.user, 'sys_styles')

  c = _appflags(request)

  if colorid is None:
    template = loader.get_template('apps_customcolors.html')
    return HttpResponse(template.render(c, request))

  c['colorid'] = colorid
  template = loader.get_template('apps_customcolor_form.html')
  return HttpResponse(template.render(c, request))

@login_required
def admin_customdates(request, daterange_id=None):
  """manage custom dates"""
  RBAC.checkPerm(request.user, 'sys_styles')

  c = _appflags(request)

  if daterange_id is None:
    template = loader.get_template('apps_customdates.html')
    return HttpResponse(template.render(c, request))

  c['daterange_id'] = daterange_id
  template = loader.get_template('apps_customdate_form.html')
  return HttpResponse(template.render(c, request))

def get_fullname(user):
  return '{} {}'.format(user.first_name, user.last_name).strip()

def _users_get_raw(users, requester_is_superuser, get_groups=None, get_roles=None):
  outl = []

  profkeys = UserProfile.PROFILE_KEYS_EDIT + UserProfile.PROFILE_KEYS_VIEW

  for u in users:
    if requester_is_superuser:
      user = {
        'username': u.username,
        'id': u.id,
        'is_superuser': u.is_superuser,
        'is_active': u.is_active,
        'last_name': u.last_name,
        'full_name': get_fullname(u),
        'first_name': u.first_name,
        'date_joined': u.date_joined.isoformat(),
        'last_login': u.last_login.isoformat()
            if u.last_login else 'N/A',
        'is_password_editable': u.userprofile.is_password_editable,
      }
    else:
      user = {
        'username': u.username,
        'id': u.id,
        'full_name': get_fullname(u),
      }

    # Add in UI visible profile keys
    try:
      data = u.userprofile.user_data_obj
    except UserProfile.DoesNotExist:
      LOGGER.warning('Creating missing user profile for %s' % u.username)
      u.save()
      data = {}
    profile = dict((key, data[key]) for key in profkeys if key in data)
    user['profile'] = json.dumps(profile)

    if get_groups:
      if hasattr(u, '_arcadia_groupnames'):
        if u._arcadia_groupnames:
          user["groups"] = [{'id':int(id), 'name':name} for (id, name) in \
                            zip(u._arcadia_groupids.split(','), u._arcadia_groupnames.split(','))]
        else:
          user["groups"] = []
      else:
        user["groups"] = _groups_get_raw(u.groups.all(), requester_is_superuser)
    if get_roles:
      user["roles"] = [{'id':r.id, 'name':r.name} for r in RBAC.roles_for_user(u)]

    outl.append(user)

  return outl

def _users_get(username, requester_is_superuser, get_groups=True, get_roles=True, filter=None, start=None, length=None, search=None, ordered_column_name=None, order_direction=None):
  is_paginated = start is not None and length is not None

  if username:
    users = User.objects.filter(username=username)
  else:
    users = User.objects.all()

  if filter:
    users = users.filter(**filter)

  if search:
    users = users.filter(
      Q(username__icontains=search) |
      Q(last_name__icontains=search) |
      Q(first_name__icontains=search)
    )

  if ordered_column_name is not None:
    order_by = ordered_column_name if order_direction == 'asc' else f'-{ordered_column_name}'
    users = users.order_by(order_by)

  if is_paginated:
    next_page = start / length + 1
    paginator = Paginator(users, per_page=length)
    users = paginator.page(next_page).object_list

  users = users.select_related('userprofile')

  if GroupConcatEnabled():
    users = users.annotate(_arcadia_groupnames=GroupConcat('groups__name', separator=',')) \
                 .annotate(_arcadia_groupids=GroupConcat('groups__id', separator=','))

  outl = _users_get_raw(users, requester_is_superuser,
                        get_groups=get_groups, get_roles=get_roles)

  if is_paginated:
    return JsonResponse({
      "numberOfUsers": User.objects.count(),
      "numberOfFiltered" : paginator.count,
      "users": outl })

  else:
    return HttpResponse(json.dumps(outl), content_type="application/json")

# Update any UI configurable profile keys
def _update_user_profile(requester, user, user_profile):
  if user_profile and user.userprofile:
    should_update = False

    user_profile = json.loads(user_profile)
    user_data_obj = user.userprofile.user_data_obj

    for (key, val) in user_profile.items():
      if key in UserProfile.PROFILE_KEYS_EDIT and user_data_obj.get(key) != val:
        if key in UserProfile.PROFILE_KEYS_PERM:
          RBAC.checkPerm(requester, 'sys_editperm')

        if val:
          user_data_obj[key] = val

        elif key in user_data_obj:
          del user_data_obj[key]

        should_update = True

    if should_update:
      user.userprofile.user_data_obj = user_data_obj
      user.userprofile.save()

def _users_create(request, username):
  if not FeatureFlipper().get_flipper_value('enable_user_create'):
    return HttpResponse("Cannot create users on this system", status=500)

  users = User.objects.filter(username=username)
  if len(users) > 0:
    return HttpResponse("username already exists", status=500)

  is_superuser = request.POST.get("is_superuser", None)

  if is_superuser == 'true' and not request.user.is_superuser:
    return HttpResponse("Only admin can create admin user", status=500)

  first_name = request.POST.get('first_name', '')
  last_name = request.POST.get('last_name', '')

  if not username:
    return HttpResponse("username must be specified", status=500)

  if "password" not in request.POST:
    return HttpResponse("password must be specified", status=500)
  password = request.POST.get("password")
  validatePassword(password)

  # Can only create usernames with the valid chars. No need
  # for both validateStringNoHtml and validateUsernameCharacters
  validateUsernameCharacters(username)

  user = User.objects.create_user(username, '', password, first_name=first_name, last_name=last_name)
  if is_superuser == "true":
    user.is_superuser = True

  groups = request.POST.get("groups", None)
  if groups is not None:
    user.groups.clear()
    user.groups.add(*json.loads(groups))

  roles = request.POST.get("roles", None)
  if roles is not None:
    RBAC.update_user_roles(json.loads(roles), user)

  if user.userprofile.is_password_editable:
      user.userprofile.should_change_password = True
      user.userprofile.save()

  user.save()
  _auditUser(user, ActionType.CREATE, request=request)

  # Update any UI configurable profile keys
  _update_user_profile(request.user, user, request.POST.get('profile'))

  return _users_get(username, request.user.is_superuser)

@login_required
def bulk_user_group_associate(request: HttpRequest) -> JsonResponse:
  if not request.method == 'POST':
    raise TypeError('Endpoint requires POST')

  if not request.user.is_superuser:
    return HttpResponse("Only admins can add users to groups", status=401)

  data = json.loads(request.body)
  group_ids = data['group_ids']
  user_ids = data['user_ids']

  for group in Group.objects.filter(id__in=group_ids):
    group.user_set.add(*user_ids)

  return JsonResponse({'msg': 'success'})

def _users_update(request, username, input=None):
  user = User.objects.filter(username=username).first()

  if user is None:
    return HttpResponse("username not found", status=500)

  put = input or QueryDict(request.body)

  # If the username is being altered check to make sure that it only
  # has valid characters inside of it
  new_username = put.get("username", None)
  if new_username is not None and username != new_username:
    if not request.user.is_superuser and request.user.username != username:
      return HttpResponse("Only admins can change usernames", status=500)
    validateUsernameCharacters(new_username)

  has_manage_user_perm = 'sys_editperm' in RBAC.get_user_obj_perms(request.user)
  def check_update_perm(attribute_name):
    if not has_manage_user_perm and not request.user.is_superuser:
      raise PermissionException(request.user, f"Missing \"manage roles and users\" permission required to update {attribute_name}")

  # Found a valid user in 'users[0]'
  needs_save = False
  is_active = put.get("is_active", None)
  if is_active is not None and (is_active == 'true') != user.is_active:
    if user.is_superuser:
      if not request.user.is_superuser:
        return HttpResponse("Only admin can change another admin's active status", status=500)
    else:
      check_update_perm("is_active")

    if is_active == "false":
      if user.username == request.user.username:
        return HttpResponse("It is not possible to inactivate yourself", status=500)
      user.is_active = False
    else:
      user.is_active = True
    needs_save = True

  # UPDATE IS_SUPERUSER LOGIC:
  # If superuser status is changing, we should check for a few conditions to determine
  #  whether the user making the update is valid do so.
  is_superuser = put.get("is_superuser", None)
  if is_superuser is not None and (is_superuser == "true") != user.is_superuser:
    # Scenario 1: User is a not superuser and tries to update either their own
    #  or another user's status to make them a super user.
    #  We don't want to allow that to happen so we'll reject their request.
    if not request.user.is_superuser:
      return HttpResponse("Only admins can change admin status", status=500)
    # Scenario 2: User is a superuser and tries to update another user's status to make them a super user,
    #  but we're explicilty not allowing this because Admin permissions are controlled by external groups.
    if (request.user.is_superuser) \
            and (os.getenv('CDV_MODE') == 'cdw' or getattr(settings, 'SAML_REQUIRED_GROUPS_CHECK', False)):
      return HttpResponse("Admin status controlled explicitly by external Group membership. "
                          "Please ensure this user is in a specified Admin group or update your Site Settings "
                          "configuration to include new Admin groups before updating their Admin status.",
                          status=500)
    if is_superuser == "true":
      user.is_superuser = True
    else:
      user.is_superuser = False
    needs_save = True

  # UPDATE PASSWORD LOGIC:
  #  - Admin should be able to update another user's password (both for admin and normal user)
  #  - Admins and normal users should not be able to update their own password without providing the old password
  old_password = put.get("old_password", None)
  new_password = put.get("new_password", None)
  confirm_password = put.get("confirm_password", None)

  if new_password:
    # If the password has been altered check to make sure it meets
    # complexity and length requirements
    validatePassword(new_password)

    # User changing own password
    if old_password is not None:
      if (confirm_password is None):
        return HttpResponse("Password confirmation is missing", status=500)

      if (new_password != confirm_password):
        return HttpResponse("Passwords do not match", status=500)

      is_authenticated = authenticate(username=username, password=old_password) is not None

      if not is_authenticated:
        return HttpResponse("The current password specified is incorrect", status=500)

      user.userprofile.should_change_password = False
      user.userprofile.save()

    # User changing password for someone else (or user trying to change own password at the wrong place)
    else:
      if request.user.username == username:
        return HttpResponse("Password can't be changed without specifying old password. Click on the user icon in the right corner of the navbar to change your password.", status=500)

      if not request.user.is_superuser:
        return HttpResponse("Only admins can change other users passwords", status=500)

      user.userprofile.should_change_password = True
      user.userprofile.save()

    user.set_password(new_password)
    update_session_auth_hash(request, user)
    needs_save = True

  if new_username is not None:
    user.username = new_username
    needs_save = True

  groups = put.get('groups')
  if groups:
    old_groups = user.groups.values_list('id', flat=True)
    new_groups = json.loads(groups)
    if sorted(old_groups) != sorted(new_groups):
      check_update_perm('groups')
      user.groups.set(new_groups)
      needs_save = True

  roles = put.get("roles", None)
  if roles is not None:
    old_roles = [role.id for role in RBAC.roles_for_user(user)]
    new_roles = json.loads(roles)
    if not sorted(old_roles) == sorted(new_roles):
      check_update_perm('roles')
      RBAC.update_user_roles(new_roles, user)


  def update_attribute(attribute_name):
    if attribute_name in put:
      new_attribute_value = put.get(attribute_name, '')
      if new_attribute_value != getattr(user,attribute_name):
        check_update_perm(attribute_name)
        setattr(user, attribute_name, new_attribute_value)
        return True
    return False

  needs_save = update_attribute('first_name') or needs_save
  needs_save = update_attribute('last_name') or needs_save

  if needs_save:
    _auditUser(user, ActionType.UPDATE, request=request)
    user.save()

  # Update any UI configurable profile keys
  _update_user_profile(request.user, user, put.get('profile'))

  return _users_get(user.username, request.user.is_superuser)

def _users_delete(request, username):
  if not request.user.is_superuser:
    return HttpResponse("Only admin can delete user", status=500)

  if username is not None:
    users = User.objects.filter(username=username)
  else:
    usernames = json.loads(QueryDict(request.body).get('users', None))
    if usernames is None:
      return HttpResponse("No users selected", status=500)
    else:
      users = User.objects.filter(username__in=usernames)

  if len(users) < 1:
    return HttpResponse("Username not found", status=500)

  for u in users:
    RBAC.update_user_roles([], u)
    _auditUser(u, ActionType.DELETE, request=request)

  users.delete()

  return HttpResponse('[]')

def _groups_get_raw(groups_objects, user_is_superuser, get_users=None, get_roles=None, start=None, length=None, search=None, ordered_column_name=None, order_direction=None):
  outl = []

  is_paginated = start is not None and length is not None

  if search:
    groups_objects = groups_objects.filter(
      Q(name__icontains=search)
    )

  if ordered_column_name is not None:
    order_by = ordered_column_name if order_direction == 'asc' else f'-{ordered_column_name}'
    groups_objects = groups_objects.order_by(order_by)

  if is_paginated:
    next_page = start / length + 1
    paginator = Paginator(groups_objects, per_page=length)
    groups_objects = paginator.page(next_page).object_list

  for g in groups_objects:
    group = {
      'id': g.id,
      'name': g.name,
    }

    if get_users:
      group["users"] = _users_get_raw(g.user_set.all(), user_is_superuser)
    if get_roles:
      all_roles = RBAC.load_roles()
      g_roles = [{'id': r.id, 'name': r.name} for r in all_roles if g.name in r.groupSet()]
      group["roles"] = g_roles

    outl.append(group)

  if is_paginated:
    return {
      "numberOfGroups": Group.objects.count(),
      "numberOfFiltered" : paginator.count,
      "groups": outl}
  else:
    return outl

def _groups_create(request):
  name = request.POST.get("name", None)
  groups = Group.objects.filter(name=name)
  if len(groups) > 0:
    return HttpResponse("Group name already exists", status=500)

  validateStringNoHtml(name)

  group = Group.objects.create(name=name)

  users = request.POST.get("users", None)
  if users is not None:
    group.user_set.clear()
    group.user_set.add(*json.loads(users))

  roles = request.POST.get("roles", None)
  if roles is not None:
    RBAC.update_group_roles(json.loads(roles), group)

  group.save()
  return HttpResponse(json.dumps(_groups_get_raw([group], request.user.is_superuser, True, True)), content_type="application/json")

def _groups_update(request, group_id, input=None):
  groups = Group.objects.filter(id=group_id)
  if len(groups) < 1:
    return HttpResponse("Group cannot found", status=500)

  if len(groups) > 1:
    return HttpResponse("More than one group cannot be edited at a time",
                        status=500)

  put = input or QueryDict(request.body)

  # Found a valid group in 'group[0]'
  new_groupname = put.get("name", None)

  validateStringNoHtml(new_groupname)

  if new_groupname is not None:
    groups[0].name = new_groupname

  users = put.get("users", None)
  if users is not None:
    groups[0].user_set.clear()
    groups[0].user_set.add(*json.loads(users))

  roles = put.get("roles", None)
  if roles is not None:
    RBAC.update_group_roles(json.loads(roles), groups[0])

  groups[0].save()

  return HttpResponse(json.dumps(_groups_get_raw(groups, request.user.is_superuser, True, True)), content_type="application/json")

def _groups_delete(request, group_id):
  if group_id is not None:
    groups = Group.objects.filter(id=group_id)
  else:
    ids = json.loads(QueryDict(request.body).get('groups', None))
    if ids is None:
      return HttpResponse("No groups selected", status=500)
    else:
      groups = Group.objects.filter(id__in=ids)

  if len(groups) < 1:
    return HttpResponse("Groups are not found", status=500)

  for g in groups:
    RBAC.update_group_roles([], g)

  groups.delete()
  return HttpResponse('[]')

@login_required
def groups(request, group_id=None):
  return groups_internal(request, group_id)

def groups_internal(request, group_id=None, input=None, filter=None):
  if group_id is not None:
    try:
      group_id = int(group_id)
    except:
        return HttpResponse("Invalid group ID.", status=500)

  if request.method == 'GET':
    start = request.GET.get('start')
    if start is not None:
      start = int(start)

    length = request.GET.get('length')
    if length is not None:
      length = int(length)

    search = request.GET.get('search[value]', None)

    ordered_column_number = request.GET.get('order[0][column]')
    if ordered_column_number is not None:
      ordered_column_number = int(ordered_column_number)

    ordered_column_name = request.GET.get(f'columns[{ordered_column_number}][name]', None)

    order_direction = request.GET.get('order[0][dir]', None)

    if request.GET.get('summary', False):
      get_details = False
    else:
      try:
        RBAC.checkPerm(request.user, 'sys_editperm')
        get_details = True
      except:
        get_details = False

    if group_id is None:
      groups = Group.objects.prefetch_related(
        'user_set',
        'user_set__userprofile'
      ).all()
      if filter:
        groups = groups.filter(**filter)
    else:
      groups = Group.objects.filter(id=group_id)
    return HttpResponse(json.dumps(_groups_get_raw(groups, request.user.is_superuser, get_details, get_details, start=start, length=length, search=search, ordered_column_name=ordered_column_name, order_direction=order_direction)), content_type="application/json")

  RBAC.checkPerm(request.user, 'sys_editperm')
  if request.method == 'POST':
    return _groups_create(request)

  if request.method == 'PUT':
    return _groups_update(request, group_id, input)

  if request.method == 'DELETE':
    return _groups_delete(request, group_id)

@login_required
def groups_basic(request):
  groups = Group.objects.all().values('id', 'name')

  return JsonResponse(list(groups), safe=False)

@login_required
def users_basic(request, username=None):
  outl = []
  try:
    RBAC.checkPerm(request.user, 'sys_viewroles')
    users = User.objects.all()
  except:
    users = [User.objects.get(id=request.user.id)]

  for u in users:
    user = {
      'username': u.username,
      'id': u.id,
      'full_name': get_fullname(u),
    }
    outl.append(user)

  return HttpResponse(json.dumps(outl), content_type="application/json")

@login_required
def users(request, username=None):
  """
  REST endpoint for users
  """

  return users_internal(request, username)

def users_internal(request, username=None, input=None, filter=None):
  # Users can update their own passwords
  if request.method == 'PUT':
    return _users_update(request, username, input)

  # If user has system edit permissions, treat their GET calls like a superuser
  if request.method == 'GET':
    RBAC.checkPerm(request.user, 'sys_viewroles')
    start = request.GET.get('start')
    if start is not None:
      start = int(start)

    length = request.GET.get('length')
    if length is not None:
      length = int(length)

    search = request.GET.get('search[value]', None)

    ordered_column_number = request.GET.get('order[0][column]')
    if ordered_column_number is not None:
      ordered_column_number = int(ordered_column_number)

    ordered_column_name = request.GET.get(f'columns[{ordered_column_number}][name]', None)

    order_direction = request.GET.get('order[0][dir]', None)

    if request.GET.get('summary', False):
      get_groups = False
      get_roles = False
    else:
      get_groups = True
      get_roles = True

    # If configured so that users with manage dataset can configure the FA,
    # then we can't require 'manage users & roles' permission to read the
    # list of users because these are required to configure the FA.
    if not getattr(settings, 'MANAGE_DS_FA', False):
      perms_per_user = RBAC.get_user_obj_perms(request.user)
      has_edit_permission = 'sys_editperm' in perms_per_user
      has_view_roles_permission = 'sys_viewroles' in perms_per_user
      if not has_edit_permission and not has_view_roles_permission:
        username = request.user.username

    return _users_get(username,
                      'sys_editperm' in RBAC.get_user_obj_perms(request.user),
                      get_groups=get_groups,
                      get_roles=get_roles,
                      filter=filter, start=start, length=length, search=search,
                      ordered_column_name=ordered_column_name, order_direction=order_direction)

  # All other requests need permisssion
  RBAC.checkPerm(request.user, 'sys_editperm')

  if request.method == 'POST':
    return _users_create(request, username)

  if request.method == 'DELETE':
    return _users_delete(request, username)

  return HttpResponse('Invalid request method', status=500)

@login_required
def appedit(request, app_id=''):
  """edit or build an app (formerly appgroup)"""
  template = loader.get_template('appgroup_edit.html')
  c = _appflags(request)
  c['appgroup_id'] = app_id
  return HttpResponse(template.render(c, request))

@login_required
def homepage(request):
  """manage homepage"""

  app_info = get_app(request)
  home_app_id = app_info["app_id"]
  home_app_name = app_info["app_name"]
  home_app_type = app_info["app_type"]

  context = _appflags(request)
  context.update({
    "app_id": home_app_id or "",
    "app_name": home_app_name or ("Untitled" if home_app_id else "Data Visualization Welcome Page"),
    "app_type": home_app_type or "Default Home Page",

    "default_app_id": app_info["default_app_id"] or "-1",
    "default_app_name": app_info["default_app_name"] or ("Untitled" if app_info["default_app_id"] else "Data Visualization Welcome Page"),
    "default_app_type": app_info["default_app_type"] or "Default Home Page",

    "is_admin": request.user.is_superuser,
    # Need this to know whether to replace app_id text when default_app_id is changed
    "original_app_id": app_info["app_id"] or "",
  })

  return render(request, 'admin_homepage.html', context)

directaccess_types = ['arcengine', 'arcengine_v2', 'impala', 'impyla','hive', 'hyve', 'sqlite', 'vertica', 'ksql', 'snowflake', 'duckdb', 'druid'];

@login_required
def data(request, template="apps_data.html"):
  """data"""
  datasource_form = DataSourceForm()
  viz_import_form = ImportUploadFileForm()
  dataimportflag = request.GET.get("dataimport", 'false')
  table_name = request.GET.get("table_name")
  template = loader.get_template(template)
  schedule_groups = schedule_group_get()
  dataconnection_id = request.GET.get("dataconnection_id")
  dataconnection_name = request.GET.get("dataconnection_name")
  database_name = request.GET.get("database_name")
  datatable_name = request.GET.get("datatable_name")

  c = {
    'datasource_form': datasource_form,
    'dataimport': dataimportflag,
    'schedule_groups': schedule_groups,
    'viz_import_form':viz_import_form,
    'scheduler_enabled': _SCHEDULER_ENABLED,
    'all_conntypes': ALL_CONNECTION_TYPES,
    'dashboard_template_enabled': 'true' if getattr(settings, 'DASHBOARD_TEMPLATE', False)
                                          else 'false',
    'dataconnection_id': '',
    'dataconnection_name': '',
    'database_name': '',
    'datatable_name': '',
    'table_name': '',
  }

  if dataconnection_id:
      c['dataconnection_id'] = dataconnection_id

  if dataconnection_name:
    c['dataconnection_name'] = dataconnection_name

  if database_name:
    c['database_name'] = database_name

  if datatable_name:
    c['datatable_name'] = datatable_name

  if table_name:
    c['table_name'] = table_name

  #Retrieve cached queries to be populated into direct access
  timestamp = ''
  for k, v in request.GET.items():
    if k == 'timestamp':
      timestamp = v

  key = 'accel-queries-%s-%s' % (request.session.session_key, timestamp)
  data = cache.get(key)
  if data:
    c['queries'] = data.get('queries')

  c.update(_appflags(request))

  embed = request.GET.get("embed")
  if embed == "true":
    c['embed'] = True

  return HttpResponse(template.render(c, request))

@login_required
def datasql(request):
  """sql access to data connections"""
  return data(request, template="apps_datasql.html")

@login_required
def unregister(request, dataset_id=None):
  if not request.user.is_superuser:
    return HttpResponse("Only an admin user can unregister", status=500)

  reports.views_settings.unregister()
  return HttpResponseRedirect("%s/apps/" % settings.ARC_BASEURL)

@login_required
def register(request, dataset_id=None):
  if not request.user.is_superuser:
    return HttpResponse("Only an admin user can register", status=500)

  firstname = request.GET.get('firstname')
  lastname = request.GET.get('lastname')
  emailaddress = request.GET.get('emailaddress')
  organization = request.GET.get('organization')
  role = request.GET.get('role')
  userid = request.GET.get('userid', '')

  if not firstname or not lastname or not emailaddress:
    raise Exception('Invalid parameters')

  reports.views_settings.register(firstname, lastname, emailaddress,
      organization, role, userid)
  return HttpResponseRedirect("%s/apps/" % settings.ARC_BASEURL)

@login_required
def dataset(request, dataset_id=None):
  """dataset"""
  template = loader.get_template('apps_dataset.html')

  schedule_groups = schedule_group_get()

  c = _appflags(request)
  c.update({
      'dataset_id': dataset_id,
      'schedule_groups': schedule_groups,
      'scheduler_enabled': _SCHEDULER_ENABLED,
      'email_addresses': json.dumps(EmailJobContent.all_addresses(request.user)),
  })
  return HttpResponse(template.render(c, request))

def fetchDataConnectionByName(dataconnection_name):
  return DataConnection.objects.filter(dataconnection_name=dataconnection_name)

def fetchDatasetsByDataConnection(dataconnection_id):
  return DataSet.objects.filter(dataconnection__id=dataconnection_id)

#
# A user can get to the dashboard builder URL with a table name or a query
# instead of a dataset id. The user can optionally specify a data connection
# or have the settings.DATASET_AUTOCREATE_DATACONNECTION connection be used.
#
# This is used to support the use-case where an external SQL editor can
# redirect a user to CDV with the context of the query or the tabel the user
# wants to visualize.
#
# Given these parameters, this function returns an existing dataset built on
# this query/table, or creates a new one. Note that query matches have to be
# exact, as we do not understand enough of the query semantically to do
# approximate matching well.
#
def getOrCreateDatasetFromQuery(user, dataconnection_name_or_id, query, fq_tablename, request=None):
  is_directacess_disabled = not FeatureFlipper().get_flipper_value('enable_direct_access')
  is_dataset_autocreate_disabled = not getattr(settings, 'ENABLE_DATASET_AUTOCREATE', False)
  if is_directacess_disabled and is_dataset_autocreate_disabled:
    LOGGER.warn("autocreate not enabled")
    return None

  if not query and not fq_tablename:
    LOGGER.warn("No query or table specified")
    return None

  if query:
    query = query.strip()
    query = query.strip(' ;')
    if not query[:6].lower() == "select":
      LOGGER.warn("Invalid query specified %s", query)
      return None

  if fq_tablename:
    fields = fq_tablename.split(".")
    if len(fields) != 2:
      LOGGER.warn("Invalid table specified %s", fq_tablename)
      return None

  if not dataconnection_name_or_id:
    dataconnection_name_or_id = settings.DATASET_AUTOCREATE_DATACONNECTION

  if not dataconnection_name_or_id:
    return None

  dc = fetchDataConnectionByName(dataconnection_name_or_id)
  if len(dc) == 0:
    try:
      dc = DataConnection.objects.get(id=int(dataconnection_name_or_id))
    except Exception as e:
      LOGGER.exception(e)
      return None
  else:
    dc = dc[0]

  try:
    dc.checkPerm(user, 'dc_explore')
  except PermissionException:
    LOGGER.warn("User has no permission to get or create dataset")
    return None

  dsl = fetchDatasetsByDataConnection(int(dc.id))
  if len(dsl) > 0:
    for ds in dsl:
      if fq_tablename:
        if ds.dataset_detail == fq_tablename:
          return ds.id

      if query:
        try:
          dsinfo = json.loads(ds.dataset_info)
          if len(dsinfo) != 1:
            continue

          if dsinfo[0]["tablename"] == "text.sql":
            if dsinfo[0]["select"] == query:
              return ds.id
        except:
          pass

  # We are going to create a dataset below if we can. Verify the user has
  # explore permissions on the dataconnection
  dc.checkPerm(user, 'dc_explore')

  conn = get_dataconn(dc.id)

  if query:
    properties = conn.properties()
    if query and not properties.supports_limit:
      raise Exception("Cannot auto-create dataset on a data connection " +
          "that does not support limit clauses")

    if properties.supports_limit_0:
      limit = 0
    else:
      limit = 1

    if query:
      cmd = f"select * from ({query}) cdv_t1 limit {limit}"
    else:
      cmd = f"select * from ({fq_tablename}) cdv_t1 limit {limit}"

    results = conn.sync_runquery(cmd, username=user.username)
  else:
    results = conn.table_columns(fq_tablename, username=user.username)

  if not results or not 'colnames' in results:
    LOGGER.warn("Could not resolve columns for dataset creation")
    return None

  colnames = results['colnames']
  coltypes = results['coltypes']

  if len(colnames) == 0:
    LOGGER.warn("Could not resolve columns for dataset creation")
    return None

  dsData = {}
  dsData['dataconnection_id'] = dc.id
  dsData['dataset_type'] = 'singletable'
  dsData['dataset_settings'] = {}
  dsData['dataset_lvname'] = ''

  if query:
    dsData['dataset_detail'] = 'Created from SQL'
  else:
    dsData['dataset_detail'] = fq_tablename

  dsData['description'] = 'Auto-created for SQL apps'

  if query:
    queryId = str(uuid.uuid4())
    dsData['dataset_name'] = 'autogen %s' % (queryId)
    dsData['dataset_info'] = [
      {
        "tablename": "text.sql",
        "select": query,
        "columns": []
      }
    ]

  else:
    dsData['dataset_name'] = fq_tablename
    dsData['dataset_info'] = [
      {
        "tablename": fq_tablename,
        "columns": []
      }
    ]

  for idx, colname in enumerate(colnames):
    dsData['dataset_info'][0]["columns"].append({
      "name": colname,
      "alias": colname,
      "type": coltypes[idx],
      "isdim": (coltypes[idx] != "BIGINT" and coltypes[idx] != "REAL" and \
          coltypes[idx] != "DOUBLE")
    })

  dsData['dataset_settings'] = json.dumps(dsData['dataset_settings'])
  dsData['dataset_info'] = json.dumps(dsData['dataset_info'])

  if not request:
    request = create_request(user=user)

  d_id_or_error = datasets.views.dataset_create_or_clone(dsData, request)
  if type(d_id_or_error) == type(1):
    return d_id_or_error

  return None

@login_required
def builder(request, report_id=''):
  """Build a report"""
  if report_id == '0':
    report_id = ''

  c = _appflags(request)
  c.update(_get_context(request))
  mobile = c['mobile']
  if mobile:
    # App builder is not available in mobile mode - return app in view mode
    # instead
    return app(request, report_id=report_id)

  report_type = request.GET.get("report_type", 'dashboard')

  if report_type == 'dashboard':
    # Handle new dashboard requests
    return dashboardedit(request, report_id=report_id)

  ds = None
  if report_id != '':
    # pylint: disable=E1101
    thisapp = Report.objects.get(id=int(report_id))

    ds = thisapp.dataset
    if ds != None:
      ds.checkPerm(request.user, 'ds_appedit')

    if thisapp.workspace:
      thisapp.workspace.checkPerm(request.user, 'ws_appedit')

    if thisapp.report_type == 'dashboard':
      # Editing a dashboard - redirect
      return dashboardedit(request, report_id=report_id)

  dataset_id = request.GET.get("dataset_id", ds.id if ds else None)
  # TODO:  When is this dataset_id different that ds.id ???
  #        If different, we need to check that one as well

  template = loader.get_template('app_detail.html')
  c['sheet_id'] = ''
  c['is_edit'] = True
  c['dataset_id'] = dataset_id
  c['report_id'] = report_id
  c['disable_v1_bars'] = getattr(settings, 'DISABLE_V1_BARS', True)
  c['globalFilters'] = json.dumps([])
  c['dashboard_save_template_enabled'] = \
        getattr(settings, 'DASHBOARD_TEMPLATE', False) and \
        getattr(settings, 'DASHBOARD_SAVE_AS_TEMPLATE', False)
  return HttpResponse(template.render(c, request))

# get full list of AVs for a dataconnection along with base table info and job info
# about when they are scheduled
@login_required
def aview_list(request, dc_id=None):
  from jobs.views import _get_next_scheduled_refresh
  if not dc_id:
    return HttpResponse('Invalid request', status=500)

  # Actual dataconnection model object
  django_dc = DataConnection.objects.get(pk=dc_id)
  django_dc.checkPerm(request.user, 'dc_aviews')

  dataconn = get_dataconn(connection_id=dc_id, django_dc=django_dc)

  table_dict = {}
  datarequest_params = {'type': SPECIAL_CMDS['SHOW_AVIEWS']}
  raw_resp = get_existing_tables(datarequest_params, table_dict, request.user, dataconn, aviews=True, return_raw=True)

  # maintain pre AE 2.12.1 behavior
  if len(raw_resp.get('colnames')) != 3:
    return HttpResponse(json_dumps(raw_resp), content_type="application/json")

  refresh_schedule = _get_next_scheduled_refresh(dc_id)
  rows = []
  for name, table in table_dict.items():
    next_schedule = refresh_schedule.get(name)
    if next_schedule:
      table['nextScheduledRefresh'] = next_schedule.isoformat()
    else:
      table['nextScheduledRefresh'] = None

    raw_dims = table.get('dimensions')
    table['dimensions'] = [dim.get('name') for dim in raw_dims]
    table['dimension_cardinalities'] = [dim.get('cardinality') for dim in raw_dims]

    table['base'] = table.get('base_lview_name', table.get('base_table_name'))

    rows.append([name, table])

  data = {
    'colnames': ['name', 'Metadata'],
    'rows': rows,
  }

  return JsonResponse(data)
