# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.template import loader
from django.views.decorators.csrf import csrf_exempt
from django.contrib.auth.models import User
from django.utils import timezone, dateparse
from django.http import HttpResponse, HttpResponseRedirect, QueryDict

import json
from rbac.permdefs import PermissionException
from sqlrun import views
from .models import APIKey
from .models import lookup_apikey
from .models import _generate_apikey, _default_expire_time
from rbac.permcache import RBAC
from .utils import _appflags
from vizlib.misc import check_login_or_use_apikey
import re
import logging
from datasets.models import DataConnection

LOGGER = logging.getLogger("arcweb")

def _dataapi_default_limit():
  if not hasattr(settings, 'DATA_API_DEFAULT_ROW_LIMIT'):
    return 1000

  return settings.DATA_API_DEFAULT_ROW_LIMIT

def _dataapi_max_limit():
  if not hasattr(settings, 'DATA_API_MAX_ROW_LIMIT'):
    return 1000

  return settings.DATA_API_MAX_ROW_LIMIT

#
# Support a data api that lets users specify a dataset id, dimensions,
#     aggregates, and filters
#
@csrf_exempt
def dataapi(request):
  if getattr(settings, 'DATA_API_ENABLED', False) == False:
    LOGGER.warning("DataAPI called but not enabled")
    return HttpResponse('DataAPI is not enabled', status=500)

  LOGGER.info("DataAPI called")

  err = check_login_or_use_apikey(request)
  if (err != None):
    return err

  if request.method == 'GET':
    requestParams = request.GET
  elif request.method == 'POST':
    requestParams = request.POST

  query = requestParams.get('query')
  dataconnection_id = requestParams.get('dataconnection_id')
  
  if query and dataconnection_id:
      dc = DataConnection.objects.get(id=dataconnection_id)
      if 'dc_explore' not in dc.getPerms(request.user):
          return HttpResponse(status=403)

      params = {
        'query': query,
        'dataconnection_id': dataconnection_id
      }
  else:
    dataset = requestParams.get("dataset", None)
    if dataset is None:
      LOGGER.warning('DataAPI: dataset not specified')
      return HttpResponse('dataset not specified', status=500)

    dataset = int(dataset)

    limit = requestParams.get("limit", None)
    if limit is None:
      limit = _dataapi_default_limit()
    else:
      limit = int(limit)
      if limit > _dataapi_max_limit():
        limit = _dataapi_max_limit()

    dimensions = requestParams.get("dimensions", None)
    if dimensions is None:
      diml = []
    else:
      diml = dimensions.split(',')

    aggregates = requestParams.get("aggregates", None)
    if aggregates is None:
      aggl = []
    else:
      aggl = aggregates.split(',')

    filters = requestParams.get("filters", None)
    filtersl = []
    filterexpr = []
    if filters is not None:
      filtersl = filters.split(',')

    if not diml and not aggl:
      LOGGER.warning('DataAPI: columns not specified')
      return HttpResponse('columns not specified', status=500)

    dsreq = {
      'version': 1,
      'type': 'SQL',
      'limit': limit,
      'dimensions': [{'type': 'SIMPLE', 'expr': '[' + x + ']'} for x in diml],
      'aggregates': [{'expr': x} for x in aggl],
      'filters': filtersl,
      'dataset_id': dataset,
    }

    params = {
      'dsreq': json.dumps(dsreq),
    }

  data = views.jsonselect_work(request, params)
  if 'data' in data:
    return HttpResponse(json.dumps(data['data']), content_type="text/json")

  if 'error' in data:
    LOGGER.warning('DataAPI: Error in query response: %s' % data['error'])
    return HttpResponse(data['error'], status=500)

  LOGGER.error('DataAPI: Unexpected response')
  return HttpResponse('Unexpected response', status=500)




# APIKey REST API below

def _apikeys_get(apikey_hash=None):
  if apikey_hash:
    apikeys = APIKey.objects.filter(apikey_hash=apikey_hash)
  else:
    apikeys = APIKey.objects.all()
  data = [ i.as_dict for i in apikeys ]
  return HttpResponse(json.dumps(data), content_type="application/json")

def _apikeys_create_or_update(request):
  username = request.POST.get("username", None)
  apikey_hash = request.POST.get("apikey", None)
  active = request.POST.get("active", None)
  expires = request.POST.get("expires", None)

  new_key = None
  try:
    apikey = APIKey.objects.get(apikey_hash=apikey_hash)
  except:
    new_key = _generate_apikey()
    apikey = APIKey()
    apikey.set_apikey(new_key)

  try:
    user = User.objects.get(username=username)
  except:
    LOGGER.error("Error saving API Key: Invalid username")
    return HttpResponse("Invalid username", status=500)

  try:
    expire_time = dateparse.parse_datetime(expires)
  except:
    LOGGER.error("Error saving API Key: Invalid expire date")
    return HttpResponse("Invalid expire date", status=500)

  try:
    if active == 'true':
      active = 1
    if active == 'false':
      active = 0
    active = bool(int(active))
  except:
    LOGGER.error("Error saving API Key: Invalid active")
    return HttpResponse("Invalid active", status=500)

  apikey.user = user
  apikey.expire_time = expire_time
  apikey.active = active

  try:
    apikey.save()
  except Exception as e:
    LOGGER.error("Error saving API Key: %s", str(e))
    return HttpResponse("Error saving API Key", status=500)

  apikey_dict = apikey.as_dict
  apikey_dict["secret_apikey"] = new_key
  return HttpResponse(json.dumps(apikey_dict), content_type="application/json")


def _apikeys_delete(request, apikey=None):
  if apikey is not None:
    apikeys = APIKey.objects.filter(apikey_hash=apikey)
  else:
    apikeys_list = json.loads(QueryDict(request.body).get('apikeys_list', None))
    if apikeys_list is None:
      return HttpResponse("No apikeys selected", status=500)
    else:
      apikeys = APIKey.objects.filter(apikey_hash__in=apikeys_list)
  if len(apikeys) < 1:
    return HttpResponse("apikeys are not found", status=500)
  apikeys.delete()
  return HttpResponse('[]')


@login_required
def apikeys(request, apikey=None):
  """
  REST endpoint for apikeys
  """
  # All  requests need permission
  RBAC.checkPerm(request.user, 'sys_editperm')

  if request.method == 'GET':
    return _apikeys_get(apikey)

  if request.method == 'POST':
    return _apikeys_create_or_update(request)

  if request.method == 'DELETE':
    return _apikeys_delete(request)

  return HttpResponse('Invalid request method', status=500)



@login_required
def admin_apikeys(request):
  """manage apikeys"""

  RBAC.checkPerm(request.user, 'sys_editperm')

  c = _appflags(request)
  c["current_user"] = request.user
  c["new_expire"] = _default_expire_time().isoformat()

  template = loader.get_template('apps_apikeys.html')
  return HttpResponse(template.render(c, request))

@csrf_exempt
def demo_apikeys(request):
  """create apikeys for demo purposes"""
  if request.method != 'POST':
    return HttpResponse('Invalid request_type', status=500)

  try:
    user = User.objects.get(username=request.POST.get("username", None))
    RBAC.checkPerm(user, 'sys_editperm')
    if not user.check_password(request.POST.get("password", None)):
      LOGGER.error("Error creating API Key: Invalid password")
      return HttpResponse("Invalid authentication", status=500)
  except PermissionException as e:
      LOGGER.error(e)
      return HttpResponse(e, status=403)
  except:
    LOGGER.error("Error creating API Key: Invalid username")
    return HttpResponse("Invalid authentication", status=500)

  return _apikeys_create_or_update(request)
