# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Home page related requests
"""
import datetime
import json
import logging
import re

from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.db.models import Count, Sum
from django.http import JsonResponse, HttpRequest
from django.db.models.manager import BaseManager
from django.views.decorators.http import etag
from django.views.decorators.cache import cache_control


from arcwebbase.etag_caching import etag_metadata_func
from datasets.models import DataSet, DataConnection
from extracts.views import ExtractTracker
from reports.models import Report, AppGroup, ReportUsage, ReportImage, AppGroupAppMembership
from reports.views import _get_report_data
from reports.views_appgroups import get_appgroup_data, appgroup_starred
from sql.models import Query
from typing import Callable
from util.get_dataset_ids_accessible_by_user import get_dataset_ids_accessible_by_user
from util.get_workspace_ids_accessible_by_user import get_workspace_ids_accessible_by_user
from util.get_sample_dashboards_query import get_sample_dashboards_query

import os

LOGGER = logging.getLogger('arcweb')

COLLECT_RELATED_USING_PREFETCH = os.getenv('COLLECT_RELATED_USING_PREFETCH', 'true') == 'true'

def collect_related(report_query: BaseManager[Report]):
  """
  Add the necessary related data to a Report query. This might change later, right now we need
  full dataset, dataconnection and workspace information.

  For many reports prefetch_related might be more effective, despite of more queries (the amount of data
  transferred is the minimal).

  For few reports select_related might be faster (a single query, with eventual duplicated data)
  """

  if COLLECT_RELATED_USING_PREFETCH:
    return report_query.prefetch_related('dataset', 'dataset__dataconnection', 'workspace')
  else:
    return report_query.select_related('dataset', 'dataset__dataconnection', 'workspace')

@login_required
@cache_control(max_age=120)
def get_statistics(request: HttpRequest):
  dataset_ids_accessible_by_user = get_dataset_ids_accessible_by_user(request.user)
  workspace_ids_accessible_by_user = get_workspace_ids_accessible_by_user(request.user)

  dashboards_viewable_by_user = _get_dashboards_viewable_by_user(
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user
  )

  statistics = do_get_statistics(
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user,
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    user=request.user
  )

  return JsonResponse(statistics)

def _get_number_of_dataconnections_viewable_by_user(user: User, user_dataset_ids=None):
  if user.is_superuser:
    return DataConnection.objects.count()

  valid_dataconnection_ids = set()
  for dataconnection in DataConnection.objects.only('dataconnection_type', 'create_user').all():
    perms = dataconnection.getPerms(user)
    if len(perms) > 0:
      valid_dataconnection_ids.add(dataconnection.id)

  user_connection_ids = DataSet.objects.filter(id__in=user_dataset_ids)\
    .values_list('dataconnection_id', flat=True).distinct()

  for dc_id in user_connection_ids:
    valid_dataconnection_ids.add(dc_id)

  return len(valid_dataconnection_ids)

@login_required
@etag(etag_metadata_func)
def get_number_of_recently_created_resources(request):
  dataset_ids_accessible_by_user = get_dataset_ids_accessible_by_user(request.user)
  workspace_ids_accessible_by_user = get_workspace_ids_accessible_by_user(request.user)

  dashboards_viewable_by_user = _get_dashboards_viewable_by_user(
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user
  )

  number_of_recently_created_resources = do_get_number_of_recently_created_resources(
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user,
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
  )

  return JsonResponse(number_of_recently_created_resources)

number_of_block_tiles = 10

@login_required
@etag(etag_metadata_func)
def get_recently_created_dashboards_by_user(request):
  dataset_ids_accessible_by_user = get_dataset_ids_accessible_by_user(request.user)
  workspace_ids_accessible_by_user = get_workspace_ids_accessible_by_user(request.user)

  dashboards_viewable_by_user = _get_dashboards_viewable_by_user(
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user
  )

  recently_created_dashboards_by_user = do_get_recently_created_dashboards_by_user(
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    user=request.user
  )

  return JsonResponse({ 'data': recently_created_dashboards_by_user })

@login_required
@cache_control(max_age=120)
def get_last_viewed_reports_by_user(request):
  dataset_ids_accessible_by_user = get_dataset_ids_accessible_by_user(request.user)
  workspace_ids_accessible_by_user = get_workspace_ids_accessible_by_user(request.user)

  dashboards_viewable_by_user = _get_dashboards_viewable_by_user(
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user
  )

  last_viewed_reports_by_user = do_get_last_viewed_reports_by_user(
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    user=request.user
  )

  return JsonResponse({ 'data': last_viewed_reports_by_user })

def get_accessible_appgroups(accessible_workspace_ids: list[int]):
  apps = AppGroup.objects \
    .filter(group_type__in=["default", "sample-app"], group_parent_id=None)

  if accessible_workspace_ids is not None:
    apps = apps.filter(workspace_id__in=accessible_workspace_ids)

  return apps

@login_required
@cache_control(max_age=120)
def get_most_popular_reports(request):
  dataset_ids_accessible_by_user = get_dataset_ids_accessible_by_user(request.user)
  workspace_ids_accessible_by_user = get_workspace_ids_accessible_by_user(request.user)

  dashboards_viewable_by_user = _get_dashboards_viewable_by_user(
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user
  )

  most_popular_reports = do_get_most_popular_reports(
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user,
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    user=request.user
  )
  return JsonResponse({ 'data': most_popular_reports })

@login_required
@etag(etag_metadata_func)
def get_samples(request):
  dataset_ids_accessible_by_user = get_dataset_ids_accessible_by_user(request.user)
  workspace_ids_accessible_by_user = get_workspace_ids_accessible_by_user(request.user)

  samples_dashboards = do_get_samples_dashboards(
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user,
    user=request.user
  )

  sample_apps = do_get_sample_apps(request.user)

  return JsonResponse({ 'data': sample_apps + samples_dashboards })

@login_required
@etag(etag_metadata_func)
def get_favorite_dashboards(request):
  dataset_ids_accessible_by_user = get_dataset_ids_accessible_by_user(request.user)
  workspace_ids_accessible_by_user = get_workspace_ids_accessible_by_user(request.user)

  dashboards_viewable_by_user = _get_dashboards_viewable_by_user(
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user
  )

  favorite_dashboards = do_get_favorite_dashboards(
    request=request,
    dashboards_viewable_by_user=dashboards_viewable_by_user
  )

  return JsonResponse({
    "data": favorite_dashboards,
  })

def _get_dashboards_viewable_by_user(dataset_ids_accessible_by_user=None, workspace_ids_accessible_by_user=None):
  query = Report.objects \
    .filter(report_type='dashboard') \
    .exclude(report_mode='template')

  if dataset_ids_accessible_by_user is not None:
    query = query.filter(dataset_id__in=dataset_ids_accessible_by_user)

  if workspace_ids_accessible_by_user is not None:
    query = query.filter(workspace_id__in=workspace_ids_accessible_by_user)

  return query

def _get_snapshot_update_times_by_report_id(report_ids: list[int]):
  snapshots = ReportImage.objects.filter(report_id__in=report_ids).values_list('report__id', 'update_time')

  snapshot_update_time_by_report_id = {}

  if snapshots is not None:
    for report_image in snapshots:
      report_id = report_image[0]
      update_time = report_image[1].strftime("%Y.%m.%d.%H.%M.%S")
      snapshot_update_time_by_report_id[report_id] = update_time

  return snapshot_update_time_by_report_id

def _get_reports_by_app_or_dashboard_ids_getter(user: User, snapshot_update_times_by_report_id: dict[int, str], dashboards_viewable_by_user: BaseManager[Report]):
  return lambda app_or_dashboard_ids: _get_reports_by_app_or_dashboard_ids(
    app_or_dashboard_ids=app_or_dashboard_ids,
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    get_report_data=_get_dashboard_data_getter(user, snapshot_update_times_by_report_id),
    get_app_data=_get_app_data_getter(user, snapshot_update_times_by_report_id)
  )

def _get_reports_by_app_or_dashboard_ids(
  app_or_dashboard_ids: list[int],
  dashboards_viewable_by_user: BaseManager[Report],
  get_report_data: Callable,
  get_app_data: Callable
):
  dashboard_ids = [app_or_dashboard_id for app_or_dashboard_id in app_or_dashboard_ids if app_or_dashboard_id > 0]
  dashboards = dashboards_viewable_by_user.filter(id__in=dashboard_ids)
  dashboards = collect_related(dashboards)
  dashboard_by_dashboard_ids = dict([(dashboard.id, dashboard) for dashboard in dashboards])

  app_ids = [-app_or_dashboard_id for app_or_dashboard_id in app_or_dashboard_ids if app_or_dashboard_id < 0]
  # num_apps is read by get_appgroup_data, permissions are also enforced there (although we might be able to do it better here)
  apps = AppGroup.objects.filter(id__in=app_ids).select_related('workspace').annotate(num_apps=Count('appgroupappmembership'))
  app_by_app_ids = dict([(-app.id, app) for app in apps])

  reports = []

  for report_id in app_or_dashboard_ids:
    is_dashboard = report_id > 0

    report_data = None
    if is_dashboard:
      dashboard = dashboard_by_dashboard_ids.get(report_id)
      if dashboard:
        report_data = get_report_data(dashboard)
    else:
      app = app_by_app_ids.get(report_id)
      if app:
        report_data = get_app_data(app)

    if report_data:
      reports.append(report_data)

  return reports

def _get_dashboard_data_getter(user, snapshot_update_times_by_report_id):
  extractTracker = ExtractTracker()

  def get_dashboard_data(dashboard):
    report_data = _get_report_data(dashboard, user, cached_ril=snapshot_update_times_by_report_id)

    return {
      "id": report_data.get('id'),
      "report": {
        "id": report_data.get('id'),
        "name": report_data.get('name'),
        "title": report_data.get('title'),
        "type": "dashboard",
        "baseImg": report_data.get('baseImg'),
        "img": report_data.get('img'),
        "editUrl": report_data.get('edit_url'),
        "url": report_data.get('url'),
        "updated": {
          "date": report_data.get('updated'),
          "by": report_data.get('updated_by')
        },
        "dataset": {
          "id": report_data.get('dataset_id'),
          "name": report_data.get('dataset_name')
        },
        "workspace": report_data.get('workspace_name'),
        "permissions": report_data.get('perm'),
      },
      "datasetInfo": {
        "name": dashboard.dataset.dataset_name,
        "extract": {
          "isSource": extractTracker.isExtractSource(dashboard.dataset),
          "isTarget": extractTracker.isExtractTarget(dashboard.dataset)
        }
      }
    }

  return get_dashboard_data

def _get_app_data_getter(user, snapshot_update_times_by_report_id):
  def get_app_data(app: AppGroup):
    report_data = get_appgroup_data(app, user, cached_ril=snapshot_update_times_by_report_id)

    return {
      "id": report_data.get('id'),
      "report": {
        "id": report_data.get('id'),
        "name": report_data.get('name'),
        "title": report_data.get('title'),
        "type": "app",
        "baseImg": report_data.get('baseImg'),
        "editUrl": report_data.get('edit_url'),
        "url": report_data.get('url'),
        "updated": {
          "date": report_data.get('updated'),
          "by": report_data.get('updated_by')
        },
        "workspace": report_data.get('workspace_name'),
        "permissions": report_data.get('perm'),
        "rw": report_data.get('rw')
      }
    }

  return get_app_data

def _get_reports_last_viewed_by_user(user):
  report_ids_last_viewed_by_user = []

  try:
    user_data = json.loads(user.userprofile.user_data)
    report_ids_last_viewed_by_user = [int(report_id) for report_id in user_data['last_viewed']]
    report_ids_last_viewed_by_user.reverse()
  except:
    pass

  return report_ids_last_viewed_by_user

@login_required
@etag(etag_metadata_func)
def get_datasets(request):
  limit = int(request.GET.get('limit'))
  datasets_created_by_user = DataSet.objects.filter(create_user=request.user.username).order_by('-update_time')[:limit]

  datasets = []
  for dataset in datasets_created_by_user:
    datasets.append({
      "id": dataset.id,
      "name": _get_datasetname_or_return_query_if_autogen(dataset),
      "tables": dataset.get_tablenames(),
      "numberOfFields": len(json.loads(dataset.dataset_info)[0]['columns']),
    })

  return JsonResponse({ "datasets":  datasets })

def _get_datasetname_or_return_query_if_autogen(dataset):
  if dataset.dataset_description == "Auto-created for SQL apps" and \
    (re.fullmatch("autogen [a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}", dataset.dataset_name) != None) and \
    dataset.dataset_settings == "{}":
    return json.loads(dataset.dataset_info)[0]['select']
  return dataset.dataset_name

@login_required
@etag(etag_metadata_func)
def get_home_page_info(request: HttpRequest):
  """
  This function is not exposed through an URL as of VIZ-7.2.6.
  It reflects an experimental stage where it would result in a lower database queries count for a non-admin user.

  The database query count decrease is the result of performing dataset and workspace access checks only once (as opposed
  to performing the checks for each of the other view functions in this module).

  It returns the union of the individual view function responses.

  At the time of writing this function generates 44 queries for an admin and 48 queries for a non-admin user.
  We might be able to reduce these numbers later on.

  """
  dataset_ids_accessible_by_user = get_dataset_ids_accessible_by_user(request.user)
  workspace_ids_accessible_by_user = get_workspace_ids_accessible_by_user(request.user)

  dashboards_viewable_by_user = _get_dashboards_viewable_by_user(
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user
  )

  statistics = do_get_statistics(
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user,
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    user=request.user
  )

  number_of_recently_created_resources = do_get_number_of_recently_created_resources(
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user,
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
  )

  recently_created_dashboards_by_user = do_get_recently_created_dashboards_by_user(
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    user=request.user
  )

  last_viewed_reports_by_user = do_get_last_viewed_reports_by_user(
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    user=request.user
  )

  most_popular_reports = do_get_most_popular_reports(
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user,
    dashboards_viewable_by_user=dashboards_viewable_by_user,
    user=request.user
  )

  samples_dashboards = do_get_samples_dashboards(
    dataset_ids_accessible_by_user=dataset_ids_accessible_by_user,
    workspace_ids_accessible_by_user=workspace_ids_accessible_by_user,
    user=request.user
  )

  sample_apps = do_get_sample_apps(request.user)

  favorite_dashboards = do_get_favorite_dashboards(
    request=request,
    dashboards_viewable_by_user=dashboards_viewable_by_user
  )

  return JsonResponse({
    "statistics": statistics,
    "number_of_recently_created_resources": number_of_recently_created_resources,
    "recently_created_dashboards_by_user": recently_created_dashboards_by_user,
    "last_viewed_reports_by_user": last_viewed_reports_by_user,
    "most_popular_reports": most_popular_reports,
    "samples": sample_apps + samples_dashboards,
    "favorite_dashboards": favorite_dashboards
  })

def do_get_statistics(dashboards_viewable_by_user, workspace_ids_accessible_by_user, dataset_ids_accessible_by_user, user):
  # For now, use the number of appgroups as the number of apps.
  apps = get_accessible_appgroups(workspace_ids_accessible_by_user)

  # TODO check subselect:
  views = ReportUsage.objects \
      .filter(report__in=dashboards_viewable_by_user.all()) \
      .aggregate(sum=Sum('count'))

  queries_created_by_user = Query.objects\
    .filter(create_user=user.username) \
    .order_by('-create_time').count()

  if dataset_ids_accessible_by_user is not None:
    dataset_count = len(dataset_ids_accessible_by_user)
  else:
    # we are superuser, have to count them:
    dataset_count = DataSet.objects.count()

  dataconnections_count = _get_number_of_dataconnections_viewable_by_user(user, dataset_ids_accessible_by_user)

  dashboards_count = dashboards_viewable_by_user.all().count()

  return {
    "dashboards": dashboards_count,
    "datasets": dataset_count,
    "apps": apps.count(),
    "views": views['sum'],
    "dataConnections": dataconnections_count,
    "queries": queries_created_by_user,
  }

def do_get_number_of_recently_created_resources(dashboards_viewable_by_user, dataset_ids_accessible_by_user, workspace_ids_accessible_by_user):
  today = datetime.date.today()
  timedelta = datetime.timedelta(days=int(7))

  dashboards_created_recently = dashboards_viewable_by_user.filter(create_time__gt=(today-timedelta))

  datasets_viewable_by_user = DataSet.objects.only('id')
  if dataset_ids_accessible_by_user is not None:
    datasets_viewable_by_user = datasets_viewable_by_user.filter(id__in=dataset_ids_accessible_by_user)
  datasets_created_recently = datasets_viewable_by_user.filter(create_time__gt=(today-timedelta))

  # For now, use created appgroups as number of apps
  apps_created_recently = get_accessible_appgroups(workspace_ids_accessible_by_user).filter(create_time__gt=(today-timedelta))

  return {
    "datasets": datasets_created_recently.count(),
    "dashboards": dashboards_created_recently.count(),
    "apps": apps_created_recently.count()
  }

def do_get_recently_created_dashboards_by_user(dashboards_viewable_by_user: BaseManager[Report], user):
  dashboards_created_by_user = dashboards_viewable_by_user.filter(create_user=user)
  dashboards_created_by_user = dashboards_created_by_user.order_by("-create_time")[:number_of_block_tiles]
  dashboards_created_by_user = collect_related(dashboards_created_by_user)

  dashboard_ids = [dashboard.id for dashboard in dashboards_created_by_user]
  snapshot_update_times_by_report_id = _get_snapshot_update_times_by_report_id(dashboard_ids)
  get_dashboard_data = _get_dashboard_data_getter(user=user, snapshot_update_times_by_report_id=snapshot_update_times_by_report_id)

  return [get_dashboard_data(dashboard) for dashboard in dashboards_created_by_user]

def do_get_last_viewed_reports_by_user(dashboards_viewable_by_user, user):
  last_viewed_report_ids_by_user = _get_reports_last_viewed_by_user(user=user)[:number_of_block_tiles]

  snapshot_update_times_by_report_id = _get_snapshot_update_times_by_report_id(last_viewed_report_ids_by_user)
  get_reports_by_app_or_dashboard_ids_getter = _get_reports_by_app_or_dashboard_ids_getter(user, snapshot_update_times_by_report_id, dashboards_viewable_by_user)

  return get_reports_by_app_or_dashboard_ids_getter(app_or_dashboard_ids=last_viewed_report_ids_by_user)

def do_get_most_popular_reports(workspace_ids_accessible_by_user, dashboards_viewable_by_user, user):
  apps_viewable_by_user = get_accessible_appgroups(workspace_ids_accessible_by_user)

  dashboard_and_app_ids_viewable_by_user = list(dashboards_viewable_by_user.values_list('id', flat=True))
  dashboard_and_app_ids_viewable_by_user.extend([-app['id'] for app in apps_viewable_by_user.values('id')])
  most_popular_dashboard_and_app_ids = list(
    ReportUsage.objects \
      .values('report_or_appgroup_id') \
      .filter(report_or_appgroup_id__in=dashboard_and_app_ids_viewable_by_user) \
      .annotate(num_views=Sum('count')) \
      .order_by('-num_views')[:number_of_block_tiles] \
      .values_list('report_or_appgroup_id', flat=True)
  )

  snapshot_update_times_by_report_id = _get_snapshot_update_times_by_report_id(most_popular_dashboard_and_app_ids)
  get_reports_by_app_or_dashboard_ids_getter = _get_reports_by_app_or_dashboard_ids_getter(user, snapshot_update_times_by_report_id, dashboards_viewable_by_user)

  return get_reports_by_app_or_dashboard_ids_getter(app_or_dashboard_ids=most_popular_dashboard_and_app_ids)

def do_get_samples_dashboards(dataset_ids_accessible_by_user, workspace_ids_accessible_by_user, user):
  try:
    samples_dashboards = get_sample_dashboards_query(
      dataset_ids_accessible_by_user,
      workspace_ids_accessible_by_user
    )
    samples_dashboards = samples_dashboards.order_by("create_time")[:number_of_block_tiles]
    samples_dashboards = collect_related(samples_dashboards)

    dashboard_ids = [dashboard.id for dashboard in samples_dashboards]
    snapshot_update_times_by_report_id = _get_snapshot_update_times_by_report_id(dashboard_ids)
    get_dashboard_data = _get_dashboard_data_getter(user=user, snapshot_update_times_by_report_id=snapshot_update_times_by_report_id)

    samples_dashboards = [get_dashboard_data(dashboard) for dashboard in samples_dashboards]
  except:
    samples_dashboards = []

  return samples_dashboards

def do_get_sample_apps(user: User):
  sample_apps = [app for app in AppGroup.objects.filter(group_type='sample-app')]

  if len(sample_apps) == 0:
    return []

  snapshot_update_times_by_report_id = _get_snapshot_update_times_by_report_id([app.id for app in sample_apps])

  get_app_data = _get_app_data_getter(user, snapshot_update_times_by_report_id)

  return [get_app_data(app) for app in sample_apps]

def do_get_favorite_dashboards(request, dashboards_viewable_by_user: BaseManager[Report]):
  favorites_appgroup = appgroup_starred(request)
  favorite_dashboard_ids = list(
      AppGroupAppMembership.objects \
        .filter(appgroup_group__id=favorites_appgroup.id) \
        .order_by('appgroup_order') \
        .values_list('appgroup_app', flat=True)
  )
  favorite_dashboards = dashboards_viewable_by_user.filter(id__in=favorite_dashboard_ids)
  favorite_dashboards = collect_related(favorite_dashboards)

  dashboard_ids = [dashboard.id for dashboard in favorite_dashboards]
  snapshot_update_times_by_report_id = _get_snapshot_update_times_by_report_id(dashboard_ids)
  get_dashboard_data = _get_dashboard_data_getter(user=request.user, snapshot_update_times_by_report_id=snapshot_update_times_by_report_id)

  return [get_dashboard_data(dashboard) for dashboard in favorite_dashboards]