# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, StreamingHttpResponse
from django.conf import settings

import vizlib.querylog as querylog

from rbac.permcache import RBAC

from dataconn.dataconn import get_dataconn
from util.log_bundling import main as log_bundling

import os
import json
import time
import tempfile
import dateutil.parser
import tarfile
import shutil
import datetime
import csv

MAX_QUERYLOGS_COUNT = 100000

def __processLogEntry(d, data, max_days=120):
  username = d.get('username')
  ts = datetime.datetime.strptime(d.get('ts', '').split()[0], '%Y-%m-%d')
  today = datetime.datetime.today()
  if (today - ts).days >= max_days:
    return True

  ts_str = ts.strftime('%Y-%m-%d');
  if not data['activities'].get(ts_str):
    data['activities'][ts_str] = 1
  else:
    data['activities'][ts_str] += 1

  if not data['users'].get(ts_str):
    data['users'][ts_str] = {}

  if username:
    if not data['users'][ts_str].get(username):
      data['users'][ts_str][username] = 1
    else:
      data['users'][ts_str][username] += 1

  dashboard_id = d.get('app') or d.get('visual')
  if not dashboard_id:
    return
  
  # In very rare cases somehow the logs contains the app or visual ids as string not as number
  # to avoid duplication in the code, I convert it to integer
  if isinstance(dashboard_id, str):
    dashboard_id = int(dashboard_id)

  if not data['dashboards'].get(ts_str):
    data['dashboards'][ts_str] = {}

  if not data['dashboards'][ts_str].get(dashboard_id):
    data['dashboards'][ts_str][dashboard_id] = 1
  else:
    data['dashboards'][ts_str][dashboard_id] += 1

def __toCsv(data):
  rows = []

  class Writer:
    def write(self, str):
      rows.append(str)

  w = csv.writer(Writer())
  w.writerow(['type', 'day', 'id', 'count']);

  for day in sorted(data['activities'].keys(), reverse=True):
    day_data = data['activities'][day]
    w.writerow(['Activities', day, day, day_data])

  for day in sorted(data['users'].keys(), reverse=True):
    day_data = data['users'][day]
    for user, count in day_data.items():
      w.writerow(['Users', day, user, count])

  for day in sorted(data['dashboards'].keys(), reverse=True):
    day_data = data['dashboards'][day]
    for dashboard_id, count in day_data.items():
      w.writerow(['Dashboards', day, dashboard_id, count])

  return ''.join(rows)

@login_required
def activeusers(request):
  """Fetch list of active users by day"""

  RBAC.checkPerm(request.user, 'sys_viewlogs')

  days = int(request.GET.get('days', '30'))

  data = {
    'users': {},
    'dashboards': {},
    'activities': {},
  }

  done = False

  # First, read ongoing queries
  for logEntry in querylog.read_ongoing_queries():
    if __processLogEntry(logEntry, data, max_days=days):
      done = True
      break

  # If we are not at the limit, then read historical queries if we can
  logfileIndex = None
  while not done:
    logReader = querylog.LogReader(fileindex=logfileIndex)
    if not logReader.file_exists():
      break

    if logfileIndex is None:
      logfileIndex = 1
    else:
      logfileIndex += 1

    for logEntry in logReader.read_json_backwards():
      if __processLogEntry(logEntry, data, max_days=days):
        done = True
        break

  return HttpResponse(__toCsv(data), content_type="text/plain")

@login_required
def querylogs(request, querycount=100):
  """Fetch last few activity logs"""

  RBAC.checkPerm(request.user, 'sys_viewlogs')

  app_id = request.GET.get('app_id', None)
  if app_id:
    app_id = int(app_id)

  visuals_only = (request.GET.get('visuals_only', '') == "true");
  edits_only = (request.GET.get('edits_only', '') == "true");
  searchl = [x.lower() for x in request.GET.get('search', '').split()]

  def allowLog(logEntry):
    if app_id and app_id != logEntry.get('app'):
      return False

    if visuals_only and not logEntry.get('visual'):
      return False

    if edits_only and logEntry.get('action') == querylog.ActionType.READ:
      return False

    if searchl:
      l = logEntry.copy()
      if l.get('action_id'):
        del l['action_id']

      if l.get('component_type'):
        l['component_type'] = querylog.ComponentType.getStr(l.get('component_type'))

      if l.get('action'):
        l['action'] = querylog.ActionType.getStr(l.get('action'))

      if l.get('query_state'):
        l['query_state'] = querylog.QueryState.getStr(l.get('query_state'))

      d = ' '.join([str(x) for x in l.values()]).lower()

      for s in searchl:
        if d.find(s) < 0:
          return False

    return True

  querycount = int(querycount)
  if querycount < 1:
    return HttpResponse('Invalid query count %s' % querycount, status=500)

  if querycount > MAX_QUERYLOGS_COUNT:
    querycount = MAX_QUERYLOGS_COUNT

  tracked = {}

  # First, read ongoing queries
  in_data = querylog.read_ongoing_queries()
  data = []
  for logEntry in in_data:
    if not allowLog(logEntry):
      continue

    data.append(logEntry)
    qid = logEntry.get('action_id')
    if qid:
      tracked[qid] = True

  # If we are not at the limit, then read historical queries if we can
  logfileIndex = None
  while len(data) < querycount:
    logReader = querylog.LogReader(fileindex=logfileIndex)
    if not logReader.file_exists():
      break

    if logfileIndex is None:
      logfileIndex = 1
    else:
      logfileIndex += 1

    for logEntry in logReader.read_json_backwards():
      if not allowLog(logEntry):
        continue

      qid = logEntry.get('action_id')
      if qid:
        if tracked.get(qid):
          # Don't fetch older logs for the same query id
          continue

      logReader.markStartedQueryCanceled(logEntry)

      data.append(logEntry)
      tracked[qid] = True
      if len(data) >= querycount:
        break

  return HttpResponse(json.dumps(data), content_type="application/json")

@login_required
def serverlogs(request):

  if request.method == 'POST':
    parms = request.POST
  elif request.method == 'GET':
    parms = request.GET
  else:
    return HttpResponse('Invalid request', status=500);

  RBAC.checkPerm(request.user, 'sys_viewlogs')

  # Helper for file access error return
  def err(text):
    return HttpResponse('Logging settings error: %s' % text, status=500)

  # Helper function to return a file download response
  def logfileResponse(logfile, prefix):
    logfile.seek(0, os.SEEK_END)
    size = logfile.tell()
    logfile.seek(0, os.SEEK_SET)

    timetext = time.strftime('%Y_%m_%d_%H_%M_%S')
    response = StreamingHttpResponse(logfile)
    response['Content-Length'] = size
    response['Content-Disposition'] = 'attachment; filename="%s_%s.log"' % (prefix, timetext)

  # Get depreciation warning with this method
  # wrapper = FileWrapper(logfile)
  # response = HttpResponse(wrapper, content_type='text/plain')

    return response

  # Data connection log getter
  dcid = parms.get('dcid')
  if dcid is not None:

    if not getattr(settings, 'ARCENGINE_LOG_UI_ENABLE', False):
      return HttpResponse('Invalid request', status=500)

    dc = get_dataconn(dcid)

    try:
      if request.method == 'GET':

        # Read the full log into a temp file
        logfile = tempfile.NamedTemporaryFile('r+');
        ctoken = None
        # Limit to last 100mb at 4mb per fetch
        startp = 100 * 1024 * 1024
        max_fetch = 4 * 1024 * 1024
        while True:
          rsp = dc.serverlog(username=request.user.username, max_fetch=max_fetch, ctoken=ctoken, startp=startp)
          if len(rsp['data']):
            ctoken = rsp['ctoken']
            logfile.write(rsp['data'])
            startp = None
          else:
            break
        return logfileResponse(logfile, 'arcengine')

      ctoken = parms.get('ctoken')
      try:
        startp = int(parms.get('startp', '0'))
      except:
        startp = 0
      max_fetch = int(parms.get('max_fetch', '100000'))
      rsp = dc.serverlog(username=request.user.username,
                         max_fetch=max_fetch, ctoken=ctoken, startp=startp)
      return HttpResponse(json.dumps(rsp, ensure_ascii=getattr(settings, 'JSON_ENSURE_ASCII', False)), content_type="application/json")
    except Exception as e:
      if str(e) == "Connection reset: Invalid method name: 'GetServerLog'":
        return HttpResponse('Log access not supported on this Arcengine version.' +
                            '  Please upgrade to version 2.5.4 or later.', status=500)
      raise

  # Webserver log handling.  Currently just dealing with active
  # log file and not handling the issues around RotatingLogHandler, etc.

  if not getattr(settings, 'ARCVIZ_LOG_UI_ENABLE', False):
    return HttpResponse('Invalid request', status=500)

  logname = settings.ARCVIZ_LOGFILE
  if not logname:
    return err('no filename found')

  logfile = open(logname, 'r')

  if request.method == 'GET':
    return logfileResponse(logfile, 'arcviz')

  logfile.seek(0, os.SEEK_END)
  size = logfile.tell()
  logfile.seek(0, os.SEEK_SET)

  try:
    ctoken = int(parms.get('ctoken', '0'))
  except:
    ctoken = 0

  try:
    startp = int(parms.get('startp', '0'))
  except:
    startp = 0

  # Continuation takes precedence over starting point but if not
  # then startp indicates distance from end of log to start
  if not ctoken and startp and (startp < size):
    ctoken = size - startp

  max_fetch = int(parms.get('max_fetch', '100000'))

  offset = min(ctoken, size)
  count = min(size - offset, max_fetch)
  logfile.seek(offset, os.SEEK_SET)
  data = logfile.read(count)

  rsp = {'ctoken':logfile.tell(), 'data':data}

  return HttpResponse(json.dumps(rsp, ensure_ascii=getattr(settings, 'JSON_ENSURE_ASCII', False)), content_type="application/json")

@login_required
def bundledlogs(request):
  if request.method == 'GET':
    params = request.GET
    # params are expected to include the following keys:
    # cookie: JSON object that has these keys:
    #     name - name of cookie to be updated on successful log retrieval
    #     value - unique identifier for this particular request
    # dcName: name of data connection -- passed to options[dcname]
    # startTime: beginning of logging window in ISO-8601-- parsed and passed to options[start_time]
    # duration: number of seconds of logs to retrieve --passed to options[duration]
    # nodeList: comma-delimited list of nodes -- passed to options[nodes]
    # coordinatorsOnly: boolean of whether to restrict nodes-- passed to options[coordinators_only]
    # maxBytes: number of bytes to zip -- passed to options[max_bytes_to_zip]
    # logLevels: comma-delimited list of log levels to retrieve -- passed to options[levels]
    # processes: comma-delimited list of processes for which to retrieve logs -- passed to options[processes]
  else:
    return HttpResponse('Invalid request', status=500)

  RBAC.checkPerm(request.user, 'sys_viewlogs')

  # Valid option list specified in parse method of log_bundling.py
  options = {}

  cookie = json.loads(params.get('cookie'))
  cookie_name = cookie.get('name')
  cookie_base_value = json.loads(cookie.get('value'))
  dcName = params.get('dcName')

  if dcName is None:
    return HttpResponse('Must specify data connection', status=500)
  else:
    options['dcname'] = dcName

  start_time = params.get('startTime')
  if start_time is None:
    return HttpResponse('Must include start time', status=500)
  else:
    try:
      options['start_time'] = dateutil.parser.parse(start_time)
    except ValueError:
      return HttpResponse('Invalid start time -- should be in ISO-8601 format', status=500)

  duration = params.get('duration')
  if duration is None:
    return HttpResponse('Must include a duration', status=500)

  try:
    int_duration = int(duration)
  except ValueError:
    return HttpResponse('Duration must be an integer', status=500)

  if int_duration < 1:
    return HttpResponse('Duration must be at least one', status=500)

  options['duration'] = duration

  options['nodes'] = params.get('nodeList')

  if params.get('coordinatorOnly') == 'true':
    options['coordinators_only'] = True

  max_bytes = params.get('maxBytes')
  if max_bytes is not None:
    max_bytes = float(max_bytes)
    if max_bytes <= 0:
      return HttpResponse('Max megabytes to zip must be positive', status=500)
    options['max_bytes_to_zip'] = max_bytes * 10 ** 6 #convert from MB

  levels = params.get('logLevels')
  if levels is None:
    return HttpResponse('At least one log level must be specified', status=500)
  else:
    options['levels'] = levels

  processes = params.get('processes')
  if processes is None:
    return HttpResponse('At least one process must be specified', status=500)
  else:
    options['processes'] = processes

  directory_path = tempfile.mkdtemp()
  options['output_dir'] = directory_path
  response = HttpResponse()

  try:
    data = log_bundling(options=options, arcviz=True)
    files_for_tar = data.get('files')
    maxed_out = data.get('maxBytesHit')
  except Exception as e:
    msg = str(e)
    if msg == 'Connection reset: Invalid method name: \'GetGlogServerLogs\'':
      msg = 'Bundled log access not supported on this Arcengine version. Please upgrade to version 2.9.3 or later.'
    response = HttpResponse(msg)
    response.set_cookie(
      cookie_name, 
      json.dumps('%s ERROR: %s' % (cookie_base_value, msg)),
      samesite=settings.SESSION_COOKIE_SAMESITE,
      secure=settings.SESSION_COOKIE_SECURE
    )
    response['Content-Disposition'] = 'attachment; filename=error.txt'
    return response
  # bundle files and make tar ball available for download
  if len(files_for_tar) > 0:
    response = HttpResponse(content_type='application/x-gzip')
    archive_base_name = 'logs.%s' % str(int(time.time()))
    response['Content-Disposition'] = 'attachment; filename=%s.tar.gz' % archive_base_name
    archive = tarfile.open(fileobj=response, mode='w:gz')
    for filename, file in files_for_tar.items():
      path = os.path.join(archive_base_name, filename)
      archive.add(file.name, path)
    archive.close()
    # Used to alert ArcViz of download availability
    if maxed_out is True:
      limit_msg = 'WARNING: The logs requested exceed the \'Max Bytes to Zip\' limit of %s bytes and so were only ' \
        'partially downloaded. Consider restricting your search criteria or increasing the byte limit.' % max_bytes
      response.set_cookie(
        cookie_name, 
        json.dumps('%s %s' % (cookie_base_value, limit_msg)),
        samesite=settings.SESSION_COOKIE_SAMESITE,
        secure=settings.SESSION_COOKIE_SECURE
      )
    else:
      response.set_cookie(
        cookie_name, 
        json.dumps('%s SUCCESS' % cookie_base_value),
        samesite=settings.SESSION_COOKIE_SAMESITE,
        secure=settings.SESSION_COOKIE_SECURE
      )
  else:
    response = HttpResponse(status=204)
    msg = 'No log files met the search criteria. Consider expanding them.'
    response.set_cookie(
      cookie_name, 
      json.dumps('%s WARNING: %s' % (cookie_base_value, msg)),
      samesite=settings.SESSION_COOKIE_SAMESITE,
      secure=settings.SESSION_COOKIE_SECURE
    )

  shutil.rmtree(directory_path)

  return response
