# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
#
# Make remote calls to external REST apis and fetch data back for
# usage within visuals.
#
# The input here is a JSON structure with a few fields
# "accessKey"
# "request"
# "response_colname"
#
# accessKey and request are passed through to the REST call as accessKey and request
#   respectively.
#   - accessKey (from cviz_rest): Unique identifier for particular model
#   - request.data.colnames (from cviz_rest): Specifies which columns are passed to the rest api
#   - [OPTIONAL] request.config (from cviz_rest): Passes any configs to the model
# response_colname (from cviz_rest) is used to returns just the columns in the response_colname back
#   to the caller.
#
#
# The calls are POST requests made to a specified URL and the format
# of the call that is a JSON structure with the following structure
# {
#   "accessKey": "xxdsrwererdsfd",
#   "request": {
#     "params": {
#       "username": "username",
#       "key1": "value1",
#       "key2": "value2",
#     },
#     "config": {
#       "key1": "value1",
#       "key2": "value2",
#     },
#     "data": {
#       "colnames": ["name1", "name2"],
#       "coltypes": ["STRING", "INT"],
#       "rows": [
#         ["row1col1", "row1col2"],
#         ["row2col1", "row2col2"],
#         ["row3col1", "row3col2"],
#       ]
#     }
#   }
# }
#
#
# Responses are expected in the following structure:
# {
#   "response": {
#     "data": {
#       "colnames": ["name1", "name2"],
#       "coltypes": ["STRING", "INT"],
#       "rows": [
#         ["row1col1", "row1col2"],
#         ["row2col1", "row2col2"],
#         ["row3col1", "row3col2"],
#       ]
#     }
#   }
# }
#

from django.contrib.auth.decorators import login_required
from django.http import HttpResponse

import requests
import json

import vizlib.cviz_remotecall

@login_required
def fetchdata(request):
  data = json.loads(request.POST.get("data", None))
  response = []

  for call_setting in data:
    row_result = restCall(call_setting, request.user.username)
    response.append(row_result)

  return HttpResponse(json.dumps(response), content_type="application/json")

def restCall(call_setting, username):
  url = call_setting.get("url")
  err = vizlib.cviz_remotecall.validURL(url)
  if err is not None:
    raise Exception(f'Error with cviz_rest URL: {err}')

  if username:
    if not call_setting['payload'].get('params'):
      call_setting['payload']['params'] = {}

    call_setting['payload']['params']['username'] = username

  r = requests.post(url,
    headers={'Content-Type': 'application/json'},
    data=json.dumps(call_setting.get("payload")))

  response_colname = call_setting.get("response_colname", [])

  if r.status_code != 200:
    raise Exception(
        "Error return back from remote call: %s" % r.status_code)

  if response_colname:
    predictions = json.loads(r.text)
    colnames = predictions.get("response").get("data").get("colnames")
    idx_found = None
    for idx, colname in enumerate(colnames):
      if colname == response_colname:
        idx_found = idx
        break

    if idx_found is None:
      raise Exception("Expected column '%s' not returned in remote call" % \
          response_colname)

    rows = predictions.get("response").get("data").get("rows")
    return [row[idx_found] for row in rows]
