# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Views for URL ShortUrl
"""

from django.http import HttpResponse, JsonResponse
from django.contrib.auth.decorators import login_required
from django.db import IntegrityError
from django.db.models import Q
from django.core.paginator import Paginator
from django.shortcuts import get_object_or_404

from functools import reduce
import json
import logging
from urllib import parse

from apps.models_shortener import ShortUrl
from apps.views import app, appgroup


LOGGER = logging.getLogger('arcweb')
REPORT_TYPES_FILTERS = ["picklist", "daterange"]

@login_required
def urlShortener(request):
  if request.method == 'POST':
    return _post_short_url(request)

  elif request.method == 'GET':
    return _get_short_url(request)

  elif request.method == 'DELETE':
    return _delete_short_url(request)

  elif request.method == 'PATCH':
    return _update_short_url(request)
       
def _post_short_url(request):
  try:
    data = json.loads(request.body)
    dashboard_id = data['dashboardId']
    type = data['type']
    parameters = data['parameters']
    slug = data['slug']
    is_created_in_email = data['isCreatedInEmail']
    user = request.user

    params = dict(parse.parse_qsl(parse.urlsplit(parameters).path))

    if type == 'dashboard':
      report_id = dashboard_id
      app_id = None
    else:
      report_id = None
      app_id = dashboard_id

    shortener = ShortUrl.objects.create(
      url_data={
          'dashboardId': dashboard_id,
          'type': type,
          'parameters': params,
          'is_created_in_email': is_created_in_email,
      },
      slug=slug,
      creator=user,
      report_id=report_id,
      app_id=app_id,
    )
    shortener.save()

    return HttpResponse("success", content_type="text/plain")

  except IntegrityError as e:
    if 'unique constraint' in str(e):
      return HttpResponse(f"Slug '{slug}' already exists", status=500)
    else:
      return HttpResponse(str(e), status=500)

def _get_short_url(request):
  aliases_data = []
  aliases = ShortUrl.objects.select_related('creator', 'report', 'app').all()

  dashboard_id = request.GET.get('dashboardId')
  if dashboard_id and dashboard_id.isnumeric():
    shortened_rows = ShortUrl.objects.filter(url_data__dashboardId=int(dashboard_id))
    shortened_row_data = []
    for shortened_row in shortened_rows:
      parameters = shortened_row.url_data.get('parameters', [])
      is_created_in_email = shortened_row.url_data.get('is_created_in_email', False)
      shortened_row_data.append({
        'slug': shortened_row.slug,
        'parametersLength': len(parameters) > 0,
        'parameters': parameters,
        'is_created_in_email': is_created_in_email
      })

    return JsonResponse(shortened_row_data, safe=False)

  for alias in aliases:
    dashboard_id = alias.url_data.get('dashboardId')
    visual_type = alias.url_data.get('type')
    creator_name = alias.creator.username if alias.creator else 'User has been deleted'
    report_id = alias.report_id
    app_id = alias.app_id

    if visual_type == 'dashboard' and report_id is not None:
      dashboard_name = alias.report.report_name
      alias_count = ShortUrl.objects.filter(report_id=report_id).count()
    elif visual_type == 'dashboard' and report_id is None:
      dashboard_name = 'Dashboard has been deleted'
      alias_count = 0
    elif visual_type == 'apps' and app_id is not None:
      dashboard_name = alias.app.group_name
      alias_count = ShortUrl.objects.filter(app_id=app_id).count()
    elif visual_type == 'apps' and app_id is None:
      dashboard_name = 'App has been deleted'
      alias_count = 0
    else:
      continue  # skip aliases with unknown visual type

    parameters = alias.url_data.get('parameters', {})
    is_created_in_email = alias.url_data.get('is_created_in_email', {})
    aliases_data.append({
      'slug': alias.slug,
      'parameters': parameters,
      'times_followed': alias.times_followed,
      'last_viewed': alias.last_viewed,
      'dashboardId': dashboard_id,
      'alias_count': alias_count,
      'create_time': alias.create_time,
      'update_time': alias.update_time,
      'creator_name': creator_name,
      'visual_type': visual_type,
      'dashboard_name': dashboard_name,
      'is_created_in_email': is_created_in_email,
    })

  return JsonResponse({
    "aliases": aliases_data })
  

def _delete_short_url(request):
  slug = request.GET.get('slug')
  short_url = ShortUrl.objects.filter(slug=slug).first()
  
  if request.user.is_superuser or short_url.creator_id == request.user.id:
    short_url.delete()
    return HttpResponse('success', content_type="text/plain")
  elif not short_url:
    return HttpResponse('The specified alias does not exist', status=404)
  else:
    return HttpResponse("Only admins or the alias creator can delete this alias", status=401)

def _update_short_url(request):
  update_type = request.GET.get('updateType')

  if update_type == 'managing':
    origin_alias = request.GET.get('slug')
    new_alias = request.GET.get('newAlias')
    data = request.GET.get('url_data')

    shorturl_row_to_update = get_object_or_404(ShortUrl, slug=origin_alias)

    if new_alias:
      shorturl_row_to_update.slug = new_alias

    if data:
      url_data = json.loads(data)
      report_type = url_data['type']

      shorturl_row_to_update.url_data = url_data

      if report_type == 'dashboard':
        shorturl_row_to_update.report_id = url_data['dashboardId']
        shorturl_row_to_update.app_id = None
      else:
        shorturl_row_to_update.app_id = url_data['dashboardId']
        shorturl_row_to_update.report_id = None

    shorturl_row_to_update.save()

    return HttpResponse("success", content_type="text/plain")

  elif update_type == 'email':
    alias = request.GET.get('slug')
    shorturl_row_to_update = get_object_or_404(ShortUrl, slug=alias)
    shorturl_row_to_update.url_data['is_created_in_email'] = True
    shorturl_row_to_update.save()
        
    return HttpResponse("success", content_type="text/plain")

  elif update_type == 'alias':
    current_alias = request.GET.get('currentAlias')
    new_alias = request.GET.get('newAlias')

    shorturl_row_to_update = get_object_or_404(ShortUrl, slug=current_alias)
    shorturl_row_to_update.slug = new_alias
    shorturl_row_to_update.save(update_fields=['slug'])

    return HttpResponse("success", content_type="text/plain")

@login_required
def redirect_url_view(request, slug):
  try:
    shortener = ShortUrl.objects.get(slug=slug)
    ShortUrl.update_usage_count_and_last_viewed(id=shortener.id)

    url_data = shortener.url_data
    id = url_data['dashboardId']
    type = url_data['type']

    request.GET._mutable = True
    request.GET.update(shortener.url_data['parameters'])

    if type == 'dashboard':
      return app(request, report_id=id)
    else:
      return appgroup(request, appgroup_id=id)

  except Exception as e:
    return HttpResponse('Invalid URL Alias', status=500)