# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import io
import json
import logging
from util.decorator_custom import login_required_401
import reports
import os
import tempfile
import threading
import urllib.parse
import uuid
import vizlib.workq
import vizlib.handles
import glob
import shutil
import re
import xlsxwriter
import csv

from django.views.decorators.cache import cache_page, cache_control
from django.contrib.auth.decorators import login_required
from django.utils.cache import patch_cache_control
from django.http import HttpResponse, HttpResponseRedirect, FileResponse, StreamingHttpResponse
from vizlib.querylog import logAction, ComponentType, QueryState, ActionType
from django.conf import settings
from django.shortcuts import redirect
from importlib import import_module
from django.contrib import auth
from dateutil import parser

import asyncio
from reports.models import AppGroup, AppGroupAppMembership
from util.perfcounters import APPROX_COUNTERS
from util.request import create_request
from urllib.parse import urlencode
from html import unescape
from asyncio import coroutine
from vizlib.arcprof import ArcProf
from vizlib import misc
from pyppeteer.errors import PageError
from sqlrun.views import jsonselect_work, _csvRequestToParams, _getDelimiter
from dataconn.dataconn import get_dataconn
from dataconn.dataconn_dbapi2 import DbApi2Conn


SessionStore = import_module(settings.SESSION_ENGINE).SessionStore
LOGGER = logging.getLogger("arcweb")

# will be assigned values in init():
CHROMIUM_PDF_AVAILABLE = False
CHROMIUM_PATH = None
pyppeteer, PdfReader, PdfWriter, PageMerge, IndirectPdfDict = (None,) * 5

PDF_OPTIONS = getattr(settings, 'PDF_OPTIONS', {})

snapshot_queue = APPROX_COUNTERS.get('snapshot.queue')
snapshot_workstarted = APPROX_COUNTERS.get('snapshot.workstarted')
snapshot_workcompleted = APPROX_COUNTERS.get('snapshot.workcompleted')

def init():
  global CHROMIUM_PDF_AVAILABLE
  global CHROMIUM_PATH
  global pyppeteer, PdfReader, PdfWriter, PageMerge, IndirectPdfDict

  if PDF_OPTIONS.get('ENABLE_CHROMIUM_PDF', False):
    try:

      # Patch for pyppeteer issue #175
      def patch_pyppeteer():
        import pyppeteer.connection
        original_method = pyppeteer.connection.websockets.connect

        def new_method(*args, **kwargs):
            kwargs['ping_interval'] = None
            kwargs['ping_timeout'] = None
            return original_method(*args, **kwargs)

        pyppeteer.connection.websockets.connect = new_method

      patch_pyppeteer()

      import pyppeteer
      pyppeteer.DEBUG = True  # print suppressed errors as error log

      from pdfrw import PdfReader, PdfWriter, PageMerge, IndirectPdfDict

      CHROMIUM_PDF_AVAILABLE = True
      CHROMIUM_PATH = PDF_OPTIONS.get('CHROMIUM_PATH', None)
      if CHROMIUM_PATH is None:
        CHROMIUM_PDF_AVAILABLE = False
        LOGGER.debug('Chromium Path: %s not accessible' % str(CHROMIUM_PATH))
      elif CHROMIUM_PATH == "*":
        CHROMIUM_PDF_AVAILABLE = True
        LOGGER.debug('Chromium Path: %s available' % str(CHROMIUM_PATH))
      else:
        CHROMIUM_PDF_AVAILABLE = os.path.isfile(CHROMIUM_PATH) and os.access(CHROMIUM_PATH, os.X_OK)
        LOGGER.debug('Chromium Path (%s) accessible? %s' % (CHROMIUM_PATH, CHROMIUM_PDF_AVAILABLE))
    except Exception as e:
      LOGGER.debug('Chromium Exception')
      LOGGER.debug(str(e))
      CHROMIUM_PDF_AVAILABLE = False

  if not CHROMIUM_PDF_AVAILABLE:
    LOGGER.info('Chromium PDF not available')

init()

CHROMIUM_HEADLESS = True
GENERATE_PDF = True

if settings.DEBUG_SNAPSHOT_HEADED:
  CHROMIUM_HEADLESS = False
  GENERATE_PDF = False

try:
  from PIL import Image
  _PIL_ENABLED = True
except Exception as e:
  _PIL_ENABLED = False
  LOGGER.warning('PIL not available. Thumbnails not enabled')

_SNAPSHOT_RUNNER:vizlib.workq.AsyncRunner = None
_ADHOC_SNAPSHOT_RUNNER:vizlib.workq.AsyncRunner = None
_SNAPSHOT_RUNNER_LOCK = threading.Lock()

def get_thumbnail_bytes(image_path: str):
    image = Image.open(image_path)
    width = image.width
    image = image.crop(box = (0, 0, width, (width * 144 / 180)))
    image.thumbnail((180*2, 144*2), Image.LANCZOS) #doubling image size for retina display

    thumbnailString = io.BytesIO()
    image.save(thumbnailString, "png", optimize=True)
    thumbnail = thumbnailString.getvalue()
    thumbnailString.close()
    return thumbnail

def _init_runner():
  # In Ambari installations, we run cherrypy with daemonize=1. That leads to
  # importing this module pre-daemonizing, and then forking without threads.
  # This in turn means we cannot initialize _SNAPSHOT_RUNNER at import time
  # and instead have to do it at a later place, protected by _SNAPSHOT_RUNNER_LOCK

  global _SNAPSHOT_RUNNER
  global _ADHOC_SNAPSHOT_RUNNER
  global _CSV_XLSX_RUNNER

  # This is the concurrency of threads that will be forking chromium and
  # fetching apps for thumbnails. Set to 1 for now
  if _SNAPSHOT_RUNNER is None:
    with _SNAPSHOT_RUNNER_LOCK:
      if _SNAPSHOT_RUNNER is None:
        threadcount = getattr(settings, 'SNAPSHOT_WORKERS', 1)
        _SNAPSHOT_RUNNER = vizlib.workq.AsyncRunner(threadcount=threadcount, name="snapshot")

        threadcount = getattr(settings, 'ADHOC_SNAPSHOT_WORKERS', 1)
        _ADHOC_SNAPSHOT_RUNNER = vizlib.workq.AsyncRunner(threadcount=threadcount, name="adhoc.snapshot")

        threadcount = getattr(settings, 'CSV_XLSX_WORKERS', 10)
        _CSV_XLSX_RUNNER = vizlib.workq.AsyncRunner(threadcount=threadcount, name="csv_xlsx")


def _enabled():
  if not _PIL_ENABLED:
    return False

  enabled = reports.views_settings.setting_get_single("site_thumbnails_enable")
  if enabled is None:
    return False

  if enabled != "true":
    return False

  return True

def _saferemove_snapshotdir(out_dir):
  # verify this looks like a snapshot directory, and then remove it
  LOGGER.info("Removing %s", out_dir)
  if os.path.isdir(out_dir):
    try:
      shutil.rmtree(out_dir, ignore_errors=True)
    except Exception as e:
      LOGGER.warning("cannot remove temporary folder %s, %s" % (out_dir, str(e)))

def chromium_pdf_available():
  return CHROMIUM_PDF_AVAILABLE

def takeSnapshot(request, app_id=None):
  if not app_id:
    raise Exception('Invalid parameters')

  if not _enabled():
    return HttpResponse('ok')

  try:
    app = reports.models.Report.objects.get(id=int(app_id))
  except Exception as e:
    raise Exception('Invalid app_id %s' % app_id)

  app.dataset.checkPerm(request.user, 'ds_appedit')

  return _snapshot_take(request, app)

# For systems migrating from non having thumbnails to having thumbnails,
# permit a way for an admin user to take a thumbnail of every app in the
# system
@login_required_401
def groupsnapshot(request):
  if not _enabled():
    return HttpResponse('ok')

  if not request.user.is_superuser:
    return HttpResponse('ok')

  idsParam = request.GET.get('ids')
  ids = set()
  if idsParam:
    ids = set(map(int, idsParam.split(',')))

  config = {}
  saveType = request.GET.get('savetype', 'thumbnail').lower()
  config['saveType'] = saveType

  if len(ids) > 0:
    apps = reports.models.Report.objects.filter(id__in=ids)
    for app in apps:
      _snapshot_take(request, app, config)
    return HttpResponse('ok')


  takeall = request.GET.get('all', 'false').lower()
  apps = reports.models.Report.objects.all()
  for app in apps:
    if takeall == "true":
      _snapshot_take(request, app, config)
      continue

    ril = reports.models.ReportImage.objects.filter(report=app)
    if len(ril) == 0:
      _snapshot_take(request, app, config)

  return HttpResponse('ok')

# This is a call from the client asking us if the snapshot he requested is ready
# Resolve the snapshot id to a .progress filename using the client's session data and
# check for a non-zero file
@login_required_401
def snapshotReady(request, snapshotId=None):
  if not snapshotId:
    raise Exception('Invalid parameters')

  snapshotData = request.session.get('snapshotData', {})
  request_snapshot_data = snapshotData.get(snapshotId)
  out_dir = request_snapshot_data['outDir']

  if not out_dir or not os.path.isdir(out_dir):
    # this scenario should not have happened
    LOGGER.exception('Output path associated with snapshot ID %s is not valid, '
                     'output path: %s' % (snapshotId, out_dir))
    raise Exception('Output directory of the snapshot is not valid')

  errorFilename = os.path.join(out_dir, 'snapshot.error')
  if os.path.exists(errorFilename):
    with open(errorFilename) as f:
      errorStr = f.read()


    return HttpResponse(errorStr, status=500)

  progressFilename = os.path.join(out_dir, 'snapshot.progress')
  if os.path.exists(progressFilename):
    # thread is still in progress as the file exists
    return HttpResponse('notready')

  if len(glob.glob(os.path.join(out_dir, '*'))) == 0:
    # if .progress file has not been created yet
    return HttpResponse('notready')

  return HttpResponse('ready')

# Convert a PNG to PDF
# Need to create a white background to handle the transparency (alpha chanel)
def _convert_to_pdf(infiles, outfile):
  imgs = []
  for f in infiles:
    i = Image.open(f)
    i.load()
    b = Image.new("RGB", i.size, (255,255,255))
    b.paste(i, mask=i.split()[3])
    imgs.append(b)
  imgs[0].save(outfile, "PDF", resolution=120, append_images=imgs[1:],
               save_all=True)

# Resolve the snapshot id to the output filename using the client's session data and
# fetch its contents
@login_required_401
def snapshotFetch(request, snapshotId=None):
  if not snapshotId:
    raise Exception('Invalid parameters')

  snapshotData = request.session.get('snapshotData', {})
  request_snapshot_data = snapshotData.get(snapshotId)
  out_dir = request_snapshot_data['outDir']
  imgformat = request_snapshot_data.get('uploadFormat')
  if imgformat == 'csv':
    outfile = os.path.join(out_dir, request_snapshot_data.get('csvfilename'))
  elif imgformat == 'xlsx':
    outfile = os.path.join(out_dir, request_snapshot_data.get('xlsxfilename'))
  else:
    outfile = os.path.join(out_dir, 'output.format')
  if not out_dir:
    raise Exception('Snapshot id %s not found' % snapshotId)

  if not os.path.isdir(out_dir):
    raise Exception('Snapshot associated with id %s not found' % snapshotId)
  if not os.path.exists(outfile):
    raise Exception('Error during fetching snapshot: file (%s) is not found' % outfile)

  f = open(outfile, 'rb')
  if imgformat == 'xlsx':
    filename = request_snapshot_data.get('xlsxfilename')
    response = FileResponse(
        f,
        content_type='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
      )
    response['Content-Disposition'] = f'attachment; filename={filename}'
  elif imgformat == 'csv':
    filename = request_snapshot_data.get('csvfilename')
    response = FileResponse(f, content_type='text/csv')
    response['Content-Disposition'] = f'attachment; filename={filename}'
  else:
    if imgformat in ('pdf-all-sheets', 'pdf-current', 'pdf-one-viz'):
      response = FileResponse(f, content_type='application/pdf')
      ending = 'pdf'
    else:
      response = FileResponse(f, content_type="image/png")
      ending = 'png'

    filename = re.sub('[\.\s]+', '_', request.GET.get('title', ''))
    response['Content-Disposition'] = 'attachment; filename="%s.%s"' % (filename, ending)

  return response

@login_required_401
def template_snapshot(request, app_id=None):
  if not app_id:
    raise Exception('Invalid parameters')
  if request.method == 'GET':
    ri = reports.models.ReportTemplate.objects.get(id=int(app_id))
    response = HttpResponse(bytes(ri.img), content_type="image/png")

    # max_age value in seconds below is 1 year
    patch_cache_control(response, private=True, max_age=31536000)
    return response

# Fetch the stored thumbnail for an appgroup
# This resolves to the thumbnail for the first dashboard in this appgroup
@login_required
@cache_page(3600)  # cache thumbnail for 1 hour to reduce database queries
def snapshot_ag(request, appgroup_id=None):
  if not appgroup_id:
    raise Exception('Invalid parameters')

  if request.method != 'GET':
    raise Exception('Invalid parameters')

  subgroups = AppGroup.objects \
    .filter(group_parent__id=appgroup_id).order_by('group_subgroup_order')
  subgroup_ids = [s.id for s in subgroups][0:1]
  apps = AppGroupAppMembership.objects \
    .filter(appgroup_group__id__in=subgroup_ids) \
    .order_by('appgroup_order')
  if len(apps) > 0:
    first_app = apps[0]
    first_app_id = first_app.appgroup_app_id
    return _snapshot_get(request, first_app_id)

  return HttpResponseRedirect(f"{settings.ARC_BASEURL}/static/img/icon_app.png?v=6")

def snapshot(request, app_id=None):
  # Permit unauthenticated fetches of thumbnails if
  # ALLOW_THUMBNAIL_GET_WITHOUT_AUTH is set
  if not request.user.is_authenticated:
    if settings.ALLOW_THUMBNAIL_GET_WITHOUT_AUTH and request.method == 'GET':
      pass
    else:
      return redirect('%s?next=%s' % (settings.LOGIN_URL, request.path))

  return snapshot_internal(request, app_id=app_id)

@cache_page(3600)  # cache thumbnail for 1 hour to reduce database queries
def snapshot_internal(request, app_id=None):
  if not app_id:
    raise Exception('Invalid parameters')

  try:
    app = reports.models.Report.objects.get(id=int(app_id))
  except Exception as e:
    raise Exception('Invalid app_id %s' % app_id)

  if request.method == 'GET':
    return _snapshot_get(request, app_id)

  if not _enabled():
    return HttpResponse('ok')

  if request.method == 'POST':
    return _snapshot_take(request, app, request.POST, is_adhoc=True)

  raise Exception('Invalid request method')

def _snapshot_get(request, app_id):
  app_id = int(app_id)

  if not settings.ALLOW_THUMBNAIL_GET_WITHOUT_AUTH:
    # Ensure the user has permissions on the report being requested
    d = reports.views.report_get_internal(request, app_id)
    if len(d) == 0:
      raise Exception('Invalid app_id %s' % app_id)

  try:
    app = reports.models.Report.objects.get(id=app_id)
  except Exception as e:
    raise Exception('Invalid app_id %s' % app_id)

  ril = reports.models.ReportImage.objects.filter(report=app)
  if len(ril) == 0:
    raise Exception("Thumbnail for app %s not found" % app_id)

  ri = ril[0]
  response = HttpResponse(bytes(ri.img), content_type="image/png")

  # max_age value in seconds below is 1 year
  patch_cache_control(response, private=True, max_age=31536000)
  response.remove_vary = True  # for thumbnails remove the Vary: Cookie header
  return response

def create_snapshotrunner(appId=None, app=None, config=None, request=None, user=None):
  if app is None:
    try:
      app = reports.models.Report.objects.get(id=int(appId))
    except Exception as e:
      LOGGER.error(e)
      raise Exception('Invalid app_id %s' % appId)

  outConfig = config or {}

  if not CHROMIUM_PDF_AVAILABLE:
    LOGGER.info("Cannot generate snapshots")
    return None

  if outConfig['uploadFormat'] in ['pdf-current', 'png', 'unknown']:
    runner = _SnapshotRunner_chromium(app=app, config=outConfig, user=user, request=request)
  else:
    runner = _Runner_csv_xlsx(app=app, config=outConfig, request=request)
  return runner

class _SnapshotRunner_common(vizlib.workq.ProgressRunner):
  # Used to log failed/success snapshots
  def _auditSnapshot(self, query_state, error_detail=''):
    is_screenshot = hasattr(self, 'save_screenshot')
    is_csv_or_xlsx = ('uploadFormat' in self._config) and self._config['uploadFormat'] in ['csv', 'xlsx']
    if (is_screenshot and self.save_screenshot) or is_csv_or_xlsx:
      component_type = ComponentType.SNAPSHOT
    else:
      component_type = ComponentType.THUMBNAIL

    logAction(component_type=component_type,
        component_id=self._app.id, request=self._request,
        error_detail=error_detail, query_state=query_state,
        action=ActionType.CREATE)

  # Override
  def progress(self):
    """Returns a tuple that indicates (steps_completed, total_steps)"""
    return (self._completed, 1)

  # Override
  def results(self):
    """Returns results of the function"""
    return self._results

  def get_app_width_height(self):
    db_width = -1
    db_height = -1
    if self._config.get('saveType') == 'screenshot':
      db_width = int(self._config.get('dbWidth', -1))
      db_height = int(self._config.get('dbHeight', -1))

    if db_width < 0 or db_height < 0:
      appwidth = None
      try:
        app = reports.models.Report.objects.get(id=int(self._app.id))
        appwidth = json.loads(app.report_data).get('user_settings', {}).get('dashboard_width', None)
      except Exception as e:
        LOGGER.exception(str(e))

      SNAPSHOT_DEFAULT_WIDTH = getattr(settings, 'SNAPSHOT_DEFAULT_WIDTH', 1600)
      SNAPSHOT_DEFAULT_HEIGHT = getattr(settings, 'SNAPSHOT_DEFAULT_HEIGHT', 1280)
      if appwidth:
        db_width = int(appwidth) + 10
        db_height = int(db_width * SNAPSHOT_DEFAULT_HEIGHT / SNAPSHOT_DEFAULT_WIDTH) + 50
      else:
        db_width = SNAPSHOT_DEFAULT_WIDTH
        db_height = SNAPSHOT_DEFAULT_HEIGHT

    return [db_width, db_height]

class _SnapshotRunner_chromium(_SnapshotRunner_common):
  def __init__(self,
      request=None,
      user=None,
      appId=None,
      app=None,
      config=None):

    if app is None:
      try:
        self._app = reports.models.Report.objects.get(id=int(appId))
      except Exception as e:
        LOGGER.error(e)
        raise Exception('Invalid app_id %s' % appId)
    else:
      self._app = app

    if request:
      self._request = request
      self._user = request.user
    if user:
      # the request is only needed for the audit log
      # mock a different one
      client_ip = misc.get_request_source_ip(request) if request else ''
      self._request = create_request(user, client_ip)
      self._user = user

    self._profile = ArcProf('SnapshotRunnerChromium', username=self._user.username)
    self._config = config or {}
    self._completed = 0
    self._results = ""
    self.save_screenshot = config.get('saveType') == 'screenshot'
    self._user_settings = self._app.get_data().get('user_settings', {})

    self._querystring = config.get('paramString', '')

    dpi = PDF_OPTIONS.get('CHROMIUM_PDF_DPI', 100)
    page_formats = pyppeteer.page.Page.PaperFormats
    page_type = self._user_settings.get('print_pdf_size', 'letter')
    page_orientation = self._user_settings.get('print_pdf_orientation', 'portrait')
    page_scale = int(self._user_settings.get('print_pdf_scale', '0'))
    page_format = page_type.capitalize()
    enable_header_footer = self._user_settings.get('print_pdf_enable_header_footer', '') == 'true'

    if page_orientation == 'landscape':
      page_width = page_formats[page_type.lower()]['height'] * dpi
    else:
      page_width = page_formats[page_type.lower()]['width'] * dpi

    app_width_height = self.get_app_width_height()

    if page_scale == 0:
      scale = page_width / app_width_height[0]
    else:
      scale = page_scale / 100.0

    scale = max(scale, 0.1)
    scale = min(scale, 2.0)

    self._pdf_options = {
      'format': page_format,
      'landscape': page_orientation == 'landscape',
      'scale': scale,
      'displayHeaderFooter': enable_header_footer,
    }

    if enable_header_footer:
      self._pdf_options['headerTemplate'] = PDF_OPTIONS.get('HEADER_TEMPLATE', '')
      self._pdf_options['footerTemplate'] = PDF_OPTIONS.get('FOOTER_TEMPLATE', '')

    self._pdf_options['printBackground'] = True
    self.browser = None

  # Override
  def profile(self):
    return self._profile

  # Override
  def run(self):
    global snapshot_workstarted
    global snapshot_workcompleted

    # Ensure we never throw an exception from run, we don't want to kill the thread
    # that is running these
    try:
      snapshot_workstarted += 1
      self._run_internal()
    except Exception as e:
      LOGGER.exception(e)
    finally:
      snapshot_workcompleted += 1

  async def create_browser(self, tmp_dir):
    launch_options = {
      'args': [
                '--no-sandbox',
                '--disable-setuid-sandbox',
                '--disable-gpu',
                '--hide-scrollbars',
                '--disable-web-security',
                '--disable-dev-shm-usage',
                '--ignore-certificate-errors',
                '--user-data-dir=%s' % tmp_dir,
                ],
      'headless': CHROMIUM_HEADLESS, # Note if headless=False can't save pdf
      'handleSIGINT': False,
      'handleSIGTERM': False,
      'handleSIGHUP': False,
      'slowMo': 1,
      'logLevel': PDF_OPTIONS.get('LOGLEVEL', logging.INFO)
    }

    if CHROMIUM_PATH != "*":
      launch_options['executablePath'] = CHROMIUM_PATH

    # Launch the Browser
    LOGGER.info("Launching chromium")
    browser = await pyppeteer.launch(launch_options)
    return browser

  async def perform_snapshotting(self, procedure):
    await procedure()

  def _run_internal(self):
    """Start running the task"""
    logfile = None
    asyncio.set_event_loop(asyncio.new_event_loop())

    tmp_dir = self._config['outDir']
    errorFilename = os.path.join(tmp_dir, 'snapshot.error')
    progressFilename = os.path.join(tmp_dir, 'snapshot.progress')
    LOGGER.info("tmp_dir=%s", tmp_dir)

    if not CHROMIUM_PDF_AVAILABLE:
      LOGGER.error('Chromium PDF not available.  Can not generate PDF.')
      return

    LOGGER.debug("Starting PDF generation")

    try:
      # Create a session with a 10-minute timeout for this user. This is used
      # by the chromium invocation below to call back into this django server
      session = SessionStore()
      session.set_expiry(settings.SNAPSHOT_TIMEOUT)
      session[auth.BACKEND_SESSION_KEY] = \
          'django.contrib.auth.backends.ModelBackend'
      session[auth.SESSION_KEY] = self._user.id
      session[auth.HASH_SESSION_KEY] = self._user.get_session_auth_hash()
      session['CSRF_COOKIE'] = self._user.get_session_auth_hash()
      session.create()

      ip = reports.views_settings.setting_get_single("site_thumbnails_ip")
      if not ip:
        LOGGER.error('Thumbnails IP not available. Thumbnails not enabled')
        return

      domain = urllib.parse.urlparse(ip).hostname
      if domain is None:
        LOGGER.error('Cannot fetch domain from %s. Thumbnails not enabled' % ip)
        return

      app_width_height = self.get_app_width_height()

      self.logfilename = os.path.join(tmp_dir, 'browser.log')
      logfile = open(self.logfilename, 'w')

      # We implement a .progress file here that indicates to snapshotReady
      # in views_snapshot.py that work is still ongoing
      with open(progressFilename, 'w') as inprogress:
        pass

      gflags = {
        'gflag.is_snapshot': 1,
      }

      limit = self._config.get('csv_download_rows')
      if limit:
        gflags['gflag.limit'] = int(limit)

      url_params = dict(urllib.parse.parse_qsl(self._querystring))
      params = urlencode({**url_params, **gflags})
      app_id = str(self._app.id)
      url = f'{ip}/arc/apps/app/{app_id}?{params}'
      image_files = []

      @coroutine
      async def createCSVorXLSXfile(page, filetype):
        file_value = (await page.evaluate('arc.misc.createCSVorXLSXfile', self._config.get('csv_download_rows', '5000'), self._config.get('include_xlsx'), filetype))
        data_loaded = file_value['returnData'];
        name_to_file = {}
        for current_name, current_value in list(data_loaded.items()):
          current_file = tempfile.NamedTemporaryFile(suffix='.csv',
                                                    dir=tmp_dir,
                                                    delete=False)
          current_file.seek(0);

          with open(current_file.name, 'w') as open_file:
            open_file.write(''.join(current_value))

          name_to_file[current_name] = current_file.name
          current_file.close()

        if filetype == 'csv':
          self._config['csv_names_to_filepath'] = name_to_file

          # yield name_to_file

        if filetype == 'xlsx':
          file_name = os.path.join(tmp_dir, '0.xlsx')
          LOGGER.debug('create xlsx for sheet %s' % file_name)
          workbook = xlsxwriter.Workbook(file_name, {'nan_inf_to_errors': True})
          for name, xlsx_file in name_to_file.items():
            with open(xlsx_file, 'rb') as csv_file:
              delimiter = file_value['delimiters'].get(name, ',')
              reader = csv.reader((str(x.replace(b'\x00', b''), 'UTF-8') for x in csv_file), delimiter=delimiter, quotechar='"')
              sheet_name = re.sub('[\[\]:\*\?/\\\]', '_', name)
              if len(sheet_name) > 31:
                sheet_name = sheet_name[0:31]
              current_sheet = workbook.add_worksheet(sheet_name)
              for r_idx, row in enumerate(reader):
                current_sheet.write_row(r_idx, 0, row)

          workbook.close()

          self._config['xlsx_name_to_filepath'] = file_name

          # yield file_name

      # This is an async function that runs Chromium and generates pdfs
      async def chromesnapshot():
        self.browser = await self.create_browser(tmp_dir=tmp_dir)
        await self.perform_snapshotting(take_snapshot)

      async def take_snapshot():
        browser = self.browser
        CHROMIUM_SNAPSHOT_TIMEOUT = settings.SNAPSHOT_TIMEOUT*1000

        # Create a new Page
        page = await browser.newPage()
        page.setDefaultNavigationTimeout(CHROMIUM_SNAPSHOT_TIMEOUT)

        # Capture console.log messages and send to debug log and arcviz log
        @page.on('console')
        def log_console(msg):
          if logfile:
            logfile.write("%s: %s\n" % (msg.type, msg.text))
            logfile.flush()
          logConsoleMessage(msg)

        def logConsoleMessage(msg):
          level = logging.ERROR if msg.type == 'error' else logging.DEBUG

          def get_info(arg):
            info = ''
            if 'description' in arg._remoteObject:
              info = arg._remoteObject['description']
            elif 'value' in arg._remoteObject:
              info = arg._remoteObject['value']
            return info

          info = [get_info(arg) for arg in msg.args]
          msg_text = '\n'.join(info).strip()

          if not msg_text:
            # no useful info could be extract from the arguments
            # use the message text:
            msg_text = msg.text

          LOGGER.log(level, 'Javascript console log while generating PDF:\n%s: %s' % (msg.type, msg_text))

        # Capture error messages and send to debug log and arcviz log
        @page.on('pageerror')
        def log_pageerror(err):
          msg = 'Page Error while generating PDF:\n'
          msg += err.args[0] + '\n'
          if logfile:
            logfile.write(msg)
            logfile.flush()
          LOGGER.error(msg)

        # Set session cookie
        await page.setCookie({"name": settings.SESSION_COOKIE_NAME, "value": session.session_key, "domain": domain, "path": "/"})
        await page.setCookie({"name": settings.CSRF_COOKIE_NAME, "value": session.get('CSRF_COOKIE', None), "domain": domain, "path": "/"})

        # Set the Viewport to be a bit larger than the Dashboard
        await page.setViewport({"width": int(app_width_height[0]) + 200, "height": int(app_width_height[1]) + 200})

        LOGGER.info("PDF/PNG/CSV Generation URL: %s" % url)
        # Navigate to the correct URL
        try:
          await page.goto(url, {'timeout': 60000, 'waitUntil': "networkidle0"})
        except PageError:
          err = 'Wrong URL provided for snapshot'

          VIZ_PORT = os.environ.get('VIZ_PORT', '38888')
          default_ip = f'http://127.0.0.1:{VIZ_PORT}'

          if (ip != default_ip):
            err += f', try switching back to the default: {default_ip}'

          raise Exception(err)

        # Turn off animations
        await page.addStyleTag({'content': '''* { -webkit-transition: none !important; transition: none !important; -webkit-animation: none !important; animation: none !important; }'''})

        # Wait for all the visuals to load
        LOGGER.debug("Waiting for visuals to load")

        await page.waitForFunction('arc.misc.allVisualsLoaded()', {'polling': 200, 'timeout': CHROMIUM_SNAPSHOT_TIMEOUT })

        # Determine number of sheets and which sheets to use for pdf
        num_of_sheets = await page.evaluate('''() => $(".sheet-handle").length''')
        if self._app.report_type == 'dashboard' and \
            self._config['uploadFormat'] == 'pdf-all-sheets':
          sheets_to_print = range(num_of_sheets)
        else:
          sheets_to_print = [int(self._config.get('sheet_id', 1))-1]

        # Iterate over each sheet
        for s in sheets_to_print:
          LOGGER.debug("create pdf for sheet: %s" % s)

          # Reset counter values used in allVisualsLoaded
          # if loading any sheet other than the first
          if len(sheets_to_print) > 1:
            LOGGER.debug("Running arc.misc.resetVisualLoadedCounters and changing sheets")
            await page.evaluate('() => { arc.misc.resetVisualLoadedCounters() }')
            await page.evaluate('() => { let s = $(".sheet-handle")[%s]; s && s.click() }' % s)

            await asyncio.sleep(1)
            LOGGER.info("Waiting for sheet visuals to load")
            await page.waitForFunction('arc.misc.allVisualsLoaded()', {'polling': 200, 'timeout': CHROMIUM_SNAPSHOT_TIMEOUT})

          # Convert mapboxgl maps to PNGs to allow PDF generation
          # Note: This does not work with the RH7 'headless_shell'
          # must use a different Chromium
          maps = await page.JJ('div.mapboxgl-map')
          if maps:
            LOGGER.info("Waiting for maps to render")
            await asyncio.sleep(15)  # for now pause if we have maps to allow them to fully render

          for map in maps:
            LOGGER.info("Converting map to png")
            map_png_data = await map.screenshot({'encoding': 'base64'})
            if isinstance(map_png_data, bytes):
              map_png_data = map_png_data.decode('ascii')
            map_img = '<img src="data:image/png;base64,' + map_png_data + '" />'
            await page.evaluate('''(el,img) => $(el).replaceWith(img)''', map, map_img)

          # Generate PDF for sheet
          if GENERATE_PDF:
            LOGGER.info("Generating PDF")
            outFilename = os.path.join(tmp_dir, 'sheet_%s.pdf' % s)
            image_files.append(outFilename)
            self._pdf_options['path'] = outFilename
            await page.pdf(self._pdf_options)

          # Generate a PNG while we are at it
          #  used in email attachments
          # Note: fullPage=True cab create a huge png
          # Emulate print media for css styling - better for PNG screenshots
          await page.emulateMedia('print')

          LOGGER.info("Generating PNG")
          pngoutFilename = os.path.join(tmp_dir, '0.png')
          await page.screenshot({'path': pngoutFilename, 'fullPage': False})

          # Generate CSV
          LOGGER.info("Generating CSV")
          if self._config.get('include_csv'):
            await createCSVorXLSXfile(page, 'csv')

          # Generate XLSX (from CSV)
          LOGGER.info("Generating XLSX")
          if self._config.get('include_xlsx'):
            await createCSVorXLSXfile(page, 'xlsx')

      # Run Chromium in an async event loop and wait until it's done
      try:
        asyncio.get_event_loop().run_until_complete(asyncio.wait_for(chromesnapshot(),
            settings.SNAPSHOT_TIMEOUT))
      except asyncio.TimeoutError as t:
        # re-raise for a better error message
        raise Exception("Snapshot took too long to generate")

      # Combine the sheet pdfs if needed
      outfile = os.path.join(tmp_dir, 'output.format')
      if len(image_files) == 1:
        shutil.copyfile(image_files[0], outfile)
      else:
        writer = PdfWriter()
        for f in image_files:
          writer.addpages(PdfReader(f).pages)
        writer.write(outfile)
      os.link(outfile, outfile + ".pdf")

      # Copy PNG to output file
      if self._config.get('uploadFormat') in ['png', 'unknown']:
        shutil.copyfile(os.path.join(tmp_dir, "0.png"), outfile)

      # Thumbnail
      if not self.save_screenshot:
        # updating thumbnails when doing pdf or image download or saving artifacts

        image_path = os.path.join(tmp_dir, "0.png")
        thumbnail = get_thumbnail_bytes(image_path)

        ril = reports.models.ReportImage.objects.filter(report=self._app)
        if len(ril) == 0:
          ri = reports.models.ReportImage(report=self._app, img=thumbnail)
        else:
          ri = ril[0]
          ri.img = thumbnail

        ri.save()
        LOGGER.info('Thumbnail for app %s created' % self._app.id)

      else:
        os.remove(progressFilename)

    except Exception as e:
      LOGGER.exception(e)
      self._results = "Error: %s" % str(e)

      try:
        with open(errorFilename, 'w') as err:
          err.write(self._results)
      except:
        pass

      if progressFilename:
        try:
          os.remove(progressFilename)
        except:
          pass

    finally:
      if logfile:
        try:
          logfile.close()
        except:
          pass

      try:
        self._close_browser()
      except:
        LOGGER.exception('Error closing browser')

      if not self.save_screenshot:
        # this is the case of thumbnails
        # by this time they were either persisted to the DB
        # or the errors stored on self._results
        _saferemove_snapshotdir(tmp_dir)

      self.run_callback(self._results)

      self._completed = 1

  def _close_browser(self):
    LOGGER.info("Closing browser")
    browser = self.browser
    if browser:
      try:
        asyncio.get_event_loop().run_until_complete(
          asyncio.wait_for(self.browser.close(), 10)
        )
      except Exception as e:
        LOGGER.warn(e)
      finally:
        if browser.process and browser.process.pid:
          LOGGER.info(
            "Terminating browser process with pid %s", browser.process.pid
          )
          browser.process.terminate()
          browser.process.wait()
        self.browser = None

  def run_callback(self, error=None):
    if error:
      status = QueryState.CANCELED
      error += "\n"
      if os.path.exists(self.logfilename):
        with open(self.logfilename, 'r') as lg_file:
          log_contents = lg_file.read()
        error += log_contents
      self._config["error"] = error
    else:
      status = QueryState.COMPLETE
    self._auditSnapshot(status, error)

class _Runner_csv_xlsx(_SnapshotRunner_common):
  def __init__(self, app=None, config=None, request=None):
    self._user = request.user
    self._app = app

    self._request = request
    self._config = config or {}
    self._params = _csvRequestToParams(config)
    self._results = ""
    self._profile = ArcProf('SnapshotRunnerCSV_XLSX', username=self._user.username)

  # Override
  def profile(self):
    return self._profile

  # Override
  def run(self):
    global snapshot_workstarted
    global snapshot_workcompleted

    # Ensure we never throw an exception from run, we don't want to kill the thread
    # that is running these
    try:
      snapshot_workstarted += 1
      self._run_internal()
    except Exception as e:
      LOGGER.exception(e)
    finally:
      snapshot_workcompleted += 1

  def _run_internal(self):
    tmp_dir = self._config['outDir']
    errorFilename = os.path.join(tmp_dir, 'snapshot.error')
    progressFilename = os.path.join(tmp_dir, 'snapshot.progress')

    LOGGER.info("tmp_dir=%s", tmp_dir)

    # We implement a .progress file here that indicates to snapshotReady
    # in views_snapshot.py that work is still ongoing
    with open(progressFilename, 'w') as inprogress:
      pass

    try:
      if self._request.POST.get('uploadFormat') == 'csv':
        self._params['csvdownload'] = 'true'
      elif self._request.POST.get('uploadFormat') == 'xlsx':
        self._params['xlsxdownload'] = 'true'
      
      jsonselect_returned_value = jsonselect_work(
        self._request,
        self._params,
        ignore_data_from_session=True
      )

      if 'error' in jsonselect_returned_value:
        error = jsonselect_returned_value['error']
        raise Exception(f'{error}')

      if self._request.POST.get('uploadFormat') == 'csv':
        if 'response' in jsonselect_returned_value:
          self.create_csv_file(jsonselect_returned_value['response'])
      elif self._request.POST.get('uploadFormat') == 'xlsx':
        if 'response' in jsonselect_returned_value:
          self.create_xslx_file(jsonselect_returned_value['response'])

      os.remove(progressFilename)

    except Exception as e:
      LOGGER.exception(e)
      self._results = "Error: %s" % str(e)

      try:
        with open(errorFilename, 'w') as err:
          err.write(self._results)
      except Exception as write_exception:
        LOGGER.exception(write_exception)
        pass

      if progressFilename:
        try:
          os.remove(progressFilename)
        except Exception as file_reove_exception:
          LOGGER.exception(file_reove_exception)
          pass
    
    self.run_callback(self._results)
    self._completed = 1
  
  def run_callback(self, error=None):
    if error:
      status = QueryState.CANCELED
      error += "\n"
      self._config["error"] = error
    else:
      status = QueryState.COMPLETE
    self._auditSnapshot(status, error)

  # Override
  def get_app_width_height(self):
    pass


  def create_csv_file(self, input_data):
    output_dir = self._config['outDir']
    file_path = os.path.join(output_dir, self._params['csvfilename'])

    if isinstance(input_data, HttpResponse) and input_data["Content-Type"] == 'text/csv':
      with open(file_path, 'wb') as f:
        f.write(input_data.content)

    elif isinstance(input_data, StreamingHttpResponse):
      with open(file_path, 'wb') as csvfile:
        for chunk in input_data.streaming_content:
          csvfile.write(chunk)
    else:
      raise ValueError("Unsupported data type. Expected HttpResponse or StreamingHttpResponse.")

  def create_xslx_file(self, input_data):
    output_dir = self._config['outDir']
    file_path = os.path.join(output_dir, self._params['xlsxfilename'])

    with open(file_path, 'wb') as file:
      if isinstance(input_data, FileResponse):
        for chunk in input_data.streaming_content:
          file.write(chunk)
        return None

    raise ValueError("Unsupported input_data type. Excpected a FileResponse object.")

def get_snapshot_temp_dir_path(prefix: str) -> str:
  os_temp_dir_absolute_path = tempfile.gettempdir()
  snapshot_temp_dir_path = os.path.join(os_temp_dir_absolute_path, prefix)
  return snapshot_temp_dir_path

def _snapshot_take(request, app, config=None, is_adhoc=False):
  global snapshot_queue

  if not config:
    config = {}

  outConfig = {}
  returnData = {}
  outConfig['uploadFormat'] = 'unknown'
  params = urllib.parse.parse_qs(config.get('paramString', ''))
  
  snapshot_files_temp_dir_path = get_snapshot_temp_dir_path(settings.SNAPSHOT_TMP_DIR_PREFIX)
  os.makedirs(snapshot_files_temp_dir_path, exist_ok=True)
  
  outConfig['outDir'] = tempfile.mkdtemp(dir=snapshot_files_temp_dir_path, suffix="_snapshot")
  LOGGER.info("Created outDir %s", outConfig['outDir'])

  snapshot_uuid = str(uuid.uuid4().int)
  if config.get('saveType') == 'screenshot':
    outConfig['saveType'] = 'screenshot'
    outConfig['dbWidth'] = float(config.get('dbWidth', -1))
    outConfig['dbHeight'] = float(config.get('dbHeight', -1))
    outConfig['paramString'] = config.get('paramString', '')

    if request:
      outConfig['uploadFormat'] = config.get('downloadFormat')
      snapshotData = request.session.get('snapshotData', {})
      snapshotData[snapshot_uuid] = outConfig
      request.session['snapshotData'] = snapshotData

    if outConfig['uploadFormat'] in ['pdf-current', 'png']:
      outConfig['sheet_id'] = params.get('sheet', [1])[0]
    returnData['snapshotId'] = snapshot_uuid

  elif config.get('uploadFormat') in ['csv', 'xlsx']:
    outConfig = config.copy()
    outConfig['outDir'] = tempfile.mkdtemp(dir=snapshot_files_temp_dir_path, suffix="_snapshot")
    snapshotData = request.session.get('snapshotData', {})
    snapshotData[snapshot_uuid] = outConfig
    request.session['snapshotData'] = snapshotData

    returnData['snapshotId'] = snapshot_uuid


  runner = create_snapshotrunner(app=app, config=outConfig, request=request)
  snapshot_queue += 1

  if runner:
    _init_runner()
    if 'uploadFormat' in outConfig and outConfig['uploadFormat'] in ['csv', 'xlsx']:
      data_connection = get_dataconn(outConfig['dataconnection_id'])
      if isinstance(data_connection, DbApi2Conn):
        misc.initAsyncConnectionThreadpool(outConfig['dataconnection_id'])
        runner_csv_xlsx = misc.get_async_runner(outConfig['dataconnection_id'])
        runner_csv_xlsx.queue_untracked(runner)
      else:
        _CSV_XLSX_RUNNER.queue_untracked(runner)
    elif is_adhoc:
      _ADHOC_SNAPSHOT_RUNNER.queue_untracked(runner)
    else:
      _SNAPSHOT_RUNNER.queue_untracked(runner)

  return HttpResponse(json.dumps(returnData), content_type='application/json')

