# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Visuals page related requests
"""

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.db.models import Q, Count
from django.http import HttpResponse, JsonResponse, HttpRequest
from django.template import loader
from django.shortcuts import render

from apps.utils import get_should_change_password, _appflags
from reports.models import Report, AppGroup, AppGroupAppMembership
from reports.views_appgroups import appgroup_starred
from rbac.models import RBAC
from util.get_dataset_ids_accessible_by_user import get_dataset_ids_accessible_by_user, is_dataset_superuser
from util.get_workspace_ids_accessible_by_user import get_workspace_ids_accessible_by_user, is_workspace_superuser
from util.get_sample_dashboards_query import get_sample_dashboards_query
from workspace.models import get_or_create_private_ws, Workspace

import json

@login_required
def page(request: HttpRequest):
  """this is the 'Visuals' tab """
  should_change_password = get_should_change_password(request.user)
  if should_change_password:
    return render(request, 'change_password.html')

  context = _appflags(request)

  mobile = context['mobile']
  if mobile:
    template = loader.get_template('apps_mobile.html')
    return HttpResponse(template.render(context, request))

  mode = request.GET.get('mode', 'new')

  user_private_workspace = get_or_create_private_ws(request.user)
  static_side_bar_items = {
    'all': {
        'id': -1,
        'name': 'All',
        'desc': '',
    },
    'favorites': {
      'id': -2,
      'name': 'My Favorites',
      'desc': 'Mark your frequently used dashboards as favorites. These bookmarks make it easy to find your most-used content in a subsequent work session.'
    },
    'private': {
      'id': user_private_workspace.id,
      'name': 'Private',
      'desc': user_private_workspace.description,
      'editable': user_private_workspace.editable,
      'perms':  list(user_private_workspace.getPerms(request.user)),
      'acl' : user_private_workspace.acl_obj,
    }
  }

  has_samples = AppGroup.objects.filter(group_type='samples').exists()
  if has_samples:
    static_side_bar_items['samples'] = {
      'id': -4,
      'name': 'Samples',
      'desc': 'Get started by looking through our samples and see how they use a combination of visuals and filters to provide business insights in sample datasets.',
    }

  public_workspace = Workspace.objects.filter(name="Public", editable=False).first()
  if public_workspace and (request.user.is_superuser or 'sys_misc' in RBAC.get_user_obj_perms(request.user) or 'ws_appview' in public_workspace.getPerms(request.user)):
    static_side_bar_items['public'] = {
      'id': public_workspace.id,
      'name': public_workspace.name,
      'desc': public_workspace.description,
      'editable': public_workspace.editable,
      'perms':  list(public_workspace.getPerms(request.user)),
      'acl' : public_workspace.acl_obj,
    }

  context['static_side_bar_items'] = json.dumps(static_side_bar_items)

  if mode == 'old':
    myapps = request.GET.get("myapps")
    initialsort = request.GET.get("initialsort")
    group = request.GET.get('group', '')
    grouptype = request.GET.get('grouptype', '')
    displaymode = request.GET.get('displaymode', '')
    appids = request.GET.get('appids', '')
    if myapps:
      context['myapps'] = 1
    if initialsort:
      context['initialsort'] = initialsort
    context['group'] = group
    context['grouptype'] = grouptype
    context['displaymode'] = displaymode
    context['show_apps_sidebar'] = settings.SHOW_APPS_SIDEBAR
    context['enable_dataconn_arcengine'] = settings.ENABLE_DATACONN_ARCENGINE
    context['appids'] = appids

    template = loader.get_template('apps_groups.html')
  else:
    template = loader.get_template('apps_visuals_page.html')

  return HttpResponse(template.render(context, request))

@login_required
def get_report_count_per_group(request: HttpRequest):
    reports = Report.objects.filter(Q(report_type='dashboard') | \
                              Q(shared_visual_dashboards__isnull=False))
    app_groups = AppGroup.objects.filter(group_parent__id=None, group_type__in=['default', 'sample-app'])

    user = request.user

    user_workspace_ids = None
    user_dataset_ids = None

    if not is_workspace_superuser(user):
      user_workspace_ids = get_workspace_ids_accessible_by_user(user)
      # limit the report search to workspaces and datasets accessible to the user
      reports = reports.filter(workspace_id__in=user_workspace_ids)

      # limit the report search to workspaces and datasets accessible to the user
      app_groups = app_groups.filter(workspace_id__in=user_workspace_ids)

    if not is_dataset_superuser(user):
      user_dataset_ids = get_dataset_ids_accessible_by_user(user)
      reports = reports.filter(dataset_id__in=user_dataset_ids)

    reports = reports.values('workspace_id').annotate(report_count=Count('id'))
    workspace_id_to_report_count = { report['workspace_id']:report['report_count'] for report in reports }

    app_groups = app_groups.values('workspace_id').annotate(app_count=Count('id'))
    workspace_id_to_app_count = { appGroup['workspace_id']:appGroup['app_count'] for appGroup in app_groups }

    favorites_appgroup = appgroup_starred(request)

    favorite_dashboard_ids = list(
      AppGroupAppMembership.objects \
        .filter(appgroup_group__id=favorites_appgroup.id) \
        .order_by('appgroup_order') \
        .values_list('appgroup_app', flat=True)
    )

    favorites_report_count = Report.objects.filter(id__in=favorite_dashboard_ids, report_type='dashboard').count()

    samples_report_count = None

    try:
        samples_report_count = get_sample_dashboards_query(
          user_dataset_ids,
          user_workspace_ids
        ).count()

        samples_report_count += AppGroup.objects.filter(group_type="sample-app").count()
    except:
      samples_report_count: None

    return JsonResponse({
      "favorites_report_count": favorites_report_count,
      "workspace_id_to_report_count": workspace_id_to_report_count,
      "workspace_id_to_app_count": workspace_id_to_app_count,
      "samples_report_count": samples_report_count
    })
