# Cloudera Data Visualization - settings.py

This is the documentation for managing Django startup settings in Cloudera Data Visualization. 
 
Here you will find more detailed notes and guidance around certain settings and their usage.

## Using and Configuring External Databases as a Metastore for Cloudera Data Visualization

### Setting up your External Database for Cloudera Data Visualization 

The following external relational databases and versions are supported today in Cloudera Data Visualization:
- MySQL (version 5.7 only; 8.x not supported yet)
    - MariaDB (Using MySQL libraries - 10.2+)
- Postgres (version 9.6+)
- Oracle Database (version 12.2+)

For additional notes on testing and usage for each type of database, take a look at these pages:
- MySQL: https://docs.djangoproject.com/en/3.2/ref/databases/#mysql-notes
- Postgres: https://docs.djangoproject.com/en/3.2/ref/databases/#postgresql-notes
- Oracle:
    - https://code.djangoproject.com/wiki/OracleTestSetup
    - https://docs.djangoproject.com/en/3.2/ref/databases/#oracle-notes

If needed for local testing, you can pull these Docker images:
- MySQL database: https://hub.docker.com/_/mysql
    - NOTE: Make sure to use a 5.x tag when pulling.
- Postgres: https://hub.docker.com/_/postgres
- Oracle: https://hub.docker.com/_/oracle-database-enterprise-edition
    - NOTE: This will create an Oracle 12.x database.

The next sub-sections provide steps for creating a database and a database user/password that Cloudera Data Visualization 
can use when connecting. Other configurations such as transaction isolation levels are not covered here 
and should only be adjusted if there are performance issues or other issues.

#### MySQL/MariaDB

From MySQL shell:

```
CREATE DATABASE dataviz DEFAULT CHARACTER SET utf8;
CREATE USER 'viz'@'%' IDENTIFIED BY '<password>';
GRANT ALL ON viz.* TO 'viz'@'%';
```

#### Postgres

From command line prompt:

```
createuser viz --createdb --pwprompt
createddb dataviz -U viz
```

Also double check that the newly created database is using `client_encoding` 'UTF8'.

#### Oracle

From Oracle SQLPlus shell:

```
CREATE user dataviz identified by <password>;
GRANT create sequence to dataviz;
GRANT create session to dataviz;
GRANT create table to dataviz;
GRANT create view to dataviz;
GRANT create procedure to dataviz;
GRANT create trigger to dataviz;
GRANT execute on sys.dbms_crypto to dataviz;
GRANT execute on sys.dbms_lob to dataviz;
GRANT execute on sys.dbms_random to dataviz;
```

### Configuring your environment/application to use your External Database

### Configuration in CDSW / CML 

From your CDSW or CML Project Setting menu, add the following environment variables 
and corresponding values that can be used to connect to your external database successfully. 
The example below shows what a complete configuration might look like.

![](./cdswcml_external_db_settings_example.png)

After making these changes and saving, restart your Cloudera Data Visualization application for the new setting to take effect.

If the Application does not start-up properly, check the logs to see if there are any tracebacks or hints of where the error occurred. 
Typically a failure to connect will happen for 3 reasons:

- Your database connection configurations aren't correct. 
    - You should check these manually just in case there was a typo or mistake.
- Your application won't start because the initial Django migrations didn't run.
    - First check your database to see if any tables were created. There should be at least 42 tables.
    - If no tables have been created, you will need open a Session in your project to manually run migrations.
    
        ```
        # Navigate to .arc where the settings_local.py is located
        cd .arc
        
        # Set environment variables so we can properly run the database migrations
        export PYTHONPATH=/opt/vizapps/venv/lib64/python3.9/site-packages/arcweb

        export BASEDIR=/home/cdsw/.arc
        export VENV_PATH=/opt/vizapps/venv
        export DJANGO_SETTINGS_MODULE=settings_local
        export VIZ_PORT=8100
        export CDV_MODE=cdsw
        
        export DIR=/opt/vizapps/tools/arcviz
        export PYTHONPATH=${PYTHONPATH}:${BASEDIR}
        
        # Manually run database migrations in Data Viz to initialize the database structure
        ${DIR}/arcviz migrate
        ```
- Your database isn't reachable due to networking inbound/outbound firewall rules. 
    - Open a Session in your project, and run a `curl` command against your hostname and port to see if any successful response comes back.
        ```
        (app-root) sh-4.4$ curl -v 172.20.0.1:1521
        * Rebuilt URL to: 172.20.0.1:1521/
        *   Trying 172.20.0.1...
        * TCP_NODELAY set
        * Connected to 172.20.0.1 (172.20.0.1) port 1521 (#0)
        > GET / HTTP/1.1
        > Host: 172.20.0.1:1521
        > User-Agent: curl/7.61.1
        > Accept: */*
        > 
        * Empty reply from server
        * Connection #0 to host 172.20.0.1 left intact
        curl: (52) Empty reply from server
        (app-root) sh-4.4$ 
        ```

### Configuration for Local Testing

First, export these environment variables in your terminal or you can also set these in your ~/.bash_profile 
if that's easier to maintain long term:

```
export METADATA_STORE'mysql'
export METASTORE_DB_NAME='mysql'
export METASTORE_DB_USER='<your mysqluser>'
export METASTORE_DB_PASSWORD='<your mysqlpassword>'
export METASTORE_DB_HOST=<your hostname i.e. localhost>
export METASTORE_DB_PORT=3306
```

For Oracle database setups, there's usually a requirement to use an SID to connect rather than a port/host combination, 
which will make the environment variables looking a little different:

```
export METADATA_STORE='oracle'
export METASTORE_DB_NAME='ORCLPDB1.localdomain'
export METASTORE_DB_USER='<your oracleuser>'
export METASTORE_DB_PASSWORD='<your oraclepassword>'
```

If you can connect to Oracle without an SID, you would need to set the environment variables this way:

```
export METASTORE_DB_ORACLE_USE_SID=false
export METADATA_STORE='oracle'
export METASTORE_DB_NAME='<your oracledb>'
export METASTORE_DB_USER='<your oracleuser>'
export METASTORE_DB_PASSWORD='<your oraclepassword>'
export METASTORE_DB_HOST=<your hostname i.e. localhost>
export METASTORE_DB_PORT=1521
```

Then startup Data Viz like you would normally do with `scons runserver` or `scons runcpserver`.

If there's a failure during start-up, check the stdout to see if there are any tracebacks or hints of where the error occurred,
and also refer to the debugging steps mentioned previously for different ideas of how to proceed.

### Configuration for Docker Testing

```More notes coming soon.```