# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.http import HttpRequest
from vizlib.misc import get_request_source_ip

import functools
import logging
import threading

LOGGER = logging.getLogger('arcweb.access_info')
LOGGER.setLevel(logging.INFO)


class AccessInfo():
  def __init__(self, username='', user_id='', client_ip=''):
    self.username = username
    self.user_id = user_id
    self.client_ip = client_ip


EMPTY = AccessInfo()
_STORAGE = threading.local()


def create_from_request(request: HttpRequest):
  user = None
  if hasattr(request, 'user'):
    user = request.user

  username = user.username if user else ''
  user_id = user.id if user else ''
  client_ip = get_request_source_ip(request)

  return AccessInfo(username, user_id, client_ip)


def store(access_info: AccessInfo):
  '''
  Stores information useful for auditing for the invoking thread.
  '''
  _STORAGE.access_info = access_info


def get() -> AccessInfo:
  ai = getattr(_STORAGE, 'access_info', None)
  if ai is None:
    LOGGER.debug('access_info unavailable')
    return EMPTY
  return ai


def clear():
  '''
  Clears previously stored information for the invoking thread
  '''
  _STORAGE.access_info = None

def store_access_info(func):
  '''
  Decorator useful for passing the current access info to a function
  that is going to be executed in another thread.
  '''

  # retrieve the access info of the current thread
  ai = get()

  @functools.wraps(func)
  def access_info_enabled_func(*args, **kwargs):
    try:
      store(ai)
      return func(*args, **kwargs)
    finally:
      clear()

  return access_info_enabled_func