# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from django.http import JsonResponse, HttpResponse
from reports.views_settings import setting_get_single
from django.conf import settings
from arcwebbase.ai_service_client import send_ai_request

from django.contrib.auth.decorators import login_required
import json
import asyncio

@login_required
def validate_ai_service_connection(request):
    if request.method == 'POST':
        try:
            body = json.loads(request.body)
            model_type = body.get('tab')
            profile_id = body.get('id')
            settings_to_test = body.get('settings')

            keys_to_check = [f"{model_type}_api_key", f"{model_type}_auth_key", f"{model_type}_bedrock_credentials"]
            placeholder = settings.DEFAULT_PASSWORD_PLACEHOLDER

            if profile_id:
                keys_with_placeholder = [
                    key for key in keys_to_check
                    if settings_to_test.get(key) == placeholder
                ]

                if keys_with_placeholder:
                    saved_settings_json = setting_get_single(f"ai_{model_type}_settings")
                    saved_settings = json.loads(saved_settings_json)
                    saved_values = saved_settings['profileSettings'][profile_id]['values']

                    for key in keys_with_placeholder:
                        if key in saved_values:
                            settings_to_test[key] = saved_values[key]

            prof_data = {
                'settings': settings_to_test,
                'input': get_test_input(model_type),
                'model_type': model_type
            }

            try:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                response = asyncio.get_event_loop().run_until_complete(get_first_ai_response(prof_data))
            except asyncio.TimeoutError:
                raise Exception("The AI request timed out while waiting for a response.")
            finally:
                loop.run_until_complete(loop.shutdown_asyncgens())
                loop.run_until_complete(asyncio.sleep(0))
                loop.close()

            status_code = response['status_code']
            response_body = response['response_body']
            req_info = response.get('req_info')

            if status_code == 200:
                return JsonResponse({'success': True, 'error': None}, status=status_code)
            else:
                resp = {'success': False, 'error': response_body}
                if req_info:
                    resp['clear_text_request'] = req_info
                return JsonResponse(resp, status=status_code)
        except json.JSONDecodeError:
            return JsonResponse({'success': False, 'error': 'Invalid JSON payload'}, status=400)
        except ValueError as e:
            return JsonResponse({'success': False, 'error': str(e)}, status=400)
        except Exception as e:
            return JsonResponse({'success': False, 'error': f'Unexpected error: {str(e)}'}, status=500)
    else:
        return HttpResponse('Invalid request method', status=405)

def get_test_input(model_type: str):
    return "Hi" if model_type == 'embeddings' else [{'role': 'user', 'content': 'Hi'}]

async def get_first_ai_response(profile_data):
    """
    Grab only the first response from the async generator, then close the generator
    so that the aiohttp.ClientSession is properly cleaned up.
    """
    gen = send_ai_request(profile_data)
    try:
        first_response = await gen.__anext__()
        return first_response
    except StopAsyncIteration:
        raise Exception("No response received from AI service.")
    finally:
        await gen.aclose()

