# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from arcwebbase.ai_settings import get_ai_settings
from django.conf import settings
from util.ai import bedrock
from arcwebbase.ai_support import ChunkProcessor

import aiohttp
import os
import re

CLEAR_TEXT_REQUEST = os.environ.get('RETURN_CLEAR_TEXT_REQUEST', '0') == '1'

class AIServiceClient:
    def __init__(self, engine_config):
        self.engine_config = engine_config

    def get_handler(self, engine_type: str):
        """Return an asynchronous request handler."""
        if engine_type == 'bedrock':
            return bedrock.bedrock_comm
        return self.default_http_handler

    async def default_http_handler(self, url, headers, upstream_req_body):
        """Default handler for standard HTTP requests using aiohttp with streaming capability."""

        timeout_seconds = getattr(settings, 'HTTP_PROXY_TIMEOUT', 300)
        client_timeout = aiohttp.ClientTimeout(total=timeout_seconds)

        async with aiohttp.ClientSession(timeout=client_timeout) as session:
            async with session.request(
                method="POST",
                url=url,
                headers=headers,
                json=upstream_req_body,
                ssl=settings.HTTP_PROXY_SSL_VERIFY
            ) as response:
                if upstream_req_body.get('stream', False):
                    # Streaming response
                    async for chunk in response.content.iter_chunked(1024):
                        yield response.status, chunk.decode('utf-8')
                else:
                    # Non-streaming response
                    content_type = response.headers.get('Content-Type', '').lower()
                    if 'application/json' in content_type:
                        data = await response.json()
                    else:
                        data = await response.text()

                    yield response.status, data

    async def send_request(self, profile_data, model_type, engine_type):
        url = self.engine_config.get('url', '').format(**profile_data)

        headers_template = self.engine_config.get('headers', {})

        p = self.engine_config.get('payload', {})
        payload_template = p.get(model_type, p)

        headers = format_recursive(headers_template, profile_data)
        payload = format_recursive(payload_template, profile_data)

        if profile_data.get('REQUEST_TRANSFORMER'):
            headers, payload = profile_data.get('REQUEST_TRANSFORMER')(headers, payload)

        try:
            handler = self.get_handler(engine_type)
            stream_obj = headers if engine_type == 'bedrock' else payload
            streaming = stream_obj.get('stream', False)
            req_info = { 'headers': headers, 'body': payload } if CLEAR_TEXT_REQUEST else None # This includes sensitive information, use it ONLY for debugging

            response_generator = handler(
                url=url,
                headers=headers,
                upstream_req_body=payload
            )

            if streaming:
                processor = ChunkProcessor()

            async for status, body in response_generator:
                if status >= 400:
                    response = extract_error_message(body)
                elif streaming:
                    response = processor.process_chunk(body)
                else:
                    if profile_data.get('RESPONSE_TRANSFORMER'):
                        response = profile_data.get('RESPONSE_TRANSFORMER')(body)
                    else:
                        response = body

                resp = {
                    'status_code': status,
                    'response_body': response
                }

                if req_info:
                    resp['req_info'] = req_info

                yield resp

                # Break after the first response for non-streaming or error
                if not streaming or status >= 400:
                    break
        except Exception as e:
            yield {
                'status_code': 500,
                'response_body': f'Exception occurred: {str(e)}'
            }

async def send_ai_request(profile_data):
    profSettings = profile_data.get('settings')
    model_type = profile_data['model_type']
    engine_type = profSettings.get(f'{model_type}_engine')
    engine_config = get_engine_config(engine_type)

    ss = get_ai_settings(profSettings, model_type)
    ss['INPUT'] = profile_data.get('input')

    client = AIServiceClient(engine_config)

    async for response in client.send_request(ss, model_type, engine_type):
        yield response

def get_shared_openai_caii_config():
    return {
        'url': '{URL}',
        'headers': {
            'Authorization': 'Bearer {SECRET}',
            'Content-Type': 'application/json',
        },
        'payload': {
            'embeddings': {
                'model': '{MODEL}',
                'input': '{INPUT}',
            },
            'completion': {
                'model': '{MODEL}',
                'temperature': '{TEMPERATURE}',
                'messages': '{INPUT}',
                'stream': '{STREAM}',
            },
        },
    }

ENGINE_CONFIGS = {
    'open_ai': get_shared_openai_caii_config(),
    'caii': get_shared_openai_caii_config(),
    'azure_ai': {
        'url': '{URL}',
        'headers': {
            'api-key': '{SECRET}',
            'Content-Type': 'application/json',
        },
        'payload': {
            'embeddings': {
                'input': '{INPUT}',
            },
            'completion': {
                'messages': '{INPUT}',
                'stream': '{STREAM}',
            },
        },
    },
    'llama': {
        'url': '{URL}',
        'headers': '{HEADERS}',
        'payload': {
            'accessKey': '{SECRET}',
            'request': '{REQUEST}'
        },
    },
    'bedrock': {
        'headers': {
            'model_id': '{MODEL}',
            'credentials': '{CREDENTIALS}',
            'stream': '{STREAM}',
            'custom_get_chunk_text': '{CUSTOM_GET_CHUNK_TEXT}',
        },
        'payload': {
            'inputText': '{INPUT}',
        },
    },
    'other': {
        # headers and payload are generated with transformers, secret and input are provided
        'url': '{URL}',
        'headers': {
            'secret': '{SECRET}',
        },
        'payload': {
            'input': '{INPUT}',
        },
    },
}

# Pattern to match exact placeholders like "{SECRET}"
EXACT_PLACEHOLDER_PATTERN = re.compile(r'^\{(\w+)\}$')

# Pattern to match embedded placeholders like "Bearer {SECRET}"
EMBEDDED_PLACEHOLDER_PATTERN = re.compile(r'\{(\w+)\}')

# Unique sentinel value
UNRESOLVED = object()

def format_recursive(obj, context):
    if isinstance(obj, dict):
        new_dict = {}
        for k, v in obj.items():
            formatted_key = format_recursive(k, context)
            formatted_value = format_recursive(v, context)
            # Skip this key if the value is unresolved.
            if formatted_value is UNRESOLVED:
                continue
            new_dict[formatted_key] = formatted_value
        return new_dict
    elif isinstance(obj, list):
        new_list = []
        for item in obj:
            formatted_item = format_recursive(item, context)
            # Skip unresolved items in lists if desired.
            if formatted_item is UNRESOLVED:
                continue
            new_list.append(formatted_item)
        return new_list
    elif isinstance(obj, str):
        # Check for an exact placeholder (the whole string)
        exact_match = EXACT_PLACEHOLDER_PATTERN.match(obj)
        if exact_match:
            key = exact_match.group(1)
            if key not in context:
                return UNRESOLVED
            return format_recursive(context.get(key), context)

        # Replace embedded placeholders
        def replace_placeholder(match):
            key = match.group(1)
            # If the key is not found, remove the placeholder
            return str(context.get(key, ""))
        return EMBEDDED_PLACEHOLDER_PATTERN.sub(replace_placeholder, obj)
    else:
        return obj

def extract_error_message(data):
    # If data is a string from response.text() or chunk
    if isinstance(data, str):
        return data.strip() or "An unknown error occurred."

    # If data is a JSON object
    if isinstance(data, dict):
        if 'error' in data:
            error_info = data['error']
            if isinstance(error_info, dict):
                message = error_info.get('message', '')
                code = error_info.get('code', '')
                return f"{message} (Code: {code})" if code else message
            else:
                # If 'error' is a string
                return error_info
        elif 'message' in data:
            return data['message']
        elif 'detail' in data:
            return data['detail']
        else:
            # For unknown structures
            return str(data)
    else:
        # If data is neither a dict nor a string
        return "An unknown error occurred."

def get_engine_config(engine_type):
    config = ENGINE_CONFIGS.get(engine_type)
    if not config:
        raise ValueError(f'Unknown engine type: {engine_type}')
    return config
