# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import util.cml_models
import os
import util.caii_models
import json
import logging
from functools import partial
from reports.views_settings import setting_get_single
from jobs.views_encryption import decrypt_secret
from util.ai import bedrock_transformers as bt
from django.conf import settings
from dataconn.cml_dependencies import get_cmldata

LOGGER = logging.getLogger("arcweb.ai_settings")

OPENAI_EMBEDDING_URL = 'https://api.openai.com/v1/embeddings'
OPENAI_COMPLETION_URL = 'https://api.openai.com/v1/chat/completions'

BEDROCK_MODELS = {
    'embeddings': ['amazon.titan-embed-text-v1', 'amazon.titan-embed-text-v2:0'],
    'completion': ['amazon.titan-text-lite-v1', 'amazon.titan-text-express-v1', 'anthropic.claude-3-5-sonnet-20241022-v2:0', 'anthropic.claude-3-5-haiku-20241022-v1:0', 'anthropic.claude-3-opus-20240229-v1:0']
}

def get_ai_settings(active_settings: dict, setting_key: str) -> dict:
  """
  These are needed configs per engine type:
  CAII => SECRET, URL, MODEL, RESPONSE_TRANSFORMER, TEMPERATURE (for completion only)
  OpenAI Azure => SECRET, URL
  Llama => SECRET (access key), Auth key (optional), URL, REQUEST_TRANSFORMER, RESPONSE_TRANSFORMER, EXTRA_ARGS, INPUT + RESPONSE PARAMS
  OpenAI => SECRET, MODEL, TEMPERATURE (for completion only)
  Bedrock => AWS Credentials, MODEL, REQUEST_TRANSFORMER and RESPONSE_TRANSFORMER
  Other => SECRET, URL, custom REQUEST_TRANSFORMER and RESPONSE_TRANSFORMER functions from AI Engine Settings
  """

  ai_engine = active_settings[f'{setting_key}_engine']
  def_settings = {}

  if ai_engine == 'bedrock':
    model = active_settings.get(f'{setting_key}_bedrock_model')
    def_settings['MODEL'] = model
    def_settings['CREDENTIALS'] = try_decrypt_key(active_settings[f'{setting_key}_bedrock_credentials'])

    if setting_key == 'completion':
      def_settings['STREAM'] = active_settings.get(f'{setting_key}_stream', False)

    if setting_key in BEDROCK_MODELS and model not in BEDROCK_MODELS[setting_key]:
        transformers = active_settings[f'{setting_key}_bedrock_custom_transformers']
        if transformers:
          locs = {}

          exec(transformers, None, locs)

          def_settings['REQUEST_TRANSFORMER'] = locs.get('REQUEST_TRANSFORMER')
          def_settings['RESPONSE_TRANSFORMER'] = locs.get('RESPONSE_TRANSFORMER')
          def_settings['CUSTOM_GET_CHUNK_TEXT'] = locs.get('CUSTOM_GET_CHUNK_TEXT')

          return def_settings

    if setting_key == 'completion':
      def_settings['REQUEST_TRANSFORMER'] = bt.completion_request_transformer
      def_settings['RESPONSE_TRANSFORMER'] = bt.completion_response_transformer

    if setting_key == 'embeddings':
      def_settings['RESPONSE_TRANSFORMER'] = bt.embedding_response_transformer


    return def_settings

  api_key = active_settings.get(f'{setting_key}_api_key')
  if api_key:
    def_settings['SECRET'] = try_decrypt_key(api_key)

  if ai_engine == 'open_ai':
    def_settings['URL'] = OPENAI_EMBEDDING_URL if setting_key == 'embeddings' else OPENAI_COMPLETION_URL
    def_settings['MODEL'] = active_settings[f'{setting_key}_model'].strip()

    if setting_key == 'completion':
      def_settings['TEMPERATURE'] = float(active_settings[f'{setting_key}_temperature'])
      def_settings['STREAM'] = active_settings.get(f'{setting_key}_stream', False)

    return def_settings


  def_settings['URL'] = active_settings[f'{setting_key}_service_url'].strip()
  if ai_engine == 'caii':
    def_settings['MODEL'] = active_settings[f'{setting_key}_model'].strip()

    is_cml_auth_mode_jwt = active_settings.get(f'{setting_key}_cml_auth_mode', 'api_key') == 'jwt'
    if is_cml_auth_mode_jwt:
      cmldata = get_cmldata()
      if not cmldata:
        raise Exception("Can't authenticate with JWT because could not import cmldata library")
      def_settings['SECRET'] = cmldata.get_jwt()

    if setting_key == 'completion':
      def_settings['TEMPERATURE'] = float(active_settings[f'{setting_key}_temperature'])
      def_settings['STREAM'] = active_settings.get(f'{setting_key}_stream', False)

    def_settings['RESPONSE_TRANSFORMER'] = util.caii_models.err_resp_transformer

    return def_settings

  if ai_engine == 'azure_ai':
    if setting_key == 'completion':
      def_settings['STREAM'] = active_settings.get(f'{setting_key}_stream', False)
    return def_settings

  if ai_engine == 'llama':
    input_param = active_settings[f'{setting_key}_input_param']
    response_param = active_settings[f'{setting_key}_response_param']
    request_transformer = util.cml_models.completion_request_transformer
    response_transformer = util.cml_models.embedding_response_transformer if setting_key == 'embeddings' else util.cml_models.completion_response_transformer
    auth_api = active_settings.get(f'{setting_key}_auth_key', '')

    def_settings['HEADERS'] = {
      'Content-Type': 'application/json'
    }

    if not auth_api:
      auth_api = os.environ.get("CDSW_APIV2_KEY", "")

    if auth_api:
      def_settings['HEADERS']['Authorization'] = f'Bearer {try_decrypt_key(auth_api)}'

    def_settings['REQUEST'] = {
      input_param: '{INPUT}'
    }

    def_settings['RESPONSE_TRANSFORMER'] = partial(response_transformer, param_name=response_param)

    if setting_key == 'completion':
      extra_args_str = active_settings.get(f'{setting_key}_extra_arg', '{}') or '{}'

      try:
        extra_args = json.loads(extra_args_str)
      except json.JSONDecodeError:
        LOGGER.error(f"Error decoding JSON: {str(e)}")
        extra_args = {}

      def_settings['REQUEST'].update(extra_args)
      def_settings['REQUEST_TRANSFORMER'] = partial(request_transformer, param_name=input_param)

    return def_settings

  # engine is "other", getting and processing the custom code
  ai_settings = active_settings[f'{setting_key}_custom_engine_settings']
  locs = {}

  exec(ai_settings, None, locs)

  locs['SECRET'] = def_settings['SECRET']
  locs['URL'] = def_settings['URL']

  return locs

def get_ai_settings_prop(model_type: str = '', profile: str = '') -> dict:
    if not model_type:
      return {}

    try:
      profs = json.loads(setting_get_single(f'ai_{model_type}_settings'))
      default_prof = profs.get('defaultProf', {})
      profile_settings = profs.get('profileSettings', {})
    except json.JSONDecodeError as e:
      LOGGER.error(f"Error decoding JSON: {str(e)}")
      return {}

    active_profile = profile_settings.get(profile, profile_settings.get(default_prof['value']))

    if not active_profile:
      return {}

    active_values = active_profile.get('values', {})

    return active_values

def try_decrypt_key(key):
  # The key can be an encrypted value if it comes from the profile (db), or a non-crypted one if it comes from the profile_validation request
  try:
    secret_data = json.loads(key)
    result = decrypt_secret(**secret_data)
  except json.JSONDecodeError:
    result = key

  return result