# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from util.ff_flipper import FeatureFlipper
from vizlib import querylog
import json

import logging

LOGGER = logging.getLogger('arcweb.arcwebbase.ai_support')


def initialize():
  querylog.REDACTORS.append(redact_query)


def is_log_redacting_enabled():
  return FeatureFlipper().get_flipper_value('redact_ai_assistant_logs')


def get_query_dict(query: str):
  if query and query[0] == '{' and query[-1] == '}':
    # looks like JSON:
    try:
      queryDict = json.loads(query)
      return queryDict
    except:
      pass


def redact_query(query: str):
  if not is_log_redacting_enabled():
    return False, None

  queryDict = get_query_dict(query)
  if queryDict:
    aiQuery: bool = queryDict.get('aiQuery', False)
    if aiQuery:
      body = queryDict.get('body')
      if body:
          queryDict['body'] = 'REDACTED'
          return True, json.dumps(queryDict)

  return False, None

class ChunkProcessor:
  def __init__(self):
    self.partial_chunk = ""

  def process_chunk(self, chunk):
    # Split the chunk by "data:" lines
    lines = chunk.split("data: ")
    result = []

    for line in lines:
      stripped_line = line.strip()

      if stripped_line:  # Ensure it's not an empty line
        try:
          # Handle the [DONE] signal
          if stripped_line == "[DONE]":
            LOGGER.info("Streaming complete.")
            return result  # End the processing

          # If there is a partial chunk from the previous response, prepend it
          if self.partial_chunk:
            line = self.partial_chunk + line
            self.partial_chunk = ""  # Reset after prepending

          # Parse the JSON data
          json_data = json.loads(line)

          choices = json_data.get('choices', [])
          if not choices:
              LOGGER.info("No choices found in chunk; ignoring.")
              continue

          # Handle the final chunk with "finish_reason": "stop"
          finish_reason = choices[0]['delta'].get('finish_reason', None)
          if finish_reason == "stop":
            LOGGER.info("Completion finished with 'stop' reason.")
            continue  # Skip further processing as content is empty

          # Extract content if present
          content = choices[0]['delta'].get('content', '')
          if content:
            result.append(content)

        except json.JSONDecodeError:
          # If JSON is incomplete, store it for the next chunk
          self.partial_chunk = line
          LOGGER.info(f"Incomplete JSON, storing for next chunk")

    return result