# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import copy
import tiktoken
from typing import List
from django.http import HttpRequest, JsonResponse
from arcwebbase.async_util import django_run_async
from arcwebbase.ai_settings import get_ai_settings_prop

citation_instruction = '\nInclude a line at the end of your response in the format "@@@Citations: 3,5,7@@@" to indicate which row(s) you used from the provided source data. If you did not use any particular rows, you can omit this line.'

def num_tokens_from_string(string: str, encoding_name: str = "cl100k_base") -> int:
    # Returns the number of tokens in a text string.
    encoding = tiktoken.get_encoding(encoding_name)
    num_tokens = len(encoding.encode(string))

    return num_tokens

def get_context_string(rows: List[List[str]], embedding_source_aliases: List[str], context_placeholder: str):
    result = ''

    for row_index, row in enumerate(rows):
        row_result = ''

        for column_index, column in enumerate(row):
            column = column or context_placeholder

            if column:
                alias = embedding_source_aliases[column_index]
                row_result += f"{alias} {row_index + 1}:\n{column}\n"

        if row_result:
            result += row_result + '\n'  # Add extra newline between rows

    if not result:
        raise Exception('Context data is missing or empty.')

    return result

async def get_ai_settings_prop_async(model_type: str, profile: str):
  return await django_run_async(get_ai_settings_prop, model_type, profile)

async def validate_and_transform_request(req: dict) -> dict:
    req = copy.deepcopy(req)
    req_body = req.get('body')
    req_context = req.get('modelType')
    req_profile = req.get('profile')

    if not req_body:
       raise Exception('Request body is missing.')
    if not req_context or not req_profile:
       raise Exception('Model type or profile id is missing.')

    ups_req_body = req_body.get('upstreamReqBody', {})
    config = req_body.get('config', {})

    # check token limitation
    if req_context == 'embeddings':
        term = ups_req_body.get('input', '')
        max_tokens = int(config.get('embeddingMaxTokens', 0))
        num_tokens = num_tokens_from_string(term)

        if max_tokens > 0 and num_tokens > max_tokens:
            error_message = 'This query exceeds token limitations in the visual settings. This can be increased by changing the "Maximum tokens" under "Embeddings" Settings. Increasing this limitation does not change any token limitations of the LLM or embeddings model'
            raise Exception(error_message)

    if req_context == 'completion':
        user_messages = ups_req_body.get('input', [])
        source_info = config.get('sourceInfo', {})
        rows = source_info.get('filteredRows', [])
        aliases = source_info.get('aliases', [])
        param_selection = config.get('paramSelection', 'all')
        context_placeholder = source_info.get('contextPlaceholder', '')
        embedding_context = get_context_string(rows, aliases, context_placeholder)
        concatenated_messages = "".join(message['content'] for message in user_messages)
        max_tokens = int(config.get('completionMaxTokens', 0))
        num_tokens = num_tokens_from_string(embedding_context + concatenated_messages)

        if max_tokens > 0 and num_tokens > max_tokens:
            error_message = 'This query exceeds token limitations in the visual settings. This can be increased by changing the "Maximum tokens" under "Completion" Settings. Increasing this limitation does not change any token limitations of the LLM or embeddings model'

            def remove_message_and_calculate_token():
                user_messages.pop(0)
                concatenated_messages = "".join(message['content'] for message in user_messages) # save tokens for each messages
                return num_tokens_from_string(concatenated_messages + embedding_context)

            def remove_context_and_calculate_token():
                rows.pop(0)
                embedding_context = get_context_string(rows, aliases, context_placeholder)
                return num_tokens_from_string(embedding_context + concatenated_messages)

            def throw_overflow_error():
                raise Exception(error_message)

            overflow_policy = config.get('overflowPolicy', 'throw-error')

            if overflow_policy == 'throw-error':
                return throw_overflow_error()
            elif overflow_policy == 'remove-conversation-first':
                while num_tokens > max_tokens and len(user_messages) > 1:
                    num_tokens = remove_message_and_calculate_token()

                # truncate context if num_tokens still exceeds max_tokens after removing all the messages
                if num_tokens > max_tokens:
                    while num_tokens > max_tokens and rows:
                        num_tokens = remove_context_and_calculate_token()

                # throw and error if num_tokens still exceeds max_tokens
                if num_tokens > max_tokens:
                    return throw_overflow_error()
            elif overflow_policy == 'truncate-context-first':
                while num_tokens > max_tokens and rows:
                    num_tokens = remove_context_and_calculate_token()

                # remove old conversations if num_tokens still exceeds max_tokens after removing all the embedding context
                if num_tokens > max_tokens:
                    while num_tokens > max_tokens and len(user_messages) > 1:
                        num_tokens = remove_message_and_calculate_token()

                    if num_tokens > max_tokens:
                        return throw_overflow_error()

        system_message = {'role': 'system'}
        context_prompt = source_info.get('contextPrompt', '')
        if param_selection == 'used-by-llm':
            context_prompt += citation_instruction

        if '<<context>>' in context_prompt:
            system_message['content'] = context_prompt.replace('<<context>>', embedding_context)
        else:
            system_message['content'] = embedding_context

        conversation = user_messages.copy()
        conversation.insert(len(user_messages) - 1, system_message)

        ups_req_body['input'] = conversation

    req['body'] = ups_req_body
    req['settings'] = await get_ai_settings_prop_async(req_context, req_profile)

    return req