# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging

from djangosaml2.backends import Saml2Backend
from django.conf import settings
from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader, module_from_spec
from arcwebbase.auth.saml.saml_post_authenticate import saml_post_authenticate
from arcwebbase.auth.saml.update_user_groups import update_user_groups
from arcwebbase.auth.common.extend_user_profile_with_authentication_source import extend_user_profile_with_authentication_source

# create logger
LOGGER = logging.getLogger("arcweb")


SAML_REQUIRED_GROUPS_CHECK = 'SAML_REQUIRED_GROUPS_CHECK'
SAML_REQUIRED_GROUPS_ATTR = 'SAML_REQUIRED_GROUPS'
SAML_REQUIRED_ADMIN_GROUPS_ATTR = 'SAML_REQUIRED_ADMIN_GROUPS'
SAML_GROUP_ATTRIBUTE_NAME = 'SAML_GROUP_ATTRIBUTE_NAME'


class VizSaml2Backend(Saml2Backend):
  # TODO: Update with VIZ-234
  # For validating if the user has the right groups we need the attributes info
  # and the user info as well, currently the djangosaml2 lib does not provide
  # a hook for getting all this info. The update_user is providing it in the
  # 0.17.2 verison of the lib, but that is deprecated in the django3 version.
  # So till they provide a better hook we have to do the same logic as in the
  # original authenticate method. Unfortunately it changes from the current
  # 0.17.2 version we use to the django3 version of the library so we need to
  # update it when we update the djangosaml2 lib on the django3 update.
  # Update the get_attributes function based on this:
  # https://github.com/IdentityPython/djangosaml2/blob/master/djangosaml2/backends.py#L110
  def authenticate(self, request, session_info=None, attribute_mapping=None, **kwargs):
    attributes = self.get_attributes(session_info, attribute_mapping)
    LOGGER.debug("SAML attributes from assertion response:")
    LOGGER.debug(attributes)
    if attributes is None:
      return None
    # Base authentication process will trigger _update_user where we'll check for group assertion and sync users to the
    #   the appropriate groups they match with
    user = super().authenticate(request, session_info, attribute_mapping, **kwargs)
    if not user:
      LOGGER.debug("SAML user not authorized by SAML Identity Provider settings. "
                   "Please check with Identity Provider to investigate further.")
      return None

    # Update User Profile with indicating authentication source as SAML
    user = extend_user_profile_with_authentication_source(user, 'SAML')

    # Get SAML group attribute name using SAML_GROUP_ATTRIBUTE_NAME setting
    group_attribute_name = getattr(settings, SAML_GROUP_ATTRIBUTE_NAME, 'groups')

    # Check if Group assertion attribute is in SAML assertion, and if require group settings are present
    group_attribute_in_assertion = group_attribute_name in attributes
    is_verify_require_groups_on = hasattr(settings, SAML_REQUIRED_ADMIN_GROUPS_ATTR) and \
                                  hasattr(settings, SAML_REQUIRED_GROUPS_ATTR)

    # If Required Groups check is turned on for authorizing User access to CDV,
    #  and the group attribute is present in the SAML assertion,
    #  then proceed with Require Group and Admin Group checking
    if getattr(settings, 'SAML_REQUIRED_GROUPS_CHECK', False) and group_attribute_in_assertion:
      if is_verify_require_groups_on:
        # If group attribute is set, and if admin and normal (required) groups are specified,
        #  assign them an Admin or non-Admin appropriately
        LOGGER.debug("SAML Require Group checking is turned on, and user is a member of at least 1 group. "
                     "Syncing admin and regular user lists.")
        if self.has_group_from_admin_list(attributes):
          user = self.set_user_is_admin(user, True)
          LOGGER.debug("SAML Require Group checking verified user is in at least one Admin group. Login authorized.")
          return user

        if self.has_group_from_regular_list(attributes):
          user = self.set_user_is_admin(user, False)
          LOGGER.debug("SAML Require Group checking verified user is in at least one Normal group. Login authorized.")
          return user

        # Required Group checks enabled, but User is not a member of any group in the SAML_REQUIRED_GROUPS_ATTR
        #  or SAML_REQUIRED_ADMIN_GROUPS_ATTR group list. Returning None will result in a 403 forbidden error,
        #  disabling user authentication.
        LOGGER.debug("SAML user %s is a member of some groups according to SAML assertion, but not a member of "
                     "any in the Normal User or Admin User group list. "
                     "Adding user to any group in the Normal User group list setting: %s "
                     "or Admin User group list setting: %s will allow them to login." %
                     (user.username, SAML_REQUIRED_GROUPS_ATTR, SAML_REQUIRED_ADMIN_GROUPS_ATTR))
        return None
      else:
        # Required Group checks enabled, but group assertion missing which means they aren't in a group we can check for
        #  Normal or Admin user access. Returning None will result in a 403 forbidden error,
        #  disabling user authentication.
        LOGGER.debug("SAML user %s is not part of any groups according to SAML assertion. "
                     "Adding user to any group in the Normal User group list setting: %s "
                     "or Admin User group list setting: %s will allow them to login." %
                     (user.username, SAML_REQUIRED_GROUPS_ATTR, SAML_REQUIRED_ADMIN_GROUPS_ATTR))
        return None

    else:
      # Require Group verification not turned on. Proceed with authentication, user
      LOGGER.debug("SAML Require Group checking is turned off. Login authorized.")
      return user

  def has_group_from_admin_list(self, attributes):
    # The admin group list is required
    return self.are_groups_overlapping(attributes, SAML_REQUIRED_ADMIN_GROUPS_ATTR)

  def has_group_from_regular_list(self, attributes):
    # We only check the regular user groups if specified
    # VIZ-2388 -> We need to check for None in addition to empty array
    if not getattr(settings, SAML_REQUIRED_GROUPS_ATTR, None):
      return True
    return self.are_groups_overlapping(attributes, SAML_REQUIRED_GROUPS_ATTR)

  def are_groups_overlapping(self, attributes, groups_list_name):
    groups_to_verify = getattr(settings, groups_list_name)
    groups_on_user = attributes['groups']
    return len(set(groups_on_user).intersection(set(groups_to_verify))) > 0

  def set_user_is_admin(self, user, is_admin):
    if is_admin != user.is_superuser:
      user.is_superuser = is_admin
      user.save(update_fields=['is_superuser'])

    return user

  def get_attributes(self, session_info, attribute_mapping):
    # Added after Django 3.x changes from djangosaml2
    idp_entityid = session_info['issuer']

    SESSION_INFO_ATTRIBURTE_FOR_SAML = 'ava'
    if session_info is None or attribute_mapping is None:
      return None

    if SESSION_INFO_ATTRIBURTE_FOR_SAML not in session_info:
      return None

    return self.clean_attributes(session_info[SESSION_INFO_ATTRIBURTE_FOR_SAML], idp_entityid)

  def _update_user(self, user, attributes, attribute_mapping, force_save=False):

    # Always save a brand new user instance
    if user.pk is None:
      user = self.save_user(user)

    # No attributes to set on the user instance, nothing to update
    if not attribute_mapping:
      return user

    # Setup the post authenticate process (default or from plugin specified in settings)
    SAML_POST_AUTHENTICATE = saml_post_authenticate
    fullpath = getattr(settings, 'SAML_POST_AUTHENTICATE_PLUGIN', None)
    if fullpath:
      try:
        spec = spec_from_loader('saml_post_authenticate', SourceFileLoader('saml_post_authenticate', fullpath))
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        SAML_POST_AUTHENTICATE = module.post_authenticate
      except Exception as e:
        LOGGER.exception('Error loading SAML post authenticate plugin')

    force_save |= SAML_POST_AUTHENTICATE(user, attributes, attribute_mapping, settings, LOGGER, update_user_groups)

    # Save user instance if new user or updates made to group membership
    if force_save:
      user = self.save_user(user)

    return super(VizSaml2Backend, self)._update_user(user, attributes, attribute_mapping, force_save)