# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import grp
import logging
import pwd
from ctypes import *

import pam
from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import User, Group
from arcwebbase.auth.saml.backend import VizSaml2Backend

# create logger
LOGGER = logging.getLogger("arcweb")


class Saml2Backend(VizSaml2Backend):
  def __init__(self):
    LOGGER.warning('WARNING: This arcwebbase.backends.Saml2Backend is deprecated in favor of '
                   'arcweb.arcwebbase.backends.VizBaseSaml2Backend. Please update your Advanced Settings '
                   '(Site Settings menu) AUTHENTICATION_BACKENDS setting to switch to the new authentication backend.')

    super().__init__()


from django_auth_ldap.backend import LDAPBackend
class LdapBackend(LDAPBackend):

  def get_or_create_user(self, username, ldap_user):
    """
    This must return a (User, created) 2-tuple for the given LDAP user.
    username is the Django-friendly username of the user. ldap_user.dn is
    the user's DN and ldap_user.attrs contains all of their LDAP attributes.
    """

    LOGGER.warning('WARNING: This arcwebbase.backends.LDAPBackend is deprecated in favor of '
                   'arcweb.arcwebbase.backends.VizBaseLDAPBackend. Please update your Advanced Settings '
                   '(Site Settings menu) AUTHENTICATION_BACKENDS setting to switch to the new authentication backend.')

    model = self.get_user_model()
    username_field = getattr(model, 'USERNAME_FIELD', 'username')

    # Obtain the name for the local django copy (if it gets created)
    localname = username.lower()
    key = getattr(settings, 'LDAP_LOCAL_USERNAME_FROM_ATTR', None)
    if getattr(settings, 'LDAP_LOCAL_USERNAME_AS_TYPED', None):
      localname = username
    elif key:
      try:
        val = ldap_user.attrs.get(key)
        localname = val[0]
        if username.lower() != localname.lower():
          raise Exception('Names can only differ in case')
      except Exception as e:
        LOGGER.warning('Error using %s from LDAP attribute %s to create local username for %s: %s' %
                       (val, key, username, e))
        localname = username.lower()
    kwargs = {
        username_field + '__iexact': username,
        'defaults': {username_field: localname}
    }

    return model.objects.get_or_create(**kwargs)

#
# This backend needs getgrouplist() to provide unix group membership information
# os.getgrouplist() is provided in Python 3.3+ only, we provide an
# implementation if not available
# For linux we use ctypes and call the libc function directly
# As a fallback (osx) we use a pure-python implementation
#

_libc = None

def _getgrouplist1(user, group):
  # On linux use ctypes to load libc.so.6 and use getgrouplist directly
  # 1 Million in 37.30 seconds
  MAXGROUPS = 1000
  groups = (c_uint * MAXGROUPS)()
  ngroups = c_uint(MAXGROUPS)
  grpcnt = _libc.getgrouplist(user, group, groups, pointer(ngroups))
  return [groups[i] for i in range(0, grpcnt)]


def _getgrouplist2(user, group):
  # Pure-python implementation
  # 1 Million in 78.56 seconds
  groups = [group]
  groups.extend([g.gr_gid for g in grp.getgrall() if
                 user in g.gr_mem and g.gr_gid != group])
  return groups


# logic to get the correct version of getgrouplist()
try:
  from os import getgrouplist
except ImportError:
  try:
    _libc = cdll.LoadLibrary("libc.so.6")  # gnu libc RHEL6
    getgrouplist = _getgrouplist1
  except OSError:
    getgrouplist = _getgrouplist2


def get_unix_groups(username):
  """
  Return a list of unix group names for a user
  """
  user = pwd.getpwnam(username)
  gnames = [grp.getgrgid(g).gr_name for g in
            getgrouplist(username, user.pw_gid)]
  LOGGER.debug('get_unix_groups(%s) = %s' % (username, gnames))
  return gnames


class UnixUser(object):
  """
  An object containing passwd(5) info + a list of group_names
  """
  def __init__(self, username):
    user_pwd = pwd.getpwnam(username)
    self.pw_name = user_pwd.pw_name
    self.pw_uid = user_pwd.pw_uid
    self.pw_gid = user_pwd.pw_gid
    self.pw_gecos = user_pwd.pw_gecos
    self.pw_dir = user_pwd.pw_dir
    self.pw_shell = user_pwd.pw_shell
    self.group_names = get_unix_groups(username)


class PAMUser(User):
  """
  PAMUser is a proxy for A Django User object with 2 additional properties
  is_unix_user:  Use this property to check if this has a unix_user property
                 hasattr(user,'unix_user') is not efficient
                 because it executes unix_user()
  unix_user: A UnixUser object
  """

  is_unix_user = True
  _unix_user = None  # cache unix_user for a given instance of PAMUser

  class Meta:
    app_label = 'PAMUser'
    proxy = True

  @property
  def unix_user(self):
    # Note: No caching between instances for now. permcache.py takes care of this
    if self._unix_user is not None:
      return self._unix_user
    self._unix_user = UnixUser(self.username)
    return self._unix_user


# PAM Auth backend inherits from the standard ModelBackend
class PamBackend(ModelBackend):
  """
  A Django Authentication Backend for PAM
  """

  def authenticate(self, username=None, password=None, **kwargs):
    # Check the username/password and return a User.
    p = pam.pam()
    pam_service = getattr(settings, 'PAM_AUTH_SERVICE', 'login')
    if p.authenticate(username, password, service=pam_service):
      try:
        user = PAMUser.objects.get(username=username)
      except PAMUser.DoesNotExist:
        if getattr(settings, 'PAM_AUTH_CREATE_USER', False):
          user, newuser = PAMUser.objects.get_or_create(username=username)
          user.save()
          LOGGER.warning('PamBackend: created user "%s"' % (username))
        else:
          LOGGER.warning(
            'PamBackend: user "%s" authenticated via PAM, but does not exist.' % (
            username))
          return None
      # Try to add unix_user and unix_user.group_names attributes
      LOGGER.info('PamBackend: user "%s" authenticated via PAM' % (username))
      return user
    LOGGER.warning('PamBackend: "%s" authentication failed: %s: %s' % (
    username, p.code, p.reason))
    return None

  def get_user(self, user_id):
    try:
      return PAMUser.objects.get(pk=user_id)
    except PAMUser.DoesNotExist:
      return None
