# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
etag header support to allow "smart" browser caching of metadata
"""

# By adding the "@etag(etag_func)' decorator to "get" views of APIS
# django will add an etag header with a unique etag to the response.
# This allows the browser to cache the results of these APIs.
# The browser will send the etag with the next request and if it has not changed
# the django server will return a 304 status which will tell the
# browser to use the cached data.
#
# The etag will be changed if any of the metadata is modified causing fresh
# metadata to be fetched and cached.
#
# In this implementation the etag is the comprised of the md5 of the
# concatenation of three strings:
# 1) The session_id. This is unique per browser session and per user so the
#    cached data will only be used for the same user.
# 2) A random string that changes if the user's groups change.  User group
#    membership is handled by etag_check_user_groups() called from the
#    RBAC permcache when users groups are enumerated.
# 3) A random string that is changed whenever the underlying model that holds
#    data used in the API is modified. It also get changed if the rbac.Role,
#    metadata changes, in case the user no longer has permission to view some
#    data.
#
# The views that use etags are split into two classes based one the models
# that contribute to the JSON response
# 1) METADATA for reports/report, report/appgroup, and datasets/dataset APIs
# 2) ASSET for views based on: StaticAsset, CustomCSS, ColorPalette,
#    ParameterSet, DateRanges, AppSetting models
#
# There are two new settings:
# ETAGS_FOR_AJAX_ENABLE - Enable etags
# ETAGS_AJAX_WITH_ARGS_ENABLE - Enable etags for api call with arguments




from hashlib import md5

import re
from django.conf import settings
from django.db.models.signals import post_save, post_delete, m2m_changed
from django.dispatch import receiver
from django.utils.crypto import get_random_string
from django.core.cache import cache

# These are lists of the Models that should cause us to invalidate the
# browser cache by creating a new etag
ETAGS_METADATA_CLASS_LIST = ['reports.models.Report',
                             'reports.models.ReportImage',
                             'datasets.models.DataSet',
                             'datasets.models.Extract',
                             'datasets.models.DataConnection',
                             'datasets.models.Segment',
                             'datasets.models.FilterAssociation',
                             'datasets.models.Event',
                             'reports.models.AppGroup',
                             'reports.models.AppGroupAppMembership',
                             'reports.models.UserAppGroupMembership',
                             'rbac.models.Role',
                             'reports.models.ColorPalette',
                             'reports.models.DateRanges',
                             'workspace.models.Workspace',
                             ]

ETAGS_ASSETS_CLASS_LIST = ['reports.models.StaticAsset',
                           'reports.models.CustomCSS',
                           'reports.models.ParameterSet',
                           'reports.models.AppSetting',
                           ]

# This is a regex to extract the class name from the sender
CLASSNAME_RE = re.compile(r'class[^a-zA-Z]+([a-zA-Z\.]+)')

USER_GROUP_CACHE = {}

ETAGS_FOR_AJAX_ENABLE = getattr(settings, "ETAGS_FOR_AJAX_ENABLE", False)
ETAGS_AJAX_WITH_ARGS_ENABLE = getattr(settings, "ETAGS_AJAX_WITH_ARGS_ENABLE", False)

def _etag_set(key):
  val = get_random_string(32)
  cache.set(key, val)
  return val

def _etag_get(key):
  val = cache.get(key)
  if not val:
    val = _etag_set(key)
  return val


def _etag_func(request, *args, **kwargs):
  etag_class = kwargs["_ETAG_CLASS"]
  del kwargs["_ETAG_CLASS"]
  if not ETAGS_FOR_AJAX_ENABLE:
    return None
  if not ETAGS_AJAX_WITH_ARGS_ENABLE:
    # Disable caching of requests for specific objects
    if any(args) or kwargs:
      return None
  if not hasattr(request, 'session') or not hasattr(request, 'user'):
    return None
  session_key = request.session.session_key
  if session_key is None:
    return None
  user_key = _etag_get('USER_%s_CACHE_KEY' % request.user.id)
  etag_key = _etag_get('ETAG_%s_CACHE_KEY' % etag_class)
  etag = md5((session_key + user_key + etag_key).encode()).hexdigest()
  return etag


def etag_metadata_func(request, *args, **kwargs):
  return _etag_func(request, _ETAG_CLASS="METADATA", *args, **kwargs)


def etag_asset_func(request, *args, **kwargs):
  return _etag_func(request, _ETAG_CLASS="ASSETS", *args, **kwargs)


def etag_check_user_groups(user, groups):
  oldgroups = USER_GROUP_CACHE.get(user.id)
  if oldgroups != groups:
    USER_GROUP_CACHE[user.id] = groups
    _etag_set('USER_%s_CACHE_KEY' % user.id)


@receiver(post_delete)
@receiver(m2m_changed)
@receiver(post_save)
def update_etag(sender, **kwargs):
  sender_class = CLASSNAME_RE.search(str(sender)).group(1)
  if sender_class == 'django.contrib.auth.models.User':
    _etag_set('USER_%s_CACHE_KEY' % kwargs['instance'].id)
  if sender_class in ETAGS_METADATA_CLASS_LIST:
    _etag_set('ETAG_METADATA_CACHE_KEY')
  if sender_class in ETAGS_ASSETS_CLASS_LIST:
    _etag_set('ETAG_ASSETS_CACHE_KEY')

