# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import gc
import logging
import os
import psutil
import time

LOGGER = logging.getLogger('arcweb.gcmonitor')


class GCCallback:
  """
  Track time spent in the garbage collector.

  We assume garbage collector stops the world: when it runs, no other threads execute.
  Because of this there is no synchronization implemented, we assume that the only instance
  of this class is always accessed by a single thread only (the one performing the garbage collection).
  """

  def __init__(self):
    self.process = psutil.Process(os.getpid())
    self.reset()

  def reset(self):
    self.start_ns = None
    self.total_ns = 0
    self.last_log_ns = time.perf_counter_ns()

  def register(self):
    if not self.enabled():
      self.reset()
      gc.callbacks.append(self)
      LOGGER.info('GC: monitor registered')

  def unregister(self):
    if self.enabled():
      gc.callbacks.remove(self)
      LOGGER.info('GC: monitor unregistered')

  def enabled(self):
    return self in gc.callbacks

  def get_psutil_line(self):
    """
    Return memory usage and cpu load percentage (averaged on last 1, 5 and 15 minutes)
    """
    memory_mb = self.process.memory_info().rss / 1024 / 1024

    cpu_count = psutil.cpu_count()
    cpu_percentage = ["{:.2f}".format(load_avg / cpu_count * 100)
                      for load_avg in psutil.getloadavg()]

    msg = f"psutil: memory_rss_mb={memory_mb}, cpu_percentage: {' '.join(cpu_percentage)}"
    return msg

  def __call__(self, phase, info):
    now = time.perf_counter_ns()
    if phase == "start":
      self.start_ns = now

    elif phase == "stop" and self.start_ns is not None:
      duration = now - self.start_ns
      self.total_ns += duration
      self.start_ns = None

      # log stats every minute
      if now > self.last_log_ns + 60_000_000_000:
        stats = gc.get_stats()

        gc_time = self.total_ns*1e-9
        time_period = (now - self.last_log_ns)*1e-9

        msgs = [f'GC: gc_time={gc_time:.6e}s/{time_period:.6e}s, gen_info=']
        for stat in stats:
          msgs.append(
            f"[{stat['collections']} {stat['collected']} {stat['uncollectable']}]")

        gc_log = ' '.join(msgs)
        psutil_log = self.get_psutil_line()

        # these two consecutive log lines can be used to asses logging health.
        # If the second line's timestamp is considerably larger than the first's
        # then logging is overloaded: handlers are synchronized and publishers
        # need to wait for the others to finish in order to have their logs consumed.
        LOGGER.info(gc_log)
        LOGGER.info(psutil_log)

        self.total_ns = 0
        self.last_log_ns = now

INSTANCE = GCCallback()