# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
#  Initialization for Arcviz
#  Start Threads, etc.
#
# After upgrading Django we may be able to use AppConfig.ready instead:
# https://docs.djangoproject.com/en/dev/ref/applications/#django.apps.AppConfig.ready
#

from datetime import timedelta
import datetime
import logging
import tempfile
import shutil
from zoneinfo import ZoneInfo
from arcwebbase import util
from django.db import models
from apps.models import UserProfile
from usage.events_upload import collect_and_send_usage_events
from django.conf import settings
from django.core.cache import cache
from apps.models_shortener import ShortUrl
from django.db.models import Count, Q, Exists, OuterRef
from reports.views_settings import initialize_settings, setting_get_single
from reports.models import Report, ReportUsage
from searches.models import Search
from django.db import close_old_connections
from sqlrun.views_cache_statistic import CacheStatistics
from vizlib.querylog import ActionType, ComponentType
from rbac.permcache import RBAC
from util.ff_flipper import FeatureFlipper
from reports.notification_center import NotificationCenter
from jobs.models import JobSchedule, JobLog, JobTarget, STATUS_RUNNING, STATUS_QUEUED
from usage.views import load_initial_usage_data, handle_continuous_usage_data
from datasets.models import DataSet, DatasetSegment, Segment
from apps.views_snapshot import get_snapshot_temp_dir_path

# Logging handling for allowing multiple processes to log to a single stream.  This is
# picked up from the python logging cookbook:
# https://docs.python.org/2/howto/logging-cookbook.html#sending-and-receiving-logging-events-across-a-network

import os
import pickle
import logging.handlers
import socketserver
import struct
import json
import threading
import time

from usage.events_tracker import eventstracker_initial_load
from django.utils import timezone
from importlib import import_module

LOGGER = logging.getLogger("arcweb")

def setup_gc_monitor():
  if os.getenv('GC_MONITOR_ENABLED', 'no') != 'yes':
    LOGGER.info('GC: monitor NOT enabled')
    return
  from arcwebbase import gcmonitor
  LOGGER.info('GC: monitor enabled')
  gcmonitor.INSTANCE.register()

class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    """Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack('>L', chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = self.unPickle(chunk)
            record = logging.makeLogRecord(obj)
            self.handleLogRecord(record)

    def unPickle(self, data):
        return pickle.loads(data)

    def handleLogRecord(self, record):
        # if a name is specified, we use the named logger rather than the one
        # implied by the record.
        if self.server.logname is not None:
            name = self.server.logname
        else:
            name = record.name
        logger = logging.getLogger(name)
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        logger.handle(record)

class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """
    Simple TCP socket-based logging receiver
    """

    allow_reuse_address = 1

    def __init__(self, host='localhost',
                 port=logging.handlers.DEFAULT_TCP_LOGGING_PORT,
                 handler=LogRecordStreamHandler):
        socketserver.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.abort = 0
        self.timeout = 1
        self.logname = None

    def serve_until_stopped(self):
        import select
        abort = 0
        while not abort:
            rd, wr, ex = select.select([self.socket.fileno()],
                                       [], [],
                                       self.timeout)
            if rd:
                self.handle_request()
            abort = self.abort

class AuditFormatter(logging.Formatter):

  def format(self, record):
    return json.dumps(record.audit_record)

class _AuditPreprocessAndFilter(logging.Filter):
  # Here we pick up the app load and do the accounting for report counts/user profile
  # Since we previously didn't log this action to the audit log, we filter it out
  # for now.
  def filter(self, record):
    audit = record.audit_record
    if audit.get('component_type') == ComponentType.APP and \
       audit.get('component_id') == audit.get('app') and \
       audit.get('query_state') == 0 and \
       audit.get('action') == ActionType.READ:

      try:
        reportid = int(audit.get('app'))
      except ValueError:
        return 0

      appgroup_id = None
      if audit.get('appgroup'):
        try:
          appgroup_id = int(audit.get('appgroup'))
        except:
          pass

      close_old_connections()
      ReportUsage.update_reports_count(reportid)

      if appgroup_id:
        # Note that we are using negative numbers to indicate appgroups
        # Positive numbers indicate dashboards & visuals
        UserProfile.update_last_viewed(audit.get('username'), -appgroup_id)
        ReportUsage.update_reports_count(-appgroup_id)
      else:
        UserProfile.update_last_viewed(audit.get('username'), reportid)

      return 0

    return 1

# All log messages come through this filter -- either when sent by the worker
# processes or when handled by the master process file handler.
class GlobalFilter(logging.Filter):

  # Here we add the process ID to the log record
  def filter(self, record):
    if not hasattr(record, 'persistent_id'):
      record.persistent_id = settings.PERSISTENT_PROCESS_ID
    return 1

def setup_bulk_update():
  interval = getattr(settings, 'BULK_UPDATE_INTERVAL', 60)
  models = [UserProfile, ReportUsage, Search, CacheStatistics, ShortUrl]

  def bulk_update():
    sleep_duration = 0
    prev_count = 0
    idle_threshold = getattr(settings, 'BULK_UPDATE_IDLE_THRESHOLD', 6)
    while True:
      sleep_interval = min(interval, 10)
      time.sleep(sleep_interval)
      sleep_duration += sleep_interval
      curr_count = sum(ReportUsage.reports_count.values())
      if sleep_duration >= interval or \
         curr_count - prev_count < idle_threshold:

        # LOGGER.info("writing report usage changes sleep_duration:%s curr_count:%s prev_count:%s idle_threshold:%s",
        #     sleep_duration, curr_count, prev_count, idle_threshold)

        close_old_connections()
        for m in models:
          try:
            m.commit_changes()
          except:
            LOGGER.exception('cannot commit %s model changes' % m.__name__)
        sleep_duration = 0
        prev_count = 0
      else:
        prev_count = curr_count

      try:
        handle_continuous_usage_data()
      except:
        LOGGER.exception('cannot update usage data')

  try:
    load_initial_usage_data()
  except:
    LOGGER.exception('cannot load initial usage data')

  thread = threading.Thread(target=bulk_update)
  thread.daemon = True
  thread.start()

def session_cleanup_and_repeat():
  cleanup_interval = getattr(settings, 'SESSION_CLEANUP_INTERVAL')
  if cleanup_interval <= 0:
    return

  def cleanup_sessions():
    engine = import_module(settings.SESSION_ENGINE)
    while True:
      time.sleep(cleanup_interval)
      try:
        engine.SessionStore.clear_expired()
      except Exception as e:
        LOGGER.exception('Failed to sessions: %s' % str(e))

  # clear out session if authentication method has system that syncs users groups in authentication method
  has_group_sync = ('arcwebbase.backends.PamBackend' in settings.AUTHENTICATION_BACKENDS or \
    'arcweb.arcwebbase.basebackends.VizBaseRemoteBackend' in settings.AUTHENTICATION_BACKENDS or \
    'arcweb.arcwebbase.backends.Saml2Backend' in settings.AUTHENTICATION_BACKENDS or \
    'arcweb.arcwebbase.basebackends.VizBaseSaml2Backend' in settings.AUTHENTICATION_BACKENDS or \
    'arcwebbase.auth.saml.backend.VizSaml2Backend' in settings.AUTHENTICATION_BACKENDS)

  if has_group_sync:
    try:
      LOGGER.info('deleting all sessions ...')
      util.delete_all_sessions()
      LOGGER.info('deleting sessions done.')
    except:
      LOGGER.exception('error while deleting sessions')


  thread = threading.Thread(target=cleanup_sessions, name='THREAD-Session_Cleanup')
  thread.daemon = True
  thread.start()

def _remove_orphan_visuals_and_filters(run_once=False):
  while True:
    time.sleep(getattr(settings, 'BULK_REMOVE_ORPHAN_REPORTS', 24*60*60))
    try:
      a_day_before = datetime.datetime.now(ZoneInfo('UTC')) - timedelta(days=1)
      # Note:
      # A visual is orphan if the report data doesn't contains the dashboard_id.
      # Usually happens when somebody starts to creating a dashboard, but leaves the page
      # without saving. EXTRA: Only non shared visuals should be deleted.
      # When a visual is a shared one, the report data doesn't contain the dashboard_id,
      # but has a list of shared_visual_dashboards. If it is empty list - it means,
      # it was added to a dashboard which wasn't saved -> no item in the list.
      # (If it is null, it is not shared visual).

      orphans = Report.objects.exclude(report_type='dashboard') \
                              .exclude(report_data__contains='dashboard_id') \
                              .filter(Q(shared_visual_dashboards__isnull=True)) \
                              .filter(create_time__lte=a_day_before)
      ids = list(orphans.values_list('id', flat=True))
      if len(ids):
        orphans.delete()
        LOGGER.info('Orphan reports (ids: %s) were successfully removed'  % ids)
    except Exception as e:
      LOGGER.exception('Failed to remove orphan reports: %s' % str(e))
    if run_once:
      break


def setup_clean_orphan_visuals_and_filters():
  # When a user starts to create a dashboard and leaves the page the previously created visuals' and filters'
  # data, will stay in the database. This removes it!

  thread = threading.Thread(target=_remove_orphan_visuals_and_filters, name='THREAD-Remove_Orphan_Visuals_and_Filters')
  thread.daemon = True
  thread.start()

def setup_clean_notification():
  # Remove all notification but the last 50 every 24 hours
  # This should be remove when implementing the expiry date of notifications in notification center V3
  def remove_old_notification():
    while True:
      time.sleep(getattr(settings, 'BULK_REMOVE_READ_NOTIFICATION_INTERVAL', 24*60*60))
      NotificationCenter().clean_notifications()

  thread = threading.Thread(target=remove_old_notification)
  thread.daemon = True
  thread.start()

def aux_remove_unnamed_versions():
      deletion_interval_sec = float(setting_get_single('unnamed_versions_deletion_interval')) * 60 * 60 * 24
      time.sleep(deletion_interval_sec)
      LOGGER.info('Cleaning unnamed versions')

      retention_seconds = float(setting_get_single('unnamed_versions_max_retention')) * 60 * 60 * 24 
      retention_time_from_now = timezone.now() - timezone.timedelta(seconds=retention_seconds)

      DataSet.objects.filter(
        is_named_version=False,
        is_active_version=False,
        update_time__lt=retention_time_from_now
      ).delete()

      DatasetSegment.objects.filter(
        ~Exists(
          DataSet.objects.filter(
            version_id=OuterRef('dataset_version_id')
          )
        )
      ).delete()

      Segment.objects.annotate(
        dataset_count=Count('dataset')
      ).filter(dataset_count=0).delete()

def setup_clean_unnamed_versions():
  def remove_unnamed_versions():
    while True:
      aux_remove_unnamed_versions()

  thread = threading.Thread(target=remove_unnamed_versions, name='UnnamedVersionCleanup')
  thread.daemon = True
  thread.start()

def log_job_table_stats():
  LOGGER.info('job log jobs_joblog rows: %d', JobLog.objects.count())
  LOGGER.info('job log jobs_jobschedule rows: %d', JobSchedule.objects.count())
  LOGGER.info('job log jobs_jobtarget rows: %d', JobTarget.objects.count())

def setup_clean_old_joblogs():
  def remove_old_joblogs():
    while True:
      # bulk_max_interval is days
      max_interval_sec = float(setting_get_single('bulk_max_interval')) * 60 * 60 * 24

      LOGGER.info('job log cleanup will run in %s seconds', max_interval_sec)
      time.sleep(max_interval_sec)

      max_entries = int(setting_get_single('bulk_max_entries'))
      retention_days = float(setting_get_single('bulk_max_retention'))

      do_remove_old_joblogs(max_entries, retention_days)

  thread = threading.Thread(target=remove_old_joblogs, name='OldJobCleanup')
  thread.daemon = True
  thread.start()

def do_remove_old_joblogs(max_entries: int, retention_days: float):
  LOGGER.info('job log cleanup start')
  log_job_table_stats()
  retention_seconds = retention_days * 60 * 60 * 24

  retention_time_from_now = timezone.now() - timezone.timedelta(seconds=retention_seconds)

  # filter for deletion candidates (logs that are neither RUNNING nor QUEUED):
  not_running_and_not_queued = ~Q(status=STATUS_RUNNING) & ~Q(status=STATUS_QUEUED)

  # image is needed because it is a file field so Django does some checks when deleting:
  base_joblog_query = JobLog.objects.only('id', 'image').filter(not_running_and_not_queued)

  # Query by retention days
  base_joblog_query.filter(end_time__lt=retention_time_from_now) \
    .delete()

  # Newest logs
  newest_jobs = base_joblog_query.order_by('-end_time')[:max_entries]
  newest_job_ids = [joblog.id for joblog in newest_jobs]

  # delete everything except the newest logs:
  base_joblog_query.exclude(id__in=newest_job_ids) \
    .delete()

  # delete ad-hoc job schedules that do not have any logs left:
  # (related JobTargets will be automatically deleted by Django)
  ad_hoc_schedules_with_no_logs = JobSchedule.objects.only('id')\
    .filter(schedule_group_id__isnull=True) \
    .exclude(last_status__isnull=True) \
    .exclude(last_status=STATUS_QUEUED) \
    .annotate(logs=Count('joblog'))\
    .filter(logs=0)

  ad_hoc_schedules_with_no_logs.delete()

  LOGGER.info('job log cleanup end')
  log_job_table_stats()

def remove_old_snapshot_dirs(run_once=False):
  while True:
    time.sleep(getattr(settings, 'BULK_REMOVE_OLD_SNAPSHOT_DIRS_INTERVAL', 1*60*60))
    LOGGER.info('Snapshot tmp directories cleanup start')
    one_hour_ago = datetime.datetime.now() - timedelta(hours=1)

    snapshot_files_temp_dir_path = get_snapshot_temp_dir_path(settings.SNAPSHOT_TMP_DIR_PREFIX)

    for dirname in os.listdir(snapshot_files_temp_dir_path):
      dir_path = os.path.join(snapshot_files_temp_dir_path, dirname)

      if not os.path.isdir(dir_path):
        continue

      modified_time = datetime.datetime.fromtimestamp(os.path.getmtime(dir_path))
      if modified_time < one_hour_ago:
        try:
          shutil.rmtree(dir_path, ignore_errors=True)
        except Exception as e:
          LOGGER.warning(f'cannot remove temporary folder {dir_path}, {str(e)}')
    if run_once:
      break

def cleanup_snapshot_tmp_dirs(run_once=False):
  thread = threading.Thread(
    target=remove_old_snapshot_dirs,
    kwargs={'run_once': run_once},
    name="snapshot_tmp_dirs_cleanup_thread"
  )
  thread.daemon = True
  thread.start()
  

def arcviz_initialize():
  """
  Initialization for Arcviz
  """

  # Do a cache lookup here so we don't race to cache initialize in the worker threads
  cache.get('Cache initialize')

  logging_port = settings.ARCVIZ_LOGGING_PORT
  logging_host = getattr(settings, 'ARCVIZ_LOGGING_HOST', '127.0.0.1')

  LOGGER.info('The application logs are displayed in time zone: %s' % settings.TIME_ZONE)

  # Log all db backend queries if specified
  def _setup_db_logging():
    if not settings.LOG_ORM_QUERIES:
      return

    from django.db import connection
    l = logging.getLogger('django.db.backends')
    l.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    l.addHandler(ch)

  # Setup the audit logger / activity log
  def _setup_audit_logger():

    logger = logging.getLogger('audit')
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    if settings.IS_MASTER or not logging_port:
      logger.addFilter(_AuditPreprocessAndFilter(name='audit'))
      if not settings.AUDIT_LOGFILE:
        fh = None
      else:
        fh = logging.handlers.RotatingFileHandler(settings.AUDIT_LOGFILE,
                                                  maxBytes=settings.AUDIT_FILE_MAXBYTES,
                                                  backupCount=settings.AUDIT_FILE_MAXCOUNT)

    else:
      fh = logging.handlers.SocketHandler(host=logging_host, port=logging_port)

    if fh:
      fh.setLevel(logging.DEBUG)
      fh.setFormatter(AuditFormatter())
      logger.addHandler(fh)

    settings.AUDIT_LOGGER = logger

  # Setup the main arcviz logger
  def _setup_arcviz_logger():
    if not settings.ARCVIZ_LOGFILE:
      return

    logger = logging.getLogger('arcweb')
    logger.setLevel(getattr(logging, settings.ARCVIZ_LOGLEVEL, logging.DEBUG))
    logger.propagate = False

    if settings.IS_MASTER or not logging_port:
      fh = logging.handlers.RotatingFileHandler(settings.ARCVIZ_LOGFILE,
                                                maxBytes=settings.ARCVIZ_FILE_MAXBYTES,
                                                backupCount=settings.ARCVIZ_FILE_MAXCOUNT)

    else:
      fh = logging.handlers.SocketHandler(host=logging_host, port=logging_port)

    from reports.handlers import NotificationCenterHandler
    notification_center_handler = NotificationCenterHandler()

    formatter = logging.Formatter(settings.ARCFMT_PID)
    fh.setFormatter(formatter)
    fh.addFilter(GlobalFilter(name='globalfilter'))
    logger.addHandler(fh)
    logger.addHandler(NotificationCenterHandler)

    # Now we can pick up the buffered pre_init log records
    bh = next((h for h in logger.handlers if h.get_name() == 'pre_init_buffer'), None)
    if bh:
      logger.removeHandler(bh)
      for r in bh.buffer:
        logger.handle(r)
      bh.close()

    # Set up the additional loggers to log the same handlers as the main arcweb logger
    for entry in settings.ADDITIONAL_LOGGERS:
      if type(entry) == type(''):
        name = entry
        level = logger.level
      elif type(entry) in [type(()), type([])] and len(entry) == 2:
        name = entry[0]
        level = getattr(logging, entry[1], logging.DEBUG)
      else:
        logger.debug("Invalid entry in ADDITIONAL_LOGGERS: %s" % (str(entry)))
        continue
      al = logging.getLogger(name)
      al.setLevel(level)
      al.propagate = False
      logger.debug("Set logger %s to level %s" % (name, level))
      for handler in logger.handlers:
        al.addHandler(handler)

  # On the master process, start up the log record receiver from the other work processes
  if settings.IS_MASTER and logging_port:
    try:
      tcpserver = LogRecordSocketReceiver(host=logging_host, port=logging_port)
      logth = threading.Thread(target=tcpserver.serve_until_stopped, name="Log server")
      logth.daemon = True
      logth.start()
    except Exception as e:
      if not ('Address already in use' in str(e) and os.environ.get('RUN_MAIN') == 'true'):
        raise

  _setup_audit_logger()
  _setup_arcviz_logger()

  # Ensure default site settings are present
  initialize_settings(getattr(settings, 'SITE_SETTING_DEFAULTS', {}))

  # Start the rbac poller
  RBAC.pollstart(polltime=getattr(settings, 'RBAC_POLL_SECONDS', 300))

  # The tasks started after this point only run in the master process
  if not settings.IS_MASTER:
    return

  # Setup the kerberos ticket renewer if needed
  if getattr(settings, 'KERBEROS_PRINCIPAL', None):
    from dataconn.thriftconn import kt_renewer_from_settings
    kt_renewer_from_settings(settings, logging.getLogger('arcweb'))

  # Only import sentry if if needed
  if getattr(settings, 'SENTRY', None):
    from dataconn import sentryif
    sentryif.startSentry(settings.SENTRY)

  # Setup the view refresher
  from jobs.views_refresh import setup_view_refresher
  setup_view_refresher()

  # Setup the cloning related dashboard id lists
  if FeatureFlipper().get_flipper_value('enable_clone_info_badge'):
    from reports.utils import setup_clone_related_report_ids
    setup_clone_related_report_ids()

  # Start scheduler checking if enabled
  from jobs.view_schedule_checker import ScheduleChecker
  ScheduleChecker()

  # start bulk update thread for report usage and last viewd dashboard by users
  setup_bulk_update()

  setup_clean_notification()

  setup_clean_old_joblogs()

  eventstracker_initial_load()

  collect_and_send_usage_events()

  _setup_db_logging()

  setup_clean_orphan_visuals_and_filters()

  session_cleanup_and_repeat()

  setup_gc_monitor()
  setup_clean_unnamed_versions()

  cleanup_snapshot_tmp_dirs()
