# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Middleware is a framework of hooks into Django's request/response processing.
It's a system for globally altering Django's input or output.
"""

from arcwebbase import access_info, middleware_storage
import logging
import time
import datetime
from time import sleep
import threading
import json
import os
from urllib.parse import urlparse
from threading import Lock

import re
import jwt
import requests

from arcwebbase.util import is_health_check_agent
from arcwebbase.validator import escapeCarets
from cryptography.hazmat.primitives import serialization
from django.conf import settings
from django.contrib.auth import login, logout
from django.contrib.auth.models import User
from django.core.cache import cache
from django.core.exceptions import MiddlewareNotUsed
from django.http import Http404, HttpResponseForbidden
from django.http import HttpRequest, HttpResponse
from django.middleware.csrf import get_token
from django.core.exceptions import SuspiciousFileOperation
from rbac.permdefs import PermissionException
from django.core.exceptions import PermissionDenied
from django.core.exceptions import RequestDataTooBig

from django.contrib.auth.middleware import RemoteUserMiddleware

from util.ff_url_flippers import UrlFlipper
import util.perfcounters as perfcounters
import vizlib.util
from vizlib.arcprof import ArcProf

LOGGER = logging.getLogger("arcweb")

STATS = None

requestsConcurrencyOverflow = perfcounters.APPROX_COUNTERS.get(
    'requests.concurrency_overflow')

responseTotalSize = perfcounters.APPROX_COUNTERS.get('response.totalsize')
responseSize = perfcounters.TimeSeriesCounter('response.size',
    timePeriodSeconds=60,
    timePeriodRetentionSeconds=60*60)

def getRequestStatsData():
  if not STATS:
    return {}

  return STATS.fetch()

def is_request_targets_static_endpoint(request):
  return request.path.startswith("/arc/static") or request.path.startswith(settings.STATIC_URL)

class RequestStats(object):
  def __init__(self):
    self.lock = threading.Lock()
    self.countsByMinute = perfcounters.TimeSeriesCounter('http.requests',
        timePeriodSeconds=60,
        timePeriodRetentionSeconds=60*60)
    self.active = 0
    self.maxActiveLevel = 200
    self.concurrentRequests = [0] * self.maxActiveLevel
    self.concurrentRequestsTimes = [0] * self.maxActiveLevel
    self.start_time = datetime.datetime.now()
    self.httpRequestsHistogram = perfcounters.APPROX_COUNTERS.getHistogram(name='http_requests',
        buckets=[1, 5, 10, 30, 60, 61])

    global STATS
    if not STATS:
      STATS = self

    self._httpRequestsTracker = vizlib.util.RequestStatsTracker("http")

  def fetch(self):
    t = datetime.datetime.now()
    uptime = str(t.replace(microsecond=0) - self.start_time.replace(microsecond=0))

    d = {
      'start_time': self.start_time.strftime("%Y-%m-%d %H:%M:%S"),
      'current_time': t.strftime("%Y-%m-%d %H:%M:%S"),
      'uptime': uptime,
      'counters': {},
      'histograms': {},
      'requests': {
        'queues': {}
      },
    }

    with self.lock:
      d.update({
        'concurrentRequests': self.concurrentRequests[:],
        'concurrentRequestsTimes': self.concurrentRequestsTimes[:],
      })

    maxIdx = 0
    for idx, c in enumerate(d['concurrentRequests']):
      if c > 0:
        maxIdx = idx

    d['concurrentRequests'] = d['concurrentRequests'][1:maxIdx+1]
    d['concurrentRequestsTimes'] = d['concurrentRequestsTimes'][1:maxIdx+1]

    return d

  def requestTracker(self, requestData, profile):
    return self._httpRequestsTracker.tracker(requestData, profile)

  def startRequest(self):
    self.countsByMinute += 1

    startTime = time.time()

    with self.lock:
      self.active += 1
      active = self.active
      if active < self.maxActiveLevel:
        self.concurrentRequests[active] += 1
      else:
        requestsConcurrencyOverflow += 1

    return (active, startTime)

  def stopRequest(self, h):
    (active, startTime) = h
    runtime = time.time() - startTime

    bucket = self.httpRequestsHistogram.getBucket(runtime)
    with self.lock:
      self.active -= 1
      self.concurrentRequestsTimes[active] += runtime
      self.httpRequestsHistogram.addToBucket(bucket)

class ProfileMiddleware(object):
  def __init__(self, get_response):
    self.get_response = get_response
    self.stats = RequestStats()
    self._idAllocator = vizlib.util.IdAllocator()

  def __call__(self, request):
    global responseTotalSize
    global responseSize

    id = self._idAllocator.allocateId()

    requestData = json.dumps({
      'id': id,
      'query': f'{request.method} {request.path}'
    })

    profile = ArcProf("http", username=request.user.username)
    with self.stats.requestTracker(requestData, profile):
      h = self.stats.startRequest()
      try:
        response = self.get_response(request)

        if not response.streaming:
          l = len(response.content)
          responseTotalSize += l
          responseSize += l

        return response

      finally:
        self.stats.stopRequest(h)

# https://stackoverflow.com/questions/43496658/django-catch-requestdatatoobig-exception
# If the request body is too big then request.body will throw a
# RequestDataTooBig exception.
class RequestDataTooBigMiddleware(object):
    def __init__(self, get_response):
      self.get_response = get_response

    def __call__(self, request):
      response = self.process_request(request)
      return response


    def process_request(self, request):
        try:
            request.body
        except RequestDataTooBig as e:
          # This error can be swallowed if another (for example CsrfViewMiddleware) middleware runs into it.
          # Here is how: .pyenv/versions/v3.9.0/lib/python3.9/site-packages/django/core/handlers/exception.py: convert_exception_to_response
          #  def convert_exception_to_response(get_response):
          #     """
          #    ...
          #     All exceptions will be converted.
          #     ...
          #     This decorator is automatically applied to all middleware to ensure that
          #     no middleware leaks an exception and that the next middleware in the stack
          #     can rely on getting a response instead of an exception.
          #     """
          # Using the CsrfViewMiddleware this error in the que never ends up in the ExceptionHandler's process_exception cycle.
          # We need to handle it separately to able to tell what is the current limitation and the "how to handle instruction" to the user!
            limit = getattr(settings, 'DATA_UPLOAD_MAX_MEMORY_SIZE')
            helper_text = f"To modify the current limit({limit} bytes) set the desired DATA_UPLOAD_MAX_MEMORY_SIZE value in the Advanced Setting or contact your system administrator!"
            return HttpResponse(f"{str(e)}: {helper_text}", status=413)

        return  self.get_response(request)

# Middleware class to return response on permission error
class ExceptionHandler(object):
  def __init__(self, get_response):
    self.get_response = get_response

  def __call__(self, request):
    response = self.get_response(request)
    return response

  def process_exception(self, request, e):
    if isinstance(e, PermissionException):
      return HttpResponse(str(e), status=500)
    elif isinstance(e, SuspiciousFileOperation):
      LOGGER.exception(e)
      return HttpResponse('invalid path')
    elif isinstance(e, PermissionDenied):
      # This is raised by the SAML authentication backend if the user is not None
      return HttpResponse('Forbidden', status=403)
    LOGGER.exception("Exception on request processing")
    if getattr(settings, 'DEBUG', False):
      raise e
    if isinstance(e, Http404):
      raise e
    return HttpResponse(escapeCarets(str(e)), status=500)

IGNORE_URLS = set([
  "/",
  "/arc/apps",
  "/arc/apps/",
  "/arc/apps/login?next=/arc/apps/",
])

class StoreRequestInfoMiddleware(object):
  def __init__(self, get_response):
    self.get_response = get_response

  def __call__(self, request):
    if 'Host' in request.headers:
      middleware_storage.HOSTNAME = request.headers['Host']
    response = self.get_response(request)
    return response

class StoreAccessInfoMiddleware(object):
  def __init__(self, get_response):
    self.get_response = get_response

  def __call__(self, request):
    try:
      request_access_info = access_info.create_from_request(request)
      access_info.store(request_access_info)

      response = self.get_response(request)

      return response
    finally:
      access_info.clear()


# Middleware to log requests
class AccessLogMiddleware(object):
  def __init__(self, get_response):
    self.get_response = get_response

  def __call__(self, request):
    self.process_request(request)
    response = self.get_response(request)
    response = self.process_response(request, response)
    return response

  def process_request(self, request):
    request._arc_start_time = time.time()
    return None

  def process_response(self, request, response):
    response['Server'] = 'Arcviz'

    httpUserAgent = request.META.get('HTTP_USER_AGENT')

    if settings.ACCESS_LOG:

      # Skip logging frequent CML pings
      if settings.SKIP_CML_LOGGING and \
         not request.user.username and \
         is_health_check_agent(httpUserAgent) and \
         request.get_full_path() in IGNORE_URLS:
        return response

      now = time.time()
      LOGGER.info('%s %s "%s %s %s" %s %s %.3f "%s" "%s"',
         request.META.get('REMOTE_ADDR'),
         request.user.username or "-",
         request.method,
         request.get_full_path(),
         request.META.get('SERVER_PROTOCOL'),
         response.status_code,
         len(response.content) if not response.streaming else "-",
         now - getattr(request, '_arc_start_time', now),
         request.META.get('HTTP_REFERER', '-'),
         httpUserAgent)

    return response

class ArcvizSecureMiddleware:
  """
  Middleware to Secure Arcviz and optionally allow anonymous access

  This middleware allows securing arcviz for four user classes
  based on the URL's accessed and http method used.
  In addition it optionally allows unauthenticated sessions to automatically be
  logged in as a default user, allowing anonymous access

  The four user classes are:
    all users - the "any_list" allows access for all sessions, both
     unauthenticated and authenticated sessions.  This allows unauthenticated
     sessions to access the login page.
    authenticated users - the "authenticated_list" allows access to "normal"
      users who are authenticated
    trustedauth users - the "trusteduser_list" allows access for users who have
      been authenticated using the trusedtedauth backend.  This allows these
      sessions to be limited to the visuals
    default users - If the session does not match any of the above lists, the
      "default_user_list" allows access by automatically logging in a "default"
      user - defined in the settings.

  All the lists are comprised of pairs of regular expressions that match
  the URL being accessed and the http method being used.  This is a bit complex
  to configure, but gives maximum flexibility and control.  To make it easier
  there is a base_list that contains urls that are needed for basic read access
  to view visuals
  """

  secure_config = getattr(settings, "ARCVIZ_SECURE", {})

  # This is the default list of allowed URLs
  base_list = ((r"""
                   /arc/reports/columns.*$ |
                   /arc/reports/dateranges$ |
                   /arc/reports/customcss(/\d+)?$ |
                   /arc/reports/customcssraw(/\d+)?$ |
                   /arc/reports/customjsraw(/\d+)?$ |
                   /arc/reports/colorpalette(/\d+)?$ |
                   /arc/reports/staticasset(/raw)?(/\d+)?$ |
                   /arc/apps/customcolors/\d+$
                   /arc/apps/customstyles/\d+$ |
                   /arc/apps/staticassets/\d+$ |
                   /arc/datasets/segment(/\d+)?$ |
                   /arc/datasets/event(/\d+)?$ |
                   /arc/datasets/filterassociation(\?.+)?$ |
                   /arc/datasets/dataset(/\d+)?$ |
                   /arc/sqlrun/asyncstatus(\?.+)?$ |
                   /arc/sqlrun/asyncresults(\?.+)?$ |
                   /arc/sqlrun/asynccancel(\?.+)?$ |
                   /arc/logout$""", ["GET"]),
               (r"""
                   /arc/sqlrun/jsonselect_parallel$ |
                   /arc/sqlrun/jsonselect_parallel_async$ |
                   /arc/sqlrun/jsonselect$""", ["POST"]),
               )

  # Compile all the regex's and create the lists
  all_list = [(re.compile(p), m) for p, m in secure_config.get("all_list", [])]

  authenticated_list = [(re.compile(p, re.X), m) for p, m in
                        secure_config.get("authenticated_list", [])]

  trustedauth_list = [(re.compile(p, re.X), m) for p, m in
                      secure_config.get("trustedauth_list", [])]

  default_user_list = [(re.compile(p, re.X), m) for p, m in
                       secure_config.get("default_user_list", [])]

  if secure_config.get("include_base_urls", True):
    base_list = [(re.compile(p, re.X), m) for p, m in base_list]
    authenticated_list.extend(base_list)
    trustedauth_list.extend(base_list)
    default_user_list.extend(base_list)

  force_embed_urls = secure_config.get("force_embed_urls",
                      r"""/arc/apps/app/\d+(\?.+)?$ |
                          /arc/trustedauth/trusted/\S+/app(group)?/\d+(\?.+)?$ |
                          /arc/trustedauth/trusted/\S+/redirect$ |
                          /arc/jobs/adminlogs
                      """)

  force_embed_re = re.compile(force_embed_urls, re.X)

  def __init__(self, get_response):
    if not self.secure_config.get('enabled', False):
      raise MiddlewareNotUsed
    self.get_response = get_response

  def __call__(self, request):
    response = self.process_request(request)
    if response is None:
      response = self.get_response(request)
    return response

  @classmethod
  def _check_embed(cls, request):
    """
    If the "force_embed" setting is True and the request matches the URLs
    in "force_embed_urls", modify the request to include "force_embed=true"
    :param request: a Django request object
    :return:
    """
    if cls.secure_config.get('force_embed', False) and \
        cls.force_embed_re.match(request.path):
      request.GET._mutable = True
      request.GET['embed'] = "true"

  @classmethod
  def _check_url(cls, request, rulelist):
    """
    Check a request against a rule list
    :param request: a Django request object
    :param rulelist: List of rules to check against
    :return:
    """
    for r, m in rulelist:
      if request.method in m:
        if r.match(request.path) is not None:
          return True
    return False

  @classmethod
  def _login_default_user(cls, request, default_user):
    """
    login the default user.  This creates an authenticated session.
    :param request:
    :param default_user:
    :return:
    """
    default_user.backend = 'django.contrib.auth.backends.ModelBackend'
    login(request, default_user)
    # Note: cache the csrf token and reuse for simultaneous unauthenticated
    # requests as in the case of multiple iframes. See django/middleware/csrf.py.
    key = "csrftoken_%s_%s_%s_%s" % (default_user.username,
                                     request.META.get('REMOTE_HOST', "host"),
                                     request.META.get('REMOTE_ADDR', "addr"),
                                     request.META.get('HTTP_USER_AGENT', "agent"),
                                     )

    csrftoken = cache.get(key)
    if not csrftoken:
      newtoken = get_token(request)
      # use add() instead of set() to ensure a single csrf token
      cache.add(key, newtoken, 30)
      sleep(1)  # allow cache to settle
      # get the token from the cache.
      # might not be "newtoken" if another request set cache first.
      csrftoken = cache.get(key)
    request.META.update({
      "CSRF_COOKIE_USED": True,
      "CSRF_COOKIE": csrftoken,
    })
    request.session.set_expiry(cls.secure_config.get('session_expiry', 300))

  def process_request(self, request):
    """
    See the Django Docs:
    http://django.readthedocs.io/en/1.6.x/topics/http/middleware.html
    """
    if not self.secure_config.get('enabled', False):
      return None

    force_default_user = self.secure_config.get("force_default_user", True)

    # Step 1: check all_list
    if self._check_url(request, self.all_list):
      LOGGER.debug("Secure access (any user) %s: %s, %s" % (
        request.user.username, request.method, request.path))
      self._check_embed(request)
      return None

    default_username = self.secure_config.get('default_user', None)

    if request.user.is_authenticated and not force_default_user:
      # Step 2 check for TrustedUser
      if request.session._session.get('_auth_user_backend', '') == \
                               'trustedauth.backends.TrustedAuthBackend' and \
                               request.user.username != default_username:
        if self._check_url(request, self.trustedauth_list):
          LOGGER.debug("Secure access (trustedauth user) %s: %s, %s" % (
            request.user.username, request.method, request.path))
          self._check_embed(request)
          return None
      # Step 3: Check for regular authenticated user
      elif request.user.username != default_username:
        if self._check_url(request, self.authenticated_list):
          LOGGER.debug("Secure access (authenticated user) %s: %s, %s" % (
            request.user.username, request.method, request.path))
          return None

    if default_username is None:
      # No default user and not authenticated so Error
      LOGGER.warn("Secure access denied (no default user) %s, %s" % (
        request.method, request.path))
      return HttpResponseForbidden(self.secure_config.get('message', 'Forbidden'))

    # Step 4: handle default_user
    # Get default user
    default_user = User.objects.get(username=default_username)

    if request.user.is_authenticated:
      if request.user.username != default_user.username:
        # not the default user so logout
        logout(request)

    # Login default user if not already authenticated
    if not request.user.is_authenticated:
      # Login as default user
      if not default_user.is_active:
        LOGGER.warn("Secure access denied (default user is not active) %s, %s" % (
          request.method, request.path))
        return HttpResponseForbidden(self.secure_config.get('message', 'Forbidden'))
      self._login_default_user(request, default_user)

    # Check default_user_list:
    if request.user.is_authenticated and \
        self._check_url(request, self.default_user_list):
      LOGGER.debug("Secure access (default user) %s: %s, %s" % (
        request.user.username, request.method, request.path))
      self._check_embed(request)
      return None

    # request did not match any list so return 403
    LOGGER.warn("Secure access denied (default user) %s: %s, %s" % (
      request.user.username, request.method, request.path))
    return HttpResponseForbidden(self.secure_config.get('message', 'Forbidden'))

# Middleware to remove the Vary header
# needs to be above the Session Middleware in settings
class RemoveVaryMiddleware:
  def __init__(self, get_response):
    self.get_response = get_response

  def __call__(self, request):
    response = self.get_response(request)
    response = self.process_response(request, response)
    return response

  def process_response(self, request, response):
    if (is_request_targets_static_endpoint(request) or hasattr(response, 'remove_vary')) \
      and 'Vary' in response:
      del response['Vary']
    return response

CML_REMOTE_USER_HEADER = "HTTP_REMOTE_USER"

class CustomHeaderMiddleware(RemoteUserMiddleware):
  # The settings file may contain a header name (true for C1C)
  header = settings.HTTP_HEADERNAME_FOR_TRUSTED_AUTH

  def process_request(self, request):
    if os.getenv('CDV_SUB_MODE') == 'C1C':
      _, token = request.headers.get('Authorization').split(' ')
      username = ''
      public_key_pem = ''
      try:
        token_metadata = jwt.get_unverified_header(token)
        headers = {'kid': token_metadata.get('kid', {})}
        jwt_key_server_url = os.getenv('JWT_KEY_SERVER_URL', '')
        response = requests.get(jwt_key_server_url, headers=headers)
        jwk = json.loads(response.content)
        if jwk.get('keys'):
          public_key = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk["keys"][0])).public_key()
          public_key_pem = public_key.public_bytes(encoding=serialization.Encoding.PEM,
                                                   format=serialization.PublicFormat.SubjectPublicKeyInfo)
        # The public key is not yet provided, we will have to get it as
        # an HTTP request but the endpoint does not exist yet, it will make
        # the verify false obsolete.
        jwt_info = jwt.decode(
          token,
          public_key_pem,
          algorithms=['RS256'],
          issuer='Cloudera1',
          audience='viz',
        )
        username = jwt_info.get('user')
      except Exception as e:
        LOGGER.error(f'Failed to validate JWT token: {token}')
        LOGGER.exception(e)

      request.META['HTTP_X_C1C_ACTOR_USERNAME'] = username

    # If the user is already authenticated, then don't auth again
    # This lets a user login as a different user via /arc/apps/login and
    #   not be limited to the user's CDSW/CML login
    if request.user.is_authenticated:
      return

    if not settings.HTTP_HEADERNAME_FOR_TRUSTED_AUTH:
      # If the settings file does not set it, we have to dynamically look
      # at site settings before each attempt at auth
      from reports.views_settings import setting_get_single
      enabled = setting_get_single("enable_remote_user_auth")
      if enabled != "true":
        return

      self.header = CML_REMOTE_USER_HEADER

    if not self.header:
      return

    RemoteUserMiddleware.process_request(self, request)
    if os.getenv('CDV_SUB_MODE') == 'C1C':
      _, token = request.headers.get('Authorization').split(' ')
      if request.user and hasattr(request.user, 'userprofile'):
        if (request.user.userprofile.jwt_token != token):
          request.user.userprofile.jwt_token = token
          request.user.userprofile.save(update_fields=['jwt_token'])

class FlipperMiddleware(object):
  """Updating the cookies with the URL flippers
  that has been set through GET requests
  """

  def __init__(self, get_response):
    self.get_response = get_response

  def __call__(self, request):
    response = self.get_response(request)
    flippers = UrlFlipper.get_url_flippers(request)
    if flippers:
      response.set_cookie(
        'flippers',
        json.dumps(flippers),
        max_age=None,
        samesite=settings.SESSION_COOKIE_SAMESITE,
        secure=settings.SESSION_COOKIE_SECURE
      )

    return response

def iframe_embed_middleware(get_response):

  def middleware(request):
    response = get_response(request)
    if os.getenv('IS_EMBED') == 'true':
      # Temporary for all domains
      response['Content-Security-Policy'] = 'frame-ancestors *'
      response['Access-Control-Allow-Origin'] = '*'
    return response

  return middleware

class DynamicCSRFMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def add_trusted_origin(self, origin):
      if origin not in settings.CSRF_TRUSTED_ORIGINS:
        csrf_lock = Lock()

        if origin not in settings.CSRF_TRUSTED_ORIGINS:
          with csrf_lock:
            updated_origins = settings.CSRF_TRUSTED_ORIGINS + [origin]
            settings.CSRF_TRUSTED_ORIGINS = updated_origins

    def __call__(self, request: HttpRequest):
        if getattr(settings, "SKIP_DYNAMIC_CSRF_MIDDLEWARE", False):
          return self.get_response(request)
        
        origin = request.META.get('HTTP_ORIGIN')

        if origin:
          self.add_trusted_origin(origin)

        response = self.get_response(request)

        return response
