# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Template context processors. These provide global template variables
that are available to use in templates for substitution.
"""

from reports.constants import FINAL_SETTINGS
from reports.utils import replace_all_ai_api_keys_with_placeholder
from django.conf import settings
from django.contrib.staticfiles import finders
from util.ff_url_flippers import UrlFlipper
from reports.views_settings import get_settings
from apps.views_snapshot import chromium_pdf_available
from util.ff_flipper import FeatureFlipper
from apps.models import UserProfile
from dataconn.cml_dependencies import get_cmldata

import json
import os
import logging
LOGGER = logging.getLogger("arcweb")

def static_url(unused_request):
  """
  Return template variables available globally for static urls.
  """

  return {
    'arc_static': settings.STATIC_URL.rstrip("/")
  }

def arc_css_js_version(unused_request):
  """
  Define a common version number for all arcadia css/js resources
  This is rev'ed once per external release.
  """

  v = getattr(settings, 'STATIC_FILES_SERVED_VERSION', False)
  if v:
    return {
      'static_version': v,
      'favicon_version': 2,
    }

  from apps.utils import get_version
  v = get_version()
  return {
    'static_version': v[1],
    'favicon_version': 2
  }

def auth_backend(request):
  """
  Return auth backend type and whether change password is allowed.
  """
  authtype = request.session._session.get('_auth_user_backend', 'Unknown')
  allow_change_password = authtype in [ 'django.contrib.auth.backends.ModelBackend' ]

  if request.user and request.user.is_authenticated and request.user.username == settings.AUTOLOGIN_USERNAME:
    allow_change_password = False

  return {
    'auth_backend': authtype,
    'allow_change_password': allow_change_password
  }

def user_authentication(request):
  """
  Returns the user authentication details
  """
  user_authentication = ""
  is_default_authentication_mode = False
  if hasattr(request.user, 'userprofile'):
    is_default_authentication_mode = request.user.userprofile.authentication_source == UserProfile.AUTH_DEFAULT
    auth_dictionary = dict((auth_type, str) for auth_type, str in UserProfile.AUTHENTICATION_SOURCES)
    if request.user.userprofile.authentication_source == UserProfile.AUTH_SAML or request.user.userprofile.authentication_source == UserProfile.AUTH_REMOTE:
      user_authentication = 'SSO'
    else:
      user_authentication = auth_dictionary[request.user.userprofile.authentication_source]
  return {
      'user_authentication_mode': user_authentication,
      'is_default_authentication_mode': is_default_authentication_mode
  }

# These settings are coming from the django metadata database.
# Some are set up through site settings and some are set up
# at install time
def app_settings(request):
  if hasattr(request,"user") and request.user.is_superuser:
    app_settings_list = get_settings(True)
  else:
    app_settings_list = get_settings(False)
  app_settings_dict = dict((s['setting_key'], s['setting_value'])
                           for s in app_settings_list
                              if s['setting_key'].find('password') == -1)

  # Detect IE browser from request or if we request it from settings file
  user_agent = request.META.get('HTTP_USER_AGENT', '').lower()
  ie_detected = 'trident' in user_agent or 'msie' in user_agent

  final_settings = FINAL_SETTINGS.copy()

  final_settings.update(app_settings_dict)

  # Initialize dependent settings
  if final_settings['page_profile_enabled'] != 'true':
    final_settings['page_profile_appview_enabled'] = 'false'

  final_settings = replace_all_ai_api_keys_with_placeholder(final_settings)

  return {
    'app_settings': final_settings,
  }

def userprofile(request):
  if not request.user.is_authenticated:
    return {}
  return {
    'userprofile': request.user.userprofile.user_data_obj
  }

# These are settings that are a subset of the settings in the settings.py file
# that are needed to be passed on to the JS side through templates
# To call these, use settings.<settingname> in the python, and in a template,
# use arcviz_settings.<settingname>
# Note these are also delivered to the js environment so that one can also
# use arc.globals.arcviz_settings.<settingname>
def arcviz_settings(request):
  """
  Return a dict of the specified settings
  Makes it easy to add more settings to template context
  """
  sd = {}

  if request.user.is_authenticated:
    # User is authenticated
    settings_list = [
      'DATA_API_ENABLED',
      'CUSTOM_EMAIL_TEMPLATE',
      'ENFORCE_LICENSE',
      'ENABLE_ADVANCED_SETTINGS',
      'ENABLE_TIMEWARP',
      'ENABLE_VISUAL_ACTION',
      'ENABLE_WEBHOOK',
      'ENABLE_JIRA',
      'ENABLE_VISUAL_TIMELINE',
      'ENABLE_DSREQ_PERF_DISPLAY',
      'COMPRESS_DSREQ',
      'ENABLE_API_KEYS',
      'HELP_SHOW_WHATSNEW',
      'HELP_SHOW_GETSTARTED',
      'HELP_SHOW_DOCUMENTATION',
      'HELP_SHOW_SUPPORT_WEB_PAGE',
      'HELP_SHOW_SUPPORT_EMAIL',
      'HELP_SHOW_VERSION',
      'HELP_MENU_LINKS',
      'INCLUDE_CANTEEN',
      'PROFILE_PROXY_USERNAME',
      'MANAGE_DS_FA',
      'ENABLE_IE_OPTIMISATIONS',
      'ENABLE_DATAIMPORT_RW',
      'DSREQ_METHOD',
      'SESSION_COOKIE_SECURE',
      'SESSION_COOKIE_SAMESITE',
      'ENABLE_URL_NAMES',
      'ENABLE_DATACONN_ARCENGINE',
      'ENABLE_DATACONN_HIVE',
      'ENABLE_DATACONN_HYVE',
      'ENABLE_DATACONN_IMPALA',
      'ENABLE_DATACONN_IMPYLA',
      'START_TIME',
      'DATA_EXTRACTS_TARGET_TABLENAME_PREFIX',
      'DEFAULT_WORKSPACE',
      'SAML_REQUIRED_GROUPS_CHECK',
      'EMBED_URL_PREFIX',
      'DATACONNECTION_CONCURRENCY_DEFAULT',
      'DATACONNECTION_CONCURRENCY_MAX',
      'DATACONNECTION_CONCURRENCY_USER_DEFAULT',
      'DATACONNECTION_CONCURRENCY_USER_MAX',
      'DATACONNECTION_QUERY_TIMEOUT_MINUTES_DEFAULT',
      'DATACONNECTION_QUERY_TIMEOUT_MINUTES_MAX',
      'UPLOAD_MAXROWS_DEFAULT',
      'UPLOAD_MAXROWS_MAX',
      'OBJECT_URL_CLEANUP_SECONDS',
    ]

    # List of settings that have default values (not necessarily present in settings.py)
    pairs = [('USE_VIZ_COLNAME_ALIASES', True)]

    sd = dict([(k, getattr(settings, k)) for k in settings_list if hasattr(settings, k)])
    sd.update((dict((s, getattr(settings, s, d)) for (s, d) in pairs)))

    # Special handling for dsreq obfuscation--want the 'no flag present' to be the
    # active case but make the setting value default to True
    if getattr(settings, 'CLEARTEXT_DSREQ', True):
      sd['ctds'] = 1

    # Similar for server side dsreq generation -- want the 'ENFORCE' case to be
    # the default when no flag is present
    if getattr(settings, 'DSREQ_METHOD', None) != 'ENFORCE':
      sd['inclds'] = 1

    # Check if Chromium PDF is enabled and available
    sd['chromium_pdf_available'] = chromium_pdf_available()

  # add from env variables
  sd['CDV_MODE'] = os.getenv('CDV_MODE')

  sd['IS_CDV_MODE_CML'] = get_cmldata() != None

  # DWX-15352: public or private in CDW
  sd['CLOUD_PLATFORM_TYPE'] = os.getenv('CLOUD_PLATFORM_TYPE', 'unknown')
  sd['IS_GLOBAL_EMBED'] = os.getenv('IS_EMBED', 'false') == 'true'

  sd['CDV_SUB_MODE'] = os.getenv('CDV_SUB_MODE')

  return { "arcviz_settings":sd, "arcviz_settings_json":json.dumps(sd) }

def feature_flipper(request):
  url_flippers = UrlFlipper.get_url_flippers(request)
  feature_flipper = FeatureFlipper()
  feature_flipper.set_url_flippers(url_flippers)
  all_flippers = [f.__dict__ for f in feature_flipper.flippers]

  return {
          'flippers_full_info': json.dumps(all_flippers),
          'flippers': _get_flippers_dict(all_flippers),
        }

def _get_flippers_dict(flippers):
    flippers_dict = {}
    for flipper in flippers:
      flippers_dict.update({flipper["id"]: flipper["value"]})

    return flippers_dict

def javascript_directory(_):
  static = static_url(_)['arc_static']

  is_production_mode = os.getenv('IS_CDV_JAVASCRIPT_PRODUCTION', 'true').lower() == 'true'

  directory = "production" if is_production_mode else "development"

  return {
    "javascript_directory": f'{static}/js/build/{directory}',
    "include_canteen": getattr(settings, "INCLUDE_CANTEEN", None)
  }