# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from configparser import ConfigParser
import io
import logging
import os
import sys

LOG = logging.getLogger('arcweb.props')

def info(msg: str):
  LOG.info(msg)

  # we might not have logging set up.
  # So output the messages to stderr (because it is unbuffered)
  sys.stderr.write(f'{msg}\n')

def ini_content(prop_file: io.IOBase):
  yield '[DEFAULT]\n'
  yield from prop_file


def read_properties_file(property_file_path: str) -> dict:
  """
  A property file has the same structure as an ini file with the exception of sections.
  We mock an ini file content (using ini_content) by injecting a section header "DEFAULT"
  """
  config = ConfigParser()
  # maintain property key casing (default behavior converts them to lowercase)
  config.optionxform = lambda key: key
  with open(property_file_path, "tr") as f:
    config.read_file(ini_content(f))

  return config.defaults()


def read_properties():
  DATAVIZ_PROPERTIES = os.getenv('DATAVIZ_PROPERTIES', 'dataviz.properties')
  CONF_DIR = os.getenv('CONF_DIR', os.getcwd())
  DATAVIZ_PROPERTIES_PATH = os.path.join(CONF_DIR, DATAVIZ_PROPERTIES)

  info(f'DATAVIZ_PROPERTIES={DATAVIZ_PROPERTIES}')
  info(f'CONF_DIR={CONF_DIR}')
  info(f'DATAVIZ_PROPERTIES_PATH={DATAVIZ_PROPERTIES_PATH}')

  if os.path.exists(DATAVIZ_PROPERTIES_PATH):
    info(f'reading properties file {DATAVIZ_PROPERTIES_PATH}')
    properties = read_properties_file(DATAVIZ_PROPERTIES_PATH)
    return properties
  else:
    info(f'properties file does not exist: {DATAVIZ_PROPERTIES_PATH}')


def update_env_with_properties():
  """
  Reads dataviz.properties and updates os.environ with the values found there.
  Existing environ values are NOT overridden (i.e. the precedence of environment variables
  is higher than the one of the properties)
  """
  properties = read_properties()
  if properties:
    _update_environ(properties)


def _update_environ(properties: dict):
  info('environment update start')
  env = os.environ

  for key, value in properties.items():
    if key in env:
      # msg = f'  ignoring property {key} because environment already has value "{env[key]}"'
      # passwords might be logged, so rather do not do it:
      msg = f'  ignoring property {key} because environment already has a value.'
      info(msg)
    else:
      info(f'  defining variable {key}')
      env[key] = value

  info('environment update end')