# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import unittest
from unittest import mock
import json
import asyncio

from django.test import TestCase, Client
from django.urls import reverse
from django.contrib.auth.models import User

from arcwebbase.ai_profile_tester import validate_ai_service_connection

class ValidateAIServiceConnectionTests(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(username='testuser', password='testpass')
        self.client = Client()
        self.url = reverse('validate_ai_profile')

    def test_get_request_method_not_allowed(self):
        self.client.login(username='testuser', password='testpass')

        # Attempt to access the view with GET
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response.content.decode(), 'Invalid request method')

    def test_unauthenticated_access(self):
        response = self.client.post(self.url)
        self.assertEqual(response.status_code, 302)

    def test_invalid_json_payload(self):
        self.client.login(username='testuser', password='testpass')

        response = self.client.post(self.url, data="Invalid JSON", content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {'success': False, 'error': 'Invalid JSON payload'})

    @mock.patch('arcwebbase.ai_profile_tester.send_ai_request')
    def test_successful_request(self, mock_send_ai_request):
        self.client.login(username='testuser', password='testpass')

        async def single_response_generator(*args, **kwargs):
            yield {
                'status_code': 200,
                'response_body': {},
                'req_info': None
            }

        mock_send_ai_request.side_effect = single_response_generator

        data = {
            'tab': 'completion',
            'id': None,
            'settings': {
                'completion_api_key': 'test-key',
                'completion_model': 'test-model'
            }
        }

        response = self.client.post(self.url, data=json.dumps(data), content_type='application/json')

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'success': True, 'error': None})

        # Assert that send_ai_request was called with correct parameters
        expected_prof_data = {
            'settings': data['settings'],
            'input': [{'role': 'user', 'content': 'Hi'}],
            'model_type': 'completion'
        }
        mock_send_ai_request.assert_called_once_with(expected_prof_data)

    @mock.patch('arcwebbase.ai_profile_tester.send_ai_request')
    def test_ai_service_error_response(self, mock_send_ai_request):
        self.client.login(username='testuser', password='testpass')

        async def error_response_generator(*args, **kwargs):
            yield {
                'status_code': 400,
                'response_body': 'Some error occurred',
                'req_info': None
            }

        mock_send_ai_request.side_effect = error_response_generator

        data = {
            'tab': 'completion',
            'id': None,
            'settings': {
                'completion_api_key': 'test-key',
                'completion_model': 'test-model'
            }
        }

        response = self.client.post(self.url, data=json.dumps(data), content_type='application/json')

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {'success': False, 'error': 'Some error occurred'})

    @mock.patch('arcwebbase.ai_profile_tester.send_ai_request')
    def test_ai_service_timeout(self, mock_send_ai_request):
        self.client.login(username='testuser', password='testpass')

        async def timeout_generator(*args, **kwargs):
            raise asyncio.TimeoutError()
            yield  # ensures it's an async generator

        mock_send_ai_request.side_effect = timeout_generator

        data = {
            'tab': 'completion',
            'id': None,
            'settings': {
                'completion_api_key': 'test-key',
                'completion_model': 'test-model'
            }
        }

        response = self.client.post(self.url, data=json.dumps(data), content_type='application/json')

        self.assertEqual(response.status_code, 500)
        self.assertJSONEqual(response.content, {
            'success': False,
            'error': 'Unexpected error: The AI request timed out while waiting for a response.'
        })

    @mock.patch('arcwebbase.ai_profile_tester.send_ai_request')
    def test_exception_handling(self, mock_send_ai_request):
        self.client.login(username='testuser', password='testpass')

        async def exception_generator(*args, **kwargs):
            raise Exception('Test exception')
            yield

        mock_send_ai_request.side_effect = exception_generator

        data = {
            'tab': 'completion',
            'id': None,
            'settings': {
                'completion_api_key': 'test-key',
                'completion_model': 'test-model'
            }
        }

        response = self.client.post(self.url, data=json.dumps(data), content_type='application/json')

        self.assertEqual(response.status_code, 500)
        self.assertJSONEqual(response.content, {'success': False, 'error': 'Unexpected error: Test exception'})

    @mock.patch('arcwebbase.ai_profile_tester.setting_get_single')
    @mock.patch('arcwebbase.ai_profile_tester.send_ai_request')
    def test_placeholder_handling(self, mock_send_ai_request, mock_setting_get_single):
        self.client.login(username='testuser', password='testpass')

        mock_setting_get_single.return_value = json.dumps({
            'profileSettings': {
                'profile1': {
                    'values': {
                        'completion_api_key': 'saved-key',
                        'completion_model': 'saved-model'
                    }
                }
            }
        })

        async def single_response_generator(*args, **kwargs):
            yield {
                'status_code': 200,
                'response_body': {},
                'req_info': None
            }

        mock_send_ai_request.side_effect = single_response_generator

        data = {
            'tab': 'completion',
            'id': 'profile1',
            'settings': {
                'completion_api_key': '********',
                'completion_model': 'saved-model'
            }
        }

        # Mock settings.DEFAULT_PASSWORD_PLACEHOLDER
        with mock.patch('django.conf.settings.DEFAULT_PASSWORD_PLACEHOLDER', '********'):
            response = self.client.post(self.url, data=json.dumps(data), content_type='application/json')

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'success': True, 'error': None})

        # Assert that send_ai_request was called with the merged settings
        expected_settings = {
            'completion_api_key': 'saved-key',
            'completion_model': 'saved-model'
        }
        expected_prof_data = {
            'settings': expected_settings,
            'input': [{'role': 'user', 'content': 'Hi'}],
            'model_type': 'completion'
        }
        mock_send_ai_request.assert_called_once_with(expected_prof_data)

    def test_missing_required_fields(self):
        self.client.login(username='testuser', password='testpass')

        # Send a request with missing 'tab' field
        data = {
            'id': None,
            'settings': {
                'completion_api_key': 'test-key',
                'completion_model': 'test-model'
            }
        }

        response = self.client.post(self.url, data=json.dumps(data), content_type='application/json')

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {
            'success': False,
            'error': "Unknown engine type: None"
        })

    @mock.patch('arcwebbase.ai_profile_tester.send_ai_request')
    def test_model_type_embeddings(self, mock_send_ai_request):
        self.client.login(username='testuser', password='testpass')

        async def single_response_generator(*args, **kwargs):
            yield {
                'status_code': 200,
                'response_body': {},
                'req_info': None
            }

        mock_send_ai_request.side_effect = single_response_generator

        data = {
            'tab': 'embeddings',
            'id': None,
            'settings': {
                'embeddings_api_key': 'test-key',
                'embeddings_model': 'test-model'
            }
        }

        response = self.client.post(self.url, data=json.dumps(data), content_type='application/json')

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'success': True, 'error': None})

        expected_prof_data = {
            'settings': data['settings'],
            'input': 'Hi',
            'model_type': 'embeddings'
        }
        mock_send_ai_request.assert_called_once_with(expected_prof_data)
