# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import unittest
from unittest import mock
import asyncio
import os

from arcwebbase.ai_service_client import (
    AIServiceClient,
    send_ai_request,
    format_recursive,
    extract_error_message,
    get_engine_config,
    ENGINE_CONFIGS,
    CLEAR_TEXT_REQUEST,
    UNRESOLVED,
)

from arcwebbase.ai_settings import get_ai_settings
from django.conf import settings
from util.ai import bedrock

class TestAIServiceClient(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.loop.close()

    def test_init(self):
        engine_config = {'key': 'value'}
        client = AIServiceClient(engine_config)
        self.assertEqual(client.engine_config, engine_config)

    @mock.patch('arcwebbase.ai_service_client.bedrock.bedrock_comm')
    def test_get_handler_for_bedrock(self, mock_bedrock_comm):
        """Test that get_handler returns bedrock_comm when engine_type is 'bedrock'."""
        engine_config = {}
        client = AIServiceClient(engine_config)
        engine_type = 'bedrock'

        handler = client.get_handler(engine_type)
        self.assertEqual(handler, bedrock.bedrock_comm)

    def test_get_handler_for_non_bedrock(self):
        """Test that get_handler returns default_http_handler when engine_type != 'bedrock'."""
        engine_config = {}
        client = AIServiceClient(engine_config)
        engine_type = 'open_ai'

        handler = client.get_handler(engine_type)
        self.assertEqual(handler, client.default_http_handler)

    @mock.patch('arcwebbase.ai_service_client.settings')
    @mock.patch('aiohttp.ClientSession')
    def test_default_http_handler_json_response(self, mock_client_session, mock_settings):
        engine_config = {}
        client = AIServiceClient(engine_config)
        url = 'http://example.com/api'
        headers = {'Content-Type': 'application/json'}
        upstream_req_body = {'key': 'value'}
        response_data = {'result': 'success'}
        status = 200

        mock_settings.HTTP_PROXY_SSL_VERIFY = True
        mock_settings.HTTP_PROXY_TIMEOUT = 300

        mock_session = mock.MagicMock()
        mock_client_session.return_value.__aenter__.return_value = mock_session

        mock_response = mock.MagicMock()
        mock_session.request.return_value.__aenter__.return_value = mock_response

        mock_response.headers = {'Content-Type': 'application/json'}
        mock_response.status = status
        mock_response.json = mock.AsyncMock(return_value=response_data)
        mock_response.text = mock.AsyncMock(return_value='')

        # Simulate non-streaming
        upstream_req_body['stream'] = False

        async def consume_response():
            async for res in client.default_http_handler(url, headers, upstream_req_body):
                return res

        result = self.loop.run_until_complete(consume_response())
        self.assertEqual(result, (status, response_data))

        mock_session.request.assert_called_with(
            method="POST",
            url=url,
            headers=headers,
            json=upstream_req_body,
            ssl=mock_settings.HTTP_PROXY_SSL_VERIFY
        )

    @mock.patch('arcwebbase.ai_service_client.settings')
    @mock.patch('aiohttp.ClientSession')
    def test_default_http_handler_text_response(self, mock_client_session, mock_settings):
        engine_config = {}
        client = AIServiceClient(engine_config)
        url = 'http://example.com/api'
        headers = {'Content-Type': 'text/plain'}
        upstream_req_body = {'key': 'value'}
        response_text = 'some text response'
        status = 200

        mock_settings.HTTP_PROXY_SSL_VERIFY = True
        mock_settings.HTTP_PROXY_TIMEOUT = 300

        mock_session = mock.MagicMock()
        mock_client_session.return_value.__aenter__.return_value = mock_session

        mock_response = mock.MagicMock()
        mock_session.request.return_value.__aenter__.return_value = mock_response

        mock_response.headers = {'Content-Type': 'text/plain'}
        mock_response.status = status
        mock_response.text = mock.AsyncMock(return_value=response_text)

        upstream_req_body['stream'] = False

        async def consume_response():
            async for res in client.default_http_handler(url, headers, upstream_req_body):
                return res

        result = self.loop.run_until_complete(consume_response())
        self.assertEqual(result, (status, response_text))

        mock_session.request.assert_called_with(
            method="POST",
            url=url,
            headers=headers,
            json=upstream_req_body,
            ssl=mock_settings.HTTP_PROXY_SSL_VERIFY
        )

    @mock.patch('arcwebbase.ai_service_client.AIServiceClient.get_handler')
    def test_send_request_success(self, mock_get_handler):
        engine_config = {
            'url': '{URL}',
            'headers': {
                'Authorization': 'Bearer {SECRET}',
                'Content-Type': 'application/json',
            },
            'payload': {
                'completion': {
                    'model': '{MODEL}',
                    'messages': '{INPUT}',
                },
            },
        }
        client = AIServiceClient(engine_config)

        profile_data = {
            'URL': 'http://example.com/api',
            'SECRET': 'secret-token',
            'MODEL': 'model-name',
            'INPUT': 'input-message',
        }
        model_type = 'completion'
        engine_type = 'open_ai'

        async def mock_handler(url, headers, upstream_req_body):
            yield (200, {'response': 'data'})

        mock_get_handler.return_value = mock_handler

        result_gen = client.send_request(profile_data, model_type, engine_type)

        async def consume_response():
            async for res in result_gen:
                return res

        result = self.loop.run_until_complete(consume_response())

        expected_headers = {
            'Authorization': 'Bearer secret-token',
            'Content-Type': 'application/json',
        }
        expected_payload = {
            'model': 'model-name',
            'messages': 'input-message',
        }

        self.assertEqual(result['status_code'], 200)
        self.assertEqual(result['response_body'], {'response': 'data'})
        if CLEAR_TEXT_REQUEST:
            self.assertEqual(result['req_info']['headers'], expected_headers)
            self.assertEqual(result['req_info']['body'], expected_payload)
        else:
            self.assertNotIn('req_info', result)

    @mock.patch('arcwebbase.ai_service_client.AIServiceClient.get_handler')
    def test_send_request_with_transformers(self, mock_get_handler):
        engine_config = {
            'url': '{URL}',
            'headers': {
                'Authorization': 'Bearer {SECRET}',
                'Content-Type': 'application/json',
            },
            'payload': {
                'completion': {
                    'model': '{MODEL}',
                    'messages': '{INPUT}',
                },
            },
        }
        client = AIServiceClient(engine_config)

        profile_data = {
            'URL': 'http://example.com/api',
            'SECRET': 'secret-token',
            'MODEL': 'model-name',
            'INPUT': 'input-message',
            'REQUEST_TRANSFORMER': mock.Mock(return_value=({'transformed': 'headers'}, {'transformed': 'payload'})),
            'RESPONSE_TRANSFORMER': mock.Mock(return_value={'transformed': 'response'}),
        }
        model_type = 'completion'
        engine_type = 'open_ai'

        async def mock_handler(url, headers, upstream_req_body):
            yield (200, {'response': 'data'})

        mock_get_handler.return_value = mock_handler

        result_gen = client.send_request(profile_data, model_type, engine_type)

        async def consume_response():
            async for res in result_gen:
                return res

        result = self.loop.run_until_complete(consume_response())

        profile_data['REQUEST_TRANSFORMER'].assert_called()
        profile_data['RESPONSE_TRANSFORMER'].assert_called_with({'response': 'data'})

        self.assertEqual(result['status_code'], 200)
        self.assertEqual(result['response_body'], {'transformed': 'response'})

    @mock.patch('arcwebbase.ai_service_client.AIServiceClient.get_handler')
    def test_send_request_exception(self, mock_get_handler):
        engine_config = {
            'url': '{URL}',
            'headers': {},
            'payload': {},
        }
        client = AIServiceClient(engine_config)

        profile_data = {
            'URL': 'http://example.com/api',
        }
        model_type = 'completion'
        engine_type = 'open_ai'

        # Mock handler to raise exception
        async def mock_handler(url, headers, upstream_req_body):
            raise Exception('Test exception')
            yield

        mock_get_handler.return_value = mock_handler

        result_gen = client.send_request(profile_data, model_type, engine_type)

        async def consume_response():
            async for res in result_gen:
                return res

        result = self.loop.run_until_complete(consume_response())

        self.assertEqual(result['status_code'], 500)
        self.assertIn('Exception occurred: Test exception', result['response_body'])

class TestHelperFunctions(unittest.TestCase):
    def test_format_recursive_with_exact_placeholder(self):
        obj = '{PLACEHOLDER}'
        context = {'PLACEHOLDER': 'value'}
        result = format_recursive(obj, context)
        self.assertEqual(result, 'value')

    def test_format_recursive_with_embedded_placeholder(self):
        obj = 'Bearer {TOKEN}'
        context = {'TOKEN': 'abc123'}
        result = format_recursive(obj, context)
        self.assertEqual(result, 'Bearer abc123')

    def test_format_recursive_with_missing_placeholder(self):
        obj = 'Bearer {TOKEN}'
        context = {}
        result = format_recursive(obj, context)
        self.assertEqual(result, 'Bearer ')

    def test_format_recursive_with_nested_dict(self):
        obj = {
            'header': {
                'Authorization': 'Bearer {TOKEN}'
            },
            'data': {
                'input': '{INPUT}'
            }
        }
        context = {'TOKEN': 'abc123', 'INPUT': 'some input'}
        result = format_recursive(obj, context)
        expected = {
            'header': {
                'Authorization': 'Bearer abc123'
            },
            'data': {
                'input': 'some input'
            }
        }
        self.assertEqual(result, expected)

    def test_format_recursive_with_exact_placeholder_missing(self):
        obj = '{MISSING}'
        context = {}
        result = format_recursive(obj, context)
        # Expect UNRESOLVED when an exact placeholder is missing.
        self.assertIs(result, UNRESOLVED)

    def test_format_recursive_dict_skips_unresolved(self):
        obj = {
            'header': 'Bearer {TOKEN}',
            'footer': '{MISSING}'
        }
        context = {'TOKEN': 'abc123'}
        result = format_recursive(obj, context)
        expected = {
            'header': 'Bearer abc123'
        }
        self.assertEqual(result, expected)

    def test_extract_error_message_with_string(self):
        data = 'An error occurred.'
        result = extract_error_message(data)
        self.assertEqual(result, 'An error occurred.')

    def test_extract_error_message_with_empty_string(self):
        data = ''
        result = extract_error_message(data)
        self.assertEqual(result, 'An unknown error occurred.')

    def test_extract_error_message_with_error_in_dict(self):
        data = {'error': 'Error message'}
        result = extract_error_message(data)
        self.assertEqual(result, 'Error message')

    def test_extract_error_message_with_error_dict(self):
        data = {'error': {'message': 'Error message', 'code': '123'}}
        result = extract_error_message(data)
        self.assertEqual(result, 'Error message (Code: 123)')

    def test_extract_error_message_with_message_in_dict(self):
        data = {'message': 'Error message'}
        result = extract_error_message(data)
        self.assertEqual(result, 'Error message')

    def test_extract_error_message_with_detail_in_dict(self):
        data = {'detail': 'Error detail'}
        result = extract_error_message(data)
        self.assertEqual(result, 'Error detail')

    def test_extract_error_message_with_unknown_structure(self):
        data = {'unexpected': 'data'}
        result = extract_error_message(data)
        self.assertEqual(result, str(data))

    def test_get_engine_config_known_engine(self):
        engine_type = 'open_ai'
        result = get_engine_config(engine_type)
        self.assertEqual(result, ENGINE_CONFIGS[engine_type])

    def test_get_engine_config_unknown_engine(self):
        engine_type = 'unknown_engine'
        with self.assertRaises(ValueError) as context:
            get_engine_config(engine_type)
        self.assertIn('Unknown engine type: unknown_engine', str(context.exception))

class TestSendAIRequest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.loop.close()

    @mock.patch('arcwebbase.ai_service_client.get_engine_config')
    @mock.patch('arcwebbase.ai_service_client.get_ai_settings')
    @mock.patch('arcwebbase.ai_service_client.AIServiceClient.send_request')
    def test_send_ai_request_success(self, mock_send_request, mock_get_ai_settings, mock_get_engine_config):
        profile_data = {
            'settings': {
                'completion_engine': 'open_ai',
                'completion_api_key': 'secret-key',
                'completion_model': 'model-name',
                'completion_temperature': '0.7',
            },
            'model_type': 'completion',
            'input': 'test input',
        }

        engine_config = {'some': 'config'}
        mock_get_engine_config.return_value = engine_config

        ai_settings = {'some': 'settings'}
        mock_get_ai_settings.return_value = ai_settings

        async def single_response_generator(*args, **kwargs):
            yield {'status_code': 200, 'response_body': {'data': 'response'}}

        mock_send_request.side_effect = single_response_generator

        async def consume_first():
            async for response in send_ai_request(profile_data):
                return response
            return None

        result = self.loop.run_until_complete(consume_first())

        mock_get_engine_config.assert_called_with('open_ai')
        mock_get_ai_settings.assert_called_with(profile_data['settings'], 'completion')

        ai_settings_with_input = ai_settings.copy()
        ai_settings_with_input['INPUT'] = 'test input'

        mock_send_request.assert_called_once_with(ai_settings_with_input, 'completion', 'open_ai')
        self.assertEqual(result, {'status_code': 200, 'response_body': {'data': 'response'}})
