# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from unittest import mock
from unittest.mock import AsyncMock

import copy
import json
import asyncio
from django.test import TestCase
from arcwebbase.ai_validator import (
    num_tokens_from_string, 
    get_context_string, 
    validate_and_transform_request
)

class TestAiValidator(TestCase):
    def setUp(self):
        self.valid_req = {
            'body': {
                'upstreamReqBody': {
                    'input': [{'content': 'User message 1'}, {'content': 'User message 2'}]
                },
                'config': {
                    'embeddingMaxTokens': 1000,
                    'completionMaxTokens': 100,
                    'overflowPolicy': 'throw-error',
                    'sourceInfo': {
                        'filteredRows': [
                            ['first column', 'second column'],
                            ['third column', '']
                        ],
                        'aliases': ['Alias1', 'Alias2'],
                        'contextPlaceholder': 'Placeholder',
                        'contextPrompt': '<<context>> - This is the context'
                    }
                }
            },
            'modelType': 'completion',
            'profile': 'open-ai',
        }
        self.example_rows = self.valid_req['body']['config']['sourceInfo']['filteredRows']
        self.example_aliases = self.valid_req['body']['config']['sourceInfo']['aliases']
        self.context_placeholder = self.valid_req['body']['config']['sourceInfo']['contextPlaceholder']
        
        # Create an event loop for asynchronous tests
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.loop.close()

    def test_num_tokens_from_string(self):
        result = num_tokens_from_string("test string", "cl100k_base")
        self.assertEqual(result, 2)

    def test_get_context_string(self):
        result = get_context_string(self.example_rows, self.example_aliases, self.context_placeholder)
        expected_result = ('Alias1 1:\nfirst column\n'
                           'Alias2 1:\nsecond column\n\n'
                           'Alias1 2:\nthird column\n'
                           'Alias2 2:\nPlaceholder\n\n')
        self.assertEqual(result, expected_result)

    def test_get_context_string_empty_data(self):
        empty_rows = []
        with self.assertRaises(Exception) as context:
            get_context_string(empty_rows, self.example_aliases, self.context_placeholder)
        self.assertTrue('Context data is missing or empty.' in str(context.exception))

    @mock.patch('arcwebbase.ai_validator.get_ai_settings_prop_async', new_callable=AsyncMock, return_value={})
    def test_validate_and_transform_request_valid(self, mock_get_ai_settings_prop_async):
        transformed_req = self.loop.run_until_complete(
            validate_and_transform_request(copy.deepcopy(self.valid_req))
        )

        expected_messages = [
            {'content': 'User message 1'},
            {'role': 'system', 'content': ('Alias1 1:\nfirst column\n'
                                            'Alias2 1:\nsecond column\n\n'
                                            'Alias1 2:\nthird column\n'
                                            'Alias2 2:\nPlaceholder\n\n - This is the context')},
            {'content': 'User message 2'}
        ]

        self.assertEqual(transformed_req['body']['input'], expected_messages)

    def test_validate_and_transform_request_token_limit_exceeded(self):
        req_exceeding_tokens = copy.deepcopy(self.valid_req)
        req_exceeding_tokens['modelType'] = 'embeddings'
        req_exceeding_tokens['body']['upstreamReqBody']['input'] = 'This is a test input'
        req_exceeding_tokens['body']['config']['embeddingMaxTokens'] = 3

        with self.assertRaises(Exception) as context:
            self.loop.run_until_complete(validate_and_transform_request(req_exceeding_tokens))
        self.assertIn('This query exceeds token limitations', str(context.exception))
        
    @mock.patch('arcwebbase.ai_validator.get_ai_settings_prop_async', new_callable=AsyncMock, return_value={})
    def test_validate_and_transform_request_overflow_policy_remove_conversation(self, mock_get_ai_settings_prop_async):
        req_with_messages = copy.deepcopy(self.valid_req)
        req_with_messages['body']['config']['overflowPolicy'] = 'remove-conversation-first'
        req_with_messages['body']['config']['completionMaxTokens'] = 60
        req_with_messages['body']['upstreamReqBody']['input'] = [
            {'content': f'User message {i}'} for i in range(1, 9)
        ]

        transformed_req = self.loop.run_until_complete(
            validate_and_transform_request(req_with_messages)
        )

        # Ensure system message is inserted, maintaining list length equal to original
        self.assertEqual(len(transformed_req['body']['input']), 8)

    @mock.patch('arcwebbase.ai_validator.get_ai_settings_prop_async', new_callable=AsyncMock, return_value={})
    def test_validate_and_transform_request_overflow_policy_truncate_context(self, mock_get_ai_settings_prop_async):
        req_with_messages = copy.deepcopy(self.valid_req)
        req_with_messages['body']['config']['overflowPolicy'] = 'truncate-context-first'

        transformed_req = self.loop.run_until_complete(
            validate_and_transform_request(req_with_messages)
        )

        # Ensure context is truncated as per the overflow policy
        self.assertEqual(len(transformed_req['body']['input']), 3)

    def test_validate_and_transform_request_throw_error_overflow(self):
        req_throw_error = copy.deepcopy(self.valid_req)
        req_throw_error['body']['config']['overflowPolicy'] = 'throw-error'
        req_throw_error['body']['config']['completionMaxTokens'] = 5

        with self.assertRaises(Exception) as context:
            self.loop.run_until_complete(validate_and_transform_request(req_throw_error))
        self.assertIn('This query exceeds token limitations', str(context.exception))
