# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import unittest
import json
from unittest.mock import patch, MagicMock
from arcwebbase.ai_service_client import ChunkProcessor

LOGGER_PATH = 'arcwebbase.ai_support.LOGGER'

class TestChunkProcessor(unittest.TestCase):
    def setUp(self):
        self.processor = ChunkProcessor()

    @patch(LOGGER_PATH)
    def test_process_chunk_complete_json_single_line(self, mock_logger):
        """
        A single well-formed JSON line, no partial chunk carried over.
        """
        chunk = """data: {"choices":[{"delta":{"content":"Hello "}}]}"""
        results = self.processor.process_chunk(chunk)

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0], "Hello ")

        # No partial leftover, no logs about incomplete JSON
        self.assertEqual(self.processor.partial_chunk, "")
        mock_logger.info.assert_not_called()

    @patch(LOGGER_PATH)
    def test_process_chunk_multiple_lines(self, mock_logger):
        """
        Multiple well-formed data lines in one chunk.
        """
        chunk = (
            'data: {"choices":[{"delta":{"content":"Hello"}}]}\n'
            'data: {"choices":[{"delta":{"content":" World"}}]}'
        )
        results = self.processor.process_chunk(chunk)

        self.assertEqual(results, ["Hello", " World"])
        self.assertEqual(self.processor.partial_chunk, "")

    @patch(LOGGER_PATH)
    def test_process_chunk_incomplete_json(self, mock_logger):
        """
        Test that an incomplete JSON line is stored as partial_chunk
        and no content is returned yet.
        """
        chunk = 'data: {"choices":[{"delta":'
        results = self.processor.process_chunk(chunk)

        # No valid JSON extracted, results should be empty
        self.assertEqual(results, [])
        # partial_chunk should contain the incomplete JSON
        self.assertIn('"choices":[{"delta":', self.processor.partial_chunk)
        mock_logger.info.assert_called_with(
            "Incomplete JSON, storing for next chunk"
        )

    @patch(LOGGER_PATH)
    def test_process_chunk_complete_after_partial(self, mock_logger):
        """
        The second chunk completes the JSON started in the partial chunk.
        """
        # First chunk is incomplete
        chunk1 = 'data: {"choices":[{"delta":{"content":"Hel'
        _ = self.processor.process_chunk(chunk1)  # We don't care about the result here

        self.assertTrue(self.processor.partial_chunk)

        # Second chunk completes it
        chunk2 = 'lo"}}]}'
        results = self.processor.process_chunk(chunk2)

        # The partial is now combined, making a valid JSON:
        # {"choices":[{"delta":{"content":"Hello"}}]}
        self.assertEqual(results, ["Hello"])
        self.assertEqual(self.processor.partial_chunk, "")

    @patch(LOGGER_PATH)
    def test_process_chunk_finish_reason_stop(self, mock_logger):
        """
        If finish_reason == "stop", the chunk has no content to append.
        """
        # JSON with finish_reason = "stop"
        chunk = (
            'data: {"choices":[{"delta":{"finish_reason":"stop"}}]}'
        )
        results = self.processor.process_chunk(chunk)

        # No new content is appended, and we skip further processing
        self.assertEqual(results, [])
        mock_logger.info.assert_called_with("Completion finished with 'stop' reason.")

    @patch(LOGGER_PATH)
    def test_process_chunk_done_signal(self, mock_logger):
        """
        The presence of '[DONE]' means we end the processing immediately
        and return whatever we've accumulated so far.
        """
        # We'll do a two-line chunk to simulate partial content plus [DONE]
        chunk = (
            'data: {"choices":[{"delta":{"content":"Some content"}}]}\n'
            'data: [DONE]'
        )
        results = self.processor.process_chunk(chunk)

        # The JSON line is valid, so "Some content" is appended,
        # then we see [DONE], so we return immediately.
        self.assertEqual(results, ["Some content"])
        mock_logger.info.assert_any_call("Streaming complete.")

    @patch(LOGGER_PATH)
    def test_process_chunk_with_empty_stripped_line(self, mock_logger):
        """
        If after splitting 'data:' we get an empty line, it should be ignored.
        """
        chunk = 'data:  \n \n data:  {"choices":[{"delta":{"content":"Hello"}}]}'
        results = self.processor.process_chunk(chunk)
        self.assertEqual(results, ["Hello"])
        # No partial chunk
        self.assertEqual(self.processor.partial_chunk, "")
        mock_logger.info.assert_not_called()