# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import gc
import re
import unittest
from unittest import mock

from arcwebbase import gcmonitor


class TestGCMonitor(unittest.TestCase):

  def tearDown(self) -> None:
    gcmonitor.INSTANCE.unregister()
    return super().tearDown()

  def test_enabling(self):
    c = gcmonitor.INSTANCE

    c.register()
    self.assertTrue(c.enabled())
    c.unregister()
    self.assertFalse(c.enabled())

    # it is fine to register/unregister multiple times
    c.register()
    c.register()
    self.assertTrue(c.enabled())
    self.assertEqual(gc.callbacks.count(c), 1)

    c.unregister()
    c.unregister()
    self.assertFalse(c.enabled())

    # registering resets the last_log time:
    time1 = gcmonitor.INSTANCE.last_log_ns
    c.unregister()
    c.register()
    self.assertGreater(gcmonitor.INSTANCE.last_log_ns, time1)

  def test_logging(self):
    c = gcmonitor.GCCallback()

    c('start', {})
    start_time = c.start_ns
    with (mock.patch.object(gcmonitor, 'time')) as time_mock:
      perf_mock: mock.Mock = time_mock.perf_counter_ns
      perf_mock.return_value = start_time + 5
      c('stop', {})
      self.assertEqual(c.total_ns, 5)

      perf_mock.return_value = start_time + 20
      c('start', {})
      perf_mock.return_value = start_time + 30
      c('stop', {})
      self.assertEqual(c.total_ns, 15)

      # now test logging:
      with (mock.patch.object(gcmonitor, 'LOGGER')) as logger_mock:
        perf_mock.return_value = start_time + 60_000_000_000 + 10
        c('start', {})
        perf_mock.return_value = start_time + 60_000_000_000 + 20
        c('stop', {})
        self.assertEqual(logger_mock.info.call_count, 2)
        # the collected time was cleared
        self.assertEqual(c.total_ns, 0)

        gc_info_str = logger_mock.info.call_args_list[0].args[0]
        m = re.search('gc_time=([^ ]+)s\/', gc_info_str)
        logged_time = float(m[1])

        # we should have logged 25 nanoseconds
        self.assertAlmostEqual(logged_time*1.e9, 25)
