# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from apps.tests.userrole import base
from datasets.models import DataSet,Segment, DatasetSegment
from arcwebbase import initialize, gcmonitor
from django.db.models import Min
from django.utils import timezone
from zoneinfo import ZoneInfo
from datetime import timezone as timezone_new
from jobs.models import JobSchedule, JobLog, JobTarget, ScheduleGroup
from jobs import models
from unittest import mock
from unittest.mock import patch
from util.ff_flipper import FeatureFlipper
from vizlib.tests.test_report_count import run_with_db_query_check2
from vizlib.tests import test_report_operations
from vizlib.tests.test_sqlrun import Common
from reports.models import Report
from apps.views_snapshot import get_snapshot_temp_dir_path
from django.conf import settings

import datetime
import django.db
import django.test
import threading
import gc
import tempfile
import time
import os

MAX_INTERVAL = 3
MAX_ENTRIES = 2
MAX_RETENTION = 4

def enable_dataset_versioning():
  FeatureFlipper().set_flipper('enable_version_control', True, 'DB')
  FeatureFlipper().set_flipper('enable_dataset_version_control', True, 'DB')


class TestJobCleanup(django.test.TestCase):

  def _mock_settings(self, settings_getMock: mock.Mock, bulk_max_interval=MAX_INTERVAL, bulk_max_entries=MAX_ENTRIES, bulk_max_retention=MAX_RETENTION):
    """
    Configure settings_get_single mock to return the constants MAX_XXX defined above
    """
    arguments = locals()

    def mock_settings(setting_name):
      value = arguments[setting_name]
      return value

    settings_getMock.side_effect = mock_settings

  def _make_datetime(self, dt_string: str):
    naive_date = datetime.datetime.fromisoformat(dt_string)
    return naive_date.replace(tzinfo=timezone_new.utc)

  def _setup_timezone(self, timezoneMock: mock.Mock):
    timezoneMock.now.return_value = self._make_datetime('2024-02-14T08:00')

    # timedelta is the original implementation:
    timezoneMock.timedelta.side_effect = timezone.timedelta

  def _set_endtime(self, jl: JobLog, dt_string: str):
    cursor = django.db.connection.cursor()
    dt = self._make_datetime(dt_string)
    cursor.execute(
      'update jobs_joblog set end_time=%s where id=%s', (dt, jl.id))
    cursor.fetchall()

  def _create_test_data(self):
    user1 = base.make_user('user1')

    # there are a few predefined schedule groups. Pick one:
    schedule_group = ScheduleGroup.objects.all()[0]

    JobSchedule.objects.create(job_name='scheduled job',
                               creator=user1,
                               owner=user1,
                               schedule_group=schedule_group)

    # will be kept because some logs will be preserved
    job1 = JobSchedule.objects.create(
      job_name='job1', creator=user1, owner=user1)

    # will be removed because it has no logs and has a status
    job2 = JobSchedule.objects.create(
      job_name='job2', creator=user1, owner=user1, last_status=models.STATUS_COMPLETED)

    # will be kept because it has no status yet (it was just created)
    JobSchedule.objects.create(
      job_name='job3', creator=user1, owner=user1)

    # will be kept because it has queued status (no logs might have been created yet)
    JobSchedule.objects.create(
      job_name='job4', creator=user1, owner=user1, last_status=models.STATUS_QUEUED)

    # will be kept because job1 is preserved
    JobTarget.objects.create(schedule=job1)

    # will be removed because job2 is removed
    JobTarget.objects.create(schedule=job2)

    jl1 = JobLog.objects.create(schedule=job1, status=models.STATUS_RUNNING)
    jl2 = JobLog.objects.create(schedule=job1, status=models.STATUS_QUEUED)
    jl30 = JobLog.objects.create(schedule=job1, status=models.STATUS_COMPLETED)
    jl31 = JobLog.objects.create(schedule=job1, status=models.STATUS_COMPLETED)
    jl32 = JobLog.objects.create(schedule=job1, status=models.STATUS_CANCELED)
    jl33 = JobLog.objects.create(schedule=job1, status=models.STATUS_FAILED)

    # old, will be kept because of status
    self._set_endtime(jl1, '2020-01-01')

    # old, too, but kept because of status
    self._set_endtime(jl2, '2020-01-01')

    # old and completed, will be deleted because of age
    self._set_endtime(jl30, '2020-01-01')

    # new, but will be deleted because of max entries
    self._set_endtime(jl31, '2024-02-12')

    # newer, will be kept
    self._set_endtime(jl32, '2024-02-13')

    # newest, will be kept
    self._set_endtime(jl33, '2024-02-14')

    self.assertEquals(JobSchedule.objects.filter(
      schedule_group__isnull=False).count(), 1)
    self.assertEquals(JobSchedule.objects.filter(
      schedule_group__isnull=True).count(), 4)
    self.assertEquals(JobTarget.objects.count(), 2)
    self.assertEquals(JobLog.objects.count(), 6)

    pass

  def setUp(self) -> None:
    self.timezonePatcher = mock.patch('arcwebbase.initialize.timezone')
    self.setting_getPatcher = mock.patch(
      'arcwebbase.initialize.setting_get_single')
    self.timePatcher = mock.patch('arcwebbase.initialize.time')
    self.threadingPatcher = mock.patch('arcwebbase.initialize.threading')

    self.timezoneMock = self.timezonePatcher.start()
    self.setting_getMock = self.setting_getPatcher.start()
    self.timeMock = self.timePatcher.start()
    self.threadingMock = self.threadingPatcher.start()

  def tearDown(self) -> None:
    self.timezonePatcher.stop()
    self.setting_getPatcher.stop()
    self.timePatcher.stop()
    self.threadingPatcher.stop()

  def test_test_with_large_max_entries(self):
    # preserved logs will NOT be limited by maxentries
    self._do_test_logs_removal(max_entries=100, expected_remaining_logs=5)

  def test_test_with_small_max_entries(self):
    # preserved logs will be limited by maxentries
    self._do_test_logs_removal(max_entries=2, expected_remaining_logs=4)

    # from the newer jobs only two are kept because of the max_entries=2 setting:
    this_year_start = self._make_datetime('2024-01-01')
    self.assertEquals(JobLog.objects.filter(
      end_time__gt=this_year_start).count(), 2)

    # these remaing two are both newer than 2024-02-12:
    min_end_time = JobLog.objects.filter(end_time__gt=this_year_start).aggregate(
      min_end_time=Min('end_time'))['min_end_time']
    self.assertGreater(min_end_time, self._make_datetime('2024-02-12T08:00'))

  @mock.patch('arcwebbase.initialize.do_remove_old_joblogs')
  def test_parameter_passing(self, do_remove_old_joblogsMock: mock.Mock):
    self.threadingMock.Thread.assert_not_called()
    initialize.setup_clean_old_joblogs()
    self.threadingMock.Thread.assert_called_once()

    iteration_end_msg = 'end iteration'
    self.timeMock.sleep.side_effect = [None, Exception(iteration_end_msg)]
    remove_old_joblogs = self.threadingMock.Thread.call_args_list[0].kwargs['target']

    # have settings_get return the BULK_* properties defined in the constats
    # MAX_INTERVAL, MAX_ENTRIES and MAX_RETENTION
    self._mock_settings(self.setting_getMock)

    try:
      # mimic thread execution
      run_with_db_query_check2(remove_old_joblogs)

    except Exception as e:
      # hopefully it is our exception that we catch
      if e.args != (iteration_end_msg,):
        # is not, rethrow it:
        raise e

    sleep_func_arg = self.timeMock.sleep.call_args_list[0].args[0]
    day_seconds = 60 * 60 * 24

    self.assertEquals(sleep_func_arg, MAX_INTERVAL * day_seconds,
                      f'cleanup interval is {MAX_INTERVAL} days')

    do_remove_old_joblogsMock.assert_called_once_with(
      MAX_ENTRIES, MAX_RETENTION)

  def _do_test_logs_removal(self, max_entries: int, expected_remaining_logs: int):

    self._setup_timezone(self.timezoneMock)

    self._create_test_data()

    initialize.do_remove_old_joblogs(max_entries, retention_days=4)

    self.assertEquals(JobLog.objects.count(), expected_remaining_logs)

    # JobLogs with status RUNNING and QUEUED are kept even despite of being very old
    self.assertEquals(JobLog.objects.filter(
      status=models.STATUS_RUNNING).count(), 1)
    self.assertEquals(JobLog.objects.filter(
      status=models.STATUS_QUEUED).count(), 1)

    # the repeating schedule was kept
    self.assertEquals(JobSchedule.objects.filter(
      schedule_group__isnull=False).count(), 1)

    # one of the ad-hoc jobs was removed
    self.assertEquals(JobSchedule.objects.filter(
      schedule_group__isnull=True).count(), 3)

    # preserved jobs:
    self.assertEquals(JobSchedule.objects.filter(
      job_name__in=('job1', 'job3', 'job4')).count(), 3)

    # the job target belonging to the deleted job was deleted, too:
    self.assertEquals(JobTarget.objects.count(), 1)

class TestCGMonitorSetup(django.test.TestCase):

  def setUp(self) -> None:
    super().setUp()

    # store original setting
    self.GC_MONITOR_ENABLED = os.getenv('GC_MONITOR_ENABLED')
    gcmonitor.INSTANCE.unregister()

  def clear_env(self):
    if 'GC_MONITOR_ENABLED' in os.environ:
      del os.environ['GC_MONITOR_ENABLED']

  def tearDown(self):
    if self.GC_MONITOR_ENABLED:
      os.environ['GC_MONITOR_ENABLED'] = self.GC_MONITOR_ENABLED
    else:
        self.clear_env()

    gcmonitor.INSTANCE.unregister()
    super().tearDown()

  def test_default_setup(self):

    # not started by default:
    self.clear_env()
    initialize.setup_gc_monitor()

    self.assertFalse(gcmonitor.INSTANCE.enabled())

  def test_configured_setup(self):

    os.environ['GC_MONITOR_ENABLED'] = 'yes'
    initialize.setup_gc_monitor()

    self.assertTrue(gcmonitor.INSTANCE.enabled())


class TestCleanupUnnamedVersions(django.test.TestCase):
    def tearDown(self) -> None:
      FeatureFlipper.instance = None
      return super().tearDown()

    @patch('arcwebbase.initialize.setting_get_single')
    @patch('arcwebbase.initialize.time.sleep', return_value=None)  # Mock sleep to skip waiting
    def test_cleanup_unnamed_versions(self, mock_sleep, mock_setting_get_single):
        enable_dataset_versioning()
        util = test_report_operations.TestDashboardOperations()

        admin = base.make_user('admin', is_superuser=True)
        admin_client = base.create_client(admin)
        self.admin_client = admin_client

        # Should work with string inputs, hence the ','
        mock_setting_get_single.side_effect = lambda key: {
            'unnamed_versions_deletion_interval': '0.175',
            'unnamed_versions_max_retention': '7'
        }[key]

        now = timezone.now()
        old_time = now - timezone.timedelta(days=10)

        dc_id1 = util.create_dataconnection(admin_client)
        ds1_id = util.create_dataset(admin_client, dc_id1, 'ds1')
        ds2_id = util.create_dataset(admin_client, dc_id1, 'ds2')

        dataset1 = DataSet.objects.get(id=ds1_id)
        dataset2 = DataSet.objects.get(id=ds2_id)

        sqlrun_common = Common()
        segment1 = sqlrun_common.make_segment(
        admin_client, dataset1, 's1', '[A] in ("doug", "veronica")')
        segment2 = sqlrun_common.make_segment(
        admin_client, dataset1, 's2', '[AA] in ("dougA", "veronicaA")')
        segment3 = sqlrun_common.make_segment(
        admin_client, dataset1, 's3', '[AA] in ("dougA", "veronicaA")')

        self.s_2_1 = sqlrun_common.make_segment(
        admin_client, dataset2, 's3', '[AA] in ("dougA", "veronicaA")')

        unlinked_segment = Segment.objects.create()

        # update the update_time forcibly to be able to test the old dataset deletion part of the script
        dataset4 = DataSet.objects.get(id=4)
        DataSet.objects.filter(id=dataset4.id).update(update_time=old_time)

        segment_count = Segment.objects.all().count()
        self.assertEqual(segment_count, 5, "Initially there should be 5 segments in total (ds1:3, ds2:1, unlinked:1)")


        named_dataset_id = DataSet.objects.filter(id=3).update(is_named_version=True, update_time=old_time)
        named_dataset = DataSet.objects.get(id=named_dataset_id)


        initialize.aux_remove_unnamed_versions()

        self.assertFalse(DataSet.objects.filter(id=dataset4.id).exists(), "Old unnamed dataset version should be deleted")
        self.assertTrue(DataSet.objects.filter(id=dataset2.id).exists(), "Recent dataset should not be deleted")
        self.assertTrue(DataSet.objects.filter(id=named_dataset.id).exists(), "Named dataset should not be deleted")
        self.assertFalse(Segment.objects.filter(id=unlinked_segment.id).exists(), "Unlinked segment should be deleted")
        segment_count = Segment.objects.all().count()
        self.assertEqual(segment_count, 4, "Segment_count should equal to 4 after orphan deletion before dataset delete")

        DataSet.objects.filter(version_group_id=dataset1.id).delete()

        initialize.aux_remove_unnamed_versions()

        segment_count = Segment.objects.all().count()
        self.assertEqual(segment_count, 1, "Segment_count should be reduced to 1 after dataset version with 3 segments deleted")

        dataset2_segment = Segment.objects.filter(datasetsegment__dataset_version_id=dataset2.id).first()

        self.assertTrue(DatasetSegment.objects.filter(segment_id=dataset2_segment.id).exists(), "Non-orphaned segment association should not be deleted")
        self.assertTrue(Segment.objects.filter(id=dataset2_segment.id).exists(), "Non-orphaned segment should not be deleted")
        self.assertFalse(DatasetSegment.objects.filter(segment_id=segment1['id']).exists(), "Orphaned segment association should be deleted")
        self.assertFalse(Segment.objects.filter(id=segment1['id']).exists(), "Orphaned segment should be deleted")

class TestCleanUpOrphanVisuals(django.test.TestCase):
    @patch('time.sleep', return_value=None)  # Mock time.sleep to skip waiting
    def test_orphan_cleanup(self, mock_sleep):
        now = datetime.datetime.now(ZoneInfo("UTC"))
        two_days_ago = now - datetime.timedelta(days=2)
        one_hour_ago = now - datetime.timedelta(hours=1)

        orphan_2_days = Report.objects.create(
            report_name="Orphan 2 Days",
            report_data="{}",
            report_type='table',
            shared_visual_dashboards=None,
        )
        orphan_1_hour = Report.objects.create(
            report_name="Orphan 1 Hour",
            report_data="{}",
            report_type='table',
            shared_visual_dashboards=None,
        )
        regular_report = Report.objects.create(
            report_name="Regular Report",
            report_type='table',
            report_data='{"dashboard_id": "123"}',
            shared_visual_dashboards=None,
        )

        Report.objects.filter(id=orphan_2_days.id).update(create_time=two_days_ago)
        Report.objects.filter(id=orphan_1_hour.id).update(create_time=one_hour_ago)
        Report.objects.filter(id=regular_report.id).update(create_time=two_days_ago)

        initialize._remove_orphan_visuals_and_filters(run_once=True)

        self.assertFalse(Report.objects.filter(id=orphan_2_days.id).exists())
        self.assertTrue(Report.objects.filter(id=orphan_1_hour.id).exists())
        self.assertTrue(Report.objects.filter(id=regular_report.id).exists())

def dynamic_timestamp(delta):
  return (datetime.datetime.now() - delta).timestamp()



class TestCleanUpSnapshotTmpDirs(django.test.TestCase):
  @patch('time.sleep', return_value=None)
  def test_tmp_dir_older_than_1_hour_is_deleted(self, mock_sleep):
    snapshot_files_temp_dir_path = get_snapshot_temp_dir_path(settings.SNAPSHOT_TMP_DIR_PREFIX)
    os.makedirs(snapshot_files_temp_dir_path, exist_ok=True)
    # Create a snapshot dir older than 1 hour (should be deleted)
    old_dir = tempfile.mkdtemp(dir=snapshot_files_temp_dir_path, suffix='_snapshot')
    old_timestamp = dynamic_timestamp(datetime.timedelta(hours=2))
    os.utime(old_dir, (old_timestamp, old_timestamp))

    # Create a recent snapshot dir (should not be deleted)
    recent_dir = tempfile.mkdtemp(dir=snapshot_files_temp_dir_path, suffix='_snapshot')
    recent_timestamp = dynamic_timestamp(datetime.timedelta(minutes=30))
    os.utime(recent_dir, (recent_timestamp, recent_timestamp))

    initialize.remove_old_snapshot_dirs(run_once=True)

    self.assertFalse(os.path.exists(old_dir), 'Old snapshot dir was not deleted.')
    self.assertTrue(os.path.exists(recent_dir), 'Recent snapshot dir was incorrectly deleted.')
