# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import django.test
import json

from apps.tests.userrole import base
from arcwebbase import middleware
from django.conf import settings
from django.test.client import Client
from django.http.request import HttpHeaders
from unittest import mock
from vizlib.arcprof import ArcProf
from vizlib.tests import test_report_operations


class TestAccessLogging(django.test.TestCase):

  def setUp(self):
    self.old_access_log_setting = settings.ACCESS_LOG
    if not settings.ACCESS_LOG:
      settings.ACCESS_LOG = True

  def tearDown(self):
    settings.ACCESS_LOG = self.old_access_log_setting

  @mock.patch('arcwebbase.middleware.LOGGER')
  def test_root_url_redirection(self, log_mock: mock.Mock):
    client = django.test.client.Client()
    client.get('/')

    # "regular" requests are logged
    log_mock.info.assert_called_once()

  @mock.patch('arcwebbase.middleware.LOGGER')
  def test_root_url_redirection_for_healthcheck(self, log_mock: mock.Mock):
    client = django.test.client.Client(HTTP_USER_AGENT='Go-http-client/1.1')
    client.get('/')

    # health check requests are not logged
    log_mock.info.assert_not_called()

from vizlib import util
class TestProfileMiddleware(django.test.TestCase):

  def setUp(self):
    # test clients re-initialize the middlewares
    # because of this we can have multiple trackers in the
    # trackers list
    util.ALL_REQUEST_STATS_TRACKERS.clear()

  def test_authenticated_access(self):
    username='user1'
    user1 = base.make_user(username)
    client = base.create_client(user1)
    self._do_test_user_info(client, username)

  def test_unauthenticated_access(self):
    client = Client()
    username=''
    self._do_test_user_info(client, username)

  def _do_test_user_info(self, client: Client, expected_username: str):

    # this call is needed to have the tracker object created
    response = client.get('/stats_data/all')
    if expected_username == '':
      # unauthenticated access, we get redirected to the login page
      self.assertEqual(response.status_code, 302)
    else:
      # content type is text/json instead of application/json
      response_json = json.loads(response.content.decode())
      current_request = response_json['requests']['queues']['http']['curr'][0]

      self.assertEqual(current_request['username'], expected_username)

    http_tracker = [t for t in util.ALL_REQUEST_STATS_TRACKERS if t._name == 'http'][0]

    startRequestSpy = mock.Mock(wraps=http_tracker.startRequest)
    with mock.patch.object(http_tracker, 'startRequest', startRequestSpy):
      client.get('/stats_data/all')

    startRequestSpy.assert_called_once()
    # locate the profile data passed to the tracker
    profile: ArcProf = startRequestSpy.call_args_list[0].args[1]
    self.assertEquals(profile.username(), expected_username)


class TestStoreAccessInfoMiddleware(django.test.TestCase):
  def setUp(self):
    self.spy = mock.MagicMock(wraps=middleware.access_info)

  def _do_test_access_info(self, client: Client, expected_username: str):
    with mock.patch.object(middleware, 'access_info', new=self.spy):
      client.get('/stats_data/all')
      self.spy.create_from_request.assert_called_once()
      self.spy.store.assert_called_once()
      self.spy.clear.assert_called_once()

      self.assertEquals(self.spy.store.call_args.args[0].username, expected_username)

  def test_authenticated_access(self):
    username='user1'
    user1 = base.make_user(username)
    client = base.create_client(user1)
    self._do_test_access_info(client, 'user1')

  def test_unauthenticated_access(self):
    client = Client()
    self._do_test_access_info(client, '')

class TestCsrfProtection(django.test.TestCase):

  def setUp(self):
    self.util = test_report_operations.TestDashboardOperations()

    admin1 = base.make_user('admin1', is_superuser=True)
    client = Client(enforce_csrf_checks=True)
    client.login(username=admin1.username, password=admin1.username)

    # need to start with a GET that has a form so that the CSRF cookie is set
    response = client.get('/apps/login')
    csrf_token = str(response.context['csrf_token'])
    client.defaults.update(HttpHeaders.to_wsgi_names({'X-CSRFTOKEN':csrf_token}))
    self.client = client

  def create_dataconnection(self, dc_name: str, headers: dict, secure=False):
    dc_payload = {
        'dataconnection_type': 'arcengine',
        'dataconnection_info': '{"PARAMS":{"HOST":"10.0.0.211","PORT":"31050"}}',
        'dataconnection_name': dc_name,
    }
    response = self.client.post('/datasets/dataconnection',
                                data=dc_payload,
                                headers=headers,
                                secure=secure)
    self.assertEquals(response.status_code, 200)
    dc_id = response.json()['id']
    return dc_id

  def test_no_headers(self):
    self.create_dataconnection(dc_name='dc1', headers={})

  def test_with_origin(self):
    self.create_dataconnection(dc_name='dc1', headers={'Origin':'http://viza'})

  def test_secure_with_origin(self):
    self.create_dataconnection(dc_name='dc1', headers={'Origin':'https://vizb'}, secure=True)

  def test_with_referer(self):
    # insecure requests do not have the referrer checked
    self.create_dataconnection(dc_name='dc1', headers={'Referer':'http://vizc/x'})

  # we do not protect against referer checks
  # def test_secure_with_referer(self):
  #   self.create_dataconnection(dc_name='dc1', headers={'Referer':'https://vizd/y'}, secure=True)

