# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import unittest
import asyncio
from unittest import mock

from arcwebbase import views_proxy
from util.perfcounters import APPROX_COUNTERS

class TestProxyComm(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.original_value = APPROX_COUNTERS.get('proxy.http_requests.current').value
        APPROX_COUNTERS.get('proxy.http_requests.current').value = 0

    async def asyncTearDown(self):
        APPROX_COUNTERS.get('proxy.http_requests.current').value = self.original_value

    async def test_non_streaming_success(self):
        """
        If streaming is disabled, we expect exactly one chunk with a merged proxy response.
        """
        req = {
            'body': {'input': 'Hello!'},
            'modelType': 'completion',
            'settings': {
                'completion_stream': False
            }
        }

        async def mock_upstream_comm(prof_data):
            # This simulates a single chunk from upstream, with status_code=200
            yield {
                'status_code': 200,
                'response_body': {'msg': 'Upstream OK'},
                'req_info': {'info': 'some request data'}
            }

        # proxy_comm is an async generator
        gen = views_proxy.proxy_comm(req, mock_upstream_comm)
        chunks = []
        async for chunk in gen:
            chunks.append(chunk)

        # We expect exactly one chunk
        self.assertEqual(len(chunks), 1)
        final_chunk = chunks[0]

        # final_chunk should be of form: {'status_code': 200, 'response_body': {...}}
        self.assertEqual(final_chunk['status_code'], 200)
        response_body = final_chunk['response_body']

        self.assertIn('request', response_body)
        self.assertIn('response', response_body)

        self.assertEqual(response_body['request']['body']['input'], 'Hello!')
        self.assertEqual(response_body['response']['status'], 200)
        self.assertEqual(response_body['response']['body'], {'msg': 'Upstream OK'})
        self.assertEqual(response_body['clear_text_request'], {'info': 'some request data'})

        self.assertNotIn('settings', response_body['request'])

        # Performance counter check
        self.assertEqual(APPROX_COUNTERS.get('proxy.http_requests.current').value, 0)

    async def test_streaming_pass_through(self):
        """
        If streaming is enabled, proxy_comm should yield the chunks exactly as upstream provides them.
        """
        req = {
            'body': {'input': 'Hello Stream'},
            'modelType': 'completion',
            'settings': {
                'completion_stream': True
            }
        }

        async def mock_upstream_comm(prof_data):
            yield {'status_code': 200, 'response_body': 'chunk 1'}
            yield {'status_code': 200, 'response_body': 'chunk 2'}

        gen = views_proxy.proxy_comm(req, mock_upstream_comm)
        chunks = []
        async for chunk in gen:
            chunks.append(chunk)

        # Because streaming_enabled=True, we pass each chunk through untouched
        self.assertEqual(len(chunks), 2)
        self.assertEqual(chunks[0]['status_code'], 200)
        self.assertEqual(chunks[0]['response_body'], 'chunk 1')
        self.assertEqual(chunks[1]['status_code'], 200)
        self.assertEqual(chunks[1]['response_body'], 'chunk 2')

        self.assertEqual(APPROX_COUNTERS.get('proxy.http_requests.current').value, 0)

    async def test_non_200_status(self):
        """
        If upstream_comm yields a non-200 status, the proxy_comm in non-streaming mode
        wraps that in 'error'.
        """
        req = {
            'body': {'input': 'Bad input'},
            'modelType': 'completion',
            'settings': {
                'completion_stream': False
            }
        }

        async def mock_upstream_comm(prof_data):
            yield {
                'status_code': 400,
                'response_body': {'detail': 'Bad Request'},
                'req_info': None
            }

        gen = views_proxy.proxy_comm(req, mock_upstream_comm)
        chunks = []
        async for chunk in gen:
            chunks.append(chunk)

        self.assertEqual(len(chunks), 1)
        final_chunk = chunks[0]
        self.assertEqual(final_chunk['status_code'], 400)

        proxy_response = final_chunk['response_body']
        self.assertEqual(proxy_response['request']['body']['input'], 'Bad input')
        self.assertEqual(proxy_response['response']['status'], 400)

        # Because it's a non-200, the code uses the 'error' key
        self.assertEqual(proxy_response['response']['error'], {'detail': 'Bad Request'})
        self.assertNotIn('clear_text_request', proxy_response)

    async def test_upstream_exception(self):
        """
        If the upstream_comm raises an exception, it propagates out of proxy_comm.
        """
        req = {
            'body': {'input': 'Raises error'},
            'modelType': 'completion',
            'settings': {
                'completion_stream': False
            }
        }

        async def mock_upstream_comm(prof_data):
            raise RuntimeError("Upstream broke")
            yield

        gen = views_proxy.proxy_comm(req, mock_upstream_comm)

        with self.assertRaises(RuntimeError) as ctx:
            async for _chunk in gen:
                pass
        self.assertIn("Upstream broke", str(ctx.exception))

        self.assertEqual(APPROX_COUNTERS.get('proxy.http_requests.current').value, 0)
