# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Validation routines for models
"""

from django.core.exceptions import ValidationError

import re
import reports

# This is a big-hammer approach to JS string sanitization
# - don't permit < strings in a bunch of places
def validateStringNoHtml(s):
  if s is None:
    return

  if s.find('<') >= 0:
    raise ValidationError('Invalid character in field %s' % s)

# This is a smaller-hammer approach to JS string sanitization
# - don't permit </ or /> strings in a bunch of places - this makes it
#   so that we cannot close any HTML tags and hence not get any
#   executable scripts etc. while also permitting <<varname>> strings
#   that we allow for arc functionality
def validateStringNoEndingTagHtml(s):
  if s is None:
    return

  if s.find('</') >= 0:
    raise ValidationError('Invalid character in field %s' % s)

  if s.find('/>') >= 0:
    raise ValidationError('Invalid character in field %s' % s)

# Have only a limited set of characters allowed for a username
def validateUsernameCharacters(s):
  if s is None:
    return

  if re.search(r'[^0-9a-zA-Z_\.\-]', s, ) is not None:
    raise ValidationError('Usernames are limited to alphanumerics, periods, underscores, and dashes')

def validateNoHtml(m):
  m.clean_fields()

def validatePassword(pw):
  default_min_length = 5
  min_length = int(reports.views_settings.setting_get_single('min_pw_length') or default_min_length)
  if pw == '':
    raise ValidationError('Password must be specified')
  elif len(pw) < min_length:
    raise ValidationError("Password must be at least {number} characters".format(number=str(min_length)))
  elif reports.views_settings.setting_get_single('pw_require_number') == "true" and not re.search(r'\d', pw):
    raise ValidationError("Password must contain a number")
  elif reports.views_settings.setting_get_single('pw_require_symbol') == "true" and not re.search(r'\W', pw):
    raise ValidationError("Password must contain a symbol")

def escapeCarets(s):
  if not isinstance(s, str):
    return s

  carets = {
    '<': '&lt;',
    '>': '&gt;',
  }

  return ''.join([carets.get(char, char) for char in s])