# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
views for the arcweb dashboard
"""

from django.conf import settings
from django.http import HttpResponse, HttpResponseRedirect
from django.contrib.auth.decorators import login_required
import arcwebbase.middleware
import json

from arcwebbase.util import is_health_check_agent
from sqlrun.views_cache_statistic import CacheStatistics
from datasets.models import DataConnection
from dataconn.dataconn_dbapi2 import CONCURRENCY_TRACKER

from util.perfcounters import APPROX_COUNTERS

from rbac.permcache import RBAC

import os
import psutil

import vizlib.util

def home(request, error_msg=None):
  """homepage"""

  httpUserAgent = request.META.get('HTTP_USER_AGENT')

  # Do not send a redirect for unautheticated agents performing health checks:
  if is_health_check_agent(httpUserAgent) and not request.user.username:
    return HttpResponse('<a href="%s/apps/">apps</a>' % settings.ARC_BASEURL)

  return HttpResponseRedirect("%s/apps/" % settings.ARC_BASEURL)

@login_required
def stats_data(request, all=None):
  """webserver statistics"""


  has_permission = request.user.is_superuser or ('sys_viewlogs' in RBAC.get_user_obj_perms(request.user))
  if all != "all" or not has_permission:
    d = {
      'requests': {
        'queues': {}
      }
    }

    vizlib.util.updateRequestStats(d['requests']['queues'], user=request.user)
    return HttpResponse(json.dumps(d), content_type="text/json")

  d = arcwebbase.middleware.getRequestStatsData()
  d['counters'].update(APPROX_COUNTERS.getAllCounters())
  d['histograms'].update(APPROX_COUNTERS.getAllHistograms())
  # d['tscounters'] = APPROX_COUNTERS.getAllTimeSeriesCounters()

  updateConnnectionCounters(d['counters'], d['requests'])
  updateSystemStats(d['counters'])
  updateQueryCacheStatistic(d['counters'])

  vizlib.util.updateRequestStats(d['requests']['queues'])

  return HttpResponse(json.dumps(d), content_type="text/json")

def updateConnnectionCounters(dcounters, drequests):
  dcl = DataConnection.objects.all()
  idToName = {}
  for dc in dcl:
    idToName[dc.id] = dc.dataconnection_name

  data = CONCURRENCY_TRACKER.fetchAll()
  for [connection_id, active, total, runtime] in data:
    dcounters[f"conn_{connection_id}.name"] = idToName.get(connection_id, "<unknown>")
    dcounters[f"conn_{connection_id}.active"] = active
    dcounters[f"conn_{connection_id}.total"] = total
    dcounters[f"conn_{connection_id}.runtime"] = round(runtime, 3)

def updateSystemStats(d):
  process = psutil.Process(os.getpid())
  d['memory.rss_mb'] = int(process.memory_info().rss / (1024 * 1024))

  avg = psutil.getloadavg()
  cpucount = psutil.cpu_count()
  pct = [(x / cpucount) * 100 for x in psutil.getloadavg()]
  d['cpu.count'] = cpucount
  d['cpu.loadavg_pct_mins_1'] = round(pct[0], 3)
  d['cpu.loadavg_pct_mins_5'] = round(pct[1], 3)
  d['cpu.loadavg_pct_mins_15'] = round(pct[2], 3)

def updateQueryCacheStatistic(d):
  d['query_stat.total_number_of_queries_returned'] = CacheStatistics.sum_total_number_of_queries_returned()
  d['query_stat.returned_from_cache'] = CacheStatistics.sum_number_of_queries_returned_from_cache()
  d['query_stat.returned_from_backend'] = CacheStatistics.sum_number_of_queries_returned_from_backend()


