# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
The CDV administrative command line interface
"""

import imp
import os
import sys
import re

from cli.core import table, flags, misc, output

# plugins is not used here, but because of the way we dynamically load in .py files from
# the plugins module, we need to import the module
# pylint: disable-msg=W0611
from cli import plugins

flags.FLAGS.DEFINE_boolean('verbose', False, 'verbose output')
flags.FLAGS.DEFINE_string('connection', None, 'Connection name')

# IMPORTANT: 'readonly' needs to be implemented so we can make safe calls from the UI
# backend in response to GET calls
flags.FLAGS.DEFINE_boolean('readonly', False, 'only readonly commands can be executed')

def print_help():
  """Display help text for commands"""
  t = table.Table('Available Commands', ['Name', 'Description'])
  for cmd in plugins.PLUGIN_COMMANDS_ORDERED:
    help_text = plugins.PLUGIN_HELP.get(cmd)
    t.add_row([cmd, help_text])
  t.display()

def execute_command(cmdargs, args):
  """Execute the command specified"""
  if len(cmdargs) == 1 and cmdargs[0] == 'help':
    raise Exception('No command specified')

  is_help = False
  if cmdargs[0] == 'help':
    is_help = True
    cmdargs = cmdargs[1:]

  if len(cmdargs) > 1:
    original_cmd = f'{cmdargs[0]} {cmdargs[1]}'

  cmdargs[0] = plugins.load_plugins([cmdargs[0]])
  if len(cmdargs) == 1:
    print_help()
    return

  cmd = f'{cmdargs[0]} {cmdargs[1]}'
  cmd_tuple = plugins.PLUGIN_COMMANDS.get(cmd)
  if not cmd_tuple:
    possible_cmds = []
    for c in plugins.PLUGIN_COMMANDS:
      if c.startswith(cmd):
        possible_cmds.append(c)
    if len(possible_cmds) == 0:
      raise Exception(f"Command '{cmd}' not found")
    if len(possible_cmds) > 1:
      raise Exception(f"Command '{cmd}' not found. Multiple options {possible_cmds}")
    cmd = possible_cmds[0]
    cmd_tuple = plugins.PLUGIN_COMMANDS.get(cmd)

  if original_cmd != cmd:
    output.verbose(f"{original_cmd} -> {cmd}")

  fn, readonly, pattern, help_text = cmd_tuple
  if is_help:
    output.display(fn.__doc__)
    return

  full_pattern = f"^{pattern}$"
  args = args[2:]
  if not re.match(full_pattern, ' '.join(args)):
    raise Exception(f"Invalid arguments to command '{cmd}': '{' '.join(args)}'\n'{cmd}':\n  {help_text}\n{fn.__doc__}")

  if flags.FLAGS.readonly and not readonly:
    raise Exception("acli is running in readonly mode but this command " \
      "requires write access")

  fn(args)

def run(argsl):
  try:
    argsl = initialize(argsl)
    run_cmd(argsl)
  except Exception as e:
    if flags.FLAGS.verbose:
      raise
    output.error(str(e))
    return 1

def initialize(argsl):
  """Setup flags and initialize acli"""

  try:
    argsl = flags.FLAGS.parse(argsl)  # parse flags
    argsl = argsl[1:]
  except Exception as e:
    raise Exception(f'{e}\\nUsage: {argsl[0]} ARGS\\n{argsl}')

  return argsl

def run_cmd(argsl):
  """Run the specified acli command"""

  cmdargs = [x for x in argsl if not x.startswith('-')]
  if len(cmdargs) == 0:
    plugins.load_plugins()
    print_help()
    return
  execute_command(cmdargs, argsl)

