# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
#
# simple gflags-like functionality that can be used to declare and parse
# command line arguments. See usage in arcweb/cli/acli.py
#

class _FLAGS(object):
  def __init__(self):
    self._internal_flags = {}

  def __getattr__(self, name):
    if name in self._internal_flags:
      return self._internal_flags[name]['value']

    raise AttributeError(f"Could not find {name}")

  def DEFINE_boolean(self, name, default, desc, short_name=None):
    self._internal_flags[name] = {
      'default': default,
      'desc': desc,
      'type': 'boolean',
      'short_name': short_name,
      'value': default
    }

  def DEFINE_integer(self, name, default, desc, short_name=None):
    self._internal_flags[name] = {
      'default': default,
      'desc': desc,
      'type': 'integer',
      'short_name': short_name,
      'value': default
    }

  def DEFINE_string(self, name, default, desc, short_name=None):
    self._internal_flags[name] = {
      'default': default,
      'desc': desc,
      'type': 'string',
      'short_name': short_name,
      'value': default
    }

  def DEFINE_list(self, name, default, desc, short_name=None):
    self._internal_flags[name] = {
      'default': default,
      'desc': desc,
      'type': 'list',
      'short_name': short_name,
      'value': default
    }

  def parse(self, args):
    new_args = []

    idx = 0
    while idx < len(args):
      arg = args[idx]
      idx += 1

      handled = False
      for name in self._internal_flags:
        data = self._internal_flags[name]
        type = data['type']
        short_name = data['short_name']

        if arg == f"--{name}":
          handled = True

          if type == 'boolean':
            data['value'] = True
            break

          raise Exception(f"Invalid type {data.type} for flag {arg}")

        if arg == f"-{short_name}":
          handled = True

          if type == 'boolean':
            data['value'] = True
            break

          nextarg = args[idx]
          idx += 1

          if type == 'string':
            data['value'] = nextarg
            break

          if type == 'integer':
            data['value'] = int(nextarg)
            break

          if type == 'list':
            data['value'] = [x.strip() for x in nextarg.split(",")]
            break

        if arg == f"--no-{name}":
          handled = True

          if type == 'boolean':
            data['value'] = False
            break

          raise Exception(f"Invalid type {data.type} for flag {arg}")

        if arg.startswith(f"--{name}="):
          (l, p, r) = arg.partition("=")
          handled = True

          if type == 'boolean':
            if r.lower() == "true":
              data['value'] = True
              break

            if r.lower() == "false":
              data['value'] = False
              break

            raise Exception(f"Invalid value {arg} for flag of type boolean")

          if type == 'string':
            data['value'] = r
            break

          if type == 'integer':
            data['value'] = int(r)
            break

          if type == 'list':
            data['value'] = [x.strip() for x in r.split(",")]
            break

          raise Exception(f"Invalid type {data.type} for flag {arg}")

      if not handled:
        new_args.append(arg)

    return new_args

# Declare a singleton that is used globally to declare and parse flags
FLAGS = _FLAGS()

