# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Controlled output helpers
"""

import sys
import os
import errno
import time

from collections import deque

from cli.core import flags

_MSGSQ = None

_DISPLAY_TIMESTRING = False

def start_capture():
  """
  Start capturing messages output from this module. Used when the output
  has to be sent to an external client by arcthrift.py
  """

  global _MSGSQ

  _MSGSQ = deque([])

def stop_capture():
  """
  Stop capturing messages output from this module, and return the current
  set of captured messages.
  """
  global _MSGSQ

  m = _MSGSQ
  _MSGSQ = None
  return m

def display_timestrings(display_ts=True):
  """Display all messages with time strings"""
  global _DISPLAY_TIMESTRING

  _DISPLAY_TIMESTRING = display_ts

def time_string():
  """return a string representation of the current time"""
  return time.strftime("%Y-%m-%d.%H:%M:%S")

def display(msg, newline=True):
  """display message unconditionally"""

  if type(msg) != type('') and type(msg) != type(''):
    msg = str(msg)

  if _MSGSQ is not None:
    if newline:
      _MSGSQ.append(f"{msg}\n")
    else:
      _MSGSQ.append(msg)

  if _DISPLAY_TIMESTRING:
    sys.stdout.write(f"{time_string()} ")

  sys.stdout.write(msg)

  if newline:
    sys.stdout.write("\n")
  try:
    sys.stdout.flush()
  except IOError as e:
    if e.errno == errno.EPIPE:
      # pylint: disable-msg=W0212
      os._exit(1)

def write(msg):
  """helper function used by core.table"""
  display(msg, newline=False)

def verbose(msg, newline=True):
  """display verbose messages"""
  if flags.FLAGS.verbose:
    display(msg, newline)

def info(msg):
  """display informational messages"""
  display(msg)

def warning(msg):
  """display warning messages"""
  display(f'warning: {msg}')

def error(msg):
  """display error messages"""
  display(f'error: {msg}')

def progress(msg):
  """display progress messages"""
  display(f'progress: {msg}')

def progress_dot():
  """display progress dot"""
  display(".", newline=False)

