# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Helper functions to run local and remote commands, and to copy files to remote nodes.
"""

import subprocess
import os
import sys
import threading

from cli.core import flags
from cli.core import output

def exec_cmd(argsl, verbose=True):
  """
  Execute the command specified, replacing this process. This call does not return on success.
  """
  if verbose or flags.FLAGS.verbose:
    output.info(' '.join(argsl))

  os.execl(argsl[0], *argsl)

def run_cmd(argsl, combined_output=False, verbose=False, exception_on_error=True,
  exception_on_ctrlc=False, stdout_only=False, display_streaming_output=False,
  streaming_prefix='', streaming_suffix='', exitcode_only=False, stdin_devnull=True,
  streaming_stdout=sys.stdout, streaming_stderr=sys.stderr):

  """
  Run the specifid command. Return the exitcode, stdout and stderr.
  """
  stdout = subprocess.PIPE
  if combined_output:
    stderr = subprocess.STDOUT
  else:
    stderr = subprocess.PIPE

  try:
    # When this file is not imported from acli, flags.FLAGS.verbose may not be set
    # This try block is a workaround for that case
    if verbose or flags.FLAGS.verbose:
      output.info(' '.join(argsl))
  except Exception:
    pass

  def process_output(out, outl, streaming_file, streaming_prefix, streaming_suffix):
    """
    Read in data from the process output pipe specified in 'out' and add it
    to the list in 'outl'. If 'display_streaming_output' is set to True, also
    stream output to the 'streaming_file' file handle
    """

    for line in iter(out.readline, b''):
      if outl is not None:
        outl.append(line)
      if display_streaming_output:
        streaming_file.write(f'{streaming_prefix}{line}{streaming_suffix}')
        streaming_file.flush()
    out.close()

  try:
    with open(os.devnull) as devnull:
      if stdin_devnull:
        p = subprocess.Popen(argsl, stdin=devnull, stdout=stdout, bufsize=1,
          stderr=stderr)
      else:
        p = subprocess.Popen(argsl, stdout=stdout, bufsize=1, stderr=stderr)
      if exitcode_only:
        outl = None
      else:
        outl = []
      # pylint: disable=E1101
      out_thread = threading.Thread(target=process_output, args=(p.stdout, outl, streaming_stdout,
        streaming_prefix, streaming_suffix))
      out_thread.start()

      err_thread = None
      if exitcode_only:
        errl = None
      else:
        errl = []
      if stderr == subprocess.PIPE:
        # pylint: disable=E1101
        err_thread = threading.Thread(target=process_output, args=(p.stderr, errl, streaming_stderr,
          streaming_prefix, streaming_suffix))
        err_thread.start()

      # pylint: disable=E1101
      p.wait()

      # Wait to read all the input
      out_thread.join()
      if err_thread:
        err_thread.join()

      if outl:
        out = ''.join(outl)
      else:
        out = ''

      if errl:
        err = ''.join(errl)
      else:
        err = ''

      # pylint: disable=E1101
      exitcode = p.returncode
  except Exception as e:
    exitcode = 1
    out = ""
    err = str(e)

  if verbose:
    if exitcode:
      if verbose or flags.FLAGS.verbose:
        output.info(f'exitcode: {exitcode}')
    if not display_streaming_output:
      if out:
        output.info(f'stdout: {out}')
      if err:
        output.info(f'stderr: {err}')

  if exception_on_error and exitcode:
    if exitcode == 130 and not exception_on_ctrlc:
      pass
    else:
      raise Exception(f"Exiting because of exitcode {exitcode}. Command {argsl}")

  if stdout_only:
    return out

  if exitcode_only:
    return exitcode

  if combined_output:
    return exitcode, out

  return exitcode, out, err

