# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Provides a in-memory 'Table' object
"""

import json
import re
import sys
import itertools
import operator

from cli.core import flags, output

TABLE_CLI = 'table'
TABLE_JSON = 'json'
TABLE_CSV = 'csv'
TABLE_TSV = 'tsv'

flags.FLAGS.DEFINE_string('search', None, 'Search string for output tables', short_name='s')
flags.FLAGS.DEFINE_boolean('search_ignorecase', True,
  'Ignore case when the "search" option is uses')
flags.FLAGS.DEFINE_list('columns', None, 'List of columns to display')
flags.FLAGS.DEFINE_string('table_formatter', TABLE_CLI,
  'Formatter to use for tables. "table", "json" and "csv" are permitted values')
flags.FLAGS.DEFINE_string('csv_separator', ",",
  'When the "csv" table formatter is in use, this setting determines the separator to use')
flags.FLAGS.DEFINE_boolean('csv_headers', False,
  'When the "csv" table formatter is in use, this flag determines whether headers are displayed')

class _CsvFormatter:
  """
  Format a table for csv output
  """

  def __init__(self, writer=output, csvsep=None, csvhdrs=None):
    self._writer = writer

    # If parameters not given, use defaults from flags
    if csvsep:
      self._separator = csvsep
    else:
      self._separator = flags.FLAGS.csv_separator
    if csvhdrs is not None:
      self._showhdrs = csvhdrs
    else:
      self._showhdrs = flags.FLAGS.csv_headers

  def write_prefix(self):
    """write out the prefix"""
    pass

  def write_table_name(self, name):
    """write out the table name"""
    pass

  def write_column_headers(self, headersl, unused_widths, alignl=None):
    """write out column headers"""
    if self._showhdrs:
      self.write_row(headersl)

  def write_row(self, coll, alignl=None):
    """write out a single row"""
    self._writer.write(f"{self._separator.join(coll)}\n")

  def write_suffix(self):
    """write out the suffix"""
    pass

  def start_rows(self):
    """rows will begin now"""
    pass

  def end_rows(self):
    """rows are done"""
    pass

class _JsonFormatter:
  """
  Format a table for json output
  """

  def __init__(self, writer=output):
    self._writer = writer
    self._first_row = True

  def write_prefix(self):
    """write out the json prefix"""
    self._writer.write("{\n")

  def write_table_name(self, name):
    """write out the table name"""
    if name is not None:
      self._writer.write(f'  "name": "{name}",\n')

  def write_column_headers(self, headersl, unused_widths, alignl=None):
    """write out column headers"""
    self._writer.write(f'  "headers": {json.dumps(headersl, ensure_ascii=False)},\n')

  def write_row(self, coll, alignl=None):
    """write out a single row"""
    if self._firstRow:
      self._firstRow = False
    else:
      self._writer.write(",")

    self._writer.write("    ");
    self._writer.write(f'    {json.dumps(coll, ensure_ascii=False)}\n')

  def write_suffix(self):
    """write out the json suffix"""
    self._writer.write("}\n")

  def start_rows(self):
    """rows will begin now"""
    self._firstRow = True
    self._writer.write('  "rows": [\n')

  def end_rows(self):
    """rows are done"""
    self._writer.write("\n  ]\n")

class _CliFormatter:
  """
  Format a table for human readable output from a command line interface
  """

  def __init__(self, writer=output, maxcolwidth=0, multirow=False):
    self._writer = writer
    self._row_count = 0
    self._widths = None
    self._maxcolwidth = maxcolwidth
    self._multirow = multirow

  def write_table_name(self, name):
    """print the table name"""
    if name is not None:
      self._writer.write(f"{name}\n")

  def _one_row(self, coll, alignl):
    more = True
    while more:
      more = False
      data = coll
      coll = []
      for col, maxwidth, align in zip(data, self._widths, alignl):
        if self._multirow:
          coll.append(col[maxwidth:])
        if len(col) > maxwidth:
          col = col[:maxwidth]
          more = True

        if align == 'l':
          self._writer.write('| ' + col + ' ' * (maxwidth - len(col)) + ' ')
        else:
          self._writer.write('| ' + ' ' * (maxwidth - len(col)) + col + ' ')
      self._writer.write('|\n')

  def write_column_headers(self, headersl, widths, alignl=None):
    """print the column headers"""

    if self._maxcolwidth == 0:
      self._widths = widths
    else:
      self._widths = [min(w, self._maxcolwidth) for w in widths]

    if alignl is None:
      alignl = 'l' * len(headersl)

    total_width = 2 + sum(self._widths) + 3 * (len(self._widths) - 1) + 2
    self._writer.write('+' + '-' * (total_width - 2) + '+\n')

    if len(headersl) == 0:
      self._writer.write('|')
    self._one_row(headersl, alignl)

    self._writer.write('+' + '-' * (total_width - 2) + '+\n')

  def write_suffix(self):
    """print any tail information"""
    total_width = 2 + sum(self._widths) + 3 * (len(self._widths) - 1) + 2
    self._writer.write('+' + '-' * (total_width - 2) + '+\n')
    self._writer.write(f'{self._row_count} rows\n')

  def write_row(self, coll, alignl=None):
    """print a single row"""
    if alignl is None:
      alignl = 'l' * len(coll)
    self._one_row(coll, alignl)
    self._row_count += 1

  def start_rows(self):
    """no-op"""
    pass

  def end_rows(self):
    """no-op"""
    pass

  def write_prefix(self):
    """no-op"""
    pass

def table_from_json_string(table_string):
  """
  Given a json table string created by 'print', convert it back into a table

  {
    "name": "HBase Table arcmeta"
    "headers": ["Family", "Min Versions", "Max Versions"]
    "rows": [
      ["d", "0", "10"]
    ]
  }
  """

  j = json.loads(table_string)
  t = Table(str(j['name']), j['headers'])
  for row in j['rows']:
    t.add_row(row)
  return t

def tables_from_json_string(table_string):
  """
  Given a json table string created by 'print', convert it back into a table

  {
    "name": "HBase Table arcmeta"
    "headers": ["Family", "Min Versions", "Max Versions"]
    "rows": [
      ["d", "0", "10"]
    ]
  }
  """

  tables = []

  json_tables = json.loads(table_string)
  for j in json_tables:
    t = Table(str(j['name']), j['headers'])
    for row in j['rows']:
      t.add_row(row)
    tables.append(t)
  return tables

def table_from_string(table_name=None, table_string=None):
  """
  Given a string that was created by 'print', convert it back into a table
  """

  headers = []
  rows = []
  lastline = ""
  table_found = False
  for line in table_string.split("\n"):
    line = line.strip()
    if line.startswith('+'):
      table_found = True
      if not table_name:
        table_name = lastline
      continue
    if not table_name:
      lastline = line
    if not line.startswith('|'):
      continue
    line = line.strip("|")
    fields = line.split("|")
    fields = [x.strip() for x in fields]
    if not headers:
      headers = fields
      continue
    rows.append(fields)

  if not table_found:
    return None

  t = Table(table_name, headers)
  for row in rows:
    t.add_row(row)
  return t

class Table:
  """
  Object that represents an in-memory table for display
  """

  def __init__(self, tablename, headers):
    self._tablename = tablename
    self._headers = headers
    self._rows = []
    self._max_col_widths = [0 for unused_i in range(0, len(headers))]
    self._adjust_max_widths(headers)
    self._maxcolwidth = 0
    self._multirow = False

  def _adjust_max_widths(self, coll):
    """
    When a new row is added in, check to see if max column widths need to be increased.
    """
    for idx, col in enumerate(coll):
      l = len(col)
      if l <= self._max_col_widths[idx]:
        continue
      self._max_col_widths[idx] = l

  def set_colparms(self, maxcolwidth=None, multirow=None):
    """Set the column formatting parameters"""
    if maxcolwidth is not None:
      self._maxcolwidth = maxcolwidth
    if multirow is not None:
      self._multirow = multirow

  def rotate(self, maxrows=None):
    """Return a table rotated 90 degrees"""
    rcnt = len(self._rows)
    if maxrows:
      rcnt = min(rcnt, maxrows)

    hdrs = ['Column Name'] + ['Row %d' % r for r in range(1, rcnt + 1)]
    htab = Table(tablename=self._tablename, headers=hdrs)
    for col in range(0, len(self._headers)):
      row = [self._headers[col]] + [self._rows[r][col] for r in range(0, rcnt)]
      htab.add_row(row)
    return htab

  def sort(self, colpositionsl, reverse=False, sort_limit=None):
    """Given a iterator of column positions, sort by it"""
    self._rows = sorted(self._rows, key=operator.itemgetter(*colpositionsl), reverse=reverse)
    if sort_limit:
      self._rows = self._rows[:sort_limit]

  def add_row(self, coll):
    """
    Add another row to this table
    """
    coll = [str(col) for col in coll]
    self._rows.append(coll)
    self._adjust_max_widths(coll)

  def headers(self):
    """Accessor"""
    return self._headers

  def table_name(self):
    """Accessor"""
    return self._tablename

  def max_column_widths(self):
    """Accessor"""
    return self._max_col_widths

  def get_rows(self):
    """Accessor"""
    return self._rows

  def get_table_output(self, table_format=None, alignl=None, csvsep=None, csvhdrs=None):
    """Get table output in a string"""

    class Writer:
      """class used to capture table output"""
      def __init__(self):
        self._data = []

      def write(self, msg):
        """capture a single message"""
        self._data.append(msg)

      def data(self):
        """return table data"""
        return ''.join(self._data)

    w = Writer()
    self.display(table_format=table_format, alignl=alignl, writer=w, csvsep=csvsep, csvhdrs=csvhdrs)
    return w.data()

  def display(self, table_format=None, alignl=None, writer=output, csvsep=None, csvhdrs=None):
    """Print out the contents of this table in the format specified"""

    if table_format is None:
      table_format = flags.FLAGS.table_formatter

    if table_format == TABLE_CLI:
      f = _CliFormatter(writer=writer, maxcolwidth=self._maxcolwidth, multirow=self._multirow)
    elif table_format == TABLE_JSON:
      f = _JsonFormatter(writer=writer)
    elif table_format == TABLE_CSV or table_format == TABLE_TSV:
      f = _CsvFormatter(writer=writer, csvsep=csvsep, csvhdrs=csvhdrs)
    else:
      raise Exception(f'Invalid table format {table_format}')

    column_indices = []
    if flags.FLAGS.columns:
      for c in flags.FLAGS.columns:
        found = False
        for i, header in enumerate(self._headers):
          if header.lower() == c.lower():
            column_indices.append(i)
            found = True
            break
        if not found:
          raise Exception(f"Column '{c}' not found")

      headersl = [self._headers[i] for i in column_indices]
      widthsl = [self._max_col_widths[i] for i in column_indices]
      if alignl:
        alignl = [alignl[i] for i in column_indices]
    else:
      headersl = self._headers
      widthsl = self._max_col_widths

    f.write_prefix()
    f.write_table_name(self._tablename)
    f.write_column_headers(headersl, widthsl, alignl)
    f.start_rows();

    reFlags = 0
    if flags.FLAGS.search_ignorecase:
      reFlags = re.IGNORECASE

    for row in self._rows:
      if flags.FLAGS.columns:
        row = [row[i] for i in column_indices]

      if flags.FLAGS.search:
        for col in row:
          if re.search(flags.FLAGS.search, col, reFlags):
            f.write_row(row, alignl)
            break
      else:
        f.write_row(row, alignl)
    f.end_rows()
    f.write_suffix()

  def insert_column(self, colpos, colname, coldatal):
    """Insert data in 'coldatal' into column position 'colpos' with the column name 'colname'"""

    maxlen = len(colname)
    self._headers.insert(colpos, colname)
    idx = 0
    for row in self._rows:
      data = str(coldatal[idx])
      row.insert(colpos, data)
      maxlen = max(maxlen, len(data))
      idx += 1

    self._max_col_widths.insert(colpos, maxlen)

  def prepend_column(self, colname, coldata):
    """Prepend a column to this table with the same value 'coldata' in each row"""
    headers = [colname]
    headers.extend(self._headers)
    self._headers = headers

    rows = []
    for row in self._rows:
      new_row = [coldata]
      new_row.extend(row)
      rows.append(new_row)
    self._rows = rows

    max_col_widths = [max(len(coldata), len(colname))]
    max_col_widths.extend(self._max_col_widths)
    self._max_col_widths = max_col_widths

  def append_table(self, other_table):
    """Append another table to this one. This will only work if both tables have the same column definitions"""
    for row in other_table.get_rows():
      self.add_row(row)

  def __len__(self):
    """Number of rows in this table"""
    return len(self._rows)

