# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Provides a threadpool and a workq to schedule tasks on
"""


import threading
from collections import deque

from cli.core import output
from cli.core import flags

import traceback

flags.FLAGS.DEFINE_integer('threadcount', 20,
  'For multi-threaded commands, the number of threads used')

class Runner:
  """
  A runner object to add to a threadpool work queue
  """

  NOTRUNNING = 'not running'
  RUNNING = 'running'
  DONE = 'done'

  def __init__(self, fn, args=tuple(), kwargs=dict()):
    self.__fn = fn
    self.__args = args
    self.__kwargs = kwargs
    self.__status = Runner.NOTRUNNING
    self.__fn_return = None
    self.__exception = None

  def execute_runner(self):
    """
    Run the function associated with this object
    """
    self.__status = Runner.RUNNING
    try:
      self.__fn_return = self.__fn(*self.__args, **self.__kwargs)
    except Exception as e:
      self.__exception = e
      output.error(f"exception: {str(e)} {traceback.format_exc()}")
    self.__status = Runner.DONE

  def function_return(self):
    """Accessor"""
    return self.__fn_return

  def exception(self):
    """Accessor"""
    return self.__exception

# pylint: disable=R0924
class WorkQ:
  """
  A threadsafe work queue object. Callers add work items to this object using 'add_work'.
  During threaded execution, get_work gets the next item of work to execute
  """
  def __init__(self):
    self.work_q = deque()
    self.lock = threading.Lock()

  def get_work(self):
    """Non-blocking function to get the next work element. Returns None if no work is present"""
    with self.lock:
      try:
        return self.work_q.popleft()
      except Exception:
        return None

  def add_work(self, runner):
    """Add work to the workq in a threadsafe way"""
    with self.lock:
      self.work_q.append(runner)

  def __len__(self):
    """Return the number of elements in this workq"""
    with self.lock:
      return len(self.work_q)

class WorkQThread(threading.Thread):
  """
  Thread-derived object for running tasks in parallel
  """

  def __init__(self, workq):
    threading.Thread.__init__(self)
    self.workq = workq

  def run(self):
    """Function runner for this thread object"""
    while True:
      work = self.workq.get_work()
      if work == None:
        return
      work.execute_runner()

class ThreadPool:
  """
  Threadpool object for executing work queue elements in parallel
  """
  def __init__(self, maxlen=flags.FLAGS.threadcount):
    self._maxlen = maxlen

  def maxlen(self):
    """Max parallelism on this threadpool"""
    return self._maxlen

  def run_workq(self, workq):
    """
    Run tasks in the workq provided using threads in this threadpool
    """
    l = len(workq)
    if l > self._maxlen:
      size = self._maxlen
    else:
      size = l

    threads = []
    for unused_i in range(0, size):
      threads.append(WorkQThread(workq))

    for t in threads:
      t.start()

    for t in threads:
      t.join()

