# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
cli commands for Data Connections
"""

import json, os, traceback

from datetime import datetime
from django.conf import settings
from django.core import serializers

from cli.core import table
from cli.utils import read_json, json_to_dc_dict, create_dataconnection
from datasets.models import DataConnection
from datasets.utils import _auditDataConnection
from vizlib.querylog import ActionType

def show(argsl):
  """
  connection show

  List all dataconnections on this system
  """

  dcl = DataConnection.objects.all()

  t = table.Table('Data Connections', headers=["ID", "Name", "Type"])
  for dc in dcl:
    t.add_row([dc.id, dc.dataconnection_name, dc.dataconnection_type])
  t.display(alignl="rll")

def read(argsl):
  """
  connection read

  Returns metadata of the connections on this system in JSON format
  """
  dcs = DataConnection.objects.all()
  dcl = []
  for dc in dcs:
    serialized_dc = json.loads(serializers.serialize('json', [dc]))[0]
    dcl.append(serialized_dc)
  out = json.dumps(dcl, indent=4)
  print(out)

def create(argsl):
  """
  connection create <filename>

  Read the JSON file and create a data connection based on that. Returns an array of IDs of created objects.
  """

  if not getattr(settings, 'ENABLE_CREATE_VIA_CLI_TOOL', False):
    message = 'You have no permission to create via the CLI Tool. You can enable it via the Advanced Settings, please check the documentation'
    print(message)
    return message

  try:
    list_dc_json = read_json(argsl[0])
    source_filename = os.path.basename(argsl[0])
  except OSError:
    print(f'File {argsl[0]} not found.\nCheck in the current working directory {os.getcwd()}')
    return

  created_ids = []

  if not isinstance(list_dc_json, list):
    list_dc_json = [list_dc_json]
  for dc_json in list_dc_json:
    try:
      if 'fields' in dc_json.keys():
        dc_json = dc_json['fields']
      dc_dict = json_to_dc_dict(dc_json, source_filename)
      if 'create_user' not in dc_dict.keys():
        dc_dict['create_user'] = os.environ.get('HADOOP_USER_NAME', 'CLI user')
        dc_dict['update_user'] = os.environ.get('HADOOP_USER_NAME', 'CLI user')
      create_dataconnection(dc_dict)
      _auditDataConnection(DataConnection.objects.last(), ActionType.CREATE, username=dc_dict['create_user'])
      created_ids.append(DataConnection.objects.last().id)
    except ValueError as eVal:
      print(str(eVal))
    except KeyError as eKey:
      print(f"The json does not contain all compulsory fields:\n {str(eKey)}")
    except:
      print(f"The data connection {dc_dict['dc_name']} could not be saved")
  print(f"Successfully created the following data connections: {','.join(map(str, created_ids))}")
  return created_ids

def delete(argsl):
  """
  connection delete <ID>

  Delete a connection and associated datasets with a specified ID
  """

  if not getattr(settings, 'ENABLE_DELETE_VIA_CLI_TOOL', False):
    message = 'You have no permission to delete via the CLI Tool. You can enable it via the Advanced Settings, please check the documentation'
    print(message)
    return  message

  try:
    dc_id = int(argsl[0])
    dc = DataConnection.objects.get(id = dc_id)
    print("Please first see the object you are deleting with command read! Are you sure you want to delete this object? The referenced objects will be deleted too! Y for yes")
    action = input()
    if action == "Y" or action == "y":
      dc.delete()
      _auditDataConnection(dc, ActionType.DELETE, username=os.environ.get('HADOOP_USER_NAME', 'CLI user'))
      print("Deletion successful")
  except Exception as e:
    traceback.print_exc()

_COMMANDS = [
  ('show', show, True, '',
     "show all the data connections on this system"),
  ('read', read, True, '',
     "returns metadata of the connections on this system in JSON format"),
  ('create', create, False, '^.*\.json$',
     "parses a json file and establishing a new data connection. Returns an array of IDs of created objects."),
  ('delete', delete,  False, '^[0-9]+$',
    "connection delete <id>. Delete the connection with a specified ID")
]

def plugin_defn():
  """
  Return meta information about the plugin
  """

  defn = {
    'help'    :
      'Commands to work with connections',
    'commands': _COMMANDS
  }
  return defn
