# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
cli commands for Datasets
"""

import json, os, traceback
import cli.plugins.query
import cli.plugins.user
import extracts.misc
import extracts.views

from datetime import datetime
from django.conf import settings
from django.core import serializers

from apps.utils import WHATSNEWVER
from cli.core import table, flags
from cli.utils import read_json, validate_params_of_a_dataset
from dataconn.dataconn import get_dataconn
from datasets.models import DataSet, DataConnection
from datasets.views import _auditDataset
from migration.migrate import export_objects, import_objects
from vizlib.querylog import ActionType

def getDataConnectionId(name=None):
  if not name:
    connection_name = flags.FLAGS.connection
    if not connection_name:
      raise Exception("Must specify --connection")

    return getDataConnectionId(connection_name)

  dcl = DataConnection.objects.all()
  for dc in dcl:
    if str(dc.id) == name or dc.dataconnection_name == name:
      return dc.id

  return None

def getDataConnection():
  connectionId = getDataConnectionId()
  if connectionId:
    return get_dataconn(connection_id=connectionId)

  return None

def show(argsl):
  """
  dataset show

  List all datasets on this system
  """

  extractTracker = extracts.views.ExtractTracker()

  t = table.Table('Datasets', headers=["ID", "Name", "Connection", "Tags"])

  for ds in DataSet.objects.all():
    tags = []
    if extractTracker.isExtractSource(ds):
      tags.append('Extract Source')

    if extractTracker.isExtractTarget(ds):
      tags.append('Extract Target')

    t.add_row([ds.id, ds.dataset_name, ds.dataconnection.dataconnection_name, ', '.join(tags)])

  t.display(alignl='rlll')

def tabledetail(argsl):
  """
  dataset tabledetail

  Show table details for datasets on this system
  """

  t = table.Table('Datasets', headers=["ID", "Name", "Tables"])
  for ds in DataSet.objects.all():
    tablenames = ds.datasetTables()
    t.add_row([ds.id, ds.dataset_name, "; ".join(tablenames)])
  t.display(alignl='rll')

def data(argsl):
  """
  dataset data

  Show data stored in this dataset by running the equivalent of a select * on
  all columns of this dataset.
  """

  datasetId = int(argsl[0])
  srcDataset = DataSet.objects.get(id=datasetId)
  user = cli.plugins.user.get_super_user()

  (query, colnames) = extracts.misc.datasetToSQL(srcDataset, user=user)
  query = extracts.misc.addLimit(query, limit=10)
  flags.FLAGS.connection = srcDataset.dataconnection.dataconnection_name
  cli.plugins.query.run([query])

def create(args):
  """
  dataset create <filename>

  Read the data from a JSON file and create a dataset based on that. Returns an array of IDs of created objects.
  """

  if not getattr(settings, 'ENABLE_CREATE_VIA_CLI_TOOL', False):
    print('You have no permission to create via the CLI Tool. You can enable it via the Advanced Settings, please check the documentation')
    return

  try:
    datasets = read_json(args[0])
  except OSError:
    print(f'File {args[0]} not found.\nCheck in the current working directory {os.getcwd()}')
    return
  except Exception as e:
    traceback.print_exc()

  created_ids = []

  if not isinstance(datasets, list):
    datasets = [datasets]
  for dataset_data in datasets:
    try:
      if 'fields' in dataset_data.keys():
        dataset_data = dataset_data['fields']
      dataset_data = validate_params_of_a_dataset(dataset_data)
      if 'create_user' not in dataset_data.keys():
        dataset_data['create_user'] = os.environ.get('HADOOP_USER_NAME', 'CLI user')
        dataset_data['update_user'] = os.environ.get('HADOOP_USER_NAME', 'CLI user')
      dataset_data['dataset_info'] = json.dumps(dataset_data['dataset_info'])
      DataSet.objects.get_or_create(**dataset_data)
      _auditDataset(DataSet.objects.last(), ActionType.CREATE, username=dataset_data['create_user'])
      created_ids.append(DataSet.objects.last().id)
    except ValueError as eVal:
      print(f"{str(eVal)} at the dataset named '{dataset_data['dataset_name']}'")
    except Exception as e:
      traceback.print_exc()
      print(f"error occured while creating the dataset named '{dataset_data['dataset_name']}'")
  print(f"Successfully created the following datasets: {','.join(map(str, created_ids))}")
  return created_ids

def read(argsl):
  """
  dataset read

  Returns metadata of the datasets on this system in JSON format
  """

  datasets = DataSet.objects.all()
  datasetsl = list()
  for d in datasets:
    seralized_d = json.loads(serializers.serialize('json', [d]))[0]
    datasetsl.append(seralized_d)
  out = json.dumps(datasetsl, indent=4)
  print(out)

def delete(argsl):
  """
  dataset delete <ID>

  Delete a dataset with a specified ID
  """

  if not getattr(settings, 'ENABLE_DELETE_VIA_CLI_TOOL', False):
    print('You have no permission to delete via the CLI Tool. You can enable it via the Advanced Settings, please check the documentation')
    return

  try:
    ds_id = int(argsl[0])
    ds = DataSet.objects.get(id = ds_id)
    print("Please first see the object you are deleting with command read! Are you sure you want to delete this object? The referenced objects will be deleted too! Y for yes")
    action = input()
    if action == "Y" or action == "y":
      ds.delete()
      _auditDataset(ds, ActionType.DELETE, username=os.environ.get('HADOOP_USER_NAME', 'CLI user'))
      print("Deletion successful")
  except Exception as e:
    traceback.print_exc()

def export_assets(argsl):
  """
  dataset export_assets
  Exports all the dataset and dependencies

  """
  ds = DataSet.objects.all()
  data = export_objects('datasets', [i.version_id for i in ds], cli.plugins.user.get_super_user())
  data['version'] = {'Arcviz Version' : WHATSNEWVER}
  with open("dataset_export_{0}.json".format(datetime.now()), "w") as outfile:
    json.dump(data, outfile)

def import_assets(argsl):
  """
  dataset imports_assets <dataconnection_name> <filename>
  imports all the datasets and their dependencies from the specified file

  """
  try:
    data = read_json(argsl[1])
  except OSError:
    print(f'File {argsl[1]} not found.\nCheck in the current working directory {os.getcwd()}')
    return
  except Exception as e:
    print(f'JSON is invalid: {str(e)}')

  try:
    dc_name = argsl[0]
    workspace_id = argsl[2]
    dataconnection_obj = DataConnection.objects.get(dataconnection_name=dc_name)
    dataconnection_id = dataconnection_obj.id
    import_objects(data, dataconnection_id, cli.plugins.user.get_super_user(), False, True, None, None, workspace_id)
  except Exception as e:
    traceback.print_exc()

_COMMANDS = [
  ('show', show, True, '',
    "dataset show. Show all the datasets on this system"),

  ('tabledetail', tabledetail, True, '',
    "dataset tabledetail. Show tables within a dataset"),

  ('data', data, True, '(\d+)',
    "dataset data <datasetid>. Show data for the specified dataset"),

  ('create', create, False, '^.*\.json$',
    "datasets create <filename>. Read the data from file and create a dataset. Returns an array of IDs of created objects."),

  ('read', read, True, '',
    "dataset read. Returns metadata of the datasets on this system in JSON format"),

  ('export_assets', export_assets, True, '',
  "dataset export_assets. Exports datasets and their assets to a json file, returns the exported objects"),

  ('import_assets', import_assets, False, '^.+\ .*\..+$',
  "dataset import_assets <dataconnection name> <source filename>. Imports datasets using the dataconnection name from the source file (use export_assets to generate one)"),

  ('delete', delete, False, '^[0-9]+$',
    "dataset delete <id>. Delete the dataset with a specified ID"),

]

def plugin_defn():
  """
  Return meta information about the plugin
  """

  defn = {
    'help'    :
      'Commands to work with datasets',
    'commands': _COMMANDS
  }
  return defn

