# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
cli commands for reports
"""

from cli.core import table

import json
import hashlib
import typing

from django.contrib.auth.models import User
from datasets.models import DataSet, DataConnection, Extract

import cli.plugins.dataset
import cli.plugins.user
import extracts.views
import extracts.rewriter

from dataconn.dataconn import get_dataconn
from dataconn.qcntl import QueryControl
from vizlib.arcprof import ArcProf
from vizlib.dataset_request import DatasetRequest
from vizlib.expression import unbracketColumn

noTargetTableOption='--no-create-target-table'
def create(argsl):
  """
  Create a data extract.

  extract create refresh_full|refresh_partitioned [--no-create-target-table] <src_datasetid> <tgt_connectionid>
      <tgt_dbname> <tgt_tablename> <dims> <aggs> <partitions>
  """

  createTargetTable = True

  if noTargetTableOption in argsl:
    createTargetTable = False
    argsl.remove(noTargetTableOption)

  extract_type = argsl[0]
  srcDatasetId = int(argsl[1])
  tgtDataconnectionId = cli.plugins.dataset.getDataConnectionId(argsl[2])
  tgtDbname = argsl[3]
  tgtTablename = argsl[4]

  user = cli.plugins.user.get_super_user()

  dims = getStringListArg(argsl, 5)
  aggs = getStringListArg(argsl, 6)
  partitions = getStringListArg(argsl, 7)

  if not dims and not aggs:
    raise Exception("Must specify dims or aggs")

  dimData = toColumnList(dims)
  aggData = toColumnList(aggs)
  partitionData = toColumnList(partitions)

  request = MockRequest("POST", user)
  payload  = {
    "extract_type": extract_type,
    "srcDatasetId": srcDatasetId,
    "tgtDataconnectionId": tgtDataconnectionId,
    "tgtDbname": tgtDbname,
    "tgtTablename": tgtTablename,
    "dimData": dimData,
    "aggData": aggData,
    "partitionData": partitionData
  }

  if not createTargetTable:
    payload["createTargetTable"] = "false"

  request.body = json.dumps(payload)

  response = extracts.views.extract_internal(request)

  data = json.loads(response.content)
  id = data[0]["id"]
  print(f"Extract {id} created.")

def toStringList(stringInput):
  items = [x.strip() for x in stringInput.split(",")]
  # discard empty strings - can happen if the input itself is empty:
  return [x for x in items if x]

class MockRequest(object):
  def __init__(self, method, user):
    self.method = method
    self.user = user
    self.GET = {}

def tree(argsl):
  datasetIds = [int(x) for x in argsl]

  rewriteController = extracts.rewriter._getExtractRewriteController()
  if not datasetIds:
    rewriteController.display()
  else:
    for datasetId in datasetIds:
      rewriteController.display(datasetIds[0])

def rewritesInfo(argsl):
  extractIds = [int(x) for x in argsl]

  t = table.Table('Rewrites Enabled?', headers=['ID', 'Enabled'])

  for d in _fetchExtracts():
    d_id = d['id']
    if extractIds and not d_id in extractIds:
      continue

    d_config = d['extract_config']
    rewrites_status = extracts.rewriter._getRewriteStatusProperty(d_config)
    t.add_row([d_id, rewrites_status])

  t.display()

def rewritesEnable(argsl):
  extractIds = [int(x) for x in argsl]
  for extractId in extractIds:
    extracts.rewriter.updateRewriteStatusProperty(extractId, 'enabled')

def rewritesDisable(argsl):
  extractIds = [int(x) for x in argsl]
  for extractId in extractIds:
    extracts.rewriter.updateRewriteStatusProperty(extractId, 'disabled')

def rewritesDisableUntilRefreshed(argsl):
  extractIds = [int(x) for x in argsl]
  for extractId in extractIds:
    extracts.rewriter.updateRewriteStatusProperty(extractId, 'disabled_until_refreshed')

def _fetchExtract(extract_id: int):
  user = cli.plugins.user.get_super_user()

  request = MockRequest("GET", user)

  response = extracts.views.extract_internal(request, extract_id)
  return json.loads(response.content)

def _fetchExtracts(detail=False):
  user = cli.plugins.user.get_super_user()

  request = MockRequest("GET", user)
  if detail:
    request.GET["runstate"] = "true"

  response = extracts.views.extract_internal(request)
  data = json.loads(response.content)

  for d in data:
    yield d

def info(argsl):
  t = table.Table('Rewrites Information', headers=['Property', 'Value'])
  t.add_row(['Extracts Enabled', extracts.misc.enabled()])
  t.add_row(['Rewrites Enabled', extracts.misc.rewritesEnabled()])

  total = 0
  sourceDatasets = {}

  for d in _fetchExtracts():
    total += 1
    sourceDatasets[d['srcDatasetId']] = sourceDatasets.get(d['srcDatasetId'], [])
    sourceDatasets[d['srcDatasetId']].append(f'{d["tgtDbname"]}.{d["tgtTablename"]}')

  t.add_row(['Extracts Count', total])
  t.add_row(['Source Datasets', len(sourceDatasets)])

  t.display()

def showdetail(argsl):
  show(argsl, detail=True)

def show(argsl, detail=False):
  extractIds = [int(x) for x in argsl]

  cols = ["ID", "Type", "Source Dataset",
      "Target Connection", "Target Table", "Config"]
  if detail:
    cols.append("Run State")

  t = table.Table('Data Extracts', headers=cols)

  for d in _fetchExtracts(detail=detail):
    d_id = d['id']
    if extractIds and not d_id in extractIds:
      continue

    row = [d["id"], d["extract_type"], d["srcDatasetId"],
        d["tgtDataconnectionId"], f'{d["tgtDbname"]}.{d["tgtTablename"]}',
        d["extract_config"]]
    if detail:
      row.append(d["extract_runstate"])

    t.add_row(row)

  t.display()

def dropTargetTableUserChoice():
  while True:
    response = input("Delete target table? [no]: ").lower()
    if len(response) == 0 or response == 'no':
      return False
    elif response == 'yes':
      return True
    else:
      print('Please respond with "yes" or "no"')

def delete(argsl):
  user = cli.plugins.user.get_super_user()
  request = MockRequest("DELETE", user)

  deleteTable = 'true' if dropTargetTableUserChoice() else 'false'
  request.GET['deleteTable'] = deleteTable

  extract_id = int(argsl[0])

  response = extracts.views.extract_internal(request, extract_id)
  print(response.content)

def toColumnList(inputList: typing.List[str]):
    return [
        {
          'expr': x,
          'alias': unbracketColumn(x),
        }
        for x in inputList
    ]

def getStringListArg(inputList: typing.List[str], index: int, defaultValue=[]):
  if len(inputList) > index:
    return toStringList(inputList[index])
  else:
    return defaultValue

def edit(argsl):
  """
  Modify an existing data extract.

  extract edit <extract_id>
      <tgt_connectionid> <tgt_dbname> <tgt_tablename>
      <dims> <aggs> <partitions>
  """

  if len(argsl) < 3:
    print(edit.__doc__)
    return

  extract_id = int(argsl[0])
  tgtDataconnectionId = int(argsl[1])
  tgtDbname = argsl[2]
  tgtTablename = argsl[3]

  dims = getStringListArg(argsl, 4)
  aggs = getStringListArg(argsl, 5)
  partitions = getStringListArg(argsl, 6)

  extract = _fetchExtract(extract_id)[0]
  emailOpts = extract["extract_config"]["notification_email"]

  payload = {
    "tgtDataconnectionId": tgtDataconnectionId,
    "tgtDbname": tgtDbname,
    "tgtTablename": tgtTablename,
    "dimData": toColumnList(dims),
    "aggData": toColumnList(aggs),
    "partitionData": toColumnList(partitions),
    "extract_type": "refresh_partitioned" if len(partitions) > 0 else "refresh_full",
    "emailSuccess": emailOpts["email_success"],
    "emailFail": emailOpts["email_fail"],
    "emailTo": emailOpts["to_addresses"],
    "emailSubject": emailOpts["email_subject"]
  }

  user = cli.plugins.user.get_super_user()
  request = MockRequest("PUT", user)
  request.META={
    "REMOTE_ADDR": "localhost"
  }

  request.body = json.dumps(payload)

  response = extracts.views.extract_internal(request, extract_id)

  print("HTTP status = %s"%response.status_code)

  data = json.loads(response.content)
  id = data["extract_id"]
  print(f"Extract {id} updated.")

def showCreateCommands(argsl):
  _showCommands(argsl, "Create")

def showRefreshCommands(argsl):
  _showCommands(argsl, "Refresh")

def _showCommands(argsl, commandType):
  if len(argsl):
    ids = [int(id) for id in argsl[0].split(",")]
  else:
    ids = [d['id'] for d in _fetchExtracts()]

  cols = ["ID", "Type", "Connection", "Commands"]
  t = table.Table(f"{commandType} Commands", headers=cols)

  for id in ids:
    e = Extract.objects.get(id=id)
    mgr = extracts.views.extractManagerFromExtract(e)

    if commandType == "Create":
      cmds = list(mgr.getCreateCmds())
    else:
      cmds = list(mgr.getRefreshCmds())

    if cmds:
      connId = mgr.tgtDataconnectionId()
      conn = get_dataconn(connId)
      row = [id, mgr.extract_type(), conn.connection_type, json.dumps(cmds)]
      t.add_row(row)

  t.display()

def run(argsl):
  id = int(argsl[0])

  e = Extract.objects.get(id=id)
  mgr = extracts.views.extractManagerFromExtract(e)
  mgr.runRefresh()

_COMMANDS = [
  ('info', info, True, '',
    'extract info. Show information about data extracts on this system'),

  ('show', show, True, '.*',
    'extract show. Show all data extracts'),

  ('show_detail', showdetail, True, '.*',
    'extract show_detail. Show all data extracts, including runstate'),

  ('showdetail', showdetail, True, '.*',
    'alias to show_detail'),

  ('tree', tree, True, '\s*((\d+)\s*)*',
    'extract tree. Show data extracts in tree form'),

  ('create', create, True, '.*',
    'extract create <parameters>. Create a data extract'),

  ('delete', delete, True, '\d+',
    'extract delete <id>. Delete specified data extract'),

  ('edit', edit, True, '.*',
    'extract edit <id>. Modify specified data extract'),

  ('run', run, True, '\s*(\d+)\s*',
    'extract run <extract_id>. Run a data extract refresh'),

  ('show_refresh', showRefreshCommands, True, '\s*(\S*)\s*',
    'extract show_refresh [<extract_ids>]. Display the refresh commands for data extracts'),

  ('showrefresh', showRefreshCommands, True, '\s*(\S*)\s*',
    'alias to show_refresh'),

  ('show_create', showCreateCommands, True, '\s*(\S*)\s*',
    'extract show_create [<extract_ids>]. Display the create commands for data extracts'),

  ('showcreate', showCreateCommands, True, '\s*(\S*)\s*',
    'alias to show_create'),

  ('rewrites_info', rewritesInfo, True, '.*',
    'extract rewrites_info [<id> ...]. Get information about whether specific extracts are enabled for rewrites'),

  ('rewritesinfo', rewritesInfo, True, '.*',
    'alias to rewrites_info'),

  ('rewrites_enable', rewritesEnable, True, '.+',
    'extract rewrites_enable <id> ... Enable rewrites for the specified extract IDs'),

  ('rewritesenable', rewritesEnable, True, '.+',
    'alias to rewrites_enable'),

  ('rewrites_disable', rewritesDisable, True, '.+',
    'extract rewrites_disable <id> ... Disable rewrites for the specified extract IDs'),

  ('rewritesdisable', rewritesDisable, True, '.+',
    'alias to rewrites_disable'),

  ('rewrites_disable_until_refreshed', rewritesDisableUntilRefreshed, True, '.+',
    'extract rewrites_disable_until_refreshed <id> ... Disable rewrites until extract refreshed for the specified extract IDs'),
]

def plugin_defn():
  """
  Return meta information about the plugin
  """

  defn = {
    'help'    :
      'Commands to work with extracts',
    'commands': _COMMANDS
  }
  return defn
