# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
cli commands for datasets
"""

import json
import threading
import time

from cli.core import table

from datasets.models import DataConnection

import cli.plugins.dataset

def run(argsl):
  """
  query run

  Run queries on data connections
  """

  cmd = ' '.join(argsl)

  conn = cli.plugins.dataset.getDataConnection()
  result = conn.log_sync_runquery(cmd)
  rows = json.loads(result.get('rows'))

  t = table.Table('Query Results', headers=result.get('colnames'))
  for row in rows:
    t.add_row(row)
  t.display()

#
# Test cancelation with different connections
# First find a query that takes 5 seconds or more to run. Then run:
#   cli --connection=38 query run_and_cancel "<long query>"
# This will run the query and cancel it in a different thread after 2 seconds
# The query runs twice to confirm that the second run succeeds even though we
#   canceled the first one on the connection i.e. the cancelation leaves the
#   connection in a sane state
#
def run_and_cancel(argsl):
  class Canceler:
    def __init__(self):
      self._cancelFn = None

    def cancelThread(self):
      print("cancelThread sleeping")
      time.sleep(self.cancelSeconds)
      print("cancelThread up")
      if self._cancelFn:
        print("cancelThread canceling")
        self._cancelFn()

    def cancelAfterTimeout(self, cancelSeconds):
      self.cancelSeconds = cancelSeconds
      thread = threading.Thread(target=self.cancelThread)
      thread.daemon = True
      thread.start()

    def setCancelFn(self, cancelFn):
      self._cancelFn = cancelFn

  canceler = Canceler()

  cmd = ' '.join(argsl)
  conn = cli.plugins.dataset.getDataConnection()
  canceler.cancelAfterTimeout(2)

  for i in range(0, 2):
    t1 = time.time()
    try:
      result = conn.log_sync_runquery(cmd, setCancelCbFn=canceler.setCancelFn)
    except Exception as e:
      print(f"Query #{i+1} canceled in {time.time() - t1} seconds")
      print(e)
      continue

    print(f"Query #{i+1} complete in {time.time() - t1} seconds")
    print(result)

def data(argsl):
  """
  query data <tablename>

  Run a select * query on the table specified.
  """

  cmd = f"select * from {argsl[0]}"
  run([cmd])

_COMMANDS = [
  ('data', data, True, '(\S+)',
    "query data <tablename>. Run a select * query from the specified table"),

  ('run', run, True, '(.*)',
    "query run <sql>. Run the sql query specified"),

  ('run_and_cancel', run_and_cancel, True, '(.*)',
    "query run <sql>. Run the sql query specified")
]

def plugin_defn():
  """
  Return meta information about the plugin
  """

  defn = {
    'help'    :
      'Commands to work with queries',
    'commands': _COMMANDS
  }
  return defn

