# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.test import TestCase
import json
from unittest import mock

from cli.plugins import extract
from datasets.models import DataSet, Extract
from extracts.tests import tests_extracts, tests_rewrites
from util.ff_flipper import FeatureFlipper
from vizlib.tests import test_sqlrun

class TestExtract(TestCase):

  def setUp(self):
    FeatureFlipper().set_flipper('site_enable_extracts', True, 'DB')

    # one superuser is needed by the cli:
    test_sqlrun.Common.make_user(username='admin1', is_superuser=True)

    self.testsRewrites = tests_rewrites.Test_Rewrites()

    # will be accessed by the mock dataconnection:
    self.supports_dataextract_source = True
    self.supports_partitions = True
    self.supports_dataextract_target = True

  def supportsPartitions(self, _):
    return self.supports_partitions

  def make_extract(self, ds: DataSet = None):
    if not ds:
      dc = self.testsRewrites.make_dataconnection()
      ds = self.testsRewrites.make_dataset(dc, tablename='testdb1.table1')
    else:
      dc = ds.dataconnection

    e = Extract.objects.create(srcDataset=ds,
        tgtDataconnection=dc,
        tgtDbname='extractsdb',
        tgtTablename='cdv_target_1',
        extract_type='refresh_full',
        extract_config='{}',
        extract_runstate='{}'
    )
    return e

  def mockFetchDataConn(self, *args, **kwargs):
    return tests_extracts.MockDataConnection(args[0], self)

  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('cli.plugins.extract.dropTargetTableUserChoice')
  def test_delete(self, mock_userChoice, mock_fetchDataConn):
    self.do_test_delete(mock_userChoice, mock_fetchDataConn, drop_table=False)

  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('cli.plugins.extract.dropTargetTableUserChoice')
  def test_delete_and_drop_table(self, mock_userChoice, mock_fetchDataConn):
    self.do_test_delete(mock_userChoice, mock_fetchDataConn, drop_table=True)

  def do_test_delete(self, mock_userChoice: mock.Mock, mock_fetchDataConn: mock.Mock, drop_table: bool):

    e = self.make_extract()
    self.assertExtractCount(1)

    mockDataConn = tests_extracts.MockDataConnection(e.tgtDataconnection.id, self)
    mock_fetchDataConn.return_value = mockDataConn
    mock_userChoice.return_value = drop_table

    extract.delete([e.id])

    self.assertExtractCount(0)

    if drop_table:
      self.assertEquals(1, len(mockDataConn.cmds))
      self.assertEquals(['extractsdb.cdv_target_1'], mockDataConn.dropCmdParams)
    else:
      self.assertEquals(0, len(mockDataConn.cmds))

  def assertExtractCount(self, expected_count: int):
    self.assertEquals(expected_count, len(Extract.objects.all()))

  def prepare(self, mock_fetchDataConn):
    ds_columns = [{
        "name": "channel_fullname",
        "alias": "Channel Name",
      },
      {
        "name": "pdate",
        "alias": "Partition Date",
      }
    ]

    dc = self.testsRewrites.make_dataconnection()
    ds = self.testsRewrites.make_dataset(dc, tablename='testdb1.table1', columns=ds_columns)

    mockDataConn = tests_extracts.MockDataConnection(dc.id, self)
    mock_fetchDataConn.return_value = mockDataConn

    return ds, mockDataConn

  @mock.patch('extracts.misc.fetchDataConn')
  def test_create(self, mock_fetchDataConn):

    ds, mockDataConn = self.prepare(mock_fetchDataConn)

    create_args = [
      "refresh_full",
      f"{ds.id}",
      f"{ds.dataconnection.id}",
      "extractsdb",
      "cdv_extract_table1",
      "[Channel Name]",
      "sum(1)",
    ]

    self.assertExtractCount(0)
    extract.create(create_args)

    self.assertExtractCount(1)
    e = Extract.objects.all()[0]
    self.assertExtract(e,
      "refresh_full",
      ["[Channel Name]"],
      ["sum(1)"],
      []
    )

    self.assertEquals(1, len(mockDataConn.createCmdParams))
    self.assertEquals(2, len(mockDataConn.cmds))

    expectedSelect = """
    SELECT TA_0."channel_fullname" as `TA_0."channel_fullname"`, sum(1) as `sum(1)`
      FROM "testdb1"."table1"  TA_0
      GROUP BY 1 limit 0
    """
    self.verifySelect(expectedSelect, mockDataConn.cmds)

  @mock.patch('extracts.misc.fetchDataConn')
  def test_create_partitioned(self, mock_fetchDataConn):

    ds, mockDataConn = self.prepare(mock_fetchDataConn)

    create_args = [
      "refresh_partitioned",
      f"{ds.id}",
      f"{ds.dataconnection.id}",
      "extractsdb",
      "cdv_extract_table1",
      "[Channel Name]",
      "",
      "[Partition Date]",
    ]

    self.assertExtractCount(0)
    extract.create(create_args)

    self.assertExtractCount(1)
    e = Extract.objects.all()[0]
    self.assertExtract(e,
      "refresh_partitioned",
      ["[Channel Name]"],
      [],
      ["[Partition Date]"]
    )

    self.assertEquals(1, len(mockDataConn.createCmdParams))
    self.assertEquals(2, len(mockDataConn.cmds))

    expectedSelect = """
    SELECT TA_0."channel_fullname" as `TA_0."channel_fullname"`, TA_0."pdate" as `TA_0."pdate"`
      FROM "testdb1"."table1"  TA_0  limit 0
    """
    self.verifySelect(expectedSelect, mockDataConn.cmds)

  def verifySelect(self, expectedSelect, sqlCommands):
    actualSelect = [cmd[0] for cmd in sqlCommands if 'SELECT' in cmd[0]][0]
    self.testsRewrites.assertStringsEqualNoWhitespace(actualSelect, expectedSelect)


  def assertExtract(self, e: Extract, extract_type, dims, aggs, parts):
    e_config = json.loads(e.extract_config)
    self.assertEquals(extract_type, e.extract_type)
    self.assertEquals(dims, [e["expr"] for e in e_config["source"]["dimData"]])
    self.assertEquals(aggs, [e["expr"] for e in e_config["source"]["aggData"]])
    self.assertEquals(parts, [e["expr"] for e in e_config["target"].get("partitionData",[])])

  def assertConnectionCommands(self, mockDataConn: tests_extracts.MockDataConnection, drop: int, create: int):
    self.assertEquals(drop, len(mockDataConn.dropCmdParams))
    self.assertEquals(create, len(mockDataConn.createCmdParams))

  @mock.patch('extracts.misc.fetchDataConn')
  def test_edit(self, mock_fetchDataConn):

    ds, mockDataConn = self.prepare(mock_fetchDataConn)

    create_args = [
      "refresh_full",
      f"{ds.id}",
      f"{ds.dataconnection.id}",
      "extractsdb",
      "cdv_extract_table1",
      "[Channel Name]",
      "sum(1)",
    ]

    self.assertExtractCount(0)
    extract.create(create_args)
    self.assertExtractCount(1)
    self.assertConnectionCommands(mockDataConn, drop=0, create=1)

    e = Extract.objects.all()[0]

    edit_args = [
      f"{e.id}",
      f"{ds.dataconnection.id}",
      "extractsdb",
      "cdv_extract_table1",
      "[Partition Date]",
      "",
    ]
    extract.edit(edit_args)
    e.refresh_from_db()
    self.assertExtract(e,"refresh_full",["[Partition Date]"],[],[])
    self.assertConnectionCommands(mockDataConn, drop=1, create=2)

    edit_args = [
      f"{e.id}",
      f"{ds.dataconnection.id}",
      "extractsdb",
      "cdv_extract_table1",
      "",
      "sum(1)",
    ]
    extract.edit(edit_args)
    e.refresh_from_db()
    self.assertExtract(e,"refresh_full",[],["sum(1)"],[])
    self.assertConnectionCommands(mockDataConn, drop=2, create=3)

    edit_args = [
      f"{e.id}",
      f"{ds.dataconnection.id}",
      "extractsdb",
      "cdv_extract_table1",
      "[Channel Name]",
      "sum(2)",
      "[Partition Date]",
    ]
    extract.edit(edit_args)
    e.refresh_from_db()
    self.assertExtract(e,"refresh_partitioned",["[Channel Name]"],["sum(2)"],["[Partition Date]"])
    self.assertConnectionCommands(mockDataConn, drop=3, create=4)

    edit_args = [
      f"{e.id}",
      f"{ds.dataconnection.id}",
      "extractsdb",
      "cdv_extract_table1",
      "[Partition Date]",
      "count([Channel Name])",
      "",
    ]
    extract.edit(edit_args)
    e.refresh_from_db()
    self.assertExtract(e,"refresh_full",["[Partition Date]"],["count([Channel Name])"],[])
    self.assertConnectionCommands(mockDataConn, drop=4, create=5)

