# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
cli commands for Users
"""

import json, os, sys, traceback

from datetime import datetime
from django.contrib.auth.models import User
from django.conf import settings
from django.core import serializers

from apps.views import _auditUser
from cli.core import table
from cli.utils import read_json, validate_params_of_a_user
from vizlib.querylog import ActionType


def read(argsl):
  """
  user read

  Returns metadata of the users on this system in JSON format
  """

  users = User.objects.all()
  userl = list()
  for u in users:
    serialized_u = json.loads(serializers.serialize('json', [u]))[0]
    userl.append(serialized_u)
  out = json.dumps(userl, indent=4)
  print(out)

def show(argsl):
  """
  user show

  List all users on this system
  """

  userl = User.objects.all()

  t = table.Table('Users', headers=["ID", "Name", "Admin"])
  for user in userl:
    t.add_row([user.id, user.username, user.is_superuser])
  t.display(alignl="rll")

def create(argsl):
  '''
  user create <filename>

  Creates user data from a json file
  '''

  if not getattr(settings, 'ENABLE_CREATE_VIA_CLI_TOOL', False):
    print('You have no permission to create via the CLI Tool. You can enable it via the Advanced Settings, please check the documentation')
    return

  try:
    list_user_json = read_json(argsl[0])
    if not isinstance(list_user_json, list):
      list_user_json = [list_user_json]
  except OSError:
    print(f'File {argsl[0]} not found.\nCheck in the current working directory {os.getcwd()}')
    return
  except Exception as e:
    traceback.print_exc()
    return

  created_ids = []

  for u_json in list_user_json:
    try:
      u_json = validate_params_of_a_user(u_json)
      User.objects.create_user(**u_json)
      _auditUser(User.objects.last(), ActionType.CREATE, username=os.environ.get('HADOOP_USER_NAME', 'CLI user'))
      created_ids.append(User.objects.last().id)
    except ValueError as eVal:
      print(str(eVal))
      return
    except Exception as e:
      traceback.print_exc()
  print(created_ids)
  return created_ids

def delete(argsl):
  """
  user delete <ID>

  Delete a user with a specified ID
  """

  if not getattr(settings, 'ENABLE_DELETE_VIA_CLI_TOOL', False):
    print('You have no permission to delete via the CLI Tool. You can enable it via the Advanced Settings, please check the documentation')
    return

  try:
    u_id = int(argsl[0])
    user = User.objects.get(id = u_id)
    print("Please first see the object you are deleting with command read! Are you sure you want to delete this object? Y for yes")
    action = input()
    if action == "Y" or action == "y":
      user.delete()
      _auditUser(user, ActionType.DELETE, username=os.environ.get('HADOOP_USER_NAME', 'CLI user'))
      print("Deletion successful")
  except Exception as e:
    traceback.print_exc()

_COMMANDS = [
  ('show', show, True, '',
    "user show. Show all the users on this system"),
  ('create', create, False, '^.*\.json$',
   "creates a new user from a json file"),
  ('delete', delete, False, '^[0-9]+$',
   "user delete <id>. Delete the user with a specified ID"),
  ('read', read, True, '',
   "user read. Return metadata of the users in the system in JSON format")
]

"""
cli.plugins.user.get_super_user()

This function returns the first super user from the existing models on the machine.
It is necessary because for some Django functionality (export, audit etc)
We would need the user object which we would normally get from request object
Since it is not possible in CLI (no requests), it returns the first super user

An alternative could be authenticating the user every time before executing a command

"""

def get_super_user():
  users = User.objects.filter(is_superuser=True).filter(is_active=True)
  if len(users) > 0:
    return users[0]

  return None

def plugin_defn():
  """
  Return meta information about the plugin
  """

  defn = {
    'help'    :
      'Commands to work with users',
    'commands': _COMMANDS
  }
  return defn
