# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
cli commands for Reports
"""

import json, os, traceback
import cli.plugins.user

from datetime import datetime
from django.core import serializers
from django.conf import settings

from apps.utils import WHATSNEWVER
from cli.core import table
from cli.utils import read_json
from datasets.dashboard_association import DashboardAssociation
from datasets.models import DataConnection
from migration.migrate import export_objects, import_objects
from reports.models import Report
from reports.audit import auditApp
from vizlib.querylog import ActionType

authenticated_super = False
authenticated_simple = False

def show(argsl):
  """
  visual show

  List all visuals (dashboards) on this system
  """

  rl = Report.objects.all()

  t = table.Table('Data Connections', headers=["ID", "Name", "Type", "Dataset", "Connection"])
  for r in rl:
    t.add_row([r.id, r.report_name, r.report_type, r.dataset.dataset_name, r.dataset.dataconnection.dataconnection_name])
  t.display()

def delete(argsl):
  """
  visual delete <ID>

  Delete a visual (dashboard) with a specified ID
  """

  if not getattr(settings, 'ENABLE_DELETE_VIA_CLI_TOOL', False):
    print('You have no permission to delete via the CLI Tool. You can enable it via the Advanced Settings, please check the documentation')
    return

  try:
    v_id = int(argsl[0])
    report = Report.objects.get(id = v_id)
    print("Please first see the object you are deleting with command read! Are you sure you want to delete this object? Y for yes")
    action = input()
    if action == "Y" or action == "y":
      report.delete()
      auditApp(report, ActionType.DELETE, os.environ.get('HADOOP_USER_NAME', 'CLI user'))
      print("Deletion successful")
  except Exception as e:
    traceback.print_exc()

def read(argsl):
  """
  visual read

  Returns metadata of the visuals (dashboards) on this system in JSON format
  """

  visuals = Report.objects.all()
  visualsl = list()
  for v in visuals:
    serialized_v = json.loads(serializers.serialize('json', [v]))[0]
    visualsl.append(serialized_v)
  out = json.dumps(visualsl, indent=4)
  print(out)

def export_assets(argsl):
  """
  visual export_assets
  Exports all the dashboards and their dependencies

  """
  r = Report.objects.all()
  data = export_objects('dashboards', [i.id for i in r], cli.plugins.user.get_super_user())
  data['version'] = {'Arcviz Version' : WHATSNEWVER}
  with open("visual_export_{0}.json".format(datetime.now()), "w") as outfile:
    json.dump(data, outfile)

def import_assets(argsl):
  """
  visual imports_assets <dataconnection_name> <filename>
  imports all the dashboards and their dependencies from the specified file

  """
  try:
    data = read_json(argsl[1])
  except OSError:
    print(f'File {argsl[1]} not found.\nCheck in the current working directory {os.getcwd()}')
    return
  except Exception as e:
    print(f'JSON is invalid: {str(e)}')

  try:
    dc_name = argsl[0]
    workspace_id = argsl[2]
    dataconnection_obj = DataConnection.objects.get(dataconnection_name=dc_name)
    dataconnection_id = dataconnection_obj.id
    reports = import_objects(data, dataconnection_id, cli.plugins.user.get_super_user(), False, True, None, None, workspace_id)

    datasets = reports.get('datasets')
    dashboards = reports.get('dashboards')

    if datasets and dashboards:
      dashboard_ids = [d['id'] for d in dashboards]
      for dataset in datasets:
        DashboardAssociation().associate(dataset['id'], dashboard_ids)
    print(dashboard_ids)
    return dashboard_ids
  except Exception as e:
    traceback.print_exc()


_COMMANDS = [
  ('show', show, True, '',
    "visual show. Lists all visuals on this system"),
  ('read', read, True, '',
    "visual read. Returns metadata of the visuals on this system in JSON format"),
  ('export_assets', export_assets, True, '',
    "visual export_assets. Exports dashboards and their dependencies and assets to a json file, returns the object"),
  ('import_assets', import_assets, False, '^.+\ .*\..+$',
    "visual import_assets <dataconnection_name> <source_filename> <workspace_id>. Imports dashboards and their dependencies using the dataconnection name from the source file (use export_assets to generate one)"),
  ('delete', delete, False, '^[0-9]+$',
    "visual delete <id>. Delete the visual with a specified ID"),
]

def plugin_defn():
  """
  Return meta information about the plugin
  """

  defn = {
    'help'    :
      'Commands to work with visuals',
    'commands': _COMMANDS
  }
  return defn

