# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json, sys

from datasets.models import DataConnection
from django.contrib.auth.models import User
from django.contrib.auth import authenticate
from getpass import getpass
from jobs.views_encryption import encrypt_secret
from arcwebbase.validator import validatePassword


authenticated_super = False

# read and parse data from file filename
def read_json(filename):
  file = open(filename)
  json_data = json.loads(file.read())
  file.close()
  return json_data

def validate_params_of_a_dataconnection(dc_json):
  validation  = {'sqlite': {'FILE'},
                 'postgres': {'HOST', 'PORT', 'DATABASE'},
                 'mysql': {'HOST', 'PORT', 'DATABASE', 'USERNAME'},
                 'hyve': {'HOST', 'PORT', 'USERNAME'},
                 'impyla': {'HOST', 'PORT'}}
  dc_type = dc_json['dataconnection_type']
  dc_info = dc_json['dataconnection_info']
  try:
    validate_json_fields(dc_info['PARAMS'], validation[dc_type])
  except ValueError as e:
    raise ValueError(e)
  return json.dumps(dc_info)

def json_to_dc_dict(dc_json, filename):
  dc_type = dc_json['dataconnection_type']
  dc_name = dc_json['dataconnection_name'] if 'dataconnection_name' in dc_json else f'{dc_type}-from-{filename}-new-connection'
  dc_description = dc_json['dataconnection_description'] if 'dataconnection_description' in dc_json else f'connection created from {filename}'
  dc_info = validate_params_of_a_dataconnection(dc_json)
  has_password = 'dataconnection_password' in dc_json and dc_json['dataconnection_password'] is not None

  dc_password = json.dumps(encrypt_secret(dc_json['dataconnection_password'])) if has_password else None

  return {'dc_type': dc_type,
          'dc_name': dc_name,
          'dc_description': dc_description,
          'dc_info': dc_info,
          'dc_password': dc_password}

def create_dataconnection(dc_dict):
  dc = DataConnection(
    dataconnection_type=dc_dict['dc_type'],
    dataconnection_name=dc_dict['dc_name'],
    dataconnection_description=dc_dict['dc_description'],
    dataconnection_info=dc_dict['dc_info'],
    dataconnection_password=dc_dict['dc_password'],
    create_user = dc_dict['create_user'],
    update_user = dc_dict['update_user'])
  dc.save()
  print(f'A new data connection with the name "{dc.dataconnection_name}" has been created')

def validate_params_of_a_user(u_json):
  validators = {"username", "password"}
  try:
    validate_json_fields(u_json, validators)
  except ValueError as e:
    raise ValueError(e)
  validatePassword(u_json["password"])
  return u_json

def validate_params_of_a_dataset(d_json):
  validators = {"dataset_info", "dataset_name", "dataset_type", "dataconnection_id"}
  try:
    validate_json_fields(d_json, validators)
  except ValueError as e:
    raise ValueError(e)
  return d_json

def authenticate_user():
  print("Autenticate your user.")
  print("Username:")
  uname = input()
  pwd = getpass()
  user = authenticate(username=uname, password=pwd)
  if user is not None:
    print("Successful authentication")
    return user
  else:
    print("The authentication was not successful")
    sys.exit(1)

def validate_json_fields(json_object, required_fields):
  keys = set(json_object.keys())
  if not required_fields.issubset(keys):
    missing_fields = required_fields.difference(keys)
    raise ValueError(f"Some of the required fields not found in the json: {missing_fields}")