# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from typing import Dict, List
from dataconn.connection_importer import ConnectionImporter
import re
import logging
from dataconn.cml_dependencies import CML_IMPALA, CML_HIVE, get_cmldata, get_project_dataconnections
from datasets.models import DataConnection
from dataconn.cml_connection_updater import CmlConnectionUpdater
from dataconn.connections_types import DATACONN_HYVE, DATACONN_IMPYLA
from urllib.parse import urlparse
import os

LOGGER = logging.getLogger('arcweb')


class CmlConnectionImporter(ConnectionImporter):
    def __init__(self):
        self.cml_connection_types = {
            CML_HIVE: DATACONN_HYVE,
            CML_IMPALA: DATACONN_IMPYLA,
        }

        self.__cmldata = get_cmldata()
        self.username = None
        self.password = None

        try:
            self.username = self.__cmldata.get_hadoop_username()
            self.password = self.__cmldata.get_workload_password()
        except Exception as e:
            LOGGER.warning(
                f'MLExplore: Error while fetching credentials (Exception: {str(e)})')

    def connect(self):
        connections: Dict[List[str]] = {
            'new_connections': [],
            'updated_connections': [],
            'synced_out_connection': [],
        }

        connection_updater = CmlConnectionUpdater()

        if self.__cmldata is None:
            warning = 'Synced stopped because the required cml library was not imported'
            LOGGER.warning(f'MLExplore: {warning}')
            return warning, 500

        if self.password is None:
            LOGGER.warning('MLExplore: Cloudera AI workload password is not set')

        connection_updater.set_unavailabe_connections(
            self.__get_unavailable_connections())

        LOGGER.info('MLExplore: Starting connection sync')

        parsed_connections = self.__get_parsed_connections()

        for parsed_dataconnection in parsed_connections:
            try:
                cml_id = parsed_dataconnection['cml_id']
                dataconnections_in_db = DataConnection.objects.filter(
                    cml_id=cml_id)

                if not dataconnections_in_db.exists():
                    DataConnection.objects.get_or_create(
                        **parsed_dataconnection)
                    connections['new_connections'].append(
                        parsed_dataconnection['dataconnection_name'])
                else:
                    is_conn_updated: bool = connection_updater.update(
                        parsed_dataconnection, dataconnections_in_db[0]
                    )
                    if is_conn_updated:
                        connections['updated_connections'].append(
                            parsed_dataconnection['dataconnection_name'])

            except Exception as e:
                connection_name = parsed_dataconnection['dataconnection_name']
                LOGGER.warning(
                    f'Error happened while saving or updating connection: {connection_name} (Exception: {str(e)})'
                )

        connections['synced_out_connection'].extend(
            connection_updater.sync_out_connections(parsed_connections))

        LOGGER.info('MLExplore: Connection sync finished')

        return connections, 200

    def __parse_cml_data_connection(self, connection):
        try:
            connection_name = connection['name']
            conn_info = connection['connectionInfo']
            connection_type = self.cml_connection_types[connection['type'].lower(
            )]
            cml_id = int(connection['id'])
            conn_properties = self.__get_connection_properties(conn_info)
        except Exception as e:
            LOGGER.warning(
                f'MLExplore: Wrong connection format led to error in parsing the connection info (Exception: {str(e)})'
            )
            return

        use_http_transport = conn_properties.get("USE_HTTP_TRANSPORT", True)
        http_path = conn_properties.get("HTTP_PATH", "cliservice")
        use_ssl = conn_properties.get("USE_SSL", True)

        params = {
            'HOST': conn_properties.get('HOSTNAME'),
            'PORT': conn_properties.get('PORT'),
            'AUTH': 'LDAP',
            'USERNAME': self.username,
            'PASSWORD': self.password,
            'MODE': 'http' if use_http_transport else 'binary',
            'HS2_HTTP_SQLPATH': http_path,
            'SOCK': 'ssl' if use_ssl else None,
        }

        if conn_properties['IS_PRIVATE_CLOUD']:
            if conn_properties.get('USE_KERB_AUTH'):
                params = {
                    'HOST': conn_properties.get('HOSTNAME'),
                    'PORT': conn_properties.get('PORT'),
                    'AUTH': 'KERBEROS',
                    'MODE': 'http' if use_http_transport else 'binary',
                    'HS2_HTTP_SQLPATH': http_path,
                    'SOCK': 'ssl' if use_ssl else None,
                    'KERBSERV': conn_properties.get('KERBEROS_SERVICE_NAME'),
                    'KRB_HOST': conn_properties.get('KRB_HOST')
                }

        jwt_auth_enabled = conn_properties.get("JWT_AUTH_ENABLED")
        if jwt_auth_enabled is not None:
            params = {
                'HOST': conn_properties.get('HOSTNAME'),
                'PORT': conn_properties.get('PORT'),
                'AUTH': 'JWT',
                'MODE': 'http' if use_http_transport else 'binary',
                'HS2_HTTP_SQLPATH': http_path,
                'SOCK': 'ssl' if use_ssl else None,
            }

        params.update({
            'ACCOUNT': '',
            'SETTINGS': {},
            'QUERY_TIMEOUT': 600,
            'SESSION_TIMEOUT': 600,
            'SOCKET_TIMEOUT': 600,
            'IMPERSONATION': False,
            'APP_NAME': 'viz',
        })

        cml_status = {
            'isAvailable': connection['availability'],
            'doesWorkloadPasswordExist': True if self.password is not None else False
        }

        return self.parse_connection(
            connection_name, connection_type, self.password, params, cml_id, cml_status
        )

    def __get_connection_properties(self, conn_info):
        properties = {}
        url = conn_info['jdbc_url']

        jwtAuthStatus = conn_info.get('jwtAuth')
        properties["JWT_AUTH_ENABLED"] = jwtAuthStatus

        is_private_cloud = conn_info.get('isPrivateCloud')
        properties["IS_PRIVATE_CLOUD"] = is_private_cloud

        urlparsed = urlparse(url.replace("jdbc:", ""))
        host = urlparsed.hostname
        if "IMPYLA_HOSTNAME" in os.environ:
            host = os.getenv("IMPYLA_HOSTNAME")

        if host is None:
            raise ValueError("Hostname not available for connection")

        properties["HOSTNAME"] = host
        port = urlparsed.port
        if port is None:
            port = 443
        properties['PORT'] = port

        parsed_params = self.__get_params_from_jdbc(url)

        properties['USE_SSL'] = parsed_params.get('ssl')

        if is_private_cloud:
            self.__parse_pvc_jdbc_params(parsed_params, properties)

        urlpath = urlparsed.path.split(";")
        urlparams = dict((p.split("=")) for p in urlpath if "=" in p)
        http_path = urlparams.get("httpPath", None)
        if http_path is not None:
            properties["HTTP_PATH"] = http_path

        return properties

    def __is_valid_cml_connection(self, connection):
        is_available = connection['availability']
        is_supported_type = connection['type'].lower(
        ) in self.cml_connection_types
        return is_available and is_supported_type

    def __get_parsed_connections(self):
        project_connections = get_project_dataconnections(self.__cmldata)
        valid_connections = filter(
            self.__is_valid_cml_connection, project_connections
        )

        parsed_connections = [
            self.__parse_cml_data_connection(connection)
            for connection in valid_connections
        ]

        return parsed_connections

    def __get_unavailable_connections(self):
        project_connections = get_project_dataconnections(self.__cmldata)
        unavailable_connections = filter(
            lambda x: not x['availability'], project_connections)

        return unavailable_connections

    # From: https://github.infra.cloudera.com/Sense/cloudera-sense/blob/master/services/api/cml/src/cml/data_v1/hiveconnection.py
    # Parses out parameters form jdbc url, example:
    # Input: jdbc:hive2://somehost.x/;ssl=true;transportMode=http;httpPath=somehost/cdp-proxy-api/hive
    # Output: {"ssl": true, "transportMode": "http", "httpPath": "somehost/cdp-proxy-api/hive"}
    def __get_params_from_jdbc(self, jdbc_url):
        urlparsed = urlparse(jdbc_url.replace("jdbc:", ""))
        param_list = urlparsed.path.split(";")
        output = {}
        for param in param_list:
            if "=" in param:
                output[param.split("=")[0]] = param.split("=")[1]
        return output

    # From: https://github.infra.cloudera.com/Sense/cloudera-sense/blob/master/services/api/cml/src/cml/data_v1/hiveconnection.py
    def __parse_pvc_jdbc_params(self, parsedParams, properties):
        # Parse out values from jdbc in private cloud

        # sample impala kerberos
        # 'jdbc:impala://coordinator-impalavirtualwarehouse.apps.cml-gpu-vantage2-ded.kcloud.cloudera.com:443
        # /default;AuthMech=1;
        # transportMode=http;httpPath=cliservice;ssl=1;
        # KrbHostFQDN=dwx-env-e3sr4j-env.cdp.local.;KrbServiceName=hive',

        # sample hive kerberos
        # 'jdbc:hive2://hs2-hivevirtualwarehouse.apps.cml-gpu-vantage2-ded.kcloud.cloudera.com/default;
        # transportMode=http;httpPath=cliservice;socketTimeout=60;ssl=true;
        # retries=3;kerberosEnableCanonicalHostnameCheck=false;
        # principal=hive/dwx-env-e3sr4j-env.cdp.local@ROOT.COMOPS.SIT'
        properties["KERBEROS_SERVICE_NAME"] = parsedParams.get(
            "kerberos_princ_name", "impala"
        )
        properties["USE_HTTP_TRANSPORT"] = (
            parsedParams.get("transportMode", "http") == "http"
        )

        # impala vw: - example rbHostFQDN=dwx-env-e3sr4j-env.cdp.local.
        if not parsedParams.get("KrbHostFQDN") is None:
            properties["KRB_HOST"] = parsedParams.get("KrbHostFQDN").strip(".")
            properties["KERBEROS_SERVICE_NAME"] = parsedParams.get(
                "KrbServiceName", "hive"
            )

        properties["USE_KERB_AUTH"] = parsedParams.get("AuthMech", "3") == "1"

        # hive vw - example principal=hive/dwx-env-e3sr4j-env.cdp.local@ROOT.COMOPS.SIT
        if not parsedParams.get("principal") is None:
            principal = parsedParams.get("principal")
            properties["KRB_HOST"] = principal.split("/")[1].split("@")[0]
            properties["KERBEROS_SERVICE_NAME"] = principal.split("/")[0]

        # Env var overrides
        if "USE_KERB_AUTH" in os.environ:
            properties["USE_KERB_AUTH"] = os.getenv("USE_KERB_AUTH")

        if "PORT" in os.environ:
            properties["PORT"] = os.getenv("PORT")

        if "KRB_HOST" in os.environ:
            properties["KRB_HOST"] = os.getenv("KRB_HOST")

        if "KERBEROS_SERVICE_NAME" in os.environ:
            properties["KERBEROS_SERVICE_NAME"] = os.getenv(
                "KERBEROS_SERVICE_NAME")

        if "USE_HTTP_TRANSPORT" in os.environ:
            properties["USE_HTTP_TRANSPORT"] = os.getenv("USE_HTTP_TRANSPORT")

        if "IS_PRIVATE_CLOUD" in os.environ:
            properties["IS_PRIVATE_CLOUD"] = os.getenv("IS_PRIVATE_CLOUD")

        return properties
