# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import time
import logging
import threading
import re

from sqlrun.views_cache_statistic import CacheStatistics

from . import dataconn_arcengine
from . import dataconn_arcengine_v2
from . import dataconn_dataengine
from . import dataconn_impala
from . import dataconn_impyla
from . import dataconn_hive
from . import dataconn_http
from . import dataconn_hyve
from . import dataconn_sqlite
from . import dataconn_duckdb
from . import dataconn_druid
from . import dataconn_mysql
from . import dataconn_pg
from . import dataconn_sparksql
from . import dataconn_redshift
from . import dataconn_mariadb
from . import dataconn_mongodb
from . import dataconn_odbc
from . import dataconn_aster
from . import dataconn_solr
from . import dataconn_drill
from . import dataconn_oracle
from . import dataconn_mssql
from . import dataconn_ksql
from . import dataconn_bigquery
from . import dataconn_teradata
from . import dataconn_snowflake
from . import dataconn_vertica
from . import dataconn_athena
from . import dataconn_denodo
from . import dataconn_elasticsearch
from . import dataconn_cassandra

from jobs.views_encryption import decrypt_secret
from datasets.models import DataConnection
from .connections_types import DATACONN_ARCENGINE, DATACONN_HIVE, \
  DATACONN_IMPALA, DATACONN_MYSQL, DATACONN_PG, DATACONN_SQLITE, \
  DATACONN_REDSHIFT, DATACONN_MARIADB, DATACONN_SPARKSQL,  \
  DATACONN_MONGODB, DATACONN_ODBC, DATACONN_ASTER, DATACONN_DRILL, \
  DATACONN_ORACLE, DATACONN_SOLR, DATACONN_MSSQL, DATACONN_KSQL, \
  DATACONN_TERADATA, DATACONN_BIGQUERY, DATACONN_SNOWFLAKE, \
  DATACONN_VERTICA, DATACONN_ATHENA, DATACONN_DENODO, DATACONN_ELASTICSEARCH, \
  DATACONN_CASSANDRA, DATACONN_IMPYLA, DATACONN_DUCKDB, DATACONN_DRUID, \
  DATACONN_HYVE, DATACONN_ARCENGINE_V2

from django.conf import settings

LOGGER = logging.getLogger("arcweb")

def getDataConnName(connection_id=-1):
  if connection_id == -1:
    return "new"

  dc = get_django_dataconn(id=int(connection_id))
  return dc.dataconnection_name

def _get_dataconn_params(connection_id=-1, dc=None):

  if dc is None:
    dc = get_django_dataconn(id=int(connection_id))

  conn = {
    'ID': dc.id,
    'TYPE': dc.dataconnection_type,
    'NAME': dc.dataconnection_name,
    'INFO': json.loads(dc.dataconnection_info)
  }

  if dc.dataconnection_password:
    try:
      conn['INFO']['PARAMS']['PASSWORD'] = decrypt_secret(**json.loads(dc.dataconnection_password))
    except Exception as e:
      pass

  return conn

_CONCURRENCY_MIN = 1
_QUERY_TIMEOUT_MINUTES_MIN = 1

# Cleanup the following values:
DEFAULT_SETTING_VALUES = {
  'DATACONNECTION_CONCURRENCY_DEFAULT': 100,
  'DATACONNECTION_CONCURRENCY_MAX': 100,
  'DATACONNECTION_CONCURRENCY_USER_DEFAULT': 5,
  'DATACONNECTION_CONCURRENCY_USER_MAX': 100,
  'DATACONNECTION_QUERY_TIMEOUT_MINUTES_DEFAULT': 10,
  'DATACONNECTION_QUERY_TIMEOUT_MINUTES_MAX': 60,
  'UPLOAD_MAXROWS_MAX': 1000000000,
  'UPLOAD_MAXROWS_DEFAULT': 1000000,
}
def handle_setting_default_max(setting_name):
  default_name = f'{setting_name}_DEFAULT'
  max_name = f'{setting_name}_MAX'
  default_default_name = DEFAULT_SETTING_VALUES.get(default_name)
  default_max_name = DEFAULT_SETTING_VALUES.get(max_name)

  default_value = None
  try:
    default_value = getattr(settings, default_name)
    default_value = int(default_value)
  except:
    LOGGER.warn(f'Failed to parse {default_value} to float. Setting {default_name} to {default_default_name}')
    default_value = default_default_name

  max_value = None
  try:
    max_value = getattr(settings, max_name)
    max_value = int(max_value)
  except:
    LOGGER.warn(f'Failed to parse {default_value} to float. Setting {max_name} to {default_max_name}')
    max_value = default_max_name

  if default_value != None and max_value != None:
    setattr(settings, default_name, min(default_value, max_value))
    if getattr(settings, default_name) != default_value:
      LOGGER.warn(f'Set {default_name} to maximum value instead')

  elif default_value == None:
    setattr(settings, default_name, min(default_default_name, max_value))
    LOGGER.warn(f'No default value {default_name}')

  elif max_value == None:
    setattr(settings, max_name,  min(default_value, default_max_name))
    LOGGER.warn(f'No max value {default_name}')

  else:
    setattr(settings, default_name, default_default_name)
    setattr(settings, max_name, default_max_name)

handle_setting_default_max('DATACONNECTION_CONCURRENCY')
handle_setting_default_max('DATACONNECTION_CONCURRENCY_USER')
handle_setting_default_max('DATACONNECTION_QUERY_TIMEOUT_MINUTES')
handle_setting_default_max('UPLOAD_MAXROWS')

# How many queries should we run at a time against this data connection
def getDataConnConcurrency(connection_id=None):
  if connection_id == -1:
    # Connection does not exist - this is a test for a new connection
    return int(settings.DATACONNECTION_CONCURRENCY_DEFAULT)

  params = _get_dataconn_params(connection_id=connection_id)

  c = None
  try:
    c = int(params.get('INFO', {}) \
                  .get('PARAMS', {}). \
                  get('CONCURRENCY', settings.DATACONNECTION_CONCURRENCY_DEFAULT))
  except Exception as e:
    LOGGER.exception(e)

  if not c:
    try:
      c = int(settings.DATACONNECTION_CONCURRENCY_DEFAULT)
    except Exception as e:
      LOGGER.exception(e)

  if c < _CONCURRENCY_MIN:
    return _CONCURRENCY_MIN

  if c > settings.DATACONNECTION_CONCURRENCY_MAX:
    return settings.DATACONNECTION_CONCURRENCY_MAX

  return c

# How many queries should a single user be able to run at a time against this data connection
def getDataConnConcurrencyUser(connection_id=None):
  if connection_id == -1:
    # Connection does not exist - this is a test for a new connection
    return int(settings.DATACONNECTION_CONCURRENCY_USER_DEFAULT)

  params = _get_dataconn_params(connection_id=connection_id)
  c = None
  try:
    c = int(params.get('INFO', {}) \
                  .get('PARAMS', {}). \
                  get('CONCURRENCY_USER', settings.DATACONNECTION_CONCURRENCY_USER_DEFAULT))
  except Exception as e:
    LOGGER.exception(e)

  if not c:
    try:
      c = int(settings.DATACONNECTION_CONCURRENCY_USER_DEFAULT)
    except Exception as e:
      LOGGER.exception(e)

  if c < _CONCURRENCY_MIN:
    return _CONCURRENCY_MIN

  if c > settings.DATACONNECTION_CONCURRENCY_USER_MAX:
    return settings.DATACONNECTION_CONCURRENCY_USER_MAX

  return c

def getDataConnQueryTimeoutMinutes(connection_id=None):
  params = _get_dataconn_params(connection_id=connection_id)
  c = None
  try:
    c = int(params.get('INFO', {}) \
                  .get('PARAMS', {}) \
                  .get('QUERY_TIMEOUT', settings.DATACONNECTION_QUERY_TIMEOUT_MINUTES_DEFAULT))
  except Exception as e:
    LOGGER.exception(e)

  if not c:
    try:
      c = int(settings.DATACONNECTION_QUERY_TIMEOUT_MINUTES_DEFAULT)
    except Exception as e:
      LOGGER.exception(e)

  if c < _QUERY_TIMEOUT_MINUTES_MIN:
    return _QUERY_TIMEOUT_MINUTES_MIN

  if c > settings.DATACONNECTION_QUERY_TIMEOUT_MINUTES_MAX:
    return settings.DATACONNECTION_QUERY_TIMEOUT_MINUTES_MAX

  return c

def get_dataconn_properties(connection_id=-1):
  return get_dataconn(connection_id=connection_id).properties()

def supports_dataextract_target(connection_type: str):
  try:
    props = get_dataconn_type_properties(connection_type)
    return props.supports_dataextract_target
  except Exception as e:
    LOGGER.error(e)
    return False

def get_dataconn_type_properties(connection_type=None):
  if connection_type == DATACONN_ARCENGINE:
    return dataconn_arcengine.ArcEngineConnProperties()

  if connection_type == DATACONN_ARCENGINE_V2:
    return dataconn_arcengine_v2.ArcEngineV2ConnProperties()

  if connection_type == DATACONN_IMPALA:
    return dataconn_impala.ImpalaConnProperties()

  if connection_type == DATACONN_IMPYLA:
    return dataconn_impyla.ImpylaConnProperties()

  if connection_type == DATACONN_HIVE:
    return dataconn_hive.HiveConnProperties()

  if connection_type == DATACONN_HYVE:
    return dataconn_hyve.HyveConnProperties()

  if connection_type == DATACONN_SQLITE:
    return dataconn_sqlite.SqliteConnProperties()

  if connection_type == DATACONN_DUCKDB:
    return dataconn_duckdb.DuckDBConnProperties()

  if connection_type == DATACONN_DRUID:
    return dataconn_druid.DruidConnProperties()

  if connection_type == DATACONN_MYSQL:
    return dataconn_mysql.MysqlConnProperties()

  if connection_type == DATACONN_MSSQL:
    return dataconn_mssql.MsSqlConnProperties()

  if connection_type == DATACONN_MARIADB:
    return dataconn_mariadb.MariaDBConnProperties()

  if connection_type == DATACONN_PG:
    return dataconn_pg.PgConnProperties()

  if connection_type == DATACONN_REDSHIFT:
    return dataconn_redshift.RedshiftConnProperties()

  if connection_type == DATACONN_SPARKSQL:
    return dataconn_sparksql.SparkSQLConnProperties()

  if connection_type == DATACONN_MONGODB:
    return dataconn_mongodb.MongoConnProperties()

  if connection_type == DATACONN_ODBC:
    return dataconn_odbc.OdbcConnProperties()

  if connection_type == DATACONN_DRILL:
    return dataconn_drill.DrillConnProperties()

  if connection_type == DATACONN_ORACLE:
    return  dataconn_oracle.OracleConnProperties()

  if connection_type == DATACONN_SOLR:
    return dataconn_solr.SolrConnProperties()

  if connection_type == DATACONN_ASTER:
    return dataconn_aster.AsterConnProperties()

  if connection_type == DATACONN_KSQL:
    return dataconn_ksql.KsqlConnProperties()

  if connection_type == DATACONN_TERADATA:
    return dataconn_teradata.TeradataConnProperties()

  if connection_type == DATACONN_BIGQUERY:
    return dataconn_bigquery.BigQueryConnProperties()

  if connection_type == DATACONN_SNOWFLAKE:
    return dataconn_snowflake.SnowflakeConnProperties()

  if connection_type == DATACONN_VERTICA:
    return dataconn_vertica.VerticaConnProperties()

  if connection_type == DATACONN_DENODO:
    return dataconn_denodo.DenodoConnProperties()

  if connection_type == DATACONN_ELASTICSEARCH:
    return dataconn_elasticsearch.ElasticsearchConnProperties()

  if connection_type == DATACONN_CASSANDRA:
    return dataconn_cassandra.CassandraConnProperties()

  if connection_type == 'HTTP':
    return dataconn_http.HTTPConnProperties()

  from dataconn.ext.main import create_properties
  properties = create_properties(connection_type)
  if properties:
    return properties

  raise Exception('Could not identify connection engine %s' % \
      connection_type)

#
# We maintain a hash of connection ids to connections. We protect this hash
# with the _DATACONN_TABLE_LOCK below
#
# We additionally use a per-connection-id lock so we only make one attempt to
# connect to a data connection that supports reuse at a time. These locks are
# stored in _DATACONN_TABLE_CONNECTION_LOCK, and _DATACONN_TABLE_LOCK is used
# to protect access to this hash as well
#
# Note that we always grab connection locks first, and then _DATACONN_TABLE_LOCK
#
_DATACONN_TABLE = {}
_DATACONN_TABLE_CONNECTION_LOCK = {}
_DATACONN_TABLE_LOCK = threading.Lock()

def del_dataconn(connection_id=-1):
  connection_id = int(connection_id)
  try:
    del _DATACONN_TABLE[connection_id]
  except Exception:
    pass

# Get the data connection driver object from the django connection id
def get_dataconn(connection_id=-1, django_dc=None) -> dataconn_dataengine.DataEngine:
  try:
    return _get_dataconn(connection_id=connection_id, django_dc=django_dc)
  except Exception as e:
    error_message = f'Cannot connect to data connection {connection_id}'
    LOGGER.exception(error_message)
    raise Exception(error_message)

def _get_connection_lock(connection_id):
  with _DATACONN_TABLE_LOCK:
    connection_lock = _DATACONN_TABLE_CONNECTION_LOCK.get(connection_id)
    if not connection_lock:
      connection_lock = threading.Lock()
      _DATACONN_TABLE_CONNECTION_LOCK[connection_id] = connection_lock

  return connection_lock

def get_django_dataconn(id) -> DataConnection:
  if id == dataconn_http.INSTANCE.id:
    django_dc = dataconn_http.INSTANCE
  else:
    django_dc = DataConnection.objects.get(id=id)
  return django_dc

def _get_dataconn(connection_id=-1, django_dc=None):
  connection_id = int(connection_id)
  if not django_dc:
    django_dc = get_django_dataconn(connection_id)

  # LOGGER.info('Waiting for lock')
  with _DATACONN_TABLE_LOCK:
    # LOGGER.info('Done waiting for lock')
    dcinfo = _get_dataconn_delete_expired(connection_id, django_dc)
    if dcinfo:
      return dcinfo['dataconn']

  # No stored dataconn, build it

  # First, check if we reuse data connections for this connection type. If so,
  # protect against multiple threads fetching a data connection at the same time
  connection_lock = None
  conn_props = get_dataconn_type_properties(connection_type=django_dc.dataconnection_type)
  if conn_props.supports_reuse:
    connection_lock = _get_connection_lock(connection_id)
    connection_lock.acquire()

  try:
    if connection_lock:
      # Within connection_lock, see if some other thread beat us to it...
      with _DATACONN_TABLE_LOCK:
        dcinfo = _get_dataconn_delete_expired(connection_id, django_dc)
        if dcinfo:
          return dcinfo['dataconn']

    # Build a new data connection and store it away
    conn = _build_dataconn(connection_id=connection_id, dc=django_dc)
    dcinfo = {
      'dataconn': conn,
      'last_update': django_dc.update_time,
      'last_recheck': time.time()
    }

    if conn.properties().supports_reuse:
      with _DATACONN_TABLE_LOCK:
        _DATACONN_TABLE[connection_id] = dcinfo

  finally:
    if connection_lock:
      connection_lock.release()

  return dcinfo['dataconn']

def _get_dataconn_delete_expired(connection_id, django_dc):
  dcinfo = _DATACONN_TABLE.get(connection_id)
  now = time.time()

  # Ensure existing dataconn in sync with django dataconn model
  if dcinfo:
    if now - dcinfo['last_recheck'] > getattr(settings, 'DATACONN_RECHECK_SECONDS', 10):
      if dcinfo['last_update'] != django_dc.update_time:
        del_dataconn(connection_id)
        dcinfo = None
      else:
        dcinfo['last_recheck'] = now

  return dcinfo

def validate_dataconn(user, dc):
  conn = _build_dataconn(dc=dc)

  comment = None

  # Solr does not have database concept, the database is
  # defaulted to Solr in the code.
  if dc.dataconnection_type == DATACONN_SOLR or \
      dc.dataconnection_type == DATACONN_ELASTICSEARCH:
    dsreq = {'version': 1, 'type': '.tables'}
  else:
    dsreq = {'version': 1, 'type': '.databases'}
  from vizlib.dataset_request import DatasetRequest
  dataset_request = DatasetRequest(json.dumps(dsreq))

  build_request = conn.build(user.username, user.id, None,
                             dataset_request=dataset_request,
                             comment=comment)

  # Issue databases command to validate the connection.
  CacheStatistics.updateQueryStatisticHistory(dc.id, False)
  conn.execute(build_request)

def _build_dataconn(connection_id=-1, dc=None):
  connection_params = _get_dataconn_params(connection_id=connection_id, dc=dc)
  info_params = connection_params['INFO']['PARAMS']
  connection_type = connection_params['TYPE']
  conn = None
  if connection_type == DATACONN_ARCENGINE:
    conn = dataconn_arcengine.ArcEngineConn(info_params, connection_id)

  if connection_type == DATACONN_ARCENGINE_V2:
    conn = dataconn_arcengine_v2.ArcEngineV2Conn(info_params, connection_id)

  elif connection_type == DATACONN_IMPALA:
    conn = dataconn_impala.ImpalaConn(info_params, connection_id)

  elif connection_type == DATACONN_IMPYLA:
    conn = dataconn_impyla.ImpylaConn(info_params, connection_id)

  elif connection_type == DATACONN_HIVE:
    conn = dataconn_hive.HiveConn(info_params, connection_id)

  elif connection_type == DATACONN_HYVE:
    conn = dataconn_hyve.HyveConn(info_params, connection_id)

  elif connection_type == DATACONN_SPARKSQL:
    conn = dataconn_sparksql.SparkSQLConn(info_params, connection_id)

  elif connection_type == DATACONN_SQLITE:
    conn = dataconn_sqlite.SqliteConn(info_params, connection_id)

  elif connection_type == DATACONN_DUCKDB:
    conn = dataconn_duckdb.DuckDBConn(info_params, connection_id)

  elif connection_type == DATACONN_DRUID:
    conn = dataconn_druid.DruidConn(info_params, connection_id)

  elif connection_type == DATACONN_MYSQL:
    conn = dataconn_mysql.MysqlConn(info_params, connection_id)

  elif connection_type == DATACONN_MSSQL:
    conn = dataconn_mssql.MsSqlConn(info_params, connection_id)

  elif connection_type == DATACONN_MARIADB:
    conn = dataconn_mariadb.MariaDBConn(info_params, connection_id)

  elif connection_type == DATACONN_PG:
    conn = dataconn_pg.PgConn(info_params, connection_id)

  elif connection_type == DATACONN_REDSHIFT:
    conn = dataconn_redshift.RedshiftConn(info_params, connection_id)

  elif connection_type == DATACONN_MONGODB:
    conn = dataconn_mongodb.MongoConn(info_params, connection_id)

  elif connection_type == DATACONN_ODBC:
    conn = dataconn_odbc.OdbcConn(info_params, connection_id)

  elif connection_type == DATACONN_ASTER:
    conn = dataconn_aster.AsterConn(info_params, connection_id)

  elif connection_type == DATACONN_SOLR:
    conn = dataconn_solr.SolrConn(connection_params['INFO']['PARAMS'],
        connection_id)

  elif connection_type == DATACONN_DRILL:
    conn = dataconn_drill.DrillConn(info_params, connection_id)

  elif connection_type == DATACONN_ORACLE:
    conn = dataconn_oracle.OracleConn(info_params, connection_id)

  elif connection_type == DATACONN_KSQL:
    conn = dataconn_ksql.KsqlConn(info_params, connection_id)

  elif connection_type == DATACONN_TERADATA:
    conn = dataconn_teradata.TeradataConn(info_params, connection_id)

  elif connection_type == DATACONN_BIGQUERY:
    conn = dataconn_bigquery.BigQueryConn(info_params, connection_id)

  elif connection_type == DATACONN_SNOWFLAKE:
    conn = dataconn_snowflake.SnowflakeConn(info_params, connection_id)

  elif connection_type == DATACONN_ATHENA:
    conn = dataconn_athena.AthenaConn(info_params, connection_id)

  elif connection_type == DATACONN_VERTICA:
    conn = dataconn_vertica.VerticaConn(info_params, connection_id)

  elif connection_type == DATACONN_DENODO:
    conn = dataconn_denodo.DenodoConn(info_params, connection_id)

  elif connection_type == DATACONN_ELASTICSEARCH:
    conn = dataconn_elasticsearch.ElasticsearchConn(info_params, connection_id)

  elif connection_type == DATACONN_CASSANDRA:
    conn = dataconn_cassandra.CassandraConn(info_params, connection_id)

  elif connection_type == 'HTTP':
    conn = dataconn_http.HTTPConn(info_params, connection_id)

  if not conn:
    from dataconn.ext.main import create_connection
    conn = create_connection(connection_type, info_params, connection_id)

  if not conn:
    raise Exception('Could not identify connection engine %s' % \
        connection_type)

  conn.connection_name = connection_params['NAME']
  return conn

RIGHT_BRACKET_COLNAME = re.compile(r"\)")
INVALID_CHARS = re.compile(r"[^0-9a-z_]")

# remove invalid characters: "SUM(PRICE)" will become "sum_price" via these transformations
def sanitize_colname(colname, idx=0):
  if colname is None:
    return f"column_{idx}"

  colname = colname.lower()

  colname = RIGHT_BRACKET_COLNAME.sub('', colname)
  colname = (INVALID_CHARS.sub('_', colname))

  return colname

def sanitize_headers(headers, connection_type=None):
  """ Generate headers if none exist or make existing ones HDFS compatible """
  headernames = [sanitize_colname(colname, idx) for idx, colname in enumerate(headers)]
  return headers_make_unique(headernames, max_length=100)


# adapted from messytables, that has a bug: https://github.com/okfn/messytables/issues/197
def headers_make_unique(headers, max_length=None):
    """Make sure the header names are unique. For non-unique
    columns, append 1, 2, 3, ... after the name. If max_length
    is set, truncate the original string so that the headers are
    unique up to that length."""

    headers = [h.strip() for h in headers]

    new_digits_length = 0
    while not max_length or new_digits_length <= max_length:
        # If maxlength is 63 and we expect 1 digit for appending
        # numerals to make the headers unique, then truncate the
        # column names to 62 characters.
        _headers = headers
        if max_length:
            _headers = [h[0:max_length - new_digits_length].strip()
                        for h in headers]

        # For headers that are not unique, add a number to the end.
        header_counter = {}
        # VIZ change start
        new_headers = []
        for h in _headers:
            h_unique = h
            while _headers.count(h_unique) > 1 or new_headers.count(h_unique) > 0:
                header_counter[h] = header_counter.get(h, 0) + 1
                h_unique = "%s_%d" % (h, header_counter[h])
            new_headers.append(h_unique)
            # VIZ change end

        # If there is a max_length but adding a counter made a header longer
        # than max_length, we have to truncate more up front (which may
        # change which headers are nonunique) and try again.
        if max_length and (True in [len(h) > max_length for h in new_headers]):
            # Adding this counter made the new header longer than max_length.
            # We have to truncate the original headers more.
            new_digits_length += 1
            continue

        # Otherwise, the new headers are unique.
        return new_headers

    raise ValueError('''max_length is so small that the column names cannot
        be made unique.''')
