# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import sys
import json
import re
import logging

from django.conf import settings

from .dataconn_dataengine import ConnProperties, DataEngine, BuildRequestExecutor
from .dataconn_consts import DCS
from .connections_types import DATACONN_ARCENGINE, DATACONN_HIVE, \
    DATACONN_IMPALA, DATACONN_SPARKSQL, DATACONN_IMPYLA, DATACONN_HYVE, \
    DATACONN_ARCENGINE_V2
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS

import jobs

from vizlib.querylog import QueryLog
from vizlib.querylog_helper import ComponentType
from vizlib.arcprof import ArcProf
from . import arctypes

from .thrift_hs2.TCLIService.ttypes import TGetServerLogReq, TGetLicenseInfoReq

# create logger
LOGGER = logging.getLogger("arcweb")

from util.perfcounters import APPROX_COUNTERS

arcengine_create = APPROX_COUNTERS.get('arcengine.create')

KERBSERV_TABLE = {
  DATACONN_ARCENGINE : 'arcengine',
  DATACONN_ARCENGINE_V2 : 'arcengine',
  DATACONN_IMPALA : 'impala',
  DATACONN_IMPYLA : 'impala',
  DATACONN_HIVE : 'hive',
  DATACONN_HYVE : 'hive'
}

_TABLENAME_RE = re.compile(r"^\S+$")

# For tests, survive not having access to all the thrift dependencies in hs2conn
try:
  from . import hs2conn
except Exception as e:
  pass

class ArcEngineConnProperties(ConnProperties):

  def __init__(self):
    super(ArcEngineConnProperties, self).__init__()
    self.import_data_insertrows = 10000
    self.supports_deriveddata = True
    self.supports_tableproperties = True
    self.supports_nulls_order = True
    self.supports_truncate_table = True

    self.display_to_table_type = {
      "String": "string",
      "Integer": "bigint",
      "Real": "double",
      "Boolean": "boolean",
      "Timestamp": "timestamp"
    }

class ArcEngineConn(DataEngine):

  TYPEMAP = {

    'INT': arctypes.ARC_DATATYPE_INT,
    'BINARY': arctypes.ARC_DATATYPE_INT,
    'SMALLINT': arctypes.ARC_DATATYPE_INT,
    'TINYINT': arctypes.ARC_DATATYPE_INT,
    'BIGINT': arctypes.ARC_DATATYPE_BIGINT,

    'DOUBLE': arctypes.ARC_DATATYPE_DOUBLE,
    'FLOAT': arctypes.ARC_DATATYPE_DOUBLE,
    'DECIMAL': arctypes.ARC_DATATYPE_DOUBLE,

    'CHAR': arctypes.ARC_DATATYPE_STRING,
    'VARCHAR': arctypes.ARC_DATATYPE_STRING,
    'STRING': arctypes.ARC_DATATYPE_STRING,

    'DATE': arctypes.ARC_DATATYPE_TIMESTAMP,
    'TIMESTAMP': arctypes.ARC_DATATYPE_TIMESTAMP,

    'BOOLEAN': arctypes.ARC_DATATYPE_BOOLEAN,

    'MAP': arctypes.ARC_DATATYPE_STRING,
    'UNION': arctypes.ARC_DATATYPE_STRING,
    'ARRAY': arctypes.ARC_DATATYPE_STRING,
    'STRUCT': arctypes.ARC_DATATYPE_STRING,
  }

  def __init__(self, params, connection_id=-1, conn_type=DATACONN_ARCENGINE):
    global arcengine_create
    arcengine_create += 1

    self.aviews_extended = 'extended '
    self.host = params['HOST']
    port = params['PORT']
    auth = params.get('AUTH', 'Plain' if conn_type == 'hive' else 'NoSasl')
    self.username = params.get('USERNAME', '_default')
    self.password = params.get('PASSWORD', '_default')
    ssl = params.get('SOCK', '').upper() in ['SSL', 'SSLCERT']
    sslcert = ''
    if params.get('SOCK', '').upper() in ['SSLCERT']:
      sslcert = getattr(settings, 'SSL_CERT_PATH', '')
    kerbserv = params.get('KERBSERV', '')
    if len(kerbserv) == 0:
      kerbserv = KERBSERV_TABLE.get(conn_type)
    self.connection_id = connection_id
    self.connection_type = conn_type
    self.impersonation = params.get('IMPERSONATION', False)
    self.impersonation_trusted = params.get('IMPERSONATION_TRUSTED', False)
    mode = params.get('MODE', 'binary')
    if mode.upper() == 'HTTP':
      http_service = params.get('HS2_HTTP_SQLPATH') or 'cliservice'
      http_access_token = params.get('HS2_HTTP_ACCESS_TOKEN')
    else:
      http_service = None
      http_access_token = None

    # Default to True here so that old connections without this parameter
    # will continue to work since we had the cert checking disabled for
    # a while.
    sslcert_name_mismatch = params.get('SSLCERT_NAME_MISMATCH', True)

    # Merge possible custom settings with defaults
    self.conn_settings = self.default_connection_settings().copy()
    self.conn_settings.update(params.get('SETTINGS', {}))

    try:
      qdepth = int(params.get('QUEUE_DEPTH', 0))
    except Exception:
      qdepth = 0
    if qdepth < 1 or qdepth > 100:
      qdepth = 40

    try:
      timeout = int(params.get('TIMEOUT', 0))
    except Exception:
      timeout = 0
    if timeout < 10 or timeout > 300:
      if not getattr(settings, 'DJANGO_TEST_ENV', False):
        timeout = 60

    try:
      session_timeout = int(params.get('SESSION_TIMEOUT', 0))
    except Exception:
      session_timeout = 0
    if session_timeout < 0 or session_timeout > 3600:
      if not getattr(settings, 'DJANGO_TEST_ENV', False):
        session_timeout = 0

    try:
      socket_timeout = int(params.get('SOCKET_TIMEOUT', 0))
    except Exception:
      socket_timeout = 0
    if socket_timeout < 0 or socket_timeout > 3600:
      if not getattr(settings, 'DJANGO_TEST_ENV', False):
        socket_timeout = 0

    conn_config = {'max_connections':40,
                   'max_requests':40,
                   'max_reqperconn':1}
    if self.impersonation or self.impersonation_trusted:
      if not getattr(settings, 'HS2CONN_ONE_REQUEST_PER_CONN', False):
        conn_config = {'max_connections':40,
                       'max_requests':400,
                       'max_reqperconn':40}

    if conn_type == DATACONN_HIVE:
      conn_config = {
        'max_connections': qdepth,
        'max_requests': qdepth,
        'max_reqperconn': 1
      }

    # Allow queue config override
    conn_config = getattr(settings, 'HS2_QUEUE_SETTINGS', conn_config)

    if params.get('FOR_TESTING'):
      return

    pure_sasl = sys.platform == 'win32' or \
                getattr(settings, 'USE_PURE_PYTHON_SASL', False)

    self.hs2conn = hs2conn.HS2Pool(host=self.host, port=port, auth=auth,
                                   username=self.username,
                                   password=self.password,
                                   ssl=ssl, sslcert=sslcert, kerbserv=kerbserv,
                                   conntype=self.connection_type,
                                   connid=self.connection_id,
                                   settings=self.conn_settings,
                                   timeout=timeout,
                                   conn_timeout=session_timeout,
                                   sslcert_name_mismatch=sslcert_name_mismatch,
                                   pure_sasl=pure_sasl,
                                   http_service=http_service,
                                   http_access_token=http_access_token,
                                   socket_timeout=socket_timeout,
                                   impersonation_trusted=self.impersonation_trusted,
                                   **conn_config)

  def default_connection_settings(self):
    return {
      'EXEC_SINGLE_NODE_ROWS_THRESHOLD': '200000',
      'OPTIMIZE_PARTITION_KEY_SCANS': '1'
    }

  # Run a 'show [extended] analytical views' type command with checking
  # old vs new arcengine then adjust+re-run if needed.
  # We do this by detecting mis-match and switching what command we
  # use to get the aview data.
  def _run_view_command(self, viewfunc):
    try:
      rsp = viewfunc()
      if not self.aviews_extended and len(rsp.get('colnames')) == 1:
        self.aviews_extended = 'extended '
        rsp = viewfunc()
    except Exception as e:
      if self.aviews_extended and str(e).find('Encountered: EXTENDED') != -1:
        self.aviews_extended = ''
        rsp = viewfunc()
      else:
        raise
    return rsp

  # Get view state for one analytical view
  def show_analytical_view(self, viewname, username=None):
    def viewfunc():
      return self._show_analytical_view_internal(viewname, username=username)
    return self._run_view_command(viewfunc)

  def _show_analytical_view_internal(self, viewname, username=None):
    prefix = 'show %sanalytical views ' % self.aviews_extended
    parts = viewname.split('.')
    if len(parts) == 2:
      suffix = 'in `%s` like "%s"' % (parts[0], viewname)
    else:
      suffix = 'like "%s"' % viewname
    return self.sync_runquery(prefix + suffix, username=username)

  def _execute_show_analytical_views(self, build_request):
    def viewfunc():
      return self._execute_show_analytical_views_internal(build_request)
    return self._run_view_command(viewfunc)

  def _execute_show_analytical_views_internal(self, build_request):
    # In secure environments, 'show all analytical views' does not
    # work correctly. Allow a setting for enumerating through dbs individually
    # and ignoring any security (or other issues) in those cases.
    # Once 'show all analytical views' works in secure environments, we
    # can remove support for the 'SHOW_ALL_AVIEWS_ENUMERATES_DBS' setting
    if settings.SHOW_ALL_AVIEWS_ENUMERATES_DBS:
      # First, fetch the list of databases. Then, enumerate views in each db,
      # ignoring any errors (likely security-related) in the process
      query = 'show databases'
      executor = BuildRequestExecutor(build_request)
      dbs = executor.execute(query=query, async_flag=False)
      views = []

      rows_parsed = json.loads(dbs['rows'])
      for row in rows_parsed:
        db = row[0]
        query = "show %sanalytical views in `%s`" % (self.aviews_extended, db)
        executor = BuildRequestExecutor(build_request)
        try:
          v = executor.execute(query=query, async_flag=False)
          if not views:
            views = v
            # Check here as well so we don't incorrectly iterate over
            # all DBs only to do it again with 'extended' keyword.
            if not self.aviews_extended and len(v.get('colnames')) == 1:
              break
          else:
            views['rows'].extend(v['rows'])
        except Exception as e:
          if str(e).find('Encountered: EXTENDED') != -1:
            raise
          # print 'Ignoring exception', e
      return views
    else:
      query = 'show %sall analytical views' % self.aviews_extended
      executor = BuildRequestExecutor(build_request)
      return executor.execute(query=query, async_flag=False)

  def _execute_drop_analytical_views(self, build_request, async_flag, stream):
    # This function deletes jobtargets relating to an analytical view
    # before executing the query to delete an analytical view.
    try:
      view_info = build_request.dataset_request.dsreq.viewname.split('.')
      # Strip the enquoting here
      db_name = view_info[0][1:-1]
      view_name = view_info[1][1:-1]
      complete_name = db_name + '.' + view_name
    except:
      error_message = 'Error parsing viewname'
      LOGGER.exception(error_message)
      data = { 'error_msg': error_message }
      return data

    models = jobs.models
    targets = models.JobTarget.objects.filter(target_type='R')
    for target in targets:
      target_data = json.loads(target.target_data)
      if (target_data['dcid'] == self.connection_id and complete_name == target_data['view']):
        # Right now, a schedule can only contain one RefreshJobTarget, so we delete the target
        # and it's associated schedule. If it becomes possible that a JobSchedule can hold
        # RefreshJobTargets and other JobTargets (ie we implement dependencies on RefreshJobTarget,
        # than we need to change this line)
        target.schedule.delete()

    query = 'drop analytical view ' + build_request.builder.dataset_request.dsreq.viewname
    executor = BuildRequestExecutor(build_request)
    return executor.execute(query=query, async_flag=async_flag, stream=stream)

  # execute query(ies)
  def execute(self, build_request, async_flag=False, stream=False):
    query = None

    if build_request.query == SPECIAL_CMDS['REFRESH_AVIEWS']:
      query = 'refresh analytical view ' + build_request.builder.dataset_request.dsreq.viewname
    elif build_request.query == SPECIAL_CMDS['DROP_AVIEWS']:
      return self._execute_drop_analytical_views(build_request, async_flag, stream)
    elif build_request.query == SPECIAL_CMDS['SHOW_AVIEWS']:
      return self._execute_show_analytical_views(build_request)
    elif build_request.query == SPECIAL_CMDS['COMPUTE_STATS']:
      table_name = build_request.builder.dataset_request.dsreq.table_name
      tables = (self.quote_tablename(table_name) if table_name else "")
      if self.connection_type == DATACONN_HIVE:
        query = f'analyze table {tables} compute statistics'
      else:
        query = f'compute incremental stats {tables}'
    elif build_request.query == SPECIAL_CMDS['INVALIDATE_METADATA']:
      table_name = build_request.builder.dataset_request.dsreq.table_name
      query = 'invalidate metadata ' + (self.quote_tablename(table_name) if table_name else "")
    elif build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = 'show databases'
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      query = 'show tables in ' + build_request.query.split()[-1]
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            # if exception is caught above we set state to unknown
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)
    return executor.execute(query=query, async_flag=async_flag, stream=stream)

  def properties(self):
    return ArcEngineConnProperties()

  def close(self):
    pass

  def _get_tables_text(self, response, key='Tables: '):
    """
    Extract the specified text string from the 'info' field of query response
    """
    try:
      for s in response.get('info', []):
        if s.startswith(key):
          return s

    except Exception as e:
      pass

    # If we cannot find the right key, do not permit caching by returning None
    return None

  def query_results_key(self, username=None, query=None, dsreq=None,
      results=None, client_ip=None):

    if self.connection_type != DATACONN_ARCENGINE:
      return None

    if results is None:
      explain_output = self.log_sync_runquery('explain ' + query,
          username=username, client_ip=client_ip, match_views=True)
    else:
      explain_output = results
    return self._get_tables_text(explain_output)

  # Use explain to get the tables referenced by this query
  def query_table_refs(self, username=None, query=None, results=None, client_ip=None):
    if results is None:
      explain_output = self.log_sync_runquery('explain ' + query,
          username=username, client_ip=client_ip)
    else:
      explain_output = results
    key = 'TableRefs: '
    refs = self._get_tables_text(explain_output, key)
    if refs:
      refs = set(refs[len(key):].split(','))
    return refs

  def table_columns(self, table_name, username=None, client_ip=None):
    partitions = []
    hive_comments = []
    if _TABLENAME_RE.match(table_name.strip()):
      (partitions, hive_comments) = self._partitions_and_comments(table_name, username=username,
          client_ip=client_ip)

    out = self._columns(table_name, username=username, client_ip=client_ip)
    if not out:
      raise Exception("Could not fetch columns for %s" % table_name)

    # Add in partitioning information
    out['partitions'] = partitions
    out['hive_comments'] = hive_comments
    return out

  def _partitions_and_comments(self, table_name, username=None, client_ip=None):
    out = self.log_sync_runquery(
        "describe formatted %s" % self.quote_tablename(table_name),
        username=username,
        client_ip=client_ip)
    return self._describe_output_to_partitions_and_comments(out)

  def _columns(self, table_name, username=None, client_ip=None):
    query = "describe %s" % self.quote_tablename(table_name)
    rsp = self.log_sync_runquery(query,
        username=username,
        client_ip=client_ip)

    # Split but handle case where comma can be embedded in <>, ()
    def split_comma(text):
      input = text.split(',')
      olist = []
      item = ''
      while len(input):
        item += input.pop(0)
        if item.count('<') != item.count('>') or \
           item.count('(') != item.count(')'):
          item += ','
        else:
          olist.append(item)
          item = ''
      return olist

    # Populate cols list with tuple for each element within a possibly
    # complex type defintion.  Couple format: (name, type, parent)
    def get_type(name, text, parent, cols):

      type, lt, text = text.partition('<')
      if not lt:
        cols.append((name, type, parent))
        return

      # Hive translate map and array types to string
      if self.connection_type not in [DATACONN_ARCENGINE, DATACONN_IMPALA, DATACONN_IMPYLA, \
            DATACONN_ARCENGINE_V2]:
        if type in ['map', 'array']:
          type = 'string'
          cols.append((name, 'string', parent))
          return

      text = text[:-1]
      cols.append((name, type, parent))
      parent = name

      if type == 'map':
        (ktype, comma, vtype) = text.partition(',')
        cols.append((name + '.key', ktype, parent))
        get_type(name + '.value', vtype, parent, cols)
      elif type == 'array':
        cols.append((name + '.pos', 'int', parent))
        get_type(name + '.item', text, parent, cols)
      elif type == 'struct':
        for pair in split_comma(text):
          (sname, colon, stype) = pair.partition(':')
          get_type(name + '.' + sname, stype, parent, cols)
      else:
        raise Exception('Unknown complex type: %s' % type)

    colnames = []
    coltypes = []

    rows_parsed = json.loads(rsp['rows'])
    for row in rows_parsed:
      # New hive includes partition information in the output so we need
      # to exit this loop when we hit the empty column name that preceeds
      # the partition info.
      if not row[0]:
        break

      # Hive 2.1.0 seems to have dropped the blank line before the partition
      # info (ARC-10410), so also quit on that.
      if row[0].startswith('# Partition Information'):
        break

      colname = row[0]
      coltype = re.sub(r'\s', '', row[1])

      cols = []
      get_type(colname, coltype, '', cols)
      for (name, type, parent) in cols:
        colnames.append(name)
        # decimal type values will be of the format decimal(precision[,scale])
        if 'decimal' in type:
          type = self.TYPEMAP.get('DECIMAL')
        elif (type not in ['map', 'struct', 'array']):
          type = self.TYPEMAP.get(type.upper(), arctypes.ARC_DATATYPE_STRING)
        coltypes.append(type)

    data = {
      'colnames': colnames,
      'coltypes': coltypes,
      'info': rsp['info'],
      'rows': [],
    }

    return data

  def _describe_output_to_partitions_and_comments(self, out):
    """
    Given the output of a describe formatted command,
    extract the partitioning columns and column comments
    """
    partitions = []
    comments = []
    start_partitions = False
    start_comments = False
    # Resulting arcengine call returns rows with the format:
    # [u'# col_name            ', u'data_type           ', u'comment             ']
    # Thus, set comment_index = 2 to extract the column comments
    comment_index = 2
    rows_parsed = json.loads(out['rows'])
    for row in rows_parsed:
      if len(row) == 0:
        continue
      colname = row[0].strip()
      if not colname:
        continue

      # End marker for column information
      if colname.startswith("# Detailed Table Information"):
        return (partitions, comments)

      # Start marker for column comments
      if colname == "# col_name":
        start_comments = True
        continue

      # Start marker for partitions
      if colname == "# Partition Information":
        start_partitions = True
        continue

      # Ignore comment rows
      if colname.startswith("#"):
        continue

      if start_comments:
        comment = row[comment_index].strip()
        if comment == "NULL":
          comments.append("")
        else:
          comments.append(comment)

      if start_partitions:
        partitions.append(colname)

    return (partitions, comments)

  def import_data_user_has_edit_perm(self, table_name, update_user):
    """
    Find out edit perms based on any arcadia user in table properties
    This is used by both hive and arcengine
    """
    arc_imported_user = None

    try:
      output = self.log_sync_runquery(
          "describe formatted %s" % self.quote_tablename(table_name),
          username=update_user.username)

      start = False
      rows_parsed = json.loads(output['rows'])
      for row in rows_parsed:
        if len(row) < 2:
          continue
        c = row[1].strip()
        if not c:
          continue

        if c.startswith("arc_importedby"):
          arc_imported_user = row[2].strip()
    except Exception as e:
      # If table does not exist, then user has perms to create it
      if re.search('does not exist', str(e)) or \
         re.search('Table not found', str(e)):
        return True
      else:
        raise

    if arc_imported_user is not None:
      if arc_imported_user == update_user.username or update_user.is_superuser:
        return True

    return False

  def async_cancel(self, qhandle):
    self.hs2conn.async_cancel(qhandle)
    return ''

  def async_status(self, qhandle, username=None, profile=None):
    """Get the status of an async query"""

    profile = profile or ArcProf('arcengine.status', username=username)

    status = self.hs2conn.async_status(qhandle, username=username, profile=profile)
    if status.status == DCS.RUNNING:
      if status.message:

        # Check for refresh progress
        m = re.match("Refresh (\d{1,3})% Completed", status.message)
        if m:
          return {"status": "Running (%s/%s completed)" % (m.groups()[0], 100)}

        # Try normal query status format
        m = re.match(".* Complete \((\d+) out of (\d+)\)", status.message)
        if m:
          return {"status": "Running (%s/%s completed)" % m.groups()}

      # Otherwise just make something up
      return {"status": "Running (10/100 completed)"}

    if status.status == DCS.FINISHED:
      return {"status": "Finished"}
    if status.status == DCS.QUEUED:
      return {"status": "Running (0/100 completed)"}
    if status.message:
      return {"status": "Failed (%s)" % status.message}
    return {"status": "Failed"}

  def async_results(self, qhandle, username=None, profile=None):
    """Get the results of an async query"""

    profile = profile or ArcProf('arcengine.results', username=username)

    colnames, coltypes, rows, info = self.hs2conn.async_results(qhandle,
        profile=profile)
    coltypes = [self.TYPEMAP.get(coltype.upper(),
                 arctypes.ARC_DATATYPE_STRING) for coltype in coltypes]
    # In version 2.1 of Spark show tables has an extra database column
    # Hack to reform newer 'show tables' response back to one column
    if self.connection_type == DATACONN_SPARKSQL and len(colnames) == 3:
      # Pull the query out of the info data
      query = ''
      for s in info:
        if s.startswith('Query:'):
          query = s[6:]
          break
      # Spark 2.1 returns 'database', 'tablename', 'istemporary'
      # we just require 'tablename'
      if query.lower().startswith('show tables '):
        colnames = colnames[1]
        coltypes = coltypes[1]
        rows_parsed = json.loads(rows)
        rows = [[r[1]] for r in rows_parsed]

    # Hive returns column names prefixed with tablenames and this messes
    # up dataset-from-sql logic.  ARC-12066
    if self.connection_type in [DATACONN_HIVE, DATACONN_SPARKSQL]:
      if getattr(settings, 'HIVE_FIXUP_COLNAMES', True):
        colnames = [re.sub(r'^.*?\.', '', c) for c in colnames]

    data = {
      'colnames': colnames,
      'coltypes': coltypes,
      'info': info,
      'rows': rows
    }
    del rows

    return data

  def async_runquery(self, query,
      username=None, sample_pct=None, comment='', logQuery=True,
      match_views=True, hideQuery=False, qcntl=None):
    """Run an async query"""
    self.validate_query(query, username)
    conf = self.getconf(sample_pct=sample_pct, match_views=match_views)

    doasuser = username if self.impersonation or self.impersonation_trusted else None
    hnd = self.hs2conn.async_issue(query, settings=conf, comment=comment,
        doasuser=doasuser, logQuery=logQuery, hideQuery=hideQuery, qcntl=qcntl)

    return {"qhandle": "%s" % hnd}

  def log_sync_runquery(self, query,
      username=None, sample_pct=None, comment='',
      client_ip=None, match_views=True):
    with QueryLog(username=username,
                  component_type=ComponentType.DATACONNECTION,
                  dataconnection_id=self.connection_id,
                  dataconnection_type=self.connection_type,
                  query=query,
                  sync=True,
                  client_ip=client_ip) as q:
      return self.sync_runquery(query,
          username=username, sample_pct=sample_pct, comment=comment,
          logQuery=False, match_views=match_views)

  def sync_runquery(self, query, build_request=None, disable_codegen=True,
      username=None, sample_pct=None, comment='', stream=False, logQuery=True,
      match_views=True, hideQuery=False, qcntl=None, profile=None,
      setCancelCbFn=None):
    """Run a SQL query and return the results"""

    # TODO: Need to rationalize these parameters...
    res = self.async_runquery(query,
                username=username,
                sample_pct=sample_pct,
                comment=comment,
                logQuery=logQuery,
                match_views=match_views,
                hideQuery=hideQuery,
                qcntl=qcntl)
    hnd = str(res['qhandle'])
    if stream:
      return self.hs2conn.async_results_stream(hnd, stream=True)

    return self.async_results(hnd, username=username)

  def getconf(self, sample_pct=None, match_views=True):
    """Build configuration overlay from input parms"""
    conf = {}
    if self.connection_type == DATACONN_ARCENGINE and \
            'match_views' not in self.conn_settings:
      if match_views:
        conf['match_views'] = '1'
      else:
        conf['match_views'] = '0'
    return conf

  def validate_expression(self, expression, build_request, exprType):
    if exprType == 'join':
      builder = build_request.builder
      dataset = builder.dataset_request.dataset
    else:
      builder = build_request.builder.builder
      dataset = builder.dataset
    alias = build_request.builder.dataset_request.dsreq.alias
    derived_data = build_request.builder.dataset_request.dsreq.derived_data

    try:
      data = self.expression_run_validation(expression, exprType,
                           builder=builder,
                           dataset=dataset,
                           username=build_request.username,
                           derived_data=derived_data,
                           expr_alias=alias)
      data['is_valid'] = "true"
    except Exception as e:
      # we assume that if expression_type fails expression is invalid
      return {'error_msg': str(e), 'is_valid': 'false'}

    if exprType == 'column':
      # if expression is not invalid continue with partition and aggregate check
      # an exception might be raised below (we don't explicitly catch it)
      # exception will be propagated to the caller
      kwargs = {
        'builder': builder,
        'dataset': dataset,
        'username': build_request.username,
        'derived_data': derived_data,
        'expr_alias': alias
      }

      data['ispart'] = self.expression_containsPartitions(expression, **kwargs)
      try:
        data['isaggregate'] = self.expression_isAggregate(expression, **kwargs)
      except Exception as e:
        return {'error_msg': str(e), 'is_valid': 'false'}

    return data

  def expression_run_validation(self, expression, exprType, **kwargs):
    data = {}
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    derived_data = kwargs.get('derived_data')
    expr_alias = kwargs.get('expr_alias')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    check_filter = exprType == 'filter'

    if exprType == 'join':
      header_query = builder.builder.query
    else:
      header_query = builder.expr_header(expression, dataset, exprType, derived_data=derived_data, expr_alias=expr_alias,
                                       check_filter=check_filter)[0]

    tmp = self.log_sync_runquery(header_query,
                             username=username)
    if exprType == 'column':
      data['type'] = tmp.get('coltypes')[0]

    return data

  def expression_containsPartitions(self, expression, **kwargs):
    # Compute if the column expression contains only partition columns --
    #  expression must contain at least one column and all must be partitions
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    derived_data = kwargs.get('derived_data')
    expr_alias = kwargs.get('expr_alias')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    columns = builder.expr_header(expression, dataset, 'column', derived_data=derived_data, expr_alias=expr_alias)[1]

    ispart = "false"
    if len(columns):
      column_table = list(columns)[0].table
      if column_table and not (column_table.fromView or column_table.select):
        tablename = list(columns)[0].table.tablename
        cols = self.table_columns(tablename, username)
        tmp = cols.get("partitions")
        partitions = {}
        for t in tmp:
          partitions[str(t)] = True
        for col in columns:
          if col.attr["name"] not in partitions:
            ispart = "false"
            break
          else:
            ispart = "true"

    return ispart

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    derived_data = kwargs.get('derived_data')
    expr_alias = kwargs.get('expr_alias')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset, derived_data=derived_data, expr_alias=expr_alias)
    hasaggr = 'false'
    try:
      res = self.log_sync_runquery(hasaggr_query, username=username)
      # SparkSQL pre-1.6 does not raise an exception, but rather returns 2 'rows' when the explain query fails above, e.g.
      # ROW1:== Physical Plan ==
      # ROW2:org.apache.spark.sql.AnalysisException: expression 'datestring' is neither present in the group by, nor is it
      # an aggregate function. Add to group by or wrap in first() (or first_value) if you don't care which value you get.;
      if self.connection_type == DATACONN_SPARKSQL:
        rows_parsed = json.loads(res.get('rows'))
        error_string = rows_parsed and len(rows_parsed)==2 and str(rows_parsed[1])
        if error_string and 'neither present in the group by, nor is it an aggregate function' in error_string:
          hasaggr = 'true'
        elif error_string:
          raise Exception(error_string)
    except Exception as e:
      s = str(e)
      if "expression not produced by aggregation" in s or \
         'in GROUP BY' in s:
        hasaggr = 'true'
      else:
        raise

    return hasaggr

  def apply_types(self, value, type):
    try:
      if value is None:
        value = "NULL"
      elif type == "String":
        value = re.sub(r"\\", r"\\\\", value)
        value = re.sub("'", r"\'", value)
        value = "'%s'" % value
      elif type == "Timestamp":
        value = "'%s'" % value
      elif type == "Real":
        value = re.sub(",", ".", value)
    except Exception as e:
      pass

    return value

  def validate_query(self, query, username):
    """Ensure the query is allowed"""
    if not '*' in settings.RAWSQL_WHITELIST:
      if not username in settings.RAWSQL_WHITELIST:
        query = (query or '').strip()
        if not query[:6].lower() == "select" and \
            not query[:6].lower() == "create" and \
            not query[:4].lower() == "show" and \
            not query[:4].lower() == "with" and \
            not query[:7].lower() == "refresh" and \
            not query[:8].lower() == "describe" and \
            not query[:20].lower() == "drop analytical view" and \
            not query[:19].lower() == "invalidate metadata" and \
            not query[:9].lower() == "drop view" and \
            not query[:10].lower() == "drop table" and \
            not query[:11].lower() == "alter table" and \
            not query[:10].lower() == "alter view" and \
            not query[:6].lower() == "insert" and \
            not query[:7].lower() == "compute" and \
            not query[:7].lower() == "analyze" and \
            not query[:7].lower() == "explain":
          raise Exception("%s queries are not permitted" % query[:6].lower())

  def serverlog(self, username, max_fetch=1024*1024, ctoken=None, startp=None):
    """ Access server log """
    if self.connection_type != DATACONN_ARCENGINE:
      raise Exception("serverlog function not supported")

    vreq = TGetServerLogReq(maxBytes=max_fetch, continuationToken=ctoken, startPoint=startp)
    vres = self.hs2call(username, 'GetServerLog', vreq, False)
    return {'data':vres.logData, 'ctoken':vres.continuationToken}

  def get_server_info(self, username):
    if self.connection_type != DATACONN_ARCENGINE:
      raise Exception("license information function not supported")

    info_req = TGetLicenseInfoReq()
    return self.hs2call(username, 'GetLicenseInfo', info_req, False)

  def srv_name(self):
    return self.hs2conn.srv_name()

  def srv_version(self):
    return self.hs2conn.srv_version()

  def hs2call(self, username, hs2name, hs2arg, getres):
    """ Admin path to issue arbitrary hs2 calls """
    doasuser = username if self.impersonation or self.impersonation_trusted else None
    return self.hs2conn.hs2call(hs2name, hs2arg, getres, doasuser)
