# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from . import dataconn_dbapi2
from .connections_types import DATACONN_ATHENA
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from django.conf import settings
from .dataconn_dataengine import DEFAULT_UNQUOTE

import json
from . import arctypes
import logging
import re

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  import pyathena as dbapi2
except Exception:
  # LOGGER.info("Libraries missing for Athena")
  pass

class AthenaConnProperties(ConnProperties):
  TYPEMAP = {
    'tinyint': arctypes.ARC_DATATYPE_INT,
    'smallint': arctypes.ARC_DATATYPE_INT,
    'bigint': arctypes.ARC_DATATYPE_BIGINT,
    'integer': arctypes.ARC_DATATYPE_INT,

    'varbinary': arctypes.ARC_DATATYPE_BINARY,

    'real': arctypes.ARC_DATATYPE_REAL,
    'double': arctypes.ARC_DATATYPE_DOUBLE,
    'float': arctypes.ARC_DATATYPE_DOUBLE,
    'decimal': arctypes.ARC_DATATYPE_DOUBLE,

    'char': arctypes.ARC_DATATYPE_STRING,
    'varchar': arctypes.ARC_DATATYPE_STRING,
    'string': arctypes.ARC_DATATYPE_STRING,

    'date': arctypes.ARC_DATATYPE_TIMESTAMP,
    'time': arctypes.ARC_DATATYPE_TIMESTAMP,
    'time with time zone': arctypes.ARC_DATATYPE_TIMESTAMP,
    'timestamp': arctypes.ARC_DATATYPE_TIMESTAMP,
    'timestamp with time zone': arctypes.ARC_DATATYPE_TIMESTAMP,

    'boolean': arctypes.ARC_DATATYPE_BOOLEAN,

    'map': arctypes.ARC_DATATYPE_STRING,
    'array': arctypes.ARC_DATATYPE_STRING,
    'row': arctypes.ARC_DATATYPE_STRING,
    'json': arctypes.ARC_DATATYPE_STRING,
  }

  def __init__(self):
    super(AthenaConnProperties, self).__init__()
    self.supports_limit_outside_union = False
    self.supports_deriveddata = True
    self.supports_tableproperties = True
    self.supports_nulls_order = True
    self.supports_async = False
    self.quote = '"'
    self.quote_end = '"'
    self.quote_alias = '"'
    self.quote_escape_character = "\\"
    self.quote_string_constant = "'"
    self.supports_partitions = False
    self.supports_backticks = False
    self.alias_supports_backticks = False
    self.display_to_table_type = {
      "String": "string",
      "Integer": "bigint",
      "Real": "double",
      "Boolean": "boolean",
      "Timestamp": "timestamp"
    }

  # Athena does not supports quotes in its aliases
  @staticmethod
  def alias_transform(alias):
    if not alias:
      return alias

    alias = re.sub(r"[^a-zA-Z0-9_\-]", "_", alias)
    return alias

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    return AthenaConnProperties.TYPEMAP.get(native_coltype,
        arctypes.ARC_DATATYPE_STRING)

class AthenaConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_ATHENA):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    conn_params = {}

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    LOGGER.info("AthenaConn")

    if params.get('HOST'):
      conn_params['region_name'] = str(params.get('HOST').strip())
    if params.get('USERNAME'):
      conn_params['aws_access_key_id'] = str(params.get('USERNAME'))
    if params.get('PASSWORD'):
      conn_params['aws_secret_access_key'] = str(params.get('PASSWORD'))
    if params.get('DATABASE'):
      self.dbname = str(params.get('DATABASE'))
      conn_params['s3_staging_dir'] = str(params.get('DATABASE').strip())
    if not self.dbname:
      raise Exception('Database name not specified')

    if not for_testing:
      self.connection = dbapi2.connect(**conn_params)
    self.connection_type = conn_type

  def properties(self):
    return AthenaConnProperties()

  def _get_unquote_regex(self):
    if self.compiled_unquote_regex is not None:
      return self.compiled_unquote_regex
    connections_quotes_regex = getattr(settings, 'VALIDATION_QUOTES_REGEX', {})
    if self.connection_type in connections_quotes_regex:
      self.compiled_unquote_regex = connections_quotes_regex[self.connection_type]
    else:
      # adding regex for removing connection specific white list statement
      self.compiled_unquote_regex = DEFAULT_UNQUOTE + \
            [re.compile(r'/\*([^\*])*\*/'),        # sql comments: /* my comment */
             re.compile(r'extract\s*\([^\)]*\)')]  # extract statement: extract(month from [time]::timestamp)
    return self.compiled_unquote_regex

  def execute(self, build_request, async_flag=False):

    query = None
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = 'show databases'
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      query = 'show tables in ' + build_request.query.split()[-1]
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            # if exception is caught above we set state to unknown
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)
    return executor.execute(query=query,
                            async_flag=async_flag)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "in GROUP BY" not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr
