# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
# To support bigquery, run:
#   pip install --upgrade google-cloud-bigquery
#   pip install google-auth-oauthlib

from .connections_types import DATACONN_BIGQUERY
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from . import dataconn_dbapi2
from . import dataconn_dataengine

from django.conf import settings
from vizlib.querylog import QueryLog
from vizlib.querylog_helper import ComponentType

import logging
import tempfile
import os

LOGGER = logging.getLogger("arcweb")

if settings.ENABLE_DATACONN_BIGQUERY:
  try:
    from google.cloud import bigquery
  except Exception as e:
    LOGGER.info("Did not find google bigquery library")

class BigQueryConnProperties(dataconn_dataengine.ConnProperties):
  def __init__(self, conn=None):
    dataconn_dataengine.ConnProperties.__init__(self)
    self._conn = conn
    self.quote = "`"
    self.quote_end = "`"
    self.quote_string_constant = '"'
    self.quote_alias = '`'
    self.supports_async = False
    self.supports_backticks = False
    self.sb_specify_alias = True

    self.alias_supports_spaces = False

  def quote_tablename(self, fq_table_name):
    return self._conn.quote_tablename(fq_table_name)

class BigQueryConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_BIGQUERY):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)

    self.connection_type = 'bigquery'
    self._billProjectId = params.get('BILL_PROJECT')
    self._projectId = params.get('PROJECT') or params.get('BILL_PROJECT')
    credentials = params.get('PASSWORD')

    f = tempfile.NamedTemporaryFile(delete=False)
    try:
      f.write(credentials.encode())
      f.flush()

      self._bq = bigquery.Client.from_service_account_json(
          f.name,
          project=self._billProjectId)
    finally:
      f.close()
      os.unlink(f.name)

  def quote_tablename(self, fq_table_name):
    fields = fq_table_name.split(".")
    if len(fields) == 1:
      return '`%s`' % fields[0]

    dbname = fields[0]
    tablename = fields[1]
    return '`%s.%s.%s`' % (self._projectId, dbname, tablename)

  def _getTables(self, datasetId):
    out = []

    dataset = bigquery.dataset.Dataset.from_string("%s.%s" % (self._projectId, datasetId))
    tables = self._bq.list_tables(dataset)
    for table in tables:
      out.append([table.table_id])

    return out

  def _getDatasets(self):
    dbs = []
    datasets = self._bq.list_datasets(self._projectId)
    for datasetItem in datasets:
      dbs.append([datasetItem.dataset_id])

    return dbs

  def properties(self):
    return BigQueryConnProperties(conn=self)

  def close(self):
    pass

  def table_columns(self, tablename, username=None, client_ip=None):
    q = 'select * from `%s.%s` limit 0' % (self._projectId, tablename)
    return self.sync_runquery(q)

  def async_cancel(self, qhandle):
    raise Exception('async_cancel not implemented')

  def async_status(self, qhandle, username=None, profile=None):
    raise Exception('async_status not implemented')

  def async_results(self, qhandle, username=None):
    raise Exception('async_results not implemented')

  def async_runquery(self, query,
                     username=None, sample_pct=None, comment=''):
    raise Exception('async_runquery not implemented')

  def sync_runquery(self, query, disable_codegen=True,
                    build_request=None,
                    username=None, sample_pct=None, comment='',
                    stream=False, qcntl=None, profile=None,
                    setCancelCbFn=None):
    if query == '.databases':
      dbs = self._getDatasets()
      data = {
        'colnames': ['Database Name'],
        'coltypes': ['STRING'],
        'info': ['Query: %s' % query],
        'rows': dbs,
      }
      return data

    if query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      dbname = query.split()[-1]
      tables = self._getTables(dbname)
      data = {
        'colnames': ['Table Name'],
        'coltypes': ['STRING'],
        'info': ['Query: %s' % query],
        'rows': tables,
      }
      return data

    if query.startswith(SPECIAL_CMDS['ROWS']):
      tablename = query.split()[-1]
      query = 'select * from `%s.%s` limit 100' % (self._projectId, tablename)

    query = comment + query

    LOGGER.info(query)
    return self._runq(query, username=username, comment=comment)

  def _runq_logged(self, query, username=None, comment='', client_ip=''):
    with QueryLog(username=username,
                  component_type=ComponentType.DATACONNECTION,
                  dataconnection_id=self.connection_id,
                  dataconnection_type=self.connection_type,
                  query=query,
                  sync=True,
                  client_ip=client_ip) as q:
      self._runq(query, username=username, comment=comment, client_ip=client_ip)

  def _runq(self, query, username=None, comment='', client_ip=''):
    job = self._bq.query(query)
    results = job.result()

    outRows = []
    for row in results:
      outRow = []
      for col in row:
        outRow.append(str(col))

      outRows.append(outRow)

    schema = results.schema
    data = {
      'colnames': [x.name for x in schema],
      'coltypes': [x.field_type for x in schema],
      'info': ['Query: %s' % query],
      'rows': outRows,
    }

    return data

  def execute(self, build_request, async_flag=False, stream=False):
    query = build_request.query

    for exprType, cmd in VALIDATE_CMDS.items():
      if query.startswith(cmd):
        expr = query[len(cmd):].strip()
        try:
          data = self.validate_expression(expr, build_request, exprType)
        except Exception as e:
          # if exception is caught above we set state to unknown
          error_message = 'Error validating expression'
          LOGGER.exception(error_message)
          data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
        return data

    executor = dataconn_dataengine.BuildRequestExecutor(build_request)
    return executor.execute(query=query, async_flag=async_flag, stream=stream)

  # TODO XXX: BigQuery validation doesn't work because it does not
  # support aliasing with spaces. Work on fixing this
  def validate_expression(self, expression, build_request, exprType):
    data = {
      'is_valid': "true",
      'isaggregate': "false",
      'ispart': "false"
    }
    return data

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    derived_data = kwargs.get('derived_data')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset,
        derived_data=derived_data)
    hasaggr = 'false'

    try:
      self.sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if 'Cannot group by an aggregate' not in str(e):
        LOGGER.exception('could not validate')
        raise
      else:
        hasaggr = 'true'

    return hasaggr

if __name__ == "__main__":
  main()
