# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json, logging, re
from . import arctypes
from . import dataconn_dbapi2
from .connections_types import DATACONN_CASSANDRA
from .dataconn_dataengine import ConnProperties, BuildRequest, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  from cassandra.cluster import Cluster
  from cassandra.auth import PlainTextAuthProvider
except Exception as e:
  # LOGGER.info('Libraries missing for Apache Cassandra')
  pass

class CassandraConnProperties(ConnProperties):
  TYPEMAP = {
    'ascii': arctypes.ARC_DATATYPE_STRING,
    'bigint': arctypes.ARC_DATATYPE_BIGINT,
    'boolean': arctypes.ARC_DATATYPE_BOOLEAN,
    'counter': arctypes.ARC_DATATYPE_BIGINT,
    'date': arctypes.ARC_DATATYPE_TIMESTAMP,
    'decimal': arctypes.ARC_DATATYPE_DOUBLE,
    'double': arctypes.ARC_DATATYPE_DOUBLE,
    'float': arctypes.ARC_DATATYPE_REAL,
    'inet': arctypes.ARC_DATATYPE_STRING,
    'int': arctypes.ARC_DATATYPE_INT,
    'smallint': arctypes.ARC_DATATYPE_INT,
    'text': arctypes.ARC_DATATYPE_STRING,
    'time': arctypes.ARC_DATATYPE_TIMESTAMP,
    'timestamp': arctypes.ARC_DATATYPE_TIMESTAMP,
    'tinyint': arctypes.ARC_DATATYPE_INT,
    'varchar': arctypes.ARC_DATATYPE_STRING,
    'varint': arctypes.ARC_DATATYPE_BIGINT,
    'uuid': arctypes.ARC_DATATYPE_BINARY,
  }

  def __init__(self):
    super(CassandraConnProperties, self).__init__()
    self.quote = '"'
    self.quote_end = '"'
    self.quote_alias = '"'
    self.quote_escape_character = '"'
    self.quote_string_constant = "'"
    self.supports_async = False
    self.supports_reuse = False
    self.supports_partitions = False
    self.alias_supports_backticks = False
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = None
    self.string_type = "text"
    self.timestamp_type = "timestamp"
    self.integer_type = "bigint"
    self.date_type = "date"
    self.double_type = "double"
    self.real_type = "float"
    self.boolean_type = "boolean"
    self.booleantruevalue = 'true'
    self.booleanfalsevalue = 'false'
    self.supports_limit_outside_union = False
    self.supports_limit = True
    self.supports_top = False
    self.supports_groupby_colindex = False
    self.supports_groupby_literal = False
    self.supports_table_aliases = False

    self.import_data_insertrows = 10000
    self.display_to_table_type = {
      "String": "varchar(max)",
      "Integer": "bigint",
      "Real": "float",
      "Boolean": "bit",
      "Timestamp": "datetime"
    }

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    r = CassandraConnProperties.TYPEMAP.get(native_coltype)
    if r:
      return r;

    # Sometimes, we get called internally when the typemap has already been
    # computed
    if native_coltype in arctypes.ARC_DATATYPES_ALL:
      return native_coltype

    # Fallback type is string
    return arctypes.ARC_DATATYPE_STRING

class CassandraConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_CASSANDRA):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.keyspace = None
    conn_params = {}
    auth_provider = None

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    if params.get('HOST'):
      conn_params['contact_points'] = [params.get('HOST').strip()]
    if params.get('PORT'):
      conn_params['port'] = int(params.get('PORT'))
    if params.get('USERNAME') and params.get('PASSWORD'):
      auth_provider = PlainTextAuthProvider(username=params.get('USERNAME'),
                                            password=params.get('PASSWORD'))
    #if params.get('DATABASE'):
    #  self.keyspace = params.get('DATABASE').strip()

    conn_params['auth_provider'] = auth_provider

    if not for_testing:
      try:
        self.connection = Cluster(**conn_params)
      except Exception as e:
        LOGGER.error(e)
        raise Exception('Could not connect to database')

    self.connection_type = conn_type
    self.cursor = None

  def properties(self):
    return CassandraConnProperties()

  def table_columns(self, tablename, username=None, client_ip=None):
    tablename = tablename.strip()
    q = 'select * from %s limit 1' % (tablename)
    out = self.sync_runquery(q)

    q = ("select column_name from system_schema.columns where keyspace_name = "
         "'%s' and table_name = '%s' and kind = 'partition_key' allow filtering" %
        (tablename.split('.')[0], tablename.split('.')[-1]))
    partitions = [x[0] for x in self.sync_runquery(q)['rows']]

    if not out:
      raise Exception("Could not fetch columns for %s" % tablename)

    # Add in partitioning information
    out['partitions'] = partitions
    return out

  def execute(self, build_request, async_flag=False):
    query = None
    limit = None

    LOGGER.info('query %s', build_request.query)
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = "select DISTINCT keyspace_name from system_schema.keyspaces"
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      query = "select table_name from system_schema.tables where keyspace_name = '%s'" % build_request.query.split()[-1]
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data


    executor = BuildRequestExecutor(build_request)

    r = executor.execute(query=query)

    return r

  def cursor_get(self, username=None):
    if self.cursor:
      return self.cursor

    cursor = self.get_connection(username).connect()
    return cursor

  def close(self):
    if self.connection:
      self.connection.shutdown()

  def cursor_close(self, cursor):
    if not self.cursor and cursor is not None:
      cursor.shutdown()

  def sync_runquery(self, query, disable_codegen=True,
                    build_request=None,
                    username=None, sample_pct=None, comment='',
                    stream=False, qcntl=None, profile=None,
                    setCancelCbFn=None):
    LOGGER.info(query)
    return self._runq(query, username=username, comment=comment)

  def _runq(self, query, username=None, comment='', client_ip=''):
    session = None
    data = dict()
    try:
      session = self.cursor_get()
      results = session.execute(query)

      outRows = []
      for row in results:
        outRow = []
        for col in row:
          outRow.append(str(col))

        outRows.append(outRow)

      properties = self.properties()
      data = {
        'colnames': results.column_names,
        'coltypes': [properties.cursor_coltype_to_arctype(x.typename) \
            for x in results.column_types],
        'info': ['Query: %s' % query],
        'rows': outRows,
      }

    finally:
      self.cursor_close(session)
    return data

  def expression_isAggregate(self, expression, data, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "Cannot use an aggregate" not in str(e):
        raise
      else:
        hasaggr = "true"

    data['isaggregate'] = hasaggr
