# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Base class for all data backend engines. Implement the interfaces in this
interface to implement a new data engine.
"""
from vizlib.query_builder import QueryBuilder
from vizlib.arcprof import ArcProf
from .connections_types import *
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.exceptions import FieldError
from .connections_types import DATACONN_DUCKDB, DATACONN_SQLITE, DATACONN_PG

import logging, re, copy

LOGGER = logging.getLogger("arcweb")

DEFAULT_UNQUOTE = [
  re.compile(r"(\\').*?(\\')", re.DOTALL),              # single quote escaped
  re.compile(r"(?<![\\'])'.*?(?<![\\'])'", re.DOTALL),  # single quote no escape
  re.compile(r'(\\`).*?(\\`)', re.DOTALL),              # backtick escaped
  re.compile(r'(?<![\\`])`.*?(?<![\\`])`', re.DOTALL),  # backtick no escape
  re.compile(r'(\\").*?(\\")', re.DOTALL),              # double quote escaped
  re.compile(r'(?<![\\"])".*?(?<![\\"])"', re.DOTALL),  # double quote no escape
  re.compile(r'(\[).*?(\])', re.DOTALL),                # bracketted text (column name)
]

_INVALID_PATTERN = \
  getattr(settings, 'INVALID_REGEX', r'(\bfrom\b)|(--)')

_INVALID = re.compile(_INVALID_PATTERN, flags=re.IGNORECASE)

UNCACHEABLE_FUNCTIONS = [
  'now',
  'current_timestamp',
  'timeofday',
  'unix_timestamp',
  'current_database',
  'effective_user',
  'pid',
  'user',
  'uuid',
  'version',
  'rand',
]

_INVALID_CACHE_PATTERNS = [ f'(\\b{func_name} *\\( *\\))' for func_name in UNCACHEABLE_FUNCTIONS]
_INVALID_CACHE_PATTERN = re.compile(r'|'.join(_INVALID_CACHE_PATTERNS), flags=re.IGNORECASE)

# Find patterns of EXTRACT(datepart from colname)
# - we are looking for EXTRACT followed by ( and then )
_EXTRACT_PATTERN = re.compile("^(.*)(EXTRACT\([^\)]+\))(.*)$", flags=re.IGNORECASE|re.DOTALL)

class ConnProperties(object):
  def __init__(self, conn_type=None):
    self.conn_type = conn_type
    self.quote = "`"
    self.quote_end = "`"
    self.quote_alias = "`"
    self.quote_escape_character = "\\"
    self.quote_string_constant = '"'
    self.supports_partitions = True
    self.supports_async = True
    self.supports_backticks = True
    self.supports_reuse = True
    self.supports_order_outside_union = True
    self.supports_limit_outside_union = True
    self.supports_import_data_perm = True
    self.supports_import_data_dbname = 'default'
    self.import_data_insertrows = 500
    self.escape_backslash = True
    self.alias_supports_backticks = True
    self.string_type = "STRING"
    self.timestamp_type = "TIMESTAMP"
    self.integer_type = "INTEGER"
    self.date_type = "STRING"
    self.double_type = "DOUBLE"
    self.real_type = "REAL"
    self.boolean_type = "BOOLEAN"
    self.booleantruevalue = 'TRUE'
    self.booleanfalsevalue = 'FALSE'
    self.supports_deriveddata = False
    self.supports_tableproperties = False
    self.supports_nulls_order = False
    self.sb_specify_alias = True
    self.display_to_table_type = None
    self.supports_groupby_literal = True
    self.joincharforarraytype = ' '
    self.paramstyle = '?'

    self.alias_supports_spaces = True

    # ORACLE and MS SQL Server do not support LIMIT clauses
    self.supports_limit = True

    # TERADATA does not support a LIMIT clause, but does support a SAMPLE clause instead
    self.supports_limit_sample = False

    # MS SQL Server supports SELECT TOP(5), and does not support LIMIT clauses
    # Teradata supports SELECT TOP 5, without the brackets
    self.supports_top = False
    self.supports_top_brackets = True

    # MS SQL Server does not support GROUP BY 1, it needs GROUP BY colname
    self.supports_groupby_colindex = True

    # KSQL does not support ORDER BY clauses
    self.supports_orderby = True

    # CQL / Cassandra does not support Table aliases
    self.supports_table_aliases = True

    # Druid interprets limit 0 and no-limit. So for druid, limit 0 queries should
    # instead be run as limit 1.
    # Similarly, Teradata does not support TOP(0) queries and needs validation queries
    # to be TOP(1) queries instead.
    # In order to support these connections, this property controls where limit 0 or
    # limit 1 queries are generated
    self.supports_limit_0 = True

    # Indicates whether this dataconnection can be the source of an extract.
    # Note that only connections that support a get_cursor() interface can be
    # used by the extracts code
    self.supports_dataextract_source = False

    # Indicates whether this dataconnection can be the target of an extract or supports data import.
    # These connections needs to support a truncate-table equivalent, along with
    # drop/create table and insert-data
    self.supports_dataextract_target = False

    # This flag is used to decide whether we generate a TRUNCATE TABLE or a DELETE FROM
    # clause to truncate data in a particular table.
    # Note that this functionality is only used by the extracts code, so we only have
    # to validate the command for data connections that can be the target of a data
    # extract.
    self.supports_truncate_table = False

    # whether inserts can have multiple values lists:
    self.supports_multi_row_insert = True

    # whether subselects can have alias
    self.supports_subselect_alias = True

  # Snowflake does not supports ( in its aliases
  @staticmethod
  def alias_transform(alias):
    return alias

  @staticmethod
  def cursor_coltypes_to_arctypes(native_coltype):
    return native_coltype

class DataEngine(object):

  compiled_unquote_regex = None

  def properties(self):
    raise Exception('properties not implemented')

  def apply_types(self, cell, type):
    raise Exception('type conversion not implemented')

  def close(self):
    raise Exception('close not implemented')

  def swap_quotes(self, query):
    return query

  def upload_maxrows(self):
    return settings.UPLOAD_MAXROWS_DEFAULT

  def _get_unquote_regex(self):
    if self.compiled_unquote_regex is not None:
      return self.compiled_unquote_regex
    connections_quotes_regex = getattr(settings, 'VALIDATION_QUOTES_REGEX', {})
    if self.connection_type in connections_quotes_regex:
      self.compiled_unquote_regex = connections_quotes_regex[self.connection_type]
    else:
      self.compiled_unquote_regex = DEFAULT_UNQUOTE
    return self.compiled_unquote_regex

  # Return a string with quoted components removed
  # For example, convert 'abc "def" ghi' to 'abc  ghi'
  def unquote(self, s):
    for regex in self._get_unquote_regex():
      s = regex.sub(r' ', s)
    return s

  def remove_extract_expr(self, s):
    m = _EXTRACT_PATTERN.search(s)
    while m is not None:
      s = m.group(1) + m.group(3)
      m = _EXTRACT_PATTERN.search(s)

    return s

  # Validate a single string
  # This is a simple algorithm:
  #  first, take out any quoted strings
  #  second, take out any EXTRACT(x FROM y) expressions
  #  third, look for and reject the word 'from'
  def _validate_str(self, s_in, validation_type):
    s = self.unquote(s_in)
    s = self.remove_extract_expr(s)
    if validation_type == 'sql':
      m = _INVALID.search(s)
      if m is not None:
        # Don't show the invalid text when hiding query, error details, etc.
        if getattr(settings, 'HIDE_PROFILE_QUERY', False):
          s_in = '***'
        raise ValidationError('Invalid string in field %s' % s_in)
    elif validation_type == 'cache':
      m = _INVALID_CACHE_PATTERN.search(s)
      if m is not None:
        raise FieldError('Invalid cache string in field %s' % s_in)

  # Validate an array of strings
  def _validate_arr(self, arr, validation_type):
    for a in arr:
      self._validate_str(a, validation_type)

  # Validate a dsreq object. Look for any invalid dsreq values
  def validate_dsreq(self, dsreq_obj, validation_type = 'sql'):
    if not dsreq_obj:
      return

    self._validate_arr([x['expr'] for x in dsreq_obj.get('dimensions', []) if 'expr' in x], validation_type)
    self._validate_arr([x['expr'] for x in dsreq_obj.get('aggregates', []) if 'expr' in x], validation_type)
    self._validate_arr([x['expr'] for x in dsreq_obj.get('partitions', []) if 'expr' in x], validation_type)
    self._validate_arr([x for x in dsreq_obj.get('filters', [])], validation_type)
    self._validate_arr([x for x in dsreq_obj.get('having', [])], validation_type)

    # validate limit in dsreq to make sure it contains only numbers and spaces
    limit = dsreq_obj.get('limit', '')
    if limit:
      try:
        limit = int(limit)
      except:
        raise ValidationError('Invalid string in field %s' % limit)

  # Enforce the site settings row limit on the input dsreq object
  def enforce_limit(self, dsreq, csvdownload, profile):
    # Quick check avoids site setting lookup for common path
    if dsreq is None:
      return

    if dsreq.get('type') == 'createaview' or dsreq.get('type') == 'save_table':
      return

    limit = dsreq.get('limit')

    from reports.views_settings import setting_get_single
    if csvdownload:
      max_limit = int(setting_get_single('download_maxrows'))
    else:
      max_limit = int(setting_get_single('visual_maxrows'))

    if limit is None or int(limit) >= max_limit:
      msg = 'Enforcing limit %s -> %s' % (limit, max_limit)
      LOGGER.warn(msg)
      profile.record('limit', msg)
      dsreq['limit'] = max_limit

  def tablename_parts(self, fq_table_name):
    properties = self.properties()

    fields = fq_table_name.split(".")
    if len(fields) == 1:
      return (None, fields[0].lstrip(properties.quote).rstrip(properties.quote_end))

    if len(fields) == 2:
      return (fields[0].lstrip(properties.quote).rstrip(properties.quote_end),
              fields[1].lstrip(properties.quote).rstrip(properties.quote_end))

    raise Exception("Invalid table name %s" % fq_table_name)

  def quote_tablename(self, fq_table_name):
    # If table name is a select statement, don't do any quoting
    if fq_table_name.startswith('('):
      return fq_table_name

    properties = self.properties()
    fields = fq_table_name.split(".")
    if len(fields) == 1:
      if fq_table_name[0] == properties.quote and \
          fq_table_name[len(fq_table_name) - 1] == properties.quote_end:
        # Previously quoted - return the tablename as is
        return fq_table_name

      return '%s%s%s' % (properties.quote, fq_table_name, properties.quote_end)

    if len(fields[0]) >= 2 and fields[0][0] == properties.quote and \
        fields[0][len(fields[0]) - 1] == properties.quote_end:
      # Previously quoted - return the tablename as is
      return fq_table_name

    return ".".join([properties.quote + field + properties.quote_end for field in fields])

  def query_results_key(self, username=None, query=None, dsreq=None,
                        results=None, client_ip=None):
    return None

  def table_columns(self, table_name, username=None, client_ip=None):
    raise Exception('table_columns not implemented')

  def async_cancel(self, qhandle):
    raise Exception('async_cancel not implemented')

  def async_status(self, qhandle, username=None, profile=None):
    raise Exception('async_status not implemented')

  def async_results(self, qhandle, username=None):
    raise Exception('async_results not implemented')

  def async_runquery(self, query,
                     username=None, sample_pct=None, comment=''):
    raise Exception('async_runquery not implemented')

  def sync_runquery(self, query, build_request=None, disable_codegen=True,
                    username=None, sample_pct=None, comment='',
                    stream=False, qcntl=None, profile=None,
                    setCancelCbFn=None):
    raise Exception('sync_runquery not implemented')

  def insert_table_prefix(self, tablename, overwrite=False):
    isHiveType = self.connection_type in ["hive", "hyve", "impala", "impyla",
        "arcengine", "arcengine_v2"]

    if isHiveType and overwrite:
      return f"INSERT OVERWRITE {self.quote_tablename(tablename)}"

    return f"INSERT INTO {self.quote_tablename(tablename)}"

  def drop_table_cmd(self, tablename):
    sqlcmd = "DROP TABLE IF EXISTS %s" % self.quote_tablename(tablename)
    return sqlcmd

  def truncate_table_cmd(self, tablename):
    properties = self.properties()

    if properties.supports_truncate_table:
      return f"TRUNCATE TABLE {self.quote_tablename(tablename)}"

    return f"DELETE FROM {self.quote_tablename(tablename)}"

  def rename_table_cmd(self, old_table_name, new_table_name):
    if self.connection_type in [DATACONN_DUCKDB, DATACONN_SQLITE, DATACONN_PG]:
      old_table_name_parts = self.tablename_parts(old_table_name)
      new_table_name_parts = self.tablename_parts(new_table_name)

      # ALTER TABLE RENAME can only rename a table within the same database
      if new_table_name_parts[0] is None or old_table_name_parts[0] == new_table_name_parts[0]:
        sqlcmd = f'ALTER TABLE {self.quote_tablename(old_table_name)} RENAME TO {self.quote_tablename(new_table_name_parts[1])}'
        return sqlcmd

      raise ValueError('Database name is not the same for source and target table names')
    else:
      sqlcmd = f'ALTER TABLE {self.quote_tablename(old_table_name)} RENAME TO {self.quote_tablename(new_table_name)}'
      return sqlcmd

  def drop_table_no_check_cmd(self, tablename):
    sqlcmd = f'DROP TABLE {self.quote_tablename(tablename)}'
    return sqlcmd

  def create_sql_cmds(self, tablename, colnames, coltypes,
                                  importuser, importsrcfile, external_table,
                                  location, localesetting='C', type_classes=None, partitionColnames=None, partitionColtypes=None,
                                  tblproperties=[], **kwargs):
    isHiveType = self.connection_type in ["hive", "hyve", "impala", "impyla",
        "arcengine", "arcengine_v2"]

    properties = self.properties()
    collist = [f"{properties.quote}{n}{properties.quote_end} {properties.display_to_table_type[t]}"
        for n, t in zip(colnames, coltypes)]

    if external_table and isHiveType:
      external = "EXTERNAL "
    else:
      external = ""

    sqlcmd = f"CREATE {external}TABLE {self.quote_tablename(tablename)} ({','.join(collist)})"

    if not isHiveType:
      return sqlcmd

    if partitionColnames:
      p = []
      p.append(' PARTITIONED BY (')
      cols = []
      for colname, coltype in zip(partitionColnames, partitionColtypes):
        cols.append(f'{colname} {coltype}')
      p.append(', '.join(cols))
      p.append(')')

      sqlcmd += ''.join(p)

    sqlcmd += " STORED AS PARQUET "

    if location:
      sqlcmd += " LOCATION '%s' " % location

    props = copy.deepcopy(tblproperties)
    if importuser:
      props.append("'arc_importedby'='" + importuser + "'")

    if importsrcfile:
      props.append("'arc_importedfile'='" + importsrcfile + "'")

    if props:
      sqlcmd += " TBLPROPERTIES(" + ", ".join(props) + ")"
    return sqlcmd

  def build(self, username=None, userid=None, query=None, dataset_request=None,
      sample_pct=None, comment=None, qcntl=None, profile=None):
    return BuildRequest(connection=self, username=username, userid=userid,
        query=query, dataset_request=dataset_request, sample_pct=sample_pct,
        comment=comment, qcntl=qcntl, profile=profile)

  def execute(self, build_request):
    """ Dataconn engines (arcengine, impala, hive, etc...)
    should implement this method for executing a build request.
    :param build_request:
    :return: result set or query handle
    """
    raise Exception('execute not implemented')

  def validate_query(self, query):
    """
    Ensure that wuery is valid
    :param query: query to be validated
    :return: returns nothing or throws exception
    """
    raise Exception('validate_query not implemented')

  def validate_expression(self, expression, build_request, exprType):
    """
    Implement this method to validate an expression
    :param expression: expression to be validated
    :param build_request: a BuildRequest object  Note: the details of the BuildRequest may vary. Its builder will always
      be a QueryBuilder object, but that QueryBuilder object could have a builder that is an ExpressionValidator object,
      as is the case for filter and column expression types, or a statsBuilder object, as is the case for join
      expressions, which are built from differently because the expression is on a proposed dataset rather than a saved
      dataset
    :param exprType: column, filter, or query (the kind of expression being validated)
    :return: a dictionary "data"
    data['is_valid'] = "true | false | unknown"
    data['error_msg'] if is_valid is false or unknown
    data['type'] = "<coltype>"
    data['ispart'] = "true | false"
    data['isaggregate'] = "true | false"
    """
    raise Exception('validate_expression not implemented')

  def expression_run_validation(self, expression, exprType, **kwargs):
    """
    Implement this method as part of the expression validation.
    Method should raise an exception for an invalid expression. If testing a column, return object should have a key of
     type to the data object with the type of the column
    :param expression: expression for which to get data type
    :param exprType: column, filter, or query (the kind of expression being validated)
    :param kwargs: e.g. dataset
    :return: data: object which may have key type
    """
    raise Exception('expression_run_validation not implemented')

  def expression_isAggregate(self, expression, **kwargs):
    """
    Implement this method as part the of the expression validation.
    Method should return True for an aggregate expression and False for a non-aggregate expression
    and raise an exception for an invalid expression
    :param expression: expression for which to get whether aggregate or not
    :param data: return value of validate_expression
    :param kwargs: e.g dataset
    :return: Boolean <True/False>
    """
    raise Exception('expression_isAggregate not implemented')

  def expression_containsPartitions(self, expression, **kwargs):
    """
    Implement this method as part of the expression validation.
    Method should return True for an expression containing only partition columns and False otherwise
    and should raise an exception for an invalid expression
    :param expression:
    :param kwargs:
    :return: Boolean <True/False>
    """
    raise Exception('expression_containsPartitions not implemented')

class BuildRequest(object):
  """ Intantiate this object with a connection,
      a client (DJANGO) request and
      a data api request object.
      Dataconn engines (arcengine, impala, hive, etc...)
      build method is expected to return this object.
  """

  def __init__(self, connection=None, username=None, userid=None, query=None,
               dataset_request=None, sample_pct=None, comment=None, qcntl=None,
               profile=None):
    self.connection = connection
    self.query = None
    self.builder = None
    self.username = username
    self.sample_pct = sample_pct
    self.comment = comment
    self.dataset_request = dataset_request
    self.qcntl = qcntl

    if profile is None:
      self.profile = ArcProf('dataengine', username=username)
    else:
      self.profile = profile

    if dataset_request is not None:
      self.builder = QueryBuilder(dataset_request=dataset_request,
                                  connection=self.connection, qcntl=qcntl)
      self.query = self.builder.buildQuery(username, userid, query, self.sample_pct)

    if settings.USE_DATA_API:
      if self.query is not None:
        # LOGGER.debug(self.query)
        pass
      else:
        self.query = query
        LOGGER.warning('using data API without a dsreq object...')
    else:
      LOGGER.warning('not using data API...')
      if query is not None:
        self.query = query
      else:
        raise ValueError('invalid query')

class BuildRequestExecutor(object):
  """ Intantiate this object with a BuildRequest.
      Dataconn engines (arcengine, impala, hive, etc...)
      may use the execute method for single query execution
      or implement their own execution pipeline.
      Dataconn engines are responsible for implementing (a)synchronous executions.
  """

  def __init__(self, build_request):
    self._build_request = build_request


  def execute(self, query=None, async_flag=False, stream=False):
    """ Use this method to execute a single query.
        Either pass in the query explicitly or use
        BuildRequest query. """

    if query is None:
      query = self._build_request.query

    if async_flag:
      if self._build_request.connection.properties().supports_async:  # ASYNCHRONOUS QUERY EXECUTION
        try:
          kwargs = {
            'username': self._build_request.username,
            'sample_pct': self._build_request.sample_pct,
            'comment': self._build_request.comment,
            'qcntl': self._build_request.qcntl,
          }

          if self._build_request.connection.connection_type == 'ksql':
            ksqlStreamPosition = None
            try:
              ksqlStreamPosition = self._build_request.dataset_request.dsreq.ksqlStreamPosition
              if ksqlStreamPosition:
                kwargs['ksqlStreamPosition'] = ksqlStreamPosition
            except:
              pass

          res = self._build_request.connection.async_runquery(query, **kwargs)
          return res
        finally:
          self._build_request.connection.close()

      else:  # QUEUE QUERY
        from vizlib.misc import asyncrunner_run

        return asyncrunner_run(connection_id=self._build_request.connection.connection_id,
                               query=query,
                               comment=self._build_request.comment,
                               username=self._build_request.username,
                               sample_pct=self._build_request.sample_pct,
                               profile=self._build_request.profile)
    else:  # SYNCHRONOUS QUERY EXECUTION
      try:
        return self._build_request.connection.sync_runquery(query,
            build_request=self._build_request,
            username=self._build_request.username,
            sample_pct=self._build_request.sample_pct,
            comment=self._build_request.comment,
            stream=stream,
            qcntl=self._build_request.qcntl)
      finally:
        if not stream:
          self._build_request.connection.close()
