# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import logging
import re

import datetime
import decimal
import base64
import time
import threading
import bisect

import vizlib.util
from vizlib.arcprof import ArcProf
from vizlib.querylog import QueryLog
from vizlib.querylog_helper import ComponentType
from django.conf import settings
from .rows_data import RowsData
import dataconn.dataconn
from .dataconn_dataengine import DataEngine
from .connections_types import DATACONN_DUCKDB, DATACONN_MONGODB, DATACONN_ASTER,\
  DATACONN_DRILL, DATACONN_MYSQL, DATACONN_ORACLE, DATACONN_SQLITE, DATACONN_IMPYLA, \
  DATACONN_HYVE
from . import arctypes

DATETIME_TYPE = type(datetime.datetime.now())
DATETIME_DATE_TYPE = type(datetime.date.today())
DECIMAL_TYPE = type(decimal.Decimal(1))
TIME_TYPE = type(datetime.time())
DATETIME_TIMEDIFF = type(datetime.timedelta(0))
BINARY_TYPE = type(bytearray())
BYTES_TYPE = type(bytes())
ORACLE_LOB_TYPE = "<type 'cx_Oracle.LOB'>"

# create logger
LOGGER = logging.getLogger("arcweb")

TO_STRING_TYPES = [DATETIME_TYPE, DATETIME_DATE_TYPE, DECIMAL_TYPE, DATETIME_TIMEDIFF]

def type_convert(col, coltype=None, milliToMicro=False):

  type_col = type(col)

  should_UTF8_decode_the_column = ((coltype == arctypes.ARC_DATATYPE_BINARY) and (col is not None)) \
    or type_col == BYTES_TYPE or type_col == BINARY_TYPE
  if should_UTF8_decode_the_column:
    return col.decode('utf-8', errors='replace')

  should_convert_to_string_the_column = type_col in TO_STRING_TYPES or str(type_col) == ORACLE_LOB_TYPE
  if should_convert_to_string_the_column:
    return str(col)

  if type_col == TIME_TYPE:
    if milliToMicro:
      return str(datetime.time(col.hour, col.minute, col.second, col.microsecond*1000))[:-3]
    else:
      return str(col)[:-3]

  return col

_SLOW_QUERY_SECONDS = 20

class _SingleConcurrencyTracker:
  def __init__(self, conn_id):
    self._lock = threading.Lock()
    self._conn_id = conn_id
    self._active = 0
    self._total = 0
    self._runtime = 0

    capacity = dataconn.dataconn.getDataConnConcurrency(conn_id)
    perUserCapacity = dataconn.dataconn.getDataConnConcurrencyUser(conn_id)

  def queryStart(self, username, profile, query):
    startTime = time.time()

    requestData = json.dumps({
      'querytracker': profile.querytracker(),
      'username': username,
      'connection': self._conn_id,
      'query': query,
    })
    ct = (startTime, requestData, profile)

    with self._lock:
      self._active += 1
      self._total += 1

    return ct

  def queryEnd(self, ct):
    (startTime, requestData, profile) = ct
    t2 = time.time()
    runtime = t2 - startTime

    with self._lock:
      self._active -= 1
      self._runtime += runtime

    if runtime > _SLOW_QUERY_SECONDS:
      LOGGER.info("Slow query. querytracker [%s]. runtime (%s) seconds, %s",
          profile.querytracker(), runtime, requestData)

  def data(self):
    with self._lock:
      return [self._conn_id, self._active, self._total, self._runtime]

class _ConcurrencyTracker:
  def __init__(self):
    self._lock = threading.Lock()
    self._trackers = {}

  def fetchAll(self):
    with self._lock:
      trackers = list(self._trackers.values())

    return [t.data() for t in trackers]

  def fetchConnectionTracker(self, connectionId):
    with self._lock:
      return self._trackers.get(connectionId)

  def queryStart(self, conn, username, profile, query):
    conn_id = conn.connection_id
    tracker = self._trackers.get(conn_id)
    if tracker is None:
      with self._lock:
        tracker = self._trackers.get(conn_id)
        if tracker is None:
          self._trackers[conn_id] = _SingleConcurrencyTracker(conn_id)
          tracker = self._trackers[conn_id]

    return tracker.queryStart(username, profile, query)

  def queryEnd(self, conn, ct):
    conn_id = conn.connection_id
    tracker = self._trackers[conn_id]
    tracker.queryEnd(ct)

class CancelState():
  def __init__(self):
    self._canceled = False

  def cancel(self):
    self._canceled = True

  def isCanceled(self):
    return self._canceled

CONCURRENCY_TRACKER = _ConcurrencyTracker()

# See similar variables in arc_dataconnection.js
UPLOAD_MAXROWS_MIN = 0
UPLOAD_MAXROWS_MAX = settings.UPLOAD_MAXROWS_MAX

class DbApi2Conn(DataEngine):
  def __init__(self, params, connection_id=-1):
    self.connection = None
    self.cursor = None
    self.connection_id = connection_id
    self.connection_type = 'unknown_dc'
    self.connection_properties = self.properties()

    self._upload_maxrows = params.get('UPLOAD_MAXROWS', settings.UPLOAD_MAXROWS_DEFAULT)
    if self._upload_maxrows < UPLOAD_MAXROWS_MIN:
      self._upload_maxrows = UPLOAD_MAXROWS_MIN
    if self._upload_maxrows > UPLOAD_MAXROWS_MAX:
      self._upload_maxrows = UPLOAD_MAXROWS_MAX

    # Support delays per query for debugging and testing
    self._delaySeconds = getattr(settings, "WORKQ_DELAY_SECONDS", {}).get(
        f"sql_{connection_id}", 0)

  def upload_maxrows(self):
    return self._upload_maxrows

  def set_connection(self, connection):
    self.connection = connection

  def get_connection(self, username=None):
    return self.connection

  def close(self):
    if self.connection:
      # The impyla api has an issue till it is fixed this is necessary
      # https://github.com/cloudera/impyla/issues/470
      try:
        self.connection.close()
      except Exception as e:
        LOGGER.exception(e)
      self.connection = None

  # mysql driver can only handle one cursor object per connection
  def cursor_get(self, username=None):
    if self.cursor:
      return self.cursor

    c = self.cursorConfiguration(username=username)

    conn = self.get_connection(username)
    if c:
      # Support configuration settings for cursors for hive/impala
      cursor = conn.cursor(configuration=c)
    else:
      cursor = conn.cursor()

    return cursor

  def cursor_close(self, cursor):
    if not self.cursor:
      cursor.close()

  def table_columns(self, fq_table_name, username=None, client_ip=None):
    if self.connection_type == 'mssqlserver':
      query = 'select top(0) * from %s' % self.quote_tablename(fq_table_name)
    elif self.connection_type == 'teradata':
      query = 'select * from %s sample 0' % self.quote_tablename(fq_table_name)
    else:
      if self.connection_properties.supports_limit_0:
        limit_str = "limit 0"
      else:
        limit_str = "limit 1"

      query = 'select * from %s %s' % (self.quote_tablename(fq_table_name), limit_str)

    d = self.log_sync_runquery(query, username=username,
        client_ip=client_ip)
    d['partitions'] = []
    return d

  def log_sync_runquery(self, query,
      username=None, sample_pct=None, comment='', setCancelCbFn=None, **kwargs):
    query = self.swap_quotes(query)
    with QueryLog(username=username,
                  component_type=ComponentType.DATACONNECTION,
                  dataconnection_id=self.connection_id,
                  dataconnection_type=self.connection_type,
                  query=query,
                  sync=True,
                  **kwargs) as q:
      return self.sync_runquery(query,
          username=username, sample_pct=sample_pct, comment=comment,
          setCancelCbFn=setCancelCbFn)

  # Override this function to pass in a configuration option to conn.cursor()
  # This is used by the dataconn_impyla class to implement impersonation
  def cursorConfiguration(self, username=None):
    return None

  # Used to insert multiple rows into a connection in the importdata path
  def executemany(self, username, query, rows, profile=None, **kwargs):
    profile = profile or ArcProf('dbapi2.executemany', username=username)

    LOGGER.info(query)
    LOGGER.info("rows: %s" % len(rows))
    cursor = self.cursor_get(username=username)

    ct = CONCURRENCY_TRACKER.queryStart(self, username, profile, query)
    try:
      if self.connection_type in [DATACONN_IMPYLA, DATACONN_HYVE]:
        # Specifying explicitly paramstyle ‘qmark’ to prevent Impyla API parameter substitution error
        # which happens when a value contains a colon and number (e.g. :37 in “2014-07-10 18:37:00”)
        # it is considered a numeric paramstyle
        configuration = {'paramstyle': 'qmark', **kwargs}
        cursor.executemany(query, rows, configuration)
      else:
        cursor.executemany(query, rows, **kwargs)
      self.get_connection(username).commit()
    finally:
      CONCURRENCY_TRACKER.queryEnd(self, ct)

  def fetchmany(self, cursor, rowBatchSize):
    if self.connection_type == DATACONN_DUCKDB:
      try:
        return cursor.fetchmany(size=rowBatchSize)
      except:
        return None
    elif self.connection_type == DATACONN_ORACLE:
      return cursor.fetchmany(rowBatchSize)
    else:
      return cursor.fetchmany(size=rowBatchSize)

    """Run a SQL query and return or stream the results"""
  def sync_runquery(self, query, build_request=None, disable_codegen=True,
      username=None, sample_pct=None, comment='', stream=False, logQuery=True,
      hideQuery=False, qcntl=None, profile=None, setCancelCbFn=None):

      query_result_generator = self.sync_runquery_stream(query, build_request, disable_codegen, username, sample_pct, comment, stream, logQuery,
      hideQuery, qcntl, profile, setCancelCbFn)

      if stream:
        return query_result_generator

      out = []
      for out in query_result_generator:
        pass

      return out

  def sync_runquery_stream(self, query, build_request=None, disable_codegen=True,
      username=None, sample_pct=None, comment='', stream=False, logQuery=True,
      hideQuery=False, qcntl=None, profile=None, setCancelCbFn=None):
    profile = profile or ArcProf('dbapi2.sync_runquery', username=username)

    if comment:
      query = comment + '\n' + query

    profile.mark('sync_runquery')
    querytracker = profile.querytracker()

    query = self.swap_quotes(query)
    query = query.strip()
    cursor = self.cursor_get(username=username)
    cancelState = CancelState()
    profile.mark('cursor.after')

    if setCancelCbFn:
      if self.connection_type in [DATACONN_SQLITE, DATACONN_IMPYLA, DATACONN_HYVE]:

        # Setup a call that can be used to cancel the query we are going to run below.
        # Note that this happens in a separate thread from where the cursor is running
        #   the query, so we have to test this for every db.
        # For SQLite, calling interrupt() on the connection works.
        # On Impyla, we have to cancel the query in the same thread. We mark it as
        #   canceled here, and we call cur.cancel_operation() below.
        def cancelQuery():
          cancelState.cancel()

          if self.connection_type == DATACONN_SQLITE:
            self.get_connection(username).interrupt()
            return

          if self.connection_type == DATACONN_IMPYLA:
            # Note that impyla cursors are not threadsafe, so cancel
            # the query in the main thread below
            return

          # Other connection types are not supported yet (needs dev work and testing)

        setCancelCbFn(cancelQuery)

    from reports.views_settings import setting_get_single
    maxSizeMB = setting_get_single('visual_max_dataresults') or ''
    maxSizeMB = int(maxSizeMB) if maxSizeMB.isnumeric() else 10
    maxSize = maxSizeMB * 1000 * 1000
    rowsData = RowsData(maxSize)
    rowCount = None

    try:
      if hideQuery:
        LOGGER.info('querytracker [%s]. Running query: ***',
            querytracker)
      else:
        LOGGER.info('querytracker [%s]. Running query: %s',
            querytracker, query)

      ct = CONCURRENCY_TRACKER.queryStart(self, username, profile, query)

      if self._delaySeconds:
        if callable(self._delaySeconds):
          s = self._delaySeconds()
        else:
          s = self._delaySeconds

        if s:
          LOGGER.info("Delaying for %s seconds", s)
          time.sleep(s)

        if cancelState.isCanceled():
          raise Exception("Query has been canceled by user")

      try:
        if self.connection_type in [DATACONN_IMPYLA, DATACONN_HYVE]:
          # Run the query asynchronously so we can cancel it if needed.
          # See setCancelFn() usage above
          cursor.execute_async(query)

          while cursor.is_executing():
            if cancelState.isCanceled():
              cursor.cancel_operation(reset_state=True)
              raise Exception("Query canceled by user")

            time.sleep(0.1)

        else:
          cursor.execute(query)

        rowCount = cursor.rowcount
      finally:
        if setCancelCbFn:
          setCancelCbFn(None)

        CONCURRENCY_TRACKER.queryEnd(self, ct)

      profile.mark('sync_runquery.execute.end')
      if cursor.description:
        # name, type_code, display_size, internal_size, precision, scale, null_ok
        # pyodbc only provides values for name, type_code, internal_size, and null_ok
        colnames = [x[0] for x in cursor.description]
        coltypes = [x[1] or 'STRING' for x in cursor.description]
        if hasattr(self.connection_properties, 'cursor_coldesc_to_arctype'):
          coltypes = [self.connection_properties.cursor_coldesc_to_arctype(x) \
              for x in cursor.description]

        compute_coltypes = True
        for x in cursor.description:
          if x[1]:
            # In sqlite, the 2nd element of the column is None
            # This means the only good way to get column types out of this
            # db is to inspect the types of the columns in the first row.
            compute_coltypes = False
            break

        if not compute_coltypes:
          if hasattr(self.connection_properties, 'cursor_coltype_to_arctype'):
            coltypes = [self.connection_properties.cursor_coltype_to_arctype(x) \
                for x in coltypes]

          # Handle druid case where name also matters
          if hasattr(self.connection_properties, 'cursor_colname_coltype_to_arctype'):
            coltypes = [self.connection_properties.cursor_colname_coltype_to_arctype(colnames[i], type) \
                for (i, type) in enumerate(coltypes)]

        rowBatchSize = 1000
        rows = self.fetchmany(cursor, rowBatchSize)

        if rows:
          profile.mark('firstrow')

        coltypes_unknown = True # True if we have not detected all the column types
        if compute_coltypes and rows:
          coltypes = [None] * len(rows[0])


        if stream and not compute_coltypes:
          yield coltypes
          yield colnames

        while rows:
          for row in rows:
            if compute_coltypes and coltypes_unknown:
              #TODO: This does not detect timestamp, so for
              #sqlite, timestamps are shown as string types
              for colidx, col in enumerate(row):
                if coltypes[colidx] is not None: # We already know the type for this column
                  continue
                if type(col) == type(3): # int
                  coltypes[colidx] = arctypes.ARC_DATATYPE_BIGINT
                elif type(col) == type(3.0): # float
                  coltypes[colidx]=arctypes.ARC_DATATYPE_DOUBLE
                elif type(col) == type(1): # long
                  coltypes[colidx]=arctypes.ARC_DATATYPE_BIGINT
                elif isinstance(col, str): #  String
                  coltypes[colidx]=arctypes.ARC_DATATYPE_STRING
                else:  # Skip nulls and unknown types
                  pass
              coltypes_unknown = None in coltypes # check if there are still unknown coltypes

            if self.connection_type == DATACONN_DRILL:
              row = [type_convert(col, milliToMicro=True) for col in row]
              if query.startswith("show tables in"):
                row = [row[1]]
            else:
              newrow = [None for c in row]
              for colindex, col in enumerate(row):
                newrow[colindex] = type_convert(col, coltype=coltypes[colindex])
              row = newrow

            if stream:
              if compute_coltypes:
                coltypes = [t if t is not None else arctypes.ARC_DATATYPE_STRING for t in coltypes]
                compute_coltypes = False
                yield coltypes
                yield colnames

              yield row

            else:
              rowsData.addRow(row)

          rows = self.fetchmany(cursor, rowBatchSize)

        # Default to String for any col where we did not detect type
        # Should we default to arctypes.ARC_DATATYPE_UNKNOWN?
        if compute_coltypes:
          coltypes = [t if t is not None else arctypes.ARC_DATATYPE_STRING for t in coltypes]

        if self.connection_type == 'hyve':
          if getattr(settings, 'HIVE_FIXUP_COLNAMES', True):
            colnames = [re.sub(r'^.*?\.', '', c) for c in colnames]

        profile.markFromStart('query_runtime')

        data = {
          'colnames': colnames,
          'coltypes': coltypes,
          'rows': rowsData.fetchAndClose(),
          'rowcount': rowsData.rowCount(),
          'info': ['Query:%s' % query],
          'query_warnings': rowsData.getQueryWarnings(),
        }
      else:
        # Insert command wasn't working for sqlite with forcing a connection commit here
        self.get_connection(username).commit()
        data = {
          'colnames': [],
          'coltypes': [],
          'rows': '[]',
          'rowcount': rowCount,
          'info': ['Query:%s' % query],
          'query_warnings': {},
        }

      if not stream:
        yield data
        return

    except Exception as e:
      LOGGER.exception(e)
      raise

    finally:
      try:
        self.cursor_close(cursor)
      except Exception as e:
        # Log exceptions when closing cursors, but don't surface these errors
        # to the user
        LOGGER.exception('Failed to close cursor')

    # NOTREACHED
    return 'invalid'

  def validate_expression(self, expression, build_request, exprType):
    if exprType == 'join':
      builder = build_request.builder
      dataset = builder.dataset_request.dataset
    else:
      builder = build_request.builder.builder
      dataset = builder.dataset

    alias = build_request.builder.dataset_request.dsreq.alias

    derived_data = build_request.builder.dataset_request.dsreq.derived_data

    is_filter = False
    data = {}

    kwargs = {
      'builder': builder,
      'dataset': dataset,
      'username': build_request.username,
      'derived_data': derived_data,
      'expr_alias': alias
    }

    # Teradata does not permit filter predicates in projections
    if self.connection_type == 'teradata' or exprType == 'filter':
      try:
        self.filter_type(expression, **kwargs)
        data['is_valid'] = "true"
        data['is_filter'] = 'true'
        data['type'] = 'boolean'
        is_filter = True
      except Exception as e:
        if exprType == 'filter':
          data['error_msg'] = self.parseError(e)
          data['is_valid'] = 'false'
          return data
        else:
          LOGGER.exception('Not a filter expression')

    # Even if this expression seems like it could be used as a filter,
    # try and get the type of the expression
    try:
      data = self.expression_run_validation(expression, exprType, **kwargs)
      data['is_valid'] = "true"
    except Exception as e:
      if is_filter:
        # We have previously validated this expression as a filter
        return data

      # AsterDB does not pass an error message to exception object
      # instead error message is in 'args' tuple in 'e'
      data['error_msg'] = self.parseError(e)
      data['is_valid'] = "false"
      return data

    kwargs.pop('expr_alias', None)
    kwargs.pop('derived_data', None)

    if exprType == 'column':
      data['ispart'] = self.expression_containsPartitions(expression, **kwargs)

      try:
        data['isaggregate'] = self.expression_isAggregate(expression, **kwargs)
      except Exception as e:
        data['error_msg'] = self.parseError(e)
        data['is_valid'] = "false"
        return data

    return data

  def parseError(self, e):
    if self.connection_type in [DATACONN_ASTER, DATACONN_DRILL]:
      if hasattr(e, 'args') and len(e.args) == 2:
        error_msg = e.args[1]
      else:
        error_msg = str(e) or 'Invalid Expression'
    elif self.connection_type == DATACONN_MYSQL:
      if hasattr(e, 'args') and len(e.args) in  [2, 3]:
        error_msg = e.args[1]
      else:
        error_msg = str(e) or 'Invalid Expression'
    else:
      error_msg = str(e)

    return error_msg

  def expression_run_validation(self, expression, exprType, **kwargs):
    data = {}
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    derived_data = kwargs.get('derived_data')
    expr_alias = kwargs.get('expr_alias')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    if exprType == 'join':
      header_query = builder.builder.query
    else:
      header_query = builder.expr_header(expression, dataset, exprType, limit=1, derived_data=derived_data, expr_alias=expr_alias)[0]


    tmp = self.log_sync_runquery(header_query,
                             username=username)

    if exprType == 'column':
      data['type'] = tmp.get('coltypes')[0]

    return data

  def filter_type(self, expression, **kwargs):
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    derived_data = kwargs.get('derived_data')
    expr_alias = kwargs.get('expr_alias')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    header_query = builder.expr_filter(expression, dataset, limit=1, derived_data=derived_data, expr_alias=expr_alias)[0]

    tmp = self.log_sync_runquery(header_query, username=username)

  def expression_containsPartitions(self, expression, **kwargs):
    # Compute if the column expression contains only partition columns --
    #  expression must contain at least one column and all must be partitions

    if not self.connection_properties.supports_partitions:
      return 'false'

    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    expr_alias = kwargs.get('expr_alias')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    try:
      columns = builder.expr_header(expression, dataset, 'column', expr_alias=expr_alias)[1]
      ispart = "false"
      if len(columns):
        tab = list(columns)[0].table
        if tab.select:
          tablename = '(%s) t1' % tab.select
        else:
          tablename = tab.tablename
        cols = self.table_columns(tablename, username)
        tmp = cols.get("partitions")
        partitions = {}
        for t in tmp:
          partitions[str(t)] = True
        for col in columns:
          if col.attr["name"] not in partitions:
            ispart = "false"
            break
          else:
            ispart = "true"

    except Exception as e:
      LOGGER.exception(e)
      return "false"

    return ispart

  def _is_stringtype(self, headertype):
    return headertype in ['string','text'] or \
        headertype.startswith('varchar') or headertype.startswith('char')

