# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from . import arctypes
from . import dataconn_dbapi2
import logging, sys
import re

from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .connections_types import DATACONN_DRUID
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from os.path import normpath, expanduser
from vizlib.querylog import QueryLog

try:
  import pydruid.db as dbapi2
except Exception:
  pass

LOGGER = logging.getLogger("arcweb")

class DruidConnProperties(ConnProperties):
  TYPEMAP = {
    1: arctypes.ARC_DATATYPE_STRING,

    2: arctypes.ARC_DATATYPE_BIGINT,

    3: arctypes.ARC_DATATYPE_DOUBLE,
  }

  TYPENAMEMAP = {
    '__time': arctypes.ARC_DATATYPE_TIMESTAMP,
  }

  def __init__(self):
    super(DruidConnProperties, self).__init__()
    self.supports_partitions = False
    self.supports_async = False

    # Druid connections must be used from within their threads
    # We cannot store per-client connections, so simply don't reuse for now
    self.supports_reuse = False

    self.escape_backslash = False
    self.supports_backticks = False
    self.alias_supports_backticks = False
    self.quote = '"'
    self.quote_end = '"'
    self.quote_alias = '"'

    self.string_type = "STRING"
    self.timestamp_type = "TIMESTAMP"
    self.date_type = "STRING"
    self.boolean_type = "LONG"
    self.booleanfalsevalue = '0'
    self.booleantruevalue = '1'

    self.supports_deriveddata = True
    self.supports_dataextract_source = True
    self.display_to_table_type = {
      "String": "STRING",
      "Integer": "LONG",
      "Real": "FLOAT",
      "Boolean": "LONG",
      "Timestamp": "TIMESTAMP"
    }

    self.supports_limit_0 = False

  # Druid does not support ", and likely many other characters in aliases
  @staticmethod
  def alias_transform(alias):
    if not alias:
      return alias

    alias = re.sub(r"[^a-zA-Z0-9_\- ]", "_", alias)
    return alias

  @staticmethod
  def cursor_colname_coltype_to_arctype(native_colname, native_coltype):
    coltype = DruidConnProperties.TYPEMAP.get(native_coltype,
        arctypes.ARC_DATATYPE_STRING)

    return DruidConnProperties.TYPENAMEMAP.get(native_colname, coltype);

class DruidConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_DRUID):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)

    conn_params = {}

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    if params.get('HOST'):
      conn_params['host'] = str(params.get('HOST').strip())
    if params.get('PORT'):
      conn_params['port'] = str(params.get('PORT'))
    if params.get('USERNAME'):
      conn_params['user'] = str(params.get('USERNAME'))
    if params.get('PASSWORD'):
      conn_params['password'] = str(params.get('PASSWORD'))
    if params.get('DATABASE'):
      self.dbname = str(params.get('DATABASE'))
      conn_params['database'] = str(params.get('DATABASE').strip())

    if not for_testing:
      self.set_connection(dbapi2.connect(**conn_params))

    self.connection_type = conn_type

  def properties(self):
    return DruidConnProperties()

  # execute query(ies)
  def execute(self,
              build_request,
              async_flag=False,
              sample_pct=None,
              comment=None):

    LOGGER.info(build_request.query)

    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      build_request.query = """
        SELECT
          distinct TABLE_SCHEMA
        FROM INFORMATION_SCHEMA.TABLES
        WHERE
          TABLE_SCHEMA != 'INFORMATION_SCHEMA'
          AND
          TABLE_SCHEMA != 'sys'
        ORDER BY 1
      """

    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      dbname = build_request.query.split()[-1]
      build_request.query = f"""
        SELECT
          distinct TABLE_NAME
        FROM INFORMATION_SCHEMA.TABLES
        WHERE
          TABLE_SCHEMA = '{dbname}'
        ORDER BY 1
      """

    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)
    return executor.execute(query=build_request.query,
                            async_flag=async_flag)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "column must appear in the GROUP BY clause or be used in an aggregate function" not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr

  def apply_types(self, value, type):
    try:
      type = type.lower()

      if value is None:
        value = "NULL"
      elif type == "string":
        value = re.sub("'", r"''", value)
        value = "'%s'" % value
      elif type == "timestamp":
        value = "'%s'" % value
      elif type == "real":
        value = re.sub(",", ".", value)
      elif type == "boolean":
        if value:
          value = 1
        else:
          value = 0
    except Exception as e:
      pass

    return value

