# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from . import arctypes
from . import dataconn_dbapi2
import logging, sys
import re
import json

from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .connections_types import DATACONN_DUCKDB
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from os.path import normpath, expanduser
from vizlib.querylog import QueryLog
from vizlib.querylog_helper import ComponentType

try:
  import duckdb as dbapi2
except Exception:
  pass

LOGGER = logging.getLogger("arcweb")

class DuckDBConnProperties(ConnProperties):
  TYPEMAP = {
    'INT': arctypes.ARC_DATATYPE_INT,
    'INTEGER': arctypes.ARC_DATATYPE_INT,
    'TINYINT': arctypes.ARC_DATATYPE_INT,
    'SMALLINT': arctypes.ARC_DATATYPE_INT,
    'MEDIUMINT': arctypes.ARC_DATATYPE_INT,

    'BIGINT': arctypes.ARC_DATATYPE_BIGINT,

    'DOUBLE': arctypes.ARC_DATATYPE_DOUBLE,
    'REAL': arctypes.ARC_DATATYPE_DOUBLE,
    'FLOAT': arctypes.ARC_DATATYPE_DOUBLE,

    'TEXT': arctypes.ARC_DATATYPE_STRING,
    'VARCHAR': arctypes.ARC_DATATYPE_STRING,
    'STRING': arctypes.ARC_DATATYPE_STRING,

    'DATE': arctypes.ARC_DATATYPE_TIMESTAMP,
    'DATETIME': arctypes.ARC_DATATYPE_TIMESTAMP,
    'TIMESTAMP': arctypes.ARC_DATATYPE_TIMESTAMP,

    'BOOLEAN': arctypes.ARC_DATATYPE_BOOLEAN,
  }

  def __init__(self):
    super(DuckDBConnProperties, self).__init__()
    self.supports_partitions = False
    self.supports_async = False
    # DuckDB connections must be used from within their threads
    # We cannot store per-client connections, so simply don't reuse for now
    self.supports_reuse = False
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = 'main'
    self.supports_import_data_dbmod = False
    self.escape_backslash = False
    self.supports_backticks = False
    self.alias_supports_backticks = False
    self.quote = '"'
    self.quote_end = '"'
    self.quote_alias = '"'

    self.string_type = "VARCHAR"
    self.timestamp_type = "TIMESTAMP"
    self.date_type = "VARCHAR"
    self.booleanfalsevalue = 'TRUE'
    self.booleantruevalue = 'FALSE'
    self.supports_deriveddata = True
    self.supports_dataextract_source = True
    self.supports_dataextract_target = True
    self.display_to_table_type = {
      "String": "VARCHAR",
      "Integer": "BIGINT",
      "Real": "REAL",
      "Boolean": "BOOLEAN",
      "Timestamp": "TIMESTAMP"
    }

  # DuckDB does not support ", and likely many other characters in aliases
  @staticmethod
  def alias_transform(alias):
    if not alias:
      return alias

    alias = re.sub(r"[^a-zA-Z0-9_\- ]", "_", alias)
    return alias

class DuckDBConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_DUCKDB):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)

    db_path = normpath(expanduser(params['FILE']))

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    if not for_testing:
      self.set_connection(dbapi2.connect(database=db_path))

    self.dbname = 'main'
    self.connection_type = conn_type

  def properties(self):
    return DuckDBConnProperties()

  # Overrides table_columns because DuckDB gives no column type information
  # as part of select query results, and so table create cmd has to be queried
  # to figure out table columns, and for other queries, column types have to
  # to guessed through column values
  def table_columns(self, db_table_name, username=None, client_ip=None):
    properties = self.properties()
    table_name = db_table_name.split(".")[-1]
    query = "PRAGMA table_info('%s')" % self.quote_tablename(table_name)
    LOGGER.info(query)
    with QueryLog(username=username,
                  component_type=ComponentType.DATACONNECTION,
                  dataconnection_id=self.connection_id,
                  dataconnection_type=self.connection_type,
                  query=query,
                  sync=True,
                  client_ip=client_ip) as q:
      cursor = self.cursor_get()
      try:
        cursor.execute(query)
        # returned fields are:
        #   columnid, columnname, columntype, notnull, defaultvalue, primarykey
        rows = cursor.fetchall()
        colnames = [row[1] for row in rows]
        coltypes = [row[2] for row in rows]
        coltypes = [properties.TYPEMAP.get(coltype.upper(),
                     arctypes.ARC_DATATYPE_STRING) for coltype in coltypes]
        data = {
          'colnames': colnames,
          'coltypes': coltypes,
          'partitions': []
        }
        return data
      finally:
        self.cursor_close(cursor)

  # execute query(ies)
  def execute(self,
              build_request,
              async_flag=False,
              sample_pct=None,
              comment=None):

    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      build_request.query = '.databases'
      data = {
        'colnames': ['dbname'],
        'coltypes': ['STRING'],
        'rows': json.dumps([[self.dbname]])
      }
      return data

    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      build_request.query = \
          """SELECT name FROM sqlite_master WHERE type='table' OR type='view' AND NOT name='sqlite_master' ORDER BY name"""
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)
    return executor.execute(query=build_request.query,
                            async_flag=async_flag)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "column must appear in the GROUP BY clause or be used in an aggregate function" not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr

  def apply_types(self, value, type):
    try:
      if value is None:
        value = "NULL"
      elif type == "String":
        value = re.sub("'", r"''", value)
        value = "'%s'" % value
      elif type == "Timestamp":
        value = "'%s'" % value
      elif type == "Real":
        value = re.sub(",", ".", value)
      elif type == "Boolean":
        if value:
          value = 1
        else:
          value = 0
    except Exception as e:
      pass

    return value

