# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from . import dataconn_dbapi2
from . import dataconn_dataengine
from .connections_types import DATACONN_ELASTICSEARCH
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS

from vizlib.querylog import QueryLog

import logging
import json
from . import arctypes

LOGGER = logging.getLogger("arcweb")
ELASTICSEARCH_DB_NAME = "Elasticsearch"
_MAX_CONNS_PER_NODE = 25

try:
  from elasticsearch import Elasticsearch
except Exception as e:
  # LOGGER.info("Did not find Elasticsearch library")
  pass

class ElasticsearchConnProperties(dataconn_dataengine.ConnProperties):
  TYPEMAP = {
    'BIGINT': arctypes.ARC_DATATYPE_BIGINT,
    'BOOLEAN': arctypes.ARC_DATATYPE_BOOLEAN,
    'TINYINT': arctypes.ARC_DATATYPE_INT,
    'SMALLINT': arctypes.ARC_DATATYPE_INT,
    'INTEGER': arctypes.ARC_DATATYPE_INT,
    'VARCHAR': arctypes.ARC_DATATYPE_STRING,
    'TEXT': arctypes.ARC_DATATYPE_STRING,
    'KEYWORD': arctypes.ARC_DATATYPE_STRING,
    'LONG': arctypes.ARC_DATATYPE_BIGINT,
    'TIMESTAMP': arctypes.ARC_DATATYPE_TIMESTAMP,
    'DATE': arctypes.ARC_DATATYPE_TIMESTAMP,
    'STRUCT': arctypes.ARC_DATATYPE_STRING,
  }

  def __init__(self, conn=None):
    dataconn_dataengine.ConnProperties.__init__(self)
    self._conn = conn
    self.quote = '"'
    self.quote_end = '"'
    self.quote_string_constant = "'"
    self.quote_escape_character = "\\"
    self.supports_reuse = True
    self.quote_alias = '"'
    self.supports_async = False
    self.supports_backticks = False
    self.sb_specify_alias = True

    self.alias_supports_spaces = False

  def quote_tablename(self, fq_table_name):
    return self._conn.quote_tablename(fq_table_name)

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    r = ElasticsearchConnProperties.TYPEMAP.get(native_coltype)
    if r:
      return r;

    # Sometimes, we get called internally when the typemap has already been
    # computed
    if native_coltype in arctypes.ARC_DATATYPES_ALL:
      return native_coltype

    # Fallback type is string
    return arctypes.ARC_DATATYPE_STRING

class ElasticsearchConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_ELASTICSEARCH):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    conn_params = {}

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    LOGGER.info("ElasticsearchConn")

    http_auth = None
    if params.get('HOST'):
      host = str(params.get('HOST').strip())
    if params.get('PORT'):
      port = str(params.get('PORT').strip())
    if params.get('USERNAME') and params.get('PASSWORD'):
      http_auth = tuple((str(params.get('USERNAME')),
                         str(params.get('PASSWORD'))))

    use_ssl = params.get('SSLCERT_NAME_MISMATCH', False)

    self.connection_type = conn_type

    self._es = Elasticsearch(host, port=port, http_auth=http_auth,
                             use_ssl=use_ssl,
                             sniff_on_start=True,
                             sniff_on_connection_fail=True,
                             sniffer_timeout=10*60,
                             maxsize=_MAX_CONNS_PER_NODE)

  def quote_tablename(self, fq_table_name):
    fields = fq_table_name.split(".")
    if len(fields) == 1:
      return '"%s"' % fields[0]

    #Ignore the database name
    dbname = fields[0]
    tablename = fields[1]
    return '"%s"' % (tablename)

  def properties(self):
    return ElasticsearchConnProperties(conn=self)

  def close(self):
    # We support reuse of connections. Don't close connection even if asked to
    pass

  def table_columns(self, tablename, username=None, client_ip=None):
    tablename = tablename.split(".")[-1]
    q = 'select * from "%s" limit 0' % tablename
    return self.sync_runquery(q)

  def async_cancel(self, qhandle):
    raise Exception('async_cancel not implemented')

  def async_status(self, qhandle, username=None, profile=None):
    raise Exception('async_status not implemented')

  def async_results(self, qhandle, username=None):
    raise Exception('async_results not implemented')

  def async_runquery(self, query,
                     username=None, sample_pct=None, comment=''):
    raise Exception('async_runquery not implemented')

  def sync_runquery(self, query, build_request=None, disable_codegen=True,
      username=None, sample_pct=None, comment='', stream=False, logQuery=True,
      hideQuery=False, qcntl=None, profile=None, setCancelCbFn=None):
    if query.startswith(SPECIAL_CMDS['ROWS']):
      tablename = query.split()[-1]
      tablename = tablename.split(".")[-1]
      query = 'select * from "%s" limit 100' % tablename

    query = self.swap_quotes(query)
    query = query.strip()
    if hideQuery:
      LOGGER.info('***')
    else:
      LOGGER.info(query)

    return self._runq(query)

  def _runq(self, query, **params):
    properties = self.properties()

    params['format'] = 'json'
    body = { 'query': query }
    results = self._es.transport.perform_request('POST', '/_xpack/sql',
                                                 params=params, body=body)

    rows = list()
    coltypes = [properties.cursor_coltype_to_arctype(x['type'].upper()) \
            for x in results['columns']]
    colnames = [x['name'] for x in results['columns']]
    rows.extend(results['rows'])

    # Drain the cursor for now 
    while 'cursor' in results:
      body = { 'cursor': results['cursor'] }
      results = self._es.transport.perform_request('POST', '/_xpack/sql',
                                                 params=params, body=body)
      rows.extend(results['rows'])

    del results

    data = {
      'colnames': colnames,
      'coltypes': coltypes,
      'info': ['Query: %s' % query],
      'rows': rows,
    }

    return data

  def execute(self, build_request, async_flag=False, stream=False):
    query = None
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      data = {
          'colnames': ['name'],
          'coltypes': ['STRING'],
          'rows': [[ELASTICSEARCH_DB_NAME]]
        }
      return data
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      query = 'show tables'
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            # if exception is caught above we set state to unknown
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = dataconn_dataengine.BuildRequestExecutor(build_request)
    return executor.execute(query=query,
                            async_flag=async_flag)

  def validate_expression(self, expression, build_request, exprType):
    data = {
      'is_valid': "true",
      'isaggregate': "false",
      'ispart': "false"
    }
    return data

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    derived_data = kwargs.get('derived_data')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset,
        derived_data=derived_data)
    hasaggr = 'false'

    try:
      self.sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if 'Cannot group by an aggregate' not in str(e):
        LOGGER.exception('could not validate')
        raise
      else:
        hasaggr = 'true'

    return hasaggr

if __name__ == "__main__":
  main()
