# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
from . import dataconn_arcengine
from .dataconn_dataengine import ConnProperties
from .connections_types import DATACONN_HIVE
import logging

from django.conf import settings
# create logger
LOGGER = logging.getLogger("arcweb")

from util.perfcounters import APPROX_COUNTERS

hive_create = APPROX_COUNTERS.get('hive.create')

class HiveConnProperties(ConnProperties):
  def __init__(self):
    super(HiveConnProperties, self).__init__()
    self.supports_limit_outside_union = False
    self.data_import_hdfs_location = '/user/hive/data'
    self.supports_deriveddata = True
    self.display_to_table_type = {
      "String": "string",
      "Integer": "bigint",
      "Real": "double",
      "Boolean": "boolean",
      "Timestamp": "timestamp"
    }

#
# Hive w/ LLAP does not support impersonation. In the context of CDW, viz apps
# runs within a trusted domain, and can login as the logged-in user. Do that
# to run queries as the logged-in user.
#
class HiveConn(dataconn_arcengine.ArcEngineConn):
  def __init__(self, params, connection_id, conn_type=DATACONN_HIVE):
    dataconn_arcengine.ArcEngineConn.__init__(self, params, connection_id,
        conn_type=conn_type)

    global hive_create
    hive_create += 1

  def properties(self):
    return HiveConnProperties()

  def table_columns(self, table_name, username=None, client_ip=None):
    out = super(HiveConn, self).table_columns(table_name, username=username,
        client_ip=client_ip)

    # Drop the table name from the column headers
    # TODO: Is there a hive setting to do this that we can use?
    fixed_colnames = []
    for c in out['colnames']:
      if c.startswith(table_name + '.'):
        fixed_colnames.append(c[len(table_name + '.'):])
      else:
        fixed_colnames.append(c)
    out['colnames'] = fixed_colnames

    return out

  def getconf(self, sample_pct=None, match_views=None):
    """These options are not supported for Hive"""
    return None

  def default_connection_settings(self):
    return {}

  def _import_data_create(self, tablename, headernames, headertypes,
                  delimiter, location,
                  importuser, importsrcfile):
    """ Run commands necessary to import data given headernames, types etc """
    properties = self.properties()
    sqlcmds = []

    # Specific to hive: if headernames have tablename embedded, remove them.
    oldheadernames = headernames
    headernames = []
    for hn in oldheadernames:
      idx = hn.find('.')
      if idx >= 0:
        hnt = hn[idx+1:]
      else:
        hnt = hn
      headernames.append(hnt)

    createimportstr = "CREATE TABLE " + tablename + "("
    for i in range(len(headernames)):
      eachheader = headernames[i]
      eachtype = headertypes[i]
      createimportstr = createimportstr + eachheader + " " + eachtype + ", "

    createimportstr = createimportstr[:-2] + ") row format delimited fields " +\
            "terminated by '" + delimiter + "' location '" +\
            properties.data_import_hdfs_location + location + "'" +\
            "TBLPROPERTIES('arc_importedby'='" + importuser +\
            "', 'arc_importedfile' = '" + importsrcfile + "')"

    return createimportstr

  def _import_data_move_file_to_hdfs(self, src_filename, tgt_filename):
    """ Move a file from its staged version at the webserver to HDFS cluster """
    # If connection has no user name, HDFS_USER needs to be specified in the
    # django settings if it's an instant set up. Or overwrite with default
    # of 'hive'

    # Imports needed for data import
    # Only import when needed for now
    from . import webhdfs
    import getpass

    properties = self.properties()
    hdfs_host = self.host
    if hdfs_host is None:
      raise Exception("No host specified for data import in connection")
    hdfs_user = self.username
    if not hdfs_user:
      hdfs_user = getattr(settings, 'HDFS_USER', None)
    if not hdfs_user:
      hdfs_user = 'hive'

    whdfscmd = ''
    try:
      whdfscmd = 'connect'
      whdfs = webhdfs.WebHDFS(host=hdfs_host, user=hdfs_user)
      whdfscmd = 'mkdir'
      whdfs.mkdir(properties.data_import_hdfs_location + tgt_filename)
      whdfscmd = 'put'
      whdfs.put(src_filename, properties.data_import_hdfs_location + tgt_filename +
                "/" + tgt_filename, overwrite=True)
    except Exception as e:
      raise Exception("Data import could not transfer file to host %s as user "
                      "%s over webhdfs: %s failed: %s" % (hdfs_host, hdfs_user,
                      whdfscmd, str(e)))

  def import_data(self, create, tablename, headernames, headertypes,
                  delimiter, location,
                  importuser, importsrcfile,
                  onlytablenamechanged, revised_tablename,
                  oldtablename,
                  movefile, movefilesrc, movefiletgt):

    sqlcmds = []
    failuresqlcmds = []

    if create == True:
      if movefile:
        self._import_data_move_file_to_hdfs(movefilesrc, movefiletgt)
      # This is for new creation of imported data
      sqlcmd = self._import_data_create(tablename, headernames,
                  headertypes, delimiter, location, importuser, importsrcfile)
      sqlcmds.append(sqlcmd)
      failuresqlcmds.append('DROP TABLE IF EXISTS ' + tablename)
    else:
      # This is for updating already imported data

      if onlytablenamechanged:
        sqlcmd = 'DROP TABLE IF EXISTS ' + revised_tablename
        sqlcmds.append(sqlcmd)
        sqlcmds.append('ALTER TABLE ' + oldtablename + ' RENAME TO ' +
             revised_tablename)
      else:

        if movefile:
          self._import_data_move_file_to_hdfs(movefilesrc, movefiletgt)

        # 1. create new table with new types/delimiters etc
        # 2. drop oldtable
        # 3. alter table rename newtable to be oldtable (to keep tablename same)
        # 4. if revised_tablename then alter table rename to be revisedtablename
        sqlcmd = self._import_data_create(tablename, headernames,
                    headertypes, delimiter, location, importuser, importsrcfile)
        sqlcmds.append(sqlcmd)
        failuresqlcmds.append('DROP TABLE IF EXISTS ' + tablename)

        sqlcmd = 'DROP TABLE ' + oldtablename
        sqlcmds.append(sqlcmd)

        sqlcmd = 'ALTER TABLE ' + tablename +\
                  ' RENAME TO ' + oldtablename
        sqlcmds.append(sqlcmd)

        if revised_tablename:
          sqlcmd = 'DROP TABLE IF EXISTS ' + revised_tablename
          sqlcmds.append(sqlcmd)
          sqlcmd = 'ALTER TABLE ' + oldtablename +\
                      ' RENAME TO ' + revised_tablename
          sqlcmds.append(sqlcmd)

    # Now, iterate over all the commands collected above and execute them
    sqlcmd = ''
    try:
      for sqlcmd in sqlcmds:
        self.log_sync_runquery(sqlcmd, username=importuser)
    except Exception as e:
      LOGGER.error('Import error in sqlcmd %s: %s' % (sqlcmd, str(e)))
       # If table created and then error, delete table, and re-raise exception
      fsqlcmd = ''
      try:
        for fsqlcmd in failuresqlcmds:
          self.log_sync_runquery(fsqlcmd, username=importuser)
      except Exception as e:
        LOGGER.error("Failure in clean up sqlcmd %s: %s" % (fsqlcmd, str(e)))
        pass
      raise
