# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import asyncio
import datetime
import json
import logging
import threading

from dataconn.dataconn_dbapi2 import DbApi2Conn
from vizlib import workq, misc
from vizlib.arcprof import ArcProf

LOG = logging.getLogger('arcweb.conn.http')

"""
This module simulates an "HTTP" dataconnection that is able to perform
non-blocking HTTP communication using
the aiohttp library. The client can use this service by sending
dsreq payloads to the dataconnection with id "-10".

The AsynRunner cleanup utility thread, which this module inherits is
only performs the handle cleanup. Task cancelling assumes blocking threads,
but here the only thread in use is the ASYNC_LOOP of the views_proxy module.

Fortunately the aiohttp default client timeout is 5 minutes. So unfinished connections
will be cleaned up after this time.

"""

def get_views_proxy():
  # can not import on top level because of circular dependency
  from arcwebbase import views_proxy
  return views_proxy

def get_send_ai_request():
  from arcwebbase.ai_service_client import send_ai_request
  return send_ai_request

def get_validate_and_transform_request():
  from arcwebbase.ai_validator import validate_and_transform_request
  return validate_and_transform_request

ASYNC_LOOP_LOCK = threading.Lock()
ASYNC_LOOP: asyncio.AbstractEventLoop = None

def get_async_loop():
  global ASYNC_LOOP
  with ASYNC_LOOP_LOCK:
    if ASYNC_LOOP is None:

      ASYNC_LOOP = asyncio.new_event_loop()

      def run_loop():
        try:
          LOG.info('Proxy AIO Loop start')
          asyncio.set_event_loop(ASYNC_LOOP)
          ASYNC_LOOP.run_forever()
        finally:
          LOG.info('Proxy AIO Loop end')

      threading.Thread(
        target=run_loop, name='HttpProxyAsyncLoop', daemon=True).start()

    return ASYNC_LOOP


class Dataconnection():
  def __init__(self, name, type, info):
    self.id = -10
    self.dataconnection_name = name
    self.dataconnection_type = type
    self.dataconnection_info = info
    self.dataconnection_password = None
    self.update_time = datetime.datetime.now()
    self.cache_sequence = 1

  def cache_settings(self):
    return {}


INSTANCE = Dataconnection(
  'HTTP', 'HTTP', json.dumps({'PARAMS': {'CONCURRENCY': 1, 'CONCURRENCY_USER': 1}}))


class HttpBuildRequest:
  def __init__(self, connection, username, userid,
               query, dataset_request, sample_pct,
               comment, qcntl, profile):
    self.connection = connection
    self.dataset_request = dataset_request
    data = dataset_request.dsreq.data['data']
    self.query = json.dumps(data)
    self.username = username
    self.profile = profile


class HTTPRunner(workq.ProgressRunner):
  def __init__(self, async_runner: workq.AsyncRunner, username: str, profile: ArcProf):
    self._results = None
    self._chunks = []
    self._task: asyncio.Task = None
    self._async_runner = async_runner
    self._profile = profile or ArcProf('httprunner', username=username)
    self._results = "Error: could not get results"
    self._completed = 0

  def toDisplayDict(self):
    return {
      'query': 'async HTTP request',
    }

  def __str__(self):
    return json.dumps(self.toDisplayDict())

  def __repr__(self):
    return json.dumps(self.toDisplayDict())

  def profile(self):
    return self._profile

  def progress(self):
    return (self._completed, 1)

  def complete(self):
    self._completed = 1

  def async_runner(self):
    return self._async_runner

  def interrupt(self):
    self._results = 'Error: HTTP request has been cancelled by user'

    if self._task:
      loop = get_async_loop()
      def async_cancel():
        LOG.info('cancelling query %s', self.profile().querytracker())
        self._task.cancel()

      loop.call_soon_threadsafe(async_cancel)
    pass

  def set_results(self, results):
    self._results = results

  def set_chunks(self, chunks):
    if chunks:
      self._chunks.extend(chunks)

  def get_chunks(self):
    return self._chunks

  def reset_chunks(self):
    self._chunks = []

  def results(self):
    return self._results

  def set_task(self, task: asyncio.Task):
    self._task = task


class HTTPConnProperties():
  def __init__(self):
    self.supports_reuse = True
    self.supports_async = False
    self.supports_dataextract_target = False


class HTTPConn(DbApi2Conn):
  def __init__(self, params, connection_id, conn_type='HTTP'):
    self.connection_id = connection_id
    self.connection_type = conn_type
    self._properties = HTTPConnProperties()

  def close(self):
    pass

  def properties(self):
    return self._properties

  def build(self, username=None, userid=None, query=None, dataset_request=None,
            sample_pct=None, comment=None, qcntl=None, profile=None):
    return HttpBuildRequest(connection=self, username=username, userid=userid,
                            query=query, dataset_request=dataset_request, sample_pct=sample_pct,
                            comment=comment, qcntl=qcntl, profile=profile)

  def execute(self, build_request: HttpBuildRequest, async_flag=True):

    misc.initAsyncConnectionThreadpool(self.connection_id)
    qhandle = workq.HANDLE_ID_ALLOCATOR.allocateId()
    async_runner = misc.get_async_runner(self.connection_id)

    http_runner = HTTPRunner(async_runner, build_request.username, build_request.profile)
    workq_runner = workq.Runner(
      http_runner, async_runner.queryTimeoutSeconds())
    workq_request_data = async_runner.workq().createRequestData(workq_runner)

    async_runner.start_tracking(qhandle, workq_runner)

    req = build_request.dataset_request.dsreq.data['data']
    run(req, workq_request_data)

    return {"qhandle": "%s" % qhandle}


def run(req: dict, workq_request_data: workq.RequestData):

  loop = get_async_loop()
  loop.call_soon_threadsafe(run_in_loop, req, workq_request_data)


def run_in_loop(req: dict, workq_request_data: workq.RequestData):
  task = asyncio.get_event_loop().create_task(
    _proxy_async(req, workq_request_data))

  http_runner: HTTPRunner = workq_request_data.runner().progress_runner()

  # used to enable cancelling:
  http_runner.set_task(task)


async def _proxy_async(req: dict, workq_request_data: workq.RequestData):
  views_proxy = get_views_proxy()
  send_ai_request = get_send_ai_request()
  validate_and_transform_request = get_validate_and_transform_request()

  workq_runner = workq_request_data.runner()
  http_runner: HTTPRunner = workq_runner.progress_runner()

  http_runner.async_runner().workq().statsStart(workq_request_data)
  workq_runner.set_status(workq.Runner.RUNNING)
  try:
    LOG.debug('proxy comm start')
    req = await validate_and_transform_request(req)
    model_type = req.get('modelType')

    streaming_enabled = req.get('settings', {}).get(f"{model_type}_stream", False)

    async for response_chunk in views_proxy.proxy_comm(req, send_ai_request):
      response_body = response_chunk['response_body']
      if streaming_enabled:
        http_runner.set_chunks(response_body)
      else:
        http_runner.set_results({'response': response_body})

    if streaming_enabled:
      # Send the leftover chunks
      chunks = http_runner.get_chunks()
      if chunks:
        resp = {
          'response': {
            'body': {
              'choices': [
                {
                  'message': {
                    'content': "".join(chunks)
                  }
                }
              ]
            }
          }
        }

        if response_chunk and response_chunk.get('req_info'):
          resp['clear_text_request'] = response_chunk.get('req_info')

        http_runner.set_results({'response': resp})
        http_runner.reset_chunks()

    LOG.debug('proxy comm success')
    http_runner.profile().markFromStart('query_runtime')
  except Exception as e:
    LOG.warn('proxy comm error', exc_info=True)
    http_runner.set_results(f'Error: {e}')
  finally:
    http_runner.complete()
    workq_runner.set_status(workq.Runner.DONE)
    http_runner.async_runner().workq().statsStop(workq_request_data)
