# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .connections_types import DATACONN_IMPYLA, DATACONN_HYVE, \
    DATACONN_ARCENGINE_V2, isImpalaBased
from django.conf import settings
from . import arctypes
from . import dataconn_dbapi2

from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from dataconn.cml_dependencies import get_cmldata

import time
import sys
import re
import json

from util.perfcounters import APPROX_COUNTERS
import jobs

import logging
LOGGER = logging.getLogger('arcweb')

try:
  import impala.dbapi as dbapi2
except Exception:
  LOGGER.info('Libraries missing for Impyla')


impyla_create = APPROX_COUNTERS.get('impyla.create')
impyla_connection = APPROX_COUNTERS.get('impyla.connection')
hyve_create = APPROX_COUNTERS.get('hyve.create')
hyve_connection = APPROX_COUNTERS.get('hyve.connection')
arcengine_v2_create = APPROX_COUNTERS.get('arcengine_v2.create')
arcengine_v2_connection = APPROX_COUNTERS.get('arcengine_v2.connection')

_TABLENAME_RE = re.compile(r"^\S+$")

#
# Sample code:
#
#    conn = connect(host='westeros.edh.cloudera.com',
#                   port=21050,
#                   auth_mechanism='GSSAPI',
#                   use_ssl=True,
#                   kerberos_service_name='impala')
#

class ImpylaConnProperties(ConnProperties):

  TYPEMAP = {

    'INT': arctypes.ARC_DATATYPE_INT,
    'BINARY': arctypes.ARC_DATATYPE_INT,
    'SMALLINT': arctypes.ARC_DATATYPE_INT,
    'TINYINT': arctypes.ARC_DATATYPE_INT,
    'BIGINT': arctypes.ARC_DATATYPE_BIGINT,

    'DOUBLE': arctypes.ARC_DATATYPE_DOUBLE,
    'FLOAT': arctypes.ARC_DATATYPE_DOUBLE,
    'DECIMAL': arctypes.ARC_DATATYPE_DOUBLE,

    'CHAR': arctypes.ARC_DATATYPE_STRING,
    'VARCHAR': arctypes.ARC_DATATYPE_STRING,
    'STRING': arctypes.ARC_DATATYPE_STRING,

    'DATE': arctypes.ARC_DATATYPE_TIMESTAMP,
    'TIMESTAMP': arctypes.ARC_DATATYPE_TIMESTAMP,

    'BOOLEAN': arctypes.ARC_DATATYPE_BOOLEAN,

    'MAP': arctypes.ARC_DATATYPE_STRING,
    'UNION': arctypes.ARC_DATATYPE_STRING,
    'ARRAY': arctypes.ARC_DATATYPE_STRING,
    'STRUCT': arctypes.ARC_DATATYPE_STRING,
  }

  def __init__(self, conn_type=DATACONN_IMPYLA):
    super(ImpylaConnProperties, self).__init__(conn_type)
    self.import_data_insertrows = 10000
    self.supports_async = False
    self.supports_reuse = False
    self.supports_deriveddata = True
    self.supports_tableproperties = True
    self.supports_nulls_order = True
    self.supports_dataextract_source = True
    self.supports_dataextract_target = True
    self.supports_truncate_table = True

    self.display_to_table_type = {
      "String": "string",
      "Integer": "bigint",
      "Real": "double",
      "Boolean": "boolean",
      "Timestamp": "timestamp"
    }

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    return ImpylaConnProperties.TYPEMAP.get(native_coltype,
        arctypes.ARC_DATATYPE_STRING)

QUERY_PERMITTED_STARTS = [
  "select", "create", "show", "with", "refresh", "describe",
  "invalidate metadata", "analyze", "compute", "explain"
]

if settings.IMPYLA_PERMIT_ADDITIONAL_CMDS:
  QUERY_PERMITTED_STARTS.extend(settings.IMPYLA_PERMIT_ADDITIONAL_CMDS)

def queryPermitted(query):
  query = (query or '').strip().lower()

  for s in QUERY_PERMITTED_STARTS:
    if query.startswith(s):
      return True

  return False

class ImpylaConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_IMPYLA):
    global impyla_create
    global arcengine_v2_create
    global hyve_create

    if conn_type == DATACONN_IMPYLA:
      impyla_create += 1
    elif conn_type == DATACONN_HYVE:
      hyve_create += 1
    elif conn_type == DATACONN_ARCENGINE_V2:
      arcengine_v2_create += 1
    else:
      raise Exception("Invalid connection type %s" % conn_type)

    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    conn_params = {}

    self._properties = self.properties()

    self.params_settings = params.get('SETTINGS', {})
    self.for_testing = False
    if params.get('FOR_TESTING'):
      self.for_testing = True

    self.aviews_extended = 'extended '
    self.host = params['HOST']
    port = params['PORT']
    auth = params.get('AUTH', 'NOSASL')
    auth = auth.upper()
    if auth == 'KERBEROS':
      auth = 'GSSAPI'

    if auth == 'JWT':
      conn_params['jwt'] = get_cmldata().get_jwt()

    self.username = params.get('USERNAME', None)
    self.password = params.get('PASSWORD', None)
    ssl = params.get('SOCK', '').upper() in ['SSL', 'SSLCERT']
    sslcert = None
    if params.get('SOCK', '').upper() in ['SSLCERT']:
      sslcert = getattr(settings, 'SSL_CERT_PATH', None)
    kerbserv = params.get('KERBSERV', '')
    if len(kerbserv) == 0:
      if conn_type == DATACONN_IMPYLA:
        kerbserv = 'impala'
      if conn_type == DATACONN_ARCENGINE_V2:
        kerbserv = 'arcengine'
      elif conn_type == DATACONN_HYVE:
        kerbserv = 'hive'

    krb_host = params.get('KRB_HOST')
    if krb_host:
      conn_params['krb_host'] = krb_host

    self.connection_id = connection_id
    self.application_name = params.get('APP_NAME', None)
    self.connection_type = conn_type
    self.impersonation = params.get('IMPERSONATION', False)
    self.impersonation_trusted = params.get('IMPERSONATION_TRUSTED', False)
    mode = params.get('MODE', 'binary')
    if mode.upper() == 'HTTP':
      http_service = params.get('HS2_HTTP_SQLPATH') or 'cliservice'
      http_access_token = params.get('HS2_HTTP_ACCESS_TOKEN')
    else:
      http_service = None
      http_access_token = None

    # Default to True here so that old connections without this parameter
    # will continue to work since we had the cert checking disabled for
    # a while.
    sslcert_name_mismatch = params.get('SSLCERT_NAME_MISMATCH', True)

    try:
      socket_timeout = int(params.get('SOCKET_TIMEOUT', None))
    except Exception:
      socket_timeout = None

    conn_config = {'max_connections':40,
                   'max_requests':40,
                   'max_reqperconn':1}
    if self.impersonation:
      if not getattr(settings, 'HS2CONN_ONE_REQUEST_PER_CONN', False):
        conn_config = {'max_connections':40,
                       'max_requests':400,
                       'max_reqperconn':40}

    # Allow queue config override
    conn_config = getattr(settings, 'HS2_QUEUE_SETTINGS', conn_config)

    if params.get('FOR_TESTING'):
      return

    pure_sasl = sys.platform == 'win32' or \
                getattr(settings, 'USE_PURE_PYTHON_SASL', False)

    conn_params['host'] = self.host
    conn_params['port'] = int(port)
    if ssl:
      conn_params['use_ssl'] = True
      conn_params['ca_cert'] = sslcert

    conn_params['kerberos_service_name'] = kerbserv
    conn_params['user'] = self.username
    conn_params['timeout'] = socket_timeout
    conn_params['auth_mechanism'] = auth

    if mode.upper() == 'HTTP':
      conn_params['use_http_transport'] = True
      conn_params['http_path'] = http_service

    # LOGGER.info('Connection params: %s', conn_params)

    conn_params['password'] = self.password
    self.conn_params = conn_params

    self.connection_type = conn_type

  # Overridden function to get a connection object if we need one
  def get_connection(self, username=None):
    global impyla_connection
    global arcengine_v2_connection
    global hyve_connection

    if self.for_testing:
      return self.connection

    if self.connection:
      return self.connection

    try:
      if self.connection_type == DATACONN_IMPYLA:
        impyla_connection += 1
        self.conn_params['http_cookie_names'] = ['impala.session.id']
      elif self.connection_type == DATACONN_HYVE:
        hyve_connection += 1
      elif self.connection_type == DATACONN_ARCENGINE_V2:
        arcengine_v2_connection += 1

      if self.impersonation_trusted:
        self.conn_params['user'] = username
        if self.conn_params.get('password'):
          del self.conn_params['password']

      self.connection = dbapi2.connect(**self.conn_params)
    except Exception as e:
      LOGGER.error(e)
      raise Exception('Could not connect to database')

    return self.connection

  def cursorConfiguration(self, username=None):
    # Merge possible custom settings with defaults
    conn_settings = self.default_connection_settings().copy()
    conn_settings.update(self.params_settings)

    if self.impersonation:
      # Allow translation of doasuser (assumed to be arcviz username) through user profile key
      profkey = getattr(settings, 'PROFILE_IMPERSONATION_KEY', None)
      if not profkey and getattr(settings, 'PROFILE_PROXY_USERNAME', False):
        profkey = 'proxy_username'
      if profkey and username:
        _username = username
        from apps.models import profile_getkey_for_username
        username = profile_getkey_for_username(_username, profkey, _username)
        LOGGER.info(f'Opening session with login username {_username}, doasuser {username}')

      if self.connection_type == DATACONN_IMPYLA:
        conn_settings['impala.doas.user'] = username
      elif self.connection_type == DATACONN_ARCENGINE_V2:
        # Note that arcengine also uses the connection setting named impala.doas.user
        conn_settings['impala.doas.user'] = username
      else:
        conn_settings['hive.server2.proxy.user'] = username

    return conn_settings

  def default_connection_settings(self):
    if self.connection_type == DATACONN_IMPYLA:
      return self._default_connection_settings_impala()
    elif self.connection_type == DATACONN_ARCENGINE_V2:
      return self._default_connection_settings_arcengine_v2()

    return self._default_connection_settings_hive()

  def _default_connection_settings_hive(self):
    d = {}

    if self.application_name:
      d['applicationName'] = self.application_name

    return d

  def _default_connection_settings_arcengine_v2(self):
    d = {
      'EXEC_SINGLE_NODE_ROWS_THRESHOLD': '200000',
      'OPTIMIZE_PARTITION_KEY_SCANS': '1'
    }

    return d

  def _default_connection_settings_impala(self):
    d = {
      'EXEC_SINGLE_NODE_ROWS_THRESHOLD': '200000',
      'OPTIMIZE_PARTITION_KEY_SCANS': '1'
    }

    if self.application_name:
      d['CLIENT_IDENTIFIER'] = self.application_name

    return d

  def properties(self):
    return ImpylaConnProperties()

  def apply_types(self, value, type):
    try:
      if value is None:
        value = "NULL"
      elif type == "String":
        value = re.sub(r"\\", r"\\\\", value)
        value = re.sub("'", r"\'", value)
        value = "'%s'" % value
      elif type == "Timestamp":
        value = "'%s'" % value
      elif type == "Real":
        value = re.sub(",", ".", value)
    except Exception as e:
      pass

    return value

  def execute(self, build_request, async_flag=False, stream=False):
    query = build_request.query
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = "show databases"
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      # Note that the escaping of the db name is important below
      # Try using a database called default without it
      query = "show tables in `%s`" % build_request.query.split()[-1]
    elif build_request.query == SPECIAL_CMDS['INVALIDATE_METADATA']:
      table_name = build_request.builder.dataset_request.dsreq.table_name
      query = 'invalidate metadata ' + (self.quote_tablename(table_name) if table_name else "")
    elif build_request.query == SPECIAL_CMDS['COMPUTE_STATS']:
      table_name = build_request.builder.dataset_request.dsreq.table_name
      tables = (self.quote_tablename(table_name) if table_name else "")
      if self.connection_type == DATACONN_HYVE:
        query = f'analyze table {tables} compute statistics'
      else:
        query = f'compute incremental stats {tables}'
    elif build_request.query == SPECIAL_CMDS['REFRESH_AVIEWS']:
      query = 'refresh analytical view ' + build_request.builder.dataset_request.dsreq.viewname
    elif build_request.query == SPECIAL_CMDS['DROP_AVIEWS']:
      return self._execute_drop_analytical_views(build_request, async_flag)
    elif build_request.query == SPECIAL_CMDS['SHOW_AVIEWS']:
      return self._execute_show_analytical_views(build_request)
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            # if exception is caught above we set state to unknown
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': error_message + ': ' + str(e)}
          return data

    executor = BuildRequestExecutor(build_request)

    # LOGGER.info(query)

    self.validate_query(query, build_request.username)

    return executor.execute(query=query,
                            async_flag=async_flag,
                            stream=stream)

  def validate_query(self, query, username):
    """Ensure the query is allowed"""
    if not '*' in settings.RAWSQL_WHITELIST:
      if not username in settings.RAWSQL_WHITELIST:
        query = query or ''
        if not queryPermitted(query):
          raise Exception("%s queries are not permitted" % query[:6].lower())

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      s = str(e)

      # Support Hive and Impala errors to detect aggregates
      if "expression not produced by aggregation" in s or \
         'in GROUP BY' in s or \
         "Can't group on" in s:
        hasaggr = 'true'
      else:
        raise

    return hasaggr

  # Run a 'show [extended] analytical views' type command with checking
  # old vs new arcengine then adjust+re-run if needed.
  # We do this by detecting mis-match and switching what command we
  # use to get the aview data.
  def _run_view_command(self, viewfunc):
    try:
      rsp = viewfunc()
      if not self.aviews_extended and len(rsp.get('colnames')) == 1:
        self.aviews_extended = 'extended '
        rsp = viewfunc()
    except Exception as e:
      if self.aviews_extended and str(e).find('Encountered: EXTENDED') != -1:
        self.aviews_extended = ''
        rsp = viewfunc()
      else:
        raise
    return rsp

  # Get view state for one analytical view
  def show_analytical_view(self, viewname, username=None):
    def viewfunc():
      return self._show_analytical_view_internal(viewname, username=username)
    return self._run_view_command(viewfunc)

  def _show_analytical_view_internal(self, viewname, username=None):
    prefix = 'show %sanalytical views ' % self.aviews_extended
    parts = viewname.split('.')
    if len(parts) == 2:
      suffix = 'in `%s` like "%s"' % (parts[0], viewname)
    else:
      suffix = 'like "%s"' % viewname
    return self.sync_runquery(prefix + suffix, username=username)

  def _execute_show_analytical_views(self, build_request):
    def viewfunc():
      return self._execute_show_analytical_views_internal(build_request)
    return self._run_view_command(viewfunc)

  def _execute_show_analytical_views_internal(self, build_request):
    # In secure environments, 'show all analytical views' does not
    # work correctly. Allow a setting for enumerating through dbs individually
    # and ignoring any security (or other issues) in those cases.
    # Once 'show all analytical views' works in secure environments, we
    # can remove support for the 'SHOW_ALL_AVIEWS_ENUMERATES_DBS' setting
    if settings.SHOW_ALL_AVIEWS_ENUMERATES_DBS:
      # First, fetch the list of databases. Then, enumerate views in each db,
      # ignoring any errors (likely security-related) in the process
      query = 'show databases'
      executor = BuildRequestExecutor(build_request)
      dbs = executor.execute(query=query, async_flag=False)
      views = []

      rows_parsed = json.loads(dbs['rows'])
      for row in rows_parsed:
        db = row[0]
        query = "show %sanalytical views in `%s`" % (self.aviews_extended, db)
        executor = BuildRequestExecutor(build_request)
        try:
          v = executor.execute(query=query, async_flag=False)
          if not views:
            views = v
            # Check here as well so we don't incorrectly iterate over
            # all DBs only to do it again with 'extended' keyword.
            if not self.aviews_extended and len(v.get('colnames')) == 1:
              break
          else:
            views['rows'].extend(v['rows'])
        except Exception as e:
          if str(e).find('Encountered: EXTENDED') != -1:
            raise
          # print 'Ignoring exception', e
      return views
    else:
      query = 'show %sall analytical views' % self.aviews_extended
      executor = BuildRequestExecutor(build_request)
      return executor.execute(query=query, async_flag=False)

  def _execute_drop_analytical_views(self, build_request, async_flag=False, stream=False):
    # This function deletes jobtargets relating to an analytical view
    # before executing the query to delete an analytical view.
    try:
      view_info = build_request.dataset_request.dsreq.viewname.split('.')
      # Strip the enquoting here
      db_name = view_info[0][1:-1]
      view_name = view_info[1][1:-1]
      complete_name = db_name + '.' + view_name
    except:
      error_message = 'Error parsing viewname'
      LOGGER.exception(error_message)
      data = { 'error_msg': error_message }
      return data

    models = jobs.models
    targets = models.JobTarget.objects.filter(target_type='R')
    for target in targets:
      target_data = json.loads(target.target_data)
      if (target_data['dcid'] == self.connection_id and complete_name == target_data['view']):
        # Right now, a schedule can only contain one RefreshJobTarget, so we delete the target
        # and it's associated schedule. If it becomes possible that a JobSchedule can hold
        # RefreshJobTargets and other JobTargets (ie we implement dependencies on RefreshJobTarget,
        # than we need to change this line)
        target.schedule.delete()

    query = 'drop analytical view ' + build_request.builder.dataset_request.dsreq.viewname
    executor = BuildRequestExecutor(build_request)
    return executor.execute(query=query, async_flag=async_flag, stream=stream)

  def base_table_columns(self, fq_table_name, username=None, client_ip=None):
    if self.connection_properties.supports_limit_0:
      limit_str = "limit 0"
    else:
      limit_str = "limit 1"

    query = 'select * from %s %s' % (self.quote_tablename(fq_table_name), limit_str)

    d = self.log_sync_runquery(query, username=username,
        client_ip=client_ip)
    d['partitions'] = []
    return d

  def table_columns(self, table_name, username=None, client_ip=None):
    if not isImpalaBased(self.connection_type):
      return self.base_table_columns(table_name, username, client_ip)

    partitions = []
    hive_comments = []
    if _TABLENAME_RE.match(table_name.strip()):
      (partitions, hive_comments) = self._partitions_and_comments(table_name, username=username,
          client_ip=client_ip)

    out = self._columns(table_name, username=username, client_ip=client_ip)
    if not out:
      raise Exception("Could not fetch columns for %s" % table_name)

    # Add in partitioning information
    out['partitions'] = partitions
    out['hive_comments'] = hive_comments
    return out

  def _partitions_and_comments(self, table_name, username=None, client_ip=None):
    out = self.log_sync_runquery(
        "describe formatted %s" % self.quote_tablename(table_name),
        username=username,
        client_ip=client_ip)
    return self._describe_output_to_partitions_and_comments(out)

  def _columns(self, table_name, username=None, client_ip=None):
    query = "describe %s" % self.quote_tablename(table_name)
    rsp = self.log_sync_runquery(query,
        username=username,
        client_ip=client_ip)

    # Split but handle case where comma can be embedded in <>, ()
    def split_comma(text):
      input = text.split(',')
      olist = []
      item = ''
      while len(input):
        item += input.pop(0)
        if item.count('<') != item.count('>') or \
           item.count('(') != item.count(')'):
          item += ','
        else:
          olist.append(item)
          item = ''
      return olist

    # Populate cols list with tuple for each element within a possibly
    # complex type defintion.  Couple format: (name, type, parent)
    def get_type(name, text, parent, cols):

      type, lt, text = text.partition('<')
      if not lt:
        cols.append((name, type, parent))
        return

      # Hive translate map and array types to string
      if not isImpalaBased(self.connection_type):
        if type in ['map', 'array']:
          type = 'string'
          cols.append((name, 'string', parent))
          return

      text = text[:-1]
      cols.append((name, type, parent))
      parent = name

      if type == 'map':
        (ktype, comma, vtype) = text.partition(',')
        cols.append((name + '.key', ktype, parent))
        get_type(name + '.value', vtype, parent, cols)
      elif type == 'array':
        cols.append((name + '.pos', 'int', parent))
        get_type(name + '.item', text, parent, cols)
      elif type == 'struct':
        for pair in split_comma(text):
          (sname, colon, stype) = pair.partition(':')
          get_type(name + '.' + sname, stype, parent, cols)
      else:
        raise Exception('Unknown complex type: %s' % type)

    colnames = []
    coltypes = []

    rows_parsed = json.loads(rsp['rows'])
    for row in rows_parsed:
      # New hive includes partition information in the output so we need
      # to exit this loop when we hit the empty column name that preceeds
      # the partition info.
      if not row[0]:
        break

      # Hive 2.1.0 seems to have dropped the blank line before the partition
      # info (ARC-10410), so also quit on that.
      if row[0].startswith('# Partition Information'):
        break

      colname = row[0]
      coltype = re.sub(r'\s', '', row[1])

      cols = []
      get_type(colname, coltype, '', cols)
      for (name, type, parent) in cols:
        colnames.append(name)
        # decimal type values will be of the format decimal(precision[,scale])
        if 'decimal' in type:
          type = self._properties.TYPEMAP.get('DECIMAL')
        elif (type not in ['map', 'struct', 'array']):
          type = self._properties.TYPEMAP.get(type.upper(), arctypes.ARC_DATATYPE_STRING)
        coltypes.append(type)

    data = {
      'colnames': colnames,
      'coltypes': coltypes,
      'info': rsp['info'],
      'rows': [],
    }

    return data

  def _describe_output_to_partitions_and_comments(self, out):
    """
    Given the output of a describe formatted command,
    extract the partitioning columns and column comments
    """
    partitions = []
    comments = []
    start_partitions = False
    start_comments = False

    # Resulting arcengine call returns rows with the format:
    # [u'# col_name            ', u'data_type           ', u'comment             ']
    # Thus, set comment_index = 2 to extract the column comments
    comment_index = 2
    rows_parsed = json.loads(out['rows'])
    for row in rows_parsed:
      if len(row) == 0:
        continue
      colname = row[0].strip()
      if not colname:
        continue

      # End marker for column information
      if colname.startswith("# Detailed Table Information"):
        return (partitions, comments)

      # Start marker for column comments
      if colname == "# col_name":
        start_comments = True
        continue

      # Start marker for partitions
      if colname == "# Partition Information":
        start_partitions = True
        continue

      # Ignore comment rows
      if colname.startswith("#"):
        continue

      if start_comments:
        comment = row[comment_index]
        if comment:
          comment = comment.strip()
        if not comment or comment == "NULL":
          comments.append("")
        else:
          comments.append(comment)

      if start_partitions:
        partitions.append(colname)

    return (partitions, comments)

