# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging, json
import time
import pprint
import threading
from . import dataconn_dbapi2
from . import arctypes
from .connections_types import DATACONN_KSQL
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from vizlib.workq import AsyncRunner

# create logger
LOGGER = logging.getLogger('arcweb')

RESULT_TYPE_TO_ROWS_FIELD = {
  'kafka_topics': ['topics'],
  'tables': ['tables'],
  'sourceDescription': ['sourceDescription', 'fields'],
  'queryDescription': ['queryDescription'],
  'function_names': ['functions'],
  'streams': ['streams'],
  'queries': ['queries'],
  'properties': ['properties'],
}

try:
  import urllib3
  urllib3.disable_warnings()
except Exception as e:
  LOGGER.info("Did not find urllib3 for KSQL")

_QUERY_EXPIRATION_TIME = 120
_KSQL_MAX_THREADS_PER_CONN = 50

_HTTP = urllib3.PoolManager()

class KsqlConnProperties(ConnProperties):
  TYPEMAP = {
    'BOOLEAN': arctypes.ARC_DATATYPE_BOOLEAN,
    'INTEGER': arctypes.ARC_DATATYPE_INT,
    'BIGINT': arctypes.ARC_DATATYPE_BIGINT,
    'DOUBLE': arctypes.ARC_DATATYPE_DOUBLE,
    'VARCHAR': arctypes.ARC_DATATYPE_STRING,
    'MAP': arctypes.ARC_DATATYPE_STRING,
    'ARRAY': arctypes.ARC_DATATYPE_STRING,
  }

  def __init__(self):
    super(KsqlConnProperties, self).__init__()
    self.supports_async = True
    self.property_map = dict()

    self.supports_groupby_colindex = False
    self.supports_orderby = False

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    coltype_parsed = native_coltype
    if coltype_parsed.find('ARRAY') > -1:
      coltype_parsed = 'ARRAY'
    elif coltype_parsed.find('MAP') > -1:
      coltype_parsed = 'MAP'
    elif coltype_parsed.find('VARCHAR') > -1 or coltype_parsed.find('STRING') > -1:
      coltype_parsed = 'VARCHAR'

    return KsqlConnProperties.TYPEMAP.get(coltype_parsed, arctypes.ARC_DATATYPE_STRING)

  def add_property(self, name, value):
    self.property_map[name] = value

#
# Per-connection query tracker
# This also tracks all active data consumers
#
# Each running query is encapsulated by an instance of _KSQLQueryRunner, which
#   is responsible for continually getting data from the stream and stuffing
#   the data into a _KSQLDataStore instance.
#
# Consumers come in and get data from a shared _KSQLDataStore instance
#   via a _KSQLDataConsumer instance
#
#
class _KSQLQueryTracker(object):
  def __init__(self):
    # queryRunnersByQuery is a dict of _KSQLQueryRunner instances
    self.queryRunnersByQuery = {}
    self.lock = threading.Lock()
    self.dataConsumers = {}        # key'd on ID
    self.dataConsumerGlobalIdAllocator = 1

  def getQuery(self, query=None):
    if query is None:
      raise Exception("getQuery: query must be specified")

    with self.lock:
      return self.queryRunnersByQuery.get(query, None)

  def trackQuery(self, queryRunner=None):
    query = queryRunner.queryTrackerName()
    with self.lock:
      self.queryRunnersByQuery[query] = queryRunner

  # Since each dataConsumer is associated with a query/queryRunner...then
  # when the query is untracked (mostly when no one has done asyncstatus on
  # the underlying dataStore for a while), untrack the associated
  # dataConsumers as well
  def untrackQuery(self, queryRunner=None):
    query = queryRunner.queryTrackerName()

    with self.lock:
      q = self.queryRunnersByQuery.get(query, None)
      if q == queryRunner:
        # This destroys the KSQLQueryRunner instance created for each query
        del self.queryRunnersByQuery[query]
        # There is still reference to dataConsumers lying around in the
        # self.dataConsumers[byid] hash.
        # They will be removed through a check_expired call done at a new
        # DataConsumer call (This is not 100% clean since it means we will
        # always have the last few instances of stale DataConsumers lying
        # around but it's good enough for now since we do not expect a high
        # volume of DataConsumers
        # We do not force mark them as expired - they expire when there has
        # been no async_status check on them for QUERY_EXPIRATION_TIME
        # For a cache reset, we do force them as expired since that means no
        # query should access the data from them anymore

  def resetAllDataStores(self):
    """ This effective resets the connection, all queries, consumers, data stores """
    with self.lock:
      for key in self.queryRunnersByQuery:
        self.queryRunnersByQuery[key].dataStore().set_expired()
      for key in self.dataConsumers:
        self.dataConsumers[key].set_expired()

  def deleteExpiredDataConsumers(self):
    """ This deletes any DataConsumers that have been marked as expired either due to timeout or due to cache reset """
    with self.lock:
      LOGGER.debug('Data Consumers before expiry deletion %s' % len(self.dataConsumers.keys()))
      toDelete = []
      for key in self.dataConsumers:
        if self.dataConsumers[key].check_expired():
          toDelete.append(key)
      for eachToDelete in toDelete:
        del self.dataConsumers[eachToDelete]
      LOGGER.debug('Data Consumers after expiry deletion %s' % len(self.dataConsumers.keys()))

  def getDataConsumer(self, dataConsumerId=None):
    if dataConsumerId is None:
      raise Exception("getDataConsumer: dataConsumerId must be specified")

    with self.lock:
      return self.dataConsumers.get(dataConsumerId, None)

  def trackDataConsumer(self, dataConsumer=None):
    with self.lock:
      dataConsumerId = self.dataConsumerGlobalIdAllocator
      self.dataConsumerGlobalIdAllocator += 1
      self.dataConsumers[dataConsumerId] = dataConsumer

    dataConsumer.setId(dataConsumerId)

  def untrackDataConsumer(self, dataConsumer=None):
    """ no one seems to be calling this function """
    dataConsumerId = dataConsumer.id()

    with self.lock:
      if dataConsumerId in self.dataConsumers:
        del self.dataConsumers[dataConsumerId]

# Having this function instead of using time.time directly in the
# code below allows us to mock it in tests
def currentTime():
  return time.time()

class _KSQLDataStore(object):
  def __init__(self):
    self._alldone = False
    self._last_client_query = currentTime()
    self._close_conn = False
    self._startpos = 0
    self._max = 1000
    self._deleteCount = 100
    self._lock = threading.Lock()
    self._rows = []
    self._errorString = None
    self._colnames = []

  def colnames(self):
    return self._colnames

  def set_colnames(self, c):
    self._colnames = c

  def set_errorString(self, e):
    self._errorString = e

  def errorString(self):
    return self._errorString

  def set_alldone(self):
    self._alldone = True

  def alldone(self):
    return self._alldone

  def positions(self):
    with self._lock:
      return (self._alldone, self._startpos, self._startpos + len(self._rows))

  def fetch_all(self, pos=0):
    if pos < 0:
      pos = 0

    vals = []
    with self._lock:
      end = self._startpos + len(self._rows)
      if pos >= end:
        return vals, end

      if pos <= self._startpos:
        vals.extend(self._rows)
        return vals, end

      vals.extend(self._rows[pos - self._startpos:])
      return vals, end

  def append(self, v):
    self._rows.append(v)
    if len(self._rows) > self._max:
      with self._lock:
        if len(self._rows) > self._max:
          toDel = len(self._rows) - self._max + self._deleteCount
          self._startpos += toDel
          self._rows = self._rows[toDel:]

  def set_expired(self):
    self._close_conn = True

  def check_expired(self):
    # For now, terminate streaming queries if clients have not fetched from it
    # in the last _QUERY_EXPIRATION_TIME seconds.
    # Also terminate queries if KSQLDataStore is set to expired by a KSQL
    # cache reset (set_expired)
    if (currentTime() - self._last_client_query > _QUERY_EXPIRATION_TIME):
      self._close_conn = True

    return self._close_conn

  def update_last_time(self):
    self._last_client_query = currentTime()

class _KSQLDataConsumer(object):
  def __init__(self, dataStore=None, query=None, conn=None):
    self._dataStore = dataStore
    self._query = query
    self._conn = conn

    self.table_vals = []

    self._last_client_query = currentTime()
    self._close_consumer = False
    self._dataConsumerId = None
    self._curr_pos = 0

    self._conn.queryTracker().trackDataConsumer(dataConsumer=self)
    self._tracked = True

  def query(self):
    return self._query

  def setId(self, id):
    self._dataConsumerId = id

  def id(self):
    return self._dataConsumerId

  def untrack(self):
    """ no one seems to be calling this function """
    if self._tracked:
      LOGGER.debug('deallocating %s', self._dataConsumerId)
      self._conn.queryTracker().untrackDataConsumer(dataConsumer=self)
      self._tracked = False

  def dataConsumerId(self):
    return self._dataConsumerId

  def errorString(self):
    return self._dataStore.errorString()

  def alldone(self):
    (alldone, start, end) = self._dataStore.positions()
    return alldone

  def __len__(self):
    (alldone, start, end) = self._dataStore.positions()
    if self._curr_pos < start:
      self._curr_pos = start

    ret = end - self._curr_pos
    return ret

  def colnames(self):
    return self._dataStore.colnames()

  def pop_all(self):
    (data, self._curr_pos) = self._dataStore.fetch_all(self._curr_pos)
    return data

  def append(self, v):
    return self._dataStore.append(v)

  def set_expired(self):
    self._close_consumer = True

  def check_expired(self):
    # For now, terminate streaming queries if clients have not fetched from it
    # in the last _QUERY_EXPIRATION_TIME seconds.
    # Also terminate queries if DataConsumers are set to expired by a KSQL
    # cache reset (set_expired)
    if (currentTime() - self._last_client_query > _QUERY_EXPIRATION_TIME):
      self._close_consumer = True

    return self._close_consumer

  def update_last_time(self):
    self._last_client_query = currentTime()
    self._dataStore.update_last_time()

class _KSQLQueryRunner(object):
  @staticmethod
  def getQueryTrackerName(query, earliest):
    return query + ";" + str(earliest)

  @staticmethod
  def getDataConsumerAndQueryRunner(query=None, conn=None, admin=None,
      sync=False, timeout=None, runAsyncQuery=True, earliest=False):

    # Get an existing query runner for this query, or create a new one
    queryName = _KSQLQueryRunner.getQueryTrackerName(query, earliest)
    queryRunner = conn.queryTracker().getQuery(query=queryName)
    if not queryRunner:
      queryRunner = _KSQLQueryRunner(query=query, conn=conn,
          admin=admin, sync=sync, timeout=timeout, earliest=earliest)
      if runAsyncQuery:
        conn.queueQuery(queryRunner)

    # Delete any expired DataConsumers
    conn.queryTracker().deleteExpiredDataConsumers()

    # Create a new data consumer for this query
    dataConsumer = _KSQLDataConsumer(dataStore=queryRunner.dataStore(), query=queryName, conn=conn)

    return (dataConsumer, queryRunner)

  def __init__(self, query=None, conn=None, admin=None, sync=False, timeout=None,
      earliest=False):
    LOGGER.info("Created thread: %s", query)
    self._query = query
    self._dataStore = _KSQLDataStore()
    self.conn = conn
    self.timeout = timeout
    self.sync = sync
    self.admin = admin
    self.response = None
    self.nonRows = []
    self.containsErrors = False
    self._errorString = None
    self.resultData = None
    self._earliest = earliest
    self._queued = False

    self.conn.queryTracker().trackQuery(queryRunner=self)

    if admin is None:
      self.admin = self.isAdminQuery()
      self.sync = self.admin

  def queryTrackerName(self):
    return _KSQLQueryRunner.getQueryTrackerName(self._query, self._earliest)

  def errorString(self):
    return self._errorString

  def queued(self):
    return self._queued

  def setQueued(self):
    self._queued = True

  def dataStore(self):
    return self._dataStore

  def query(self):
    return self._query

  def isAdminQuery(self):
    queryStart = self._query.split()[0].lower()
    if queryStart in ['select']:
      return False

    return True

  def processResponse(self, line):
    if type(line) == type('') or type(line) == type(''):
      jsonRow = json.loads(line)
    else:
      jsonRow = line

    try:
      resultType = jsonRow.get('@type')
      if resultType in RESULT_TYPE_TO_ROWS_FIELD:
        path = RESULT_TYPE_TO_ROWS_FIELD[resultType]

        l = jsonRow
        for element in path:
          l = l[element]

        if resultType == 'queryDescription':
          self._dataStore.set_colnames(['Description'])
          self._dataStore.append([json.dumps(l)])
          return

        if type(l) == type({}):
          self._dataStore.set_colnames(['Name', 'Value'])
          for k, v in l.items():
            self._dataStore.append([k, v])

        if type(l) == type([]):
          for row in l:
            if not self._dataStore.colnames():
              if resultType == 'sourceDescription':
                self._dataStore.set_colnames(['Name', 'Type'])
              else:
                self._dataStore.set_colnames(list(row.keys()))

            elif resultType == 'sourceDescription':
              self._dataStore.append([row['name'], row['schema']['type']])
            else:
              self._dataStore.append(list(row.values()))

        return
    except Exception as e:
      LOGGER.exception(e)
      pass

    row = self.conn._get_row(jsonRow)
    if row is None:
      if type(jsonRow) is type({}):
        for k, v in jsonRow.items():
          if type(v) is type({}):
            otherValue = None
            found = False
            for k2, v2 in v.items():
              if k2 == 'statementText':
                found = True
                continue

              if k2 == "errorMessage":
                self.containsErrors = True
                otherValue = [v2]
                continue

              if k2 == "commandStatus":
                otherValue = [v2]
                continue

              if not otherValue:
                otherValue = v2

            if found:
              if type(otherValue) == type([]):
                self.resultData = listToResultData(otherValue)

      # TODO: When there are errors in the stream, they show up here.
      # We need to be able to detect when these indicate errors, and to
      # then process those errors appropriately. Kafka does not indicate
      # error states cleanly
      self.nonRows.append(line)
    else:
      self._dataStore.append(row)

  def getLines(self, response):
    data = ''

    while True:
      if response.closed:
        return

      for chunk in response.stream(1024):
        data = data + chunk.decode('utf-8')
        idx = data.find("\n")
        if idx >= 0:
          line = data[:idx]
          data = data[idx+1:]
          yield line

  def run(self):
    LOGGER.info("Running thread: %s", self._query)

    # One of these threads should be running for any active _KSQLDataStore

    try:
      self.response = self.conn._ksql_query(self._query, admin=self.admin,
               sync=self.sync, timeout=self.timeout, earliest=self._earliest)
      if type(self.response) == type([]):
        for line in self.response:
          self.processResponse(line)
        self.response = None
        return

      while True:
        if not self.response:
          break

        for line in self.getLines(self.response):
          if self._dataStore is not None and line != '':
            self.processResponse(line)
            if self._dataStore.check_expired():
              LOGGER.debug('Exiting because dataStore expired')
              return 'Streaming Query Exited'

        if self.admin:
          self.response = None
          break

        # An ksql async query with a limit will end up in this if case
        if self.response.closed:
          LOGGER.debug('KSQL Query Runner run: Done - calling _close()')
          self._close()
          self._dataStore.set_alldone()

        time.sleep(1)

    except Exception as e:
      self._errorString = "Error: %s" % str(e)
      self._dataStore.set_errorString(self._errorString)
      self.containsErrors = True
      self._dataStore.set_alldone()

      error_message = 'Exception of type %s' % type(e).__name__
      LOGGER.exception(error_message)
      raise

    finally:
      self._close()

  def _close(self):
    self.conn.queryTracker().untrackQuery(queryRunner=self)

    if self.response:
      self.response.close()
      self.response = None

  def cancel(self):
    LOGGER.debug('KSQL Query Runner cancel: Canceling ksql query')
    self._close()

class KsqlConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_KSQL):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.connection_type = conn_type

    self._queryTracker = _KSQLQueryTracker()

    host = params.get('HOST', '')
    port = params.get('PORT', '')
    self.sslverify = params.get('SSLCERT_NAME_MISMATCH', False)
    proto = ""
    if not host.startswith("http://") and not host.startswith("https://"):
        proto = "http://"
    self.username = params.get('USERNAME', '_default')
    # self.password = params.get('PASSWORD', '_default')
    self.adminurl = proto + host + ':' + port + '/ksql/'
    self.queryurl = proto + host + ':' + port + '/query/'
    self.return_type = 'json'
    self._properties = KsqlConnProperties()
    self._ksql_query_runner = AsyncRunner(threadcount=_KSQL_MAX_THREADS_PER_CONN,
        name="ksql")

  def queueQuery(self, queryRunner):
    if queryRunner.queued():
      return

    self._ksql_query_runner.queue(queryRunner)
    queryRunner.setQueued()

  def close(self):
    pass

  def queryTracker(self):
    return self._queryTracker

  def properties(self):
    return KsqlConnProperties()

  def execute(self, build_request, async_flag=False):
    LOGGER.debug("query=%s, async=%s", build_request.query, async_flag)

    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      return self._topics()

    if build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      return self._tables_and_streams(build_request)

    if build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      return self._validate(build_request)

    executor = BuildRequestExecutor(build_request)
    return executor.execute(async_flag=async_flag)

  def _ksql_query(self, query, admin=False, sync=False, timeout=None, earliest=False):
    url = self.adminurl if admin else self.queryurl
    stream = (sync == False)

    data = {'ksql': query}
    if earliest:
      data['streamsProperties'] = {'auto.offset.reset': 'earliest'}

    LOGGER.debug('_ksql_query(sync=%s, timeout=%s, stream=%s, query=%s): %s', sync, timeout, stream, query, url)
    if stream:
      response = _HTTP.request(
          'POST',
          url,
          headers={'Content-Type': 'application/json'},
          body=json.dumps(data),
          preload_content=False)
      status = response.status
    else:
      response = _HTTP.request(
          'POST',
          url,
          headers={'Content-Type': 'application/json'},
          body=json.dumps(data))
      status = response.status

    if status != 200:
      print(data, status, response, type(response))
      raise Exception(response.data)

    if stream or not admin:
      return response

    rsp = json.loads(response.data.decode('utf-8'))
    LOGGER.debug('_ksql_response:\n%s', pprint.pformat(rsp, indent=2))
    return rsp

  def _schema(self, table_name):
    try:
      result = self._ksql_query('describe %s;' % table_name, admin=True, sync=True)
      # Schema API changed from KSQL version 0.5, to 4.1 to 5.0. The code below
      # checks for each possibility in that order.
      tmp_description = result[0].get('description')
      if tmp_description:
        columns = tmp_description.get('schema')
      else:
        tmp_description = result[0].get('sourceDescription')
        columns = tmp_description.get('schema')
        if columns == None:
          columns = tmp_description.get('fields')
          return [[column.get('name', ''), column.get('schema', {}).get('type', '')] for column in columns]

    except:
      raise

    return [[column.get('name', ''), column.get('type', '')] for column in columns]

  def _topics(self):
    try:
      result = self._ksql_query('show topics;', admin=True, sync=True)
      # Version handling between different KSQL API's
      tmp_topics = result[0].get('kafka_topics') or result[0]
      topics = tmp_topics.get('topics')
    except:
      raise

    rows = [[topic.get('name', '')] for topic in topics]
    colnames = ['name']
    coltypes = ['STRING']
    data = {
      'colnames': colnames,
      'coltypes': coltypes,
      'rows': rows
    }
    return data

  def _validate(self, build_request):
    for exprType, cmd in VALIDATE_CMDS.items():
      if build_request.query.startswith(cmd):
        expression = build_request.query[len(cmd):].strip()
        try:
          data = self._validate_expression(expression, build_request, exprType)
        except Exception as e:
          # if exception is caught above we set state to unknown
          error_message = 'Error validating expression'
          LOGGER.exception(error_message)
          data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
        return data

    return data

  def _validate_expression(self, expression, build_request, exprType):
    data = {
      'is_valid': "true",
      'isaggregate': "false",
      'ispart': "false"
    }

    return data

  def _tables_and_streams(self, build_request):
    topic = build_request.query.split(' ')[1]
    tables = self._get_tables_or_streams(topic, 'tables')
    streams = self._get_tables_or_streams(topic, 'streams')

    rows = [[stream, 'stream'] for stream in streams] + \
        [[table, 'table'] for table in tables]
    colnames = ['name', 'type']
    coltypes = ['STRING', 'STRING']
    data = {
      'colnames': colnames,
      'coltypes': coltypes,
      'rows': rows
    }
    return data

  def _get_tables_or_streams(self, topic, target='streams'):
    # "Show Tables;" Response format (version < 5)
    # [{'tables':
    #   {'statementText':'show tables;',
    #    'tables': [{'name':'', 'topic':'', 'format':'JSON', 'isWindowed':false}]
    #    }
    # }]

    query = 'show ' + target + ';'
    try:
      result = self._ksql_query(query, admin=True, sync=True)
      # Version 4 had result[0].get(target) be a dict and version 5 has
      # result[0].get(target) be a list
      consumers = result[0].get(target)
      if consumers is None or isinstance(consumers, dict):
        consumers = result[0].get(target, {}).get(target, [])
      else:
        consumers = result[0].get(target, [])
    except:
      raise

    return [[cons.get('name', '')] for cons in consumers if not topic or cons.get('topic', '') == topic]

  def table_columns(self, table_name, username=None, client_ip=None):
    # Don't use the dbapi2 table_columns until "select * from limit 0" is a valid query.
    # For now, just use describe
    table = table_name.split('.')[1]

    schema = self._schema(table)

    data = {
      'colnames': [str(x[0]) for x in schema],
      'coltypes': [KsqlConnProperties.cursor_coltype_to_arctype(str(x[1])) for x in schema],
      'partitions': []
    }

    return data

  def _get_columns(self, raw_rows):
    if len(raw_rows) <= 0:
      return []

    row_obj = json.loads(raw_rows[0])

  def _get_row(self, jsonRow):
    if type(jsonRow) != type({}):
      return None

    r = jsonRow.get('row', {})
    if r:
      cols = r.get('columns', [])
      if cols:
        return cols

    return None

  def _get_rows(self, raw_rows):
    if raw_rows is None or len(raw_rows) <= 0:
      return []

    try:
      out_rows = []
      for row in raw_rows:
        jsonRow = json.loads(row)
        d = self._get_row(jsonRow)
        if d is not None:
          out_rows.append(d)
      return out_rows
    except Exception as e:
      LOGGER.exception(str(e))
      return []

  def build_stream_hash(self, build_request):
    return build_request.query

  def _sync_results(self, query, admin=False, sync=False, timeout=None):
    if admin == True:
       return self._ksql_query(query, admin=admin, sync=sync, timeout=timeout)

    (dataConsumer, runner) = _KSQLQueryRunner.getDataConsumerAndQueryRunner(
        query=query, conn=self, admin=admin, sync=sync, timeout=timeout,
        runAsyncQuery=False)
    runner.run()
    rows = dataConsumer.pop_all()
    return rows

  def async_status(self, qhandle, username=None, profile=None):
    """Get the status of an async query"""
    dataConsumerId = int(qhandle)
    dataConsumer = self.queryTracker().getDataConsumer(dataConsumerId=dataConsumerId)

    if dataConsumer is None:
      raise Exception("status: Failed (or Streaming Queries have been reset)")

    if dataConsumer.errorString():
      raise Exception(dataConsumer.errorString())

    dataConsumer.update_last_time()
    if len(dataConsumer) > 0:
      s = {"status": "Finished"}
    else:
      if dataConsumer.alldone():
        s = {"status": "StreamConsumed"}
      else:
        s = {"status": "Running (10/100 completed)"}

    return s

  def async_results(self, qhandle, username=None, profile=None):
    """Get the results of an async query"""

    from sqlrun.views import qhandleAddPrefix

    try:
      dataConsumerId = int(qhandle)
      dataConsumer = self.queryTracker().getDataConsumer(dataConsumerId=dataConsumerId)

      if dataConsumer is None:
        raise Exception("Status: Failed")

      dataConsumer.update_last_time()
      rows = dataConsumer.pop_all()

      resultData = _ResultData()
      resultRows = rows
      resultData.data['rows'] = resultRows
      if resultRows:
        row = resultData.data['rows'][0]
        resultData.data['colnames'] = dataConsumer.colnames() or ['unset'] * len(row)
        resultData.data['coltypes'] = ['string'] * len(row)
        resultData.data['ongoing_qhandle'] = qhandleAddPrefix(qhandle)

      return resultData.data
    except Exception as e:
      LOGGER.exception(str(e))
      raise

  def resetAllQueries(self, username=None):
    """ Cancel all pending queries, clear up DataConsumers and DataStores """

    self._queryTracker.resetAllDataStores()

  def async_runquery(self, query,
      username=None, sample_pct=None, comment='', logQuery=True,
      match_views=True, qcntl=None, ksqlStreamPosition=None):
    """Run an async query"""

    if not query.endswith(';'):
      query = query + ';'

    kwargs = {
      'query': query,
      'conn': self,
      'sync': False,
      'runAsyncQuery': True,
    }

    if ksqlStreamPosition == 'earliest':
      kwargs['earliest'] = True

    (dataConsumer, runner) = _KSQLQueryRunner.getDataConsumerAndQueryRunner(**kwargs)

    return {
      'qhandle': str(dataConsumer.dataConsumerId()),
    }

  def sync_runquery(self, query_url, build_request=None,
    username=None, sample_pct=None, comment='', logQuery=False, stream=True, qcntl=None,
    profile=None, setCancelCbFn=None):

    earliest = False
    colnames = []
    coltypes = []

    if build_request.query.startswith(SPECIAL_CMDS['ROWS']):
      table_name = build_request.query[1 + len(SPECIAL_CMDS['ROWS']):]
      table_name = table_name.split(".")[-1]
      stream = False
      limit = 10
      schema = self._schema(table_name)
      colnames = [s[0] for s in schema]
      coltypes = [s[1] for s in schema]
      query = "select * from %s limit %s;" % (table_name, limit)
      earliest = True

    else:
      query = build_request.query + ';'

    if build_request and build_request.dataset_request and \
        build_request.dataset_request.dstype == "getrows":
      earliest = True

    if build_request.builder:
      col_objs = list(build_request.builder.dataset_request.dataset.dsuhash.values())
      colnames = [c.attr.get('alias') or c.attr.get('name') for c in col_objs]
      coltypes = [c.attr.get('type') or c.attr.get('type') for c in col_objs]

    if stream:
      timeout = None
      sync = False
    else:
      timeout = 15
      sync = True

    (dataConsumer, runner) = _KSQLQueryRunner.getDataConsumerAndQueryRunner(
        query=query, conn=self, sync=sync, timeout=timeout, runAsyncQuery=False,
        earliest=earliest)

    rows = None

    if sync:
      try:
        if not runner.admin:
          self.queueQuery(runner)
        else:
          runner.run()

        if runner.errorString():
          raise Exception(runner.errorString())

        if runner.resultData:
          return runner.resultData.data

        # This is a sync query that is also a streaming query.
        # Wait for a few seconds
        for i in range(0, timeout):
          time.sleep(1)
          if runner.errorString():
            raise Exception('Error running query: %s' % runner.errorString())

          if len(dataConsumer) > 0:
            rows = dataConsumer.pop_all()
            break
      finally:
        runner.cancel()
    else:
      self.queueQuery(runner)
      rows = dataConsumer.pop_all()

    resultData = _ResultData()
    resultData.data['rows'] = rows
    resultData.data['colnames'] = colnames
    resultData.data['coltypes'] = coltypes

    return resultData.data

def colValue(row, colname):
  v = row.get(colname, '')
  if type(v) == type([]):
    return '\n'.join(v)

  if type(v) != type(''):
    return str(v)

  return v

def listToResultData(l):
  columns = []
  columnsH = {}

  for row in l:
    for k, v in row.items():
      if k in columnsH:
        continue

      columnsH[k] = True
      columns.append(k)

  resultData = _ResultData()
  resultData.data['colnames'] = columns
  resultData.data['coltypes'] = ['string'] * len(columns)

  for row in l:
    resultData.data['rows'].append([colValue(row, colname) for colname in columns])

  return resultData

class _ResultData(object):
  def __init__(self):
    self.data = dict()
    self.data['colnames'] = list()
    self.data['coltypes'] = list()
    self.data['rows'] = list()
    self.data['info'] = ['Query:']

  def setError(self, errorString):
    self.data['error'] = True
    if errorString is not None:
      self.data['errorString'] = errorString

