# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging
from . import dataconn_dbapi2
from .connections_types import DATACONN_MONGODB
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from . import arctypes

# create logger
LOGGER = logging.getLogger("arcweb")

MONGO_HIDE_COLLECTIONS = ["system.indexes"]


missing_modules = []
try:
  import pyodbc as dbapi2
except Exception as e:
  missing_modules.append('pyodbc')

try:
  from pymongo import MongoClient
  from pymongo.database import Database
except Exception as e:
  missing_modules.append('pymongo')

if missing_modules:
  # LOGGER.info('Libraries missing for MongoDB {}'.format(', '.join(missing_modules)))
  pass

class MongoConnProperties(ConnProperties):
  TYPEMAP = {
    "<type 'float'>" : arctypes.ARC_DATATYPE_REAL,
    "<type 'int'>"  : arctypes.ARC_DATATYPE_INT,
    "<type 'long'>" : arctypes.ARC_DATATYPE_BIGINT,
    "<type 'bool'>" : arctypes.ARC_DATATYPE_BOOLEAN,
    "<type 'str'>" : arctypes.ARC_DATATYPE_STRING,
    "<type 'datetime.datetime'>" : arctypes.ARC_DATATYPE_TIMESTAMP
  }
  def __init__(self):
    super(MongoConnProperties, self).__init__()
    self.quote_alias = '"'
    self.quote_string_constant = "'"
    self.supports_async = False
    self.supports_reuse = False
    self.supports_partitions = False
    self.alias_supports_backticks = False
    self.supports_limit_outside_union = True
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = None

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    return MongoConnProperties.TYPEMAP.get(str(native_coltype),
        arctypes.ARC_DATATYPE_STRING)

class MongoConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_MONGODB):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    conn_params = {}

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    conn_params['autocommit'] = True

    connection_string = []
    if params.get('DRIVER'):
      # driver must be enclosed in curly brackets
      driver_prefix = ''
      driver_postfix = ''
      self.driver = params.get('DRIVER', "").strip()
      if len(self.driver):
        if self.driver[0] != '{':
          driver_prefix = '{'
        if self.driver[-1] != '}':
          driver_postfix = '}'
      self.driver = driver_prefix + self.driver + driver_postfix
      connection_string.append('DRIVER='+ self.driver)
    if params.get('HOST'):
      self.host = params.get('HOST', "").strip()
      connection_string.append('HOST='+self.host)
    if params.get('PORT'):
      self.port = params.get('PORT', "").strip()
      connection_string.append('PORT='+self.port)
    if params.get('USERNAME'):
      connection_string.append('UID='+params.get('USERNAME', "").strip())
    if params.get('PASSWORD'):
      connection_string.append('PWD='+params.get('PASSWORD', "").strip())
    if params.get('DATABASE'):
      self.dbname = params.get('DATABASE').strip()
      connection_string.append('DATABASE='+ params.get('DATABASE', "").strip())

    if not for_testing:
      client = None
      try:
        self.connection = dbapi2.connect(';'.join(connection_string), **conn_params)
        client = MongoClient(host=self.host, port=int(self.port))
        self.db_names = [x.encode('UTF8') for x in client.database_names()]
        self.collections = dict()
        for db_name in self.db_names:
          db = Database(client=client, name=db_name)
          self.collections[db_name] = [x for x in db.collection_names() if x not in MONGO_HIDE_COLLECTIONS]
      except Exception as e:
        LOGGER.error(e)
        raise Exception('Could not connect to database')
      finally:
        if client is not None:
          client.close()

    self.connection_type = conn_type

  def properties(self):
    return MongoConnProperties()

  def execute(self, build_request,
              async_flag=False):

    query = None
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      # if doesn't make sense to list databases other than the currently connected one
      # data = {
      #     'colnames': ['name'],
      #     'coltypes': ['STRING'],
      #     'rows': [[x] for x in self.db_names]
      #   }
      data = {
          'colnames': ['name'],
          'coltypes': ['STRING'],
          'rows': [[self.dbname]]
        }
      return data

    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      data = {
          'colnames': ['name'],
          'coltypes': ['STRING'],
          'rows': [[x] for x in self.collections[build_request.query.split()[-1]]]
        }
      return data

    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)

    return executor.execute(query=query,
                            async_flag=async_flag)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "Aggregate functions and subqueries are not allowed in the GROUP BY clause" not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr
