# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json, logging, re
from . import dataconn_dbapi2
from .connections_types import DATACONN_MSSQL
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from . import arctypes

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  import pymssql as dbapi2
except Exception as e:
  # LOGGER.info('Libraries missing for MS SQL Server (see instructions in code)')
  pass

# (as of 11/2017, release 4.2)
# try $ pip install pymssql.  If it fails:
# $ brew unlink freetds; brew install homebrew/versions/freetds091
# $ brew link --force freetds@0.91
# $ pip install mssql

class MsSqlConnProperties(ConnProperties):
  TYPEMAP = {
    1: arctypes.ARC_DATATYPE_STRING, # (n)char, (n)varchar, (n)text,
                                     # date, datetimeoffset, datetime2
    2: arctypes.ARC_DATATYPE_BINARY, # binary, varbinary, image
    3: arctypes.ARC_DATATYPE_BIGINT, # bigint, bit, smallint, int, tinyint
    4: arctypes.ARC_DATATYPE_TIMESTAMP, # smalldatetime, datetime
    5: arctypes.ARC_DATATYPE_REAL, # numeric, decimal, smallmoney, money
  }

  def __init__(self):
    super(MsSqlConnProperties, self).__init__()
    self.quote = '\001'
    self.quote_end = '\002'
    self.quote_string_constant = "'"
    self.quote_alias = "'"
    self.supports_async = False
    self.supports_reuse = False
    self.supports_partitions = False
    self.alias_supports_backticks = False
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = None
    self.string_type = "VARCHAR(MAX)"
    self.supports_limit_outside_union = False
    self.supports_limit = False
    self.supports_top = True
    self.supports_groupby_colindex = False
    self.supports_groupby_literal = False

    self.import_data_insertrows = 10000
    self.display_to_table_type = {
      "String": "varchar(max)",
      "Integer": "bigint",
      "Real": "float",
      "Boolean": "bit",
      "Timestamp": "datetime"
    }

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    return MsSqlConnProperties.TYPEMAP.get(native_coltype,
        arctypes.ARC_DATATYPE_STRING)

class MsSqlConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_MSSQL):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    conn_params = {}

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    if params.get('HOST'):
      conn_params['host'] = params.get('HOST').strip()
    if params.get('PORT'):
      conn_params['port'] = params.get('PORT')
    if params.get('USERNAME'):
      conn_params['user'] = params.get('USERNAME')
    if params.get('PASSWORD'):
      conn_params['password'] = params.get('PASSWORD')
    if params.get('DATABASE'):
      self.dbname = params.get('DATABASE').strip()
      conn_params['database'] = params.get('DATABASE')

    if not for_testing:
      try:
        self.connection = dbapi2.connect(**conn_params)
      except Exception as e:
        LOGGER.error(e)
        raise Exception('Could not connect to database')

    self.connection_type = conn_type

  def properties(self):
    return MsSqlConnProperties()

  def swap_quotes(self, query):
    return query.replace('\001', '[').replace('\002', ']')

  def execute(self, build_request,
              async_flag=False):

    query = None
    limit = None

    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = "select name from master.dbo.sysdatabases where name not in ('master', 'tempdb', 'model', 'msdb')"
    elif build_request.query.startswith(SPECIAL_CMDS['ROWS']):
      table_name = build_request.query[1 + len(SPECIAL_CMDS['ROWS']):]
      limit = 10
      query = "select top(%s) * from %s" % (limit, self.quote_tablename(table_name))
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      # Note that the escaping of the db name is important below
      # Try using a database called default without it
      query = "select table_schema + '.' + table_name from %s.INFORMATION_SCHEMA.TABLES" % \
              self.quote_tablename(build_request.query.split()[-1])
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)

    if query:
      query = self.swap_quotes(query)
      LOGGER.info('query %s', query)
    else:
      build_request.query = self.swap_quotes(build_request.query)
      LOGGER.info('query %s', build_request.query)

    r = executor.execute(query=query,
                         async_flag=async_flag)
    return r

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "Cannot use an aggregate" not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr
