# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging, re
from . import dataconn_dbapi2
from .connections_types import DATACONN_MYSQL
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from . import arctypes

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  import pymysql as dbapi2
except Exception as e:
  LOGGER.info('Libraries missing for MySQL (pymysql)')

class MysqlConnProperties(ConnProperties):
  TYPEMAP = {
    0: arctypes.ARC_DATATYPE_REAL, # 'DECIMAL',
    1: arctypes.ARC_DATATYPE_INT, # 'TINY',
    2: arctypes.ARC_DATATYPE_INT, # 'SHORT',
    3: arctypes.ARC_DATATYPE_BIGINT, # 'LONG',
    4: arctypes.ARC_DATATYPE_REAL, # 'FLOAT',
    5: arctypes.ARC_DATATYPE_REAL, # 'DOUBLE',
    6: arctypes.ARC_DATATYPE_UNKNOWN, # 'NULL',
    7: arctypes.ARC_DATATYPE_TIMESTAMP, # 'TIMESTAMP',
    8: arctypes.ARC_DATATYPE_BIGINT, # 'LONGLONG',
    9: arctypes.ARC_DATATYPE_INT, # 'INT24',
    10: arctypes.ARC_DATATYPE_TIMESTAMP, # 'DATE',
    11: arctypes.ARC_DATATYPE_STRING, # 'TIME',
    12: arctypes.ARC_DATATYPE_TIMESTAMP, # 'DATETIME',
    13: arctypes.ARC_DATATYPE_INT, # 'YEAR',
    14: arctypes.ARC_DATATYPE_UNKNOWN, # 'NEWDATE',
    15: arctypes.ARC_DATATYPE_STRING, # 'VARCHAR',
    16: arctypes.ARC_DATATYPE_INT, # 'BIT',
    246: arctypes.ARC_DATATYPE_REAL, # 'NEWDECIMAL',
    247: arctypes.ARC_DATATYPE_UNKNOWN, # 'INTERVAL',
    248: arctypes.ARC_DATATYPE_UNKNOWN, # 'SET',
    249: arctypes.ARC_DATATYPE_STRING, # 'TINY_BLOB',
    250: arctypes.ARC_DATATYPE_STRING, # 'MEDIUM_BLOB',
    251: arctypes.ARC_DATATYPE_STRING, # 'LONG_BLOB',
    252: arctypes.ARC_DATATYPE_STRING, # 'BLOB',
    253: arctypes.ARC_DATATYPE_STRING, # 'VAR_STRING',
    254: arctypes.ARC_DATATYPE_STRING, # 'STRING',
    255: arctypes.ARC_DATATYPE_UNKNOWN # 'GEOMETRY' }
  }

  def __init__(self):
    super(MysqlConnProperties, self).__init__()
    self.quote_alias = "'"
    self.supports_async = False
    self.supports_reuse = False
    self.supports_partitions = False
    self.alias_supports_backticks = False
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = None
    self.string_type = "CHAR"
    self.timestamp_type = "DATETIME"
    self.integer_type = "SIGNED INTEGER"
    self.supports_dataextract_source = True
    self.supports_dataextract_target = True
    self.paramstyle = '%s'

    self.display_to_table_type = {
      'String': 'text',
      'Integer': 'bigint',
      'Real': 'real',
      'Boolean': 'tinyint',
      'Timestamp': 'datetime'
    }

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    return MysqlConnProperties.TYPEMAP.get(native_coltype,
        arctypes.ARC_DATATYPE_STRING)

class MysqlConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_MYSQL):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    conn_params = {}

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    if params.get('HOST'):
      conn_params['host'] = params.get('HOST').strip()
    if params.get('PORT'):
      conn_params['port'] = int(params.get('PORT'))
    if params.get('USERNAME'):
      conn_params['user'] = params.get('USERNAME')
    if params.get('PASSWORD'):
      conn_params['password'] = params.get('PASSWORD')
    if params.get('DATABASE'):
      self.dbname = params.get('DATABASE').strip()
      conn_params['db'] = params.get('DATABASE')
    conn_params['charset'] = 'utf8'
    if not for_testing:
      try:
        self.connection = dbapi2.connect(**conn_params)
      except Exception as e:
        LOGGER.error(e)
        raise Exception('Could not connect to database')

    self.connection_type = conn_type

  def properties(self):
    return MysqlConnProperties()

  def execute(self, build_request,
              async_flag=False):

    query = None
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = "show databases"
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      # Note that the escaping of the db name is important below
      # Try using a database called default without it
      query = "show tables in `%s`" % build_request.query.split()[-1]
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            data = {'is_valid': 'unknown', 'error_msg': 'Error validating expression: ' + str(e)}
          return data

    executor = BuildRequestExecutor(build_request)

    return executor.execute(query=query,
                            async_flag=async_flag)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "Can't group on" not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr

  def apply_types(self, value, type):
    try:
      if value is None:
        value = "NULL"
      elif type == "String":
        value = re.sub(r"\\", r"\\\\", value)
        value = re.sub("'", r"\'", value)
        value = "'%s'" % value
      elif type == "Timestamp":
        value = "'%s'" % value
      elif type == "Real":
        value = re.sub(",", ".", value)
      elif type == "Boolean":
        if value:
          value = 1
        else:
          value = 0
    except Exception as e:
      pass

    return value