# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging
from . import dataconn_dbapi2
from .connections_types import DATACONN_ODBC, DATACONN_TERADATA
from .dataconn_dataengine import ConnProperties
from . import arctypes
from .dataconn_dataengine import BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from django.conf import settings

# create logger
LOGGER = logging.getLogger("arcweb")

if settings.ENABLE_DATACONN_ODBC:
  # To install pyodbc:
  # $ brew install unixodbc
  # $ pip install pyodbc==3.0.10
  try:
    import pyodbc as dbapi2
  except Exception as e:
    LOGGER.info('Libraries missing for ODBC (pyodbc)')

class OdbcConnProperties(ConnProperties):
  TYPEMAP = {
    "<class 'float'>" : arctypes.ARC_DATATYPE_REAL,
    "<class 'decimal.Decimal'>": arctypes.ARC_DATATYPE_REAL,
    "<class 'int'>"  : arctypes.ARC_DATATYPE_INT,
    "<class 'long'>" : arctypes.ARC_DATATYPE_BIGINT,
    "<class 'bool'>" : arctypes.ARC_DATATYPE_BOOLEAN,
    "<class 'str'>" : arctypes.ARC_DATATYPE_STRING,
    "<class 'datetime.datetime'>" : arctypes.ARC_DATATYPE_TIMESTAMP
  }

  def __init__(self):
    super(OdbcConnProperties, self).__init__()
    self.quote_alias = '"'
    self.quote_string_constant = "'"
    self.supports_async = False
    self.supports_reuse = False
    self.supports_partitions = False
    self.alias_supports_backticks = False
    self.supports_limit_outside_union = True
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = None

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    return OdbcConnProperties.TYPEMAP.get(str(native_coltype),
        arctypes.ARC_DATATYPE_STRING)


class OdbcConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_ODBC):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    conn_params = {}

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    conn_params['autocommit'] = True

    connection_string = []
    if params.get('DSN'):
      self.dsn = params.get('DSN', "").strip()
      connection_string.append('DSN='+ self.dsn)
    if params.get('DRIVER'):
      self.driver = params.get('DRIVER', "").strip()
      connection_string.append('DSN=' + self.driver)
    if params.get('HOST'):
      self.host = params.get('HOST', "").strip()
      connection_string.append('HOST=' + self.host)
    if params.get('PORT'):
      self.port = str(params.get('PORT', "")).strip()
      connection_string.append('PORT='+self.port)
    if params.get('USERNAME'):
      connection_string.append('UID='+params.get('USERNAME', "").strip())
    if params.get('PASSWORD'):
      connection_string.append('PWD='+params.get('PASSWORD', "").strip())
    if params.get('DATABASE'):
      self.dbname = params.get('DATABASE').strip()
      connection_string.append('DATABASE='+ params.get('DATABASE', "").strip())
    if params.get('Authentication'):
      connection_string.append('AuthenticationType='+ params.get('Authentication').strip())

    if conn_type != DATACONN_TERADATA and not for_testing:
      try:
        self.connection = dbapi2.connect(';'.join(connection_string), **conn_params)
      except Exception as e:
        LOGGER.error(e)
        raise

    self.connection_type = conn_type

  def properties(self):
    return OdbcConnProperties()

  def execute(self, build_request, async_flag=False):

    query = None

    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = 'show databases'
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      query = 'show tables in ' + self.quote_tablename(build_request.query.split()[-1])
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          return self.validate_expression(expression, build_request, exprType)


    executor = BuildRequestExecutor(build_request)
    LOGGER.debug('query: %s' % build_request.query or query)
    return executor.execute(query=query, async_flag=async_flag)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query,
                               username=username)
    except Exception as e:
      err_msg = e
      if hasattr(e, 'args') and len(e.args) == 2:
        err_msg = e.args[1]

      err_msg = str(err_msg)
      if "expected something like a name or a Unicode delimited" in err_msg:
        # Teradata
        hasaggr = "false"
      elif "may not contain aggregate functions" in err_msg:
        hasaggr = "true"
      elif "not allowed in GROUP BY" in err_msg:
        hasaggr = "true"
      elif "is not being grouped" in err_msg: # Drill TODO need to use a different query because agg expres don't fail
        hasaggr = "false"
      else:
        error_message = 'Error validating for an aggregate expression'
        LOGGER.exception(error_message)
        raise Exception(error_message)

      return hasaggr
