# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
from .connections_types import DATACONN_ORACLE
from django.conf import settings
from . import dataconn_dbapi2
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from . import arctypes
import logging

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  import cx_Oracle as dbapi2
except Exception as e:
  # LOGGER.info('Libraries missing for Oracle DB (cx_Oracle)')
  pass

class OracleConnProperties(ConnProperties):
  TYPEMAP = {
    "<class 'cx_Oracle.NUMBER'>": arctypes.ARC_DATATYPE_REAL,
    "<class 'cx_Oracle.FIXED_CHAR'>": arctypes.ARC_DATATYPE_STRING,
    "<class 'cx_Oracle.STRING'>": arctypes.ARC_DATATYPE_STRING,
    "<class 'cx_Oracle.TIMESTAMP'>": arctypes.ARC_DATATYPE_TIMESTAMP,
    "<class 'cx_Oracle.DATETIME'>": arctypes.ARC_DATATYPE_TIMESTAMP,
    "<class 'cx_Oracle.LOB'>": arctypes.ARC_DATATYPE_STRING
  }
  def __init__(self):
    super(OracleConnProperties, self).__init__()
    self.quote = '"'
    self.quote_end = '"'
    self.quote_alias = '"'
    self.quote_escape_character = "'"
    self.quote_string_constant = "'"
    self.supports_partitions = False
    self.supports_async = False
    self.supports_backticks = False
    self.supports_reuse = False
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = 'public'
    self.alias_supports_backticks = False
    self.string_type = "varchar"
    self.date_type = "date"
    self.double_type = "double precision"
    self.supports_nulls_order = True
    self.supports_order_outside_union = False
    self.supports_limit_outside_union = False

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    return OracleConnProperties.TYPEMAP.get(str(native_coltype), arctypes.ARC_DATATYPE_STRING)

class OracleConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_ORACLE):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    conn_params = {}

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    if params.get('HOST'):
      conn_params['host'] = str(params.get('HOST').strip())
    if params.get('PORT'):
      conn_params['port'] = str(params.get('PORT'))
    if params.get('USERNAME'):
      conn_params['user'] = str(params.get('USERNAME'))
    if params.get('PASSWORD'):
      conn_params['password'] = str(params.get('PASSWORD'))
    if params.get('DATABASE'):
      self.dbname = str(params.get('DATABASE'))
      conn_params['database'] = str(params.get('DATABASE').strip())
    if not self.dbname:
      raise Exception('Database name not specified')

    if not for_testing:
      dsn = "%s/%s@%s:%s/%s" % (conn_params['user'],
                                conn_params['password'],
                                conn_params['host'],
                                conn_params['port'],
                                conn_params['database'])
      self.connection = dbapi2.connect(str(dsn))

      self.cursor = self.connection.cursor()
      self.cursor.execute("ALTER SESSION SET NLS_TERRITORY = 'AMERICA'")
      self.cursor.execute("ALTER SESSION SET NLS_DATE_FORMAT = 'YYYY-MM-DD HH24:MI:SS'"
                     " NLS_TIMESTAMP_FORMAT = 'YYYY-MM-DD HH24:MI:SS.FF'"
                      + (" TIME_ZONE = 'UTC'" if settings.USE_TZ else ''))

    self.connection_type = conn_type

  def properties(self):
    return OracleConnProperties()

  def execute(self, build_request, async_flag=False):

    query = None
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = "SELECT owner from all_tables UNION SELECT owner from all_views" # there is no DB in ORACLE, this is users
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      schema = build_request.query.split()[-1]
      query = "SELECT table_name FROM all_tables WHERE owner = '%s'" % schema + " UNION " + \
              "SELECT view_name FROM all_views WHERE owner = '%s'" % schema
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          return self.validate_expression(expression, build_request, exprType)

    executor = BuildRequestExecutor(build_request)

    return executor.execute(query=query,
                            async_flag=async_flag)

  def table_columns(self, fq_table_name, username=None, client_ip=None):
    query = 'select * from %s where rownum <= 0' % self.quote_tablename(fq_table_name)

    d = self.log_sync_runquery(query, username=username,
                                   client_ip=client_ip)
    d['partitions'] = []
    return d

  def expression_isAggregate(self, expression, **kwargs):
    # see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = 'false'
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "not a GROUP BY expression" not in str(e):
        raise
      else:
        hasaggr = 'true'

    return hasaggr
