# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.


from . import dataconn_dbapi2
from .connections_types import DATACONN_PG
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from django.conf import settings
from .dataconn_dataengine import DEFAULT_UNQUOTE

from . import arctypes
import logging
import re
import threading

from django.conf import settings

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  import psycopg2
  import psycopg2.pool
except Exception:
  LOGGER.info("Libraries missing for Postgres (psycopg2)")

POOL_MAP = {}
POOL_LOCK = threading.Lock()

MAXCONN = getattr(settings, 'PGPOOL_MAXCONN', 20)
WAITCONN = getattr(settings, 'PGPOOL_WAITCONN', 60)

class PgPool:
  def __init__(self, params, minconn=0, maxconn=MAXCONN):
    self.params = params
    self.minconn = minconn
    self.maxconn = maxconn
    self.lock = threading.Lock()
    self.cv = threading.Condition(self.lock)
    self.pool = psycopg2.pool.ThreadedConnectionPool(minconn, maxconn, **params)

  def putconn(self, conn, close=False):
    with self.lock:
      self.pool.putconn(conn, close=close)
      self.cv.notify()

  def getconn(self):
    with self.lock:
      while True:
        try:
          return self.pool.getconn()
        except Exception as e:
          if str(e).find('connection pool exhausted') != -1:
            if not self.cv.wait(timeout=WAITCONN):
              raise Exception('Timeout waiting for available connection')
          else:
            raise

class PgConnProperties(ConnProperties):
  # Refer to mappings here: http://doxygen.postgresql.org/include_2catalog_2pg__type_8h.html
  TYPEMAP = {
    1043:   arctypes.ARC_DATATYPE_STRING,
    20:     arctypes.ARC_DATATYPE_BIGINT,
    21:     arctypes.ARC_DATATYPE_BIGINT,
    23:     arctypes.ARC_DATATYPE_BIGINT,
    700:    arctypes.ARC_DATATYPE_DOUBLE,
    701:    arctypes.ARC_DATATYPE_DOUBLE,
    1114:   arctypes.ARC_DATATYPE_TIMESTAMP,
    16:     arctypes.ARC_DATATYPE_BOOLEAN,
  }

  def __init__(self):
    super(PgConnProperties, self).__init__()
    self.quote = '"'
    self.quote_end = '"'
    self.quote_alias = '"'
    self.quote_escape_character = '"'
    self.quote_string_constant = "'"
    self.supports_partitions = False
    self.supports_async = False
    self.supports_backticks = False
    self.supports_reuse = False
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = 'public'
    self.alias_supports_backticks = False
    self.string_type = "TEXT"
    self.date_type = "date"
    self.double_type = "double precision"
    self.supports_nulls_order = True
    self.supports_dataextract_source = True
    self.supports_dataextract_target = True
    self.paramstyle = '%s'

    self.display_to_table_type = {
      'String': 'text',
      'Integer': 'bigint',
      'Real': 'real',
      'Boolean': 'boolean',
      'Timestamp': 'timestamp'
    }

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    return PgConnProperties.TYPEMAP.get(native_coltype,
        arctypes.ARC_DATATYPE_STRING)

class PgConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_PG):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    self.pool = None
    conn_params = {}

    self.for_testing = False
    if params.get('FOR_TESTING'):
      self.for_testing = True

    if params.get('HOST'):
      conn_params['host'] = str(params.get('HOST').strip())
    if params.get('PORT'):
      conn_params['port'] = str(params.get('PORT'))
    if params.get('USERNAME'):
      conn_params['user'] = str(params.get('USERNAME'))
    if params.get('PASSWORD'):
      conn_params['password'] = str(params.get('PASSWORD'))
    if params.get('DATABASE'):
      self.dbname = str(params.get('DATABASE'))
      conn_params['database'] = str(params.get('DATABASE').strip())
    if params.get('SOCK', '').upper() in ['SSL', 'SSLCERT']:
      conn_params['sslmode'] = 'require'
    if not self.dbname:
      raise Exception('Database name not specified')

    if not self.for_testing:
      # 'seq' here allows for forcing new pool on sequence change (like cache flush)
      # not clear this is needed
      pairs = [('id', connection_id), ('seq', 1)] + list(conn_params.items())
      pairs.sort()
      key = ','.join('%s:%s' % (key, val) for (key, val) in pairs)
      if key not in POOL_MAP:
        with POOL_LOCK:
          if key not in POOL_MAP:
            POOL_MAP[key] = PgPool(conn_params)
      self.pool = POOL_MAP[key]
    self.connection_type = conn_type

  def cursor_get(self, username=None):
    if self.for_testing:
      return self.cursor
    if not self.connection:
      self.connection = self.pool.getconn()
    return self.connection.cursor()

  def cursor_close(self, cursor):
    cursor.close()

  def close(self):
    if self.connection and self.pool:
      self.pool.putconn(self.connection)
      self.connection = None

  def properties(self):
    return PgConnProperties()

  def _get_unquote_regex(self):
    if self.compiled_unquote_regex is not None:
      return self.compiled_unquote_regex
    connections_quotes_regex = getattr(settings, 'VALIDATION_QUOTES_REGEX', {})
    if self.connection_type in connections_quotes_regex:
      self.compiled_unquote_regex = connections_quotes_regex[self.connection_type]
    else:
      # adding regex for removing connection specific white list statement
      self.compiled_unquote_regex = DEFAULT_UNQUOTE + \
            [re.compile(r'/\*([^\*])*\*/'),        # sql comments: /* my comment */
             re.compile(r'extract\s*\([^\)]*\)')]  # extract statement: extract(month from [time]::timestamp)
    return self.compiled_unquote_regex

  def execute(self, build_request, async_flag=False):

    query = None
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = "SELECT distinct table_schema FROM information_schema.tables " + \
          "WHERE table_schema NOT IN ('pg_catalog', 'information_schema')"
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      if self.connection_type == DATACONN_PG:
        query = "SELECT table_name FROM information_schema.tables " + \
          "WHERE table_schema IN ('%s') " % build_request.query.split()[-1] + \
          "UNION ALL " + \
          "SELECT matviewname FROM pg_matviews;"
      else:
        query = "SELECT table_name FROM information_schema.tables " + \
          "WHERE table_schema IN ('%s')" % build_request.query.split()[-1]
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)

    return executor.execute(query=query,
                            async_flag=async_flag)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "not allowed in GROUP BY" not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr

  def apply_types(self, value, type):
    try:
      if value is None:
        value = "NULL"
      elif type == "String":
        value = re.sub("'", r"''", value)
        value = "'%s'" % value
      elif type == "Timestamp":
        value = "'%s'" % value
      elif type == "Real":
        value = re.sub(",", ".", value)
    except Exception as e:
      pass

    return value