# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from . import dataconn_dbapi2
from .connections_types import DATACONN_SNOWFLAKE
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from django.conf import settings
from .dataconn_dataengine import DEFAULT_UNQUOTE

from . import arctypes
import logging
import re
import threading

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  # Requires the snowflake-connector-python python package
  # to be available
  import snowflake.connector as dbapi2
except Exception:
  # LOGGER.info("Libraries missing for snowflake")
  pass

class SnowflakeConnProperties(ConnProperties):
  # Refer to mappings here:
  # https://docs.snowflake.net/manuals/user-guide/python-connector-api.html#id4
  TYPEMAP = {
    0: arctypes.ARC_DATATYPE_BIGINT,
    1: arctypes.ARC_DATATYPE_REAL,
    2: arctypes.ARC_DATATYPE_STRING,
    3: arctypes.ARC_DATATYPE_TIMESTAMP,
    4: arctypes.ARC_DATATYPE_TIMESTAMP,
    5: arctypes.ARC_DATATYPE_STRING,
    6: arctypes.ARC_DATATYPE_TIMESTAMP,
    7: arctypes.ARC_DATATYPE_TIMESTAMP,
    8: arctypes.ARC_DATATYPE_TIMESTAMP,
    9: arctypes.ARC_DATATYPE_STRING,  # OBJECT (map?)
    10: arctypes.ARC_DATATYPE_STRING, # ARRAY
    11: arctypes.ARC_DATATYPE_BINARY,
    12: arctypes.ARC_DATATYPE_TIMESTAMP,
    13:	arctypes.ARC_DATATYPE_BOOLEAN,
  }

  def __init__(self):
    super(SnowflakeConnProperties, self).__init__()
    self.quote = '"'
    self.quote_end = '"'
    self.quote_alias = '"'
    self.quote_escape_character = "\\"
    self.quote_string_constant = "'"
    self.supports_partitions = False
    self.supports_async = False
    self.supports_backticks = False
    self.supports_reuse = True
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = 'public'
    self.alias_supports_backticks = False
    self.string_type = "TEXT"
    self.date_type = "date"
    self.double_type = "double precision"
    self.supports_nulls_order = True
    self.supports_dataextract_source = True
    self.supports_dataextract_target = True
    self.paramstyle = '%s'

    self.display_to_table_type = {
      "String": "string",
      "Integer": "bigint",
      "Real": "double",
      "Boolean": "boolean",
      "Timestamp": "timestamp"
    }

  # Snowflake does not supports ( in its aliases
  @staticmethod
  def alias_transform(alias):
    if not alias:
      return alias

    alias = re.sub(r"[^a-zA-Z0-9_\- ]", "_", alias)
    return alias

  @staticmethod
  def cursor_coldesc_to_arctype(native_coldesc):
    native_coltype = native_coldesc[1]
    r = SnowflakeConnProperties.TYPEMAP.get(native_coltype)
    if r is arctypes.ARC_DATATYPE_TIMESTAMP:
      return r

    if native_coldesc[4] or native_coldesc[5]:
      return arctypes.ARC_DATATYPE_REAL

    if r:
      return r

    # Sometimes, we get called internally when the typemap has already been
    # computed
    if native_coltype in arctypes.ARC_DATATYPES_ALL:
      return native_coltype

    # Fallback type is string
    return arctypes.ARC_DATATYPE_STRING

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    r = SnowflakeConnProperties.TYPEMAP.get(native_coltype)
    if r:
      return r;

    # Sometimes, we get called internally when the typemap has already been
    # computed
    if native_coltype in arctypes.ARC_DATATYPES_ALL:
      return native_coltype

    # Fallback type is string
    return arctypes.ARC_DATATYPE_STRING

class SnowflakeConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_SNOWFLAKE):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    self.connection_type = conn_type
    self.params = params
    self._connectionLock = threading.Lock()
    self._connectionIncn = 1

    if params.get('FOR_TESTING'):
      return

    conn_params = {'client_session_keep_alive':True}
    if params.get('ACCOUNT'):
      conn_params['account'] = str(params.get('ACCOUNT').strip())
    else:
      raise Exception('Account name not specified')

    if params.get('USERNAME'):
      conn_params['user'] = str(params.get('USERNAME'))
    else:
      raise Exception('User name not specified')

    if params.get('PASSWORD'):
      conn_params['password'] = str(params.get('PASSWORD'))
    else:
      raise Exception('Password not specified')

    if params.get('ROLE'):
      conn_params['role'] = str(params.get('ROLE'))

    if params.get('WAREHOUSE'):
      conn_params['warehouse'] = str(params.get('WAREHOUSE'))

    if params.get('DATABASE'):
      conn_params['database'] = str(params.get('DATABASE'))
    else:
      raise Exception('Database not specified')

    self.conn_params = conn_params
    self.makeConnection(self._connectionIncn)

  def makeConnection(self, incn):
    with self._connectionLock:
      if incn != self._connectionIncn:
        return

      LOGGER.debug('Authenticating')
      self._connectionIncn += 1
      self.connection = dbapi2.connect(**self.conn_params)
      self._setSessionParameters(self.params.get('SETTINGS', {}))

  def _setSessionParameters(self, settings):
    self.log_sync_runquery("ALTER SESSION SET CLIENT_SESSION_KEEP_ALIVE=TRUE")
    for s in settings:
      self.log_sync_runquery("ALTER SESSION SET %s=%s" % (s, settings[s]))

  def properties(self):
    return SnowflakeConnProperties()

  def close(self):
    # We support reuse of connections. Don't close connection even if asked to
    pass

  def _get_unquote_regex(self):
    if self.compiled_unquote_regex is not None:
      return self.compiled_unquote_regex
    connections_quotes_regex = getattr(settings, 'VALIDATION_QUOTES_REGEX', {})
    if self.connection_type in connections_quotes_regex:
      self.compiled_unquote_regex = connections_quotes_regex[self.connection_type]
    else:
      # adding regex for removing connection specific white list statement
      self.compiled_unquote_regex = DEFAULT_UNQUOTE + \
            [re.compile(r'/\*([^\*])*\*/'),        # sql comments: /* my comment */
             re.compile(r'extract\s*\([^\)]*\)')]  # extract statement: extract(month from [time]::timestamp)
    return self.compiled_unquote_regex

  def execute(self, build_request, async_flag=False):

    query = None
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = "SELECT distinct schema_name FROM information_schema.schemata"
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      query = "SELECT table_name FROM information_schema.tables " + \
          "WHERE table_schema IN ('%s')" % build_request.query.split()[-1]
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)

    incn = self._connectionIncn

    try:
      return executor.execute(query=query, async_flag=async_flag)
    except Exception as e:
      error_message = str(e)
      LOGGER.exception(error_message)
      if 'Authentication token has expired' in error_message:
        pass
      else:
        raise

    # Authentication token has expired - make a new connection and try again
    LOGGER.info("try making new connection")
    self.makeConnection(incn)
    return executor.execute(query=query, async_flag=async_flag)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      error_message = str(e)
      LOGGER.exception(error_message)
      if 'is not a valid group by' not in error_message:
        raise
      else:
        hasaggr = "true"

    return hasaggr

  def apply_types(self, value, type):
    try:
      if value is None:
        value = "NULL"
      elif type == "String":
        value = re.sub("'", r"\'", value)
        value = "'%s'" % value
      elif type == "Timestamp":
        value = "'%s'" % value
      elif type == "Boolean":
        if value:
          value = 1
        else:
          value = 0
    except Exception as e:
      pass

    return value
