# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging, json, requests, re
import time
import datetime
from . import dataconn_nosqlapi
from . import arctypes
from .connections_types import DATACONN_SOLR
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS
from datetime import datetime
from django.conf import settings
from requests.auth import HTTPBasicAuth
from vizlib.querylog import actionlog_create
from urllib.parse import urlparse
from posixpath import join as urljoin

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  from .solrparser import solrlistener
  from requests_kerberos import HTTPKerberosAuth, OPTIONAL
except Exception as e:
  LOGGER.info("Libraries missing for Solr parser")

SOLR_DB_NAME = "Solr"
SOLR_CLOUD = "cloud"
SOLR_ROWS = 1000

class SolrConnProperties(ConnProperties):
  TYPEMAP = {
    float: arctypes.ARC_DATATYPE_REAL,
    int: arctypes.ARC_DATATYPE_INT,
    int: arctypes.ARC_DATATYPE_BIGINT,
    bool: arctypes.ARC_DATATYPE_BOOLEAN,
    str: arctypes.ARC_DATATYPE_STRING,
    'datetime': arctypes.ARC_DATATYPE_TIMESTAMP,
    'text': arctypes.ARC_DATATYPE_STRING,
  }
  def __init__(self):
    super(SolrConnProperties, self).__init__()
    self.supports_async = False
    self.joincharforarraytype = ','
    self.property_map = dict()
    self.supports_dataextract_target = True

    self.display_to_table_type = {
      'String': 'string',
      'Integer': 'pint',
      'Real': 'pfloat',
      'Boolean': 'boolean',
      'Timestamp': 'pdate'
    }

  @staticmethod
  def cursor_coltypes_to_arctypes(native_coltype):
    return SolrConnProperties.TYPEMAP.get(native_coltype, arctypes.ARC_DATATYPE_STRING)

  def add_property(self, name, value):
    self.property_map[name] = value

# A SolrConn is created every time a request is issued
class SolrConn(dataconn_nosqlapi.NoSqlApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_SOLR):
    dataconn_nosqlapi.NoSqlApi2Conn.__init__(self, connection_id)
    self.connection_type = conn_type
    self.solr_type = params.get('SOLRTYPE', SOLR_CLOUD)

    host = params.get('HOST', "")
    self.sock = params.get('SOCK', 'normal')
    self.sslverify = params.get('SSLCERT_NAME_MISMATCH', False)
    if not host.startswith("http://") and not host.startswith("https://"):
      if self.sock.startswith("ssl"):
        host = "https://" + host
      else:
        host = "http://" + host
    url = urlparse(host)
    port = params.get('PORT', "")
    if port:
      url = url._replace(netloc="{}:{}".format(url.hostname, port))
    url = url._replace(path=urljoin(url.path, 'solr/'))
    self.username = params.get('USERNAME', '_default')
    self.password = params.get('PASSWORD', '_default')
    self.baseurl = url.geturl()
    self.return_type = "json"
    self.impersonation = params.get('IMPERSONATION', False)
    self.auth = params.get('AUTH', '')
    self._properties = SolrConnProperties()
    self._properties.add_property('baseurl', self.baseurl)
    self._properties.add_property('wt', self.return_type)
    self._properties.joincharforarraytype = params.get('SOLR_JOINCHAR', ',')

  def close(self):
    pass

  def properties(self):
    return self._properties

  # Example of a faceted Stats query is this:
  # /select?q=*:*&rows=0&stats=true&stats.field=rules_count&stats.facet=locator_str
  def fieldStats(self, build_request, fieldName, querySuffix):
    return self._fieldStatsInternal(build_request, fieldName, querySuffix, minmax=False)

  def fieldStatsMinMax(self, build_request, fieldName, querySuffix):
    return self._fieldStatsInternal(build_request, fieldName, querySuffix, minmax=True)

  def _fieldStatsInternal(self, build_request, fieldName, querySuffix, minmax = False):
    collection = build_request.builder.builder.collection

    if build_request and build_request.dataset_request and build_request.dataset_request.dataset:
      c = build_request.dataset_request.dataset.getColumn(fieldName)
      if c:
        if c.attr.get('type') != 'string':
          if minmax == False:
            q = self.baseurl + collection + \
                '/select?wt=json&rows=0&stats=true&stats.field=%s' % fieldName
            q = q + querySuffix
            LOGGER.info('solr query: %s', q)
            data = self._solr_response(q, build_request.username).get('stats').get('stats_fields').get(fieldName)
            return data
          else:
            q = self.baseurl + collection + \
                '/select?wt=json&rows=1&fl=%s&sort=%s asc' % (fieldName, fieldName)
            q = q + querySuffix
            LOGGER.info('solr query: %s', q)
            mindatadoc = self._highlighting_enabled_docs(q, build_request.username, dimensions)
            if mindatadoc:
              mindata = mindatadoc[0].get(fieldName)
            else:
              return None

            q = self.baseurl + collection + \
                '/select?wt=json&rows=1&fl=%s&sort=%s desc' % (fieldName, fieldName)
            q = q + querySuffix
            LOGGER.info('solr query: %s', q)
            maxdatadoc = self._highlighting_enabled_docs(q, build_request.username, dimensions)
            if maxdatadoc:
              maxdata = maxdatadoc[0].get(fieldName)
            else:
              return None

            data = {}
            data['min'] = mindata
            data['max'] = maxdata
            return data

    return None

  def compute_autotime_gap(self, minTime, maxTime):
    dmin = datetime.strptime(minTime[0:19], '%Y-%m-%dT%H:%M:%S')
    dmax = datetime.strptime(maxTime[0:19], '%Y-%m-%dT%H:%M:%S')
    td = dmax - dmin
    minutes = 60
    hours = 3600
    days = 24 * hours
    years = 365 * days
    tdSeconds = td.seconds + td.days * days

    # Auto-compute the gap parameter in solr
    if tdSeconds > 20 * years:
      gap = '1YEAR'
    elif tdSeconds > 10 * years:
      # Minimum of 10 * 12 = 120 points
      # Maximum of 20 * 12 = 240 points
      gap = '1MONTH'
    elif tdSeconds > 5 * years:
      # Minimum of 26 * 5 = 130 points
      # Maximum of 26 * 10 = 260 points
      gap = '14DAY'
    elif tdSeconds > 2 * years:
      # Minimum of 52 * 2 = 104 points
      # Maximum of 52 * 5 = 260 points
      gap = '7DAY'
    elif tdSeconds > 1 * years:
      # Minimum of 120 points
      # Maximum of 120 * 2 = 240 points
      gap = '3DAY'
    elif tdSeconds > 180 * days:
      # Minimum of 180 points
      # Maximum of 365 points
      gap = '1DAY'
    elif tdSeconds > 90 * days:
      # Minimum of 90 * 2 = 180 points
      # Maximum of 180 * 2 = 360 points
      gap = '12HOUR'
    elif tdSeconds > 45 * days:
      # Minimum of 45 * 4 = 180 points
      # Maximum of 90 * 4 = 360 points
      gap = '6HOUR'
    elif tdSeconds > 10 * days:
      # Minimum of 10 * 6 = 60 points
      # Maximum of 45 * 6 = 270 points
      gap = '4HOUR'
    elif tdSeconds > 5 * days:
      # Minimum of 5 * 12 = 60 points
      # Maximum of 10 * 12 = 120 points
      gap = '2HOUR'
    elif tdSeconds > 1 * days:
      # Minimum of 24 points
      # Maximum of 5 * 24 = 120 points
      gap = '1HOUR'
    elif tdSeconds > 12 * hours:
      # Minimum of 12 * 6 = 72 points
      # Maximum of 24 * 6 = 144 points
      gap = '10MINUTE'
    elif tdSeconds > 6 * hours:
      # Minimum of 6 * 12 = 72 points
      # Maximum of 12 * 12 = 144 points
      gap = '5MINUTE'
    elif tdSeconds > 3 * hours:
      # Minimum of 3 * 30 = 90 points
      # Maximum of 6 * 30 = 180 points
      gap = '2MINUTE'
    elif tdSeconds > 1 * hours:
      # Minimum of 1 * 60 = 60 points
      # Maximum of 3 * 60 = 180 points
      gap = '1MINUTE'
    elif tdSeconds > 30 * minutes:
      # Minimum of 30 * 2 = 60 points
      # Maximum of 60 * 2 = 120 points
      gap = '30SECOND'
    elif tdSeconds > 15 * minutes:
      # Minimum of 15 * 4 = 60 points
      # Maximum of 30 * 4 = 120 points
      gap = '15SECOND'
    elif tdSeconds > 5 * minutes:
      # Minimum of 5 * 12 = 60 points
      # Maximum of 15 * 12 = 180 points
      gap = '5SECOND'
    elif tdSeconds > 1 * minutes:
      # Minimum of 60 points
      # Maximum of 5 * 60 = 300 points
      gap = '1SECOND'
    else:
      # Minimum of 1 point
      # Maximum of 60 points
      gap = '1SECOND'

    return gap

  def timeQuery(self, build_request, fieldName, minTime, maxTime, gap_override, querySuffix):
    if minTime is None or maxTime is None:
      # If compute stats results in an invalid time range, run a query that will
      # return empty results
      minTime = '1970-01-01T00:00:00Z'
      maxTime = '1970-01-01T00:00:00Z'
      gap = '1000YEAR'
    elif str(gap_override) != 'AUTO':
      gap = gap_override
    else:
      gap = self.compute_autotime_gap(minTime, maxTime)

    roundup = re.search('\d+(\S+)', gap).group(1)
    collection = build_request.builder.builder.collection
    q = self.baseurl + collection + '/select?'

    params = [
      'wt=json',
      'rows=0',
      'facet=on',
      'facet.range.start=%s/%s' % (minTime, roundup),
      'facet.range.end=%s' % maxTime,
      'facet.range=%s' % fieldName,
      'facet.limit=%s' % (50000),
      'facet.range.gap=%%2B%s' % gap,
    ]

    q += '&'.join(params)
    q = q + querySuffix

    return q


  def log_sync_runquery(self, query_urls,
      username=None, sample_pct=None, comment='', setCancelCbFn=None, **kwargs):
      try:
        for query_url in query_urls:
          self._solr_response(query_url, username)
      except Exception as e:
        e = str(e)
        if 'Solr instance is not running in SolrCloud mode' in e:
          e = 'This feature is supported for Solr running in SolrCloud mode.'
        elif 'solr.DenseVectorField' in query_url and 'error processing commands' in e:
          e = 'Note Vector field type is supported for Solr 9.0+'
        raise Exception(e)

  # used for data upload
  def create_sql_cmds(self, tablename, colnames, coltypes,
                                importuser, importsrcfile, external_table,
                                location, localesetting='C', type_classes=None,
                                vector_similarity_fn='euclidean', hnsw_max_conn='16',
                                hnsw_beam_width='100', **kwargs):
    return self.create_collection_cmds(tablename, colnames, coltypes, importuser,
      importsrcfile, localesetting, type_classes, vector_similarity_fn,
      hnsw_max_conn, hnsw_beam_width)

  def create_collection_cmds(self, collection_name, colnames,
                              coltypes, importuser, importsrcfile,
                              localesetting, type_classes, vector_similarity_fn,
                              hnsw_max_conn, hnsw_beam_width):
    cmds = []
    # 1. create config
    create_collection_config_cmd = self.baseurl + 'admin/configs?action=CREATE&'
    confName = f'{collection_name}Config'
    conf_params = [
      f'name={confName}',
      'baseConfigSet=_default',
      'configSetProp.immutable=false',
      'wt=json',
      'omitHeader=true',
    ]
    create_collection_config_cmd += '&'.join(conf_params)
    cmds.append(create_collection_config_cmd)

    # 2. create collection
    create_collection_cmd = self.baseurl + 'admin/collections?action=CREATE&'
    collection_params = [
      f'name={collection_name}',
      f'config={confName}',
      'numShards=1',
    ]
    create_collection_cmd += '&'.join(collection_params)
    cmds.append(create_collection_cmd)

    # 3. define vector field types and add language code to Text type
    knnvector_type_to_dim = {}
    new_coltypes = coltypes
    for i, type in enumerate(coltypes):
      if type == 'Vector' and type_classes:
        vector_type_dim = type_classes[i].dimension
        vector_type_name = f'knn_vector_{vector_type_dim}'
        knnvector_type_to_dim[vector_type_name] = vector_type_dim
        new_coltypes[i] = vector_type_name

      elif type == 'Text':
        # en is default, otherwise e.g. hu from hu_HU
        lang_code = 'en' if localesetting == 'C' else localesetting.split('_')[0]
        new_coltypes[i] = f'text_{lang_code}'
      else:
        new_coltypes[i] = self._properties.display_to_table_type[type]

    req = {
      'url': f'{self.baseurl}{collection_name}/schema',
      'body': {
        'add-field-type' : []
      }
    }

    for type_name, dim in knnvector_type_to_dim.items():
      type_params = {
                      'name': type_name,
                      'class': 'solr.DenseVectorField',
                      'vectorDimension': dim,
                      'similarityFunction': vector_similarity_fn,
                      'knnAlgorithm': 'hnsw',
                      'hnswMaxConnections': hnsw_max_conn,
                      'hnswBeamWidth': hnsw_beam_width
                    }
      req['body']['add-field-type'].append(type_params)

    new_vector_types_cmd = 'json:'+json.dumps(req)
    if len(knnvector_type_to_dim):
      cmds.append(new_vector_types_cmd)

    # 4. add fields to schema
    add_fields_req = {
      'url': f'{self.baseurl}{collection_name}/schema',
      'body': {
        'add-field' : []
      }
    }

    for i, colname in enumerate(colnames):
      field_params = {
        "name": colname,
        "type": new_coltypes[i],
        "multiValued": False
      }
      add_fields_req['body']['add-field'].append(field_params)

    collection_schema_cmd = 'json:'+json.dumps(add_fields_req)
    cmds.append(collection_schema_cmd)

    return cmds

  def sync_runquery(self, query_url, build_request=None,
      username=None, sample_pct=None, comment='', logQuery=False, stream=False, qcntl=None,
      profile=None, setCancelCbFn=None):
    """Run a SOLR query and return the results"""

    querySuffix = ''

    solrBuilder = None

    if build_request and build_request.builder:
      solrBuilder = build_request.builder.builder
      if solrBuilder.isSegmented:
        segments = []
        for s in solrBuilder.solr_query:
          # The call to getSolrSegments seems to be asynchronous. For now, wait a maximum of 1 second to get the segments.
          # If the segments are still empty, just return an empty query
          segs = solrlistener.getSolrSegments(s)
          start_time = time.time()
          while((len(segs) == 0) and time.time() - start_time < 1):
            pass
          segments.extend(segs)

        LOGGER.info('query string %s', solrBuilder.solr_query)
        LOGGER.info('query string segments %s', segments)
        datal = []
        for s in segments:
          querySuffix = '&q=' + s
          datal.append(self._internal_runquery(query_url, build_request=build_request,
              username=username, sample_pct=sample_pct, comment=comment,
              logQuery=logQuery, stream=stream, solrBuilder=solrBuilder,
              querySuffix=querySuffix, segmentName=s))

        for i in range(1, len(datal)):
          datal[0].data['rows'].extend(datal[i].data['rows'])

        rows_concatted = datal[0].data if len(datal) > 0 else []
        return rows_concatted

    data = self._internal_runquery(query_url, build_request=build_request,
        username=username, sample_pct=sample_pct, comment=comment,
        logQuery=logQuery, stream=stream, solrBuilder=solrBuilder,
        querySuffix=querySuffix)
    return data.data

  def _highlighting_enabled_docs(self, query_url, username, dimensions):
    solr_response = self._solr_response(query_url, username)
    docs = solr_response.get('response').get('docs')

    if 'highlighting' in solr_response:
      highlighting = solr_response.get('highlighting')
      for doc in docs:
        highlighting_obj = highlighting[doc['id']]
        for column,value in highlighting_obj.items():
          if column in dimensions:
            doc[column] = value

        # Remove id used for highlighting, if it is not in dimensions shelf
        if 'id' not in dimensions:
          del doc['id']
    return docs

  def _internal_runquery(self, query_url, build_request=None,
      username=None, sample_pct=None, comment='', logQuery=False, stream=False,
      solrBuilder=None, querySuffix='', segmentName=''):

    LOGGER.info('solr query: %s%s', query_url, querySuffix)
    isDate = False
    if query_url.startswith('columnstats'):
      c = query_url[len('columnstats') + 1:]
      stats = self.fieldStats(build_request, c, '&q=*:*')

      data = _ResultData()
      data.dataAddColumn('min', 'STRING')
      data.dataAddColumn('max', 'STRING')
      if stats is not None:
        data.dataAddRow([stats.get('min', 0), stats.get('max', 0)])
      return data

    if query_url.startswith('autotime'):
      possibleSuffix = query_url[len('autotime') + 1:].strip()

      if possibleSuffix:
        querySuffix += '&' + possibleSuffix

      facetName = build_request.builder.builder.dimensions[0]
      stats = self.fieldStatsMinMax(build_request, facetName, querySuffix)

      if stats:
        isDate = True
        query_url = self.timeQuery(build_request, facetName, stats.get('min'), stats.get('max'),
                                   solrBuilder.autotime_override, querySuffix)
        LOGGER.info('replaced solr query: %s', query_url)
      else:
        data = self._dataFromFacets(solrBuilder.dimensions, solrBuilder.aggregates)
        if solrBuilder.isSegmented:
          data.dataInsertSegment(solrBuilder.segmentIndex, segmentName)
        return data

    else:
      query_url = query_url + querySuffix

    getrows = None
    if build_request.dataset_request:
      getrows = build_request.dataset_request.dstype == 'getrows'
    if build_request.builder:
      collection = build_request.builder.builder.collection
    else:
      collection = None

    hasAggregates = False
    facets = []
    aggregates = None
    if build_request.builder and build_request.builder.builder.aggregates:
      hasAggregates = True
      aggregates = build_request.builder.builder.aggregates
      facets = build_request.builder.builder.dimensions

    dimensions = build_request.builder.builder.dimensions if build_request.builder else None
    distinct = build_request.dataset_request.dsreq.distinct \
      if (build_request.dataset_request and hasattr(build_request.dataset_request.dsreq, 'distinct')) else None

    if logQuery:
      actionlog_create(query=query_url,
                       dataconnection_id=self.connection_id,
                       dataconnection_type=self.connection_type,
                       sync=True)

    try:
      if getrows:
        data = self._highlighting_enabled_docs(query_url, username, dimensions)
      elif not collection:
        SolrConn._process_admin_query(query_url, facets)
        if facets:
          data = self._solr_response(query_url, username).get('facet_counts').get('facet_pivot') or \
                 self._solr_response(query_url, username).get('facet_counts').get('facet_ranges')
        else:
          data = self._highlighting_enabled_docs(query_url, username, dimensions)
      elif facets:
        if isDate:
          data = self._solr_response(query_url, username).get('facet_counts').get('facet_ranges')
        else:
          # Read faceted stats
          data = self._solr_response(query_url, username).get('facet_counts').get('facet_pivot')
      else:
        if hasAggregates:
          agg_field, agg_type = aggregates[0]

          data = _ResultData()
          rsp = self._solr_response(query_url, username)
          stats = self._solr_response(query_url, username).get('stats').get('stats_fields')

          row = []
          for agg_field, agg_type in aggregates:
            if agg_type == 'recordcount':
              data.dataAddColumn('Record Count', 'INT')
              row.append([rsp.get('response').get('numFound')])
            else:
              data.dataAddColumn(agg_type, 'FLOAT')
              row.append([stats.get(agg_field).get(agg_type, 0)])
          data.dataAddRow(row)
          return data
        else:
          data = self._highlighting_enabled_docs(query_url, username, dimensions)

      rs = self._process_results(data, facets, collection, distinct, dimensions, isDate, aggregates)
      if solrBuilder and solrBuilder.isSegmented:
        rs.dataInsertSegment(solrBuilder.segmentIndex, segmentName)

      return rs
    except:
      import traceback
      traceback.print_exc()
      raise

  @staticmethod
  def _process_admin_query(query, facets):
    splits = query.split('&')
    collection = splits[0]
    pat = '(.+?)/select'
    m = re.match(pat, collection)
    collection = m.group(1).split('/')[-1]
    for split in splits:
      if 'facet.field=' in split:
        facets.append(split.split('=')[1])
      elif 'facet.range=' in split:
        facets.append(split.split('=')[1])

  def execute(self, build_request, async_flag=False):
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      data = {
          'colnames': ['name'],
          'coltypes': ['STRING'],
          'rows': [[SOLR_DB_NAME]]
        }
      return data

    if build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      collections = []
      try:
        collections = self._collections(build_request.username)
      except:
        raise

      if self.solr_type == SOLR_CLOUD:
        rows = [[collection['name'], collection['encompassingcollections']] for collection in collections]
        colnames = ['name', 'encompassingcollections']
        coltypes = ['STRING', 'STRING']
      else:
        rows = [[collection] for collection in collections]
        colnames = ['name']
        coltypes = ['STRING']
      data = {
        'colnames': colnames,
        'coltypes': coltypes,
        'rows': rows
      }
      return data

    executor = BuildRequestExecutor(build_request)
    return executor.execute()

  def table_columns(self, table_name, username=None, client_ip=None):
    collection = table_name.split('.')[1]
    schema = self._schema(collection, username, True)
    data = {
          'colnames': [str(x['name']) for x in schema],
          'coltypes': [str(x['type']) for x in schema],
          'partitions': []
        }
    return data

  def _dataFromFacets(self, facets, aggregates, distinct=False):
    data = _ResultData()
    for f in facets:
      data.dataAddColumn(f, 'STRING')

    if not distinct and aggregates:
      for agg_field, agg_type in aggregates:
        if agg_type != 'recordcount':
          data.dataAddColumn(agg_type, 'FLOAT')
        else:
          data.dataAddColumn('Record Count', 'INT')

    return data

  # Used to flatten arbitrary depth facet_pivot object when query is faceted stats.
  # Accepts facet_pivot object as an argument. Returns dict of depth 0.
  # e.g., if input is ->
  # "facet_pivot":{
  #        "Position,Absences":[
  #           {
  #              "field":"Position",
  #              "value":"Production Technician I",
  #              "count":1507,
  #              "pivot":[
  #                 {
  #                    "field":"Absences",
  #                    "value":4,
  #                    "count":121,
  #                    "stats":{
  #                       "stats_fields":{
  #                          "Salary":{
  #                             "mean":56122.454545454544
  #                          }}}},
  #                 ...
  # }
  # Then output is ->
  # {'0*value': 'Production Technician I',
  #  '0*count': 1507,
  #  '0*pivot_0*value': 4,
  #  '0*pivot_0*count': 121,
  #  '0*pivot_0*stats_stats_fields_Salary_mean': 56122.454545454544,
  #  ...
  # }
  @staticmethod
  def flattenStatsRspJson(nested_json):
    out = {}

    def flatten(json_obj, name=''):
      if type(json_obj) is dict:
        for key in json_obj:
          if key not in ['field']:
            flatten(json_obj[key], f'{name}{key}_')
      elif type(json_obj) is list:
        i = 0
        for key in json_obj:
          flatten(key, f'{name}{i}*')
          i += 1
      else:
        out[name[:-1]] = json_obj

    flatten(nested_json)
    return out

  def _process_results(self, results, facets, collection, distinct, dimensions, isDate, aggregates):
    # if isDate == False
    #   results are of the form [{[],[],...}, {[],[],...}, ...]
    #   i.e. a list(dict())
    #   where length of list is limit clause (number or rows)
    #   and length of each dictionary in list is the number of rows
    #
    # if isDate == True
    #   results are of the form {
    #     'timestamp1': count1,
    #     'timestamp2': count2,
    #     ...
    #   }

    data = _ResultData()

    if facets:
      data = self._dataFromFacets(facets, aggregates, distinct)

      rows = results.get(','.join(facets))
      if not rows:
        return data

      if isDate:
        for k, v in rows.items():
          if k in ['gap', 'start', 'end']:
            continue

          for i in range(int(len(v) / 2)):
            data.dataAddRow([v[i * 2], v[i * 2 + 1]])

        data.dataSortRows()
        return data

      if aggregates:
        flat_stats_response = self.flattenStatsRspJson(rows)
        facet_fields_count = len(facets)

        # Leave only deepest recordcount and stats key-value pairs. facet_pivot object
        # contains recordcount and stats for each facet level, but we need only
        # the deepest values in visuals.
        agg_list = {}
        for k,v in flat_stats_response.items():
          key_list = k.split('*')
          if len(key_list) == facet_fields_count + 1:
            if (key_list)[facet_fields_count].startswith('stats') or (key_list)[facet_fields_count].startswith('count'):
              agg_list[k] = v

        set_of_ids_dict = {}
        for k in agg_list.keys():
          trim_stats_or_count = k.split('*')[:-1]
          only_id = '*'.join(trim_stats_or_count)
          set_of_ids_dict[only_id] = None

        list_of_ids = list(set_of_ids_dict.keys()) # pivots

        agg_strings = [f'{agg[0]}_{agg[1]}' for agg in aggregates] # agg column names order in UI
        for id in list_of_ids:
          row = []

          # Adding dimension values
          for i in range(facet_fields_count):
            depth_id_list = id.split('*')[:i + 1]
            depth_id_str = '*'.join(depth_id_list)
            row.append(flat_stats_response[f'{depth_id_str}*value'])


          # Adding stats values
          aggs = {}
          for key, value in agg_list.items():
            key_pivot_part_list = key.split('*')[:-1]
            key_pivot_part_str = '*'.join(key_pivot_part_list)
            if key_pivot_part_str == id:
              aggs[key] = value

          # Aggs order in json responce is different from UI's, this loop puts values in correct order
          for agg_string in agg_strings:
            if re.match("sum\(1\)\sas\s'.+'", agg_string):
              row.extend([v for k, v in aggs.items() if k.endswith('*count')])
            else:
              row.extend([v for k, v in aggs.items() if k.split('*')[-1] == f'stats_stats_fields_{agg_string}'])

          data.dataAddRow(row)
        return data

      for row in rows:
        firstFacet = row.get('value', '')
        if len(facets) == 1:
          data.dataAddRow([firstFacet, row.get('count', 0)])
          continue

        pivotRows = row.get('pivot', [])
        if pivotRows:
          for pivotRow in pivotRows:
            secondFacet = pivotRow.get('value')
            data.dataAddRow([firstFacet, secondFacet, pivotRow.get('count', 0)])

      return data

    if dimensions:
      schemaList = dimensions
    else:
      schemaSet = set()
      schemaList = []
      for row in results:
        for k in list(row.keys()):
          if k in schemaSet:
            continue
          schemaSet.add(k)
          schemaList.append(k)

    # create a map of colname to colindex
    colmap = dict()

    # create array of colnames and coltypes from schema
    for index, colname in enumerate(schemaList):
      colmap[colname] = index
      data.dataAddColumn(colname, 'UNKNOWN')

    # process results
    for row in results:
      # start with empty values
      row_data = [""] * len(schemaList)
      for column in row:
        data_point = self.properties().joincharforarraytype.join(str(v) for v in row[column]) if type(row[column]) is list else row[column]
        colindex = colmap[column]
        row_data[colindex] = data_point
        if data.data["coltypes"][colindex] == 'UNKNOWN':
          coltype =  SolrConnProperties.cursor_coltypes_to_arctypes(type(data_point))
          data.data["coltypes"][colindex] = coltype

      data.dataAddRow(row_data)

    return data

  def _solr_response(self, url: str, doAs=None, *fields, **kwargs):
    default = kwargs.get('default')
    function = kwargs.get('function')

    doAsKey = getattr(settings, 'SOLR_IMPERSONATION_KEY', 'doAs')

    body = None
    if url.startswith('json:'):
      req = json.loads(url[len('json:'):])
      url = req['url']
      body = req['body']

    # without strong authentication like kerberos we specify the username
    solrauth = None
    if self.auth == '':
      solrauth = None
      if self.username != '' and self.password == '':
        url = url + ('&user.name=%s' % (self.username))

    if self.auth == 'nosasl' and  self.username != '' and self.password != '':
      solrauth = HTTPBasicAuth(self.username, self.password)
    else:
      solrauth = None

    if self.auth == 'kerberos':
      solrauth = HTTPKerberosAuth(mutual_authentication=OPTIONAL)

    if doAs is not None and self.impersonation:
      url = url + '&' + doAsKey + '=' + doAs

    LOGGER.debug('_solr_response: %s', url)
    if body is None:
      response = requests.get(url, allow_redirects=True, auth=solrauth, verify=(not self.sslverify))
    else:
      response = requests.post(url, allow_redirects=True, auth=solrauth, verify=(not self.sslverify),
                               json=body)

    if response.status_code != 200:
      errtext = ''
      try:
        errtext = response.json().get('error', {}).get('msg')
      except Exception:
        errtext = "%s:%s:%s" % (response.status_code, response.reason, response.text)

      raise Exception('Error reading Solr data: ' + errtext)
    else:
      data = response.json()
      for field in fields:
        data = data.get(field, default)

      if function is not None:
        data = list(map(function, data))

      return data

  # return list of collections
  def _collections(self, username=None):
    collections_list = []
    if self.solr_type == SOLR_CLOUD:
      url = self.baseurl + ("admin/collections?action=LIST&wt=%s" % self.return_type)
      collections_list_raw = self._solr_response(url, username, 'collections', default=[])
      for collection in collections_list_raw:
        collections_list.append(
          {'name':collection,
           'encompassingcollections': ''
          })
      url = self.baseurl + ("admin/collections?action=CLUSTERSTATUS&wt=%s" % self.return_type)
      aliases_list_raw = self._solr_response(url, username, 'cluster', 'aliases', default=[])
      for alias in aliases_list_raw:
        collections_list.append(
          {'name': alias,
           'encompassingcollections':aliases_list_raw[alias]
          })
      return collections_list
    else:
      url = self.baseurl + ("admin/cores?&indexInfo=false&wt=%s" % self.return_type)
      return self._solr_response(url, username, 'status', default=[], function=str)

  # return schema for a collection
  def _schema(self, collection, username=None, asList=True):
    def solr_arcviz_type_map(solr_type):
      """
      convert solr type to a type that can be consumed by arcviz.
      We are converting all list of values types (e.g. tdoubles, floats, strings)
      to string because it is not clear if arcviz is capable of processing a
      list of values at this moment.
      """
      solr_type_map = {
        "double":"double",
        "tdouble":"double",
        "pdouble":"double",
        "float":"float",
        "tfloat":"float",
        "pfloat":"float",
        "long":"bigint",
        "tlong":"bigint",
        "plong":"bigint",
        "int":"int",
        "tint":"int",
        "pint":"int",
        "date":"timestamp",
        "tdate":"timestamp",
        "pdate":"timestamp",
        "boolean":"boolean",
        "binary":"binary",
      }
      if solr_type.lower().startswith('knn_vector_'):
        return 'vector'
      elif solr_type.lower().startswith('text_'):
        return 'text'
      elif solr_type.lower() in solr_type_map:
        return solr_type_map[solr_type.lower()]
      else:
        return 'string'

    schema = []
    schema_dict = dict()

    url = self.baseurl + "%s/schema?&wt=%s" % (collection, self.return_type)
    response = self._solr_response(url, username, 'schema', 'fields', default=None)
    if response:
      for field in response:
        schema.append({'name': field['name'], 'type': solr_arcviz_type_map(field['type'])})
        schema_dict[field['name']] = solr_arcviz_type_map(field['type'])
    return schema if asList else schema_dict

  # return list of shard names and states for a collection
  def _shards(self, collection, username=None):
    shards = []
    url = self.baseurl + ("admin/collections?action=CLUSTERSTATUS&wt=%s" % self.return_type) if self.baseurl else None
    if url:
      try:
        shards_dict = self._solr_response(url, username, 'cluster', 'collections', collection, 'shards', default=[])
        for shard in shards_dict:
          shards.append({'name':shard, 'state':shards_dict[shard]['state']})
      except:
        raise
    return shards

  # return list of replicas for a collection and shard
  def _replicas(self, collection, shard, username=None):
    replicas = []
    url = self.baseurl + ("/admin/collections?action=CLUSTERSTATUS&wt=%s" % self.return_type) if self.baseurl else None
    if url:
      try:
        replicas_dict = self._solr_response(url, username, 'cluster', 'collections', collection, 'shards', shard, 'replicas',
                                                default=[])
        for replica in replicas_dict:
          replicas.append({'replica': replica, 'core': replicas_dict[replica]['core']})
      except:
        raise
    return replicas

class _ResultData(object):
  def __init__(self):
    self.data = dict()
    self.data['colnames'] = list()
    self.data['coltypes'] = list()
    self.data['rows'] = list()
    self.data['info'] = ['Query:']

  def rows(self):
    return self.data['rows']

  def dataInsertSegment(self, segmentIndex, segmentData):
    self.data['colnames'].insert(segmentIndex, 'Segment')
    self.data['coltypes'].insert(segmentIndex, 'STRING')
    for row in self.data['rows']:
      row.insert(segmentIndex, segmentData)

  def dataAddColumn(self, colname, coltype):
    self.data['colnames'].append(colname)
    self.data['coltypes'].append(coltype)

  def dataAddRow(self, row):
    self.data['rows'].append(row)

  def dataSortRows(self):
    self.data['rows'].sort()

  def join(self, rs):
    self.data['colnames'].extend(rs.data['colnames'][1:])
    self.data['coltypes'].extend(rs.data['coltypes'][1:])
    h = {}
    for r in self.data['rows']:
      h[r[0]] = r

    for r in rs.data['rows']:
      h[r[0]].extend(r[1:])

