# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from . import arctypes
from . import dataconn_dbapi2
import logging, sys
import math
import re
import json

from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .connections_types import DATACONN_SQLITE
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from os.path import abspath, normpath, dirname, join, expanduser
from vizlib.querylog import QueryLog
from vizlib.querylog_helper import ComponentType

try:
  from sqlite3 import dbapi2
except Exception:
  pass

LOGGER = logging.getLogger("arcweb")

def vector_distance(v1s: str, v2s: str):
  """
  Calculates the Euclidean distance between two vectors.
  The vectors are expected to be JSON-serialized number arrays (e.g. "[1,2,..]")
  """
  try:
    v1 = json.loads(v1s)
    v2 = json.loads(v2s)
    sum = 0
    for v1v,v2v in zip(v1,v2):
      sum += (v1v-v2v)**2
    return math.sqrt(sum)
  except Exception as e:
    LOGGER.error(e, stack_info=True)
    raise e

class SqliteConnProperties(ConnProperties):
  TYPEMAP = {
    'INT': arctypes.ARC_DATATYPE_INT,
    'INTEGER': arctypes.ARC_DATATYPE_INT,
    'TINYINT': arctypes.ARC_DATATYPE_INT,
    'SMALLINT': arctypes.ARC_DATATYPE_INT,
    'MEDIUMINT': arctypes.ARC_DATATYPE_INT,
    'BIGINT': arctypes.ARC_DATATYPE_BIGINT,
    'DOUBLE': arctypes.ARC_DATATYPE_DOUBLE,
    'REAL': arctypes.ARC_DATATYPE_DOUBLE,
    'FLOAT': arctypes.ARC_DATATYPE_DOUBLE,
    'TEXT': arctypes.ARC_DATATYPE_STRING,
    'STRING': arctypes.ARC_DATATYPE_STRING,
    'DATE': arctypes.ARC_DATATYPE_TIMESTAMP,
    'DATETIME': arctypes.ARC_DATATYPE_TIMESTAMP,
    'TIMESTAMP': arctypes.ARC_DATATYPE_TIMESTAMP,
    'BOOLEAN': arctypes.ARC_DATATYPE_BOOLEAN,
  }

  def __init__(self):
    super(SqliteConnProperties, self).__init__()
    self.supports_partitions = False
    self.supports_async = False
    # sqlite connections must be used from within their threads
    # We cannot store per-client connections, so simply don't reuse for now
    self.supports_reuse = False
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = 'main'
    self.supports_import_data_dbmod = False
    self.escape_backslash = False

    self.string_type = "TEXT"
    self.timestamp_type = "TEXT"
    self.date_type = "TEXT"
    self.booleanfalsevalue = '0'
    self.booleantruevalue = '1'
    self.supports_deriveddata = True
    self.supports_dataextract_source = True
    self.supports_dataextract_target = True
    self.display_to_table_type = {
      "String": "TEXT",
      "Integer": "INTEGER",
      "Real": "DOUBLE",
      "Boolean": "BOOLEAN",
      "Timestamp": "TIMESTAMP"
    }
    self.version = dbapi2.sqlite_version

    # sqlite limit 0 does not return the right column types
    self.supports_limit_0 = False

class SqliteConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_SQLITE):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    if getattr(sys, '_MEIPASS', False):
      app_path = sys._MEIPASS
    elif getattr(sys, 'frozen', False):
      app_path = dirname(sys.executable)
    else:
      app_path = "."
    db_path = abspath(join(app_path,normpath(expanduser(params['FILE']))))

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    if not for_testing:

      connection = dbapi2.connect(database=db_path)
      connection.create_function("vector_distance", 2, vector_distance)
      self.set_connection(connection)

    self.dbname = 'main'
    self.connection_type = conn_type

  def properties(self):
    return SqliteConnProperties()

  # Overrides table_columns because SQLite gives no column type information
  # as part of select query results, and so table create cmd has to be queried
  # to figure out table columns, and for other queries, column types have to
  # to guessed through column values
  def table_columns(self, db_table_name, username=None, client_ip=None):
    properties = self.properties()
    table_name = db_table_name.split(".")[-1]
    query = 'PRAGMA table_info(%s)' % self.quote_tablename(table_name)
    with QueryLog(username=username,
                  component_type=ComponentType.DATACONNECTION,
                  dataconnection_id=self.connection_id,
                  dataconnection_type=self.connection_type,
                  query=query,
                  sync=True,
                  client_ip=client_ip) as q:
      cursor = self.cursor_get()
      try:
        cursor.execute(query)
        # returned fields are:
        #   columnid, columnname, columntype, notnull, defaultvalue, primarykey
        rows = cursor.fetchall()
        colnames = [row[1] for row in rows]
        coltypes = [row[2] for row in rows]
        coltypes = [properties.TYPEMAP.get(coltype.upper(),
                     arctypes.ARC_DATATYPE_STRING) for coltype in coltypes]
        data = {
          'colnames': colnames,
          'coltypes': coltypes,
          'partitions': []
        }
        return data
      finally:
        self.cursor_close(cursor)

  # execute query(ies)
  def execute(self,
              build_request,
              async_flag=False,
              sample_pct=None,
              comment=None,
              stream=False):

    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      build_request.query = '.databases'
      data = {
        'colnames': ['dbname'],
        'coltypes': ['STRING'],
        'rows': json.dumps([[self.dbname]])
      }
      return data

    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      build_request.query = \
          """SELECT name FROM %s.sqlite_master WHERE type='table' OR type='view' AND NOT name='sqlite_sequence' ORDER BY name""" % \
          build_request.query.split()[-1]
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)
    return executor.execute(query=build_request.query,
                            async_flag=async_flag,
                            stream=stream)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "aggregate functions are not allowed in the GROUP BY clause" not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr

  def apply_types(self, value, type):
    try:
      if value is None:
        value = "NULL"
      elif type == "String":
        value = re.sub("'", r"''", value)
        value = "'%s'" % value
      elif type == "Timestamp":
        value = "'%s'" % value
      elif type == "Real":
        value = re.sub(",", ".", value)
      elif type == "Boolean":
        if value:
          value = 1
        else:
          value = 0
    except Exception as e:
      pass

    return value


