# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from . import dataconn_odbc
from .connections_types import DATACONN_TERADATA
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from .dataconn_dataengine import BuildRequestExecutor

import re, json
import logging

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  import teradata
except Exception as e:
  # LOGGER.info("Did not find Teradata library")
  pass

class TeradataConnProperties(dataconn_odbc.OdbcConnProperties):
  def __init__(self):
    super(TeradataConnProperties, self).__init__()
    self.quote = '"'
    self.quote_end = '"'
    self.quote_alias = '"'
    self.quote_escape_character = '"'
    self.quote_string_constant = "'"
    self.supports_limit = False
    self.supports_limit_sample = True
    self.supports_partitions = False
    self.supports_async = False
    self.supports_backticks = False
    self.supports_reuse = False
    self.alias_supports_backticks = False
    self.string_type = "VARCHAR(1024)"
    self.boolean_type = "INTEGER"
    self.date_type = "DATE"
    self.real_type = "FLOAT"

    self.supports_top = True
    self.supports_top_brackets = False

    self.display_to_table_type = {
      "String": "varchar(1024)",
      "Integer": "bigint",
      "Real": "decimal(20, 6)",
      "Boolean": "byteint",
      "Timestamp": "timestamp"
    }

    self.supports_limit_0 = False

class TeradataConn(dataconn_odbc.OdbcConn):
  def __init__(self, params, connection_id, conn_type=DATACONN_TERADATA):
    super(TeradataConn, self).__init__(params, connection_id, conn_type=conn_type)

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    self.connection_type = conn_type

    if not for_testing:
      udaExec = teradata.UdaExec(appName="Arcadia", version="1.0",
          logConsole=False)
      self.connection = udaExec.connect(method="odbc", dsn=params.get("DSN"),
          username=params.get("USERNAME"), password=params.get("PASSWORD"))

    self.p = TeradataConnProperties()

  def properties(self):
    return TeradataConnProperties()

  def create_sql_cmds(self, tablename, headernames, headertypes,
                                  importuser, importsrcfile, external_table,
                                  location, localesetting='C', type_classes=None, **kwargs):
    colist = []
    for n, t in zip(headernames, headertypes):
      colist.append("{quote}{name}{quote_end} {type}".format(
          quote=self.p.quote, quote_end=self.p.quote_end, name=n,
          type=self.p.display_to_table_type[t]))

    sqlcmd = "CREATE MULTISET TABLE {tblname} ({colist})".format(
              tblname=self.quote_tablename(tablename), colist=",".join(colist))

    return sqlcmd

  def execute(self, build_request, async_flag=False):

    query = None
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = "SELECT DISTINCT \"databasename\" FROM dbc.tablesv order by databasename asc"
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      schema = build_request.query.split()[-1]
      query = "SELECT tablename FROM dbc.tablesv where tablekind in ('T', 'V') and databasename='%s' order by tablename asc" % (schema)
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          data = self.validate_expression(expression, build_request, exprType)
          return json.dumps(data)


    executor = BuildRequestExecutor(build_request)

    LOGGER.info(query)
    return executor.execute(query=query,
                            async_flag=async_flag)

