# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from . import dataconn_dbapi2
from .connections_types import DATACONN_VERTICA
from .dataconn_dataengine import ConnProperties, BuildRequestExecutor
from .data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from django.conf import settings
from .dataconn_dataengine import DEFAULT_UNQUOTE

import json
from . import arctypes
import logging
import re

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  import vertica_python as dbapi2
except Exception:
  # LOGGER.info("Libraries missing for Vertica")
  pass

class VerticaConnProperties(ConnProperties):
  TYPEMAP = {
    0:     arctypes.ARC_DATATYPE_UNKNOWN,
    1:     arctypes.ARC_DATATYPE_UNKNOWN,
    2:     arctypes.ARC_DATATYPE_UNKNOWN,
    3:     arctypes.ARC_DATATYPE_UNKNOWN,
    4:     arctypes.ARC_DATATYPE_UNKNOWN,
    5:     arctypes.ARC_DATATYPE_BOOLEAN,
    6:     arctypes.ARC_DATATYPE_BIGINT,
    7:     arctypes.ARC_DATATYPE_REAL,
    8:     arctypes.ARC_DATATYPE_STRING,
    9:     arctypes.ARC_DATATYPE_STRING,
    10:     arctypes.ARC_DATATYPE_TIMESTAMP,
    11:     arctypes.ARC_DATATYPE_TIMESTAMP,
    12:     arctypes.ARC_DATATYPE_TIMESTAMP,
    13:     arctypes.ARC_DATATYPE_TIMESTAMP,
    14:     arctypes.ARC_DATATYPE_TIMESTAMP,
    15:     arctypes.ARC_DATATYPE_TIMESTAMP,
    16:     arctypes.ARC_DATATYPE_DOUBLE,
    17:     arctypes.ARC_DATATYPE_STRING,
    18:     arctypes.ARC_DATATYPE_UNKNOWN,
  }

  def __init__(self):
    super(VerticaConnProperties, self).__init__()
    self.quote = '"'
    self.quote_end = '"'
    self.quote_alias = '"'
    self.quote_escape_character = '"'
    self.quote_string_constant = "'"
    self.supports_partitions = False
    self.supports_async = False
    self.supports_backticks = False
    self.supports_reuse = False
    self.supports_import_data_perm = False
    self.supports_import_data_dbname = 'public'
    self.alias_supports_backticks = False
    self.string_type = "VARCHAR"
    self.date_type = "date"
    self.double_type = "double precision"
    self.supports_nulls_order = True

  @staticmethod
  def cursor_coltype_to_arctype(native_coltype):
    return VerticaConnProperties.TYPEMAP.get(native_coltype,
        arctypes.ARC_DATATYPE_STRING)

class VerticaConn(dataconn_dbapi2.DbApi2Conn):
  def __init__(self, params, connection_id, conn_type=DATACONN_VERTICA):
    dataconn_dbapi2.DbApi2Conn.__init__(self, params, connection_id)
    self.dbname = None
    conn_params = {}

    for_testing = False
    if params.get('FOR_TESTING'):
      for_testing = True

    LOGGER.info("VerticaConn")

    if params.get('HOST'):
      conn_params['host'] = str(params.get('HOST').strip())
    if params.get('PORT'):
      conn_params['port'] = int(params.get('PORT'))
    if params.get('USERNAME'):
      conn_params['user'] = str(params.get('USERNAME'))
    if params.get('PASSWORD'):
      conn_params['password'] = str(params.get('PASSWORD'))
    if params.get('DATABASE'):
      self.dbname = str(params.get('DATABASE'))
      conn_params['database'] = str(params.get('DATABASE').strip())
    if not self.dbname:
      raise Exception('Database name not specified')

    if not for_testing:
      self.connection = dbapi2.connect(**conn_params)
    self.connection_type = conn_type

  def properties(self):
    return VerticaConnProperties()

  def _get_unquote_regex(self):
    if self.compiled_unquote_regex is not None:
      return self.compiled_unquote_regex
    connections_quotes_regex = getattr(settings, 'VALIDATION_QUOTES_REGEX', {})
    if self.connection_type in connections_quotes_regex:
      self.compiled_unquote_regex = connections_quotes_regex[self.connection_type]
    else:
      # adding regex for removing connection specific white list statement
      self.compiled_unquote_regex = DEFAULT_UNQUOTE + \
            [re.compile(r'/\*([^\*])*\*/'),        # sql comments: /* my comment */
             re.compile(r'extract\s*\([^\)]*\)')]  # extract statement: extract(month from [time]::timestamp)
    return self.compiled_unquote_regex

  def execute(self, build_request, async_flag=False):

    query = None
    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      query = "SELECT distinct table_schema FROM v_catalog.tables "
    elif build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      query = "SELECT table_name FROM v_catalog.all_tables " + \
          "WHERE table_type in ('TABLE','VIEW') AND " + \
          "schema_name ILIKE ('%s') ORDER BY table_name" % build_request.query.split()[-1]
    elif build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            LOGGER.exception(e)
            data = {'is_valid': 'unknown', 'error_msg': 'Error validating expression: ' + str(e)}
          return data

    executor = BuildRequestExecutor(build_request)

    return executor.execute(query=query,
                            async_flag=async_flag)

  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = builder.expr_hasAggregate(expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if "in the GROUP BY" not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr
