# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from dataconn.arctypes import ARC_DATATYPE_STRING
from dataconn.dataconn_dbapi2 import DbApi2Conn
from dataconn.data_api_cmds import SPECIAL_CMDS, VALIDATE_CMDS
from dataconn.dataconn_dataengine import BuildRequest, BuildRequestExecutor, ConnProperties
from vizlib.dataset import ModelDataset
from vizlib.validator_builder import ExpressionValidator

import logging
import re

LOGGER = logging.getLogger('arcweb.dataconn.ext')

class ExtensionConnProperties(ConnProperties):

  def __init__(self):
    super(ExtensionConnProperties, self).__init__()

  def get_typemap(self) -> dict:
    raise NotImplementedError()

  def cursor_coltype_to_arctype(self, native_type: str):
    return self.get_typemap().get(native_type, ARC_DATATYPE_STRING)


class ExtensionDataConn(DbApi2Conn):
  def __init__(self, params: dict, connection_id: int, conn_type: str):
    DbApi2Conn.__init__(self, params, connection_id)

    self.connection_type = conn_type
    self.conn_params = params

  # Overridden to create the connection object lazily
  def get_connection(self, username=None):
    if self.connection:
      return self.connection

    try:
      self.connection = self.create_connection(self.conn_params, username)
    except Exception as e:
      LOGGER.error(e)
      raise Exception('Could not connect to database')

    return self.connection

  def properties(self) -> ExtensionConnProperties:
    raise NotImplementedError()

  def create_connection(self, params: dict, username: str = None):
    '''
    Retrun a DB API 2.0 connection instance
    '''
    raise NotImplementedError()

  def show_databases(self, build_request: BuildRequest):
    raise NotImplementedError()

  def show_tables(self, build_request: BuildRequest, database: str):
    raise NotImplementedError()

  def execute(self,
              build_request: BuildRequest,
              async_flag=False,
              sample_pct=None,
              comment=None,
              stream=False):

    if build_request.query == SPECIAL_CMDS['SHOW_DATABASES']:
      return self.show_databases(build_request)

    if build_request.query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
      dbname = build_request.dataset_request.dsreq.dbname
      return self.show_tables(build_request, dbname)

    if build_request.query.startswith(SPECIAL_CMDS['VALIDATE']):
      for exprType, cmd in VALIDATE_CMDS.items():
        if build_request.query.startswith(cmd):
          expression = build_request.query[len(cmd):].strip()
          try:
            data = self.validate_expression(expression, build_request, exprType)
          except Exception as e:
            # if exception is caught above we set state to unknown
            error_message = 'Error validating expression'
            LOGGER.exception(error_message)
            data = {'is_valid': 'unknown', 'error_msg': '%s: %s' % (error_message, str(e))}
          return data

    executor = BuildRequestExecutor(build_request)
    return executor.execute(query=build_request.query,
                            async_flag=async_flag,
                            stream=stream)
  # used by data import to create the values list to insert.
  #   - type is a messytable type
  #   - value is read from the import source
  # TODO: this function has so many copies, should it go to DataEngine?
  def apply_types(self, value, type:str):
    try:
      if value is None:
        value = "NULL"
      elif type == "String":
        value = re.sub(r"\\", r"\\\\", value)
        value = re.sub("'", r"\'", value)
        value = "'%s'" % value
      elif type == "Timestamp":
        value = "'%s'" % value
      elif type == "Real":
        value = re.sub(",", ".", value)
    except Exception as e:
      pass

    return value

  # TODO: same as above, at least 10 copies:
  def expression_isAggregate(self, expression, **kwargs):
    # Now see if expression has aggregation
    builder = kwargs.get('builder')
    dataset = kwargs.get('dataset')
    username = kwargs.get('username')
    if dataset is None or builder is None or username is None:
      raise ValueError('builder, dataset and username are required')

    hasaggr_query = self.aggregate_detection_query(builder, expression, dataset)
    hasaggr = "false"
    try:
      self.log_sync_runquery(hasaggr_query, username=username)
    except Exception as e:
      if self.aggregate_detection_error() not in str(e):
        raise
      else:
        hasaggr = "true"

    return hasaggr

  def aggregate_detection_query(self, builder: ExpressionValidator, expression: str, dataset: ModelDataset):
    return builder.expr_hasAggregate(expression, dataset)

  def aggregate_detection_error(self):
    """
    Error message that can be used to detect when an aggregate is being used in a group by expression
    """
    raise NotImplementedError()
