# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from dataconn.ext.base import ExtensionConnProperties, ExtensionDataConn
from vizlib.dataset import ModelDataset
from vizlib.validator_builder import ExpressionValidator

import dataconn.arctypes as arctypes
import json
import logging
import re

LOG = logging.getLogger('arcweb.dataconn.ext.phoenix')

TYPEMAP = {
  'INTEGER': arctypes.ARC_DATATYPE_INT,
  'UNSIGNED_INT': arctypes.ARC_DATATYPE_INT,
  'UNSIGNED_LONG': arctypes.ARC_DATATYPE_INT,
  'TINYINT': arctypes.ARC_DATATYPE_INT,
  'UNSIGNED_TINYINT': arctypes.ARC_DATATYPE_INT,
  'SMALLINT': arctypes.ARC_DATATYPE_INT,
  'UNSIGNED_SMALLINT': arctypes.ARC_DATATYPE_INT,

  'BIGINT': arctypes.ARC_DATATYPE_BIGINT,

  'FLOAT': arctypes.ARC_DATATYPE_DOUBLE,
  'UNSIGNED_FLOAT': arctypes.ARC_DATATYPE_DOUBLE,
  'DOUBLE': arctypes.ARC_DATATYPE_DOUBLE,
  'UNSIGNED_DOUBLE': arctypes.ARC_DATATYPE_DOUBLE,
  'DECIMAL': arctypes.ARC_DATATYPE_DOUBLE,

  'CHAR': arctypes.ARC_DATATYPE_STRING,
  'VARCHAR': arctypes.ARC_DATATYPE_STRING,

  'DATE': arctypes.ARC_DATATYPE_TIMESTAMP,
  'UNSIGNED_DATE': arctypes.ARC_DATATYPE_TIMESTAMP,
  'TIME': arctypes.ARC_DATATYPE_TIMESTAMP,
  'UNSIGNED_TIME': arctypes.ARC_DATATYPE_TIMESTAMP,
  'TIMESTAMP': arctypes.ARC_DATATYPE_TIMESTAMP,
  'UNSIGNED_TIMESTAMP': arctypes.ARC_DATATYPE_TIMESTAMP,

  'BOOLEAN': arctypes.ARC_DATATYPE_BOOLEAN,

  'BINARY': arctypes.ARC_DATATYPE_BINARY,
  'VARBINARY': arctypes.ARC_DATATYPE_BINARY,
}


class PhoenixConnProperties(ExtensionConnProperties):

  def __init__(self):
    super(PhoenixConnProperties, self).__init__()
    self.supports_partitions = False
    self.supports_async = False
    self.supports_reuse = False
    self.quote_alias = '"'
    self.quote_escape_character = '"'
    self.quote_string_constant = "'"
    self.quote = '"'
    self.quote_end = '"'
    self.supports_groupby_colindex = False
    self.supports_deriveddata = True
    self.supports_dataextract_source = True
    self.supports_dataextract_target = False # might reconsider this later
    # sadly, we can insert rows only one-by-one :-(
    self.supports_multi_row_insert = False
    # subselect aliases does not seem to work:
    self.supports_subselect_alias = False

    # used in dataset field casts:
    self.string_type = "VARCHAR"
    self.boolean_type = "BOOLEAN"
    self.timestamp_type = "TIMESTAMP"
    self.integer_type = "INTEGER"
    self.real_type = "DOUBLE"

    # needed for data import:
    self.display_to_table_type = {
      "String": "varchar",
      "Integer": "bigint",
      "Real": "double",
      "Boolean": "boolean",
      "Timestamp": "timestamp",
      # duplicate each type so that they can be used as primary key:
      "String PRIMARY KEY": "varchar PRIMARY KEY",
      "Integer PRIMARY KEY": "bigint PRIMARY KEY",
      "Real PRIMARY KEY": "double PRIMARY KEY",
      "Boolean PRIMARY KEY": "boolean PRIMARY KEY", # this does not make too much sense, though
      "Timestamp PRIMARY KEY": "timestamp PRIMARY KEY",
    }

  def get_typemap(self):
    return TYPEMAP


class PhoenixDataConn(ExtensionDataConn):
  def __init__(self, params: dict, connection_id: int, conn_type='phoenix'):
    ExtensionDataConn.__init__(
      self, params, connection_id, conn_type=conn_type)

  def properties(self):
    return PhoenixConnProperties()

  def create_connection(self, params: dict, username: str = None):
    host = params.get('HOST')
    port = params.get('PORT')
    conn_username = params.get('USERNAME')
    password = params.get('PASSWORD')
    auth = params.get('AUTH')
    socket_type = params.get('SOCK','')
    http_path = params.get('HS2_HTTP_SQLPATH', '')
    ssl_verify = not params.get('SSLCERT_NAME_MISMATCH', False)
    impersonation = params.get('IMPERSONATION', False)

    protocol = 'http'
    if 'ssl' in socket_type:
      protocol = 'https'

    authentication = None
    if auth == 'plain':
      authentication = 'BASIC'
    elif auth == 'kerberos':
      authentication = 'SPNEGO'
      if impersonation:
        conn_username = username

    database_url = f'{protocol}://{host}'
    if port:
      database_url += f':{port}'
    database_url += http_path

    dbapi2 = self.dbapi2_module()
    connection = dbapi2.connect(
      database_url,
      autocommit=True,
      user=conn_username,
      password=password,
      authentication=authentication,
      verify=ssl_verify)

    return connection

  def dbapi2_module(self):
    # phoenixdb module must be available on PYTHONPATH
    import phoenixdb
    return phoenixdb

  def show_databases(self, build_request):
    build_request.query = 'SELECT distinct TABLE_SCHEM FROM SYSTEM.CATALOG'
    result = self.execute(build_request)

    # replace the "None" schema by the empty string
    rows = json.loads(result['rows'])
    rows = [[''] if row == [None] else row for row in rows]
    result['rows'] = json.dumps(rows)

    return result

  def show_tables(self, build_request, database: str):
    if database:
      build_request.query = f"SELECT distinct TABLE_NAME FROM SYSTEM.CATALOG WHERE TABLE_SCHEM='{database}'"
    else:
      build_request.query = f"SELECT distinct TABLE_NAME FROM SYSTEM.CATALOG WHERE TABLE_SCHEM is NULL"

    return self.execute(build_request)

  # used for creating tables during data import:
  def create_sql_cmds(self, tablename, colnames, coltypes,
        importuser, importsrcfile, external_table,
        location, localesetting='C', type_classes=None, partitionColnames=None, partitionColtypes=None,
        **kwargs):

    # TODO: ask the user for the primary key
    # make the first column the primary key:
    coltypes = [f'{coltypes[0]} PRIMARY KEY', *coltypes[1:]]
    return super().create_sql_cmds(tablename, colnames, coltypes,
        importuser, importsrcfile, external_table,
        location, localesetting, type_classes, partitionColnames, partitionColtypes,
        **kwargs)

  def insert_table_prefix(self, tablename, overwrite=False):
    return f'UPSERT INTO {self.quote_tablename(tablename)}'

  def rename_table_cmd(self, old_table_name, new_table_name):
    # TODO: no rename?
    # TODO: if so, figure out how to delete views and eventually the table behind it
    sqlcmd = f'CREATE VIEW {self.quote_tablename(new_table_name)} AS SELECT * FROM {self.quote_tablename(old_table_name)}'
    return sqlcmd

  def aggregate_detection_query(self, builder: ExpressionValidator, expression: str, dataset: ModelDataset):
    expr_header = builder.expr_header(expression, dataset, 'column', add_expr=True)
    inner_query = expr_header[2][1]
    outer_select: str = expr_header[2][2]
    expression_alias = '"__test_expr__"'
    outer_select_with_alias = re.sub(' FROM $', f' as {expression_alias} FROM ', outer_select)
    return f'{outer_select_with_alias} {inner_query} GROUP BY {expression_alias}'

  def aggregate_detection_error(self):
    return 'Aggregate expressions may not be used in GROUP BY'
