# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import django.test
import json

from unittest import mock
from dataconn.dataconn_dbapi2 import DbApi2Conn
from dataconn.ext import main, base

from dataconn.ext.phoenix.phoenix_ext import PhoenixExtension
from dataconn.ext.phoenix.phoenix import PhoenixDataConn
from dataconn.dataconn_dataengine import BuildRequest
from vizlib.dataset_request import DatasetRequest


class TestPhoenix(django.test.TestCase):

  @django.test.override_settings(DATACONN_EXTENSIONS=[PhoenixExtension()])
  @mock.patch.object(PhoenixDataConn, 'dbapi2_module')
  @mock.patch.object(base, 'BuildRequestExecutor')
  def test_show_databases(self, executor_mock: mock.Mock, dbapi2_module: mock.Mock):
    conn: DbApi2Conn = main.create_connection('phoenix', {}, 10)

    executor_mock.return_value.execute.return_value = {
      'rows': json.dumps([[None], ['SYSTEM']])
    }

    self.assertIsNotNone(conn.execute(BuildRequest(query='.databases')))

    # query execution was performed:
    executor_mock.return_value.execute.assert_called_once()
    # arguments:
    args = executor_mock.return_value.execute.call_args_list[0].kwargs

    expected_query = 'SELECT distinct TABLE_SCHEM FROM SYSTEM.CATALOG'
    self.assertEqual(args['query'], expected_query)

  @django.test.override_settings(DATACONN_EXTENSIONS=[PhoenixExtension()])
  @mock.patch.object(PhoenixDataConn, 'dbapi2_module')
  @mock.patch.object(base, 'BuildRequestExecutor')
  def test_show_tables(self, executor_mock: mock.Mock, dbapi2_module: mock.Mock):
    conn: DbApi2Conn = main.create_connection('phoenix', {}, 10)

    ds_req = {
      'type': '.tables',
      'dbname': 'SYSTEM'
    }
    ds_req_obj = DatasetRequest(json.dumps(ds_req))
    br = BuildRequest(connection=conn, dataset_request=ds_req_obj)
    self.assertIsNotNone(conn.execute(br))

    # query execution was performed:
    executor_mock.return_value.execute.assert_called_once()
    # arguments:
    args = executor_mock.return_value.execute.call_args_list[0].kwargs

    expected_query = "SELECT distinct TABLE_NAME FROM SYSTEM.CATALOG WHERE TABLE_SCHEM='SYSTEM'"
    self.assertEqual(args['query'], expected_query)

  @django.test.override_settings(DATACONN_EXTENSIONS=[PhoenixExtension()])
  @mock.patch.object(PhoenixDataConn, 'dbapi2_module')
  @mock.patch.object(base, 'BuildRequestExecutor')
  def test_show_tables_for_default_namespace(self, executor_mock: mock.Mock, dbapi2_module: mock.Mock):
    conn: DbApi2Conn = main.create_connection('phoenix', {}, 10)

    ds_req = {
      'type': '.tables',
      'dbname': ''
    }
    ds_req_obj = DatasetRequest(json.dumps(ds_req))
    br = BuildRequest(connection=conn, dataset_request=ds_req_obj)
    self.assertIsNotNone(conn.execute(br))

    # query execution was performed:
    executor_mock.return_value.execute.assert_called_once()
    # arguments:
    args = executor_mock.return_value.execute.call_args_list[0].kwargs

    expected_query = "SELECT distinct TABLE_NAME FROM SYSTEM.CATALOG WHERE TABLE_SCHEM is NULL"
    self.assertEqual(args['query'], expected_query)
