# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from dataconn.ext.base import ExtensionConnProperties, ExtensionDataConn
from vizlib.dataset import ModelDataset
from vizlib.validator_builder import ExpressionValidator

import dataconn.arctypes as arctypes
import json
import logging
import re

LOG = logging.getLogger('arcweb.dataconn.ext.trino')

TYPEMAP = {
    'INTEGER': arctypes.ARC_DATATYPE_INT,
    'TINYINT': arctypes.ARC_DATATYPE_INT,
    'SMALLINT': arctypes.ARC_DATATYPE_INT,
    'BIGINT': arctypes.ARC_DATATYPE_BIGINT,

    'REAL': arctypes.ARC_DATATYPE_DOUBLE,
    'DOUBLE': arctypes.ARC_DATATYPE_DOUBLE,
    'DECIMAL': arctypes.ARC_DATATYPE_DOUBLE,

    'CHAR': arctypes.ARC_DATATYPE_STRING,
    'VARCHAR': arctypes.ARC_DATATYPE_STRING,

    'DATE': arctypes.ARC_DATATYPE_TIMESTAMP,
    'TIME': arctypes.ARC_DATATYPE_TIMESTAMP,
    'TIMESTAMP': arctypes.ARC_DATATYPE_TIMESTAMP,

    'BOOLEAN': arctypes.ARC_DATATYPE_BOOLEAN,

    'VARBINARY': arctypes.ARC_DATATYPE_BINARY,
}


class TrinoConnProperties(ExtensionConnProperties):

    def __init__(self):
        super(TrinoConnProperties, self).__init__()
        self.quote = '"'
        self.quote_end = '"'
        self.quote_alias = '"'
        self.quote_escape_character = '"'
        self.quote_string_constant = "'"
        self.supports_partitions = False
        self.supports_async = False
        self.supports_backticks = False
        self.supports_reuse = False
        self.supports_import_data_perm = False
        self.supports_import_data_dbname = 'public'
        self.alias_supports_backticks = False
        self.string_type = "VARCHAR"
        self.date_type = "date"
        self.double_type = "double"
        self.supports_nulls_order = True
        self.supports_dataextract_source = True
        self.supports_dataextract_target = True

        self.display_to_table_type = {
            'String': 'varchar',
            'Integer': 'integer',
            'Real': 'double',
            'Boolean': 'boolean',
            'Timestamp': 'timestamp'
        }

    def cursor_coltype_to_arctype(self, native_type: str):
        # types look like this: timestamp(3) or decimal(10,2)
        typemap_key = native_type.split('(', maxsplit=2)[0].upper()

        return TYPEMAP.get(typemap_key, arctypes.ARC_DATATYPE_STRING)


class TrinoDataConn(ExtensionDataConn):
    def __init__(self, params: dict, connection_id: int, conn_type='trino'):
        ExtensionDataConn.__init__(
            self, params, connection_id, conn_type=conn_type)

    def properties(self):
        return TrinoConnProperties()

    def create_connection(self, params: dict, username: str = None):
        from trino.auth import BasicAuthentication

        host = params.get('HOST')
        port = params.get('PORT')
        user = params.get('USERNAME')
        conn_username = params.get('USERNAME')
        password = params.get('PASSWORD')
        auth = params.get('AUTH')
        socket_type = params.get('SOCK', '')
        # http_path = params.get('HS2_HTTP_SQLPATH', '')
        ssl_verify = not params.get('SSLCERT_NAME_MISMATCH', False)
        impersonation = params.get('IMPERSONATION', False)

        protocol = 'http'
        if 'ssl' in socket_type:
            protocol = 'https'

        authentication = None
        if auth == 'plain':
            authentication = 'BASIC'
        elif auth == 'kerberos':
            authentication = 'SPNEGO'
            if impersonation:
                conn_username = username

        database_url = f'{protocol}://{host}'
        auth = BasicAuthentication(user, password)

        properties = {
            'host': database_url,
            'port': port,
            'user': conn_username,
        }

        if (protocol == 'https'):
            properties['http_scheme'] = 'https'
            properties['auth'] = auth

        dbapi2 = self.dbapi2_module()
        connection = dbapi2.connect(**properties)

        return connection

    def dbapi2_module(self):
        from trino import dbapi
        return dbapi

    def show_databases(self, build_request):
        sql_query = 'SHOW CATALOGS'
        build_request.query = sql_query
        result = self.execute(build_request)

        # replace the "None" schema by the empty string
        rows = json.loads(result['rows'])

        ret_result = []
        for row in rows:
            catalog = row[0]
            if catalog != 'system':
                build_request.query = f'SHOW SCHEMAS FROM {catalog}'
                temp = self.execute(build_request)
                for schema in json.loads(temp['rows']):
                    ret_result.append([f'{catalog}.{schema[0]}'])

        result['rows'] = json.dumps(ret_result)
        return result

    def show_tables(self, build_request, database: str):
        if database:
            build_request.query = f"SHOW TABLES FROM {database}"

        return self.execute(build_request)

    def aggregate_detection_query(self, builder: ExpressionValidator, expression: str, dataset: ModelDataset):
        expr_header = builder.expr_header(
            expression, dataset, 'column', add_expr=True)
        inner_query = expr_header[2][1]
        outer_select: str = expr_header[2][2]
        expression_alias = '"__test_expr__"'
        outer_select_with_alias = re.sub(
            ' FROM $', f' as {expression_alias} FROM ', outer_select)
        return f'{outer_select_with_alias} {inner_query} GROUP BY {expression_alias}'

    def aggregate_detection_error(self):
        return 'Aggregate expressions may not be used in GROUP BY'

    def tablename_parts(self, fq_table_name):
        properties = self.properties()
        db_name = fq_table_name.rsplit('.', 1)[0]
        table_name = fq_table_name.rsplit('.', 1)[1]
        return (db_name.lstrip(properties.quote).rstrip(properties.quote_end),
                table_name.lstrip(properties.quote).rstrip(properties.quote_end))

    def apply_types(self, value, type: str):
        try:
            if value is None:
                value = "NULL"
            elif type == "String":
                value = re.sub("'", r"''", value)
                value = "'%s'" % value
            elif type == "Timestamp":
                value = "timestamp '%s'" % value
            elif type == "Real":
                if type(value) == str: # no idea if this can happen.
                    value = re.sub(",", ".", value)
        except Exception as e:
            pass

        return value
