# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

"""
Provide a class to connect to a HS2 server and communicate with it

Provide both sync and async access
Use multiple connections
Handles failed connection--On error, resets the connection

Client specifies how many connections, requests per connection, total active requests.

Currenty no cursor type support--we always fetch full result sets (and put a limit
on the size size the viz can not deal with too much data anyway)

Request flow:
  Created through async_request() (via _newreq()), added to global request list

  Removed when:
    error when issued though async_request()
    async_status() returns error
    as part of async_results() completion
    upon async_cancel()

  Background thread to clean up requests which are abandoned by the front end

  Request accounting is strictly through _issue_requests routine.

Notes on impersonation:

  For impersonation, the current 'arcviz' username (not the username/password used
  to initiate the connection) is passed to the target for each request.

  Note that this requires the 'arcviz' user to be provided in each sync and async
  request.

  The target will use this name for authorization checks, like via the 'sentry'
  system.

  In impersonation mode sessions can not be shared among arcviz users.  Since we
  currently have a 1-1 mapping of connections to sessions this mode will create
  many more connections as well.  This has some impacts:

    We might consider forcing the connection timeout logic active so that we
    don't leave around unused connections.  Currently when we require a
    connection for a new user, we will force recycle an inactive connection.

    We must send each request on a particular connection.  This makes the 
    queueing system harder:

      We can be below the total command limit but not have any available
      connections, so we need to wait for a connection to become available
      by having it's command count go to zero.

      When a command finishes, we need to check waiting requests against
      all available connections rather than just the one on which the
      command just finished.

      We can get starvation if all current connections are kept busy.  To
      fix this we need to prevent issue of new commands if we have
      a command waiting for an available connection over some threshold.

TODO:
  There are two places where completed requests can occupy a slot which could
  prevent other requests being issued:
    1) If the client delays asking for results from a completed request
    2) If queued invalid request is sent via _issue_requests(),
         we wait until the client to request status
  Possible solutions:
    For (1), we could make the polling process actually monitor status and pull the
    results, allowing the completed request to be closed out from the db.
    For (2), similarly we leave around the status but remove the request from the
    active count.

"""

import time
import logging
import threading
import re
import codecs

from distutils.version import LooseVersion

from .dataconn_consts import DCS
from .connections_types import DATACONN_HIVE, DATACONN_IMPALA, DATACONN_ARCENGINE, \
    DATACONN_SPARKSQL, DATACONN_IMPYLA, DATACONN_ARCENGINE_V2, isImpalaBased, isArcEngineBased
from django.conf import settings

import vizlib.querylog
import vizlib.lrt
from vizlib.arcprof import ArcProf
from .rows_data import RowsData

# pylint: disable=W0401,W0614
from .thrift_hs2.TCLIService import TCLIService
from .thrift_hs2.TCLIService.ttypes import *
from .thrift_hs2.TCLIService.ExtendedMetadata.ttypes import *

# pylint: enable=W0401,W0614

from .thriftconn import thrift_protocol

LOGGER = logging.getLogger("arcweb")
LOGGER_MSG_LIMIT = 10000

from util.perfcounters import APPROX_COUNTERS

hs2_poolcreate = APPROX_COUNTERS.get('hs2.poolcreate')
hs2_conncreate = APPROX_COUNTERS.get('hs2.conncreate')
hs2_opensession = APPROX_COUNTERS.get('hs2.opensession')
hs2_queries_submit = APPROX_COUNTERS.get('hs2.queries_submit')
hs2_queries_retry = APPROX_COUNTERS.get('hs2.queries_retry')

class DCStatus(object):
  def __init__(self, status=DCS.ERROR, message='', progress=0):
    self.status = status
    self.message = message
    self.progress = progress

# True if arcengine version string indicates new extended metadata functionality supported
def supports_extmeta(name, ver):
  if name != 'Arcengine':
    return False
  if ver.startswith('arcengine'):
    ver = ver[len('arcengine'):]
  return LooseVersion(ver) >= LooseVersion('2.9.2')

class ServerErrorException(Exception):
  pass

class ConnectionErrorException(Exception):
  pass

def tstatus_pass(s):
  """ Return True if TStatus indicates success """
  return (s.statusCode == TStatusCode.SUCCESS_STATUS or
          s.statusCode == TStatusCode.SUCCESS_WITH_INFO_STATUS)

def tstatus_parse(s):
  """
    Parse the status, info messages, and error info from the TStatus structure
    Returns (pass/fail, info text, error text) tuple
  """

  success = True
  inftxt = []
  errtxt = None

  if s.infoMessages:
    inftxt = s.infoMessages

  success = tstatus_pass(s)
  if not success:
    errtxt = ("(%s, %s) %s" % (s.sqlState, s.errorCode, s.errorMessage)).strip('\n')

    # The error can contain the query, so hide it if configured to do so
    if getattr(settings, 'HIDE_PROFILE_QUERY', False):
      if errtxt.startswith('(HY000, None) AnalysisException:'):
        errtxt = '(HY000, None) AnalysisException: ***'

  return (success, inftxt, errtxt)

def exception_on_connection_failure(req, rsp):
  """
    Raise exception on an error that indicates a failure at
    the connection/session level so that we can close and
    re-establish the connection.

    Currently we only check for the impala timeout response
  """
  if rsp.__class__.__name__ in ["TExecuteStatementResp",
                                "TGetAnalyticalViewsRecommendationsResp"] and \
     rsp.status.statusCode != TStatusCode.SUCCESS_STATUS and \
     rsp.status.errorMessage is not None and \
     (rsp.status.errorMessage.startswith("Client session expired due to more than") or \
      rsp.status.errorMessage.startswith("TSocket read 0 bytes") or \
      rsp.status.errorMessage.startswith("Invalid SessionHandle:")):
    raise ConnectionErrorException(rsp.status.errorMessage)

def exception_on_error(obj, rsp):
  """Raise an exception on error"""

  if not rsp:
    LOGGER.info("Exception. No response received")
    obj.log("Exception. No response received")
    raise Exception("Connection error")
  success, inftxt, errtxt = tstatus_parse(rsp.status)
  if not success:
    LOGGER.info("Server Error. Status: %s", errtxt)
    obj.log("Server Error. Status: %s" % errtxt)
    raise ServerErrorException(errtxt)

  return (success, inftxt, errtxt)

def get_coltype_name(ttype):
  """get a coltype string from the coltype"""
  if ttype.mapEntry:
    return 'MAP'

  if ttype.unionEntry:
    return 'UNION'
  if ttype.arrayEntry:
    return 'ARRAY'

  if ttype.userDefinedTypeEntry:
    return 'UDT'

  if ttype.structEntry:
    return 'STRUCT'

  if ttype.primitiveEntry:
    typename = TTypeId._VALUES_TO_NAMES.get(ttype.primitiveEntry.type, 'unknown')
    if typename.endswith("_TYPE"):
      return typename[:-5]
    else:
      return typename

  return "unknown"

# Table indexed by TTypeDesc primitive type to function which extracts the
# corresponding element from a TColumn or TColumnValue element.

TTYPETOELEMENT = {
  TTypeId.BOOLEAN_TYPE:   ((lambda cv: cv.boolVal), str),
  TTypeId.BINARY_TYPE:    ((lambda cv: cv.binaryVal), str),

  TTypeId.TINYINT_TYPE:   ((lambda cv: cv.byteVal), str),
  TTypeId.SMALLINT_TYPE:  ((lambda cv: cv.i16Val), str),
  TTypeId.INT_TYPE:       ((lambda cv: cv.i32Val), str),
  TTypeId.BIGINT_TYPE:    ((lambda cv: cv.i64Val), str),

  TTypeId.FLOAT_TYPE:     ((lambda cv: cv.doubleVal), str),
  TTypeId.DOUBLE_TYPE:    ((lambda cv: cv.doubleVal), str),

  TTypeId.TIMESTAMP_TYPE: ((lambda cv: cv.stringVal), (lambda x: x)),
  TTypeId.DATE_TYPE:      ((lambda cv: cv.stringVal), (lambda x: x)),

  TTypeId.STRING_TYPE:    ((lambda cv: cv.stringVal), (lambda x: x)),
  TTypeId.VARCHAR_TYPE:   ((lambda cv: cv.stringVal), (lambda x: x)),
  TTypeId.CHAR_TYPE:      ((lambda cv: cv.stringVal), (lambda x: x)),

  TTypeId.DECIMAL_TYPE:   ((lambda cv: cv.stringVal), (lambda x: x)),

  TTypeId.ARRAY_TYPE:     ((lambda cv: cv.stringVal), (lambda x: x)),
  TTypeId.MAP_TYPE:       ((lambda cv: cv.stringVal), (lambda x: x)),
  TTypeId.STRUCT_TYPE:    ((lambda cv: cv.stringVal), (lambda x: x)),
  TTypeId.UNION_TYPE:     ((lambda cv: cv.stringVal), (lambda x: x)),
}

# ----------------------------------------------------------------------------------------

class HS2Request(object):
  """Represent one HS2 request"""
  def __init__(self, query, conf_settings, pool, conn, qhandle, doasuser,
      logQuery=True, qcntl=None):
    self.query = query
    self.comment = ''
    self.rewrite = ''
    self.conf_settings = conf_settings
    self.pool = pool
    self.conn = conn
    self.qhandle = qhandle
    self.ophandle = None
    self.doasuser = doasuser
    self.active = False
    self.poll_time = time.time()
    self.status = DCS.QUEUED
    self.message = ''
    self.sequence = 0
    self.logQuery = logQuery
    self.qcntl = qcntl
    self.extresp = None
    self.webserver_url = None
    self.computeActionType(query)

    # Allow translation of doasuser (assumed to be arcviz username) through user profile key
    if doasuser:
      profkey = getattr(settings, 'PROFILE_IMPERSONATION_KEY', None)
      if not profkey and getattr(settings, 'PROFILE_PROXY_USERNAME', False):
        profkey = 'proxy_username'
      if profkey:
        from apps.models import profile_getkey_for_username
        self.doasuser = profile_getkey_for_username(doasuser, profkey, doasuser)

    self.querylog = vizlib.querylog.actionlog_create(
        query=query,
        username=self.doasuser,
        action=self.actionType,
        dataconnection_id=self.pool.connid(),
        dataconnection_type=self.pool.conntype(),
        sync=False,
        component_type=vizlib.querylog.ComponentType.DATACONNECTION,
        component_id=self.pool.connid())

    if self.isLongRunningTask:
      self.lrt = vizlib.lrt.LongRunningTask(
          tasktype=vizlib.lrt.TaskType.REFRESH_AVIEW,
          taskname=query.split()[-1],
          taskid=qhandle,
          connectionid=self.pool.connid(),
          cancelcb=self.cancelRequest)
    else:
      self.lrt = None

  def cancelRequest(self):
    self.pool.async_cancel(self.qhandle)

  def computeActionType(self, query):
    self.isLongRunningTask = False
    if not query:
      self.actionType = vizlib.querylog.ActionType.UNKNOWN
      return

    # Queries have comments. Get to the start of the query
    m = re.match(r"^\s*\/\*.+?\*\/\s*\n(.+)$", query, re.MULTILINE)
    if not m:
      # No comments - use the base query
      q = query.upper()
    else:
      q = m.group(1).upper()

    if q.startswith('EXPLAIN'):
      self.actionType = vizlib.querylog.ActionType.READ
      return

    if q.startswith('DESCRIBE'):
      self.actionType = vizlib.querylog.ActionType.READ
      return

    if q.startswith('SHOW'):
      self.actionType = vizlib.querylog.ActionType.READ
      return

    if q.startswith('SELECT'):
      self.actionType = vizlib.querylog.ActionType.READ
      return

    if q.startswith('DROP'):
      self.actionType = vizlib.querylog.ActionType.DELETE
      return

    if q.startswith('CREATE'):
      self.actionType = vizlib.querylog.ActionType.CREATE
      return

    if q.startswith('REFRESH'):
      self.actionType = vizlib.querylog.ActionType.UPDATE
      self.isLongRunningTask = True
      return

    self.actionType = vizlib.querylog.ActionType.UNKNOWN

  def logQueryQueued(self):
    if self.logQuery:
      self.querylog.logQueryQueued()

  def logQueryStart(self):
    if self.logQuery:
      self.querylog.logQueryStart()

  def logQueryEnd(self, **kwargs):
    if self.logQuery:
      self.querylog.logQueryEnd(**kwargs)
    if self.lrt:
      self.lrt.done()

  def log(self, msg):
    if self.conn is None:
      idstr = "x_%s" % self.qhandle
    else:
      idstr = "%s_%s" % (self.conn.idstr, self.qhandle)
    if self.doasuser:
      idstr += " (%s) " % self.doasuser
    self.pool.log(msg, idstr=idstr)

class HS2Client(object):
  """represent one connection"""
  def __init__(self, pool, id):
    self.pool = pool
    self.id = id
    self.client = None
    self.rsttext = None
    self.session = None
    self.needrst = False
    self.protocol = None
    self.doasuser = None
    self.rsttext = None
    self.idstr = "%s" % id
    self.lock = threading.Lock()
    self.reqcnt = 0
    self.idle_time = time.time()

  def log(self, msg):
    idstr = self.idstr
    if self.doasuser:
      idstr += " (%s) " % self.doasuser
    self.pool.log(msg, idstr=idstr)

INSTANCE = 0
INSTLOCK = threading.Lock()

class HS2Pool(object):
  """Pool of client side connections for HS2"""

  def _poller(self):
    """polling thread to clean up abandonded requests"""
    while self._polling:
      try:
        time.sleep(1)
        now = time.time()
        for key in list(self._requests.keys()):
          req = self._requests.get(key)
          if req and not req.active and (now - req.poll_time > self._timeout):
            req.log("Timeout request due to client inactivity")
            self.async_cancel(req.qhandle)

        # Close idle connections
        if self._conn_timeout > 0:
          with self._lock:
            for conn in self._connections + [self._adminconn]:
              if conn.protocol and (conn.reqcnt == 0) and (now - conn.idle_time > self._conn_timeout):
                conn.log("Close idle connection")
                self._clsconn(conn)

      except Exception as e:
        self.log("Exception during polling: %s" % str(e))

  def log(self, msg, idstr=None):
    LOGGER.info("HS2 %s %s" %
                 (self.idstr if idstr is None else ("%s_%s" % (self.idstr, idstr)),
                  msg[:LOGGER_MSG_LIMIT]))

  def __init__(self, host=None, port=None, auth=None,
                     username='', password='',
                     ssl=False, sslcert=None, kerbserv=None,
                     conntype=None,
                     connid=None,
                     settings=None,
                     timeout=60, max_connections=1, max_reqperconn=1, max_requests=1,
                     conn_timeout=0, sslcert_name_mismatch=False, pure_sasl=False,
                     http_service=None, http_access_token=None,
                     socket_timeout=0,
                     impersonation_trusted=None):

    global INSTANCE
    with INSTLOCK:
      INSTANCE += 1
      self._instance = INSTANCE

    self._host = host
    self._port = port
    self._username = username
    self._password = password
    self._settings = settings
    self._auth = auth
    self._ssl = ssl
    self._sslcert = sslcert
    self._sslcert_name_mismatch = sslcert_name_mismatch
    self._kerbserv = kerbserv
    self._conntype = conntype
    self._connid = connid
    self._timeout = timeout
    self._conn_timeout = conn_timeout
    self._socket_timeout = None if not socket_timeout else socket_timeout
    self._impersonation_trusted = impersonation_trusted
    self._max_connections = max_connections
    self._max_requests = max_requests
    self._max_reqperconn = max_reqperconn
    self._srv_name = ''
    self._srv_ver = ''
    self._extmeta = None
    self._queued = []
    self._requests = dict()
    self._reqcnt = 0
    self._lock = threading.Lock()
    self._handle = int(time.time())
    self._polling = False
    self._pollth = None
    self._sequence = 0
    self._pure_sasl = pure_sasl
    self._http_service = http_service
    self._http_access_token = http_access_token

    self.idstr = "%s:%s_%s" % (host, port, self._instance)
    # List of connections
    self._connections = [HS2Client(self, i) for i in range(0, self._max_connections)]

    self._adminconn = HS2Client(self, 'A')

    self.log("HS2Pool Created with %d connections for %d requests and limit of %d request per connection" %
             (self._max_connections, self._max_requests, self._max_reqperconn))

    global hs2_poolcreate
    hs2_poolcreate += 1

  def __del__(self):
    self.close()

  def close(self):
    """clean up"""
    for conn in self._connections:
      self._clsconn(conn)
    if self._polling:
      self._polling = False
      if self._pollth is not None:
        self._pollth.join()

  def srv_name(self):
    # Should only be not set when we've never connected
    if not self._srv_name:
      with self._lock:
        self._rstconn(self._adminconn)
    return self._srv_name or 'Unknown'

  def srv_version(self):
    # Should only be not set when we've never connected
    if not self._srv_ver:
      with self._lock:
        self._rstconn(self._adminconn)
    return self._srv_ver or 'Unknown'

  # Format a list of TCacheInvalidationObjects into text
  def _format_tablesigs(self, objects):
    return 'Tables: %s' % \
             ','.join('%s.%s:%s+%s' % \
               (t.databaseName, t.tableName, t.crc or 0, t.tldt or 0) for t in objects)

  # Parse the above formatted text back into a list of TCacheInvalidationObjects
  def _parse_tablesigs(self, text):
    cinv = []
    if text and text.startswith('Tables: '):
      for tt in text[len('Tables: '):].split(','):
        db_tb, crc_tldt = tt.split(':')
        db, tb = db_tb.split('.')
        crc, tldt = crc_tldt.split('+')
        if crc == '0':
          crc = None
        cinv.append(TCacheInvalidationObject(databaseName=db, tableName=tb, tldt=tldt, crc=crc))
    return cinv

  def _clsconn(self, conn):
    """close a connection"""
    conn.log("Closing connection")

    # First set all requests associated with this connection to RESET
    for hnd in self._requests:
      req = self._requests.get(hnd)
      if req and req.conn == conn and req.status in [DCS.RUNNING, DCS.FINISHED]:
        req.status = DCS.ERROR
        req.message = 'Connection reset'

    if conn.protocol:
      try:
        conn.protocol.trans.close()
      except Exception as e:
        pass
      finally:
        conn.protocol = None
        conn.doasuser = None

  def _rstconn(self, conn):
    """Reset a connection"""

    conn.needrst = False
    conn.rsttext = None
    conn.log("Reset connection")
    # Make sure the polling thread is active
    if not self._polling and self._timeout and not self._pollth:
      self._polling = True
      self._pollth = threading.Thread(target=self._poller)
      self._pollth.daemon = True
      self._pollth.start()

    # Helper function to do one info request
    def inforeq(infotype):
      rsp = conn.client.GetInfo(TGetInfoReq(sessionHandle=conn.session,infoType=infotype))
      exception_on_error(conn, rsp)
      return rsp.infoValue

    # Now close existing connection if needed
    doasuser = conn.doasuser
    self._clsconn(conn)

    # Reconnect
    protocol = None
    try:
      login_username = self._username
      if self._impersonation_trusted and doasuser:
        login_username = doasuser

      conn.log("Connecting to %s" % self._host)
      global hs2_conncreate
      hs2_conncreate += 1

      protocol = thrift_protocol(host=self._host, port=self._port, auth=self._auth,
                                 username=self._username, password=self._password,
                                 ssl=self._ssl, sslcert=self._sslcert,
                                 kerbserv=self._kerbserv,
                                 sslcert_name_mismatch=self._sslcert_name_mismatch,
                                 pure_sasl=self._pure_sasl,
                                 http_service=self._http_service,
                                 http_access_token=self._http_access_token,
                                 normal_timeout=self._socket_timeout)

      client = TCLIService.Client(protocol)

      # If doasuser set, need to add it into the connection settings
      config = self._settings
      _username = login_username
      if (not self._impersonation_trusted) and doasuser:
        config = {} if not self._settings else self._settings.copy()
        if isImpalaBased(self._conntype):
          config['impala.doas.user'] = doasuser
        else:
          config['hive.server2.proxy.user'] = doasuser

      self.log("Opening session with login username %s, username %s, doasuser %s" % \
          (login_username, _username, doasuser))

      global hs2_opensession
      hs2_opensession += 1

      req = TOpenSessionReq(
        client_protocol = TProtocolVersion.HIVE_CLI_SERVICE_PROTOCOL_V7,
        username=login_username,
        password=self._password,
        configuration=config)
      rsp = client.OpenSession(req)
      exception_on_error(conn, rsp)
      conn.client = client
      conn.session = rsp.sessionHandle

      # Grab the server version
      self._srv_name = inforeq(TGetInfoType.CLI_DBMS_NAME).stringValue
      self._srv_ver = inforeq(TGetInfoType.CLI_DBMS_VER).stringValue

      # Check for new style arcengine extended metadata support
      extmeta = None
      if isArcEngineBased(self._conntype) and supports_extmeta(self._srv_name, self._srv_ver):
        extmeta = 'New style'

      # Test if this arcengine supports the extended metadata function
      if isArcEngineBased(self._conntype) and extmeta is None:
        try:
          rsp = conn.client.GetResultSetExtendedMetadata(
            TGetResultSetExtendedMetadataReq(operationHandle=
              TOperationHandle(hasResultSet=True,
                               modifiedRowCount=None,
                               operationType=0,
                               operationId=THandleIdentifier(secret=b'1' * 16,
                                                             guid=b'1' * 16))))
          extmeta = 'Old style'
        except Exception as e:
          if not str(e).startswith('Invalid method name:'):
            raise

      # Update extended metadata support method
      self._extmeta = extmeta

      # This marks the connection has active
      conn.doasuser = doasuser
      conn.protocol = protocol
      conn.idle_time = time.time()
      conn.log("Connected to: %s %s" % (self._srv_name, self._srv_ver))
    except Exception as e:
      LOGGER.exception("Exception on connection reset")
      conn.rsttext = str(e)
      conn.log("Exception on connection reset: %s" % str(e))
      if protocol:
        protocol.trans.close()

  def _onethrift(self, req, fn, arg):
    """Issue one thrift call for a request.  Reset connection on error"""
    rsp = None
    conn = req.conn
    error = None
    try:
      req.active = True
      conn.lock.acquire()
      if conn.protocol:
        rsp = fn(arg)
        exception_on_connection_failure(req, rsp)
    except Exception as e:
      # On error, try to close the connection if it's there
      error = str(e)
      if type(e) == ConnectionErrorException:
        conn.log("Exception on thrift op: %s" % error)
      else:
        LOGGER.exception("Exception on thrift op")


      # Indicate no response on exception
      rsp = None

      if conn.protocol:
        try:
          conn.protocol.trans.close()
        except Exception:
          pass
        conn.protocol = None
    finally:
      req.poll_time = time.time()
      req.active = False
      conn.lock.release()

    # Done now if no error
    if rsp:
      return rsp

    # On error, reset connection under global lock, then raise error
    try:
      self._lock.acquire()
      if conn.protocol is None:
        self._rstconn(conn)
    except Exception:
      pass
    finally:
      self._lock.release()

    raise Exception(("Connection reset: " + error) if error else "Connection reset")

  def conntype(self):
    return self._conntype

  def connid(self):
    return self._connid

  # Fill in the arcengine specific extended metadata portion of the request
  def _build_extmetadatareq(self, req):
    thext = TExtMetadataReq()
    qcntl = req.qcntl

    fa_list = []
    if qcntl.fa_expr and qcntl.fa_tables is not None:
      fa_list = [TFilterAssociation(filterExpr=qcntl.fa_expr, tableNames=qcntl.fa_tables)]

    thext.filterAssociations = fa_list

    thext.fallbackEnabled = qcntl.fallback_dcid is not None
    thext.systemFunctionsAllowed = qcntl.system_functions_allowed

    sinj = []
    for tn in qcntl.ds_tables:
      (dbname, tbname) = tn.split('.')
      sinj.append(TSQLInjectionObject(databaseName=dbname, tableName=tbname))
    thext.sqlInjectionObjectsInput = sinj
    thext.performSQLInjectionChecks = (len(sinj) > 0)

    thext.cacheInvalidationObjectsInput = self._parse_tablesigs(qcntl.tables_text)
    return thext

  def _newreq(self, query, settings, doasuser, logQuery=True, hideQuery=False, qcntl=None):
    """Allocate new request and try to issue via a connection"""

    with self._lock:
      h = self._handle
      self._handle += 1

    req = HS2Request(query, settings, self, None, "%s_%s" % (self._instance, h), doasuser,
        logQuery=logQuery, qcntl=qcntl)
    self._lock.acquire()
    self._requests[req.qhandle] = req
    self._queued.append(req.qhandle)
    self._issue_requests(newreq=req, hideQuery=hideQuery, locked=True)
    return req

  def _getreq(self, qhandle, ftxt):
    req = self._requests.get(qhandle)
    if not req:
      self.log("Handle not found for %s" % ftxt, idstr=qhandle)
      raise Exception("Invalid query handle %s" % qhandle)
    req.poll_time = time.time()
    req.log(ftxt)
    return req

  def _sndreq(self, req):
    # Request activation -- send on its assigned connection to the DB
    tries = 0
    limit = 2
    rsp = None

    # Track order in which requests are emitted for test purposes
    self._sequence += 1
    req.sequence = self._sequence

    # Do one retry on connection reset when issuing command
    while tries <= limit and not rsp:
      global hs2_queries_submit
      hs2_queries_submit += 1

      if tries > 0:
        global hs2_queries_retry
        hs2_queries_retry += 1

      try:
        tries += 1

        threq = TExecuteStatementReq(sessionHandle=req.conn.session,
                                     statement=req.query, confOverlay=req.conf_settings,
                                     runAsync=True)
        # Allow auth check on logical views to follow actual base table
        if isArcEngineBased(self._conntype):
          if re.match(r'^\s*explain\s+', req.query, re.I):
            threq.byPassAuth = True
          if req.qcntl and self._extmeta == 'New style':
            threq.extMetadataReq = self._build_extmetadatareq(req)
        rsp = self._onethrift(req, req.conn.client.ExecuteStatement, threq)
        req.extresp = getattr(rsp, 'extMetadataResp', None)
        if req.extresp and req.qcntl:
          req.qcntl.exec_method = TQueryExecMethod._VALUES_TO_NAMES[req.extresp.queryExecMethod]
          req.qcntl.tables_text = self._format_tablesigs(req.extresp.cacheInvalidationObjects)

        exception_on_error(req, rsp)
        req.status = DCS.RUNNING
        req.ophandle = rsp.operationHandle
        req.webserver_url = getattr(rsp, 'webserverUrl', None)
      except Exception as e:
        self.log("Error running query: %s" % str(e))
        if type(e) not in [ConnectionErrorException, ServerErrorException]:
          LOGGER.exception('Exception on command issue')
        if tries >= limit or not self._retryableException(e):
          req.status = DCS.ERROR
          req.message = str(e)
          break
        time.sleep(1)

  def _retryableException(self, e):
    s = str(e).lower()
    if s.find('connection error') >= 0:
      return True

    if s.find('connection reset') >= 0:
      return True

    if s.find('gateway_timeout') >= 0:
      return True

    if s.find('error 504') >= 0:
      return True

    return False

  def _delreq(self, req):
    # Close out the operation
    try:
      req.status = DCS.CLOSED
      req.logQueryEnd(rewrite=req.rewrite)
      req.log("Delete request")
      del self._requests[req.qhandle]
      if req.ophandle:
        self._onethrift(req, req.conn.client.CloseOperation,
                        TCloseOperationReq(req.ophandle))
        req.ophandle = None
      req.log("Deleted")
    except Exception as e:
      pass

    # Send possible next requests if this one was holding a slot
    if req.conn is not None and req.conn != self._adminconn:
      self._issue_requests(delreq=req)

  def _issue_requests(self, newreq=None, delreq=None, hideQuery=False, locked=False):
    """Issue as many waiting requests as we can, special logging for newreq"""

    # Process the request lists/counts under global lock
    index = 0
    lost = []
    issue = []

    if not locked:
      self._lock.acquire()

    # Now that we have global lock, we can do the accounting associated
    # with the deleted request if present and it was holding a slot
    if delreq and delreq.conn:
      conn = delreq.conn
      delreq.conn = None
      self._reqcnt -= 1
      conn.reqcnt -= 1
      if conn.reqcnt == 0:
        conn.idle_time = time.time()

    try:
      while self._reqcnt < self._max_requests and index < len(self._queued):
        handle = self._queued[index]
        index += 1
        req = self._requests.get(handle)
        if not req:
          lost.append(handle)
          continue

        # If not doing impersonation, just look for available connection
        # Else
        #   If we have a matching connection, can only consider using that one
        #   Else look for first an unused then an idle connection

        connlist = self._connections
        if not req.doasuser:
          self.log("Running query without impersonation")
          conn = next((c for c in connlist if c.reqcnt < self._max_reqperconn), None)
        else:
          self.log(f"Running query as {req.doasuser}")
          conn = next((c for c in connlist if c.doasuser == req.doasuser), None)
          if conn:
            conn = conn if conn.reqcnt < self._max_reqperconn else None
          else:
            conn = next((c for c in connlist if c.protocol == None), None)
            if not conn:
              conn = next((c for c in connlist if c.reqcnt == 0), None)

        # If we found a connection for this request, set up for issue
        if conn:
          req.conn = conn
          if conn.protocol is not None and conn.doasuser != req.doasuser:
            conn.needrst = True
          conn.doasuser = req.doasuser
          conn.reqcnt += 1
          self._reqcnt += 1
          conn.log("Found connection. Request count: " + \
              f"{conn.reqcnt} / {self._max_reqperconn}. Concurrency: {self._reqcnt}")
          issue.append(req)

      # We have a list of lost handles and a list of requests to issue
      for handle in lost:
        self._queued.remove(handle)
      for req in issue:
        self._queued.remove(req.qhandle)

    finally:
      self._lock.release()

    # Now that all counts/lists are consistent, do the blocking code
    for handle in lost:
      self.log("Request not found for issue queued request", idstr=handle)

    newreq_issued = False
    for req in issue:
      req.logQueryStart()
      if req == newreq:
        newreq_issued = True
        if hideQuery:
          req.log("Async issue: ***")
        else:
          req.log("Async issue: %s" % req.query)
      else:
        req.log("Issue queued request")
        # TODO: Currently, we don't support the cache validated / fallback responses on
        #       the queued query path, so we disable those functions here.
        if req.qcntl:
            req.qcntl.tables_text = None
            req.qcntl.fallback_dcid = None

      # Do reset under global lock
      conn = req.conn
      self._lock.acquire()
      try:
        if not conn.protocol or conn.needrst:
          self._rstconn(conn)
      except Exception as e:
        conn.log("Exception on reset: %s" % str(e))
      finally:
        self._lock.release()

      if conn.protocol:
        self._sndreq(req)
      else:
        req.message = conn.rsttext or "Unable to connect to database"
        req.status = DCS.ERROR

    if newreq and not newreq_issued:
      newreq.logQueryQueued()

  def hs2call(self, hs2name, hs2arg, getres, doasuser, qcntl=None):
    """ Issue a single hs2 request via the private admin connection """
    # TODO -- handle case where impersonation is set and admin connection
    #         is used by two different users simultaniously.  Window is
    #         pretty small, so we leave it for now.
    req = None
    try:
      self._lock.acquire()
      req = HS2Request(None, None, self, None, "%s_%s" % (self._instance, self._handle), doasuser, qcntl=qcntl)
      self._handle += 1
      req.conn = self._adminconn
      req.status = DCS.RUNNING
      if not req.conn.protocol or (req.conn.doasuser != req.doasuser):
        req.conn.doasuser = req.doasuser
        self._rstconn(req.conn)
    except Exception as e:
      LOGGER.exception(str(e))
    finally:
      self._lock.release()

    if not req.conn.protocol:
      raise Exception('Unable to connect to database')

    # Mark the connection as in use
    req.conn.idle_time = time.time()

    tries = 0
    limit = 2
    rsp = None
    while tries <= limit and not rsp:
      try:
        tries += 1
        hs2arg.sessionHandle = req.conn.session
        hs2fn = getattr(req.conn.client, hs2name)
        if not hs2fn:
          raise Exception('Unknown function name: %s' % hs2name)

        req.log("Issue HS2Request %s" % hs2name)
        rsp = self._onethrift(req, hs2fn, hs2arg)
        req.extresp = getattr(rsp, 'extMetadataResp', None)
        req.status = DCS.RUNNING
        if getres:
          exception_on_error(req, rsp)
          req.ophandle = rsp.operationHandle
      except Exception as e:
        if (str(e) != 'Connection error' and not str(e).startswith('Connection reset')) or tries >= limit:
          raise
        time.sleep(1)

    if getres and rsp.operationHandle:
      self._requests[req.qhandle] = req
      return self.async_results(req.qhandle)
    else:
      self._delreq(req)
      return rsp
 
  def run(self, statement, settings=None, doasuser=None, qcntl=None):
    """run a single command, returning a result tuple"""
    self.log("Sync issue: %s" % statement)
    hnd = self.async_issue(statement, settings, doasuser, qcntl=qcntl)
    res = self.async_results(hnd)
    return res


  def async_issue(self, statement, settings=None, comment='', doasuser=None,
      logQuery=True, hideQuery=False, qcntl=None):
    """run a single command, returning a handle string to use for getting results"""

    if comment:
      statement = comment + '\n' + statement

    req = self._newreq(statement, settings=settings, doasuser=doasuser,
        logQuery=logQuery, hideQuery=hideQuery, qcntl=qcntl)
    req.comment = comment
    if not req.conn or req.status == DCS.QUEUED:
      req.log("Async queue: %s" % statement)
      return req.qhandle

    if req.status != DCS.RUNNING:
      if not req.message:
        req.message = "Missing message on status: %s" % req.status
        req.log(req.message)
      self._delreq(req)
      raise Exception(req.message)

    req.log("Issued")
    return req.qhandle

  def async_cancel(self, qhandle):
    """cancel a previously issued command"""
    req = self._getreq(qhandle, "Async cancel")
    if req.status == DCS.RUNNING:
      # TODO: A hack here...hive doesn't cancel correctly
      if self._max_reqperconn == 1 and \
         not isImpalaBased(self._conntype):
        self._rstconn(req.conn)
      else:
        self._onethrift(req, req.conn.client.CancelOperation, TCancelOperationReq(req.ophandle))
      req.status == DCS.ERROR
      req.message = 'Request cancelled'
      req.log("Cancelled")
    self._delreq(req)

  def async_status(self, qhandle, username=None, profile=None):
    """return status for previous issued command"""

    profile = profile or ArcProf('hs2.status', username=username)

    req = self._getreq(qhandle, "Async status")
    profile.mark('async_status.enter')

    # No polling needed if request is not active
    if req.status != DCS.RUNNING:
      req.log("Async status: %s (%s)" % (req.status, req.message))
      profile.mark('async_status.notrunning')
      if req.status == DCS.ERROR:
        self._delreq(req)
        profile.mark('async_status.delreq')
      return DCStatus(req.status, req.message)

    rsp = self._onethrift(req, req.conn.client.GetOperationStatus,
                               TGetOperationStatusReq(req.ophandle))
    profile.mark('async_status.getoperationstatus')
    if rsp:
      waiting = rsp.status and rsp.status.statusCode == TStatusCode.SUCCESS_STATUS and \
                rsp.operationState and rsp.operationState == TOperationState.RUNNING_STATE

      success, inftxt, errtxt = tstatus_parse(rsp.status)
      req.message = '' if not inftxt else "\n".join(inftxt)
      if waiting:
        req.status = DCS.RUNNING
        # For non-hive, we can get progress from the GetLog command
        if self._srv_name.find('Hive') == -1 and self._srv_name.find('Spark') == -1:
          logrsp = self._onethrift(req, req.conn.client.GetLog,
                                     TGetLogReq(req.ophandle))
          exception_on_error(req, logrsp)
          req.message = str(logrsp.log).strip('\n')
      elif success:
        req.status = DCS.FINISHED
      else:
        req.status = DCS.ERROR
        req.message = errtxt

    if req.message:
      req.log("Async status: %s (%s)" % (req.status, req.message))
    else:
      req.log("Async status: %s" % req.status)
    status = DCStatus(req.status, req.message)
    if status.status not in [DCS.RUNNING, DCS.FINISHED]:
      self._delreq(req)
    return status

  def async_results(self, qhandle, profile=None):

    out = []
    for out in self.async_results_stream(qhandle, stream=False, profile=profile):
      pass
    return out

  #
  # Because we use this function in streaming and non-streaming mode,
  # we use yield to return data for both cases (python doesn't let you
  # mix the two)
  #
  def async_results_stream(self, qhandle, stream=False, profile=None):

    profile = profile or ArcProf('hs2.results')

    req = self._getreq(qhandle, "Async results")
    profile.mark('async_results.enter')

    # Ensure request is finished
    t1 = time.time()
    waitcnt = 0
    while DCS.active(req.status):
      status = self.async_status(qhandle)
      if DCS.active(status.status):
        waitcnt += 1
        sleeptime = 0.1 if waitcnt < 50 else 0.25 if waitcnt < 100 else 1.0
        time.sleep(sleeptime)

    t2 = time.time()

    if isImpalaBased(self._conntype) and req.ophandle:
      # Use the returned coordinator node webserver url if we have it
      if req.webserver_url and getattr(settings, 'PROFILE_ARCENGINE_WEBSERVER_URL', True):
        coord_url = req.webserver_url
      else:
        coord_url = 'http://%s:%s' % (self._host, int(self._port) + 3950)
      profile_url = str('%s/query_plan?query_id=%s:%s' % (
        coord_url,
        codecs.encode(req.ophandle.operationId.guid[7::-1], "hex").decode().lstrip("0"),
        codecs.encode(req.ophandle.operationId.guid[15:7:-1], "hex").decode().lstrip("0")))

      if not getattr(settings, 'HIDE_PROFILE_QUERY', False):
        profile.record('query_profile_url', '<a target="_blank" href="%s">Profile</a>' % profile_url)

    profile.mark('async_results.wait')
    req.log('Async results: wait time %0.6f' % (t2 - t1))

    if req.status != DCS.FINISHED:
      self._delreq(req)
      raise Exception(req.message or 'Error during results fetch')

    # If no op handle, no error but nothing to do
    colnames = []
    coltypes = []

    from reports.views_settings import setting_get_single
    maxSizeMB = setting_get_single('visual_max_dataresults') or ''
    maxSizeMB = int(maxSizeMB) if maxSizeMB.isnumeric() else 10
    maxSize = maxSizeMB * 1000 * 1000
    rowsData = RowsData(maxSize)

    info = []
    if not req.ophandle or not req.ophandle.hasResultSet:
      req.log("Async results: no data returned")
      self._delreq(req)
      if stream:
        yield coltypes
        yield colnames
        return

      yield (colnames, coltypes, rowsData.fetchAndClose(), info)
      return

    # get the schema as a list of ('column name', 'column type') tuples
    try:
      ta = time.time()
      rsp = self._onethrift(req, req.conn.client.GetResultSetMetadata,
                                 TGetResultSetMetadataReq(operationHandle=req.ophandle))
      tb = time.time()
      profile.mark('async_results.fetch.columns')

      unused_status, info, unused_errtxt = exception_on_error(req, rsp)
      fncs = []
      decodes = []
      if rsp.schema:
        for c in rsp.schema.columns:
          colnames.append(c.columnName)
          coltypes.append(get_coltype_name(c.typeDesc.types[0]))
          ttype = c.typeDesc.types[0].primitiveEntry.type
          fnc, decode = TTYPETOELEMENT[ttype]
          fncs.append(fnc)
          decodes.append(decode)
        if stream:
          yield coltypes
          yield colnames

      # now the rows
      more = True
      itercount = 0
      ishive = self._conntype in [DATACONN_HIVE, DATACONN_SPARKSQL]
      while more:
        itercount += 1
        ta = time.time()
        rsp = self._onethrift(req, req.conn.client.FetchResults,
                                   TFetchResultsReq(operationHandle=req.ophandle,
                                   orientation=TFetchOrientation.FETCH_NEXT, maxRows=10000))
        tb = time.time()
        profile.mark('async_results.fetch.rows.%s' % (itercount))
        exception_on_error(req, rsp)

        more = rsp.hasMoreRows
        if len(rsp.results.rows) == 0 and rsp.results.columns and len(rsp.results.columns):
          columns = [fn(cl) for (fn,cl) in zip(fncs, rsp.results.columns)]
          profile.record('async_results.fetch.rows.count', len(columns[0].values))
          more = more or (ishive and (len(columns[0].values) > 0))
          for r in range(0, len(columns[0].values)):
            if r % 8 == 0:
              idx = r // 8
              nul = [0 if len(col.nulls) <= idx else col.nulls[idx] for col in columns]
            msk = 1 << (r % 8)
            rowdata = ["NULL" if n & msk else d(c.values[r]) for (n,c,d) in zip(nul, columns, decodes)]
            if stream:
              yield rowdata
            else:
              rowsData.addRow(rowdata)
        else:
          profile.record('async_results.fetch.rows.count', len(rsp.results.rows))
          more = more or (ishive and (len(rsp.results.rows) > 0))
          for row in rsp.results.rows:
            rowdata = []
            for col in row.colVals:
              if col.i32Val is not None and col.i32Val.value is not None:
                rowdata.append(col.i32Val.value)
              elif col.byteVal is not None and col.byteVal.value is not None:
                rowdata.append(col.byteVal.value)
              elif col.i16Val is not None and col.i16Val.value is not None:
                rowdata.append(col.i16Val.value)
              elif col.i64Val is not None and col.i64Val.value is not None:
                rowdata.append(col.i64Val.value)
              elif col.boolVal is not None and col.boolVal.value is not None:
                rowdata.append(col.boolVal.value)
              elif col.doubleVal is not None and col.doubleVal.value is not None:
                rowdata.append(col.doubleVal.value)
              elif col.stringVal is not None and col.stringVal.value is not None:
                rowdata.append(col.stringVal.value)
              else:
                rowdata.append(None)
            rowdata = ["NULL" if c is None else str(c) for c in rowdata]
            if stream:
              yield rowdata
            else:
              rowsData.addRow(rowdata)
    except Exception as e:
      req.log("Exception on results fetch: %s" % str(e))
      self._delreq(req)
      raise
    
    # See about extended info
    if self._extmeta:
      try:

        if self._extmeta == 'Old style' and not req.extresp:
          ta = time.time()
          rsp = self._onethrift(req, req.conn.client.GetResultSetExtendedMetadata,
                                TGetResultSetExtendedMetadataReq(operationHandle=req.ophandle))
          exception_on_error(req, rsp)
          tb = time.time()
          profile.mark('async_results.extended_metadata')
          req.extresp = rsp.extendedMetadata

        ext = req.extresp
        if ext and ext.rewrStmt and ext.rewrStmt != 'N/A':
          req.rewrite = ext.rewrStmt
          if not getattr(settings, 'HIDE_PROFILE_QUERY', False):
            info.append("Rewrite: %s" % req.rewrite)
          else:
            info.append("Rewrite: ***")
          if req.rewrite.count(' FROM ') == 1:
            beg = req.rewrite.find(' FROM ') + 6
            end = req.rewrite[beg:].find(' ')
            aview = req.rewrite[beg:beg+end]
          else:
            aview = '(see profile)'
          profile.record('analytical_view_name', aview)

        if getattr(ext, 'infoMessages', None):
          info.extend(ext.infoMessages)
        if getattr(ext, 'cacheInvalidationObjects', None):
          info.append(self._format_tablesigs(ext.cacheInvalidationObjects))
        if getattr(ext, 'sqlInjectionObjects', None):
          info.append('TableRefs: %s' % \
                      ','.join(set('%s.%s' % (t.databaseName, t.tableName) \
                                   for t in ext.sqlInjectionObjects \
                                   if t.tableName and t.databaseName)))

      # Just log errors for now
      except Exception as e:
        LOGGER.exception("Error getting extended metadata")
        pass

    # Done, close the request
    info.append("Sequence:%d" % req.sequence)
    query = "***"
    if not getattr(settings, 'HIDE_PROFILE_QUERY', False):
      if not req.comment:
        query = req.query
      else:
        query = req.query[len(req.comment) + 1:]

    info.append("Query:%s" % query)
    t3 = time.time()
    req.log("Async results: fetch time %0.6f, %d cols / %d rows returned" % \
        ((t3 - t2), len(colnames), rowsData.rowCount()))
    profile.mark('async_results.fetch.all')

    self._delreq(req)

    if stream:
      return

    yield (colnames, coltypes, rowsData.fetchAndClose(), info)
    return
