# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

"""
Connect to Sentry service and keep our roles in sync

Currently enabled via settings SENTRY structure.

Periodically poll sentry roles:
  For sentry each role,
    Ensure corresponding arcviz role exists
    Ensure corresponding arcviz role 'sentry_groups' contains current groups
  For each non-sentry role,
    Ensure 'sentry_groups' is empty
    Could consider deleting the role if local users and groups are empty

"""

import sys
import json
import time
import logging
import traceback
import threading


from django.conf import settings
from django.db import close_old_connections

from rbac.models import Role

# pylint: disable=W0401,W0614
from . import thrift_sentry
sys.path.extend(thrift_sentry.__path__)
from .thrift_sentry.sentry_policy_service import SentryPolicyService
from .thrift_sentry.sentry_policy_service.ttypes import *
from .thrift_sentry.sentry_common_service.ttypes import *
from .thrift_sentry.sentry_common_service.constants import *
# pylint: enable=W0401,W0614

from .thriftconn import thrift_protocol, kt_renewer

# Translate sentry site property names to our config names
SITEPARMS = {
  'sentry.service.security.mode'             : 'AUTH',
  'sentry.service.client.server.rpc-address' : 'HOST',
  'sentry.service.client.server.rpc-port'    : 'PORT',
}

LOGGER = logging.getLogger("arcweb")
SENTRYIF = None

def startSentry(conf):
  global SENTRYIF, LOGGER

  try:
    # If sentry-site.xml file given, we load some options from there
    sitefile = conf.get('SENTRY-SITE')
    if sitefile:
      import xml.etree.ElementTree as ET
      tree = ET.parse(sitefile)
      for p in tree.getroot().findall('property'):
        key = p.find('name').text
        val = p.find('value').text
        our_key = SITEPARMS.get(key)
        if our_key:
          conf[our_key] = val

    SENTRYIF = SentryIF(conf, LOGGER)
    SENTRYIF.activate()
  except Exception as e:
    LOGGER.error("Error on Sentry I/F startup: %s\n%s" %
                 (str(e), traceback.format_exc()))

class SentryIF(object):
  def __init__(self, conf, logger, mock_client=None):
    self._conf = conf
    self._logger = logger
    self._mock_client = mock_client
    self._pure_sasl = sys.platform == 'win32' or \
                      getattr(settings, 'USE_PURE_PYTHON_SASL', False)

  def log(self, msg, lvl=logging.INFO):
    if self._logger:
      self._logger.log(lvl, msg)

  def activate(self):
    # Required settings
    self._host = self._conf['HOST']
    self._auth = self._conf['AUTH']

    # optional settings
    self._username = self._conf.get('USERNAME', '')
    self._password = self._conf.get('PASSWORD', '')
    self._sentry_clientname = self._conf.get('SENTRY_CLIENTNAME', 'arcadia')

    try:
      self._port = int(self._conf.get('PORT'))
    except Exception as e:
      self._port = 8038

    socktype = self._conf.get('SOCKTYPE', 'normal').upper()
    self._ssl = False
    self._sslcert = ''
    self._sslcert_name_mismatch = False
    if socktype == 'SSL':
      self._ssl = True
      self._sslcert = self._conf.get('SSLCERT')
      self._sslcert_name_mismatch = self._conf.get('SSLCERT_NAME_MISMATCH', False)

    self._kerbserv = 'sentry'
    self._multiplex = 'SentryPolicyService'

    try:
      self._pollfreq = int(self._conf.get('POLL_FREQUENCY'))
    except Exception as e:
      self._pollfreq = (5 * 60)

    self._pollth = threading.Thread(target=self.poller)
    self._pollth.daemon = True
    self._pollth.start()
    
  def poller(self):
    while True:
      try:
        version = TSENTRY_SERVICE_V1
        self.log("Connecting to SENTRY host at: %s:%s" % (self._host, self._port))
        if self._mock_client:
          client = self._mock_client
        else:
          protocol = thrift_protocol(host=self._host, port=self._port, auth=self._auth,
                                     username=self._username, password=self._password,
                                     ssl=self._ssl, sslcert=self._sslcert,
                                     kerbserv=self._kerbserv, multiplex=self._multiplex,
                                     sslcert_name_mismatch=self._sslcert_name_mismatch,
                                     pure_sasl=self._pure_sasl)

          client = SentryPolicyService.Client(protocol)
        self.log("Connected!")

        while True:
          close_old_connections()
          req = TListSentryRolesRequest(protocol_version=version,
                                        requestorUserName=self._sentry_clientname)
          rsp = client.list_sentry_roles_by_group(req)

          # retry with version 2 of the protocol
          if rsp.status.value == TSENTRY_STATUS_THRIFT_VERSION_MISMATCH and \
             version == TSENTRY_SERVICE_V1:
            version = TSENTRY_SERVICE_V2
            continue

          if rsp.status.value != TSENTRY_STATUS_OK:
            raise Exception("Error %d from server: %s" % (rsp.status.value, rsp.status.message))

          srlist = []
          for sr in rsp.roles:
            srlist.append(sr.roleName)
            self.oneRole(sr.roleName, set(g.groupName for g in sr.groups))

          # Now for roles not named in the Sentry system, we need to make sure
          # their 'sentry_groups' setting is empty.
          for role in Role.objects.exclude(name__in=srlist):
            if len(role.sentryGroupSet()) != 0:
              role.sentry_groups = json.dumps([])
              role.save()
              self.log("Clearing groups '%s' for role '%s' not found in Sentry" %
                       (str(list(role.sentryGroupSet())), role.name))

          time.sleep(self._pollfreq)

      except Exception as e:
        self.log("Error from SENTRY: %s\n%s" % (str(e), traceback.format_exc()))

        # Should do initially fast retry then slower
        time.sleep(self._pollfreq)
        
  # Process one role from Sentry
  def oneRole(self, name, groups):
    role, created = Role.objects.get_or_create(name=name)
    if created:
      role.description = "Role mirrored from Sentry"
      role.privileges = json.dumps([{'ptype':'system','perms':[]}])
      role.users = json.dumps([])
      role.groups = json.dumps([])
      role.sentry_groups = json.dumps([])
      role.save()
      self.log("Created role '%s' to match sentry role" % name)
    ogroups = role.sentryGroupSet()
    if ogroups != groups:
      role.sentry_groups = json.dumps(list(groups))
      role.save()
      self.log("Updated role '%s' groups from %s to %s" % (name, str(ogroups), str(groups)))

