# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
from django.test import TestCase
from unittest.mock import Mock
from datasets.models import DataConnection
from jobs.views_encryption import encrypt_secret

from dataconn.connection_updater import ConnectionUpdater

class ConnectionUpdaterTest(TestCase):
    def test_when_no_change_should_not_update(self):
        # Given
        old_attr: Mock = self.__get_connection_attr_mock('', 'password', '', '', {})
        new_attr: dict = self.__get_connection_attr_dict('', 'password', '', '', {})
        # When
        did_update: bool = ConnectionUpdater().update(new_attr, old_attr)
        # Then
        self.assertFalse(did_update)

    def test_when_password_change_should_update(self):
        # Given
        old_attr: Mock = self.__get_connection_attr_mock('', 'old_password', '', '', {})
        new_attr: dict = self.__get_connection_attr_dict('', 'new_password', '', '', {})
        # When
        did_update: bool = ConnectionUpdater().update(new_attr, old_attr)
        # Then
        self.assertTrue(did_update)

    def test_when_username_change_should_update(self):
        # Given
        old_attr: Mock = self.__get_connection_attr_mock('', 'password', 'user1', '', {})
        new_attr: dict = self.__get_connection_attr_dict('', 'password', 'user2', '', {})
        # When
        did_update: bool = ConnectionUpdater().update(new_attr, old_attr)
        # Then
        self.assertTrue(did_update)

        self.assertEqual(json.loads(old_attr.dataconnection_info)['PARAMS']['USERNAME'], 'user2')

    def test_when_name_extension_change_should_update(self):
        # Given
        old_attr: Mock = self.__get_connection_attr_mock('', 'password', '', 'extend-1', {})
        new_attr: dict = self.__get_connection_attr_dict('', 'password', '', 'extend-2', {})
        # When
        did_update: bool = ConnectionUpdater().update(new_attr, old_attr)
        # Then
        self.assertTrue(did_update)

        self.assertEqual(json.loads(old_attr.dataconnection_info)['PARAMS']['CONNECTIONNAME_EXTENDED'], 'extend-2')

    def test_when_connection_name_change_should_update(self):
        # Given
        old_attr: Mock = self.__get_connection_attr_mock('conn1', 'password', '', '', {})
        new_attr: dict = self.__get_connection_attr_dict('conn2', 'password', '', '', {})
        # When
        did_update: bool = ConnectionUpdater().update(new_attr, old_attr)
        # Then
        self.assertTrue(did_update)

        self.assertEqual(old_attr.dataconnection_name, 'conn2')

    def test_when_cml_status_change_should_update(self):
        # Given
        old_attr: Mock = self.__get_connection_attr_mock('', 'password', '', '', {'isAvailable': True})
        new_attr: dict = self.__get_connection_attr_dict('', 'password', '', '', {'isAvailable': False})
        # When
        did_update: bool = ConnectionUpdater().update(new_attr, old_attr)
        # Then
        self.assertTrue(did_update)

        self.assertFalse(old_attr.cml_status['isAvailable'])

    def __get_connection_attr_mock(self, name: str, password: str, username: str, conn_extend: str, cml_status: dict) -> Mock:
        conn: Mock = Mock()

        conn.dataconnection_name = name
        conn.dataconnection_password = json.dumps(encrypt_secret(password))
        
        info: dict = {
            'PARAMS': {
                'USERNAME': username,
                'CONNECTIONNAME_EXTENDED': conn_extend,
            },
        }
        conn.dataconnection_info = json.dumps(info)
        conn.cml_status = cml_status

        return conn

    def __get_connection_attr_dict(self, name: str, password: str, username: str, conn_extend: str, cml_status: dict) -> dict:
        conn: dict = {}
        
        conn['dataconnection_name'] = name
        conn['dataconnection_password'] = json.dumps(encrypt_secret(password))

        info: dict = {
            'PARAMS': {
                'USERNAME': username,
                'CONNECTIONNAME_EXTENDED': conn_extend,
            },
        }
        conn['dataconnection_info'] = json.dumps(info)
        conn['cml_status'] = cml_status

        return conn

    def __save_conn_in_db(self, mock_conn: Mock) -> None:
        name: str = mock_conn.dataconnection_name
        password: dict = mock_conn.dataconnection_password
        info: dict = mock_conn.dataconnection_info
        cml_status: dict = mock_conn.cml_status

        conn = DataConnection(dataconnection_name=name, dataconnection_password=password, dataconnection_info=info, cml_status=cml_status)
        conn.save()