# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
tests for dataconn.py
"""

from django.test import TestCase
from dataconn import dataconn
from dataconn import dataconn_sqlite
from dataconn import dataconn_druid
from dataconn import dataconn_impyla

class Test_Dataconn(TestCase):
  def test_sanitize(self):
    self.assertEqual(dataconn.sanitize_colname(None), "column_0")
    self.assertEqual(dataconn.sanitize_colname(None, 6), "column_6")
    self.assertEqual(dataconn.sanitize_colname(None, idx=3), "column_3")

    self.assertEqual(dataconn.sanitize_colname("ABC"), "abc")
    self.assertEqual(dataconn.sanitize_colname("customer_name"), "customer_name")
    self.assertEqual(dataconn.sanitize_colname("customer name"), "customer_name")
    self.assertEqual(dataconn.sanitize_colname("sum(price)"), "sum_price")

    hdrs = ["abc" ,"def"]
    result = ["abc", "def"]
    self.assertEqual(dataconn.sanitize_headers(hdrs, None), result)

    hdrs = ["customer_name", "sum(price)"]
    result = ["customer_name", "sum_price"]
    self.assertEqual(dataconn.sanitize_headers(hdrs, None), result)

    hdrs = ["customer_name", "customer name", None, "sum(price)", "sum(price)"]
    result = ["customer_name_1", "customer_name_2", "column_2", "sum_price_1", "sum_price_2"]
    self.assertEqual(dataconn.sanitize_headers(hdrs, None), result)

    hdrs = ["customer_name", "Customer Name", "Customer Name 2"]
    result = ["customer_name_1", "customer_name_2", "customer_name_2_1"]
    self.assertEqual(dataconn.sanitize_headers(hdrs, None), result)

    hdrs = ["customer_name", "Customer Name 2", "Customer Name"]
    result = ["customer_name_1", "customer_name_2", "customer_name_3"]
    self.assertEqual(dataconn.sanitize_headers(hdrs, None), result)

  def test_quotetablename(self):
    conn = dataconn_sqlite.SqliteConn({
      "FILE": "/nodir/nofile",
      "FOR_TESTING": True,
    }, 1)

    tablename = "foo.bar"
    f = conn.quote_tablename(tablename)
    self.assertEquals(f, "`foo`.`bar`")

    # Confirm that quoted names are not quoted again i.e. this function
    # can be called idempotently
    f = conn.quote_tablename(f)
    self.assertEquals(f, "`foo`.`bar`")

    # Confirm that quoting is data connection specific
    conn = dataconn_druid.DruidConn({
      "FOR_TESTING": True,
    }, 1)

    tablename = "foo.bar"
    f = conn.quote_tablename(tablename)
    self.assertEquals(f, '"foo"."bar"')

  def test_tablenameparts(self):
    conn = dataconn_sqlite.SqliteConn({
      "FILE": "/nodir/nofile",
      "FOR_TESTING": True,
    }, 1)

    tablename = "foo.bar"
    f = conn.tablename_parts(tablename)
    self.assertEquals(len(f), 2)
    self.assertEquals(f[0], "foo")
    self.assertEquals(f[1], "bar")

    tablename = "`foo`.`bar`"
    f = conn.tablename_parts(tablename)
    self.assertEquals(len(f), 2)
    self.assertEquals(f[0], "foo")
    self.assertEquals(f[1], "bar")

    tablename = "`foo`.bar"
    f = conn.tablename_parts(tablename)
    self.assertEquals(len(f), 2)
    self.assertEquals(f[0], "foo")
    self.assertEquals(f[1], "bar")

    tablename = "foo.`bar`"
    f = conn.tablename_parts(tablename)
    self.assertEquals(len(f), 2)
    self.assertEquals(f[0], "foo")
    self.assertEquals(f[1], "bar")

    # Confirm that quoting is data connection specific
    conn = dataconn_druid.DruidConn({
      "FOR_TESTING": True,
    }, 1)

    tablename = "foo.`bar`"
    f = conn.tablename_parts(tablename)
    self.assertEquals(len(f), 2)
    self.assertEquals(f[0], "foo")
    self.assertEquals(f[1], "`bar`")

    tablename = '"foo"."bar"'
    f = conn.tablename_parts(tablename)
    self.assertEquals(len(f), 2)
    self.assertEquals(f[0], "foo")
    self.assertEquals(f[1], "bar")

  def test_droptable(self):
    conn = dataconn_sqlite.SqliteConn({
      "FILE": "/nodir/nofile",
      "FOR_TESTING": True,
    }, 1)

    cmd = conn.drop_table_cmd("foo.bar")
    self.assertEquals(cmd, "DROP TABLE IF EXISTS `foo`.`bar`")

    # Confirm that quoting is data connection specific
    conn = dataconn_druid.DruidConn({
      "FOR_TESTING": True,
    }, 1)

    cmd = conn.drop_table_cmd("foo.bar")
    self.assertEquals(cmd, 'DROP TABLE IF EXISTS "foo"."bar"')

  def test_truncate(self):
    conn = dataconn_sqlite.SqliteConn({
      "FILE": "/nodir/nofile",
      "FOR_TESTING": True,
    }, 1)

    cmd = conn.truncate_table_cmd("foo.bar")
    self.assertEquals(cmd, "DELETE FROM `foo`.`bar`")

    # Confirm truncate-table form for connections that support it
    conn = dataconn_impyla.ImpylaConn({
      "HOST": "host",
      "PORT": "port",
      "AUTH": "auth",
      "FOR_TESTING": True,
    }, 1)

    cmd = conn.truncate_table_cmd("foo.bar")
    self.assertEquals(cmd, "TRUNCATE TABLE `foo`.`bar`")

  def test_create(self):
    conn = dataconn_sqlite.SqliteConn({
      "FILE": "/nodir/nofile",
      "FOR_TESTING": True,
    }, 1)

    cmd = conn.create_sql_cmds("foo.bar", ["c1", "c2"], ["String", "Integer"],
        "user", "srcfile", True, "loc1")
    self.assertEquals(cmd, "CREATE TABLE `foo`.`bar` (`c1` TEXT,`c2` INTEGER)")

    conn = dataconn_impyla.ImpylaConn({
      "HOST": "host",
      "PORT": "port",
      "AUTH": "auth",
      "FOR_TESTING": True,
    }, 1)

    cmd = conn.create_sql_cmds("foo.bar", ["c1", "c2"], ["String", "Integer"],
        "user", "srcfile", True, "loc1")
    self.assertEquals(cmd,
        "CREATE EXTERNAL TABLE `foo`.`bar` (`c1` string,`c2` bigint) " + \
        "STORED AS PARQUET  LOCATION 'loc1'  TBLPROPERTIES("           + \
        "'arc_importedby'='user', 'arc_importedfile'='srcfile')")

    cmd = conn.create_sql_cmds("foo.bar", ["c1", "c2"], ["String", "Integer"],
        "user", "srcfile", False, "loc1")
    self.assertEquals(cmd,
        "CREATE TABLE `foo`.`bar` (`c1` string,`c2` bigint) "         + \
        "STORED AS PARQUET  LOCATION 'loc1'  TBLPROPERTIES("          + \
        "'arc_importedby'='user', 'arc_importedfile'='srcfile')")

    cmd = conn.create_sql_cmds("foo.bar", ["c1", "c2"], ["String", "Integer"],
        "user", None, False, "loc1")
    self.assertEquals(cmd,
        "CREATE TABLE `foo`.`bar` (`c1` string,`c2` bigint) "         + \
        "STORED AS PARQUET  LOCATION 'loc1'  TBLPROPERTIES("          + \
        "'arc_importedby'='user')")

    cmd = conn.create_sql_cmds("foo.bar", ["c1", "c2"], ["String", "Integer"],
        None, None, False, None)
    self.assertEquals(cmd,
        "CREATE TABLE `foo`.`bar` (`c1` string,`c2` bigint) " + \
        "STORED AS PARQUET ")

    cmd = conn.create_sql_cmds("foo.bar", ["c1", "c2"], ["String", "Integer"],
        None, None, False, None, partitionColnames=["year", "month"],
        partitionColtypes=["Integer", "String"])
    self.assertEquals(cmd,
        "CREATE TABLE `foo`.`bar` (`c1` string,`c2` bigint) " + \
        "PARTITIONED BY (year Integer, month String) "       + \
        "STORED AS PARQUET ")

  def test_droptable_nocheck(self):
    conn = dataconn_sqlite.SqliteConn({
      'FILE': '/nodir/nofile',
      'FOR_TESTING': True,
    }, 1)

    cmd = conn.drop_table_no_check_cmd('foo.bar')
    self.assertEquals(cmd, 'DROP TABLE `foo`.`bar`')

  def test_renametable(self):
    conn = dataconn_sqlite.SqliteConn({
      'FILE': '/nodir/nofile',
      'FOR_TESTING': True,
    }, 1)

    cmd = conn.rename_table_cmd('foo.bar', 'newbar')
    self.assertEquals(cmd, 'ALTER TABLE `foo`.`bar` RENAME TO `newbar`')

    # sqlite strips out new table name's db part
    cmd = conn.rename_table_cmd('foo.bar', 'foo.newbar')
    self.assertEquals(cmd, 'ALTER TABLE `foo`.`bar` RENAME TO `newbar`')

    # sqlite raises an exception
    with self.assertRaises(ValueError) as exc:
      cmd = conn.rename_table_cmd('foo.bar', 'notfoo.bar')
    self.assertEqual('Database name is not the same for source and target table names', str(exc.exception))

    conn = dataconn_impyla.ImpylaConn({
      'HOST': 'host',
      'PORT': 'port',
      'AUTH': 'auth',
      'FOR_TESTING': True,
    }, 1)

    cmd = conn.rename_table_cmd('foo.bar', 'newbar')
    self.assertEquals(cmd, 'ALTER TABLE `foo`.`bar` RENAME TO `newbar`')

    cmd = conn.rename_table_cmd('foo.bar', 'foo.newbar')
    self.assertEquals(cmd, 'ALTER TABLE `foo`.`bar` RENAME TO `foo`.`newbar`')

    cmd = conn.rename_table_cmd('foo.bar', 'notfoo.bar')
    self.assertEquals(cmd, 'ALTER TABLE `foo`.`bar` RENAME TO `notfoo`.`bar`')
