# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import asyncio
import django.test
import json
import time

from apps.tests.userrole import base
from dataconn import dataconn_http
from dataconn.dataconn import get_django_dataconn, get_dataconn
from django.core.cache import cache
from reports.views_settings import setting_update_single
from unittest import mock
from unittest.mock import AsyncMock
from vizlib.dataset_request import DatasetRequest
from vizlib import misc
from vizlib.tests.test_sqlrun import get_json_response

ASYNC_RETRY_TIMEOUT_SECONDS = 2


class TestDataconnHttpBasic(django.test.TestCase):
  @mock.patch('arcwebbase.ai_validator.validate_and_transform_request', new_callable=AsyncMock)
  @mock.patch('arcwebbase.views_proxy.proxy_comm')
  def test_instance(self, proxy_comm, mock_validate):
    mock_validate.side_effect = lambda req: req
    async def single_response(*args, **kwargs):
      yield {'status_code': 200, 'response_body': 'MOCK_RESULT'}

    proxy_comm.side_effect = single_response

    self.assertEquals(get_django_dataconn(dataconn_http.INSTANCE.id), dataconn_http.INSTANCE)

    conn = get_dataconn(connection_id=dataconn_http.INSTANCE.id)
    self.assertEquals(conn.connection_type, 'HTTP')

    dataset_request = DatasetRequest({'type': 'HTTP', 'data': {'a': 'b'}})
    build_req = conn.build(dataset_request=dataset_request)

    excute_handle = conn.execute(build_req)
    qhandle = excute_handle.get('qhandle')
    self.assertIsNotNone(qhandle)

    async_runner = misc.get_async_runner(dataconn_http.INSTANCE.id)
    runner = async_runner.tracker[qhandle]

    for i in range(10):
      time.sleep(0.1)
      if runner.status() == runner.DONE:
        break

    self.assertEqual(runner.results()['response'], 'MOCK_RESULT')

  async def delayed_proxy_comm(self, *args):
    await asyncio.sleep(5)
    yield (200, 'DELAYED MOCK RESPONSE')

  def test_cancelling(self):
    views_proxy_mock = mock.Mock()
    views_proxy_mock.proxy_comm = self.delayed_proxy_comm
    get_views_proxy_mock = mock.Mock(return_value=views_proxy_mock)

    with mock.patch('dataconn.dataconn_http.get_views_proxy', get_views_proxy_mock):
      self._test_canceling()

  def _test_canceling(self):

    conn = get_dataconn(connection_id=dataconn_http.INSTANCE.id)
    dataset_request = DatasetRequest({'type': 'HTTP', 'data': {'a': 'b'}})
    build_req = conn.build(dataset_request=dataset_request)

    excute_handle = conn.execute(build_req)
    qhandle = excute_handle.get('qhandle')

    async_runner = misc.get_async_runner(dataconn_http.INSTANCE.id)
    runner = async_runner.tracker[qhandle]

    http_runner = runner.progress_runner()

    # wait for the task attribute to be set
    for i in range(10):
      time.sleep(0.1)
      if getattr(http_runner, '_task'):
        break

    start = time.time()
    runner.markCancelled()
    for i in range(60):
      print(runner.status())
      time.sleep(0.1)
      if runner.status() == runner.DONE:
        break

    elapsed_time = time.time() - start
    self.assertIn('cancelled', runner.results())
    # the async task waits 5 seconds. If cancelling works, we should get here faster
    self.assertLess(elapsed_time, 5)


class TestDataconnHttp(django.test.TestCase):

  def setUp(self):
    user1 = base.make_user('user1')
    self.client = base.create_client(user1)

    # needed by the jsonselect tests:
    setting_update_single('visual_maxrows', 100, False)

  def tearDown(self):
    # settings are cached. So even if database changes are discarded, the cache persists:
    cache.clear()

  def _post(self, url: str, data: dict):
    response = self.client.post(url, data=data)
    response_json = get_json_response(response)
    return response_json

  def jsonselect_qhandle_status(self, qhandle: str, start_time: float):
    url = '/sqlrun/p1/asyncstatus_async'
    response_json = self._post(url, {'qhandles': qhandle})
    return self.async_retry(url, response_json, start_time)

  def jsonselect_start(self, payload: dict = None):
    start_time = time.time()
    url = '/sqlrun/jsonselect_parallel_async'
    response_json = self._post(url, data=payload)
    response_json = self.async_retry(url, response_json, start_time)

    qhandle = response_json[0].get('response', {}).get('qhandle')

    while True:
      response_json = self.jsonselect_qhandle_status(qhandle, start_time)

      if isinstance(response_json, list):
        results = response_json[0].get('results')
        if results == 'true':
          response = response_json[0]
          return response['response']['response'], response['profile']

      time.sleep(0.1)

  def async_retry(self, url: str, response_json: dict, start_time: float):
    while isinstance(response_json, dict) and response_json.get('type') == 'async':
      if time.time() > start_time + ASYNC_RETRY_TIMEOUT_SECONDS:
        raise Exception(f'Async retries exceeded timeout of {ASYNC_RETRY_TIMEOUT_SECONDS}s')

      request_id = response_json.get('id')
      time.sleep(0.1)

      response_json = self._post(url, data={'async': json.dumps({'id': request_id})})
    return response_json

  @mock.patch('arcwebbase.ai_validator.validate_and_transform_request', new_callable=AsyncMock)
  @mock.patch('arcwebbase.views_proxy.proxy_comm')
  def test_web_interface(self, proxy_comm, mock_validate):
    mock_validate.side_effect = lambda req: req

    async def single_response(*args, **kwargs):
      yield {'status_code': 200, 'response_body': {'mock_response': 'MOCK_RESULT'}}

    proxy_comm.side_effect = single_response

    payload = {
      'data': json.dumps([
        {
          'params': {
            'async': 'true',
            'dataconnection_id': dataconn_http.INSTANCE.id,
            'dsreq': json.dumps({
              'version': 1,
              'type': 'HTTP',
              'data': {
                'a': 'b'
              }
            })
          }
        }
      ])
    }

    response, profile = self.jsonselect_start(payload)

    # we received the mock response
    self.assertEquals(response, {'mock_response': 'MOCK_RESULT'})

    # our data was passed to the mock
    self.assertEquals(proxy_comm.call_args[0][0], {'a': 'b'})

    query_runtime_line = [p for p in profile if p[0] == 'query_runtime']
    self.assertIsNotNone(query_runtime_line[0][1])
