# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
tests for dataconn_ksql.py
"""

from django.test import TestCase, RequestFactory
from django.contrib.auth.models import User
from datasets.models import DataSet, DataConnection
from django.test.client import Client

from dataconn import dataconn_ksql

import json

class Test_KSQL_Tracker(TestCase):
  def __init__(self, *args, **kwargs):
    super(Test_KSQL_Tracker, self).__init__(*args, **kwargs)

  # initial unit test setup
  def setUp(self):
    pass

  # called after unit tests are executed
  def tearDown(self):
    pass

  def test_query_tracking(self):
    class QueryRunner(object):
      def __init__(self, query):
        self._query = query

      def queryTrackerName(self):
        return self._query

      def query(self):
        return self._query

    tracker = dataconn_ksql._KSQLQueryTracker()
    q1 = QueryRunner('q1')
    q2 = QueryRunner('q2')

    tracker.trackQuery(q1)
    self.assertEqual(tracker.getQuery(query='q1'), q1)
    self.assertEqual(tracker.getQuery(query='q2'), None)

    tracker.trackQuery(q2)
    self.assertEqual(tracker.getQuery(query='q1'), q1)
    self.assertEqual(tracker.getQuery(query='q2'), q2)

  def test_data(self):
    conn = dataconn_ksql.KsqlConn({}, 1)
    dataStore = dataconn_ksql._KSQLDataStore()

    c1 = dataconn_ksql._KSQLDataConsumer(dataStore=dataStore, conn=conn)
    c2 = dataconn_ksql._KSQLDataConsumer(dataStore=dataStore, conn=conn)

    self.assertEqual(len(c1), 0)
    self.assertEqual(len(c2), 0)

    dataStore.append('data1')

    self.assertEqual(len(c1), 1)
    self.assertEqual(len(c2), 1)

    self.assertEqual(c1.pop_all(), ['data1'])
    self.assertEqual(c1.pop_all(), [])
    self.assertEqual(len(c1), 0)
    self.assertEqual(len(c2), 1)

    dataStore.append('data2')

    self.assertEqual(len(c1), 1)
    self.assertEqual(len(c2), 2)

    self.assertEqual(c1.pop_all(), ['data2'])
    self.assertEqual(c1.pop_all(), [])

    self.assertEqual(c2.pop_all(), ['data1', 'data2'])
    self.assertEqual(c2.pop_all(), [])

    c3 = dataconn_ksql._KSQLDataConsumer(dataStore=dataStore, conn=conn)
    c4 = dataconn_ksql._KSQLDataConsumer(dataStore=dataStore, conn=conn)

    self.assertEqual(c3.pop_all(), ['data1', 'data2'])
    self.assertEqual(c3.pop_all(), [])

    self.assertEqual(c4.pop_all(), ['data1', 'data2'])
    self.assertEqual(c4.pop_all(), [])

  def test_query_runners(self):
    conn = dataconn_ksql.KsqlConn({}, 1)

    (c1, r1) = dataconn_ksql._KSQLQueryRunner.getDataConsumerAndQueryRunner(query='q1', conn=conn,
        timeout=None, runAsyncQuery=False)
    (c2, r2) = dataconn_ksql._KSQLQueryRunner.getDataConsumerAndQueryRunner(query='q1', conn=conn,
        timeout=None, runAsyncQuery=False)

    self.assertEqual(r1, r2)
    self.assertEqual(c2.id(), c1.id() + 1)

    dataStore = r1.dataStore()
    dataStore.append('data1')
    dataStore.append('data2')

    self.assertEqual(c1.pop_all(), ['data1', 'data2'])
    self.assertEqual(c1.pop_all(), [])

    self.assertEqual(c2.pop_all(), ['data1', 'data2'])
    self.assertEqual(c2.pop_all(), [])

