namespace cpp impala
namespace java org.apache.impala.thrift

// Information related to determining if a Django cache entry
// is fresh or not. 'databaseName' and 'tablename' uniquely
// identify a table. 'tldt' (Transient Last DDL Timestamp) tracks
// the time at which last ddl change was made to the table. 'crc'
// tracks changes to the data in the table. 'crc' is optional
// since an empty table does not have a 'crc'.
struct TCacheInvalidationObject {
  1: required string databaseName
  2: required string tableName
  3: required string tldt
  4: optional string crc
}

// Information about the tables in a dataset. 'databaseName' and
// 'tableName' uniquely identify the table. 'referencedColumns'
// is currently not used but will be used for detecting column
// injection in the future.
struct TSQLInjectionObject {
  1: required string databaseName
  2: required string tableName
  3: optional list<string> referencedColumns
}

// Represents 'filter association' from ArcViz.
// Filter association is a set of predicates that must be applied
// in the query for a user. A filter association structure captures
// the predicate and the table name to which that predicate should
// be applied.
struct TFilterAssociation {
  1: required string filterExpr       // of the form: "a > 5 and b < 10"
  2: required list<string> tableNames // list of schema qualified table names
}

// Extended metadata request to ArcEngine requesting additional information
// regarding aviews used, Django cache invalidation, SQL injection and
// manner of query execution.
// Cache invalidation and SQL injection info for this query as cached in
// ArcViz Django cache at the time the query was issued.
// Also passed in are filter associations that must be applied, whether
// fallback is enabled on this connection and whether the user is allowed
// to call system functions like 'version', 'pid' etc.
struct TExtMetadataReq {
  1: required list<TCacheInvalidationObject> cacheInvalidationObjectsInput
  2: required list<TSQLInjectionObject> sqlInjectionObjectsInput
  3: required list<TFilterAssociation> filterAssociations
  4: required bool fallbackEnabled
  5: required bool systemFunctionsAllowed
  6: required bool performSQLInjectionChecks
}

// Information regarding how/where the query should be executed
enum TQueryExecMethod {
  ARCENGINE_BASE,        // Query has some features that require it to be
                         // executed by arcengine
  ARCENGINE_AVIEW,       // Query routes to aview(s)
  DJANGO_CACHE,          // Query can be served from the django cache
  FALLBACK               // Query must be executed by the fallback engine
  SQL_INJECTION_DETECTED // Query will not be executed
}

// The structure captures information regarding the query and the objects in it
struct TExtMetadataResp {
  // Did the query route to aview(s)?
  1: required bool aviewUsed
  // The rewritten SQL statement if the query could be routed to aview(s)
  2: required string rewrStmt
  // CRCs and TLDTs for objects in the query. The info is used for Django cache
  // invalidation.
  3: required list<TCacheInvalidationObject> cacheInvalidationObjects
  // How/where the query will be executed
  4: required TQueryExecMethod queryExecMethod
  // Details of the query execution method
  5: required string queryExecDetails
}
