include "Logging.thrift"

namespace cpp impala
namespace java org.apache.impala.thrift

// The applications supported for retrieving log files
enum TLogApplication {
  CATALOGD,
  STATESTORED,
  ARCENGINED
}

// A continuation token.  Since logs are huge, we don't
// pass all the log files back in one iteration. This token
// will be sent back to the caller.  On the next call, the
// caller will send back the same token so the next set of
// log information can be returned.  The caller should not
// change anything within the token or bad things could
// happen.
struct TContinuationToken {
  // log file name
  1: required string fileName

  // size of log file.  Needed for when the log file
  // is still in use.  This fileSize will mark the
  // end of data that will be sent back so that the
  // caller won't continuously call in for new data
  // that is constantly being written.
  2: required i64 fileSize

  // offset into current log file
  3: required i64 offset

  // current node containing log files
  4: required string nodeName

  // The file time of the log file.  This is a convenience
  // field which also could have been grabbed from the fileName
  5: required i64 fileTime
}

struct TServerLogSharedReq {

  // Max bytes to read.  Default 1MB
  1: optional i32 maxBytes = 1000000

  // Deprecated: Starting point for retrieving just the tail of the log.
  // Example set to 1MB to get (up to) the 1MB from the end of file.
  // Should not be used with timestamp
  2: optional i32 startPoint = 0

  // Continuation token.  Returned from the server on each read.
  // Use the returned value on the next call to continue the read
  // from the previous point.
  3: optional TContinuationToken continuationToken

  // Starting time of where to fetch information
  4: optional i64 timestamp = 0

  // Amount of time to fetch after "timestamp" (default is 2 hours)
  5: optional i32 duration = 7200

  // For use if we are tailing the log
  6: optional bool continueForever = false

  // List of nodes of log files to retrieve.
  7: optional list<string> nodes

  // Only look at coordinator nodes
  8: optional bool coordinators_only = false
}

// Structure for access to server log(s) to simplify client side
// error analysis/reporting.
struct TGlogServerLogInternalReq {
  1: required TServerLogSharedReq sharedReq

  // Type of log file (e.g. INFO, WARNING)
  2: required Logging.TLogLevel logLevel

  3: required TLogApplication app
}

// Structure to get profile for a given query
struct TProfilesServerLogInternalReq {
  1: required TServerLogSharedReq sharedReq

  2: optional string queryId
}

struct TServerLogSharedResp {
  // continuation token.  If blank, there is no more data to
  // send back.
  1: optional TContinuationToken continuationToken

  // String (could be very long) containing log lines.
  2: required string logData

  // Error message (blank if all is good).
  3: optional string errorMessage

  // Node name where the logData was fetched.
  4: optional string nodeName

  // File name where the logData was fetched.
  5: optional string fileName

  // nodes that still need to be fetched
  6: optional list<string> nodesToFetch
}

