# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

"""
Support for establishing thrift connection and protocol layer
Clients then build the particular service using the protocol object
"""

import os
import sys
import time
import base64
import logging
import threading
import itertools
import subprocess
import re

from socket import gethostbyname

from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.transport.TTransport import TTransportException
from thrift.protocol import TBinaryProtocol

# Kerberos renewer stuff -- we can run 'kinit ...' every so often to
# ensure we have a valid ticket

# Global array of active renewers, indexed by (principal, keytab) tuple
KT_RENEWERS = {}

# Structure to hold info for kerberos ticket renewer thread
class kt_control(object):
  def __init__(self, principal, keytab, frequency, logger=None):
    self.principal = principal
    self.keytab = keytab
    self.frequency = frequency
    self.logger = logger
    self.th = None
    self.lastrun = 0

  def get_kinit_command(self):
    cmd = ['kinit', '-r', str(self.frequency * 2), '-k', '-t', self.keytab, self.principal]
    return cmd

  def run_kinit(self):
    """
    Return True if kinit succeeded, False otherwise
    """
    cmd = self.get_kinit_command()
    if self.logger:
      self.logger.info("Kerberos reinit: %s" % " ".join(cmd))

    subp = subprocess.Popen(cmd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            close_fds=True,
                            text=True,
                            bufsize=-1)
    subp.wait()
    if subp.returncode != 0:
      if self.logger:
        self.logger.error("Couldn't reinit from keytab! `kinit' exited with %s.\n%s\n%s" % (
                            subp.returncode,
                            "\n".join(subp.stdout.readlines()),
                            "\n".join(subp.stderr.readlines())))
      return False

    return True

  def kt_renewer(self):
    while True:

      success = self.run_kinit()
      if not success:
        return

      self.lastrun = time.time()
      time.sleep(self.frequency)

  def begin(self):
    self.th = threading.Thread(target=self.kt_renewer, name="Kerberos Renewer")
    self.th.daemon = True
    self.th.start()

    # Wait a bit for the initial ticket renewal
    for i in range(0, 20):
      time.sleep(0.1)
      if self.lastrun != 0:
        break

def kt_renewer(principal, keytab, frequency=(60 * 60 * 12), logger=None):
  global KT_RENEWERS

  index = (principal, keytab)
  ktr = KT_RENEWERS.get(index)
  if ktr is None:
    ktr = kt_control(principal, keytab, frequency, logger)
    KT_RENEWERS[index] = ktr
    ktr.begin()

  return ktr

# Start the kerberos ticket renewer if needed
def kt_renewer_from_settings(settings, logger):
  try:
    principal = getattr(settings, 'KERBEROS_PRINCIPAL', None)
    keytab = getattr(settings, 'KERBEROS_KEYTAB', None)
    if principal and keytab:
      # Default to 12 hour renewal period
      try:
        frequency = int(getattr(settings, 'KERBEROS_RENEW_SECONDS'))
      except Exception:
        frequency = 60 * 60 * 12
      # from dataconn.thriftconn import kt_renewer
      kt_renewer(principal, keytab, frequency, logger)
  except Exception as e:
    logger.error("Failed to activate kerberos ticket renewer: %s\n%s" %
                 (str(e), traceback.format_exc()))


# Routine to create a thrift connection and protocol
def thrift_protocol(host=None, port=None, username=None, password=None, auth=None,
                    ssl=None, sslcert=None, kerbserv=None, multiplex=None,
                    sslcert_name_mismatch=False, pure_sasl=False,
                    http_service=None, http_access_token=None,
                    open_timeout=10, normal_timeout=None):

  try:
    port = int(port)
  except Exception as e:
    raise Exception("Invalid port: '%s'" % port)

  if port < 0 or port > 65535:
    raise Exception("Port must be from 0 to 65535, not '%s'" % port)

  if not host:
    raise Exception("No host name specified")

  try:
    gethostbyname(host)
  except Exception as e:
    raise Exception("Unable to resolve host '%s': %s" % (host, str(e)))

  auth_modes = ['PLAIN', 'SASL', 'NOSASL', 'LDAP', 'KERBEROS', 'NONE']
  if auth is not None:
    auth = auth.upper()
    if auth not in auth_modes:
      raise Exception("Invalid authorization mode: %s, must be one of %s" % (auth, auth_modes))


  # Get correct socket type
  if http_service:
    from .thrift_http.thttp_client import THttpClient
    http_mode = 'https' if ssl else 'http'
    url = '%s://%s:%s' % (http_mode, host, port)
    socket = THttpClient(url, http_service)
    if http_access_token:
      socket.setCustomHeaders({'Authorization' : 'Basic ' + base64.standard_b64encode(('token:' + http_access_token).encode()).decode()})
    elif auth in ['SASL', 'KERBEROS']:
      socket.set_kerberos_auth()
    elif os.getenv('CDV_SUB_MODE') == 'C1C':
      user = User({'username': username})
      socket.set_bearer_auth(user.userprofile.jwt_token)
    else:
      socket.set_basic_auth(username, password)
    if ssl and sslcert:
      socket.set_verify(sslcert)
    else:
      socket.set_verify(False)
  elif not ssl:
    socket = TSocket.TSocket(host, port)
  else:
    from thrift.transport import TSSLSocket
    if not sslcert:
      socket = TSSLSocket.TSSLSocket(host, int(port), validate=False)
    else:
      socket = TSSLSocket.TSSLSocket(host, int(port), validate=True, ca_certs=str(sslcert))

  if auth in [None, 'NOSASL', 'NONE'] or http_service:
    transport = TTransport.TBufferedTransport(socket)
  else:

    # Only import the sasl stuff when needed
    if pure_sasl:
      from . import saslppwrapper as sasl
    else:
      import sasl

    from thrift_sasl import TSaslClientTransport

    saslc = sasl.Client()
    saslc.setAttr("host", str(host))
    if auth in ['PLAIN', 'SASL', 'LDAP']:
      saslc.setAttr("username", str(username) or "username")
      saslc.setAttr("password", str(password) or "password")
      saslc.init()
      transport = TSaslClientTransport((lambda: saslc), "PLAIN", socket)
    else:
      # TODO: Might need kinit here to ensure kerberos tick is valid with something like:
#     import subprocess
#     cmd = ['kinit', '-r', '1d', '-k', '-t', '/home/ec2-user/shant.keytab', 'shant@EC2.INTERNAL']
#     p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
#     exitcode = p.wait()
#     if exitcode != 0:
#       raise Exception("Error from kinit: %d" % exitcode)

      saslc.setAttr("service", str(kerbserv))
      saslc.init()
      transport = TSaslClientTransport((lambda: saslc), "GSSAPI", socket)

  try:
    if open_timeout is not None:
      socket.setTimeout(open_timeout * 1000)
    transport.open()
  except TTransportException as e:
    # This is kind of a hack...we want to allow the 'host commonName mismatch'
    # error but there is no option for this into the transport.  Fortunately
    # this check is the final action of the open() routine, so we can catch the
    # thrown exception and not fail in this case.
    erstr = 'Hostname we connected to "' + host + '"' + \
            " doesn't match certificate provided commonName"
    if sslcert_name_mismatch and str(e)[0:len(erstr)] == erstr:
      transport.is_valid = True
    else:
      raise

  socket.setTimeout(None if normal_timeout is None else (normal_timeout * 1000))

  if multiplex:
    protocol = TBinaryProtocol.TBinaryProtocol(transport)
  else:
    protocol = TBinaryProtocol.TBinaryProtocolAccelerated(transport)

  # Wrap the base protocol if needed (import
  if multiplex:
    from thrift.protocol import TMultiplexedProtocol
    protocol = TMultiplexedProtocol.TMultiplexedProtocol(protocol, multiplex)

  return protocol

