# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json

from datasets.models import DataSet
from reports.audit import auditApp, get_dc_info
from reports.models import Report, get_visuals_picklist
from vizlib.querylog import ActionType, ComponentType, logAction
import collections
import django.apps


class DashboardAssociation:
    def associate(self, dataset_id, dashboard_ids):
        dataset = DataSet.objects.get(id=dataset_id)
        dashboards = Report.objects.filter(id__in=dashboard_ids)

        associated_dashboards = [d.id for d in dashboards if dataset_id in d.get_datasets_ids_in_dashboard()]
        old_dashboards = json.loads(dataset.dashboards)
        dataset.dashboards = json.dumps(old_dashboards + associated_dashboards)
        dataset.save()

    def get_user_dashboards(self, dataset_id, user):
        dashboards_in_workspace = []
        all_dashboards = self.get_dashboards(dataset_id=dataset_id)
        for dashboard in all_dashboards:
            if dashboard.can_view_report(user):
                dashboards_in_workspace.append(dashboard)

        return dashboards_in_workspace

    def get_dashboards(self, dataset_id):
        associated_dashboard_ids = json.loads(
            DataSet.objects.get(id=dataset_id).dashboards
        )
        dashboards = list(Report.objects.filter(id__in=associated_dashboard_ids))
        return dashboards

    def remove_dashboard_from_visual_dataset(self, visual_id):
        dashboards = Report.objects.filter(report_type="dashboard")
        for dashboard in dashboards:
            (visuals, pickers) = dashboard.get_visuals_in_dashboard()
            if visual_id in visuals:
                visual = Report.objects.get(id=visual_id)
                dataset = DataSet.objects.get(id=visual.dataset_id)
                dataset.remove_dashboard_from_dataset(dashboard)
                return

    def get_possible_removed_reports(self, dataset_id, user):
        possible_deleted_reports = set[Report]()
        not_owned_reports = set[Report]()

        dataset = DataSet.objects.only('dashboards').get(id=dataset_id)
        associated_dashboards = json.loads(dataset.dashboards)

        dashboards = self.get_reports(associated_dashboards)

        dashboard_to_visuals = dict[int, list[int]]()
        visual_ids = list[int]()
        for dashboard in dashboards:
            visuals,_ = dashboard.get_visuals_in_dashboard()
            dashboard_to_visuals[dashboard.id] = visuals
            visual_ids.extend(visuals)

        visual_to_dataset = dict[int,int]()
        visuals_with_dataset = Report.objects.only('dataset_id').filter(id__in=set(visual_ids))
        for visual in visuals_with_dataset:
            visual_to_dataset[visual.id] = visual.dataset_id

        dashboard_individually_deleted_visuals = dict[int,int]()
        for dashboard_id, visuals in dashboard_to_visuals.items():
            datasets = {visual_to_dataset[visual] for visual in visuals if visual in visual_to_dataset}
            dataset_count = len(datasets)

            if dataset_count > 1:
                affected_visuals = [visual for visual in visuals if visual_to_dataset.get(visual) == dataset_id]
                dashboard_individually_deleted_visuals[dashboard_id] = affected_visuals

        visible_report_ids = list[int]()
        invisible_report_ids = list[int]()

        for dashboard in dashboards:

            if dashboard.id in dashboard_individually_deleted_visuals:
                # the dashboard will be kept, instead, some of the contained visuals will be deleted
                reports = set()
                extra_report_ids = dashboard_individually_deleted_visuals[dashboard.id]
            else:
                # the whole dashboard will be removed
                reports = {dashboard}
                extra_report_ids = []

            if dashboard.can_view_report(user):
                possible_deleted_reports.update(reports)
                visible_report_ids.extend(extra_report_ids)
            else:
                not_owned_reports.update(reports)
                invisible_report_ids.extend(extra_report_ids)

        visible_reports = set(self.get_reports(visible_report_ids))
        invisible_reports = set(self.get_reports(invisible_report_ids))

        possible_deleted_reports.update(visible_reports)
        not_owned_reports.update(invisible_reports)

        # Create copies before iterating to avoid re-checking items
        possible_deleted_reports_copy = possible_deleted_reports.copy()
        not_owned_reports_copy = not_owned_reports.copy()

        # Up until that line, we are only checking if the user can view the dashboards.
        # However, the permissions for the linked visuals’ workspace
        # might be different from those for the dashboard.
        for report in possible_deleted_reports_copy:
            if not report.can_view_report(user):
                # The user can't view the linked visual itself,
                # move it to not_owned_reports if not already present
                not_owned_reports.add(report)
                possible_deleted_reports.remove(report)

        for report in not_owned_reports_copy:
            if report.can_view_report(user):
                # The user can view the linked visual itself,
                # move it to possible_deleted_reports if not already present
                possible_deleted_reports.add(report)
                not_owned_reports.remove(report)

        return possible_deleted_reports, not_owned_reports

    def __get_removing_candidates_in_dashboard(self, dashboard, dataset_id):
        possible_deleted_reports = []
        datasets_in_dashboard = list(set(dashboard.get_datasets_ids_in_dashboard()))
        if len(datasets_in_dashboard) == 1:
            possible_deleted_reports.append(dashboard)
        else:
            visuals = self._get_visuals_of_dataset_in_dashboard_light(dashboard, dataset_id)
            possible_deleted_reports += visuals

        return possible_deleted_reports

    def get_reports(self, report_ids: list[int]):
        """
        Retrieves reports with all the data that is sent to the client.
        This is quite a lot. Dedicated endpoints (e.g. possible_deleted_reports) could use a
        reduced set.
        """
        return Report.objects\
            .filter(id__in=report_ids)\
            .select_related('workspace')\
            .select_related('dataset__dataconnection')\
            .only(
                'report_name', 'report_description','report_type', 'create_time', 'create_user',
                'update_time', 'update_user', 'report_mode', 'report_data', 'shared_visual_dashboards',
                'dataset__dataset_name','dataset__dataset_detail', 'dataset__dataset_type', 'dataset__create_user',
                'dataset__dataconnection__dataconnection_name', 'dataset__dataconnection__dataconnection_type', 'dataset__dataconnection__create_user',
                'workspace__name', 'workspace__acl', 'workspace__description','workspace__editable', 'workspace__private_user_id'
            )

    @staticmethod
    def _get_visuals_of_dataset_in_dashboard(dashboard, dataset_id):
        (visual_ids, _) = dashboard.get_visuals_in_dashboard()
        visuals = Report.objects\
            .filter(id__in=visual_ids)\
            .filter(dataset_id=dataset_id)\
            .select_related('workspace')\
            .select_related('dataset')\
            .select_related('dataset__dataconnection')
        return list(visuals)
    
    @staticmethod
    def _get_visuals_of_dataset_in_dashboard_light(dashboard, dataset_id):
        (visual_ids, _) = dashboard.get_visuals_in_dashboard()
        visuals = Report.objects.filter(
            id__in=visual_ids,
            dataset_id=dataset_id
        ).values('id', 'report_type', 'report_name')

        return list(visuals)

    def handle_dataset_deletion(self, dataset, request=None):
        dashboards = self.get_dashboards(dataset.id)
        for dashboard in dashboards:
            datasets = list(set(dashboard.get_datasets_ids_in_dashboard()))
            if dashboard.dataset_id == dataset.id and len(datasets) > 1:
                datasets.remove(dataset.id)
                dashboard.dataset_id = sorted(datasets, key=datasets.count, reverse=True)[0]
            removing_candidates = self.__get_removing_candidates_in_dashboard(dashboard, dataset.id)
            for report in removing_candidates:
                if not isinstance(report, dict) and report.category() == 'dashboard':
                    visuals = self._get_visuals_of_dataset_in_dashboard(report, dataset.id)
                    for visual in visuals:
                        auditApp(visual, ActionType.DELETE, app_id=report.id, request=request)
                    auditApp(report, ActionType.DELETE, app_id=dashboard.id, request=request)

                else:
                    dcid, dctype, dcname = get_dc_info(dataset)
                    logAction(
                        action=ActionType.DELETE, 
                        component_type=ComponentType.APP,
                        app_id=dashboard.id,
                        visual_id=report['id'],
                        visual_type=report['report_type'],
                        visual_name=report['report_name'],
                        component_id=report['id'], 
                        dataset_id=dataset.id, 
                        dataset_name=dataset.dataset_name,
                        dataconnection_id=dcid,
                        dataconnection_type=dctype, 
                        dataconnection_name=dcname,
                        request=request
                    )

            dashboard.remove_visual(*removing_candidates, dataset_id=dataset.id)
            dashboard.save()

def init_associate_dashboards_to_datasets(apps:django.apps.registry.Apps=None):
    if apps:
        DataSet_historical = apps.get_model('datasets', 'dataset')
        Report_historical = apps.get_model('reports', 'report')
        _init_associate_dashboards_to_datasets(DataSet_historical, Report_historical)
    else:
        _init_associate_dashboards_to_datasets(DataSet, Report)

def _init_associate_dashboards_to_datasets(DataSet: DataSet, Report: Report):

    datasets_dashboards_dict = collections.defaultdict(list)

    dashboards = Report.objects.only(
        'report_data').filter(report_type='dashboard')

    # prefetch report dataset information
    report_dataset_data = Report.objects.values_list('id', 'dataset_id').exclude(report_type='dashboard')
    report_datasets = {report_id: dataset_id for report_id, dataset_id in report_dataset_data}

    for dashboard in dashboards:
        dashboard_data = json.loads(dashboard.report_data)

        (visuals, _) = get_visuals_picklist(dashboard_data)
        dashboard_datasets = {report_datasets.get(visual_id) for visual_id in visuals}

        for dataset in dashboard_datasets:
            datasets_dashboards_dict[dataset].append(dashboard.id)

    # we must update ALL datasets, not only the ones that have associated dashboards,
    # but also the ones that currently do not.
    datasets = DataSet.objects.only('id')

    # does not help, all fields are retrieved on save because of the invocation
    # of validateNoHtml() in DataSet.save()
    # (eventually one could fix DataSet.save() to consult update_fields)
    # datasets = datasets.only('dashboards')

    for ds in datasets:
        dashboards_ids = datasets_dashboards_dict[ds.id]
        dashboards_text = json.dumps(dashboards_ids)

        # because of foreign key validation, this save involves a select on the dataconnections table.
        # So saving involves two SQL statements.
        # ds.dashboards = dashboards_text
        # ds.save(update_fields=['dashboards'])

        # bulk update does not do foreign key validation, so we have one SQL less:
        DataSet.objects.filter(id=ds.id).update(dashboards=dashboards_text)
