# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from tests.migrationtests_base import MigrationsTestBase, delete_field
from datasets.migration_utils import table_column_exists
from vizlib.tests.test_report_operations import dataset_payload
from django.db.backends.base.base import BaseDatabaseWrapper

import django.db

connection: BaseDatabaseWrapper = django.db.connection

class TestMigrations(MigrationsTestBase):

  def test_0018(self):
    self.run_migration('datasets', '0017')
    tables_1 = set(self.introspection().table_names())

    self.run_migration('datasets', '0018')
    tables_2 = set(self.introspection().table_names())

    # no new tables:
    self.assertEqual(tables_1, tables_2)

    # just being paranoid: datasetsegments was NOT created:
    self.assertTableExists('datasets_datasetsegment', exists=False)

    # segment.dataset_id is present in the database
    self.assertTrue(table_column_exists(connection, 'datasets_segment', 'dataset_id'))

  def test_0021(self):

    self.run_migration('datasets', '0017')

    # segment.dataset_id is present in the database
    self.assertTrue(table_column_exists(connection, 'datasets_segment', 'dataset_id'))

    # polulate the database:
    DataConnection = self.project_state.apps.get_model('datasets', 'dataconnection')
    DataSet = self.project_state.apps.get_model('datasets', 'dataset')
    Segment = self.project_state.apps.get_model('datasets', 'segment')

    dc1 = DataConnection.objects.create(dataconnection_name='dc1')
    ds1 = DataSet.objects.create(dataset_name='ds1', dataconnection=dc1)
    ds2 = DataSet.objects.create(dataset_name='ds2', dataconnection=dc1)
    self.assertIsNotNone(ds1.pk)
    self.assertIsNotNone(ds2.pk)
    self.assertNotEqual(ds1.pk, ds2.pk)

    s1 = Segment.objects.create(dataset=ds1, segment_name='s1')
    s2 = Segment.objects.create(dataset=ds2, segment_name='s2')

    self.assertEqual(DataSet.objects.count(), 2)
    self.assertEqual(Segment.objects.count(), 2)

    self.run_migration('datasets', '0018')
    self.run_migration('datasets', '0019')
    DataSet = self.project_state.apps.get_model('datasets', 'dataset')
    ds1 = DataSet.objects.filter(dataset_name='ds1')[0]
    # verify new fields:
    self.assertIsNotNone(ds1.version_id)
    self.assertIsNotNone(ds1.version_group_id)
    self.assertIsNotNone(ds1.version_name)
    self.assertEqual(ds1.version_id, ds1.pk)
    self.assertEqual(ds1.version_group_id, ds1.pk)

    self.run_migration('datasets', '0020')
    # new table created:
    self.assertTableExists('datasets_datasetsegment', exists=True)
    # new model created:
    DataSegment = self.project_state.apps.get_model(
      'datasets', 'datasetsegment')
    self.assertEqual(DataSegment.objects.count(),2)
    dds1 = DataSegment.objects.filter(segment_id=s1.pk)[0]
    self.assertEquals(dds1.dataset_version_id, ds1.version_id)

    self.run_migration('datasets', '0021')
    # verify that the dataset field was removed from the model:
    Segment = self.project_state.apps.get_model('datasets', 'segment')
    segment_fields = [f.name for f in Segment._meta.fields]
    self.assertNotIn('dataset', segment_fields)

    # verify that segment.dataset_id is removed from the database:
    self.assertFalse(table_column_exists(connection, 'datasets_segment', 'dataset_id'))

  def test_0021_skipping(self):

    self.run_migration('datasets', '0018')

    # segment.dataset_id is present in the database
    self.assertTrue(table_column_exists(connection, 'datasets_segment', 'dataset_id'))

    # in versions prior to 7.2.7 the migration could actually succeed (on sqlite and mysql and also postgres, if the database was initially empty)
    # simulate that scenario by dropping the dataset_id field. In this case the migrations should all be skipped.

    Segment = self.project_state.apps.get_model('datasets','segment')
    delete_field(self.schema_editor(), Segment, field_name='dataset')

    # segment.dataset_id should be gone:
    self.assertFalse(table_column_exists(connection, 'datasets_segment', 'dataset_id'))

    # migrations 0019-0021 should be skipped now:
    self.run_migration('datasets', '0019')
    self.assertFalse(table_column_exists(connection, 'datasets_dataset', 'version_id'))

    self.run_migration('datasets', '0020')
    # no table created:
    self.assertTableExists('datasets_datasetsegment', exists=False)

    self.run_migration('datasets', '0021')
    # no table created:
    self.assertTableExists('datasets_datasetsegment', exists=False)

  def test_0022(self):
    self.run_migration('datasets', '0021')

    # dataset.dataset_tablenames is not present in the database
    self.assertFalse(table_column_exists(connection, 'datasets_dataset', 'dataset_tablenames'))

    # polulate the database:
    DataConnection = self.project_state.apps.get_model('datasets', 'dataconnection')
    DataSet = self.project_state.apps.get_model('datasets', 'dataset')

    dc1 = DataConnection.objects.create(dataconnection_name='dc1')

    dataset_types = ['singletable', 'multitable', 'singletable-sql', 'multitable-sql']
    for dataset_type in dataset_types:
      ds_payload = dataset_payload(ds_name=f'ds_{dataset_type}', dc_id=dc1, dataset_type=dataset_type)
      ds = DataSet.objects.create(dataset_name=ds_payload['dataset_name'], dataconnection=dc1,
                                  dataset_type=ds_payload['dataset_type'], dataset_info=ds_payload['dataset_info'],
                                  dataset_detail=ds_payload['dataset_detail'])
      self.assertIsNotNone(ds.pk)

    self.run_migration('datasets', '0022')
    DataSet = self.project_state.apps.get_model('datasets', 'dataset')

    # dataset.dataset_tablenames is present in the database
    self.assertTrue(table_column_exists(connection, 'datasets_dataset', 'dataset_tablenames'))

    expected_ds_tablenames = ['["default.table_1"]',
                              '["default.table_1", "default.table_2"]',
                              '["main.abc"]',
                              '["main.abc", "default.table_2"]'
                            ]
    for idx, dataset_type in enumerate(dataset_types):
      ds = DataSet.objects.filter(dataset_name=f'ds_{dataset_type}')[0]

      # verify dataset_tablenames value
      self.assertEqual(ds.dataset_tablenames, expected_ds_tablenames[idx])
