# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Models for datasets
"""

from util.string_operations import extract_tablenames_from_sqlquery
from django.db import models
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.utils.timezone import now

from arcwebbase.validator import validateNoHtml, validateStringNoHtml
import json
import uuid

from dataconn.connections_types import *
#from south.signals import post_migrate
from django.dispatch import receiver
from django.db.models.signals import post_save, pre_save

_POST_MIGRATION_FUNCTIONS = []

def enque_migration(f):
  if f not in _POST_MIGRATION_FUNCTIONS:
    _POST_MIGRATION_FUNCTIONS.append(f)

#@receiver(post_migrate)
def rx_post_migration(**kwargs):
  global _POST_MIGRATION_FUNCTIONS
  for f in _POST_MIGRATION_FUNCTIONS:
    f()

  _POST_MIGRATION_FUNCTIONS = []

class DataConnection(models.Model):
  """Data Connection"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(DataConnection, self).save(*args, **kwargs)

  # one of 'arcengine', 'hive', 'sqlite', 'postgres', 'mysql' etc.
  dataconnection_type = models.CharField(max_length=100, blank=True,
      validators=[validateStringNoHtml])

  dataconnection_name = models.CharField(max_length=1000, blank=True,
      validators=[validateStringNoHtml])
  dataconnection_description = models.TextField(max_length=4000, blank=True,
      validators=[validateStringNoHtml])

  # this information is never sent back to the client
  dataconnection_password = models.TextField(max_length=4000, null=True, blank=True,
      validators=[validateStringNoHtml])

  # json data for connection information
  dataconnection_info = models.TextField(max_length=4000, blank=True,
      validators=[validateStringNoHtml])

  # json data for settings
  dataconnection_settings = models.TextField(max_length=4000, blank=True,
      validators=[validateStringNoHtml])

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Cache sequence number
  cache_sequence = models.IntegerField(default=0)

  # It is used to distingush connections in MLExplore form-factor
  cml_id = models.BigIntegerField(default=None, null=True, blank=True)

  # Is the dataconnection in sync or not (used in the form factors)
  cml_status = models.TextField(default='{}')

  # Get the cache settings for this connection.
  def cache_settings(self):
    return json.loads(self.dataconnection_info or '{}').get('PARAMS', {}).get('CACHE', {})

  # Get the fallback settings for this connection.
  def fallback_settings(self):
    return json.loads(self.dataconnection_info or '{}').get('PARAMS', {}).get('FALLBACK', {})

  # Permission type for objects of this class
  @staticmethod
  def perm_type():
    return 'dataconn'

  # Check if user has given permission on connection
  def checkPerm(self, user, perm):
    from rbac.permcache import RBAC
    RBAC.checkPerm(user, perm, self)

  # Get all permissions a user has on this connection
  def getPerms(self, user):
    from rbac.permcache import RBAC
    return RBAC.get_user_obj_perms(user, self)

  # Set up initial cache settings based on default/settings values
  def migrate_cache(self):

    # A map of connection types to old style settings
    settings_map = {

      DATACONN_ARCENGINE : 'CACHE_RECHECK_SECONDS',
      DATACONN_SPARKSQL  : 'HIVE_CACHE_ENABLE',
      DATACONN_HIVE      : 'HIVE_CACHE_ENABLE',
      DATACONN_HYVE      : 'HIVE_CACHE_ENABLE',
      DATACONN_IMPALA    : 'IMPALA_CACHE_ENABLE',
      DATACONN_IMPYLA    : 'IMPALA_CACHE_ENABLE',
      DATACONN_REDSHIFT  : 'REDSHIFT_CACHE_ENABLE',
      DATACONN_ASTER     : 'ASTER_CACHE_ENABLE',
      DATACONN_SOLR      : 'SOLR_CACHE_ENABLE',
      DATACONN_DRILL     : 'DRILL_CACHE_ENABLE',
      DATACONN_ORACLE    : 'ORACLE_CACHE_ENABLE',
    }

    try:
      info = json.loads(self.dataconnection_info or '{}')
    except:
      info = {}

    params = info.get('PARAMS')
    if params is None:
      params = {}
      info['PARAMS'] = params

    dctype = self.dataconnection_type
    try:
      setting = int(getattr(settings, settings_map.get(dctype), 0))
    except:
      setting = 0

    if isArcEngineBased(dctype):
      cache = {'ENABLED':1, 'RETENTION':setting}
    else:
      cache = {'ENABLED':setting, 'RETENTION': (60 * 60 * 24)}

    params['CACHE'] = cache
    self.dataconnection_info = json.dumps(info)

# reset cached data connection on any update
@receiver(post_save, sender=DataConnection)
def dataconnection_post_save(sender, instance, **kwargs):
  if isinstance(instance, DataConnection):
    from dataconn.dataconn import del_dataconn
    del_dataconn(instance.id)

# Do the custom cache migration for data connection
def migrate_cache():
  for dc in DataConnection.objects.all():
    dc.migrate_cache()
    dc.save()

@receiver(pre_save, sender=DataConnection)
def check_dataconnection_name(sender, instance, **kwargs):
    counter = 0
    new_name = instance.dataconnection_name

    while DataConnection.objects.filter(dataconnection_name=new_name).exclude(pk=instance.pk).exists():
        counter += 1
        new_name = f"{instance.dataconnection_name}_{counter}"

    instance.dataconnection_name = new_name

class DataSet(models.Model):
  """Dataset"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(DataSet, self).save(*args, **kwargs)

  # dataconnection this dataset belongs to
  dataconnection = models.ForeignKey(DataConnection, on_delete=models.CASCADE, blank=True, null=True)

  dataset_name = models.CharField(max_length=1000,
      validators=[validateStringNoHtml])

  # One of 'singletable', 'customsql', 'multipletables'
  dataset_type = models.CharField(max_length=20,
      validators=[validateStringNoHtml])

  # Contents are interpreted differently based on 'dataset_type'
  # For a dataset_type of 'singletable', this is 'db.table'
  dataset_detail = models.TextField(max_length=20000,
      validators=[validateStringNoHtml])

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Description
  dataset_description = models.TextField(max_length=20000, blank=True,
      validators=[validateStringNoHtml])

  # json data
  dataset_info = models.TextField(max_length=20000, blank=True)

  # List of tablenames the dataset points at
  dataset_tablenames = models.TextField(max_length=10000, blank=True)

  # Dataset -> LogicalView (LV)
  dataset_lvinfo = models.TextField(max_length=20000, null=True, blank=True)

  # Permissions - default is that everyone had read-write access
  dataset_everyone_perm = models.CharField(max_length=20, default='readwrite',
      validators=[validateStringNoHtml])

  # Permissions - default is that dataset-creator has setupperm access
  dataset_create_user_perm = \
      models.CharField(max_length=20, default='setupperm',
          validators=[validateStringNoHtml])

  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

  # Cache sequence number
  cache_sequence = models.IntegerField(default=0)

  # A general place for settings (too many places assume dataset_info is
  # a list of tables, so we put settings here).
  dataset_settings = models.TextField(max_length=20000, default='{}')

  # Setting to mark if search is enabled
  search_enabled = models.BooleanField(default=False)

  # List of dashboards ids that are using dataset
  dashboards = models.TextField(max_length=32*1024, blank=True, default='[]')

  # In Version Control we set a unique id for every new version, to be able to handle segment and extract versioning
  version_id = models.IntegerField(editable=False, null=True)

  # In Version Control we set the id of the origin dataset as version_group_id
  version_group_id = models.IntegerField(default=None, blank=True, null=True)

  # Because of the Version Control we set (only) one of the versions active from the same version group
  is_active_version = models.BooleanField(default=True)

  # In Version Control we set the creation time as version name by default and then the user can rename it to a named version
  version_name = models.CharField(max_length=100, null=True, blank=True)

  is_named_version = models.BooleanField(default=False)

  segments = models.ManyToManyField('Segment', through='DatasetSegment', related_name='dataset')


  # Given a dataset, return the tablesnames the dataset points at
  def get_tablenames(self):
    if self.dataset_type == "singletable":
      if self.dataset_detail == "Created from SQL":
        tl = json.loads(self.dataset_info)
        return extract_tablenames_from_sqlquery(tl[0].get("select"))

      return [self.dataset_detail]

    tl = json.loads(self.dataset_info)
    tablenames = []
    for table in tl:
      if table.get("tablename") == "text.sql":
        tablenames.extend(extract_tablenames_from_sqlquery(table.get("select")))
      else:
        tablenames.append(table.get("tablename"))

    return tablenames

  # Given a dataset, return the tables and sql queries that it points at
  def datasetTables(self):
    if self.dataset_type == "singletable":
      if self.dataset_detail == "Created from SQL":
        tl = json.loads(self.dataset_info)
        return [tl[0].get("select")]

      return [self.dataset_detail]

    tl = json.loads(self.dataset_info)
    tablenames = []
    for table in tl:
      if table.get("tablename") == "text.sql":
        tablenames.append(table.get("select"))
      else:
        tablenames.append(table.get("tablename"))

    return tablenames

  def add_dashboard_to_dataset(self, dashboard_id):
      associated_dashboards = json.loads(self.dashboards)

      if dashboard_id in associated_dashboards:
          return

      associated_dashboards.append(dashboard_id)
      self.dashboards = associated_dashboards
      self.save()

  # Checks if the dashboard should be removed and removes if that is the case
  def remove_dashboard_from_dataset(self, dashboard):
      if not self.__should_remove_dashboard(dashboard):
          return

      self.__detaching_dashboard_from_dataset(self, dashboard)

  @staticmethod
  def remove_dashboard_from_all_datasets(dashboard):
      datasets = DataSet.objects.all()
      for dataset in datasets:
          DataSet.__detaching_dashboard_from_dataset(
              dataset, dashboard
          )

  # Removing the dashboard from the dashboards attached to a dataset
  @staticmethod
  def __detaching_dashboard_from_dataset(dataset, dashboard):
      associated_dashboards = json.loads(dataset.dashboards)

      if dashboard.id in associated_dashboards:
        associated_dashboards.remove(dashboard.id)

        dataset.dashboards = associated_dashboards
        dataset.save()

  def __should_remove_dashboard(self, dashboard):
      datasets = dashboard.get_datasets_ids_in_dashboard()
      datasets_occurrences = datasets.count(self.id)

      return datasets_occurrences <= 1

  # Get the cache settings for this dataset.  This can come from the
  # connection or from our own settings
  def cache_settings(self, dc=None):
    cache = json.loads(self.dataset_settings or '{}').get('CACHE', {})
    if not cache.get('OVERRIDE', 0):
      dc = dc or self.dataconnection
      cache = dc.cache_settings()
    return cache

  def user_has_perms(self, user_obj, perm):
    """ Check if the user has the specified permissions on the dataset """
    perm_code_name = "datasets.%s_%s_%s" % ("can", perm, self.id)
    return user_obj.has_perm(perm_code_name)

  def highest_user_perm(self, user_obj):
    """ Check if the user has the specified permissions on the dataset """
    # Return values can be none, setupperm, readwrite, readonly
    for perm in ('setupperm', 'readwrite', 'readonly'):
      if self.user_has_perms(user_obj, perm):
        return perm
      if self.dataset_everyone_perm == perm:
        return perm
      if self.create_user == user_obj.username:
        if self.dataset_create_user_perm == perm:
          return perm
    return 'none'

  # Permission type for objects of this class
  @staticmethod
  def perm_type():
    return 'dataset'

  # Check if user has given permission on connection
  def checkPerm(self, user, perm):
    from rbac.permcache import RBAC
    RBAC.checkPerm(user, perm, self)

  # Get all permissions a user has on this connection
  def getPerms(self, user):
    from rbac.permcache import RBAC
    return RBAC.get_user_obj_perms(user, self)

  # This only check if the user herself has the permission, does not
  # check if the everyone/creator flag values
  @staticmethod
  def user_has_perms_on_dataset(username, perm, dataset_id):
    """ Check if the user has the specified permissions on the dataset """

    # If no dataset id has been specified, then user has all permissions
    if dataset_id == '':
      return True

    perm_code_name = "datasets.%s_%s_%s" % ("can", perm, int(dataset_id))

    user_type = ContentType.objects.get(app_label="auth", model="user")
    user_obj = user_type.get_object_for_this_type(username=username)

    if user_obj.has_perm(perm_code_name):
      return True
    else:
      return False

  @staticmethod
  def highest_user_perm_on_dataset(username, dataset_id):
    """ Check if the user has the specified permissions on the dataset """
    # Return values can be none, setupperm, readwrite, readonly

    # If no dataset id has been specified, then give default of rw perm
    if dataset_id == '':
      return 'readwrite'

    ds = DataSet.objects.get(id=dataset_id)
    user_type = ContentType.objects.get(app_label="auth", model="user")
    user_obj = user_type.get_object_for_this_type(username=username)

    for perm in ('setupperm', 'readwrite', 'readonly'):
      if DataSet.user_has_perms_on_dataset(username, perm, dataset_id):
        return perm
      if ds.dataset_everyone_perm == perm:
        return perm
      if ds.create_user == username:
        if ds.dataset_create_user_perm == perm:
          return perm

    return 'none'

  @staticmethod
  def user_has_edit_perms_on_dataset(username, dataset_id):
    """ Check if user has RW perms or setupuser perms on the dataset"""
    if (DataSet.highest_user_perm_on_dataset(username, dataset_id) == 'readonly' or
       DataSet.highest_user_perm_on_dataset(username, dataset_id) == 'none'):
      return False
    else:
      return True

  @staticmethod
  def group_has_perms_on_dataset(gname, perm, dataset_id):
    """ Check if the group has the specified permissions on the dataset """

    # If no dataset id has been specified, then that group has all permissions
    if dataset_id == '':
      return True

    permission_code_name = "%s_%s_%s" % ("can", perm, int(dataset_id))

    group_type = ContentType.objects.get(app_label="auth", model="group")
    group_obj = group_type.get_object_for_this_type(name=gname)

    if group_obj.permissions.filter(codename=permission_code_name):
      return True
    else:
      return False

  @staticmethod
  def highest_group_perm_on_dataset(gname, dataset_id):
    """ Check if the user has the specified permissions on the dataset """
    # Return values can be none, setupperm, readwrite, readonly

    # If no dataset id has been specified, then group has default RW perms
    if dataset_id == '':
      return 'readwrite'

    for perm in ('setupperm', 'readwrite', 'readonly'):
      if DataSet.group_has_perms_on_dataset(gname, perm, dataset_id):
        return perm

    return 'none'

@receiver(post_save, sender=DataSet)
def set_version_id(sender, instance, created, **kwargs):
  if created:
    instance.version_name = instance.create_time
    instance.version_id = instance.id
    instance.version_group_id = instance.version_group_id or instance.id
    instance.save(update_fields=['version_id', 'version_group_id', 'version_name'])

class Segment(models.Model):
  """Segments on datasets"""
  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(Segment, self).save(*args, **kwargs)

  # segment name
  segment_name = models.CharField(max_length=1000, null=True, blank=True,
      validators=[validateStringNoHtml])

  # json structure describing the segment
  segment_data = models.TextField(max_length=32*1024, null=True, blank=True)

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

class DatasetSegment(models.Model):
  dataset = models.ForeignKey(DataSet, on_delete=models.CASCADE)
  segment = models.ForeignKey(Segment, on_delete=models.CASCADE)
  dataset_version_id = models.IntegerField(null=True)

  class Meta:
    unique_together = ('dataset', 'segment', 'dataset_version_id')


class Event(models.Model):
  """Events on dataset"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(Event, self).save(*args, **kwargs)

  # dataset this event is based on
  dataset = models.ForeignKey(DataSet, on_delete=models.CASCADE, blank=True, null=True)

  # segment name
  event_name = models.CharField(max_length=1000, null=True, blank=True,
      validators=[validateStringNoHtml])

  # json structure describing the segment
  event_data = models.TextField(max_length=32*1024, null=True, blank=True)

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

class FilterAssociation(models.Model):
  """ Filter Associations on datasets """

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(FilterAssociation, self).save(*args, **kwargs)

  # dataset this filter assocation is based on
  dataset = models.ForeignKey(DataSet, on_delete=models.CASCADE, blank=True, related_name='filterassociations', null=True)


  # filter association name (this is the name of the underlying view)
  filterassociation_name = models.CharField(max_length=1000, null=True, blank=True,
      validators=[validateStringNoHtml])

  # json structure describing the filter association  (including segments)
  filterassociation_data = models.TextField(max_length=32*1024, null=True, blank=True,
      validators=[validateStringNoHtml])

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True,blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

  def user_has_filter_perms(self, user_obj):
    """ Check if the user has the specified permissions on the dataset """

    perm_code_name = "datasets.%s_%s%s_%s%s" % ("applied", "dataset", self.dataset.id, "fa", self.id)
    # Super user does not get assigned any filter association permissions
    if user_obj.is_superuser:
      return False
    # This will also check if any group that this user belongs to has this perm
    return user_obj.has_perm(perm_code_name)

  def user_not_group_has_filter_perms(self, user_obj):
    """ Check if user, and not the group has specified FA's on the dataset """

    perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", self.dataset.id, "fa", self.id)
    # Super user does not get assigned any filter association permissions
    if user_obj.is_superuser:
      return False
    # This will only check if that user, not any group that it belongs to,
    # has that permission
    return user_obj.user_permissions.filter(codename=perm_code_name).exists()



  @staticmethod
  def user_has_filter_perms_on_dataset(filterassociation_id, username, dataset_id, user_obj=None):
    """ Check if the user has the specified permissions on the dataset """

    # If no dataset id has been specified, then user has all permissions
    if dataset_id == '':
      return True

    perm_code_name = "datasets.%s_%s%s_%s%s" % ("applied", "dataset", int(dataset_id), "fa", filterassociation_id)

    if user_obj is None:
      user_type = ContentType.objects.get(app_label="auth", model="user")
      user_obj = user_type.get_object_for_this_type(username=username)

    # Super user does not get assigned any filter association permissions
    if user_obj.is_superuser:
      return False

    # This will also check if any group that this user belongs to has this perm
    if user_obj.has_perm(perm_code_name):
      return True
    else:
      return False

  @staticmethod
  def user_not_group_has_filter_perms_on_dataset(filterassociation_id,
                              username, dataset_id, user_obj=None):
    """ Check if user, and not the group has specified FA's on the dataset """

    # If no dataset id has been specified, then user has all permissions
    if dataset_id == '':
      return True

    perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", int(dataset_id), "fa", filterassociation_id)

    if user_obj is None:
      user_type = ContentType.objects.get(app_label="auth", model="user")
      user_obj = user_type.get_object_for_this_type(username=username)

    # Super user does not get assigned any filter association permissions
    if user_obj.is_superuser:
      return False

    # This will only check if that user, not any group that it belongs to,
    # has that permission
    if not hasattr(user_obj, '_arc_user_perm_cache'):
      user_obj._arc_user_perm_cache = set(user_obj.user_permissions.all().values_list('codename', flat=True))
    return perm_code_name in user_obj._arc_user_perm_cache

  @staticmethod
  def group_has_filter_perms_on_dataset(filterassociation_id,
                                 gname, dataset_id):
    """ Check if the group has the specified FA on the dataset """

    # If no dataset id has been specified, then that group has all permissions
    if dataset_id == '':
      return True

    perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", int(dataset_id), "fa", filterassociation_id)

    group_type = ContentType.objects.get(app_label="auth", model="group")
    group_obj = group_type.get_object_for_this_type(name=gname)

    # Special group 'everyone' has no filter associations
    if gname == 'everyone':
      return False

    if group_obj.permissions.filter(codename=perm_code_name):
      return True
    else:
      return False

class Extract(models.Model):
  """Data Extracts"""

  # dataset this extract is based on
  srcDataset = models.ForeignKey(DataSet, on_delete=models.CASCADE)

  # dataconnection this extract is saved to
  tgtDataconnection = models.ForeignKey(DataConnection, on_delete=models.PROTECT,
      null=True, blank=True)

  # dbname and tablename for the target table
  tgtDbname = models.CharField(max_length=1000, blank=True, validators=[validateStringNoHtml])
  tgtTablename = models.CharField(max_length=1000, blank=True, validators=[validateStringNoHtml])

  # Type of extract - current options are refresh_full and refresh_partitioned
  extract_type = models.CharField(max_length=100, blank=True, validators=[validateStringNoHtml])

  # extract config
  extract_config = models.TextField(max_length=4*1024, blank=True)

  # extract runstate
  extract_runstate = models.TextField(max_length=32*1024, blank=True)

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  def jobName(self):
    return f"Data Extract: {self.id}"
