# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from rbac.models import Role
import datetime
import django.db
import django.test
import json

from apps.tests.userrole import base
from apps.tests.tests_views_visuals_page import grant_read_access_on_dataset
from dataconn import dataconn
from datasets import views_connections
from datasets.models import DataConnection
from django.contrib.auth.models import User
from django.urls import reverse
from sqlrun.views_cache_statistic import CacheStatistics
from unittest import mock
from usage.views import create_usage_dataconnection, USAGE_NAME
from util.tests.test_get_dataset_ids_accessible_by_user import grant_explore_on_dataconnection
from vizlib.tests import test_report_operations
from vizlib.tests.util import run_with_db_query_check


class TestDataconnection(django.test.TestCase):
    def setUp(self):
        admin = base.make_user('admin1', is_superuser=True)
        admin_client = base.create_client(admin)
        self.admin_client = admin_client
        self.util = test_report_operations.TestDashboardOperations()
        self.dc_1_id = self._create_dataconnection(admin_client, 'DC1')

    @mock.patch('dataconn.dataconn._build_dataconn', wraps=getattr(dataconn, '_build_dataconn'))
    def test_dataconnection_get(self, build_dataconn: mock.Mock):
        dataconn_list, db_query_count_1 = self._dataconnection_get()
        self.assertEquals(len(dataconn_list), 1)

        # create a second dataconnection:
        self._create_dataconnection(self.admin_client, 'DC2')
        dataconn_list, db_query_count_2 = self._dataconnection_get()
        self.assertEquals(len(dataconn_list), 2)

        # DB queries must not increase
        self.assertEquals(db_query_count_2, db_query_count_1)

        # dataconnection objects must not be built:
        self.assertEquals(build_dataconn.call_count, 0)

    def test_dataconnection_get_by_id(self):

        # create a second dataconnection:
        dc2_id = self._create_dataconnection(self.admin_client, 'DC2')

        dataconn_list, _ = self._dataconnection_get()
        self.assertEquals(len(dataconn_list), 2)

        dataconn_list = self._dataconnection_get_by_id(dc2_id)
        self.assertEquals(len(dataconn_list), 1)

    def test_used_time_retrieval(self):
        stats = CacheStatistics.get_by_dataconnection(self.dc_1_id)
        update_time = datetime.datetime.now()
        stats.update_time = update_time
        stats.save()

        dc_1 = DataConnection.objects.prefetch_related(
            'cachestatisticmodel_set').get(id=self.dc_1_id)

        result, db_query_count = run_with_db_query_check(
            lambda: views_connections.get_last_used_time(dc_1))
        self.assertEquals(
            result, stats.update_time.isoformat(timespec='milliseconds'))
        # prefetched data must be used
        self.assertEquals(db_query_count, 0)

    def test_used_time_retrieval_no_stats(self):

        dc_1 = DataConnection.objects.prefetch_related(
            'cachestatisticmodel_set').get(id=self.dc_1_id)

        result, db_query_count = run_with_db_query_check(
            lambda: views_connections.get_last_used_time(dc_1))
        self.assertEquals(
            result, dc_1.update_time.isoformat(timespec='milliseconds'))
        # prefetched data must be used
        self.assertEquals(db_query_count, 0)

    def test_when_no_permission_user_get_should_return_zero(self):
        # Given
        no_permission_user = base.make_user('no-permission-user')
        no_permission_client = base.create_client(no_permission_user)
        # When
        response, _ = run_with_db_query_check(
            lambda: no_permission_client.get(
                '/datasets/dataconnection?import_data=true'))

        self.assertEquals(response.status_code, 200)
        # Then
        self.assertEquals(len(response.json()), 0)

    def test_when_user_with_permission_for_dc_get_should_return_one(self):
        # Given: The user has permission for only one dc
        # create a second dataconnection:
        self._create_dataconnection(self.admin_client, 'DC2')
        dataconn_list, _ = self._dataconnection_get()
        self.assertEquals(len(dataconn_list), 2)

        # Create role for DC2
        role = Role()
        role.name = "DC2-only"
        role.privileges = json.dumps([{"ptype": "dataconn",
                                       "dclist": ["2"],
                                       "perms": ["dc_upload",
                                                 "dc_explore",
                                                 "dc_aviews"]}])
        role.save()

        # user with permission for only DC2
        some_permission_user = base.make_user('no-permission-user')
        some_permission_client = base.create_client(some_permission_user)
        role.add_users([some_permission_user.username])
        # When
        response, query_count_1 = run_with_db_query_check(
            lambda: some_permission_client.get(
                '/datasets/dataconnection?import_data=true'))

        self.assertEquals(response.status_code, 200)
        # Then
        self.assertEquals(len(response.json()), 1)
        self.assertEquals(response.json()[0]['name'], 'DC2')

        # create a third dataconnection for which the user does not have access
        create_usage_dataconnection()
        # self._create_dataconnection(self.admin_client, 'DC3')
        response, query_count_2 = run_with_db_query_check(
            lambda: some_permission_client.get(
                '/datasets/dataconnection?import_data=true'),
            asserter=self,
            expected_query_count=query_count_1 # query count must not change
            )

        self.assertEquals(response.status_code, 200)
        self.assertEquals(len(response.json()), 1)
        self.assertEqual(query_count_1, query_count_2, 'query count must not change')

        # viewing the usage dataconnection requires sys_viewlogs:
        self.grant_viewlogs_permission(some_permission_user.username)

        response, query_count_3 = run_with_db_query_check(
            lambda: some_permission_client.get(
                '/datasets/dataconnection?import_data=true'))

        self.assertEquals(response.status_code, 200)
        dc_names = {dc['name'] for dc in response.json()}
        self.assertEqual(len(dc_names), 2) # usage database should show up
        self.assertEqual(dc_names, set(['DC2', USAGE_NAME]))
        self.assertEqual(query_count_2, query_count_3, 'query count must not change')

    def grant_viewlogs_permission(self, username: str):
        role = Role()
        role.name = f"sys_viewlog_for_{username}"
        role.privileges = json.dumps([
                {"ptype": "system", "perms": ["sys_viewlogs"]}
            ])
        role.save()
        role.add_users([username])

    def _create_dataconnection(self, client: django.test.Client, dc_name: str):
        dc_payload = {
            'dataconnection_type': 'arcengine',
            'dataconnection_info': '{"PARAMS":{"HOST":"10.0.0.211","PORT":"31050"}}',
            'dataconnection_name': dc_name,
        }
        return self.util.create_dataconnection_with_payload(client, dc_payload)

    def _dataconnection_get(self) -> tuple[dict, int]:
        response, query_count = run_with_db_query_check(
            lambda: self.admin_client.get(
                '/datasets/dataconnection?import_data=true'))

        self.assertEquals(response.status_code, 200)
        return response.json(), query_count

    def _dataconnection_get_by_id(self, dc_id: int):
        response, _ = run_with_db_query_check(
            lambda: self.admin_client.get(
                f'/datasets/dataconnection/{dc_id}'))

        self.assertEquals(response.status_code, 200)
        return response.json()
    
    def _clone_connection(self, client, dataconnection_id) -> int:
        url = reverse('clone-dataconnection-id', args=[dataconnection_id])
        response = client.post(url)
        self.assertEquals(response.status_code, 200)
        return response.json()
    
    def test_dataconnection_edition_with_existing_name(self):
        payload_update = {
          'dataconnection_type': 'arcengine',
          'dataconnection_info': '{"PARAMS":{"HOST":"10.0.0.211","PORT":"31050"}}',
          'dataconnection_name': 'DC',
        }

        dc_id = self._create_dataconnection(self.admin_client, 'DC')
        dc2_id = self._create_dataconnection(self.admin_client, 'DC2')
        self.util.edit_dataconnection_with_payload(self.admin_client, payload_update, dc2_id)

        dc2 = self._dataconnection_get_by_id(dc2_id)[0]
        self.assertEquals(dc2['name'], 'DC_1')

        dc3_id = self._create_dataconnection(self.admin_client, 'DC3')
        self.util.edit_dataconnection_with_payload(self.admin_client, payload_update, dc3_id)

        dc3 = self._dataconnection_get_by_id(dc3_id)[0]
        self.assertEquals(dc3['name'], 'DC_2')

    def test_dataconnection_name_uniqueness_when_cloning(self):
        dc_id = self._create_dataconnection(self.admin_client, 'DC')
        dc_clone1 = self._clone_connection(self.admin_client, dc_id)
        self.assertEquals(dc_clone1['name'], 'Clone of DC')

        dc_clone2 = self._clone_connection(self.admin_client, dc_id)
        self.assertEquals(dc_clone2['name'], 'Clone of DC_1')

        dc_clone3 = self._clone_connection(self.admin_client, dc_id)
        self.assertEquals(dc_clone3['name'], 'Clone of DC_2')

class TestDataconnectionAccess(django.test.TestCase):

  def setUp(self):
    super().setUp()
    util = test_report_operations.TestDashboardOperations()
    self.util = util

    # create admin
    admin = base.make_user('admin', is_superuser=True)
    admin_client = base.create_client(admin)
    self.admin = admin
    self.admin_client = admin_client

    # create user1
    user1 = base.make_user('user1', is_superuser=False)
    user1_client = base.create_client(user1)
    self.user1 = user1
    self.user1_client = user1_client

    dc_id1 = util.create_dataconnection(admin_client)
    self.dc_id1 = dc_id1

    ds1_id = util.create_dataset(admin_client, dc_id1, 'ds1')
    self.ds1_id = ds1_id

  def get_dc_ids_through_datasets(self, user: User, expected_query_count: int=None):

    # groups are cached on the user object. Clear them:
    if hasattr(user, '_arcadia_groupset'):
      del user._arcadia_groupset

    result, _ = run_with_db_query_check(
      lambda: views_connections.dc_ids_accessible_through_datasets(user),
      asserter=self,
      expected_query_count=expected_query_count
    )
    return result

  def test_admin(self):
    """
    dc id collection is used for permission-based filtering.
    It is completely OK to return None for admin users, as those must not have their
    dataconnections filtered
    """
    self.assertIsNone(self.get_dc_ids_through_datasets(self.admin, expected_query_count=0))

  def test_user_dataconnections_visible_through_datasets(self):

    expected_dc_ids = set() # user has no visible datasets, so no connections:
    actual_ds_ids = self.get_dc_ids_through_datasets(self.user1, expected_query_count=3)
    self.assertEqual(actual_ds_ids, expected_dc_ids)

    # grant permissions on a dataconnection so that the user can create datasets:
    dc_id2 = self.util.create_dataconnection(self.admin_client, 'DC 2')
    grant_explore_on_dataconnection(self.user1, dc_id2)

    # create a dataset:
    ds_1 = self.util.create_dataset(self.user1_client, dc_id2, 'User-created dataset - 1')

    expected_dc_ids = set([dc_id2])
    actual_ds_ids = self.get_dc_ids_through_datasets(self.user1, expected_query_count=3)
    self.assertEqual(actual_ds_ids, expected_dc_ids)

    # create a second dataset to see if query count is constant
    ds_2 = self.util.create_dataset(self.user1_client, dc_id2, 'User-created dataset - 2')
    expected_dc_ids = set([dc_id2])
    actual_ds_ids = self.get_dc_ids_through_datasets(self.user1, expected_query_count=3)
    self.assertEqual(actual_ds_ids, expected_dc_ids)


    # grant access to the admin-created dataset
    grant_read_access_on_dataset(self.user1, self.dc_id1, self.ds1_id)
    expected_ds_ids = {self.dc_id1, dc_id2}
    actual_ds_ids = self.get_dc_ids_through_datasets(self.user1, expected_query_count=3)
    self.assertEqual(actual_ds_ids, expected_ds_ids)