from django.test import TestCase
from datasets.models import DataSet, DataConnection
from reports.models import Report
from datasets.dashboard_association import DashboardAssociation, \
    init_associate_dashboards_to_datasets
from vizlib.tests.util import run_with_db_query_check
from workspace.models import Workspace
from json import dumps, loads




def dashboard_data(visual_ids: list[int]):
    widgets = [
        {
            "col": 1,
            "row": 1,
            "size_x": 32,
            "size_y": 20,
            "id": f"uri-1-widget-{visual_id}"
        } for visual_id in visual_ids
    ]

    report_data = {
        "dashboard_sheets": [
            {
                "order": 1,
                "sheet_id": 1,
                "sheet_handle_title":
                "Sheet 1",
                "behaviors": {},
                "visual_widgets": widgets,
                "control_widgets": []
             }
        ]
    }

    return dumps(report_data)
class DashboardAssociationTest(TestCase):
    def setUp(self):
        self.dc = DataConnection.objects.create()

    def run_with_db_query_check(self, action: callable, expected_query_count:int=None):
        return run_with_db_query_check(action, asserter=self, expected_query_count=expected_query_count)

    def test_when_associate_should_add_to_existing_associated_dashboards(self):
        # Given
        ds = DataSet.objects.create(dataset_name='dataset1', dataconnection=self.dc,
                                    dataset_detail='test', dataset_type='test', dashboards='[10,11]')
        ws = Workspace.objects.create()
        table1 = Report.objects.create(report_type='table', dataset=ds,
                              workspace=ws, report_data='{}')
        r = Report.objects.create(workspace=ws, dataset=ds, report_type='dashboard',
                                  report_data=dashboard_data([table1.id]))
        # When
        self.run_with_db_query_check(
            lambda: DashboardAssociation().associate(ds.id, [r.id])
        )
        # Then
        new_ds = DataSet.objects.get(id=ds.id)
        self.assertEqual(loads(new_ds.dashboards), [10, 11, r.id])

    def test_when_init_associate_and_wrong_dashboards_should_overwrite_correctly(self):
        # Given
        ds = DataSet.objects.create(dataset_name='dataset1', dataconnection=self.dc,
                                    dataset_detail='test', dataset_type='test',
                                    dashboards='[10,11]')
        ds2 = DataSet.objects.create(dataset_name='ds2', dataconnection=self.dc,
                                     dataset_detail='tst', dataset_type='test',
                                     dashboards='[]')

        ws = Workspace.objects.create()
        table_1 = Report.objects.create(report_type='table', dataset=ds,
                              workspace=ws, report_data='{}')
        table_2 = Report.objects.create(report_type='table', dataset=ds2,
                              workspace=ws, report_data='{}')

        r = Report.objects.create(workspace=ws, dataset=ds,
                                  report_type='dashboard',
                                  report_data=dashboard_data([table_1.id, table_2.id]))
        # When
        self.run_with_db_query_check(init_associate_dashboards_to_datasets, expected_query_count=5)
        # Then
        ds.refresh_from_db()
        ds2.refresh_from_db()
        self.assertEqual(loads(ds.dashboards), [r.id])
        self.assertEqual(loads(ds2.dashboards), [r.id])

        ds3 = DataSet.objects.create(dataset_name='ds3', dataconnection=self.dc,
                            dataset_detail='tst', dataset_type='test',
                            dashboards='[10,11]')

        # there is 1 update per dataset:
        self.run_with_db_query_check(init_associate_dashboards_to_datasets, expected_query_count=5 + 1)
        ds3.refresh_from_db()
        # no visuals were created for ds3:
        self.assertEqual(loads(ds3.dashboards), [])

        # create a new dashboard to test dependence on dashboard count:

        table_1_2 = Report.objects.create(report_type='table', dataset=ds,
                              workspace=ws, report_data='{}')

        r_3 = Report.objects.create(workspace=ws, dataset=ds,
                                  report_type='dashboard',
                                  report_data=dashboard_data([table_1_2.id]))

        # same query count as before:
        self.run_with_db_query_check(init_associate_dashboards_to_datasets, expected_query_count=5 + 1)
        ds.refresh_from_db()
        ds2.refresh_from_db()
        ds3.refresh_from_db()
        self.assertEqual(set(loads(ds.dashboards)), set([r.id, r_3.id]))
        self.assertEqual(loads(ds2.dashboards), [r.id])
        self.assertEqual(loads(ds3.dashboards), [])

    def test_when_init_associate_and_correct_dashboards_should_overwrite_correctly(self):
        # Given
        ds = DataSet.objects.create(dataset_name='dataset1', dataconnection=self.dc,
                                    dataset_detail='test', dataset_type='test',
                                    dashboards='[3]')
        ds2 = DataSet.objects.create(dataset_name='ds2', dataconnection=self.dc,
                                     dataset_detail='tst', dataset_type='test',
                                     dashboards='[3]')

        ws = Workspace.objects.create()
        table_1 = Report.objects.create(report_type='table', dataset=ds,
                              workspace=ws, report_data='{}')
        table_2 = Report.objects.create(report_type='table', dataset=ds2,
                              workspace=ws, report_data='{}')

        r = Report.objects.create(workspace=ws, dataset=ds,
                                  report_type='dashboard',
                                  report_data=dashboard_data([table_1.id, table_2.id]))
        # When
        init_associate_dashboards_to_datasets()
        # Then
        ds.refresh_from_db()
        ds2.refresh_from_db()
        self.assertEqual(loads(ds.dashboards), [r.id])
        self.assertEqual(loads(ds2.dashboards), [r.id])
