# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import unittest
import json
from django.contrib.auth.models import User, Group, Permission
from django.http import HttpRequest

from . import views
from datasets.models import DataSet, Segment, FilterAssociation, DataConnection

class MockGroup(object):
  def __init__(self, name=None):
    self.name = name

  def values_list(self, attr, flat=False):
    return []

class MockUser(object):
  def __init__(self, username=None, is_superuser=True):
    self.username = username
    self.is_superuser = is_superuser
    self.id = 10000
    self.groups = MockGroup()
    self.is_authenticated = True

  def has_perm(self, perm):
    return True

class MockRequest(object):
  def __init__(self, user, POST):
    self.user = user
    self.POST = POST
    self.META = {
      "REMOTE_ADDR": "1.2.3.4"
    }

class TestBasic(unittest.TestCase):
  def _create_dataconnection(self):
    dc = DataConnection(dataconnection_type="sqlite",
    dataconnection_info = '{"PARAMS": {"FOR_TESTING": "true", "DATABASE": "tmp", "FILE": "/tmp/sqlite3_test_file", "HOST": "unknown", "PORT": 99, "AUTH": "NoSasl"}}')
    dc.save()
    return dc

  def test_clone_dataset(self):
    print("Verifying dataset clone")

    # 0. create a dataconnection, users
    dc = self._create_dataconnection()
    users = []
    for i in range(0, 4):
      user = User(username=f"user_dsclone_{i+1}")
      user.save()
      users.append(user)
      if i == 0:
        self.assertGreater(user.id, 0)
      else:
        self.assertEquals(user.id, users[0].id + i)

    groups = []
    for i in range(0, 4):
      group = Group(name=f"group_dsclone_{i+1}")
      group.save()
      groups.append(group)
      if i == 0:
        self.assertGreater(group.id, 0)
      else:
        self.assertEquals(group.id, groups[0].id + i)

    # 1. create a dataset
    d = {
      "dataconnection_id": dc.id,
      "dataset_name": "test dataset",
      "dataset_type": "singletable",
      "dataset_info": "abc",
      "dataset_detail": "def"
    }

    user = MockUser('testuser', is_superuser=True)
    request = HttpRequest()
    request.user = user
    ds_id = views.dataset_create_or_clone(d, request)
    self.assertGreater(ds_id, 0)

    # 2. add 2 segments
    POST = {
      "dataset_id": ds_id,
      "name": "segment name",
      "data": "segment data"
    }

    request = MockRequest(user, POST)
    response = views._segment_create(request)
    j = json.loads(response.content)
    self.assertGreater(j[0]["id"], 0)
    segment_id = j[0]["id"]
    self.assertEquals(j[0]["name"], "segment name")
    self.assertEquals(j[0]["data"], "segment data")

    POST = {
      "dataset_id": ds_id,
      "name": "segment name 2",
      "data": "segment data 2"
    }

    request = MockRequest(user, POST)
    response = views._segment_create(request)
    j = json.loads(response.content)
    self.assertEquals(j[0]["id"], segment_id + 1)

    # 3. verify segments are as expected
    d = DataSet.objects.get(id=ds_id)
    sl = Segment.objects.filter(datasetsegment__dataset_id=d.id, datasetsegment__dataset_version_id=d.version_id)
    self.assertEquals(len(sl), 2)
    ids = [s.id for s in sl]
    ids.sort()
    self.assertEquals(ids[0], segment_id)
    self.assertEquals(ids[1], segment_id + 1)

    # 4. create 2 filter associations
    d = {
      "dataset_id": ds_id,
      "name": "fa name",
      "data": "fa data",
      "users": f"[{users[0].id}, {users[1].id}]",
      "groups": f"[{groups[0].id}, {groups[1].id}]"
    }
    response = views._filterassociation_create(d, request=request)
    j = json.loads(response.content)
    self.assertGreater(j[0]["id"], 0)
    fa_id = j[0]["id"]
    self.assertEquals(j[0]["name"], "fa name")
    self.assertEquals(j[0]["data"], "fa data")

    d = {
      "dataset_id": ds_id,
      "name": "fa name 2",
      "data": "fa data 2",
      "users": f"[{users[2].id}, {users[3].id}]",
      "groups": f"[{groups[2].id}, {groups[3].id}]",
      "version_id": ds_id
    }
    response = views._filterassociation_create(d, request=request)
    j = json.loads(response.content)
    self.assertEquals(j[0]["id"], fa_id + 1)

    # 5. verify filter associations are as expected
    fal = FilterAssociation.objects.filter(dataset__id=ds_id)
    self.assertEquals(len(fal), 2)
    ids = [fa.id for fa in fal]
    ids.sort()
    self.assertEquals(ids[0], fa_id)
    self.assertEquals(ids[1], fa_id + 1)

    # 6. clone the dataset
    clone_d = {
      "dataset_id": ds_id
    }

    dsclone_id = views.dataset_create_or_clone(clone_d, request=request)

    # 7. verify the dataset and the segments
    self.assertEquals(dsclone_id, ds_id + 1)

    # 8. Both the old dataset and the new dataset should have 2 segments
    # segments ids should be as expected

    sl = Segment.objects.filter(datasetsegment__dataset_id=ds_id, datasetsegment__dataset_version_id=d['version_id'])
    self.assertEquals(len(sl), 2)
    ids = [s.id for s in sl]
    ids.sort()
    self.assertEquals(ids[0], segment_id)
    self.assertEquals(ids[1], segment_id + 1)

    d_clone = DataSet.objects.get(id=dsclone_id)
    sl = Segment.objects.filter(datasetsegment__dataset_id=d_clone.id, datasetsegment__dataset_version_id=d_clone.version_id)
    self.assertEquals(len(sl), 2)
    ids = [s.id for s in sl]
    ids.sort()
    self.assertEquals(ids[0], segment_id + 2)
    self.assertEquals(ids[1], segment_id + 3)

    # 9. Both the old dataset and the new dataset should have 2 FAs
    # FA ids should be as expected
    fal = FilterAssociation.objects.filter(dataset__id=ds_id)
    self.assertEquals(len(fal), 2)
    ids = [fa.id for fa in fal]
    ids.sort()
    self.assertEquals(ids[0], fa_id)
    self.assertEquals(ids[1], fa_id + 1)

    for fa in fal:
      (fa_users, fa_groups, _) = views._filterassociation_members(fa)
      fa_users = list(fa_users)
      fa_users.sort()

      if fa.id == fa_id or fa.id == fa_id + 2:
        self.assertEquals(fa_users[0], users[0].id)
        self.assertEquals(fa_users[1], users[1].id)

      fa_groups = list(fa_groups)
      fa_groups.sort()
      if fa.id == fa_id + 1 or fa.id == fa_id + 3:
        self.assertEquals(fa_groups[0], groups[2].id)
        self.assertEquals(fa_groups[1], groups[3].id)

    fal = FilterAssociation.objects.filter(dataset__id=dsclone_id)
    self.assertEquals(len(fal), 2)
    ids = [fa.id for fa in fal]
    ids.sort()
    self.assertEquals(ids[0], fa_id + 2)
    self.assertEquals(ids[1], fa_id + 3)

    for fa in fal:
      (fa_users, fa_groups, _) = views._filterassociation_members(fa)
      fa_users = list(fa_users)
      fa_users.sort()

      if fa.id == fa_id or fa.id == fa_id + 2:
        self.assertEquals(fa_users[0], users[0].id)
        self.assertEquals(fa_users[1], users[1].id)

      fa_groups = list(fa_groups)
      fa_groups.sort()
      if fa.id == fa_id + 1 or fa.id == fa_id + 3:
        self.assertEquals(fa_groups[0], groups[2].id)
        self.assertEquals(fa_groups[1], groups[3].id)

