# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import django.test
import django.db
import json
import time

from datasets.models import DataSet
from rbac.models import Role
from reports.models import Report
from util.ff_flipper import FeatureFlipper
from vizlib.tests import test_report_operations
from vizlib.tests.test_report_count import full_access_only_one_dataset_role
from vizlib.tests.test_report_operations import make_user, create_client
from vizlib.tests.test_report_count import run_with_db_query_check
from vizlib.tests.test_sqlrun import Common
from datasets.dashboard_association import DashboardAssociation

def enable_dataset_versioning():
  FeatureFlipper().set_flipper('enable_version_control', True, 'DB')
  FeatureFlipper().set_flipper('enable_dataset_version_control', True, 'DB')

class TestDatasetDeleting(django.test.TestCase):

  def test_delete_dataset_query_count_is_stable(self):
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    util = test_report_operations.TestDashboardOperations()

    dc_id = util.create_dataconnection(admin_client)
    ds_1 = util.create_dataset(admin_client, dc_id, 'ds_1')
    ds_2 = util.create_dataset(admin_client, dc_id, 'ds_2')

    visual_1 = util.create_visual(admin_client, ds_1)
    visual_2 = util.create_visual(admin_client, ds_2)
    visual_3 = util.create_visual(admin_client, ds_2)

    dashboard_id = util.create_dashboard(
      admin_client, ds_1, [visual_1, visual_2, visual_3])

    response, _ = run_with_db_query_check(lambda: admin_client.delete(
      f'/datasets/dataset/{ds_1}'), expected_query_count=34, asserter=self)
    self.assertEqual(response.status_code, 200)

  def test_delete_with_broken_dashboard(self):
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    util = test_report_operations.TestDashboardOperations()

    dc_id = util.create_dataconnection(admin_client)
    ds_1 = util.create_dataset(admin_client, dc_id, 'ds_1')
    ds_2 = util.create_dataset(admin_client, dc_id, 'ds_2')

    visual_1 = util.create_visual(admin_client, ds_1)
    visual_2 = util.create_visual(admin_client, ds_2)
    visual_3 = util.create_visual(admin_client, ds_2)

    dashboard_id = util.create_dashboard(
      admin_client, ds_1, [visual_1, visual_2, visual_3])

    # break the dashboard:
    Report.objects.get(id=visual_3).delete()

    payload = {
      'dataset_id': ds_1
    }
    response, _ = run_with_db_query_check(lambda: admin_client.get(
      '/datasets/possible_deleted_reports', payload))

    visuals_to_delete = [v['id'] for v in response.json()]
    self.assertEquals(visuals_to_delete, [visual_1])

    response, _ = run_with_db_query_check(lambda: admin_client.delete(
      f'/datasets/dataset/{ds_1}'))

    self.assertEqual(response.status_code, 200)
    self.assertEqual(Report.objects.filter(id=visual_1).count(), 0)
    self.assertEqual(Report.objects.filter(
      id__in=[visual_2, dashboard_id]).count(), 2)
    self.assertEqual(DataSet.objects.filter(id=ds_1).count(), 0)
    self.assertEqual(DataSet.objects.filter(id=ds_2).count(), 1)

  def create_dashboards(self, admin_client: django.test.client.Client, ds_1: int, ds_2: int):

    util = test_report_operations.TestDashboardOperations()

    visual_1_1 = util.create_visual(admin_client, ds_1)
    visual_1_2 = util.create_visual(admin_client, ds_1)
    visual_1_3 = util.create_visual(admin_client, ds_1)
    visual_1_4 = util.create_visual(admin_client, ds_1)

    visual_2_1 = util.create_visual(admin_client, ds_2)
    visual_2_2 = util.create_visual(admin_client, ds_2)

    db_1 = util.create_dashboard(admin_client, ds_1, [visual_1_1])
    # mixed dataset dashboard
    db_2 = util.create_dashboard(admin_client, ds_1, [visual_1_2, visual_2_1])
    # make these dashboards public:
    util.move_to_public_workspace(admin_client, db_1)
    util.move_to_public_workspace(admin_client, db_2)

    # dashboards private to admin:
    util.create_dashboard(admin_client, ds_1, [visual_1_3])
    # mixed dataset dashboard
    util.create_dashboard(admin_client, ds_1, [visual_1_4, visual_2_2])

  def possible_deleted_reports(self, client: django.test.client.Client, ds_1: int, expected_items: int):
    payload = {
      'dataset_id': ds_1
    }
    response, _ = run_with_db_query_check(lambda: client.get(
      '/datasets/possible_deleted_reports', payload))
    self.assertEqual(response.status_code, 200)
    self.assertEqual(len(response.json()), expected_items, 'possible_deleted_reports count')

  def test_pre_delete_performance(self):
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    user1 = make_user('user1', is_superuser=False)
    user1_client = create_client(user1)

    util = test_report_operations.TestDashboardOperations()

    dc_id = util.create_dataconnection(admin_client)
    ds_1 = util.create_dataset(admin_client, dc_id, 'ds_1')
    ds_2 = util.create_dataset(admin_client, dc_id, 'ds_2')

    util.setup_user_role(user1, dc_id, ds_1, ds_2)

    self.create_dashboards(admin_client, ds_1, ds_2)
    self.assertEqual(Report.objects.count(), 4 + 2 + 4, 'ds_1 visuals + ds_2 visuals + dashboards')

    self.possible_deleted_reports(user1_client, ds_1, 4)
    user_query_count_1 = len(django.db.connection.queries)
    self.possible_deleted_reports(admin_client, ds_1, 4)
    admin_query_count_1 = len(django.db.connection.queries)

    # create an other set of dashboards and check the DB query count:
    self.create_dashboards(admin_client, ds_1, ds_2)
    self.assertEqual(Report.objects.count(), 2*(4 + 2 + 4), 'ds_1 visuals + ds_2 visuals + dashboards')

    self.possible_deleted_reports(user1_client, ds_1, 2*4)
    user_query_count_2 = len(django.db.connection.queries)
    self.possible_deleted_reports(admin_client, ds_1, 2*4)
    admin_query_count_2 = len(django.db.connection.queries)

    self.assertEquals(user_query_count_2, user_query_count_1 + 0, '0 extra query per dashboard for non-admin')
    self.assertEquals(admin_query_count_2, admin_query_count_1 + 0, '0 extra query per dashboard for admin')

class TestSegmentREST(django.test.TestCase):

  def setUp(self) -> None:
    super().setUp()
    self.common_setup()

  def tearDown(self) -> None:
    # feature flipper values are cached:
    FeatureFlipper.instance = None
    return super().tearDown()

  def common_setup(self):
    util = test_report_operations.TestDashboardOperations()
    self.util = util

    # crete admin
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    self.admin_client = admin_client

    # create normal user
    user1 = make_user('user1', is_superuser=False)
    self.user1 = user1
    user1_client = create_client(user1)
    self.user1_client = user1_client

    # create impala connection and dataset
    dc_id = self.util.create_dataconnection(admin_client, connection_name='DC1')
    self.dc_id = dc_id
    self.ds_id_1 = util.create_dataset(admin_client, dc_id, 'impala_dataset_1')
    self.ds_id_2 = util.create_dataset(admin_client, dc_id, 'impala_dataset_2')

    sqlrun_common = Common()
    ds_1 = DataSet.objects.get(id=self.ds_id_1)
    ds_2 = DataSet.objects.get(id=self.ds_id_2)
    self.s1 = sqlrun_common.make_segment(
      admin_client, ds_1, 's1', '[A] in ("doug", "veronica")')
    self.s11 = sqlrun_common.make_segment(
      admin_client, ds_1, 's11', '[AA] in ("dougA", "veronicaA")')
    self.s2 = sqlrun_common.make_segment(
      admin_client, ds_1, 's2', '[A] in ("alan", "sean", "amanda")')
    fa1 = sqlrun_common.make_fa(
      admin_client, ds_1, 'fa_s1', self.s1.get('id'), users=[user1.id])
    fa2 = sqlrun_common.make_fa(
      admin_client, ds_2, 'fa_s2', self.s2.get('id'), users=[user1.id])

    role_1 = full_access_only_one_dataset_role(dc_id, self.ds_id_1)
    role_2 = full_access_only_one_dataset_role(dc_id, self.ds_id_2)
    role_1.add_users([user1.username])
    role_2.add_users([user1.username])

  def segment_get(self, client: django.test.Client, dataset_id=None, expected_query_count:int=None) -> list[dict]:
    query_parameters = {}
    if dataset_id:
      query_parameters['dataset_id']=dataset_id

    def action():
      return client.get('/datasets/segment', data=query_parameters)

    result, _ = run_with_db_query_check(action, asserter=self, expected_query_count=expected_query_count)
    self.assertEquals(result.status_code, 200)
    return result.json()

  def filterassociation_get(self, client: django.test.Client, dataset_id=None, expected_query_count:int=None) -> list[dict]:
    query_parameters = {}
    if dataset_id:
      query_parameters['dataset_id']=dataset_id

    def action():
      return client.get('/datasets/filterassociation', data=query_parameters)

    result, query_count = run_with_db_query_check(action, asserter=self, expected_query_count=expected_query_count)
    self.assertEquals(result.status_code, 200)
    return result.json(), query_count

  def segment_delete(self, client: django.test.Client, segment_id: int, ds_id=int) -> list[dict]:
    def action():
      return client.delete(f'/datasets/segment/{segment_id}?dataset_id={ds_id}')

    result, _ = run_with_db_query_check(action, asserter=self)
    self.assertEquals(result.status_code, 200)
    return result.json()

  def dataset_version_change(self, client: django.test.Client, dataset_id: int, target_version_id: int):
    data = {
      'master_id': dataset_id,
      'version_id': target_version_id,
    }
    response = client.post('/datasets/dataset/version_control/', data=data)
    self.assertEquals(response.status_code, 200)

  def revert_dataset_version(self):
    # retrieve the snapshot created for the dataset when deleting the segment:
    snapshot_id = DataSet.objects.filter(version_group_id=self.ds_id_1).exclude(id=self.ds_id_1).values_list('id', flat=True)[0]
    self.dataset_version_change(self.admin_client, dataset_id=self.ds_id_1, target_version_id=snapshot_id)

  def test_get_admin(self):

    enable_dataset_versioning()

    segments = self.segment_get(self.admin_client, expected_query_count=3)
    self.assertEqual(len(segments), 3)

    segments = self.segment_get(self.admin_client, dataset_id=self.ds_id_1, expected_query_count=4)
    self.assertEqual(len(segments), 3)
    segments = self.segment_get(self.admin_client, dataset_id=self.ds_id_2, expected_query_count=4)
    self.assertEqual(len(segments), 0)

    self.segment_delete(self.admin_client, segment_id=self.s1['id'], ds_id=self.ds_id_1)

    segments = self.segment_get(self.admin_client, expected_query_count=3)
    self.assertEqual(len(segments), 2)

    segments = self.segment_get(self.admin_client, dataset_id=self.ds_id_1, expected_query_count=4)
    self.assertEqual(len(segments), 2)
    segments = self.segment_get(self.admin_client, dataset_id=self.ds_id_2, expected_query_count=4)
    self.assertEqual(len(segments), 0)

    self.revert_dataset_version()

    segments = self.segment_get(self.admin_client, expected_query_count=3)
    self.assertEqual(len(segments), 3)

    segments = self.segment_get(self.admin_client, dataset_id=self.ds_id_1, expected_query_count=4)
    self.assertEqual(len(segments), 3)
    segments = self.segment_get(self.admin_client, dataset_id=self.ds_id_2, expected_query_count=4)
    self.assertEqual(len(segments), 0)

  def test_get_user(self):

    enable_dataset_versioning()

    segments = self.segment_get(self.user1_client, expected_query_count=4+2) # extra queries: roles+groups+accessible dataset ids
    self.assertEqual(len(segments), 3)

    segments = self.segment_get(self.user1_client, dataset_id=self.ds_id_1, expected_query_count=5+1)
    self.assertEqual(len(segments), 3)
    segments = self.segment_get(self.user1_client, dataset_id=self.ds_id_2, expected_query_count=5+1)
    self.assertEqual(len(segments), 0)

    self.segment_delete(self.admin_client, segment_id=self.s1['id'], ds_id=self.ds_id_1)

    segments = self.segment_get(self.user1_client, expected_query_count=4+2)
    self.assertEqual(len(segments), 2)

    segments = self.segment_get(self.user1_client, dataset_id=self.ds_id_1, expected_query_count=5+1)
    self.assertEqual(len(segments), 2)
    segments = self.segment_get(self.user1_client, dataset_id=self.ds_id_2, expected_query_count=5+1)
    self.assertEqual(len(segments), 0)

    self.revert_dataset_version()

    segments = self.segment_get(self.user1_client, expected_query_count=4+2)
    self.assertEqual(len(segments), 3)

    segments = self.segment_get(self.user1_client, dataset_id=self.ds_id_1, expected_query_count=5+1)
    self.assertEqual(len(segments), 3)
    segments = self.segment_get(self.user1_client, dataset_id=self.ds_id_2, expected_query_count=5+1)
    self.assertEqual(len(segments), 0)

  def test_get_admin_filterassociations(self):

    fa_list, query_count_1 = self.filterassociation_get(self.admin_client)
    fa_names = {fa['name'] for fa in fa_list}
    self.assertEqual(fa_names, {'fa_s1', 'fa_s2'})

    fa_list, query_count_2 = self.filterassociation_get(self.admin_client, dataset_id=self.ds_id_1)
    fa_names = {fa['name'] for fa in fa_list}
    self.assertEqual(fa_names, {'fa_s1'})

    # verify that the query number does not depend on the datasets
    # it DOES depend on the number of filterassociations :-(
    self.util.create_dataset(self.admin_client, self.dc_id, 'DS 3')

    fa_list, _ = self.filterassociation_get(self.admin_client, expected_query_count=query_count_1)
    fa_names = {fa['name'] for fa in fa_list}
    self.assertEqual(fa_names, {'fa_s1', 'fa_s2'})

    fa_list, _ = self.filterassociation_get(self.admin_client, dataset_id=self.ds_id_1, expected_query_count=query_count_2)
    fa_names = {fa['name'] for fa in fa_list}
    self.assertEqual(fa_names, {'fa_s1'})

  def test_get_user_filterassociations(self):

    fa_list, query_count_1 = self.filterassociation_get(self.user1_client)
    fa_names = {fa['name'] for fa in fa_list}
    self.assertEqual(fa_names, {'fa_s1', 'fa_s2'})

    fa_list, query_count_2 = self.filterassociation_get(self.user1_client, dataset_id=self.ds_id_1)
    fa_names = {fa['name'] for fa in fa_list}
    self.assertEqual(fa_names, {'fa_s1'})

    # verify that the query number does not depend on the datasets
    # it DOES depend on the number of filterassociations :-(
    self.util.create_dataset(self.admin_client, self.dc_id, 'DS 3')

    fa_list, _ = self.filterassociation_get(self.user1_client, expected_query_count=query_count_1)
    fa_names = {fa['name'] for fa in fa_list}
    self.assertEqual(fa_names, {'fa_s1', 'fa_s2'})

    fa_list, _ = self.filterassociation_get(self.user1_client, dataset_id=self.ds_id_1, expected_query_count=query_count_2)
    fa_names = {fa['name'] for fa in fa_list}
    self.assertEqual(fa_names, {'fa_s1'})

class TestReportCreationPermissions(django.test.TestCase):
  def test_super_user(self):
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)

    result, _ = run_with_db_query_check(lambda: admin_client.get('/datasets/dataset/report_creation_permissions'), asserter=self, expected_query_count=2)

    self.assertEqual(
      result.json(),
      {
        "canCreateDashboard": True,
        "canCreateApp": True,
      }
    )

  def test_user_with_only_view_permission(self):
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)

    util = test_report_operations.TestDashboardOperations()
    dc_id = util.create_dataconnection(admin_client)
    dataset = util.create_dataset(admin_client, dc_id, 'ds_1')

    role = Role()
    role.name = "view-dataset"
    role.privileges = json.dumps([
      {
        "ptype": "dataset",
        "dcid": str(dc_id),
        "dslist": [
            str(dataset)
        ],
        "perms": [
            "ds_appview",
        ]
      }
    ])
    role.save()

    user = make_user('view_permission')
    client = create_client(user)
    role.add_users([user.username])

    result, _ = run_with_db_query_check(lambda: client.get('/datasets/dataset/report_creation_permissions'), asserter=self, expected_query_count=5)

    self.assertEqual(
      result.json(),
      {
        "canCreateDashboard": False,
        "canCreateApp": True,
      }
    )

  def test_user_with_edit_permission(self):
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)

    util = test_report_operations.TestDashboardOperations()
    dc_id = util.create_dataconnection(admin_client)
    dataset = util.create_dataset(admin_client, dc_id, 'ds_2')

    role = Role()
    role.name = "edit-dataset"
    role.privileges = json.dumps([
      {
        "ptype": "dataset",
        "dcid": str(dc_id),
        "dslist": [
            str(dataset)
        ],
        "perms": [
            "ds_appview",
            "ds_appedit"
        ]
      }
    ])
    role.save()

    user = make_user('edit_permission')
    client = create_client(user)
    role.add_users([user.username])

    result, _ = run_with_db_query_check(lambda: client.get('/datasets/dataset/report_creation_permissions'), asserter=self, expected_query_count=5)

    self.assertEqual(
      result.json(),
      {
        "canCreateDashboard": True,
        "canCreateApp": True,
      }
    )

class TestGetDatasetsWithDataPageParam(django.test.TestCase):
  def test_dataset_info_fetched_and_tablenames_values(self):
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    util = test_report_operations.TestDashboardOperations()

    dc_id = util.create_dataconnection(admin_client)
    util.create_dataset(admin_client, dc_id, 'ds_1', 'singletable')
    util.create_dataset(admin_client, dc_id, 'ds_2', 'multitable')
    util.create_dataset(admin_client, dc_id, 'ds_3', 'singletable-sql')
    util.create_dataset(admin_client, dc_id, 'ds_4', 'multitable-sql')
    expected_tablenames = ['["default.table_1"]',
                           '["default.table_1", "default.table_2"]',
                           '["main.abc"]',
                           '["main.abc", "default.table_2"]'
                           ]

    def check_response(response, data_page):
      dl = response.json()
      for idx, ds in enumerate(dl):
        if data_page:
          self.assertNotIn(
            'info',
            ds,
            'dataset_info should not be in the response.'
          )
        else:
          self.assertIn(
            'info',
            ds,
            'dataset_info should be in the response.'
          )

        self.assertEquals(ds['tablenames'], expected_tablenames[idx])

    response, _ = run_with_db_query_check(lambda: admin_client.get(
      '/datasets/dataset?get_associate_dashboard=true&data_page=true'), expected_query_count=5, asserter=self)
    check_response(response, True)

    response2, _ = run_with_db_query_check(lambda: admin_client.get(
      '/datasets/dataset?get_associate_dashboard=true'), expected_query_count=5, asserter=self)
    check_response(response2, False)
