# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.test import TestCase

from django.test import TestCase

from django.core.exceptions import ValidationError

from datasets.models import DataSet, Segment, Event, FilterAssociation, DataConnection

INVALID_FIELD = 'a<b'
INVALID_MSG = 'Invalid character in field a<b'

class ModelValidation(TestCase):
  def test_dataset(self):
    # First, test success
    ds = DataSet(
        dataset_name='ds1',
        dataset_type='singletable',
        dataset_detail='ds detail',
        dataset_description='ds description',
        # dataset_info is allowed to contain invalid characters
        dataset_info=INVALID_FIELD,
        dataset_everyone_perm='rw',
        dataset_create_user_perm='sp',
        create_user='user1',
        update_user='user2')
    ds.save()

    # Now, check that invalid characters are being validated
    ds = DataSet(
        dataset_name=INVALID_FIELD,
        dataset_type=INVALID_FIELD,
        dataset_detail=INVALID_FIELD,
        dataset_description=INVALID_FIELD,
        dataset_info='ds info',
        dataset_everyone_perm=INVALID_FIELD,
        dataset_create_user_perm=INVALID_FIELD,
        create_user=INVALID_FIELD,
        update_user=INVALID_FIELD)
    with self.assertRaises(ValidationError):
      try:
        ds.save()
      except ValidationError as v:
        self.assertEqual(v.messages, [INVALID_MSG] * 8)
        raise

  def test_event(self):
    # First, test success
    event = Event(
        event_name='event name',
        event_data='event data',
        create_user='user1',
        update_user='user2')
    event.save()

    event = Event(
        event_name=INVALID_FIELD,
        event_data=INVALID_FIELD,
        create_user=INVALID_FIELD,
        update_user=INVALID_FIELD)
    with self.assertRaises(ValidationError):
      try:
        event.save()
      except ValidationError as v:
        self.assertEqual(v.messages, [INVALID_MSG] * 3)
        raise

  def test_segment(self):
    # First, test success
    seg = Segment(
        segment_name='seg name',
        # Note that segment_data permits invalid characters
        segment_data=INVALID_FIELD,
        create_user='user1',
        update_user='user1')
    seg.save()

    seg = Segment(
        segment_name=INVALID_FIELD,
        segment_data='seg data',
        create_user=INVALID_FIELD,
        update_user=INVALID_FIELD)
    with self.assertRaises(ValidationError):
      try:
        seg.save()
      except ValidationError as v:
        self.assertEqual(v.messages, [INVALID_MSG] * 3)
        raise

  def test_dataconnection(self):
    # First, test success
    dc = DataConnection(
        dataconnection_type='arcengine',
        dataconnection_name='default',
        dataconnection_description='auto-created default connection',
        dataconnection_info = '{}',
        create_user='user1',
        update_user='user2')
    dc.save()

    # Now, check that invalid characters are being validated
    dc = DataConnection(
        dataconnection_type=INVALID_FIELD,
        dataconnection_name=INVALID_FIELD,
        dataconnection_description=INVALID_FIELD,
        dataconnection_info = INVALID_FIELD,
        create_user=INVALID_FIELD,
        update_user=INVALID_FIELD)
    with self.assertRaises(ValidationError):
      try:
        dc.save()
      except ValidationError as v:
        self.assertEqual(v.messages, [INVALID_MSG] * 6)
        raise

  def test_filterassociation(self):
    # First, test success
    fa = FilterAssociation(
        filterassociation_name='fa name',
        filterassociation_data='fa data',
        create_user='user1',
        update_user='user2')
    fa.save()

    fa = FilterAssociation(
        filterassociation_name=INVALID_FIELD,
        filterassociation_data=INVALID_FIELD,
        create_user=INVALID_FIELD,
        update_user=INVALID_FIELD)
    with self.assertRaises(ValidationError):
      try:
        fa.save()
      except ValidationError as v:
        self.assertEqual(v.messages, [INVALID_MSG] * 4)
        raise

