# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import logging
import jobs
from django.http import HttpResponse
from django.http.request import QueryDict
from django.conf import settings
from django.utils.html import escape
from rbac.permcache import RBAC
from vizlib.querylog import ActionType, logAction, ComponentType
from dataconn.connections_types import KNOWN_CONNECTION_TYPES
from dataconn import dataconn
from datasets.models import DataSet, DataConnection, dataconnection_post_save
from jobs.views_encryption import encrypt_secret


LOGGER = logging.getLogger("arcweb")

HISTORY_ETL = """
CREATE TABLE IF NOT EXISTS {target_table}
  PARTITIONED BY (start_date)
  SORT BY (start_time, end_time)
  STORED AS PARQUET
  AS SELECT *, to_date(start_time)
  AS start_date
  FROM arcadia_catalog.completed_queries;
INSERT INTO {target_table}
PARTITION (start_date)
  SELECT *, to_date(start_time)
    AS start_date
    FROM arcadia_catalog.completed_queries
    WHERE end_time > (SELECT max(end_time) FROM {target_table});
"""
#TODO: Invistigate to see if we can just delete it.
def get_default_dataconnection():
  dc_count = DataConnection.objects.count()
  if dc_count == 0:
    _migration_dataconnection()
  dcl = DataConnection.objects.filter(dataconnection_type='arcengine').order_by('id')[:1]
  if len(dcl) == 1:
    return dcl[0]
  return None

def _migration_dataconnection(request=None):
  if not hasattr(settings, 'ARCSQL_SERVER'):
    return

  dl = DataSet.objects.all()
  if len(dl) == 0:
    # If there are no datasets in existence, do not auto-create
    # a data connection
    return

  server = settings.ARCSQL_SERVER.split(":")
  if len(server) == 1:
    host = server[0]
  elif len(server) == 2:
    host = server[0]
    port = int(server[1])
  else:
    # We don't understand the contents of ARCSQL_SERVER - do not autocreate
    # a data connection
    return

  dc = DataConnection(
      dataconnection_type='arcengine',
      dataconnection_name='default',
      dataconnection_description='auto-created default connection',
      dataconnection_info = '{"PARAMS": {"HOST": "%s", "PORT": %s, "AUTH": "NoSasl"}}' % \
          (host, port))
  dc.save()
  if request:
    _auditDataConnection(dc, ActionType.CREATE, request=request)
  else:
    _auditDataConnection(dc, ActionType.CREATE, username='util admin')

def _auditDataConnection(dataconnection, action, request=None, username=None):
  logAction(action=action, component_type=ComponentType.DATACONNECTION,
      component_id=dataconnection.id, request=request, username=username,
      dataconnection_id=dataconnection.id, dataconnection_name=dataconnection.dataconnection_name,
      dataconnection_type=dataconnection.dataconnection_type)

# Common validation for data connection create/update
def _validate_request_and_get_dc(request, dataconnection_id=None, input=None):
  dc = None
  if dataconnection_id is not None:
    dcl = DataConnection.objects.filter(id=int(dataconnection_id))
    if len(dcl) == 0:
      return None, HttpResponse(
          "Data Connection with id '%s' does not exist" % dataconnection_id,
          status=500)
    dc = dcl[0]

  if request.POST:
    data = request.POST
  else:
    data = input or QueryDict(request.body)

  dataconnection_type = data.get('dataconnection_type', None)
  dataconnection_password = data.get('dataconnection_password', None)
  dataconnection_name = data.get('dataconnection_name', '')
  dataconnection_description = data.get('dataconnection_description', '')
  dataconnection_info = data.get('dataconnection_info', '')

  if not dataconnection_type in KNOWN_CONNECTION_TYPES:
    return None, HttpResponse("Invalid data connection type '%s'" % escape(dataconnection_type), status=500)

  if not dc:
    # New connection is being created. Confirm that the name is unique
    dcl = DataConnection.objects.filter(dataconnection_name=dataconnection_name)
    if len(dcl) > 0:
      return None, HttpResponse(
          "There is an existing data connection with the same name",
          status=500)

  prev_etl = None
  if not dc:
    # New connection - create a data connection
    dc = DataConnection(
        dataconnection_type = dataconnection_type,
        dataconnection_name = dataconnection_name,
        dataconnection_description = dataconnection_description,
        dataconnection_info = dataconnection_info)
  else:
    prev_etl = json.loads(dc.dataconnection_info or '{}').get('PARAMS', {}).get('ARCENGINE_HISTORY_ETL')
    dc.dataconnection_type = dataconnection_type
    dc.dataconnection_name = dataconnection_name
    dc.dataconnection_description = dataconnection_description
    dc.dataconnection_info = dataconnection_info

  if dataconnection_password is not None:
    dc.dataconnection_password = json.dumps(encrypt_secret(str(dataconnection_password)))

  # Sort of a hack here--for validation path, we return the validation result
  # as the 'error' object so that the callers will return before doing any
  # actual update of the dc object being validated.
  #
  # Note that we could support a 'validate then save' mode by returning the
  # (dc, None) tuple, but that is not needed at this point.
  #
  if data.get('do_validate') == "true":
    try:
      dataconn.validate_dataconn(request.user, dc)
      return None, HttpResponse('{"status": "success"}', content_type="application/json")
    except Exception as e:
      error_message = str(e)
      LOGGER.exception(error_message)
      return None, HttpResponse(error_message, status=500)

  curr_etl = json.loads(dc.dataconnection_info or '{}').get('PARAMS', {}).get('ARCENGINE_HISTORY_ETL', {})

  # Delete the old job present and something has changed
  skip_etl = False
  if prev_etl and prev_etl.get('ENABLED'):
    skip_etl = all(prev_etl.get(attr) == curr_etl.get(attr) for attr in ['ENABLED', 'TARGET_TABLE', 'JOB_NAME', 'SCHEDULE'])
    if not skip_etl:
      try:
        jobs.models.JobScheduleRepeat.objects.get(id=prev_etl.get('JOB_ID')).delete()
      except:
        LOGGER.exception('Unable to delete connection history ETL ID %s' % prev_etl.get('JOB_ID'))

  # If an ETL setting has changed, create the current job as needed
  if not skip_etl and curr_etl.get('ENABLED'):
    params = {
      'schedule_name': curr_etl.get('SCHEDULE'),
      'job_name': curr_etl.get('JOB_NAME'),
      'run_now': False,
      'dataconnection_id': dc.id,
      'sql': HISTORY_ETL.format(**{'target_table': curr_etl.get('TARGET_TABLE')})
    }
    etl_data = jobs.views.scriptScheduleInternal(request.user, params)
    curr_etl['JOB_ID'] = etl_data['schedule_id']
    info = json.loads(dc.dataconnection_info or '{}')
    info['PARAMS']['ARCENGINE_HISTORY_ETL'] = curr_etl
    dc.dataconnection_info = json.dumps(info)
  return dc, None
