# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Views on datasets
"""
from django.http import HttpRequest, HttpResponse, QueryDict, JsonResponse
from django.db.models import Count, F, Q
from django.utils.html import escape
from django.forms.models import model_to_dict
from django.db import transaction

from datasets.models import DataSet, DatasetSegment, Segment, Event, FilterAssociation, DataConnection
from datasets.dashboard_association import DashboardAssociation

from reports.models import Report
import reports.views
import extracts.views
from workspace.models import Workspace

from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.views.decorators.csrf import csrf_exempt

from django.views.decorators.http import etag
from arcwebbase.etag_caching import etag_metadata_func

from vizlib.querylog import logAction, ComponentType, ActionType
from rbac.permcache import RBAC
from rbac.permdefs import PermissionException
from vizlib.misc import check_login_or_use_apikey
from datasets.utils import get_default_dataconnection
from reports.utils import _get_report_data
from util.get_dataset_ids_accessible_by_user import get_dataset_ids_accessible_by_user, get_dataset_ids_managable_by_user, is_dataset_superuser, get_dataset_query_for_permission_check

import json
import re
import logging
from vizlib.dataset_lvinfo import dataset_update_lvinfo
import uuid

from dataconn.connections_types import ALL_CONNECTION_TYPES
from dataconn.cml_connection_importer import CmlConnectionImporter
from django.db.models import F
from util.ff_flipper import FeatureFlipper

from rbac.models import Group


LOGGER = logging.getLogger("arcweb")

HISTORY_ETL = """
CREATE TABLE IF NOT EXISTS {target_table}
  PARTITIONED BY (start_date)
  SORT BY (start_time, end_time)
  STORED AS PARQUET
  AS SELECT *, to_date(start_time)
  AS start_date
  FROM arcadia_catalog.completed_queries;
INSERT INTO {target_table}
PARTITION (start_date)
  SELECT *, to_date(start_time)
    AS start_date
    FROM arcadia_catalog.completed_queries
    WHERE end_time > (SELECT max(end_time) FROM {target_table});
"""

def _auditDataset(dataset, action, request=None, username=None):
  # Handle old-style datasets without data connections
  if dataset.dataconnection:
    dcid = dataset.dataconnection.id
    dctype = dataset.dataconnection.dataconnection_type
    dcname = dataset.dataconnection.dataconnection_name
  else:
    dcid = ''
    dctype = ''
    dcname = ''

  logAction(action=action, component_type=ComponentType.DATASET,
      component_id=dataset.id, dataset_id=dataset.id, dataset_name=dataset.dataset_name,
      dataconnection_id=dcid, dataconnection_name=dcname, dataconnection_type=dctype, request=request, username=username)

def _auditSegment(segment, action, dataconnection, request=None):
  logAction(action=action, component_type=ComponentType.SEGMENT, component_id=segment.id,
      dataconnection_id=dataconnection.id, dataconnection_name=dataconnection.dataconnection_name,
      dataconnection_type=dataconnection.dataconnection_type, request=request)

def _auditFilterAssociation(segment, action, dataconnection, request=None):
  logAction(action=action, component_type=ComponentType.FILTERASSOCIATION, component_id=segment.id,
      dataconnection_id=dataconnection.id, dataconnection_name=dataconnection.dataconnection_name,
      dataconnection_type=dataconnection.dataconnection_type, request=request)

def _segment_delete(request, segment_id):
  """Delete a segment"""
  dataset_id = request.GET.get('dataset_id', None)
  dataset = DataSet.objects.get(pk=dataset_id)

  if segment_id is None:
    return HttpResponse('No segment id specified', status=500)

  if dataset_id is None:
    return HttpResponse('No dataset id specified', status=500)

  try:
    # pylint: disable=E1101
    s = Segment.objects.get(id=segment_id)
    if (FeatureFlipper().get_flipper_value('enable_version_control') and FeatureFlipper().get_flipper_value('enable_dataset_version_control')):
      # Since 7.2.6, because of dataset versioning feature we are not deleting the segment just remove the association record from the switch table called Dataset_DatasetSegment.
      # If the feature is enabled, we create a snapshot/version about the dataset before we delete the record from the switch table
      _dataset_update(request, dataset_id)

    dss = DatasetSegment.objects.get(dataset=dataset.id, segment_id=s.id, dataset_version_id=dataset.version_id)
    dss.delete()

    return HttpResponse('[]', content_type="application/json")
  except Report.DoesNotExist:
    return HttpResponse('[]', content_type="application/json")

def _segment_update(request, segment_id, input=None):
  """Update segment details"""
  if segment_id is None:
    return HttpResponse('No segment id specified', status=500)

  segment_id = int(segment_id)

  try:
    # pylint: disable=E1101
    s = Segment.objects.get(id=segment_id)
    dataset_version_id = request.GET.get('dataset_version_id', None)
    dataset = DataSet.objects.get(version_id=dataset_version_id)

    if (FeatureFlipper().get_flipper_value('enable_version_control') and FeatureFlipper().get_flipper_value('enable_dataset_version_control')):
      dataset.checkPerm(request.user, 'ds_manage')

      s = _segment_create(request, s, dataset.id)
      dss = DatasetSegment.objects.get(dataset=dataset.id, segment_id=segment_id, dataset_version_id=dataset_version_id)
      dss.delete()


    put = input or QueryDict(request.body)
    name = put.get('name', None)
    data = put.get('data', None)

    if name is None:
      return HttpResponse("Invalid name", status=500)
    if data is None:
      return HttpResponse("Invalid data", status=500)

    s.segment_name = name
    s.segment_data = data
    s.update_user = request.user.username
    s.save()

    _auditSegment(s, ActionType.UPDATE, dataset.dataconnection, request=request)

    json_data = []
    json_data.append(_get_segment_data(s))
    return HttpResponse(json.dumps(json_data), content_type="application/json")
  except Segment.DoesNotExist:
    return HttpResponse('Could not find segment %s' % segment_id, status=500)
  except DataSet.DoesNotExist:
    return HttpResponse("Dataset does not exist with version_id: %s" % dataset_version_id, status=500)
  except DatasetSegment.DoesNotExist:
    return HttpResponse('Could not find dataset_segment', status=500)
  except PermissionException as e:
      LOGGER.error(e)
      return HttpResponse(e, status=403)

def _segment_create(request, versioned_segment=None, versioned_dataset_id=None):
  active_dataset_id = versioned_dataset_id or request.POST.get("dataset_id", None)
  """Create segment"""

  if versioned_segment:
    name = versioned_segment.segment_name
    data = versioned_segment.segment_data
  else:
    name = request.POST.get("name", None)
    data = request.POST.get("data", None)

  if active_dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)
  if name is None:
    return HttpResponse("Invalid name", status=500)
  if data is None:
    return HttpResponse("Invalid data", status=500)
  active_dataset_id = int(active_dataset_id)
  # pylint: disable=E1101
  try:
    active_dataset = DataSet.objects.get(id=active_dataset_id)
  except DataSet.DoesNotExist:
    return HttpResponse("Dataset %s does not exist" % active_dataset_id, status=500)
  active_dataset.checkPerm(request.user, 'ds_manage')

  s = Segment(segment_name=name,
              segment_data=data, create_user=request.user.username)

  if (FeatureFlipper().get_flipper_value('enable_version_control') and FeatureFlipper().get_flipper_value('enable_dataset_version_control')):
    _dataset_update(request, active_dataset_id)

  s.save()

  dss = DatasetSegment(dataset=active_dataset, segment=s, dataset_version_id=active_dataset.version_id)
  dss.save()

  if(FeatureFlipper().get_flipper_value('enable_version_control') and FeatureFlipper().get_flipper_value('enable_dataset_version_control') and versioned_segment):
    return s

  _auditSegment(s, ActionType.CREATE, active_dataset.dataconnection, request=request)

  json_data = []
  json_data.append(_get_segment_data(s))

  return HttpResponse(json.dumps(json_data), content_type="application/json")

from django.db import transaction

def _segments_clone(request, old_dataset_id, new_dataset):
    old_dataset = DataSet.objects.get(id=old_dataset_id)
    segment_links = DatasetSegment.objects.filter(
        dataset_id=old_dataset_id,
        dataset_version_id=old_dataset.version_id
    )

    with transaction.atomic():
        for link in segment_links:
            old_segment = link.segment
            new_segment = Segment.objects.create(
                create_user=request.user.username,
                update_user=request.user.username,
                segment_name=old_segment.segment_name,
                segment_data=old_segment.segment_data,
                uuid=uuid.uuid4()
            )

            DatasetSegment.objects.create(
                dataset=new_dataset,
                segment=new_segment,
                dataset_version_id=new_dataset.version_id
            )

            new_segment.save()
            _auditSegment(new_segment, ActionType.CREATE, new_dataset.dataconnection, request)


def _get_segment_data(s):
  """Return json'ified segment data"""
  sd = {
      'id': s.id,
      'name': s.segment_name,
      'data': s.segment_data,
      'created': s.create_time.isoformat(timespec='milliseconds'),
      'created_by': s.create_user,
      'updated': s.update_time.isoformat(timespec='milliseconds'),
      'updated_by': s.update_user,
  }
  return sd

def _segment_get(request, segment_id=None, dataset_id=None, dataset=None, filter=None):
  """Get segment details"""

  user = request.user
  segments_query = \
    Q(datasetsegment__dataset_version_id=F('datasetsegment__dataset__version_id'))\
    &\
    Q(datasetsegment__dataset__is_active_version=True)

  # We may be provided with a segment id, a dataset id, or nothing
  # pylint: disable=E1101
  if segment_id:
    # Segment id provided - get information about that segment
    sl = Segment.objects.filter(id=int(segment_id))

  elif dataset_id is not None:

    # Dataset id provided - return all segments for that dataset
    # pylint: disable=E1101
    try:
      dataset_id = int(dataset_id)
      d = dataset or DataSet.objects.get(id=dataset_id)
    except DataSet.DoesNotExist:
      return (None, "Dataset %s does not exist" % dataset_id)
    except:
      raise Exception('Invalid dataset ID')

    # TODO: Viewing the app should not require the viz side loading
    # the segment so when we fix that, this check should be for 'ds_manage'
    if not is_dataset_superuser(user):
      d.checkPerm(user, 'ds_appview')
    # sl = Segment.objects.filter(datasetsegment__dataset_id=dataset_id, datasetsegment__dataset_version_id=d.version_id)
    segments_query &= Q(datasetsegment__dataset__in=[dataset_id])
    sl = Segment.objects.filter(segments_query)
  else:
    if not is_dataset_superuser(user):
      dataset_ids = get_dataset_ids_accessible_by_user(user)
      segments_query &= Q(datasetsegment__dataset__in=dataset_ids)

    sl = Segment.objects.filter(segments_query)
    if filter:
      sl = sl.filter(**filter)

  json_data = []
  for s in sl:
    json_data.append(_get_segment_data(s))

  return (json_data, None)

def _segment_get_http(request, segment_id, dataset_id=None, filter=None):
  """Get segment details"""

  data, err = _segment_get(request, segment_id, dataset_id=dataset_id, filter=filter)
  if err:
    return HttpResponse(err, status=500)

  return HttpResponse(json.dumps(data), content_type="application/json")

def _dataset_delete(request, dataset_id):
  """Delete a dataset"""
  if dataset_id is None:
    return HttpResponse("Invalid dataset id %s" % dataset_id, status=500)

  dataset_id = int(dataset_id)
  # pylint: disable=E1101
  dls = list(
    DataSet.objects\
      .filter(version_group_id=dataset_id)\
      .select_related('dataconnection')\
      .only(
        'id', 'dataconnection__id', 'dataconnection__dataconnection_type', 'dataconnection__dataconnection_name'
      )
  )
  if len(dls) == 0:
      return HttpResponse("Invalid dataset id %s" % dataset_id, status=500)

  errors = []
  for ds in dls:
        # Check permissions for each dataset
        try:
            ds.checkPerm(request.user, 'ds_manage')
        except Exception as e:
            errors.append(str(e))
            continue

        put = QueryDict(request.body)
        delete_table = put.get('delete_table', None)
        if delete_table == "true":
            # Attempt to drop the table associated with the dataset
            try:
                from dataconn.dataconn import get_dataconn
                dc = get_dataconn(ds.dataconnection_id)
                dc.sync_runquery(query="drop table %s" % ds.dataset_detail, username=request.user.username)
            except Exception as e:
                errors.append(str(e))
                continue

        # Audit the deletion action
        _auditDataset(ds, ActionType.DELETE, request=request)
        DashboardAssociation().handle_dataset_deletion(ds, request)
        ds.delete()

  return HttpResponse('[]', content_type="application/json")



def _dataset_check_column_rename(ds, old_info, request):
  """
    Handle a column display name change for a dataset.

    We should have used a hidden and fixed internal name for each column
    in a dataset and made all uses of the column lookup through the dataset
    to get the display name.

    Instead, the display name gets incorporated into expressions in lots
    of different places (report expressions, segment filters, segment entities,
    event filters, etc.).

    So if we change a display name and want all the references to continue
    to work, we need to find and fix all of those.

    Text replace on json strings seems pretty bad but not really worse than
    doing full parse on the json objects then text replace on specific elements
  """

  # Extend list of text replacement pairs built from old/new column names
  def add_pairs(pairs, columns, template):
    for old, new in columns:
      pairs.append((re.compile(re.escape(template.format(old)), re.IGNORECASE),
                    template.format(new)))

  # Perform old/new text replacement, return (<changed>, <new text>) tuple
  def col_rename(text, pairs):
    save = text
    for old, new in pairs:
      text = old.sub(new, text)
    return (text != save, text)

  # Get list of renamed columns
  columns = []
  old_info = json.loads(old_info or '[]')
  new_info = json.loads(ds.dataset_info or '[]')
  for old_table in old_info:
    for new_table in new_info:
      if old_table.get('tablename') == new_table.get('tablename') and \
         old_table.get('table_alias') == new_table.get('table_alias'):

        # Found same table in old and new dataset info -- check columns
        for old_col in old_table.get('columns',[]):
          for new_col in new_table.get('columns',[]):
            if old_col.get('name') == new_col.get('name') and \
               old_col.get('expr', '') == new_col.get('expr', '') and \
               old_col.get('basecol') == new_col.get('basecol'):

              # Found same column in old and new dataset info for this table
              old_name = old_col.get('alias') or old_col.get('name')
              new_name = new_col.get('alias') or new_col.get('name')
              if old_name != new_name:
                columns.append((old_name, new_name))
              break
        break

  if len(columns) == 0:
    return

  pairs = []
  add_pairs(pairs, columns, '[{0}]')

  # Renamed column could be in the expression text for a computed column
  (changed, data) = col_rename(ds.dataset_info, pairs)
  if changed:
    ds.dataset_info = data
    ds.save()
    # We already logged an audit message, we do not need another one

  for s in Segment.objects.filter(datasetsegment__dataset_id=ds.id, datasetsegment__dataset_version_id=ds.version_id):
    (changed, data) = col_rename(s.segment_data, pairs)
    if changed:
      s.segment_data = data
      s.save()

  for e in Event.objects.filter(dataset_id=ds.id):
    (changed, data) = col_rename(e.event_data, pairs)
    if changed:
      e.event_data = data
      e.save()

  # pairs without space
  add_pairs(pairs, columns, '"daterange_base_field":"{0}"')
  add_pairs(pairs, columns, '"picklist_base_field":"{0}"')
  add_pairs(pairs, columns, '"picklist_output_parameter":"{0}"')
  add_pairs(pairs, columns, '"picklist_filter_columns":"{0}"')
  add_pairs(pairs, columns, '"picklist_title":"{0}"')
  add_pairs(pairs, columns, '"dataset_colname":"{0}"')
  add_pairs(pairs, columns, '"filter_column":"{0}"')

  # pairs with space
  add_pairs(pairs, columns, '"daterange_base_field": "{0}"')
  add_pairs(pairs, columns, '"picklist_base_field": "{0}"')
  add_pairs(pairs, columns, '"picklist_output_parameter": "{0}"')
  add_pairs(pairs, columns, '"picklist_filter_columns": "{0}"')
  add_pairs(pairs, columns, '"picklist_title": "{0}"')
  add_pairs(pairs, columns, '"dataset_colname": "{0}"')
  add_pairs(pairs, columns, '"filter_column": "{0}"')

  add_pairs(pairs, columns, " as '{0}'" + '":"')
  for r in Report.objects.filter(dataset_id=ds.id):
    (changed, data) = col_rename(r.report_data, pairs)
    if changed:
      r.report_data = data
      r.save()
      reports.views.auditApp(r, ActionType.UPDATE, request=request)

@etag(etag_metadata_func)
def _dataset_get(request, dataset_id, updatedLV=False, filter=None, search=False,
    get_counts=False, dataconnection_id=None, version_group_id=None, is_create_new_ds_version=True):
  return dataset_get_internal(request, dataset_id, updatedLV=updatedLV, inpds=None,
      filter=filter, search=search, get_counts=get_counts, dataconnection_id=dataconnection_id, version_group_id=version_group_id, is_create_new_ds_version=is_create_new_ds_version)

def user_dataset_ids(user: User, dataconnection_id=None) -> list[int]:
  dataset_query = DataSet.objects.only('id', 'dataconnection_id', 'create_user', 'version_group_id')
  if dataconnection_id:
    dataset_query = dataset_query.filter(dataconnection_id=dataconnection_id)

  return [d.id for d in dataset_query if 'ds_appview' in d.getPerms(user)]

def dataset_get_internal(request, dataset_id, updatedLV=False, inpds=None,
    filter=None, search=False, get_counts=False, dataconnection_id=None, version_group_id=None, is_create_new_ds_version=True):
  segments = []
  serialized_segments = []
  events = []
  filterassociations = []
  extractTracker = extracts.views.ExtractTracker()
  summary = request.GET.get('summary') == 'true'
  detail = ''
  data_page = request.GET.get('data_page', None) == 'true'
  # Retrieve only the necessary fields from the DataSet model
  fields = [
              'id',
              'dataset_name',
              'dataset_type',
              'dataset_detail',
              'create_time',
              'create_user',
              'update_time',
              'update_user',
              'dataset_description',
              'dataset_tablenames',
              'dataset_lvinfo',
              'dataset_settings',
              'search_enabled',
              'dataconnection__dataconnection_name',
              'dataconnection__dataconnection_type',
              'dataconnection__create_user',
              'is_active_version',
              'version_name',
              'version_id',
              'version_group_id',
            ]

  if data_page:
    queryset = DataSet.objects.select_related('dataconnection').only(*fields)
  else:
    fields.append('dataset_info')
    queryset = DataSet.objects.select_related('dataconnection').only(*fields)

  # pylint: disable=E1101
  if dataset_id is None:
    if dataconnection_id:
      queryset = queryset.filter(dataconnection_id=dataconnection_id, is_active_version=True)
    else:
      queryset = queryset.filter(is_active_version=True)


    if not request.user.is_superuser:
      user_ds_ids = user_dataset_ids(request.user, dataconnection_id)
      queryset = queryset.filter(id__in=user_ds_ids)

    dl = queryset.order_by('dataset_name')

    if filter:
      dl = dl.filter(**filter)

  else:
    version_group_id = request.GET.get('version_group_id', None) or version_group_id

    if version_group_id:
      dl = queryset.filter(version_group_id=version_group_id) \
                  .order_by('-version_id')

    else:
      dataset_id = int(dataset_id)
      if inpds is None:
        ds_filter = {'id': dataset_id}
        if is_create_new_ds_version:
          ds_filter['is_active_version'] = True
        dl = queryset.filter(id=dataset_id)
        detail = request.GET.get("detail", 'false')

      else:
        dl = [inpds]
        detail = 'true'

    if detail == 'true' and len(dl) == 1:
      d = dl[0]
      segments = Segment.objects.filter(datasetsegment__dataset_id=dataset_id, datasetsegment__dataset_version_id=d.version_id)
      for s in segments:
        serialized_segments.append(_get_segment_data(s))

      events, err = _event_get(request, dataset_id=dataset_id, dataset=d)
      if err:
        return HttpResponse(err, status=500)

        # The js side only uses the filter assocations for the dataset detail page,
        # not for visual rendering or editing.
        # For the dataset detail page, the filter associations are fetched via
        # the specific URL rather than as part of the 'detail' dataset level URL.
        # Since generating the FA list is expensive, we leave it out of this path.

  json_data = []
  default_dc = None

  # We need to only return datasets for which search is enabled
  if search:
    dl = dl.filter(search_enabled=True)

  if summary:
    dl = dl.only('id', 'dataset_name', 'dataset_type', 'dataset_detail', 'dataconnection_id', 'dataconnection__dataconnection_name', 'dataconnection__dataconnection_type', 'create_user', 'version_group_id')
  else:
    extract_sources = {ds.id: extractTracker.isExtractSource(ds) for ds in dl}
    extract_targets = {ds.id: extractTracker.isExtractTarget(ds) for ds in dl}

  for d in dl:
    # Request for specific dataset does permission check rather than filter
    if dataset_id is not None:
      d.checkPerm(request.user, 'ds_appview')

    dc = d.dataconnection
    if not dc:
      if not default_dc:
        default_dc = get_default_dataconnection()
      dc = default_dc

    if summary:
      dd = {
        'id': d.id,
        'name': d.dataset_name,
        'dc_id': dc.id,
        'dc_name': dc.dataconnection_name,
        'dc_type': dc.dataconnection_type,
        'created_by': d.create_user,
      }
    else:
      dsperms = d.getPerms(request.user)
      dcperms = dc.getPerms(request.user)
      lvinfo = d.dataset_lvinfo if hasattr(d, 'dataset_lvinfo') else None
      lvname = ""
      lvcolumns = ""
      if lvinfo:
        try:
          lvinfo_obj = json.loads(lvinfo)[0]
          lvname = lvinfo_obj.get('lvname', "")
          lvcolumns = lvinfo_obj.get('columns', "")
        except:
          pass

      dd = {
        'id': d.id,
        'dc_id': dc.id,
        'dc_type': dc.dataconnection_type,
        'dc_name': dc.dataconnection_name,
        'dc_type_display': ALL_CONNECTION_TYPES.get(dc.dataconnection_type, "unknown"),
        'name': d.dataset_name,
        'type': d.dataset_type,
        'detail': d.dataset_detail,
        'detail_display': d.dataset_detail,
        'description': d.dataset_description,
        'settings': d.dataset_settings,
        'lvname': lvname,
        'lvcolumns': lvcolumns,
        'created': d.create_time.isoformat(timespec='milliseconds'),
        'created_by': d.create_user,
        'updated': d.update_time.isoformat(timespec='milliseconds'),
        'updated_by': d.update_user,
        'perms': list(dsperms.union(dcperms)),
        'updatedLV': updatedLV,
        'extractSource': extract_sources[d.id],
        'extractTarget': extract_targets[d.id],
        'version_name': d.version_name,
        'version_id': d.version_id,
        'version_group_id': d.version_group_id,
        'tablenames': d.dataset_tablenames
      }

      if not data_page:
        dd['info'] = d.dataset_info

    if dataset_id is not None:
      dd['segments'] = serialized_segments
      dd['events'] = events
      dd['filterassociations'] = filterassociations

    json_data.append(dd)

  if get_counts:
    user = request.user
    if user.is_superuser:
      report_counts = get_report_counts_for_admin()
    else:
      dataset_ids = [dd['id'] for dd in json_data]
      report_counts = get_report_counts(user, dataset_ids)

    for dd in json_data:
      dd['report_count'] = report_counts.get(dd['id'], 0)

  if inpds:
    return json_data

  return HttpResponse(json.dumps(json_data), content_type="application/json")

def get_report_counts(user: User, dataset_ids: list):
  user_workspaces = []
  for ws in Workspace.objects.only('acl').all():
    if user.is_superuser or 'sys_misc' in RBAC.get_user_obj_perms(user) or 'ws_appview' in ws.getPerms(user):
      user_workspaces.append(ws.id)

  results = Report.objects\
    .filter(report_type='dashboard') \
    .filter(dataset_id__in=dataset_ids) \
    .filter(workspace_id__in=user_workspaces) \
    .values('dataset_id') \
    .annotate(report_count = Count('id'))

  return {r['dataset_id']:r['report_count'] for r in results}

def get_report_counts_for_admin():
  results = Report.objects\
    .filter(report_type='dashboard') \
    .values('dataset_id') \
    .annotate(report_count = Count('id'))

  return {r['dataset_id']:r['report_count'] for r in results}


def _get_detail(dataset_detail, dataset_info):
  if dataset_detail == 'Created from SQL':
    return json.loads(dataset_info)[0].get('select')
  else:
    return dataset_detail

def _dataset_clone(dataset_id, request):
  dsl = DataSet.objects.filter(id=dataset_id)
  if len(dsl) == 0:
    return HttpResponse("Invalid dataset %s" % dataset_id)

  if len(dsl) > 1:
    return HttpResponse("Internal error while fetching dataset %s" % dataset_id)

  ds = dsl[0]
  ds.checkPerm(request.user, 'ds_manage')

  # Finally, create the dataset
  ds.id = None
  ds.dataset_name = f"Clone of {ds.dataset_name}"
  ds.create_user = request.user.username
  ds.update_user = request.user.username
  ds.uuid = uuid.uuid4()
  ds.dashboards = json.dumps([])
  ds.version_group_id = None

  ds.save()
  _auditDataset(ds, ActionType.CREATE, request)

  _segments_clone(request, old_dataset_id=dataset_id, new_dataset=ds)
  _filterassociations_clone(request, old_dataset_id=dataset_id, new_dataset=ds)

  return ds.id

def dataset_create(requestData, request):
  """Create dataset"""
  dataconnection_id = requestData.get("dataconnection_id", None)
  dataset_name = requestData.get("dataset_name", None)
  dataset_type = requestData.get("dataset_type", None)
  dataset_info = requestData.get("dataset_info", None)
  dataset_lvinfo = requestData.get("dataset_lvinfo", None)
  dataset_detail = requestData.get("dataset_detail", None)
  dataset_description = requestData.get("description", '')
  dataset_settings = requestData.get("dataset_settings", '{}')
  dashboards = requestData.get("dashboards", [])
  version_name = requestData.get("version_name", None)
  version_group_id = requestData.get("version_group_id", None)
  is_active_version = requestData.get("is_active_version", 1)
  create_user = requestData.get("create_user", request.user.username)

  if dataconnection_id is None:
    return HttpResponse("Invalid data connection information", status=500)
  if dataset_name is None:
    return HttpResponse("Invalid dataset name", status=500)
  if dataset_type is None:
    return HttpResponse("Invalid dataset type", status=500)
  if dataset_detail is None:
    return HttpResponse("Invalid dataset detail", status=500)

  new_dataset = _dataset_create_internal(
    dataconnection_id=dataconnection_id,
    dataset_settings=dataset_settings,
    dataset_name=dataset_name,
    dataset_type=dataset_type,
    dataset_info=dataset_info,
    dataset_lvinfo=dataset_lvinfo,
    dataset_detail=dataset_detail,
    dataset_description=dataset_description,
    dashboards=dashboards,
    version_name=version_name,
    version_group_id=version_group_id,
    is_active_version=is_active_version,
    create_user=create_user,
    user=request.user,
    request=request
  )

  return new_dataset.id

def _dataset_create_internal(
    dataconnection_id,
    dataset_settings,
    dataset_name,
    dataset_type,
    dataset_info,
    dataset_lvinfo,
    dataset_detail,
    dataset_description,
    dashboards,
    version_name,
    version_group_id,
    is_active_version,
    create_user,
    user,
    request
  ):
  dcl = DataConnection.objects.filter(id=dataconnection_id)
  if len(dcl) == 0:
    return HttpResponse("Invalid data connection information", status=500)
  dc = dcl[0]
  dc.checkPerm(user, 'dc_explore')
  # Fill in search settings
  search_enabled = _get_search_enabled(dataset_settings)

  # Finally, create the dataset
  d = DataSet(dataconnection=dc,
              dataset_name=dataset_name,
              dataset_type=dataset_type,
              dataset_info=dataset_info,
              dataset_lvinfo=dataset_lvinfo,
              dataset_detail=dataset_detail,
              dataset_description=dataset_description,
              dataset_settings=dataset_settings,
              search_enabled=search_enabled,
              dashboards=dashboards,
              version_name=version_name,
              version_group_id=version_group_id,
              is_active_version=is_active_version,
              create_user=create_user)
  d.update_user = user.username
  d.dataset_tablenames = json.dumps(d.get_tablenames())
  d.save()

  _auditDataset(d, ActionType.CREATE, request)

  return d

def dataset_create_or_clone(requestData, request):
  dataset_id = requestData.get("dataset_id", None)
  if dataset_id:
    return _dataset_clone(int(dataset_id), request)
  return dataset_create(requestData, request)

def update_foreign_keys_param(foreign_keys, conn_id, username):
  if foreign_keys is None:
    return
  else:
    foreign_keys = json.loads(foreign_keys)
  from dataconn.dataconn import get_dataconn
  dc = get_dataconn(conn_id)
  def set_foreign_key(p_table, child_table, fks, add_key):
    try:
      params = ','.join(["'fkey_%s.%s'='%s'" %
                          (p_table, i[0],
                            'pkey_%s.%s' % (child_table, i[1]) if add_key else '')
                         for i in fks])
      if params != '':
        query = "ALTER TABLE %s set TBLPROPERTIES (%s)" % (p_table, params)
        dc.sync_runquery(query=query, username=username)
    except Exception as e:
      msg = 'Cannot %s foreign key: %s' % ('add' if add_key else 'remove', str(e))
      LOGGER.exception(msg)
      raise Exception(msg)

  for pair in foreign_keys:
    for child_t, fk_dict in list(pair[1].items()):
      set_foreign_key(pair[0], child_t, fk_dict['add'], True)
      set_foreign_key(pair[0], child_t, fk_dict['rm'], False)

# Provide an admin interface for updating dataset info
def dataset_update_info(ds, new_info, new_lvname=None, foreign_keys=None, request=None):
  old_info = ds.dataset_info
  ds.dataset_info = new_info
  updatedLV = False
  # update lvinfo (ArcEngine only)
  if ds.dataconnection.dataconnection_type == 'arcengine':
    updatedLV = dataset_update_lvinfo(ds,
                                      ds.update_user,
                                      lvname=new_lvname,
                                      new_ds_info=new_info,
                                      old_ds_info=old_info)
    update_foreign_keys_param(foreign_keys,
                              ds.dataconnection.id,
                              ds.update_user)
  ds.dataset_tablenames = json.dumps(ds.get_tablenames())
  ds.save()
  _auditDataset(ds, ActionType.UPDATE, request=request)


  # Now check for renamed columns in this dataset
  try:
    _dataset_check_column_rename(ds, old_info, request)
  except Exception:
    pass
  return updatedLV

def _get_search_enabled(dataset_settings):
  return json.loads(dataset_settings).get('SEARCH', {}).get('enabled', False)

@transaction.atomic
@login_required
def swap_records(request):
  if request.method == "POST":
    master_id = request.POST.get('master_id')
    version_id = request.POST.get('version_id')
    master_version = DataSet.objects.get(id=master_id)
    version_to_revert = DataSet.objects.get(id=version_id)

    master_version.checkPerm(request.user, 'ds_manage')
    version_to_revert.checkPerm(request.user, 'ds_manage')

    master_uuid = getattr(master_version, 'uuid')
    version_to_revert_uuid = getattr(version_to_revert, 'uuid')

    for field in DataSet._meta.fields:
        field_name = field.name
        if field_name not in ['id', 'uuid', 'is_active_version']:
            placeholder = getattr(master_version, field_name)
            setattr(master_version, field_name, getattr(version_to_revert, field_name))
            setattr(version_to_revert, field_name, placeholder)


    setattr(master_version, 'uuid', 'placeholder')
    master_version.save()
    setattr(version_to_revert, 'uuid', master_uuid)
    version_to_revert.save()
    setattr(master_version, 'uuid', version_to_revert_uuid)
    master_version.save()
    return _dataset_get(request, dataset_id=master_id, version_group_id=master_id )
  else:
    return HttpResponse('Invalid request type. Only POST requests accepted', status=400)

def create_new_version(ds, request):
  dataset_dict = model_to_dict(ds)

  undesired_fields = ['id', 'dataconnection', 'uuid', ]
  for field in undesired_fields:
        dataset_dict.pop(field, None)

  dataset_dict['dataconnection_id'] = ds.dataconnection.id
  dataset_dict['version_group_id'] = ds.version_group_id
  dataset_dict['is_active_version'] = 0
  dataset_dict['create_time'] = ds.create_time

  return dataset_create(dataset_dict, request=request)

def _get_ds_update_data(request_user, ds, put):
  ds_update_data = DataSet()

  ds_update_data.version_name = put.get('version_name', None)
  ds_update_data.dataset_name = put.get("dataset_name", None)
  ds_update_data.dataset_type = put.get("dataset_type", None)
  ds_update_data.dataset_info = put.get("dataset_info", None)
  ds_update_data.dataset_detail = put.get("dataset_detail", None)
  ds_update_data.dataset_lvname = put.get("dataset_lvname", None)
  ds_update_data.dataset_dcid = put.get("dataconnection_id", None)
  ds_update_data.foreign_keys = put.get('foreign_keys', None)
  ds_update_data.is_named_version = True if put.get('is_named_version', False) == 'true' else False

  if ds_update_data.dataset_name is None:
    return HttpResponse("Invalid dataset name", status=500)
  if ds_update_data.dataset_type is None:
    return HttpResponse("Invalid dataset type", status=500)
  if ds_update_data.dataset_detail is None:
    return HttpResponse("Invalid dataset detail", status=500)

    # Check for switched dataconnection ID
  ds_update_data = _add_dc_info(ds_update_data, ds)

    # Some dataset updates require connection explore permission
  if 'dc_explore' not in ds_update_data.dc_forthisds.getPerms(request_user):
    # Like changing dataconnection id
    if ds.dataconnection_id != ds_update_data.dc_forthisds.id:
      ds_update_data.dc_forthisds.checkPerm(request_user, 'dc_explore')

    # Or changing the set of dataset tables:
    if ds.dataset_info:
      oldtabs = set(t['tablename'] for t in json.loads(ds.dataset_info))
    else:
      oldtabs = set([ds.dataset_detail])
    if ds_update_data.dataset_info:
      newtabs = set(t['tablename'] for t in json.loads(ds_update_data.dataset_info))
    else:
      newtabs = set([ds_update_data.dataset_detail])
    if oldtabs != newtabs:
      ds_update_data.dc_forthisds.checkPerm(request_user, 'dc_explore')

    # DJC -- Should validate info and dataset type here and other options
    #   type must be valid
    #   name can change but should not match other dataset
    #   info only valid for multipletables
    #   detail must be tablename for type = singletable
    # Check for duplicates

    if ds_update_data.dataset_info:
      old_custom_styles = []
      for table in json.loads(ds.dataset_info):
        for column in table.get('columns', []):
          display_format = column.get('displayformat', {})
          if display_format.get('type') == 'js':
            old_custom_styles.append(display_format)

      for table in json.loads(ds_update_data.dataset_info):
        for column in table.get('columns', []):
          display_format = column.get('displayformat', {})
          if display_format.get('type') == 'js' and display_format not in old_custom_styles:
            RBAC.checkPerm(request_user, 'sys_styles')
            enable_styles = reports.views_settings.setting_get_single('site_enable_custom_styles')
            if enable_styles == 'false':
              msg = 'Failed to update dataset. Custom JS styles can only be set if site-wide custom styles are enabled.'
              return HttpResponse(msg, status=403)
  return ds_update_data

def _add_dc_info(ds_update_data, ds):
  if ds.dataconnection_id is not None:
    ds_update_data.dcl_forthisds = DataConnection.objects.filter(id=int(ds.dataconnection_id))
    if len(ds_update_data.dcl_forthisds) == 0:
      return HttpResponse("Invalid dataconnection information", status=500)
    ds_update_data.dc_forthisds = ds_update_data.dcl_forthisds[0]
  else:
    ds_update_data.dc_forthisds = ds.dataconnection
  return ds_update_data

def _dataset_update(request, dataset_id=None, input=None, is_create_new_ds_version=True):
  """Update dataset"""
  if dataset_id is None:
    return HttpResponse('No dataset id specified', status=500)

  dataset_id = int(dataset_id)
  try:
    # pylint: disable=E1101
    ds = DataSet.objects.get(id=dataset_id)
    ds.checkPerm(request.user, 'ds_manage')

    if FeatureFlipper().get_flipper_value('enable_version_control') and FeatureFlipper().get_flipper_value('enable_dataset_version_control'):
      is_create_new_ds_version = request.GET.get('is_create_new_ds_version', 'true').lower() == 'true'
      if is_create_new_ds_version:

        new_ds_id = create_new_version(ds, request)
        new_ds = DataSet.objects.get(id=new_ds_id)
        new_ds_version_id = new_ds.version_id

        new_ds.create_time = ds.create_time
        new_ds.create_user = ds.create_user
        new_ds.save()

        segments = Segment.objects.filter(datasetsegment__dataset_id=dataset_id, datasetsegment__dataset_version_id=ds.version_id)
        if segments.exists():
          for segment in segments:
            dataset_segment_instance = DatasetSegment(dataset_id=dataset_id, segment_id=segment.id, dataset_version_id=new_ds_version_id)
            dataset_segment_instance.save()

        is_dataset_request = QueryDict(request.body).get('dataset_name', None)
        if not is_dataset_request:
          return new_ds_id

    put = input or QueryDict(request.body)
    if request.content_type == 'application/json':
      put = json.loads(request.body.decode('utf-8'))

    dsud = _get_ds_update_data(request.user, ds, put)

    # Finally, update the dataset
    ds.version_name = dsud.version_name or ds.version_name
    ds.dataset_name = dsud.dataset_name
    ds.dataset_type = dsud.dataset_type
    ds.dataset_detail = dsud.dataset_detail
    ds.update_user = request.user.username
    ds.dataset_description = put.get("description", ds.dataset_description)
    ds.dataset_settings = put.get("dataset_settings", ds.dataset_settings)
    ds.dataconnection = dsud.dc_forthisds
    ds.search_enabled = _get_search_enabled(ds.dataset_settings)
    ds.is_named_version = dsud.is_named_version
    setattr(ds, 'foreign_keys', getattr(dsud, 'foreign_keys', None))
    setattr(ds, 'dataset_lvname', getattr(dsud, 'dataset_lvname', None))

    # Update the info via common interface
    try:
      updatedLV = dataset_update_info(ds, dsud.dataset_info, new_lvname=dsud.dataset_lvname, foreign_keys=ds.foreign_keys, request=request)
      return _dataset_get(request, ds.id, updatedLV=updatedLV, is_create_new_ds_version=is_create_new_ds_version)
    except Exception as e:
      return HttpResponse(str(e), status=500)

  except DataSet.DoesNotExist:
    return HttpResponse('Could not find dataset %s' % dataset_id, status=500)

@login_required
def dataset(request, dataset_id=None):
  """
  REST endpoint for datasets
  """
  return dataset_internal(request, dataset_id)

def dataset_internal(request, dataset_id=None, input=None, filter=None):
  if request.method == 'GET':
    search = (request.GET.get('search', "false") == "true")
    get_counts = request.GET.get('get_associate_dashboard', 'false') == 'true'
    dataconnection_id = request.GET.get('dataconnection_id', None)
    return _dataset_get(request, dataset_id, filter=filter, search=search, get_counts=get_counts, dataconnection_id=dataconnection_id)

  if request.method == 'PUT':
    return _dataset_update(request, dataset_id, input)

  if request.method == 'POST':
    d_id_or_error = dataset_create_or_clone(request.POST, request=request)
    if type(d_id_or_error) == type(1):
      return _dataset_get(request, d_id_or_error)

    return d_id_or_error

  if request.method == 'DELETE':
    ret = _dataset_delete(request, dataset_id)
    return ret

  return HttpResponse('Invalid request method', status=500)

@login_required
def segment(request, segment_id=None):
  """
  REST endpoint for segments
  """
  return segment_internal(request, segment_id)

def segment_internal(request, segment_id=None, input=None, filter=None):
  if request.method == 'GET':
    return _segment_get_http(request, segment_id,
        dataset_id=request.GET.get("dataset_id", None), filter=filter)

  if request.method == 'PUT':
    return _segment_update(request, segment_id, input)

  if request.method == 'POST':
    return _segment_create(request)

  if request.method == 'DELETE':
    return _segment_delete(request, segment_id)

  return HttpResponse('Invalid request method', status=500)

def _datasetperm_add(unused_request, dataset_id):
  """ Set up custom permissions for the dataset - for ro, rw, setupperm"""

  perm = "setupperm"
  perm_code_name = "%s_%s_%s" % ("can", perm, int(dataset_id))
  perm_name = "%s %s %s" % ("Can", perm, int(dataset_id))

  # pylint: disable=E1101
  try:
    # If Permissions have already been set up for this dataset, nothing to do..
    Permission.objects.get(codename=perm_code_name)
    return
  except Permission.DoesNotExist:
    # Else, set up RO , RW, SP perms for this dataset as custom perms
    content_type = ContentType.objects.get_for_model(DataSet)
    permission = Permission.objects.create(codename=perm_code_name,
                                           name=perm_name,
                                           content_type=content_type)
    permission.save()

    for perm in ('readwrite', 'readonly'):
      perm_code_name = "%s_%s_%s" % ("can", perm, int(dataset_id))
      perm_name = "%s %s %s" % ("Can", perm, int(dataset_id))
      permission = Permission.objects.create(codename=perm_code_name,
                                             name=perm_name,
                                             content_type=content_type)
      permission.save()

def _userperm_get(request, dataset_id=None, query_user=None):
  """ Get all the users that have any permissions on this dataset """
  # Or this may be called to ask a specific user's permission on a dataset

  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)

  ds = DataSet.objects.get(id=dataset_id)
  ds.checkPerm(request.user, 'ds_appview')

  # If you can't see the dataset (have no perms on the dataset, not even RO),
  # then you can't see who has permissions on this dataset

  fastr = 'applied_dataset' + str(dataset_id) + '_fa'

  # pylint: disable=E1101
  if query_user is not None:
    # Append this user even if it has no permissions on this dataset since
    # it was asked 'what perm does this user have'
    json_data = []
    ul = User.objects.filter(username=query_user)
    if len(ul) == 0:
      return HttpResponse("No user called %s exists" % query_user, 500)

    permstring = DataSet.highest_user_perm_on_dataset(request.user.username,
                                                      dataset_id)
    faid = ''
    fp = ul[0].user_permissions.filter(codename__startswith=fastr)
    if len(fp) != 0:
      faid = (re.search("_fa(\d+)", fp[0].codename)).group(1)

    uu = {
        'username': query_user,
        'perm': permstring,
        'filterassociation_id': faid
    }
    json_data.append(uu)

  else:
    ds = DataSet.objects.get(id=dataset_id)
    ul = User.objects.all()
    datasetidstr = '_' + str(dataset_id)
    json_data = []
    # Return only the permissions that have been explicitly set in users,
    # not those users that might have permissions because their group has perms
    for u in ul:

      #append filter association if it exists
      faid = ''
      fp = u.user_permissions.filter(codename__startswith=fastr)
      if len(fp) != 0:
        faid = (re.search('_fa(\d+)', fp[0].codename)).group(1)

      if u.username == ds.create_user:
        permstring = ds.dataset_create_user_perm
        if len(permstring) != 0:
          uu = {
              'username': u.username,
              'perm': permstring,
              'filterassociation_id': faid
          }
          json_data.append(uu)
        continue

      po = u.user_permissions.filter(codename__endswith=datasetidstr)
      if len(po) != 0 and po[0].codename.startswith('can_'):
        permstring = (re.search('can_(\w+)_', po[0].codename)).group(1)
        uu = {
            'username': u.username,
            'perm': permstring,
            'filterassociation_id': faid
        }
        json_data.append(uu)

  return HttpResponse(json.dumps(json_data), content_type="application/json")

def _groupperm_get(request, dataset_id=None):
  """ Get all the groups that have any permissions on this dataset """

  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)

  ds = DataSet.objects.get(id=dataset_id)
  ds.checkPerm(request.user, 'ds_appview')

  fastartstr = 'applied_dataset' + str(dataset_id) + '_fa'

  gl = Group.objects.all()

  json_data = []
  for g in gl:
    faid = ''
    fp = g.permissions.filter(codename__startswith=fastartstr)
    if len(fp) != 0:
      faid = (re.search('_fa(\d+)', fp[0].codename)).group(1)

    permstring = DataSet.highest_group_perm_on_dataset(g.name, dataset_id)
    if permstring != 'none':
      gd = {
          'gname': g.name,
          'perm': permstring,
          'filterassociation_id': faid
      }
      json_data.append(gd)

  # pylint: disable=E1101
  faid = ''
  if ds.dataset_everyone_perm != 'none':
    gd = {
        'gname': 'everyone',
        'perm': ds.dataset_everyone_perm,
        'filterassociation_id': faid
    }
    json_data.append(gd)

  return HttpResponse(json.dumps(json_data), content_type="application/json")

def _userperm_add(request, dataset_id=None):
  """ Add a permission to a user """
  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)

  DataSet.objects.get(id=dataset_id).checkPerm(request.user, 'ds_manage')

  uname = request.POST.get("username", None)
  perm = request.POST.get("perm", None)
  faid = request.POST.get("filterassociation_id", None)
  if faid == 'None':
    faid = None

  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)
  if uname is None:
    return HttpResponse("Invalid user name", status=500)
  if perm is None:
    return HttpResponse("Invalid permission", status=500)

  # if this is first time, then custom perms will be set up for dataset
  _datasetperm_add(request, dataset_id)

  # Remove all permissions this user has already and then save the newly
  # added permission
  for eachperm in ('setupperm', 'readwrite', 'readonly'):
    user_type = ContentType.objects.get(app_label="auth", model="user")
    user_obj = user_type.get_object_for_this_type(username=uname)
    if DataSet.user_has_perms_on_dataset(uname, eachperm, dataset_id):
      # If this is because the group has permissions and not the user,
      # the following sentences end up being a NOOP
      perm_code_name = "%s_%s_%s" % ("can", eachperm, int(dataset_id))
      permission = Permission.objects.get(codename=perm_code_name)
      user_obj.user_permissions.remove(permission)
      user_obj.save()

  # remove existing fa perms that the user has because a user can have only
  # one FA applied
  fas = FilterAssociation.objects.all()
  for eachfa in fas:
    if FilterAssociation.user_has_filter_perms_on_dataset(eachfa.id,
                                                          uname, dataset_id):
      # If this is because the group has permissions and not the user,
      # the following sentences end up being a NOOP
      perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", int(dataset_id), "fa", eachfa.id)
      permission = Permission.objects.get(codename=perm_code_name)
      user_obj.user_permissions.remove(permission)
      user_obj.save()

  user_type = ContentType.objects.get(app_label="auth", model="user")
  user_obj = user_type.get_object_for_this_type(username=uname)

  # Add RW/RO/Modify perm
  perm_code_name = "%s_%s_%s" % ("can", perm, int(dataset_id))
  permission = Permission.objects.get(codename=perm_code_name)
  user_obj.user_permissions.add(permission)

  # Add FA perm
  if faid is not None:
    perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", dataset_id,
                                       "fa", faid)
    permission = Permission.objects.get(codename=perm_code_name)
    user_obj.user_permissions.add(permission)

  user_obj.save()

  return HttpResponse('[]', content_type="application/json")

def _groupperm_add(request, dataset_id=None):
  """ Add a permission to a group """
  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)

  gname = request.POST.get("groupname", None)
  perm = request.POST.get("perm", None)
  faid = request.POST.get("filterassociation_id", None)
  if faid == 'None':
    faid = None

  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)
  if gname is None:
    return HttpResponse("Invalid group name", status=500)
  if perm is None:
    return HttpResponse("Invalid permission", status=500)

  if not perm in ['readonly', 'readwrite', 'setupperm']:
    return HttpResponse("Invalid permission %s" % escape(perm), status=500)

  DataSet.objects.get(id=dataset_id).checkPerm(request.user, 'ds_manage')

  _datasetperm_add(request, dataset_id)

  ds = DataSet.objects.get(id=dataset_id)

  # Special case for the everyone group:
  if gname == 'everyone':
    # pylint: disable=E1101
    if faid is not None:
      return HttpResponse("No filter associations are allowed on the 'everyone'"
                          " group", status=500)
    ds.dataset_everyone_perm = perm
    ds.save()
    return HttpResponse('[]', content_type="application/json")

  # Group already has this permission on the dataset, return with success
  if DataSet.highest_group_perm_on_dataset(gname, dataset_id) == perm and \
      faid is None:
    return HttpResponse('[]', content_type="application/json")

  # Remove all permissions this group has already and then save the newly
  # added permission
  for eachperm in ('setupperm', 'readwrite', 'readonly'):
    group_type = ContentType.objects.get(app_label="auth", model="group")
    group_obj = group_type.get_object_for_this_type(name=gname)
    if DataSet.group_has_perms_on_dataset(gname, eachperm, dataset_id):
      perm_code_name = "%s_%s_%s" % ("can", eachperm, int(dataset_id))
      permission = Permission.objects.get(codename=perm_code_name)
      group_obj.permissions.remove(permission)
      group_obj.save()

  # Do removal of FA for groups etc
  fas = FilterAssociation.objects.filter(dataset=ds)
  for eachfa in fas:
    if FilterAssociation.group_has_filter_perms_on_dataset(eachfa.id, gname, dataset_id):
      perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", int(dataset_id), "fa", eachfa.id)
      permission = Permission.objects.get(codename=perm_code_name)
      group_obj.permissions.remove(permission)
      group_obj.save()

  group_type = ContentType.objects.get(app_label="auth", model="group")
  group_obj = group_type.get_object_for_this_type(name=gname)

  perm_code_name = "%s_%s_%s" % ("can", perm, int(dataset_id))
  permission = Permission.objects.get(codename=perm_code_name)
  group_obj.permissions.add(permission)

  # Add FA perm
  if faid is not None:
    perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", dataset_id,
                                       "fa", faid)
    permission = Permission.objects.get(codename=perm_code_name)
    group_obj.permissions.add(permission)

  group_obj.save()

  return HttpResponse('[]', content_type="application/json")

def _userperm_delete(request, dataset_id=None, uname=None):
  """ Remove a user from having any permissions on a dataset """
  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)

  DataSet.objects.get(id=dataset_id).checkPerm(request.user, 'ds_manage')

  if uname is None:
    return HttpResponse("Invalid user name", status=500)

  user_type = ContentType.objects.get(app_label="auth", model="user")
  user_obj = user_type.get_object_for_this_type(username=uname)

  for eachperm in ('setupperm', 'readwrite', 'readonly'):
    if DataSet.user_has_perms_on_dataset(uname, eachperm, dataset_id):
      perm_code_name = "%s_%s_%s" % ("can", eachperm, int(dataset_id))
      permission = Permission.objects.get(codename=perm_code_name)
      user_obj.user_permissions.remove(permission)

  # Remove any associated FAs too
  ds = DataSet.objects.get(id=dataset_id)
  fas = FilterAssociation.objects.filter(dataset=ds)
  for eachfa in fas:
    if FilterAssociation.user_not_group_has_filter_perms_on_dataset(eachfa.id, uname, dataset_id):
      perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", int(dataset_id), "fa", eachfa.id)
      permission = Permission.objects.get(codename=perm_code_name)
      user_obj.user_permissions.remove(permission)

  user_obj.save()
  return HttpResponse('[]', content_type="application/json")

def _groupperm_delete(request, dataset_id=None, gname=None):
  """ Remove a group from having any permissions on a dataset """
  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)

  if gname is None:
    return HttpResponse("Invalid group name", status=500)

  ds = DataSet.objects.get(id=dataset_id)
  ds.checkPerm(request.user, 'ds_manage')

  # Special case for the everyone group:
  if gname == 'everyone':
    # pylint: disable=E1101
    ds.dataset_everyone_perm = 'none'
    ds.save()
    return HttpResponse('[]', content_type="application/json")

  # Remove all permissions
  for eachperm in ('setupperm', 'readwrite', 'readonly'):
    if DataSet.group_has_perms_on_dataset(gname, eachperm, dataset_id):
      perm_code_name = "%s_%s_%s" % ("can", eachperm, int(dataset_id))
      permission = Permission.objects.get(codename=perm_code_name)
      group_type = ContentType.objects.get(app_label="auth", model="group")
      group_obj = group_type.get_object_for_this_type(name=gname)
      group_obj.permissions.remove(permission)
      group_obj.save()

  # Remove all FAs
  fal = FilterAssociation.objects.filter(dataset=ds)
  for eachfa in fal:
    if FilterAssociation.group_has_filter_perms_on_dataset(eachfa.id, gname, dataset_id):
      perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", int(dataset_id), "fa", eachfa.id)
      permission = Permission.objects.get(codename=perm_code_name)
      group_obj.permissions.remove(permission)
      group_obj.save()

  return HttpResponse('[]', content_type="application/json")

@login_required
def userperms(request, dataset_id=None, username_in=None):
  """
  REST endpoint for user permissions
  """
  if request.method == 'GET':
    return _userperm_get(request, dataset_id, username_in)

  if request.method == 'POST':
    ret = _userperm_add(request, dataset_id)
    return ret

  if request.method == 'DELETE':
    ret = _userperm_delete(request, dataset_id, username_in)
    return ret

  return HttpResponse('Invalid request method', status=500)

@login_required
def groupperms(request, dataset_id=None, groupname_in=None):
  """
  REST endpoint for group permissions
  """
  if request.method == 'GET':
    return _groupperm_get(request, dataset_id)

  if request.method == 'POST':
    ret = _groupperm_add(request, dataset_id)
    return ret

  if request.method == 'DELETE':
    ret = _groupperm_delete(request, dataset_id, groupname_in)
    return ret

  return HttpResponse('Invalid request method', status=500)

def _event_delete(request, event_id):
  """Delete an event"""
  if event_id is None:
    return HttpResponse('No event id specified', status=500)

  try:
    # pylint: disable=E1101
    e = Event.objects.get(id=event_id)

    e.dataset.checkPerm(request.user, 'ds_manage')

    e.delete()
    return HttpResponse('[]', content_type="application/json")
  except Event.DoesNotExist:
    return HttpResponse('[]', content_type="application/json")

def _event_update(request, event_id):
  """Update event details"""
  if event_id is None:
    return HttpResponse('No event id specified', status=500)

  event_id = int(event_id)
  try:
    # pylint: disable=E1101
    e = Event.objects.get(id=event_id)

    e.dataset.checkPerm(request.user, 'ds_manage')

    put = QueryDict(request.body)
    name = put.get('name', None)
    data = put.get('data', None)

    if name is None:
      return HttpResponse("Invalid name", status=500)
    if data is None:
      return HttpResponse("Invalid data", status=500)

    e.event_name = name
    e.event_data = data
    e.update_user = request.user.username
    e.save()

    json_data = []
    json_data.append(_get_event_data(e))
    return HttpResponse(json.dumps(json_data), content_type="application/json")
  except Event.DoesNotExist:
    return HttpResponse('Could not find event %s' % event_id, status=500)

def _event_create(request):
  """Create event"""

  dataset_id = request.POST.get("dataset_id", None)
  name = request.POST.get("name", None)
  data = request.POST.get("data", None)

  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)
  if name is None:
    return HttpResponse("Invalid name", status=500)
  if data is None:
    return HttpResponse("Invalid data", status=500)

  dataset_id = int(dataset_id)
  # pylint: disable=E1101
  try:
    d = DataSet.objects.get(id=dataset_id)
  except DataSet.DoesNotExist:
    return HttpResponse("Dataset %s does not exist" % dataset_id, status=500)

  d.checkPerm(request.user, 'ds_manage')

  e = Event(dataset=d, event_name=name, event_data=data,
            create_user=request.user.username)
  e.save()
  json_data = []
  json_data.append(_get_event_data(e))
  return HttpResponse(json.dumps(json_data), content_type="application/json")

def _get_event_data(e):
  """Return json'ified event data"""
  sd = {
      'id': e.id,
      'dataset_id': e.dataset_id,
      'name': e.event_name,
      'data': e.event_data,
      'created': e.create_time.isoformat(timespec='milliseconds'),
      'created_by': e.create_user,
      'updated': e.update_time.isoformat(timespec='milliseconds'),
      'updated_by': e.update_user,
  }
  return sd

def _event_get_http(request, event_id=None, dataset_id=None):
  """Get event details"""

  data, err = _event_get(request, event_id=event_id, dataset_id=dataset_id)
  if err:
    return HttpResponse(err, status=500)

  return HttpResponse(json.dumps(data), content_type="application/json")

def _event_get(request, event_id=None, dataset_id=None, dataset=None):
  """Get event details"""
  # We may be provided with a event id, a dataset id, or nothing
  # pylint: disable=E1101
  if event_id:
    # Event id provided - get information about that event
    el = Event.objects.filter(id=int(event_id))
  else:
    if dataset_id is None:
      return (None, "events: Dataset id not provided")

    # Dataset id provided - return all events for that dataset
    dataset_id = int(dataset_id)
    try:
      d = dataset or DataSet.objects.get(id=dataset_id)
    except DataSet.DoesNotExist:
      return (None, "Dataset %s does not exist" % dataset_id)

    d.checkPerm(request.user, 'ds_appview')

    el = Event.objects.filter(dataset=d)

  json_data = []
  for e in el:
    json_data.append(_get_event_data(e))

  return json_data, None

@login_required
def event(request, event_id=None):
  """
  REST endpoint for events
  """

  if request.method == 'GET':
    return _event_get_http(request, event_id,
        dataset_id=request.GET.get("dataset_id", None))

  if request.method == 'PUT':
    return _event_update(request, event_id)

  if request.method == 'POST':
    return _event_create(request)

  if request.method == 'DELETE':
    return _event_delete(request, event_id)

  return HttpResponse('Invalid request method', status=500)

def _filterassociation_delete(request, filterassociation_id):
  """Delete a filter association"""
  if filterassociation_id is None:
    return HttpResponse('No filterassociation id specified', status=500)

  try:
    # pylint: disable=E1101
    fa = FilterAssociation.objects.get(id=filterassociation_id)

    fa.dataset.checkPerm(request.user, 'ds_manage')

    # First delete the associated permission..
    perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", fa.dataset_id,
                                       "fa", fa.id)
    permission = Permission.objects.get(codename=perm_code_name)
    permission.delete()

    _auditFilterAssociation(fa, ActionType.DELETE, fa.dataset.dataconnection, request=request)

    # ..then delete the Filter Association
    fa.delete()

    return HttpResponse('[]', content_type="application/json")
  except FilterAssociation.DoesNotExist:
    return HttpResponse('[]', content_type="application/json")

# Return a tuple of users and groups and corresponding permission for this fa
def _filterassociation_members(fa):
  perm = Permission.objects.get(codename="applied_dataset%d_fa%s" % (fa.dataset_id, fa.id))
  uset = User.objects.filter(user_permissions=perm).values_list('id', flat=True)
  gset = Group.objects.filter(permissions=perm).values_list('id', flat=True)
  return (set(uset), set(gset), perm)

def _filterassociation_update(request, filterassociation_id, input=None):
  """Update filterassociation details"""
  if filterassociation_id is None:
    return HttpResponse('No filterassociation id specified', status=500)

  filterassociation_id = int(filterassociation_id)
  try:
    # pylint: disable=E1101
    fa = FilterAssociation.objects.get(id=filterassociation_id)

    fa.dataset.checkPerm(request.user, 'ds_manage')

    put = input or QueryDict(request.body)
    name = put.get('name', None)
    data = put.get('data', None)
    unew = put.get('users')
    gnew = put.get('groups')

    if name is None:
      return HttpResponse("Invalid name", status=500)
    if data is None:
      return HttpResponse("Invalid data", status=500)

    fa.filterassociation_name = name
    fa.filterassociation_data = data
    fa.update_user = request.user.username
    fa.save()

    _auditFilterAssociation(fa, ActionType.UPDATE, fa.dataset.dataconnection, request=request)

    (uold, gold, perm) = _filterassociation_members(fa)
    if unew is not None:
      unew = set(json.loads(unew))
      for uid in uold.difference(unew):
        User.objects.get(id=uid).user_permissions.remove(perm)
      for uid in unew.difference(uold):
        User.objects.get(id=uid).user_permissions.add(perm)
    if gnew is not None:
      gnew = set(json.loads(gnew))
      for gid in gold.difference(gnew):
        Group.objects.get(id=gid).permissions.remove(perm)
      for gid in gnew.difference(gold):
        Group.objects.get(id=gid).permissions.add(perm)

    json_data = []
    json_data.append(_get_filterassociation_data(fa))
    return HttpResponse(json.dumps(json_data), content_type="application/json")
  except FilterAssociation.DoesNotExist:
    return HttpResponse('Could not find filterassociation %s' %
                        filterassociation_id, status=500)

def _filterassociations_clone(request, old_dataset_id=None, new_dataset=None):
  fas = FilterAssociation.objects.filter(dataset__id=old_dataset_id)
  for old_fa in fas:
    (users, groups, _) = _filterassociation_members(old_fa)

    data = {
      "dataset_id": new_dataset.id,
      "name": old_fa.filterassociation_name,
      "data": old_fa.filterassociation_data,
      "users": json.dumps(list(users)),
      "groups": json.dumps(list(groups))
    }

    _filterassociation_create(data, request=request)

def _createFAPermission(dataset_id, fa_id):
  perm_code_name = "%s_%s%s_%s%s" % ("applied", "dataset", dataset_id,
                                     "fa", fa_id)
  perm_name = "%s %s %s %s %s"  % ("Applied", "dataset", dataset_id,
                                   "filter assocation", fa_id)
  content_type = ContentType.objects.get_for_model(DataSet)
  permission = Permission.objects.create(codename=perm_code_name,
                                         name=perm_name,
                                         content_type=content_type)
  permission.save()
  return permission

def getFAPermissionNames(dataset_id, fa_id):
  return (perm_code_name, perm_name)

def _filterassociation_create(requestData, request):
  """Create filterassociation"""

  dataset_id = requestData.get("dataset_id", None)
  name = requestData.get("name", None)
  data = requestData.get("data", None)
  users = requestData.get("users", None)
  groups = requestData.get("groups", None)

  if dataset_id is None:
    return HttpResponse("Invalid dataset", status=500)
  if name is None:
    return HttpResponse("Invalid name", status=500)
  if data is None:
    return HttpResponse("Invalid data", status=500)

  dataset_id = int(dataset_id)
  try:
    d = DataSet.objects.get(id=dataset_id)
  except DataSet.DoesNotExist:
    return HttpResponse("Dataset %s does not exist" % dataset_id, status=500)

  d.checkPerm(request.user, 'ds_manage')

  fa = FilterAssociation(dataset=d, filterassociation_name=name,
                         filterassociation_data=data,
                         create_user=request.user.username)
  fa.save()

  _auditFilterAssociation(fa, ActionType.CREATE, fa.dataset.dataconnection, request=request)

  # Also create a permission for this filter that may be applied to users
  # and groups
  permission = _createFAPermission(dataset_id, fa.id)

  # Only add users to the filter association that exist
  if users is not None:
    users = json.loads(users)
    for uid in users:
      try:
        User.objects.get(id=uid).user_permissions.add(permission)
      except User.DoesNotExist:
        pass
  if groups is not None:
    groups = json.loads(groups)
    for gid in groups:
      try:
        Group.objects.get(id=gid).permissions.add(permission)
      except Group.DoesNotExist:
        pass

  json_data = []
  json_data.append(_get_filterassociation_data(fa))
  return HttpResponse(json.dumps(json_data), content_type="application/json")

def _get_filterassociation_data(fa, detail=True):
  """Return json'ified filterassociation data"""

  fad = {
      'id': fa.id,
      'dataset_id': fa.dataset_id,
      'name': fa.filterassociation_name,
      'data': fa.filterassociation_data,
      'created': fa.create_time.isoformat(timespec='milliseconds'),
      'created_by': fa.create_user,
      'updated': fa.update_time.isoformat(timespec='milliseconds'),
      'updated_by': fa.update_user,
  }

  if detail:
    (uset, gset, perm) = _filterassociation_members(fa)
    fad['users'] = list(uset)
    fad['groups'] = list(gset)

  return fad

def _filterassociation_get_http(request, filterassociation_id=None,
    dataset_id=None, detail=True, filter=None):
  """Get filterassociation details"""
  data, err = _filterassociation_get(request,
      filterassociation_id=filterassociation_id, dataset_id=dataset_id,
      detail=detail, filter=filter)
  if err:
    return HttpResponse(err, status=500)

  return HttpResponse(json.dumps(data), content_type="application/json")

def _filterassociation_get(request, filterassociation_id=None,
    dataset_id=None, dataset=None, detail=True, filter=None):
  """Get filterassociation details"""

  # We may be provided with a filterassociation id, a dataset id, or nothing
  if filterassociation_id:
    # FilterAssociation id provided - get info about that filterassociation
    el = FilterAssociation.objects.filter(id=int(filterassociation_id))

  elif dataset_id is not None:
    # Dataset id provided - return all filterassociations for that dataset
    dataset_id = int(dataset_id)
    try:
      d = dataset or DataSet.objects.get(id=dataset_id)
    except DataSet.DoesNotExist:
      return (None, "Dataset %s does not exist" % dataset_id)

    # TODO: Viewing the app should not require the viz side loading
    # the fa so when we fix that, this check should be for 'ds_manage'
    if not is_dataset_superuser(request.user):
      d.checkPerm(request.user, 'ds_appview')
    el = FilterAssociation.objects.filter(dataset=d)

  else:
    if is_dataset_superuser(request.user):
      el = FilterAssociation.objects.all()
    else:
      idlist = get_dataset_ids_managable_by_user(request.user)
      el = FilterAssociation.objects.filter(dataset_id__in=idlist)
    if filter:
      el = el.filter(**filter)

  json_data = []
  for e in el:
    json_data.append(_get_filterassociation_data(e, detail=detail))
  return json_data, None

@login_required
def filterassociation(request, filterassociation_id=None):
  """
  REST endpoint for filterassociations
  """
  return filterassociation_internal(request, filterassociation_id)

def filterassociation_internal(request, filterassociation_id=None, input=None,
  detail=True, filter=None):
  if request.method == 'GET':
    return _filterassociation_get_http(request, filterassociation_id,
        dataset_id=request.GET.get("dataset_id", None), detail=detail, filter=filter)

  if request.method == 'PUT':
    return _filterassociation_update(request, filterassociation_id, input)

  if request.method == 'POST':
    return _filterassociation_create(request.POST, request)

  if request.method == 'DELETE':
    return _filterassociation_delete(request, filterassociation_id)

  return HttpResponse('Invalid request method', status=500)

@csrf_exempt
def dataset_cache_reset(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  if request.method != 'POST':
    return HttpResponse('Invalid request method', status=500)

  if not id:
    return HttpResponse('No ID specified', status=500)

  LOGGER.info("Clearing result cache for dataset %s", id)
  for ds in DataSet.objects.filter(id=id):
    ds.checkPerm(request.user, 'ds_manage')
    ds.cache_sequence += 1
    ds.update_user = request.user.username
    ds.save()
    _auditDataset(ds, ActionType.UPDATE, request=request)

  return HttpResponse()

@csrf_exempt
def dataconnection_cache_reset(request, id=None):

  err = check_login_or_use_apikey(request)
  if err:
    return err

  if request.method != 'POST':
    return HttpResponse('Invalid request method', status=500)

  if not id:
    return HttpResponse('No ID specified', status=500)

  for dc in DataConnection.objects.filter(id=id):
    dc.checkPerm(request.user, 'dc_explore')
    if dc.dataconnection_type == 'ksql':
      try:
        from dataconn.dataconn import get_dataconn
        dcobj = get_dataconn(dc.id)
        dcobj.resetAllQueries()
      except Exception as e:
        return HttpResponse(content="<pre>" + str(e) + "</pre>", status=500)
    dc.cache_sequence += 1
    dc.save()

  return HttpResponse()

def _get_datasetname_or_return_query_if_autogen(dataset):
  if dataset.dataset_description == "Auto-created for SQL apps" and (re.fullmatch("autogen [a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}", dataset.dataset_name) != None) and dataset.dataset_settings == "{}":
    return json.loads(dataset.dataset_info)[0]['select']
  return dataset.dataset_name

@login_required
def possible_deleted_reports(request):
	if request.method == 'GET':
		dataset_id = int(request.GET.get('dataset_id'))
		possible_delete_reports, not_owned_reports = DashboardAssociation().get_possible_removed_reports(dataset_id, request.user)

		reports_data = []
		for r in possible_delete_reports:
			rd = _get_report_data(r, request.user)
			reports_data.append(rd)

	public_attr = ['id', 'created', 'created_by', 'updated', 'updated_by', 'type', 'perm', 'workspace_name']
	for r in not_owned_reports:
		report = _get_report_data(r, request.user)
		report = dict((attr, report[attr]) for attr in public_attr if attr in report)
		reports_data.append(report)

	return HttpResponse(json.dumps(reports_data), content_type="application/json")

@login_required
def get_associated_dashboards(request, dataset_id=None):
  """Get associated dashboard ids"""
  if dataset_id is not None:
    associated_dashboard_ids =json.loads(DataSet.objects.get(id=dataset_id).dashboards)
    return HttpResponse(json.dumps({'dashboardIds': associated_dashboard_ids}), content_type="application/json")
  return HttpResponse(status=404)

@login_required
def sync_cml_connections(request):
    if request.method == 'GET':
        connections, status_code = CmlConnectionImporter().connect()

        if status_code == 200:
          return HttpResponse(json.dumps(connections), content_type='application/json', status=status_code)
        else:
          return HttpResponse(connections, status=status_code)

@login_required
def get_report_creation_permissions(request: HttpRequest):
  is_superuser = is_dataset_superuser(request.user)

  has_dataset_view_permission = is_superuser
  has_dataset_edit_permission = is_superuser

  if not is_superuser:
    datasets = get_dataset_query_for_permission_check()

    for dataset in datasets:
      perms = dataset.getPerms(request.user)

      if 'ds_appedit' in perms:
        has_dataset_view_permission = True
        has_dataset_edit_permission = True
        break

      if 'ds_appview' in perms:
        has_dataset_view_permission = True

  return JsonResponse({
    "canCreateDashboard": has_dataset_edit_permission,
    "canCreateApp": has_dataset_view_permission,
  })