# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from datasets.models import DataConnection, DataSet
from datasets.utils import _migration_dataconnection, _validate_request_and_get_dc, _auditDataConnection
from rbac.permcache import RBAC
from django.http import HttpResponse, HttpRequest
from vizlib.misc import reinitAsyncConnectionThreadpool
from vizlib.querylog import ActionType
from dataconn.connections_types import ALL_CONNECTION_TYPES
from django.contrib.auth.models import User
from django.db.models import F
from django.contrib.auth.decorators import login_required
from util.get_dataset_ids_accessible_by_user import get_dataset_query_for_permission_check, is_dataset_superuser
import json
import dataconn.dataconn as dataconn
import os
import logging

LOGGER = logging.getLogger("arcweb")


@login_required
def dataconnection(request, dataconnection_id=None):
    return dataconnection_internal(request, dataconnection_id)


def dataconnection_internal(request, dataconnection_id=None, input=None, filter=None):
    if request.method == 'GET':
        get_type = request.GET.get('get_type')
        import_data = request.GET.get('import_data')

        if get_type == 'all':
            order_by = request.GET.get('order_by')
            if order_by == 'last_used':
                # not every dataconnection can have cachestatistic model so in case it missing the connection update_time will be used
                return _dataconnection_get(request, None, None, order_by=[F('cachestatisticmodel__update_time').desc(nulls_last=None), '-update_time'], import_data=import_data)
            else:
                return _dataconnection_get(request, None)

        return _dataconnection_get(request, dataconnection_id, filter=filter, import_data=import_data)

    if request.method == 'PUT':
        return _dataconnection_update(request, dataconnection_id, input)

    if request.method == 'POST':
        return _dataconnection_create(request)

    if request.method == 'DELETE':
        return _dataconnection_delete(request, dataconnection_id)

    return HttpResponse('Invalid request method', status=500)


def _dataconnection_get(request, dataconnection_id, filter=None, order_by=['dataconnection_name'], import_data=None):
    """Get dataconnection details"""
    if dataconnection_id is None:
        dcl = DataConnection.objects.prefetch_related(
            'cachestatisticmodel_set').all().order_by(*order_by)
        if filter:
            dcl = dcl.filter(**filter)
        elif len(dcl) == 0:
            # We have no data connections. Auto-create one based on ARCSQL_SERVER
            # to support migrations of existing systems
            _migration_dataconnection(request=request)
            dcl = DataConnection.objects.all()
    else:
        dataconnection_id = int(dataconnection_id)
        dcl = DataConnection.objects.filter(id=dataconnection_id)

    json_data = []
    user_dc_ids = dc_ids_accessible_through_datasets(request.user)
    for dc in dcl:
        # If no connection level permissions, look for permissions on
        # some dataset for this connection
        if not does_user_have_acess_to_connection(request, dc, user_dc_ids):
            continue

        dataconnectionDto = createDataConnectionObject(
            dc, request.user, import_data)
        json_data.append(dataconnectionDto)

    val = json.dumps(limitResult(request, json_data))
    return HttpResponse(val, content_type="application/json")

def dc_ids_accessible_through_datasets(user: User):
  if is_dataset_superuser(user):
      return None

  accessible_dc_ids = set[str]()

  datasets = get_dataset_query_for_permission_check()
  for dataset in datasets:
    if 'ds_appview' in dataset.getPerms(user):
      accessible_dc_ids.add(dataset.dataconnection_id)

  return accessible_dc_ids


def _dataconnection_update(request, dataconnection_id, input=None):
    RBAC.checkPerm(request.user, 'sys_editconn')

    dataconnection_id = int(dataconnection_id)
    dc, error = _validate_request_and_get_dc(
        request, dataconnection_id=dataconnection_id, input=input)
    if not dc:
        return error

    dataconn.del_dataconn(dataconnection_id)
    dc.update_user = request.user.username
    dc.save()

    reinitAsyncConnectionThreadpool(dataconnection_id)

    _auditDataConnection(dc, ActionType.UPDATE, request=request)
    return HttpResponse(
        '{"status": "success", "id": %s, "name": "%s"}' % (dc.id, dc.dataconnection_name),
        content_type="application/json",
    )


def _dataconnection_delete(request, dataconnection_id):
    RBAC.checkPerm(request.user, 'sys_editconn')

    if not dataconnection_id:
        return HttpResponse('Must specify a data connection id to delete', status=500)

    dcl = DataConnection.objects.filter(id=int(dataconnection_id))
    if len(dcl) == 0:
        return HttpResponse("Data Connection with id '%s' does not exist" % dataconnection_id, status=500)
    dc = dcl[0]

    dl = DataSet.objects.filter(dataconnection__id=int(dataconnection_id))
    if len(dl) > 0:
        return HttpResponse('Cannot delete a data connection with datasets in it', status=500)

    dataconn.del_dataconn(dataconnection_id)
    _auditDataConnection(dc, ActionType.DELETE, request=request)
    dc.delete()
    return HttpResponse('[]', content_type="application/json")


def _dataconnection_create(request):
    RBAC.checkPerm(request.user, 'sys_editconn')

    dc, error = _validate_request_and_get_dc(request)
    if not dc:
        return error
    dc.create_user = request.user.username
    dc.update_user = request.user.username
    dc.save()
    _auditDataConnection(dc, ActionType.CREATE, request=request)
    return HttpResponse(
        '{"status": "success", "id": %s, "name": "%s"}' % (dc.id, dc.dataconnection_name),
        content_type="application/json",
    )


def clone_dataconnection(request: HttpRequest, dataconnection_id=None):
    try:
        dataconnection = DataConnection.objects.get(id=int(dataconnection_id))
        # copy and  recreate the dataconnection
        dataconnection.id = None
        dataconnection.dataconnection_name = f"Clone of {dataconnection.dataconnection_name}"
        dataconnection.create_user = request.user.username
        dataconnection.update_user = request.user.username
        dataconnection.cml_id = None
        dataconnection.cml_status = {}
        dataconnection.save()
        return HttpResponse(json.dumps(createDataConnectionObject(dataconnection, request.user, import_data=True)), content_type="application/json")
    except Exception as exception:
        LOGGER.error("Failed to clone dataconnection: %s", exception.__str__())
        return HttpResponse("Failed to clone dataconnection: %s" % exception.__str__(), status=500)


def does_user_have_acess_to_connection(request, connection: DataConnection, user_dc_ids: set[int]):
    perms = connection.getPerms(request.user)
    if len(perms) > 0:
        return True

    # If no connection level permissions, look for permissions on
    # some dataset for this connection
    if user_dc_ids and connection.id in user_dc_ids:
        return True

    # The usage analysis sets the permissions in dataconnection_settings field
    if connection.dataconnection_settings:
        try:
            permissions = json.loads(connection.dataconnection_settings)[
                'permissions']
            for permission in permissions:
                RBAC.checkPerm(request.user, permission)
            return True
        except:
            pass

    return False


def limitResult(request, array):
    limit = request.GET.get('limit')
    if limit:
        limit = min(int(limit), len(array))
        array = array[:limit]
    return array


def createDataConnectionObject(dataconnection: DataConnection, user: User, import_data) -> dict:
    perms = dataconnection.getPerms(user)
    info = dataconnection.dataconnection_info
    info_obj = json.loads(dataconnection.dataconnection_info)
    writable = True

    if 'FILE' in info_obj['PARAMS']:
        dataconnection_file_path = info_obj['PARAMS']['FILE']
        writable = os.access(dataconnection_file_path, os.W_OK)

    supportsImportData = False

    if import_data is not None:
        if writable:
            try:
                supportsImportData = dataconn.get_dataconn_type_properties(
                    connection_type=dataconnection.dataconnection_type).supports_dataextract_target
                pass
            except:
                pass

    if info_obj.get("PARAMS", {}).get("PASSWORD"):
        del info_obj["PARAMS"]["PASSWORD"]

    is_jwt_auth = info_obj.get("PARAMS", {}).get('AUTH') == 'JWT'
    info = json.dumps(info_obj)
    dataConnectionObject = {
        'id': dataconnection.id,
        'type': dataconnection.dataconnection_type,
        'typeName': ALL_CONNECTION_TYPES.get(dataconnection.dataconnection_type, dataconnection.dataconnection_type),
        'name': dataconnection.dataconnection_name,
        'description': dataconnection.dataconnection_description,
        'created': dataconnection.create_time.isoformat(timespec='milliseconds'),
        'created_by': dataconnection.create_user,
        'updated': dataconnection.update_time.isoformat(timespec='milliseconds'),
        'updated_by': dataconnection.update_user,
        'cmlId': dataconnection.cml_id,
        'info': info,
        'perms': list(perms),
        'cmlStatus': dataconnection.cml_status,
        'isPasswordSet': dataconnection.dataconnection_password is not None or is_jwt_auth,
        'lastUsed': get_last_used_time(dataconnection),
        'supportsImportData': supportsImportData,
        'writable': writable,
    }
    return dataConnectionObject


def get_last_used_time(dataconnection: DataConnection):
    cache_stat = None
    # this uses the prefetched data
    stats = dataconnection.cachestatisticmodel_set.all()

    if len(stats) > 0:
        cache_stat = stats[0]

    last_discovered = None
    if cache_stat is not None and cache_stat.update_time is not None:
        last_discovered = cache_stat.update_time.isoformat(
            timespec='milliseconds')
    else:
        last_discovered = dataconnection.update_time.isoformat(
            timespec='milliseconds')
    return last_discovered
