# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Base class for all data backend engines. Implement the interfaces in this
interface to implement a new data engine.
"""

from dataconn.dataconn import get_dataconn, sanitize_headers
from dataconn.connections_types import *
from sqlrun.models import CacheStatisticModel
from sqlrun.views_cache_statistic import CacheStatistics
from .datasource_csv import DataSourceCSV
from .datasource_csv import _MT_TO_DISPLAY
from .datasource_csv import ERROR_STATUS, FILLED_STATUS, SKIPPED_STATUS

import logging
import os
import re
import json
import datetime
from decimal import Decimal
from json import JSONEncoder


LOGGER = logging.getLogger("arcweb")

_DELIM_TO_TXT = {
  '\t': 'tab',
  '|': 'pipe',
  ' ': 'space',
  ',': 'comma',
  ';': 'semicolon',
  ':': 'colon',
  '\001': 'control-A',
  '\002': 'control-B',
  None: ''
}

_TXT_TO_DELIM = dict([ (v, k) for k, v in list(_DELIM_TO_TXT.items())])

def cmp_version(version1, version2):
  def normalize(v):
    return [int(x) for x in v.split(".")[:3]]

  version1 = normalize(version1)
  version2 = normalize(version2)
  return (version1 > version2) - (version1 < version2)

class MultipleJsonEncoders():
  """
  Combine multiple JSON encoders
  """
  def __init__(self, *encoders):
    self.encoders = encoders
    self.args = ()
    self.kwargs = {}

  def default(self, obj):
    for encoder in self.encoders:
      try:
        return encoder(*self.args, **self.kwargs).default(obj)
      except TypeError:
        pass
    raise TypeError(f'Object of type {obj.__class__.__name__} is not JSON serializable')

  def __call__(self, *args, **kwargs):
    self.args = args
    self.kwargs = kwargs
    enc = json.JSONEncoder(*args, **kwargs)
    enc.default = self.default
    return enc

class DateTimeEncoder(json.JSONEncoder):
  """JSON serializer for objects not serializable by default json code"""
  def default(self, obj):
    if isinstance(obj, (datetime.date, datetime.datetime)):
      return obj.isoformat()
    return super().default(obj)

class DecimalEncoder(json.JSONEncoder):
  def default(self, obj):
    if isinstance(obj, Decimal):
      return str(obj)
    return super().default(obj)

class DataImport(object):

  def __init__(self,
               uploadedfile=None,
               dataconnection_id=None,
               sourcefile=None,
               username=None,
               fill_missing_columns=False,
               data=None):
    self.tablename = None
    self.database = None
    self.location = None
    self.external_table = False
    self.error_report = None
    self.sample_rows = None
    self.uploadedfile = uploadedfile
    self.dataconnection_id = dataconnection_id
    self.fill_missing_columns = fill_missing_columns
    self.sourcefile = sourcefile
    self.username = username
    self.database = ""
    self.vector_similarity_fn = 'euclidean'
    self.hnsw_max_conn = '16'
    self.hnsw_beam_width = '100'
    if uploadedfile is not None:
      self.datasource_csv = DataSourceCSV(source_filename=uploadedfile)
      self.tablename = os.path.basename(self.datasource_csv.source_filename)
    else:
      self.datasource_csv = None
    if dataconnection_id is not None:
      self.dataconn = get_dataconn(self.dataconnection_id)
      if isImpalaBased(self.dataconn.connection_type):
        self.database = 'default'
      elif self.dataconn.connection_type in (DATACONN_SQLITE, DATACONN_DUCKDB):
        self.database = 'main'
      self.datasource_csv.dataconnection_type = self.dataconn.connection_type
    if data:
      self.update_values(data)

  def total_rows(self):
    if self.datasource_csv:
      return self.datasource_csv.total_rows

    return 0

  def reset_file(self, preserve_errors=False):
    self.datasource_csv.reset_file(preserve_errors=preserve_errors)
    self.datasource_csv.headers = sanitize_headers(self.datasource_csv.headers,
                                                    self.dataconn.connection_type)

  def check_for_actions(self, data):
    redo_header = False
    recalculate = False
    keys = list(data.keys())
    if 'datasource_has_header' in keys:
      old_has_header = True if int(data.get('datasource_has_header')) else False
      if old_has_header != self.datasource_csv.has_header:
        recalculate = True
        redo_header = True

    if 'datasource_headers' in keys:
      header_dict = data.get('datasource_headers')
      old_coltypes = [c["coltype"] for c in header_dict]
      if old_coltypes != self.datasource_csv.types:
        recalculate = True

    if 'datasource_col_delimiter' in keys:
      try:
        old_col_delim = _TXT_TO_DELIM[data.get('datasource_col_delimiter', '')]
      except KeyError:
        old_col_delim = str(data.get('datasource_col_delimiter', ''))
      if old_col_delim != self.datasource_csv.delimiter:
        recalculate = True
        redo_header = True

    if 'datasource_locale' in keys:
      old_localesetting = data.get('datasource_locale')
      if old_localesetting != self.datasource_csv.localesetting:
        recalculate = True

    if 'datasource_skip_bad_rows' in keys:
      old_skipped_bad_rows = True if int(
        data.get('datasource_skip_bad_rows')) else False
      if old_skipped_bad_rows != self.datasource_csv.skip_bad_rows and\
          old_skipped_bad_rows is True:
        recalculate = True
      elif old_skipped_bad_rows != self.datasource_csv.skip_bad_rows and \
          old_skipped_bad_rows is False and self.error_report["report"]:
        # this is a hack to avoid calling get_bad_rows
        # if the skipped_bad_rows was False and user changes to True, then we
        # know that there won't be any erro, so we zero the number of error
        # lines and set the skipped lines to number of error lines. This way
        # we avoid calling get_bad_rows function
        self.error_report["report"] = {ERROR_STATUS:0,
                                       SKIPPED_STATUS:self.error_report["report"][ERROR_STATUS],
                                       FILLED_STATUS:0}

    if 'datasource_fill_missing_columns' in keys:
      old_fill_missing_columns = True if int(
        data.get('datasource_fill_missing_columns')) else False
      if old_fill_missing_columns != self.fill_missing_columns:
        recalculate = True

    if 'datasource_escapechar' in keys:
      old_escapecahr = data.get('datasource_escapechar')
      if old_escapecahr != self.datasource_csv.escapechar:
        recalculate = True

    return recalculate, redo_header

  def update_values(self, data):
    keys = list(data.keys())
    if 'datasource_uploaded_file' in keys and not self.datasource_csv:
      self.datasource_csv = DataSourceCSV(source_filename=data.get('datasource_uploaded_file'))
      self.uploadedfile = data.get('datasource_uploaded_file')

    if 'datasource_dataconnection_id' in keys:
      self.dataconnection_id = data.get('datasource_dataconnection_id')
      self.dataconn = get_dataconn(self.dataconnection_id)
      self.datasource_csv.dataconnection_type = self.dataconn.connection_type

    if 'datasource_dataconnection_name' in keys:
      self.dataconnection_name = data.get('datasource_dataconnection_name')

    if 'datasource_tablename' in keys:
      self.tablename = data.get('datasource_tablename')

    if 'datasource_database' in keys:
      self.database = data.get('datasource_database')

    self.location = data.get('datasource_location', None)

    if 'datasource_external_table' in keys:
      self.external_table = data.get('datasource_external_table') == "1"

    if 'datasource_src_file' in keys:
      self.sourcefile = data.get('datasource_src_file')

    if 'datasource_has_header' in keys:
      self.datasource_csv.has_header = True if int(data.get('datasource_has_header')) else False

    if 'datasource_headers' in keys:
      header_dict = data.get('datasource_headers')
      self.datasource_csv.headers = [c["colname"] for c in header_dict]
      self.datasource_csv.types = [c["coltype"] for c in header_dict]

    if 'datasource_col_delimiter' in keys:
      try:
        self.datasource_csv.delimiter = _TXT_TO_DELIM[data.get('datasource_col_delimiter', '')]
      except KeyError:
        self.datasource_csv.delimiter = str(data.get('datasource_col_delimiter', ''))

    if 'datasource_num_header_rows' in keys:
      self.datasource_csv.header_offset = int(data.get('datasource_num_header_rows'))

    if 'datasource_locale' in keys:
      self.datasource_csv.localesetting = str(data.get('datasource_locale'))

    if 'datasource_vector_similarity_fn' in keys:
      self.vector_similarity_fn = str(data.get('datasource_vector_similarity_fn'))

    if 'datasource_hnsw_max_conn' in keys:
      self.hnsw_max_conn = str(data.get('datasource_hnsw_max_conn'))

    if 'datasource_hnsw_beam_width' in keys:
      self.hnsw_beam_width = str(data.get('datasource_hnsw_beam_width'))

    if 'datasource_skip_bad_rows' in keys:
      self.datasource_csv.skip_bad_rows = True if int(
        data.get('datasource_skip_bad_rows')) else False

    if 'datasource_fill_missing_columns' in keys:
      self.fill_missing_columns = True if int(
        data.get('datasource_fill_missing_columns')) else False

    if 'datasource_escapechar' in keys:
      self.datasource_csv.escapechar = "\\" if int(
        data.get('datasource_escapechar')) else None

    if 'datasource_sample_rows' in keys:
      self.sample_rows = data.get('datasource_sample_rows')

    if 'datasource_errors' in keys:
      self.error_report = data.get('datasource_errors')

  def get_dict(self, recalculate=True):
    """ Get a dict representing imported data"""
    try:
      delimiter = _DELIM_TO_TXT[self.datasource_csv.delimiter]
    except KeyError:
      delimiter = self.datasource_csv.delimiter

    data = {
      'datasource_id': 1,
      'datasource_has_header': 1 if self.datasource_csv.has_header else 0,
      'datasource_col_num': self.datasource_csv.col_num,
      'datasource_col_delimiter': delimiter,
      'datasource_num_header_rows': self.datasource_csv.header_offset,
      'datasource_type': self.datasource_csv.types,
      'datasource_uploaded_file': self.datasource_csv.source_filename,
      'datasource_src_file': self.sourcefile,
      'datasource_tablename': self.tablename,
      'datasource_database': self.database,
      'datasource_location': self.location,
      'datasource_external_table': self.external_table,
      'datasource_locale': self.datasource_csv.localesetting,
      'datasource_skip_bad_rows': 1 if self.datasource_csv.skip_bad_rows else 0,
      'datasource_fill_missing_columns': 1 if self.fill_missing_columns else 0,
      'datasource_escapechar': 1 if self.datasource_csv.escapechar else 0,
      'datasource_dataconnection_id': self.dataconnection_id,
      'datasource_dataconnection_name': self.dataconn.connection_name,
      'datasource_dataconnection_type': self.dataconn.connection_type,
      'datasource_headers': list(map(lambda x, y: {'colname':x, 'coltype':y},
                                self.datasource_csv.headers, self.datasource_csv.types)),
      'datasource_vector_similarity_fn': self.vector_similarity_fn,
      'datasource_hnsw_max_conn': self.hnsw_max_conn,
      'datasource_hnsw_beam_width': self.hnsw_beam_width
    }
    if recalculate:
      self.error_report = self.datasource_csv.get_bad_rows_report(
                              fill_missing_columns=self.fill_missing_columns)
      if self.error_report["report"]["error"] > 1000:
        self.error_report["bad_rows"] = self.error_report["bad_rows"][0:1000]
    self.sample_rows = []
    for row in self.datasource_csv.sample_rows(
                            count=100,
                            fill_missing_columns=self.fill_missing_columns):
      self.sample_rows.append(["%s" % c.value if c.value is not None
                                              else "NULL" for c in row])
    data['datasource_sample_rows'] = self.sample_rows
    data['datasource_errors'] = self.error_report
    return data

  @property
  def _get_full_tablename(self):
    if self.dataconn.connection_type in (DATACONN_SQLITE, DATACONN_DUCKDB, DATACONN_SOLR):
      fq_table_name = self.tablename
    else:
      fq_table_name = "%s.%s" % (
        self.database, self.tablename)
    return fq_table_name

  @property
  def create_table_cmd(self):
    return self.dataconn.create_sql_cmds(self._get_full_tablename,
                                        self.datasource_csv.headers,
                                        self.datasource_csv.types,
                                        self.username,
                                        self.datasource_csv.source_filename,
                                        self.external_table,
                                        self.location,
                                        localesetting=self.datasource_csv.localesetting,
                                        type_classes=self.datasource_csv._guessed_messytable_types,
                                        vector_similarity_fn=self.vector_similarity_fn,
                                        hnsw_max_conn=self.hnsw_max_conn,
                                        hnsw_beam_width=self.hnsw_beam_width)

  def drop_table(self, data):
    try:
      sqlcmd = self.dataconn.drop_table_cmd(self._get_full_tablename)
      self.dataconn.log_sync_runquery(sqlcmd, username=self.username)
    except Exception as e:
      error_message = 'Drop table failed'
      LOGGER.exception(error_message)
      msg = "%s: %s" % (error_message, str(e))
      raise Exception(msg)

  def dataconnection(self):
    return self.dataconn

  def create_table(self, data):
    importtype = data.get('datasource_importtype', 'create')
    if getattr(settings, 'ENABLE_DATAIMPORT_RW', False) == False:
      if importtype != 'create':
        raise Exception('Overwrite or Append in Data Import is not allowed')

    if importtype == 'append':
      return
    if importtype == 'overwrite':
      self.drop_table(data)
    try:
      sqlcmd = self.create_table_cmd
      self.dataconn.log_sync_runquery(sqlcmd, username=self.username)
    except Exception as e:
      error_message = 'Creating table failed'
      LOGGER.exception(error_message)
      e = str(e)
      msg = "%s: %s" % (error_message, e)
      raise Exception(msg)

  def insert_cmd(self, count=None):
    insert_command = self.dataconn.insert_table_prefix(self._get_full_tablename)
    sqlcmd = insert_command + " values "
    if count is None:
      count = self.datasource_csv.total_rows
    while True:
      insert_values = []
      for row in self.datasource_csv.row_batch(count=count,
                                fill_missing_columns=self.fill_missing_columns):
        insert_values.append("(%s)" % ", ".join(
          ["%s" % self.dataconn.apply_types(c.value, _MT_TO_DISPLAY[type(c.type)])
                    for c in row]))
      if not insert_values:
        break
      yield sqlcmd + ",".join(insert_values)

  def insert_data_teradata(self, batch_count):
    if batch_count is None:
      batch_count = self.datasource_csv.total_rows

    queryl = ["INSERT INTO %s values " % \
      self.dataconn.quote_tablename(self._get_full_tablename)]
    firstTime = True

    try:
      loop_iter = 0

      while True:
        loop_iter += 1
        rows = self.datasource_csv.row_batch(count=batch_count,
            fill_missing_columns=self.fill_missing_columns)

        insertRows = []
        for row in rows:
          insertRows.append([self.dataconn.apply_types(c.value, _MT_TO_DISPLAY[type(c.type)])
              for c in row])

        if len(insertRows) == 0:
          break

        if firstTime:
          firstTime = False
          queryl.append('(')
          queryl.extend(['?,' for x in insertRows[0]])
          queryl[len(queryl) - 1] = '?)'
          query = ''.join(queryl)

        self.dataconn.executemany(username, query, insertRows, batch=True)
        CacheStatistics.updateQueryStatisticHistory(self.dataconnection_id, False)

    except Exception as e:
      msg = "Insert data failed for line %s to line %s: %s" % \
            ((loop_iter -1) * batch_count,
              loop_iter * batch_count,
              str(e))
      LOGGER.exception(msg)
      raise Exception(msg)

  def insert_data_solr(self, batch_count):
    if batch_count is None:
      batch_count = self.datasource_csv.total_rows

    colnames = self.datasource_csv.headers

    try:
      loop_iter = 0

      while True:
        req = {
          'url': f'{self.dataconn.baseurl}{self._get_full_tablename}/update?commit=true&wt=json',
          'body': []
        }
        loop_iter += 1
        rows = self.datasource_csv.row_batch(count=batch_count,
            fill_missing_columns=self.fill_missing_columns)

        for row in rows:
          row_obj = {}
          for idx, colname in enumerate(colnames):
            value = row[idx].value
            row_obj[colname] = value
          req['body'].append(row_obj)

        if len(req['body']) == 0:
          break

        # Serialize the Decimal or DateTime object using custom JSON encoder
        insert_cmd = 'json:'+json.dumps(req, cls=MultipleJsonEncoders(DateTimeEncoder, DecimalEncoder))
        self.dataconn.log_sync_runquery([insert_cmd], username=self.username)
    except Exception as e:
      msg = "Insert data failed for line %s to line %s: %s" % \
            ((loop_iter -1) * batch_count,
              loop_iter * batch_count,
              str(e))
      LOGGER.exception(msg)
      raise Exception(msg)

  def insert_data(self, batch_count):
    if self.dataconn.connection_type == DATACONN_TERADATA:
      return self.insert_data_teradata(batch_count)

    if self.dataconn.connection_type == DATACONN_SOLR:
      return self.insert_data_solr(batch_count)

    try:
      loop_iter = 1
      if self.dataconn.connection_type == DATACONN_SQLITE and \
          cmp_version(self.dataconn.properties().version, '3.7.11') < 0:
        # if sqlite version is lower than 3.7.11 disable multi row insert
        # because it is not supported
        batch_count = 1
      elif not self.dataconn.properties().supports_multi_row_insert:
        batch_count = 1

      for loop_iter, sqlcmd in enumerate(self.insert_cmd(batch_count), 1):
        self.dataconn.sync_runquery(sqlcmd, username=self.username,
                                    logQuery=False, hideQuery=True)
    except Exception as e:
      msg = "Insert data failed for line %s to line %s: %s" % \
            ((loop_iter -1) * batch_count,
              loop_iter * batch_count,
              str(e))
      LOGGER.error(msg)
      raise Exception(msg)

  def compute_state_if_possible(self):
    if isArcEngineBased(self.dataconn.connection_type):
      try:
        sqlcmd = 'compute incremental stats %s' % self._get_full_tablename
        self.dataconn.sync_runquery(sqlcmd, username=self.username)
      except Exception as e:
        # dont' fail the import if compute stat is failing
        LOGGER.exception('Compute stat failed on %s table: %s' %
                         (self._get_full_tablename, str(e)))

  def remove_uploaded_file(self):
    if os.path.exists(self.uploadedfile):
      try:
        os.remove(self.uploadedfile)
      except Exception as e:
        LOGGER.error("could not remove uploaded file: %s" % str(e))
        pass
