# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import csv
import decimal
import locale
import logging
import os

from django.utils.encoding import smart_str
from inspect import isclass

from messytables import CSVRowSet, headers_guess, offset_processor
from messytables import StringType, BoolType
from messytables.types import CellType
import messytables.commas
from collections import defaultdict
import datetime
import dateutil.parser as parser

from dataconn.connections_types import supportsVectorDataType

# create logger
LOGGER = logging.getLogger("arcweb")

try:
  from messytables.core import Cell
except Exception as e:
  LOGGER.exception(str(e))

_TRUTHYVALUES = ['yes', '1', 'true', 't', 'y', 'on']
_FASLYVALUES = ['no', '0', 'false', 'f', 'n', 'off']


ERROR_STATUS = "error"
SKIPPED_STATUS = "skipped"
FILLED_STATUS = "filled"
PASS_STATUS = "no_error"



class DataSourceCSV():
  """DataSource CSV"""

  def __init__(self,
               source_filename=None,
               source_filehandle=None,

               dataconnection_id=None,
               dataconnection_type=None,

               types=None,
               headers=None,
               header_offset=None,
               has_header=None,

               delimiter=None,
               quotechar=None,
               encoding="utf-8",
               doublequote=True,
               lineterminator=None,
               skipinitialspace=None,
               escapechar=None,
               localesetting='C',
               skip_bad_rows=False):

    # Parameters
    self.source_filename = source_filename or None
    self.source_filehandle = source_filehandle or None
    self.dataconnection_type = dataconnection_type

    # defaults
    self.types = types or None
    self.headers = headers or None
    self.delimiter = delimiter or None
    self.quotechar = quotechar or None
    self.encoding = encoding or None
    self.doublequote = doublequote or True
    self.lineterminator = lineterminator or None
    self.skipinitialspace = skipinitialspace or None
    self.escapechar = escapechar or None
    self.skip_bad_rows = skip_bad_rows
    self.has_header = has_header or None
    self.localesetting  = localesetting

    # Read Only Properties
    self.row_counter = 0
    self.total_rows = 0
    self.byte_counter = 0
    self.total_bytes = 0
    # should we make this local variables to reset_file function
    self.guessed_headers = None
    self.guessed_types = None
    self.col_num = None


    # Private Properties
    self.window = 10000
    self.header_offset = header_offset or None
    # self._guessed_num_header_rows = None
    self._guessed_messytable_types = None
    self._row_set = None
    self._csvfile = None
    self._rowiter = None
    self._col_error = []
    self._row_error = None

    self.dataconnection_type = dataconnection_type or None

  def reset_file_handler(self):
    self._csvfile.seek(0, os.SEEK_SET)

  def reset_file(self, preserve_errors=False):
    locale.setlocale(locale.LC_NUMERIC, self.localesetting)
    if self._csvfile:
      # reset file to beginning
      self._csvfile.seek(0, os.SEEK_SET)

    self.row_counter = 0
    self.total_rows = 0
    self.byte_counter = 0
    self.total_bytes = 0
    self._rowiter = None
    if not preserve_errors:
      self.rows_report = {}
    if self._csvfile is None or self._csvfile.closed:
      if self.source_filehandle is not None:
        self._csvfile = self.source_filehandle
      else:
        self._csvfile = open(self.source_filename, 'rUb')

    # get line count
    for self.total_rows, l in enumerate(self._csvfile, 1):  # note: may want to use linedelimiter
      pass
    # Seek to end of file to count bytes
    self._csvfile.seek(0, os.SEEK_END)
    self.total_bytes = self._csvfile.tell()
    # reset file to beginning
    self._csvfile.seek(0, os.SEEK_SET)
    self._row_set = CSVRowSetLocal("table", self._csvfile,
                            delimiter=self.delimiter,
                            quotechar=self.quotechar,
                            encoding=self.encoding,
                            window=self.window,
                            doublequote=self.doublequote,
                            lineterminator=self.lineterminator,
                            skipinitialspace=self.skipinitialspace,
                            escapechar=self.escapechar)
    # Guess headers and find offset
    self._guessed_header_offset, guessed_headers = headers_guess(self._row_set.sample)

    # Add offset processor to skip headers. Add one to begin with content, not the header:
    if self.header_offset is None:
      self.header_offset = self._guessed_header_offset

    # if the has_header attribute was set to False
    if self.has_header is False:
      guessed_headers = [None] * len(guessed_headers)
    # if no information is available for has_header
    elif self.has_header is None:
      self.has_header = False if self.total_rows == 1 else self._guessed_has_header
      if self.has_header is False:
        guessed_headers = [None] * len(guessed_headers)

    if self.headers is None:
      self.headers = guessed_headers

    self._row_set.register_processor(
      offset_processor(self.header_offset + 1 if self.has_header else 0))

    # Guess Types
    GUESSTYPES = [ self.mt_instance(_DISPLAY_TO_MT[t]) for t in list(_MT_TO_DISPLAY.values()) ]
    self._guessed_messytable_types = type_guess(self._row_set.sample, types=GUESSTYPES,
                                                strict=True, dataconnection_type=self.dataconnection_type)
    self.guessed_types = [_MT_TO_DISPLAY[type(k)] for k in self._guessed_messytable_types]

    if self.types is None:
      self.types = self.guessed_types
    self._row_set.register_processor(types_processor(self.mt_types,
                                                     error=self._col_error,
                                                     strict=True))

    # not registering the headers processor because the information of how many
    # columns each row has will be lost. header processor is filling the missing
    # columns with None
    #self._row_set.register_processor(headers_processor(self.headers))

    if self.delimiter is None:
      self.delimiter = self._row_set.delimiter

    if self.lineterminator is None:
      self.lineterminator = self._row_set.lineterminator

    if len(self.headers) == len(self.types):
      self.col_num = len(self.headers)
    else:
      # taking the maximum occurrence of column numbers as the column number
      self.col_num = max(list(self.sample_stat.items()), key=lambda k_v: k_v[1])[0]
      # making header and type the same size as self.col_num
      self.headers += [None] * (self.col_num - len(self.headers))
      self.headers =  self.headers[:self.col_num]
      self.types += [_MT_TO_DISPLAY[StringType]] * (self.col_num - len(self.types))
      self.types =  self.types[:self.col_num]

  def row_status(self, row, fill_missing_columns=True):
    # are we allowed to handle missing columns
    if self._row_error:
      return ERROR_STATUS
    col_num = len(row)
    if self._col_error or \
        col_num > self.col_num or \
        (col_num < self.col_num and not fill_missing_columns):
      return SKIPPED_STATUS if self.skip_bad_rows else ERROR_STATUS
    elif col_num < self.col_num and fill_missing_columns:
      return FILLED_STATUS
    else:
      return PASS_STATUS

  def sample_rows(self, count, fill_missing_columns=True):
    sample_iter = iter(self.rows(fill_missing_columns=fill_missing_columns))
    for i in range(count):
      row = next(sample_iter, None)
      if row:
        yield row

  @property
  def sample_stat(self):
    col_num_stat = {}
    for row in self.row_batch(count=self.window, handle_bad_rows=False):
      row_len = len(row)
      if row_len in col_num_stat:
        col_num_stat[row_len] += 1
      else:
        col_num_stat[row_len] = 1

    self._rowiter = None
    return col_num_stat

  def rows(self, handle_bad_rows=True, fill_missing_columns=True):
    """
    Wrapper for row_set that keeps track of line number and
    approximate bytes read.
    Also keep track of typecast errors
    """
    try:
      for self.row_counter, row in enumerate(self._row_set, 1):
        self._row_error = False
        if handle_bad_rows:
          if self.skip_bad_rows and self._col_error:
            continue
          if self.skip_bad_rows and self.col_num != len(row) and not fill_missing_columns:
            continue
          if fill_missing_columns:
            row = row + [Cell(None, type=self.mt_instance(_DISPLAY_TO_MT[t]))
                            for t in self.types[len(row):]]
        yield row
    except Exception as e:
      self._row_error = True
      self.reset_file()
      yield []

  def row_batch(self, count=10, handle_bad_rows=True, fill_missing_columns=True):
    if not self._rowiter:
      self._rowiter=iter(self.rows(handle_bad_rows=handle_bad_rows,
                                   fill_missing_columns=fill_missing_columns))
    for i in range(count):
      row = next(self._rowiter, None)
      if row:
        yield row

  def get_bad_rows_report(self, fill_missing_columns=True):
    bad_rows = []
    report = {ERROR_STATUS:0,
              SKIPPED_STATUS:0,
              FILLED_STATUS:0,
              PASS_STATUS:0}
    for row_counter, row in enumerate(self.rows(handle_bad_rows=False), 1):
      status = self.row_status(row, fill_missing_columns)
      report[status] += 1
      if status == ERROR_STATUS:
        error = {"num_of_col":len(row),
                 "type_errors":["Cannot read the line. Parsing has stopped. "
                                "This might be due to missing quotes"]
                    if self._row_error else list(self._col_error),
                 "status":status}
        bad_rows.append({row_counter:error})

    return {"report":report,
            "bad_rows":bad_rows}

  def mt_instance(self, cls):
    """
    This creates a instance of a Messytables Type.
    It handles passing arguments to __init__ if needed.
    """
    if cls is _DISPLAY_TO_MT["Boolean"]:
      return cls(true_values=_TRUTHYVALUES, false_values=_FASLYVALUES)
    else:
      return cls()

  @property
  def mt_types(self):
    return [ self.mt_instance(_DISPLAY_TO_MT[t]) for t in self.types ]

  @property
  def py_types(self):
    return [t.result_type for t in self.mt_types]

  @property
  def js_headers(self):
    return list(zip(self.headers,self.type))

  @property
  def _guessed_has_header(self):
    return self._row_set.has_header




class DecimalType(CellType):
  """ Decimal number, ``decimal.Decimal`` or float numbers. """
  guessing_weight = 4
  result_type = decimal.Decimal

  def cast(self, value):
    if value in ('', None):
      return None
    try:
      return decimal.Decimal(value)
    except:
      value = locale.atof(value)
      # converting value to string;
      # did not use str because it may change the value
      value = "%s" % value
      return decimal.Decimal(value)

# rewriting IntergerType to handle Norwegian integer
class IntegerType(CellType):
  """ An integer field. """
  guessing_weight = 6
  result_type = int

  def cast(self, value):
    if value in ('', None):
      return None

    try:
      value = locale.atof(value)
    except:
      return locale.atoi(value)

    if value.is_integer():
      return int(value)
    else:
      raise ValueError()

# class DateUtilType(DateUtilType):
class DateUtilType(CellType):
  """ The date util type uses the dateutil library to
  parse the dates. The advantage of this type over
  DateType is the speed and better date detection. However,
  it does not offer format detection.

  Do not use this together with the DateType

  Modified to return a datetime.time if only a time is given
  """
  guessing_weight = 3
  result_type = datetime.datetime

  class NoDefaultDate(object):
    # This class is used as the default value for the dateutil.parser
    # to allow the parser to return time values without defaulting to the
    # current date
    year=datetime.MINYEAR
    month=1
    day=1

    def replace(self, **fields):
      if not set(fields).issuperset(set(['year', 'month', 'day'])):
        return datetime.time(**dict([(k, v) for k,v in list(fields.items())
                       if k in [ 'hour', 'minute', 'second', 'microsecond' ] ]))
      return datetime.datetime(datetime.MINYEAR, 1, 1).replace(**fields)

  def cast(self, value):
    if value in ('', None):
      return None
    return parser.parse(value, default=self.NoDefaultDate())


class VectorType(CellType):
  """ A vector field -- list of doubles. Used for Solr only """

  def __init__(self) -> None:
      super().__init__()
      self.dimension = None

  guessing_weight = 3
  result_type = list

  def cast(self, value):
    if value in ('', None):
      return None

    try:
      if value.startswith('[') and value.endswith(']'):
        elemlist = value[1:len(value)-1].split(',')
        result = []
        decimal_type = DecimalType()
        for e in elemlist:
          result.append(float(decimal_type.cast(e)))
        self.dimension = len(result)
        return result
    except:
      raise ValueError()

    raise ValueError()

class TextType(StringType):
  """ String and Text types are same in CDV, however Text
  is saved as TextFieldType and String as StringFieldType
  into Solr. String values are guessed as String automatically,
  but Text can only be selected explicitly by user. Used for
  Solr only """

  guessing_weight = 0

def types_processor(types, error=None, strict=False):
  """ Apply the column types set on the instance to the
  current row, attempting to cast each cell to the specified
  type.

  Strict means that casting errors are not ignored"""

  def apply_types(row_set, row):
    if types is None:
      return row
    error[:] = []
    for i, (cell, type) in enumerate(zip(row, types), 1):
      try:
        cell.value = type.cast(cell.value)
      except Exception as e:
        if strict and type and error is not None:
          # should we remove if not raise in any situation
          error.append({"error": "%s: %s" % (e.__class__.__name__ ,str(e)),
                        "col_num":i})
        else:
          raise
      finally:
        cell.type = type
    return row
  return apply_types

TYPES = [StringType, DecimalType, IntegerType, BoolType, VectorType]

# from messytables types.py - modified type_instances
def type_guess(rows, types=TYPES, strict=False, dataconnection_type=None):
  """ The type guesser aggregates the number of successful
  conversions of each column to each type, weights them by a
  fixed type priority and select the most probable type for
  each column based on that figure. It returns a list of
  ``CellType``. Empty cells are ignored.

  Strict means that a type will not be guessed
  if parsing fails for a single cell in the column."""
  guesses = []
  if not supportsVectorDataType(dataconnection_type):
    for type_instance in types:
      if isinstance(type_instance, VectorType) or isinstance(type_instance, TextType):
        types.remove(type_instance)
  type_instances = [i if isclass(t) else t for t in types for i in t.instances()]
  if strict:
    at_least_one_value = []
    for ri, row in enumerate(rows):
      diff = len(row) - len(guesses)
      for _ in range(diff):
        typesdict = {}
        for type in type_instances:
          if isinstance(type, VectorType):
            type = VectorType()
          typesdict[type] = 0
        guesses.append(typesdict)
        at_least_one_value.append(False)
      for ci, cell in enumerate(row):
        if not cell.value:
          continue
        at_least_one_value[ci] = True
        for type in list(guesses[ci].keys()):
          if not type.test(cell.value):
            guesses[ci].pop(type)
    # no need to set guessing weights before this
    # because we only accept a type if it never fails
    for i, guess in enumerate(guesses):
      for type in guess:
        guesses[i][type] = type.guessing_weight
    # in case there were no values at all in the column,
    # we just set the guessed type to string
    for i, v in enumerate(at_least_one_value):
      if not v:
        guesses[i] = {StringType(): 0}
  else:
    for i, row in enumerate(rows):
      diff = len(row) - len(guesses)
      for _ in range(diff):
        guesses.append(defaultdict(int))
      for i, cell in enumerate(row):
        # add string guess so that we have at least one guess
        guesses[i][StringType()] = guesses[i].get(StringType(), 0)
        if not cell.value:
          continue
        for type in type_instances:
          if type.test(cell.value):
            guesses[i][type] += type.guessing_weight
    _columns = []
  _columns = []
  for guess in guesses:
    # this first creates an array of tuples because we want the types to be
    # sorted. Even though it is not specified, python chooses the first
    # element in case of a tie
    # See: http://stackoverflow.com/a/6783101/214950
    guesses_tuples = [(t, guess[t]) for t in guess]
    _columns.append(max(guesses_tuples, key=lambda t_n: t_n[1] + (1 - type_instances.index(t_n[0])/10))[0])
  return _columns


# overriding two functions of CSVRowSet class
class CSVRowSetLocal(CSVRowSet):
  def __init__(self, name, fileobj, delimiter=None, quotechar=None,
               encoding='utf-8', window=None, doublequote=None,
               lineterminator=None, skipinitialspace=None,
               #escapechar = "\\"):
                escapechar = None):
    self.escapechar = escapechar
    self.has_header = None
    self.__dialect = None
    super(CSVRowSetLocal, self).__init__(name,
                                         fileobj,
                                         delimiter=delimiter,
                                         quotechar=quotechar,
                                         encoding=encoding,
                                         window=window,
                                         doublequote=doublequote,
                                         lineterminator=lineterminator,
                                         skipinitialspace=skipinitialspace)

    try:
      delim = '\n'  # NATIVE
      sample = delim.join(self._sample)
      sniffer = csv.Sniffer()
      self.has_header = sniffer.has_header(smart_str(sample))
    except Exception as e:
      # on the exception we consider no header
      self.has_header = False

  @property
  def _dialect(self):
    delim = '\n'  # NATIVE
    sample = delim.join(self._sample)
    try:
      dialect = csv.Sniffer().sniff(sample,
                                    delimiters=['\t', ',', ';', '|'])  # NATIVE
      dialect.delimiter = messytables.commas.native_string(dialect.delimiter)
      dialect.quotechar = messytables.commas.native_string(dialect.quotechar)
      dialect.lineterminator = delim
      dialect.doublequote = True
      if self.delimiter is None:
        self.delimiter = dialect.delimiter
      if self.lineterminator is None:
        self.lineterminator = dialect.lineterminator
    except csv.Error:
      dialect = csv.excel

    return dialect

  @property
  def _overrides(self):
    d = {}
    if self.delimiter:
      d['delimiter'] = self.delimiter
    if self.quotechar:
      d['quotechar'] = self.quotechar
    if self.doublequote:
      d['doublequote'] = self.doublequote
    if self.lineterminator:
      d['lineterminator'] = self.lineterminator
    if self.escapechar:
      d['escapechar'] = self.escapechar
    if self.skipinitialspace is not None:
      d['skipinitialspace'] = self.skipinitialspace
    return d

# Messytables to Type Name mapping
_MT_TO_DISPLAY = {
  StringType: "String",
  IntegerType: "Integer",
  DecimalType: "Real",
  BoolType: "Boolean",
  DateUtilType:"Timestamp",
  VectorType: "Vector",
  TextType: "Text",
}
_DISPLAY_TO_MT = dict([ (v, k) for k, v in list(_MT_TO_DISPLAY.items())])
