# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import django.test
import json
import os
import datetime
import re

from datasets.models import DataConnection
from datasources.dataimport import DataImport, DecimalEncoder, DateTimeEncoder, MultipleJsonEncoders
from unittest import mock
from decimal import Decimal

def normalize_sql(sql):
  return re.sub(r"\s+", " ", sql).strip()

class TestDataimport(django.test.TestCase):
  def setUp(self):
    super().setUp()
    self.test_csv_input = 'test_input.csv'

    # Edge cases for testing postgres
    self.test_data_rows = [
      # 1. Regular string with single quotes
      ["John's data", 123.45, "2025-01-01 12:00:00", None, True],

      # 2. Empty string
      ["", 0.0, "2025-01-01 00:00:00", None, False],

      # 3. String with special characters (backslashes and single quotes)
      ["back\\slash", 100.5, "2025-01-01 12:34:56", None, True],

      # 4. Null value (None in Python, will be converted to SQL NULL)
      [None, None, None, None, None],

      # 5. Float with NaN (Not a Number)
      [None, float('nan'), "2025-01-01 12:00:00", None, None],

      # 6. Float with infinity
      [None, float('inf'), "2025-01-01 12:00:00", None, None],

      # 7. Valid integer, to check if it's treated properly
      [None, 12345, "2025-01-01 00:00:01", None, True],

      # 8. Timestamp formatted correctly (should match a real timestamp)
      ["Alice", 987.65, "2025-01-01 12:30:45", None, False],

      # 9. String with Unicode characters (for testing UTF-8 compatibility)
      ["José", 56.78, "2025-01-01 12:45:00", None, True],

      # 10. Boolean values (True and False)
      ["Bob", 99.99, "2025-01-01 00:00:00", None, True],
      ["Alice", 99.99, "2025-01-01 00:00:00", None, False],
    ]

  def tearDown(self):
    super().tearDown()
    if os.path.isfile(self.test_csv_input):
      os.remove(self.test_csv_input)

  def test_solr_import(self):
    dc1 = DataConnection.objects.create(
      dataconnection_name='test_solr_connection',
      dataconnection_type='solr',
      dataconnection_info=json.dumps({'PARAMS': {}}))

    # create csv input
    with open(self.test_csv_input, mode='wt') as csv_file:
      csv_file.writelines(['col1, col2, col3, col4, col5, col6\n'])
      csv_file.writelines(['2, "a", 1.2, 2024-06-27T19:25:35, "[-0.2, 0.45, 1.5]", true\n'])

    # import the CSV file:
    di = DataImport(uploadedfile=self.test_csv_input,
                    dataconnection_id=dc1.id,
                    data={
                        'datasource_tablename': 'test_table'
                    }
                    )
    di.reset_file()

    with mock.patch.object(di.dataconn, 'log_sync_runquery') as runquery_mock:
      di.create_table({})
      runquery_mock.assert_called_once()
      createCollectionArgs = runquery_mock.call_args_list[0].args
      createCollectionPayload = json.loads(createCollectionArgs[0][3][len('json:'):])
      fields = createCollectionPayload['body']['add-field']
      colnames = [field['name'] for field in fields]
      coltypes = [field['type'] for field in fields]
      self.assertEquals(colnames, ['col1', 'col2', 'col3', 'col4', 'col5', 'col6'])
      self.assertEquals(coltypes, ['pint', 'string', 'pfloat', 'pdate', 'knn_vector_3', 'boolean'])

      runquery_mock.reset_mock()
      di.insert_data(batch_count=100)
      runquery_mock.assert_called_once()
      dataUploadArgs = runquery_mock.call_args_list[0].args
      uploadPayload = json.loads(dataUploadArgs[0][0][len('json:'):])
      colvalues = [colvalue for colname, colvalue in uploadPayload['body'][0].items()]
      self.assertEquals(colvalues, [2, 'a', '1.2', '2024-06-27T19:25:35', [-0.2, 0.45, 1.5], True])

  def test_postgres_import(self):
    dc2 = DataConnection.objects.create(
        dataconnection_name='test_postgres_connection',
        dataconnection_type='postgres',
        dataconnection_info=json.dumps({'PARAMS': {'DATABASE':'test_db_postgres'}}))

    with open(self.test_csv_input, mode='wt') as csv_file:
      csv_file.writelines(['Column1, Column2, Column3, Column4, Column5\n'])
      for row in self.test_data_rows:
        # Convert None to empty string for CSV compatibility
        row = [str(item) if item is not None else '' for item in row]
        csv_file.writelines([','.join(row) + '\n'])

    di = DataImport(uploadedfile=self.test_csv_input,
                    dataconnection_id=dc2.id,
                    data={
                        'datasource_tablename': 'test_table',
                        'database': 'test_db_postgres'
                    })
    di.reset_file()

    with mock.patch.object(di.dataconn, 'sync_runquery') as runquery_mock:
      di.create_table({})
      runquery_mock.assert_called_once()
      runquery_mock.reset_mock()

      di.insert_data(batch_count=100)
      runquery_mock.assert_called_once()
      dataUploadQuery = runquery_mock.call_args_list[0].args[0]
      expectedDataUploadQuery = """INSERT INTO ""."test_table" values ('John''s data', 123.45, '2025-01-01 12:00:00', '', True),('', 0.0, '2025-01-01 00:00:00', '', False),('back\slash', 100.5, '2025-01-01 12:34:56', '', True),('', NULL, NULL, '', NULL),('', NaN, '2025-01-01 12:00:00', '', NULL),('', Infinity, '2025-01-01 12:00:00', '', NULL),('', 12345, '2025-01-01 00:00:01', '', True),('Alice', 987.65, '2025-01-01 12:30:45', '', False),('José', 56.78, '2025-01-01 12:45:00', '', True),('Bob', 99.99, '2025-01-01 00:00:00', '', True),('Alice', 99.99, '2025-01-01 00:00:00', '', False)"""
      
      self.maxDiff = None
      self.assertEqual(dataUploadQuery, expectedDataUploadQuery)

class TestDecimalEncoder(django.test.TestCase):
  def test_decimal_encoder(self):
    obj = {'decimal': Decimal('1.11')}
    with self.assertRaises(TypeError):
      json.dumps(obj)
    self.assertEquals(json.dumps(obj, cls=DecimalEncoder), '{"decimal": "1.11"}')

class TestDateTimeEncoder(django.test.TestCase):
  def test_datetime_encoder(self):
    obj = {'datetime': datetime.datetime(2020, 5, 17)}
    with self.assertRaises(TypeError):
      json.dumps(obj)
    self.assertEquals(json.dumps(obj, cls=DateTimeEncoder), '{"datetime": "2020-05-17T00:00:00"}')

    obj = {'date': datetime.date(2020, 5, 17)}
    with self.assertRaises(TypeError):
      json.dumps(obj)
    self.assertEquals(json.dumps(obj, cls=DateTimeEncoder), '{"date": "2020-05-17"}')

class TestMultipleJsonEncoders(django.test.TestCase):
  def test_multiple_json_encoders(self):
    obj = {'datetime': datetime.datetime(2020, 5, 17), 'decimal': Decimal('1.11')}

    with self.assertRaises(TypeError):
      json.dumps(obj, cls=DecimalEncoder)

    with self.assertRaises(TypeError):
      json.dumps(obj, cls=DateTimeEncoder)

    self.assertEquals(
      json.dumps(obj, cls=MultipleJsonEncoders(DateTimeEncoder, DecimalEncoder)),
      '{"datetime": "2020-05-17T00:00:00", "decimal": "1.11"}'
    )