# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.test import TestCase
from datasources.datasource_csv import DataSourceCSV
from io import BytesIO
from decimal import Decimal
from dateutil.parser import parse as datetime_parse
from dataconn.connections_types import *
from datasets.models import DataConnection
from pprint import pprint
from unittest import skip

TESTS_ENABLED=True

try:
  from messytables.commas import native_string
except Exception as e:
  # Disable tests if incorrect messytables version is found
  TESTS_ENABLED=False

# CSV parser tests
# Run with: python manage.py test datasources


# Helper functions

def _mt_rows_to_string(rows):
  """
  Return a string representation of rows
  """
  outputstr = ""
  for row in rows:
    rowstring = ",".join([str(c.value) for c in row])
    outputstr += rowstring + "\n"
  return outputstr

def _mt_rows_to_lists(rows):
  """
  Return a list of lists with each list being the values of a row.
  """
  outputlist = []
  for row in rows:
    outputlist.append([c.value for c in row])
  return outputlist


def _D(decimalstring):
  """
  Cast to decimal.Decimal
  """
  return Decimal(decimalstring)

def _DT(datestring):
  """
  Parse a string and cast to datetime
  """
  return datetime_parse(datestring)



# Tests

class CSVTest(TestCase):

  def test_basic_parsing(self):
    """
    Test basic parsing of:
    Integer, String, Real, Boolean, Timestamp
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1,col2,col3,col4,col5
123,ABC1,123.456,No,2016-01-01 12:00
223,ABC2,123.456,Yes,2015-06-02 13:00
323,ABC3,123.456,No,2014-03-19 01:00
423,ABC4,123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    OUTPUTSTR = \
"""
123,ABC1,123.456,False,2016-01-01 12:00:00
223,ABC2,123.456,True,2015-06-02 13:00:00
323,ABC3,123.456,False,2014-03-19 01:00:00
423,ABC4,123.456,True,2013-07-14 12:00:00
523,ABC5,123.456,False,2012-11-24 12:00:00
"""[1:]

    EXPECTED_HEADERS = ["col1","col2","col3","col4","col5"]
    EXPECTED_TYPES = ["Integer","String","Real","Boolean","Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename = "Test1", source_filehandle = csvfilehandle)
    ds.reset_file()

    outputstr = _mt_rows_to_string(ds.rows())

    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)
    self.assertEqual(outputstr, OUTPUTSTR)


  def test_basic_parsing_list(self):
    """
    Test basic parsing of:
    Integer, String, Real, Boolean, Timestamp
    Use list of values to compare
    """
    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1,col2,col3,col4,col5
123,ABC1,123.456,No,2016-01-01 12:00
223,ABC2,123.456,Yes,2015-06-02 13:00
323,ABC3,123.456,No,2014-03-19 01:00
423,ABC4,123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    OUTPUTLIST = \
      [
        [123, "ABC1", _D("123.456"), False, _DT("2016-01-01 12:00:00")],
        [223, "ABC2", _D("123.456"), True, _DT("2015-06-02 13:00:00")],
        [323, "ABC3", _D("123.456"), False, _DT("2014-03-19 01:00:00")],
        [423, "ABC4", _D("123.456"), True, _DT("2013-07-14 12:00:00")],
        [523, "ABC5", _D("123.456"), False, _DT("2012-11-24 12:00:00")],
      ]

    EXPECTED_HEADERS = ["col1", "col2", "col3", "col4", "col5"]
    EXPECTED_TYPES = ["Integer", "String", "Real", "Boolean", "Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename="Test1",
                       source_filehandle=csvfilehandle)
    ds.reset_file()

    outputlist = _mt_rows_to_lists(ds.rows())

    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)
    self.assertEqual(outputlist, OUTPUTLIST)

  def test_override_headers_and_types(self):
    """
    Test overriding headers and types.
    Override types with compatible types.
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1,col2,col3,col4,col5
123,ABC1,123.456,No,2016-01-01 12:00
223,ABC2,123.456,Yes,2015-06-02 13:00
323,ABC3,123.456,No,2014-03-19 01:00
423,ABC4,123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    TEST_HEADERS=["column_a","column_b","column_c","column_d","column_e"]
    TEST_TYPES = ["Real","String","String","String","String"]


    OUTPUTSTR = \
"""
123,ABC1,123.456,No,2016-01-01 12:00
223,ABC2,123.456,Yes,2015-06-02 13:00
323,ABC3,123.456,No,2014-03-19 01:00
423,ABC4,123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    EXPECTED_HEADERS=["column_a","column_b","column_c","column_d","column_e"]
    EXPECTED_TYPES = ["Real","String","String","String","String"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename = "Test1", headers=TEST_HEADERS,
                       types=TEST_TYPES, source_filehandle = csvfilehandle)
    ds.reset_file()

    outputstr = _mt_rows_to_string(ds.rows())

    self.assertEqual(outputstr, OUTPUTSTR)
    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)


  def test_single_col(self):
    """
    Test single column.  ARC-1563
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
123
223
323
423
523
"""[1:]

    OUTPUTSTR = \
"""
123
223
323
423
523
"""[1:]

    EXPECTED_HEADERS=["column_0"]
    EXPECTED_TYPES = ["Integer"]

    csvfilehandle = BytesIO(TESTCSV)
    from .dataimport import DataImport
    ds = DataImport(uploadedfile = "Test1")

    # Mocking data connection
    ds.dataconn = DataConnection()
    ds.dataconn.connection_type = DATACONN_ARCENGINE
    ds.datasource_csv.source_filehandle = csvfilehandle
    ds.reset_file()

    outputstr = _mt_rows_to_string(ds.datasource_csv.rows())

    self.assertEqual(ds.datasource_csv._guessed_has_header, False)
    self.assertEqual(outputstr, OUTPUTSTR)
    self.assertEqual(ds.datasource_csv.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.datasource_csv.types, EXPECTED_TYPES)


  def test_single_col_with_headers(self):
    """
    Test single column with header.  ARC-1563 ARC-1595
    Note: CSV header detection does not work in this case.
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
Numbers
123
223
323
423
523
"""[1:]

    OUTPUTSTR = \
"""
123
223
323
423
523
"""[1:]

    EXPECTED_HEADERS=["Numbers"]
    EXPECTED_TYPES = ["Integer"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename = "Test1", has_header=True, source_filehandle =csvfilehandle)
    ds.reset_file()

    outputstr = _mt_rows_to_string(ds.rows())

    self.assertEqual(ds._guessed_has_header, False)
    self.assertEqual(ds.has_header, True)
    self.assertEqual(outputstr, OUTPUTSTR)
    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)



  def test_doublequotes_with_commas(self):
    """
    Test double-quoted strings
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1,col2,col3,col4,col5
123,"ABC1,part2",123.456,No,2016-01-01 12:00
223,",ABC2",123.456,Yes,2015-06-02 13:00
323,"ABC3,",123.456,No,2014-03-19 01:00
423,ABC4,123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    OUTPUTSTR = \
"""
123,ABC1,part2,123.456,False,2016-01-01 12:00:00
223,,ABC2,123.456,True,2015-06-02 13:00:00
323,ABC3,,123.456,False,2014-03-19 01:00:00
423,ABC4,123.456,True,2013-07-14 12:00:00
523,ABC5,123.456,False,2012-11-24 12:00:00
"""[1:]

    EXPECTED_HEADERS = ["col1","col2","col3","col4","col5"]
    EXPECTED_TYPES = ["Integer","String","Real","Boolean","Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename = "Test1", source_filehandle = csvfilehandle)
    ds.reset_file()

    outputstr = _mt_rows_to_string(ds.rows())

    self.assertEqual(outputstr, OUTPUTSTR)
    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)

  def test_singlequotes_with_commas(self):
    """
    Test parsing single-quoted strings
    ARC-1677
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1,col2,col3,col4,col5
123,'ABC1,part2',123.456,No,2016-01-01 12:00
223,',ABC2',123.456,Yes,2015-06-02 13:00
323,'ABC3,',123.456,No,2014-03-19 01:00
423,ABC4,123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    OUTPUTSTR = \
"""
123,ABC1,part2,123.456,False,2016-01-01 12:00:00
223,,ABC2,123.456,True,2015-06-02 13:00:00
323,ABC3,,123.456,False,2014-03-19 01:00:00
423,ABC4,123.456,True,2013-07-14 12:00:00
523,ABC5,123.456,False,2012-11-24 12:00:00
"""[1:]

    EXPECTED_HEADERS = ["col1","col2","col3","col4","col5"]
    EXPECTED_TYPES = ["Integer","String","Real","Boolean","Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename = "Test1", source_filehandle = csvfilehandle)
    ds.reset_file()

    outputstr = _mt_rows_to_string(ds.rows())

    self.assertEqual(outputstr, OUTPUTSTR)
    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)

  def test_semicolon_delimiter_parsing(self):
    """
    Test semicolon delim
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1;col2;col3;col4;col5
123;ABC1;123.456;No;2016-01-01 12:00
223;ABC2;123.456;Yes;2015-06-02 13:00
323;ABC3;123.456;No;2014-03-19 01:00
423;ABC4;123.456;Yes;2013-07-14 12:00
523;ABC5;123.456;No;2012-11-24 12:00
"""[1:]

    OUTPUTSTR = \
"""
123,ABC1,123.456,False,2016-01-01 12:00:00
223,ABC2,123.456,True,2015-06-02 13:00:00
323,ABC3,123.456,False,2014-03-19 01:00:00
423,ABC4,123.456,True,2013-07-14 12:00:00
523,ABC5,123.456,False,2012-11-24 12:00:00
"""[1:]

    EXPECTED_HEADERS = ["col1","col2","col3","col4","col5"]
    EXPECTED_TYPES = ["Integer","String","Real","Boolean","Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename = "Test1", source_filehandle = csvfilehandle)
    ds.reset_file()

    outputstr = _mt_rows_to_string(ds.rows())

    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)
    self.assertEqual(outputstr, OUTPUTSTR)

  @skip("No swedish in UBI")
  def test_comma_decimal_semicolon_delimiter_parsing(self):
    """
    Test comma decimal, semicolon delim
    ARC-1910
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1;col2;col3;col4;col5
123;ABC1;123,456;No;2016-01-01 12:00
223;ABC2;123,456;Yes;2015-06-02 13:00
323;ABC3;123,456;No;2014-03-19 01:00
423;ABC4;123,456;Yes;2013-07-14 12:00
523;ABC5;123,456;No;2012-11-24 12:00
"""[1:]

    TEST_TYPES = ["Integer","String","Real","Boolean","Timestamp"]

    OUTPUTSTR = \
"""
123,ABC1,123.456,False,2016-01-01 12:00:00
223,ABC2,123.456,True,2015-06-02 13:00:00
323,ABC3,123.456,False,2014-03-19 01:00:00
423,ABC4,123.456,True,2013-07-14 12:00:00
523,ABC5,123.456,False,2012-11-24 12:00:00
"""[1:]

    EXPECTED_HEADERS = ["col1","col2","col3","col4","col5"]
    EXPECTED_TYPES = ["Integer","String","Real","Boolean","Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename = "Test1",
                       localesetting='sv_SE',
                       source_filehandle = csvfilehandle)
    ds.reset_file()

    outputstr = _mt_rows_to_string(ds.rows())

    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)
    self.assertEqual(outputstr, OUTPUTSTR)


  def test_newline_string_list(self):
    """
    Test multiple line strings
    ARC-1535
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1,col2,col3,col4,col5
123,"ABC1
part2",123.456,No,2016-01-01 12:00
223,"
ABC2",123.456,Yes,2015-06-02 13:00
323,"ABC3
",123.456,No,2014-03-19 01:00
423,"ABC4

part2",123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    OUTPUTLIST = \
      [
        [123, "ABC1\npart2", _D("123.456"), False, _DT("2016-01-01 12:00:00")],
        [223, "\nABC2", _D("123.456"), True, _DT("2015-06-02 13:00:00")],
        [323, "ABC3\n", _D("123.456"), False, _DT("2014-03-19 01:00:00")],
        [423, "ABC4\n\npart2", _D("123.456"), True, _DT("2013-07-14 12:00:00")],
        [523, "ABC5", _D("123.456"), False, _DT("2012-11-24 12:00:00")],
      ]

    EXPECTED_HEADERS = ["col1", "col2", "col3", "col4", "col5"]
    EXPECTED_TYPES = ["Integer", "String", "Real", "Boolean", "Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename="Test1",
                       source_filehandle=csvfilehandle)
    ds.reset_file()

    outputlist = _mt_rows_to_lists(ds.rows())

    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)
    self.assertEqual(outputlist, OUTPUTLIST)


  def test_singlequote_list(self):
    """
    Test singlequote within a doublequote string
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1,col2,col3,col4,col5
123,"ABC1'part2",123.456,No,2016-01-01 12:00
223,ABC2,123.456,Yes,2015-06-02 13:00
323,ABC3,123.456,No,2014-03-19 01:00
423,ABC4,123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    OUTPUTLIST = \
      [
        [123, "ABC1'part2", _D("123.456"), False, _DT("2016-01-01 12:00:00")],
        [223, "ABC2", _D("123.456"), True, _DT("2015-06-02 13:00:00")],
        [323, "ABC3", _D("123.456"), False, _DT("2014-03-19 01:00:00")],
        [423, "ABC4", _D("123.456"), True, _DT("2013-07-14 12:00:00")],
        [523, "ABC5", _D("123.456"), False, _DT("2012-11-24 12:00:00")],
      ]

    EXPECTED_HEADERS = ["col1", "col2", "col3", "col4", "col5"]
    EXPECTED_TYPES = ["Integer", "String", "Real", "Boolean", "Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename="Test1",
                       source_filehandle=csvfilehandle)
    ds.reset_file()

    outputlist = _mt_rows_to_lists(ds.rows())

    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)
    self.assertEqual(outputlist, OUTPUTLIST)

  def test_escapechar_comma_list(self):
    """
    Test escapechar before comma
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1,col2,col3,col4,col5
123,ABC1\\,part2,123.456,No,2016-01-01 12:00
223,ABC2,123.456,Yes,2015-06-02 13:00
323,ABC3,123.456,No,2014-03-19 01:00
423,ABC4,123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    OUTPUTLIST = \
      [
        [123, "ABC1,part2", _D("123.456"), False, _DT("2016-01-01 12:00:00")],
        [223, "ABC2", _D("123.456"), True, _DT("2015-06-02 13:00:00")],
        [323, "ABC3", _D("123.456"), False, _DT("2014-03-19 01:00:00")],
        [423, "ABC4", _D("123.456"), True, _DT("2013-07-14 12:00:00")],
        [523, "ABC5", _D("123.456"), False, _DT("2012-11-24 12:00:00")],
      ]

    EXPECTED_HEADERS = ["col1", "col2", "col3", "col4", "col5"]
    EXPECTED_TYPES = ["Integer", "String", "Real", "Boolean", "Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename="Test1",
                       source_filehandle=csvfilehandle,
                       has_header=True,
                       escapechar="\\",
                       headers=EXPECTED_HEADERS)
    ds.reset_file()

    outputlist = _mt_rows_to_lists(ds.rows())

    # pprint(outputlist)

    self.assertTrue(ds.has_header)
    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)
    self.assertEqual(outputlist, OUTPUTLIST)

  def test_escapechar_quote_list(self):
    """
    Test escapechar before comma
    """

    if not TESTS_ENABLED:
      return

    TESTCSV = \
b"""
col1,col2,col3,col4,col5
123,"ABC1\\"part2",123.456,No,2016-01-01 12:00
223,ABC2,123.456,Yes,2015-06-02 13:00
323,ABC3,123.456,No,2014-03-19 01:00
423,ABC4,123.456,Yes,2013-07-14 12:00
523,ABC5,123.456,No,2012-11-24 12:00
"""[1:]

    OUTPUTLIST = \
      [
        [123, 'ABC1"part2', _D("123.456"), False, _DT("2016-01-01 12:00:00")],
        [223, "ABC2", _D("123.456"), True, _DT("2015-06-02 13:00:00")],
        [323, "ABC3", _D("123.456"), False, _DT("2014-03-19 01:00:00")],
        [423, "ABC4", _D("123.456"), True, _DT("2013-07-14 12:00:00")],
        [523, "ABC5", _D("123.456"), False, _DT("2012-11-24 12:00:00")],
      ]

    EXPECTED_HEADERS = ["col1", "col2", "col3", "col4", "col5"]
    EXPECTED_TYPES = ["Integer", "String", "Real", "Boolean", "Timestamp"]

    csvfilehandle = BytesIO(TESTCSV)
    ds = DataSourceCSV(source_filename="Test1",
                       escapechar="\\",
                       source_filehandle=csvfilehandle)
    ds.reset_file()

    outputlist = _mt_rows_to_lists(ds.rows())

    # pprint(TESTCSV)
    # pprint(outputlist)

    self.assertTrue(ds.has_header)
    self.assertEqual(ds.headers, EXPECTED_HEADERS)
    self.assertEqual(ds.types, EXPECTED_TYPES)
    self.assertEqual(outputlist, OUTPUTLIST)


