# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.http import HttpResponse, QueryDict
from django.utils.html import escape
from django.template import loader
from django.contrib.auth.decorators import login_required
from django.conf import settings
from dataconn.dataconn import get_dataconn
from .dataimport import DataImport
from apps.utils import _appflags
from .imports.url_request import UrlRequest
from .imports import url_csv, url_opendata
import traceback

import base64
import datetime
import json
import locale
import logging
import os
import re
import time

# create logger
LOGGER = logging.getLogger("arcweb")

def handle_exception_and_response(e, api_message, log_message, status_code):
  stack_trace = escape(traceback.format_exc())
  error_message = escape(f'{log_message}: {str(e)}\n{stack_trace}')

  LOGGER.exception(escape(f'{api_message} {log_message}: {str(e)}'))

  return HttpResponse(f'{error_message} {str(e)}', status=status_code)


def _datasource_save_uploaded_file(in_file):
  """
  Save uploaded file to known location in filesystem

  in_file supports:
    .name: name of the file
    .chunks(): return an iterable of chunks, where each chunk is a byte string
  """

  filename = re.sub(r'\W', '_', in_file.name.split('.')[0].lower())
  todaydir = '%s' % datetime.date.today()
  media_root = getattr(settings, 'MEDIA_ROOT', '')
  upload_directory = os.path.join(media_root, "uploaded_files", todaydir)
  if not os.path.exists(upload_directory):
    os.makedirs(upload_directory)

  # Create unique name for storing file by appending timestamp
  uniq_filename = os.path.join(upload_directory,
                               "%s_%d" % (filename, time.time()))
  with open(uniq_filename, 'wb+') as destination:
    for chunk in in_file.chunks():
      destination.write(chunk)

  return uniq_filename

@login_required
def dataimport_html(request, datasource_id=None):
  """ Respond with html template for data import """
  return dataimport(request, datasource_id, 'html')

@login_required
def dataimport_json(request, datasource_id=None):
  """ Respond with html template for data import """
  return dataimport(request, datasource_id, 'json')

@login_required
def dataimport(request, datasource_id=None, format_type='json'):
  """
  REST endpoint for datasources - imported data
  """
  try:
    loc = locale.getlocale(locale.LC_NUMERIC)
    if request.method == 'POST':
      return _datasource_create(request)
    if request.method in ('PUT'):
      data = QueryDict(request.body).dict()
      if data['datasource_import_confirmed'] == "true":
        return _datasource_insert(request, datasource_id, format_type)
      else:
        return _datasource_update(request, datasource_id, format_type)

    error_message = 'Invalid request method'
    return handle_exception_and_response(
      Exception({error_message}),
      api_message='DataImport API: ',
      log_message={error_message},
      status_code=500
    )
  except Exception as e:
    return handle_exception_and_response(
      e,
      api_message='DataImport API: ',
      log_message='Internal server error',
      status_code=500
    )
  finally:
    locale.setlocale(locale.LC_NUMERIC, loc)

def _datasource_insert(request, datasource_id, format_type='json'):
  if datasource_id is None:
    error_message = 'No datasource id specified'
    return handle_exception_and_response(
        Exception({error_message}),
        api_message='DataImport API - Insert: ',
        log_message={error_message},
        status_code=500
    )

  try:
    data = QueryDict(request.body).dict()
    data['datasource_headers'] = json.loads(data.get('datasource_headers'))
    data = dict(list(request.session['arc_import_data'].items()) + list(data.items()))
  except Exception as e:
    return handle_exception_and_response(
      e,
      api_message='DataImport API - Insert',
      log_message='No valid data about import available for this session',
      status_code=500
    )

  ds = DataImport(username=request.user.username, data=data)
  ds.reset_file(preserve_errors=True)

  dataconn = ds.dataconnection()
  if ds.total_rows() > dataconn.upload_maxrows():
    raise Exception(f"Too many rows {ds.total_rows()} being inserted. " +
        f"This data connection is limited to uploading {dataconn.upload_maxrows()} rows.")

  ds.create_table(data=data)
  ds.insert_data(batch_count=settings.IMPORT_BATCH_INSERT)
  ds.compute_state_if_possible()
  ds.remove_uploaded_file()
  return HttpResponse([])

def _datasource_update(request, datasource_id, format_type='json'):
  """ Update an existing datasource """
  LOGGER.info(f'Updating data for datasource ID: {datasource_id}')
  if datasource_id is None:
    error_message = 'No datasource id specified'
    return handle_exception_and_response(
      Exception({error_message}),
      api_message='DataImport API - Update: ',
      log_message={error_message},
      status_code=500
    )

  try:
    data = QueryDict(request.body).dict()
    data['datasource_headers'] = json.loads(data.get('datasource_headers'))
    data = dict(list(request.session['arc_import_data'].items()) + list(data.items()))
  except Exception as e:
    return handle_exception_and_response(
      e,
      api_message='DataImport API - Update',
      log_message='No valid data about import available for this session',
      status_code=500
    )

  ds = DataImport(data=data)
  recalculate, redo_headers = ds.check_for_actions(
                                            request.session['arc_import_data'])
  if redo_headers:
    # forcing to recalculate header and types
    ds.datasource_csv.headers = None
    ds.datasource_csv.types = None
  ds.reset_file()
  data = ds.get_dict(recalculate)
  request.session['arc_import_data'] = data
  LOGGER.debug('DataImport API - Update: Updated data successfully')
  return HttpResponse(json.dumps(data), content_type = "application/json")

def _datasource_create(request):
  """Create an imported datasource """
  LOGGER.debug('DataImport API - Create: Creating a new imported datasource')
  dataType = request.POST.get('data-type')
  if dataType not in ['csv', 'url']:
    error_message = f'Invalid data type specified: {escape(dataType)}'
    return handle_exception_and_response(
      Exception({error_message}),
      api_message='DataImport API - Create: ',
      log_message={error_message},
      status_code=500
    )

  dataconnection_id = request.POST.get('datasource_active_dataconnid', None)
  if dataconnection_id is None:
    error_message = 'No active data connection in request'
    return handle_exception_and_response(
      Exception({error_message}),
      api_message='DataImport API - Create: ',
      log_message={error_message},
      status_code=500
    )
  else:
    dataconnection_id = int(dataconnection_id)
    dataconn = get_dataconn(dataconnection_id)
    if not dataconn.properties().supports_dataextract_target:
      error_message = 'DataImport API - Create: This dataconnection type does not support import ' \
                          'data yet.'
      return handle_exception_and_response(
        Exception({error_message}),
        api_message='DataImport API - Create: ',
        log_message={error_message},
        status_code=500
      )

  if dataType == 'url':
    from reports.views_settings import setting_get_single
    urlsEnabled = setting_get_single('site_enable_download_from_url') or 'true'
    if urlsEnabled != 'true':
      error_message = f'DataImport API - Create: {error_message}'
      return handle_exception_and_response(
        Exception({error_message}),
        api_message='DataImport API - Create: ',
        log_message={error_message},
        status_code=500
      )

  if dataType == 'csv':
    try:
      # File contents
      source_filecontents = request.FILES['datasource_file']

      # Name of original file as was on the user's system
      source_filename = source_filecontents.name
    except Exception as e:
      error_message = 'No file specified for importing'
      return handle_exception_and_response(
        e,
        api_message='DataImport API - Create: ',
        log_message={error_message},
        status_code=500
      )
  elif dataType == 'url':
    try:
      url = request.POST.get('url', '')
      if not url:
        error_message = 'Invalid URL specified'
        return handle_exception_and_response(
          Exception({error_message}),
          api_message='DataImport API - Create: ',
          log_message={error_message},
          status_code=500
        )

      source_filename = url
      url_request = UrlRequest(url)

      possible_handlers = [
          url_csv.CsvUrlToData,
          url_opendata.OpenDataUrlToData]

      for h in possible_handlers:
        d = h(url_request)
        if d.handled():
          source_filecontents = d
          break

      if not source_filecontents:
        raise Exception('Cannot handle data at URL')

    except Exception as e:
      error_message = 'Could not fetch file from URL'
      return handle_exception_and_response(
        e,
        api_message='DataImport API - Create: ',
        log_message={error_message},
        status_code=500
      )

  # check if there is a previous session; if yes, delete the file and info
  imported_data = request.session.get('arc_import_data', None)
  if imported_data:
    ds = DataImport(uploadedfile=imported_data['datasource_uploaded_file'])
    ds.remove_uploaded_file()
    del request.session['arc_import_data']

  try:
    uploadedfile = _datasource_save_uploaded_file(source_filecontents)
  except Exception as e:
    error_message = f'Uploaded file {source_filename} could not be saved: {str(e)}'
    return handle_exception_and_response(
      e,
      api_message='DataImport API - Create: ',
      log_message={error_message},
      status_code=500
    )
  ds = DataImport(dataconnection_id=dataconnection_id,
                  uploadedfile=uploadedfile,
                  sourcefile=source_filename)
  ds.reset_file()
  data = ds.get_dict()
  template = loader.get_template('data_import_edit.html')
  c = {'datasources': json.dumps(data)}
  c.update(_appflags(request))

  del data['datasource_sample_rows']
  request.session['arc_import_data'] = data
  return HttpResponse(template.render(c, request))
