# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import itertools
import uuid

from django.contrib.auth.models import User
from django.conf import settings

from dataconn.dataconn import get_dataconn, sanitize_headers

from extracts import EXTRACT_TYPE_FULL
import extracts.misc
import extracts.rewriter

class ExtractTargetFull(object):
  def __init__(self, tgtDataconnectionId, tgtDbname, tgtTablename, config, user):
    self._stats = {}
    self._user = user
    source = config['source']
    target = config['target']
    self._dataconnectionId = tgtDataconnectionId
    self._dataconnection = extracts.misc.fetchDataConn(tgtDataconnectionId)
    dbname = tgtDbname
    tablename = tgtTablename
    self._fqtablename = self._dataconnection.quote_tablename(f"{dbname}.{tablename}")
    self._tempFqtablename = self._dataconnection.quote_tablename(f"{dbname}.{tablename}tmp{uuid.uuid4().hex}")
    self._dimData = source.get('dimData')
    self._aggData = source.get('aggData')

  def stats(self):
    return self._stats

  def dataconnection(self):
    return self._dataconnection

  def dataconnectionId(self):
    return self._dataconnectionId

  def runQuery(self, query, rows=None):
    if rows:
      return self._dataconnection.executemany(self._user.username, query, rows)

    return self._dataconnection.log_sync_runquery(query, username=self._user.username)

  def validate(self):
    (dbname, tablename) = self._dataconnection.tablename_parts(self._fqtablename)
    extracts.misc.validateTargetTableName(dbname, tablename)
    extracts.misc.validateTargetDataConnection(self._dataconnection)

  def createCommands(self, createTempTable = False):
    properties = self._dataconnection.properties()

    tgtColnames = sanitize_headers([x['alias'] \
        for x in itertools.chain(self._dimData, self._aggData)], properties)
    tgtColtypes = [x['type'] for x in itertools.chain(self._dimData, self._aggData)]

    importuser = None
    importsrcfile = None
    external_table = False
    location = None
    tablename = self._tempFqtablename if createTempTable else self._fqtablename
    createCmd = self._dataconnection.create_sql_cmds(tablename, tgtColnames, tgtColtypes,
        importuser, importsrcfile, external_table, location)

    return [createCmd]

  def rowBatch(self, extractSource, srcQuery):
    srcConn = extractSource.dataconnection()
    srcCursor = srcConn.cursor_get(username=self._user.username)
    srcCursor.execute(srcQuery)
    rowBatchSize = settings.IMPORT_BATCH_INSERT

    while True:
      if srcConn.connection_type == 'duckdb':
        try:
          rows = srcCursor.fetchmany(size=rowBatchSize)
          if rows:
            yield rows
          else:
            return

        except:
          return
      else:
        rows = srcCursor.fetchmany(size=rowBatchSize)
        if rows:
          yield rows
        else:
          return

  def refreshCommands(self, extractSource, testSelectCommand=True):
    (srcQuery, colnames) = extracts.misc.datasetToSQL(extractSource.dataset(), user=self._user,
        dims=[x['expr'] for x in self._dimData],
        aggs=[x['expr'] for x in self._aggData])

    if extractSource.dataconnectionId() == self.dataconnectionId():
      # First, verify that the select query can still be run on the source table
      if testSelectCommand:
        testQuery = extracts.misc.addLimit(srcQuery, limit=None,
            connProperties=extractSource.dataconnection().properties())
        yield [testQuery, None]

      # create temp table identical to target table name
      createTempTableCmd = self.createCommands(True)[0]
      yield [createTempTableCmd, None]

      # insert into temp table
      yield [f"{self._dataconnection.insert_table_prefix(self._tempFqtablename)} {srcQuery}", None]

      # drop target table
      yield [self._dataconnection.drop_table_no_check_cmd(self._fqtablename), None]

      # rename temp table to target table name
      yield [self._dataconnection.rename_table_cmd(self._tempFqtablename, self._fqtablename), None]

      return

    # Different data connections - use an insert-values statement
    colCount = len(colnames)
    paramstyle = self._dataconnection.properties().paramstyle
    insert_table_name = self._tempFqtablename

    valuesPlaceholder = f'({",".join([f"{paramstyle}"] * colCount)})'
    insertCmd = f'{self._dataconnection.insert_table_prefix(insert_table_name)} values {valuesPlaceholder}'

    tgtConnMaxRows = self._dataconnection.upload_maxrows()
    insertCount = 0
    self._stats["insertCount"] = insertCount

    firstTime = True
    for rows in self.rowBatch(extractSource, srcQuery):
      if firstTime:
        firstTime = False
        # Note that we create temp table identical to target table only after the first
        # batch of information has come out of the source side. In this way, if there is an issue with the source
        # connection, we do not delete data from the target before finding this out
        createTempTableCmd = self.createCommands(True)[0]
        yield [createTempTableCmd, None]

      insertCount += len(rows)
      if insertCount > tgtConnMaxRows:
        raise Exception(f"Too many rows {insertCount} being inserted. " +
            f"This data connection is limited to uploading {tgtConnMaxRows} rows.")

      self._stats["insertCount"] = insertCount

      insert_cmd, rows = extracts.misc.combineInserts(
        self._dataconnection.connection_type, insertCmd, valuesPlaceholder, rows)
      yield [insert_cmd, rows]

    if firstTime:
      firstTime = False
      # Handle the case where the row count on the source side is 0
      yield [self._dataconnection.truncate_table_cmd(self._fqtablename), None]
    else:
      # successfully inserted data into temp table
      # drop target table
      yield [self._dataconnection.drop_table_no_check_cmd(self._fqtablename), None]

      # rename temp table to target table name
      yield [self._dataconnection.rename_table_cmd(self._tempFqtablename, self._fqtablename), None]

class ExtractSourceFull(object):
  def __init__(self, srcDatasetId, config, user):
    self._stats = {}
    self._user = user
    datasetId = srcDatasetId
    self._dataset = extracts.misc.fetchDataset(datasetId)
    self._dataconnection = extracts.misc.fetchDataConn(self._dataset.dataconnection.id)

  def stats(self):
    return self._stats

  def dataconnection(self):
    return self._dataconnection

  def dataconnectionId(self):
    return self._dataset.dataconnection.id

  def validate(self):
    extracts.misc.validateSourceDataConnection(self._dataconnection)

  def dataset(self):
    return self._dataset

class ExtractManagerFull(object):
  def __init__(self, id, srcDatasetId, tgtDataconnectionId, tgtDbname, tgtTablename,
      config, runstate):
    self._stats = {}
    self._id = id
    self._srcDatasetId = srcDatasetId
    self._tgtDataconnectionId = tgtDataconnectionId
    self._tgtDbname = tgtDbname
    self._tgtTablename = tgtTablename
    self._config = config
    self._runstate = runstate
    self._user = User.objects.get(username=self._config['username'])
    self._extractSource = ExtractSourceFull(srcDatasetId, config, self._user)
    self._extractTarget = ExtractTargetFull(tgtDataconnectionId, tgtDbname, tgtTablename,
        config, self._user)

  def stats(self):
    s = {}
    s.update(self._stats)
    s.update(self._extractSource.stats())
    s.update(self._extractTarget.stats())
    return s

  def validate(self):
    extracts.misc.checkEnabled()
    self._extractSource.validate()
    self._extractTarget.validate()

  def id(self):
    return self._id

  def extract_type(self):
    return EXTRACT_TYPE_FULL

  def srcDatasetId(self):
    return self._srcDatasetId

  def tgtDataconnectionId(self):
    return self._tgtDataconnectionId

  def tgtDbname(self):
    return self._tgtDbname

  def tgtTablename(self):
    return self._tgtTablename

  def config(self):
    return self._config

  def runstate(self):
    return self._runstate

  def runCreate(self):
    createCmds = self._extractTarget.createCommands()
    for cmd in createCmds:
      self._extractTarget.runQuery(cmd)

  def runRefresh(self):
    cmds = self._extractTarget.refreshCommands(self._extractSource)
    for (cmd, rows) in cmds:
      self._extractTarget.runQuery(cmd, rows)
    extracts.rewriter.enableRewritesIfDisabledUntilRefreshed(self._id, self._config)

  def getCreateCmds(self):
    yield self._extractTarget.createCommands()

  def getRefreshCmds(self):
    if self._extractSource.dataconnectionId() != self._extractTarget.dataconnectionId():
      return

    cmds = self._extractTarget.refreshCommands(self._extractSource, testSelectCommand=False)
    for (cmd, rows) in cmds:
      yield cmd

