# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json

from django.contrib.auth.models import User
from django.http import HttpResponse
from django.conf import settings

from datasets.models import DataSet, DataConnection
from vizlib.query_builder import QueryBuilder
from dataconn.qcntl import QueryControl
from dataconn.arctypes import ARC_DATATYPE_TO_NOMINAL
from dataconn.connections_types import DATACONN_IMPYLA, DATACONN_HYVE
from vizlib.dataset_request import DatasetRequest
from vizlib.dataset import ModelDataset
from dataconn.dataconn import get_dataconn
from util.ff_flipper import FeatureFlipper

def enabled():
  return FeatureFlipper().get_flipper_value('site_enable_extracts')

def checkEnabled():
  if not enabled():
    raise Exception("Data Extracts not enabled")

def rewritesEnabled():
  f = FeatureFlipper()
  return f.get_flipper_value('site_enable_extracts') and \
         f.get_flipper_value('site_enable_extract_rewrites')

def HttpResponseError(err):
  return HttpResponse(err, status=500, content_type="text/plain")

def fetchUser(username):
  return User.objects.get(username=username)

def fetchDataset(datasetId):
  dsl = DataSet.objects.filter(id=datasetId)
  if len(dsl) != 1:
    raise Exception(f"Invalid dataset {datasetId}")

  return dsl[0]

def fetchDataConn(connectionId):
  return get_dataconn(connection_id=connectionId)

def fetchDataConnectionObject(connectionId):
  dcl = DataConnection.objects.filter(id=connectionId)
  if len(dcl) != 1:
    raise Exception(f"Invalid data connection {connectionId}")

  return dcl[0]

def datasetToSQL(srcDataset, user=None, srcConn=None, srcProperties=None,
    dims=None, aggs=None, filters=None):
  if not srcConn:
    srcConn = fetchDataConn(srcDataset.dataconnection.id)
    srcProperties = srcConn.properties()

  srcColnames = []

  qcntl = QueryControl()
  if not dims and not aggs:
    tables = json.loads(srcDataset.dataset_info or "[]")
    for table in tables:
      srcColnames.extend([(c.get('alias') or c.get('name')) for c in table["columns"] \
          if not c.get('hidden') and not c.get('exprhasagg')])

    dims = [f'[{x}]' for x in srcColnames]
  else:
    srcColnames.extend(dims)
    srcColnames.extend(aggs)

  if dims is None:
    dims = []

  if aggs is None:
    aggs = []

  if filters is None:
    filters = []

  dsreq = {
    'version': 1,
    'type': 'SQL',
    'dimensions': [
      {
        'type': 'SIMPLE',
        'expr': dim,
      } for dim in dims
    ],
    'aggregates': [
      {
        'type': 'SIMPLE',
        'expr': agg,
      } for agg in aggs
    ],
    'filters': filters,
    'dataset_id': srcDataset.id,
  }

  ds = ModelDataset(srcDataset.dataset_info, srcProperties)
  datasetRequest = DatasetRequest(dsreq, ds=ds)
  builder = QueryBuilder(dataset_request=datasetRequest,
                         connection=srcConn,
                         qcntl=qcntl)

  query = builder.buildQuery(user.username, user.id, None, None)
  return (query, srcColnames)

def addLimit(query, limit=None, connProperties=None):
  f = query.split()
  if len(f) >= 2:
    possible_limit = f[-2]
    if possible_limit.lower() == "limit":
      return query

  if limit is None:
    if not connProperties:
      raise Exception("No limit and no connection properties")

    if connProperties.supports_limit:
      if connProperties.supports_limit_0:
        limit = 0
      else:
        limit = 1
    else:
      raise Exception("Invalid dataconnection for adding limit")

  return f"{query} limit {limit}"

def getColumnTypes(conn=None, connProperties=None, query=None, username=None):
  query = addLimit(query, limit=None, connProperties=connProperties)
  result = conn.log_sync_runquery(query, username=username)

  if not result.get('coltypes'):
    raise Exception("Could not compute column types")

  return [ARC_DATATYPE_TO_NOMINAL.get(c, "STRING") for c in result["coltypes"]]

def validateSourceDataConnection(srcConn):
  srcProperties = srcConn.properties()
  if not srcProperties.supports_dataextract_source:
    raise Exception("The source data connection does not support extracts")

def validateTargetDataConnection(tgtConn):
  tgtProperties = tgtConn.properties()
  if not tgtProperties.supports_dataextract_target:
    raise Exception("The target data connection does not support extracts")

def validateTargetTableName(dbname, tablename):
  prefix = settings.DATA_EXTRACTS_TARGET_TABLENAME_PREFIX
  if prefix and not tablename.startswith(prefix):
    raise Exception(f"Target table name must start with {prefix}")

  if tablename == prefix:
    raise Exception(f"Invalid target table name {tablename}")

def combineInserts(connection_type, insertCmd, valuesPlaceholder, rows):
  """
  impyla executemany() is implemented as a sequence of individual executes
  This results in as many HDFS files being created as rows are inserted.

  This utility returns a version of the insert command that inserts multiple rows at once.
  """

  # Override insertCmd for hyve/impyla, because executemany differs for these conns
  if connection_type in [DATACONN_IMPYLA, DATACONN_HYVE]:
    for i in range(1, len(rows)):
      insertCmd = f'{insertCmd}, {valuesPlaceholder}'

    # flatten rows:
    if len(rows) > 1:
      flattened_rows = []
      for row in rows:
        flattened_rows += list(row)
      rows = [flattened_rows]

  return insertCmd, rows