# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import hashlib
import itertools

from cli.core import table

from django.contrib.auth.models import User
from django.conf import settings

import extracts.misc
import extracts.rewriter
from dataconn.dataconn import sanitize_headers
from datasets.models import Extract

from extracts import EXTRACT_TYPE_PARTITIONED

def fetchExtract(extractId):
  return Extract.objects.get(id=extractId)

def quoteValue(colvalue, coltype):
  coltype = coltype.upper()
  if coltype == "STRING" or coltype == "TIMESTAMP":
    return f'"{colvalue}"'

  return colvalue

class ExtractTargetPartitioned(object):
  def __init__(self, tgtDataconnectionId, tgtDbname, tgtTablename, config, runstate, user):
    self._stats = {}
    self._user = user
    self._runstate = runstate
    source = config['source']
    target = config['target']
    self._dataconnectionId = tgtDataconnectionId
    self._dataconnection = extracts.misc.fetchDataConn(tgtDataconnectionId)
    dbname = tgtDbname
    tablename = tgtTablename
    self._fqtablename = self._dataconnection.quote_tablename(f"{dbname}.{tablename}")

    properties = self._dataconnection.properties()
    if not properties.supports_partitions:
      raise Exception("Partitioned data extracts can only be run on target " +
          "connections that support partitions")

    self._dimData = source.get('dimData')
    self._aggData = source.get('aggData')
    self._partitionData = target.get('partitionData')

    allColData = self._dimData + self._aggData + self._partitionData
    allColAliases = [x['alias'] for x in allColData]
    tgtColnames = sanitize_headers(allColAliases, properties)
    self.aliasToTargetColumn = {alias: colName for alias, colName in zip(allColAliases, tgtColnames)}

    self._partitionColnames = [x['alias'] for x in self._partitionData]
    self._partitionSignatures = runstate.get('partitionSignatures')

    if self._partitionSignatures is None:
      self._partitionSignatures = {}
      runstate['partitionSignatures'] = self._partitionSignatures

  def stats(self):
    return self._stats

  def dataconnectionId(self):
    return self._dataconnectionId

  def runQuery(self, query, rows=None):
    if rows:
      return self._dataconnection.executemany(self._user.username, query, rows)

    return self._dataconnection.log_sync_runquery(query, username=self._user.username)

  def validate(self):
    if not self._dimData and not self._aggData:
      raise Exception("Invalid partitioned config. Dims or aggs must be specified")

    if not self._partitionData:
      raise Exception("Invalid partitioned config. Partitions must be specified")

    (dbname, tablename) = self._dataconnection.tablename_parts(self._fqtablename)
    extracts.misc.validateTargetTableName(dbname, tablename)
    extracts.misc.validateTargetDataConnection(self._dataconnection)

  def partitionData(self):
    return self._partitionData.copy()

  def partitionColnames(self):
    return self._partitionColnames.copy()

  def partitionSignatures(self):
    return self._partitionSignatures.copy()

  def resetPartitionSignatures(self):
    self._partitionSignatures = {}
    self._runstate['partitionSignatures'] = self._partitionSignatures

  def partitionSignaturesMatch(self, partition):
    if partition in self._partitionSignatures:
      return self._partitionSignatures[partition] == self._sourcePartitionSignatures[partition]

    return False

  def setSinglePartitionSignatureFromSource(self, partition):
    k = json.dumps(partition)
    if k in self._sourcePartitionSignatures:
      self._partitionSignatures[k] = self._sourcePartitionSignatures[k]
    else:
      if k in self._partitionSignatures:
        del self._partitionSignatures[k]

  def setSourcePartitionSignatures(self, sourcePartitionSignatures):
    self._sourcePartitionSignatures = sourcePartitionSignatures.copy()

  def createCommands(self):
    properties = self._dataconnection.properties()

    dimsAndAggs = self._dimData + self._aggData
    tgtColtypes = [x['type'] for x in dimsAndAggs]
    tgtColnames = [self.aliasToTargetColumn[c['alias']] for c in dimsAndAggs]

    tgtPartitionColtypes = [x['type'] for x in self._partitionData]
    tgtPartitionColnames = [self.aliasToTargetColumn[c['alias']] for c in self._partitionData]

    importuser = None
    importsrcfile = None
    external_table = False
    location = None
    # In CDP Impala, by default tables are transactional tables with the insert_only property, override it to
    # 'transactional'='false' to be able to drop partitions
    createCmd = self._dataconnection.create_sql_cmds(self._fqtablename, tgtColnames, tgtColtypes,
        importuser, importsrcfile, external_table, location,
        partitionColnames=tgtPartitionColnames,
        partitionColtypes=tgtPartitionColtypes,
        tblproperties=["'transactional'='false'"])

    return [createCmd]

  def rowBatch(self, extractSource, srcQuery):
    srcConn = extracts.misc.fetchDataConn(extractSource.dataconnectionId())
    srcCursor = srcConn.cursor_get(username=self._user.username)
    srcCursor.execute(srcQuery)
    rowBatchSize = settings.IMPORT_BATCH_INSERT

    while True:
      if srcConn.connection_type == 'duckdb':
        try:
          rows = srcCursor.fetchmany(size=rowBatchSize)
          if rows:
            yield rows
          else:
            return

        except:
          return
      else:
        rows = srcCursor.fetchmany(size=rowBatchSize)
        if rows:
          yield rows
        else:
          return

  def _dropPartitionCmd(self, tablename, partitionClause):
    return f"ALTER TABLE {tablename} DROP IF EXISTS PARTITION ({partitionClause})"

  def getRefreshCmdTemplate(self, extractSource):
    k = [f'<<{col["alias"]}>>' for col in self._partitionData]
    (srcQuery, colnames) = extracts.misc.datasetToSQL(extractSource.dataset(), user=self._user,
        dims=[x['expr'] for x in self._dimData],
        aggs=[x['expr'] for x in self._aggData],
        filters=[f"{col['alias']}={quoteValue(colvalue, col['type'])}" \
          for col, colvalue in zip(self._partitionData, k)])

    p = []
    for col, colvalue in zip(self._partitionData, k):
      p.append(f"{col['alias']}={quoteValue(colvalue, col['type'])}")
    partitionClause = f"{','.join(p)}"
    partitionPhrase = f"PARTITION ({partitionClause})"

    insertOverwrite = self._dataconnection.insert_table_prefix(self._fqtablename, overwrite=True)
    cmd = f"{insertOverwrite} {partitionPhrase} {srcQuery}"

    return [cmd]

  def refreshCommands(self, extractSource):
    # Iterate over a snapshot of partition signatures because they change
    # in the caller for this generator
    signaturesCopy = self._partitionSignatures.copy()
    for k in signaturesCopy:
      if not k in self._sourcePartitionSignatures:
        k = json.loads(k)
        p = []
        for col, colvalue in zip(self._partitionData, k):
          tgtColName = self.aliasToTargetColumn[col['alias']]
          p.append(f"{tgtColName}={quoteValue(colvalue, col['type'])}")
        partitionClause = f"{','.join(p)}"

        yield ['DROP', k, self._dropPartitionCmd(self._fqtablename, partitionClause), None]
        yield ['SIGN', k, None, None]

    for k in self._sourcePartitionSignatures:
      if self.partitionSignaturesMatch(k):
        continue

      k = json.loads(k)
      p = []
      for col, colvalue in zip(self._partitionData, k):
        tgtColName = self.aliasToTargetColumn[col['alias']]
        p.append(f"{tgtColName}={quoteValue(colvalue, col['type'])}")
      partitionClause = f"{','.join(p)}"

      (srcQuery, colnames) = extracts.misc.datasetToSQL(extractSource.dataset(), user=self._user,
        dims=[x['expr'] for x in self._dimData],
        aggs=[x['expr'] for x in self._aggData],
        filters=[f"[{col['alias']}]={quoteValue(colvalue, col['type'])}" \
          for col, colvalue in zip(self._partitionData, k)])

      partitionPhrase = f"PARTITION ({partitionClause})"

      if extractSource.dataconnectionId() == self.dataconnectionId():
        insertOverwrite = self._dataconnection.insert_table_prefix(self._fqtablename, overwrite=True)
        yield ['INSERT', k, f"{insertOverwrite} {partitionPhrase} {srcQuery}", None]
        yield ['SIGN', k, None, None]
        continue

      insertNoOverwrite = self._dataconnection.insert_table_prefix(self._fqtablename, overwrite=False)
      insertPrefix = f"{insertNoOverwrite} {partitionPhrase}"

      # Different data connections - use an insert-values statement
      colCount = len(colnames)
      paramstyle = self._dataconnection.properties().paramstyle

      valuesPlaceholder = f'({",".join([f"{paramstyle}"] * colCount)})'
      insertCmd = f'{insertPrefix} values {valuesPlaceholder}'

      tgtConnMaxRows = self._dataconnection.upload_maxrows()
      insertCount = 0
      self._stats["insertCount"] = insertCount

      firstTime = True
      for rows in self.rowBatch(extractSource, srcQuery):
        insertCount += len(rows)
        if insertCount > tgtConnMaxRows:
          raise Exception(f"Too many rows {insertCount} being inserted. " +
              f"This data connection is limited to uploading {tgtConnMaxRows} rows.")

        if firstTime:
          firstTime = False
          yield ['DROP', k, self._dropPartitionCmd(self._fqtablename, partitionClause), None]

        self._stats["insertCount"] = insertCount

        insert_cmd, rows = extracts.misc.combineInserts(
          self._dataconnection.connection_type, insertCmd, valuesPlaceholder, rows)
        yield ['INSERT', k, insert_cmd, rows]

      if firstTime:
        firstTime = False
        yield ['DROP', k, self._dropPartitionCmd(self._fqtablename, partitionClause), None]

      yield ['SIGN', k, None, None]

class ExtractSourcePartitioned(object):
  def __init__(self, datasetId, config, user):
    self._stats = {}
    self._user = user
    self._partitionColnames = []
    self._partitionSignatures = {}
    self._fqtablename = None

    self._datasetId = datasetId
    self._dataset = extracts.misc.fetchDataset(self._datasetId)
    self._dataconnection = extracts.misc.fetchDataConn(self._dataset.dataconnection.id)

    tables = json.loads(self._dataset.dataset_info)

    self.aliasToColumnName = self.computeAliasToColumnName(tables)
    table = tables[0]
    if table.get('select'):
      raise Exception("Dataset defined by select query can not be used as partitioned extract source")

    self._fqtablename = table['tablename']

    properties = self._dataconnection.properties()
    if not properties.supports_partitions:
      raise Exception("Partitioned data extracts can only be run on source " +
          "connections that support partitions")

  def computeAliasToColumnName(self, tables):
    result = {}
    for table in tables:
      for column in table['columns']:
        name = column.get('name')
        alias = column.get('alias')
        if name and alias:
          result[alias] = name
    return result

  def stats(self):
    return self._stats

  def dataconnectionId(self):
    return self._dataset.dataconnection.id

  def validate(self):
    extracts.misc.validateSourceDataConnection(self._dataconnection)

  def dataset(self):
    return self._dataset

  def tablename(self):
    return self._fqtablename

  def _convertHashToSignatures(self, h):
    h2 = {}

    for k in h:
      h[k].sort()

      if type(h[k][0]) == type([]):
        s = ','.join([','.join(x) for x in h[k]])
      else:
        s = ','.join(h[k])

      h2[k] = hashlib.sha256(s.encode('utf-8')).hexdigest()

    return h2

  def getIndices(self, srcList, tgtList):
    tgtList = [self.aliasToColumnName[alias] for alias in tgtList]
    indices = []
    for elem in tgtList:
      if not elem in srcList:
        raise Exception(f"Could not find {elem} in list {srcList}")

      indices.append(srcList.index(elem))

    return indices

  def repartitionExtractSource(self, colnames):
    indices = self.getIndices(self._partitionColnames, colnames)
    colPartitioning = self._repartition(self._partitionSignatures, indices)

    t = ExtractSourcePartitioned(self._datasetId, {}, self._user)

    t._partitionColnames = colnames
    t._partitionSignatures = self._convertHashToSignatures(colPartitioning)
    return t

  def _repartition(self, h, indices):
    h2 = {}
    for partition, v in h.items():
      newKey = self._getSubTuple(partition, indices)
      if newKey in h2:
        h2[newKey].append(v)
      else:
        h2[newKey] = [v]

    return h2

  def _getSubTuple(self, t, indices):
    t = json.loads(t)

    l = []
    for idx in indices:
      l.append(t[idx])

    return json.dumps(l)

  def partitionColnames(self):
    return self._partitionColnames.copy()

  def partitionSignatures(self):
    return self._partitionSignatures.copy()

  def coltype(self, colname):
    return self._colnameToType.get(colname)

  def initialize(self):
    data = self._dataconnection.table_columns(self._fqtablename, username=self._user.username)
    self._partitions = data.get('partitions')
    self._colnames = data.get('colnames')
    self._coltypes = data.get('coltypes')

    self._colnameToType = {}
    for colname, coltype in zip(self._colnames, self._coltypes):
      self._colnameToType[colname] = coltype

    result = self._dataconnection.log_sync_runquery(f"SHOW FILES IN {self._fqtablename}",
        username=self._user.username)
    rows = json.loads(result.get('rows'))
    colnames = result.get('colnames')

    try:
      index_path = colnames.index("Path")
      index_size = colnames.index("Size")
      index_partition = colnames.index("Partition")
    except ValueError as e:
      raise Exception(f"Expected column not found in output from SHOW FILES: {colnames}: {e}")


    firstTime = True
    partitionColnames = []

    partitionFiles = {}
    partitionSignatures = {}

    for row in rows:
      path = row[index_path]
      size = row[index_size]
      if len(row) == 2:
        continue

      partition = row[index_partition].strip()
      partitionVals = []

      if not partition:
        continue

      partitionSplits = partition.split("/")
      for ps in partitionSplits:
        lr = ps.split("=")
        partitionVals.append(lr[1])
        if firstTime:
          partitionColnames.append(lr[0])

      if firstTime:
        firstTime = False

      partition = json.dumps(partitionVals)
      if partition in partitionFiles:
        partitionFiles[partition].append([path, size])
      else:
        partitionFiles[partition] = [[path, size]]

    self._partitionColnames = partitionColnames
    self._partitionSignatures = self._convertHashToSignatures(partitionFiles)

class ExtractManagerPartitioned(object):
  def __init__(self, id, srcDatasetId, tgtDataconnectionId, tgtDbname, tgtTablename, config,
        runstate):
    self._stats = {}
    self._id = id
    self._srcDatasetId = srcDatasetId
    self._tgtDataconnectionId = tgtDataconnectionId
    self._tgtDbname = tgtDbname
    self._tgtTablename = tgtTablename
    self._config = config
    self._runstate = runstate
    self._user = User.objects.get(username=self._config['username'])
    self._extractSource = ExtractSourcePartitioned(self._srcDatasetId, self._config['source'],
        self._user)
    self._extractTarget = ExtractTargetPartitioned(tgtDataconnectionId, tgtDbname, tgtTablename,
        self._config, self._runstate, self._user)

  def stats(self):
    s = {}
    s.update(self._stats)
    s.update(self._extractSource.stats())
    s.update(self._extractTarget.stats())
    return s

  def validate(self):
    extracts.misc.checkEnabled()
    self._extractSource.validate()
    self._extractTarget.validate()

  def id(self):
    return self._id

  def extract_type(self):
    return EXTRACT_TYPE_PARTITIONED

  def srcDatasetId(self):
    return self._srcDatasetId

  def tgtDataconnectionId(self):
    return self._tgtDataconnectionId

  def tgtDbname(self):
    return self._tgtDbname

  def tgtTablename(self):
    return self._tgtTablename

  def config(self):
    return self._config

  def runstate(self):
    return self._runstate

  def runCreate(self):
    # We are going to drop and create the target table. Forget previous signatures
    self._extractTarget.resetPartitionSignatures()

    createCmds = self._extractTarget.createCommands()
    for cmd in createCmds:
      self._extractTarget.runQuery(cmd)

  def runRefresh(self):
    self._extractSource.initialize()
    es2 = self._extractSource.repartitionExtractSource(self._extractTarget.partitionColnames())
    self._extractTarget.setSourcePartitionSignatures(es2.partitionSignatures())

    if self._extractSource.dataconnectionId() == self._extractTarget.dataconnectionId():
      refreshCommands = self._extractTarget.refreshCommands(self._extractSource)
      for (type, partition, cmd, rows) in refreshCommands:
        if type == "SIGN":
          self._extractTarget.setSinglePartitionSignatureFromSource(partition)

          # Save away state
          e = fetchExtract(self._id)
          e.extract_runstate = json.dumps(self._runstate)
          e.save()

          continue

        self._extractTarget.runQuery(cmd, rows)

      extracts.rewriter.enableRewritesIfDisabledUntilRefreshed(self._id, self._config)
      return

    raise Exception("Partitioned extracts across connections is not yet supported")

  def getCreateCmds(self):
    yield self._extractTarget.createCommands()

  def getRefreshCmds(self):
    if self._extractSource.dataconnectionId() != self._extractTarget.dataconnectionId():
      return []

    for cmd in self._extractTarget.getRefreshCmdTemplate(self._extractSource):
      yield cmd

