# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import copy
import itertools
import re
import threading
import logging
import typing

import extracts.misc
from sqlrun.jsonselect_cache import jsonselect_get_segment
from vizlib.expression import Expression, unbracketColumn
from vizlib.filter_association import UserFilter
from dataconn.dataconn import sanitize_headers
from datasets.models import Extract
from dataconn.connections_types import isImpalaBased

from dataconn.dataconn import get_dataconn
from django.contrib.auth.models import User

from vizlib.dataset import ModelDataset

#
# TODO:
# - Stacked data extracts
#

LOGGER = logging.getLogger('arcweb')

RewriteStatusType = typing.Literal['enabled', 'disabled', 'disabled_until_refreshed']

#
# Check if we need to rewrite the query to use a data extract
# If true, a few parameters may change - the data connection the dsreq gets directed to
#   and the dsreq. Note that the dataset remains the original dataset since the
#   rewrite happens within ModelDataset by changing the dataset definition to point
#   at the table associated with the Extract (this is necessary because we may not have
#   a dataset at all for an Extract and it is very cumbersome to create one)
#
def handleRewrites(dataconnection_id, dsreq, user: User, ds: object, django_ds, conn):
  # LOGGER.info('handleRewrites %s', dataconnection_id)
  if not extracts.misc.rewritesEnabled():
    return (dataconnection_id, dsreq)

  if not dsreq:
    return (dataconnection_id, dsreq)

  # Check if you need to rewrite the query to use a data extract
  rewriteController = _getExtractRewriteController()
  if not rewriteController:
    return (dataconnection_id, dsreq)

  rewriteController.setupRewriteRules(dsreq, user=user, ds=ds, django_ds=django_ds, conn=conn)
  dsreqRewriter = dsreq.get('dsreqRewriter')
  if not dsreqRewriter:
    return (dataconnection_id, dsreq)

  # We are rewriting the query to use a data extract
  # Adjust the data connection to use if necessary
  extractConnectionId = dsreqRewriter.getExtractDataConnectionId()
  if extractConnectionId == dataconnection_id:
    return (dataconnection_id, dsreq)

  dataconnection_id = extractConnectionId
  return (dataconnection_id, dsreq)

# Helper function used below, and in tests to create mocks
def fetchExtracts(datasetIds):
  el = Extract.objects.values('id', 'srcDataset__id', 'tgtDataconnection__id',
      'tgtDbname', 'tgtTablename', 'extract_config',
      'srcDataset__dataconnection__dataconnection_type',
      'tgtDataconnection__dataconnection_type')

  if datasetIds:
    el = el.filter(srcDataset__id__in=datasetIds)

  return el

def _getRewriteStatusProperty(extractConfig):
  return extractConfig.get('rewrite_properties', {}).get('rewrite_status', 'disabled_until_refreshed')

def _updateRewriteStatusPropertyConfig(extractConfig, rewrite_status: RewriteStatusType):
  properties = extractConfig.get('rewrite_properties', {})
  properties['rewrite_status'] = rewrite_status
  extractConfig['rewrite_properties'] = properties

def updateRewriteStatusProperty(extractId, rewrite_status: RewriteStatusType):
  try:
    extract = Extract.objects.get(id=extractId)
  except:
    return False

  extractConfig = json.loads(extract.extract_config)
  if _getRewriteStatusProperty(extractConfig) == rewrite_status:
    return False

  _updateRewriteStatusPropertyConfig(extractConfig, rewrite_status)

  extract.extract_config = json.dumps(extractConfig)
  extract.save()

  resetExtractRewriteController()

  if rewrite_status == 'disabled_until_refreshed':
    LOGGER.info(f'Rewrites disabled until refreshed for extract {extractId}')
  else:
    LOGGER.info(f'Rewrites {rewrite_status} for extract {extractId}')

  return True

def enableRewritesIfDisabledUntilRefreshed(extractId, extractConfig):
  if _getRewriteStatusProperty(extractConfig) == 'disabled_until_refreshed':
      updateRewriteStatusProperty(extractId, 'enabled')

def _dimTracker(colname):
  return {
    'type': 'dim',

    # Strip out leading (s and trailing )s
    'dim': unbracketColumn(colname, pre='(', post=')')
  }

def _aggTracker(aggType, aggField):
  return {
    'type': 'agg',
    'aggType': aggType,
    'aggField': unbracketColumn(aggField, pre='(', post=')'),
    'exprStr': Expression.buildSimpleFnStr(aggType, aggField)
  }

class _MultipleSumProcessor:
  def __init__(self):
    self._sumProcessor = _SumProcessor()

  def processExpression(self, rewriteTarget, expr):
    # Break up the expression with one or more sums in it
    exprParts = expr.sumsOnly()
    if not exprParts:
      return expr

    # Rewrite each sum() expression individually
    newExprParts = []
    for exprPart in exprParts:
      if type(exprPart) == Expression:
        exprPart.setBaseExpr(rewriteTarget.findRewriteCol(exprPart.baseExpr()))
        newExprParts.append(exprPart.baseExpr())
      else:
        newExprParts.append(exprPart)

    # Combine the parts into a whole
    expr.setBaseExpr(''.join(newExprParts))
    return expr

  def processRewriteNeeds(self, expressionStr, atLeastOnes):
    expr = Expression(expressionStr)
    exprParts = expr.sumsOnly()
    if not exprParts:
      return False

    for exprPart in exprParts:
      if type(exprPart) == Expression:
        atLeastOnes.append([_aggTracker('sum', exprPart.simpleFnArg())])

    return True

class _SumProcessor:
  def processExpression(self, rewriteTarget, expr):
    extractColumn = rewriteTarget.findRewriteCol(expr.baseExpr())
    if extractColumn:
      expr.setBaseExpr(extractColumn)
    return expr

  def processRewriteNeeds(self, expressionStr, atLeastOnes):
    (aggType, aggField) = Expression(expressionStr).simpleFnComponents()
    atLeastOnes.append([_aggTracker(aggType, aggField)])
    return True

class _AvgProcessor:
  # make avg([col]) into a sum([col]) / sum(1)
  def processExpression(self, rewriteTarget, expr):
    col = expr.simpleFnArg()
    sumBaseExpr = f'sum({expr.simpleFnArg()})'
    rewriteSumBaseExpr = rewriteTarget.findRewriteCol(sumBaseExpr)
    rewriteSum1 = rewriteTarget.findRewriteCol('sum(1)')

    if rewriteSumBaseExpr and rewriteSum1:
      expr.setBaseExpr(f'{rewriteSumBaseExpr} / {rewriteSum1}')
    return expr

  def processRewriteNeeds(self, expressionStr, atLeastOnes):
    (aggType, aggField) = Expression(expressionStr).simpleFnComponents()

    # avg requires a sum on the column...
    atLeastOnes.append([_aggTracker('sum', aggField)])

    # ... AND a sum(1)
    atLeastOnes.append([_aggTracker('sum', '1')])

    return True

class _MinMaxProcessor:
  # min/max can work with the column aggregated up, or with the column as a dimension
  def processExpression(self, rewriteTarget, expr):
    newBaseExpr = rewriteTarget.findRewriteCol(expr.baseExpr())
    if newBaseExpr:
      expr.setBaseExpr(newBaseExpr)
      return expr

    arg = unbracketColumn(expr.simpleFnArg())
    newAggCol = rewriteTarget.findRewriteCol(arg)
    if newAggCol:
      # yes, the column is present.
      # however, we must refer it through the alias, not the physical column name
      expr.setSimpleFnArg(f'[{arg}]')
    return expr

  def processRewriteNeeds(self, expressionStr, atLeastOnes):
    (aggType, aggField) = Expression(expressionStr).simpleFnComponents()

    # min/max requires the extract to have a column as a dimension OR min/max(col) aggregation
    atLeastOnes.append([
      _dimTracker(aggField),
      _aggTracker(aggType, aggField)
    ])

    return True

class _NdvProcessor:
  def processExpression(self, rewriteTarget, expr):
    # ndv can work with the column as a dimension
    arg = unbracketColumn(expr.simpleFnArg())
    newAggCol = rewriteTarget.findRewriteCol(arg)
    if newAggCol:
      expr.setSimpleFnArg(f'[{arg}]') # must use alias to reference column
    return expr

  def processRewriteNeeds(self, expressionStr, atLeastOnes):
    (aggType, aggField) = Expression(expressionStr).simpleFnComponents()
    atLeastOnes.append([_dimTracker(aggField)])
    return True

#
# 'count' supports 2 distinct use-cases
#
# 1. count([col]) can be rewritten iff count[col] is an aggregate in the extract.
# - in this case, we have to rewrite the aggregation as sum([count_col])
#
# 2. count(distinct [col]) can be rewritten iff [col] is a dimension in the extract
#
class _CountProcessor:
  def _processExpressionCount(self, rewriteTarget, expr):
    col = expr.simpleFnArg()
    sumBaseExpr = f'sum({expr.simpleFnArg()})'
    rewriteSumBaseExpr = rewriteTarget.findRewriteCol(sumBaseExpr)
    if rewriteSumBaseExpr:
      expr.setBaseExpr(rewriteSumBaseExpr)
    return expr

  def _processExpressionDistinct(self, rewriteTarget, expr):
    lookfor = unbracketColumn(expr.countDistinctArg())
    newAggCol = rewriteTarget.findRewriteCol(lookfor)
    if newAggCol:
      expr.setCountDistinctArg(f'[{lookfor}]') # must use column alias
    return expr

  def processExpression(self, rewriteTarget, expr):
    if expr.isCountDistinct():
      return self._processExpressionDistinct(rewriteTarget, expr)

    return self._processExpressionCount(rewriteTarget, expr)

  def _processRewriteNeedsCount(self, expr, atLeastOnes):
    (aggType, aggField) = expr.simpleFnComponents()
    atLeastOnes.append([_aggTracker('sum', aggField)])
    return True

  def _processRewriteNeedsDistinct(self, expr, atLeastOnes):
    colname = expr.countDistinctArg()
    atLeastOnes.append([_dimTracker(colname)])
    return True

  def processRewriteNeeds(self, expressionStr, atLeastOnes):
    expr = Expression(expressionStr)
    # LOGGER.info('processRewriteNeeds %s', expr.isCountDistinct())
    if not expr.isCountDistinct():
      return self._processRewriteNeedsCount(expr, atLeastOnes)

    return self._processRewriteNeedsDistinct(expr, atLeastOnes)

class _AggExprProcessor:
  AGG_TYPE_TO_PROCESSOR = None
  AGG_TYPE_TO_PROCESSOR_LOCK = threading.Lock()
  AGG_TYPE_MULTIPLE_SUM_PROCESSOR = None

  @classmethod
  def _initAggProcessors(cls):
    if not _AggExprProcessor.AGG_TYPE_TO_PROCESSOR:
      with _AggExprProcessor.AGG_TYPE_TO_PROCESSOR_LOCK:
        if not _AggExprProcessor.AGG_TYPE_TO_PROCESSOR:
          _AggExprProcessor.AGG_TYPE_MULTIPLE_SUM_PROCESSOR = _MultipleSumProcessor()
          _AggExprProcessor.AGG_TYPE_TO_PROCESSOR = {
            'sum': _SumProcessor(),
            'avg': _AvgProcessor(),
            'min': _MinMaxProcessor(),
            'max': _MinMaxProcessor(),
            'ndv': _NdvProcessor(),
            'count': _CountProcessor(),
          }

  @classmethod
  def getAggProcessor(cls, aggType):
    _AggExprProcessor._initAggProcessors()
    return _AggExprProcessor.AGG_TYPE_TO_PROCESSOR.get(aggType)

  def __init__(self, expr):
    self._expr = expr

  def aggProcessor(self):
    aggType = self._expr.simpleFn()
    # LOGGER.info('aggProcessor %s', self._expr.components())
    if aggType:
      return _AggExprProcessor.getAggProcessor(aggType)

    if self._expr.sumsOnly():
      _AggExprProcessor._initAggProcessors()
      return _AggExprProcessor.AGG_TYPE_MULTIPLE_SUM_PROCESSOR

    return None

  def processExpression(self, rewriteTarget):
    aggProcessor = self.aggProcessor()
    if not aggProcessor:
      return self._expr.fullExpressionStr()

    expr = aggProcessor.processExpression(rewriteTarget, self._expr)
    return expr.fullExpressionStr()

  def processRewriteNeeds(self, atLeastOnes):
    aggProcessor = self.aggProcessor()
    if not aggProcessor:
      return False

    if not aggProcessor.processRewriteNeeds(self._expr.baseExpr(), atLeastOnes):
      return False

    return True

class _DsreqRewriter:
  def __init__(self, rewriteTarget=None):
    self._rewriteTarget = rewriteTarget
    self._rewrittenDataset = [{
      'tablename': self._rewriteTarget.fqTableName(),
      'columns': self._rewriteTarget.columns()
    }]

  def display(self):
    LOGGER.info('_DsreqRewriter.display() %s', json.dumps(self._rewrittenDataset, indent=2))

  def getExtractDataset(self):
    return self._rewrittenDataset

  def getExtractID(self):
    return self._rewriteTarget.getExtractID()

  def getExtractDataConnectionId(self):
    return self._rewriteTarget.getDataConnectionId()

  def processExpression(self, expressionStr):
    expr = Expression(expressionStr)
    aggExprProcessor = _AggExprProcessor(expr)
    return aggExprProcessor.processExpression(self._rewriteTarget)

__EXTRACT_REWRITE_CONTROLLER = None
__EXTRACT_REWRITE_CONTROLLER_LOCK = threading.Lock()

#
# Return a singleton instance of _ExtractRewriteController
# Note that we reuse this instance till something changes, when we
#   expect that 'resetExtractRewriteController' below will be called
#   and a new instance of _ExtractRewriteController will be created
#
def _getExtractRewriteController():
  global __EXTRACT_REWRITE_CONTROLLER

  e = __EXTRACT_REWRITE_CONTROLLER
  if e:
    return e

  with __EXTRACT_REWRITE_CONTROLLER_LOCK:
    if not __EXTRACT_REWRITE_CONTROLLER:
      __EXTRACT_REWRITE_CONTROLLER = _ExtractRewriteController()

    return __EXTRACT_REWRITE_CONTROLLER

def resetExtractRewriteController():
  global __EXTRACT_REWRITE_CONTROLLER

  if not __EXTRACT_REWRITE_CONTROLLER:
    return

  with __EXTRACT_REWRITE_CONTROLLER_LOCK:
    __EXTRACT_REWRITE_CONTROLLER = None

class _ExtractProcessor:
  def __init__(self, data):
    self._data = data
    self._dimAlias = {}
    self._aggAlias = {}
    self._columns = []
    self._stats = {}

  def findRewriteCol(self, baseExpr):
    for column in self._columns:
      if column.get('alias') == baseExpr:
        return column.get('name')

    return None

  def fqTableName(self):
    return f"{self._data['tgtDbname']}.{self._data['tgtTablename']}"

  def dimAlias(self):
    return self._dimAlias

  def setDimAlias(self, d):
    self._dimAlias = d

  def aggAlias(self):
    return self._aggAlias

  def setAggAlias(self, d):
    self._aggAlias = d

  def columns(self):
    return self._columns

  def addColumns(self, l):
    self._columns.extend(l)

  def stats(self):
    return self._stats

  def setStats(self, d):
    self._stats = d

  def getDataConnectionId(self):
    return self._data['tgtDataconnection__id']

  def getExtractID(self):
    return self._data['id']

def expandAllComputedFields(expr, modelDataset):
  exprExpanded = expr
  for cname in Expression.findColumns(expr):
    col = modelDataset.getColumn(cname)
    expandedColExpr = col.attr.get('expr')
    if expandedColExpr:
      exprExpanded = exprExpanded.replace(f'[{cname}]', f'({expandedColExpr})')

  if exprExpanded != expr:
    return expandAllComputedFields(exprExpanded, modelDataset)
  return expr

class _ExtractRewriteController:
  def __init__(self, datasetIds=None):
    self.__treeByDatasetCol = {}
    self.__treeByExtract = {}
    self.__datasetIds = datasetIds
    self.__extractsWithAggs = set()
    self.__extractsWithDimsOnly = set()
    self._computeTrees()

  def getExtracts(self):
    return self.__treeByExtract

  # Are the source and target of a data extract in compatible connections?
  # - yes if they are the same connection type
  # - also yes if they are both one of impala/arcengine
  def dataConnectionsCompatible(self, dcType1, dcType2):
    if not dcType1 or not dcType2:
      return False

    if dcType1 == dcType2:
      return True

    if isImpalaBased(dcType1) and isImpalaBased(dcType2):
      return True

    return False

  def _computeTrees(self):
    treeByDatasetCol = {}
    treeByExtract = {}

    el = fetchExtracts(self.__datasetIds)
    for d in el:
      if not self.dataConnectionsCompatible(
          d['srcDataset__dataconnection__dataconnection_type'],
          d['tgtDataconnection__dataconnection_type']):
        continue

      d['extract_config'] = json.loads(d['extract_config'])
      if _getRewriteStatusProperty(d['extract_config']) != 'enabled':
        continue

      extractId = d['id']

      # Make a clone
      treeByExtract[extractId] = _ExtractProcessor(copy.deepcopy(d))

      srcDatasetId = d.get('srcDataset__id')

      treeByDatasetCol[srcDatasetId] = treeByDatasetCol.get(srcDatasetId, {
        'extracts': [],
        'dims': {},
        'aggs': {},
      })

      treeByDatasetCol[srcDatasetId]['extracts'].append(extractId)

      extractConfig = d.get('extract_config', {})
      source = extractConfig.get('source', {})
      target = extractConfig.get('target', {})
      dims = [x.get('expr') for x in source.get('dimData', [])]
      aggs = [x.get('expr') for x in source.get('aggData', [])]
      partitions = [x.get('expr') for x in target.get('partitionData', [])]

      if aggs:
        self.__extractsWithAggs.add(d['id'])
      else:
        self.__extractsWithDimsOnly.add(d['id'])

      extract_dims = source.get('dimData', [])
      extract_parts = target.get('partitionData', [])
      extract_aggs = source.get('aggData', [])

      all_extract_columns = extract_dims + extract_parts + extract_aggs
      all_extract_colnames = sanitize_headers([c.get('alias') for c in all_extract_columns])

      treeByExtract[extractId].setDimAlias(dict(
        [(x.get('expr'), all_extract_colnames[idx]) for idx, x in enumerate(extract_dims + extract_parts)]
      ))

      aggs_index = len(extract_dims) + len(extract_parts)
      treeByExtract[extractId].setAggAlias(dict(
        [(x.get('expr'), all_extract_colnames[idx]) for idx, x in enumerate(extract_aggs, start=aggs_index)]
      ))

      treeByExtract[extractId].addColumns([
        {
          'name': alias,
          'alias': unbracketColumn(expr),
          'isdim': True,
        }
        for (expr, alias) in treeByExtract[extractId].dimAlias().items()
      ])

      # Add in aggregate columns as aliases to themselves for table substitutions
      # in ModelDataset.process_expr later on
      treeByExtract[extractId].addColumns([
        {
          'name': alias,
          'alias': alias,
          'isdim': True,
        }
        for (expr, alias) in treeByExtract[extractId].aggAlias().items()
      ])

      # Add in aggregate expressions for substitutions in ModelDataset.process_expr
      # later on
      treeByExtract[extractId].addColumns([
        {
          'name': Expression.applyAggToAlias(aggExprStr, alias),
          'alias': unbracketColumn(aggExprStr),
          'isdim': False,
        }
        for (aggExprStr, alias) in treeByExtract[extractId].aggAlias().items()
      ])

      treeByExtract[extractId].setStats({
        'dimCount': len(dims),
        'aggCount': len(aggs),
        'partitionCount': len(partitions)
      })

      for dim in itertools.chain(dims, partitions):
        treeByDatasetCol[srcDatasetId]['dims'][dim] = \
          treeByDatasetCol[srcDatasetId]['dims'].get(dim, [])
        treeByDatasetCol[srcDatasetId]['dims'][dim].append(d['id'])

      for agg in aggs:
        treeByDatasetCol[srcDatasetId]['aggs'][agg] = \
          treeByDatasetCol[srcDatasetId]['aggs'].get(agg, [])
        treeByDatasetCol[srcDatasetId]['aggs'][agg].append(d['id'])

    self.__treeByDatasetCol = treeByDatasetCol
    self.__treeByExtract = treeByExtract

  def display(self, datasetId=None):
    if not datasetId:
      LOGGER.info('_ExtractRewriteController.display() %s',
          json.dumps(self.__treeByDatasetCol, indent=4))
      return

    if not self.__treeByDatasetCol.get(datasetId):
      LOGGER.info('_ExtractRewriteController.display()')
      return

    LOGGER.info('_ExtractRewriteController.display() %s',
        json.dumps(self.__treeByDatasetCol[datasetId], indent=4))

  def findSegmentIds(self, dsreq, user: User):
    segmentIds = set()

    # segments in the filter shelf:
    for gs in dsreq.get('global_segments',[]):
      if gs.get('type') == 'SEGMENT':
        for s in gs.get('segments',[]):
          segmentIds.add(s.get('seg_id'))

    # filter associations:
    user_filter = UserFilter(user_id=user.id, dataset_id=dsreq.get('dataset_id'))
    if user_filter.association:
      fa_data = json.loads(user_filter.association.filterassociation_data)
      for segment in fa_data['applied_segments']:
        segmentIds.add(segment['id'])

    return list(segmentIds)

  def segmentExpressions(self, segmentIds: typing.List[int]):
    expressions = []

    segments = [jsonselect_get_segment(segmentId) for segmentId in segmentIds]
    for segment in segments:
      segment_data = json.loads(segment.segment_data)
      filters = segment_data.get('filters', [])
      for filter in filters:
        expressions.append(filter)

    return expressions

  def setupRewriteRules(self, dsreq, user: User, ds: object, django_ds, conn):
    # LOGGER.info('setupRewriteRules %s', json.dumps(dsreq, indent=2))
    if dsreq.get('type') != 'SQL':
      return dsreq

    mds = ds
    if not mds:
      mds = ModelDataset(django_ds.dataset_info, conn.properties())

    # TODO: stop rewrites for derived data, window functions

    # treat segment expressions as filters:
    segmentIds = self.findSegmentIds(dsreq, user=user)

    datasetId = dsreq.get('dataset_id')

    extractCols = self.__treeByDatasetCol.get(datasetId)
    if not extractCols:
      return dsreq

    atLeastOnes = []

    allCols = []

    atLeastOnesExpandedComputedFields = []
    allColsExpandedComputedFields = []
    expandedDimensions = []
    dimsExpanded = False

    filterExprs = dsreq.get('filters', [])
    if len(segmentIds) > 0:
      filterExprs = filterExprs + self.segmentExpressions(segmentIds)
    for filterExpr in filterExprs:
      for colname in Expression.findColumns(filterExpr):
        dimFinder = _dimTracker(f'[{colname}]')
        atLeastOnes.append([dimFinder])
        allCols.append(dimFinder)
        atLeastOnesExpandedComputedFields.append([dimFinder])
        allColsExpandedComputedFields.append(dimFinder)

    dimExprs = [x.get('expr') for x in dsreq.get('dimensions', []) if x.get('type') == 'SIMPLE']
    if len(dimExprs) != len(dsreq.get('dimensions', [])):
      # There are some non-SIMPLE dimensions - SEGMENT expressions, for example
      # We don't support these in rewrites
      return dsreq

    for idx, dimExprStr in enumerate(dimExprs):
      for colname in Expression.findColumns(dimExprStr):
        dimFinder = _dimTracker(f'[{colname}]')
        atLeastOnes.append([dimFinder])
        allCols.append(dimFinder)

        atLeastOnesExpandedComputedFields.append([dimFinder])
        allColsExpandedComputedFields.append(dimFinder)
        dimension = copy.deepcopy(dsreq.get('dimensions')[idx])
        expandedDimensions.append(dimension)

        column = mds.getColumn(colname)
        expandedDimExprStr = column.attr.get('expr')

        if expandedDimExprStr: # column is computed field
          dimsExpanded = True
          expandedDimExprStr = expandAllComputedFields(expandedDimExprStr, mds) # expand further all computed fields
          dimension['expr'] = dimension['expr'].replace(f'[{colname}]', f'({expandedDimExprStr})')
          expandedDimensions[idx] = dimension
          atLeastOnesExpandedComputedFields.remove([dimFinder])
          allColsExpandedComputedFields.remove(dimFinder)
          for colname in Expression.findColumns(expandedDimExprStr):
            dimFinder = _dimTracker(f'[{colname}]')
            atLeastOnesExpandedComputedFields.append([dimFinder])
            allColsExpandedComputedFields.append(dimFinder)

    aggExprStrs = [x.get('expr') for x in dsreq.get('aggregates', [])]
    for aggExprStr in aggExprStrs:
      for colname in Expression.findColumns(aggExprStr):
        dimFinder = _dimTracker(f'[{colname}]')
        allCols.append(dimFinder)
        allColsExpandedComputedFields.append(dimFinder)

      expr = Expression(aggExprStr)
      aggExprProcessor = _AggExprProcessor(expr)
      aggAtleastOnes = []
      if not aggExprProcessor.processRewriteNeeds(aggAtleastOnes):
        return dsreq

      atLeastOnes.extend(aggAtleastOnes)
      atLeastOnesExpandedComputedFields.extend(aggAtleastOnes)

    # LOGGER.info('allCols %s', json.dumps(allCols, indent=2))
    # LOGGER.info('atLeaseOnes %s', json.dumps(atLeastOnes, indent=2))

    #
    # Note that first extracts with aggregates are prioritized over extracts with dims only because
    # the ones with aggregates are going to be smaller because of the GROUP BYs. Then, among extract
    # dims computed field dims are prioritized over base field dims because computed ones contain
    # computation

    extractsByPriority = []

    # We cannot match extracts with aggregates if there are no aggregates in the dsreq
    if dsreq.get('aggregates'):
      extractsByPriority.extend(self._lookForMatchingExtractsWithAggs(extractCols, atLeastOnes))

      if not extractsByPriority and dimsExpanded:
        extractsByPriority.extend(self._lookForMatchingExtractsWithAggs(extractCols, atLeastOnesExpandedComputedFields))
        if extractsByPriority:
          dsreq['dsreqRewriter'] = _DsreqRewriter(rewriteTarget=extractsByPriority[0])
          dsreq['dimensions'] = expandedDimensions
          return dsreq

    extractsByPriority.extend(self._lookForMatchingExtractsWithDimsOnly(extractCols, allCols))

    if extractsByPriority:
      dsreq['dsreqRewriter'] = _DsreqRewriter(rewriteTarget=extractsByPriority[0])
    elif dimsExpanded:
      extractsByPriority.extend(self._lookForMatchingExtractsWithDimsOnly(extractCols, allColsExpandedComputedFields))
      if extractsByPriority:
        dsreq['dsreqRewriter'] = _DsreqRewriter(rewriteTarget=extractsByPriority[0])
        dsreq['dimensions'] = expandedDimensions

    return dsreq

  def _lookForMatchingExtractsWithDimsOnly(self, extractCols, allCols):
    extractDims = extractCols.get('dims')

    possibleExtracts = set(extractCols.get('extracts'))
    possibleExtracts = possibleExtracts.intersection(self.__extractsWithDimsOnly)
    if not possibleExtracts:
      return []

    for expr in allCols:
      # Note that based on how allCols is constructed, dimCol['type'] is 'dim'
      possibleExtracts = possibleExtracts.intersection(extractDims.get(expr['dim']) or [])
      if not possibleExtracts:
        return []

    extractsByPriority = [self.__treeByExtract.get(x) for x in possibleExtracts]
    extractsByPriority.sort(key = lambda x: x.stats().get('dimCount', 0))

    return extractsByPriority

  def _lookForMatchingExtractsWithAggs(self, extractCols, atLeastOnes):
    extractDims = extractCols.get('dims')
    extractAggs = extractCols.get('aggs')

    possibleExtracts = set(extractCols.get('extracts'))
    possibleExtracts = possibleExtracts.intersection(self.__extractsWithAggs)
    if not possibleExtracts:
      return []

    for atLeastOne in atLeastOnes:
      exprPossibleExtracts = set([])
      for expr in atLeastOne:
        if expr['type'] == 'dim':
          exprPossibleExtracts = exprPossibleExtracts.union(extractDims.get(expr['dim']) or [])
          continue

        if expr['type'] == 'agg':
          exprPossibleExtracts = exprPossibleExtracts.union(extractAggs.get(expr['exprStr']) or [])
          continue

      possibleExtracts = possibleExtracts.intersection(exprPossibleExtracts)
      if not possibleExtracts:
        return []

    extractsByPriority = [self.__treeByExtract.get(x) for x in possibleExtracts]
    extractsByPriority.sort(key = lambda x: x.stats().get('dimCount', 0))
    return extractsByPriority

# There are some functions that are private to this file, but should still be tested.
# Those are listed in TEST_ENTRYPOINTS
TEST_ENTRYPOINTS = {
  '_dimTracker': _dimTracker,
  '_aggTracker': _aggTracker,
  '_getExtractRewriteController': _getExtractRewriteController
}

