# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from django.test import TestCase, Client
from jobs.models import ScheduleGroup, JobSchedule
from apps.tests.userrole import base
from unittest.mock import patch
from vizlib.tests import util

import json

def create_dataextract():
    from datasets.models import Extract, DataSet, DataConnection
    from util.ff_flipper import FeatureFlipper
    FeatureFlipper().set_flipper('site_enable_extracts', True, 'DB')
    dc = DataConnection.objects.create(dataconnection_type='impala')
    ds = DataSet.objects.create(
        dataconnection=dc, dataset_name='name', dataset_detail='detail', dataset_type='TYPE')
    extract = Extract.objects.create(
        srcDataset=ds, tgtDataconnection=dc, extract_config={}, extract_runstate={})

    return extract


def create_role_with_permissions(user, permissions):
    from rbac.models import Role
    import json
    role = Role()
    role.name = "user"
    role.privileges = json.dumps(permissions)
    role.add_users([user.username])


class ExtractGetTest(TestCase):

    def extracts(self, client: Client, extract_id=None, dataset_id=None, runstate=False):
        data = {}
        if runstate:
            data['runstate'] = 'true'
        if dataset_id:
            data['dataset_id'] = dataset_id

        path = '/datasets/extract'
        if extract_id:
            path += f'/{extract_id}'

        response, query_count = util.run_with_db_query_check(lambda: client.get(path, data))
        return response, query_count

    @patch('rbac.permcache.RBAC')
    def test_when_get_extracts_should_contain_schedule_object(self, rbac):
        # Given
        rbac.get_user_obj_perms.return_value = ["ds_appview"]

        sg_10 = ScheduleGroup(schedule_name='schedule_a', run_interval='')
        sg_10.save()

        e = create_dataextract()
        JobSchedule.objects.create(schedule_group=sg_10, extract=e)

        user = base.make_user('user')
        client = base.create_client(user)
        # When
        response, _ = self.extracts(client)
        # Then
        resp = response.json()[0]
        self.assertEqual(resp['scheduleId'], 1)
        self.assertEqual(resp['scheduleGroupName'], 'schedule_a')

    @patch('rbac.permcache.RBAC')
    def test_when_get_extracts_with_multiple_schedule_should_return_add_jobsschdule_with_schedule_group(self, rbac):
        # Given
        rbac.get_user_obj_perms.return_value = ["ds_appview"]

        sg_10 = ScheduleGroup(schedule_name='schedule_a', run_interval='')
        sg_10.save()

        e = create_dataextract()
        JobSchedule.objects.create(extract=e)
        JobSchedule.objects.create(schedule_group=sg_10, extract=e)

        user = base.make_user('user')
        client = base.create_client(user)
        # When
        response, _ = self.extracts(client)
        # Then
        resp = response.json()[0]
        self.assertEqual(resp['scheduleId'], 2)
        self.assertEqual(resp['scheduleGroupName'], 'schedule_a')

    @patch('rbac.permcache.RBAC')
    def test_when_get_extracts_and_multiple_schedules_with_no_group_should_return_extract_with_last_created_schedule_id(self, rbac):
        # Given
        rbac.get_user_obj_perms.return_value = ["ds_appview"]

        sg_10 = ScheduleGroup(schedule_name='schedule_a', run_interval='')
        sg_10.save()

        e = create_dataextract()
        JobSchedule.objects.create(extract=e)
        JobSchedule.objects.create(extract=e)

        user = base.make_user('user')
        client = base.create_client(user)
        # When
        response, _ = self.extracts(client)
        # Then
        resp = response.json()[0]
        self.assertEqual(resp['scheduleId'], 2)

    @patch('rbac.permcache.RBAC')
    def test_when_get_extracts_and_multiple_schedules_with_group_should_return_extract_with_first_created_schedule_id(self, rbac):
        # Given
        rbac.get_user_obj_perms.return_value = ["ds_appview"]

        sg_10 = ScheduleGroup(schedule_name='schedule_a', run_interval='')
        sg_10.save()

        e = create_dataextract()
        JobSchedule.objects.create(extract=e, schedule_group=sg_10)
        JobSchedule.objects.create(extract=e, schedule_group=sg_10)

        user = base.make_user('user')
        client = base.create_client(user)
        # When
        response, _ = self.extracts(client)
        # Then
        resp = response.json()[0]
        self.assertEqual(resp['scheduleId'], 1)

    @patch('rbac.permcache.RBAC')
    def test_when_get_extracts_and_schedules_have_no_group_should_return_extract_without_schedulegroup_info(self, rbac):
        # Given
        rbac.get_user_obj_perms.return_value = ["ds_appview"]
        sg_10 = ScheduleGroup(schedule_name='schedule_a', run_interval='')
        sg_10.save()

        e = create_dataextract()
        JobSchedule.objects.create(extract=e)
        JobSchedule.objects.create(extract=e)

        user = base.make_user('user')
        client = base.create_client(user)
        # When
        response, _ = self.extracts(client)
        # Then
        resp = response.json()[0]
        self.assertEqual(resp.get('scheduleGroupName'), None)
        self.assertEqual(resp.get('scheduleGroupId'), None)

    def test_when_get_extracts_with_no_ds_appview_permission_should_return_empty_result(self):
        # Given
        sg_10 = ScheduleGroup(schedule_name='schedule_a', run_interval='')
        sg_10.save()

        e = create_dataextract()
        JobSchedule.objects.create(extract=e)
        JobSchedule.objects.create(extract=e)
        user = base.make_user('user')
        create_role_with_permissions(user, [])
        client = base.create_client(user)

        # When
        response, _ = self.extracts(client)
        # Then
        resp = response.json()
        self.assertEqual(len(resp), 0)

    def test_when_get_extracts_with_correct_ds_appview_premission_should_return_ds_extracts_user_has_premission_to(self):
        # Given
        sg_10 = ScheduleGroup(schedule_name='schedule_a', run_interval='')
        sg_10.save()

        e = create_dataextract()
        JobSchedule.objects.create(extract=e)
        JobSchedule.objects.create(extract=e)

        e2 = create_dataextract()
        JobSchedule.objects.create(extract=e2)

        user = base.make_user('user')
        client = base.create_client(user)
        create_role_with_permissions(user, [{"ptype": "dataset", "dcid": str(e.srcDataset.dataconnection.id), "dslist": [
                                     str(e.srcDataset.id)], "perms": ["ds_manage", "ds_appview", "ds_appedit"]}])
        # When
        response, _ = self.extracts(client)
        # Then
        resp = response.json()
        self.assertEqual(len(resp), 1)
        self.assertEqual(resp[0].get('id'), e.id)

    def test_retrieval_by_id(self):
        # Given
        sg_10 = ScheduleGroup(schedule_name='schedule_a', run_interval='')
        sg_10.save()

        e = create_dataextract()
        JobSchedule.objects.create(extract=e)
        JobSchedule.objects.create(extract=e)

        e2 = create_dataextract()
        JobSchedule.objects.create(extract=e2)

        user = base.make_user('user')
        client = base.create_client(user)
        create_role_with_permissions(user, [{"ptype": "dataset", "dcid": str(e.srcDataset.dataconnection.id), "dslist": [
                                     str(e.srcDataset.id)], "perms": ["ds_manage", "ds_appview", "ds_appedit"]}])
        # When
        response, query_count_1 = self.extracts(client, extract_id=e.id)
        # Then
        resp = response.json()
        self.assertEqual(len(resp), 1)
        self.assertEqual(resp[0].get('id'), e.id)
        self.assertEqual(query_count_1, 6 + 1) # two extra queries because of permission check

        # requesting the extract we have no access to:
        response, _ = self.extracts(client, extract_id=e2.id)
        self.assertTrue(response.status_code, 500)
        self.assertTrue('You do not have the necessary permission on this dataset' in response.content.decode() )

        response, query_count_2 = self.extracts(client, dataset_id=e.srcDataset.id)
        resp = response.json()
        self.assertEqual(len(resp), 1)
        self.assertEqual(resp[0].get('id'), e.id)
        self.assertEqual(query_count_2, 5 + 2 + 1) # additional query for the dataset to check the access

        # requesting the dataset we have no access to:
        response, _ = self.extracts(client, dataset_id=e2.srcDataset.id)
        self.assertTrue(response.status_code, 500)
        self.assertTrue('You do not have the necessary permission on this dataset' in response.content.decode() )



    def test_runstate_retrieval(self):
        # Given
        sg_10 = ScheduleGroup(schedule_name='schedule_a', run_interval='')
        sg_10.save()

        e = create_dataextract()

        user = base.make_user('user')
        client = base.create_client(user)
        create_role_with_permissions(user, [{"ptype": "dataset", "dcid": "-1", "dslist": ["-1"], "perms": ["ds_manage", "ds_appview", "ds_appedit"]}])
        # When
        response, query_count_1 = self.extracts(client)
        # Then
        resp = response.json()
        self.assertEqual(len(resp), 1)
        self.assertEqual(resp[0].get('id'), e.id)
        self.assertEqual(query_count_1, 8)

        # When
        response, query_count_2 = self.extracts(client,runstate=True)
        # Then
        resp = response.json()
        self.assertEqual(len(resp), 1)
        self.assertEqual(resp[0].get('id'), e.id)
        # query count must not increase:
        self.assertEqual(query_count_2, 8)

        # having extra data must not result in increased query count:
        JobSchedule.objects.create(extract=e)
        JobSchedule.objects.create(extract=e)

        e2 = create_dataextract()
        runstate = {
            "a":"b"
        }
        e2.extract_runstate=json.dumps(runstate)
        e2.save()
        JobSchedule.objects.create(extract=e2)

        # When
        response, query_count_3 = self.extracts(client,runstate=True)
        # Then
        resp = response.json()
        self.assertEqual(len(resp), 2)
        self.assertEqual(resp[0].get('id'), e.id)
        self.assertEqual(resp[1].get('id'), e2.id)

        self.assertEqual(resp[1].get('extract_runstate'), runstate)
        # query count must not increase:
        self.assertEqual(query_count_3, 8)

