# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
These tests verify the python code in arcweb/extracts/
"""

import json

from django.test import TestCase, RequestFactory
from django.conf import settings
from datasets.models import DataConnection, DataSet

import extracts.views
import extracts.misc
import extracts.partitioned
import extracts.full
from util.ff_flipper import FeatureFlipper

from unittest import mock

class MockUser(object):
  def __init__(self, username=None, test=None):
    self.test = test
    self.username = username

class MockRequest(object):
  def __init__(self, username=None, test=None):
    self.test = test
    self.user = MockUser(username)

class MockDataSet(object):
  def __init__(self, id, test):
    self.test = test
    self.id = id
    self.dataconnection = self.test.getDataConnection(self.id + 1000)
    self.dataset_type = 'singletable'
    self.dataset_info = json.dumps([
      {
        'tablename': 'db1.src_table1',
        'columns': [
          {
            'alias': 'col1',
          },
          {
            'alias': 'col2',
          },
          {
            'alias': 'Customer Name',
            'name': 'customer',
          },
          {
            'alias': 'Product Name',
            'name': 'product',
          },
          {
            'alias': 'Price in USD',
            'name': 'price',
          },
          {
            'alias': 'Sales year',
            'name': 'year'
          },
          {
            'alias': 'Sales month',
            'name': 'month'
          },
        ],
      },
    ])

class MockProperties(object):
  def __init__(self, dc, conn_type = None):
    self.conn_type = conn_type
    self.dc = dc
    self.supports_dataextract_source = self.dc.test.supports_dataextract_source
    self.supports_partitions = self.dc.test.supportsPartitions(self)
    self.supports_dataextract_target = self.dc.test.supports_dataextract_target
    self.quote_alias = '`'
    self.quote = '"'
    self.quote_end = '"'
    self.quote_escape_character = "\\"
    self.supports_table_aliases = True
    self.alias_supports_spaces = True
    self.alias_supports_backticks = True
    self.supports_top = False
    self.sb_specify_alias = True
    self.supports_orderby = True
    self.supports_limit = True
    self.supports_limit_0 = True
    self.supports_groupby_colindex = True
    self.paramstyle = '?'

class MockDataConnection(object):
  def __init__(self, id, test):
    self.test = test
    self.id = id
    self.connection_type = "impala"
    self.cmds = []
    self.createCmdParams = []
    self.dropCmdParams = []
    self.dropNoCheckCmdParams = []
    self.renameCmdParams = []

  def quote_tablename(self, name):
    return name

  def properties(self):
    return MockProperties(self)

  def tablename_parts(self, fqtablename):
    (l, _, r) = fqtablename.partition(".")
    return [l, r]

  def log_sync_runquery(self, query, username=None):
    self.cmds.append([query, username])
    if query.startswith("SHOW FILES IN"):
      return {
        "rows": json.dumps([
          ["path1", "12k", "year=2020/month=3"],
          ["path2", "12k", "year=2019/month=2"],
          ["path3", "12k", "year=2021/month=3"],
          ["path4", "12k", "year=2020/month=2"],
          ["path5", "12k", "year=2020/month=1"],
        ]),
        "colnames": ["Path", "Size", "Partition"],
      }

    return {
      "coltypes": ["STRING", "INT", "REAL"]
    }

  def truncate_table_cmd(self, tablename):
    return f"TRUNCATE TABLE {tablename}"

  def rename_table_cmd(self, old_table_name, new_table_name):
    self.renameCmdParams.append([old_table_name, new_table_name])
    return 'rename e f'

  def create_sql_cmds(self, tablename, colnames, coltypes,
        importuser, importsrcfile, external_table,
        location, partitionColnames=None, partitionColtypes=None, tblproperties=[]):

    self.createCmdParams.append([tablename, colnames, coltypes,
        partitionColnames, partitionColtypes])
    return "create a b"

  def drop_table_cmd(self, tablename):
    self.dropCmdParams.append(tablename)
    return "drop c d"

  def drop_table_no_check_cmd(self, tablename):
    self.dropNoCheckCmdParams.append(tablename)
    return "drop c d"

  def insert_table_prefix(self, tablename, overwrite=False):
    if overwrite:
      return f"INSERT OVERWRITE {self.quote_tablename(tablename)}"

    return f"INSERT INTO {self.quote_tablename(tablename)}"

  def upload_maxrows(self):
    return settings.UPLOAD_MAXROWS_DEFAULT

  def cursor_get(self, username):
    self.cmds.append(['cursor_get', username])
    return MockCursor(self)

  def executemany(self, username, query, rows):
    self.cmds.append([f'{query} {rows}', username])

  def table_columns(self, tablename, username=None):
    self.cmds.append([f'table_columns {tablename}', username])
    return {
      "colnames": ["customer", "product", "price", "year", "month"],
      "coltypes": ["String", "String", "Real", "Integer", "Integer"],
      "partitions": ["year", "month"],
    }

class MockCursor(object):
  def __init__(self, dc):
    self.dc = dc
    self.rows = [
      None,
      [
        [9, 'nine'],
      ],
      [
        [7, 'seven'],
        [8, 'eight'],
      ],
      None,
      [
        [5, 'five'],
        [6, 'six'],
      ],
      None,
      [
        [3, 'three'],
        [4, 'four'],
      ],
      [
        [1, 'one'],
        [2, 'two'],
      ],
    ]

  def execute(self, cmd):
    self.dc.cmds.append([cmd])

  def fetchmany(self, size=None):
    self.dc.cmds.append(['fetchmany'])
    return self.rows.pop()

class MockExtractInvalidType(object):
  def __init__(self, test):
    self.test = test
    self.id = 1
    self.extract_config = json.dumps({})
    self.extract_runstate = json.dumps({})
    self.extract_type = "abcd"

class MockExtractFull(object):
  def __init__(self, test):
    self.test = test
    self.id = 1
    srcDatasetId = test.srcDatasetId
    tgtDataconnectionId = test.tgtDataconnectionId

    self.srcDataset = MockDataSet(srcDatasetId, test)
    self.tgtDataconnection = test.getDataConnection(tgtDataconnectionId)

    self.tgtDbname = test.tgtDbname
    self.tgtTablename = test.tgtTablename

    self.extract_type = "refresh_full"

    self.extract_config = json.dumps({
      "username": "foo_user",
      "source": {
        "dimData": [
          {
            'expr': '[customer]',
            'alias': 'customer',
            'type': 'String',
          },
          {
            'expr': '[product]',
            'alias': 'product',
            'type': 'String',
          },
        ],
        "aggData": [
          {
            'expr': 'sum([price])',
            'alias': 'price',
            'type': 'Real',
          },
        ],
      },
      "target": {
        "partitionData": [
        ],
      },
      'rewrite_properties': {
        'rewrite_status': 'disabled_until_refreshed'
      }
    })

    self.extract_runstate = json.dumps({
    })

class MockExtractPartitioned(object):
  def __init__(self, extractId, test):
    self.cmds = []

    self.test = test
    self.id = extractId
    srcDatasetId = test.srcDatasetId
    tgtDataconnectionId = test.tgtDataconnectionId

    self.srcDataset = MockDataSet(srcDatasetId, test)
    self.tgtDataconnection = test.getDataConnection(tgtDataconnectionId)

    self.tgtDbname = test.tgtDbname
    self.tgtTablename = test.tgtTablename

    self.extract_type = "refresh_partitioned"

    self.extract_config = json.dumps({
      "username": "foo_user",
      "source": {
        "dimData": [
          {
            'expr': '[Customer Name]',
            'alias': 'Customer Name',
            'type': 'String',
          },
          {
            'expr': '[Product Name]',
            'alias': 'Product Name',
            'type': 'String',
          },
        ],
        "aggData": [
          {
            'expr': 'sum([Price in USD])',
            'alias': 'Product Price',
            'type': 'Real',
          },
        ],
      },
      "target": {
        "partitionData": [
          {
            'expr': '[Sales year]',
            'alias': 'Sales year',
            'type': 'Integer',
          },
        ],
      }
    })

    self.extract_runstate = json.dumps({
    })

  def save(self):
    self.cmds.append(['save', self.extract_runstate])

class Test_Extracts(TestCase):
  def __init__(self, *args, **kwargs):
    super(Test_Extracts, self).__init__(*args, **kwargs)

  def setUp(self):
    self.username = 'foo_user'
    self.request = MockRequest(self.username, self)
    FeatureFlipper().set_flipper('site_enable_extracts', True, 'DB')

  def tearDown(self):
    FeatureFlipper.instance = None
    return super().tearDown()

  def test_invalidType(self):
    extract = MockExtractInvalidType(self)
    with self.assertRaisesRegex(Exception, "Invalid type abcd"):
      extracts.views.extractManagerFromExtract(extract)

  def test_addLimit(self):
    class Props(object):
      def __init__(self):
        self.supports_limit = True
        self.supports_limit_0 = True

    query = extracts.misc.addLimit('select a from b', connProperties = Props())
    self.assertEquals(query, 'select a from b limit 0')

    query = extracts.misc.addLimit('select a from b limit 6', connProperties = Props())
    self.assertEquals(query, 'select a from b limit 6')

    query = extracts.misc.addLimit('select a from b', limit = 9, connProperties = Props())
    self.assertEquals(query, 'select a from b limit 9')

    query = extracts.misc.addLimit('select a from b', limit = 9)
    self.assertEquals(query, 'select a from b limit 9')

    with self.assertRaisesRegex(Exception, "No limit and no connection properties"):
      query = extracts.misc.addLimit('select a from b')

    class Props2(object):
      def __init__(self):
        self.supports_limit = True
        self.supports_limit_0 = False

    query = extracts.misc.addLimit('select a from b', connProperties = Props2())
    self.assertEquals(query, 'select a from b limit 1')

    class Props3(object):
      def __init__(self):
        self.supports_limit = False

    with self.assertRaisesRegex(Exception, "Invalid dataconnection for adding limit"):
      extracts.misc.addLimit('select a from b', connProperties = Props3())

  def test_getColumnTypes(self):
    class Props(object):
      def __init__(self):
        self.supports_limit = True
        self.supports_limit_0 = True

    class Conn(object):
      def __init__(self):
        self.cmds = []

      def log_sync_runquery(self, query, username=None):
        self.cmds.append([query, username])
        return {
          "coltypes": ["STRING", "INT", "REAL"]
        }

    query = 'select c from d'
    username = "user_abc"
    conn = Conn()

    coltypes = extracts.misc.getColumnTypes(conn=conn,
        connProperties=Props(), query=query, username=username)

    self.assertEquals(coltypes, ['String', 'Integer', 'Real'])
    self.assertEquals(len(conn.cmds), 1)
    self.assertEquals(conn.cmds[0][0], f"{query} limit 0")
    self.assertEquals(conn.cmds[0][1], username)

class Test_Extracts_FullBase(TestCase):
  def __init__(self, *args, **kwargs):
    super(Test_Extracts_FullBase, self).__init__(*args, **kwargs)
    self.conns = {}
    self.extracts = {}
    self.srcDatasetId = 2
    self.tgtDataconnectionId = 3
    self.extractId = 12

  def getDataConnection(self, dataConnectionId):
    if self.conns.get(dataConnectionId):
      return self.conns[dataConnectionId]

    self.conns[dataConnectionId] = self.createDataConnection(dataConnectionId)
    return self.conns[dataConnectionId]

  def createDataConnection(self, dataConnectionId: int):
    return MockDataConnection(dataConnectionId, self)

  def setUp(self):
    self.username = 'foo_user'
    self.request = MockRequest(self.username, self)
    self.tgtDbname = "db_foo"
    self.tgtTablename = "cdv_bar"
    self.supports_dataextract_source = True
    self.supports_partitions = True
    self.supports_partitions_by_id = {}
    self.supports_dataextract_target = True

  def supportsPartitions(self, properties):
    s = self.supports_partitions_by_id.get(properties.dc.id)
    if s is not None:
      return s

    return self.supports_partitions

  @mock.patch('extracts.misc.enabled')
  def test_disabled(self, mock_enabled):
    mock_enabled.side_effect = self.mock_enabled
    with self.assertRaisesRegex(Exception, "Data Extracts not enabled"):
      self.test_base()

  def test_invalidTable(self):
    self.tgtTablename = "bar"
    with self.assertRaisesRegex(Exception, "Target table name must start with cdv_"):
      self.test_base()

  def test_invalidTable2(self):
    self.tgtTablename = "cdv_"
    with self.assertRaisesRegex(Exception, "Invalid target table name cdv_"):
      self.test_base()

  def test_invalidSrcDataConn(self):
    self.supports_dataextract_source = False
    with self.assertRaisesRegex(Exception, "The source data connection does not support extracts"):
      self.test_base()

  @mock.patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'site_enable_extracts', 'label': 'Enable Data Extracts', 'value': True, 'level': 'SETTINGS'}])
  def test_invalidTgtDataConn(self):
    self.supports_dataextract_target = False
    with self.assertRaisesRegex(Exception, "The target data connection does not support extracts"):
      self.test_base()

  @mock.patch('extracts.full.User')
  @mock.patch('extracts.misc.fetchDataset')
  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'site_enable_extracts', 'label': 'Enable Data Extracts', 'value': True, 'level': 'SETTINGS'}])
  def test_base(self, mock_fetchDataConn, mock_fetchDataset, mock_User):
    mock_fetchDataset.side_effect = self.mock_fetchDataset
    mock_fetchDataConn.side_effect = self.mock_fetchDataConn

    extract = MockExtractFull(self)
    self.mgr = extracts.views.extractManagerFromExtract(extract)
    self.assertEquals(type(self.mgr), extracts.full.ExtractManagerFull)

  def mock_enabled(self, *args, **kwargs):
    return False

  def mock_fetchDataset(self, *args, **kwargs):
    return MockDataSet(args[0], self)

  def mock_fetchDataConn(self, *args, **kwargs):
    return self.getDataConnection(args[0])

class Test_Extracts_Full(Test_Extracts_FullBase):
  def __init__(self, *args, **kwargs):
    super(Test_Extracts_Full, self).__init__(*args, **kwargs)

  # Note that not supporting partitions does not have an impact on "full" extracts
  @mock.patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'site_enable_extracts', 'label': 'Enable Data Extracts', 'value': True, 'level': 'SETTINGS'}])
  def test_noPartitionSupport(self):
    self.supports_partitions = False
    self.test_base()

  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'site_enable_extracts', 'label': 'Enable Data Extracts', 'value': True, 'level': 'SETTINGS'}])
  def test_create(self, mock_fetchDataConn):
    mock_fetchDataConn.side_effect = self.mock_fetchDataConn

    self.test_base()
    self.mgr.runCreate()
    srcDataConn = self.conns[1002]
    tgtDataConn = self.conns[3]

    self.assertEquals(len(srcDataConn.cmds), 0)
    self.assertEquals([x[0] for x in tgtDataConn.cmds],
      ['create a b'])
    self.assertEquals(tgtDataConn.createCmdParams,
      [
        ['db_foo.cdv_bar',
          ['customer', 'product', 'price'],
          ['String', 'String', 'Real'],
          None,
          None
        ]
      ])

    self.assertEquals(tgtDataConn.dropCmdParams, [])

  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'site_enable_extracts', 'label': 'Enable Data Extracts', 'value': True, 'level': 'SETTINGS'}])
  def test_runSameConn(self, mock_fetchDataConn):
    mock_fetchDataConn.side_effect = self.mock_fetchDataConn

    self.tgtDataconnectionId = 1002
    self.test_base()
    self.mgr._extractTarget._tempFqtablename = 'db_foo.temp_cdv_bar'
    self.mgr.runRefresh()

    dataConn = self.conns[1002]
    self.assertEquals([x[0] for x in dataConn.cmds],
      [
        'SELECT [customer] as `[customer]`, [product] as `[product]`, sum([price]) as `sum([price])`\n  FROM "db1"."src_table1"  TA_0 \n  GROUP BY 1, 2 limit 0',
        'create a b',
        'INSERT INTO db_foo.temp_cdv_bar SELECT [customer] as `[customer]`, [product] as `[product]`, sum([price]) as `sum([price])`\n  FROM "db1"."src_table1"  TA_0 \n  GROUP BY 1, 2',
        'drop c d',
        'rename e f'
      ])
    self.assertEquals(dataConn.createCmdParams,
      [
        ['db_foo.temp_cdv_bar',
          ['customer', 'product', 'price'],
          ['String', 'String', 'Real'],
          None,
          None
        ]
      ])
    self.assertEquals(dataConn.dropNoCheckCmdParams, ['db_foo.cdv_bar'])
    self.assertEquals(dataConn.renameCmdParams, [['db_foo.temp_cdv_bar', 'db_foo.cdv_bar']])

  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'site_enable_extracts', 'label': 'Enable Data Extracts', 'value': True, 'level': 'SETTINGS'}])
  def test_runDiffConn(self, mock_fetchDataConn):
    mock_fetchDataConn.side_effect = self.mock_fetchDataConn

    self.test_base()
    self.mgr._extractTarget._tempFqtablename = 'db_foo.temp_cdv_bar'
    self.mgr.runRefresh()

    srcConn = self.conns[1002]
    self.assertEquals([x[0] for x in srcConn.cmds],
      [
        'cursor_get',
        'SELECT [customer] as `[customer]`, [product] as `[product]`, sum([price]) as `sum([price])`\n  FROM "db1"."src_table1"  TA_0 \n  GROUP BY 1, 2',
        'fetchmany',
        'fetchmany',
        'fetchmany',
      ])

    tgtConn = self.conns[3]
    self.assertEquals([x[0] for x in tgtConn.cmds],
      [
        'create a b',
        "INSERT INTO db_foo.temp_cdv_bar values (?,?,?) [[1, 'one'], [2, 'two']]",
        "INSERT INTO db_foo.temp_cdv_bar values (?,?,?) [[3, 'three'], [4, 'four']]",
        'drop c d',
        'rename e f'
      ])
    self.assertEquals(tgtConn.createCmdParams,
      [
        ['db_foo.temp_cdv_bar',
          ['customer', 'product', 'price'],
          ['String', 'String', 'Real'],
          None,
          None
        ]
      ])
    self.assertEquals(tgtConn.dropNoCheckCmdParams, ['db_foo.cdv_bar'])
    self.assertEquals(tgtConn.renameCmdParams, [['db_foo.temp_cdv_bar', 'db_foo.cdv_bar']])


  @mock.patch('extracts.views._extract_get')
  @mock.patch('extracts.rewriter.resetExtractRewriteController')
  def test_retrieval_no_reset(self, mock_resetExtractRewriteController: mock.MagicMock, mock_extract_get: mock.MagicMock):

    request = MockRequest()
    request.method='GET'
    request.GET={}

    extracts.views.extract_internal(request)

    mock_extract_get.assert_called_once()
    mock_resetExtractRewriteController.assert_not_called()


  @mock.patch('extracts.views._extract_get')
  @mock.patch('extracts.rewriter.resetExtractRewriteController')
  def test_retrieval_with_reset(self, mock_resetExtractRewriteController: mock.MagicMock, mock_extract_get: mock.MagicMock):

    request = MockRequest()
    request.method='GET'
    request.GET={
      'reset_controller': 'true'
    }

    extracts.views.extract_internal(request)

    mock_extract_get.assert_called_once()
    mock_resetExtractRewriteController.assert_called_once()

  @mock.patch('extracts.views._extract_get')
  @mock.patch('extracts.views.extractManagerFromExtract')
  @mock.patch('extracts.views._process_cols_data')
  @mock.patch('extracts.misc.fetchDataConn')
  def test_create_using_rest(self, mock_fetchDataConn, mock_process_cols_data, mock_extractManagerFromExtract, mock_extract_get: mock.MagicMock):
    self._do_test_create_using_rest(mock_fetchDataConn, mock_process_cols_data, mock_extractManagerFromExtract, mock_extract_get, createTargetTable=False)

  @mock.patch('extracts.views._extract_get')
  @mock.patch('extracts.views.extractManagerFromExtract')
  @mock.patch('extracts.views._process_cols_data')
  @mock.patch('extracts.misc.fetchDataConn')
  def test_create_using_rest_no_target_table(self, mock_fetchDataConn, mock_process_cols_data, mock_extractManagerFromExtract, mock_extract_get: mock.MagicMock):
    self._do_test_create_using_rest(mock_fetchDataConn, mock_process_cols_data, mock_extractManagerFromExtract, mock_extract_get, createTargetTable=True)

  def _do_test_create_using_rest(self, mock_fetchDataConn, mock_process_cols_data, mock_extractManagerFromExtract, mock_extract_get: mock.MagicMock, createTargetTable: bool):
    mock_fetchDataConn.side_effect = self.mock_fetchDataConn
    mock_process_cols_data.return_value = {'dimData':[],'aggData':[],'partitionData':[]}
    extractManagerMock = mock.MagicMock()
    mock_extractManagerFromExtract.return_value = extractManagerMock

    dc = DataConnection(dataconnection_type = 'sqlite')
    dc.save()
    dataset = DataSet(
      dataconnection=dc,
      dataset_name='Source DS',
      dataset_type='SQLite',
      dataset_detail='test dataset')
    dataset.save()

    request = MockRequest()
    request.user.username = 'admin'
    request.user.is_superuser = True
    request.method='POST'

    payload = {
      'srcDatasetId': dataset.id,
      'tgtDataconnectionId': dc.id,
      'tgtDbname': 'target_db',
      'tgtTablename': 'target_table',
      'extract_type': 'refresh_full',
      'dimData': [],
      'aggData': [{'expr':'sum(1)'}]
    }
    if not createTargetTable:
      payload['createTargetTable'] = 'false'
    request.body = json.dumps(payload)

    extracts.views.extract_internal(request)

    mock_extract_get.assert_called_once()

    if not createTargetTable:
      extractManagerMock.runCreate.assert_not_called()
    else:
      extractManagerMock.runCreate.assert_called_once()


  @mock.patch('extracts.rewriter.updateRewriteStatusProperty')
  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'site_enable_extracts', 'label': 'Enable Data Extracts', 'value': True, 'level': 'SETTINGS'}])
  def test_disabled_until_refreshed_to_enabled(self, mock_fetchDataConn, mock_updateRewriteStatusProperty):

    mock_fetchDataConn.side_effect = self.mock_fetchDataConn

    self.test_base()
    self.mgr.runRefresh()

    # if extract rewrite status is disabled_until_refreshed, then next refresh sets status to enabled
    mock_updateRewriteStatusProperty.assert_called_with(1, 'enabled')

class Test_Extracts_Partitioned(Test_Extracts_FullBase):
  def __init__(self, *args, **kwargs):
    super(Test_Extracts_Partitioned, self).__init__(*args, **kwargs)

  def setUp(self):
    super().setUp()
    FeatureFlipper().set_flipper('site_enable_extracts', True, 'DB')

  def tearDown(self):
    FeatureFlipper.instance = None
    return super().tearDown()

  def getExtract(self, extractId):
    if self.extracts.get(extractId):
      return self.extracts[extractId]

    self.extracts[extractId] = MockExtractPartitioned(extractId, self)
    return self.extracts[extractId]

  # Note that not supporting partitions *does* have an impact on "partitioned" extracts
  def test_noPartitionSupport(self):
    self.supports_partitions_by_id[1002] = False
    with self.assertRaisesRegex(Exception, "Partitioned data extracts can only " + \
        "be run on source connections that support partitions"):
      self.test_base()

  def test_noPartitionSupportTarget(self):
    self.supports_partitions_by_id[3] = False
    with self.assertRaisesRegex(Exception, "Partitioned data extracts can only " + \
        "be run on target connections that support partitions"):
      self.test_base()

  @mock.patch('extracts.partitioned.User')
  @mock.patch('extracts.misc.fetchDataset')
  @mock.patch('extracts.misc.fetchDataConn')
  def test_base(self, mock_fetchDataConn, mock_fetchDataset, mock_User):
    mock_fetchDataset.side_effect = self.mock_fetchDataset
    mock_fetchDataConn.side_effect = self.mock_fetchDataConn

    extract = self.getExtract(self.extractId)
    self.mgr = extracts.views.extractManagerFromExtract(extract)
    self.assertEquals(type(self.mgr), extracts.partitioned.ExtractManagerPartitioned)

  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'site_enable_extracts', 'label': 'Enable Data Extracts', 'value': True, 'level': 'SETTINGS'}])
  def test_create(self, mock_fetchDataConn):
    mock_fetchDataConn.side_effect = self.mock_fetchDataConn

    self.test_base()
    self.mgr.runCreate()
    srcDataConn = self.conns[1002]
    tgtDataConn = self.conns[3]

    self.assertEquals(len(srcDataConn.cmds), 0)
    self.assertEquals([x[0] for x in tgtDataConn.cmds],
      ['create a b'])
    self.assertEquals(tgtDataConn.createCmdParams,
      [
        ['db_foo.cdv_bar',
          ['customer_name', 'product_name', 'product_price'],
          ['String', 'String', 'Real'],
          ['sales_year'],
          ['Integer']
        ]
      ])

    self.assertEquals(tgtDataConn.dropCmdParams, [])

  def mock_fetchExtract(self, extractId):
    return self.getExtract(extractId)

  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('extracts.misc.fetchDataset')
  @mock.patch('extracts.partitioned.fetchExtract')
  def test_runSameConn(self, mock_fetchExtract, mock_fetchDataset, mock_fetchDataConn):
    self.do_test_runSameConn(mock_fetchExtract, mock_fetchDataset, mock_fetchDataConn)

  def do_test_runSameConn(self, mock_fetchExtract, mock_fetchDataset, mock_fetchDataConn):
    mock_fetchExtract.side_effect = self.mock_fetchExtract
    mock_fetchDataset.side_effect = self.mock_fetchDataset
    mock_fetchDataConn.side_effect = self.mock_fetchDataConn

    self.tgtDataconnectionId = 1002
    self.test_base()
    self.mgr.runRefresh()

    dataConn = self.conns[1002]
    self.assertEquals([x[0] for x in dataConn.cmds],
      [
        'table_columns db1.src_table1',
        'SHOW FILES IN db1.src_table1',
        'INSERT OVERWRITE db_foo.cdv_bar PARTITION (sales_year=2020) SELECT TA_0."customer" as `TA_0."customer"`, TA_0."product" as `TA_0."product"`, sum((TA_0."price")) as `sum((TA_0."price"))`\n  FROM "db1"."src_table1"  TA_0 \n  WHERE (TA_0."year"=2020)\n  GROUP BY 1, 2',
        'INSERT OVERWRITE db_foo.cdv_bar PARTITION (sales_year=2019) SELECT TA_0."customer" as `TA_0."customer"`, TA_0."product" as `TA_0."product"`, sum((TA_0."price")) as `sum((TA_0."price"))`\n  FROM "db1"."src_table1"  TA_0 \n  WHERE (TA_0."year"=2019)\n  GROUP BY 1, 2',
        'INSERT OVERWRITE db_foo.cdv_bar PARTITION (sales_year=2021) SELECT TA_0."customer" as `TA_0."customer"`, TA_0."product" as `TA_0."product"`, sum((TA_0."price")) as `sum((TA_0."price"))`\n  FROM "db1"."src_table1"  TA_0 \n  WHERE (TA_0."year"=2021)\n  GROUP BY 1, 2',
      ])

    self.assertEquals(self.mgr._extractTarget.partitionSignatures(),
        {
          '["2020"]': '41d78bf2037a58c6fda244289bffdb8271625b8b1a7e562e637f645e5b102e75',
          '["2019"]': 'cd522283c28fd83b876a20e40375a7e74a5fc9b83678c9498a1b350344377988',
          '["2021"]': '1e5d2f1c4a77f8f23d221afaa5fb62cba466f96f061ead343e57b3e40ac574d2'
        })

    # Reset commands, and run refresh again
    dataConn.cmds = []
    self.mgr.runRefresh()

    dataConn = self.conns[1002]
    self.assertEquals([x[0] for x in dataConn.cmds],
      [
        'table_columns db1.src_table1',
        'SHOW FILES IN db1.src_table1',
      ])

    # Pretend one of the partition signatures is different, and run refresh again
    self.mgr._extractTarget._partitionSignatures['["2020"]'] = '12'
    dataConn.cmds = []
    self.mgr.runRefresh()

    self.assertEquals([x[0] for x in dataConn.cmds],
      [
        'table_columns db1.src_table1',
        'SHOW FILES IN db1.src_table1',
        'INSERT OVERWRITE db_foo.cdv_bar PARTITION (sales_year=2020) SELECT TA_0."customer" as `TA_0."customer"`, TA_0."product" as `TA_0."product"`, sum((TA_0."price")) as `sum((TA_0."price"))`\n  FROM "db1"."src_table1"  TA_0 \n  WHERE (TA_0."year"=2020)\n  GROUP BY 1, 2',
      ])

    # Pretend two of the partition signatures are different, and we have an extra 3rd one,
    # and run refresh again
    self.mgr._extractTarget._partitionSignatures['["2019"]'] = '34'
    self.mgr._extractTarget._partitionSignatures['["2020"]'] = '12'
    self.mgr._extractTarget._partitionSignatures['["1900"]'] = '12'
    dataConn.cmds = []
    self.mgr.runRefresh()

    self.assertEquals([x[0] for x in dataConn.cmds],
      [
        'table_columns db1.src_table1',
        'SHOW FILES IN db1.src_table1',
        'ALTER TABLE db_foo.cdv_bar DROP IF EXISTS PARTITION (sales_year=1900)',
        'INSERT OVERWRITE db_foo.cdv_bar PARTITION (sales_year=2020) SELECT TA_0."customer" as `TA_0."customer"`, TA_0."product" as `TA_0."product"`, sum((TA_0."price")) as `sum((TA_0."price"))`\n  FROM "db1"."src_table1"  TA_0 \n  WHERE (TA_0."year"=2020)\n  GROUP BY 1, 2',
        'INSERT OVERWRITE db_foo.cdv_bar PARTITION (sales_year=2019) SELECT TA_0."customer" as `TA_0."customer"`, TA_0."product" as `TA_0."product"`, sum((TA_0."price")) as `sum((TA_0."price"))`\n  FROM "db1"."src_table1"  TA_0 \n  WHERE (TA_0."year"=2019)\n  GROUP BY 1, 2',
      ])

  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('extracts.misc.fetchDataset')
  @mock.patch('extracts.partitioned.fetchExtract')
  def test_runDiffConn(self, mock_fetchExtract, mock_fetchDataset, mock_fetchDataConn):
    mock_fetchExtract.side_effect = self.mock_fetchExtract
    mock_fetchDataset.side_effect = self.mock_fetchDataset
    mock_fetchDataConn.side_effect = self.mock_fetchDataConn

    self.test_base()
    with self.assertRaisesRegex(Exception, "Partitioned extracts across connections " +
        "is not yet supported"):
      self.mgr.runRefresh()

class MockDataConnectionImpala719(MockDataConnection):
  def __init__(self, id, test):
    MockDataConnection.__init__(self, id, test)

  def log_sync_runquery(self, query, username=None):
    self.cmds.append([query, username])
    if query.startswith("SHOW FILES IN"):
      return {
        "rows": json.dumps([
          ["path1", "12k", "year=2020/month=3", "", "a"],
          ["path2", "12k", "year=2019/month=2", "", "b"],
          ["path3", "12k", "year=2021/month=3", "", "c"],
          ["path4", "12k", "year=2020/month=2", "", "b"],
          ["path5", "12k", "year=2020/month=1", "", "a"],
        ]),
        # simulate extra columns present in SHOW FILES output:
        "colnames": ["Path", "Size", "Partition", "EC Policy", "extra column"],
      }

    return {
      "coltypes": ["STRING", "INT", "REAL", "STRING", "STRING"]
    }

class Test_Extracts_Partitioned_On_Impala719(Test_Extracts_Partitioned):
  def __init__(self, *args, **kwargs):
    super(Test_Extracts_Partitioned_On_Impala719, self).__init__(*args, **kwargs)

  def createDataConnection(self, dataConnectionId: int):
    return MockDataConnectionImpala719(dataConnectionId, self)

class MockDataConnectionBroken(MockDataConnection):
  def __init__(self, id, test):
    MockDataConnection.__init__(self, id, test)

  def log_sync_runquery(self, query, username=None):
    self.cmds.append([query, username])
    if query.startswith("SHOW FILES IN"):
      return {
        "rows": json.dumps([
          ["path1", "12k", "year=2020/month=3"],
        ]),
        # simulate case when partition information is missing:
        "colnames": ["Path", "Size", "extra column"],
      }

    return {
      "coltypes": ["STRING", "INT", "REAL"]
    }

class Test_Extracts_Partitioned_On_Broken(Test_Extracts_Partitioned):
  def __init__(self, *args, **kwargs):
    super(Test_Extracts_Partitioned_On_Broken, self).__init__(*args, **kwargs)

  def createDataConnection(self, dataConnectionId: int):
    return MockDataConnectionBroken(dataConnectionId, self)

  @mock.patch('extracts.misc.fetchDataConn')
  @mock.patch('extracts.misc.fetchDataset')
  @mock.patch('extracts.partitioned.fetchExtract')
  def test_runSameConn(self, mock_fetchExtract, mock_fetchDataset, mock_fetchDataConn):
    # partition refreshing can not be run if SHOW FILES is missing columns
    with self.assertRaisesRegex(Exception, "Expected column not found in output from SHOW FILES"):
      self.do_test_runSameConn(mock_fetchExtract, mock_fetchDataset, mock_fetchDataConn)

  def test_runDiffConn(self):
    self.skipTest('Scenario is tested in other classes')