# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import copy
import json
import os

from django.test import TestCase, RequestFactory
from django.test.client import Client

from dataconn.dataconn_dataengine import BuildRequest
from datasets.models import DataSet, DataConnection, Segment

from vizlib.dsreq import DsreqCreator
from dataconn.qcntl import QueryControl
from vizlib.arcprof import ArcProf
from vizlib.dataset_request import DatasetRequest
from vizlib.tests import test_sqlrun

import extracts.misc
import extracts.rewriter
import vizlib.util

from vizlib.expression import Expression

from unittest import mock
from util.ff_flipper import FeatureFlipper

import logging

LOGGER = logging.getLogger('arcweb')
PRINT_BUILD_REQUEST = os.environ.get('PRINT_BUILD_REQUEST','0') == '1'
class _MockRewriteTarget:
  def __init__(self, test, expectedExpressions, newExpressions):
    self.test = test
    self.expectedExpressions = expectedExpressions
    self.newExpressions = newExpressions

  def findRewriteCol(self, expressionStr):
    self.test.assertEquals(expressionStr, self.expectedExpressions.pop(0))
    return self.newExpressions.pop(0)

class _MockExtract:
  def __init__(self, id, test):
    self.id = id
    self.test = test

    self.data = {
      'id': id,
      'srcDataset__id': None,
      'tgtDataconnection__id': None,
      'tgtDbname': None,
      'tgtTablename': None,
      'extract_config': None,
      'srcDataset__dataconnection__dataconnection_type': 'impala',
      'tgtDataconnection__dataconnection_type': 'impala',
    }

class _MockUser:
  def __init__(self, username=None, id=None, test=None):
    self.username = username
    self.id = id
    self.test = test

class _MockProperties:
  def __init__(self, dc, conn_type = None):
    self.conn_type = conn_type
    self.dc = dc
    self.supports_dataextract_source = self.dc.test.supports_dataextract_source
    self.supports_partitions = self.dc.test.supportsPartitions(self)
    self.supports_dataextract_target = self.dc.test.supports_dataextract_target
    self.quote_alias = '`'
    self.quote = '"'
    self.quote_end = '"'
    self.quote_escape_character = "\\"
    self.supports_table_aliases = True
    self.alias_supports_spaces = True
    self.alias_supports_backticks = True
    self.supports_top = False
    self.sb_specify_alias = True
    self.supports_orderby = True
    self.supports_limit = True
    self.supports_limit_0 = True
    self.supports_groupby_colindex = True
    self.property_map = getattr(self.dc.test,'property_map', None)

class _MockDataConnection:
  def __init__(self, id, test):
    self.test = test
    self.id = id

class _MockDataConnectionObj:
  def __init__(self, id, test, dc_type="impala"):
    self.test = test
    self.id = id
    self.connection_type = dc_type

  def properties(self):
    return _MockProperties(self)

  def build(self, username=None, userid=None, query=None, dataset_request=None,
      sample_pct=None, comment=None, qcntl=None, profile=None):

      return BuildRequest(connection=self, username=username, userid=userid,
        query=query, dataset_request=dataset_request, sample_pct=sample_pct,
        comment=comment, qcntl=qcntl, profile=profile)

class _MockDataSet:
  def __init__(self, id, test):
    self.test = test
    self.id = id
    self.dataconnection = _MockDataConnection(self.id + 3000, test)
    self.dataconnection_obj = _MockDataConnectionObj(self.id + 3000, test)
    self.dataset_type = 'singletable'
    self.dataset_lvinfo = None
    self.dataset_info = None

class Test_BaseComponents(TestCase):
  def __init__(self, *args, **kwargs):
    super(Test_BaseComponents, self).__init__(*args, **kwargs)

  def test_dimTracker(self):
    testEntrypoints = extracts.rewriter.TEST_ENTRYPOINTS
    self.assertEquals(testEntrypoints['_dimTracker']('abc'), {
      'type': 'dim',
      'dim': 'abc',
    })

    self.assertEquals(testEntrypoints['_dimTracker']('[abc]'), {
      'type': 'dim',
      'dim': '[abc]',
    })

    self.assertEquals(testEntrypoints['_dimTracker']('([abc])'), {
      'type': 'dim',
      'dim': '[abc]',
    })

    self.assertEquals(testEntrypoints['_dimTracker']('(([abc]))'), {
      'type': 'dim',
      'dim': '[abc]',
    })

    self.assertEquals(testEntrypoints['_aggTracker']('sum', '1'), {
      'type': 'agg',
      'exprStr': 'sum(1)',
      'aggField': '1',
      'aggType': 'sum',
    })

    self.assertEquals(testEntrypoints['_aggTracker']('sum', '[abc]'), {
      'type': 'agg',
      'exprStr': 'sum([abc])',
      'aggField': '[abc]',
      'aggType': 'sum',
    })

    self.assertEquals(testEntrypoints['_aggTracker']('sum', '([abc])'), {
      'type': 'agg',
      'exprStr': 'sum(([abc]))',
      'aggField': '[abc]',
      'aggType': 'sum',
    })

class Test_FeatureFlipper(TestCase):
  def __init__(self, *args, **kwargs):
    super(Test_FeatureFlipper, self).__init__(*args, **kwargs)

  def test_settings(self):
    FeatureFlipper().set_flipper('site_enable_extract_rewrites', False, 'DB')
    self.assertFalse(extracts.misc.rewritesEnabled())

    FeatureFlipper().set_flipper('site_enable_extracts', False, 'DB')
    FeatureFlipper().set_flipper('site_enable_extract_rewrites', True, 'DB')
    self.assertFalse(extracts.misc.rewritesEnabled())

    FeatureFlipper().set_flipper('site_enable_extracts', True, 'DB')
    FeatureFlipper().set_flipper('site_enable_extract_rewrites', True, 'DB')
    self.assertTrue(extracts.misc.rewritesEnabled())

class Test_Rewrite_Helpers(TestCase):
  def __init__(self, *args, **kwargs):
    super(Test_Rewrite_Helpers, self).__init__(*args, **kwargs)

  def setUp(self):
    FeatureFlipper().set_flipper('site_enable_extracts', True, 'DB')
    FeatureFlipper().set_flipper('site_enable_extract_rewrites', True, 'DB')

  def test_unsupportedAgg(self):
    p = extracts.rewriter._AggExprProcessor.getAggProcessor('unsupported')
    self.assertTrue(p is None)

  def test_SumProcessor(self):
    sumProcessor = extracts.rewriter._AggExprProcessor.getAggProcessor('sum')
    atLeastOnes = []
    sumProcessor.processRewriteNeeds('sum([abc])', atLeastOnes)
    self.assertEquals(atLeastOnes, [[{
      'type': 'agg',
      'aggType': 'sum',
      'aggField': '[abc]',
      'exprStr': 'sum([abc])'
    }]])

    expr = Expression('sum([abc])')
    rewrites = _MockRewriteTarget(self, ['sum([abc])'], ['sum([sum_abc])'])
    rewrittenExpr = sumProcessor.processExpression(rewrites, expr)
    self.assertEquals(rewrittenExpr.fullExpressionStr(),
      'sum([sum_abc])')

  def test_AvgProcessor(self):
    avgProcessor = extracts.rewriter._AggExprProcessor.getAggProcessor('avg')
    atLeastOnes = []
    avgProcessor.processRewriteNeeds('avg([abc])', atLeastOnes)
    self.assertEquals(atLeastOnes, [
      [
        {
          'type': 'agg',
          'aggType': 'sum',
          'aggField': '[abc]',
          'exprStr': 'sum([abc])'
        }
      ],
      [
        {
          'type': 'agg',
          'aggType': 'sum',
          'aggField': '1',
          'exprStr': 'sum(1)'
        }
      ]
    ])

    expr = Expression('sum([abc])')
    rewrites = _MockRewriteTarget(self, ['sum([abc])', 'sum(1)'],
        ['sum([sum_abc])', 'sum([record_count])'])
    rewrittenExpr = avgProcessor.processExpression(rewrites, expr)
    self.assertEquals(rewrittenExpr.fullExpressionStr(),
      'sum([sum_abc]) / sum([record_count])')

  def test_MinProcessor(self):
    minProcessor = extracts.rewriter._AggExprProcessor.getAggProcessor('min')
    atLeastOnes = []
    minProcessor.processRewriteNeeds('min([abc])', atLeastOnes)
    self.assertEquals(atLeastOnes, [
      [
        {
          'type': 'dim',
          'dim': '[abc]',
        },
        {
          'type': 'agg',
          'aggType': 'min',
          'aggField': '[abc]',
          'exprStr': 'min([abc])'
        }
      ]
    ])

    expr = Expression('min([abc])')
    rewrites = _MockRewriteTarget(self, ['min([abc])', 'abc'],
        [None, 'new_abc'])
    rewrittenExpr = minProcessor.processExpression(rewrites, expr)
    # self.assertEquals(rewrittenExpr.fullExpressionStr(), 'min([new_abc])')
    self.assertEquals(rewrittenExpr.fullExpressionStr(), 'min([abc])') # see VIZ-1965

    expr = Expression('min([abc])')
    rewrites = _MockRewriteTarget(self, ['min([abc])'], ['min([min_abc])'])
    rewrittenExpr = minProcessor.processExpression(rewrites, expr)
    self.assertEquals(rewrittenExpr.fullExpressionStr(), 'min([min_abc])')

  def test_MaxProcessor(self):
    maxProcessor = extracts.rewriter._AggExprProcessor.getAggProcessor('max')
    atLeastOnes = []
    maxProcessor.processRewriteNeeds('max([abc])', atLeastOnes)
    self.assertEquals(atLeastOnes, [
      [
        {
          'type': 'dim',
          'dim': '[abc]',
        },
        {
          'type': 'agg',
          'aggType': 'max',
          'aggField': '[abc]',
          'exprStr': 'max([abc])'
        }
      ]
    ])

    expr = Expression('max([abc])')
    rewrites = _MockRewriteTarget(self, ['max([abc])', 'abc'],
        [None, 'new_abc'])
    rewrittenExpr = maxProcessor.processExpression(rewrites, expr)
    # self.assertEquals(rewrittenExpr.fullExpressionStr(), 'max([new_abc])')
    self.assertEquals(rewrittenExpr.fullExpressionStr(), 'max([abc])') # see VIZ-1965

    expr = Expression('max([abc])')
    rewrites = _MockRewriteTarget(self, ['max([abc])'], ['max([max_abc])'])
    rewrittenExpr = maxProcessor.processExpression(rewrites, expr)
    self.assertEquals(rewrittenExpr.fullExpressionStr(), 'max([max_abc])')

  def test_NdvProcessor(self):
    ndvProcessor = extracts.rewriter._AggExprProcessor.getAggProcessor('ndv')
    atLeastOnes = []
    ndvProcessor.processRewriteNeeds('ndv([abc])', atLeastOnes)
    self.assertEquals(atLeastOnes, [
      [
        {
          'type': 'dim',
          'dim': '[abc]',
        }
      ]
    ])

    expr = Expression('ndv([abc])')
    rewrites = _MockRewriteTarget(self, ['abc'], ['new_abc'])
    rewrittenExpr = ndvProcessor.processExpression(rewrites, expr)
    # self.assertEquals(rewrittenExpr.fullExpressionStr(), 'ndv([new_abc])')
    self.assertEquals(rewrittenExpr.fullExpressionStr(), 'ndv([abc])') # see VIZ-1965

  def test_CountProcessor(self):
    cntProcessor = extracts.rewriter._AggExprProcessor.getAggProcessor('count')

    atLeastOnes = []
    cntProcessor.processRewriteNeeds('count(distinct ([abc]))', atLeastOnes)
    self.assertEquals(atLeastOnes, [
      [
        {
          'type': 'dim',
          'dim': '[abc]',
        }
      ]
    ])

    atLeastOnes = []
    cntProcessor.processRewriteNeeds('count(distinct [abc])', atLeastOnes)
    self.assertEquals(atLeastOnes, [
      [
        {
          'type': 'dim',
          'dim': '[abc]',
        }
      ]
    ])

    exprStr = 'count(distinct ([abc]))'
    expr = Expression(exprStr)
    self.assertEqual(expr.simpleFn(), 'count')
    self.assertTrue(expr.isCountDistinct())
    self.assertEqual(expr.simpleFnArg(), 'distinct ([abc])')

    expr = Expression('count(distinct [abc])')
    self.assertEqual(expr.simpleFn(), 'count')
    self.assertTrue(expr.isCountDistinct())
    self.assertEqual(expr.simpleFnArg(), 'distinct [abc]')
    rewrites = _MockRewriteTarget(self, ['abc'], ['new_abc'])
    rewrittenExpr = cntProcessor.processExpression(rewrites, expr)
    # self.assertEquals(rewrittenExpr.fullExpressionStr(), 'count(distinct [new_abc])')
    self.assertEquals(rewrittenExpr.fullExpressionStr(), 'count(distinct [abc])') # VIZ-1965

class Test_Rewrites(TestCase):
  def __init__(self, *args, **kwargs):
    super(Test_Rewrites, self).__init__(*args, **kwargs)
    self.setup_props()

  def setup_props(self):
    self.supports_dataextract_source = True
    self.supports_partitions = True
    self.supports_partitions_by_id = {}
    self.supports_dataextract_target = True

  def setUp(self):
    extracts.rewriter.resetExtractRewriteController()

    FeatureFlipper().set_flipper('site_enable_extracts', True, 'DB')
    FeatureFlipper().set_flipper('site_enable_extract_rewrites', True, 'DB')

    self.setup_props()

    self.dsreq = None
    self.originalQuery = None
    self.rewrittenQuery = None

    self.datasetConfig = [
      {
        'tablename': 'db1.src_table1',
        'columns': [
          {
            'name': 'channel',
            'alias': 'channel',
            'isdim': 'true'
          },
          {
            'name': 'channel_shortname',
            'alias': 'channel_shortname',
            'isdim': 'true'
          },
          {
            'name': 'channel_fullname',
            'alias': 'channel_fullname',
            'isdim': 'true'
          },
          {
            'name': 'protagonist',
            'alias': 'protagonist',
            'isdim': 'true'
          },
          {
            'name': 'program',
            'alias': 'program',
            'isdim': 'true'
          },
          {
            'name': 'datestring',
            'alias': 'datestring',
            'isdim': 'true'
          },
          {
            'name': 'price',
            'type': 'DOUBLE',
            'isdim': False,
            'alias': 'price',
            'expr': '',
            'hidden': False,
            'detaildata': True,
            'ispartition': False,
            'useroverrodeisdim': False,
            'dflagg': 'avg'
          },
          {
            'name': 'revenue',
            'type': 'DOUBLE',
            'isdim': False,
            'alias': 'revenue',
            'expr': '',
            'dflagg': 'sum'
          },
          {
            'name': 'uid',
            'type': 'DOUBLE',
            'isdim': False,
            'alias': 'uid',
            'expr': '',
            'dflagg': 'min'
          },
          {
            'name': 'sid',
            'type': 'DOUBLE',
            'isdim': False,
            'alias': 'sid',
            'expr': '',
            'dflagg': 'max'
          }
        ],
      },
    ]

    self.extractConfig = [
      {
        'srcDataset__id': 4001,
        'tgtDataconnection__id': 330,
        'tgtDbname': 'extractdb',
        'tgtTablename': 'extracttable_dim',
        'extract_config': json.dumps({
          'source': {
            'dimData': [
              {
                'expr': '[channel_fullname]',
                'alias': 'channel_fullname'
              },
              {
                'expr': '[channel_shortname]',
                'alias': 'channel_shortname'
              }
            ],
          },
          'target': {
            'partitionData': [
              {
                'expr': '[datestring]',
                'alias': 'datestring'
              }
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        })
      },
      {
        'srcDataset__id': 4001,
        'tgtDataconnection__id': 330,
        'tgtDbname': 'extractdb',
        'tgtTablename': 'extracttable',
        'extract_config': json.dumps({
          'source': {
            'dimData': [
              {
                'expr': '[channel]',
                'alias': 'channel'
              }
            ],
            'aggData': [
              {
                'expr': 'sum(1)',
                'alias': 'record_count'
              },
              {
                'expr': 'sum([price])',
                'alias': 'price'
              },
              {
                'expr': 'avg([price])',
                'alias': 'should_not_use_price'
              },
              {
                'expr': 'sum([revenue])',
                'alias': 'revenue'
              },
              {
                'expr': 'min([uid])',
                'alias': 'uid'
              },
              {
                'expr': 'max([sid])',
                'alias': 'sid'
              },
            ]
          },
          'target': {
            'partitionData': [
              {
                'expr': '[datestring]',
                'alias': 'datestring'
              }
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        })
      }
    ]

  def supportsPartitions(self, properties):
    return True

  def assertStringsEqualNoWhitespace(self, str1, str2):
    msg = '\nactual:\n' + str1 + '\ngolden:\n' + str2
    str1 = ''.join(str1.split())  # remove all whitespace characters
    str2 = ''.join(str2.split())  # remove all whitespace characters
    self.assertEqual(str1, str2, msg=msg)

  def mock_fetchExtracts(self, datasetIds):
    extracts = []

    startingExtractId = 12

    for extractConfig in self.extractConfig:
      e = _MockExtract(startingExtractId, self)
      startingExtractId += 1
      e.data.update(extractConfig)
      extracts.append(e)

    return [e.data for e in extracts]

  def create_build_request(self, dsreq, ds:DataSet=None, user=None):
    if not ds:
      ds = _MockDataSet(4001, self)
      ds.dataset_info = json.dumps(self.datasetConfig)
      dc = ds.dataconnection_obj
    else:
      dc = _MockDataConnectionObj(ds.dataconnection.id, self)
      dc.connection_type = ds.dataconnection.dataconnection_type

    if not user:
      user = _MockUser('admin', 5001, self)

    query = ''
    qcntl = QueryControl()
    profile = ArcProf('testrewrite', username=user.username)

    dataset_request = DatasetRequest(dsreq, django_ds=ds)
    build_request = dc.build(user.username, user.id, query,
                             dataset_request=dataset_request,
                             qcntl=qcntl,
                             profile=profile)
    if PRINT_BUILD_REQUEST:
      print('build_request.request: '+build_request.query)
    return build_request

  def runtest(self, dsreq, originalQuery, rewrittenQuery, ds:DataSet=None, user=None):
    if not user:
      user = _MockUser('admin', 5001, self)

    build_request = self.create_build_request(dsreq, ds=ds, user=user)
    self.assertStringsEqualNoWhitespace(build_request.query, originalQuery)

    mds = build_request.dataset_request.dataset
    extracts.rewriter.handleRewrites(100, dsreq, user=user, ds=mds, django_ds=None, conn=None)

    build_request = self.create_build_request(dsreq, ds=ds, user=user)
    self.assertStringsEqualNoWhitespace(build_request.query, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_enabledisable(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    testEntrypoints = extracts.rewriter.TEST_ENTRYPOINTS
    getExtractRewriteController = testEntrypoints['_getExtractRewriteController']
    rewriteController = getExtractRewriteController()

    data = rewriteController.getExtracts()
    self.assertEquals(len(data), 2)
    self.assertEquals([x for x in data.keys()], [12, 13])
    self.assertEquals(data[12].fqTableName(), 'extractdb.extracttable_dim')
    self.assertEquals(data[13].fqTableName(), 'extractdb.extracttable')

    config = json.loads(self.extractConfig[0]['extract_config'])
    extracts.rewriter._updateRewriteStatusPropertyConfig(config, 'disabled_until_refreshed')
    self.extractConfig[0]['extract_config'] = json.dumps(config)
    extracts.rewriter.resetExtractRewriteController()

    rewriteController = getExtractRewriteController()
    data = rewriteController.getExtracts()
    self.assertEquals(len(data), 1)
    self.assertEquals([x for x in data.keys()], [13])
    self.assertEquals(data[13].fqTableName(), 'extractdb.extracttable')

    config = json.loads(self.extractConfig[0]['extract_config'])
    extracts.rewriter._updateRewriteStatusPropertyConfig(config, 'disabled')
    self.extractConfig[0]['extract_config'] = json.dumps(config)
    extracts.rewriter.resetExtractRewriteController()

    rewriteController = getExtractRewriteController()
    data = rewriteController.getExtracts()
    self.assertEquals(len(data), 1)
    self.assertEquals([x for x in data.keys()], [13])
    self.assertEquals(data[13].fqTableName(), 'extractdb.extracttable')

    config = json.loads(self.extractConfig[0]['extract_config'])
    extracts.rewriter._updateRewriteStatusPropertyConfig(config, 'enabled')
    self.extractConfig[0]['extract_config'] = json.dumps(config)
    extracts.rewriter.resetExtractRewriteController()

    rewriteController = getExtractRewriteController()
    data = rewriteController.getExtracts()
    self.assertEquals(len(data), 2)
    self.assertEquals([x for x in data.keys()], [12, 13])
    self.assertEquals(data[12].fqTableName(), 'extractdb.extracttable_dim')
    self.assertEquals(data[13].fqTableName(), 'extractdb.extracttable')

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_noaggs_nomatch(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addDimension("[datestring] as 'datestring'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, TA_0."datestring" as `datestring`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, originalQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_noaggs_match(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel_fullname] as 'channel_fullname'") \
      .addDimension("[datestring] as 'datestring'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel_fullname" as `channel_fullname`, TA_0."datestring" as `datestring`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel_fullname" as `channel_fullname`, TA_0."datestring" as `datestring`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_filter_nomatch(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addDimension("[datestring] as 'datestring'") \
      .addFilter("[protagonist] = 'Tom Cruise'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, TA_0."datestring" as `datestring`
        FROM "db1"."src_table1"  TA_0
        WHERE (TA_0."protagonist" = 'Tom Cruise')
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, originalQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_filter_match(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel_fullname] as 'channel_fullname'") \
      .addDimension("[datestring] as 'datestring'") \
      .addFilter("upper([channel_shortname]) > 'A'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel_fullname" as `channel_fullname`, TA_0."datestring" as `datestring`
        FROM "db1"."src_table1"  TA_0
        WHERE (upper((TA_0."channel_shortname")) > 'A')
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel_fullname" as `channel_fullname`, TA_0."datestring" as `datestring`
        FROM "extractdb"."extracttable_dim"  TA_0
        WHERE (upper((TA_0."channel_shortname")) > 'A')
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_filter_agg_match(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter("[datestring] = '2012-02-02'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, sum(1) as `Record Count`
        FROM "db1"."src_table1"  TA_0
        WHERE (TA_0."datestring" = '2012-02-02')
        GROUP BY 1
        LIMIT 5000
    """

    rewrittenQuery = r"""
    SELECT TA_0."channel" as `channel`, sum((TA_0."record_count")) as `Record Count`
      FROM "extractdb"."extracttable"  TA_0
      WHERE (TA_0."datestring" = '2012-02-02')
      GROUP BY 1
      LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_noaggs_match_alias(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel_fullname] as 'c1'") \
      .addDimension("[datestring] as 'd2'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel_fullname" as `c1`, TA_0."datestring" as `d2`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel_fullname" as `c1`, TA_0."datestring" as `d2`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_nodims(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addAggregate("sum(1) as 'Record Count'") \
      .getdsreq()

    originalQuery = r"""
      SELECT sum(1) as `Record Count`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT sum((TA_0."record_count")) as `Record Count`
        FROM "extractdb"."extracttable"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  # Verify that, if disabled, rewrites don't work
  # This is the same test as test_nodims above, but with the feature flipper turned off,
  #     either with site_enable_extracts or site_enable_extract_rewrites disabled
  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_nodims_rewrites_disabled(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addAggregate("sum(1) as 'Record Count'") \
      .getdsreq()

    originalQuery = r"""
      SELECT sum(1) as `Record Count`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    FeatureFlipper().set_flipper('site_enable_extract_rewrites', False, 'DB')
    self.runtest(dsreq, originalQuery, originalQuery)

    FeatureFlipper().set_flipper('site_enable_extracts', False, 'DB')
    FeatureFlipper().set_flipper('site_enable_extract_rewrites', True, 'DB')
    self.runtest(dsreq, originalQuery, originalQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_sum1(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addDimension("[datestring] as 'datestring'") \
      .addAggregate("sum(1) as 'Record Count'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, TA_0."datestring" as `datestring`, sum(1) as `Record Count`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1, 2
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel" as `channel`, TA_0."datestring" as `datestring`, sum((TA_0."record_count")) as `Record Count`
        FROM "extractdb"."extracttable"  TA_0
        GROUP BY 1, 2
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_sum1_alias(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'c1'") \
      .addDimension("[datestring] as 'd2'") \
      .addAggregate("sum(1) as 'r3'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `c1`, TA_0."datestring" as `d2`, sum(1) as `r3`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1, 2
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel" as `c1`, TA_0."datestring" as `d2`, sum((TA_0."record_count")) as `r3`
        FROM "extractdb"."extracttable"  TA_0
        GROUP BY 1, 2
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_sumprice(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addDimension("[datestring] as 'datestring'") \
      .addAggregate("sum([price]) as 'sum_price'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, TA_0."datestring" as `datestring`, sum((TA_0."price")) as `sum_price`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1, 2
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel" as `channel`, TA_0."datestring" as `datestring`, sum((TA_0."price")) as `sum_price`
        FROM "extractdb"."extracttable"  TA_0
        GROUP BY 1, 2
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_avgprice(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addDimension("[datestring] as 'datestring'") \
      .addAggregate("avg([revenue]) as 'avg_revenue'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, TA_0."datestring" as `datestring`, avg((TA_0."revenue")) as `avg_revenue`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1, 2
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel" as `channel`, TA_0."datestring" as `datestring`, sum((TA_0."revenue")) / sum((TA_0."record_count")) as `avg_revenue`
        FROM "extractdb"."extracttable"  TA_0
        GROUP BY 1, 2
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_minmaxndv_dimensions(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addAggregate("max([datestring]) as 'max_datestring'") \
      .addAggregate("min([channel]) as 'min_channel'") \
      .addAggregate("ndv([datestring]) as 'ndv_datestring'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, max((TA_0."datestring")) as `max_datestring`, min((TA_0."channel")) as `min_channel`, ndv((TA_0."datestring")) as `ndv_datestring`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel" as `channel`, max((TA_0."datestring")) as `max_datestring`, min((TA_0."channel")) as `min_channel`, ndv((TA_0."datestring")) as `ndv_datestring`
        FROM "extractdb"."extracttable"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_min_aggregates(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    # Note that min(uid) is expected to rewrite, but max(uid) is not because of the min
    # aggregation being in self.extractConfig, but the max aggregation not being in there
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addAggregate("min([uid]) as 'min_uid'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, min((TA_0."uid")) as `min_uid`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel" as `channel`, min((TA_0."uid")) as `min_uid`
        FROM "extractdb"."extracttable"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addAggregate("max([uid]) as 'max_uid'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, max((TA_0."uid")) as `max_uid`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, originalQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_max_aggregates(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    # Note that max(sid) is expected to rewrite, but min(sid) is not
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addAggregate("max([sid]) as 'max_sid'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, max((TA_0."sid")) as `max_sid`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel" as `channel`, max((TA_0."sid")) as `max_sid`
        FROM "extractdb"."extracttable"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addAggregate("min([sid]) as 'min_sid'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, min((TA_0."sid")) as `min_sid`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, originalQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_countdistinct(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    # Note that max(sid) is expected to rewrite, but min(sid) is not
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addAggregate("sum(1) as 'total'") \
      .addAggregate("count(distinct ([channel])) as 'count_channel'") \
      .getdsreq()

    originalQuery = r"""
      SELECT sum(1) as `total`, count(distinct ((TA_0."channel"))) as `count_channel`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT sum((TA_0."record_count")) as `total`, count(distinct ((TA_0."channel"))) as `count_channel`
        FROM "extractdb"."extracttable"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_multiplesums(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    # "(SUM((TA_0.`past_due`))/Sum(1))*100"

    # Note that max(sid) is expected to rewrite, but min(sid) is not
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel] as 'channel'") \
      .addAggregate("(SUM([revenue])/Sum(1))*100 as 'calc'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel" as `channel`, (SUM((TA_0."revenue"))/Sum(1))*100 as `calc`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel" as `channel`, (sum((TA_0."revenue"))/sum((TA_0."record_count")))*100 as `calc`
        FROM "extractdb"."extracttable"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)


  def mockExtracts(self, extracts):
    """
    Rewrite the test extractConfigs so that they match the needs of the current test.
    The first item of self.extractConfig is used as a template. It is copied and
    the attributes overwritten by the ones passed in the argument list
    """
    template = self.extractConfig[0]

    new_configs = []
    for extractForTest in extracts:
      extractConfig = copy.deepcopy(template)
      for prop in extractForTest.keys():
        if prop == 'extract_config':
          extractConfig[prop] = json.dumps(extractForTest[prop])
        else:
          extractConfig[prop] = extractForTest[prop]
      new_configs.append(extractConfig)

    self.extractConfig = new_configs

  def mockDatasetColumns(self, columns):
    """
    Reconfigure the dataset used by the test to have the provideded columns.
    """
    self.datasetConfig[0]['columns'] = columns

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_rewrite_alias_with_space(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([{
      'extract_config':{
        'source':{
          'dimData':[
            {
              'expr': '[The Channel fullname]',
              'alias': 'The Channel fullname'
            }
          ]
        },
        'rewrite_properties': {'rewrite_status': 'enabled'}
      }
    }])

    self.mockDatasetColumns([
        {
        'name': 'channel_fullname',
        'alias': 'The Channel fullname',
        'isdim': 'true'
      },
    ])


    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[The Channel fullname] as 'Channel FQN'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel_fullname" as `Channel FQN`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."the_channel_fullname" as `Channel FQN`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_rewrite_alias_no_space(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([{
      'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[FancyName]',
                'alias': 'FancyName'
              },
            ],
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
    }])

    self.mockDatasetColumns([
        {
        'name': 'channel_fullname',
        'alias': 'FancyName',
        'isdim': 'true'
      },
    ])


    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[FancyName] as 'Fancy Channel Name'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel_fullname" as `Fancy Channel Name`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."fancyname" as `Fancy Channel Name`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)


  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_shorter_extract_wins(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'tgtTablename': 'extract_2_dims',
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[FancyName]',
                'alias': 'FancyName'
              },
              {
                'expr': '[channel_shortname]',
                'alias': 'channel_shortname'
              },
            ],
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
      {
        'tgtTablename': 'extract_1_dim',
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[FancyName]',
                'alias': 'FancyName'
              },
            ],
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
    ])

    self.mockDatasetColumns([
      {
        'name': 'channel_fullname',
        'alias': 'FancyName',
        'isdim': 'true'
      },
      {
        'name': 'channel_shortname',
        'alias': 'channel_shortname',
        'isdim': 'true'
      },
    ])


    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[FancyName] as 'Fancy Channel Name'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel_fullname" as `Fancy Channel Name`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."fancyname" as `Fancy Channel Name`
        FROM "extractdb"."extract_1_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)


  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_global_segments_detected(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'tgtTablename': 'extract_1_dim',
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[FancyName]',
                'alias': 'FancyName'
              },
            ],
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
      {
        'tgtTablename': 'extract_3_dim',
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[FancyName]',
                'alias': 'FancyName'
              },
              {
                'expr': '[Channel Shortname]',
                'alias': 'Channel Shortname'
              },
              {
                'expr': '[Subscription Price]',
                'alias': 'Subscription Price'
              },
            ],
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      }
    ])

    self.mockDatasetColumns([
      {
        'name': 'channel_fullname',
        'alias': 'FancyName',
        'isdim': 'true'
      },
      {
        'name': 'channel_shortname',
        'alias': 'Channel Shortname',
        'isdim': 'true'
      },
      {
        'name': 'price',
        'alias': 'Subscription Price',
        'isdim': 'false'
      },
    ])

    segment_data = {
      'group': '',
      'entities':[],
      'filters': [
          '[Channel Shortname] LIKE \'ABC%\'',
          '[Subscription Price] < 100',
        ]
    }
    segment = Segment(segment_data=json.dumps(segment_data))
    segment.save()

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[FancyName] as 'Fancy Channel Name'") \
      .addGlobalSegment(segment_id=segment.id) \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel_fullname" as `Fancy Channel Name`
        FROM "db1"."src_table1"  TA_0
        WHERE (((TA_0."channel_shortname") LIKE 'ABC%' AND (TA_0."price") < 100))
        LIMIT 5000
    """

    # rewriting should pick up the second extract:
    rewrittenQuery = r"""
      SELECT TA_0."fancyname" as `Fancy Channel Name`
        FROM "extractdb"."extract_3_dim"  TA_0
        WHERE (((TA_0."channel_shortname") LIKE 'ABC%' AND (TA_0."subscription_price") < 100))
        LIMIT 5000
    """
    self.runtest(dsreq, originalQuery, rewrittenQuery)

  def make_dataconnection(self):
    return DataConnection.objects.create(
      dataconnection_name="test_dc", dataconnection_info="", dataconnection_type="impala"
    )

  def make_dataset(self, dc: DataConnection, tablename, columns=[]):
    table = {"tablename": tablename, "columns": columns}
    ds = DataSet.objects.create(
      dataset_name=tablename,
      dataconnection=dc,
      dataset_detail="default.grouptest",
      dataset_type="singletable",
      dataset_info=json.dumps([table]),
    )
    return ds

  def make_dataset_with_filter_association(self, table_name, columns, segment_filters):
    """
    This utility does not use mocks. It creates real entities (users, dataconnections, datasets)
    and uses rest endpoints to create a segment and a filter association.
    """

    common = test_sqlrun.Common()

    user_admin = test_sqlrun.Common.make_user("admin", is_superuser=True)
    user1 = test_sqlrun.Common.make_user("user1")

    client_admin = Client()
    client_admin.login(
      username=user_admin.username, password=user_admin.username
    )

    dc = self.make_dataconnection()
    ds = self.make_dataset(dc, table_name, columns=columns)

    segment = common.make_segment(
      client_admin, ds, "Test Segment", segment_filters
    )
    _ = common.make_fa(
      client_admin,
      ds,
      "Test Segment associations",
      segment["id"],
      [user1.id],
    )

    return ds, user1

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_filter_associations_detected(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dataset_columns = [
      {
        "name": "channel_fullname",
        "alias": "FancyName",
        "isdim": "true"
      },
      {
        "name": "channel_sname",
        "alias": "Channel Shortname",
        "isdim": "true"
      },
      {
        "name": "price",
        "alias": "Subscription Price",
        "isdim": "false"
      },
    ]

    segment_filters = [
      "[Channel Shortname] LIKE 'ABC%'",
      "[Subscription Price] < 100",
    ]

    ds, user1 = self.make_dataset_with_filter_association(
      table_name="db1.src_table1",
      columns=dataset_columns,
      segment_filters=segment_filters
    )

    self.mockExtracts([
      {
        'tgtTablename': 'extract_1_dim',
        'srcDataset__id': ds.id,
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[FancyName]',
                'alias': 'FancyName'
              },
            ],
          },
        },
        'rewrite_properties': {'rewrite_status': 'enabled'}
      },
      {
        'tgtTablename': 'extract_3_dim',
        'srcDataset__id': ds.id,
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[FancyName]',
                'alias': 'FancyName'
              },
              {
                'expr': '[Channel Shortname]',
                'alias': 'Channel Shortname'
              },
              {
                'expr': '[Subscription Price]',
                'alias': 'Subscription Price'
              },
            ],
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      }
    ])

    dsreq = (
      DsreqCreator()
      .datasetId(ds.id)
      .addDimension("[FancyName] as 'Fancy Channel Name'")
      .getdsreq()
    )

    originalQuery = """
      SELECT TA_0."channel_fullname" as `Fancy Channel Name`
        FROM "db1"."src_table1"  TA_0
        WHERE (((TA_0."channel_sname") LIKE 'ABC%' AND (TA_0."price") < 100))
        LIMIT 5000
    """

    # rewriting should use the second extract
    rewrittenQuery = r"""
      SELECT TA_0."fancyname" as `Fancy Channel Name`
        FROM "extractdb"."extract_3_dim"  TA_0
        WHERE (((TA_0."channel_shortname") LIKE 'ABC%' AND (TA_0."subscription_price") < 100))
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery, ds=ds, user=user1)


  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_filter_associations_do_not_interfere(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    dataset_columns = [
      {
        "name": "channel_fullname",
        "alias": "FancyName",
        "isdim": "true"
      },
      {
        "name": "channel_sname",
        "alias": "Channel Shortname",
        "isdim": "true"
      },
      {
        "name": "price",
        "alias": "Subscription Price",
        "isdim": "false"
      },
    ]

    segment_filters = [
      "[Channel Shortname] LIKE 'ABC%'",
      "[Subscription Price] < 100",
    ]

    ds, _ = self.make_dataset_with_filter_association(
      table_name="db1.src_table1",
      columns=dataset_columns,
      segment_filters=segment_filters
    )

    self.mockExtracts([
      {
        'tgtTablename': 'extract_1_dim',
        'srcDataset__id': ds.id,
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[FancyName]',
                'alias': 'FancyName'
              },
            ],
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
    ])

    dsreq = (
      DsreqCreator()
      .datasetId(ds.id)
      .addDimension("[FancyName] as 'Fancy Channel Name'")
      .getdsreq()
    )

    # run the query as a different user (who does not have the filter associated)
    user2 = test_sqlrun.Common.make_user('user2')

    originalQuery = """
      SELECT TA_0."channel_fullname" as `Fancy Channel Name`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = """
      SELECT TA_0."fancyname" as `Fancy Channel Name`
        FROM "extractdb"."extract_1_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery, ds=ds, user=user2)


  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_non_unique_aliases(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[channel_fullname]',
                'alias': 'channel_fullname'
              },
              {
                'expr': '[Channel Fullname]',
                'alias': 'Channel Fullname'
              },
              {
                'expr': '[Sum 1]',
                'alias': 'Sum 1'
              },
            ],
            'aggData': [
              {
                'expr': 'sum(1)',
                'alias': 'sum(1)'
              }
            ]
          },
          'target': {
            'partitionData': [
              {
                'expr': '[sum?1]',
                'alias': 'sum?1'
              }
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
    ])

    self.mockDatasetColumns([
      {
        'name': 'channel_fullname',
        'alias': 'channel_fullname',
        'isdim': 'true'
      },
      {
        'name': 'channel_fullname',
        'alias': 'Channel Fullname',
        'isdim': 'true'
      },
      {
        'name': 'channel_fullname',
        'alias': 'Sum 1',
        'isdim': 'true'
      },
    ])


    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[channel_fullname] as 'channel_fullname'") \
      .addDimension("[Channel Fullname] as 'Channel Fullname'") \
      .addDimension("[Sum 1] as 'Sum 1'") \
      .addAggregate("sum(1) as 'Record Count'") \
      .getdsreq()

    originalQuery = r"""
      SELECT TA_0."channel_fullname" as `channel_fullname`, TA_0."channel_fullname" as `Channel Fullname`, TA_0."channel_fullname" as `Sum 1`, sum(1) as `Record Count`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1, 2, 3
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."channel_fullname_1" as `channel_fullname`, TA_0."channel_fullname_2" as `Channel Fullname`, TA_0."sum_1_1" as `Sum 1`, sum((TA_0."sum_1_3")) as `Record Count`
        FROM "extractdb"."extracttable_dim"  TA_0
        GROUP BY 1, 2, 3
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_min_max_with_alias(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[Channel Fullname]',
                'alias': 'Channel Fullname'
              },
            ],
          },
          'target': {
            'partitionData': [
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
    ])

    self.mockDatasetColumns([
      {
        'name': 'physical_channel',
        'alias': 'Channel Fullname',
        'isdim': 'true'
      },
    ])


    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addAggregate("max([Channel Fullname])") \
      .addAggregate("min([Channel Fullname])") \
      .getdsreq()

    originalQuery = r"""
      SELECT max((TA_0."physical_channel")) as `max((TA_0."physical_channel"))`, min((TA_0."physical_channel")) as `min((TA_0."physical_channel"))`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT max((TA_0."channel_fullname")) as `max((TA_0."channel_fullname"))`, min((TA_0."channel_fullname")) as `min((TA_0."channel_fullname"))`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_ndv_with_alias(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[Channel Fullname]',
                'alias': 'Channel Fullname'
              },
            ],
          },
          'target': {
            'partitionData': [
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
    ])

    self.mockDatasetColumns([
      {
        'name': 'physical_channel',
        'alias': 'Channel Fullname',
        'isdim': 'true'
      },
    ])


    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addAggregate("ndv([Channel Fullname])") \
      .getdsreq()

    originalQuery = r"""
      SELECT ndv((TA_0."physical_channel")) as `ndv((TA_0."physical_channel"))`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT ndv((TA_0."channel_fullname")) as `ndv((TA_0."channel_fullname"))`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_count_with_alias(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[Channel Fullname]',
                'alias': 'Channel Fullname'
              },
            ],
          },
          'target': {
            'partitionData': [
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
    ])

    self.mockDatasetColumns([
      {
        'name': 'physical_channel',
        'alias': 'Channel Fullname',
        'isdim': 'true'
      },
    ])


    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addAggregate("count([Channel Fullname])") \
      .addAggregate("count( distinct [Channel Fullname])") \
      .getdsreq()

    originalQuery = r"""
      SELECT count((TA_0."physical_channel")) as `count((TA_0."physical_channel"))`, count( distinct (TA_0."physical_channel")) as `count( distinct (TA_0."physical_channel"))`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT count((TA_0."channel_fullname")) as `count((TA_0."channel_fullname"))`, count( distinct (TA_0."channel_fullname")) as `count( distinct (TA_0."channel_fullname"))`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_computed_field_expansion(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[Cereal Name]',
                'alias': 'Cereal Name'
              },
            ],
          },
          'target': {
            'partitionData': [
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
    ])

    self.mockDatasetColumns([
      {
        'name': 'cereal_name',
        'alias': 'Cereal Name',
        'isdim': 'true'
      },
      {
        'alias': 'Lowercase Cereal Name',
        'expr': 'lower([Cereal Name])',
        'isdim': 'true'
      },
      {
        'alias': 'Doubled Name',
        'expr': 'concat([Cereal Name], " : ", [Lowercase Cereal Name])',
        'isdim': 'true'
      },
    ])

    # expression in query - works
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("lower([Cereal Name]) as 'column 1'") \
      .getdsreq()

    originalQuery = r"""
      SELECT lower((TA_0."cereal_name")) as `column 1`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT lower((TA_0."cereal_name")) as `column 1`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

    # expression moved to calculated field - fails
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Lowercase Cereal Name] as 'column 2'") \
      .getdsreq()

    originalQuery = r"""
      SELECT lower((TA_0."cereal_name")) as `column 2`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT (lower((TA_0."cereal_name"))) as `column 2`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_nested_computed_field_expansion(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[Cereal Name]',
                'alias': 'Cereal Name'
              },
            ],
          },
          'target': {
            'partitionData': [
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
    ])

    self.mockDatasetColumns([
      {
        'name': 'cereal_name',
        'alias': 'Cereal Name',
        'isdim': 'true'
      },
      {
        'alias': 'Lowercase Cereal Name',
        'expr': 'lower([Cereal Name])',
        'isdim': 'true'
      },
      {
        'alias': 'Doubled Name',
        'expr': 'concat([Cereal Name], " : ", [Lowercase Cereal Name])',
        'isdim': 'true'
      },
    ])

    # calculated field uses other calculated field
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Doubled Name] as 'column 3'") \
      .getdsreq()

    originalQuery = r"""
      SELECT concat((TA_0."cereal_name"), " : ", (lower((TA_0."cereal_name")))) as `column 3`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT (concat((TA_0."cereal_name"), " : ", (lower((TA_0."cereal_name"))))) as `column 3`
        FROM "extractdb"."extracttable_dim"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_computed_field_expansion_dim_and_agg(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'tgtTablename': 'extracttable',
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[Cereal Name]',
                'alias': 'Cereal Name'
              },
            ],
            'aggData': [
              {
                'expr': 'max([revenue])',
                'alias': 'revenue'
              }
            ]
          },
          'target': {
            'partitionData': [
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
      {
        'tgtTablename': 'extracttable_dim',
        'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[Cereal Name]',
                'alias': 'Cereal Name'
              },
              {
                'expr': '[revenue]',
                'alias': 'revenue'
              },
            ],
          },
          'target': {
            'partitionData': [
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
    ])

    self.mockDatasetColumns([
      {
        'name': 'cereal_name',
        'alias': 'Cereal Name',
        'isdim': 'true'
      },
      {
        'alias': 'Lowercase Cereal Name',
        'expr': 'lower([Cereal Name])',
        'isdim': 'true'
      },
      {
        'alias': 'Doubled Name',
        'expr': 'concat([Cereal Name], " : ", [Lowercase Cereal Name])',
        'isdim': 'true'
      },
      {
        'name': 'revenue',
        'type': 'DOUBLE',
        'isdim': False,
        'alias': 'revenue',
        'expr': '',
        'dflagg': 'sum'
      }
    ])

    # computed field dimension and aggregate => should match aggregated extract over dimensions only extract
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Doubled Name] as 'column 3'") \
      .addAggregate("max([revenue]) as 'max_revenue'") \
      .getdsreq()

    originalQuery = r"""
      SELECT concat((TA_0."cereal_name"), " : ", (lower((TA_0."cereal_name")))) as `column 3`, max((TA_0."revenue")) as `max_revenue`
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT (concat((TA_0."cereal_name"), " : ", (lower((TA_0."cereal_name"))))) as `column 3`, max((TA_0."revenue")) as `max_revenue`
        FROM "extractdb"."extracttable"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_computed_field_over_expansion(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([
      {
        'extract_config': {
          'source': {
            'dimData': [
              {
                'expr': '[Cereal Name]',
                'alias': 'Cereal Name'
              },
            ],
          },
          'target': {
            'partitionData': [
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      },
      {
        'tgtTablename': 'extracttable_computed',
        'extract_config': {
          'source': {
            'dimData': [
              {
                'expr': '[Doubled Name]',
                'alias': 'Doubled Name'
              },
            ],
          },
          'target': {
            'partitionData': [
            ]
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
      }
    ])

    self.mockDatasetColumns([
      {
        'name': 'cereal_name',
        'alias': 'Cereal Name',
        'isdim': 'true'
      },
      {
        'alias': 'Lowercase Cereal Name',
        'expr': 'lower([Cereal Name])',
        'isdim': 'true'
      },
      {
        'alias': 'Doubled Name',
        'expr': 'concat([Cereal Name], " : ", [Lowercase Cereal Name])',
        'isdim': 'true'
      },
    ])

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Doubled Name] as 'column 3'") \
      .getdsreq()

    # prioritize extract with computed fields
    originalQuery = r"""
      SELECT concat((TA_0."cereal_name"), " : ", (lower((TA_0."cereal_name")))) as `column 3`
        FROM "db1"."src_table1"  TA_0
        LIMIT 5000
    """

    rewrittenQuery = r"""
      SELECT TA_0."doubled_name" as `column 3`
        FROM "extractdb"."extracttable_computed"  TA_0
        LIMIT 5000
    """

    self.runtest(dsreq, originalQuery, rewrittenQuery)

  @mock.patch('extracts.rewriter.fetchExtracts')
  def test_aggregates_on_constants(self, mock_fetchExtracts):
    mock_fetchExtracts.side_effect = self.mock_fetchExtracts

    self.mockExtracts([{
      'extract_config':{
          'source': {
            'dimData': [
              {
                'expr': '[FancyName]',
                'alias': 'FancyName'
              },
            ],
          },
          'rewrite_properties': {'rewrite_status': 'enabled'}
        }
    }])

    self.mockDatasetColumns([
        {
        'name': 'channel_fullname',
        'alias': 'FancyName',
        'isdim': 'true'
      },
    ])


    aggs=[
      "sum(1) as 'sum'",
      "count(1) as 'count'",
      "count(distinct (1)) as 'count(distinct (1))'",
      "min(1) as 'min'",
      "max(1) as 'max'",
      "avg(1) as 'avg'",
      # "group_concat(1) as 'group_concat'", # no AggregateProcessor, also not recognized as a function
    ]

    for agg in aggs:
      self._do_test_agg_(agg)

  def _do_test_agg_(self, agg: str):
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[FancyName] as 'Fancy Channel Name'") \
      .addAggregate(agg) \
      .getdsreq()

    agg_with_backquotes = agg.replace("'","`")
    originalQuery = f"""
      SELECT TA_0."channel_fullname" as `Fancy Channel Name`, {agg_with_backquotes}
        FROM "db1"."src_table1"  TA_0
        GROUP BY 1
        LIMIT 5000
    """

    rewrittenQuery = f"""
      SELECT TA_0."fancyname" as `Fancy Channel Name`, {agg_with_backquotes}
        FROM "extractdb"."extracttable_dim"  TA_0
        GROUP BY 1
        LIMIT 5000
    """
    self.runtest(dsreq, originalQuery, rewrittenQuery)
