# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from cli.plugins import extract
from extracts.full import ExtractManagerFull
from extracts.tests import tests_rewrites
from datasets.models import DataSet, Extract
from reports.views_settings import setting_update_single
from unittest import mock
from util.ff_flipper import FeatureFlipper
from vizlib.dsreq import DsreqCreator
from vizlib.tests import test_report_operations
from vizlib.tests.test_report_operations import make_user, create_client
from vizlib.tests.test_sqlrun import run_dslist
from vizlib import workq

import django.test
import extracts.misc
import json


"""
Integration-like tests for query rewriting. Requests are posted through the HTTP endpoint.
Some mocking is still necessary.
"""


class TestRewritesIntegration(django.test.TransactionTestCase):
  @classmethod
  def tearDownClass(cls):
    # recreate the test DB because with TransactionTestCase all tables get truncated after each test case:
    from django.db import connection
    connection._connections['default'].creation.create_test_db(
      autoclobber=True)

    # clear the cache as it can disturb other tests
    # (visual_maxrows gets cached)
    from django.core.cache import cache
    cache.clear()

  def setUp(self) -> None:
    super().setUp()

    self.asserter = tests_rewrites.Test_Rewrites()

    self.workq_off = mock.patch.object(workq._WorkQ, 'add_work')
    FeatureFlipper().set_flipper('site_enable_extracts', True, 'DB')
    FeatureFlipper().set_flipper('site_enable_extract_rewrites', True, 'DB')

    self.common_setup()

  def tearDown(self) -> None:
    if isinstance(workq._WorkQ.add_work, mock.Mock):
      self.workq_off.stop()
    super().tearDown()

  def common_setup(self):
    util = test_report_operations.TestDashboardOperations()
    self.util = util

    # crete admin
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    self.admin_client = admin_client

    # create impala connection and dataset
    dc_id_1 = self.create_dataconnection(admin_client)
    self.dc_id_1 = dc_id_1
    self.ds_id_1 = util.create_dataset(
      admin_client, dc_id_1, 'impala_dataset_1')

    # this is the extract target
    self.dc_id_2 = self.create_dataconnection(
      admin_client, name='Extract Target DC')

    columns = [
      {
        "name": "channel",
        "type": "STRING",
        "alias": "Channel Name"
      },
      {
        "name": "datestring",
        "type": "STRING",
        "alias": "Date"
      },
    ]
    self.update_dataset(self.ds_id_1, 'testdb.test_table_1', columns=columns)
    self.create_extract()

    # needed by dataengine. Since this is stored in the DB and the DB is reset between tests
    # it should not affect the other tests
    setting_update_single('visual_maxrows', 100, False)

  def update_dataset(self, ds_id_1: int, tablename: str, columns: list[dict]):
    ds = DataSet.objects.get(id=ds_id_1)

    ds.dataset_info = json.dumps(
      [{"tablename": tablename, "columns": columns}])
    ds.dataset_detail = tablename
    ds.save()

  def create_dataconnection(self, admin_client, name='test DC'):
    impala_payload = {
      'dataconnection_type': 'impyla',
      'dataconnection_info': json.dumps({
        "PARAMS": {
          "HOST": "10.0.0.211",
          "PORT": "31050",
        }
      }),
      'dataconnection_name': name,
      'dataconnection_password': 'abc'
    }
    return self.util.create_dataconnection_with_payload(admin_client, impala_payload)

  def create_extract(self):
    create_args = [
      "refresh_full",
      f"{self.ds_id_1}",
      f"{self.dc_id_2}",
      "extractsdb",
      "cdv_extract_table1",
      "[Channel Name]",
      "sum(1)",
    ]

    existing_extracts = set(Extract.objects.values_list('id', flat=True))

    with mock.patch.object(extracts.misc, 'getColumnTypes') as getColumnTypes_mock:
      getColumnTypes_mock.return_value = ['String', 'Integer']
      with mock.patch.object(ExtractManagerFull, 'runCreate'):
        extract.create(create_args)

    current_extracts = set(Extract.objects.values_list('id', flat=True))

    self.assertEqual(len(current_extracts - existing_extracts), 1)
    new_extract_id = next(iter(current_extracts - existing_extracts))
    extract.rewritesEnable([new_extract_id])

  def extract_query(self, ds_response: list[dict]):
    """
    Extract the query from a data request response
    """
    query = [line[1]
             for line in ds_response[0]['profile'] if line[0] == 'query'][0]
    return query

  def test_dsreq_without_dataset_object(self):

    dsreq_1 = DsreqCreator() \
        .datasetId(self.ds_id_1) \
        .addDimension("[Channel Name] as 'Channel'") \
        .addAggregate("sum(1) as 'Record Count'") \
        .getdsreq()

    self.workq_off.start()

    results = run_dslist(self.admin_client, dslist=[
                         dsreq_1], wait_for_result=False)
    query = self.extract_query(results)
    print(query)

    expected_query = """
    SELECT TA_0.`channel_name` as `Channel`, sum((TA_0.`sum_1`)) as `Record Count`
      FROM `extractsdb`.`cdv_extract_table1` TA_0
      GROUP BY 1
      LIMIT 100
    """
    self.asserter.assertStringsEqualNoWhitespace(expected_query, query)

  def test_dsreq_with_dataset_object(self):

    dsreq_1 = DsreqCreator() \
        .datasetId(self.ds_id_1) \
        .addDimension("[lowercase channel name] as 'Low Channel'") \
        .addAggregate("sum(1) as 'Record Count'") \
        .getdsreq()

    dsreq_1['dataset_object'] = json.dumps([{
      'tablename': 'does not matter',
      'columns': [
        {
          'alias': 'Channel Name',
          'name': 'channel'
        },
        {
          'alias': 'lowercase channel name',
          'expr': 'lower([Channel Name])'
        }
      ]
    }])

    self.workq_off.start()

    results = run_dslist(self.admin_client, dslist=[
                         dsreq_1], wait_for_result=False)
    query = self.extract_query(results)
    print(query)

    expected_query = """
    SELECT (lower((TA_0.`channel_name`))) as `Low Channel`, sum((TA_0.`sum_1`)) as `Record Count`
      FROM `extractsdb`.`cdv_extract_table1` TA_0
      GROUP BY 1
      LIMIT 100
    """
    self.asserter.assertStringsEqualNoWhitespace(expected_query, query)
