# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import logging

from django.http import HttpRequest, HttpResponse
from django.contrib.auth.decorators import login_required
from django.db.models import Min, Max

from vizlib.querylog import logAction, ComponentType, ActionType
from datasets.models import DataSet, DataConnection, Extract
from util import perm
import extracts.misc
import extracts.rewriter

LOGGER = logging.getLogger("arcweb")

@login_required
def extract(request, extract_id=None):
  """
  REST endpoint for extracts
  """
  return extract_internal(request, extract_id)

def extract_internal(request, extract_id=None):
  if extract_id:
    extract_id = int(extract_id)

  if request.method == 'GET':
    response = _extract_get(request, extract_id)

    if request.GET.get("reset_controller") == "true":
      extracts.rewriter.resetExtractRewriteController()

    return response

  extracts.misc.checkEnabled()

  if request.method == 'POST':
    try:
      return _extract_create(request)
    finally:
      extracts.rewriter.resetExtractRewriteController()

  if request.method == 'PUT':
    return _extract_update(request, extract_id)

  if request.method == 'DELETE':
    try:
      ret = _extract_delete(request, extract_id)
      return ret
    finally:
      extracts.rewriter.resetExtractRewriteController()

  return HttpResponse('Invalid request method', status=500)

def extractManagerFromExtract(e):
  config = json.loads(e.extract_config)
  runstate = json.loads(e.extract_runstate)

  from extracts import EXTRACT_TYPE_PARTITIONED, EXTRACT_TYPE_FULL
  from extracts.full import ExtractManagerFull
  from extracts.partitioned import ExtractManagerPartitioned

  mgr = None

  if e.extract_type == EXTRACT_TYPE_PARTITIONED:
    mgr = ExtractManagerPartitioned(e.id, e.srcDataset.id, e.tgtDataconnection.id, e.tgtDbname,
        e.tgtTablename, config, runstate)
  elif e.extract_type == EXTRACT_TYPE_FULL:
    mgr = ExtractManagerFull(e.id, e.srcDataset.id, e.tgtDataconnection.id, e.tgtDbname,
        e.tgtTablename, config, runstate)

  if not mgr:
    raise Exception(f"Invalid type {e.extract_type}")

  mgr.validate()
  return mgr

def _process_cols_data(dimData, aggData, partitionData, user, srcDatasetId):
  # Make sure we discard any additional json elements passed in
  dimData = [
    {
      'expr': x['expr'],
      'alias': x.get('alias', x['expr'])
    }
    for x in dimData
  ]

  aggData = [
    {
      'expr': x['expr'],
      'alias': x.get('alias', x['expr'])
    }
    for x in aggData
  ]

  partitionData = [
    {
      'expr': x['expr'],
      'alias': x.get('alias', x['expr'])
    }
    for x in partitionData
  ]

  # Get the column types of the columns
  dims = [x['expr'] for x in dimData]
  aggs = [x['expr'] for x in aggData]
  partitions = [x['expr'] for x in partitionData]
  if not dims and not aggs:
    raise Exception("Must specify dims or aggs")

  coltypes = _getColumnTypes(user, srcDatasetId, dims, aggs, partitions)

  dimColtypes = coltypes[:len(dims)]
  partitionColtypes = coltypes[len(dims):len(dims) + len(partitions)]
  aggColtypes = coltypes[len(dims) + len(partitions):]

  for i, d in enumerate(dimData):
    d['type'] = dimColtypes[i]

  for i, a in enumerate(aggData):
    a['type'] = aggColtypes[i]

  for i, p in enumerate(partitionData):
    p['type'] = partitionColtypes[i]

  return {'dimData': dimData, 'aggData': aggData, 'partitionData': partitionData}

def _generate_extract_config(de_data, user, srcDatasetId):
  processedColsData = _process_cols_data(de_data.get('dimData'),
                          de_data.get('aggData'),
                          de_data.get('partitionData', []),
                          user,
                          srcDatasetId
                        )
  dimData = processedColsData.get('dimData')
  aggData = processedColsData.get('aggData')
  partitionData = processedColsData.get('partitionData')

  return {
      'username': user.username,
      'source': {
        'dimData': dimData,
        'aggData': aggData,
      },
      'target': {
        'partitionData': partitionData
      },
      'notification_email': {
        'email_success': de_data.get('emailSuccess'),
        'email_fail': de_data.get('emailFail'),
        'to_addresses': de_data.get('emailTo', []),
        'email_subject': de_data.get('emailSubject'),
      },
      'rewrite_properties': {
        'rewrite_status': de_data.get('rewriteStatus', 'disabled_until_refreshed').lower()
      }
    }

def _extract_create(request):
  data = json.loads(request.body)
  extract_type = data.get('extract_type')
  srcDatasetId = data.get('srcDatasetId')
  tgtDataconnectionId = data.get('tgtDataconnectionId')
  tgtDbname = data.get('tgtDbname')
  tgtTablename = data.get('tgtTablename')
  createTargetTable = data.get('createTargetTable', "true")

  srcDataset = DataSet.objects.get(id=int(srcDatasetId))
  tgtDataconnection = DataConnection.objects.get(id=int(tgtDataconnectionId))

  _checkPermInternal(request.user, srcDataset, tgtDataconnection)

  extract_config = json.dumps(_generate_extract_config(data, request.user, srcDatasetId))

  extract_runstate = json.dumps({})

  e = Extract(srcDataset=srcDataset,
              tgtDataconnection=tgtDataconnection,
              tgtDbname=tgtDbname,
              tgtTablename=tgtTablename,
              extract_type=extract_type,
              extract_config=extract_config,
              extract_runstate=extract_runstate)
  e.create_user = request.user.username
  e.update_user = request.user.username

  mgr = extractManagerFromExtract(e)
  if createTargetTable == "true":
    mgr.runCreate()
  e.save()

  return _extract_get(request, e.id)

def _extract_get(request: HttpRequest, extract_id):
  # In the front-end, if extracts are disabled, we may still end up getting a few
  # get requests for extracts in the backend. In these cases, we just return an
  # empty list (instead of erroring out)
  if not extracts.misc.enabled():
    return HttpResponse(json.dumps([]), content_type="application/json")

  runstate = (request.GET.get("runstate", None) == "true")

  user = request.user

  dataset_id = request.GET.get("dataset_id", None)
  if dataset_id:
    dataset_id = int(dataset_id)
    perm.check_dataset_permission(user, dataset_id, 'ds_appview')

  projections = ['extract_type','tgtDbname', 'tgtTablename', 'extract_config', 'srcDataset__dataset_name','tgtDataconnection__dataconnection_name', 'srcDataset__version_group_id']
  if runstate:
    projections.append('extract_runstate')

  base_extract_query = Extract.objects\
      .select_related('srcDataset')\
      .select_related('tgtDataconnection')

  if not extract_id:
    if dataset_id:
      el = base_extract_query.filter(srcDataset__id=dataset_id)
    elif not user.is_superuser:
      dataset_ids = perm.user_dataset_ids(user)
      el = base_extract_query.filter(srcDataset_id__in=dataset_ids)
    else:
      el = base_extract_query
  else:
    el = base_extract_query.filter(id=extract_id)
    if not user.is_superuser:
      # we need extra fields to be able to check the permission
      projections.append('srcDataset__dataconnection_id')
      projections.append('srcDataset__create_user')

  el = el.only(*projections)

  extract_ids = list[int]()
  _extracts = list[dict]()
  for e in el:

    if extract_id and not user.is_superuser:
      # have to check the permission
      e.srcDataset.checkPerm(user, 'ds_appview')

    extract_ids.append(e.id)
    ej = {
      "id": e.id,
      "extract_type": e.extract_type,
      "srcDatasetId": e.srcDataset.id,
      "srcDatasetName": e.srcDataset.dataset_name,
      "tgtDataconnectionName": e.tgtDataconnection.dataconnection_name,
      "tgtDataconnectionId": e.tgtDataconnection.id,
      "tgtDbname": e.tgtDbname,
      "tgtTablename": e.tgtTablename,
      "extract_config": json.loads(e.extract_config),
    }

    if runstate:
      ej["extract_runstate"] = json.loads(e.extract_runstate)

    _extracts.append(ej)

  from jobs.models import JobSchedule
  latest_schedules = JobSchedule.objects\
    .filter(extract_id__in=extract_ids)\
    .values('extract_id')\
    .annotate(schedule_id=Max('id'))

  latest_repeating_schedules = JobSchedule.objects\
    .filter(extract_id__in=extract_ids)\
    .filter(schedule_group__isnull=False)\
    .values('extract_id')\
    .annotate(schedule_id=Min('id'))

  repeating_schedule_ids = []
  extract_schedule_info = {}

  for ls in latest_schedules:
    extract_schedule_info[ls['extract_id']] = {
      'scheduleId': ls['schedule_id']
    }

  for lrs in latest_repeating_schedules:
    repeating_schedule_ids.append(lrs['schedule_id'])

  repeating_schedules = JobSchedule.objects\
    .filter(id__in=repeating_schedule_ids)\
    .select_related('schedule_group')\
    .only('schedule_group__schedule_name', 'extract_id')

  for rs in repeating_schedules:
    extract_schedule_info[rs.extract_id] = {
      'scheduleId': rs.id,
      'scheduleGroupId': rs.schedule_group.id,
      'scheduleGroupName': rs.schedule_group.schedule_name,
    }

  for ej in _extracts:
    schedule_info = extract_schedule_info.get(ej['id'])
    if schedule_info:
      ej.update(schedule_info)

  return HttpResponse(json.dumps(_extracts), content_type="application/json")

# Permission checks for any operation that needs to manage or run an extract
# User must have manage permissions on the source dataset
#   and data extract permissions on the source and target datasets
def checkPerm(user, extract):
  _checkPermInternal(user, extract.srcDataset, extract.tgtDataconnection)

def _checkPermInternal(user, srcDataset, tgtDataconnection):
  srcDataset.checkPerm(user, 'ds_manage')
  srcDataset.dataconnection.checkPerm(user, 'dc_aviews')
  tgtDataconnection.checkPerm(user, 'dc_aviews')

def _areColsListsEq(cols1, cols2):
  return sorted([x['expr'] for x in cols1]) ==  sorted([x['expr'] for x in cols2])

def _auditExtract(extract, action, request):
  logAction(action=action, component_type=ComponentType.EXTRACT,
    component_id=extract.id, request=request,
    dataconnection_id=extract.tgtDataconnection_id, dataset_id=extract.srcDataset_id,
    dataconnection_type=extract.tgtDataconnection.dataconnection_type)

def _extract_update(request, extract_id):
  extract = Extract.objects.filter(id=extract_id).first()
  if not extract_id or not extract:
    return HttpResponse('Invalid extract id', status=500)

  checkPerm(request.user, extract)
  old_tgtDataconnectionId = extract.tgtDataconnection_id
  old_tgtDbname = extract.tgtDbname
  old_tgtTablename = extract.tgtTablename

  data = json.loads(request.body)
  tgtDataconnectionId = int(data.get('tgtDataconnectionId'))
  tgtDbname = data.get('tgtDbname')
  tgtTablename = data.get('tgtTablename')
  extract_type = data.get('extract_type')

  new_extract_config = _generate_extract_config(data, request.user, extract.srcDataset_id)
  is_extract_updated = False
  del_recreate_table = False

  if tgtDataconnectionId != extract.tgtDataconnection_id:
    tgtDataconnection = DataConnection.objects.get(id=int(tgtDataconnectionId))
    tgtDataconnection.checkPerm(request.user, 'dc_aviews')
    extract.tgtDataconnection_id = tgtDataconnectionId
    is_extract_updated = True
    del_recreate_table = True

  if tgtDbname != extract.tgtDbname:
    extract.tgtDbname = tgtDbname
    is_extract_updated = True
    del_recreate_table = True

  if tgtTablename != extract.tgtTablename:
    extract.tgtTablename = tgtTablename
    is_extract_updated = True
    del_recreate_table = True

  extract_config = json.loads(extract.extract_config)
  is_extract_config_updated = False
  if not _areColsListsEq(extract_config.get('source').get('dimData'),
            new_extract_config.get('source').get('dimData')):
    extract_config['source']['dimData'] = new_extract_config.get('source').get('dimData')
    is_extract_config_updated = True
    del_recreate_table = True

  if not _areColsListsEq(extract_config.get('source').get('aggData'),
            new_extract_config.get('source').get('aggData')):
    extract_config['source']['aggData'] = new_extract_config.get('source').get('aggData')
    is_extract_config_updated = True
    del_recreate_table = True

  if not _areColsListsEq(extract_config.get('target').get('partitionData'),
            new_extract_config.get('target').get('partitionData')):
    extract_config['target']['partitionData'] = new_extract_config.get('target').get('partitionData')
    extract.extract_type = extract_type
    is_extract_config_updated = True
    del_recreate_table = True

  if extract_config.get('notification_email', None) != new_extract_config.get('notification_email'):
    extract_config['notification_email'] = new_extract_config.get('notification_email')
    is_extract_config_updated = True

  is_rewrite_props_updated = False
  if extract_config.get('rewrite_properties', None) != new_extract_config.get('rewrite_properties'):
    extract_config['rewrite_properties'] = new_extract_config.get('rewrite_properties')
    is_extract_config_updated = True
    is_rewrite_props_updated = True

  if is_extract_config_updated:
    extract.extract_config = json.dumps(extract_config)
    is_extract_updated = True

  if is_extract_updated:
    extract.extract_runstate = json.dumps({})
    extract.update_user = request.user.username

    if del_recreate_table:
      # drop old target table
      old_tgtDataconnection = extracts.misc.fetchDataConn(old_tgtDataconnectionId)
      old_fqtablename = old_tgtDataconnection.quote_tablename(f'{old_tgtDbname}.{old_tgtTablename}')
      drop_old_table_cmd = old_tgtDataconnection.drop_table_cmd(old_fqtablename)
      old_tgtDataconnection.log_sync_runquery(drop_old_table_cmd, username=request.user.username)

      # create new target table
      mgr = extractManagerFromExtract(extract)
      mgr.runCreate()

    extract.save()
    _auditExtract(extract, ActionType.UPDATE, request)

    if is_rewrite_props_updated:
      extracts.rewriter.resetExtractRewriteController()

  return HttpResponse(json.dumps({'extract_id': extract_id}), content_type="application/json")

def _extract_delete(request, extract_id):
  e = Extract.objects.get(id=extract_id)

  checkPerm(request.user, e)

  e.delete()

  if request.GET.get('deleteTable','false') == 'true':
    _drop_target_table(e, username=request.user.username)

  return HttpResponse('ok', content_type="text/plain")

def _drop_target_table(e: Extract, username: str):
  tgtDataconnectionId = e.tgtDataconnection_id
  tgtDbname = e.tgtDbname
  tgtTablename = e.tgtTablename

  tgtDataconnection = extracts.misc.fetchDataConn(tgtDataconnectionId)
  fqtablename = tgtDataconnection.quote_tablename(f'{tgtDbname}.{tgtTablename}')
  drop_table_cmd = tgtDataconnection.drop_table_cmd(fqtablename)

  tgtDataconnection.log_sync_runquery(drop_table_cmd, username=username)

def _getColumnTypes(user, datasetId, dims=[], aggs=[], partitions=[]):
  ds = extracts.misc.fetchDataset(datasetId)
  srcConn = extracts.misc.fetchDataConn(ds.dataconnection.id)
  srcProperties = srcConn.properties()

  alldims = []
  allcols = []
  alldims.extend(dims)
  allcols.extend(dims)
  alldims.extend(partitions)
  allcols.extend(partitions)

  allcols.extend(aggs)

  if not alldims and not aggs:
    raise Exception("Must specify dimensions or aggregates")

  (srcQuery, _) = extracts.misc.datasetToSQL(ds, user=user,
      dims=alldims, aggs=aggs)
  coltypes = extracts.misc.getColumnTypes(conn=srcConn,
      connProperties=srcProperties, query=srcQuery, username=user.username)

  return coltypes

class ExtractTracker(object):
  def __init__(self):
    self._extracts = Extract.objects.all()
    self._extractsByTable = {}
    self._extractsByDataset = {}
    for e in self._extracts:
      fqTablename = f"{e.tgtDbname}.{e.tgtTablename}"
      self._extractsByTable[(e.tgtDataconnection_id, fqTablename)] = e
      dsid = e.srcDataset_id
      if not self._extractsByDataset.get(dsid):
        self._extractsByDataset[dsid] = []
      self._extractsByDataset[dsid] = self._extractsByDataset[dsid].append(e)

  def isExtractSource(self, ds):
    return ds.id in self._extractsByDataset

  def isExtractTarget(self, ds):
    tl = json.loads(ds.dataset_tablenames)
    for tablename in tl:
      if (ds.dataconnection.id, tablename) in self._extractsByTable:
        return True

    return False

