#!python -tt
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import os
import sys
import glob
import re
import sqlite3

ADMIN_USERNAME = os.environ.get("VIZ_ADMIN_USERNAME", "vizapps_admin")
ADMIN_PASSWORD = os.environ.get("VIZ_ADMIN_PASSWORD", "vizapps_admin")

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

# Setup the environment before django imports
if 'DJANGO_SETTINGS_MODULE' not in os.environ:
  os.environ['DJANGO_SETTINGS_MODULE'] = 'arcwebbase.settings_instant'

from django.conf import settings
from django.core.management import execute_from_command_line
from django.contrib.auth.models import User
from django.http import HttpRequest

import util.migrations
import util.app
import reports.views_settings
import json

from reports.models import AppGroup, ReportTemplate, Report
from datasets.models import DataConnection, DataSet
from migration.migrate import DashboardSync
from reports.dash_template import create_template
from migration.views import import_file
from django.test import RequestFactory
from reports.views_appgroups import _user_appgroup_update, single_appgroup_app_update, appgroup_starred
try:
  from collections import OrderedDict
except:
  from ordereddict import OrderedDict



def db_exists():
  return os.path.exists(settings.DATABASES['default']['NAME']) and os.stat(settings.DATABASES['default']['NAME']).st_size != 0

def checks():
  if db_exists():
    print('DB already exists.')
    execute_from_command_line(["manage.py", "migrate", "--fake-initial"])
    # Create sample apps
    create_sample_apps(False, True)
    create_dashboard_templates()

    # Create missing tables in samples database if upgrading.
    # Note that we do not touch existing tables in the database
    insert_tables_sqlite()
    # Enable this setting even if DB already exists since we want to add this
    # setting for existing installs too
    reports.views_settings.setting_update_single('site_enable_visual_transitions',
      'true', False)
    return False
  return True

def create_dashboard_templates():

  user = User.objects.get(username=ADMIN_USERNAME)
  for filename in glob.iglob('%s/../samples/dashboard_templates/*' % THISDIR):
    print(('Creating dashboard template using %s' % os.path.basename(filename)))
    with open(filename) as f:
      templates = json.loads(f.read())

    print(('Creating dashboard template form %s file' % os.path.basename(filename)))
    for template in templates:
      # create report template
      data = json.loads(template['fields']['data'])

      # setting the dataset uuid to something that exists, so the dataset does not
      # get created. Note that this dataset ID will be switched upon creating a template
      data['datasets'][0]['fields']['uuid'] = DataSet.objects.all()[0].uuid

      r = create_template(data, user, deep_clone=True)
      report = Report.objects.get(id=r[DashboardSync.name][0]['id'])

      # create template record for the above template
      t = ReportTemplate()
      for field, value in template['fields'].items():
        if field == 'report':
          setattr(t, field, report)
        else:
          setattr(t, field, value)
      t.save()

# Helper function to check if dataset requires upgrade
# Checks if latest dataset has the same settings as
# the incoming dataset.
def _upgrade_dataset_check(ds_dict, data):
  version_tag_regex = r'-v\d+$'
  import_ds_list = []
  upgrade_needed = False
  # For each dataset in the app group, find all versions of that dataset
  # Get the latest version and then check against its dataset_setting
  # If different, add a new dataset version.
  for dataset_model in data['datasets']:
    ds_name = dataset_model['fields']['dataset_name']
    base_name = re.sub(version_tag_regex, '', ds_name)
    if base_name in ds_dict:
      # Find latest version
      latest_ver = ds_dict[ds_name]['latest_version']
      latest_settings = ds_dict[ds_name]['dataset_settings']
      m = re.search(r'-v(\d+)$', latest_ver)
      try:
        if m:
          new_vnum = int(m.group(1)) + 1
        else:
          new_vnum = 2
      except ValueError:
        # Use length of array instead
        new_vnum = len(ds_dict[ds_name]) + 1

      # Need to create new dataset object
      dataset_model['fields']['dataset_name'] += '-v' + str(new_vnum)
      if latest_settings != dataset_model['fields']['dataset_settings']:
        print('Upgrading %s dataset settings' % latest_ver)
        upgrade_needed = True

      import_ds_list.append(dataset_model)
    else:
      # Sample dataset does not exist, add it here
      import_ds_list.append(dataset_model)
      upgrade_needed = True

  if len(import_ds_list) > 0 and upgrade_needed:
    return import_ds_list
  else:
    return None

def create_sample_apps(skip_thumbnails, upgrade_datasets):
  def create_appgroup(gnames):
    # Create parent appgroup
    a = AppGroup.objects.filter(group_type='samples').first()
    if not a:
      a = AppGroup(
        group_name='Sample Dashboards',
        group_type='samples',
        create_user=ADMIN_USERNAME,
        update_user=ADMIN_USERNAME,
        group_options="{\"user_settings\":{\"menu_on_left\":\"false\",\"navigate_to_apps_page\":\"false\",\"container_css_class\":\"\",\"css_styles\":\"[]\",\"inline_css\":\"\"}}")
      a.save()
    samples_parent = a
    gnames.remove('Sample Dashboards')

    # create subgroups
    for g in gnames:
      if not AppGroup.objects.filter(group_name=g).exists():
        a = AppGroup(
            group_parent=samples_parent,
            group_name=g,
            group_type='default',
            create_user=ADMIN_USERNAME,
            update_user=ADMIN_USERNAME)
        a.save()

  def get_dataset_dict():
    dataconnection_obj = DataConnection.objects.get(dataconnection_name='samples')
    dataconnection_id = dataconnection_obj.id
    filtered_ds = DataSet.objects.filter(dataconnection_id=dataconnection_id)
    ds_list = {}
    version_tag_regex = '-v\d+$'

    # For each dataset in the data connection, add to a dictionary of base names
    # These base names do not include a version tag (eg 'Inspector View-v3' base name: 'Inspector View')
    # Store the dataset_settings and info of latest version so it can be compared if changed
    for ds in filtered_ds:
      base_name = re.sub(version_tag_regex, '', ds.dataset_name)
      if base_name in ds_list:
        if ds.dataset_name > ds_list[base_name]['latest_version']:
          ds_list[base_name] = {
            'latest_version': ds.dataset_name,
            'dataset_settings': ds.dataset_settings,
          }
      else:
        ds_list[base_name] = {
          'latest_version': ds.dataset_name,
          'dataset_settings': ds.dataset_settings,
        }

    return ds_list

  if skip_thumbnails:
    generate_thumbnails = True
  else:
    generate_thumbnails = False

  user = User.objects.get(username=ADMIN_USERNAME)
  request = HttpRequest()
  request.user = user

  # create app group and subgroups in the same order of _MEMBERSHIP_METADATA.json
  with open('%s/../samples/dashboards/_MEMBERSHIP_METADATA.json' % THISDIR, 'r') as f:
    app_group_mapping = json.loads(f.read(), object_pairs_hook=OrderedDict)

  create_appgroup(list(app_group_mapping.keys()))
  groups = dict(AppGroup.objects.values_list('group_name', 'id'))

  ds_dict = get_dataset_dict()

  # creating sample dashboards in the same order of the _MEMBERSHIP_METADATA.json
  # Note that the order of artifacts within a subgroup will be reversed
  # becasue the last artifact that gets created will be on top
  for group_name, filenames in list(app_group_mapping.items()):
    for filename in filenames:

      print(('Creating sample artifacts using %s' % filename))
      fullpath = '%s/../samples/dashboards/%s' % (THISDIR, filename)

      with open(fullpath) as f:
        sample = json.loads(f.read())

      if upgrade_datasets:
        new_ds_list = _upgrade_dataset_check(ds_dict, sample)

        if new_ds_list != None:
          report = import_file(request, dc_name='samples', filename=fullpath,
                           sanity_check='False', dry_run='False', workspace_id=1,
                           limit_search_scope='True',
                           skip_thumbnails=skip_thumbnails,
                           generate_thumbnails=generate_thumbnails,
                           upgraded_datasets=new_ds_list)
        else:
          print('No change detected, skipping sample.')
          continue
      else:
        report = import_file(request, dc_name='samples', filename=fullpath,
                         sanity_check='False', dry_run='False', workspace_id=1,
                         limit_search_scope='True',
                         skip_thumbnails=skip_thumbnails,
                         generate_thumbnails=generate_thumbnails)

      if report.status_code != 200:
        raise Exception('import failed for %s file' % filename)

      report = json.loads(report.content)
      ids = [r['id'] for r in report[DashboardSync.name]]
      for id in ids:
        response = single_appgroup_app_update(request, groups[group_name], id, True)
        if response and response.status_code != 200:
          raise Exception('Cannot add artifact with %s ID to %s app group: %s'
                          % (id, group_name, response.content))

  print ('Creating sample apps...')

  with open('%s/../samples/apps/_APPS_METADATA.json' % THISDIR, 'r') as f:
    apps_mapping = json.loads(f.read(), object_pairs_hook=OrderedDict)

  for filename in list(apps_mapping.values()):
    fullpath = '%s/../samples/apps/%s' % (THISDIR, filename)
    with open(fullpath) as f:
      app = json.loads(f.read())

    if upgrade_datasets:
      new_ds_list = _upgrade_dataset_check(ds_dict, app)

      if new_ds_list != None:
        report = import_file(request, dc_name='samples', filename=fullpath,
                         sanity_check='False', dry_run='False', workspace_id=1,
                         limit_search_scope='True',
                         skip_thumbnails=skip_thumbnails,
                         generate_thumbnails=generate_thumbnails,
                         upgraded_datasets=new_ds_list)
      else:
        print('No change detected, skipping app.')
        continue

    else:
      report = import_file(request, dc_name='samples', filename=fullpath,
                       sanity_check='False', dry_run='False', workspace_id=1,
                       limit_search_scope='True',
                       skip_thumbnails=skip_thumbnails,
                       generate_thumbnails=generate_thumbnails)

    if report.status_code != 200:
      raise Exception('import failed for sample app')

    # figuring out the sample app appgroup id
    report = json.loads(report.content)
    appgroup_id = None
    for appgroup in report['appgroups']:
      if appgroup['name'] == 'Sample App':
        appgroup_id = appgroup['id']
        break

    # making the sample app available in the side menu
    data = {'data': json.dumps([{"groupid":appgroup_id,"in_userlist":True}])}
    request = RequestFactory().post('/arc/sqlrun/jsonselect', data=data)
    request.user = user
    _user_appgroup_update(request)

def insert_tables_sqlite():
  """Inserts tables in SQLite database.

    This function will include new tables in samples database,
    but does not update existing tables.
    Steps:
    1. Get target sqlite file in destination
    2. Fetch all JSON files with tabular data in samples directory
    3. Open connection and get all database tables
    4. Add table if filename (coinciding with tablename) is not in db
    5. Commit database changes
  """
  samplessqlite = os.path.join("samples", "samples.sqlite")

  # ARCADIA_DATA_PATH env is set for CDH installs
  if 'ARCADIA_DATA_PATH' in os.environ:
    samplefilepath_dst = os.path.join(os.environ['ARCADIA_DATA_PATH'],
                                      samplessqlite)
  else:
    samplefilepath_dst = os.path.join(
      os.path.dirname(os.path.abspath(settings.DATABASES['default']['NAME'])), \
      samplessqlite)

  # Get list of updated sample data to be imported into database
  samples_src_dir = os.path.join(THISDIR, "../samples/")
  json_files = glob.glob(os.path.join(samples_src_dir, '*.json'))

  # Open SQLite connection and query tables
  with sqlite3.connect(samplefilepath_dst) as conn:
    cursor = conn.cursor()
    cmd = "SELECT tbl_name FROM sqlite_master where type='table'"

    try:
      # Fetchall returns list of tuples, get first element of each tuple
      tables = [x[0] for x in cursor.execute(cmd).fetchall()]

      # Loop through updated samples and add table if missing in current database
      for filepath in json_files:
        abspath_dir, filename = os.path.split(filepath)
        sample_tbl, sep, ext = filename.partition('.')
        if sample_tbl not in tables:
          print('%s missing in db, adding' % sample_tbl)
          with open(filepath) as f:
            data = f.read()

          j = json.loads(data)
          if 'headers' in j:
            headers = j['headers']
            column_names = []
            column_types = []
            for c in headers:
              l, sep, r = c.rpartition(" ")
              column_names.append(l)
              column_types.append(r)
          else:
            raise Exception("Missing headers in json file")
          rows = j['rows']
          cmd = create_table_cmd(sample_tbl, column_names, column_types)
          cursor.execute(cmd)
          insert_rows(cursor, sample_tbl, rows, column_types)
          conn.commit()
        else:
          print('%s in db, skipping insert' % sample_tbl)

    except Exception as e:
      print("Failed on cmd: %s" % cmd)
      raise


def create_table_cmd(tablename, cnames, ctypes):

  nameset = {}
  for idx, c in enumerate(cnames):
    if not c in nameset:
      nameset[c] = idx
      continue
    cnames[idx] = c + "_" + str(idx)

  columns = [(x, y) for x, y in zip(cnames, ctypes)]

  creates = ["CREATE TABLE %s (" % tablename]
  creates.append(", ".join(["%s %s" % (x[0], x[1]) for x in columns]))
  creates.append(")")
  create_cmd = ''.join(creates)
  return create_cmd

def insert_rows(cursor, tablename, rows, ctypes):
  # sqlite can only insert 500 rows at a time
  while len(rows):
    rows_current_batch = rows[:500]
    rows = rows[500:]

    inserts = ["INSERT INTO %s VALUES " % tablename]
    for row in rows_current_batch:
      inserts.append("(")
      display_row = [_insert_display(coldata, coltype) \
                     for coldata, coltype in zip(row, ctypes)]
      inserts.append(",".join(display_row))
      inserts.append(")")
      inserts.append(",")
    insert_cmd = ''.join(inserts[:-1])
    cursor.execute(insert_cmd)

def _insert_display(coldata, coltype):
  if coltype == "STRING" or coltype == "TIMESTAMP":
    return "'%s'" % coldata.replace("'", "''")
  return coldata

def main(server_addr, server_port):
  # Run validation checks
  if not checks():
    return

  # Run syncdb to create the tables
  execute_from_command_line(["manage.py", "migrate", "--fake-initial"])

  # Load the superuser in
  user = User.objects.create_superuser(ADMIN_USERNAME, '', ADMIN_PASSWORD)
  request = HttpRequest()
  request.user = user
  appgroup_starred(request=request)

  # Create the samples connection
  util.migrations.create_sample_dataconnection()

  # Create sample apps
  create_sample_apps(False, True)
  create_dashboard_templates()

  # Setup as instant
  reports.views_settings.setting_update_single('installation_type', 'instant',
      False)
  reports.views_settings.setting_update_single('registration_url',
      'https://registration.arcadiadata.com/register', False)
  reports.views_settings.setting_update_single('site_enable_visual_transitions',
      'true', False)

  loopback_server_url = "http://%s:%s" % (server_addr, server_port)
  # no need to set call back url if it's already set. also not overwrite any user disablement of thumbnails
  # TODO only turn on thumbnail for osx for now, remember to fix arc_site.js in arc.site.siteSettings
  if sys.platform == 'darwin' and reports.views_settings.setting_get_single('site_thumbnails_ip') != loopback_server_url :
    reports.views_settings.setting_update_single('site_thumbnails_ip', loopback_server_url, True)
    reports.views_settings.setting_update_single('site_thumbnails_enable', 'true', False)
  # Ask for registration
  reports.views_settings.unregister()

#
# Work-in-progress. Be able to create samples anew on an installed system
#
# After these two steps, replacing the call to main() below with a call to
# 'createSamples()' will re-create samples
#
def createSamples(skip_thumbnails=False, force=False, upgrade=False):
  # If force is true, we will delete older samples. Otherwise, we only install
  # samples if there are no installed samples.
  dcl = DataConnection.objects.filter(dataconnection_name='samples')
  if dcl and len(dcl) > 0:
    if force:
      dcl.delete()
    else:
      return

  # Create the samples connection and datasets
  ds_ids = util.migrations.create_sample_dataconnection()

  # Create sample apps
  create_sample_apps(skip_thumbnails, upgrade)
  create_dashboard_templates()

if __name__ == "__main__":
  main()
