# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import time
import json
from django.core import serializers

from arcwebbase import access_info
from datasets.models import Extract
from vizlib.querylog import QueryLog, ComponentType, ActionType
import extracts.views
import extracts.misc

#
# See comments in arc/data/arc_extractmgr.js that describe the notion
#   of data extracts
#

class DataCopyJobRunner(object):
  def __init__(self, target_data, target_order, run_parameters, joblog):
    self.target_data = json.loads(target_data)
    self.target_order = target_order
    self.run_parameters = run_parameters
    self.joblog = joblog

    self.extractId = self.target_data.get('extractId')
    self.extract = Extract.objects.get(id=self.extractId)

  def run(self, request=None):

    # access_info has the user who initiated the current run.
    # if that is unavailable (e.g. for scheduled runs), then we
    # can use the user stored in the run data:
    username = access_info.get().username or self.target_data.get('username','')

    with QueryLog(request=request,
                  action=ActionType.UPDATE,
                  component_type=ComponentType.EXTRACT,
                  component_id=self.extract.id,
                  query=f"Data Extract: {self.target_data}",
                  username=username,
                  sync=True):
      self._run_work()

  def _run_work(self):
    self.joblog.add_task('Checking if data extracts is enabled for data connection')
    extracts.misc.checkEnabled()

    t1 = time.time()

    # self.run_parameters[extract] used for sending success/fail extract emails,
    # adding extract details before mgr.runRefresh() in case refresh fails
    _extract = serializers.serialize('json', [self.extract, ])
    self.run_parameters.update({'extract': json.loads(_extract)})

    self.joblog.update_log(last_task=f"Data Extract job started for extract {self.extractId}",
                            run_parameters=self.run_parameters)

    mgr = extracts.views.extractManagerFromExtract(self.extract)
    self.joblog.add_task(task="Preparing to extract data")
    mgr.runRefresh()
    self.joblog.add_task(task="Refreshed extract target")
    stats = mgr.stats()
    insertCount = stats.get("insertCount")

    if insertCount is None:
      insertMsg = ""
    else:
      insertMsg = f"{insertCount} rows inserted. "

    diff = int(time.time() - t1)

    self.joblog.update_log(
        last_task=f"Data Extract job complete for extract {self.extractId}. {insertMsg}Runtime: {diff}s",
        target_data=self.target_data,
        target_order=self.target_order,
        run_parameters=self.run_parameters)

