# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import threading
import os
import json, re, tempfile, time, urllib.request, urllib.parse, urllib.error, base64, html
import shutil
from datetime import datetime
from django.db import models
from django.template import loader
from django.core.mail import EmailMultiAlternatives
from django.conf import settings
from django.contrib.auth.models import User
from django.core.files import File
from django.db.models.signals import pre_delete, pre_save
from django.dispatch import receiver

from jobs.views_helpers import CancelTask, setup_email, send_email
from apps.views_snapshot import create_snapshotrunner
from jobs.views_refresh import _enque_refresh
from email.mime.image import MIMEImage
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email import encoders
from reports.models import Report
from reports.views_settings import setting_get_single
from apps.templatetags.site_setting_name import site_setting_name
from sqlrun.views_jsonselect import JsonSelectRunner
from vizlib.workq import AsyncRunner, JobProgressRunner
from vizlib.accel import Shelf
from dataconn.dataconn import get_dataconn
from vizlib.query_builder import QueryBuilder
from dataconn.qcntl import QueryControl
from vizlib.dataset_request import DatasetRequest
from vizlib.dataset import ModelDataset
from apps.views_remotecalls import restCall
from util.ff_flipper import FeatureFlipper
from util.request import create_request
from datasets.models import Extract

from .datacopy import DataCopyJobRunner

import io, csv

from django.test.client import RequestFactory
from dataconn.connections_types import DATACONN_SOLR
from django.core import serializers
import logging

LOGGER = logging.getLogger("arcweb")

try:
  from croniter import croniter
  _SCHEDULER_ENABLED = True
except Exception as e:
  _SCHEDULER_ENABLED = False
  LOGGER.warning('Croniter not available. Scheduler not enabled')

try:
  if _SCHEDULER_ENABLED and not getattr(settings, 'DJANGO_TEST_ENV', False) and not FeatureFlipper().get_flipper_value('enable_jobs'):
    _SCHEDULER_ENABLED = False
except Exception as e:
  pass

"""
TO CREATE NEW JOB TARGET:
1) Create new class: `class SnapshotJobTarget(JobTarget):`
2) Include a `run_job_target(self, run_parameters, joblog)` function for the class
   Be sure to joblog.update_log(last_task, target_data, run_parameters) often
3) Add new TARGET_TYPES and JobTarget.CHILDREN
4) Will probably want custom @classmethod to create instances of specific JobTarget type

ORGANIZATION:
                          ScheduleGroup
Remembers intervals and next time job should be run links to multiple JobSchedules
                                 |
                                 |
                                 |
      JobSchedule (can be either JobScheduleNow or JobScheduleRepeat)
                                 |
                ---------------------------------
                |                |                |
                |                |                |
                |                |                |
Schedule has list of JobTargets that run according to target_order
                |                |                |
eg: 0:TriggerJobTarget-> 1:SnapshotJobTarget-> 2:EmailJobTarget
                |                |                |
                |                |                |
           -------------------------------------------
           |           |            |                |
           |           |            |                |
    Each schedule has logs generated each JobSchedule is run
           |           |            |                |
          log1        log2         log3             log4


OPTIONAL:
There's a JobContent parent class that can be used to store info that
would be useful to recall for future JobTargets

******************SCHEDULES PARENT CLASS***************************************
 Needs a "run", "progress" and "results" function to use emulate
 vizlib.workq.ProgressRunner

"""
# Have a set of workers for emails/triggers
_JOB_RUNNER_FAST = None
_JOB_RUNNER_FAST_LOCK = threading.Lock()
# Have another set of workers for slow jobs, like ones with refreshes
_JOB_RUNNER_SLOW = None
_JOB_RUNNER_SLOW_LOCK = threading.Lock()

# Statuses are shared among JobLogs and Schedule
STATUS_RUNNING = "R"
STATUS_COMPLETED = "C"
STATUS_FAILED = "F"
STATUS_PAUSED = "P"
STATUS_CANCELED = "X"
STATUS_QUEUED = "Q"
STATUSES = (
  (STATUS_RUNNING, "Still Running"),
  (STATUS_COMPLETED, "Completed"),
  (STATUS_FAILED, "Error"),
  (STATUS_PAUSED, "Paused"),
  (STATUS_CANCELED, "Canceled"),
  (STATUS_QUEUED, "Queued"),
)

STATUS_MAPS = {
  STATUS_RUNNING: "Still Running",
  STATUS_COMPLETED: "Completed",
  STATUS_FAILED: "Error",
  STATUS_PAUSED: "Paused",
  STATUS_CANCELED: "Canceled",
  STATUS_QUEUED: "Queued",
}

EMAIL_BODY_REGEX = re.compile('<messagebody>([\s\S]*)<\/messagebody>')

class NoOwnerException(Exception):
  def __str__(self):
    return 'This job has no owner'

"""
SCHEDULES:

When a job is scheduled, a new instance of this class is created (and tracked in
  the ORM) to track the properties of this schedule and the next run for this job

'next_run_time' is calculated everytime it needs to be, based on 'create_time' and
  the previous value of 'next_run_time'. This is done when a schedule group is created,
  and additionally in the ScheduleChecker thread.
"""
class ScheduleGroup(models.Model):
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  creator = models.ForeignKey(User, default=None, blank=True, null=True, on_delete=models.SET_NULL)
  schedule_name = models.CharField(max_length=200, null=True, default=None, unique=True)
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  run_interval = models.CharField(max_length=200, blank=True, null=True, default=None)
  next_run_time = models.DateTimeField(null=True, blank=True)

  def run_schedule_group(self):
    for schedule in self.jobschedule_set.filter(finished=False):
      if schedule.needs_to_be_run():
        try:
          schedule.add_to_queue()
        except NoOwnerException as e:
          LOGGER.debug(e, exc_info=True)

  def update_next_run_time(self, current_run_time=None):
    if not self.next_run_time:
      iter = croniter(self.run_interval, self.create_time)
    else:
      if current_run_time:
        iter = croniter(self.run_interval, current_run_time)
      else:
        iter = croniter(self.run_interval, datetime.now(self.create_time.tzinfo))
    if iter:
      self.next_run_time = iter.get_next(datetime)
      self.save()

class JobSchedule(models.Model):
  # Create time
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  creator = models.ForeignKey(User, on_delete=models.SET_NULL, default=1, null=True)
  owner = models.ForeignKey(User, on_delete=models.SET_NULL, default=None, blank=True, null=True, related_name="owned_jobschedule_set")
  # ScheduleGroup that the JobSchedule belongs to. adHoc jobs do not belong to a
  # ScheduleGroup.
  schedule_group = models.ForeignKey(ScheduleGroup, blank=True, null=True, on_delete=models.CASCADE)
  # Update time
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  job_name = models.CharField(max_length=200, null=True, default=None)
  run_counter = models.IntegerField(default=0)
  success_counter = models.IntegerField(default=0)
  max_success_run = models.IntegerField(null=True, blank=True)
  # Record previous run stats for quick retrieval
  # for repeating schedule, last_finished_time updates if cancelled or fails excessively
  last_finished_time = models.DateTimeField(default=None, null=True, blank=True)
  last_status = models.CharField(max_length=1, choices=STATUSES, default=None, blank=True, null=True)
  finished = models.BooleanField(default=False)
  # Just used for handling errors
  last_started_time = models.DateTimeField(default=None, null=True, blank=True)
  consecutive_fails = models.IntegerField(default=0)
  # last finished job is just the "target_order" of the last completed JobTarget
  # could be used to only re-run incompleted jobs
  last_finished_job = models.IntegerField(default=None, blank=True, null=True)
  extract = models.ForeignKey(Extract, on_delete=models.CASCADE, null=True)
  trigger_target = models.OneToOneField("JobTarget", on_delete=models.SET_NULL, null=True)

  def init_job_runner(self):
    # In Ambari installations, we run cherrypy with daemonize=1. That leads to
    # importing this module pre-daemonizing, and then forking without threads.
    # This in turn means we cannot initialize _SNAPSHOT_RUNNER at import time
    # and instead have to do it at a later place, protected by _SNAPSHOT_RUNNER_LOCK
    global _JOB_RUNNER_FAST, _JOB_RUNNER_SLOW

    if _JOB_RUNNER_FAST is None:
      with _JOB_RUNNER_FAST_LOCK:
        if _JOB_RUNNER_FAST is None:
          # Have more than 1 thread running since some tasks involve sleeping times
          # where other schedules can run
          worker_count = getattr(settings, 'JOB_RUNNER_FAST_WORKERS', 3)
          _JOB_RUNNER_FAST = AsyncRunner(threadcount=worker_count, name="jobs_fast")

    if _JOB_RUNNER_SLOW is None:
      with _JOB_RUNNER_SLOW_LOCK:
        if _JOB_RUNNER_SLOW is None:
          # Only have 2 threads for now. Unlikely that we'll have more than two
          # Arcadia connections, and right now we only allow one refresh job per
          # data connection
          worker_count = getattr(settings, 'JOB_RUNNER_SLOW_WORKERS', 2)
          _JOB_RUNNER_SLOW = AsyncRunner(threadcount=worker_count, name="jobs_slow")

  def get_targets_and_latest_log(self):
    latest_log = self.joblog_set.first()
    if latest_log:
      latest_log = json.loads(serializers.serialize('json', [latest_log]))[0]
    return {
      'schedule_log': latest_log,
      'targets': json.loads(serializers.serialize('json', self.jobtarget_set.all())),
    }

  def pause_schedule(self):
    self.last_status = STATUS_PAUSED
    self.finished = True
    self.save(update_fields=['last_status', 'finished', 'update_time'])
    LOGGER.debug('Schedule #%s Paused' % self.id)

  def unpause_schedule(self):
    job_log = self.joblog_set.first()
    if job_log:
      self.last_status = job_log.status
    else:
      self.last_status = None
    self.finished = False
    self.save(update_fields=['last_status', 'finished', 'update_time'])
    LOGGER.debug('Schedule #%s Unpaused' % self.id)

  def fail_schedule(self, message):
    self._update_schedule(STATUS_FAILED)
    self.job_log.update_log(end_time=True,
                            run_parameters=self.run_parameters,
                            status=STATUS_FAILED,
                            error_message=message)

  def cancel_schedule(self):
    self._update_schedule(STATUS_CANCELED)

  def _update_schedule(self, new_status):
    self.last_finished_time = datetime.now(self.create_time.tzinfo)
    if self.last_status != STATUS_PAUSED:
      self.last_status = new_status
      LOGGER.debug('Schedule #%s %s' % (self.id, new_status))
    self.save(update_fields=['last_finished_time', 'last_status', 'update_time'])
    if not self.schedule_group:
      self.finished = True
      self.save(update_fields=['finished'])
      LOGGER.debug('Schedule #%s Marked As Finished' % self.id)

  # Return true if it's been longer than the run_interval since last_finished_time
  # And if the scheduler hasn't run more than maximum times
  def needs_to_be_run(self):
    current_time = datetime.now(self.create_time.tzinfo)

    # Make sure the schedule is not paused, probably redundant
    not_paused = self.last_status != STATUS_PAUSED

    # Make sure no error, and if there's been an error,
    # Wait 5 minutes to run it if failed so not run immediately but also allow for
    # next run if schedule for 10 minute intervals
    # (queuing might mean job is less than 10 minutes apart)
    # Don't run if 5 failed attempts in a row
    no_error_or_waited_ten_min = self.last_status != STATUS_FAILED
    if not no_error_or_waited_ten_min and self.last_started_time is not None:
      no_error_or_waited_ten_min = (current_time - self.last_started_time).seconds >= 300

    # Don't run scheduler if it's still running unless it's been 'running' for a day
    not_running = self.last_status != STATUS_RUNNING
    seems_stalled = (current_time - self.update_time).days > 0

    run_interval = self.schedule_group and self.schedule_group.run_interval
    is_adhoc = not run_interval
    if is_adhoc:
      past_run_interval = None
    else:
      if self.last_finished_time:
        last_time = self.last_finished_time
      else:
        last_time = self.create_time

      next_timestamp = croniter(run_interval, last_time).get_next()
      next_dateobject = datetime.fromtimestamp(next_timestamp, current_time.tzinfo)
      past_run_interval = next_dateobject < current_time

    # run if schedule hasn't reached max runs yet
    no_max_run = not self.max_success_run
    not_run_max = no_max_run or self.max_success_run > self.success_counter

    # Mark schedule as "finished" if it's reached maximum run or if paused
    # So it doesn't clutter up check_and_run_schedules
    if not not_run_max or not not_paused:
      self.finished = True
      self.save()

    return not_paused and \
           no_error_or_waited_ten_min and \
           (not_running or seems_stalled) and \
           not_run_max and \
           (is_adhoc or past_run_interval)

  # Runs all tasks asynchronously
  def add_to_queue(self, run_parameters=None, force_run=None, editor_id=None):
    self.init_job_runner()

    # If parent schedule doesn't have next_run_time, update it
    if self.schedule_group and self.schedule_group.next_run_time == None:
      self.schedule_group.update_next_run_time()

    if not run_parameters:
      run_parameters = {
        'skip_next_targets': False,
        'force_run': force_run,
        'creator_id': self.creator.id if self.creator else "Creator user was removed",
        'creator_username': self.creator.username if self.creator else "Creator user was removed",
        # owner else should not happen since run button is disabled - this is just handling
        'owner_id': self.owner.id if self.owner else "Error: This job has no owner",
        'owner_username': self.owner.username if self.owner else "Error: This job has no owner",
      }

    if editor_id:
      run_parameters['editor_id'] = editor_id
      run_parameters['editor_username'] = User.objects.get(id=editor_id).username

    self.run_parameters = run_parameters
    if self.last_status != STATUS_PAUSED:
      self.last_status = STATUS_QUEUED
      self.save()

    contents_archive = {
      'run_parameters': self.run_parameters,
    }
    self.job_log = self.joblog_set.create(contents_archive=json.dumps(contents_archive),
                                          task_list=json.dumps({'tasks': [{'message':'Starting Task', 'date': str(datetime.now(self.create_time.tzinfo))}]}),
                                          status=STATUS_QUEUED,
                                          )
    LOGGER.debug('Schedule #%s Queued' % self.id)

    targets_to_run = self.jobtarget_set.all()
    involves_data_movement = False
    for target in targets_to_run:
      if target.target_type == JobTarget.REFRESH or target.target_type == JobTarget.DATACOPY:
        involves_data_movement = True

    if self.owner:
      if involves_data_movement:
        _JOB_RUNNER_SLOW.queue_untracked(JobProgressRunner(self, username=self.owner.username))
      else:
        _JOB_RUNNER_FAST.queue_untracked(JobProgressRunner(self, username=self.owner.username))
    else:
      error_message = 'Job has no owner'
      self.fail_schedule(error_message)
      raise NoOwnerException(error_message)

  """
  AsyncRunner methods
  # RUN: will sequentially run each job_target associated with a schedule
  # PROGRESS: Returns back completed, total targets
  # RESULTS: Returns completed/total targets as string
  """
  def run(self):
    self.run_counter += 1
    self.job_log.update_log(status=STATUS_RUNNING, start_time=True)
    if self.last_status != STATUS_PAUSED:
      self.last_status = STATUS_RUNNING

    self.last_started_time = datetime.now(self.create_time.tzinfo)
    self.save(update_fields=['run_counter', 'last_status', 'last_started_time', 'update_time'])
    targets_to_run = self.jobtarget_set.all()
    self.total_targets = len(targets_to_run)
    self.completed = 0

    # sleep for 30 seconds to allow any other processes to run (especially needed for force_runs)
    if self.run_parameters.get('force_run'):
      time.sleep(30)

    dataCopyJobTarget = None
    emailJobTarget = None
    try:
      # Begin running each target, updating schedule when succeeds
      LOGGER.info('Schedule #%s Running' % self.id)
      for target in targets_to_run:
        if target.target_type == JobTarget.DATACOPY: # check if this schedule runs DataCopyJobTarget
          dataCopyJobTarget = target.CHILDREN[target.target_type].objects.get(id=target.id)
        elif target.target_type == JobTarget.EMAIL:
          emailJobTarget = target

        skip_reason = self.run_parameters.get('skip_next_targets')
        if skip_reason:
          self.job_log.update_log(last_task='Skipping next targets: %s' % skip_reason,
                                  run_parameters=self.run_parameters)
          break
        target.run_counter += 1
        target.save(update_fields=['run_counter', 'update_time'])
        target.find_and_run_target(self.run_parameters, self.job_log)
        self.last_finished_job = self.completed
        self.save(update_fields=['last_finished_job', 'update_time'])
        self.completed += 1
      # Final accounting for successfully completed jobs
      self.success_counter += 1

      # Don't change last status if paused
      if self.last_status != STATUS_PAUSED:
        self.last_status = STATUS_COMPLETED

      self.last_finished_time = datetime.now(self.create_time.tzinfo)
      self.consecutive_fails = 0
      self.save(update_fields=['success_counter', 'last_status',
                               'last_finished_time', 'finished',
                               'consecutive_fails', 'update_time'])
      self.job_log.update_log(end_time=True,
                              run_parameters=self.run_parameters,
                              status=STATUS_COMPLETED,
                              )
      self.needs_to_be_run() # updates finished

      # If email not set up properly, can't send email, so just forget this part
      try:
        if dataCopyJobTarget:
          dataCopyJobTarget.send_notification_email(self, None)
      except Exception as e:
          extract_id = self.run_parameters['extract'][0].get('pk')
          last_task_msg = f'Data Extract Completed: {self.job_log.get_last_task()}. '
          last_task_msg += f'Sending "SUCCESS" email has failed for extract #{extract_id}: {str(e)}'
          self.job_log.add_task(last_task_msg)

    except CancelTask as e:
      self.last_finished_time = datetime.now(self.create_time.tzinfo)
      self.save(update_fields=['last_finished_time', 'update_time'])
      # pause schedule if ad-hoc so it doesn't automatically run in 5 minutes
      if not self.schedule_group:
        self.pause_schedule()
      else:
        # Don't change status if paused
        if self.last_status != STATUS_PAUSED:
          self.last_status = STATUS_CANCELED
          self.save(update_fields=['last_status', 'update_time'])

        self.job_log.update_log(end_time=True,
                                run_parameters=self.run_parameters,
                                status=STATUS_CANCELED)
    except Exception as e:
      LOGGER.exception(f'Error Running Schedule #{self.id}')

      # Record failed run
      paused_text = ''
      if self.consecutive_fails >= 5:
        self.pause_schedule()
        paused_text = '\nThis job has been paused due to 5 or more consecutive failed runs'
      elif self.last_status != STATUS_PAUSED:
        self.last_status = STATUS_FAILED
        self.save(update_fields=['last_status', 'update_time'])
      self.consecutive_fails += 1
      self.last_finished_time = datetime.now(self.create_time.tzinfo)
      self.save(update_fields=['last_finished_time', 'consecutive_fails', 'update_time'])
      self.job_log.update_log(end_time=True,
                              run_parameters=self.run_parameters,
                              status=STATUS_FAILED,
                              error_message=str(e),
                              )

      # Try to send error email
      self.job_log = JobLog.objects.get(id=self.job_log.id)
      latest_contents = json.dumps(json.loads(self.job_log.contents_archive), indent=2)
      email_message = 'Error Running Schedule #%d ' % self.id
      email_message += "(" + self.job_name + ")"
      email_message += paused_text
      email_message += "\n\nLast Run Task:\n"
      email_message += self.job_log.get_last_task()
      email_message += "\n\nError:\n"
      if self.run_parameters.get('error'):
        email_message += self.run_parameters['error']
      else:
        email_message += json.loads(latest_contents).get('error', 'No Error Message')
      email_message += '\n\nEmail Log #%s:\n' % self.job_log.id
      email_message += latest_contents

      # If email not set up properly, can't send email, so just forget this part
      try:
        if dataCopyJobTarget:
          dataCopyJobTarget.send_notification_email(self, email_message)
        else:
          subject = self.run_parameters.get("email_subject", self.job_name) + " (JOB ERROR)"
          error_to = json.loads(emailJobTarget.target_data).get('error_to_addresses', settings.EMAIL_HOST_USER)
          send_email(subject, email_message, error_to, settings.EMAIL_HOST_USER)
      except Exception as e:
        if dataCopyJobTarget:
          extract_id = self.run_parameters['extract'][0].get('pk')
          last_task_msg = f'Data Extract Failed: {self.job_log.get_last_task()}. '
          last_task_msg += f'Sending "FAIL" email has failed for extract #{extract_id}: {str(e)}'
          self.job_log.add_task(last_task_msg)

  def progress(self):
    return (self.completed, self.total_targets)

  def results(self):
    return "{0}/{1}".format(*self.progress())


"""
SCHEDULES: Run immediately
"""
class JobScheduleNow(JobSchedule):
  class Meta:
    proxy = True

@receiver(pre_save, sender=JobScheduleNow)
def set_adhoc_properties(sender, instance, **kwargs):
  # Only run when creating instance
  if not instance.id:
    instance.job_name = "Adhoc "
    instance.max_success_run = 1

"""
SCHEDULES: Run at certain intervals
"""
class JobScheduleRepeat(JobSchedule):
  class Meta:
    proxy = True


"""
****************TARGET JOB PARENT CLASS***************************************
"""
class JobTarget(models.Model):
  class Meta:
    ordering = ['target_order']
  # Create time
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  # Update time
  update_time = models.DateTimeField(auto_now=True, blank=True)
  # defines order JobTargets run for each schedule
  target_order = models.IntegerField(default=0)
  # The schedule that triggered this job target
  schedule = models.ForeignKey('JobSchedule', on_delete=models.CASCADE, null=True, blank=True)
  run_counter = models.IntegerField(default=0)
  success_counter = models.IntegerField(default=0)
  # minimum_time_interval stores minimum wait before running again
  minimum_time_interval = models.IntegerField(default=None, null=True, blank=True)
  last_finished_time = models.DateTimeField(default=None, null=True, blank=True)

  EMAIL = "E"
  SNAPSHOT = "S"
  TRIGGER = "T"
  REFRESH = "R"
  SCRIPT = "C"
  DATACOPY = "D"

  TARGET_TYPES = (
    (EMAIL, "Email"),
    (SNAPSHOT, "Snapshot"),
    (TRIGGER, "Trigger Assessment"),
    (REFRESH, "Refresh Analytical Views"),
    (SCRIPT, "Script"),
    (DATACOPY, "Data Copy"),
  )
  target_type = models.CharField(max_length=4, choices=TARGET_TYPES, default=EMAIL, db_index=True)
  target_data = models.TextField(default="", blank=True)

  """
  Returns dictionary with right properties for creating new JobTarget object.
  Note that schedule and target_type must be specified
  """
  @classmethod
  def job_target_properties(cls, target_type=None, schedule=None,
      target_order = None, minimum_time_interval=None):

    if not target_type:
      raise Exception('target_type must be specified to create job target')

    if not schedule:
      raise Exception('schedule must be specified to create job target')

    # If target_order is not specified, make it last
    if not target_order:
      target_order = schedule.jobtarget_set.count()

    return {
      # Required
      'target_type': target_type,
      'schedule': schedule,
      # Optional
      'minimum_time_interval': minimum_time_interval,
      'target_order': target_order,
    }

  # True if has passed minimum_time_interval
  def needs_to_be_run(self):
    no_min_interval = not self.minimum_time_interval
    if no_min_interval:
      past_minimum_interval = None
    else:
      current_time = datetime.now(self.create_time.tzinfo)
      last_time = self.last_finished_time
      if last_time:
        past_minimum_interval = (current_time - last_time).total_seconds() > self.minimum_time_interval
      else:
        past_minimum_interval = True

    return (no_min_interval or past_minimum_interval)

  """
  Finds the specific JobTarget instance and runs the instance
  Also creates and saves log for each job_target run
  """
  def find_and_run_target(self, run_parameters, job_log):
    # Since relationships point to JobTarget, need to figure out specific type of job target
    # to specific job target object to run_job_target function
    child_instance = self.CHILDREN[self.target_type].objects.get(id=self.id)
    task_name = '#%s: %s (id #%d)' % (child_instance.target_order + 1,
                                      child_instance._meta.verbose_name,
                                      child_instance.id)
    if run_parameters.get('force_run') or self.needs_to_be_run():
      try:
        job_log.update_log(last_task='Starting Task %s' % task_name,
                           target_order=child_instance.target_order,
                           target_data=child_instance.target_data,
                           run_parameters=run_parameters
                           )
        child_instance.run_job_target(run_parameters, job_log)
        child_instance.success_counter += 1
        child_instance.last_finished_time = datetime.now(child_instance.create_time.tzinfo)
        child_instance.save()
        job_log.update_log(last_task='Finished Task %s' % task_name,
                           target_order=child_instance.target_order,
                           target_data=child_instance.target_data,
                           run_parameters=run_parameters,
                           last_completed_target=self.target_order,
                           )
      except CancelTask as e:
        run_parameters['canceled'] = task_name

        outDir = run_parameters.get('image')
        if outDir:
          self.safely_remove_tmp_folder(pathToDir=outDir)

        raise CancelTask('Canceled')
      except Exception as e:
        LOGGER.exception('Error Running Target #%d ' % self.id)
        LOGGER.exception(e)
        error_message = str(e)
        run_parameters['last_task'] = job_log.get_last_task()
        run_parameters['last_target'] = task_name
        if run_parameters.get('error'):
          run_parameters['error'] = run_parameters.get('error') + error_message
        else:
          run_parameters['error'] = error_message

        outDir = run_parameters.get('image')
        if outDir:
          self.safely_remove_tmp_folder(pathToDir=outDir)

        raise Exception('Job Target Failed: {0}'.format(error_message))
    else:
      # Skip next targets if not needed to run
      # Else, next dependent targets will be run without dependencies
      run_parameters['skip_next_targets'] = 'Task %s does not need to be run' % task_name

  def safely_remove_tmp_folder(self, pathToDir):
    try:
      shutil.rmtree(pathToDir, ignore_errors=True)
    except Exception as e:
      LOGGER.warning("cannot remove temporary folder %s, %s" % (pathToDir, str(e)))

  def overwrite_values(self, form, property_form=None):
    specific_child = self.CHILDREN[self.target_type]
    if hasattr(specific_child, 'get_editable_target_data'):
      td = specific_child.get_editable_target_data(form)

    is_updated = False
    target_data = json.loads(self.target_data)
    for key, value in td.items():
      old_target_value = target_data.get(key)
      if old_target_value != value:
        is_updated = True
        target_data[key] = value

    if is_updated:
      self.target_data = json.dumps(target_data)

    if hasattr(specific_child, 'get_editable_properties'):
      prop = specific_child.get_editable_properties(form)
      for key, value in prop.items():
        old_prop_value = getattr(self, key)
        if old_prop_value != value:
          is_updated = True
          setattr(self, key, value)

    if is_updated:
      self.save()

    return target_data

  def _add_trigger_content(self, array_of_content=None, name=""):
    array_of_content = array_of_content or []
    for cont in array_of_content:
      job_content = EmailJobContent.objects.filter(field=name, value=cont)
      if len(job_content) == 0:
        self.schedule.creator.jobcontent_set.create(field=name, value=cont)
      else:
        self.schedule.creator.jobcontent_set.add(*job_content)
    return self

"""
TRIGGER TARGET: Specific proxy class for assessing triggers
"""
class TriggerJobTarget(JobTarget):
  class Meta:
    ordering = ['target_order']
    proxy = True

  """
  Private methods for making dsreq
  """

  def _get_having(self, filter_shelf):
    # Seems to be what is happening in base.addFiltersToRb
    default_having = []
    for filter in filter_shelf:
      custom_expr = filter.get('custom_expr')
      # Not the best way to check for aggregates, but can't use expr_hasAggregate in validator
      if custom_expr and Shelf(None, None, None).expressionIsAggregate(custom_expr):
        default_having.append(custom_expr)
    return default_having

  def _get_shelf_els(self, shelf_json, is_aggregate=None, is_filter=None):
    dsreq_array = []
    replace_brackets = re.compile('\[|\]')
    for el in shelf_json:
      el_obj = {}
      # Add order if there's one
      el_order = el.get('order')
      if el_order and (el_order.get('ascending') != None):
        el_obj['order'] = {
          'asc': el_order.get('ascending', False),
          'pri': el_order.get('priority', 1),
        }
      # Add expr
      expression_for_trigger = el.get('expression_for_trigger')
      custom_expression = el.get('custom_expr')
      if expression_for_trigger:
        expression = expression_for_trigger
      elif custom_expression:
        expression = custom_expression
      else:
        # type only changes for events and segments loooking at arc_sql's request_builder
        el_obj['type'] = 'SIMPLE'
        # Now painfully parse expression
        function_name = el.get('fn_name')
        column = el.get('dataset_colname')

        if not column:
          app_id = json.loads(self.target_data)['trigger_id']
          raise Exception('Please re-save visual #%d. Cannot parse shelf element %s' %
                          (app_id, shelf_json))
        elif function_name:
          expression = function_name + "([" + column + "])"
        else:
          expression = "[" + column + "]"

      if is_filter:
        if not Shelf(None, None, None).expressionIsAggregate(expression):
          dsreq_array.append(expression)
      else:
        col_name = replace_brackets.sub('', expression)
        col_name = el.get('col_alias', col_name)
        el_obj['expr'] = expression + ' as "' + col_name + '"'
        if is_aggregate and not hasattr(self, '_first_aggregate'):
          self._first_aggregate = col_name

        dsreq_array.append(el_obj)

    return dsreq_array

  def _get_segments(self, saved_segments):
    results = []
    for seg in saved_segments:
      results.append({
        'seg_id': seg['id'],
        'negate': seg['negate'],
      })
    return results

  """
  Custom method for making snapshots
  """
  @classmethod
  def create_trigger(cls, *args, **kwargs):
    cleaned_kwargs = cls.job_target_properties(
        target_type=JobTarget.TRIGGER,
        schedule=kwargs.get("schedule"),
        target_order=kwargs.get("target_order"),
        minimum_time_interval=kwargs.get("minimum_time_interval"))

    target_data = cls.get_target_data(kwargs)

    cls._check_trigger_visual(target_data['trigger_id'])

    cleaned_kwargs["target_data"] = json.dumps(target_data)

    trigger_target = cls.objects.create(*args, **cleaned_kwargs)
    schedule = trigger_target.schedule
    schedule.trigger_target = trigger_target
    schedule.save()
    return trigger_target

  @classmethod
  def get_editable_target_data(cls, form):
    return cls.get_target_data(form)

  @staticmethod
  def get_target_data(arg):
      return {
      'trigger_id': int(arg['trigger_id']),
      'trigger_type': arg['trigger_type'],
      'trigger_conditions': arg.get('trigger_conditions', []),
    }

  @classmethod
  def get_editable_properties(cls, form):
      return {
      'minimum_time_interval': int(form['minimum_time_interval'])
    }

  @classmethod
  def _check_trigger_visual(cls, visual_id):
    trigger_visual = Report.objects.filter(id=visual_id)
    if trigger_visual.count():
      trigger_visual = trigger_visual.first()
      visual_type = trigger_visual.report_type
      if visual_type == 'table':
        # Also, check that visual isn't solr, since parsing that dsreq isn't currently supported
        if trigger_visual.dataset.dataconnection.dataconnection_type != DATACONN_SOLR:
          return trigger_visual
        else:
          raise Exception('Solr not supported')
      else:
        raise Exception('Visual #%d is %s, not table type' % (visual_id, visual_type))
    else:
      raise Exception('No visual #%d' % visual_id)

  """
  Gets trigger
  NEED TO MODIFY EVERY TIME DSREQ OR JSONSELECT_WORK INPUTS ARE CHANGED!
  """
  def run_job_target(self, run_parameters, joblog):
    joblog.add_task('Get data for table visual')
    trigger_params = json.loads(self.target_data)
    table_visual = TriggerJobTarget._check_trigger_visual(trigger_params['trigger_id'])
    joblog.add_task('Creating inputs for jsonselect_work')
    report_data = json.loads(table_visual.report_data)['report_data']
    sample_pct = report_data.get('sample_pct', '')
    if sample_pct.lower() == 'off':
      sample_pct = ''
    dsreq = {
      'aggregates': self._get_shelf_els(report_data['aggregates_shelf'], is_aggregate=True),
      'dataset_id': table_visual.dataset_id,
      'derived_data': report_data['report_derived_data'],
      'dimensions': self._get_shelf_els(report_data['dimensions_shelf']),
      'distinct': False, # only on for picklists and maybe extensions?
      'events': [],
      'filters': self._get_shelf_els(report_data['filters_shelf'], is_filter=True),
      'global_segments': [
        {
          # TODO: Discard global_segments -- don't think it's used anymore
          'segments': self._get_segments(report_data.get('selected_segments', [])),
          'type': 'SEGMENT',
       }
      ],
      'having': self._get_having(report_data['filters_shelf']),
      'limit': Report.get_limit(report_data),
      'nulls_order': report_data['user_settings'].get('nulls_order', 'Default'),
      'partitions': [], # probably none, only used in incremental queries
      'session': [], # only used for paths
      'solr_query': [], # not supported, would have to also generate query
      'time': [], # only used for paths, don't need
      # According to default dsreq in arc.requestBuilder
      'type': 'SQL', # think this is always the case? Only SOLR is alternative option
      'version': 1,
    }
    jsonselect_params = {
      'dsreq': json.dumps(dsreq),
      'cache_read': 'false',
      'sample_pct': sample_pct,
      'report_id': table_visual.id,
      'dataset_id': table_visual.dataset_id,
      # Don't need:
      #   query (built from dsreq),
      #   dataconnection_id (gets from dsreq.dataset_id))
      #   async (this should always be false?)
    }
    joblog.add_task('Running query')
    fake_request = RequestFactory() # needed for @login_required for jsonselect_work
    fake_request.user = self.schedule.owner
    fake_request.META = {
      'REMOTE_ADDR': None,
    }

    # Mimic jsonselect_init_request from jsonselect_parallel
    fake_request._mdict = {}
    fake_request._mdlock = _JOB_RUNNER_FAST_LOCK
    jsonselect_runner = JsonSelectRunner(request=fake_request, params=jsonselect_params,
        wait_for_results=JsonSelectRunner.WAIT_LONG)
    jsonselect_runner.run()
    data = jsonselect_runner._results
    run_parameters['trigger_id'] = table_visual.id
    run_parameters['trigger_name'] = report_data['report_title']
    run_parameters['trigger_data'] = data
    joblog.update_log(last_task='Getting first measure of data result', run_parameters=run_parameters)
    if data.get('error'):
      raise Exception("Trigger Assessment Failed: " + data.get('error'))
    else:
      data_results = data['data']
      colnames = data_results['colnames']
      if not hasattr(self, '_first_aggregate'):
        raise Exception('No element in measures shelf')
      first_aggregate_idx = colnames.index(self._first_aggregate)
      rows = json.loads(data_results['rows'])
      if len(rows) == 0:
        raise Exception('No data in this visual')

      first_agg_value = rows[0][first_aggregate_idx]

      # Handle cases where measure is 'Remote call'
      if isinstance(first_agg_value, str):
        tables = json.loads(table_visual.dataset.dataset_info)
        for table in tables:
          for column in table['columns']:
            if column.get('alias') == self._first_aggregate:
              # regexp from cvizRestRE
              column_expr = column.get('expr', '')
              cviz_rest_data = re.match('\bcviz_rest\((.*)?\)\s*$', column_expr, re.IGNORECASE)
              if not cviz_rest_data:
                  raise Exception('Malformed remote data settings: %s' % column_expr)

              cviz_rest_data = json.loads(cviz_rest_data.groups()[0])

              # This is a copy of cvizRestSettingsToRestCall + fetchRemoteDataWork
              # NOTE: Params are not supported because a single visual should not have associated parameters
              remoteCallSettings = {
                'version': 2.0,
                'response_colname': cviz_rest_data['response_colname'],
                'url': cviz_rest_data['url'],
                'payload': {
                  'request': {
                    'data': {},
                  },
                },
              }

              api_config = cviz_rest_data.get('config')
              if api_config:
                remoteCallSettings['payload']['request']['config'] = api_config

              api_accessKey = cviz_rest_data.get('accessKey')
              if api_accessKey:
                remoteCallSettings['payload']['accessKey'] = api_accessKey

              api_colnames = cviz_rest_data.get('colnames', [])
              if api_colnames:
                remoteCallSettings['payload']['request']['data']['colnames'] = api_colnames

              colnames = data_results['colnames']
              api_colIndices = []
              api_coltypes = []
              api_row = []  # just need first row
              coltypes = data_results.get('coltypes', [])
              for col in api_colnames:
                api_idx = colnames.index(col)
                api_colIndices.append(api_idx)
                api_coltypes.append(coltypes[api_idx])
                api_row.append(rows[0][api_idx])

              remoteCallSettings['payload']['request']['data']['coltypes'] = api_coltypes
              remoteCallSettings['colIndices'] = api_colIndices
              remoteCallSettings['payload']['request']['data']['rows'] = [api_row]
              first_agg_value = restCall(remoteCallSettings, fake_request.user.username)[0]

      run_parameters['trigger_data_value'] = first_agg_value
      joblog.update_log(last_task='Parsing trigger_fullconditional', run_parameters=run_parameters)

      trigger_conditions = trigger_params.get('trigger_conditions')

      # adapter for old trigger format - edited jobs are updating themselves
      if trigger_conditions is None:
        trigger_conditional = trigger_params.get('trigger_conditional')
        trigger_threshold = trigger_params.get('trigger_threshold')
        trigger_conditions = [{'0':{'alert-equality': trigger_conditional, 'alert-trigger-value': trigger_threshold}}]
        trigger_params['trigger_conditions'] = trigger_conditions

      trigger_type = trigger_params['trigger_type']

      if trigger_type == 'D':
        trigger_type_name = 'Change in ' + self._first_aggregate
      else:
        trigger_type_name = self._first_aggregate

      joblog.update_log(last_task='Getting current trigger value', run_parameters=run_parameters)
      completed_jobs = []
      if trigger_type == 'D' or self.minimum_time_interval == -1:
        completed_jobs = self.schedule.joblog_set.filter(status=STATUS_COMPLETED, last_completed_target__gte=self.target_order)
      if trigger_type == 'D':
        # If calculating difference, can't calculate if job is being run first time
        if completed_jobs.count() == 0:
          run_parameters['skip_next_targets'] = 'Conditional Not Triggered (First run)'
          return
        else:
          previous_job_log = json.loads(completed_jobs.first().contents_archive)
          previous_value = float(previous_job_log['run_parameters']['trigger_data_value'])
          current_value = float(run_parameters['trigger_data_value'])
          trigger_value = current_value - previous_value
          run_parameters['trigger_difference'] = trigger_value
      else:
        trigger_value = float(run_parameters['trigger_data_value'])

      run_parameters['trigger_value'] = trigger_value

      joblog.add_task('Add words to run_parameters for using in email')
      run_parameters.update(trigger_params)
      joblog.update_log(last_task='Comparing trigger measure to threshold', run_parameters=run_parameters)

      condition_results = []
      logical_operator = ''
      fullconditional = ''
      for idx, condition in enumerate(trigger_conditions):
        condition_values = condition[str(idx)]
        equality = condition_values['alert-equality']
        value_to_compare = float(condition_values['alert-trigger-value'])
        if equality == '>':
          trigger_active = trigger_value > value_to_compare
        elif equality == '<':
          trigger_active = trigger_value < value_to_compare
        elif equality == '>=':
          trigger_active = trigger_value >= value_to_compare
        elif equality == '<=':
          trigger_active = trigger_value <= value_to_compare
        elif equality == '=':
          trigger_active = trigger_value == value_to_compare
        else:
          trigger_active = False
        condition_results.append(trigger_active);

        fullconditional = f'{fullconditional}{logical_operator}{trigger_type_name} {equality} {value_to_compare} '
        logical_operator = 'AND '

      run_parameters['trigger_fullconditional'] = fullconditional.rstrip()

      if False in condition_results: # MVP: 'AND' operator between all conditions
        trigger_active = False
      else:
        trigger_active = True

      if trigger_active:
        run_parameters['triggered_time'] = datetime.now(self.create_time.tzinfo).strftime('%x %X %Z')
        # email only on first trigger check
        if self.minimum_time_interval == -1:
          if completed_jobs.count() > 0:
            previous_job_log = json.loads(completed_jobs.first().contents_archive)['run_parameters']
            is_previous_job_active = previous_job_log.get('last_trigger_active')
            if is_previous_job_active:
              run_parameters['skip_next_targets'] = 'Email already sent for active trigger'
              run_parameters['first_trigger_value'] = previous_job_log.get('first_trigger_value')
              run_parameters['first_triggered_time'] = previous_job_log.get('first_triggered_time')
              run_parameters['first_trigger_fullconditional'] = previous_job_log.get('first_trigger_fullconditional')
          else:
            run_parameters['first_trigger_value'] = trigger_value
            run_parameters['first_triggered_time'] = run_parameters['triggered_time']
            run_parameters['first_trigger_fullconditional'] = run_parameters['trigger_fullconditional']

      if not trigger_active:
        run_parameters['skip_next_targets'] = 'Conditional Not Triggered'

      run_parameters['last_trigger_active'] = trigger_active

"""
Script target: Specific proxy class for running scripts (for now, only SQL scripts)
"""
class ScriptTarget(JobTarget):
  class Meta:
    ordering = ['target_order']
    proxy = True

  """
  Custom method for creating scripts
  """
  @classmethod
  def createScript(cls, schedule=None, script_type=None, script=None,
      email_address=None, dataconnection_id=None, user=None):
    params = cls.job_target_properties(
        target_type=JobTarget.SCRIPT, schedule=schedule)

    target_data = {
      # required
      'script_type': script_type,
      'script': script,
      'user_name': user.username,
      'email_address': email_address,
    }

    if dataconnection_id is not None:
      target_data['dataconnection_id'] = dataconnection_id

    params["target_data"] = json.dumps(target_data)

    return cls.objects.create(**params)

  """
  Run script
  """
  def run_job_target(self, run_parameters, joblog):
    joblog.add_task('Running SQL')

    runParams = json.loads(self.target_data)
    if runParams['script_type'] != "sql":
      raise Exception("Unknown script type %s" % (runParams['script_type']))

    conn = get_dataconn(connection_id=runParams['dataconnection_id'])
    script = runParams['script']
    email_address = runParams.get('email_address')

    for q in script.split(';'):
      if q.rstrip('\n'):
        result = conn.sync_runquery(q, username=runParams['user_name'])
        if result.get('rows'):
          result['rows'] = result['rows'][0:64]
        LOGGER.info("Raw SQL job is completed")
        run_parameters['results'] = result

    if email_address:
      email_address = email_address.strip()

    if email_address:
      setup_email()
      msg = EmailMultiAlternatives(
          subject='SQL results',
          body="SQL Command: %s" % script,
          from_email=settings.EMAIL_HOST_USER,
          to=[email_address])

      output = io.StringIO()
      writer = csv.writer(output)
      for row in result.get('rows', []):
        writer.writerow(row)

      csv_attachment = MIMEText(output.getvalue(), _subtype='csv')
      csv_attachment.add_header('Content-Disposition', 'attachment', filename='results.csv')
      msg.attach(csv_attachment)

      status = msg.send()
      LOGGER.info("Email sent. %s", status)

    LOGGER.info("Raw SQL job is completed. Email: %s", email_address)

    joblog.update_log(last_task='SQL Execution Complete',
                      target_data=runParams,
                      target_order=self.target_order,
                      run_parameters=run_parameters)

"""
DATACOPY target: Specific proxy class for running data copy operations
"""
class DataCopyJobTarget(JobTarget):
  class Meta:
    ordering = ['target_order']
    proxy = True

  """
  Custom method for creating a data copy operation
  """
  @classmethod
  def createDataCopy(cls, request, schedule=None, extractId=None, srcdataset_url=None, jobs_url=None):
    params = cls.job_target_properties(
        target_type=JobTarget.DATACOPY, schedule=schedule)

    target_data = {
      'extractId': extractId,
      'username': request.user.username,
      'srcdataset_url': srcdataset_url,
      'jobs_url': jobs_url
    }

    params["target_data"] = json.dumps(target_data)

    LOGGER.info(f"User {request.user.username} created data extract job for extract {extractId}")
    return cls.objects.create(**params)

  """
  Run data copy
  """
  def run_job_target(self, run_parameters, joblog, request=None):
    d = DataCopyJobRunner(self.target_data, self.target_order, run_parameters, joblog)
    return d.run(request)

  """
  Send notification email
  """
  def send_notification_email(self, schedule, fail_email_msg):
    target_data = json.loads(self.target_data)
    extract = schedule.run_parameters['extract'][0]
    de_id = extract.get('pk')
    de_notification_email = json.loads(extract['fields']['extract_config']).get('notification_email', None)

    if de_notification_email:
      de_email_success = True if de_notification_email.get('email_success') == 'true' else False
      de_email_fail = True if de_notification_email.get('email_fail') == 'true' else False
      de_to_addresses = de_notification_email.get('to_addresses', settings.EMAIL_HOST_USER)
      de_email_subject = de_notification_email.get('email_subject', f'[<<extract_status>>] {settings.DEFAULT_EMAIL_PREFIX} Data Extract: <<extract_id>>')

      status = ''
      email_message = None

      if de_email_success and schedule.last_status == STATUS_COMPLETED:
        status = 'SUCCESS'
        schedule.job_log = JobLog.objects.get(id=schedule.job_log.id)
        email_message = f'Completed Running Schedule #{schedule.id} '
        email_message += f'({schedule.job_name})'
        email_message += "\n\nLast Run Task:\n"
        email_message += schedule.job_log.get_last_task()
        email_message += f'\n\nEmail Log #{schedule.job_log.id}:\n'
        latest_contents = json.dumps(json.loads(schedule.job_log.contents_archive), indent=2)
        email_message += latest_contents
      if de_email_fail and fail_email_msg:
        status = 'FAIL'
        email_message = fail_email_msg

      if email_message:
        srcdataset_url = target_data.get('srcdataset_url', None)
        jobs_url = target_data.get('jobs_url', None)

        if srcdataset_url:
          email_message += f'\n\nSource dataset link: {srcdataset_url}'

        if jobs_url:
          email_message += f'\nJobs page link: {jobs_url}'

        subject = de_email_subject.replace('<<extract_status>>', status).replace('<<extract_id>>', str(de_id))
        last_task_msg = f'Data Extract Completed: {schedule.job_log.get_last_task()}. '

        if setting_get_single("email_enabled") == 'false':
          last_task_msg += f'Didn\'t send email for extract #{de_id}: '
          last_task_msg += f'Emailing visuals and extract refresh status not enabled in {site_setting_name(False)}'
        else:
          send_email(subject, email_message, de_to_addresses, settings.EMAIL_HOST_USER)
          last_task_msg += f'"{status}" email has sent for extract #{de_id}'
        schedule.job_log.add_task(last_task_msg)

"""
SNAPSHOT TARGET: Specific proxy class for generating snapshots and CSV's
                 Make CSV by loading page because don't want to parse dsreq in backend
"""
class SnapshotJobTarget(JobTarget):
  class Meta:
    ordering = ['target_order']
    proxy = True

  """
  Custom method for making snapshots and CSVs
  """
  @classmethod
  def create_snapshot(cls, *args, **kwargs):
    cleaned_kwargs = cls.job_target_properties(
        target_type=JobTarget.SNAPSHOT,
        schedule=kwargs.get("schedule"),
        target_order=kwargs.get("target_order"),
        minimum_time_interval=kwargs.get("minimum_time_interval"))

    target_data = {
      # required
      'app_id': kwargs['app_id'],
      'include_csv': kwargs.get('image_type').count(EmailJobTarget.IMAGE_TYPES['attachCSV']),
      'include_xlsx': kwargs.get('image_type').count(EmailJobTarget.IMAGE_TYPES['attachXLSX']),
      'csv_download_rows': kwargs.get('csv_download_rows', '5000'),
      # optional
      'dbWidth': float(kwargs.get('dbWidth', -1)),
      'dbHeight': float(kwargs.get('dbHeight', -1)),
      'email_subject': '[Error] ' + kwargs.get('email_subject', ''),
      'error_to_addresses': kwargs.get('error_to_addresses', []),
      'reply_to_address': kwargs.get('reply_to_address', ''),
      'paramString': kwargs.get('paramString', 'embed=true'),
      'uploadFormat': kwargs.get('pdf_option')
        if kwargs.get('image_type').count(EmailJobTarget.IMAGE_TYPES['attachPDF'])
        else 'unknown'
    }

    cleaned_kwargs["target_data"] = json.dumps(target_data)
    return cls.objects.create(*args, **cleaned_kwargs)

  """
  Creates Snapshot and CSV
  """
  def send_error_email(self):
    try:
      snapshot_params = json.loads(self.target_data)
      error_to_addresses = snapshot_params.get('error_to_addresses', [])
      if not error_to_addresses:
        LOGGER.info('Snapshot failed. No email notification specified')
        return

      from_address = '{0} <{1}>'.format(self.schedule.owner.username, setting_get_single('email_host_user'))
      subject = snapshot_params.get('email_subject', '')

      ip = setting_get_single('site_thumbnails_ip')
      url = '{0}arc/apps/app/{1}?{2}'.format(ip, snapshot_params['app_id'], snapshot_params['paramString'])

      body = 'Attempted to snapshot {0}'.format(url)
      body += '\nUnable to send email. Please check job logs for details.'

      setup_email()
      msg = EmailMultiAlternatives(subject=subject,
                                   body=body,
                                   from_email=from_address,
                                   to=error_to_addresses)
      count = msg.send()
      LOGGER.info('Email sent. Count of messages: %s', (count))
    except Exception as e:
      LOGGER.exception('Failed to send email indicating an error with taking a snapshot')

  def run_job_target(self, run_parameters, joblog):
    joblog.add_task('Updating snapshot params')
    snapshot_params = json.loads(self.target_data)
    snapshot_params['outDir'] = tempfile.mkdtemp()
    # add predefined properties to params
    snapshot_params['saveType'] = 'screenshot'
    snapshot_params['forJobTarget'] = 1

    joblog.add_task('Generating Snapshots')
    mock_request = create_request(user=self.schedule.owner)
    runner = create_snapshotrunner(appId=snapshot_params['app_id'],
                              user=self.schedule.owner,
                              config=snapshot_params,
                              request=mock_request)

    if not runner:
      snapshot_params['error'] = "Cannot generate snapshots on this system. " + \
          "Configuration not complete."
      self.send_error_email()
      raise Exception('Error when generating snapshot')

    runner.run()
    run_parameters['image'] = snapshot_params['outDir']
    run_parameters['csv_names_to_filepath'] = snapshot_params.get('csv_names_to_filepath', {})
    run_parameters['xlsx_name_to_filepath'] = snapshot_params.get('xlsx_name_to_filepath')
    run_parameters['error'] = snapshot_params.get('error')
    joblog.update_log(last_task='Saving error and image into run_parameters',
                      target_data=snapshot_params,
                      target_order=self.target_order,
                      run_parameters=run_parameters)

    if snapshot_params.get('error'):
      self.send_error_email()
      raise Exception('Error when generating snapshot - %s' % snapshot_params.get('error'))

"""
EMAIL TARGET HELPER FUNCTIONS
"""
def _process_image(match, idx, attached_images):
  idx['count'] += 1
  image_text = match.group()
  groups = match.groups()
  full_src_text = groups[0]
  type = groups[1]
  src_text = groups[2]
  image_name = 'image_%s.%s' % (idx['count'], type)

  attached_images.append({
    'name': image_name,
    'data': src_text,
    'type': type,
  })

  return image_text.replace(full_src_text, 'cid:%s' % image_name)

def _process_html_images(template_value, attached_images):
  idx = { 'count': 0 } # need to store count in object so mutable
  image_finder = re.compile("<img .*?src=[\"'](data:image/(.+?);base64,(.+?))[\"'].*?>")
  return image_finder.sub(lambda match: _process_image(match, idx, attached_images),
                          template_value)

"""
EMAIL TARGET: Specific proxy class for email jobs
"""
class EmailJobTarget(JobTarget):

  IMAGE_TYPES = {
    'attachPNG': 'A',
    'attachPDF': 'D',
    'attachCSV': 'C',
    'attachXLSX': 'X',
    'embedded': 'E',
  }

  class Meta:
    ordering = ['target_order']
    proxy = True

  """
  Custom function for creating email
  """
  @classmethod
  def create_email(cls, *args, **kwargs):
    cleaned_kwargs = cls.job_target_properties(
        target_type=JobTarget.EMAIL,
        schedule=kwargs.get("schedule"),
        target_order=kwargs.get("target_order"),
        minimum_time_interval=kwargs.get("minimum_time_interval"))

    trigger_content_items = {
      'to_addresses': kwargs.get("to_addresses", []),
      'cc_addresses': kwargs.get("cc_addresses", []),
      'error_to_addresses': kwargs.get("error_to_addresses", []),
    }

    reply_to_address = kwargs.get("reply_to_address")
    if reply_to_address:
      trigger_content_items['reply_to_address'] = [reply_to_address]

    readonly_data = cls.get_readonly_target_data(kwargs)
    editable_data = cls.get_editable_target_data(kwargs)
    readonly_data.update(editable_data)
    target_data = readonly_data

    cleaned_kwargs["target_data"] = json.dumps(target_data)
    email = cls.objects.create(*args, **cleaned_kwargs)

    for key, value in list(trigger_content_items.items()):
      email._add_trigger_content(value, key)

    template_id = target_data.get('email_template')
    if template_id != '':
      CustomEmailTemplate.add_schedule(schedule_id=email.schedule.id, template_id=template_id)

    return email

  @classmethod
  def get_readonly_target_data(cls, arg):
    return {
      'app_id': arg.get("app_id"),
      'image_type': arg.get("image_type", cls.IMAGE_TYPES['embedded']),
      'visual_url': arg.get("visual_url"),
      'location_pathname': arg.get("location_pathname"),
    }

  @staticmethod
  def get_editable_target_data(arg):
    return {
      'email_subject': arg.get("email_subject", settings.DEFAULT_EMAIL_PREFIX + ' Visual Email'),
      'message': arg.get("message", ''),
      'to_addresses': arg.get("to_addresses", []),
      'cc_addresses': arg.get("cc_addresses", []),
      'error_to_addresses': arg.get("error_to_addresses", []),
      'reply_to_address': arg.get("reply_to_address", ''),
      'email_template': arg.get('email_template', ''),
      'url_params': arg.get('url_params', 'embed=true'),
    }

  def _replace_special_params(self, message, replacements, escape_text=False):
    # Place longer ones first to keep shorter substrings from matching where the longer ones should take place
    # For instance given the replacements {'ab': 'AB', 'abc': 'ABC'} against the string 'hey abc', it should produce
    # 'hey ABC' and not 'hey ABc'
    # from http://stackoverflow.com/questions/6116978/python-replace-multiple-strings#comment60793027_6117124
    substrs = sorted(replacements, key=len, reverse=True)

    parameter_delimiters = '\\<\\<{0}\\>\\>|\\&lt\\;\\&lt\\;{0}\\&gt\\;\\&gt\\;'
    escaped_regex = [parameter_delimiters.format(re.escape(str)) for str in substrs]
    regex_string = '|'.join(escaped_regex)

    # Create a big OR regex that matches any of the substrings to replace
    regexp = re.compile('<clouderaembed>[\s\S]*?</clouderaembed>|' + regex_string)

    # For each match, look up the new string in the replacements
    get_replacement = re.compile('\\<|\\>|\\&lt\\;|\\&gt\\;')
    image_replacement = replacements.get('clouderaembed', '')
    def replace_string(match):
      base_str = get_replacement.sub('', match.group(0))
      if base_str.startswith('clouderaembed'):
        return image_replacement
      else:
        replacement_string = replacements.get(base_str, '')
        if escape_text:
          try:
            replacement_string = html.escape(replacement_string, quote=True)
          except:
            pass

        return str(replacement_string)

    return regexp.sub(replace_string, message)

  def _parse_params(self, url_params):
    # params are in the format of embed=true&param.test=test
    url_parameters_split = []
    # Duplicate of param_name in apps.views._get_param_context
    param_name = 'param.'
    for url_param in url_params.split('&'):
      kv = url_param.split('=')
      k = urllib.parse.unquote(kv[0])
      if k.startswith(param_name):
        v = urllib.parse.unquote(kv[1])
        url_parameters_split.append({
          'key': k.replace(param_name, ''),
          'value': v,
        })
    return url_parameters_split

  """
  SENDS EMAIL:
  """
  def run_job_target(self, run_parameters, joblog):
    joblog.add_task('Loading email information')
    target_data = json.loads(self.target_data)

    joblog.add_task("Setting up email")
    setup_email()
    joblog.add_task("Assigning run_parameters values")
    run_parameters.update(target_data)
    email_host_user = setting_get_single('email_host_user')
    if email_host_user.endswith('>'):
      run_parameters["from_address"] = '{0}'.format(email_host_user)
    else:
      run_parameters["from_address"] = '{0} <{1}>'.format(self.schedule.owner.username, email_host_user)

    try:
      report = Report.objects.get(id=run_parameters["app_id"])
      if report.report_type == 'dashboard':
        name = report.report_name
      else:
        name = json.loads(report.report_data)["report_data"]["report_title"]
    except:
      name = run_parameters['image'].name

    # Replace whitespaces to make sure file name and cid name match for embedded
    run_parameters['name'] = re.sub('\s+', "_", name) or 'Untitled'

    image_type = run_parameters['image_type']

    # If snapshot was taken, save the image
    msg_img = {
      'pdf_img': '',
      'png_img': '',
    }
    is_pdf = image_type.count(self.IMAGE_TYPES['attachPDF'])
    is_embedded = (image_type.count(self.IMAGE_TYPES['embedded']))
    is_attach = image_type.count(self.IMAGE_TYPES['attachPNG'])
    outDir = run_parameters.get('image')
    if outDir:
      # Convert snapshot to PDF if PDF download type chosen
      outfile = os.path.join(outDir, 'output.format')
      if is_pdf:
        img = open(outfile, 'rb')
        pdf_name = run_parameters['name'] + '.pdf'
        pdf_img = MIMEApplication(img.read(), _subtype='pdf')
        pdf_img.add_header('Content-Disposition', 'attachment; filename="%s"' % pdf_name)
        msg_img['pdf_img'] = pdf_img
        # save image into joblog
        joblog.image.save(pdf_name, File(img), save=True)
        img.close()

      # add in image if embedded or attached
      if is_embedded or is_attach:
        embedded_name = run_parameters['name'] + '.png'
        img = open(os.path.join(outDir, '0.png'), 'rb')

        mime_image = MIMEImage(img.read(), _subtype='png')
        mime_image.add_header('Content-ID', '<{0}>'.format(embedded_name))
        msg_img['png_img'] = mime_image
        disposition_type = 'inline' if is_embedded else 'attachment'
        mime_image.add_header('Content-Disposition', disposition_type, filename=embedded_name)

        # save image into joblog
        joblog.image.save(embedded_name if embedded_name else "image", File(img), save=True)
        img.close()

    joblog.add_task("Updating message")

    run_parameters['start_time'] = joblog.start_time.strftime('%x %X %Z')
    run_parameters['run_time'] = str(joblog.end_time - joblog.start_time)
    if self.schedule.schedule_group:
      run_parameters['run_interval'] = self.schedule.schedule_group.schedule_name
      run_parameters['run_interval_cron'] = self.schedule.schedule_group.run_interval
      run_parameters['next_run_time'] = self.schedule.schedule_group.next_run_time.strftime('%x %X %Z')

    url_parameters = run_parameters.get('url_params')
    parameter_hash = {}
    if url_parameters:
      url_parm_keys = self._parse_params(url_parameters)
      parameters_for_string = []

      for param in url_parm_keys:
        key = param['key']
        value = param['value']
        parameter_hash[key] = value
        parameters_for_string.append('{}: {}'.format(key, value))

      url_parameters = '\n'.join(parameters_for_string)

    full_url = run_parameters.get('visual_url')
    relative_url = run_parameters.get('location_pathname')
    url_params = run_parameters.get('url_params')
    if url_params:
      full_url += '?' + url_params

      # Add query to relative_url if there's more than embed=true
      relative_query = re.sub('embed=true&?', '', url_params)
      if relative_query:
        relative_url += '?' + relative_query

    # In custom templates, escape filename especially since quotes can mess up image embedding
    clouderaembed = ''
    if is_embedded:
      clouderaembed = '<img src="cid:{0}" alt="{0}" style="max-width:600px;">'.format(embedded_name)

    special_params = {
      'trigger_conditions': run_parameters.get('trigger_conditions'),
      'trigger_fullconditional': run_parameters.get('trigger_fullconditional'),
      'trigger_data_value': run_parameters.get('trigger_data_value'),
      'trigger_difference': run_parameters.get('trigger_difference'),
      'triggered_time': run_parameters.get('triggered_time'),
      'trigger_id': run_parameters.get('trigger_id'),
      'trigger_name': run_parameters.get('trigger_name'),
      'run_interval': run_parameters.get('run_interval'),
      'run_interval_cron': run_parameters.get('run_interval_cron'),
      'next_run_time': run_parameters.get('next_run_time'),
      'minimum_time_interval': self.minimum_time_interval,
      'cloudera_user_id': self.schedule.owner.username,
      'app_id': run_parameters.get('app_id'),
      'app_name': name,
      'app_url': full_url,
      'relative_url': relative_url,
      'start_time': run_parameters.get('start_time'),
      'run_time': run_parameters.get('run_time'),
      'params': url_parameters,
      'clouderaembed': clouderaembed,
    }

    special_params.update(parameter_hash)

    run_parameters['message'] = self._replace_special_params(run_parameters["message"],
                                                            special_params)
    run_parameters['email_subject'] = self._replace_special_params(run_parameters["email_subject"],
                                                                  special_params)

    joblog.update_log(last_task="Getting email template", run_parameters=run_parameters)
    email_template = target_data.get('email_template')

    # Try to use custom template. Otherwise, use default template
    # If use custom templates, need to escape replacement text
    template_attachments = []
    try:
      template_object = CustomEmailTemplate.email_templates('template_value',
                                                            id=int(email_template),
                                                            no_json=True)[0]

      # Substitute body, adding in <br/> for newlines
      newline_regex = re.compile('\n|\r')
      escaped_message = html.escape(run_parameters["message"], quote=True)

      run_parameters['message'] = '<br/>' + newline_regex.sub('<br/>', escaped_message) + '<br/>'

      html_string = EMAIL_BODY_REGEX.sub(run_parameters['message'],
                                        template_object.get('template_value', ''))

      html_string =_process_html_images(html_string, template_attachments)

      html_content = self._replace_special_params(html_string,
                                                  special_params,
                                                  escape_text=True)
    except:
      html_content = loader.get_template('send_visual.html').render(run_parameters)
      html_dict = {}
      html_dict.update(special_params)
      html_dict.update(run_parameters)
      html_content = loader.get_template('send_visual.html').render(html_dict)

    reply_to = run_parameters.get('reply_to_address')
    reply_to = [ reply_to ] if reply_to else None
    msg = EmailMultiAlternatives(subject=run_parameters["email_subject"],
                                # Body needs to have non-empty text
                                #  in order for template to render properly
                                body=name + ':\n' + run_parameters['message'],
                                from_email=run_parameters["from_address"],
                                to=run_parameters["to_addresses"],
                                cc=run_parameters.get('cc_addresses'),
                                reply_to=reply_to)
    msg.attach_alternative(html_content, "text/html")

    joblog.add_task("Attaching image if necessary")
    for type, attachment_image in list(msg_img.items()):
      if attachment_image:
        msg.attach(attachment_image)

    joblog.add_task("Attaching csv if necessary")
    csv_names_to_filepath = run_parameters.get('csv_names_to_filepath')
    if csv_names_to_filepath:
      for real_name, temp_name in list(csv_names_to_filepath.items()):
        opened_file = open(temp_name)
        csv_attachment = MIMEApplication(opened_file.read())
        opened_file.close()
        csv_attachment.add_header('Content-Disposition', 'attachment', filename='%s.csv' % real_name)
        msg.attach(csv_attachment)
        os.remove(temp_name)

    joblog.add_task("Attaching xlsx if necessary")
    xlsx_name_to_filepath = run_parameters.get('xlsx_name_to_filepath')
    if xlsx_name_to_filepath:
      xlsx_attachment = MIMEBase('application', 'octet-stream')
      xlsx_attachment.set_payload(open(xlsx_name_to_filepath, 'rb').read())
      encoders.encode_base64(xlsx_attachment)
      if run_parameters['name'] == 'Untitled':
        filename = run_parameters['app_id']
      else:
        filename = run_parameters['name']
      xlsx_attachment.add_header('Content-Disposition', 'attachment', filename='%s.xlsx' % filename)
      msg.attach(xlsx_attachment)

    joblog.add_task("Attaching custom template images if necessary")
    for attachment in template_attachments:
      new_image_data = attachment.get('data', '')
      new_image_name = attachment.get('name', '')
      new_image_type = attachment.get('type', 'png')
      new_image = MIMEImage(base64.b64decode(new_image_data), _subtype=new_image_type)
      new_image.add_header('Content-ID', '<%s>' % new_image_name)
      new_image.add_header('Content-Disposition', 'inline',
                          filename=new_image_name)
      msg.attach(new_image)

    joblog.update_log(run_parameters=run_parameters, last_task="Sending email")
    status = msg.send()

    self.safely_remove_tmp_folder(pathToDir=outDir)

    return 'Email Successfully Sent: {0}'.format(status)

"""
REFRESH JOB TARGET: Specific proxy class for putting refreshes on the refresh queue,
and then monitoring them until completion
"""
class RefreshJobTarget(JobTarget):
  class Meta:
    ordering = ['target_order']
    proxy = True

  @classmethod
  def create_refresh_job(cls, *args, **kwargs):
    cleaned_kwargs = cls.job_target_properties(
        target_type=JobTarget.REFRESH,
        schedule=kwargs.get("schedule"),
        target_order=kwargs.get("target_order"),
        minimum_time_interval=kwargs.get("minimum_time_interval"))

    target_data = {
      'dcid': kwargs['dcid'],
      'view': kwargs['view'],
      'error_to_addresses': kwargs.get('error_to_addresses', []),
    }

    cleaned_kwargs["target_data"] = json.dumps(target_data)

    return cls.objects.create(*args, **cleaned_kwargs)

  def send_error_email(self):
    try:
      target_data = json.loads(self.target_data)
      view = target_data['view']
      error_to_addresses = target_data.get('error_to_addresses', [])
      if not error_to_addresses:
        LOGGER.info('View refresh failed. No email notification specified.')
        return

      from_address = '{0} <{1}>'.format(self.schedule.owner.username,
          setting_get_single('email_host_user'))
      subject = '[Error] Analytical View refresh failed'
      ip = setting_get_single('site_thumbnails_ip')
      joblogs_url = '{0}arc/jobs/adminlogs'.format(ip)
      body = 'Analytical view refresh failed for view {0}. Please check job logs for details. {1}' \
          .format(view, joblogs_url)

      setup_email()
      msg = EmailMultiAlternatives(subject=subject,
                                   body=body,
                                   from_email=from_address,
                                   to=error_to_addresses)
      count = msg.send()
      LOGGER.info('Email sent. Count of messages: %s', (count))
    except Exception as e:
      LOGGER.exception('Failed to send email indicating an error with refreshing analytical views')

  """
  Adds a new refresh target onto the queue
  """
  def run_job_target(self, run_parameters, joblog):
    joblog.add_task('Queueing up Analytical View Refresh')
    target_data = json.loads(self.target_data)
    view = target_data['view']

    refresh = _enque_refresh(self, joblog)

    LOGGER.info('ENQUEUED')
    pollSeconds = 2
    while True:
      done = True
      allFinished = True

      allStatus = refresh.status_list()
      LOGGER.info('STATUS %s', allStatus)

      for s in allStatus:
        statusStr = s.get('status', {}).get('status', '')
        LOGGER.info('CHECKING %s', statusStr)
        if statusStr != 'Finished':
          allFinished = False

        if statusStr.startswith('Running'):
          done = False
          break

      if not done:
        time.sleep(pollSeconds)
        continue

      # done is True
      if allFinished:
        msg = 'Successful Analytical View Refresh for view {0}'.format(view)
        LOGGER.info(msg)
        return msg

      # done is True
      # allFinished is False
      self.send_error_email()
      return 'Error refreshing AV'

"""
Define all JobTarget children
"""
JobTarget.CHILDREN = {
  JobTarget.SNAPSHOT: SnapshotJobTarget,
  JobTarget.EMAIL: EmailJobTarget,
  JobTarget.TRIGGER: TriggerJobTarget,
  JobTarget.REFRESH: RefreshJobTarget,
  JobTarget.SCRIPT: ScriptTarget,
  JobTarget.DATACOPY: DataCopyJobTarget,
}


"""
********SCHEDULE TARGET LOGS PARENT CLASS***************************************
Logs all the job_target events that have been run
"""
class JobLog(models.Model):
  class Meta:
    ordering = ['-end_time']
  start_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  end_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  schedule = models.ForeignKey('JobSchedule', on_delete=models.CASCADE)
  # archives target_data with a property that remembers run_parameters
  contents_archive = models.TextField(blank=True, default="")
  error_message = models.TextField(blank=True, default="")
  task_list = models.TextField(blank=True, default=json.dumps({'tasks': []}))
  last_completed_target = models.IntegerField(default=None, null=True, blank=True)
  status = models.CharField(max_length=1, choices=STATUSES, default=STATUS_RUNNING)
  image = models.FileField(upload_to="job/email", blank=True, null=True, default="")

  def update_log_owner(self, id:int, name:str):
    try:
      contents_archive = json.loads(self.contents_archive)
    except:
      LOGGER.error("Could not parse contents archive for job log")
      contents_archive = None


    if contents_archive:
      run_parameters = contents_archive.get('run_parameters', {})
      run_parameters['owner_id'] = id
      run_parameters['owner_username'] = name
      contents_archive['run_parameters'] = run_parameters

    self.contents_archive = json.dumps(contents_archive)
    self.save(update_fields=['contents_archive'])

  def update_log(self, last_task=None, target_data=None, target_order=None,
                 run_parameters=None, end_time=None, status=None, error_message=None,
                 contents_archive=None, last_completed_target=None, start_time=None):

    # Make target_order string because can't use number as key
    target_order = str(target_order)
    # refetch model from database each time or else we won't realize things are cancelled
    self = JobLog.objects.get(id=self.id)

    fields_to_save = []

    if last_task:
      # update task list
      self.add_task(task=last_task, save=False)
      fields_to_save.append('task_list')

    if end_time:
      self.end_time = datetime.now(self.start_time.tzinfo)
      fields_to_save.append('end_time')
      LOGGER.info('Schedule #%s Finished' % self.schedule.id)
      LOGGER.info(self.contents_archive)

    if start_time:
      self.start_time = datetime.now(self.start_time.tzinfo)
      fields_to_save.append('start_time')
      LOGGER.info('Schedule #%s Started' % self.schedule.id)
      LOGGER.debug(self.contents_archive)

    if error_message:
      self.error_message = error_message
      fields_to_save.append('error_message')

    if last_completed_target is not None:
      self.last_completed_target = last_completed_target
      fields_to_save.append('last_completed_target')

    if isinstance(contents_archive, str):
      contents_archive = json.loads(contents_archive)
    if isinstance(target_data, str):
      target_data = json.loads(target_data)
    if isinstance(run_parameters, str):
      run_parameters = json.loads(run_parameters)

    if contents_archive or run_parameters or target_data:
      old_contents = json.loads(self.contents_archive)

      if contents_archive:
        old_contents.update(contents_archive)
      if target_data and target_order != None:
        old_target_data = old_contents.get(target_order, {})
        old_target_data.update(target_data)
        old_contents[target_order] = old_target_data
      elif target_data:
        old_params = old_contents.get('run_parameters', {})
        old_params.update(target_data)
        old_contents['run_parameters'] = run_parameters
      if run_parameters:
        old_params = old_contents.get('run_parameters', {})
        old_params.update(run_parameters)
        old_contents['run_parameters'] = run_parameters

      self.contents_archive = json.dumps(old_contents)
      fields_to_save.append('contents_archive')

    self.save(update_fields=fields_to_save)

    # Test both schedule and log for cases where we don't update schedule because it's paused
    if self.schedule.last_status == STATUS_CANCELED or self.status == STATUS_CANCELED:
      raise CancelTask('Canceled')
    elif status:
      self.status = status
      self.save(update_fields=['status'])

  def get_last_task(self):
    last_task = ''
    task_list = json.loads(self.task_list)
    if task_list and len(task_list['tasks']):
      last_task = task_list['tasks'][-1]

    if isinstance(last_task, dict):
      return last_task.get('message')

    return last_task

  def add_task(self, task=None, save=True):
      LOGGER.info('Schedule #%s Last Run Task: %s' % (self.schedule.id, task))

      tasks = []
      if self.task_list != '':
        task_list = json.loads(self.task_list)

        if task_list:
          tasks = task_list['tasks']

      tasks.append({'message': task, 'date': str(datetime.now(self.start_time.tzinfo))})
      self.task_list = json.dumps({'tasks': tasks})

      if save:
        self.save(update_fields=['task_list'])

@receiver(pre_delete, sender=JobLog)
def delete_image(sender, instance, **kwargs):
  # Pass false so FileField doesn't save the model.
  if instance.image:
    instance.image.delete(False)


"""
****************TEXT HISTORY PARENT CLASS***************************************
User input from job_target for autocompletion purpose
"""
class JobContent(models.Model):
  # Create time
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  # Source trigger
  creator = models.ManyToManyField(User, default=1)
  # Type of field
  field = models.CharField(max_length=200, blank=True, default=None)
  # Value of the field (used for autocompleting)
  value = models.TextField(blank=True, default="")

"""
TEXT HISTORY: Emailer, for to/cc fields
field: cc_addresses and to_addresses
"""
class EmailJobContent(JobContent):
  class Meta:
    proxy = True

  @classmethod
  def all_addresses(cls, user):
    return list(set(user.jobcontent_set.filter(field__in=['to_addresses', 'cc_addresses']).values_list('value', flat=True)))

"""
EMAIL TEMPLATE: For storing email templates
field: custom_email_templates
value: {
  update_time (date)
  schedule_ids (array)
  template_name (string)
  template_description (string)
  template_value (string)
  template_default_message (html string)
}
"""
class CustomEmailTemplate(JobContent):
  class Meta:
    proxy = True

  """
  expect:
  CustomEmailTemplate.email_templates('id', 'update_time', 'schedule_ids', 'adhoc_schedule_ids',
                                      'template_name', 'template_description',
                                      'template_value', 'template_default_message',
                                      id=1, no_json=True)
  """
  @classmethod
  def email_templates(cls, *args, **kwargs):
    templates = cls.objects.filter(field='custom_email_templates')
    template_id = kwargs.get('id')
    if template_id:
      templates = templates.filter(id=template_id)
    elif template_id == 0:
      templates = []

    results = []

    for t in templates:
      template_info = json.loads(t.value)

      # If user passed in specific values, only return those
      # otherwise, return all
      if len(args):
        current_info = {}
        for key in args:
          try:
            current_info[key] = getattr(t, key)
          except:
            current_info[key] = template_info.get(key)

      else:
        current_info = {
          'id': t.id,
          'update_time': template_info.get('update_time'),
          'schedule_ids': template_info.get('schedule_ids', []),
          'adhoc_schedule_ids': template_info.get('adhoc_schedule_ids', []),
          'template_name': template_info.get('template_name'),
          'template_description': template_info.get('template_description'),
          'template_value': template_info.get('template_value'),
          'template_default_message': template_info.get('template_default_message'),
        }

      results.append(current_info)

    if kwargs.get('no_json'):
      return results
    else:
      return json.dumps(results)

  @classmethod
  def add_schedule(cls, schedule_id=None, template_id=None):
    template = cls.objects.filter(id=template_id)
    schedule = JobSchedule.objects.filter(id=schedule_id)
    if template.exists() and schedule.exists():
      schedule_type = 'adhoc_schedule_ids'
      if schedule.first().schedule_group:
        schedule_type = 'schedule_ids'

      template = template.first()
      template_info = json.loads(template.value)
      try:
        template_info[schedule_type].append(schedule_id)
      except:
        template_info[schedule_type] = [schedule_id]
      template.value = json.dumps(template_info)
      template.save()


# TODO: when we start allowing editing of schedules,
# add in check that schedule includes email target that uses template
@receiver(pre_save, sender=CustomEmailTemplate)
def cleanup_custom_email_template(sender, instance, **kwargs):
  # check that schedule with_ids still exist
  instance_json = json.loads(instance.value or '{}')
  instance_json['schedule_ids'] = [id for id in instance_json.get('schedule_ids', []) if JobSchedule.objects.filter(id=id).exists()]
  instance_json['adhoc_schedule_ids'] = [id for id in instance_json.get('adhoc_schedule_ids', []) if JobSchedule.objects.filter(id=id).exists()]

  template_value = instance_json.get('template_value', '')

  # store message
  found_value = EMAIL_BODY_REGEX.search(template_value)
  if found_value:
    instance_json['template_default_message'] = found_value.groups()[0].strip()
  else:
    instance_json['template_default_message'] = None

  instance_json['template_value'] = template_value
  instance.value = json.dumps(instance_json)
