# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from datetime import datetime, timedelta
from django.contrib.auth.models import User
import django.test
from django.test.client import Client
import json

from apps.tests.userrole import base
from cli.plugins import extract as extract_cli
from cli.plugins.tests import test_extract
from datasets.models import Extract
import jobs.models as models
from jobs.models import JobSchedule, JobTarget, JobLog, ScheduleGroup, JobScheduleRepeat
from jobs.tests import test_schedules
from jobs import views
from vizlib.tests.util import run_with_db_query_check
from extracts.tests.tests_extract_get import create_dataextract, create_role_with_permissions

import unittest.mock as mock
import django.db

def grant_all_dataconn_permissions(user: User, dc_id: int):
  create_role_with_permissions(user, [{"ptype": "dataconn", "dclist": [str(dc_id)], "perms": ["dc_aviews", "dc_upload", "dc_explore"]}])

def grant_all_dataset_permissions(user: User, dc_id: int, ds_id: int):
  create_role_with_permissions(user, [{"ptype": "dataset", "dcid": str(dc_id), "dslist": [str(ds_id)], "perms": ["ds_manage", "ds_appview", "ds_appedit"]}])

class TestJobsViews(django.test.TestCase):

  def create_extract(self, mock_fetchDataConn: mock.Mock):
    extractHelper = test_extract.TestExtract()
    extractHelper.setUp()

    dataset, dataconn = extractHelper.prepare(mock_fetchDataConn)
    create_args = [
      "refresh_full",
      f"{dataset.id}",
      f"{dataconn.id}",
      "extractsdb",
      "cdv_extract_table1",
      "[Channel Name]",
      "sum(1)",
    ]

    extract_count = len(Extract.objects.all())
    extract_cli.create(create_args)

    all_extracts = list(Extract.objects.all())
    self.assertEquals(1 + extract_count, len(all_extracts))
    extract = all_extracts[extract_count]

    return extract

  def assertOwner(self, expected_owner: str, extract: Extract, jobschedule: JobSchedule, jobtarget: JobTarget):

    extract_config = json.loads(extract.extract_config)
    self.assertEquals(extract_config["username"], expected_owner)

    self.assertEquals(jobschedule.owner.username, expected_owner)

    target_data = json.loads(jobtarget.target_data)
    self.assertEquals(target_data['username'], expected_owner)

  @mock.patch('extracts.misc.fetchDataConn')
  def test_update_owner(self, mock_fetchDataConn: mock.Mock):

    extract = self.create_extract(mock_fetchDataConn)

    schedule_group = test_schedules.TestSchedules().create_schedule_group()
    user1 = base.make_user('user1')

    client = Client()
    # admin1 is a user created by extract testing utility
    client.login(username='admin1', password='admin1')

    response = client.post('/jobs/scheduledatacopy', data={
      'extractId': extract.id,
      'scheduleId': schedule_group.id
    })

    response = json.loads(response.content.decode())
    schedule_id = response['schedule_id']
    jobtarget_id = response['jobtarget_id']

    jobschedule = JobSchedule.objects.get(id=schedule_id)
    jobtarget = JobTarget.objects.get(id=jobtarget_id)

    self.assertOwner('admin1', extract, jobschedule, jobtarget)

    edit_owner_url = f'/jobs/editjobowner/{schedule_id}'
    response = client.post(edit_owner_url, data={
      'is_bulk_change': 'false',
      'newowner_id': user1.id,
    })

    extract.refresh_from_db()
    jobschedule.refresh_from_db()
    jobtarget.refresh_from_db()

    self.assertOwner('user1', extract, jobschedule, jobtarget)

  @mock.patch('rbac.permcache.RBAC')
  def test_when_get_data_copy_logs_should_get_latest_log_of_extract(self, rbac):
    # Given
    rbac.get_user_obj_perms.return_value = ["ds_appview"]

    e = create_dataextract()
    sch = JobSchedule.objects.create(extract=e)
    JobLog.objects.create(schedule=sch, error_message='1')
    JobLog.objects.create(schedule=sch, error_message='2')
    user = base.make_user('user')
    client = base.create_client(user)
    # When
    response, _ = self.datacopylogs(client)
    # Then
    resp = response.json()[0]
    self.assertEqual(resp['error_message'], '2')

  def datacopylogs(self, client: django.test.client.Client, expected_query_count=None, **kwargs):
    result, query_count = run_with_db_query_check(
      lambda: client.get('/jobs/datacopylogs', data=kwargs),
      expected_query_count,
      self
      )
    return result, query_count

  @mock.patch('rbac.permcache.RBAC')
  def test_when_get_data_copy_logs_should_contain_last_successful_run(self, rbac):
    from django.utils import timezone
    # Given
    rbac.get_user_obj_perms.return_value = ["ds_appview"]

    e = create_dataextract()
    sch = JobSchedule.objects.create(extract=e)
    JobLog.objects.create(schedule=sch, error_message='1', status='C')
    log1 = JobLog.objects.create(schedule=sch, error_message='2', status='C')

    e2 = create_dataextract()
    sch2 = JobSchedule.objects.create(extract=e2)
    dt = timezone.now() + timedelta(hours=1)
    log2 = JobLog(schedule=sch2, error_message='11', status='C')
    log2.save()
    JobLog.objects.filter(id=log2.id).update(end_time=dt)
    tmp = JobLog(schedule=sch2, error_message='12', status='R')
    tmp.save()

    user = base.make_user('user')
    client = base.create_client(user)
    # When
    response, _ = self.datacopylogs(client)
    # Then
    resp = response.json()
    self.assertEqual(resp[0]['error_message'], '12')
    self.assertEqual(resp[0]['lastSuccessfulRun'], dt.strftime('%Y-%m-%d %H:%M:%S'))
    self.assertEqual(resp[1]['error_message'], '2')
    self.assertEqual(resp[1]['lastSuccessfulRun'], log1.end_time.strftime('%Y-%m-%d %H:%M:%S'))

  def test_when_data_copy_logs_with_no_permission_shoule_return_empyty(self):
    # Given
    e = create_dataextract()
    sch = JobSchedule.objects.create(extract=e)
    JobLog.objects.create(schedule=sch, error_message='1', status='C')
    JobLog.objects.create(schedule=sch, error_message='2', status='C')

    user = base.make_user('user')
    client = base.create_client(user)
    create_role_with_permissions(user, [])
    # When
    response, _ = self.datacopylogs(client, expected_query_count=5)

    # Then
    resp = response.json()
    self.assertEqual(len(resp), 0)

  def test_when_data_copy_logs_with_permission_shoule_return_only_jobs_have_dataset_with_dsappview_perm(self):
    # Given
    e = create_dataextract()
    sch = JobSchedule.objects.create(extract=e)
    JobLog.objects.create(schedule=sch, error_message='1', status='C')

    e2 = create_dataextract()
    sch2 = JobSchedule.objects.create(extract=e2)
    JobLog.objects.create(schedule=sch2, error_message='1', status='C')
    user = base.make_user('user')
    client = base.create_client(user)
    grant_all_dataset_permissions(user, dc_id=e.srcDataset.dataconnection.id, ds_id=e.srcDataset.id)
    # When
    response, _ = self.datacopylogs(client, expected_query_count=8)

    # Then
    resp = response.json()
    self.assertEqual(len(resp), 1)
    self.assertEqual(resp[0]['extractId'], e.id)

  def test_admin_access(self):
    # Given
    e = create_dataextract()
    sch = JobSchedule.objects.create(extract=e)
    JobLog.objects.create(schedule=sch, error_message='1', status='C')

    e2 = create_dataextract()
    sch2 = JobSchedule.objects.create(extract=e2)
    JobLog.objects.create(schedule=sch2, error_message='1', status='C')
    user = base.make_user('admin1', is_superuser=True)
    client = base.create_client(user)
    # When
    response, _ = self.datacopylogs(client, expected_query_count=5)

    # Then
    resp = response.json()
    self.assertEqual(len(resp), 2)

  def test_dataset_filtering_admin(self):
    user = base.make_user('admin1', is_superuser=True)
    client = base.create_client(user)

    self.do_test_dataset_filtering(client, 5)

  def test_dataset_filtering_user(self):
    user = base.make_user('user1', is_superuser=False)
    client = base.create_client(user)
    grant_all_dataset_permissions(user, dc_id=-1, ds_id=-1)


    self.do_test_dataset_filtering(client, 5 + 3) # dataset, user groups + rbac are loaded

  def do_test_dataset_filtering(self, client: django.test.Client, expected_query_count: int):
    # Given
    e = create_dataextract()
    sch = JobSchedule.objects.create(extract=e)
    JobLog.objects.create(schedule=sch, error_message='1', status='C')

    e2 = create_dataextract()
    sch2 = JobSchedule.objects.create(extract=e2)
    JobLog.objects.create(schedule=sch2, error_message='1', status='C')

    # When
    response, query_count = self.datacopylogs(client, expected_query_count, srcDatasetId=e2.srcDataset.id)

    # Then
    resp = response.json()
    self.assertEqual(len(resp), 1)
    self.assertEqual(resp[0]['extractId'], e2.id)

  def test_extract_filtering_admin(self):
    user = base.make_user('admin1', is_superuser=True)
    client = base.create_client(user)

    self.do_test_extract_filtering(client, 5)

  def test_extract_filtering_user(self):
    user = base.make_user('user1', is_superuser=False)
    client = base.create_client(user)
    grant_all_dataset_permissions(user, dc_id=-1, ds_id=-1)
    grant_all_dataconn_permissions(user, dc_id=-1)

    self.do_test_extract_filtering(client, 8)

  def do_test_extract_filtering(self, client: django.test.Client, expected_query_count: int):

    # Given
    e = create_dataextract()
    sch = JobSchedule.objects.create(extract=e)
    JobLog.objects.create(schedule=sch, error_message='1', status='C')

    e2 = create_dataextract()
    sch2 = JobSchedule.objects.create(extract=e2)
    JobLog.objects.create(schedule=sch2, error_message='1', status='C')

    # When
    response, _ = self.datacopylogs(client, expected_query_count, extractId=e2.id)

    # Then
    resp = response.json()
    self.assertEqual(len(resp), 1)
    self.assertEqual(resp[0]['extractId'], e2.id)

class TestJobsPage(django.test.TestCase):

  def create_job_schedule(self, schedule_group: models.ScheduleGroup, user: User):
    job = models.JobScheduleRepeat()
    job.schedule_group = schedule_group
    job.creator = user
    job.owner = user
    job.save()

    return job

  def create_job_log(self, job: models.JobScheduleRepeat):
    log = models.JobLog()
    log.schedule = job
    log.start_time = datetime.now()
    log.save()

    return log
  
  def create_email_target(self, job: models.JobScheduleRepeat):
    email_target = models.EmailJobTarget()
    email_target.schedule = job
    email_target.target_type = 'E'

    target_data = {
      'app_id': 1, 
      'include_csv': 0, 
      'include_xlsx': 0, 
      'csv_download_rows': 5000,
      'dbWidth': 100, 
      'dbHeight': 100, 
      'email_subject': 'Subject', 
      'error_to_addresses': [], 
      'reply_to_address': '', 
      'paramString': 'embed=true&sheet=1', 
      'uploadFormat': 'unknown'
    }
    email_target.target_data = json.dumps(target_data)
    email_target.save()
    return email_target

  def create_job_schedule_and_log(self, schedule_group: models.ScheduleGroup, user: User):
    job  = self.create_job_schedule(schedule_group, user)
    self.create_job_log(job)

    return job

  def create_multiple_logs(self, job: models.JobScheduleRepeat):
    logs = []

    for _ in range(20):
      log = self.create_job_log(job)
      logs.append(log)

    return logs

  def test_get_adminlogs(self):

    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)

    url = '/jobs/adminlogs'

    response = admin1_client.get(url)
    self.assertEquals(response.status_code, 200)

  def test_get_job_detail_page(self):

    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)
    job = self.create_job_schedule_and_log(None, admin1)

    url = f'/jobs/job/{job.id}'

    response = admin1_client.get(url)
    self.assertEquals(response.status_code, 200)

  def test_get_job(self):
    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)

    groups = models.ScheduleGroup.objects.all()

    self.create_job_schedule_and_log(groups[1], admin1)
    job_schedule = self.create_job_schedule_and_log(None, admin1)

    url = f'/jobs/api/jobs/{job_schedule.id}'

    # test retrieval for regular job
    response = admin1_client.get(url)
    self.assertEqual(response.status_code, 200)

    # test retrieval for job with missing data
    job_schedule.owner = None
    job_schedule.creator = None
    job_schedule.save()
    response = admin1_client.get(url)
    self.assertEqual(response.status_code, 200)

    # test job not found
    not_found_url = '/jobs/api/jobs/123456789'
    response = admin1_client.get(not_found_url)
    self.assertEqual(response.status_code, 404)

  def test_get_jobs(self):

    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)

    url = '/jobs/api/jobs?is_scheduled=true' # collects joblog and schedule data

    groups = models.ScheduleGroup.objects.all()
    self.create_job_schedule_and_log(groups[1], admin1)

    response, call_count_1 = run_with_db_query_check(lambda: admin1_client.get(url))

    # new schedule for the same schedule group:
    self.create_job_schedule_and_log(groups[1], admin1)

    response, call_count_2 = run_with_db_query_check(lambda: admin1_client.get(url))
    response_json = response.json()
    self.assertEquals(len((response_json['jobs'])),2)

    # the old method requires 1 extra query per job schedule
    query_increment = 0

    # adding a new schedule results in additional db query:
    self.assertEquals(call_count_2, call_count_1 + query_increment, f'{query_increment} additional DB query')

    # schedule for another schedule group:
    j1 = self.create_job_schedule_and_log(groups[2], admin1)

    response, call_count_3 = run_with_db_query_check(lambda: admin1_client.get(url))
    response_json = response.json()
    self.assertEquals(len((response_json['jobs'])),3)

    self.assertEquals(call_count_3, call_count_2 + query_increment, f'{query_increment} additional DB query')

    # jobs do not have the schedule group embedded, they only have the ID:
    for schedule in response_json['jobs']:
      sch_id = schedule['schedule_id']
      self.assertIsNone(schedule.get('schedule_group'), f'schedule group for schedule {sch_id}')
      self.assertIsNotNone(schedule['schedule_group_id'], f'schedule group id for schedule {sch_id}')

    # ad-hoc job
    self.create_job_schedule_and_log(None, admin1)

    response, call_count_4 = run_with_db_query_check(lambda: admin1_client.get(url))
    response_json = response.json()
    self.assertEquals(len((response_json['jobs'])),3)

    # new ad-hoc Job log: extra query
    self.assertEquals(call_count_4, call_count_3 + query_increment, f'{query_increment} additional DB query')
    self._test_non_admin_user()

  def test_get_jobs_for_nonadmin(self):
    self._test_non_admin_user()

  def _test_non_admin_user(self):
    user1 = base.make_user('user1', is_superuser=False)
    user1_client = base.create_client(user1)

    url = '/jobs/api/jobs?is_scheduled=true' # collects the same data from the db

    response, _ = run_with_db_query_check(lambda: user1_client.get(url))
    response_json = response.json()

    # non-admin does not see the logs created by other users
    self.assertEquals(len((response_json['jobs'])),0)

    # scheduled job
    groups = models.ScheduleGroup.objects.all()
    self.create_job_schedule_and_log(groups[3], user1)

    # ad-hoc job
    self.create_job_schedule_and_log(None, user1)

    response, _ = run_with_db_query_check(lambda: user1_client.get(url))
    response_json = response.json()
    # owned jobs are visible
    self.assertEquals(len((response_json['jobs'])),1)

  def test_get_job_status(self):
    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)

    groups = models.ScheduleGroup.objects.all()

    self.create_job_schedule_and_log(groups[1], admin1)
    job_schedule = self.create_job_schedule_and_log(None, admin1)

    url = f'/jobs/api/jobs/{job_schedule.id}/status'

    response = admin1_client.get(url)
    self.assertEqual(response.status_code, 200)

  def test_get_job_logs(self):
    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)

    groups = models.ScheduleGroup.objects.all()

    self.create_job_schedule_and_log(groups[1], admin1)
    job_schedule = self.create_job_schedule_and_log(None, admin1)

    url = f'/jobs/api/jobs/{job_schedule.id}/logs'

    response = admin1_client.get(url)
    self.assertEqual(response.status_code, 200)

    # test not found
    not_found_url = '/jobs/api/jobs/123456789/logs'
    response = admin1_client.get(not_found_url)
    self.assertEqual(response.status_code, 404)

    # test pagination
    job_schedule2 = self.create_job_schedule(None, admin1)
    logs = self.create_multiple_logs(job_schedule2)

    response1, _ = run_with_db_query_check(
    lambda: admin1_client.get(f'/jobs/api/jobs/{job_schedule2.id}/logs', {
      'page': 2,
      'page_size': 10,
    }))

    self.assertEquals(response1.status_code, 200)
    page1 = response1.json()
    self.assertEquals(len(page1['logs']), 10)
    self.assertEquals(page1['total'], len(logs))

  def test_get_email_templates(self):
    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)

    url = '/jobs/emailtemplates'
    response = admin1_client.get(url)
    self.assertEqual(response.status_code, 200)

  def test_job_entitlement(self):
    user1 = base.make_user('user1', is_superuser=False)
    user1_client = base.create_client(user1)

    user2 = base.make_user('user2', is_superuser=False)
    user2_client = base.create_client(user2)

    groups = models.ScheduleGroup.objects.all()

    self.create_job_schedule_and_log(groups[1], user1)
    job_schedule = self.create_job_schedule_and_log(None, user1)

    job_url = f'/jobs/api/jobs/{job_schedule.id}'
    job_logs_url = f'/jobs/api/jobs/{job_schedule.id}/logs'
    job_status_url = f'/jobs/api/jobs/{job_schedule.id}/status'

    # the owner should be able to retrieve the job
    response = user1_client.get(job_url)
    self.assertEqual(response.status_code, 200)

    # the other regular user should not be able to retrieve the job
    response = user2_client.get(job_url)
    self.assertEqual(response.status_code, 403)

    # the owner should be able to retrieve the job status
    response = user1_client.get(job_status_url)
    self.assertEqual(response.status_code, 200)

    # the other regular user should not be able to retrieve the job status
    response = user2_client.get(job_status_url)
    self.assertEqual(response.status_code, 403)

    # the owner should be able to retrieve the job logs
    response = user1_client.get(job_logs_url)
    self.assertEqual(response.status_code, 200)

    # the other regular user should not be able to retrieve the job logs
    response = user2_client.get(job_logs_url)
    self.assertEqual(response.status_code, 403)

  def test_get_latest_logs(self):
    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)
    groups = models.ScheduleGroup.objects.all()
    self.create_job_schedule_and_log(groups[1], admin1)

    url = f'/jobs/api/logs/latest'
    response1, queries = run_with_db_query_check(lambda: admin1_client.get(url))

    self.assertEquals(response1.status_code, 200)
    self.assertEquals(queries, 7)

    # the data contains the details of a log and the related job
    latest_logs = response1.json()
    data = latest_logs.get('data')[0]
    self.assertIsNotNone(data.get('log'))
    self.assertIsNotNone(data.get('job'))

  def test_edit_job(self):
    admin1 = base.make_user('admin1', is_superuser=True)
    client = base.create_client(admin1)
    groups = models.ScheduleGroup.objects.all()

    self.create_job_schedule_and_log(groups[1], admin1)
    job_schedule = self.create_job_schedule_and_log(None, admin1)
    email_target = self.create_email_target(job_schedule)

    url = f'/jobs/editjob/{job_schedule.id}'
    edited = json.dumps({
      'job_name': 'Edited job',
      'schedule_group': 1,
      'email_subject': 'Edited subject',
      'message': 'Edited content',
      'to_addresses':['saddress@dummy.com'],
      'cc_addresses':['arcadia.qe@yahoo.com'],
      'error_to_addresses': [],
      'reply_to_address': '',
      'url_params': 'embed=true&sheet=1', 
    })
   
    response, count = run_with_db_query_check(lambda: client.post(url, data={'emailFormData': edited}))

    self.assertEqual(response.status_code, 200)
    self.assertEqual(count, 10)

    job_schedule.refresh_from_db()
    email_target.refresh_from_db()
    target_data = json.loads(email_target.target_data)
    self.assertEqual(job_schedule.job_name, 'Edited job')
    self.assertEqual(job_schedule.schedule_group.id, 1)
    self.assertEqual(target_data['email_subject'], 'Edited subject')
    self.assertEqual(target_data['to_addresses'], ['saddress@dummy.com'])

  def test_edit_job_with_new_schedule(self):
    admin1 = base.make_user('admin1', is_superuser=True)
    client = base.create_client(admin1)
    groups = models.ScheduleGroup.objects.all()

    self.create_job_schedule_and_log(groups[1], admin1)
    job_schedule = self.create_job_schedule_and_log(None, admin1)
    self.create_email_target(job_schedule)

    url = f'/jobs/editjob/{job_schedule.id}'
    edited = json.dumps({
      'job_name': 'Edited job',
      'email_subject': 'Edited subject',
      'message': 'Edited content',
      'to_addresses':['saddress@dummy.com'],
      'cc_addresses':['arcadia.qe@yahoo.com'],
      'error_to_addresses': [],
      'reply_to_address': '',
      'url_params': 'embed=true&sheet=1', 
      'schedule_group': -2,
      'schedule_group_name': 'New schedule group',
      'run_interval': '1 1 1 1 1'
    })
   
    response, count = run_with_db_query_check(lambda: client.post(url, data={'emailFormData': edited}))

    self.assertEqual(response.status_code, 200)
    self.assertEqual(count, 11)
    job_schedule.refresh_from_db()
    self.assertEqual(job_schedule.schedule_group.schedule_name, 'New schedule group')
    self.assertEqual(job_schedule.schedule_group.run_interval, '1 1 1 1 1')


class TestJobsFilters(django.test.TestCase):
  def create_admin_user_and_its_client(self, username: str):
    self.admin_usr = base.make_user(username, is_superuser=True)
    self.admin_client = base.create_client(self.admin_usr)

  def create_job_schedule(self, **parameters: dict):
    return JobScheduleRepeat.objects.create(**parameters)

  def update_job_schedule_with_a_trigger(self, job_schedule, **trigger_parameters):
    trigger_target = JobTarget.objects.create(**trigger_parameters)
    job_schedule.trigger_target = trigger_target
    job_schedule.save()

  def setUp(self):
    self.create_admin_user_and_its_client("admin1")

    schedule_group = ScheduleGroup.objects.filter(id=1).first()
    common_parameters = {
          'creator': self.admin_usr,
          'owner': self.admin_usr,
          'schedule_group': schedule_group,
        }
    self.job_with_threshold = self.create_job_schedule(job_name="test_job_with_threshold", **common_parameters)
    self.job_without_threshold = self.create_job_schedule(job_name="test_job_without_threshold", **common_parameters)

    trigger_parameters = {
      'target_type': JobTarget.TRIGGER,
      'schedule': self.job_with_threshold,
      'target_data': ""
    }
    self.update_job_schedule_with_a_trigger(self.job_with_threshold, **trigger_parameters)

  def test_return_all_jobs_without_filters(self):
    url = f'/jobs/api/jobs'
    response = self.admin_client.get(url)
    jobs_resp = json.loads(response.content).get("jobs")
    self.assertEquals(len(jobs_resp), 2)

  def test_threshold_only_filter_works_when_true(self):
    url = f'/jobs/api/jobs?threshold_only=true'
    response = self.admin_client.get(url)
    jobs_resp = json.loads(response.content).get("jobs")
    job_name_resp = jobs_resp[0].get("job_name")
    self.assertEquals(len(jobs_resp), 1)
    self.assertEquals(job_name_resp, self.job_with_threshold.job_name)

  def test_threshold_only_filter_works_when_false(self):
    url = f'/jobs/api/jobs?threshold_only=false'
    response = self.admin_client.get(url)
    jobs_resp = json.loads(response.content).get("jobs")
    self.assertEquals(len(jobs_resp), 2)
