# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from datetime import datetime
from django.contrib.auth.models import User
import django.test

from jobs.models import STATUS_COMPLETED, JobLog

from apps.tests.userrole import base
import jobs.models as models
import json


class TestLogs(django.test.TestCase):
  def _create_schedule(self, schedule_group: models.ScheduleGroup, user: User):
    params = {
      'creator': user,
      'owner': user,
      'schedule_group': schedule_group,
      'job_name': 'Test job',
    }
    schedule = models.JobScheduleRepeat.objects.create(**params)
    return schedule
  
  def _create_log(self, schedule: models.JobScheduleRepeat):
    run_parameters = {
      'skip_next_targets': True,
      'force_run': True,
      'creator_id': self.user.id,
      'creator_username': self.user.username,
      'owner_id': self.user.id,
      'owner_username': self.user.username,
    }
    contents_archive = {
      'run_parameters': run_parameters,
    }
  
    log = JobLog()
    log.schedule = schedule
    log.start_time = datetime.now()
    log.task_list = json.dumps({'tasks': [{'message': 'Starting Task'}]})
    log.contents_archive = json.dumps(contents_archive)
    log.save()

    return log

  def _prepare(self):
    admin = base.make_user('admin1', is_superuser=True)
    self.admin = admin

    user = base.make_user('user', is_superuser=True)
    self.user = user
    groups = models.ScheduleGroup.objects.all()

    schedule = self._create_schedule(groups[1], admin)
    self.schedule = schedule
    schedule.save()
  
  def test_update_log(self):
    self._prepare()
    log = self._create_log(schedule=self.schedule)

    # build the new contents archive
    run_parameters = {
      'skip_next_targets': False,
      'force_run': False,
      'creator_id': self.admin.id,
      'creator_username': self.admin.username,
      'owner_id': self.admin.id,
      'owner_username': self.admin.username,
      'editor_id': self.user.id,
      'editor_username': self.user.username,
    }
    contents_archive = {
      'run_parameters': run_parameters,
    }
  
    log.update_log(last_task='Finished', status=STATUS_COMPLETED, end_time=datetime.now(), 
                   contents_archive=json.dumps(contents_archive))
    log.refresh_from_db()
    
    self.assertEquals(log.status, STATUS_COMPLETED)
    self.assertEquals(log.get_last_task(), 'Finished')

    updated_run_parameters = json.loads(log.contents_archive)['run_parameters']
    if run_parameters:
      self.assertEquals(updated_run_parameters['creator_username'], self.admin.username)
      self.assertEquals(updated_run_parameters['editor_username'], self.user.username)
      self.assertEquals(updated_run_parameters['skip_next_targets'], False)
      self.assertEquals(updated_run_parameters['force_run'], False)

  def test_update_log_owner(self):
    self._prepare()
    log = self._create_log(schedule=self.schedule)

    log.update_log_owner(self.admin.id, self.admin.username)
    log.refresh_from_db()

    updated_run_parameters = json.loads(log.contents_archive)['run_parameters']
    if updated_run_parameters:
      self.assertEquals(updated_run_parameters['owner_username'], self.admin.username)
      self.assertEquals(updated_run_parameters['owner_id'], self.admin.id)
  
  def test_add_task(self):
    self._prepare()
    log = self._create_log(schedule=self.schedule)
  
    log.add_task('Finished')
    log.refresh_from_db()
    
    self.assertEquals(len(json.loads(log.task_list)['tasks']), 2)
    self.assertEquals(log.get_last_task(), 'Finished')

  def test_get_last_task(self):
    self._prepare()
    log = self._create_log(schedule=self.schedule)
    
    self.assertEquals(log.get_last_task(), 'Starting Task')
