# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from datetime import datetime, timedelta
from django.contrib.auth.models import User
import django.test

import jobs.models
from jobs.models import ScheduleGroup, JobScheduleRepeat
from jobs.view_schedule_checker import ScheduleChecker
from reports.views_settings import setting_update_single

from apps.tests.userrole import base
import unittest.mock as mock
from unittest.mock import patch
from jobs.models import JobSchedule
import json
from django.conf import settings


class TestSchedules(django.test.TestCase):
  def create_schedule(self, schedule_group: ScheduleGroup, user: User, job_name: str):
    params = {
      'creator': user,
      'owner': user,
      'schedule_group': schedule_group,
      'job_name': job_name,
    }
    schedule = JobScheduleRepeat.objects.create(**params)
    return schedule

  def create_schedule_group(self):
    sg_10 = ScheduleGroup(
      schedule_name='every 10 minutes', run_interval='0,10,20,30,40,50 * * * *'
    )
    sg_10.save()
    return sg_10

  def prepare(self):
    '''
    Create a schedule that runs every 10 minutes.
    Create two users and one job for each of them.
    Manipulate times so that the jobs will be executed on the first iteration of ScheduleChecker.
    '''

    # switch off ScheduleChecker thread to avoid interference
    setting_update_single('scheduling_enabled', 'false', 0)
    schedule_checker = ScheduleChecker()
    self.schedule_checker = schedule_checker
    self.assertIsNone(getattr(schedule_checker, 'interval', None))

    sg_10 = self.create_schedule_group()

    user1 = base.make_user('user1')
    self.user1 = user1
    user2 = base.make_user('user2')
    self.user2 = user2
    s1 = self.create_schedule(sg_10, user1, job_name='schedule 1')
    s2 = self.create_schedule(sg_10, user2, job_name='schedule 2')

    # modify dates so that execution is triggered
    now = datetime.now(sg_10.create_time.tzinfo)
    now_11_min = now - timedelta(minutes=11)
    sg_10.next_run_time = now_11_min
    s1.create_time = now_11_min
    s2.create_time = now_11_min
    sg_10.save()
    s1.save()
    s2.save()

    # create the worker pools so that we can spy on them
    s1.init_job_runner()

  def test_normal_operation(self):
    self.prepare()

    with mock.patch.object(
      jobs.models._JOB_RUNNER_FAST, 'queue_untracked'
    ) as runner_spy:
      self.schedule_checker.check_and_run_schedules()
      # both jobs were queued:
      self.assertEquals(runner_spy.call_count, 2)

  def test_user_of_first_job_removed(self):
    self.prepare()

    user1 = self.user1
    user1.delete()
    user1.save()

    with mock.patch.object(
      jobs.models._JOB_RUNNER_FAST, 'queue_untracked'
    ) as runner_spy:
      self.schedule_checker.check_and_run_schedules()
      # the second job should be queued
      self.assertEquals(runner_spy.call_count, 1)

  def create_dataextract(self):
    from datasets.models import Extract, DataSet, DataConnection
    from util.ff_flipper import FeatureFlipper
    FeatureFlipper().set_flipper('site_enable_extracts', True, 'DB')
    dc = DataConnection.objects.create()
    ds = DataSet.objects.create(dataconnection=dc, dataset_name='name', dataset_detail='detail', dataset_type='TYPE')
    Extract.objects.create(srcDataset=ds, tgtDataconnection=dc, extract_config={})

  @patch('jobs.views.run_schedule')
  @patch('extracts.views.checkPerm')
  def test_data_copy_schedule_with_manual_extracts_should_contain_extract_id(self, check_perm, run):
    # Given
    user = base.make_user('user')
    client = base.create_client(user)
    self.create_dataextract()
    # When
    response = client.post('/jobs/scheduledatacopy', {'extractId': '1'})
    # Then
    self.assertEquals(response.status_code, 200)
    self.assertEquals(JobSchedule.objects.first().extract.id, 1)

  @patch('extracts.views.checkPerm')
  def test_data_copy_schedule_with_recurring_extracts_should_contain_extract_id(self, check_perm):
    # Given
    user = base.make_user('user')
    client = base.create_client(user)
    g = self.create_schedule_group()
    self.create_dataextract()
    # When
    response = client.post('/jobs/scheduledatacopy', {'extractId': '1', 'scheduleId': g.id})
    # Then
    self.assertEquals(response.status_code, 200)
    self.assertEquals(JobSchedule.objects.first().extract.id, 1)

  @patch('vizlib.workq._WorkQThread')
  def test_email_snapshot_now_should_create_schedule_with_null_extract_id(self, run):
    # Given
    setting_update_single('email_enabled', 'true', False)
    user = base.make_user('user')
    client = base.create_client(user)
    # When
    formValues = json.dumps({"visual_url": "ok",
                             "email_subject": 1,
                             "image_type": ""})
    response = client.post('/jobs/email/22',
                           {'formType': 'emailnow',
                            'emailFormData': formValues})
    # Then
    self.assertEquals(response.status_code, 200)
    self.assertEquals(JobSchedule.objects.first().job_name, 'Adhoc Email: 1')
    self.assertEquals(JobSchedule.objects.first().extract, None)

  @patch('vizlib.workq._WorkQThread')
  @patch('jobs.views.time')
  def test_script_schedule_with_no_schedule_should_create_schedule_with_null_extract_id(self, time, run):
    # Given
    settings.ENABLE_SQL_SCHEDULER = True
    user = base.make_user('user')
    client = base.create_client(user)
    time.time.return_value = 1
    # When
    response = client.post('/jobs/schedulescript', {})
    # Then
    self.assertEquals(response.status_code, 200)
    self.assertEquals(JobSchedule.objects.first().job_name, 'Adhoc sql: 1')
    self.assertEquals(JobSchedule.objects.first().extract, None)

  @patch('vizlib.workq._WorkQThread')
  def test_viewrefresh_enque_should_create_schedule_with_null_extract_id(self, run):
    # Given
    user = base.make_user('user')
    client = base.create_client(user)
    # When
    formValues = json.dumps({'view': 1, 'dcid': 1})
    post = {"opcode": "enque", "formValues": formValues}
    response = client.post('/jobs/viewrefresh', post)
    # Then
    self.assertEquals(response.status_code, 200)
    self.assertEquals(JobSchedule.objects.first().job_name, 'Adhoc Refresh: 1')
    self.assertEquals(JobSchedule.objects.first().extract, None)
