# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from apps.tests.userrole import base
from django.db.transaction import Atomic
from jobs.models import JobSchedule
from reports.models import Report
from reports.views_settings import setting_update_single
from unittest import mock

import django.test
import jobs.models
import json
import logging
import time

LOG = logging.getLogger(f'arcweb.{__name__}')

old_Atomic__exit__ = Atomic.__exit__


def delayed_Atomic__exit(self, exc_type, exc_value, traceback):
  """
  Slows down transaction commit by 2 seconds
  """
  LOG.info('ATOMIC EXIT sleep start')
  time.sleep(2)
  LOG.info('ATOMIC EXIT sleep end')
  return old_Atomic__exit__(self, exc_type, exc_value, traceback)


class TestViews(django.test.TransactionTestCase):
  @classmethod
  def setUpClass(cls):
    # make sure runners are real:
    jobs.models._JOB_RUNNER_FAST = None
    jobs.models._JOB_RUNNER_SLOW = None

  @classmethod
  def tearDownClass(cls):
    # recreate the test DB because with TransactionTestCase all tables get truncated after each test case:
    from django.db import connection
    connection._connections['default'].creation.create_test_db(
      autoclobber=True)

    # clear the cache as it can disturb other tests
    # (visual_maxrows gets cached)
    from django.core.cache import cache
    cache.clear()

  def _create_visual(self, report_type: str, report_data: dict = None):
    if report_data is None:
      report_data = {
        'report_data': {
          'report_title': f'visual_{report_type}'
        }
      }

    r = Report(report_type=report_type, report_data=json.dumps(report_data))
    r.save()
    return r

  def wait_for_completion(self, schedule: JobSchedule, timeout=5):
    LOG.info('waiting for schedule to complete')
    start = time.time()
    while schedule.joblog_set.count() == 0 or schedule.joblog_set.first().status != jobs.models.STATUS_COMPLETED:
      if (time.time() - start) > timeout:
        self.fail('Timed out waiting for schedule to complete')
      LOG.info('sleeping to allow schedule to complete')
      time.sleep(0.5)

    LOG.info('schedule completed')

  def test_emailnow_with_slow_database(self):
    """
    Checks that the email processing background thread does its job even when
    committing the database changes takes long. The test mimics this by artificially
    slowing down the Atomic context manager.
    """
    setting_update_single('email_enabled', 'true', False)
    user = base.make_user('user')
    client = base.create_client(user)

    self.assertEqual(0, JobSchedule.objects.count())

    table_visual = self._create_visual(report_type='table')

    # patch Django transaction context manager so that it commits later
    with mock.patch.object(Atomic, '__exit__', new=delayed_Atomic__exit):

      formValues = json.dumps({"visual_url": "ok",
                              "email_subject": 'test email',
                               "image_type": ""})
      response = client.post(f'/jobs/email/{table_visual.id}',
                             {'formType': 'emailnow',
                              'emailFormData': formValues})

    self.assertEquals(response.status_code, 200)

    self.assertEqual(1, JobSchedule.objects.count())
    schedule = JobSchedule.objects.first()
    self.assertEqual(schedule.job_name, 'Adhoc Email: test email')

    # will fail the test on timeout (i.e. if the job does not complete)
    self.wait_for_completion(schedule)
