# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging
import time
from jobs.models import ScheduleGroup, _SCHEDULER_ENABLED
from django.db import close_old_connections
from datetime import datetime
from reports.views_settings import setting_get_single
from threading import Thread

LOGGER = logging.getLogger("arcweb")

class ScheduleChecker(object):
  """ Threaded Schedule Checker
  This will create a thread that checks for schedules that need to be run
  """
  SCHEDULING_ENABLED = False

  def __init__(self, interval=300):
    """ Constructor
    :type interval: int
    :param interval: Check interval, in seconds
    """
    try:
      scheduling_enabled = setting_get_single("scheduling_enabled")
      ScheduleChecker.SCHEDULING_ENABLED = scheduling_enabled == 'true' or scheduling_enabled == None 
    except:
      # Exception should occur only when running builds to create instant packages
      pass

    if _SCHEDULER_ENABLED and self.SCHEDULING_ENABLED:
      self.interval = interval
      thread = Thread(target=self.run, args=(), name="ScheduleChecker")
      thread.daemon = True                            # Daemonize thread
      thread.start()                                  # Start the execution

  def run(self):
    """ Method that runs forever """
    while _SCHEDULER_ENABLED and self.SCHEDULING_ENABLED:
      # Wait before running check so that doesn't interfere with page load
      time.sleep(self.interval)
      try:
        close_old_connections()
        self.check_and_run_schedules()
      except Exception as e:
        # If there is an exception log it, but continue running the
        # ScheduleChecker Thread
        LOGGER.exception("ScheduleChecker caught exception.")

  def check_and_run_schedules(self):
    if not ScheduleGroup.objects.first():
      return

    # Remember when the check and run was intialized
    tzinfo = ScheduleGroup.objects.first().create_time.tzinfo
    now = datetime.now(tzinfo)
    # Update the next_run_times for groups that haven't run
    new_schedule_groups = ScheduleGroup.objects.filter(next_run_time__isnull=True)
    for schedule_group in new_schedule_groups:
      schedule_group.update_next_run_time()
    # Run only the repeating jobs that should be run
    groups_to_run = ScheduleGroup.objects.filter(next_run_time__lte=now)
    for schedule_group in groups_to_run:
      schedule_group.update_next_run_time(now)
      schedule_group.run_schedule_group()


