# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.shortcuts import render
from django.db import transaction
from django.core.paginator import Paginator
from django.core.exceptions import BadRequest, PermissionDenied
from extracts.views import extract
from jobs.models import JobSchedule, JobScheduleNow, \
    JobScheduleRepeat, EmailJobTarget, JobTarget, \
    SnapshotJobTarget, ScheduleGroup, TriggerJobTarget, \
    RefreshJobTarget, JobLog, ScriptTarget, DataCopyJobTarget, \
    STATUS_MAPS, User
from jobs.views_helpers import setup_email, email_self, CancelTask
from apps.utils import _appflags
from jobs.models import STATUS_PAUSED, STATUS_CANCELED, STATUS_COMPLETED, \
      _SCHEDULER_ENABLED, CustomEmailTemplate, EmailJobContent
from jobs.views_refresh import _fetch_refresh_status
from jobs.views_schedulegroups import schedule_group_get, user_can_schedule
from datasets.models import DataSet, DataConnection, Extract
from vizlib.querylog import logAction, ComponentType, ActionType
from datetime import datetime
from apps.templatetags.site_setting_name import site_setting_name
from util.ff_flipper import FeatureFlipper
from util import perm

from django.core import serializers
from django.db.models import Q, Max
from django.views.decorators.csrf import csrf_exempt
from vizlib.misc import check_login_or_use_apikey
from django.urls import reverse
from reports.views_settings import setting_get_single
from django.template import loader
from django.http import HttpRequest, HttpResponse, JsonResponse, QueryDict
from django.contrib.auth.decorators import login_required
from django.conf import settings
from bs4 import BeautifulSoup

import extracts.views
import extracts.misc
from apps.config import product_name

import json
import time
import logging

LOGGER = logging.getLogger("arcweb")

def HttpResponseError(err):
  return HttpResponse(err, status=500, content_type="text/plain")

def postOnly(f):
  def w(*args, **kwargs):
    request = args[0]
    if request.method != 'POST':
      raise Exception("Invalid request")

    return f(*args, **kwargs)

  return w

def _schedule_now(user, job_name, extract_id = None):
  schedule = JobScheduleNow.objects.create(creator=user, owner=user, extract_id=extract_id)
  schedule.job_name += job_name
  schedule.save(update_fields=['job_name', 'update_time'])
  return schedule

def _auditCustomEmail(customstyle, action, request=None):
  logAction(action=action, component_type=ComponentType.CUSTOMEMAIL,
      component_id=customstyle.id, request=request)

@login_required
@postOnly
def email_snapshot(request, app_id=None):
  if setting_get_single("email_enabled") == 'false':

    return HttpResponse(f"Emailing visual not enabled in {site_setting_name(False)}", status=403)
  return_data = {}

  form_values = json.loads(request.POST.get("emailFormData", "{}"))
  form_type = request.POST.get("formType")

  # Create new schedule with either adhoc, scheduled, or alerted email
  app_id = int(app_id)
  form_values["app_id"] = app_id
  location = reverse('apps-app-id', args=[app_id])

  if form_values.get('visual_url') == None:
    form_values["visual_url"] = request.build_absolute_uri(location)

  with transaction.atomic():
    if form_type == 'emailnow':
      schedule = _schedule_now(request.user, 'Email: {0}'.format(form_values["email_subject"]))
    else:
      try:
        schedule_group = _get_or_create_schedule_group(request.user, form_values)
      except BadRequest as e:
        transaction.set_rollback(True)
        return HttpResponse(str(e), status=400)
      except PermissionDenied as e:
        transaction.set_rollback(True)
        return HttpResponse(str(e), status=403)

      parameters = {
        'creator': request.user,
        'owner': request.user,
        'schedule_group': schedule_group,
        'job_name': form_values['job_name']
      }
      schedule = JobScheduleRepeat.objects.create(**parameters)

      # Add trigger assessment
      if form_type == 'trigger':
        try:
          TriggerJobTarget.create_trigger(schedule=schedule, **form_values)
        except Exception as e:
          transaction.set_rollback(True)
          return HttpResponse(str(e), status=400)

    # Take snapshot or send email immediately
    if form_values["image_type"]:
      SnapshotJobTarget.create_snapshot(schedule=schedule, **form_values)

    EmailJobTarget.create_email(schedule=schedule, **form_values)

  # For adhoc email, don't wait for view_schedule_checker. Immediately queue it.
  if form_type == 'emailnow':
    if schedule.needs_to_be_run():
      schedule.add_to_queue()

  return_data["schedule_id"] = schedule.id
  return HttpResponse(json.dumps(return_data), content_type="application/json")

def _get_schedule_group(scheduleName):
  schedule_groups = ScheduleGroup.objects.filter(schedule_name=scheduleName)
  if schedule_groups.count() == 0:
    raise Exception('Could not fetch schedule group with name %s' % scheduleName)

  if schedule_groups.count() > 1:
    raise Exception('Too many schedule groups with name %s' % scheduleName)

  return schedule_groups.first()

# From the form_values object:
#   Uses schedule_group - to get a group
#    OR
#   Uses user and schedule_group and run_interval and schedule_group_name - to create a group
def _get_or_create_schedule_group(user, form_values):
  schedule_group_id = form_values.get('schedule_group', '')
  schedule_group_id = str(schedule_group_id)

  if schedule_group_id == '':
    schedule_group = None
  elif schedule_group_id == '-2':
    if user_can_schedule(user):
      schedule_group_name = form_values.get('schedule_group_name')
      schedule_group_interval = form_values.get('run_interval')

      if not schedule_group_name or not schedule_group_interval:
        raise BadRequest("Missing interval data")

      schedule_group =\
        ScheduleGroup.objects.create(creator=user,\
          schedule_name=schedule_group_name,\
          run_interval=schedule_group_interval)
      schedule_group.update_next_run_time()
    else:
      raise PermissionDenied('You do not have permission to create new intervals')
  else:
    schedule_group = ScheduleGroup.objects.filter(id=schedule_group_id).first()

  return schedule_group

def rerun_schedule(schedule, target_data, user_id):
  if target_data:
    target_data = json.loads(target_data)

    for target in schedule.jobtarget_set.all():
      new_target_data = target_data.get(str(target.target_order))
      if new_target_data:
        target.target_data = new_target_data
        target.save()

  schedule.add_to_queue(force_run=True, editor_id=user_id)

@csrf_exempt
@postOnly
def api_rerun_schedules_list(request):
  res = check_login_or_use_apikey(request)
  if res is not None:
    return res

  schedule_ids = request.POST.get('schedule_ids')
  schedule_names = request.POST.get('schedule_names')
  if not (schedule_ids or schedule_names) or (schedule_ids and schedule_names):
    return HttpResponseError('either schedule_ids or schedule_names '
                        'should be specified')
  elif schedule_ids:
    schedule_ids = schedule_ids.split(',')
  elif schedule_names:
    schedule_names = schedule_names.split(',')
    schedule_ids = JobSchedule.objects.filter(job_name__in=schedule_names).\
                        values_list('id', flat=True)

  return rerun_schedules_list(request, schedule_ids)

@login_required
@postOnly
def rerun_schedules_list(request, schedules=None):
  # If schedule already exists, rerun the event
  if schedules is None:
    schedules = json.loads(request.POST.get('schedules'))
  successful_run = []
  skipped = []
  user = request.user
  for sch in schedules:
    try:
      schedule_id = int(sch)
      schedule = JobSchedule.objects.filter(id=schedule_id)
    except:
      return HttpResponse('Invalid schedule parameter', status=400)

    if schedule:
      schedule = schedule[0]
    else:
      # job id does not exist
      skipped.append(str(sch))
      continue

    if user_can_schedule(user) or schedule.owner == user:
      rerun_schedule(schedule, request.POST.get(sch), user.id)
      successful_run.append(str(sch))
    else:
      skipped.append(str(sch))

  success_text = ', '.join(successful_run)
  if len(skipped):
    fail_text = '(skipped %s)' % ', '.join(skipped)
  else:
    fail_text = ''

  return HttpResponse('Rerunning Jobs: {0} {1}'.format(success_text, fail_text), status=200)

@login_required
@postOnly
def pause_or_resume(request):
  # If schedule already exists, rerun the event
  schedules = json.loads(request.POST.get('schedules'))
  resume_only = request.POST.get('resume_only')
  successful_run = []
  skipped = []
  user = request.user
  for sch in schedules:
    schedule_id = int(sch)
    schedule = JobSchedule.objects.filter(id=schedule_id)[0]
    last_status_paused = schedule.last_status == STATUS_PAUSED

    if (user_can_schedule(user) or schedule.owner == user) and (not resume_only or last_status_paused):
      if last_status_paused:
        schedule.unpause_schedule()
      else:
        schedule.pause_schedule()
      successful_run.append(sch)
    else:
      skipped.append(sch)

  success_text = ', '.join(successful_run)
  if len(skipped):
    fail_text = '(skipped %s)' % ', '.join(skipped)
  else:
    fail_text = ''

  return HttpResponse('Resumed/Paused Jobs: {0} {1}'.format(success_text, fail_text), status=200)

@login_required
def cancel_job(request, schedule_id=None, job_log_id=None):
  user = request.user
  schedule_id = int(schedule_id)
  schedule = JobSchedule.objects.filter(id=schedule_id)[0]

  if user_can_schedule(user) or schedule.owner == user:
    # In case scheduler isn't actually running anymore, mark log as cancelled directly
    # Need to do this before updating schedule or update_log will throw exception
    if job_log_id:
      job_log_id = int(job_log_id)
      job_log = JobLog.objects.filter(id=job_log_id)[0]
      try:
        job_log.update_log(end_time=True, status=STATUS_CANCELED)
      except CancelTask as e:
       raise Exception('Cannot cancel an already cancelled task.')

    schedule.cancel_schedule()

  return HttpResponse('Canceled Job {0}'.format(schedule_id), status=200)

@login_required
@postOnly
def test_email(request):
  try:
    request_json = json.loads(request.body)

    email_prefix = request_json.get("email_prefix", settings.DEFAULT_EMAIL_PREFIX)
    to_address = request_json["EMAIL_HOST_USER"]
    email_subject = email_prefix + " Test Email"
    email_body = f"Testing {product_name} email settings"
    status = email_self(email_subject, email_body, request_json, to_address)
    return HttpResponse('Email Status: {0} sent'.format(status, status=200))
  except Exception as e:
    return HttpResponse('Email Send Failed: {0}'.format(repr(e)), status=401)

@login_required
def jobs(request):
  if request.method == 'DELETE':
    return _delete_items(request.body, JobSchedule, request.user)
  else:
    return _get_jobs(request)

@login_required
def job_logs(request):
  if request.method == 'DELETE':
    return _delete_items(request.body, JobLog, request.user)

def _delete_items(ids, item_type, user):
  successfully_deleted = []
  no_id_or_skip = []
  if isinstance(ids, bytes):
    ids = ids.decode()
  ids_to_delete = ids.split(',')
  item_type_string = item_type._meta.object_name
  for id in ids_to_delete:
    try:
      item_to_delete = item_type.objects.filter(id=id)
    except ValueError as e:
      return HttpResponse('Invalid input for request body', status=400)

    if len(item_to_delete):
      item_to_delete = item_to_delete[0]
      if item_type_string == 'JobLog':
        created_by_user = item_to_delete.schedule.owner == user
      elif item_type_string == 'ScheduleGroup':
          created_by_user = item_to_delete.creator == user
      elif item_type_string == 'JobSchedule':
        created_by_user = item_to_delete.owner == user
      if user_can_schedule(user) or created_by_user:
        # Delete schedule if ad-hoc and no other runs
        if item_type_string == 'JobLog' and\
           not item_to_delete.schedule.schedule_group and\
           item_to_delete.schedule.run_counter <= 1:
            item_to_delete = item_to_delete.schedule
            id = '%s (Job %s)' % (id, item_to_delete.id)
        item_to_delete.delete()
        successfully_deleted.append(id)
      else:
        no_id_or_skip.append(id)
    else:
      no_id_or_skip.append(id)

  success_text = ', '.join(successfully_deleted)

  if item_type_string == 'JobLog':
    delete_type = 'Log'
  elif item_type_string == 'JobSchedule':
    delete_type = 'Job'
  elif item_type_string == 'ScheduleGroup':
    delete_type = 'Interval'

  successfully_deleted_number = len(successfully_deleted)
  successfully_deleted_message = ''
  if successfully_deleted_number >= 1:
    plural_modifier = 's' if successfully_deleted_number > 1 else ''
    successfully_deleted_message = 'Deleted {0}{1}: {2} '.format(delete_type,
                                                                 plural_modifier,
                                                                 success_text)

  if len(no_id_or_skip):
    failed_message = ', '.join(no_id_or_skip)
    if successfully_deleted_number >= 1:
      failed_message = '[skipped {0}]'.format(failed_message)
    else:
      # If no items were deleted, return back error message
      return HttpResponseError('Unable to delete: ' + failed_message)
  else:
    failed_message = ''

  return HttpResponse(successfully_deleted_message + failed_message, status=200)

@transaction.atomic
def edit_job(request, schedule_id=None):
  if not schedule_id:
    return HttpResponse('Missing job id', status=400)

  job_schedule = JobSchedule.objects.filter(id=schedule_id).first()

  if not job_schedule:
    return HttpResponse('Job not found', status=404)

  if not user_can_schedule(request.user) and not job_schedule.owner == request.user:
    return HttpResponse('You do not have permission to edit this job', status=403)

  form_data = json.loads(request.POST.get("emailFormData", "{}"))

  # 1.1) Update EmailJobTarget
  email_target = job_schedule.jobtarget_set.filter(target_type='E').first()
  if not email_target:
    return HttpResponse('Email target not found', status=400)

  email_target.overwrite_values(form_data)

  # 1.2) Update TriggerJobTarget
  # It is not mandatory for an email job to have a trigger target
  trigger_target = job_schedule.jobtarget_set.filter(target_type='T').first()
  if trigger_target:
    trigger_target.overwrite_values(form_data)

  # 2) Update JobSchedule
  is_schedule_updated = False

  # Update JobSchedule.job_name if needed
  job_name = form_data.get('job_name', None)
  if job_name and job_schedule.job_name != job_name:
    is_schedule_updated = True
    job_schedule.job_name = job_name

  try:
    # Update ScheduleGroup if needed
    schedule_group = _get_or_create_schedule_group(request.user, form_data)
  except BadRequest as e:
    transaction.set_rollback(True)
    return HttpResponse(str(e), status=400)
  except PermissionDenied as e:
    transaction.set_rollback(True)
    return HttpResponse(str(e), status=403)

  if not schedule_group and job_schedule.schedule_group_id:
    transaction.set_rollback(True)
    return HttpResponse('Schedule group not found', status=400)

  if schedule_group and job_schedule.schedule_group_id != schedule_group.id:
    is_schedule_updated = True
    job_schedule.schedule_group_id = schedule_group.id
    form_data['schedule_group_id'] = schedule_group.id

  if is_schedule_updated:
    job_schedule.save()

  return JsonResponse(form_data)

def edit_job_owner(request, schedule_id=None):
  job_list = JobSchedule.objects.filter(id=schedule_id)
  if not schedule_id or not job_list.exists():
    return HttpResponse('Invalid job id', status=500)

  newowner_id = int(request.POST.get('newowner_id', ''))
  if newowner_id == '':
    newowner_id = request.user.id
  else:
    is_bulk_change = request.POST.get('is_bulk_change', False)
    if is_bulk_change == 'true':
      job_schedule = job_list.first()
      current_owner_id = job_schedule.owner_id
      job_list = JobSchedule.objects.filter(owner=current_owner_id)

  newowner_name = User.objects.get(id=newowner_id).username
  jobs = []
  for job in job_list:
    if job.owner_id != newowner_id:
      job.owner_id = newowner_id
      jobs.append({'schedule_id': job.id, 'newowner_id': job.owner_id})
      job.save()

      logs = JobLog.objects.filter(schedule_id=job.id)
      for log in logs:
        log.update_log_owner(newowner_id, newowner_name)

      if job.job_name.startswith('Data Extract: ') or \
        job.job_name.startswith('Adhoc Data Extract: '):
        update_extract_owner(job, newowner_name)

  return HttpResponse(json.dumps(jobs), content_type="application/json")

def update_extract_owner(job: JobSchedule, newowner_name: str):
  for job_target in job.jobtarget_set.all():
    if job_target.target_type == JobTarget.DATACOPY:

      target_data = json.loads(job_target.target_data)
      target_data["username"] = newowner_name
      job_target.target_data = json.dumps(target_data)
      job_target.save()

      extractId = target_data.get('extractId')
      extract = Extract.objects.get(id=extractId)
      extract_config = json.loads(extract.extract_config)
      extract_config["username"] = newowner_name
      extract.extract_config = json.dumps(extract_config)
      extract.save()

def get_logs_for_schedules(schedule_ids: list[int]):
  log_ids = get_log_ids_for_schedules(schedule_ids)

  logs = JobLog.objects.filter(id__in=log_ids)

  schedule_logs = dict[int,JobLog]()
  for log in logs:
    schedule_logs[log.schedule_id] = log

  return schedule_logs

def get_log_ids_for_schedules(schedule_ids: list[int]):

  # We want to get the most recent log for each schedule.
  # Here we assume that the most recent log is the one with the highest id.
  rows = JobSchedule.objects\
    .filter(id__in=schedule_ids) \
    .values_list('id') \
    .annotate(max_log_id=Max('joblog__id'))

  log_ids = [row[1] for row in rows if row[1] is not None]
  return log_ids

def get_targets_and_latest_log(schedule: JobSchedule, latest_log: JobLog):
  if latest_log:
    latest_log = json.loads(serializers.serialize('json', [latest_log]))[0]
  return {
    'schedule_log': latest_log,
    'targets': json.loads(serializers.serialize('json', schedule.jobtarget_set.all())),
    }

@login_required
def render_job_overview_page(request):
  context = _appflags(request)
  return render(request, 'job_overview_page.html', context)

@login_required
def render_job_page(request, schedule_id):
  context = _appflags(request)
  context.update({ "id": schedule_id })
  return render(request, 'partial_admin_job.html', context)

@login_required
def get_email_templates(_):
  email_templates = json.loads(CustomEmailTemplate.email_templates('id', 'template_name', 'template_description', 'template_default_message'))

  return JsonResponse({ 'email_templates': email_templates })

def _get_pagination_params(request: HttpRequest):
  max_page_size = int(settings.JOBS_MAX_LOGS)

  page = request.GET.get('page', '')
  page = int(page) if page.isdigit() else 1

  page_size = request.GET.get('page_size', '')
  page_size = min(int(page_size), max_page_size) if page_size.isdigit() else max_page_size

  return page, page_size

def _parse_job_data(job: JobSchedule):
  owner_username = job.owner.username if job.owner is not None else None
  owner_id = job.owner.id if job.owner is not None else None
  creator_username = job.creator.username if job.creator is not None else None
  creator_id = job.creator.id if job.creator is not None else None

  job_data = {
    'creator_id': creator_id,
    'creator_username': creator_username,
    'job_name': job.job_name,
    'last_finished_time': job.last_finished_time,
    'last_started_time': job.last_started_time,
    'extract_id': job.extract_id,
    'owner_id': owner_id,
    'owner_username': owner_username,
    'schedule_id': job.id,
    'scheduler_status': job.last_status,
    'schedule_group_id': job.schedule_group_id,
    'targets': json.loads(serializers.serialize('json', job.jobtarget_set.all())),
  }

  return job_data

def _get_jobs(request: HttpRequest):
  sort = request.GET.get('sort', '-update_time')
  status_filter= request.GET.get('status')
  owner_filter= request.GET.get('owner')
  is_extract_filter= request.GET.get('is_extract')
  last_finished_start_filter = request.GET.get('last_finished_time_start')
  last_finished_end_filter = request.GET.get('last_finished_time_end')
  is_scheduled_filter = request.GET.get('is_scheduled', '').lower()
  search =  request.GET.get('search')
  threshold_only = request.GET.get('threshold_only', "False")
  page, page_size = _get_pagination_params(request)

  jobs = JobSchedule.objects \
    .select_related('creator', 'owner') \
    .prefetch_related('jobtarget_set') \
    .order_by(sort)

  if not user_can_schedule(request.user):
    jobs = jobs.filter(owner=request.user)

  if status_filter:
    jobs = jobs.filter(last_status__in=status_filter.split( ','))
  if owner_filter:
    jobs = jobs.filter(owner_id__in=owner_filter.split(','))
  if last_finished_start_filter:
    jobs = jobs.filter(last_finished_time__gte=last_finished_start_filter)
  if last_finished_end_filter:
    jobs = jobs.filter(last_finished_time__lte=last_finished_end_filter)
  if is_scheduled_filter in ['true', 'false']:
    jobs = jobs.filter(schedule_group_id__isnull=is_scheduled_filter == 'false')
  if is_extract_filter in ['true', 'false']:
    jobs = jobs.filter(extract_id__isnull=is_extract_filter == 'false')
  if search:
    jobs = jobs.filter(Q(job_name__icontains=search) | Q(id__icontains=search))
  if threshold_only.lower() == "true":
    jobs = jobs.filter(trigger_target__isnull=False)

  paginator = Paginator(jobs, per_page=page_size)
  jobs = paginator.get_page(page).object_list

  jobs_data = []
  for job in jobs:
    job_data = _parse_job_data(job)
    jobs_data.append(job_data)

  return JsonResponse({
    "jobs": list(jobs_data),
    "total": paginator.count
  })

@login_required
def get_latest_logs(request: HttpRequest):
  max_logs = request.GET.get('max_logs', settings.JOBS_MAX_LOGS)
  max_logs = int(max_logs)

  logs = JobLog.objects.all()
  jobs = JobSchedule.objects.select_related('creator', 'owner')

  if not user_can_schedule(request.user):
    jobs = jobs.filter(owner=request.user)
    logs = logs.filter(schedule__owner=request.user)

  jobs_with_most_recent_logs = logs.values('schedule_id').annotate(ut=Max('schedule__update_time')).order_by('-ut')[:max_logs]
  job_ids = [log['schedule_id'] for log in jobs_with_most_recent_logs]
  jobs = jobs.filter(id__in=job_ids)
  logs_for_schedules = get_logs_for_schedules(job_ids)

  jobs = jobs.prefetch_related('jobtarget_set').order_by('-update_time')

  parsed_jobs_with_logs = []
  for job in jobs:
    log = logs_for_schedules.get(job.id)
    job_data = _parse_job_data(job)

    log_data = {
      'id': log.id,
      'schedule_id': log.schedule_id,
      'contents_archive': log.contents_archive,
      'status': log.status,
      'start_time': log.start_time,
      'end_time': log.end_time,
      'task_list': log.task_list,
      'error_message': log.error_message,
    }

    parsed_job_and_log = {
      "job": job_data,
      "log": log_data,
    }
    parsed_jobs_with_logs.append(parsed_job_and_log)

  return JsonResponse({'data': parsed_jobs_with_logs})

@login_required
def get_job(request, schedule_id=None):
  job = JobSchedule.objects.filter(id=schedule_id).select_related('owner', 'creator', 'schedule_group').prefetch_related('jobtarget_set').first()

  if job is None:
    return HttpResponse('Job not found', status=404)

  user = request.user
  if user_can_schedule(user) or job.owner == user:
    job_data = _parse_job_data(job)

    if job.schedule_group:
      job_data['schedule_group'] = {
        'id': job.schedule_group.id,
        'schedule_name': job.schedule_group.schedule_name,
        'run_interval': job.schedule_group.run_interval,
        'next_run_time': str(job.schedule_group.next_run_time),
      }

    return JsonResponse(job_data)
  else:
    return HttpResponse('You do not have permission to view this job', status=403)

@login_required
def get_job_logs_for_schedule(request, schedule_id=None):
  sort = request.GET.get('sort', '-id')
  page, page_size = _get_pagination_params(request)
  status_filter= request.GET.get('status')
  start_time_filter = request.GET.get('start_time')
  end_time_filter = request.GET.get('end_time')

  job = JobSchedule.objects.filter(id=schedule_id).select_related('owner').first()

  if job is None:
    return HttpResponse('Job not found', status=404)

  user = request.user
  if user_can_schedule(user) or job.owner == user:
    logs = JobLog.objects.filter(schedule=schedule_id).order_by(sort)

    if status_filter:
      logs = logs.filter(status__in=status_filter.split(','))
    if start_time_filter:
      logs = logs.filter(start_time__gte=start_time_filter)
    if end_time_filter:
      logs = logs.filter(end_time__lte=end_time_filter)

    paginator = Paginator(logs, per_page=page_size)
    logs = paginator.get_page(page).object_list

    return JsonResponse({
      "logs": list(logs.values()),
      "total": paginator.count
    })
  else:
    return HttpResponse('You do not have permission to access the logs for this job', status=403)

@login_required
def get_job_status(request, schedule_id=None):
  job = JobSchedule.objects.filter(id=schedule_id).first()

  if job is None:
    return HttpResponse('Job not found', status=404)

  status = job.last_status

  user = request.user
  if user_can_schedule(user) or job.owner == user:
    return JsonResponse({ "status":  status })
  else:
    return HttpResponse('You do not have permission to access the logs for this job', status=403)


def getDataCopyJobs(user):
  job_targets = JobTarget.objects.select_related("schedule").select_related('schedule__schedule_group').filter(target_type=JobTarget.DATACOPY)

  extract_ids = set()
  job_targets_extract_ids = {}
  for jt in job_targets:
    try:
      j = json.loads(jt.target_data)
    except:
      LOGGER.warning(f"Invalid JobTarget with id {jt.id}")
      continue

    extractId = j.get('extractId')
    if not extractId:
      continue

    if extractId in extract_ids:
      previousJobTarget = job_targets_extract_ids[extractId]
      if not previousJobTarget.schedule.schedule_group:
        job_targets_extract_ids[extractId] = jt
    else:
      extract_ids.add(extractId)
      job_targets_extract_ids[extractId] = jt

  extracts = Extract.objects.select_related('srcDataset').filter(id__in=extract_ids)

  extracts_perms = {}
  for e in extracts:
    extracts_perms[e.id] = e.srcDataset.getPerms(user)

  l = []

  for extractId, jt in job_targets_extract_ids.items():
    # Only return extracts on datasets that are visible to the user
    try:
      dsperms = extracts_perms[extractId]
    except:
      continue
    if 'ds_appview' not in dsperms:
      continue

    l.append((extractId, jt))

  return l

@login_required
def dataCopyJobs(request):
  extracts.misc.checkEnabled()

  srcDatasetId = request.GET.get("srcDatasetId", None)
  if srcDatasetId:
    srcDatasetId = int(srcDatasetId)
    srcDataset = DataSet.objects.get(id=srcDatasetId)
    srcDataset.checkPerm(request.user, 'ds_appview')

  r = []
  for extractId, jt in getDataCopyJobs(request.user, srcDatasetId):
    if srcDatasetId:
      extract = es.get(extractId)
      if not extract:
        continue

      if extract.srcDataset.id != srcDatasetId:
        continue

    r.append({
      "id": jt.id,
      "extractId": extractId,
      "srcDatasetId": srcDatasetId,
      "scheduleId": jt.schedule.id,
      "scheduleGroupId": jt.schedule.schedule_group.id,
      "scheduleGroupName": jt.schedule.schedule_group.schedule_name,
    })

  return HttpResponse(json.dumps(r), content_type="application/json")


@login_required
def dataCopyLogs(request):
  # In the front-end, if extracts are disabled, we may still end up getting a few
  # get requests for extracts in the backend. In these cases, we just return an
  # empty list (instead of erroring out)
  if not extracts.misc.enabled():
    return HttpResponse(json.dumps([]), content_type="application/json")

  srcDatasetId = request.GET.get('srcDatasetId')
  if srcDatasetId:
    srcDatasetId = int(srcDatasetId)
    perm.check_dataset_permission(request.user, srcDatasetId, 'ds_appview')

  extractId = request.GET.get('extractId')
  if extractId:
    extractId = int(extractId)
    check_extract_permission(request.user, extractId)

  logs = getDataCopyLogs(request.user, srcDatasetId, extractId)
  return HttpResponse(json.dumps(logs), content_type="application/json")

def check_extract_permission(user: User, extract_id: int):
  if user.is_superuser:
    return

  extract = Extract.objects.select_related('srcDataset__dataconnection','tgtDataconnection')\
    .only(
      'srcDataset__create_user',
      'srcDataset__version_group_id',
      'srcDataset__dataconnection__create_user',
      'srcDataset__dataconnection__dataconnection_type',
      'tgtDataconnection__create_user',
      'tgtDataconnection__dataconnection_type',
    ) \
    .get(id=extract_id)
  extracts.views.checkPerm(user, extract)

def getDataCopyLogs(user: User, srcDatasetId: int, extractId: int):
  """
  Returns the latest log for every data extract.
  """
  logs = []

  # latest logs:
  latest_logs = JobLog.objects\
    .values('schedule__extract_id')\
    .annotate(log_id=Max('id')) \
    .order_by('schedule__extract_id') # without ordering the groupby gets end_time

  # last succesful runs time:
  latest_success = JobLog.objects\
    .values('schedule__extract_id')\
    .filter(status=STATUS_COMPLETED)\
    .annotate(last_success_time=Max('end_time')) \
    .order_by('schedule__extract_id') # without ordering the groupby gets end_time

  if(extractId):
    latest_logs = latest_logs.filter(schedule__extract_id=extractId)
    latest_success = latest_success.filter(schedule__extract_id=extractId)
  else:
    latest_logs = latest_logs.filter(schedule__extract_id__isnull=False)
    latest_success = latest_success.filter(schedule__extract_id__isnull=False)

  if srcDatasetId:
    latest_logs = latest_logs.filter(schedule__extract__srcDataset_id=srcDatasetId)
    latest_success = latest_success.filter(schedule__extract__srcDataset_id=srcDatasetId)

  # single dataset and extract permissions are checked before calling this function
  if not (user.is_superuser or extractId or srcDatasetId):
    user_ds_ids = perm.user_dataset_ids(user)
    latest_logs = latest_logs.filter(schedule__extract__srcDataset_id__in=user_ds_ids)
    latest_success = latest_success.filter(schedule__extract__srcDataset_id__in=user_ds_ids)

  logToExtractId = {}
  lastSuccessfulRun = {}
  timeFormatStr = "%Y-%m-%d %H:%M:%S"

  for success in latest_success:
    success_time = success.get('last_success_time')
    if success_time:
      lastSuccessfulRun[success['schedule__extract_id']] = success_time.strftime(timeFormatStr)

  joblog_ids = []
  for log in latest_logs:
    extract_id = log['schedule__extract_id']
    log_id = log['log_id']
    joblog_ids.append(log_id)
    logToExtractId[log_id] = extract_id

  jobLogs = JobLog.objects.filter(id__in=joblog_ids).only('status', 'start_time', 'end_time', 'error_message', 'task_list')

  for jobLog in jobLogs:
    log_extract_id = logToExtractId.get(jobLog.id)

    status = STATUS_MAPS.get(jobLog.status, "Unknown")

    el = {
      "extractId": log_extract_id,
      "runtime": (jobLog.end_time - jobLog.start_time).total_seconds(),
      "start_time": jobLog.start_time.strftime(timeFormatStr),
      "end_time": jobLog.end_time.strftime(timeFormatStr),
      "error_message": jobLog.error_message,
      "task_list": jobLog.task_list,
      "status": status,
    }

    lastSuccessTime = lastSuccessfulRun.get(log_extract_id)
    if lastSuccessfulRun:
      el["lastSuccessfulRun"] = lastSuccessTime

    logs.append(el)

  return logs

@login_required
@postOnly
def dataCopySchedule(request):
  extracts.misc.checkEnabled()

  params = request.POST.copy()

  extractId = params.get('extractId', None)
  if extractId is None:
    raise Exception("extractId not specified")
  else:
    extractId = int(extractId)

  extract = Extract.objects.get(id=extractId)
  if not extract:
    raise Exception("Invalid extractId")

  scheduleId = params.get('scheduleId', None)
  if scheduleId is not None:
    scheduleId = int(scheduleId)

  extracts.views.checkPerm(request.user, extract)

  jobName = extract.jobName()

  # DELETE
  if scheduleId == 0:
    # Delete any previous jobs for this extract
    js = extract.jobschedule_set.first()
    if js:
      LOGGER.info(f"Deleting job {js.id} for extract with id {extract.id}")
      js.delete()

    return HttpResponse(json.dumps({}), content_type="application/json")

  sgs = None
  if scheduleId is not None:
    sgs = ScheduleGroup.objects.filter(id=scheduleId)
    if not sgs.exists():
      raise Exception("Invalid schedule group %s" % scheduleId)
  else: # Manual
    jobName = f'Adhoc {jobName}'

  job_schedule = extract.jobschedule_set.first()
  de_notification_email = json.loads(extract.extract_config).get('notification_email', None)
  de_to_addresses = None
  if de_notification_email:
    de_to_addresses = de_notification_email.get('to_addresses')

  # CREATE
  if not job_schedule:
    params = {
      'creator': request.user,
      'owner': request.user,
      'schedule_group': sgs[0] if sgs else None,
      'job_name': jobName,
      'extract_id': extractId,
    }

    job_schedule = JobSchedule.objects.create(**params)

    ds_location = reverse('apps-dataset-by-id', args=[extract.srcDataset_id])
    srcdataset_url = f'{request.build_absolute_uri(ds_location)}#tab_extracts'
    jobs_location = reverse('jobs-render-job-overview-page')
    jobs_url = f'{request.build_absolute_uri(jobs_location)}'

    jobTarget = DataCopyJobTarget.createDataCopy(
        schedule=job_schedule,
        extractId=extractId,
        request=request,
        srcdataset_url=srcdataset_url,
        jobs_url=jobs_url)

    # Needed for email input suggestions
    if de_to_addresses:
      jobTarget._add_trigger_content(de_to_addresses, 'to_addresses')

    return_data = {
      'schedule_id': job_schedule.id,
      'jobtarget_id': jobTarget.id
    }

    if not scheduleId: # Manual run
      run_schedule(job_schedule, return_data)
    return HttpResponse(json.dumps(return_data), content_type="application/json")

  else: # EDIT
    old_schedule_id = job_schedule.schedule_group_id
    if old_schedule_id != scheduleId:
      job_schedule.schedule_group_id = sgs[0].id if sgs else None
      update_fields = ['schedule_group_id', 'update_time']

      if job_schedule.job_name != jobName:
        job_schedule.job_name = jobName
        update_fields.append('job_name')

      # If by previous code job_schedule created as JobScheduleNow, then undo it to just JobSchedule.
      # Only difference between them -> for JobScheduleNow max_success_run is 1 and finished is True after
      # it was run once, hence won't run anymore
      if job_schedule.max_success_run: # JobScheduleNow
        job_schedule.max_success_run = None
        job_schedule.finished = False

        # JobScheduleNow can have STATUS_PAUSED, if it was cancelled or failed 5+ times,
        # unpause it so it starts running per schedule
        job_log = job_schedule.joblog_set.first()
        if job_log:
          job_schedule.last_status = job_log.status
        else:
          job_schedule.last_status = None

        update_fields.extend(['max_success_run', 'finished', 'last_status'])

      job_schedule.save(update_fields=update_fields)

    # Needed for email input suggestions
    jobTarget = DataCopyJobTarget.objects.filter(schedule_id=job_schedule.id).first()
    if de_to_addresses:
      jobTarget._add_trigger_content(de_to_addresses, 'to_addresses')

    return_data = {
      'schedule_id': job_schedule.id,
      'jobtarget_id': jobTarget.id
    }
    return HttpResponse(json.dumps(return_data), content_type="application/json")

def run_schedule(schedule, return_data):
    schedule.add_to_queue()
    specific_job_log = schedule.job_log
    if specific_job_log:
      return_data['joblog_id'] = specific_job_log.id


@login_required
@postOnly
def scriptSchedule(request):
  if not settings.ENABLE_SQL_SCHEDULER:
    raise Exception("Cannot schedule SQL jobs. " \
        "Please modify the arcviz configuration to enable it")

  params = request.POST.copy()
  params['run_now'] = True

  return_data = scriptScheduleInternal(request.user, params)
  return HttpResponse(json.dumps(return_data), content_type="application/json")

def scriptScheduleInternal(user, params):

  schedule_name = params.get('schedule_name')
  job_name = params.get('job_name')
  dataconnection_id = params.get('dataconnection_id')
  sql = params.get('sql')
  email_address = params.get('email_address')

  run_now = params.get('run_now')

  if schedule_name:
    schedule_group = _get_schedule_group(schedule_name)
    parameters = {
      'creator': user,
      'owner': user,
      'schedule_group': schedule_group,
      'job_name': job_name,
    }

    schedule = JobScheduleRepeat.objects.create(**parameters)
  else:
    schedule = _schedule_now(user, 'sql: %s' % (time.time()))

  ScriptTarget.createScript(
      schedule=schedule,
      script_type="sql",
      script=sql,
      email_address=email_address,
      dataconnection_id=dataconnection_id,
      user=user)

  return_data = {
    'schedule_id': schedule.id
  }

  if run_now:
    schedule.add_to_queue()
    specific_job_log = schedule.job_log
    if specific_job_log:
      return_data['joblog_id'] = specific_job_log.id

  return return_data

def _delete_email_template(request, template_id):
  try:
    # pylint: disable=E1101
    email_template = CustomEmailTemplate.objects.get(id=template_id)
  except Exception:
    return HttpResponseError('Invalid email template id')

  _auditCustomEmail(email_template, ActionType.DELETE, request=request)

  # Set dependent schedule templates to empty string
  email_template_info = CustomEmailTemplate.email_templates('schedule_ids', 'adhoc_schedule_ids', id=template_id, no_json=True)[0]

  # If there's no schedule_id or adhoc_schedule_ids, it's returned as None
  # Can't concat that to an array, so set to empty array instead
  repeating_jobs = email_template_info.get('schedule_ids') or []
  adhoc_jobs = email_template_info.get('adhoc_schedule_ids') or []
  for sch in (repeating_jobs + adhoc_jobs):
    emails = JobSchedule.objects.filter(id=sch).first().jobtarget_set.filter(target_type=JobTarget.EMAIL)

    for em in emails.all():
      new_data = json.loads(em.target_data)
      if new_data.get('email_template') == str(template_id):
        del new_data['email_template']
        em.target_data = json.dumps(new_data)
        em.save(update_fields=['target_data'])

  email_template.delete()
  return HttpResponse("[]", content_type="application/json")

def _edit_email_template(request, template_id):
  name = request.POST.get('name', None)
  description = request.POST.get('description', None)
  value = request.POST.get('template', None)
  schedule_id = request.POST.get('schedule_id', None)
  if value:
    s = BeautifulSoup(value, "lxml")
    if len(s.find_all('script')) > 0:
      raise Exception('Invalid template')

    for a in s.find_all('a', href=True):
      s_a = BeautifulSoup(a['href'], "lxml")
      if len(s_a.find_all('script')) > 0:
        raise Exception('Invalid template')

  if template_id is None:
    # Create
    t = CustomEmailTemplate.objects.create(field='custom_email_templates')
    t.creator.add(request.user)
    schedule_ids = []
    if schedule_id:
      schedule_ids.append(schedule_id)

    t.value = json.dumps({
      'update_time': t.create_time.strftime('%x %X %Z'),
      'schedule_ids': schedule_ids,
      'template_name': name or '',
      'template_description': description or '',
      'template_value': value or '',
    })
    t.save()
    _auditCustomEmail(t, ActionType.CREATE, request=request)
    return HttpResponse(json.dumps([t.id]), content_type="application/json")

  # Update
  try:
    # pylint: disable=E1101
    template = CustomEmailTemplate.objects.get(id=template_id)
  except Exception:
    return HttpResponseError('Invalid template id')

  template_values = json.loads(template.value)
  new_schedule_ids = template_values.get('schedule_ids', [])

  if schedule_id:
    new_schedule_ids.append(schedule_id)

  template_values['schedule_ids'] = new_schedule_ids
  template_values['update_time'] = datetime.now(template.create_time.tzinfo).strftime('%x %X %Z')

  if name != None:
    template_values['template_name'] = name


  if description != None:
    template_values['template_description'] = description


  if value != None:
    template_values['template_value'] = value


  template.value = json.dumps(template_values)
  template.save()
  _auditCustomEmail(template, ActionType.UPDATE, request=request)
  return HttpResponse(json.dumps([template.id]), content_type="application/json")


def _get_email_templates(request, template_id):
  if template_id == None:
    template = loader.get_template('admin_custom_email.html')
  else:
    template = loader.get_template('admin_custom_email_form.html')

  c = _appflags(request)
  c.update({
    'scheduler_enabled': _SCHEDULER_ENABLED,
  })

  if template_id == None:
    c['email_templates'] = CustomEmailTemplate.email_templates('id',
                                                               'template_name',
                                                               'template_description',
                                                               'schedule_ids')
  else:
    c['email_template'] = CustomEmailTemplate.email_templates('id',
                                                              'template_name',
                                                              'template_description',
                                                              'template_value',
                                                              id=template_id)
    c['templateid'] = template_id

  c['show_code_view'] = 'true' if getattr(settings, "EMAIL_TEMPLATE_SHOW_CODE_VIEW", True) \
                          else 'false'

  return HttpResponse(template.render(c, request))

@login_required
def preview_custom_email(request, template_id=None):
  params = request.GET
  template_id = template_id or params.get('template_id')
  if template_id:
    template_info = CustomEmailTemplate.email_templates('template_value', 'template_name', id=template_id, no_json=True)[0]
    template_text = template_info.get('template_value', '')
    template_name = template_info.get('template_name', '')
  else:
    template_name = 'Default template'
    template_text = loader.get_template('send_visual_full.html')\
                          .render({}, request)
  template = loader.get_template('preview_email_template.html')
  context = {
    'template_text': template_text,
    'template_name': template_name,
  }
  return HttpResponse(template.render(context, request))


@login_required
def custom_email(request, template_id=None):
  if template_id != None:
    template_id = int(template_id)

  if not user_can_schedule(request.user):
    return HttpResponse('You do not have permission to update email templates', status=403)

  if request.method == 'GET':
    return _get_email_templates(request, template_id)

  if request.method == 'DELETE':
    return _delete_email_template(request, template_id)

  if request.method == 'POST':
    return _edit_email_template(request, template_id)

  return HttpResponse('Invalid request', status=400)

"""
ANALYTICAL VIEW REFRESH REST END POINT HERE
"""
@login_required
@postOnly
def viewrefresh(request):
  form_values = json.loads(request.POST.get('formValues') or "{}")
  opcode = request.POST.get('opcode')
  if opcode == 'enque':
    schedule = _schedule_now(request.user, 'Refresh: {0}'.format(form_values["view"]))
    RefreshJobTarget.create_refresh_job(schedule=schedule, **form_values)

    # For adhoc view refresh, don't wait for the queue. Just go ahead and
    # push it into the queue

    return_data = {
      'dcid': form_values['dcid'],
      'view': form_values['view'],
      'schedule_id': schedule.id,
    }

    # Not sure if this if statement is needed here
    if schedule.needs_to_be_run():
      schedule.add_to_queue()
      specific_job_log = schedule.job_log
      if specific_job_log:
        return_data['joblog_id'] = specific_job_log.id

    return HttpResponse(json.dumps(return_data), content_type="application/json")

  elif opcode == 'schedule':
    try:
      schedule_group = _get_or_create_schedule_group(request.user, form_values)
    except BadRequest as e:
      return HttpResponse(str(e), status=400)
    except PermissionDenied as e:
      return HttpResponse(str(e), status=403)

    parameters = {
      'creator': request.user,
      'owner': request.user,
      'schedule_group': schedule_group,
      'job_name': form_values['job_name']
    }
    schedule = JobScheduleRepeat.objects.create(**parameters)

    RefreshJobTarget.create_refresh_job(schedule=schedule, **form_values)

    return_data = {
      "schedule_id": schedule.id,
      "next_run": schedule_group.next_run_time.isoformat(),
    }
    return JsonResponse(return_data)

  elif opcode == 'fetch':
    return _fetch_refresh_status(form_values)

  else:
    return HttpResponseError('Invalid request')

def _get_next_scheduled_refresh(dc_id, av_name=None):
  # Retrieve all targets for repeating AV refresh jobs on this DC that don't have a status of paused
  targets = RefreshJobTarget.objects.filter(target_type='R', target_data__contains='"dcid": %s,' % dc_id,
    schedule__schedule_group__isnull=False).exclude(schedule__last_status='P')\
    .prefetch_related('schedule', 'schedule__schedule_group')

  if av_name:
    targets = targets.filter(target_data__contains='"view": "' + av_name + '",')

  next_refresh_hash = {}
  for target in targets:
    view = json.loads(target.target_data).get('view')
    if next_refresh_hash.get(view) is None:
      next_refresh_hash[view] = target.schedule.schedule_group.next_run_time
    else:
      next_refresh_hash[view] = min(next_refresh_hash.get(view), target.schedule.schedule_group.next_run_time)

  return next_refresh_hash

"""
POST /arc/jobs/api/extracts/create
POST parameters:
  - data: dict with items
    - scheduleId (optional): if given then job will be created for the extract
    - srcDatasetId
    - tgtDataconnectionId
    - tgtDbname
    - tgtTablename
    - dimData: list of dict which contains Dimension column's expr(which is "[<column_name>]")
              and alias(which is "<column_name>") as JSON string. Dimension columns are from source dataset.
              E.g.: "[{\"expr\":\"[col_a]\",\"alias\":\"col_a\"},{\"expr\":\"[col_b]\",\"alias\":\"col_b\"}]"
    - aggData: list of dict which contains Measure column's expr(which is "<agg_function>([<column_name>])")
              and alias(which is "<column_name>") as JSON string. Measure columns are from source dataset.
              E.g.: "[{\"expr\":\"sum([protein_grams])\",\"alias\":\"protein_grams\"}]"
    - partitionData: list of dict which contains Partition column's expr(which is "[<column_name>]")
              and alias(which is "<column_name>") as JSON string. Partition columns are from source dataset,
              and are used to partition and incrementally refresh the target tables
              E.g.: "[{\"expr\":\"[col_a]\",\"alias\":\"col_a\"}]"
E.g. data={
            "scheduleId": 2,
            "srcDatasetId": 9,
            "tgtDataconnectionId": 26,
            "tgtDbname": "main",
            "tgtTablename": "cdv_vv16",
            "dimData": "[{\"expr\":\"[cereal_name]\",\"alias\":\"cereal_name\"},{\"expr\":\"[manufacturer_code]\",\"alias\":\"manufacturer_code\"}]",
            "aggData": "[{\"expr\":\"sum([protein_grams])\",\"alias\":\"protein_grams\"},{\"expr\":\"sum([fat_grams])\",\"alias\":\"fat_grams\"}]",
            "partitionData": "[]"
          }
"""
@csrf_exempt
def api_extracts_create(request):
  if not FeatureFlipper().get_flipper_value('site_enable_extracts'):
    return HttpResponse(f'Data Extracts not enabled in {site_setting_name(False)}', status=403)

  err = check_login_or_use_apikey(request)
  if err:
    return err

  if request.method == 'POST':
    try:
      data = json.loads(request.POST.get('data'))
    except:
      data = {}
    if not data:
      msg = """Expected json data like:
                {
                  "scheduleId": 2,
                  "srcDatasetId": 9,
                  "tgtDataconnectionId": 26,
                  "tgtDbname": "main",
                  "tgtTablename": "cdv_vv16",
                  "dimData": "[{\"expr\":\"[cereal_name]\",\"alias\":\"cereal_name\"},{\"expr\":\"[manufacturer_code]\",\"alias\":\"manufacturer_code\"}]",
                  "aggData": "[{\"expr\":\"sum([protein_grams])\",\"alias\":\"protein_grams\"},{\"expr\":\"sum([fat_grams])\",\"alias\":\"fat_grams\"}]",
                  "partitionData": "[]"
                }
            """
      return HttpResponse(msg, status=400)

    schedule_id = data.pop('scheduleId', None)

    # Create extract
    if len(json.loads(data.get('partitionData', []))) == 0:
        data['extract_type'] = 'refresh_full'
    else:
        data['extract_type'] = 'refresh_partitioned'

    request.POST = QueryDict('', mutable=True)
    request.POST.update(data)
    rsp = extracts.views.extract_internal(request)

    if rsp.status_code != 200:
      return rsp

    return_obj = {'extract' : json.loads(rsp.content)[0]}

    # Create job
    if schedule_id is not None:
      extract_id = return_obj.get('extract', {}).get('id')
      request.POST.update({'extractId': extract_id, 'scheduleId': schedule_id})
      rsp = dataCopySchedule(request)
      if rsp.status_code != 200:
        return rsp
      return_obj['jobtarget'] = json.loads(rsp.content)

    return HttpResponse(json.dumps(return_obj), content_type='application/json')

  else:
    return HttpResponse('Invalid request type. Only POST requests accepted', status=400)

"""
POST /arc/jobs/api/extracts/run
POST parameters:
  - extract_ids: comma-separated ints
"""
@csrf_exempt
def api_extracts_run(request):
  if not FeatureFlipper().get_flipper_value('site_enable_extracts'):
    return HttpResponse(f'Data Extracts not enabled in {site_setting_name(False)}', status=403)

  err = check_login_or_use_apikey(request)
  if err:
    return err

  if request.method == 'POST':
    extract_ids = request.POST.get('extract_ids')
    if not extract_ids:
      return HttpResponseError('extract_ids should be specified')
    elif extract_ids:
      extract_ids = extract_ids.split(',')

    successful_run = []
    skipped = []

    for e_id in extract_ids:
      try:
        extract_id = int(e_id)
        extract = Extract.objects.filter(id=extract_id)
      except:
        return HttpResponse(f'Invalid extract id: {e_id}', status=400)

      if extract:
        extract = extract[0]
        request.POST = QueryDict('', mutable=True)
        request.POST.update({'extractId': extract_id})
        dataCopySchedule(request)
        successful_run.append(str(e_id))
      else:
        # extract id does not exist
        skipped.append(str(e_id))

    success_text = ', '.join(successful_run)
    if len(skipped):
      fail_text = '(skipped %s)' % ', '.join(skipped)
    else:
      fail_text = ''

    return HttpResponse(f'Ran Extracts: {success_text} {fail_text}', status=200)
  else:
    return HttpResponse('Invalid request type. Only POST requests accepted', status=400)
