# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging
import base64
import os
from django.conf import settings
LOGGER = logging.getLogger("arcweb")

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.backends import default_backend

from sqlrun.jsonselect_cache import jsonselect_cached_function

def encrypt_secret(secret):
  # Generate a salt for use in the PBKDF2 hash
  salt = base64.b64encode(os.urandom(12))  # Recommended method from cryptography.io
  # Set up the hashing algo
  kdf = PBKDF2HMAC(
      algorithm=SHA256(),
      length=32,
      salt=salt,
      iterations=100000,  # This stretches the hash against brute forcing
      backend=default_backend(),  # Typically this is OpenSSL
  )
  # Derive a binary hash and encode it with base 64 encoding
  hashed_pwd = base64.b64encode(kdf.derive(settings.SECRET_KEY.encode()))

  # Set up AES in CBC mode using the hash as the key
  f = Fernet(hashed_pwd)
  encrypted_secret = f.encrypt(secret.encode())

  # Store the safe inputs in the DB
  return {
    'secret': encrypted_secret.decode(),
    'algo': 'pbkdf2_sha256',
    'iterations': '100000',
    'salt': salt.decode()
  }

def decrypt_secret(secret, salt, algo='pbkdf2_sha256', iterations='100000'):
  cache_key = (secret, salt, algo, iterations)
  return jsonselect_cached_function(cache_key, decrypt_secret_0, secret=secret, salt=salt, algo=algo, iterations=iterations)

def decrypt_secret_0(secret, salt, algo='pbkdf2_sha256', iterations='100000'):
  # Set up the Key Derivation Formula (PBKDF2)
  if isinstance(salt, str):
    salt = salt.encode()
  if isinstance(secret, str):
    secret = secret.encode()

  kdf = PBKDF2HMAC(
      algorithm=SHA256(),
      length=32,
      salt=salt,
      iterations=int(iterations),
      backend=default_backend(),
  )
  # Generate the key from the user's password
  key = base64.b64encode(kdf.derive(settings.SECRET_KEY.encode()))

  # Set up the AES encryption again, using the key
  f = Fernet(key)

  # Decrypt the secret!
  return f.decrypt(secret).decode()
