# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.conf import settings
from reports.views_settings import setting_get_single
import json
import logging
from django.core.mail import EmailMultiAlternatives

LOGGER = logging.getLogger('arcweb')

try:
  from jobs.views_encryption import decrypt_secret
except Exception as e:
  LOGGER.exception(str(e))

"""
Cancel Error
"""
class CancelTask(Exception):
  """Raise exception for cancelling task + schedule"""

def email_self(subject, body, email_settings=None, to_address=None):
  setup_email(email_settings)
  if not to_address:
    to_address = getattr(settings, 'EMAIL_HOST_USER')

  msg = EmailMultiAlternatives(subject=subject,
                               body=body,
                               to=[to_address], from_email=to_address)
  return msg.send()

def send_email(subject, body, to_addresses, from_email, email_settings=None):
  setup_email(email_settings)
  msg = EmailMultiAlternatives(subject=subject,
                            body=body,
                            from_email=from_email,
                            to=to_addresses)
  msg.send()

# Create or update attributes
def configure_settings(setting_name, setting_value):
  value_absent = not hasattr(settings, setting_name)
  if value_absent or (hasattr(settings, setting_name) and getattr(settings, setting_name) != setting_value):
    setattr(settings, setting_name, _process_email_value(setting_value))

def _process_email_value(value):
  if value == 'true':
    return True
  elif value == 'false':
    return False
  else:
    try:
        return int(value)
    except ValueError:
        return value

# Set up all email settings
def setup_email(post_values=None):
  email_settings = ['EMAIL_HOST', 'EMAIL_PORT',
                    'EMAIL_HOST_USER', 'EMAIL_HOST_PASSWORD',
                    'EMAIL_PROTOCOL',
                    ]
  default_email_settings = {'EMAIL_PROTOCOL': 'EMAIL_USE_TLS'}
  for setting in email_settings:
    is_password = setting.find("PASSWORD") != -1
    try:
      value = post_values[setting]
      if (value == '' and not is_password) or value == None:
        raise Exception('empty value')
      if is_password:
        # it's possible we are sending those
        # asterix so we have values but those are wrong
        # so we need the saved value from db
        if value == settings.DEFAULT_PASSWORD_PLACEHOLDER:
          value = setting_get_single(setting.lower())
          if value:
            # decrypt the value from db
            value = decrypt_secret(**json.loads(value))
    except:
      value = setting_get_single(setting.lower())
      if value and is_password:
        value = decrypt_secret(**json.loads(value))

    if value == None:
      try:
        value = default_email_settings[setting]
      except:
        raise Exception('Missing email site setting: %s' % setting)

    if setting == 'EMAIL_PROTOCOL':
      # Since TLS and SSL are mutually exclusive as of Django 1.7a1,
      # disable SSL if TLS enabled

      # reset
      configure_settings('EMAIL_USE_TLS', 'false')
      configure_settings('EMAIL_USE_SSL', 'false')

      # set the only one that's true
      if value != 'NEITHER':
        configure_settings(value, 'true')
    else:
      configure_settings(setting, value)

  # If username is specified, use that instead of email/EMAIL_HOST_USER
  if post_values:
    email_username = post_values.get('email_username')
  else:
    email_username = setting_get_single('email_username')

  if email_username:
    configure_settings('EMAIL_HOST_USER', email_username)
