# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import re
import time
import json
import logging
import threading

from django.conf import settings
from django.contrib.auth.models import User
from django.http import HttpResponse
from datasets.models import DataConnection
from dataconn.connections_types import DATACONN_ARCENGINE, DATACONN_ARCENGINE_V2, isArcEngineBased
from dataconn.dataconn import get_dataconn
from jobs.views_helpers import CancelTask
from vizlib.arcprof import ArcProf

import vizlib.workq

LOGGER = logging.getLogger("arcweb")
REFRESHERS = None

def setup_view_refresher():
  global REFRESHERS
  REFRESHERS = RefreshManager()

# Global set of refresh queues indexed by connection id
class RefreshManager(object):
  def __init__(self):
    self.lock = threading.Lock()
    self.refreshers = {}

  def get_or_create_ref(self, dcid):
    ref = self.refreshers.get(dcid)
    if ref:
      return ref

    ref = _Refresher(dcid)
    with self.lock:
      if self.refreshers.get(dcid) is None:
        self.refreshers[dcid] = ref

    return self.refreshers.get(dcid)

  def find_ref(self, dcid):
    return self.refreshers.get(dcid)

  def allrefs(self):
    return list(self.refreshers.values())

# Handle one connection, only allow one refresh per data connection
# at a time
class _Refresher(object):
  def __init__(self, dcid):
    self._dcid = dcid
    self._viewdict = {}
    self._lock = threading.Lock()
    self._dc_lock = threading.Lock()

  # Remove a view from the refresh list if present
  def _remove(self, view):
    with self._lock:
      runner = self._viewdict.get(view)
      if runner is not None:
        del self._viewdict[view]
      return runner

  # Enque a view to be refreshed
  def enque(self, user, view, schedule, joblog):
    with self._lock:
      if self._viewdict.get(view) is not None:
        joblog.add_task('Analytical View Refresh Already Queued')
        return
      runner = _RefreshRunner(self, self._dcid, user.id, view, schedule, joblog)
      self._viewdict[view] = runner

    # Some testing to verify we expect the refresh to proceed
    try:
      runner.validate()
    except:
      self._remove(view)
      raise

    try:
      with self._dc_lock:
        runner.run()
    except CancelTask as e:
      self._remove(view)
      raise CancelTask('Canceled')

  # Return a list of queue refresh jobs
  def status_list(self):
    return [r.get_status() for r in list(self._viewdict.values())]

  # Completion callback from the refresh task
  # TODO: figure out how to remove this from the workq structure
  def completed(self, view, status):
    self._remove(view)

def _enque_refresh(refresh_target, joblog=None):
  schedule = refresh_target.schedule
  target_data = json.loads(refresh_target.target_data)
  dcid = target_data['dcid']
  view = target_data['view']
  user = schedule.owner

  try:
    dc = DataConnection.objects.get(id=dcid)
  except Exception as e:
    exception_string = 'Invalid data connection id %s: %s' % (dcid, str(e))
    if joblog:
      joblog.update_log(last_task=exception_string, target_data=target_data)

  # Add dc name information if user has perms to the connection
  dc.checkPerm(user, 'dc_aviews')
  target_data['dataconn_name'] = dc.dataconnection_name
  refresh_target.target_data = json.dumps(target_data)
  refresh_target.save()

  if not isArcEngineBased(dc.dataconnection_type):
    exception_string = 'Data connection %s must be of type %s, not %s' % \
                       (dcid, DATACONN_ARCENGINE, dc.dataconnection_type)
    if joblog:
      joblog.update_log(last_task=exception_string, target_data=target_data)

  ref = REFRESHERS.get_or_create_ref(dcid)
  try:
    ref.enque(user, view, schedule, joblog)
    return ref
  except CancelTask as e:
    raise CancelTask('Canceled')
  except Exception as e:
    exception_string = 'Error starting refresh of %s: %s' % (view, str(e))
    joblog.update_log(last_task=exception_string, target_data=target_data)
    LOGGER.exception(exception_string)
    raise

def _fetch_refresh_status(request_data):
  dcid = None
  if request_data is not None and 'dcid' in request_data:
    dcid = request_data['dcid']

  rsp = []
  if not settings.IS_MASTER:
    LOGGER.debug('Refresh status fetched on non-master process?')
  elif dcid is None:
    for ref in REFRESHERS.allrefs():
      rsp.extend(ref.status_list())
  else:
    ref = REFRESHERS.find_ref(dcid)
    if ref:
      rsp.extend(ref.status_list())

  return HttpResponse(json.dumps({'views':rsp}), content_type="application/json")
    
class _RefreshRunner(vizlib.workq.ProgressRunner):
  def __init__(self, refq, dcid, userid, view, schedule, joblog):
    self._refq = refq
    self._dcid = dcid
    self._view = view
    self._userid = userid
    self._qhandle = None
    self._progress = 0
    self._start = 0
    self._remain = 0
    self._cancelled = False
    self._pollseconds = 2
    self._status = {'status': 'Queued'}
    self._error = None
    self._joblog = joblog
    self._schedule = schedule
    self._profile = ArcProf('RefreshRunner')

  def get_status(self):
    return {'dc_id':self._dcid,
            'view':self._view,
            'status':self._status,
            'error':self._error,
            'progress':self._progress,
            'start':self._start,
            'remain':self._remain,
            'cancelled':self._cancelled,
            'joblog_id':self._joblog.id,
            'schedule_id': self._schedule.id,
           }
    
  def set_cancel(self):
    self._cancelled = True
    self._status = {'status': 'Cancelled'}
    self._joblog.add_task('Analytical View Refresh Cancelled')

  def validate(self):
    viewname = self._view
    LOGGER.info('Validating %s refresh' % viewname)

    user = User.objects.get(id=self._userid)
    dc = get_dataconn(self._dcid)

    rsp = dc.show_analytical_view(viewname, username=user.username)
    colnames = rsp.get('colnames', [])

    dc = DataConnection.objects.get(id=self._dcid)
    if dc.dataconnection_type == DATACONN_ARCENGINE or dc.dataconnection_type == DATACONN_ARCENGINE_V2:
      rows = json.loads(rsp.get('rows', '[]'))
    else:
      rows = rsp.get('rows', [])

    if len(rows) == 1 and 'aview partition names' in colnames:
      row = rows[0]
      state = row[colnames.index('state')]
    elif len(rows) == 1 and 'Metadata' in colnames:
      row = json.loads(rows[0][colnames.index('Metadata')])
      state = row.get('state')
    else:
      raise Exception("Unable to get analytical view state")

    # These states just need normal refresh
    if state in ['UNUSABLE', 'USABLE', 'STALE', 'FRESH']:
      return

    if not getattr(settings, 'REFRESH_RECREATE_INVALID_AVIEWS', True):
      LOGGER.info('AView %s state %s -- skipping attempt to recreate aview' % (viewname, state))
      return

    # See if we can re-create the view.
    LOGGER.info('AView %s state %s -- attempting to recreate aview' % (viewname, state))
    ptext = ''
    if 'aview partition names' in colnames:
      partitions = json.loads(row[colnames.index('aview partition names')])
      ptext = 'partitioned by (%s)' % str(', '.join(partitions))

    select = ''
    if 'select' in colnames:
      select = row[colnames.index('select')]

    testname = viewname + '_' + str(int(time.time() * 1000))

    if select != '':
      cmd = 'create analytical view %s %s stored as parquet as %s' % \
            (dc.quote_tablename(testname), ptext, select)
      rsp = dc.sync_runquery(cmd, username=user.username)

      # If it worked, then drop the test name, drop the view name and re-create using view name
      dc.sync_runquery('drop analytical view ' + dc.quote_tablename(testname), username=user.username)
      dc.sync_runquery('drop analytical view ' + dc.quote_tablename(viewname), username=user.username)

      cmd = 'create analytical view %s %s stored as parquet as %s' % \
            (dc.quote_tablename(viewname), ptext, select)
      rsp = dc.sync_runquery(cmd, username=user.username)

  def _refresh(self):
    # Since we may have been queue for a while, re-validate some stuff
    user = User.objects.get(id=self._userid)
    django_dc = DataConnection.objects.get(id=self._dcid)
    django_dc.checkPerm(user, 'dc_aviews')
    
    self._start = time.time()
    dc = get_dataconn(self._dcid)

    update_string = 'Starting %s refresh' % self._view
    LOGGER.info(update_string)
    self._joblog.add_task(update_string)

    query = 'refresh analytical view ' + dc.quote_tablename(self._view)
    if dc.connection_type == DATACONN_ARCENGINE_V2:
      self._results = dc.sync_runquery(query, username=user.username)
      return

    rsp = dc.async_runquery(query, username=user.username)
    self._qhandle = rsp.get('qhandle')

    done = False
    try:
      while not done:
        time.sleep(self._pollseconds)
        self._status = dc.async_status(self._qhandle, username=user.username)
        done = not self._status['status'].split()[0] == 'Running'
        if not done:
          m = re.findall(r'\((\d+)/100 completed\)', self._status['status'])
          if m:
            self._progress = min(100, max(1, int(m[0])))
            elapsed = time.time() - self._start
            self._remain = max(self._pollseconds, int(((100.0 / self._progress) - 1) * elapsed))
            self._joblog.add_task('Progress is at %s' % self._progress)
    except CancelTask as e:
      dc.async_cancel(self._qhandle)
      self.set_cancel()
      raise CancelTask('Canceled')

    if self._status['status'] == 'Finished':
      self._results = dc.async_results(self._qhandle, username=user.username)
    else:
      raise Exception('Unexpected status on refresh: %s' % self._status)
              
  # Override
  def profile(self):
    return self._profile

  # Override
  def run(self):
    if not self._cancelled:
      self._status = {'status': 'Starting'}
      self._joblog.add_task(self._status)

      try:
        self._refresh()
      except CancelTask as e:
        raise CancelTask('Canceled')
      except Exception as e:
        exception_string = 'Exception on %s refresh' % self._view
        self._joblog.add_task(exception_string)
        LOGGER.exception(exception_string)
        self._error = str(e)
        self._status = {'status': 'Exception %s' % self._error}
        self._joblog.add_task('Refresh of view %s failed: %s' % (self._view, self._error))
        threading.Timer(4.0, lambda: self._refq.completed(self._view, self._status)).start()
        raise

      found = False
      error_string = ''
      rows = self._results.get('rows')

      # Backwards compatibility with pre 4.2 arcengine
      if len(rows[0]) == 1:
        # Put off completion so that front-end polling sees finished message
        threading.Timer(4.0, lambda: self._refq.completed(self._view, self._status)).start()
        self._joblog.add_task('Refresh of view %s completed' % self._view)
        return

      for viewname, status, comment in rows:
        if self._view == viewname.strip():
          found = True
          if status == 'was refreshed':
            # Put off completion so that front-end polling sees finished message
            threading.Timer(4.0, lambda: self._refq.completed(self._view, self._status)).start()
            self._joblog.add_task('Refresh of view %s completed' % self._view)
          else:
            error_string = 'Refresh of view %s completed with error: %s' % (self._view, comment)

          break

      if not found or error_string:
        if not found:
          error_string = 'No message from arcengine on fate of refresh of view %s' % self._view
        self._joblog.add_task(error_string)
        self._error = error_string
        self._status = {'status': 'Errored'}
        # Put off completion so that front-end polling notices error
        threading.Timer(4.0, lambda: self._refq.completed(self._view, self._status)).start()

        raise Exception(error_string)

  # Override
  def progress(self):
    """Returns a tuple that indicates (steps_completed, total_steps)"""
    return (0, 1)

  # Override
  def results(self):
    """Returns results of the function"""
    return self._results

