# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.db import IntegrityError
import jobs
from django.http import HttpResponse
import json
from django.contrib.auth.decorators import login_required
from django.core import serializers
from datetime import datetime

from rbac.permcache import RBAC

def user_can_schedule(user):
  return 'sys_schedules' in RBAC.get_user_obj_perms(user)

@login_required
def schedule_group(request, schedule_group_id=None):
  # REST endpoint for adding/getting schedule groups
  if request.method == 'DELETE':
    return jobs.views._delete_items(request.body, jobs.models.ScheduleGroup, request.user)
  elif request.method == 'GET':
    schedule_groups = {
      'schedule_groups': schedule_group_get(id=schedule_group_id),
    }
    return HttpResponse(json.dumps(schedule_groups), content_type="application/json")
  elif request.method == 'POST':
    if not user_can_schedule(request.user):
      return HttpResponse("You do not have permission to change schedule intervals", status=500)

    form_entries = json.loads(request.POST.get('formEntries'))
    form_entries['creator'] = request.user
    try:
      if schedule_group_id:
        schedule = jobs.models.ScheduleGroup.objects.filter(id=schedule_group_id)
        schedule.update(**form_entries)
        text = 'Updated'
        schedule = schedule[0]

        # Update next run time
        tzinfo = schedule.create_time.tzinfo
        now = datetime.now(tzinfo)
        schedule.update_next_run_time(now)
      else:
        schedule = jobs.models.ScheduleGroup.objects.create(**form_entries)
        schedule.update_next_run_time()
        text = 'Created'
    except IntegrityError as e:
      if 'UNIQUE constraint' in e.args[0]:
        return HttpResponse('A schedule group with this name already exists', status=400)
      else:
        return HttpResponse(e.args[0], status=500)

    return HttpResponse('Successfully {0} Interval #{1}'.format(text, schedule.id), status=200)

  return HttpResponse('Invalid request method', status=500)

def schedule_group_get(id=None, omit_details=False):
    schedule_groups = jobs.models.ScheduleGroup.objects.all()
    if id:
      schedule_groups = schedule_groups.filter(id=id)

    if not omit_details:
      schedule_groups = schedule_groups.select_related("creator").prefetch_related("jobschedule_set")

    all_schedule_groups = []
    for group in schedule_groups:

      if omit_details:
        jobschedule_set = None
      else:
        jobschedule_set = json.loads(serializers.serialize('json', group.jobschedule_set.all(), fields=('id', 'job_name')))

      group_json = { 'id': group.id,
          'schedule_name': group.schedule_name,
          'run_interval': group.run_interval,
          'next_run_time': str(group.next_run_time),
          'jobschedule_set': jobschedule_set,
        }

      creator = None if omit_details else group.creator
      if creator:
        group_json['creator'] = creator.id
        group_json['owner_username'] = creator.username

      all_schedule_groups.append(group_json)

    return all_schedule_groups
