# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from datasets.models import DataConnection, DataSet, Segment, Event, DatasetSegment
from reports.models import CustomCSS, Report, StaticAsset, AppGroup, \
  AppGroupAppMembership, DateRanges, ReportImage, ColorPalette
from reports.models import ENABLE_URL_NAMES
from django.core import serializers
from django.db.models import Q
from django.db import transaction
from base64 import b64decode
from rbac.models import RBAC
from django.test import RequestFactory
from workspace.views import workspace_internal
from workspace.models import Workspace
from rbac.permdefs import PermissionException
from django.core.exceptions import ObjectDoesNotExist
from datasources.dataimport import cmp_version

import re, json, logging, importlib

appgroup_migration = importlib.import_module('reports.migrations.0005_appgroup_workspace')
dashboard_migration = importlib.import_module('reports.migrations.0007_fix_app_scope_filters')
add_dataset_tablenames_migration = importlib.import_module('datasets.migrations.0022_add_dataset_tablenames_field')
LOGGER = logging.getLogger('arcweb')
IMPORT_ROLE_MAP = {
  DataSet:     {'create':'dc_explore',
                'update':'ds_manage'
               },
  Segment:      'ds_manage',
  Event:        'ds_manage',
  Report:       'ds_appedit',
  CustomCSS:    'sys_styles',
  StaticAsset:  'sys_styles'
}

class ParseException(Exception):
  def __init__(self, obj_name, id, exc):
    self.obj_name = obj_name
    self.id = id
    self.exc = exc

  def __str__(self):
    return "Cannot parse %s data with ID %s: %s => %s" % \
           (self.obj_name, self.id, self.exc.__class__.__name__, self.exc)

# Sanity check Exception
class MissingTablesException(Exception):
  def __init__(self, missing_tables):
    self.missing_tables = missing_tables

  def __str__(self):
    return "Required tables do not exist. List of missing tables: %s" % \
           ', '.join(self.missing_tables)

class NotUniqueException(Exception):
  def __str__(self):
    return 'Artifacts from this file have already been imported into multiple ' \
           'workspaces. Please check "Disregard artifacts that exist in other ' \
           'workspaces" checkbox and try again.'

# check if the user has the permission for performing the "action"
def check_perm(obj, user, action, new_dataset_objs, old_obj_data, dest_workspace_id):
  def check_workspace_permission():
    # make sure the user has edit permission on the source workspace
    request = RequestFactory()
    request.user = user
    request.method = 'GET'
    request.META = {}

    try:
      error_msg = 'An artifact will be moved from %s workspace during the import.' \
                  ' Edit or manage permission is required on %s workspace' % \
                  (obj.workspace.name, obj.workspace.name)
      workspace_info = json.loads(workspace_internal(request, obj.workspace_id).content)[0]
      perms = workspace_info['perms']
      if 'ws_appedit' not in perms and 'ws_manage' not in perms:
        raise PermissionException(user, error_msg, Workspace)

      #make sure the user has edit or manage permission on destination workspace
      error_msg = 'Artifacts will be created on the selected workspace.' \
                  ' Edit or manage permission is required on %s workspace' % \
                  Workspace.objects.get(id=dest_workspace_id).name
      workspace_info = json.loads(workspace_internal(request, dest_workspace_id).content)[0]
      perms = workspace_info['perms']
      if 'ws_appedit' not in perms and 'ws_manage' not in perms:
        raise PermissionException(user, error_msg, Workspace)
    except:
      raise PermissionException(user, error_msg, Workspace)

  if type(obj) == DataSet:
    if action['create']:
      dc = DataConnection.objects.get(pk=obj.dataconnection_id)
      dc.checkPerm(user, IMPORT_ROLE_MAP[DataSet]['create'])
      new_dataset_objs[old_obj_data['pk']] = obj
    elif action['data_update']:
      obj.checkPerm(user, IMPORT_ROLE_MAP[DataSet]['update'])
  elif type(obj) == Event:
    if action['create'] or action['data_update']:
      # if obj.dataset_id is None and we have reached this point then the user
      # has the permission to create the dataset; thus, she should have
      # the permission to do event for that dataset
      if obj.dataset_id is not None:
        ds = DataSet.objects.get(pk=obj.dataset_id)
        if ds.dataconnection is None:
          ds.dataconnection = DataConnection.objects.get(id=ds.dataconnection_id)
        ds.checkPerm(user, IMPORT_ROLE_MAP[type(obj)])
  elif type(obj) == Report:
    if action['create'] or action['data_update']:
      if obj.dataset_id is not None:
        ds = DataSet.objects.get(pk=obj.dataset_id)
      else:
        ds = new_dataset_objs[old_obj_data['fields']['dataset']]

      if ds.dataconnection is None:
        ds.dataconnection = DataConnection.objects.get(id=ds.dataconnection_id)
      ds.checkPerm(user, IMPORT_ROLE_MAP[type(obj)])
    if action['workspace_update']:
      check_workspace_permission()

  elif type(obj) == AppGroup:
    if action['workspace_update']:
      check_workspace_permission()
  elif type(obj) == Segment:
    # if obj.dataset_id is None and we have reached this point then the user
    # has the permission to create the dataset; thus, she should have
    # the permission to do segments for that dataset
    return True
  elif type(obj) in IMPORT_ROLE_MAP:
    if action['create'] or action['data_update']:
      RBAC.checkPerm(user, IMPORT_ROLE_MAP[type(obj)])


def is_sql(table):
  select_query = table.get('select')
  return select_query is not None

def get_table_source(table, is_trino):
  select_query = table.get('select')
  tablesource_from_query = re.findall('(FROM\s+)([^\s]+)', select_query, flags=re.IGNORECASE)[0][1]

  if is_trino:
    db_n_table_splitted = tablesource_from_query.rsplit('.', 1)
    return (db_n_table_splitted[0], db_n_table_splitted[1])

  return tablesource_from_query.split('.')

def dataset_sanity_check(data, username, dc_id, dc_type):
  isKSQL = (dc_type == 'ksql')
  is_trino = dc_type == 'trino'
  from util.avtool import jsonselect_runner, SHOW_TABLES_DSREQ, SHOW_DATABASES_DSREQ
  # check if all required tables are present in the data connection
  databases = []
  tables = []
  for dataset in data[DatasetSync.name]:
    dataset_info = json.loads(dataset['fields']['dataset_info'])
    for table in dataset_info:
      if is_sql(table):
        tablesource = get_table_source(table, is_trino)
        databases.append(tablesource[0])
        tables.append('.'.join(tablesource))
      else:
        if is_trino:
          databases.append(table['tablename'].rsplit('.', 1)[0])
        else:
          databases.append(table['tablename'].split('.')[0])

        if isKSQL:
          tables.append(table['tablename'].split('.')[1])
        else:
          tables.append(table['tablename'])

  jrunner = jsonselect_runner(username, dc_id=dc_id)
  table_names = []
  # For ksql, there are no databases (topics instead), so don't follow usual
  # route. Instead, get a list of all tables and streams
  if isKSQL:
    dsreq = SHOW_TABLES_DSREQ
    response = jrunner.run(dsreq)
    table_names += ["%s" % (t[0][0]) for t in json.loads(response.content)["rows"]]
    if not set(tables) <= set(table_names):
      raise MissingTablesException(list(set(tables) - set(table_names)))
    return

  # get the list of all databases
  databases = list(set(databases))
  all_databases = jrunner.run(SHOW_DATABASES_DSREQ)
  response_rows = json.loads(all_databases.content)["rows"]
  if type(response_rows) == str:
    response_rows = json.loads(response_rows)
  all_databases = [db[0] for db in response_rows]
  common_databases = list(set(databases) & set(all_databases))

  # get all tables names in the common databases
  for db in common_databases:
    dsreq = SHOW_TABLES_DSREQ
    dsreq["dbname"] = db
    response = jrunner.run(dsreq)
    response_rows = json.loads(response.content)["rows"]
    if type(response_rows) == str:
      response_rows = json.loads(response_rows)
    table_names += ["%s.%s" % (db, t[0]) for t in response_rows]

  # check if tables is subset of table_names
  if not set(tables) <= set(table_names):
    raise MissingTablesException(list(set(tables) - set(table_names)))

def get_data(class_obj, ids, excluded_fields=None, type='json'):
  excluded_fields = excluded_fields or []

  # serialize the data of an object
  filter = {'id__in': ids}
  objl = class_obj.objects.filter(**filter)
  all_fields = [f.name for f in class_obj._meta.fields]
  fields = tuple(set(all_fields) - set(excluded_fields))
  return serializers.serialize(type, objl, fields=fields)

# Add sheet information for pre-4.1.5 dashboards
# Add global_control_widgets for import/export
def addDashboardSheetData(new_report_data):
  new_report_data['global_control_widgets'] = []
  new_report_data.setdefault('control_widgets', [])
  new_report_data['dashboard_sheets'] = [{
    'sheet_id': 1,
    'sheet_handle_title': 'Sheet 1',
    'order': 1,
    'visual_widgets': new_report_data.get('dashboard_widgets', []),
    'control_widgets': new_report_data.get('control_widgets', []),
    'behaviors': new_report_data.get('click_behavior', {}),
  }]

  return new_report_data

class SyncBase(object):
  """
  parent class for migrating models. Each migratable model will inherit this class
  """
  name = None
  model = None
  dependencies_model = []
  excluded_fields = ['update_time',
                     'update_user',
                     'create_user',
                     'create_time']

  @classmethod
  def get_filter_column(cls, class_obj):
    # return the filter for finding the objects to be exported
    return {'id__in':cls.ids}

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    # update the field data prior to import. This could be any required update
    # such as updating the ids, and so on
    return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, obj):
    # return the name of the object (e.g. visual name, dashboard names)
    return {'name': cls.name}

  @classmethod
  def perform_deferred_update(cls, obj, ids_map):
    raise Exception('The method is not implemented')

class EventSync(SyncBase):
  ids = []
  name = "events"
  model = Event

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    if field_name == 'dataset':
      return 'dataset_id', ids_map[DatasetSync.name][field_value], True
    else:
      return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, data):
    return {'name': data['fields']['event_name']}

class SegmentSync(SyncBase):
  ids = []
  name = "segments"
  model = Segment

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    if field_name == 'dataset':
      return 'dataset_id', ids_map[DatasetSync.name][field_value], True
    elif field_name == 'segment_data':
      cmp_value = False
      field_value = json.loads(field_value)
      if obj_value:
        obj_value = json.loads(obj_value)
        if obj_value == field_value:
          cmp_value = True
      return field_name, json.dumps(field_value), cmp_value
    else:
      return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, data):
    return {'name': data['fields']['segment_name']}

  def update_ids_with_parse(segment_ids_map, values_string):
    segment_items = json.loads(values_string)
    updated_segment_items = SegmentSync.update_ids(segment_ids_map, segment_items)
    updated_values_string = json.dumps(updated_segment_items)
    return updated_values_string

  def update_ids(segment_ids_map, segment_items):
    return [[segment_ids_map[int(segment_id)], include_segment] for segment_id, include_segment in segment_items]

  def has_new_segment_ids(segment_ids_map):
    return all(segment_ids_map.values())

  @classmethod
  def update_field(cls, field_value, keys, ids_map):
    try:
      selected_segments = field_value['report_data']['selected_segments']
      for s in selected_segments:
        s['id'] = ids_map[SegmentSync.name][s['id']]
    except:
      pass

    # updating the segment id in the case that segment is dragged into the shelves
    try:
      for key, value in list(field_value['report_data'].items()):
        if type(value) == list:
          for v in value:
            if type(v) == dict and 'filter_data' in v:
              for f in v['filter_data']:
                if 'id' in f:
                  f['id'] = ids_map[SegmentSync.name][f['id']]
    except:
      pass

    if SegmentSync.has_new_segment_ids(ids_map[SegmentSync.name]):
      try:
        for shelf in field_value['report_data']['filters_shelf']:
          segment_ids_map = ids_map[SegmentSync.name]

          if SegmentSync.has_new_segment_ids(segment_ids_map):
            shelf['custom_expr'] = SegmentSync.update_ids_with_parse(segment_ids_map, shelf['custom_expr'])
            shelf['expression_for_trigger'] = SegmentSync.update_ids_with_parse(segment_ids_map, shelf['expression_for_trigger'])
            shelf['filter_input']['value_picker']['expr'] = SegmentSync.update_ids_with_parse(segment_ids_map, shelf['filter_input']['value_picker']['expr'])
            shelf['filter_input']['expr'] = SegmentSync.update_ids_with_parse(segment_ids_map, shelf['filter_input']['expr'])

            expression = shelf['filter_input']['expr']
            shelf['filter_input']['expression']['expr'] = f'{expression} as Segment'
            shelf['filter_input']['value_picker']['values'] = SegmentSync.update_ids(segment_ids_map, shelf['filter_input']['value_picker']['values'])

      except:
        pass

class DatasetSync(SyncBase):
  ids = []
  version_ids = []
  name = "datasets"
  model = DataSet
  dependencies_model = [SegmentSync, EventSync]
  excluded_fields = SyncBase.excluded_fields + ['dataset_everyone_perm',
                                                'dataset_create_user_perm',
                                                'dataset_lvinfo']

  @classmethod
  def get_filter_column(cls, class_obj):
    if class_obj in cls.dependencies_model:
      if class_obj == SegmentSync:
        return {'datasetsegment__dataset_version_id__in':cls.version_ids}
      else:
        return {'dataset__in':cls.ids}
    else:
      return {'id__in':cls.ids}

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    if field_name == 'dataconnection':
      return 'dataconnection_id', ids_map['dataconnection'], True
    if field_name == 'dashboards':
      # In VIZ-741 we did change the logic how we export datasets
      # Since than we export the datasets without the associated dashboard ids.
      # So here we are skip the 'obj_value == field_value' check
      # to pevent unnecessary dataset replacings.
      return field_name, field_value, True
    else:
      return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, data):
    return {'name': data['fields']['dataset_name']}

  @classmethod
  def extract_ids(cls, data):
    dataset_ids = []
    for viz in data:
      dataset_ids += [viz['fields']['dataset']]

    return list(set(dataset_ids))

  @classmethod
  def extract_version_ids(cls, ds_ids):
    return list(DataSet.objects.filter(pk__in=ds_ids).values_list('version_id', flat=True))


class DateRangesSync(SyncBase):
  ids = []
  name = "dateranges"
  model = DateRanges

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, data):
    return {'name': data['fields']['dateranges_name']}

  @classmethod
  def extract_ids(cls, data, key):
    date_ranges_names = []
    for viz in data:
      try:
        report_data = json.loads(viz['fields'][key])
        date_ranges_names.append(
          report_data['report_data']['user_settings']['oneclick_dateranges']
                                )
      except:
        continue

    return date_ranges_names

class StaticAssetSync(SyncBase):
  ids = []
  name = "staticasset"
  model = StaticAsset

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, data):
    return {'name': data['fields']['staticasset_name']}

  @staticmethod
  def get_staticasset_ids(data):
    p = re.compile(r'/reports/staticasset/raw/(\d+)\.')
    ids = []
    try:
      ids = p.findall(data)
    except:
      pass
    finally:
      return ids

  @classmethod
  def extract_ids(cls, data, key_list):
    staticasset_ids = []
    for viz in data:
      for k in key_list:
        staticasset_ids += cls.get_staticasset_ids(viz['fields'][k])
    return staticasset_ids

  @classmethod
  def update_field(cls, field_value, ids_map):
    for old_static_id in cls.get_staticasset_ids(field_value):
      field_value = re.sub(
        r'/reports/staticasset/raw/%s\.' % old_static_id,
        r'/reports/staticasset/raw/%s.' % ids_map[StaticAssetSync.name][int(old_static_id)],
        field_value)
    return field_value

class ColorPaletteSync(SyncBase):
  ids = []
  name = 'colorpalette'
  model = ColorPalette

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, data):
    return {'name': data['fields']['colorpalette_name']}

  @classmethod
  def extract_ids(cls, data):
    color_palette_ids = []
    for viz in data:
      report_data = json.loads(viz['fields']['report_data'])
      if 'colorSettings' in report_data['report_data']['settings'] and \
                  report_data['report_data']['settings']['colorSettings']:
        color_palette_ids.append(report_data['report_data']['settings']['colorSettings']['id'])

    return color_palette_ids

class CustomeCssSync(SyncBase):
  ids = []
  name = "customcss"
  model = CustomCSS
  dependencies_model = [StaticAssetSync]
  excluded_fields = SyncBase.excluded_fields

  @classmethod
  def get_filter_column(cls, class_obj):
    data = get_data(cls.model, cls.ids, type='python')
    if class_obj == StaticAssetSync:
      staticasset_ids = StaticAssetSync.extract_ids(data, ['customcss_value', 'customcss_js'])
      StaticAssetSync.ids += staticasset_ids
      return {'id__in': staticasset_ids}
    else:
      return {'id__in':cls.ids}

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    if field_name in ('customcss_value', 'customcss_js'):
      field_value = StaticAssetSync.update_field(field_value, ids_map)
    return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, data):
    return {'name': data['fields']['customcss_name']}

  @classmethod
  def extract_ids(cls, data, key):
    p = re.compile(r'cssid\\":\s*(\\")?(\d+)')
    css_ids = []
    for viz in data:
      css_ids += [int(i[1]) for i in p.findall(viz['fields'][key])]

    return css_ids

  @classmethod
  def update_field(cls, field_value, keys, ids_map):
    css_style = field_value
    try:
      # getting css_style json by going down the field_value nested dictionary
      # keys is the list of nested key in field_value
      for k in keys[:-1]: css_style = css_style[k]
      css_style = json.loads(css_style[keys[-1]])

      # updating the css ids in the new css_style dictionary
      for css in css_style:
        css['cssid'] = ids_map[CustomeCssSync.name][css['cssid']]

      # updating the json value of field_value
      for k in keys[:-1]: field_value = field_value[k]
      field_value[keys[-1]] = json.dumps(css_style)
    except:
      pass

class ReportImageSync(SyncBase):
  ids = []
  name = "reportimage"
  model = ReportImage
  dependencies_model = []
  excluded_fields = ['update_time']

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    if field_name == 'report':
      if field_value in ids_map[VisualSync.name]:
        return 'report_id', ids_map[VisualSync.name][field_value], True
      elif field_value in ids_map[DashboardSync.name]:
        return 'report_id', ids_map[DashboardSync.name][field_value], True
      else:
        raise "Thumbnails cannot find id mapping for report id %s" % field_value
    elif field_name == 'img':
      new_value = b64decode(field_value)
      return field_name, new_value, obj_value == new_value
    else:
      return field_name, field_value, obj_value == field_value

class VisualSync(SyncBase):
  ids = []
  name = "visuals"
  model = Report
  dependencies_model = [StaticAssetSync,
                        DateRangesSync,
                        CustomeCssSync,
                        ColorPaletteSync,
                        ReportImageSync,
                        DatasetSync]

  @classmethod
  def get_filter_column(cls, class_obj):
    data = get_data(cls.model, cls.ids, type='python')
    if class_obj == StaticAssetSync:
      staticasset_ids = StaticAssetSync.extract_ids(data, ['report_data'])
      StaticAssetSync.ids += staticasset_ids
      return {'id__in': staticasset_ids}
    elif class_obj == DateRangesSync:
      date_range_names = DateRangesSync.extract_ids(data, 'report_data')
      DateRangesSync.ids += date_range_names
      return {'dateranges_name__in':date_range_names}
    elif class_obj == CustomeCssSync:
      css_ids = CustomeCssSync.extract_ids(data, 'report_data')
      CustomeCssSync.ids += css_ids
      return {'id__in':css_ids}
    elif class_obj == DatasetSync:
      dataset_ids = DatasetSync.extract_ids(data)
      DatasetSync.ids += dataset_ids
      dataset_version_ids = DatasetSync.extract_version_ids(DatasetSync.ids)
      DatasetSync.version_ids += dataset_version_ids
      return {'id__in':dataset_ids}
    elif class_obj == ColorPaletteSync:
      color_palette_ids = ColorPaletteSync.extract_ids(data)
      ColorPaletteSync.ids += color_palette_ids
      return {'id__in': color_palette_ids}
    elif class_obj == ReportImageSync:
      return {'report_id__in': cls.ids}
    else:
      return {'id__in':cls.ids}

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    if field_name == 'dataset':
      return 'dataset_id', ids_map[DatasetSync.name][field_value], True
    elif field_name == 'report_data':
      cmp_value = False
      field_value = StaticAssetSync.update_field(field_value, ids_map)

      field_value = json.loads(field_value)
      CustomeCssSync.update_field(field_value,
                                  ['report_data', 'user_settings', 'css_styles'],
                                  ids_map)
      SegmentSync.update_field(field_value,
                               ['report_data', 'selected_segments'],
                               ids_map)

      # updating the link visual if there is any
      report_data = field_value['report_data']
      track_report = False
      try:
        old_id = report_data['click_behaviors']['primary']['report_id']
        if old_id in ids_map[VisualSync.name] and \
            ids_map[VisualSync.name][old_id] is not None:
          report_data['click_behaviors']['report_id'] = \
            ids_map[VisualSync.name][old_id]
        else:
          # the target visual has not been created yet, we need to update the
          # ID later when the target visual has been created
          track_report = True

      except:
        pass

      # updating dashboard_id which should be done as a deferred action
      # given that at this point we don't have the dashboard ids
      try:
        old_id = report_data['dashboard_id']
        track_report = True
      except:
        pass

      if obj_value:
        obj_value = json.loads(obj_value)

        # making the dashboard_id of obj_value the same as
        # the new data (field_value) to make sure the difference
        # does not come from dashboard_id value
        try:
          if 'dashboard_id' not in report_data:
            del obj_value['report_data']['dashboard_id']
          else:
            obj_value['report_data']['dashboard_id'] = report_data['dashboard_id']
        except:
          pass

        if obj_value == field_value:
          cmp_value = True

      return field_name, json.dumps(field_value), (cmp_value, track_report)
    elif field_name == 'shared_visual_dashboards':
      return field_name, field_value, True
    elif field_name == 'parent_report':
      return field_name, field_value, True
    else:
      return field_name, field_value, obj_value == field_value

  @classmethod
  def perform_deferred_update(cls, obj, ids_map):
    field_value = json.loads(obj.report_data)
    report_data = field_value['report_data']

    # updating click behavior
    try:
      old_id = report_data['click_behaviors']['primary']['report_id']
      report_data['click_behaviors']['primary']['report_id'] = \
        ids_map[VisualSync.name][old_id]
    except:
      # This may happen if the link visual is not exported with the visuals
      pass

    # updating dashboard id
    try:
      old_id = report_data['dashboard_id']
      report_data['dashboard_id'] = ids_map[DashboardSync.name][old_id]
    except:
      pass

    obj.report_data = json.dumps(field_value)
    obj.save()

  @classmethod
  def get_name(cls, data):
    response = {}
    try:
      report_data = json.loads(data['fields']['report_data'])
      response['type'] = report_data['report_type']
      response['name'] = report_data['report_data']['report_title']
    except KeyError:
      response['name'] = data['fields']['report_name']
    finally:
      return response

  @classmethod
  def extract_ids(cls, data, no_id_mapping=True):
    p = re.compile(r"^uri-(\d+)-widget-(\d+)$")
    visual_ids = []
    visual_in_dashboards = {}
    for viz in data:
      visual_in_dashboards[viz['pk']] = []
      report_data = json.loads(viz['fields']['report_data'])
      if 'global_control_widgets' in report_data:
        visual_ids += report_data['global_control_widgets']

      if 'dashboard_sheets' in report_data:
        # 4.1.5 release and beyonds
        for sheet in report_data['dashboard_sheets']:
          try:
            visual_ids += [int(p.search(x['id']).group(2))
                           for x in sheet['visual_widgets']]
          except:
            continue

          try:
            visual_ids += [int(i) for i in sheet['control_widgets']]
          except:
            continue

          try:
            for id, config in list(sheet['behaviors']['reports'].items()):
              if 'primary' in config and 'report_id' in config['primary']:
                linked_id = int(config['primary']['report_id'])
                if Report.objects.get(id=linked_id).report_type == 'dashboard':
                  DashboardSync.ids.append(linked_id)
                else:
                  visual_ids.append(linked_id)
          except:
            continue

      else:
        try:
          ids = [int(p.search(x['id']).group(2))
                         for x in report_data['dashboard_widgets']]
          visual_ids += ids
          for id in ids:
            if id not in visual_in_dashboards:
              visual_in_dashboards[id] = []
            visual_in_dashboards[id].append(viz['pk'])
        except KeyError:
          continue
        try:
          ids = [int(i) for i in report_data['control_widgets']]
          visual_ids += ids
          for id in ids:
            if id not in visual_in_dashboards:
              visual_in_dashboards[id] = []
            visual_in_dashboards[id].append(viz['pk'])
        except KeyError:
          continue
    if no_id_mapping:
      return list(set(visual_ids))
    else:
      return visual_in_dashboards

class DashboardSync(SyncBase):
  ids = []
  name = "dashboards"
  model = Report
  dependencies_model = [StaticAssetSync,
                        CustomeCssSync,
                        ReportImageSync,
                        VisualSync]

  @classmethod
  def get_filter_column(cls, class_obj):
    data = get_data(cls.model, cls.ids, type='python')
    if class_obj == StaticAssetSync:
      staticasset_ids = StaticAssetSync.extract_ids(data, ['report_data'])
      StaticAssetSync.ids += staticasset_ids
      return {'id__in': staticasset_ids}
    elif class_obj == VisualSync:
      visual_ids = VisualSync.extract_ids(data)
      VisualSync.ids += list(set(visual_ids))
      return {'id__in':visual_ids}
    elif class_obj == CustomeCssSync:
      css_ids = CustomeCssSync.extract_ids(data, 'report_data')
      CustomeCssSync.ids += css_ids
      return {'id__in':css_ids}
    elif class_obj == ReportImageSync:
      return {'report_id__in': cls.ids}
    else:
      return {'id__in':cls.ids}

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    if field_name == 'dataset':
      return 'dataset_id', ids_map[DatasetSync.name][field_value], True
    elif field_name == 'report_data':
      cmp_value = False
      field_value = StaticAssetSync.update_field(field_value, ids_map)
      field_value = json.loads(field_value)
      CustomeCssSync.update_field(field_value,
                                  ['user_settings', 'css_styles'],
                                  ids_map)
      track_report = DashboardSync.update_field(field_value, [], ids_map)
      if 'dashboard_sheets' not in field_value:
        # prior 4.1.5, populate some default values
        field_value = addDashboardSheetData(field_value)

      if 'control_widgets' in field_value:
        field_value.pop('control_widgets', None)

      if obj_value:
        obj_value = json.loads(obj_value)
        if obj_value == field_value:
          cmp_value = True
      return field_name, json.dumps(field_value), (cmp_value, track_report)
    elif field_name == 'shared_visual_dashboards':
      try:
        ids = [ids_map[VisualSync.name][int(id)] for id in json.loads(field_value)]
        return field_name, json.dumps(ids), True
      except:
        return field_name, field_value, True
    else:
      return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, data):
    return {'name': data['fields']['report_name']}

  @classmethod
  def perform_deferred_update(cls, obj, ids_map):
    field_value = json.loads(obj.report_data)
    for sheet in field_value['dashboard_sheets']:
      for id, config in list(sheet['behaviors']['reports'].items()):
        try:
          old_id = int(config['primary']['report_id'])
          if old_id in ids_map[VisualSync.name]:
            config['primary']['report_id'] = ids_map[VisualSync.name][old_id]
          else:
            config['primary']['report_id'] = ids_map[DashboardSync.name][old_id]
        except:
          continue

    obj.report_data = json.dumps(field_value)
    obj.save()

  @classmethod
  def update_field(cls, field_value, keys, ids_map):
    widget_pattern = re.compile(r"^uri-(\d+)-widget-(\d+)$")
    def update_dashboard_widget(value):
      for i in range(len(value)):
        w = widget_pattern.search(value[i]['id'])
        old_id = int(w.group(2))
        value[i]['id'] = \
          'uri-%s-widget-%s' % (w.group(1), ids_map[VisualSync.name][old_id])

    def update_control_widget(value):
      ids = []
      if 'control_widgets' in value:
        for i in range(len(value['control_widgets'])):
          try:
            ids.append(ids_map[VisualSync.name][int(value['control_widgets'][i])])
          except KeyError:
            pass

        value['control_widgets'] = ids

    update_dashboard_widget(field_value['dashboard_widgets'])

    # update global_control_widgets
    if 'global_control_widgets' in field_value:
      picklist_ids = []
      for id in field_value['global_control_widgets']:
        picklist_ids.append(ids_map[VisualSync.name][id])
      field_value['global_control_widgets'] = picklist_ids

    if 'time_control_settings' in field_value:
      # Need to get updated dataset id for time warp widget
      # For now, we expect to receive only 1 dataset reference
      time_warp_ids = {}
      time_warp_settings = field_value['time_control_settings']
      for k in time_warp_settings:
        if k == 'dataset_id':
          ds_id = time_warp_settings[k]
          time_warp_ids.update({'dataset_id': ids_map[DatasetSync.name][ds_id]})

      field_value['time_control_settings'] = {**time_warp_settings, **time_warp_ids}

    track_report = False
    if 'dashboard_sheets' in field_value:
      # 4.1.5 release and beyond
      for sheet in field_value['dashboard_sheets']:
        update_dashboard_widget(sheet['visual_widgets'])
        update_control_widget(sheet)
        if 'behaviors' in sheet and \
            'reports' in sheet['behaviors']:
          new_click_behavior = {}
          for id, config in list(sheet['behaviors']['reports'].items()):
            id = int(id)
            try:
              old_id = int(config['primary']['report_id'])
              if old_id in ids_map[VisualSync.name] and \
                  ids_map[VisualSync.name][old_id] is not None:
                config['primary']['report_id'] = ids_map[VisualSync.name][old_id]
              else:
                # the id where this dashboard is linked into is not available yet
                # need to update it later when all ids are available
                track_report = True
            except:
              pass

            # this should not happen, but apparently there is a bug in
            # populating click_behavior dict
            if id in ids_map[VisualSync.name]:
              new_click_behavior[str(ids_map[VisualSync.name][id])] = config
          sheet['behaviors']['reports'] = new_click_behavior
    else:

      if 'click_behavior' in field_value and \
         'reports' in field_value['click_behavior']:
        new_click_behavior = {}
        for id, config in list(field_value['click_behavior']['reports'].items()):
          id = int(id)
          # this should not happen, but apparently there is a bug in populating click_behavior dict
          if id in ids_map[VisualSync.name]:
            new_click_behavior[str(ids_map[VisualSync.name][id])] = config
        field_value['click_behavior']['reports'] = new_click_behavior

    return track_report

class AppGroupAppMembershipSync(SyncBase):
  ids = []
  name = "appgroupmembership"
  model = AppGroupAppMembership
  dependencies_model = [DashboardSync, VisualSync]

  @classmethod
  def get_filter_column(cls, class_obj):
    dashboard_visual_ids = AppGroupAppMembershipSync.model.objects.filter(
              **{'appgroup_group_id__in': AppGroupAppMembershipSync.ids}
        ).values_list('appgroup_app_id', flat=True)
    if class_obj == DashboardSync:
      ids = DashboardSync.model.objects.filter(
                **{'id__in':dashboard_visual_ids,
                   'report_type':'dashboard'}).values_list('id', flat=True)
      DashboardSync.ids += ids
      return {'id__in': ids}
    elif class_obj == VisualSync:
      ids = [i for i in dashboard_visual_ids if i not in DashboardSync.ids]
      VisualSync.ids += ids
      return {'id__in': ids}
    else:
      return AppGroupSync.get_filter()

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    if field_name == 'appgroup_app':
      if field_value in ids_map[VisualSync.name]:
        id = ids_map[VisualSync.name][field_value]
      elif field_value in ids_map[DashboardSync.name]:
        id = ids_map[DashboardSync.name][field_value]
      else:
        assert "cannot find %s id in ids map"
      return 'appgroup_app_id', id, True
    elif field_name == 'appgroup_group':
      return 'appgroup_group_id', ids_map[AppGroupSync.name][field_value], True
    else:
      return field_name, field_value, obj_value == field_value

  @classmethod
  def extract_ids(cls, filter):
    objl = AppGroupSync.model.objects.filter(filter)
    ids = []
    for obj in objl:
      ids.append(obj.pk)
    return ids

class AppGroupSync(SyncBase):
  ids = []
  name = "appgroups"
  model = AppGroup
  dependencies_model = [StaticAssetSync, CustomeCssSync, AppGroupAppMembershipSync]

  @classmethod
  def get_filter(cls):
    filter_dict = {'id__in':cls.ids, 'group_parent_id__in':cls.ids}
    filter = Q()
    for key in filter_dict:
      filter |= Q(**{key: filter_dict[key]})
    return filter

  @classmethod
  def get_filter_column(cls, class_obj):
    data = get_data(cls.model, cls.ids, type='python')
    if class_obj == StaticAssetSync:
      staticasset_ids = StaticAssetSync.extract_ids(data, ['group_options'])
      StaticAssetSync.ids += staticasset_ids
      return {'id__in': staticasset_ids}
    elif class_obj == CustomeCssSync:
      css_ids = CustomeCssSync.extract_ids(data, 'group_options')
      CustomeCssSync.ids += css_ids
      return {'id__in':css_ids}
    elif class_obj == AppGroupAppMembershipSync:
      ids = AppGroupAppMembershipSync.extract_ids(AppGroupSync.get_filter())
      AppGroupAppMembershipSync.ids += ids
      return {'appgroup_group_id__in':ids}
    else:
      return AppGroupSync.get_filter()

  @classmethod
  def prepare_for_import(cls, field_name, field_value, obj_value, ids_map):
    if field_name == 'dataset':
      return 'dataset_id', ids_map[DatasetSync.name][field_value], True
    elif field_name == 'group_options':
      cmp_value = False
      field_value = StaticAssetSync.update_field(field_value, ids_map)
      if field_value:
        field_value = json.loads(field_value)
        CustomeCssSync.update_field(field_value,
                                    ['user_settings', 'css_styles'],
                                    ids_map)
      if obj_value:
        obj_value = json.loads(obj_value)
        if obj_value == field_value:
          cmp_value = True
      return field_name, json.dumps(field_value), cmp_value
    elif field_name == 'group_parent' and field_value is not None:
      return 'group_parent_id', ids_map[AppGroupSync.name][field_value], True
    else:
      return field_name, field_value, obj_value == field_value

  @classmethod
  def get_name(cls, data):
    return {'name': data['fields']['group_name']}

# recursive call to iterate through all models and their dependencies
def all_model_iter(class_obj, recursive_call=False):
  if not recursive_call:
    yield class_obj, class_obj
  for c in class_obj.dependencies_model:
    yield c, class_obj
    if c.dependencies_model:
      for cc in all_model_iter(c, recursive_call=True):
        yield cc[0], cc[1]

# order in which the objects should be created on import
import_order = [DatasetSync,
                SegmentSync,
                EventSync,
                StaticAssetSync,
                CustomeCssSync,
                ColorPaletteSync,
                DateRangesSync,
                VisualSync,
                DashboardSync,
                ReportImageSync,
                AppGroupSync,
                AppGroupAppMembershipSync]

class_name_map = dict((c.name, c) for c in import_order)

def export_objects(class_name, ids, user):
  """
   main function for exporting objects (models) with their dependencies
  :param class_name: name of corresponding class for migrating a model
  :param ids: list of model ids
  :return:
  """

  class_object = class_name_map[class_name]
  objs = class_object.model.objects.filter(id__in=ids)
  if len(objs) != len(ids):
    objs_ids = objs.values_list('id', flat=True)
    raise ValueError('%s with ID(s) %s does not exist' %
            (class_name, ', '.join([str(i) for i in ids if i not in objs_ids])))
  data = dict((c.name, []) for c, p in all_model_iter(class_object))
  for c in import_order:
    if c.model == DataSet:
      c.version_ids = []
    c.ids = []
    data[c.name] = []
  class_object.ids = ids
  for c, p in all_model_iter(class_object):
    filter = p.get_filter_column(c)
    if isinstance(filter, dict):
      objl = c.model.objects.filter(**filter)
    else:
      objl = c.model.objects.filter(filter)
    if c.model == DataSet:
      for ds in objl:
        ds.checkPerm(user, 'ds_appedit')
    all_fields = [f.name for f in c.model._meta.fields]
    fields = tuple(set(all_fields) - set(c.excluded_fields))

    resource_data = serializers.serialize('python', objl, fields=fields)

    if c.model == Segment:
      for segment in resource_data:
        segment_data = json.loads(segment['fields']['segment_data'])
        dataset_id = DatasetSegment.objects.filter(segment_id=segment['pk']).select_related('dataset').first().dataset.version_group_id
        segment_data['dataset_id'] = dataset_id
        updated_segment_data = json.dumps(segment_data)
        segment['fields']['segment_data'] = updated_segment_data

    data[c.name] += resource_data
  return data

def create_obj_from_dic(class_obj, obj_json, ids_map, workspace_limit, workspace_id, track_artifacts):
  # check if the artifact is already imported

  filter_keys = {}
  if class_obj == DatasetSync:
    filter_keys.update({'dataconnection': ids_map['dataconnection']})

  elif class_obj == EventSync:
    filter_keys.update({'dataset': ids_map[DatasetSync.name][obj_json['fields']['dataset']]})

  elif class_obj == SegmentSync:
    segment_data = json.loads(obj_json['fields']['segment_data'])
    dataset_id = segment_data.get('dataset_id')
    dataset_version_id = ids_map[DatasetSync.name].get(dataset_id, None)
    filter_keys.update({'datasetsegment__dataset_version_id': dataset_version_id})
    if 'dataset' in obj_json['fields']:
      segment_data = json.loads(obj_json['fields']['segment_data'])
      segment_data['dataset_id'] = obj_json['fields']['dataset']
      obj_json['fields']['segment_data'] = json.dumps(segment_data)
      obj_json['fields'].pop('dataset')


  elif class_obj in (VisualSync, DashboardSync):
    filter_keys.update({'dataset__dataconnection__id': ids_map['dataconnection']})
    if workspace_limit is not None:
      filter_keys.update({'workspace_id': workspace_limit})

  elif class_obj == AppGroupSync and workspace_limit is not None:
    filter_keys.update({'workspace_id': workspace_limit})

  elif class_obj == ReportImageSync:
    try:
      id = ids_map[VisualSync.name][obj_json['fields']['report']]
    except:
      id = ids_map[DashboardSync.name][obj_json['fields']['report']]

    filter_keys.update({'report': id})

  matched_objects = class_obj.model.objects.filter(imported_uuid=obj_json['fields']['uuid'], **filter_keys)

  if not matched_objects.exists():
    matched_objects = class_obj.model.objects.filter(uuid=obj_json['fields']['uuid'], **filter_keys)


  if len(matched_objects) > 1 and not workspace_limit:
    raise NotUniqueException()

  action = {'create': False,
            'data_update': False,
            'workspace_update': False,
            'thumbnail_update': False}
  if len(matched_objects) >= 1:
    obj = matched_objects[0]
  else:
    obj = class_obj.model()
    action['create'] = True
    obj.workspace_id = workspace_id

  for f_name, f_value in list(obj_json['fields'].items()):
    if f_name in 'uuid':
      imported_uuid = f_value
      continue
    elif f_name in ('imported_uuid', 'workspace', 'group_visible'):
      continue
    f_name, f_value, cmp_boolean = class_obj.prepare_for_import(
                                                          f_name,
                                                          f_value,
                                                          getattr(obj, f_name),
                                                          ids_map)
    if f_name == 'parent_report':
      f_value = None
    if f_name != 'dashboards':
      # Since the introduction of versioning we have two extra identifier on the datasets.
      # During the import process if it is a data update / Replace action on a dataset
      # we do not want to update these fields.
      if class_obj == DatasetSync and cmp_boolean is False and f_name in ['version_id', 'version_group_id']:
        continue
      else:
        setattr(obj, f_name, f_value)
    if type(cmp_boolean) is tuple:
      if cmp_boolean[1]:
        # the artifact is not updated fully (one example would be the
        # dashboard/visual that this artifact is linked into was not available)
        # We are going to track this artifact to update it later by calling
        # the perform_deferred_update function of that artifact class
        track_artifacts.append({'name': class_obj.name,
                             'id': obj_json['pk']})
      cmp_boolean = cmp_boolean[0]

    if cmp_boolean is False:
      action['data_update'] = True

  # Here we only allow setting of the url name if it is unique, otherwise clear it.
  if ENABLE_URL_NAMES and class_obj is DashboardSync:
    try:
      report_data = json.loads(obj.report_data or '{}')
      dashboard_url_name = report_data.get('user_settings', {}).get('dashboard_url_name')
      if dashboard_url_name:
        idlist = list(Report.objects.filter(report_type='dashboard',
                                            dashboard_url_name=dashboard_url_name)
                                    .exclude(id=getattr(obj, 'id', 0))
                                    .values_list('id', flat=True))
        if len(idlist):
          LOGGER.warning('Dashboard URL name "%s" already used by IDs %s' % (dashboard_url_name, idlist))
          report_data['user_settings']['dashboard_url_name'] = ''
          obj.report_data = json.dumps(report_data)

    except Exception as e:
      LOGGER.exception('Error on dashboard_url_name check')

  # Same url name logic for AppGroup
  if ENABLE_URL_NAMES and class_obj is AppGroupSync:
    try:
      group_options = json.loads(obj.group_options or '{}')
      group_url_name = group_options.get('user_settings', {}).get('app_url_name')
      if group_url_name:
        idlist = list(AppGroup.objects.filter(group_url_name=group_url_name)
                                    .exclude(id=getattr(obj, 'id', 0))
                                    .values_list('id', flat=True))
        if len(idlist):
          LOGGER.warning('App Group URL name "%s" already used by IDs %s' % (group_url_name, idlist))
          group_options['user_settings']['app_url_name'] = ''
          obj.group_options = json.dumps(group_options)

    except Exception as e:
      LOGGER.exception('Error on group_url_name check')

  setattr(obj, 'imported_uuid', imported_uuid)
  return obj, action

@transaction.atomic
def import_objects(data,
                   dataconnection_id,
                   user,
                   dry_run,
                   skip_permission_check,
                   ids_map=None,
                   report_mode=None,
                   workspace_id=None,
                   limit_search_scope=False,
                   skip_thumbnails=False):
  """
  main function for importing objects (models) with their dependencies
  :param data: json data representing the models to be imported
  :param dataconnection_id: id of the connection to bes used for creating datasets
  :param user: user that is issuing the import command
  :param dry_run: if True just do the import evaluation
  :return:
  """
  if 'version' in data:
    version = '.'.join(data['version']['Arcviz Version'].split('.')[0:2])
  else:
    raise Exception('The json file does not contain the version number')

  ids_map = ids_map or {}
  all_filters = None
  report = {}
  import_order_names = [c.name for c in import_order]
  keys = list(data.keys())
  keys.remove('version')
  keys.sort(key=lambda x: import_order_names.index(x))
  if skip_thumbnails:
    keys.remove(ReportImageSync.name)
  ids_map['dataconnection'] = dataconnection_id
  track_artifacts = []
  new_dataset_obj = {}
  new_shared_visual_dashboards = {}
  parent_group = None

  for obj_name in keys:
    sync_model = class_name_map[obj_name]
    objs_data = data[obj_name]
    ids_map[sync_model.name] = {}
    report[sync_model.name] = []
    if obj_name == AppGroupSync.name:
      # in the case of app group, make sure the parent app group goes first
      objs_data = sorted(objs_data,
                         key=lambda x: -1
                            if x == 'null' or
                              not isinstance(x['fields']['group_parent'], int)
                            else x['fields']['group_parent'])
      if len(objs_data) > 0 and cmp_version(version, '4.4') < 0:
        # apps are exported prior to 4.4; need to make sure the parent group does not
        # have any dashboard. we keep track of the parent group and perform the operation
        # later on. Note that prior to 4.4 you can only export one group, so we are sure
        # there is only one parent group and that is the first one on the list
        parent_group = objs_data[0]
    for obj_data in objs_data:
      # do changes
      try:
        if limit_search_scope:
          workspace_search_limit = workspace_id
        else:
          workspace_search_limit = None
        obj, action = create_obj_from_dic(sync_model,
                                          obj_data,
                                          ids_map,
                                          workspace_search_limit,
                                          workspace_id,
                                          track_artifacts)
      except NotUniqueException:
        raise
      except Exception as e:
        LOGGER.exception(str(e))
        raise ParseException(obj_name, obj_data['pk'], e)

      if obj_name in (DashboardSync.name, VisualSync.name, AppGroupSync.name):
        if obj.workspace_id != workspace_id:
          action['workspace_update'] = True
        if report_mode is not None:
          obj.report_mode = report_mode

      if obj_name == DatasetSync.name:
        obj.version_group_id = obj.id
        obj.dataset_tablenames = json.dumps(add_dataset_tablenames_migration.get_tablenames(obj))

      if obj_name == ReportImageSync.name and \
         action['data_update'] and \
         obj.report_id is not None:
        # mark the action for corresponding visual/dashboard to indicate its
        # thumbnail will be update.
        if obj.report_id in ids_map[VisualSync.name].values():
          parent_report = [viz_report for viz_report in report[VisualSync.name]
                           if viz_report['id'] == obj.report_id][0]
        else:
          parent_report = [viz_report for viz_report in report[DashboardSync.name]
                           if viz_report['id'] == obj.report_id][0]
        parent_report['action']['thumbnail_update'] = True

      if obj_name == DashboardSync.name:
        if cmp_version(version, '4.4') < 0:
          if not all_filters:
            all_filters = dict(Report.objects.filter(report_type__in=['picklist', 'daterange'])
                                              .values_list('id', 'report_data'))
          report_data = dashboard_migration._populateGlobalControlWidgets(
              json.loads(obj.report_data), all_filters)
          obj.report_data = json.dumps(report_data)

      if action['create']:
        obj.create_user = user.username
        obj.update_user = user.username
      elif action['data_update'] or action['workspace_update']:
        obj.update_user = user.username

      if not skip_permission_check:
        check_perm(obj, user, action, new_dataset_obj, obj_data, workspace_id)

      if obj_name in (DashboardSync.name, VisualSync.name, AppGroupSync.name):
        # need to update workspace_id after check_perm call because check_perm call
        # will treat the current obj.workspace_id as the source workspace_id
        obj.workspace_id = workspace_id

      if not dry_run:

        obj.save()


      # recording the visuals that may need update to their
      # shared_visual_dashboards values
      if obj_name == DashboardSync.name and obj.shared_visual_dashboards is not None:
        for viz_id in json.loads(obj.shared_visual_dashboards):
          if viz_id not in new_shared_visual_dashboards:
            new_shared_visual_dashboards[viz_id] = []
          new_shared_visual_dashboards[viz_id].append(obj.pk)
      ids_map[sync_model.name][obj_data['pk']] = obj.pk
      rep = {
        'id': obj.pk,
        'action': action
      }
      rep.update(sync_model.get_name(obj_data))
      report[sync_model.name].append(rep)

      if not dry_run:

        obj.save()
        if obj_name == 'segments':
          segment_data = json.loads(obj_data['fields']['segment_data'])
          dataset_id = segment_data.get('dataset_id')
          dataset_version_id = ids_map[DatasetSync.name].get(dataset_id, None)
          ds = DataSet.objects.get(version_id=int(dataset_version_id))
          datasetSegment, created = DatasetSegment.objects.get_or_create(dataset=ds, segment_id=obj.id, dataset_version_id=ds.version_id)
          if created:
            LOGGER.info("A new DatasetSegment was created with the id:%s", datasetSegment.id)
          else:
            LOGGER.info("DatasetSegment already exists with id:%s", datasetSegment.id)

  if not dry_run and parent_group:
    # apps are exported prior to 4.4; need to make sure the parent group does not
    # have any dashboard in it
    parent_group = AppGroup.objects.get(id=ids_map[AppGroupSync.name][parent_group['pk']])

    # check if update_parentgroup is already run
    already_run = False
    subgroup = AppGroup.objects.filter(group_parent=parent_group.id, imported_uuid='upgrade-auto')
    if subgroup.count() > 0:
      already_run = True

    if not already_run:
      appgroup_migration.update_parentgroup(parent_group)
    else:
      # put all dashboards in the parent_group to the new subgroup created as part of previous
      # import. Also set that subgroup to be the first subgroup
      subgroup = subgroup[0]
      parent_group.appgroupappmembership_set.update(appgroup_group_id=subgroup.id)
      from django.db.models import Min
      min_group_order = AppGroup.objects.filter(group_parent=parent_group.id).\
          aggregate(Min('group_subgroup_order'))
      min_group_order = list(min_group_order.values())[0]
      subgroup.group_subgroup_order = min_group_order / 2
      subgroup.save()


  # Updating shared_visual_dashboards values for visuals. Note that, this update
  # cannot be done in the previous loop because the ID of dashboards is not
  # available yet when we are creating/updating the visuals. The order in which
  # the objects get created is such that visuals get created first and then dashboards.
  viz_to_update = list(new_shared_visual_dashboards.keys())
  for viz in report[VisualSync.name]:
    if viz['id'] in viz_to_update:
      if viz['action']['create'] is not True:
        viz_obj = Report.objects.get(id=viz['id'])
        viz_obj.shared_visual_dashboards = json.loads(viz_obj.shared_visual_dashboards or '[]')
        for id in viz_obj.shared_visual_dashboards:
          if id in new_shared_visual_dashboards[viz['id']] or id is None:
            continue
          else:
            # remove any id from shared_visual_dashboards that is not valid. Note
            # that we do not know IDs of dashboards that are removed/added on
            # this system.
            try:
              Report.objects.get(id=id)
            except ObjectDoesNotExist:
              viz_obj.shared_visual_dashboards.remove(id)
              viz['action']['data_update'] = True

        # adding the remaining IDs in new_shared_visual_dashboards
        new_ids = list(set(
                new_shared_visual_dashboards[viz['id']] +
                viz_obj.shared_visual_dashboards))
        if set(viz_obj.shared_visual_dashboards)^set(new_ids):
          viz_obj.shared_visual_dashboards = new_ids
          viz['action']['data_update'] = True

      if not dry_run:
        if viz['action']['create']:
          viz_obj = Report.objects.get(id=viz['id'])
          viz_obj.shared_visual_dashboards = new_shared_visual_dashboards[viz['id']]
          viz_obj.save()
        elif viz['action']['data_update']:
          viz_obj.update_user = user.username
          viz_obj.save()

  # if the artifacts are imported from prior 4.1.5 version, process all visuals
  # to decide which one should be linked and which one should not. Note that by
  # default all visuals are private
  # check to see how many dashboards visual is used in
  #   - if used in none, should be linked
  #   - if used in 1, should be unlinked
  #   - if used in 2 or more, should be linked
  if len(data[VisualSync.name]) > 0 and \
                  'shared_visual_dashboards' not in data[VisualSync.name][0]['fields']:
    visual_dash_ids = VisualSync.extract_ids(data[DashboardSync.name],
                                             no_id_mapping=False)
    inv_ids_map = dict((v, k) for k, v in list(ids_map[VisualSync.name].items()) if v is not None)
    for viz in report[VisualSync.name]:
      if viz['id'] is None:
        continue
      old_id = inv_ids_map[viz['id']]
      if old_id not in visual_dash_ids:
        # visual is not used in any dashboard
        if not dry_run:
          viz_obj = Report.objects.get(id=ids_map[VisualSync.name][old_id])
          viz_obj.shared_visual_dashboards = []
          viz_obj.save()

      elif len(visual_dash_ids[old_id]) > 1:
        # visual is used in multiple dashboards
        viz['action']['data_update'] = True
        if not dry_run:
          viz_obj = Report.objects.get(id=ids_map[VisualSync.name][old_id])
          viz_obj.shared_visual_dashboards = [ids_map[DashboardSync.name][id] for
                                            id in visual_dash_ids[old_id]]
          viz_obj.save()
      else:
        if not dry_run:
          viz_obj = Report.objects.get(id=ids_map[VisualSync.name][old_id])
          report_data = json.loads(viz_obj.report_data)
          if 'dashboard_id' not in report_data:
            report_data['dashboard_id'] = \
              ids_map[DashboardSync.name][visual_dash_ids[old_id][0]]
          viz_obj.report_data = json.dumps(report_data)
          viz_obj.save()

  if not dry_run:
    # performing any deferred action
    for artifact in track_artifacts:
      sync_model = class_name_map[artifact['name']]
      old_id = artifact['id']
      obj = sync_model.model.objects.get(id=ids_map[sync_model.name][old_id])
      sync_model.perform_deferred_update(obj, ids_map)

  return report
