# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import django.test
from django.core.files.uploadedfile import SimpleUploadedFile
import json
from unittest import mock

from apps.tests.userrole import base
from datasets.models import DataConnection, Segment, DataSet
from vizlib.tests import test_report_operations


class TestMigrate(django.test.TestCase):

    def prepare(self):
        admin = base.make_user('admin', is_superuser=True)
        admin_client = base.create_client(admin)

        dc_id = test_report_operations.TestDashboardOperations().create_dataconnection(admin_client)
        dc_name = DataConnection.objects.get(id=dc_id)

        return admin_client, dc_name

    @mock.patch('migration.views.import_objects')
    def test_api_import(self, import_objects_mock: mock.Mock):

        import_objects_mock.return_value={
            "visuals": [],
            "dashboards": [],
        }

        admin_client, dc_name = self.prepare()
        response = admin_client.post('/migration/api/import/',{
            'dataconnection_name': dc_name.dataconnection_name,
            'sanity_check': 'False',
            'dry_run': 'False',
            'json_data': json.dumps({
                "version": {
                    "Arcviz Version": "7.1.2"
                }
            })
        })
        self.assertEquals(response.status_code, 200)

        # there is a dry run first, then the real one:
        self.assertEquals(import_objects_mock.call_count, 2)
        self.assertEquals(import_objects_mock.call_args_list[0].args[3], True)
        self.assertEquals(import_objects_mock.call_args_list[1].args[3], False)

    @mock.patch('migration.views.dataset_sanity_check')
    @mock.patch('migration.views.import_objects')
    def test_ui_import_with_sanity_check(self, import_objects_mock: mock.Mock, sanity_check_mock: mock.Mock):

        import_objects_mock.return_value={
            "visuals": [],
            "dashboards": [],
        }

        admin_client, dc_name = self.prepare()

        artifacts = {
                "version": {
                    "Arcviz Version": "7.1.2"
                }
            }

        file = SimpleUploadedFile("artifacts.json", json.dumps(artifacts).encode())

        response = admin_client.post('/migration/migrate',{
            'dataconnection_name': dc_name.dataconnection_name,
            'sanity_check': 'True',
            'dry_run': 'False',
            'import_file': file
        })
        self.assertEquals(response.status_code, 200)
        sanity_check_mock.assert_called_once()

        # there is only a dry run:
        self.assertEquals(import_objects_mock.call_count, 1)

        # dry_run passed to import method is True despite of POST argument above
        self.assertEquals(import_objects_mock.call_args_list[0].args[3], True)

    def get_artifacts(self):
        return {
            "dashboards":[
                {
                    "model": "reports.report",
                    "pk": 1728,
                    "fields": {
                        "report_name": "Dashboard on dataset with segment",
                        "report_description": "placeholder desc",
                        "dataset": 125,
                        "workspace": 2,
                        "report_type": "dashboard",
                        "uuid": "b3086154-204d-4db2-91ce-251bbf708840",
                        "report_data": json.dumps({
                            "dashboard_sheets": [
                                {
                                    "sheet_id": 1,
                                    "visual_widgets": [
                                        {
                                            "id": "uri-1-widget-1727"
                                        }
                                    ]
                                }
                            ],
                            "dashboard_widgets": [
                                {
                                "id": "uri-1-widget-1727"
                                }
                            ],
                        })
                    }
                }
            ],
            "datasets":[
                {
                    "model": "datasets.dataset",
                    "pk": 125,
                    "fields": {
                        "dataconnection": 2,
                        "dataset_name": "Cereals Clone",
                        "dataset_detail": "{}",
                        "dataset_type": "singletable",
                        "uuid": "55c88221-6bbd-4800-9531-bdddd146b71f",
                    }
                }
            ],
            "segments":[
                {
                    "model": "datasets.segment",
                    "pk": 48,
                    "fields": {
                        "dataset": 125,
                        "segment_name": "Segment cloned with dataset",
                        "uuid": "cba852ed-2241-4a16-9c88-4d7ed905e16c",
                        "segment_data": json.dumps({
                            "applyToNewVisuals": False,
                            "id": 8,
                            "dataset_id": 86 # this is probably a bug due to cloning
                        })
                    }
                }
            ],
            "visuals": [
              {
                "model": "reports.report",
                "pk": 1727,
                "fields": {
                    "report_name": "",
                    "report_description": "",
                    "report_type": "table",
                    "dataset": 125,
                    "uuid": "db6e2df8-412e-4f00-8a6b-1d8c2f070e13",
                    "report_data": json.dumps({
                        "report_type": "table",
                        "report_data": {
                            "core": {
                                "viz_type": "table",
                            }
                        }
                    })
                }
              }
            ],
            "version": {
                "Arcviz Version": "7.1.2"
            }
        }

    def test_api_import_with_segments(self):

        admin_client, dc_name = self.prepare()

        artifacts = self.get_artifacts()

        self.assertEquals(len(Segment.objects.all()), 0, 'no segments')

        response = admin_client.post('/migration/api/import/',{
            'dataconnection_name': dc_name.dataconnection_name,
            'sanity_check': 'False',
            'dry_run': 'False',
            'json_data': json.dumps(artifacts)
        })

        self.assertEquals(response.status_code, 200)
        self.assertEquals(len(Segment.objects.all()), 1, '1 segment imported')



    def test_when_import_dashboard_to_existing_dataset_should_associate_the_new_dashboard_and_keep_the_old_ones(self):
        # Given
        admin_client, dc = self.prepare()
        ds = DataSet.objects.create(dataset_name='dataset1', dataconnection=dc, dataset_detail='test', dataset_type='singletable', uuid='55c88221-6bbd-4800-9531-bdddd146b71f')
        ds.dashboards = json.dumps([10, 11])
        ds.save()

        artificate = {
            "dashboards": [
                {
                    "model": "reports.report",
                    "pk": 1728,
                    "fields": {
                        "report_name": "TEST",
                        "report_description": "placeholder desc",
                        "dataset": ds.id,
                        "workspace": 2,
                        "report_type": "dashboard",
                        "uuid": "b3086154-204d-4db2-91ce-251bbf708840",
                        "report_data": json.dumps({
                            "dashboard_sheets": [
                                {
                                    "sheet_id": 1,
                                    "visual_widgets": [
                                        {
                                            "id": "uri-1-widget-1727"
                                        }
                                    ]
                                }
                            ],
                            "dashboard_widgets": [
                                {
                                "id": "uri-1-widget-1727",
                                }
                            ],
                        })
                    }
                }
            ],
            "datasets": [
                {
                    "model": "datasets.dataset",
                    "pk": ds.id,
                    "fields": {
                        "dataconnection": ds.dataconnection.id,
                        "dataset_name": ds.dataset_name,
                        "dataset_detail": ds.dataset_detail,
                        "dataset_type": ds.dataset_type,
                        "uuid": "55c88221-6bbd-4800-9531-bdddd146b71f",
                        "dashboards": "[]",
                    }
                }
            ],
            "segments": [],
            "visuals": [
              {
                "model": "reports.report",
                "pk": 1727,
                "fields": {
                    "report_name": "",
                    "report_description": "",
                    "report_type": "table",
                    "dataset": ds.id,
                    "uuid": "db6e2df8-412e-4f00-8a6b-1d8c2f070e13",
                    "report_data": json.dumps({
                        "report_type": "table",
                        "report_data": {
                            "core": {
                                "viz_type": "table",
                            }
                        }
                    })
                }
              }
            ],
            "version": {
                "Arcviz Version": "7.1.2"
            }
        }
        # When
        admin_client.post('/migration/api/import/', {
            'dataconnection_name': dc.dataconnection_name,
            'sanity_check': 'False',
            'dry_run': 'False',
            'json_data': json.dumps(artificate)
        })
        # Then
        new_ds = DataSet.objects.get(id=ds.id)
        self.assertEqual(json.loads(new_ds.dashboards), [10, 11, 2])
