# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Views on migration
"""
import reports

from distutils.util import strtobool
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpRequest
from django.template import loader
from datasets.dashboard_association import DashboardAssociation
from migration.migrate import class_name_map,\
                              MissingTablesException,\
                              export_objects,\
                              import_objects,\
                              dataset_sanity_check,\
                              ParseException,\
                              VisualSync,\
                              DashboardSync
from datasets.models import DataConnection
from django.core.exceptions import ObjectDoesNotExist
from version.version import get_version_info
from datasources.views import _datasource_save_uploaded_file
from rbac.permdefs import PermissionException, PERMDEFS
from vizlib.misc import check_login_or_use_apikey
from django.views.decorators.csrf import csrf_exempt
from apps.utils import _appflags
from workspace.models import Workspace, get_or_create_private_ws, get_default_workspace
from apps.views_snapshot import _snapshot_take

import json
import logging


LOGGER = logging.getLogger('arcweb')

def merge_dicts(dict1, dict2):
  merge_dict = {}
  for k in set(list(dict1.keys()) + list(dict2.keys())):
    l = dict1.get(k, [])
    l.extend(x for x in dict2.get(k, []) if x not in l)
    merge_dict[k] = l
  return merge_dict

@csrf_exempt
def export_api(request):
  res = check_login_or_use_apikey(request)
  if res is not None:
    return res

  try:
    return export(request, api_call=True)
  except Exception as e:
    error_message = str(e)
    LOGGER.exception(error_message)
    return HttpResponse(error_message, status=500)

@csrf_exempt
def import_api(request):
  res = check_login_or_use_apikey(request)
  if res is not None:
    return res

  if request.method != 'POST':
    return HttpResponse('only post request is accepted for import', status=500)

  try:
    uploaded_data = get_uploaded_data(request)
  except ValueError as ve:
    print('\nERROR`: %s\n' %str(ve))
    return HttpResponse(str(ve), status=406)
  except Exception as e:
    return HttpResponse('Uploaded file could not be saved:%s\n' %
                        (str(e)), status=500)

  dc_name = request.POST.get('dataconnection_name', None)
  sanity_check = request.POST.get('sanity_check', 'True')
  dry_run = request.POST.get('dry_run', 'True')
  workspace = request.POST.get('workspace', 'Private')
  limit_search_scope = request.POST.get('limit_search_workspace', 'False')
  skip_thumbnails = request.POST.get('skip_thumbnails', 'False')
  generate_thumbnails = request.POST.get('generate_thumbnails', 'False')
  if workspace == 'Private':
    workspace_id = get_or_create_private_ws(request.user).id
  else:
    try:
      workspace_id = Workspace.objects.get(name=workspace).id
    except ObjectDoesNotExist:
      return HttpResponse('%s workspace does not exist' % workspace)

  res = _import(request, dc_name=dc_name, uploaded_data=uploaded_data,
                sanity_check=sanity_check, dry_run=dry_run,
                workspace_id=workspace_id, limit_search_scope=limit_search_scope,
                skip_thumbnails=skip_thumbnails,
                generate_thumbnails=generate_thumbnails)
  if res.status_code != 200:
    return HttpResponse('Error happened during import: %s' % res.content, status=500)

  return HttpResponse(res.content, content_type='application/json')

@login_required
def migrate(request):
  # function for filling the required parameters (except dry_run parameter) and
  # dispatching the import request generated by UI
  error = None
  if request.method == 'POST':
    # First request for import. User selected the file and submitted the form
    try:
      uploaded_file_path = _datasource_save_uploaded_file(request.FILES['import_file'])
    except ValueError as ve:
      return HttpResponse(str(ve), status=406)
    except Exception as e:
      return HttpResponse('Uploaded file could not be saved:%s\n' %
                          (str(e)), status=500)

    dc_name = request.POST.get('dataconnection_name', None)
    sanity_check = request.POST.get('sanity_check', None)
    sanity_check = 'True' if sanity_check else 'False'
    workspace_id = request.POST.get('workspace_id', None)
    limit_search_scope = request.POST.get('limit_search', 'False')
    thumbnails_action = request.POST.get('thumbnails_action', '1')
    thumbnails_action = int(thumbnails_action)
    if thumbnails_action == 1:
      skip_thumbnails = False
      generate_thumbnails = False
    elif thumbnails_action == 2:
      skip_thumbnails = True
      generate_thumbnails = True
    else:
      skip_thumbnails = True
      generate_thumbnails = False
    dry_run = 'True'

  elif request.method == 'PUT':
    # second and last request for import. User accepts and confirms the planned import
    try:
      uploaded_file_path = request.session['arc_import_objects']['uploaded_file_path']
      dc_name = request.session['arc_import_objects']['dc_name']
      sanity_check = request.session['arc_import_objects']['sanity_check']
      workspace_id = request.session['arc_import_objects']['workspace_id']
      limit_search_scope = request.session['arc_import_objects']['limit_search_scope']
      skip_thumbnails = request.session['arc_import_objects']['skip_thumbnails']
      generate_thumbnails = request.session['arc_import_objects']['generate_thumbnails']
      dry_run = request.GET.get('dry_run', 'True')
    except Exception as e:
      return HttpResponse('Cannot get the session parameters; %s' % str(e), status=500)
  else:
    return HttpResponse('Unknown request method', status=500)

  uploaded_data = parse_file(uploaded_file_path)

  res = _import(request, dc_name=dc_name, uploaded_data=uploaded_data,
                sanity_check=sanity_check, dry_run=dry_run,
                workspace_id=workspace_id, limit_search_scope=limit_search_scope,
                skip_thumbnails=skip_thumbnails,
                generate_thumbnails=generate_thumbnails)
  if res.status_code != 200:
    # if the error happens, we still let
    # the render happens and then display the error
    error = res.content.decode()
    res.content = '{}'

  request.session['arc_import_objects'] = {
    'uploaded_file_path': uploaded_file_path,
    'dc_name': dc_name,
    'sanity_check': sanity_check,
    'workspace_id': workspace_id,
    'limit_search_scope': limit_search_scope,
    'skip_thumbnails': skip_thumbnails,
    'generate_thumbnails': generate_thumbnails
  }
  template = loader.get_template('admin_migrate.html')
  c = { 'objects_to_import': res.content.decode(), 'error_on_post': error }
  c.update(_appflags(request))

  if request.method == 'POST':
    return HttpResponse(template.render(c, request))
  else:
    return HttpResponse(res.content, content_type='application/json')

def get_uploaded_data(request):
  if 'import_file' in request.FILES:
    uploaded_file_path = _datasource_save_uploaded_file(request.FILES['import_file'])
    data = parse_file(uploaded_file_path)
  elif request.POST.get('json_data'):
    data = json.loads(request.POST.get('json_data'))
  else:
    raise ValueError('The request has to contain either an uploaded file or a stringified json in the form encoded body under the json_data key')

  return data

def parse_file(file_path):
  with open(file_path, 'r') as f:
    return json.loads(f.read())

def merge_export_objects_to_data(obj_name, ids, user, data):
  result = export_objects(obj_name, ids, user)
  data = merge_dicts(data, result)
  return data

@login_required
def export(request, api_call=False):
  # handle export request
  data = {}
  if request.method != 'GET':
    return HttpResponse('Only GET method is supported', status=500)

  dry_run = request.GET.get('dry_run', 'True')
  try:
    dry_run = strtobool(dry_run)
  except:
    return HttpResponse('dry_run argument should be true or false', status=500)

  if not api_call:
    filename = request.GET.get('filename', None)
    if filename is None:
      return HttpResponse('filename should be specified', status=500)
  else:
    # in api call the client will write the content into a file
    filename = 'dummy_file'

  try:
    for obj_name in list(class_name_map.keys()):
      ids = json.loads(request.GET.get('%s' % obj_name, '[]'))
      if not isinstance(ids, list):
        return HttpResponse('%s values should have list format' % obj_name, status=500)
      if ids:
        data = merge_export_objects_to_data(obj_name, ids, request.user, data)
  except PermissionException as e:
    msg = 'Cannot export the visual artifacts: ' \
          'You do not have the necessary permission. "%s" role is required' % PERMDEFS[e.perm]
    LOGGER.error(msg)
    return HttpResponse(msg, status=500)

  if data:
    dashboard_fields = json.loads(data['dashboards'][0]['fields']['report_data'])
    if 'time_control_settings' in dashboard_fields:
      dataset_id_for_time_control = dashboard_fields['time_control_settings']['dataset_id']
      if dataset_id_for_time_control not in [ds['pk'] for ds in data['datasets']] :
        merge_export_objects_to_data('datasets', [dataset_id_for_time_control], request.user, data)

  for i in range(0,len(data['datasets'])):
    data['datasets'][i]['fields']['dashboards'] = []

  if dry_run:
    response = {}
    for key, value in list(data.items()):
      response[key] = []
      for v in value:
        report = {'id':v['pk']}
        report.update(class_name_map[key].get_name(v))
        response[key].append(report)

    response['version'] = get_version_info()
    return HttpResponse(json.dumps(response), content_type='application/json')
  else:
    data['version'] = get_version_info()
    response = HttpResponse(content_type='text/plain')
    response['Content-Disposition'] = 'attachment; filename="%s.json"' % filename
    json.dump(data, response)
    return response

def import_file(request, dc_name, filename, sanity_check, dry_run, workspace_id,
            limit_search_scope, skip_thumbnails, generate_thumbnails=False,
            upgraded_datasets=None):
  data = parse_file(filename)
  return _import(request, dc_name, data, sanity_check, dry_run, workspace_id,
    limit_search_scope, skip_thumbnails, generate_thumbnails,
    upgraded_datasets)

def _import(request, dc_name, uploaded_data, sanity_check, dry_run, workspace_id,
            limit_search_scope, skip_thumbnails, generate_thumbnails=False,
            upgraded_datasets=None):

  data = uploaded_data

  # Handling import request. This function can be either called by UI through
  # migrate function defined above or directly by passing the arguments through URL
  try:
    dry_run = strtobool(dry_run)
  except:
    return HttpResponse('dry_run argument should be true or false', status=500)

  try:
    sanity_check = strtobool(sanity_check)
  except:
    return HttpResponse('sanity_check argument should be true or false', status=500)

  try:
    limit_search_scope = strtobool(limit_search_scope)
  except:
    return HttpResponse('limit_search argument should be true or false', status=500)

  try:
    if isinstance(skip_thumbnails, str):
      skip_thumbnails = strtobool(skip_thumbnails)
  except:
    return HttpResponse('limit_search argument should be true or false', status=500)

  try:
    if isinstance(generate_thumbnails, str):
      generate_thumbnails = strtobool(generate_thumbnails)
  except:
    return HttpResponse('generate_thumbnails argument should be true or false', status=500)

  if dc_name is None:
    return HttpResponse('data connection name should be specified', status=500)
  else:
    try:
      dataconnection_obj = DataConnection.objects.get(dataconnection_name=dc_name)
      dataconnection_id = dataconnection_obj.id
      dataconnection_type = dataconnection_obj.dataconnection_type
    except ObjectDoesNotExist:
      return HttpResponse('data connection "%s" does not exist' % dc_name, status=500)

  if not workspace_id:
    workspace_id = get_default_workspace(request.user, 'ws_appedit').id
  else:
    workspace_id = int(workspace_id)

  if sanity_check:
    try:
      dataset_sanity_check(data, request.user.username, dataconnection_id, dataconnection_type)
    except MissingTablesException as e:
      msg = 'Data Table Compatibility Check Failed. %s' % str(e)
      LOGGER.error(msg)
      return HttpResponse(msg, status=500)
    except Exception as e:
      msg = 'Cannot Perform Data Table Compatibility Check. %s' % str(e)
      LOGGER.exception(msg)
      return HttpResponse(msg, status=500)

  if upgraded_datasets != None:
    data['datasets'] = upgraded_datasets

  # first call with dry_run equals to true to make sure
  # the user has all required permissions. This is to avoid performing
  # partial import in the case the user has some but not all required permissions.
  try:
    report = import_objects(data, dataconnection_id, request.user, True, False,
                            workspace_id=workspace_id,
                            limit_search_scope=limit_search_scope,
                            skip_thumbnails=skip_thumbnails)
  except PermissionException as e:
    if e.obj == Workspace:
      msg = """ %s.""" % e.perm
    else:
      msg = str(e) + "."
      msg += """ "%s" role is required.""" % PERMDEFS[e.perm]
    LOGGER.exception(msg)
    return HttpResponse(msg, status=500)
  except ParseException as e:
    error_message = str(e)
    LOGGER.exception(error_message)
    return HttpResponse(error_message, status=500)
  except Exception as e:
    error_message = str(e)
    LOGGER.exception(error_message)
    return HttpResponse(error_message, status=500)

  if not dry_run:
    report = import_objects(data, dataconnection_id, request.user, dry_run, True,
                            workspace_id=workspace_id,
                            limit_search_scope=limit_search_scope,
                            skip_thumbnails=skip_thumbnails)
    if generate_thumbnails:
      viz_dash_ids = [viz['id'] for viz in report[VisualSync.name]]
      viz_dash_ids += [viz['id'] for viz in report[DashboardSync.name]]

      mock_request = HttpRequest()
      mock_request.user = request.user
      for id in viz_dash_ids:
         app = reports.models.Report.objects.get(id=int(id))
         _snapshot_take(mock_request, app)


    datasets = report.get('datasets')
    dashboards = report.get('dashboards')

    if datasets and dashboards:
      dashboard_ids = [d['id'] for d in dashboards]
      for dataset in datasets:
        DashboardAssociation().associate(dataset['id'], dashboard_ids)

  if generate_thumbnails:
    # set all thumbnail_update to true for visuals and dashboards
    for viz in report[VisualSync.name]:
      viz['action']['thumbnail_update'] = True
    for viz in report[DashboardSync.name]:
      viz['action']['thumbnail_update'] = True


  return HttpResponse(json.dumps(report), content_type='application/json')
