# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

"""
Migrate old style permission to new role based scheme
"""

import os
import sys
import json
import copy
import django.apps.registry

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

from datasets.models import DataSet, DataConnection
from datasets.utils import get_default_dataconnection
from django.contrib.auth.models import User, Group, Permission
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned

DRYRUN = False
try:
  from rbac.models import Role
  from rbac.permdefs import PERMLISTS
except:
  PERMLISTS = {'system':[], 'dataconn':[], 'dataset':[]}
  settings.EVERYONE_GROUP = 'Everyone'
  DRYRUN = True

def fatal(msg):
  raise Exception(msg)

defroles = [

  {'name' : 'System Admin',
   'description' : 'Full access to all features',
   'groups' : [],
   'privs' : [
      {'ptype' : 'system',
       'perms' : PERMLISTS['system'],
      },
      {'ptype' : 'dataconn',
       'perms' : PERMLISTS['dataconn'],
       'dclist' : ["-1"],
      },
      {'ptype' : 'dataset',
       'perms' : PERMLISTS['dataset'],
       'dcid'  : "-1",
       'dslist': ["-1"],
      },
    ],
  },

  {'name' : 'Database admin',
   'description' : 'Connection level management and dataset creation',
   'groups' : [],
   'privs' : [
      {'ptype' : 'system',
       'perms' : ['sys_viewlogs'],
      },
      {'ptype' : 'dataconn',
       'perms' : PERMLISTS['dataconn'],
       'dclist' : ["-1"],
      },
      {'ptype' : 'dataset',
       'perms' : PERMLISTS['dataset'],
       'dcid'  : "-1",
       'dslist': ["-1"],
      },
    ],
  },

  {'name' : 'Analyst',
   'description' : 'Create and manage visual apps',
   'groups' : [],
   'privs' : [
      {'ptype' : 'system',
       'perms' : [],
      },
      {'ptype' : 'dataset',
       'perms' : ['ds_appedit', 'ds_appview'],
       'dcid'  : "-1",
       'dslist': ["-1"],
      },
    ],
  },

  {'name' : 'Visual Consumer',
   'description' : 'View access to visual apps',
   'groups' : [],
   'privs' : [
      {'ptype' : 'system',
       'perms' : [],
      },
      {'ptype' : 'dataset',
       'perms' : ['ds_appview'],
       'dcid'  : "-1",
       'dslist': ["-1"],
      },
    ],
  },
]

def printRole(rd, dcByID=None, dsByID=None):

  def dcname(id):
    id = int(id)
    if id == -1:
      return "All Connections"
    if dcByID and dcByID.get(id):
      return dcByID.get(id).dataconnection_name
    return "Connection %s" % id

  def dsname(id):
    id = int(id)
    if id == -1:
      return "All Datasets"
    if dsByID and dsByID.get(id):
      return dsByID.get(id).dataset_name
    return "Dataset %s" % id

  print()
  print("Role: %s" % rd.get('name'))

  ulist = rd.get('users', [])
  if len(ulist):
    print("  Users:      %s" % ", ".join(ulist))

  glist = rd.get('groups', [])
  if len(glist):
    print("  Groups:     %s" % ", ".join(glist))

  # Don't need to print dataconnection names if in the role name
  printconn = rd.get('name').find('on connection') == -1
  for p in rd.get('privs')[1:]:

    # Don't need to print the data connections since the dataset entry
    # names the same connection
    if False and  p['ptype'] == 'dataconn':
      if printconn:
        print("  Connection: %s" % \
              ", ".join(dcname(id) for id in p['dclist']))

    if p['ptype'] == 'dataset':
      if printconn:
        print("  Connection: %s" % dcname(p['dcid']))
      print("  Dataset:    %s" % ", ".join(dsname(id) for id in p['dslist']))

def createRoles(roles):
  for dr in roles:
    try:
      (r, created) = Role.objects.get_or_create(name=dr['name'], defaults={
        'description': dr['description'],
        'groups': json.dumps(dr.get('groups', [])),
        'users': json.dumps(dr.get('users', [])),
        'privileges': json.dumps(dr['privs'])})
      if created:
        print("Created role: %s" % r.name)
    except MultipleObjectsReturned:
      pass

def defaultRoles(dryrun=False):
  if not dryrun:
    createRoles(defroles)
  else:
    for rd in defroles:
      printRole(rd)

# Migration routine for old style permissions
def migratePerms(dryrun=False, merge=False, apps:django.apps.registry.Apps=None):

  global DataSet
  global DataConnection

  if apps:
    # this is when we migrate the database:
    DataSet = apps.get_model('datasets', 'DataSet')
    DataConnection = apps.get_model('datasets', 'DataConnection')

  dsl = DataSet.objects.all()

  # If no datasets, assume new system and just need default roles
  if len(dsl) == 0:
    defaultRoles(dryrun)
    return

  default_dc = get_default_dataconnection()
  dcl = DataConnection.objects.all()

  dcByID = {}
  for dc in dcl:
    if not dc.dataconnection_name:
      dc.dataconnection_name = "Data connection %s" % dc.id
      if not dryrun:
        dc.save()
    dcByID[dc.id] = dc

  dsByID = {}
  for ds in dsl:
    if ds.dataconnection_id is None:
      if not default_dc:
        fatal("Unable to locate default data connection")
      ds.dataconnection_id = default_dc.id
      if not dryrun:
        ds.save()
    if not ds.dataset_name:
      ds.dataset_name = "Dataset %s" % ds.id
      if not dryrun:
        ds.save()
    if dcByID.get(ds.dataconnection_id) is None:
      fatal("Dataset %s (ID %s) unknown dataconnection id (%s)" %
             (ds.dataset_name, ds.id, ds.dataconnection_id))
    dsByID[ds.id] = ds

  # Don't do anything if some roles are already present
  if not dryrun and len(Role.objects.all()) > 0:
    print("Existing roles found, exit without additional role migration")
    return

  if not dryrun:
    createRoles(defroles[0:1])

  ptypes = ['setupperm', 'readwrite', 'readonly']

  roles = {}
  for p in ptypes:
    roles[p] = []

  # Get membership list for each group
  groupMembers = {}
  for g in Group.objects.all():
    groupMembers[g.name] = g.user_set.values_list('username', flat=True)

  # Helper function to remove any user entries that are members of the
  # goups at the same permission level.
  # Also remove any superusers as they have all permissions
  def mergeUsers(m):
    udrop = set()
    for u in m.uset:
      ul = User.objects.filter(username=u)
      if len(ul) == 1 and ul[0].is_superuser:
        udrop.add(u)
        continue
      for g in m.gset:
        if g == settings.EVERYONE_GROUP or u in groupMembers[g]:
          udrop.add(u)
          break
    if len(udrop):
      m.uset = m.uset.difference(udrop)

  for dc in dcl:
    for ds in [ds for ds in dsl if ds.dataconnection_id == dc.id]:

      # Get the users and groups that have each permission on this dataset
      members = {}
      uset = set()
      gset = set()
      for perm in ptypes:
        codename = "%s_%s_%s" % ("can", perm, ds.id)
        users = User.objects.filter(user_permissions__codename=codename)
        groups = Group.objects.filter(permissions__codename=codename)
        m = type('', (), {})
        m.uset = set(u.username for u in users).difference(uset)
        m.gset = set(g.name for g in groups).difference(gset)
        uset.update(m.uset)
        gset.update(m.gset)
        members[perm] = m

      # If dataset creator is not a superuser and is valid with valid perm, add this
      # user into the list of users for the perm.
      # (superusers have all perms so no point in adding those)
      cperm = ds.dataset_create_user_perm
      clist = User.objects.filter(username=ds.create_user)
      if len(clist) == 1 and not clist[0].is_superuser and cperm in ptypes:
        members[cperm].uset.update([clist[0].username])

      # Everyone perm -- Add 'everyone' group to appropriate permission level
      eperm = ds.dataset_everyone_perm
      if eperm in ptypes and settings.EVERYONE_GROUP:
        members[eperm].gset.update([settings.EVERYONE_GROUP])
        members[eperm].uset = set([])
      else:
        eperm = None

      # Now we've built the set of users at each permission level for this ds
      for perm in ptypes:
        m = members[perm]
        mergeUsers(m)

        # Skip this permission if no one was assigned (typically 'readonly')
        if len(m.uset) or len(m.gset):
          found = False
          for mdef in roles[perm]:
            if m.uset == mdef.uset and m.gset == mdef.gset and mdef.dcid == dc.id:
              found = True
              mdef.dsids.append(ds.id)
              break

            # Old systems put the dataset creator at 'setupperm' but we don't want one
            # role per creator, so merge all creators into a single role
            elif merge and perm == 'setupperm' and mdef.dcid == dc.id:
              mdef.uset.update(m.uset)
              mdef.gset.update(m.gset)
              mdef.dsids.append(ds.id)
              mergeUsers(mdef)
              found = True
              break

          if not found:
            mdef = type('', (), {})
            mdef.uset = m.uset
            mdef.gset = m.gset
            mdef.dcid = dc.id
            mdef.dsids = [ds.id]
            roles[perm].append(mdef)

        # As we process the permissions in decreasing order, once we have
        # a permission with 'everyone', we can stop processing the more
        # restrictive ones.
        if perm == eperm:
          break


  # Now create the roles per data connection
  # Roles are based on the default roles above
  # One role for all dataset creators
  # As many as needed for readwrite and readonly
  for dc in dcl:
    for pi, perm in enumerate(ptypes):

      # Template for this ptype
      roleTemplate = defroles[pi + 1]

      # Grab all the definitions for this dc of this ptype
      for i, mdef in enumerate([mdef for mdef in roles[perm] if mdef.dcid == dc.id]):
        rdef = copy.deepcopy(roleTemplate)

        # Name and members
        rdef['name'] = '%s on connection %s' % (rdef['name'], dc.dataconnection_name)
        if i > 0:
          rdef['name'] += ' set %s' % i
        rdef['users'] = list(mdef.uset)
        rdef['groups'] = list(mdef.gset)

        # If this role has a dataconn type privilege, set the dc id
        if rdef['privs'][1]['ptype'] == 'dataconn':
          rdef['privs'][1]['dclist'] = [str(dc.id)]

        # Now dataset privilege entries either 'All Datasets' or one per ds
        # Note--setupperm uses 'All datasets' so no update from template needed
        dsindex = len(rdef['privs']) - 1
        rdef['privs'][dsindex]['dcid'] = str(dc.id)
        if not merge or perm != 'setupperm':
          allds = [ds.id for ds in dsl if ds.dataconnection_id == mdef.dcid]
          if len(mdef.dsids) != len(allds):
            dstempl = rdef['privs'][dsindex]
            dsprivs = []
            for dsid in mdef.dsids:
              dspriv = copy.deepcopy(dstempl)
              dspriv['dslist'] = [str(dsid)]
              dsprivs.append(dspriv)
            rdef['privs'] = rdef['privs'][0:dsindex] + dsprivs
        if dryrun:
          printRole(rdef, dcByID, dsByID)
        else:
          createRoles([rdef])

def main():
  """
  Can be invoked through the Django management utility:
  python -m manage util migrateperms
  """

  import optparse
  parser = optparse.OptionParser(usage="")

  parser.add_option('--defroles', dest='defroles', action='store_true',
    help='Setup default roles')

  parser.add_option('--reset', dest='reset', action='store_true',
    help='Delete all existing roles before migrate and/or default creation')

  parser.add_option('--dryrun', dest='dryrun', action='store_true', default=DRYRUN,
    help='No database update--just print roles that would be created')

  parser.add_option('--merge', dest='merge', action='store_true',
    help='Merge dataset admins into single role')

  parser.add_option('--migrate', dest='migrate', action='store_true',
    help='Run migration algorithm')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  if len(sys.argv) == 1:
    sys.argv.append("--help")

  options, args = parser.parse_args()

  if len(args) > 0:
    fatal('Extra options: %s' % str(args))

  if not options.dryrun and options.reset:
    Role.objects.all().delete()

  if options.defroles:
    defaultRoles(options.dryrun)

  if options.migrate:
    migratePerms(options.dryrun, options.merge)
