# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

# Models for datasets

from django.db import models
from django.conf import settings

from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from django.core.cache import cache

from django.contrib.auth.models import User, Group

from rbac.permcache import RBAC, groups_for_user
from rbac.permdefs import PERMDEFS, PERMLISTS, PERMTYPES
from functools import cmp_to_key

import json
import time
import logging

from arcwebbase.validator import validateNoHtml, validateStringNoHtml

LOGGER = logging.getLogger("arcweb")

# Validates a privilege entry
# Also updates the permissions to include those forced by higher permissions
def validate_priv(priv):
  ptype = priv.get('ptype')
  if ptype not in PERMTYPES:
    raise Exception("Invalid privilege type: %s" % ptype)
  perms = priv.get('perms', [])
  forced = set()
  for perm in perms:
    if perm not in PERMLISTS[ptype]:
      raise Exception("Invalid permission for type %s: %s" % (ptype, perm))
    forced.update(PERMLISTS.get(perm, []))
  for perm in forced:
    if perm not in perms:
      perms.append(perm)

# Build the index tuple for a privilege dict or op dict
def privkey(p):
  ptype = p.get('ptype')
  key = [ptype]
  if ptype == 'dataconn':
    key.extend(sorted(set(p.get('dclist'))))
  elif ptype == 'dataset':
    key.append(p.get('dcid'))
    key.extend(sorted(set(p.get('dslist'))))
  elif ptype == 'role':
    key.extend(sorted(set(p.get('rolelist'))))
  return tuple(key)

# Turn the key back into a privilege
def keypriv(key):
  ptype = key[0]
  priv = {'ptype':ptype}
  if ptype == 'dataconn':
    priv['dclist'] = key[1:]
  elif ptype == 'dataset':
    priv['dcid'] = key[1]
    priv['dslist'] = key[2:]
  elif ptype == 'role':
    priv['rolelist'] = key[1:]
  return priv

# Validate and condense a list of privileges:
#   combine permissions for identical entries, drop entries with empty permission list
def condense_privs(privlist):

  def keycmp(ka, kb):
    ta = ka[0]
    tb = kb[0]
    if ta != tb:
      return (PERMTYPES.index(ta) > PERMTYPES.index(tb)) - \
             (PERMTYPES.index(ta) < PERMTYPES.index(tb))
    return (len(ka) > len(kb)) - (len(ka) < len(kb))

  # Combine identical entries, ensuring we always have a 'system' entry
  phash = {}
  for p in privlist + [{'ptype':'system', 'perms':[]}]:
    validate_priv(p)
    key = privkey(p)
    if phash.get(key) is None:
      phash[key] = set()
    phash[key].update(p.get('perms'))

  # Turn hash back into privilege list
  condlist = []
  for key in sorted(list(phash.keys()), key=cmp_to_key(keycmp)):
    perms = phash[key]
    p = keypriv(key)
    if len(perms) or p['ptype'] == 'system':
      p['perms'] = list(perms)
      condlist.append(p)
  return condlist

# TODO (DJC): On dataset/dataconnection delete, we don't automatically remove
# the references from the items names inside of the role privilege structure.
# Either need to add that or add a background cleaner to walk the roles and
# remove non-existent datasets/dataconnections named in the privilege elements.
class Role(models.Model):
  """Role model"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    if self.privileges:
      self.privileges = json.dumps(condense_privs(json.loads(self.privileges)))
    self.reset()
    super(Role, self).save(*args, **kwargs)
    RBAC.set_changed()

  # role name
  name = models.CharField(max_length=1000, null=True, blank=True,
      validators=[validateStringNoHtml])

  # role description
  description = models.CharField(max_length=1000, null=True, blank=True,
      validators=[validateStringNoHtml])

  # json formatted list of privileges
  privileges = models.TextField(max_length=32*1024, null=True, blank=True,
      validators=[validateStringNoHtml])

  # json formatted lists of users/groups assigned to this role
  users = models.TextField(max_length=8*1024, null=True, blank=True,
      validators=[validateStringNoHtml])
  groups = models.TextField(max_length=8*1024, null=True, blank=True,
      validators=[validateStringNoHtml])

  # Groups from sentry if this role matches a sentry role
  sentry_groups = models.TextField(max_length=8*1024, null=True, blank=True,
      validators=[validateStringNoHtml])

  # Flag to indicate user must be member of all groups, not just any of the listed groups
  require_all_groups = models.BooleanField(default=False)

  # Permission type for objects of this class
  @staticmethod
  def perm_type():
    return 'role'

  # Check if user has given permission on connection
  def checkPerm(self, user, perm):
    RBAC.checkPerm(user, perm, self)

  # Get all permissions a user has on this connection
  def getPerms(self, user):
    return RBAC.get_user_obj_perms(user, self)

  def userSet(self):
    if not hasattr(self, '_userset'):
      self._userset = set(json.loads(self.users or '[]'))
    return self._userset

  def groupSet(self):
    if not hasattr(self, '_groupset'):
      self._groupset = set(json.loads(self.groups or '[]'))
    return self._groupset

  def sentryGroupSet(self):
    if not hasattr(self, '_sentry_groupset'):
      self._sentry_groupset = set(json.loads(self.sentry_groups or '[]'))
    return self._sentry_groupset

  def privList(self):
    if not hasattr(self, '_privlist'):
      self._privlist = json.loads(self.privileges or '[]')
    return self._privlist

  # Reset parsed version of json objects
  def reset(self):
    for attr in ['_userset', '_groupset', '_sentry_groupset', '_privlist']:
      if hasattr(self, attr):
        delattr(self, attr)

  # Add a privilege
  def add_priv(self, ptype, perms, **kwargs):
    data = {'ptype':ptype, 'perms':perms}
    data.update(kwargs)
    if data.get('dcid'):
      data['dcid'] = str(data['dcid'])
    for key in ['dclist', 'dslist', 'rolelist']:
      if data.get(key):
        data[key] = [str(x) for x in data[key]]
    self.privileges = json.dumps(self.privList() + [data])
    self.save()

  def add_users(self, names):
    to_add = set(names)
    current = self.userSet()
    if not to_add.issubset(current):
      self.users = json.dumps(list(current.union(to_add)))
      self.save()

  def add_groups(self, names):
    to_add = set(names)
    current = self.groupSet()
    if not to_add.issubset(current):
      self.groups = json.dumps(list(current.union(to_add)))
      self.save()

  def delete_users(self, names):
    to_del = set(names)
    current = self.userSet()
    if to_del.intersection(current):
      self.users = json.dumps(list(current.difference(to_del)))
      self.save()

  def delete_groups(self, names):
    to_del = set(names)
    current = self.groupSet()
    if to_del.intersection(current):
      self.groups = json.dumps(list(current.difference(to_del)))
      self.save()

  def migrate_settings_styles(self, forward):
    # Here we use the fact the condense_privs always forces
    # a system privilege as the first entry
    #
    # Forward  -> expand sys_styles to [sys_styles + sys_settings]
    # Backward -> drop the new sys_settings permission
    #
    privlist = condense_privs(self.privList())
    perms = privlist[0].get('perms')
    updated = False
    if forward:
      if 'sys_styles' in perms:
        perms.append('sys_settings')
        updated = True
    else:
      if 'sys_settings' in perms:
        perms.remove('sys_settings')
        updated = True
    if updated:
      self.privileges = json.dumps(privlist)
      self.save()

  # Make sure the dataset and dataconnection id's named in the
  # privileges are valid and fix up the role if not
  def fixup(self, dsids, dcids, roleids):
    newprivs = []
    changed = False
    for priv in self.privList():
      newprivs.append(priv)
      ptype = priv.get('ptype')

      # For role, process the connection id list
      if ptype == 'role':
        rolelist = [int(id) for id in priv.get('rolelist')]
        rolefilt = [str(id) for id in rolelist if id == -1 or id in roleids]
        if len(rolefilt) == 0:
          changed = True
          del newprivs[-1]
        elif len(rolelist) != len(rolefilt):
          changed = True
          priv['rolelist'] = rolefilt

      # For data connect, process the connection id list
      elif ptype == 'dataconn':
        dclist = [int(id) for id in priv.get('dclist')]
        dcfilt = [str(id) for id in dclist if id == -1 or id in dcids]
        if len(dcfilt) == 0:
          changed = True
          del newprivs[-1]
        elif len(dclist) != len(dcfilt):
          changed = True
          priv['dclist'] = dcfilt

      # For dataset, process the dataset list but also look
      # at the connection id
      elif ptype == 'dataset':
        dcid = int(priv.get('dcid'))
        dslist = [int(id) for id in priv.get('dslist')]

        # If data connection id no longer valid, clear out the ds id list
        if dcid != -1 and dcid not in dcids:
          dsfilt = []
        else:
          dsfilt = [str(id) for id in dslist if id == -1 or dsids.get(id) == dcid]

        if len(dsfilt) == 0:
          changed = True
          del newprivs[-1]
        elif len(dslist) != len(dsfilt):
          changed = True
          priv['dslist'] = dsfilt

        # Create new privilege rows for any dataset that has changed data connection
        for id in [id for id in dslist if id != -1 and str(id) not in dsfilt]:
          ds_dcid = dsids.get(id)
          if ds_dcid in dcids and ds_dcid != dcid:
            changed = True
            newprivs.append({'ptype':ptype,
                             'dcid':str(ds_dcid),
                             'dslist':[str(id)],
                             'perms':priv['perms']})

    if changed:
      self.privileges = json.dumps(newprivs)
      self.save()
    return changed

  def user_has_role(self, user):
    """ See if a user has this role """
    if user.username in self.userSet():
      return True

    # Require all groups means user must be member of all groups named in the role
    gset = self.groupSet()
    if gset and self.require_all_groups:
      uset = groups_for_user(user)
      if settings.EVERYONE_GROUP:
        uset = uset.union([settings.EVERYONE_GROUP])
      return gset.intersection(uset) == gset

    gset.update(self.sentryGroupSet())
    if settings.EVERYONE_GROUP in gset:
      return True
    if gset.intersection(groups_for_user(user)):
      return True

    # Do case insensitive match on sentry group names if requested
    if getattr(settings, 'SENTRY', {}).get('GROUP_NAME_IGNORE_CASE', False):
      glower = set(g.lower() for g in self.sentryGroupSet())
      ulower = set(g.lower() for g in groups_for_user(user))
      if glower.intersection(ulower):
        return True

    return False

  def perms_for_obj(self, obj=None):
    """ Return set of permissions that this role defines for the given object """

    # Get a function to see if this privilege applies to this object
    def priv_is_for_object(priv, obj):
      ptype = priv.get('ptype')
      if ptype != RBAC.obj_perm_type(obj):
        return False
      if ptype == 'system':
        return True
      if ptype == 'dataconn':
        dclist = priv.get('dclist', [])
        return (str(obj.id) in dclist or '-1' in dclist)
      if ptype == 'dataset':
        dslist = priv.get('dslist', [])
        dcid = priv.get('dcid')
        return (dcid == '-1' or dcid == str(obj.dataconnection_id)) and \
               (str(obj.version_group_id) in dslist or '-1' in dslist)
      if ptype == 'role':
        rolelist = priv.get('rolelist', [])
        return (str(obj.id) in rolelist)
      return False

    perms = set()
    for priv in self.privList():
      if priv_is_for_object(priv, obj):
        perms.update(priv.get('perms', []))

    return perms


# Reset permission cache on any role update
@receiver(post_save, sender=Role)
def role_post_save(sender, instance, **kwargs):
  if isinstance(instance, Role):
    RBAC.reset()

@receiver(post_delete, sender=Role)
def role_post_delete(sender, instance, **kwargs):
  if isinstance(instance, Role):
    RBAC.reset()

# SAML User
class SAMLUser(User):
  """
  SAMLUser is a proxy for a user, which an added method to set group membership
  """
  class Meta:
    proxy = True
    db_table = 'auth_user'

  def add_groups(self, groups):
    LOGGER.debug("adding saml groups")
    if getattr(settings, 'SAML_CREATE_UNKNOWN_GROUP', False):
      for name in groups:
        Group.objects.get_or_create(name=name)
    self.groups.clear()
    self.groups.add(*list(Group.objects.filter(name__in=groups)))
    pass
