# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

# Permission cache and support for RBAC

import time
import logging
import threading

from .permdefs import PermissionException, PERMLISTS
from datasets.models import DataSet, DataConnection
from dataconn.dataconn import supports_dataextract_target

from django.conf import settings
from django.core.cache import cache
from django.db import close_old_connections
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from arcwebbase.etag_caching import etag_check_user_groups

from vizlib.misc import ChangeChecker

LOGGER = logging.getLogger('arcweb')

# Routine to get the set of groups for a user
def groups_for_user(user):
  if hasattr(user, '_arcadia_groupset'):
    return user._arcadia_groupset

  if hasattr(user, '_arcadia_groupnames'):
    if user._arcadia_groupnames:
      groups = set(user._arcadia_groupnames.split(','))
    else:
      groups = set()
  else:
    groups = set(user.groups.values_list('name', flat=True))

  if hasattr(user, 'ldap_user') or hasattr(user, 'is_unix_user'):
    key = "rbac.remote_groups_for_user.%s" % user.id
    remote = cache.get(key)
    if remote is None:
      remote = set()
      if hasattr(user, 'ldap_user'):
        remote.update(user.ldap_user.group_names)
      # Add in the UNIX groups if present
      if hasattr(user, 'unix_user') and hasattr(user.unix_user, 'group_names'):
        remote.update(user.unix_user.group_names)
      cache.set(key, remote, getattr(settings, 'REMOTE_GROUP_CACHE_TIMEOUT', 0))
      LOGGER.debug('External groups for %s : %s' % (user.username, remote))

    groups.update(remote)

  # Here when the input user object is not obtained from the arcviz request
  # structure but rather from a User.objects.get() type of call, the objects
  # corresponding to users requiring LDAP or PAM authentication do not have
  # the group information filled in.

  # We detect this type of user object because it does not have a usable
  # password but also does not have the above 'ldap_user' or 'unix_user'
  # attributes.

  # Local user objects would have a usable password
  elif hasattr(user, 'has_usable_password') and not user.has_usable_password():
    key = "rbac.ldap_groups_for_user.%s" % user.id
    more_groups = cache.get(key)
    if more_groups is None:
      more_groups = set([])
      try:
        if ('django_auth_ldap.backend.LDAPBackend' in settings.AUTHENTICATION_BACKENDS or \
            'arcwebbase.backends.LdapBackend' in settings.AUTHENTICATION_BACKENDS or \
            'arcweb.arcwebbase.basebackends.VizBaseLDAPBackend' in settings.AUTHENTICATION_BACKENDS) and \
          not getattr(settings, 'LDAP_PREVENT_EXTENDED_ROLE_LOOKUP', False):
          from django_auth_ldap.backend import LDAPBackend
          lu = LDAPBackend().populate_user(user.username)
          more_groups.update(lu.ldap_user.group_names)

        if 'arcwebbase.backends.PamBackend' in settings.AUTHENTICATION_BACKENDS and \
          not getattr(settings, 'PAM_PREVENT_EXTENDED_ROLE_LOOKUP', False):
          from arcwebbase.backends import get_unix_groups
          more_groups.update(get_unix_groups(user.username))

        cache.set(key, more_groups,
                  getattr(settings, 'AUTH_LDAP_GROUP_CACHE_TIMEOUT', 0))

        LOGGER.debug('External group lookup for %s : %s' % (user.username, more_groups))

      except:
        LOGGER.exception('getting groups for user "%s" failed' %
                         user.username)

    groups.update(more_groups)

  user._arcadia_groupset = groups
  etag_check_user_groups(user, groups)
  return user._arcadia_groupset

# TODO (DJC): Note that caching in memory like this has an issue for multi-processing
# as another process could save/update a role, call the reset but this would not be
# noticed by this process.

# Class to provide top level role caching and permission checking
class _PermCache():
  def __init__(self):
    self.event = threading.Event()
    self.polltime = 0
    self.pollth = None
    self.pollcnt = 0
    self.last_recheck = 0
    self.last_update = 0
    self.change_checker = ChangeChecker('rbac_update_time',
                                        getattr(settings, 'RBAC_WORKER_CHECK_SECONDS', 5))
    self.roles = None

    self.reset()

  def set_changed(self):
    self.change_checker.set_changed()

  def pollstart(self, polltime=0):
    if polltime:
      self.polltime = polltime
    if self.polltime and not self.pollth:
      self.pollth = threading.Thread(target=self.poller, name="Role sync")
      self.pollth.daemon = True
      self.pollth.start()

  def pollstop(self):
    if self.pollth:
      self.polltime = 0
      self.pollkick()
      self.pollth.join()
      self.pollth = None

  def pollwait(self, limit=0):
    start = time.time()
    now = start
    count = self.pollcnt
    while (self.pollcnt == count) and ((now - start) < limit):
      time.sleep(.1)
      now = time.time()
    return (self.pollcnt != count)

  def reset(self, poll=True):
    self.roles = None;
    self.userHash = {}
    self.permHash = {}
    if poll:
      self.event.set()

  # A routine to ensure roles are kept in sync when dataset or dataconnections
  # are removed.
  def poller(self):
    while self.polltime:
      try:
        self.event.wait(self.polltime)
        self.event.clear()
        changed = False
        if settings.IS_MASTER:
          close_old_connections()
          roles = self.load_roles()
          rlids = set(r.id for r in roles)
          dsids = dict(DataSet.objects.values_list('id', 'dataconnection_id'))
          dcids = set(DataConnection.objects.values_list('id', flat=True))
          for r in roles:
            if r.fixup(dsids, dcids, rlids):
              changed = True
              LOGGER.info("Role %s synced to db" % r.id)
      except:
        LOGGER.exception("Error in role poller")
      self.pollcnt += 1

      # Reload the permissions cache each poll cycle
      if not changed:
        self.reset(poll=False)

  # Signal the poller to run
  def pollkick(self):
    self.pollstart()
    self.event.set()

  # Provide access to role objects but avoid circular import
  def load_roles(self):
    self.pollstart()
    if self.roles is None or self.change_checker.is_changed():
      from rbac.models import Role
      self.roles = Role.objects.all()
    return self.roles

  def obj_perm_type(self, obj):
    return 'system' if obj is None else obj.perm_type()

  def all_perms_for_obj(self, obj):
    return set(PERMLISTS.get(self.obj_perm_type(obj), []))

  def get_user_obj_perms(self, user, obj=None):
    ptype = self.obj_perm_type(obj)
    remove_dc_upload = ptype == 'dataconn' and not supports_dataextract_target(obj.dataconnection_type)

    if user.is_superuser:
      perms = self.all_perms_for_obj(obj)
      if remove_dc_upload:
        perms.discard('dc_upload')
      return perms

    # Reset cache if roles have changed
    if self.change_checker.is_changed():
      self.reset(poll=False)
      self.load_roles()

    # Delete any cached entries for this user if groups have changed
    groups = groups_for_user(user)
    if self.userHash.get(user.id) != groups:
      self.userHash[user.id] = groups
      self.permHash[user.id] = {}

    hash = self.permHash[user.id]
    index = (ptype, 0 if obj is None else obj.id)
    perms = hash.get(index)
    if perms is None:
      perms = set()
      if ptype == 'workspace':
        perms.update(obj.perms_for_user(user, groups))
      else:
        for r in self.load_roles():
          if r.user_has_role(user):
            perms.update(r.perms_for_obj(obj))

      # Force users to have full access to the datasets that they create
      # But only when they still have 'dc_explore' on the data connection (ARC-5897)
      allperms = self.all_perms_for_obj(obj)
      if ptype == 'dataset' and obj.create_user == user.username:
        if not allperms.issubset(perms):
          if 'dc_explore' in obj.dataconnection.getPerms(user):
            perms.update(allperms)

      # Similarly for data connections, force creators to have full access
      # But only when they still have system level permission to edit connections
      elif ptype == 'dataconn' and obj.create_user == user.username:
        if not allperms.issubset(perms):
          if 'sys_editconn' in self.get_user_obj_perms(user):
            perms.update(allperms)

      if remove_dc_upload:
        perms.discard('dc_upload')
      hash[index] = perms

    if remove_dc_upload:
      perms.discard('dc_upload')
    return perms

  def checkPerm(self, user, perm, obj=None):
    self.pollstart()
    if perm not in self.get_user_obj_perms(user, obj):
      raise PermissionException(user, perm, obj)

  def update_group_roles(self, roles, group):
    for r in self.load_roles():
      if r.id in roles:
        r.add_groups([group.name])
      else:
        r.delete_groups([group.name])

  def update_user_roles(self, roles, user):
    for r in self.load_roles():
      if r.id in roles:
        r.add_users([user.username])
      else:
        r.delete_users([user.username])

  def roles_for_user(self, user):
    return [r for r in self.load_roles() if r.user_has_role(user)]

# Global permission manager
RBAC = _PermCache()

# On dataset/connection delete or dataset update, kick the poller to fix up the roles
@receiver(post_delete, sender=DataSet)
def dataset_post_delete(sender, instance, **kwargs):
  if isinstance(instance, DataSet):
    RBAC.pollkick()

@receiver(post_save, sender=DataSet)
def dataset_post_save(sender, instance, **kwargs):
  if isinstance(instance, DataSet):
    RBAC.pollkick()

@receiver(post_delete, sender=DataConnection)
def dataconnection_post_delete(sender, instance, **kwargs):
  if isinstance(instance, DataConnection):
    RBAC.pollkick()
