# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import django.db.models
import django.test

from datasets.models import DataConnection, DataSet
from rbac import migrateperms
from rbac.models import Role
from unittest import mock


class TestMigrateperms(django.test.TestCase):

  def setUp(self):
    self.dc1 = DataConnection.objects.create(
      dataconnection_name='dc1', dataconnection_type='arcengine')
    ds1 = DataSet.objects.create(
      dataset_name='ds1', dataconnection=self.dc1, dataset_type='singletable', dataset_detail='[]')
    self.ds1 = ds1
    # simulate a dataset without a dataconnection:
    ds1.dataconnection = None
    ds1.save()
    ds1.refresh_from_db()

  def test_role_creation(self):
    # db migration creates the default roles
    self.assertTrue(Role.objects.count() > 0, 'default roles exist')

    # delete roles
    Role.objects.all().delete()
    self.assertTrue(Role.objects.count() == 0, 'no roles exist')

    migrateperms.migratePerms()
    self.assertTrue(Role.objects.count() > 0, 'default roles created')

  def test_dataconnection_assigning(self):

    self.assertEquals(DataConnection.objects.count(), 1)
    self.assertEquals(DataSet.objects.count(), 1)

    self.assertIsNone(self.ds1.dataconnection)

    migrateperms.migratePerms()

    # no new dataconnection is created
    self.assertEquals(DataConnection.objects.count(), 1)
    self.ds1.refresh_from_db()
    # since dc1 is an arcengine connection, it is picked as the default
    self.assertEqual(self.ds1.dataconnection.dataconnection_name, 'dc1')

  def test_new_connection_creation(self):
    self.dc1.delete()
    self.assertEquals(DataConnection.objects.count(), 0)

    migrateperms.migratePerms()

    # a new data connection was created:
    self.assertEquals(DataConnection.objects.count(), 1)
    self.ds1.refresh_from_db()
    self.assertEqual(self.ds1.dataconnection.dataconnection_name, 'default')

  def test_management_command(self):
    from django.core.management import load_command_class

    command = load_command_class('apps', 'util')

    with mock.patch('apps.management.commands.util.exit') as exit_mock:
      with mock.patch('apps.management.commands.util.print'):
        with mock.patch.object(migrateperms, 'migratePerms') as migratePerms_mock:

          command.run_from_argv(['django-admin', 'util', 'migrateperms', '--migrate', '--dryrun', '--merge'])

          # expected arguments
          dryrun = True
          merge = True

          migratePerms_mock.assert_called_once_with(dryrun, merge)

          # exit was called with 0 (=no error)
          self.assertEqual(exit_mock.call_args_list[0].args,(0,))


