# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Roles url support for rbac

Note - this module knows too much about the internals of the role structure
       and one could consider udating the model to provide better abstraction
       but the overall role system is small and contained in a few modules
       so fine for now.
"""

from django.http import HttpResponse, JsonResponse

from rbac.models import Role, condense_privs
from rbac.permcache import RBAC

from django.contrib.auth.decorators import login_required
from django.conf import settings
from datasets.models import DataSet

from vizlib.querylog import logAction, ComponentType, ActionType
from django.core.paginator import Paginator
from django.db.models import Q

import json
import logging

LOGGER = logging.getLogger("arcweb")

def _auditRole(role, action, request):
  logAction(action=action, component_type=ComponentType.ROLE, 
            component_id=role.id, request=request)

@login_required
def roles_dspublish(request):

  if request.method != 'POST':
    return HttpResponse("Invalid request", status=500)

  rdict = {}
  dsdict = {}
  user = request.user
  data = json.loads(request.POST.get("data", "[]"))

  # Process one ds pub entry.  Note that the input here is generated by the UI
  # and is expected to be correctly formatted so we just through an exception
  # on any error.
  # This updates the role but does not save it so that any error encountered
  # results in no changes.
  def onepub(d):
    rid = d['id']
    dsid = int(d['dsid'])
    perms = d['perms']

    # Get the role
    role = rdict.get(rid)
    if role is None:
      role = Role.objects.get(id=rid)
      rdict[rid] = role

    # Get the dataset end ensure we have manage permission on it
    ds = dsdict.get(dsid)
    if ds is None:
      ds = DataSet.objects.get(id=dsid)
      dsdict[dsid] = ds
      ds.checkPerm(user, 'ds_manage')

    # Find or create the privilege for this dataset
    found = False
    privlist = role.privList()
    for priv in role.privList():
      if priv.get('ptype') == 'dataset' and priv.get('dslist') == [str(dsid)]:
        found = True
        break
    if not found:
      priv = {'ptype':'dataset',
              'dcid':str(ds.dataconnection_id),
              'dslist':[str(dsid)],
              'perms':[],
             }
      privlist.append(priv)

    # If user does not have system level role edit permissions, then
    # For any changed permissions, ensure user has corresponding grant permission
    if 'sys_editperm' not in RBAC.get_user_obj_perms(user):
      for perm in set(priv.get('perms')).symmetric_difference(set(perms)):
        role.checkPerm(user, 'rg' + perm)

    # Now we can store the new permissions for this privilage entry
    priv['perms'] = perms

  # Process the list of updates
  for u in data:
    try:
      onepub(u)
    except Exception as e:
      LOGGER.exception('Error on dataset publish')
      return HttpResponse('Error on dataset publish: %s' % str(e), status=500)

  # Now save the updated privileges for each role
  for role in list(rdict.values()):
    role.privileges = json.dumps(role.privList())
    _auditRole(role, ActionType.UPDATE, request=request)
    role.save()

  return HttpResponse('Success')


@login_required
def roles(request, role_id=None):
  """
  REST endpoint for role access
  """
  return roles_internal(request, role_id)

def roles_internal(request, role_id=None, filter=None):
  """
  Generic interface to front end for one role:
    {
      name: <particular role name>
      desc: <role description>
      id: <id for this role>
      privs: [<list of privileges for this role>]
      users : [<list of user names assigned to this role>]
      groups : [<list of group names assigned to this role>]
    }

  The DB structure for roles matches the above interface object with the lists
  stored in json format.

  Each individual privilege is itself a structure:
    {
      ptype: One of 'system', 'dataconn', 'dataset'
      perms: [<list of permissions for this privilege>]

      When type is 'dataconn':
      dclist: [<list of data connection IDs>]

      When type is 'dataset'
      dcid: <data connection ID>
      dslist: [<list of dataset IDs>
    }

  Note that we use data connection and dataset IDs rather than names so the no
  action is required on name change.  TODO: should probably remove the IDs when
  dataconn/dataset is deleted.

  We'll use the special -1 ID to indicate "ALL" objects.

  Similarly, we need a special 'Everyone' group to indicate all users.  However,
  in LDAP mode, we have the option of specifying a group to indicate which users
  are allowed to log in which is essentially the 'Everyone' group.

  We use user/group names as these might come from an external source like LDAP
  and so may not be present in the Django model.

  The set of privilege types and the permissions within each type are specified
  at the top of dataset/models.py, but here also are thoughts on possible
  permissions (needs more thought):

  System level permissions:
    Admin:
      Edit users/groups
      Edit permissions
    Data connection creator
      Create/Edit data connections

  Data connection permissions:
    Admin:
      Edit permissions
      Edit connection (this is same as data connection creator)
    Dataset Creator
      Upload data
      Browse connection / create dataset
    Dataset viewer (Do we need this one?)
      See all datasets on a connection

  Dataset permissions:
      Add joins (this is basically browsing a connection)
      Accelerate (is this a data connection admin?)
      Materialize (flatten joins, cube building, etc.)
    Admin
      Edit permissions
      Filter assocations
    Report creator
      Create reports
      Edit attributes
      Create segments / events
    Report viewer
      View reports

  """

  def validate_role(id):
    if not id:
      raise Exception("No role id specified")
    roles = Role.objects.filter(id=int(id))
    if len(roles) == 0:
      raise Exception("Role with id '%s' not found" % id)
    return roles[0]

  if request.method == 'GET':
    is_paginated = False
    if role_id:
      roles = [validate_role(role_id)]
    else:
      page = request.GET.get('page')
      page_size = request.GET.get('page_size')
      sort = request.GET.get('sort', 'name')
      order = request.GET.get('order', 'asc')
      search = request.GET.get('search')

      if order == 'desc':
        sort = f'-{sort}'

      roles = Role.objects.all().order_by(sort)

      if search:
        roles = roles.filter(
          Q(name__icontains=search) |
          Q(description__icontains=search)
        )

      is_paginated = page is not None and page_size is not None
      if is_paginated:
        page = int(page) if page.isdigit() else None
        page_size = int(page_size) if page_size.isdigit() else None
        paginator = Paginator(roles, per_page=page_size)
        roles = paginator.get_page(page).object_list

      if filter:
        roles = roles.filter(**filter)
    result = []
    for r in roles:
      rlres = {}
      rlres['id'] = r.id
      rlres['name'] = r.name
      rlres['desc'] = r.description or ''
      rlres['privs'] = json.loads(r.privileges or '[]')
      rlres['users'] = json.loads(r.users or '[]')
      rlres['groups'] = json.loads(r.groups or '[]')
      rlres['sentry_groups'] = json.loads(r.sentry_groups or '[]')
      rlres['perms'] = list(RBAC.get_user_obj_perms(request.user, r))
      rlres['require_all_groups'] = r.require_all_groups
      result.append(rlres)

    if is_paginated:
      return JsonResponse({
      "roles": list(result),
      "total": paginator.count if is_paginated else len(roles)
    }) 
    else:
      return HttpResponse(json.dumps(result))
    

    

  RBAC.checkPerm(request.user, 'sys_editperm')
  # below this point we either are role manager or superuser

  if request.method == 'PUT':
    return HttpResponse("Invalid request", status=500)

  if request.method == 'POST':
    response = {}

    roles = json.loads(request.POST.get("data"))

    for recieved_role in roles:
      role_id = recieved_role.get('id')
      name = recieved_role.get('name')
      description = recieved_role.get('desc')
      users = json.dumps(recieved_role.get('users', []))
      groups = json.dumps(recieved_role.get('groups', []))
      require_all_groups = recieved_role.get('require_all_groups', False)

      try:
        has_privileges = len(recieved_role.get('privs')) > 0
        condensed_privileges = condense_privs(recieved_role.get('privs'))
        privileges = json.dumps(condensed_privileges)
      except Exception as e:
        LOGGER.error("Error updating role '%s':" % str(e))
        return HttpResponse("Error updating role", status=500)

      action_type = ActionType.UPDATE if role_id else ActionType.CREATE

      role = Role() if action_type == ActionType.CREATE else Role.objects.get(id=role_id)

      if request.user.is_superuser and not name:
        return HttpResponse("No role name specified", status=500)

      is_name_already_in_use = False
      existing_roles = Role.objects.filter(name=name)
      if action_type == ActionType.CREATE:
        if(len(existing_roles) > 0):
          is_name_already_in_use = True
      else:
        if len(existing_roles) > 1 or (len(existing_roles) == 1 and existing_roles[0].id != role_id):
          is_name_already_in_use = True

      if is_name_already_in_use:
          LOGGER.error("Role name '%s' already in use" % name)
          return HttpResponse("Role name already in use", status=500)

      if action_type == ActionType.CREATE:
        role.sentry_groups = json.dumps([])

      role.name = name
      role.description = description
      role.privileges = privileges
      role.users = users
      role.groups = groups
      role.require_all_groups = require_all_groups

      role.save()

      _auditRole(role, action_type, request=request)

      recieved_role['id'] = role.id

    response['status'] = "success"
    response['data'] = roles

    return HttpResponse(json.dumps(response), content_type="application/json")

  if request.method == 'DELETE':
    if role_id:
      role = validate_role(role_id)
      _auditRole(role, ActionType.DELETE, request=request)
      role.delete()
      return HttpResponse('{"status": "success"}', content_type="application/json")
    else:
      return HttpResponse("No role ID specified", status=500)

  return HttpResponse('Invalid request method', status=500)
