# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from migration.migrate import export_objects, VisualSync, import_objects, DashboardSync
import logging, re, json, random
from .models import ReportTemplate, ReportImage, Report
from datasets.models import DataSet
from dataconn.dataconn import get_dataconn
from vizlib.dataset import ModelDataset
from workspace.models import get_default_workspace
import copy

LOGGER = logging.getLogger("arcweb")

def _publish_template(param_context, report_id, user):
  template = ReportTemplate.objects.get(report=Report.objects.get(id=report_id))
  template_data = template.data
  for k, v in param_context['params'].items():
    if v == '_arc_record_count_':
      # if the picked column is Record Count then reset the expression to be sum(1)
      template_data = re.sub('[^"]*_arc_agg_[^"]*',
                             r"sum(1) as 'Record Count'\\", template_data)
    else:
      template_data = template_data.replace("<<%s>>" % k, v)
  template_data = json.loads(template_data)
  for key in list(template_data.keys()):
    for obj in template_data[key]:
      obj['fields']['uuid'] = "_template_%s" % random.randrange(10000,99999)
  if len(template_data['datasets']) != 1:
    raise Exception("cannot load template ")
  # don't need to create dataset, use the dataset with dataset_id
  # here we are assuming there is only one dataset. However, note that one
  # can save a dashboard with multiple datasets as template. Maybe, later
  # we should support loading template with multiple datasets
  if param_context['dssw']:
    dataset_id_map = param_context['dssw']
  else:
    # no dataset switching, new dashboard will use the same dataset as the template
    dataset_id_map = {template_data['datasets'][0]['pk']:template_data['datasets'][0]['pk']}
  del template_data['datasets']
  del template_data['segments']

  dataset = DataSet.objects.get(id=list(dataset_id_map.values())[0])
  report = import_objects(template_data,
                          dataset.dataconnection_id,
                          user,
                          False,
                          True,
                          ids_map={'datasets':dataset_id_map},
                          workspace_id=get_default_workspace(user).id)
  return report

def create_template(template_data, user, deep_clone=False):
  # change UUID of the data to make sure such UUID does not exist on the system
  # thus new artifact gets created instead of updating a new one.
  for key in list(template_data.keys()):
    for obj in template_data[key]:
      obj['fields']['uuid'] = "_template_%s" % random.randrange(10000,99999)

  # Here we are assuming there is only one dataset. However, note that one
  # can save a dashboard with multiple datasets as template. Maybe, later
  # we should support loading template with multiple datasets as well
  if len(template_data['datasets']) > 1:
    raise Exception('Dashboards using multiple datasets are not supported')


  # Creating dashboard with parameterized data using import function. Modifying
  # template_data to avoid recreating/creating dataset. Also only create visuals
  # and dashboards if deep_clone is set to False
  dataset_id = template_data['datasets'][0]['pk']
  ids_map = {'datasets':{dataset_id:dataset_id}}
  del template_data['datasets']
  del template_data['segments']
  del template_data['events']
  if not deep_clone:
    # we are only creating visuals and dashboards and reuse other dependencies
    # like customcss. This scenario happens when we are creating a new template
    # on the same system. But, if we are calling this function to initialize
    # templates shipped by the product we need to create all dependencies
    ids_map['customcss'] = {}
    ids_map['dateranges'] = {}
    for c in template_data['customcss']:
      ids_map['customcss'][c['pk']] = c['pk']
    for c in template_data['dateranges']:
      ids_map['dateranges'][c['pk']] = c['pk']
    del template_data['customcss']
    del template_data['dateranges']
  report = import_objects(template_data,
                          DataSet.objects.get(id=dataset_id).dataconnection_id,
                          user,
                          False,
                          True,
                          ids_map=ids_map,
                          report_mode='template',
                          workspace_id=1)
  return report

def parameterize_and_save(name, desc, report_id, obj_name, dsreq, user):
  dsreqs = json.loads(dsreq)
  ids = [report_id]
  export_dict = export_objects(obj_name, ids, user)
  parameterize_columns_dict, export_dict = parameterize(export_dict, dsreqs)
  num_dims = 0
  num_aggs = 0
  for dataset_id, value in parameterize_columns_dict.items():
    num_dims += len(value['dimensions'])
    num_aggs += len(value['aggregates'])
  try:
    img = ReportImage.objects.get(report_id=report_id)
    img = img.img
  except ReportImage.DoesNotExist:
    img = None

  # create a new parameterized dashboard, get its record in report table
  # (by searching its id), and then create entry in reporttempate model
  import_report = create_template(copy.deepcopy(export_dict), user)
  report = Report.objects.get(id=import_report[DashboardSync.name][0]['id'])
  r = ReportTemplate(report=report,
                     name=name,
                     description=desc,
                     data=json.dumps(export_dict),
                     params=json.dumps(
                            {"columns": parameterize_columns_dict,
                             "no_dims":num_dims,
                             "no_aggs":num_aggs,
                             "report_dataset_id":report.dataset_id}),
                     img=img)
  r.save()

def _extract_dim_msr(dsreq_list):
  # return the dimension and aggregate objects used in dashboard

  def get_col_objects(col_names):
    col_objects = {}
    for dataset_id in col_names:
      dataset = DataSet.objects.get(id=dataset_id)
      model_dataset = ModelDataset(
        dataset.dataset_info,
        get_dataconn(connection_id=dataset.dataconnection.id).properties())
      col_objects[dataset_id] = [model_dataset.getColumn(n)
                          for n in col_names[dataset_id] if n != 'Record Count']
      if 'Record Count' in col_names[dataset_id]:
        record_count = model_dataset.tables[0].columns[0]
        record_count.attr['alias'] = 'Record Count'
        record_count.attr['type'] = 'BIGINT'
        record_count.attr['name'] = '_record_count'
        col_objects[dataset_id].append(record_count)

    return col_objects

  # dictionaries for holding datasets' column names
  # note that here we are supporting having multiple datasets in a dashboard (for
  # future work. Later on we raise exception if there is more than one dataset.
  dim_names= {}
  agg_names = {}
  filter_names = {}

  for dsreq in dsreq_list:
    if type(dsreq) != dict:
      dsreq = json.loads(dsreq)
    dataset_id = dsreq['dataset_id']
    if dataset_id not in dim_names:
      dim_names[dataset_id] = []
      agg_names[dataset_id] = []
      filter_names[dataset_id] = []
    for d in dsreq['dimensions']:
      if d['type'] == 'SIMPLE':
        column_names = _get_colname(expr=d['expr'])
        dim_names[dataset_id] = list(set(dim_names[dataset_id] + column_names))
      elif d['type'] == 'SEGMENT':
        raise Exception("having segment is not supported")
    for a in dsreq['aggregates']:
      column_names = _get_colname(expr=a['expr'])
      agg_names[dataset_id] = list(set(agg_names[dataset_id] + column_names))
    for f in dsreq['filters'] + dsreq['having']:
      column_names = _get_colname(expr=f)
      filter_names[dataset_id] = list(set(filter_names[dataset_id] + column_names))

  for dataset_id in dim_names:
    # aggregates are the column names that are only used in visual aggregates
    # thus taking out the ones that are already used in dimensions
    agg_names[dataset_id] = list(set(agg_names[dataset_id]) - set(dim_names[dataset_id]))

    filter_names[dataset_id] = list(set(filter_names[dataset_id]) - set(dim_names[dataset_id]))
    filter_names[dataset_id] = list(set(filter_names[dataset_id]) - set(agg_names[dataset_id]))

  return {"dimensions":get_col_objects(dim_names),
          "aggregates":get_col_objects(agg_names),
          "filters":get_col_objects(filter_names)}

def _get_colname(pill=None, expr=None):
  if expr is None and pill is not None:
    if 'expression_for_trigger' in pill and pill['expression_for_trigger'] != "(1 = 1)":
      expr = pill['expression_for_trigger']
    elif 'custom_expr' in pill:
      expr = pill['custom_expr']
    elif 'dataset_colname' in pill:
      expr = '[%s]' % pill["dataset_colname"]

  if expr is not None and expr.startswith('sum(1)'):
    return ['Record Count']
  elif expr is not None:
    return list(set(re.findall(r'\[([^(>\])]*)\]', expr)))
  elif pill is not None and 'filter_column' in pill:
    # this will happen when a column is dragged to filter shelf with no value
    return [pill['filter_column']]
  else:
    raise Exception("cannot extract column name from % s" % pill)

def _update_pill(pill, name, replace, shelf, reset_filter=False):
  if shelf == 'filters_shelf' and 'expression_for_trigger' in pill and pill['expression_for_trigger'] == "(1 = 1)":
    pill['dataset_colname'] = "<<%s>>" % replace
    pill['filter_column'] = "<<%s>>" % replace
    if 'custom_expr' in pill:
      del pill['custom_expr']
  elif shelf == 'filters_shelf' and reset_filter:
    pill['expression_for_trigger'] = "(1 = 1)"
    pill['dataset_colname'] = "<<%s>>" % replace
    if 'custom_expr' in pill:
      del pill['custom_expr']
  elif shelf == 'filters_shelf':
    pill['filter_column'] = "<<%s>>" % replace
    if 'custom_expr' in pill:
      del pill['custom_expr']
  else:
    if 'expression_for_trigger' in pill:
      expr = pill['expression_for_trigger']
    elif 'custom_expr' in pill:
      expr = pill['custom_expr']
    elif 'dataset_colname' in pill:
      expr = '[%s]' % pill['dataset_colname']
    else:
      raise Exception("pill does not have one of the required key")

    if re.search(r'\[(%s)\]' % name, expr):
      pill['expression_for_trigger'] = re.sub(r'%s' % name,
                                              '<<%s>>' % replace,
                                              expr)
    else:
      pill['expression_for_trigger'] = '[<<%s>>]' % replace

    pill['custom_expr'] = pill['expression_for_trigger']

def parameterize(report, dsreqs):
  """
  replace the column names and titles used in visuals with parameteres.
  The flow is:
    1. get the dictionary of all columns used in visuals which is done by calling
    _extract_dim_msr function
    2. loop through each visual
      2.1 if the dateset visual is visited for the first time, then create a
          name mapping between column names and its parameter
      2.2 go through each shelf of visual and each pill, then update the column name
      with a parameter using column name mapping
  :param report:
  :param dsreqs:
  :return:
  """

  def get_columns_attr(cols, type, start=0):
    params = []
    if type == 'dimensions':
      alias = 'dim'
    elif type == 'aggregates':
      alias = 'agg'
    elif type == 'filters':
      alias = 'filter'
    else:
      alias = None
    for i in range(start, start+ len(cols)):
      params.append(dict([('name', '_arc_%s_%s_%s' % (alias, dataset_id, i)),
                          ('orig_name', cols[i].attr['alias']),
                          ('type', cols[i].attr['type']),
                          ('visual_ids', []),
                          ('visual_types', []),
                          ('shelves', [])]))
      column_name_map_index_hash[cols[i].attr['alias']] = dict([('index', i),
                                                                ('type', type)])
    return params

  def get_col_map(col_name):
    # returns the parameter name to replace colname variable in shelf
    col = None
    if col_name in column_name_map_index_hash:
      col = column_name_map[dataset_id][column_name_map_index_hash[col_name]['type']] \
        [column_name_map_index_hash[col_name]['index']]
      col['visual_ids'].append(viz_id)
      if viz_type not in col['visual_types']:
        col['visual_types'].append(viz_type)
      if shelf not in col['shelves']:
        col['shelves'].append(shelf)
    return col

  def get_name_map(col_name):
    # returns the parameter name to replace colname variable in shelf
    col = get_col_map(col_name)
    if col is not None:
      return col['name']
    else:
      return None

  column_name_map = {}
  # hashing name of column to its index in column_name_map variable to avoid
  # nested loops, which gets populated during call to get_columns_attr
  column_name_map_index_hash = {}

  # get the list of all dimension and measures objects used in all viz
  dim_measures = _extract_dim_msr(dsreqs)

  for viz in report[VisualSync.name]:
    dataset_id = viz['fields']['dataset']
    viz_id  = viz['pk']
    viz_type = viz['fields']['report_type']
    if dataset_id not in column_name_map:
      # if this dataset is visited for the first time, create name mapping for it
      # and add it to the column_name_map
      dims = dim_measures['dimensions'][dataset_id]
      aggs = dim_measures['aggregates'][dataset_id]
      filters = dim_measures['filters'][dataset_id]

      column_name_map[dataset_id] = {
        'dimensions': get_columns_attr(dims, 'dimensions'),
        'aggregates': get_columns_attr(aggs, 'aggregates'),
        'filters': get_columns_attr(filters, 'filters'),
        'geo': [],
        'date': []
      }

    report_data = json.loads(viz['fields']['report_data'])
    shelves = [s['name'] for s in report_data['report_data']['core']['shelves']]
    for shelf in shelves:
      for pill in report_data['report_data'][shelf]:
        try:
          colnames = _get_colname(pill=pill)
        except:
          raise Exception('cannot get column name for pill %s in visual ID %s' %
                          (pill, viz_id))
        for col_name in colnames:
          if col_name is None:
            continue
          elif get_name_map(col_name) is None:
            # the col_name does not exist in the column_name_map variable, which
            # happens if the col_name does not exist in the dsreq. I am not sure
            # why this wil happen, but to handle this situation for the filter shelf
            # we are going to add this column to the column_name_map. Note that the
            # call to get_columns_attr will populate the index in
            # column_name_map_index_hash.
            if shelf == 'filters_shelf':
              fake_dsreq = {'dataset_id':dataset_id,
                            'dimensions':[],
                            'aggregates':[],
                            'filters':['[%s]' % col_name],
                            'having':[]}
              col = _extract_dim_msr([fake_dsreq])
              column_name_map[dataset_id]['filters'] += \
                  get_columns_attr(col['filters'][dataset_id],
                                   'filters',
                                   start=len(column_name_map[dataset_id]['filters']))
              _update_pill(pill, col_name, get_name_map(col_name), shelf)
            else:
              raise Exception("'%s' in visual with %s ID not found, column_name_map: %s" %
                                (col_name, viz_id, column_name_map))
          else:
            # get the parameter for colname variable
            _update_pill(pill, col_name, get_name_map(col_name), shelf)

            # adding some heuristic
            if shelf == 'geo_shelf' and \
               get_name_map(col_name) not in column_name_map[dataset_id]['geo']:
              column_name_map[dataset_id]['geo'].append(get_name_map(col_name))

            if viz_type == 'calendar-heatmap' and \
               get_col_map(col_name)['type'] == 'TIMESTAMP' and \
               get_name_map(col_name) not in column_name_map[dataset_id]['date']:
              column_name_map[dataset_id]['date'].append(get_name_map(col_name))

    # update title
    for p in re.findall(r'<<(.*):?>>', report_data['report_data']['report_title']):
      # take out default value if there is any
      col_name = p.split(':')[0]
      report_data['report_data']['report_title'] = \
                            re.sub(r'<<%s>>' % p,
                                   r'<<%s>>' % get_name_map(col_name),
                                   report_data['report_data']['report_title'])

    if viz_type == 'picklist':
      col_name = report_data['report_data']['user_settings']['picklist_base_field']
      name_map = get_name_map(col_name)
      report_data['report_data']['user_settings']['picklist_base_field'] = '<<%s>>' % name_map
      report_data['report_data']['user_settings']['picklist_output_parameter'] = '<<%s>>' % name_map
      report_data['report_data']['user_settings']['picklist_title'] = '<<%s>>' % name_map

    viz['fields']['report_data'] = json.dumps(report_data)
  return column_name_map, report

