# -*- coding: utf-8 -*-
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from __future__ import unicode_literals

from django.db import migrations
import json, logging, os, tempfile, copy

# For dashboards pre-4.1.5 (before sheets), introduce dashboard_sheets
# For dashboards pre-Banff (pre ARC-8988), populate global_control_widgets (fix app scope filters)
def populate_global_control_widgets(apps, schema_editor):
  Report = apps.get_model('reports', 'report')
  all_dashboards = Report.objects.filter(report_type='dashboard')
  all_filter_data = dict(Report.objects.filter(report_type__in=['picklist', 'daterange']).values_list('id', 'report_data'))
  for db in all_dashboards:
    try:
      report = json.loads(db.report_data)
      report = _populateGlobalControlWidgets(report, all_filter_data)
      db.report_data = json.dumps(report)
      db.save()
    except Exception as e:
      msg = 'ERROR: Cannot update dashboard report ID: %s' % db.id
      print(msg)
      print(str(e))
      logging.exception(msg)

def _populateGlobalControlWidgets(report, all_filter_data):
  # migrating pre-Banff dashboards to post-Banff dashboards
  dashboard_widgets = []
  control_widgets = []
  app_scope_filters = report['global_control_widgets']
  for sheet in report['dashboard_sheets']:
    sheet_filters = []
    for control_widget in sheet['control_widgets']:
      if is_filter_app_scope(control_widget, all_filter_data):
        app_scope_filters.append(control_widget)
      else:
        sheet_filters.append(control_widget)
    sheet['control_widgets'] = sheet_filters
    dashboard_widgets += sheet['visual_widgets']
    control_widgets += sheet_filters
  report['global_control_widgets'] = list(set(app_scope_filters))
  # Keep dashboard_widgets and control_widgets updated (TODO: Delete them in ARC-9378)
  report['dashboard_widgets'] = dedup_dashboard_widgets(dashboard_widgets)
  report['control_widgets'] = list(set(control_widgets + app_scope_filters))
  return report

def dedup_dashboard_widgets(dashboard_widgets):
  dedup_dashboard_widgets = []
  dedup_dashboard_widget_ids = {}
  for dw in dashboard_widgets:
    if dw['id'] not in dedup_dashboard_widget_ids:
      dedup_dashboard_widget_ids[dw['id']] = True
      dedup_dashboard_widgets.append(copy.deepcopy(dw))
  return dedup_dashboard_widgets

def is_filter_app_scope(report_id, all_filter_data):
  try:
    report_data = all_filter_data.get(int(report_id))
    if not report_data:
      return False
    report_data = json.loads(report_data)
    report_data = report_data['report_data']
    return report_data['user_settings'].get('filter_appgroup_scope', 'false') == 'true'
  except Exception as e:
    return False

class Migration(migrations.Migration):
  dependencies = [
    ('reports', '0006_report_has_css_styles'),
  ]

  operations = [
      migrations.RunPython(populate_global_control_widgets)
  ]
