# -*- coding: utf-8 -*-
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from __future__ import unicode_literals

from django.db import migrations
import json, os, tempfile, math, copy

# No more area range setting, replaced with radius range
def update_map_area_setting(apps, schema_editor):
  Report = apps.get_model('reports', 'report')
  map_interactive_reports = Report.objects.filter(report_type='leaflet')

  for report in map_interactive_reports:
    report_data = json.loads(report.report_data)
    report_data_internal = report_data.get('report_data', {})
    user_settings = report_data_internal.get('user_settings', {})
    circle_area_values = user_settings.get('circle_area_range')
    circle_radius_values = user_settings.get('circle_radius_range')
    tile_style = user_settings.get('tile-style')
    if tile_style == 'mapbox':
      tile_style = 'mapboxgl'
      user_settings['tile-style'] = tile_style

    if circle_radius_values:
      continue

    if not circle_area_values:
      circle_area_values = '100-10000'

    area_range = circle_area_values.split('-')
    low_area = int(area_range[0])
    high_area = int(area_range[1])
    low_radius = math.sqrt(low_area / math.pi)
    high_radius = math.sqrt(high_area / math.pi)
    radius_range_value = '{}-{}'.format(int(low_radius), int(high_radius))
    user_settings['circle_radius_range'] = radius_range_value
    report_data_internal['user_settings'] = user_settings
    report_data['report_data'] = report_data_internal
    report_data = json.dumps(report_data)
    report.report_data = report_data
    report.save(update_fields=['report_data'])

def update_map_shelf(apps, schema_editor):
  Report = apps.get_model('reports', 'report')
  map_reports = Report.objects.filter(report_type='map')

  for report in map_reports:
    report_data = json.loads(report.report_data)
    report_data_internal = report_data.get('report_data', {})
    user_settings = report_data_internal.get('user_settings', {})
    geo_shelf_data = report_data_internal.get('geo_shelf')
    is_bubble = user_settings.get('map_bubbles') == 'true'
    if geo_shelf_data:
      for shelf_element in geo_shelf_data:
        if shelf_element.get('geo_cast') == 'zipcode':
          is_bubble = True

    if is_bubble and not report_data_internal.get('size_shelf'):
      old_color_shelf = report_data_internal.get('aggregates_shelf')
      report_data_internal['aggregates_shelf'] = []
      report_data_internal['size_shelf'] = old_color_shelf

    report_data['report_data'] = report_data_internal
    report_data = json.dumps(report_data)
    report.report_data = report_data
    report.save(update_fields=['report_data'])

# from forever ago, but figured it was worth getting rid of any remaining single value reports
def update_single_value_to_kpi(apps, schema_editor):
  Report = apps.get_model('reports', 'report')
  single_value_reports = Report.objects.filter(report_type='single_value')
  for report in single_value_reports:
    report['report_type'] = 'kpi'
    report.save()

# kpi label font label id being changed to avoid conflicts
def update_kpi_label_setting(apps, schema_editor):
  Report = apps.get_model('reports', 'report')
  kpi_reports = Report.objects.filter(report_type='kpi')

  for report in kpi_reports:
    report_data = json.loads(report.report_data)
    report_data_internal = report_data.get('report_data', {})
    user_settings = report_data_internal.get('user_settings', {})
    label_font_value = user_settings.get('label_font_size')
    kpi_label_font_size = user_settings.get('kpi_label_font_size')
    if kpi_label_font_size:
      continue

    if label_font_value == '':
      label_font_value = '14'

    user_settings['kpi_label_font_size'] = label_font_value
    report_data_internal['user_settings'] = user_settings
    report_data['report_data'] = report_data_internal
    report_data = json.dumps(report_data)
    report.report_data = report_data
    report.save(update_fields=['report_data'])

def update_word_cloud_tooltips(apps, schema_editor):
  Report = apps.get_model('reports', 'report')
  wordcloud_reports = Report.objects.filter(report_type='wordcloud')

  for report in wordcloud_reports:
    report_data = json.loads(report.report_data)
    report_data_internal = report_data.get('report_data', {})
    user_settings = report_data_internal.get('user_settings', {})
    show_tooltips = user_settings.get('show_tooltips')
    display_tooltips = user_settings.get('show_tooltip')
    if show_tooltips == 'false' or display_tooltips == 'false':
      user_settings['show_tooltip'] = 'false'
    report_data_internal['user_settings'] = user_settings
    report_data['report_data'] = report_data_internal
    report_data = json.dumps(report_data)
    report.report_data = report_data
    report.save(update_fields=['report_data'])

class Migration(migrations.Migration):
  dependencies = [
    ('reports', '0010_css_add_user_info_timestamps'),
  ]

  operations = [
    migrations.RunPython(update_map_area_setting),
    migrations.RunPython(update_map_shelf),
    migrations.RunPython(update_single_value_to_kpi),
    migrations.RunPython(update_kpi_label_setting),
    migrations.RunPython(update_word_cloud_tooltips),
  ]
  
