# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from tests.migrationtests_base import MigrationsTestBase
from unittest.mock import patch, Mock

import copy
import json

AI_EMBEDDINGS_SETTING_VALUE = {
    'defaultProf': {
        'value': 'caii_embed_value',
        'label': 'caii_embed'
    },
    'profileSettings': {
        'caii_embed_value': {
            'label': 'caii_embed',
            'values': {
                'embeddings_engine': 'caii',
                'embeddings_api_key': 'some-key',
                'embeddings_model': 'some-model',
                'embeddings_service_url': 'some-url'
            }
        },
        'azure_embed_value': {
            'label': 'azure_embed',
            'values': {
                'embeddings_engine': 'azure_ai',
                'embeddings_api_key': 'some-key',
                'embeddings_service_url': 'some-url'
            }
        }
    }
}

AI_COMPLETION_SETTING_VALUE = {
    'defaultProf': {
        'value': 'azure_comp_value',
        'label': 'azure_comp'
    },
    'profileSettings': {
        'caii_comp_value': {
            'label': 'caii_comp',
            'values': {
                'completion_engine': 'caii',
                'completion_api_key': 'some-key',
                'completion_model': 'some-model',
                'completion_service_url': 'some-url',
                'completion_temperature': '1'
            }
        },
        'azure_comp_value': {
            'label': 'azure_comp',
            'values': {
                'completion_engine': 'azure_ai',
                'completion_api_key': 'some-key',
                'completion_service_url': 'some-url'
            }
        }
    }
}

def test_migration_is_applied(test_case_instance, is_cml_env):
    test_case_instance.run_migration('reports', '0043')

    AppSetting = test_case_instance.project_state.apps.get_model('reports', 'appsetting')

    ai_embed_setting = AppSetting.objects.create(setting_key='ai_embeddings_settings',
                                                    setting_value=json.dumps(AI_EMBEDDINGS_SETTING_VALUE),
                                                    setting_read_admin_only=False)
    ai_comp_setting = AppSetting.objects.create(setting_key='ai_completion_settings',
                                                    setting_value=json.dumps(AI_COMPLETION_SETTING_VALUE),
                                                    setting_read_admin_only=False)

    test_case_instance.assertIsNotNone(ai_embed_setting.pk)
    test_case_instance.assertIsNotNone(ai_comp_setting.pk)

    test_case_instance.run_migration('reports', '0044')
    AppSetting = test_case_instance.project_state.apps.get_model('reports', 'appsetting')

    ai_embed_setting = AppSetting.objects.filter(setting_key=f'ai_embeddings_settings').first()
    ai_comp_setting = AppSetting.objects.filter(setting_key=f'ai_completion_settings').first()

    if is_cml_env:
        # check caii profiles have new cml_auth_mode property with default value
        # api_key because we are on CML environment
        embed_setting_new_value = copy.deepcopy(AI_EMBEDDINGS_SETTING_VALUE)
        embed_setting_new_value['profileSettings']['caii_embed_value']['values']['embeddings_cml_auth_mode'] = 'api_key'
        comp_setting_new_value = copy.deepcopy(AI_COMPLETION_SETTING_VALUE)
        comp_setting_new_value['profileSettings']['caii_comp_value']['values']['completion_cml_auth_mode'] = 'api_key'

        test_case_instance.assertEqual(json.loads(ai_embed_setting.setting_value), embed_setting_new_value)
        test_case_instance.assertEqual(json.loads(ai_comp_setting.setting_value), comp_setting_new_value)
    else:
        # check caii profiles don't have new cml_auth_mode property because we are on
        # non CML environment
        test_case_instance.assertEqual(json.loads(ai_embed_setting.setting_value), AI_EMBEDDINGS_SETTING_VALUE)
        test_case_instance.assertEqual(json.loads(ai_comp_setting.setting_value), AI_COMPLETION_SETTING_VALUE)
class TestMigrations(MigrationsTestBase):
    @patch('reports.migrations.0044_set_caii_profile_cml_auth_mode_to_api_key.get_cmldata')
    def test_0044_non_cml_env(self, get_cmldata):
        get_cmldata.return_value = None # Not CML env
        test_migration_is_applied(self, False)

    @patch('reports.migrations.0044_set_caii_profile_cml_auth_mode_to_api_key.get_cmldata')
    def test_0044_cml_env_and_no_profiles(self, get_cmldata):
        get_cmldata.return_value = Mock() # CML env

        self.run_migration('reports', '0043')

        AppSetting = self.project_state.apps.get_model('reports', 'appsetting')

        ai_embed_settingExists = AppSetting.objects.filter(setting_key='ai_embeddings_settings').exists()
        ai_comp_settingExists = AppSetting.objects.filter(setting_key='ai_completion_settings').exists()

        # check there are no profiles in the db
        self.assertFalse(ai_embed_settingExists)
        self.assertFalse(ai_comp_settingExists)

        # check there is no error message, there was an error because of a missing ai_setting_obj check
        self.run_migration('reports', '0044')

    @patch('reports.migrations.0044_set_caii_profile_cml_auth_mode_to_api_key.get_cmldata')
    def test_0044_cml_env_and_there_are_profiles(self, get_cmldata):
        get_cmldata.return_value = Mock() # CML env
        test_migration_is_applied(self, True)
