# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Models for reports
"""

from django.db import models, transaction
from django.contrib.auth.models import User
from django.db.models import F
from django.utils import timezone
from datasets.models import DataSet
from workspace.models import Workspace
from rbac.models import RBAC

import datetime
from zoneinfo import ZoneInfo 

import json
import uuid

from arcwebbase.validator import validateNoHtml, validateStringNoHtml, validateStringNoEndingTagHtml
from django.core.cache import cache
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from django.conf import settings
from threading import Lock
import logging

ENABLE_URL_NAMES = getattr(settings, 'ENABLE_URL_NAMES', False)
LOGGER = logging.getLogger("arcweb")

class StaticAsset(models.Model):
  """User-uploaded static assets"""
  def save(self, *args, **kwargs):
    super(StaticAsset, self).save(*args, **kwargs)

  # mime type
  staticasset_mimetype = models.CharField(max_length=100, blank=True, null=True,
      validators=[validateStringNoHtml])

  # Name
  staticasset_name = models.CharField(max_length=1000, blank=True, null=True,
      validators=[validateStringNoHtml])

  # Description
  staticasset_description = models.TextField(max_length=16000, default='', blank=True, null=True,
      validators=[validateStringNoHtml])

  # Size
  staticasset_size = models.IntegerField(blank=True, null=True)

  # json object representing apps using this asset
  staticasset_apps = models.TextField(max_length=16000, blank=True, null=True,
      validators=[validateStringNoHtml])

  # Contents
  staticasset_contents = models.TextField(max_length=100*1024*1024, blank=True, null=True)

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

class CustomCSS(models.Model):
  """Saved color palettes"""

  def save(self, *args, **kwargs):
    #no validateNoHTML is added here because customJS can have HTML strings inside of it.
    super(CustomCSS, self).save(*args, **kwargs)

  customcss_name = models.CharField(max_length=1000, blank=True, null=True,
      validators=[validateStringNoHtml])

  customcss_description = models.TextField(max_length=16000, default='', blank=True, null=True,
      validators=[validateStringNoHtml])

  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True, db_index=True,
                                 validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True, db_index=True,
                                 validators=[validateStringNoHtml])
  # CSS stylesheet
  customcss_value = models.TextField(max_length=16000, blank=True, null=True,
      validators=[validateStringNoHtml])

  # JS code
  customcss_js = models.TextField(max_length=16000, blank=True, null=True)

  # If we modify the format in which we store 'value' in the future, this field
  # will help us interpret 'colorpalette_value'
  customcss_type = models.IntegerField(blank=True, null=True)

  # Should this stylesheet apply by default to all apps?
  customcss_default = models.BooleanField(default=False, blank=True)

  # uuid
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

class ColorPalette(models.Model):
  """Saved color palettes"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(ColorPalette, self).save(*args, **kwargs)

  colorpalette_name = models.CharField(max_length=1000, blank=True, null=True,
      validators=[validateStringNoHtml])
  colorpalette_value = models.TextField(max_length=16000, blank=True, null=True,
      validators=[validateStringNoHtml])

  # If we modify the format in which we store 'value' in the future, this field
  # will help us interpret 'colorpalette_value'
  colorpalette_type = models.IntegerField(blank=True, null=True)

  colorpalette_description = models.TextField(max_length=16000, default='', blank=True, null=True,
      validators=[validateStringNoHtml])

  # uuid
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)
  #uuid = models.CharField(max_length=36, null=True, blank=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

class ParameterSet(models.Model):
  """
  Saved parameter sets.

  Users can store away lists of parameter key/value names stored as a url string
  (example: param.sysid=123&param.component_type=volume) and then access this set
  of parameters by using a arcparam.paramset.<setname>=<anyvalue> URL query string.
  """

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(ParameterSet, self).save(*args, **kwargs)

  parameterset_name = models.CharField(max_length=1000, blank=True, null=True,
      validators=[validateStringNoHtml])

  # parameterset_value may contain < characters, so we can't restrict possible values
  parameterset_value = models.TextField(max_length=16000, blank=True, null=True)

  # If we modify the format in which we store 'value' in the future, this field
  # will help us interpret 'parameterset_type'
  parameterset_type = models.IntegerField(blank=True, null=True)

class DateRanges(models.Model):
  """Saved date ranges"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(DateRanges, self).save(*args, **kwargs)

  dateranges_name = models.CharField(max_length=1000, blank=True, null=True,
      validators=[validateStringNoHtml])
  dateranges_value = models.TextField(max_length=16000, blank=True, null=True,
      validators=[validateStringNoHtml])

  # If we modify the format in which we store 'value' in the future, this field
  # will help us interpret 'colorpalette_value'
  dateranges_type = models.IntegerField(blank=True, null=True)

  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

class AppSetting(models.Model):
  """App Settings"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    cache.set('site-setting.%s' % self.setting_key, self.setting_value)

    super(AppSetting, self).save(*args, **kwargs)

  # Key-value pairs to track settings
  setting_key = models.CharField(max_length=1000, blank=True, null=True,
      validators=[validateStringNoHtml])

  # Note that html values are permitted
  setting_value = models.TextField(max_length=16000, blank=True, null=True)

  # Are only admins permitted to read this value ?
  setting_read_admin_only = models.BooleanField(default=True, blank=True)

class AppGroup(models.Model):
  """Groups of apps"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)

    if ENABLE_URL_NAMES:
      # If group_url_name is set, make sure it is unique
      group_options = {}
      try:
        group_options = json.loads(self.group_options)
      except Exception as e:
        pass

      self.group_url_name = group_options.get('user_settings', {}).get('app_url_name')
      if self.group_url_name:
        idlist = list(AppGroup.objects.filter(group_url_name=self.group_url_name)
                                    .exclude(id=self.id)
                                    .values_list('id', flat=True))
        if len(idlist):
          raise(Exception('App URL name "%s" already used by IDs %s' % (self.group_url_name, idlist)))

    super(AppGroup, self).save(*args, **kwargs)

  def permitted_update(self, user):
    # Always allow admin users to update app groups
    if user.is_superuser or 'sys_misc' in RBAC.get_user_obj_perms(user):
      return True

    # Users are always allowed to update their favorites (starred)
    if self.group_type == 'starred' and self.group_username == user.username:
      return True

    # Check appgroup workspace permisisons
    return 'ws_appedit' in self.workspace.getPerms(user)

  # name of group
  group_name = models.CharField(max_length=1000, blank=True,
      validators=[validateStringNoHtml],
      null=True)

  # Group type. Set to 'starred' for per-user apps. In the future,
  # we can use this to track 'auto' groups like the most popular apps,
  # all apps in a datasetwith etc.
  group_type = models.CharField(max_length=100, blank=True,
      null=True, validators=[validateStringNoHtml])

  # User. For groups of group_type 'starred', this is the username for whom
  # this is the starred group. Note that this is the only time 'group_username'
  # is set. For all other cases, it is set to null.
  group_username = models.CharField(max_length=1000, blank=True, null=True,
      validators=[validateStringNoHtml])

  # parent group key. For root groups, this is null
  group_parent = models.ForeignKey("AppGroup", on_delete=models.CASCADE, blank=True, null=True, related_name='subgroups')

  # ordering of subgroups within groups
  group_subgroup_order = models.FloatField(blank=True, null=True)

  # json data with other details
  group_options = models.TextField(max_length=4*1024, blank=True,
      validators=[validateStringNoHtml])

  # Allow apps to be loaded by name as well as ID
  group_url_name = models.CharField(max_length=100,
      validators=[validateStringNoHtml], blank=True, null=True, default='')

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

  # workspace for this app group
  workspace = models.ForeignKey(Workspace,
                                on_delete=models.SET_DEFAULT, default=1)

class UserAppGroupMembership(models.Model):
  """Mapping of users to visible groups"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(UserAppGroupMembership, self).save(*args, **kwargs)

  # group key
  usergroup_group = models.ForeignKey(AppGroup, on_delete=models.CASCADE, blank=True, null=True)

  # user
  usergroup_username = models.CharField(max_length=100, blank=True, null=True,
      validators=[validateStringNoHtml])

  # Order of app group within user's groups
  usergroup_order = models.FloatField(blank=True, null=True)

def _validateReportData(s):
  if s is None:
    return

  try:
    j = json.loads(s)
  except Exception as e:
    return

  # Note that we permit <<varname:default>> strings in report_title &
  # report_subtitle, which makes this call NoEndingTagHtml instead of
  # the simpler NoHtml
  report_title = j.get('report_title')
  validateStringNoEndingTagHtml(report_title)
  report_subtitle = j.get('report_subtitle')
  validateStringNoEndingTagHtml(report_subtitle)

def get_visuals_picklist(report_data):
  try:
    # Try and get the list of visuals within a dashboard
    visuals = set()
    pickers = set()

    for sheet in report_data.get('dashboard_sheets', []):
      for visualWidget in sheet.get('visual_widgets', []):
        visualWidgetId = int(visualWidget.get('id').rpartition("-")[-1])
        visuals.add(visualWidgetId)

      for controlWidget in sheet.get('control_widgets', []):
        pickers.add(int(controlWidget))

    for controlWidget in report_data.get('global_control_widgets', []):
      pickers.add(int(controlWidget))

    return (list(visuals), list(pickers))
  except Exception as e:
    LOGGER.error(e)

  return ([], [])

class Report(models.Model):
  """Dashboard, Visual, or Picker"""

  # name of the report
  report_name = models.CharField(max_length=1000, blank=True)

  # description of the report
  report_description = models.TextField(max_length=2000, blank=True,
      validators=[validateStringNoHtml])

  # dataset it is based on
  dataset = models.ForeignKey(DataSet, on_delete=models.CASCADE, blank=True, null=True)

  # workspace for this report
  workspace = models.ForeignKey(Workspace, blank=True, null=True,
                                on_delete=models.SET_DEFAULT, default=1)

  # 'flow' etc.
  report_type = models.CharField(max_length=100,
      validators=[validateStringNoHtml])

  # template etc.
  report_mode = models.CharField(max_length=100,
      validators=[validateStringNoHtml], blank=True, null=True, default='')

  # Allow dashboards to be loaded by name as well as ID
  dashboard_url_name = models.CharField(max_length=100,
      validators=[validateStringNoHtml], blank=True, null=True, default='')

  # json structure describing the report
  report_data = models.TextField(max_length=32*1024, validators=[_validateReportData])

  # json structure that only exists for linked visuals or dashboards
  # in linked visuals: array of dashboard id's that use it
  # in dashboard: array of shared visuals
  shared_visual_dashboards = models.TextField(max_length=32*1024, blank=True, null=True,)

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  parent_report = models.ForeignKey('self', on_delete=models.SET_NULL, blank=True, null=True, default=None)
  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

  # flag so we can find reports with custom css quickly
  has_css_styles = models.BooleanField(default=False)

  # cached dictionary of json data contained in report_data field
  # initialized to None to avoid parsing if not needed
  # populated on first call to report.get_data
  _temp_data = None

  # Store information about search or bookmark
  report_search_text = models.TextField(null=True, blank=True, max_length=1000)

  # Get the css_styles entry for a report
  def get_css_styles(self, report_data=None):
    try:
      report_data = report_data or json.loads(self.report_data)
      if self.report_type == 'dashboard':
        css_styles = report_data['user_settings']['css_styles']
      else:
        css_styles = report_data['report_data']['user_settings']['css_styles']
    except Exception as e:
      css_styles = None
    if css_styles == '[]':
      css_styles = None
    return css_styles

  # Get the list of visuals and filters in this dashboard
  def get_visuals_in_dashboard(self, report_data=None):
    if self.report_type != "dashboard":
      return ([], [])

    if not report_data:
      try:
        report_data = json.loads(self.report_data)
      except Exception:
        return ([], [])

    return get_visuals_picklist(report_data)

  # Get the list of non unique visual's datasets within a dashboard
  def get_datasets_ids_in_dashboard(self):
    (visuals, pickers) = self.get_visuals_in_dashboard()
    reports_in_dashboard = list(Report.objects.filter(id__in=visuals).only('id', 'dataset_id'))

    return [report.dataset_id for report in reports_in_dashboard]

  def save(self, *args, **kwargs):
    validateNoHtml(self)

    # Set some fields from the big json structure into distinct columns
    report_data = {}
    try:
      report_data = json.loads(self.report_data)
      self.has_css_styles = self.get_css_styles(report_data) is not None
    except Exception as e:
      pass

    if ENABLE_URL_NAMES:
      # If dashboard_url_name is set, make sure it is unique
      self.dashboard_url_name = ''
      if self.report_type == 'dashboard':
        self.dashboard_url_name = report_data.get('user_settings', {}).get('dashboard_url_name')
        if self.dashboard_url_name:
          idlist = list(Report.objects.filter(report_type='dashboard',
                                              dashboard_url_name=self.dashboard_url_name)
                                      .exclude(id=self.id)
                                      .values_list('id', flat=True))
          if len(idlist):
            raise(Exception('Dashboard URL name "%s" already used by IDs %s' % (self.dashboard_url_name, idlist)))

    # Make sure shared_visual_dashboards is either jsonified array or None
    self.shared_visual_dashboards = self.shared_visual_dashboards or None
    super(Report, self).save(*args, **kwargs)

  # TODO: update to handle case of deleteing linked visual from one sheet of a dashboard while
  # remaining in another currently this is handled on the frontend at dashboard save, but if this
  # is added here, can simply hit the endpoint and remove the need to handle object relationships
  # on the front end.
  def delete(self, dash_id=None, audit_function=None, *args, **kwargs):
    """
    1) Choose pre delete action based on type of report
    2) Delete related objects from the database
    3) Delete report from the database
    """
    if self.category() == 'dashboard':
      self._delete_dashboard(audit_function)
    elif self.category() == 'linked':
      self._delete_linked_viz()
    else:
      self._delete_viz(dash_id)

    if audit_function:
      audit_function(self)
    super(Report, self).delete(*args, **kwargs)

  def category(self):
    if self.report_type == 'dashboard':
      return 'dashboard'
    elif not self.shared_visual_dashboards:
      return 'unlinked'
    else:
      return 'linked'

  def _permit_category(self, *categories):
    if self.category() not in categories:
      raise TypeError('Invalid Report method for {0}'.format(self.category()))

  def _delete_linked_viz(self):
    """Deleted linked visual only if not included in any dashboards"""
    self._permit_category('linked')

    msg = "Cannot delete Linked Visual {0} because it is still included in dashboards: {1}"
    dashboards = ', '.join([str(id) for id in self.shared_dashboards()])
    if not self.is_unused():
      raise Exception(msg.format(self.id, dashboards))

  def _delete_viz(self, dash_id):
    """
    Removes a visual from a dashboard prior to deletion from DB.
    1) from dashboard builder: dash_id is passed in and visual is removed
    2) as a result of dashboard deletion: no action needed

    """
    self._permit_category('unlinked')

    if dash_id:
      Report.objects.filter(id=dash_id).first().remove_visual(self).save()

  def _delete_dashboard(self, audit_function):
    """
    1) Delete all unlinked visuals
    2) Remove dashboard from shared_visual_dashboards of linked visuals
    3) Update parent_report_id of children
    """
    self._permit_category('dashboard')
    self.set_parent_id_on_children()

    for report in Report.objects.filter(id__in=self.all_report_ids()):
      # if linked viz remove from this dashboard
      if report.category() == 'linked':
        report.remove_shared(self)
        report.save()
      else:
        if audit_function:
          audit_function(report, app_id=self.id)
        report.delete()

    DataSet.remove_dashboard_from_all_datasets(self)

  def set_parent_id_on_children(self):
    children = Report.objects.filter(parent_report_id=self.id)
    for child in list(children):
      child.parent_report_id = self.parent_report_id
      child.save()

  def is_unused(self):
    """Checks if linked visual is used in any dashboards"""
    self._permit_category('linked')
    return not self.shared_dashboards()

  def remove_shared(self, *visuals):
    """Remove shared dashboard from linked viz or vice versa"""
    self._permit_category('dashboard', 'linked')
    viz_ids = [viz.id if hasattr(viz, 'id') else viz for viz in visuals]
    new_shared = set(self.shared_dashboards()) - set(viz_ids)
    self.shared_dashboards(list(new_shared))

  def shared_dashboards(self, dash_ids=None):
    """Getter/setter for JSON column shared_visual_dashboards takes and returns list<int>"""
    self._permit_category('linked', 'dashboard')

    if dash_ids is None:
      return json.loads(self.shared_visual_dashboards)
    else:
      self.shared_visual_dashboards = json.dumps(dash_ids)

  def get_data(self):
    """Return cached report data or parse and cache report_data if first call"""
    self._temp_data = self._temp_data or json.loads(self.report_data)
    return self._temp_data

  def sync_temp_data(self):
    if self._temp_data is not None:
      self.report_data = json.dumps(self._temp_data)


  def all_sheets(self):
    """Return a list<dict> of all sheets contained in a dashboard"""
    self._permit_category('dashboard')
    return self.get_data().get('dashboard_sheets', [])

  def process_ids(self, arr):
    _flatten = lambda lst: [item for sublist in lst for item in sublist]
    # dashboard widget id format for visual id = 10: url-2-widget-10
    _parse_id = lambda id: int(id) if str(id).isdigit() else int(id.get('id').split('-')[-1])
    processed_arr = _flatten(arr)
    processed_arr = [_parse_id(id) for id in processed_arr]
    processed_arr = list(set(processed_arr))
    return processed_arr

  def all_filter_ids(self):
    """Returns list<int> of ids of all control_widget objects in a dashboard"""
    self._permit_category('dashboard')
    widget_ids = self.process_ids(
      [sheet.get('control_widgets', []) for sheet in self.all_sheets()] +
      [self.get_data().get('global_control_widgets', [])]
    )
    return widget_ids

  def all_visual_ids(self):
    """Returns list<int> of ids of all visual_widget objects in a dashboard"""
    self._permit_category('dashboard')
    visual_ids = self.process_ids([sheet.get('visual_widgets', []) for sheet in self.all_sheets()])
    return visual_ids

  def all_report_ids(self):
    """
    Returns list<int> of ids of all report objects (visuals & pickers) in a dashboard
    """
    self._permit_category('dashboard')

    widget_ids = self.all_filter_ids()
    visual_ids = self.all_visual_ids()

    return visual_ids + widget_ids

  def remove_visual(self, *visuals, dataset_id=None):
    """
    Removes a visual/picker from a dashboard.  Takes any number of ids <int>/<str> or report objects
    """
    self._permit_category('dashboard')
    viz_ids = set([
      int(v['id']) if isinstance(v, dict) else int(v.id) if hasattr(v, 'id') else int(v)
      for v in visuals
    ])
    all_remaining_control_widgets = []
    all_remaining_dashboard_widgets = []

    dashboard_data = self.get_data()
    global_control_widgets = dashboard_data.get('global_control_widgets', [])
    remaining_global_control_widgets = [p for p in global_control_widgets if not int(p) in viz_ids]
    dashboard_data.update({'global_control_widgets': remaining_global_control_widgets})
    all_remaining_control_widgets += remaining_global_control_widgets

    for sheet in self.all_sheets():
      v_ids = sheet.get('visual_widgets', []) #visuals
      v_remaining = [v for v in v_ids if not int(v.get('id').split('-')[-1]) in viz_ids]
      sheet['visual_widgets'] = v_remaining
      all_remaining_dashboard_widgets += v_remaining
      p_ids = sheet.get('control_widgets', []) #pickers
      p_remaining = [p for p in p_ids if not int(p) in viz_ids]
      sheet['control_widgets'] = p_remaining
      all_remaining_control_widgets += p_remaining

    dashboard_data.update({'dashboard_widgets': all_remaining_dashboard_widgets})

    self.remove_shared(*viz_ids)

    #loop modifies report data so sync temp data and report_data
    self.sync_temp_data()

    if dataset_id is not None:
      dataset = DataSet.objects.get(id=dataset_id)
      dataset.remove_dashboard_from_dataset(self)
    else:
      visuals_datasets_ids = [visual.dataset_id for visual in visuals] # all visuals has the same dataset_id - we can simply pass it as param
      for dataset in DataSet.objects.filter(id__in=visuals_datasets_ids):
        dataset.remove_dashboard_from_dataset(self)

    return self

  def increment_report_usage(self):
    """increment tracked report usage"""
    n = datetime.datetime.now(ZoneInfo('UTC'))
    d = datetime.datetime(n.year, n.month, n.day, 0, 0, 0, 0, tzinfo=ZoneInfo('UTC'))

    rul = self.reportusage_set.filter(timewindow_start=d)
    if rul.exists():
      ru = rul[0]
      ru.count = F("count") + 1
    else:
      ru = ReportUsage(report=self, timewindow_start=d, count=1)
      ru.save()

  @staticmethod
  def get_limit(report_data, report_type=None):
    default_limit = 5000
    max_rowcount = report_data['user_settings'].get('max_rowcount', '')
    if max_rowcount != '':
      minimum_limit = min(default_limit, int(max_rowcount))
    else:
      minimum_limit = default_limit

    # Only table viz has limit shelf and it is not constrained by above setting
    if report_type == 'table':
      table_limit = report_data.get('limit', '')
      if table_limit != '':
        minimum_limit = int(table_limit)

    shelves = report_data['core']['shelves']
    for shelf in shelves:
      for el in report_data[shelf['name']]:
        el_limit = el.get('order_limit')
        if el_limit:
          minimum_limit = min(minimum_limit, int(el_limit))

    return minimum_limit

  def get_relatives_chain(self, recursion_direction, include_self=False):
    info_objects = []
    if include_self:
        info_objects.append({
          'report_id': self.pk,
          'parent_report_id': self.parent_report_id,
          'report_name': self.report_name,
          })
    if recursion_direction == 'parents':
      report_set = Report.objects.filter(pk=self.parent_report_id)
    else:
      report_set = Report.objects.filter(parent_report_id=self.pk)

    for parent in report_set:
      _info_objects = parent.get_relatives_chain(recursion_direction,include_self=True)
      if len(_info_objects) > 0:
        info_objects += _info_objects
    return info_objects

  def get_immediate_children(self):
    children_info_objects = []

    for child in Report.objects.filter(parent_report_id=self.id):
      children_info_objects.append({
        'report_id': child.pk,
        'parent_report_id': self.id,
        'report_name': child.report_name,
        })
    return children_info_objects

  def can_view_report(self, user):
    workspace_permissions = self.workspace.getPerms(user)

    can_view_report = "ws_appview" in workspace_permissions or 'sys_misc' in RBAC.get_user_obj_perms(user)

    return can_view_report

class ReportTemplate(models.Model):
  """Report"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(ReportTemplate, self).save(*args, **kwargs)

  # report id that represent the template
  report = models.ForeignKey(Report, on_delete=models.CASCADE, blank=True, null=True)

  # name of the report
  name = models.CharField(max_length=1000, blank=True)

  # description of the report
  description = models.TextField(max_length=2000, blank=True,
                                 validators=[validateStringNoHtml])

  # json structure describing the report
  data = models.TextField(max_length=32 * 1024)

  # json structure describing the report
  params = models.TextField(max_length=32 * 1024)

  # Image of template
  img = models.BinaryField(blank=True, null=True)

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
                                 validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
                                 validators=[validateStringNoHtml])

class ReportImage(models.Model):
  """ReportImage"""

  def save(self, *args, **kwargs):
    super(ReportImage, self).save(*args, **kwargs)

  # Report
  report = models.ForeignKey(Report, null=True, blank=True, on_delete=models.CASCADE)

  # Image
  img = models.BinaryField(blank=True, null=True)

  # Update time
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)

  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

class ReportUser(models.Model):
  """Add report to user's list"""
  # Report
  report = models.ForeignKey(Report, blank=True, null=True, on_delete=models.CASCADE)
  # User
  user = models.ForeignKey(User, blank=True, null=True, on_delete=models.CASCADE)
  # Order in user's list
  order = models.IntegerField(blank=True, null=True)

  @staticmethod
  def user_to_reports(username):
    """Get a user's reports"""
    # pylint: disable=E1101
    try:
      u = User.objects.get(username=username)
    except User.DoesNotExist:
      return []

    try:
      rul = ReportUser.objects.filter(user=u).order_by('order')
      return rul
    except ReportUser.DoesNotExist:
      return []

  @staticmethod
  def max_order(username):
    """Get the max 'order' value for a user"""
    # pylint: disable=E1101
    try:
      u = User.objects.get(username=username)
    except User.DoesNotExist:
      return 0

    try:
      agg = ReportUser.objects.filter(user=u).aggregate(models.Max('order'))
      return agg.get('order__max', 0)
    except ReportUser.DoesNotExist:
      return 0

  @staticmethod
  def report_to_user(reportid, username):
    """Get a user's reports"""
    # pylint: disable=E1101
    try:
      r = Report.objects.get(id=int(reportid))
    except Report.DoesNotExist:
      return []

    try:
      u = User.objects.get(username=username)
    except User.DoesNotExist:
      return []

    try:
      rul = ReportUser.objects.filter(report=r, user=u)
      return rul
    except ReportUser.DoesNotExist:
      return []

class ReportUsage(models.Model):
  """ReportUsage"""

  # report if this is tracking a dashboard, None if we are tracking an appgroup
  report = models.ForeignKey(Report, null=True, blank=True, on_delete=models.CASCADE)

  # A single ReportUsage record stores data about exactly 1 report or exactly 1 appgroup
  #   - report id if positive, appgroup id if negative
  report_or_appgroup_id = models.IntegerField(db_index=True, blank=True, null=True)

  timewindow_start = models.DateTimeField(blank=True, null=True)
  count = models.IntegerField(blank=True, null=True)

  # keep track of how many times each report has been viewed in memory. This
  # information will be written into the arcviz database periodically
  # by bulk update thread.
  reports_count = {}
  lock = Lock()

  @classmethod
  def update_reports_count(cls, report_id):
    with cls.lock:
      if report_id not in cls.reports_count:
        cls.reports_count[report_id] = 1
      else:
        cls.reports_count[report_id] += 1

  @classmethod
  def commit_changes(cls):
    if cls.reports_count == {}:
      return

    cls._commit()

  @classmethod
  @transaction.atomic
  def _commit(cls):
    with cls.lock:
      reports_count = cls.reports_count
      cls.reports_count = {}

    n = datetime.datetime.now(ZoneInfo('UTC'))
    d = datetime.datetime(n.year, n.month, n.day, 0, 0, 0, 0, tzinfo=ZoneInfo('UTC'))

    for r_id, count in reports_count.items():
      rul = ReportUsage.objects.filter(report_or_appgroup_id=r_id, timewindow_start=d)
      if rul.exists():
        ru = rul[0]
        ru.count = F("count") + count
      else:
        # report ids < 0 mean a appgroup id - save these with a null report id
        if r_id < 0:
          r = None
        else:
          try:
            r = Report.objects.get(id=r_id)
          except Report.DoesNotExist:
            continue
        ru = ReportUsage(report=r, report_or_appgroup_id=r_id, timewindow_start=d, count=count)

      ru.save()

class AppGroupAppMembership(models.Model):
  """Mapping of app groups to apps"""

  def save(self, *args, **kwargs):
    super(AppGroupAppMembership, self).save(*args, **kwargs)

  # group key
  appgroup_group = models.ForeignKey(AppGroup, on_delete=models.CASCADE, blank=True, null=True)

  # app key
  appgroup_app = models.ForeignKey(Report, on_delete=models.CASCADE, blank=True, null=True)

  # Order of app within group. Note that this is a float, and that
  # only relative orders make sense in the implementation
  appgroup_order = models.FloatField(blank=True, null=True)

  # uuid
  #uuid = models.CharField(max_length=36, null=True, blank=True)
  uuid = models.CharField(max_length=36, unique=True, default=uuid.uuid4)

  # uuid of imported object
  imported_uuid = models.CharField(max_length=100, null=True, blank=True)

@receiver(post_save, sender=AppSetting)
@receiver(post_delete, sender=AppSetting)
def delete_appsetting_cache(sender, instance, **kwargs):
  # deleting the caches set in get_settings function in views_settings
  cache.delete_many(['site-setting.True_None',
                     'site-setting.False_None',
                     'site-setting.True_%s' % instance.setting_key,
                     'site-setting.False_%s' % instance.setting_key])

class AppNotificationCenter(models.Model):
  '''persistent Notifications'''

  STATUS_CHOICES = [
    ('R', 'Read'),
    ('U', 'Unread'),
    ('P', 'Pinned'),
  ]

  id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
  notification_type = models.CharField(max_length=10)
  body = models.CharField(max_length=1000)
  title = models.CharField(max_length=200, default='')
  status = models.CharField(max_length=1, choices=STATUS_CHOICES, default='U')
  timestamps = models.JSONField('TimeStamps', default=dict)
  last_occurrence = models.DateTimeField(default=timezone.now)