# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from datetime import datetime, timedelta
from typing import List, Literal, TypedDict
import uuid
from django.db.models import QuerySet
from reports.models import AppNotificationCenter
import logging
from threading import Lock
import time
from django.utils import timezone
from django.conf import settings
from reports.views_settings import setting_get_single

LOGGER = logging.getLogger("arcweb")

NotificationType = Literal['error', 'warning', 'info']
StatusType = Literal['R', 'U', 'P']


class Notification(TypedDict):
    id: uuid.UUID
    type: NotificationType
    title: str
    body: str
    status: StatusType
    timestamps: dict


class PartialNotification(TypedDict):
    title: str
    body: str


class NotificationCenter:
    class __NotificationCenter:
        def __init__(self) -> None:
            self.__notifications = self.__restore_notifications_from_db()

        def add_info(self, info: List[PartialNotification], should_save_in_db: bool=False) -> None:
            self.__add_notifications('info', info, should_save_in_db)

        def add_warnings(self, warnings :List[PartialNotification], should_save_in_db: bool=False) -> None:
            self.__add_notifications('warning', warnings, should_save_in_db)

        def add_errors(self, errors: List[PartialNotification], should_save_in_db: bool=False) -> None:
            self.__add_notifications('error', errors, should_save_in_db)

        def get_info(self) -> List[dict]:
            return [n for n in self.get_all_notifications() if n['type'] == 'info']

        def get_warnings(self) -> List[dict]:
            return [n for n in self.get_all_notifications() if n['type'] == 'warning']

        def get_errors(self) -> List[dict]:
            return [n for n in self.get_all_notifications() if n['type'] == 'error']

        def get_all_notifications(self) -> List[dict]:
            return self.__notifications

        def read_notifications(self, notification_ids: List) -> None:
            notifications = AppNotificationCenter.objects.filter(id__in=notification_ids).filter(status='U')

            def batch(iterable, n=1):
                l = len(iterable)
                for index in range(0, l, n):
                    yield iterable[index:min(index + n, l)]

            for chunk in batch(notifications, 100):
                for notification in chunk:
                    notification.status = 'R'
                AppNotificationCenter.objects.bulk_update(chunk, ['status'])

            notifications_to_update = [notification for notification in self.get_all_notifications() if (notification['id'] in notification_ids) and notification['status'] == 'U']
            for notification in notifications_to_update:
                notification['status'] = 'R'

        def pin_notification(self, notification_id: uuid) -> None:
            notification = AppNotificationCenter.objects.filter(id=notification_id).first()
            if notification and notification.status == 'R':
                notification.status = 'P'
                notification.save()

            notification: dict = [n for n in self.get_all_notifications() if n['id'] == notification_id][0]
            if notification['status'] == 'R':
                notification['status'] = 'P'

        def unpin_notification(self, notification_id: uuid) -> None:
            notification = AppNotificationCenter.objects.filter(id=notification_id).first()
            if notification and notification.status == 'P':
                notification.status = 'R'
                notification.save()

            notification: dict = [n for n in self.get_all_notifications() if n['id'] == notification_id][0]
            if notification['status'] == 'P':
                notification['status'] = 'R'

        def clean_notifications(self):
            LOGGER.info('Cleaning read notification')

            number_of_retention_in_seconds = min(
                float(setting_get_single('notification_max_retention') * 24*60*60),
                int(getattr(settings, 'BULK_REMOVE_READ_NOTIFICATION_INTERVAL', 24*60*60))
            )
            max_retention_date = timezone.make_aware(datetime.now()-timedelta(seconds=number_of_retention_in_seconds))

            self.__notifications = [n for n in self.__notifications if timezone.make_aware(datetime.fromtimestamp(n['last_occurrence'] / 1000)) >= max_retention_date]

            AppNotificationCenter.objects.filter(last_occurrence__lte=max_retention_date).delete()

        def __save_in_db(self, notification: Notification):
            AppNotificationCenter(id=notification['id'], notification_type=notification['type'], title=notification['title'], body=notification['body'], timestamps=notification['timestamps']).save()

        def __get_duplicate_notification(self, notification: PartialNotification):
            duplicates = [n for n in self.__notifications if n['title'] == notification['title'] and n['body'] == notification['body']]

            if len(duplicates) > 1:
                raise Exception

            if len(duplicates) == 1:
                return duplicates[0]

            return None

        def __add_notifications(self, notification_type: NotificationType, notifications: List[PartialNotification], should_save_in_db: bool):
            for notification in notifications:
                    duplicate_notification = self.__get_duplicate_notification(notification)
                    if duplicate_notification:
                        ts = int(time.time()*1000)
                        duplicate_notification['timestamps']['timestamps'].insert(0, ts)

                        try:
                            db_notification = AppNotificationCenter.objects.get(id=duplicate_notification['id'])
                        except AppNotificationCenter.DoesNotExist:
                            db_notification = None

                        if db_notification:
                            db_notification.timestamps['timestamps'].insert(0, ts)
                            db_notification.last_occurrence = ts
                    else:
                        notification_dict = self.__to_notification_dict(notification['title'], notification['body'], notification_type)
                        self.__notifications.insert(0, notification_dict)
                        if should_save_in_db: self.__save_in_db(notification_dict)

        def __restore_notifications_from_db(self) -> List[dict]:
            db_notifications: QuerySet = AppNotificationCenter.objects.all()
            return [self.__db_notification_to_dict(notification) for notification in db_notifications]

        def __db_notification_to_dict(self, notification) -> Notification:
            return {
                'id': str(notification.id),
                'title': notification.title,
                'body': notification.body,
                'status': notification.status,
                'timestamps': notification.timestamps,
                'last_occurrence': int(notification.last_occurrence.timestamp() * 1000),
                'type': notification.notification_type,
                'source': 'backend',
                'hasErrorSync': False,
            }

        def __to_notification_dict(self, title: str, body: str, notifications_type: NotificationType) -> Notification:
            ts = int(time.time() * 1000)

            return {
                'id': str(uuid.uuid4()),
                'title': title,
                'body': body,
                'status': 'U',
                'timestamps': {'timestamps': [ts]},
                'last_occurrence': ts,
                'type': notifications_type,
                'source': 'backend',
                'hasErrorSync': False,
            }

    instance: __NotificationCenter = None
    _lock = Lock()


    def __new__(cls: __NotificationCenter) -> __NotificationCenter:
        if NotificationCenter.instance is None:
            with cls._lock:
                # Another thread could have created the instance
                # before we acquired the lock. So check that the
                # instance is still nonexistent.
                if NotificationCenter.instance is None:
                    NotificationCenter.instance = NotificationCenter.__NotificationCenter()

        return NotificationCenter.instance
