# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from typing import List
from django.test import TestCase
from django.conf import settings

from reports.notification_center import NotificationCenter
from reports.models import AppNotificationCenter


class NotificationCenterTest(TestCase):
    def setUp(self):
        NotificationCenter.instance = None

    def tearDown(self):
        NotificationCenter.instance = None

    def test_when_add_info_notification_should_be_stored(self):
        # Given
        info:List[dict] = [{'title': 'info1-title', 'body': 'info1-body'}, {'title': 'info2-title', 'body': 'info2-body'}]
        NotificationCenter().add_info(info)
        # When
        info_notifications:List[dict] = NotificationCenter().get_info()
        # Then
        info1:dict = info_notifications[1]
        info2:dict = info_notifications[0]
        self.__is_notification_attr_matching(info1, 'info1-title', 'info1-body', 'U')
        self.__is_notification_attr_matching(info2, 'info2-title', 'info2-body', 'U')

    def test_when_add_warnings_notification_should_be_stored(self):
        # Given
        warnings:List[dict] = [{'title': 'warning1-title', 'body': 'warning1-body'}, {'title': 'warning2-title', 'body': 'warning2-body'}]
        NotificationCenter().add_warnings(warnings)
        # When
        warnings_notifications:List[dict] = NotificationCenter().get_warnings()
        # Then
        warning1:dict = warnings_notifications[1]
        warning2:dict = warnings_notifications[0]
        self.__is_notification_attr_matching(warning1, 'warning1-title', 'warning1-body', 'U')
        self.__is_notification_attr_matching(warning2, 'warning2-title', 'warning2-body', 'U')

    def test_when_add_errors_notification_should_be_stored(self):
        # Given
        errors:List[dict] = [{'title': 'error1-title', 'body': 'error1-body'}, {'title': 'error2-title', 'body': 'error2-body'}]
        NotificationCenter().add_errors(errors)
        # When
        errors_notifications:List[dict] = NotificationCenter().get_errors()
        # Then
        error1:dict = errors_notifications[1]
        error2:dict = errors_notifications[0]
        self.__is_notification_attr_matching(error1, 'error1-title', 'error1-body', 'U')
        self.__is_notification_attr_matching(error2, 'error2-title', 'error2-body', 'U')

    def test_when_add_notification_to_db_should_store_properly(self):
        # Given
        info:List[dict] = [{'title': 'info1-title', 'body': 'info1-body'}]
        warnings:List[dict] = [{'title': 'warning1-title', 'body': 'warning1-body'}]
        errors:List[dict] = [{'title': 'error1-title', 'body': 'error1-body'}]
        # When
        NotificationCenter().add_info(info, True)
        NotificationCenter().add_warnings(warnings, True)
        NotificationCenter().add_errors(errors, True)
        # Then
        info1 = AppNotificationCenter.objects.filter(notification_type='info').first()
        warning1 = AppNotificationCenter.objects.filter(notification_type='warning').first()
        error1 = AppNotificationCenter.objects.filter(notification_type='error').first()

        self.assertEqual(info1.title, 'info1-title')
        self.assertEqual(warning1.title, 'warning1-title')
        self.assertEqual(error1.title, 'error1-title')

    def test_when_retrieve_notification_from_db_should_parse_it_correctly(self):
        # Given
        AppNotificationCenter(notification_type='info', title='info1-title', body='info1-body', status='R').save()
        AppNotificationCenter(notification_type='warning', title='warning1-title', body='warning1-body').save()
        AppNotificationCenter(notification_type='error', title='error1-title', body='error1-body').save()
        # When
        notifications:List[dict] = NotificationCenter().get_all_notifications()
        # Then
        self.__is_notification_attr_matching(notifications[0], 'info1-title', 'info1-body', 'R')
        self.__is_notification_attr_matching(notifications[1], 'warning1-title', 'warning1-body', 'U')
        self.__is_notification_attr_matching(notifications[2], 'error1-title', 'error1-body', 'U')

    def test_when_read_notifications_should_update_attr(self):
        # Given
        info:List[dict] = [{'title': 'info1-title', 'body': 'info1-body'}]
        # When
        NotificationCenter().add_info(info, True)
        notification_id = NotificationCenter().get_info()[0]['id']
        NotificationCenter().read_notifications([notification_id])
        # Then
        notification_status = NotificationCenter().get_info()[0]['status']
        self.assertEqual(notification_status, 'R')
        self.assertEqual(AppNotificationCenter.objects.filter(id=notification_id).first().status, 'R')

    def test_when_pin_and_unpin_notification_should_update_attr(self):
        # Given
        info:List[dict] = [{'title': 'info1-title', 'body': 'info1-body'}]
        # When
        NotificationCenter().add_info(info, True)
        notification_id = NotificationCenter().get_info()[0]['id']
        NotificationCenter().read_notifications([notification_id])
        NotificationCenter().pin_notification(notification_id)
        # Then
        notification_status = NotificationCenter().get_info()[0]['status']
        self.assertEqual(notification_status, 'P')
        self.assertEqual(AppNotificationCenter.objects.filter(id=notification_id).first().status, 'P')

        NotificationCenter().unpin_notification(notification_id)
        notification_status = NotificationCenter().get_info()[0]['status']
        self.assertEqual(notification_status, 'R')
        self.assertEqual(AppNotificationCenter.objects.filter(id=notification_id).first().status, 'R')

    def test_when_clean_notification_should_delete_all_notifications(self):
        # Given
        info:List[dict] = [{'title': 'info1-title', 'body': 'info1-body'}]
        warnings:List[dict] = [{'title': 'warning1-title', 'body': 'warning1-body'}]
        errors:List[dict] = [{'title': 'error1-title', 'body': 'error1-body'}]
        # When
        NotificationCenter().add_info(info, True)
        NotificationCenter().add_warnings(warnings, True)
        NotificationCenter().add_errors(errors, True)

        setattr(settings, 'notification_max_retention', 0)
        setattr(settings, 'BULK_REMOVE_READ_NOTIFICATION_INTERVAL', 0)

        NotificationCenter().clean_notifications()
        # Then
        self.assertEqual(len(NotificationCenter().get_all_notifications()), 0)
        self.assertEqual(len(AppNotificationCenter.objects.all()), 0)

    def __is_notification_attr_matching(self, notification: dict, title: str, body: str, status: bool):
        self.assertNotEqual(notification['id'], None)
        self.assertEqual(notification['title'], title)
        self.assertEqual(notification['body'], body)
        self.assertEqual(notification['status'], status)
