# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
from django.test import TestCase
from django.conf import settings
from datetime import datetime
from django.contrib.auth.models import User
from reports.models import Report
from jobs.views_encryption import encrypt_secret
from unittest.mock import patch
from parameterized import parameterized_class
from reports.utils import (
    _get_report_data,
    replace_dummy_api_key_for_save,
    replace_ai_api_key_with_placeholder,
    replace_all_ai_api_keys_with_placeholder
)

def getMockedSecret(value='api_key'):
    return f'{{"secret": "encrypted_{value}", "algo": "pbkdf2_sha256", "iterations": "100000", "salt": "+cCqslI9Tw2dI6Yj"}}'

class TestReportUtils(TestCase):
    def setUp(self):
        # Create a test user
        self.user = User.objects.create_user(
            username='testuser',
            email='testuser@example.com',
            password='password',
            is_superuser=False,
            is_staff=False,
            is_active=True
        )
        # Create a test report with sample data
        self.report = Report(
            report_name="Census Dashboard",
            report_type="dashboard",
            create_time=datetime(2022, 6, 23, 11, 49, 16, 408335),
            update_time=datetime(2022, 6, 23, 11, 49, 16, 408348),
            id="66",
            dataset=None,
            report_data=json.dumps({
                "numColumns": 64,
                "dashboard_widgets": [
                    {"col": 1, "row": 9, "size_x": 21, "size_y": 11, "id": "uri-1-widget-37"},
                    {"col": 1, "row": 25, "size_x": 45, "size_y": 32, "id": "uri-2-widget-36"},
                    {"col": 46, "row": 25, "size_x": 19, "size_y": 16, "id": "uri-3-widget-39"},
                ],
                "user_settings": {
                    "dashboard_width": "1500",
                    "single_visual_dashboard": "false",
                    "permit_csv_download_dashboard": "true",
                    "permit_email_report": "true"
                }
            })
        )
        # Old API key value (before any changes)
        self.old_value = json.dumps({
            "profileSettings": {
                "profile1": {
                    "values": {"completion_api_key": getMockedSecret("old_api_key"), "completion_auth_key": getMockedSecret("old_auth_key") }
                }
            }
        })
        # New API key value with placeholder
        self.new_value = json.dumps({
            "profileSettings": {
                "profile1": {
                    "values": {"completion_api_key": settings.DEFAULT_PASSWORD_PLACEHOLDER, "completion_auth_key": settings.DEFAULT_PASSWORD_PLACEHOLDER}
                }
            }
        })

    def test_get_report_data(self):
        with self.assertLogs('arcweb.reports.utils', level='WARNING') as log:
            report_data = _get_report_data(self.report, self.user)

            # Assert that a warning was logged
            self.assertEqual(len(log.output), 1)
            self.assertIn("Report %s requires default data connection" % self.report.id, log.output[0])

            # Assert that the returned report data is a dictionary and contains correct 'id'
            self.assertIsInstance(report_data, dict)
            self.assertIn('id', report_data)
            self.assertEqual(report_data['id'], self.report.id)

    def test_replace_dummy_api_key_for_save(self):
        """
        Test that when the new value contains the placeholder,
        it is replaced with the old actual API key.
        """
        result = replace_dummy_api_key_for_save(self.new_value, self.old_value, "completion")
        result_dict = json.loads(result)

        self.assertEqual(
            result_dict['profileSettings']['profile1']['values']['completion_api_key'],
            getMockedSecret("old_api_key"),
            "The placeholder for the API key should be replaced with the old API key"
        )

        self.assertEqual(
            result_dict['profileSettings']['profile1']['values']['completion_auth_key'],
            getMockedSecret("old_auth_key"),
            "The placeholder for the auth key should be replaced with the old auth key"
        )

    
    # Common variables
    BASE_PROFILE = {
        "profileSettings": {
            "profile1": {
                "values": {
                    "completion_api_key": "old_api_key",
                    "completion_auth_key": "old_auth_key"
                }
            }
        }
    }

    EMPTY_PROFILE = {
        "profileSettings": {
            "profile1": {}
        }
    }

    NEW_KEYS_BOTH = {"new_api_key": "new_api_key", "new_auth_key": "new_auth_key"}
    NEW_KEYS_API_ONLY = {"new_api_key": "new_api_key_only", "new_auth_key": ""}
    NEW_KEYS_AUTH_ONLY = {"new_api_key": "", "new_auth_key": "new_auth_key_only"}
    EMPTY_KEYS = {"new_api_key": "", "new_auth_key": ""}

    @parameterized_class([
        {
            "description": "Replace old API and auth key with new encrypted ones",
            "new_keys": NEW_KEYS_BOTH,
            "old_profile": json.dumps(BASE_PROFILE),
        },
        {
            "description": "Adds new encrypted API and auth keys for the first time",
            "new_keys": NEW_KEYS_BOTH,
            "old_profile": json.dumps({}),
        },
        {
            "description": "Properly saves empty API and auth keys",
            "new_keys": EMPTY_KEYS,
            "old_profile": json.dumps(BASE_PROFILE),
        },
        {
            "description": "Properly saves empty API and auth keys for the first time",
            "new_keys": EMPTY_KEYS,
            "old_profile": json.dumps({}),
        },
        {
            "description": "Replace only API key with new encrypted one",
            "new_keys": NEW_KEYS_API_ONLY,
            "old_profile": json.dumps(BASE_PROFILE),
        },
        {
            "description": "Replace only auth key with new encrypted one",
            "new_keys": NEW_KEYS_AUTH_ONLY,
            "old_profile": json.dumps(BASE_PROFILE),
        },
        {
            "description": "Replace only API key with new encrypted one (missing auth key)",
            "new_keys": {"new_api_key": "new_api_key_only_missing_auth_key"},
            "old_profile": json.dumps(BASE_PROFILE),
        },
        {
            "description": "Replace only auth key with new encrypted one (missing api key)",
            "new_keys": {"new_auth_key": "new_auth_key_only_missing_api_key"},
            "old_profile": json.dumps(BASE_PROFILE),
        },
    ])
    class TestReplaceDummyApiKeyWhenUpdated(TestCase):
        @patch('reports.utils.encrypt_secret')
        def test_replace_dummy_api_key(self, mock_encrypt_secret):
            # Mock encrypt_secret to return a mocked encrypted value
            mock_encrypt_secret.side_effect = lambda key: getMockedSecret(key)

            non_placeholder_value = {
                "profileSettings": {
                    "profile1": {
                        "values": {}
                    }
                }
            }

            new_api_key = self.new_keys.get("new_api_key")
            new_auth_key = self.new_keys.get("new_auth_key")

            # Set only non-None values in the profile settings
            if new_api_key is not None:
                non_placeholder_value['profileSettings']['profile1']['values']["completion_api_key"] = new_api_key
                encrypted_new_api_key = json.dumps(getMockedSecret(new_api_key))
            if new_auth_key is not None:
                non_placeholder_value['profileSettings']['profile1']['values']["completion_auth_key"] = new_auth_key
                encrypted_new_auth_key = json.dumps(getMockedSecret(new_auth_key))

            result = replace_dummy_api_key_for_save(
                json.dumps(non_placeholder_value),
                self.old_profile,
                "completion"
            )
            result_dict = json.loads(result)

            # Assertions
            if new_api_key is not None:
                self.assertEqual(
                    result_dict['profileSettings']['profile1']['values'].get('completion_api_key'),
                    encrypted_new_api_key,
                    f"{self.description} - The placeholder for the API key should be replaced with the encrypted new API key.",
                )
            if new_auth_key is not None:
                self.assertEqual(
                    result_dict['profileSettings']['profile1']['values'].get('completion_auth_key'),
                    encrypted_new_auth_key,
                    f"{self.description} - The placeholder for the auth key should be replaced with the encrypted new auth key.",
                )

    def test_replace_dummy_api_key_for_deleting_profile_settings(self):
        """
        Test that when 'profileSettings' is missing in either new_value or old_value,
        the function returns new_value unchanged.
        """
        # new_value missing 'profileSettings'
        new_value_no_profile = json.dumps({})
        result = replace_dummy_api_key_for_save(new_value_no_profile, self.old_value, "completion")

        self.assertEqual(result, new_value_no_profile)
        
    def test_replace_dummy_api_key_for_save_missing_values_key(self):
        """
        Test that the function handles profiles without a 'values' key without raising exceptions.
        """
        new_value_missing_values = json.dumps({
            "profileSettings": {
                "profile1": {
                    # 'values' key is missing here
                }
            }
        })
        result = replace_dummy_api_key_for_save(new_value_missing_values, self.old_value, "completion")
        result_dict = json.loads(result)

        self.assertNotIn(
            'values',
            result_dict['profileSettings']['profile1'],
            "The function should not add a 'values' key where it was missing."
        )


    @patch('reports.utils.encrypt_secret')
    def test_replace_dummy_api_key_for_save_multiple_profiles(self, mock_encrypt_secret):
        """
        Test that the function correctly processes multiple profiles with different conditions
        for both API and auth keys.
        """
        # Mock encrypt_secret to return different values based on input
        def mock_encrypt(value):
            if value == 'new_api_key_profile2':
                return getMockedSecret('new_api_key_profile2')
            elif value == 'new_auth_key_profile2':
                return getMockedSecret('new_auth_key_profile2')
            else:
                return getMockedSecret(value)

        mock_encrypt_secret.side_effect = mock_encrypt

        # Test input JSON with placeholders and new keys for both API and Auth keys
        new_value_multiple_profiles = json.dumps({
            "profileSettings": {
                "profile1": {
                    "values": {
                        "completion_api_key": settings.DEFAULT_PASSWORD_PLACEHOLDER,
                        "completion_auth_key": settings.DEFAULT_PASSWORD_PLACEHOLDER
                    }
                },
                "profile2": {
                    "values": {
                        "completion_api_key": "new_api_key_profile2",
                        "completion_auth_key": "new_auth_key_profile2"
                    }
                }
            }
        })

        old_value_multiple_profiles = json.dumps({
            "profileSettings": {
                "profile1": {
                    "values": {
                        "completion_api_key": getMockedSecret("old_api_key_profile1"),
                        "completion_auth_key": getMockedSecret("old_auth_key_profile1")
                    }
                },
                "profile2": {
                    "values": {
                        "completion_api_key": getMockedSecret("old_api_key_profile2"),
                        "completion_auth_key": getMockedSecret("old_auth_key_profile2")
                    }
                }
            }
        })

        result = replace_dummy_api_key_for_save(
            new_value_multiple_profiles, old_value_multiple_profiles, "completion"
        )
        result_dict = json.loads(result)

        self.assertEqual(
            result_dict['profileSettings']['profile1']['values']['completion_api_key'],
            getMockedSecret("old_api_key_profile1"),
            "Profile1 should have the old API key replaced correctly (placeholder replaced with old value)."
        )
        self.assertEqual(
            result_dict['profileSettings']['profile1']['values']['completion_auth_key'],
            getMockedSecret("old_auth_key_profile1"),
            "Profile1 should have the old Auth key replaced correctly (placeholder replaced with old value)."
        )

        encrypted_new_api_key = json.dumps(getMockedSecret("new_api_key_profile2"))
        encrypted_new_auth_key = json.dumps(getMockedSecret("new_auth_key_profile2"))
        
        self.assertEqual(
            result_dict['profileSettings']['profile2']['values']['completion_api_key'],
            encrypted_new_api_key,
            "Profile2 should have the new API key encrypted and stored correctly."
        )
        self.assertEqual(
            result_dict['profileSettings']['profile2']['values']['completion_auth_key'],
            encrypted_new_auth_key,
            "Profile2 should have the new Auth key encrypted and stored correctly."
        )

    def test_replace_dummy_api_key_for_save_source_key_removal(self):
        """
        Test that the 'source' key is removed from the profile data when 'source_id' is present,
        and that the API key is replaced with the one from the source profile.
        """

        # new value with cloned profile and source property
        new_value_with_source = json.dumps({
            "profileSettings": {
                "profile1": {
                    "label": "Profile 1",
                    "values": {
                        "completion_api_key": settings.DEFAULT_PASSWORD_PLACEHOLDER,
                        "completion_auth_key": settings.DEFAULT_PASSWORD_PLACEHOLDER
                    }
                },
                "profile2": {
                    "label": "Profile 2",
                    "values": {"completion_api_key": settings.DEFAULT_PASSWORD_PLACEHOLDER}
                },
                "profile1Copy": {
                    "source": "profile1",
                    "label": "Profile 1 Copy",
                    "values": {
                        "completion_api_key": settings.DEFAULT_PASSWORD_PLACEHOLDER,
                        "completion_auth_key": settings.DEFAULT_PASSWORD_PLACEHOLDER
                    }
                }
            }
        })

        # old value with sourced profile
        old_value_with_source_profile = json.dumps({
            "profileSettings": {
                "profile1": {
                    "label": "Profile 1",
                    "values": {
                        "completion_api_key": getMockedSecret("api_key_for_profile1"),
                        "completion_auth_key": getMockedSecret("auth_key_for_profile1")
                    }
                },
                "profile2": {
                    "label": "Profile 2",
                    "values": {"completion_api_key": getMockedSecret("api_key_for_profile2")}
                }
            }
        })

        result = replace_dummy_api_key_for_save(
            new_value_with_source,
            old_value_with_source_profile,
            "completion"
        )
        result_dict = json.loads(result)

        self.assertNotIn(
            'source',
            result_dict['profileSettings']['profile1Copy'],
            "The 'source' key should be removed from profile1Copy."
        )

        self.assertEqual(
            result_dict['profileSettings']['profile1Copy']['values']['completion_api_key'],
            getMockedSecret("api_key_for_profile1"),
            "The 'completion_api_key' in profile1Copy should match the API key from profile1."
        )
        self.assertEqual(
            result_dict['profileSettings']['profile1Copy']['values']['completion_auth_key'],
            getMockedSecret("auth_key_for_profile1"),
            "The 'completion_auth_key' in profile1Copy should match the Auth key from profile1."
        )

    def test_replace_ai_api_key_with_placeholder(self):
        """
        Test that the AI API key is replaced with the placeholder
        when a key is present in the profiles.
        """
        profiles = {
            "profile1": {
                "values": {
                    "completion_api_key": getMockedSecret("some_key"),
                    "completion_auth_key": getMockedSecret("some_auth_key")
                    }
            }
        }
        replace_ai_api_key_with_placeholder(profiles, "completion")
        
        # Assert that both API and auth keys are replaced with the placeholder
        self.assertEqual(
            profiles["profile1"]["values"]["completion_api_key"],
            settings.DEFAULT_PASSWORD_PLACEHOLDER,
            "The API key in profile1 should be replaced with the default placeholder."
        )
        self.assertEqual(
            profiles["profile1"]["values"]["completion_auth_key"],
            settings.DEFAULT_PASSWORD_PLACEHOLDER,
            "The auth key in profile1 should be replaced with the default placeholder."
        )

    def test_replace_ai_api_key_with_placeholder_no_value(self):
        """
        Test that if the AI API key is empty, it remains unchanged
        (i.e., not replaced with the placeholder).
        """
        profiles = {
            "profile1": {
                "values": {"completion_api_key": "", "completion_auth_key": ""}
            }
        }
        replace_ai_api_key_with_placeholder(profiles, "completion")
        
        self.assertEqual(
            profiles["profile1"]["values"]["completion_api_key"],
            "",
            "The API key in profile1 should remain unchanged as an empty string."
        )
        self.assertEqual(
            profiles["profile1"]["values"]["completion_auth_key"],
            "",
            "The auth key in profile1 should remain unchanged as an empty string."
        )
    
    def test_replace_all_ai_api_keys_with_placeholder(self):
        """
        Test that all AI API keys in the settings dictionary are replaced with the placeholder.
        """
        settings_dict = {
            "ai_completion_settings": json.dumps({
                "profileSettings": {
                    "profile1": {
                        "values": {
                            "completion_api_key": getMockedSecret("some_key"),
                            "completion_auth_key": getMockedSecret("some_auth_key")
                        }
                    }
                }
            }),
            "ai_embeddings_settings": json.dumps({
                "profileSettings": {
                    "profile1": {
                        "values": {
                            "embeddings_api_key": getMockedSecret("some_other_key"),
                            "embeddings_auth_key": getMockedSecret("some_other_auth_key")
                        }
                    }
                }
            })
        }

        result = replace_all_ai_api_keys_with_placeholder(settings_dict)
        result_completion_dict = json.loads(result["ai_completion_settings"])
        
        self.assertEqual(
            result_completion_dict["profileSettings"]["profile1"]["values"]["completion_api_key"],
            settings.DEFAULT_PASSWORD_PLACEHOLDER,
            "The completion API key in profile1 should be replaced with the default placeholder."
        )
        self.assertEqual(
            result_completion_dict["profileSettings"]["profile1"]["values"]["completion_auth_key"],
            settings.DEFAULT_PASSWORD_PLACEHOLDER,
            "The completion auth key in profile1 should be replaced with the default placeholder."
        )

        result_embeddings_dict = json.loads(result["ai_embeddings_settings"])

        self.assertEqual(
            result_embeddings_dict["profileSettings"]["profile1"]["values"]["embeddings_api_key"],
            settings.DEFAULT_PASSWORD_PLACEHOLDER,
            "The embeddings API key in profile1 should be replaced with the default placeholder."
        )
        self.assertEqual(
            result_embeddings_dict["profileSettings"]["profile1"]["values"]["embeddings_auth_key"],
            settings.DEFAULT_PASSWORD_PLACEHOLDER,
            "The embeddings auth key in profile1 should be replaced with the default placeholder."
        )

    def test_replace_all_ai_api_keys_with_placeholder_no_profile(self):
        """
        Test that if there are no profiles in the settings dictionary,
        the function returns the original settings without changes.
        """
        settings_dict = {
            "ai_completion_settings": json.dumps({})
        }
        result = replace_all_ai_api_keys_with_placeholder(settings_dict)

        self.assertEqual(
            result,
            settings_dict,
            "The settings should remain unchanged when no profiles are present."
        )
