# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import django.test
from django.test.client import Client
import json

from rbac.models import Role
from rbac.permcache import RBAC
from reports.models import Report
from reports.views import can_user_access_dataset, report

from vizlib.tests import test_report_operations as test_utils

helper = test_utils.TestDashboardOperations()
from vizlib.tests.util import run_with_db_query_check


def make_sysadmin_role():
  role = Role()
  role.name = "Test Sysadmin"
  role.privileges = json.dumps(
    [
      {
        "ptype": "system",
        "perms": [
          "sys_workspace",
          "sys_schedules",
          "sys_viewlogs",
          "sys_editperm",
          "sys_styles",
          "sys_viewroles",
          "sys_settings",
          "sys_editconn",
          "sys_misc"
        ]
      },
      {
        "ptype": "dataconn",
        "dclist": [
          "-1"
        ],
        "perms": [
          "dc_upload",
          "dc_explore",
          "dc_aviews"
        ]
      },
      {
        "ptype": "dataset",
        "dcid": "-1",
        "dslist": [
          "-1"
        ],
        "perms": [
          "ds_appedit",
          "ds_appview",
          "ds_manage"
        ]
      }
    ]
  )
  role.save()
  return role

def make_ds_viewer_role(role_name, dc_id: str, ds_ids: list[str]):
  role = Role()
  role.name = role_name
  role.privileges = json.dumps(
      [
          # no system permissions:
          {'ptype': 'system', 'perms': []},
          # view rights for specified datasets
          {
              'ptype': 'dataset',
              'dcid': str(dc_id),
              'dslist': [str(ds_id) for ds_id in ds_ids],
              'perms': ['ds_appview'],
          },
      ]
  )
  role.save()
  return role

class GetReportsTest(django.test.TestCase):

  def setUp(self):
    self.admin_user = test_utils.make_user('admin', is_superuser=True)
    self.admin_client = test_utils.create_client(self.admin_user)

    self.normal_user = test_utils.make_user('user1', is_superuser=False)
    self.normal_client = test_utils.create_client(self.normal_user)

    self.normal_user_2 = test_utils.make_user('user2', is_superuser=False)
    self.normal_client_2 = test_utils.create_client(self.normal_user_2)

    self.sysadmin = test_utils.make_user('sysadmin1', is_superuser=False)
    sysadmin_role = make_sysadmin_role()
    sysadmin_role.add_users([self.sysadmin.username])
    sysadmin_role.save()
    self.sysadmin_client = test_utils.create_client(self.sysadmin)

    dc_id = helper.create_dataconnection(self.admin_client)

    (self.ds1_id, db_1) = self.create_dataset_and_dashboard(dc_id, 'Dataset 1')
    (self.ds2_id, _) = self.create_dataset_and_dashboard(dc_id, 'Dataset 2')
    (self.ds3_id, db_3) = self.create_dataset_and_dashboard(dc_id, 'Dataset 3')

    helper.move_to_public_workspace(self.admin_client, db_1)
    helper.move_to_public_workspace(self.admin_client, db_3)

    ds1_view_role = make_ds_viewer_role(
      'ds1_viewer', dc_id, [self.ds1_id, self.ds2_id])
    ds1_view_role.add_users([self.normal_user.username])
    ds1_view_role.save()

  def create_dataset_and_dashboard(self, dc_id, ds_name):
    ds_id = helper.create_dataset(self.admin_client, dc_id, ds_name)

    visual_id = helper.create_visual(self.admin_client, ds_id)
    picklist_id = helper.create_visual(
      self.admin_client, ds_id, visual_type='picklist')
    dashboard_id = helper.create_dashboard(
      self.admin_client, ds_id, [visual_id, picklist_id])

    return (ds_id, dashboard_id)


  def _get(self, client: Client, url: str, expected_query_count:int=None) -> list[Report]:
    response, _ = run_with_db_query_check(
      lambda: client.get(url),
      asserter=self,
      expected_query_count=expected_query_count
      )
    self.assertEquals(response.status_code, 200)
    return response.json()

  def user_with_dataset_permission_get(self, url: str, expected_query_count:int=None) -> list[Report]:
    return self._get(self.normal_client, url, expected_query_count)

  def user_without_dataset_permission_get(self, url: str) -> list[Report]:
    return self._get(self.normal_client_2, url)

  def test_sysdmin_get(self):
    # sys admins only get dashboards that they have both dataset and workspace permission for
    response = self._get(self.sysadmin_client, '/reports/report')
    self.assertEquals(2, len(response))

    # extra queries compared to admin: group+role+accessible datasets
    response = self._get(self.sysadmin_client, '/reports/report', expected_query_count=10+3)
    self.assertEquals(2, len(response))

  def test_admin_get(self):
    # dashboards: as admins we get all of them
    response = self._get(self.admin_client, '/reports/report')
    self.assertEquals(3, len(response))

    # first run creates reports_appgroupappmembership entries, generates lots of extra queries
    response = self._get(self.admin_client, '/reports/report', expected_query_count=10)
    self.assertEquals(3, len(response))

  def _test_adminlike_get(self, client: Client):
    # private workspace access
    response = self._get(client, f'/reports/report?dataset_id={self.ds2_id}')
    self.assertEquals(1, len(response))
    self.assertEquals(response[0]['report_type'], 'dashboard')

    # all=true returns all non-filters (dashboards+regular visuals)
    response = self._get(client,
      f'/reports/report?dataset_id={self.ds2_id}&all=true')
    self.assertEquals(2, len(response))
    self.assertEquals(set(['dashboard', 'table']), set(
      [r['report_type'] for r in response]))

    # all=true can skip dashboards if asked:
    response = self._get(client,
      f'/reports/report?dataset_id={self.ds2_id}&all=true&exclude_dashboards=true')
    self.assertEquals(1, len(response))
    self.assertEquals(response[0]['report_type'], 'table')

    # global all:
    response = self._get(client, f'/reports/report?all=true')
    self.assertEquals(6, len(response))
    self.assertEquals(set(['dashboard', 'table']), set(
      [r['report_type'] for r in response]))

  def test_user_get(self):
    # dashboards: we only get the dashboard from the public workspace
    response = self.user_with_dataset_permission_get('/reports/report')
    self.assertEquals(1, len(response))

    # extra queries compared to admin: group+role+accessible datasets+workspaces
    response = self.user_with_dataset_permission_get('/reports/report', expected_query_count=14)
    self.assertEquals(1, len(response))

    # no reports returned because user has no dataset permissions
    response = self.user_without_dataset_permission_get('/reports/report')
    self.assertEquals(0, len(response))

    # public workspace access - should work
    response = self.user_with_dataset_permission_get(f'/reports/report?dataset_id={self.ds1_id}')
    self.assertEquals(1, len(response))
    self.assertEquals(response[0]['report_type'], 'dashboard')

    # public workspace access but no dataset permission - should not work
    response = self.user_without_dataset_permission_get(f'/reports/report?dataset_id={self.ds1_id}')
    self.assertEquals(0, len(response))

    # all=true returns non-dashboards and non-filters
    response = self.user_with_dataset_permission_get(
      f'/reports/report?dataset_id={self.ds1_id}&all=true&exclude_dashboards=true')
    self.assertEquals(1, len(response))
    self.assertEquals(response[0]['report_type'], 'table')

    response = self.user_without_dataset_permission_get(
      f'/reports/report?dataset_id={self.ds1_id}&all=true&exclude_dashboards=true')
    self.assertEquals(0, len(response))

    # private workspace access - not allowed
    response = self.user_with_dataset_permission_get(f'/reports/report?dataset_id={self.ds2_id}')
    self.assertEquals(0, len(response))

    # all=true returns non-dashboards and non-filters
    response = self.user_with_dataset_permission_get(
      f'/reports/report?dataset_id={self.ds2_id}&all=true')
    self.assertEquals(0, len(response))

    # public workspace but private dataset:
    response = self.user_with_dataset_permission_get(f'/reports/report?dataset_id={self.ds3_id}')
    self.assertEquals(0, len(response))

    # all=true returns non-filters
    response = self.user_with_dataset_permission_get(
      f'/reports/report?dataset_id={self.ds3_id}&all=true')
    self.assertEquals(0, len(response))

    # global all:
    response = self.user_with_dataset_permission_get(f'/reports/report?all=true')
    self.assertEquals(2, len(response))
    self.assertEquals(set(['dashboard', 'table']), set(
      [r['report_type'] for r in response]))

    response = self.user_without_dataset_permission_get( f'/reports/report?all=true')
    self.assertEquals(0, len(response))


class TestDatasetPermissionCheck(django.test.TestCase):
  def setUp(self):
    self.helper = GetReportsTest()
    self.helper.setUp()

  def test_ds_perm_check(self):
    RBAC.reset()
    access_granted, db_query_count = run_with_db_query_check(
      lambda: can_user_access_dataset(self.helper.normal_user, self.helper.ds1_id))

    # there must be only three DB queries issued:
    """
    DB queries count 3
      SELECT "datasets_dataset"."id", "datasets_dataset"."dataconnection_id", "datasets_dataset"."create_user" FROM "datasets_dataset" WHERE "datasets_dataset"."id" = 1 LIMIT 21
      SELECT "auth_group"."name" FROM "auth_group" INNER JOIN "auth_user_groups" ON ("auth_group"."id" = "auth_user_groups"."group_id") WHERE "auth_user_groups"."user_id" = 2
      SELECT "rbac_role"."id", "rbac_role"."name", "rbac_role"."description", "rbac_role"."privileges", "rbac_role"."users", "rbac_role"."groups", "rbac_role"."sentry_groups", "rbac_role"."require_all_groups" FROM "rbac_role"
    """
    self.assertEqual(db_query_count, 3)

    self.assertTrue(access_granted)

    # normal_user has no access to dataset 3:
    access_granted = can_user_access_dataset(self.helper.normal_user, self.helper.ds3_id)
    self.assertFalse(access_granted)

class TestDashboardDeleting(django.test.TestCase):

  def test_delete_broken_dashboard(self):
    admin = test_utils.make_user('admin', is_superuser=True)
    admin_client = test_utils.create_client(admin)
    util = test_utils.TestDashboardOperations()

    dc_id = util.create_dataconnection(admin_client)
    ds_1 = util.create_dataset(admin_client, dc_id, 'ds_1')

    visual_1 = util.create_visual(admin_client, ds_1)
    visual_2 = util.create_visual(admin_client, ds_1)

    dashboard_id = util.create_dashboard(
      admin_client, ds_1, [visual_1, visual_2])

    # break the dashboard
    Report.objects.get(id=visual_2).delete()

    response, _ = run_with_db_query_check(
      lambda: admin_client.delete(f'/reports/report/{dashboard_id}'))
    self.assertEqual(response.status_code, 200)
    self.assertEqual(Report.objects.filter(
      id__in=[dashboard_id, visual_1]).count(), 0)