# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import logging

from django.conf import settings
from typing import Literal
from util.ff_flipper import FeatureFlipper
from reports.models import Report
from reports import constants
from workspace.views import update_private_name
from datasets.utils import get_default_dataconnection
from jobs.views_encryption import encrypt_secret

LOGGER = logging.getLogger('arcweb.reports.utils')

def _get_report_data(r, user, views_counts=None,
                     user_reports=None,
                     include_data=False,
                     cached_ril=None,
                     get_clone_info=None):
  """Get a dict representing report data"""

  # This should never be needed
  default_dataconnection = None
  if not r.dataset or not r.dataset.dataconnection:
    LOGGER.warning("Report %s requires default data connection" % r.id)
    default_dataconnection = get_default_dataconnection()

  views_counts = views_counts or {}
  user_reports = user_reports or {}
  descendants_count = 'n/a'
  is_clone = False

  if FeatureFlipper().get_flipper_value('enable_clone_info_badge'):
    if get_clone_info:
      if not constants.ALL_CLONE_REPORT_IDS and not constants.ALL_PARENT_REPORT_IDS:
        setup_clone_related_report_ids()

      descendants_count = len(Report.get_relatives_chain(r, 'children')) if r.id in constants.ALL_PARENT_REPORT_IDS else 0
      is_clone =  r.id in constants.ALL_CLONE_REPORT_IDS

  rd = {
      'id': r.id,
      'name': r.report_name,
      'description': r.report_description,
      'type': r.report_type,
      'created': r.create_time.isoformat(),
      'report_tags': 'applications' if r.report_type == 'dashboard' else "",
      'created_by': r.create_user,
      'updated': r.update_time.isoformat(),
      'updated_by': r.update_user,
      'views_count': views_counts.get(r.id, 0),
      'descendants_count': descendants_count,
      'is_clone': is_clone,
      'mode':r.report_mode,
  }
  if r.id in user_reports:
    rd['in_userlist'] = True

  # dataset id will exist for dashboards if all reports are
  # part of same dataset, else it'll be set to none
  try:
    rd['dataset_id'] = r.dataset.id
    rd['dataset_name'] = r.dataset.dataset_name
    if r.dataset.dataconnection:
      rd['dataconn_name'] = r.dataset.dataconnection.dataconnection_name
    elif default_dataconnection:
      rd['dataconn_name'] = default_dataconnection.dataconnection_name
    else:
      rd['dataconn_name'] = ''
    rd['dataset_detail'] = r.dataset.dataset_detail
    rd['dataset_type'] = r.dataset.dataset_type

  except Exception:
    rd['dataset_id'] = ''
    rd['dataset_name'] = ''
    rd['dataset_detail'] = ''
    rd['dataset_type'] = ''

  if include_data:
    rd['data'] = r.report_data
    rd['report_search_text'] = r.report_search_text
  try:
    o = json.loads(r.report_data)
  except Exception:
    o = {}

  if r.report_type == "dashboard":
    (visuals, pickers) = r.get_visuals_in_dashboard(report_data=o)
    rd['visuals'] = visuals
    rd['pickers'] = pickers
  else:
    rd['visuals'] = []
    rd['pickers'] = []

  rd['report_type'] = r.report_type
  rd['edit_url'] = "%s/apps/builder/%s" % (settings.ARC_BASEURL, r.id)

  if r.shared_visual_dashboards:
    # Keep as json so python doesn't evaluate empty array as falsy value
    rd['shared_visual_dashboards'] = r.shared_visual_dashboards
  else:
    rd['dashboard_id'] = o.get('dashboard_id', '')

  if rd['report_type'] == "dashboard":
    rd['title'] = o.get('report_title', '')
  else:
    viz_report_data = o.get('report_data', {})
    rd['title'] = viz_report_data.get('report_title', '')
    rd['dashboard_id'] = viz_report_data.get('dashboard_id', '')

  rd['url'] = "%s/apps/app/%s" % (settings.ARC_BASEURL, r.id)

  img = None
  baseImg = "%s/static/img/icon_%s.png?v=6" % (settings.ARC_BASEURL,
      r.report_type)
  if cached_ril is not None and r.id in cached_ril:
    img = "%s/apps/snapshot/%s?v=%s" % (settings.ARC_BASEURL, r.id,
        cached_ril[r.id])
  else:
    img = baseImg

  rd['img'] = img
  rd['baseImg'] = baseImg
  rd['workspace_id'] = r.workspace_id

  workspace_name_desc = update_private_name(r.workspace, user)
  rd['workspace_name'] = workspace_name_desc.get('name')
  rd['workspace_perm'] = list(r.workspace.getPerms(user))

  if r.dataset and r.dataset.dataconnection:
    dc = r.dataset.dataconnection
  else:
    dc = default_dataconnection
  dsperms = set([]) if not r.dataset else r.dataset.getPerms(user)
  dcperms = set([]) if not dc else dc.getPerms(user)
  rd['dataset_perm'] = list(dsperms.union(dcperms))
  dsperms = r.workspace.maskPerms(user, dsperms)
  rd['perm'] = list(dsperms.union(dcperms))

  return rd

def setup_clone_related_report_ids():
  constants.ALL_CLONE_REPORT_IDS = set()
  constants.ALL_PARENT_REPORT_IDS = set()
  reports = Report.objects.all()
  for report in reports.iterator():
    if report.parent_report_id:
      constants.ALL_CLONE_REPORT_IDS.add(report.id)
      constants.ALL_PARENT_REPORT_IDS.add(report.parent_report_id)
      
keys_to_replace_completion = ['completion_api_key', 'completion_auth_key', 'completion_bedrock_credentials']
keys_to_replace_embeddings = ['embeddings_api_key', 'embeddings_auth_key', 'embeddings_bedrock_credentials']
ModelType = Literal['completion', 'embeddings']
      
def replace_dummy_api_key_for_save(new_value, old_value, model_type: ModelType):
  new_val_dict = json.loads(new_value)
  old_val_dict = json.loads(old_value)

  if 'profileSettings' not in new_val_dict:
    return new_value

  for profile, data in new_val_dict['profileSettings'].items():
    if 'values' in data:
      source_id = data.get('source', '')

      keys = api_keys_to_replace(model_type)
      for key in keys:
        if key in data['values']:
          if data['values'][key] == settings.DEFAULT_PASSWORD_PLACEHOLDER and old_val_dict.get('profileSettings'):
            source_profile = source_id if source_id else profile
            data['values'][key] = old_val_dict['profileSettings'][source_profile]['values'][key]
          else:
            data['values'][key] = json.dumps(encrypt_secret(str(data['values'][key])))
          
      if source_id:
        del data['source']

  return json.dumps(new_val_dict)

def replace_ai_api_key_with_placeholder(profiles, model_type: ModelType):
  keys = api_keys_to_replace(model_type)

  for profile, data in profiles.items():
    if 'values' in data:
      for key in keys:
        if key in data['values'] and data['values'][key]:
          data['values'][key] = settings.DEFAULT_PASSWORD_PLACEHOLDER

def replace_all_ai_api_keys_with_placeholder(settings_dict):
    def process_settings(settings_key, placeholder_key):
        if settings_key in settings_dict:
            settings = json.loads(settings_dict[settings_key])
            if 'profileSettings' in settings:
                replace_ai_api_key_with_placeholder(settings['profileSettings'], placeholder_key)
                settings_dict[settings_key] = json.dumps(settings)

    process_settings('ai_completion_settings', 'completion')
    process_settings('ai_embeddings_settings', 'embeddings')

    return settings_dict

def api_keys_to_replace(model_type: ModelType):
  if model_type == 'embeddings':
    return keys_to_replace_embeddings
  elif model_type == 'completion':
    return keys_to_replace_completion
  else:
    LOGGER.warning(f"Invalid model_type '{model_type}' provided to api_keys_to_replace. Returning an empty list.")
    return []