# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Report related CRUD operations
"""

from django.http import HttpResponse, JsonResponse

from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.core.cache import cache
from django.db.models import Sum, Q
from django.db import connection, IntegrityError, transaction
import urllib.parse

import logging

LOGGER = logging.getLogger('arcweb')

import json

import base64

from collections import defaultdict

from dataconn import dataconn
from .views_appgroups import appgroup_starred
from .audit import auditApp
from vizlib.querylog import ActionType
import apps.views_snapshot
from vizlib.accel import viewrec_accel_info, viewrec_checkapps
from vizlib.misc import get_request_source_ip
from .views_sitedata import dateranges_get_data, colorpalette_get_data

from datasets.models import DataSet
from reports.models import Report, ReportUser, ReportUsage, ReportImage, AppGroup, AppGroupAppMembership
from workspace.models import Workspace, get_default_workspace
from migration.migrate import addDashboardSheetData

from django.conf import settings
from django.views.decorators.http import etag
from arcwebbase.etag_caching import etag_metadata_func, _etag_set
from django.views.decorators.cache import cache_control
from django.core.exceptions import PermissionDenied

from rbac.permdefs import PermissionException, PERMDEFS
from reports.dash_template import parameterize_and_save, _publish_template
from .models import ReportTemplate
from rbac.models import RBAC

import datasets.views
from datasets.dashboard_association import DashboardAssociation
import uuid

from datetime import datetime

from reports.utils import _get_report_data
from reports import constants

from util.get_dataset_ids_accessible_by_user import get_dataset_ids_accessible_by_user
from util.get_workspace_ids_accessible_by_user import get_workspace_ids_accessible_by_user

REPORT_TYPES_FILTERS = ["picklist", "daterange"]

# Different default for sqlite -- mysql, etc. can be much higher.
if connection.vendor == 'sqlite':
  QUERY_FILTER_COUNT = getattr(settings, 'QUERY_FILTER_COUNT', 1000)
else:
  QUERY_FILTER_COUNT = getattr(settings, 'QUERY_FILTER_COUNT', 5000)

def _reportuser_to_json(rul, user):
  """Given an array or ReportUser objects, return a json representation"""
  json_data = []
  for ru in rul:
    r = ru.report
    rd = _get_report_data(r, user)
    rd['order'] = ru.order
    rd['username'] = ru.user.username
    json_data.append(rd)
  return HttpResponse(json.dumps(json_data), content_type="application/json")

def can_user_access_dataset(user: User, dataset_id):
  dataset = DataSet.objects.only('id','dataconnection_id','create_user', 'version_group_id').get(id=dataset_id)
  return 'ds_appview' in dataset.getPerms(user)

@login_required
def myapps(request, report_id=None):
  """
  REST endpoint for myapps handling.
  If called with a GET request and no report_id, we return all
    apps for this user
  If called with a GET request and a valid report id, we return the
    object representing this report-user mapping, if it exists. If no object
    exists, we return an empty array.
  If called with POST, we create a new report-user mapping.
  If called with DELETE, we delete any existing report-user mapping.
  For POST and DELETE, the returned data is the same as would be
    returned by a later GET - i.e. for successful deletes, we return
    an empty list, and for successful POST calls, we return the
    object representing the report-user mapping.
  """

  if report_id is None:
    if request.method == 'GET':
      # pylint: disable=E1101
      rul = ReportUser.user_to_reports(request.user.username)
      return _reportuser_to_json(rul, request.user)

    return HttpResponse('report_id must be specified', status=500)

  if request.method not in ['GET', 'POST', 'DELETE']:
    return HttpResponse('invalid request_type', status=500)

  try:
    report_id = int(report_id)
  except Exception:
    return HttpResponse('report_id "%s" is invalid' % report_id, status=500)

  # pylint: disable=E1101
  try:
    r = Report.objects.get(id=report_id)
  except Exception:
    return HttpResponse('report_id "%s" not found' % report_id, status=500)

  # Retrieve the current mapping, if any exists, between this report
  # and user
  rul = ReportUser.report_to_user(report_id,
                                                 request.user.username)

  # If this is a GET request, just return the current mapping
  if request.method == 'GET':
    return _reportuser_to_json(rul, request.user)

  if rul:
    # There is an existing mapping. Delete the mapping if requested,
    # and return
    if request.method == 'DELETE':
      for ru in rul:
        ru.delete()
      rul = []
    return _reportuser_to_json(rul, request.user)

  # There is no existing mapping. If the request is for a DELETE,
  # return now...
  if request.method == 'DELETE':
    return _reportuser_to_json(rul, request.user)

  # ...otherwise, create the mapping and return it
  ru = _save_user_apps(username=request.user.username, report=r)
  rul = [ru]
  return _reportuser_to_json(rul, request.user)

@login_required
def columns(request, table_name=None, no_cache=False, dataconnection_id=None):
  if table_name is None:
    return HttpResponse('No table name specified', status=500)
  no_cache = int(no_cache)
  if dataconnection_id is None:
    dataconnection_id = request.GET.get("dataconnection_id")
  client_ip = get_request_source_ip(request)
  cols = columns_internal(request.user.username, table_name=table_name,
      no_cache=no_cache, dataconnection_id=dataconnection_id,
      client_ip=client_ip)

  return HttpResponse(json.dumps(cols), content_type="application/json")

def columns_internal(username, table_name, no_cache=False,
    dataconnection_id=None, client_ip=None):
  """Return table columns"""

  # Fetch results from the cache, if available
  cache_key = base64.b64encode(("columns.v1.%s.%s" % \
      (dataconnection_id, table_name)).encode())

  if no_cache:
    cache.delete(cache_key)

  val = cache.get(cache_key)
  if val:
    return val

  dc = dataconn.get_dataconn(connection_id=dataconnection_id)
  out = dc.table_columns(table_name, username=username, client_ip=client_ip)

  # Insert the response into the cache, and return it
  cache.add(cache_key, out, timeout=settings.CACHE_RECHECK_SECONDS)

  return out

def check_permission_on_contained_visuals(report: Report, user: User):
  if report.category() == 'dashboard':
    (visual_ids, _) = report.get_visuals_in_dashboard()
    visuals = Report.objects.only('id', 'dataset__id').filter(id__in=visual_ids).select_related('dataset')
    for visual in visuals:
      if visual.dataset:
        visual.dataset.checkPerm(user, 'ds_appedit')

def report_delete(request, report_id):

  if not report_id:
    return HttpResponse('Must specify id to delete', status=500)

  report = Report.objects.filter(id=report_id).first()

  if not report:
    return HttpResponse('Invalid report id', status=500)

  if report.dataset:
    report.dataset.checkPerm(request.user, 'ds_appedit')

  report.workspace.checkPerm(request.user, 'ws_appedit')

  # get dashboard id if deleteing from dashboard
  source_dash = request.GET.get('source_dashboard')
  dash_id = int(source_dash) if source_dash else None

  check_permission_on_contained_visuals(report, request.user)

  try:
    auditApp_delete = lambda r, app_id='': auditApp(r, ActionType.DELETE, app_id=app_id, request=request)
    report.delete(dash_id=dash_id, audit_function = auditApp_delete)
    return HttpResponse(report.report_data, content_type="application/json")
  except PermissionDenied as err:
    # handles attempted delete of linked_viz when still in dashboards
    return HttpResponse(err.message, status=500)

def report_usage(request, report_id=None):
  # TODO: filter by user?
  # TODO: handle report_id
  views_counts = {}
  if not report_id:
    # pylint: disable=E1101
    rul = ReportUsage.objects      \
                 .values('report_id')             \
                 .annotate(cnt=Sum('count'))
    for ru in rul:
      views_counts[ru['report_id']] = ru['cnt']

  val = json.dumps(views_counts)
  return HttpResponse(val, content_type="application/json")

@etag(etag_metadata_func)
@cache_control(max_age=5)
def report_get(request, report_id, filter=None):
  """Get a report"""

  dataset_id = request.GET.get('dataset_id')
  include_all = request.GET.get('include_all') == 'true'
  if dataset_id:
    try:
      dataset_id = int(dataset_id)
    except Exception as e:
      raise Exception('invalid dataset id')

  workspace_id = request.GET.get('workspace_id')
  if workspace_id:
    try:
      workspace_id = int(workspace_id)
    except Exception as e:
      raise Exception('invalid workspace id')

  report_ids = request.GET.get('report_ids')
  if report_ids:
      try:
        report_ids = json.loads(report_ids)
        report_ids = list(map(int, report_ids))
      except Exception as e:
        raise Exception('invalid report id')

  get_assets = request.GET.get('assets', 'false') == 'true'
  appgroup_id = request.GET.get('appgroup', None)
  detail = request.GET.get('detail')
  all = request.GET.get('all')
  get_clone_info = request.GET.get('getcloneinfo',None)
  report_type = request.GET.get('report_type')
  if detail is not None:
    detail = (detail == 'true') or (detail == '1')
  exclude_dashboards = request.GET.get('exclude_dashboards', 'false') == 'true'
  rsp = report_get_internal(request,
                            report_id,
                            report_type=report_type,
                            dataset_id=dataset_id,
                            get_assets=get_assets,
                            detail=detail,
                            all=all,
                            include_all=include_all,
                            filter=filter,
                            appgroup_id=appgroup_id,
                            get_clone_info=get_clone_info,
                            exclude_dashboards=exclude_dashboards,
                            workspace_id=workspace_id,
                            report_ids=report_ids
                            )

  if type(rsp) != HttpResponse:
    rsp = HttpResponse(json.dumps(rsp), content_type="application/json")
  return rsp

# Internal function to get report data...might return HttpResponse in case of error
def report_get_internal(request,
                        report_id,
                        report_type=None,
                        dataset_id=None,
                        get_assets=False,
                        detail=None,
                        all=None,
                        include_all=None,
                        filter=None,
                        appgroup_id=None,
                        get_clone_info=None,
                        exclude_dashboards=False,
                        workspace_id=None,
                        report_ids=None
                        ):

  # If detail is not specified, then we only include report data on fetch of specific report
  if detail is not None:
    include_data = detail
  else:
    include_data = (report_id is not None)

  views_counts = {}
  if not report_id:
    # pylint: disable=E1101
    rul = ReportUsage.objects      \
                 .values('report_id')             \
                 .annotate(cnt=Sum('count'))
    for ru in rul:
      views_counts[ru['report_id']] = ru['cnt']

  user_reports = {}
  favorites_appgroup = appgroup_starred(request)
  favorites = AppGroupAppMembership.objects \
      .filter(appgroup_group__id=favorites_appgroup.id)
  if report_id:
    favorites = favorites.filter(appgroup_app__id=report_id)
  for ag in favorites:
    user_reports[ag.appgroup_app_id] = 1

  ril = None
  if report_id:
    rl = Report.objects.filter(id=report_id) \
                       .select_related('workspace', 'dataset', 'dataset__dataconnection')
  elif report_ids:
    rl = Report.objects.filter(id__in=report_ids) \
      .select_related('workspace', 'dataset', 'dataset__dataconnection')
  else:
    # When all reports are asked for, do not return filter controls or dashboard templates
    # Controls only show up in the context of the dashboards they are in
    #
    # For dataset-specific fetch, we filter out unlinked visuals unless include_all is True
    #
    # For global fetch, when including detail we don't filter out non-shared visuals.  This
    # is for the dashboard edit mode sidebar populate with all visuals.
    if dataset_id:
      if not can_user_access_dataset(request.user, dataset_id):
        # no permissions for the specified dataset:
        rl = Report.objects.none()
      elif all:
        rl = Report.objects \
          .filter(dataset_id=dataset_id) \
          .exclude(report_type__in=REPORT_TYPES_FILTERS)
        if exclude_dashboards:
          rl = rl.exclude(report_type='dashboard')
      else:
        rl = Report.objects \
          .filter(dataset_id=dataset_id) \
          .filter(report_type='dashboard')
    elif all:
      rl = Report.objects.exclude(report_type__in=REPORT_TYPES_FILTERS)

    elif report_type:
      rl = Report.objects.filter(report_type=report_type)
    elif filter:
      rl = Report.objects.filter(**filter)
    elif workspace_id:
      if workspace_id == -4:
        samples_appgroup = AppGroup.objects.get(group_type='samples')
        samples_sub_appgroup = AppGroup.objects.filter(group_parent__id=samples_appgroup.id).select_related('workspace')

        samples_report_ids = list(
          AppGroupAppMembership.objects \
            .filter(appgroup_group__id=samples_appgroup.id) \
            .order_by('appgroup_order') \
            .values_list('appgroup_app', flat=True)
        )

        for sample_sub_appgroup in samples_sub_appgroup:
          samples_report_ids.extend(
            list(
              AppGroupAppMembership.objects \
                .filter(appgroup_group__id=sample_sub_appgroup.id) \
                .order_by('appgroup_order') \
                .values_list('appgroup_app', flat=True)
            )
          )

        rl = Report.objects.filter(id__in=samples_report_ids)
      elif workspace_id == -2:
        favorite_dashboard_ids = list(
          favorites \
            .order_by('appgroup_order') \
            .values_list('appgroup_app', flat=True)
        )

        rl = Report.objects.filter(id__in=favorite_dashboard_ids, report_type='dashboard')
      else:
        rl = Report.objects \
          .filter(workspace_id=workspace_id) \
          .filter(
            Q(report_type='dashboard') | \
            Q(shared_visual_dashboards__isnull=False)
          )
    else:
      rl = Report.objects.filter(Q(report_type='dashboard') | \
                                 Q(shared_visual_dashboards__isnull=False))

    # We are trying to filter the reports as much as possible.
    # When dataset_id or workspace_id is provided they are already heavily filtered.
    # When those are not provided, the next best thing we can do in the case of a non admin user
    # is to filter by permission, for that we need to collect all datasets and workspaces.
    if not request.user.is_superuser and not dataset_id and not workspace_id:
      allowed_dataset_ids = get_dataset_ids_accessible_by_user(request.user)
      if allowed_dataset_ids is not None and len(allowed_dataset_ids) < QUERY_FILTER_COUNT:
        rl = rl.filter(dataset_id__in=set(allowed_dataset_ids))

      allowed_workspace_ids = get_workspace_ids_accessible_by_user(request.user)
      if allowed_workspace_ids is not None and len(allowed_workspace_ids) < QUERY_FILTER_COUNT:
        rl = rl.filter(workspace_id__in=set(allowed_workspace_ids))

    rl = rl.prefetch_related('workspace', 'dataset', 'dataset__dataconnection')

    rl = rl.exclude(report_mode='template')

  if report_id:
    rl = Report.objects.filter(id=report_id) \
                       .select_related('workspace', 'dataset', 'dataset__dataconnection')

  if report_ids:
    rl = Report.objects.filter(id__in=report_ids) \
      .select_related('workspace', 'dataset', 'dataset__dataconnection')

  json_data = []
  r = None

  # Track report images
  cached_ril = None
  if report_id is None:
    idset = set(r.id for r in rl)
    if len(idset) < QUERY_FILTER_COUNT:
      ril = ReportImage.objects.filter(report_id__in=idset).values_list('report__id', 'update_time')
    else:
      ril = ReportImage.objects.values_list('report__id', 'update_time')
    cached_ril = dict((ri[0], ri[1].strftime("%Y.%m.%d.%H.%M.%S")) for ri in ril if ri[0] in idset)

  for r in rl:
    rd = _get_report_data(r, request.user, views_counts=views_counts,
                          user_reports=user_reports, include_data=include_data,
                          cached_ril=cached_ril, get_clone_info=get_clone_info)
    if request.user.is_superuser or ('ds_appview' in rd['perm'] and r.can_view_report(request.user)):
      json_data.append(rd)

  if report_id is not None and r is not None and len(json_data) == 0:
    report_type_simple = r.report_type if r.report_type == 'dashboard' else 'visual'
    return HttpResponse("You do not have permissions to view %s %s" % (report_type_simple, report_id), status=401)

  if report_id is not None and r is not None and \
     r.report_type == 'dashboard' and \
     get_assets:

    # This is where we audit the dashboard loads
    auditApp(r, ActionType.READ, appgroup_id=appgroup_id, request=request)

    # Track all the datasets that this dashboard uses
    dsdict = {}
    dsdict[r.dataset_id] =r.dataset

    moreids = set([])
    moreids.update(rd['visuals'])
    moreids.update(rd['pickers'])
    rl = Report.objects.filter(id__in=moreids) \
                       .prefetch_related('workspace', 'dataset', 'dataset__dataconnection')

    baseurl = settings.ARC_BASEURL
    assets = {}
    for r in rl:
      rd = _get_report_data(r, request.user, views_counts=views_counts,
                            user_reports=user_reports, include_data=include_data,
                            cached_ril=cached_ril, get_clone_info=get_clone_info)
      if 'ds_appview' in rd['perm'] or include_all:
        assets["%s/reports/report/%s" % (baseurl, rd.get('id'))] = [rd]
        if dsdict.get(r.dataset_id) is None:
          dsdict[r.dataset_id] = r.dataset

    for ds in list(dsdict.values()):
      dsinfo = datasets.views.dataset_get_internal(request, ds.id, updatedLV=False, inpds=ds)
      if type(dsinfo) == type([]):
        assets["%s/datasets/dataset/%s?detail=true" % (baseurl, ds.id)] = dsinfo

    assets["%s/reports/colorpalette" % baseurl] = colorpalette_get_data()
    assets["%s/reports/dateranges" % baseurl] = dateranges_get_data();

    json_data[0]['assets'] = assets

  return json_data

def update_dataset(dataset_id, report, request):
  # pylint: disable=E1101
  dataset = DataSet.objects.get(id=dataset_id)

  if not dataset:
    return HttpResponse('Could not find dataset with id %s' % dataset_id, status=500)

  dataset.checkPerm(request.user, 'ds_appedit')

  report_data = json.loads(report.report_data)['report_data']
  dashboard_id = report_data.get('dashboard_id')

  if dashboard_id:
    if report.dataset.id != int(dataset_id):
      DashboardAssociation().remove_dashboard_from_visual_dataset(report.id)
      dataset.add_dashboard_to_dataset(dashboard_id)

  report.dataset = dataset
  report.update_user = request.user.username
  auditApp(report, ActionType.UPDATE, request=request)
  report.save()

def dataset_patch(request):
    patch_data = json.loads(request.body.decode("utf-8"))
    report_ids = patch_data['report_ids']
    try:
        with transaction.atomic():
            # pylint: disable=E1101
            user = request.user
            dataset_id = patch_data['dataset_id']
            for report_id in report_ids:
              report = Report.objects.get(id=int(report_id))
              update_dataset(dataset_id, report, request)
            return HttpResponse(report_ids, content_type="text/plain")
    except IntegrityError as err:
        return HttpResponse("Failed to update reports(reportIds: %s) error: %s" % (report_ids, err.message), status=500)


def report_save(request, report_id, return_https=True):
  """
  Save a report
  POST request can include shared_visual_only property,
  in which case only shared_visual_dashboards will be updated
  """
  dataset_id = request.POST.get('dataset_id', '')
  dataset_id = int(dataset_id)
  try:
    # pylint: disable=E1101
    dataset = DataSet.objects.get(id=dataset_id)
  except Exception:
    return HttpResponse('Invalid dataset id "%s"' % dataset_id, status=500)

  if not dataset:
    return HttpResponse('Could not find dataset with id %s' % dataset_id,
                        status=500)

  dataset.checkPerm(request.user, 'ds_appedit')

  workspace_id = request.POST.get('workspace_id', '')
  if not workspace_id:
    workspace = get_default_workspace(request.user)
  else:
    try:
      # pylint: disable=E1101
      workspace = Workspace.objects.get(id=int(workspace_id))
    except Exception:
      return HttpResponse('Invalid workspace id', status=500)

  workspace.checkPerm(request.user, 'ws_appedit')

  if report_id is None:
    report_id = request.POST.get('report_id', None)

  report_description = request.POST.get('report_description', '')
  report_type = request.POST.get('report_type', None)
  report_data = request.POST.get('report_data', None)
  shared_visual_dashboards = request.POST.get('shared_visual_dashboards', None)
  shared_visual_only = request.POST.get('shared_visual_only', None)
  report_search_text = request.POST.get('report_search_text', None)
  img = request.POST.get('img', None)

  if report_type == 'extension':
    RBAC.checkPerm(request.user, 'sys_styles')

  if not shared_visual_only and (not report_type or not report_data):
    return HttpResponse('Invalid report data "%s", "%s", "%s"' % \
        (report_id, report_type, report_data), status=500)

  try:
    # pylint: disable=E1101
    if report_id:
      report_id = int(report_id)
      r = Report.objects.get(id=report_id)
    else:
      r = Report()
    if not shared_visual_only:
      r.report_description = report_description
      r.report_type = report_type
      r.report_data = report_data
      r.update_user = request.user.username
      # Also update datasets if they don't match
      if not r.dataset or r.dataset.id != int(dataset_id):
        r.dataset = dataset
        DashboardAssociation().remove_dashboard_from_visual_dataset(report_id)

      r.report_search_text = report_search_text
      r.workspace.checkPerm(request.user, 'ws_appedit')
      r.workspace = workspace

    # Currently, only allow linking/unlinking independent of other saved actions
    elif (r.shared_visual_dashboards != shared_visual_dashboards):
      if shared_visual_dashboards:
        # Remove dashboard_id for newly linked visuals
        new_report_data = json.loads(r.report_data)
        if new_report_data['report_data'].get('dashboard_id'):
          del new_report_data['report_data']['dashboard_id']
          r.report_data = json.dumps(new_report_data)

        for dashboard_id in json.loads(shared_visual_dashboards):
          dashboard = Report.objects.filter(id=dashboard_id)
          if dashboard.exists():
            dashboard = dashboard.first()
            dashboard_shared_visuals = json.loads(dashboard.shared_visual_dashboards or '[]')
            if report_id not in dashboard_shared_visuals:
              dashboard_shared_visuals.append(report_id)
              dashboard.shared_visual_dashboards = json.dumps(dashboard_shared_visuals)
              dashboard.save(update_fields=['shared_visual_dashboards'])
      else:
        current_dash_id = int(request.POST.get('dashboard_id'))

        # Update dashboard_id for newly unlinked visual
        new_report_data = json.loads(r.report_data)
        new_report_data['report_data']['dashboard_id'] = current_dash_id
        r.report_data = json.dumps(new_report_data)

        # UI shouldn't allow this
        # If a visual is being unshared, go into each dashboard and remove that visual
        if not shared_visual_dashboards:
          shared_dashboard_list = json.loads(r.shared_visual_dashboards or '[]')
          for dash_id in shared_dashboard_list:
            # Don't remove visual from current dashboard
            if current_dash_id != dash_id:
              curr_dash = Report.objects.filter(id=dash_id)
              if curr_dash.exists():
                curr_dash = curr_dash.first()
                shared_visual_list = json.loads(curr_dash.shared_visual_dashboards or '[]')
                shared_visual_list.remove(report_id)
                curr_dash.shared_visual_dashboards = json.dumps(shared_visual_list)
                dash_report_data = json.loads(curr_dash.report_data)

                new_widgets = []
                for viz in dash_report_data['dashboard_widgets']:
                  viz_id = int(viz['id'].split('-')[-1])
                  if viz_id != report_id:
                    new_widgets.append(viz)

                dash_report_data['dashboard_widgets'] = new_widgets
                curr_dash.report_data = json.dumps(dash_report_data)
                curr_dash.save()

    r.shared_visual_dashboards = shared_visual_dashboards
  except Report.DoesNotExist:
    return HttpResponse("Report %s does not exist" % report_id, status=500)

  auditApp(r, ActionType.UPDATE, request=request)
  r.save()

  # Disable snapshots for filters on save
  if not shared_visual_only and (report_type not in REPORT_TYPES_FILTERS):
    apps.views_snapshot.takeSnapshot(request, r.id)

  if return_https:
      return HttpResponse(r.id, content_type="text/plain")
  else:
      return r


def create_dashboard(dashboard_title, request):
  dataset_id = int(request.POST.get('dataset_id'))
  dataset = DataSet.objects.get(id=dataset_id)
  dataset.checkPerm(request.user, 'ds_appedit')
  report_id = request.POST.get('report_id')
  report_data = { 'report_title': dashboard_title, }
  workspace = get_default_workspace(request.user)
  dashboard = Report.objects.create(report_type='dashboard',
                                  report_name=dashboard_title,
                                  dataset=dataset,
                                  report_data=json.dumps(report_data),
                                  create_user=request.user,
                                  update_user=request.user,
                                  workspace=workspace)
  return dashboard

def add_visuals_to_dashboard(new_visuals, dashboard_title, is_json_response, request):
  # FIRST PART
  # Create or get dashboard
  report_id = request.POST.get('report_id')
  sheet_order = request.POST.get('sheet_order')
  search_text = request.POST.get('search_text')

  if report_id:
    report_id = int(report_id)
    dashboard = Report.objects.get(id=report_id)
    dashboard.workspace.checkPerm(request.user, 'ws_appedit')
    dashboard.dataset.checkPerm(request.user, 'ds_appedit')
    dashboard.update_user = request.user
    #add search text to dashboard
  else:
    dashboard = create_dashboard(dashboard_title, request)
    report_id = dashboard.id

  # SECOND PART
  # Add visuals to dashboard
  new_visual_ids = [visual.id for visual in new_visuals]

  # Add dashboard to the dataset's dahboard
  visuals_datasets_ids = [visual.dataset_id for visual in new_visuals]
  for dataset in DataSet.objects.filter(id__in=visuals_datasets_ids):
    dataset.add_dashboard_to_dataset(report_id)

  dashboard_data = json.loads(dashboard.report_data)
  if not dashboard_data.get('dashboard_widgets'):
    dashboard_data['dashboard_widgets'] = []

  if not dashboard_data.get('global_control_widgets'):
    dashboard_data['global_control_widgets'] = []

  # We are no longer using control_widgets on dashboard level
  # dashboard_data['global_control_widget] used for store app scoped filters
  # dashboard_data['dashboard_sheet']['control_widgets] used for individual sheets filters

  dashboard_widget = dashboard_data['dashboard_widgets']

  # Use default widget size from arc_dashboard.js
  numColumns = 64
  widgetXColumns = numColumns / 2
  widgetYColumns = 20
  start_row = 1
  number_added = 0
  current_sheet = addDashboardSheetData({})['dashboard_sheets'][0]
  if sheet_order:
    sheet_order = int(sheet_order)
    if not dashboard_data.get('dashboard_sheets'):
      dashboard_data = addDashboardSheetData(dashboard_data)

    sheets = dashboard_data.get('dashboard_sheets')
    if sheets:
      current_sheet = sheets[sheet_order - 1]
    else:
      dashboard_data['dashboard_sheets'] = [current_sheet]
      current_sheet['sheet_order'] = 1
  else:
    dashboard_data['dashboard_sheets'] = [current_sheet]
    dashboard_data['numColumns'] = 64

  # Get lowest visual on sheet
  for existing_viz in current_sheet['visual_widgets']:
    if existing_viz['row'] + existing_viz['size_y'] >= start_row:
      start_row = existing_viz['row'] + existing_viz['size_y'] + 1

  # Add newly created visuals to dashboard
  for viz_id in new_visual_ids:
    new_widget = {
      'col': (number_added % 2) * widgetXColumns + 1,
      'row': start_row,
      'size_x': widgetXColumns,
      'size_y': widgetYColumns,
      'id': 'uri-1-widget-%s' % viz_id,
    }
    current_sheet['visual_widgets'].append(new_widget)
    dashboard_widget.append(new_widget)
    if number_added % 2:
      start_row += widgetYColumns

    number_added += 1

  dashboard.report_data = json.dumps(dashboard_data)
  dashboard.save()

  data = {
      'report_id': report_id,
      'new_visual_ids': new_visual_ids,
    }

  if is_json_response:
    return JsonResponse(data)
  else:
    return data

def add_threshold_report(report, request):
  request.POST._mutable = True
  request.POST['dataset_id'] = report.dataset_id
  dashboard_title = request.POST.get('report_title', 'Dashboard containing trigger visual')
  return add_visuals_to_dashboard([report], dashboard_title, False, request)

@login_required
def add_search_report(request):
  """
  REST endpoint for adding search reports
  """
  if request.method != 'POST':
    return HttpResponse('Invalid request method', status=500)

  dashboard_id = request.POST.get('report_id')
  new_saved_visuals = []
  new_visuals = json.loads(request.POST.get('new_reports', '[]'))
  search_text = request.POST.get('search_text')
  dashboard_title = search_text if \
                    search_text else \
                    'Search Dashboard (%s)' % datetime.now().strftime('%Y/%m/%d')

  request.POST._mutable = True
  request.POST['report_id'] = None
  # Save each new visual
  for new_viz in new_visuals:
    request.POST['dataset_id'] = new_viz.get('dataset_id')
    request.POST['report_type'] = new_viz.get('report_type')
    request.POST['report_search_text'] = new_viz.get('report_search_text')
    report_data = json.loads(new_viz['report_data'])
    request.POST['report_data'] = json.dumps(report_data)
    r = report_save(request, None, False)
    if hasattr(r, 'id'):
      new_saved_visuals.append(r)
    else:
      # There's an error response, return directly
      return r

  request.POST['report_id'] = dashboard_id
  return add_visuals_to_dashboard(new_saved_visuals, dashboard_title, True, request)

@login_required
def report(request, report_id=None):
  """
  REST endpoint for reports
  """
  return report_internal(request, report_id)

def report_internal(request, report_id=None, filter=None):
  if request.method == 'DELETE':
    return report_delete(request, report_id)

  if request.method == 'GET':
    return report_get(request, report_id, filter=filter)

  if request.method == 'POST':
    return report_save(request, report_id)

  if request.method == 'PATCH':
    return dataset_patch(request)

  return HttpResponse('Invalid request method', status=500)

def get_report_accel(r):
  """
  Return acceleration info record for a report

    last accel attemp:
      timestamp
      status
    last accel access:
      timestamp
      view used
      rows read
      time taken
    last non-accel access:
      timestamp
      view used
      rows read
      time taken
    other --
      counts --
        number of requests
        served by django cache
        served by accel
        non-accel
  """
  accel = {}

@login_required
def reports_accel(request):
  if request.method != 'POST':
    return HttpResponse('Invalid request method', status=500)

  appids = request.POST.get('appids')
  if appids is None:
    return HttpResponse('Invalid request', status=500)
  appids = set(int(id) for id in json.loads(appids))

  reports = Report.objects.filter(id__in=appids)
  moreids = set([])

  rdtable = {}
  for r in reports:
    rd = _get_report_data(r, request.user,
                          include_data=True)
    if 'ds_appview' not in rd['perm']:
      continue

    if rd['report_type'] == 'dashboard':
      moreids.update(rd['visuals'])
      moreids.update(rd['pickers'])

    rdtable[r.id] = rd

  # Add in the visuals and pickers from any named dashboards
  if moreids.difference(appids):
    reports = Report.objects.filter(id__in=moreids.difference(appids))
    for r in reports:
      rd = _get_report_data(r, request.user,
                            include_data=True)
      if 'ds_appview' not in rd['perm']:
        continue

      rdtable[r.id] = rd

  check = viewrec_checkapps(request.user, list(rdtable.keys()))
  info = viewrec_accel_info(list(rdtable.keys()), days=30)
  for id in list(info.keys()):
    rdtable[id]['accel'] = info[id]

  # This is a little odd going through the recommendation engine
  # but it gives us the ability to show which apps need views
  # refreshed, etc.  Still working on how to best show app
  # status.

  for app_data in list(check['appstat'].values()):
    app_results = defaultdict(lambda: {'successes': [], 'failures': []})
    views = []
    for stat in app_data.get('stats'):
      id = stat.get('id')
      d = {}
      status = stat.get('status')
      d['recommend_status'] = status
      viewname = stat.get('viewname')
      if status == 'ROUTED':
        d['rewrite'] = viewname
      else:
        d['rewrite'] = False
        d['rewrite_status'] = stat.get('reason', 'Unknown')

      if status in ['ROUTED', 'RECOMMENDATION', 'AWAITING_REFRESH']:
        if viewname not in views:
          views.append(viewname)
          app_results[id]['successes'].append(d)
      else:
        app_results[id]['failures'].append(d)

    for app_id in app_results:
      successes = app_results[app_id]['successes']
      if successes:
        info[app_id]['aviews'] = successes
      else:
        info[app_id]['aviews'] = app_results[app_id]['failures']

  rsp = [rdtable[id] for id in sorted(rdtable.keys())]
  return HttpResponse(json.dumps(rsp), content_type="application/json")

@login_required
def clone(request):
  """Clone a visual - POST request accepts dashboard_id and report_id"""
  add_clone_of = request.POST.get('add_clone_of')
  report_id = request.POST.get('report_id')
  dashboard_id = request.POST.get('dashboard_id')
  workspace_id = request.POST.get('workspace_id')
  workspace = None
  if workspace_id:
    new_workspace = Workspace.objects.filter(id=workspace_id)
    if new_workspace.exists():
      new_workspace = new_workspace.first()
      new_workspace.checkPerm(request.user, 'ws_appedit')
      workspace = new_workspace

  report = clone_widget(report_id, dashboard_id, workspace, request.user, True, add_clone_of)
  rd = _get_report_data(report, request.user)
  return HttpResponse(json.dumps([rd]), content_type="application/json")

def clone_widget(report_id, dashboard_id, workspace, user, should_clone_dashboards, add_clone_of=False):
  report = Report.objects.filter(id=report_id)

  if not report.exists():
    return HttpResponse('Invalid report id "%s"' % report_id, status=500)

  report = report.first()
  if should_clone_dashboards or report.report_type != 'dashboard':
    report = clone_report(report, user, add_clone_of=add_clone_of)

  if dashboard_id:
    report_data = json.loads(report.report_data)
    report_data['report_data']['dashboard_id'] = int(dashboard_id)
    report.report_data = json.dumps(report_data)

  if workspace:
    report.workspace = workspace
  report.save()
  return report

@login_required
def create(request, dataset_id=None):
  """Create a new report for a dataset"""
  return create_internal(request, dataset_id)

def create_internal(request, dataset_id=None):

  try:
    # pylint: disable=E1101
    dataset = DataSet.objects.get(id=int(dataset_id))
  except Exception:
    return HttpResponse('Invalid dataset id "%s"' % dataset_id, status=500)

  dataset.checkPerm(request.user, 'ds_appedit')

  workspace_id = request.POST.get('workspace_id', '')
  if not workspace_id:
    workspace = get_default_workspace(request.user)
  else:
    try:
      # pylint: disable=E1101
      workspace = Workspace.objects.get(id=int(workspace_id))
    except Exception:
      return HttpResponse('Invalid workspace id', status=500)

  workspace.checkPerm(request.user, 'ws_appedit')

  dashboard_only_str = request.POST.get('dashboard_only', '')
  if dashboard_only_str.strip().lower() == "true":
    dashboard_only = True
  else:
    dashboard_only = False

  visual_type = request.POST.get('visual_type')
  if not visual_type:
    visual_type = "table"

  if visual_type == 'extension':
    RBAC.checkPerm(request.user, 'sys_styles')

  report_data = {
    "report_data": {
      "report_title": request.POST.get('report_title', "Untitled")
    },
    "report_type": visual_type
  }
  createlimit = request.POST.get('create_limit', '100')
  if createlimit != '':
    report_data["report_data"]["limit"] = createlimit

  r = Report(
      report_name='',
      report_description='',
      dataset=dataset,
      workspace=workspace,
      report_type=visual_type,
      report_data=json.dumps(report_data),
      create_user=request.user.username,
      update_user=request.user.username)

  r.save()
  auditApp(r, ActionType.CREATE, request=request)

  json_data = []
  rd = _get_report_data(r, request.user)
  json_data.append(rd)

  if not dashboard_only:
    # Track in this user's list of reports
    _save_user_apps(username=request.user.username, report=r)

  is_trigger = request.POST.get('is_trigger')
  if is_trigger:
    threshold_report = add_threshold_report(r, request)
    return HttpResponse(json.dumps(threshold_report), content_type="application/json")

  return HttpResponse(json.dumps(json_data), content_type="application/json")

def _save_user_apps(username=None, report=None):
  if not report:
    return None

  if not username:
    return None

  user = User.objects.get(username=username)

  ru = ReportUser(user=user, report=report)
  max_order = ReportUser.max_order(user.username)
  if max_order:
    ru.order = max_order + 1
  else:
    ru.order = 1
  ru.save()
  return ru


def updateVisualReport(report, report_details, user):
  if report_details:
    for report_property in report_details:
      setattr(report, report_property, dict(report_details).get(report_property))
    report.update_user = user.username
  return report


# Only used for single visuals, not dashboards
# Note for not_yet_saved_report_details: when somebody press clone on an modified dashbaord - the original reports's details
# might not yet in the database - we need to save it during clone on the new cloned object
def clone_report(report, user, keep_workspace=False, add_clone_of=False, not_yet_saved_report_details=None):
  # Make compatible with app_clone.py
  if user:
    report.dataset.checkPerm(user, 'ds_appedit')

    if keep_workspace:
      report.workspace.checkPerm(user, 'ws_appedit')
    else:
      report.workspace = get_default_workspace(user)

  if not_yet_saved_report_details:
    # Use the unsaved reportdata instead of the copy if available
    report = updateVisualReport(report, not_yet_saved_report_details, user)

  # Remove the primary key, so the next save will create a new app id
  report.pk = None
  # assign new uuid to satisfy unique constraint
  report.uuid = uuid.uuid4()
  report.imported_uuid = None
  # Don't clone shared visual properties
  report.shared_visual_dashboards = None

  # If add_clone_of is true, add "Clone of" to beginning of title
  # Used when cloning from sidebar - not combined with not_yet_saved_report_details
  if add_clone_of:
        new_report_data = json.loads(report.report_data if not not_yet_saved_report_details else not_yet_saved_report_details.get('report_data'))
        new_report_data['report_data']['report_title'] = 'Clone of {0}'.format(new_report_data['report_data']['report_title'])
        report.report_data = json.dumps(new_report_data)

  report.create_user = user.username
  report.update_user = user.username

  report.save()
  return report

def find_report(report_id, array_of_reports):
    return next(filter(lambda x: int(x['report_id']) == int(report_id), array_of_reports), None)

# Find all control widgets within the 'dashboard_data' json structure passed in,
# and create clones of them. Because control widgets are completely private to a
# app, we do not share them across apps (unlike visuals, which we do share across
# apps)
#
# keep_workspace is True for app_clone util
#
# If we're moving visuals, a move_tracker is passed in and is used to track:
#  - reports_to_move: a dict of successfully moved visuals, includes cloned linked visuals
#  - reports_to_clone: array of objects that store old and new ids for cloned linked visuals
#  - reports_failed: visuals that were not cloned due to permissioning issues
#  Note: we don't clone filters or unlinked visuals when moving dashboard, only if cloning
#
# Note that there is a risk we end up with a few control widgets created, but not
# associated with a dashboard.
#
# TODO: We need to sweep up unused control widgets periodically, and delete them
# from the system

def move_or_clone_dashboard(dashboard_data, user=None, clone_requests=None, unshared_visuals=None, keep_workspace=True, move_tracker=None, not_yet_saved_report_modifications=None):
  clone_requests = clone_requests or [] # shared_visuals
  unshared_visuals = unshared_visuals or []
  parsed_dashboard_data = json.loads(dashboard_data)
  dashboard_sheets = parsed_dashboard_data.get('dashboard_sheets', [])
  not_yet_saved_report_modifications = not_yet_saved_report_modifications or []

  # If we're trying to move workspaces without workspaces provided, nothing to do
  if move_tracker:
    workspace = move_tracker.get('workspace')
    current_dashboard_id = move_tracker.get('report_id')
    if not workspace or not current_dashboard_id:
      return {
        'dashboard_data': dashboard_data,
      }

  # Clone and track the new ids in dashboard_sheets
  new_shared_visuals = []

  # Setting list of control_widgets and dashboard_widgets
  global_control_widgets_all = []
  new_control_widgets_all = []
  new_visual_widgets_all = []
  new_behaviors_all = []
  not_yet_saved_report_modifications_dict = dict((report['report_id'], report) for report in not_yet_saved_report_modifications)
  for control_widget in parsed_dashboard_data.get('global_control_widgets', []):
    not_yet_saved_report_details = not_yet_saved_report_modifications_dict.get(control_widget, None)
    new_r_id = move_or_clone_control_widget(control_widget, move_tracker, user,
                                            keep_workspace, unshared_visuals, not_yet_saved_report_details)
    if new_r_id:
      global_control_widgets_all.append(new_r_id)
      new_control_widgets_all.append(new_r_id)
  for dashboard_sheet in dashboard_sheets:
    new_control_widgets = []
    new_visual_widgets = []
    new_behaviors = []

    # If moving, move every picklist
    # If cloning, clone and add new cloned value to unshared_visuals
    for control_widget in dashboard_sheet.get('control_widgets', []):
      not_yet_saved_report_details = find_report(control_widget, not_yet_saved_report_modifications)
      new_r_id = move_or_clone_control_widget(control_widget, move_tracker, user,
                                              keep_workspace, unshared_visuals, not_yet_saved_report_details)

      # Always add new_r_id regardless of moving/cloning
      if new_r_id:
        new_control_widgets_all.append(new_r_id)
        new_control_widgets.append(new_r_id)

    # If moving, clone requested link visual
    # If cloning, clone every unlinked visual and requested link visual
    for visual_data in dashboard_sheet.get('visual_widgets', []):
      cloned = False
      visual_components = visual_data.get('id').split('-')
      visual_id = int(visual_components.pop())
      visual = Report.objects.filter(id=visual_id)
      if visual.exists():
        visual = visual.first()
        not_yet_saved_report_details = not_yet_saved_report_modifications_dict.get(visual_id, None)
        try:
          # If shared visual, check if user wanted cloned
          if visual.shared_visual_dashboards:
            if visual_id in clone_requests:
              cloned = True

              # If we're moving, need to update old visual to exclude current dashboard id
              if move_tracker:
                new_dashboards = json.loads(visual.shared_visual_dashboards)
                new_dashboards = [dash_id for dash_id in new_dashboards if dash_id != current_dashboard_id]
                visual.shared_visual_dashboards = json.dumps(new_dashboards)
                visual.save()

              visual = clone_report(report=visual, user=user, keep_workspace=keep_workspace, not_yet_saved_report_details=not_yet_saved_report_details)
            else:
              if not_yet_saved_report_details:
                visual = updateVisualReport(visual, not_yet_saved_report_details, user)
                visual.save()
              new_shared_visuals.append(visual.id)
          elif not move_tracker:
            cloned = True
            visual = clone_report(report=visual, user=user, not_yet_saved_report_details=not_yet_saved_report_details)
            unshared_visuals.remove(visual_id)

          # If cloned, mutate shared_visuals/unshared_visuals list
          # If not cloned, and we're moving, check permissions
          if cloned:
            unshared_visuals.append(visual.id)
            visual_data['id'] = '%s-%s' % ('-'.join(visual_components), visual.id)
          elif move_tracker:
            visual.workspace.checkPerm(user, 'ws_appedit')

          # If we're moving, let's move workspaces
          if move_tracker:
            visual.workspace = workspace

            if cloned:
              visual_report_data = json.loads(visual.report_data)
              visual_report_data['report_data']['dashboard_id'] = current_dashboard_id
              visual.report_data = json.dumps(visual_report_data)
              move_tracker['reports_to_clone'].append({
                'old_id': visual_id,
                'new_id': visual.id,
              })

            visual.save()
            move_tracker['reports_to_move'][visual.id] = True

          new_visual_widgets_all.append(visual_data)
          new_visual_widgets.append(visual_data)
        except Exception as e:
          error_message = 'Skipping visual %s because lack permissions' % visual_id
          print(error_message)
          print((str(e)))
          if move_tracker:
            move_tracker['reports_failed'].append('%s (%s)' % (visual_id, error_message))


            # TODO: Fix clone of behaviors
    # for behavior in dashboard_sheet.get('behaviors', []):
      # print 'figure out which id belongs to what'

    dashboard_sheet['control_widgets'] = new_control_widgets
    dashboard_sheet['visual_widgets'] = new_visual_widgets
    # dashboard_sheets['behaviors'] = new_behaviors

  parsed_dashboard_data['dashboard_sheets'] = dashboard_sheets
  parsed_dashboard_data['global_control_widgets'] = global_control_widgets_all
  parsed_dashboard_data['dashboard_widgets'] = new_visual_widgets_all
  parsed_dashboard_data['click_behaviors'] = new_behaviors_all
  return {
    'dashboard_data': json.dumps(parsed_dashboard_data),
    'dashboards_shared_visuals': json.dumps(new_shared_visuals),
  }

def move_or_clone_control_widget(control_widget, move_tracker, user, keep_workspace, unshared_visuals, not_yet_saved_report_details):
  control_widget_appid = int(control_widget)
  # pylint: disable=E1101
  try:
    r = Report.objects.get(id=control_widget_appid)
    if move_tracker:
      workspace = move_tracker.get('workspace')
      new_r_id = control_widget_appid
      r.workspace = workspace
      r.save()
      move_tracker['reports_to_move'][new_r_id] = True
    else:
      new_r = clone_report(report=r, user=user, keep_workspace=keep_workspace, not_yet_saved_report_details=not_yet_saved_report_details)
      new_r_id = int(new_r.id)
      unshared_visuals.append(new_r_id)
  except Exception as e:
    error_message = 'The dashboard filter list is incorrect for filter id  %s' % control_widget_appid
    LOGGER.exception(error_message)
    if move_tracker:
      move_tracker['reports_failed'].append(error_message)
  return new_r_id

@login_required
def clone_widgets(request):
  """
  Clone widgets whose ids are set as part of request array.
  Ignore dashboard type reports if provided.
  """

  request_data = json.loads(request.body)
  original_widgets = request_data.get('original_widgets', [])
  workspace_id = request_data.get('workspace_id')
  dashboard_id = request_data.get('dashboard_id')
  if not workspace_id:
    workspace = get_default_workspace(request.user)
  else:
    try:
      workspace = Workspace.objects.get(id=int(workspace_id))
    except Exception:
      return HttpResponse('Invalid workspace id', status=500)

  workspace.checkPerm(request.user, 'ws_appedit')
  unique_original_widgets = list(set(original_widgets))
  map_original_widgets_to_cloned = {}

  for ow in unique_original_widgets:
    newWidgetId = clone_widget(ow, dashboard_id, workspace, request.user, False).id
    if newWidgetId != None:
      map_original_widgets_to_cloned[ow] = newWidgetId

  return HttpResponse(json.dumps(map_original_widgets_to_cloned), content_type="application/json")

@login_required
def savedashboard(request):
  """ Save a dashboard into the reports database """
  clone_data = json.loads(request.POST.get('clone_data', 'false'))
  dashboard_id = request.POST.get('dashboard_id', '')
  dashboard_name = request.POST.get('dashboard_name', '')
  dashboard_description = request.POST.get('dashboard_description', '')
  dashboard_data = request.POST.get('dashboard_data', '')
  dashboards_shared_visuals = request.POST.get('dashboards_shared_visuals', '[]')
  unshared_visuals = json.loads(request.POST.get('unshared_visuals', '[]'))
  dashboard_dataset_id = request.POST.get('dashboard_dataset_id', '')
  try:
    dataset = DataSet.objects.get(id=int(dashboard_dataset_id))
  except Exception:
    return HttpResponse('Invalid dataset id "%s"' % dashboard_dataset_id,
                        status=500)

  dataset.checkPerm(request.user, 'ds_appedit')

  # TODO - consider create workspace on demand if needed
  dashboard_workspace_id = request.POST.get('dashboard_workspace_id', '')
  if not dashboard_workspace_id or clone_data:
    workspace = get_default_workspace(request.user)
  else:
    try:
      workspace = Workspace.objects.get(id=int(dashboard_workspace_id))
    except Exception:
      return HttpResponse('Invalid workspace id', status=500)

  if not clone_data:
    workspace.checkPerm(request.user, 'ws_appedit')

  # Create a dashboard if id has not been specified
  if dashboard_id and not clone_data:
    # pylint: disable=E1101
    r = Report.objects.get(id=int(dashboard_id))
    r.workspace.checkPerm(request.user, 'ws_appedit')
    r.report_name = dashboard_name.strip()
    r.report_description = dashboard_description
    r.shared_visual_dashboards = dashboards_shared_visuals
    r.report_data = dashboard_data
    r.dataset = dataset
    r.workspace = workspace
    r.update_user = request.user.username
    action = ActionType.UPDATE
  else:
    action = ActionType.CREATE
    if clone_data:
      # We are cloning an existing dashboard. Create copies of all the control widgets
      dashboard_object = move_or_clone_dashboard(dashboard_data=dashboard_data, user=request.user, clone_requests=clone_data.get('shared_visuals_to_clone'), unshared_visuals=unshared_visuals, keep_workspace=False, not_yet_saved_report_modifications=clone_data.get('not_saved_report_data_of_original_dashboard'))
      dashboard_data = dashboard_object.get('dashboard_data')
      dashboards_shared_visuals = dashboard_object.get('dashboards_shared_visuals')

    r = Report(
        report_name=dashboard_name,
        report_description=dashboard_description,
        dataset=dataset,
        workspace=workspace,
        report_type="dashboard",
        report_data=dashboard_data,
        shared_visual_dashboards=dashboards_shared_visuals,
        create_user=request.user.username,
        update_user=request.user.username,
        parent_report_id=clone_data.get('parent_dashboard_id') if clone_data else None)

  r.save()

  # Add this dashboard to shared_visual_dashboards of all it's shared visuals
  dashboard_id = r.id


  if clone_data and constants.ALL_CLONE_REPORT_IDS and constants.ALL_PARENT_REPORT_IDS:
    constants.ALL_CLONE_REPORT_IDS.add(dashboard_id)
    constants.ALL_PARENT_REPORT_IDS.add(r.parent_report_id)

  for shared_visual_id in json.loads(dashboards_shared_visuals):
    visual_instance = Report.objects.filter(id=shared_visual_id)
    if visual_instance.exists():
      visual_instance = visual_instance.first()
      dataset = DataSet.objects.get(id=visual_instance.dataset_id)
      dataset.add_dashboard_to_dataset(dashboard_id)
      shared_visuals = json.loads(visual_instance.shared_visual_dashboards)

      if dashboard_id not in shared_visuals:
        shared_visuals.append(dashboard_id)
        visual_instance.shared_visual_dashboards = json.dumps(shared_visuals)
        visual_instance.save()

  # Add dashboard id to unshared visuals and move to dashboard's existing workspace
  if unshared_visuals:
    for unshared_visual_id in unshared_visuals:
      visual_instance = Report.objects.filter(id=unshared_visual_id)
      if visual_instance.exists():
        visual_instance = visual_instance.first()
        dataset = DataSet.objects.get(id=visual_instance.dataset_id)
        dataset.add_dashboard_to_dataset(dashboard_id)
        report_data = json.loads(visual_instance.report_data)
        report_data['report_data']['dashboard_id'] = dashboard_id
        visual_instance.report_data = json.dumps(report_data)
        visual_instance.workspace = workspace
        visual_instance.save()

  auditApp(r, action, request=request)
  apps.views_snapshot.takeSnapshot(request, r.id)

  # Track in this user's list of reports
  _save_user_apps(username=request.user.username, report=r)

  return_data = {
    'id': r.id,
    'workspace_id': r.workspace.id,
    'workspace_name': r.workspace.name,
    'workspace_perm': list(r.workspace.getPerms(request.user)),
  }

  return HttpResponse(json.dumps(return_data), content_type="application/json")


def _move_report(report_data, workspace, user):
  result = {
    'reports_to_clone': [],
    'reports_to_move': {},
    'reports_failed': [],
    'workspace': workspace
  }

  reports_to_clone = report_data.get('reports_to_clone', [])
  reports_to_move = report_data.get('reports_to_move', [])
  unshared_visuals = report_data.get('unsharedVisuals', [])
  current_report = report_data.get('report_id')
  result['report_id'] = current_report

  is_app = report_data.get('type') == 'apps'

  try:
    if current_report is not None:
      if is_app:
        appgroup_id = int(current_report)
        report = AppGroup.objects.filter(Q(id=appgroup_id) | Q(group_parent__id=appgroup_id))
        if not user.is_superuser and 'sys_misc' not in RBAC.get_user_obj_perms(user):
          report.first().workspace.checkPerm(user, 'ws_appedit')
        report.update(workspace=workspace)

        # Update etag since report.update does not trigger post_save signal
        _etag_set('ETAG_METADATA_CACHE_KEY')
      else:
        report = Report.objects.get(id=int(current_report))

        if not user.is_superuser and 'sys_misc' not in RBAC.get_user_obj_perms(user):
          report.workspace.checkPerm(user, 'ws_appedit')

        report.dataset.checkPerm(user, 'ds_appedit')

        report.workspace = workspace

        fields_to_save = ['workspace']

        # Now if is dashboard, use move_or_clone_dashboard to clone shared visuals
        if report.report_type == 'dashboard':
          dashboard_object = move_or_clone_dashboard(report.report_data, user, reports_to_clone, unshared_visuals, False, result)
          report.report_data = dashboard_object.get('dashboard_data')
          report.shared_visual_dashboards = dashboard_object.get('dashboards_shared_visuals', json.dumps(reports_to_move))
          fields_to_save += ['report_data', 'shared_visual_dashboards']

        report.save(update_fields=fields_to_save)
  except Exception as e:
    result['reports_failed'].append('%s (%s)' % (current_report, str(e)))
    result['error'] = str(e)

  # Move any newly added unlinked visuals if moving from dashboard view/edit page
  for visual_id in (reports_to_move + unshared_visuals):
    if not result['reports_to_move'].get(visual_id):
      single_report_data = {
        'report_id': visual_id
      }

      current_result = _move_report(single_report_data, workspace, user)
      if not isinstance(result, dict):
        move_error = 'Skipping moving of visual %s: ' % visual_id
        if isinstance(current_result, HttpResponse):
          move_error += current_result.content
        else:
          move_error += current_result

        result['reports_failed'].append('%s (%s)' % (visual_id, move_error))

        print(move_error)
      elif len(current_result.get('reports_failed')):
        result['reports_failed'] += current_result.get('reports_failed')
      else:
        result['reports_to_move'][visual_id] = True

  del result['workspace']
  return result

@login_required
def move_workspace(request):
  if request.method == 'POST':
    reports_to_move = json.loads(request.POST.get('dashboards_info', '[]'))
    workspace_id = request.POST.get('current_workspace_id')
    results = []

    if not workspace_id:
      workspace = get_default_workspace(request.user)
    else:
      try:
        workspace = Workspace.objects.get(id=int(workspace_id))
      except Exception:
          return HttpResponse('Invalid workspace id', status=500)

      if not request.user.is_superuser and 'sys_misc' not in RBAC.get_user_obj_perms(request.user):
        try:
          workspace.checkPerm(request.user, 'ws_appedit')
        except Exception:
          return HttpResponse("You don't have the necessary permissions to edit this workspace", status=500)

    for report in reports_to_move:
      result = _move_report(report, workspace, request.user)
      results.append(result)

    return HttpResponse(json.dumps(results), content_type="application/json")
  else:
    return HttpResponse("Invalid request", status=500)

@login_required
def save_template(request, report_id=None):
  try:
    name = request.POST.get('name', '')
    desc = request.POST.get('description', '')
    report_id = int(request.POST.get('report_id'))
    obj_name = request.POST.get('report_type')
    dsreqs = request.POST.get('dsreqs')
    parameterize_and_save(name, desc, report_id, obj_name, dsreqs, request.user)
  except PermissionException as e:
    msg = 'Cannot save the template: ' \
          'You do not have the necessary permission. "%s" role is required' % \
          PERMDEFS[e.perm]
    LOGGER.error(msg)
    return HttpResponse(msg, status=500)
  except Exception as e:
    import traceback
    traceback.print_exc()
    LOGGER.error(str(e))
    return HttpResponse(str(e), status=500)

  return HttpResponse("Template has been saved.", status=200)

@login_required
def get_matched_templates(request, showall=False):
  temps = ReportTemplate.objects.all()
  reports = []
  no_dims = int(request.GET.get('nDims', 0))
  no_aggs = int(request.GET.get('nAggs', 0))
  for t in temps:
    p = json.loads(t.params)
    if (p['no_aggs'] == no_aggs and p['no_dims'] == no_dims) or showall:
      if t.img:
        img = True
      else:
        img = False
      reports.append({"report_id":t.report.id,
                      "name":t.name,
                      "desc":t.description,
                      "params":t.params,
                      "id":t.id,
                      "img":img})
  return HttpResponse(json.dumps({"templates":reports}), content_type="application/json")

@login_required
def convert_to_new_dashboard(request):
  # convert the current view of template to real dashboard by
  #   1. getting url parameters
  #   2. import the template being used to create a new dashboard
  #   3. replace all the parameter of the new imported dashboard with the url
  #      parameters from step 1
  from apps.views import _get_param_context
  report_id = int(request.POST.get('report_id'))
  url_param_dict = urllib.parse.parse_qs(urllib.parse.urlsplit(request.POST['url_params']).query)
  param_dict = {}
  for k, v in url_param_dict.items():
    v = ''.join(v)
    url_param_dict[k] = v
    param_dict[k[len('param.'):]] = v
  c = _get_param_context(request, is_dashboard=True, param_dict=url_param_dict)
  c['params'] = param_dict
  report = _publish_template(c, report_id, request.user)
  return HttpResponse(report['dashboards'][0]['id'])

@login_required
def get_clone_info(request, report_id=None):
  report = Report.objects.filter(id=report_id)
  if not report.exists():
    return HttpResponse('Invalid report id "%s"' % report_id, status=500)
  report = report.first()
  ancestors_chain = Report.get_relatives_chain(report, 'parents')
  immediate_children = Report.get_immediate_children(report)
  children_chain = Report.get_relatives_chain(report, 'children')
  return JsonResponse({
    "ancestorsChain": ancestors_chain,
    "immediateChildrenList": immediate_children,
    "descendantsCount": len(children_chain
    )})
