# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
AppGroup related CRUD operations
"""

from django.http import HttpResponse

from django.contrib.auth.decorators import login_required
from django.db.models import Max, Count, Min

from django.views.decorators.http import etag
from arcwebbase.etag_caching import etag_metadata_func
from rbac.models import RBAC

import json
import threading
from vizlib.querylog import logAction, ComponentType, ActionType
from vizlib.misc import global_lock

from reports.models import AppGroupAppMembership, AppGroup, \
    Report, ReportUser, UserAppGroupMembership

from django.contrib.auth.models import User
from django.db import connection
from django.db.models import Q, Count
from django.conf import settings
from workspace.models import get_default_workspace

from django.db.models.signals import post_delete
from django.dispatch import receiver

def _auditAppGroup(appgroup, action, request=None):
  logAction(action=action, component_type=ComponentType.APPGROUP,
     component_id=appgroup.id, appgroup_name = appgroup.group_name, request=request)

@login_required
def myapps(request, app_id=None):
  """
  REST endpoint for myapps handling.
  If called with POST, we create a new report-user mapping.
  If called with DELETE, we delete any existing report-user mapping.
  """
  if app_id is None:
    # we don't support not having an appid
    raise Exception('Invalid app id')

  if request.method == 'POST':
    app_present = True
  elif request.method == 'DELETE':
    app_present = False
  else:
    raise Exception('Invalid request method')

  app_id = int(app_id)
  a = appgroup_starred(request)

  # Don't favorite if report isn't linked visual or dashboard
  report = Report.objects.filter(id=app_id)
  if not report.exists():
    return HttpResponse('Dashboard does not exist', status=500)
  elif report.first().report_type != 'dashboard':
    return HttpResponse('Only dashboards can be favorited', status=500)

  save_response = single_appgroup_app_update(request, a.id, app_id, app_present)

  if isinstance(save_response, HttpResponse):
    return save_response
  else:
    return HttpResponse('[]', content_type="application/json")

def _get_user_appgroups(username):
  # pylint: disable=E1101
  usergroups = UserAppGroupMembership.objects.filter(
      usergroup_username=username)
  appgroup_ids = {}
  for usergroup in usergroups:
    if usergroup.usergroup_group_id:
      appgroup_ids[usergroup.usergroup_group_id] = usergroup.usergroup_order
  return appgroup_ids

# Get summary information for all groups visible to this user
def _appgroup_get_all(request, group_type=None):
  username = request.user.username
  is_superuser = request.user.is_superuser

  # Ensure that the favorites group exists for this user
  appgroup_starred(request)
  appgroup_ids = _get_user_appgroups(username)
  al = AppGroup.objects \
    .filter(Q(group_username=username) | Q(group_username__isnull=True))

  if group_type is not None:
    al = al.filter(group_type=group_type)

  al = al.select_related('workspace')

  if connection.vendor != 'oracle':
    al = al.annotate(num_apps=Count('appgroupappmembership'))

  apps = AppGroupAppMembership.objects.all()
  data = _appgroups_to_json(request.user, None, al, apps,
      appgroup_ids)

  # If only updatable apps are requested, filter the list
  if not is_superuser:
    rw = request.GET.get("rw", '')
    if rw.lower() == "true":
      data = [x for x in data if x['rw'] == True]

  return HttpResponse(json.dumps(data), content_type="application/json")

def appgroup_create_sample():
  # pylint: disable=E1101
  al = AppGroup.objects \
      .filter(group_type='samples')
  if len(al) > 0:
    return None

  a = AppGroup(
      group_name='Samples',
      group_type='samples',
      create_user='vizapps_admin',
      update_user='vizapps_admin')
  a.save()
  return a.id

# Atomically, create a 'starred' AppGroup instance for the given username
# and return it. Migrate from older style starred apps if necessary.
StarredLockProcesses = global_lock('appgroup_starred')
StarredLockThreads = threading.Lock()

def _appgroup_create_starred(request):
  # pylint: disable=E1101
  username = request.user.username
  with StarredLockThreads:
    with StarredLockProcesses:
      al = AppGroup.objects \
          .filter(group_type='starred') \
          .filter(group_username=username)
      if len(al) == 0:
        a = AppGroup(
            group_name='Starred',
            group_type='starred',
            group_username=username,
            create_user=username,
            update_user=username)
        a.save()

  if len(al) > 0:
    return al

  _auditAppGroup(a, ActionType.CREATE, request=request)

  # Migrate from older reports data
  rul = ReportUser.user_to_reports(username)
  if len(rul) > 0:
    for ru in rul:
      ag = AppGroupAppMembership(
          appgroup_group = a,
          appgroup_app = ru.report,
          appgroup_order = ru.order)
      ag.save()
    rul.delete()

  return AppGroup.objects \
      .filter(group_type='starred') \
      .filter(group_username=username)

# Get from a username to a list of myapps visuals. If necessary, migrate from
# previous representation of myapps in ReportUser to a new representation in
# AppGroup
def appgroup_starred(request):
  # pylint: disable=E1101
  al = AppGroup.objects \
      .order_by() \
      .filter(group_type='starred') \
      .filter(group_username=request.user.username) \
      .select_related('workspace')

  if len(al) == 0:
    # We do not have a 'starred' group for this user. Create one
    al = _appgroup_create_starred(request)

  if len(al) == 0:
    raise Exception('Internal error creating a starred group')

  if len(al) > 1:
    raise Exception(
        'Internal error - multiple starred groups for user %s' % request.user.username)

  return al[0]

def get_appgroup_data(a: AppGroup, user, cached_ril=None):
  """Get a dict representing appgroup data"""

  ad = _appgroups_to_json(user, None, [a], None)[0]
  del ad['group_options']
  ad['id'] = a.id
  ad['updated_by'] = ad['update_user']
  ad['title'] = ad['name']
  ad['url'] = "%s/apps/appgroup/%s" % (settings.ARC_BASEURL, a.id)
  ad['edit_url'] = "%s/apps/appedit/%s" % (settings.ARC_BASEURL, a.id)
  ad['type'] = 'apps'
  ad['report_type'] = "apps"

  if cached_ril and cached_ril.get(a.id):
    ad['img'] = f"{settings.ARC_BASEURL}/apps/snapshot_ag/{a.id}"

  return ad

# Get from a username to a list of myapps visuals. If necessary, migrate from
# previous representation of myapps in ReportUser to a new representation in
# AppGroup
def _appgroup_get_starred(request):
  a = appgroup_starred(request)
  # pylint: disable=E1101
  apps = AppGroupAppMembership.objects \
      .filter(appgroup_group__id=a.id).order_by('appgroup_order')
  data = _appgroups_to_json(request.user, None, [a], apps)
  return HttpResponse(json.dumps(data), content_type="application/json")

# Get a listing of all appgroups, or detail about a particular appgroup
@etag(etag_metadata_func)
def _appgroup_get(request, appgroup_id):
  if appgroup_id is None:
    return _appgroup_get_all(request=request)

  appgroup_id = int(appgroup_id)
  if appgroup_id == -2:
    return _appgroup_get_starred(request)

  # pylint: disable=E1101
  al = AppGroup.objects \
      .filter(id=appgroup_id) \
      .select_related('workspace')

  apps = AppGroupAppMembership.objects \
      .filter(appgroup_group__id=appgroup_id).order_by('appgroup_order')

  appgroup_ids = _get_user_appgroups(request.user.username)
  data = _appgroups_to_json(request.user, appgroup_id, al, apps, appgroup_ids, detail=True)

  if len(al) == 0:
    return HttpResponse(json.dumps(data), content_type="application/json")

  parent_group = al[0]

  a_subgroups = AppGroup.objects \
      .filter(group_parent__id=appgroup_id) \
      .select_related('workspace')

  for a_subgroup in a_subgroups:
    apps = AppGroupAppMembership.objects \
        .filter(appgroup_group__id=a_subgroup.id).order_by('appgroup_order')
    data.extend(_appgroups_to_json(request.user,
        a_subgroup.id, [a_subgroup], apps,
        detail=True))

  return HttpResponse(json.dumps(data), content_type="application/json")

def _appgroups_to_json(user, appgroup_id, al: list[AppGroup], apps,
      appgroup_ids=None, detail=False):
  data = []
  parent_groups = {}
  child_groups = {}
  appgroup_ids = appgroup_ids or {}
  for a in al:
    workspace_perm = list(a.workspace.getPerms(user))
    d = {
      'group_id': a.id,
      'group_type': a.group_type,
      'name': a.group_name,
      'username': a.group_username,
      'subgroup_order': a.group_subgroup_order,
      'group_options': a.group_options,
      'create_user': a.create_user,
      'update_user': a.update_user,
      'created': a.create_time.isoformat(),
      'updated': a.update_time.isoformat(),
      'workspace_name': a.workspace.name,
      'workspace_id': a.workspace.id,
      'workspace_perm': workspace_perm,
      'baseImg': "%s/static/img/icon_app.png?v=6" % settings.ARC_BASEURL,
      'perm': workspace_perm,
      'num_apps': getattr(a, 'num_apps') if hasattr(a, 'num_apps') else a.appgroupappmembership_set.count(),
      'apps': [],
    }

    if d['group_type'] != 'starred' and 'ws_appview' not in d['perm'] and 'sys_misc' not in RBAC.get_user_obj_perms(user):
      continue

    d['rw'] = a.permitted_update(user)

    if a.id in appgroup_ids:
      d['in_userlist'] = "true"
      d['userlist_order'] = appgroup_ids[a.id]
    else:
      d['in_userlist'] = "false"

    if a.group_parent_id:
      d['group_parent_id'] = a.group_parent_id
      if a.group_parent_id in child_groups:
        child_groups[a.group_parent_id].append(d)
      else:
        child_groups[a.group_parent_id] = [d]

      # If we are retrieving a particular appgroup, add it to the final
      # set: no parent group is forthcoming in this request for us to
      # attach as a subgroup
      if appgroup_id is not None:
        data.append(d)
    else:
      data.append(d)
      parent_groups[a.id] = d

    if apps:
      dashboardsList = [(ap.appgroup_app_id, ap.appgroup_order) for ap in apps if ap.appgroup_group_id == a.id]
      dashboardsAppOrderDict = dict(dashboardsList)
      dashboardIDs = [dash[0] for dash in dashboardsList]
      if detail == False:
        d['apps'].extend(sorted(dashboardsList, key=lambda l: l[1] if l[1] is not None else -1))
      else:
        # In order to avoid one query per dashboard and dashboard image, we
        # create a query set which does a join between reports and report image
        # first, and then go through it to get name, type, img
        dashboardInfo = Report.objects.filter(id__in=dashboardIDs) \
                                      .select_related('workspace', 'dataset') \
                                      .prefetch_related('reportimage_set')
        dapps = []
        for eachDashboard in dashboardInfo:
          did = eachDashboard.id
          try:
            if not user.is_superuser and 'sys_misc' not in RBAC.get_user_obj_perms(user):
              eachDashboard.workspace.checkPerm(user, 'ws_appview')
            eachDashboard.dataset.checkPerm(user, 'ds_appview')
            first_image = eachDashboard.reportimage_set.first()
            if first_image:
              img = "%s/apps/snapshot/%s?v=%s" % \
                    (settings.ARC_BASEURL, did, first_image.update_time.strftime("%Y.%m.%d.%H.%M.%S"))
            else:
              img = "%s/static/img/icon_%s.png?v=6" % (settings.ARC_BASEURL, eachDashboard.report_type)
            dapps.append([did, dashboardsAppOrderDict[did], eachDashboard.report_name, img])
          except:
            dapps.append([did, dashboardsAppOrderDict[did]])
        d['apps'].extend(sorted(dapps, key=lambda l: l[1] if l[1] is not None else -1))

  for parent_id, subgroups in child_groups.items():
    if parent_id in parent_groups:
      parent_groups[parent_id]['subgroups'] = subgroups

  return data

# Delete an appgroup
def _appgroup_delete(appgroup_id, request):
  if not appgroup_id:
    raise Exception("DELETE must specify an appgroup id")

  # pylint: disable=E1101
  al = AppGroup.objects.filter(id=appgroup_id)
  for a in al:
    if a.permitted_update(request.user):
      _auditAppGroup(a, ActionType.DELETE, request=request)
      a.delete()
    else:
      raise Exception("User %s is not permitted to delete app %s" % \
          (request.user.username, a.id))
  return HttpResponse('[]', content_type="application/json")

def _group_data_to_appgroup(group_data, user, group_parent=None):

  # Note that for updates, we do not always send in every parameter
  # within 'group_data'. When these parameters are not passed in,
  username = user.username
  group_id = group_data.get('group_id')
  group_name = group_data.get('group_name')
  group_username = group_data.get('group_username')
  group_subgroup_order = group_data.get('group_subgroup_order')
  group_options = group_data.get('group_options')
  workspace = group_data.get('workspace_id')

  if group_username:
    # Only a particular user is allowed to set his favs
    if group_username != username:
      raise Exception("Invalid setting for group_username")
    if group_name:
      raise Exception("Favorites cannot be named")

  if group_id is None or group_id <= 0:
    a = AppGroup(
          group_name=group_name,
          group_type='default',
          group_username=group_username,
          group_subgroup_order=group_subgroup_order,
          group_options=group_options,
          create_user=username,
          update_user=username)
    if group_parent:
      a.group_parent = group_parent

    if workspace:
      a.workspace_id = int(workspace)
    else:
      a.workspace = get_default_workspace(user)

    return a

  # pylint: disable=E1101
  al = AppGroup.objects.filter(id=group_id)
  if len(al) == 0:
    raise Exception('Could not find appgroup (%s) with id (%s)' % (group_name, group_id))

  a = al[0]
  if a.permitted_update(user):
    if group_name is not None:
      a.group_name = group_name
    if group_username is not None:
      a.group_username = group_username
    if group_subgroup_order is not None:
      a.group_subgroup_order = group_subgroup_order
    if group_options is not None:
      a.group_options = group_options
    if workspace:
      a.workspace_id = int(workspace)
    a.update_user = username
    return a

  raise Exception("User %s is not permitted to update app %s" % \
      (username, group_id))


# Update an appgroup
def _appgroup_update(request):
  workspace_id = request.POST.get('workspace_id')
  updated_apps_in_subgroups = request.POST.get('updated_apps_in_subgroups',
                                               None)
  if updated_apps_in_subgroups:
    _appgroup_app_update(request, updated_apps_in_subgroups)

  in_deleted_groups = request.POST.get('deleted_groups', None)
  if in_deleted_groups:
    deleted_groups = json.loads(in_deleted_groups)
  else:
    deleted_groups = []

  data = request.POST.get('data', None)
  if not data:
    raise Exception("Invalid parameters to update appgroups")

  in_appgroupl = json.loads(data)
  appgroups = []
  return_data = {}
  for in_appgroup in in_appgroupl:
    a = _group_data_to_appgroup(in_appgroup, request.user)
    appgroups.append(a);
    if a.id:
      action = ActionType.UPDATE
    else:
      action = ActionType.CREATE
    a.save()
    # Assume 1st appgroup in the list is the main one and the remaining
    # ones are subgroups
    if 'saved_app_id' not in return_data:
      return_data['saved_app_id'] = a.id
    _auditAppGroup(a, action, request=request)

  # Next, create the subgroups
  for idx, in_appgroup in enumerate(in_appgroupl):
    subgroups = in_appgroup.get('subgroups')
    if not subgroups:
      continue
    for subgroup in subgroups:
      subgroup['workspace_id'] = workspace_id
      a = _group_data_to_appgroup(subgroup, request.user, appgroups[idx])
      newsubgroup = False
      if not a.id:
        newsubgroup = True
      a.save()
      if newsubgroup:
        subgroup_apps = subgroup.get('apps_in_subgroup', None)
        if subgroup_apps:
          for eachapp in subgroup_apps:
            single_appgroup_app_update(request,
                                       a.id,
                                       eachapp.get('app_id', None),
                                       eachapp.get('app_present', None),
                                       eachapp.get('order', None))

  # Delete deleted groups
  for groupid in deleted_groups:
    # pylint: disable=E1101
    AppGroup.objects.filter(id=int(groupid)).delete()

  # Setup association of users to appgroup
  for idx, in_appgroup in enumerate(in_appgroupl):
    if in_appgroup.get('in_userlist') == "false":
      # Delete the current association
      # pylint: disable=E1101
      UserAppGroupMembership.objects.filter(
          usergroup_group=appgroups[idx],
          usergroup_username=request.user.username).delete();

    if in_appgroup.get('in_userlist') == "true":
      # Create a new mapping unless already in user's list
      uaobjs = UserAppGroupMembership.objects.filter(
          usergroup_group=appgroups[idx],
          usergroup_username=request.user.username)
      if len(uaobjs) > 0:
        continue

      # Find the right order
      order = int(in_appgroup.get('order', 0))
      if order <= 0:
        order = _next_userappgroup_order(request.user.username)

      ua = UserAppGroupMembership(
        usergroup_group=appgroups[idx],
        usergroup_username=request.user.username,
        usergroup_order=order)
      ua.save()

  return HttpResponse(json.dumps(return_data), content_type="application/json")

@login_required
def appgroup(request, appgroup_id=None):
  """REST endpoint for app group"""

  if request.method == 'GET':
    return _appgroup_get(request, appgroup_id)

  if request.method == 'POST':
    return _appgroup_update(request=request)

  if request.method == 'DELETE':
    return _appgroup_delete(appgroup_id, request=request)

  return HttpResponse('Invalid request method', status=500)

def single_appgroup_app_update(request, group_id, app_id, app_present, order=None,
                               failed_update=None):
  # pylint: disable=E1101

  # If we know app is not dashboard, go ahead and quit
  if failed_update and failed_update.get(app_id):
    return

  appgroup = AppGroup.objects.filter(id=group_id)
  if not appgroup:
    return HttpResponse('Invalid appgroup', status=500)

  user = User.objects.get(username=request.user.username)
  if not appgroup.first().permitted_update(user):
    return HttpResponse("User is not permitted to update appgroup %s" % \
        appgroup.group_name, status=403)

  app = Report.objects.filter(id=app_id)
  if not app:
    return HttpResponse('Invalid app', status=500)

  app = app.first()
  if app.report_type != 'dashboard':
    if failed_update != None:
      failed_update[app.id] = True
      return
    else:
      return HttpResponse('Only dashboards can be added to apps', status=500)

  if not app_present:
    AppGroupAppMembership.objects \
        .filter(appgroup_group__id=group_id) \
        .filter(appgroup_app__id=app_id) \
        .delete()
    return

  if order is not None:
    order = float(order)
  else:
    m = AppGroupAppMembership.objects \
        .filter(appgroup_group__id=group_id).aggregate(Min('appgroup_order'))
    min_order = m['appgroup_order__min']
    if min_order:
      order = min_order / 2
    else:
      order = 1

  AppGroupAppMembership.objects \
      .filter(appgroup_group__id=group_id) \
      .filter(appgroup_app__id=app_id) \
      .delete()

  a = AppGroupAppMembership(
      appgroup_group = appgroup[0],
      appgroup_app=app,
      appgroup_order=order)
  a.save()
  _auditAppGroup(appgroup[0], ActionType.UPDATE, request=request)

# Create or update an appgroup - to - app association
def _appgroup_app_update(request, json_data):
  if not json_data:
    raise Exception("Invalid parameters to update appgroups")

  data = json.loads(json_data)
  failed = {}
  for d in data:
    group_id = d['group_id']
    app_id = d['app_id']
    app_present = d['app_present']
    order = d.get('order', None)
    single_appgroup_app_update(request, group_id, app_id, app_present, order, failed)

  return HttpResponse(json.dumps(failed), content_type="application/json")

def _appgroup_app_get(appgroup_id, app_id):
  appgroup_id = int(appgroup_id)
  app_id = int(app_id)
  if (appgroup_id <= 0) and (app_id <= 0):
    return HttpResponse('Invalid request')
  # pylint: disable=E1101
  if (appgroup_id > 0) and (app_id > 0):
    apps = AppGroupAppMembership.objects \
        .filter(appgroup_group__id=appgroup_id) \
        .filter(appgroup_app__id=app_id)
  elif appgroup_id > 0:
    apps = AppGroupAppMembership.objects.filter(appgroup_group__id=appgroup_id)
  else:
    apps = AppGroupAppMembership.objects.filter(appgroup_app__id=app_id)

  data = []
  for a in apps:
    if a.appgroup_app and a.appgroup_group:
      data.append({
        'app_id': a.appgroup_app.id,
        'group_id': a.appgroup_group.id,
        'order': a.appgroup_order
      })
  return HttpResponse(json.dumps(data), content_type="application/json")

@login_required
def appgroup_app(request, appgroup_id=None, app_id=None):
  if request.method == 'POST':
    json_data = request.POST.get('data', None)
    return _appgroup_app_update(request, json_data)

  if request.method == 'GET':
    return _appgroup_app_get(appgroup_id, app_id)

  return HttpResponse('Invalid request method', status=500)

# For the specified user, return the next appgroup order
def _next_userappgroup_order(username):
  agg = UserAppGroupMembership.objects \
          .filter(usergroup_username=username) \
          .aggregate(Max('usergroup_order'))
  max_order = agg.get('usergroup_order__max', 0)
  if max_order:
    order = max_order + 1
  else:
    order = 1
  return order

# Create or update a user - to - appgroup association
def _user_appgroup_update(request):
  data = json.loads(request.POST.get('data', '[]'))
  for d in data:
    appgroup_id = int(d['groupid'])
    in_userlist = bool(d['in_userlist'])
    order = d.get('order', None)

    # pylint: disable=E1101
    appgroup = AppGroup.objects.filter(id=appgroup_id)
    if not appgroup:
      return HttpResponse('Invalid appgroup', status=500)

    # Delete the current entry
    al = UserAppGroupMembership.objects \
        .filter(usergroup_username=request.user.username) \
        .filter(usergroup_group__id=appgroup_id)
    for a in al:
      a.delete()

    # If in userlist, add in a new entry with the specified or calculated order
    if in_userlist:
      if order is not None:
        order = int(order)
      else:
        order = _next_userappgroup_order(request.user.username)

      a = UserAppGroupMembership(
          usergroup_group = appgroup[0],
          usergroup_username = request.user.username,
          usergroup_order=order)
      a.save()
  return HttpResponse('[]', content_type="application/json")

@login_required
def user_appgroup(request):
  if request.method == 'POST':
    return _user_appgroup_update(request)

  return HttpResponse('Invalid request method', status=500)

# Delete user's favorites
@receiver(post_delete, sender=User)
def workspace_post_user_delete(sender, instance, **kwargs):
  if isinstance(instance, User):
    AppGroup.objects.filter(group_type='starred', group_username=instance.username).delete()