# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
AppSetting related CRUD operations
"""

from django.http import HttpResponse, JsonResponse

from django.contrib.auth.decorators import login_required
from django.core.exceptions import ValidationError

import json
import logging
import re
import time

from reports.constants import FINAL_SETTINGS
from reports.models import Report, AppSetting
from reports.utils import replace_dummy_api_key_for_save, replace_ai_api_key_with_placeholder
from apps.models import UserProfile
from rbac.models import RBAC
from django.core.cache import cache
from django.forms.models import model_to_dict
from django.core.validators import URLValidator
from util.ff_flipper import FeatureFlipper

from django.conf import settings

LOGGER = logging.getLogger("arcweb")

try:
  from jobs.views_encryption import encrypt_secret
  from jobs.views_encryption import decrypt_secret
except Exception as e:
  LOGGER.exception(str(e))

# We cache settings that come from the db - we cannot do this across process
# restarts because we don't know if the db has changed. Handle this by adding
# in a per-process factor the the cache key name that changes each time the
# process restarts
CACHE_PER_PROCESS_KEY = getattr(settings, 'CACHE_PER_PROCESS_KEY', time.time())

KEY = 'site-setting-%s-1.0.{0}_{1}' % (CACHE_PER_PROCESS_KEY)

def unregister():
  setting_update_single('registered', 'False', False)
  setting_update_single('registered_firstname', '', False)
  setting_update_single('registered_lastname', '', False)
  setting_update_single('registered_email', '', False)
  setting_update_single('registered_organization', '', False)
  setting_update_single('registered_role', '', False)
  setting_update_single('registered_userid', '', False)

def register(firstname, lastname, emailaddress, organization, role, userid):
  setting_update_single('registered', 'True', False)
  setting_update_single('registered_firstname', firstname, False)
  setting_update_single('registered_lastname', lastname, False)
  setting_update_single('registered_email', emailaddress, False)
  setting_update_single('registered_organization', organization, False)
  setting_update_single('registered_role', role, False)
  setting_update_single('registered_userid', userid, False)

def registered_userid():
  return setting_get_single('registered_userid')

def is_registered():
  return setting_get_single('registered') == 'True'

def initialize_settings(defaults):
  for (key, val) in defaults.items():
    if setting_get_single(key) is None:
      if type(val) not in [type(tuple()), type(list())]:
        read_admin_only = False
      else:
        read_admin_only = val[1]
        val = val[0]
      setting_update_single(key, val, read_admin_only)

def setting_keyname(setting):
  return 'site-setting-%s.%s' % (CACHE_PER_PROCESS_KEY, setting)

def setting_get_single(setting_key):
  default_value = FINAL_SETTINGS.get(setting_key)
  key = setting_keyname(setting_key)
  cache_value = cache.get(key)
  if cache_value is not None:
    return default_value if cache_value == 'None' else cache_value

  sl = AppSetting.objects.filter(setting_key=setting_key)
  from django.db.utils import OperationalError
  try:
    if len(sl) == 0:
      cache.set(key, 'None')
      return default_value
  except OperationalError:
    cache.set(key, 'None')
    return default_value

  cache.set(key, sl[0].setting_value)
  return sl[0].setting_value

# Return a dict of site global settings provided to all apps
def setting_get_globals():

  res = {
    'site_hide_refresh_button_from_connection_explorer': setting_get_single('site_hide_refresh_button_from_connection_explorer'),
    'site_enable_column_rename': setting_get_single('site_enable_column_rename'),
    'site_enable_app_acceleration':
            'true' if getattr(settings, 'ENABLE_APP_ACCELERATION', False) else 'false',
    'scheduler_site_setting_enabled': setting_get_single('scheduling_enabled') in ['true', None],
  }

  return {k: v if v is not None else '' for k, v in res.items()}

def get_settings(is_superuser, setting_key=None):
  # pylint: disable=E1101

  key = KEY.format(is_superuser, setting_key)
  cache_value = cache.get(key)
  if cache_value is not None:
    return cache_value

  if is_superuser:
    if setting_key is None:
      sl = AppSetting.objects.all()
    else:
      sl = AppSetting.objects.filter(setting_key=setting_key)
  else:
    if setting_key is None:
      sl = AppSetting.objects.filter(setting_read_admin_only=False)
    else:
      sl = AppSetting.objects.filter(setting_key=setting_key) \
          .filter(setting_read_admin_only=False)

  sl = list(sl.values())
  cache.set(key, sl)
  return sl

# Get settings
def setting_get(is_superuser, setting_key=None):
  sl = get_settings(is_superuser, setting_key=setting_key)

  data = []
  secret_setting_list = ['email_host_password', 'mapbox_access_token', 'google_client_id', 'jira_password']
  
  for s in sl:
    setting_key = s['setting_key']
    setting_value = s['setting_value']

    if setting_key == 'email_host_password':
      setting_value = decrypt_secret(**json.loads(setting_value))
    
    if setting_value and (setting_key in secret_setting_list):
      setting_value = settings.DEFAULT_PASSWORD_PLACEHOLDER
    
    if setting_key in ['ai_completion_settings', 'ai_embeddings_settings']:
      try:
        dat = json.loads(setting_value)
      except json.JSONDecodeError as e:
        LOGGER.error(f"Error decoding JSON: {str(e)}")
        dat = {}
       
      if 'profileSettings' in dat:
        replace_ai_api_key_with_placeholder(dat['profileSettings'], 'completion' if setting_key == 'ai_completion_settings' else 'embeddings')
        setting_value = json.dumps(dat)

    data.append({
        'setting_key': setting_key,
        'setting_value': setting_value
    })

  val = json.dumps(data)
  return HttpResponse(val, content_type="application/json")

# Delete settings
def _setting_delete(is_superuser, setting_key):
  if not is_superuser:
    return HttpResponse('Only an admin user can delete settings')

  if not setting_key:
    return HttpResponse('Setting must be specified')

  if is_superuser:
    sl = AppSetting.objects.filter(setting_key=setting_key)
  else:
    sl = AppSetting.objects.filter(setting_key=setting_key)\
      .filter(setting_read_admin_only=False)

  cache.delete(setting_keyname(setting_key))
  sl.delete()
  return HttpResponse('[]', content_type="application/json")


def enable_usage_analytics_callback(request, value):
    from usage.views import enable_usage_analytics
    enable_usage_analytics(request, to_boolean(value))

change_listeners = {
    'enable_usage_analytics': enable_usage_analytics_callback,
}

def call_change_listener(request, key, value):
  # call the change listener callback if it exists for the given setting or flipper
  callback = change_listeners.get(key)
  if callback:
    callback(request, value)

# Update settings
def _setting_update(request, setting_key=None):

  RBAC.checkPerm(request.user, 'sys_settings')

  data = request.body
  settings_n_flippers = json.loads(data)
  settings = settings_n_flippers.get('settings', None)
  flippers = settings_n_flippers.get('flippers', None)

  if not data:
    raise Exception("Invalid parameters to update settings")

  for s in settings:
    key = s.get('setting_key')
    value = s.get('setting_value')
    old_value = setting_get_single(key)
    
    if key == 'ai_completion_settings':
      value = replace_dummy_api_key_for_save(value, old_value, "completion")
      
    if key == 'ai_embeddings_settings':
      value = replace_dummy_api_key_for_save(value, old_value, "embeddings")
      

    if key == 'min_pw_length' and not re.match(r'^\d+$', value):
      raise ValidationError("Minimum password length must be an integer")
    elif key == 'min_pw_length' and int(value) < 5:
      raise ValidationError("Minimum password length must be at least five.")

    if key.find('password') != -1:
      value = json.dumps(encrypt_secret(str(value)))

    # Strip extra spaces from value
    if key in ['email_host', 'email_host_user', ]:
      value = value.strip()

    read_admin_only = s.get('setting_read_admin_only', True)

    try:
      setting_update_single(key, value, read_admin_only)

      call_change_listener(request, key, value)
    except:
      setting_update_single(key, old_value, read_admin_only)
      LOGGER.exception(f'could not set {key} from {str(old_value)} to {str(value)}')


  for f in flippers:
    value = to_boolean(f['value'])
    old_value = FeatureFlipper().get_flipper_value(f['id'])

    try:
      FeatureFlipper().set_flipper(f['id'], value, 'DB')

      call_change_listener(request, f['id'], f['value'])
    except:
      FeatureFlipper().set_flipper(f['id'], old_value, 'DB')
      LOGGER.exception(f'could not set {f["id"]} from {str(old_value)} to {str(value)}')


  return HttpResponse('[]', content_type="application/json")

def to_boolean(val):
  if val.lower() == 'false':
    return False

  return True

def setting_update_single(setting_key, setting_value, setting_read_admin_only):
  if not setting_key:
    raise Exception("Setting must be specified")

  if setting_value is None:
    raise Exception("Setting value must be specified")

  if setting_key == 'doc_link_url' or setting_key == 'support_web_page_link_url':
    try:
      URLValidator()(setting_value)
    except ValidationError as e:
      raise Exception("URL value is not valid.")

  sl = AppSetting.objects.filter(setting_key=setting_key)
  if len(sl):
    s = sl[0]
    s.setting_value = setting_value
    s.setting_read_admin_only = setting_read_admin_only
  else:
    s = AppSetting(setting_key=setting_key, setting_value=setting_value,
        setting_read_admin_only=setting_read_admin_only)
  s.save()
  value = [model_to_dict(s)]

  #update single setting and user combo cache set by get_settings function
  cache.set(KEY.format(True, setting_key), value)
  cache.set(KEY.format(False, setting_key), value)

  #invalidate all settings cache set by get_settings function
  cache.set(KEY.format(True, None), None)
  cache.set(KEY.format(False, None), None)

  #update the single setting and no user comb cache set by setting_get_single function
  cache.set(setting_keyname(setting_key), s.setting_value)

@login_required
def setting(request, setting_key=None):
  """REST endpoint for app group"""

  if request.method == 'GET':
    return setting_get(request.user.is_superuser,
        setting_key)

  if request.method == 'POST':
    return _setting_update(request)

  if request.method == 'DELETE':
    return _setting_delete(request.user.is_superuser, setting_key)

  return HttpResponse('Invalid request method', status=500)

def use_derived_data(is_superuser):
  derived_data_setting = get_settings(is_superuser, "site_enable_derived_data")
  if not derived_data_setting:
    return False

  return str(derived_data_setting[0]['setting_value']) == "true"

@login_required
def user_data(request, setting_key=None):
  # REST endpoint for updating user_data of a user profile
  # Since values in user_data are typically unrelated to each other,
  # this endpoint is built to allow updating only one setting at a time

  # Current list of items stored in user_data:
  # homeapp: app id of homepage app
  # last_viewed: list of last viewed apps for a user
  # new_release_modal_seen: if the user has seen the new release modal
  # incompatible_browser_warning_dismissed: if the user has seen the incompatibly (IE) browser warning

  if request.method == 'GET':
    if setting_key not in UserProfile.PROFILE_KEYS_EDIT + UserProfile.PROFILE_KEYS_VIEW:
      return HttpResponse('Invalid profile key', status=500)
    return user_data_get(request, setting_key)

  if setting_key not in UserProfile.PROFILE_KEYS_EDIT:
    return HttpResponse('Invalid profile key', status=500)

  if setting_key in UserProfile.PROFILE_KEYS_PERM:
    RBAC.checkPerm(request.user, 'sys_editperm')

  if request.method == 'POST':
    return user_data_update(request, setting_key)

  if request.method == 'DELETE':
    return user_data_delete(request, setting_key)

  return HttpResponse('Invalid request method', status=500)

def user_data_get(request, setting_key):
  # Fetches data for a single setting in user_data
  user = request.user
  user_data = {}
  try:
    user_data = json.loads(user.userprofile.user_data)
  except:
    pass

  target_data = {}

  if setting_key is not None and setting_key in user_data:
    target_data[setting_key] = user_data[setting_key]

  return HttpResponse(json.dumps(target_data), content_type="application/json")

def user_data_update(request, setting_key):
  # Updates a single setting in user_data
  user = request.user
  user_profile = user.userprofile

  data = request.POST.get(setting_key, None)
  if not data:
    return HttpResponse("Invalid Data or Missing User Setting", status=400)

  data = json.loads(data)
  response = []

  # Specific homeapp error message handling here
  if setting_key == "homeapp":
    if not data == -1 and not Report.objects.filter(id = data).exists():
      return HttpResponse("Dashboard # " + str(data) + " does not exist", status=404)

  _user_data_update(user_profile, setting_key, data)

  # Specific homeapp response handling here
  if setting_key == "homeapp":
    response = get_app_name(data)

  return HttpResponse(json.dumps(response), content_type="application/json")

def _user_data_update(user_profile, key, value):
  if key == 'last_viewed':
    LOGGER.exception('invalid call. _user_data_update is called with last_viewed key')
    return

  user_data = {}
  try:
    user_data = json.loads(user_profile.user_data)
  except:
    pass

  user_data[key] = value
  user_profile.user_data = json.dumps(user_data)
  user_profile.save()

def user_data_delete(request, setting_key):
  user = request.user
  user_profile = user.userprofile
  user_data = {}
  try:
    user_data = json.loads(user.userprofile.user_data)
  except:
    pass

  response = []
  if setting_key is not None and setting_key in user_data:

    # Specific homeapp deletion handled here
    if setting_key == "homeapp":
      user_data[setting_key] = -1
    else:
      del user_data[setting_key]

    user_profile.user_data = json.dumps(user_data)
    user_profile.save()

    # Specific homeapp response handling here
    if setting_key == "homeapp":
      response = get_app_name(-1)

  return HttpResponse(json.dumps(response), content_type="application/json")

def get_app_name(app_id):
  app_id = int(app_id)
  if app_id == -1:
    return ["N/A", "Data Visualization Welcome Page", "Default Home Page"]
  else:
    app = Report.objects.get(id = app_id)
    app_type = app.report_type
    report_data = json.loads(app.report_data)
    try:
      return [app_id, report_data["report_title"], app_type]
    except:
      return [app_id, report_data["report_data"]["report_title"], app_type]

# Update all user's homepage (only admin can do)
def _default_homepage_update(request, value):
  if not request.user.is_superuser and 'sys_misc' not in RBAC.get_user_obj_perms(request.user):
    return HttpResponse("Only admins can modify this", status=401)
  elif not value == -1 and not Report.objects.filter(id = value).exists():
    return HttpResponse("Dashboard # " + str(value) + " does not exist", status=404)
  else:
    try:
      defaultApp = AppSetting.objects.get(setting_key='default_app')
      defaultApp.setting_value = value
    except:
      defaultApp = AppSetting(setting_key='default_app', setting_value=value)

    defaultApp.save()
    cache.set(setting_keyname('default_app'), value)
    app_data = get_app_name(value)
  
    return HttpResponse(json.dumps(app_data), content_type="application/json")
  
def _set_homepage_override_update(request):
  if not request.user.is_superuser and 'sys_misc' not in RBAC.get_user_obj_perms(request.user):
    return HttpResponse("Only admins can modify this", status=401)
  else:
    disable_override = request.POST.get("disable_override", False) == "true"

    try:
      override = AppSetting.objects.get(setting_key='disable_homepage_override')
      override.setting_value = disable_override
    except:
      override = AppSetting(setting_key='disable_homepage_override', setting_value=disable_override)
  
    override.save()
    cache.set(setting_keyname('disable_homepage_override'), disable_override)

    return JsonResponse({'disable_homepage_override': disable_override})

@login_required
def assign_default_homepage(request):
  """REST endpoint for admin assigning default homepage"""
  if request.method == 'POST':
    app_id = request.POST.get("app_id", -1)
    return _default_homepage_update(request, int(app_id))

  if request.method == 'DELETE':
    return _default_homepage_update(request, -1)

  return HttpResponse('Invalid request method', status=500)

@login_required
def set_homepage_override(request):
  if request.method == 'POST':    
    return _set_homepage_override_update(request)
  
  return HttpResponse('Invalid request method', status=500)
