# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Site-specific object definitions
"""

from django.http import HttpResponse
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.core.serializers.json import DjangoJSONEncoder

import json
import base64
import re
import logging

from reports.models import ColorPalette, DateRanges, CustomCSS, StaticAsset
from vizlib.querylog import logAction, ComponentType, ActionType
from rbac.permcache import RBAC

from django.views.decorators.http import etag
from arcwebbase.etag_caching import etag_asset_func, etag_metadata_func

def _auditStaticAsset(asset, action, request=None):
  logAction(action=action, component_type=ComponentType.STATICASSET,
      component_id=asset.id, request=request)

def _auditCustomStyle(customstyle, action, request=None):
  logAction(action=action, component_type=ComponentType.CUSTOMSTYLE,
      component_id=customstyle.id, request=request)

def _auditCustomColor(customstyle, action, request=None):
  logAction(action=action, component_type=ComponentType.CUSTOMCOLOR,
      component_id=customstyle.id, request=request)

def _auditCustomDaterange(daterange, action, request=None):
  logAction(action=action, component_type=ComponentType.CUSTOMDATERANGE,
      component_id=daterange.id, request=request)

def colorpalette_get_data(cp_id=None):
  if cp_id is None:
    cl = ColorPalette.objects.all()
  else:
    cl = ColorPalette.objects.filter(id=cp_id)
  data = []
  for c in cl:
    data.append({
      'name': c.colorpalette_name,
      'description': c.colorpalette_description,
      'value': c.colorpalette_value,
      'type': c.colorpalette_type,
      'id': c.id
    })
  return data

@login_required
@etag(etag_metadata_func)
def _colorpalette_get(request, cp_id):
  data = colorpalette_get_data(cp_id)
  return HttpResponse(json.dumps(data), content_type="application/json")

def _colorpalette_update(request, cp_id):
  name = request.POST.get('name', None)
  description = request.POST.get('description', None)
  value = request.POST.get('value', None)
  type = request.POST.get('type', None)
  if cp_id is None:
    # Create
    c = ColorPalette(
        colorpalette_name=name,
        colorpalette_description=description,
        colorpalette_value=value,
        colorpalette_type=type)
    c.save()
    _auditCustomColor(c, ActionType.CREATE, request=request)
    return HttpResponse(json.dumps([c.id]), content_type="application/json")

  # Update
  try:
    # pylint: disable=E1101
    cp = ColorPalette.objects.get(id=cp_id)
  except Exception:
    return HttpResponse('Invalid color palette id', status=500)

  cp.colorpalette_name = name
  cp.colorpalette_description = description
  cp.colorpalette_value = value
  cp.colorpalette_type = type
  cp.save()
  _auditCustomColor(cp, ActionType.UPDATE, request=request)
  return HttpResponse(json.dumps([cp.id]), content_type="application/json")

def _colorpalette_delete(request, cp_id):
  try:
    # pylint: disable=E1101
    cp = ColorPalette.objects.get(id=cp_id)
  except Exception:
    return HttpResponse('Invalid color palette', status=500)

  _auditCustomColor(cp, ActionType.DELETE, request=request)
  cp.delete()
  return HttpResponse("[]", content_type="application/json")

def colorpalette(request, cp_id=None):
  if cp_id is not None:
    cp_id = int(cp_id)

  if request.method == 'GET':
    return _colorpalette_get(request, cp_id)

  RBAC.checkPerm(request.user, 'sys_styles')

  if request.method == 'POST':
    return _colorpalette_update(request, cp_id)

  if request.method == 'DELETE':
    return _colorpalette_delete(request, cp_id)

  return HttpResponse('Invalid request method', status=500)

def dateranges_get_data(daterange_id=None):
  if daterange_id is None:
    drl = DateRanges.objects.all()
  else:
    drl = DateRanges.objects.filter(id=daterange_id)
  data = []
  for dr in drl:
    data.append({
      'name': dr.dateranges_name,
      'value': dr.dateranges_value,
      'type': dr.dateranges_type,
      'id': dr.id
    })
  return data

@login_required
@etag(etag_metadata_func)
def _dateranges_get(request, daterange_id=None):
  data = dateranges_get_data(daterange_id=daterange_id)
  return HttpResponse(json.dumps(data), content_type="application/json")

def _dateranges_update(request, daterange_id):
  name = request.POST.get('name', None)
  value = request.POST.get('value', None)
  type = request.POST.get('type', None)
  if daterange_id is None:
    # Create
    dr = DateRanges(
        dateranges_name=name,
        dateranges_value=value,
        dateranges_type=type)
    dr.save()

    _auditCustomDaterange(dr, ActionType.CREATE, request=request)
    return HttpResponse(json.dumps([dr.id]), content_type="application/json")

  # Update
  try:
    # pylint: disable=E1101
    dr = DateRanges.objects.get(id=daterange_id)
  except Exception:
    return HttpResponse('Invalid color palette id', status=500)

  dr.dateranges_name = name
  dr.dateranges_value = value
  dr.dateranges_type = type
  dr.save()
  _auditCustomDaterange(dr, ActionType.UPDATE, request=request)
  return HttpResponse(json.dumps([dr.id]), content_type="application/json")

def _dateranges_delete(request, daterange_id):
  try:
    # pylint: disable=E1101
    dr = DateRanges.objects.get(id=daterange_id)
  except Exception:
    return HttpResponse('Invalid daterange', status=500)

  dr.delete()
  _auditCustomDaterange(dr, ActionType.DELETE, request=request)
  return HttpResponse("[]", content_type="application/json")

def dateranges(request, daterange_id=None):
  if daterange_id is not None:
    daterange_id = int(daterange_id)

  if request.method == 'GET':
    return _dateranges_get(request, daterange_id)

  RBAC.checkPerm(request.user, 'sys_styles')

  if request.method == 'POST':
    return _dateranges_update(request, daterange_id)

  if request.method == 'DELETE':
    return _dateranges_delete(request, daterange_id)

  return HttpResponse('Invalid request method', status=500)

@etag(etag_asset_func)
def _customcss_raw_get(request, css_id):
  try:
    # pylint: disable=E1101
    css = CustomCSS.objects.get(id=int(css_id))
  except Exception:
    return HttpResponse('Invalid css id', status=500)

  if not css:
    return HttpResponse('Could not find custom style with id %s' % css_id,
                        status=500)

  return HttpResponse(css.customcss_value, content_type="text/css")

_VALID_MIMETYPES = set([
  'image/jpeg',
  'image/png',
  'image/gif',
  'image/svg+xml',
  'image/x-icon',
  'text/javascript',
  'application/javascript',
  'application/x-javascript',
  'text/x-json',
  'application/json',
  'application/vnd.geo+json',
  'text/css',
  'application/pdf',
  'application/msword',
  'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
  'application/vnd.ms-excel',
  'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
  'font/otf',
  'font/sfnt',
  'font/ttf',
  'font/woff',
  'font/woff2',
  'video/avi',
  'video/quicktime',
  'video/mpeg'
])

def _remove_scripts_from_contents(contents, asset_identifier, decode_result=False):
  """
  Removes scripts from provided file contents and decode it if decode_result is True.

  Args:
    contents (base64 str): file contents
    decode_result (bool): if True will return result contents as byte str else as base64 str
  """
  script_pattern = b"<script.*?>.*?<\/script>"
  decoded_contents = base64.b64decode(contents)
  noscript_decoded_contents = re.sub(script_pattern, b'', decoded_contents, 0, re.DOTALL)

  if decoded_contents != noscript_decoded_contents:
    LOGGER = logging.getLogger('arcweb')
    warning = f'Stripped scripts from contents of the static asset {asset_identifier}:'
    removed_texts = re.findall(script_pattern, decoded_contents, flags=re.DOTALL)
    for text in removed_texts:
      warning += f'\n{text}'
    LOGGER.warning(warning)

  if decode_result:
    return noscript_decoded_contents

  return base64.b64encode(noscript_decoded_contents).decode('ascii')

@login_required
@etag(etag_asset_func)
def staticasset_raw(request, asset_id=None):
  if asset_id is None:
    return HttpResponse('Unspecified asset', status=500)

  asset_id = int(asset_id)
  try:
    # pylint: disable=E1101
    asset = StaticAsset.objects.get(id=asset_id)
  except Exception:
    return HttpResponse('Invalid asset id', status=500)

  mimetype = asset.staticasset_mimetype
  if mimetype not in _VALID_MIMETYPES:
    return HttpResponse('Invalid asset id', status=500)

  return HttpResponse(_remove_scripts_from_contents(asset.staticasset_contents, asset_id, True),
      content_type=asset.staticasset_mimetype)

@login_required
def staticasset(request, asset_id=None):
  if asset_id is not None:
    asset_id = int(asset_id)

  if request.method == 'GET':
    return _staticasset_get(request, asset_id)

  RBAC.checkPerm(request.user, 'sys_styles')

  if request.method == 'POST':
    return _staticasset_update(request, asset_id)

  if request.method == 'DELETE':
    return _staticasset_delete(request, asset_id)

  return HttpResponse('Invalid request method', status=500)

def _staticasset_to_data(asset):
  return {
    'id': asset.id,
    'mimetype': asset.staticasset_mimetype,
    'name': asset.staticasset_name,
    'description': asset.staticasset_description,
    'size': asset.staticasset_size,
    'apps': asset.staticasset_apps,
    'update_user': asset.update_user,
  }

@etag(etag_asset_func)
def _staticasset_get(request, asset_id=None):
  if asset_id is None:
    assetl = StaticAsset.objects.all()
    data = []
    for asset in assetl:
      data.append(_staticasset_to_data(asset))

    return HttpResponse(json.dumps(data), content_type="application/json")

  try:
    # pylint: disable=E1101
    asset = StaticAsset.objects.get(id=asset_id)
  except Exception:
    return HttpResponse('Invalid asset id', status=500)

  data = []
  data.append(_staticasset_to_data(asset))
  return HttpResponse(json.dumps(data), content_type="application/json")

def _staticasset_update(request, asset_id):
  name = request.POST.get('name', None)
  description = request.POST.get('description', None)
  mimetype = request.POST.get('mimetype', None)
  contents = request.POST.get('contents', None)

  if contents and mimetype not in _VALID_MIMETYPES:
    return HttpResponse(f'Invalid mime type for asset: {mimetype}', status=500)

  if asset_id is None:
    # Create
    asset = StaticAsset(
        staticasset_mimetype = mimetype,
        staticasset_name = name,
        staticasset_description = description,
        create_user = request.user.username,
        update_user = request.user.username)

    if contents:
      contents = _remove_scripts_from_contents(contents, name)
      asset.staticasset_contents = contents
      asset.staticasset_size = len(contents)
      limit = getattr(settings, 'DATA_UPLOAD_MAX_MEMORY_SIZE')
      error_message = f"To modify the current limit({limit} bytes) set the desired DATA_UPLOAD_MAX_MEMORY_SIZE value in the Advanced Setting or contact your system administrator!"
      if asset.staticasset_size > getattr(settings, 'DATA_UPLOAD_MAX_MEMORY_SIZE'):
        return HttpResponse(f"Asset too large: {error_message}", status=500)

    asset.save()
    _auditStaticAsset(asset, ActionType.CREATE, request=request)
    return HttpResponse(json.dumps([asset.id]), content_type="application/json")

  # Update
  try:
    # pylint: disable=E1101
    asset = StaticAsset.objects.get(id=asset_id)
  except Exception:
    return HttpResponse('Invalid asset id', status=500)

  asset.staticasset_name = name
  asset.staticasset_description = description
  asset.update_user = request.user.username

  if contents:
    contents = _remove_scripts_from_contents(contents, asset_id)
    asset.staticasset_contents = contents
    asset.staticasset_size = len(contents)
    asset.staticasset_mimetype = mimetype

  asset.save()
  _auditStaticAsset(asset, ActionType.UPDATE, request=request)
  return HttpResponse(json.dumps([asset.id]), content_type="application/json")

def _staticasset_delete(request, asset_id):
  asset_id = int(asset_id)

  try:
    # pylint: disable=E1101
    asset = StaticAsset.objects.get(id=asset_id)
  except Exception:
    return HttpResponse('Invalid asset id', status=500)

  _auditStaticAsset(asset, ActionType.DELETE, request=request)
  asset.delete()
  return HttpResponse("[]", content_type="application/json")

@etag(etag_asset_func)
def _customcss_get(request, css_id=None):
  if css_id is None:
    cssl = CustomCSS.objects.all()
    data = []
    for css in cssl:
      data.append({
        'id': css.id,
        'name': css.customcss_name,
        'description': css.customcss_description,
        'css': css.customcss_value,
        'js': css.customcss_js,
        'author': css.create_user,
        'last_editor': css.update_user,
        'last_edited': css.update_time.isoformat()
      })
    return HttpResponse(json.dumps(data), content_type="application/json")

  try:
    # pylint: disable=E1101
    css = CustomCSS.objects.get(id=css_id)
  except Exception:
    return HttpResponse('Invalid css id', status=500)

  if not css:
    return HttpResponse('Could not find custom style with id %s' % css_id,
                        status=500)

  return HttpResponse(json.dumps({
    'id': css.id,
    'name': css.customcss_name,
    'description': css.customcss_description,
    'css': css.customcss_value,
    'js': css.customcss_js
  }), content_type="application/json")

def _customcss_delete(request, css_id):
  try:
    # pylint: disable=E1101
    css = CustomCSS.objects.get(id=css_id)
  except Exception:
    return HttpResponse('Invalid css id', status=500)

  _auditCustomStyle(css, ActionType.DELETE, request=request)
  css.delete()
  return HttpResponse("[]", content_type="application/json")

def _customcss_update(request, css_id):
  name = request.POST.get('name', None)
  value = request.POST.get('css', None)
  description = request.POST.get('description', None)
  js = request.POST.get('js', None)
  if css_id is None:
    # Create
    c = CustomCSS(
        customcss_name=name,
        customcss_value=value,
        customcss_description=description,
        customcss_js=js,
        create_user=request.user.username,
        update_user=request.user.username)
    c.save()
    _auditCustomStyle(c, ActionType.CREATE, request=request)
    return HttpResponse(json.dumps([c.id]), content_type="application/json")

  # Update
  try:
    # pylint: disable=E1101
    css = CustomCSS.objects.get(id=css_id)
  except Exception:
    return HttpResponse('Invalid css id', status=500)

  css.customcss_name = name
  css.customcss_value = value
  css.customcss_description = description
  css.customcss_js = js
  css.update_user = request.user.username
  css.save()
  _auditCustomStyle(css, ActionType.UPDATE, request=request)
  return HttpResponse(json.dumps([css.id]), content_type="application/json")

@login_required
def customcss(request, css_id=None):
  if css_id is not None:
    css_id = int(css_id)

  if request.method == 'GET':
    return _customcss_get(request, css_id)

  RBAC.checkPerm(request.user, 'sys_styles')

  if request.method == 'POST':
    return _customcss_update(request, css_id)

  if request.method == 'DELETE':
    return _customcss_delete(request, css_id)

  return HttpResponse('Invalid request method', status=500)

@login_required
def customcss_raw(request, css_id=None):
  if css_id is None:
    return HttpResponse("Invalid css_id", status=500)

  return _customcss_raw_get(request, css_id)

@etag(etag_asset_func)
def _customjs_raw_get(request, js_id):
  try:
    # pylint: disable=E1101
    js = CustomCSS.objects.get(id=int(js_id))
  except Exception:
    return HttpResponse('Invalid js id', status=500)

  if not js:
    return HttpResponse('Could not find custom style with id %s' % js_id,
                        status=500)

  return HttpResponse(js.customcss_js, content_type="application/javascript")

@login_required
def customjs_raw(request, js_id=None):
  if js_id is None:
    return HttpResponse("Invalid js id", status=500)

  return _customjs_raw_get(request, js_id)
