#!/usr/bin/python -tt
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import sys
import json
import itertools

SQLITE_DB = "samples.sqlite"

import sqlite3

class SqliteTable(object):
  def __init__(self, tablename):
    self.tablename = tablename
    self.conn = sqlite3.connect(SQLITE_DB)

  def _run_cmd(self, cmd):
    cursor = self.conn.cursor()
    try:
      cursor.execute(cmd)
    except Exception as e:
      print(cmd)
      raise
    self.conn.commit()

  def create(self, columns):
    self.column_names = [x[0] for x in columns]

    nameset = {}
    for idx, c in enumerate(self.column_names):
      if not c in nameset:
        nameset[c] = idx
        continue
      self.column_names[idx] = c + "_" + str(idx)

    self.column_types = [x[1].upper() for x in columns]

    self.columns = [(x, y) for x, y in zip(self.column_names, self.column_types)]

    creates = ["CREATE TABLE %s (" % self.tablename]
    creates.append(", ".join(["%s %s" % (x[0], x[1]) for x in self.columns]))
    creates.append(")")
    self.create_cmd = ''.join(creates)

    self._run_cmd(self.create_cmd)

  def _insert_display(self, coldata, coltype):
    if coltype == "STRING" or coltype == "TIMESTAMP":
      return "'%s'" % coldata.replace("'", "''")
    return coldata

  def insert(self, rows):
    # sqlite can only insert 500 rows at a time
    while len(rows):
      rows_current_batch = rows[:500]
      rows = rows[500:]

      inserts = ["INSERT INTO %s VALUES " % self.tablename]
      for row in rows_current_batch:
        inserts.append("(")
        display_row = [self._insert_display(coldata, coltype) \
            for coldata, coltype in zip(row, self.column_types)]
        inserts.append(",".join(display_row))
        inserts.append(")")
        inserts.append(",")
      self.insert_cmd = ''.join(inserts[:-1])
      self._run_cmd(self.insert_cmd)

def load_file(filename):
  with open(filename) as f:
    data = f.read()

  tablename = None
  l, sep, r = filename.partition(".")
  if sep:
    tablename = l
  else:
    tablename = filename

  t = SqliteTable(tablename)

  j = json.loads(data.decode("ISO-8859-1"))
  if 'headers' in j:
    headers = j['headers']
    columns = []
    for c in headers:
      l, sep, r = c.rpartition(" ")
      columns.append((l, r))
  else:
    colnames = j['colnames']
    coltypes = j['coltypes']
    for i, colname in enumerate(colnames):
      columns.append((colname, coltypes[i]))
  rows = j['rows']
  t.create(columns)
  t.insert(rows)

def main():
  for filename in sys.argv[1:]:
    load_file(filename)

if __name__ == "__main__":
  main()

