# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Models for searches
"""

from django.db import models, transaction

from datasets.models import DataSet
from django.contrib.auth.models import User
from threading import Lock

class Search(models.Model):
  search_text = models.TextField(max_length=1000)

  # dataset this search was performed on
  datasets = models.ManyToManyField(DataSet, blank=True)

  # Once there's rating, should store data for that rating
  search_data = models.TextField(default='{}')

  user_rating = models.IntegerField(null=True)

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True)
  create_user = models.ForeignKey(User, on_delete=models.CASCADE, related_name='created_search')

  # Update time
  update_time = models.DateTimeField(auto_now=True)
  update_user = models.ForeignKey(User, on_delete=models.CASCADE, related_name='updated_search')

  # keep track of search history in memory. This
  # information will be written into the arcviz database periodically
  # by bulk update thread.
  bulk_search_history = {}
  lock = Lock()

  @classmethod
  def update_search_history(cls, create_user, search_text, dataset):
    with cls.lock:
      search_key = f'{create_user}{search_text}'
      if search_key not in cls.bulk_search_history:
        cls.bulk_search_history[search_key] = {
          'create_user': create_user,
          'update_user': create_user,
          'search_text': search_text,
          'datasets': [dataset] if dataset else [],
        }
      elif dataset:
        cls.bulk_search_history[search_key]['datasets'].append(dataset)

  @classmethod
  def commit_changes(cls):
    if cls.bulk_search_history == {}:
      return

    cls._commit()

  @classmethod
  @transaction.atomic
  def _commit(cls):
    with cls.lock:
      bulk_search_history = cls.bulk_search_history
      cls.bulk_search_history = {}

    for _key, search_values in bulk_search_history.items():
      searchUser = Search.objects.filter(create_user=search_values['create_user'])
      search = searchUser.filter(search_text=search_values['search_text'])
      if search.exists():
        search = search.first()
        search.save()
      else:
        search = Search.objects.create(
            search_text=search_values['search_text'],
            create_user=search_values['create_user'],
            update_user=search_values['update_user'],
          )

      dataset = search_values['datasets']
      if dataset:
        search.datasets.add(*dataset)

    # Delete old search if there's too many
    if searchUser.count() > 2000:
      searchUser.order_by('update_time')[:500].delete()

class Bookmark(models.Model):
  # Links to Search model
  search_text = models.TextField(max_length=1000)
  dataset = models.ForeignKey(DataSet, on_delete=models.CASCADE)
  bookmark_data = models.TextField(default='{}')

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True)
  create_user = models.ForeignKey(User, on_delete=models.CASCADE, related_name='created_bookmark')

  # Update time
  update_time = models.DateTimeField(auto_now=True)
  update_user = models.ForeignKey(User, on_delete=models.CASCADE, related_name='updated_bookmark')

# Storing frequency of used words
class SearchElement(models.Model):
  search_type = models.CharField(db_index=True, max_length=200)
  search_word = models.TextField(max_length=1000)

  bookmarks = models.ManyToManyField(Bookmark, blank=True)
  searches = models.ManyToManyField(Search, blank=True)
  datasets = models.ManyToManyField(DataSet)

  user = models.ForeignKey(User, on_delete=models.CASCADE)
  create_time = models.DateTimeField(auto_now_add=True)
  update_time = models.DateTimeField(auto_now=True)
  used_number = models.IntegerField(default=1)
