# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

# Search-related urls

from django.http import HttpResponse

from django.contrib.auth.decorators import login_required
from django.conf import settings

from vizlib.querylog import logAction, ComponentType, ActionType

from .models import Search, Bookmark
from datasets.models import DataSet
from arcwebbase.processors import feature_flipper

import json
import logging

LOGGER = logging.getLogger("arcweb")

def _auditSearch(action, search_text, request=None):
  logAction(action=action, component_type=ComponentType.SEARCH,
      request=request, query=search_text)

@login_required
def search(request):
  """
  REST endpoint for role access
  """

  flippers = feature_flipper(request)
  if not flippers['flippers']['site_enable_search']:
    return HttpResponse('Invalid request', status=500)

  if request.method == 'GET':
    return _search_get(request)

  if request.method == 'POST':
    return _search_post(request)

  if request.method == 'DELETE':
    return _bulk_delete(request)

  return HttpResponse('Invalid request method', status=500)

def _bulk_delete(request):
  bookmark_ids = request.GET.get('bookmark_ids','').split(',')
  user_id = request.user.id
  bookmarks = Bookmark.objects.filter(id__in=bookmark_ids, create_user_id=user_id)
  bookmark_ids = list(bookmarks.values('id'))
  bookmarks.delete()
  return HttpResponse(json.dumps(bookmark_ids), content_type="application/json")

def _search_write(bookmark_id, user, search_data, dataset_id, favorite, search_text=''):
  search_text = search_text.strip()
  if not search_text:
    return {}

  dataset = None
  if dataset_id:
    try:
      dataset = DataSet.objects.get(id=dataset_id)
    except DataSet.DoesNotExist:
      pass

  LOGGER.debug('Bookmark: %s, Favorite: %s' % (bookmark_id, favorite))
  if bookmark_id and not favorite:
    # remove bookmark
    try:
      Bookmark.objects.get(id=bookmark_id, create_user=user).delete()
    except Bookmark.DoesNotExist:
      LOGGER.info('No bookmark %s' % (bookmark_id))
      pass

    return {
      'bookmark_id': bookmark_id,
    }

  Search.update_search_history(create_user=user, search_text=search_text, dataset=dataset)

  bookmark = None
  if favorite and dataset:
    bookmark = Bookmark.objects.filter(search_text=search_text, bookmark_data=search_data, create_user=user, dataset=dataset)
    if bookmark.exists():
      bookmark = bookmark.first()
      bookmark.save()
    else:
      bookmark = Bookmark.objects.create(search_text=search_text, bookmark_data=search_data, create_user=user, update_user=user, dataset=dataset)

  return {
    'bookmark_id': bookmark and bookmark.id,
  }

def _search_iter(request, favorite):
  create_user = request.user
  username = create_user.username
  if favorite:
    bookmarks = Bookmark.objects.filter(create_user=create_user, dataset__search_enabled=True)\
                                .order_by('-update_time')\
                                .select_related('dataset')
    perm_checked_dataset = set()
    has_view_permission = set()
    for bookmark in bookmarks:
      b_dataset = bookmark.dataset
      b_dataset_id = b_dataset.id
      if b_dataset_id in perm_checked_dataset:
        can_view = b_dataset_id in has_view_permission
      else:
        perm_checked_dataset.add(b_dataset.id)
        can_view = 'ds_appview' in b_dataset.getPerms(request.user)
        if can_view:
          has_view_permission.add(b_dataset_id)

      if can_view:
        data = {
          'username': username,
          'bookmark_id': bookmark.id,
          'search_text': bookmark.search_text,
          'search_data': bookmark.bookmark_data,
          'dataset_id': b_dataset.id,
          'dataset_name': b_dataset.dataset_name,
          'favorite': 'true',
          'search_time': bookmark.update_time.isoformat()
        }

        yield data

    return

  searches = Search.objects.filter(create_user=create_user)\
                           .order_by('-update_time')[:1000]\
                           .prefetch_related('datasets')
  for search in searches:
    data = {
      'username': username,
      'search_text': search.search_text,
      'search_time': search.update_time.isoformat()
    }

    if search.datasets:
      data['dataset_ids'] = list(search.datasets.values_list('id', flat=True))

    yield data

def _search_get(request):
  searches = []
  search_text = request.GET.get('search_text', '')
  favorite = request.GET.get('favorite', False)
  favorite = (str(favorite).lower() == "true")

  _auditSearch(ActionType.READ, search_text, request=request)

  for s in _search_iter(request, favorite):
    searches.append(s)

  return HttpResponse(json.dumps(searches), content_type="application/json")

def _search_post(request):
  bookmark_id = request.POST.get('bookmark_id', None)
  search_text = request.POST.get('search_text', None)
  search_data = request.POST.get('search_data', None)
  if not search_text:
    return HttpResponse('[]', content_type="application/json")

  dataset_id = request.POST.get('dataset_id', None)
  favorite = request.POST.get('favorite', False)
  favorite = (str(favorite).lower() == "true")

  _auditSearch(ActionType.UPDATE, search_text, request=request)
  results = _search_write(bookmark_id, request.user, search_data, dataset_id, favorite, search_text)
  return HttpResponse(json.dumps(results), content_type="application/json")

