# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

# Search-related urls

from django.db.models.query import QuerySet
from util.string_operations import extract_tablenames_from_sqlquery
from django.http import HttpResponse, QueryDict

from django.contrib.auth.decorators import login_required
from django.conf import settings

from vizlib.querylog import logAction, ComponentType, ActionType

from .models import Query, Bookmark
from datasets.models import DataConnection

import json
import logging

LOGGER = logging.getLogger("arcweb")

def _auditQuery(action, query, request=None):
  logAction(action=action, component_type=ComponentType.RAW_QUERY,
      request=request, query=query)

@login_required
def query(request):
  """
  REST endpoint for query access
  """
  if not settings.ENABLE_QUERIES:
    return HttpResponse('Invalid request', status=500)

  if request.method == 'GET':
    return _query_get(request)

  if request.method == 'POST':
    return _query_post(request)

  if request.method == 'DELETE':
    return _query_delete(request)

  return HttpResponse('Invalid request method %s' % request.method, status=500)

def _favorite_delete(bookmark_id, username, query, dataconnection_id):
  query = query or ''
  query = query.strip()
  if not query:
    return

  dataconnection = None
  if dataconnection_id:
    try:
      dataconnection = DataConnection.objects.get(id=dataconnection_id)
    except DataConnection.DoesNotExist:
      pass

  try:
    bookmarks = Bookmark.objects.filter(create_user=username, dataconnection=dataconnection, query=query)
    bookmarks.delete()
  except Exception:
    pass

def _query_write(bookmark_id, username, query, dataconnection_id, favorite):
  query = query or ''
  query = query.strip()
  if not query:
    return

  dataconnection = None
  if dataconnection_id:
    try:
      dataconnection = DataConnection.objects.get(id=dataconnection_id)
    except DataConnection.DoesNotExist:
      pass

  LOGGER.debug('Bookmark: %s, Favorite: %s' % (bookmark_id, favorite))
  if bookmark_id and not favorite:
    # remove bookmark
    try:
      Bookmark.objects.get(id=bookmark_id).delete()
    except Bookmark.DoesNotExist:
      LOGGER.info('No bookmark %s' % (bookmark_id))
      pass

    return

  Query(query=query, create_user=username, dataconnection=dataconnection).save()

  if favorite and dataconnection:
    bookmarks = Bookmark.objects.filter(create_user=username, dataconnection=dataconnection, query=query)
    if bookmarks:
      bookmarks[0].save()
      return

    Bookmark(query=query, create_user=username, dataconnection=dataconnection).save()

def _query_iter(request, query, dataconnection_id, favorite):
  dc = None
  if dataconnection_id:
    dl = DataConnection.objects.filter(id=int(dataconnection_id))
    if len(dl) != 1:
      return

    dc = dl[0]

  username = request.user.username

  if favorite:
    bookmarks = Bookmark.objects.filter(create_user=username, dataconnection=dc).order_by('-update_time')
    for bookmark in bookmarks:
      data = {
        'bookmark_id': bookmark.id,
        'query': bookmark.query,
        'dataconnection_id': bookmark.dataconnection.id,
        'dataconnection_name': bookmark.dataconnection.dataconnection_name,
        'favorite': 'true',
        'query_time': bookmark.update_time.timestamp()
      }

      yield data

    return

  queries = Query.objects.filter(create_user=username, dataconnection=dc).order_by('-create_time')
  for query in queries:
    data = {
      'query': query.query,
      'query_time': query.create_time.timestamp()
    }

    yield data

def get_query_info(query):
  return {
    'id': query.id,
    'query': query.query,
    'queryTime': query.create_time.timestamp(),
    'dataconnectionId': query.dataconnection.id,
    'dataconnectionName': query.dataconnection.dataconnection_name,
    'dataconnectionType': query.dataconnection.dataconnection_type,
    'tableNames': extract_tablenames_from_sqlquery(query.query),
    'favorite': type(query) == Bookmark,
  }

def add_distinct_queries_to_dict(queries: QuerySet, limit: int, query_dict: dict = {}):
  max_number_of_query_to_add = limit - len(query_dict)
  if max_number_of_query_to_add > 0 and len(queries) > 0:
    for query in queries:
      if max_number_of_query_to_add <= 0:
        break
      key = "connection_id: " + str(query.dataconnection.id) + " " +  query.query
      if key not in query_dict.keys():
        key = "connection_id: " + str(query.dataconnection.id) + " " +  query.query
        query_dict.update({key: get_query_info(query)})
        max_number_of_query_to_add -= 1
  return query_dict

def _query_get(request):
  username = request.user.username
  if request.GET.get('get_type') == 'get_by_id':
    query_id = int(request.GET.get('query_id'))
    favorite = request.GET.get('favorite') == 'true'
    data = None
    if favorite:
      data = Bookmark.objects.filter(create_user=username).filter(id=query_id).first()
    else:
      data = Query.objects.filter(create_user=username).filter(id=query_id).first()

    return HttpResponse(json.dumps({'Query': data.query, 'id': data.id}), content_type="application/json")

  queries = []
  if request.GET.get('get_type') == 'all':
    limit = int(request.GET.get('limit'))
    query_dict:dict = {}
    not_bookmarked_queries = Query.objects.filter(create_user=username).order_by('-create_time')

    # Improvement opp: do not store queries in different places, but only one, and bookmark could be a flag which can easily filtered
    if request.GET.get('should_sort'):
      bookmarked_queries = Bookmark.objects.filter(create_user=username).order_by('-create_time')
      bookmarked_query_dict = add_distinct_queries_to_dict(bookmarked_queries, limit, {})
      query_dict = add_distinct_queries_to_dict(not_bookmarked_queries,  limit, bookmarked_query_dict)
    else:
      query_dict = add_distinct_queries_to_dict(not_bookmarked_queries, limit, {})
    queries = list(query_dict.values())

  else:
    query = request.GET.get('query', '')
    dataconnection_id = request.GET.get('dataconnection_id', None)
    favorite = request.GET.get('favorite', False)
    favorite = (str(favorite).lower() == "true")

    _auditQuery(ActionType.READ, query, request=request)

    for s in _query_iter(request, query, dataconnection_id, favorite):
      queries.append(s)

  return HttpResponse(json.dumps(queries), content_type="application/json")

def _query_delete(request):
  params = QueryDict(request.body)
  query = params.get('query', None)
  bookmark_id = params.get('bookmark_id', None)
  dataconnection_id = params.get('dataconnection_id', None)
  _auditQuery(ActionType.DELETE, query, request=request)
  _favorite_delete(bookmark_id, request.user.username, query, dataconnection_id)
  return HttpResponse(json.dumps([]), content_type="application/json")

def _query_post(request):
  bookmark_id = request.POST.get('bookmark_id', None)
  query = request.POST.get('query', None)
  if not query:
    return HttpResponse('[]', content_type="application/json")

  dataconnection_id = request.POST.get('dataconnection_id', None)
  result = request.POST.get('result', None)
  favorite = request.POST.get('favorite', False)
  favorite = (str(favorite).lower() == "true")

  _auditQuery(ActionType.UPDATE, query, request=request)
  _query_write(bookmark_id, request.user.username, query, dataconnection_id, favorite)
  return HttpResponse('[]', content_type="application/json")
