# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

# ---- Functions to cache django model objects in the thread parallel execution mode -----

from django.contrib.auth.models import User
from django.db.models.query import QuerySet
from datasets.models import Segment, FilterAssociation

import threading

THREAD_LOCAL = threading.local()

# A general function to cache the results of a function for the
# jsonselect path.  Mostly for django model lookups but can be
# used generally.


def jsonselect_cached_function(key, f, **kwargs):
  request = getattr(THREAD_LOCAL, 'request', None)
  if request is None:
    res = f(**kwargs)
  else:
    mdict = request._mdict
    with request._mdlock:
      try:
        res = mdict[key]
      except KeyError:
        res = f(**kwargs)
        if isinstance(res, QuerySet):
          len(res)
        mdict[key] = res
  return res


def jsonselect_init_request(request):
  request._mdict = {}
  request._mdlock = threading.Lock()

def jsonselect_cache_put(key, value):
  """
  This function assumes jsonselect_init_request and jsonselect_set_request were already called.
  It does not use syncronization as it is assumed no other threads received this request yet
  """
  if not hasattr(THREAD_LOCAL, 'request'):
    return

  THREAD_LOCAL.request._mdict[key] = value

def jsonselect_set_request(request):
  THREAD_LOCAL.request = request


def jsonselect_clr_request(request):
  THREAD_LOCAL.request = None


def jsonselect_get_segment(id):
  return jsonselect_cached_function(('segment', id), Segment.objects.get, id=id)


def jsonselect_get_user(id):
  request = getattr(THREAD_LOCAL, 'request', None)
  if request is None or request.user.id != id:
    return jsonselect_cached_function(('user', id), User.objects.get, id=id)
  else:
    return request.user


def jsonselect_get_falist(id):
  return jsonselect_cached_function(('falist', id), FilterAssociation.objects.filter, dataset_id=id)

# ---- End of cached lookup stuff
