# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from apps.tests.userrole import base
from datetime import datetime
from django.core.cache import cache
from django.test import TestCase
from extracts.tests import tests_rewrites
from reports.views_settings import setting_get_single, setting_update_single
from sqlrun.views import restore_column_name_casing, _csvStream
from sqlrun import views
from unittest import mock
from vizlib.dsreq import DsreqCreator
from vizlib.tests import test_sqlrun

import json
import tempfile
import xlrd

def get_xlsx_content(filename: str):
  book = xlrd.open_workbook(filename)
  sheet = book.sheet_by_index(0)

  def cell_value(cell: xlrd.sheet.Cell):
    if cell.ctype == xlrd.XL_CELL_EMPTY:
      return None
    if cell.ctype == xlrd.XL_CELL_DATE:
      return xlrd.xldate.xldate_as_datetime(cell.value, book.datemode)

    return cell.value

  for row in sheet.get_rows():
    yield [cell_value(cell) for cell in row]



def create_helper():
  helper = tests_rewrites.Test_Rewrites()
  helper.supports_dataextract_source = True
  helper.supports_partitions = True
  helper.supports_partitions_by_id = {}
  helper.supports_dataextract_target = True

  return helper
class TestColumnNameCaseRestoring(TestCase):

  def test_success_case(self):
    columns = [
      {
        "name": "channel",
        "alias": "Channel"
      },
      {
        "name": "datestring",
        "alias": "Date"
      },
    ]
    helper = create_helper()

    dc = helper.make_dataconnection()
    ds = helper.make_dataset(dc, 'testdb.test_table', columns=columns)

    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Channel] as 'Channel Name'") \
      .addDimension("[Date] as 'Date (string)'") \
      .addAggregate("sum(1) as 'Record Count'") \
      .getdsreq()

    build_request = helper.create_build_request(dsreq, ds=ds)

    columns = restore_column_name_casing(build_request, ['channel name', 'date (string)', 'record count'])
    self.assertEquals(columns, ['Channel Name', 'Date (string)', 'Record Count'])

    # no replacing if column count is different
    columns = restore_column_name_casing(build_request, ['channel name', 'date (string)'])
    self.assertEquals(columns, ['channel name', 'date (string)'])

    # no replacing if column names differ (ignoring case)
    columns = restore_column_name_casing(build_request, ['channel name', 'date (string)', 'sum(1)'])
    self.assertEquals(columns, ['Channel Name', 'Date (string)', 'sum(1)'])

    # no replacing if connection type is something else than Hive or Impala:
    build_request.connection.connection_type='duckdb'
    columns = restore_column_name_casing(build_request, ['channel name', 'date (string)', 'record count'])
    self.assertEquals(columns, ['channel name', 'date (string)', 'record count'])

    # tolerates missing build_request or other missing information
    # in this case the column names must be returned unchanged:
    columns = restore_column_name_casing(None, ['some column'])
    self.assertEquals(columns, ['some column'])

    # tolerates missing connection
    mock_build_request = mock.Mock()
    mock_build_request.connection = None
    columns = restore_column_name_casing(mock_build_request, ['some column'])
    self.assertEquals(columns, ['some column'])

    # tolerates missing builder
    mock_build_request = mock.Mock()
    mock_build_request.connection.connection_type = 'impala'
    mock_build_request.builder = None
    columns = restore_column_name_casing(mock_build_request, ['some column'])
    self.assertEquals(columns, ['some column'])

    # tolerates missing select builder
    mock_build_request = mock.Mock()
    mock_build_request.connection.connection_type = 'impala'
    mock_build_request.builder.builder = None
    columns = restore_column_name_casing(mock_build_request, ['some column'])
    self.assertEquals(columns, ['some column'])

class TestStreamBase(TestCase):

  def setUp(self):
    self.ts_1 = '2024-11-20 16:30:00.123456'
    self.date_1 = '2024-11-30'
    self.mock_column_names = ['row_num', 'msg', 'ts_1', 'date_1']

    self.csv_excel_cache_result = setting_get_single('csv_excel_cache_result')
    self.visual_max_dataresults = setting_get_single('visual_max_dataresults')

  def tearDown(self):
    setting_update_single('csv_excel_cache_result', self.csv_excel_cache_result, False)
    setting_update_single('visual_max_dataresults', self.visual_max_dataresults, False)

  def configure_mock_connection(self, conn: mock.Mock):
    mock_column_types = ['INT', 'STRING', 'TIMESTAMP', 'TIMESTAMP']

    mock_data=[
      [1, 'A', self.ts_1, None],
      [2, None, None, None],
      [None, 'C', None, self.date_1],
    ]
    conn.execute.return_value = iter([mock_column_types, self.mock_column_names] + mock_data)

  def assert_cached_info(self, cached_info: dict):

    self.assertIsNotNone(cached_info)

    self.assertListEqual(cached_info['colnames'], ['row_num', 'msg', 'ts_1', 'date_1'])
    self.assertListEqual(cached_info['coltypes'], ['INT', 'STRING', 'TIMESTAMP', 'TIMESTAMP'])

    expected_cached_rows=[
      f'[1,"A","{self.ts_1}",null]',
      f'[2,null,null,null]',
      f'[null,"C",null,"{self.date_1}"]'
    ]
    # the cached value is a JSON-serialized list of lists:
    expected_cached_rows_string = f'[{",".join(expected_cached_rows)}]'
    self.assertEquals(cached_info['rows'],expected_cached_rows_string)

class TestCsvStream(TestStreamBase):

  def assert_csv_content(self, actual_rows: list[str], has_format: bool):
    expected_rows = [
      'row_num,msg,ts_1,date_1',
      '1,A,2024-11-20 16:30:00.123456,',
      '2,,,',
    ]

    if has_format:
      # if there is a format settings list (even if it consists of nulls only)
      # dates are parsed as datetimes an are formatted as follows:
      str(datetime.fromisoformat(self.date_1))

      expected_rows.append(',C,,2024-11-30 00:00:00')
    else:
      expected_rows.append(',C,,2024-11-30')


    self.assertListEqual(actual_rows, expected_rows)

  def test_timestamp_handling_no_formats_list(self):
    format_settings = None
    self.do_test_timestamp_handling(format_settings = format_settings)

  def test_timestamp_handling_formats_list_with_nulls_only(self):
    format_settings = [None]*4
    self.do_test_timestamp_handling(format_settings = format_settings)

  def test_cache_disabling(self):
    setting_update_single('csv_excel_cache_result', 'false', False)
    self.do_test_timestamp_handling(format_settings = None, cache_expected=False)

  def test_cache_size_limit(self):
    setting_update_single('visual_max_dataresults', '0', False)
    self.do_test_timestamp_handling(format_settings = None, cache_expected=False)

  def test_timestamp_handling_non_trivial(self):

    date_format = {
      "type":"date",
      "result":"dddd, MMMM Do, YYYY",
      "dateoptions":"dddd, MMMM Do, YYYY",
      "dateformat":"dddd, MMMM Do, YYYY",
      "colName":"time"}
    format_settings = [None, None, date_format, date_format]

    # date formats are not applied :-(
    self.do_test_timestamp_handling(format_settings=format_settings)

  def do_test_timestamp_handling(self, format_settings, cache_expected=True):

    conn = mock.MagicMock()
    self.configure_mock_connection(conn)

    cache.clear()
    result = _csvStream(conn=conn,
                        build_request=mock.MagicMock(),
                        async_flag=False,
                        language_code='en_US',
                        quotation_mark='"',
                        format_settings=format_settings,
                        cache_key='a')

    csv_lines = [line.strip() for line in result]

    self.assert_csv_content(csv_lines, format_settings is not None)

    cached_info = cache.get('a')
    if cache_expected:
      self.assert_cached_info(cached_info)
    else:
      self.assertIsNone(cached_info)

  @mock.patch.object(views,'connectionSupportsStreaming')
  @mock.patch.object(views,'get_dataconn')
  def test_download_through_http(self, get_dataconn: mock.Mock, connectionSupportsStreaming: mock.Mock):
    """
    Instead of invoking _csvStream directly this method uses an HTTP POST and GET to generate a CSV
    """
    connectionSupportsStreaming.return_value = True
    self.configure_mock_connection(get_dataconn.return_value)

    # dataset does not really matter as connection is mocked, and the mocked
    # data is independent of the input
    helper = create_helper()

    dc = helper.make_dataconnection()
    ds = helper.make_dataset(dc, 'testdb.test_table', columns=[])
    dsreq = DsreqCreator() \
      .datasetId(ds.id) \
      .getdsreq()

    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)


    format_settings = [None]*4
    filename = 'test.csv'
    params = {}
    params['csvdownload'] = 'true'
    params['csvfilename'] = filename
    params['async'] = 'false'
    params['language_code'] = 'en_US'
    params['dsreq'] = json.dumps(dsreq)
    params['formattingSettings'] = json.dumps(format_settings)
    rsp = admin1_client.post('/sqlrun/csvdownloadStore', data=params)
    assert(rsp.status_code == 200)

    rsp = admin1_client.get('/sqlrun/jsonselect?csvdownload=true', stream=True)
    assert(rsp.status_code == 200)

    result_csv = tempfile.NamedTemporaryFile(dir='/tmp').name
    test_sqlrun.save_response_to_file(rsp, result_csv)

    with open(result_csv, mode='rt', encoding='utf-8') as f:
      csv_lines = f.readlines()

    # strip newlines:
    csv_lines = [line.strip() for line in csv_lines]

    self.assert_csv_content(csv_lines, format_settings is not None)

class TestXlsxStream(TestStreamBase):

  def setUp(self):
    super().setUp()
    self.start_mocks()

  def tearDown(self):
    self.stop_mocks()
    super().tearDown()

  def create_mock(self, owner: object, property: str):
    patch =  mock.patch.object(owner, property)
    self.patches.append(patch)
    return patch.start()

  def start_mocks(self):
    self.patches = list[mock._patch]()
    css_mock = self.create_mock(views,'connectionSupportsStreaming')
    css_mock.return_value = True

    self.get_dataconn_mock = self.create_mock(views,'get_dataconn')

  def stop_mocks(self):
    for patch in self.patches:
      patch.stop()


  def assert_xlsx_content(self, actual_content: list[list[object]]):

    # 1. excel stores dates as floats. On request, xlrd can convert them to datetimes.
    # 2. datetime precision is milliseconds (microseconds are lost)
    # 3. numbers are stored as floats (there is no int type)
    expected_content = [
      self.mock_column_names,
      [1.0, 'A', datetime.fromisoformat('2024-11-20 16:30:00.123000'), None],
      [2.0, None, None, None],
      [None, 'C', None, datetime.fromisoformat(self.date_1)],
    ]
    self.assertListEqual(actual_content, expected_content)


  def test_download_default_caching(self):
    self.do_test_download(cache_should_exist=True)

  def test_download_caching_off(self):
    setting_update_single('csv_excel_cache_result', 'false', False)
    self.do_test_download(cache_should_exist=False)

  def test_download_caching_prevented_by_max_dataresults(self):
    setting_update_single('visual_max_dataresults', '0', False)
    self.do_test_download(cache_should_exist=False)

  def do_test_download(self, cache_should_exist:bool):
    """
    Instead of invoking _csvStream directly this method uses an HTTP POST and GET to generate a CSV
    """
    self.configure_mock_connection(self.get_dataconn_mock.return_value)

    # dataset does not really matter as connection is mocked, and the mocked
    # data is independent of the input
    helper = create_helper()

    dc = helper.make_dataconnection()
    ds = helper.make_dataset(dc, 'testdb.test_table', columns=[])
    dsreq = DsreqCreator() \
      .datasetId(ds.id) \
      .getdsreq()

    admin1 = base.make_user('admin1', is_superuser=True)
    admin1_client = base.create_client(admin1)

    format_settings = [None]*4
    params = {}
    params['xlsxdownload'] = 'true'
    params['async'] = 'false'
    params['language_code'] = 'en_US'
    params['dsreq'] = json.dumps(dsreq)
    params['formattingSettings'] = json.dumps(format_settings)
    rsp = admin1_client.post('/sqlrun/xlsxdownloadStore', data=params)
    assert(rsp.status_code == 200)

    cache.clear()
    # set up a spy to get the cache_key:
    with mock.patch.object(views, 'getXlsxResponse', wraps=views.getXlsxResponse) as spy:
      rsp = admin1_client.get('/sqlrun/jsonselect?xlsxdownload=true', stream=True)
      assert(rsp.status_code == 200)
      cache_key = spy.call_args.kwargs['cache_key']

    result_file = tempfile.NamedTemporaryFile(dir='/tmp').name
    test_sqlrun.save_response_to_file(rsp, result_file)

    lines = list(get_xlsx_content(result_file))
    self.assert_xlsx_content(lines)

    # check cache:
    cached_value = cache.get(cache_key)
    if cache_should_exist:
      self.assert_cached_info(cached_value)
    else:
      self.assertIsNone(cached_value)

