# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from django.contrib.auth.models import User
from datasets.models import DataSet
from django.test import TransactionTestCase
from jobs import views_encryption
from rbac.models import Role
from reports.views_settings import setting_update_single
from sqlrun import jsonselect_cache
from unittest import mock
from vizlib.tests.util import run_with_db_query_check
from vizlib.dsreq import DsreqCreator
from vizlib.tests import test_report_operations
from vizlib.tests.test_report_count import full_access_only_one_dataset_role
from vizlib.tests.test_report_operations import make_user, create_client
from vizlib.tests.test_sqlrun import get_json_response, run_dslist, Common
from vizlib import workq, misc

import json


def grant_dc_explore(user: User, dc_id: int):
  role = Role()
  role.name = f'explore_on_dc_id_{dc_id}'
  role.privileges = json.dumps(
    [
      # no system permissions:
      {'ptype': 'system', 'perms': []},
      {
          'ptype': 'dataconn',
          'dclist': [f'{dc_id}'],
          'perms': ['dc_explore'],
      },
    ]
  )
  role.add_users([user.username])
  role.save()


class TestJsonselectCaching(TransactionTestCase):
  @classmethod
  def tearDownClass(cls):
    # recreate the test DB because with TransactionTestCase all tables get truncated after each test case:
    from django.db import connection
    connection._connections['default'].creation.create_test_db(
      autoclobber=True)

    # clear the cache as it can disturb other tests
    # (visual_maxrows gets cached)
    from django.core.cache import cache
    cache.clear()

  def setUp(self) -> None:
    super().setUp()
    self.common_setup()
    self.cache_off = mock.patch.object(
      jsonselect_cache, 'jsonselect_set_request')
    self.workq_off = mock.patch.object(workq._WorkQ, 'add_work')

  def tearDown(self) -> None:
    if isinstance(jsonselect_cache.jsonselect_set_request, mock.Mock):
      self.cache_off.stop()
    if isinstance(workq._WorkQ.add_work, mock.Mock):
      self.workq_off.stop()

    super().tearDown()

  def common_setup(self):
    util = test_report_operations.TestDashboardOperations()
    self.util = util

    # crete admin
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    self.admin_client = admin_client

    # create normal user
    user1 = make_user('user1', is_superuser=False)
    self.user1 = user1
    user1_client = create_client(user1)
    self.user1_client = user1_client

    # create impala connection and dataset
    dc_id = self.create_dataconnection(admin_client)
    self.dc_id = dc_id
    self.ds_id_1 = util.create_dataset(admin_client, dc_id, 'impala_dataset_1')
    self.ds_id_2 = util.create_dataset(admin_client, dc_id, 'impala_dataset_2')

    # needed by dataengine. Since this is stored in the DB and the DB is reset between tests
    # it should not affect the other tests
    setting_update_single('visual_maxrows', 100, False)

    sqlrun_common = Common()
    ds_1 = DataSet.objects.get(id=self.ds_id_1)
    ds_2 = DataSet.objects.get(id=self.ds_id_2)
    self.s1 = sqlrun_common.make_segment(
      admin_client, ds_1, 's1', '[A] in ("doug", "veronica")')
    self.s11 = sqlrun_common.make_segment(
      admin_client, ds_1, 's11', '[AA] in ("dougA", "veronicaA")')
    self.s2 = sqlrun_common.make_segment(
      admin_client, ds_1, 's2', '[A] in ("alan", "sean", "amanda")')
    fa1 = sqlrun_common.make_fa(
      admin_client, ds_1, 'fa_s1', self.s1.get('id'), users=[user1.id])
    fa2 = sqlrun_common.make_fa(
      admin_client, ds_2, 'fa_s2', self.s2.get('id'), users=[user1.id])

    role_1 = full_access_only_one_dataset_role(dc_id, self.ds_id_1)
    role_2 = full_access_only_one_dataset_role(dc_id, self.ds_id_2)
    role_1.add_users([user1.username])
    role_2.add_users([user1.username])
    role_1.save()
    role_2.save()

  def update_dataset(self, ds_id_1: int, tablename: str, columns: list[dict]):
    ds = DataSet.objects.get(id=ds_id_1)

    ds.dataset_info = json.dumps(
      [{"tablename": tablename, "columns": columns}])
    ds.dataset_detail = tablename

    ds.save()

  def create_dataconnection(self, admin_client):
    impala_payload = {
      'dataconnection_type': 'impyla',
      'dataconnection_info': json.dumps({
        "PARAMS": {
          "HOST": "10.0.0.211",
          "PORT": "31050",
        }
      }),
      'dataconnection_name': 'Impala Judit',
      'dataconnection_password': 'abc'
    }
    return self.util.create_dataconnection_with_payload(admin_client, impala_payload)

  def mock_response(self, sync_runquery: mock.Mock):
    def mock_result(*args, **kwargs):
      return {
          'colnames': ['Channel', 'Date', 'Record Count'],
          'coltypes': ['STRING', 'STRING'],
          'info': [f'Query: {args[0]}'],
          'query_warnings': {},
          'rowcount': 2,
          'rows': [
              ['ABC', '2024-06-12', 2],
              ['CBS', '2024-06-13', 3]
          ]
      }
    sync_runquery.side_effect = mock_result

  @mock.patch('dataconn.dataconn_dbapi2.DbApi2Conn.sync_runquery')
  def test_success_case(self, sync_runquery: mock.Mock):
    columns = [
      {
        "name": "channel",
        "alias": "Channel"
      },
      {
        "name": "datestring",
        "alias": "Date"
      },
    ]
    self.update_dataset(self.ds_id_1, 'testdb.test_table_1', columns=columns)
    self.update_dataset(self.ds_id_2, 'testdb.test_table_2', columns=columns)

    self.mock_response(sync_runquery)

    dsreq_1 = DsreqCreator() \
        .datasetId(self.ds_id_1) \
        .addDimension("[Channel] as 'Channel Name'") \
        .addDimension("[Date] as 'Date (string)'") \
        .addAggregate("sum(1) as 'Record Count'") \
        .addGlobalSegment(self.s11.get('id')) \
        .getdsreq()

    dsreq_2 = DsreqCreator() \
        .datasetId(self.ds_id_2) \
        .addDimension("[Channel] as 'Channel Name'") \
        .addDimension("[Date] as 'Date (string)'") \
        .addAggregate("sum(1) as 'Record Count'") \
        .getdsreq()

    # there are at least 3 dataconnection creations on initialization.
    # do that first so that it does not disturb the numbers below.
    misc.initAsyncConnectionThreadpool(self.dc_id)

    # uncomment the line below to see uncached behavior
    # self.cache_off.start()
    self.workq_off.start()

    def act(dslist: list[dict]):
      with mock.patch.object(views_encryption, 'decrypt_secret_0', wraps=views_encryption.decrypt_secret_0) as decrypt_secret_0:

        # we only test request processing now. To test the whole reqest-response cycle, set wait_for_result=True
        result = run_dslist(
          self.user1_client,
          dslist=dslist,
          wait_for_result=False)

        self.assertEqual(decrypt_secret_0.call_count,
                        1, 'A single decrypt call')
        return result

    dslist = [dsreq_1, dsreq_1, dsreq_2]
    result_1, query_count_1 = run_with_db_query_check(lambda: act(dslist))
    self.assertEqual(len(result_1), 3)

    # double the requests
    dslist = dslist * 2
    result_2, query_count_2 = run_with_db_query_check(lambda: act(dslist))
    self.assertEqual(len(result_2), 6)

    # query count depends on query rewriting. Some tests enable query rewriting and
    # do not switch it off.
    # so we do not check the query count, but want it to be constant
    self.assertEqual(query_count_2, query_count_1, 'DB query count must not increase')


  @mock.patch('dataconn.dataconn_dbapi2.DbApi2Conn.sync_runquery')
  def test_tolerates_missing_dsreq(self, sync_runquery: mock.Mock):
    # direct access sends requests that do not have a dsreq object
    data = [{
      "params": {
        "query": "select sum(1) from cereals limit 100",
        "dsreq": "",
        "dataconnection_id": self.dc_id
      }
    }]
    self.mock_response(sync_runquery)

    # query execution needs explore permission on the dataconnection
    grant_dc_explore(self.user1, self.dc_id)

    # workq must be able to execute, it must not be mocked:
    # self.workq_off.start()

    rsp = self.user1_client.post(
      '/sqlrun/jsonselect_parallel', data={'data': json.dumps(data)})
    self.assertEqual(rsp.status_code, 200)
    json_response = get_json_response(rsp)
    self.assertEquals(len(json_response[0]['response']['rows']), 2)
