# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Views for the sqlrun app
"""

from django.http import HttpResponse, FileResponse
from django.http import StreamingHttpResponse
from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.core.cache import cache
from util.decorator_custom import login_required_401
from datasets.models import DataSet, DataConnection
from django.core.exceptions import ValidationError
from django.core.exceptions import FieldError
from django.contrib.auth.models import User

from dataconn import connections_types as CT
from dataconn.dataconn_dataengine import BuildRequest
from dataconn.dataconn import get_dataconn
from dataconn.qcntl import QueryControl
from sqlrun.views_cache_statistic import CacheStatistics
from vizlib.misc import asyncquery_cancel
from vizlib.misc import asyncquery_status
from vizlib.misc import asyncquery_results
from vizlib.misc import add_handle_data
from vizlib.misc import get_handle_data
from vizlib.misc import asyncrunner_stop_tracking
from vizlib.misc import get_request_source_ip
from vizlib.misc import obfuscate
from vizlib.arcprof import ArcProf
from vizlib.querylog import mark_query_canceled, redact_query

from util.json_utils import json_dumps

import vizlib
import vizlib.handles
import vizlib.lrt
import vizlib.accel
import extracts.rewriter

from vizlib.dataset_request import DatasetRequest
from vizlib.dataset import ModelDataset
from vizlib.dataset import _LowercaseDict as ldict

import time
import json
import urllib.request, urllib.error, urllib.parse
import re
import datetime
import csv
import threading
import tempfile
from dateutil import parser

from hashlib import md5

import locale
from babel.numbers import get_decimal_symbol

from bs4 import BeautifulSoup
import logging

from dataconn.data_api_cmds import SPECIAL_CMDS, NORM_CMDS, VALIDATE_CMDS
from vizlib.querylog import querylog_create, querylog_fetch, ActionType, ComponentType
from util.streaming_json_str import StreamingResponseIfPossible

from dataconn.dataconn import get_django_dataconn
from dataconn.dataconn_arcengine import ArcEngineConn
from dataconn.dataconn_arcengine_v2 import ArcEngineV2Conn
from dataconn.dataconn_impala import ImpalaConn
from dataconn.dataconn_impyla import ImpylaConn
from dataconn.dataconn_hive import HiveConn
from dataconn.dataconn_hyve import HyveConn
from dataconn.dataconn_ksql import KsqlConn
from dataconn.hs2conn import supports_extmeta
from dataconn.dataconn_sqlite import SqliteConn
from reports.views_settings import setting_get_single
from dataconn.rows_data import RowsData
from sqlrun.jsonselect_cache import jsonselect_cached_function
from vizlib.cviz_remotecall import cviz_is_rest_expr
from vizlib.cviz_remotecall import cviz_validate_column
from vizlib.cviz_remotecall import cviz_modify_dsreq

import util.perfcounters as perfcounters

import xlsxwriter
from datasets.utils import get_default_dataconnection

from apps.views_static_model_helper import get_static_dataset_model


datareqCountByMinute = perfcounters.TimeSeriesCounter('data.requests',
    timePeriodSeconds=60, timePeriodRetentionSeconds=60*60)

LOGGER = logging.getLogger("arcweb")

_VALIDATE_CMDS = set([NORM_CMDS['SQL'], NORM_CMDS['MR']])

def connectionIsArcEngineBased(conn):
  return type(conn) in [ArcEngineConn, ArcEngineV2Conn]

def connectionSupportsStreaming(conn):
  return type(conn) in [ArcEngineConn, ImpalaConn, ImpylaConn, HyveConn, ArcEngineV2Conn, SqliteConn]

def connectionSupportsExtMetadata(conn):
  if type(conn) is ArcEngineConn:
    return supports_extmeta(conn.srv_name(), conn.srv_version())

  return (type(conn) is ArcEngineV2Conn)

CONN_SUPPORTS_COMMENTS = set([ImpalaConn, ImpylaConn, ArcEngineConn, ArcEngineV2Conn])
def connectionSupportsComments(conn):
  return type(conn) in CONN_SUPPORTS_COMMENTS

def tskey(key):
  """Build the timestamp cache key corresponding to the query value cache key"""
  return md5(('timestamp.' + key).encode()).hexdigest()

def tsput(key, ttxt):
  """Store (tables_text, timestamp) tuple associated with cache key and json input"""
  try:
    val = (ttxt, time.time())
    cache.set(tskey(key), val, None)
  except:
    pass

def tsget(key):
  """Retrieve (tables_text, timestamp) associated with a cache key"""
  val = None
  try:
    val = cache.get(tskey(key))
  except:
    pass
  return val or (None, 0)

def jsonselect_get_dataset(id):
  if isinstance(id, str):
    id = int(id)

  if id < 0:
    return jsonselect_cached_function(('dataset', id), get_static_dataset_model, id=id)
  return jsonselect_cached_function(('dataset', id), DataSet.objects.get, id=id)

def jsonselect_get_dataconn(id):
  return jsonselect_cached_function(('dataconn', id), get_django_dataconn, id=id)

@login_required
def longrunningtask(request):
  """Return a list of recent long running tasks"""
  return HttpResponse(json.dumps(vizlib.lrt.getTasks()),
      content_type="application/json")

@login_required
def viewrec(request):
  return vizlib.accel.viewrec_request(request)

@login_required
def viewrec_sql(request):
  return vizlib.accel.viewrec_request_sql(request)

@login_required
def viewrec_queries(request):
  return vizlib.accel.viewrec_request_queries(request)


def _populate_querylog_args(args):
  query = args['query']
  if query == SPECIAL_CMDS['REFRESH_AVIEWS']:
    args['action'] = ActionType.UPDATE
    args['component_id'] = args['dataconnection_id']
    args['component_type'] = ComponentType.DATACONNECTION
  elif query == SPECIAL_CMDS['DROP_AVIEWS']:
    args['action'] = ActionType.UPDATE
    args['component_id'] = args['dataconnection_id']
    args['component_type'] = ComponentType.DATACONNECTION
  elif query == SPECIAL_CMDS['SHOW_AVIEWS']:
    args['action'] = ActionType.READ
    args['component_id'] = args['dataconnection_id']
    args['component_type'] = ComponentType.DATACONNECTION
  elif query == SPECIAL_CMDS['INVALIDATE_METADATA']:
    args['action'] = ActionType.UPDATE
    args['component_id'] = args['dataconnection_id']
    args['component_type'] = ComponentType.DATACONNECTION
  elif query == SPECIAL_CMDS['COMPUTE_STATS']:
    args['action'] = ActionType.UPDATE
    args['component_id'] = args['dataconnection_id']
    args['component_type'] = ComponentType.DATACONNECTION
  elif query == SPECIAL_CMDS['SHOW_DATABASES']:
    args['action'] = ActionType.READ
    args['component_id'] = args['dataconnection_id']
    args['component_type'] = ComponentType.DATACONNECTION
  elif query.startswith(SPECIAL_CMDS['SHOW_TABLES']):
    args['action'] = ActionType.READ
    if args['dataset_id']:
      args['component_id'] = args['dataset_id']
      args['component_type'] = ComponentType.DATASET
    else:
      args['component_id'] = args['dataconnection_id']
      args['component_type'] = ComponentType.DATACONNECTION

def _getDelimiter(language_code):
  return ',' if get_decimal_symbol(language_code) == '.' else ';'

def _convert_from_unicode(input):
  if isinstance(input, str):
    return input.encode('utf8', 'replace')
  else:
    # no need to encode if not unicode
    return input

def _convert_row_from_unicode(row):
  return [_convert_from_unicode(s) for s in row]

# Convert periods to commas if locale is one that uses commas as decimal separator
def _make_decimal_comma(val, col_is_number=None, idx=None):
  val_string = str(val)
  is_numeric = re.search('^\d*\.?\d*$', val_string)
  contains_digits = re.search('\d', val_string)
  if is_numeric and contains_digits:
    if col_is_number != None:
      # Mark the column as numeric if it hasn't already been marked as non-numeric
      col_is_number[idx] = col_is_number.get(idx, True) and True
    return str(val).replace('.', ',')

  return val

def _convert_val_to_locale(val, col_is_number, idx, use_col_is_number):
  # We've already processed 100 rows and know which columns might have numbers
  if use_col_is_number:
    if col_is_number.get(idx):
      return _make_decimal_comma(val)
    else:
      return val

  # Haven't processed 100 rows yet, assume each column might be a number
  # Pass in col_is_number to keep track of which columns actually turn out to have numbers
  else:
    return _make_decimal_comma(val, col_is_number, idx)

# Return array with properly formatted row
def _convert_row_to_locale(row, types, decimal_symbol, col_is_number, use_col_is_number, quotation_setting):
  # Add quotation mark
  quotation = ''
  if quotation_setting == 'true':
    quotation = "'"
  for idx, x in enumerate(row):
    try:
      if types[idx] == 'STRING':
        try:
          str_is_number = float(x).is_integer() or True
        except:
          str_is_number = False
        if str_is_number == True:
          row[idx] = quotation + str(x) + quotation

      if decimal_symbol == '.':
        return row
      else:
        return [ _convert_val_to_locale(val,
                                        col_is_number,
                                        idx,
                                        use_col_is_number)
                for idx, val in enumerate(row) ]
    except:
      LOGGER.error(f'Error in parsing row {row} while constructing CSV file')
      return row

def _jsonToCsvResponse(data, language_code, quotation_mark, filename='Untitled.csv', format_settings=None, build_request: BuildRequest = None):
  response = HttpResponse(content_type='text/csv')
  response['Content-Disposition'] = 'attachment; filename="%s"' % filename

  j = data
  writer = csv.writer(response, delimiter=_getDelimiter(language_code))

  # Try to write title, and if fail, try to write encoded version of title
  try:
    writer.writerow(restore_column_name_casing(build_request, j['colnames']))
  except:
    writer.writerow(_convert_row_from_unicode(j['colnames']))

  # Find numbers for first 100 rows
  col_is_number = {}
  row_idx = 0
  decimal_symbol = get_decimal_symbol(language_code)
  coltypes = j['coltypes']
  rows_parsed = json.loads(j['rows'])
  for r in rows_parsed:
    row_idx += 1
    formatted_row = _convert_row_to_locale(r, coltypes, decimal_symbol, col_is_number, row_idx >= 100, quotation_mark)

    try:
      writer.writerow(formatted_row)
    except:
      writer.writerow(_convert_row_from_unicode(formatted_row))
  return response

@login_required_401
def csvdownloadStore(request):
  params = request.POST
  outParams = _csvRequestToParams(params)
  request.session['csvDownload'] = outParams
  return HttpResponse('ok')

def _csvStream(conn, build_request, async_flag, language_code, quotation_mark, format_settings=None, cache_key=None):
  class pseudoWriter(object):
    def write(self, msg):
      return msg

  writer = csv.writer(pseudoWriter(), delimiter=_getDelimiter(language_code))
  col_is_number = {}
  row_idx = 0
  decimal_symbol = get_decimal_symbol(language_code)
  row_generator = conn.execute(build_request, async_flag=async_flag, stream=True)
  CacheStatistics.updateQueryStatisticHistory(build_request.connection.connection_id, False)
  # first row from stream are column types
  coltypes = next(row_generator)

  should_cache = setting_get_single('csv_excel_cache_result') == 'true'
  cached_results = {'rows': []}
  # Calculating the maxSize here to avoid circular import
  maxSizeMB = setting_get_single('visual_max_dataresults') or ''
  maxSizeMB = int(maxSizeMB) if maxSizeMB.isnumeric() else 10
  maxSize = maxSizeMB * 1000 * 1000
  rows_data = RowsData(maxSize)

  cached_results['coltypes'] = coltypes
  rows = row_generator
  for row_idx, row in enumerate(rows):
    if (row_idx > 0):
      if should_cache:
        try:
          rows_data.addRow(row)
        except:
          # discard cache accumulated up to this point:
          rows_data = None
          should_cache = False
      if format_settings is not None and len(format_settings) > 0:
        col_idx = 0
        formatted_cells = []
        for cell in row:
          ct = coltypes[col_idx]
          if ct == 'TIMESTAMP' and cell is not None:
            try:
              cell = parser.isoparse(cell)
            except:
              LOGGER.error('Error in formatting {cell} with the time format parser')

          if format_settings[col_idx] is not None:
            cell = _format_value(cell, format_settings[col_idx])
          formatted_cells.append(cell)
          col_idx += 1
        row = formatted_cells
    else:
      cached_results['colnames'] = row
      # this is the first row, try restoring column name casing
      row = restore_column_name_casing(build_request, row)

    use_col_is_number = row_idx >= 100
    formatted_row = row
    formatted_row = _convert_row_to_locale(formatted_row, coltypes, decimal_symbol, col_is_number, use_col_is_number, quotation_mark)

    try:
      yield writer.writerow(formatted_row)
    except:
      yield writer.writerow(_convert_row_from_unicode(formatted_row))

  if should_cache:
    cached_results['rows'] = rows_data.fetchAndClose()
    cache.set(cache_key, cached_results, None)
    tsput(cache_key, None)

def restore_column_name_casing(build_request: BuildRequest, column_headers):
  """
  Hive and Impala do not preserve column casing. This method tries to restore the original casing
  using request information
  """
  if not build_request:
    return column_headers

  if not hasattr(build_request,'connection'):
    return column_headers

  if not hasattr(build_request.connection,'connection_type'):
    return column_headers

  if build_request.connection.connection_type not in \
      [CT.DATACONN_IMPALA, CT.DATACONN_IMPYLA, CT.DATACONN_HIVE, CT.DATACONN_HYVE]:
    return column_headers

  if not hasattr(build_request,'builder'):
    return column_headers

  if not hasattr(build_request.builder,'builder'):
    return column_headers

  # select_builder: vizlib.select_builder.BaseSelectBuilder
  select_builder = build_request.builder.builder
  if not select_builder:
    return column_headers

  projections = select_builder.dimensions + select_builder.aggregates

  if len(projections) != len(column_headers):
    # unlikely, but rather do not do anything if we missed a column
    return column_headers

  _, quoted_aliases = select_builder.selectlist(projections)

  aliases = [quoted_alias[1:-1] for quoted_alias in quoted_aliases]

  fixed_column_headers = []
  for header, alias in zip(column_headers, aliases):
    if header.lower() == alias.lower():
      fixed_column_headers.append(alias)
    else:
      fixed_column_headers.append(header)

  return fixed_column_headers

def _csvRequestToParams(params):
  outParams = {}

  # csvdownload must be sync
  outParams['async'] = 'false'

  # use site setting for cache_read, otherwise default to not using cache
  outParams['cache_read'] = str(params.get('cache_read', 'false'))

  outParams['language_code'] = params.get('language_code', locale.getlocale())

  outParams['query'] = params.get('query', '')
  outParams['csvfilename'] = params.get('csvfilename', 'Untitled.csv')
  outParams['xlsxfilename'] = params.get('xlsxfilename', 'Untitled.xlsx')
  outParams['quotation_mark'] = params.get('quotation_mark', 'false')
  outParams['view_only'] = params.get('view_only', False)
  outParams['sample_pct'] = params.get('sample_pct', '')
  outParams['dataconnection_id'] = params.get('dataconnection_id', None)
  outParams['parameters_case_sensitive'] = params.get('parameters_case_sensitive')
  outParams['is_edit'] = params.get('is_edit')
  outParams['report_id'] = params.get('report_id')
  outParams['dataset_id'] = params.get('dataset_id')
  outParams['parent_report_id'] = params.get('parent_report_id')

  # DATA API JSON object: dsreq
  outParams['dsreq'] = params.get('dsreq')
  outParams['formattingSettings'] = params.get('formattingSettings', '[]') or '[]'

  return outParams

def xlsxdownloadStore(request):
  params = request.POST
  outParams = _csvRequestToParams(params)
  request.session['xlsxDownload'] = outParams
  return HttpResponse('ok')

def getXlsxResponse(
  conn=None,
  build_request={},
  cached_data=None,
  filename='Untitled.xlsx',
  async_flag=False,
  stream=False,
  format_settings=None,
  cache_key=None
):
  # Create a temporary file for the new workbook.
  output = tempfile.SpooledTemporaryFile()

  # Even though the final file will be in memory the module uses temp
  # files during assembly for efficiency. To avoid this on servers tha
  # don't allow temp files, for example the Google APP Engine, set the
  # 'in_memory' Workbook() constructor option as shown in the docs.
  workbook = xlsxwriter.Workbook(output, {'nan_inf_to_errors': True})
  worksheet = workbook.add_worksheet()

  should_cache = setting_get_single('csv_excel_cache_result') == 'true'
  cached_results = {'rows': []}
  # Calculating the maxSize here to avoid circular import
  maxSizeMB = setting_get_single('visual_max_dataresults') or ''
  maxSizeMB = int(maxSizeMB) if maxSizeMB.isnumeric() else 10
  maxSize = maxSizeMB * 1000 * 1000
  rows_data = RowsData(maxSize)



  if cached_data is not None:
    coltypes = cached_data['coltypes']
    colnames = cached_data['colnames']
    colnames = restore_column_name_casing(build_request, colnames)
    worksheet.write_row(0, 0, colnames)
    rows = cached_data['rows']
    if isinstance(rows, str):
      rows = json.loads(rows)

  elif stream:
    row_generator = conn.execute(build_request, async_flag=async_flag, stream=stream)
    CacheStatistics.updateQueryStatisticHistory(build_request.connection.connection_id, False)
    # first row from stream are column types
    coltypes = next(row_generator)
    # second row is header row with column names
    colnames = next(row_generator)
    colnames = restore_column_name_casing(build_request, colnames)
    worksheet.write_row(0, 0, colnames)
    rows = row_generator
  else:
    result_dict = conn.execute(build_request, async_flag=async_flag)
    CacheStatistics.updateQueryStatisticHistory(build_request.connection.connection_id, False)
    coltypes = result_dict['coltypes']
    colnames = result_dict['colnames']
    colnames = restore_column_name_casing(build_request, colnames)
    worksheet.write_row(0, 0, colnames)
    if isinstance(result_dict.get('rows'), str):
      rows = json.loads(result_dict['rows'])
    elif isinstance(result_dict.get('rows'), (list, dict)):
      rows = result_dict['rows']

  cached_results['colnames'] = colnames
  cached_results['coltypes'] = coltypes

  # Start row counter at 1 to avoid overwriting the header row
  for row_idx, row in enumerate(rows, start=1):
    if should_cache:
        try:
          rows_data.addRow(row)
        except:
          # discard cache accumulated up to this point:
          rows_data = None
          should_cache = False

    for col_idx, cell in enumerate(row):
      ct = coltypes[col_idx]
      # Keeping the original value to log it, to show the original value
      # instead of the wrongly parsed or formatted
      original_cell_value = cell
      correct_time_format = False

      if ct == 'TIMESTAMP' and cell is not None:
        try:
          cell = parser.isoparse(cell)
          if cell:
            correct_time_format = True
        except:
          LOGGER.error(f'Error in formatting {original_cell_value} with the time format parser')

      if (format_settings is not None and len(format_settings) > 0 and format_settings[col_idx] is not None):
        try:
          cell = _format_value(cell, format_settings[col_idx])
        except:
          LOGGER.error(f'Error in formatting {original_cell_value}, with formatting setting {format_settings[col_idx]}')
      try:
        if ct == 'TIMESTAMP' and correct_time_format:
          date_format = workbook.add_format({'num_format': 'yyyy-mm-dd hh:mm:ss'})
          try:
            worksheet.write_datetime(row_idx, col_idx, cell, date_format)
          except Exception as e:
            worksheet.write(row_idx, col_idx, cell)
            LOGGER.error(f'Error in writing date fromate: {e}. For cell value: {original_cell_value}')
        else:
          worksheet.write(row_idx, col_idx, cell)
      except Exception as e:
        LOGGER.error(e)
        LOGGER.error(f'Unable to write to excel sheet, the cell value is {original_cell_value}')


  #cache the result if possible
  if should_cache:
    cached_results['rows'] = rows_data.fetchAndClose()
    cache.set(cache_key, cached_results, None)
    tsput(cache_key, None)

  # Close the workbook before sending the data.
  workbook.close()

  # Rewind the buffer.
  output.seek(0)
  response = FileResponse(
    output,
    content_type='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
  )
  response['Content-Disposition'] = 'attachment; filename=%s' % filename

  return response

@login_required_401
def jsonselect(request, params=None):
  """Run a query and get back results in json"""
  data = jsonselect_work(request, params)
  if 'response' in data:
    return data['response']

  if 'data' in data:
    return HttpResponse(json_dumps(data['data']), content_type="application/json")

  if 'error' in data:
    return HttpResponse(data['error'], status=500)

  return HttpResponse('Invalid response', status=500)

def jsonselectRequestIsASync(params):
  async_flag = (params.get("async", '') == 'true')
  csvdownload = (params.get("csvdownload", '') == 'true')
  xlsxdownload = (params.get("xlsxdownload", '') == 'true')

  # async mode is not compatible with csvdownload
  if csvdownload or xlsxdownload:
    async_flag = False

  return async_flag

def paramsToVariables(params):
  dataconnection_id = params.get("dataconnection_id", None)
  dataset_id = params.get("dataset_id")

  # DATA API JSON object: dsreq
  dsreq = params.get("dsreq")

  # This has gotten messy over time: dataset_id and dataconnection_id might
  # be present in the request but the dsreq within the request can also have
  # a dataset_id and in that case, we should get the dataconnection (and id) from
  # that

  dsreq_obj = None
  # To able to create visuals on nonexisting datasets, the following HACK is used
  # On the frontend a temporaly dataset object is generated, and sent via the usual dsreq
  temp_dataset_object = None

  # Ultimately, we always need a data connection but dataset is optional dsreq_obj = None
  if dsreq and len(dsreq):
    dsreq_obj = json.loads(dsreq)
    dataset_id = dsreq_obj.get('dataset_id', dataset_id)
    if dataset_id is None:
      temp_dataset_object = dsreq_obj.get('dataset_object')

  django_ds = None
  if dataset_id:
    django_ds = jsonselect_get_dataset(dataset_id)
    dataconnection_id = django_ds.dataconnection_id
  elif temp_dataset_object:
      django_ds = DataSet(
                dataset_name='Temporary Dataset for Profiling',
                dataset_type='singletable',
                dataset_info=temp_dataset_object,
                dataset_detail='Temporary Dataset for Profiling')

  if dataconnection_id:
    django_dc = jsonselect_get_dataconn(dataconnection_id)
    # Materialize the data connection within the dataset from the
    # cached model structure rather than re-fetching
    if django_ds:
      django_ds.dataconnection = django_dc

  else:
    django_dc = get_default_dataconnection()

  dataconnection_id = django_dc.id
  conn = get_dataconn(connection_id=dataconnection_id, django_dc=django_dc)
  return (conn, dsreq, dsreq_obj, django_ds, dataset_id, django_dc, dataconnection_id)

# Check if we need to rewrite the query to use a data extract
# If the rewrite does happen, we need to update a set of parameters in the caller (jsonselect_work below),
#   so we accept and return a bunch of values
def _handleExtractRewrites(conn, dsreq_obj, django_dc, dataconnection_id, dataset_request,
    profile, request, ds, viz_params, django_ds, user: User):
  if dsreq_obj:
    (rewritten_dataconnection_id, dsreq_obj) = extracts.rewriter.handleRewrites(
        dataconnection_id, dsreq_obj, user=user, ds=ds, django_ds=django_ds, conn=conn)

    dsreqRewriter = dsreq_obj.get('dsreqRewriter')
    if dsreqRewriter:
      profile.record('rewritten_extract', dsreqRewriter.getExtractID())
      # if we found an extract then we can not use ds - as it uses the original dataset
      dataset_request = DatasetRequest(dsreq_obj, ds=None, params=viz_params, django_ds=django_ds)

      if dataconnection_id != rewritten_dataconnection_id:
        # Re-fetch all the dataconnection related objects
        dataconnection_id = rewritten_dataconnection_id
        conn = get_dataconn(connection_id=dataconnection_id)
        django_dc = jsonselect_get_dataconn(dataconnection_id)

        # Check permissions against this new dataconnection
        dataset_request.validateAccess(request.user, django_dc, django_ds)

  return (conn, dsreq_obj, django_dc, dataconnection_id, dataset_request)

@login_required_401
def jsonselect_work(request, params=None, profile=None, ignore_data_from_session=False):
  global datareqCountByMinute
  datareqCountByMinute += 1

  if not params:
    # A hack to allow GET and POST request
    # DJC -- should move all of these in the js side to POST
    if request.method == 'GET':
      params = request.GET
    elif request.method == 'POST':
      params = request.POST
    else:
      return {
        'error': 'Invalid request method'
      }

  profile = profile or ArcProf('jsonselect', username=request.user.username)
  querytracker = profile.querytracker()

  # Set up these formerly global flags here so we response to unit test config changes
  DSREQ_VALIDATION = getattr(settings, 'DSREQ_VALIDATION', True)
  ARCENGINE_VALIDATE_ACCESS = getattr(settings, 'ARCENGINE_VALIDATE_ACCESS', False)

  # csvdownload has to be a clickable GET request to get it working correctly
  # on the client side. GET requests have 2K size requirements. So, we use
  # the session store to store parameters using csvdownloadStore, and then
  # we retrieve them here
  csvdownload = (params.get("csvdownload", '') == 'true')
  xlsxdownload = (params.get("xlsxdownload", '') == 'true')
  if (csvdownload or xlsxdownload) and not ignore_data_from_session:
    params = request.session.get('csvDownload') if csvdownload else request.session.get('xlsxDownload')
    language_code = params.get('language_code', locale.getlocale())

  if ignore_data_from_session:
    language_code = params.get('language_code', locale.getlocale())

  # Fetch from the cache, if possible
  query = params.get("query", '')
  csvfilename = params.get("csvfilename", 'Untitled.csv')
  xlsxfilename = params.get("xlsxfilename", 'Untitled.xlsx')
  quotation_mark = params.get("quotation_mark", 'false')
  view_only = params.get("view_only", False)
  sample_pct = params.get("sample_pct", '')
  cache_read = str(params.get("cache_read", 'true'))
  viz_params = params.get("viz_params", None)
  report_id = params.get("report_id")
  parent_report_id = params.get("parent_report_id")
  no_fallback = params.get('no_fallback', False)

  if settings.NO_CACHE:
    cache_read = 'false'

  async_flag = jsonselectRequestIsASync(params)

  (conn, dsreq, dsreq_obj, django_ds, dataset_id, django_dc, dataconnection_id) = \
      paramsToVariables(params)

  # Ultimately, we always need a data connection but dataset is optional dsreq_obj = None
  if dsreq_obj:
    if dsreq_obj.get('type') == VALIDATE_CMDS['column']:
      colname = dsreq_obj.get('validate')
      if cviz_is_rest_expr(colname):
        return cviz_validate_column(dsreq_obj, colname)

    # Temp dataset object is used to profile data and create frequency/histogram visuals
    # on data without dataset.
    # For secrity reason, I am checking dc_explore in this scenario
    if dsreq_obj.get('dataset_id') is None and dsreq_obj.get('dataset_object'):
      django_dc.checkPerm(request.user, 'dc_explore')

  # Account for new style viz_params delivery
  new_params = []
  dfunc = dict if params.get('parameters_case_sensitive') else ldict
  if viz_params:
    viz_params = json.loads(viz_params)
    if type(viz_params) == type([]):
      new_params = viz_params
      viz_params = dfunc((p['key'], p['value']) for p in new_params if p.get('value'))
    else:
      viz_params = dfunc((k, v) for (k, v) in viz_params.items())

  # Server side data request generation modes:
  #   ENFORCE  - User with no dataconn perm and view only dataset perm, use server side request
  #   VALIDATE - For non-edit request, compare recieved request with service side request
  method = getattr(settings, 'DSREQ_METHOD', None)
  if method in ['VALIDATE', 'ENFORCE']:

    # Determine if this is a view only user
    vonly = False
    if django_dc.getPerms(request.user) == set([]):
      if not django_ds:
        raise Exception('Invalid request')
      vonly = django_ds.getPerms(request.user) == set(['ds_appview'])

    # ENFORCE mode means view only viewers use server side data request
    if method == 'ENFORCE' and vonly:
      dsreq_obj = vizlib.accel.check_dsreq(request, params, None, new_params, conn, django_ds, dfunc)

    # Otherwise we just validate the view mode data requests for the visuals
    elif not params.get('is_edit') and dsreq_obj and dsreq_obj['type'] == NORM_CMDS['SQL']:
      vizlib.accel.check_dsreq(request, params, dsreq_obj, new_params, conn, django_ds, dfunc)

  # The RAWSQL_WHITELIST was the legacy method to allow the user to run
  # queries outside the data-api.
  #
  # This is used in the 'Raw table' visual and in the connection explorer to
  # get table sample data.
  #
  # From a security standpoint, exploring the connection gives full access
  # to the data on the connection, so that is the permission that enables
  # this capability.
  #
  if not dsreq_obj and query:
    if not '*' in settings.RAWSQL_WHITELIST:
      if not request.user.username in settings.RAWSQL_WHITELIST:
        django_dc.checkPerm(request.user, 'dc_explore')
        if conn.connection_type != 'ksql':
          q = query.lower().strip()

          permitted = False
          if q == "":
            permitted = True
          else:
            for start in settings.DIRECTSQL_QUERIES_PERMITTED:
              if q.startswith(start):
                permitted = True
                break

          if not permitted:
            raise Exception('The requested query is not permitted. %s' % q)

  # SQL validation and caching validation
  if DSREQ_VALIDATION and dsreq_obj is not None:
    conn.validate_dsreq(dsreq_obj, validation_type = 'sql')

  # Check for now() or similar patterns in query to indicate invalid for caching
  try:
    conn.validate_dsreq(dsreq_obj, validation_type = 'cache')
  except FieldError:
    cache_read = 'false'

  # enable comments for engines that we know support it
  if connectionSupportsComments(conn):
    comment = '/* widget=%s, app=%s, ds=%s */ ' % \
        (report_id, parent_report_id, dataset_id)
  else:
    comment = ''

  dataset_request = None
  ds = None
  if dsreq_obj:
    ds = None
    # Since DataProfiling feature we can send temp_dataset_object
    # to run similar queries as on datasets, without saving the temp obj
    dso = dsreq_obj.get('dataset_object')
    if dso:
      # Can only specify the dataset definition on the expression validation
      # code path within the dataset field edit
      valid_operations = list(VALIDATE_CMDS.values())
      valid_operations.append(NORM_CMDS['getrows'])
      valid_operations.append(NORM_CMDS['SQL'])
      if dsreq_obj.get('type') not in valid_operations or not django_ds:
        raise Exception('Invalid operation')
      django_ds.checkPerm(request.user, 'ds_manage')
      ds = ModelDataset(dso, conn.properties(), params=viz_params)

    conn.enforce_limit(dsreq_obj, csvdownload or xlsxdownload, profile)
    cviz_modify_dsreq(dsreq_obj)
    dataset_request = DatasetRequest(dsreq_obj, ds=ds, params=viz_params, django_ds=django_ds)
    dataset_request.validateAccess(request.user, django_dc, django_ds)

  elif not query:
    return {
      'error': 'No response'
    }

  else:
    from vizlib.accel import parameterReplace
    query = parameterReplace(query, viz_params)

  # Should init with request, dsreq, ds, dc, django_ds, django_dc
  qcntl = QueryControl()

  results_key = None
  explain_results = None
  arcengine_extmeta = False

  # Determine if arcengine connection is current (i.e. supports new style extended metadata operations)
  # If so, setting explain_results here disables the older style arcengine custom query path operations:
  # Query cache re-check, SQL injection tables check, Fallback check.  These are all handled
  # now in arcengine, so now we just need to make sure the query control structure has the
  # information to deliver to arcengine along with the query.
  if connectionSupportsExtMetadata(conn):
    arcengine_extmeta = True
    explain_results = 'Error'
    tables_text = None
    qcntl.fa_engine = True

  # Check if we need to rewrite the query to use a data extract
  (conn, dsreq_obj, django_dc, dataconnection_id, dataset_request) = \
      _handleExtractRewrites(conn, dsreq_obj, django_dc, dataconnection_id, dataset_request, \
        profile, request, ds, viz_params, django_ds, user=request.user)

  # Use Dataconnection Object to build and execute queries
  # a build_request ALWAYS has a query associated with it
  # when dataset_request is not None the query comes from QueryBuilder
  # and hence it has an associated builder as well
  profile.mark('build.start')
  if getattr(settings, 'ENABLE_DSREQ_PERF_DISPLAY', False):
    profile.record('dataset request', dsreq)

  try:
    build_request = conn.build(request.user.username, request.user.id, query,
                              dataset_request=dataset_request,
                              sample_pct=sample_pct,
                              comment=comment,
                              qcntl=qcntl,
                              profile=profile)
  except Exception as e:
    error = f'Error in the query build: {e}'
    LOGGER.error(error)
    return {'error': error}

  profile.mark('build.complete')

  if view_only:
    return {
      'data': build_request.query
    }

  dsreqRewriter = None
  if dsreq_obj:
    dsreqRewriter = dsreq_obj.get('dsreqRewriter')

  rewriteLogStr = ''
  if dsreqRewriter:
    rewriteLogStr = f'Rewritten to Extract: {dsreqRewriter.getExtractID()}. '

  query_to_log = redact_query(build_request.query)

  LOGGER.info("querytracker [%s]. %sScheduling query: %s",
      querytracker, rewriteLogStr, query_to_log)
  query_key = build_request.query

  # Include sequence keys into the cache key
  dcseq = django_dc.cache_sequence
  dsseq = 0 if not django_ds else django_ds.cache_sequence

  dataconnection_info = json.loads(django_dc.dataconnection_info or '{}')
  user_dependent_cache = dataconnection_info.get("PARAMS", {}).get("USER_DEPENDENT_CACHE")

  if user_dependent_cache:
    query_key = f'{request.user.username}.{query_key}'

  cache_key = md5(("jsonselect.v1.%s.%s.%s.%s.%s.%s" % \
                (dcseq, dsseq, str(dataconnection_id), query_key, str(sample_pct), qcntl.fa_expr)).encode()).hexdigest()

  client_ip = get_request_source_ip(request)

  args = {
    'username': request.user.username,
    'user_id': request.user.id,
    'visual_id': report_id,
    'visual_name': params.get('report_name', ''),
    'dataset_id': dataset_id,
    'dataset_name': params.get('dataset_name', ''),
    'app_id': parent_report_id,
    'app_name': params.get('parent_report_name', ''),
    'appgroup_id': params.get('appgroup_id', ''),
    'appgroup_name': params.get('appgroup_name', ''),
    'dataconnection_id': dataconnection_id,
    'dataconnection_type': conn.connection_type,
    'dataconnection_name': params.get('dataconnection_name', ''),
    'query': build_request.query,
    'sync': not async_flag,
    'client_ip': client_ip
  }

  if dsreqRewriter:
    args['rewrite_target'] = dsreqRewriter.getExtractID()

  _populate_querylog_args(args)
  query = '***'
  if not getattr(settings, 'HIDE_PROFILE_QUERY', False):
    query = build_request.query

  profile.record('query', query)

  # Track query with action_id till queryHandle is obtained
  ql = querylog_create(**args)
  if not ql.isNull():
    ql.track(ql.action_id)
  ql.logQueryStart()

  # Pick up cache settings from dataset or connection
  if django_ds:
    cache_settings = django_ds.cache_settings(dc=django_dc)
  else:
    cache_settings = django_dc.cache_settings()
  if not cache_settings.get('ENABLED', 0):
    cache_read = "false";
  # read results from cache, if key is there
  if cache_read.lower() == "true":
    profile.mark('cache.start')
    cached_val = cache.get(cache_key)
    if cached_val:
      tables_text, ts = tsget(cache_key)
      now = time.time()
      format_settings = json.loads(params.get('formattingSettings', '[]'))

      if ts and (now > ts) and (now - ts < cache_settings.get('RETENTION', 0)):
        ql.logQueryEnd(cache=True)

        profile.mark('cache.fetch')
        profile.markFromStart('query_runtime')
        if dataset_id:
          CacheStatistics.updateQueryStatisticHistory(DataSet.objects.get(id=dataset_id).dataconnection_id, True)

        if csvdownload:
          return {
            'response': _jsonToCsvResponse(
              cached_val,
              language_code,
              quotation_mark,
              filename=csvfilename,
              format_settings=format_settings,
              build_request=build_request
            )
          }

        if xlsxdownload:
          return {
            'response': getXlsxResponse(
              cached_data=cached_val,
              filename=xlsxfilename,
              format_settings=format_settings,
            )
          }

        return {
          'data': cached_val,
          'profile': profile.results()
        }

      if explain_results != 'Error':
        try:
          if connectionIsArcEngineBased(conn):
            # For arcengine, save the full results of the explain for use in possible
            # use in sql injection check below
            explain_results = conn.log_sync_runquery(query='EXPLAIN ' + build_request.query,
                username=request.user.username, client_ip=client_ip, comment=comment)

          results_key = conn.query_results_key(username=request.user.username,
            query=build_request.query, dsreq=dsreq, results=explain_results, client_ip=client_ip)

          profile.mark('cache.recheck')
          if results_key is not None:
            if results_key == tables_text:
              tsput(cache_key, results_key)
              ql.logQueryEnd(cache_check=True)
              profile.mark('cache.tsupdate.1')

              profile.markFromStart('query_runtime')

              if csvdownload:
                return {
                  'response': _jsonToCsvResponse(
                    cached_val,
                    language_code,
                    quotation_mark,
                    filename=csvfilename,
                    format_settings=format_settings,
                    build_request=build_request
                    )
                }
              if xlsxdownload:
                return {
                  'response': getXlsxResponse(
                    cached_data=cached_val,
                    filename=xlsxfilename,
                    format_settings=format_settings)
                }

              return {
                'data': cached_val,
                'profile': profile.results()
              }
        except Exception as e:
          explain_results = 'Error'
          LOGGER.info("Cache validation check failed: %s (%s)" % (build_request.query, e))

  if ARCENGINE_VALIDATE_ACCESS and dsreq_obj and \
    dsreq_obj.get('type') in _VALIDATE_CMDS and \
    explain_results != 'Error' and \
    connectionIsArcEngineBased(conn):

    try:
      # Get explain results separately so we can also use for fallback check
      if explain_results is None:
        explain_results = conn.log_sync_runquery(query='EXPLAIN ' + build_request.query,
            username=request.user.username, client_ip=client_ip, comment=comment)

      tablerefs = conn.query_table_refs(
          username=request.user.username, query=build_request.query,
          results=explain_results, client_ip=client_ip)

      profile.mark('sqlinj.getinfo')
      # Convert the result of the explain into a list of tables
      # Handle case of dataset on an empty table (it doesn't show up
      # in the info list
      if tablerefs is None:
        queryTables = set()
      else:
        queryTables = tablerefs

      # Get the list of tables from the dataset
      mds = ModelDataset(django_ds.dataset_info, conn.properties())
      datasetTables = set([x.tablename for x in mds.tables])

      # Verify that the tables in the query are a subset of tables in the dataset
      tablesInQueryNotInDataset = queryTables - datasetTables
      profile.mark('sqlinj.complete')
      if len(tablesInQueryNotInDataset) > 0:
        LOGGER.info('Tables outside of dataset used: (%s) (%s)' % (queryTables, datasetTables))
        raise ValidationError('Tables outside of dataset used: %s' % \
                              ','.join(tablesInQueryNotInDataset))
    except Exception as e:
      explain_results = 'Error'
      LOGGER.info("Arcengine injection validation check failed: %s (%s)" % (build_request.query, e))

  # Here for Arcengine, we check to see if we should fall back to
  # another connection:
  fallback_dcid = None
  if dsreq_obj and dsreq_obj.get('type') == NORM_CMDS['SQL'] and \
     explain_results != 'Error' and \
     connectionIsArcEngineBased(conn) and no_fallback is False:
    fallback_dcid = django_dc.fallback_settings().get('CONNECTION_ID')
    if fallback_dcid:
      try:
        # Get explain results if we don't have them yet
        if explain_results is None:
          profile.mark('fallback_explain')
          explain_results = conn.log_sync_runquery(query='EXPLAIN ' + build_request.query,
              username=request.user.username, client_ip=client_ip, comment=comment)

        profile.mark('fallback_check')
        explrows = explain_results.get('rows', [['None']])
        aview_used = False
        for r in explrows:
          if r[0] == 'REWR :':
            aview_used = True
            break

        # Ensure the fallback connection is valid
        try:
          fallback_dc = DataConnection.objects.get(id=fallback_dcid)
        except:
          profile.mark('fallback conn %s not valid' % fallback_dcid)
          aview_used = False

        if aview_used:
          fallback_dcid = None
        else:
          profile.mark('fallback to connection id %s (%s)' % (fallback_dcid, fallback_dc.dataconnection_name))

          # Ensure we have the results key now because it won't be available via
          # the fallback connection results.
          if results_key is None:
            results_key = conn.query_results_key(username=request.user.username,
              query=build_request.query, dsreq=dsreq, results=explain_results, client_ip=client_ip)

          # Create a new build request based on the fallback connection.  Note that the cache key still
          # corresponds to the query as built for the arcengine connection but should be ok.
          dataconnection_id = fallback_dcid
          conn = get_dataconn(fallback_dcid)

          # TODO -- Knowledge of connection can deal with comment should be in builder not client
          if not connectionSupportsComments(conn):
            comment = ''

          dataset_request = DatasetRequest(dsreq, None, params=viz_params)
          build_request = conn.build(request.user.username, request.user.id, query,
                                     dataset_request=dataset_request,
                                     sample_pct=sample_pct,
                                     comment=comment,
                                     qcntl=qcntl)

      except Exception as e:
        explain_results = 'Error'
        LOGGER.info("Arcengine fallback check failed: %s (%s)" % (build_request.query, e))

  # Here is where we fill in the query control structure to contain information to fill in
  # the new style extended metadata request structure
  if arcengine_extmeta:
    if dsreq_obj and dsreq_obj.get('type') == NORM_CMDS['SQL'] and no_fallback is False:
      fallback_dcid = django_dc.fallback_settings().get('CONNECTION_ID')
    # Currently can't support fallback with streaming download
    if csvdownload:
      fallback_dcid = None

    qcntl.system_functions_allowed = getattr(settings, 'ARCENGINE_SYSTEM_FUNCTIONS_ALLOWED', True)
    qcntl.fallback_dcid = fallback_dcid
    qcntl.tables_text = tables_text
    if tables_text:
      profile.mark('cache.recheck')

    # Get the list of tables from the dataset
    if ARCENGINE_VALIDATE_ACCESS and dsreq_obj and dsreq_obj.get('type') in _VALIDATE_CMDS and dataset_request:
      # Tables test not valid if any table is 'table from SQL'
      if not any(x.select for x in dataset_request.dataset.tables):
        qcntl.ds_tables = set([x.tablename for x in dataset_request.dataset.tables if not x.fromView])

  streamCSV = False
  streamXlsx = False
  if connectionSupportsStreaming(conn):
    if csvdownload:
      streamCSV = True
    if xlsxdownload:
      streamXlsx = True

  # EXECUTE
  if streamCSV:
    try:
      response = StreamingHttpResponse(
        _csvStream(
          conn,
          build_request,
          async_flag,
          language_code,
          quotation_mark,
          format_settings=json.loads(params.get('formattingSettings', '[]')),
          cache_key=cache_key,
        ),
        content_type="text/csv"
      )
      response['Content-Disposition'] = 'attachment; filename="%s"' % csvfilename
      profile.markFromStart('query_runtime')
      return {
        'response': response
      }
    except Exception as e:
      return {
        'error': str(e)
      }

  elif xlsxdownload:
    profile.markFromStart('query_runtime')
    try:
      return {
        'response': getXlsxResponse(
          conn,
          build_request,
          filename=xlsxfilename,
          async_flag=async_flag,
          stream=streamXlsx,
          format_settings=json.loads(params.get('formattingSettings', '[]')),
          cache_key=cache_key,
        ),
      }
    except Exception as e:
      return {
        'error': str(e)
      }
  else:
    profile.mark('execute.pre')
    try:
      val = conn.execute(build_request, async_flag=async_flag)
      CacheStatistics.updateQueryStatisticHistory(build_request.connection.connection_id, False)
      profile.mark('execute.post')

      # New style arcengine extmeta raises exception on fallback so in that case
      # for no exception, we clear out the fallback_dcid
      if arcengine_extmeta:
        fallback_dcid = None

    except Exception as e:
      profile.mark('execute.done')

      # Here we check for special exception conditions returned from arcengine
      if qcntl.exec_method == 'DJANGO_CACHE':
        ql.logQueryEnd(cache_check=True)
        tsput(cache_key, tables_text)
        profile.mark('cache.tsupdate.2')
        return {
          'data': cached_val,
          'profile': profile.results()
        }

      if qcntl.exec_method == 'FALLBACK':
        fallback_dc = DataConnection.objects.get(id=fallback_dcid)
        profile.mark('fallback to connection id %s (%s)' % (fallback_dcid, fallback_dc.dataconnection_name))

        # Create a new build request based on the fallback connection.  Note that we pick up the
        # results_key from the tables_text returned by arcengine for use in storing the results
        # that we get from the fallback engine.
        results_key = qcntl.tables_text
        dataconnection_id = fallback_dcid
        conn = get_dataconn(fallback_dcid, django_dc=fallback_dc)

        # Fallback runs with new query control to reset all features.  Only new feature
        # enabled on fallback path is to apply the FA in the engine if the fallback
        # engine is also arcengine (which seems weird).
        qcntl = QueryControl()
        if connectionSupportsExtMetadata(conn):
          qcntl.fa_engine = True

        # TODO -- Knowledge of connection can deal with comment should be in builder not client
        if not connectionSupportsComments(conn):
          comment = ''

        dataset_request = DatasetRequest(dsreq, None, params=viz_params)
        build_request = conn.build(request.user.username, request.user.id, query,
                                   dataset_request=dataset_request,
                                   sample_pct=sample_pct,
                                   comment=comment,
                                   qcntl=qcntl,
                                   profile=profile)
        try:
          val = conn.execute(build_request, async_flag=async_flag)
          CacheStatistics.updateQueryStatisticHistory(build_request.connection.connection_id, False)
        except Exception as e:
          ql.logQueryEnd(error=True, error_detail=str(e))
          LOGGER.exception(e)

          return {
            'error': str(e)
          }

      else:
        ql.logQueryEnd(error=True, error_detail=str(e))

        return {
          'error': str(e)
        }

  if val:
    # Clear out the results_key if we don't need to preserve it from the fallback check
    if fallback_dcid is None:
      results_key = None

    if async_flag:
      if 'qhandle' in val:
        qhandle = val['qhandle']

        # update tracking id from action_id to qhandle
        ql.untrack(ql.action_id)
        ql.track(qhandle)
        profile.mark('execute.handle')

        add_handle_data(qhandle, {
          'username': request.user.username,
          'connection_id': dataconnection_id,
          'cache_key': cache_key,
          'results_key': results_key,
        })

        return {
          'data': val,
          'profile': profile.results(),
          'qhandle': qhandleAddPrefix(qhandle),
        }
      else:
        ql.logQueryEnd(dataconnection_fbid=fallback_dcid)
    else:
      ql.logQueryEnd(dataconnection_fbid=fallback_dcid)
      format_settings = json.loads(params.get('formattingSettings', '[]'))
      if val and val.get('rows') and format_settings:
        if isinstance(val.get('rows'), str):
          rows_data = json.loads(val['rows'])
        elif isinstance(val.get('rows'), (list, dict)):
          rows_data = val['rows']
        rows = _format_values(
          rows_data,
          format_settings,
        )
        val['rows'] = json.dumps(rows)
      cache.set(cache_key, val, None)
      profile.mark('cache.post')

      if results_key is None:
        results_key = conn.query_results_key(username=request.user.username,
            results=val, client_ip=client_ip)
      tsput(cache_key, results_key or 'NOT SUPPORTED')
      profile.mark('cache.tsupdate.3')

    profile.markFromStart('query_runtime')

    if csvdownload:
      return {
        'response': _jsonToCsvResponse(
          val,
          language_code,
          quotation_mark,
          filename=csvfilename,
          format_settings=json.loads(params.get('formattingSettings', '[]')),
          build_request=build_request
        )
      }

    return {
      'data': val,
      'profile': profile.results()
    }

  return {
    'error': 'No response'
  }

def _format_value(value, format_setting):
  if format_setting is None:
    return value
  formatted = value
  format_type = format_setting.get('type')
  if (format_type not in ['js', 'date']):
    try:
      format_spec = format_setting.get('result')

      if format_type == 'currency':
        format_spec = format_spec.replace(format_setting.get('currency'), '')

      if (type(value) == str and value.isnumeric()):
        value = float(value)
        if (value.is_integer()):
          value = int(value)

      formatted = format(value, format_spec)

      if format_type == 'currency':
        formatted = format_setting.get('currency') + formatted
    except:
      return value
  return formatted

def _format_values(rows, format_settings):
  if (format_settings is not None and len(format_settings) > 0):
    for row in rows:
      for i in range(len(row)):
        format_setting = format_settings[i]
        row[i] = _format_value(row[i], format_setting)
  return rows

def qhandleAddPrefix(qhandle):
  return 'p' + str(settings.PERSISTENT_PROCESS_ID) + '.' + qhandle

def qhandleRemovePrefix(qhandle):
  if qhandle.startswith('p'):
    i = qhandle.find(".")
    if i > 0:
      return qhandle[i+1:]

  return qhandle

def parse_queries_active():
  """
  parse queries from arcengine logs
  """

  data = urllib.request.urlopen(
    "http://localhost:25000/queries").read()

  s = BeautifulSoup(data)

  queryid_re = re.compile("query_id=(.+?)\"")

  tables = s.find_all("table")
  if len(tables) != 3:
    raise Exception("Invalid data received")

  now = datetime.datetime.now()
  for row in tables[0].find_all("tr"):
    cols = row.find_all("td")
    if len(cols) != 10:
      continue
    profile = str(cols[-1])
    m = queryid_re.search(profile)
    if m:
      queryid = m.group(1)
    else:
      queryid = ""
    cols = [x.get_text() for x in cols]
    user, db, stmt, query_type, start_time, progress, state, \
    unused_rowcount, unused_profile, unused_action = cols
    idx = start_time.find(".")
    if idx > 0:
      start_time = start_time[:idx]
    start_dt = datetime.datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")
    tdelta = now - start_dt
    yield queryid, user, db, stmt, query_type, start_time.rstrip("."), tdelta, \
        progress, state

# Helper function to return results of async status/cancel/results obfuscated if needed
def _format_results(data):
  s = json.dumps(data)
  if getattr(settings, 'CLEARTEXT_DSREQ', True):
    return HttpResponse(s, content_type='application/json')

  s = obfuscate(s)
  return HttpResponse(s, content_type='application/octet-stream')

@login_required_401
def asynccancel(request):
  """Cancel async queries"""
  qhandles = request.POST.get("qhandles", '')
  if qhandles:
    qhandles = [qhandleRemovePrefix(x) for x in qhandles.split(',')]
    responsel = []

    for qhandle in qhandles:
      t1 = time.time()
      (data, status) = _asynccancel_single(qhandle, request.user.username)
      t2 = time.time()
      LOGGER.info('_asynccancel_single(%s): %s', qhandle, t2 - t1)
      if status == 0:
        responsel.append({
          'response': data
        })
      else:
        responsel.append({
          'error': data
        })
      mark_query_canceled(qhandle)

    return _format_results(responsel)

  return HttpResponse(data, status=500)

def _asynccancel_single(qhandle, username):
  try:
    out = asyncquery_cancel(qhandle, username)
    return (out, 0)
  except Exception:
    return ('Could not cancel %s' % qhandle, 500)

@login_required_401
def asyncstatus(request):
  """Get the status of an async query"""
  client_ip = get_request_source_ip(request)
  username = request.user.username

  qhandles = request.POST.get("qhandles", '')
  if qhandles:
    qhandles = [qhandleRemovePrefix(x) for x in qhandles.split(',')]
    (data, status) = asyncstatus_work(qhandles, username, client_ip)
    return StreamingResponseIfPossible(data)

  return HttpResponse('must specify qhandles', status=500)

def asyncstatus_work(qhandles, username, client_ip, inp_profile=None):
  responsel = []

  for qhandle in qhandles:
    # If we don't have an input profile, then use a new profile for each qhandle
    profile = inp_profile or ArcProf('asyncstatus', username=username)
    profile.mark('status.start')
    (data, status) = _asyncstatus_single(qhandle, username=username, profile=profile)
    profile.mark('status.done')

    # Handle query completion (status == 1) by returning the results
    if status == 1:
      profile.mark('results.start')
      (data, status) = _asyncresults_single(qhandle, username, client_ip,
          profile=profile)
      profile.mark('results.done')

      if status == 0:
        responsel.append({
          'response': data,
          'profile': profile.results(),
          'results': 'true',
        })
      else:
        responsel.append({
          'error': data,
          'profile': profile.results(),
        })
      del data
      asyncrunner_stop_tracking(qhandle, username)

    # (status = 3) is for error return (no results are available)
    elif status == 3:
      responsel.append({
          'error': data.get('error', 'Unknown error response'),
          'profile': profile.results(),
        })

    else:
      # A status of 2 is only returned for KSQL connections, and indicates
      # that all the data in the stream has been consumed successfully
      if status == 2:
        responsel.append({
          'response': {
            'status': 'StreamConsumed',
          },
        })
      elif status == 0:
        responsel.append({
          'response': data
        })
      else:
        responsel.append({
          'error': data
        })

  return (responsel, 0)

def _asyncstatus_single(qhandle, username=None, profile=None):
  profile = profile or ArcProf('asyncstatus_single', username=username)

  try:
    out = asyncquery_status(qhandle, username=username, profile=profile)
    if 'status' in out:
      status = out['status'].split()[0]
      if status == 'StreamConsumed':
        return (None, 2)
      elif status == 'Finished':
        return (None, 1)
      elif status == 'Failed':
        return ({'error': out['status']}, 3)
      elif status == 'Running':
        m = re.search(r"\((\d+)/(\d+) completed\)", out['status'])
        if m is not None:
          return (
              {
                'status': status,
                'progress': m.group(1),
                'total': m.group(2),
              }, 0)

        if out.get('chunk_text'):
          return ({'status': status, 'chunk_text': out.get('chunk_text'), 'type': 'async'}, 0)

      return ({'status': status}, 0)

    return (out, 500)
  except Exception as e:
    LOGGER.exception('Async status error')
    return ("<pre>" + str(e) + "</pre>", 500)

@login_required_401
def asyncresults(request):
  """Get the results of an async query"""
  client_ip = get_request_source_ip(request)
  username = request.user.username
  qhandles = request.POST.get("qhandles", '')
  if qhandles:
    responsel = []
    qhandles = [qhandleRemovePrefix(x) for x in qhandles.split(',')]

    for qhandle in qhandles:
      t1 = time.time()
      (data, status) = _asyncresults_single(qhandle, username, client_ip)
      t2 = time.time()
      LOGGER.info('_asyncresults_single(%s): %s', qhandle, t2 - t1)
      if status == 0:
        responsel.append({
          'response': data
        })
      else:
        responsel.append({
          'error': data
        })
      del data
      asyncrunner_stop_tracking(qhandle, username)

    return StreamingResponseIfPossible(responsel)

  return HttpResponse('must specify qhandles', status=500)

def _asyncresults_single(qhandle, username, client_ip, profile=None):

  profile = profile or ArcProf('asyncresults_single', username=username)
  profile.mark('async_results.single')

  ql = querylog_fetch(qhandle)
  rewrite = ''
  fallback_dcid = None
  error_logged = False

  try:
    (val, run_profile) = asyncquery_results(qhandle, username, profile=profile)
    if run_profile and run_profile.firststamp != profile.firststamp:
      for e in run_profile.results():
        profile.record(*e)

    profile.mark('async_results.finished')
    if isinstance(val, str):
      raise Exception(val)

    info = val.get('info', [])
    try:
      for l in info:
        if l.startswith('Rewrite: '):
          rewrite = l[len('Rewrite: '):]
    except:
      pass

    # Cache non-error results
    profile.mark('async_results.cachepre')
    hdata = get_handle_data(qhandle, username)
    if hdata:
      cache_key = hdata['cache_key']
      results_key = hdata['results_key']
      dataconnection_id = hdata['connection_id']
      cache.set(cache_key, val, None)
      profile.mark('async_results.cachepost')

      # Get the results key if we're not preserving the value as obtained
      # from the connection fallback check
      # Otherwise, track that we used the fallback dataconnection id
      if results_key is None:
        conn = get_dataconn(dataconnection_id)
        results_key = conn.query_results_key(username=username,
            results=val, client_ip=client_ip)
      else:
        fallback_dcid = dataconnection_id

      tsput(cache_key, results_key or 'NOT SUPPORTED')
      profile.mark('async_results.cachepostpost')

    return (val, 0)

  except Exception as e:
    LOGGER.exception('Async results error')
    if ql:
      ql.logQueryEnd(rewrite=rewrite, dataconnection_fbid=fallback_dcid, error=True, error_detail=str(e))
      error_logged = True

    return (str(e), 500)

  finally:
    if ql:
      if not error_logged:
        ql.logQueryEnd(rewrite=rewrite, dataconnection_fbid=fallback_dcid)
      ql.untrack(qhandle)

