# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json

from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.http import HttpResponse
from util.decorator_custom import login_required_401

from util.http_async import Async
import vizlib.misc

from . import views_jsonselect
from . import views

from util.perfcounters import APPROX_COUNTERS

import logging

LOGGER = logging.getLogger("arcweb")

ASYNC_CALLS = Async(
    name="http_worker",
    concurrency=settings.HTTP_DATACONNECTION_WORKER_THREADCOUNT,
    buckets = [1, 5, 10, 30, 60, 120, 180, 240, 300, 301],
    optimistic_return_check_seconds=0.1)

jsonselect_counter = APPROX_COUNTERS.get('async.jsonselect_parallel_async')
asyncstatus_counter = APPROX_COUNTERS.get('async.asyncstatus_async')
asyncresults_counter = APPROX_COUNTERS.get('async.asyncresults_async')
asynccancel_counter = APPROX_COUNTERS.get('async.asynccancel_async')
asynccancel_qt_counter = APPROX_COUNTERS.get('async.asynccancel_qt')

@login_required_401
@ASYNC_CALLS
def jsonselect_parallel_async(request):
  global jsonselect_counter
  jsonselect_counter += 1

  return views_jsonselect.jsonselect_parallel(request)

@login_required_401
@ASYNC_CALLS
def asyncstatus_async(request):
  global asyncstatus_counter
  asyncstatus_counter += 1

  qhandles = request.POST.get("qhandles", '')
  return views.asyncstatus(request)

@login_required_401
@ASYNC_CALLS
def asyncresults_async(request):
  global asyncresults_counter
  asyncresults_counter += 1

  qhandles = request.POST.get("qhandles", '')
  LOGGER.info("asyncresults_async qhandles: %s", qhandles)

  return views.asyncresults(request)

@login_required_401
@ASYNC_CALLS
def asynccancel_async(request):
  global asynccancel_counter
  asynccancel_counter += 1

  qhandles = request.POST.get("qhandles", '')
  LOGGER.info("asynccancel_async qhandles: %s", qhandles)

  return views.asynccancel(request)

def asynccancel_qt_cancel_all(user, allusers=None):
  connection_ids = set()
  returnData = []
  total = 0

  for connection_id in vizlib.misc.asyncrunner_connection_ids():
    LOGGER.info("asynccancel_qt_cancel_all: connection %s", connection_id)
    cnt = vizlib.misc.asyncrunner_qt_cancel(user=user,
        connection_id=connection_id, allusers=allusers)

    if cnt:
      connection_ids.add(connection_id)
      returnData.append({
        'connection_id': connection_id,
        'cnt': cnt
      });

      total += cnt
      LOGGER.info("Canceled %s queries from connection %s", cnt, connection_id)

  for connection_id in connection_ids:
    vizlib.misc.asyncrunner_reset_pending_work(connection_id)

  LOGGER.info("Canceled a total of %s queries", total)
  return HttpResponse(json.dumps(returnData), content_type="application/json")

@login_required_401
@ASYNC_CALLS
def asynccancel_qt(request, all=None):
  global asynccancel_qt_counter
  asynccancel_qt_counter += 1

  allusers = (all == "all")

  data = json.loads(request.POST.get("data", '[]'))
  LOGGER.info("asynccancel_qt: Canceling %s queries (%s)", len(data), data)

  cancelAll = False

  if len(data) == 1 and \
     data[0].get('connectionId') == '*' and \
     data[0].get('querytracker') == '*':

     cancelAll = True

  connection_ids = set()

  returnData = []

  if cancelAll:
    return asynccancel_qt_cancel_all(request.user, allusers=allusers)

  for d in data:
    connection_id = int(d.get('connectionId', -1))
    if connection_id <= 0:
      continue

    querytracker = d['querytracker']

    LOGGER.info("asynccancel_qt: querytracker [%s]. Canceling", querytracker)
    cnt = vizlib.misc.asyncrunner_qt_cancel(user=request.user,
        connection_id=connection_id, querytracker=querytracker,
        allusers=allusers)

    if cnt:
      returnData.append({
        'connectionId': connection_id,
        'querytracker': querytracker,
        'cnt': cnt
      })

      connection_ids.add(connection_id)

  for connection_id in connection_ids:
    vizlib.misc.asyncrunner_reset_pending_work(connection_id)

  return HttpResponse(json.dumps(returnData), content_type="application/json")
