# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import datetime
import logging
from django.db import transaction
from sqlrun.models import CacheStatisticModel
from django.db.models.aggregates import Sum
from datasets.models import DataConnection
from dataconn import dataconn
from threading import Lock

LOGGER = logging.getLogger("arcweb")

class CacheStatisticHistoryItem:
    def __init__(self, dataconnection_id, returned_from_cache):
        self.dataconnection_id = dataconnection_id
        self.number_of_queries_returned_from_cache = 1 if returned_from_cache else 0
        self.number_of_queries_returned_from_backend = 0 if returned_from_cache else 1

class CacheStatistics:
    bulk_querystatistic_history = {}
    lock = Lock()

    @staticmethod
    def get_by_dataconnection(dataconnection_id):
      dataconnections = DataConnection.objects.filter(id=dataconnection_id)
      if (len(dataconnections)==1):
         dataconnection = dataconnections[0]
      else:
         return None
      try:
        querystatistic, _ = CacheStatisticModel.objects.get_or_create(dataconnection=dataconnection);
        return querystatistic
      except Exception as e:
        LOGGER.warning('Failed to create cache statistic entity: %s', str(e))
        return None

    @staticmethod
    def sum_total_number_of_queries_returned():
      return CacheStatisticModel.objects.all().aggregate(Sum('total_number_of_queries_returned'))['total_number_of_queries_returned__sum']

    @staticmethod
    def sum_number_of_queries_returned_from_cache():
      return CacheStatisticModel.objects.all().aggregate(Sum('number_of_queries_returned_from_cache'))['number_of_queries_returned_from_cache__sum']

    @staticmethod
    def sum_number_of_queries_returned_from_backend():
      return CacheStatisticModel.objects.all().aggregate(Sum('number_of_queries_returned_from_backend'))['number_of_queries_returned_from_backend__sum']

      # keep track of search history in memory. This
      # information will be written into the arcviz database periodically
      # by bulk update thread.
    @classmethod
    def updateQueryStatisticHistory(cls, dataconnection_id, returned_from_cache):
        if dataconnection_id is not None:
          item = cls.bulk_querystatistic_history.get(dataconnection_id);
          with cls.lock:
            if item:
              if returned_from_cache:
                  item.number_of_queries_returned_from_cache += 1
              else:
                  item.number_of_queries_returned_from_backend += 1
              cls.bulk_querystatistic_history[dataconnection_id] = item
            else:
              cls.bulk_querystatistic_history[dataconnection_id] = CacheStatisticHistoryItem(dataconnection_id, returned_from_cache)

    @classmethod
    def commit_changes(cls):
      if cls.bulk_querystatistic_history == {}:
        return
      cls._commit()

    @classmethod
    @transaction.atomic
    def _commit(cls):
      bulk_querystatistic_history = cls.bulk_querystatistic_history
      cls.bulk_querystatistic_history = {}

      for _key, value in bulk_querystatistic_history.items():
        with cls.lock:
          statistic_model = CacheStatistics.get_by_dataconnection(value.dataconnection_id)
          if statistic_model is not None:
            statistic_model.increment_from_cache(value.number_of_queries_returned_from_cache)
            statistic_model.increment_from_backend(value.number_of_queries_returned_from_backend)
            statistic_model.save()
