# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.conf import settings
from util.decorator_custom import login_required_401
from sqlrun import jsonselect_cache
from sqlrun import views
from django.core.exceptions import ValidationError

import vizlib.workq

import threading
import logging
import time

from vizlib.misc import get_request_source_ip
from vizlib.misc import unobfuscate
from vizlib.arcprof import ArcProf
from rbac.permcache import groups_for_user
import json
from util.streaming_json_str import StreamingResponseIfPossible

from sqlrun.views import paramsToVariables, jsonselectRequestIsASync
from dataconn.dataconn_dbapi2 import DbApi2Conn

LOGGER = logging.getLogger('arcweb.views.jsonselect')

_JSONSELECT_RUNNER = None
_JSONSELECT_RUNNER_LOCK = threading.Lock()

def _init_runner():
  # In Ambari installations, we run cherrypy with daemonize=1. That leads to
  # importing this module pre-daemonizing, and then forking without threads.
  # This in turn means we cannot initialize _JSONSELECT_RUNNER at import time
  # and instead have to do it at a later place, protected by _JSONSELECT_RUNNER_LOCK

  global _JSONSELECT_RUNNER

  # This is the concurrency of threads that will be running jsonselect calls
  if _JSONSELECT_RUNNER is None:
    with _JSONSELECT_RUNNER_LOCK:
      if _JSONSELECT_RUNNER is None:
        _JSONSELECT_RUNNER = vizlib.workq.AsyncRunner(
            threadcount=settings.JSONSELECT_THREADCOUNT, name="jsonselect")

class _SyncObject(object):
  def __init__(self, count=None):
    self._count = count
    self._lock = threading.Lock()
    self._cv = threading.Condition(self._lock)

  def decrement(self):
    with self._lock:
      self._count -= 1
      self._cv.notify()

  def waitall(self):
    with self._lock:
      while self._count > 0:
        self._cv.wait()

_json_script_escapes = {
    ord('>'): '\\u003E',
    ord('<'): '\\u003C',
    ord('&'): '\\u0026',
}

def escape_html(str):
  return str.translate(_json_script_escapes)

class JsonSelectRunner(vizlib.workq.ProgressRunner):
  WAIT_LONG = 1
  WAIT_SHORT = 2
  WAIT_NONE = 3

  def __init__(self, request, params=None, sync=None, wait_for_results=None, profile=None):
    self._request = request
    self._params = params
    self._sync = sync
    self._results = ""

    # Used for TriggerJobTarget
    self._wait_for_results = wait_for_results or JsonSelectRunner.WAIT_SHORT

    self._profile = profile or ArcProf('jsonselect.runner', username=request.user.username)
    self._querytracker = self._profile.querytracker()
    self._profile.mark('runner.create')

  def __str__(self):
    return f"{self._request.method} {self._request.path}"

  def __repr__(self):
    return f"{self._request.method} {self._request.path}"

  # Override
  def profile(self):
    return self._profile

  # Override
  def run(self):
    """Start running the task"""

    self._profile.mark('runner.start')
    t1 = time.time()
    jsonselect_cache.jsonselect_set_request(self._request)
    try:
      data = views.jsonselect_work(self._request, self._params, profile=self._profile)

      if 'qhandle' in data:
        client_ip = get_request_source_ip(self._request)
        username = self._request.user.username
        qhandle = data['qhandle']
        self._profile.mark('query.%s' % qhandle)

        # Wait for .2 seconds and get back results if they exist
        if self._wait_for_results == JsonSelectRunner.WAIT_LONG:
          sleep_time = 30 # check every half minute
          loop_times = 240 # run for 2 hours
        elif self._wait_for_results == JsonSelectRunner.WAIT_SHORT:
          sleep_time = .05
          loop_times = 4
        else:
          sleep_time = .05
          loop_times = 0

        for i in range(0, loop_times):
          time.sleep(sleep_time)
          (asyncdata, status) = views.asyncstatus_work([views.qhandleRemovePrefix(qhandle)],
              username, client_ip, inp_profile=self._profile)
          self._profile.mark('status_fetch.mark')

          # If we got the results back, return them.
          # Otherwise, return the original qhandle by falling out of the
          # loop here
          if 'results' in asyncdata[0]:
            self._profile.timestamp('finish')
            self._results = {
              'data': asyncdata[0]['response'],
              'profile': asyncdata[0].get('profile', [])
            }
            return

          if 'error' in asyncdata[0]:
            self._profile.timestamp('finish')
            self._results = {
              'error': asyncdata[0]['error'],
              'profile': asyncdata[0].get('profile', [])
            }
            return

      if 'response' in data:
        self._results = {
          'error': 'Cannot fetch direct responses in batch queries'
        }
        return

      if 'error' in data:
        self._results = {
          'error': escape_html(data['error'])
        }
        return

      if 'data' in data:
        self._profile.timestamp('finish')
        dj = data['data']
        if 'qhandle' in dj:
          dj['qhandle'] = views.qhandleAddPrefix(dj['qhandle'])
          LOGGER.info("Returning qhandle: %s", dj['qhandle'])

        self._results = {
          'data': dj,
          'profile': self._profile.results(),
        }
        return

      self._results = {
        'error': 'Invalid response'
      }

    except ValidationError as e:
      # removing the list from string when there is only one message
      msg = str(e.messages[0]) if len(e.messages) == 1 else str(e)
      LOGGER.exception("Exception on request procesing")
      self._results = {
        'error': 'error: %s' % escape_html(msg)
      }
    except Exception as e:
      LOGGER.exception("Exception on request procesing")
      self._results = {
        'error': 'error: %s' % escape_html(str(e))
      }
    finally:
      self._profile.mark('runner.complete')
      LOGGER.info("querytracker [%s]. JsonSelectRunner complete. duration: %.3f",
          self._querytracker, time.time() - t1)
      jsonselect_cache.jsonselect_clr_request(self._request)
      self._completed = 1
      # No sync in TriggerJobTarget
      if self._sync:
        self._sync.decrement()

  # Override
  def progress(self):
    """Returns a tuple that indicates (steps_completed, total_steps)"""
    return (self._completed, 1)

  # Override
  def results(self):
    """Returns results of the function"""
    return self._results

def prepopulate_cache(datal: list[dict]):
  from datasets.models import DataConnection, DataSet
  LOGGER.info('populating cache')

  dataconnection_ids = set[int]()
  dataset_ids = set[int]()

  for d in datal:
    params = d['params']

    dataconnection_id = params.get("dataconnection_id")
    if dataconnection_id:
      dataconnection_ids.add(dataconnection_id)

    dataset_id = params.get("dataset_id")
    if not dataset_id:
      # need to look into dsreq:
      dsreq_str = params.get('dsreq')
      if dsreq_str:
        dsreq = json.loads(dsreq_str)
        dataset_id = dsreq.get('dataset_id')
    if dataset_id:
      dataset_ids.add(dataset_id)

  ds_list = DataSet.objects.filter(id__in=dataset_ids).prefetch_related('datasetsegment_set__segment', 'filterassociations')
  for ds in ds_list:
    # dataconnection information might not be sent. collect it from the datasets:
    dataconnection_ids.add(ds.dataconnection_id)

    jsonselect_cache.jsonselect_cache_put(('dataset', ds.id), ds)
    jsonselect_cache.jsonselect_cache_put(('falist', ds.id), list(ds.filterassociations.all()))
    for datasetsegment in ds.datasetsegment_set.all():
      jsonselect_cache.jsonselect_cache_put(('segment', datasetsegment.segment.id), datasetsegment.segment)

  dc_list = DataConnection.objects.filter(id__in=dataconnection_ids)
  for dc in dc_list:
    jsonselect_cache.jsonselect_cache_put(('dataconn', dc.id), dc)

  LOGGER.info('cache populated')

@login_required_401
def jsonselect_parallel(request):
  """Run a set of queries and get back results in json"""

  params = request.POST
  data = params.get('data', '[]')

  querytracker = params.get('querytracker')
  if not querytracker:
    querytracker = f"00-{time.time()}"

  t1 = time.time()
  LOGGER.info("querytracker [%s]. jsonselect_parallel_async start", querytracker)

  _init_runner()

  # We know each thread will do permission checks which cause group lookups
  # To avoid a bunch of LDAP lookups running in parallel, here we do a lookup
  # to ensure the thread permission checks will use the cached value
  groups_for_user(request.user)

  # Setup the request level thread shared structures
  jsonselect_cache.jsonselect_init_request(request)

  if not getattr(settings, 'CLEARTEXT_DSREQ', True):
    data = unobfuscate(data)
  datal = json.loads(data)

  # enable caching for the current thread
  jsonselect_cache.jsonselect_set_request(request)
  prepopulate_cache(datal)

  #
  # There are some connection types where we know that queries just get
  #   added to the connection-specific queues. In these cases, it is safe to
  #   simply call jaonselect_work once for each query here. In these cases,
  #   it is also better to call things synchronously so that queries run
  #   in the order they are sent in (important for top to bottom, left to
  #   right scheduling in the presence of concurrency limits)
  #
  # For other connection types, jsonselect_work may itself call into the
  #   backend data connection in order to run explain queries or to schedule
  #   the actual query. In these cases, we need to schedule JsonSelectRunner
  #   threads so we don't block this thread, and all those calls can happen
  #   in parallel.
  #
  # Note that sync queries should always run in parallel threads so we don't block
  # a thread from this threadpool.
  #
  # Deal with both cases below.
  #

  parallelPath = False
  if not parallelPath:
    for d in datal:
      params = d['params']
      if not jsonselectRequestIsASync(params):
        parallelPath = True
        break

      conn = paramsToVariables(params)[0]
      if not isinstance(conn, DbApi2Conn):
        parallelPath = True
        break

  if parallelPath:
    sync = _SyncObject(count=len(datal))

    runnerl = []
    for idx, d in enumerate(datal):
      profile = ArcProf('jsonselect.runner.parallel', querytracker=f"{querytracker}.{idx}",
          username=request.user.username)
      runner = JsonSelectRunner(request=request, params=d['params'], sync=sync,
          wait_for_results=JsonSelectRunner.WAIT_SHORT,
          profile=profile)
      _JSONSELECT_RUNNER.queue_untracked(runner)
      runnerl.append(runner)

    sync.waitall()

  else:
    runnerl = []
    for idx, d in enumerate(datal):
      profile = ArcProf('jsonselect.runner', querytracker=f"{querytracker}.{idx}",
          username=request.user.username)
      runner = JsonSelectRunner(request, params=d['params'],
          wait_for_results=JsonSelectRunner.WAIT_NONE,
          profile=profile)
      runnerl.append(runner)
      runner.run()

  responsel = []
  for runner in runnerl:
    data = runner.results()
    if 'error' in data:
      responsel.append({
        'error': escape_html(data['error']), # escape for XSS
      })
    else:
      responsel.append({
        'response': data['data'],
        'profile': data.get('profile', '')
      })
    if 'data' in data:
      del data['data']
    del data

  del runnerl

  LOGGER.info("querytracker [%s]. jsonselect_parallel_async complete. duration: %.3f",
      querytracker, time.time() - t1)
  return StreamingResponseIfPossible(responsel)
