import $ from 'jquery';
import arc from 'arc';
import QueryColumnManager, {FILTERTYPES} from './QueryColumnManagerClass';
import {PICKLIST_TYPES} from './QueryColumnManagerClass';

/**
 * Picklist Column Manager Class
 *
  This class is used to manage the behavior of a single picklist filter column in the table of recent
  arcengine queries.
 * Properties are the same as the parent Query Column Manager Class, but the structure of data is
 * different for this subclass
 * @property {object} data = {
    exclude: boolean,                 -whether the selected list is of values to be excluded
    values: array of strings,         -list of selected values, internal
    aliases: array of strings,        -list of selected values, user facing
    extraOption: boolean,             -whether a column-specific extra option should be applied
  }

 * @method drawFilterDetails          -Render the picklist filter in the column header and attach
                                       listeners to ensure user changes to the filter are detected
 * @method updateSummary              -Update the summary visible at the top of the picklist filter
                                       even when the dropdown is collapsed
 * @method generateFilterExpression   -Generate a SQL where clause based on the current filter values
 * @method generateFilterQuery        -Generate SQL select expression to retrieve list of options
                                       for the filter
 * @method _setValuesFromPicklistUI   -Based on the currently selected DOM items, generate
                                       a value of the _data property that reflects the current
                                       filter status
 **/

class PicklistColumnManager extends QueryColumnManager {
  constructor(displayName, columnName, config={}) {
    super(displayName, columnName, config);
  }

  /*
    function to actually draw the picklist filter
     internalValues {array}     -array of values taken by the column field
     updateOnly {boolean}       -whether to only re-render the picklist dropdown options
     displayValMapping {object} -Mapping of internal values to user-friendly display values
     $filter {object}           -optional, jQuery object of the filter if rendered previously
   */
  drawFilterDetails(internalValues, updateOnly, displayValMapping, $filter) {
    const self = this;

    function paintPicklist() {
      let hasNoneValue;
      let dataValues;
      const aliasesHash = {};

      if (!updateOnly) {
        $filter = $($('#limited-picklist-template').html());
      } else {
        $filter.find('.picklist-list').remove();
      }

      const $list = $('<ul class="picklist-list"></ul>');

      const excluded = self._data && self._data.exclude;

      if (self._data && self._data.aliases && self._data.values) {
        dataValues = self._data.values;

        self._data.aliases.forEach(function addToHashes(displayVal, i) {
          aliasesHash[displayVal] = true; //use for fast lookup later
          displayValMapping[dataValues[i]] = displayVal;
        });
      }

      if (excluded) {
        $filter.find('.picklist-exclude').prop('checked', true);
      }

      if (!updateOnly) {
        self.updateSummary($filter);
        const optionDisplay = self._extraOption;
        const extraOption = self._data && self._data.extraOption;

        if (optionDisplay) {
          const $optionSel = $filter.find('.extra-option');
          $optionSel.removeClass('arc-invisible').prop('checked', extraOption);
          $optionSel.siblings('.viz-sel-label').text(optionDisplay.label);
        }
      }

      if (self._filterType === FILTERTYPES.PARTIAL_MATCH_PICKLIST) {
        hasNoneValue = internalValues.some(array => !array.length);
        internalValues = internalValues.flat().filter(
          function removeDuplicates(internalValue, index, self) {
            return self.indexOf(internalValue) === index;
        });
      }

      //eslint-disable-next-line no-undef
      internalValues = [...new Set([ ...(dataValues || []), ...internalValues])];
      $filter.find('.count').text(internalValues.length);

      if (hasNoneValue && internalValues.indexOf('None') === -1) {
        internalValues.unshift('None');
      }

      internalValues.forEach(function addSelectOption(internalValue) {
        const label = displayValMapping[internalValue] || internalValue;
        $list.append(arc.misc.createSelectableOption({
          VALUE: internalValue,
          LABEL: label,
          CHECKED: aliasesHash[label],
        }));
      });

      $filter.find('.dropdown-picklist-header').after($list);
      $filter.find('.dropdown-content').removeClass('arc-invisible');
      $filter.find('.fa-spinner').addClass('arc-invisible');

      if (!updateOnly) {
        self._$header.append($filter);
        $filter.offOn('click.filter', function updateFilterVal(evt) {
          if (evt.target.getAttribute('type') === 'checkbox') {
            self._setValuesFromPicklistUI();
            if ($(evt.target).hasClass('picklist-exclude') && !self._data.values) {
              return;
            }

            self.updateSummary($filter);
            $('#query-list-table').trigger({type: 'filterUpdate'});
          }
        });

        $filter.find('.fa-search').offOn('click.search', searchFilterValues);
        $filter.find('.show-all-options').offOn('click.displayHidden', () =>
          $filter.find('.picklist-list').find('li').removeClass('arc-invisible')
        );
        $filter.find('.picklist-searchbox').offOn('keypress.search', function detectEnter(evt) {
          const keycode = (evt.keyCode ? evt.keyCode : evt.which);
          if (keycode === 13) {
            searchFilterValues();
          }
        });
      }
    }

    function searchFilterValues() {
      const searchTerm = $filter.find('.picklist-searchbox').val();
      const $listItems = $filter.find('.picklist-list').find('li');
      $filter.find('.show-all-options').toggleClass('disabled', searchTerm);
      if (!searchTerm) {
        $listItems.removeClass('arc-invisible');
      } else {
        $listItems.each(function toggleVisibility() {
          const $this = $(this); //eslint-disable-line no-invalid-this
          const text = $this.find('.viz-sel-label').text();
          $this.toggleClass('arc-invisible', !text.startsWith(searchTerm));
        });
      }

      $filter.find('.count').text($filter.find('li:visible').length);
    }

    if (PICKLIST_TYPES.includes(this._filterType)) {
      paintPicklist();
    } else {
      return;
    }
  }

  /* function to update summary of picklist status
     $filter {object}           -optional, jQuery object of the filter, only passed in if not yet
                                 attached to DOM
   */
  updateSummary($filter) {
    if (!$filter) {
      $filter =this._$header.find('.selected-control');
    }

    const excluded = this._data.exclude;
    const selections = this._data.aliases || [];
    const selectedCount = selections.length;
    const selectionString = selections.join(', ');
    arc.misc.updatePicklistSummary($filter, {
      individualSelections: selectedCount > 0,
      selectedCount,
      excluded,
      selectionString,
    });
  }

  generateFilterExpression() {
    if (this._filterFn) {
      return this._filterFn(this._columnName, this._data, this._extraOption);
    }

    let sql;
    const values = (this._data.values || []).map(val => `'${val}'`);
    if (this._filterType === FILTERTYPES.PICKLIST) {
      sql = arc.misc.whereClauseFromValues(this._columnName, values,
        this._data.exclude);
      if (this._extraOption && !this._data.extraOption) {
        sql = `${sql} AND (${this._columnName} ${this._extraOption.defaultFilter})`;
      }
    }

    return sql;
  }

  generateFilterQuery() {
    if (PICKLIST_TYPES.includes(this._filterType)) {
      return `SELECT DISTINCT(${this._columnName}) `;
    }
  }

  _setValuesFromPicklistUI() {
    const self = this;
    const data = {};
    self._data = {};

    this._$header.find('.picklist-list').find('input:checked').each(function registerSelection() {
      const $this = $(this); //eslint-disable-line no-invalid-this
      const value = $this.data('value');
      // intentionally not using jquery data to avoid type coercion of numeric strings to numbers
      const alias = $this.attr('data-alias');
      if (data.values) {
        data.values.push(value);
        data.aliases.push(alias);
      } else {
        data.values = [value];
        data.aliases = [alias];
      }
    });

    data.exclude = this._$header.find('.picklist-exclude').prop('checked');
    if (this._extraOption) {
      data.extraOption = this._$header.find('.extra-option').prop('checked');
    }

    self._data = data;
  }

}

export default PicklistColumnManager;