import $ from 'jquery';
import arc from 'arc';

export const FILTERTYPES = {
  PICKLIST: 'picklist',

  // suitable for values that are arrays where looking to match on one array element
  PARTIAL_MATCH_PICKLIST: 'picklist-partial',
  RANGE: 'range',
  DATERANGE: 'daterange',
  NONE: 'none',
};

export const PICKLIST_TYPES = [FILTERTYPES.PICKLIST, FILTERTYPES.PARTIAL_MATCH_PICKLIST];

/* Can be expanded to include other types from https://datatables.net/reference/option/columns.type
 and even more if the appropriate datatables plugin is added -- see
https://datatables.net/plug-ins/sorting/ */
export const COLTYPES = {
  DATETIME: 'date',
  NUMBER: 'num',
  FORMATTED_NUMBER: 'num-fmt',
  HTML: 'html',
  STRING: 'string',
};

export const EXTERNAL_VALUE = 'External';
export const BLANK_USER_DISPLAY = 'unknown';

export const DATETIME_FILTER_TYPES = {
  QUICK: 'quick_filters',
  RELATIVE: 'relative',
  ABSOLUTE: 'absolute',
};

/**
 * Query Column Manager Class
 *
  This class is used to manage the behavior of a single column in the table of recent arcengine
  queries. It is responsible for how the column values display, sort and filter. Some details
  of the filter widget display and behavior are delegated the subclasses Picklist Column Manager and
  Range Column Manager
 * @property {string} _displayName             -Name displayed for column header
 * @property {string} _columnName              -Name of database column used to pull data
 * @property {number} _colIdx                  -Index of database column in server response
 * @property {string} _columnType              -Type of dataTables column, value from COLTYPES,
 * @property {string} _filterType              -Type of filter to display, from FILTERTYPES
 * @property {boolean} _orderable              -Whether to permit sorting on column
 * @property {boolean} _hidden                 -Whether to hide column from user
 * @property {string} _className               -CSS class to apply to td and th elements in column
 * @property {Object} _extraOption = {         -Config to add an additional input to picklist filter
    label: string,                              Label for checkbox for input
    defaultFilter: string,                      SQL filter to append to query unless checkbox checked
   }
 * @property {function} _colParseFn            -optional function to extract column value from
                                                database value, if not present, just use db value as-is
                                                This results in the internal value used for column
                                                sorting
 * @property {function} _displayFn             -optional function to modify the internal value to the
                                                user-facing value that will be seen in filters
 * @property {function} _makeCellContentFn     -optional function to add any needed HTML around the
                                                user-facing value when displaying it in table cell
 * @property {function} _filterFn              -optional custom function to generate SQL where
                                                clause for filter based on current filter state
 * @property {boolean} _skipRetrieval          -Whether values need to actually be retrieved for this
                                                column (in some cases one database column populates
                                                several table columns, so only one needs a retrieval
                                                to occur)
 * @property {object} _data                    -Current filter state, structure varies by subclass
 * @property {object} _defaultFilter           -Default value for data
 * @property {object} _$header                 -jQuery object of th element in column header generated
                                                by dataTables
 * @property {function} _retrievalFn           -function to get column value from the relevant SQL
                                                response row, running colParseFn as necessary
 * @property {boolean} _valuesFetched          -Whether we have an up-to-date list of all possible
                                                filter values

 * @method setValuesFetched                   -set _valuesFetched
 * @method generateColDef                     -use properties to generate a DataTables column
                                               definition
 * @method generateColValueMapping            -Generate hash with internal column values as keys
                                               and user facing values as values
 * @method drawFilter                         -Render filter in header of column
 * @method resetFilter                        -Reset filter to default value or no value at all
 **/

class QueryColumnManager {
  constructor(displayName, columnName, config={}) {
    let initialValue;
    const filterHash = config.filterHash;
    if (filterHash) {
      initialValue = filterHash[displayName];
    }

    this._displayName = displayName; // Name to show the user
    this._columnName = columnName; // Name of corresponding raw column in completed queries table
    if (config.columnType && !Object.values(COLTYPES).includes(config.columnType)) {
      throw new Error('Invalid column type');
    }

    this._columnType = config.columnType; // Column data type  -- must be from COLTYPES enum
    const filterType = config.filterType || FILTERTYPES.PICKLIST;
    if (!Object.values(FILTERTYPES).includes(filterType)) {
      throw new Error('Invalid filter type');
    }

    this._filterType = filterType; // what kind of filter UI is needed
    this._orderable = !!config.orderable; // boolean - do we should support sort on this column
    this._hidden = config.hidden;
    this._className = config.className;
    /* Object for extra checkbox on filter display -- has key label (string of checkbox label) and
      defaultFilter -- Sql filter clause to use if checkbox not checked.
     */
    this._extraOption = config.extraOption;
    /*function to parse data from raw column val if needed - must take a single parameter, the
      raw row from the completed queries table, and return an object with values stored under
      a key that is displayName of the column where they will show on the front end
    */
    this._colParseFn = config.colParseFn;
    this._displayFn = config.displayFn; // function to generate user-facing value
    this._makeCellContentFn = config.makeCellContentFn; //function to generate cell HTML
    this._filterFn = config.filterFn; // custom function to generate sql filter clause
    /* boolean - skip retrieving value because its retrieval is bundled with that of
     another table column that is parsed from the same column of the completed queries table*/
    this._skipRetrieval = config.skipRetrieval;
    this._data = config.defaultFilter || initialValue || {}; // current filter state
    this._defaultFilter = config.defaultFilter;
    this._$header = null; // jquery object of header cell for column
    this._valuesFetched = false;

    const self = this;
    if (this._skipRetrieval) {
      this._retrievalFn = (row) => {}; //eslint-disable-line no unused-vars
    } else if (!this._colParseFn) {
      this._retrievalFn = function addColProperty(rawRow) {
        const data = {};
        data[self._displayName] = rawRow[self._colIdx];
        return data;
      };
    } else {
      this._retrievalFn = function getParsedVal(rawRow) {
        return self._colParseFn(rawRow[self._colIdx]);
      };
    }
  }

  get displayName() {
    return this._displayName;
  }

  get columnName() {
    return this._columnName;
  }

  get data() {
    return this._data;
  }

  get colParseFn() {
    return this._colParseFn;
  }

  get retrievalFn() {
    return this._retrievalFn;
  }

  setColIdx(value) {
    this._colIdx = value;
  }

  setValuesFetched(value) {
    this._valuesFetched = value;
  }

  hasColumnIndex() {
    return this._colIdx !== undefined;
  }

  /* This function generates a column definition for use by the datatables jquery library
  and includes info on how to appropriately display, sort, and filter the column content
  */
  generateColDef() {
    const self = this;
    const colDef = {
      orderable: this._orderable,
      type: this._columnType,
      data: this._displayName,
      title: this._displayName,
      visible: !this._hidden,
      className: this._className,

      //  https://datatables.net/manual/data/orthogonal-data
      render: function colRender(data, type, row) {
        let value = data;
        // If display or filter data is requested, format for display
        if ((type === 'display' || type === 'filter') && self._displayFn) {
          value = self._displayFn(data, row);
        }

        if ((type === 'display') && self._makeCellContentFn) {
          value = self._makeCellContentFn(value, row);
        }

        return value;
      }
    };

    return colDef;
  }

  /* This function generates a map of internal column values for sorting to their user-facing
    display values. It accepts a DataTables values object.
  */
  generateColValueMapping(values) {
    const internalValues = values.sort();
    const displayValueMapping = {};
    if (this._displayFn) {
      values.forEach((val) => displayValueMapping[val] = this._displayFn(val));
    }

    return {
      internalValues,
      displayValueMapping,
    };
  }

  drawFilter($header, internalValues, displayValMapping = {}, updateOnly = false) {
    this._$header = $header;
    const self = this;
    let $filter;
    if (updateOnly) {
      $filter = $header.find('.selected-control');
    } else {
      $header.find('.selected-control').remove(); // remove old filter
    }

    this.drawFilterDetails(internalValues, updateOnly, displayValMapping, $filter);

    $filter = $header.find('.selected-control');
    const $listWrapper = $filter.find('.picklist-list-wrapper');

    addListeners();

    function addListeners() {
      $filter.find('.selected-picklist-link')
        .offOn('click.dropdown', function toggleDropdown(evt) {
          if ($listWrapper.is(':visible')) {
            $listWrapper.hide();
          } else {
            $listWrapper.show();
            $header.siblings().each(function hideOtherDropdowns() {
              $(this).find('.picklist-list-wrapper').hide(); //eslint-disable-line no-invalid-this
          });

          const hasRangeWidget = !!$filter.find('.filter-range-widget').length;
          const hasPicklistValues = !!$filter.find('.picklist-list').find('li').length;

          if (self._data && !self._valuesFetched && (self._data.values && !self._data.exclude) ||
            (self._data.min && self._data.max && self._filterType === FILTERTYPES.RANGE)) {
            $listWrapper.find('.dropdown-content').addClass('arc-invisible');
            $listWrapper.find('.fa-spinner').removeClass('arc-invisible');
            self._$header.trigger('filterValuesFetch');
          } else if (self._sliderConfig && !hasRangeWidget) {
            self.paintRangeFilter($filter);
          } else if (self._filterType === FILTERTYPES.RANGE && !hasRangeWidget ||
            (PICKLIST_TYPES.includes(self._filterType) && !hasPicklistValues)) { //initial load
            $listWrapper.find('.dropdown-content').addClass('arc-invisible');
            $listWrapper.find('.fa-spinner').removeClass('arc-invisible');
          }
        }

        evt.stopPropagation();
      });
      
      if ($listWrapper.length) {
        //eslint-disable-next-line no-undef
        $(document).offOn(`click.close-${self._displayName}`, function hideFilter(evt) {
          if (!$.contains($listWrapper[0], evt.target)) { //clicking off dropdown should close it
            const $target = $(evt.target);                // unless click falls in datepicker modal
            if ($target.closest('.calendar-date').length ||
              $target.closest('.daterangepicker').length) {
              return;
            }

            $listWrapper.hide();
          }
        });
      }


      $filter.find('.clear-filter').offOn('click.clear', () => self.resetFilter());
    }

  }

  resetFilter(trigger = true) {
    const newVal = this._defaultFilter || {};
    if (arc.misc.isIdentical(this._data, newVal)) {
      return;
    }

    this._data = newVal;
    if (PICKLIST_TYPES.includes(this._filterType)) {
      this._$header.find('input:checked').each(function uncheckCheckbox() {
        $(this).prop('checked', false); //eslint-disable-line no-invalid-this
      });
    } else if (this._filterType === FILTERTYPES.DATERANGE) {
      this.showDateFilterWithDefaults(DATETIME_FILTER_TYPES.QUICK, true);
    }

    this.updateSummary();

    if (trigger) {
     $('#query-list-table').trigger({type: 'filterUpdate'});
    }
  }

  drawFilterDetails() {
    console.log('drawFilterDetails not implemented on subclass of QueryColumnManager');
  }

  updateSummary() {
    console.log('updateSummary not implemented on subclass of QueryColumnManager');
  }

  generateFilterExpression() {
    console.log('generateFilterExpression not implemented on subclass of QueryColumnManager');
  }

  generateFilterQuery() {
    console.log('generateFilterQuery not implemented on subclass of QueryColumnManager');
  }

  paintRangeFilter() {
    console.log('paintRangeFilter not implemented on subclass of QueryColumnManager');
  }

  showDateFilterWithDefaults() {
    console.log('showDateFilterWithDefaults not implemented on subclass of QueryColumnManager');
  }
}

export default QueryColumnManager;

