import arc from 'arc';
import $ from 'jquery';
import PicklistColumnManager from './PicklistColumnManagerClass';
import RangeColumnManager from './RangeColumnManagerClass';
import {EXTERNAL_VALUE, FILTERTYPES, COLTYPES, DATETIME_FILTER_TYPES,
  BLANK_USER_DISPLAY} from './QueryColumnManagerClass';

const HISTORY_TABLE = 'arcadia_catalog.completed_queries';
/**
 * Query Table Manager Class
 *
    This class is used to manage the table of recent arcengine queries that is shown as part of
    the 'Smart Acceleration' workflow. It generates a request to retrieve the list of queries
    to populate the table, and when the results return it parses them and populates the table. It
    also submits the required request once the user has chosen which queries to accelerate.

    Individual column level logic, like display formatting, special sorting, etc. is handled by
    the QueryColumnManager class. A QueryTableManager maintains an array of QueryColumnManagers, one
    for each column, that references to handle that logic.
 * @property {string} _dcId               -the id of the data connection for which we are retrieving
                                           queries
 * @property {object} _$elem              -jQuery element where the table of queries is to be rendered
 * @property {object} _selectedRowHash = {
    queryId {string}: boolean             -Object telling which queries should show as selected
   },                                      on initial render (relevant if reloading a stored state)
 * @property {object} _datasetPerms = {
    datasetId {string}: boolean           -Object storing whether user has app_view permission
   },                                      each dataset (controls whether db/visual links shown)
 * @property {string} _filterLimit        -Maximum number of queries to retrieve at at time
 * @property {object} _savedFilterData = {
    displayName {string}: data {object}   -displayName property of associated Query Column Manager
   }                                       and saved data value (see that class for structure)
 * @property {array} _columnMgrs          -array of Query Column Manager objects to handle columns
 * @property {object} _queryTable         -dataTable object that controls the table rendering and
                                           behavior
 * @method _submitQueriesForAcceleration  -Submits post request to server with list of queries for
                                           acceleration and on response navigates to acceleration
                                           status page
 * @method _initializeTable               -Helper function to constructor, creates dataTable
 * @method _setUpColumns                  -Helper function to constructor, initiates Query Column
                                           Managers that are populated into _columnMgrs, giving
                                           each the column-specific display and filter logic
 * @method _setUpListeners                -Helper function to constructor, adds listeners to table
 * @method _parseQueries                  -Receives list of queries from server and parses them
                                           for population into table
 * @method _buildQuery                    -Constructs SQL query that will populate table
 * @method _buildFilterValueQuery         -Constructs SQL query that will populate potential filter
                                           values for a particular column filter
 * @method _resetColumnFilters            -Reset column-level filters for every column in the table
 **/
class QueryTableManager {
  constructor(dcId, $elem, state, perms) {
    this._dcId = dcId;
    this._$elem = $elem; //jQuery element
    this._selectedRowHash = {}; //hash with row ids for selected rows as keys, boolean value
    this._datasetPerms = JSON.parse(perms); //hash with dataset ids as keys, whether user has app_view
                                            // permission on them as the value
    this._filterLimit = '5000';

    let parsedState;
    let filterData;
    if (state) {
      parsedState = JSON.parse(state);
      filterData = parsedState.filters;
      this._filterLimit = parsedState.filterLimit;
      this._selectedRowHash = parsedState.selected;
      if (filterData) {
        this._savedFilterData = filterData;
      }
    }

    $('#limit-returned-queries').val(this._filterLimit);

    this._setUpColumns();

    this._initializeTable(function savedState() {
      if (parsedState) {
        return parsedState.dataTables;
      }

      return null;
    });
  }

  _initializeTable(stateLoadCb) {
    const self = this;

    $.fn.dataTable.ext.order['dom-checkbox'] = function orderByCheckboxState(settings, col) {
      return this.api().column( col, {order:'index'} ).nodes().map(
        (td) => $('input', td).prop('checked') ? '1' : '0'
      );
    };

    let columnDefs = [{
      orderDataType: 'dom-checkbox',
      defaultContent: '<input type="checkbox" class="row-select">',
    }];
    columnDefs = columnDefs.concat(self._columnMgrs.map((mgr) => mgr.generateColDef()));

    this._queryTable = this._$elem.DataTable({
      lengthChange: false,
      pageLength: 100,
      dom: 'ript',
      stateSave: true,
      stateLoadCallback: stateLoadCb,
      order: [[1, 'desc']],
      select: { style: 'api' },
      language: {
        loadingRecords: '',
        emptyTable: `No queries are available on this connection \
                     in the selected time period`,
      },
      rowId: 'Query Details',
      ajax: {
        url: arc.frontend.globals.urls.sqlRunUrl,
        type: 'POST',
        beforeSend(xhr) {
          xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
            $('.loading-shield').addClass('loading');
            $('#query-list-table_wrapper tbody').addClass('blur');
        },
        data: function getRequestParams() {
          return {
            dataconnection_id: self._dcId,
            query: self._buildQuery(),
            cache_read: false,
          };
        },
        dataSrc: (response) => self._parseQueries(response),
        error: function handleError(xhr, error) {
          if (error !== 'abort') { // Don't display error when user intentionally canceled
            arc.report.alertMessage(`Failed to fetch query list for data connection ${self._dcId}`,
              error);
          }

          $('.loading-shield').removeClass('loading');
          $('#query-list-table_wrapper tbody').removeClass('blur');
        },
      },
      columns: columnDefs,
      drawCallback: function paintFilters() {
        const $header = $(this.api().table().header());
        const colCount = this.api().columns(':visible').count();
        if ($header.find('tr').length === 1) {
          const row = arc.misc.createFragmentElement({ tagName: 'tr' });
          for (let i = 0; i < colCount; i++) {
            const config = { tagName: 'th' };
            if (i === 5 || i === 9) {
              config.classes = ['left-border'];
            } else if (i === 6) {
              config.classes = ['nowrap'];
            } else if (i === 11) {
              config.classes = ['drop-left'];
            }

            const e = arc.misc.createFragmentElement(config);
            if (i === 0) {
              e.innerHTML = '<input type="checkbox" class="allrow-select">';
              row.appendChild(e);
            } else {
              row.appendChild(e);
            }
          }

          $header.append($(row));
        }

        const rowCount = this.api().rows().count();
        if (rowCount === Number(self._filterLimit)) {
          arc.report.alertMessageSmall(`Display limited to ${rowCount} entries`, 'alert-warning');
        }

        const $filterCells = self._$elem.find('thead tr:last-of-type').find('th');
        $('.loading-shield').removeClass('loading');
        $('#query-list-table_wrapper tbody').removeClass('blur');
        self._columnMgrs.forEach((colMgr) => colMgr.setValuesFetched(false));

        let visibleColIndex = 0;
        this.api()
          .columns()
          // eslint-disable-next-line array-callback-return
          .every(function paintFilter(index) {
            if (index === 0) { // doesn't apply for checkbox column
              return;
            }

            if (this.visible()) {
              visibleColIndex += 1;
              const colMgr = self._columnMgrs[index - 1];
              const values = this.data().unique();
              const {
                internalValues,
                displayValueMapping
              } = colMgr.generateColValueMapping(values.toArray());
              const $cell = $($filterCells[visibleColIndex]);
              $cell.data('colMgrIdx', index - 1);
              colMgr.drawFilter($cell, internalValues, displayValueMapping);
            }
          });

        this.api()
            .rows(function isSelected(idx, data) {
              return !!self._selectedRowHash[data['Query Details']];
            })
            // eslint-disable-next-line array-callback-return
            .every(function checkBox() {
              const $row = $(this.node());
              $row.find('input').prop('checked', true);
            })
            .select();
      },
      createdRow: (row, data) => $(row).data('query', data),
    });

    this._setUpListeners();
  }

  _parseQueries(response) {
    const queries = [];
    const self = this;
    const colnames = response.colnames;
    if (this._columnMgrs[0] && !this._columnMgrs[0].hasColumnIndex()) {
      this._columnMgrs.forEach(
        columnMgr => columnMgr.setColIdx(colnames.indexOf(columnMgr.columnName))
      );
    }

    JSON.parse(response.rows).forEach(function parseRawRow(rawRow) {
      let queryData = {};
      self._columnMgrs.forEach(function addColData(columnMgr) {
        queryData = Object.assign(queryData, columnMgr.retrievalFn(rawRow));
      });

      queries.push(queryData);
    });

    return queries;
  }

  _buildQuery() {
    let filters = this._columnMgrs.map(colMgr => colMgr.generateFilterExpression())
                                  .filter(expr => expr); //filter out empty strings
    filters = filters.length ? 'AND '.concat(filters.join(' AND ')) : '';

    const query = `SELECT * FROM ${HISTORY_TABLE}
                WHERE (query_type = 'QUERY' ${filters})
                LIMIT ${this._filterLimit}`;
    return query;
  }

  _setUpColumns() {
    const columnMgrs = [];
    const btnHtml = $('#query-detail-btn').html();
    const self = this;

    function parseTableList(listStr) {
      const listStrPattern = /^\[(.+)\]$/;
      if (listStr && listStr.match(listStrPattern)) {
        return listStr.match(listStrPattern)[1].split(', ');
      } else {
        return [];
      }
    }

    function displayAVs(val, row) { //eslint-disable-line no-unused-vars
      if (!val.length || !Array.isArray(val)) {
        return '<span class="text-faint">None</span>';
      } else {
        const str = val.map(avName => arc.clone_template('av-with-detail-link', { avName }).html())
                       .join('<br>');
        return str;
      }
    }

    // creates where clause looking for entries in string field formatted like [val1,val2,val3]
    // accepts name of column and data from a
    function whereClauseForTableListField(colName, data, extraOption) {
      let extraOptionClause;
      if (extraOption) {
        extraOptionClause = `${colName} ${extraOption.defaultFilter}`;
      }

      if (!data.values) {
        if (extraOption && !data.extraOption) {
          return extraOptionClause;
        }

        return '';
      }

      let sql = colName;
      let includeNone;
      let excludeNone;
      let noneSQLClause;
      const values = data.values.slice();
      const noneIndex = values.indexOf('None');
      if (noneIndex !== -1) {
        values.splice(noneIndex, 1);
        if (data.exclude) {
          excludeNone = true;
          noneSQLClause = `${colName} != '[]'`;
        } else {
          includeNone = true;
          noneSQLClause = `${colName} = '[]'`;
        }
      }

      if (data.exclude) {
        sql = sql.concat(' NOT');
      }

      if (values.length) {
        sql = sql.concat(` RLIKE '[,\\[]${values.join('|')}[],]'`);
        if (excludeNone) {
          sql = `((${sql}) AND (${noneSQLClause}))`;
        } else if (includeNone) {
          sql = `((${sql}) OR (${noneSQLClause}))`;
        }
      } else {
        sql = noneSQLClause;
      }

      if (extraOption && !data.extraOption) {
        sql = `${sql} AND (${extraOptionClause})`;
      }

      return sql;
    }

    function whereClauseForArcvizAsset(assetType, data) {
      if (!data.aliases) {
        return '';
      }

      const aliases = data.aliases.slice();
      let includeExternal;
      let excludeExternal;
      let externalSQLClause = '';

      const colName = 'sql_statement';
      const assetTypeHash = {
        dataset: 'ds',
        visual: 'widget',
        dashboard: 'app',
      };

      let sql = colName;
      if (data.exclude) {
        sql = sql.concat(' NOT');
      }

      const externalIndex = aliases.indexOf(EXTERNAL_VALUE);
      if (externalIndex !== -1) {
        if (data.exclude) {
          excludeExternal = true;
          externalSQLClause = `${colName} RLIKE '^/\\\\* widget='`; //look for arcviz comment
        } else {
          includeExternal = true;
          externalSQLClause = `${colName} RLIKE '^[^/]'`; // Don't expect starting comment
        }

        aliases.splice(externalIndex, 1);
      }

      if (aliases.length) {
        sql = sql.concat(` RLIKE '${assetTypeHash[assetType]}=(${aliases.join('|')})'`);
        if (excludeExternal) {
          sql = `((${sql}) AND (${externalSQLClause}))`;
        } else if (includeExternal) {
          sql = `((${sql}) OR (${externalSQLClause}))`;
        }

        return sql;
      }

      return externalSQLClause;
    }

    function displayDbOrVisualLink(id, row) {
      if (id !== 'None' && id !== EXTERNAL_VALUE && self._datasetPerms[row.Dataset]) {
        const url = arc.misc.getBlankTargetUrl(`/arc/apps/app/${id}`);
        return `<a href="${url}" arcperm="ds_appview" target="_blank">${id}</a>`;
      }

      return id;
    }

    function extractAssetIds(sqlString) {
      const dashboardPattern = /^\/\* widget=(\w+), app=(\w+), ds=(\w+) \*\/ \n([\s\S]*)/;
      const matches = sqlString.match(dashboardPattern);

      function getAssetIdMatch(matches, matchIdx) {
        const matchVal = matches[matchIdx];
        return  $.isNumeric(matchVal) ? Number(matchVal) : 0;
      }

      if (matches) {
        return {
          Visual: getAssetIdMatch(matches, 1),
          Dashboard: getAssetIdMatch(matches, 2),
          Dataset: getAssetIdMatch(matches, 3),
          Query: matches[4],
        };
      } else {
        return {
          Visual: -1,
          Dashboard: -1,
          Dataset: -1,
          Query: sqlString,
        };
      }
    }

    function displayValAssetId(assetId) {
      if (assetId === -1) {
        return EXTERNAL_VALUE;
      } else if (assetId === 0) {
        return 'None';
      }

      return String(assetId);
    }

    columnMgrs.push(new RangeColumnManager('Ending Timestamp', 'end_time',
      {
        filterHash: this._savedFilterData,
        filterType: FILTERTYPES.DATERANGE,
        orderable: true,
        displayFn: date => arc.misc.dateToMoment(date).format('MMM D, h:mm:ss A'),
        columnType: COLTYPES.DATETIME,
        defaultFilter: {
          quickRange: 'T',
          dateRangeType: DATETIME_FILTER_TYPES.QUICK,
        },
      }));
    columnMgrs.push(new PicklistColumnManager('Issuing User', 'effective_user',
      {
        filterHash: this._savedFilterData,
        orderable: true,
        displayFn: user => user === '' ? BLANK_USER_DISPLAY : user,
        makeCellContentFn: function displayUser(user) {
          return user === BLANK_USER_DISPLAY ?
            `<span class="text-faint">${BLANK_USER_DISPLAY}</span>` : user;
        },
        columnType: COLTYPES.USER,
      }));
    columnMgrs.push(new PicklistColumnManager('IP', 'network_address',
      {
        filterHash: this._savedFilterData,
        orderable: true,
        displayFn: val => val.slice(7),
      }));
    columnMgrs.push(new PicklistColumnManager('Base Table(s)', 'base_objects',
      {
        filterHash: this._savedFilterData,
        filterType: FILTERTYPES.PARTIAL_MATCH_PICKLIST,
        colParseFn: val => ({ 'Base Table(s)': parseTableList(val)}),
        makeCellContentFn: val => arc.misc.displayBaseTables(val),
        filterFn: whereClauseForTableListField,
        extraOption: {
          label: ' Include Arcadia Catalog',
          defaultFilter: "NOT RLIKE '[,\\[]arcadia_catalog.completed_queries[],]'"
        },
        defaultFilter: {
          extraOption: false,
        },
      }));
    columnMgrs.push(new PicklistColumnManager('Query Details', 'query_id',
      {
        filterType: FILTERTYPES.NONE,
        className: 'left-border',
        makeCellContentFn: () => btnHtml,
      }));
    columnMgrs.push(new PicklistColumnManager('Query', 'sql_statement',
      {
        skipRetrieval: true,
        hidden: true,
        filterType: FILTERTYPES.NONE,
      }));
    columnMgrs.push(new PicklistColumnManager('Matched AV(s)', 'aview_names',
      {
        filterHash: this._savedFilterData,
        filterType: FILTERTYPES.PARTIAL_MATCH_PICKLIST,
        className: 'nowrap shift-up-button',
        colParseFn: val => ({ 'Matched AV(s)': parseTableList(val)}),
        makeCellContentFn: displayAVs,
        filterFn: whereClauseForTableListField,
      }));
    columnMgrs.push(new RangeColumnManager('Duration', 'duration',
      {
        filterHash: this._savedFilterData,
        filterType: FILTERTYPES.RANGE,
        orderable: true,
        displayFn: val => `${Number(val).toPrecision(5)}s`,
      }));
    columnMgrs.push(new RangeColumnManager('# Rows', 'num_rows_fetched',
      {
        filterHash: this._savedFilterData,
        filterType: FILTERTYPES.RANGE,
        orderable: true,
        displayFn: val => Number(val),
      }));
    columnMgrs.push(new PicklistColumnManager('Dashboard', 'sql_statement',
      {
        filterHash: this._savedFilterData,
        orderable: true,
        className: 'left-border',
        colParseFn: extractAssetIds,
        displayFn: displayValAssetId,
        filterFn: (colName, data) => whereClauseForArcvizAsset('dashboard', data),
        makeCellContentFn: displayDbOrVisualLink,
      }));
    columnMgrs.push(new PicklistColumnManager('Visual', 'sql_statement',
      {
        filterHash: this._savedFilterData,
        orderable: true,
        skipRetrieval: true,
        colParseFn: extractAssetIds,
        displayFn: displayValAssetId,
        filterFn: (colName, data) => whereClauseForArcvizAsset('visual', data),
        makeCellContentFn: displayDbOrVisualLink,
      }));
    columnMgrs.push(new PicklistColumnManager('Dataset', 'sql_statement',
      {
        filterHash: this._savedFilterData,
        orderable: true,
        skipRetrieval: true,
        colParseFn: extractAssetIds,
        displayFn: displayValAssetId,
        filterFn: (colName, data) => whereClauseForArcvizAsset('dataset', data),
        makeCellContentFn: function datasetLink(id) {
          if (id !== 'None' && id !== EXTERNAL_VALUE && self._datasetPerms[id]) {
            const url = arc.misc.getBlankTargetUrl(`${arc.frontend.globals.urls.datasetEditUrl}/${id}`);
            return `<a href="${url}" target="_blank">${id}</a>`;
          }

          return id;
        },
      }));


    this._columnMgrs = columnMgrs;
  }

  _setUpListeners() {
    const self = this;
    $('#query-table-refresh').offOn('click.refresh', () => self._queryTable.ajax.reload());

    $('#query-table-reset').offOn('click.reset', () => self._resetColumnFilters());

    this._$elem.find('tbody').offOn('click.updateSelect', function toggleSelection(evt) {
      const $target = $(evt.target);
      if (!$target.closest('a').length) {
        const $row = $target.closest('tr');
        $row.toggleClass('selected');

        self._queryTable.rows('.selected').select();
        self._queryTable.rows(':not(.selected)').select(false);

        if (!$target.hasClass('row-select')) {
          const $checkbox = $row.find('.row-select');
          const isChecked = $checkbox.prop('checked');
          $checkbox.prop('checked', !isChecked);
        }
      }
    });

    const that = this;
    this._$elem.find('.allrow-select')
        .offOn('change', function() {
          let $this = $(this);
          let $inputs = that._$elem.find('tbody .row-select');
          let isChecked = $this.prop('checked');
          $inputs.each(function() {
            const $input = $(this);
            if (isChecked && !$input.prop('checked')) {
              $input.trigger('click');
            }

            if (!isChecked && $input.prop('checked')) {
              $input.trigger('click');
            }
          });
        });

    const $limitInput = $('#limit-returned-queries');

    $limitInput.offOn('change', function handleLimitUpdates() {
      const limit = $limitInput.val();
      const numLimit = Number(limit);
      if (numLimit < 0 || !Number.isInteger(numLimit)) {
        arc.report.alertMessageSmall('Limit must be a positive integer', 'alert-danger');
        return;
      }

      self._filterLimit = limit;
      self._queryTable.ajax.reload();
    });

    this._$elem.offOn('filterUpdate', () => self._queryTable.ajax.reload());

    this._$elem.offOn('filterValuesFetch', function handleFilterValuesFetch(evt) {
      const $target = $(evt.target);
      const colMgrIdx = $target.data('colMgrIdx');
      $.ajax({
        url: arc.frontend.globals.urls.sqlRunUrl,
        data: {
          dataconnection_id: self.dcId,
          query: self._buildFilterValueQuery(colMgrIdx),
        },
        type: 'POST',
        beforeSend: (xhr) => xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token),
      })
      .done(function handleFilterValues(resp) {
        const colMgr = self._columnMgrs[colMgrIdx];
        colMgr.setValuesFetched(true);
        let values = resp.rows.flat();
        if (colMgr.colParseFn) {
          values = values.map((val) => colMgr.colParseFn(val)[colMgr.displayName]);
        }

        let internalValues, displayValueMapping;
        ({ internalValues, displayValueMapping} = colMgr.generateColValueMapping(values));
        colMgr.drawFilter($target, internalValues, displayValueMapping, true);
      })
      .fail(
        err => arc.report.alertMessageSmall('Failed to retrieve full list of possible values', err)
      );


    });

    $('#query-detail-modal').offOn('show.bs.modal', function setupQueryDetailModal(evt) {
      const queryData = $(evt.relatedTarget).closest('tr').data('query');
      $('#query_id').text(queryData['Query Details']);
      $('#query_sql').text(queryData.Query);
    });

    const $avDetailModal = $('#av-sql-detail-modal');
    $avDetailModal.offOn('show.bs.modal', function setupAvQueryDetailModal(evt) {
      const avName = $(evt.relatedTarget).text();
      const dbName = avName.split('.')[0];
      const $avDef = $('#av-def');
      $avDef.addClass('arc-invisible');
      $('#av-name').text(avName);
      $avDetailModal.find('.fa-spin').removeClass('arc-invisible');
      $.ajax({
        url: arc.frontend.globals.urls.sqlRunUrl,
        type: 'POST',
        beforeSend: function (xhr) {
          xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
        },
        data: {
          dataconnection_id: self._dcId,
          query: `SHOW EXTENDED ANALYTICAL VIEWS in \`${dbName}\` LIKE '${avName}'`
        },
      })
      .fail(error => arc.report.alertMessage('Could not retrieve AV definition', error))
      .done(function populateModal(response) {
        const colNames = response.colnames;
        const metadataColIdx = colNames.indexOf('Metadata');
        const aviewObj = response.rows.map((row) => JSON.parse(row[metadataColIdx]))
                                      .filter((av) => av.name === avName); //ensure exact match
        if (!aviewObj.length) { // If AV has been deleted since query ran
          $('#av-sql-detail-modal').modal('hide');
          arc.report.alertMessageSmall(`No existing AV with name ${avName}`, 'alert-danger');
        } else {
          $avDetailModal.find('.fa-spin').addClass('arc-invisible');
          $avDef.text(aviewObj[0].definition).removeClass('arc-invisible');
        }
      });
    });

    $('#query-recommend-btn').offOn('click.accelerate', function storeQueries() {
      const maxQueries = 100;
      const rows = self._queryTable.rows({ selected: true});
      const count = rows.count();
      if (!count) {
        arc.report.alertMessageSmall('You must select at least one query to accelerate',
          'alert-warning');
        return;
      } else if (count > maxQueries) {
         arc.report.alertMessageSmall(
           `You may only accelerate up to ${maxQueries} queries at a time`, 'alert-warning');
        return;
      }

      self._queryTable.state.save();
      const filterStates = {};
      const selected = {};
      rows.ids().toArray().forEach((id) => selected[id] = true);
      self._columnMgrs.forEach(function storeState(mgr) {
        const filterData = mgr.data;
        if (filterData && Object.keys(filterData).length) {
          filterStates[mgr.displayName] = filterData;
        }
      });

      const data = {
        queries: JSON.stringify(rows.data().toArray()),
        dcId: self._dcId,
        state: JSON.stringify({
          dataTables: self._queryTable.state(),
          filters: filterStates,
          selected,
          filterLimit: self._filterLimit,
        }),
      };

      self._submitQueriesForAcceleration(data);
    });
  }

  /* Build a query that determines possible values for a particular column identified by the
  index of its column manager in the column managers list
  */
  _buildFilterValueQuery(colMgrIdx) {
    let filters = this._columnMgrs.filter((mgr, idx) => idx !== colMgrIdx)
      .map(colMgr => colMgr.generateFilterExpression())
      .filter(expr => expr);

    filters = filters.length ? 'AND '.concat(filters.join(' AND ')) : '';
    const query = `${this._columnMgrs[colMgrIdx].generateFilterQuery()} FROM ${HISTORY_TABLE}
                  WHERE (query_type = 'QUERY' ${filters})
                  LIMIT ${this._filterLimit}`;
    return query;
  }

  _resetColumnFilters() {
    this._columnMgrs.forEach(colMgr => colMgr.resetFilter(false));
    this._queryTable.ajax.reload();
  }

  _submitQueriesForAcceleration(data) {
    $.ajax({
      url: arc.frontend.globals.urls.rawSqlQueriesUrl,
      data,
      type: 'POST',
      beforeSend: function(xhr) {
        // Because of a jquery bug when handling POST requests,
        // we have to explicitly set X-CSRFToken here
        xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
      },
    })
    .done(function navigateToRecs(resp) {
      //eslint-disable-next-line no-undef
      window.location.href =
        `${arc.frontend.globals.urls.accelerateUrl}?timestamp=${resp}`;
    })
    .fail(function reportError(error) {
      arc.report.alertMessage('Failed to submit queries for acceleration', error);
    });
  }
}

export default QueryTableManager;
