import $ from 'jquery';
import arc from 'arc';
import moment from 'moment';
import QueryColumnManager, {FILTERTYPES, DATETIME_FILTER_TYPES} from './QueryColumnManagerClass';

const TIMESTAMP_FORMAT = 'YYYY-MM-DD HH:mm:ss';

const QUICK_FILTER_OPTIONS = {
  '5M': {
    display: 'last 5 mins',
    start: moment().subtract(5, 'minutes').format(TIMESTAMP_FORMAT),
    end: moment().format(TIMESTAMP_FORMAT),
  },
  '30M': {
    display: 'last 30 mins',
    start: moment().subtract(30, 'minutes').format(TIMESTAMP_FORMAT),
    end: moment().format(TIMESTAMP_FORMAT),
  },
  '1H':{
    display: 'last 1 hr',
    start: moment().subtract(1, 'hour').format(TIMESTAMP_FORMAT),
    end: moment().format(TIMESTAMP_FORMAT),
  },
  '3H': {
    display:'last 3 hrs',
    start: moment().subtract(3, 'hours').format(TIMESTAMP_FORMAT),
    end: moment().format(TIMESTAMP_FORMAT),
  },
  '6H': {
    display: 'last 6 hrs',
    start: moment().subtract(6, 'hours').format(TIMESTAMP_FORMAT),
    end: moment().format(TIMESTAMP_FORMAT),
  },
  'T': {
    display: 'Today',
    start: moment().startOf('day').format(TIMESTAMP_FORMAT),
    end: moment().endOf('day').format(TIMESTAMP_FORMAT),
  },
  'Y': {
    display: 'Yesterday',
    start: moment().startOf('day').subtract(1, 'day').format(TIMESTAMP_FORMAT),
    end: moment().endOf('day').subtract(1, 'day').format(TIMESTAMP_FORMAT),
  },
  '3D': {
    display: 'Last 3 days', //including today to be consistent with db filters
    start: moment().startOf('day').subtract(2, 'days').format(TIMESTAMP_FORMAT),
    end: moment().format(TIMESTAMP_FORMAT),
  },
  '7D': {
    display:'Last 7 days',
    start: moment().startOf('day').subtract(6, 'days').format(TIMESTAMP_FORMAT),
    end: moment().format(TIMESTAMP_FORMAT),
  },
};

const SQL_OPERATORS = {
  'greater': '>',
  'greater-or-equal': '>=',
  'equal': '=',
  'less': '<',
  'less-or-equal': '<=',
};

const SUMMARY_OPERATORS = {
  'greater': '>',
  'greater-or-equal': '\u2265',
  'equal': '=',
  'less': '<',
  'less-or-equal': '\u2264',
};

/**
 * Range Column Manager Class
 *
  This class is used to manage the behavior of a single range filter column in the table of recent
  arcengine queries.
 * Properties are the same as the parent Query Column Manager Class, but the structure of data
 * is different for this subclass
 * @property {object} data = {
    min: string,                    -minimum range value
    max: string ,                   -maximum range value
    quickRange: string,             -optional, value from QUICK_FILTER_DISPLAY
    dateRangeType: string           -optional, value from DATETIME_FILTER_TYPE
    minOper: string,                -optional, operator to apply to range min, value from
                                     SQL_OPERATORS or 'ignore'
    maxOper: string,                -optional, operator to apply to range max, value from
                                     SQL_OPERATORS or 'ignore'
  }

 * @method drawFilterDetails          -Setup the range filter. Only actually renders it though
                                       if the dropdown is expanded because jqrangeslider doesn't
                                       size properly when the DOM element isn't visible
 * @method updateSummary              -Update the summary visible at the top of the range filter
                                       even when the dropdown is collapsed
 * @method paintRangeFilter           -Actually render the range slider, only should be called
                                       when dropdown is expanded
 * @method generateFilterExpression   -Generate a SQL where clause based on the current filter range
 * @method generateFilterQuery        -Generate SQL select expression to retrieve possible range
                                       for the filter
 **/


class RangeColumnManager extends QueryColumnManager {
  constructor(displayName, columnName, config={}) {
    super(displayName, columnName, config);
  }

  /*
    function to setup the range filter
     internalValues {array}     -array of values taken by the column field
     updateOnly {boolean}       -whether to only re-render the range slider portion of the filter
     displayValMapping {object} -Mapping of internal values to user-friendly display values
     $filter {object}           -optional, jQuery object of the filter if rendered previously
   */
  drawFilterDetails(internalValues, updateOnly, displayValMapping, $filter) {
    const self = this;
    let $rangeInput;
    let $rangeSummary;

    const $queryTable = $('#query-list-table');

    function setupRangeFilter() {
      const numericValues = internalValues.map(Number);
      let min = Math.min(...numericValues);
      let max = Math.max(...numericValues);
      if (!updateOnly) {
        $filter = $($('#rangepicker-template').html());
       
        self._$header.append($filter);
        self.updateSummary();
      }

      if (internalValues.length) {
        if (min === max) {
          max += 1;
        }

        const range = max - min;
        const step = Math.pow(10, Math.round(Math.log10(range / 100)));

        // Make sure plotted range includes true min and max even with rounding based on step
        max = max + step;
        min = min > step ? min - step : 0; //for this table don't have negatives

        self._sliderConfig = {
          arrows: false,
          bounds: {
            min,
            max,
          },
          step,
          defaultValues: {
            min: (self.data.min !== undefined ? self.data.min : min),
            max: (self.data.max !== undefined ? self.data.max : max),
          },
        };

        if (updateOnly) {
          self.paintRangeFilter($filter);
        }
      }
    }

    function paintDateTimeFilter() {
      const dateRangePickerOpts = {
        format: TIMESTAMP_FORMAT,
        timePicker: true,
        timePickerIncrement: 1,
        applyButtonClass: 'choose-date-range',
      };
      let summaryText;
      $('.daterangepicker').hide();

      $filter = $($('#small-daterange-picker-template').html());

      $rangeSummary = $filter.find('.range-summary');
      self._$header.append($filter);
      $rangeInput = $filter.find('.filter-date-range');
      $rangeInput.daterangepicker(dateRangePickerOpts);
      $rangeInput.offOn('apply.daterangepicker', function updateRange(evt, picker) {
        const startDate = picker.startDate.format(TIMESTAMP_FORMAT);
        const endDate = picker.endDate.format(TIMESTAMP_FORMAT);
        const rangeText = `${startDate} - ${endDate}`;
        self._data = {
          dateRangeType: DATETIME_FILTER_TYPES.ABSOLUTE,
          min: startDate,
          max: endDate,
        };

        $rangeSummary.text(rangeText).attr('title', rangeText);
        $queryTable.trigger({type: 'filterUpdate'});
      });

      const rangeConfig = {
        $filter,
        $summary: $rangeSummary,
        $input: $rangeInput,
      };

      if (self.data.dateRangeType === DATETIME_FILTER_TYPES.QUICK) {
        self.showDateFilterWithDefaults(DATETIME_FILTER_TYPES.QUICK, false, rangeConfig);

        $filter.find(`button[value="${self.data.quickRange}"]`).addClass('active');
        $rangeSummary.text(QUICK_FILTER_OPTIONS[self.data.quickRange].display);
      } else if (self.data.dateRangeType === DATETIME_FILTER_TYPES.ABSOLUTE) {
        self.showDateFilterWithDefaults(DATETIME_FILTER_TYPES.ABSOLUTE, false, rangeConfig);
        $rangeInput.data('daterangepicker').setStartDate(self.data.min);
        $rangeInput.data('daterangepicker').setEndDate(self.data.max);
        summaryText = `${self.data.min} - ${self.data.max}`;
        $rangeSummary.text(summaryText).attr('title', summaryText);
      }

      $filter.offOn('click.dateFilter', function updateDateFilterVal(evt) {
        const $target = $(evt.target);
        const $closestDiv = $target.closest('div');
        if ($target.attr('type') === 'button') {
          if (!$target.hasClass('active')) {
            $target.addClass('active');
            $target.siblings().removeClass('active');
            $rangeSummary.text(QUICK_FILTER_OPTIONS[$target.val()].display);
            self._data = {
              dateRangeType: DATETIME_FILTER_TYPES.QUICK,
              quickRange: $target.val(),
            };
            $queryTable.trigger({type: 'filterUpdate'});
          }
        } else if ($closestDiv.hasClass('filter-type-option')) {
          if (!$closestDiv.find('i').hasClass('icon-invisible')) {
            return; //this option is already active
          }

          if ($closestDiv.hasClass('calendar-fixed-picked')) {
            self.showDateFilterWithDefaults(DATETIME_FILTER_TYPES.ABSOLUTE, false, rangeConfig);
          } else if ($closestDiv.hasClass('quick-filters-picked')) {
            self.showDateFilterWithDefaults(DATETIME_FILTER_TYPES.QUICK, true, rangeConfig);
            $queryTable.trigger({type: 'filterUpdate'});
          }
        }
      });
    }

    if (this._filterType === FILTERTYPES.RANGE) {
      setupRangeFilter();
    } else if (this._filterType === FILTERTYPES.DATERANGE) {
      paintDateTimeFilter();
    }
  }

  showDateFilterWithDefaults(filterType, applyDefaults = true, rangeConfig  = {}) {
    let defaultVal;
    let startDate;
    let endDate;
    let quickFilterOption;
    const $filter = rangeConfig .$filter || this._$header.find('.selected-control');
    const $summary = rangeConfig .$summary || $filter.find('.range-summary');
    const $input = rangeConfig .$input || $filter.find('.filter-date-range');
    switch (filterType) {
      case DATETIME_FILTER_TYPES.QUICK:
        defaultVal = 'T';
        if (applyDefaults) {
          $filter.find('button').each(function toggleButtonClass() {
            const $btn = $(this); //eslint-disable-line no-invalid-this
            $btn.toggleClass('active', $btn.val() === defaultVal);
          });

          $summary.text(QUICK_FILTER_OPTIONS[defaultVal].display);
          this._data = {
            quickRange: defaultVal,
            dateRangeType: filterType,
          };
        }

        $filter.find('.calendar-fixed-picked i').addClass('icon-invisible');
        $filter.find('.quick-filters-picked i').removeClass('icon-invisible');
        $filter.find('.quick-filters-related').removeClass('arc-invisible');
        $filter.find('.fixed-dates').addClass('arc-invisible');
        break;
      case DATETIME_FILTER_TYPES.ABSOLUTE:
        if (applyDefaults || this._data.dateRangeType !== DATETIME_FILTER_TYPES.ABSOLUTE) {
          quickFilterOption = this._data.quickRange || 'T'; //default to today
          endDate = QUICK_FILTER_OPTIONS[quickFilterOption].end;
          startDate = QUICK_FILTER_OPTIONS[quickFilterOption].start;
          $input.data('daterangepicker').setStartDate(startDate);
          $input.data('daterangepicker').setEndDate(endDate);
          $summary.text(`${startDate} - ${endDate}`);
          this._data = {
            min: startDate,
            max: endDate,
            dateRangeType: filterType,
          };
        }

        $filter.find('.calendar-fixed-picked i').removeClass('icon-invisible');
        $filter.find('.quick-filters-picked i').addClass('icon-invisible');
        $filter.find('.quick-filters-related').addClass('arc-invisible');
        $filter.find('.fixed-dates').removeClass('arc-invisible');
        break;
    }
  }

  updateSummary() {
    const $rangeSummary = this._$header.find('.range-summary');
    let summaryText;
    if (this.data.min !== undefined || this.data.max !== undefined) {
      const displayMin = this._displayFn && this.data.min !== undefined ?
        this._displayFn(this.data.min) : this.data.min;
      const displayMax = this._displayFn && this.data.max !== undefined ?
        this._displayFn(this.data.max) : this.data.max;
      if (this.data.maxOper === 'ignore') {
        if (this.data.minOper === 'ignore') {
          summaryText = '(All)';
        } else if (this.data.minOper === 'equal') {
          summaryText = this.data.min;
        } else {
          summaryText = `${SUMMARY_OPERATORS[this.data.minOper]} ${this.data.min}`;
        }
      } else if (this.data.minOper === 'ignore') {
        if (this.data.maxOper === 'equal') {
          summaryText = this.data.max;
        } else {
          summaryText = `${SUMMARY_OPERATORS[this.data.maxOper]} ${this.data.max}`;
        }
      } else {
        summaryText = `${displayMin} - ${displayMax}`;
      }
    } else if (this.data.dateRangeType === DATETIME_FILTER_TYPES.QUICK) {
      summaryText = QUICK_FILTER_OPTIONS[this.data.quickRange].display;
    } else {
      summaryText = '(All)';
    }

    $rangeSummary.text(summaryText).attr('title', summaryText);
  }

  paintRangeFilter($filter) {
    const self = this;
    const $dropDownContent = $filter.find('.dropdown-content');
    let $sliderWidget = $filter.find('.filter-range-widget');
    if ($sliderWidget.length) {
      $sliderWidget.remove();
    }

    $sliderWidget = $('<div class="filter-range-widget"></div>');
    $dropDownContent.prepend($sliderWidget);
    $sliderWidget.rangeSlider(self._sliderConfig);
    $sliderWidget.bind('userValuesChanged', function saveNewRange(evt, sliderData) {
      const currData = self._data || {};
      currData.min = sliderData.values.min;
      currData.max = sliderData.values.max;
      currData.minOper = currData.minOper || 'greater-or-equal';
      currData.maxOper = currData.maxOper || 'less-or-equal';
      self._data = currData;

      self.updateSummary();
      $('#query-list-table').trigger({type: 'filterUpdate'});
    });

    if (self._data && self._data.minOper) {
      markOperatorAsActive($filter.find('.min-range'), self._data.minOper);
    }

    if (self._data && self._data.maxOper) {
      markOperatorAsActive($filter.find('.max-range'), self._data.maxOper);
    }

    function markOperatorAsActive($inputGroup, value) {
      $inputGroup.find('.dropdown-menu').find('i').each(function toggleIconVisiblity() {
        const $this = $(this);
        $this.toggleClass('icon-invisible', $this.data('value') !== value);
      });

      $inputGroup.find('.operator-display').each(function toggleVisibility() {
        const $this = $(this);
        $this.toggleClass('arc-invisible', $this.data('value') !== value);
      });
    }

    $filter.find('.dropdown-item').offOn('click.changeOperator', function saveNewOper(evt) {
      const $div = $(evt.target);
      const val = $div.data('value');
      const $inputGroup = $div.closest('.input-group');
      const whichKey = $inputGroup.hasClass('min-range') ? 'minOper' : 'maxOper';
      const otherKey = whichKey === 'minOper' ? 'maxOper' : 'minOper';
      if (!$div.find('i').hasClass('icon-invisible')) { //already selected
        return;
      }

      markOperatorAsActive($inputGroup, val);
      const currData = self._data || {};
      currData[whichKey] = val;
      currData.min = currData.min || self._sliderConfig.defaultValues.min;
      currData.max = currData.max || self._sliderConfig.defaultValues.max;
      currData.minOper = currData.minOper || 'greater-or-equal';
      currData.maxOper = currData.maxOper || 'less-or-equal';

      if (val === 'equal') { //ignore other operator
        currData[otherKey] = 'ignore';
        markOperatorAsActive($inputGroup.siblings().first(), 'ignore');
      }

      self._data = currData;

      self.updateSummary();
      $('#query-list-table').trigger({type: 'filterUpdate'});
    });

    $filter.find('.fa-spinner').addClass('arc-invisible');
    $dropDownContent.removeClass('arc-invisible');
  }

  generateFilterExpression() {
    let sql;
    let min;
    let max;
    let minOperator;
    let maxOperator;

     //TODO: make this smarter rather than relying on a static set of options
    const QUICK_FILTER_SQL = {
      '5M': 'now() - interval 5 minute AND now()',
      '30M': 'now() - interval 30 minute AND now()',
      '1H': 'now() - interval 1 hour AND now()',
      '3H': 'now() - interval 3 hour AND now()',
      '6H': 'now() - interval 6 hour AND now()',
      'T': 'to_date(now()) AND now()',
      'Y': 'to_date(now() - interval 1 day) AND to_date(now())',
      '3D': 'to_date(now() - interval 2 day) AND now()',
      '7D': 'to_date(now() - interval 6 day) AND now()',
    };

    if (this._filterFn) {
      return this._filterFn(this._columnName, this._data, this._extraOption);
    }

    if (this._filterType === FILTERTYPES.DATERANGE) {
      if (this.data.dateRangeType === DATETIME_FILTER_TYPES.QUICK) {
        sql = `${this._columnName} BETWEEN ${QUICK_FILTER_SQL[this.data.quickRange]}`;
      } else if (this.data.dateRangeType === DATETIME_FILTER_TYPES.ABSOLUTE) {
        sql = `${this._columnName} BETWEEN '${this.data.min}' AND '${this.data.max}'`;
      }
    } else if (this._filterType === FILTERTYPES.RANGE ) {
      minOperator = this.data.minOper;
      maxOperator = this.data.maxOper;
      min = this.data.min;
      max = this.data.max;

      if (minOperator === 'ignore') {
        if (max === undefined || maxOperator === 'ignore') {
          return '';
        }

        sql = arc.misc.whereClauseFromMax(this._columnName, max, SQL_OPERATORS[maxOperator], true);
      } else if (maxOperator === 'ignore') {
        sql = arc.misc.whereClauseFromMin(this._columnName, min, SQL_OPERATORS[minOperator], true);
      } else {
        if (min === undefined || max === undefined) {
          return '';
        }

        sql = arc.misc.whereClauseFromMinMax(this._columnName, min, max, SQL_OPERATORS[minOperator],
          SQL_OPERATORS[maxOperator], true);
      }
    }

    return sql;

  }

  generateFilterQuery() {
    return `SELECT MIN(${this._columnName}), MAX(${this._columnName})`;
  }
}

export default RangeColumnManager;