window.arc = window.arc || {};

(function() {
'use strict';

/*
 * Parameter Controls
 * (Currently working through url passing,
 * potentially support post requests in the future)
 * TODO: Add in post request support
 *
 * This allows for users to customize parameter settings through the url

 * Example URL
  http://BASEURL/arc/apps/app/ID?config.OUTPUT={SETTING: VALUE}
  CURRENTLY ONLY SUPPORTS "value"
  - BASEURL: something like 127.0.0.1:8000 or sf.arcadiadata.com):
  - ID: dashboard id (should also work for picklist visual)
  - OUTPUT (can also be picklist id)
  - HASH OF SETTING/VALUES
    - SETTING: Picklist setting name
      Includes (and any other picklist settings):
        textbox: Displays textbox parameter (true/false)
        base_field: Base column field -- if passed in, will change base column to match
        title: Title of picklist, used as alias (Works with <<params>>)
        output_parameter: Key of params that is output-ed upon selecting value
        filter_columns
        itemcount: number of items to display (accepts number 1 - 10000)
        single_select: If only one value can be selected (true/false)
        display_no_selection: Allow "All" option in single select mode
          (true/false)
        multi_parm: emits distinct parameters for each item (true/false)
        dropdown: use dropdown instead of radio buttons (true/false)
        values: the labels/values users can choose
          (array of arrays, need to be valid json object, singl dimension:
          [[value1, label1], [value2, label2], [value3, label3]])
        hide_filter_when_empty:
        filter_appgroup_scope:
        filter_dataset_scope:
    - SPECIAL SETTINGS:
        hide: Hides specified picklist under certain conditions
          (conditional of output and value to hide on: state != VA | state == VA)
    - VALUE:
        See above, need designated value to work

  * Behavior:
    If a picklist with specified output doesn't exist, then it will be created
      This actually creates newe visual, with it's own id and row in django records_record
      Too many of these could clutter up the database
      Preferred method is to create the parameter, with expected output,
        and modify from there
    If existing picklist's output/title is changed,
      this will NOT undo any changes made through the url
    If value is passed in through url, and parameter being already has values
      the values will be appended.
      Duplicates will not be removed since it's not immediately obvious
        if label or value duplicates should be removed
    If user mis-types something, there will be warning
 */

arc.params = arc.params || {};

// Initialize parameters passed in
arc.params.parameters = function() {
  var params = {}, picklists = {}, hideConditions = {},
      supportedProperties = {
        values: true,
        finalize: false,
      };

  var parameters = function() { return 'arc.params.parameters'; };

  parameters.addParams = function(k, v) {
    params[k] = parseJSON(v);
    return parameters;

    function parseJSON(hashString) {
      // Need custom json parser that doesn't require quotes around keys
      // Also, only returns 1D hash
      var cleanHash = hashString.replace(/(['"])?([a-zA-Z0-9_]+)(['"])?:/g, '"$2": '),
          hashResult = JSON.parse(cleanHash);

      for (var property in hashResult) {
        stringifyAndCleanupValue(property, hashResult[property]);
      }

      return hashResult;

      function stringifyAndCleanupValue(key, value) {
        // Remove spaces and 1 quote at beginning of key and value
        value = cleanWord(value);
        if (supportedProperties[key]) {
          if (key === 'hide') {
            var hideWhen = parseHideValue(value);
            if (hideWhen) {
              hideConditions[hideWhen.parameterName] = {
                picklistName: k,
                hideWhen: hideWhen.parameterValue,
                equals: hideWhen.onEquals,
              };
            } else {
              delete hashResult[key];
            }
          } else {
            hashResult[key] = value;
          }
        } else {
          console.log(key + ' is not currently supported');
        }

        function cleanWord(string) {
          if (typeof string !== 'string') {
            string = JSON.stringify(string);
          }

          return string.replace(/^\s*["']?|["']?\s*$/g, '');
        }

        function parseHideValue(string) {
          var onEquals,
              equalSign = '==', notEqualSign = '!=',

              // Allows 1 space before/after equivalence sign
              // This way, user can specify spaces in parameter name and value
              nameAndValue = string.split(/\s?!?==?\s?/),
              name = nameAndValue[0], value = nameAndValue[1];

          if (string.indexOf(equalSign) !== -1) {
            onEquals = true;
          } else if (string.indexOf(notEqualSign) !== -1) {
            onEquals = false;
          } else {
            arc.report.alertMessage('"hide" parameter property missing != or ==: ' + string);
            return false;
          }

          return {
            parameterName: name,
            parameterValue: value,
            onEquals: onEquals,
          };
        }
      }
    }
  };

  parameters.checkParamsAndHide = function(paramsData) {
    paramsData.forEach(function (pd) {
      parameters.checkParam(pd.key, pd.value);
    });
  };

  parameters.checkParam = function(newParam, newValue) {
    var hideParam = hideConditions[newParam],
        hideVisual;
    if (hideParam) {
      var picklist = picklists[hideParam.picklistName];

      // Figure out if visual should be hidden or shown
      if (hideParam.equals) {
        if (hideParam.hideWhen === newValue) {
          hideVisual = true;
        }
      } else {
        if (hideParam.hideWhen !== newValue) {
          hideVisual = true;
        }
      }

      // Now hide or show
      if (hideVisual && picklist && picklist.viz() && picklist.viz().viz()) {
        picklist.viz().viz().viz_resetControl(true);
        picklist.viz().viz().setVisibility();
      } else if (picklist && picklist.viz() && picklist.viz().viz()) {
        picklist.viz().viz().setVisibility(true);

      }
    }
  };

  parameters.addPicklist = function(report) {
    var output = updateReportUrlReturnOutput(report);
    parameters.add(report.report_id(), report);
    if (output) {
      parameters.add(output, report);
    }
  };

  parameters.deletePicklist = function (report) {
    if (report) {
      parameters.delete(report.report_id());
      parameters.delete(report.viz().viz().getOutputParameterRaw());
    }
  };

  parameters.delete = function(paramKey) {
    delete picklists[paramKey];
  };

  parameters.add = function(k, v) {
    picklists[k.toString()] = v;
  };

  parameters.create = function (dashboard, picklistName) {
    var dataset_id = dashboard.getDatasetId(),
        picklistConfig = {
                  picklistSettings: params[picklistName],
                  output: picklistName,
                },
        viz = dashboard.dashboard_filter_controls()
                       .createControl({
                         dataset_id,
                         visual_type: 'picklist',
                         single_select: true,
                         picklistConfig,
                       });
  };

  parameters.finalize = function (dashboard) {
    Object.keys(params).forEach(function (picklistName) {
      if (!picklists[picklistName]) {
        if (supportedProperties.finalize) {
          parameters.create(dashboard, picklistName);
        } else {
          console.log('No picklist with output or ID of ' + picklistName);
        }
      }
    });
  };

  parameters.picklists = function () {
    return picklists;
  };

  parameters.params = function () {
    return params;
  };

  function updateReportUrlReturnOutput(report) {
    var viz = report.viz().viz(),
        output = viz.getOutputParameterRaw(),
    urlParams = params[output] || params[report.report_id()];
    if (urlParams) {
      viz.updateWithUrlParameters(urlParams);
    }

    return output;
  }

  return parameters;
};

}());
