window.arc = window.arc || {};

(function() {
"use strict";

arc.report = arc.report || {};

const flowFunnelVisualTypes = new Set(['flow', 'funnel']);

/*
 * When we switch from one visual to another, it is necessary to figure out
 * what to do with the data from all the shelves. This object tries to move
 * objects from one set of shelves to another set of shelves in an optimal
 * way.
 */
arc.report.shelftracker = function() {
  var info = {},
      shelf_data,
      viz_lib,
      primaryShelfHistory,
      shelfHistory = {},
      keepInShelfHistory = [
                              'bar_measure',
                              'line_measure',
                              'highlighter_measure',
                              'core',
                            ];

  var shelftracker = function() { return "arc.report.shelftracker"; };

  function filterPredicates(s) {
    if (s.shelf_type === arc.report.shelf_types.PREDICATE) return false;
    return true;
  }

  shelftracker.info = function(_) {
    if (!arguments.length) return info;
    info = _;
    return shelftracker;
  };

  shelftracker.clearHistory = function() {
    shelfHistory = {};
    primaryShelfHistory = undefined;
    return shelftracker;
  };

  shelftracker.addHistory = function(vizType, shelfInfo) {
    shelfHistory[vizType] = {};
    const shelvesToStore = shelfInfo.core.shelves.filter(filterPredicates);

    shelvesToStore.forEach(function (shelf) {
      const shelfElems = shelfInfo[shelf.name];
      shelfElems.forEach(shelfElement => shelfHistory[vizType][shelfElement.expression_for_trigger] = shelf.name);
    });

    if (shelfInfo.core.shelves.length) {
      primaryShelfHistory = $.extend(true, {}, shelfInfo);

      // Delete unnecessary properties so they don't overwrite current settings
      delete primaryShelfHistory.filter_shelf;
      Object.keys(primaryShelfHistory).forEach(infoProperty => {
        if (!infoProperty.includes('_shelf') && !keepInShelfHistory.includes(infoProperty)) {
          delete primaryShelfHistory[infoProperty];
        }
      });
    }

    return shelftracker;
  };

  shelftracker.viz_lib = function(_) {
    if (!arguments.length) return viz_lib;
    viz_lib = _;
    return shelftracker;
  };

  shelftracker.shelf_data = function(_) {
    if (!arguments.length) return shelf_data;
    shelf_data = _;
    return shelftracker;
  };

  // Return information about the shelves associated with this
  // tracker
  shelftracker._shelves_to_stored_shelves = function() {
    return shelf_data.map(function(s) {
      return {
        name: s.name,
        shelf_type: s.shelf_type,
        column_type: s.column_type
      };
    });
  };

  // Save current information about shelf elements, to use in 'restore'
  // at a later point
  shelftracker.save = function() {
    info.core.shelves = shelftracker._shelves_to_stored_shelves();
    return shelftracker;
  };

  /*
   * Move shelf elements around as necessary.
   */
  shelftracker.restore = function(switchingVisualType) {
    if (!info.core || !switchingVisualType) {
      // We are loading in a previous version of this metadata
      return shelftracker;
    }

    const newVizType = viz_lib.viz_type();
    let oldVizType = info.core.viz_type;

    if (newVizType === oldVizType) {
      return shelftracker;
    }

    // flow and funnel can only transition from/to each other
    if ((flowFunnelVisualTypes.has(oldVizType) && !flowFunnelVisualTypes.has(newVizType)) ||
        (flowFunnelVisualTypes.has(newVizType) && !flowFunnelVisualTypes.has(oldVizType))) {
      return shelftracker;
    }

    let oldShelves = info.core.shelves.filter(filterPredicates),
        shelvesToRemove,
        new_shelves = shelftracker._shelves_to_stored_shelves().filter(filterPredicates);
    const tempShelves = {};
    new_shelves.forEach(shelf => tempShelves[shelf.name] = []);

    //viz Groups are grouped by visuals with similar shelf configurations. Possible future class
    //structure.
    const vizGroups = {
      trellisAxis: ['trellis-bars', 'trellis-lines', 'trellis-areas', 'trellis-groupedbars'],
      axis: ['stackedbar', 'lines', 'area', 'bar', 'area', 'line', 'bar-horizontal', 'combo',
        'stackedbar-agg', 'lines-agg', 'stackedbar-agg-horizontal', 'scatter', 'sparklines'],
      basic: ['packed-bubbles', 'pie', 'radial', 'wordcloud', 'chatbot'],
      relational: ['network', 'chord', 'dendrogram', 'treemap', 'corelation', 'corelation-flow'],
      kpi: ['kpi', 'gauge', 'bullet'],
      analysis: ['histogram', 'boxplot'],
      geoAndDate: ['map', 'leaflet', 'calendar-heatmap'],
      flow: ['flow', 'funnel'],
      table: ['table', 'crosstab', 'list'],
      noShelfVisuals: ['link', 'extension', 'html', 'column-stats', 'rawtable', 'action', 'search'],
      filters: ['picklist', 'daterange', 'range', 'timecontrol'],
    };

    let vizGroup;
    for (let vizGroupName in vizGroups) {
      if (vizGroups[vizGroupName].includes(newVizType)) {
        vizGroup = vizGroupName;
        break;
      }
    }

    if (vizGroup === 'noShelfVisuals' || vizGroup === 'filters') {
      return shelftracker;
    }

    if (!vizGroup) {
      console.log(`Unknown visual type: ${newVizType}`);
    }

    // Only try restoring from shelf history if switching visuals
    if (switchingVisualType) {
      const savedHistory = shelfHistory[newVizType];

      if (!info.core.shelves.length && primaryShelfHistory) {
        // If the visual we're switching from doesn't have shelves (sql, link, etc),
        // switch using first available history item with shelves
        Object.assign(info, primaryShelfHistory);
        primaryShelfHistory.core = Object.assign({}, primaryShelfHistory.core);
        oldVizType = info.core.viz_type;
      }

      oldShelves = info.core.shelves.filter(filterPredicates);
      shelvesToRemove = oldShelves.filter(shelfName => !new_shelves.includes(shelfName));

      if (savedHistory) {
        oldShelves.forEach(function replaceSavedHistoryShelfInfo(shelf) {
          let currShelfElems = info[shelf.name];
          currShelfElems.forEach(function updateShelfElements(elem) {
            if (savedHistory[elem.expression_for_trigger]) {
              tempShelves[savedHistory[elem.expression_for_trigger]].push(elem);
            } else {
              reassignSingleElement(elem, tempShelves, vizGroup, newVizType, shelf);
            }
          });
        });

        Object.assign(info, tempShelves);
        shelvesToRemove.forEach(shelf => delete info[shelf.shelf_name]);
        return shelftracker;
      }
    }

    // If all 'oldShelves' and 'new_shelves' names are identical, we are done
    if (arc.arraysIdentical(
          oldShelves.map(function(d) {
            return d.name;
          }),
          new_shelves.map(function(d) {
            return d.name;
          }))) {
      return shelftracker;
    }

    var old_shelves_h = arc.misc.arrayToMap(oldShelves, d => d.name);

    var new_shelves_h = arc.misc.arrayToMap(new_shelves, function(d) {
      // Wipe out old information from new shelves
      // If a new shelf has the same name as an old shelf, we want to maintain
      // that information; so we skip over those.
      if (old_shelves_h[d.name] === undefined) {
        info[d.name] = [];
      }
      return d.name;
    });

    oldShelves.forEach(oldShelf => reassignShelfElements(tempShelves, oldShelf, new_shelves_h, vizGroup, newVizType));

    for (let shelfName in tempShelves) {
      info[shelfName] = tempShelves[shelfName];
    }

    return shelftracker;
  };

  function reassignSingleElement(elem, tempShelves, vizGroup, newVizType, oldShelf) {
    let shelfElem = arc.report.shelf.element().restore(elem, viz_lib.dataset_events(), viz_lib.dataset());
    const isAgg = shelfElem.expressionIsAggregate();

    switch (vizGroup) {
      case 'basic':
        if (tempShelves.dimensions_shelf.length < 1 && !isAgg) {
          tempShelves.dimensions_shelf.push(elem);
        } else if (tempShelves.aggregates_shelf.length < 1 && isAgg) {
          tempShelves.aggregates_shelf.push(elem);
        } else if (!isAgg && newVizType === 'pie') {
          tempShelves.dimensions_shelf.push(elem);
        } else if (isAgg && newVizType === 'radial') {
          tempShelves.aggregates_shelf.push(elem);
        } else if (['wordcloud', 'packed-bubbles'].includes(newVizType) && tempShelves.color_shelf.length < 1) {
          tempShelves.color_shelf.push(elem);
        } else if (isAgg) {
          tempShelves.tooltip_shelf.push(elem);
        } else {
          tempShelves.x_shelf.push(elem);
        }

        break;
      case 'relational':
        if ((tempShelves.dimensions_shelf.length < 2 || newVizType === 'corelation-flow') && !isAgg) {
          tempShelves.dimensions_shelf.push(elem);
        } else if (tempShelves.aggregates_shelf && isAgg && tempShelves.aggregates_shelf.length < 1) {
          tempShelves.aggregates_shelf.push(elem);
        } else if (tempShelves.link_color_shelf && tempShelves.link_color_shelf.length < 1) {
          tempShelves.link_color_shelf.push(elem);
        } else {
          tempShelves.tooltip_shelf.push(elem);
        }

        break;
      case 'kpi':
        if (!isAgg) {
          if (newVizType === 'kpi' && tempShelves.dimensions_shelf.length < 1) {
            tempShelves.dimensions_shelf.push(elem);
          } else if (tempShelves.x_shelf.length < 1) {
            tempShelves.x_shelf.push(elem);
          } else {
            tempShelves.tooltip_shelf.push(elem);
          }
        } else {
          if (tempShelves.aggregates_shelf.length < 1) {
            tempShelves.aggregates_shelf.push(elem);
          } else if (tempShelves.compare_shelf.length < 1) {
            tempShelves.compare_shelf.push(elem);
          } else {
            tempShelves.tooltip_shelf.push(elem);
          }
        }

        break;
      case 'analysis':
        //note these visuals do not really correspond well, so this is not a great mapping
        if (oldShelf.column_type === arc.report.columnTypeEnum.AGGREGATE &&
            tempShelves.aggregates_shelf.length < 1) {
          tempShelves.aggregates_shelf.push(elem);
        } else if (newVizType === 'boxplot' && !isAgg && tempShelves.dimensions_shelf.length < 1) {
          tempShelves.dimensions_shelf.push(elem);
        } else {
          tempShelves.x_shelf.push(elem);
        }

        break;
      case 'geoAndDate':
        if (isAgg && tempShelves.aggregates_shelf.length < 1) {
          tempShelves.aggregates_shelf.push(elem);
        } else if (!isAgg && tempShelves.dimensions_shelf && tempShelves.dimensions_shelf.length < 1) {
          tempShelves.dimensions_shelf.push(elem);
        } else if (!isAgg && tempShelves.geo_shelf && tempShelves.geo_shelf.length < 2) {
          tempShelves.geo_shelf.push(elem);
        } else {
          tempShelves.tooltip_shelf.push(elem);
        }

        break;
      case 'trellisAxis':
      case 'axis':
        if (!isAgg && tempShelves.x_shelf.length < 1) {
          tempShelves.x_shelf.push(elem);
        } else if (isAgg && tempShelves.y_shelf && tempShelves.y_shelf.length < 1) {
          tempShelves.y_shelf.push(elem);
        } else if (isAgg && tempShelves.bar_measure && tempShelves.bar_measure.length < 1) {
          tempShelves.bar_measure.push(elem);
        } else if (tempShelves.color_shelf && tempShelves.color_shelf.length < 1) {
          tempShelves.color_shelf.push(elem);
        } else if (isAgg && tempShelves.line_measure && tempShelves.line_measure.length < 1) {
          tempShelves.line_measure.push(elem);
        } else if (isAgg && tempShelves.highlighter_measure && tempShelves.highlighter_measure.length < 1) {
          tempShelves.highlighter_measure.push(elem);
        } else if (isAgg && newVizType !== 'sparklines') {
          tempShelves.tooltip_shelf.push(elem);
        } else if (isAgg && newVizType === 'sparklines') {
          tempShelves.y_shelf.push(elem);
        } else {
          tempShelves.x_shelf.push(elem);
        }

        break;
      case 'flow':
        if (isAgg && tempShelves.measure_shelf) {
          tempShelves.measure_shelf.push(elem);
        } else {
          tempShelves.time_shelf.push(elem);
        }
        break;

      case 'table':
        if(isAgg) {
          tempShelves.aggregates_shelf.push(elem);
        } else if (tempShelves.y_dimensions_shelf && tempShelves.y_dimensions_shelf.length < 1) {
          tempShelves.y_dimensions_shelf.push(elem);
        } else if (tempShelves.x_dimensions_shelf) {
          tempShelves.x_dimensions_shelf.push(elem);
        } else if (tempShelves.dimensions_shelf) {
          tempShelves.dimensions_shelf.push(elem);
        } else {
          tempShelves.tooltip_shelf.push(elem);
        }
    }
  }

  function reassignShelfElements(tempShelves, oldShelf, newShelvesH, vizGroup, newVizType) {
    const elems = info[oldShelf.name];

    if (newShelvesH[oldShelf.name] === undefined) {
      delete info[oldShelf.name];
    }

    // No elements remaining - we are done
    if (elems.length === 0) {
      return;
    }

    elems.forEach(elem => reassignSingleElement(elem, tempShelves, vizGroup, newVizType, oldShelf));
  }

  return shelftracker;
};

}());
