window.arc = window.arc || {};

(function() {
"use strict";

/*
 * In some cases, individual visuals need to fetch assets. If there are
 * multiple visuals on a page, it is useful to cache these assets so
 * there is exactly one copy.
 *
 * The api we support for this is:
 *
   arc.staticfetcher.fetch([
     {
       asset: asset_url,
       postprocess: postprocess_fn
     },
     {
       asset: asset_url_2,
       postprocess: postprocess_fn
     }
   ], callback_fn)

 * postprocess_fn is meant to be something like d3.csv or d3.json
 * Note that the expectation here is that the postprocess_code is identical
 * for all calls with the same asset_url.
 *
 * Fetched and postprocessed assets are tracked in arc.staticfetcher.data,
 *    key'ed by asset_url
 *
 * Assets in the middle of being fetched are tracked in
 *    arc.staticfetcher.fetchInProcess, key'ed by asset_url. The value in
 *    this map is an array of callback functions that need to be called
 *    on completion.
 */

arc.testutil = arc.testutil || {};
arc.staticfetcher = arc.staticfetcher || {};

arc.staticfetcher.fetchInProcess = {};
arc.staticfetcher.data = {};
arc.staticfetcher.forget = {};

// This object is created for each call to arc.staticfetcher.fetch()
// It keeps track of how many assets need to be tracked till the
// combined_callback can be called. It is first initialized with total() (which
// is number of assets, and therefore same as remaining) and then repeatedly
// called with decrement() every time an asset is retrieved. decrement() will
// eventually call the combined_callback when remaining assets become 0.
arc.staticfetcher.multistageCallback = function() {
  var q = function() { return "arc.staticfetcher.multistate_callback"; };

  var remaining,
      total,
      callback;

  q.total = function(_) {
    if (!arguments.length) return total;
    if (remaining !== undefined) {
      throw new Error("'total' cannot be called multiple times");
    }
    total = _;
    remaining = _;
    return q;
  };

  q.callback = function(_) {
    if (!arguments.length) return callback;
    callback = _;
    return q;
  };

  q.decrement = function(error) {
    if (remaining === undefined) {
      throw new Error("'decrement' called without 'total'");
    }
    remaining--;
    if (remaining < 0) {
      throw new Error("'decrement' called too many times. Original count was " + total);
    }
    if (error) {
      callback(error);
      return true;
    }
    if (remaining === 0) {
      callback();
      return true;
    }
    return false;
  };

  return q;
};

/*
 * Forget about a previously cached asset. Called when we want to reload
 * from the backend explicitly.
 */
arc.staticfetcher.forget = function(assets) {
  assets.forEach(function(a) {
    arc.staticfetcher.forget[a.asset] = true;
  });
};

/*
 * Update with pre-defined front-end changes
 */
arc.staticfetcher.update = function(assets) {
  assets.forEach(function(a) {
    const currentData = arc.staticfetcher.data[a.asset];
    if (currentData) {
      a.cb(currentData);
    }
  });
};

/*
 * Populate existing assets
 */
arc.staticfetcher.populate = function(assets) {
  Object.keys(assets).forEach(k => {
    arc.staticfetcher.data[k] = assets[k];
  });
};

// Multiple calls to arc.staticfetcher.fetch will create multiple q.
// q.decrement() is the callback for every time an asset loads.
// For each individual q (which itself refers to the group of assets from a
// single call to arc.staticfetcher.fetch), when the remaining assets gets to 0
// (after decrementing each time), the initial callback specified in the fetch()
// will be called
arc.staticfetcher.fetch = function(assets, cb) {
  var q = arc.staticfetcher.multistageCallback();
  var remainingap = [];
  assets.forEach(function(asset_processor) {
    var asset = asset_processor.asset;
    var postprocess = asset_processor.postprocess;
    if (arc.staticfetcher.forget[asset] ||
        (arc.staticfetcher.data[asset] === undefined)) {
      // Mark the asset as needing to be fetched
      remainingap.push(asset_processor);
    }
  });

  function gatherAndCallback(error) {
    // The callback is made when all assets are fetched successfully
    // Collect the data up and make the callback
    var data = [error];
    assets.forEach(function(asset_processor) {
      var asset = asset_processor.asset;
      data.push(arc.staticfetcher.data[asset]);
    });

    cb.apply(undefined /* this */, data);
  }

  if (remainingap.length === 0) {
    gatherAndCallback();
  }

  // Setup the success case
  q.total(remainingap.length)
   .callback(gatherAndCallback);

  function makeCallbacks(asset, error) {
    var cbs = arc.staticfetcher.fetchInProcess[asset];
    if (!cbs) {
      return;
    }
    cbs.forEach(function(eachcb) {
      eachcb(error);
    });
    delete arc.staticfetcher.fetchInProcess[asset];
  }

  // Start fetching the remaining data
  remainingap.forEach((asset_processor) => {
    const { asset } = asset_processor;
    const p = arc.staticfetcher.fetchInProcess[asset];
    if (p === undefined) {
      arc.staticfetcher.fetchInProcess[asset] = [q.decrement];
      const { postprocess } = asset_processor;
      const qf = arc.testutil.queue || d4.queue;
      const qu = qf();

      const t1 = new Date();

      // Fetch the required assets
      qu.defer(postprocess, asset)
        .await((error, _asset_data) => {
          const t2 = new Date();
          if (error) {
            arc.perf.PAGEPERF.addProfile({
              type: 'staticfetch',
              time: t2 - t1,
              error,
            });

            // On error, call all error routines immediately
            makeCallbacks(asset, error);
            return;
          }

          arc.perf.PAGEPERF.addProfile({
            id: 0,
            type: 'staticfetch',
            time: t2 - t1,
            profile: [
              [
                'url',
                asset,
              ]
            ],
          });

          arc.staticfetcher.data[asset] = _asset_data;
          delete arc.staticfetcher.forget[asset];
          makeCallbacks(asset);
        });
    } else {
      // We are fetching this asset currently in some other call. Wait
      // for it to get here
      p.push(q.decrement);
    }
  });
};
}());
