window.arc = window.arc || {};

(function() {
'use strict';

function alertMessage(lib, msg, detail, optionalAlertConfig) {
  return arc.report.contextualAlertMessage(lib, msg, detail, optionalAlertConfig);
}

arc.globals = arc.globals || {};
arc.data = arc.data || {};
arc.data.STATUS_RUNNER = arc.data.STATUS_RUNNER || {};
arc.data.CANCEL_RUNNER = arc.data.CANCEL_RUNNER || {};
arc.data.RESULTS_RUNNER = arc.data.RESULTS_RUNNER || {};

// Enumeration of query states
arc.queryStates = {
  RUNNING: 'RUNNING',
  CANCEL: 'CANCEL',
  CANCELLING: 'CANCELLING',
  CANCELLED: 'CANCELLED',
  FINISHED: 'FINISHED',
  ERROR: 'ERROR',
};

function queryRunner() {
  if (arc.data.QUERY_RUNNER) {
    return arc.data.QUERY_RUNNER;
  }

  // Initialize counters
  arc.counters.initialize('dataapi_start');
  arc.counters.initialize('dataapi_end');
  arc.counters.initialize('dataapi_query_start');
  arc.counters.initialize('dataapi_query_end');

  arc.data.QUERY_RUNNER = arc.data.queryParallelRunner();
  return arc.data.QUERY_RUNNER;
}

function groupWarningColumns(warnings, colnames) {
  const warningRowIndices = {};
  return warnings.reduce((allWarnings, warning) => {
    if (!warningRowIndices[warning.row_idx]) {
      warningRowIndices[warning.row_idx] = warning.row_idx.toString();
      return [...allWarnings, {
        rowIdx: warning.row_idx,
        columnName: [colnames[warning.column_idx]]
      }];
    }

    allWarnings[warning.row_idx].columnName.push(colnames[warning.column_idx]);

    return allWarnings;
  }, []);
}

function getWarningText({ query_warnings, rows, colnames }, type) {
  const groupedWarnings = groupWarningColumns(query_warnings[type], colnames);
  const warningList = groupedWarnings.reduce((acc, message) => `${acc}<li>row: {${arc.misc.escapeHTML(rows[message.rowIdx])}},
    column: <strong>${arc.misc.escapeHTML(message.columnName.toString())}</strong>\n</li>`, '');
  return `<div><p>The following rows contain <strong>${type}</strong>
  values that are not being displayed:</p><ul style="list-style-type:none">${warningList}</ul></div>`;
}

function formatQueryWarningMessage(query_data) {
  let infinityWarningText = '';
  let nanWarningText = '';
  if (query_data.query_warnings.inf.length) {
    infinityWarningText = getWarningText(query_data, 'inf');
  }

  if (query_data.query_warnings.nan.length) {
    nanWarningText = getWarningText(query_data, 'nan');
  }

  return infinityWarningText + nanWarningText;
}

function showInfinityOrNanValueWarning(query_data) {
  arc.report.alertMessageSmallBase({
    type: 'alert-warning',
    alertHeader: 'Attention: Infinity or NaN values ',
    alertMessage: formatQueryWarningMessage(query_data),
    fadeOut: false,
    useHTML: true
  });
}

function queryRun(callId, params, cb, profileId) {
  var paramsCopy = $.extend({}, params);
  queryRunner().add(callId, paramsCopy, cb, profileId);
}

function qhandleProcessId(qhandle) {
  return qhandle.substr(0, qhandle.indexOf('.'));
}

function cancelRunner(qhandle) {
  let processId = qhandleProcessId(qhandle);

  if (arc.data.CANCEL_RUNNER[processId]) {
    return arc.data.CANCEL_RUNNER[processId];
  }

  arc.data.CANCEL_RUNNER[processId] =
      arc.data.batch()
              .type('Cancel')
              .url(`${arc.frontend.globals.urls.sqlRunUrl}/../${processId}/asynccancel_async`);

  return arc.data.CANCEL_RUNNER[processId];
}

function asyncCancel(qhandle, cb, profileId) {
  cancelRunner(qhandle).add(qhandle, qhandle, cb, profileId);
}

function statusRunner(qhandle) {
  let processId = qhandleProcessId(qhandle);

  if (arc.data.STATUS_RUNNER[processId]) {
    return arc.data.STATUS_RUNNER[processId];
  }

  arc.data.STATUS_RUNNER[processId] =
      arc.data.batch()
              .type('Status')
              .url(`${arc.frontend.globals.urls.sqlRunUrl}/../${processId}/asyncstatus_async`);

  return arc.data.STATUS_RUNNER[processId];
}

function asyncStatus(qhandle, cb, profileId) {
  statusRunner(qhandle).add(qhandle, qhandle, cb, profileId);
}

function resultsRunner(qhandle) {
  let processId = qhandleProcessId(qhandle);

  if (arc.data.RESULTS_RUNNER[processId]) {
    return arc.data.RESULTS_RUNNER[processId];
  }

  arc.data.RESULTS_RUNNER[processId] =
      arc.data.batch()
              .type('Results')
              .url(`${arc.frontend.globals.urls.sqlRunUrl}/../${processId}/asyncresults_async`);

  return arc.data.RESULTS_RUNNER[processId];
}

function asyncResults(qhandle, cb) {
  resultsRunner(qhandle).add(qhandle, qhandle, cb);
}

function $ajax(params) {
  return new Promise((resolve, reject) => {
    $.ajax(params)
      .then(resolve)
      .fail(reject);
  });
}

function prepareSessionForFileDownload(url, data) {
  return $ajax({
    url,
    data,
    type: 'POST',
    beforeSend: function (xhr) {
      // Because of a jquery bug when handling PUT requests,
      // we have to explicitly set X-CSRFToken here
      xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
    },
  });
}

function requestFileAsBlob(url) {
  return arc.frontend.vizHttp.get({
    url,
    responseType: 'blob',
  }).catch(async (error) => {
    const err = await error.response.data.text();
    throw err;
  });
}

function downloadFile(blob, fileName) {
  const link = document.createElement('a');
  link.download = fileName;
  const objectUrl = URL.createObjectURL(blob);
  link.href = objectUrl;
  link.dispatchEvent(new MouseEvent('click'));

  // object URLs need to be revoked in order to avoid memory leak
  const cleanupTimeoutSeconds = arc.frontend.globals.settings.objectUrlCleanupSeconds;
  console.info(`Generated object url "${objectUrl}". Will revoke it after ${cleanupTimeoutSeconds} seconds.`);
  setTimeout(() => {
    // eslint-disable-next-line no-console
    console.info('Revoking object url', objectUrl);
    URL.revokeObjectURL(objectUrl);
  },
  cleanupTimeoutSeconds * 1000);
}

arc.data.seqnum = 0;

arc.data.runQueryParams = function(config) {
  let {
    rb,
    successCb,
    failCb,
    lib,
    errorReportingCb,
    timeoutSeconds,
    timeoutCb,
    queryCmd,
    loader,
  } = config;

  return arc.data.runQuery(rb, successCb, failCb, lib, errorReportingCb,
    timeoutSeconds, timeoutCb, queryCmd, loader);
};

/*
 * This interface should be simpler:
 *
 *   At most there should be two completion callbacks -- pass/fail.  Better would be a
 *   single callback with appropriate status to indicate what happened.
 *
 *   Should take only an arc.data.queryCmd() rather than an arc.requestBuilder() since the
 *   former is the full container for the host request parameters.
 */
arc.data.runQuery = function(rb, successCb, failCb, lib, errorReportingCb,
    timeoutSeconds, timeoutCb, queryCmd, loader) {
  var queryid = arc.data.seqnum++;

  const dataset_id = (rb && rb.dsreq() && rb.dsreq().dataset_id) ||
                     (queryCmd && queryCmd.dataset_id());

  // no idea what lib is supposed to be here
  // the query state tries to hide the loading indicator using it
  var qstate = queryState({
    seqnum: 'unmanaged_' + queryid,
    queryid,
    dataset_id,
    errorReportingCb,
    loader,
    lib: lib?.hideLoader ? lib : undefined,
  });

  if (queryCmd === undefined) {
    queryCmd = arc.data.queryCmd()
                  .dsreq(JSON.stringify(rb.dsreq()))
                  .sync(false);
  }

  var params = queryCmd.getParams();

  let callMade = false;

  if (loader) {
    loader.cancelCb(function() {
            qstate.cancel();
            if (failCb) {
              failCb('Cancelled');
            }
          });

    loader.reset();
    loader.showProgress();
  }

  if (timeoutSeconds) {
    // Handle timeouts in the UI
    // TODO: Cancel the backend query
    setTimeout(function() {
      if (callMade) {
        return;
      }

      callMade = true;
      qstate.cancel();

      if (failCb) {
        failCb('Timeout waiting for response');
      }

      if (timeoutCb) {
        timeoutCb();
      }
    }, timeoutSeconds * 1000);
  }

  var baseQuery = arc.data.baseQuery(qstate,
      function(...args) {
        if (callMade) {
          return;
        }

        callMade = true;
        successCb(...args);
      },

      function(...args) {
        if (callMade) {
          return;
        }

        callMade = true;
        if (failCb) {
          failCb(...args);
        }
      },
      lib);

  baseQuery.run(params);
  return baseQuery;
};

/*
 * Help track query state for a single query. This helper object handles
 *   state change reporting, url generation, and error reporting.
 *
 * config is:
 *   activeQueryManager: query manager
 *   seqnum: sequence number
 *   lib: library object for alerts and for loader
 *   query: query
 *   dataset_id: for error reporting
 *   queryid: query id
 *   errorReportingCb: function to use for error reporting
 */
function queryState(config) {
  var qstate = function() { return 'queryState'; };

  let {
    activeQueryManager,
    seqnum,
    lib,
    query,
    queryid,
    dataset_id,
    errorReportingCb,
  } = config;

  var state = arc.queryStates.RUNNING,
      startTime = Date.now(),
      processingResponse,
      streamConsumed,
      queryStartTime = new Date(),
      qhandle;

  function logState(initial, oldstate, newstate) {
    if (arc.queryStateLogging) {
      if (initial) {
        console.log('Query ' + queryid + ' state set to ' + state);
      } else {
        console.log('Query ' + queryid + ' state update from ' + oldstate +
            ' to ' + newstate);
      }
    }
  }

  logState(true);

  qstate.url = function(code) {
    let processId = qhandleProcessId(qhandle);

    return {
      url: `${arc.frontend.globals.urls.sqlRunUrl}/../${processId}/async${code}_async`,
      paramsToSend: {
        qhandles: qhandle,
      },
    };
  };

  qstate.query = function() {
    return query;
  };

  qstate.queryid = function() {
    return queryid;
  };

  qstate.seqnum = function() {
    return seqnum;
  };

  qstate.running = function() {
    return ((state === arc.queryStates.RUNNING) && !processingResponse);
  };

  qstate.state = function() {
    return state;
  };

  // Number of seconds since started
  qstate.elapsed = function() {
    return ((Date.now() - startTime) / 1000.0);
  };

  qstate.qhandle = function(_) {
    if (!arguments.length) return qhandle;
    qhandle = _;
    return qstate;
  };

  qstate.streamConsumed = function(_) {
    if (!arguments.length) return streamConsumed;
    streamConsumed = _;
    return qstate;
  };

  // Cancel a running query
  qstate.cancel = function() {
    if (state === arc.queryStates.RUNNING) {
      qstate.update(arc.queryStates.CANCEL);
      if (queryid) {
        arc.data.QUERY_RUNNER.cancel(queryid);
      }
    }
  };

  // Transition a query tracking structure to a new state
  qstate.update = function(newstate) {
    const oldstate = state;

    logState(false, oldstate, newstate);

    state = newstate;

    const completedQueryStates = [
        arc.queryStates.CANCELLED,
        arc.queryStates.ERROR,
        arc.queryStates.FINISHED,
    ];

    if (completedQueryStates.includes(newstate)) {
      if (activeQueryManager) {
        activeQueryManager.markInactive(qstate.seqnum);
      }

      const t = new Date(),
            diff = t - queryStartTime;

      arc.counters.increment('queryCount');
      arc.counters.data.queryTime += diff;
      arc.counters.data.maxQueryTime = Math.max(diff, arc.counters.data.maxQueryTime);
    }

    if (newstate !== arc.queryStates.RUNNING) {
      if (oldstate === arc.queryStates.RUNNING) {
        if (lib) {
          lib.hideLoader();
        }
      }

      if (activeQueryManager) {
        activeQueryManager.markInactive(seqnum);
      }
    }
  };

  // Hide any loader window, report an error, and do cleanup
  qstate.reportQueryError = function(message, error) {

    // Not needed if query state already transitioned from running
    if (state !== arc.queryStates.RUNNING) {
      return;
    }

    if (errorReportingCb) {
      errorReportingCb(qstate, message, error.response || error);
    } else {
      const qhandleMsg = qhandle ? `${qhandle}: ` : '';

      const dmsg = dataset_id ? `Dataset ${dataset_id}. ` : '';

      alertMessage(lib, `${qhandleMsg}${dmsg}${message}`, error.response || error);
    }

    qstate.update(arc.queryStates.ERROR);

    return qstate;
  };

  // Hide the loader and do successCb() to render the data via a
  // timeout callback so that display is updated with the loader
  // changes during rendering
  qstate.renderQueryResults = function(successCb, _data) {
    if (lib) {
      lib.hideLoader();
    }

    processingResponse = true;
    if (_data.query_warnings?.inf.length || _data.query_warnings?.nan.length) {
      showInfinityOrNanValueWarning(_data);
    }

    arc.sqlrun.fetchDataRows(_data, function(data) {
      setTimeout(function runAfter0seconds() {
        if (state === arc.queryStates.RUNNING) {
          successCb(data, qstate);
        }

        qstate.update(arc.queryStates.FINISHED);
      }, 0);
    });

    return qstate;
  };

  return qstate;
}

arc.data.queryCmd = function() {
  var queryCmd = function() {
    return 'arc.data.queryCmd';
  };

  var query,
      timeoutSeconds,
      dsreq,
      cache_read = arc.globals.query.cache_read,
      sync,
      sample_pct,
      viz_params,
      widget_id,
      widget_name,
      app_id,
      app_name,
      appgroup_id,
      appgroup_name,
      dataset_id,
      dataset_name,
      skip_dsreq,
      dataconnection_id,
      dataconnection_name;

  queryCmd.query = function(_) {
    if (!arguments.length) return query;
    query = _;
    return queryCmd;
  };

  queryCmd.timeoutSeconds = function(_) {
    if (!arguments.length) return timeoutSeconds;
    timeoutSeconds = _;
    return queryCmd;
  };

  queryCmd.dsreq = function(_) {
    if (!arguments.length) return dsreq;
    dsreq = _;
    return queryCmd;
  };

  queryCmd.cache_read = function(_) {
    if (!arguments.length) return cache_read;
    cache_read = _;
    return queryCmd;
  };

  queryCmd.sync = function(_) {
    if (!arguments.length) return sync;
    sync = _;
    return queryCmd;
  };

  queryCmd.sample_pct = function(_) {
    if (!arguments.length) return sample_pct;
    sample_pct = _;
    return queryCmd;
  };

  queryCmd.vizParams = function(_) {
    if (!arguments.length) return viz_params;
    viz_params = _;
    return queryCmd;
  };

  queryCmd.widget_id = function(_) {
    if (!arguments.length) return widget_id;
    widget_id = _;
    return queryCmd;
  };

  queryCmd.widget_name = function(_) {
    if (!arguments.length) return widget_name;
    widget_name = _;
    return queryCmd;
  };

  queryCmd.app_id = function(_) {
    if (!arguments.length) return app_id;
    app_id = _;
    return queryCmd;
  };

  queryCmd.app_name = function(_) {
    if (!arguments.length) return app_name;
    app_name = _;
    return queryCmd;
  };

  queryCmd.appgroup_id = function(_) {
    if (!arguments.length) return appgroup_id;
    appgroup_id = _;
    return queryCmd;
  };

  queryCmd.appgroup_name = function(_) {
    if (!arguments.length) return appgroup_name;
    appgroup_name = _;
    return queryCmd;
  };

  queryCmd.dataset_id = function(_) {
    if (!arguments.length) return dataset_id;
    dataset_id = _;
    return queryCmd;
  };

  queryCmd.dataset_name = function(_) {
    if (!arguments.length) return dataset_name;
    dataset_name = _;
    return queryCmd;
  };

  queryCmd.dataconnection_id = function(_) {
    if (!arguments.length) return dataconnection_id;
    dataconnection_id = _;
    return queryCmd;
  };

  queryCmd.dataconnection_name = function(_) {
    if (!arguments.length) return dataconnection_name;
    dataconnection_name = _;
    return queryCmd;
  };

  queryCmd.skip_dsreq = function(_) {
    if (!arguments.length) return skip_dsreq;
    skip_dsreq = _;
    return queryCmd;
  };

  queryCmd.getParams = function() {
    var params = {
      query: query || '',
      dsreq: dsreq || '',
      async: (sync ? 'false' : 'true'),
      cache_read: (cache_read ? 'true' : 'false'),
      parameters_case_sensitive: arc.globals.parameters_case_sensitive,
      is_edit: arc.globals.is_edit,
    };

    if ((sample_pct) && (sample_pct !== 'Off')) {
      params.sample_pct = sample_pct;
    }

    if (viz_params) {
      params.viz_params = viz_params;
    }

    if (widget_id) {
      params.report_id = widget_id;
    }

    if (widget_name) {
      params.report_name = widget_name;
    }

    if (app_id) {
      params.parent_report_id = app_id;
    }

    if (app_name) {
      params.parent_report_name = app_name;
    }

    if (appgroup_id) {
      params.appgroup_id = appgroup_id;
    }

    if (appgroup_name) {
      params.appgroup_name = appgroup_name;
    }

    if (dataset_id) {
      params.dataset_id = dataset_id;
    }

    if (dataset_name) {
      params.dataset_name = dataset_name;
    }

    if (dataconnection_id) {
      params.dataconnection_id = dataconnection_id;
    }

    if (dataconnection_name) {
      params.dataconnection_name = dataconnection_name;
    }

    // For server side dsreq generation/validation, we shift the
    // dsreq phase information to the outer params structure
    if (arc.globals.arcviz_settings.DSREQ_METHOD && params.dsreq) {
      let dsnew = JSON.parse(params.dsreq);
      if (dsnew.phase) {
        params.phase = dsnew.phase;
        delete dsnew.phase;
        params.dsreq = JSON.stringify(dsnew);
      }
    }

    // Here is where we drop the dsreq and query from the request if skip_dsreq is set
    if (skip_dsreq) {
      params.dsreq = '';
      params.query = '';
    }

    return params;
  };

  return queryCmd;
};

/*
 * Every visual has one of these query state manager objects.
 *
 * As the visual fires off queries, this object allocates and tracks the
 *   current queryState object.
 */
arc.data.activeQueryManager = function(lib_) {
  var activeQueryManager = function() { return 'arc.data.activeQueryManager'; };

  var lib = lib_,

      // Every new query is given a sequence number for tracking
      querySequence = 0,

      // The currently known query sequence, and query, are stored here
      activeSequence,
      activeQuery;

  // Return any currently active query
  activeQueryManager.activeQuery = function() {
    return activeQuery;
  };

  // Mark a particular query as being inactive
  activeQueryManager.markInactive = function(seqnum) {
    if (seqnum === activeSequence) {
      activeSequence = undefined;
      activeQuery = undefined;
    }

    return activeQueryManager;
  };

  // Allocate a query tracking structure
  activeQueryManager.queryStateAllocate = function(_lib, query, errorReportingCb) {
    const reportId = _lib ? _lib.report_id() : 'unknown';
    const queryid = `${reportId}.${querySequence}`;
    const qstate = queryState({
      errorReportingCb,
      activeQueryManager,
      seqnum: querySequence,
      lib: _lib,
      queryid,
      query,
      loader: _lib && _lib.loader(),
    });

    querySequence += 1;

    if (arc.queryStateLogging) {
      // eslint-disable-next-line no-console
      console.log(`Query ${queryid}`);
    }

    activeQuery = qstate;
    activeSequence = qstate.seqnum();

    return qstate;
  };

  return activeQueryManager;
};

/*
 * This object provides two important functions
 *
 * run() -> execute the sql query and make the success cb on completion
 *
 * exportCSV() -> use a streaming interface to download csv directly from
 *     the server.
 */
arc.data.baseQuery = function(qstate_, successCb_, failCb_, lib_, chunkCb) {
  var baseQuery = function() {
    return 'arc.data.baseQuery';
  };

  var qstate = qstate_,
      successCb = successCb_,
      failCb = failCb_,
      lib = lib_,
      apiParams;

  baseQuery.qstate = qstate;

  function showUnformattedColumnsWarning(formattingSettings) {
    if (!formattingSettings) {
      return;
    }

    const cannotBeFormatted = formattingSettings.filter(Boolean).reduce((acc, setting) => {
      if (['js', 'date'].includes(setting.type)) {
        acc.push(setting);
      }

      return acc;
    }, []);

    if (cannotBeFormatted.length > 0) {
      arc.report.alertMessageSmallBase({
        type: 'alert-warning',
        alertHeader: '',
        alertMessage: `Display formatter won't be applied on the following columns:
          <strong>${cannotBeFormatted.map((setting) => setting.colName).join(', ')}</strong>`,
        fadeOut: true,
        useHTML: true
      });
    }
  }

  baseQuery.exportCSV = function(
    filename,
    languageCode,
    dsreqOverride,
    csvQueryLimit,
    quotationSetting,
    formattingSettings,
    forceFullDownload
  ) {
    apiParams.csvfilename = filename;
    apiParams.language_code = languageCode;
    apiParams.cache_read = arc.globals.query.cache_read;
    apiParams.quotation_mark = quotationSetting;
    apiParams.parameters_case_sensitive = arc.globals.parameters_case_sensitive;
    apiParams.is_edit = arc.globals.is_edit;

    if (dsreqOverride) {
      apiParams.dsreq = dsreqOverride;
    }

    showUnformattedColumnsWarning(formattingSettings);

    if (forceFullDownload) {
      const dsreq = JSON.parse(apiParams.dsreq);
      dsreq.dimensions = [];
      dsreq.aggregates = [];
      apiParams.dsreq = JSON.stringify(dsreq);
    }

    if (csvQueryLimit) {
      if (apiParams.dsreq) {
        const dsreq = JSON.parse(apiParams.dsreq);
        dsreq.limit = csvQueryLimit;
        apiParams.dsreq = JSON.stringify(dsreq);
      } else if (apiParams.query) {
        apiParams.query = arc.misc.addLimitToSQLQuery(
          apiParams.query,
          csvQueryLimit
        );
      }
    }

    if (formattingSettings && !forceFullDownload) {
      apiParams.formattingSettings = JSON.stringify(formattingSettings);
    } else {
      apiParams.formattingSettings = null;
    }

    const pingUrl = arc.frontend.globals.urls.csvDownloadUrl;
    const downloadUrl = `${arc.frontend.globals.urls.sqlRunUrl}?csvdownload=true`;
    const alertMessageGroupHandler = arc.report.createAlertMessageGroup();

    prepareSessionForFileDownload(pingUrl, apiParams)
      .then(() => alertMessageGroupHandler.addDelayedMessage('Sending query ...', 'alert-info'))
      .then(() => requestFileAsBlob(downloadUrl))
      .then((response) => downloadFile(response.data, apiParams.csvfilename))
      .then(() => alertMessageGroupHandler.addFinalMessage('Download Finished!', 'alert-success'))
      .catch((reqOrError = {}) => {
        const errorMessage = 'Failed to download CSV';
        const detail = typeof reqOrError === 'string' ? reqOrError : reqOrError.responseText || reqOrError.message || '';

        alertMessage(lib, errorMessage, detail);

        alertMessageGroupHandler.addFinalMessage(errorMessage, 'alert-danger');
      });
  };

  baseQuery.exportFileViaSnapshot = function ({
    filename,
    languageCode = null,
    dsreqOverride = null,
    queryLimit = null,
    quotationSetting = null,
    formattingSettings = null,
    forceFullDownload = false,
    fileFormat = 'csv', // 'csv' or 'xlsx'
  }) {
    const snapshotAdmin = arc.snapshot().is_dashboard(false);
    const apiParamsCopy = Object.assign({}, apiParams);

    if (fileFormat === 'csv') {
      apiParamsCopy.csvfilename = filename;
      apiParamsCopy.language_code = languageCode;
      apiParamsCopy.quotation_mark = quotationSetting;
      apiParamsCopy.parameters_case_sensitive = arc.globals.parameters_case_sensitive;
      apiParamsCopy.is_edit = arc.globals.is_edit;
    } else if (fileFormat === 'xlsx') {
      apiParamsCopy.xlsxfilename = filename;
    }

    apiParamsCopy.cache_read = arc.globals.query.cache_read;

    if (dsreqOverride) {
      apiParamsCopy.dsreq = dsreqOverride;
    }

    showUnformattedColumnsWarning(formattingSettings);

    if (forceFullDownload && apiParamsCopy.dsreq) {
      const dsreq = JSON.parse(apiParamsCopy.dsreq);
      dsreq.dimensions = [];
      dsreq.aggregates = [];
      apiParamsCopy.dsreq = JSON.stringify(dsreq);
    }

    if (queryLimit) {
      if (apiParamsCopy.dsreq) {
        const dsreq = JSON.parse(apiParamsCopy.dsreq);
        dsreq.limit = queryLimit;
        apiParamsCopy.dsreq = JSON.stringify(dsreq);
      } else if (apiParamsCopy.query) {
        apiParamsCopy.query = arc.misc.addLimitToSQLQuery(apiParamsCopy.query, queryLimit);
      }
    }

    if (formattingSettings && (!forceFullDownload || fileFormat === 'xlsx')) {
      apiParamsCopy.formattingSettings = JSON.stringify(formattingSettings);
    } else {
      apiParamsCopy.formattingSettings = null;
    }

    apiParamsCopy.uploadFormat = fileFormat;

    const url = `/arc/apps/snapshot/${apiParamsCopy.report_id}`;
    arc.frontend.vizHttp.post({
      url,
      data: new URLSearchParams(apiParamsCopy).toString(),
      headers: {
        'Content-Type': 'application/x-www-form-urlencoded',
        'X-CSRFToken': arc.globals.csrf_token,
      },
    })
      .then((response) => {
        const resp = response.data;
        arc.report.alertMessageSmallBase({
          type: 'alert-success',
          alertHeader: 'Generating the requested file',
          alertMessage: 'The file will be available in a few seconds...',
          fadeOut: false,
        });

        const { snapshotId } = resp;
        const fileName = arc.misc.getTimestampFilename('test_title');

        function downloadSnapshot() {
          const snapshotLink = document.createElement('a');
          snapshotLink.download = fileName;
          snapshotLink.href = `${arc.frontend.globals.urls.snapshotFetchUrl}/${snapshotId}?title=${fileName}`;
          if (typeof MouseEvent === 'function') {
            snapshotLink.dispatchEvent(new MouseEvent('click'));
          } else {
            $(snapshotLink)[0].click();
          }

          arc.report.alertMessageSmallHide();
        }

        snapshotAdmin.downloadSnapshotIfReady(
          snapshotId,
          apiParamsCopy.report_id,
          downloadSnapshot,
          0
        );
      })
      .catch((error) => {
        arc.report.alertMessageSmallHide();
        arc.report.alertMessage(
          `Failed to download ${fileFormat.toUpperCase()}`,
          error.response?.data || '',
          { alert_report_id: apiParamsCopy.report_id }
        );
      });
  };

  baseQuery.exportXlsx = function (filename, formattingSettings, queryRowLimit, dsreqOverwrite) {
    const excelApiParams = Object.assign({}, apiParams);
    excelApiParams.xlsxfilename = filename;
    excelApiParams.cache_read = arc.globals.query.cache_read;

    showUnformattedColumnsWarning(formattingSettings);

    // dsreqOverwrite is used to overwrite dsreq for expanded scenarios
    if (dsreqOverwrite) {
      excelApiParams.dsreq = dsreqOverwrite;
    }

    if (queryRowLimit) {
      if (excelApiParams.dsreq) {
        const dsreq = JSON.parse(excelApiParams.dsreq);
        dsreq.limit = queryRowLimit;
        excelApiParams.dsreq = JSON.stringify(dsreq);
      } else if (excelApiParams.query) {
        excelApiParams.query = arc.misc.addLimitToSQLQuery(
          excelApiParams.query,
          queryRowLimit
        );
      }
    }

    if (formattingSettings) {
      excelApiParams.formattingSettings = JSON.stringify(formattingSettings);
    }

    const pingUrl = arc.frontend.globals.urls.xlsxdownloadUrl;
    const downloadUrl = `${arc.frontend.globals.urls.sqlRunUrl}?xlsxdownload=true`;
    const alertMessageGroupHandler = arc.report.createAlertMessageGroup();

    prepareSessionForFileDownload(pingUrl, excelApiParams)
      .then(() => alertMessageGroupHandler.addDelayedMessage('Sending query ...', 'alert-info'))
      .then(() => requestFileAsBlob(downloadUrl))
      .then((response) => downloadFile(response.data, excelApiParams.xlsxfilename))
      .then(() => alertMessageGroupHandler.addFinalMessage('Download Finished!', 'alert-success'))
      .catch((reqOrError = {}) => {
        const errorMessage = 'Failed to download Excel';
        const detail = typeof reqOrError === 'string' ? reqOrError : reqOrError.responseText || reqOrError.message || '';

        alertMessage(lib, errorMessage, detail);

        alertMessageGroupHandler.addFinalMessage(errorMessage, 'alert-danger');
      });
  };

  function handleResponse(resolve, reject) {
    if (!resolve) {
      resolve = () => {};
    }

    if (!reject) {
      reject = () => {};
    }

    const loader = lib ? lib.loader() : undefined;
    const report_id = lib ? lib.report_id() : undefined;

    createResponseController(
      qstate, successCb, loader, report_id, resolve, reject, chunkCb
    ).pollAndFetch();
  }

  baseQuery.run = function(parms) {
    apiParams = $.extend({}, parms);

    if (qstate.qhandle()) {
      handleResponse();
      return;
    }

    /*
     * NOTE: the 'dataapi_start' counter is used in arcweb/apps/views_snapshot.py
     * It is incremented from a few other places as well, so we have a distinct
     * 'dataapi_query_start' counter to track running queries and pop a message
     * onto dashboards for long-running queries.
     */
    arc.counters.increment('dataapi_start');
    arc.counters.increment('dataapi_query_start');
    if (arc.globals.dashboard) {
      arc.globals.dashboard.queryStarted();
    }

    function decrement() {
      arc.counters.increment('dataapi_end');
      arc.counters.increment('dataapi_query_end');
    }

    var t1 = new Date();

    new Promise(function(resolve, reject) {
      queryRun(qstate.queryid(), parms,
          function(error, response) {

        var t2 = new Date();

        arc.perf.PAGEPERF.addProfile({
          type: 'QueryRun',
          time: t2 - t1,
        });

        if (error) {
          reject(new Error('Could not load data', { cause: error }));
          qstate.reportQueryError('Could not load data', error);
          arc.counters.increment('viz_error_end');
          if (failCb) {
            failCb(error);
          }

          return;
        }

        try {
          if (response.response.qhandle) {
            qstate.qhandle(response.response.qhandle);
            handleResponse(resolve, reject);
            return;
          }
        } catch (err) {
          reject(new Error(err));
          console.log(err);
          qstate.reportQueryError('Could not parse response', response);
          return;
        }

        // Here rather than a handle back, we just got back the results
        // (happens when the result obtained from cache).  So we need
        // to ensure no cancel request before taking the successCb path.
        if (qstate.state() !== arc.queryStates.RUNNING) {
          if (qstate.state() === arc.queryStates.CANCEL) {
            console.log('Canceling query');
            resolve();
          } else {
            reject(new Error(`Query state is ${qstate.state()}`));
          }

          qstate.update(arc.queryStates.CANCELLED);
          return;
        }

        /*
         * KSQL query results contain a qhandle so further streaming results
         * can be fetched.
         */
        if (response.response.ongoing_qhandle) {
          qstate.qhandle(response.response.ongoing_qhandle);
        }

        qstate.renderQueryResults(successCb, response.response);
        resolve();
      }, (lib ? lib.report_id() : arc.report.REPORTID_NEW));
    })
    .then(decrement)
    .catch(function(error) {
      // Question: we reject all the default errors in the call and catch it here. It is intended?
      if (error.message !== 'Cancelled') {
        qstate.reportQueryError('Error fetching data', error);
      }

      decrement();
    });
  };

  return baseQuery;
};

function createResponseController(qstate_, successCb_, loader_, report_id_,
    resolve_, reject_, chunkCb) {
  var responseController = function() {
    return 'responseController';
  };

  var qstate = qstate_,
      successCb = successCb_,
      loader = loader_,
      report_id = report_id_,
      resolve = resolve_,
      reject = reject_;

  if (!resolve) {
    resolve = () => {};
  }

  if (!reject) {
    reject = () => {};
  }

  var unload_handle;

  function handleResponse(error, response) {
    if (!qstate.running()) {
      reject(new Error('Query is no longer running'));
      qstate.update(arc.queryStates.CANCELLED);
      return;
    }

    if (error) {
      reject(new Error(error));
      qstate.reportQueryError('Could not load data', error);
      return;
    }

    var o;
    try {
      resolve();
      o = response.response;
      qstate.renderQueryResults(successCb, o);
    } catch (err) {
      reject(new Error(err));
      qstate.reportQueryError('Could not parse response', response);
    }
  }

  function fetchResponse() {
    asyncResults(qstate.qhandle(), handleResponse);
  }

  // Poll for async query results and process it when available.
  // Cancel query if requested.
  responseController.pollAndFetch = function() {

    // Drop the unload function if needed
    function drop_unload() {
      if (unload_handle) {
        arc.report.pagenav.del_unload(unload_handle);
        unload_handle = undefined;
      }
    }

    // Return false if no cancel request, else return true with cancel posted
    function checkCancel() {
      if (qstate.state() === arc.queryStates.CANCEL) {
        qstate.update(arc.queryStates.CANCELLING);
        asyncCancel(qstate.qhandle(), function(error, response) {
          qstate.update(arc.queryStates.CANCELLED);
        }, report_id);

        drop_unload();
        return true;
      }

      return false;
    }

    if (checkCancel()) {
      reject(new Error('Cancelled'));
      return responseController;
    }

    if (!unload_handle) {
      unload_handle = arc.report.pagenav.add_unload(function () {
        if (qstate.state() === arc.queryStates.RUNNING && qstate.qhandle()) {
          qstate.update(arc.queryStates.CANCELLING);
          const d = qstate.url('cancel');
          $.ajax({
            url: d.url,
            data: d.paramsToSend,
            type: 'POST',
            beforeSend: function(xhr) {
              // Because of a jquery bug when handling PUT requests,
              // we have to explicitly set X-CSRFToken here
              xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
            },
          });
        }
      });
    }

    asyncStatus(qstate.qhandle(), function(error, response) {
      if (checkCancel()) {
        reject(new Error('Cancelled'));
        return;
      }

      if (!qstate.running()) {
        reject(new Error('Query is no longer running'));
        drop_unload();
        return;
      }

      if (error) {
        reject(new Error(error));
        drop_unload();
        qstate.reportQueryError('Could not load data', error);
        return;
      }

      var o;
      try {
        o = response.response;
        if (o.status) {
          if (o.status === 'StreamConsumed') {
            resolve();
            qstate.streamConsumed(true);
            successCb(undefined, qstate);
            return;
          }

          if ((o.status === 'Finished') || (o.status.substr(0, 3) === '(HY')) {
            fetchResponse();
          } else {
            if (qstate.state() === arc.queryStates.RUNNING) {
              if (o.chunk_text && chunkCb) {
                chunkCb(o.chunk_text);
              }

              if (loader && o.progress) {
                loader.total(Number(o.total));
                loader.progress(Number(o.progress));
              }
            }

            // Slow down polling as we wait longer
            var elapsed = qstate.elapsed();
            var nextpoll = (elapsed < 2)  ? 0.5 :
                           (elapsed < 10) ? 1 :
                           (elapsed < 60) ? 5 : 10;

            setTimeout(responseController.pollAndFetch, nextpoll * 1000);
          }
        } else {
          resolve();
          drop_unload();
          handleResponse(undefined, response);
        }
      } catch (err) {
        reject(new Error(err));
        console.log(err);
        drop_unload();
        qstate.reportQueryError('Could not parse response', response);
      }
    }, report_id);

    return responseController;
  };

  return responseController;
}

}());
