window.arc = window.arc || {};

(function() {
"use strict";

arc.datasets = arc.datasets || {};

function extractsEnabled() {
  return arc.frontend.featureFlipper.getValueById('site_enable_extracts');
}

/*
  Column name notes:

    We want every column in a dataset to have a unique 'Display Name'.  This is the
    name displayed in Dimensions/Measures lists and on the report shelves.

    For example, with a column like 'user_id', we get:

      Column Name          | Display Name   | In expression   | In SQL
      db.tab.user_id       | User Id        | [User Id]       | db.tab.user_id

    Note that when a dimension table exists more than once in the dataset, it must
    be refered to by the table alias so the SQL name must be: 'table_alias.user_id'.

    DJC -- Should not need 'basecol' any more for custom expression
*/

arc.datasets.quoteColname = function(colname) {
  if (colname[0] !== '[') {
    return `[${colname}]`;
  }
};

// Nearly everywhere we use fully qualified tablenames (<dbname>.<tablename>).  This
// function extracts just the tablename portion.
arc.datasets.drop_dbname = function(fqname) {
  return fqname.substr(fqname.indexOf('.') + 1);
};

arc.datasets.just_dbname = function(fqname) {
  return fqname.substr(0, fqname.indexOf('.'));
};

// Strip outer quotes of a string, less overhead compared to RegEx
arc.datasets.stripQuotes = function(str) {
  if (str && str.length > 2 && "`'\"".indexOf(str[0]) !== -1 && str[0] === str[str.length - 1]) {
    str = str.substr(1, str.length - 2);
  }
  return str;
};

// True if type is a complex type
arc.datasets.is_complex_type = function(type) {
  return (['map', 'struct', 'array'].indexOf((type || '').toLowerCase()) !== -1);
};

// Routine to find a matching entry in table of lists
arc.datasets.findListEntry = function(table, text, index, xlat) {
  for (var i = 0; i < table.length; i++) {
    if (text === table[i][index]) {
      return (xlat === undefined ? table[i] : table[i][xlat]);
    }
  }
};

// A table to convert between display text and agg function
arc.datasets.dflagg_table = [
    ['Sum', 'sum'],
    ['Count', 'count'],
    ['Approx Distinct Count', 'ndv'],
    ['Exact Distinct Count', 'countdistinct'],
    ['Minimum', 'min'],
    ['Maximum', 'max'],
    ['Average', 'avg'],
    ['String Concat', 'group_concat']
  ];

// List of valid column attributes (value 2 indicates custom attribute).
// If "ispart" is true -> the col is partition col of base table that the dataset is based on, while
// only the non partition col can have "ispartition" value True and be used as partition for AVs.
arc.datasets.ColumnAttrNames = {
  name: 1,
  type: 1,
  hivecomment: 1,
  isdim: 1,
  ispart: 1,
  alias: 1,    // Could be custom if doesn't match name
  parent: 1,   // Indicates parent column of complex type sub-column
  terms: 1,          // search terms
  searchable: 1,     // search setting for indexed columns
  recordcount: 1,    // indicates that this column is the Record Count column
  colors: 1,
  comment: 2,
  basecol: 2,
  expr: 2,
  hidden: 2,
  cast: 2,
  dflagg: 2,
  exprhasagg: 2,
  dimh: 2,
  ispartition: 2,
  geotype: 2,
  detaildata: 2,
  isSensitive: 2,
  displayformat: 2,
  useroverrodeisdim: 2,
};

// Represents a column within a table within a dataset
arc.datasets.Column = function() {
  var table,
      attr = {};

  var Column = function() { return 'arc.datasets.Column'; };

  Column.valuesNeedQuoting = function() {
    let dc_type;

    if (table) {
      let ds = table.dataset();
      if (ds) {
        dc_type = ds.dataconnection_type();
      }
    }

    return !arc.report.measure_types[Column.columnType()];
  };

  Column.isRemoteColumn = function() {
    return arc.cviz.isCvizExpr(attr.expr);
  };

  Column.remoteDataSettings = function() {
    if (!Column.isRemoteColumn()) {
      return undefined;
    }

    return arc.cviz.cvizSettingsFromExpr(attr.expr);
  };

  Column.isGeo = function() {
    if (Column.isHierarchy()) {
      return false;
    }

    const geoType = Column.attr('geotype');
    const hasGeoType = geoType && geoType !== 'none';

    return (hasGeoType || arc.geo.autocastGeoType(Column.display_name()));
  };

  /*
   * Normally, a column's display_name() can be used as its expression.
   * However, when a new column is made, we need to be able to access its
   * values at times, and because the column is new in the UI and has not
   * yet been communicated to the backend, we need to use the underlying
   * expression for the column.
   */
  Column.expr = function() {
    return attr.expr || attr.name;
  };

  Column.isSensitive = function() {
    return attr.isSensitive === true;
  };

  Column.isAggregate = function() {
    if (Column.isHierarchy()) {
      return false;
    }

    return (Column.attr('isdim') === false);
  };

  Column.showInDetailData = function() {
    if (Column.isHierarchy()) {
      return false;
    }

    if (Column.isComplexType()) {
      return false;
    }

    if (Column.attr('hidden')) {
      return false;
    }

    if (Column.attr('detaildata') === false) {
      return false;
    }

    return true;
  };

  Column.coltype = function() {
    return Column.attr('cast') || Column.attr('type');
  };

  Column.isHierarchy = function() {
    let t = Column.attr('type');
    return (t && (t.toLowerCase() === 'hierarchy'));
  };

  // Return true if column has at least one custom attribute
  Column.has_custom_attribute = function() {
    // See if alias doesn't match column name
    if (attr.name !== attr.alias) return true;

    // See if any attributes are custom
    for (var k in attr) {
      if (arc.datasets.ColumnAttrNames[k] > 1) {
        return true;
      }
    }

    return false;
  };

  Column.isBaseCol = function () {
    // If column name is empty, it's a cloned or hierarchy column and can be deleted
    // Also, set up expression editor id if this is a cloned column
    // else delete autocomplete on/off button
    return attr.name;
  };

  Column.coltype_class = function() {
    // TODO: Remove the need to create shelf element and port the functionality here.
    var shelf_elem = arc.report.shelf.element()
            .dataset_colname(attr.name)
            .dataset_coltype(attr.type)
            .fn_cast(attr.cast);
    return shelf_elem.dataset_coltype_class();
  };

  Column.display_name = function () {
    return attr.alias || attr.name;
  };

  Column.defaultAggregate = function () {
    return attr.dflagg || 'SUM';
  };

  function quoteIfNecessary(s) {
    if (s.search(/\s/) >= 0) {
      return '"' + s + '"';
    }

    return s;
  }

  Column.terms = function() {
    return (attr.terms === '') ? '' : attr.terms ||
        quoteIfNecessary(Column.display_name());
  };

  Column.isDimension = function () {
    if (Column.isHierarchy()) {
      return false;
    } else if (Column.isComplexType() || Column.hasComplexParent()) {
      return true;
    }

    return Column.attr('isdim');
  };

  Column.isComplexType = function () {
    return arc.datasets.is_complex_type(attr.type);
  };


  Column.getComplexParent = function () {
    if (attr.parent) {
      return table.get_column(attr.parent);
    }

    return undefined;
  };

  Column.hasComplexParent = () => !!Column.getComplexParent();

  // Clone column.  'name' set indicates base vs derived and affects result
  Column.clone = function() {

    const c = arc.datasets.Column()
               .table(table);

    // Set up attributes for new column
    const a = Object.assign(c.attr(), attr);

    // Remove hive comments for cloned columns, copy them into user comments if
    // user comment's don't exist
    if (a.hivecomment && !a.comment) {
      a.comment = a.hivecomment;
    }
    delete a.hivecomment;

    // Remove search settings when cloning, otherwise we end up with multiple
    // columns containing the same search terms
    a.terms = undefined;
    a.searchable = false;

    // Cloned columns have no name plus when cloning a base column, we set up
    // the cloned column's expr and basecol entries from this base.
    a.name = '';
    if (attr.name) {
      a.expr = '[' + (attr.alias || attr.name) + ']';
      a.basecol = attr.name;
    }

    a.useroverrodeisdim = false;

    // Build valid alias
    var alias;
    var i = 0;
    do {
      i++;
      alias = ((i === 1) ? "Copy of " : "Copy " + i + " of ") + (attr.alias || attr.name);
    } while (table.dataset().get_column(alias) !== undefined);
    a.alias = alias;

    return c;
  };

  Column.table = function(_) {
    if (!arguments.length) return table;
    table = _;
    return Column;
  };

  Column.columnType = function() {
    return attr.cast || attr.type;
  };

  Column.attr = function(k, v) {
    switch (arguments.length) {
      case 0: return attr;

      case 1: {
        if (!arc.datasets.ColumnAttrNames[k]) {
          console.log("Unknown field access: ", k);
          return undefined;
        }

        return attr[k];
      }

      case 2: {
        if (!arc.datasets.ColumnAttrNames[k]) {
          console.log("Unknown field update:", k, "with", v);
          return undefined;
        }

        if (v === undefined) {
          delete attr[k];
        } else {
          attr[k] = v;
        }

        if (k === 'alias') {
          table.dataset().rebuild_hashes();
        }

        return Column;
      }
    }
  };

  Column.build_dsobject = function() {
    return Object.assign({}, attr);
  };

  // Helper inline function
  const _checkStr = prop => prop && typeof prop === 'string';

  Column.init_from_dsobject = function(dso) {
    // Copy `dso` to new object
    attr = Object.assign({}, dso);
    if (_checkStr(attr.alias)) {
      attr.alias = arc.datasets.stripQuotes(attr.alias);
    }

    attr.type = _checkStr(attr.type) ? attr.type.toUpperCase() : attr.type;
    attr.cast = _checkStr(attr.cast) ? attr.cast.toUpperCase() : attr.cast;

    return Column;
  };

  Column.init_from_old_dsobject = function(dso) {
    attr = Object.assign({}, dso[2]);
    attr.name = dso[1];
    attr.type = _checkStr(dso[0]) ? dso[0].toUpperCase() : dso[0];
    if (dso.hide) {   // Rename hide -> hidden
      attr.hidden = dso.hide;
    }
    delete attr.hide;
    if (attr.alias) {
      attr.alias = arc.datasets.stripQuotes(attr.alias);
    }
    return Column;
  };

  return Column;
};

// Represents a table within a dataset
arc.datasets.Table = function() {
  var tablename,
      dataset,
      sample_data,

      // Alias for this table within the dataset.  Note that a table is present in
      // the dataset once for each time is it joined--each instance must have
      // a unique name/alias and can have different column overrides.  This means
      // the table can exist once in the dataset without an alias.
      table_alias,

      // A 'table' can be the result of a select statement which is stored here
      table_select,

      // Parent table and join columns.
      // DJC -- Parent name might be alias which means if parent is given a new
      //        alias, we need to fix up this entry.
      parent_tablename,
      join_pairs = [],
      join_exprs = [],
      join_type = "left outer join",
      columns = [],
      foreign_keys = {};

  var Table = function() { return 'arc.datasets.Table'; };

  Table.tablename = function(_) {
    if (!arguments.length) return table_alias || tablename;
    tablename = _;
    return Table;
  };

  Table.table_select = function(_) {
    if (!arguments.length) return table_select;
    table_select = _;
    return Table;
  };

  Table.dataset = function(_) {
    if (!arguments.length) return dataset;
    dataset = _;
    return Table;
  };

  Table.sample_data = function(_) {
    if (!arguments.length) return sample_data;
    sample_data = _;
    return Table;
  };

  Table.columns = function(_) {
    if (!arguments.length) return columns;
    columns = _;
    return Table;
  };

  Table.tablename_noalias = function() {
    return tablename;
  };

  Table.tablename_withalias = function() {
    if (table_select) {
      return "(" + table_select + ") " + (table_alias || tablename).replace(".", "_");
    }
    return table_alias ? tablename + " " + table_alias : tablename;
  };

  Table.table_alias = function(_) {
    if (!arguments.length) return table_alias;
    table_alias = _;
    return Table;
  };

  Table.is_duplicate = function() {
    return table_alias || false;
  };

  Table.table_select = function(_) {
    if (!arguments.length) return table_select;
    table_select = _;
    return Table;
  };

  Table.parent_tablename = function(_) {
    if (!arguments.length) return parent_tablename;
    parent_tablename = _;
    return Table;
  };

  Table.join_pairs = function(_) {
    if (!arguments.length) return join_pairs;
    join_pairs = _;
    return Table;
  };

  Table.join_exprs = function(_) {
    if (!arguments.length) return join_exprs;
    join_exprs = _;
    return Table;
  };

  // Currently we only use one join type
  Table.join_type = function(_) {
    if (!arguments.length) return join_type;
    join_type = _;
    return Table;
  };

  Table.foreign_keys = function(tablename, columns) {
    if (!arguments.length) return foreign_keys;
    foreign_keys[tablename] = columns;
    return Table;
  };

  Table.dimensionsOnly = function() {
    return columns.filter(c => c.isDimension());
  };

  Table.dimensionsAndHierarchies = function() {
    return columns.filter(c => c.isDimension() || c.isHierarchy());
  };

  Table.measuresOnly = function() {
    return columns.filter(c => !c.isDimension() && !c.isHierarchy());
  };

  // Computed columns don't have names
  Table.base_columns = function() {
    return columns.filter(function(c) {return c.attr('name');});
  };

  Table.partitions = function() {
    var partnames = [];
    columns.forEach(function(c) {
      if (c.attr('ispart')) {
        partnames.push(c.attr('alias') || c.attr('name'));
      }
    });
    return partnames;
  };

  Table.init_complex_columns = function() {
    columns.forEach(c => {
      var name = c.attr('name') || '';
      var cpos = name.lastIndexOf('.');
      if (cpos !== -1) {
        c.attr('parent', name.substr(0, cpos));
        c.attr('isdim', true);
      }
    });
  };

  Table.init_from_url_response = function(coldata) {

    var qr = arc.queryResults().data(coldata);

    qr.columns().forEach(function(c) {
      var column = arc.datasets.Column()
                    .table(Table)
                    .attr('name', c[0])
                    .attr('type', c[1])
                    .attr('hivecomment', c[2])
                    .attr('isdim', !(arc.report.measure_types[c[1]]));
      columns.push(column);
    });

    qr.partitions().forEach(function(p) {
      var column = Table.get_column(p);
      if (column) {
        column.attr('ispart', true);
        column.attr('isdim', true);
      }
    });

    if (arc.dataconnection.connectionSupportsComplexTypes(Table.dataset().dataconnection_type())) {
      Table.init_complex_columns();
    }

    return Table;
  };

  Table.update_from_url_response = function(coldata) {
    let changed = false;
    const qr = arc.queryResults().data(coldata);
    const newcols = {};
    const nice = dataset && dataset.has_title_cased_names();

    // Build list of new columns
    qr.columns().forEach(function(c) {
        var newcol = arc.datasets.Column()
                    .table(Table)
                    .attr('name', c[0])
                    .attr('type', c[1])
                    .attr('hivecomment', c[2])
                    .attr('isdim', !(arc.report.measure_types[c[1]]));
        newcols[c[0]] = newcol;
      });

    // Process each base column for possible update/remove
    var dropcols = [];
    Table.base_columns().forEach(function(col) {

      // Flag if partition setting changed
      if ((col.attr('ispart') === true) !==
          (qr.partitions().indexOf(col.attr('name')) !== -1)) {
        changed = true;
      }

      // Clear current partition setting (updated later)
      col.attr('ispart', undefined);

      // Check against current column info
      var newcol = newcols[col.attr('name')];
      if (!newcol) {
        // Existing column has been removed from the table
        dropcols.push(col);
      } else if (col.attr('type') === newcol.attr('type') &&
        col.attr('hivecomment') === newcol.attr('hivecomment')) {
        // No change in name or type or hivecomment
        delete newcols[col.attr('name')];
      } else {
        // Type has changed or hivecomment has changed
        changed = true;
        col.attr('type', newcol.attr('type'));
        col.attr('isdim', newcol.attr('isdim'));
        col.attr('hivecomment', newcol.attr('hivecomment'));
        if (col.attr('type') === col.attr('cast')) {
          col.attr('cast', undefined);
        }
        delete newcols[col.attr('name')];
      }
    });

    // Drop all removed columns
    dropcols.forEach(function(col) {
      changed = true;
      var idx = columns.indexOf(col);
      if (idx !== -1) {
        columns.splice(idx, 1);
      }
    });

    // Add all remaining new columns
    for (var newname in newcols) {
      changed = true;
      if (nice) {
        newcols[newname].attr().alias = dataset.title_casing_names(newname);
      }
      columns.push(newcols[newname]);
    }

    // Process partitions
    qr.partitions().forEach(function(p) {
      var column = Table.get_column(p);
      if (column) {
        column.attr('ispart', true);
      }
    });

    Table.init_complex_columns();
    return changed;
  };

  Table.init_from_query = function(cnames, ctypes, query) {

    table_select = query;
    for (var i = 0; i < cnames.length; i++) {
      var column = arc.datasets.Column()
                    .table(Table)
                    .attr('name', cnames[i])
                    .attr('type', ctypes[i])
                    .attr('alias', cnames[i])
                    .attr('isdim', !(arc.report.measure_types[ctypes[i]]));
      columns.push(column);
    }
    Table.init_complex_columns();
    return Table;
  };

  Table.init_from_add_sq = function(cnames, ctypes) {

    for (var i = 0; i < cnames.length; i++) {
      var column = arc.datasets.Column()
                    .table(Table)
                    .attr('name', cnames[i])
                    .attr('type', ctypes[i])
                    .attr('isdim', !(arc.report.measure_types[ctypes[i]]));
      columns.push(column);
    }
    Table.init_complex_columns();
    return Table;
  };

  Table.get_colnames = function() {
    var colnames = [];
    columns.forEach(function(d) {
      if (d.attr('name')) {
        colnames.push(d.attr('name'));
      }
    });
    return colnames;
  };

  Table.get_column = function(colname, skipcol) {
    colname = colname.toLowerCase();
    for (var i = 0; i < columns.length; i++) {
      var col = columns[i];
      if (col === skipcol) {
        continue;
      }
      if (col.attr('name') && col.attr('name').toLowerCase() === colname) {
        return col;
      }
      if (col.attr('alias') && col.attr('alias').toLowerCase() === colname) {
        return col;
      }
    }
    return undefined;
  };

  // Clone a column
  Table.clone_column = function(srccol) {
    // Ensure the source column is present in this table
    var newcol;
    var idx;
    idx = columns.indexOf(srccol);
    if (idx !== -1) {
      newcol = srccol.clone();
      columns.splice(idx + 1, 0, newcol);
      dataset.add_column(newcol);
    }
    return newcol;
  };

  // Drop a column -- only allowed on cloned columns (i.e. with no name (only alias))
  Table.drop_column = function(col) {
    if (col.attr('name')) {
      return;
    }

    // Find the source column and drop it
    const idx = columns.indexOf(col);
    if (idx !== -1) {
      columns.splice(idx, 1);
    }

    dataset.rebuild_hashes();
  };

  Table.build_dsobject = function(skipcol) {
    const dso = {};
    dso.tablename = tablename;
    dso.table_alias = table_alias;
    if (parent_tablename) {
      dso.parent_tablename = parent_tablename;
      dso.join_pairs = join_pairs;
      dso.join_exprs = join_exprs;
      dso.join_type = join_type;
    } else {
      dso.enable_join_elim = dataset.enable_join_elim();
      dso.enable_incremental_results = dataset.enable_incremental_results();
    }
    if (table_select) {
      dso.select = table_select;
    }

    dso.columns = [];
    columns.forEach(function(c) {
      if (c !== skipcol) {
        dso.columns.push(c.build_dsobject());
      }
    });
    return dso;
  };

  Table.init_from_dsobject = function(dso) {
    tablename = dso.tablename;
    table_alias = dso.table_alias;
    if (dso.parent_tablename) {
      parent_tablename = dso.parent_tablename;
      join_pairs = dso.join_pairs || [];
      join_exprs = dso.join_exprs || [];
      join_type = dso.join_type || "left outer join";
    }
    if (dso.select) {
      table_select = dso.select;
    }
    if (dso.columns) {
      dso.columns.forEach(function(dsocol) {
        var col = arc.datasets.Column()
                    .init_from_dsobject(dsocol)
                    .table(Table);
        columns.push(col);
      });
    } else {

      // Old style dataset column load
      var oldway = function(dsocol, isdim) {
        var col = arc.datasets.Column()
                    .init_from_old_dsobject(dsocol)
                    .table(Table)
                    .attr('isdim', isdim);
        columns.push(col);
      };

      dso.dimensions.forEach(function(c) {oldway(c, true);});
      dso.aggregates.forEach(function(c) {oldway(c, false);});
      dso.partitions.forEach(function(p) {
        var c = Table.get_column(p);
        if (c) {
          c.attr('ispart', true);
        }
      });
    }

    return Table;
  };

  // Make copy of table. Not going so far as to also clone associated dataset
  // and columns, so if you update one of those, it will affect original as well
  Table.clone = function() {
    const newTable = arc.datasets.Table();
    newTable.tablename(tablename)
            .dataset(dataset)
            .sample_data(arc.misc.OBJ.sliceRecursive(sample_data))
            .table_alias(table_alias)
            .table_select(table_select)
            .parent_tablename(parent_tablename)
            .join_pairs(arc.misc.OBJ.sliceRecursive(join_pairs))
            .join_exprs(arc.misc.OBJ.sliceRecursive(join_exprs))
            .join_type(join_type)
            .columns(columns)
            .foreign_keys(arc.misc.OBJ.sliceRecursive(foreign_keys));

    return newTable;
  };

  return Table;
};

// Given a data connection id and a table name, return the GET URL that
// will fetch information about the columns in this table
arc.datasets.dataset_columns_url = function(dc_id, tablename, no_cache) {
  return `${arc.frontend.globals.urls.datasetColumnsUrl}/${no_cache ? '1' : '0'}/${tablename}?dataconnection_id=${dc_id}`;
};

arc.datasets.datasetObjectToTableNames = function(d) {
  if (d.type === 'singletable') {
    return [d.detail_display];
  }

  const { tablenames } = d;
  try {
    if (tablenames) {
      return JSON.parse(tablenames);
    }
  } catch (err) {
    // Could not parse table names
    console.error(err);
  }

  return [];
};

arc.datasets.Dataset = function() {
  var id,
      dataconnection_id,
      dataconnection_type,
      dataconnection_type_display,
      dataconnection_name,
      dsname,
      dstype,
      info,
      segments,
      settings = {},
      detail,
      detail_display,
      description = "",
      lvname = "",
      lvcolumns,
      perms = [],
      tables = [],
      lvtable, // this is a Table object with the associated LV (if any)
      needs_column_info = false,
      enable_join_elim,
      enable_incremental_results,
      ksql_stream_position,
      fullTextScore,
      recordCount,
      extractSource,
      extractTarget,
      hasTitleCasedNames = false,
      version_name,
      version_id;

  var colhash = {},    // Column name -> list of tables with this column
      dsuhash = {},    // Display name -> Column table
      tabhash = {};    // Table name (no dbname) -> list of tables matching this name

  var Dataset = function() { return "arc.datasets.Dataset"; };

  Dataset.getRecordCount = function() {
    if (recordCount) {
      return recordCount;
    }

    let terms;
    if (settings.SEARCH && settings.SEARCH.recordCount) {
      terms = settings.SEARCH.recordCount.terms;
    } else {
      terms = 'count rows';
    }

    recordCount = arc.datasets.Column();
    recordCount.attr('name', 'Record Count')
               .attr('expr', 'sum(1)')
               .attr('isdim', false)
               .attr('type', 'BIGINT')
               .attr('recordcount', true)
               .attr('terms', terms)
               .attr('expr', 'SUM(1)');
    return recordCount;
  };

  Dataset.getFullTextScore = function() {
    if (fullTextScore) {
      return fullTextScore;
    }

    let terms = 'search score';

    fullTextScore = arc.datasets.Column();
    fullTextScore.attr('name', 'Search Score')
                 .attr('expr', 'score()')
                 .attr('isdim', true)
                 .attr('type', 'DOUBLE')
                 .attr('terms', terms)
                 .attr('expr', 'SCORE()');
    return fullTextScore;
  };

  Dataset.findColumn = function(colname) {
    let c;

    let found = Dataset.columnIterator(function(column, table) {
      if (column.display_name() === colname) {
        c = column;
        return true;
      }

      return false;
    });

    return c;
  };

  Dataset.defaultAgg = function() {
    return Dataset.searchSettings().defaultAgg;
  };

  // Make the 'cb' call with each column in the dataset
  Dataset.columnIterator = function(cb) {
    let done = false;

    done = cb(Dataset.getRecordCount(), undefined);

    if (done) {
      return done;
    }

    if (Dataset.dataconnection_type() === 'elasticsearch') {
      done = cb(Dataset.getFullTextScore(), undefined);
      if (done) {
        return done;
      }
    }

    Dataset.tables().some(function(table) {
      return table.columns().some(function(column) {
        if (column.attr('hidden')) {
          return;
        }

        return cb(column, table);
      });
    });
  };

  Dataset.enable_join_elim = function(_) {
    if (!info) {
      if (!arguments.length) {
        return undefined;
      }

      return Dataset;
    }

    var tmp_info;
    try {
      tmp_info = JSON.parse(info);
    } catch (err) {
      console.log(err);
    }

    if (!arguments.length) {
      // getter may need to read from info (when enable_join_elim not set)
      if (!enable_join_elim && tmp_info) {
        enable_join_elim = tmp_info[0].enable_join_elim;
      }

      return enable_join_elim; // may be undefined
    } else { // setter updates info with new value
      enable_join_elim = _;
      tmp_info[0].enable_join_elim = _;
      info = JSON.stringify(tmp_info);
    }

    return Dataset;
  };

  /*
   * For datasets created from SQL, store away the SQL in the first table
   * object. This keeps internal state updated so when we call Dataset.save(),
   * the information gets stored in the dataset object.
   */
  Dataset.table_select = function(_) {
    if (!info) {
      console.log('Error: info not set. Cannot save table select');
      return Dataset;
    }

    var tmp_info;
    try {
      tmp_info = JSON.parse(info);
    } catch (err) {
      console.log(err);
    }

    tmp_info[0].select = _;
    info = JSON.stringify(tmp_info);

    return Dataset;
  };

  Dataset.ksql_stream_position = function(_) {
    if (!info) {
      if (!arguments.length) {
        return undefined;
      }

      return Dataset;
    }

    var tmp_info;
    try {
      tmp_info = JSON.parse(info);
    } catch (err) {
      console.log(err);
    }

    if (!arguments.length) {
      // getter may need to read from info (when ksql_stream_position not set)
      if (!ksql_stream_position && tmp_info) {
        ksql_stream_position = tmp_info[0].ksql_stream_position;
      }

      return ksql_stream_position; // may be undefined
    } else { // setter updates info with new value
      ksql_stream_position = _;
      tmp_info[0].ksql_stream_position = _;
      info = JSON.stringify(tmp_info);
    }

    return Dataset;
  };

  Dataset.enable_incremental_results = function(_) {
    if (!info) {
      if (!arguments.length) {
        return undefined;
      }

      return Dataset;
    }

    var tmp_info;
    try {
      tmp_info = JSON.parse(info);
    } catch (err) {
      console.log(err);
    }

    if (!arguments.length) {
      // getter may need to read from info (when enable_incremental_results not set)
      if (!enable_incremental_results && tmp_info) {
        enable_incremental_results = tmp_info[0].enable_incremental_results;
      }

      return enable_incremental_results; // may be undefined
    } else { // setter updates info with new value
      enable_incremental_results = _;
      tmp_info[0].enable_incremental_results = _;
      info = JSON.stringify(tmp_info);
    }

    return Dataset;
  };

  Dataset.needs_column_info = function(_) {
    if (!arguments.length) return needs_column_info;
    needs_column_info = _;
    return Dataset;
  };

  Dataset.dataconnection_id = function(_) {
    if (!arguments.length) return dataconnection_id;
    dataconnection_id = _;
    return Dataset;
  };

  Dataset.dataconnection_type = function(_) {
    if (!arguments.length) return dataconnection_type;
    dataconnection_type = _;
    return Dataset;
  };

  Dataset.dataconnection_type_display = function(_) {
    if (!arguments.length) return dataconnection_type_display;
    dataconnection_type_display = _;
    return Dataset;
  };

  Dataset.dataconnection_name = function(_) {
    if (!arguments.length) return dataconnection_name;
    dataconnection_name = _;
    return Dataset;
  };

  Dataset.dataset_id = function(_) {
    if (!arguments.length) return id;
    id = _;
    return Dataset;
  };

  Dataset.dataset_name = function(_) {
    if (!arguments.length) return dsname || ("Dataset " + id);
    dsname = _;
    return Dataset;
  };

  Dataset.dataset_type = function(_) {
    if (!arguments.length) return dstype;
    dstype = _;
    return Dataset;
  };

  Dataset.dataset_perms = function(_) {
    if (!arguments.length) return perms;
    perms = _;
    return Dataset;
  };

  Dataset.dataset_info = function(_) {
    if (!arguments.length) return info;
    info = _;
    return Dataset;
  };

  Dataset.description = function(_) {
    if (!arguments.length) return description;
    description = _;
    return Dataset;
  };

  Dataset.lvname = function(_) {
    if (!arguments.length) return lvname;
    lvname = _;
    return Dataset;
  };

  Dataset.dataset_lvcolumns = function(_) {
    if (!arguments.length) return lvcolumns;
    lvcolumns = _;
    return Dataset;
  };

  Dataset.dataset_lvtable = function(_) {
    if (!arguments.length) return lvtable;
    lvtable = _;
    return Dataset;
  };

  Dataset.dataset_detail = function(_) {
    if (!arguments.length) return detail;
    detail = _;
    return Dataset;
  };

  Dataset.version_name = function(_) {
    if (!arguments.length) return version_name;
    version_name = _;
    return Dataset;
  };

  Dataset.version_id = function(_) {
    if (!arguments.length) return version_id;
    version_id = _;
    return Dataset;
  };

  Dataset.dataset_detail_display = function(_) {
    if (!arguments.length) return detail_display;
    detail_display = _;
    return Dataset;
  };

  Dataset.tables = function(_) {
    if (!arguments.length) return tables;
    tables = _;
    return Dataset;
  };

  Dataset.segments = function(_) {
    // Segment is currently empty if created for search visual
    if (!arguments.length) return segments || [];
    segments = _;
    return Dataset;
  };

  Dataset.searchSettings = function(_) {
    if (!arguments.length) {
      return settings.SEARCH || {};
    }

    settings.SEARCH = {
      enabled: _.enabled ? true : false,
      group: _.group,
      expansions: _.expansions,
      suggestions: _.suggestions,
      defaultAgg: _.defaultAgg,
      recordCount: {
        terms: Dataset.getRecordCount().attr('terms'),
      },
    };

    return Dataset;
  };

  Dataset.tags = function() {
    const datasetTags = [];

    if (extractsEnabled()) {
      if (extractSource) {
        datasetTags.push(arc.extracts.extractSourceBadge);
      }

      if (extractTarget) {
        datasetTags.push(arc.extracts.extractTargetBadge);
      }
    }

    return datasetTags;
  };

  Dataset.isExtractSource = function() {
    return extractsEnabled() && extractSource;
  };

  Dataset.isExtractTarget = function() {
    return extractsEnabled() && extractTarget;
  };

  Dataset.defaultDateTime = function() {
    return Dataset.timeSettings().defaultDateTime;
  };

  Dataset.defaultTimeInterval = function() {
    return Dataset.timeSettings().defaultTimeInterval;
  };

  Dataset.defaultDateTimeFormat = function() {
    return Dataset.timeSettings().defaultDateTimeFormat;
  };

  Dataset.timeSettings = function(_) {
    if (!arguments.length) {
      return settings.TIME || {};
    }

    settings.TIME = {
      defaultDateTime: _.defaultDateTime,
      defaultDateTimeFormat: _.defaultDateTimeFormat,
      defaultTimeInterval: _.defaultTimeInterval,
      defaultAgg: _.defaultAgg,
      frequencies: _.frequencies,
      increments: _.increments,
      selectionPos: _.selectionPos,
      selectionRange: _.selectionRange,
    };
  };

  Dataset.validateTimeControl = function() {
    let timeSettings,
        dim,
        agg,
        dimCol,
        aggCol;

    timeSettings = Dataset.timeSettings();
    dim = timeSettings.defaultDateTime;
    agg = timeSettings.defaultAgg;

    dimCol = Dataset.findColumn(dim);
    aggCol = Dataset.findColumn(agg);

    if (dimCol === undefined) {
      return false;
    }

    if (!dimCol.isDimension()) {
      return false;
    }

    if (aggCol === undefined) {
      return false;
    }

    if (aggCol.isDimension()) {
      return false;
    }

    if (timeSettings.frequencies.length === 0) {
      return false;
    }

    if (timeSettings.increments.length === 0) {
      return false;
    }

    return true;
  };

  Dataset.get_cache_settings = function() {
    return settings.CACHE || {};
  };

  Dataset.put_cache_settings = function(cache) {
    settings.CACHE = {OVERRIDE: cache.OVERRIDE ? 1 : 0,
                      ENABLED: cache.ENABLED ? 1 : 0,
                      RETENTION: parseInt(cache.RETENTION)};
  };

  // Check to see if a table name is unique in the dataset
  Dataset.validate_table_alias = function(table, alias) {
    for (var i = 0; i < tables.length; i++) {
      var t = tables[i];
      if (t !== table && t.tablename_noalias() === alias) {return false;}
      if (t !== table && t.table_alias() === alias) {return false;}
    }
    return true;
  };

  // Set up an alias for this table if needed
  Dataset.set_table_alias = function(table) {
    let tablename = table.tablename_noalias();
    table.table_alias(undefined);

    if (!Dataset.validate_table_alias(table, tablename)) {
        var tabonly = arc.datasets.drop_dbname(tablename);
        var iter = 0;
        var alias;
        do {
          iter++;
          alias = tabonly + "_" + iter;
        } while (!Dataset.validate_table_alias(table, alias));
        table.table_alias(alias);
    }
  };

  Dataset.validate_user_input_table_alias = function(table, alias) {
    alias = alias.trim();
    if (!alias) {
      throw new Error(`Table alias can't be blank`);
    }

    // Check alias is not the same as table tablename it's based on
    const dbName = arc.datasets.just_dbname(table.tablename_noalias());
    const dbPlusNewAlias = `${dbName}.${alias}`;
    if (!Dataset.validate_table_alias(table, dbPlusNewAlias)) {
      throw new Error(`Table name/alias ${alias} in use`);
    }

    // Check there is no alias same as this given alias
    if (!Dataset.validate_table_alias(table, alias)) {
      throw new Error(`Table name/alias ${alias} in use`);
    }
  };

  // Lookup a column in the dataset
  // DJC -- not clear how to handle column names existing in more than one table
  //        if colhash was guarenteed to be up to date and also included the aliases,
  //        then we could just use that, rather than a search.
  Dataset.get_column = function(fullname) {
    if (!fullname) {
      return undefined;
    }

    fullname = fullname.toLowerCase();
    if (dsuhash[fullname]) {
      return dsuhash[fullname];
    }

    let m;
    // The two regexes below match dot separated database, table and column name groups where
    // only difference is for Solr column name can start with not only a letter but
    // underscore as well
    if (dataconnection_type === 'solr') {
      m = fullname.match(
          /([a-zA-Z][a-zA-Z0-9_]*\.)?([a-zA-Z][a-zA-Z0-9_]*\.)?([a-zA-Z_][a-zA-Z0-9_]*)/
        );
    } else {
      m = fullname.match(
          /([a-zA-Z][a-zA-Z0-9_]*\.)?([a-zA-Z][a-zA-Z0-9_]*\.)?([a-zA-Z][a-zA-Z0-9_]*)/
        );
    }

    // Give up if didn't match the entire name
    if (!m || m.length !== 4 || m[0] !== fullname) {
      return undefined;
    }
    var dbname = m[1] ? m[1].slice(0, -1) : undefined;
    var tbname = m[2] ? m[2].slice(0, -1) : undefined;
    var clname = m[3];

    // For column name only, look first at column hash
    if (!tbname) {
      if (colhash[clname]) {
        return colhash[clname][0].get_column(clname);
      }
    }

    // If we have database.table, find the table
    for (var i = 0; i < tables.length; i++) {
      var table = tables[i];
      if (dbname) {
        if (table.tablename_noalias() === dbname + '.' + tbname) {
          return table.get_column(clname);
        }
      } else if (tbname) {
        if ((table.tablename_alias() === tbname) ||
            (arc.datasets.drop_dbname(table.tablename_noalias()) === tbname)) {
          return table.get_column(clname);
        }
      } else {
        var column = table.get_column(clname);
        if (column) {
          return column;
        }
      }
    }
    return undefined;
  };

  // Ensure a given alias is ok, return error text if not
  Dataset.check_alias = function(col, alias) {
    // Currently, just ensure alias not already in use
    alias = alias.toLowerCase();
    if (dsuhash[alias] && (dsuhash[alias] !== col)) {
      return 'Field alias `' + alias + '` in use';
    }
    return '';
  };

  // Pull list of column names from an expression (must be in [])
  Dataset.colnames_from_expr = function(exprtext) {
    let info = { colnames:[], columns:[], sqltext:"" };
    let m = [];
    const regex = /(.*?)\[(.+?)\](.*)/;

    do {
      info.sqltext = exprtext;
      m = exprtext.match(regex);
      if (m && m.length === 4) {
        exprtext = m[1];
        const column = Dataset.get_column(m[2]);
        if (column) {
          info.colnames.push(m[2]);
          info.columns.push(column);
          if (column.attr('name')) {
            exprtext += Dataset.get_nonambiguous_colname(column.table().tablename_noalias(),
                                                         column.attr('name'))
                               .replace('[', '')
                               .replace(']', '');
          } else {
            exprtext += '';
          }
        } else {
          exprtext += m[2];
        }
        exprtext += m[3];
      }
    } while (m && m.length === 4);

    return info;
  };

  // Convert 'my_col' -> 'My Col'
  Dataset.title_casing_names = function(name) {
    if (!name) return name;

    let newString = [];
    let prevIsSpace = false;
    let char;
    for (let i = 0; i < name.length; i++) {
      char = name.charAt(i);

      // For complex types, we strip everything before the '.'
      // TODO -- we should be smarter about recognizing these
      if (char === '.') {
        newString = [];
        prevIsSpace = true;
        continue;
      }

      if (char === '_') {
        newString.push(' ');
        prevIsSpace = true;
      } else {
        if (prevIsSpace || i === 0) {
          newString.push(char.toUpperCase());
        } else {
          newString.push(char);
        }

        prevIsSpace = false;
      }
    }

    return newString.join('');
  };

  Dataset.has_title_cased_names = function() {
    return hasTitleCasedNames;
  };

  Dataset.toggle_title_cased_names = function() {
    hasTitleCasedNames = (!hasTitleCasedNames);
  };

  // Toggle between base names and nicer ones: 'my_col' -> 'My Col'
  Dataset.toggle_colnames = function() {
    const convertColName = (col) => {
      const baseName = col.attr('name');

      if (baseName) {
        const displayedName = col.attr('alias');
        const titleCasedName = Dataset.title_casing_names(baseName);
        if (hasTitleCasedNames) {
          if (!displayedName || displayedName === baseName) {
            if (titleCasedName !== displayedName) {
              col.attr().alias = titleCasedName;
            }
          }
        } else if (titleCasedName === displayedName) {
          col.attr().alias = baseName;
        }
      }
    };

    tables.forEach((table) => {
      table.base_columns().forEach((c) => {
        convertColName(c);
      });
    });

    Dataset.rebuild_hashes();
  };

  // Add one column into the global dataset structures
  Dataset.add_column = function(column) {
    var table = column.table();
    var alias = column.attr('alias');
    var colname = column.attr('name');

    // Ensure we have a unique alias
    alias = alias || colname || '';

    // Prefix with table name if computed alias not unique in dataset
    if (dsuhash[alias.toLowerCase()]) {
      var tn = arc.datasets.drop_dbname(table.tablename());
      let c = ' ';
      if (dataconnection_type === 'bigquery') {
        c = '_';
      }

      if (dsuhash[(tn + c + alias).toLowerCase()]) {
        tn = table.tablename_noalias().replace('.', ' ');
      }

      alias = tn + c + alias;
    }

    // DJC -- Should clean this up by removing the column.attr() call into
    //        rebuild_hashes().
    // We don't want to use the standard column.attr('alias', alias) as
    // that routine calling back into rebuild_hashes.
    column.attr().alias = alias;

    dsuhash[alias.toLowerCase()] = column;

    if (colname) {
      colname = colname.toLowerCase();
      if (colhash[colname] === undefined) {
        colhash[colname] = [];
      }
      colhash[colname].push(table);
    }

    return Dataset;
  };

  // Local function to rebuild the lookup hashes
  Dataset.rebuild_hashes = function() {

    colhash = {};
    tabhash = {};
    dsuhash = {};

    var onetab = function(t) {
      // Add all column names to the column hash
      t.columns().forEach(function(c) {
        Dataset.add_column(c);
      });

      // And the tablename (without dbname) to the table hash
      var tabonly = arc.datasets.drop_dbname(t.tablename());
      if (tabhash[tabonly] === undefined) {
        tabhash[tabonly] = [];
      }
      tabhash[tabonly].push(t);
    };

    tables.forEach(function(t) {
      onetab(t);
    });
  };

  // Get the shortest version of column which is non-ambiguous
  // DJC -- might want asserts if input is not valid
  Dataset.get_nonambiguous_colname = function(tablename, colname) {

    var coltick = arc.dataconnection._enquote_tablename(colname, dataconnection_type);
    colname = colname.toLowerCase();
    var tabonly = arc.datasets.drop_dbname(tablename);
    if ((colhash[colname] === undefined) || (tabhash[tabonly] === undefined)) {
      return coltick;
    }

    if (colhash[colname].length === 1) {   // column name is unique
      return coltick;
    }

    if (tabhash[tabonly].length === 1) {   // tablename is unique
      return tabonly + '.' + coltick;
    }

    return tablename + '.' + coltick;     // need full <dbname>.<tabname>.<colname>
  };

  Dataset.get_table = function(tablename) {
    var retval;
    tables.forEach(function(t) {
      if (t.tablename() === tablename) retval = t;
    });
    return retval;
  };

  // Build the select list elements for all columns as seen by app-builder
  Dataset.all_columns_select_text = function() {
    var sellist = [];
    var fromlist = [];
    var tabinfo = [];
    var cnt;

    var onecol = function(tab, col) {
      var colsql = [];
      if (!col.attr('hidden') && !col.attr('exprhasagg')) {
        cnt++;
        var alias = col.attr('alias');
        var cast = col.attr('cast');
        var name = col.attr('name');
        var expr = col.attr('expr');
        if (expr) {
          name = expr;
        } else {
          name = tab.tablename() + '.`' + name + '`';
        }
        if (cast) {
          colsql.push('cast(' + name + ' as ' + cast + ')');
        } else {
          colsql.push(name);
        }
        if (alias) {
          colsql.push(' as `' + alias + '`');
        }
        var text = colsql.join(' ');
        var info = Dataset.colnames_from_expr(text);
        text = info.sqltext;

        sellist.push(text);
      }
    };

    tables.forEach(function(table, idx) {
      cnt = 0;
      table.dimensionsAndHierarchies().forEach(function(col) {onecol(table, col);});
      table.measuresOnly().forEach(function(col) {onecol(table, col);});

      // Track tablename and number of items from this table
      tabinfo.push([arc.datasets.drop_dbname(table.tablename()), cnt]);

      // Now build the join portion
      if (idx === 0) {
        fromlist.push(table.tablename_withalias());
      } else {
        var joinpreds = [];
        fromlist.push(table.join_type());
        fromlist.push(table.tablename_withalias());
        fromlist.push("on");
        table.join_pairs().forEach(function(p) {
          joinpreds.push("(" + table.parent_tablename() + '.`' + p[0] +
                         "`=" + table.tablename() + '.`' + p[1] + "`)");
        });
        table.join_exprs().forEach(function(e) {
          joinpreds.push("(" + e + ")");
        });
        fromlist.push(joinpreds.join(' and '));
      }
    });

    return [sellist.join(', '), fromlist.join(' and '), tabinfo];
  };

  Dataset.update_lv = function() {
    if (lvname && lvcolumns) {
      var dsoobject = {
        'tablename': lvname,
        'columns': lvcolumns
      };
      lvtable = arc.datasets.Table()
        .dataset(Dataset)
        .tablename(lvname)
        .init_from_dsobject(dsoobject);
    }
    return Dataset;
  };

  Dataset.init_from_fetch = function(response, columninfo) {
    function populate_tables(response, columninfo) {
      Dataset.update_lv();
      tables = [];
      var table, tablename;
      if (dstype === "singletable" && info.length === 0) {
        tablename = response.detail;
        table = arc.datasets.Table()
          .dataset(Dataset)
          .tablename(tablename);
        if (columninfo) {
          table.init_from_url_response(columninfo);
          needs_column_info = false;
        } else {
          needs_column_info = true;
        }
        tables.push(table);
      } else {
        JSON.parse(info).forEach(function(i) {
          table = arc.datasets.Table()
            .dataset(Dataset)
            .init_from_dsobject(i);
          tables.push(table);
        });
      }
    }

    dataconnection_id = response.dc_id;
    dataconnection_type = response.dc_type;
    dataconnection_type_display = response.dc_type_display;
    dataconnection_name = response.dc_name;
    id = response.id;
    dsname = response.name;
    dstype = response.type;
    perms = response.perms;
    info = response.info;
    settings = JSON.parse(response.settings);
    detail = response.detail;
    detail_display = response.detail_display;
    description = response.description;
    lvname = response.lvname;
    lvcolumns = response.lvcolumns;
    extractSource = response.extractSource;
    extractTarget = response.extractTarget;
    hasTitleCasedNames = settings.HAS_TITLE_CASED_NAMES;
    version_name = response.version_name;
    version_id = response.version_id;

    // clean up settings.SEARCH.suggestions and settings.SEARCH.expansions
    const searchSettings = Dataset.searchSettings();
    const fixedSuggestions = searchSettings.suggestions || [];
    const fixedExpansions = searchSettings.expansions || [];
    searchSettings.suggestions = fixedSuggestions.filter(s => s.s);

    // Empty values for expansions are permitted.
    // This is a way to suck up contextual words for better matching
    //    without actually substituting anything
    // For example: in multiple population datasets, you could expand US in one
    //    and Dutch in the other, for 'US population' and 'Dutch population' to
    //    score better in the appropriate dataset
    searchSettings.expansions = fixedExpansions.filter((s) => s.key);
    Dataset.searchSettings(searchSettings);
    Dataset.segments(response.segments);

    populate_tables(response, columninfo);

    Dataset.rebuild_hashes();
    return Dataset;
  };

  // Drop a table from the dataset -- also need drop any descendent tables
  Dataset.drop_table = function(tablename) {

    var droplist = [tablename];

    // Iterate until we don't drop any more tables
    var onepass = function() {
      var startlen = tables.length;
      var newtables = [];
      tables.forEach(function(t) {

        // If parent has been dropped, drop this one as well
        if (droplist.indexOf(t.parent_tablename()) !== -1) {
          droplist.push(t.tablename());
        }

        // Otherwise if not in drop list, we can keep this one
        else if (droplist.indexOf(t.tablename()) === -1) {
          newtables.push(t);
        }
      });
      tables = newtables;

      // we are not dropping foreign/primary keys of deleted table because of its
      //global effect. If we decide to do so, here is the place to build the
      // foreign keys to be removed
      return (tables.length !== startlen);
    };

    while (onepass());
    Dataset.rebuild_hashes();
    return Dataset;
  };

  Dataset.add_table = function(table) {
    tables.push(table);
    Dataset.rebuild_hashes();
    return Dataset;
  };

  // Update an existing dataset table to point to a different database table
  Dataset.update_tablename = function(table, newname) {

    let oldname = table.tablename();
    let parents = tables.filter(t => t.parent_tablename() === oldname);
    table.tablename(newname);
    Dataset.set_table_alias(table);
    let finalname = table.tablename();
    parents.forEach(t => t.parent_tablename(finalname));

    // The legacy dataset_detail item always matches the tablename for the
    // first table in the dataset
    if (table === tables[0]) {
      Dataset.dataset_detail(newname);
    }
  };

  Dataset.update_tablealias = function(table, alias) {
    const oldTbAlias = table.table_alias();
    const children = tables.filter((t) => t.parent_tablename() === oldTbAlias);
    table.table_alias(alias);
    const newTbAlias = table.table_alias();
    children.forEach((t) => t.parent_tablename(newTbAlias));

    // replace old table alias with new in column name aliases
    table.columns().forEach((col) => {
      col.attr().alias = col.attr().alias.replace(oldTbAlias, newTbAlias);
    });

    Dataset.rebuild_hashes();
  };

  Dataset.fetch_and_add_table = function(tablename, cb, error_cb) {

    let url = arc.datasets.dataset_columns_url(dataconnection_id, tablename, true);

    arc.staticfetcher.forget([{asset:url}]);
    arc.staticfetcher.fetch([{
        asset: url,
        postprocess: d3.json,
      }], function(error, columninfo) {
        if (error) {
          console.log(error);
          arc.report.alertMessage("Failed to load dataset table column detail",
              error.response ? error.response : error);
          if (error_cb) {
            error_cb(error);
          }
          return;
        }

        // Build and add new table structure
        var table = arc.datasets.Table()
                       .dataset(Dataset)
                       .tablename(tablename)
                       .init_from_url_response(columninfo);

        // Set up an alias for this table if needed
        Dataset.set_table_alias(table);

        if (arc.globals.site_enable_column_rename) {
          table.columns().forEach(function (c) {
            c.attr().alias = Dataset.title_casing_names(c.attr('name'));
          });

          Dataset.toggle_title_cased_names();
        }

        Dataset.add_table(table);
        cb(table);
      });
      return Dataset;
    };

  Dataset.update_tables = function(cb) {

    var idx = 0;
    var responses = {};

    var nextcall = function(table, tablename) {
      if (table.table_select()) {
        arc.datasets.createFromSql()
           .dataconnection_id(dataconnection_id)
           .dataconnection_type(dataconnection_type)
           .sql(table.table_select())
           .errorCb(function(errorMsg) {
             cb(false, errorMsg);
           })
           .fetchColumns(function(r) {
             nextresp(undefined, r);
           });

        return;
      }

      d4.queue()
        .defer(d3.json,
               arc.datasets.dataset_columns_url(dataconnection_id, tablename, true))
        .await(nextresp);
    };

    var nextresp = function(error, colinfo) {

      // Error terminates process
      if (error) {
        console.log(error);
        if (cb) {
          cb(false, "Failed to load column detail for " + tables[idx].tablename());
        }
        return;
      }

      // Save this response and issue next if needed
      responses[tables[idx].tablename()] = colinfo;
      idx++;
      if (idx < tables.length) {
        nextcall(tables[idx], tables[idx].tablename_noalias());
        return;
      }

      // Here we are all done -- update table info and do callback
      var changed = false;
      tables.forEach(function(t) {
        var r = responses[t.tablename()];
        if (r) {
          changed |= t.update_from_url_response(r);
        }
      });
      if (changed) {
        Dataset.rebuild_hashes();
      }
      if (cb) {
        cb(changed, null);
      }
    };

    // Start off the table info fetching process
    nextcall(tables[0], tables[0].tablename());
  };

  Dataset.add_sq_table = function(sql, parent, alias, jp, cnames, ctypes) {
    // Build and add new table structure
    var table = arc.datasets.Table()
                   .tablename('(' + sql + ')')
                   .dataset(Dataset)
                   .table_alias(alias)
                   .parent_tablename(parent)
                   .join_pairs(jp)
                   .init_from_add_sq(cnames, ctypes);

    console.log("Added table: ", table.tablename());
    Dataset.add_table(table);
    return Dataset;
  };

  Dataset.buildFK = function() {
    var tabdata = [];
    tables.forEach(function(t) {
      tabdata.push([t.tablename(), t.foreign_keys() || {}]);
    });
    return JSON.stringify(tabdata);
  };

  Dataset.buildInfoObj = function(skipcol) {
    var tabdata = [];
    tables.forEach(function(t) {
      tabdata.push(t.build_dsobject(skipcol));
    });

    if (enable_incremental_results) {
      tabdata[0].enable_incremental_results = enable_incremental_results;
    }

    if (ksql_stream_position) {
      tabdata[0].ksql_stream_position = ksql_stream_position;
    }

    if (enable_join_elim) {
      tabdata[0].enable_join_elim = enable_join_elim;
    }

    return tabdata;
  };

  Dataset.buildInfo = function(skipcol) {
    return JSON.stringify(Dataset.buildInfoObj(skipcol));
  };

  Dataset.cloneDatasetOnServer = () => {
    if (!id) {
      throw new Error('Cannot clone unsaved dataset');
    }

    const data = {
      dataset_id: id,
    };

    return arc.frontend.vizHttp.post({
        url: arc.frontend.globals.urls.datasetUrl,
        data,
        headers: {
          'X-CSRFTOKEN': arc.globals.csrf_token,
        },
        isOldEndpoint: true,
      })
      .then((result) => result.data[0].id)
      .catch((error) => {
        throw new Error(`Failed to clone dataset: ${
          error.responseText ?
          error.responseText :
          error
        }`);
      });
  };

  Dataset.save = function(done_cb, errorDisplayCb) {
    if (!done_cb) {
      done_cb = function() { };
    }

    if (!errorDisplayCb) {
      errorDisplayCb = (error_msg) => arc.report.alertMessage(
        `Failed to ${id ? 'update' : 'create'} dataset: `, error_msg
        );
    }

    settings.HAS_TITLE_CASED_NAMES = hasTitleCasedNames;

    const post_data = {
      csrfmiddlewaretoken: arc.globals.csrf_token,
      dataconnection_id,
      dataset_name: dsname.trim(),
      dataset_type: dstype.trim(),
      dataset_info: info.trim(),
      dataset_settings: JSON.stringify(settings),
      dataset_lvname: lvname.trim(),
      dataset_detail: detail.trim(),
      description: description.trim(),
      version_name: version_name?.trim(),
      version_id,
      foreign_keys: {}
    };

    // Switch to multipletables type and flag to build info of table data items
    let postinfo = false;
    if (tables.length > 1) {
      dstype = 'multipletables';
      post_data.dataset_type = dstype;
      postinfo = true;
    } else if (tables.length === 1) {
      dstype = 'singletable';
      post_data.dataset_type = dstype;
      postinfo = true;
      /* -- Always save dataset info
      // For single table, switch to post info if there are any custom columns
      var allcols = tables[0].columns();
      for (var i = 0; i < allcols.length; i++) {
        if (allcols[i].has_custom_attribute()) {
          postinfo = true;
          break;
        }
      }
      */
    }

    // Build dataset info if needed
    if (postinfo) {
      post_data.dataset_info = Dataset.buildInfo();
      post_data.foreign_keys = Dataset.buildFK();
    } else {
      post_data.dataset_info = "";
    }

    if (id) {
      post_data.dataset_id = id;
    }

    let url = arc.frontend.globals.urls.datasetUrl;
    if (id) {
      url += "/" + id;
      arc.staticfetcher.forget([
        {
          asset: url,
        },
        {
          asset: `${url}?detail=true`
        },
      ]);
    }

    // save event
    var error_msg;
    $.ajax({
      url: url,
      data: post_data,

      // 'PUT' updates a event, 'POST' creates a new one
      type: id ? 'PUT' : 'POST',
      beforeSend: function(xhr) {
        // Because of a jquery bug when handling PUT requests,
        // we have to explicitly set X-CSRFToken here
        xhr.setRequestHeader("X-CSRFToken", arc.globals.csrf_token);
      }
    })
    .done(function(resp) {
      if (!id) {
        id = resp[0].id;
      }
      if (lvname.trim() && lvname.trim() !== 'DELETE') {
        if (resp[0].updatedLV) {
          arc.report.alertMessageSmall('Created/updated associated logical view "' +
              lvname.trim() + '"', "alert-success");
        }
      } else if (lvname.trim() === 'DELETE') {
        // clear LV name
        lvname = '';
        lvcolumns = '';
        lvtable = undefined;
        arc.report.alertMessageSmall("Deleted associated logical view.", "alert-success");
      }
    })
    .fail((xhr) => {
      // If there is a problem with the network(readyState==0)
      // there will be no responseText and displaying the XHR as object not meaningful
      const networkOrUnknownError = (xhr.readyState === 0 ? 'Network error' : xhr);
      error_msg = xhr.responseText ?  xhr.responseText : networkOrUnknownError;
      errorDisplayCb(error_msg);
    })
    .always(function() {
      done_cb(error_msg, Dataset);
    });

    return Dataset;
  };

  Dataset.delete = function(delete_table, cb) {
    let url = arc.frontend.globals.urls.datasetUrl;
    url += "/" + id;

    // save new segment
    $.ajax({
      url: url + '?' + $.param({"delete_table": delete_table ? "true" : "false"}),
      type: 'DELETE',
      beforeSend: function(xhr) {
        // Because of a jquery bug when handling DELETE requests,
        // we have to explicitly set X-CSRFToken here
        xhr.setRequestHeader("X-CSRFToken", arc.globals.csrf_token);
      }
    })
    .fail(function(req) {
      arc.report.alertMessage("Failed to delete dataset",
          req.responseText ? req.responseText : req);
    })
    .always(function() {
      cb();
    });
    return Dataset;
  };

  Dataset.reset_cache = function(cb) {
    let url = arc.frontend.globals.urls.datasetUrl;
    url += "/cache_reset/" + id;

    $.ajax({
      url: url,
      type: 'POST',
      beforeSend: function(xhr) {
        // Because of a jquery bug when handling DELETE requests,
        // we have to explicitly set X-CSRFToken here
        xhr.setRequestHeader("X-CSRFToken", arc.globals.csrf_token);
      }
    })
    .done(function() {
      if (cb) {
        cb();
      }
    })
    .fail(function(rsp) {
      if (cb) {
        cb(rsp.responseText || rsp);
      }
    });
    return Dataset;
  };

  Dataset.set_version = function(cb) {
    let url = arc.frontend.globals.urls.datasetUrl;
    url += `/version_control/${id}`;

    $.ajax({
      url,
      type: 'POST',
      beforeSend(xhr) {
        xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
      }
    })
    .done(() => {
      if (cb) {
        cb();
      }
    })
    .fail((rsp) => {
      if (cb) {
        cb(rsp.responseText || rsp);
      }
    });
    return Dataset;
  };

  return Dataset;
};

arc.datasets.createFromTable = (
  dataconnectionId,
  dataconnectionType,
  datasetName,
  databaseName,
  tableName,
  successCb,
  errorCb,
  ) => {
    const datasetToSave = arc.datasets.Dataset()
          .dataconnection_id(dataconnectionId)
          .dataconnection_type(dataconnectionType)
          .dataset_name(datasetName)
          .dataset_type('singletable')
          .dataset_info('')
          .dataset_detail(`${databaseName}.${tableName}`);

    datasetToSave.fetch_and_add_table(datasetToSave.dataset_detail(),
          () => datasetToSave.save(successCb),
          errorCb);
};

function clearColumnErrors() {
  // Remove any existing alert text
  const $editColumnModal = $('#edit-column-modal');
  $editColumnModal.find('.alert-container .alert-error-messages').empty();
  $editColumnModal.find('.alert-container .alert-number-errors').text(0);
  $editColumnModal.find('.column-edit-form .form-text.with-errors').empty();
  $editColumnModal.find('.column-edit-form .form-group').removeClass('has-error');
}

arc.datasets.populateEditColumnModal = function(col, table_editor, displayOptions) {
  clearColumnErrors();
  const $editColumnModal = $("#edit-column-modal");
  $editColumnModal.find('.alert').hide();
  $editColumnModal.find('a[href=\\#basic]').tab('show');
  $editColumnModal.modal('show');
  const name = col.attr('name') || col.attr('basecol');
  const display = col.attr('alias') || '';
  const comment = col.attr('comment') || col.attr('hivecomment') || '';
  const dflagg = col.attr('dflagg') || arc.misc.defaultAgg(col.attr('type'));
  const expr = col.attr('expr') || '';
  const isDim = col.attr('isdim') ? 'dimension' : 'measure';
  const ishidden = col.attr('hidden');
  const geoType = col.attr('geotype') || 'none';
  const dataset = col.table().dataset();
  const ispartition = col.attr('ispartition') || false;
  const isSensitive = col.attr('isSensitive') || false;
  const displayFormat = col.attr('displayformat');
  const exprhasagg = col.attr('exprhasagg');

  displayOptions.populateValues(displayFormat, 'fromDatasets');

  $('#override-validation').prop('checked', true);
  $editColumnModal.find('[arcid=validate-spinner]').hide();

  $('#expr-has-agg').prop('checked', exprhasagg === true);

  let dctype = col.table().dataset().dataconnection_type();

  let isDetailData = col.attr('detaildata');
  if (isDetailData === undefined) {
    // Default for cases where this is not setup
    isDetailData = true;
  }

  // Include "aggregate" for cassandra
  if (dctype === 'cassandra') {
    arc.datasets.dflagg_table.push(['Aggregate', 'agg']);
  }

  // Fill in the template with the available agg functions
  const $aggList = $("[arcid=dsattr-default-agg]").empty();
  const $aggAdditions = arc.datasets.dflagg_table.map(d => `<option>${d[0]}</option>`);
  $aggList.append($aggAdditions);

  const $geotype = $("[arcid=dsattr-geotype]").empty();
  const $geoAdditions = arc.geo.geoTypes.map(
    t => `<option value="${t}">${arc.geo.geoTypeToDisplayName(t)}</option>`
  );
  $geotype.append($geoAdditions);

  if (table_editor) {
    table_editor.destroy();
    table_editor = null;
  }

  $editColumnModal.find("span[arcid=base-column]").text(name);
  $editColumnModal.find("input[arcid=column-display]").val(display);
  $editColumnModal.find("textarea[arcid=column-comment]").val(comment);
  $editColumnModal.find("[arcid=column-expression]").text(expr);
  $editColumnModal.find("[arcid=column-visibility]").prop("checked", !ishidden);
  $editColumnModal.find('[arcid=column-sensitivity]').prop('checked', isSensitive);
  $editColumnModal.find("[arcid=column-detaildata]").prop("checked", isDetailData);
  $editColumnModal.find('#ecmodal-user-overrode').prop('checked', col.attr('useroverrodeisdim'));

  if (col.attr('ispart')) {
    $editColumnModal.find('span[arcid=base-column]')
      .append($('#dsattr-partition-key-template').html());
  }

  $editColumnModal.find('[data-toggle="tooltip"]').tooltip();

  const $partitionColumn = $editColumnModal.find("[arcid=column-partition]");
  if (!arc.dataconnection.isArcEngine(dctype)) {
    $partitionColumn.closest('.form-check').hide();
  } else {
    if (col.attr('ispart')) {
      $partitionColumn.closest('.form-check').hide();
    } else {
      $partitionColumn.closest('.form-check').show();
      $partitionColumn.prop('checked', ispartition);
    }
  }

  $editColumnModal.find("[name=radio-category]").prop("checked", function() {
    return $(this).val() === isDim;
  });

  const aggval = arc.datasets.findListEntry(arc.datasets.dflagg_table, dflagg, 1, 0);
  $editColumnModal.find("[arcid=dsattr-default-agg]")
                  .find(`:contains(${aggval})`)
                  .prop("selected", "selected");

  $editColumnModal.find(`[arcid=dsattr-geotype] option[value="${geoType}"]`)
                  .prop("selected", "selected");

  // make field edit modal width expand/contract depending
  // on which tab is selected
  $editColumnModal.find('a[data-toggle="tab"]')
    .off('shown.bs.tab').on('shown.bs.tab', function updateModalSize(e) {
      const $this = $(this);
      const $dialog = $this.closest('.modal-dialog');
      if ($this.attr('href') === '#expression') {
        $dialog.addClass('modal-lg');
      } else {
        $dialog.removeClass('modal-lg');
      }
  });

  $editColumnModal.find('#ecmodal-dim-radio, #ecmodal-agg-radio')
    .off('click.toggleDimVMeasure').on('click.toggleDimVMeasure', function setUserOverrode() {
      $editColumnModal.find('#ecmodal-user-overrode').prop('checked', true);
  });

  // If column name is empty, it's a cloned column and can be deleted
  // Also, set up expression editor id if this is a cloned column
  // else delete autocomplete on/off button
  const $autocompleteButton = $editColumnModal.find(".on-off-autocomplete");
  const $editorContainer = $editColumnModal.find("#expression-editor-container");
  const $editor = $editColumnModal.find(".column-expression-edit");
  if (!col.isBaseCol()) {
    $editColumnModal.find("[arcid=column-edit-remove]").show();
    $editColumnModal.find("[arcid=edit-expression-section]").show();
    $editColumnModal.find("#expression-tab").show();

    $editorContainer.empty();
    const id = `expression-editor-${name}`;
    $editor.attr("id", id);
    $autocompleteButton.show();
    const expression = $editor.text().trim();
    arc.globals.is_edit = true;
    table_editor = arc.misc.expressionEditor(dataset, table_editor, id, $editColumnModal,
      {includeCalculated: true});
    table_editor.value(expression);
  } else {
    $editColumnModal.find("[arcid=column-edit-remove]").hide();
    $editColumnModal.find("[arcid=edit-expression-section]").hide();
    $editColumnModal.find("#expression-tab").hide();
    $autocompleteButton.hide();
    table_editor = null;
  }
  return table_editor;
};

// For the above modal, validate the column changes
arc.datasets.validateColumnEditChanges = function(col, table_editor, displayOptions, mustValidate,
    success_cb) {

  const dctype = col.table().dataset().dataconnection_type();

  const $editColumnModal = $("#edit-column-modal");
  clearColumnErrors();

  // Show the validation spinner
  $editColumnModal.find('[arcid=validate-spinner]').show();

  // Pluck the current values from the modal
  let alias = $editColumnModal.find("input[arcid=column-display]").val().trim();
  const comment = $editColumnModal.find("[arcid=column-comment]").val().trim();
  let expr = table_editor ? table_editor.value().trim() :
    $editColumnModal.find(".column-expression-edit .ace_content").text().trim();
  const esplit = arc.misc.splitExprAndAlias(expr);
  if (esplit.alias) {
    expr = esplit.expr;
    alias = alias || esplit.alias;
    $editColumnModal.find("[arcid=column-display]").val(alias);
  }
  const dflagg = arc.datasets.findListEntry(arc.datasets.dflagg_table,
    $editColumnModal.find("[arcid=dsattr-default-agg]").val().trim(), 0, 1);

  const geoType = $editColumnModal.find("[arcid=dsattr-geotype]").val() || 'none';

  const displayformat = displayOptions.parseDisplayFormat(false);

  if (displayformat.type === 'js') {
    // Make sure to test the function
    displayOptions.updateSample();
  }

  const newAttrs = {
    alias,
    expr,
    isdim: ($editColumnModal.find("[name=radio-category]:checked").val() === 'dimension'),
    hidden: !($editColumnModal.find("[arcid=column-visibility]").is(":checked")),
    isSensitive: ($editColumnModal.find('[arcid=column-sensitivity]').is(':checked')),
    detaildata: $editColumnModal.find("[arcid=column-detaildata]").is(":checked"),
    ispartition: $editColumnModal.find("[arcid=column-partition]").is(":checked"),
    useroverrodeisdim: $editColumnModal.find('#ecmodal-user-overrode').is(':checked'),
    dflagg: dflagg,
    geotype: geoType,
    displayformat: displayformat,
    type: col.coltype(),
  };

  arc.admin.columnColorHandler(col).fetch(newAttrs);

  if (comment !== col.attr("hivecomment")) {
    newAttrs.comment = comment;
  }

  // This is the completion function called after the validation process
  // Update the modal and do the success callback as appropriate
  var validate_done = function(params) {
    let { err, fieldLevelError, $errField, type, ispart, info } = params;
    let { hasagg } = params;

    const loc = $errField ? $errField.closest('.form-group') :
      $editColumnModal.find('.tab-content');

    if (err) {
      if (mustValidate) {
        if ($errField && fieldLevelError) {
          $errField.next('.form-text.with-errors').text(err);
          $errField.closest('.form-group').addClass('has-error');
        } else {
          loc.after(arc.clone_template('col-val-failed',
              {errtext: arc.misc.escapeHTML(err)}).html());
        }
      } else {
        if (success_cb) {
          newAttrs.ispartition = false;
          newAttrs.exprhasagg = $('#expr-has-agg').prop('checked') ? true : false;

          if (newAttrs.exprhasagg) {
            newAttrs.isdim = false;
          }

          success_cb({
            newAttrs,
            ispart: newAttrs.ispartition,
            hasagg: newAttrs.exprhasagg,
            type: newAttrs.type,
            loc,
          });
        }
      }

      return;
    }

    // Remove any existing alert text
    clearColumnErrors();
    $editColumnModal.find('[arcid=validate-spinner]').hide();

    // Update modal dimension vs measure
    if (col) {
      if (!col.attr('name')) {
        // XXX TODO: KSQL validation support
        // Detection of ksql aggregations does not work right now
        // Added later: Hack below was causing partition key to be set to true
        // for derived fields for KSQL, adding more to the hack for now
        if (dctype === 'ksql') {
          hasagg = false;
          ispart = false;
          type = 'TEXT';
        }

        if (hasagg) {
          newAttrs.exprhasagg = true;
        } else {
          newAttrs.exprhasagg = false;
        }
        $('#expr-has-agg').prop('checked', newAttrs.exprhasagg === true);

        newAttrs.ispart = ispart;
        newAttrs.type = type;

        if (!newAttrs.useroverrodeisdim) {
          if (hasagg) {
            newAttrs.isdim = false;
          }

          newAttrs.isdim = !(arc.report.measure_types[type]);
          const selectorId = `#ecmodal-${newAttrs.isdim ? 'dim' : 'agg'}-radio`;
          $editColumnModal.find(selectorId).prop('checked', true);
        }
      }
    }

    if (success_cb) {
      success_cb({
        newAttrs,
        ispart: ispart || newAttrs.ispartition,
        hasagg,
        type,
        loc,
      });
    }
  };

  // Now proceed with the validation.
  const table = col.table();
  const dataset = table.dataset();
  let err, $errField;
  if (alias) {
    err = dataset.check_alias(col, alias);
    if (err) {
      return validate_done({ err });
    }
  }

  err = displayOptions.errorMsg();
  if (displayformat.type === 'js' && err) {
    return validate_done({ err: `Error thrown by Custom JS Format Function: ${err}` });
  }

  if (alias.includes('(') || alias.includes(')')) {
    return validate_done({ err: 'Display name can\'t contain parenthesis',
      fieldLevelError: true,
      $errField });
  }

  // Base column update (alias, comment, and dflagg only) -- no further checking
  if (col.isBaseCol()) {
    return validate_done({});
  }

  // Computed column validation
  if (!alias) {
    $errField = $('[arcid=column-display]');
    return validate_done({ err: 'Must specify display name for computed fields',
      fieldLevelError: true, $errField });
  }

  if (!expr) {
    return validate_done({ err: 'Must specify expression for computed fields' });
  }

  if (arc.cviz.isCvizExpr(expr)) {
    let s;

    try {
      s = arc.cviz.cvizSettingsFromExpr(expr);
    } catch (e) {
      return validate_done({ err: 'Remote call settings are not valid JS: ' + e });
    }

    if (!s) {
      return validate_done({ err: 'Invalid remote call settings' });
    }

    if (!s.url) {
      return validate_done({ err: 'URL must be specified for remote call' });
    }
  }

  if (!mustValidate) {
    return validate_done({
      err: 'validation not required',
    });
  }

  // So far ok so do back end expression validation through host
  // (which will call the validate_done() with the validation results
  arc.validate_expression(expr, dataset, validate_done, 'column', {
    alias,
    skipcol: col,
    inclds: true,
  });
};

// Apply the column edit changes to the existing column
arc.datasets.applyColumnEditChanges = function(col, params) {
  const newAttrs = params.newAttrs;
  const old_alias = col.attr('alias') || col.attr('name');
  const keys = Object.keys(newAttrs);
  keys.forEach(function(k) {
    col.attr(k, newAttrs[k]);
  });
};

arc.datasets.getAssociatedDashboards = function (datasetId) {
  const url = `${arc.frontend.globals.urls.datasetUrl}/associated_dashboards/${datasetId}`;

  return arc.frontend.vizHttp.get({
    url,
    headers: {
      'X-CSRFTOKEN': arc.globals.csrf_token,
    },
  })
  .then(({ data }) => data)
  .catch((error) => {
    throw new Error(
      `Failed to get associated dashboards (error: ${error.responseText ?? error
      })`
    );
  });
};
}());
