window.arc = window.arc || {};

(function() {
'use strict';

/*
 * Track a set of extensions, applied to a single visual or a single app.
 */
arc.extensions = function() {
  var extensionsInOrder = [],
      extensionsHash = {},
      inlineExtensions = [];

  var extensions = function() {
    return 'arc.extensions';
  };

  // This function is used in testing. We assign inline JS code in tests.
  extensions.setInlineJsAssets = function(assets, viz_lib) {
    inlineExtensions =
        assets.map(function(asset) {
          return arc.extension()
                    .js(asset)
                    .processJs(viz_lib);
        });
  };

  // Return an extension by ID
  extensions.byUniqueId = function(_) {
    return extensionsHash[_];
  };

  // Return an extension by ID
  extensions.byExtensionId = function(_) {
    return extensionsInOrder.filter(function(e) {
      return e.extensionId() === _;
    });
  };

  // Return a list of extensions to iterate through
  extensions.all = function() {
    return extensionsInOrder.concat(inlineExtensions);
  };

  // Remove an extension from the list of extensions being
  // tracked
  extensions.removeUniqueId = function(uniqueId) {
    var newExtensions = [];
    extensionsInOrder.forEach(function(e) {
      if (e.uniqueId() === uniqueId) {
        return;
      }

      newExtensions.push(e);
    });

    extensionsInOrder = newExtensions;
    delete extensionsHash[uniqueId];

    return extensions;
  };

  // Add in an extension to the list of extensions being tracked.
  extensions.add = function(extension) {
    extensionsInOrder.push(extension);
    extensionsHash[extension.uniqueId()] = extension;
  };

  // Call out to the metastore and fetch the JS associated with all
  // extensions tracked by this object.
  extensions.fetchExtensions = function(viz_lib, cb, from_cache) {
    var extensionIds =
      extensionsInOrder.map(function(e) {
                         return e.extensionId();
                       })
                       .filter(function(e) {
                         return e !== undefined;
                       });

    if (extensionIds.length === 0) {
      // Nothing to fetch - make the callback
      if (cb) {
        cb(false);
      }

      return;
    }

    if (!from_cache) {
      arc.misc.customJsUnload(extensionIds);
    }

    var freshJs = 0;
    var t1 = new Date();

    if (arc.testutil && arc.testutil.testMode) {
      return;
    }

    arc.misc.customJsFetcher(extensionIds, function(error, assets, ids) {
      var t2 = new Date();
      if (error) {
        if (viz_lib) {
          viz_lib.alertMessage(`Error fetching custom js: ${error.responseText}`);
        } else {
          arc.report.alertMessage(
            'Error fetching custom js',
            `${error.responseText}. Custom style does not exist or may have been deleted.`
          );
        }

        return;
      }

      var jsById = {};
      assets.forEach(function(asset, idx) {
        var jsId = ids[idx];
        extensions.byExtensionId(jsId).forEach(function(e) {
          var doLoad = true;

          if (e.js()) {
            // We already had some js associated with this extension.
            // Check if it is different from the new js
            if (e.js() === asset) {
              doLoad = false;
            } else {
              ++freshJs;
            }
          } else {
            if (asset) {
              ++freshJs;
            }
          }

          if (doLoad) {
            // In dashboard, viz_lib is not set, and JS extensions do not
            // make sense (but CSS extensions do).
            // Do not process js within added extensions.
            if (viz_lib) {
              e.js(asset);
              e.processJs(viz_lib);

              logLoad(e, viz_lib, t2 - t1);
            }
          }
        });
      });

      if (cb) {
        cb(freshJs > 0);
      }
    });
  };

  function logLoad(e, viz_lib, timeTaken) {
    var id = e.extensionId();
    var appId = viz_lib.report_id();
    const extensionUrl = `${arc.frontend.globals.urls.customextensionEditUrl}/${id}`;
    const appUrl = `${arc.frontend.globals.urls.viewDashboardUrl}/${appId}`;
    arc.perf.PAGEPERF.addProfile({
      id,
      type: 'Load Extension',
      time: timeTaken,
      profile: [
        [
          'App',
          `<a target="_blank" href="${arc.misc.getBlankTargetUrl(appUrl)}">${appId}'</a>`,
        ],
        [
          'Extension',
          `<a target="_blank" href="${arc.misc.getBlankTargetUrl(extensionUrl)}">${id}'</a>`,
        ],
      ],
    });
  }

  return extensions;
};

arc.extensionId = 1;

/*
 * Track a single extension, storing the associated extension ID
 * and javascript, and with helpers to process the JS extension
 * into code that is bound with the appropriate arcapi object.
 */
arc.extension = function() {
  var js,
      extensionId,
      code,
      settings,
      error,
      savedSettings,
      uniqueId = arc.extensionId++;

  var extension = function() {
    return 'arc.extension';
  };

  extension.uniqueId = function(_) {
    if (!arguments.length) return uniqueId;
    uniqueId = _;
    return extension;
  };

  extension.extensionId = function(_) {
    if (!arguments.length) return extensionId;
    extensionId = _;
    return extension;
  };

  // Fetch only
  extension.code = function() {
    return code;
  };

  extension.settings = function(_) {
    if (!arguments.length) {
      var s;

      if (settings) {
        s = settings.slice();
      } else {
        s = [];
      }

      if (savedSettings) {
        $.each(savedSettings, function(k, v) {
          s.forEach(function(ss) {
            if (ss.id === k) {
              ss.value = v;
            }
          });
        });
      }

      return s;
    }

    settings = _;
    return extension;
  };

  extension.savedSettings = function(_) {
    if (!arguments.length) return savedSettings;
    savedSettings = {};
    $.each(_, function(k, v) {
      savedSettings[k] = v;
    });

    return extension;
  };

  extension.settingsToSave = function(newSettings) {
    var s = {};

    if (settings) {
      settings.forEach(function(setting) {
        s[setting.id] = setting.value;
      });
    }

    if (savedSettings) {
      $.each(savedSettings, function(k, v) {
        s[k] = v;
      });
    }

    if (newSettings) {
      $.each(newSettings, function(k, v) {
        s[k] = v;
      });
    }

    return s;
  };

  extension.error = function(_) {
    if (!arguments.length) return error;
    error = _;
    return extension;
  };

  extension.js = function(_) {
    if (!arguments.length) return js;
    js = _;
    return extension;
  };

  function getValue(setting) {
    if (savedSettings) {
      if (setting.id in savedSettings) {
        return savedSettings[setting.id];
      }
    }

    if (setting.value !== undefined) {
      return setting.value;
    }

    return setting.defaultValue;
  }

  function processSettings(settings) {
    settings.forEach(function(setting) {
      // Setup default values for 'type' and 'displayName'
      if (!setting.type) {
        setting.type = 'string';
      }

      if (!setting.displayName) {
        setting.displayName = setting.id;
      }

      setting.value = getValue(setting);
    });

    return settings;
  }

  // Process the JS associated with this extension. This means the following
  // things:
  // 1. create and bind the arcapi object to the function
  // 2. See if the version number associated with the function object is
  //    what it needs to be. See 'checkVersion' below.
  // 3. Check if the function object is supported in the context of this
  //    visual i.e. if the visual type is the same as this function object.
  //    See 'checkSupported' below.
  // 4. Process and handle the custom settings associated with this visual
  //    as it relates to this extension. See 'handleSettings' below.
  // 5. Store away the bound code associated with this extension for use
  //    later. See the association of the bound function to the 'code' object
  //    below.
  // 6. In all this processing, if there are any errors we find that makes
  //    this extension unusable, we set 'error'. This variable is accessed and
  //    used to display errors.
  extension.processJs = function(viz_lib) {
    extension.error(undefined);
    code = undefined;
    extension.settings(undefined);

    if (!js) {
      return extension;
    }

    if (js.trim().length === 0) {
      return extension;
    }

    // Setup arcapi. Fresh for each extension in case some extensions
    // modify it (this is not supported or expected, but it can happen)
    var arcapi = {};
    arcapi.getExtensionId = function() {
      return arcapi.unsupportedDriver.report_id() + '*' + extensionId;
    };

    arcapi.chartId = function() {
      return viz_lib.chart_div();
    };

    arcapi.dataResult = function() {
      return viz_lib.dataResult();
    };

    arcapi.visualType = function() {
      return viz_lib.viz_type();
    };

    arcapi.addScripts = arc.misc.addScripts;

    arcapi.addStyles = arc.misc.addStyles;

    arcapi.sendParameters = function(paramObjects, config={}) {
      if (config.version && config.version === '2') {
        viz_lib.dashboard_click(paramObjects);
        return;
      }

      const scopeConfig = config;
      if (!(paramObjects instanceof Array)) {
        // Deprecate receiving of old format params
        console.log('Params received in old format. These will be deprecated soon.');
        paramObjects = arc.params.convertOldParamHashToNewParamObjects(
          paramObjects, viz_lib, scopeConfig, arcapi);
      }

      viz_lib.dashboard_click(paramObjects);
    };

    arcapi.getParameters = function(config={}) {
      let retParams = {};
      let currentParams;
      if (arc.globals.is_dashboard) {
        if (arc.globals.dashboard.params_history()) {
          currentParams = arc.globals.dashboard.params_history().current_params()[0];
          if (currentParams) {
            currentParams = currentParams.slice(0);
          }
        }
      } else {
        currentParams = viz_lib.getParams().getkvsList();
      }

      if (config.version && config.version === '2') {
        retParams = arc.params.convertOldParamArrToNewParamObjects(currentParams);
      } else {
        console.log('Params requested in old format. These will be deprecated soon.');
        currentParams.forEach(function (cp) {
          retParams[cp.key] = cp.value;
        });
      }

      return retParams;
    };

    /**
     * NOTE: Received callback executes synchronously when dashboard params get updated.
     * @param config
     * config.callbackName- name of callback when being registered
     * config.callbackFn  - function to be called when params are updated but not yet broadcasted.
     * config.callbackFn(
     *  paramObjects,     - array of version 2 format paramObjects that are about to be distributed.
     *  doNotRefresh      - Will the visuals be refreshed or just dashboard history be updated.
     * )
     * return {
     *  paramObjects,     - new array of version 2 that should be broadcasted instead.
     *  doNotRefresh      - Should the visuals be refreshed or just dashboard history be updated.
     * }
     */
    arcapi.beforeParamDistributionCb = function(config={}) {
      if (!config.callbackFn) {
        console.log('Incomplete argument to arcapi.beforeParamDistributionCb.' +
                    'Expected config.callbackFn to be a function.');
        return;
      }

      if (!config.callbackName) {
        console.log('Incomplete argument to arcapi.beforeParamDistributionCb.' +
                    'Expected config.callbackName to be a string.');
        return;
      }

      if (arc.globals.is_dashboard) {
        return arc.globals.ms_dashboard.registerParamDistributionCb(config);
      }
    };

    // Unsupported API interface. In place for POCs
    arcapi.unsupportedDriver = viz_lib;

    let origParamsFn = viz_lib.params;
    arcapi.unsupportedDriver.params = function () {
      console.warn('This function is deprecated.' + '\n' +
                  'Please use arcapi.getParameters() to get current parameters ' +
                  'or arcapi.sendParameters(arrayOfParamObjects) to send new parameter values.');
      return origParamsFn.apply(this, arguments);
    };

    var _settings;
    arcapi.settings = function() {
      return _settings;
    };

    arcapi.getSetting = function(_) {
      var s;

      _settings.some(function(setting) {
        if (setting.id === _) {
          s = setting;
          return true;
        }

        return false;
      });

      if (s) {
        return s.value;
      }

      return undefined;
    };

    // Verify that the version number of the extension is supported by
    // this version of arcviz. If folks just write in random JS instead
    // of following the api, this is the first place we see errors.
    function checkVersion(fn) {
      if (fn.version === undefined) {
        error = 'Extension is not supported because it does not include a ' +
            '"version" string';
        viz_lib.alertMessage(error);
        return false;
      }

      if (fn.version === '1') {
        return true;
      }

      error = 'Error registering custom js. Version number ' +
          'is unsupported: ' + fn.version;
      viz_lib.alertMessage(error);
      return false;
    }

    // Verify that this extension is supported in the context of this
    // visual. Either there is not 'supported' member in the function
    // object, or there is a supported member and it includes a visualType
    // setting that works.
    function checkSupported(fn) {
      if (!fn.supported) {
        return true;
      }

      if (!fn.supported.visualType) {
        return true;
      }

      var v = fn.supported.visualType;
      if (typeof (v) === typeof ([])) {
        if (v.indexOf(arcapi.visualType()) < 0) {
          error = 'Extension not valid because it only works ' +
              'with visuals of types ' + v.join(', ');
          return false;
        }

        return true;
      }

      if (v !== arcapi.visualType()) {
        error = 'Extension not valid because it only works ' +
            'with visuals of type ' + fn.supported.visualType;
        return false;
      }

      return true;
    }

    // Check and process the settings object for this extension.
    // Among other things, associated the stored settings for this extension
    // in the context of this visual.
    function handleSettings(fn) {
      if (fn.settings) {
        _settings = fn.settings();
        processSettings(_settings);
      } else {
        _settings = [];
      }

      _settings.forEach(function(setting) {
        if (setting.error) {
          if (error) {
            error = error + ', "' + setting.displayName + '"';
          } else {
            error = 'Invalid settings for "' + setting.displayName + '"';
          }
        }
      });
    }

    try {
      const extensionFunc = Function('arcapi', js).bind(undefined, arcapi); // eslint-disable-line no-new-func
      const fn = extensionFunc();

      if (!checkVersion(fn)) {
        return extension;
      }

      if (!checkSupported(fn)) {
        return extension;
      }

      extension.error(undefined);
      handleSettings(fn);
      code = fn;
      extension.settings(_settings);
    } catch (e) {
      error = 'Error registering custom js: ' + e.toString();
      viz_lib.alertMessage(error);
      extension.error(error);
      return;
    }

    return extension;
  };

  return extension;
};

}());
