window.arc = window.arc || {};

(function() {
  'use strict';

arc.perf = arc.perf || {};

const loadSkeleton = `
  <div style="font-size:18px;">
    You are running <span id="current-load-active-count-user"></span> active queries and
    <span id="current-load-pending-count-user"></span> pending queries.<br/>
    Across all users, there are <span id="current-load-active-count"></span> active queries
    and <span id="current-load-pending-count"></span> pending queries.
  </div>
  <div style="font-size:14px;">
    Note that queries can be pending if you are at the per-user or the overall concurrency
    limits on a data connection. The table below includes information about each connection
    with active queries on it.
  </div>
  <div class="perf-section-title">
    <strong>Active Queries Summary</strong>
  </div>
  <div id="current-load-table">
  </div>
  <div class="perf-section-title">
    <strong>Active Queries Detail</strong>
  </div>
  <div id="current-load-queries">
  </div>
`;

arc.perf.cancelQuery = function(connectionId, querytracker, allusers) {
  let url;

  if (allusers) {
    url = `${arc.frontend.globals.urls.asynCancelQtUrl}/all`;
  } else {
    url = `${arc.frontend.globals.urls.asynCancelQtUrl}`;
  }

  const toCancel = [{
    connectionId,
    querytracker,
  }];

  const paramsToSend = {
    data: JSON.stringify(toCancel),
  };

  arc.sqlrun.httpAsyncCall({
    url,
    paramsToSend,
    doneCb: (responsel) => {
      let total = 0;
      responsel.forEach((d) => {
        total += d.cnt;
      });

      if (total === 1) {
        arc.report.alertMessageSmall(`Canceled ${total} query`, 'alert-success');
      } else {
        arc.report.alertMessageSmall(`Canceled ${total} queries`, 'alert-success');
      }
    },
    failCb: () => {
      arc.report.alertMessageSmall('Failed to cancel queries', 'alert-warning');
    },
  });
};

arc.perf.drawQueues = function(queues, divId, allUsers) {
  let colnames;

  if (allUsers) {
    colnames = ['Queue', 'Per User Capacity', 'Capacity (All)', 'Running (All)', 'Queued (All)'];
  } else {
    colnames = ['Queue', 'Capacity (Yours)', 'Running (Yours)', 'Queued (Yours)',
        'Capacity (All)', 'Running (All)', 'Queued (All)'];
  }

  const rows = [];
  const queueNames = Object.keys(queues).sort();
  let activeQueriesUser = 0;
  let pendingQueriesUser = 0;
  let activeQueriesAll = 0;
  let pendingQueriesAll = 0;

  queueNames.forEach((queueName) => {
    const qData = queues[queueName];

    const {
      active,
      capacity,
      perUserCapacity,
      pendingCount
    } = qData;

    if (queueName !== 'http') {
      activeQueriesAll += active;
      pendingQueriesAll += pendingCount;

      activeQueriesUser += qData.curr.length;
      pendingQueriesUser += qData.pending.length;
    }

    if (active || pendingCount) {
      if (allUsers) {
        rows.push([queueName, perUserCapacity || '-', capacity || '-', active, pendingCount]);
      } else {
        rows.push([queueName, perUserCapacity || '-', qData.curr.length, qData.pending.length,
            capacity || '-', active, pendingCount]);
      }
    }
  });

  arc.misc.drawTable(divId, rows, colnames, [
      {
        key: 'table_wrap_column_contents',
        value: 'true'
      },
      {
        key: 'table_full_width',
        value: 'true'
      },
  ]);

  return {
    activeQueriesUser,
    pendingQueriesUser,
    activeQueriesAll,
    pendingQueriesAll,
  };
};

arc.perf.drawQueries = function(queues, divId, showCancelAction) {
  const colnames = ['Queue', 'Query Tracker', 'Start Time', 'Runtime', 'Username',
      'Status'];

  if (showCancelAction) {
    colnames.push('Action');
    colnames.push('Query');
  } else {
    colnames.push('Query');
  }

  const rows = [];
  const queueNames = Object.keys(queues).sort();

  queueNames.forEach((queueName) => {
    const qData = queues[queueName];
    qData.curr.forEach((q) => {
      const j = JSON.parse(q.request);
      const info = j.runner ?? j;

      if (showCancelAction) {
        rows.push([queueName, q.querytracker, q.startTime,
            q.runtimeOngoing, q.username, 'Running', '', info.query]);
      } else {
        rows.push([queueName, q.querytracker, q.startTime,
            q.runtimeOngoing, q.username, 'Running', info.query]);
      }
    });

    qData.pending.forEach((q) => {
      const j = JSON.parse(q.request);
      const info = j.runner ?? j;

      if (showCancelAction) {
        rows.push([queueName, q.querytracker, q.startTime, q.runtimeOngoing, q.username,
            'Pending', '', j.query || info.query]);
      } else {
        rows.push([queueName, q.querytracker, q.startTime, q.runtimeOngoing, q.username,
            'Pending', j.query || info.query]);
      }
    });
  });

  arc.misc.drawTable(divId, rows, colnames, [
      {
        key: 'table_full_width',
        value: 'true'
      },
      {
        key: 'display_tooltips_table',
        value: 'true'
      },
  ], {
      showTooltip: true,
      orderBy: {
        colname: 'Start Time',
        order: arc.report.shelf.order()
        .priority(1)
        .ascending(true)
      }
    });

  if (showCancelAction) {
    d3.selectAll(`#${divId} table tbody tr td:nth-child(7)`)
        .each(function() {
          const $this = $(this);
          $this.html(`<a class="cancel-query">Cancel</a>`);
        })
        .on('click.cancel', function() {
          const $this = $(this);
          const connectionId = $this.closest('tr').find('td:nth-child(1)').text();
          const querytracker = $this.closest('tr').find('td:nth-child(2)').text();
          arc.report.alertMessageSmall('Canceling query', 'alert-success');
          arc.perf.cancelQuery(connectionId, querytracker);
        });
  }

  // Setup 'Download' button within perf modal
  $('#perf-modal .current-load-download')
      .offOn('click.download', () => {
        arc.misc.downloadCsv('load.txt', colnames, rows);
      });
};

arc.perf.currentLoad = function() {
  const currentLoad = function() {
    return 'arc.perf.currentLoad';
  };

  let statsData;

  currentLoad.start = function() {
    currentLoad.draw();

    $('#perf-modal .current-load-refresh')
        .offOn('click.download', () => {
          currentLoad.draw();
        });

    $('#perf-modal .current-load-cancel-all')
        .offOn('click.download', () => {
          arc.perf.cancelQuery('*', '*');
        });
  };

  currentLoad.draw = function() {
    d4.queue()
      .defer(d3.json, arc.frontend.globals.urls.statsDataUrl)
      .await((error, _) => {
        if (error) {
          arc.report.alertMessage('Failed to load server stats',
              error.response ? error.response : error, { alertContainerSelector: '.perf-contents' });
          return;
        }

        statsData = _;
        paint();
      });

    return currentLoad;
  };

  function paint() {
    $('#perf-modal .current-load-download')
        .off('click.download');

    $('#current-load-info').empty();
    $('#current-load-info').append(loadSkeleton);

    const queueNames = Object.keys(statsData.requests.queues).sort();
    const queues = {};
    queueNames.forEach((queueName) => {
      const qData = statsData.requests.queues[queueName];

      if (queueName.endsWith('async.workq') &&
          queueName.startsWith('conn_')) {
        const displayQueueName = queueName.substring(5, queueName.indexOf('.'));
        queues[displayQueueName] = qData;
      }
    });

    const s = arc.perf.drawQueues(queues, 'current-load-table', 'Connection');
    const {
      activeQueriesUser,
      pendingQueriesUser,
      activeQueriesAll,
      pendingQueriesAll,
    } = s;

    $('#current-load-active-count-user').text(activeQueriesUser);
    $('#current-load-pending-count-user').text(pendingQueriesUser);
    $('#current-load-active-count').text(activeQueriesAll);
    $('#current-load-pending-count').text(pendingQueriesAll);

    arc.perf.drawQueries(queues, 'current-load-queries', true);
  }

  return currentLoad;
};
}());
