window.arc = window.arc || {};

(function() {
'use strict';

arc.errors = arc.errors || {};

let errorPopoverInstalled = false;

/*
 * All current page errors are tracked here.
 *
 * They are displayed when a user clicks on one of the error icons on the page.
 *
 * They are cleared when the user requests it.
 *
 * Per-visual errors are cleared when a visual redraws.
 */
arc.errors.pageErrorTracker = (function() {
  let p = function() { return 'pageErrorTracker'; };

  let errors = {};

  p.clearAllErrors = function() {
    errors = {};
  };

  p.clearErrors = function(visualId) {
    let e = errors[visualId];
    if (e) {
      delete errors[visualId];
    }

    return p;
  };

  p.totalErrorCount = function() {
    let cnt = 0;

    Object.keys(errors).forEach(function(v) {
      cnt += errors[v].length;
    });

    return cnt;
  };

  p.addError = function(visualId, msg, detail = '', errorConfig = {}) {
    let e = errors[visualId];
    if (!e) {
      errors[visualId] = [];
      e = errors[visualId];
    }

    e.push({
      msg,
      detail,
    });

    if ((!arc.testutil || !arc.testutil.ignoreAlerts) && !errorConfig.noConsoleLog) {
      console.error(`${msg}: ${detail} (ID: ${visualId})`);
    }

    arc.report.logAlertMessages('alert-danger', {
      type: 'big',
      header: msg,
      message: detail,
    });

    return p;
  };

  p.getErrors = function(visualId, visualId2) {
    let e = errors[visualId] || [];

    if (visualId === 'Page') {
      Object.keys(errors).forEach(function(v) {
        if (v === visualId) {
          return;
        }

        e = e.concat(errors[v]);
      });
    }

    if (!visualId2) {
      return e;
    }

    return e.concat(errors[visualId2] || []);
  };

  return p;
}());

arc.errors.displayPageLevelAlert = function(msg, detail) {
  arc.errors.initialize();
  arc.errors.reportError(undefined, msg, detail);
};

arc.errors.clearAllErrors = function() {
  $('.page-error-state')
      .removeClass('page-error-state');
  $('.error-state')
      .removeClass('error-state');
  arc.errors.pageErrorTracker.clearAllErrors();
};

arc.errors.initialize = function() {
  if (errorPopoverInstalled) {
    return;
  }

  errorPopoverInstalled = true;

  $('.super-modal .viz-error-msg-wrapper')
      .addClass('viz-error-page-wrapper');

  let $pageErrors = $('.viz-error-page-wrapper .viz-error-msg');
  if ($pageErrors.length > 0) {
    arc.errors.setupErrorPopover(undefined, $pageErrors);
  }

  if (arc.globals.jserrors_enabled) {
    window.onerror = function(message, source, lineno, colno, error) {
      arc.report.alertMessage('Execution error',
          `Source: ${source}, Line: ${lineno}, Col: ${colno}, Error: ${error}, Message: ${message}`);
    };
  }

  $('body, .super-modal')
      .off('click.error-popover')
      .on('click.error-popover', function(e) {
        const $target = $(e.target);
        if ($target.closest('.viz-error-popover').length > 0) {
          return;
        }

        arc.errors.dismissPopovers();
      });
};

const allPopoversSel = '.arc-viz-container .viz-error-msg, .viz-error-page-wrapper .viz-error-msg';

/*
 * Specialized handling to be able to get to .arc-viz-container for both visuals
 * and filters.
 */
function getContainer(container_div_id) {
  const $c = $(`#${container_div_id}`);
  if ($c.hasClass('arc-viz-container')) {
    // Filters
    return $c;
  }

  // Visuals
  return $c.closest('.arc-viz-container');
}

arc.errors.setupVisualErrorPopover = function(vizDriver, container_div_id) {
  const $elem = getContainer(container_div_id).find('.viz-error-msg');
  arc.errors.setupErrorPopover(vizDriver, $elem);
};

function visualId(vizDriver) {
  if (vizDriver) {
    return `${vizDriver.getErrorPrefix()} ${vizDriver.report_id()}`;
  }

  return 'Page';
}

arc.errors.setupErrorPopover = function(vizDriver, $elem) {
  let placement = 'right';
  if (!vizDriver) {
    // Page-level errors open down
    placement = 'bottom';
  } else if ($elem.closest('.super-modal').length > 0) {
    // In-modal errors open down
    placement = 'bottom';
  }

  if ($elem.length === 0) {
    arc.errors.initialize();
    return;
  }

  $elem.data('vizDriver', vizDriver);

  if ($elem.data && $elem.data('bs.popover')) {
    // There is a popover already - don't redefine it
    arc.errors.initialize();

    return;
  }

  $elem
      .popover({
        animation: false,
        html: true,
        trigger: 'manual',
        placement,
        container: 'body',
        sanitize: false,
        title() {
          const storedVizDriver = $elem.data('vizDriver');
          if (storedVizDriver) {
            return `Error - ${visualId(storedVizDriver)}`;
          }

          let cnt = arc.errors.pageErrorTracker.totalErrorCount();
          return `${cnt} Errors`;
        },

        content() {
          const storedVizDriver = $elem.data('vizDriver');
          let visualId2;
          if (storedVizDriver && ($elem.closest('.super-modal').length > 0)) {
            // If this popover is for a in-modal visual, add in page-level
            // errors here
            visualId2 = visualId();
          }

          const errors = arc.errors.pageErrorTracker.getErrors(visualId(storedVizDriver),
              visualId2);
          if (errors.length === 0) {
            return '';
          }

          const lis = errors.map(e => arc.misc.safe_substitute(
            $('#viz-single-error-template ul').html(),
            {
              SUMMARY: e.msg,
              DETAIL: e.detail,
            }
          ));

          return arc.misc.safe_substitute(
            $('#viz-error-popover-template').html(),
            {
              HTML_LIST: lis.join(''),
            }
          );
        },
      })
      .offOn('shown.bs.popover', function() {
        const storedVizDriver = $elem.data('vizDriver');
        const $that = $(this);

        $('.popover.show .popover-close')
            .offOn('click', () => {
              $that.popover('hide');
            });

        $('.popover.show .clear-error')
            .offOn('click', () => {
              $that.popover('hide');
              arc.errors.pageErrorTracker.clearErrors(visualId(storedVizDriver));
              if (storedVizDriver) {
                if ($that.closest('.super-modal').length > 0) {
                  // Also clear page-level errors if we are within a super modal
                  arc.errors.pageErrorTracker.clearErrors(visualId());

                  $('.super-modal')
                      .removeClass('page-error-state');
                }
              } else {
                arc.errors.clearAllErrors();
                return;
              }

              if (arc.errors.pageErrorTracker.totalErrorCount() === 0) {
                $('.page-error-state')
                    .removeClass('page-error-state');
              }

              $that.closest('.error-state')
                   .removeClass('error-state');
            });

        $('.popover.show .copy-error')
            .offOn('click', function() {
              const errors = arc.errors.pageErrorTracker.getErrors(visualId(storedVizDriver));

              // Concat all errors together to copy/paste in email etc
              const str = errors.map((e) => [e.msg, e.detail].join('\n')).join('\n');

              arc.misc.copyToClipboard(str);

              const $btn = $(this);
              $btn.attr('data-toggle', 'tooltip')
                  .attr('title', 'copied')
                  .tooltip('show')
                  .one('mousemove', () => {
                    $btn.attr('title', null)
                        .attr('data-original-title', null)
                        .attr('data-toggle', null);

                    $btn.tooltip('dispose');
                  });
            });
      })
      .offOn('click', function(e) {
        const _this = this;
        const $this = $(this);
        if ($this.closest('.popover-close').length > 0) {
          $this.popover('hide');
          return;
        }

        $(allPopoversSel)
            .each(function() {
              if (this === _this) {
                return;
              }

              $(this).popover('hide');
            });

        $this.popover('toggle');
        e.stopPropagation();
      });

  arc.errors.initialize();
};

arc.errors.clearErrors = function(vizDriver) {
  if (vizDriver) {
    const $vizContainer = getContainer(vizDriver.chart_div());
    $vizContainer.removeClass('error-state');
  }

  arc.errors.pageErrorTracker.clearErrors(visualId(vizDriver));

  // Check if all errors on dashboard/visual have been cleared
  if (arc.errors.pageErrorTracker.totalErrorCount() === 0) {

    // Remove error icon from page if there are no errors left
    $('.page-error-state').removeClass('page-error-state');
    $('.error-state').removeClass('error-state');
  }
};

arc.errors.reportError = function(vizDriver, msg, detail, errorConfig) {
  if (vizDriver && !vizDriver.report_id()) {
    arc.report.alertMessage(msg, detail, errorConfig);
    return;
  }

  arc.errors.pageErrorTracker.addError(visualId(vizDriver), msg, detail, errorConfig);

  // Show page-level error mark for all errors
  $('.viz-error-page-wrapper .viz-error-msg')
    .closest('.main-app')
    .addClass('error-state page-error-state');

  if (!vizDriver) {
    // Detail data modal opens on top of any existing modal
    const $detailModal = $('#datadetail-modal');
    if ($detailModal.is(':visible')) {
      $detailModal.addClass('error-state page-error-state');
      $detailModal.find('.viz-error-msg-wrapper')
                  .addClass('viz-error-page-wrapper');
    } else {
      // For page-level errors, mark super-modal error state as well
      $('.super-modal')
        .addClass('error-state')
        .addClass('page-error-state');
    }

    let $pageErrors = $('.viz-error-page-wrapper .viz-error-msg');
    if ($pageErrors.length > 0) {
      arc.errors.setupErrorPopover(vizDriver, $pageErrors);
    }
  } else {
    const $vizContainer = getContainer(vizDriver.chart_div());
    $vizContainer.addClass('error-state');
  }
};

arc.errors.dismissPopovers = function() {
  let $a = $(allPopoversSel);
  if ($a.length > 0) {
    $a.popover('hide');
  }
};

}());
