import { formatDialect, sql } from 'sql-formatter';

window.arc = window.arc || {};

(function() {
'use strict';

arc.perf = arc.perf || {};

arc.perf.MAX_PROFILES = 100;
arc.perf.PROFILE_DELETE_EXTRA = 20;

arc.perf.tracker = function() {
  const tracker = function() {
    return 'arc.perf.tracker';
  };

  const profiles = [];
  let currentLoad;

  tracker.addProfile = function(p) {
    p.profileTime = new Date().toISOString();
    profiles.push(p);

    if (profiles.length > arc.perf.MAXPROFILES) {
      profiles.splice(profiles.length - arc.perf.MAX_PROFILES -
          arc.perf.PROFILE_DELETE_EXTRA);
    }

    return tracker;
  };

  tracker.profiles = function() {
    return profiles;
  };

  tracker.display = function() {
    console.log(JSON.stringify(profiles));

    return tracker;
  };

  const profileColnames = [
    'Time', 'ID', 'Type', 'Duration (seconds)', 'Detail', 'Error',
  ];

  const profileColtypes = [
    'STRING', 'INTEGER', 'STRING', 'REAL', 'STRING', 'STRING',
  ];

  const counterColnames = [
    'Name', 'Value',
  ];

  const counterColtypes = [
    'STRING', 'STRING',
  ];

  function profileRows(config) {
    const rows = [];
    // eslint-disable-next-line no-shadow
    const profiles = arc.perf.PAGEPERF.profiles();

    config = config || {};
    config.forTable = config.forTable || false;
    if (!config.joinStr) {
      if (config.forTable) {
        config.joinStr = '<br>';
      } else {
        config.joinStr = ';';
      }
    }

    profiles.forEach((profile) => {
      const pl = profile.profile;
      let p;

      if (pl) {
        p = pl.map((v) => v.map((e, i) => {
                    if (config.forTable) {
                      return `<span class="perf-${i}">${e}</span>`;
                    }

                    if (i < v.length - 1) {
                      return `${e}:`;
                    }

                    return e;
                  })
                  .join('')).join(config.joinStr);
      } else {
        p = '';
      }

      rows.push([
          profile.profileTime,
          profile.id ? profile.id : '0',
          profile.type,
          profile.time / 1000,
          p,
          profile.error ? profile.error : '',
      ]);
    });

    const t = new Date().toISOString();
    $.each(arc.counters.data, (k, v) => {
      rows.push([
          t,
          '',
          'counter',
          0,
          `${k} ${v}`,
          '',
      ]);
    });

    return rows;
  }

  function drawPerfViz(divId, inRows) {
    const $d = $(`#${divId}`).empty();

    const $showall = $d.parent().find('.perf-showall');

    if (inRows.length === 0) {
      return;
    }

    function timeGetter(d) {
      return d[0];
    }

    function idGetter(d) {
      return d[1];
    }

    function typeGetter(d) {
      return d[2];
    }

    function durationGetter(d) {
      return Number(d[3]);
    }

    const noCounterRows = inRows.filter((d) => (typeGetter(d) !== 'counter'));

    if (noCounterRows.length === 0) {
      return;
    }

    const baseRows = noCounterRows;

    $showall.off('click')
            .on('click', () => {
              drawRows(baseRows);
              drawPerfTable('perf-table', inRows);
            });

    let lastDraw = {
      startTime: undefined,
      endTime: undefined,
    };

    function drawRows(rows) {
      $d.empty();

      if (rows.length === baseRows.length) {
        $showall.hide();
      } else {
        $showall.show();
      }

      rows.forEach((d) => {
        d.endTime = new Date(timeGetter(d));

        d.startTime = new Date(timeGetter(d));
        d.startTime.setMilliseconds(d.startTime.getMilliseconds() - (durationGetter(d) * 1000));
      });

      const idList = arc.misc.arrayUnique(rows, idGetter);
      idList.push('');

      const margin = { top: 5, bottom: 20, left: 40, right: 30 };
      const rowHeight = 15;

      const outerWidth = $d.width();
      const innerWidth = outerWidth - margin.left - margin.right;
      const innerHeight = idList.length * rowHeight;
      const outerHeight = innerHeight + margin.top + margin.bottom;

      const container = d3.select(`#${divId}`)
                        .append('svg')
                        .attr({ width: outerWidth, height: outerHeight });

      const svg = container.append('g')
                            .attr('transform', `translate(${margin.left},${margin.top})`);

      const brush =
          arc.visual.brush()
             .svg(svg)
             .initialize();

      const g = svg.append('g');

      const firstRow = rows[0];
      let minTime = firstRow?.startTime;
      let maxTime = firstRow?.endTime;

      rows.forEach((d) => {
        if (d.startTime < minTime) {
          minTime = d.startTime;
        }

        if (d.endTime > maxTime) {
          maxTime = d.endTime;
        }
      });

      const yScale = d3.scale.ordinal()
                     .domain(idList)
                     .rangePoints([10, innerHeight - 10]);

      const xScale = d3.time.scale()
                     .domain([minTime, maxTime])
                     .range([0, innerWidth]);

      const colorScale = arc.misc.dimensionColorScale();
      colorScale.init(true, {
        'Set data result': '#e31a1c',
        Query: '#1f78b4',
        Cancel: '#333',
        'Draw visual': '#33a02c',
        'Table rows processed': '#ff7f00',
        Parameters: 'darkgreen',
      });
      colorScale.range([
        '#a6cee3',
        '#b2df8a',
        '#fb9a99',
        '#fdbf6f',
        '#cab2d6',
        '#6a3d9a',
        '#ffff99',
        '#b15928',
      ]);

      function drawMarker(elem) {
        d3.select(elem)
          .append('circle')
          .attr('cx', (d) => xScale(d.endTime))
          .attr('cy', (d) => yPoint(d))
          .attr('r', () => 5)
          .style('fill', (d) => colorScale(typeGetter(d)));
      }

      function drawLine(elem) {
        d3.select(elem)
          .append('line')
          .attr('stroke', (d) => colorScale(typeGetter(d)))
          .attr('stroke-opacity', strokeOpacityScale)
          .attr('stroke-width', widthScale)
          .attr('x1', (d) => xScale(d.endTime))
          .attr('x2', (d) => xScale(d.endTime))
          .attr('y1', (d) => {
            const y = yPoint(d) - 5;
            return y;
          })
          .attr('y2', (d) => {
            const y = yPoint(d) + 5;
            return y;
          });
      }

      function drawEndMark(elem, d) {
        const t = typeGetter(d);
        if (t === 'Parameters') {
          drawMarker(elem, d);
        }

        drawLine(elem, d);
      }

      function strokeOpacityScale() {
        return '0.8';
      }

      function widthScale() {
        return '3';
      }

      const yAxis = d3.svg.axis()
                    .scale(yScale)
                    .orient('left');
      svg.append('g')
         .classed('arc-viz-axis', true)
         .call(yAxis);

      const xAxis = d3.svg.axis()
                    .scale(xScale)
                    .orient('bottom')
                    .ticks(5)
                    .tickFormat(d3.time.format('%H:%M:%S.%L'));
      svg.append('g')
         .attr('transform', `translate(0,${innerHeight - 10})`)
         .classed('arc-viz-axis', true)
         .call(xAxis);

      function yPoint(d) {
        return yScale(idGetter(d));
      }

      const endMarks =
      g.selectAll('.nosel')
         .data(rows)
         .enter()
           .append('g')
           .each(function(d) {
             drawEndMark(this, d);
           });

      const durationMarks =
      g.selectAll('.nosel')
         .data(rows)
         .enter()
           .append('line')
           .attr('stroke', (d) => colorScale(typeGetter(d)))
           .attr('stroke-opacity', strokeOpacityScale)
           .attr('stroke-width', widthScale)
           .attr('x1', (d) => xScale(d.startTime))
           .attr('x2', (d) => xScale(d.endTime))
           .attr('y1', yPoint)
           .attr('y2', yPoint);

      // Create a driver object for the tooltip_from_kvs function
      const table = arc.visual.plugins.table();
      const driver = arc.report.driver(table);

      const tooltips =
      arc.tooltip()
         .dom_id('#tooltip')
         .addSelection(endMarks)
         .addSelection(durationMarks)
         .text_fn((d) => driver.tooltip_from_kvs([
            {
              key: 'Duration',
              value: durationGetter(d),
              formatter: d3.format('.3f'),
            },
            {
              key: 'Id',
              value: idGetter(d),
            },
            {
              key: 'Type',
              value: typeGetter(d),
            },
            {
              key: 'Detail',
              value: d[4] ? d[4] : '',
              ishtml: true,
            },
            {
              key: 'Error',
              value: d[5] ? d[5] : '',
            },
           ]))
         .activate();

      // Make tooltips visible within the performance modal
      $(tooltips.dom_id()).css('z-index', '1100');

      endMarks.on('click.tooltip', (d, i) => {
        d3.event.stopPropagation();
        tooltips.updateTooltipText(d, i);
        tooltips.freeze();
      });

      durationMarks.on('click.tooltip', (d, i) => {
        d3.event.stopPropagation();
        tooltips.updateTooltipText(d, i);
        tooltips.freeze();
      });

      brush.xScale(xScale)
           .yScale(yScale)
           .cb((e) => {
             if (arc.misc.isIdentical(e[0], lastDraw.startTime) &&
                 arc.misc.isIdentical(e[1], lastDraw.endTime)) {
               return;
             }

             lastDraw = {
               startTime: e[0],
               endTime: e[1],
             };

             // eslint-disable-next-line no-shadow
             const rows = baseRows.filter((d) => {
               if (d.startTime > e[1]) {
                 return false;
               }

               if (d.endTime < e[0]) {
                 return false;
               }

               return true;
             });

             if (tooltips) {
               tooltips.disable();
             }

             drawRows(rows);
             drawPerfTable('perf-table', rows);
           });

      brush.draw();
    }

    drawRows(baseRows);
    drawPerfTable('perf-table', inRows);
  }

  function drawCloneInfo() {
    const dashboardId = arc.globals.ms_dashboard.dashboard_id();
    if (dashboardId) {
      getCloneInfo(dashboardId);
    }
  }

  function getCloneInfo(dashboardId) {
    const $ancestorsListContainer = $('#clone-ancestors-table');
    const $childrenListContainer = $('#clone-children-table');

    $.ajax({
      type: 'GET',
      url: `${arc.frontend.globals.urls.reportCloneInfoUrl}/${dashboardId}`,
    })
      .done((resp) => {
        $ancestorsListContainer.empty();
        $childrenListContainer.empty();
        $ancestorsListContainer.append(cloneChainReducer(resp.ancestorsChain));
        $childrenListContainer.append(cloneChainReducer(resp.immediateChildrenList));
      });
  }

  function cloneChainReducer(relativesList) {
    return relativesList.reduce((acc, rel) => {
      const parentId = rel.parent_report_id || 'Origin';
      return `${acc}<tr><td><a href="/arc/apps/app/${rel.report_id}">${rel.report_id}</a>` +
        `</td><td>${parentId}</td><td>${arc.misc.escapeHTML(rel.report_name)}</td></tr>`;
    }, '');
  }

  function drawPerfCounters() {
    const drawPerfCounter = (tableContainerId, title, rows) => {
      const $tableContainer = $(`#${tableContainerId}`);
      $tableContainer.empty();
      const table = arc.visual.plugins.table();
      const driver = arc.report.driver(table)
        .params(arc.params.tracker());

      driver.user_settings().find_by_id('table_template')
        .value('table-template-6');

      const keyColumn = arc.data.column()
        .colname(title)
        .coltype('string')
        .colproperty(arc.report.columnTypeEnum.DIMENSION)
        .colsource('dimensions_shelf')
        .colindex(0);

      const valueColumn = arc.data.column()
        .colname('')
        .coltype('string')
        .colproperty(arc.report.columnTypeEnum.DIMENSION)
        .colsource('dimensions_shelf')
        .colindex(1);

      const columns = [keyColumn, valueColumn];

      const result = arc.data.result();
      result.columns(columns);
      result.rows(rows);
      result.finalize();

      table.viz_setDataResult(result);

      table.setConfiguration({
        cellsHtml: true,
        disableProfile: true,
        showTooltip: false,
        ordering: false
      });
      table.viz_drawVisual(tableContainerId);
    };

    const leftId = 'perf-counters-left';
    const rightId = 'perf-counters-right';

    const avgRender = arc.counters.data.viz_drawVisual ?
      (arc.counters.data.renderTime / arc.counters.data.viz_drawVisual) : 0;

    const renderPerfomanceInformation = [
      ['Average Render Time', `${(avgRender / 1000).toFixed(3)}s`],
      ['Total Render Time', `${arc.counters.data.renderTime / 1000}s`],
      ['Render Count', arc.counters.data.viz_drawVisual],
      ['Max Render Time', `${arc.counters.data.maxRenderTime / 1000}s`],
    ];

    const avgQuery = arc.counters.data.queryCount ?
      (arc.counters.data.queryTime / arc.counters.data.queryCount) : 0;

    const queryPerformanceInformation = [
      ['Average Query Time', `${(avgQuery / 1000).toFixed(3)}s`],
      ['Total Query Time', `${arc.counters.data.queryTime / 1000}s`],
      ['Query Count', arc.counters.data.queryCount],
      ['Max Query Time', `${arc.counters.data.maxQueryTime / 1000}s`],
      ['Total Result Rows', arc.counters.data.dataResultTotalRows],
    ];

    drawPerfCounter(leftId, 'Rendering Engine', renderPerfomanceInformation);
    drawPerfCounter(rightId, 'Query Engine', queryPerformanceInformation);
  }

  function drawPerfTable(divId, rows) {
    $(`#${divId}`).empty();
    const table = arc.visual.plugins.table();
    const driver = arc.report.driver(table)
                      .params(arc.params.tracker());

    const settings = driver.user_settings();
    const perfSettingsValues = {
      enable_pagination: 'false',
      table_template: 'table-template-6',
    };

    $.each(perfSettingsValues, (k, v) => {
      settings.find_by_id(k)
              .value(v);
    });

    const cols = [];
    profileColnames.forEach((colname, coli) => {
      cols.push(arc.data.column()
                .colname(colname)
                .coltype(profileColtypes[coli])
                .colproperty(arc.report.columnTypeEnum.DIMENSION)
                .colsource('dimensions_shelf')
                .colindex(coli));
    });

    const result = arc.data.result();
    result.columns(cols);
    result.rows(rows);
    result.finalize();

    // create the visual by setting the data and then drawing it
    table.viz_setDataResult(result);
    table.setConfiguration({
      cellsHtml: true,
      disableProfile: true,
      showTooltip: false,
      enablePagination: true
    });

     table.viz_drawVisual(`${divId}`);
  }

  function showProfiles() {
    const rows = profileRows({
      forTable: true,
    });

    drawPerfCounters();

    drawPerfViz('perf-viz', rows);
  }

  tracker.modal = function(initialTab) {
    const $modal = $('#perf-modal');
    arc.misc.handleAlertContainerActivationOnModal($modal);
    if (arc.misc.modalShown($modal)) {
      return;
    }

    $('#perf-viz').empty();
    $('#perf-counters').empty();

    $('#perf-modal .nav-tabs a').offOn('click.nav', function(e) {
      e.preventDefault();
      $(this).tab('show');
    });

    const $queriesTable = $('#perf-queries-table');
    $queriesTable.empty();
    $queriesTable.css('height', `${Math.max($(window).height() - 300, 300)}px`);

    const $usageTable = $('#perf-usage-table');
    $usageTable.empty();
    $usageTable.css('height', `${Math.max($(window).height() - 300, 300)}px`);

    const $table = $('#perf-table');
    $table.empty();
    $table.css('height', `${Math.max($(window).height() - 550, 300)}px`);

    $modal.modal('show', {
      keyboard: true,
    });

    $modal.off('shown.bs.modal')
          .on('shown.bs.modal', () => {
      // Activate the initial tab
      initialTab = initialTab || 'nav-tab-profile';
      $(`#perf-modal .${initialTab}`).tab('show');

      if (initialTab === 'nav-tab-profile') {
        showProfiles();
      }
    });

    if (!arc.globals.dashboard) {
      // Don't show the layout and historical perf tab for
      // pages that don't have a dashboard
      $('#perf-modal').find('.perf-contents').find('.nav-item').each((_idx, li) => {
        if (!li.querySelector('[href="#perf-tab-profiles"]') &&
          !li.querySelector('[href="#perf-tab-queries"]') &&
          !li.querySelector('[href="#perf-tab-counters"]')) {
            li.hidden = true;
        }
      });
    }

    $('#perf-modal ul.nav-tabs a')
        .offOn('shown.bs.tab', function() {
          if ($(this).hasClass('nav-tab-profile')) {
            showProfiles();
            return;
          }

          if ($(this).hasClass('nav-tab-usage')) {
            arc.frontend.legacyBridge.events.next({
              type: 'open-perf-usage-table',
              payload: {
                dashboardId: arc.globals.dashboard.dashboard_id()
              }
            });

            return;
          }

          if ($(this).hasClass('nav-tab-queries')) {
            showQueries('perf-queries-table');
            return;
          }

          if ($(this).hasClass('nav-tab-layout')) {
            showLayout();
            return;
          }

          if ($(this).hasClass('nav-tab-clones')) {
            drawCloneInfo();
          }

          if ($(this).hasClass('nav-tab-current-load')) {
            if (currentLoad === undefined) {
              currentLoad = arc.perf.currentLoad();
            }

            currentLoad.start();
          }
        });
  };

  $('.data-section ul.nav-tabs a')
    .off('shown.bs.tab')
    .on('shown.bs.tab', function() {
      if ($(this).hasClass('nav-tab-queries')) {
        $('#perf-queries-table-data').empty();
        const report_id = arc.report.getPrimaryReport().report_id();
        if (report_id) {
          showQueries('perf-queries-table-data', report_id);
        }

        return;
      }
    });

  function showLayout() {
    const l = arc.visual.dashboardResizer()
                 .dashboard(arc.globals.dashboard)
                 .modalDivId('perf-layout-display')
                 .forResize(false)
                 .paintLayout();

    const c = l.colors();
    const datasetIds = c.domain();
    const datasetColors = c.range();
    const $d = $('#perf-layout-display .perf-datasets');
    $d.empty();

    const datasetIdToName = datasetIdToNameMapping();

    if (datasetIds.length > 0) {
      $d.append('<h5>Datasets</h5>');
    }

    const reportToDataset = reportDatasetMapping();
    const datasetCounts = {};
    Object.keys(reportToDataset).forEach((k) => {
      datasetCounts[reportToDataset[k]] = (datasetCounts[reportToDataset[k]] || 0) + 1;
    });

    const data = datasetIds.map((datasetId, idx) => [
      datasetId, datasetColors[idx % datasetColors.length], datasetCounts[datasetId] || 0]);

    data.sort((a1, a2) => {
      const v1 = a1[2];
      const v2 = a2[2];

      // eslint-disable-next-line no-nested-ternary
      return (v1 > v2) ? -1 : ((v1 < v2) ? 1 : 0);
    });

    // eslint-disable-next-line no-undef
    const domain = d4.extent(data.map((d) => d[2]));
    const range = (domain[0] === domain[1]) ? [100, 100] : [40, 100];

    // eslint-disable-next-line no-undef
    const wScale = d4.scaleLinear()
                     .domain(domain)
                     .range(range);

    data.forEach((d) => {
      const bgColor = d[1];
      const fgColor = arc.colorpalette.colorForBrightness(bgColor, 1);
      const datasetId = d[0];

      // eslint-disable-next-line no-use-before-define
      $d.append(arc.misc.safe_substitute(datasetDisplayHtml, {
        FG_COLOR: fgColor,
        BG_COLOR: bgColor,
        DATASET_URL: arc.misc.getBlankTargetUrl(arc.frontend.globals.urls.datasetEditUrl),
        DATASET_ID: datasetId,
        DATASET_NAME: arc.misc.escapeHTML(datasetIdToName[datasetId]) || '',
        WIDTH: wScale(d[2]),
      }));
    });
  }

  function datasetIdToNameMapping() {
    const datasetIdToName = {};
    if (!arc.globals.dashboard) {
      return datasetIdToName;
    }

    const reports = arc.globals.dashboard.reports().getAllReports();
    reports.forEach((report) => {
      const dataset = report.dataset();
      if (dataset) {
        datasetIdToName[dataset.dataset_id()] = dataset.dataset_name();
      }
    });

    return datasetIdToName;
  }

  function datasetIdToConnectionMapping() {
    const datasetIdToConnection = {};
    if (!arc.globals.dashboard) {
      return datasetIdToConnection;
    }

    const reports = arc.globals.dashboard.reports().getAllReports();
    reports.forEach((report) => {
      const dataset = report.dataset();
      if (dataset) {
        datasetIdToConnection[dataset.dataset_id()] = {
          type: dataset.dataconnection_type_display(),
          id: dataset.dataconnection_id(),
        };
      }
    });

    return datasetIdToConnection;
  }

  function reportDatasetMapping() {
    const reportToDataset = {};
    if (!arc.globals.dashboard) {
      return reportToDataset;
    }

    const reports = arc.globals.dashboard.reports().getAllReports();
    reports.forEach((report) => {
      reportToDataset[report.report_id()] = report.dataset_id();
    });

    return reportToDataset;
  }

  function initializeTable(queryTableId) {
    const tableSelector = `#${queryTableId}`;
    $(tableSelector).empty();

    const table = arc.visual.plugins.table();
    const driver = arc.report.driver(table).params(arc.params.tracker());

    driver.user_settings().find_by_id('table_template').value('table-template-6');

    return { table, driver };
  }

  function createColumns() {
    const colnames = ['Time', 'Dataset', 'Visual', 'Connection', 'Type', 'Runtime', 'Query'];
    const coltypes = ['string', 'string', 'string', 'string', 'string', 'integer', 'string'];

    return colnames.map((colname, idx) => arc.data.column()
      .colname(colname)
      .coltype(coltypes[idx])
      .colproperty(arc.report.columnTypeEnum.DIMENSION)
      .colsource('dimensions_shelf')
      .colindex(idx));
  }

  function createResult(rows) {
    const cols = createColumns();

    const result = arc.data.result();
    result.columns(cols);
    result.rows(rows);
    result.finalize();

    return result;
  }

  function buildRows(rows, downloadRows, profile, datasetId, visualId, runtime, query) {
    function createLink(href, id) {
      return `<a href="${href}" target="_blank">${id}</a>`;
    }

    function escapeHTML(text) {
      return arc.misc.escapeHTML(text) || '';
    }

    const htmlWrappedQuery =
      `<span>
        <a data-toggle="popover" class="query-sql-link">Show SQL</a>
        <span class="query-sql-detail arc-invisible">${escapeHTML(query)}</span>
      </span>`;

    const dataconnectionInfo = datasetIdToConnectionMapping()[datasetId] || {};
    const dataConnectionLink = createLink(`${arc.frontend.globals.urls.dataUrl}?dataconnection_id=${dataconnectionInfo.id}`, dataconnectionInfo.type);
    const datasetLink = createLink(`${arc.frontend.globals.urls.datasetEditUrl}/${datasetId}`, datasetId);
    const visualLink = createLink(`${arc.frontend.globals.urls.viewDashboardUrl}/${visualId}`, visualId);

    rows.push(
      [
        profile.profileTime,
        datasetLink,
        visualLink,
        dataConnectionLink,
        dataconnectionInfo.type,
        runtime,
        htmlWrappedQuery
      ]
    );
    downloadRows.push(
      [
        profile.profileTime,
        datasetId,
        visualId,
        dataconnectionInfo.id,
        dataconnectionInfo.type,
        runtime,
        query
      ]
    );
  }

  function processProfileRows(p, rows, downloadRows, reportToDataset, queryRuntime) {
    const visualId = p.id;
    const datasetId = reportToDataset[visualId] || '';

    const query = getProfileValue(p.profile, 'query');

    buildRows(
        rows,
        downloadRows,
        p,
        datasetId,
        visualId,
        queryRuntime,
        query
    );
  }

  function getProfileValue(profileList, valueKey) {
    const profileLine = profileList.find((p) => p[0] === valueKey);
    if (profileLine) {
      return profileLine[1];
    }
  }

  function getQueryRuntime(p, statusProfilesForQuery) {
    const queryRuntime = getProfileValue(p.profile, 'query_runtime');
    if (queryRuntime !== undefined) {
      return queryRuntime;
    }

    const queryTracker = getProfileValue(p.profile, 'querytracker');
    const status = statusProfilesForQuery.find(
      (statusProfile) => statusProfile.profile.some(
        (sp) => sp[0] === 'querytracker' && sp[1] === queryTracker
      )
    );

    if (status) {
      return getProfileValue(status.profile, 'query_runtime');
    }
  }

  function showQueries(queryTableId, visualId) {
    const { table, driver } = initializeTable(queryTableId);

    const reportToDataset = reportDatasetMapping();
    const rows = [];
    const downloadRows = [];
    const statusProfilesForQuery = profiles.filter((p) => {
      if (visualId && p.id !== visualId) {
        return false;
      }

      if (p.type === 'Status' &&
          p.profile &&
          p.profile.some((e) => e[0] === 'query') &&
          p.profile.some((e) => e[0] === 'querytracker') &&
          p.profile.some((e) => e[0] === 'query_runtime')
      ) {
        return true;
      }

      return false;
    });

    profiles.forEach((p) => {
      if (visualId && p.id !== visualId) {
        return;
      }

      if (p.type !== 'Query' || !p.profile) {
        return;
      }

      const queryRuntime = getQueryRuntime(p, statusProfilesForQuery);
      processProfileRows(p, rows, downloadRows, reportToDataset, queryRuntime);
    });

    rows.reverse();
    downloadRows.reverse();

    const result = createResult(rows);

    table.viz_setDataResult(result);

    driver.sb_colnames(['Time', 'Dataset', 'Visual', 'Connection', 'Runtime', 'Query']);

    driver.sb_rows(downloadRows);

    table.setConfiguration({
      cellsHtml: true,
      disableProfile: true,
      showTooltip: false,
      fullWidth: true,
      drawCallback: () => {
        arc.misc.setupQueryPopover('.query-sql-link', '.query-sql-detail');
      },
    });

    table.viz_drawVisual(queryTableId);

    $('button.queries-download').offOn('click.download', () => {
      table.viz_lib().exportCSV({});
    });
  }

  arc.perf.showQueries = showQueries;

  function counterRows() {
    const rows = [];

    $.each(arc.counters.data, (k, v) => {
      rows.push([
        k,
        v,
      ]);
    });

    return rows;
  }

  // Note that this may be called multiple times in the dashboard-edit page,
  // so make sure any changes here are designed to be called multiple times
  tracker.setup = function() {
    if (!arc.globals.page_profile_enabled) {
      return tracker;
    }

    if (!arc.globals.is_edit && !arc.globals.page_profile_appview_enabled) {
      return tracker;
    }

    arc.shortcut.remove('Ctrl+Shift+G');
    arc.shortcut.add('Ctrl+Shift+G', () => tracker.modal());

    // Setup 'Download' button within perf modal
    $('#perf-modal .perf-download')
        .offOn('click.download', () => {
          arc.misc.downloadCsv('profile.txt', profileColnames, profileRows(), profileColtypes);
        });

    $('#perf-modal .counter-download')
        .offOn('click.download', () => {
          arc.misc.downloadCsv('counters.txt', counterColnames, counterRows(), counterColtypes);
        });

    return tracker;
  };

  return tracker;
};

const datasetDisplayHtml = `
  <a target="_blank" href="%DATASET_URL%/%DATASET_ID%"
     data-toggle="tooltip" data-placement="left" title="%DATASET_NAME%">
    <div class="perf-dataset-display"
         style="color:%FG_COLOR%; background-color:%BG_COLOR%; width: %WIDTH%px">
      %DATASET_ID%
    </div>
  </a>
`;

arc.perf.PAGEPERF = arc.perf.tracker();
arc.perf.PAGEPERF.addProfile({
  type: 'startup',
  time: 0,
});
}());
