window.arc = window.arc || {};

(function() {
'use strict';

arc.cviz = arc.cviz || {};

// Similar regexp used in trigger cviz_rest_data
const cvizRestRE = /\bcviz_rest\((.*)?\)\s*$/i;

arc.cviz.isCvizExpr = function(expr) {
  if (!expr) {
    return false;
  }

  const m = expr.match(cvizRestRE);
  return ((m !== undefined) && (m !== null));
};

arc.cviz.cvizSettingsFromExpr = function(expr) {
  const m = expr.match(cvizRestRE);
  if (!m) {
    return undefined;
  }

  let s = m[1];
  if (s[0] === "'") {
    if (s[s.length - 1] === "'") {
      s = s.substring(1, s.length - 1);
    } else {
      throw "Malformed remote data settings: " + expr;
    }
  } else if (s[0] === '"') {
    if (s[s.length - 1] === '"') {
      s = s.substring(1, s.length - 1);
    } else {
      throw "Malformed remote data settings: " + expr;
    }
  } else {
    throw "Malformed remote data settings: " + expr;
  }

  return JSON.parse(s);
};

arc.cviz.remoteCaller = function() {
  const r = function() { return 'arc.cviz.remoteCaller'; };

  let dataset,
      viz_lib,
      results;

  const remoteCallCols = [];
  const remoteCallColIndices = [];
  const remoteCallSettings = [];

  r.dataset = function(_) {
    if (!arguments.length) return dataset;
    dataset = _;
    return r;
  };

  r.viz_lib = function(_) {
    if (!arguments.length) return viz_lib;
    viz_lib = _;
    return r;
  };

  r.results = function(_) {
    if (!arguments.length) return results;
    results = _;
    return r;
  };

  r.fetchRemoteData = function(dsreq_obj, completedCb) {
    trackRemoteData(dsreq_obj);

    if (remoteCallCols.length > 0) {
      fetchRemoteDataWork(results, completedCb);
      return;
    }

    completedCb();
  };

  // Similar code is used for trigger remoteCallSettings
  function cvizRestSettingsToRestCall(s) {
    /*
     * See description of the expected call syntax in
     *   arcweb/apps/views_remotecalls.py
     *
     * The URL key is where the REST api endpoint is
     * payload.accessKey is a special param for CML REST models
     * payload.request is actually sent as data to the model
     */
    const r = {
      version: "2.0",
      response_colname: s.response_colname,
      url: s.url,
      payload: {
        request: {
          data: {},
        },
      },
    };

    if (s.config !== undefined) {
      r.payload.request.config = s.config;
    }

    if (s.accessKey !== undefined) {
      r.payload.accessKey = s.accessKey;
    }

    if (s.colnames !== undefined) {
      r.payload.request.data.colnames = s.colnames;
    }

    if (arc.globals.ms_dashboard) {
      const viz_params = arc.globals.ms_dashboard.params_history().current_params()[0];
      r.payload.request.params = {};
      viz_params.forEach(function(p) {
        r.payload.request.params[p.key] = p.value;
      });
    }

    return r;
  }

  function trackRemoteData(dsreq_obj) {
    if (!dsreq_obj) {
      return;
    }

    dsreq_obj.dimensions.forEach(function(dim, dimIdx) {
      if (dim.expr === undefined) {
        return;
      }

      const d = arc.grammar._colnames.parse(dim.expr);
      d.cols.forEach(function(col) {
        const dscol = dataset.findColumn(col);
        if (dscol && dscol.isRemoteColumn()) {
          remoteCallCols.push(dscol);
          remoteCallColIndices.push(dimIdx);
          const s = dscol.remoteDataSettings();
          remoteCallSettings.push(cvizRestSettingsToRestCall(s));
        }
      });
    });

    dsreq_obj.aggregates.forEach(function(agg, aggIdx) {
      if (agg.expr === undefined) {
        return;
      }

      const d = arc.grammar._colnames.parse(agg.expr);
      d.cols.forEach(function(col) {
        const dscol = dataset.findColumn(col);
        if (dscol && dscol.isRemoteColumn()) {
          remoteCallCols.push(dscol);
          remoteCallColIndices.push(dsreq_obj.dimensions.length + aggIdx);
          const s = dscol.remoteDataSettings();
          remoteCallSettings.push(cvizRestSettingsToRestCall(s));
        }
      });
    });
  }

  // Similar code is used for trigger remoteCallSettings
  function fetchRemoteDataWork(results, cb) {
    remoteCallSettings.forEach(function(s) {
      s.colIndices = [];
      s.payload.request.data.coltypes = [];
      s.payload.request.data.colnames.forEach(function(colname) {
        const i = results.colnames.indexOf(colname);
        if (i === -1) {
          throw 'Could not fetch remote data. Did not find column: ' + colname;
        }

        s.payload.request.data.coltypes.push(results.coltypes[i]);
        s.colIndices.push(i);
      });
    });

    remoteCallSettings.forEach(function(s) {
      s.payload.request.data.rows = [];
      results.rows.forEach(function(row) {
        s.payload.request.data.rows.push(s.colIndices.map((i) => row[i]));
      });
    });

    $.ajax({
      url: arc.frontend.globals.urls.remotecallsUrl,
      data: {
        "data": JSON.stringify(remoteCallSettings),
      },
      type: 'POST',
      beforeSend: function(xhr) {
        xhr.setRequestHeader("X-CSRFToken", arc.globals.csrf_token);
      }
    })
    .fail(function(rsp, textstatus, errorthrown) {
      console.log(rsp, textstatus, errorthrown);
      remoteCallSettings.forEach(function(s, idx) {
        const colIdx = remoteCallColIndices[idx];
        results.rows.some(function(row, rowIdx) {
          row[colIdx] = 'Error fetching data';
        });
      });

      cb();
    })
    .done(function(rsp) {
      remoteCallSettings.forEach(function(s, idx) {
        const remoteRows = rsp[idx];
        if (!remoteRows) {
          viz_lib.alertMessage("Error fetching remote data");
          return;
        }

        const colIdx = remoteCallColIndices[idx];
        results.rows.some(function(row, rowIdx) {
          if (remoteRows[rowIdx] === undefined) {
            viz_lib.alertMessage("Error fetching remote data");
            return true;
          }

          row[colIdx] = remoteRows[rowIdx];
        });
      });

      cb();
    });
  }

  return r;
};

}());
