window.arc = window.arc || {};

(function() {
'use strict';

arc.sqlrun = arc.sqlrun || {};

/*
 * We are called from 2 places:
 * In one, response.response.rows is a string that has to be JSONified
 * In the other code path, response.rows is a string that has to be JSONified
 *
 * In both cases, we want to build for a future where the actual data has to be
 *   fetched from the server when we are here vs simply calling JSON.parse on
 *   the response - so, build the interface with a cb mechanism and insist on
 *   the cb being passed in.
 *
 * Additionally, this code is a placeholder for real streaming results.
 *
 * The idea is that instead of returning result rows directly from a json
 * response, what we should do is to return a handle to a cursor. The client
 * can then ask for results and we can instantiate a StreamingHttpResponse
 * on the server side to return results in a chunked manner. We never have to
 * hold the memory necessary to hold the entire resultset in memory on the
 * client side.
 *
 * We will have to do this via a dedicated threadpool so we never block this
 * behind new queries since we know the results are already available in this
 * case.
 */
arc.sqlrun.fetchDataRows = function(response, cb) {
  if (response.response && response.response.rows) {
    if (typeof (response.response.rows) === typeof ('')) {
      response.response.rows = JSON.parse(response.response.rows);
    }
  } else if (response.rows) {
    if (typeof (response.rows) === typeof ('')) {
      response.rows = JSON.parse(response.rows);
    }
  }

  cb(response);
};

arc.sqlrun.runSQLWithParams = function(params, successCb, errorCb) {
  function cb(errorMsg, response) {
    if (errorMsg) {
      errorCb(errorMsg);
      return;
    }

    successCb(response);
  }

  return arc.sqlrun.runSQLWithParamsParallel(
    [params],
    [cb]);
};

/*
 * All (most :) access to the backend asking for data come through here.
 * (there are some exceptions for csv/xls download and for acceleration)
 *
 * We use async http interfaces to send the request over and get results back.
 * See additional notes in arc.sqlrun.httpAsyncCall()
 */
arc.sqlrun.runSQLWithParamsParallel = function(paramsl, cbl, finallyCb) {
  const formattedParams = paramsl.map(function(p) {
    return {
      params: p,
    };
  });

  let dtos = JSON.stringify(formattedParams);
  if (!arc.globals.arcviz_settings.ctds) {
    dtos = arc.misc.obfuscate(dtos);
  }

  let querytracker = window.crypto.getRandomValues(new Uint32Array(1))[0];
  let dashboardId = 0;
  if (arc && arc.globals && arc.globals.dashboard) {
    dashboardId = arc.globals.dashboard.dashboard_id();
    if (dashboardId === undefined) {
      dashboardId = 0;
    }
  }

  querytracker = `${dashboardId}-${querytracker}`;
  const url = `${arc.frontend.globals.urls.sqlParallelAsyncUrl}/${querytracker}`;

  const paramsToSend = {
    querytracker,
    data: dtos,
  };

  const startTime = new Date();

  function doneCb(responsel) {
    const t2 = new Date();
    if (finallyCb) {
      finallyCb();
    }

    if (!arc.globals.arcviz_settings.ctds) {
      responsel = JSON.parse(arc.misc.unobfuscate(responsel));
    }

    if (!Array.isArray(responsel)) {
      const response = typeof responsel === 'object' ? JSON.stringify(responsel) : responsel;
      // eslint-disable-next-line no-use-before-define
      failCb(`The result is not in the expected (array) format: ${response}`);
      return;
    }

    responsel.forEach(function(r, idx) {
      var p = paramsl[idx];
      var error = r.error;

      if (error) {
        cbl[idx](error);

        arc.perf.PAGEPERF.addProfile({
          id: p.report_id,
          type: 'Query',
          time: t2 - startTime,
          error: r.error,
          profile: r.profile,
        });

        return;
      }

      arc.sqlrun.fetchDataRows(r, function(_r) {
        const t2 = new Date();
        cbl[idx](undefined, _r);
        arc.perf.PAGEPERF.addProfile({
          id: p.report_id,
          type: 'Query',
          time: t2 - startTime,
          error: r.error,
          profile: r.profile,
        });
      });
    });
  }

  function failCb(req) {
    const t2 = new Date();
    if (finallyCb) {
      finallyCb();
    }

    const errorMsg = req.responseText ? req.responseText : req;
    paramsl.forEach(function(p) {
      arc.perf.PAGEPERF.addProfile({
        id: p.report_id,
        type: 'Query',
        time: t2 - startTime,
        error: errorMsg,
      });
    });

    cbl.forEach(function(cb) {
      cb(errorMsg);
    });
  }

  arc.sqlrun.httpAsyncCall({
    url,
    paramsToSend,
    doneCb,
    failCb,
  });
};

/*
 * http async calls.
 *
 * The corresponding implementation on the server side here is in util/http_async.py and
 * sqlrun/views_async.py. Look at that to understand how the backend works
 *
 * What we do is this:
 * - First, we sent out a request to a backend interface that supports the http_async protocol
 * - Next, we look at the response to see whether it's a response that says come-back-for-more
 *   or whether it's our results
 * - If it's the results, we are done
 * - If it's come back for more, we wait for a bit and then go back with the request id.
 *   We use a backoff algorithm that maxes out at going back for results every 5 seconds
 *
 */
arc.sqlrun.httpAsyncCall = function(_) {
  const {
    url,
    paramsToSend,
    doneCb,
    failCb,
    contentType,
  } = _;

  const backoff = [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 1, 1, 1, 1, 2, 2, 2, 2, 5];
  let backoffIdx = 0;

  function doneMaybeRetryCb(response) {
    if (response.type !== 'async') {
      doneCb(response);
      return;
    }

    if (response.error) {
      failCb(response.error);
      return;
    }

    if (backoffIdx < backoff.length - 1) {
      ++backoffIdx;
    }

    const asyncId = response.id;
    const params = {
      async: JSON.stringify({
        type: 'async',
        id: asyncId,
      })
    };

    setTimeout(() => {
      $.ajax({
        url,
        data: params,
        type: 'POST',
        beforeSend: function (xhr) {
          xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
        }
      })
      .done(doneMaybeRetryCb)
      .fail(function(req) {
        const errorMsg = req.responseText ? req.responseText : req;
        failCb(errorMsg);
      });
    }, backoff[backoffIdx] * 1000);
  }

  const ajaxParams = {
    url,
    data: paramsToSend,
    type: 'POST',
    beforeSend: function (xhr) {
      xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
    }
  };
  if (contentType) {
    ajaxParams.contentType = contentType;
  }

  $.ajax(ajaxParams)
  .done(doneMaybeRetryCb)
  .fail((xhr) => {
    // If there is a problem with the network(readyState==0)
    // there will be no responseText and displaying the XHR as object not meaningful
    const networkOrUnknownError = (xhr.readyState === 0 ? 'Network error' : xhr);
    const errorMsg = xhr.responseText ?  xhr.responseText : networkOrUnknownError;
    failCb(errorMsg);
  });
};

}());
