window.arc = window.arc || {};

(function() {
'use strict';

/*

---
TODO:

  Should have cancel refresh option

  Really need refresh progress display

  Should have 'Strength of recommendation' or 'Confidence' number
  for each recommendation (including if base is not on parquet)

  App-Scope and Dataset-Scope parameters

  Handle parameter pickers

  Size info into view display?

---

  App acceleration mananager

  Accepts a list of app IDs to be processed -- could come from
  dataset or app group.

  Note that it is individual visuals (and picklist/datepickers)
  that are accelerated.

  Dashboard apps show the status of the individual components
  and overall summary of components.  Acceleration for the dashboard
  is done by accelerating the components.

  The flow for acceleration has the following steps:

    1) Get the execution history and recommendation status for each
       of the input apps.

    2) Select specific apps to get the detailed recommendations for.

    3) From the detailed recommendations display, one can create/refresh/
       delete specific analytical views or operate on the entire set by
       using the 'Create all' and 'Refresh all' buttons.

  Note that this flow results in getting recommendations twice but this
  seems ok as the recommendations can change based on the set of input apps.

*/

arc.admin = arc.admin || {};

/* Common routine to get timeleft text from refresh status response */
arc.admin.refreshTimeLeftText = function(reftask, fetchtime) {
  let text = 'Unknown';
  if (reftask.cancelled) {
    text = 'Cancelling...';
  } else if (reftask.progress > 0) {
    const remain = Math.max(fetchtime || 1, reftask.remain);
    if (remain <= 90) {
      text = `About ${remain} second${remain > 1 ? 's' : ''} left`;
    } else if (remain <= (90 * 60)) {
      text = `About ${Math.round(remain / 60)} minutes left`;
    } else {
      const hours = Math.floor(remain / 3600);
      const mins = Math.round((remain % 3600) / 60);
      text =
        `About ${hours} hour${hours > 1 ? 's' : ''} and ${mins} minute${mins > 1 ? 's' : ''} left`;
    }
  } else {
    text = 'Waiting to start...';
  }

  return text;
};

/* Routine to return detailed scoring data for an aview */
arc.admin.computeAviewScore = function(aview) {
  if (aview.error_message) {
    return { overallScore: 0 };
  }

  // Return a score array, where each item in the array is the score deduction
  // for each rule. The default reduction is 0 for each rule.
  const deductions = {
      lowDataReduction: 0,
      notPartitioned: 0,
      hasExpensiveAggs: 0,
      isLargeAndExpensive: 0,
  };

  const basetables = aview.basetables || [];
  const dimension_cardinalities = aview.dimension_cardinalities || [];
  const dimensions = aview.dimensions || [];
  const incrementallyRefreshable = aview.incrementally_refreshable || '';
  const hasExpensiveAggregates = (aview.expensive_aggregates || []).length > 0;
  const expensiveAggregates = arc.misc.makeEnglishList(aview.expensive_aggregates || []);
  const nonLVBasetables = basetables.filter((b) => !b.is_an_lview);
  const hasJoinMaterialization = nonLVBasetables.length > 1;
  const hasFileFormatChange = nonLVBasetables.some((t) => t.file_format !== 'PARQUET');
  let maxDimCardinality = 0;
  let maxDimName;
  dimension_cardinalities.forEach((dim, i) => {
    if ($.isNumeric(dim) && Number(dim) > maxDimCardinality) {
      maxDimCardinality = Number(dim);
      maxDimName = dimensions[i];
    }
  });
  const maxBaseTableRowCount =
    Math.max(...basetables.map((d) => ($.isNumeric(d.row_count) ? Number(d.row_count) : -1)));

  // If don't have actual row count because AV doesn't exist yet, estimate based on dim cardinality
  const avRowCount = aview.row_count > 0 ? aview.row_count : maxDimCardinality;
  const lowDataReduction = (avRowCount > 0 &&
    maxBaseTableRowCount > 0 &&
    (avRowCount / maxBaseTableRowCount) > 0.4);

  const baseTableIsLarge = maxBaseTableRowCount > 10 ** 7;

  if (lowDataReduction && !hasFileFormatChange && !hasJoinMaterialization) {
    deductions.lowDataReduction -= 1;
  }

  if (!incrementallyRefreshable && baseTableIsLarge) {
    deductions.notPartitioned -= 1;
  }

  if (hasExpensiveAggregates) {
    if (lowDataReduction) {
      deductions.hasExpensiveAggs -= 1;
    }

    if (!incrementallyRefreshable && baseTableIsLarge) {
      deductions.isLargeAndExpensive -= 1;
    }
  }

  const defaultScore = 5;

  // Consider any score reductions from the deductions
  const overallScore = defaultScore + Object.values(deductions).reduce((a, b) => (a + b));

  return {
    overallScore,
    deductions,
    incrementallyRefreshable,
    dataReduction: !lowDataReduction,
    baseTableIsLarge,
    maxBaseTableRowCount,
    maxDimCardinality,
    maxDimName,
    hasExpensiveAggregates,
    expensiveAggregates,
    hasJoinMaterialization,
    hasFileFormatChange
  };
};

/* Method to generate five stars with appropriate number filled in based on passed score */
arc.admin.generateStars = function(rating) {
  let html = '<div style="padding-bottom: 5px">';
  for (let i = 0; i < rating; i++) {
    html += '<i class="fa fa-star fa-1 arc-rating"></i>';
  }

  for (let i = rating; i < 5; i++) {
    html += '<i class="fa fa-star fa-star-0 fa-1 arc-rating"></i>';
  }

  html += '</div>';

  return html;
};

/* Method to generate html for next aview refresh */
arc.admin.generateRefreshString = function(aview) {
  let refreshString = '';
  const nextRefresh = aview.nextScheduledRefresh;
  if (nextRefresh) {
    const dateString = moment(nextRefresh).format('MMM D YYYY h:mm A');
    refreshString += `<span title="${dateString}">`;
    refreshString +=
      `<i class="fa fa-calendar"></i> ${arc.misc.friendlyDateSince(nextRefresh)}</span>`;
  }

  return refreshString;
};

/* Method to populate detail modal for an analytical view */
arc.admin.populateAVDetailModal = ($row, selected, accelContext = false, rawQueries = false) => {
  const $modal = $('#detail-accel-modal');
  $modal.find('.modal-title').text(accelContext ? 'Recommendation Details' : 'AV Details');
  let viewdef = $row.data('viewdef');
  const values = $row.data('substitutions');
  const scoreData = $row.data('score');
  const score = scoreData.overallScore;
  const tableTemplate = $('#single-table-info-template').html();
  const $baseTableContainer = $('#aview-tab-basetable').find('.arc-form-horizontal');
  const $aviewTableContainer = $('#aview-tab-aviewtable').find('.arc-form-horizontal');
  const $appsContainer = $('#aview-tab-app').find('.arc-form-horizontal');
  const $vizTab = $('#aview-viz-sql-tab');
  $('#aview-detail-name').html(viewdef.name);
  $('#aview-detail-dim').html(values.HTML_DIMENSIONS);
  $('#aview-detail-agg').html(values.MEASURES);
  $('#aview-detail-sql').html(values.SQL);
  $('.aview-detail-rating').text(score);
  $('.aview-score-total span').text(score);
  $('div.perfect-score').toggleClass('arc-invisible', score !== 5);
  $('#aview-detail-star-container').html(arc.admin.generateStars(score));

  $baseTableContainer.empty();
  $appsContainer.empty();

  const uncomputedBaseTables = [];
  const uncomputedLVs = [];
  let $computeStatsBtns;
  const warning_template = $('#missing-stats-warning-template').html();

  function createTableSection(basetable) {
    let uncomputed = false;
    if (basetable.row_count === -1) {
      uncomputed = true;
      if (basetable.is_an_lview) {
        uncomputedLVs.push(basetable.name);
      } else {
        uncomputedBaseTables.push(basetable.name);
      }
    }

    const format = basetable.is_an_lview ? 'N/A for logical view' : basetable.file_format;
    const diskSize = basetable.is_an_lview ? 'N/A for logical view' :
      arc.misc.formatAsFileSize(basetable.size_on_disk);
    const $section = $(arc.misc.safe_substitute(tableTemplate, {
      ID: basetable.name.replace('.', '-'),
      NAME: basetable.name,
      FORMAT: format,
      COUNT: uncomputed ? 'Not computed' : arc.misc.formatReadableNumber(basetable.row_count),
      ROW_SIZE: arc.misc.sizeDisplay(basetable.row_size, 'Not computed'),
      MEMORY_SIZE: arc.misc.sizeDisplay(basetable.in_mem_size, 'Not computed'),
      DISK_SIZE: diskSize,
      PARTITIONS: basetable.partitions.length ? basetable.partitions : 'None'
    }));

    if (uncomputed) {
      $section.find('.arc-invisible').removeClass('arc-invisible');
    }

    return $section;
  }

  const basetables = viewdef.basetables || [];
  const sections = basetables.map((basetable) => createTableSection(basetable));

  $baseTableContainer.append(sections);

  function populateWarnings() {
    $modal.find('.missing-stats-warning').remove();
    const nameList = arc.misc.makeEnglishList(
      [viewdef.name].concat(uncomputedBaseTables, uncomputedLVs)
    );
    let dc_id;
    let dc_type;
    let activeDataconn;
    if (viewdef.dc_id && viewdef.dc_type) {
      dc_id = viewdef.dc_id;
      dc_type = viewdef.dc_type;
    } else {
      activeDataconn = arc.apps_data.get_active_dataconnection();
      dc_id = activeDataconn.id;
      dc_type = activeDataconn.type;
    }

    const $warning = $(arc.misc.safe_substitute(warning_template, { LIST: nameList }))
      .data('tables', uncomputedBaseTables)
      .data('dc_id', dc_id)
      .data('dc_type', dc_type);
    $('#aview-tab-score, #aview-tab-basetable, #aview-tab-aviewtable').prepend($warning);
    $computeStatsBtns = $('.compute-stats');
    $computeStatsBtns.off('click.getstats').on('click.getstats', requestStats);
  }

  if (uncomputedBaseTables.length) {
    populateWarnings();
  } else {
    $modal.find('.missing-stats-warning').remove();
  }

  // eslint-disable-next-line no-shadow
  function populateScoreDetail(scoreData) {
    const scoreValues = scoreData.deductions;
    const advantages = ['hasFileFormatChange',
      'hasJoinMaterialization',
      'incrementallyRefreshable',
      'dataReduction'];
    const expensiveAggs = scoreData.expensiveAggregates;
    const count = arc.misc.formatReadableNumber(scoreData.maxBaseTableRowCount);
    const dimCard = arc.misc.formatReadableNumber(scoreData.maxDimCardinality);
    const dimPct = Math.round(
      (scoreData.maxDimCardinality * 100) / scoreData.maxBaseTableRowCount
    );

    function showAdvantage(advantage) {
      const $targetCell = $(`#${advantage}`);
      if (scoreData[advantage]) {
        $targetCell.find('span').each(function toggleSuccessVisibility() {
          const $span = $(this);
          $span.toggleClass('arc-invisible', !$span.hasClass('success'));
        });
        const $successSpan = $targetCell.find('.success');
        const old_format = viewdef.basetables[0].file_format.toLowerCase();
        switch (advantage) {
          case 'hasFileFormatChange':
            $successSpan.attr('data-content',
              `Format was converted from ${old_format} to parquet.`);
            break;
          case 'hasJoinMaterialization':
            $successSpan.attr('data-content',
              `${viewdef.basetables.filter((t) => !t.is_an_lview).length} base tables were joined.`);
            break;
          case 'dataReduction':
            $successSpan.removeClass('bg-success bg-secondary');
            if (uncomputedBaseTables.length) {
              $successSpan.addClass('bg-secondary').text('TBD');
              $successSpan.attr('data-content',
                'The data reduction impact is unknown because base table stats need to be computed.');
            } else {
              $successSpan.text('Yes');
              $successSpan.addClass('bg-success').attr('data-content',
                `The analytical view has low dimension cardinality (${dimCard})
                relative to the row count of the largest base table (${count})."`);
            }

            break;
          default:
            break;
        }
      } else {
        $targetCell.find('span').each(function toggleFailureVisibility() {
          const $span = $(this);
          $span.toggleClass('arc-invisible', $span.hasClass('success'));
        });
        const $failureSpan = $targetCell.find('.failure');
        if (advantage === 'incrementallyRefreshable') {
          $failureSpan.removeClass('bg-danger bg-secondary');
          if (scoreValues.notPartitioned === 0) {
            if (uncomputedBaseTables.length) {
              $failureSpan.addClass('bg-secondary').attr('data-content',
                'With base table sizes yet to be computed, the impact is uncertain.');
            } else {
              const msg = `With a maximum base table size of ${count} rows,` +
                ` this will not signficantly impact refresh times.`;
              $failureSpan.addClass('bg-secondary').attr('data-content', msg);
            }
          } else {
            $failureSpan.addClass('bg-danger').attr('data-content',
              `With base table sizes of up to ${count} rows, this could lead to slow refreshes.`);
          }
        } else if (advantage === 'dataReduction') {
          $failureSpan.attr('data-content',
            `The analytical view has high dimension cardinality (${dimCard})
            relative to the row count of the largest base table (${count})."`);
        }
      }
    }

    advantages.forEach(showAdvantage);

    Object.keys(scoreValues).forEach(function displaySubtraction(key) {
      const $div = $(`#aview-detail-rule-${key}`);
      if (scoreValues[key]) {
        $div.removeClass('arc-invisible');
        $div.find('span').text('- 1');
        let explanation;
        switch (key) {
          case 'lowDataReduction':
            explanation = `Aview dimension column ${scoreData.maxDimName} has cardinality ${dimCard} which is ${dimPct}% of the base table row ` +
              `count and above the target threshold of 40%. An analytical view with such low ` +
              `data reduction that is not offset by a file format change or by join ` +
              `materialization is less advantageous.`;
            break;
          case 'notPartitioned':
            explanation = `${'This analytical view is less advantageous because it is not ' +
              'incrementally refreshable and is large ('}${count} rows).`;
            break;
          case 'hasExpensiveAggs':
            explanation = `Aview dimension column ${scoreData.maxDimName} has cardinality ${dimCard} which is ${dimPct}% of the base table row ` +
              `count and above the target threshold of 40%. The combination of such low data ` +
              `reduction and the use of expensive aggregates (${expensiveAggs}) makes this ` +
              `analytical view less advantageous.`;
            break;
          case 'isLargeAndExpensive':
            explanation = `The combination of expensive aggregates (${expensiveAggs}) and a large base table (${count} rows) lessens the advantage this analytical view.`;
            break;
          default:
            break;
        }

        $div.find('i').attr('data-content', explanation);
      } else {
        $div.addClass('arc-invisible');
      }
    });
  }

  function requestStats(e) {
    $computeStatsBtns.prop('disabled', true);
    const $div = $(e.target).closest('div');
    const tables = $div.data('tables');
    const dc_id = $div.data('dc_id');
    const dc_type = $div.data('dc_type');
    const $spinner = $('#detail-accel-modal .spinner-container');
    let computeResponseCount = 0;
    let computeSuccessCount = 0;
    let expectedTableStatsResponseCount = tables.length;
    let tableStatsResponseCount = 0;
    let tableStatsSuccessCount = 0;
    let expectedColumnStatsResponseCount = tables.length;
    let columnStatsResponseCount = 0;
    let columnStatsSuccessCount = 0;
    const errMsgs = [];
    const rowCounts = {};
    $spinner.removeClass('arc-invisible');

    function stopProgressIndicators() {
      $computeStatsBtns.prop('disabled', false);
      $spinner.addClass('arc-invisible');
      if (errMsgs.length) {
        arc.report.alertMessageSmall(`Error computing stats: ${errMsgs.join(' ')}`,
          'alert-danger');
      } else {
        $('.missing-stats-warning').addClass('arc-invisible');
      }
    }

    function makeRequestBuilder(table) {
      function fetchTableStatsCb(errorMsg, response) {
        tableStatsResponseCount += 1;
        if (errorMsg) {
          expectedColumnStatsResponseCount -= 1;
          errMsgs.push(`Failed to retrieve table stats on ${table} with message ${errorMsg}`);
        } else {
          tableStatsSuccessCount += 1;
          const $countDiv = $(`#${table.replace('.', '-')} .row-count`);
          const rowCount = response.rows[0][0];
          rowCounts[table] = rowCount;
          $countDiv.find('.value').text(rowCount);
          $countDiv.find('.text-warning').addClass('arc-invisible');
          arc.apps_data.fetchColumnStats({
            cache: false,
            dcid: dc_id,
            dctype: dc_type,
            fullname: table,
            cb: fetchColumnStatsCb,
          });
        }

        const tableStatsDone = tableStatsResponseCount === expectedTableStatsResponseCount;
        if (tableStatsDone && tableStatsSuccessCount === 0) {
          stopProgressIndicators();
        }
      }

      function fetchColumnStatsCb(error, response) {
        columnStatsResponseCount += 1;
        if (error) {
          errMsgs.push(`Failed to retrieve column stats on ${table} with message ${error}`);
        } else {
          columnStatsSuccessCount += 1;
          const $tableDiv = $(`#${table.replace('.', '-')}`);
          const rowSize = response.rows.reduce(
            (acc, currRow) => Number.parseFloat(currRow[5]) + acc, 0
          );
          const $rowSizeDiv = $tableDiv.find('.row-size');
          const $inMemSizeDiv = $tableDiv.find('.mem-size');
          $rowSizeDiv.find('.value').text(arc.misc.sizeDisplay(rowSize));
          $rowSizeDiv.find('.text-warning').addClass('arc-invisible');
          const inMemSize = rowSize * rowCounts[table];
          $inMemSizeDiv.find('.value').text(arc.misc.sizeDisplay(inMemSize));
          $inMemSizeDiv.find('.text-warning').addClass('arc-invisible');

          if (!accelContext) {
            for (let i = 0; i < viewdef.basetables.length; i++) {
              const basetable = viewdef.basetables[i];
              if (basetable.name === table) {
                basetable.row_size = rowSize;
                basetable.in_mem_size = inMemSize;
                basetable.row_count = parseInt(rowCounts[table], 10);
                break;
              }
            }

            // eslint-disable-next-line no-shadow
            const scoreData = arc.admin.computeAviewScore(viewdef);
            const $viewRow = $(`tr[data-arc-aview="${viewdef.name}"]`);
            $viewRow.find('.arc-aview-rating')
              .html(arc.admin.generateStars(scoreData.overallScore));
            $viewRow.data('score', scoreData);
            populateScoreDetail(scoreData);
          }
        }

        if (columnStatsResponseCount === expectedColumnStatsResponseCount) {
          if (columnStatsSuccessCount === 0) {
            stopProgressIndicators();
          } else if (accelContext) {
            const data = { appids: JSON.stringify(Object.keys(selected)) };
            $.ajax({
              url: arc.frontend.globals.urls.viewRecUrl,
              data,
              type: 'POST',
              beforeSend(xhr) {
                xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
              }
            })
              .fail((rsp, textstatus) => {
                errMsgs.push(rsp.responseText || textstatus);
                stopProgressIndicators();
              })
              .done(recommendationCb);
          } else {
            $('#detail-accel-modal').find('.missing-stats-warning').remove();
            const rb = arc.requestBuilder();
            rb.requestType('.compute_stats');
            rb.tableName(viewdef.name);
            const doNothing = () => null;

            arc.report.sql_runner.run(dc_id, '', doNothing, false, rb.dsreq());
          }
        }
      }

      function recommendationCb(rsp) {
        rsp.viewdefs.forEach(updateViewRow);
        viewdef = rsp.viewdefs.find((def) => def.name === viewdef.name);
        const $currRow = $(`#arc-view-id-${viewdef.name.replace('.', '-')}`);
        arc.admin.populateAVDetailModal($currRow, selected, accelContext, rawQueries);
        stopProgressIndicators();
      }

      function updateViewRow(aview) {
        // eslint-disable-next-line no-shadow
        const $row = $(`#arc-view-id-${aview.name.replace('.', '-')}`);
        // eslint-disable-next-line no-shadow
        const scoreData = arc.admin.computeAviewScore(aview);
        $row.find('.arc-aview-rating').html(arc.admin.generateStars(scoreData.overallScore));
        $row.data('viewdef', aview)
          .data('score', scoreData);
      }

      function metadataComputeCb(errorMsg) {
        computeResponseCount += 1;
        if (errorMsg) {
          errMsgs.push(`Failed to compute stats on ${table} failed with message ${errorMsg}`);
          expectedTableStatsResponseCount -= 1;
          expectedColumnStatsResponseCount -= 1;
        } else {
          computeSuccessCount += 1;
          arc.apps_data.fetchTableStats({
            cache: false,
            dcid: dc_id,
            dctype: dc_type,
            fullname: table,
            cb: fetchTableStatsCb,
          });
        }

        if (computeResponseCount === tables.length && computeSuccessCount === 0) {
          stopProgressIndicators();
        }
      }

      const rb = arc.requestBuilder();
      rb.requestType('.compute_stats');
      rb.tableName(table);
      arc.report.sql_runner.run(dc_id, '', metadataComputeCb, false, rb.dsreq());
    }

    // For each base table we need to queue compute stats, and once that completes,
    // we have to actually retrieve the metadata, both at the table and column level.
    // This can only handle materialized base tables -- to get metadata on the aview,
    // then need to run through recommendation engine again.
    // Don't want to display any errors to the user until all these calls have completed, but
    // when counting must keep in mind that one failure on a particular table will prevent the
    // any further calls on that table.
    tables.forEach(makeRequestBuilder);
  }

  function populateAVMetaData() {
    const uncomputed = viewdef.row_count === -1 && viewdef.state !== 'UNUSABLE';
    $aviewTableContainer.empty();
    const placeholder = viewdef.state === 'UNUSABLE' && viewdef.row_count === -1 ?
      'TBD' : 'Not computed';
    let row_count;
    if (uncomputed) {
      row_count = 'Not computed';
    } else if (viewdef.state === 'RECOMMENDED' ||
      (viewdef.state === 'UNUSABLE' && viewdef.row_count !== -1)) {
      row_count = `${arc.misc.formatReadableNumber(viewdef.row_count)} (estimated)`;
    } else if (viewdef.state === 'UNUSABLE' && viewdef.row_count === -1) {
      row_count = 'TBD';
    } else {
      row_count = arc.misc.formatReadableNumber(viewdef.row_count);
    }

    const memSize = arc.misc.sizeDisplay(viewdef.in_mem_size, placeholder,
      viewdef.state === 'RECOMMENDED');
    const diskSize = viewdef.state === 'UNUSABLE' ?
      'TBD' : arc.misc.sizeDisplay(viewdef.size_on_disk);
    const partitionStr = viewdef.partitions && viewdef.partitions.length ?
      viewdef.partitions.join(',') : 'None';
    const $aviewData = $(arc.misc.safe_substitute(tableTemplate, {
      NAME: viewdef.name,
      FORMAT: viewdef.file_format,
      COUNT: row_count,
      ROW_SIZE: arc.misc.sizeDisplay(viewdef.row_size, placeholder),
      MEMORY_SIZE: memSize,
      DISK_SIZE: diskSize,
      PARTITIONS: partitionStr,
    }));

    if (uncomputed) {
      $aviewData.find('.arc-invisible').removeClass('arc-invisible');
    }

    $aviewTableContainer.append($aviewData);
  }

  populateScoreDetail(scoreData);
  populateAVMetaData();

  if (accelContext) {
    const appTemplate = $('#single-app-sql-template').html();
    const appData = values.aviewApps.map((app) => arc.misc.safe_substitute(appTemplate, app));
    $appsContainer.append(appData);
    $vizTab.find('a').text(rawQueries ? 'Query SQL' : 'Visual SQL');
  }

  $vizTab.toggleClass('arc-invisible', !accelContext);
  $modal.find('[data-toggle="tooltip"]').tooltip();
  $modal.find('[data-toggle="popover"]').popover();
};

arc.admin.generateAvDimensionsHtml = function(aview) {
  const templatePartition = $('#accel-dim-partition-template').html();
  const partitions = aview.partitions || [];
  const dimensions = aview.dimensions.slice().filter((d) => !partitions.includes(d));

  partitions.forEach((p) => {
    dimensions.push(arc.misc.safe_substitute(templatePartition, { PARTITION: p }));
  });

  return dimensions.join(', ') || 'None';
};

arc.admin.setupAVDetailListeners = function(selected, accelContext = false, useRawQueries = false) {
  const $avDetailModal = $('#detail-accel-modal');

  function setupAVDetailModal(evt) {
    $('#aview-detail-modal-tabs a:first').tab('show');
    const $row = $(evt.relatedTarget).closest('tr');
    arc.admin.populateAVDetailModal($row, selected, accelContext, useRawQueries);
  }

  $avDetailModal.offOn('show.bs.modal', setupAVDetailModal);
  $avDetailModal.find('a[data-toggle="tab"]')
                .offOn('shown.bs.tab', function updateModalSize() {
                  const $this = $(this);
                  const $dialog = $this.closest('.modal-dialog');
                  if (['#aview-tab-sql', '#aview-tab-app'].includes($this.attr('href'))) {
                    $dialog.addClass('modal-lg');
                  } else {
                    $dialog.removeClass('modal-lg');
                  }
                });
};

arc.admin.accelManager = function() {
  let reports;
  let reportsByID = {};
  const $spinner = $('#accel-waiting-spinner');
  let appstat = {};
  let viewdefs = [];
  const dbexpanded = {};
  let refreshing = {};
  const cancelled = {};
  const pollseconds = 2;
  let usesRawQueries = false;
  // eslint-disable-next-line no-unused-vars
  let fromDirect = false;
  let idlist;

  // Per connection info returned from the server
  let dc_info = {};

  // This table is used to support the global av options.  This is harder because
  // we support multiple data connections, so we need to track information per connection
  const dcTable = {};

  const aviewManager = arc.admin.aviewManager();

  let templateTable;
  let templateRow;
  const icon_map = {
    dashboard: 'th-large',
    visual: 'bar-chart',
    filter: 'filter'
  };

  const accelManager = function() { return 'accelManager'; };

  accelManager.initialize = function(div, appids, queries, dcId, fromDirectAccess) {
    if (queries) {
      arc.qdict = {};
      JSON.parse(queries).forEach((q, i) => { arc.qdict[i] = q; });
      usesRawQueries = true;
      if (fromDirectAccess) {
        fromDirect = true;
      }
    }

    idlist = appids.split('_').map(Number);
    idlist.forEach((id) => { dbexpanded[id] = true; });
    fetchRefresh();
    if (usesRawQueries) {
      rawQueriesStart(arc.qdict, dcId);
    } else {
      accelManager.appAccelStart();
    }

    if (fromDirectAccess) {
      $('.direct-hidden').addClass('arc-invisible');
    }
  };

  accelManager.idlist = function(_) {
    if (!arguments.length) return idlist;
    idlist = _;
    return accelManager;
  };

  function hideSpinner() {
    $spinner.addClass('arc-invisible');
  }

  function showSpinner() {
    $spinner.removeClass('arc-invisible');
  }

  // Handle rename of recommended view, return true if name is changed
  function updateViewName(viewdef, newname) {
    const oldname = viewdef.name;
    if (oldname !== newname) {
      viewdef.name = newname;
      Object.keys(appstat).forEach((id) => {
        appstat[id].stats.forEach((stat) => {
          if (stat.viewname === oldname) {
            stat.viewname = newname;
          }
        });
      });
    }

    return (oldname !== newname);
  }

  // Update the database for the recommended AVs, returns true if changed
  function updateConnectionDatabase(dc_id, new_viewdb, force) {
    const old_viewdb = dcTable[dc_id].select_viewdb;
    if (force || old_viewdb !== new_viewdb) {
      dcTable[dc_id].select_viewdb = new_viewdb;
      viewdefs.filter((v) => v.state === 'RECOMMENDED' && v.dc_id === dc_id)
              .forEach((v) => {
        const parts = v.name.split('.');
        const newname = `${new_viewdb || v.engine_viewdb}.${parts[parts.length - 1]}`;
        updateViewName(v, newname);
      });
    }

    return (old_viewdb !== new_viewdb);
  }

  // Fixup the refresh status for a row in the analytical view table
  function paintRefreshStatus($row) {
    const viewdef = $row.data('viewdef');
    const name = `${viewdef.dc_id}.${viewdef.name}`;
    const refresh = refreshing[name];
    const $refreshLink = $row.find('.aview-refresh');
    const $refreshIcon = $row.find('.fa-refresh');
    const $refactive = $row.find('.aview-progress-active');

    if (!refresh || refresh.error) {
      $refreshIcon.removeClass('fa-spin');
      if (['FRESH', 'RECOMMENDED'].indexOf(viewdef.state) !== -1) {
        $refreshLink.addClass('disabled');
      } else {
        $refreshLink.removeClass('disabled');
      }

      $refactive.addClass('arc-invisible');
      cancelled[name] = undefined;
    } else if (refresh.status && refresh.status.status === 'Finished') {
      updateRowState($row, 'FRESH');
    } else {
      $refreshIcon.addClass('fa-spin');
      $refreshLink.addClass('disabled');
      $refactive.removeClass('arc-invisible');

      const $timeleft = $row.find('.aview-time-left');
      const $progress = $row.find('.progress-bar');

      // If task is cancelled, used the status frozen in the front-end
      const cancelStatus = cancelled[name];
      if (cancelStatus) {
        refresh.cancelled = true;
        refresh.progress = cancelStatus.progress;
      }

      $progress.attr('style', `width: ${refresh.progress}%`);
      $timeleft.text(arc.admin.refreshTimeLeftText(refresh, pollseconds));
    }
  }

  function swapToDesiredClass($label, targetClass) {
    const classList = 'bg-warning bg-danger bg-info bg-success';
    $label.removeClass(classList)
          .addClass(targetClass);
  }

  function updateAssociatedAppStatuses(viewName, state) {
    Object.keys(appstat).forEach(function checkApp(appId) {
      appstat[appId].stats.forEach(function updateViewState(stat) {
        if (stat.viewname === viewName) {
          stat.status = state;
        }
      });
    });

    paintAppStatTable();
  }

  function updateRowState($row, state) {
    const elementStates = {
      RECOMMENDED: {
        disableCreate: false,
        disableRefresh: true,
        disableDelete: true,
        disableSchedule: true,
        spinRefresh: false,
        hideProgressBar: true,
        text: 'Recommended',
        class: 'bg-info',
        appState: 'RECOMMENDATION',
      },
      FRESH: {
        disableCreate: true,
        disableRefresh: true,
        disableDelete: false,
        disableSchedule: false,
        spinRefresh: false,
        hideProgressBar: true,
        text: 'Up to date',
        class: 'bg-success',
        appState: 'ROUTED',
      }
    };

    const newStates = elementStates[state];

    // Update actions section
    $row.find('.aview-create').toggleClass('disabled', newStates.disableCreate);
    $row.find('.aview-refresh').toggleClass('disabled', newStates.disableRefresh);
    $row.find('.aview-delete').toggleClass('disabled', newStates.disableDelete);
    $row.find('.aview-schedule').toggleClass('disabled', newStates.disableSchedule);
    $row.find('.fa-refresh').toggleClass('fa-spin', newStates.spinRefresh);
    $row.find('.aview-progress-active').addClass('arc-invisible', newStates.hideProgressBar);

    // Update saved data
    const viewdef = $row.data('viewdef');
    viewdef.state = state;

    const $viewStatusLabel = $row.find('.badge');
    swapToDesiredClass($viewStatusLabel, newStates.class);
    $viewStatusLabel.text(newStates.text);

    // update app status
    const viewName = viewdef.name;
    updateAssociatedAppStatuses(viewName, newStates.appState);
  }

  function fetchRefresh(isInitial = false) {
    $.ajax({
        url: arc.frontend.globals.urls.viewRefreshUrl,
        data: { opcode: 'fetch' },
        type: 'POST',
        beforeSend(xhr) {
          // Because of a jquery bug when handling PUT requests,
          // we have to explicitly set X-CSRFToken here
          xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
        }
      })

      .fail(() => {
      })
      .done((rsp) => {
        // Rebuild list of refreshing views
        refreshing = {};
        rsp.views.forEach((s) => {
          refreshing[`${s.dc_id}.${s.view}`] = s;
          if (s.error) {
            arc.report.alertMessageSmall(`Refresh of ${s.view} failed: ${s.error}`, 'alert-danger');
          }
        });

        $('.aviews-tables-container tbody tr').each((idx, tr) => {
          const $row = $(tr);
          const viewdef = $row.data('viewdef');
          const refresh = refreshing[`${viewdef.dc_id}.${viewdef.name}`];

          if (refresh) {
            paintRefreshStatus($row);
          }
        });
      })
      .always(function keepPollingAsNecessary(rsp) {
        // If this is are very first request, give it one more poll even if nothing is actively
        // refreshing to allow for lagginess
        if ((!(rsp.views && rsp.views.length === 0) || isInitial) &&
          $('#accel-view-selection').is(':visible')) {
            setTimeout(fetchRefresh, pollseconds * 1000);
        }
      });
  }

  accelManager.appAccelStart = function() {
    reports = undefined;
    reportsByID = {};
    const data = { appids: JSON.stringify(idlist) };
    showSpinner();
    $.ajax({
      url: arc.frontend.globals.urls.viewRecUrl,
      data,
      type: 'POST',
      beforeSend(xhr) {
        // Because of a jquery bug when handling PUT requests,
        // we have to explicitly set X-CSRFToken here
        xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
      }
    })
    .fail(function failureCb(rsp, textstatus, errorthrown) {
      textstatus = rsp.responseText || textstatus;
      arc.report.alertMessageSmall(
        `Error get view recommendations: ${textstatus}. ${errorthrown}`,
        'alert-danger'
      );
    })
    .always(hideSpinner)
    .done(gotResponse);
  };

  function clickDashboardArrow(e) {
    const $row = $(e.currentTarget).closest('tr');
    const $tbody = $row.closest('tbody');
    const id = $row.data('id');
    const fa = $row.find('.fa-angle-right, .fa-angle-down');
    if (fa.hasClass('fa-angle-right')) {
      dbexpanded[id] = true;
      fa.removeClass('fa-angle-right');
      fa.addClass('fa-angle-down');
      $tbody.find(`[arc-owning-appid=${id}]`).show();
    } else {
      delete dbexpanded[id];
      fa.addClass('fa-angle-right');
      fa.removeClass('fa-angle-down');
      $tbody.find(`[arc-owning-appid=${id}]`).hide();
    }
  }

  function displayAppName(app) {
    if (app.title) {
      return `${app.title} (ID: ${app.query_id || app.id})`;
    }

    const data = JSON.parse(app.data);
    if (app.report_type === 'picklist') {
      return `Column filter on ${data.report_data.user_settings.picklist_base_field}`;
    }

    if (app.report_type === 'daterange') {
      return `Date filter ${data.report_data.user_settings.daterange_base_field}`;
    }

    return `Untitled (ID: ${app.id})`;
  }

  function rawQueriesStart(queries, dcId) {
    reports = [];
    reportsByID = {};
    idlist = [];
    const qdict = {};
    Object.keys(queries).forEach((key) => {
      const queryData = queries[key];
      const id = parseInt(key, 10);
      const report = {
        id,
        query_id: queryData['Query Details'],
        title: 'Query',
        report_type: 'raw-query',
      };
      reports.push(report);
      reportsByID[id] = report;
      idlist.push(id);
      qdict[id] = queryData.Query;
    });

    const data = {
      qdict: JSON.stringify(qdict),
      dcid: dcId,
    };
    showSpinner();
    $.ajax({
      url: arc.frontend.globals.urls.rawSqlViewRecUrl,
      data,
      type: 'POST',
      beforeSend(xhr) {
        // Because of a jquery bug when handling PUT requests,
        // we have to explicitly set X-CSRFToken here
        xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
      }
    })
    .fail(function handleError(rsp, textstatus, errorthrown) {
      textstatus = rsp.responseText || textstatus;
      arc.report.alertMessageSmall(
        `Error get view recommendations: ${textstatus}. ${errorthrown}`, 'alert-danger'
      );
    })
    .always(hideSpinner)
    .done(gotResponse);
  }

  // This function sets a listener on the element with ID specified
  function expandOrCollapseAll(id) {
    const $btn = $(`#${id}`);
    $btn.off('click.toggleExpandAll')
        .on('click.toggleExpandAll', function expand() {
          const isExpanded = this.className.includes('fa-angle-double-down');
          const $tbody = $('.expanding-table tbody');
          const $fa = $tbody.find('.fa-angle-right, .fa-angle-down');

          $fa.toggleClass('fa-angle-down', !isExpanded)
             .toggleClass('fa-angle-right', isExpanded);
          $btn.toggleClass('fa-angle-double-down', !isExpanded)
              .toggleClass('fa-angle-double-right', isExpanded);

          if (isExpanded) {
            $tbody.find('.indented').hide();
          } else {
            $tbody.find('.indented').show();
          }
      });
  }

  // Build one row for a view definition
  function oneViewRow(aview) {
    const { name } = aview;
    const rinfo = refreshing[`${aview.dc_id}.${name}`];
    const hasRefreshError = rinfo && rinfo.error;
    const state = rinfo && !rinfo.error ? arc.admin.aviewStates.REFRESHING :
      arc.admin.translateAviewState(aview.state, hasRefreshError, aview.error_message);
    const measures = aview.aggregates;

    aview.expensive_aggregates.forEach(function addExpAgg(agg) {
      if (!measures.includes(agg)) {
        measures.push(agg);
      }
    });

    const substitutions = {
      VIEW: name,
      HTML_STATE: state,
      SQL: aview.select,
      HTML_DIMENSIONS: arc.admin.generateAvDimensionsHtml(aview),
      MEASURES: measures.join(', ') || 'None',
      DETAIL_BTN_CLASS: aview.error_message ? 'disabled' : '',
    };

    const $row = $(arc.misc.safe_substitute(templateRow, substitutions));
    const scoreData = arc.admin.computeAviewScore(aview);
    $row.find('.arc-aview-rating').html(arc.admin.generateStars(scoreData.overallScore));

    const refreshLink = $row.find('.aview-refresh');
    const createLink = $row.find('.aview-create');
    const deleteLink = $row.find('.aview-delete');
    const scheduleLink = $row.find('.aview-schedule');

    if (aview.state === 'RECOMMENDED') {
      $row.find('[arc-view-name]').attr({ contenteditable: true, spellcheck: false })
                                 .addClass('view-rec-name');
      refreshLink.addClass('disabled');
      deleteLink.addClass('disabled');
      scheduleLink.addClass('disabled');
    } else {
      createLink.addClass('disabled');
    }

    if (aview.state === 'FRESH') {
      refreshLink.addClass('disabled');
    }

    substitutions.aviewApps = [];

    Object.keys(appstat).forEach(function checkApp(appId) {
      appstat[appId].stats.forEach(function getSQL(app) {
        if (app.viewname === aview.name) {
          substitutions.aviewApps.push({
            NAME: displayAppName(reportsByID[app.id]),
            QUERY: app.query,
          });
        }
      });
    });

    $row.data('viewdef', aview)
        .data('substitutions', substitutions)
        .data('score', scoreData);
    paintRefreshStatus($row);
    return $row;
  }

  function normalizeTemplates(rsp) {
    const $table = $('#accel-table-template');
    const $row = $('#accel-table-row-template').find('tbody');
    const version = rsp.old_style ? 'v2_9_1' : 'v2_5';
    $table.find(`.arcengine_${version}`).remove();
    $row.find(`.arcengine_${version}`).remove();
    templateTable = $table.html();
    templateRow = $row.html();
  }

  function paintViews() {
    const aviews = viewdefs;

    // Collect up views by tables - this helps display all the views for a
    // single table next to each other
    const avhash = {};
    aviews.forEach((aviewObj) => {
      if (!(aviewObj.base in avhash)) {
        avhash[aviewObj.base] = [];
      }

      avhash[aviewObj.base].push(aviewObj);
    });

    // Remove any previously painted tables
    const $tables_container = $('.aviews-tables-container');
    $tables_container.empty().off();

    // Enable tooltips for the create/refresh/delete buttons
    $tables_container.tooltip({
      selector: '[data-toggle="tooltip"]',
      placement: 'bottom',
      delay: {
       show: 500,
       hide: 100
      }
    });

    // Draw each base table and the views for that table
    Object.keys(avhash).sort().forEach((tablename) => {
      const $base_container = $(arc.misc.safe_substitute(
          templateTable,
          {
            TABLENAME: tablename
          }
      ));

      const link_table = {};
      $tables_container.append($base_container);
      avhash[tablename].sort().forEach((aview) => {
        const $row = oneViewRow(aview);
        if (link_table[aview.name] === undefined) {
          link_table[aview.name] = true;
          $row.attr('id', `arc-view-id-${aview.name.replace('.', '-')}`);
        }

        $base_container.find('tbody').append($row);
      });
    });

    setupViewTableEvents();
  }

  function clickCreateView(e) {
    const td = $(e.currentTarget).closest('td');
    td.find('.aview-create').find('.fa').addClass('fa-spin');
    td.find('.aview-create').addClass('disabled');
    const viewdef = td.closest('tr').data('viewdef');
    const quoted_name = arc.dataconnection._enquote_tablename(viewdef.name, 'arcengine');

    // For location string for this connection if on has been selected
    let cmdstr = `CREATE ANALYTICAL VIEW ${quoted_name}${viewdef.full_sql}`;
    const { dc_id } = viewdef;
    const lockey = dcTable[dc_id].location_key;
    const locstr = lockey ? dc_info[dc_id].LOCATIONS[lockey] : '';
    const parts = viewdef.name.split('.');
    if (locstr && parts.length === 2) {
      cmdstr = cmdstr.replace('PARQUET',
                              `PARQUET LOCATION '${locstr}/${parts[0]}.db/${parts[1]}'`);
    }

    arc.report.sql_runner.run(dc_id, cmdstr,
        (errorMsg) => {
          if (errorMsg) {
            td.find('.aview-create').find('.fa').removeClass('fa-spin');
            td.find('.aview-create').removeClass('disabled');
            arc.report.alertMessageSmall(`Unable to create analytical view '${viewdef.name}' : ${errorMsg}`, 'alert-danger');
          } else {
            viewdef.state = 'UNUSABLE';
            Object.keys(appstat).forEach(function checkApp(id) {
              appstat[id].stats.forEach(function updateViewStatus(stat) {
                if (stat.viewname === viewdef.name && stat.status === 'RECOMMENDATION') {
                  stat.status = 'AWAITING_REFRESH';
                }
              });
            });
            paintViews();
            paintAppStatTable();
          }
          }, false, undefined, undefined, false);
  }

  function populateSQLModal(evt) {
    const $row = $(evt.relatedTarget).closest('tr');
    const displayType = $row.data('display-type');
    const $modal = $('#viz-sql-modal');
    $modal.find('h3').text(`${displayType} SQL`);
    $modal.find('.sql').text($row.data('query'));
    $modal.find('h4').text(displayAppName(reportsByID[$row.data('id')]));
  }

  function populateAVModal(evt) {
    const $row = $(evt.relatedTarget).closest('tr');
    $('#query_id').text(reportsByID[$row.data('id')].query_id);
    $('#query_sql').text($row.data('query'));
  }

  /*
   * Here we update the global option model based on the current recommendation response
   * as contained in the viewdefs list.
   * Note that we might also update the view names from the previously selected database
   * for each connection
   */
  function updateGlobalOptionsModal() {
    // Text to indicate use default database from arcengine recommendation response
    const dfldb = '(Use database from base table)';

    // Get list of connections IDs present in the current recommendation response
    // At one time, planned on supporting options only for connections where we
    // got some recommended aviews (not existing, etc.) but that is hard to track
    // and not consistent with other buttons, so we don't filter the viewdefs.
    const vrecs = viewdefs;
    const vrids = new Set(vrecs.map((v) => v.dc_id));

    // If no AVs recommended, then there are no options to update.
    const $btn = $('#accel-globaloptions-btn');
    if (vrecs.length === 0) {
      $btn.hide();
      return;
    }

    // Update the dcTable with entries for any newly added data connections
    // Also save the engine recommended view database
    vrecs.forEach((v) => {
      v.engine_viewdb = v.name.split('.')[0];
      if (dcTable[v.dc_id] === undefined) {
        dcTable[v.dc_id] = { dc_id: v.dc_id, dc_name: v.dc_name };
      }
    });

    // Populate the data connection selector within the modal
    const $dcdiv = $('#accel-create-aview-connection');
    const $options = $dcdiv.find('select');
    $options.empty();
    vrids.forEach((id) => $options.append($(`<option value=${id}>${dcTable[id].dc_name}</option>`)));

    // If only one connection, then we can hide the selector
    $dcdiv.toggleClass('arc-invisible', vrids.size === 1);

    // Update the recommended views database name from what has been selected
    vrids.forEach((dc_id) => {
      const dbname = dcTable[dc_id].select_viewdb;
      if (dbname && dbname !== dfldb) {
        updateConnectionDatabase(dc_id, dbname, true);
      }
    });

    // When the connection selector changes, update the database and location selectors
    const $modal = $('#accel-globaloptions-modal');
    const $dbsel = $modal.find('#accel-create-aview-database select');
    const $lcdiv = $modal.find('#accel-create-aview-location');
    const $lcsel = $lcdiv.find('select');

    $options.offOn('change', () => {
      const dc_id = Number($options.val());

      $modal.find('.modal-title')
            .text(`AV creation options for connection: ${dcTable[dc_id].dc_name}`);

      const dbs = dc_info[dc_id].DATABASES || [];
      $dbsel.empty();
      $dbsel.append($(`<option value="${dfldb}">${dfldb}</option>`));
      dbs.forEach((db) => $dbsel.append($(`<option value="${db}">${db}</option>`)));
      if (dcTable[dc_id].select_viewdb) {
        $dbsel.val(dcTable[dc_id].select_viewdb);
      }

      const locs = dc_info[dc_id].LOCATIONS || {};
      const keys = Object.keys(locs);
      $lcsel.empty();
      $lcsel.append($(`<option value="" selected>(System default)</option>`));
      keys.forEach((key) => $lcsel.append($(`<option value="${key}">${key}</option>`)));
      $lcdiv.toggleClass('arc-invisible', keys.length === 0);
      if (dcTable[dc_id].location_key) {
        $lcsel.val(dcTable[dc_id].location_key);
      }
    });

    $btn.show();
    $btn.offOn('click', () => {
      if (arc.misc.modalShown($modal)) {
        return;
      }

      $options.trigger('change');
      $modal.modal();
    });

    $('#accel-globaloptions-save').offOn('click', () => {
      const dc_id = Number($options.val());
      const dbname = $dbsel.val();

      dcTable[dc_id].location_key = $lcsel.val();

      if (updateConnectionDatabase(dc_id, dbname === dfldb ? '' : dbname)) {
        paintViews();
        paintAppStatTable();
      }

      $modal.find('.close').trigger('click');
    });
  }

  // Set up event handles for the views table
  function setupViewTableEvents() {
    const $div = $('.aviews-tables-container');

    // Kind of a weird way to handle the 'all views' buttons (by
    // just triggering a click on all the individual icons) but works
    // ok.  Some concerns about error handling.
    $('#accel-refreshviews-btn').off().on('click', () => {
      $div.find(".aview-refresh:not('.disabled')").trigger('click');
    });

    $('#accel-makeviews-btn').off().on('click', () => {
      $div.find(".aview-create:not('.disabled')").trigger('click');
    });

    arc.admin.setupAVDetailListeners(Object.keys(reportsByID), true, usesRawQueries);

    $('#viz-sql-modal').offOn('show.bs.modal', populateSQLModal);
    $('#query-detail-modal').offOn('show.bs.modal', populateAVModal);

    // Set up for toggle SQL visibility
    $div.on('click', '.av-sql-link', function() {
      $(this).next('.av-sql-detail').toggle();
      $(this).text((i, text) => (text === 'Show SQL' ? 'Hide SQL' : 'Show SQL'));
    });

    $div.on('blur', 'span[contenteditable]', (e) => {
      const viewdef = $(e.currentTarget).closest('tr').data('viewdef');
      const newname = $(e.currentTarget).text();
      if (updateViewName(viewdef, newname)) {
        // Kind of brute force here but the view name is used in
        // multiple places so easiest to do full rebuild/redraw
        // eslint-disable-next-line no-undef
        $span.closest('tr').data('viewdef').name = newname;

        paintViews();
        paintAppStatTable();
      }
    });

    $div.on('keyup keydown', 'span[contenteditable]', (e) => {
      const $span = $(e.currentTarget);
      if (e.type === 'keydown') {
        if (e.which === 13) {
          e.stopImmediatePropagation();
          return false;
        }
      }

      if (e.which === 27) {
        $span.text($span.attr('arc-view-name'));
        return false;
      }

      if (e.which === 13) {
        $span.trigger('blur');
        return false;
      }
    });

    $div.on('click', ".aview-create:not('.disabled')", clickCreateView);

    $div.on('click', ".aview-refresh:not('.disabled')", function() {
        const $row = $(this).closest('tr');
        const viewdef = $row.data('viewdef');
        refreshView(viewdef);
        paintRefreshStatus($row);
        fetchRefresh(true);
      });

    $div.on('click', ".aview-delete:not('.disabled')", function() {
         const $row = $(this).closest('tr');
         dropView($row);
      });

    $div.on('click', '.aview-cancel-btn', function() {
        const $row = $(this).closest('tr');
        const viewdef = $row.data('viewdef');

        cancelRefresh(viewdef);
        paintRefreshStatus($row);
      });

    $div.on('click', ".aview-schedule:not('.disabled')", function() {
        const viewdef = $(this).closest('tr').data('viewdef');
        aviewManager.dataconnection_id(viewdef.dc_id);
        arc.frontend.legacyBridge.events.next({
          type: 'open-analytical-job-modal',
          payload: { dataConnectionId: viewdef.dc_id, viewName: viewdef.name }
        });
      });
  }

  function refreshView(viewdef) {
    const data = { opcode: 'enque',
                 formValues: JSON.stringify({ dcid: viewdef.dc_id,
                   view: viewdef.name }) };
    $.ajax({
        url: arc.frontend.globals.urls.viewRefreshUrl,
        data,
        type: 'POST',
        beforeSend(xhr) {
          // Because of a jquery bug when handling PUT requests,
          // we have to explicitly set X-CSRFToken here
          xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
        }
      })
      .fail((rsp, textstatus, errorthrown) => {
        textstatus = rsp.responseText || textstatus;
        arc.report.alertMessageSmall(`Refresh error: ${textstatus}. ${errorthrown}`, 'alert-danger');
      })
      .done((rsp) => {
        // Create an initial refreshing status record so we can paint the status
        refreshing[`${rsp.dcid}.${rsp.view}`] = {
          dc_id: rsp.dcid,
          view: rsp.view,
          schedule_id: rsp.schedule_id,
          joblog_id: rsp.joblog_id,
          status: 'Status: Queued',
          progress: 0,
        };
      });
  }

  function cancelRefresh(viewdef) {
    const name = `${viewdef.dc_id}.${viewdef.name}`;

    // Don't send multiple AJAX requests if already cancel procedure is already
    // in progress
    if (cancelled[name]) {
      return;
    }

    let schedule_id;
    let joblog_id;
    let progress;
    const viewInfo = refreshing[name];

    if (viewInfo) {
      schedule_id = viewInfo.schedule_id;
      joblog_id = viewInfo.joblog_id;
      progress = viewInfo.progress;
    }

    if ((!schedule_id && schedule_id !== 0) || (!joblog_id && joblog_id !== 0)) {
      return;
    }

    cancelled[name] = { progress, };
    const url = `${arc.frontend.globals.urls.cancelJobUrl}/${schedule_id}/${joblog_id}`;
    $.ajax({
        url,
        data: {},
        type: 'POST',
        beforeSend(xhr) {
          // Because of a jquery bug when handling PUT requests,
          // we have to explicitly set X-CSRFToken here
          xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
        }
      })
      .fail((rsp, textstatus, errorthrown) => {
        cancelled[name] = undefined;
        textstatus = rsp.responseText || textstatus;
        arc.report.alertMessageSmall(`Error cancelling refresh: ${textstatus}. ${errorthrown}`, 'alert-danger');
      })
      .done(() => {
        arc.report.alertMessageSmall(`View ${viewdef.name} refresh cancelled`,
            'alert-success');
      });
  }

  function dropView($row) {
    const c = arc.misc.deleteConfirmation();
    const viewdef = $row.data('viewdef');

    function dropFetchCb(errorMsg) {
      c.hideModal();
      if (errorMsg) {
        arc.report.alertMessageAllHide();
        arc.report.alertMessage(`Failed to drop aview ${viewdef.name}`,
            errorMsg);
      } else {
        arc.report.alertMessageSmall(`View ${viewdef.name} deleted`,
            'alert-success');
      }

      updateRowState($row, 'RECOMMENDED');
    }

    const rb = arc.requestBuilder();
    rb.requestType('.drop_aviews');
    rb.tableName(arc.dataconnection._enquote_tablename(viewdef.name, 'arcengine'));
    c.objectType('Analytical View')
     .objectName(viewdef.name)
     .askUser(() => {
       arc.report.sql_runner.run(viewdef.dc_id,
           '', dropFetchCb, false, rb.dsreq());
     });
  }

  // Paint the apps portion of the recommend response
  function paintAppStatTable() {
    let templ;
    let $table;

    if (usesRawQueries) {
      $table = $('#query-accel-response-table');
      templ = $('#query-accel-response-template').find('tbody').html();
    } else {
      $table = $('#app-accel-response-table');
      templ = $('#app-accel-response-template').find('tbody').html();
    }

    const $tbody = $table.find('tbody');
    $tbody.empty();
    $tbody.off();

    function onerow(app, dbid) {
      app.title = arc.misc.escapeHTML(app.title);
      let dbcaret = '';
      let idtext = arc.misc.appsToLinks([app.id], 2).replace('href=', 'target="_blank" href=');
      let { title } = app;
      let displayType = '';
      let queryData;

      const display_statuses = [];
      if (app.report_type === 'dashboard') {
        dbcaret = dbexpanded[app.id] ?
                    '<a class="link"><span class="fa fa-angle-down"></span></a>' :
                    '<a class="link"><span class="fa fa-angle-right"></span></a>';
      } else {
        appstat[app.id].stats.forEach(function processAviewStat(stat) {
          const status = generateDisplayStatus(stat);
          if (!display_statuses.includes(status)) {
            display_statuses.push(status);
          }
        });
      }

      function generateDisplayStatus(stat) {
        if (stat.label_code === 'danger') {
          return `<span class="badge badge-${stat.label_code}">${stat.reason}</span>`;
        }

        const name = stat.viewname;
        return `<span class="badge badge-${stat.label_code}" arc-view-name=${name}>${name}</span>`;
      }

      if (dbid) {
        title = app.title;
        displayType = 'visual';
      } else if (app.report_type === 'raw-query') {
        queryData = arc.qdict[app.id];
        displayType = 'query';
      } else {
        displayType = 'dashboard';
      }

      if (app.report_type === 'picklist' || app.report_type === 'daterange') {
        idtext = app.id;
        displayType = 'filter';
        title = `${app.report_type} for app ${dbid}`;
        const data = JSON.parse(app.data);
        if (app.report_type === 'picklist') {
          idtext = 'Filter';
          title = data.report_data.user_settings.picklist_base_field;
        } else {
          idtext = 'Date Filter';
          title = data.report_data.user_settings.daterange_base_field;
        }
      }

      let $tr;
      if (usesRawQueries) {
        $tr = $(arc.misc.safe_substitute(templ,
          {
            TIMESTAMP: arc.misc.dateToMoment(queryData.Timestamp).format('MMM D, h:mm:ss A'),
            ORDER: queryData['Query Details'],
            USER: queryData['Issuing User'],
            IP_ADDRESS: queryData.IP,
            HTML_BASETABLES: arc.misc.displayBaseTables(queryData['Base Table(s)']),
            HTML_APP_STATUS: display_statuses.join(' '),
          }));
      } else {
        $tr = $(arc.misc.safe_substitute(templ,
          {
            HTML_APP_CARET: dbcaret,
            APP_ID: app.id,
            APP_ICON: icon_map[displayType],
            HTML_APP_IDTEXT: idtext,
            HTML_APP_TITLE: title,
            HTML_APP_STATUS: display_statuses.join(' '),
            APP_DBID: dbid,
            APP_CLASS: displayType === 'dashboard' ? '' : 'indented',
            SQL_CLASS: displayType === 'dashboard' ? 'arc-invisible' : '',
          }));
      }

      $tr.data('id', app.id);
      $tr.data('display-type', `${displayType[0].toUpperCase()}${displayType.substring(1)}`);
      $tr.data('query', appstat[app.id].query);

      if (dbid !== 0 && !dbexpanded[dbid]) {
         $tr.hide();
      }

      return $tr;
    }

    // Translate status to label color code
    Object.keys(appstat).forEach((id) => {
      const app = reportsByID[id];
      if (app.report_type !== 'dashboard') {
        const status_map = {
          ROUTED: ['success', 'Accelerated'],
          FAILED: ['danger', 'Errors'],
          RECOMMENDATION: ['info', 'Recommended'],
          AWAITING_REFRESH: ['warning', 'Need refresh'],
          NO_RECOMMENDATION: ['danger', 'Unable to accelerate'],
        };
        appstat[app.id].stats.forEach(function translateStatus(stat) {
          [stat.label_code, stat.status_desc] = status_map[stat.status] || status_map.FAILED;
        });
      }
    });

    idlist.forEach((id) => {
      const app = reportsByID[id];
      if (!app || app.report_type === 'picklist' || app.report_type === 'daterange') {
        return;
      }

      // Overwrite dashboard status with summary
      let dbapps = [];
      if (app.report_type === 'dashboard') {
        dbapps = app.pickers.concat(app.visuals);
        const dbcodes = {};
        const dbdescs = {};
        // eslint-disable-next-line no-shadow
        dbapps.forEach((id) => {
          appstat[id].stats.forEach(function updateStatusTotals(aviewStat) {
            const key = aviewStat.label_code;
            if (dbcodes[key] === undefined) {
              dbcodes[key] = 0;
              dbdescs[key] = aviewStat.status_desc;
            }

            dbcodes[key] += 1;
          });
        });
        const spans = [];
        Object.keys(dbcodes).sort().forEach((key) => {
          spans.push(`<span class="badge badge-${key}">${dbcodes[key]} ${dbdescs[key]}</span>`);
        });
        appstat[app.id].stats[0].status = spans.join(' ');
      }

      $tbody.append(onerow(app, 0));
      if (app.report_type === 'dashboard') {
        dbapps.forEach((cid) => {
          $tbody.append(onerow(reportsByID[cid], app.id));
        });
      }
    });

    // Set up events handlers for apps table
    $tbody.on('mouseenter', 'span[arc-view-name]', (e) => {
      const name = $(e.currentTarget).attr('arc-view-name');
      $('.aviews-tables-container').find(`span[arc-view-name="${name}"]`)
                                   .addClass('rview-highlight');
    });
    $tbody.on('mouseleave', 'span[arc-view-name]', (e) => {
      const name = $(e.currentTarget).attr('arc-view-name');
      $('.aviews-tables-container').find(`span[arc-view-name="${name}"]`)
                                   .removeClass('rview-highlight');
    });
    $tbody.on('click', 'span[arc-view-name]', (e) => {
      const name = $(e.currentTarget).attr('arc-view-name');
      $.scrollTo($(`#arc-view-id-${name.replace('.', '-')}`), 500);
    });
    $tbody.on('click', 'td .fa-angle-right, td .fa-angle-down', clickDashboardArrow);

    expandOrCollapseAll('expand-all-rec-status');
  }

  function gotResponse(rsp) {
    normalizeTemplates(rsp);
    viewdefs = rsp.viewdefs;
    appstat = rsp.appstat;
    reports = rsp.reports;

    if (reports) {
      reports.forEach((r) => { reportsByID[r.id] = r; });
    }

    dc_info = rsp.dc_info;
    updateGlobalOptionsModal();

    paintViews();
    paintAppStatTable();

    $('#accel-view-selection').show();
  }

  return accelManager;
};
}());
