/*
 * Extracts are copies of a dataset materialized to the same or a different data
 *   connection.
 *
 * Datasets can have multiple extracts associated with them
 * Extracts are defined by:
 *   a source dataset
 *   a target connection
 *   a target table
 *   a list of dimensions, aggregates and partitions
 *
 * If partitions are specified on the target of an extract, and if both the
 *   source and target datasets support partitioning, then incremental refreshes
 *   are run. Incremental refreshes rely on the output of 'SHOW FILES IN <tablename>'
 *   on the source table being different for a particular partition if data in the
 *   partition needs to be refreshed. This is not always guaranteed, so there is some
 *   administrative willpower relied upon to configure incremental refreshes only
 *   in those cases where this is reasonable.
 *
 * Information about extracts is stored in the Extract model in datasets/models
 *
 * If no schedule is associated with an extract, then the only way to run an
 *   an extract is to click on 'Run Now' in the UI. In many cases, this is sufficient
 *   and no schedule is necessary since the source dataset almost never changes.
 *
 * The implementation of extracts is currently only associated with the source
 *   dataset. To use the extract, a new dataset has to be created off the target
 *   data table, and then used explicitly. No automatic relationship is setup
 *   between the source & target tables.
 *
 * The code in this file (arc_extractmgr.js), arcweb/extracts/ and arcweb/jobs/datacopy.py
 *   helps manage extracts in the frontend and the backend.
 */

window.arc = window.arc || {};

(function() {
'use strict';

arc.admin = arc.admin || {};

function Extract(config) {
  const e = function() { return 'Extract'; };

  const {
    srcDatasetId,
    tgtDataconnectionId,
    tgtDbname,
    tgtTablename,
    dimData,
    aggData,
    partitionData,
    extract_type,
    emailSuccess,
    emailFail,
    emailTo,
    emailSubject,
    rewriteStatus
  } = config;

  e.create = function(successCb, failCb) {
    // casting emailSuccess/emailFail to str. Otherwise, bool val getting
    // converted to python bool True/False on backend by json, and being
    // saved as json bool true/false in db instead of "true"/"false" as before
    const extract = {
      srcDatasetId,
      tgtDataconnectionId,
      tgtDbname,
      tgtTablename,
      dimData,
      aggData,
      partitionData,
      extract_type,
      emailSuccess: String(emailSuccess),
      emailFail: String(emailFail),
      emailTo,
      emailSubject,
      rewriteStatus
    };

    const url = arc.frontend.globals.urls.extractUrl;
    $.ajax({
      url,
      data: JSON.stringify(extract),
      type: 'POST',
      beforeSend(xhr) {
        xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
      },
    })
    .done((resp) => {
      const extractData = resp[0];
      successCb(extractData);
    })
    .fail(failCb);
  };

  return e;
}

function deleteExtract(extractId, successCb, failCb, deleteTable = false) {
  const url = `${arc.frontend.globals.urls.extractUrl}/${extractId}?deleteTable=${deleteTable}`;
  $.ajax({
    url,
    type: 'DELETE',
    beforeSend(xhr) {
      xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
    },
  })
  .fail(failCb)
  .done(successCb);
}

function editExtract(extractId, editData, successCb, failCb) {
  const url = `${arc.frontend.globals.urls.extractUrl}/${extractId}`;
  $.ajax({
    url,
    data: JSON.stringify(editData),
    type: 'PUT',
    beforeSend(xhr) {
      xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
    },
  })
  .fail(failCb)
  .done((resp) => {
    successCb(resp.extract_id);
  });
}

function deleteJobAndExtract(extractId, successCb, failCb, deleteTable = false) {
  // First, delete the job associated with this extract. Then delete the extract
  createOrDeleteExtractJob(extractId, 0, () => {
    deleteExtract(extractId, successCb, failCb, deleteTable);
  }, failCb);
}

function rerunJob(scheduleId, successCb, failCb) {
  const data = {
    schedules: JSON.stringify([scheduleId]),
  };

  $.ajax({
    url: arc.frontend.globals.urls.rerunJobUrl,
    data,
    type: 'POST',
    beforeSend(xhr) {
      xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
    },
  })
  .done((_data) => {
    arc.report.alertMessageSmall('Data Extract has been scheduled', 'alert-success');

    if (successCb) {
      successCb(_data);
    }
  })
  .fail((_data) => {
    const errorMessage = `Failed to rerun job: ${_data.responseText}`;
    arc.report.alertMessage(errorMessage);

    if (failCb) {
      failCb(errorMessage);
    }
  });
}

/*
 * A scheduleId of 0 deletes any existing extract job.
 * A scheduleId of undefined creates a adhoc job that runs now.
 * A scheduleId > 0 creates a job with the specified schedule.
 */
function createOrDeleteExtractJob(extractId, scheduleId, successCb, failCb) {
  const data = {
    extractId,
    scheduleId,
  };

  $.ajax({
    url: arc.frontend.globals.urls.dataCopyUrl,
    data,
    type: 'POST',
    beforeSend: (xhr) => {
      xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
    },
  })
  .fail((req) => {
    let err = '';
    if (req.responseText) {
      err = `: ${req.responseText.substring(0, 200)}`;
    }

    const fullError = `Failed to setup a data copy schedule${err}`;
    arc.report.alertMessageSmall(fullError, 'alert-danger');

    if (failCb) {
      failCb(fullError);
    }
  })
  .done((_data) => {
    if (scheduleId === undefined) {
      arc.report.alertMessageSmall('Data Extract has started', 'alert-success');
    } else {
      arc.report.alertMessageSmall('Data Extract has been scheduled', 'alert-success');
    }

    if (successCb) {
      successCb(_data);
    }
  });
}

arc.admin.dataExtractManager = function() {
  const m = function() { return 'arc.admin.dataExtractManager'; };

  const $createExtractBtn = $('#create-extract-btn');
  const $refreshDisplay = $('#refresh-extract-display');
  const $noExtractsLabel = $('.extracts-none');
  const $extractCounter = $('#extracts-count');
  const $tableContainer = $('#extracts-table-container');
  const $tableTemplate = $('#extracts-table-template');
  const $tableRowTemplate = $('#extracts-table-row-template tbody');

  let builder;
  let dataconnections = {};
  let tablesToDataset = {};
  let datasetIdToName = {};
  let extracts = [];
  let extractIdToExtract = {};
  let extractLogs = {};

  m.dataconnections = function() {
    return dataconnections;
  };

  m.builder = function(_) {
    if (!arguments.length) return builder;
    builder = _;
    return m;
  };

  m.initialize = function() {
    $createExtractBtn
        .offOn('click.createExtract', () => {
          arc.extractCreator()
              .dataExtractManager(m)
              .create();
        });

    $refreshDisplay
        .offOn('click.refreshDisplay', () => {
          m.fetchAndPaint(true);
        });

    m.fetchAndPaint();

    return m;
  };

  m.fetchAndPaint = function(reset_controller) {
    $refreshDisplay.addClass('disabled');
    $refreshDisplay.find('.fa-refresh').addClass('fa-spin');

    function failCb() {
      $refreshDisplay.removeClass('disabled');
      $refreshDisplay.find('.fa-refresh').removeClass('fa-spin');
    }

    function successCb() {
      $refreshDisplay.removeClass('disabled');
      $refreshDisplay.find('.fa-refresh').removeClass('fa-spin');
      paint();
    }

    fetch(successCb, failCb, reset_controller);
  };

  function _collistToStr(collist) {
    return collist.map((d) => d.expr.replace(/[[\]]/g, '')).join(', ');
  }

  function paint() {
    let runningCount = 0;

    const len = extracts.length;
    if (len) {
      $noExtractsLabel.addClass('arc-invisible');
      $tableContainer.html($tableTemplate.html());
      $extractCounter.text(extracts.length);
    } else {
      $noExtractsLabel.removeClass('arc-invisible');
      $tableContainer.empty();
      $extractCounter.empty();
    }

    const contentsHtml = $tableRowTemplate.html();
    extracts.forEach((extract) => {
      const dskey = `${extract.tgtDataconnectionId}:${extract.tgtDbname}.${extract.tgtTablename}`;
      const datasets = tablesToDataset[dskey] || [];

      const htmlDatasets = datasets.map(
        (ds) => {
          const target = '_blank';
          const url = arc.misc.getBlankTargetUrl(`${arc.frontend.globals.urls.datasetEditUrl}/${ds.id}`, target);
          return `<a title="${arc.misc.escapeHTML(ds.name)}" target="${target}" href="${url}">${arc.misc.escapeHTML(ds.name)}</a>`;
        }
      );

      htmlDatasets.push(`<a data-id="${extract.id}" ` +
          'class="create-dataset btn viz-btn-xs btn-secondary">Create Dataset</a>');

      const data = extractToMap(extract, htmlDatasets);
      const $row = $(arc.misc.safe_substitute(contentsHtml, data));
      $tableContainer.find('tbody').append($row);

      // Highlight the row received in the URL search parameter
      if (String(extract.id) === arc.globals.urlparams.highlight_extract) {
        $row.addClass('table-primary');
      }

      if (['Still Running', 'Queued'].includes(data.STATE)) {
        runningCount += 1;
      }
    });

    setupCreateDataset();
    setupEdit();
    setupDelete();
    setupRun();
    setupInfo();

    if (runningCount > 0) {
      const nextRefreshDelaySeconds = 5;
      setTimeout(
          () => {
            m.fetchAndPaint();
          },

          nextRefreshDelaySeconds * 1000
      );
    }
  }

  function extractToMap(extract, htmlDatasets) {
    let partitionsHdr = '';
    let partitions = '';
    let aggHdr = '';

    if (extract.extract_config.target && extract.extract_config.target.partitionData) {
      partitions = _collistToStr(extract.extract_config.target.partitionData);
      if (partitions) {
        partitionsHdr = '<br>Partitions: ';
      }
    }

    if (extract.extract_config.source.aggData.length > 0) {
      aggHdr = '<br>Measures: ';
    }

    const dc = dataconnections[extract.tgtDataconnectionId];
    const extractLog = extractLogs[extract.id];

    let statusMessage = '';
    if (extractLog) {
      const taskList = JSON.parse(extractLog.task_list).tasks || [];
      statusMessage = extractLog.error_message || taskList[taskList.length - 1]?.message || '';
    }

    const {
      state,
      htmlState,
      stateLabel,
    } = arc.extracts.stateStringsFromLog(extractLog);

    const { notification_email, rewrite_properties } = extract.extract_config;

    const rewrite_status = rewrite_properties?.rewrite_status ?? 'disabled_until_refreshed';

    return {
      ID: extract.id,
      DATACONNECTION: dc ? dc.name : extract.tgtDataconnectionId,
      TABLE: `${extract.tgtDbname}.${extract.tgtTablename}`,
      HTML_DATASETS: htmlDatasets.join('<br>'),
      STATE: state,
      HTML_STATE: htmlState,
      STATE_LABEL: stateLabel,
      JOB_PAGE_LINK: extract.scheduleId && htmlState !== 'Never run' ? `/arc/jobs/job/${extract.scheduleId}` : '#tab_extracts',
      NEVER_RUN: htmlState === 'Never run' ? 'pointer-events: none;' : '',
      LASTRUN: extractLog ? (extractLog.lastSuccessfulRun || '') : '',
      REWRITE_ENABLED: rewrite_status === 'disabled_until_refreshed' ? 'disabled until refreshed' : rewrite_status,
      SCHEDULE: extract.scheduleGroupName || 'Manual',
      DIMS: _collistToStr(extract.extract_config.source.dimData),
      AGGS: _collistToStr(extract.extract_config.source.aggData),
      HTML_AGGS_HDR: aggHdr,
      HTML_PARTITIONS_HDR: partitionsHdr,
      PARTITIONS: partitions,
      RUNTIME: extractLog ? moment.duration(extractLog.runtime, 'seconds').humanize() : '',
      STATUS_MSG: statusMessage,
      SHOW_PARTITIONS: partitions !== '',
      SHOW_STATUS_MSG: statusMessage !== '',
      EMAIL_SUCCESS: notification_email?.email_success || 'false',
      EMAIL_FAIL: notification_email?.email_fail || 'false',
      EMAIL_TO: String(notification_email?.to_addresses),
      EMAIL_SUBJECT: notification_email?.email_subject,
    };
  }

  function setupCreateDataset() {
    $tableContainer.find('.create-dataset')
      .offOn('click.createDataset', function() {
        const extractId = Number($(this).attr('data-id'));
        const extract = extractIdToExtract[extractId];
        if (!extract) {
          arc.report.alertMessageSmall(`No extract with id ${extractId}`, 'alert-danger');
          return;
        }

        const dc = dataconnections[extract.tgtDataconnectionId];
        if (!dc) {
          arc.report.alertMessageSmall(
            `Could not find data connection with id ${extract.tgtDataconnectionId}`,
            'alert-danger'
          );
          return;
        }

        arc.frontend.legacyBridge.events.next({
          type: 'create-dataset-from-table-modal',
          payload: {
          databaseName: extract.tgtDbname,
          tableName: extract.tgtTablename,
          connection: { id: dc.id, type: dc.type },
          onSuccess: m.fetchAndPaint,
        }
      });
    });
  }

  function extractDeleteConfirmation(extractId) {
    const discardChanges = function() { return 'arc.admin.extractDeleteConfirmation'; };

    discardChanges.hideModal = function() {
      $('#discard-changes-modal').modal('hide');
    };

    discardChanges.askUser = function(confirmedCb) {
      const $modal = $('#discard-changes-modal');
      if (arc.misc.modalShown($modal)) {
        return;
      }

      const modalHtml = `
      <div class="modal-header">
        <h4 class="modal-title">Confirmation</h4>
        <a class="close" data-dismiss="modal">&times;</a>
      </div>
      <div class="modal-body">
        <h4>
          Confirm that you want to delete this extract
          by clicking the "Delete Extract" button.
        </h4>
        <h4>
          Note that deleting this extract will NOT remove the target table.
          If you want to also delete the extract target table,
          select the checkbox below.
        </h4>
        <h4>
          <div class="form-check">
            <input class="form-check-input" type="checkbox" value="" id="deleteExtractTableCheckbox">
            <label class="form-check-label" for="deleteExtractTableCheckbox">
              Delete extract table
            </label>
          </div>
        </h4>
      </div>
      <div class="modal-footer">
        <button type="button" class="btn btn-secondary" data-dismiss="modal">Cancel</button>
        <button type="button" class="btn btn-primary discard-changes-yes">Delete Extract</button>
      </div>
      `;
      $modal.find('.modal-content').html(modalHtml);

      $modal.modal('show');
      const $yes = $modal.find('.discard-changes-yes');

      $yes.offOn('click', () => {
            const deleteTable = $('.modal-content #deleteExtractTableCheckbox').is(':checked');
            discardChanges.hideModal();
            if (deleteTable) {
              const extract = extracts.find((e) => e.id === extractId);
              const tableName = `${extract.tgtDbname}.${extract.tgtTablename}`;

              const deleteTableConfirmation = arc.misc.deleteConfirmation();
                deleteTableConfirmation
                .objectType('Extract Table')
                .objectName(tableName)
                .askUser(() => {
                  deleteTableConfirmation.hideModal();
                  confirmedCb(deleteTable);
                });
            } else {
              confirmedCb(deleteTable);
            }
          });
    };

    return discardChanges;
  }

  function setupDelete() {
    $tableContainer
        .find('.delete')
        .offOn('click.delete', function() {
          const extractId = Number($(this).attr('data-id'));
          const confirmDelete = extractDeleteConfirmation(extractId);

          confirmDelete.askUser((deleteTable) => {
            function successCb() {
              arc.report.alertMessageSmall('The data extract has been deleted', 'alert-success');
              m.fetchAndPaint();
            }

            function failCb(resp) {
              arc.report.alertMessage(`Failed to delete extract: ${resp.responseText || ''}`);
              m.fetchAndPaint();
            }

            deleteJobAndExtract(extractId, successCb, failCb, deleteTable);
          });
        });
  }

  function setupEdit() {
    $tableContainer
        .find('.edit')
        .offOn('click.edit', function() {
          const extractId = Number($(this).attr('data-id'));
          const extract = extractIdToExtract[extractId];
          const scheduleGroupId = extract.scheduleGroupId || 0; // 0 for Manual
          arc.extractCreator()
            .dataExtractManager(m)
            .edit(extract, scheduleGroupId);
        });
  }

  function setupRun() {
    $tableContainer
        .find('.run')
        .offOn('click.run', function() {
          const confirmRun = arc.misc.simpleUserConfirmation({
            confirmTitle: 'Confirmation',
            confirmQus: 'Please confirm that you would like to run this extract ' +
                'by clicking on "Run Extract" below.',
            confirmText: 'Run Extract',
          });

          const extractId = Number($(this).attr('data-id'));
          const { scheduleId } = extractIdToExtract[extractId];

          confirmRun.askUser(() => {
            function successCb() {
              m.fetchAndPaint();
            }

            function failCb() {
              m.fetchAndPaint();
            }

            if (scheduleId) {
              rerunJob(scheduleId, successCb, failCb);
            } else {
              createOrDeleteExtractJob(extractId, undefined, successCb, failCb);
            }
          });
        });
  }

  function setupInfo() {
    const detailSel = d3.select($tableContainer[0])
                        .selectAll('.details');

    function detailParams(d, i, elem) {
      const extractId = Number($(elem).closest('tr').attr('data-id'));
      const extract = extractIdToExtract[extractId];
      const data = extractToMap(extract, []);

      const kvs = [
        {
          key: 'Extract ID',
          value: data.ID,
        },
        {
          key: 'Target Connection',
          value: data.DATACONNECTION,
        },
        {
          key: 'Target Table',
          value: data.TABLE,
        },
        {
          key: 'Last Runtime',
          value: data.RUNTIME,
        },
        {
          key: 'Rewrite',
          value: data.REWRITE_ENABLED,
        },
        {
          key: 'Dimensions',
          value: data.DIMS,
        },
        {
          key: 'Measures',
          value: data.AGGS,
        },
      ];

      if (data.SHOW_PARTITIONS) {
        kvs.push({
          key: 'Partitions',
          value: data.PARTITIONS,
        });
      }

      kvs.push({
        key: 'State',
        value: data.STATE,
      });

      if (data.SHOW_STATUS_MSG) {
        kvs.push({
          key: 'Last status',
          value: data.STATUS_MSG,
        });
      }

      kvs.push({
        key: 'Email success',
        value: data.EMAIL_SUCCESS,
      });

      kvs.push({
        key: 'Email fail',
        value: data.EMAIL_FAIL,
      });

      let emailToVal = '-';
      let emailSubjectVal = '-';
      if (data.EMAIL_SUCCESS === 'true' || data.EMAIL_FAIL === 'true') {
        emailToVal = data.EMAIL_TO;
        emailSubjectVal = data.EMAIL_SUBJECT;
      }

      kvs.push({
        key: 'Email to',
        value: emailToVal,
      });

      kvs.push({
        key: 'Email subject',
        value: emailSubjectVal,
      });

      return kvs;
    }

    arc.marks()
       .addSelection(detailSel)
       .params_fn(detailParams)
       .text_fn((d, i, elem) => arc.misc.tooltipFromKVs(undefined, detailParams(d, i, elem)))
       .marker_name('details')
       .activate();

    const labelSel = d3.select($tableContainer[0])
                  .selectAll('.badge');

    function labelParams(d, i, elem) {
       const extractId = Number($(elem).closest('tr').attr('data-id'));
       const extract = extractIdToExtract[extractId];
       const data = extractToMap(extract, []);

       const kvs = [];

       if (data.STATUS_MSG) {
         kvs.push({
           key: 'Last Message',
           value: data.STATUS_MSG,
         });
       }

       return kvs;
    }

    arc.marks()
       .addSelection(labelSel)
       .params_fn(labelParams)
       .text_fn((d, i, elem) => arc.misc.tooltipFromKVs(undefined, labelParams(d, i, elem)))
       .marker_name('label')
       .activate();
  }

  function fetch(successCb, failCb, reset_controller = false) {
    const datasetId = builder.dataset().dataset_id();

    let extract_url = `${arc.frontend.globals.urls.extractUrl}?dataset_id=${datasetId}&details=true`;
    if (reset_controller) {
      extract_url += '&reset_controller=true';
    }

    const assets = [
      {
        asset: arc.frontend.globals.urls.datasetUrl,
        postprocess: d3.json,
      },
      {
        asset: extract_url,
        postprocess: d3.json,
      },
      {
        asset: `${arc.frontend.globals.urls.dataCopyLogsUrl}?srcDatasetId=${datasetId}`,
        postprocess: d3.json,
      },
    ];

    // Always re-fetch datasets, extracts and job logs
    // - this page can be used to create/delete extracts and to create datasets, so we have
    //   to make sure we always fetch the latest state
    arc.staticfetcher.forget(assets);

    // Always use cached dataconnections by adding this to assets after the
    // call to 'forget'
    // - dataconnection do not change that often, and if they do, require
    //   a page reload before those changes are reflected here
    assets.push({
      asset: arc.frontend.globals.urls.dataconnectionUrl,
      postprocess: d3.json,
    });

    arc.staticfetcher.fetch(assets,
        (error, _datasets, _extracts, _extractLogs, _dataconnections) => {
      if (error) {
        arc.report.alertMessageSmall('Error fetching datasets', 'alert-danger');
        if (failCb) {
          failCb(error);
        }

        return;
      }

      extracts = _extracts;
      extractIdToExtract = Object.fromEntries(extracts.map((e) => [e.id, e]));
      if (_extractLogs) {
        extractLogs = Object.fromEntries(_extractLogs.map((e) => [e.extractId, e]));
      }

      dataconnections = {};
      _dataconnections.forEach((dc) => {
        dataconnections[dc.id] = dc;
      });

      tablesToDataset = {};
      datasetIdToName = {};
      _datasets.forEach((ds) => {
        datasetIdToName[ds.id] = ds.name;

        const tablenames = arc.datasets.datasetObjectToTableNames(ds);
        tablenames.forEach((tablename) => {
          const dskey = `${ds.dc_id}:${tablename}`;
          if (tablesToDataset[dskey]) {
            tablesToDataset[dskey].push(ds);
          } else {
            tablesToDataset[dskey] = [ds];
          }
        });
      });

      successCb();
    });
  }

  return m;
};

arc.extractCreator = function() {
  const c = function() { return 'arc.extractCreator'; };

  let dataExtractManager;

  const $dataConnectionSelect = $('.extract-dataconnection');
  const $dbnameSelect = $('#extract-db-name');
  const $hdrTemplate = $('#createview-column-hdr-template');
  const $colTemplate = $('#createextract-column-template ul');
  const $createExtractSchedules = $('#extract-refresh');

  const $modal = $('#create-extract-modal');
  const $form = $('#create-extract-form');
  const $partitionContainer = $('#create-extract-parts');
  const partitionDivSel = '#create-extract-partsholder';
  const $partitionDiv = $(partitionDivSel);
  const $targetPartitions = $('#use-target-partitions');
  const $submitExtract = $('#submit-extract-confirmation');
  const $tgtDataconnection = $('#create-extract-dataconnection select');
  const $tgtDbname = $('#extract-db-name');
  const $tgtTablename = $('#create-extract-name input');
  const $emailNotification = $('#email-notification-type');
  const $emailUponSuccess = $emailNotification.find('#email-upon-success');
  const $emailUponFailure = $emailNotification.find('#email-upon-failure');
  const $emailTo = $('#extract-advanced-tab').find('#email-to');
  const $emailSubject = $('#extract-advanced-tab').find('#extract-email-subject');
  const $emailToGroup = $('#extract-advanced-tab').find('.email-to');
  const $emailSubjectGroup = $('#extract-advanced-tab').find('.email-subject');
  const $targetPartitionsDiv = $targetPartitions.closest('.form-check');
  const $dimsAggsPartitionsDiv = $form.find('.createview-attributes');
  const $queryRewriteEnabled = $form.find('#query-rewrite-status');
  const searchBoxSel = '#create-extract-form .col-searchbox';
  const aggDivSel = '#create-extract-aggs';
  const dimDivSel = '#create-extract-dims';

  let builder;
  let dataconnections;
  let schedules;
  const defaultEmailSubject = `[<<extract_status>>] ${arc.site.settings.email_prefix().value()} Data Extract: <<extract_id>>`;

  c.dataExtractManager = function(_) {
    if (!arguments.length) return dataExtractManager;
    dataExtractManager = _;
    return c;
  };

  c.create = function() {
    const srcDsId = dataExtractManager.builder().dataset().dataset_id();
    setupAndPaintModal(false, srcDsId);
  };

  c.edit = function(extract, scheduleGroupId, fetchAndRefreshJobsTable) {
    setupAndPaintModal(true,
        extract.srcDatasetId,
        extract,
        scheduleGroupId,
        fetchAndRefreshJobsTable);
  };

  function getDimAggPart() {
    const dataset = builder.dataset();
    const $dimDiv = $(dimDivSel);
    const $aggDiv = $(aggDivSel);
    const $dimInputs = $dimDiv.find('.table-list input:checked');
    const $aggInputs = $aggDiv.find('.table-list input:checked');
    const $dimAggInputs = $dimInputs.add($aggInputs);
    const $partitionInputs = $partitionDiv.find('.table-list input:checked');
    const tgtDataconnectionId = Number($tgtDataconnection.val());

    const dc = dataconnections[tgtDataconnectionId];
    if (!dc) {
      return {
        error: `Invalid data connection ${tgtDataconnectionId}`,
      };
    }

    const tgtSupportsPartitions = arc.dataconnection.supportsPartitions(dc.type);

    const dimData = [];
    const aggData = [];
    $dimAggInputs.each(function() {
      const $d = $(this);
      const colname = $d.attr('data-value');

      if (colname === 'Record Count') {
          aggData.push({
            expr: 'sum(1)',
            alias: colname,
          });

        return;
      }

      const $selectedAggLi = $d.parent().find('#agg-funcs li.selected');
      if ($selectedAggLi.length) {
        const aggAction = $selectedAggLi.find('a').attr('agg-action');
        const colAliasExpr = getAggColAliasExprPerConn(aggAction, colname);
        aggData.push({
          expr: colAliasExpr.expr,
          alias: colAliasExpr.alias,
        });
      } else {
        dimData.push({
          expr: `[${colname}]`,
          alias: colname,
        });
      }
    });

    const partitionData = [];
    $partitionInputs.each(function() {
      const $d = $(this);
      const colname = $d.attr('data-value');

      if (tgtSupportsPartitions && $targetPartitions.prop('checked')) {
        partitionData.push({
          expr: `[${colname}]`,
          alias: colname,
        });
      } else {
        dimData.push({
          expr: `[${colname}]`,
          alias: colname,
        });
      }
    });

    return {
      dimData,
      aggData,
      partitionData,
    };
  }

  function doCreate() {
    // save default email settings if email settings are hidden when emailing is disabled
    const emailSuccess = $emailUponSuccess.is(':checked') || false;
    const emailFail = $emailUponFailure.is(':checked') || false;
    const emailTo = $emailTo.length !== 0 ? $emailTo.tokenfield('getTokens').map((em) => em.value) : [];
    const emailSubject = $emailSubject.val() || defaultEmailSubject;
    const r = getDimAggPart();
    if (r.error) {
      arc.report.alertMessageSmall(r.error, 'alert-danger');
      return;
    }

    const dataset = builder.dataset();

    const {
      dimData,
      aggData,
      partitionData,
    } = r;

    let extract_type;
    if (partitionData.length === 0) {
      extract_type = 'refresh_full';
    } else {
      extract_type = 'refresh_partitioned';
    }

    const extract = Extract({
      srcDatasetId: dataset.dataset_id(),
      tgtDataconnectionId: Number($tgtDataconnection.val()),
      tgtDbname: $tgtDbname.val(),
      tgtTablename: $tgtTablename.val(),
      dimData,
      aggData,
      partitionData,
      extract_type,
      emailSuccess: String(emailSuccess),
      emailFail: String(emailFail),
      emailTo,
      emailSubject,
      rewriteStatus: $queryRewriteEnabled.find('input[name=rewrite_status]:checked').val()
    });

    if ($submitExtract.hasClass('disabled')) {
      // Someone clicked on the button twice in quick succession
      // and we are processing the second click while processing
      // the first one

      /* eslint-disable no-console */
      console.log('multiple clicks');
      /* eslint-enable no-console */

      return;
    }

    $submitExtract.addClass('disabled');

    function successCb(extractData) {
      const extractId = extractData.id;
      const scheduleId = Number($createExtractSchedules.val());

      if (scheduleId === 0) {
        // No schedule to create - we are done
        $modal.modal('hide');
        dataExtractManager.fetchAndPaint();
        $submitExtract.removeClass('disabled');
        return;
      }

      createOrDeleteExtractJob(extractId, scheduleId, () => {
            // success
            $modal.modal('hide');
            dataExtractManager.fetchAndPaint();
            $submitExtract.removeClass('disabled');
          }, () => {
            // failure to create schedule. Note that we previously created the extract,
            // so this is not a complete failure
            $modal.modal('hide');
            arc.report.alertMessage('Failed to create scheduled job for refreshing extract');
            dataExtractManager.fetchAndPaint();
            $submitExtract.removeClass('disabled');
          });
    }

    function failCb(resp) {
      arc.report.alertMessage('Failed to create extract', resp.responseText || '', { alertContainerSelector: '#create-extract-modal' });
      $submitExtract.removeClass('disabled');
    }

    extract.create(successCb, failCb);
  }

  function doEdit(extract, fetchAndRefreshJobsTable) {
    const $confirmEditModal = $('#edit-extract-confirmation-modal');
    if (arc.misc.modalShown($confirmEditModal)) {
      return;
    }

    let emailSuccess;
    let emailFail;
    let emailTo;
    let emailSubject;

    if (arc.globals.email_enabled) {
      emailSuccess = $emailUponSuccess.is(':checked');
      emailFail = $emailUponFailure.is(':checked');
      emailTo = $emailTo.tokenfield('getTokens').map((em) => em.value);
      emailSubject = $emailSubject.val();
    } else {
      // save saved email settings if email settings are hidden when emailing is disabled
      const notificationEmail = extract.extract_config.notification_email;
      emailSuccess = notificationEmail.email_success;
      emailFail = notificationEmail.email_fail;
      emailTo = notificationEmail.to_addresses;
      emailSubject = notificationEmail.email_subject;
    }

    const r = getDimAggPart();
    if (r.error) {
      arc.report.alertMessageSmall(r.error, 'alert-danger');
      return;
    }

    const {
      dimData,
      aggData,
      partitionData,
    } = r;

    let extract_type;
    if (partitionData.length === 0) {
      extract_type = 'refresh_full';
    } else {
      extract_type = 'refresh_partitioned';
    }

    const editData = {
      tgtDataconnectionId: Number($tgtDataconnection.val()),
      tgtDbname: $tgtDbname.val(),
      tgtTablename: $tgtTablename.val(),
      dimData,
      aggData,
      partitionData,
      extract_type,
      emailSuccess: String(emailSuccess),
      emailFail: String(emailFail),
      emailTo,
      emailSubject,
      rewriteStatus: $queryRewriteEnabled.find('input[name=rewrite_status]:checked').val()
    };

    const $editExtractWarning = $('#edit-extract-warning');
    $editExtractWarning.hide();
    if (editData.tgtDataconnectionId !== extract.tgtDataconnectionId ||
        editData.tgtDbname !== extract.tgtDbname ||
        editData.tgtTablename !== extract.tgtTablename ||
        areColsListsNotEq(editData.aggData, extract.extract_config.source.aggData) ||
        areColsListsNotEq(editData.dimData, extract.extract_config.source.dimData) ||
        areColsListsNotEq(editData.partitionData, extract.extract_config.target.partitionData)) {
      $editExtractWarning.show();
    }

    function areColsListsNotEq(cols1, cols2) {
      const originalNewCols = JSON.parse(JSON.stringify(cols1));
      const sortedExprs1 = cols1.map((col) => col.expr).sort();
      const sortedExprs2 = cols2.map((col) => col.expr).sort();
      return sortedExprs1.length !== sortedExprs2.length ||
        sortedExprs1.some((expr, idx) => {
          const oldExpr = sortedExprs2[idx];
          if (expr === 'sum(1)') {
            const oldALias = cols2.find((col) => col.expr === 'sum(1)')?.alias;
            if (oldALias) {
              // keep old alias of cli extract to prevent target table deletion
              // cli extract alias -- sum(1)
              // UI extract alias -- Record Count
              cols1.find((col) => col.expr === 'sum(1)').alias = oldALias;
              return false;
            }
          }

          // new expr will have [] around aggregated column names, but if extract is created
          // via cli then names are not surrounded by [],
          // if expressions match without [] then they are the same and we should't
          // delete existing target table
          if (expr.replace(/[[\]]/g, '') === oldExpr) {
            cols1.find((col) => col.expr === expr).expr = oldExpr;
            return false;
          }

          // now we use shelf elem funtion to get alias and expr, however there might be
          // saved extracts without space after count(distinct
          if (expr.replace('count(distinct ', 'count(distinct') === oldExpr) {
            return false;
          }

          // previously created agg extracts alias contain only colname instead of agg_fn(colname)
          // if expr match we shouldn't update alias and shouldn't delete existing target table
          if (expr === oldExpr) {
            const oldAlias = cols2.find((col) => col.expr === expr).alias;
            cols1.find((col) => col.expr === expr).alias = oldAlias;
            return false;
          }

          // Old and new expresions aren't the same.
          // Keep new aliases because at least one column has changed
          // than we better use new aliases from now on
          cols1 = originalNewCols;
          return true;
        });
    }

    const $yes = $confirmEditModal.find('.discard-changes-yes');
    $yes.offOn('click', () => {
      $confirmEditModal.modal('hide');

      // do edit
      if ($submitExtract.hasClass('disabled')) {
        // Someone clicked on the button twice in quick succession
        // and we are processing the second click while processing
        // the first one

        /* eslint-disable no-console */
        console.log('multiple clicks');
        /* eslint-enable no-console */

        return;
      }

      $submitExtract.addClass('disabled');

      function successCb(extractId) {
        const scheduleId = Number($createExtractSchedules.val());
        createOrDeleteExtractJob(
            extractId,
            scheduleId > 0 ? scheduleId : undefined, // undefined for Manual
            () => {
              // success
              $modal.modal('hide');
              arc.report.alertMessageSmall('Successfully edited extract', 'alert-success');
              dataExtractManager?.fetchAndPaint();
              fetchAndRefreshJobsTable?.();
              $submitExtract.removeClass('disabled');
            },
            () => {
              // failure to create schedule. Note that we previously edited the extract,
              // so this is not a complete failure
              $modal.modal('hide');
              arc.report.alertMessage('Failed to edit scheduled job for refreshing extract');
              dataExtractManager?.fetchAndPaint();
              fetchAndRefreshJobsTable?.();
              $submitExtract.removeClass('disabled');
            }
          );
      }

      function failCb(resp) {
        arc.report.alertMessage('Failed to edit extract', resp.responseText ? resp.responseText : '', { alertContainerSelector: '#create-extract-modal' });
        $submitExtract.removeClass('disabled');
      }

      editExtract(extract.id, editData, successCb, failCb);
    });

    $confirmEditModal.modal('show');

    // add backdrop on top of Edit Extract modal from which Confirmation modal is opened
    $('.modal-backdrop').each((index, elem) => {
      $(elem).css('z-index', `${(1040 + index * 20)}`);
    });
    $confirmEditModal.css('z-index', '1070');
  }

  function setupFormValidator() {
    // links all dim and agg checkboxes' errors
    const $dimAggAttrErrVal = $('#create-extract-attr-error-val');
    // validate there is at least one dim or agg checked
    $dimsAggsPartitionsDiv.offOn('change', () => {
      const r = getDimAggPart();
      if (r.error) {
        $dimAggAttrErrVal.val('');
      } else if (r.dimData.length || r.aggData.length) {
        $dimAggAttrErrVal.val('valid');
      } else {
        $dimAggAttrErrVal.val('');
      }

      $dimAggAttrErrVal.trigger('change.bs.validator');
    });

    const $submitExtractTooltip = $submitExtract.parent();
    $submitExtractTooltip.offOn('mouseover.tooltip', () => {
      if ($submitExtract.hasClass('disabled')) {
        $submitExtractTooltip.tooltip('enable');
      } else {
        $submitExtractTooltip.tooltip('disable');
      }
    });

    $form.validator({
      custom: {
        requiredemail: ($el) => {
          const formGroup = $el.closest('.col-sm-10');
          const helptext = $el.parent().siblings('.with-errors');
          if ($emailNotification.find('input:checkbox:checked').length > 0) {
            const hasEmail = $el.tokenfield('getTokens').length;
            const validEmail = $el.tokenfield('getTokens').every((email) => /\S+@\S+\.\S+/.test(email.value));

            // Hack to validate and show error message given token inputs
            if (hasEmail) {
              if (!validEmail) {
                formGroup.addClass('has-error');
                return ' ';
              }

              formGroup.removeClass('has-error');
              helptext.text('');
            } else {
              formGroup.addClass('has-error');
              return ' ';
            }
          } else {
            formGroup.removeClass('has-error');
            helptext.text('');
          }
        },
      },
    });
  }

  function setupTargetTablePrefix() {
    const tablePrefix = arc.globals.data_extracts_target_tablename_prefix;
    if (tablePrefix) {
      const $tablenameContainer = $('#create-extract-name');
      $tablenameContainer.find('input')
                         .attr('pattern', `${tablePrefix}.*`);
      $tablenameContainer.find('span.prefix')
                         .text(` (starting with ${tablePrefix})`);
    }
  }

  function setupAndPaintModal(isEdit,
        srcDatasetId,
        extract,
        scheduleGroupId,
        fetchAndRefreshJobsTable) {
    if (arc.misc.modalShown($modal)) {
      return;
    }

    const assets = [
      {
        asset: arc.frontend.globals.urls.dataconnectionUrl,
        postprocess: d3.json,
      },
      {
        asset: arc.frontend.globals.urls.scheduleGroupUrl,
        postprocess: d3.json,
      },
      {
        asset: `${arc.frontend.globals.urls.datasetUrl}/${srcDatasetId}?detail=true`,
        postprocess: d3.json,
      },
    ];

    arc.staticfetcher.fetch(assets, (error, _dataconnections, _schedules, _dataset) => {
      if (error) {
        arc.report.alertMessageSmall('Error fetching dataconnections', 'alert-danger');
        return;
      }

      dataconnections = {};
      _dataconnections.forEach((dc) => {
        dataconnections[dc.id] = dc;
      });
      paintDataConnections();

      schedules = _schedules;
      paintSchedules();

      const dataset = arc.datasets.Dataset()
                          .dataset_id(srcDatasetId)
                          .init_from_fetch(_dataset[0]);
      builder = arc.reportbuilder().dataset_id(srcDatasetId).dataset(dataset);

      $(searchBoxSel).val('');
      if (!isEdit) { // create
        // reset the inputs
        $tgtTablename.val('');
        $targetPartitions.prop('checked', true);
        $createExtractSchedules.val('0');
        $queryRewriteEnabled.find('input[value=disabled_until_refreshed]').prop('checked', true);
      }

      paintDimAgg();
      paintEmailSettings();
      setupTargetTablePrefix();
      setupFormValidator();
      toggleExtFormTitleSubmitBtn(isEdit ? 'Edit Data Extract' : 'Create Data Extract');

      if (isEdit) {
        // fill in extract details
        $form.find('[data-toggle="popover"]').popover();
        setFormData(extract, scheduleGroupId);
      }
    });

    $form.offOn('submit', (event) => {
      if (event.isDefaultPrevented()) {
        return;
      }

      event.preventDefault();
      event.stopPropagation();
      if (isEdit) {
        doEdit(extract, fetchAndRefreshJobsTable);
      } else {
        doCreate();
      }
    });

    $modal.modal('show');
    arc.misc.handleAlertContainerActivationOnModal($modal);
  }

  function setFormData(extract, scheduleGroupId) {
    $createExtractSchedules.val(scheduleGroupId);

    $dataConnectionSelect.val(extract.tgtDataconnectionId);
    $dataConnectionSelect.trigger('chosen:updated');

    checkPartitionSupportAndLoadDatabases(extract.tgtDbname);

    $tgtTablename.val(extract.tgtTablename);
    // if previous modal had wrong tgt table name,
    // error was still there in new modal without following ln
    $tgtTablename.trigger('change.bs.validator');

    extract.extract_config.source.dimData.forEach((dim) => {
      const colname = dim.alias;

      // dimData can contain not only dimensions but non-aggregated measures and partitions
      // where $targetPartitions wasn't checked
      const $colsInput = $dimsAggsPartitionsDiv.find(`.table-list input[data-value="${colname}"]`);

      if ($colsInput.length <= 1 && !$colsInput.prop('checked')) { // first column with this name
        $colsInput.prop('checked', true);
      } else { // duplicate col
        const columnObj = arc.datasets.Column().init_from_dsobject(dim);
        const $newColLi = $(arc.misc.safe_substitute(
          $colTemplate.html(),
          {
            DISPLAY_NAME: colname,
            VALUE: colname,
            CDT_field_wrapper: '',
            HTML_CDT_expand: '',
            HTML_coltype_class: columnObj.coltype_class(),
            AlIGN_attr_content_wrapper: 'form-check',
          }
        ));

        const $newColInput = $newColLi.find('input.aview-column-checkbox');
        $newColInput.prop('checked', true);

        $newColLi.find('.remove-col').parent().show();

        filterAggsPerDCtype($newColLi);

        $colsInput.last().closest('.field-wrapper').parent().after($newColLi);
      }
    });

    extract.extract_config.source.aggData.forEach((agg) => {
      // extract created by cli has alias sum(1) for expr sum(1)
      if (agg.expr === 'sum(1)') {
        $dimsAggsPartitionsDiv.find(`.table-list input[data-value="Record Count"]`).prop('checked', true);
      } else {
        const [colname, aggAction] = getColNameAndAggAction(agg.alias, agg.expr);

        // Previously created extract's alias has column name only instead of agg_fn(colname).
        // Get col label from expr
        const colLabel = getAggColAliasExprPerConn(aggAction, colname).alias;

        const $colsInput = $dimsAggsPartitionsDiv.find(`.table-list input[data-value="${colname}"]`);
        const columnObj = arc.datasets.Column().init_from_dsobject(agg);
        if ($colsInput.length <= 1 && !$colsInput.prop('checked')) { // first column with this name
          $colsInput.prop('checked', true);

          // set agg function
          const $aggAction = $colsInput.closest('.attr-content-wrapper').find(`[agg-action="${aggAction}"]`);
          const $aggLi = $aggAction.parent();
          const $colRow = $colsInput.closest('.attr-content-wrapper');
          const $icon = $colRow.find('.dim-icon-type');

          $aggLi.addClass('selected');
          $aggAction.closest('.properties-menu > div').addClass('property-selected');
          $colRow.find('label').text(colLabel);

          const classList = $icon.attr('class').split(' ');
          classList.forEach((elem) => {
            if (elem.startsWith('type-')) $icon.removeClass(elem);
          });
          $icon.addClass(columnObj.coltype_class());
        } else { // duplicate col
          const $newColLi = $(arc.misc.safe_substitute(
            $colTemplate.html(),
            {
              DISPLAY_NAME: colLabel,
              VALUE: colname,
              CDT_field_wrapper: '',
              HTML_CDT_expand: '',
              HTML_coltype_class: columnObj.coltype_class(),
              AlIGN_attr_content_wrapper: 'form-check',
            }
          ));

          const $newColInput = $newColLi.find('input.aview-column-checkbox');
          $newColInput.prop('checked', true);

          // set agg function
          const $aggAction = $newColLi.find(`[agg-action="${aggAction}"]`);
          const $aggLi = $aggAction.parent();
          $aggLi.addClass('selected');
          $aggAction.closest('.properties-menu > div').addClass('property-selected');

          $newColLi.find('.remove-col').parent().show();

          filterAggsPerDCtype($newColLi);

          $colsInput.last().closest('.field-wrapper').parent().after($newColLi);
        }
      }
    });

    if (extract.extract_config.target.partitionData.length > 0) {
      $targetPartitions.prop('checked', true);
      extract.extract_config.target.partitionData.forEach((p) => {
        $(`${partitionDivSel} .table-list input[data-value="${p.alias}"]`).prop('checked', true);
      });
    } else {
      // unpartitioned extract - indicate it by making sure the checkbox is unset
      $targetPartitions.prop('checked', false);
    }

    // trigger validator -- there is at least one dim or agg checked
    $dimsAggsPartitionsDiv.trigger('change');

    // notification email details
    const notificationEmail = extract.extract_config.notification_email;
    if (notificationEmail) {
      $emailUponSuccess.prop('checked', notificationEmail.email_success === 'true');
      $emailUponFailure.prop('checked', notificationEmail.email_fail === 'true');
      $emailTo.tokenfield('setTokens', notificationEmail.to_addresses);
      $emailSubject.val(notificationEmail.email_subject);
    }

    handleEmailNotificationChange();

    const rewrite_status = extract.extract_config.rewrite_properties?.rewrite_status ?? 'disabled_until_refreshed';
    $queryRewriteEnabled.find(`input[value=${rewrite_status}]`).prop('checked', true);
  }

  function handleEmailNotificationChange() {
    if ($emailNotification.find('input:checkbox:checked').length > 0) {
      $emailToGroup.removeClass('disabled-feature');
      $emailTo.prop('disabled', false);

      $emailSubjectGroup.removeClass('disabled-feature');
      $emailSubject.attr('required', true);
    } else {
      $emailToGroup.addClass('disabled-feature');
      $emailTo.prop('disabled', true);

      $emailSubjectGroup.addClass('disabled-feature');
      $emailSubject.attr('required', false);
    }

    $emailTo.trigger('change.bs.validator');
    $emailSubject.trigger('change.bs.validator');
  }

  function paintEmailSettings() {
    $emailNotification.offOn('change', () => {
      handleEmailNotificationChange();
    });

    // email input field with typeahead and tokens
    const engine = new Bloodhound({
      // we don't need the local, but removing it breaks the tokenizer
      local: arc.globals.email_addresses?.map((d) => ({
          value: d,
        })),

      datumTokenizer(d) {
        return Bloodhound.tokenizers.whitespace(d.value);
      },

      queryTokenizer: Bloodhound.tokenizers.whitespace,
    });

    engine.initialize();

    $modal.find('.email-addresses').tokenfield({
      inputType: 'text',
      createTokensOnBlur: true,
      typeahead: [
          null,
          {
            source: engine.ttAdapter(),
          },
      ],
    });

    $emailNotification.find('input:checkbox').prop('checked', false);
    $emailNotification.trigger('change');
    $emailTo.tokenfield('setTokens', []);
    $emailSubject.val(defaultEmailSubject);
  }

  function paintDataConnections() {
    $dataConnectionSelect.empty();
    $dataConnectionSelect.append('<option selected disabled hidden></option>'); // No option is selected initially
    Object.values(dataconnections).forEach((dc) => {
      if (!dc.perms.includes('dc_aviews')) {
        return;
      }

      const option = arc.misc.safe_substitute(
          '<option value="%VALUE%" data-perms="%PERMS%">%NAME% </option>',
          {
            NAME: dc.name,
            VALUE: dc.id,
            PERMS: dc.perms.join(' '),
          }
      );

      $dataConnectionSelect.append(option);

      if (!arc.dataconnection.supportsDataExtractsTarget(dc.type) || (dc.type === 'sqlite' &&
        arc.globals.arcviz_settings.CDV_MODE === 'cdw')) {
        $dataConnectionSelect.find(`[value=${dc.id}]`).attr('disabled', true);
      }
    });

    $dataConnectionSelect.chosen({
      disable_search_threshold: 5
    });

    $dataConnectionSelect.parent().find('.chosen-container').offOn('click', addTooltip);
    $dataConnectionSelect.parent().find('.chosen-container .chosen-search-input')
                                  .offOn('keyup.tooltipAdder', addTooltip);

    function addTooltip() {
      Object.values(dataconnections).forEach((dc) => {
        let popoverText;

        if (dc.type === 'sqlite' && arc.globals.arcviz_settings.CDV_MODE === 'cdw') {
          popoverText = 'SQLite data connection is read-only in Cloudera Data Warehouse';
        } else if (!arc.dataconnection.supportsDataExtractsTarget(dc.type)) {
          popoverText = `${dc.type} data connection does not support extracting data`;
        }

        const $dcli = $dataConnectionSelect.parent()
                      .find(`.chosen-container li:contains("${dc.name}")`)
                      .filter(function() {
                          return $.trim($(this).text()) === dc.name;
                        });

        if (popoverText && !$dcli.find('span').length) {
          $dcli.append(`<span class="info-icon" data-toggle="popover" data-content="${popoverText}"></span>`)
            .offOn('click', '.info-icon', (e) => {
              e.preventDefault();
            })
            .find('[data-toggle="popover"]')
            .popover({
              boundary: 'window',
              container: 'body',
              placement: 'top',
              trigger: 'hover'
            });
        }
      });
    }

    $dataConnectionSelect.trigger('chosen:updated');

    $dataConnectionSelect.offOn('change', () => {
      checkPartitionSupportAndLoadDatabases();

      // trigger dims/aggs validator -- getDimAggPart() used in dims/aggs
      // validator and returns invalid dc error if no dc selected yet
      $dimsAggsPartitionsDiv.trigger('change');
    });

    disableDbnameSelect();
  }

  function paintDimAgg() {
    const $dimDiv = $(dimDivSel);
    const $aggDiv = $(aggDivSel);
    const $searchbox = $(searchBoxSel);
    const dataset = builder.dataset();

    $dimDiv.empty();
    $aggDiv.empty();
    $partitionDiv.empty();

    builder.paintAttributes(dataset,
        {
          $dimDiv,
          $aggDiv,
          $partDiv: $partitionDiv,
          headersOnly: true,
          selectAllOption: true,
        });

    const templateCol = $colTemplate.html();

    function paintCol(col) {
      const isCdt = col.isComplexType();
      const colname = col.display_name();
      const $li = $(arc.misc.safe_substitute(
                  templateCol,
                  {
                    DISPLAY_NAME: colname,
                    VALUE: colname,
                    CDT_field_wrapper: isCdt ? 'cdt-field-wrapper' : '',
                    HTML_CDT_expand: isCdt ? '<a class="link"><i class="fa fa-angle-down expand-icon"></i></a>' : '',
                    HTML_coltype_class: col.coltype_class(),
                    AlIGN_attr_content_wrapper: isCdt ? 'align-items-center' : 'form-check',
                  }
                ));

      if (isCdt) {
        $li.find('.aview-column-checkbox').remove();
        $li.find('.dropdown').remove();
      }

      if (colname === 'Record Count') {
        $li.find('.dropdown').remove();
      }

      return $li;
    }

    function paintColumns($container, columns) {
      const cdt = {};
      columns.forEach((col) => {
        const $li = paintCol(col);
        if (col.isComplexType()) {
          cdt[col.attr('name')] = {
            col,
            $li,
          };
        }

        const parent = col.getComplexParent();
        if (parent) {
          const parent_name = parent.display_name();
          if (cdt[parent_name]) {
            cdt[parent_name].$li.find('.child-table-list').first().append($li);
          }
        } else {
          $container.append($li);
        }
      });
    }

    const templateHdr = $hdrTemplate.html();
    $dimDiv.append(templateHdr);
    $aggDiv.append(templateHdr);
    $partitionDiv.append(templateHdr);

    $aggDiv.find('.table-list').append(paintCol(dataset.getRecordCount()));

    dataset.tables().forEach((table) => {
      paintColumns($dimDiv.find('.table-list'),
          table.dimensionsOnly()
               .filter((col) => (!col.attr('name') || !col.attr('ispart')) &&
                   !col.attr('ispartition') && !col.isHierarchy()));

      paintColumns($aggDiv.find('.table-list'),

          // Note that we are only plotting measures by calling 'measuresOnly()'
          // and ignoring hierarchies. For backward compatibility reasons, aggregates()
          // returns hierarchies as well.
          table.measuresOnly()
               .filter((col) => !col.attr('ispartition')));

      paintColumns(
          $partitionDiv.find('.table-list'),
          table.base_columns()
               .filter((col) => col.attr('ispart') || col.attr('ispartition'))
        );

      return true;
    });

    $partitionDiv.find('.table-list li .dropdown').remove();
    const $partitionCols = $partitionDiv.find('.table-list-value');
    if ($partitionCols.length === 0) {
      $partitionContainer.hide();
    } else {
      $partitionContainer.show();
    }

    filterAggsPerDCtype($form);

    function updateIconTypeClass($icon, newTypeClass) {
      const classList = $icon.attr('class').split(' ');
      classList.forEach((elem) => {
        if (elem.startsWith('type-')) $icon.removeClass(elem);
      });
      $icon.addClass(newTypeClass);
    }

    $dimsAggsPartitionsDiv.offOn('click', '.col-action', (event) => {
      const $this = $(event.target);
      if ($this.hasClass('agg-fn')) {
        const $parent = $this.parent();
        const $colRow = $this.closest('.attr-content-wrapper');
        const $icon = $colRow.find('.dim-icon-type');
        const colName = $colRow.find('input').attr('data-value');

        if ($parent.hasClass('selected')) {
          $parent.removeClass('selected');
          $this.closest('.properties-menu > div').removeClass('property-selected');

          $colRow.find('label').text(colName);

          updateIconTypeClass($icon, $icon.attr('data-original-coltype'));
        } else {
          $parent.parent().find('li').removeClass('selected');
          $parent.addClass('selected');
          $this.closest('.properties-menu > div').addClass('property-selected');

          const aggAction = $this.attr('agg-action');
          const aggColAliasExprPerConn = getAggColAliasExprPerConn(aggAction, colName);
          const label = aggColAliasExprPerConn.alias;
          $colRow.find('label').text(label);

          updateIconTypeClass($icon, aggColAliasExprPerConn.iconType);
        }
      } else if ($this.hasClass('duplicate-col')) {
        const $col = $this.closest('.field-wrapper').parent();
        const $clone_col = $col.clone();
        $clone_col.find('.remove-col').parent().show();
        $col.after($clone_col);
        $clone_col.find('.dropdown > a').dropdown('toggle');
      } else if ($this.hasClass('remove-col')) {
        const $col = $this.closest('.field-wrapper').parent();
        $col.remove();
      } else if ($this.hasClass('table-list-value')) {
        $this.closest('.attr-content-wrapper').find('input').trigger('click');
      }
    });

    $form.find('.table-list-value')
         .addClass('noselect');

    // reset dim and agg error input to ''
    $dimsAggsPartitionsDiv.trigger('change');

    // setup search for columns
    const searchConfig = {
      $searchbox,
      search_fn() {
        clearTimeout(searchConfig.maxTimeId);
        searchConfig.maxTimeId = undefined;

        const searchTerm = $searchbox.val().toLowerCase() || '';
        $('.createview-attributes .table-list-value').each(function() {
          if (this.textContent.trim().toLowerCase().search(searchTerm) >= 0) {
            $(this).closest('li').show();
          } else {
            $(this).closest('li').hide();
          }
        });
      },
    };

    arc.report.triggerOnSearch(searchConfig);
  }

  function filterAggsPerDCtype($colsDiv) {
    const dctype = builder.dataset().dataconnection_type();
    const isArcengineCompatible = arc.dataconnection.isArcEngineOrImpala(dctype);

    $colsDiv
      .find('[data-conn]')
      .each(function() {
        const $this = $(this);
        const supportedConns = $this.attr('data-conn').split(' ');
        if (!(supportedConns.indexOf(dctype) >= 0)) {
          $this.addClass('arc-invisible');
        }
      });

    if (!isArcengineCompatible) {
      $colsDiv.find('[data-conn="arcengine"]')
        .addClass('arc-invisible', true);
    }

    $colsDiv
      .find('[data-conn-exclude]')
      .each(function() {
        const $this = $(this);
        const notSupportedConns = $this.attr('data-conn-exclude').split(' ');
        if (notSupportedConns.indexOf(dctype) >= 0) {
          $this.addClass('arc-invisible');
        }
      });
  }

  function toggleExtFormTitleSubmitBtn(name) {
    $form.find('.modal-title').text(name);
    $submitExtract.html(`${name} <i class="arc-invisible create-confirmation-spinner fa fa-spinner fa-spin"></i>`);
  }

  function paintSchedules() {
    $createExtractSchedules.empty();
    $createExtractSchedules.append('<option value="0">Manual</option>');
    schedules.schedule_groups.forEach((g) => {
      const option = `<option value="${g.id}">${arc.misc.escapeHTML(g.schedule_name)}</option>`;
      $createExtractSchedules.append(option);
    });
  }

  function checkPartitionSupportAndLoadDatabases(selectedDb) {
    disableDbnameSelect();

    const dcId = $dataConnectionSelect.val();
    if (!dcId) {
      return;
    }

    // Hide or show support for partitioning data depending on the target
    // data connection
    let tgtSupportsPartitions;
    const dc = dataconnections[dcId];
    if (dc) {
      tgtSupportsPartitions = arc.dataconnection.supportsPartitions(dc.type);
    }

    if (tgtSupportsPartitions) {
      $targetPartitionsDiv.show();
    } else {
      $targetPartitionsDiv.hide();
    }

    $dbnameSelect.empty();
    arc.apps_data.load_databases_cb(Number(dcId), (_error, data) => {
      if (_error) {
        /* eslint-disable no-console */
        console.error('Failed to fetch databases');
        /* eslint-enable no-console */
        return;
      }

      $dbnameSelect.empty();
      $dbnameSelect.append('<option selected disabled hidden></option>'); // No option is selected initially
      data.rows.forEach((row) => {
        const option = arc.misc.safe_substitute(
            '<option value="%VALUE%">%NAME%</option>',
            {
              NAME: row[0] === '' ? '<<DEFAULT>>' : row[0],
              VALUE: row[0],
            }
        );
        $dbnameSelect.append(option);
      });

      $dbnameSelect.chosen({
        disable_search_threshold: 5,
        allow_single_deselect: true,
        search_contains: true,
        placeholder_text_single: `Default language: ${arc.misc.getLanguage()}`
      });

      // If there is exactly one db, select it by default. Otherwise, do not set
      // a default (since we do not know where the user would like to create a
      // data extract)
      if (data.rows.length === 1) {
        $dbnameSelect.val(data.rows[0]);
      }

      // In Edit modal, select existing extract's target db
      if (selectedDb) {
        $dbnameSelect.val(selectedDb);
      }

      $dbnameSelect.attr('disabled', false);
      $dbnameSelect.trigger('chosen:updated');
      $dbnameSelect.trigger('change.bs.validator');
    });
  }

  function disableDbnameSelect() {
    $dbnameSelect.chosen('destroy');
    $dbnameSelect.val('');
    $dbnameSelect.attr('disabled', true);
  }

  function getAggColAliasExprPerConn(aggAction, colName) {
    // use correct names according to conn here => conn.count...
    const dctype = builder.dataset().dataconnection_type();
    const conn = arc.dataconnection.types[dctype]();

    // build a shelf element so we can get the alias and expr
    const shelfElem = arc.report.shelf.element()
                                .dataset_colname(colName);
    shelfElem.setShelfFnNameFromShelfId(aggAction, conn);
    const aliasExpr = shelfElem.getExpressionComponents();

    // TO-DO: improve aggregated column type guessing (also in visual shelves )
    // Now, on UI we update column type to string if aggregate is string concat,
    // otherwise to integer
    aliasExpr.iconType = 'type-integer';
    if (aggAction === 'shelf-agg-groupconcat') {
      aliasExpr.iconType = 'type-string';
    }

    // strip "'"
    const { alias } = aliasExpr;
    if ((alias[0] === '\'') && (alias[alias.length - 1] === '\'')) {
      aliasExpr.alias = alias.substr(1, alias.length - 2);
    }

    return aliasExpr;
  }

  function getColNameAndAggAction(alias, expr) {
    const dctype = builder.dataset().dataconnection_type();
    const conn = arc.dataconnection.types[dctype]();

    let colname;
    let aggFn = expr.substring(0, expr.indexOf('('));

    if (conn.expr_appxmedian && conn.expr_appxmedian('').includes(aggFn)) {
      // e.g. approx_percentile([colname],0.5)
      colname = expr.substring(expr.indexOf('(') + 1, expr.indexOf(','));

      // cli extract's aggregated columns expr doesn't have square braces
      // around them, but UI extracts do
      colname = arc.dataconnection.types.stripBrackets(colname);

      return [colname, 'shelf-agg-appx-median'];
    }

    if (alias.includes('countdistinct') || alias.includes('count(distinct')) {
      aggFn = 'countdistinct';
      colname = expr.substring(expr.lastIndexOf('(') + 1, expr.indexOf(')'));
    } else {
      colname = expr.substring(expr.indexOf('(') + 1, expr.indexOf(')'));
    }

    // cli extract's aggregated columns expr doesn't have square braces
    // around them, but UI extracts do
    colname = arc.dataconnection.types.stripBrackets(colname);

    // eslint-disable-next-line default-case
    switch (aggFn) {
      case conn.sum:
        return [colname, 'shelf-agg-sum'];

      case conn.count:
        return [colname, 'shelf-agg-count'];

      case conn.ndv:
        return [colname, 'shelf-agg-ndv'];

      case 'countdistinct':
        return [colname, 'shelf-agg-countdistinct'];

      case conn.min:
        return [colname, 'shelf-agg-min'];

      case conn.max:
        return [colname, 'shelf-agg-max'];

      case conn.avg:
        return [colname, 'shelf-agg-avg'];

      case conn.appx_median:
        return [colname, 'shelf-agg-appx-median'];

      case conn.stddev:
        return [colname, 'shelf-agg-stddev'];

      case conn.stddev_pop:
        return [colname, 'shelf-agg-stddev-pop'];

      case conn.variance:
        return [colname, 'shelf-agg-variance'];

      case conn.variance_pop:
        return [colname, 'shelf-agg-variance-pop'];

      case conn.group_concat:
        return [colname, 'shelf-agg-groupconcat'];
    }
  }

  return c;
}
}());
