window.arc = window.arc || {};

(function() {
"use strict";

/*
 * View Management.
 *
 * Fetch and display views on the data connections page and the dataset page.
 *
 * If there are views currently being refreshed, poll status every 10 seconds
 * and update the page to display whether refresh is still ongoing.
 *
 * In the datasets page, associated with the create button, and paint and
 * handle the create aview modal.
 */

arc.admin = arc.admin || {};

arc.admin.aviewStates = {
    REFRESHING: '<span class="badge badge-warning badge-pill">Refreshing</span>',
    FRESH: '<span class="badge badge-success badge-pill">Up to date</span>',
    STALE: '<span class="badge badge-warning badge-pill">Needs refresh</span>',
    UNUSABLE: '<span class="badge badge-warning badge-pill">Needs initial refresh</span>',
    RECOMMENDED: '<span class="badge badge-info badge-pill">Recommended</span>',
    INVALID: '<span class="badge badge-danger badge-pill">INVALID</span>'
};

arc.admin.translateAviewState = function(state, hasRefreshError, detail) {
    let errorMessage = hasRefreshError ? '</br><span class="text-danger">Refresh failed</span>' :
      '';
    let states = arc.admin.aviewStates;
    let html = '';

    if (states[state]) {
      html = states[state] + errorMessage;
    } else {
      html = `<span class="badge badge-danger badge-pill">${state}</span>`;
    }

    if (detail) {
      html = html.replace('span', `span data-toggle="tooltip" title="${detail}"`);
    }

    return html;
};

arc.admin.aviewManager = function() {
  var aviewManager = function() { return "aviewManager"; };

  const defaultSelectedInputs = {
    all: false,
    tables: {},
    views: {},
  };

  var dataconnection_id,
      dataconnection_type,
      dataconnection_info,
      arcengineVersion,
      tasks = [],
      builder,
      allDatasets,
      tables,
      lvtable,
      fetchThreadActive = false,
      fetchSeconds = 2,
      disabled,
      clickedDuringFetch,
      cancelled = {},
      aviewDisplayConfig,
      parsedAviews = [],
      unfilteredBaseDbs = {},
      unfilteredBasetables = [],
      avhash = {},
      tablenames = [],
      totalStatus = {},
      tableInfo = {},
      selectedInputs = defaultSelectedInputs,
      collapsedTables = {};

  const defaultDisplayConfig = {
    style: 'grouped-table',
    dbFilter: null,
    tableFilter: null,
    statusFilter: null,
    order: [
      { field: 'baseName', direction: 'asc' },
      { field: 'name', direction: 'asc' },
    ],
  };

  aviewManager.tables = function(_) {
    if (!arguments.length) {
      // if an associated LV exists prepend to list of tables
      if (tables.length && lvtable) {
        var newArray = tables.slice();
        newArray.unshift(lvtable);
        return newArray;
      } else {
        return tables;
      }
    }
    tables = _;
    return aviewManager;
  };

  aviewManager.allDatasets = function(_) {
    if (!arguments.length) return allDatasets;
    allDatasets = _;
    return aviewManager;
  };

  aviewManager.lvtable = function(_) {
    if (!arguments.length) return lvtable;
    lvtable = _;
    return aviewManager;
  };

  aviewManager.disable = function() {
    disabled = true;
    return aviewManager;
  };

  aviewManager.disabled = function(_) {
    if (!arguments.length) return disabled;
    disabled = _;
    return aviewManager;
  };

  aviewManager.dataconnection_info = function() {
    return dataconnection_info;
  };

  aviewManager.dataconnection_id = function(_) {
    if (!arguments.length) return dataconnection_id;
    dataconnection_id = _;
    return aviewManager;
  };

  aviewManager.dataconnection_type = function(_) {
    if (!arguments.length) return dataconnection_type;
    dataconnection_type = _;
    return aviewManager;
  };

  function get_dataconnection_id() {
    if (dataconnection_id) {
      return dataconnection_id;
    } else if (builder) {
      return builder.dataset().dataconnection_id();
    }

    return undefined;
  }

  function get_dataconnection_type() {
    if (dataconnection_type) {
      return dataconnection_type;
    } else if (builder) {
      return builder.dataset().dataconnection_type();
    }

    return undefined;
  }

  aviewManager.aviewDisplayConfig = function(_) {
    if (!arguments.length) return aviewDisplayConfig;
    aviewDisplayConfig = _;
    return aviewManager;
  };

  aviewManager.builder = function(_) {
    if (!arguments.length) return builder;
    builder = _;
    return aviewManager;
  };

  aviewManager.selectedInputs = function(_) {
    if (!arguments.length) return selectedInputs;
    selectedInputs = _;
    return aviewManager;
  };

  aviewManager.loadViews = function() {
    setAViewButton();
    arc.admin.setupAVDetailListeners({});
    aviewDisplayConfig = aviewDisplayConfig || defaultDisplayConfig;

    if (parsedAviews.length) {
      buildAVHash();
      paintViews();
      return aviewManager;
    }

    fetchAndPaintViews();
    return aviewManager;
  };

  function aviewRowsToObj(aview, colNames) {
    var aviewObj = {};
    colNames.forEach(function(d, i) {
      aviewObj[d] = aview[i];
    });
    return aviewObj;
  }

  function stripBackticks(s) {
    // ArcEngine returns the table name wrapped inside backticks if the table name
    // has a trailing or leading underscore. This can potentially happen to db names
    // too. We strip those out here.
    var removeBacktick = /^`.+`$/;
    return (removeBacktick.test(s) ? s.substring(1, s.length-1) : s);
  }

  function clearSelectedInputs() {
    selectedInputs.all = false;
    selectedInputs.tables = {};
    selectedInputs.views = {};
  }

  function paintViews() {
    if (disabled) {
      // This aview manager has been disabled. Don't do anything
      return;
    }

    var refreshViews = {};
    var dcid = get_dataconnection_id();
    tasks.forEach(function(task) {
      if (dcid === task.dc_id) {
        refreshViews[task.view] = task;
      }
    });

    // Display the view count in the badge next to the 'Analytical Views' tab
    const totalCount = parsedAviews.length;
    $("#aviews-count").text(totalCount);
    const $treeTable = $('#aview-tree-table');
    const $statusBoxes = $('.av-status-box');

    const sortFieldToColumnClass = {
      baseName: 'base-table-name',
      name: 'av-name',
      overallScore: 'star-rating',
      nextScheduledRefresh: 'next-run',
      lastSuccessfulRefresh: 'last-run',
    };

    function updateStatusFilter(evt) {
      const $thisBox = $(evt.currentTarget);
      $statusBoxes.removeClass('active');
      $thisBox.addClass('active');
      aviewDisplayConfig.statusFilter = $thisBox.data('statuses');
      paintViews();
    }

    // Populate table select dropdown and status totals
    if (arcengineVersion === '2.12.1' && totalCount > 0) {
      // Empty base db and table select fields
      const $dbSelect = $('#av-database-select');
      const $tableSelect = $('#av-table-select');
      const $selects = $('.base-filter');
      $dbSelect.empty();
      $tableSelect.empty();
      let $tableOptions = [$('<option selected>All</option>')];
      let $dbOptions = [$('<option selected>All</option>')];
      let addTableOption = function(tableName) {
        $tableOptions.push($(`<option>${tableName}</option>`));
      };

      Object.keys(unfilteredBaseDbs).forEach(
        (dbName) => $dbOptions.push($(`<option>${dbName}</option>`))
      );

      $dbSelect.append($dbOptions);
      let currentDbFilter = aviewDisplayConfig.dbFilter;
      if (currentDbFilter) {
        $dbSelect.val(currentDbFilter);
        unfilteredBaseDbs[currentDbFilter].forEach(addTableOption);
      } else {
        unfilteredBasetables.forEach(addTableOption);
      }

      $tableSelect.append($tableOptions);

      if (aviewDisplayConfig.tableFilter) {
        $tableSelect.val(aviewDisplayConfig.tableFilter);
      }

      $dbSelect.trigger('chosen:updated');
      $tableSelect.trigger('chosen:updated');
      $selects.offOn('change.filter', function updateBaseFilter(evt) {
        const $filter = $(evt.currentTarget);
        const val = $filter.val();
        const filterVal = val === 'All' ? null : val;
        if ($filter.attr('id') === 'av-database-select') {
          aviewDisplayConfig.dbFilter = filterVal;
        } else {
          aviewDisplayConfig.tableFilter = filterVal;
        }

        buildAVHash();
        paintViews();
      });

      const upToDateCount = totalStatus.FRESH || 0;
      const refreshCount = (totalStatus.STALE || 0) + (totalStatus.UNUSABLE || 0) +
        (totalStatus.REFRESHING || 0);
      const invalidCount = totalStatus.INVALID || 0;
      $('#up-to-date').toggleClass('arc-invisible', upToDateCount === 0)
                      .data('statuses', ['FRESH'])
                      .find('.av-status-number')
                      .text(upToDateCount);
      $('#need-refresh').toggleClass('arc-invisible', refreshCount === 0)
                      .data('statuses', ['STALE', 'UNUSABLE', 'REFRESHING'])
                      .find('.av-status-number')
                      .text(refreshCount);
      $('#have-error').toggleClass('arc-invisible', invalidCount === 0)
                      .data('statuses',['INVALID'])
                      .find('.av-status-number')
                      .text(invalidCount);
      $('#total-count').find('.av-status-number')
                       .text(upToDateCount + refreshCount + invalidCount);
      $('.av-status-container').removeClass('arc-invisible');
      $('#aview-summary').removeClass('arc-invisible').data('statuses', null);
      $statusBoxes.offOn('click.filter', updateStatusFilter);
      const $sortableHeaders = $treeTable.find('th.sortable');
      $sortableHeaders.removeClass('sort-asc sort-desc');
      $sortableHeaders.offOn('click.sort', function updateTableSort(event) {
        const sortField = $(event.currentTarget).data('avfield');
        const sortCount = aviewDisplayConfig.order.length;
        const currentSort = aviewDisplayConfig.order[sortCount - 1];
        const isGrouped = aviewDisplayConfig.style === 'grouped-table';
        const newSort = { field: sortField, direction: 'asc' };

        function toggleSort(dir) {
          if (dir === 'asc') {
            return 'desc';
          } else {
            return 'asc';
          }
        }

        // If already sorting on that field, we can just toggle sort direction and can get away
        // with just repainting -- no need to regenerate the avhash
        if (sortField === currentSort.field) {
          currentSort.direction = toggleSort(currentSort.direction);
        } else if (sortField === 'baseName' && isGrouped) { // reducing down to a single sort
          aviewDisplayConfig.order.pop();
          const firstSort = aviewDisplayConfig.order[0];
          firstSort.direction = toggleSort(firstSort.direction);
          buildAVHash();
        } else {
          if (isGrouped && sortCount === 1) {
            aviewDisplayConfig.order.push(newSort);
          } else {
            aviewDisplayConfig.order[sortCount - 1] = newSort;
          }

          buildAVHash();
        }

        paintViews();
      });


      $('#bulk-av-refresh, #bulk-av-delete').addClass('disabled');

      $('#bulk-av-refresh, #bulk-av-delete').offOn('click.bulkAction', function handleBulk(evt) {
        if ($(this).hasClass('disabled')) {
          evt.stopPropagation();
          return;
        }

        if ($(this).attr('id') === 'bulk-av-delete') {
          dropViews(getSelectedAVs(), clearSelectedInputs);
        } else {
          refreshAVs(getSelectedAVs(), clearSelectedInputs);
        }
      });
    }

    function getSelectedAVs() {
      let avList = [];
      $('.av-child-row').find(':checked').each(function addToList() {
        avList.push($(this).closest('tr').data('arc-aview'));
      });

      return avList;
    }

    // Remove any previously painted tables
    const $unsortableTablesContainer = $('.aviews-tables-container');
    $unsortableTablesContainer.empty();
    $treeTable.children('tbody').remove();
    $treeTable.toggleClass('arc-invisible', arcengineVersion !== '2.12.1' || totalCount === 0);
    if (totalCount === 0) {
      $('#aview-summary').addClass('arc-invisible');
      const $datasetLevelMsg = $('.views-none.dataset-level');
      if ($datasetLevelMsg.length) {
        $datasetLevelMsg.removeClass('arc-invisible');
      } else {
        $('.views-none').removeClass('arc-invisible');
      }

      return;
    }

    let templateTableHtml, templateRowHtml, getTableRowSubs, getViewRowSubs,
      $tablesContainer, $tbody, highlightedClass;
    // Paint the actual tables displaying each view

    const isFlat = aviewDisplayConfig.style === 'flat-table';
    if (arcengineVersion === '2.12.1') {
      // populate header stuff
      templateTableHtml = $('#aviews-table-v2-12-1-table-row-template table').html();
      const $templateRow = $('#aviews-table-v2-12-1-av-row-template tbody');

      if (aviewDisplayConfig.order[1]) {
        highlightedClass = sortFieldToColumnClass[aviewDisplayConfig.order[1].field];
      } else if (aviewDisplayConfig.style === 'flat-table') {
        highlightedClass = sortFieldToColumnClass[aviewDisplayConfig.order[0].field];
      }

      // handle any hiding associated with collapse
      if (isFlat) {
        $templateRow.find('.av-child-row').removeClass('.arc-invisible');
      }

      templateRowHtml = $templateRow.html();
      $tablesContainer = $treeTable;
      const $toggleAll = $tablesContainer.find('.table-collapse-all, .table-expand-all');

      $toggleAll.toggleClass('arc-invisible', isFlat);
      if (isFlat) {
        $tbody = $('<tbody></tbody>');
        $tablesContainer.append($tbody);
      }

      $('#select-app-all').prop('checked', !!selectedInputs.all);

      getTableRowSubs = function(tablename) {
        let statusHashHtml = '';
        const classMappings = {
          'badge-success': {
            statuses: ['FRESH'],
            display: 'up to date',
          },
          'badge-warning': {
            statuses: ['STALE', 'UNUSABLE', 'REFRESHING'],
            display: 'need refresh',
          },
          'badge-danger': {
            statuses: ['INVALID'],
            display: 'have errors',
          },
        };
        const tableData = tableInfo[tablename];
        Object.keys(classMappings).forEach(function addStatusCountBadge(badgeClass) {
          const statuses = classMappings[badgeClass].statuses;
          const display = classMappings[badgeClass].display;
          const count = statuses.reduce(
            (acc, status) => { return (tableData.statusHash[status] || 0) + acc; }, 0);
          if (count > 0) {
            const title = `${count} AV(s) ${display}`;
            statusHashHtml +=
              ` <span class="badge ${badgeClass}" data-toggle="tooltip" title="${title}">`;
            statusHashHtml += `${count}</span>`;
          }
        });
        const checkedAttr =
          selectedInputs.all || (selectedInputs.tables && selectedInputs.tables[tablename]) ?
          'checked' : '';
        const substitutions = {
          CHECKED_ATTR: checkedAttr,
          TABLENAME: tablename,
          LV_CLASS: tableData.isLview ? '' : 'arc-invisible',
          AV_COUNT: tableData.totalCount,
          HTML_STATE_TOTAL: statusHashHtml,
        };
        return substitutions;
      };

      getViewRowSubs = function(aview) {
        const checkedAttr = selectedInputs.views && selectedInputs.views[aview.name] ?
          'checked' : '';
        return {
          CHECKED_ATTR: checkedAttr,
          HTML_RATING: arc.admin.generateStars(aview.overallScore),
          TABLENAME: aview.base,
          TABLE_CLASS: aviewDisplayConfig.style === 'flat-table' ? '' : 'arc-invisible',
          PRIMARY_SORT_CLASS: aviewDisplayConfig.style === 'flat-table' ? '' : 'sorting_p',
          LV_CLASS: !!aview.base_lview_name ? '' : 'arc-invisible',
          HTML_NEXT_SCHEDULED_REFRESH: arc.admin.generateRefreshString(aview),
          DETAIL_BTN_CLASS: aview.error_message ? 'disabled' : '',
        };
      };
    } else {
      $tablesContainer = $unsortableTablesContainer;
      const $templateTable = $('#aviews-table-template');
      const $templateRow = $('#aviews-table-row-template tbody');
      const templatePartition = $('#aview-dim-partition-template').html();
      tablenames = Object.keys(avhash);

      //Currently, no way to tell what version of arcengine is being run. Thus,
      //we hardcode the aview formatting based upon the number of columns returned.
      //In arcengine v2_5, there are 15 columns. In arcengine v2_2_4, there are 7 columns.
      //Older versions have fewer columns.
      const aviewConfigs = {
        2.5: {
          columnWidths: [15, 15, 5, 5, 5, 5, 5, 10, 10, 10, 15],
          remove: '.arcengine_v2_2_4',
        },
        '2.2.4': {
          columnWidths: [25, 15, 10, 15, 15, 20],
          remove: '.arcengine_v2_5'
        },
        default: {
          columnWidths: [40, 15, 10, 35],
          remove: '.arcengine_v2_2_4, .arcengine_v2_5'
        }
      };

      const currAviewConfig = aviewConfigs[arcengineVersion];
      const removalSelector = currAviewConfig.remove;
      $templateTable.find(removalSelector).remove();
      $templateRow.find(removalSelector).remove();
      const columnWidths = currAviewConfig.columnWidths;

      $templateTable.find('th').each((i) => $(this).css('width', `${columnWidths[i]}%`));

      templateTableHtml = $templateTable.html();
      templateRowHtml = $templateRow.html();
      getTableRowSubs = function(tablename) {
        return { TABLENAME: tablename };
      };
      getViewRowSubs = function(aview) {
        let measures, dimensions, partitions, last_success_refresh,
            new_partitions, deleted_partitions, updated_partitions, total_partitions;
        //Arcengine 2.5: structure is now JSON, partitions has changed
        if (arcengineVersion === '2.5') {
          var processArrayOrEmptyString = function(s) {
            if (s === 'N/A') {
              return '';
            }
            s = JSON.parse(s);
            if (s) {
              s = s.join(', ');
            }
            return s;
          };

          measures = processArrayOrEmptyString(aview.aggregates);
          dimensions = processArrayOrEmptyString(aview.dimensions);
          partitions = processArrayOrEmptyString(aview['aview partition names']);

          deleted_partitions = aview['number of deleted base table partitions'];
          new_partitions = aview['number of newly added base table partitions'];
          updated_partitions = aview['number of updated base table partitions'];
          total_partitions = aview['total number of aview partitions'];
        } else if (arcengineVersion === '2.2.4') {
          measures = aview.aggregates.replace(/,/g, ', ');
          dimensions = aview.dimensions.replace(/,/g, ', ');
          partitions = aview.partitions;
        }

        if (partitions) {
          if (dimensions) dimensions+=', ';

          partitions.split(',').forEach(function(p, i, a) {
            dimensions+=arc.misc.safe_substitute(
              templatePartition,
              {
                PARTITION: p,
              });

            if (i !== a.length - 1)
              dimensions += ', ';
          });
        }

        return {
          NEW_PARTITIONS: new_partitions,
          DELETED_PARTITIONS: deleted_partitions,
          UPDATED_PARTITIONS: updated_partitions,
          TOTAL_PARTITIONS: total_partitions,
          SQL: aview.select,
          HTML_DIMENSIONS: dimensions || 'N/A',
          HTML_MEASURES: measures || 'N/A',
        };
      };
    }

    var refreshingCount = 0;
    let stateRefreshing = arc.admin.aviewStates.REFRESHING;
    if (arcengineVersion === '2.12.1') {
      const firstSortDirection = aviewDisplayConfig.order[0].direction;
      const firstSortField = aviewDisplayConfig.order[0].field;
      const secondSortDirection = aviewDisplayConfig.order[1] ?
        aviewDisplayConfig.order[1].direction : undefined;
      const secondSortField = aviewDisplayConfig.order[1] ?
        aviewDisplayConfig.order[1].field : undefined;
      let firstKeyList = Object.keys(avhash).sort();
      if (firstSortDirection === 'desc') {
        firstKeyList.reverse();
      }

      // update header sort arrows
      $treeTable.find(`[data-avfield=${firstSortField}]`).addClass(`sort-${firstSortDirection}`);
      if (secondSortField) {
        $treeTable.find(`[data-avfield=${secondSortField}]`)
                  .addClass(`sort-${secondSortDirection}`);
      }

      firstKeyList.forEach(function printTableSection(firstKeyItem) {
        let $container;

        // print table row if necessary
        if (aviewDisplayConfig.style === 'grouped-table') {
          const displayedStatuses = aviewDisplayConfig.statusFilter;
          if (displayedStatuses) {
            if (!displayedStatuses.some(status => tableInfo[firstKeyItem].statusHash[status])) {
              return;
            }
          }
          const $tableRow = $(arc.misc.safe_substitute(templateTableHtml,
            getTableRowSubs(firstKeyItem))); // when grouped, firstKeyItem is name of base table
          $tableRow.find(`.${highlightedClass}`).addClass('sorting_1');
          $tablesContainer.append($tableRow);
          $container = $tablesContainer.find('tbody').last();
        } else {
          $container = $tbody;
        }

        let firstKeyVal = avhash[firstKeyItem];
        if (Array.isArray(firstKeyVal)) {
          firstKeyVal.forEach((aview) => printViewRow(aview, $container));
        } else {
          let secondKeyList = Object.keys(firstKeyVal).sort();
          if (secondSortDirection === 'desc') {
            secondKeyList.reverse();
          }

          secondKeyList.forEach(function printAviewRows(key) {
            firstKeyVal[key].forEach((aview) => printViewRow(aview, $container));
          });
        }
      });

      setupSelectionButtons();
      setupTableStyleButtons();
      setupCollapsibility();

      // Make fixes as needed based on change in list
      if ($('.av-sel:checked').length > 0) {
        $('#bulk-av-refresh, #bulk-av-delete').removeClass('disabled');
      }

      if ($treeTable.find('.av-sel:not(:checked)').length > 0) {
        $('#select-app-all').prop('checked', false);
        selectedInputs.all = false;
      }

      $treeTable.find('.av-parent-row').each(function updateSelectAll() {
        const $this = $(this);
        if ($this.siblings().find('.av-sel:not(:checked)').length > 0) {
          $this.find('.av-sel').prop('checked', false);
          selectedInputs.tables[$this.find('.truncated-needs-title').attr('title')] = false;
        }
      });
    } else {
      tablenames.sort().forEach(function printTable(tablename) {
        const tableData = avhash[tablename];
        $tablesContainer.append(arc.misc.safe_substitute(
            templateTableHtml,
            getTableRowSubs(tablename)
          ));
        const $tbody = $tablesContainer.find('tbody').last();
        tableData.views.forEach((aview) => printViewRow(aview, $tbody));
      });
    }

    if (refreshingCount > 0) {
      fetchAndPaintLater();
    }

    // Enable tooltips for the refresh/delete buttons
    $tablesContainer.find('[data-toggle="tooltip"]').tooltip({
      delay: {
       show: 500,
       hide: 100
      }
    });

    function printViewRow(aview, $tbody) {
      let hasRefreshError = false;
      let name = aview.name;
      let reftask = refreshViews[name];
      if (reftask && reftask.error) {
        hasRefreshError = true;
        arc.report.alertMessageSmall(`Analytical view refresh failed: ${reftask.error}`,
          'alert-danger');
      }

      let state = reftask && !reftask.error ? stateRefreshing :
        arc.admin.translateAviewState(aview.state, hasRefreshError, aview.error_message);
      const shownStatuses = aviewDisplayConfig.statusFilter;
      if (shownStatuses && !shownStatuses.includes(aview.state)) {
        return;
      }

      let refresh_string;
      if (arcengineVersion === '2.5') {
        refresh_string = aview['last successful refresh'];
      } else if (arcengineVersion === '2.12.1') {
        refresh_string = aview.lastSuccessfulRefresh;
      }

      let last_success_refresh = refresh_string ? arc.misc.friendlyDateSince(refresh_string)
        : refresh_string;
      let viewRowSubs = {
        VIEW: name,
        HTML_STATE: state,
        HTML_LAST_SUCCESS_REFRESH: last_success_refresh || 'N/A',
      };
      viewRowSubs = $.extend(viewRowSubs, getViewRowSubs(aview));
      const $row = $(arc.misc.safe_substitute(templateRowHtml, viewRowSubs));
      if (arcengineVersion === '2.12.1') {
        let substitutions = {
          VIEW: name,
          HTML_STATE: state,
          HTML_DIMENSIONS: arc.admin.generateAvDimensionsHtml(aview),
          MEASURES: aview.aggregates.join(', ') || 'None',
          SQL: aview.select,
        };
        $row.data('viewdef', aview)
            .data('substitutions', substitutions)
            .data('score', aview.scoreData);

        $row.find(`.${highlightedClass}`).addClass('sorting_1');

        $row.toggleClass('arc-invisible', !!collapsedTables[aview.base_table_name]);
      }

      $tbody.append($row);

      var refreshLink = $tbody.find('.aview-refresh').last();
      var row = $tbody.find('tr').last();
      var refactive = row.find('.aview-progress-active');
      refactive.addClass('arc-invisible');
      if (state === stateRefreshing) {
        ++refreshingCount;
        refreshLink.addClass('disabled');
        refreshLink.find('.fa-refresh').addClass('fa-spin');
        refactive.removeClass('arc-invisible');
        var timeleft = row.find('.aview-time-left');
        var progress = row.find('.progress-bar');

        // If task is cancelled, used the status frozen in the front-end
        let cancelStatus = cancelled[reftask.view];
        if (cancelStatus) {
          reftask.cancelled = true;
          reftask.progress = cancelStatus.progress;
        }

        progress.attr('style', `width: ${reftask.progress}%`);
        timeleft.text(arc.admin.refreshTimeLeftText(reftask, fetchSeconds));
      } else {
        cancelled[name] = undefined;
      }

      if (['FRESH', 'INVALID'].includes(aview.state)) {
        refreshLink.addClass('disabled');
      }

      if (aview.state === 'INVALID') {
        $tbody.find('.aview-schedule').last().addClass('disabled');
      }

      $tbody.find('.av-sql-link').last().offOn('click.avSql', function() {
        $(this).next('.av-sql-detail').toggleClass('show');

        $(this).text(function(i, text) {
          return text === 'Show SQL' ? 'Hide SQL' : 'Show SQL';
        });
      });
    }

    setupViewButtons();
  }

  // Collect up views by tables - this helps display all the views for a
  // single table next to each other
  function buildAVHash() {

    //start from scratch
    avhash = {};
    totalStatus = {};
    tableInfo = {};
    tablenames = [];

    if (arcengineVersion === '2.12.1') {
      const tableFilter = aviewDisplayConfig.tableFilter;
      const dbFilter = aviewDisplayConfig.dbFilter;
      const firstGroup = aviewDisplayConfig.order[0].field;
      const secondGroup =
        aviewDisplayConfig.order[1] ? aviewDisplayConfig.order[1].field : undefined;
      parsedAviews.forEach(function handleGrouping(aviewObj) {
        const status = aviewObj.state;
        let secondValue, baseDb, baseTable;
        const baseName = aviewObj.baseName;
        [baseDb, baseTable] = baseName.split('.');
        if (dbFilter && baseDb !== dbFilter) {
          return;
        }

        if (tableFilter && baseTable !== tableFilter) {
          return;
        }

        const firstValue = aviewObj[firstGroup];
        if (secondGroup) {
          secondValue = aviewObj[secondGroup];
        }

        // create hash based on values of the fields used to group and sort the AV table
        // In the case of a table with style flat-table, there is just one level of organization,
        // that of the column on which the table is sorted.  In the case of a grouped-table style,
        // there are two levels of organization: the base table and then a secondary sort on the
        // the sorted column
        if (!avhash[firstValue]) {
          if (secondGroup) {
            avhash[firstValue] = {};
            avhash[firstValue][secondValue] = [aviewObj];
          } else {
            avhash[firstValue] = [aviewObj];
          }
        } else {
          if (secondGroup) {
            if (avhash[firstValue][secondValue]) {
              avhash[firstValue][secondValue].push(aviewObj);
            } else {
              avhash[firstValue][secondValue] = [aviewObj];
            }
          } else {
            avhash[firstValue].push(aviewObj);
          }
        }

        totalStatus[status] = (totalStatus[status] || 0) + 1;

        tableInfo[baseName] = tableInfo[baseName] || {};
        tableInfo[baseName].statusHash = tableInfo[baseName].statusHash || {};
        tableInfo[baseName].statusHash[status] = (tableInfo[baseName].statusHash[status] || 0) + 1;
        tableInfo[baseName].totalCount = (tableInfo[baseName].totalCount || 0) + 1;
        if (aviewObj.base_lview_name) {
          tableInfo[baseName].isLview = true;
        }
      });
    } else {
      parsedAviews.forEach(function processAview(aviewObj) {
        const baseName = aviewObj.baseName;

        if (!(baseName in avhash)) {
          avhash[baseName] = {
            count: 0,
            views: [],
          };
        }

        avhash[baseName].views.push(aviewObj);
        ++avhash[baseName].count;
      });
    }

  }

  function refreshOneView(viewname, failureCb, successCb, alwaysCb) {
    var data = {
      opcode: 'enque',
      formValues: JSON.stringify({
        dcid: get_dataconnection_id(),
        view: viewname
      }),
    };

    $.ajax({
        url: arc.frontend.globals.urls.viewRefreshUrl,
        data: data,
        type: 'POST',
        beforeSend: function(xhr) {
          // Because of a jquery bug when handling PUT requests,
          // we have to explicitly set X-CSRFToken here
          xhr.setRequestHeader("X-CSRFToken", arc.globals.csrf_token);
        }
      })
      .fail(failureCb)
      .done(successCb)
      .always(alwaysCb);
  }

  function refreshAVs(views, refreshCb) {
    let refreshingViews = [];
    let viewIdx = 0;
    function refreshNextView() {
      const view = views[viewIdx];
      if (!view) {
        arc.report.alertMessageSmall(
          `View(s) ${arc.misc.makeEnglishList(refreshingViews)} are being refreshed`,
          'alert-info',
          undefined,
          true
        );
        return;
      }

      function failedRefreshCb(rsp, textstatus, errorthrown) {
        textstatus = rsp.responseText || textstatus;
        arc.report.alertMessage(`Failed to refresh aview ${viewname}`,
            textstatus ? textstatus : errorthrown);
      }

      function refreshQueuedCb(rsp) {
        refreshingViews.push(view);
        tasks.push({
          dc_id: rsp.dcid,
          view: rsp.view,
          schedule_id: rsp.schedule_id,
          joblog_id: rsp.joblog_id,
          status: 'Status: Queued',
          progress: 0,
        });

        if (refreshCb) {
          refreshCb();
        }

        paintViews();
      }

      function responseCb() {
        ++viewIdx;
        refreshNextView();
      }

      refreshOneView(view, failedRefreshCb, refreshQueuedCb, responseCb);
    }

    refreshNextView();
  }

  function refreshCancel(viewname) {
    // Don't send multiple AJAX requests if already cancel procedure is already
    // in progress
    if (cancelled[viewname]) {
      return;
    }

    let schedule_id, joblog_id, progress;
    tasks.some(function(task) {
      if (task.view === viewname) {
        schedule_id = task.schedule_id;
        joblog_id = task.joblog_id;
        progress = task.progress;
        return true;
      }
      return false;
    });

    // Require that we need both the schedule_id and the joblog_id to cancel
    if ((!schedule_id && schedule_id !== 0) || (!joblog_id && joblog_id !== 0)) {
      return;
    }

    cancelled[viewname] = { progress, };
    const url = `${arc.frontend.globals.urls.cancelJobUrl}/${schedule_id}/${joblog_id}`;
    $.ajax({
      url: url,
      data: {},
      type: 'POST',
      beforeSend: function(xhr) {
        // Because of a jquery bug when handling PUT requests,
        // we have to explicitly set X-CSRFToken here
        xhr.setRequestHeader("X-CSRFToken", arc.globals.csrf_token);
      }
    })
    .fail((rsp, textstatus, errorthrown) => {
      textstatus = rsp.responseText || textstatus;
      arc.report.alertMessageAllHide();
      arc.report.alertMessage("Failed to cancel aview refresh " + viewname,
          textstatus ? textstatus : errorthrown);
      cancelled[viewname] = undefined;
    })
    .done(() => {
      paintViews();
    });
  }

  function dropView(viewname, cb) {
    let rb = arc.requestBuilder();
    rb.requestType('.drop_aviews');
    rb.tableName(arc.dataconnection._enquote_tablename(viewname, 'arcengine'));
    arc.report.sql_runner.run(get_dataconnection_id(), '', cb, false, rb.dsreq());
  }

  function dropViews(views, deleteCb) {
    var c = arc.misc.deleteConfirmation();
    let deletedViews = [];

    c.objectType('Analytical View(s)')
     .objectName(arc.misc.makeEnglishList(views))
     .askUser(function queueDeletions() {
       $('#delete-confirmation-yes').prop('disabled', true);
       let viewIdx = 0;
       function dropNextView() {
         const view = views[viewIdx];
         if (!view) {
           c.hideModal();
           fetchAndPaintViews(function deleteSuccessCb() {
             arc.report.alertMessageSmall(
               `View(s) ${arc.misc.makeEnglishList(deletedViews)} deleted`,
               'alert-success'
             );

             if (deleteCb) {
               deleteCb();
             }
           });

           return;
         }

         dropView(view, function dropFetchCb(errorMsg) {
           if (errorMsg) {
             arc.report.alertMessage(`Failed to drop aview ${view}`, errorMsg);
           } else {
             deletedViews.push(view);
           }

           ++viewIdx;
           dropNextView();
         });
       }

       dropNextView();
     });
  }

  function setupViewButtons() {
    $(".aview-refresh").off("click")
                       .on("click", function() {
                         const $this = $(this);
                         if ($this.hasClass('disabled')) {
                           return;
                         }

                         var viewname = $this.closest("td").attr("data-arc-aview");
                         refreshAVs([viewname]);
                       });

    $(".aview-delete").off("click")
                      .on("click", function() {
                         var viewname = $(this).closest("td").attr("data-arc-aview");
                         dropViews([viewname]);
                      });

    $('.aview-cancel-btn').offOn('click.cancelRefresh', function handleRefreshCancel(evt) {
                            const viewname = $(this).closest('tr').attr('data-arc-aview');
                            if (cancelled[viewname]) {
                              evt.stopPropagation();
                              return;
                            }

                            refreshCancel(viewname);
                          });

    $('.aview-schedule').off('click')
                        .on('click', function() {
                          const $this = $(this);
                          if ($this.hasClass('disabled')) {
                            return;
                          }

                          const viewName = $this.closest('tr').attr('data-arc-aview');
                          arc.frontend.legacyBridge.events.next({
                            type: 'open-analytical-job-modal',
                            payload: { dataConnectionId: get_dataconnection_id(), viewName }
                          });
                        });
  }

  function setAViewButton() {
    const assets = [
      {
        asset: `${arc.frontend.globals.urls.dataconnectionUrl}/${aviewManager.dataconnection_id()}`,
        postprocess: d3.json,
      },
    ];

    arc.staticfetcher.fetch(assets, function(error, data) {
      if (!error) {
        dataconnection_info = data[0];
      }
    });

    $("#create-aview-btn")
      .off("click")
      .on("click", function() {
        var aviewCreator = arc.admin.aviewCreator()
                                    .aviewManager(aviewManager);
        if (builder) {
          aviewCreator.builder(builder)
                      .tables(aviewManager.tables());
        } else {
          aviewCreator.dataconnection_id(aviewManager.dataconnection_id());
        }

        aviewCreator.create(/*createCb*/ fetchAndPaintViews);
      });
  }

  function setupSelectionButtons() {
    const $aviewTable = $('#aview-tree-table');
    const $allCheckBoxes = $aviewTable.find('.av-sel');
    const $checkAllBox = $('#select-app-all');
    $aviewTable.offOn('click.checkbox', function handleCheckClick(evt) {
      const $target = $(evt.target);
      if (!$target.hasClass('av-sel')) {
        return;
      }

      const isChecked = $target.prop('checked');
      const $closestRow = $target.closest('tr');

      if (isChecked) {
        $('#bulk-av-refresh, #bulk-av-delete').removeClass('disabled');
      }

      function adjustSelectedViews(_, row) {
        selectedInputs.views[$(row).data('arc-aview')] = isChecked;
      }

      if ($target.attr('id') === 'select-app-all') {
        $allCheckBoxes.prop('checked', isChecked);
        selectedInputs.all = isChecked;
        $aviewTable.find('.av-parent-row').each(function adjustSelectedTables() {
          selectedInputs.tables[$(this).find('.truncated-needs-title').attr('title')] = isChecked;
        });
        $aviewTable.find('.av-child-row').each(adjustSelectedViews);
      } else if ($closestRow.hasClass('av-parent-row')) {
        const tableName = $closestRow.find('.truncated-needs-title').attr('title');
        selectedInputs.tables[tableName] = isChecked;
        const $viewRows = $closestRow.siblings('.av-child-row');
        $viewRows.find('.av-sel').prop('checked', isChecked);
        $viewRows.each(adjustSelectedViews);
      } else if ($closestRow.hasClass('av-child-row')) {
        selectedInputs.views[$closestRow.data('arc-aview')] = isChecked;
        const $parentRow = $closestRow.siblings('.av-parent-row');
        const baseTableName = $parentRow.find('.truncated-needs-title').attr('title');
        const $parentRowCheckbox = $parentRow.find('.av-sel');
        if (!isChecked && $parentRowCheckbox.prop('checked')) {
          $parentRowCheckbox.prop('checked', false);
          selectedInputs.tables[baseTableName] = false;
        }
      }

      if (!isChecked && $checkAllBox.prop('checked')) {
        $checkAllBox.prop('checked', false);
        selectedInputs.all = false;
      }

      if (!isChecked && $('.av-sel:checked').length === 0) {
        $('#bulk-av-refresh, #bulk-av-delete').addClass('disabled');
      }
    });
  }

  function setupCollapsibility() {
    const $body = $('body');
    $body.on('click.collapseTable','.table-collapse', function tableCollapse() {
      const $this = $(this);
      const $parentBody = $this.closest('.av-parent-body');
      collapsedTables[$parentBody.find('.truncated-needs-title').attr('title')] = true;
      $this.removeClass('table-collapse')
           .addClass('table-expand');
      $parentBody.find('.av-child-row').addClass('arc-invisible');
    });

    $body.on('click.expandTable','.table-expand', function tableExpand() {
      const $this = $(this);
      const $parentBody = $this.closest('.av-parent-body');
      collapsedTables[$parentBody.find('.truncated-needs-title').attr('title')] = false;
      $this.addClass('table-collapse')
             .removeClass('table-expand');
      $parentBody.find('.av-child-row').removeClass('arc-invisible');
    });

    const $childRows = $('table').find('.av-child-row');

    // expand collapse all
    $body.on('click.collapseAll','.table-collapse-all', function collapseAll() {
      const $aviewTable = $('#aview-tree-table');
      $aviewTable.find('.av-parent-row').each(function adjustCollapsedTables() {
        collapsedTables[$(this).find('.truncated-needs-title').attr('title')] = true;
      });
      $(this).removeClass('table-collapse-all')
            .addClass('table-expand-all');
      $('table .table-collapse').addClass('table-expand')
            .removeClass('table-collapse');
      $childRows.addClass('arc-invisible');
    });

    // expand all
    $body.on('click.expandAll','.table-expand-all', function expandAll() {
      collapsedTables = {};
      $(this).addClass('table-collapse-all')
            .removeClass('table-expand-all');
      $('table .table-expand').addClass('table-collapse')
            .removeClass('table-expand');
      $childRows.removeClass('arc-invisible');
    });

  }

  function setupTableStyleButtons() {
    const $styleOptions = $('.av-table-type > a');
    $styleOptions.offOn('click.toggle', function handleToggle(evt) {
      const $target = $(evt.target);
      if ($target.hasClass('selected')) {
        return;
      }

      if ($target.attr('id') === 'flat-table' ) {
        selectedInputs.tables = {};
        collapsedTables = {};
      }

      $styleOptions.removeClass('selected');
      $target.addClass('selected');
      toggleTableStyle($target.attr('id'));
    });

    $styleOptions.removeClass('selected');
    $(`#${aviewDisplayConfig.style}`).addClass('selected');
  }

  function toggleTableStyle(style) {
    aviewDisplayConfig.style = style;
    if (style === 'grouped-table') {
      aviewDisplayConfig.order.unshift({
        field: 'baseName',
        direction: 'asc',
      });
    } else {
      aviewDisplayConfig.order = aviewDisplayConfig.order.slice(0, 1);
    }

    buildAVHash();
    paintViews();
  }

  function fetchAndPaintViews(alertCb) {
    fetchThreadActive = false;

    if (disabled) {
      // This aview manager has been disabled. Don't do anything
      return;
    }

    function fetchCb(_aviews) {
      if (disabled) {
        console.log('ignoring stale fetch');
        return;
      }

      function handleAlertCb() {
        if (alertCb) {
          alertCb();
        }
      }

      handleAlertCb();

      function parseViews(rawResponse) {
        parsedAviews = [];
        unfilteredBaseDbs = {};
        unfilteredBasetables = [];
        let hasExtraMetadata = false;
        const aviewCols = rawResponse.colnames;
        const metadataColIdx = rawResponse.colnames.indexOf('Metadata');
        // Distinguish between arcengine v2_12_1 content and prior
        if (metadataColIdx !== -1) {
          hasExtraMetadata = true;
          arcengineVersion = '2.12.1';
        } else {
          switch(rawResponse.colnames.length) {
            case 15:
              arcengineVersion = '2.5';
              break;
            case 7:
              arcengineVersion = '2.2.4';
              break;
            default:
              arcengineVersion = 'default';
          }
        }

        let thash = {};
        if (tables) {
          tables.forEach(t => thash[t.tablename().toLowerCase()] = t);

          if (lvtable) {
            thash[lvtable.tablename().toLowerCase()] = lvtable;
          }
        }

        const dc_id = get_dataconnection_id();
        const dc_type = get_dataconnection_type();

        rawResponse.rows.forEach(function parseViewRow(row) {
          let aviewObj;
          if (hasExtraMetadata) {
            aviewObj = row[metadataColIdx];
            aviewObj.base = aviewObj.base_table_name || aviewObj.base_lview_name;
          } else {
            aviewObj = aviewRowsToObj(row, aviewCols);
          }

          aviewObj.baseName = getFormattedBaseName(aviewObj);

          if (tables && !(aviewObj.baseName in thash)) {
            return;
          }

          if (hasExtraMetadata) {
            if (aviewObj.base_lview_name) {
              aviewObj.isLview = true;
            }

            if (aviewObj.nextScheduledRefresh) {
              aviewObj.nextScheduledRefresh = Date.parse(aviewObj.nextScheduledRefresh);
            }

            aviewObj.partitions = aviewObj.partitioning_columns;
            aviewObj.basetables.forEach(base => base.partitions = base.partitioning_columns);
            const scoreData = arc.admin.computeAviewScore(aviewObj);
            aviewObj.overallScore = scoreData.overallScore;
            aviewObj.scoreData = scoreData;
            if (aviewObj.last_successful_refresh) {
              aviewObj.lastSuccessfulRefresh = Date.parse(aviewObj.last_successful_refresh);
            }

            let baseDb, baseTable;
            [baseDb, baseTable] = aviewObj.base.split('.');
            if (!unfilteredBaseDbs[baseDb]) {
              unfilteredBaseDbs[baseDb] = [baseTable];
            } else if (!unfilteredBaseDbs[baseDb].includes(baseTable)) {
              unfilteredBaseDbs[baseDb].push(baseTable);
            }

            if (!unfilteredBasetables.includes(baseTable)) {
              unfilteredBasetables.push(baseTable);
            }
          }

          aviewObj.dc_id = dc_id;
          aviewObj.dc_type = dc_type;

          function getFormattedBaseName(aviewObj) {
            const parsedTableName = aviewObj.base.split('.');
            return `${stripBackticks(parsedTableName[0])}.${stripBackticks(parsedTableName[1])}`;
          }

          parsedAviews.push(aviewObj);
        });
      }

      parseViews(_aviews);
      buildAVHash();

      $.ajax({
        url: arc.frontend.globals.urls.viewRefreshUrl,
        data: { opcode: 'fetch' , formValues: JSON.stringify({}) },
        type: 'POST',
        beforeSend: function(xhr) {
          // Because of a jquery bug when handling PUT requests,
          // we have to explicitly set X-CSRFToken here
          xhr.setRequestHeader("X-CSRFToken", arc.globals.csrf_token);
        }
      })
      .fail(function(rsp, textstatus, errorthrown) {
        if (disabled) {
          console.log('ignoring stale fetch for refresh status');
          return;
        }

        textstatus = rsp.responseText || textstatus;
        arc.report.alertMessageAllHide();
        arc.report.alertMessage("Failed to fetch refresh status",
            textstatus + ". " + errorthrown);
      })
      .done(function(rsp) {
        tasks = rsp.views;
      })
      .always(function () {
        $(".views-loading").addClass("arc-invisible");
        $(".views-none").addClass("arc-invisible");
        paintViews();
      });
    }

    $.ajax({
      url: `${arc.frontend.globals.urls.aviewListUrl}/${dataconnection_id}`,
      type: 'GET',
      beforeSend: function(xhr) {
        // Because of a jquery bug when handling PUT requests,
        // we have to explicitly set X-CSRFToken here
        xhr.setRequestHeader('X-CSRFToken', arc.globals.csrf_token);
      }
    })
    .fail(function(rsp, textstatus, errorthrown) {
      if (disabled) {
        console.log('ignoring stale fetch for avs');
        return;
      }

      /*
       * This call to fetch all views tends to take a lot of time.
       * If users navigate away from the page, the call gets cancelled
       * and we see an error flash on screen before navigating to the
       * next page.
       *
       * In order to prevent this invalid error notification, we check
       * for rsp.status to be non-zero
       */
      if (rsp.status === 0) {
        return;
      }

      textstatus = rsp.responseText || textstatus;
      arc.report.alertMessageAllHide();

      arc.report.alertMessage("Failed to fetch analytical views",
          textstatus + ". " + errorthrown);
    })
    .done(fetchCb);
  }

  function fetchAndPaintLater() {
    // This aview manager has been disabled. Don't do anything
    if (disabled) {
      return;
    }

    if (!fetchThreadActive) {
      fetchThreadActive = true;
      setTimeout(fetchAndPaintViews, fetchSeconds * 1000);
    }
  }

  return aviewManager;
};

arc.admin.aviewCreator = function() {
  var creator = function() { return "arc.admin.aviewCreator"; };

  var dataconnection_id,
      dataconnection,
      selected_dataset,
      datasetsById = {},
      tables,
      lvtable,
      builder,
      aviewManager,
      selectedTableName,
      colsByName = {},
      createCb,
      selectedDbName = "",
      dbTables = {},
      dbTableDelimiter = ".",
      $templateOption = $("#aview-table-item select").html(),
      $modal = $("#create-aview-modal"),
      $dimDiv = $modal.find("#create-aview-dims"),
      $aggDiv = $modal.find("#create-aview-aggs"),
      $partDiv = $modal.find("#create-aview-parts"),
      $attrErrorBlock = $modal.find("#create-aview-attr-error");

  creator.dataconnection_id = function(_) {
    if (!arguments.length) return dataconnection_id;
    dataconnection_id = _;
    return creator;
  };

  creator.tables = function(_) {
    if (!arguments.length) return tables;
    tables = _;
    return creator;
  };

  creator.builder = function(_) {
    if (!arguments.length) return builder;
    builder = _;
    return creator;
  };

  creator.aviewManager = function(_) {
    if (!arguments.length) return aviewManager;
    aviewManager = _;
    return creator;
  };

  function displayOrHideDatasets() {
    var $dataset_dropdown = $modal.find("#create-aview-dataset");
    if (builder) {
      $dataset_dropdown.hide();
      return;
    }

    $dataset_dropdown.show();

    var $sel = $dataset_dropdown.find("select");
    $sel.empty();
    datasetsById = {};

    aviewManager.allDatasets().forEach(function(d) {
      if (d.dc_id !== dataconnection_id) {
        return;
      }

      datasetsById[d.id] = d;
      var html = arc.misc.safe_substitute(
          $templateOption,
          {
            VALUE: d.name,
          });
      $sel.append($(html).val(d.id)[0]);
    });

    $sel.off("change")
      .on("change", function() {
        selectedTableName = undefined;
        displayTables();
        buildNameInputAndAttrs();
      });
  }

  function displayTables() {
    var $sel = $modal.find("#create-aview-table select");
    $sel.empty();

    var response, selectedDatasetId, dataset_lvtable;
    if (builder) {
      lvtable = builder.dataset().lvname();
    } else {
      //If there is no dataset associated with this modal, create our dataset on
      //the fly
      selectedDatasetId = $modal.find("#create-aview-dataset select").val();
      response = datasetsById[selectedDatasetId];
      selected_dataset = arc.datasets.Dataset()
                    .dataset_id(selectedDatasetId)
                    .dataconnection_id(dataconnection_id)
                    .init_from_fetch(response);
      dataset_lvtable = selected_dataset.dataset_lvtable();
      tables = selected_dataset.tables();
      if (dataset_lvtable) {
        tables.unshift(selected_dataset.dataset_lvtable());
      }
      lvtable = selected_dataset.lvname();
    }

    tables.forEach(function(t) {
      if (!selectedTableName) {
        selectedTableName = t.tablename();
      }
      $sel.append(arc.misc.safe_substitute(
          $templateOption,
          {
            VALUE: t.tablename(),
          }));
    });

    $sel.prop('disabled', tables.length === 1);

    $sel.off("change")
        .on("change", function() {
          selectedTableName = $(this).val();
          buildNameInputAndAttrs();
        });

    const $dbSel = $modal.find("#create-aview-database select");
    $dbSel.off('change.updateDatasetName')
          .on('change.updateDatasetName', function() {
            selectedDbName = $(this).val();
            let viewName = $modal.find('#create-aview-table option:selected').text()
              .split(dbTableDelimiter)[1];
            const dc_id = getDcId();
            tableLoad(dc_id, selectedDbName).then(function updateModal(tableList) {
                updateViewName(viewName, tableList);
              }).catch(function showAlert(error) {
                arc.report.alertMessage('Could not return dataset tables', error);
                return;
              });
          });
  }

  function updateViewName(viewName, tableList) {
    dbTables = {};
    if (tableList && tableList.rows) {
      tableList.rows.forEach(function(t) {
        dbTables[t[0]] = 1;
      });
    }

    let viewIndex = 1;
    let complimentedViewName = `${viewName}_av_${viewIndex}`;
    while (complimentedViewName in dbTables) {
      ++viewIndex;
      complimentedViewName = `${viewName}_av_${viewIndex}`;
    }

    $modal.find('#create-aview-name input').val(complimentedViewName);
  }

  function tableLoad(dc_id, selectedDbName) {
    return new Promise(function(resolve, reject) {
      arc.apps_data.load_tables_cb(dc_id, selectedDbName, generateHandler(resolve, reject));
    });
  }

  function getDcId() {
    return builder ? builder.dataset().dataconnection_id() : dataconnection_id;
  }

  function generateHandler(resolve, reject) {
    function responseHandler(error, values) {
      if (error) {
        reject(error);
      }

      resolve(values);
    }

    return responseHandler;
  }

  function buildNameInputAndAttrs(cb) {
    $modal.find("#aview-db-name-spinner").removeClass("arc-invisible");

    var tableName = $modal.find('#create-aview-table option:selected').text(),
        splitfields = tableName.split(dbTableDelimiter),
        viewName = splitfields[1];
    selectedDbName = splitfields[0];

    const dc_id = getDcId();

    function dbLoad(dc_id, selectedDbName) {
      return new Promise(function(resolve, reject) {
        arc.apps_data.load_databases_cb(dc_id, generateHandler(resolve, reject));
      });
    }

    Promise.all([tableLoad(dc_id, selectedDbName), dbLoad(dc_id, selectedDbName)])
      .then(function populateModal(returnValues) {
        if (aviewManager.disabled()) {
          // This aview manager has been disabled. Don't do anything
          return;
        }

        let [tableList, dbList] = returnValues;
        updateViewName(viewName, tableList);

        const $dbSel = $modal.find('#aview-db-name').empty();
        let optionsList = [];
        if (dbList && dbList.rows) {
          dbList.rows.forEach(function addDbOption(db) {
            let name = db[0];
            let html = $templateOption.replace('%VALUE%', name);
            optionsList.push(html);
          });

          $dbSel.append($(optionsList.join('\n')));
          $dbSel.val(selectedDbName);
        }

        loadAttrAndValidator(cb);
      })
      .catch(function alert(error) {
        arc.report.alertMessage('Could not return data', error);
        return;
      });
  }

  function loadAttrAndValidator(cb) {
    paintDimAgg();

    var validationOptions = {
      custom: {
        norepeattables: function ($el) {
          return dbTables[$el.val().trim()] && 'Cannot enter duplicate table name.';
        },
        atleastonedimagg: function() {
          const $dimPicked = $dimDiv.find('input.aview-column-checkbox:checked');
          const $aggPicked = $aggDiv.find('input.aview-column-checkbox:checked');
          var dimlen = $dimPicked.length;
          var agglen = $aggPicked.length;

          // At least one non-partition column must be picked
          if ((dimlen > 0) || (agglen > 0)) {
            $dimDiv.find('input')
                   .off('.bs.validator')
                   .removeData(['bs.validator.errors', 'bs.validator.deferred',
                      'bs.validator.previous']);
            $aggDiv.find('input')
                   .off('.bs.validator')
                   .removeData(['bs.validator.errors', 'bs.validator.deferred',
                      'bs.validator.previous']);
            $attrErrorBlock.addClass("arc-invisible");
            return false;
          }

          $attrErrorBlock.addClass("has-error")
                         .removeClass("arc-invisible");
          return 'Please select at least one field.';
        },
      },
    };

    $modal.find("#create-aview-form")
          .validator("destroy")
          .off("submit")
          .validator(validationOptions)
          .on("submit", function(e) {
            if (!e.isDefaultPrevented()) {
              e.preventDefault();
              e.stopPropagation();
              $modal.find(".create-confirmation-spinner").removeClass("arc-invisible");
              doCreate();
            }
          });

    $attrErrorBlock.addClass("arc-invisible");

    $modal.find("#aview-db-name-spinner").addClass("arc-invisible");

    if (cb) {
      cb();
    }
  }

  function paintDimAgg() {
    colsByName = {};
    $dimDiv.empty();
    $aggDiv.empty();
    $partDiv.empty();

    //If there is no dataset associated with this modal, create our builder on
    //the fly
    var attribute_builder = builder || arc.reportbuilder().dataset(selected_dataset);

    attribute_builder.paintAttributes(attribute_builder.dataset(),
        {
          $dimDiv: $dimDiv,
          $aggDiv: $aggDiv,
          $partDiv: $partDiv,
          headersOnly: true,
          selectAllOption: true,
        });

    var templateCol = $("#createview-column-template ul").html();
    function paintCol($container, colname) {
      $container.append(arc.misc.safe_substitute(
          templateCol,
          {
            VALUE: colname,
          }));
    }

    function paintColumns($container, columns) {
      columns.forEach(function(col) {
        var colname = col.attr('alias') || col.attr('name');
        colsByName[colname] = col;
        paintCol($container, colname);
      });
    }

    var templateHdr = $("#createview-column-hdr-template").html();
    $dimDiv.append(templateHdr);
    $aggDiv.append(templateHdr);
    $partDiv.append(templateHdr);

    var tablename = $("#create-aview-table option:selected").text();
    tables.some(function(table) {
      if (table.tablename() === tablename) {
        paintColumns($dimDiv.find(".table-list"),
            table.dimensionsOnly()
                 .filter(function(col) {
                   return (!col.attr('name') || !col.attr('ispart')) &&
                     !col.attr('ispartition') && !col.isHierarchy();
                 }));
        paintCol($aggDiv.find(".table-list"), 'Record Count');
        paintColumns($aggDiv.find(".table-list"),

            // Note that we are only plotting measures by calling 'measuresOnly()'
            // and ignoring hierarchies. For backward compatibility reasons, aggregates()
            // returns hierarchies as well.
            table.measuresOnly()
                 .filter(function(col) {
                   return !col.attr('ispartition');
                 }));
        paintColumns($partDiv.find(".table-list"),
            table.base_columns()
                 .filter(function(col) {
                   return col.attr('ispart') || col.attr('ispartition');
                 }));
        const $partitionCheckboxes = $partDiv.find('input[type="checkbox"]');

        // Check all partition checkboxes assuming there are any beyond just the select all one
        // in the header
        if ($partitionCheckboxes.length > 1) {
          $partitionCheckboxes.prop('checked', true);
        }

        return true;
      }
    });

    $modal.find('.aview-column-checkbox').on('change.checkstatus', function updateSelectAll(evt) {
      const $target = $(evt.target);
      const $selectAll = $target.closest('.attribute-wrapper').find('.select-all-fields');
      if ($selectAll.prop('checked') && !$target.prop('checked')) {
        $selectAll.prop('checked', false);
      }
    });

    $modal.find('.table-list-value')
          .off("click")
          .on("click", function(e, i) {
            var $checkbox = $(this).prev("input");
            $checkbox.prop("checked", ($checkbox.prop("checked") === false))
                     .trigger("change");
          });

    $partDiv.find('input').removeAttr('data-atleastonedimagg');

  }

  function doCreate() {
    const $viewNameForm = $modal.find('#create-aview-name');
    const viewName = $viewNameForm.find('input').val();

    const $viewLocationForm = $modal.find('#create-aview-location');
    let viewLocation = $viewLocationForm.find('select').val();

    selectedDbName = $modal.find('#aview-db-name').val();
    let complimentedViewName = selectedDbName + dbTableDelimiter + viewName;

    var dataset = builder ? builder.dataset() : selected_dataset;
    var rb = arc.requestBuilder();
    rb.limit(undefined);
    rb.requestType('createaview');
    if (selectedTableName === lvtable) {
      rb.useAssocLV(true);
    } else {
      rb.useAssocLV(false);
    }
    rb.datasetId(dataset.dataset_id());
    rb.tableName(complimentedViewName);

    if (viewLocation) {
      rb.viewLocation(`${viewLocation}/${selectedDbName}.db/${viewName}`);
    }

    const $dimPicked = $dimDiv.find('input.aview-column-checkbox:checked');
    const $aggPicked = $aggDiv.find('input.aview-column-checkbox:checked');
    const $partPicked = $partDiv.find('input.aview-column-checkbox:checked');

    function addShelfElementToRb(rb, shelf_elem, columnType, isPartition) {
      var se = arc.report.shelfElementPostProcess({
        shelf_elem: shelf_elem,
        shelf_type: arc.report.shelf_types.COLUMN,
        shelf_property: columnType,
        discardFunctionName: true,
      });
      if (columnType === arc.report.columnTypeEnum.DIMENSION) {
        if (isPartition) {
          rb.addPartition(se.getExpression());
        } else {
          rb.addDimension(se.getExpression());
        }
      } else {
        rb.addAggregate(se.getExpression());
      }
    }

    function addColToRb(rb, col, columnType, isPartition) {
      var shelf_elem = arc.report.columnToShelfElement(dataset, col);
      addShelfElementToRb(rb, shelf_elem, columnType, isPartition);
    }

    $dimPicked.each(function() {
      var colname = $(this).attr("data-value");
      var col = colsByName[colname];
      addColToRb(rb, col, arc.report.columnTypeEnum.DIMENSION);
    });
    $aggPicked.each(function() {
      var colname = $(this).attr("data-value");
      if (colname === 'Record Count') {
        addShelfElementToRb(rb,
            arc.report.shelf.record_count_element(dataset.dataconnection_type()).dataset(dataset),
            arc.report.columnTypeEnum.AGGREGATE);
        return;
      }
      var col = colsByName[colname];
      addColToRb(rb, col, arc.report.columnTypeEnum.AGGREGATE);
    });
    $partPicked.each(function() {
      var colname = $(this).attr("data-value");
      var col = colsByName[colname];
      addColToRb(rb, col, arc.report.columnTypeEnum.DIMENSION,
          true /* isPartition */);
    });

    arc.report.sql_runner.run(dataset.dataconnection_id(), '',
      function(errorMsg, resp) {
        function alertCb() {
          arc.report.alertMessageSmall(`Analytical view ${complimentedViewName} created`,
              "alert-success");
        }

        if (errorMsg) {
          arc.report.alertMessageAllHide();
          arc.report.alertMessage("Failed to create analytical view " + complimentedViewName,
              errorMsg);
        } else {
          $viewNameForm.val("");
          if (createCb) {
            const avMgr = creator.aviewManager();
            if (avMgr) {
              const selections = avMgr.selectedInputs();
              selections.all = false;
              selections.tables[selectedTableName] = false;
            }
            createCb(alertCb);
          }
        }
        $modal.modal('hide');

      }, false, rb.dsreq());
  }

  function paintModal() {
    if (arc.misc.modalShown($modal)) {
      return;
    }

    // Need a dataset in order to open the create analytical view modal.
    displayOrHideDatasets();
    if (builder) {
      if (!builder.dataset()){
        return noDatasetsAlert();
      }
    } else if ($.isEmptyObject(datasetsById)) {
      return noDatasetsAlert();
    }

    const dc = aviewManager.dataconnection_info();
    const locs = ((dc && dc.info ? JSON.parse(dc.info) : {}).PARAMS || {}).LOCATIONS || {};
    const keys = Object.keys(locs);
    const $locsel = $modal.find('#create-aview-location');
    $locsel.toggleClass('arc-invisible', keys.length === 0);
    if (keys.length) {
      const $select = $locsel.find('select');
      $select.empty();
      $select.append($(`<option value="" selected>(System default)</option>`));
      keys.forEach(key => $select.append($(`<option value="${locs[key]}">${key}</option>`)));
    }

    let $modalSpinner = $modal.find('.arc-spinner');
    let $modalForm = $modal.find('#aview-form');
    $modalForm.addClass('arc-invisible');
    $modalSpinner.removeClass('arc-invisible');
    $modal.modal('show');

    displayTables();

    buildNameInputAndAttrs(function() {
      $modal.find(".create-confirmation-spinner").addClass("arc-invisible");
      $modal.on("hidden.bs.modal", function() {
        $modal.find("#create-aview-name input").val("");
      });
      $modal.find('#create-confirmation').addClass('disabled');
      $modalSpinner.addClass("arc-invisible");
      $modalForm.removeClass("arc-invisible");
    });
  }

  function noDatasetsAlert() {
    arc.report.alertMessageSmall('Please create a dataset before creating an analytical view.',
      'alert-info');
  }

  creator.create = function(_createCb) {
    createCb = _createCb;

    paintModal();

    return creator;
  };

  return creator;
};
}());
