import arc from 'arc';

export const ParamEnums = {
  DataTypeEnum: {
    // UNKNOWN is considered as STRING. We do not want it to be set as standard datatype
    // BOOLEAN is not being used at the moment. We currently consider it as string in code.
    STRING: 'STRING',
    NUMERIC: 'NUMERIC',
    TIMESTAMP: 'TIMESTAMP',
    BOOLEAN: 'BOOLEAN',
  },

  ValueTypeEnum: {
    SINGLE: 'single',
    MULTI: 'multi',
    RANGE: 'range',
  },

  ParamTypeEnum: { // default PARAM. Maybe we will derive different classes instead
    PARAM: 'param',
    PIVOT: 'pivot',
    CONFIG: 'config',
    ARCPARAM: 'arcparam',
  },

  ClassTypeEnum: {
    Param: 'Param',
    ColParam: 'ColParam',
    PivotParam: 'PivotColParam',
    CustomParam: 'CustomParam',
    RangeValueParam: 'RangeParam',
    DateRangeParam: 'DateRangeParam',
  },

  EntityTypeEnum: {
    FILTER: 'filter',
    VISUAL: 'visual',
    SHEET: 'sheet',
    DASHBOARD: 'dashboard',
    APP: 'app',
    URL: 'url', // maybe url shouldn't be needed
  },

  ScopeTypeEnum: {
    DATASET: 'dataset',
    EXPLICIT: 'explicit',
    GLOBAL: 'global',
  },

  DateEnums: {
    DateRangeTypeEnum: {
      FIXED: 'fixed',
      RELATIVE: 'relative',
      QUICK: 'quick_filters',
    },
    DateTimeFormats: {
      DATE: ['YYYY-MM-DD'],
      DATETIME: ['YYYY-MM-DD HH:mm:ss', 'YYYY-MM-DDTHH:mm:ss[Z]'],
    }
  },
  getKeyForValue(value, enumName) {
    let keyName;
    if (!enumName) {
      return undefined;
    }

    Object.keys(enumName).some(function findValueInEnum(key) {
      if (enumName[key] === value) {
        keyName = key;
        return true;
      }

      return false;
    });

    return keyName;
  }
};

/**
 * Base Param Object Class
 * @property {string} _outputParameter  - Defines the root keyname of the parameter being emitted.
 * @property {object} _source = {       - Stores the properties of emitting entity and column
    column: {                           - Represents the base column for which param is emitted
      baseField,                        - actual column name in dataset
      dataType,                         - ParamEnums.DataTypeEnum.*
      connectionType,                   - 'solr', 'ksql', etc.
    },
    entity: {                           - Stores the properties of emitting entity = visual/filter
      entityId,                         - report_id of source visual or filter. For URL its 0.
      entityType,                       - ParamEnums.EntityTypeEnum.*
      sheetId,                          - id of the sheet from which param was emitted
      dashboardId,                      - id of the db    from which param was emitted
      appId,                            - id of the app from which param was emitted (if in app)
      datasetId,                        - id of the dataset of source entity.
    }
 }
 * @property {object} _scope = {        - scope where the param should be applicable
    scopeType,                          - ParamEnums.ScopeTypeEnum.*
    isAppScope,                         - true/false
 }
 * @property {object} _visibility = {
    entityType,                         - ParamEnums.EntityTypeEnum.*
 }
 * @property {object} _meta = {
    valueType,                          - ParamEnums.ValueTypeEnum.*
    paramType,                          - ParamEnums.ParamTypeEnum.*
    keyType,                            - arc.report.columnTypeEnum.*
 }
 * @property {object} _state = {
    isIgnore,                           - true/false (helps paramsHistory to delete params)
    hideExtension,                      - true/false
    // if hideExtension is true, when converting to oldParamFormat, extension is skipped
    // hideExtension is set to true for dashboards with conflicting params
 }
 * @property {array} _value             - stores values in their original format (as in database)
 * @property {array} _isExcluded        - true/false applicable only for multi-value param.
 *
 * @property {boolean} _includeNulls    - true/false
 *
 * @method initializeParam              - Initializes the object (helper to constructor)
 * @method toOldParamFormat             - Converts param object to older json format
 */
export class Param {
  /**
   *
   * @param lib                         - base object corresponding to source visual/filter entity
   * @param paramConfig = {             - object used to initialise Param object
     baseField,
     colDataType,
     entityType,
     entityId,
     arcapi,                            - passed when param is created from custom JS
     connectionType,
     datasetId,
     isAppScope,
     valueType,
     keyType,
     paramType,
     outputParameter,
     exclude,
     hideExtension,
     value,
     isIgnore,
     isDatasetSegment,
   }
   */
  constructor(lib, paramConfig = {}) {
    this.name = 'Param';

    const sourceObj = {
      baseField: paramConfig.baseField,
      colDataType: paramConfig.colDataType || 'UNKNOWN',
      entityType: paramConfig.entityType || ParamEnums.EntityTypeEnum.FILTER,
      connectionType: paramConfig.connectionType,
      datasetId: paramConfig.datasetId,
      sheetId: paramConfig.sheetId || (
            arc.globals.ms_dashboard && arc.globals.ms_dashboard.getCurrentlyActiveSheet() &&
            arc.globals.ms_dashboard.getCurrentlyActiveSheet().sheet_id()
          ) || undefined,
      dashboardId: paramConfig.dashboardId ||
          (arc.globals.ms_dashboard && arc.globals.ms_dashboard.dashboard_id()) ||
          undefined,
      appId: paramConfig.appId ||
          (arc.globals.appgroup && arc.globals.appgroup.appgroup_id()) ||
          undefined,
    };

    this._isDatasetSegment = paramConfig.isDatasetSegment;

    const entityIdValue = paramConfig.entityId ||
                        (paramConfig.arcapi && paramConfig.arcapi.getExtensionId());
    if (entityIdValue !== null && entityIdValue !== undefined) {
      sourceObj.entityId = entityIdValue;
    }

    const scopeType = paramConfig.scopeType ||
      lib?.scopetype() ||
      ParamEnums.ScopeTypeEnum.EXPLICIT;

    const scopeObj = {
      isAppScope: false,
      scopeType,
    };

    if (paramConfig.isAppScope !== undefined) {
      scopeObj.isAppScope = !!paramConfig.isAppScope;
    } else if (lib) {
      scopeObj.isAppScope = lib.isAppScope();
    }

    this._setScope(scopeObj);

    if (typeof paramConfig.appId !== 'undefined') {
      sourceObj.appId = paramConfig.appId;
    } else {
      let appId;
      if (arc.globals.appgroup) {
        // We are in a launched app context
        appId = arc.globals.appgroup.appgroups()[0].group_id;
      } else if (arc.globals.ms_dashboard) {
        // We are in a dashboard but not in a launched app context
        appId = arc.globals.ms_dashboard.dashboard_id();
      } else if (arc.globals.reports) {
        // We are in a single visual mode and not in a launched app context
        const primaryReport = arc.globals.reports.getPrimaryReport();
        if (primaryReport) {
          appId = primaryReport.report_id();
        }
      }

      sourceObj.appId = appId;
    }

    if (lib) {
      if (!sourceObj.entityId) {
        sourceObj.entityId = lib.report_id();
      }

      if (typeof paramConfig.entityType === 'undefined') {
        sourceObj.entityType = ['daterange', 'range', 'picklist'].includes(lib.viz_type()) ?
          ParamEnums.EntityTypeEnum.FILTER :
          ParamEnums.EntityTypeEnum.VISUAL;
      }

      if (!sourceObj.baseField) {
        const baseFieldSetting = lib.user_settings().find_by_id('picklist_base_field') ||
          lib.user_settings().find_by_id('daterange_base_field');
        if (baseFieldSetting) {
          sourceObj.baseField = lib.substituteParams(baseFieldSetting.value());
        }
      }

      const dataset = lib.dataset();
      if (dataset) {
        if (typeof paramConfig.colDataType === 'undefined' && sourceObj.baseField) {
          const col = dataset.findColumn(sourceObj.baseField);
          if (col) {
            // col.coltype() - in brush.cb in visual_bar.js
            // col.attr('type') - in getColtype() in picklist.js
            sourceObj.colDataType = col.coltype ? col.coltype() : col.attr('type');
          }
        }

        if (!sourceObj.connectionType) {
          sourceObj.connectionType = dataset.dataconnection_type();
        }

        if (!sourceObj.datasetId && scopeType === ParamEnums.ScopeTypeEnum.DATASET) {
          sourceObj.datasetId = dataset.dataset_id();
        }
      }
    }

    this._setSource(sourceObj);

    const metaObj = {
      valueType: paramConfig.valueType || ParamEnums.ValueTypeEnum.SINGLE,
      keyType: paramConfig.keyType || arc.report.columnTypeEnum.DIMENSION,
      paramType: paramConfig.paramType || ParamEnums.ParamTypeEnum.PARAM,
    };

    this._setMeta(metaObj);

    this._state = {
      isIgnore: !!paramConfig.isIgnore,
      hideExtension: !!paramConfig.hideExtension, // TODO: Remove its need for Multi-select params
    };

    this._setValue(paramConfig.value || []);
    this._outputParameter = paramConfig.outputParameter || '';
    this._isExcluded = !!paramConfig.isExcluded;
    this._includeNulls = !!paramConfig.includeNulls;
  }

  get meta() {
    // We do not want user to be able to edit object accessed from getter
    return arc.misc.extend({}, this._meta);
  }

  _setMeta(metaObj = {}) {
    if (!this._meta) {
      this._meta = {};
    }

    let key = ParamEnums.getKeyForValue(metaObj.valueType, ParamEnums.ValueTypeEnum);
    if (key) {
      this._meta.valueType = ParamEnums.ValueTypeEnum[key];
    } else if (typeof metaObj.valueType !== 'undefined') {
      Param.throwParamException('Invalid value type.');
    }

    key = ParamEnums.getKeyForValue(metaObj.keyType, arc.report.columnTypeEnum);
    if (key) {
      this._meta.keyType = arc.report.columnTypeEnum[key];
    } else if (typeof metaObj.keyType !== 'undefined') {
      Param.throwParamException('Invalid key type.');
    }

    key = ParamEnums.getKeyForValue(metaObj.paramType, ParamEnums.ParamTypeEnum);
    if (key) {
      this._meta.paramType = ParamEnums.ParamTypeEnum[key];
    } else if (typeof metaObj.keyType !== 'undefined') {
      Param.throwParamException('Invalid param type.');
    }
  }

  get source() {
    const sourceObj = {
      colDataType: this._source.column.dataType,
      baseField: this._source.column.baseField,
      connectionType: this._source.column.connectionType,

      entityType: this._source.entity.entityType,
      entityId: this._source.entity.entityId,
      sheetId: this._source.entity.sheetId,
      dashboardId: this._source.entity.dashboardId,
      appId: this._source.entity.appId,
      datasetId: this._source.entity.datasetId,
    };

    return sourceObj;
  }

  _setSource(sourceObj = {}) {
    if (!this._source) {
      this._source = {
        column: {},
        entity: {},
      };
    }

    if (sourceObj.baseField) {
      this._source.column.baseField = sourceObj.baseField;
    }

    if (typeof sourceObj.connectionType !== 'undefined') {
      this._source.column.connectionType = sourceObj.connectionType;
    }

    if (typeof sourceObj.entityType !== 'undefined') {
      this._source.entity.entityType = sourceObj.entityType;
    }

    if (typeof sourceObj.entityId !== 'undefined') {
      this._source.entity.entityId = sourceObj.entityId;
    }

    if (typeof sourceObj.datasetId !== 'undefined') {
      this._source.entity.datasetId = sourceObj.datasetId;
    }

    if (sourceObj.colDataType) {
      this._setDataType(sourceObj.colDataType);
    }

    if (typeof sourceObj.sheetId !== 'undefined') {
      this._source.entity.sheetId = sourceObj.sheetId;
    }

    if (typeof sourceObj.dashboardId !== 'undefined') {
      this._source.entity.dashboardId = sourceObj.dashboardId;
    }

    if (typeof sourceObj.appId !== 'undefined') {
      this._source.entity.appId = sourceObj.appId;
    }
  }

  get dataType() {
    return this._source.column.dataType;
  }

  _setDataType(val) {
    let _dataType;
    const upperCaseVal = String(val).toUpperCase();
    const key = ParamEnums.getKeyForValue(upperCaseVal, ParamEnums.DataTypeEnum);
    if (key) {
      _dataType = ParamEnums.DataTypeEnum[key];
    } else if (upperCaseVal === 'UNKNOWN') {
      _dataType = ParamEnums.DataTypeEnum.STRING;
    }

    if (!_dataType) {
      _dataType = arc.misc.isString(val, this._source.column.connectionType) ?
        ParamEnums.DataTypeEnum.STRING :
        ParamEnums.DataTypeEnum.NUMERIC;
    }

    this._source.column.dataType = _dataType;
  }

  get isDatasetSegment() {
    return this._isDatasetSegment;
  }

  get scope() {
    return {
      scopeType: this._scope.scopeType,
      isAppScope: this._visibility.entityType === ParamEnums.EntityTypeEnum.APP,
    };
  }

  _setScope(scopeObj = {}) {
    if (!this._scope) {
      this._scope = {};
    }

    if (!this._visibility) {
      this._visibility = {};
    }

    if (ParamEnums.getKeyForValue(scopeObj.scopeType, ParamEnums.ScopeTypeEnum)) {
      this._scope.scopeType = scopeObj.scopeType;
    } else {
      this.throwParamException(`Invalid scopeType: ${scopeObj.scopeType}`);
    }

    this._visibility.entityType = scopeObj.isAppScope ?
      ParamEnums.EntityTypeEnum.APP : ParamEnums.EntityTypeEnum.SHEET;
  }

  get isExcluded() {
    return this._isExcluded;
  }

  get includeNulls() {
    return this._includeNulls;
  }

  get outputParameter() {
    return this._outputParameter;
  }

  get hideExtension() {
    return this._state.hideExtension;
  }

  get value() {
    return this._value && this._value.slice(0);
  }

  _setValue(val) {
    if (Param._isValidValue(val)) {
      this._updateValue(val);
    } else {
      Param.throwParamException('Invalid param value format.');
    }
  }

  get isIgnore() {
    return this._state.isIgnore;
  }

  _updateValue(val) {
    this._value = val.slice(0);
  }

  static _isValidValue(val) {
    if (!(val instanceof Array)) {
      return false;
    }

    // TODO: Check for dataType matches
    return true;
  }

  static throwParamException(message) {
    throw new Error(message);
  }

  toJSON() {
    const jsonifiedObject = {};
    Object.keys(this).forEach((key) => {
      jsonifiedObject[key] = this[key];
    });
    jsonifiedObject._className = ParamEnums.ClassTypeEnum[this.name];
    delete jsonifiedObject.name;
    return jsonifiedObject;
  }

  importFromJSON(jsonifiedObject = {}) {
    /* eslint-disable no-undef */
    const parsedObject = $.extend({}, jsonifiedObject);
    /* eslint-enable no-undef */
    delete parsedObject._className;
    delete parsedObject.name;
    console.log({ parsedObject });
    Param._populateObject(parsedObject, this);
  }

  static _populateObject(src = {}, dest = {}) {
    /*  eslint no-param-reassign: [
          "error", { "props": true, "ignorePropertyModificationsFor": ["dest"] }
        ]
    */
    Object.getOwnPropertyNames(src).forEach((key) => {
      if (typeof src[key] !== 'object') {
        dest[key] = src[key];
      } else {
        dest[key] = dest[key] || {};
        Param._populateObject(src[key], dest[key]);
      }
    });
  }

  toOldParamFormat() {
    const oldFormatParam = {};
    if (this.isIgnore) {
      oldFormatParam[`${this._outputParameter}.ignore`] = [true];
      return oldFormatParam;
    }

    oldFormatParam[`${this._outputParameter}.include_nulls`] = [String(this.includeNulls)];

    if (!this.value.length) {
      return oldFormatParam;
    }

    if (this._meta.valueType === ParamEnums.ValueTypeEnum.SINGLE || this.shouldHideExtension()) {
      oldFormatParam[this._outputParameter] = this.value;
    } else if (this._meta.valueType === ParamEnums.ValueTypeEnum.MULTI) {
      oldFormatParam[`${this._outputParameter}.data`] = this.value;
      oldFormatParam[`${this._outputParameter}.exclude`] = this._isExcluded ? ['not in'] : ['in'];
    }

    return oldFormatParam;
  }

  shouldHideExtension() {
    return this._state.hideExtension && this._value &&
      this._value.length === 1 && !this._isExcluded;
  }
}

/**
 * @property {array} _alias             - stores alias labels for selected values
 */
export class CustomParam extends Param {
  constructor(lib, paramConfig = {}) {
    super(lib, paramConfig);

    this.name = 'CustomParam';

    this._alias = [];
    this._setAlias(paramConfig.alias || paramConfig.value || []);
  }

  get alias() {
    return this._alias && this._alias.slice(0);
  }

  _setAlias(val) {
    if (Param._isValidValue(val)) {
      this._alias = val.slice(0);
    } else {
      Param.throwParamException('Invalid param alias format.');
    }
  }

  toOldParamFormat() {
    const oldFormatParam = super.toOldParamFormat();
    if (this.isIgnore || !this.value.length) {
      return oldFormatParam;
    }

    if ([
      ParamEnums.ValueTypeEnum.SINGLE,
      ParamEnums.ValueTypeEnum.MULTI
    ].includes(this._meta.valueType)) {
      oldFormatParam[`${this.outputParameter}.alias`] = this.alias;
    }

    return oldFormatParam;
  }
}
