import arc from 'arc';
import { ParamEnums, Param } from './paramClass';

/**
 * @property {object} _value = [       - Array containing 2 values (start, end)
 ]
 _value[0] => start
 _value[1] => end
 */
export class RangeValueParam extends Param {
  constructor(lib, paramConfig = {}) {
    super(lib, paramConfig);

    this.name = 'RangeValueParam';

    this._source.column.dataType = ParamEnums.DataTypeEnum.NUMERIC;
    this._meta.valueType = ParamEnums.ValueTypeEnum.RANGE;
  }

  get value() {
    return [{
      start: this._value && this._value[0],
      end: this._value && this._value[1],
    }];
  }

  _setValue(valArr) {
    const defaultValue = {
      start: null,
      end: null,
    };

    if (Param._isValidValue(valArr) && (!valArr[0] || RangeValueParam._isValidValue(valArr[0]))) {
      const val = valArr[0] || defaultValue;
      this._value = [Number(val.start), Number(val.end)];
    } else {
      Param.throwParamException('Invalid param value format.');
    }
  }

  static _isValidValue(val) {
    return typeof val === 'object' &&
      (!(val instanceof Array)) &&
      typeof val.start !== 'undefined' &&
      typeof val.end !== 'undefined';
  }

  toOldParamFormat() {
    const oldFormatParam = {};
    if (this.isIgnore) {
      oldFormatParam[`${this._outputParameter}.ignore`] = [true];
      return oldFormatParam;
    }

    if (this._value) {
      if (this.hideExtension) {
        oldFormatParam[this._outputParameter] = this._value.slice(0);
      } else {
        oldFormatParam[`${this._outputParameter}.start`] = [this._value[0]];
        oldFormatParam[`${this._outputParameter}.end`] = [this._value[1]];
      }

      oldFormatParam[`${this._outputParameter}.include_nulls`] = [String(this.includeNulls)];
    }

    return oldFormatParam;
  }
}

/**
 * @property {object} _value = [       - Array containing 2 values (start, end)
 ]
 _value[0] => start (should always be a string with either date, datetime or timestamp)
 _value[1] => end   (should always be a string with either date, datetime or timestamp)
 * @property {object} _meta = {
   dateRangeType
 }
 * @property {object} _config = {
   datetimeFormat,
   relativeDate,
   quickFilters,
 }
 */
export class DateRangeParam extends RangeValueParam {
  constructor(lib, paramConfig = {}) {
    super(lib, paramConfig);

    this.name = 'DateRangeValueParam';
    this._config = this._config || {};
    this._meta.dateRangeType = ParamEnums.DateEnums.DateRangeTypeEnum.FIXED;
    this._source.column.dataType = ParamEnums.DataTypeEnum.TIMESTAMP;

    let { dateRangeType, datetimeFormat } = paramConfig;
    if (lib) {
      if (!dateRangeType && lib.viz_type() === 'daterange') {
        dateRangeType = lib.viz().getDateRangeType();
      }

      if (!datetimeFormat) {
        const showTimeSetting = lib.user_settings().find_by_id('show_time');
        const isDateTime = (showTimeSetting && showTimeSetting.isChecked) ||
                           dateRangeType === ParamEnums.DateEnums.DateRangeTypeEnum.QUICK;
        datetimeFormat = isDateTime ?
          ParamEnums.DateEnums.DateTimeFormats.DATETIME[0] :
          ParamEnums.DateEnums.DateTimeFormats.DATE[0];
      }
    }

    if (dateRangeType) {
      this._setDateRangeType(dateRangeType);
    }

    if (datetimeFormat) {
      this._setDatetimeFormat(datetimeFormat);
    }

    if (paramConfig.relativeDate) {
      // calc the relative date instead of saving the hardcoded values from config
      const data = paramConfig.relativeDate;

      arc.filter.helpers.fetchRelative(data, data.datetimeFormat);

      this._setValue([{ start: data.startDate, end: data.endDate }]);
      this._setRelativeDate(data);
    }

    if (paramConfig.quickFilters) {
      this._setQuickFilters(paramConfig.quickFilters);
    }
  }

  get dateRangeType() {
    return this._meta.dateRangeType;
  }

  _setValue(valArr) {
    const defaultValue = {
      start: null,
      end: null,
    };

    if (Param._isValidValue(valArr) && (!valArr[0] || RangeValueParam._isValidValue(valArr[0]))) {
      const val = valArr[0] || defaultValue;
      this._value = [val.start, val.end];
    } else {
      Param.throwParamException('Invalid param value format.');
    }
  }

  _setDateRangeType(dateRangeType) {
    const key = ParamEnums.getKeyForValue(dateRangeType, ParamEnums.DateEnums.DateRangeTypeEnum);
    if (!key) {
      throw new Error('Invalid date range type');
    }

    this._meta.dateRangeType = ParamEnums.DateEnums.DateRangeTypeEnum[key];
  }

  get config() {
    return arc.misc.extend({}, this._config);
  }

  get datetimeFormat() {
    return this._config.datetimeFormat;
  }

  _setDatetimeFormat(val) {
    if (DateRangeParam._isValidDateTimeFormat(val)) {
      this._config.datetimeFormat = val;
    } else {
      Param.throwParamException('Invalid format for datetimeFormat');
    }
  }

  get relativeDate() {
    if (this._config && this._config.relativeDate) {
      return arc.misc.extend({}, this._config.relativeDate);
    }

    return null;
  }

  _setRelativeDate(val) {
    if (DateRangeParam._isValidRelativeDate(val)) {
      this._config.relativeDate = val;

      this._setDateRangeType(ParamEnums.DateEnums.DateRangeTypeEnum.RELATIVE);
      if (val.datetimeFormat) {
        this._setDatetimeFormat(val.datetimeFormat);
      }
    } else {
      Param.throwParamException('Invalid format for relative date');
    }
  }

  get quickFilters() {
    if (this._config && this._config.quickFilters) {
      return arc.misc.extend({}, this._config.quickFilters);
    }

    return null;
  }

  _setQuickFilters(val) {
    if (DateRangeParam._isValidQuickFilter(val)) {
      this._config.quickFilters = arc.misc.extend({}, val);
      this._setDateRangeType(ParamEnums.DateEnums.DateRangeTypeEnum.QUICK);
      if (val.datetimeFormat) {
        this._setDatetimeFormat(val.datetimeFormat);
      }
    } else {
      Param.throwParamException('Invalid format for quick filter');
    }
  }

  static _isValidRelativeDate(val) {
    // TODO: Make validation for startDate/endDate more stringent
    // Possible values: new Date().getTime(), moment.utc(date), moment(date)
    return typeof val === 'object' &&
      typeof val.startDate !== 'undefined' &&
      typeof val.endDate !== 'undefined';
  }

  static _isValidQuickFilter(val) {
    return typeof val === 'object' && typeof val.name !== 'undefined';
  }

  static _isValidDateTimeFormat(val) {
    return val &&
      (ParamEnums.DateEnums.DateTimeFormats.DATE.includes(val) ||
        ParamEnums.DateEnums.DateTimeFormats.DATETIME.includes(val));
  }

  toOldParamFormat() {
    const paramOldFormat = super.toOldParamFormat();
    let configObj;
    if (this.isIgnore) {
      return paramOldFormat;
    }

    if (Object.keys(this._config).length) {
      // relative, quick filter params

      if (typeof this._config.relativeDate !== 'undefined') {
        configObj = {
          [arc.misc.relativeFilterSuffix]: this._config.relativeDate,
        };
        paramOldFormat[`${this._outputParameter}.config`] = [JSON.stringify(configObj)];
      }

      if (typeof this._config.quickFilters !== 'undefined') {
        configObj = {
          '.btn': this._config.quickFilters,
        };
        paramOldFormat[`${this._outputParameter}.config`] = [JSON.stringify(configObj)];
      }
    }

    return paramOldFormat;
  }
}
