import { stemmer } from 'stemmer';

window.arc = window.arc || {};

(function() {
'use strict';

arc.search = arc.search || {};
arc.search.rule = arc.search.rule || {};

let verbose = false;

arc.search.setVerbose = function(_) {
  verbose = _;
};

const forceDimensionWords = new Set([
  'base',
  'simple',
  'unaggregated',
  'raw',
  'dimensional',
  'dimension',
]);

// Special words that can be either meausre or dimension
const nullWords = new Set([
  'null',
  'nulls',
]);

const booleanWords = new Set([
  'true',
  'false',
]);

const dimensionType = {
  IGNORE: 'ignore',
};

const filterType = {
  FIXED: 'fixed',
};

const datetimeType = {
  RELATIVE: 'relative',
};

const tokenRule = {
  UNQUOTED_STRING: 'unquoted_string',
};

const compareType = {
  CONTAINS: 'contains',
  EQUALITY: 'equality',
  INEQUALITY: 'inequality',
  STARTSWITH: 'startswith',
  ENDSWITH: 'endswith',
  EQ: '=',
  NE: '!=',
  GT: '>',
  GTE: '>=',
  LT: '<',
  LTE: '<=',
  BETWEEN: 'between',
};

// If a new type is added, also add to arc_search_highlight > typesData and search_menus
// to handle underlining and context menu
const elementType = {
  FILTER: 'filter',
  DATETIME: 'datetime',
  DIMENSION: 'dimension',
  AGGREGATE: 'aggregate',
  TREND: 'trend',
  DISTINCT: 'distinct',
  ORDER: 'order',
  EACH: 'each',
  EXPANSION: 'expansion',
  VISUALTYPE: 'visual type',
  TOP: 'top',
  BOTTOM: 'bottom',
  NEW: 'new',
  OLD: 'old',
  KEYVALUE: 'keyvalue',
  CONTROL: 'control',
  COMPARE: 'compare',
  CONNECTOR: 'connector',
  STR: 'str',
  IGNORE: 'ignore',
  AGGREGATION: 'aggregation',
  NEGATION: 'negation',
  UNKNOWN: 'unknown',
};

const termType = {
  COLUMN: 'column',
};

const tokenizerRegex = /\s+/;
function getWordCount(string) {
  return string.split(tokenizerRegex).length;
}

// Add a multiplier for each additional word over the subtractValue of words
// so multi-word counts subtly more than just single word matches
// eg: 1 word - add 1
//     2 words - add 2.5
function calculateMultiwordScore(phrase, multiplier = 1.5, subtractValue = 1) {
  const multiWordAmplifier = (getWordCount(phrase) - subtractValue) * multiplier;
  return subtractValue + Math.max(multiWordAmplifier, 0);
}

function quoteValue(column, value) {
  if (typeof (value) !== 'string') {
    if (column.valuesNeedQuoting()) {
      value += '';
    } else {
      return value;
    }
  }

  const lv = value.toLowerCase();
  if (nullWords.has(lv)) {
    return 'NULL';
  }

  if (value[0] !== "'" && (column.valuesNeedQuoting() || !$.isNumeric(value))) {
    return `'${value}'`;
  }

  return value;
}

function aggExpression(c, aggregation) {
  const displayName = arc.misc.quoteColumn(c.display_name());
  if (c.attr('exprhasagg')) {
    return displayName;
  }

  if (aggregation === undefined) {
    aggregation = c.defaultAggregate();
  }

  if (aggregation) {
    const lc = aggregation.toLowerCase();

    if (lc === 'countdistinct') {
      const dcType = c.table().dataset().dataconnection_type();
      return arc.dataconnection.exactCount(dcType, displayName);
    }

    if (lc === 'approximate_count') {
      const dcType = c.table().dataset().dataconnection_type();
      return arc.dataconnection.approximateOrExactCount(dcType, displayName);
    }
  }

  return c.attr('recordcount') ? 'sum(1)' :
    `${aggregation}(${displayName})`;
}

function phraseToTerms(phrase) {
  const searchTokenizer = arc.grammar._searchtokens;
  const { tokens } = searchTokenizer.parse(phrase, {
    startRule: 'strings',
  });

  tokens.forEach((t) => {
    if (t.phrase) {
      t.phraseLowerCase = t.phrase.toLowerCase();
    } else {
      t.phraseLowerCase = '';
    }
  });

  const q = query().setTokens(tokens)
                 .initialize();
  return q;
}

function phraseToQuery(phrase) {
  // remove trailing question marks
  phrase = phrase.replace(/\?+$/, '');

  const searchTokenizer = arc.grammar._searchtokens;
  const { tokens } = searchTokenizer.parse(phrase);

  tokens.forEach((t) => {
    if (t.phrase) {
      t.phraseLowerCase = t.phrase.toLowerCase();
    } else {
      t.phraseLowerCase = '';
    }
  });

  const q = query().setTokens(tokens)
                 .initialize();
  return q;
}

arc.search.rule.base = function() {
  const r = function() { return 'arc.search.rule'; };

  let tree;

  r.tree = function(_) {
    if (!arguments.length) return tree;
    tree = _;
    return r;
  };

  r.process = function() {
    throw new Error('Rule not implemented');
  };

  r.forEachUnquotedString = function(cb) {
    r.forEachLeafWithoutElement((node) => {
      const token = node.data();
      if (token.rule === tokenRule.UNQUOTED_STRING) {
        cb(node);
      }
    });
  };

  r.forEachLeafWithoutElement = function(cb) {
    tree.walkLeaves((node) => {
      const d = node.data();
      if (d === undefined) {
        return;
      }

      if (d.element) {
        return;
      }

      cb(node);
    });
  };

  r.forEachLeaf = function(cb) {
    tree.walkLeaves((node) => {
      cb(node);
    });
  };

  r.data = function() {
    const datal = [];

    r.forEachLeaf((node) => {
      const d = node.data();
      if (d === undefined) {
        return;
      }

      datal.push(d);
    });

    return datal;
  };

  return r;
};

function printTree(elementTypes) {
  let r = arc.search.rule.base();

  r.process = function() {
    console.log('START tree');

    r.forEachLeaf(function(node) {
      const token = node.data();

      if (elementTypes) {
        const {
          element,
        } = token;

        if (elementTypes.find(element.type) < 0) {
          return;
        }
      }

      console.log(JSON.stringify(Object.assign({}, token)));
    });

    console.log('END tree');
    };

  return r;
}

function wordExpansionsRule(expansionTermsTracker) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Word expansions';
  };

  r.process = function() {
    let substitutions = 0;

    r.forEachLeafWithoutElement((node) => {
      const token = node.data();
      const w = token.phrase;

      if (!w) {
        return;
      }

      let v;
      try {
        v = expansionTermsTracker.findStemmedTerm(w, node);
        if (v === undefined) {
          return;
        }
      } catch (e) {
        console.log(w, e);
        return;
      }

      token.type = elementType.EXPANSION;
      token.highlightNote = `Expanded into: ${v}`;

      // 1 point for each word in the phrase, and 1 extra point to
      // prioritize expansion matches over individual filters or dims/aggs
      const searchTokenizer = arc.grammar._searchtokens;
      const { tokens } = searchTokenizer.parse(w, {
                          startRule: 'strings',
                        });

      substitutions += tokens.length + 1;

      token.element = {
        source: token.type,
        phrase: token.phrase,
        type: token.type,
      };

      const q = phraseToQuery(v);
      q.tree().children().forEach((c) => {
        const ct = c.data();
        if (!ct) {
          return;
        }

        if (ct.startIndex !== undefined) {
          delete ct.startIndex;
        }

        if (ct.endIndex !== undefined) {
          delete ct.endIndex;
        }

        node.addChild(c);
      });
    });

    return substitutions;
  };

  return r;
}

// Connector words that we can ignore if they're before a visual type
const typePrefix = new Set([
    'as',
]);

const twoWordVisual = {
  box: 'plot',
  correl: 'heatmap',
  group: 'bar',
  tree: 'map',
  interact: 'map',
  pack: 'bubbl',
  calendar: 'heatmap',
};

const stemToVisualType = {
  area: 'trellis-areas',
  bar: 'trellis-bars',
  boxplot: 'boxplot',
  box: 'boxplot',
  bubbl: 'packed-bubbles',
  bullet: 'bullet',

  calendar: 'calendar-heatmap',
  chord: 'chord',
  combo: 'combo',
  combin: 'combo',
  correl: 'corelation',
  corel: 'corelation',
  heatmap: 'corelation',

  // 'corelation',
  // 'corelation-flow',
  crosstab: 'crosstab',
  dendrogram: 'dendrogram',
  gaug: 'gauge',
  group: 'trellis-groupedbars',
  groupedbar: 'trellis-groupedbars',
  histogram: 'histogram',
  kpi: 'kpi',
  interact: 'leaflet',
  leaflet: 'leaflet',
  line: 'trellis-lines',
  map: 'map',
  network: 'network',
  pack: 'packed-bubbles',
  pie: 'pie',
  radial: 'radial',
  scatter: 'scatter',
  sparklin: 'sparklines',
  tabl: 'table',
  data: 'table',
  treemap: 'treemap',
  tree: 'treemap',
};

function processVisualType() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Identify visual type specification';
  };

  r.process = function() {
    let substitutions = 0;

    r.forEachUnquotedString((node) => {
      const token = node.data();
      const fullPhrase = token.phraseLowerCase.split('-');
      const lc = fullPhrase[0];
      const s = stemmer(lc);
      const vizType = stemToVisualType[s];

      if (!vizType) {
        return;
      }

      let { phrase } = token;
      if (twoWordVisual[s]) {
        let nextNode;
        let nextToken;

        // Handle hyphenated words
        if (fullPhrase[1]) {
          nextToken = {
            phraseLowerCase: fullPhrase[1].toLowerCase(),
            phrase: fullPhrase[1],
          };
        } else {
          nextNode = node.next();
          nextToken = nextNode && nextNode.data();
        }

        const stem = nextToken && stemmer(nextToken.phraseLowerCase);
        if (stem && stem === twoWordVisual[s]) {
          phrase = `${phrase} ${nextToken.phrase}`;

          if (nextNode) {
            combineNodes(node, nextNode);
          }
        }
      }

      const nextNode = node.next();
      const nextWord = nextNode && nextNode.data().phraseLowerCase;
      const nextWordIsVisual = nextWord && stemmer(nextWord) === 'visual';
      const endsWithVisual = nextWordIsVisual && nextNode.next() === undefined;
      if (endsWithVisual) {
        combineNodes(node, nextNode);
      } else if ((node.prev() !== undefined) && (node.next() !== undefined)) {
        return;
      }

      const prevWord = node.prev();
      if (prevWord) {
        const prevToken = prevWord.data();
        if (typePrefix.has(prevToken.phraseLowerCase)) {
          prevWord.remove();
        }
      }

      token.type = elementType.VISUALTYPE;
      token.highlightNote = 'Sets visual type';
      substitutions += 1;

      token.element = {
        source: token.type,
        phrase,
        type: token.type,
        vizType,
      };
    });

    return substitutions;
  };

  return r;
}

function processDistinct() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Identify distinct request';
  };

  r.process = function() {
    r.forEachUnquotedString((node) => {
      const token = node.data();
      const lc = token.phrase.toLowerCase();
      const slc = stemmer(lc);

      // uniqu is used instead of unique
      // Read more about stemmer and how to use it: https://www.npmjs.com/package/stemmer
      if ((slc === 'uniqu') || (slc === 'distinct')) {
        token.type = elementType.DISTINCT;
        token.highlightNote = 'Show distinct values';

        token.element = {
          source: token.type,
          phrase: token.phrase,
          type: token.type,
        };
      }
    });
  };

  return r;
}

function processTrending() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Identify trending request';
  };

  r.process = function() {
    r.forEachUnquotedString((node) => {
      const token = node.data();
      const lc = token.phrase.toLowerCase();

      if (stemmer(lc) === 'trend') {
        token.type = elementType.TREND;
        token.highlightNote = 'Show trend lines';

        token.element = {
          source: token.type,
          phrase: token.phrase,
          type: token.type,
        };
      }
    });
  };

  return r;
}

function processEach() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Identify "each" request';
  };

  r.process = function() {
    r.forEachUnquotedString((node) => {
      const token = node.data();
      const lc = token.phrase.toLowerCase();

      if (stemmer(lc) === 'each') {
        token.type = elementType.EACH;
        token.highlightNote = 'Granular grouping';

        token.element = {
          source: token.type,
          phrase: token.phrase,
          type: token.type,
        };
      }
    });
  };

  return r;
}

function processTop() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Identify TopK';
  };

  r.process = function() {
    const substitutions = 0;

    r.forEachLeafWithoutElement((node) => {
      const token = node.data();

      if ((token.type === elementType.TOP) || (token.type === elementType.BOTTOM)) {
        token.element = {
          source: token.type,
          phrase: token.phrase,
          type: token.type,
          k: token.k,
        };
      }
    });

    return substitutions;
  };

  return r;
}

function whereClauseFromValues(column, values, filter_exclude, numeric) {
  let needToLower = false;
  values = values.map((v) => {
    const quotedValue = quoteValue(column, v);
    const currNeedLower = !numeric && quotedValue !== 'NULL' && !$.isNumeric(quotedValue);
    needToLower = needToLower || currNeedLower;
    if (currNeedLower) {
      try {
        return quotedValue.toLowerCase();
      } catch (e) {
        return quotedValue;
      }
    }

    return quotedValue;
  });

  const columnExp = arc.misc.quoteOrCastColumnName(column, needToLower);
  return arc.misc.whereClauseFromValues(columnExp, values, filter_exclude);
}

function processExplicitFilter(knownTerms) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Process explicit filters';
  };

  /*
   * Handle strings like:
   * year:2010
   * state:CA
   * TODO: need to test/handle something like product:"skin care"
   */
  r.process = function() {
    let substitutions = 0;

    r.forEachLeafWithoutElement((node) => {
      const token = node.data();

      if (token.type !== elementType.KEYVALUE) {
        return;
      }

      const k = token.key;
      const v = token.value;

      const c = knownTerms.findFullTerm(k);
      if (!c) {
        return;
      }

      let expr;
      if (v.indexOf('*') >= 0) {
        expr = `(${arc.misc.approximateClauseFromValues(c, [v])})`;
      } else {
        expr = `(${whereClauseFromValues(c, [v])})`;
      }

      token.element = {
        source: 'fixedfilter',
        phrase: token.phrase,
        type: elementType.FILTER,
        subtype: filterType.FIXED,
        name: arc.misc.quoteColumn(c.display_name()),
        column: c,
        values: [v],
        expr,
      };

      token.element.displayExpr = token.element.expr;
      substitutions += calculateMultiwordScore(`${k} ${v}`);
    });

    return substitutions;
  };

  return r;
}

function nextElementByType(node, elementTypes) {
  let n;

  node.walkLeaves((curr) => {
    const t = curr.data();
    if (curr === node) {
      return;
    }

    if (t) {
      const e = t.element;
      if (e) {
        if (elementTypes.indexOf(e.type) >= 0) {
          n = curr;
          return true;
        }
      }
    }
  });

  return n;
}

function nextDimOrAgg(node) {
  return nextElementByType(node, [elementType.DIMENSION, elementType.AGGREGATE]);
}

function nextDim(node) {
  return nextElementByType(node, [elementType.DIMENSION]);
}

function nextFilter(node) {
  return nextElementByType(node, [elementType.FILTER]);
}

function combinePrevConnectors(node) {
  const token = node.data();

  if (!token) {
    return;
  }

  let curr = node.prev();
  while (curr && curr.data().type === elementType.CONNECTOR) {
    token.startIndex = curr.data().startIndex;
    curr.remove();
    curr = node.prev();
  }
}

function connectedStrings(node, cb) {
  let curr = node.next();
  while (curr) {
    const next = curr.next();
    const t = curr.data();

    if (t.type === elementType.CONNECTOR) {
      if (isAnd(t)) {
        node.data().connector = 'AND';
      }

      curr.remove();
    } else if ((t.type === elementType.STR) && (t.element === undefined)) {
      const ret = cb(curr);
      if (ret) {
        // early return
        return;
      }

      curr.remove();
    } else {
      return;
    }

    curr = next;
  }
}

function processDashboardFilters(knownTerms) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Process dashboard filters';
  };

  r.process = function() {
    let substitutions = 0;

    r.forEachLeafWithoutElement((node) => {
      const token = node.data();

      if (token.type !== elementType.KEYVALUE) {
        return;
      }

      const k = token.key;
      const v = token.value;

      const lw = k.toLowerCase();
      if (lw !== elementType.FILTER) {
        return;
      }

      const c = knownTerms.findFullTerm(v);
      if (!c) {
        return undefined;
      }

      token.type = elementType.CONTROL;
      token.element = {
        source: 'control',
        phrase: token.phrase,
        type: elementType.CONTROL,
        name: v,
        column: c,
      };

      // Add 1.5 pt for each additional word
      substitutions += calculateMultiwordScore(v);
    });

    return substitutions;
  };

  return r;
}

function processAggregations(dataset) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Process expressions with aggregates';
  };

  r.process = function() {
    let substitutions = 0;
    const connType = dataset.dataconnection_type();
    r.forEachLeafWithoutElement((node) => {
      const token = node.data();

      const isSupportedAgg = token.aggregation &&
                             (arc.report.addAndCheckAggregates(token.aggregation, connType) ||
                              token.aggregation === 'COUNTDISTINCT' ||
                              token.aggregation === 'APPROXIMATE_COUNT');
      const isSupportedPhrase = arc.report.addAndCheckAggregates(token.phraseLowerCase, connType);
      if (!isSupportedAgg && !isSupportedPhrase) {
        return;
      }

      if (!token.aggregation) {
        token.aggregation = token.phrase.toUpperCase();
        token.type = elementType.AGGREGATE;
      }

      let dimagg;
      dimagg = nextDimOrAgg(node);

      if (dimagg) {
        const aggToken = dimagg.data();
        const { element } = aggToken;
        const { column } = element;

        // Don't associate values that already have aggregates
        if (column.attr('exprhasagg') || column.attr('recordcount')) return;

        aggToken.type = elementType.AGGREGATE;
        element.type = elementType.AGGREGATE;

        aggToken.startIndex = token.startIndex;

        element.aggregation = token.aggregation;
        element.expr = aggExpression(column, token.aggregation);
        aggToken.startIndex = token.startIndex;

        node.remove();
        substitutions += calculateMultiwordScore(aggToken.phrase);
      }
    });

    return substitutions;
  };

  return r;
}

function isDateDimension(node) {
  const token = node.data();
  if (!token) {
    return undefined;
  }

  const elem = token.element;
  if (!elem) {
    return undefined;
  }

  const column = elem.column;
  if (!column) {
    return undefined;
  }

  return (column.coltype() === 'TIMESTAMP') ? column : undefined;
}

function getDateColumnAndMarkAsIgnored(dataset, node, markIgnore = true) {
  let column;

  node.walkPrevLeaves((curr) => {
    column = isDateDimension(curr);
    if (column) {
      if (markIgnore) {
        const elem = curr.data().element;
        elem.subtype = dimensionType.IGNORE;
      }

      return true;
    }
  });

  if (!column) {
    node.walkLeaves((curr) => {
      column = isDateDimension(curr);
      if (column) {
        if (markIgnore) {
          const elem = curr.data().element;
          elem.subtype = dimensionType.IGNORE;
        }

        return true;
      }
    });
  }

  if (!column) {
    const dt = dataset.defaultDateTime();
    if (dt) {
      column = dataset.findColumn(dt);
    }
  }

  return column;
}

function findPreviousNodeByType(node, elementTypes, nodesToIgnore) {
  let n;

  node.walkPrevLeaves((curr) => {
    const token = curr.data();
    if (token) {
      const elem = token.element;
      if (elem) {
        if (elementTypes.indexOf(elem.type) >= 0) {
          const ignore = nodesToIgnore && nodesToIgnore.some(function(_) {
            return (_ === curr);
          });

          if (!ignore) {
            n = curr;
            return true;
          }
        }
      }
    }
  });

  return n;
}

function findPreviousAgg(node) {
  return findPreviousNodeByType(node, [elementType.AGGREGATE]);
}

function findPreviousDim(node, nodesToIgnore) {
  return findPreviousNodeByType(node, [elementType.DIMENSION], nodesToIgnore);
}

function findPreviousDimOrAgg(node) {
  return findPreviousNodeByType(node, [elementType.DIMENSION, elementType.AGGREGATE]);
}

function removeAggregation(prevElement) {
  prevElement.expr = prevElement.expr.replace(prevElement.aggregation, '');
  delete prevElement.aggregation;
  prevElement.type = elementType.DIMENSION;
}

function processComparisons(knownTerms) {
  const r = arc.search.rule.base();
  let substitutions = 0;

  const addedTokens = [];
  const nodesToRemove = [];

  r.ruleName = function() {
    return 'Process simple comparisons';
  };

  // Operators can be !=, =, >, >=, <, <=, or empty
  //           empty operator implies "between" filter
  // When the column is a *measure*, the behavior is as follows:
  //  - No dimensions - don't filter by aggregate measure because this needs group by of dim
  //  - Measure preceded by "base/simple/unaggregated/raw" – filter as dimension
  //  - All other cases, filter by user-specified aggregation or default aggregation
  function handleSimpleCompare(node, operator) {
    const token = node.data();
    const prev = node.prev();

    if (!prev) {
      return;
    }

    const prevToken = prev.data();
    if (!prevToken) {
      return;
    }

    const prevElement = prevToken.element;
    if (!prevElement) {
      return;
    }

    if ((prevElement.type === elementType.DIMENSION) ||
        (prevElement.type === elementType.AGGREGATE)) {
      const c = prevElement.column;
      nodesToRemove.push(prev);

      // Handle special cases
      // If no dimensions and prevElement is aggregate, make the prevElement dimension
      let expr;
      let deleteElem = true;
      if (prevElement.type === elementType.AGGREGATE) {
        const otherDim = findPreviousDim(prev, nodesToRemove);
        if (!otherDim) {
          deleteElem = false;
          removeAggregation(prevElement);
        } else {
          deleteElem = !!findPreviousAgg(prev);
        }
      }

      if (operator) {
        let exprRHS;
        if (token.term && !$.isNumeric(token.term)) {
          const nodeColumn = knownTerms.findStemmedTerm(token.term, node);

          if (nodeColumn) {
            token.term = arc.misc.quoteColumn(nodeColumn.display_name());
            token.termType = termType.COLUMN;
            exprRHS = token.term;
          }
        }

        if (!exprRHS) {
          exprRHS = quoteValue(c, token.term);
        }

        expr = `(${prevElement.expr} ${operator} ${exprRHS})`;
        substitutions += calculateMultiwordScore(expr);
      } else {
        const colExpr = prevElement.expr;
        let greaterThanTerm;
        let lessThanTerm;

        const gtColumn = knownTerms.findStemmedTerm(token.greater_than_term);
        if (gtColumn) {
          greaterThanTerm = arc.misc.quoteColumn(gtColumn.display_name());
        } else {
          greaterThanTerm = quoteValue(c, token.greater_than_term);
        }

        const ltColumn = knownTerms.findStemmedTerm(token.less_than_term);
        if (ltColumn) {
          lessThanTerm = arc.misc.quoteColumn(ltColumn.display_name());
        } else {
          lessThanTerm = quoteValue(c, token.less_than_term);
        }

        expr = `(${colExpr} >= ${greaterThanTerm}) and (${colExpr} <= ${lessThanTerm})`;
        substitutions += calculateMultiwordScore(`${greaterThanTerm} ${colExpr} ${lessThanTerm}`);
      }

      const newElement = {
        source: elementType.COMPARE,
        phrase: token.term,
        type: elementType.FILTER,
        subtype: filterType.FIXED,
        column: c,
        values: [token.term],
        name: prevElement.name,
        expr,
      };

      token.type = elementType.FILTER;
      token.element = newElement;

      if (operator && deleteElem) {
        prev.remove();
      }

      if ((operator === '=') || (operator === '!=')) {
        const equality = operator === '=';
        const values = [newElement.phrase];

        connectedStrings(node, (curr) => {
          const { phrase } = curr.data();
          if (!$.isNumeric(phrase)) {
            // trigger early exit from connectedStrings
            return true;
          }

          values.push(phrase);
          token.endIndex = curr.data().endIndex;
        });

        if (values.length > 1) {
          newElement.values = values;
          newElement.expr = `(${whereClauseFromValues(c, values, !equality, true)})`;
        }
      }
    }
  }

  function handleStartsWith(node, atEnd) {
    const token = node.data();

    let dimNode = node.prev();
    while (dimNode) {
      const dimToken = dimNode.data();
      if (!dimToken ||
        !dimToken.element ||
        dimToken.element.type !== elementType.DIMENSION) {
        dimNode = dimNode.prev();
      } else {
        break;
      }
    }

    if (!dimNode) {
      return;
    }

    const dimToken = dimNode.data();
    const dimElement = dimToken.element;
    if (dimElement.type === elementType.DIMENSION) {
      const c = dimElement.column;
      nodesToRemove.push(dimNode);

      const newElement = {
        source: elementType.COMPARE,
        phrase: token.term,
        type: elementType.FILTER,
        subtype: filterType.FIXED,
        column: c,
        name: dimElement.name,
        values: [],
      };

      const addValue = function(v) {
        if (atEnd) {
          if (v.length > 0 && v[0] !== '*') {
            v = `*${v}`;
          }
        } else if (v.length > 0 && v[v.length - 1] !== '*') {
          v = `${v}*`;
        }

        newElement.values.push(v);
        const where = arc.misc.approximateClauseFromValues(c, newElement.values);
        if (where) {
          newElement.expr = `(${(token.negate ? 'not ' : '') + where})`;
        }
      };

      addValue(token.term);

      if (!newElement.expr) {
        return;
      }

      substitutions += calculateMultiwordScore(token.term);
      token.type = elementType.FILTER;
      token.element = newElement;
      token.startIndex = dimToken.startIndex;

      dimNode.remove();

      connectedStrings(node, (curr) => {
        const t = curr.data();
        addValue(t.phrase);
      });
    }
  }

  function handleEquality(node, fuzzy) {
    return handleEqualComparison(node, fuzzy, true);
  }

  function handleInequality(node, fuzzy) {
    return handleEqualComparison(node, fuzzy, false);
  }

  function handleEqualComparison(node, fuzzy, equality) {
    const token = node.data();
    const isBooleanWord = token.phraseLowerCase &&
                          booleanWords.has(token.phraseLowerCase);
    const isSpecialTerm = token.phraseLowerCase &&
                          nullWords.has(token.phraseLowerCase);
    const isNumeric = $.isNumeric(token.phrase) || isSpecialTerm;

    let prev;
    if (isBooleanWord) {
      prev = findPreviousDimOrAgg(node);
      const prevToken = prev.data() || {};
      const prevElement = prevToken.element || {};

      // Only accept measure column if is boolean, otherwise, use previous dimension
      const isAgg = prevElement.type === elementType.AGGREGATE;
      if (isAgg &&
          !arc.report.boolean_types[prevElement.column.columnType()]) {
        prev = findPreviousDim(node);
      } else if (isAgg) {
        removeAggregation(prevElement);
      }
    } else if (isNumeric) {
      prev = findPreviousDimOrAgg(node);
    } else {
      prev = findPreviousDim(node);
    }

    if (!prev) {
      return;
    }

    const prevToken = prev.data();
    if (!prevToken) {
      return;
    }

    const prevElement = prevToken.element;
    if (!prevElement) {
      return;
    }

    if (prevElement.type === elementType.DIMENSION || isSpecialTerm) {
      const c = prevElement.column;
      nodesToRemove.push(prev);

      const newElement = {
        source: elementType.COMPARE,
        phrase: `${prevElement.phrase} ${token.phrase}`,
        type: elementType.FILTER,
        subtype: filterType.FIXED,
        column: prevElement.column,
        name: prevElement.name,
        values: [],
      };

      const addValue = function(v) {
        if (fuzzy) {
          v = `*${v}*`;
        }

        newElement.values.push(v);

        let isFuzzy;
        newElement.values.some((val) => {
          if (val.indexOf('*') >= 0) {
            isFuzzy = true;
            return true;
          }

          return false;
        });

        let expr;
        if (isFuzzy) {
          expr = `(${arc.misc.approximateClauseFromValues(c, newElement.values, !equality)})`;
        } else {
          expr = `(${whereClauseFromValues(c, newElement.values, !equality)})`;
        }

        newElement.expr = expr;
      };

      addValue(token.term);

      // In some connection types (sqlite for example), booleans are returned as 0/1
      if (isBooleanWord && token.phraseLowerCase === 'false') {
        addValue('0');
      } else if (isBooleanWord && token.phraseLowerCase === 'true') {
        addValue('1');
      }

      substitutions += calculateMultiwordScore(token.term);

      token.type = elementType.FILTER;
      token.element = newElement;

      connectedStrings(node, (curr) => {
        const t = curr.data();
        token.endIndex = t.endIndex;
        addValue(t.phrase);
      });
    } else {
      handleSimpleCompare(node, equality ? '=' : '!=');
    }
  }

  r.process = function() {
    r.forEachLeafWithoutElement((node) => {
      const token = node.data();

      if (token.type !== elementType.COMPARE) {
        return;
      }

      if (token.subtype === compareType.CONTAINS) {
        handleEquality(node, true /* fuzzy */);
        return;
      }

      if (token.subtype === compareType.EQUALITY) {
        handleEquality(node);
        return;
      }

      if (token.subtype === compareType.INEQUALITY) {
        handleInequality(node);
        return;
      }

      if (token.subtype === compareType.STARTSWITH) {
        handleStartsWith(node);
        return;
      }

      if (token.subtype === compareType.ENDSWITH) {
        handleStartsWith(node, true);
        return;
      }

      if (token.subtype === compareType.NE) {
        handleSimpleCompare(node, '!=');
        return;
      }

      if (token.subtype === compareType.EQ) {
        handleSimpleCompare(node, '=');
        return;
      }

      if (token.subtype === compareType.GT) {
        handleSimpleCompare(node, '>');
        return;
      }

      if (token.subtype === compareType.GTE) {
        handleSimpleCompare(node, '>=');
        return;
      }

      if (token.subtype === compareType.LT) {
        handleSimpleCompare(node, '<');
        return;
      }

      if (token.subtype === compareType.LTE) {
        handleSimpleCompare(node, '<=');
        return;
      }

      if (token.subtype === compareType.BETWEEN) {
        handleSimpleCompare(node);
      }
    });

    addedTokens.forEach(token => r.tree().addChild(arc.tree.node(token)));

    nodesToRemove.forEach(function(token) {
      const nextToken = token.next();
      if (nextToken) {
        nextToken.data().startIndex = token.data().startIndex;
      }

      token.remove();
    });

    return substitutions;
  };

  return r;
}

function combineNodes(startNode, endNode) {
  if (startNode === endNode) {
    return;
  }

  let tmp = startNode.next();
  while (true) {
    if (tmp === undefined) {
      return;
    }

    // Absorb tmp into startNode
    startNode.data().endIndex = tmp.data().endIndex;

    const n = tmp.next();
    tmp.remove();
    if (tmp === endNode) {
      return;
    }

    tmp = n;
  }
}

/*
 * datetime_string:
 *     YYYY-M-D
 *     YYYY/M/D
 *     M-D-YYYY
 *     M/D/YYYY
 *     YYYY-M-DTHH:mm:ss
 *
 *     YYYY-M-D HH:mm:ss
 *
 *     MMMM D YYYY
 *     MMM D YYYY
 *     D MMMM YYYY
 *     D MMM YYYY
 *
 * <between/from> <datetime_string> <and/to> <datetime_string>
 *
 * date_period:
 *     century/decade/year/quarter/month/day/hour/minute
 *
 * <last/previous/next> <number>? <date_period>
 *     (<starting/ending> <with/on/from/at> <datetime_string>)?
 *
 * MMM YYYY   (example: dec 2010)
 * MMMM YYYY  (example: december 2010)
 *
 */
function processDateTimeFilters(_dataset) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Process date/time filters';
  };

  const dataset = _dataset;
  const conn = arc.dataconnection.types[dataset.dataconnection_type()]();

  function nextString(node) {
    let curr = node.next();

    while (curr) {
      const t = curr.data();

      if (t.type === elementType.CONNECTOR) {
        curr = curr.next();
      } else {
        if (t.element) {
          return undefined;
        }

        if (t.type === elementType.STR) {
          return curr;
        }

        curr = curr.next();
      }
    }

    return undefined;
  }

  function foundTime(startNode, endNode, phrase, m) {
    const token = startNode.data();

    token.element = {
      type: elementType.DATETIME,
      source: 'datetime',
      phrase,
      moment: m,
    };

    combineNodes(startNode, endNode);
  }

  const tokenFormats3 = [
    'MMMM D YYYY',
    'MMM D YYYY',
    'D MMMM YYYY',
    'D MMM YYYY',
  ];

  const tokenFormats2 = [
    'YYYY-M-D HH:mm:ss',
  ];

  const tokenFormats1 = [
    'YYYY-M-D',
    'YYYY/M/D',
    'M-D-YYYY',
    'M/D/YYYY',
    'YYYY-M-DTHH:mm:ss',
  ];

  const tokenFormatsMonth = [
    'MMM YYYY',
    'MMMM YYYY',
  ];

  r.process = function() {
    let substitutions = 0;

    const dc_type = dataset.dataconnection_type();

    // If there is no default datetime and there is no datetime in the list of dimensions
    // we have already parsed, then we can't process anything here.
    if (getDateColumnAndMarkAsIgnored(dataset, r.tree(), false) === undefined) {
      return;
    }

    let fmt = dataset.defaultDateTimeFormat();
    if ((fmt === undefined) || (fmt === '')) {
      fmt = 'YYYY-MM-DD';
    }

    const nodes = [];
    const relativeNodes = [];

    r.forEachLeafWithoutElement((node) => {
      const token = node.data();

      if ((token.type === elementType.DATETIME) &&
          (token.subtype === datetimeType.RELATIVE)) {
        relativeNodes.push(node);
        return;
      }

      if (token.type !== elementType.STR) {
        return;
      }

      const n1 = nextString(node);
      let n2;
      if (n1 !== undefined) {
        n2 = nextString(n1);
      }

      if (n1 && n2) {
        const s = `${token.phrase} ${n1.data().phrase} ${n2.data().phrase}`;
        const m = moment.utc(s, tokenFormats3, true);
        if (m.isValid()) {
          foundTime(node, n2, s, m);
          nodes.push(node);
          return;
        }
      }

      if (n1) {
        const s = `${token.phrase} ${n1.data().phrase}`;
        let m = moment.utc(s, tokenFormats2, true);
        if (m.isValid()) {
          foundTime(node, n1, s, m);
          nodes.push(node);
          return;
        }

        m = moment.utc(s, tokenFormatsMonth, true);
        if (m.isValid()) {
          foundTime(node, n1, s, m);
          nodes.push(node);
          return;
        }
      }

      const s = token.phrase;
      const m = moment.utc(s, tokenFormats1, true);
      if (m.isValid()) {
        foundTime(node, node, s, m);
        nodes.push(node);
      }
    });

    relativeNodes.forEach((node) => {
      const relativeToken = node.data();
      const mStart = moment.utc();
      const mEnd = moment.utc();

      if (relativeToken.period_prevnext === 'curr') {
        mStart.startOf(relativeToken.period);
        mEnd.endOf(relativeToken.period);
      } else if (relativeToken.period_prevnext === 'prev') {
        mStart.subtract(relativeToken.period_count, relativeToken.period)
              .startOf(relativeToken.period);

        mEnd.subtract(1, relativeToken.period)
            .endOf(relativeToken.period);
      } else {
        // relativeToken.period_prevnext === 'next'
        mStart.add(1, relativeToken.period)
              .startOf(relativeToken.period);

        mEnd.add(relativeToken.period_count, relativeToken.period)
            .endOf(relativeToken.period);
      }

      relativeToken.type = elementType.FILTER;
      relativeToken.subtype = filterType.FIXED;
      relativeToken.ignoreDups = true;

      const column = getDateColumnAndMarkAsIgnored(dataset, node);
      relativeToken.element = {
        type: elementType.FILTER,
        subtype: filterType.FIXED,
        name: arc.misc.quoteColumn(column.display_name()),
        column,
        duplicateExpr: periodExpr(conn, column, relativeToken.period),
        expr: arc.misc.timestampRange(dc_type, column.display_name(),
            column.attr('type'), mStart.format(fmt), mEnd.format(fmt), fmt),
      };

      combinePrevConnectors(node);
    });

    if (nodes.length > 0) {
      substitutions += nodes.length;
    }

    if (nodes.length === 2) {
      const firstNode = nodes[0];
      const prev = firstNode.prevLeaf();
      const next = firstNode.nextLeaf();
      let prevPhrase;
      let nextPhrase;

      if (prev && prev.data()) {
        prevPhrase = prev.data().phrase;
      }

      if (next && next.data()) {
        nextPhrase = prev.data().phrase;
      }

      /*
       * We have 2 dates in one of the three formats below:
       *   between <date> <something> <date>
       *   from <date> <something> <date>
       *   <date> to <date>
       *
       * process the two dates as a range
       */
      if ((nextPhrase === 'to') ||
          (prevPhrase === 'between') ||
          (prevPhrase === 'from')) {
        const m1 = nodes[0].data().element.moment;
        const m2 = nodes[1].data().element.moment;

        const t1 = m1.format(fmt);
        let t2;

        if (tokenFormatsMonth.indexOf(m2._f) >= 0) {
          // The second part of the range is a month filter of the form
          // "oct 1940". Go to the end of october in that case, not the start.
          t2 = m2.endOf('month').format(fmt);
        } else {
          t2 = m2.format(fmt);
        }

        const token = nodes[0].data();
        token.type = elementType.FILTER;
        token.subtype = filterType.FIXED;
        token.ignoreDups = true;

        const column = getDateColumnAndMarkAsIgnored(dataset, nodes[0]);
        token.element = {
          type: elementType.FILTER,
          subtype: filterType.FIXED,
          name: arc.misc.quoteColumn(column.display_name()),
          column,
          expr: arc.misc.timestampRange(
            dc_type,
            column.display_name(),
            column.attr('type'), t1, t2, fmt
          ),
        };

        nodes[0].data().endIndex = nodes[1].data().endIndex;
        nodes[1].remove();

        combinePrevConnectors(nodes[0]);
        return;
      }
    }

    /*
     * 1, 3, or more absolute dates are in play. Add in exact filters
     * for the dates.
     */
    nodes.forEach((node) => {
      const token = node.data();
      const m = token.element.moment;
      let expr;

      const column = getDateColumnAndMarkAsIgnored(dataset, node);

      if (tokenFormatsMonth.indexOf(m._f) >= 0) {
        /*
         * This is a month string i.e. december 2009. Put in a range from
         * the start to the end of december 2009.
         */
        const t1 = m.format(fmt);
        const t2 = m.endOf('month').format(fmt);
        expr = arc.misc.timestampRange(dc_type, column.display_name(),
          column.attr('type'), t1, t2, fmt);
      }

      if (!expr) {
        const t1 = m.format(fmt);
        expr = arc.misc.timestampEqual(
          dc_type,
          column.display_name(),
          column.attr('type'), t1, fmt
        );
      }

      token.type = elementType.FILTER;
      token.subtype = filterType.FIXED;
      token.ignoreDups = true;
      token.element = {
        type: elementType.FILTER,
        name: arc.misc.quoteColumn(column.display_name()),
        column,
        expr,
      };

      combinePrevConnectors(node);
    });

    return substitutions;
  };

  return r;
}

function processTerms(knownTerms, stem) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return `Process known terms${(stem ? '' : ' with stemming')}`;
  };

  r.process = function() {
    let substitutions = 0;

    r.forEachLeafWithoutElement((node) => {
      const token = node.data();
      const w = token.phrase;

      if (!w) {
        return;
      }

      let c;
      if (stem) {
        c = knownTerms.findStemmedTerm(w, node);
      } else {
        c = knownTerms.findFullTerm(w, node);
      }

      if (c === undefined) {
        return;
      }

      substitutions += 1;

      // If previous word is forceDimensionWords, make dimension
      const prevWord = node.prev();
      let forceDim = false;

      if (prevWord) {
        const prevToken = prevWord.data();
        forceDim = prevToken &&
                   forceDimensionWords.has(prevToken.phraseLowerCase);
      }

      const isDim = c.isDimension() || c.isGeo();
      if (isDim || forceDim) {
        token.element = {
          source: stem ? 'stemmedterm' : 'term',
          phrase: token.phrase,
          type: elementType.DIMENSION,
          name: arc.misc.quoteColumn(c.display_name()),
          column: c,
          expr: arc.misc.quoteColumn(c.display_name()),
        };

        if (forceDim) {
          // Combine the dimension transformation word with this token
          token.startIndex = prevWord.data().startIndex;
          prevWord.remove();
        }

        return;
      }

      token.element = {
        source: stem ? 'stemmedterm' : 'term',
        phrase: token.phrase,
        type: elementType.AGGREGATE,
        name: arc.misc.quoteColumn(c.display_name()),
        aggregation: c.defaultAggregate(),
        column: c,
        expr: aggExpression(c, c.defaultAggregate()),
      };
    });

    return substitutions;
  };

  return r;
}

const datetimePeriods = {};

createDateTimePeriods();

function createDateTimePeriods() {
  datetimePeriods.year = 'year';
  datetimePeriods[stemmer('year')] = 'year';
  datetimePeriods.yearly = 'year';
  datetimePeriods[stemmer('yearly')] = 'year';
  datetimePeriods['year-over-year'] = 'year';

  datetimePeriods.month = 'month';
  datetimePeriods[stemmer('month')] = 'month';
  datetimePeriods.monthly = 'month';
  datetimePeriods[stemmer('monthly')] = 'month';

  datetimePeriods.quarter = 'quarter';
  datetimePeriods[stemmer('quarter')] = 'quarter';
  datetimePeriods.quarterly = 'quarter';
  datetimePeriods[stemmer('quarterly')] = 'quarter';

  datetimePeriods.week = 'week';
  datetimePeriods[stemmer('week')] = 'week';
  datetimePeriods.weekly = 'weekly';
  datetimePeriods[stemmer('weekly')] = 'weekly';

  datetimePeriods.date = 'day';
  datetimePeriods[stemmer('date')] = 'day';
  datetimePeriods.day = 'day';
  datetimePeriods[stemmer('day')] = 'day';
  datetimePeriods.daily = 'day';
  datetimePeriods[stemmer('daily')] = 'day';

  datetimePeriods.hour = 'hour';
  datetimePeriods[stemmer('hour')] = 'hour';
  datetimePeriods.hourly = 'hour';
  datetimePeriods[stemmer('hourly')] = 'hour';
}

const defaultDateFormat = {
  dateformat: 'MM/DD/YYYY HH:mm:ss',
  dateoptions: 'MM/DD/YYYY HH:mm:ss',
  result: 'MM/DD/YYYY HH:mm:ss',
  type: 'date',
};

function hourExpression(conn, column) {
  if (conn.expr_fullhour) {
    // Set display format
    const displayFormat = column.attr('displayformat');
    if (!displayFormat || displayFormat.type !== 'date') {
      column.attr('displayformat', defaultDateFormat);
    }

    return conn.expr_fullhour(column.display_name());
  }

  return dateExpression(conn, column);
}

function dateExpression(conn, column) {
  let data = arc.dataconnection.dateConvert({
    columnName: column.display_name(),
    connection: conn,
    unit: 'DATE',
  });

  if (data && data.expr) {
    return data.expr;
  }

  // Set display format
  const displayFormat = column.attr('displayformat');
  if (!displayFormat || displayFormat.type !== 'date') {
    column.attr('displayformat', defaultDateFormat);
  }

  return `${conn.date}([${column.display_name()}])`;
}

function monthExpression(conn, column, includeYear) {
  let expr;

  if (includeYear) {
    expr = conn.expr_yearmonth(column.display_name());
  } else if (conn.expr_month) {
    expr = conn.expr_month(column.display_name());
  } else {
    expr = `${conn.month}([${column.display_name()}])`;
  }

  return expr;
}

function weekExpression(conn, column, includeYear) {
  let expr;

  if (conn.expr_yearweek && includeYear) {
    expr = conn.expr_yearweek(column.display_name());
  } else if (conn.expr_week) {
    expr = conn.expr_week(column.display_name());
  } else {
    expr = dateExpression(conn, column);
  }

  return expr;
}

function yearExpression(conn, column) {
  let expr;

  if (conn.expr_year) {
    expr = conn.expr_year(column.display_name());
  } else {
    expr = `${conn.year}([${column.display_name()}])`;
  }

  return expr;
}

function quarterExpression(conn, column, includeYear) {
  if (includeYear) {
    return conn.expr_yearquarter(column.display_name());
  }

  return conn.expr_quarter(column.display_name());
}

function periodExpr(conn, column, period) {
  if (period === 'year') {
    return yearExpression(conn, column);
  }

  if (period === 'month') {
    return monthExpression(conn, column);
  }

  if (period === 'quarter') {
    return quarterExpression(conn, column);
  }

  return undefined;
}

function processDateTimeGroups(_dataset) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Process date/time groups';
  };

  const dataset = _dataset;

  r.process = function() {
    let substitutions = 0;

    const dt = dataset.defaultDateTime();
    let column = dataset.findColumn(dt);
    const dc_type = dataset.dataconnection_type();
    const conn = arc.dataconnection.types[dc_type]();

    if (conn === undefined) {
      return;
    }

    let fmt = dataset.defaultDateTimeFormat();
    if ((fmt === undefined) || (fmt === '')) {
      fmt = 'YYYY-MM-DD';
    }

    function getNearbyDateCol(node) {
      const prevDimension = findPreviousDim(node);
      if (!setIfDate(prevDimension)) {
        const nextDimension = nextDim(node);
        if (setIfDate(nextDimension)) {
          combineNodes(node, nextDimension);
        }
      } else {
        const tmpData = node.data();
        combineNodes(prevDimension, node);
        tmpData.startIndex = prevDimension.data().startIndex;
        prevDimension.data(tmpData);
      }

      function setIfDate(dim) {
        const colToken = dim && dim.data();
        const col = colToken && colToken.element && colToken.element.column;
        if (col && col.coltype() === 'TIMESTAMP') {
          column = col;
          return true;
        }
      }
    }

    function yearElement(token) {
      const expr = yearExpression(conn, column);
      column.attr('displayformat', '.0f');
      return exprToElement(token, `${expr} as year`, 'year');
    }

    function quarterElement(token, includeYear) {
      const expr = quarterExpression(conn, column, includeYear);
      return exprToElement(token, `${expr} as quarter`, 'quarter');
    }

    function monthElement(token, includeYear) {
      const expr = monthExpression(conn, column, includeYear);
      return exprToElement(token, `${expr} as month`, 'month');
    }

    function weekElement(token, includeYear) {
      const expr = weekExpression(conn, column, includeYear);
      return exprToElement(token, `${expr} as week`, 'week');
    }

    function dateElement(token) {
      const expr = dateExpression(conn, column);
      return exprToElement(token, `${expr} as date`, 'date');
    }

    function hourElement(token) {
      const expr = hourExpression(conn, column);
      return exprToElement(token, `${expr} as hour`, 'hour');
    }

    function exprToElement(token, expr, name) {
      const element = {
        source: 'datetime_groups',
        phrase: token.phrase,
        type: elementType.DIMENSION,
        name,
        column,
        expr,
      };

      return element;
    }

    const includedPeriods = {};
    let startIndex;
    let endIndex;
    r.forEachUnquotedString((node) => {
      let token = node.data();
      let lc = token.phrase.toLowerCase();

      if (datetimePeriods[lc] === undefined) {
        lc = stemmer(lc);
      }

      if (datetimePeriods[lc] === undefined) {
        return;
      }

      const p = datetimePeriods[lc];

      if (p) {
        getNearbyDateCol(node);
      }

      if (!column) {
        return;
      }

      token = node.data();
      startIndex = startIndex === undefined ?
                   token.startIndex :
                   Math.min(token.startIndex, startIndex);
      endIndex = endIndex === undefined ?
                 token.endIndex :
                 Math.max(token.endIndex, endIndex);
      includedPeriods[p] = token;
    });

    let token;
    if (includedPeriods.hour) {
      token = includedPeriods.hour;
      token.element = hourElement(token);
    } else if (includedPeriods.day) {
      token = includedPeriods.day;
      token.element = dateElement(token);
    } else if (includedPeriods.week && includedPeriods.year) {
      token = includedPeriods.week;
      token.element = weekElement(token, 'includeYear');
    } else if (includedPeriods.week) {
      token = includedPeriods.week;
      token.element = weekElement(token);

      r.tree().addChild(arc.tree.node({
        phrase: '',
        type: elementType.DIMENSION,
        rule: 'datetime_groups',
        startIndex: token.startIndex,
        endIndex: token.endIndex,
        element: yearElement(token),
      }));
    } else if (includedPeriods.month && includedPeriods.year) {
      token = includedPeriods.month;
      token.element = monthElement(token, 'includeYear');
    } else if (includedPeriods.month) {
      token = includedPeriods.month;
      token.element = monthElement(token);

      r.tree().addChild(arc.tree.node({
        phrase: '',
        type: elementType.DIMENSION,
        rule: 'datetime_groups',
        startIndex: token.startIndex,
        endIndex: token.endIndex,
        element: yearElement(token),
      }));
    } else if (includedPeriods.quarter && includedPeriods.year) {
      token = includedPeriods.quarter;
      token.element = quarterElement(token, 'includeYear');
    } else if (includedPeriods.quarter) {
      token = includedPeriods.quarter;
      token.element = quarterElement(token);

      r.tree().addChild(arc.tree.node({
        phrase: '',
        type: elementType.DIMENSION,
        rule: 'datetime_groups',
        startIndex: token.startIndex,
        endIndex: token.endIndex,
        element: yearElement(token),
      }));
    } else if (includedPeriods.year) {
      token = includedPeriods.year;
      token.element = yearElement(token);
    }

    if (token) {
      substitutions += 1;
      token.type = token.element.type;
      token.startIndex = startIndex;
      token.endIndex = endIndex;
    }

    return substitutions;
  };

  return r;
}

const ignoreWords = {};

createIgnoreWords();

function createIgnoreWords() {
  'show me this that when why which where how what was a were'
      .split(/\s+/).forEach((w) => {
    ignoreWords[w] = true;
  });

  'as an the of in on by for is are to over with between has contains'
      .split(/\s+/).forEach((w) => {
    ignoreWords[w] = true;
  });
}

function processIgnoreWords(_stem) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return `Ignore simple words${_stem ? '' : ' with stemming'}`;
  };

  const stem = _stem;

  r.process = function() {
    r.forEachLeafWithoutElement((node) => {
      const token = node.data();
      let w = token.phraseLowerCase;

      if (!w) {
        return;
      }

      if (stem) {
        w = stemmer(w);
      }

      if (ignoreWords[w]) {
        token.element = {
          source: 'ignore',
          phrase: token.phrase,
          type: elementType.IGNORE,
        };
      }
    });
  };

  return r;
}

function processOrdering() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Process Ordering';
  };

  const supportedAsc = new Set(['ascend', 'asc']);
  const supportedDesc = new Set(['descend', 'desc']);
  const supportedOrder = new Set(['order', 'sort']);

  r.process = function() {
    let ascOrder;
    let orderSpecified;

    /*
     * First, look for any usage of ascending/descending/ordering
     */
    r.forEachLeafWithoutElement(function(node) {
      let token = node.data();
      let w = token.phrase.toLowerCase();

      const s = stemmer(w);
      if (supportedAsc.has(s)) {
        ascOrder = true;
      } else if (supportedDesc.has(s)) {
        ascOrder = false;
      } else if (supportedOrder.has(s)) {
        orderSpecified = true;
      }
    });

    if ((ascOrder === undefined) && (orderSpecified === undefined)) {
      return;
    }

    r.forEachLeafWithoutElement((node) => {
      const token = node.data();
      const w = token.phrase.toLowerCase();

      let s = stemmer(w);
      if (!supportedAsc.has(s) && !supportedDesc.has(s) && !supportedOrder.has(s)) {
        return;
      }

      if (orderSpecified && supportedOrder.has(s)) {
        // If only "order" or "sort" is specified, assume ascending
        if (ascOrder === undefined) {
          s = 'asc';
        } else {
          node.remove();
          return;
        }
      }

      if (supportedAsc.has(s)) {
        ascOrder = true;
      } else if (supportedDesc.has(s)) {
        ascOrder = false;
      }

      let n = nextDimOrAgg(node);
      if (!n) {
        n = findPreviousDimOrAgg(node);
      }

      if (!n) {
        return;
      }

      token.type = elementType.ORDER;

      token.element = {
        source: token.type,
        phrase: token.phrase,
        type: token.type,
      };

      const { element } = n.data();
      if (element.order) {
        return;
      }

      if (element.type === elementType.AGGREGATE) {
        element.order = {
          ascending: ascOrder === undefined ? false : ascOrder,
          limit: 5000,
        };
      } else {
        element.order = {
          ascending: ascOrder === undefined ? true : ascOrder,
          limit: 5000,
        };
      }
    });
  };

  return r;
}

function processUnknowns(_searchText) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Process unknowns';
  };

  const searchText = _searchText;

  r.process = function() {
    r.forEachLeafWithoutElement((node) => {
      const token = node.data();
      const w = token.phrase;

      if (token.type === elementType.CONNECTOR) {
        node.remove();
        return;
      }

      token.element = {
        source: 'unknown',
        phrase: token.phrase,
        type: elementType.UNKNOWN,
        name: w,
        token,
        tooltip: searchText,
      };
    });
  };

  return r;
}

function postProcessTokens() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Post-process tokens';
  };

  r.process = function() {
    r.forEachLeaf((node) => {
      const token = node.data();

      if (token && token.element) {
        token.element.token = token;
        token.type = token.element.type;
      }
    });
  };

  return r;
}

function processTopk(_config) {
  const r = arc.search.rule.base();

  const config = _config || {};
  const {
    ifNoAggregate,
  } = config;

  r.ruleName = function() {
    return 'Process TopK';
  };

  r.process = function() {
    let firstAggregate;
    let firstSortDim;
    let firstDim;
    let topkToken;
    let bottomkToken;
    let eachCount = 0;

    r.forEachLeaf((node) => {
      const token = node.data();
      const { element } = token;

      if (!element) {
        return;
      }

      if (element.type === elementType.EACH) {
        eachCount += 1;
      }

      if (!firstAggregate && (element.type === elementType.AGGREGATE)) {
        firstAggregate = element;
        return;
      }

      if (!topkToken && (element.type === elementType.TOP)) {
        topkToken = token;
        return;
      }

      if (!bottomkToken && (element.type === elementType.BOTTOM)) {
        bottomkToken = token;
        return;
      }

      if (!firstSortDim && (element.type === elementType.DIMENSION)) {
        if (element.order) {
          firstSortDim = element;
        } else {
          firstDim = element;
        }
      }
    });

    const topOrBottomToken = topkToken || bottomkToken;
    if (!topOrBottomToken) {
      return;
    }

    const topOrBottomElement = topOrBottomToken.element;

    if (firstAggregate) {
      if (ifNoAggregate) {
        return;
      }

      topOrBottomElement.aggregateExpr = firstAggregate.expr;
      firstAggregate.order = {
        ascending: (topkToken === undefined),
        limit: topOrBottomElement.k,
      };

      if (eachCount > 0) {
        firstAggregate.order_limit_granular = true;
      }
    } else {
      if (!ifNoAggregate) {
        return;
      }

      if (!firstSortDim && !firstDim) {
        return;
      }

      // handle cases where there's only dimension without sort word
      if (!firstSortDim) {
        firstSortDim = firstDim;
        firstSortDim.order = {
          ascending: (topkToken === undefined),
        };
      }

      firstSortDim.order.limit = topOrBottomElement.k;
    }
  };

  return r;
}

function addDistinctIfNoAggregate() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Add distinct dimension modified if no aggregates present';
  };

  let aggCount = 0;
  let distinctCount = 0;
  let topCount = 0;
  let firstDimension;

  r.process = function() {
    r.forEachLeaf((node) => {
      const token = node.data();
      const { element } = token;

      if (!element) {
        return;
      }

      if ((element.type === elementType.TOP) || (element.type === elementType.BOTTOM)) {
        topCount += 1;
      }

      if (element.type === elementType.DISTINCT) {
        distinctCount += 1;
      }

      if (element.type === elementType.AGGREGATE) {
        aggCount += 1;
      }

      if (element.type === elementType.DIMENSION) {
        firstDimension = token;
      }
    });

    if (aggCount !== 0) {
      // Aggregates present, distinct does not need to be applied
      return;
    }

    // If top is specified but there are no aggregates, apply distinct to the dimensions
    distinctCount += topCount;

    if (distinctCount === 0) {
      // No distinct, rule does not apply
      return;
    }

    /*
     * No aggregates present. If there is a dimension, apply distinct to it
     */
    if (firstDimension) {
      if (firstDimension.element && firstDimension.element.expr) {
        firstDimension.element.expr = `distinct ${firstDimension.element.expr}`;
      }
    }
  };

  return r;
}

function addFilterDimensionsIfTopKAndNoDim() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Add filter dimensions if topk and no other dimensions present';
  };

  let dimCount = 0;
  let topCount = 0;

  r.process = function() {
    const filterDimColumns = [];

    r.forEachLeaf((node) => {
      const token = node.data();
      const { element } = token;

      if (!element) {
        return;
      }

      if ((element.type === elementType.TOP) || (element.type === elementType.BOTTOM)) {
        topCount += 1;
      }

      if (element.type === elementType.DIMENSION) {
        dimCount += 1;
      }

      if (element.type === elementType.FILTER) {
        if (element.column.isDimension()) {
          filterDimColumns.push(element.column);
        }
      }
    });

    if (topCount === 0) {
      // No topk, rule does not apply
      return;
    }

    if (dimCount !== 0) {
      // Dimension present, we do not have to go through the filters
      return;
    }

    /*
     * No dimensions present. If there are filters, add the dimension from
     * the first filter here.
     */
    filterDimColumns.some((c) => {
      r.tree().addChild(arc.tree.node(dimensionToken(c, r.ruleName())));
      return true;
    });
  };

  return r;
}

function addFilterDimensionsIfNoDimAgg() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Add filter dimensions if none present';
  };

  let dimAggCount = 0;

  r.process = function() {
    const filterDimColumns = [];

    r.forEachLeaf((node) => {
      const token = node.data();
      const { element } = token;

      if (!element) {
        return;
      }

      if ((element.type === elementType.DIMENSION) ||
          (element.type === elementType.AGGREGATE)) {
        dimAggCount += 1;
      }

      if (element.type === elementType.FILTER) {
        if (element.column.isDimension()) {
          filterDimColumns.push(element.column);
        }
      }
    });

    if (dimAggCount !== 0) {
      // Dimension or aggregate present, we do not have to go through the filters
      return;
    }

    /*
     * No dimensions present. If there are filters, add dimensions used in the
     * filters here.
     */
    filterDimColumns.forEach((c) => {
      r.tree().addChild(arc.tree.node(dimensionToken(c, r.ruleName())));
    });
  };

  return r;
}

function dimensionToken(column, source = 'unknown', phrase = '') {
  return {
    phrase,
    type: elementType.DIMENSION,
    element: {
      source,
      phrase,
      type: elementType.DIMENSION,
      name: arc.misc.quoteColumn(column.display_name()),
      column,
      expr: arc.misc.quoteColumn(column.display_name()),
    },
  };
}

function aggregateToken(column, source = 'unknown', phrase = '') {
  return {
    phrase,
    type: elementType.AGGREGATE,
    element: {
      source,
      phrase,
      type: elementType.AGGREGATE,
      name: arc.misc.quoteColumn(column.display_name()),
      aggregation: column.defaultAggregate(),
      column,
      expr: aggExpression(column, column.defaultAggregate()),
    },
  };
}

function addTrendDimension(dataset) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Add trend dimension';
  };

  r.process = function() {
    let dimCount = 0;
    let trendCount = 0;

    r.forEachLeaf((node) => {
      const token = node.data();
      const { element } = token;

      if (!token.element) {
        return;
      }

      if (element.type === elementType.DIMENSION) {
        dimCount += 1;
      }

      if (element.type === elementType.TREND) {
        trendCount += 1;
      }
    });

    if (trendCount === 0) {
      return;
    }

    if (dimCount > 0) {
      return;
    }

    const dt = dataset.defaultDateTime();
    const column = dataset.findColumn(dt);
    const dc_type = dataset.dataconnection_type();
    const conn = arc.dataconnection.types[dc_type]();

    if (column) {
      const timeToken = dimensionToken(column, r.ruleName());
      switch (dataset.defaultTimeInterval()) {
        case 'Hour':
          timeToken.element.expr = `${hourExpression(conn, column)} as date`;
          break;
        case 'Date':
          timeToken.element.expr = `${dateExpression(conn, column)} as date`;
          break;
        case 'Week':
          timeToken.element.expr = `${weekExpression(conn, column, 'includeYear')} as date`;
          break;
        case 'Month':
          timeToken.element.expr = `${monthExpression(conn, column, 'includeYear')} as month`;
          break;
        case 'Quarter':
          timeToken.element.expr = `${quarterExpression(conn, column, 'includeYear')} as quarter`;
          break;
        case 'Year':
          timeToken.element.expr = `${yearExpression(conn, column)} as year`;
          break;
        default:
          break;
      }

      r.tree().addChild(arc.tree.node(timeToken));
    }
  };

  return r;
}

function addFirstAggregate(_dataset) {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Track first aggregate';
  };

  const dataset = _dataset;

  r.process = function() {
    let firstAggregate;
    let distinctCount = 0;

    r.forEachLeaf((node) => {
      const token = node.data();
      const { element } = token;

      if (!element) {
        return;
      }

      if (element.type === elementType.DISTINCT) {
        distinctCount += 1;
      }

      if (!firstAggregate && (element.type === elementType.AGGREGATE)) {
        firstAggregate = element;
      }
    });

    if ((distinctCount === 0) && (firstAggregate === undefined)) {
      const token = defaultAggregateToken();
      if (token) {
        r.tree().addChild(arc.tree.node(token));
      }
    }
  };

  function defaultAggregateToken() {
    let c;

    const defaultAgg = dataset.defaultAgg();
    if (defaultAgg) {
      c = dataset.findColumn(defaultAgg);
    }

    if (c === undefined) {
      return undefined;
    }

    return aggregateToken(c, r.ruleName());
  }

  return r;
}

function processMultipleFiltersAsDimensions() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Add dimensions for multiple filters';
  };

  r.process = function() {
    const tracker = {};
    const duplicates = {};
    const addedTokens = [];

    r.forEachLeaf((node) => {
      const token = node.data();
      const { element } = token;

      if (!element) {
        return;
      }

      if (element.type !== elementType.FILTER) {
        return;
      }

      if (token.ignoreDups) {
        return;
      }

      if (element.column && element.column.isAggregate()) {
        return;
      }

      if (duplicates[element.name]) {
        // Already handled
        return;
      }

      if (tracker[element.name]) {
        // Duplicate found
        duplicates[element.name] = element;
        addedTokens.push(createDimension(element.column,
            element.name,
            element.duplicateExpr ? element.duplicateExpr : element.name));
        return;
      }

      // Track for later duplicate finding
      tracker[element.name] = element;
    });

    // TODO: removing addition of dimensions for now
    // addedTokens.forEach(token => r.tree().addChild(arc.tree.node(token)));
  };

  return r;
}

function createDimension(column, name, expr) {
  const element = {
    source: 'duplicates',
    phrase: '',
    type: elementType.DIMENSION,
    name,
    column,
    expr: expr || arc.misc.quoteColumn(column.display_name()),
  };

  const token = {
    phrase: '',
    type: elementType.DIMENSION,
    element,
  };

  return token;
}

function handleComparisons() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Handle comparisons';
  };

  r.process = function() {
    const comparisonNodes = [];

    r.forEachUnquotedString((node) => {
      const token = node.data();
      const lc = token.phrase.toLowerCase();

      if ((lc === 'compare') ||
          (lc === 'vs') ||
          (lc === 'versus') ||
          (lc === 'compared')) {
        comparisonNodes.push(node);
        node.remove();
      }
    });
  };

  return r;
}

const negativeWords = {};

(function() {
  negativeWords[stemmer('not')] = true;
  negativeWords[stemmer('negate')] = true;
  negativeWords[stemmer('except')] = true;
  negativeWords[stemmer('exclude')] = true;
  negativeWords[stemmer('but')] = true;
}());

function isAnd(token) {
  return ((token.type === elementType.CONNECTOR) &&
          (token.phraseLowerCase === 'and'));
}

function attachAndComparisons() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Handle AND comparisons';
  };

  r.process = function() {
    r.forEachLeaf((node) => {
      const token = node.data();

      if (!isAnd(token)) {
        return;
      }

      const n = node.next();
      if (!n) {
        return;
      }

      const ntoken = n.data();
      if (!ntoken) {
        return;
      }

      ntoken.connector = 'AND';
    });
  };

  return r;
}

function handleNegationsPre() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Handle filter negations';
  };

  r.process = function() {
    r.forEachLeaf((node) => {
      const token = node.data();

      if (token.type !== elementType.STR) {
        return;
      }

      if (token.element) {
        return;
      }

      const sw = stemmer(token.phraseLowerCase);
      if (!negativeWords[sw]) {
        return;
      }

      token.type = elementType.NEGATION;

      token.element = {
        source: 'negation',
        phrase: token.phrase,
        type: elementType.NEGATION,
        name: token.phrase,
        token,
      };
    });
  };

  return r;
}

function handleNegationsPost() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Handle filter negations';
  };

  r.process = function() {
    r.forEachLeaf((node) => {
      const token = node.data();

      if (token.type !== elementType.NEGATION) {
        return;
      }

      const filterNode = nextFilter(node);
      if (!filterNode) {
        return;
      }

      const filterToken = filterNode.data();
      const filterElement = filterToken.element;

      if (filterToken.negate) {
        return;
      }

      filterToken.negate = true;
      filterElement.expr = `(not ${filterElement.expr})`;
      filterElement.displayExpr = filterElement.expr;
    });
  };

  return r;
}

function showmeFinalize(config) {
  let {
    highlights,
    showmeVisualConfig,
    elements,
  } = config;

  let r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Handle default sorting';
  };

  r.process = function() {
    r.tree().walk(function(node) {
      const token = node.data();

      if (node.isLeaf()) {
        const { element } = token;
        if (!element) {
          return;
        }

        if (element.type === elementType.TREND) {
          showmeVisualConfig.settings.trendline_linear = 'true';
        }

        if (element.source === 'datetime_groups') {
          showmeVisualConfig.settings.trendline_linear = 'true';
        }

        if (element.type === elementType.VISUALTYPE) {
          showmeVisualConfig.vizType = element.vizType;
          showmeVisualConfig.vizTypeElement = element;
        }

        if (element.type === elementType.UNKNOWN) {
          config.unknownCount += 1;
        }
      } else {
        if (!token) {
          return;
        }

        if (token.highlightNote) {
          highlights.push(token);
        }
      }
    });

    // Remove t.element so there's no circular reference
    r.data().map(function(t) {
              const el = t.element;
              delete t.element;
              return el;
            })
            .forEach(function(e) {
              if (e && (e.type !== elementType.IGNORE) && (e.subtype !== dimensionType.IGNORE)) {
                elements.push(e);
              }
            });

    const hasAggDim = elements.some(function(e) {
      return (e.type === elementType.DIMENSION ||
              e.type === elementType.AGGREGATE);
    });

    // If there's no aggregates or dimensions, give a score of 0
    if (!hasAggDim) {
      config.score = 0;
    }
  };

  return r;
}

/*
 * If nothing in the dimensions/aggs are sorted, then sort descending by the first
 * aggregate.
 */
function processDefaultSorting() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Handle default sorting';
  };

  r.process = function() {
    let orderCnt = 0;
    let sortAgg = true;
    let sortDim = false;
    let firstDimElement;
    const dateElems = [];
    let firstAggElement;

    r.forEachLeaf((node) => {
      const token = node.data();
      const { element } = token;

      if (!element) {
        return;
      }

      if (element.type === elementType.TREND) {
        sortDim = true;
        sortAgg = false;
      }

      if (element.source === 'datetime_groups') {
        dateElems.push(element);
      }

      if (element.type === elementType.AGGREGATE) {
        firstAggElement = element;
      }

      if (element.type === elementType.DIMENSION) {
        firstDimElement = element;
      }

      if (element.order) {
        orderCnt += 1;
      }
    });

    if (orderCnt === 0) {
      if (dateElems.length) {
        dateElems.forEach((dateEl) => {
          dateEl.order = {
            ascending: true,
            limit: 5000,
          };
        });
      } else if (sortAgg && firstAggElement) {
        firstAggElement.order = {
          ascending: false,
          limit: 5000,
        };
      } else if (sortDim && firstDimElement) {
        firstDimElement.order = {
          ascending: true,
          limit: 5000,
        };
      }
    }
  };

  return r;
}

function handleDuplicates() {
  const r = arc.search.rule.base();

  r.ruleName = function() {
    return 'Remove any duplicates';
  };

  r.process = function() {
    const filters = {};
    const orderedFilters = [];
    const elementsByName = {};

    // Find duplicate filter columns (maybe with different values)
    // and duplicate aggregates and dimensions.
    r.forEachLeaf((node) => {
      const token = node.data();
      const { element } = token;

      if (!element) {
        return;
      }

      if (element.type === elementType.FILTER) {
        // We want to handle multiple filters with the same column and
        // different values by placing the values in an OR
        if (filters[element.name]) {
          node.remove();
          filters[element.name].push(element);
        } else {
          filters[element.name] = [element];
          orderedFilters.push(element.name);
        }

        return;
      }

      // Remove duplicate if not order element or
      // if element.name && element.type already exist
      if (element.type !== elementType.ORDER &&
          elementsByName[element.name] &&
          elementsByName[element.name].type === element.type) {
        // preserve the order clause
        if (element.order && !elementsByName[element.name].order) {
          elementsByName[element.name].order = element.order;
        }

        // ignore duplicates
        node.remove();
        return;
      }

      elementsByName[element.name] = element;
    });

    // Combine duplicate filter values with an OR clause
    orderedFilters.forEach((filterName) => {
      const f = filters[filterName];

      f.forEach((e, i) => {
        e.displayExpr = e.expr;
        if (i === 0) {
          return;
        }

        if (e.values) {
          f[0].values = f[0].values || [];
          e.values.forEach((v) => {
            f[0].values.push(v);
          });
        }

        let connector = 'OR';
        if (f[0].token && f[0].token.connector) {
          connector = f[0].token.connector;
        } else if (e.token && e.token.connector) {
          connector = e.token.connector;
        }

        f[0].expr = `(${f[0].expr} ${connector} ${e.expr})`;
        if (f[0].additionalElements) {
          f[0].additionalElements.push(e);
        } else {
          f[0].additionalElements = [e];
        }

        // Combine if possible
        const firstEndIndex = f[0].token.endIndex;
        const currStartIndex = e.token.startIndex;
        const currEndIndex = e.token.endIndex;
        if (firstEndIndex + 1 === currStartIndex) {
          f[0].token.endIndex = Math.max(firstEndIndex, currEndIndex);
        }
      });
    });
  };

  return r;
}

function query() {
  const q = function() { return 'query'; };

  let tokens;
  let tree;
  let score = 0;

  q.setTokens = function(_) {
    tokens = _;
    return q;
  };

  q.tree = function() {
    return tree;
  };

  q.initialize = function() {
    tree = arc.tree.node();

    tokens.forEach((t) => {
      tree.addChild(arc.tree.node().data(t));
    });

    return q;
  };

  q.applyRule = function(rule) {
    const _score = rule.tree(tree)
                       .process();

    if (_score) {
      score += _score;
    }
  };

  q.compilePhase2 = function(config) {
    const {
      finalCb,
      dataset,
    } = config;

    // Add in all filter columns with multiple filter values as dimensions
    q.applyRule(processMultipleFiltersAsDimensions(dataset));

    // Add in a default aggregate if none is present
    q.applyRule(addFirstAggregate(dataset));

    // Add in filter dimensions if no dimensions or aggregates are present
    q.applyRule(addFilterDimensionsIfNoDimAgg());

    // Add in a dimension if trending is requested and no dimension is present
    q.applyRule(addTrendDimension(dataset));

    // Apply top-k to aggregate
    q.applyRule(processTopk());

    // Add in filter dimensions if n topk is present and no dimensions are present
    q.applyRule(addFilterDimensionsIfTopKAndNoDim());

    // Remove any duplicates
    q.applyRule(handleDuplicates());

    // Handle any negations of filters
    q.applyRule(handleNegationsPost());

    // Handle default sorting
    q.applyRule(processDefaultSorting());

    // Apply top-k in cases with no aggregate
    q.applyRule(processTopk({
      ifNoAggregate: true,
    }));

    // Add in distinct if no aggregates
    q.applyRule(addDistinctIfNoAggregate());

    if (finalCb) {
      const finalConfig = {
        highlights: [],
        showmeVisualConfig: {
          settings: {},
        },
        elements: [],
        unknownCount: 0,
        score,
      };

      q.applyRule(showmeFinalize(finalConfig));

      finalCb(finalConfig);
    }
  };

  q.compilePhase1 = function(config) {
    const {
      dataset,
      expansionTermsTracker,
      knownTerms,
      searchText,
    } = config;

    if (verbose) {
      q.applyRule(printTree());
    }

    // Look for any dataset-level word expansions
    q.applyRule(wordExpansionsRule(expansionTermsTracker));

    // Process explicit filters i.e. year:2010 etc
    q.applyRule(processExplicitFilter(knownTerms));

    // Process create-control filters i.e. filter:year etc
    q.applyRule(processDashboardFilters(knownTerms));

    // Handle any dim/agg specs using stemmed words
    q.applyRule(processTerms(knownTerms, true /* stem */));

    // Handle any absolute date/time filters
    q.applyRule(processDateTimeFilters(dataset));

    // Look for and apply any usage of trending in the query
    q.applyRule(processTrending());

    // Look for and apply the 'each' keyword to do granular top/bottom k
    q.applyRule(processEach());

    // Process any top/bottom phrases
    q.applyRule(processTop());

    // Look for and apply any visual type specification
    q.applyRule(processVisualType());

    // Pick up specific aggregations
    q.applyRule(processAggregations(dataset));

    // Look for and apply any usage of distinct/unique in the query
    // Note that this must be after the aggregation step i.e. after count distinct has been handled
    q.applyRule(processDistinct());

    if (verbose) {
      q.applyRule(printTree());
    }

    // Pick up comparison filters
    q.applyRule(processComparisons(knownTerms));

    // Now, process any ignore words
    q.applyRule(processIgnoreWords());

    // Stemmed ignore words
    q.applyRule(processIgnoreWords(true /* stem */));

    // Look for groupings by time
    /* Supported times:
    year/yearly/year-over-year
    month/monthly (automatically adds year as color)
    month and year
    quarter/quarterly (automatically adds year as color)
    quarter and year
    week/weekly (automatically adds year as color)
    week and year
    day/date/daily
    hour/hourly
    */
    q.applyRule(processDateTimeGroups(dataset));

    // Handle any comparisons e.g. population this year vs last year
    q.applyRule(handleComparisons());

    // Handle any negations of filters
    q.applyRule(handleNegationsPre());

    // AND joins between filter conditions get set here
    q.applyRule(attachAndComparisons());

    // handle '<dim|agg> ordered' and 'order by <dim|agg>' phrases
    q.applyRule(processOrdering());

    // And finally mark any remaining tokens as unknown
    q.applyRule(processUnknowns(searchText));

    q.applyRule(postProcessTokens());
  };

  q.checkFilters = function(config) {
    const {
      dataset,
      searchableColumns,
      timeoutSeconds,
      timeoutCb,
      checkFiltersDoneCb,
    } = config;

    const filters = [];
    const unknowns = [];

    tree.walkLeaves((node) => {
      const token = node.data();
      const { element } = token;

      if (!element) {
        return;
      }

      switch (element.type) {
        case elementType.FILTER:
          if (element.subtype !== filterType.FIXED) {
            const { column } = token.element;
            const searchable = column.attr('searchable');
            if (!searchable) {
              break;
            }

            filters.push(token);
          }

          break;

        case elementType.UNKNOWN:
          if (token.phrase) {
            if (searchableColumns.length > 0) {
              unknowns.push(token);
            }
          }

          break;

        default:
          break;
      }
    });

    let queryCnt = (unknowns.length * searchableColumns.length) + filters.length;
    if (queryCnt === 0) {
      checkFiltersDoneCb();
      return;
    }

    unknowns.forEach((token) => {
      const { element } = token;
      const phrase = element.phrase || token.phrase;

      searchableColumns.forEach((c) => {
        const { column } = c;
        const v = quoteValue(column, phrase);

        const where = arc.misc.approximateClauseFromValues(column,
                [phrase]);
        if (where === undefined) {
          fail();
          return;
        }

        token.rb = arc.data_requests.distinctColumnValues(
            null,
            dataset,
            arc.misc.quoteColumn(column.display_name()),
            100,
            column.attr('type')
        );
        token.rb.filters(where);

        function success(data) {
          const rowCount = data.rows.length;
          if ((rowCount > 0) && (token.type === elementType.UNKNOWN)) {
            score += 1;

            // Remove any pre-set connector values so that a search like
            // show me the populations of CA and WA
            // leads to a OR filter between the two values
            token.connector = undefined;

            token.type = elementType.FILTER;
            element.type = elementType.FILTER;
            element.name = arc.misc.quoteColumn(column.display_name());
            element.column = column;
            element.values = [phrase];
            element.expr = where;
            element.token = token;
          }

          queryCnt -= 1;
          if (queryCnt === 0) {
            checkFiltersDoneCb();
          }
        }

        function fail() {
          queryCnt -= 1;
          if (queryCnt === 0) {
            checkFiltersDoneCb();
          }
        }

        function errorReporting() {
          // Ignore errors
          console.log('Error', column.display_name(), phrase);
        }

        function _timeoutCb() {
          timeoutCb({
            dataset,
            column,
            phrase,
          });
        }

        arc.data.runQuery(token.rb, success, fail, undefined, errorReporting,
            timeoutSeconds, _timeoutCb);
      });
    });

    filters.forEach((token) => {
      const { element } = token;
      const { column } = element;
      const where = arc.misc.approximateClauseFromValues(column,
          [element.phrase || token.phrase]);

      token.rb = arc.data_requests.distinctColumnValues(
        null,
        dataset,
        arc.misc.quoteColumn(column.display_name()),
        100,
        column.attr('type')
      );
      token.rb.filters(where);

      function success(data) {
        const rowCount = data.rows.length;
        if (rowCount === 0) {
          element.type = elementType.UNKNOWN;
        } else {
          element.expr = where;
        }

        queryCnt -= 1;
        if (queryCnt === 0) {
          checkFiltersDoneCb();
        }
      }

      function fail() {
        element.type = elementType.UNKNOWN;

        queryCnt -= 1;
        if (queryCnt === 0) {
          checkFiltersDoneCb();
        }
      }

      arc.data.runQuery(token.rb, success, fail);
    });
  };

  q.data = function() {
    return arc.search.rule.base()
              .tree(tree)
              .data();
  };

  return q;
}

function knownTermsTracker() {
  const k = function() { return 'knownTermsTracker'; };

  const searchTokenizer = arc.grammar._searchtokens;
  const fullMultipleWords = {};
  let stemmedMultipleWords = {};

  k.finalize = function() {
    stemmedMultipleWords = {};
    Object.keys(fullMultipleWords).forEach((keyWord) => {
      const sk = stemmer(keyWord);
      if (stemmedMultipleWords[sk] === undefined) {
        stemmedMultipleWords[sk] = {};
        stemmedMultipleWords[sk] = {
          data: [],
        };
      }

      fullMultipleWords[keyWord].data.forEach((f) => {
        stemmedMultipleWords[sk].data.push({
          value: f.value,
          allWords: f.allWords.map(w => stemmer(w)),
        });
      });
    });

    Object.keys(stemmedMultipleWords).forEach((keyWord) => {
      const v = stemmedMultipleWords[keyWord].data;
      v.sort((a, b) => (b.allWords.length - a.allWords.length));
    });
  };

  k.addTerm = function(phrase, value) {
    phrase = phrase.toLowerCase();

    const { tokens } = searchTokenizer.parse(phrase, {
      startRule: 'strings',
    });

    if (tokens.length === 0) {
      return;
    }

    const allWords = tokens.map(t => t.phrase);
    const firstPhrase = tokens[0].phrase;

    if (!fullMultipleWords[firstPhrase]) {
      fullMultipleWords[firstPhrase] = {
        data: [],
      };
    }

    fullMultipleWords[firstPhrase].data.push({
      value,
      allWords,
    });
  };

  k.findFullTerm = function(phrase, node) {
    return findTerm(phrase, node, false);
  };

  k.findStemmedTerm = function(phrase, node) {
    return findTerm(phrase, node, true);
  };

  function findTerm(phrase, node, stem) {
    phrase = phrase.toLowerCase();

    const { tokens } = searchTokenizer.parse(phrase, {
                        startRule: 'strings',
                      });
    if (tokens.length === 0) {
      return undefined;
    }

    function tokensToPhrases(tokenList) {
      const phrases = tokenList.map(t => t.phrase.toLowerCase())
                               .filter(p => p.length > 0);
      if (stem) {
        return phrases.map(w => stemmer(w));
      }

      return phrases;
    }

    const phrases = tokensToPhrases(tokens);

    const terms = stem ? stemmedMultipleWords : fullMultipleWords;

    const currPhrase = phrases[0];

    const t = terms[currPhrase];
    if (t === undefined) {
      // The first word does not match - no match
      return undefined;
    }

    let possibleMatches = terms[currPhrase].data;

    /*
     * We want to find the longest possible phrase that matches.
     * We may find shorter phrases that match along the way - keep track
     *   of the longest one we have found currently in currMatch
     */
    let currMatch;
    let commitedNode;

    // Note that 'node' may be undefined if we are simply matching a term,
    // but are not dealing with subsequent terms
    let currNode = node;

    while (possibleMatches.length > 0) {
      const newPossibleMatches = [];

      for (let idx = 0; idx < possibleMatches.length; ++idx) {
        const m = possibleMatches[idx];

        const {
          value,
          allWords,
        } = m;

        if (arc.misc.isIdentical(phrases, allWords)) {
          currMatch = value;
          commitedNode = currNode;
        } else if (phrases.length < allWords.length) {
          if (arc.misc.isIdentical(
              phrases,
              allWords.slice(0, phrases.length)
            )) {
            newPossibleMatches.push(m);
          }
        }
      }

      if (!node) {
        break;
      }

      possibleMatches = newPossibleMatches;
      if (possibleMatches.length > 0) {
        if (currNode) {
          currNode = currNode.next();
        }

        if (!currNode) {
          break;
        }

        const token = currNode.data();
        if ((token.type !== elementType.STR) && (token.type !== elementType.CONNECTOR)) {
          break;
        }

        const addedPhrases = tokensToPhrases([token]);
        phrases.push(...addedPhrases);
      }
    }

    if (currMatch !== undefined) {
      if (commitedNode !== node) {
        combineNodes(node, commitedNode);
      }
    }

    return currMatch;
  }

  return k;
}

arc.search.searchCompiler = function() {
  const search = function() { return 'arc.search.searchCompiler'; };

  let dataset;

  let storedQuery;
  const knownTerms = knownTermsTracker();
  let searchableColumns = [];

  search.searchableColumns = function() {
    return searchableColumns;
  };

  search.dataset = function(_) {
    if (!arguments.length) return dataset;
    dataset = _;
    return search;
  };

  search.searchColumnsText = function() {
    const searchText = searchableColumns.map(s => (
      arc.misc.quoteColumn(s.column.display_name())
    )).join(', ');

    if (searchableColumns.length === 0) {
      return 'No searchable columns in dataset';
    }

    return `Searched through ${searchText}`;
  };

  search.initialize = function() {
    searchableColumns = [];

    dataset.columnIterator((column) => {
      const searchable = column.attr('searchable');
      if (searchable) {
        searchableColumns.push({
          column,
          searchable,
        });
      }

      const terms = column.terms();

      if (!terms) {
        return;
      }

      let q;

      try {
        q = phraseToTerms(terms);
      } catch (e) {
        return;
      }

      q.tree().walkLeaves((node) => {
        const token = node.data();
        if (token === undefined) {
          return;
        }

        const t = token.phrase;
        knownTerms.addTerm(t, column);
      });

      knownTerms.finalize();
    });

    searchableColumns.sort((a, b) => a.searchable - b.searchable);

    return search;
  };

  function expansionToTracker(expansions) {
    const expansionTermsTracker = knownTermsTracker();

    expansions.forEach((expansion) => {
      expansionTermsTracker.addTerm(expansion.key, expansion.value);
    });

    expansionTermsTracker.finalize();
    return expansionTermsTracker;
  }

  search.processSearchString = function(s, config) {
    const {
      finalCb,
      validateFilters,
      timeoutSeconds,
      timeoutCb,
    } = config;

    const expansions = dataset.searchSettings().expansions || [];
    const expansionTermsTracker = expansionToTracker(expansions);

    storedQuery = phraseToQuery(s);
    storedQuery.compilePhase1({
      dataset,
      expansionTermsTracker,
      knownTerms,
      searchText: search.searchColumnsText(),
    });

    function finalProcessing() {
      storedQuery.compilePhase2({
        finalCb,
        dataset,
      });

      // query.applyRule(printTree());
    }

    if (validateFilters) {
      storedQuery.checkFilters({
        dataset,
        searchableColumns,
        timeoutSeconds,
        timeoutCb,
        checkFiltersDoneCb: finalProcessing,
      });
    } else {
      finalProcessing();
    }

    return search;
  };

  return search;
};

arc.search.score = function(config) {
  let {
    text,
    datasets,
    validateFilters,
    timeoutSeconds,
    timeoutCb,
    eachCb,
    cb,
  } = config;

  text = text.trim();
  if (text.length === 0) {
    return;
  }

  const scoredConfigs = [];
  const expectedCbs = datasets.length;
  let cbs = 0;

  function afterEachCb() {
    if (eachCb) {
      eachCb(scoredConfigs);
    }

    cbs += 1;
    if (cbs !== expectedCbs) {
      return;
    }

    scoredConfigs.sort((a, b) => {
      if (b.score === a.score) {
        // Smaller unknownCount is better
        return a.unknownCount - b.unknownCount;
      }

      // Bigger score is better
      return b.score - a.score;
    });

    cb(scoredConfigs);
  }

  // Try out the search string on each dataset we have.
  // Score all of the results, and make a callback with the
  //   results in descending order of score.
  datasets.forEach((dataset, idx) => {
    const sh = arc.search.searchCompiler();
    sh.dataset(dataset)
      .initialize();

    try {
      sh.processSearchString(text, {
        validateFilters,
        timeoutSeconds,
        timeoutCb,
        finalCb(_config) {
          scoredConfigs[idx] = _config;
          scoredConfigs[idx].dataset = dataset;
          scoredConfigs[idx].validateFilters = validateFilters;
          afterEachCb();
        },
      });
    } catch (e) {
      afterEachCb();
      console.log(e);
    }
  });
};
}());
