window.arc = window.arc || {};

(function() {
'use strict';

arc.expansionTracker = function() {
  const expansionTracker = function() {
    return 'expansionTracker';
  };

  let lib;
  let dataResult;
  let dataRequest;

  let originalDimensions;
  let originalFilters;
  let originalRows = [];

  let dimensionShelfElements = [];
  let colTypes = [];
  let indexOfFirstRowDimension;
  let indexOfFirstExpandableDimension;

  let expandedRowInfoByRowNumber = {};

  let rowNumberByExpandableDimensionValues = {};

  // Contains expanded dimension values and dimension values where expansion can't be enabled
  //  crosstab has a Column shelf which contains dimensions but expansion can't be enabled
  let notCollapsedDimensionValuesByRowNumber = {};

  let numberOfRows = 0;

  const getExprForCollapsedDimensionByShelfElement = (shelfElement) => {
    const expression = shelfElement.getExpressionComponents();

    const colType = shelfElement.dataset_coltype();
    const placeholder = arc.misc.isString(colType) ? '"x"' : '1';

    return `${placeholder} as ${expression.alias}`;
  };

  const deepCopyRows = (rows) => rows.map((row) => row.map((col) => col));

  expansionTracker.viz_lib = function(_) {
    if (!arguments.length) return lib;
    lib = _;
    return expansionTracker;
  };

  expansionTracker.firstExpansionDimensionIndex = function() {
    return (indexOfFirstExpandableDimension);
  };

  expansionTracker.enabled = function() {
    return (indexOfFirstExpandableDimension !== undefined);
  };

  // This is called before the visual requests the data
  // Here we store the initial state, filters, dimensions and shelf informations.
  // Then we update the dimensions of the data request object
  // because of these changes, the data will be requested as if everything were collapsed.
  expansionTracker.onGetDataRequest = function(_) {
    if (!lib || !_) {
      return;
    }

    expandedRowInfoByRowNumber = {};
    notCollapsedDimensionValuesByRowNumber = {};
    rowNumberByExpandableDimensionValues = {};

    dataRequest = _;
    const requestBuilder = dataRequest.rb();

    originalFilters = requestBuilder.filters().slice(0);

    const dimensions = dataRequest.rb().dimensions();
    originalDimensions = dimensions.map((dimension) => ({ ...dimension }));

    // lib -> `lib` is `arc_reports_base`
    const shelves = lib.shelf_data();

    indexOfFirstRowDimension = lib.viz_type() === 'crosstab' ? lib.getShelfElements('x_dimensions_shelf').length : 0;

    // we can expand only dimensions
    // not measures and tooltips (those are shelf types too)
    const dimensionShelves = shelves
      .filter(
        (shelf) => ['dimensions_shelf', 'x_dimensions_shelf', 'y_dimensions_shelf'].includes(shelf.name)
      );

    dimensionShelfElements = dimensionShelves
      .map((shelf) => lib.getShelfElements(shelf.name))
      .flat();

    indexOfFirstExpandableDimension = dimensionShelfElements
      .findIndex((shelfElement) => shelfElement.enableExpansion());

    if (indexOfFirstExpandableDimension === -1) {
      indexOfFirstExpandableDimension = undefined;
    } else {
      lib.fullyExpandedDsreq(dataRequest.dsreq());

      const newDimensions = dimensions.map((dimension, index) => {
        const isExpandable = index >= indexOfFirstExpandableDimension;

        if (!isExpandable) {
          return { ...dimension };
        }

        const shelfElement = dimensionShelfElements[index];
        const newExpr = getExprForCollapsedDimensionByShelfElement(shelfElement);

        return { ...dimension, expr: newExpr };
      });

      requestBuilder.dimensions(newDimensions);

      const dsreq = requestBuilder.dsreq();
      dataRequest.dsreq(dsreq);
      lib.dsreq(dsreq);

      dataRequest.finalize();
    }

    return expansionTracker;
  };

  const createOrGetRowNumber = (row, indexOfFirstCollapsedDimension) => {
    const expandableDimensionValues = row.slice(
      indexOfFirstRowDimension,
      dimensionShelfElements.length
    );

    let rowNumber = rowNumberByExpandableDimensionValues[expandableDimensionValues];

    if (rowNumber === undefined) {
      numberOfRows += 1;
      rowNumber = numberOfRows;

      rowNumberByExpandableDimensionValues[expandableDimensionValues] = rowNumber;

      const notCollapsedDimensionValues = row.slice(0, indexOfFirstCollapsedDimension);
      notCollapsedDimensionValuesByRowNumber[rowNumber] = notCollapsedDimensionValues;
    }

    return rowNumber;
  };

  const insertExpandAndCollapseIcons = (row, rowNumber, indexOfFirstCollapsedDimension) => {
    const isAlreadyExpanded = rowNumber in expandedRowInfoByRowNumber;

    for (
      let columnIndex = indexOfFirstCollapsedDimension;
      columnIndex <= dimensionShelfElements.length - 1;
      ++columnIndex
    ) {
      let subtotalTitle = '';

      if (columnIndex === indexOfFirstCollapsedDimension) {
        subtotalTitle = lib.user_settings().find_by_id('crosstab_total_str')?.value() || '';
      }

      const className = `crosstab-${isAlreadyExpanded ? 'collapse' : 'expand'}`;
      const icon = `<i class="fa fa-${isAlreadyExpanded ? 'minus' : 'plus'}-square-o"></i>`;

      row[columnIndex] = `<span data-row-number="${rowNumber}" data-column-index="${columnIndex}" class="${className}">${subtotalTitle} ${icon}</span>`;
    }
  };

  // This is called after the request,
  // which was prepared in the onGetDataRequest funciton returned the data.
  // We store some initial state like the column types, and the original rows.
  // Then register the rows which prepares the expand and collapse function to
  // to identify which row was clicked.
  // Lastly replace the missing ( marked as x or 1 ) values with expand and collapse icons.
  expansionTracker.onSetDataResult = function(_) {
    if (!lib) {
      return;
    }

    if (indexOfFirstExpandableDimension === undefined) {
      return expansionTracker;
    }

    dataResult = _;

    colTypes = dataResult.columns().map((column) => column.coltype());

    const rows = dataResult.rows();

    originalRows = deepCopyRows(rows);

    rows.forEach((row) => {
      const rowNumber = createOrGetRowNumber(row, indexOfFirstExpandableDimension);

      insertExpandAndCollapseIcons(row, rowNumber, indexOfFirstExpandableDimension);
    });

    return expansionTracker;
  };

  const redrawVisualAndScrollTo = () => {
    const $chartDiv = $(`#${lib.chart_div()}`);
    let $sel;
    let scrollTop;
    let scrollLeft;

    if (arc.globals.is_dashboard) {
      $sel = $chartDiv.find(lib.viz_type() === 'crosstab' ? '.slick-viewport' : '.dataTables_scrollBody');
    } else {
      $sel = $('#main');
    }

    if ($sel?.length > 0) {
      scrollTop = $sel.scrollTop();
      scrollLeft = $sel.scrollLeft();
    }

    lib.viz().viz_drawVisual(lib.chart_div());

    $sel = $chartDiv.find(lib.viz_type() === 'crosstab' ? '.slick-viewport' : '.dataTables_scrollBody');

    if ($sel?.length > 0) {
      $sel.scrollTop(scrollTop);
      $sel.scrollLeft(scrollLeft);
    }
  };

  // Called after and expansion or collapse.
  // We register every row, to have the information
  // when the user clicks and expand or collapse icon.
  // Then we have to add the expand and collapse icons.
  // Afterwards we overwrite the data results object with the modified rows.
  // Lastly we trigger a redraw that will use the modified rows to render the visual.
  const onExpansionChange = () => {
    // First we register and add the icons to the original rows,
    // where everything was collapsed
    const rows = deepCopyRows(originalRows);
    rows.forEach((row) => {
      const rowNumber = createOrGetRowNumber(row, indexOfFirstExpandableDimension);
      insertExpandAndCollapseIcons(row, rowNumber, indexOfFirstExpandableDimension);
    });

    // Then we get the expanded rows that were added whenever a user clicked the expand icon.
    Object.values(expandedRowInfoByRowNumber)
      // We have to sort them by row number to not mess up the order.
      .sort((rowA, rowB) => rowA.rowNumber - rowB.rowNumber)
      .forEach((expandedRowInfo) => {
        const copiedRows = deepCopyRows(expandedRowInfo.rows);

        // Register then add the expand and collapse icons to these rows also
        copiedRows.forEach((row) => {
          const rowNumber = createOrGetRowNumber(row, expandedRowInfo.columnNumber);
          insertExpandAndCollapseIcons(row, rowNumber, expandedRowInfo.columnNumber);

          row.isFullyExpanded = expandedRowInfo.isFullyExpanded;
          // Insert them into the rows at the correct index
          rows.splice(expandedRowInfo.rowNumber + 1, 0, row);
        });
      });

    dataResult.rows(rows);

    redrawVisualAndScrollTo();
  };

  // This is called when an expand icon is cliked.
  // First we identify which row was clicked,
  // Then update the dimensions of the request object so that the currently collapsed
  // column will be requested.
  // We also update the filters of the request object, because we don't want to request all the data
  // only the parts where the column has the specific data that was expanded.
  // After the request object is prepared the request is sent.
  // When the data is retuerned it is saved into the expandedRowsByRowNumber.
  expansionTracker.expand = function(clickedElement, rowIndex) {
    const $clickedElement = $(clickedElement);
    const clickedRowNumber = Number($clickedElement.attr('data-row-number'));
    const clickedColumnIndex = Number($clickedElement.attr('data-column-index'));
    const notCollapsedDimensionValues = notCollapsedDimensionValuesByRowNumber[clickedRowNumber];
    if (!notCollapsedDimensionValues) {
      return expansionTracker;
    }

    const isFullyExpanded = clickedColumnIndex === originalDimensions.length - 1;

    const newDimensions = originalDimensions.map((dimension, index) => {
      const isDimensionExpansionEnabled = index >= indexOfFirstExpandableDimension;
      if (!isDimensionExpansionEnabled) {
        return { ...dimension };
      }

      const shelfElement = dimensionShelfElements[index];
      const isDimensionExpanded = clickedColumnIndex >= index;
      if (isDimensionExpanded) {
        const expression = shelfElement.getExpressionComponents();
        return { ...dimension, expr: `${expression.expr} as ${expression.alias}` };
      }

      const newExpr = getExprForCollapsedDimensionByShelfElement(shelfElement);
      return { ...dimension, expr: newExpr };
    });

    const requestBuilder = dataRequest.rb();
    requestBuilder.dimensions(newDimensions);

    const newFilters = [...originalFilters];
    notCollapsedDimensionValues.forEach((_, index) => {
      if (index >= indexOfFirstRowDimension) {
        const shelfElement = dimensionShelfElements[index];
        const { expr } = shelfElement.getExpressionComponents();
        const value = notCollapsedDimensionValues[index];

        const dimensionFilter = arc.misc.inClause(
          expr,
          shelfElement.dataset_coltype(),
          value,
          lib.dataconnection_type()
        );
        newFilters.push(dimensionFilter);
      }
    });
    requestBuilder.replaceFilters(newFilters);

    const dsreq = requestBuilder.dsreq();
    dataRequest.dsreq(dsreq);
    lib.dsreq(dsreq);

    dataRequest.finalize();

    lib.runSqlCmd(
      undefined,
      (data) => {
        for (let i = 0; i <= clickedColumnIndex; ++i) {
          colTypes[i] = data.coltypes[i];
        }

        expandedRowInfoByRowNumber[clickedRowNumber] = {
          rowNumber: rowIndex,
          columnNumber: clickedColumnIndex + 1,
          isFullyExpanded,
          rows: data.rows,
        };

        onExpansionChange();
      },
      JSON.stringify(dsreq)
    );
  };

  // This is called when a user clicks a collapse icon.
  expansionTracker.collapse = function(clickedElement) {
    const $clickedElement = $(clickedElement);
    const clickedRowNumber = Number($clickedElement.attr('data-row-number'));
    const clickedColumnIndex = Number($clickedElement.attr('data-column-index'));
    const isCollapseAllButton = clickedRowNumber === -1;
    // clickedExpandedDimensionValues is only undefined if the collapse all button
    // which only exists in a table visual
    const clickedExpandedDimensionValues = notCollapsedDimensionValuesByRowNumber[clickedRowNumber]
      ?.slice(indexOfFirstRowDimension);

    // If the user clicked the collapse all button on a table visual's header row
    // We collapse every row that was added by clicking on an expand icon that was
    // either at the same index or a higher index than the currecntly clicked collapse all icon.
    if (isCollapseAllButton) {
      Object.entries(expandedRowInfoByRowNumber).forEach(([rowNumber, { columnNumber }]) => {
        if (columnNumber > clickedColumnIndex) {
          delete expandedRowInfoByRowNumber[rowNumber];
        }
      });
    } else {
      // We go through each row
      Object.entries(notCollapsedDimensionValuesByRowNumber)
        .forEach(([rowNumber, notCollapsedDimensionValues]) => {
          // To identify if this row is either the in the row as the current row
          // or if the current row is a child of the clicked row,
          // we compare the columns of the row that are not collapsed
          // (meaning they have a value not just an x or 1 and they are dimensions, not measures).
          // Example:
          //                              | inspection_type |
          //                              | Complaint       |
          // postal_code | city | address | Record Count    |
          //         [-] |  [-] |     [-] |               1 |
          //       92671 |  [-] |     [-] |               1 |
          //             |   37 |     [+] |               1 |
          //       94158 |  [-] |     [-] |               2 |
          //             |   39 |     [+] |               2 |
          // In this case we have 5 rows
          // ['Complaint']
          // ['Complaint', 92671]
          // ['Complaint', 92671, 37]
          // ['Complaint', 94158]
          // ['Complaint', 94158, 39]

          const expandedDimensionValues = notCollapsedDimensionValues
            .slice(indexOfFirstRowDimension);

          const length = Math.min(
            clickedExpandedDimensionValues?.length,
            expandedDimensionValues.length
          );

          const isClickedOrChildRow = arc.arraysIdentical(
            clickedExpandedDimensionValues?.slice(0, length),
            expandedDimensionValues.slice(0, length)
          );

          if (isClickedOrChildRow) {
            const expandedRow = expandedRowInfoByRowNumber[rowNumber];

            if (expandedRow?.columnNumber > clickedColumnIndex) {
              delete expandedRowInfoByRowNumber[rowNumber];
            }
          }
        });
    }

    onExpansionChange();

    return expansionTracker;
  };

  return expansionTracker;
};
}());
