window.arc = window.arc || {};

(function () {
  'use strict';

  arc.report = arc.report || {};

  arc.report.columnstats = function () {
    const columnstats = arc.report.base();

    columnstats.shelf_data([]);

    let trackedColumns = [];
    let colVisualOverrides = {};

    let coldata = [];
    let globalDatasetFilters = [];
    let fetch_complete = false;
    let report;

    // Note that we persist trackedColumns, so the ids below
    // cannot be changed without breaking existing visuals
    const COL_NAME = 1;
    const COL_TYPE = 2;
    const COL_NDV = 3;
    const COL_MIN = 4;
    const COL_MAX = 5;
    const COL_NUM_NULLS = 6;
    const COL_DETAIL = 7;
    const COL_DETAIL_TYPE = 8;
    const COL_BUCKETSIZE = 9;

    const NDV_TOPK_THRESHOLD = 500;
    const TOPK_COUNT = 20;

    const COL_DETAIL_HISTOGRAM = 'histogram';
    const COL_DETAIL_TOPK = 'topk';
    const COL_DETAIL_NONE = 'none';

    const user_settings = arc.visual.defaultSettings({
      noAxes: true,
      noTrellis: true,
    })
      .removeSetting([
        'visual_show_detail_data',
      ]);
    columnstats.user_settings(user_settings);

    // Called once at startup time
    columnstats.setup = function () {
      $('.app-table').empty();
      d3.select(`#${columnstats.shelves_dom_id()}`)
        .selectAll('tr')
        .remove();
      $('#viz-refresh').show();
      return columnstats;
    };

    let saved_settings = {};

    // Dummy function added because columnstats's base object is not same as driver (ARC-9501)
    columnstats.report = function (_) {
      if (!arguments.length) return report;
      report = _;
      return columnstats;
    };

    // Called to return an object that can later be passed into
    // columnstats.restore to restore state
    columnstats.save = function () {
      let settings = saved_settings;
      settings = columnstats.save_basics(settings);

      settings.viztype_settings = settings.viztype_settings || {};
      settings.viztype_settings[columnstats.viz_type()] =
        settings.viztype_settings[columnstats.viz_type()] || {};

      const s = settings.viztype_settings[columnstats.viz_type()];
      s.colVisualOverrides = colVisualOverrides;
      s.trackedColumns = trackedColumns;

      saved_settings = settings;
      return saved_settings;
    };

    columnstats.restore = function (settings, switchingVisualType) {
      if (settings === undefined) {
        return columnstats;
      }

      saved_settings = settings;
      columnstats.restore_basics(settings, switchingVisualType);

      if (settings.viztype_settings &&
        settings.viztype_settings[columnstats.viz_type()]) {
        const s = settings.viztype_settings[columnstats.viz_type()];
        if (s.colVisualOverrides) {
          colVisualOverrides = s.colVisualOverrides;
        }

        if (s.trackedColumns) {
          trackedColumns = s.trackedColumns;
        }
      }

      return columnstats;
    };

    columnstats.refresh = function () {
      // NOTE: this counter is used in arcweb/apps/views_snapshot.py
      arc.counters.increment('viz_drawVisual_start');
      columnstats.redraw();

      // NOTE: this counter is used in arcweb/apps/views_snapshot.py
      arc.counters.increment('viz_drawVisual_end');
      return columnstats;
    };

    // This visual makes a large number of distinct calls to the backend. If a redraw
    // call occurs and we decide to start again from scratch, then we need to interrupt
    // the existing chain of calls. We interrupt this chain by setting and checking
    // statsVersion below, which gets incremented everytime interrupting the current chain
    // is necessary
    let currentStatsVersion = 1;

    function initializeState() {
      globalDatasetFilters = [];
      coldata = [];
      fetch_complete = false;
    }

    columnstats.redraw = function () {
      const chart_div = this.chart_div();

      // Reset the sqlQuery parameter so a refresh happens when we
      // switch visual types
      columnstats.sqlQuery(null);
      columnstats.undraw();

      currentStatsVersion += 1;
      const statsVersion = currentStatsVersion;

      const divid = `colstatstable_div${columnstats.report_id()}-${statsVersion}`;

      const d3_chart_div = d3.select(`#${chart_div}`);
      const height = d3_chart_div.node().getBoundingClientRect().height - 30;

      // Make a div with some padding, and the add in the table div to it
      d3_chart_div
        .append('div')
        .style('height', `${height}px`)
        .style('overflow', 'scroll')
        .append('div')
        .attr('id', divid);

      // Get dataset level filters in datasetFilters
      const datasetFilters = [];
      columnstats.datasetFiltersCb((expr) => {
        datasetFilters.push(expr);
      });

      columnstats.handleSolrParam((expr) => {
        datasetFilters.push(expr);
      });

      if (arc.arraysIdentical(datasetFilters, globalDatasetFilters) &&
        fetch_complete) {
        drawColumns(`#${divid}`, statsVersion);
        return columnstats;
      }

      initializeState();
      globalDatasetFilters = datasetFilters;

      function afterCols() {
        drawColumns(`#${divid}`, statsVersion);
        fetchStats(`#${divid}`, undefined, statsVersion);
      }

      const colDetailTypes = [COL_DETAIL_HISTOGRAM, COL_DETAIL_TOPK, COL_DETAIL_NONE];
      if (trackedColumns && trackedColumns.length > 0) {
        coldata = [];
        trackedColumns.forEach((tc) => {
          const row = ['', tc[1], tc[2], undefined, '', '', '', '', '', ''];
          if (colDetailTypes.indexOf(tc[COL_DETAIL_TYPE]) >= 0) {
            row[COL_DETAIL_TYPE] = tc[COL_DETAIL_TYPE];
          }

          coldata.push(row);
        });

        afterCols();
        return columnstats;
      }

      if (columnstats.dataset()) {
        populateColnamesFromDataset();
        afterCols();
        return columnstats;
      }

      const sb = getSelectBuilder();
      sb.limit(0);
      columnstats.runSqlCmd(sb.getSql(), (o) => {
        columnstats.populateFromSqlOutput(o);
        populateColnamesFromSbresult();
        afterCols();
      });

      return columnstats;
    };

    function populateColnamesFromSbresult() {
      const sb_colnames = columnstats.sb_colnames();
      const sb_coltypes = columnstats.sb_coltypes();

      coldata = [];
      sb_colnames.forEach((d, i) => {
        coldata.push(['', d, sb_coltypes[i].toLowerCase(), undefined, '', '', '', '', '', '']);
      });

      trackedColumns = coldata;

      return columnstats;
    }

    function populateColnamesFromDataset() {
      function onetab(t) {
        t.columns().forEach((c) => {
          const coltype = c.attr('type');
          if (arc.report.coltypeIsHierarchy(coltype)) {
            return;
          }

          // Only allow compute column stats on base table columns
          if (!c.attr('exprhasagg') && c.isBaseCol()) {
            coldata.push([
              '',
              c.attr('alias') || c.attr('name'),
              c.attr('type').toLowerCase(),
              undefined,
              '',
              '',
              '',
              '',
              '',
              '',
            ]);
          }
        });
      }

      coldata = [];
      columnstats.dataset().tables().forEach((t) => {
        onetab(t);
      });

      trackedColumns = coldata;

      return columnstats;
    }

    function populateBaseColdata(o, start_col) {
      if (o.rows.length > 0) {
        const d = o.rows[0];
        let coli = 0;
        for (let i = start_col; i < start_col + d.length / 4; ++i) {
          coldata[i][COL_MIN] = d[coli];
          coli += 1;
          coldata[i][COL_MAX] = d[coli];
          coli += 1;
          coldata[i][COL_NDV] = Number(d[coli]);
          coli += 1;
          coldata[i][COL_NUM_NULLS] = Number(d[coli]);
          coli += 1;
        }
      }

      return columnstats;
    }

    function fetchDetail(container_div, currentcol, statsVersion, oneTime) {
      if (statsVersion !== currentStatsVersion) {
        return;
      }

      let expr;
      const col_count = coldata.length;

      const connectionType = columnstats.dataconnection_type();

      if (!currentcol) {
        currentcol = 0;
      }

      if (currentcol >= col_count) {
        // Normal termination. Indicate that we fetched all the data
        fetch_complete = true;
        return columnstats;
      }

      if (connectionType === 'solr') {
        // solr column stat details is not implemented. This code will mock
        // an empty response to fill the details column
        coldata[currentcol][COL_DETAIL] = JSON.parse('{"colnames":[], "rows":[]}');
        drawColumns(container_div, statsVersion);
        if (currentcol < col_count) {
          fetchDetail(container_div, currentcol + 1, statsVersion);
        }

        return columnstats;
      }

      if (connectionType === 'solr') {
        // solr column stat details is not implemented. This code will mock
        // an empty response to fill the details column
        coldata[currentcol][COL_DETAIL] = JSON.parse('{"colnames":[], "rows":[]}');
        drawColumns(container_div, statsVersion);
        if (currentcol < col_count) {
          fetchDetail(container_div, currentcol + 1, statsVersion);
        }

        return columnstats;
      }

      const sb = getSelectBuilder();
      const colname = coldata[currentcol][COL_NAME];
      const coltype = coldata[currentcol][COL_TYPE];
      let visualType = COL_DETAIL_NONE;

      if (colname in colVisualOverrides) {
        const t = colVisualOverrides[colname];
        if (t === COL_DETAIL_TOPK) {
          visualType = COL_DETAIL_TOPK;
        } else if (t === COL_DETAIL_HISTOGRAM) {
          visualType = COL_DETAIL_HISTOGRAM;
        }
      }

      if (visualType === COL_DETAIL_NONE) {
        if ((coldata[currentcol][COL_NDV] > 0) &&
          (coldata[currentcol][COL_NDV] <= NDV_TOPK_THRESHOLD) &&
          !arc.report.coltypeIsNumber(coltype)) {
          visualType = COL_DETAIL_TOPK;
        }

        if (arc.report.coltypeIsNumber(coltype)) {
          visualType = COL_DETAIL_HISTOGRAM;
        }
      }

      coldata[currentcol][COL_DETAIL_TYPE] = visualType;

      if (visualType === COL_DETAIL_TOPK) {
        if (columnstats.dataset()) {
          expr = sb.addTablenamesFromExpression(arc.dataconnection.types.brackets(colname));
          sb.addProjection(expr);
        } else {
          expr = colname;
          sb.addProjection(colname);
        }

        sb.addFilter(sb.addTablenamesFromExpression(
          `${arc.dataconnection.types.brackets(expr)} IS NOT NULL`
        ));

        sb.addProjection('sum(1) as Records');
        if (connectionType === 'drill') {
          sb.addOrderBy('2 desc');
          sb.addGroupWithoutProjection(`\`${colname}\``);
        } else if (connectionType === 'oracle') {
          sb.addOrderBy('2 desc');
          sb.addGroupWithoutProjection(`"${colname}"`);
        } else if (['hyve', 'hive'].includes(connectionType)) {
          sb.addOrderBy('Records desc');
          sb.addGroupWithoutProjection(`${colname}`);
        } else {
          sb.addOrderBy('2 desc');
          sb.addGroupWithoutProjection('1');
        }

        sb.limit(TOPK_COUNT);
        columnstats.runSqlCmd(sb.getSql(), (o) => {
          coldata[currentcol][COL_DETAIL] = o;
          drawColumns(container_div, statsVersion);
          if (!oneTime) {
            fetchDetail(container_div, currentcol + 1, statsVersion);
          }

          return columnstats;
        }, undefined);

        return columnstats;
      }

      if (visualType === COL_DETAIL_HISTOGRAM) {
        const range = coldata[currentcol][COL_MAX] - coldata[currentcol][COL_MIN];

        // If we're drawing all the columns, skip to the next column if range is invalid
        if (!arc.misc.isNumeric(range)) {
          if (!oneTime) {
            fetchDetail(container_div, currentcol + 1, statsVersion);
          }

          return columnstats;
        }

        const bucketsize = Math.max(Math.floor(range / 10), 1);
        coldata[currentcol][COL_BUCKETSIZE] = bucketsize;
        let cast_col;
        if (connectionType === 'mysql') {
          cast_col = `cast(([${colname}] - ${coldata[currentcol][COL_MIN]})/${bucketsize} as signed)`;
        } else {
          cast_col = `cast(([${colname}] - ${coldata[currentcol][COL_MIN]})/${bucketsize} as integer)`;
        }

        cast_col = sb.addTablenamesFromExpression(cast_col);
        expr = `${cast_col} as bucket`;
        let sql_query;
        if (connectionType === 'oracle') {
          expr = expr.replace(/`/g, '"'); // use double quoter oracle and subselect query
          sb.addProjection(expr);
          sql_query = `select bucket, sum(1) from (${sb.getSql()
            }) group by bucket order by bucket asc`;
        } else if (connectionType === 'drill') {
          sb.addProjection(expr);
          sql_query = `select bucket, sum(1) from (${sb.getSql()
            }) group by bucket order by bucket asc`;
        } else {
          sb.addProjection(expr);
        }

        // It seems like in HIVE we can not use column indices for GROUP BY and ORDER BY queries
        const orderBy = ['hyve', 'hive'].includes(connectionType) ? 'bucket asc' : '1 asc';
        const groupBy = ['hyve', 'hive'].includes(connectionType) ? `${cast_col}` : '1';

        sb.addFilter(sb.addTablenamesFromExpression(
          `${arc.dataconnection.types.brackets(colname)} IS NOT NULL`
        ));
        sb.addProjection('sum(1)');
        sb.addOrderBy(orderBy);
        sb.addGroupWithoutProjection(groupBy);
        columnstats.runSqlCmd(sql_query || sb.getSql(), (o) => {
          coldata[currentcol][COL_DETAIL] = o;
          drawColumns(container_div, statsVersion);
          if (!oneTime) {
            fetchDetail(container_div, currentcol + 1, statsVersion);
          }

          return columnstats;
        }, undefined);

        return columnstats;
      }

      if (!oneTime) {
        setTimeout(() => {
          fetchDetail(container_div, currentcol + 1, statsVersion);
        }, 0);
      }

      return columnstats;
    }

    function numNulls(expr) {
      return `sum(case when ${expr} IS NULL then 1 else 0 end)`;
    }

    function countDistinctFn(expr) {
      return arc.dataconnection.approximateOrExactCount(
        columnstats.dataconnection_type(), expr
      );
    }

    function getSelectBuilder() {
      const sb = columnstats.getSelectBuilder();
      globalDatasetFilters.forEach((expr) => {
        sb.addFilter(sb.expression(expr));
      });

      return sb;
    }

    function fetchStats(container_div, currentcol, statsVersion) {
      if (statsVersion !== currentStatsVersion) {
        return;
      }

      let columns_at_a_time;
      const col_count = coldata.length;

      if (columnstats.dataconnection_type() === 'solr') {
        columns_at_a_time = 1;
      } else {
        columns_at_a_time = 5;
      }

      if (!currentcol) {
        currentcol = 0;
      }

      if (currentcol >= col_count) {
        return fetchDetail(container_div, undefined, statsVersion);
      }

      const sb = getSelectBuilder();
      const ds = columnstats.dataset();
      const rb = columnstats.getRequestBuilder();
      let type = 'columnstats';
      columnstats.dsreq(JSON.stringify(rb.dsreq()));
      let done = 0;
      const solr_query = [];
      if (ds) {
        let table;
        for (let i = currentcol; i < currentcol + columns_at_a_time; ++i) {
          if (i < col_count) {
            const name = coldata[i][COL_NAME];

            // Stop if we're about to change tables...
            if (!table) {
              table = ds.get_column(name).table();
            } else if (table !== ds.get_column(name).table()) {
              break;
            }

            done += 1;
            let expr = sb.addTablenamesFromExpression(arc.dataconnection.types.brackets(name));
            if (columnstats.dataconnection_type() === 'oracle') {
              expr = expr.replace(/`/g, '"');
            } else if (columnstats.dataconnection_type() === 'solr') {
              // solr column state queries will be generated in django
              solr_query.push(`columnstats:${name}`);
              type = 'SOLR';
            }

            sb.addProjection(`min(${expr})`);
            sb.addProjection(`max(${expr})`);
            sb.addProjection(countDistinctFn(expr));
            sb.addProjection(numNulls(expr));
          }
        }
      } else {
        for (let j = currentcol; j < currentcol + columns_at_a_time; ++j) {
          if (j < col_count) {
            done += 1;
            let cname = coldata[j][COL_NAME];
            if (columnstats.dataconnection_type() === 'oracle') {
              cname = cname.replace(/`/g, '"');
            }

            sb.addProjection(`min(${cname})`);
            sb.addProjection(`max(${cname})`);
            sb.addProjection(countDistinctFn(cname));
            sb.addProjection(numNulls(cname));
          }
        }
      }

      rb.requestType(type);
      const dsreq = rb.dsreq();
      dsreq.solr_query = solr_query;
      columnstats.runSqlCmd(sb.getSql(), (o) => {
        populateBaseColdata(o, currentcol);
        drawColumns(container_div, statsVersion);

        currentcol += done;
        fetchStats(container_div, currentcol, statsVersion);
      }, JSON.stringify(dsreq));

      return columnstats;
    }

    function drawColumns(container_div, statsVersion) {
      if (statsVersion !== currentStatsVersion) {
        return;
      }

      let viewMode = true;
      if (arc.globals.is_edit && !arc.globals.dashboard) {
        viewMode = false;
      }

      const timestamp = Date.now();
      d3.selectAll(container_div)
        .selectAll('table')
        .data([1])
        .enter()
        .append('table')
        .attr('id', `data-table${timestamp}`)
        .attr('class', 'colstats-table table table-width-auto table-striped ' +
          'table-bordered table-sm')
        .classed('colstats-view-table', viewMode);

      const table = d3.selectAll(container_div)
        .select('table');
      table.selectAll('thead')
        .data([1])
        .enter()
        .append('thead');
      table.selectAll('tbody')
        .data([1])
        .enter()
        .append('tbody');

      const thead = table.selectAll('thead');
      const tbody = table.selectAll('tbody');

      const headers = [
        'Actions', 'Field', 'Type', 'Distinct Values',
        'Minimum', 'Maximum', 'Number of nulls', 'Detail',
      ];

      // append the header row
      thead.selectAll('tr')
        .data([1])
        .enter()
        .append('tr')
        .selectAll('th')
        .data(headers)
        .enter()
        .append('th')
        .style('min-width', (d, i) => (i === 0 ? '80px' : '20px'))
        .style('width', (d, i) => (i === 0 ? '80px' : '20px'))
        .attr('align', 'left')
        .text((d) => d);

      // create a row for each object in the sb_rows
      let rows =
        tbody.selectAll('tr')
          .data(coldata);

      rows.exit()
        .remove();

      rows.enter()
        .append('tr')
        .classed('column-data', true);

      rows = tbody.selectAll('tr')
        .attr('data-colindex', (d, i) => i)
        .attr('data-colname', (d) => d[COL_NAME])
        .attr('data-coltype', (d) => d[COL_TYPE]);

      // create a cell in each row for each column
      let cells = rows.selectAll('td')
        .data((d, i) => coldata[i].slice(0, 8));

      const nameTemplate = $('#colstats-type-template').html();
      const actionsTemplate = $('#colstats-actions-template').html();

      function columnData(d, i) {
        if (i === 0) {
          return actionsTemplate;
        }

        if (i === 1) {
          return arc.misc.safe_substitute(nameTemplate, {
            COLNAME: d,
            LINK_CLASS: arc.globals.is_edit ? 'link-active' : 'link-inactive',
          });
        }

        return arc.misc.escapeHTML(d);
      }

      // update existing cells
      cells.filter((d, i) => (i !== COL_DETAIL))
        .html(columnData);

      // add new cells
      cells.enter()
        .append('td')
        .filter((d, i) => (i !== COL_DETAIL))
        .html(columnData);

      $(container_div).find('tr.column-data a.down-column')
        .last()
        .css('opacity', '0')
        .css('cursor', 'default');

      rows.selectAll('a.top-column')
        .on('click.top', function () {
          const $tr = $(this).closest('tr');
          const colindex = Number($tr.attr('data-colindex'));

          if (colindex <= 0) {
            return;
          }

          const r = trackedColumns[colindex];
          trackedColumns.splice(colindex, 1);
          trackedColumns.splice(0, 0, r);

          coldata = trackedColumns;
          drawColumns(container_div, statsVersion);
          columnstats.page_needs_save();
        });

      function drawColumnsAgain() {
        coldata = trackedColumns;
        drawColumns(container_div, statsVersion);
        columnstats.page_needs_save();
      }

      rows.selectAll('a.up-column')
        .on('click.up', function () {
          const $tr = $(this).closest('tr');
          const colindex = Number($tr.attr('data-colindex'));

          if (colindex <= 0) {
            return;
          }

          const r = trackedColumns[colindex];
          trackedColumns[colindex] = trackedColumns[colindex - 1];
          trackedColumns[colindex - 1] = r;

          drawColumnsAgain();
        });

      rows.selectAll('a.down-column')
        .on('click.down', function () {
          const $tr = $(this).closest('tr');
          const colindex = Number($tr.attr('data-colindex'));

          if (colindex >= (trackedColumns.length - 1)) {
            return;
          }

          const r = trackedColumns[colindex];
          trackedColumns[colindex] = trackedColumns[colindex + 1];
          trackedColumns[colindex + 1] = r;

          drawColumnsAgain();
        });

      rows.selectAll('a.remove-column')
        .on('click.remove', function () {
          const $tr = $(this).closest('tr');
          const colindex = Number($tr.attr('data-colindex'));

          trackedColumns.splice(colindex, 1);

          drawColumnsAgain();
        });

      rows.selectAll('a.set-topk')
        .on('click.changetype', function () {
          const $tr = $(this).closest('tr');
          const colname = $tr.attr('data-colname');
          const colindex = Number($tr.attr('data-colindex'));

          colVisualOverrides[colname] = COL_DETAIL_TOPK;
          fetchDetail(container_div, colindex, statsVersion, true);
          columnstats.page_needs_save();
        });

      rows.selectAll('a.set-histogram')
        .on('click.changetype', function () {
          const $tr = $(this).closest('tr');
          const colname = $tr.attr('data-colname');
          const colindex = Number($tr.attr('data-colindex'));

          colVisualOverrides[colname] = COL_DETAIL_HISTOGRAM;
          fetchDetail(container_div, colindex, statsVersion, true);
          columnstats.page_needs_save();
        });

      cells =
        rows.selectAll('td')
          .filter((d, i) => (i === COL_DETAIL))
          .each(function (d) {
            d3.select(this)
              .selectAll('svg')
              .remove();
            d3.select(this)
              .selectAll('div')
              .remove();
            if (d === undefined || d === '' || arc.misc.check_if_value_null(d)) {
              return;
            }

            const coldata_elem = d3.select(this.parentNode).datum();
            const qr = arc.queryResults().data(d);
            const colnames = qr.colnames();
            const extent = d3.extent(d.rows, (de) => de[1]);

            extent[0] = Math.min(0, extent[0]);
            const h = d3.scale.linear()
              .domain([extent[1], extent[0]])
              .range([90, 5]);

            const rect =
              d3.select(this)
                .append('svg')
                .attr('width', `${20 + (d.rows.length * 15)}px`)
                .attr('height', '100px')
                .selectAll('rect')
                .data(d.rows)
                .enter()
                .append('rect')
                .attr('x', (dx, i) => 10 + (i * 15))
                .attr('y', (dy) => 95 - h(Number(dy[1])))
                .attr('width', 12)
                .attr('height', (dh) => h(Number(dh[1])))
                .style('fill', () => {
                  if (coldata_elem[COL_DETAIL_TYPE] === COL_DETAIL_HISTOGRAM) {
                    return 'rgb(253, 141, 60)';
                  }

                  return 'rgb(107, 174, 214)';
                })
                .each(function () {
                  const $tr = $(this).closest('tr');

                  // For numeric types, permit flipping between topk and histograms
                  const coltype = $tr.attr('data-coltype');
                  if (!arc.report.coltypeIsNumber(coltype)) {
                    return;
                  }

                  if (coldata_elem[COL_DETAIL_TYPE] === COL_DETAIL_HISTOGRAM) {
                    $tr.addClass(COL_DETAIL_HISTOGRAM);
                    $tr.removeClass(COL_DETAIL_TOPK);
                  } else {
                    $tr.addClass(COL_DETAIL_TOPK);
                    $tr.removeClass(COL_DETAIL_HISTOGRAM);
                  }
                });

            const numFormat = d3.format(',');
            arc.tooltip()
              .dom_id('#tooltip')
              .text_fn((dtf, i) => {
                const kvs = [];

                if (coldata_elem[COL_DETAIL_TYPE] ===
                  COL_DETAIL_HISTOGRAM) {
                  kvs.push({
                    key: 'Histogram',
                    value: '',
                  });
                } else {
                  kvs.push({
                    key: `Top ${TOPK_COUNT}`,
                    value: '',
                  });
                }

                let val;
                if (coldata_elem[COL_DETAIL_TYPE] ===
                  COL_DETAIL_HISTOGRAM) {
                  const min = Number(coldata_elem[COL_MIN]);
                  const bucketsize = Number(coldata_elem[COL_BUCKETSIZE]);
                  val = `${min + (i * bucketsize)}`;
                  val += ' - ';
                  val += (min + ((i + 1) * bucketsize));
                  kvs.push({
                    key: 'Bucket Range',
                    value: val,
                  });
                  kvs.push({
                    key: 'Frequency',
                    value: numFormat(dtf[1]),
                  });
                } else {
                  kvs.push({
                    key: colnames[0],
                    value: columnstats.d_string(dtf[0]),
                  });
                  kvs.push({
                    key: colnames[1],
                    value: numFormat(dtf[1]),
                  });
                }

                return columnstats.tooltip_from_kvs(kvs);
              })
              .addSelection(rect)
              .activate();
          });

      return columnstats;
    }

    return columnstats;
  };
}());
