/* eslint-disable no-unused-vars */
window.arc = window.arc || {};

(function() {
'use strict';

arc.visual = arc.visual || Object.create(null);
arc.visual.plugins = arc.visual.plugins || Object.create(null);
arc.visual.extensions = arc.visual.extensions || [];

arc.visual.plugins.combined = function() {
  let saved_settings = Object.create(null);
  let data_result;
  let lib;
  let trellis;
  let sort_orders_asc = Object.create(null);

  const combined = arc.visual.api();

  function setMultipleStackedBarsToPrimaryAxis(measureShelfElements) {
    if (measureShelfElements.length > 1) {
      const secondaryElements = measureShelfElements
                                  .filter((element) => element.secondaryaxis());
      if (secondaryElements.length) {
        const names = secondaryElements.map((e) => e.col_alias() || e.dataset_colname()).toString();
        lib.alertMessage(`Multiple bar measures must share the same axis. Thee following bar measure${secondaryElements.length > 1 ? 's were' : ' was'} moved to the primary axis: ${names}.`);
        secondaryElements.forEach((element) => element.secondaryaxis(false));
      }
    }
  }

  // New requirement: move all measures to primary axis if number of measures is greater than 1.
  // This function is added to handle previously saved measures with secondary data.
  function handleSavedSecondaryAxisSettingOnBarMeasures(settings) {
    if (settings?.bar_measure?.length > 1) {
      const barMeasureWithSecondaryAxis =
        settings.bar_measure.filter((barMeasure) => barMeasure.secondaryaxis);
      barMeasureWithSecondaryAxis.forEach((barMeasure) => {
        barMeasure.secondaryaxis = false;
      });

      // Note: it would be nice to add warning somewhere,
      // but during the setup other function clear all the alarms
      // and it wouldn't be visible (in this case check_shelf_contents will clear all alarm)
      // TODO create specific alarm (hierarch) - which shouldn't be cleared automatically.
    }
  }

  const shelf_data = [
    {
      name: 'x_shelf',
      display_name: 'Dimension',
      shelf_type: arc.report.shelf_types.COLUMN,
      column_type: arc.report.columnTypeEnum.DIMENSION,
    },
    {
      name: 'bar_measure',
      display_name: 'Bar Measure',
      shelf_type: arc.report.shelf_types.COLUMN,
      column_type: arc.report.columnTypeEnum.AGGREGATE,
      onShelfListChange: setMultipleStackedBarsToPrimaryAxis
    },
    {
      name: 'line_measure',
      display_name: 'Line Measure',
      shelf_type: arc.report.shelf_types.COLUMN,
      column_type: arc.report.columnTypeEnum.AGGREGATE,
    },
    {
      name: 'highlighter_measure',
      display_name: 'Highlight Mark Measure',
      shelf_type: arc.report.shelf_types.COLUMN,
      column_type: arc.report.columnTypeEnum.AGGREGATE,
    },
    {
      name: 'tooltip_shelf',
      display_name: 'Tooltips',
      shelf_type: arc.report.shelf_types.COLUMN,
      column_type: arc.report.columnTypeEnum.AGGREGATE,
    },
    {
      name: 'filters_shelf',
      display_name: 'Filters',
      display_html: arc.report.filters_shelf_html(),
      shelf_type: arc.report.shelf_types.PREDICATE,
      column_type: arc.report.columnTypeEnum.UNKNOWN,
      onShelfListChange: arc.report.updateFilterShelfLabel,
    },
  ];

  // Will return true for all names that are not in the array
  // Used to filter out shelves or columns not relevant to the visual type with
  // specified shelf
  const isColumn = {
    bar_measure(shelfName) {
      return ['line_measure', 'highlighter_measure'].indexOf(shelfName) === -1;
    },

    line_measure(shelfName) {
      return ['bar_measure', 'highlighter_measure'].indexOf(shelfName) === -1;
    },

    highlighter_measure(shelfName) {
      return ['bar_measure', 'line_measure'].indexOf(shelfName) === -1;
    },
  };

  const bar_shelf_data = shelf_data.filter(function filterUnrelatedShelves(sd) {
    return isColumn.bar_measure(sd.name);
  });

  const line_shelf_data = shelf_data.filter(function filterUnrelatedShelves(sd) {
    return isColumn.line_measure(sd.name);
  });

  const highlight_shelf_data = shelf_data.filter(function filterUnrelatedShelves(sd) {
    return isColumn.highlighter_measure(sd.name);
  });

  const linePlugin = arc.visual.plugins.bar();
  const barPlugin = arc.visual.plugins.bar();
  const highlightPlugin = arc.visual.plugins.bar();
  let barConfig = Object.create(null);
  let lineConfig = Object.create(null);
  let highlightConfig = Object.create(null);

  runAllPlugins((plugin, config) => {
    plugin.combined_visual(true);
    config.combined_visual = true;
  });

  const markTypeSetting = arc.visual.setting()
                                  .id('highlight_mark_type')
                                  .tab('Marks')
                                  .type(arc.visual.settingTypeEnum.RADIO)
                                  .group('Highlight Mark Type')
                                  .add_name('CIRCLE')
                                  .add_label('Circle')
                                  .add_name('DOT')
                                  .add_label('Dot')
                                  .add_name('STAR')
                                  .add_label('Star')
                                  .value('CIRCLE');

  // Used in arc_visual_bar.js>create_highlights function to control size of marks
  const markRadiusSetting = arc.visual.setting()
                                    .id('highlight_mark_radius')
                                    .tab('Marks')
                                    .type(arc.visual.settingTypeEnum.SLIDER)
                                    .add_name('MARK_RADIUS')
                                    .add_label('Max Highlight Radius')
                                    .min(0)
                                    .max(200)
                                    .value('20');

  const useArea = arc.visual.setting()
                          .id('use_area')
                          .tab('Marks')
                          .type(arc.visual.settingTypeEnum.CHECKBOX)
                          .add_name('USE_AREA')
                          .add_label('Convert line to area visual')
                          .value('false');

  const user_settings = barPlugin.user_settings();
  user_settings.removeSetting('trellis_font_size');
  user_settings.removeSetting('use_dataset_color_settings');
  user_settings.removeSetting('dim_to_color_mapping');
  linePlugin.user_settings(user_settings);
  highlightPlugin.user_settings(user_settings);

  user_settings.addSetting([markTypeSetting, useArea, markRadiusSetting]);

  linePlugin.mark_type('line');
  barPlugin.mark_type('bar');
  highlightPlugin.mark_type('highlight');
  linePlugin.shelf_data(line_shelf_data);
  barPlugin.shelf_data(bar_shelf_data);
  highlightPlugin.shelf_data(highlight_shelf_data);
  linePlugin.y_shelf('line_measure');
  barPlugin.y_shelf('bar_measure');
  highlightPlugin.y_shelf('highlighter_measure');

  makeSettingHidden('dynamic_bar_width');
  makeSettingHidden('legend_visibility');
  makeSettingHidden('legend_width_override');

  // TODO
  makeSettingHidden('trendline_linear');
  makeSettingHidden('trendline_color');
  makeSettingHidden('trendline_decimal');
  makeSettingHidden('scale_dim_independent');
  makeSettingHidden('scale_agg_independent');

  // Only supporting log scales right now
  scale_type().names(['LINEAR', 'LOG10']);
  scale_type().labels(['Linear', 'log<sub>10</sub>']);

  function scale_type() {
    return user_settings.find_by_id('scale_type');
  }

  function use_area() {
    return user_settings.find_by_id('use_area');
  }

  const constraints = {
    minimums: {
      x_shelf: 1,
    },
    maximums: {
      x_shelf: 1,
      highlighter_measure: 1,
    },
  };

  combined.viz_lib = function(_) {
    if (!arguments.length) return lib;
    lib = _;
    runAllPlugins((plugin, config) => {
      plugin.viz_lib(lib);
    });

    lib.shelf_data(shelf_data)
       .user_settings(user_settings)
       .chart_type_graded(0);
    return combined;
  };

  combined.viz_getSettings = function() {
    let settings = saved_settings;
    settings = lib.save_basics(settings);
    settings.sort_orders_asc = sort_orders_asc;
    saved_settings = settings;
    return settings;
  };

  combined.viz_setSettings = function(settings, switchingVisualType) {
    if (settings === undefined) {
      return combined;
    }

    handleSavedSecondaryAxisSettingOnBarMeasures(settings);
    saved_settings = settings;
    lib.restore_basics(settings, switchingVisualType);
    if ('sort_orders_asc' in saved_settings) {
      sort_orders_asc = saved_settings.sort_orders_asc;
    }

    return combined;
  };

  combined.viz_drawDesigner = function(shelves_div) {
    lib.shelves_dom_id(shelves_div)
       .base_setup();

    // Define messages to be included in shelves
    const xMessage = [
      lib.createPopoverMessages('xAxis'),
      lib.createPopoverMessages('combinedXAxis'),
    ];
    const barMeasure = [
      lib.createPopoverMessages('yAxis'),
      lib.createPopoverMessages('barOnly'),
    ];
    const highlightMeasure = [
      lib.createPopoverMessages('yAxis'),
      lib.createPopoverMessages('highlightOnly'),
    ];
    const lineMeasure = [
      lib.createPopoverMessages('yAxis'),
      lib.createPopoverMessages('lineOnly'),
    ];

    // Create those popovers
    lib.addShelfDescription(constraints, {
      x_shelf: xMessage,
      bar_measure: barMeasure,
      line_measure: lineMeasure,
      highlighter_measure: highlightMeasure,
    });

    return combined;
  };

  combined.viz_getDataRequest = function() {
    // If multiple aggregates, don't allow colors
    // Check that there's something in either the line or bar shelf
    if (!onlyAggsInMeasureShelves()) {
      return null;
    }

    return lib.getCheckedDataRequest(constraints);

    function onlyAggsInMeasureShelves() {
      let allAggregates = true;
      let atLeastOneColumn = false;
      lib.shelfElementCb('bar_measure', checkElement);
      lib.shelfElementCb('line_measure', checkElement);
      lib.shelfElementCb('highlighter_measure', checkElement);

      if (!atLeastOneColumn) {
        lib.alertMessage('Please add measure to Bar, Line, or Highlight Mark shelf.');
      }

      if (!allAggregates) {
        lib.alertMessage('Please only specify aggregates in the ' +
            'Bar and Line Measure shelf.');
      }

      return allAggregates && atLeastOneColumn;

      function checkElement(shelf_element) {
        atLeastOneColumn = true;
        if (!shelf_element.expr_hasagg() && !shelf_element.expressionIsAggregate()) {
          allAggregates = false;
        }
      }
    }
  };

  /*
   * Set the data this object should use to draw out the visual
   */
  combined.viz_setDataResult = function(result) {
    data_result = result;
    runAllPlugins((plugin, config) => {
      // Only include relevant columns so base.compute_agg_properties doesn't
      // filter out nulls from other columns
      const filteredColumns = result.columns()
                                    .filter(function filterUnrelatedMeasures(col) {
                                      return isColumn[plugin.y_shelf()](col.colsource());
                                    });

      const newResult = arc.data.result()
                                .columns(filteredColumns)
                                .rows(result.rows());

      plugin.viz_setDataResult(newResult);
    }, true);

    return combined;
  };

  combined.viz_drawVisual = function(viz_div) {
    user_settings.find_by_id('secondary_min_agg_axis').disable();
    user_settings.find_by_id('secondary_max_agg_axis').disable();
    const $chart_div = $(`#${viz_div}`);
    let $barVisual;
    barConfig = Object.create(null);
    lineConfig = Object.create(null);
    highlightConfig = Object.create(null);

    lib.chart_div(viz_div);
    lib.undraw();
    lib.colors(lib.contrast_colors());

    if (use_area().isChecked) {
      linePlugin.mark_type('area');
    } else {
      linePlugin.mark_type('line');
    }

    runAllPlugins((plugin, config) => {
      config.combined_visual = true;
      config.rows = data_result.rows();
    }, true);

    if (barConfig.rows.length === 0) {
      return combined;
    }

    lib.unselectOnClick();

    const validationResult = runAllPlugins((plugin, config) => {
      if (!plugin.init_viz(config) ||
          !plugin.setup_colors(config) ||
          !plugin.setup_dims_aggs(config)) {
        return false;
      }

      return true;
    }, true);

    const combinedCols = {
      bar: lib.shelf_columns(barConfig.columns, 'bar_measure'),
      line: lib.shelf_columns(lineConfig.columns, 'line_measure'),
      highlight: lib.shelf_columns(highlightConfig.columns, 'highlighter_measure'),
    };
    const colNumbers = {
      bar: combinedCols.bar.length,
      line: combinedCols.line.length,
      highlight: combinedCols.highlight.length,
    };

    const y_secondary_labels = [];
    const y_labels = [];
    const y_primary_axis = {};
    const y_cols = [];
    const y_secondary_cols = [];
    const ycol_idxs = {};

    const originalColors = lib.colors();
    const originalColorsLength = originalColors.length;

    // Hack to calculate colors for lines without overlapping with bars
    // and calculate colors for highlights without overlapping with bars + lines
    runAllPlugins((plugin, config) => {
      let markType = plugin.mark_type();
      markType = markType === 'area' ? 'line' : markType;
      let numberColorsToSkip = 0;
      if (markType === 'line') {
        numberColorsToSkip = colNumbers.bar;
      } else if (markType === 'highlight') {
        numberColorsToSkip = colNumbers.bar + colNumbers.line;
      }

      if (colNumbers[markType] === 0 || !validationResult[markType]) {
        config.missingYColOrInvalid = true || !config.rows.length;
      } else {
        lib.filterDateNulls(config);
        config.missingYColOrInvalid = !config.rows.length;
        if (config.missingYColOrInvalid) {
          return;
        }

        const idxLastColor = numberColorsToSkip % originalColorsLength;

        if (markType !== 'bar') {
          lib.colorScale(arc.misc.dimensionColorScale());
        }

        lib.colors(originalColors.slice(idxLastColor)
                                 .concat(originalColors.slice(0, idxLastColor)));
        plugin.setup_colors(config);
        combinedCols[markType].forEach(processColumns);
      }
    }, true);

    // Reset colors to original after calculating colors for line/highlights
    lib.colors(originalColors);

    let newTrellis;
    let max_slots;
    let unique_dims;

    // Only make bar dynamic
    barPlugin.makeDimDynamic(true);
    runAllPlugins((plugin, config) => {
      // Set up trellis
      config.viz_settings = {
        legend_contextmenu_shown: false,
        show_cb: [],
        hide_cb: [],
      };

      if (newTrellis) {
        config.trellis = newTrellis;
      } else {
        trellis = plugin.setupTrellis(viz_div, config, config.viz_settings);
        if (trellis.error_state()) {
          return undefined;
        }

        newTrellis = trellis;
        config.trellis = trellis;
        config.trellis.draw_skeleton(() => {
          plugin.viz_drawVisual(viz_div);
        });
      }

      plugin.get_unique_dims(config, config.rows);
      plugin.organize_cells(config, config.rows);
      plugin.compute_dim_maxvalues(config, config.rows);

      // Want to get the maximum max_slots values among all 3 visual types
      // Also want all to have same unique_dims because that's used to get dim
      // for tooltips
      if (!unique_dims || config.unique_dims.length > unique_dims.length) {
        max_slots = config.max_slots;
        unique_dims = config.unique_dims;
      }

      return plugin;
    });

    // TODO: At some point allow colors either as aggregates or dimensions
    // if (barCols.length > 1) {
    //   barPlugin.mark_type('groupedbar');
    // }

    function adjustDimSz(scale, data, dimSz, minBarSz) {
      for (let i = 1; i < data.length; i++) {
        const dist = Math.abs(scale(data[i].datecol) - scale(data[i - 1].datecol));
        if (dist < dimSz) {
          if (dist < minBarSz) {
            dimSz = minBarSz;
          } else {
            dimSz = dist;
          }
        }
      }

      return dimSz;
    }

    runAllPlugins((plugin, config) => {
      // Set max_slots the same for all visual types.
      // However, recalculate dim_sz because that's dependent on visual type
      config.max_slots = max_slots;
      config.unique_dims = unique_dims;
      plugin._compute_dim_sz(config, max_slots);

      config.has_secondary_agg_cols = y_secondary_cols.length > 0;
      config.color_secondary_agg_cols = y_secondary_cols;
      config.coldisplayformat = y_cols[0].coldisplayformat();
      config.primary_y_col = y_cols[0];
      config.secondary_y_col = y_secondary_cols[0] || null;
      config.secondary_coldisplayformat = y_secondary_cols[0] &&
          y_secondary_cols[0].coldisplayformat();
      config.y_all_indices = y_cols.map((col) => col.colindex());
      config.y_second_all_indices = y_secondary_cols.map((col) => col.colindex());
      plugin.compute_cell_size(config, viz_div);
    });

    const scales = combinedScales(validationResult);
    let combined_visual_config;
    runAllPlugins((plugin, config) => {
      // Set same width for all visuals
      config.width = scales.width;

      // Need to make distinct configs for each plugin, because labels get mutated
      combined_visual_config = {
        y_labels: $.extend(true, [], y_labels),
        y_secondary_labels: $.extend(true, [], y_secondary_labels),
        bar_padding_px: scales.bar_padding_px,
        bar_padding: scales.bar_padding,
      };

      config.agg_scale_secondary = scales.y_scale_secondary;
      plugin.x_scale(scales.x_scale);
      plugin.y_scale(scales.y_scale);
      config.x_scale = scales.x_scale;
      config.y_scale = scales.y_scale;
      config.dim_scale = scales.x_scale;
      config.agg_scale = scales.y_scale;
      if (config.dim_is_date) {
        config.dim_sz = adjustDimSz(config.dim_scale, config.rows, config.dim_sz, config.minBarSz);
      }

      plugin.combined_visual(combined_visual_config);
      config.combined_visual = combined_visual_config;

      // Draw visual
      // TODO: Currently not using xy_rows because it's incorrect
      //       and trellising not supported anyways
      config.trellis.draw_visuals((cell_sel, xy_rows) => {
        // Delete duplicate axis/axis label
        $chart_div.find('.arc-viz-axis').remove();
        $chart_div.find('text.arc-viz-axis-label').remove();
        if (!config.rows.length) {
          return combined;
        }

        plugin.compute_axes(config);
        plugin.draw_base_visual(config, cell_sel, config.rows, config.viz_settings);
      });
    });

    return combined;

    // HELPER FUNCTIONS

    function processColumns(col) {
      const colSource = col.colsource();

      let config = barConfig;
      if (colSource === 'line_measure') {
        config = lineConfig;
      } else if (colSource === 'highlighter_measure') {
        config = highlightConfig;
      }

      if (colSource in ycol_idxs) {
        ycol_idxs[colSource].push(col.colindex());
      } else {
        ycol_idxs[colSource] = [col.colindex()];
      }

      const c = config.color(col.colname());

      const colLabel = lib.Label(col.colname(), col.coldesc(), c);
      if (col.secondaryaxis()) {
        y_secondary_labels.push(colLabel);
        y_secondary_cols.push(col);
      } else {
        y_primary_axis[col.colindex()] = true;
        y_labels.push(colLabel);
        y_cols.push(col);
      }
    }

    function combinedScales(validationOutcome) {
      let plugin;
      let config;
      if (!barConfig.missingYColOrInvalid && validationOutcome.bar) {
        plugin = barPlugin;
        config = barConfig;
      } else if (!lineConfig.missingYColOrInvalid && validationOutcome.line) {
        plugin = linePlugin;
        config = lineConfig;
      } else {
        plugin = highlightPlugin;
        config = highlightConfig;
      }

      const combinedPrimaryRows = [];
      const combinedSecondaryRows = [];
      const dimCol = lib.shelf_columns(config.columns, 'x_shelf');
      const dimIdx = dimCol[0].colindex();
      const xIdx = 0;
      const yIdx = 1;
      const xRange = [0, config.width];
      const yRange = [config.height, 0];
      const xScaleConfig = Object.create(null);
      const yScaleConfig = Object.create(null);
      let yScaleSecondaryConfig = Object.create(null);

      function isPrimaryAxis(colIdx) {
        return y_primary_axis[colIdx];
      }

      // Creating combined scale from original set of rows mean
      // we use maximum dimensions. Howeer, have to check for nulls again
      const yAxesExtrema = {
        primary: {
          min: 0,
          max: 0,
        },
        secondary: {
          min: 0,
          max: 0,
        },
      };
      let isStackedBar = false;
      data_result.rows().forEach((r) => {
        const dimValue = r[dimIdx];

        // If dimension is date and not number, no need to addDimFilter
        if (config.dim_is_date && arc.misc.check_if_value_null(dimValue)) {
          return;
        }

        $.each(ycol_idxs, (k, v) => {
          const isArea = useArea.isChecked && k === 'line_measure';
          const isPrimary = isPrimaryAxis(v[0]);
          const extrema = yAxesExtrema[isPrimary ? 'primary' : 'secondary'];
          if ((k === 'bar_measure' || isArea) && (v.length > 1)) {
            isStackedBar = true;
            let maxpos = 0;

            // TODO: we are currently ignoring negative values here
            let maxneg = 0;

            let rowTotals = 0;
            v.forEach((colIdx) => {
              const x = Number(r[colIdx]);
              rowTotals += x;

              // If measure isn't number, no need to add
              if (arc.misc.check_if_value_null(x)) {
                return;
              }

              if (x > 0) {
                maxpos += x;
              } else {
                maxneg += x;
              }

              extrema.max = Math.max(rowTotals, extrema.max);
              extrema.min = Math.min(rowTotals, extrema.min);
            });

            if (isPrimary) {
              combinedPrimaryRows.push([dimValue, maxpos]);
            } else {
              combinedSecondaryRows.push([dimValue, maxpos]);
            }
          } else {
            v.forEach((colIdx) => {
              extrema.max = Math.max(Number(r[colIdx]), extrema.max);
              extrema.min = Math.min(Number(r[colIdx]), extrema.min);
              if (isPrimaryAxis(colIdx)) {
                combinedPrimaryRows.push([dimValue, Number(r[colIdx])]);
              } else {
                combinedSecondaryRows.push([dimValue, Number(r[colIdx])]);
              }
            });
          }
        });
      });

      const padCalc = plugin._computeDynamicDimPadding(config);

      xScaleConfig.rows = combinedPrimaryRows;
      xScaleConfig.idx = xIdx;
      xScaleConfig.range = xRange;
      xScaleConfig.is_date = config.dim_is_date;
      xScaleConfig.is_datetime = config.dim_is_datetime;
      xScaleConfig.date_format = config.date_format;
      xScaleConfig.is_dim = true;

      xScaleConfig.width_for_each_bar = Math.min(padCalc.width_for_each_bar,
          padCalc.max_bar_width);
      xScaleConfig.bar_padding =
          (padCalc.width_for_each_bar > padCalc.max_bar_width) ?
              padCalc.space_percent_of_width :
              false;

      yScaleConfig.rows = combinedPrimaryRows;
      yScaleConfig.idx = yIdx;
      yScaleConfig.range = yRange;
      yScaleConfig.is_log10scale = config.agg_is_log10;
      yScaleConfig.agg_scale_zero = user_settings.find_by_id('scale_agg_zero').isChecked;
      yScaleConfig.minScale = plugin.getAggScaleMin;
      yScaleConfig.maxScale = plugin.getAggScaleMax;

      lineConfig.barDimSz = barConfig.dim_sz ? barConfig.dim_sz : xScaleConfig.width_for_each_bar;
      highlightConfig.barDimSz = lineConfig.barDimSz;

      yScaleSecondaryConfig = $.extend(Object.create(null), yScaleConfig);
      yScaleSecondaryConfig.rows = combinedSecondaryRows;
      yScaleSecondaryConfig.isSecondaryAxis = true;

      if (isStackedBar) {
        yScaleConfig.domain = [yAxesExtrema.primary.min, yAxesExtrema.primary.max];
        yScaleSecondaryConfig.domain = [yAxesExtrema.secondary.min, yAxesExtrema.secondary.max];
      }

      const x_scale = lib.idx_to_scale_config(xScaleConfig);
      const y_scale = lib.idx_to_scale_config(yScaleConfig);
      const y_scale_secondary = lib.idx_to_scale_config(yScaleSecondaryConfig);

      return {
        x_scale,
        y_scale,
        y_scale_secondary,
        combinedPrimaryRows,
        combinedSecondaryRows,
        bar_padding_px: (padCalc.width_for_each_bar - padCalc.max_bar_width) / 2,
        bar_padding: xScaleConfig.bar_padding,
        width: config.width,
      };
    }
  };

  return combined;

  // HELPER FUNCTIONS

  // Since there's multiple visual config/plugins being used (bar/line/highlight)
  // this prevents duplication of code. Takes a function and runs for all plugins
  function runAllPlugins(functionToRun, ignoreMissingCol) {
    let lineResult;
    let barResult;
    let highlightResult;
    if (!barConfig.missingYColOrInvalid || ignoreMissingCol) {
      barResult = functionToRun(barPlugin, barConfig);
    }

    if (!lineConfig.missingYColOrInvalid || ignoreMissingCol) {
      lineResult = functionToRun(linePlugin, lineConfig);
    }

    if (!highlightConfig.missingYColOrInvalid || ignoreMissingCol) {
      highlightResult = functionToRun(highlightPlugin, highlightConfig);
    }

    return {
      bar: barResult,
      line: lineResult,
      highlight: highlightResult,
    };
  }

  function makeSettingHidden(settingName) {
    const setting = user_settings.find_by_id(settingName);
    if (setting) {
      setting.hidden(true);
    }
  }
};

arc.visual.plugins.combined_extension = function() {
  const ext = function() { return 'arc.visual.plugins.combined_extension'; };

  ext.viz_name = function() {
    return 'combo';
  };

  ext.viz_tooltip = function() {
    return 'combined bar/line';
  };

  ext.viz_background = function() {
    return 'url(../../static/img/chart-icons.png?v=6) top no-repeat';
  };

  ext.viz_background_position = function() {
    return '0 0';
  };

  ext.viz_img_large_src = function() {
    return '../static/img/icon_combo.png?v=1';
  };

  ext.viz = arc.visual.plugins.combined;

  return ext;
};

arc.visual.extensions.push(arc.visual.plugins.combined_extension());
}());
