window.arc = window.arc || {};

(function() {
'use strict';

/*
 * In order to create parameters using an 'extension' widget, use an html element
 * that includes the 'arc-parameter' class, and use the 'arc-parameter' data
 * element to setup parameters.
 *
 * Look at the example html code below for an html widget that sends out
 * parameters

   <a class='arc-parameter' data-arc-parameter='{"state":"California", "abc.ignore": true}'>
    California ($2.2B)
  </a>,
   <a class='arc-parameter' data-arc-parameter='{"state":"Texas", "abc": "ABC"}'>
    Texas ($1.6B)
  </a> and
   <a class='arc-parameter' data-arc-parameter='{"state":"New York"}'>New York ($1.3B)</a>
   are the three US states with a GSP (Gross State Product - the GDP equivalent for states) > $1B.

   The three US states with the lowest GSP are
   <a class='arc-parameter' data-arc-parameter='{"state":"Vermont"}'>Vermont ($30.7M)</a>,
   <a class='arc-parameter' data-arc-parameter='{"state":"Montana"}'>Montana ($45.8M)</a> and
   <a class='arc-parameter' data-arc-parameter='{"state":"Rhode Island"}'>Rhode Island ($45.9M)</a>

 *
 * Notice how Texas sends out multiple parameters, and how California deletes
 * one of those additional parameters, but the other states do not.
 */

arc.visual = arc.visual || {};
arc.visual.user_settings = arc.visual.user_settings || {};
arc.visual.plugins = arc.visual.plugins || {};
arc.visual.extensions = arc.visual.extensions || [];

arc.visual.plugins.extension = function() {
  let saved_settings = {};
  // eslint-disable-next-line no-unused-vars
  let data_result;
  let lib;
  // eslint-disable-next-line no-unused-vars
  let shelves_div;
  let html_contents = '';
  let $editor;
  let $applyButton;

  const html = arc.visual.api();

  // Remove 'Visual Axes and Labels' and 'Tooltips' options
  // in Style section of visual editor
  const userSettings = arc.visual.defaultSettings({
                                   noAxes: true,
                                   noTrellis: true,
                                   noTooltip: true,
                                 })
                                 .removeSetting([
                                   'visual_show_detail_data',
                                 ]);

  const getIsEditorShown = () => arc.globals.is_edit && lib.inModal();

  const setUpVisual = (vizDiv, htmlToDisplay) => {
    const containerId = html.container_id();

    const substitutedHTML = lib.substituteParams(
      htmlToDisplay,
      [htmlToDisplay, `<<arc_container_id:${containerId}>>`],
      true
    );

    // Extract HTML
    const htmlcontent = $.parseHTML(`<div>\n${substitutedHTML}\n</div>`);

    // Extract Scripts
    const content = $.parseHTML(`<div>\n${substitutedHTML}\n</div>`, null, /* keepScripts */ true);
    const scripts = $(content).find('script').map(function () {
      return $(this).text();
    }).get();
    const masterscript = `// ${arc.frontend.globals.product.name} Extension \n\n${scripts.join('\n')}`;

    // Setup arcapi
    const arcapi = {};
    arcapi.chartId = function() {
      return vizDiv;
    };

    arcapi.container_id = html.container_id;
    arcapi.set_state = lib.set_viz_state;
    arcapi.get_state = lib.get_viz_state;
    arcapi.set_dashboard_params = lib.dashboard_click;
    arcapi.hide_border = html.hide_border;
    arcapi.hide_title = html.hide_title;

    // Create a function bound with arcapi set
    // eslint-disable-next-line no-new-func
    const masterfunc = Function('arcapi', masterscript).bind(undefined, arcapi);
    const $vizDiv = $(`#${vizDiv}`);

    // Materialize the HTML
    $vizDiv
      .append($(htmlcontent)
      .attr('id', containerId));

    // Hide the "Save CSV" button
    $vizDiv
      .closest('.gridster .gs-w')
      .find('a.save-csv')
      .remove();

    // Run the scripts
    if (masterscript) {
      window.addEventListener('load', masterfunc);
    }

    setupLinks();
  };

  const setUpEditor = (viz_div) => {
    const $viz = $(`#${viz_div}`);

    const previewContainerTemplate = '<div id="viz-html-preview" data-test-id="viz-html-preview"></div>';

    const $editorContainer = $(`
    <div id="viz-html-editor-container">
      <div>
        <div id="viz-html-editor-autocomplete" data-test-id="viz-html-editor-autocomplete"></div>
        <div id="viz-html-editor"></div>
      </div>
      <div>
        <button id="refresh-html-preview" type="button" class="btn btn-primary btn-sm" data-test-id="refresh-html-preview">
          <span class="fa fa-refresh"></span> Refresh preview
        </button>
        ${previewContainerTemplate}
      </div>
    </div
    `);

    $viz.append($editorContainer);

    const currContent = $editor ? $editor.value() : html_contents;

    $editor = arc.utils.autocompleteEditor(
      'viz-html-editor',
      'html',
      {
        autocompleteOnOffButton: $('#viz-html-editor-autocomplete'),
        autocompleteType: 'html',
        aceOptions: {
          showPrintMargin: false,
        }
      }
    );

    $editor.value(currContent);

    $applyButton = $editorContainer
      .closest('.super-modal')
      .find('#apply-button');

    $applyButton.offOn('click.applyChanges', () => {
      const newValue = $editor.value();

      if (newValue !== html_contents) {
        html_contents = newValue;
        lib.page_needs_save();
      }
    });

    const $refreshButton = $('#refresh-html-preview');
    setUpVisual('viz-html-preview', $editor.value());

    $refreshButton.offOn('click', () => {
      $('#viz-html-preview').replaceWith(previewContainerTemplate);

      setUpVisual('viz-html-preview', $editor.value());
    });

    $editor.focusEditor();
  };

  html.viz_lib = function(_) {
    if (!arguments.length) return lib;
    lib = _;
    lib.shelf_data([])
       .user_settings(userSettings)
       .refresh_visible(false);
    return html;
  };

  html.viz_getSettings = function() {
    const settings = lib.save_basics(saved_settings);

    settings.viz_type = settings.viz_type || {};
    settings.viz_type.html = settings.viz_type.html || {};
    settings.viz_type.html.html_contents = getIsEditorShown() ? $editor.value() : html_contents;

    saved_settings = settings;

    return settings;
  };

  html.viz_setSettings = function(settings, switchingVisualType) {
    if (settings === undefined) {
      return html;
    }

    saved_settings = settings;
    lib.restore_basics(settings, switchingVisualType);

    const hasContents = settings.viz_type &&
      settings.viz_type.html &&
      settings.viz_type.html.html_contents;

    if (hasContents) {
      html_contents = settings.viz_type.html.html_contents;
    }

    return html;
  };

  html.viz_drawDesigner = function(_shelves_div) {
    const hasManageCustomStylesPerm = arc.globals.perms.sysperms.includes('sys_styles');
    shelves_div = _shelves_div;
    lib.shelves_dom_id(_shelves_div)
       .base_setup();

    const $shelves = $('#app-shelves');
    if (hasManageCustomStylesPerm) {
      $shelves.html(arc.misc.vizHTMLInfo());
    } else {
      $shelves.html(arc.misc.vizHTMLInfoNoPermission());
    }

    $shelves
      .find('.viz-html-view')
      .offOn('click.fullscreen', () => {
        const widgetId = lib
          .report()
          .reportWidgetId();

          $(`#${widgetId} .viz-modal-link`).trigger('click');
      });

    return html;
  };

  html.viz_getDataRequest = function() {
    arc.counters.increment('dataapi_start');
    return arc.data.request()
                   .request_type('ignore');
  };

  /*
   * Set the data this object should use to draw out the visual
   */
  html.viz_setDataResult = function(result) {
    arc.counters.increment('dataapi_end');
    data_result = result;
    return html;
  };

  html.viz_drawVisual = function(viz_div) {
    lib.chart_div(viz_div);
    lib.undraw();

    if (getIsEditorShown()) {
      setUpEditor(viz_div);
    } else {
      $editor = null;
      setUpVisual(viz_div, html_contents);
    }

    return html;
  };

  function setupLinks() {
    var $viz = $("#" + lib.chart_div());
    var parameters = $viz.find(".arc-parameter");
    parameters.off("click.arc-parameter")
              .on("click.arc-parameter", function() {
                var params = {};

                var p = $(this).data("arc-parameter");
                if (p !== undefined) {
                  params = p;
                }

                if (Object.keys(params).length > 0) {
                  lib.dashboard_click(params);
                }
              });
  }

  html.container_id = function() {
    return "viz-html-uid-" + lib.unique_report_id();
  };

  html.hide_border = function () {
    //var container_id = "viz-html-uid-" + lib.unique_report_id();
    var container_id = html.container_id();
    $("#" + container_id).closest(".gridster .gs-w").attr("style", "border: solid 0px #ddd");
    $("#" + container_id).closest(".gridster .gs-w").removeClass("gridster-border");
  };

  html.hide_title = function () {
    //var container_id = "viz-html-uid-" + lib.unique_report_id();
    var container_id = html.container_id();
    $("#" + container_id).closest(".gridster .gs-w").find(".app-title").hide();
  };

  html.viz_destroy = function() {
    const vizDiv = lib.chart_div();
    const $viz = $(`#${vizDiv}`);
    $viz.find('.arc-parameter').off('click.arc-parameter');
    $viz.find('.viz-html-editor').remove();

    if ($applyButton) {
      $applyButton.off('click.applyChanges');
      $applyButton = undefined;
    }

    if ($editor) {
      $editor.destroy();
    }
  };

  return html;
};

arc.visual.plugins.extension_extension = function() {
  var ext = function() { return "arc.visual.plugins.extension_extension"; };

  ext.viz_name = function() {
    return "extension";
  };

  ext.viz_tooltip = function() {
    return "extension";
  };

  ext.viz_background = function() {
    return "url(../../static/img/chart-icons.png?v=6) top no-repeat";
  };

  ext.viz_background_position = function() {
    return "0 0";
  };

  ext.viz_img_large_src = function() {
    return "../static/img/icon_extension.png?v=1";
  };

  ext.viz = arc.visual.plugins.extension;

  return ext;
};

arc.visual.extensions.push(arc.visual.plugins.extension_extension());
}());
