import L from 'leaflet';
import 'leaflet.heat';
import 'leaflet.markercluster';
import 'leaflet.gridlayer.googlemutant';
import mapboxgl from 'mapbox-gl';
import 'mapbox-gl/dist/mapbox-gl.css';

window.arc = window.arc || {};

(function() {
'use strict';

arc.visual = arc.visual || {};
arc.visual.plugins = arc.visual.plugins || {};
arc.visual.extensions = arc.visual.extensions || [];
arc.testutil = arc.testutil || {};

arc.visual.plugins.leaflet = function() {
  var leaflet = arc.visual.api();
  var saved_settings = {};
  var supported_types = 'latitude/longitude, or ZIP Code';
  var baseGoogleURL = 'https://maps.google.com/maps/api/js?';
  var allArcTooltips = [];
  var mapArtifacts;

  var mapType,
      dataResult,
      lib,
      map,
      localServerAddress,
      googleLayer,
      zipcodeLatLong,
      google_api_id,
      google_client_id,
      mapboxAccessToken,
      defaultStartLocation = [-98.5794797, 39.8283459], //geographic center of the US
      defaultStartZoom = 2,
      maxVisibleZoom = 23,
      colorRange = [undefined, undefined],
      colorDomain = {},
      colorIsDim = false,
      colorIsNumeric = false,
      colorCol,
      aggRange = [undefined, undefined],
      aggCol,
      choroplethFilter = null,
      choroplethRenderedFeatures = {},
      mapPreloaded = false,
      noRedraw = false,
      fullRedraw = false,
      $vizDiv,
      preserveMap,
      mapDiv,
      customIconId = 'arcadia-custom-icon',
      routeSourceIds = {},
      autoUpdateLocation = true,
      choroplethDataHash = {};

  /**
   * Info about  noRedraw/fullRedraw/preserveMap/mapPreloaded (all
   * noRedraw: If true indicates all redraw handled by setting changeCB. Don't run through
   * fullRedraw: Need to redraw everything including the base map
   * mapPreloaded: indicates base map was not thrown away
   * preserveMap: used in custom map redraw to indicate to both preserve the base map as well as
   *    important data such as color range and domain.
   */

  const sourceIdObject = {
      main: 'arcadia-main-map-geoJSON',
      route: 'arcadia-routes-geoJSON',
      startMarks: 'arcadia-start-marks-geoJSON',
      midMarks: 'arcadia-mid-marks-geoJSON',
      endMarks: 'arcadia-end-marks-geoJSON',
      cluster: 'arcadia-cluster-geoJSON',
    };

  const layerIdObject = {
    circle: 'arcadia-circles',
    pins: 'arcadia-pins',
    heatmap: 'arcadia-heatmap',
    routes: 'arcadia-routes',
    cluster: 'arcadia-clusters',
    clusterText: 'arcadia-cluster-text',
    clusterSingle: 'arcadia-unclustered',
    startMarks: 'arcadia-startMarks',
    midMarks: 'arcadia-midMarks',
    endMarks: 'arcadia-endMarks',
  };

  const defaultMapboxToken = '';

  if (!arc.testutil.leafletManualLoad) {
    google_api_id = arc.globals.app_settings.google_api_id;
    google_client_id = arc.globals.app_settings.google_client_id;
    mapboxAccessToken = arc.globals.app_settings.mapbox_access_token;
    localServerAddress = arc.globals.app_settings.local_mapbox_server;
  }

  const showMapData = arc.settings.mapInteractive.showMapData();

  showMapData.addChangeCb('updateMapInfoDiv', function() {
    if (showMapData.value() === 'true') {
      createMapInfoDiv();
    } else {
      removeMapInfoDiv();
    }

    noRedraw = true;
  });

  //Map style Options
  const mapServer = arc.settings.mapInteractive.mapServer();
  const mapLocalKey = arc.settings.mapInteractive.localKey();
  const mapServerAddress = arc.settings.mapInteractive.visualAtlasServer();
  const mapStyles = arc.settings.mapInteractive.mapStyles();
  const mapboxStyle = arc.settings.mapInteractive.mapboxStyle();

  const completeRedrawSettings = [mapServer, mapLocalKey, mapStyles, mapboxStyle, mapServerAddress];
  completeRedrawSettings.forEach(setting => setting.addChangeCb('redraw', () => fullRedraw = true));

  //Heatmap Layer Settings
  const heatmapEnabled = arc.settings.mapInteractive.heatmapEnabled();
  const heatmapLegend = arc.settings.mapInteractive.heatmapLegend();
  const intensityMag = arc.settings.mapInteractive.intensityMagnification();
  const heatmapRadius = arc.settings.mapInteractive.heatmapRadius();
  const heatmapVisibleZoom = arc.settings.mapInteractive.heatmapVisibleZoom();
  heatmapEnabled.addDependents([heatmapLegend, intensityMag, heatmapRadius]);

  //Cluster Options
  const cluster_enabled = arc.settings.mapInteractive.clusterEnabled();
  const cluster_color = arc.settings.mapInteractive.clusterColor();
  const cluster_radius =  arc.settings.mapInteractive.clusterRadius();
  cluster_enabled.addDependents([cluster_color, cluster_radius]);

  //Circle Layer Options
  const circle_enabled = arc.settings.mapInteractive.circleEnabled();
  const circleLegendArea = arc.settings.mapInteractive.circleAreaLegend();
  const circleLegendColor = arc.settings.mapInteractive.colorLegend();
  const circleRadiusRange = arc.settings.mapInteractive.circleRadiusRange();
  const circleVisibleZoom = arc.settings.mapInteractive.circleVisibleZoom();
  circle_enabled.addDependents([circleLegendArea, circleLegendColor, circleRadiusRange]);

  circleRadiusRange.addChangeCb('adjustSetting', function() {
    if (mapServer.value() !== 'mapboxgl' || !map || !map.getLayer('arcadia-circles')) {
      return;
    }

    noRedraw = true;
    const circleRadiusProperty = createRadiusProperty();
    map.setPaintProperty('arcadia-circles', 'circle-radius', circleRadiusProperty);

    if (circleLegendArea.value() === 'true') {
      const zoomRange = circleVisibleZoom.get_range();
      removeGLSizeLegend();
      createGLSizeLegend(undefined, zoomRange[0], zoomRange[1]);
    }
  });

   //Marker Options
  const markerType = arc.settings.mapInteractive.markerType();
  const customIcon = arc.settings.mapInteractive.customIcon();
  const startMarkers = arc.settings.mapInteractive.startMarker();
  const middleMarkers = arc.settings.mapInteractive.middleMarker();
  const endMarkers = arc.settings.mapInteractive.endMarker();
  const routesEnabled = arc.settings.mapInteractive.routesEnabled();
  const markersEnabled = arc.settings.mapInteractive.markersEnabled();
  const markersPermanentLabels = arc.settings.mapInteractive.markerPermanentLabels();
  const customIconSize = arc.settings.mapInteractive.customIconSize();

  customIcon.addChangeCb('changeCustomIcon', function() {
    if (customIcon.value() === '' || mapServer.value() !== 'mapboxgl') {
      return;
    }

    if (map.hasImage(customIconId)) {
      map.removeImage(customIconId);
    }

    map.loadImage(customIcon.value(), function(error, image) {
          if (error) {
            lib.alertMessage('Error: custom icon could not load');
            return;
          }

          if (!map.hasImage(customIconId)) {
            map.addImage(customIconId, image);
          }
        });

  });

  //Choropleth Layer
  const choroplethLayerId = arc.settings.mapInteractive.choroplethLayerId();
  const choroplethOpacity = arc.settings.mapInteractive.choroplethOpacity();
  const includeChoroplethLegend = arc.settings.mapInteractive.includeChoroplethLegend();
  const startZoom = arc.settings.mapInteractive.startZoom();
  const startCenter = arc.settings.mapInteractive.startCenter();
  const rememberMapView = arc.settings.mapInteractive.saveZoomAndLocation();

  includeChoroplethLegend.addChangeCb('toggleLegend', function() {
    noRedraw = true;
    if (mapServer.value() !== 'mapboxgl' || mapType !== 'custom' || !colorCol) {
      return;
    }

    if (includeChoroplethLegend.value() === 'true') {
      createGLColorLegend(undefined, 0, 22);
    } else {
      removeGLColorLegend();
    }
  });

  choroplethOpacity.addChangeCb('adjustChoroplethOpacity', function() {
    noRedraw = true;
    if (mapServer.value() !== 'mapboxgl' || mapType !== 'custom') {
      return;
    }

    updateChoroplethOpacity();
  });

  const mapboxGLSettings = [choroplethLayerId, choroplethOpacity, includeChoroplethLegend,
    mapboxStyle, heatmapVisibleZoom, circleVisibleZoom, showMapData,
  ];

  mapServer.addDependents(mapboxGLSettings, ['google', 'mapbox'], ['mapboxgl']);

  const holdLocationManualMove = arc.settings.mapInteractive.holdLocationOnManualMove();
  const user_settings = arc.visual.defaultSettings({ noTrellis: true })
                                .addSetting([arc.colorpalette.setting(),
                                                arc.colormap.measureRangesSetting(),
                                                showMapData,
                                                mapServer,
                                                mapLocalKey,
                                                mapServerAddress,
                                                mapStyles,
                                                mapboxStyle,
                                                choroplethLayerId,
                                                choroplethOpacity,
                                                includeChoroplethLegend,
                                                rememberMapView,
                                                startCenter,
                                                startZoom,
                                                heatmapEnabled,
                                                intensityMag,
                                                heatmapRadius,
                                                heatmapVisibleZoom,
                                                heatmapLegend,
                                                cluster_enabled,
                                                cluster_color,
                                                cluster_radius,
                                                circle_enabled,
                                                circleLegendArea,
                                                circleLegendColor,
                                                circleRadiusRange,
                                                circleVisibleZoom,
                                                markersEnabled,
                                                markersPermanentLabels,
                                                markerType,
                                                routesEnabled,
                                                startMarkers,
                                                middleMarkers,
                                                endMarkers,
                                                customIcon,
                                                customIconSize,
                                                holdLocationManualMove,
                                              ])
                                  .removeSetting([
                                    'axis_label_font_size',
                                    'axis_ticks_font_size',
                                    'label_font_size',
                                    'label_color',
                                  ]);

  arc.addColorSettings(user_settings, {
    includeMeasureRangeMapSetting: true,
  });

  var shelf_data = [
    {
      name: 'geo_shelf',
      display_name: 'Geo',
      shelf_type: arc.report.shelf_types.COLUMN,
      column_type: arc.report.columnTypeEnum.UNKNOWN,
      geoShelf: true,
    },
    {
      name: 'aggregates_shelf',
      display_name: 'Measure',
      shelf_type: arc.report.shelf_types.COLUMN,
      column_type: arc.report.columnTypeEnum.AGGREGATE,
    },
    {
      name: 'color_shelf',
      display_name: 'Colors',
      shelf_type: arc.report.shelf_types.COLUMN,
      column_type: arc.report.columnTypeEnum.UNKNOWN,
    },
    {
      name: 'tooltip_shelf',
      display_name: 'Tooltips',
      shelf_type: arc.report.shelf_types.COLUMN,
      column_type: arc.report.columnTypeEnum.UNKNOWN,
    },
    {
      name: 'filters_shelf',
      display_name: 'Filters',
      display_html: arc.report.filters_shelf_html(),
      shelf_type: arc.report.shelf_types.PREDICATE,
      column_type: arc.report.columnTypeEnum.UNKNOWN,
      onShelfListChange: arc.report.updateFilterShelfLabel,
    },
  ];

  var constraints = {
    maximums: {
      aggregates_shelf: 1,
      color_shelf: 1,
    },
  };

  leaflet.viz_lib = function(_) {
    if (!arguments.length) return lib;
    lib = _;
    lib.shelf_data(shelf_data)
       .user_settings(user_settings);
    return leaflet;
  };

  leaflet.viz_getSettings = function () {
    var settings = saved_settings;
    settings = lib.save_basics(settings);
    saved_settings = settings;
    return settings;
  };

  leaflet.viz_setSettings = function (settings, switchingVisualType, siteSettings) {
    if (settings === undefined) {
      return leaflet;
    }

    // Only needed for unit tests
    if (siteSettings) {
      siteSettings.forEach(function(s) {
        switch (s.setting_key) {
          case 'google_api_id': {
            google_api_id = s.setting_value;
            break;
          }

          case 'google_client_id': {
            google_client_id = s.setting_value;
            break;
          }

          case 'mapbox_access_token': {
            mapboxAccessToken = s.setting_value;
            break;
          }

          case 'local_mapbox_server': {
            if (s.setting_value !== '') {
              localServerAddress = s.setting_value;
            }
          }
        }
      });
    }

    saved_settings = settings;
    lib.restore_basics(settings, switchingVisualType);

    return leaflet;
  };

  leaflet.viz_drawDesigner = function(shelves_div) {
    //This method is called by report driver to set-up our shelves
    lib.shelves_dom_id(shelves_div)
       .base_setup();

    // Define messages to be included in shelves
    const geoMessages = [
      'Drag <b>' + supported_types +
      '</b>  dimension(s) or measure(s) field to this shelf ' +
        'to designate location, then assign a Geo type to each field. For ' +
        'ZIP codes, only five digit US ZIP codes are supported.',
      lib.createPopoverMessages('multipleGeoConfigs'),
    ];

    const aggregateMessages = [
      'Assign the <b>measure</b> to be plotted on the map.',
      'This measure is used to compute the heatmap intensity, the value shown in ' +
      'cluster aggregation, and the size of circle markers.',
      'If no measures are specified, then a value of 1 will be used.',
    ];

    const colorMessages = [
      'Add a <b>dimension</b> or <b>measure</b> field from the Data menu to specify the color of ' +
      'the circle layer, pins, and routes by common identifier. This shelf accepts a single field.',
    ];

    lib.addShelfDescription(constraints, {
      geo_shelf: geoMessages,
      aggregates_shelf: aggregateMessages,
      color_shelf: colorMessages,
    });
    return leaflet;
  };

  leaflet.viz_getDataRequest = function() {
    let geoColObject = {};

    lib.shelfElementCb('geo_shelf', function(shelf_element, i) {
      var cast = shelf_element.geo_cast();
      if (geoColObject[cast] === undefined) {
        geoColObject[cast] = true;
      }
    });

    mapType = arc.geo.computeMapType(undefined, mapServer.value() === 'mapboxgl', geoColObject);

    if (mapType === 'error') {
      lib.alertMessage(`Error: ${arc.geo.typeError}`);
      return null;
    }

    lib.check_shelf_contents(constraints); // throw an error but still run query

    let dataRequestConfig = {};
    if (mapType === 'custom') {
      let currMap = mapArtifacts;
      let inModal = arc.globals.isInModal();
      if (currMap && currMap.modalState !== inModal) {
        choroplethFilter = null;
      } else if (choroplethFilter) {
        dataRequestConfig.incrementalFilter = choroplethFilter;
      }
    }

    let dataRequest = lib.shelvesToDataRequest(undefined, dataRequestConfig);
    if (mapType === 'none') {
      dataRequest.request_type('ignore');
    }

    return dataRequest;
  };

  leaflet.viz_setDataResult = function(result) {
    dataResult = result;
    return leaflet;
  };

  leaflet.viz_fetchCollateral = function(viz_div, isWindowResize, drawCb) {

    const geoColumns = lib.shelf_columns(dataResult.columns(), 'geo_shelf');
    mapType = arc.geo.computeMapType(geoColumns, true);

    if (mapType === 'error') {
      lib.alertMessage(`Error: ${arc.geo.typeError}`);
    }

    //Fetch ZIP code data here if we need to, but only do it once
    if (!leaflet.collateral_is_available()) {
      leaflet.fetch_collateral(function() {
        drawCb(viz_div, isWindowResize);
      });
    } else {
      drawCb(viz_div, isWindowResize);
    }

    return leaflet;
  };

  function resetInitialData() {
    colorRange = [undefined, undefined];
    choroplethDataHash = {};
    choroplethRenderedFeatures = {};
    colorDomain = {};
    colorCol = undefined;
    aggCol = undefined;
  }

  /**
   * Determins when base map should be preserved versus getting rid of it.
   * Preserving the base map is important for streaming cases and changing data in the base map
   * Should mostly keep the base map and change the layers on top of it.
   * @param currMap {object} the current saved map if there is one.
   * @param inModal {boolean}
   * @param isEdit {boolean}
   * @param isWindowResize {boolean}
   * @returns {boolean}
   */
  function deleteMap(currMap, inModal, isEdit, isWindowResize) {
    if (fullRedraw || (isWindowResize && mapServer.value() === 'mapboxgl') ||
      (currMap && currMap.modalState !== inModal)) {
      preserveMap = false;
      fullRedraw = false;
      return true;
    }

    if (preserveMap) {
      return false;
    }

    if (inModal && isEdit) {
      return true;
    }

    if (mapServer.value() !== 'mapboxgl' && (isWindowResize || (inModal && isEdit))) {
      return true;
    } else if (mapServer.value() !== 'mapboxgl') {
      return false;
    }

    if (mapServer.value() === 'mapboxgl' && mapType === 'custom') {
      return true;
    }

    return false;
  }

  leaflet.viz_drawVisual = function(viz_div, isWindowResize) {
    if (noRedraw) {
      noRedraw = false;
      return;
    }

    // report_id is undefined for visuals generated by search
    // Need to make sure the cached mapArtifact not referenced in that case
    if (!lib.report_id()) {
      mapArtifacts = undefined;
    }

    lib.chart_div(viz_div);
    $vizDiv = $(`#${viz_div}`);
    const inModal = arc.globals.isInModal();
    const isEdit = arc.globals.is_edit;
    let currMap = mapArtifacts;

    if (deleteMap(currMap, inModal, isEdit, isWindowResize)) {
      mapArtifacts = undefined;
    }

    if (!mapArtifacts || arc.testutil.testMode) {
      mapArtifacts = undefined;
      lib.undraw();
    }

    if (!(mapType === 'custom' && preserveMap)) {
      resetInitialData();
    }

    let config = {};
    config.vizDiv = viz_div;
    config.map_div = `map-div-${lib.report_id()}-${viz_div}`;
    mapDiv = config.map_div;

    d3.select(`#${viz_div}`)
      .selectAll(`#${config.map_div}`)
      .data([1])
      .enter()
        .append('div')
        .attr('style', 'height:inherit')
        .attr('id', config.map_div);

    config.d3_map_div = d3.select(`#${config.map_div}`);

    //instead of adding the rows to our config, lets generate new arrays in our
    //config object for easy passing of values straight into the leaflet plugins
    config.points = [];
    config.pointsForHeatMap = [];
    config.latLongPoints = [];
    config.popup_info = [];
    config.columns = dataResult.columns();
    config.latRange = [undefined, undefined];
    config.longRange = [undefined, undefined];
    config.routeColors = {};
    aggCol = lib.first_column(config.columns, 'aggregates_shelf');
    colorCol = lib.first_column(config.columns, 'color_shelf');
    config.latCol = config.columns.filter(col => col.colgeotype() === 'latitude');
    config.longCol = config.columns.filter(col => col.colgeotype() === 'longitude');
    config.pointGeoJSONData = { type: 'FeatureCollection', features: [] };
    config.startGeoJSON = { type: 'FeatureCollection', features: [] };
    config.midGeoJSON = { type: 'FeatureCollection', features: [] };
    config.endGeoJSON = { type: 'FeatureCollection', features: [] };
    config.routeGeoJSON = {};
    config.mapServer = mapServer.value();
    config.includeTooltips = user_settings.find_by_id('show_tooltip').value() === 'true';

    const illegalVals = {
      numIllegalGeo: 0,
      numIllegalAggs: 0,
      numIllegalColors: 0,
    };

    let noLatLong = ['none', 'custom'].includes(mapType);
    if (!colorCol && aggCol && mapType === 'custom') {
      colorCol = aggCol;
    }

    if (colorCol) {
      colorIsDim = colorCol.colproperty() === arc.report.columnTypeEnum.DIMENSION;
      colorIsNumeric = colorCol.isNumeric();
    }

    if (mapType === 'custom') {
      config.choroCol = lib.first_column(config.columns, 'geo_shelf');
    }

    if (mapLocalKey.value() !== '') {
      switch (mapServer.value()) {
        case 'google':
          google_api_id = mapLocalKey.value();
          break;
        case 'mapbox':
        case 'mapboxgl':
          mapboxAccessToken = mapLocalKey.value();
          break;
      }
    }

    if (mapboxAccessToken === '' && mapServer.value() === 'mapboxgl') {
      mapboxAccessToken = defaultMapboxToken;
    }

    let alert_message = '';
    let sb_rows = dataResult.rows();

    if (mapType === 'custom') {
      config.rows = sb_rows;
    } else if (sb_rows.length === 0) {
      noLatLong = true;
      mapType = 'none';
    }

    if (mapType === 'zipcode') {
      let zip_index;

      config.columns.forEach(col => {
        if (col.colgeotype() === 'zipcode') {
          zip_index = col.colindex();
        }
      });

      sb_rows.forEach(d => {
        const lat_long = zipcodeLatLong[`${d[zip_index]}`];
        if (lat_long) {
          let [long, lat] = lat_long;
          processRow(config, d, lat, long);
        } else {
          illegalVals.numIllegalGeo++;
        }
      });

      if (illegalVals.numIllegalGeo > 0) {
        alert_message = alert_message + 'Ignoring ' + illegalVals.numIllegalGeo +
          ' row(s) with invalid ZIP codes.';
      }
    } else if (!noLatLong && sb_rows.length > 0) {
      let latIndex, longIndex;
      latIndex = config.latCol.length > 0 ? config.latCol[0].colindex() : null;
      longIndex = config.longCol.length > 0 ? config.longCol[0].colindex() : null;

      sb_rows.forEach((d) => {
        let lat, long;
        if (latIndex !== null) {
          lat = Number(d[latIndex]);
        }

        if (longIndex !== null) {
          long = Number(d[longIndex]);
        }

        processRow(config, d, lat, long, illegalVals);
      });
    }

    if (routesEnabled.value() === 'true' && config.mapServer === 'mapboxgl' &&
      config.points.length > 1) {
      for (let color in config.routeColors) {
        const pointsArray = config.routeColors[color];
        if (pointsArray.length > 1) {
          config.endGeoJSON.features.push(pointsArray[pointsArray.length - 1]);
          config.midGeoJSON.features =
            config.midGeoJSON.features.concat(pointsArray.slice(1, pointsArray.length - 1));
        }
      }
    }

    if (illegalVals.numIllegalGeo > 0) {
      alert_message = alert_message + 'Ignoring ' + illegalVals.numIllegalGeo +
                        ' row(s) with non-numeric latitudes and longitudes. ';
    }

    if (illegalVals.numIllegalAggs > 0) {
      alert_message = alert_message + 'Setting ' + illegalVals.numIllegalAggs +
                        ' aggregate(s) with non-numeric data to 0.';
    }

    if (illegalVals.numIllegalColors > 0) {
      alert_message = alert_message + 'Setting ' + illegalVals.numIllegalColors +
                      ' row(s) to 0. ';
    }

    // Show alerts only in edit mode.
    // These are really warnings that should be flagged while Dashboard development is ongoing.
    if (alert_message && arc.globals.is_edit) {
      lib.alertMessage(alert_message, undefined, { noConsoleLog: true });
    }

    if (mapArtifacts && !arc.testutil.testMode) {
      map = mapArtifacts.mapvar;
      mapPreloaded = true;

      if (autoUpdateLocation &&  config.longRange[0] && config.longRange[0] && config.latRange[0] &&
        config.latRange[1]) {
        let minPoint = [config.longRange[0], config.latRange[0]];
        let maxPoint = [config.longRange[1], config.latRange[1]];
        if (mapServer.value() !== 'mapboxgl') {
          minPoint.reverse();
          maxPoint.reverse();
          map.fitBounds([minPoint, maxPoint]);
        } else {
          const startData = { range: [minPoint, maxPoint] };
          map.fitBounds(startData.range, { padding: { top: 40, bottom: 50, left: 25, right: 25 } });
        }
      }
    } else {
      mapArtifacts = {
        mapvar: undefined,
        zoomlevel: undefined,
        lines: [],
        markers: [],
        modalState: inModal,
      };

      let startData, minPoint, maxPoint;

      if (config.longRange[0] && config.longRange[0] && config.latRange[0] && config.latRange[1]) {
        minPoint = [config.longRange[0], config.latRange[0]];
        maxPoint = [config.longRange[1], config.latRange[1]];
        startData = { range: [minPoint, maxPoint] };
      } else if (noLatLong) {
        const centerPoint = startCenter.value().split(',').map(val => Number(val));
        startData = { center: centerPoint, zoom: Number(startZoom.value()) };
      } else {
        startData = { center: defaultStartLocation, zoom: defaultStartZoom };
      }

      // create map object and initialize tile server with base map and all that jazz
      map = initializeMap(config.mapServer, mapStyles.value(), config.latLongPoints, config.map_div,
        mapboxStyle.value(), startData);
      mapPreloaded = false;
      choroplethDataHash = {};
      choroplethRenderedFeatures = {};

      if (showMapData.value() === 'true' && mapServer.value() === 'mapboxgl') {
        createMapInfoDiv();
      }

      if (holdLocationManualMove.value() === 'true') {
        map.on('dragend', () => autoUpdateLocation = false);
        map.on('wheel', () => autoUpdateLocation = false);
      }

      mapArtifacts.mapvar = map;
    }

    if (!map) {
      lib.alertMessage('Error: map creation failed - no map object');
      return leaflet;
    }

    const savePositionData = config.mapServer === 'mapboxgl' && noLatLong &&
      rememberMapView.value() === 'true' && isEdit;

    if (savePositionData) {
      rememberPositionData();
    }

    if (mapType === 'none') {
      // There is no geo data so nothing to place on the map;
      return leaflet;
    }

    if (mapType === 'custom') {
      if (choroplethLayerId.value() !== '' && config.choroCol) {
        createChoropleth(config, sb_rows);
      } else {
        lib.alertMessage('Error: Please specify layer ID in Settings -> Choropleth to visualize ' +
          'a choropleth');
      }

      return leaflet;
    }

    if (config.mapServer === 'mapboxgl' && config.points.length > 0) {
      if (mapPreloaded) {
        updateData(config);
        loadData(config);
      } else {
        map.on('load', () => loadData(config));
      }
    } else if (config.points.length < 1) {
      return leaflet;
    }

    if (heatmapEnabled.value() === 'true') {
      createHeatmap(config);
    }

    if (cluster_enabled.value() === 'true') {
      createClusterGroups(config);
    }

    if (markersEnabled.value() === 'true' && routesEnabled.value() !== 'true') {
      createMarkers(config);
    }

    if (routesEnabled.value() === 'true') {
      createRoutes(config);
    }

    if (circle_enabled.value() === 'true') {
      createCircles(config);
    }

    return leaflet;
  };

  function initializeMap(mapServer, mapStyle, points, mapDiv, customMapStyle, startData) {
    let map;
    if (mapServer === 'google') {
      map = initializeGoogle(mapStyle, points, mapDiv);
    } else {
      map = initializeMapboxGL(mapStyle, points, mapDiv, customMapStyle, startData);
    }

    return map;
  }

  function initializeGoogle(mapStyle, points, mapDiv) {
    if (points.length === 0) {
      map = L.map(mapDiv, { maxZoom: 18 });
      map.setView(defaultStartLocation.reverse(), 3);
    } else {
      map = L.map(mapDiv, {
                maxZoom: 18,
                minZoom: 3,
                maxBounds: L.latLngBounds([-90, 180], [90, -180]),
              })
              .fitBounds(points)
              .on('unload', destroyArcadiaTooltips);
    }

    function buildGoogleLayer() {
      let googleMutantObject;
      if (mapStyle === 'default') {
        googleMutantObject = { type: 'roadmap' };
      } else if (mapStyle === 'satellite') {
        googleMutantObject = { type: 'satellite' };
      } else {
        let googleMapStyle = arc.settings.mapInteractive.mapStyleOptions[mapStyle];
        googleMutantObject = { type: 'roadmap', styles: googleMapStyle };
      }

      googleLayer = L.gridLayer.googleMutant(googleMutantObject)
                                .on('load', function() {
                                  arc.testutil.tilesLoaded = true;
                                  googleLayer.off('load');
                                });

      googleLayer.addTo(map);
    }

    if ($('head script[src^="' + baseGoogleURL + '"]').length > 0) {
      var existingURL = $('head script[src^="' + baseGoogleURL + '"]').first().attr('src');
      arc.misc.addScripts([existingURL], buildGoogleLayer);
    } else {
      var key = 'v=3';

      if (!google_api_id && !google_client_id) {
        lib.alertMessage('Error: No Google Maps API Key or Client Key Entered.');
      } else {
        if (google_api_id) {
          key += (`&key=${google_api_id}`);
        }

        if (google_client_id) {
          key += (`&client=${google_client_id}`);
        }

        arc.misc.addScripts([baseGoogleURL + key], buildGoogleLayer);
      }
    }

    if (points && points.length > 0) {
      map.fitBounds(points);
    }

    return map;
  }

  function initializeMapboxGL(mapStyle, points, mapDiv, customMapStyle, startData) {
    let map, style;

    if (localServerAddress) {
      mapboxgl.config.API_URL = localServerAddress;
    }

    if (mapServerAddress.value() !== '') {
      mapboxgl.config.API_URL = mapServerAddress.value();
    }

    const mapboxStyleOptions = {
        dark: 'mapbox://styles/mapbox/dark-v9',
        default: 'mapbox://styles/mapbox/streets-v10',
        light: 'mapbox://styles/mapbox/light-v9',
        subtle: 'mapbox://styles/mapbox/basic-v9',
        muted: 'mapbox://styles/mapbox/outdoors-v10',
        satellite: 'mapbox://styles/mapbox/satellite-v9',
      };

    googleLayer = null;
    if (!mapboxAccessToken) {
      lib.alertMessage('Error: No Mapbox Access Token Entered.');
      return null;
    } else {
      mapboxgl.accessToken = mapboxAccessToken;
    }

    if (customMapStyle && customMapStyle !== '') {
      style = customMapStyle;
    } else {
      style = mapboxStyleOptions[mapStyle];
    }

    const initialMapObject = {
      container: mapDiv,
      style: style,
      attributionControl: false,
      preserveDrawingBuffer: true,
    };

    // const initialMapObject = { container: mapDiv, style: style, attributionControl: false };

    if (startData.center) {
      initialMapObject.center = startData.center;
      initialMapObject.zoom = startData.zoom;
    }

    if (points.length === 0) {
      map = new mapboxgl.Map(initialMapObject);
    } else {
      map = new mapboxgl.Map(initialMapObject);
    }

    if (startData.range) {
      map.fitBounds(startData.range, { padding: { top: 40, bottom: 50, left: 25, right: 25 } });
    }

    return map;
  }

  function updateNumericRange(d, range) {
    if (range[0] === undefined || d < range[0]) {
      range[0] = d;
    }

    if (range[1] === undefined || d > range[1]) {
      range[1] = d;
    }
  }

  function processRow(config, d, lat, long, illegalVals) {
    if (!lat || !long || (lat && !arc.misc.isNumeric(lat)) || (long && !arc.misc.isNumeric(long))) {
      illegalVals.numIllegalGeo += 1;
      return;
    }

    let rowData = [];
    if (lat && long) {
      lat = Number(lat);
      long = Number(long);
      updateNumericRange(lat, config.latRange);
      updateNumericRange(long, config.longRange);
      rowData.push(lat, long);
      config.latLongPoints.push([lat, long]);
    }

    let agg, color;
    if (aggCol) {
      agg = d[aggCol.colindex()];
    }

    if (colorCol) {
      color = d[colorCol.colindex()];
    }

    // Need to check for zero aggregates. Use a val of 1 if no aggregates are specified.
    // Solr connections cannot support 2 dimensions and 1 aggregate, and so
    // this line of code is needed for Solr functionality.
    if (agg === undefined) {
      agg = 1;
    } else if (!arc.misc.isNumeric(agg)) {
        agg = 0;
        illegalVals.numIllegalAggs++;
      } else {
        agg = Number(agg);
      }

    rowData.push(agg);
    updateNumericRange(agg, aggRange);

    if (colorCol) {
      if (colorIsNumeric) {
        if (color === undefined || arc.misc.check_if_value_null(color)) {
          illegalVals.numIllegalColors++;
          color = 0;
        }

        color = Number(color);
        updateNumericRange(color, colorRange);
      }

      if (colorIsDim) {
        colorDomain[color] = true;
      }

      rowData.push(color);
    }

    if (lat && long) {
      config.points.push(rowData);
    } else {
      return;
    }

    config.popup_info.push(d);

    const featurePoint = { type: 'Feature', properties: { agg: agg, color: color, row: d },
      geometry: { type: 'Point', coordinates: [long, lat] },
    };

    config.pointGeoJSONData.features.push(featurePoint);

    if (colorIsDim && routesEnabled.value() === 'true') {
      if (config.routeColors[color]) {
        config.routeColors[color].push(featurePoint);
        config.routeGeoJSON[color].geometry.coordinates.push([long, lat]);
      } else {
        config.startGeoJSON.features.push(featurePoint);
        config.routeColors[color] = [featurePoint];
        config.routeGeoJSON[color] = { type: 'Feature', properties: { color: color },
          geometry: { type: 'LineString', coordinates: [[long, lat]] }, };
      }
    } else if (routesEnabled.value() === 'true') {
      if (config.routeColors.singleRoute) {
        config.routeColors.singleRoute.push(featurePoint);
        config.routeGeoJSON.singleRoute.geometry.coordinates.push([long, lat]);
      } else {
        const routeColor = lib.colors()[0];
        config.startGeoJSON.features.push(featurePoint);
        config.routeColors.singleRoute = [];
        config.routeGeoJSON.singleRoute = { type: 'Feature', properties: { color: routeColor },
          geometry: { type: 'LineString', coordinates: [[long, lat]] }, };
      }
    }

    if (heatmapEnabled.value() === 'true') {
      let modified_point = rowData.slice(0, 3);// heatmap doesn't care about color info
      modified_point[2] = modified_point[2] * (Number(intensityMag.value()));
      config.pointsForHeatMap.push(modified_point);
    }
  }

  function loadData(config) {
    if (!map.getSource(sourceIdObject.main)) {
      map.addSource(sourceIdObject.main, {
        type: 'geojson',
        data: config.pointGeoJSONData,
      });
    }

    if (routesEnabled.value() === 'true' && !map.getSource(sourceIdObject.startMarks)) {
      map.addSource(sourceIdObject.startMarks, { type: 'geojson', data: config.startGeoJSON });
      map.addSource(sourceIdObject.midMarks, { type: 'geojson', data: config.midGeoJSON });
      map.addSource(sourceIdObject.endMarks, { type: 'geojson', data: config.endGeoJSON });
    }
  }

  function updateData(config) {
    if (map && map.getSource(sourceIdObject.main)) {
      map.getSource(sourceIdObject.main).setData(config.pointGeoJSONData);
    }

    if (map && map.getSource(sourceIdObject.startMarks)) {
      map.getSource(sourceIdObject.startMarks).setData(config.startGeoJSON);
      map.getSource(sourceIdObject.midMarks).setData(config.midGeoJSON);
      map.getSource(sourceIdObject.endMarks).setData(config.endGeoJSON);
    }
  }

  function createMapInfoDiv() {
    let $mapInfoDiv = $(`#${mapDiv}`).find('.map-info-div');
    if ($mapInfoDiv.length > 0) {
      return;
    }

    $mapInfoDiv = $('<div class="map-info-div"></div>');
    const mapInfoStyle = {
      position: 'absolute',
      backgroundColor: '#eee',
      bottom: '0px',
      right: '0px',
      height: '20px',
      width: '280px',
      opacity: '0.7',
      paddingLeft: '5px',
    };

    $mapInfoDiv.css(mapInfoStyle);
    $(`#${mapDiv}`).append($mapInfoDiv);

    const updateMapInfo = () => {
      const currZoom = String(map.getZoom()).slice(0, 4);
      const currCenter = map.getCenter();
      const currLat = String(currCenter.lat).slice(0, 12);
      const currLong = String(currCenter.lng).slice(0, 12);
      $mapInfoDiv.text(`center: ${currLat}, ${currLong}, zoom: ${currZoom}`);
    };

    if (map.loaded()) {
      updateMapInfo();
    } else {
      map.on('load', updateMapInfo);
    }

    map.on('zoomend', updateMapInfo);
    map.on('moveend', updateMapInfo);
  }

  function removeMapInfoDiv() {
    const $mapInfoDiv = $(`#${mapDiv}`).find('.map-info-div');

    if ($mapInfoDiv.length < 1) {
      return;
    }

    $mapInfoDiv.remove();
  }

  function rememberPositionData() {
    const updateMapLocation = () => {
      const currZoom = map.getZoom();
      const currCenter = map.getCenter();
      startZoom.value(currZoom);
      startCenter.value(`${currCenter.lng}, ${currCenter.lat}`);
    };

    map.on('load', updateMapLocation);
    map.on('zoomend', updateMapLocation);
    map.on('moveend', updateMapLocation);
  }

  const addHeatLayer = function(config) {
    const colors = lib.colors();
    const colors_length = colors.length;
    const ordinal_domain = d3.range(colors_length);

    const heat_domain = [
      aggRange[0] * (Number(intensityMag.value())),
      aggRange[1] * (Number(intensityMag.value())),
    ];

    const heatmap_range = d3.scale.linear()
                        .range(heat_domain)
                        .domain([0, colors_length - 1]);

    const color_scale = d3.scale.ordinal()
                        .range(colors)
                        .domain(ordinal_domain);

    const gradient_scale = d3.scale.linear()
                        .range([0, 1])
                        .domain([0, colors_length - 1]);

    const gradient = {};
    ordinal_domain.forEach((d) => {
      gradient[gradient_scale(d)] = color_scale(d);
    });

    // eslint-disable-next-line new-cap
    const heat = new L.heatLayer(
      config.pointsForHeatMap,
      {
        radius: Number(heatmapRadius.value()),
        gradient,
      }
    );

    heat.getV = function() {
      const maxZoom = this.options.maxZoom === undefined ?
        this._map.getMaxZoom() : this.options.maxZoom;
      return 2 ** Math.max(0, Math.min(maxZoom - this._map.getZoom(), 12));
    };

    heat.addTo(map);

    if (heatmapLegend.value() === 'true') {
      // eslint-disable-next-line new-cap
      const legend = new L.control({ position: 'bottomright' });
      const legendExtraHorizontalMargin = 40;
      const legendValues = [heatmap_range(0), heat.getV()];

      const legend_width = lib.dim_axis_max_textlen(
        legendValues.map((d) => lib.tickFormatDefaultNumeric(d)), lib.xaxis_label_maxlen()
      ) + legendExtraHorizontalMargin;

      legend.onAdd = function () {
        const div = L.DomUtil.create('div', 'info legend heatmap_legend');
        div.innerHTML += (`<svg class="arc-viz-legend" width="${legend_width}"></svg>`);
        return div;
      };

      legend.addTo(map);

      const lsvg = config.d3_map_div
                      .select('div.heatmap_legend')
                      .select('.arc-viz-legend');

      const legend_config = {
        legendType: 'continuous',
        legendPositioning: 'Right',
        svg: lsvg,
        width: 0,
        color: color_scale,
        legends: legendValues,
        color_palette: colors,
        legend_width,
        sort_domain: true,
      };

      lib.draw_legend(legend_config);
      lsvg.attr('height', `${legend_config.height}px`);

      map.on('zoomend', () => {
        legend_config.legends = [heatmap_range(0), heat.getV()];
        lib.draw_legend(legend_config);
      });
    }
  };

  function createHeatmap(config) {
    if (config.mapServer === 'mapboxgl') {
      createGLHeatmap();
    } else {
      addHeatLayer(config);
    }
  }

  function createGLHeatmap() {

    if (map.getLayer(layerIdObject.heatmap)) {
      return;
    }

    const colors = lib.colors(),
        ordinalDomain = d3.range(colors.length);

    const colorScale = d3.scale.ordinal()
                        .range(colors)
                        .domain(ordinalDomain);

    const gradientScale = d3.scale.linear()
                        .range([0, 1])
                        .domain([0, colors.length - 1]);

    var gradient = {};
    ordinalDomain.forEach(function(d) {
      gradient[gradientScale(d)] = colorScale(d);
    });

    let zeroColor = arc.colorpalette.convertToRGBAString(gradient[0], 0);

    const heatMapColor = ['interpolate', ['linear'], ['heatmap-density'], 0, zeroColor];
    let gradientValues = Object.keys(gradient).sort((a, b) => Number(a) - Number(b));
    gradientValues.shift();
    gradientValues.forEach(val =>
      heatMapColor.push(Number(val), arc.colorpalette.convertToRGBAString(gradient[val])));
    const zoomLevels = heatmapVisibleZoom.get_range();
    const lowerVal = aggRange[0] < aggRange[1] ? aggRange[0] : 0;
    const heatMapObject = {
      id: layerIdObject.heatmap,
      type: 'heatmap',
      source: sourceIdObject.main,
      minzoom: zoomLevels[0],
      maxzoom: zoomLevels[1] === (maxVisibleZoom - 1) ? maxVisibleZoom : zoomLevels[1],
      paint: {
        'heatmap-weight': [
          'interpolate',
          ['linear'],
          ['get', 'agg'],
          lowerVal, 0,
          aggRange[1], Number(intensityMag.value()) / 1000,
        ],
        'heatmap-radius': Number(heatmapRadius.value()),
        'heatmap-color': heatMapColor,
      },
    };

    if (mapPreloaded) {
      map.addLayer(heatMapObject);
    } else {
      map.on('load', () => map.addLayer(heatMapObject));
    }
  }

  function createClusterGroups(config) {
    if (config.mapServer === 'mapboxgl') {
      createGLClusterGroups(config);
    } else {
      addClusterGroups(config);
    }
  }

  var addClusterGroups = function(config) {

    const clusterColor = cluster_color.value(),
      styling = `"background-color: ${clusterColor};"`,
      radius = Number(cluster_radius.value());

    var buildClusterTooltips = function() {
      var d3_markers = config.d3_map_div
                         .selectAll('img[alt^="marker_for_cluster"]');
      var popup_indices = [];

      if (d3_markers.size()) {
        var n = d3_markers.each(function(d, i) {
          popup_indices.push(Number(d3.select(this)
                                .attr('alt')
                                .substring('marker_for_cluster'.length)));
        });

        createTooltips(config, d3_markers, popup_indices);
      }
    };

    let clusterDisplayFormat = aggCol ? aggCol.coldisplayformat() : null;

    var clusterGroup = new L.MarkerClusterGroup({
      maxClusterRadius: radius,
      zoomCb: buildClusterTooltips,
      iconCreateFunction: function (cluster) {
        var markers = cluster.getAllChildMarkers();

        var n = 0;
        for (var i = 0; i < markers.length; i++) {
          n += Number(markers[i].number);
        }

        if (clusterDisplayFormat) {
          n = clusterDisplayFormat(n);
        }

        return L.divIcon({ html: '<div style =' + styling + '><span>' + n +
          '</span></div>', className: 'marker-cluster', iconSize: L.point(40, 40), });
      },
    });

    config.points.forEach(function(point, i) {
      if (colorCol) {
        point.pop();
      }

      var m = L.marker(point, { alt: 'marker_for_cluster' + i });
      if (point[2]) {
        m.number = point[2];
      } else {
        m.number = 1;
      }

      clusterGroup.addLayer(m);
    });

    clusterGroup.on('spiderfied', buildClusterTooltips);

    map.addLayer(clusterGroup);
    buildClusterTooltips();
  };

  function createGLClusterGroups(config) {
    const clusterColor = cluster_color.value(),
      clusterRadius = Number(cluster_radius.value()),
      clusterLabelColor = arc.colorpalette.colorForBrightness(clusterColor);

    const clusterSourceObject = {
      type: 'geojson',
      data: config.pointGeoJSONData,
      cluster: true,
      clusterMaxZoom: 14,
      clusterRadius: clusterRadius,
    };

    const clusterLayerObject = {
      id: layerIdObject.cluster,
      type: 'circle',
      source: sourceIdObject.cluster,
      filter: ['has', 'point_count'],
      paint: {
          'circle-color': clusterColor,
          'circle-radius': 15,
        },
    };

    const clusterTextObject = {
      id: layerIdObject.clusterText,
      type: 'symbol',
      source: sourceIdObject.cluster,
      filter: ['has', 'point_count'],
      layout: {
          'text-field': '{point_count_abbreviated}',
          'text-font': ['DIN Offc Pro Medium', 'Arial Unicode MS Bold'],
          'text-size': 12,
        },
      paint: {
          'text-color': clusterLabelColor,
        },
    };

    const singlePointObject = {
      id: layerIdObject.clusterSingle,
      type: 'circle',
      source: sourceIdObject.cluster,
      filter: ['!', ['has', 'point_count']],
      paint: {
          'circle-color': clusterColor,
          'circle-radius': 10,
          'circle-stroke-width': 1,
          'circle-stroke-color': '#fff',
        },
    };

    if (mapPreloaded) {
      if (map.getSource(sourceIdObject.cluster)) {
        map.getSource(sourceIdObject.cluster).setData(config.pointGeoJSONData);
      } else {
        map.addSource(sourceIdObject.cluster, clusterSourceObject);
        map.addLayer(clusterLayerObject);
        map.addLayer(clusterTextObject);
        map.addLayer(singlePointObject);
      }

      if (config.includeTooltips) {
        const textFn = getTextOrParamFn(config.columns, layerIdObject.clusterSingle);
        const paramsFn = getTextOrParamFn(config.columns, layerIdObject.clusterSingle, true);
        createGLTooltips(layerIdObject.clusterSingle, textFn, map, paramsFn);
      }
    } else {
      map.on('load', function() {
        map.addSource(sourceIdObject.cluster, clusterSourceObject);
        map.addLayer(clusterLayerObject);
        map.addLayer(clusterTextObject);
        map.addLayer(singlePointObject);
      });

      if (config.includeTooltips) {
        const textFn = getTextOrParamFn(config.columns, layerIdObject.clusterSingle);
        const paramsFn = getTextOrParamFn(config.columns, layerIdObject.clusterSingle, true);
        createGLTooltips(layerIdObject.clusterSingle, textFn, map, paramsFn);
      }
    }
  }

  function createTooltips(config, selection, n, setup_cb) {
    var params_fn = function(d, i) {
      if (n) {
        i = n[i];
      }

      return lib.params_from_columns(config.popup_info[i], config.columns);
    };

    var text_fn = function(d, i) {
      if (n) {
        i = n[i];
      }

      return lib.text_from_columns(config.popup_info[i], config.columns);
    };

    var tooltip =
      arc.tooltip()
         .dom_id('#tooltip')
         .addSelection(selection)
         .text_fn(text_fn)
         .activate();

    var contextmenu =
      lib.contextmenu('primary')
       .addTooltip(tooltip)
       .text_fn(text_fn)
       .params_fn(params_fn)
       .setup_cb(setup_cb)
       .activate();

    allArcTooltips.push(tooltip);
    allArcTooltips.push(contextmenu);
  }

  function createGLTooltips(mapLayerId, textFn, map, paramsFn) {
    const tooltip = arc.tooltip()
      .dom_id('#tooltip')
      .addSelection(map)
      .text_fn(textFn)
      .mapLayerId(mapLayerId)
      .activate();

    const contextmenu = lib.contextmenu('primary')
                            .addTooltip(tooltip)
                            .text_fn(textFn)
                           .params_fn(paramsFn)
                            .mapLayerId(mapLayerId)
                            .highlight_on_select(false)
                           .activate();

    allArcTooltips.push(tooltip);
    allArcTooltips.push(contextmenu);
  }

  function destroyArcadiaTooltips() {
    allArcTooltips.forEach(function(d) {
      d.destroy();
    });

    allArcTooltips = [];
  }

  function addMarkerToPoint(point, markerType, color, customIconUrl, arcColoredIcons, colorLabel) {
    let marker;
    let markerIcon;
    if (arcColoredIcons) {
      marker = arcColoredIcons[color];
    }

    let markerHTMLStyle;

    if (markerIcon === undefined) {
      if (markerType === 'circle') {
        const defaultRadius = 5;
        marker = L.circleMarker([point[0], point[1]], {
          color: color,
          fillColor: color,
          fillOpacity: 0.5,
          className: 'marker-pins',
          radius: defaultRadius,
          weight: 1,
        });
      } else if (markerType === 'custom' && customIconUrl !== '') {
        markerIcon = L.icon({
          iconUrl: customIconUrl,
          iconSize:     [32, 40],
          iconAnchor:   [16, 40],
          popupAnchor:  [-3, -16],
          className: 'marker-pins',
        });
        marker = L.marker([point[0], point[1]], { alt: 'marker_point', icon: markerIcon });
      } else {
        if (mapServer.value() !== 'mapboxgl') {
          markerHTMLStyle = arc.settings.mapInteractive.markerOptions.basic;
          markerHTMLStyle = arc.misc.safe_substitute(markerHTMLStyle, { ICONCOLOR: color });
          markerIcon = L.divIcon({
            className: 'marker-pins',
            iconAnchor: [14, 35],
            labelAnchor: [-6, 0],
            popupAnchor: [0, -36],
            html: markerHTMLStyle,
          });
          marker = L.marker([point[0], point[1]], { alt: 'marker_point', icon: markerIcon });
        } else {
          if (point[0] && point[1]) {
            marker = new mapboxgl.Marker({ color: color }).setLngLat([point[1], point[0]]);
          }
        }
      }

      if (arcColoredIcons) {
        arcColoredIcons[color] = marker;
      }
    }

    if (marker) {
      marker.addTo(map);
      if (markersPermanentLabels.value() === 'true' && colorLabel) {
        let colorSuffix = colorLabel.split(' ').join('');
        let className = 'arc-leaflet-tooltip ' + colorSuffix;
        marker.bindTooltip(colorLabel, {
          permanent: true,
          opacity: 0.8,
          direction: 'bottom',
          className: className,
        });
        const selector = '.arc-leaflet-tooltip.' + colorSuffix;
        $(selector).css('background', color);
      }

      mapArtifacts.markers.push(marker);
    }
  }

  var addMarkers = function(config) {
    let colorFn = get_color_fn(config);
    let arcColoredIcons = {};
    var d3_markers;

    //Limit query sizes to 2000 for marker layer performance. Queries with 2000+
    //rows seem to utilize over 1GB of RAM
    if (config.points.length > 2000) {
      lib.alertMessage('Error: For performance reasons, the marker layer ' +
        'cannot be drawn for queries with more than 2000 rows.');
      return;
    }

    config.points.forEach(function(point) {
      let colorLabel = point[3];
      let color = colorFn(colorLabel);
      let customURL = customIcon.value();
      addMarkerToPoint(point, markerType.value(), color, customURL, arcColoredIcons, colorLabel);
    });

    $('.marker_pins > svg > g').css('pointer-events', 'none');
    d3_markers = config.d3_map_div.selectAll('.marker-pins');
    createTooltips(config, d3_markers);
  };

  function createGLMarkers(config) {
    let d3Markers = config.d3_map_div.selectAll('.mapboxgl-marker');
    d3Markers.remove();
    if (markerType.value() === 'circle') {
      createGLCircles(config, layerIdObject.pins, 5);
      return;
    }

    if (markerType.value() === 'custom' && customIcon.value() !== '') {
      createGLSymbols(config, customIcon.value());
      return;
    }

    if (config.points.length > 2000) {
      lib.alertMessage('Error: For performance reasons, the marker layer ' +
        'cannot be drawn for queries with more than 2000 rows.');
      return;
    }

    const colorFn = get_color_fn(config);
    const arcColoredIcons = {};
    config.points.forEach(function(point) {
      let colorLabel = point[3];
      let color = colorFn(colorLabel);
      let customURL = customIcon.value();
      addMarkerToPoint(point, markerType.value(), color, customURL, arcColoredIcons, colorLabel);
    });

    d3Markers = config.d3_map_div.selectAll('.mapboxgl-marker');
    createTooltips(config, d3Markers);
  }

  function createMarkers(config) {
    if (config.mapServer === 'mapboxgl') {
      createGLMarkers(config);
    } else {
      addMarkers(config);
    }
  }

  function createRoutes(config) {
    if (config.mapServer === 'mapboxgl') {
      addGLRoutes(config);
    } else {
      addRoutes(config);
    }
  }

  var addRoutes = function(config) {

    //Limit query sizes to 2000 for marker layer performance. Queries with 2000+
    //rows seem to utilize over 1GB of RAM
    if (config.points.length > 2000 && config.mapServer !== 'mapboxgl') {
      lib.alertMessage('Error: For performance reasons, the route layer ' +
        'cannot be drawn for queries with more than 2000 rows.');
      return;
    }

    const latlngs = {};
    let color_fn = get_color_fn(config);

    if (config.mapServer === 'mapboxgl') {
      addGLRoutes(config, color_fn);
      return;
    }

    config.points.forEach(function(point) {

      // TODO: This seems a little strange that we have hardcoded
      // idx 3 to be that for color. There should be a indirect
      // reference to index retrieved from the color shelf somehow
      let color = color_fn(point[3]);

      if (!latlngs[color]) {
        latlngs[color] = {
          colorLabel: point[3],
          points: [],
        };
      }

      latlngs[color].points.push([point[0], point[1]]);
    });

    let bounds;
    for (let color in latlngs) {
      const pointList = latlngs[color].points;
      const pointListColor = latlngs[color].colorLabel;

      if (markersEnabled.value() === 'true') {
        let currMarker;
        for (let idx = 0; idx < pointList.length; idx++) {
          if (idx === 0) {
            currMarker = startMarkers.value();
          } else if (idx === pointList.length - 1) {
            currMarker = endMarkers.value();
          } else {
            currMarker = middleMarkers.value();
          }

          if (currMarker !== 'none') {
            addMarkerToPoint([pointList[idx][0], pointList[idx][1]],
                             currMarker, color, customIcon.value(),
                             undefined, pointListColor);
          }
        }
      }

      let polyline = L.polyline(pointList, { color: color }).addTo(map);
      mapArtifacts.lines.push(polyline);
      if (bounds === undefined) {
        bounds = polyline.getBounds();
      } else {
        bounds.extend(polyline.getBounds());
      }
    }

    // Make the tooltips be smoother by keeping the internal elements not selective
    $('.marker_pins > svg > g').css('pointer-events', 'none');
    let d3_markers = config.d3_map_div.selectAll('.marker-pins');
    createTooltips(config, d3_markers);

    if (bounds) {
      // zoom the map to the polyline
      map.fitBounds(bounds);
    }
  };

  function addGLRoutes(config) {
    const colorFn = get_color_fn(config);

    const updateRoutes = function() {
      for (let c in config.routeGeoJSON) {
        const routeId = `arcadia-routes-${c}`;
        routeSourceIds[routeId] = `arcadia-route-source-${c}`;
        let currGeoJSON = config.routeGeoJSON[c];
        if (map.getLayer(routeId)) {
          const currSource = map.getSource(routeSourceIds[routeId]);
          if (currSource) {
            currSource.setData(currGeoJSON);
            return;
          }

          map.removeLayer(routeId);
        }

        map.addSource(routeSourceIds[routeId], { type: 'geojson', data: currGeoJSON });

        map.addLayer({
          id: `arcadia-routes-${c}`,
          type: 'line',
          source: routeSourceIds[routeId],
          layout: {
            'line-join': 'round',
            'line-cap': 'round',
          },
          paint: {
            'line-color': c === 'singleRoute' ? currGeoJSON.properties.color : colorFn(c),
            'line-width': 2,
          },
        });
      }
    };

    if (mapPreloaded) {
      updateRoutes();
    } else {
      map.on('load', updateRoutes);
    }

    const customIconUrl = customIcon.value();
    let d3Markers = config.d3_map_div.selectAll('.mapboxgl-marker');
    d3Markers.remove();

    createRouteGLMarkers(config, startMarkers.value(), config.startGeoJSON, colorFn, 'startMarks',
      customIconUrl);
    createRouteGLMarkers(config, middleMarkers.value(), config.midGeoJSON, colorFn, 'midMarks',
      customIconUrl);
    createRouteGLMarkers(config, endMarkers.value(), config.endGeoJSON, colorFn, 'endMarks',
      customIconUrl);
  }

  function createRouteGLMarkers(config, markerType, geoJSON, colorFn, objectId, customIconURL) {
    const layerId = layerIdObject[objectId];
    const sourceId = sourceIdObject[objectId];
    if (markerType === 'none') {
      return;
    } else if (markerType === 'circle') {
      createGLCircles(config, layerId,  5, sourceId);
    } else if (markerType === 'custom' && customIconURL !== '') {
      createGLSymbols(config, customIconURL, sourceId, layerId);
    } else {
      geoJSON.features.forEach(function (feature) {
        if (!feature) return;
        addMarkerToPoint(feature.geometry.coordinates.reverse(), markerType,
          colorFn(feature.properties.color), customIconURL);
      });

      let d3Markers = config.d3_map_div.selectAll('.mapboxgl-marker');
      createTooltips(config, d3Markers);
    }
  }

  function createCircles(config) {
    if (config.mapServer === 'mapboxgl') {
      createGLCircles(config);
    } else {
      addCircleLayer(config);
    }
  }

  var addCircleLayer = function(config) {
    //Limit query sizes to 2000 for marker layer performance. Queries with 2000+
    //rows seem to utilize over 1GB of RAM mapboxgl has no problem
    if (config.points.length > 2000) {
      lib.alertMessage('Error: For performance reasons, the circle layer ' +
        'cannot be drawn for queries with more than 2000 rows.');
      return;
    }

    var d3_markers = config.d3_map_div.selectAll('.circle_marker');
    d3_markers.remove();

    let color_fn = get_color_fn(config);
    const radiusRange = d3.scale.linear().range(circleRadiusRange.get_range())
                                         .domain(aggRange);

    let aggToRadiusLength = (agg) => radiusRange(agg);

    config.points.forEach((point, i) => {
      let r = aggToRadiusLength(point[2]);
      let color = color_fn(point[3]);
      let circle = L.circleMarker([point[0], point[1]], {
          color: color,
          fillColor: color,
          fillOpacity: 0.5,
          className: 'circle_marker',
          radius: r,
          strokeWidth: 1,
      });

      circle.addTo(map);
    });

    d3_markers = config.d3_map_div.selectAll('.circle_marker');

    var setup_cb =  function() {
      d3_markers.classed('viz-fully-visible', false);
    };

    createTooltips(config, d3_markers, null, setup_cb);
    if (circleLegendArea.value() === 'true' && aggCol) {
      let areaLegend;
      if (config.mapServer !== 'mapboxgl') {
        areaLegend = new L.control({ position: 'bottomright' });
      }

      const midValue = (aggRange[0] + aggRange[1]) / 2;
      let labels = arc.misc.arrayUnique([aggRange[0], midValue, aggRange[1]]);
      const horizontalSpacing = 5;
      let legendWidth = getLegendWidth(labels,
        2 * aggToRadiusLength(aggRange[1]) + horizontalSpacing * 3);
      areaLegend.onAdd = function (map) {
        let div = L.DomUtil.create('div', 'info legend circle_area_legend');
        div.innerHTML += '<svg class="arc-viz-legend"></svg>';
        return div;
      };

      areaLegend.addTo(map);

      let areaSvg = config.d3_map_div
              .select('div.circle_area_legend')
              .select('.arc-viz-legend');

      const radiusRange = circleRadiusRange.get_range();
      const smallR = radiusRange[0];
      const largeR = radiusRange[1];
      const areaRange = [Math.PI * smallR * smallR, Math.PI * largeR * largeR];
      const areaScale = d3.scale.linear().range(areaRange).domain(aggRange);

      let area_legend_config = {
        legendType: 'area',
        legendPositioning: 'Right',
        svg: areaSvg,
        width: 0,
        legends: labels,
        legend_column: aggCol,
        legend_width: legendWidth,
        area_scale: areaScale,
        legend_title: aggCol.colname(),
      };

      lib.draw_legend(area_legend_config);
    }

    if (circleLegendColor.value() === 'true' && colorCol) {
      let colorLegend;
      config.drawnColorLegend = true;
      colorLegend = new L.control({ position: 'bottomright' });
      const legend_extra_horizontal_margin = 60;

      if (Object.keys(colorDomain).length > 200) {
        //Dom can't hold so many elements, also if your legend has more than 100 elements
        //it will not be super useful.
        lib.alertMessage('Error: legend has too many elements, can not show.');
        return;
      }

      let correct_legend_vals = [];
      let colorMappings = {};
      if (colorIsDim) {
        correct_legend_vals = Object.keys(colorDomain);
        correct_legend_vals.forEach((v, i) => {
          colorMappings[v] = color_fn(v);
        });
      } else {
        let colors = lib.colors();
        let legend_values = d3.range(colors.length);
        let domain_to_color_range = d3.scale.linear()
                .range([0, colors.length])
                .domain(colorRange);
        let colFormatter = (colorCol ? colorCol.coldisplayformat() : null) ||
          domain_to_color_range.tickFormat(colors.length);
        let uniqueLegendValues = {}, biggest_legend_str = '';
        legend_values.forEach((d, i) => {
          let val = colFormatter(domain_to_color_range.invert(d));
          if (i === (legend_values.length - 1)) {
            val += '+';
          } else {
            val = '≥' + val;
          }

          if (!uniqueLegendValues[val]) {
            uniqueLegendValues[val] = true;
            correct_legend_vals.push(val);
            if (val.length > biggest_legend_str.length) {
              biggest_legend_str = val;
            }
          }
        });

        correct_legend_vals.forEach(function(v, i) {
          colorMappings[v] = colors[i];
        });
      }

      let legend_width = getLegendWidth(correct_legend_vals, legend_extra_horizontal_margin);

      colorLegend.onAdd = function (map) {
        let div = L.DomUtil.create('div', 'info legend circle_color_legend');
        div.innerHTML += '<svg class="arc-viz-legend" width="' + legend_width + '"></svg>';
        return div;
      };

      colorLegend.addTo(map);

      var colorSvg = config.d3_map_div
                      .select('div.circle_color_legend')
                      .select('.arc-viz-legend');

      var color_scale = arc.misc.dimensionColorScale()
                                .init(false, colorMappings);

      var color_legend_config = {
        legendType: 'discrete',
        legendPositioning: 'Right',
        svg: colorSvg,
        width: 0,
        color: color_scale,
        legends: correct_legend_vals,
        legend_title: colorCol ? colorCol.colname() : '',
        sort_domain: false,
        legend_width: legend_width,
      };

      lib.draw_legend(color_legend_config);
    }
  };

  function createRadiusProperty(setRadius) {
    let circleRadius;
    const radiusRange = circleRadiusRange.get_range();
    if (setRadius) {
      circleRadius = setRadius;
    } else if (aggRange[1] && aggRange[0] && aggRange[1] - aggRange[0] > 0) {
      circleRadius = [
        'interpolate',
        ['linear'], ['get', 'agg'],
        aggRange[0], radiusRange[0],
        aggRange[1], radiusRange[1],
      ];
    } else {
      circleRadius = radiusRange[0];
    }

    return circleRadius;
  }

  function createCircleColorProperty(config) {
    const colorFn = get_color_fn(config);
    let circleColor;
    if (colorIsNumeric) {
      circleColor = ['interpolate', ['linear'], ['get', 'color']];
      const colorStep = (colorRange[1] - colorRange[0]) / (lib.colors().length - 1);
      for (let colVal = colorRange[0]; colVal < colorRange[1]; colVal += colorStep) {
        circleColor.push(colVal, colorFn(colVal));
      }

      circleColor.push(colorRange[1], colorFn(colorRange[1]));
    } else if (colorCol) {
      circleColor = ['match', ['get', 'color']];
      Object.keys(colorDomain).forEach(color => circleColor.push(color, colorFn(color)));

      // other color
      circleColor.push('#fff');
    } else {
      circleColor = arc.colorpalette.convertToRGBAString(lib.colors()[0]);
    }

    return circleColor;
  }

  function createGLCircles(config, layerId, setRadius, sourceId) {
    layerId = layerId || layerIdObject.circle;
    sourceId = sourceId || sourceIdObject.main;
    if (map.getLayer(layerId)) {
      return;
    }

    const circleRadius = createRadiusProperty(setRadius);
    const circleColor = createCircleColorProperty(config);
    let minZoom, maxZoom;
    const circleZoomLevels = circleVisibleZoom.get_range();
    if (setRadius) {
      minZoom = 0;
      maxZoom = maxVisibleZoom;
    } else {
      minZoom = circleZoomLevels[0];
      maxZoom = circleZoomLevels[1] === (maxVisibleZoom - 1) ?
                maxVisibleZoom : circleZoomLevels[1];
    }

    const circleLayerObject = {
        id: layerId,
        type: 'circle',
        source: sourceId,
        minzoom: minZoom,
        maxzoom: maxZoom,
        paint: {
            'circle-color': circleColor,
            'circle-radius': circleRadius,
            'circle-stroke-width': 2,
            'circle-stroke-color': circleColor,
            'circle-stroke-opacity': 1,
            'circle-opacity': 0.8,
        },
      };

    if (mapPreloaded) {
      map.addLayer(circleLayerObject);
    } else {
      map.on('load', () => map.addLayer(circleLayerObject));

      if (config.includeTooltips) {
        const textFn = getTextOrParamFn(config.columns, layerId);
        const paramsFn = getTextOrParamFn(config.columns, layerId, true);
        createGLTooltips(layerId, textFn, map, paramsFn);
      }
    }

    const hasAreaLegend = $(`#${mapDiv}`).find('.circle_area_legend').length > 0;
    if (circleLegendArea.value() === 'true' && aggCol && !setRadius && !hasAreaLegend) {
      createGLSizeLegend(config, circleZoomLevels[0], circleZoomLevels[1]);
    } else if ((circleLegendArea.value() !== 'true' || !aggCol) && hasAreaLegend) {
      removeGLSizeLegend();
    }

    const hasCircleLegend = $(`#${mapDiv}`).find('.circle_color_legend').length > 0;
    if (circleLegendColor.value() === 'true' && colorCol && !setRadius && !hasCircleLegend) {
      createGLColorLegend(config, circleZoomLevels[0], circleZoomLevels[1]);
    } else if ((circleLegendColor.value() !== 'true' || !colorCol) && hasCircleLegend) {
      removeGLColorLegend();
    }
  }

  function createGLSymbols(config, customImage, sourceId, layerId) {
    sourceId = sourceId || sourceIdObject.main;
    layerId = layerId || layerIdObject.pins;

    if (map.getLayer(layerId)) {
      return;
    }

    const iconSize = Number(customIconSize.value()) / 100;
    const symbolIconObject = {
      id: layerId,
      type: 'symbol',
      source: sourceId,
      layout: {
        'icon-image': customIconId,
        'icon-size': iconSize,
        'icon-allow-overlap': true,
      },
    };

    const addImageAndIcons = function() {
      if (map.hasImage(customIconId)) {
        map.addLayer(symbolIconObject);
      } else {
        map.loadImage(customImage, function(error, image) {
          if (error) {
            lib.alertMessage('Error: custom icon could not load');
            return;
          }

          if (!map.hasImage(customIconId)) {
            map.addImage(customIconId, image);
          }

          map.addLayer(symbolIconObject);
        });
      }
    };

    if (mapPreloaded) {
      addImageAndIcons();

      if (config.includeTooltips) {
        const textFn = getTextOrParamFn(config.columns, layerId);
        const paramsFn = getTextOrParamFn(config.columns, layerId, true);
        createGLTooltips(layerId, textFn, map, paramsFn);
      }
    } else {
      map.on('load', addImageAndIcons);

      if (config.includeTooltips) {
        const textFn = getTextOrParamFn(config.columns, layerId);
        const paramsFn = getTextOrParamFn(config.columns, layerId, true);
        createGLTooltips(layerId, textFn, map, paramsFn);
      }
    }
  }

  function createChoroplethFilter(config, layerId) {
    const choroLayer = map.getLayer(layerId);
    if (!choroLayer) {
      lib.alertMessage('Error: Invalid choropleth Layer ID specified in Settings -> Choropleth');
      return;
    }

    const renderedFeatures = map.queryRenderedFeatures({ layers: [layerId] });
    if (renderedFeatures.length === 0) {
      return;
    }

    const filteredFeatureArrays = [[]];
    let filteredArrayCount = 0;
    renderedFeatures.forEach(function(rf) {
      let filteredRenderedFeatures = filteredFeatureArrays[filteredArrayCount];
      if (filteredRenderedFeatures.length >= 10000) {
        filteredArrayCount++;
        filteredRenderedFeatures = filteredFeatureArrays[filteredArrayCount];
      }

      if (!(choroplethDataHash[rf.id])) {
        if (choroplethRenderedFeatures[rf.id]) {
          choroplethRenderedFeatures[rf.id]++;
        } else {
          filteredRenderedFeatures.push(rf);
          choroplethRenderedFeatures[rf.id] = 1;
        }
      }
    });

    if (filteredFeatureArrays[0].length === 0) {
      return;
    }

    let visibleIdArrays = [];
    if (config.choroCol.coltype() === 'STRING') {
      filteredFeatureArrays.forEach(function(filterdArray) {
        visibleIdArrays.push(filterdArray.map(feature => `"${(feature.id)}"`));
      });
    } else {
      filteredFeatureArrays.forEach(function(filterdArray) {
        visibleIdArrays.push(filterdArray.map(feature => feature.id));
      });
    }

    const colName = config.choroCol.colname();
    let filterStrings = visibleIdArrays.map(ids => `[${colName}] in (${ids.join(',')})`);
    let filterString = filterStrings.join(' and ');
    return filterString;
  }

  function updateChoroplethOpacity() {
    const layerId = choroplethLayerId.value();
    const layerOpacity = Number(choroplethOpacity.value());
    if (layerId === '' || !map.getLayer(layerId)) {
      return;
    }

    map.setPaintProperty(layerId, 'fill-opacity', layerOpacity);
  }

  function paintChoropleth(layerId, rowsData, config) {
    const choroLayer = map.getLayer(layerId);
    let sourceLayerId, sourceId;

    if (!choroLayer) {
      lib.alertMessage('Error: Invalid choropleth Layer ID specified in Settings -> Choropleth');
      return;
    } else {
      sourceLayerId = choroLayer.sourceLayer;
      sourceId = choroLayer.source;
    }

    const layerData = {
      layerId: layerId,
      sourceLayerId: sourceLayerId,
      sourceId: sourceId,
    };

    try {
      paintViewPort(rowsData, layerData, config);
    } catch (err) {
      console.log(err);
      return false;
    }

    return 'painted';
  }

  function createChoropleth(config) {
    const layerId = choroplethLayerId.value();
    let rowData = config.rows.slice();

    const updateChoropleth = () => {
      const updateFilter = createChoroplethFilter(config, layerId);
      if (updateFilter) {
        choroplethFilter = updateFilter;
        preserveMap = true;
        lib.refresh(true);
      }
    };

    if (map && map.style.loaded()) {
      paintChoropleth(layerId, rowData, config);
      updateChoropleth();
      updateChoroplethOpacity();
    } else {
      map.on('load', () => {
        paintChoropleth(layerId, rowData, config);
        updateChoropleth();
        updateChoroplethOpacity();
      });
      map.on('moveend', updateChoropleth);
      map.on('zoomend', updateChoropleth);

      if (config.includeTooltips) {
        const textFn = getTextOrParamFn(config.columns, layerId, undefined, true);
        const paramsFn = getTextOrParamFn(config.columns, layerId, true, true);
        createGLTooltips(layerId, textFn, map, paramsFn);
      }
    }
  }

  function paintViewPort(inData, layerData, config) {
    let numberOfInvalidShapeIds = 0;
    function updateColorInfo(row) {
      let color = row[colorCol.colindex()];
      if (colorCol) {
        if (colorIsNumeric) {
          if (color === undefined || arc.misc.check_if_value_null(color)) {
            color = 0;
          }

          color = Number(color);
          updateNumericRange(color, colorRange);
        }

        if (colorIsDim) {
          colorDomain[color] = true;
        }
      }
    }

    function updateFeatureState(row) {
      const colorInfo = colorCol ? row[colorCol.colindex()] : null;
      associatedData = {
        color: colorInfo,
        row: row,
      };

      const shapeId = row[config.choroCol.colindex()];
      if (!arc.misc.isNumeric(shapeId) || Number(shapeId) < 0) {
        numberOfInvalidShapeIds += 1;
        return;
      }

      map.setFeatureState({ source: layerData.sourceId, sourceLayer: layerData.sourceLayerId,
        id: row[config.choroCol.colindex()], }, associatedData);
      choroplethDataHash[row[config.choroCol.colindex()]] = true;
    }

    if (colorCol) {
      inData.forEach(updateColorInfo);
    }

    let associatedData, colorProperty;
    if (colorCol) {
      colorProperty = makeColorProperty(config, get_color_fn(config, true), true);
    } else {
      colorProperty = 'rgba(0,0,0,0)';
    }

    map.setPaintProperty(layerData.layerId, 'fill-color', colorProperty);

    inData.forEach(updateFeatureState);
    choroplethFilter = null;
    preserveMap = false;

    if (includeChoroplethLegend.value() === 'true' && !config.drawnColorLegend && colorCol) {
      createGLColorLegend(config, undefined, undefined, true);
    }

    if (numberOfInvalidShapeIds > 0) {
      lib.alertMessage(`Ignoring ${numberOfInvalidShapeIds} invalid shape id(s)`, undefined,
        { noConsoleLog: true });
    }
  }

  function makeColorProperty(config, colorFn, isChoropleth) {
    let colorProperty;
    if (isChoropleth && !colorCol) {
      colorCol = config.choroCol;
    }

    if (colorIsNumeric) {
      let colorInterpolation;
      if (isChoropleth) {
        colorProperty = ['case', ['!=', ['feature-state', 'color'], null]];
        colorInterpolation = ['interpolate', ['linear'], ['to-number', ['feature-state', 'color']]];
      } else {
        colorProperty = ['case', ['!=', ['get', 'color'], null]];
        colorInterpolation = ['interpolate', ['linear'], ['get', 'color']];
      }

      const colorStep = (colorRange[1] - colorRange[0]) / (lib.colors().length - 1);
      for (let colVal = colorRange[0]; colVal < colorRange[1]; colVal += colorStep) {
        colorInterpolation.push(colVal, colorFn(colVal));
      }

      colorInterpolation.push(colorRange[1], colorFn(colorRange[1]));
      colorProperty.push(colorInterpolation);
      colorProperty.push('rgba(0, 0, 0, 0)');

    } else if (colorCol) {
      if (isChoropleth) {
        colorProperty = ['match', ['feature-state', 'color']];
      } else {
        colorProperty = ['match', ['get', 'color']];
      }

      Object.keys(colorDomain).forEach(color => colorProperty.push(color, colorFn(color)));

      // If there is no match
      colorProperty.push('rgba(0, 0, 0, 0)');
    } else {
      colorProperty = arc.colorpalette.convertToRGBAString(lib.colors()[0]);
    }

    return colorProperty;
  }

  function createGLColorLegend(config, minZoom, maxZoom, isChoropleth) {
    if (!colorCol) {
      return;
    }

    const colorFn = get_color_fn(config, isChoropleth);
    const legendExtraHorizontalMargin = 50;
    const legendTitle = colorCol.colname();

    if (Object.keys(colorDomain).length > 200) {
      lib.alertMessage('Error: legend failed to load, too many values');
      return;
    }

    if ($(`#${mapDiv}`).find('.circle_color_legend').length > 0) {
      $(`#${mapDiv}`).find('.circle_color_legend').remove();
    }

    let colorLegendConfig;
    let correctLegendVals = [];
    let colorMappings = {};
    if (colorIsDim) {
      correctLegendVals = Object.keys(colorDomain);
      correctLegendVals.forEach((v) => colorMappings[v] = colorFn(v));
    } else {
      correctLegendVals = colorRange;
    }

    let legendWidth = getLegendWidth(correctLegendVals, legendExtraHorizontalMargin, legendTitle);
    const side = $(`#${mapDiv}`).find('.circle_area_legend').length > 0 ? 'left' : 'right';
    const d3MapDiv = d3.select(`#${mapDiv}`);

    d3MapDiv.append('div').attr('class', 'info legend circle_color_legend')
                       .style('position', 'absolute')
                       .style(side, '10px')
                       .style('bottom', '20px')
         .append('svg').attr('width', legendWidth)
                       .attr('class', 'arc-viz-legend');

    let colorSvg = d3MapDiv.select('div.circle_color_legend')
                           .select('.arc-viz-legend');
    if (colorIsDim) {
      const colorScale = arc.misc.dimensionColorScale().init({}, colorMappings);
      colorLegendConfig = {
        legendType: 'discrete',
        legendPositioning: 'Right',
        svg: colorSvg,
        width: 0,
        color: colorScale,
        legends: correctLegendVals,
        legend_title: legendTitle,
        sort_domain: false,
        legend_width: legendWidth,
      };
    } else {
      colorLegendConfig = {
        legendType: 'continuous',
        legendPositioning: 'Right',
        svg: colorSvg,
        width: 0,
        color: colorFn,
        legend_title: legendTitle,
        sort_domain: false,
        legend_width: legendWidth,
        legend_column: colorCol,
        color_palette: lib.colors(),
      };
    }

    lib.draw_legend(colorLegendConfig);

    if (maxZoom && minZoom) {
      map.on('load', function() {
        hideAndShowLegend(map.getZoom(), minZoom, maxZoom, '.circle_color_legend');
        map.on('zoomend', function() {
          hideAndShowLegend(map.getZoom(), minZoom, maxZoom, '.circle_color_legend');
        });
      });
    }
  }

  function removeGLColorLegend() {
    $(`#${mapDiv}`).find('.circle_color_legend').remove();
  }

  function createGLSizeLegend(config, minZoom, maxZoom) {
    const radiusRange = circleRadiusRange.get_range();
    const midValue = (aggRange[0] + aggRange[1]) / 2;
    let labels = arc.misc.arrayUnique([aggRange[0], midValue, aggRange[1]]);
    const gapSpacing = 10;
    const smallR = radiusRange[0];
    const largeR = radiusRange[1];
    let legendWidth = getLegendWidth(labels, 2 * (largeR + gapSpacing), aggCol.colname());
    const titleSpace = 25;
    const extraSpacing = (largeR - smallR) / 2;
    const location = $(`#${mapDiv}`).find('.circle_color_legend').length > 0 ? 'left' : 'right';
    const legendHeight = ((largeR + smallR) * 3) + titleSpace + gapSpacing + extraSpacing;
    const d3MapDiv = d3.select(`#${mapDiv}`);
    d3MapDiv.append('div').attr('class', 'info legend circle_area_legend')
                           .style('position', 'absolute')
                           .style(location, '10px')
                           .style('bottom', '20px')
                           .style('height', `${legendHeight}px`)
             .append('svg').attr('width', `${legendWidth}px`)
                           .attr('class', 'arc-viz-legend');

    let areaSvg = d3MapDiv.select('div.circle_area_legend')
                          .select('.arc-viz-legend');

    const areaRange = [Math.PI * smallR * smallR, Math.PI * largeR * largeR];
    const areaScale = d3.scale.linear().range(areaRange).domain(aggRange);

    let area_legend_config = {
      legendType: 'area',
      legendPositioning: 'Right',
      svg: areaSvg,
      width: 0,
      legends: labels,
      legend_column: aggCol,
      legend_width: legendWidth,
      area_scale: areaScale,
      legend_title: aggCol.colname(),
    };

    lib.draw_legend(area_legend_config);

    if (minZoom && maxZoom) {
      map.on('load', function() {
        hideAndShowLegend(map.getZoom(), minZoom, maxZoom, '.circle_area_legend');
        map.on('zoomend', function() {
          hideAndShowLegend(map.getZoom(), minZoom, maxZoom, '.circle_area_legend');
        });
      });
    }
  }

  function removeGLSizeLegend() {
    $(`#${mapDiv}`).find('.circle_area_legend').remove();
  }

  function hideAndShowLegend(zoomLevel, minZoom, maxZoom, idString) {
    if (zoomLevel < minZoom || zoomLevel > maxZoom) {
      $vizDiv.find(idString).addClass('arc-invisible');
    } else {
      $vizDiv.find(idString).removeClass('arc-invisible');
    }
  }

  let get_color_fn = function(config, isChoropleth) {
    const colors = lib.colors();
    const colorSettings = lib.colorSettings();
    let colorScale;

    if (!colorCol) {
      return function() {
        return colors[0];
      };
    } else if (colorIsDim) {
      colorScale = d3.scale.ordinal().range(colors).domain(Object.keys(colorDomain));

      return function (colorVal) {
        return arc.colorpalette.convertToRGBAString(colorScale(colorVal));
      };
    } else {
      const colorFunc = arc.colorpalette.aggScale(colorRange, colors, colorSettings);
      colorScale = function(aggVal) {
        return arc.colorpalette.convertToRGBAString(colorFunc(aggVal));
      };

      colorScale.domain = () => colorFunc.domain();
      return colorScale;
    }
  };

  function getTextOrParamFn(cols, layerId, isParamFn, inFeatureState) {
    const textOrParamsFn = function (e) {
      let features = map.queryRenderedFeatures(e.point, { layers: [layerId] });
      if (!features.length) {
        return;
      }

      let feature = features[0];
      let rowData;
      if (inFeatureState) {
        rowData = feature.state.row;
      } else {
        rowData = feature.properties.row;

        try {
          rowData = JSON.parse(rowData);
        } catch (err) {
          rowData = [];
        }

        rowData = rowData.map(val => val[0] === '"' ? val.slice(1, val.length - 1) : val);
      }

      if (rowData && rowData.length > 0) {
        if (isParamFn) {
          return lib.params_from_columns(rowData, cols);
        } else {
          return lib.text_from_columns(rowData, cols);
        }
      } else {
        return '';
      }
    };

    return textOrParamsFn;
  }

  function getLegendWidth(labels, extraMargin, legendTitle = '') {
    let legendWidth;
    let titleMargin = 10;
    const maxLegendWidth = 250;
    const legendFontSize = lib.getLegendFontSize() ? lib.getLegendFontSize() : 10;
    const titleWidth = ((legendFontSize / 1.7) * legendTitle.length) + titleMargin;
    let longestLabel = 0;
    labels.forEach(label => longestLabel = Math.max(longestLabel, String(label).length));
    const nonTitleWidth = extraMargin + ((legendFontSize / 1.7) * longestLabel);
    legendWidth = Math.max(nonTitleWidth, titleWidth);
    if (legendWidth > maxLegendWidth) {
      legendWidth = maxLegendWidth;
    }

    return legendWidth;
  }

  leaflet.collateral_is_available = function() {
    if (mapType === 'zipcode') {
      if (zipcodeLatLong) {
        return true;
      }

      return false;
    }

    //normal lat-long map does not need any other info
    return true;
  };

  leaflet.fetch_collateral = function(cb) {
    const map_base_url = arc.frontend.globals.urls.mapBaseUrl;
    var zipcodes_asset = {
      asset: map_base_url + '/United States/zipcodes.json',
      postprocess: d3.json,
    };
    if (mapType === 'zipcode') {
      arc.staticfetcher.fetch([zipcodes_asset],
          function(error, _zipcodeLatLong) {
            if (error) {
              lib.alertMessage('Error fetching map metadata');
              return;
            }

            zipcodeLatLong = _zipcodeLatLong;
            cb();
          });

      return leaflet;
    }

    return leaflet;
  };

  leaflet.viz_destroy = function() {
    // The Google Maps plugin leaks memory right now. Periodically check to see
    // if there is an upgrade that allows fewer memory leaks.
    if (map) {
      map.remove();
    }
  };

  return leaflet;
};

arc.visual.plugins.leaflet_extension = function() {
  var ext = function () {
    return 'arc.visual.plugins.leaflet_extension';
  };

  ext.viz_name = function() {
    return 'leaflet';
  };

  ext.viz_tooltip = function() {
    return 'interactive map';
  };

  ext.viz = arc.visual.plugins.leaflet;
  return ext;
};

arc.visual.extensions.push(arc.visual.plugins.leaflet_extension());

}());
