window.arc = window.arc || {};

(function() {
'use strict';

arc.canvas = arc.canvas || {};

let pixelRatio;

arc.canvas.highlightLineWidth = 2;
arc.canvas.highlightStrokeStyle = '#999';

arc.canvas.colorFetcher = function() {
  let _nextColor = 1;

  function nextColor() {
    const ret = [];
    if (_nextColor < 16777215) {
      // eslint-disable-next-line no-bitwise
      ret.push(_nextColor & 0xff); // R
      // eslint-disable-next-line no-bitwise
      ret.push((_nextColor & 0xff00) >> 8); // G
      // eslint-disable-next-line no-bitwise
      ret.push((_nextColor & 0xff0000) >> 16); // B
      _nextColor += 2;
    }

    const col = `rgb(${ret.join(',')})`;
    return col;
  }

  return nextColor;
};

arc.canvas.pixelRatio = function() {
  if (pixelRatio) {
    return pixelRatio;
  }

  const ctx = document.createElement('canvas').getContext('2d');
  let dpr = window.devicePixelRatio || 1;
  const bsr = ctx.webkitBackingStorePixelRatio ||
            ctx.mozBackingStorePixelRatio ||
            ctx.msBackingStorePixelRatio ||
            ctx.oBackingStorePixelRatio ||
            ctx.backingStorePixelRatio || 1;

  if (arc.testutil && arc.testutil.testMode) {
    // This is different for Jenkins vs when running tests on Chrome locally
    dpr = 1;
  }

  pixelRatio = dpr / bsr;
  return pixelRatio;
};

arc.canvas.createContainer = function(cell_sel, cellConfig) {
  const {
    width,
    height,
  } = cellConfig;

  cellConfig.overlayContainer =
      cell_sel
          .append('div')
          .classed('arc-viz-overlay-container', true)
          .style('position', 'relative')
          .style('width', `${width}px`)
          .style('height', `${height}px`);
};

arc.canvas.axisOverlay = function(config) {
  const {
    overlayContainer,
    width,
    height,
    margin,
  } = config;

  overlayContainer.selectAll('.overlay-axis')
                  .remove();

  const svg = overlayContainer.append('svg')
                        .classed('overlay-axis', true)
                        .style('position', 'absolute')
                        .style('left', '0px')
                        .style('top', '0px')
                        .attr('width', `${width}px`)
                        .attr('height', `${height}px`)
                        .append('g')
                        .attr('transform', `translate(${margin.left},${margin.top})`);

  return svg;
};

arc.canvas.drawAllCanvas = function(config, cellConfig, drawCanvas, inTransition) {
  // Draw the main canvas...
  drawCanvas(config, cellConfig, cellConfig.mainCanvasConfig);

  if (!inTransition) {
    // ...the hidden canvas
    drawCanvas(config, cellConfig, cellConfig.hiddenCanvasConfig);

    // ...and the highlight canvas
    drawCanvas(config, cellConfig, cellConfig.highlightCanvasConfig);
  }
};

arc.canvas.createAllCanvas = function(cellConfig) {
  const {
    overlayContainer,
    width,
    height,
    margin,
    mainCanvasOnly,
  } = cellConfig;

  cellConfig.mainCanvasConfig = {};

  cellConfig.hiddenCanvasConfig = {
    hidden: true,
  };

  cellConfig.highlightCanvasConfig = {
    highlight: true,
  };

  // Create the main canvas...
  cellConfig.mainCanvasConfig.canvas = createCanvas({
    overlayContainer,
    width,
    height,
    margin,
  });
  cellConfig.mainCanvasConfig.canvas.classed('maincanvas', true);

  if (mainCanvasOnly) {
    return;
  }

  // ... the hidden canvas
  cellConfig.hiddenCanvasConfig.canvas = createCanvas({
    overlayContainer,
    width,
    height,
    margin,
    hidden: true,
  });
  cellConfig.hiddenCanvasConfig.canvas.classed('hiddencanvas', true);

  // ... and finally, the highlight canvas
  // We keep the highlight canvas separate so it's easy to clear
  // it out when we want to end the highlight
  cellConfig.highlightCanvasConfig.canvas = createCanvas({
    overlayContainer,
    width,
    height,
    margin,
  });
  cellConfig.highlightCanvasConfig.canvas.classed('highlightcanvas', true);
};

arc.canvas.adjustPixelRatio = function(canvasSel, width, height) {
  const c = canvasSel.node();
  const ratio = arc.canvas.pixelRatio();

  const maxWidth = arc.misc.isNumeric(arc.globals.app_settings.maximumCanvasWidth) ?
    Number(arc.globals.app_settings.maximumCanvasWidth) : 8000;
  const maxHeight = 8000;

  if (height > maxHeight) {
    console.log('truncating canvas height to 8k pixels');
  }

  if (width > maxWidth) {
    console.log(`truncating canvas width to ${Math.floor(maxWidth / 1000)}k pixels`);
  }

  height = Math.min(height, maxHeight);
  width = Math.min(width, maxWidth);

  c.width = width * ratio;
  c.height = height * ratio;
  c.style.width = `${width}px`;
  c.style.height = `${height}px`;

  const context = c.getContext('2d');
  context.setTransform(ratio, 0, 0, ratio, 0, 0);
};

function createCanvas(config) {
  const {
    overlayContainer,
    hidden,
    width,
    height,
    margin,
  } = config;

  const canvas = overlayContainer.append('canvas')
                              .style('position', 'absolute')
                              .style('left', '0px')
                              .style('top', '0px');

  if (hidden) {
    canvas.style('display', 'none');
  }

  arc.canvas.adjustPixelRatio(canvas, width, height);

  const c = canvas.node();
  const context = c.getContext('2d');
  context.translate(margin.left, margin.top);

  return canvas;
}

arc.canvas.handleTooltips = function(tooltipConfig) {
  const {
    mouseoverCanvas,
    hiddenCanvas,
    hiddenColorMap,
    permitPivot,
    text_fn,
    params_fn,
    highlight_fn,
    viz_lib,
    mark_type,
  } = tooltipConfig;

  const hiddenContext = hiddenCanvas.node().getContext('2d');

  function pos_to_row(elem) {
    if (!elem) {
      return undefined;
    }

    const canvasOffset = $(elem).offset();
    if (!canvasOffset) {
      return undefined;
    }

    // eslint-disable-next-line no-undef
    const evt = d4.event || d3.event;
    if (!evt) {
      return undefined;
    }

    let mouseX = evt.pageX - canvasOffset.left;
    let mouseY = evt.pageY - canvasOffset.top;

    const ratio = arc.canvas.pixelRatio();

    mouseX *= ratio;
    mouseY *= ratio;

    const t1 = new Date();
    const color = hiddenContext.getImageData(mouseX, mouseY, 1, 1).data;
    const t2 = new Date();
    if (t2 - t1 > 100) {
      console.log(`slow getImageData ${(t2 - t1)} ms`);
    }

    const colorKey = `rgb(${color[0]},${color[1]},${color[2]})`;
    const nodeData = hiddenColorMap[colorKey];

    if (nodeData) {
      return nodeData;
    }

    return undefined;
  }

  function canvas_text_fn(d, i, elem) {
    const r = pos_to_row(elem);
    if (!r) {
      return undefined;
    }

    return text_fn(r);
  }

  function canvas_show_highlight_fn(contextmenu, active_paramsl, elem) {
    return highlight(elem);
  }

  function canvas_hide_highlight_fn() {
    highlight_fn();
  }

  function canvas_highlight_fn(d, i, elem) {
    return highlight(elem);
  }

  function highlight(elem) {
    const row = pos_to_row(elem);

    if (!row) {
      highlight_fn();

      return undefined;
    }

    return highlight_fn(row);
  }

  function canvas_params_fn(d, i, elem, dimsOnly) {
    const r = pos_to_row(elem);
    if (!r) {
      return undefined;
    }

    return params_fn(r, dimsOnly);
  }

  const marks =
    arc.marks()
       .viz_lib(viz_lib)
       .mark_type(mark_type)
       .permitPivot(permitPivot)
       .marker_name('primary')
       .addSelection(mouseoverCanvas)
       .text_changes_on_move(true)
       .text_fn(canvas_text_fn)
       .params_fn(canvas_params_fn)
       .tooltip_highlight_cb(canvas_highlight_fn)
       .highlight_cb(canvas_highlight_fn)
       .cm_hide_cb(canvas_hide_highlight_fn)
       .cm_show_cb(canvas_show_highlight_fn);

  marks.activate();
  return marks;
};
}());
