import colorbrewer from 'colorbrewer';

window.arc = window.arc || {};

(function() {
'use strict';

arc.colorpalette = arc.colorpalette || {};

function getSwatchHeight(colors) {
  const maxPaletteSize = 200;
  const defaultSwatchHeight = 12;

  if (!colors || colors.length === 0) {
    return defaultSwatchHeight;
  }

  const swatchHeight = maxPaletteSize / colors.length;
  if (swatchHeight > defaultSwatchHeight) {
    return defaultSwatchHeight;
  }

  return swatchHeight;
}

arc.colorpalette.getOrderedColorPalette = function(bgColor) {
  let colors;
  colors = bgColor.colors.slice();
  if (bgColor.isReversed) {
    colors.reverse();
  }

  colors = arc.colorpalette.modifyBasedOnFirstColor(colors, bgColor.firstColor);
  return colors;
};

arc.colorpalette.modifyBasedOnFirstColor = function(colors, firstColor) {
  if (!firstColor) {
    return colors;
  }

  const startingPoint = colors.indexOf(firstColor);
  if (startingPoint <= 0) {
    return colors;
  }

  const c = colors.slice(startingPoint);
  c.push(...colors.slice(0, startingPoint));

  return c;
};

arc.colorpalette.setting = function() {
  const defaultFirstColorPalette = ['#0097a7', '#7dcad2', '#546e7a', '#b0bec5', '#ed8111',
        '#f6c088', '#e54954', '#f2a4a9', '#ad73b1', '#d6b9d8', '#2375b8',
        '#91badb', '#3e8111', '#b0cb73',
  ];

  let savedColors;
  let savedColorSettings;
  let savedIsReversed;
  let allColors;
  let selectedColors;
  let firstColor = 'white';
  // eslint-disable-next-line no-unused-vars
  let paletteContainer;

  const setting = arc.visual.setting()
                   .id('color_palette')
                   .section('Style')
                   .tab('Color')
                   .type(arc.visual.settingTypeEnum.CUSTOM)
                   .add_name('COLOR_PALETTE')
                   .value(fetch_fn()); // setting is used in fetch_fn() -> getColors()

  setting.selectedColors = function(_) {
    if (!arguments.length) return selectedColors;
    selectedColors = _;
    return setting;
  };

  setting.savedIsReversed = function(_) {
    if (!arguments.length) return savedIsReversed;
    savedIsReversed = _;
    return setting;
  };

  function setFirstColor(_color) {
    let color = _color;

    if (color === undefined) {
      if (savedColors && savedColors.length > 0) {
        color = savedColors[0];
      } else {
        color = 'white';
      }
    }

    firstColor = color;
    const $firstColor = $(`#${setting.div_id()}`).find('.first-color');
    $firstColor.css('background-color', color);
  }

  const msg = 'To set this value, click on a palette below to select it first,' +
      ' and then click on the individual color swatches within the selected ' +
      'color palettes.<br><br>For visuals that use only one color, this is the ' +
      'color that will be used. For visuals that use categorical colors, the ' +
      'palette starts at this color and wraps around. For visuals that use only ' +
      'graded colors, this has no effect.';

  const checkbox_template = `
    <div class="form-check color-palettes">
      <input class="form-check-input" id="colorPalettes" type="checkbox" data-test-id="reverse-color-order-checkbox">
      <label class="form-check-label" for="colorPalettes">
        Reverse color order
        <span class="info-icon" data-toggle="popover"
              data-html="true" data-placement="bottom" data-container="body"
              data-content="Use this checkbox to reverse the order of colors in the palettes below">
        </span>
      </label>
      <div>
        <span class="first-color"></span>
        <span>First color within palette
          <span class="info-icon" data-toggle="popover"
                data-html="true" data-placement="bottom" data-container="body"
                data-content="${msg}">
          </span>
        </span>
      </div>
      <div class="colorpalettes-setting" data-test-id="colorpalettes-setting"></div>
    </div>
  `;

  setting.disable_fn(() => {
    $(`#${setting.div_id()}`).find('.settings-palette').css('pointer-events', 'none');
    $(`#${setting.div_id()}`).find('input').prop('disabled', true);
  });

  setting.enable_fn(() => {
    $(`#${setting.div_id()}`).find('.settings-palette').css('pointer-events', '');
    $(`#${setting.div_id()}`).find('input').prop('disabled', false);
  });

  setting.activeColors = function() {
    return savedColors;
  };

  setting.allColors = function() {
    getAllColors();
    return allColors;
  };

  setting.saveIntoVisual = function() {
    const data = JSON.parse(setting.value());
    const colors = data.colors || [];
    const colorSettings = data.colorSettings || {};
    const isReversed = data.isReversed || false;
    const lib = setting.viz_lib();

    if (lib && colors.length > 0) {
      let gradedColors = colors.slice(0);
      if (isReversed) {
        gradedColors = gradedColors.reverse();
      }

      const contrastColors = arc.colorpalette.modifyBasedOnFirstColor(
          gradedColors, data.firstColor
        );

      lib.reverse_colors(isReversed);
      lib.colorSettings(colorSettings);

      lib.colors(contrastColors);
      if (this.id() === 'link_color_palette') {
        lib.linkColors(contrastColors);
        return;
      }

      if (lib.colors_could_be_graded_or_contrast()) {
        lib.contrast_colors(contrastColors);
        lib.graded_colors(gradedColors);
        return;
      }

      if (lib.chart_type_graded() === 0) {
        lib.contrast_colors(contrastColors);
        return;
      }

      if (lib.chart_type_graded()) {
        lib.graded_colors(gradedColors);
      }
    }
  };

  let opacity = 100;
  setting.setOpacity = function(v) {
    opacity = Number(v);
    drawOpacity();
  };

  function drawOpacity() {
    $(`#${setting.div_id()}`)
        .find('.settings-swatch, .first-color')
        .css('opacity', opacity / 100);
  }

  function convertToSwatches(colors) {
    return {
      colors,
      colorSettings: {},
      swatchHeight: getSwatchHeight(colors),
    };
  }

  function getAllColors() {
    if (allColors !== undefined) {
      return;
    }

    allColors = [];
    allColors.push(convertToSwatches(defaultFirstColorPalette));
    allColors.push(convertToSwatches(['#0097a7', '#546e7a', '#ed8111', '#e54954', '#ad73b1',
        '#2375b8', '#3e8111', '#b0cb73',
    ]));
    allColors.push(convertToSwatches(['#E0F7FA', '#B2EBF2', '#80DEEA', '#4DD0E1', '#26C6DA',
        '#00BCD4', '#00ACC1', '#0097A7', '#00838F', '#006064',
    ]));
    allColors.push(convertToSwatches(['#eceff1', '#cfd8dc', '#b0bec5', '#90a4ae', '#78909c',
        '#607d8b', '#546e7a', '#455a64', '#37474f', '#263238',
    ]));
    allColors.push(convertToSwatches(['#feffdb', '#e4f4d4', '#bee5cd', '#8bd0c2', '#65c1bb',
        '#32acb1', '#0097a7', '#00838f', '#006064',
    ]));
    allColors.push(convertToSwatches(['#1f77b4', '#aec7e8', '#ff7f0e', '#ffbb78', '#2ca02c',
        '#98df8a', '#d62728', '#ff9896', '#9467bd', '#c5b0d5', '#8c564b',
        '#c49c94', '#e377c2', '#f7b6d2', '#7f7f7f', '#c7c7c7', '#bcbd22',
        '#dbdb8d', '#17becf', '#9edae5',
    ]));
    allColors.push(convertToSwatches(arc.report.contrast_colors()));
    allColors.push(convertToSwatches(['#393b79', '#5254a3', '#6b6ecf', '#9c9ede', '#637939',
        '#8ca252', '#b5cf6b', '#cedb9c', '#8c6d31', '#bd9e39', '#e7ba52',
        '#e7cb94', '#843c39', '#ad494a', '#d6616b', '#e7969c', '#7b4173',
        '#a55194', '#ce6dbd', '#de9ed6',
    ]));
    allColors.push(convertToSwatches(['#3182bd', '#6baed6', '#9ecae1', '#c6dbef', '#e6550d',
        '#fd8d3c', '#fdae6b', '#fdd0a2', '#31a354', '#74c476', '#a1d99b',
        '#c7e9c0', '#756bb1', '#9e9ac8', '#bcbddc', '#dadaeb', '#636363',
        '#969696', '#bdbdbd', '#d9d9d9',
    ]));
    allColors.push(convertToSwatches(['#d73027', '#fed976', '#1a9850',
    ]));
    allColors.push(convertToSwatches(['#d73027', '#1a9850',
    ]));
    allColors.push(convertToSwatches([
          '#133450', '#2976b6', '#469870', '#B1CA78', '#E34B57', '#EB8128',
          '#AC75B0', '#D4DCD8',
    ]));

  [
    'Spectral', 'RdYlGn', 'Accent', 'Dark2',
    'Paired', 'Pastel1', 'Pastel2', 'Set1', 'Set2', 'Set3',
    'YlGn', 'YlGnBu', 'GnBu', 'BuGn', 'PuBuGn', 'PuBu', 'BuPu', 'RdPu',
    'PuRd', 'OrRd', 'YlOrRd', 'YlOrBr', 'Purples', 'Blues',
    'Greens', 'Oranges', 'Reds', 'Greys', 'PuOr', 'BrBG', 'PRGn', 'PiYG',
    'RdBu', 'RdGy', 'RdYlBu',
  ].forEach(
    (schemeName) => {
      const colorScheme = colorbrewer[schemeName];
      const mostNumberOfColorsInScheme = Math.max(...Object.keys(colorScheme));
      const colors = colorScheme[mostNumberOfColorsInScheme];
      const colorsAsSwatches = convertToSwatches(colors);

      allColors.push(colorsAsSwatches);
    }
  );

    allColors.push(convertToSwatches(arc.report.highcard_colors()));

    const site_colors = arc.colorpalette.get();
    site_colors.forEach((c) => {
      try {
        allColors.push({
          colors: c.dimColors,
          colorSettings: c.colorSettings,
          swatchHeight: c.swatchHeight,
          name: c.name,
        });
      } catch (err) {
        console.log('could not parse', c.value);
      }
    });
  }

  // Highlight the color palette that matches to last stored (just before refresh) palette
  setting.highlightLastStoredPalette = function() {
    const colorSetting = JSON.parse(setting.value()).colorSettings;
    const colorName = colorSetting.name;

    d3.selectAll('span.settings-palette')
      .classed('settings-color-selected', false)
      .filter((d) => arc.arraysIdentical(d.colors, savedColors ||
                      getColors()) && d.name === colorName)
      .each((d) => {
        savedColors = d.colors;
        setFirstColor(firstColor);
      })
      .classed('settings-color-selected', true);
  };

  function paint_fn(d3_sel) {
    paletteContainer = d3_sel;
    const div_id = setting.div_id();

    let $settingSection = $(`#${div_id}`);

    const $sel = $settingSection.find('.color-palettes');
    if ($sel.find('.settings-palette').length > 0) {
      $sel.remove();
    }

    d3_sel.append('div')
          .attr('id', div_id)
          .attr('data_setting_id', 'color_palette');
    $settingSection = $(`#${div_id}`);
    $settingSection.append(checkbox_template);

    const $input = $settingSection.find('input');
    $input.off('change')
          .on('change', function() {
      savedIsReversed = this.checked;
      setting.value(fetch_fn());
      setting.makeChangeCbs();
      paint_fn(d3_sel);
    });

    if (savedIsReversed) {
      $input.prop('checked', savedIsReversed);
    }

    getAllColors();

    const d3Colors =
      d3.select(`#${div_id} .colorpalettes-setting`)
        .selectAll('.nosel')
        .data(allColors)
        .enter()
          .append('span')
          .attr('class', 'settings-palette')
          .attr('data-toggle', 'tooltip')
          .attr('data-placement', 'top')
          .attr('title', (d) => d.name)
          .attr('data-test-id', 'settings-palette')
          .on('click', function(d) {
            const $this = $(this);
            const setClickedFirstColor = $this.hasClass('settings-color-selected');

            d3Colors.classed('settings-color-selected', false);
            $this.addClass('settings-color-selected');
            savedColors = d.colors;
            selectedColors = d.colors;
            savedColorSettings = d.colorSettings;

            if (setClickedFirstColor) {
              const $target = $(d3.event.target);
              if ($target.hasClass('settings-swatch')) {
                const color = $target.attr('data-color');
                setFirstColor(color);
              }
            } else {
              setFirstColor();
            }

            setting.value(fetch_fn());
            setting.makeChangeCbs();
          });

    d3Colors.selectAll('.settings-swatch')
            .data((d) => {
              const displayColors = d.colors.map((c) => [c, d.swatchHeight]);

              if (savedIsReversed) {
                displayColors.reverse();
              }

              return displayColors;
            })
            .enter()
              .append('span')
              .attr('class', 'settings-swatch')
              .style('height', (d) => `${d[1]}px`)
              .attr('data-color', (d) => d[0])
              .style('background-color', (d) => d[0]);

    setting.highlightLastStoredPalette();

    drawOpacity();

    return setting;
  }

  function getColors() {
    if (selectedColors) {
      return selectedColors;
    }

    let report_colors = defaultFirstColorPalette;
    const lib = setting && setting.viz_lib();

    if (lib) {
      if (setting.id() === 'link_color_palette') {
        report_colors = lib.linkColors();
      } else if (lib.chart_type_graded() === 0) {
        report_colors = lib.contrast_colors();
      } else if (lib.chart_type_graded()) {
        report_colors = lib.graded_colors();
      } else {
        report_colors = lib.colors();
      }
    }

    return report_colors;
  }

  function fetch_fn() {
    const data = {};
    data.colors = savedColors || getColors();
    data.colorSettings = savedColorSettings || {};
    data.isReversed = savedIsReversed || false;
    data.firstColor = firstColor;

    return JSON.stringify(data);
  }

  setting.loadColor = function() {
    const data = {};
    data.colorSettings = savedColorSettings || {};
    data.isReversed = savedIsReversed || false;
    data.firstColor = firstColor;

    const lib = setting.viz_lib();
    if (lib.chart_type_graded() === 0) {
      data.colors = lib.contrast_colors();
    } else if (lib.chart_type_graded()) {
      data.colors = lib.graded_colors();
    } else if (this.id() === 'link_color_palette') {
      data.colors = lib.linkColors();
    } else {
      data.colors = lib.colors();
    }

    if (data.isReversed) {
      data.colors = data.colors.reverse();
    }

    setting.value(JSON.stringify(data));
    return setting;
  };

  function valueCb(value) {
    const data = JSON.parse(value);
    savedColors = data.colors || [];
    savedColorSettings = data.colorSettings || {};
    savedIsReversed = data.isReversed || false;
    firstColor = data.firstColor;
    if (!firstColor) {
      setFirstColor();
    } else {
      setFirstColor(firstColor);
    }
  }

  setting.valueCb(valueCb)
         .paint_fn(paint_fn)
         .fetch_fn(fetch_fn)
         .repaint_fn(paint_fn);

  return setting;
};

arc.colorpalette.data = [];
arc.colorpalette.dataFetched = false;

function numProgression(start, end, step, max) {
  return Math.floor(start + (((end - start) / max) * step));
}

function rgbToHex(r, g, b) {
  // explanation of algo see: https://ryanclark.me/rgb-to-hex-via-binary-shifting/
  // eslint-disable-next-line no-bitwise
  return `#${((1 << 24) + (r << 16) + (g << 8) + b).toString(16).slice(1)}`;
}

// takes something of the form rgb(22, 23, 27) and returns [22, 23, 27].
// Also handles 'rgba(23,14,54,0.4)' => [23,14,45,0.4].
function rgbStringToRGB(rgbString) {
  const split = rgbString.split('(')[1].split(')')[0].split(',');
  return split.map((v) => parseInt(v.trim(), 10));
}

function rgbStringToHex(rgbString) {
  return rgbToHex(...rgbStringToRGB(rgbString));
}

function hexToRGB(hex) {
  hex = hex.trim();

  // Expand shorthand form (e.g. "03F") to full form (e.g. "0033FF")
  const shorthandRegex = /^#?([a-f\d])([a-f\d])([a-f\d])$/i;
  hex = hex.replace(shorthandRegex, (m, r, g, b) => r + r + g + g + b + b);

  const result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex);
  return result ? [
      parseInt(result[1], 16),
      parseInt(result[2], 16),
      parseInt(result[3], 16),
  ] : null;
}

function hexToRGBAString(hex, opacity) {
  const rgba = hexToRGB(hex);
  rgba.push(opacity);
  return `rgba(${rgba.join(',')})`;
}

function rgbStringToRGBAString(rgbString, opacity) {
  const rgba = rgbStringToRGB(rgbString);
  rgba.push(opacity);
  return `rgba(${rgba.join(',')})`;
}

function convertToRGBAString(color, opacity = 1) {
  if (!color) {
    console.log('invalid color');
    return 'rbga(0, 0, 0, 0)';
  }

  if (Array.isArray(color)) {
    if (color.length === 4) {
      return `rgba(${color.join(',')}`;
    }

    color.push(opacity);
    return `rgba(${color.join(',')}`;
  }

  if (color.includes('rgba')) {
    return color;
  }

  if (color.includes('rgb')) {
    return rgbStringToRGBAString(color, opacity);
  }

  return hexToRGBAString(color, opacity);
}

function computeColors(p) {
  let c;
  try {
    c = JSON.parse(p.value);
  } catch (e) {
    console.log(`Could not parse value: ${p.value} for the palette: `, p);
  }

  if (!(c && c.colors && c.colors.length === 3)) {
    if (c && c.colors && c.colors.length > 3) {
      // Use first 3 colors hence forth
      c.colors = c.colors.slice(0, 3);
    } else {
      return {
        dimColors: [],
        midColors: [],
        endColors: [],
      };
    }
  }

  const colorCount = 21;
  const rgbColors = [];
  for (let idx = 0; idx < 3; idx++) {
    if (beginsWithPoundSign(c.colors[idx].color)) {
       rgbColors.push(hexToRGB(c.colors[idx].color));
    } else {
      rgbColors.push(rgbStringToRGB(c.colors[idx].color));
    }
  }

  // position is between 0 and 100
  let mid = Math.floor((c.colors[1].position / 100) * colorCount);
  mid = Math.min(mid, colorCount - 1);
  mid = Math.max(mid, 0);

  const dimColors = [];
  let i;
  for (i = 0; i < mid; ++i) {
    dimColors.push(rgbToHex(
      numProgression(rgbColors[0][0], rgbColors[1][0], i, mid),
      numProgression(rgbColors[0][1], rgbColors[1][1], i, mid),
      numProgression(rgbColors[0][2], rgbColors[1][2], i, mid)
    ));
  }

  dimColors.push(rgbToHex(rgbColors[1][0], rgbColors[1][1], rgbColors[1][2]));

  const midColors = dimColors.slice(0);
  const endColors = [];
  const diff = colorCount - mid - 2;
  for (i = mid + 1; i < colorCount; ++i) {
    const nextColor = rgbToHex(
      numProgression(rgbColors[1][0], rgbColors[2][0], i - mid - 1, diff),
      numProgression(rgbColors[1][1], rgbColors[2][1], i - mid - 1, diff),
      numProgression(rgbColors[1][2], rgbColors[2][2], i - mid - 1, diff)
    );

    dimColors.push(nextColor);
    endColors.push(nextColor);
  }

  return {
    dimColors,
    midColors,
    endColors,
  };
}

arc.colorpalette.dimScale = function(domain, colors, colorSettings) {
  const colorScale = arc.misc.dimensionColorScale();
  colorScale.init(colorSettings.fixedDimensionColors, colorSettings.dimensionMappings)
            .domain(domain)
            .range(colors);

  return colorScale;
};

arc.colorpalette.aggScale = function(domain, colors, colorSettings) {
  const defaultColor = '#ffffff';

  let ranges;
  if (colorSettings && colorSettings.measureRanges) {
    ranges = colorSettings.measureRanges;
  }

  let colorScaleFunction;
  if (colorSettings && colorSettings.value && colorSettings.type !== 1) {
    const value = JSON.parse(colorSettings.value);

    // Doesn't make sense to support reverse in color palette with values
    let c;
    let split;
    let flipValues = false;

    if (colorSettings.type === 100) {
      const colorValues = JSON.parse(colorSettings.value).colors;
      const colorsWithValues = colorValues.filter((colorInfo) => colorInfo.value !== '');

      // if two or more of the colors have values associated with them, it does not make sense to
      // flip the palette. If one or no colors have values associated with them, then it makes sense
      // to flip the palette.
      if (colorsWithValues.length === 0) {
        split = Math.ceil(colors.length / 2);
      } else if (colorsWithValues.length === 1) {
        split = Math.ceil(colors.length / 2);
        if (colorValues[1].value === '') {
          const lowerColor = rgbStringToRGB(colorValues[0].color);
          if (lowerColor !== colors[0]) {
            // if there is only one value at either end then we need to flip the end values
            flipValues = true;
          }
        }
      } else {
        c = computeColors(colorSettings);
      }
    } else if (colorSettings.type === 102) {
      split = computeColors(colorSettings).midColors.length;
    } else {
      split = Math.floor(colors.length / 2);
    }

    if (!c) {
      c = {
        dimColors: colors,
        midColors: colors.slice(0, split),
        endColors: colors.slice(split),
      };
    }

    if (value && value.colors.length === 3) {
      let v0 = value.colors[0].value;
      let v1 = value.colors[1].value;
      let v2 = value.colors[2].value;

      if (flipValues) {
        [v0, v2] = [v2, v0];
      }

      if (v0 === undefined || v0 === '') {
        v0 = domain[0];
      } else {
        v0 = Number(v0);
      }

      if (v2 === undefined || v2 === '') {
        v2 = domain[1];
      } else {
        v2 = Number(v2);
      }

      if (v1 === undefined || v1 === '') {
        v1 = v0 + ((v2 - v0) / 2);
      } else {
        v1 = Number(v1);
      }

      // In case the minimum is higher than v1, or maximum is lower than v1
      v0 = Math.min(v0, v1);
      v2 = Math.max(v1, v2);

      const lLow = d3.scale.linear()
                   .domain([v0, v1])
                   .rangeRound([0, c.midColors.length - 1]);

      const lHigh = d3.scale.linear()
                    .domain([v1, v2])
                    .rangeRound([0, c.endColors.length - 1]);

      colorScaleFunction = function(v) {
        // If value doesn't exist, return white
        if (!v && !$.isNumeric(v)) {
          return defaultColor;
        }

        // Is v below the v0 -> v2 range
        if (v < v0) {
          return c.dimColors[0];
        }

        // Is v above the v0 -> v2 range
        if (v > v2) {
          return c.dimColors[c.dimColors.length - 1];
        }

        // Now we know we are in range. See our relationship to v1 to
        // decide whether to use lLow or lHigh
        if (v <= v1) {
          return c.midColors[lLow(v)];
        }

        return c.endColors[lHigh(v)];
      };

      colorScaleFunction.values = function() {
        return [v2, v1, v0];
      };

      colorScaleFunction.dimColors = function() {
        return c.dimColors;
      };

      colorScaleFunction.midColors = function() {
        return c.midColors;
      };

      colorScaleFunction.endColors = function() {
        return c.endColors;
      };

      colorScaleFunction.domain = function() {
        return domain;
      };

      colorScaleFunction.invert = function(rangeValue) {
        if (rangeValue < split) {
          return lLow.invert(rangeValue);
        }

        return lHigh.invert(rangeValue - split);
      };

      return colorScaleFunction;
    }
  }

  const l = d3.scale.linear()
            .domain(domain)
            .rangeRound([0, colors.length - 1]);

  colorScaleFunction = function(v) {
    const [min, max] = domain;
    // Range can be a simple number or percentage (with % sign at the end).
    // In the last scenario we need the value as percentage for the comparison!
    const calculatePercentageValue = (value) => ((value - min) * 100) / (max - min);
    const convertRangeAndValueForComparison = (range, value) => {
      if (range?.endsWith?.('%')) {
        return { range: parseFloat(range.replace(/%$/, ''), 10), value: calculatePercentageValue(value) };
      }

      return { range: parseFloat(range, 10), value };
    };

    let found;
    if (ranges) {
      ranges.some((range) => {
        const { range: rangeLeft, value: valueToLeftComparison } =
          convertRangeAndValueForComparison(range.rangeLeft, v);
        const { range: rangeRight, value: valueToRightComparison } =
          convertRangeAndValueForComparison(range.rangeRight, v);

        if (Number.isNaN(rangeLeft) && Number.isNaN(rangeRight)) {
          return;
        }

        if (!Number.isNaN(rangeLeft)) {
          if ((range.rangeLeftType === 'gte') && (valueToLeftComparison < rangeLeft)) {
            return;
          }

          if ((range.rangeLeftType === 'gt') && (valueToLeftComparison <= rangeLeft)) {
            return;
          }
        }

        if (!Number.isNaN(rangeRight)) {
          if ((range.rangeRightType === 'lte') && (valueToRightComparison > rangeRight)) {
            return;
          }

          if ((range.rangeRightType === 'lt') && (valueToRightComparison >= rangeRight)) {
            return;
          }
        }

        found = range.color;
        return true;
      });
    }

    if (found !== undefined) {
      return found;
    }

    return colors[l(v)];
  };

  colorScaleFunction.invert = function(rangeValue) {
    return l.invert(rangeValue);
  };

  colorScaleFunction.domain = function() {
    return l.domain();
  };

  return colorScaleFunction;
};

arc.colorpalette.fetch = function(cb) {
  function processColorPallete(p) {
    let dimColors = [];

    p.colorSettings = {
      name: p.name,
      type: p.type,
      value: p.value,
      id: p.id,
    };

    if (p.type === 100 || p.type === 102) {
      const c = computeColors(p);
      if (!c.dimColors) {
        return;
      }

      dimColors = c.dimColors;
    } else if (p.value[0] === '{') {
      const parsedValue = JSON.parse(p.value);
      dimColors = parsedValue.colors;
    } else {
      dimColors = p.value.split('\n');
    }

    p.swatchHeight = getSwatchHeight(dimColors);
    p.dimColors = dimColors;

    p.aggColors = function(domain) {
      return arc.colorpalette.aggScale(domain, p.dimColors);
    };
  }

  function onReceivingResponse(error, colorpalettes) {
    if (error) {
      arc.report.alertMessage('Failed to fetch color palettes');
    }

    arc.colorpalette.data = colorpalettes;
    arc.colorpalette.data.forEach((p) => {
      try {
        processColorPallete(p);
      } catch (err) {
        console.log('could not parse', p);
      }
    });

    if (cb) {
      cb();
    }
  }

  if (arc.testutil && typeof arc.testutil.mockColorPaletteData !== 'undefined') {
    onReceivingResponse(null, arc.testutil.mockColorPaletteData);
    return;
  }

  if (arc.colorpalette.dataFetched) {
    return;
  }

  arc.colorpalette.dataFetched = true;

  const assets = [
    {
      asset: arc.frontend.globals.urls.colorpaletteUrl,
      postprocess: d3.json,
    },
  ];

  arc.staticfetcher.fetch(assets, onReceivingResponse);
};

arc.colorpalette.get = function() {
  return arc.colorpalette.data;
};

// Convert a rgb array to brightness. We use this to figure out
// a contrast color for the text
function brightness(rgbArray) {
  // Standard formula to get from rgb to perceived brightness
  // Note that green is most visible to the human eye, followed by red
  //   and then blue
  let perceived_brightness = (rgbArray[0] * 299) + (rgbArray[1] * 587) + (rgbArray[2] * 114);
  perceived_brightness /= 255000;
  return perceived_brightness;
}

function beginsWithPoundSign(string) {
  return string.trim()[0] === '#';
}

// Depending on the brightness of the background color, use
// white or black text
arc.colorpalette.colorForBrightness = function(colorString, opacity) {
  try {
    if (opacity <= 0.5) {
      return '#222';
    }

    let rgbArray;

    if (beginsWithPoundSign(colorString)) {
      rgbArray = hexToRGB(colorString);
    } else {
      rgbArray = rgbStringToRGB(colorString);
    }

    if (brightness(rgbArray) >= 0.5) {
      return '#222';
    }

    return '#eee';
  } catch (err) {
    // If we can't parse the color, use a dark font
    return '#222';
  }
};

const memoizedColors = {};

arc.colorpalette.colorToRGBA = function(str, opacity) {
  opacity = opacity || '';
  const s = `${str}-${opacity}`;
  if (memoizedColors[s]) {
    return memoizedColors[s];
  }

  const ctx = document.createElement('canvas').getContext('2d');
  ctx.fillStyle = str;
  const f = ctx.fillStyle;
  const rgb = [
      parseInt(f.slice(1, 3), 16),
      parseInt(f.slice(3, 5), 16),
      parseInt(f.slice(5, 7), 16),
  ];
  const c = `rgba(${rgb.join(',')},${opacity})`;
  memoizedColors[s] = c;
  return c;
};

arc.colorpalette.rgbStringToRGB = rgbStringToRGB;
arc.colorpalette.rgbToHex = rgbToHex;
arc.colorpalette.hexToRGB = hexToRGB;
arc.colorpalette.rgbStringToHex = rgbStringToHex;
arc.colorpalette.hexToRGBAString = hexToRGBAString;
arc.colorpalette.rgbStringToRGBAString = rgbStringToRGBAString;
arc.colorpalette.convertToRGBAString = convertToRGBAString;
}());
