window.arc = window.arc || {};

(function() {
'use strict';

/*
 * Extension JS API:
 *
 * An extentions JS module should return an object that may implement
 *   the following endpoints:
 *
 * version: This must exist and reports on the version of the API
 *   that this object implements. Currently supported version strings
 *   are '1'. Note that this should be a string, not a number.
 *
 * supported: If this member exists, it is expected to be an object that
 *   contains a member called 'visualType'. This member should point at
 *   a string with the type of visual supported, or an array with strings
 *   containing the types of visuals supported.

     Example:
       supported: {
         visualType: 'trellis-bars'
       },

 * The current list of supported visuals is:
    table
    crosstab
    trellis-bars
    trellis-lines
    trellis-areas
    trellis-groupedbars
    singlevalue
    kpi
    packed-bubbles
    scatter
    flow
    funnel
    pie
    radial
    chord
    corelation
    corelation-flow
    calendar-heatmap
    map
    sparklines
    timecontrol
    link
    extension
    column-stats
    rawtable

 * settings: If a member exists with this name, it is expected to be a
 *   function that returns an array. Each array element is expected to be
 *   an object with the following members. In the list below, example values
 *   are also provided.

       id: 'Column Name',           // This is the id of this setting
                                    // Default value: none. This value must be
                                    //   specified

       displayName: 'Column Name',  // Name to use for displaying this setting.
                                    // Default value: same as id

       type: 'string',              // The type of the variable. Currently
                                    //   supported types are 'string' and 'boolean'
                                    // Default value: 'string'

       defaultValue: '',            // The default value for this setting.
                                    // Default value: none. This setting is
                                    //   optional

  Example settings value:
      settings: function() {
        return [
          {
            id: 'Column Name'
          },
          {
            id: 'Max bar width',
            defaultValue: '50'
          },
          {
            id: 'Color',
            defaultValue: 'steelblue'
          }
        ];
      },

 *
 * disableDraw: If a member exists with this name and is a function that returns
 *   a boolean true, then we disable to drawing of the visual.
 *
 * beforeDraw: If a member exists with this name it is expected to be
 *   a function that gets invoked before the draw event on the underlying
 *   visual.
 *
 * afterDraw: If a member exists with this name it is expected to be
 *   a function that gets invoked before the draw event on the underlying
 *   visual.
 *
 * afterDetailedDataRequest: If a member exists with this name it is expected to be
 *   a function that gets invoked before the data request for detailed data is sent
 *   out to the backend.
 *
 * The extension JS module also has access to an 'arcapi' object. That
 *   object supports the following functions:
 *
 * arcapi.settings(): This function returns an array of objets. Each object
 *   has a 'id' member that contains the id of a setting, and a 'value' object
 *   that has the value of that setting. Settings are returned in the same
 *   order as this JS extension specifies in its 'settings' function.
 *
 * arcapi.getSetting(settingName): This function returns the value of the the
 *   specified setting.
 *
 * arcapi.chartId(): This function returns the id attribute of the DOM element
 *   used to draw the visual.
 *   Example Usage: $('#' + arcapi.chartId()) will return a jquery selector to
 *     the DOM element.
 *
 * arcapi.dataResult(): This function returns data result object that gives the
 *   extension access to the raw data being used to plot the visual. This
 *   object supports the arc.data.result interface documented below.
 *
 * arcapi.addStyles(filepaths, cb): This function add in link elements for the
 *   additional styles specified in the filepaths array. The cb function is
 *   invoked after load completes.
 *
 * arcapi.addScripts(filepaths, cb): This function add in script elements for
 *   the additional styles specified in the filepaths array. The cb function is
 *   invoked after load completes.
 *
 * arcapi.sendParameters(params_hash, config): This function takes in a hash and send
 *   out the keys and values of the hash as parameters to other visuals
 *   on this page.

 *   To send parameters with dataset scope, you must include your parameters in the
 *   parameters hash. Then, for you config object, pass in the following:
 *
 *    {datasetScope:
 *        {dataset_id1: ['col_name1', 'col_name2'],
 *         dataset_id2: ['col_name1']
 *     }
 *
 *   For example:
 *     {datasetScope:
 *         {2: ['cereal_name', 'manufacturer_code'],
 *          4: ['state']
 *          5: []
 *     }
 *
 *   Sending an empty array for a dataset_id will untrack that dataset's dataset-
 *   scoped parameters sent from the visual via arcapi.sendParameters
 *
 * arcapi.getParameters(): This function return a hash that contains key-value
 *   mappings for all current parameters
 *
 * Internal Only:
 * afterDataRequest(data_request): If a member exists with this name and is a
 *   function, we call it with the data request object that the base visual
 *   asks for. This function can modify data_request and return a modified one,
 *   which is used for the backend data request.
 *
 * Extension CSS API:
 *
 * The best practice for CSS customizations is that customizations
   * should always use a custom selector at the beginning of every rule
   * can use any class names below
   * can use native DOM element selections
   * you may need to use !important to override existing rules
 *
 * arc-app-container:
 *   Container class for apps. One of these exist on the page
 *   when an app is being rendered.
 *
 * arc-viz-container:
 *   Container class for visuals. One of these exist on the page
 *   for every visual on the screen.
 *
 * arc-viz-type-<<visualtype>>:
 *   Container class for a particular visual type. For example,
 *   tables will have an arc-viz-type-table class.
 *
 * arc-dashboard-title-container:
 * arc-dashboard-subtitle-container:
 *   Container class for a div that surrounds the dashboard title,
 *   and another that surrounds the subtitle. Used to set dashboard
 *   title background colors, font type/size and text alignment.

 * arc-dashboard-title:
 * arc-dashboard-subtitle:
 *   Container class for a div that surrounds the dashboard title text,
 *   and another that surrounds the subtitle text. Used to access text values.

 * arc-viz-title-container:
 * arc-viz-subtitle-container:
 *   Container class for a div that surrounds the visual title,
 *   and another that surrounds the subtitle. Used to set visual
 *   title background color, font type/size and text alignment.

 * arc-viz-title:
 * arc-viz-subtitle:
 *   Container class for a div that surrounds the visual title text,
 *   and another that surrounds the subtitle text. Used to access text values.
 *
 * arc-viz-axis:
 *   Applied to the axis.
 *
 * arc-viz-legend:
 *   Applied to the color/mark legend.
 *
 * arc-viz-axis-label:
 *   Applied to the axis label.
 */

arc.testutil = arc.testutil || {};
arc.css = arc.css || {};
arc.css.settings = arc.css.settings || {};

arc.css.settings.fetchCssStyles = function(userSettings) {
  var s = userSettings.find_by_id('css_styles');
  if (!s) {
    return [];
  }
  s.process_value();
  return s.getCssIds();
};

arc.css.settings.fetchInlineCss = function(userSettings) {
  var s = userSettings.find_by_id('inline_css');
  if (!s) {
    return undefined;
  }

  return s.value();
};

/*
 * Converts an array of settings dictionaries into settings
 * Each setting is:
 *   id: id
 *   displayName: display name
 *   type: 'string' | 'boolean'
 *   subtype: 'multiline' (valid for 'string' types)
 *   value: value
 *   validate: js function that validates
 */
arc.css.settings.specificationToSettings = function(spec) {
  let settings = arc.visual.settings();

  if (!spec) {
    return settings;
  }

  spec.forEach(function(s) {
    let type = s.type.toUpperCase(),
        subtype = s.subtype && s.subtype.toUpperCase();

    if (type === 'DROPDOWN') {
      let v = arc.visual.setting()
                 .id(s.id)
                 .section('Style')
                 .tab('Base Settings')
                 .group(s.displayName)
                 .type(arc.visual.settingTypeEnum.DROPDOWN)
                 .value(s.value);
      if (s.validate) {
        v.validate_fn(s.validate);
      }

      s.names.forEach(function(n, i) {
        v.add_name(n);
        v.add_label(s.labels ? s.labels[i] : n);
      });

      settings.addSetting(v);
      return;
    }

    if (type === 'STRING') {
      let v = arc.visual.setting()
                 .id(s.id)
                 .section('Style')
                 .tab('Base Settings')
                 .type(arc.visual.settingTypeEnum.STRING)
                 .add_name(s.id)
                 .add_label(s.displayName)
                 .value(s.value);
      if (s.validate) {
        v.validate_fn(s.validate);
      }

      if (subtype === 'MULTILINE') {
        v.is_textarea(true);
      }

      settings.addSetting(v);
      return;
    }

    if (type === 'BOOLEAN') {
      let v = arc.visual.setting()
                 .id(s.id)
                 .section('Style')
                 .tab('Base Settings')
                 .type(arc.visual.settingTypeEnum.CHECKBOX)
                 .add_name(s.id)
                 .add_label(s.displayName)
                 .value(s.value);
      if (s.validate) {
        v.validate_fn(s.validate);
      }

      settings.addSetting(v);
    }
  });

  return settings;
};

arc.css.settings.cssStyles = function() {
  /*
   * TODO: this needs refactoring.
   * Pre release 4.0, all settings happened via modals. Custom Styles settings
   *   happened via a nested modal that needed to track a lot of state because
   *   the proposed modifications only applied after the parent modal had the 'Apply'
   *   button checked.
   * This complication does not exist anymore, and substantially simplifies the
   * issues being handled below.
   *
   * There are a few complicating factors here:
   *
   * 1. The setting is stored as a string. It is a string that JSON.parses into
   *    an array of hashes, where each hash has a cssid and a cssSettings hash.
   *    The call to JSON.parse and JSON.stringify convert back and forth from
   *    the string to the JSON structure below.
   * 2. A single extension can be added in multiple times to a visual. Hence,
   *    we cannot track extensions uniquely within a visual by using the
   *    extension ID. Instead, we create a 'uniqueId' within the arc.extension
   *    object and use it to uniquely identify individual instances of
   *    extensions
   */

  var customStyles = arc.model.CustomStyles(),
      extensions = arc.extensions();

  var setting = arc.visual.setting()
                   .id('css_styles')
                   .section('Style')
                   .tab('Custom Style')
                   .type(arc.visual.settingTypeEnum.CUSTOM)
                   .add_name('CSS_STYLES')
                   .add_label('CSS styles')
                   .value('[]');

  setting.extensions = function() {
    return extensions;
  };

  function extensionsToValue() {
    var data = currentExtensions().map(function(e) {
      let newSettings = {};

      if (activeExtension) {
        if (e.uniqueId() === activeExtension.uniqueId()) {
          activeExtensionSettings.fillValues(newSettings);
        }
      }

      let toSave = e.settingsToSave(newSettings);

      return {
        cssid: e.extensionId(),
        settings: toSave,
      };
    });

    return JSON.stringify(data);
  }

  function valueToExtensions(value) {
    extensions = arc.extensions();
    var v = JSON.parse(value);

    v.forEach(function(s) {
      var cssid,
          cssSettings;

      if ((typeof (s) === 'string') || (typeof (s) === 'number')) {
        cssid = s;
        cssSettings = {};
      } else {
        cssid = s.cssid;
        cssSettings = s.settings;
      }

      var e = arc.extension()
                 .extensionId(cssid)
                 .savedSettings(cssSettings);
      extensions.add(e);

      if (arc.testutil && arc.testutil.ignoreJsCrossOrigin) {
        return;
      }

      extensions.fetchExtensions(setting.viz_lib(), undefined /* cb */, true);
    });
  }

  function currentExtensions() {
    const es = [];
    extensions.all().forEach((e) => { es.push(e); });
    return es;
  }

  // When a new value is set, re-initializa tracking variables
  setting.valueCb(function(_) {
    // Setup 'extensions' based on the saved CSS IDs
    valueToExtensions(_);

    setting.paintTable();
    return setting;
  });

  setting.getCssIds = function() {
    return currentExtensions().map(function(e) {
      return e.extensionId();
    });
  };

  setting.changeCb = function(cbname, cb) {
    return setting.addChangeCb(cbname, cb);
  };

  function getExtension(uniqueId) {
    return extensions.byUniqueId(uniqueId);
  }

  function errorContent($elem) {
    var uniqueid = $elem.closest('tr').attr('data-uniqueid');
    var e = getExtension(uniqueid);
    if (!e) {
      return '';
    }

    if (e.error()) {
      return e.error();
    }

    return '';
  }

  function popoverContent($elem) {
    var uniqueid = $elem.closest('tr').attr('data-uniqueid');
    var e = getExtension(uniqueid);
    if (!e) {
      return '';
    }

    var settings = e.settings();
    if (!settings || (settings.length === 0)) {
      return 'Nothing to configure';
    }

    var cssname = $elem.closest('tr').attr('data-cssstylename');
    var templateIn = $('#css-setting-popover').html();
    var templateItem = $('#css-setting-item table').html();
    var data = settings.map(function(s) {
      return arc.misc.safe_substitute(templateItem, {
        LABEL: s.displayName,
        CONTENT: s.value,
      });
    });
    var c = arc.misc.safe_substitute(
        templateIn,
        {
          'TITLE': 'Settings for ' + cssname,
          'HTML_TABLE': '<table class="popover-table">' + data.join('') +
              '</table>'
        });
    return c;
  }

  var activeExtension,
      activeExtensionSettings;

  function clickSettings($elem) {
    $elem.popover('hide');
    var uniqueid = $elem.closest('tr').attr('data-uniqueid');
    var extension = getExtension(uniqueid);
    if (!extension) {
      return;
    }

    var settings = extension.settings();
    if (!settings || (settings.length === 0)) {
      return;
    }

    activeExtension = extension;

    var cssname = $elem.closest('tr').attr('data-cssstylename');
    activeExtensionSettings = arc.css.settings.specificationToSettings(settings);

    var extensionSettingsModal = arc.report.settingsModal();
    extensionSettingsModal.viz_lib(setting.viz_lib())
                    .settings_modal_id('settings-css-' + uniqueid)
                    .settings_modal_title('Custom JS Settings for ' + cssname)
                    .settings_modal_holder('css-settings-modal-holder')
                    .user_settings(activeExtensionSettings);

    extensionSettingsModal.setup(function() {
      var newSettings = {};
      activeExtensionSettings.fillValues(newSettings);
      extension.savedSettings(newSettings);
      extension.processJs(setting.viz_lib());
      activeExtension = undefined;
      activeExtensionSettings = undefined;
      setting.paintTable();
    });

    extensionSettingsModal.show();
  }

  function handleSettingsEdit($tbody) {
    $tbody.find('.fa-times')
          .off('click')
          .on('click', function() {
            var $tr = $(this).closest('tr');
            var uniqueid = Number($tr.attr('data-uniqueid'));
            extensions.removeUniqueId(uniqueid);
            $tr.remove();
            setting.process_value();
            setting.makeChangeCbs();
          });

    $tbody.find('.fa-cog')
          .popover('dispose')
          .popover({
            container: 'body',  // makes popover be on top
            html: true,
            sanitize: false,
            trigger: 'hover',
            placement: 'bottom',
            delay: {
              show: '50'
            },
            title: '',
            content: function() {
              return popoverContent($(this));
            }
          })
          .off('click')
          .on('click', function() {
            clickSettings($(this));
          });

    $tbody.find('.fa-exclamation')
          .popover('dispose')
          .popover({
            container: 'body',  // makes popover be on top
            html: true,
            sanitize: false,
            trigger: 'hover',
            placement: 'bottom',
            delay: {
              show: '50'
            },
            title: '',
            content: function() {
              return errorContent($(this));
            }
          });
  }

  // This function is exposed for testing purposes
  setting.paintTable = function(modified) {
    var $settingsDiv = $('#' + setting.div_id());
    var $tbody = $settingsDiv.find('tbody');
    var templateIn = $('#custom-css-styles-template .baserow tbody').html();

    function addRow(e) {
      var settings_class;
      var remove_class = '';
      if (!setting.mutable()) {
        remove_class = 'arc-invisible';
      }

      var settings_error = 'arc-invisible';
      if (e.error()) {
        settings_error = 'settings-error';
      }
      var settings = e.settings();
      if (!settings || (settings.length === 0)) {
        settings_class = 'settings-disabled';
      } else {
        if (setting.mutable()) {
          settings_class = 'settings-enabled';
        } else {
          settings_class = 'settings-disabled';
        }
      }

      $tbody.append(arc.misc.safe_substitute(
          templateIn,
          {
            'UNIQUEID': e.uniqueId(),
            'STYLEID': e.extensionId(),
            'STYLENAME': customStyles.cssname(e.extensionId()) || '',
            'SETTINGS_ENABLED': settings_class,
            'REMOVE_ENABLED': remove_class,
            'SETTINGS_ERROR': settings_error,
          }));
    }

    $tbody.empty();
    var es = currentExtensions();
    if (es.length === 0) {
      $settingsDiv.find('table').addClass('arc-invisible');
    } else {
      $settingsDiv.find('table').removeClass('arc-invisible');
      es.forEach(function(e) {
        addRow(e);
      });
    }

    handleSettingsEdit($tbody);
  };

  // Initialize variables
  let displayPicklist = []; // ids of assets to display
  let localStyleMap = {}; // Dictionary of assets by id
  const $modal = $('#pickcss-modal');
  let $ul = $modal.find('ul');
  const templateIn = $('#pickcss-item-template').html();

  // Add event handlers
  let sortOrder = 1; // Ascending = 1; Descending = -1
  const $customCssHeader = $('#custom-css-header');
  const $sortNameIcon = $customCssHeader.find('.sort-by-name-icon i');
  const $sortIdIcon = $customCssHeader.find('.sort-by-id-icon i');
  const $sortEditDateIcon = $customCssHeader.find('.sort-by-editdate-icon i');
  const $sortEditorIcon = $customCssHeader.find('.sort-by-editor-icon i');
  const $assetName = $customCssHeader.find('.customcss-asset-name');
  const $idHeader = $customCssHeader.find('.customcss-id');
  const $editDateHeader = $customCssHeader.find('.customcss-editdate');
  const $editorHeader = $customCssHeader.find('.customcss-editor');

  const replaceAssetList = (keys) => {
    $ul.empty();
    displayPicklist = [];
    keys.forEach(id => {
      $ul.append(arc.misc.safe_substitute(
        templateIn,
        {
          'LABEL': localStyleMap[id].name,
          'VALUE': id,
          'EDITDATE': localStyleMap[id].last_edited,
          'FORMATTEDDATE': arc.misc.fromNow(localStyleMap[id].last_edited),
          'LASTEDITOR': localStyleMap[id].last_editor
        })
      );
      displayPicklist.push(id);
    });
  };

  function resetIcon(_sortIcon) {
    _sortIcon.removeClass('fa-sort-amount-desc');
    _sortIcon.addClass('fa-sort-amount-asc');
    sortOrder = 1;
  }

  function flipSortIcon(_sortIcon) {
    _sortIcon.toggleClass('fa-sort-amount-asc');
    _sortIcon.toggleClass('fa-sort-amount-desc');
    sortOrder = -sortOrder;
  }

  function hideAllIcons() {
    $sortNameIcon.parent().addClass('arc-invisible');
    $sortIdIcon.parent().addClass('arc-invisible');
    $sortEditDateIcon.parent().addClass('arc-invisible');
    $sortEditorIcon.parent().addClass('arc-invisible');
  }

  function toggleCustomCssElements(isVisible) {

    // Reset search bar and sort icons
    $modal.find('input.custom-css-search').val('');
    hideAllIcons();
    $modal.find('.search-textbox').toggleClass('arc-invisible', !isVisible);
    $('#custom-css-header').toggleClass('arc-invisible', !isVisible);
    $('.pickcss-list-item-cont').toggleClass('arc-invisible', !isVisible);
    $('.customcss-add').removeAttr('disabled', !isVisible);
  }

  function handleAddStyle() {
    if (arc.misc.modalShown($modal)) {
      return;
    }

    customStyles.invalidate();
    customStyles.fetch(handleAddStylePostFetch);
  }

  function handleAddStylePostFetch() {
    if (arc.misc.modalShown($modal)) {
      return;
    }

    $ul.empty();
    $.each(customStyles.styles(), function(k, v) {
      $ul.prepend(arc.misc.safe_substitute(
          templateIn,
          {
            'LABEL': v.name,
            'VALUE': k,
            'EDITDATE': v.last_edited,
            'FORMATTEDDATE': arc.misc.fromNow(v.last_edited),
            'LASTEDITOR': v.last_editor
          }));
    });

    if ($ul.find('li').length === 0) {
      $modal.find('.customcss-none')
            .removeClass('arc-invisible');
      toggleCustomCssElements(false);
    } else {
      $modal.find('.customcss-none')
            .addClass('arc-invisible');
      toggleCustomCssElements(true);
    }

    const newExtensions = arc.extensions();

    let $cssadd = $modal.find('.customcss-add');
    $cssadd.offOn('click.addCss', () => {
      // Don't permit two clicks
      $cssadd.off('click');

      var $elems = $('#pickcss-modal li input:checked');
      $elems.each(function() {
        var cssid = $(this).data('value');

        var e = arc.extension()
                   .extensionId(cssid)
                   .savedSettings({});
        newExtensions.add(e);
      });

      newExtensions.fetchExtensions(setting.viz_lib(), function() {
        newExtensions.all().forEach(function(e) {
          extensions.add(e);
        });

        setting.paintTable(true);

        setting.process_value();
        setting.makeChangeCbs();

        $modal.modal('hide');
      }, true);
    });

    // Sort handlers for picklist
    // On click, sort by asset name, redraw picklist, flip sort icon,
    $assetName.offOn('click.assetName', function handleSortAssetName() {
      if ($sortNameIcon.parent().hasClass('arc-invisible')) {
        hideAllIcons();
        resetIcon($sortNameIcon);
        $sortNameIcon.parent().removeClass('arc-invisible');
      } else {
        flipSortIcon($sortNameIcon);
      }
      displayPicklist.sort(function (a, b) {
        const firstName = localStyleMap[a].name;
        const secondName = localStyleMap[b].name;
        return sortOrder * firstName.localeCompare(secondName, 'en');
      });
      replaceAssetList(displayPicklist);
    });

    $idHeader.offOn('click.id', function handleSortById() {
      if ($sortIdIcon.parent().hasClass('arc-invisible')) {
        hideAllIcons();
        resetIcon($sortIdIcon);
        $sortIdIcon.parent().removeClass('arc-invisible');
      } else {
        flipSortIcon($sortIdIcon);
      }
      displayPicklist.sort(function (a, b) {
        return sortOrder * (localStyleMap[a].id - localStyleMap[b].id);
      });
      replaceAssetList(displayPicklist);
    });

    $editDateHeader.offOn('click.editDate', function handleSortEditDate() {
      if ($sortEditDateIcon.parent().hasClass('arc-invisible')) {
        hideAllIcons();
        resetIcon($sortEditDateIcon);
        $sortEditDateIcon.parent().removeClass('arc-invisible');
      } else {
        flipSortIcon($sortEditDateIcon);
      }
      displayPicklist.sort(function (a, b) {
        const firstDate = localStyleMap[a].last_edited;
        const secondDate = localStyleMap[b].last_edited;
        return sortOrder * (moment(firstDate) - moment(secondDate));
      });
      replaceAssetList(displayPicklist);
    });

    $editorHeader.offOn('click.editor', function handleSortEditor() {
      if ($sortEditorIcon.parent().hasClass('arc-invisible')) {
        hideAllIcons();
        resetIcon($sortEditorIcon);
        $sortEditorIcon.parent().removeClass('arc-invisible');
      } else {
        flipSortIcon($sortEditorIcon);
      }
      displayPicklist.sort(function (a, b) {
        const firstEditor = localStyleMap[a].last_editor;
        const secondEditor = localStyleMap[b].last_editor;
        return sortOrder * firstEditor.localeCompare(secondEditor, 'en');
      });
      replaceAssetList(displayPicklist);
    });

    $modal.modal('show');
  }

  // Key-value function that should be applied on each style object
  function addStyleHtmlRow(k, v) {
    $ul.prepend(
      arc.misc.safe_substitute(
        templateIn,
        {
          'LABEL': v.name,
          'VALUE': k,
          'EDITDATE': v.last_edited,
          'FORMATTEDDATE': arc.misc.fromNow(v.last_edited),
          'LASTEDITOR': v.last_editor
        }
      ));
    displayPicklist.push(k);
  }

  function paintSetting(tab_d3_sel) {
    var container = tab_d3_sel.append('div')
                              .attr('id', setting.div_id());
    if (setting.hidden()) {
      container.attr('class', 'arc-invisible');
    }

    var $settingsDiv = $('#' + setting.div_id());
    $settingsDiv.append($('#custom-css-styles-template .basetable').html());

    const settingsContainer = $(tab_d3_sel.node()).closest('.settings-container')[0];
    const settingsContainerId = settingsContainer != null ? settingsContainer.id : '';
    const customStyleBelongsToVisual = settingsContainerId === 'dashboard-visual-style-sidebar';
    if (!customStyleBelongsToVisual) { // belongs to a dashboard or app
      $settingsDiv.find('#custom-style-info').removeClass('arc-invisible');
    }

    if (!setting.mutable()) {
      $settingsDiv.find('.viz-customcss-addstyle')
                  .addClass('arc-invisible');
    }

    // Paint the settings table now with the information we have
    setting.paintTable();

    // Then, fetch id -> name mappings from the metadata, and fetch the
    // settings table again after the mapping is available
    customStyles.fetch(function initializeStylesModal(stylesMap) {
      $ul.empty();
      $.each(stylesMap, addStyleHtmlRow);
      displayPicklist.reverse();
      localStyleMap = stylesMap;

      $('.custom-css-search')
        .offOn('input.customStyles', function () {
          if (this.value === '') {
            $ul.empty();
            displayPicklist = [];
            $.each(localStyleMap, addStyleHtmlRow);
            displayPicklist.reverse();
          } else {
            const result = arc.css.searchStyles(this.value, localStyleMap);
            replaceAssetList(result);
          }

          // Clear sort icons
          hideAllIcons();
        });
        setting.paintTable();
    });

    var $addstyle = $settingsDiv.find('.viz-customcss-addstyle');

    $addstyle.offOn('click.addStyle', () => {
      handleAddStyle();
    });

    return setting;
  }

  setting.paint_fn(paintSetting);
  setting.fetch_fn(extensionsToValue);

  return setting;
};

arc.css.settings.containerCssClass = function(isFilter) {
  var viz_type = (isFilter ? 'filter' : 'visual');
  return arc.visual.setting()
            .id('container_css_class')
            .section('Style')
            .tab('Custom Style')
            .type(arc.visual.settingTypeEnum.STRING)
            .add_name('CONTAINER_CSS_CLASS')
            .add_label('Enter CSS classes for the ' + viz_type)
            .info('Separate multiple classes using a space.' +
              ' These classes add to the existing CSS classes of the ' +
              viz_type + '\'s container.')
            .value('');
};

arc.css.settings.inlineCss = function() {
  const setting = arc.visual.setting()
                   .id('inline_css')
                   .section('Style')
                   .tab('Custom Style')
                   .type(arc.visual.settingTypeEnum.CUSTOM)
                   .add_name('INLINE_CSS')
                   .add_label('Inline CSS')
                   .value('');

  let cssEditor;

  setting.cleanupSetting = function() {
    if (cssEditor) {
      cssEditor.destroy();
      cssEditor = undefined;
    }

    return setting;
  };

  setting.resetValue = function() {
    if (arc.testutil.testWithoutAce || !cssEditor) {
      setting.value('');
    } else {
      cssEditor.value('');
    }

    return setting;
  };

  function cssEditorSetup($container) {
    if (arc.globals.enable_custom_styles && (arc.globals.is_edit || arc.globals.screen_appgroup)) {
      const $autocompleteOnOffButton = $container.find('.autocomplete');

      const editorSettings = {
        autocompleteType: 'customstyle',
        autocompleteOnOffButton: $autocompleteOnOffButton,
        excludeAutocomplete: ['keyWordCompleter'],
        aceOptions: {
          theme: 'ace/theme/chrome',
          wrap: false,
        }
      };
      setting.cleanupSetting();

      cssEditor = arc.utils.autocompleteEditor(`${setting.div_id()}-css`, 'css', editorSettings);
      cssEditor.value(setting.value());

      if (setting.mutable()) {
        const session = cssEditor.getSession();

        let contents;

        // When modifying the contents of the inline css textbox,
        // we only redraw the page after the user has not made
        // any changes for 2 seconds. This means that as the user
        // is continually typing data, we are not continually
        // redrawing
        let incarnation = 1;
        session
          .on('change', () => {
            const v = setting.value();
            contents = session.getValue();
            if (v === contents) {
              return;
            }

            setting.value(contents);
            if (setting.viz_lib()) {
              setting.viz_lib().page_needs_save();
            } else {
              arc.report.pagenav.needs_save(true);
            }

            contents = v;
            incarnation += 1;
            const storedIncarnation = incarnation;
            setTimeout(() => {
              if (storedIncarnation === incarnation) {
                setting.process_value();
                setting.makeChangeCbs();
              }
            }, 2000);
          });
      } else {
        cssEditor.aceEditor().setReadOnly(true);
      }
    }
  }

  function paintSetting(tab_d3_sel) {
    const divId = setting.div_id();

    const container =
      tab_d3_sel.append('div')
                .attr('id', divId);
    if (setting.hidden()) {
      container.classed('arc-invisible', true);
    }

    const containerLabel = container.append('div')
             .attr('id', `${divId}-label`);
    containerLabel.append('label')
                  .text('Inline CSS');

    containerLabel.append('span')
                  .classed('autocomplete', true)
                  .classed('pull-right', true)
                  .style('margin-top', 'auto');

    container.append('div')
             .classed('inline_css_style', true)
             .attr('id', `${divId}-css`);
    cssEditorSetup($(container.node()));
  }

  function repaintSetting() {
    const $container = $(`${setting.div_id()}`);
    cssEditorSetup($container);
  }

  function fetchSetting() {
    if (arc.testutil.testWithoutAce) {
      return setting.value();
    }

    if (cssEditor) {
      if (setting.mutable()) {
        return cssEditor.getSession().getValue();
      }

      return cssEditor.value();
    }

    return setting.value();
  }

  setting.paint_fn(paintSetting);
  setting.repaint_fn(repaintSetting);
  setting.fetch_fn(fetchSetting);

  setting.changeCb = function(cbname, cb) {
    return setting.addChangeCb(cbname, cb);
  };

  return setting;
};

arc.css.CUSTOMCSS_MANAGER = (function() {
  var css = function() { return 'arc.css.CUSTOMCSS_MANAGER'; };

  var activeVisualStyles = [],
      activeVisualCssIds = {},
      styleElement1,
      styleElement2,
      appSettings,
      appgroupSettings;

  /*
   * Apply custom CSS classes to the container.
   *
   * One complication here is that in edit mode, users may be modifying the
   *   classes that apply. So we have to remove any classes we applied before.
   *
   * Another complication here is that arcadia may have some classes set on
   *   the container element. We want to make sure we ignore these class names
   *   when applying or removing previously-applied container classes.
   */
  css.applyContainerClasses = function($sel, in_classnames) {
    if (!arc.globals.apply_custom_styles) {
      // Do not apply any styles if global settings disable them
      return;
    }

    // Read previously set classes as an array
    var existingClassNames = [];
    var d = $sel.attr('data-arc-classnames');
    if (d) {
      existingClassNames = d.trim().split(/\s+/);
    }

    // Parse new classes as an array
    var classNames = in_classnames.trim().split(/\s+/);

    // Remove previously set classes that no longer apply
    existingClassNames.forEach(function(classname) {
      if (!(classname in classNames)) {
        $sel.removeClass(classname);
      }
    });

    var newExistingClassNames = [];
    classNames.forEach(function(classname) {
      // For classnames that don't exist, track them
      if (!$sel.hasClass(classname)) {
        newExistingClassNames.push(classname);
        $sel.addClass(classname);
        return;
      }

      // For classnames that do exist, tract the ones that exist
      // because we added them in
      if (classname in existingClassNames) {
        newExistingClassNames.push(classname);
        return;
      }

      // Note that we are ignoring class names that previously exist
      // on the element - these are arcadia-specified classes and should
      // not be later removed by being added to 'data-arc-classnames' below
    });

    $sel.attr('data-arc-classnames', newExistingClassNames.join(' '));
  };

  css.appSettings = function(_) {
    if (!arguments.length) return appSettings;
    appSettings = _;
    return css;
  };

  css.appgroupSettings = function(_) {
    if (!arguments.length) return appgroupSettings;
    appgroupSettings = _;
    return css;
  };

  function initialize() {
    // Create two elements to track CSS so we can transition to new styles
    // without a flash of unstyled content
    styleElement1 = document.createElement('style');
    styleElement1.setAttribute('id', 'arc-customcss1');
    styleElement1.type = 'text/css';
    styleElement1.innerHTML = '';

    styleElement2 = document.createElement('style');
    styleElement2.setAttribute('id', 'arc-customcss2');
    styleElement2.type = 'text/css';
    styleElement2.innerHTML = '';

    document.getElementsByTagName('head')[0].appendChild(styleElement1);
    document.getElementsByTagName('head')[0].appendChild(styleElement2);
  }

  function cssIdToName(cssId) {
    return 'arc-customcss-' + cssId;
  }

  function cssUrl(cssId) {
    return `${arc.frontend.globals.urls.customCssRawUrl}/${cssId}`;
  }

  function setInlineStyle(inlineStyles) {
    var activeStyleElement,
        newStyleElement;

    if (inlineStyles.indexOf('<') >= 0) {
      inlineStyles = '';
    }

    // For inline CSS, we have one active and one inactive element at
    // all times. We switch between the active and inactive so we don't
    // get flashes of unstyled content when we overwrite an existing set
    // of styles
    if (styleElement1.innerHTML === '') {
      activeStyleElement = styleElement2;
      newStyleElement = styleElement1;
    } else {
      activeStyleElement = styleElement1;
      newStyleElement = styleElement2;
    }

    if (activeStyleElement.innerHTML !== inlineStyles) {
      newStyleElement.innerHTML = inlineStyles;
      activeStyleElement.innerHTML = '';
    }
  }

  var customcssClassName = 'customcsslink';
  css.paint = function(cb_after_load) {
    if (!styleElement1) {
      initialize();
    }

    var newStyles = activeVisualStyles.join('\n');

    setInlineStyle(newStyles);

    var filepaths = [];
    $.each(activeVisualCssIds, function(cssId, v) {
      if (cssId === undefined) {
        return;
      }

      filepaths.push(cssUrl(cssId));
    });

    // We need to remove older included CSS files.
    // Find all current CSS files and remove the classname from them
    var $css = $('.' + customcssClassName);
    $css.removeClass(customcssClassName);

    // Add the classname to all current CSS classes (and add them in if necessary)
    arc.frontend.loadResources('link', filepaths, customcssClassName, cb_after_load);

    // Remove css links that no longer exist
    $css.filter(':not(.' + customcssClassName + ')')
        .remove();

    return css;
  };

  function applyCssFromApp() {
   if (arc.globals.appgroup) {
      // apply appgroup-specific settings
      var css = arc.globals.appgroup.getCss();
      if (css) {
        activeVisualStyles.push(css);
      }

      // now apply any custom styles specified within the appgroup
      applyCssFromSettings(appgroupSettings);
    }
  }

  function applyCssFromDashboard() {
    if (arc.globals.dashboard) {
      // apply dashboard-specific settings
      var css = arc.globals.dashboard.getCss();
      if (css) {
        activeVisualStyles.push(css);
      }

      // now apply any custom styles specified within the dashboard
      applyCssFromSettings(appSettings);
    }
  }

  function applyCssFromSettings(userSettings) {
    if (!userSettings) {
      return;
    }

    var s = arc.css.settings.fetchInlineCss(userSettings);
    if (s) {
      activeVisualStyles.push(s);
    }

    var cssl = arc.css.settings.fetchCssStyles(userSettings);
    cssl.forEach(function(cssId) {
      if (cssId !== undefined) {
        activeVisualCssIds[cssId] = true;
      }
    });
    $('.fouc-appgroup').removeClass('fouc-appgroup');
  }

  // Typically, the list of visuals comes from arc.globals.reports
  // For testing purposes, we also support getting the list of visuals
  // passed in as a parameter.
  css.applyCustomCss = function(vizl, cb) {
    // The 'styles=off' URL parameter turns off custom CSS and javascript
    // This is expected to be very useful for debugging customer issues
    if (!arc.globals.apply_custom_styles) {
      // remove fouc, this is called multiple times in dashboard -
      // once for every visual and once for the dashboard
      $('.fouc').removeClass('fouc');
      return;
    }

    // Re-fetch styles and css ids from active visuals
    activeVisualStyles = [];
    activeVisualCssIds = {};

    if (!vizl) {
      vizl = [];
    }

    if (arc.globals.reports) {
      arc.globals.reports.vizCallback(function(viz) {
        vizl.push(viz);
      });
    }

    applyCssFromDashboard();

    vizl.forEach(function(viz) {
      applyCssFromSettings(viz.user_settings());
    });

    applyCssFromApp();

    css.paint(function() {
      $('.fouc').removeClass('fouc');
      if (cb) {
        cb();
      }
    });

    return css;
  };

  return css;
}());

}());
