arc.datasetcustomcolorsetting = function() {
const setting = arc.visual.setting();

function getColorShelfElements() {
  const vizLib = setting.viz_lib();
  if (!vizLib) {
    return undefined;
  }

  const colorShelf = vizLib.shelves().getShelf('color_shelf');
  return colorShelf?.elements?.();
}

function shallowComparisonOfBackgroundObject(obj1, obj2) {
  /* Example:
  colorSetting: {}
  colors: ['#eceff1', '#cfd8dc', '#b0bec5', '#90a4ae', '#78909c',
   '#607d8b', '#546e7a', '#455a64', '#37474f', '#263238']
  customColorsEnabled:true
  dimensionMappings: {C: 'rgb(55, 71, 79)', H: 'rgb(84, 110, 122)'}
  fixedDimensionColors: false
  isReversed: false
  opacity: 60
   */
  if ((!obj1 && obj2) || (!obj2 && obj1)) {
    return true;
  }

  if (obj1 && obj2) {
    for (const [key, value] of Object.entries(obj1)) {
      if (JSON.stringify(value) !== JSON.stringify(obj2[key])) {
        return true;
      }
    }
  }

  return false;
}

function getCustomColorConfig() {
    const colorShelfElements = getColorShelfElements();
    const element = colorShelfElements?.length === 1 && colorShelfElements?.[0];
    const previousColorSetting = element?.backgroundColor?.();
    element?.setupShelfElementBackgroundColor?.();
    const colorSetting = element?.backgroundColor?.();
    const changed = shallowComparisonOfBackgroundObject(colorSetting, previousColorSetting);

    return { changed,
             customColorSetting: colorSetting.customColorsEnabled ? colorSetting : undefined };
}

function isOnlyOneElementPresentInColorShelf() {
  return getColorShelfElements()?.length === 1;
}

// There is an option to define color palette/opacity/color mapping etc
// on dataset's field (arc_admin_dataset.js).
// If any of this field used as a color_shelf element we use by default the color config.
// It can be overwritten but first the "Use default dataset checkbox" should be unchecked!
function applyDatasetCustomColoringConfigAndBlockEditing(
  colorconfig, changed = false
) {
  if (setting && colorconfig) {
    if (colorconfig && colorconfig.customColorsEnabled) {
      setting.getDependents().forEach((affectedSetting) => {
        if (colorconfig.customColorsEnabled) {
          let newValue;
          switch (affectedSetting.id()) {
            case 'color_palette':
              newValue = JSON.stringify({
                colors: colorconfig.colors,
                colorSettings: colorconfig.colorSettings,
                isReversed: colorconfig.isReversed,
                firstColor: colorconfig.firstColor,
              });
              break;
            case 'fixed_color_legend':
              newValue = colorconfig.fixedDimensionColors ? 'true' : 'false';
              break;
            case 'color_opacity':
              newValue = colorconfig.opacity.toString();
              break;
            case 'measure_ranges':
              newValue = JSON.stringify({
                ranges: colorconfig?.colorSettings?.measureRanges || [],
              });
              break;
            case 'dim_to_color_mapping':
              // eslint-disable-next-line no-case-declarations
              const dimensionMapDefined =
                colorconfig.dimensionMappings &&
                Object.keys(colorconfig?.dimensionMappings).length;
              // eslint-disable-next-line no-case-declarations
              const dimensionMap = dimensionMapDefined ?
                colorconfig.dimensionMappings :
                {};
              newValue = JSON.stringify({
                mappings: dimensionMap,
                isChecked: dimensionMapDefined,
              });
              break;
            default:
              break;
          }

          affectedSetting?.value?.(newValue);

          if (affectedSetting.id() === 'color_palette') {
            affectedSetting.saveIntoVisual();
          }

          affectedSetting?.repaintControl();
        }
      });

      // We have to save this change without user action
      // otherwise the saved_data().report_info will overwrite it, when restore_data is called
      // (because the saved value differs from the newly requested dataset's color!)
      if (changed) {
        setting.viz_lib().report().viz().save();
      }
    }
  }
}

function processCustomColorSettings(_value) {
  // Settings only enabled if and only one element present in the color shelf.
  // If more are present the coloring based on the combination, not one individual setting.
  if (isOnlyOneElementPresentInColorShelf()) {
    // changed means that in the background since the last view,
    // the dataset color setting was changed on the dataset page -> we should use that new value
    const { changed, customColorSetting } = getCustomColorConfig();
    // handle the scenario when we have saved report_info with custom color,
    // but in the dataset the custom color was deleted!
    if (changed && !customColorSetting && _value === 'true') {
      setting.value('false');
      return;
    }

    if (customColorSetting) {
      setting.show();
      if (_value === 'true') {
        applyDatasetCustomColoringConfigAndBlockEditing(customColorSetting, changed);
      }
    } else {
      // if there is only one element without custom -> hide the feature
      setting.hide();
    }
  } else {
    // if there is no element in the color_shelf, or more than one, the setting is useless
    // so we hide it!
    setting.hide();
  }
}

setting
  .id('use_dataset_color_settings')
  .section('Style')
  .tab('Color')
  .type(arc.visual.settingTypeEnum.CHECKBOX)
  .add_label('Use the dataset level defined color settings')
  .info(
    'The dataset level field custom color configurations can be turn on and off'
  )
  .value('false')
  .valueCb(processCustomColorSettings)
  .hide();

return setting;
};
