window.arc = window.arc || {};

(function() {
'use strict';

arc.detaildata = arc.detaildata || {};

const columnNameRegex = /(\[\w+\]) as ['"`]\w+['"`]/;
const columnNameNoAliasRegex = /^\[(\w+)\]$/;

arc.detaildata.display = function() {
  let display = function() { return 'arc.detaildata.display'; };

  let activeQueryManager = arc.data.activeQueryManager(),
      modalRowMax = 200,
      detailDataColMax = 1000,
      dimensions,
      predicates,
      viz,
      contextMenu,
      combineWithOr = false,
      baseQuery;

  display.viz = function(_) {
    if (!arguments.length) return viz;
    viz = _;
    return display;
  };

  display.contextMenu = function(_) {
    if (!arguments.length) return contextMenu;
    contextMenu = _;
    return display;
  };

  display.combineWithOr = function(_) {
    if (!arguments.length) return combineWithOr;
    combineWithOr = _;
    return display;
  };

  display.showDetail = function() {
    const $dataDetailModal = $('#datadetail-modal');
    if (arc.misc.modalShown($dataDetailModal)) {
      console.log('modal already open');
      return;
    }

    cleanupTables();
    initializePredicatesAndDimensions();
    showDetailInternal();
  };

  function showDetailInternal(withinModal) {
    const rb = buildRequestBuilder(true);

    let deduper = arc.misc.columnDeduper();
    deduper.initializeWithRB(rb);

    rb.dimensions(deduper.dedupedDims());
    rb.aggregates(deduper.dedupedAggs());

    runQueryAndDisplayData(rb, withinModal);
  }

  function initializePredicatesAndDimensions() {
    dimensions = [];
    predicates = [];

    let colnameParser = arc.grammar._colnames;

    /*
     * Predicates in detail data should include visual filters...
     */
    try {
      let vizDsreq = JSON.parse(viz.shelvesToDataRequest().dsreq());
      let vizFilters = vizDsreq.filters;
      if (vizFilters) {
        vizFilters.forEach(f => predicates.push(f));
      }
    } catch (err) {
      /*
       * There may be some cases where the expected dsreq is not created.
       * Defensively, permit those cases to log an error and continue on
       * with the filters in the marks
       */
      console.log(err);
    }

    /*
     * ...and they should include dimensional predicates from the mark clicked on
     */
    let activeParams = contextMenu.params();
    activeParams.forEach(function(p) {
      p.forEach(function(c) {
        let e = c.expr;
        let custom_e = c.custom_expr;
        if (!e || custom_e) {
          return;
        }

        if (c.keytype === arc.report.columnTypeEnum.DIMENSION) {
          const match = columnNameNoAliasRegex.exec(e);
          dimensions.push({
            expr: match ? `${e} as '${match[1]}'` : e,
            colname: c.key,
            coltype: 'STRING',
          });

          let v = c.value;
          if (arc.misc.isString(c.valuetype)) {
            v = "'" + v + "'";
          }

          let pred = arc.misc.whereClauseFromValues(
              e, [v], undefined, false, viz.dataconnection_type(),
              c.valuetype);
          predicates.push(pred);
          return;
        }
        let d = colnameParser.parse(e);
        let cols = d.cols;
        cols.forEach(function(c) {
          dimensions.push({
            expr: '[' + c + ']',
            colname: c,
            coltype: 'STRING',
          });
        });
      });
    });

    predicates = arc.misc.arrayUnique(predicates);
    if (combineWithOr && predicates.length > 1) {
      const newPredicate = predicates.join(' or ');
      predicates = [newPredicate];
    }
  }

  function predicatesToParams(rb) {
    rb.replaceFilters([]);

    predicates.forEach(function(pred) {
      rb.addFilter(pred);
    });
  }

  function processDataset() {
    let dataset = viz.dataset();
    const removedCols = [];
    dataset.tables().forEach(function(t) {
      t.columns().forEach(function(c) {
        if (c.attr('hidden')) {
          return;
        }

        let colname = c.display_name();
        if (c.showInDetailData()) {
          let expr = '[' + colname + "] as '" + colname + "'";
          /**
           * if the column has a custom expression with aggregate functions
           * (e.g sum, avg etc.) The idea here is to get all the columns
           * from the formula and add them one by one as dimensions.
           *
           * but...👇
           */
          if (c.attr('exprhasagg')) {

            /**
             * 👆...it might be added from activeParams, but we need to
             * remove the base column because we're about to add
             * the columns from the formula.
             *
             * THE REASON: is because when we want the detailed data, we want to
             * see the underlying data from the table, we're not interrested
             * in aggregated values.
             */
            dimensions = dimensions.filter(dim => dim.colname !== colname);

            const colsFromExpr = arc.grammar._colnames.parse(c.expr()) || {};
            if (colsFromExpr.cols && colsFromExpr.cols.length > 0) {
              dimensions = dimensions.concat(colsFromExpr.cols.map((col) => ({
                expr: `[${col}] as '${col}'`,
                colname: col
              })));
            }
            removedCols.push({
              colname,
              colsFromExpr: colsFromExpr.cols
            });
          } else {
            dimensions.push({
              expr: expr,
              colname: colname,
              coltype: c.attr('type'),
            });
          }
        }
      });
    });

    function createRowText(removedCol) {
      return `<li>
        <strong>
          ${arc.misc.escapeHTML(removedCol.colname)}
        </strong>:
        See ${removedCol.colsFromExpr.map(arc.misc.escapeHTML).join(', ')}
      </li>`;
    }

    if (removedCols.length > 0) {
      const $alert = $('#datadetail-complex-expr-alert');
      $alert.html(
        `The following fields are not displayed because their expression contains aggregates.
        Instead, the columns used in the expressions are individually displayed.:
        <ul>
          ${removedCols.map(createRowText).join('')}
        </ul>
      `);
      $alert.removeClass('arc-invisible');
    }
  }

  function buildRequestBuilder(enforceRowLimit = false) {
    let rb = viz.getFilteredRequest();
    rb.resetHaving();
    if (enforceRowLimit) {
      rb.limit(modalRowMax);
    }

    predicatesToParams(rb);

    // When the user shows detailed data on a specific cell,
    // the column of the cell is pushed onto the `dimensions` array
    // with the value of `expr: [column_name]`
    // Problem with that is because processDataset() adds the rest
    // of the columns with `expr: [column_name] as 'column_name'`
    processDataset();

    // Check that array is unique by only taking the column name
    // in square brackets, resolving the disparity mentioned above.
    // Just in case, also return the `expr` if regex fails.
    dimensions = arc.misc.arrayUnique(dimensions, d => {
      const match = columnNameRegex.exec(d.expr);
      if (match) {
        return match[1];
      }

      return d.expr;
    }, true);

    if (arc.globals.dashboard) {
      detailDataColMax = arc.globals.dashboard.user_settings()
                                              .find_by_id('detailed_data_max_columns')
                                              .value();
    }

    let colIndex = 0;
    while (colIndex < detailDataColMax &&
           colIndex < dimensions.length) {
      rb.addDimension(dimensions[colIndex].expr);
      colIndex++;
    }

    return rb;
  }

  function cleanupTables() {
    viz.removeDataTables('datadetail-viz');
    $('#datadetail-viz > :not(.viz-permanent)').remove();

    const $dataDetailModal = $('#datadetail-modal');
    $dataDetailModal
        .off('shown.bs.modal');
  }

  function runQueryAndDisplayData(rb, withinModal) {
    const $dataDetailModal = $('#datadetail-modal');
    if (!withinModal && arc.misc.modalShown($dataDetailModal)) {
      return;
    }

    $dataDetailModal.find('.filters-section').show();

    const $alert = $dataDetailModal.find('#datadetail-alert');
    const $spinner = $dataDetailModal.find('#datadetail-spinner');
    $alert.addClass('arc-invisible');
    $spinner.show();

    if (!withinModal) {
      $dataDetailModal.modal('show', {
        keyboard: true,
      });

      $dataDetailModal
          .offOn('hidden.bs.modal', function() {
            cleanupTables();
            const $alertComplexExpr = $dataDetailModal
              .find('#datadetail-complex-expr-alert');
            $alertComplexExpr.addClass('arc-invisible');
            $('#datadetail-viz').removeAttr('data-microquery-version');
          });

      contextMenu.hide();
    }

    const $filters = $dataDetailModal.find('.filter-pills');
    $filters.empty();

    let predicatesH = {};
    predicates.forEach(function(pred, idx) {
      idx = String(idx);
      predicatesH[idx] = pred;
      $filters.append(arc.misc.safe_substitute(
          $('#removable-filter-pill-template').html(),
          {
            TEXT: pred,
            IDX: idx,
          }));
    });

    $filters.find('.filter-pill .remove')
        .offOn('click.remove', function() {
          let $filterPill = $(this).closest('.filter-pill');
          let idx = $filterPill.attr('data-idx');
          delete predicatesH[idx];

          predicates = [];
          Object.keys(predicatesH).sort().forEach(function(k) {
            predicates.push(predicatesH[k]);
          });

          showDetailInternal(true /* withinModal */);
        });

    setupTableViz();
    if (withinModal) {
      displayDetailDataTable();
    } else {
      $dataDetailModal.offOn('shown.bs.modal', displayDetailDataTable);
    }

    function displayDetailDataTable() {
      let dataRequest = arc.data.request();
      dataRequest.rb(rb);
      dataRequest.dsreq(rb.dsreq());
      dataRequest.columns(dimensions.map(function(d, i) {
        return arc.data.column()
                  .colname(d.colname)
                  .colindex(i);
      }));
      dataRequest = viz.afterDetailedDataRequestCb(dataRequest);
      rb = dataRequest.rb();

      let dctype = viz.dataconnection_type();
      if (dctype === 'ksql') {
        let runner = arc.ext.streamingQueryRunner({
          lib: driver,
        });

        runner.setRequestBuilder(rb);
        runner.dataCb(function (config) {
          let {
            result,
          } = config;

          displayFetchedData(dataRequest, result);
        });

        runner.startQuerying();
      } else {
        baseQuery = viz.runQuery(rb, function successCb(data) {
          displayFetchedData(dataRequest, data);
        }, function failureCb() {
          // Failure callback
          $spinner.hide();
        });
      }
    }
  }

  let table,
      driver;

  function setupTableViz() {
    table = arc.visual.plugins.table();
    driver = arc.report.driver(table)
                .chart_div('datadetail-viz')
                .params(arc.params.tracker());

    let s = driver.user_settings();
    s.find_by_id('table_full_width')
     .value('true');

    s.find_by_id('table_template')
     .value('table-template-6');

    // Set the download_rows setting from the source visual
    let src = viz.user_settings().find_by_id('download_rows');
    let tgt = s.find_by_id('download_rows');
    if (src && tgt) {
      tgt.value(src.value());
    }
  }

  function displayFetchedData(dataRequest, data) {
    cleanupTables();

    const $dataDetailModal = $('#datadetail-modal');
    const $alert = $('#datadetail-alert');
    const $spinner = $('#datadetail-spinner');
    $spinner.hide();

    const minimumHeight = 300,
          offsetHeight = 400;

    let $viz = $('#datadetail-viz'),
        h = ($(window).height() - $('#datadetail-viz').position().top - offsetHeight);

    h = Math.max(h, minimumHeight);
    $viz.height(h + 'px');

    let cols = [];
    data.colnames.forEach(function(colname, coli) {
      cols.push(arc.data.column()
                .colname(dataRequest.columns()[coli].colname())
                .coltype(data.coltypes[coli])
                .colproperty(arc.report.columnTypeEnum.DIMENSION)
                .colsource('dimensions_shelf')
                .colindex(coli));
    });

    let result = arc.data.result();
    result.columns(cols);
    result.rows(data.rows);
    result.finalize();

    // create the visual by setting the data and then drawing it
    table.viz_setDataResult(result);

    driver.sb_colnames(cols.map(function(c) {
      return c.colname();
    }));

    driver.sb_rows(data.rows);
    table.setConfiguration({
      cellsHtml: false,
      showTooltip: false,
      showDetailedData: true,
    });
    table.viz_drawVisual('datadetail-viz');

    let dctype = viz.dataconnection_type();
    const modalDataMaxedOut = (dctype === 'ksql') ? false : (data.rows.length === modalRowMax);

    // Here we assume if we got the row limit that the download size would provide more rows
    if (modalDataMaxedOut) {
      $alert.text(
        `Only ${modalRowMax} rows are being displayed. Click download button above to see more.`
      );
      $alert.removeClass('arc-invisible');
    }

    arc.misc.toggleColWarningByTotalCount(result, display.viz());

    $dataDetailModal.find('.detaildata-download')
        .offOn('click', function() {

          // Make sure to get full dataset for CSV download.  Any row limits particular
          // to download will be enforced in exportCSV
          if (modalDataMaxedOut) {
            const query = viz.runQuery(buildRequestBuilder(), function exportAllRows(data) {
              driver.sb_rows(data.rows);
              driver.exportCSV({ query, forceFullDownload: true });
            });
          } else {
            driver.exportCSV({ query: baseQuery, forceFullDownload: true });
          }
        });
  }

  return display;
};

}());
