window.arc = window.arc || {};

(function() {
'use strict';

arc.transition = arc.transition || {};

let transitionIdx = 10000;

const MAX_TRANSITION_MARKS = 1000;

function setupTooltips(cellConfig, drawConfig) {
  const {
    text_fn,
    params_fn,
    lib,
  } = drawConfig;

  const {
    hiddenCanvasConfig,
    highlightCanvasConfig,
    overlayContainer,
  } = cellConfig;

  function highlight_fn(row) {
    if (highlightCanvasConfig.drawRows) {
      const rows = row ? [row] : [];
      highlightCanvasConfig.drawRows(rows);
    }
  }

  if (cellConfig.marks) {
    cellConfig.marks.destroy();
  }

  cellConfig.marks = arc.canvas.handleTooltips({
    mouseoverCanvas: overlayContainer,
    hiddenCanvas: hiddenCanvasConfig.canvas,
    hiddenColorMap: hiddenCanvasConfig.colorMap,
    text_fn,
    params_fn,
    highlight_fn,
    viz_lib: lib,
  });
}

function drawVisualTransition(config, cellConfig, drawConfig, oldData, newData) {
  const {
    lib,
    drawCanvas,
    drawCanvasAxes,
    dataToGroup,
  } = drawConfig;

  /*
   * The transitions themselves may start after a bit of time. So start with drawing
   * the old state of the canvas synchronously.
   */
  cellConfig.computedData = oldData;
  arc.canvas.drawAllCanvas(config, cellConfig, drawCanvas, /* inTransition */ true);

  /*
   * Note that we do not transition axes as of right now. If we need to draw axes,
   * do it now.
   */
  if (drawCanvasAxes) {
    cellConfig.computedData = newData;
    arc.canvas.drawAllCanvas(config, cellConfig, drawCanvasAxes, /* inTransition */ true);
  }

  const transition = arc.transition.transition()
                      .visualId(lib.report_id())
                      .select(cellConfig.mainCanvasConfig.canvas.node())
                      .profileData([
                        [
                          'rows',
                          newData.length,
                        ],
                        [
                          'visual',
                          lib.viz_type(),
                        ],
                      ]);

  /*
   * Match the individual elements in oldData with newData for smoother transitions.
   */
  const oldDataReordered = [];

  {
    const groupToIdx = {};
    newData.forEach((d, i) => {
      groupToIdx[dataToGroup(d)] = i;
    });

    oldData.forEach((d) => {
      const newIdx = groupToIdx[dataToGroup(d)];
      if (newIdx !== undefined) {
        oldDataReordered[newIdx] = d;
      }
    });
  }

  // Setup the interpolator
  // eslint-disable-next-line no-undef
  const interpolation = d4.interpolate(oldDataReordered, newData);

  transition
    .drawFn((t) => {
      let rowsData;

      let inTransition = true;

      if (t >= 1) {
        /*
         * Note that if we are not in transition i.e. this is the final
         * render, then we use the passed-in newData. This way, we do not
         * need to trust that the interpolator returns newData in its
         * final run.
         */
        inTransition = false;
        rowsData = newData;
      } else {
        rowsData = interpolation(t);
      }

      cellConfig.computedData = rowsData;
      arc.canvas.drawAllCanvas(config, cellConfig, drawCanvas, inTransition);
      if (!inTransition) {
        if (drawCanvasAxes) {
          arc.canvas.drawAllCanvas(config, cellConfig, drawCanvasAxes, inTransition);
        }

        setupTooltips(cellConfig, drawConfig);
      }
    })
    .start();
}

/*
 * This function wraps up a lot of functionality to do with drawing canvas visuals,
 * and transitioning them in a smooth way from their previous state to their current
 * one.
 */
arc.transition.drawTransition = function(config, cellConfig, drawConfig) {
  const {
    drawCanvas,
    startTransitionGenerator,
    previousDataTracker,

    // Note that 'drawCanvasAxes' can be undefined if the canvas has no axes to draw
    drawCanvasAxes,

    // The variables below are not used by 'drawTransition', but are used by other
    // functions that 'drawTransition' function calls, and hence are a part of the
    // contract of 'drawTransition'.
    //
    // lib,
    // dataToGroup,
  } = drawConfig;

  const {
    computedData,
    cellIdx,

    // The variables below are not used by 'drawTransition', but are used by other
    // functions that 'drawTransition' function calls, and hence are a part of the
    // contract of 'drawTransition'.
    //
    // hiddenCanvasConfig,
    // highlightCanvasConfig,
    // overlayContainer,
  } = cellConfig;

  // First, get previously tracked data, and set the previously tracked
  // data for the next iteration.
  let previousComputedData = previousDataTracker.getData(cellIdx);
  previousDataTracker.setData(computedData, cellIdx);

  if (!previousComputedData) {
    if (!previousDataTracker.enabled()) {
      /*
       * If we do not have previous data, and if transitions are not enabled,
       * then just draw the canvas and exit out.
       */
      arc.canvas.drawAllCanvas(config, cellConfig, drawCanvas);
      if (drawCanvasAxes) {
        arc.canvas.drawAllCanvas(config, cellConfig, drawCanvasAxes);
      }

      setupTooltips(cellConfig, drawConfig);
      return;
    }

    /*
     * If we do not have previous data, and if transitions are enabled,
     * then generate the previous data and use it as a part of the transition.
     */
    previousComputedData = [];
    computedData.forEach((d) => {
      const startd = startTransitionGenerator(d);
      if (startd !== undefined) {
        previousComputedData.push(startd);
      }
    });
  }

  // Draw the canvas with transitions
  drawVisualTransition(config, cellConfig, drawConfig, previousComputedData, computedData);
};

arc.transition.transition = function() {
  const f = function() { return 'arc.misc.transition'; };

  const idx = transitionIdx;
  transitionIdx += 1;
  const startTime = new Date();
  let visualId;
  let select;
  let transitionName = 'viz';
  let drawFn;
  let duration = 250;
  let profileData;

  f.duration = function(_) {
    if (!arguments.length) return duration;
    duration = _;
    return f;
  };

  f.drawFn = function(_) {
    if (!arguments.length) return drawFn;
    drawFn = _;
    return f;
  };

  f.visualId = function(_) {
    if (!arguments.length) return visualId;
    visualId = _;
    return f;
  };

  f.select = function(_) {
    if (!arguments.length) return select;
    select = _;
    return f;
  };

  f.transitionName = function(_) {
    if (!arguments.length) return transitionName;
    transitionName = _;
    return f;
  };

  f.profileData = function(_) {
    if (!arguments.length) return profileData;
    profileData = _;
    profileData.push([
      'transitionIdx',
      idx,
    ]);
    return f;
  };

  f.idx = function() {
    return idx;
  };

  f.start = function() {
    setTimeout(startInternal, 0);
    return f;
  };

  function startInternal() {
    // eslint-disable-next-line no-undef
    d4.select(select)
      .transition(transitionName)
      .on('start', () => {
        arc.counters.increment('transitionStart');
      })
      .on('end', f.complete)
      .on('interrupt', f.complete)
    // eslint-disable-next-line no-undef
      .ease(d4.easeLinear)
      .duration(duration)
      .attrTween('data-transition', () => function(t) {
          const t1 = new Date();

          drawFn(t);

          const t2 = new Date();

          arc.perf.PAGEPERF.addProfile({
            type: 'transition',
            id: idx,
            time: t2 - t1,
            profile: profileData,
          });

          return `${t}`;
        });
  }

  f.complete = function() {
    arc.counters.increment('transitionEnd');
    const t4 = new Date();
    arc.perf.PAGEPERF.addProfile({
      type: 'transition',
      id: visualId,
      time: t4 - startTime,
      profileData,
    });
  };

  return f;
};

/*
 * Keep track of data across calls to viz_drawVisual below
 * so we can transition from one set of marks to another
 * smoothly, if so desired.
 */
arc.transition.dataTracker = function() {
  const f = function() { return 'arc.misc.dataTracker'; };

  let data = [];
  let enabled;
  let lib;
  const maxRowCount = MAX_TRANSITION_MARKS;

  // When transitioning a visual between fullscreen and otherwise,
  // remembering previous drawn state leads to non-smooth transitions.
  // In this case, we start transitions from scratch by forgetting
  // previous state
  let lastKnownInModal;

  f.enabled = function(_) {
    if (!arguments.length) {
      if (!arc.globals.enable_visual_transitions) {
        return false;
      }

      if (enabled && arc.globals.dashboard) {
        // Dashboard settings override visual settings
        return arc.globals.dashboard.enableAnimations();
      }

      return enabled;
    }

    enabled = _;
    return f;
  };

  f.enableOnRowCount = function(rowCount) {
    if ((lib && !lib.enableAnimations()) ||
        (rowCount > maxRowCount)) {
      f.enabled(false);
    } else {
      f.enabled(true);
    }

    return f;
  };

  f.lib = function(_) {
    if (!arguments.length) return lib;
    lib = _;
    lastKnownInModal = lib.inModal();
    return f;
  };

  f.cleanup = function() {
    data = [];
    lastKnownInModal = lib.inModal();
  };

  f.trellis = function(trellis) {
    if (f.enabled()) {
      data = data.slice(0, trellis.trellisCellsCount());
    } else {
      data = [];
    }
  };

  f.setData = function(d, i) {
    if (f.enabled()) {
      data[i] = d;
    }
  };

  f.getData = function(i) {
    if (f.enabled()) {
      if (lastKnownInModal !== lib.inModal()) {
        f.cleanup();
      }

      return data[i];
    }

    return undefined;
  };

  return f;
};
}());
