window.arc = window.arc || {};

(function() {
"use strict";

/*
 * In shelf element expressions, in display format string, and in filter base
 * field * expressions, we permit variable substitution via the usage of
 * <<varname>> or <<varname:defaultvalue>> syntax.
 *
 * If settings.ENABLE_SCRIPT_PARAMETERS is set to true, we also permit <script>
 * tags in these locations to be executed, and the <script> tag being replaced
 * by the return value of the script before execution of the next phase.
 *   (Note that <script> tags are not secure, and should not be in used in
 *   externally facing production. They are meant to permit quick progress
 *   during POCs and for feedback about missing features to be given to
 *   engineering)
 *
 * This implementation is done via a grammer in viztemplates.g, which has been
 * translated by peg.js and the parser made available in
 * arc.grammar._viztemplates.  We use the output of the parser here to actually
 * convert entered expressions etc. into a string.
 *
 * The parser invoked by arc.grammar._viztemplates.parse return an array of two
 * elements. The first element is a dict of key-value pairs that are default
 * parameters for variables. The second element is a series of elements, each of
 * which is of type 'str' (indicating raw string), 'variable' (indicating a
 * variable to be substituted, or 'code' (indicating js code to be executed)
 */

arc.misc = arc.misc || {};

arc.misc.viztemplates = function() {
  var vt = function() { return "arc.misc.viztemplates"; };

  const m = moment.utc(),
        mlocal = moment();

  var kvs = {
        'arc.sys.username': arc.globals.username,
        'arc.sys.utcdate': m.format('YYYY-MM-DD'),
        'arc.sys.utcdatetime': m.format('YYYY-MM-DD HH:mm:ss'),
        'arc.sys.localdate': mlocal.format('YYYY-MM-DD'),
        'arc.sys.localdatetime': mlocal.format('YYYY-MM-DD HH:mm:ss'),
      },
      defaults = {};

  vt.addTemplate = function(template) {
    if (!template) {
      return;
    }
    var s = arc.grammar._viztemplates.parse(template);
    Object.keys(s[0]).forEach(key => defaults[key] = s[0][key]);
    return vt;
  };

  vt.addKV = function(k, v) {
    kvs[arc.params.convertKey(k)] = v;
    return vt;
  };

  function defaultSubstitute(newvalue, varname) {
    return newvalue;
  }

  vt.transformInput = function(input, substituteFn, leave_code_alone,
      parametersSubstituted) {
    if (!substituteFn) {
      substituteFn = defaultSubstitute;
    }

    var s = arc.grammar._viztemplates.parse(input);
    var parseBits = s[1];

    var outputl = [];

    function getSubstitutedParamValue(hash, varname) {
      const casedKeys = arc.globals.parameters_case_sensitive ?
          Object.keys(hash) :
          Object.keys(hash)
                .map(k => k.toLowerCase());

      const casedVarname = arc.globals.parameters_case_sensitive ? varname : varname.toLowerCase();

      if (casedKeys.includes(casedVarname)) {
        if (parametersSubstituted) {
          parametersSubstituted.push({
            key: varname,
            value: hash[arc.params.convertKey(varname)],
          });
        }

        return substituteFn(hash[casedVarname], varname);
      }

      return undefined;
    }

    function getVar(varname) {
      const substitutedValue = getSubstitutedParamValue(kvs, varname);
      if (substitutedValue) {
        return substitutedValue;
      }

      return getSubstitutedParamValue(defaults, varname);
    }

    var special_value = false;
    parseBits.forEach(function(p) {
      // 'str' elements are copied out as-is
      if (p.element === "str") {
        outputl.push(p.str);
        // Push non-substituted parameters to paramsSubstituted as well.
        if (typeof getVar(p.str) === 'undefined') {
          if (typeof getVar(p.str + '.data') === 'undefined') {
            getVar(p.str + '.start');
            getVar(p.str + '.end');
          }
        }
        return;
      }

      // 'variable' elements are replaced with variable values
      if (p.element === "variable") {
        var v = getVar(p.varname);
        if (v === '_arc_record_count_') {
          // if the value of the param is a special one indicating `Record Count`,
          // then reset the expression to `sum(1)`. Currently this scenario happens
          // only in dashboard template
          special_value = true;
        }
        if (v !== undefined) {
          outputl.push(v);
          return;
        }
        outputl.push("<<" + p.varname + ">>");
        return;
      }

      // 'code' elements are replaced with executed code
      if (p.element === "code") {
        // Check if settings.ENABLE_SCRIPT_PARAMETERS is set
        // or if we have been asked to leave code alone
        if (leave_code_alone || !arc.globals.enable_script_parameters) {
          outputl.push("<script>" + p.code + "</script>");
          return;
        }

        var arcapi = {};
        arcapi.getVar = getVar;
        try {
          var extensionFunc = new Function("arcapi", p.code)
              .bind(undefined, arcapi);
          var f = extensionFunc();
          if (f !== undefined) {
            outputl.push(f);
          }
        } catch (e) {
          outputl.push(p.code);
        }
      }
    });

    if (special_value) {
      return 'sum(1)';
    }
    return outputl.join("");
  };

  return vt;
};

}());
