/*
 * Given an expression of the type:
 *   a AND b AND c=12 AND d=error
 *
 * generate a where clause that can be used in a select clause
 *
 * This can be plugged in into a search box on the top of a dashboard that leads to
 *   where predicates being emitted by it
 */

{
  this.searchPredicate = function() {
    var s = function() { return 'searchPredicate'; };

    var colnames,
        coltypes,
        defaultSearchColnames,
        collections = {};

    var columnsHash;

    s.columns = function(_colnames, _coltypes) {
      colnames = _colnames;
      coltypes = _coltypes;

      columnsHash = {};

      colnames.forEach(function(colname, i) {
        columnsHash[colname] = coltypes[i];
      });

      return s;
    };

    s.colnames = function(_) {
      if (!arguments.length) return colnames;
      colnames = _;
      return s;
    };

    s.coltypes = function(_) {
      if (!arguments.length) return coltypes;
      coltypes = _;
      return s;
    };

    s.addSearchCollection = function(collectionName, collectionColnames) {
      collections[collectionName] = collectionColnames;
      return s;
    };

    s.defaultSearchColnames = function(_) {
      if (!arguments.length) return defaultSearchColnames;
      defaultSearchColnames = _;
      return s;
    };

    /*
     * Given a set of columns, create a where perdicate for each column in turn
     * and then OR the predicates together
     */
    function multiCol(colnames, value, exactMatch) {
      var wheres = [];
      colnames.forEach(function(colname) {
        if (exactMatch) {
          wheres.push(colname + " = '" + value + "'");
        } else {
          wheres.push(colname + " like '%" + value + "%'");
        }
      });

      return wheres.join(" OR ");
    }

    function toColumns(colname) {
      if (colname.endsWith('*')) {
        while (colname.endsWith('*')) {
          colname = colname.slice(0, colname.length - 1);
        }

        return colnames.filter(function(c) {
          return c.startsWith(colname);
        });
      }

      return [colname];
    }

    s.convertPredicatesToWhereClause = function(predicate) {
      if (predicate.type === 'AND') {
        return "(" + this.convertPredicatesToWhereClause(predicate.left) +
          ") AND (" + this.convertPredicatesToWhereClause(predicate.right) + ")";
      }

      if (predicate.type === 'OR') {
        return "(" + this.convertPredicatesToWhereClause(predicate.left) +
          ") OR (" + this.convertPredicatesToWhereClause(predicate.right) + ")";
      }

      if (predicate.type === 'NOT') {
        return " NOT (" + this.convertPredicatesToWhereClause(predicate.operand) + ")";
      }

      if (predicate.type === 'APPROX') {
        if (collections[predicate.key]) {
          return multiCol(collections[predicate.key], predicate.value);
        }

        return multiCol(toColumns(predicate.key), predicate.value);
      }

      if (predicate.type === 'EXACT') {
        if (collections[predicate.key]) {
          return multiCol(collections[predicate.key], predicate.value, true);
        }

        return multiCol(toColumns(predicate.key), predicate.value, true);
      }

      return multiCol(defaultSearchColnames, predicate.value);
    };

    return s;
  };
}

start
  = implicit_and

implicit_and
  = left:logical_or ws+ right:implicit_and { return {type: "AND", left:left, right:right} }
  / logical_or

logical_or
  = left:logical_and ws+ "or"i ws+ right:logical_or { return {type: "OR", left:left, right:right} }
  / logical_and

logical_and
  = left:factor ws+ "and"i ws+ right:logical_and { return {type: "AND", left:left, right:right} }
  / factor

factor
  = "not"i ws+ operand:factor { return {type: "NOT", operand: operand } }
  / primary

primary
  = doubleequality
  / equality
  / token
  / "(" logical_or:logical_or ")" { return logical_or; }

doubleequality
  = s1:colname ws* "==" ws* s2:token { return {type: "EXACT", key: s1.value, value: s2.value} }

equality
  = s1:colname ws* "=" ws* s2:token { return {type: "APPROX", key: s1.value, value: s2.value} }

token
  = unquoted_str
  / singlequoted_str
  / doublequoted_str

singlequoted_str
  = "'" s:singlequoted_str_inside "'" { return {value: s}; }

doublequoted_str
  = '"' s:doublequoted_str_inside '"' { return {value: s}; }

colname
  = c:[\*a-zA-Z0-9_]+ { return {value: c.join("")}; }

unquoted_str
  = c:[\.a-zA-Z0-9_\,\;\*]+ { return {value: c.join("")}; }

singlequoted_str_inside
  = c:[ \t\.a-zA-Z0-9_\(\)\,\;\*"]+ { return c.join(""); }

doublequoted_str_inside
  = c:[ \t\.a-zA-Z0-9_\(\)\,\;\*']+ { return c.join(""); }

ws
  = [ \t\n]

