/*
 * We create the following tokens from phrases entered in:
 *
 * comma: ,
 *
 * top: top <whole>
 * bottom: bottom <whole>
 * top: top
 * bottom: bottom
 *
 * (that/which/who) start(s) with <str>
 * (that/which/who) begin(s) with <str>
 * starting with <str>
 * beginning with <str>
 * (that/which/who) end(s) with <str>
 * ending with <str>
 *
 * Comparisons:
 *   is/are equal to <str|number>
 *   is/are equals to <str|number>
 *   is/are in/among/within <str|number>
 *   equals/is/are <str|number>
 *
 *   is/are not equal/equals to <str|number>
 *   is/are not in/among/within <str|number>
 *   is_str not <str|number>
 *   != <str|number>
 *
 *   >= <numeric>
 *   not less than <numeric>
 *   <= <numeric>
 *   not greater than <numeric>
 *   (is) greater / longer / over / bigger / > / larger / higher (than) <number>
 *   (is) lesser / shorter / under / smaller / < / lower (than) <number>
 *
 * or
 * and
 *
 * <key>:<value>
 *
 * average / avg / mean
 * median
 * sum / total
 * count distinct / count unique / distinct count / unique count
 * count
 * approximate count / approx count
 * min / minimum
 * max / maximum
 * population <standard deviation>
 * standard deviation / stddev
 * population <variance>
 * variance
 *
 * whitespace: 1 or more of [ \t\r\n\xA0]
 *
 * str: "<str>"
 *      Note that quoted strings can contain whitespace, and \, ", ' escaped with \
 * str: '<str>'
 *      Note that quoted strings can contain whitespace, and \, ", ' escaped with \
 * str: 1 or more of non-whitespace characters
 */

{
  input = input + ' ';
  var _in = input,
      tokens = [],
      addComma,
      lastPhraseStart,
      lastWhitespaceColumn;

  var newTokens = [];

  function setToken(token) {
    newTokens = [token];
  }

  function agg(aggType) {
    setToken({
      phrase: aggType,
      type: 'aggregation',
      aggregation: aggType,
    });
  }

  function subjectCompare(str, type, term) {
    newTokens = [];
    newTokens.push({
      type: 'str',
      rule: 'subjectcompare',
      phrase: str,
    });

    newTokens.push({
      type: 'compare',
      subtype: type,
      term: term,
      phrase: type + ' ' + term,
    });
  }

  function compare(type, term) {
    setToken({
      type: 'compare',
      subtype: type,
      term: term,
      phrase: type + ' ' + term,
    });
  }

  function getBetween(r1, r2, t, c) {
    var r1_number = +(r1);
    var r2_number = +(r2);

    setToken({
      source: t,
      type: 'compare',
      subtype: 'between',
      phrase: t + ' ' + r1 + ' ' + c + ' ' + r2,
      term: t + ' ' + r1 + ' ' + c + ' ' + r2,
      greater_than_term: (r1_number > r2_number) ? r2 : r1,
      less_than_term: (r1_number > r2_number) ? r1 : r2,
    });
  }

  function createConnector(type) {
    return {
      phrase: type,
      type: 'connector',
      subtype: type,
    };
  }

  function connector(type) {
    setToken(createConnector(type));
  }

  function topBottom(type, count) {
    setToken({
      phrase: type + ' ' + count,
      type: type,
      k: +count,
    });
  }

  function handleNewTokens() {
    if (newTokens.length > 0) {
      var newToken;
      newTokens.forEach(function(_) {
        newToken = _;
        if (newToken.startIndex === undefined) {
          newToken.startIndex = lastPhraseStart - 1;
          newToken.endIndex = lastWhitespaceColumn - 1;
          if (addComma) {
            newToken.endIndex -= 1;
          }
        }

        tokens.push(newToken);
      });

      if (addComma) {
        var t = createConnector(',');
        t.startIndex = newToken.endIndex + 1;
        t.endIndex = newToken.endIndex + 1;

        addComma = false;
      }

      newTokens = [];
    }
  }

  function finalProcessingTokens(tokens) {
    if (tokens.length > 0) {
      var t = tokens[tokens.length - 1];
      t.endIndex = t.endIndex - 1;
    }

    return {tokens: tokens};
  }
}

expr
  = whitespace? (onephrase)* {
    if (tokens.length > 0) {
      var t = tokens[tokens.length - 1];
      t.endIndex = _in.length;
    }

    return finalProcessingTokens(tokens);
  }

strings
  = whitespace? (allstrings_ws)* allstrings_process? {
    if (tokens.length > 0) {
      var t = tokens[tokens.length - 1];
      t.endIndex = _in.length;
    }

    return finalProcessingTokens(tokens);
  }

allstrings_raw
  = allstrings_simple {
    lastPhraseStart = column();
  }

allstrings_process
  = allstrings_raw {
    handleNewTokens();
  }

allstrings_ws
  = allstrings_raw whitespace_marked {
    handleNewTokens();
  }

comma_or_whitespace
  = comma:(comma_str whitespace_marked* / comma_str* whitespace_marked) {
    return comma;
  }

onephrase_raw
  = (relative_dates comma:comma_or_whitespace / allaggs comma:comma_or_whitespace / keyvalue comma:comma_or_whitespace / alltop comma:comma_or_whitespace / allcompares comma:comma_or_whitespace / allconnectors comma:comma_or_whitespace / allstrings comma:comma_or_whitespace) {

    try {
      if (comma.join('') !== '') {
        addComma = true;
      }
    } catch (err) {
    }

    lastPhraseStart = column();
  }

whitespace_marked
  = whitespace {
    lastWhitespaceColumn = column();
  }

onephrase
  = onephrase_raw {
    handleNewTokens();
  }

/*
 * date_period:
 *     century/decade/year/quarter/month/day/hour/minute
 *
 * <this/current/last/previous/next> <whole>? <date_period>
 *     (<starting/ending> <with/on/from/at> <datetime_string>)?
 */
relative_dates
  = (curr_relative_dates / prev_relative_dates / next_relative_dates)

curr_relative_dates
  = curr_str whitespace p:datetime_period {
    setToken({
      type: 'datetime',
      subtype: 'relative',
      rule: 'curr_relative_dates',
      phrase: 'this ' + p,
      period_prevnext: 'curr',
      period: p.toLowerCase(),
    });
  }

prev_relative_dates
  = prev_relative_dates_num / prev_relative_dates_nonum

prev_relative_dates_num
  = previous_str whitespace s:whole+ whitespace p:datetime_period {
    setToken({
      type: 'datetime',
      subtype: 'relative',
      rule: 'prev_relative_dates',
      phrase: 'last ' + s.join('') + ' ' + p,
      period_prevnext: 'prev',
      period_count: +(s.join('')),
      period: p.toLowerCase(),
    });
  }

prev_relative_dates_nonum
  = previous_str whitespace p:datetime_period {
    setToken({
      type: 'datetime',
      subtype: 'relative',
      rule: 'prev_relative_dates',
      phrase: 'last ' + p,
      period_prevnext: 'prev',
      period_count: 1,
      period: p.toLowerCase(),
    });
  }

next_relative_dates
  = next_relative_dates_num / next_relative_dates_nonum

next_relative_dates_num
  = next_str whitespace s:whole+ whitespace p:datetime_period {
    setToken({
      type: 'datetime',
      subtype: 'relative',
      rule: 'next_relative_dates',
      phrase: 'next ' + s.join('') + ' ' + p,
      period_prevnext: 'next',
      period_count: +(s.join('')),
      period: p.toLowerCase(),
    });
  }

next_relative_dates_nonum
  = next_str whitespace s:whole+ whitespace p:datetime_period {
    setToken({
      type: 'datetime',
      subtype: 'relative',
      rule: 'next_relative_dates',
      phrase: 'next ' + p,
      period_prevnext: 'next',
      period_count: 1,
      period: p.toLowerCase(),
    });
  }

curr_str
  = "current"i / "this"i

previous_str
  = "last"i / "previous"i / "prev"i

next_str
  = "next"i

datetime_period
  = year / quarter / month / week / day / hour / minute / second

year
  = year_str {
    return 'year';
  }

year_str
  = "years"i / "year"i

quarter
  = quarter_str {
    return 'quarter';
  }

quarter_str
  = "quarters"i / "quarter"i

month
  = month_str {
    return 'month';
  }

month_str
  = "months"i / "month"i

week
  = week_str {
    return 'week';
  }

week_str
  = "weeks"i / "week"i

day
  = day_str {
    return 'day';
  }

day_str
  = "days"i / "day"i

hour
  = hour_str {
    return 'hour';
  }

hour_str
  = "hours"i / "hour"i

minute
  = minute_str {
    return 'minute';
  }

minute_str
  = "minutes"i / "minute"i

second
  = second_str {
    return 'second';
  }

second_str
  = "seconds"i / "second"i

/*
 * Implementations for
 * starts with <str>
 * ends with <str>
 * is <str>
 * is equal to <str>
 * is equals to <str>
 * equals <str>
 * between <num> and <num>
 * from <num> to <num>
 */
allcompares
  = (subject_compare / between_compare / from_numeric / contains / greater_equal / lesser_equal / greater / lesser / notequals / equals / startswith / endswith)

allconnectors
  = (comma / or / and / of / from / between / to / the)

comma
  = comma_str {
    connector(',');
  }

of
  = of_str {
    connector('of');
  }

between
  = between_str {
    connector('between');
  }

from
  = from_str {
    connector('from');
  }

to
  = to_str {
    connector('to');
  }

the
  = the_str {
    connector('the');
  }

or
  = or_str {
    connector('or');
  }

and
  = and_str {
    connector('and');
  }

startswith
  = s:(not_str whitespace)? ((that_str whitespace)? starts_str whitespace with_str whitespace allstrings) {
    // replace the string token
    var replaceToken = newTokens[newTokens.length - 1];
    var t = replaceToken;

    // and replace it with a compare node
    replaceToken = {
      source: 'startswith',
      type: 'compare',
      subtype: 'startswith',
      startIndex: t.startIndex,
      term: t.phrase,
      phrase: t.phrase,
    };

    if (s !== null) {
      replaceToken.negate = true;
    }

    setToken(replaceToken);
  }

endswith
  = s:(not_str whitespace)? ((that_str whitespace)? ends_str whitespace with_str whitespace allstrings) {
    // replace the string token
    var replaceToken = newTokens[newTokens.length - 1];
    var t = replaceToken;

    // and replace it with a compare node
    replaceToken = {
      source: 'startswith',
      type: 'compare',
      subtype: 'endswith',
      startIndex: t.startIndex,
      term: t.phrase,
      phrase: t.phrase,
    };

    if (s !== null) {
      replaceToken.negate = true;
    }

    setToken(replaceToken);
  }

real_number
  = a:negative? s:real+ t:numeric_suffix? {
    var m = +1;
    if (a) {
      m = -1;
    }

    if (t) {
      var n = 1;

      if ((t === 'k') || (t === 'K')) {
        n = 1000;
      } else if ((t === 'm') || (t === 'M')) {
        n = 1000 * 1000;
      } else if ((t === 'g') || (t === 'G') || (t === 'b') || (t === 'B')) {
        n = 1000 * 1000 * 1000;
      } else if ((t === 't') || (t === 'T')) {
        n = 1000 * 1000 * 1000 * 1000;
      }

      return (+(s.join('')) * n) * m;
    }

    return +(s.join('')) * m;
  }

subject_compare
 = s:(string / sstring / unquoted_subject_str) whitespace? c:comparison_operator whitespace? r:(real_number / allstrings) {
   subjectCompare(s, c, r);
 }

between_compare
  = (t:between_str whitespace r1:(real_number / allstrings) whitespace c:and_str whitespace r2:(real_number / allstrings)) {
    getBetween(r1, r2, t, c);
  }

from_numeric
  = (t:from_str whitespace r1:real_number whitespace c:to_str whitespace r2:real_number) {
    getBetween(r1, r2, t, c);
  }

greater_equal
  = (((is_str whitespace)? greater_equal_str whitespace?) / ((is_str whitespace)? not_str whitespace lesser_str whitespace? than_str? whitespace?)) r:real_number {
    compare('>=', r);
  }

lesser_equal
  = (((is_str whitespace)? lesser_equal_str whitespace?) / ((is_str whitespace)? not_str whitespace greater_str whitespace? than_str? whitespace?)) r:real_number {
    compare('<=', r);
  }

greater
  = ((is_str whitespace)? greater_str whitespace? than_str? whitespace? r:real_number) {
    compare('>', r);
  }

lesser
  = ((is_str whitespace)? lesser_str whitespace? than_str? whitespace? r:real_number) {
    compare('<', r);
  }

contains
  = contains_str whitespace allstrings {
    // replace the string token
    var replaceToken = newTokens[newTokens.length - 1];
    var t = replaceToken;

    // and replace it with a compare node
    replaceToken = {
      type: 'compare',
      subtype: 'contains',
      startIndex: t.startIndex,
      term: t.phrase,
      phrase: t.phrase,
    };

    setToken(replaceToken);
  }

notequals
  = ((is_str whitespace not_str whitespace equal_str whitespace to_str) /
     (is_str whitespace not_str whitespace in_str) /
     (is_str whitespace not_str whitespace among_str) /
     (is_str whitespace not_str whitespace within_str) /
     is_str whitespace not_str /
     notequal_str) whitespace r:(real_number / allstrings) {

    if (typeof (r) === 'number') {
      compare('!=', r);
      return;
    }

    // replace the string token
    var replaceToken = newTokens[newTokens.length - 1];
    var t = replaceToken;

    // and replace it with a compare node
    replaceToken = {
      type: 'compare',
      subtype: 'inequality',
      startIndex: t.startIndex,
      term: t.phrase,
      phrase: t.phrase,
    };

    setToken(replaceToken);
  }

equals
  = ((is_str whitespace equal_str whitespace to_str) /
     (is_str whitespace in_str) /
     (is_str whitespace among_str) /
     (is_str whitespace within_str) /
     is_str /
     equal_str) whitespace r:(real_number / allstrings) {

    if (typeof (r) === 'number') {
      compare('=', r);
      return;
    }

    // replace the string token
    var replaceToken = newTokens[newTokens.length - 1];
    var t = replaceToken;

    // and replace it with a compare node
    replaceToken = {
      type: 'compare',
      subtype: 'equality',
      startIndex: t.startIndex,
      term: t.phrase,
      phrase: t.phrase,
    };

    setToken(replaceToken);
  }

/*
 * Implementation for
 * key:value
 */
keyvalue
  = (k:allstrings ":" v:allstrings) {
    setToken({
      type: 'keyvalue',
      rule: 'keyvalue',
      key: k,
      value: v,
      phrase: k + ':' + v,
    });
  }

/*
 * Implementations for
 * top: top <whole>
 * bottom: bottom <whole>
 * top: top
 * bottom: bottom
 * oldest: old <whole>
 * newest: newest <whole>
 * oldest: old
 * newest: newest
 */
alltop
  = (topk / bottomk / top1 / bottom1 / oldest / newest)

/*
 * Implementations for
 * str: "<str>"
 *      Note that quoted strings can contain whitespace, and \, ", ' escaped with \
 * str: '<str>'
 *      Note that quoted strings can contain whitespace, and \, ", ' escaped with \
 * str: 1 or more of non-whitespace characters except for , and :
 */
allstrings
  = (string / sstring / unquoted_string)

allstrings_simple
  = (string / sstring / unquoted_string_simple)

top1
  = top {
    topBottom('top', '1');
  }

bottom1
  = bottom {
    topBottom('bottom', '1');
  }

topk
  = top whitespace* s:whole+ {
    topBottom('top', s.join(''));
  }

old1
  = oldest {
    topBottom('old', '1');
  }

new1
  = newest {
    topBottom('new', '1');
  }

oldk
  = oldest whitespace* s:whole+ {
    topBottom('old', s.join(''));
  }

newk
  = newest whitespace* s:whole+ {
    topBottom('new', s.join(''));
  }

oldest
  = "oldest"i

newest
  = "latest"i / "newest"i

comma_str
  = ","

of_str
  = "of"i

or_str
  = "or"i / "versus"i / "vs"i

and_str
  = "and"i

article_str
  = "an"i / "a"i / "the"i

with_str
  = "with"i

ends_str
  = "ends"i / "ending"i / "end"i

lesser_str
  = "lesser"i / "shorter"i / "under"i / "<" / "smaller"i / "lower"i / "less"i

greater_equal_str
  = ">" whitespace? "="

lesser_equal_str
  = "<" whitespace? "="

greater_str
  = "greater"i / "longer"i / "over"i / "bigger"i / ">" / "larger"i / "higher"i / "more"i

than_str
  = "than"i / "then"i

that_str
  = "that"i / "which"i / "who"i

starts_str
  = "starts"i / "starting"i / "start"i / "begins"i / "beginning"i / "begin"i

notequal_str
  = "!=" / "excluding"i / "excludes"i / "exclude"i

equal_str
  = "equals"i / "equal"i / "==" / "=" / "including"i / "includes"i / "include"i / "match"i

is_str
  = "is"i / "are"i

not_str
  = "not"i / "don't"i / "doesn't"i

in_str
  = "in"i

among_str
  = "among"i / "amongst"i

within_str
  = "within"i

between_str
  = "between"i

from_str
  = "from"i

to_str
  = "to"i

similar_str
  = "similar"i

like_str
  = "like"i

contains_str
  = "contains"i / (is_str whitespace like_str) / like_str / (is_str whitespace similar_str whitespace to_str) / (similar_str whitespace to_str) / similar_str

the_str
  = "the"i

top
  = "top"i / "highest"i / "longest"i / "best"i / "most"i

bottomk
  = bottom whitespace* s:whole+ {
    topBottom('bottom', s.join(''));
  }

bottom
  = "bottom"i / "lowest"i / "shortest"i / "worst"i / "least"i

numeric_suffix
  = c:[kKmMbBgGtT] {return c; }

negative
  = "-"

real
  = c:[0-9\.] { return c; }

whole
  = c:[0-9] { return c; }

allaggs
  = approximate_count / average / population_stddev / stddev / population_variance / variance / median / sum / countdistinct / count / min / max

population_stddev
  = population_stddev_str {
    agg('STDDEV_POP');
  }

population_stddev_str
  = population_str whitespace stddev_str

population_str
  = "population"i

approximate_count
  = approximate_count_str {
    agg('APPROXIMATE_COUNT');
  }

approximate_count_str
  = ("approximate"i / "approx"i) whitespace "count"i

stddev
  = stddev_str {
    agg('STDDEV');
  }

stddev_str
  = ("stddev"i / (("standard"i / "std"i) whitespace ("deviation"i / "dev"i)))

population_variance
  = population_variance_str {
    agg('VARIANCE_POP');
  }

population_variance_str
  = population_str whitespace variance_str

variance
  = variance_str {
    agg('VARIANCE');
  }

variance_str
  = "variance"i

average
  = average_str {
    agg('AVG');
  }

average_str
  = "average"i / "avg"i / "mean"i

median
  = median_str {
    agg('MEDIAN');
  }

median_str
  = "median"i

sum
  = sum_str {
    agg('SUM');
  }

sum_str
  = "sum"i / "total"i

countdistinct
  = countdistinct_strs {
    agg('COUNTDISTINCT');
  }

countdistinct_strs
  = (count_str whitespace distinct_str) / (distinct_str whitespace count)

distinct_str
  = "distinct"i / "unique"i

count
  = count_str {
    agg('COUNT');
  }

count_str
  = "count"i

min
  = min_str {
    agg('MIN');
  }

min_str
  = "minimum"i / "min"i

max
  = max_str {
    agg('MAX');
  }

max_str
  = "maximum"i / "max"i

string
  = "\"" chars:char* "\"" {
    setToken({
      type: 'str',
      rule: 'string',
      phrase: chars.join(''),
    });

    return chars.join('');
  }

sstring
  = "'" chars:char* "'" {
    setToken({
      type: 'str',
      rule: 'sstring',
      phrase: chars.join(''),
    });

    return chars.join('');
  }

/*
 * Used when looking for words like:
 * channel=nflhd
 * OR
 * population>=2M
 *
 * Handle the lack of spaces in these terms
 */
unquoted_subject_str
  = chars:notspecialsubject+ {
    setToken({
      type: 'str',
      rule: 'unquoted_string',
      phrase: chars.join(''),
    });

    return chars.join('');
  }

/*
 * commas and colons at the end of an unquoted string are handled
 * in a special manner
 */
unquoted_string
  = chars:notspecial+ {
    setToken({
      type: 'str',
      rule: 'unquoted_string',
      phrase: chars.join(''),
    });

    return chars.join('');
  }

unquoted_string_simple
  = chars:notspecial_simple+ {
    setToken({
      type: 'str',
      rule: 'unquoted_string',
      phrase: chars.join(''),
    });

    return chars.join('');
  }

comparison_operator
  = ('>=' / '>' / '<=' / '<' / '=' / '!=')

notspecialsubject
  = c:[^,:\!\=\>\< \t\r\n\xA0] { return c; }

notspecial
  = c:[^,: \t\r\n\xA0] { return c; }

notspecial_simple
  = c:[^ \t\r\n\xA0] { return c; }

char
  = unescaped
  / escape
    sequence:(
        '"'
      / "'"
      / "\\"
    )
    { return sequence; }

escape
  = "\\"

unescaped
  = c:[^\0-\x1F\x22\x27\x5C] { return c; }

whitespace
  = whitespace_char+

whitespace_char
  = w:[ \t\r\n\xA0]
