/* eslint-disable import/no-extraneous-dependencies */
const path = require('path');
const fs = require('fs');
const {
  printResultSummary,
  printFailedTests,
  runQunitWithPage,
} = require('node-qunit-puppeteer');
const puppeteer = require('puppeteer');

const DEBUG_ENABLED = process.env.DEBUG_TEST_EXECUTION === 'true';

function debugLog(msg) {
  if (!DEBUG_ENABLED) {
    return;
  }

  // eslint-disable-next-line no-console
  console.log(msg);
}

const getHtmlFileAbsolutePathsFromDirectory = () => {
  const directoryAbsolutePath = path.join(
    __dirname,
    '../../../../qunit/templates/tests',
  );

  const filePaths = fs.readdirSync(directoryAbsolutePath);

  return filePaths.map((filePath) => filePath.slice(0, filePath.length - 5));
};

const coverageDir = path.join(__dirname, '..', '..', '.nyc_output');

const runTest = async (browser, targetUrl, combinesResults) => {
  let result = {
    modules: {},
  };

  const page = await browser.newPage();

  debugLog(`Loading ${targetUrl}`);

  result = await runQunitWithPage(page, { targetUrl });

  const coverage = await page.evaluate(() => JSON.stringify(window.__coverage__));

  if (coverage?.length > 0) {
    if (!fs.existsSync(coverageDir)) {
      fs.mkdirSync(coverageDir);
    }

    const url = new URL(targetUrl);

    const coverageFile = `${coverageDir}/${
      url.pathname.split('/').slice(-1)[0].split('.')[0]
    }.json`;
    debugLog(`Saving coverage to ${coverageFile}`);
    fs.writeFileSync(coverageFile, coverage);
  }

  await page.close();

  Object.entries(result.modules).forEach(([moduleName, testModule]) => {
    // eslint-disable-next-line no-console
    console.log(
      `${moduleName} ${testModule.failed === 0 ? 'passed' : 'FAILED'}.`,
    );

    combinesResults.modules[moduleName] = testModule;

    testModule.tests.forEach(({ passed, failed, total, runtime }) => {
      combinesResults.stats.passed += passed;
      combinesResults.stats.failed += failed;
      combinesResults.stats.total += total;
      combinesResults.stats.runtime += runtime;

      combinesResults.totalTests += total;
    });
  });
};

const runAll = async () => {
  const filterRegexp =
    process.argv.length > 2 ? new RegExp(process.argv[2]) : null;
  const testUrls = getHtmlFileAbsolutePathsFromDirectory().map(
    (filePath) => `http://localhost:8090/arc/qunit/${filePath}`,
  );

  const combinedResults = {
    modules: {},
    totalTests: 0,
    stats: {
      passed: 0,
      failed: 0,
      total: 0,
      runtime: 0,
    },
  };

  const browser = await puppeteer.launch({
    args: ['--disable-dev-shm-usage', '--no-sandbox'],
  });

  for (const url of testUrls) {
    // skip urls that do not match the filter
    if (!filterRegexp || filterRegexp.test(url)) {
      // eslint-disable-next-line no-await-in-loop
      await runTest(browser, url, combinedResults);
    }
  }

  await browser.close();

  if (combinedResults.stats.failed > 0) {
    printFailedTests(combinedResults, console);
    printResultSummary(combinedResults, console);
    process.exit(1);
  } else {
    printResultSummary(combinedResults, console);
  }
};

runAll().catch((error) => {
  // eslint-disable-next-line no-console
  console.error(error);
  process.exit(1);
});
