QUnit.module('framework_arc_misc', {});

QUnit.test('nulls', (assert) => {
  assert.equal(arc.misc.check_if_value_null('abc'), false);
  assert.equal(arc.misc.check_if_value_null(null), true);
  assert.equal(arc.misc.check_if_value_null(undefined), true);
  assert.equal(arc.misc.check_if_value_null('NULL'), true);
  assert.equal(arc.misc.check_if_value_null('NaN'), true);
  assert.equal(arc.misc.check_if_value_null('inf'), true);
});

QUnit.test('binary search', (assert) => {
  function compareValues(a, b) {
    if (a === b) {
      return 0;
    }

    if (a > b) {
      return 1;
    }

    return -1;
  }

  assert.equal(
    arc.misc.binarySearch({
      target: 5,
      array: [2],
      startIdx: 0,
      endIdx: 0,
      comparator: compareValues,
    }),
    -1,
  );
  assert.equal(
    arc.misc.binarySearch({
      target: 5,
      array: [5],
      startIdx: 0,
      endIdx: 0,
      comparator: compareValues,
    }),
    0,
  );
  assert.equal(
    arc.misc.binarySearch({
      target: 5,
      array: [2, 3, 5, 7],
      startIdx: 0,
      endIdx: 3,
      comparator: compareValues,
    }),
    2,
  );
  assert.equal(
    arc.misc.binarySearch({
      target: 5,
      array: [2, 3, 5, 7],
      startIdx: 3,
      endIdx: 3,
      comparator: compareValues,
    }),
    -1,
  );
});

QUnit.test(
  'convertToStrRep: returns empty string if colValue is null',
  (assert) => {
    const result = arc.misc.convertToStrRep(null, 'STRING');
    assert.equal(result, '', 'Null colValue should return empty string');
  },
);

QUnit.test(
  'convertToStrRep: converts numeric string to quoted string if type is STRING',
  (assert) => {
    const result = arc.misc.convertToStrRep('123', 'STRING');
    assert.equal(result, '123', 'Numeric string should be quoted');
  },
);

QUnit.test(
  'convertToStrRep: sanitizes string containing column separator',
  (assert) => {
    const result = arc.misc.convertToStrRep('Hello, world', 'STRING', {
      columnSeparator: ',',
    });
    assert.equal(
      result,
      '"Hello, world"',
      'String containing column separator should be quoted and sanitized',
    );
  },
);

QUnit.test(
  'convertToStrRep: handles string with different column separator',
  (assert) => {
    const result = arc.misc.convertToStrRep('Hello, world', 'STRING', {
      columnSeparator: ';',
    });
    assert.equal(
      result,
      'Hello, world',
      'String containing comma should not be quoted since the column separator is ;',
    );
  },
);

QUnit.test(
  'convertToStrRep: quotes strings with newline characters (\\n)',
  (assert) => {
    const result = arc.misc.convertToStrRep('Hello\nworld', 'STRING');
    assert.equal(
      result,
      '"Hello\nworld"',
      'String with newline characters should be quoted',
    );
  },
);

QUnit.test(
  'convertToStrRep: quotes strings with carriage return characters (\\r)',
  (assert) => {
    const result = arc.misc.convertToStrRep('Hello\rworld', 'STRING');
    assert.equal(
      result,
      '"Hello\rworld"',
      'String with carriage return characters should be quoted',
    );
  },
);

QUnit.test(
  'convertToStrRep: quotes strings with HTML entity &quot;',
  (assert) => {
    const result = arc.misc.convertToStrRep(
      'Hello &quot;world&quot;',
      'STRING',
    );
    assert.equal(
      result,
      'Hello &quot;world&quot;',
      "String with &quot; entity shouldn't be quoted and escaped",
    );
  },
);

QUnit.test(
  'convertToStrRep: quotes strings with HTML entity &quot; and columnSeparator',
  (assert) => {
    const result = arc.misc.convertToStrRep(
      'Hello, &quot;world&quot;',
      'STRING',
      { columnSeparator: ',' },
    );
    assert.equal(
      result,
      '"Hello, &quot;world&quot;"',
      "String with &quot; entity shouldn't be quoted and escaped and string should be quoted",
    );
  },
);

QUnit.test(
  'convertToStrRep: quotes strings with HTML entity &quot; and with comma',
  (assert) => {
    const result = arc.misc.convertToStrRep(
      'Hello, &quot;world&quot;',
      'STRING',
      { columnSeparator: ';' },
    );
    assert.equal(
      result,
      '"Hello, &quot;world&quot;"',
      "String with &quot; entity shouldn't be quoted and escaped and string should be quoted because HTML entities has columnSeparator",
    );
  },
);

QUnit.test('convertToStrRep: quotes strings with comma', (assert) => {
  const result = arc.misc.convertToStrRep('Hello, "world"', 'STRING', {
    columnSeparator: ',',
  });
  assert.equal(
    result,
    '"Hello, ""world"""',
    'String with quotes inside and comma should double the quotes and string should be quoted',
  );
});

QUnit.test(
  'convertToStrRep: quotes strings with comma but comma is not the columnSeparator',
  (assert) => {
    const result = arc.misc.convertToStrRep('Hello, "world"', 'STRING', {
      columnSeparator: ';',
    });
    assert.equal(
      result,
      'Hello, "world"',
      'String should not be quoted and quotes should not be doubled as columnSeparator is not comma',
    );
  },
);

QUnit.test(
  'convertToStrRep: converts number to string with comma as decimal separator if type is DOUBLE',
  (assert) => {
    const result = arc.misc.convertToStrRep(123.45, 'DOUBLE', {
      columnSeparator: ';',
    });
    assert.equal(
      result,
      '123,45',
      'Number should be formatted with comma as decimal separator',
    );
  },
);

QUnit.test(
  'convertToStrRep: converts number to string with dot as decimal separator if type is DOUBLE',
  (assert) => {
    const result = arc.misc.convertToStrRep(123.45, 'DOUBLE', {
      columnSeparator: ',',
    });
    assert.equal(
      result,
      '123.45',
      'Number should be formatted with dot as decimal separator',
    );
  },
);

QUnit.test(
  'convertToStrRep: converts number to string with comma as decimal separator if type is REAL',
  (assert) => {
    const result = arc.misc.convertToStrRep(123.45, 'REAL', {
      columnSeparator: ';',
    });
    assert.equal(
      result,
      '123,45',
      'Number should be formatted with comma as decimal separator',
    );
  },
);

QUnit.test(
  'convertToStrRep: converts number to string with dot as decimal separator if type is REAL',
  (assert) => {
    const result = arc.misc.convertToStrRep(123.45, 'REAL', {
      columnSeparator: ',',
    });
    assert.equal(
      result,
      '123.45',
      'Number should be formatted with dot as decimal separator',
    );
  },
);

QUnit.test(
  'convertToStrRep: converts number to string with comma as decimal separator if columnSeparator is ; and type is isNumeric',
  (assert) => {
    const result = arc.misc.convertToStrRep(123.45, null, {
      columnSeparator: ';',
    });
    assert.equal(
      result,
      '123,45',
      'Number should be formatted with comma as decimal separator',
    );
  },
);

QUnit.test(
  'convertToStrRep: converts number to string with dot as decimal separator if columnSeparator is , and type is isNumeric',
  (assert) => {
    const result = arc.misc.convertToStrRep(123.45, null, {
      columnSeparator: ',',
    });
    assert.equal(
      result,
      '123.45',
      'Number should be formatted with dot as decimal separator',
    );
  },
);

QUnit.test(
  'convertToStrRep: quotes strings with special characters',
  (assert) => {
    const result = arc.misc.convertToStrRep('Hello "world"', 'STRING');
    assert.equal(
      result,
      'Hello "world"',
      'String with quotes should not be escaped',
    );
  },
);

QUnit.test(
  'convertToStrRep: handles mixed numeric and non-numeric strings',
  (assert) => {
    const result = arc.misc.convertToStrRep('123abc', 'STRING');
    assert.equal(result, '123abc', 'Non-numeric strings should not be quoted');
  },
);

QUnit.test(
  'convertToStrRep: applies default configuration correctly',
  (assert) => {
    const result = arc.misc.convertToStrRep('1000', 'DOUBLE');
    assert.equal(
      result,
      '1000',
      'Default config should not change number formatting',
    );
  },
);
