QUnit.module('framework_arc_params', {
  beforeEach() {
    arc.globals_init().enable_script_parameters(true);
    arc.globals.params = arc.params.tracker();

    this.case_sensitive = arc.globals.parameters_case_sensitive;
  },

  afterEach() {
    arc.globals.parameters_case_sensitive = this.case_sensitive;
  },
});

QUnit.test('historylimit', (assert) => {
  const h = arc.report.params_history();

  const viz = function () {};

  const fns = [];

  // Add in 100 elements of history - confirm that things work
  for (let i = 0; i < 100; ++i) {
    h.add(
      [
        {
          key: 'k1',
          value: i,
        },
      ],
      viz,
    );

    const f = function () {
      const p = h.current_params();

      assert.equal(p.length, 1);
      assert.equal(p[0].length, 1);
      assert.equal(p[0][0].key, 'k1');
      assert.equal(p[0][0].value, i);
    };

    f();

    fns.push(f);
  }

  // Some of the history gets cut off. Check remaining history
  for (let i = 99; i >= 64; --i) {
    fns[i]();
    h.back();
  }

  // Now verify we are back to initial state (not the 63rd state)
  // This is expected because we throw away old history after a while
  const p = h.current_params();
  assert.equal(p.length, 1);
  assert.equal(p[0].length, 1);
  assert.equal(p[1], undefined);
});

QUnit.test('history', (assert) => {
  const h = arc.report.params_history();

  function verifyInitialState(p = h.current_params()) {
    assert.equal(p.length, 1);
    assert.equal(p[0].length, 0);
    assert.equal(p[1], undefined);
  }

  verifyInitialState();
  verifyInitialState(h.getCheckPointState());

  const viz = function () {};

  h.add([
    {
      key: 'k1',
      value: 'v1',
    },
  ]);

  function verifyState2(p = h.current_params()) {
    assert.equal(p.length, 1);
    assert.equal(p[0].length, 1);
    assert.equal(p[0][0].key, 'k1');
    assert.equal(p[0][0].value, 'v1');
  }

  verifyState2();

  h.add(
    [
      {
        key: 'k2',
        value: 'v2',
      },
    ],
    viz,
  );

  function verifyState3(p = h.current_params()) {
    assert.equal(p.length, 1);
    assert.equal(p[0].length, 2);
    assert.equal(p[0][0].key, 'k1');
    assert.equal(p[0][0].value, 'v1');
    assert.equal(p[0][1].key, 'k2');
    assert.equal(p[0][1].value, 'v2');
  }

  verifyState3();

  h.add(
    [
      {
        key: 'k1',
        value: 'v1.2',
      },
    ],
    viz,
  );

  function verifyState4(p = h.current_params()) {
    assert.equal(p.length, 1);
    assert.equal(p[0].length, 2);
    assert.equal(p[0][0].key, 'k2');
    assert.equal(p[0][0].value, 'v2');
    assert.equal(p[0][1].key, 'k1');
    assert.equal(p[0][1].value, 'v1.2');
  }

  verifyState4();

  h.add(
    [
      {
        key: 'ts.start',
        value: '2017-02-12',
      },
      {
        key: 'ts.end',
        value: '2017-02-16',
      },
    ],
    viz,
  );

  function verifyState5(p = h.current_params()) {
    assert.equal(p.length, 1);
    assert.equal(p[0].length, 4);
    assert.equal(p[0][0].key, 'k2');
    assert.equal(p[0][0].value, 'v2');
    assert.equal(p[0][1].key, 'k1');
    assert.equal(p[0][1].value, 'v1.2');
    assert.equal(p[0][2].key, 'ts.start');
    assert.equal(p[0][2].value, '2017-02-12');
    assert.equal(p[0][3].key, 'ts.end');
    assert.equal(p[0][3].value, '2017-02-16');
  }

  verifyState5();

  h.add(
    [
      {
        key: 'ts.start',
        value: '2017-02-15',
      },
      {
        key: 'ts.end',
        value: '2017-02-16',
      },
    ],
    viz,
  );

  function verifyState6(p = h.current_params()) {
    assert.equal(p.length, 1);
    assert.equal(p[0].length, 4);
    assert.equal(p[0][0].key, 'k2');
    assert.equal(p[0][0].value, 'v2');
    assert.equal(p[0][1].key, 'k1');
    assert.equal(p[0][1].value, 'v1.2');
    assert.equal(p[0][2].key, 'ts.start');
    assert.equal(p[0][2].value, '2017-02-15');
    assert.equal(p[0][3].key, 'ts.end');
    assert.equal(p[0][3].value, '2017-02-16');
  }

  verifyState6();

  const stateFunctions = [
    verifyState2,
    verifyState3,
    verifyState4,
    verifyState5,
    verifyState6,
  ];
  let pos = stateFunctions.length - 1;

  while (pos > 0) {
    h.back();
    pos -= 1;
    stateFunctions[pos]();
  }

  while (pos < stateFunctions.length - 1) {
    h.forward();
    pos += 1;
    stateFunctions[pos]();
  }

  // Move back, verify state, and set checkpoint
  h.back();
  h.back();
  h.back();
  stateFunctions[1]();
  h.setCheckPointState();

  // Move forward and verify state
  h.forward();
  h.forward();
  h.forward();
  stateFunctions[4]();

  // Restore to checkpoint and verify state
  h.getCheckPointState();
  stateFunctions[1]();

  h.forward();
  h.forward();
  h.forward();
  stateFunctions[4]();

  // Restore to checkpoint and verify state
  h.getCheckPointState();
  stateFunctions[1]();
});

QUnit.test('vars', (assert) => {
  let p = arc.params.tracker();
  let s = 'abcd';
  let r = 'abcd';

  assert.equal(r, p.substitute(s));

  s = 'abc <<one>> def';
  r = 'abc <<one>> def';
  assert.equal(r, p.substitute(s));

  r = 'abc 1 def';
  assert.equal(r, p.substitute(s, ['<<one:1>>']));

  assert.equal('abc two def', p.substitute(s, ['<<one:two>>']));

  arc.globals.parameters_case_sensitive = false;
  p.add({
    key: 'oNe',
    value: '1',
  });
  assert.equal(r, p.substitute(s));

  arc.globals.parameters_case_sensitive = true;
  p = arc.params.tracker();
  p.add({
    key: 'oNe',
    value: '1',
  });
  assert.equal(s, p.substitute(s));

  s = 'one <<abc:def>> two <<xy:z>> three <<ad>> four < a d > ';
  r = 'one def two z three <<ad>> four < a d > ';
  assert.equal(r, p.substitute(s));

  s = 'one <<aBc:def>> two <<XY:z>> three <<ad>> four < a d > ';
  assert.equal(r, p.substitute(s));

  s = 'one < ad <<abc:def>>';
  r = 'one < ad def';
  assert.equal(r, p.substitute(s));

  s = 'one < ad <<abc:>>';
  r = 'one < ad ';
  assert.equal(r, p.substitute(s));

  s = "one < ad <<abc:a'b'c>>";
  r = "one < ad a'b'c";
  assert.equal(r, p.substitute(s));

  s = 'one < ad <<abc:a"b"c>>';
  r = 'one < ad a"b"c';
  assert.equal(r, p.substitute(s));
});

QUnit.test('code', (assert) => {
  const p = arc.params.tracker();
  let s = "one <script> return 'abc' </script> two";
  let r = 'one abc two';
  assert.equal(r, p.substitute(s));

  s =
    "<<one:1>> <script> if (arcapi.getVar('one') === '1') { return '2'; } </script> two";
  r = '1 2 two';
  assert.equal(r, p.substitute(s));

  s = "<<one:1>> <script> if (3 < 4) { return '2'; } </script> two";
  r = '1 2 two';
  assert.equal(r, p.substitute(s));
});

QUnit.test('failed execution', (assert) => {
  const p = arc.params.tracker();
  const s = "one <script> return 'abc' </script> two";
  let r = 'one abc two';

  assert.equal(r, p.substitute(s));
  arc.globals_init().enable_script_parameters(false);
  r = "one <script> return 'abc' </script> two";

  // Don't print out errors on the console - errors are expected
  const stored_logtrace = arc.misc.logtrace;
  arc.misc.logtrace = function () {};

  assert.equal(r, p.substitute(s));
  arc.misc.logtrace = stored_logtrace;
});
