QUnit.module('framework_arc_searchstring', {
  default_colnames: [
    'ip',
    'ip_src',
    'ip_dst',
    'bytes',
    'hostname',
    'traffictype',
    'status',
    'description',
  ],
  default_coltypes: [
    'string',
    'string',
    'string',
    'integer',
    'string',
    'string',
    'string',
    'string',
  ],
  default_search_columns: ['ip', 'hostname', 'traffictype'],

  convertToWhere(predicateStr) {
    const predicate = arc.grammar._searchstring.parse(predicateStr);

    return arc.grammar._searchstring
      .searchPredicate()
      .colnames(this.default_colnames)
      .coltypes(this.default_coltypes)
      .addSearchCollection('host', ['ip', 'hostname'])
      .defaultSearchColnames(this.default_search_columns)
      .convertPredicatesToWhereClause(predicate);
  },
});

QUnit.test('simple', (assert) => {
  assert.deepEqual(arc.grammar._searchstring.parse('a b c'), {
    type: 'AND',
    left: {
      value: 'a',
    },
    right: {
      left: {
        value: 'b',
      },
      right: {
        value: 'c',
      },
      type: 'AND',
    },
  });
});

QUnit.test('simple and', (assert) => {
  const e = arc.grammar._searchstring.parse('a AND b');

  assert.deepEqual(e, {
    type: 'AND',
    left: {
      value: 'a',
    },
    right: {
      value: 'b',
    },
  });
});

QUnit.test('where clauses', function (assert) {
  let i;

  i = 'a AND b';
  assert.equal(
    this.convertToWhere(i),
    "(ip like '%a%' OR hostname like '%a%' OR traffictype like '%a%') AND (ip like '%b%' OR hostname like '%b%' OR traffictype like '%b%')",
  );

  i = 'ip=1.2.3.4 and description=error';
  assert.equal(
    this.convertToWhere(i),
    "(ip like '%1.2.3.4%') AND (description like '%error%')",
  );

  i = 'ip==1.2.3.4 and description=error';
  assert.equal(
    this.convertToWhere(i),
    "(ip = '1.2.3.4') AND (description like '%error%')",
  );

  i = "ip=1.2.3.4 and description=error or host='yahoo.com'";
  assert.equal(
    this.convertToWhere(i),
    "((ip like '%1.2.3.4%') AND (description like '%error%')) OR (ip like '%yahoo.com%' OR hostname like '%yahoo.com%')",
  );

  i = "ip*=1.2.3.4 and description=error or host='yahoo.com'";
  assert.equal(
    this.convertToWhere(i),
    "((ip like '%1.2.3.4%' OR ip_src like '%1.2.3.4%' OR ip_dst like '%1.2.3.4%') AND (description like '%error%')) OR (ip like '%yahoo.com%' OR hostname like '%yahoo.com%')",
  );
});

QUnit.test('simple or', (assert) => {
  const i = 'a OR b';
  const e = arc.grammar._searchstring.parse(i);

  assert.deepEqual(e, {
    type: 'OR',
    left: {
      value: 'a',
    },
    right: {
      value: 'b',
    },
  });
});

QUnit.test('simple not', (assert) => {
  const i = 'not b';
  const e = arc.grammar._searchstring.parse(i);

  assert.deepEqual(e, {
    type: 'NOT',
    operand: {
      value: 'b',
    },
  });
});

QUnit.test('simple precedence', (assert) => {
  assert.deepEqual(arc.grammar._searchstring.parse('a AND b OR c'), {
    type: 'OR',
    right: {
      value: 'c',
    },
    left: {
      type: 'AND',
      left: {
        value: 'a',
      },
      right: {
        value: 'b',
      },
    },
  });
});

QUnit.test('simple precedence', (assert) => {
  assert.deepEqual(arc.grammar._searchstring.parse('a AND b OR c AND not d'), {
    type: 'OR',
    left: {
      type: 'AND',
      left: {
        value: 'a',
      },
      right: {
        value: 'b',
      },
    },
    right: {
      type: 'AND',
      left: {
        value: 'c',
      },
      right: {
        type: 'NOT',
        operand: {
          value: 'd',
        },
      },
    },
  });
});

QUnit.test('long strings and', (assert) => {
  assert.deepEqual(arc.grammar._searchstring.parse('abc AND de'), {
    type: 'AND',
    left: {
      value: 'abc',
    },
    right: {
      value: 'de',
    },
  });

  assert.deepEqual(arc.grammar._searchstring.parse("'abc' AND de"), {
    type: 'AND',
    left: {
      value: 'abc',
    },
    right: {
      value: 'de',
    },
  });

  assert.deepEqual(arc.grammar._searchstring.parse("'ab c' AND de"), {
    type: 'AND',
    left: {
      value: 'ab c',
    },
    right: {
      value: 'de',
    },
  });

  assert.deepEqual(
    arc.grammar._searchstring.parse('\'ab " c\' AND "d  \' e"'),
    {
      type: 'AND',
      left: {
        value: 'ab " c',
      },
      right: {
        value: "d  ' e",
      },
    },
  );
});

QUnit.test('expr and', (assert) => {
  assert.deepEqual(arc.grammar._searchstring.parse('ip=1.2.3.4 AND de'), {
    type: 'AND',
    left: {
      type: 'APPROX',
      key: 'ip',
      value: '1.2.3.4',
    },
    right: {
      value: 'de',
    },
  });
});

QUnit.test('complex', function (assert) {
  const searchString =
    'a AND b AND colgroup=1.2.3.4 OR (c AND d and colgroup2=error) AND (NOT col2=error)';
  assert.deepEqual(arc.grammar._searchstring.parse(searchString), {
    left: {
      left: {
        value: 'a',
      },
      right: {
        left: {
          value: 'b',
        },
        right: {
          type: 'APPROX',
          key: 'colgroup',
          value: '1.2.3.4',
        },
        type: 'AND',
      },
      type: 'AND',
    },
    right: {
      left: {
        left: {
          value: 'c',
        },
        right: {
          left: {
            value: 'd',
          },
          right: {
            type: 'APPROX',
            key: 'colgroup2',
            value: 'error',
          },
          type: 'AND',
        },
        type: 'AND',
      },
      right: {
        operand: {
          type: 'APPROX',
          key: 'col2',
          value: 'error',
        },
        type: 'NOT',
      },
      type: 'AND',
    },
    type: 'OR',
  });

  this.convertToWhere(searchString);
});
