QUnit.module('visuals_arc_css', {
  beforeEach() {
    arc.testutil = arc.testutil || {};
    arc.testutil.ignoreFontLoader = true;
    arc.globals.apply_custom_styles = true;
    arc.globals.enable_custom_styles = true;
    this.table = arc.visual.plugins.table();
    this.driver = arc.report.driver(this.table);

    this.driver.chart_div('chart-div')
               .report_id(3);
    arc.frontend.globals.urls.customCssRawUrl = '/arc/css/';
  },

  afterEach() {
    $('#chart-div').empty();
    arc.globals.apply_custom_styles = false;
    arc.globals.enable_custom_styles = false;
    arc.testutil = arc.testutil || {};
    arc.testutil.ignoreFontLoader = false;
    arc.testutil.ignoreJsCrossOrigin = false;
  },
});

function drawTable(assert, table, driver, tableChecksDisabled) {
  const result = arc.data.result();
  result.rows([
    ['India', '200'],
    ['China', '400'],
  ]);
  const col1 = arc.data
    .column()
    .colname('country')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(1);
  result.columns([col1, col2]);
  result.finalize();

  // create the visual by setting the data and then drawing it
  driver.dataResult(result);
  driver.bindScripts(result);
  driver.drawVisual();

  if (!tableChecksDisabled) {
    const tables = $('#chart-div').find('table');
    assert.equal(tables.length, 3);

    const tr = tables.find('tbody tr');
    assert.equal(tr.length, 2);
  }
}

QUnit.test('Apply CSS Class', function (assert) {
  const settings = this.driver.user_settings();
  const container_css_class = settings.find_by_id('container_css_class');
  container_css_class.value('abc');

  assert.ok(!$('#main').hasClass('abc'));
  assert.ok(!$('#main').hasClass('xyz'));
  assert.ok(!$('#main').hasClass('ghi'));

  drawTable(assert, this.table, this.driver);
  assert.ok($('#main').hasClass('abc'));
  assert.ok(!$('#main').hasClass('xyz'));
  assert.ok(!$('#main').hasClass('ghi'));

  container_css_class.value('xyz');
  drawTable(assert, this.table, this.driver);
  assert.ok(!$('#main').hasClass('abc'));
  assert.ok($('#main').hasClass('xyz'));
  assert.ok(!$('#main').hasClass('ghi'));

  container_css_class.value('ghi');
  drawTable(assert, this.table, this.driver);
  assert.ok(!$('#main').hasClass('abc'));
  assert.ok(!$('#main').hasClass('xyz'));
  assert.ok($('#main').hasClass('ghi'));

  container_css_class.value('abc ghi');
  drawTable(assert, this.table, this.driver);
  assert.ok($('#main').hasClass('abc'));
  assert.ok(!$('#main').hasClass('xyz'));
  assert.ok($('#main').hasClass('ghi'));

  container_css_class.value('');
  drawTable(assert, this.table, this.driver);
  assert.ok(!$('#main').hasClass('abc'));
  assert.ok(!$('#main').hasClass('xyz'));
  assert.ok(!$('#main').hasClass('ghi'));
});

QUnit.test('Inline CSS', function (assert) {
  let css = '#main {background-color: aliceblue;}';
  const settings = this.driver.user_settings();
  settings.find_by_id('inline_css').value(css);

  drawTable(assert, this.table, this.driver);

  arc.css.CUSTOMCSS_MANAGER.applyCustomCss([this.driver]);
  let read_css = $('#arc-customcss1').html();
  assert.equal(read_css, css);
  read_css = $('#arc-customcss2').html();
  assert.equal(read_css, '');

  css = '#main {background-color: grey;}';
  settings.find_by_id('inline_css').value(css);
  arc.css.CUSTOMCSS_MANAGER.applyCustomCss([this.driver]);
  read_css = $('#arc-customcss2').html();
  assert.equal(read_css, css);
  read_css = $('#arc-customcss1').html();
  assert.equal(read_css, '');
});

QUnit.test('External CSS Stylesheets', function (assert) {
  // We cannot do cross origin fetch of a local file fetch with file:// protocol.
  // This line prevents that error.
  arc.testutil.ignoreJsCrossOrigin = true;

  const settings = this.driver.user_settings();
  settings.find_by_id('css_styles').value(JSON.stringify([1, 2]));
  arc.frontend.globals.urls.customCssRawUrl = '/cssurl';

  drawTable(assert, this.table, this.driver);

  arc.css.CUSTOMCSS_MANAGER.applyCustomCss([this.driver]);

  let $css1 = $('head link[href="/cssurl/1"]');
  assert.equal($css1.length, 1);
  assert.equal($css1.attr('href'), `${arc.frontend.globals.urls.customCssRawUrl}/1`);

  let $css2 = $('head link[href="/cssurl/2"]');
  assert.equal($css2.length, 1);
  assert.equal($css2.attr('href'), `${arc.frontend.globals.urls.customCssRawUrl}/2`);

  settings.find_by_id('css_styles').value(JSON.stringify([3, 4]));
  arc.css.CUSTOMCSS_MANAGER.applyCustomCss([this.driver]);

  $css1 = $('#arc-customcss-1');
  assert.equal($css1.length, 0);
  $css2 = $('#arc-customcss-2');
  assert.equal($css2.length, 0);

  const $css3 = $('head link[href="/cssurl/3"]');
  assert.equal($css3.length, 1);
  assert.equal($css3.attr('href'), `${arc.frontend.globals.urls.customCssRawUrl}/3`);

  const $css4 = $('head link[href="/cssurl/4"]');
  assert.equal($css4.length, 1);
  assert.equal($css4.attr('href'), `${arc.frontend.globals.urls.customCssRawUrl}/4`);
});

QUnit.test('External JS beforeDraw', function (assert) {
  const v = $('#chart-div');

  window.tracker = {
    beforeDraw: 0,
    afterDraw: 0,
  };

  // Don't specify 'version', should not impact draw
  const jsFn = function () {
    return {
      version: '1',
      beforeDraw() {
        window.tracker.beforeDraw += 1;
      },
      afterDraw() {
        window.tracker.afterDraw = window.tracker.beforeDraw + 1;
      },
    };
  };

  const jsSource = `return ${jsFn.toString()}()`;
  this.driver.setInlineJsAssets([jsSource]);

  assert.equal(window.tracker.beforeDraw, 0);
  assert.equal(window.tracker.afterDraw, 0);
  drawTable(assert, this.table, this.driver);
  assert.equal(window.tracker.beforeDraw, 1);
  assert.equal(window.tracker.afterDraw, 2);

  const t = v.find('.dataTables_scrollBody');
  assert.equal(t.length, 1, 'Table should be drawn');
});

QUnit.test('External JS disableDraw', function (assert) {
  let jsFn;
  let jsSource;
  const v = $('#chart-div');
  let t;

  // Don't specify 'version', should not impact draw
  jsFn = function () {
    return {
      disableDraw() {
        return true;
      },
    };
  };

  jsSource = `return ${jsFn.toString()}()`;
  arc.testutil = arc.testutil || {};
  arc.testutil.ignoreAlerts = true;
  this.driver.setInlineJsAssets([jsSource]);
  drawTable(assert, this.table, this.driver);
  arc.testutil.ignoreAlerts = false;
  t = v.find('.dataTables_scrollBody');
  assert.equal(t.length, 1, 'Table should be drawn');

  // Specify 'version', 'disableDraw' should disable the table
  $('#chart-div').empty();
  jsFn = function () {
    return {
      version: '1',
      disableDraw() {
        return true;
      },
    };
  };

  jsSource = `return ${jsFn.toString()}()`;
  this.driver.setInlineJsAssets([jsSource]);
  drawTable(assert, this.table, this.driver, true);
  t = v.find('.dataTables_scrollBody');
  assert.equal(t.length, 0, 'No table should be drawn');

  // Specity disableDraw() === false
  $('#chart-div').empty();
  jsFn = function () {
    return {
      version: '1',
      disableDraw() {
        return false;
      },
    };
  };

  jsSource = `return ${jsFn.toString()}()`;
  this.driver.setInlineJsAssets([jsSource]);
  drawTable(assert, this.table, this.driver);
  t = v.find('.dataTables_scrollBody');
  assert.equal(t.length, 1, 'Table should be drawn');

  // Finally, don't specify a disableDraw
  $('#chart-div').empty();
  jsFn = function () {
    return {
      version: '1',
    };
  };

  jsSource = `return ${jsFn.toString()}()`;
  this.driver.setInlineJsAssets([jsSource]);
  drawTable(assert, this.table, this.driver);
  t = v.find('.dataTables_scrollBody');
  assert.equal(t.length, 1, 'Table should be drawn');
});
