QUnit.module('visuals_arc_displayformat', {
  beforeEach() {
    this.table = arc.visual.plugins.table();
    this.driver = arc.report.driver(this.table);
    this.driver.chart_div('chart-div');
  },
  afterEach() {
    $('#chart-div').empty();
  },
});

QUnit.test('Simple Display Format', function (assert) {
  // provide the data
  const result = arc.data.result();
  result.rows([
    ['India', '20000'],
    ['China', '40000'],
  ]);
  const col1 = arc.data
    .column()
    .colname('country')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('dimensions_shelf')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .setDisplayFormat(this.driver.translatedDisplayFormat('.3s'))
    .colindex(1);
  result.columns([col1, col2]);
  result.finalize();

  // create the visual by setting the data and then drawing it
  this.table.viz_setDataResult(result);
  this.table.viz_drawVisual('chart-div');

  const tables = $('#chart-div').find('table');
  assert.equal(tables.length, 3);

  const tr = tables.find('tbody tr');
  assert.equal(tr.length, 2);

  const td = tr.find('td');
  assert.equal(td.length, 4);
  td.each(function (i) {
    const d = $(this).text();
    if (i === 0) {
      assert.equal(d, 'India');
    } else if (i === 1) {
      assert.equal(d, '20.0k');
    } else if (i === 2) {
      assert.equal(d, 'China');
    } else {
      assert.equal(i, 3);
      assert.equal(d, '40.0k');
    }
  });
});

// TODO: Failed date formatter should get caught by tryAndCatch, but currently fails
//      and triggers date_column._defaultFormat() due to _fallbackToDefault
QUnit.test('Console log formatting errors properly', function (assert) {
  // provide the data
  const result = arc.data.result();
  result.rows([
    ['India', 'Bangalore', '20000', '2000'],
    ['China', 'Chengdu', '40000', '1000'],
  ]);
  const oldStyleEmptyFormatter = this.driver.translatedDisplayFormat('');
  let modernEmptyFormatter = this.driver.translatedDisplayFormat({
    type: '',
    result: '',
  });
  assert.equal(
    modernEmptyFormatter,
    undefined,
    'Old empty display formatter returns back undefined',
  );
  modernEmptyFormatter = this.driver.translatedDisplayFormat({});
  assert.equal(
    modernEmptyFormatter,
    undefined,
    'Modern empty display formatter returns back undefined',
  );
  const modernDisplayFormat = this.driver.translatedDisplayFormat({
    type: 'number',
    result: '.2f',
    comma: false,
    decimal: '2',
    decimalall: false,
  });
  const col1 = arc.data
    .column()
    .colname('country')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('dimensions_shelf')
    .setDisplayFormat(oldStyleEmptyFormatter)
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('city')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('dimensions_shelf')
    .setDisplayFormat(modernDisplayFormat)
    .colindex(1);
  const col3 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .setDisplayFormat(modernEmptyFormatter)
    .colindex(2);
  const col4 = arc.data
    .column()
    .colname('towers')
    .coltype('INT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .setDisplayFormat(modernDisplayFormat)
    .colindex(3);

  result.columns([col1, col2, col3, col4]);
  result.finalize();

  // create the visual by setting the data and then drawing it
  this.table.viz_setDataResult(result);

  // catch console logs
  const logs = [];
  const expectedLogs = [
    'undefined display format making Bangalore NaN',
    'undefined display format making Chengdu NaN',
  ];
  // eslint-disable-next-line no-console
  console.log = function (message) {
    logs.push(message);
  };

  this.table.viz_drawVisual('chart-div');
  assert.deepEqual(logs, expectedLogs, 'Get expected console logs only');

  const tables = $('#chart-div').find('table');
  assert.equal(tables.length, 3);

  const tr = tables.find('tbody tr');
  assert.equal(tr.length, 2);

  const td = tr.find('td');
  assert.equal(td.length, 8);

  const expectedCellValue = [
    'India',
    'Bangalore',
    '20,000',
    '2000.00',
    'China',
    'Chengdu',
    '40,000',
    '1000.00',
  ];
  const actualCellValues = [];
  td.each((i, el) => {
    actualCellValues.push(el.textContent);
  });
  assert.deepEqual(
    expectedCellValue,
    actualCellValues,
    'table has correct values',
  );
});

QUnit.test('Parametrized Display Format', function (assert) {
  // provide the data
  const result = arc.data.result();
  result.rows([
    ['India', '20000'],
    ['China', '40000'],
  ]);

  const col1 = arc.data
    .column()
    .colname('country')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('dimensions_shelf')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .setDisplayFormat(this.driver.translatedDisplayFormat('<<display:.3s>>'))
    .colindex(1);
  result.columns([col1, col2]);
  result.finalize();

  // create the visual by setting the data and then drawing it
  this.table.viz_setDataResult(result);
  this.table.viz_drawVisual('chart-div');

  let tables = $('#chart-div').find('table');
  assert.equal(tables.length, 3);

  let tr = tables.find('tbody tr');
  assert.equal(tr.length, 2);

  let td = tr.find('td');
  assert.equal(td.length, 4);
  td.each(function (i) {
    const d = $(this).text();
    if (i === 0) {
      assert.equal(d, 'India');
    } else if (i === 1) {
      assert.equal(d, '20.0k');
    } else if (i === 2) {
      assert.equal(d, 'China');
    } else {
      assert.equal(i, 3);
      assert.equal(d, '40.0k');
    }
  });

  const params = arc.params.tracker().add({
    key: 'display',
    value: '.5s',
  });
  this.driver.params(params);
  col2.setDisplayFormat(this.driver.translatedDisplayFormat('<<display:.3s>>'));
  result.columns([col1, col2]);
  result.finalize();

  // create the visual by setting the data and then drawing it
  this.table.viz_setDataResult(result);
  this.table.viz_drawVisual('chart-div');

  tables = $('#chart-div').find('table');
  assert.equal(tables.length, 3);

  tr = tables.find('tbody tr');
  assert.equal(tr.length, 2);

  td = tr.find('td');
  assert.equal(td.length, 4);
  td.each(function (i) {
    const d = $(this).text();
    if (i === 0) {
      assert.equal(d, 'India');
    } else if (i === 1) {
      assert.equal(d, '20.000k');
    } else if (i === 2) {
      assert.equal(d, 'China');
    } else {
      assert.equal(i, 3);
      assert.equal(d, '40.000k');
    }
  });
});
