QUnit.module('visuals_arc_filter_params', {
  beforeEach() {
    // main chart object
    const bar = arc.visual.plugins.bar();

    // create the lib object
    const driver = arc.report.driver(bar);
    driver.chart_div('chart-div');
    driver.enable_auto_page_save_handling(false);

    bar.viz_drawDesigner('app-shelves');

    this.bar = bar;
    this.driver = driver;
  },

  afterEach() {
    $('#app-shelves').empty();
  },
});

QUnit.test('verify string escaping', (assert) => {
  // 'STRING' and 'TIMESTAMP' fields get escaped, others do not
  assert.equal(arc.report.shelf.parse_filter_value('CA', 'STRING'), "'CA'");
  assert.equal(
    arc.report.shelf.parse_filter_value("CA'CA", 'STRING'),
    "'CA\\'CA'",
  );
  assert.equal(
    arc.report.shelf.parse_filter_value("CA'CA'CA", 'STRING'),
    "'CA\\'CA\\'CA'",
  );
  assert.equal(
    arc.report.shelf.parse_filter_value("CA'CA'CA", 'string'),
    "'CA\\'CA\\'CA'",
  );
  assert.equal(
    arc.report.shelf.parse_filter_value("CA'CA'CA", 'TIMESTAMP'),
    "'CA\\'CA\\'CA'",
  );
  assert.equal(
    arc.report.shelf.parse_filter_value("CA'CA'CA", 'timestamp'),
    "'CA\\'CA\\'CA'",
  );

  assert.equal(
    arc.report.shelf.parse_filter_value("CA'CA'CA", 'INTEGER'),
    "CA'CA'CA",
  );
});

QUnit.test('Add parameters to shelves', function (assert) {
  let params = [];
  params.push({
    key: 'state_id',
    value: 12,
    keytype: arc.report.columnTypeEnum.DIMENSION,
    valuetype: 'integer',
  });

  let local_filters = this.driver.local_filters();
  assert.equal(local_filters.length, 0, 'There are no local filters');
  this.driver.addToFilters(params, false);
  local_filters = this.driver.local_filters();
  assert.equal(local_filters.length, 1, 'There is one local filter');
  assert.equal(local_filters[0], 12, '12 is filtered');

  const s1 = 'single\'nowdouble"end';
  const s1_filtered = "'single\\'nowdouble\"end'";
  params = [];
  params.push({
    key: 'state',
    value: s1,
    keytype: arc.report.columnTypeEnum.DIMENSION,
    valuetype: 'string',
  });

  this.driver.addToFilters(params, false);
  local_filters = this.driver.local_filters();
  assert.equal(local_filters.length, 2, 'There are two local filters');
  assert.equal(local_filters[0], 12, '12 is filtered');
  assert.equal(local_filters[1], s1, `verifying escaped string${s1}`);

  const s2 = 'CA';
  const s2_filtered = "'CA'";
  params = [];
  params.push({
    key: 'state',
    value: s2,
    keytype: arc.report.columnTypeEnum.DIMENSION,
    valuetype: 'string',
  });

  this.driver.addToFilters(params, false);
  local_filters = this.driver.local_filters();
  assert.equal(local_filters.length, 3, 'There are three local filters');
  assert.equal(local_filters[0], 12, '12 is filtered');
  assert.equal(local_filters[1], s1, `verifying escaped string${s1}`);
  assert.equal(local_filters[2], s2, `verifying escaped string${s2}`);

  const s3 = "'single' end'";
  const s3_filtered = "'\\'single\\' end\\''";
  params = [];
  params.push({
    key: 'state',
    value: s3,
    keytype: arc.report.columnTypeEnum.DIMENSION,
    valuetype: 'string',
  });

  this.driver.addToFilters(params, false);
  local_filters = this.driver.local_filters();
  assert.equal(local_filters.length, 4, 'There are four local filters');
  assert.equal(local_filters[0], 12, '12 is filtered');
  assert.equal(local_filters[1], s1, `verifying escaped string${s1}`);
  assert.equal(local_filters[2], s2, `verifying escaped string${s2}`);
  assert.equal(local_filters[3], s3, `verifying escaped string${s3}`);

  const stored_elements = [];
  this.driver.shelfElementCb('filters_shelf', (shelf_element) => {
    stored_elements.push(shelf_element);
  });

  assert.equal(stored_elements.length, 2, 'There are two stored elements');
  assert.equal(stored_elements[0].filter_column(), 'state_id');
  assert.equal(stored_elements[0].filter_exclude(), true);
  assert.equal(stored_elements[1].filter_column(), 'state');
  assert.equal(stored_elements[1].filter_exclude(), true);

  let d = stored_elements[0].filter_data();
  assert.equal(d.length, 1, 'There is one element in state_id');

  d = stored_elements[1].filter_data();
  assert.equal(d.length, 3, 'There are three elements in state');
  assert.equal(d[0], s1_filtered, 'verifying first filter');
  assert.equal(d[1], s2_filtered, 'verifying second filter');
  assert.equal(d[2], s3_filtered, 'verifying third filter');

  // Now that we have both text and non-text filter shelf elements, use them to
  // validate with and without NULL in both include and exclude modes.
  const f0 = stored_elements[0];
  const f1 = stored_elements[1];
  const f0text = '(12)';
  const f1text = `(${s1_filtered}, ${s2_filtered}, ${s3_filtered})`;
  assert.equal(f0.getExpression(), `[state_id] not in ${f0text}`);
  assert.equal(f1.getExpression(), `[state] not in ${f1text}`);

  f0.filter_exclude(false);
  f1.filter_exclude(false);
  assert.equal(f0.getExpression(), `[state_id] in ${f0text}`);
  assert.equal(f1.getExpression(), `[state] in ${f1text}`);

  f0.filter_data().push('NULL');
  f1.filter_data().push('NULL');
  assert.equal(
    f0.getExpression(),
    `([state_id] in ${f0text} OR [state_id] is NULL)`,
  );
  assert.equal(f1.getExpression(), `([state] in ${f1text} OR [state] is NULL)`);

  f0.filter_exclude(true);
  f1.filter_exclude(true);
  assert.equal(
    f0.getExpression(),
    `([state_id] not in ${f0text} AND [state_id] is not NULL)`,
  );
  assert.equal(
    f1.getExpression(),
    `([state] not in ${f1text} AND [state] is not NULL)`,
  );

  // Also validate NULL only case
  f0.filter_data(['NULL']);
  f1.filter_data(['NULL']);
  assert.equal(f0.getExpression(), '[state_id] is not NULL');
  assert.equal(f1.getExpression(), '[state] is not NULL');

  f0.filter_exclude(false);
  f1.filter_exclude(false);
  assert.equal(f0.getExpression(), '[state_id] is NULL');
  assert.equal(f1.getExpression(), '[state] is NULL');
});

QUnit.module('visuals_arc_params.format_list: postgres', {
  beforeEach() {
    arc.report.shelf.CONNECTION_TYPE = {
      connection_type: 'postgres',
    };
  },
});

QUnit.test('postgres: formatting filter expressions', (assert) => {
  let original = [
    "'quoted'",
    "It's a Wonderful Life",
    '&ymb#ls g@lore!',
    'nothing',
  ];
  let expected = [
    "'''quoted'''",
    "'It''s a Wonderful Life'",
    "'&ymb#ls g@lore!'",
    "'nothing'",
  ];
  let actual = arc.params.format_list(original, 'string');
  assert.deepEqual(actual, expected, 'Formatting string correctly');

  actual = arc.params.format_list(original);
  assert.deepEqual(actual, expected, 'Formatting unknown string correctly');

  original = ['2018-11-11 00:00:00', '2018-11-11 00:01:00'];
  expected = ["'2018-11-11 00:00:00'", "'2018-11-11 00:01:00'"];
  actual = arc.params.format_list(original, 'timestamp');
  assert.deepEqual(actual, expected, 'Formatting datetime correctly');

  actual = arc.params.format_list(original);
  assert.deepEqual(actual, expected, 'Formatting unknown datetime correctly');

  original = [100, 200, 300];
  actual = arc.params.format_list(original, 'bigint');
  assert.deepEqual(actual, original, 'Formatting other value correctly');

  original = [100, 200, 300];
  actual = arc.params.format_list(original);
  assert.deepEqual(actual, original, 'Formatting unknown number correctly');
});

QUnit.test('postgres: unformatting filter expressions', (assert) => {
  let original = [
    "'quoted'",
    "It's a Wonderful Life",
    '&ymb#ls g@lore!',
    'nothing',
  ];
  let expected = [
    "'''quoted'''",
    "'It''s a Wonderful Life'",
    "'&ymb#ls g@lore!'",
    "'nothing'",
  ];
  let actual = arc.params.format_list(expected, 'string', true);
  assert.deepEqual(actual, original, 'Unformatting string correctly');

  original = ['2018-11-11 00:00:00', '2018-11-11 00:01:00'];
  expected = ["'2018-11-11 00:00:00'", "'2018-11-11 00:01:00'"];
  actual = arc.params.format_list(expected, 'timestamp', true);
  assert.deepEqual(actual, original, 'Unformatting datetime correctly');

  original = [100, 200, 300];
  actual = arc.params.format_list(original, 'bigint', true);
  assert.deepEqual(actual, original, 'Unformatting other value correctly');
});

QUnit.module('visuals_arc_params.format_list: sqlite', {
  beforeEach() {
    arc.report.shelf.CONNECTION_TYPE = {
      connection_type: 'sqlite',
    };
  },
});

QUnit.test('sqlite: formatting filter expressions', (assert) => {
  let original = [
    "'quoted'",
    "It's a Wonderful Life",
    '&ymb#ls g@lore!',
    'nothing',
  ];
  let expected = [
    "'''quoted'''",
    "'It''s a Wonderful Life'",
    "'&ymb#ls g@lore!'",
    "'nothing'",
  ];
  let actual = arc.params.format_list(original, 'string');
  assert.deepEqual(actual, expected, 'Formatting string correctly');

  actual = arc.params.format_list(original);
  assert.deepEqual(actual, expected, 'Formatting unknown string correctly');

  original = ['2018-11-11 00:00:00', '2018-11-11 00:01:00'];
  expected = ["'2018-11-11 00:00:00'", "'2018-11-11 00:01:00'"];
  actual = arc.params.format_list(original, 'timestamp');
  assert.deepEqual(actual, expected, 'Formatting datetime correctly');

  actual = arc.params.format_list(original);
  assert.deepEqual(actual, expected, 'Formatting unknown datetime correctly');

  original = [100, 200, 300];
  actual = arc.params.format_list(original, 'bigint');
  assert.deepEqual(actual, original, 'Formatting other value correctly');

  original = [100, 200, 300];
  actual = arc.params.format_list(original);
  assert.deepEqual(actual, original, 'Formatting unknown number correctly');
});

QUnit.test('sqlite: unformatting filter expressions', (assert) => {
  let original = [
    "'quoted'",
    "It's a Wonderful Life",
    '&ymb#ls g@lore!',
    'nothing',
  ];
  let expected = [
    "'''quoted'''",
    "'It''s a Wonderful Life'",
    "'&ymb#ls g@lore!'",
    "'nothing'",
  ];
  let actual = arc.params.format_list(expected, 'string', true);
  assert.deepEqual(actual, original, 'Unformatting string correctly');

  original = ['2018-11-11 00:00:00', '2018-11-11 00:01:00'];
  expected = ["'2018-11-11 00:00:00'", "'2018-11-11 00:01:00'"];
  actual = arc.params.format_list(expected, 'timestamp', true);
  assert.deepEqual(actual, original, 'Unformatting datetime correctly');

  original = [100, 200, 300];
  actual = arc.params.format_list(original, 'bigint', true);
  assert.deepEqual(actual, original, 'Unformatting other value correctly');
});

QUnit.module('visuals_arc_params.format_list: oracle/other', {
  beforeEach() {
    arc.report.shelf.CONNECTION_TYPE = {
      connection_type: 'oracle/other',
    };
  },
});

QUnit.test('oracle/other: formatting filter expressions', (assert) => {
  let original = [
    "'quoted'",
    "It's a Wonderful Life",
    '&ymb#ls g@lore!',
    'nothing',
  ];
  let expected = [
    "'\\'quoted\\''",
    "'It\\'s a Wonderful Life'",
    "'&ymb#ls g@lore!'",
    "'nothing'",
  ];
  let actual = arc.params.format_list(original, 'string');
  assert.deepEqual(actual, expected, 'Formatting string correctly');

  actual = arc.params.format_list(original);
  assert.deepEqual(actual, expected, 'Formatting unknown string correctly');

  original = ['2018-11-11 00:00:00', '2018-11-11 00:01:00'];
  expected = ["'2018-11-11 00:00:00'", "'2018-11-11 00:01:00'"];
  actual = arc.params.format_list(original, 'timestamp');
  assert.deepEqual(actual, expected, 'Formatting datetime correctly');

  actual = arc.params.format_list(original);
  assert.deepEqual(actual, expected, 'Formatting unknown datetime correctly');

  original = [100, 200, 300];
  actual = arc.params.format_list(original, 'bigint');
  assert.deepEqual(actual, original, 'Formatting other value correctly');

  original = [100, 200, 300];
  actual = arc.params.format_list(original);
  assert.deepEqual(actual, original, 'Formatting unknown number correctly');
});

QUnit.test('oracle/other: unformatting filter expressions', (assert) => {
  let original = [
    "'quoted'",
    "It's a Wonderful Life",
    '&ymb#ls g@lore!',
    'nothing',
  ];
  let expected = [
    "'\\'quoted\\''",
    "'It\\'s a Wonderful Life'",
    "'&ymb#ls g@lore!'",
    "'nothing'",
  ];
  let actual = arc.params.format_list(expected, 'string', true);
  assert.deepEqual(actual, original, 'Unformatting string correctly');

  original = ['2018-11-11 00:00:00', '2018-11-11 00:01:00'];
  expected = ["'2018-11-11 00:00:00'", "'2018-11-11 00:01:00'"];
  actual = arc.params.format_list(expected, 'timestamp', true);
  assert.deepEqual(actual, original, 'Unformatting datetime correctly');

  original = [100, 200, 300];
  actual = arc.params.format_list(original, 'bigint', true);
  assert.deepEqual(actual, original, 'Unformatting other value correctly');
});
