QUnit.module('visuals_arc_map', {
  beforeEach() {
    // main chart object
    this.map = arc.visual.plugins.map();

    // create the lib object
    this.driver = arc.report.driver(this.map).chart_div('chart-div');
  },

  afterEach() {
    $('#chart-div').empty();
  },
});

const choroplethMapSettings = {
  settings: {
    colors: [
      '#ffffe5',
      '#f7fcb9',
      '#d9f0a3',
      '#addd8e',
      '#78c679',
      '#41ab5d',
      '#238443',
      '#006837',
      '#004529',
    ],
    graded_colors: [
      '#ffffe5',
      '#f7fcb9',
      '#d9f0a3',
      '#addd8e',
      '#78c679',
      '#41ab5d',
      '#238443',
      '#006837',
      '#004529',
    ],
  },
  user_settings: {
    color_palette:
      '{"colors":["#ffffe5","#f7fcb9","#d9f0a3","#addd8e","#78c679",' +
      '"#41ab5d","#238443","#006837","#004529"],"colorSettings":{},"isReversed":false}',
    legend_style: 'Right',
  },
};

function verifyNoBubbles(assert, $svg) {
  const circles = $svg.find('circle');
  assert.equal(circles.length, 0);
}

function verifyBubbles(assert, $svg) {
  const circles = $svg.find('circle');
  assert.equal(circles.length, 2);
  circles.each((i, e) => {
    const r = $(e).attr('r');
    if (i === 0) {
      assert.ok(arc.misc.within_tolerance(50, r, 2));
    } else if (i === 1) {
      assert.ok(arc.misc.within_tolerance(10, r, 2));
    }
  });
}

function testLegend(assert, $svg) {
  // This function assumes that the values 200 and 400 are used for the rows in
  // the measures. If the rows change, this function needs to be updated.

  const $legendSvg = $($svg[1]);

  assert.equal($legendSvg.length, 1, 'Legend is missing.');
  assert.equal(
    $legendSvg.is(':visible'),
    true,
    'Svg legend wrapper is not visible',
  );
  assert.equal(
    $legendSvg.find('g.viz-base-legend-g').css('display') !== 'none',
    true,
    'Base legend g object is not visible',
  );

  const labels = [400, 300, 200];
  const legendLabels = $legendSvg.find('text');

  legendLabels.each((i, d) => {
    assert.equal(
      $(d).text(),
      labels[i],
      'Legend vals not displayed or are incorrect.',
    );
  });
}

function testMapZoom(assert, $svg) {
  const $zoomSvg = $svg.find('.svg-pan-zoom_viewport');

  assert.equal($zoomSvg.length, 1, 'Missing Svg required for zoom');

  const startCoords = $zoomSvg.attr('transform');
  const event1 = new Event('mousedown', { bubbles: false, cancelable: false });
  const event2 = new Event('mouseup', { bubbles: false, cancelable: false });

  // Use this hack to set the number of times the mouse was clicked in a certain
  // area to two, which simulates a double click. The zoom-pan library does not
  // respond to dblclick events.
  event1.detail = 2;
  $svg[0].dispatchEvent(event1);
  $svg[0].dispatchEvent(event2);
  const restartQUnit = assert.async();

  setTimeout(() => {
    restartQUnit();
    const finalCoords = $zoomSvg.attr('transform');
    assert.equal(
      startCoords === finalCoords,
      false,
      'Double click zoom failed.',
    );
  }, 500);
}

function testMapPan(assert, $svg) {
  const $zoomSvg = $svg.find('.svg-pan-zoom_viewport');

  assert.equal($zoomSvg.length, 1, 'Missing Svg required for panning');

  const startCoords = $zoomSvg.attr('transform');
  const event1 = new Event('mousedown', { bubbles: false, cancelable: false });
  event1.clientX = 10;
  event1.clientY = 10;
  const event2 = new Event('mousemove', { bubbles: false, cancelable: false });
  event2.clientX = 20;
  event2.clientY = 20;
  const event3 = new Event('mouseup', { bubbles: false, cancelable: false });
  event3.clientX = 20;
  event3.clientY = 20;

  $svg[0].dispatchEvent(event1);
  $svg[0].dispatchEvent(event2);
  $svg[0].dispatchEvent(event3);
  const restartQUnit = assert.async();

  setTimeout(() => {
    restartQUnit();
    const finalCoords = $zoomSvg.attr('transform');
    assert.equal(startCoords === finalCoords, false, 'Panning has failed.');
  }, 500);
}

function createWorldMap(assert, that, cb) {
  const { map } = that;

  // provide the data
  const result = arc.data.result();
  result.rows([
    ['India', '200', '200'],
    ['China', '400', '400'],
  ]);
  const col1 = arc.data
    .column()
    .colname('country')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('country')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(1);
  const col3 = arc.data
    .column()
    .colname('xFactor')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('size_shelf')
    .colindex(2);
  result.columns([col1, col2, col3]);

  map.viz_setSettings(choroplethMapSettings);
  map.viz_setDataResult(result);

  const restartQUnit = assert.async();
  map.viz_fetchCollateral('chart-div', false, () => {
    map.viz_drawVisual('chart-div', false);
    restartQUnit();
    cb();
  });
}

QUnit.test('World choropleth', function (assert) {
  createWorldMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);

    testLegend(assert, svg);

    svg = svg.first();

    verifyNoBubbles(assert, svg);

    const state_paths = svg.find('path.viz-map-state');
    assert.equal(state_paths.length, 0);

    const paths = svg.find('path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 241);

    paths.each((i, e) => {
      const fill = $(e).attr('fill');
      if (i === 41) {
        assert.equal(
          fill,
          '#004529',
          'India has a color value based on the default palette',
        );
      } else if (i === 98) {
        assert.equal(
          fill,
          '#ffffe5',
          'China has a color value based on the default palette',
        );
      } else {
        assert.equal(fill, 'white');
      }
    });
  });
});

QUnit.test('World bubbles', function (assert) {
  // Draw bubbles (the default is a choropleth)
  this.map.map_bubbles.value('true');

  createWorldMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);
    svg = svg.first();

    const state_paths = svg.find('path.viz-map-state');
    assert.equal(state_paths.length, 0);

    const paths = svg.find('path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 241);

    paths.each((i, e) => {
      const fill = $(e).attr('fill');
      assert.equal(fill, 'white');
    });

    verifyBubbles(assert, svg);
  });
});

function createStateMap(assert, that, cb) {
  const { map } = that;

  // provide the data
  const result = arc.data.result();
  result.rows([
    ['CA', '200', '200'],
    ['AZ', '400', '400'],
  ]);
  const col1 = arc.data
    .column()
    .colname('state')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('state')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(1);
  const col3 = arc.data
    .column()
    .colname('xFactor')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('size_shelf')
    .colindex(2);
  result.columns([col1, col2, col3]);

  map.viz_setSettings(choroplethMapSettings);
  const user_settings = that.driver.user_settings();
  user_settings.find_by_id('zoom_enabled').value('true');
  user_settings.find_by_id('zoom_to_active').value('false');
  map.viz_setDataResult(result);

  const restartQUnit = assert.async();
  map.viz_fetchCollateral('chart-div', false, () => {
    map.viz_drawVisual('chart-div', false);
    restartQUnit();
    cb();
  });
}

QUnit.test('State choropleth', function (assert) {
  createStateMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);

    testLegend(assert, svg);

    svg = svg.first();

    verifyNoBubbles(assert, svg);

    const state_paths = svg.find('path.viz-map-state');
    assert.equal(state_paths.length, 0);

    const paths = svg.find('path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 53);
    paths.each((i, e) => {
      const fill = $(e).attr('fill');
      if (i === 5) {
        assert.equal(
          fill,
          '#004529',
          'CA has a color value based on the default palette',
        );
      } else if (i === 6) {
        assert.equal(
          fill,
          '#ffffe5',
          'AZ has a color value based on the default palette',
        );
      } else {
        assert.equal(fill, 'white');
      }
    });

    testMapZoom(assert, svg);
    testMapPan(assert, svg);
  });
});

QUnit.test('State bubbles', function (assert) {
  // Draw bubbles (the default is a choropleth)
  this.map.map_bubbles.value('true');
  createStateMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);
    svg = svg.first();

    const paths = svg.find('path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 53);
    paths.each((i, e) => {
      const fill = $(e).attr('fill');
      assert.equal(fill, 'white');
    });

    verifyBubbles(assert, svg);
  });
});

function createCountyMap(assert, that, cb) {
  const { map } = that;

  // provide the data
  const result = arc.data.result();
  result.rows([
    ['California', 'San Mateo County', '200', '200'],
    ['New York', 'Kings County', '400', '400'],
  ]);

  const col1 = arc.data
    .column()
    .colname('state')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('state')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('county')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('county')
    .colindex(1);
  const col3 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(2);
  const col4 = arc.data
    .column()
    .colname('xFactor')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('size_shelf')
    .colindex(3);
  result.columns([col1, col2, col3, col4]);

  map.viz_setSettings(choroplethMapSettings);
  map.viz_setDataResult(result);

  const restartQUnit = assert.async();
  map.viz_fetchCollateral('chart-div', false, () => {
    map.viz_drawVisual('chart-div', false);
    restartQUnit();
    cb();
  });
}

QUnit.test('County choropleth', function (assert) {
  createCountyMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);

    testLegend(assert, svg);

    svg = svg.first();

    verifyNoBubbles(assert, svg);

    const state_paths = svg.find('path.viz-map-state');
    assert.equal(state_paths.length, 53);

    const paths = svg.find('path.viz-map-path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 3213);
    paths.each((i, e) => {
      const fill = $(e).attr('fill');
      if (i === 1009) {
        assert.equal(
          fill,
          '#004529',
          'San Mateo County has a color value based on the default palette',
        );
      } else if (i === 1720) {
        assert.equal(
          fill,
          '#ffffe5',
          'Kings County has a color value based on the default palette',
        );
      } else {
        assert.equal(fill, 'white');
      }
    });
  });
});

QUnit.test('County bubbles', function (assert) {
  // Draw bubbles (the default is a choropleth)
  this.map.map_bubbles.value('true');

  createCountyMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);
    svg = svg.first();

    const state_paths = svg.find('path.viz-map-state');
    assert.equal(state_paths.length, 53);

    verifyBubbles(assert, svg);

    const paths = svg.find('path.viz-map-path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 3213);
    paths.each((i, e) => {
      const fill = $(e).attr('fill');
      assert.equal(fill, 'white');
    });
  });
});

function createDMAMap(assert, that, cb) {
  const { map } = that;

  // provide the data
  const result = arc.data.result();
  result.rows([
    [509, '200', '200'],
    [505, '400', '400'],
  ]);

  const col1 = arc.data
    .column()
    .colname('dma')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('dma')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(1);
  const col3 = arc.data
    .column()
    .colname('xFactor')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('size_shelf')
    .colindex(2);
  result.columns([col1, col2, col3]);

  map.viz_setSettings(choroplethMapSettings);
  map.viz_setDataResult(result);

  const restartQUnit = assert.async();
  map.viz_fetchCollateral('chart-div', false, () => {
    map.viz_drawVisual('chart-div', false);
    restartQUnit();
    cb();
  });
}

QUnit.test('DMA choropleth', function (assert) {
  createDMAMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);

    testLegend(assert, svg);

    svg = svg.first();
    verifyNoBubbles(assert, svg);

    const state_paths = svg.find('path.viz-map-state');
    assert.equal(state_paths.length, 53);

    const paths = svg.find('path.viz-map-path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 206);
    paths.each((i, e) => {
      const fill = $(e).attr('fill');
      if (i === 53) {
        assert.equal(
          fill,
          '#004529',
          'DMA 505 has a color value based on the default palette',
        );
      } else if (i === 67) {
        assert.equal(
          fill,
          '#ffffe5',
          'DMA 509 has a color value based on the default palette',
        );
      } else {
        assert.equal(fill, 'white');
      }
    });
  });
});

QUnit.test('DMA bubbles', function (assert) {
  // Draw bubbles (the default is a choropleth)
  this.map.map_bubbles.value('true');

  createDMAMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);
    svg = svg.first();

    verifyBubbles(assert, svg);

    const state_paths = svg.find('path.viz-map-state');
    assert.equal(state_paths.length, 53);
    const paths = svg.find('path.viz-map-path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 206);
    paths.each((i, e) => {
      const fill = $(e).attr('fill');
      assert.equal(fill, 'white');
    });
  });
});

function createZipcodeMap(assert, that, cb) {
  const { map } = that;

  // provide the data
  const result = arc.data.result();
  result.rows([
    ['94118', '837442', '837442'],
    ['10001', '8406000', '8406000'],
  ]);
  const col1 = arc.data
    .column()
    .colname('zipcode')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('zipcode')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(1);
  const col3 = arc.data
    .column()
    .colname('xFactor')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('size_shelf')
    .colindex(2);
  result.columns([col1, col2, col3]);

  // create the visual by setting the data and then drawing it
  map.viz_setDataResult(result);

  const restartQUnit = assert.async();
  map.viz_fetchCollateral('chart-div', false, () => {
    map.viz_drawVisual('chart-div', false);
    restartQUnit();
    cb();
  });
}

QUnit.test('ZIP Code Bubbles', function (assert) {
  // The default for ZIP code is bubbles - there is no way to get a choropleth
  createZipcodeMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);
    svg = svg.first();

    assert.equal(svg.length, 1);
    verifyBubbles(assert, svg);

    const paths = svg.find('path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 53);
  });
});

function createLatLongMap(assert, that, cb) {
  const { map } = that;

  // provide the data
  const result = arc.data.result();
  result.rows([
    ['-90.2', '42', '200', '200'],
    ['-85.4', '35', '400', '400'],
  ]);
  const col1 = arc.data
    .column()
    .colname('longitude')
    .coltype('REAL')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('longitude')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('latitude')
    .coltype('REAL')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('latitude')
    .colindex(1);
  const col3 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(2);
  const col4 = arc.data
    .column()
    .colname('xFactor')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('size_shelf')
    .colindex(3);
  result.columns([col1, col2, col3, col4]);

  // create the visual by setting the data and then drawing it
  map.viz_setDataResult(result);

  const restartQUnit = assert.async();
  map.viz_fetchCollateral('chart-div', false, () => {
    map.viz_drawVisual('chart-div', false);
    restartQUnit();
    cb();
  });
}

QUnit.test('LatLong bubbles', function (assert) {
  // The default for latlong is bubbles - there is no way to get a choropleth
  createLatLongMap(assert, this, () => {
    let svg = $('#chart-div').find('svg');

    // One svg for the map, another for the legend
    assert.equal(svg.length, 2);
    svg = svg.first();

    assert.equal(svg.length, 1);
    verifyBubbles(assert, svg);

    const paths = svg.find('path');
    assert.equal(svg.length, 1);
    assert.equal(paths.length, 241);
  });
});

function testMapSpecificity(assert, that, cb) {
  const { map } = that;

  // provide the data
  const result = arc.data.result();
  result.rows([
    ['California', '94118', '837442', '837442'],
    ['New York', '10001', '8406000', '8406000'],
  ]);

  const col1 = arc.data
    .column()
    .colname('state')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('state')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('zipcode')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('zipcode')
    .colindex(1);
  const col3 = arc.data
    .column()
    .colname('population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(2);
  const col4 = arc.data
    .column()
    .colname('xFactor')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('size_shelf')
    .colindex(3);
  result.columns([col1, col2, col3, col4]);

  // create the visual by setting the data and then drawing it
  map.viz_setDataResult(result);

  const restartQUnit = assert.async();
  map.viz_fetchCollateral('chart-div', false, () => {
    map.viz_drawVisual('chart-div', false);
    restartQUnit();
    cb();
  });
}

QUnit.test('Test Map Specificity', function (assert) {
  // The map with the higher resolution should be drawn, regardless of the
  // order inside the geo shelves
  testMapSpecificity(assert, this, () => {
    let svg = $('#chart-div').find('svg');
    svg = svg.first();

    assert.equal(svg.length, 1);
    verifyBubbles(assert, svg);

    const paths = svg.find('path');
    assert.equal(svg.find('path').length, 53);

    paths.each((_, e) => {
      const fill = $(e).attr('fill');
      assert.equal(fill, 'white');
    });
  });
});
