QUnit.module('visuals.arc_map_interactive', {
  beforeEach() {
    arc.testutil = arc.testutil || {};
    arc.testutil.leafletManualLoad = true;

    // main chart object
    this.leaflet = arc.visual.plugins.leaflet();
    // create the lib object
    this.driver = arc.report.driver(this.leaflet);
    this.driver.chart_div('chart-div');
  },

  afterEach() {
    this.driver.undraw();
    $('#chart-div').empty();
  },
});

function svgDispatchMouseover(svgElement) {
  const event = new Event('mouseover', { bubbles: false, cancelable: false });
  svgElement.dispatchEvent(event);
}

const formatter = d3.format('.3s');

const googleApiId = {
  setting_key: 'google_api_id',
  setting_value: 'AIzaSyCvGKEdXC7GbToXFXxHs8m-52_9dZv35MY',
};

const siteSettings = [googleApiId];
const colors = ['#ff0000', '#ffff00', '#00ff00'];
const colorSetting = {
  colorSettings: {},
  colors,
  reverse_colors: false,
};
const settings = {
  settings: colorSetting,
  user_settings: {
    heatmap_enabled: 'false',
    heatmap_legend: 'false',
    intensity_mag: '10000',
    heatmap_radius: '50',
    cluster_enabled: 'false',
    cluster_color: '#120c04',
    cluster_radius: '80',
    circle_enabled: 'false',
    circle_legend_area: 'false',
    circle_legend_color: 'false',
    circle_area_range: '100-10000',
    circle_radius_range: '10-100',
    marker_enabled: 'true',
    color_palette: JSON.stringify(colorSetting),
    visual_dataset_scope: 'false',
    'tile-style': 'google',
    send_aggregate_parameters: 'true',
  },
};

function createLatLongLeaflet(
  assert,
  leaflet,
  visualSettings,
  rowsToUse,
  numMeasures,
  numDims,
  cb,
) {
  const result = arc.data.result();
  const rows = [];

  rowsToUse.forEach((row, rowIndex) => {
    const rowCopy = row.concat();
    for (let i = 1; i < numMeasures + 1; i++) {
      rowCopy.push((rowIndex + 1) * i);
    }

    if (numDims === 1) {
      rowCopy.push(`dim${(rowIndex + 1) * (numMeasures + 1)}`);
    }

    rows.push(rowCopy);
  });

  result.rows(rows);
  const col1 = arc.data
    .column()
    .colname('longitude')
    .coltype('REAL')
    .setDisplayFormat(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('longitude');
  const col2 = arc.data
    .column()
    .colname('latitude')
    .coltype('REAL')
    .setDisplayFormat(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('latitude');
  const col3 = arc.data
    .column()
    .colname('Record Count')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .setDisplayFormat(formatter);
  const col4 = arc.data
    .column()
    .colname('Population')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('color_shelf')
    .setDisplayFormat(formatter);
  const col5 = arc.data
    .column()
    .colname('Country')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('color_shelf')
    .setDisplayFormat(formatter);
  const columns = [col1, col2, col3, col4].slice(0, 2 + numMeasures);
  if (numDims === 1) {
    columns.push(col5);
  }

  columns.forEach((col, i) => col.colindex(i));
  result.columns(columns);

  leaflet.viz_setSettings(visualSettings, false, siteSettings);
  leaflet.viz_setDataResult(result);

  const restartQUnit = assert.async();
  leaflet.viz_fetchCollateral('chart-div', false, (chart_div, redraw) => {
    leaflet.viz_drawVisual(chart_div, redraw);
    restartQUnit();
    cb();
  });
}

QUnit.test('General Tests', function (assert) {
  // Tests the marker layer works and that tooltips are being applied
  // Tests that the map works with 0, 1, and 2 measures
  // Tests both the Tile Style Settings (Google / Mapbox)

  const latLongRows = [
    ['-74.1', '40.7'],
    ['-122.4', '37.7'],
    ['-87.6', '41.8'],
  ];

  function evaluateMap(numMeasures) {
    const $viz_div = $('#chart-div');
    const $map_pane = $viz_div.find('.leaflet-map-pane');

    // Check to see that the map is drawing, should only be 1 map pane
    assert.equal(
      $map_pane.length,
      1,
      `Map Pane is Missing for ${numMeasures} measures.`,
    );

    // Check to see the number of markers === 3
    const $markers = $viz_div.find('.leaflet-marker-icon');
    assert.equal(
      $markers.length,
      3,
      `Some markers are missing for ${numMeasures} measures.`,
    );

    // Check to see that the tooltips are appearing
    const firstMarker = $markers[0];
    $(firstMarker).on('mouseover.testutil', () => {
      const $tooltipRows = $('#tooltip').find('tr');
      const $firstRow = $tooltipRows.first();
      const $secondRow = $($tooltipRows[1]);

      assert.equal(
        $firstRow.html(),
        `<td class="header-label">longitude</td><td class="viz-tooltip-num">${formatter(
          -74.1,
        )}</td>`,
        'Tooltip is incorrect.',
      );
      assert.equal(
        $secondRow.html(),
        `<td class="header-label">latitude</td><td class="viz-tooltip-num">${formatter(
          40.7,
        )}</td>`,
        'Tooltip is incorrect.',
      );
    });

    svgDispatchMouseover(firstMarker);
  }

  const baseSettings = JSON.parse(JSON.stringify(settings));
  // Tests 0, 1, and 2 measures
  for (let i = 0; i < 3; i++) {
    createLatLongLeaflet(
      assert,
      this.leaflet,
      baseSettings,
      latLongRows,
      i,
      0,
      () => {
        evaluateMap(i);
      },
    );
  }
});

QUnit.test('Test Heatmap Layer', function (assert) {
  // Tests that the map draws
  // Tests that the map exists with the proper settings for the legend
  const latLongRows = [
    ['-74.1', '40.7'],
    ['-122.4', '37.7'],
    ['-87.6', '41.8'],
  ];

  const legendNumberFormat = arc.report.base().tickFormatDefaultNumeric;

  function evaluateMap() {
    const $viz_div = $('#chart-div');
    const $map_pane = $viz_div.find('.leaflet-map-pane');

    // Check to see that the map is drawing, should only be 1 map pane
    assert.equal($map_pane.length, 1, 'Map Pane is Missing.');

    // Check to see the heatmap exists
    const $heatmap = $viz_div.find('.leaflet-heatmap-layer');
    assert.equal($heatmap.length, 1, 'Heatmap layer is missing.');

    // Check to see that the heatmap legend exists
    const $legend = $viz_div.find('.heatmap_legend');
    assert.equal($legend.length, 1, 'Heatmap legend is missing.');

    const $labels = $legend.find('text');

    // Check legend first value
    assert.equal(
      $labels.first().text(),
      legendNumberFormat(10000),
      'Legend last value is 10.0k',
    );

    // Check legend last value
    assert.equal(
      $labels.last().text(),
      legendNumberFormat(4096),
      'Legend first value is 4.10k',
    );
  }

  const baseSettings = JSON.parse(JSON.stringify(settings));
  const userSettings = baseSettings.user_settings;

  userSettings.marker_enabled = 'false';
  userSettings.heatmap_enabled = 'true';
  userSettings.heatmap_legend = 'true';

  createLatLongLeaflet(
    assert,
    this.leaflet,
    baseSettings,
    latLongRows,
    1,
    0,
    evaluateMap,
  );
});

QUnit.test('Test Cluster Layer', function (assert) {
  // Tests that the map exists
  // Tests cluster color setting
  // Tests that the tooltips are applied

  const latLongRows = [
    ['-74.1', '40.7'],
    ['-74.5', '40.9'],
    ['-87.6', '41.8'],
  ];

  function evaluateMap() {
    const $viz_div = $('#chart-div');
    const $map_pane = $viz_div.find('.leaflet-map-pane');

    // Check to see that the map is drawing, should only be 1 map pane
    assert.equal($map_pane.length, 1, 'Map Pane is Missing.');

    // Check to see the cluster exists, and is of the right color
    const $cluster = $viz_div.find('.leaflet-marker-pane .marker-cluster div');
    assert.equal(
      $cluster.css('background-color'),
      'rgb(18, 12, 4)',
      'Cluster is missing or is of the incorrect color.',
    );

    // Check to see the number of markers === 1
    const $markers = $viz_div.find('img.leaflet-marker-icon');
    assert.equal($markers.length, 1, 'Incorrect number of markers');

    // Check to see that the tooltips are appearing
    const firstMarker = $markers[0];
    $(firstMarker).on('mouseover.testutil', () => {
      const $tooltipRows = $('#tooltip').find('tr');
      const $firstRow = $tooltipRows.first();
      const $secondRow = $($tooltipRows[1]);

      assert.equal(
        $firstRow.html(),
        `<td class="header-label">longitude</td><td class="viz-tooltip-num">${formatter(
          -87.6,
        )}</td>`,
        'Tooltip is incorrect.',
      );
      assert.equal(
        $secondRow.html(),
        `<td class="header-label">latitude</td><td class="viz-tooltip-num">${formatter(
          41.8,
        )}</td>`,
        'Tooltip is incorrect.',
      );
    });

    svgDispatchMouseover(firstMarker);
  }

  const baseSettings = JSON.parse(JSON.stringify(settings));
  const userSettings = baseSettings.user_settings;

  userSettings.marker_enabled = 'false';
  userSettings.cluster_enabled = 'true';

  createLatLongLeaflet(
    assert,
    this.leaflet,
    baseSettings,
    latLongRows,
    1,
    0,
    evaluateMap,
  );
});

QUnit.test('Test Circle Layer', function (assert) {
  // Tests that the map exists, and tests the use with 0, 1, and 2 measures, and
  // with one measure and one dimension
  // Test Circle Area Legend
  // Test Circle Color Legend

  const latLongRows = [
    ['-74.1', '40.7'],
    ['-122.4', '37.7'],
    ['-87.6', '41.8'],
  ];

  function checkRadiusVal(svgMarker, value) {
    const targetRadius = value;

    const radiusRegex = /[a][0-9.]+[,]/;
    const d = $(svgMarker).attr('d');
    const radiusString = radiusRegex.exec(d)[0];
    const paintedRadius = formatter(
      Number(radiusString.substring(1, radiusString.length - 1)),
    );

    assert.equal(
      targetRadius,
      paintedRadius,
      `Circle marker with radius ${paintedRadius} should have radius ${targetRadius}`,
    );
  }

  function evaluateMap(numMeasures, numDims) {
    const $viz_div = $('#chart-div');
    const $map_pane = $viz_div.find('.leaflet-map-pane');

    // Check to see that the map is drawing, should only be 1 map pane
    assert.equal(
      $map_pane.length,
      1,
      `Map Pane is Missing for ${numMeasures} measures.`,
    );

    // Check to see the number of markers === 3
    const $markers = $viz_div.find('.circle_marker');
    assert.equal(
      $markers.length,
      3,
      `Some markers are missing for ${numMeasures} measures.`,
    );

    let $colorLegend;
    if (numDims === 1 || numMeasures === 2) {
      $colorLegend = $viz_div.find('.circle_color_legend .small-legend');
      assert.equal($colorLegend.length, 1, 'Circle Color Legend is missing.');
    }

    const $areaLegend = $viz_div.find('.circle_area_legend');
    if (numMeasures > 0) {
      assert.equal($areaLegend.length, 1, 'Circle Area Legend is missing.');
    }

    let scale;
    switch (numMeasures) {
      case 0:
        // Check to see markers are the same size, and if they are the right color
        $markers.each((i, marker) => {
          checkRadiusVal(marker, 10);

          if (numDims === 1) {
            const expectedColor = arc.colorpalette.convertToRGBAString(
              colors[i],
            );
            assert.equal(
              $(marker).attr('fill'),
              expectedColor,
              `Marker should be ${expectedColor}`,
            );
          } else {
            assert.equal(
              $(marker).attr('fill'),
              '#ff0000',
              'Marker should be #ff0000.',
            );
          }
        });

        if (numDims === 1) {
          assert.equal(
            $colorLegend.find('rect').length,
            3,
            'Should be three swatches in color legend',
          );

          const $labels = $colorLegend.find('text');
          assert.equal(
            $colorLegend.find('rect').length,
            3,
            'Should be three swatches in color legend',
          );
          assert.equal(
            $labels.length,
            4,
            'Should be 4 text elements, including the title, in the color legend',
          );
          assert.equal(
            $($colorLegend.find('text')[1]).text(),
            'dim3',
            'First color legend value should be dim3',
          );
        }

        break;
      case 1:
        scale = d3.scale.linear().domain([1, 3]).range([10, 100]);

        $markers.each((i, marker) => {
          checkRadiusVal(marker, scale(i + 1));

          if (numDims === 1) {
            const expectedColor = arc.colorpalette.convertToRGBAString(
              colors[i],
            );
            assert.equal(
              $(marker).attr('fill'),
              expectedColor,
              `Marker should be ${expectedColor}`,
            );
          } else {
            assert.equal(
              $(marker).attr('fill'),
              '#ff0000',
              'Marker should be #ff0000.',
            );
          }
        });

        if (numDims === 1) {
          assert.equal(
            $colorLegend.find('rect').length,
            3,
            'Should be three swatches in color legend',
          );

          const $labels = $colorLegend.find('text');
          assert.equal(
            $colorLegend.find('rect').length,
            3,
            'Should be three swatches in color legend',
          );
          assert.equal(
            $labels.length,
            4,
            'Should be 4 text elements, including the title, in the color legend',
          );
          assert.equal(
            $($colorLegend.find('text')[1]).text(),
            'dim6',
            'First color legend value should be dim6',
          );
        }

        break;
      case 2:
        scale = d3.scale.linear().domain([2, 6]).range([10, 100]);

        $markers.each((i, marker) => {
          checkRadiusVal(marker, scale((i + 1) * 2));
          const expectedColor = arc.colorpalette.convertToRGBAString(colors[i]);
          assert.equal(
            $(marker).attr('fill'),
            expectedColor,
            `Marker should be ${expectedColor}`,
          );
        });

        assert.equal(
          $($colorLegend.find('text')[1]).text(),
          `≥${formatter(2)}`,
          'First color legend value should be ≥2.00',
        );
        assert.equal(
          $($areaLegend.find('text')[1]).text(),
          `${formatter(3)}`,
          'First area legend value should be 3.00',
        );
        break;
      default:
    }

    // Check to see that the tooltips are appearing
    const firstMarker = $markers[0];
    $(firstMarker).on('mouseover.testutil', () => {
      const $tooltipRows = $('#tooltip').find('tr');
      const $firstRow = $tooltipRows.first();
      const $secondRow = $($tooltipRows[1]);
      assert.equal(
        $firstRow.html(),
        `<td class="header-label">longitude</td><td class="viz-tooltip-num">${formatter(
          -74.1,
        )}</td>`,
        'Tooltip is incorrect.',
      );
      assert.equal(
        $secondRow.html(),
        `<td class="header-label">latitude</td><td class="viz-tooltip-num">${formatter(
          40.7,
        )}</td>`,
        'Tooltip is incorrect.',
      );
    });

    svgDispatchMouseover(firstMarker);
  }

  const baseSettings = JSON.parse(JSON.stringify(settings));
  const userSettings = baseSettings.user_settings;
  userSettings.marker_enabled = 'false';
  userSettings.circle_enabled = 'true';
  userSettings.circle_legend_area = 'true';
  userSettings.circle_legend_color = 'true';

  // Tests varying measures
  for (let i = 0; i < 3; i++) {
    createLatLongLeaflet(
      assert,
      this.leaflet,
      baseSettings,
      latLongRows,
      i,
      0,
      () => {
        evaluateMap(i, 0);
      },
    );

    if (i < 2) {
      createLatLongLeaflet(
        assert,
        this.leaflet,
        baseSettings,
        latLongRows,
        i,
        1,
        () => {
          evaluateMap(i, 1);
        },
      );
    }
  }
});

function createZipcodeLeaflet(assert, leaflet, rows, cb) {
  // provide the data
  const result = arc.data.result();
  result.rows(rows);

  const col1 = arc.data
    .column()
    .colname('zipcode')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('zipcode')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('Record Count')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(1);
  result.columns([col1, col2]);

  leaflet.viz_setSettings(settings, false, siteSettings);
  leaflet.viz_setDataResult(result);

  const restartQUnit = assert.async();
  leaflet.viz_fetchCollateral('chart-div', false, (chart_div, redraw) => {
    leaflet.viz_drawVisual(chart_div, redraw);
    restartQUnit();
    cb();
  });
}

QUnit.test('ZIP code, test valid ZIP codes', function (assert) {
  // Just test to see that it draws and filters out invalid ZIP codes.
  // If the other tests work for lat/long maps, then they will work for ZIP code maps.

  const valid_rows = [
    ['10001', '1'],
    ['94102', '1'],
    ['60601', '1'],
  ];

  // Test valid ZIP codes here
  createZipcodeLeaflet(assert, this.leaflet, valid_rows, () => {
    const $viz_div = $('#chart-div');

    // Check to see that the map is drawing, should only be 1 map pane
    const map_pane = $viz_div.find('.leaflet-map-pane');
    assert.equal(map_pane.length, 1, 'Map Pane is Missing for ZIP code Map.');

    // Check to see the number of markers === 3
    const markers = $viz_div.find('.leaflet-marker-icon');
    assert.equal(
      markers.length,
      3,
      'Some markers are missing for ZIP code Map.',
    );
  });
});

QUnit.test('ZIP codes, test with invalid ZIP codes', function (assert) {
  const invalid_rows = [
    ['10001', '1'],
    ['asdf#', '1'],
    ['00000', '1'],
  ];

  // Test invalid ZIP codes here
  createZipcodeLeaflet(assert, this.leaflet, invalid_rows, () => {
    const $viz_div = $('#chart-div');
    const map_pane = $viz_div.find('.leaflet-map-pane');

    // Check to see that the map is drawing, should only be 1 map pane
    assert.equal(map_pane.length, 1, 'Map Pane is Missing for ZIP Code Map.');

    // Check to see the number of markers === 1
    const markers = $viz_div.find('.leaflet-marker-icon');
    assert.equal(
      markers.length,
      1,
      'Should only be one marker due to invalid rows.',
    );
  });
});

function testLeafletSpecificity(assert, leaflet, rows, cb) {
  // If both a zipcode and a lat-long configuration are specified, the map should
  // default to latitude and longitude
  const result = arc.data.result();
  result.rows(rows);

  const col1 = arc.data
    .column()
    .colname('zipcode')
    .coltype('STRING')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('zipcode')
    .colindex(0);
  const col2 = arc.data
    .column()
    .colname('longitude')
    .coltype('REAL')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('longitude')
    .colindex(1);
  const col3 = arc.data
    .column()
    .colname('latitude')
    .coltype('REAL')
    .colproperty(arc.report.columnTypeEnum.DIMENSION)
    .colsource('geo_shelf')
    .colgeotype('latitude')
    .colindex(2);
  const col4 = arc.data
    .column()
    .colname('Record Count')
    .coltype('BIGINT')
    .colproperty(arc.report.columnTypeEnum.AGGREGATE)
    .colsource('aggregates_shelf')
    .colindex(3);
  result.columns([col1, col2, col3, col4]);

  leaflet.viz_setSettings(settings, false, siteSettings);
  leaflet.viz_setDataResult(result);

  const restartQUnit = assert.async();
  leaflet.viz_fetchCollateral('chart-div', false, (chart_div, redraw) => {
    leaflet.viz_drawVisual(chart_div, redraw);
    restartQUnit();
    cb();
  });
}

QUnit.test('Test Map Specificity', function (assert) {
  const rows = [
    ['unused_zip', '-74.1', '40.7', 1],
    ['unused_zip', '-122.4', '37.7', 1],
    ['unused_zip', '-87.6', '41.8', 1],
  ];

  // Test invalid ZIP codes here
  testLeafletSpecificity(assert, this.leaflet, rows, () => {
    const $viz_div = $('#chart-div');

    // Check to see the number of markers === 3 since latlong configuration
    // should override zipcode configuration
    const $markers = $viz_div.find('.leaflet-marker-icon');
    assert.equal(
      $markers.length,
      3,
      'Latlong configuration is not override zipcode configuration',
    );
  });
});
