QUnit.module('visuals_arc_scatter', {
  beforeEach() {
    // main chart object
    const scatter = arc.visual.plugins.scatter();

    // create the lib object
    const driver = arc.report.driver(scatter);
    driver.chart_div('chart-div');

    // provide the data
    const result = arc.data.result();
    result.rows([
      ['Palo Alto', 'CA', '9123', '12300'],
      ['Menlo Park', 'CA', '1234', '1345'],
      ['San Mateo', 'CA', '12', '123'],
      ['Raleigh', 'NC', '24', '123'],
      ['Greensboro', 'NC', '1234', '1234'],
      ['Chapel Hill', 'NC', '4356', '32'],
    ]);

    const col0 = arc.data
      .column()
      .colname('City')
      .coltype('STRING')
      .colproperty(arc.report.columnTypeEnum.DIMENSION)
      .colsource('x_shelf')
      .colindex(0);
    const col1 = arc.data
      .column()
      .colname('State')
      .coltype('STRING')
      .colproperty(arc.report.columnTypeEnum.DIMENSION)
      .colsource('color_shelf')
      .colindex(1);
    const col2 = arc.data
      .column()
      .colname('record_count')
      .coltype('INT')
      .colproperty(arc.report.columnTypeEnum.AGGREGATE)
      .colsource('y_shelf')
      .colindex(2);
    const col3 = arc.data
      .column()
      .colname('population')
      .coltype('INT')
      .colproperty(arc.report.columnTypeEnum.AGGREGATE)
      .colsource('size_shelf')
      .colindex(3);
    result.columns([col0, col1, col2, col3]);

    // create the visual by setting the data and then drawing it
    scatter.viz_setDataResult(result);
    scatter.viz_drawVisual('chart-div');

    // store elements on the page to be tested
    this.svg = $('#chart-div svg');
    this.legendG = $('table>tbody>tr>td>div>svg>.viz-base-legend-g');
    this.xAxis = $('#chart-div .x.arc-viz-axis');
    this.yAxis = $('#chart-div .y.arc-viz-axis');

    this.canvas = d3.select('#chart-div').selectAll('canvas');
  },

  afterEach() {
    $('#chart-div').empty();
  },
});

QUnit.test('Scatter Plot and Legend Elements Present', function (assert) {
  const xAxisTicks = this.xAxis[0].getElementsByClassName('tick');
  const yAxisTicks = this.yAxis[0].getElementsByClassName('tick');

  assert.equal(this.svg.length, 3, 'There are 3 svg elements (2 are nested)');
  assert.equal(
    this.legendG.length,
    2,
    'There are two g elements that contains the legend',
  );

  assert.equal(this.xAxis.length, 1, 'There is an x axis element');
  assert.equal(this.yAxis.length, 1, 'There is an y axis element');
  assert.equal(xAxisTicks.length, 6, 'There are 6 marks on the x axis');
  assert.equal(yAxisTicks.length, 10, 'There are 10 marks on the y axis');

  assert.equal(this.canvas.size(), 3, 'There are three canvas elements');

  const plots = arc.canteen.getCanvasData(this.canvas);
  assert.equal(arc.canteen.countMethod(plots[0], 'arc'), 6);
});

QUnit.test('Scatter Plot Elements Correct', function (assert) {
  const xAxisTicks = this.xAxis[0].getElementsByClassName('tick');
  const yAxisTicks = this.yAxis[0].getElementsByClassName('tick');
  const xAxisLabels = [];
  const yAxisLabels = [];
  const xAxisLabelsExpected = [
    'Chapel Hill',
    'Greensboro',
    'Menlo Park',
    'Palo Alto',
    'Raleigh',
    'San Mateo',
  ];
  const yAxisLabelsExpected = [
    '0',
    '1,000',
    '2,000',
    '3,000',
    '4,000',
    '5,000',
    '6,000',
    '7,000',
    '8,000',
    '9,000',
  ];
  const circleSizesExpected = [
    '50',
    '21.788247511427542',
    '21.200721668327663',
    '12.065086608774397',
    '12.065086608774397',
    '10',
  ];
  const circlePositionsExpected = [
    [207, 17],
    [148, 308],
    [88, 308],
    [311, 353],
    [254, 353],
    [30, 201],
  ];
  for (let i = 0; i < xAxisTicks.length; i++) {
    xAxisLabels.push(xAxisTicks[i].textContent);
  }

  for (let j = 0; j < yAxisTicks.length; j++) {
    yAxisLabels.push(yAxisTicks[j].textContent);
  }

  assert.deepEqual(
    xAxisLabels,
    xAxisLabelsExpected,
    'The x axis labels are correct',
  );
  assert.deepEqual(
    yAxisLabels,
    yAxisLabelsExpected,
    'The y axis labels are correct',
  );

  const plots = arc.canteen.getCanvasData(this.canvas);
  assert.equal(arc.canteen.countMethod(plots[0], 'arc'), 6);
  const a = arc.canteen.methodArgs(plots[0], 'arc');

  // The difference between Mac/Linux positions
  const scatterPositionTolerance = 80;
  for (let k = 0; k < 6; k++) {
    // Check position here because we need to compare actual x/y positions to
    // range of expected values (check it's within 10px larger or higher)
    const actualX = a[k][0];
    const actualY = a[k][1];
    const expectedX = circlePositionsExpected[k][0];
    const expectedY = circlePositionsExpected[k][1];
    assert.ok(
      arc.misc.within_tolerance(expectedX, actualX, scatterPositionTolerance),
      `The ${k} circle is in the right x position:${expectedX}:${actualX}`,
    );
    assert.ok(
      arc.misc.within_tolerance(expectedY, actualY, scatterPositionTolerance),
      `The ${k} circle is in the right y position`,
    );

    assert.ok(
      arc.misc.within_tolerance(a[k][2], circleSizesExpected[k], 2),
      `The ${k} circle is the right size`,
    );
  }
});

QUnit.test('Legend Elements Correct', function (assert) {
  const rectangles = this.legendG[0].getElementsByTagName('rect');
  const labels = this.legendG[0].getElementsByTagName('text');
  const rectangleColors = [];
  const labelTexts = [];
  const rectangleColorsExpected = ['#0097a7', '#7dcad2'];
  const labelTextsExpected = ['CA', 'NC'];

  assert.equal(rectangles.length, 2, 'The legend has 2 rectangles');
  assert.equal(labels.length, 2, 'The legend has 2 labels');

  for (let i = 0; i < rectangles.length; i++) {
    rectangleColors.push(rectangles[i].style.fill);
    labelTexts.push(labels[i].textContent);
  }

  for (let i = 0; i < rectangleColors.length; i++) {
    rectangleColors[i] = arc.misc.colorToHex(rectangleColors[i]);
  }

  assert.deepEqual(
    rectangleColors,
    rectangleColorsExpected,
    'The legend has the correctly colored rectangles',
  );
  assert.deepEqual(
    labelTexts,
    labelTextsExpected,
    'The legend has the correct labels',
  );
});
