QUnit.module('visuals_arc_table', (hooks) => {
  hooks.beforeEach(function () {
    // Main chart object and driver setup
    this.table = arc.visual.plugins.table();
    this.driver = arc.report.driver(this.table).chart_div('chart-div');
  });

  hooks.afterEach(() => {
    $('#chart-div').empty();
  });

  function setupTable(col1Order, col2Order) {
    $('#chart-div').empty(); // Clear previous visualizations

    const result = arc.data.result();
    result.rows([
      ['Africa', 1901],
      ['Americas', 1900],
      ['Europe', 1901],
      ['Oceania', 1901],
      ['Africa', 1900],
      ['Asia', 1901],
      ['Asia', 1900],
      ['Europe', 1900],
      ['Americas', 1901],
      ['Oceania', 1900],
    ]);

    const col1 = arc.data
      .column()
      .colname('un_region')
      .coltype('STRING')
      .colproperty(arc.report.columnTypeEnum.DIMENSION)
      .colsource('dimensions_shelf')
      .colindex(0)
      .order(col1Order);
    const col2 = arc.data
      .column()
      .colname('year')
      .coltype('BIGINT')
      .colproperty(arc.report.columnTypeEnum.DIMENSION)
      .colsource('dimensions_shelf')
      .colindex(1)
      .order(col2Order);
    result.columns([col1, col2]);
    result.finalize();

    this.table.viz_setDataResult(result);
    this.table.viz_drawVisual('chart-div');
  }

  QUnit.test(
    'Sort by year descending, then un_region ascending',
    function (assert) {
      const col1Order = { priority: () => 2, ascending: () => true };
      const col2Order = { priority: () => 1, ascending: () => false };
      setupTable.call(this, col1Order, col2Order);

      // Extract the first column's text to verify sorting order
      const rows = $('#chart-div')
        .find('table tbody tr')
        .toArray()
        .map((row) => $(row).find('td').eq(0).text());

      const expectedOrder = [
        'Africa',
        'Americas',
        'Asia',
        'Europe',
        'Oceania', // 1901
        'Africa',
        'Americas',
        'Asia',
        'Europe',
        'Oceania', // 1900
      ];

      assert.deepEqual(
        rows,
        expectedOrder,
        'Rows are sorted correctly by year descending, then un_region ascending',
      );
    },
  );

  QUnit.test(
    'Sort by un_region ascending, then year descending',
    function (assert) {
      const col1Order = { priority: () => 1, ascending: () => true };
      const col2Order = { priority: () => 2, ascending: () => false };
      setupTable.call(this, col1Order, col2Order);

      // Extract the second column's text (year) to verify sorting order within each region
      const rows = $('#chart-div')
        .find('table tbody tr')
        .toArray()
        .map((row) => $(row).find('td').eq(1).text());

      const expectedYears = [
        '1,901',
        '1,900', // Africa
        '1,901',
        '1,900', // Americas
        '1,901',
        '1,900', // Asia
        '1,901',
        '1,900', // Europe
        '1,901',
        '1,900', // Oceania
      ];

      assert.deepEqual(
        rows,
        expectedYears,
        'Rows are sorted correctly by un_region ascending, then year descending',
      );
    },
  );

  QUnit.test('Table Element Internals', (assert) => {
    const t = arc.visual.user_settings.settingTableTemplate();
    t.unittest_validate();
    assert.expect(0);
  });

  QUnit.test('Test Table Limit Spinner', function (assert) {
    const table = arc.datasets.Table().tablename('default.census_pop');

    const dataset = arc.datasets.Dataset().add_table(table);

    this.driver.dataset(dataset);

    const shelf_id = 'dimensions_shelf';
    const shelf_contents = [
      {
        dataset_colname: 'cereal_name',
        dataset_coltype: 'STRING',
        order: {
          ascending: false,
          priority: 1,
        },
        order_limit: '1',
      },
    ];

    this.table.viz_drawDesigner('app-shelves');

    this.driver.restoreLimit({ limit: '10' });
    let { limit } = JSON.parse(this.driver.shelvesToDataRequest().dsreq());
    assert.equal(limit, 10, 'Limit Spinner is not being applied.');

    this.driver.shelves().add(shelf_id, shelf_contents);
    limit = JSON.parse(this.driver.shelvesToDataRequest().dsreq()).limit;
    assert.equal(
      limit,
      1,
      'Adding a limit to a column should override limit spinner',
    );
  });

  QUnit.test('Simple Table', function (assert) {
    const col1Order = { priority: () => 1, ascending: () => true };
    const col2Order = { priority: () => 2, ascending: () => false };
    setupTable.call(this, col1Order, col2Order);

    const tables = $('#chart-div').find('table');

    assert.equal(tables.length, 3);

    const tr = tables.find('tbody tr');
    assert.equal(tr.length, 10);

    const td = tr.find('td');
    assert.equal(td.length, 20);
  });
});
