import path from "path";
import glob from 'glob'
import { Configuration, BannerPlugin } from "webpack";
import MiniCssExtractPlugin from "mini-css-extract-plugin";
import CssMinimizerPlugin from "css-minimizer-webpack-plugin";
import RemoveEmptyScriptsPlugin from "webpack-remove-empty-scripts";
import WebpackConcatPlugin from '@mcler/webpack-concat-plugin';
import fs from 'fs';
import TerserPlugin from 'terser-webpack-plugin';

export const staticDirectoryPath = path.join(__dirname, "../..");

const copyright = fs
  .readFileSync(path.resolve(staticDirectoryPath, '../../copyright.txt'))
  .toString();

const SHARED_BETWEEN_DESKTOP_AND_MOBILE_CSS_ENTRIES = [
  "table-style",
  "arcadia",
  "home",
  "arc_navbar",
  "admin",
]

const SEPARATE_CSS_ENTRIES = [
  "new_data_page_style",
  "new_direct_access_style",
  "new_cloudera_colouring",
  "ml",
  "arcadia_embed",
  "dashboard_edit_sql",
  "smart-acceleration",
  "custom_styles",
  "dashboard",
  "app",
  "arc_sidemenu",
  "data",
  "connection_modal",
  "arc_dashboard_template",
  "browser_contextmenu",
  "mobile",
  "data-model",
  "swagger_ui",
];

const getArcJsBundleFilePaths = () => [
  "js/arc/**/*.js",
  "js/grammar/*.js",
].flatMap((pattern) => glob.sync(path.join(staticDirectoryPath, pattern)));

const getDesktopCssBundleFilePaths = () => glob.sync(path.join(staticDirectoryPath, 'scss/arc/*.scss'))
  .filter(
    (filePath) => !SEPARATE_CSS_ENTRIES.some((entry) => filePath.endsWith(`${entry}.scss`)) &&
      !SHARED_BETWEEN_DESKTOP_AND_MOBILE_CSS_ENTRIES.some((entry) => filePath.endsWith(`${entry}.scss`)
  )
);

const getArcCssChunks = () => SEPARATE_CSS_ENTRIES.reduce((entries, name) => {
  entries[name] = `../scss/arc/${name}.scss`
  return entries;
}, {});

export const commonWebpackConfiguration: Configuration = {
  entry: {
    // Our source code bundles
    'bundle-js': getArcJsBundleFilePaths(),
    'desktop-css': getDesktopCssBundleFilePaths(),
    ...getArcCssChunks(),
    'expose-node-modules-to-ext': './webpack/entries/expose-node-modules-to-ext.js',
    'shared-between-desktop-and-mobile-css': SHARED_BETWEEN_DESKTOP_AND_MOBILE_CSS_ENTRIES.map((name) => `../scss/arc/${name}.scss`),
    // 3rd party bundle from css/ext folder
    'ext-vendors-css': [
      '../css/ext/jQRangeSlider/jQRangeSlider.css',
      '../css/ext/slickGrid-2.4.44/slick.grid.css',
      '../css/ext/wan-spinner/wan-spinner.css',
      '../css/ext/daterangepicker.css',
      '../css/ext/font-awesome.css',
      '../css/ext/spectrum.css'
    ],
    canteen: './node_modules/canteen/build/canteen.min.js',
    "expose-shared-vendors-to-frontend": './ext/bootstrap-validator-0.11.9/validator.min.js',
  },
  output: {
    clean: true,
    filename: "[name].js",
  },
  module: {
    rules: [
      {
        test: /\.js$/,
        exclude: /node_modules/,
        use: {
          loader: "babel-loader",
          options: {
            presets: ["@babel/preset-env"],
          },
        },
      },
      {
        test: /\.(c|sc)ss$/i,
        use: [
          MiniCssExtractPlugin.loader,
          "css-loader",
          "sass-loader",
        ],
      }
    ],
  },
  plugins: [
    new RemoveEmptyScriptsPlugin({}),
    new MiniCssExtractPlugin({
      filename: '../../../../css/arc/[name].css',
    }),
    new BannerPlugin({ banner: copyright, exclude: ['canteen', 'node-modules', 'ext-vendor-css', 'expose-node-modules-to-ext'] }),
    // 3rd party bundles from js/ext folder
    new WebpackConcatPlugin({
      name: 'ext-vendors-js',
      filesToConcat: [
        "./ext/jquery.ui.touch-punch.js",
        "./ext/typeahead.bundle.min.js",
        "./ext/jQRangeSlider-5.7.2/jQAllRangeSliders-min.js",
        "./ext/wan-spinner/wan-spinner.js",
        "./ext/colorpicker/spectrum.min.js",
        "./ext/slickGrid-2.4.44/jquery.event.drag-2.3.0.min.js",
        "./ext/slickGrid-2.4.44/slick.core.js",
        "./ext/slickGrid-2.4.44/slick.grid.min.js",
        "./ext/slickGrid-2.4.44/slick.autotooltips.min.js",
        "./ext/daterangepicker.min.js",
        "./ext/sankey.js",
      ],
    }),
  ],
  optimization: {
    splitChunks: {
      cacheGroups: {
        // 3rd party bundle from node_modules
        vendor: {
          // Bundle everything inside node_modules, except canteen because
          // that will be only loaded for testing and not for production
          test: /[\\/]node_modules[\\/]((?!(canteen)).*)[\\/]/,
          name: 'node-modules',
          chunks: 'all',
        },
      },
    },
    minimizer: [
      new TerserPlugin({ extractComments: false }),
      new CssMinimizerPlugin(),
    ]
  },
  externals: {
    arc: "arc",
    moment: 'moment',
    jquery: "jQuery",
    d4: "d4",
    'fuse.js': "fuse.js",
    bootstrap: 'bootstrap',
    'short-unique-id': 'ShortUniqueId'
  },
  // This is a workaround for summernote version 0.8.20
  // more details at: https://github.com/summernote/summernote/issues/4200
  resolve: {
    alias: {
      jQuery: 'jquery'
    }
  }
};

if(process.env.COVERAGE=='true') {
  console.log('COVERAGE enabled');

  const istanbulLoader = {
    loader: '@ephesoft/webpack.istanbul.loader', // Must be first loader
    exclude: /(node_modules|\.test)/,
    test: /\.[tj]s$/,
  };

  commonWebpackConfiguration.module.rules.splice(0,0,istanbulLoader);
}