# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import os
import yaml

from django.conf import settings
from django.http import HttpResponse
from adminapi.views_adminapi import login_or_apikey_required

from  .constants import DOCUMENTED_ADMIN_API_ENDPOINTS, DOCUMENTED_NON_ADMIN_API_ENDPOINTS, DATA_API_ENDPOINTS


def generate_allowed_paths(endpoint_names):
    allowed_paths = []

    for endpoint in endpoint_names:
        base_path = f'/adminapi/v1/{endpoint}'
        allowed_paths.append(base_path)

        detailed_path = f'/adminapi/v1/{endpoint}/{{object_id}}'
        allowed_paths.append(detailed_path)

        if settings.CDV_ADDITIONAL_API_ENDPOINTS:
            if endpoint == 'datasets':
                sql_path = f'/adminapi/v1/{endpoint}/sql'
                allowed_paths.append(sql_path)

            if endpoint == 'visuals':
                smart_visual_path = f'/adminapi/v1/{endpoint}/smart'
                allowed_paths.append(smart_visual_path)

    return allowed_paths


def filter_openapi_spec(openapi_spec, allowed_paths):
    filtered_spec = openapi_spec.copy()
    filtered_spec['paths'] = {
        path: details
        for path, details in openapi_spec['paths'].items()
        if path in allowed_paths
    }
    return filtered_spec

@login_or_apikey_required
def openapi_spec_view(request):
    BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    file_path = os.path.join(BASE_DIR, 'static/swagger_docs/openapi-adminapi.yaml')
    with open(file_path, 'r') as file:
        openapi_spec = yaml.safe_load(file)

    if hasattr(settings, 'ADMIN_API_DEMO_LIST'):
        allowed_endpoints = (
            DOCUMENTED_ADMIN_API_ENDPOINTS
            if settings.ADMIN_API_DEMO_LIST == ['*']
            else settings.ADMIN_API_DEMO_LIST
        )
    else:
        allowed_endpoints = []
    allowed_paths = generate_allowed_paths(allowed_endpoints)
    allowed_paths += DOCUMENTED_NON_ADMIN_API_ENDPOINTS
    should_add_data_api_endpoints = hasattr(settings, 'DATA_API_ENABLED') and settings.DATA_API_ENABLED == True
    if should_add_data_api_endpoints:
        allowed_paths += DATA_API_ENDPOINTS
    filtered_spec = filter_openapi_spec(openapi_spec, allowed_paths)
    filtered_yaml = yaml.dump(filtered_spec, sort_keys=False)
    return HttpResponse(filtered_yaml, content_type='application/yaml')
