# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from django.core.management import call_command
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.base.introspection import BaseDatabaseIntrospection
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.state import ProjectState
from django.db.models import Model
from django.db.models.fields import Field

import logging
import unittest

LOG = logging.getLogger('arcweb.testmigrations')
CLEANUP_LOG = logging.getLogger('arcweb.testmigrations.cleanup')
CLEANUP_LOG.setLevel(logging.INFO) # change this to DEBUG to see more

# will be set after the first intercepted migration
connection: BaseDatabaseWrapper = None

def get_field(model: Model, field_name: str):
  fields: list[Field] = model._meta.fields
  field = [field for field in  fields if field.name == field_name][0]
  return field

def delete_field(schema_editor: BaseDatabaseSchemaEditor, model: Model, field_name: str):
  field = get_field(model, field_name)
  with schema_editor:
    schema_editor.remove_field(model, field=field)

class MigrationsTestBase(unittest.TestCase):

  def setUp(self) -> None:
    self.start_migration_interception()

  def tearDown(self) -> None:
    self.stop_migration_interception
    self.clear_database(self.project_state)

  def start_migration_interception(self):
    """
    Monkey-patch MigrationExecutor so that we have acces to the last migration state
    """
    apply_migration_old = MigrationExecutor.apply_migration
    self.apply_migration_old = apply_migration_old
    apply_migration_args = []
    self.apply_migration_args = apply_migration_args

    def apply_migration_wrapper(self, state, migration, fake=False, fake_initial=False):
      global connection
      connection = self.connection
      apply_migration_args.append([state, migration])
      return apply_migration_old(self, state, migration, fake, fake_initial)
    MigrationExecutor.apply_migration = apply_migration_wrapper

  def stop_migration_interception(self):
    """
    revert patching done in start_migration_interception
    """
    MigrationExecutor.apply_migration = self.apply_migration_old

  def introspection(self):
    """
    Utility to provide type information for introspection
    """
    introspection: BaseDatabaseIntrospection = connection.introspection
    return introspection

  def schema_editor(self):
    """
    Utility to provide typing for schema editor
    """
    schema_editor: BaseDatabaseSchemaEditor = connection.schema_editor()
    return schema_editor

  def run_migration(self, app_name: str = None, migration: str = None, **options):
    """
    Execute a migration. The migration state is saved as self.project_state
    so that it is accessible for inspection.
    """

    args = []
    if app_name:
      args.append(app_name)
    if migration:
      args.append(migration)

    call_command('migrate', *args, **options)

    self.project_state: ProjectState = self.apply_migration_args[-1][0]
    self.connection = connection

  def clear_database(self, project_state: ProjectState):
    """
    Remove all tables from the database so that migrations can be re-applied.
    Since there can be interdependencies between tables, the order in which the
    tables are deleted is important. However, instead if figuring out the dependency graph
    we blindly try to delete tables until eventually they all get deleted.
    """
    models = []
    for _, app_models in project_state.apps.all_models.items():
      for _, model in app_models.items():
        models.append(model)

    schema_editor = self.schema_editor()

    while len(models) > 0:
      remaining_models = []
      for model in models:
        try:
          with schema_editor:
            CLEANUP_LOG.debug('deleting %s', model)
            schema_editor.delete_model(model)
        except Exception as e:
          CLEANUP_LOG.debug('delete failed: ' + str(e))
          remaining_models.append(model)

      if len(remaining_models) == len(models) and len(models) > 0:
        CLEANUP_LOG.debug('clear_database stalled')
        CLEANUP_LOG.debug('          models = %s', ','.join([str(m) for m in models]))
        break
      else:
        models = remaining_models

    LOG.debug('dropping remaining tables')
    table_names = self.introspection().table_names()
    if connection.vendor == 'mysql':
    # because of constraints, some tables need to be dropped together on mysql:
      table_name_list = ', '.join(self.introspection().table_names())
      table_names = [table_name_list]
    if len(table_names) > 0:
      for table_name in table_names:
        LOG.debug('dropping tables %s', table_names)
        connection.cursor().execute(self.schema_editor().sql_delete_table %
                                      {'table': table_name})
    self.print_tables()

  def print_tables(self):
    tables = self.introspection().table_names()
    LOG.info('TABLES (count=%d)', len(tables))
    for t in tables:
      LOG.info('  table: %s', t)

  def assertTableExists(self, table_name: str, exists: bool):
    table_names = self.introspection().table_names()
    if exists:
      self.assertIn(table_name, table_names)
    else:
      self.assertNotIn(table_name, table_names)