# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.test import TestCase, override_settings, modify_settings, Client
from django.contrib.auth.models import User

from time import sleep

from reports.models import AppGroup
from workspace.models import Workspace
import logging

trusted_user_name='trustedUser'
trusted_user_password='password'

@override_settings(
 TRUSTED_AUTH = {
    'trusted_ips': ['127.0.0.1'],
    'trusted_users': [trusted_user_name],
    'timeout': 1,
    'single_use': True,
    'session_expiry': 30,
    'allow_superuser': False,
  }
)
@modify_settings(
  INSTALLED_APPS = {
    'append': 'trustedauth'
  },
  AUTHENTICATION_BACKENDS = {
    'append': 'trustedauth.backends.TrustedAuthBackend'
  }
)
class TrustedAuthTest(TestCase):
  @classmethod
  def setUpTestData(cls) -> None:
    trusted_user = User.objects.create_user(username=trusted_user_name, password=trusted_user_password)
    trusted_user_workspace = Workspace.objects.create(
      name='private',
      private_user=trusted_user,
      editable=False
    )
    cls.trusted_user_app_id = AppGroup.objects.create(
      group_name='trusted app',
      group_type='app',
      workspace=trusted_user_workspace,
      create_user=trusted_user_name,
    ).id

    other_username = 'other'
    other_user = User.objects.create_user(username=other_username, password=other_username)
    other_user_workspace = Workspace.objects.create(
      name='private',
      private_user=other_user,
      editable=False
    )
    cls.other_user_app_id = AppGroup.objects.create(
      group_name='other app',
      group_type='app',
      create_user=other_username,
      workspace=other_user_workspace
    ).id

  def test_when_invalid_token_and_valid_app_should_return_403(self):
    client = Client()

    invalid_token = "0mCy6uVjTS66Q_8XzXadPw"

    with self.assertLogs('arcweb.trustedauth', level='DEBUG') as log:
      app_response = client.get(f'/trustedauth/trusted/{invalid_token}/appgroup/{self.trusted_user_app_id}')
      self.assertEqual(app_response.status_code, 403)
      self.assertIn('Entering _setupTrust with ticket: 0mCy6uVjTS66Q_8XzXadPw and report_id:', log.output[0])
      self.assertIn('Auth ticket is invalid or username not found in authinfo', log.output[-1])

  def test_when_timed_out_valid_token_and_valid_app_should_return_403(self):
    client = Client()

    token_response = client.post(
      '/trustedauth/getticket',
      f'username={trusted_user_name}&trustedusername={trusted_user_name}&trustedpassword={trusted_user_password}',
      'application/x-www-form-urlencoded'
    )

    token = token_response.content.decode('utf-8').strip()

    sleep(1)

    with self.assertLogs('arcweb.trustedauth', level='DEBUG') as log:
      app_response = client.get(f'/trustedauth/trusted/{token}/appgroup/{self.trusted_user_app_id}')
      self.assertEqual(app_response.status_code, 403)
      self.assertIn('Entering _setupTrust with ticket:', log.output[0])
      self.assertIn('Auth ticket is invalid or username not found in authinfo', log.output[-1])

  def test_when_valid_token_and_valid_app_but_no_permission_should_return_500_and_display_error_message(self):
    client = Client()

    token_response = client.post(
      '/trustedauth/getticket',
      f'username={trusted_user_name}&trustedusername={trusted_user_name}&trustedpassword={trusted_user_password}',
      'application/x-www-form-urlencoded'
    )

    token = token_response.content.decode('utf-8').strip()

    with self.assertLogs('arcweb.trustedauth', level='DEBUG') as log:
      app_response = client.get(f'/trustedauth/trusted/{token}/appgroup/{self.other_user_app_id}')
      self.assertEqual(app_response.status_code, 500)
      self.assertIn('User trustedUser successfully authenticated and session set.', log.output[-1])
      self.assertIn('You do not have the necessary permission on this workspace', app_response.content.decode('utf-8'))

  @override_settings(
  TRUSTED_AUTH = {
      'trusted_ips': ['127.0.0.1'],
      'trusted_users': [trusted_user_name],
      'timeout': 2,
      'single_use': False,
      'session_expiry': 30,
      'allow_superuser': False,
    }
  )
  def test_when_valid_token_and_valid_app_and_single_use_false_should_return_200_until_timeout(self):
    client = Client()

    token_response = client.post(
      '/trustedauth/getticket',
      f'username={trusted_user_name}&trustedusername={trusted_user_name}&trustedpassword={trusted_user_password}',
      'application/x-www-form-urlencoded'
    )
    token = token_response.content.decode('utf-8').strip()

    with self.assertLogs('arcweb.trustedauth', level='DEBUG') as log:
      app_response_1 = client.get(f'/trustedauth/trusted/{token}/app/{self.trusted_user_app_id}')
      self.assertEqual(app_response_1.status_code, 200)

      app_response_2 = client.get(f'/trustedauth/trusted/{token}/app/{self.trusted_user_app_id}')
      self.assertEqual(app_response_2.status_code, 200)

      sleep(2)

      app_response_3 = client.get(f'/trustedauth/trusted/{token}/app/{self.trusted_user_app_id}')
      self.assertEqual(app_response_3.status_code, 403)
      self.assertIn('Entering _setupTrust with ticket:', log.output[0])
      self.assertIn('Auth ticket is invalid or username not found in authinfo', log.output[-1])

  # Test to Check Trusted IP Handling
  def test_when_trusted_ip_and_valid_token_should_return_200(self):
    client = Client(REMOTE_ADDR='127.0.0.1')
    token_response = client.post(
      '/trustedauth/getticket',
      f'username={trusted_user_name}&trustedusername={trusted_user_name}&trustedpassword={trusted_user_password}',
      'application/x-www-form-urlencoded'
    )
    token = token_response.content.decode('utf-8').strip()
    with self.assertLogs('arcweb.trustedauth', level='DEBUG') as log:
      app_response = client.get(f'/trustedauth/trusted/{token}/app/{self.trusted_user_app_id}')
      self.assertEqual(app_response.status_code, 200)
      self.assertIn('User trustedUser successfully authenticated and session set', log.output[-1])

  # Test to Check Non-Trusted IP Handling
  def test_when_non_trusted_ip_should_return_403(self):
    client = Client(REMOTE_ADDR='192.168.0.1')

    with self.assertLogs('arcweb.trustedauth', level='DEBUG') as log:
      token_response = client.post(
        '/trustedauth/getticket',
        f'username={trusted_user_name}&trustedusername=non_trusted_user&trustedpassword={trusted_user_password}',
        'application/x-www-form-urlencoded'
      )
      self.assertEqual(token_response.status_code, 403)
      self.assertIn('Authentication failed for IP', log.output[-1])
