# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import redirect
from django.contrib.auth.models import User
from django.contrib.auth import authenticate, login, logout
from django.views.decorators.csrf import csrf_exempt
from django.middleware.csrf import get_token
from django.conf import settings
from django.core.cache import cache
from uuid import uuid4
from base64 import urlsafe_b64encode
from time import sleep
from rbac.models import RBAC

import logging

# import the real "app" to use after Ticketed authentication
from apps.views import app, appgroup

# import the real "dataapi" to use after Ticketed authentication
from apps.views_dataapi import dataapi

LOGGER = logging.getLogger("arcweb.trustedauth")

@csrf_exempt
def getticket(request, error_msg=None):
    """ Generate a ticket """
    LOGGER.debug("Entering getticket with request: %s and error_msg: %s", request, error_msg)
    try:
        ta_settings = getattr(settings, 'TRUSTED_AUTH', None)
        if ta_settings is None:
            LOGGER.error("TRUSTED_AUTH settings are not configured.")
            return HttpResponse("TRUSTED_AUTH settings are not configured.\n")
        timeout = ta_settings.get('timeout', 30)
        trusted_users = ta_settings.get('trusted_users', [])
        trusted_ips = ta_settings.get('trusted_ips', [])
        allow_superuser = ta_settings.get('allow_superuser', False)
    except Exception as e:
        LOGGER.exception("Error accessing TRUSTED_AUTH settings: %s", e)
        return HttpResponse("-1\n")

    authenticated = False
    # Check IP
    remote_ip = request.META.get('HTTP_X_REAL_IP', None) or request.environ['REMOTE_ADDR']
    LOGGER.debug("Checking remote IP: %s", remote_ip)
    if remote_ip in trusted_ips:
        authenticated = True
        LOGGER.debug("IP %s is trusted.", remote_ip)

    if not authenticated:
        # Check for trusted user
        trustedusername = request.POST.get('trustedusername', None)
        trustedpassword = request.POST.get('trustedpassword', None)
        LOGGER.debug("Checking trusted user: %s", trustedusername)
        if trustedusername in trusted_users:
            trusteduser = authenticate(username=trustedusername, password=trustedpassword)
            if trusteduser and trusteduser.is_active:
                authenticated = True
                LOGGER.debug("User %s is authenticated.", trustedusername)

    if not authenticated:
        LOGGER.warning("Authentication failed for IP: %s and trustedusername: %s", remote_ip, trustedusername)
        return HttpResponseForbidden("-1\n")

    username = request.POST.get('username', None)
    LOGGER.debug("Generating ticket for username: %s", username)
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        LOGGER.error("User %s does not exist.", username)
        return HttpResponse("-1\n")
    except Exception as e:
        LOGGER.exception("Error retrieving user %s: %s", username, e)
        return HttpResponse("-1\n")

    if user is not None:
        if not allow_superuser and (user.is_superuser or 'sys_misc' in RBAC.get_user_obj_perms(user)):
            LOGGER.warning("Superuser or sys_misc permissions not allowed for user: %s", username)
            return HttpResponse("-1\n")
        if user.is_active:
            ticket = urlsafe_b64encode(uuid4().bytes).rstrip(b"=").decode()
            key = "authticket_%s" % ticket
            value = {"username": username}
            cache.set(key, value, timeout)
            LOGGER.debug("Ticket generated for user %s: %s", username, ticket)
            return HttpResponse("%s\n" % ticket, content_type='text/plain')
    
    LOGGER.warning("User %s is not active or authentication failed.", username)
    return HttpResponse("-1\n")

def _setupTrust(request, ticket='', report_id=''):
    """ Allow login via Ticket and load app  """
    LOGGER.debug("Entering _setupTrust with ticket: %s and report_id: %s", ticket, report_id)
    try:
        ta_settings = getattr(settings, 'TRUSTED_AUTH', None)
        if ta_settings is None:
            LOGGER.error("TRUSTED_AUTH settings are not configured.")
            return HttpResponse("TRUSTED_AUTH settings are not configured.\n")
        session_expiry = ta_settings.get('session_expiry', 0)
        allow_superuser = ta_settings.get('allow_superuser', False)
        single_use = ta_settings.get('single_use', True)
    except Exception as e:
        LOGGER.exception("Error accessing TRUSTED_AUTH settings: %s", e)
        return HttpResponse("-1\n")

    key = "authticket_%s" % ticket
    authinfo = None
    max_attempts = 5
    retry_delay = 0.05  # 50ms
    attempt = 0
    
    invalid_authinfo = True

    while attempt < max_attempts and invalid_authinfo:
        authinfo = cache.get(key)
        invalid_authinfo = not authinfo or "username" not in authinfo
    
        if invalid_authinfo:
            sleep(retry_delay)

        attempt += 1
    
    if single_use:
        cache.delete(key)
    if invalid_authinfo:
        LOGGER.warning("Auth ticket is invalid or username not found in authinfo: %s", authinfo)
        return HttpResponseForbidden("Forbidden\n")

    username = authinfo['username']
    LOGGER.debug("Authenticating user: %s", username)
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        LOGGER.error("User %s does not exist.", username)
        return HttpResponseForbidden("Forbidden.\n")
    except Exception as e:
        LOGGER.exception("Error retrieving user %s: %s", username, e)
        return HttpResponseForbidden("Forbidden.\n")

    if user is not None:
        user.backend = 'trustedauth.backends.TrustedAuthBackend'
        if user.username != request.user.username:
            LOGGER.debug("Logging out current user %s", request.user.username)
            logout(request)
        if not allow_superuser and user.is_superuser:
            LOGGER.warning("Superuser %s login attempt is not allowed.", username)
            return HttpResponse("-1\n")
        if user.is_active:
            if not request.user.is_authenticated:
                LOGGER.debug("Logging in user: %s", username)
                login(request, user)
                # Note: cache the csrf token and reuse for simultaneous unauthenticated requests
                key = "csrftoken_%s" % user.username
                csrftoken = cache.get(key)
                if not csrftoken:
                    newtoken = get_token(request)
                    cache.add(key, newtoken, 30) # use add() instead of set() to ensure a single csrf token
                    sleep(1) # allow cache to settle
                    csrftoken = cache.get(key) # get the token from the cache. may not be "newtoken" if other process set cache first.
                request.META.update({
                    "CSRF_COOKIE_USED": True,
                    "CSRF_COOKIE": csrftoken,
                })
            request.session['trustedauth'] = True
            request.session.set_expiry(session_expiry)
            LOGGER.debug("User %s successfully authenticated and session set.", username)
            # success
            return None

    LOGGER.warning("User %s authentication failed.", username)
    return HttpResponseForbidden("Forbidden.\n")

def trustedapp(request, ticket='', report_id=''):
  errResponse = _setupTrust(request, ticket=ticket)
  if errResponse:
    return errResponse

  # call apps.app
  return app(request, report_id=report_id)


def trustedappgroup(request, ticket='', appgroup_id=''):
  errResponse = _setupTrust(request, ticket=ticket)
  if errResponse:
    return errResponse

  # call apps.appgroup
  return appgroup(request, appgroup_id=appgroup_id)


def trustedapi(request, ticket=''):
  errResponse = _setupTrust(request, ticket=ticket)
  if errResponse:
    return errResponse

  # call sqlrun.dataapi
  return dataapi(request)

def trustedredirect(request, ticket=''):
    LOGGER.debug("Entering trustedredirect with ticket: %s", ticket)
    errResponse = _setupTrust(request, ticket=ticket)
    if errResponse:
        LOGGER.error("Error in _setupTrust: %s", errResponse)
        return errResponse
    next_url = request.GET.get('next', '')
    querystring = request.GET.copy()
    querystring.pop('next', None)
    if next_url and next_url.startswith("/arc"):
        LOGGER.debug("Redirecting to %s with querystring %s", next_url, querystring.urlencode())
        return redirect('%s?%s' % (next_url, querystring.urlencode()))
    LOGGER.warning("Forbidden access attempt with next_url: %s", next_url)
    return HttpResponseForbidden('Forbidden')

