# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.conf import settings
import threading
import time
import traceback
import logging

from .models import EventsTracker
from util.ff_flipper import FeatureFlipper
from util.querylog_loader import QueryLogLoader
from util.querylog_summarizer import QueryLogSummarizer
from .events_upload import are_env_vars_set

LOGGER = logging.getLogger('arcweb')
user_sessions = []

def eventstracker_initial_load():
  if not FeatureFlipper().get_flipper_value('enable_cml_usage_events'):
    return

  if not are_env_vars_set():
    LOGGER.info('EventsTracker is disabled because missing environment variable(s)')
    return

  def collect_data_for_eventstracker():
    LOGGER.info('start')
    try:
      while True:
        event_tracker_runner()
        time.sleep(settings.EVENTSTRACKER_BACKEND_DATA_TO_USAGE_DB)
    except Exception as e:
      LOGGER.error('Stop sending usage information: ' + ' '.join(traceback.format_exception_only(e.__class__,e)))

  thread = threading.Thread(target=collect_data_for_eventstracker, name='EventsTrackerThread')
  thread.daemon = True
  thread.start()

def log_to_database(data):
  # create or update usage_eventstracker row with data collected in the backend
  for summary_data in data:
    try:
      # Attempt to fetch the existing row
      existing_row = EventsTracker.objects.get(time=summary_data['time'])

      # Do not update the non_visual_page and non_visual_page_load_sec 
      # they are coming from the frontend
      for field, value in summary_data.items():
        if field in ["non_visual_page", "non_visual_page_load_sec", "errors"]:
          continue
        setattr(existing_row, field, value)

      if "errors" in summary_data:
        existing_errors = existing_row.errors or {}  # Ensure existing errors are not None
        new_errors = summary_data["errors"]

        # Update only 'query' and 'python', leaving 'visual' and 'js' unchanged
        existing_errors["query"] = new_errors.get("query", existing_errors.get("query", 0))
        existing_errors["python"] = new_errors.get("python", existing_errors.get("python", 0))

        setattr(existing_row, "errors", existing_errors)

      existing_row.save()

    except EventsTracker.DoesNotExist:
      # If no row exists, create a new one with all the provided data
      EventsTracker.objects.create(**summary_data)

    except Exception as e:
      LOGGER.error(f"Error while updating or creating row for time: {summary_data['time']}: {e}", exc_info=True)


# the user_session_loader gets called every single time
# if a user loging out (signals)
def user_session_loader(user_id, session_duration):
  user_sessions.append((str(user_id), session_duration.total_seconds()))


def user_session_cleanup():
  global user_sessions
  user_sessions = []

def event_tracker_runner():
  global user_sessions

  loader = QueryLogLoader()

  # On first load - collect the query logs until the beginning of the day
  if not EventsTracker.objects.exists():
    logs_data = loader.load_logs(time_frame='initial')
  else:
    # Otherwise, collect the query logs for the last hour
    logs_data = loader.load_logs()

  summarizer = QueryLogSummarizer()
  result = summarizer.summarize(logs_data, user_sessions)

  if result:
    log_to_database(result)
    user_session_cleanup()