# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.conf import settings
import logging
import json
import threading
import time
from util.ff_flipper import FeatureFlipper
import os
from datetime import timedelta
from usage.models import EventsTracker
from django.utils import timezone
from django.core.serializers.json import DjangoJSONEncoder
import requests
import arcwebbase.middleware_storage as ms

LOGGER = logging.getLogger("arcweb")

app_id = None
def get_application_id():
  hostnames = set()

  while True:
    hostname = ms.HOSTNAME

    if hostname is not None:
      if hostname not in hostnames:
        LOGGER.info(' new hostname=%s',hostname)
        hostnames.add(hostname)

        if '.' in hostname:
          subdomain = hostname.split('.')[0]
          LOGGER.info('  calculated subdomain=%s', subdomain)

          app_id = get_project_info(subdomain)

          LOGGER.info('app_id=%s', app_id)

          if app_id is not None:
            return app_id

    time.sleep(settings.HOSTNAME_POLL_INTERVAL)

def get_project_info(subdomain):
  apps = get_v1_applications()

  for app in apps:
    if app.get('subdomain') == subdomain:
      return app['id']

  return None

def are_env_vars_set():
  required_env_vars = ['CDSW_DOMAIN', 'CDSW_API_KEY', 'PROJECT_OWNER', 'CDSW_PROJECT']
  missing_env_vars = []

  for env_var in required_env_vars:
    if not os.environ.get(env_var):
      missing_env_vars.append(env_var)

  if missing_env_vars:
    LOGGER.info("Missing environment variables: %s", ', '.join(missing_env_vars))
    FeatureFlipper().set_flipper('enable_cml_usage_events', False, 'SETTINGS')
    return False
  else:
    return True

def get_v1_applications():
  if not are_env_vars_set():
    return []

  cdsw_domain = os.environ.get('CDSW_DOMAIN')
  apiv1_key = os.environ.get('CDSW_API_KEY')
  project_owner = os.environ.get('PROJECT_OWNER')
  project_slug = os.environ.get('CDSW_PROJECT')

  url = f"https://{cdsw_domain}/api/v1/projects/{project_owner}/{project_slug}/applications"
  res = req_v1(url, apiv1_key)
  return res

def req_v1(url, apiv1_key):
  res = requests.get(
    url,
    headers={"Content-Type": "application/json"},
    auth=(apiv1_key, ""),
    verify=False
  )
  check_response(res)
  return res.json()

def collect_and_send_usage_events():
  if not FeatureFlipper().get_flipper_value('enable_cml_usage_events'):
    return
  def send_events_to_cml_api(events_data):
    api_key = os.environ.get('CDSW_API_KEY')
    api_url = os.environ.get('CDSW_API_URL')
    headers = {
      'Content-Type': 'application/json'
    }
    response = requests.post(
      api_url + '/metricsreport',
      headers=headers,
      auth=(api_key, ""),
      data=events_data,
      verify=False
    )
    check_response(response)

  def get_events_for_last_24_hours_safe():
    try:
      get_events_for_last_24_hours()
    except Exception as e:
      LOGGER.error(str(e))

  def get_events_for_last_24_hours():
    global app_id

    if app_id is None:
      app_id = get_application_id()
    
    while True:
      engine_id = os.environ.get('CDSW_ENGINE_ID')
      current_datetime = timezone.now()
      start_of_previous_day = current_datetime - timedelta(days=1)
      events_for_last_24_hours = EventsTracker.objects.filter(
        timestamp__range=(start_of_previous_day, current_datetime))

      events_details = []
      for event in events_for_last_24_hours:
        events_detail = {
          'version': event.version,
          'timestamp': event.timestamp,
          'time': event.time,
          'domain': event.domain,
          'connectionTypes': event.connectionTypes,
          'visualTypes': event.visualTypes,
          'datasetTypes': event.datasetTypes,
          'numberUsers': event.numberUsers,
          'userTotalTime': event.userTotalTime,
          'non_visual_page': event.non_visual_page,
          'non_visual_page_load_sec': event.non_visual_page_load_sec,
          'errors': event.errors,
          'query_load_sec': event.query_load_sec,
          'features': event.features
        }

        events_details.append(events_detail)

      events_data = {
        "engineId": engine_id,
        "applicationId": app_id,
        "eventType": "CDVEvent",
        "eventDetails": json.dumps(events_details, cls=DjangoJSONEncoder)
      }

      events_data_json = json.dumps(events_data, cls=DjangoJSONEncoder)
      send_events_to_cml_api(events_data_json)

      time.sleep(settings.EVENTSTRACKER_DATA_TO_CML)

  thread = threading.Thread(target=get_events_for_last_24_hours_safe, name='EventsTrackerSendingThread')
  thread.daemon = True
  thread.start()

def check_response(res: requests.Response):
  if not res.status_code // 100 == 2:
    LOGGER.info(res.json())
    LOGGER.error('Error with API endpoint, stopping events logging')
    FeatureFlipper().set_flipper('enable_cml_usage_events', False, 'SETTINGS')
