# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.db import models, transaction
from collections import defaultdict
from datetime import datetime
from typing import Dict, TypedDict
import logging

LOGGER = logging.getLogger('arcweb')

class LogEntry(models.Model):
  new_entries = []

  @classmethod
  @transaction.atomic
  def commit_changes(cls):
    if len(cls.new_entries) > 0:
      cls.objects.bulk_create(cls.new_entries)
      cls.new_entries = []

  action_id = models.UUIDField()
  ts = models.DateTimeField()
  action = models.TextField(null=True)
  query_state = models.TextField(null=True)
  username = models.TextField(null=True)
  user_id = models.IntegerField(null=True)
  ip = models.TextField(null=True)
  component_type = models.TextField(null=True)
  component_id = models.IntegerField(null=True)
  visual_id = models.IntegerField(null=True)
  visual_type = models.TextField(null=True)
  visual_name = models.TextField(null=True)
  app_id = models.IntegerField(null=True)
  dashboard_name = models.TextField(null=True)
  appgroup_id = models.IntegerField(null=True)
  appgroup_name = models.TextField(null=True)
  dataset_id = models.IntegerField(null=True)
  dataset_name = models.TextField(null=True)
  dataconnection_id = models.IntegerField(null=True)
  dataconnection_type = models.TextField(null=True)
  dataconnection_fbid = models.TextField(null=True)
  dataconnection_name = models.TextField(null=True)
  sync = models.BooleanField(null=True)
  cache = models.BooleanField(null=True)
  cache_check = models.BooleanField(null=True)
  runtime = models.FloatField(null=True)
  query = models.TextField(null=True)
  rewrite = models.BooleanField(null=True)
  error = models.BooleanField(null=True)
  error_detail = models.TextField(null=True)
  process_id = models.IntegerField(null=True)

class Timestamp(TypedDict):
  Hour: int
  Day: int
  Month: int
  Year: int

class UsageDefaults(TypedDict, total=False):
  version: int
  time: Timestamp
  domain: str
  connectionTypes: Dict[str, int]
  visualTypes: Dict[str, int]
  datasetTypes: Dict[str, int]
  numberUsers: int
  userTotalTime: Dict[str, int]
  non_visual_page: Dict[str, int]
  non_visual_page_load_sec: Dict[str, int]
  errors: Dict[str, int]
  query_load_sec: Dict[str, int]
  features: Dict[str, int]

class EventsTracker(models.Model):
  version = models.PositiveIntegerField()
  timestamp = models.DateTimeField(null=True, auto_now_add=True)
  time = models.JSONField()
  domain = models.TextField()
  connectionTypes = models.JSONField()
  visualTypes = models.JSONField()
  datasetTypes = models.JSONField()
  numberUsers = models.PositiveIntegerField()
  userTotalTime = models.JSONField()
  non_visual_page = models.JSONField()
  non_visual_page_load_sec = models.JSONField()
  errors = models.JSONField()
  query_load_sec = models.JSONField()
  features = models.JSONField()

  @classmethod
  def update_events_tracker(
    cls,
    urls_counts,
    load_time_buckets,
    jsErrors,
    visualErrors,
  ):
    current_datetime = datetime.now()
    timestamp = {
      'Hour': current_datetime.hour,
      'Day': current_datetime.day,
      'Month': current_datetime.month,
      'Year': current_datetime.year,
    }

    # Try to fetch the existing row
    try:
      existing_row = cls.objects.get(time=timestamp)
      # Update only the necessary fields
      existing_row.non_visual_page = urls_counts
      existing_row.non_visual_page_load_sec = load_time_buckets
      existing_row.errors['js'] = int(jsErrors)
      existing_row.errors['visual'] = int(visualErrors)
      existing_row.errors = existing_row.errors
      existing_row.save()
    except cls.DoesNotExist:
      # If no row exists, create a new one with the following defaults
      usage_data: UsageDefaults = {
        'version': 1,
        'time': timestamp,
        'domain': '',
        'connectionTypes': defaultdict(int),
        'visualTypes': defaultdict(int),
        'datasetTypes': {},
        'numberUsers': 0,
        'userTotalTime': defaultdict(int),
        'non_visual_page': urls_counts,
        'non_visual_page_load_sec': load_time_buckets,
        'query_load_sec': defaultdict(int),
        'errors': {'visual': int(visualErrors), 'query': 0, 'python': 0, 'js': int(jsErrors)},
        'features': {},
      }

      cls.objects.create(**usage_data)

    except Exception as e:
      LOGGER.error(f"Error while updating or creating row for time: {timestamp}: {e}", exc_info=True)
