# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import django.test

from datetime import datetime

from usage.models import EventsTracker, UsageDefaults

from usage.events_tracker import log_to_database

import json

def create_record():
  return EventsTracker(
    version=1,
    time={},
    connectionTypes={},
    visualTypes={},
    datasetTypes={},
    numberUsers=0,
    userTotalTime={},
    non_visual_page={},
    non_visual_page_load_sec={},
    errors={},
    query_load_sec={},
    features={}
    )

# Backend update
class TestLogToDatabase(django.test.TestCase):
  databases = ['default', 'usage']

  def test_update_existing_row(self):

    self.assertEqual(EventsTracker.objects.count(), 0, 'no record')

    existing_row = create_record()
    existing_row.time = {"Hour": 2, "Day": 24, "Month": 1, "Year": 2025}
    # The backend thread should NOT update / or PARTIALLY update those fields
    existing_row.non_visual_page = {"apps_data": 0}
    existing_row.non_visual_page_load_sec = {"0": 1, "1": 0, "5": 2}
    existing_row.errors = {"visual": 1, "query": 0, "python": 0, "js": 2}  # Should partially update

    existing_row.save()
    self.assertEqual(EventsTracker.objects.count(), 1, '1 record')

    summary_data: UsageDefaults = {
      "version": 1,
      "timestamp": '2025-01-24 10:02:07.125020',
      "time": {"Hour": 2, "Day": 24, "Month": 1, "Year": 2025},
      "domain": 'test-domain',
      "connectionTypes": {"sqlite": {"1": 20, "12": 10}},
      "visualTypes": {"dashboard": 1, "crosstab": 1, "table": 4},
      "datasetTypes": {"numberOfTables": {"1": 2}, "sql": 0, "table": 2},
      "numberUsers": 1,
      "userTotalTime": {},
      "non_visual_page": {"apps_data": 1},
      "non_visual_page_load_sec": {"0": 0, "1": 1, "5": 0, "10": 0, "30": 0, "60": 0, "120": 0, "121": 0},
      "errors": {"visual": 0, "query": 1, "python": 1, "js": 0},
      "query_load_sec": {"1": 0, "5": 0, "10": 0, "30": 0, "60": 0, "120": 0},
      "features": {
        "featureFlippers": {
          "disable_visuals_for_workflows": False,
          "enable_ai_assistant": True,
          "enable_clone_info_badge": False,
        },
        "completions": {
          "azure_ai": {"count": 1},
          "open_ai": {"count": 1, "stream": 1},
          "caii": {"count": 1}
        },
        "embeddings": {
          "azure_ai": {"count": 1},
          "open_ai": {"count": 1},
          "caii": {"count": 1}
        }
      }
    }

    log_to_database([summary_data])

    # we update the existing_row data in the log_to_database
    # although the existing_row doesn't get updated automatically here
    # so we need to 'force' the refresh
    existing_row.refresh_from_db()

    self.assertEqual(EventsTracker.objects.count(), 1, '1 record')

    # assertEqual that non_visual_page & non_visual_page_load_sec did NOT change
    self.assertEqual(
      existing_row.non_visual_page,
      {"apps_data": 0},
      "non_visual_page should NOT be updated"
    )
    self.assertEqual(
      existing_row.non_visual_page_load_sec,
      {"0": 1, "1": 0, "5": 2},
      "non_visual_page_load_sec should NOT be updated"
    )

    # assertEqual that only "query" & "python" inside errors updated
    expected_errors = {"visual": 1, "query": 1, "python": 1, "js": 2}
    self.assertEqual(
      existing_row.errors,
      expected_errors,
      f"errors field was incorrectly updated: {existing_row.errors}"
    )

  def test_create_new_row(self):

    self.assertEqual(EventsTracker.objects.count(), 0, 'no record')

    summary_data: UsageDefaults = {
      "version": 1,
      "timestamp": '2025-01-24 10:02:07.125020',
      "time": {"Hour": 2, "Day": 24, "Month": 1, "Year": 2025},
      "domain": 'test-domain',
      "connectionTypes": {"sqlite": {"1": 20, "12": 10}},
      "visualTypes": {"dashboard": 1, "crosstab": 1, "table": 4},
      "datasetTypes": {"numberOfTables": {"1": 2}, "sql": 0, "table": 2},
      "numberUsers": 1,
      "userTotalTime": {},
      "non_visual_page": {"apps_data": 1},
      "non_visual_page_load_sec": {"0": 0, "1": 1, "5": 0, "10": 0, "30": 0, "60": 0, "120": 0, "121": 0},
      "errors": {"visual": 0, "query": 1, "python": 1, "js": 0},
      "query_load_sec": {"1": 0, "5": 0, "10": 0, "30": 0, "60": 0, "120": 0},
      "features": {
        "featureFlippers": {
          "disable_visuals_for_workflows": False,
          "enable_ai_assistant": True,
          "enable_clone_info_badge": False,
        },
        "completions": {
          "azure_ai": {"count": 1},
          "open_ai": {"count": 1, "stream": 1},
          "caii": {"count": 1}
        },
        "embeddings": {
          "azure_ai": {"count": 1},
          "open_ai": {"count": 1},
          "caii": {"count": 1}
        }
      }
    }

    log_to_database([summary_data])

    self.assertEqual(EventsTracker.objects.count(), 1, '1 record')

    existing_row = EventsTracker.objects.get(time=summary_data["time"])
    existing_row.refresh_from_db()

    self.assertEqual(existing_row.version, summary_data["version"])
    self.assertEqual(existing_row.time, summary_data["time"])
    self.assertEqual(existing_row.domain, summary_data["domain"])
    self.assertEqual(existing_row.connectionTypes, summary_data["connectionTypes"])
    self.assertEqual(existing_row.visualTypes, summary_data["visualTypes"])
    self.assertEqual(existing_row.datasetTypes, summary_data["datasetTypes"])
    self.assertEqual(existing_row.numberUsers, summary_data["numberUsers"])
    self.assertEqual(existing_row.userTotalTime, summary_data["userTotalTime"])
    self.assertEqual(existing_row.non_visual_page, summary_data["non_visual_page"])
    self.assertEqual(existing_row.non_visual_page_load_sec, summary_data["non_visual_page_load_sec"])
    self.assertEqual(existing_row.errors, summary_data["errors"])
    self.assertEqual(existing_row.query_load_sec, summary_data["query_load_sec"])
    self.assertEqual(existing_row.features, summary_data["features"])


# FrontEnd update
class TestUpdateEventsTracker(django.test.TestCase):
  databases = ['default', 'usage']

  def test_update_existing_row(self):
    """Test if save_frontend_events_tracker_data processes and calls update_events_tracker correctly."""
    self.assertEqual(EventsTracker.objects.count(), 0, 'no record')
    current_datetime = datetime.now()
    existing_row = create_record()

    # FrontEnd thread should NOT update THESE fields
    existing_row.version = '1'
    existing_row.time = {
      'Hour': current_datetime.hour,
      'Day': current_datetime.day,
      'Month': current_datetime.month,
      'Year': current_datetime.year,
    }
    existing_row.domain = "test-domain"
    existing_row.connectionTypes = {"sqlite": {"1": 20, "12": 10}}
    existing_row.visualTypes = {"dashboard": 1, "crosstab": 1, "table": 4}
    existing_row.datasetTypes = {"numberOfTables": {"1": 2}, "sql": 0, "table": 2}
    existing_row.numberUsers = 1
    existing_row.userTotalTime = {}
    existing_row.query_load_sec = {"1": 0, "5": 0, "10": 0, "30": 0, "60": 0, "120": 0}
    existing_row.errors = {"visual": 0, "query": 3, "python": 5, "js": 0}  # Should partially update
    existing_row.features = {
        "featureFlippers": {
          "disable_visuals_for_workflows": False,
          "enable_ai_assistant": True,
          "enable_clone_info_badge": False,
        },
        "completions": {
          "azure_ai": {"count": 1},
          "open_ai": {"count": 1, "stream": 1},
          "caii": {"count": 1}
        },
        "embeddings": {
          "azure_ai": {"count": 1},
          "open_ai": {"count": 1},
          "caii": {"count": 1}
        }
      }

    existing_row.save()
    self.assertEqual(EventsTracker.objects.count(), 1, '1 record')

    data = [
      {"url": "apps_activity", "loadTime": 0.5},
      {"url": "apps_datasql*sql-tab", "loadTime": 2},
      {"url": "apps_home", "loadTime": 6},
      {"url": "apps_activity", "loadTime": 11}
    ]
    post_data = {
      'data': json.dumps(data),
      'jsErrors': '2',
      'visualErrors': '1'
    }

    response = self.client.post('/usage/save_frontend_events_tracker_data', post_data)

    self.assertEqual(response.status_code, 200)
    self.assertJSONEqual(response.content, {"message": "Load time data updated successfully."})

    existing_row.refresh_from_db()

    # assertEqual that non_visual_page & non_visual_page_load_sec did NOT change
    self.assertEqual(
      existing_row.non_visual_page,
      {"apps_activity": 2, "apps_datasql*sql-tab": 1, "apps_home": 1},
      "non_visual_page should be updated"
    )
    self.assertEqual(
      existing_row.non_visual_page_load_sec,
      {'0': 1, '1': 1, '10': 1, '120': 0, '121': 0, '30': 0, '5': 1, '60': 0},
      "non_visual_page_load_sec should be updated"
    )

    # assertEqual that only "visual" & "js" inside errors updated
    expected_errors = {"visual": 1, "query": 3, "python": 5, "js": 2}
    self.assertEqual(
      existing_row.errors,
      expected_errors,
      f"errors field was incorrectly updated: {existing_row.errors}"
    )

  def test_create_new_row(self):

    self.assertEqual(EventsTracker.objects.count(), 0, 'no record')
    current_datetime = datetime.now()
    data = [
      {"url": "apps_activity", "loadTime": 0.5},
      {"url": "apps_datasql*sql-tab", "loadTime": 2},
      {"url": "apps_home", "loadTime": 6},
      {"url": "apps_activity", "loadTime": 11}
    ]
    post_data = {
      'data': json.dumps(data),
      'jsErrors': '2',
      'visualErrors': '1'
    }

    response = self.client.post('/usage/save_frontend_events_tracker_data', post_data)

    self.assertEqual(response.status_code, 200)
    self.assertJSONEqual(response.content, {"message": "Load time data updated successfully."})

    self.assertEqual(EventsTracker.objects.count(), 1, '1 record')

    existing_row = EventsTracker.objects.get(time={
      'Hour': current_datetime.hour,
      'Day': current_datetime.day,
      'Month': current_datetime.month,
      'Year': current_datetime.year,
    })
    existing_row.refresh_from_db()

    # assertEqual that non_visual_page & non_visual_page_load_sec did NOT change
    self.assertEqual(
      existing_row.non_visual_page,
      {"apps_activity": 2, "apps_datasql*sql-tab": 1, "apps_home": 1},
      "non_visual_page should be updated"
    )
    self.assertEqual(
      existing_row.non_visual_page_load_sec,
      {'0': 1, '1': 1, '10': 1, '120': 0, '121': 0, '30': 0, '5': 1, '60': 0},
      "non_visual_page_load_sec should be updated"
    )

    # assertEqual that only "visual" & "js" inside errors updated
    expected_errors = {"visual": 1, "query": 0, "python": 0, "js": 2}
    self.assertEqual(
      existing_row.errors,
      expected_errors,
      f"errors field was incorrectly updated: {existing_row.errors}"
    )

