# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import logging
from math import floor
from django.utils import timezone

import vizlib.querylog as querylog
from usage.models import LogEntry
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from datasets.models import DataConnection
from django.conf import settings
from rbac.permcache import RBAC
from reports.views_settings import setting_get_single
from util.ff_flipper import FeatureFlipper
from migration.views import import_file
from apps.views_static_model_helper import get_static_dashboard_path
from .views_helper import to_usage_logentry_model
from apps.config import product_name

LOGGER = logging.getLogger('arcweb')

USAGE_NAME = 'Cloudera Data Visualization Usage'

def load_initial_usage_data():
  if not FeatureFlipper().get_flipper_value('enable_usage_analytics'):
    return

  already_loaded = LogEntry.objects.all().count() > 0
  if already_loaded:
    return

  enable_usage_analytics(None, True)


def enable_usage_analytics(request=None, enable=False):
  if request:
    RBAC.checkPerm(request.user, 'sys_viewlogs')

  if enable:
    create_usage_dataconnection()
    load_usage_data_from_logs()


def create_usage_dataconnection():
  if DataConnection.objects.filter(dataconnection_name=USAGE_NAME).exists():
    return

  connection = DataConnection(
    dataconnection_type='sqlite',
    dataconnection_name=USAGE_NAME,
    dataconnection_description=f'Automatically created and populated database with {product_name} usage logs',
    dataconnection_info=json.dumps({'PARAMS': {'FILE': settings.DATABASES['usage']['NAME'], 'CACHE': { 'ENABLED': 0, 'RETENTION': 86400}}}),
    dataconnection_settings=json.dumps({ 'permissions': ['sys_viewlogs'] }),
  )
  connection.save()


@login_required
def delete_usage_database(request):
  if request:
    RBAC.checkPerm(request.user, 'sys_viewlogs')

  LogEntry.objects.all().delete()

  return HttpResponse(status=200)


@login_required
def import_usage_dashboard(request):
  RBAC.checkPerm(request.user, 'sys_viewlogs')

  create_usage_dataconnection()

  return import_file(request, dc_name=USAGE_NAME, filename=get_static_dashboard_path("cloudera_data_visualization_usage"),
                     workspace_id=None, sanity_check='false', dry_run='false', limit_search_scope='false', skip_thumbnails='true')


def handle_continuous_usage_data():
  if FeatureFlipper().get_flipper_value('enable_usage_analytics'):
    LogEntry.commit_changes()
    delete_expired_usage_logs()
  else:
    LogEntry.new_entries = []


def load_usage_data_from_logs():
  try:
    LogEntry.objects.all().delete()

    max_days= get_max_expiration_days()
    max_batch_size = 100000
    logfileIndex = None
    while True:
      logReader = querylog.LogReader(fileindex=logfileIndex)
      if not logReader.file_exists():
        break

      if logfileIndex is None:
        logfileIndex = 1
      else:
        logfileIndex += 1

      for logEntryDict in logReader.read_json_backwards():
        logEntry = to_usage_logentry_model(logEntryDict)

        if is_expired_log(logEntry, max_days):
          break

        LogEntry.new_entries.append(logEntry)

        if len(LogEntry.new_entries) >= max_batch_size:
          LogEntry.commit_changes()

    LogEntry.commit_changes()
  except:
    LOGGER.exception('cannot load usage data')


def get_max_expiration_days():
  try:
    value = setting_get_single('usage_analytics_expiration_days')
    return float(value)
  except:
    return 0


def is_expired_log(logEntry: LogEntry, max_days=0):
  if not max_days:
    return False

  today = timezone.now()
  return (today - logEntry.ts).days > max_days


def delete_expired_usage_logs():
  max_days = get_max_expiration_days()
  expiration_seconds = floor(max_days * 86400)  # 86400 seconds in a day
  expiration_date = timezone.now() - timezone.timedelta(seconds=expiration_seconds)
  LogEntry.objects.filter(ts__lte=expiration_date).delete()
