#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Test the admin api calls
"""

import os
import sys
import json
import copy
import codecs
import base64
import optparse
import itertools
import requests
from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

# Keep a copy of all the objects for use in testing
DB = {}

def connect(prefix, username, password, apikey):

  s = requests.Session()
  # Disable https certificate verification
  requests.packages.urllib3.disable_warnings()
  s.verify = False

  """
    On the move to python 3, we're getting random intermittent errors of the form:

    urllib3.exceptions.ProtocolError: ('Connection aborted.', RemoteDisconnected('Remote end closed connection without response',))

    The discussion here looks relevant:

      https://github.com/requests/requests/issues/3458

    There is a solution mentioned here:

      https://stackoverflow.com/questions/15431044/can-i-set-max-retries-for-requests-request

    For now, using the connect error option to hopefully reconnect as needed.

    I leave this in not sure if/how correct it is.
  """

  retries = Retry(connect=3)

  s.mount('http://', HTTPAdapter(max_retries=retries))

  if apikey:
    s.headers.update({'AUTHORIZATION':'apikey %s' % apikey})
  else:
    # Get login page with csrf cookie
    url = prefix + "/arc/apps/login"
    r = s.get(url)
    s.headers['referer'] = r.url
    # Login and get sessionid
    payload = {'csrfmiddlewaretoken': s.cookies["arccsrftoken"],
               'username': username,
               'password': password}
    s.post(url, data=payload)

  return s

def fatal(text):
  """ Fatal error """
  raise Exception(text)
  sys.exit(-1)

def checkrsp(rsp):
  if rsp.status_code != 200:
    fatal("Error %s : %s" % (rsp.status_code, rsp.content[:100]))

def parse():

  parser = optparse.OptionParser()
  parser.add_option('--url', dest="url", default='http://127.0.0.1:7999/arc/adminapi/v1',
    help='URL to server')
  parser.add_option('--apikey', dest='apikey', default='',
    help='Connect with apikey rather than username/password')
  parser.add_option('--username', dest='username', default='admin',
    help='Username to connect with')
  parser.add_option('--password', dest='password', default='admin',
    help='Password to connect with')
  parser.add_option('--keep', dest='keep', action='store_true', default=False,
    help='Keep the created test objects')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if len(args) > 0:
    fatal('Extra options: %s' % str(args))

  login = options.url.replace('arc/adminapi', '')
  options.session = connect(login, options.username, options.password, options.apikey)

  return options

# Test a sequence of ops on an object:
# CREATE/UPDATE/GET/DELETE
def test_ops(options, otype, obj, update, check):

  url = options.url + '/' + otype;
  ses = options.session

  key = 'title' if otype == 'visuals' else 'username' if otype == 'users' else 'name'
  newobj = copy.deepcopy(obj)

  # Make a copy to create a new object
  del newobj['id']
  if key == 'username':
    newobj[key] = 'apitest_user_from_%s' % obj['id']
    newobj['password'] = 'old_password'
  else:
    newobj[key] = 'Admin API test %s copied from ID %s' % (otype, obj['id'])

  rsp = ses.post(url, data={'data':json.dumps([newobj])})
  checkrsp(rsp)

  newobj = rsp.json()[0]
  check(obj, newobj)

  print("Created %s %s -- %s" % (otype[0:-1], newobj['id'], newobj[key]))

  # Ensure update with 0 or 2 objects fails
  rsp = ses.post(url, data={'data':json.dumps([])})
  assert(rsp.status_code != 200)
  rsp = ses.post(url, data={'data':json.dumps([newobj, newobj])})
  assert(rsp.status_code != 200)

  # Update a field and ensure it works
  update(newobj)
  url += '/%s' % newobj['id']
  rsp = ses.post(url, data={'data':json.dumps([newobj])})
  checkrsp(rsp)
  endobj = rsp.json()[0]
  check(obj, newobj, endobj)

  if not options.keep:
    # Delete the object
    rsp = ses.delete(url)
    checkrsp(rsp)

    # Make sure it's gone
    rsp = ses.get(url)
    assert(rsp.status_code != 200)

def test_gets(options):

  url = options.url
  ses = options.session
  for t in ['visuals', 'datasets', 'connections', 'users', 'groups', 'roles',
            'segments', 'filterassociations', 'workspaces']:

    # Make sure summary and detail return the same number of objects
    rsp = ses.get(url + '/' + t)
    checkrsp(rsp)
    count = len(rsp.json())
    print("Got %s %s" % (count, t))
    assert(count > 0)

    rsp = ses.get(url + '/' + t + '?detail=1')
    objects = rsp.json()
    # Getting detail on visuals includes additional report types
    if t != 'visuals':
      assert(count == len(objects))
    DB[t] = objects
    # DB[t + 'byID'] = dict([(obj['id', obj) for obj in objects])

    # Make sure we can get a single object
    # Workspace name/desc swizzling for private workspace confuses things so
    # use the Public workspace for testing
    id = objects[-1]['id']
    if t == 'workspaces':
      id = objects[0]['id']
    urlone = url + ('/%s/%s' % (t, id))
    rsp = ses.get(urlone)
    checkrsp(rsp)
    objone = rsp.json()
    assert(len(objone) == 1)
    assert(objone[0]['id'] == id)

    # Make sure we can get a single object with details
    rsp = ses.get(urlone,  params={'detail':1})
    checkrsp(rsp)
    details = rsp.json()
    assert(len(details) == 1)
    assert(details[0]['id'] == id)
    assert(len(list(details[0].keys())) >= len(list(objone[0].keys())))

    # Test get by name (not supported in visuals)
    # Note also that not all objects have unique names as this test assumes
    if t not in ['visuals']:
      namekey = 'name'
      if t == 'users':
        namekey = 'username'
      nameval = details[0][namekey]
      rsp = ses.get(url + '/' + t + ('?name=%s' % nameval))
      checkrsp(rsp)
      objone = rsp.json()
      assert(len(objone) == 1)
      assert(objone[0]['id'] == id)

      # Ensure name not found gives error
      rsp = ses.get(url + '/' + t + ('?name=%s' % ('Not there - ' + nameval)))
      assert(rsp.status_code != 200)

    # Test get with filter
    filter = {'id':id}
    fencode = base64.b64encode(bytes(json.dumps(filter), 'ascii'))
    rsp = ses.get(url + '/' + t + ('?filter=%s' % fencode.decode('ascii')))
    checkrsp(rsp)
    objone = rsp.json()
    assert(len(objone) == 1)
    assert(objone[0]['id'] == id)

    # Ensure get empty response on no matches
    filter = {'id':id + 123456}
    fencode = base64.b64encode(bytes(json.dumps(filter), 'ascii'))
    rsp = ses.get(url + '/' + t + ('?filter=%s' % fencode.decode('ascii')))
    checkrsp(rsp)
    objnone = rsp.json()
    assert(len(objnone) == 0)

def test_visuals(options):

  tab = None
  vizzes = DB['visuals']
  for i in range(len(vizzes), 0, -1):
    v = vizzes[i - 1]
    if v.get('type') == 'table':
      tab = v
      break

  if tab is None:
    fatal("Did not find table viz")
  rsp = options.session.get(options.url + '/visuals/%s?detail=1' % tab['id'])
  checkrsp(rsp)
  tab = rsp.json()[0]

  def check(obj, newobj, endobj=None):
    if endobj is not None:
      assert(newobj['title'] == endobj['title'])
      assert(endobj['data']['report_data']['limit'] == 12345)
      assert(endobj['data']['report_data']['report_title'] == endobj['title'])

  # Build the update object request
  def update(obj):
    id = obj['id']
    data = obj['data']
    data['report_data']['limit'] = 12345
    obj.clear()
    obj['id'] = id
    obj['title'] = 'Update viz %s' % id
    obj['data'] = data

  test_ops(options, 'visuals', tab, update, check)

def test_datasets(options):

  ds = None
  dslist = DB['datasets']
  for i in range(len(dslist), 0, -1):
    d = dslist[i - 1]
    if d.get('info'):
      ds = d
      break

  if ds is None:
    fatal("Did not find dataset with populated column info")

  def check(obj, newobj, endobj=None):
    if endobj is not None:
      assert(newobj['name'] == endobj['name'])
      assert(newobj['dc_id'] == endobj['dc_id'])
      assert(endobj['info'][0]['tablename'] == 'New Tablename')

  # Build the update object request
  def update(obj):
    id = obj['id']
    dcid = obj['dc_id']
    newdc = next((dc for dc in DB['connections'] if dc['id'] != dcid), None)
    if newdc is not None:
      dcid = newdc['id']
    info = obj['info']
    info[0]['tablename'] = 'New Tablename'

    obj.clear()
    obj.update({'id':id,
                'name':'Updated dataset %s' % id,
                'dc_id':dcid,
                'info':info})

  test_ops(options, 'datasets', ds, update, check)

def test_connections(options):

  def check(obj, newobj, endobj=None):
    if endobj is not None:
      assert(newobj['name'] == endobj['name'])
      assert(endobj['password'] == None)
      assert(endobj['info']['USERNAME'] == 'New USERNAME')

  # Build the update object request
  def update(obj):
    id = obj['id']
    info = obj['info']
    info['USERNAME'] = 'New USERNAME'

    obj.clear()
    obj.update({'id':id,
                'name':'Updated connection %s' % id,
                'password':'New Password',
                'info':info})

  test_ops(options, 'connections', DB['connections'][0], update, check)

def test_users(options):

  def check(obj, newobj, endobj=None):
    if endobj is not None:
      assert(newobj['username'] == endobj['username'])
      assert(endobj['is_superuser'] == endobj['is_superuser'])
      assert(set(g['id'] for g in DB['groups'][0:3]) == set(g['id'] for g in endobj['groups']))
      assert(set(r['id'] for r in DB['roles'] [0:2]) == set(r['id'] for r in endobj['roles']))

  # Build the update object request
  def update(obj):
    id = obj['id']
    obj.clear()
    obj['id'] = id
    obj['username'] = 'updated_user_%s' % id
    obj['is_superuser'] = True
    obj['password'] = 'new_password'
    obj['old_password'] = 'old_password'
    obj['groups'] = [{'id':g['id'], 'name':g['name']} for g in DB['groups'][0:3]]
    obj['roles'] =  [{'id':r['id'], 'name':r['name']} for r in DB['roles'][0:2]]

  test_ops(options, 'users', DB['users'][0], update, check)

def test_groups(options):

  def check(obj, newobj, endobj=None):
    if endobj is not None:
      assert(newobj['name'] == endobj['name'])
      assert(set(u['id'] for u in DB['users'][0:3]) == set(u['id'] for u in endobj['users']))
      assert(set(r['id'] for r in DB['roles'][0:2]) == set(r['id'] for r in endobj['roles']))

  # Build the update object request
  def update(obj):
    id = obj['id']
    obj.clear()
    obj['id'] = id
    obj['name'] = 'updated_group_%s' % id
    obj['users'] = [{'id':u['id'], 'name':u['username']} for u in DB['users'][0:3]]
    obj['roles'] = [{'id':r['id'], 'name':r['name']} for r in DB['roles'][0:2]]

  test_ops(options, 'groups', DB['groups'][0], update, check)

def test_roles(options):

  def check(obj, newobj, endobj=None):
    if endobj is not None:
      assert(newobj['name'] == endobj['name'])
      assert(set(newobj['users']) == set(endobj['users']))
      assert(set(newobj['groups']) == set(endobj['groups']))
      assert(len(newobj['privs']) == len(endobj['privs']))
      for (n, e) in zip(newobj['privs'], endobj['privs']):
        for f in ['ptype', 'dcid']:
          assert(n.get(f) == e.get(f))
        for l in ['perms', 'dclist', 'dslist']:
          assert(set(n.get(l, [])) == set(e.get(l, [])))

  # Build the update object request
  def update(obj):
    id = obj['id']
    obj.clear()
    obj['id'] = id
    obj['name'] = 'updated_role_%s' % id
    obj['users'] = ['user_1', 'user_2']
    obj['groups'] = ['group_1', 'group_2', 'group_3']
    obj['privs'] = [{
        "perms": [
          "sys_styles",
          "sys_viewlogs"
        ],
        "ptype": "system"
      },
      {
        "perms": [
          "dc_aviews",
          "dc_upload",
          "dc_explore"
        ],
        "ptype": "dataconn",
        "dclist": [
          "-1"
        ]
      }]

  test_ops(options, 'roles', DB['roles'][0], update, check)

def test_segments(options):

  def check(obj, newobj, endobj=None):
    if endobj is not None:
      assert(newobj['name'] == endobj['name'])
      assert(newobj['dataset_id'] == endobj['dataset_id'])
      assert(newobj['data']['entities'] == [])
      assert(newobj['data']['group'] == "A group name")
      assert(newobj['data']['filters'] == 'a in ("name", "another name")')

  # Build the update object request
  def update(obj):
    id = obj['id']
    dataset_id = obj['dataset_id']
    obj.clear()
    obj['id'] = id
    obj['dataset_id'] = dataset_id
    obj['name'] = 'updated_segment_%s' % id
    obj['data'] = {
      'entities': [],
      'group': "A group name",
      'filters': 'a in ("name", "another name")',
    }

  test_ops(options, 'segments', DB['segments'][0], update, check)

def test_filterassocations(options):

  fa = DB['filterassociations'][0]
  segs = []
  for seg in DB['segments']:
    if seg['dataset_id'] == fa['dataset_id']:
      segs.append(seg)

  def check(obj, newobj, endobj=None):
    if endobj is not None:
      assert(newobj['name'] == endobj['name'])
      assert(newobj['dataset_id'] == endobj['dataset_id'])
      assert(set(newobj['groups']) == set([g['id'] for g in DB['groups'][0:2]]))
      assert(set(newobj['users']) == set([u['id'] for u in DB['users'][0:3]]))
      applied = newobj['data'].get('applied_segments')
      negate = True
      for app, seg in zip(applied, segs):
        assert(app['negate'] == negate)
        assert(app['id'] == seg['id'])
        assert(app['name'] == seg['name'])
        negate = not negate

  # Build the update object request
  def update(obj):
    id = obj['id']
    dataset_id = obj['dataset_id']
    obj.clear()
    obj['id'] = id
    obj['dataset_id'] = dataset_id
    obj['name'] = 'updated_fa_%s' % id
    obj['groups'] = [g['id'] for g in DB['groups'][0:2]]
    obj['users'] = [u['id'] for u in DB['users'][0:3]]
    applied = []
    negate = True
    for seg in segs:
      applied.append({'negate':negate, 'id':seg['id'], 'name':seg['name']})
      negate = not negate
    obj['data'] = {'applied_segments': applied}

  test_ops(options, 'filterassociations', fa, update, check)

def test_workspaces(options):

  ws = DB['workspaces'][0]

  def check(obj, newobj, endobj=None):
    if endobj is not None:
      assert(newobj['name'] == endobj['name'])
      assert(newobj['desc'] == endobj['desc'])
      assert(newobj['acl'] == endobj['acl'])

  # Build the update object request
  def update(obj):
    id = obj['id']
    obj.clear()
    obj['id'] = id
    obj['name'] = 'updated_ws_%s' % id
    obj['desc'] = 'New descrption for ws %s' % id
    obj['acl'] = [[1, 1, 'admin'], [2, 2, 'Everyone'], [2, 3, 'another group']]

  test_ops(options, 'workspaces', ws, update, check)

def main():
  options = parse()
  test_gets(options)
  test_visuals(options)
  test_datasets(options)
  test_connections(options)
  test_users(options)
  test_groups(options)
  test_roles(options)
  test_segments(options)
  test_filterassocations(options)
  test_workspaces(options)

if __name__ == "__main__":
  main()

