# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import asyncio
import json
from configparser import ConfigParser
from util.ai import bedrock_transformers as t

# logging: ADDITIONAL_LOGGERS.extend([('botocore.endpoint', 'DEBUG')])

def get_session(credentials: str):
  import botocore.credentials
  import aiobotocore.session

  session = aiobotocore.session.get_session()
  if not credentials:
    # Read from file
    return session

  config = ConfigParser()
  config.read_string(credentials)
  aws_profiles = {}

  for key in config.sections():
    aws_profiles[key] = config[key]
  session._config = {'profiles': aws_profiles}

  # workaround for AssumeRoleProvider trying to read from config files. For some reason this provider
  # does not use the config provided above, but tries to parse config files again.
  credential_providers = session._components.get_component('credential_provider').providers
  assume_role_provider = [p for p in credential_providers if isinstance(p,botocore.credentials.AssumeRoleProvider)][0]
  if assume_role_provider:
      assume_role_provider._profile_provider_builder = None

  return session

async def bedrock_comm(url: str, headers: dict, upstream_req_body: dict):
  session = get_session(headers.get('credentials', ''))
  async with session.create_client(service_name='bedrock-runtime') as brt:
    body = json.dumps(upstream_req_body)
    model_id = headers.get('model_id', '')
    stream_enabled = headers.get('stream', False)

    if stream_enabled:
      get_chuck_text_fn = headers.get('custom_get_chunk_text') or default_get_chunk_text_fn(model_id)
      if not get_chuck_text_fn:
        raise Exception("Streaming is enabled, but no function to extract chunk text is available.")

      response = await brt.invoke_model_with_response_stream(
        modelId=model_id,
        body=body
      )
      http_status = response.get('ResponseMetadata', {}).get('HTTPStatusCode', 200)
      stream = response.get('body')

      async for event in stream:
        chunk = event.get('chunk', {})
        chunk_bytes = chunk.get('bytes', b'')
        if chunk_bytes:
          chunk_data = json.loads(chunk_bytes)
          text = get_chuck_text_fn(chunk_data) # Different models have different response structure
          if text.strip():
            # Construct an OpenAI-style chunk with the text content to ensure consistent chunk handling downstream.
            openai_chunk = {
              "choices": [
                {
                  "delta": {
                      "content": text
                  },
                  "index": 0,
                  "finish_reason": None
                }
              ]
            }
            yield http_status, f"data: {json.dumps(openai_chunk)}\n\n"

      # After the loop completes, send a final chunk with finish_reason="stop"
      stop_chunk = {
        "choices": [
          {
            "delta": {},
            "index": 0,
            "finish_reason": "stop"
          }
        ]
      }
      yield http_status, f"data: {json.dumps(stop_chunk)}\n\n"

      # Then send the [DONE] marker
      yield http_status, "data: [DONE]\n\n"

    else:
      # Non-streaming case: just return the final response body as JSON
      response = await brt.invoke_model(
        modelId=model_id,
        body=body
      )
      http_status = response.get('ResponseMetadata', {}).get('HTTPStatusCode', 200)
      response_body = json.loads(await response.get('body').read())
      yield http_status, response_body

def default_get_chunk_text_fn(model_id: str):
  model_id_lower = model_id.lower()
  if "titan" in model_id_lower:
      return titan_chunk
  elif "anthropic" in model_id_lower:
      return anthropic_chunk
  else:
      return None
def titan_chunk(chunk_data):
  return chunk_data.get('outputText', '')

def anthropic_chunk(chunk_data):
  return chunk_data.get('delta', {}).get('text', '')


async def go_completion(stream_mode: bool = True):
  ''' Models
    amazon.titan-text-lite-v1
    amazon.titan-text-express-v1
    ============================
    req = {
      'inputText': 'Tell me a story about Mars in 500 words'
    }

    Any anthropic model like anthropic.claude-3-5-sonnet-20241022-v2:0
    =========================================
    messages = [
      {"content": "Hello! How can I help you?", "role": "assistant"},
      {"content": "Tell me a story about Mars in 500 words.", "role": "user"},
    ]

    req = {
      "max_tokens": 1024,
      "anthropic_version": "bedrock-2023-05-31",
      "messages": messages,
    }
  '''

  messages = [
    {"content": "Hello! How can I help you?", "role": "assistant"},
    {"content": "Tell me a story about Mars in 500 words.", "role": "user"},
  ]

  req = {
    "max_tokens": 1024,
    "anthropic_version": "bedrock-2023-05-31",
    "messages": messages,
  }

  headers = {
    'model_id': 'anthropic.claude-3-5-sonnet-20241022-v2:0',
    'stream': stream_mode
  }

  if stream_mode:
    aggregated_text = ""
    from arcwebbase.ai_support import ChunkProcessor
    processor = ChunkProcessor()
    # Process each streaming chunk.
    async for status, response in bedrock_comm(
      url='',
      headers=headers,
      upstream_req_body=req
    ):
      response = processor.process_chunk(response)
      aggregated_text += ' '.join(response)

    # Use the aggregated text to build a final response structure.
    final_resp = t.completion_response_transformer({"results": [{"outputText": aggregated_text}]})
    print("Final aggregated streaming response:", final_resp)
  else:
    # Non-streaming: process the single response.
    async for status, response in bedrock_comm(
      url='',
      headers=headers,
      upstream_req_body=req
    ):
      final_resp = t.completion_response_transformer(response)
      print("Non-streaming response:", final_resp)
      break

async def go_embedding():
  req = {
    'inputText': 'How do I bake a pie?'
  }

  async for status, response in bedrock_comm(
    url='',
    headers={
      'model_id': 'amazon.titan-embed-text-v1'
    },
    upstream_req_body=req
  ):
    resp = t.embedding_response_transformer(response)
    print(resp)
    if 'data' in resp:
      for d in resp['data']:
        print('embedding length', len(d['embedding']))
        break

def main():
  asyncio.run(go_completion())
  asyncio.run(go_embedding())

if __name__ == '__main__':
  main()
