# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

"""
Utilities related to Amazon Bedrock models.

This file might be imported from Django settings,
so keep module-level imports to a minimum!

Under no condition import anything that requires an initialized
Django environment (like model access)!
"""

def embedding_response_transformer(body: dict):
  return {
    'data': [body]
  }

def completion_request_transformer(headers: dict, body: dict):
  model_id = headers.get("model_id", "").lower()

  if "anthropic" in model_id:
    # For Anthropic Claude models, the API expects:
    #  - "anthropic_version"
    #  - "messages": a list of message objects with roles "user" or "assistant"
    #  - Optionally, a top-level "system" field containing the system prompt.
    messages = body.get("inputText", [])
    anthropic_messages = []
    system_prompt = None

    for m in messages:
      # If there's a system message, capture its content separately.
      if m["role"] == "system":
        system_prompt = m["content"]
      else:
        anthropic_messages.append({
          "role": m["role"],
          "content": [{"type": "text", "text": m["content"]}]
        })

    new_body = {
      "max_tokens": 1024,
      "anthropic_version": "bedrock-2023-05-31",
      "messages": anthropic_messages
    }

    if system_prompt is not None:
      new_body["system"] = system_prompt
  else:
    # For Titan models: concatenate the messages into a prompt string.
    messages = body.get("inputText", [])
    prompt = ""
    for m in messages:
      content = m["content"]
      role = m["role"]
      role_str = "Bot" if role == "assistant" else "User"
      prompt += f"{role_str}: {content}\n"
    prompt += "Bot:"
    new_body = {"inputText": prompt}

  return headers, new_body


def completion_response_transformer(body: dict):
  """
  Transforms the response from the bedrock API into a standard structure.

  For Titan models, it expects a 'results' key with an 'outputText' field.
  For Anthropic/Claude models, it expects a 'content' key, which is a list of
  objects that include text segments.
  """
  response = 'NO RESPONSE'
  if 'results' in body:
    # Titan-style response
    results = body.get('results')
    if results and isinstance(results, list) and 'outputText' in results[0]:
      response = results[0]['outputText']
  elif 'content' in body:
    # Anthropic/Claude-style response: join all text segments in 'content'
    content_list = body.get('content')
    if content_list and isinstance(content_list, list):
      # Each content dict is expected to have keys 'type' and 'text'
      texts = [item.get('text', '') for item in content_list if item.get('type') == 'text']
      response = ''.join(texts) if texts else response

  return {
    'choices': [
      {
        'message': {
          'content': response
        }
      }
    ]
  }
