# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
This script analyzes and suggests acceleration for apps
"""

import os
import sys

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

import re
import json

import reports.models
import datasets.models

from . import ds

from django.core import serializers

class Apps(object):
  DB_RE = re.compile("^uri-(\d+)-widget-(\d+)$")

  def __init__(self):
    self.__apps = {}
    self.__multiapps = {}

  def __len__(self):
    return len(self.__apps) + len(self.__multiapps)

  def apps(self):
    return list(self.__apps.values())

  def add_dataset_apps(self, ds_id):
    apps = reports.models.Report.objects.filter(dataset__id=ds_id)
    for app in apps:
      self.add(app.id)

  def deserialize(self, app_data, ds_maps):
    created_apps = []
    apps_json = json.loads(app_data)
    apps = []
    multiapps = []
    old_appids = {}
    new_appids = {}
    for i, app_json in enumerate(apps_json):
      if app_json['fields']['report_type'] == 'dashboard':
        multiapps.append(app_json)
        old_appids[i] = app_json['pk']
      else:
        apps.append(app_json)
        old_appids[i] = app_json['pk']

      if not app_json['fields']['dataset'] in ds_maps:
        raise Exception('New dataset id not specified for app %s, ds %s' % \
            (app_json['pk'], app_json['fields']['dataset']))
      app_json['fields']['dataset'] = ds_maps[app_json['fields']['dataset']]
      app_json['pk'] = None

    modified_json = json.dumps(apps)

    for i, app in enumerate(serializers.deserialize("json", modified_json)):
      app.save()
      # Note that all apps must be serialized before all multiapps for this
      # to work. This is currently what 'serialize' below does
      new_appids[old_appids[i]] = app.object.id
      print('App %s created' % app.object.id)
      created_apps.append(app.object.id)

    for app_json in multiapps:
      d = json.loads(app_json['fields']['report_data'])
      new_click_behavior_maps = {}
      for report_k, report_v in d['click_behavior'].get('reports', {}).items():
        report_k = int(report_k)
        if report_k in new_appids:
          new_click_behavior_maps[new_appids[report_k]] = report_v
        else:
          print('not found:', report_k, type(report_k), ':', new_appids)
      d['click_behavior']['reports'] = new_click_behavior_maps

      if 'global_control_widgets' in d:
        cl = []
        for c in d['global_control_widgets']:
          c = int(c)
          new_id = new_appids[c]
          cl.append(new_id)
        d['global_control_widgets'] = cl

      if 'control_widgets' in d:
        cl = []
        for c in d['control_widgets']:
          c = int(c)
          new_id = new_appids[c]
          cl.append(new_id)
        d['control_widgets'] = cl

      for w in d['dashboard_widgets']:
        m = Apps.DB_RE.search(w['id'])
        new_id = 'uri-%s-widget-%s' % (m.group(1), new_appids[int(m.group(2))])
        w['id'] = new_id
      app_json['fields']['report_data'] = json.dumps(d)

    modified_json = json.dumps(multiapps)
    for i, app in enumerate(serializers.deserialize("json", modified_json)):
      app.save()
      new_appids[old_appids[len(apps) + i]] = app.object.id
      print('Multi App %s created' % app.object.id)
      created_apps.append(app.object.id)
    return new_appids

  def serialize(self):
    all_apps = []
    for x in list(self.__apps.values()):
      all_apps.append(x)
    for x in list(self.__multiapps.values()):
      all_apps.append(x)

    return serializers.serialize("json",
        [x.app() for x in all_apps], indent=2)

  def cb(self, fn):
    for app in self.apps():
      fn(app)

  def add(self, app_id):
    if app_id in self.__apps:
      return
    app = reports.models.Report.objects.get(id=app_id)
    j = json.loads(app.report_data)
    if not 'dashboard_widgets' in j:
      self.__apps[app_id] = App(app)
      return
    self.__multiapps[app_id] = App(app)
    matched_ids = [Apps.DB_RE.search(x['id']) for x in j['dashboard_widgets']]
    db_app_ids = [int(m.group(2)) for m in matched_ids]
    print("IDs: %s" % str(db_app_ids))

    # Filter down to the reports that exist
    applist = reports.models.Report.objects.filter(id__in=db_app_ids)
    for r in applist:
      self.add(r.id)

    if 'global_control_widgets' in j:
      applist = reports.models.Report.objects.filter(id__in=j['global_control_widgets'])
      for r in applist:
        self.add(r.id)

    if 'control_widgets' in j:
      applist = reports.models.Report.objects.filter(id__in=j['control_widgets'])
      for r in applist:
        self.add(r.id)

  def analyze(self):
    for app in list(self.__apps.values()):
      app.analyze()

  def datasets(self):
    datasets = ds.DataSets()
    for app in self.apps():
      dataset_id = app.dataset_id()
      dataset = datasets.add(dataset_id)
      for coltype, col in app.cols():
        dataset.track_col(coltype, col)
    return datasets

def coltype_to_str(coltype):
  if coltype == App.COL_DIMENSION:
    return "Dimension"
  return "Aggregate"

class App(object):
  COL_DIMENSION = 1
  COL_AGGREGATE = 2

  def __init__(self, app):
    self.__tracked_cols = []
    self.__app = app

  def app(self):
    return self.__app

  def dataset_id(self):
    return self.__app.dataset.id

  def app_id(self):
    return self.__app.id

  def title(self):
    return self.__title

  def analyze(self):
    j = json.loads(self.__app.report_data)

    if j['report_type'] == 'daterange':
      print("Skipping analyze of datarange picker id ", self.__app.id)
      self.__title = "daterange picker %d" % self.__app.id
      return

    self.__title = j['report_data'].get('report_title', '')
    print('title', self.__title)
    self.__track_cols(App.COL_DIMENSION,
        j['report_data'].get('dimensions_shelf', []))
    self.__track_cols(App.COL_AGGREGATE,
        j['report_data'].get('aggregates_shelf', []))
    self.__track_cols(App.COL_DIMENSION,
        j['report_data'].get('filters_shelf', []))

  def cols(self):
    return self.__tracked_cols

  def __track_col(self, coltype, col):
    self.__tracked_cols.append([coltype, col])

  def __track_cols(self, coltype, coll):
    for col in coll:
      if col.get('custom_expr'):
        self.__track_col(coltype, col['custom_expr'])
        continue
      if col.get('fn_name'):
        self.__track_col(coltype,
            "%s(%s)" % (col['fn_name'], col['dataset_colname']))
        continue
      if col.get('filter_column'):
        self.__track_col(coltype, col['filter_column'])
        continue
      if col.get('dataset_colname'):
        self.__track_col(coltype, col['dataset_colname'])

