# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
This script clones apps and points them at different datasets.

For example:
1. If you have a app 3 targeted at dataset 6, you can use:
<script> --app 3 --dataset 6:8 --change
to clone a new app and make it point at dataset 8 instead

2. If you have a app 12 which is a dashboard containing many individual widgets
   that point at distinct datasets, you can use:
<script> --app 12 --dataset 6:8 --dataset 7:12 --dataset 1:33 --change
to clone new apps for each app within the dashboard, using the old:new datasets
specified to make changes, and to finally create a new dashboard that points at
all of these new reports
"""

import itertools
import re
import sys
import json
import optparse
import os
import uuid

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

import reports.views
import reports.models
import datasets.models

def parse():
  parser = optparse.OptionParser()
  parser.add_option('--app',
    action="store", dest="app",
    help="report to clone", default=None)
  parser.add_option('--dataset',
    action="append", dest="dataset",
    help="oldid:newid for dataset modifications. You can specify this option multiple times",
    default=None)
  parser.add_option("--change", action="store_true", dest="change",
    help="Actually clone apps. Without this specified, this script " \
         "prints out some identification information for the apps " \
         "and datasets identified",
    default=False)
  return parser.parse_args()

def main():
  options, args = parse()

  dbs = []
  apps = []
  db_uris = []
  db_reportids = []
  db_datasetids = []
  dataset_old_to_new = {}
  dataset_id_to_dataset = {}

  if options.app:
    options.app = int(options.app)
    r = reports.models.Report.objects.get(id=options.app)
    j = json.loads(r.report_data)
    if 'dashboard_widgets' in j:
      dbs.append(r)
      print('Dashboard Title:', j['report_title'])
      db_re = re.compile("^uri-(\d+)-widget-(\d+)$")
      matched_ids = [db_re.search(x['id']) for x in j['dashboard_widgets']]
      db_uris = [int(m.group(1)) for m in matched_ids]
      db_reportids = [int(m.group(2)) for m in matched_ids]
      db_datasetids = []
      for reportid in db_reportids:
        r = reports.models.Report.objects.get(id=reportid)
        apps.append(r)
        db_datasetids.append(r.dataset.id)
      for uri, reportid, dataset in zip(db_uris,
          db_reportids, db_datasetids):
        print('Report Detail: ID %s Dataset %s' % (int(reportid), int(dataset)))
    else:
      apps.append(r)
      title = j['report_data']['report_title']
      print('Report Title:', title)
      print('Current Dataset: %s (ID %s)' % (r.dataset.dataset_name, r.dataset.id))

  if options.dataset:
    for dataset_id in options.dataset:
      old_id, sep, new_id = dataset_id.partition(":")
      old_id = int(old_id)
      new_id = int(new_id)
      dataset_old_to_new[old_id] = new_id

    for old_id, new_id in dataset_old_to_new.items():
      d = datasets.models.DataSet.objects.get(id=new_id)
      dataset_id_to_dataset[new_id] = d
      print('Reports in dataset %s will clone to dataset %s (%s)' % \
          (old_id, new_id, d.dataset_name))

  if not options.app or not options.dataset:
    sys.exit(1)

  if not options.change:
    sys.exit(1)

  # First, clone individual apps
  apps_old_to_new = {}
  for r in apps:
    old_report_id = r.id
    clone(r, dataset_id_to_dataset[dataset_old_to_new[r.dataset.id]])
    new_report_id = r.id
    apps_old_to_new[old_report_id] = new_report_id
    print('Report %s cloned to ID %s' % (old_report_id, new_report_id))

  # Next, clone dashboards
  for r in dbs:
    j = json.loads(r.report_data)
    for i in range(0, len(db_uris)):
      j['dashboard_widgets'][i]['id'] = 'uri-%d-widget-%d' % (db_uris[i], apps_old_to_new[db_reportids[i]])
    r.report_data = json.dumps(j)
    if r.dataset_id in dataset_old_to_new:
      r.dataset_id = dataset_old_to_new[r.dataset_id]
    old_report_id = r.id

    # Clone widgets, and save away new widgets within the dashboard
    r.report_data = reports.views.move_or_clone_dashboard(r.report_data)['dashboard_data']

    clone(r)
    new_report_id = r.id
    print('Dashboard %s cloned to ID %s' % (old_report_id, new_report_id))

# Copy some of reports.views.clone_report
def clone(r, d=None):
  r.pk = None
  r.id = None
  r.uuid = uuid.uuid4()
  r.imported_uuid = None
  # Don't clone shared visual properties
  r.shared_visual_dashboards = None
  if d:
    r.dataset = d
  r.save()

if __name__ == "__main__":
  main()

