#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Test program to load and display app contents

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

import os
import re
import sys
import json
import pprint
import getpass
import optparse
import traceback

import time
from datetime import datetime, timedelta

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)
sys.path.append("%s/../vizlib" % THISDIR)

from django.conf import settings

from vizlib.accel import viewrec_accel_info, App, Apps, Shelf

from django.db.models.base import ObjectDoesNotExist
from django.contrib.auth.models import User

from datasets.models import DataSet, DataConnection
from dataconn.dataconn import get_dataconn
from dataset import ModelDataset, getDatasetInfo
from vizlib.dataset_request import DatasetRequest

from reports.models import Report

def fatal(text):
  """ Fatal error """
  estr = traceback.format_exc()
  if not estr.startswith("None"):
    print(estr)
  if text:
    print(text)
  sys.exit(-1)

def parse():

  parser = optparse.OptionParser(usage="")

  parser.add_option('--appid', dest='appids', type='int', action='append',
    help='App ID')

  parser.add_option('--dsid', dest='dsids', type='int', action='append',
    help='Dataset ID')

  parser.add_option('--all', dest='allapps', action='store_true',
    help='Process all apps')

  parser.add_option('--verbose', dest='verbose', action='store_true',
    help='Set verbose output mode')

  parser.add_option('--makepw', dest='makepw',
    help='Enter password for encryption')

  parser.add_option('--log', dest='log', action='store_true',
    help='Show app execution info from log')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if len(args) > 0:
    fatal('Extra options: %s' % str(args))

  if options.makepw:
    from django.contrib.auth.hashers import make_password
    print("PW for %s : %s" % (options.makepw, make_password(options.makepw)))
    sys.exit(0)

  return options

def main():

  options = parse()

  if options.appids is None:
    options.appids = []

  if options.dsids:
    options.appids.extend(Report.objects.filter(dataset_id__in=options.dsids).values_list('id', flat=True))

  if options.allapps:
    options.appids.extend(Report.objects.all().values_list('id', flat=True))

  # Load the listed apps, expand the dashboards
  if not options.appids:
    fatal("No App IDs specified")

  apps = Apps()
  for appid in set(options.appids):
    apps.add(appid)

  for app in apps.apps():
    print("App %d name '%s' type '%s' dataset %d" % (app.id(), app.report_title(), app.report_type(), app.dbapp.dataset_id))
    if options.verbose:
      app.pprint()

    print("Shelves:")

    dsreq = {"version": 1,
             "type": 'SQL',
             "dimensions": [],
             "aggregates": [],
             "filters": [],
             "having": [],
             "global_segments": [],
             "limit": 100000,
             "dataset_id": None}

    for sh in app.shelves:
      print("  %s" % str(sh.desc))
      for i in range(0, len(sh.elements)):
        agg, txt = sh.getExpr(i, dsreq, expand_params=False, filter_dims=True)
        print("    %d (%s): %s" % (i, "agg" if agg else "dim", txt))
    print("dsreq:")
    app.pprinter.pprint(app.getDsreq())

  if options.log:
    idlist = [app.dbapp.id for app in apps.apps()]
    info = viewrec_accel_info(idlist)
    print("Getting log info for: %s" % str(idlist))
    pprint.pprint(info)

if __name__ == "__main__":
  main()

