#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

"""
  Import or export analytical views

  This tool requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

import fnmatch
import json
import optparse
import os
import os.path
import sys
import warnings
with warnings.catch_warnings():
  warnings.filterwarnings("ignore", category=DeprecationWarning)

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

from datasets.models import DataConnection
from django.test import RequestFactory
from django.contrib.auth.models import User
from sqlrun.views import jsonselect

SHOW_VIEW_DSREQ = {'version': 1, 'limit': 100000, 'type': '.show_aviews'}
SHOW_DATABASES_DSREQ = {'version': 1, 'limit': 100000, 'type': '.databases'}
SHOW_TABLES_DSREQ = {'version': 1, 'limit': 100000, 'type': '.tables'}

# column names of "show analytical views" command
AV_NAME = 'name'
BASE_TABLE = 'base'
QUERY = 'select'
PARTITIONS = 'partitions'


class jsonselect_runner(object):
  def __init__(self, user, dc_name=None, dc_id=None):
    self.user = User.objects.get(username=user)
    self.request = RequestFactory()
    if dc_name and dc_id:
      raise ValueError("either dc_name or dc_id parameter should be set")
    elif dc_name:
      self.dc_id = DataConnection.objects.get(dataconnection_name=dc_name).id
    elif dc_id:
      self.dc_id = dc_id
    else:
      raise ValueError("either dc_name or dc_id parameter should be set")

  def run(self, dsreq_query):
    """
    run jsonselect using either dsreq or query
    if a string or unicode is passed the request is created
    with the string as the query. If a dictionary is passed
    the request will be created with dictionary as dsreq
    """
    data = {'dataconnection_id':[str(self.dc_id)]}
    data['cache_read'] = "false"
    if isinstance(dsreq_query, str):
      data['query'] = dsreq_query
    elif isinstance(dsreq_query, dict):
      data['dsreq'] = json.dumps(dsreq_query)
    else:
      raise Exception("the input argument should be either string in the "
                      "case of query or dictionary in the case of dsreq")
    # create a post request
    post_req =  self.request.post('/arc/sqlrun/jsonselect', data=data)
    post_req.user = self.user
    post_req.session = {}
    response = jsonselect(post_req)
    return response

def pprint_table(rows):
  header = rows[0]
  col_index = list(range(len(header)))

  # get the max length of each column
  col_paddings = []
  total_length = 0
  for index in col_index:
    col_max_length = max([len(row[index]) for row in rows])
    col_paddings.append(col_max_length)
    total_length += col_max_length + 3
  total_length -= 1

  # taking the header out of rows
  rows = rows[1:] if len(rows) > 1 else []

  # printing the header
  print("+" + "-" * total_length + "+")
  print("|", end=' ')
  for index in col_index:
    print(header[index].ljust(col_paddings[index]), "|", end=' ')
  print("")
  print("+" + "-" * total_length + "+")

  # printing the data rows
  for row in rows:
    print("|", end=' ')
    for index in col_index:
      print(row[index].ljust(col_paddings[index]), "|", end=' ')
    print("")

  print("+" + "-" * total_length + "+")
  print("%s rows" % len(rows))

def extract_aviews(options):
  jrunner = jsonselect_runner(options.username, options.dc_name, options.dc_id)
  response = jrunner.run(SHOW_VIEW_DSREQ)
  response = json.loads(response.content)
  aviews = response["rows"]
  name_index = response["colnames"].index(AV_NAME)
  base_index = response["colnames"].index(BASE_TABLE)
  select_index = response["colnames"].index(QUERY)
  # handling arcengine prior to 2.2.6 version
  # where there is no partition information
  partition_index = response["colnames"].index(PARTITIONS) \
    if PARTITIONS in response["colnames"] else -1
  if options.list:
    print("\nList of Analytical Views:")
    rows = [["Analytical View Name", "Base Table"]]
    for av in aviews:
      if fnmatch.fnmatch(av[name_index], options.filter):
        rows.append([av[name_index], av[base_index]])
    pprint_table(rows)
  else:
    # creating a dictionary for the views
    # that need to be exported to a file
    data = {}
    for av in aviews:
      if fnmatch.fnmatch(av[name_index], options.filter):
        partitions = av[partition_index] if partition_index != -1 else ""
        data[av[name_index]] = {"base":av[base_index],
                                "select":av[select_index],
                                "partitions":partitions}

    with open(options.path, "w") as f:
      json.dump(data, f, indent=4)

    print("\nAnalytical views are exported into %s file" % options.path)

def import_aviews(options):
  with open(options.path, "r") as f:
    aviews = json.loads(f.read())
  status = [["Analytical View Name", "Status", "Error Message"]]
  jrunner = jsonselect_runner(options.username, options.dc_name, options.dc_id)
  for av_name in list(aviews.keys()):
    try:
      av_partition = aviews[av_name][PARTITIONS]
      if fnmatch.fnmatch(av_name, options.filter):
        query = "CREATE ANALYTICAL VIEW %s " % av_name
        if av_partition:
          query += "PARTITIONED BY (%s) " % av_partition
        query += "STORED AS PARQUET AS %s;" % aviews[av_name][QUERY]
        if options.verbose:
          print("-------------------------------------------------------------")
          print("creating %s" % av_name)
          print("-------------------------------------------------------------")
          print(query)
          print("-------------------------------------------------------------")
        response = jrunner.run(query)
        status.append([av_name, "Passed", ""])
    except Exception as e:
      if options.verbose:
        print(e)
      status.append([av_name, "Failed", str(e).strip()])
      if not options.continue_on_error:
        break

  print("\nAnalytical View Creation Report:")
  pprint_table(status)

def import_aviews_sanity(options):
  with open(options.path, "r") as f:
    aviews = json.loads(f.read())

  # get the list of databases
  jrunner = jsonselect_runner(options.username, options.dc_name, options.dc_id)
  databases = jrunner.run(SHOW_DATABASES_DSREQ)
  databases = json.loads(databases.content)["rows"]

  # get the table names in each database
  table_names = []
  for db in databases:
    db = db[0]
    dsreq = SHOW_TABLES_DSREQ
    dsreq["dbname"] = db
    tables = jrunner.run(dsreq)
    tables = json.loads(tables.content)["rows"]
    table_names += ["%s.%s" % (db, t[0]) for t in tables]

  # get the name of all analytical views
  response = jrunner.run(SHOW_VIEW_DSREQ)
  response = json.loads(response.content)
  name_index = response["colnames"].index(AV_NAME)
  all_av_names = [av[name_index] for av in response["rows"]]

  # perform sanity check for each analytical view that is supposed
  # to be imported. The sanity check is to test 1) analytical view
  # with the same name does not exist, and 2) the base table exists
  print("\nSanity Check Results (no analytical view has been created)")
  status = [["Analytical View Name", "Name exists",
             "Base Table Name Exists", "Sanity Check"]]
  for av_name in list(aviews.keys()):
    base_table = aviews[av_name][BASE_TABLE]
    if fnmatch.fnmatch(av_name, options.filter):
      av_name_exists = av_name in all_av_names
      base_table_exists = base_table in table_names
      status.append([av_name, str(av_name_exists), str(base_table_exists),
                     'Passed' if base_table_exists and not av_name_exists
                              else 'Failed'])
  pprint_table(status)

def parseargs():
  desc = """The utility for importing and exporting analytical views. This tool requires
  the settings path in the environment, something like:
  export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'"""
  epilog = "Note '--username' and '--dc_name' or '--dc_id' are required"

  parser = optparse.OptionParser(epilog=epilog, description=desc)

  parser.add_option('--list', dest='list', action="store_true",
    help='show analytical views')
  parser.add_option('--username', dest='username',
    help='username')
  parser.add_option('--dc_name', dest='dc_name',
    help='name of data connection')
  parser.add_option('--dc_id', dest='dc_id',
    help='id of data connection')
  parser.add_option('--filter', dest='filter',
    help='filter results based on wildcards. For example, '
         'use my_db.tv* to match all views that starts with "tv" '
         'in my_db database, or use my_db.* to match all views in '
         "my_db database. Default is '*' to match everything", default="*")
  parser.add_option('--export', dest='export', action="store_true",
    help='export analytical views to the provided file')
  parser.add_option('--import', dest='imp', action="store_true",
    help='import analytical views from the provided file. If '
         'dry_run option is provided it only performs '
         'sanity check on the requirements of creating analytical views ')
  parser.add_option('--path', dest='path',
    help='import or export analytical views to this file. '
         'It can be either relative or absolute path')
  parser.add_option('--dry_run', dest='sanity', action="store_true",
    help='performs a sanity check in the case of import')
  parser.add_option('--continue_on_error', dest='continue_on_error',
    help='do not stop the process of importing analytical views '
         'on an import failure', action="store_true")
  parser.add_option('--verbose', dest='verbose', action="store_true",
    help='in the case of import print the create analytical view statement '
         'and possible error message')

  options, args = parser.parse_args()

  exclusive_options_count = (options.list is not None) + \
                  (options.export is not None) + \
                  (options.imp is not None)

  required_options_count = (options.username is not None) + \
                           (options.dc_name is not None) + \
                           (options.dc_id is not None)

  if not options.username:
    parser.error("options --username is required")
  elif required_options_count < 2:
    parser.error("options --username and --dc_name or --dc_id are required")
  elif required_options_count > 2:
    parser.error("options --dc_name and --dc_id are mutually exclusive. Choose one")

  if exclusive_options_count > 1:
    parser.error("options --list, --export and --import are mutually exclusive. Choose one")
  elif exclusive_options_count < 1:
    parser.error("option --list, --export or --import is required. Choose one")

  if not options.list == 1 and not options.path:
    parser.error("option --path is required with --import or --export option")

  if options.path and not os.path.isabs(options.path):
    options.path = os.path.join(os.getcwd(), options.path)

  if options.imp and not os.path.exists(options.path):
    parser.error("No such file %s" % options.path)

  return options

def main():
  options = parseargs()

  if not 'DJANGO_SETTINGS_MODULE' in os.environ:
    print("error: Please set the 'DJANGO_SETTINGS_MODULE' environment variable")
    exit(1)

  # If the settings file has a relative path for the db we need to find it
  # relative to the settings module not this file
  dsmname=os.environ['DJANGO_SETTINGS_MODULE']
  print("Using: DJANGO_SETTINGS_MODULE=%s" % dsmname)
  dsm = __import__(dsmname, fromlist=[''])
  if dsm.DATABASES['default']['ENGINE'].find("sqlite") >= 0:
    dbname = dsm.DATABASES['default']['NAME']
    if os.path.isabs(dbname):
      print("Using: db = %s" % dbname)
      from django.conf import settings
    else:
      print(dsm.__file__)
      newdbname = os.path.normpath(os.path.join(os.path.dirname(dsm.__file__),
                                                "..",
                                                dsm.DATABASES['default']['NAME']))
      from django.conf import settings
      settings.DATABASES['default']['NAME'] = newdbname
      from django.db import connection
      connection.close()
      print("Using: db = %s" % newdbname)

  if options.list or options.export:
    extract_aviews(options)

  if options.imp and options.sanity:
    import_aviews_sanity(options)
  elif options.imp:
    import_aviews(options)

if __name__ == "__main__":
  main()
