# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json

CURRENCY_TYPE = 'currency'

def cleanup_fields(fields, display_format_name='display_format', colname='dataset_colname', aliasname='dataset_colname'):
  is_changed = False
  for field in fields:
    display_format = field.get(display_format_name)
    if display_format and not isinstance(display_format, str) and display_format.get('type') == CURRENCY_TYPE:
        result = display_format.get('result')
        currency_symbol = display_format.get(CURRENCY_TYPE)
        if result and currency_symbol:
            col_name = field.get(colname)
            col_alias = field.get(aliasname)
            col_alias = f'({col_alias})' if col_alias else ''
            print(f'Updating field {col_name} {col_alias}')
            is_changed = True
            field[display_format_name] = {
                'type': CURRENCY_TYPE,
                'currencysymbol': currency_symbol,
                'currencyformat': result,
                'result': result,
                'currencyoptions': 'custom',
            }
        elif not display_format.get('currencysymbol') and not display_format.get('currencyformat') and not display_format.get('currencyoptions'):
            print(f'Missing result and currency symbol for {json.dumps(field)}')

  return (fields, is_changed)

def cleanup_datasets():
  all_datasets = DataSet.objects.all()
  for dataset in all_datasets:
    is_changed = False
    try:
      dataset_info = json.loads(dataset.dataset_info)
    except:
      is_changed = True
      dataset.dataset_info = '[]'
      dataset_info = []
    for table in dataset_info:
      (fields, is_curr_changed) = cleanup_fields(table.get('columns'), 'displayformat', 'name', 'alias')
      is_changed = is_changed or is_curr_changed
      if is_curr_changed:
          print(f'Updating columns in table {table.get("tablename")}')
          table['columns'] = fields

    if is_changed:
      print(f'Saving dataset {dataset.id}')
      dataset.dataset_info = json.dumps(dataset_info)
      dataset.save(update_fields=['dataset_info'])

def cleanup_reports():
  all_reports = Report.objects.all()
  for report in all_reports:
    is_changed = False
    report_data_full = json.loads(report.report_data)
    report_data = report_data_full.get('report_data', {})
    shelves = report_data.get('core', {}).get('shelves', [])
    for shelf in shelves:
      shelf_name = shelf.get('name')
      (fields, is_curr_changed) = cleanup_fields(report_data.get(shelf_name, []))
      is_changed = is_changed or is_curr_changed
      if is_curr_changed:
        print(f'Updating fields in shelf {shelf_name}')
        report_data[shelf_name] = fields

    if is_changed:
      print(f'Saving report {report.id}')
      report_data_full['report_data'] = report_data
      report.report_data = json.dumps(report_data_full)
      report.save(update_fields=['report_data'])

def main(apps, schema_editor):
  if apps:
    global Report
    Report = apps.get_model('reports', 'report')
    global DataSet
    DataSet = apps.get_model('datasets', 'dataset')
  else:
    from reports.models import Report
    from datasets.models import DataSet

  cleanup_reports()
  cleanup_datasets()

# Handle when script called as a file
if __name__ == '__main__':
    import django
    django.setup()

    from reports.models import Report
    from datasets.models import DataSet
    main(None, None)
