#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

"""
This script adds dashboard_sheets
"""
import os
import sys

import json

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

import optparse


try:
    import reports.models
    from migration.migrate import addDashboardSheetData
except Exception as e:
    import django
    django.setup()
    import reports.models
    from migration.migrate import addDashboardSheetData

def parse():
  # TODO: Eventually, this can be used to clean up orphaned visuals --
  #       need to check that it's not used in any triggers, homepages, etc
  parser = optparse.OptionParser()
  parser.add_option('--remove_unused_visuals',
    dest="remove_unused_visuals",
    help="Clean up unused visuals",
    default=None)
  parser.add_option('--file',
    dest="file",
    help="filename to export information to",
    default=None)
  return parser.parse_args()

# MAIN CLEANUP FUNCTIONS
def update_reports_data(report_object, logging=None, handle_unused_visual=None, finished_update_cb=None):
  if logging:
    print_log = logging.info
  else:
    def print_log(message):
      print(message)

  # visual_id_to_dashboards and dashboard_id_to_visuals is an object with sets as values
  # Use sets just in case dashboard contains multiple iterations of visual (shouldn't happen)
  visual_id_to_dashboards = {}
  explicit_linked_visuals = set()
  dashboard_id_to_visuals = {}
  picklist_ids = {}
  visual_id_to_full_screen_dashboard = {}

  # create shared visuals or make single visual dashboards
  # - If visual used in just 1 dashboard, add dashboard_id to report_data
  # - If single visual is not used in any dashboard, create dashboard
  # - If visual used in 2 or more dashboards, make shared visual
  def check_is_visual(id):
    list_of_dashboards = visual_id_to_dashboards.get(id)
    return list_of_dashboards or list_of_dashboards == set([])

  def is_linked_visual(visual_id):
    is_explicitly_linked = visual_id in explicit_linked_visuals
    return is_explicitly_linked or (len(visual_id_to_dashboards.get(visual_id, [])) > 1)

  def get_report_stats():
    # add shared_visual_dashboards (if shared, contains array of dashboards it's in. otherwise, nothing)
    all_reports = report_object.objects.all()
    # Go through all the reports, figure out picklists and dashboards
    for report in all_reports:
      try:
        report_id = report.id
        report_type = report.report_type

        if report_type == 'dashboard':
          all_visual_ids = report.all_visual_ids()
          dashboard_id_to_visuals[report_id] = set()
          for viz_id in all_visual_ids:
            dashboard_id_to_visuals[report_id].add(viz_id)
            if check_is_visual(viz_id):
              visual_id_to_dashboards[viz_id].add(report_id)
            else:
              # Shouldn't be coming into this else statement often
              # either visual created after dashboard or visual no longer exists
              visual_id_to_dashboards[viz_id] = set([report_id])

          all_filter_ids = report.all_filter_ids()
          for pick_id_int in all_filter_ids:
            picklist_ids[pick_id_int] = True
            if check_is_visual(pick_id_int):
              # Very unlikely, but just in case picklist is in multiple dashboards
              visual_id_to_dashboards[pick_id_int].add(report_id)
            else:
              visual_id_to_dashboards[pick_id_int] = set([report_id])

        elif report_type == 'picklist' or report_type == 'daterange':
          picklist_ids[report_id] = True

        elif report_id not in visual_id_to_dashboards:
          visual_id_to_dashboards[report_id] = set()

          if report['shared_visual_dashboards']:
            explicit_linked_visuals.add(report_id)

        elif report['shared_visual_dashboards']:
          explicit_linked_visuals.add(report_id)

      except Exception as e:
        msg = 'ERROR: Cannot add report ID %s to visual_id_to_dashboards ' \
              'or picklist_ids' % report_id
        print(msg)
        print(str(e))
        if logging:
          logging.exception(msg)

  def iterate_and_update():
    for viz_id, dashboards in visual_id_to_dashboards.items():
      try:
        number_of_dashboards = len(dashboards)
        if number_of_dashboards >= 1:
          current_viz = report_object.objects.filter(id=viz_id)
          if current_viz.exists():
            current_viz = current_viz.first()
            print_log('Updating visual #%s used in %s dashboards' % (viz_id, number_of_dashboards))
            if number_of_dashboards == 1 and viz_id not in explicit_linked_visuals:
              current_viz_json = json.loads(current_viz.report_data)
              dashoard_id = list(dashboards)[0]

              if current_viz_json['report_data'].get('dashboard_id') == dashoard_id:
                print_log('Visual %s not saved' % viz_id)
                continue

              current_viz_json['report_data']['dashboard_id'] = dashoard_id
              current_viz.report_data = json.dumps(current_viz_json)
              current_viz.shared_visual_dashboards = None
            else:
              current_viz_json = json.loads(current_viz.report_data)
              previous_dashboard = current_viz_json['report_data'].pop('dashboard_id', None)
              print_log('Deleted dashboard_id (now linked): %s' % previous_dashboard)
              current_viz.report_data = json.dumps(current_viz_json)

              old_shared_visuals = set(json.loads(current_viz.shared_visual_dashboards or '[]'))
              new_shared_visuals = json.dumps(list(dashboards))
              print_log('Old shared_visual_dashboards: %s' % current_viz.shared_visual_dashboards)
              print_log('New shared_visual_dashboards: %s' % new_shared_visuals)

              if previous_dashboard is None and not dashboards ^ old_shared_visuals:
                print_log('Visual %s not saved' % viz_id)
                continue

              current_viz.shared_visual_dashboards = new_shared_visuals

            current_viz.save(update_fields=['shared_visual_dashboards', 'report_data'])
            print_log('Changes saved for visual %s' % viz_id)
          else:
            print_log('Visual %s does not exist anymore' % viz_id)

        elif number_of_dashboards == 0:
          print_log('Visual %s not in dashboard' % viz_id)
          if handle_unused_visual:
              print_log('Handling visual %s' % viz_id)
              handle_unused_visual(viz_id, visual_id_to_dashboards, dashboard_id_to_visuals, picklist_ids, visual_id_to_full_screen_dashboard)
      except Exception as e:
        msg = 'ERROR: Cannot update visual ID: %s' % viz_id
        print(msg)
        print(str(e))
        if logging:
          logging.exception(msg)

    for dash_id, visuals in dashboard_id_to_visuals.items():
      try:
        needs_save = False
        current_dash = report_object.objects.filter(id=dash_id)
        if current_dash.exists():
          current_dash = current_dash.first()
          new_report_data = json.loads(current_dash.report_data)
          if 'dashboard_sheets' not in new_report_data:
            # if the dashboard has not been converted already
            needs_save = True
            current_dash.report_data = json.dumps(addDashboardSheetData(new_report_data))
            print_log('Added dashboard_sheets to dashboard ID: %s' % dash_id)

          # Add shared_visual_dashboards
          shared_visuals = [visual_id for visual_id in visuals if is_linked_visual(visual_id)]
          print_log('Added shared_visual_dashboards to dashboard: %s' % dash_id)
          print_log('Old: %s' % current_dash.shared_visual_dashboards)
          print_log('New: %s' % shared_visuals)
          new_visuals = set(shared_visuals)
          old_visuals = set(json.loads(current_dash.shared_visual_dashboards or '[]'))
          needs_save = needs_save or (new_visuals ^ old_visuals)
          current_dash.shared_visual_dashboards = json.dumps(shared_visuals)

          if needs_save:
            current_dash.save(update_fields=['shared_visual_dashboards', 'report_data'])
            print_log('Changes saved for dashboard %s' % dash_id)
          else:
            print_log('Dashboard %s not saved' % dash_id)

      except Exception as e:
        msg = 'ERROR: Cannot update dashboard ID: %s' % dash_id
        print(msg)
        print(str(e))
        if logging:
          logging.exception(msg)

  get_report_stats()

  print_log('dump of visual_id_to_dashboards dictionary: %s' % visual_id_to_dashboards)
  print_log('dump of dashboard_id_to_visuals dictionary: %s' % dashboard_id_to_visuals)

  iterate_and_update()

  if finished_update_cb:
    finished_update_cb(visual_id_to_dashboards, dashboard_id_to_visuals, picklist_ids, visual_id_to_full_screen_dashboard)

def main():
  options, args = parse()
  logging = None
  if options.file:
    import logging
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(levelname)s %(message)s',
                        filename=options.file)

  update_reports_data(reports.models.Report, logging)

if __name__ == "__main__":
  try:
    main()
  except Exception as e:
    print(str(e))
    sys.exit(1)
