# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
Utilities related to CML models.

This file might be imported from Django settings,
so keep module-level imports to a minimum!

Under no condition import anything that requires an initialized
Django environment (like model access)!
"""

def embedding_response_transformer(body: dict, param_name = 'response'):
  """
  Given a response from a CML embeddings calculator endpoint, return a structure
  matching Openai's response structure.
  """
  return {
    'data': [
      {
        'embedding': body[param_name]
      }
    ]
  }

def completion_request_transformer(headers: dict, payload: dict, param_name='prompt'):
  messages = payload.get('request', {}).get(param_name, [])

  system_prompts = [m['content'] for m in messages if m['role'] == 'system']
  system_prompt = system_prompts[0] if system_prompts else ''

  assistant_messages = [m['content'] for m in messages if m['role'] == 'assistant']
  chatbot_first_line = assistant_messages[0] if assistant_messages else ''

  user_messages = [m['content'] for m in messages if m['role'] == 'user']
  if not user_messages:
      raise ValueError("No user messages found in the payload.")

  # The last user message is the current question
  question = user_messages[-1]

  # user messages (except last) + assistant messages (excluding first)
  chat_history = list(zip(user_messages[:-1], assistant_messages[1:]))

  prompt = llama_format_message(
    system_prompt=system_prompt,
    chatbot_first_line=chatbot_first_line,
    message=question,
    history=chat_history
  )

  new_payload = payload.copy()
  new_request = new_payload.get('request', {}).copy()
  new_request[param_name] = prompt
  new_payload['request'] = new_request

  return headers, new_payload


def completion_response_transformer(body: dict, param_name = 'response'):
  """
  Given a response from a CML completion endpoint, return a structure
  matching Openai's response structure.
  """

  llama_response = body.get(param_name)
  
  if not llama_response:
    return body

  # only keep the part after the last [/INST]
  INST = '[/INST]'
  if INST in llama_response:
    start_index = llama_response.rindex(INST) + len(INST)
    llama_response = llama_response[start_index:]

  return {
    'choices': [
      {
        'message': {
           'content': llama_response
        }
      }
    ]
  }

# Adapted from https://colab.research.google.com/drive/1SSv6lzX3Byu50PooYogmiwHqf5PQN68E?usp=sharing#scrollTo=PeEh17FDLzEe
# Then https://github.infra.cloudera.com/alexsong/hack2023_demo_chatbot/blob/4ce03e1e20d65e2921ae517a4a3cc09180b29883/3_launch-application/chat-demo.py

SYSTEM_PROMPT = """<s>[INST]{system_prompt_section}[/INST]{chatbot_first_line_section}</s>"""
QA_PAIR = """<s>[INST]{user_msg}[/INST]{model_answer}</s>"""
LAST_QUESTION = """<s>[INST]{message}[/INST]"""

def llama_format_message(system_prompt: str, chatbot_first_line: str, message: str, history: list, memory_limit: int = 3) -> str:
  """
  Formats the message and history for the Llama model.

  Parameters:
      system_prompt (str): System prompt text.
      chatbot_first_line (str): Initial assistant message.
      message (str): Current user message.
      history (list): Past conversation history as a list of (user_msg, model_answer) tuples.
      memory_limit (int): Limit on how many past interactions to consider.

  Returns:
      str: Formatted message string.
  """

  if system_prompt:
      system_prompt_section = f"<<SYS>>\n{system_prompt}\n<</SYS>>"
  else:
      system_prompt_section = ''

  if chatbot_first_line:
      chatbot_first_line_section = f"\n\n{chatbot_first_line}"
  else:
      chatbot_first_line_section = ''

  if len(history) > memory_limit:
    history = history[-memory_limit:]

  formatted_message = SYSTEM_PROMPT.format(
    system_prompt_section=system_prompt_section, chatbot_first_line_section=chatbot_first_line_section
  )

  # Handle conversation history
  for user_msg, model_answer in history:
    formatted_message += QA_PAIR.format(user_msg=user_msg, model_answer=model_answer)

  # Handle the current message
  formatted_message += LAST_QUESTION.format(message=message)

  return formatted_message
