#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Data connection utility.

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

import os
import sys
import json
import optparse
import traceback

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)


from datasets.models import DataConnection

from dataconn.connections_types import DATACONN_ARCENGINE, DATACONN_HIVE, DATACONN_IMPALA, \
    DATACONN_IMPYLA, DATACONN_HYVE, DATACONN_ARCENGINE_V2
from jobs.views_encryption import encrypt_secret
valid_types = [DATACONN_ARCENGINE, DATACONN_HIVE, DATACONN_IMPALA, DATACONN_IMPYLA, \
    DATACONN_HYVE, DATACONN_ARCENGINE_V2]
valid_auth = ['Plain', 'NoSasl', 'LDAP', 'Kerberos']
valid_sock = ['Normal', 'SSL', 'SSLCert']

valid_actions = ['create', 'list']

defaults = {
  DATACONN_HIVE : {
    'port' : '10000',
    'auth' : 'Plain',
    'sock' : 'Normal',
    'cache_enable' : 0,
    'cache_retention' : (60 * 60 * 24),
  },
  DATACONN_HYVE : {
    'port' : '10000',
    'auth' : 'Plain',
    'sock' : 'Normal',
    'cache_enable' : 0,
    'cache_retention' : (60 * 60 * 24),
  },
  DATACONN_IMPALA : {
    'port' : '21050',
    'auth' : 'NoSasl',
    'sock' : 'Normal',
    'cache_enable' : 0,
    'cache_retention' : (60 * 60 * 24),
  },
  DATACONN_IMPYLA : {
    'port' : '21050',
    'auth' : 'NoSasl',
    'sock' : 'Normal',
    'cache_enable' : 0,
    'cache_retention' : (60 * 60 * 24),
  },
  DATACONN_ARCENGINE : {
    'port' : '31050',
    'auth' : 'NoSasl',
    'sock' : 'Normal',
    'cache_enable' : 1,
    'cache_retention' : (5 * 60),
  },
  DATACONN_ARCENGINE_V2 : {
    'port' : '31050',
    'auth' : 'NoSasl',
    'sock' : 'Normal',
    'cache_enable' : 1,
    'cache_retention' : (5 * 60),
  },
}

def fatal(text):
  """ Fatal error """
  estr = traceback.format_exc()
  if not estr.startswith("None"):
    print(estr)
  if text:
    print(text)
  sys.exit(-1)

def parse():

  usage = """

  %%prog <action> [options]

  <action> must be one of: %s""" % (str(valid_actions))

  parser = optparse.OptionParser(usage=usage)
  parser.add_option('--name', dest="name", default='default',
    help='Data connection name')
  parser.add_option('--type', dest="type",
    help='Data connection type.  One of: %s' % str(valid_types))
  parser.add_option('--host', dest="host", default='localhost',
    help='Host name or IP address')
  parser.add_option('--port', dest="port",
    help='Host port (default depends on connection type)')
  parser.add_option('--auth', dest="auth",
    help='Authentication mode.  One of %s (default according to type)' % str(valid_auth))
  parser.add_option('--kerbserv', dest="kerbserv",
    help='Kerberos service name (default depends on connection type)')
  parser.add_option('--sock', dest="sock",
    help='Socket type.  One of %s (default according to type)' % str(valid_sock))
  parser.add_option('--username', dest="username", default="",
    help='Username for connection')
  parser.add_option('--password', dest="password", default="",
    help='Password for connection')
  parser.add_option('--desc', dest="desc", default="",
    help='Data connection description')
  parser.add_option('--queue_depth', dest="queue_depth", type="int",
    help='Specify queue depth (from 1 to 100)')
  parser.add_option('--cache_enable', dest="cache_enable", type="int",
    help='Specify result cache enable (0 or 1)')
  parser.add_option('--cache_retention', dest="cache_retention", type="int",
    help='Specify result cache retention time (in seconds)')

  parser.add_option('--impersonation', dest="impersonation", action="store_true", default=False,
    help='Specify impersonation mode')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if len(args) == 0:
    fatal('No action given')
  if len(args) > 1:
    fatal('Extra options: %s' % str(args))
  action = args[0].lower()
  if action not in valid_actions:
    fatal("Action must be one of %s, not: '%s'" % (str(valid_actions), action))

  if action == 'create':
    if not options.type:
      fatal('Must specify connection type')
    options.type = options.type.lower()
    if options.type.lower() not in valid_types:
      fatal("Connection type must be one of %s, not '%s'" % (str(valid_types), options.type))

    # Set defaults according to connection type
    for (key, val) in defaults[options.type].items():
      if getattr(options, key) is None:
        setattr(options, key, val)

    authlow = [a.lower() for a in valid_auth]
    if options.auth.lower() not in authlow:
      fatal("Authentication mode must be one of %s, not '%s'" % (str(valid_auth), options.auth))
    options.auth = valid_auth[authlow.index(options.auth.lower())]

    socklow = [a.lower() for a in valid_sock]
    if options.sock.lower() not in socklow:
      fatal("Socket type must be one of %s, not '%s'" % (str(valid_sock), options.sock))
    options.sock = valid_sock[socklow.index(options.sock.lower())]

    if options.queue_depth:
      if options.queue_depth < 1 or options.queue_depth > 100:
        fatal("Queue depth must be from 1 to 100, not %s" % options.queue_depth)

    if options.cache_enable:
      if options.cache_enable not in [0, 1]:
        fatal("Cache enable must be 0 or 1, not %s" % options.cache_enable)

    if options.cache_retention:
      max_retention = 60 * 60 * 24 * 365
      if options.cache_retention < 0 or options.cache_retention > max_retention:
        fatal("Cache retention must be from 0 to %s, not %s" % (max_retention, options.cache_enable))

  return (options, action)

def dclist(options):
  print("Listing DCs")
  dcs = DataConnection.objects.filter()
  for dc in dcs:
    print("DC %s (id %s):" % (dc.dataconnection_name, dc.id))
    for key, val in vars(dc).items():
      if key not in ['id', 'dataconnection_name']:
        print("  %s = %s" % (key, val)) 

def dccreate(options):
  dc = DataConnection.objects.filter(dataconnection_name=options.name)
  if dc:
    fatal("Data connection with name '%s' already exists" % options.name)

  params = {"HOST":options.host, "PORT":options.port, "AUTH":options.auth,
            "SOCK":options.sock}
  if options.username:
    params["USERNAME"] = options.username
  if options.kerbserv:
    params["KERBSERV"] = options.kerbserv
  if options.queue_depth:
    params["QUEUE_DEPTH"] = options.queue_depth
  if options.impersonation is not None:
    params["IMPERSONATION"] = options.impersonation

  params["CACHE"] = {"ENABLE": options.cache_enable,
                     "RETENTION": options.cache_retention}

  dc = DataConnection(
    dataconnection_name=options.name,
    dataconnection_type=options.type,
    dataconnection_password=json.dumps(encrypt_secret(str(options.password))),
    dataconnection_description=options.desc,
    dataconnection_info=json.dumps({"PARAMS":params})
  )
  dc.save()
              
def main():
  options, action = parse()

  if action == 'list':
    dclist(options)
  elif action == 'create':
    dccreate(options)


if __name__ == "__main__":
  main()

