#!/usr/bin/python -tt
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

"""
This script analyzes and suggests acceleration for apps
"""

import os
import sys

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

import json
from . import sqlselect
from . import app
import optparse

import datasets.models

class DataSets(object):
  def __init__(self):
    self.__datasets = {}

  def add(self, dataset_id):
    if dataset_id in self.__datasets:
      return self.__datasets[dataset_id]
    self.__datasets[dataset_id] = DataSet(dataset_id)
    return self.__datasets[dataset_id]

  def datasets(self):
    return list(self.__datasets.values())

  def cb(self, fn):
    for ds in self.datasets():
      fn(ds)

class Table(object):
  def __init__(self):
    self.__tablename = None
    self.__partitions = []
    self.__columns = []

  def set_tablename(self, tablename):
    self.__tablename = tablename

  def set_partitions(self, partitions):
    self.__partitions = partitions

  def set_columns(self, columns):
    self.__columns = columns

  def display(self, display_unchanged=False):
    print(self.__tablename, self.__partitions)
    for c in self.__columns:
      if c.get('name') != c.get('alias'):
        print(' *', c.get('alias'), '<-->', (c.get('name') or c.get('expr')))
    if display_unchanged:
      for c in self.__columns:
        if c.get('name') == c.get('alias'):
          print('  ', c.get('alias'), '<-->', (c.get('name') or c.get('expr')))

class DataSet(object):
  def __init__(self, dataset_id):
    self.__dataset_id = dataset_id
    self.__tracked_cols = {}
    self.__dataset = datasets.models.DataSet.objects.get(id=dataset_id)
    self.__tables = []
    self.__process()

  def display(self):
    print(self.__dataset.__dict__)

  def __process(self):
    if not self.__dataset.dataset_info:
      table = Table()
      table.set_tablename(self.__dataset.dataset_detail)
      # table.display()
      self.__tables.append(table)
      return

    j = json.loads(self.__dataset.dataset_info)
    for t in j:
      table = Table()
      table.set_tablename(t['tablename'])
      if t.get('partitions', []):
        table.set_partitions(t.get('partitions'))
      if t.get('columns', []):
        table.set_columns(t.get('columns'))
      table.display()
      self.__tables.append(table)

  def track_col(self, coltype, col):
    if col in self.__tracked_cols:
      return
    self.__tracked_cols[col] = coltype

  def cols(self):
    return self.__tracked_cols

  def select_statement(self):
    s = sqlselect.SelectStatement()
    s.table(self.__dataset.dataset_detail)
    for k, v in self.__tracked_cols.items():
      if v == app.App.COL_DIMENSION:
        s.addGroup(k)
        continue
      if v == app.App.COL_AGGREGATE:
        s.addAggregate(k)
        continue
    return s

def parse():
  parser = optparse.OptionParser()
  parser.add_option('--dataset',
    action="append", dest="dataset",
    help="dataset to analyze", default=None)
  return parser.parse_args()

def main():
  options, args = parse()

  dsids = set()

  if options.dataset:
    for o_dataset in options.dataset:
      for dataset_id in o_dataset.split(","):
        dsids.add(int(dataset_id))

  for dsid in dsids:
    ds = DataSet(dsid)
    ds.display()

if __name__ == "__main__":
  main()

