#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Dataset update utility.

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

USAGE = """

The ds_update allows command line update operations for arcviz datasets.

Currently supported operations are:

  Define a new computed column
  Update an existing computed or base column
  Remove a compute column from the dataset

The column attributes that can be controlled are the same as are available
in the 'Attributes' tab of the visual datasets page:

  Set the expression text for a computed column:    --expr "<expression text>"
  Set the display name for a column:                --alias "<display name>"
  Set the data type for a column:                   --cast <new type>
  Set the default aggregations for a column:        --dflagg <agg function>
  Set the dimension/measure state:                  --dimension / --measure
  Control column visibility in the report builder:  --hidden / --shown

Valid types are: INT, REAL, DOUBLE, BOOLEAN, TIMESTAMP, STRING

Note that the program will perform the same expression validation that is
done in the visual environment.  This means the data connection for the
dataset must be available.

Note also that the program uses the same dataset update as the visual tool
so that dataset column name changes will be propagated to all the reports based
on the dataset.

Example:

  Perform the following set of operations on dataset 'Test Dataset':

    1) Add computed column 'Store' as 'upper([store_name])'
    2) Hide column 'store_name'
    3) Set display name for column 'age' as 'Age', cast to INT,
       set the default aggregation to 'avg',  and set as a dimension
    4) Remove computed column 'Total Sales'

    python util/ds_update.py --name 'Test Dataset'  \\
      --newcol Store --expr 'upper([store_name])'   \\
      --update store_name --hidden                  \\
      --update age --alias 'Age' --dimension --dflagg avg --cast INT   \\
      --delcol 'Total Sales'

  Make sure to include brackets ([]) when naming columns in the expression text.

"""

import os
import sys
import json
import getpass
import optparse
import traceback

from optparse import OptionValueError

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)
sys.path.append("%s/../vizlib" % THISDIR)

from django.db.models.base import ObjectDoesNotExist
from django.contrib.auth.models import User

from datasets.models import DataSet, DataConnection
from dataconn.dataconn import get_dataconn
from datasets.utils import get_default_dataconnection
from datasets.views import dataset_update_info
from vizlib.dataset import ModelDataset, getDatasetInfo, column_is_measure
from vizlib.dataset_request import DatasetRequest

# Turn off the logging in the django parts (like hs2conn.py)
import logging
logging.disable(logging.DEBUG)

valid_casts = ['INT','BOOLEAN','TIMESTAMP','DOUBLE','STRING','REAL']
valid_dflaggs = ['sum', 'count', 'ndv', 'countdistinct', 'min', 'max', 'avg', 'group_concat']

def fatal(text):
  """ Fatal error """
  estr = traceback.format_exc()
  if not estr.startswith("None"):
    print(estr)
  if text:
    print(text)
  sys.exit(-1)

# Start a new option structure
def newop(option, opt_str, value, parser):
  if not value:
    raise OptionValueError("%s requires column name parameter" % opt_str)
  col = {"op":opt_str,"name":value}
  parser.column_ops.append(col)

# Process an parameter for the current operation
def param(option, opt_str, value, parser):
  if len(parser.column_ops) == 0:
    raise OptionValueError("%s requires column name specified via --update or --newcol options" % opt_str)
  col = parser.column_ops[-1]
  if opt_str == '--cast':
    if value.upper() not in valid_casts:
      raise OptionValueError("--cast must be one of %s, not %s" % (str(valid_casts), value))
    col['cast'] = 'DOUBLE' if value.upper() == 'REAL' else value.upper()
  elif opt_str == '--dflagg':
    if value.lower() not in valid_dflaggs:
      raise OptionValueError("--dflagg must be one of %s, not %s" % (str(valid_dflaggs), value))
    col['dflagg'] = value.lower()
  elif opt_str == '--basecol':
    col['basecol'] = value
  elif opt_str == '--alias':
    col['alias'] = value
  elif opt_str == '--dimension':
    col['isdim'] = True
  elif opt_str == '--measure':
    col['isdim'] = False
  elif opt_str == '--hidden':
    col['hidden'] = True
  elif opt_str == '--shown':
    col['hidden'] = False
  elif opt_str == '--expr':
    col['expr'] = value
  elif opt_str == '--dimh':
    col['dimh'] = value
  else:
    fatal("Unknown column operation parameter: %s" % opt_str)

def parse():

  usage = """

  %prog <dataset specifier> [options]

  <dataset specifier> is either '--name <Dataset name>' or '--id <Dataset ID>'
  """

  parser = optparse.OptionParser(usage="")
  parser.column_ops = []

  parser.add_option('--name', dest='name',
    help='Dataset name')
  parser.add_option('--id', dest='id', type='int',
    help='Dataset ID')
  parser.add_option('--user', dest='username', default='admin',
    help='Perform data operations as this user')
  parser.add_option('--usage', dest='usage', action='store_true',
    help='Display usage text and exit')
  parser.add_option('--dryrun', dest='dryrun', action='store_true',
    help='Do operations but do not save updated dataset')

  # These options start a new action structure being constructed
  parser.add_option('--update', type='string', action='callback', callback=newop,
    help='Update an existing column')
  parser.add_option('--newcol', type='string', action='callback', callback=newop,
    help='Add a computed')
  parser.add_option('--delcol', type='string', action='callback', callback=newop,
    help='Remove a computed column')


  # Parameters which are applied to the current action (update, newcol)
  parser.add_option('--basecol', type='string', action='callback', callback=param,
    help='Specify display name for base column')
  parser.add_option('--alias', type='string', action='callback', callback=param,
    help='Specify display name for base column')
  parser.add_option('--cast', type='string', action='callback', callback=param,
    help='Specify data type for base column')
  parser.add_option('--dflagg', type='string', action='callback', callback=param,
    help='Specify default aggregation function for column')
  parser.add_option('--hidden', action='callback', callback=param,
    help='Specify base column is hidden in app builder')
  parser.add_option('--shown', action='callback', callback=param,
    help='Specify base column is shown in app builder')
  parser.add_option('--dimension', action='callback', callback=param,
    help='Specify base column is a dimension')
  parser.add_option('--measure', action='callback', callback=param,
    help='Specify base column is a measure')
  parser.add_option('--expr', type='string', action='callback', callback=param,
    help='Specify expression text for computed column')
  parser.add_option('--dimh', type='string', action='callback', callback=param,
    help='Specify dimension hierarchy column list for computed column')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if options.usage:
    print(USAGE)
    parser.print_help()
    sys.exit(0)

  if len(args) > 0:
    fatal('Extra options: %s' % str(args))

  if not options.name and not options.id:
    fatal("Must specify dataset name or id (--name, --id)")
  if options.name:
    if options.id is not None:
      fatal("Can not specify both dataset name and id (--name, --id)")
    dl = DataSet.objects.filter(dataset_name=options.name)
    if len(dl) == 0:
      fatal("Dataset name %s not found" % options.name)
    if len(dl) != 1:
      fatal("More than one dataset found for name %s: %s" % (options.name, [ds.id for ds in dl]))
    ds = dl[0]
  else:
    dl = DataSet.objects.filter(id=options.id)
    if len(dl) != 1:
      fatal("Dataset ID %d not found" % options.id)
    ds = dl[0]

  user = None
  try:
    user = User.objects.get(username=options.username)
  except ObjectDoesNotExist as e:
    sys.exc_clear()
  if not user:
    fatal("Username '%s' not found" % options.username)
  options.userid = user.pk

  if len(parser.column_ops) == 0:
    fatal("No column operations specified")

  return (options, ds, parser.column_ops)

# Set of routines for access to the parsed json dataset data
def colName(info):
  return info.get('alias') or info.get('name')

def findCol(name, dataset, must=True):
  for table in dataset:
    col = next((c for c in table.get('columns') if name == colName(c)), None)
    if col:
      return (table, col)
  if must:
    raise Exception
  return (None, None)

def addCol(info, dataset):
  table, col = findCol(colName(info), dataset, False)
  if col:
    raise Exception
  table, base = findCol(info.get('basecol'), dataset, True)

  table.insert(info, table.index(base))

def updateCol(info, dataset):
  table, col = findCol(colName(info), dataset)
  col.update(info)

def validateCol(options, ds, ds_id, dc, info):

  dsreq = {'type':'.validate',
           'dataset_id':ds_id,
           'validate':info.get('expr')}

  req_obj = DatasetRequest(json.dumps(dsreq), ds)
  bld_req = dc.build(username=options.username, userid=options.userid,
                     query=None, dataset_request=req_obj)
  res = dc.execute(bld_req, async_flag=False)
  res = json.loads(res)
  info['ispart'] = res['ispart'] == "true"
  info['type'] = res['type']
  if info.get('isdim') is None:
    info['isdim'] = not column_is_measure(info.get('cast') or info['type'])
  info['exprhasagg'] = res['isaggregate'] == "true"

def main():
  options, django_ds, column_ops = parse()

  connid = django_ds.dataconnection_id
  if not connid:
    connid = get_default_dataconnection().id

  dc = get_dataconn(connid)
  if not django_ds.dataset_info:
    try:
      django_ds.dataset_info = getDatasetInfo(django_ds.dataset_detail,
                                              connid, options.username)
    except Exception as e:
      fatal("Unable to fill in column details for table: %s" % django_ds.dataset_detail)
    print("Loaded columns")

  dataset = json.loads(django_ds.dataset_info)

  for op in column_ops:
    ds = ModelDataset(json.dumps(dataset), dc.properties())
    info = op.copy()
    del(info['op'])
    opcode = op['op']
    name = info['name']

    if opcode == '--newcol':
      print("Adding new column: %s" % name)
      table, col = findCol(name, dataset, False)
      if col:
        fatal("Column '%s' already exists in dataset" % colName(info))

      if info.get('alias') is not None:
        fatal("Can not specify alias for computed columns")
      info['alias'] = name
      info['name'] = ''

      if not info.get('expr'):
        if not info.get('dimh'):
          fatal("Must specify expression for new column")
      validateCol(options, ds, django_ds.dataconnection_id, dc, info)

      basecol = info.get('basecol')
      if basecol is None:
        basecols = set()
        ds.process_expr(info.get('expr'), columns=basecols)
        if len(basecols) == 0:
          fatal("New computed column requires --basecol <column> to be specified")
        for b in basecols:
          basecol = b.attr.get('name')
          if basecol:
            break
      table, base = findCol(basecol, dataset, False)
      if not base:
        fatal("Base column '%s' not found in dataset" % info.get('basecol'))
      info['basecol'] = basecol
      table['columns'].insert(table['columns'].index(base) + 1, info)

    elif opcode == '--delcol':
      print("Removing computed column: %s" % name)
      table, col = findCol(name, dataset)
      if col.get('name'):
        fatal("Can not remove base column: %s" % name)
      table['columns'].remove(col)

    elif opcode == '--update':
      print("Update column: %s" % name)
      table, col = findCol(name, dataset)

      # name is only populated for base columns.
      isbase = bool(col.get('name'))
      alias = info.get('alias')
      if alias is not None:
        if not alias:
          if isbase:
            alias = col.get('name')
          else:
            fatal("Can not remove display name for computed column")
        testt, testc = findCol(alias, dataset, must=False)
        if testc is not None and testc != col:
          fatal("Column alias '%s' already exists in dataset" % alias)
        col['alias'] = alias

      newdim = False
      if info.get('expr'):
        if isbase:
          fatal("Can not set expression for base column")
        col['expr'] = info['expr']
        validateCol(options, ds, django_ds.dataconnection_id, dc, col)
        newdim = True

      cast = info.get('cast')
      if cast is not None:
        newdim = True
        if cast == '' or cast == col.get('type'):
          if col.get('cast') is not None:
            del col['cast']
        else:
          col['cast'] = cast

      if info.get('isdim') is not None:
        col['isdim'] = info['isdim']
      elif newdim:
        col['isdim'] = not column_is_measure(col.get('cast') or col.get('type'))

      if info.get('hidden') is not None:
        col['hidden'] = info.get('hidden')

      if info.get('dflagg') is not None:
        col['dflagg'] = info.get('dflagg')

    else:
      fatal("Unknown column operation: %s" % opcode)

  if options.dryrun:
    print("New DS info: %s" % str(dataset))
  else:
    dataset_update_info(django_ds, json.dumps(dataset))

if __name__ == "__main__":
  main()

