#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import os, sys
import optparse

# Find the
THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

import datasets.models as ds_models
from datasets.utils import get_default_dataconnection
from vizlib.dataset import getDatasetInfo, refreshDatasetColumns
from django.db.models import Count
from django.core import serializers

# Turn off the logging in the django parts (like hs2conn.py)
import logging
logging.disable(logging.DEBUG)

# Turn off role polling
from django.conf import settings
settings.RBAC_POLL_SECONDS = 0

def dataset_create(dsname='', tablename=None, username='admin', dcid='', populate=False):
  dataset_type = 'singletable'
  dataset_name = dsname
  dataset_info = '[]'
  dataset_detail = tablename
  # noinspection PyBroadException
  try:
    dcid = int(dcid)
    dc = ds_models.DataConnection.objects.get(id=dcid)
  except Exception:
    dc = get_default_dataconnection()

  d = ds_models.DataSet(dataset_name=dataset_name, dataset_type=dataset_type,
                        dataset_info=dataset_info,
                        dataset_detail=dataset_detail,
                        dataconnection=dc)
  d.create_user = username
  d.update_user = username

  if populate:
    d.dataset_info = getDatasetInfo(tablename, dc.id, username)

  d.save()

  return d

def dataset_refresh(dslist=None, username='admin'):
  if dslist is None:
    dslist = []

  for dsid in set(dslist):
    dsqs = ds_models.DataSet.objects.filter(id=dsid)
    if dsqs.count() == 1:
      ds = dsqs[0]
      if refreshDatasetColumns(ds, username):
        print("Updated dataset %i" % ds.id)
      else:
        print("No changes detected for dataset %i" % ds.id)
    else:
      print("Dataset %i does not exist" % dsid)


def dataset_delete(dslist=None, force=False, deleteunused=False):
  if dslist is None:
    dslist = []
  if deleteunused:
    dslist = dslist + list(ds_models.DataSet.objects \
                           .annotate(reportcount=Count('report')) \
                           .filter(reportcount=0).values_list('id', flat=True))

  for dsid in set(dslist):
    dsqs = ds_models.DataSet.objects.filter(id=dsid)
    if dsqs.count() == 1:
      ds = dsqs[0]
      if force or ds.report_set.count() == 0:
        print("Deleting Dataset %i" % ds.id)
        ds.delete()
      else:
        print("Dataset %i is in use.  Can't Delete." % ds.id)
    else:
      print("Dataset %i does not exist" % dsid)


def dataset_list(headers=True, listunused=False):
  datasets = ds_models.DataSet.objects \
    .annotate(reportcount=Count('report')).all()
  if listunused:
    datasets = datasets.filter(reportcount=0)
  if headers:
    print("id, dataset name, table name, dataconnection_id, number of reports using this dataset")
  for ds in datasets:
    print(('%i, "%s", "%s", %s, %i' % \
          (ds.id, ds.dataset_name, ds.dataset_detail, ds.dataconnection_id,
           ds.reportcount)))


def dataset_export(dslist=None, exportfile=None, username='admin', populate=False):
  with open(exportfile, 'w') as f:
    dsl = ds_models.DataSet.objects.filter(id__in=dslist)
    if populate:
      for ds in dsl:
        if not ds.dataset_info:
          ds.dataset_info = getDatasetInfo(ds.dataset_detail, ds.dataconnection_id, username)
          ds.save()
    serializers.serialize('json', dsl, stream=f)


def dataset_import(importfile=None, dcid='', dcmaps=None):
  dc = None
  default_dc = get_default_dataconnection()
  new_dcmaps = {}
  print(dcmaps)
  try:
    if dcid is not '':
      dcid = int(dcid)
      dc = ds_models.DataConnection.objects.get(id=dcid)
    if dcmaps:
      for src_dc, dest_dc in dcmaps.items():
        new_dcmaps[src_dc]=ds_models.DataConnection.objects.get(id=dest_dc)
  except Exception as e:
    raise Exception("Error setting Dataconnection ID")

  with open(importfile, 'r') as f:
    dsl = serializers.deserialize('json', f.read())
    for ds in dsl:
      if ds.object.dataconnection_id in new_dcmaps:
        ds.object.dataconnection = new_dcmaps[ds.object.dataconnection_id]
      elif dc:
        ds.object.dataconnection = dc
      ds.object.pk = None
      print("Import %s" % ds)
      ds.save()

def dataset_switchdataconnection(dslist=None, dcswitchids=None,
                                dcswitchnames=None):
  if dcswitchids:
    src_dc_id, dest_dc_id = dcswitchids
    src_dc = ds_models.DataConnection.objects.get(id=src_dc_id)
    dest_dc = ds_models.DataConnection.objects.get(id=dest_dc_id)
  else:
    src_dc_name, dest_dc_name = dcswitchnames
    src_dc = ds_models.DataConnection.objects.get(dataconnection_name=src_dc_name)
    dest_dc = ds_models.DataConnection.objects.get(dataconnection_name=dest_dc_name)

  if src_dc == None or dest_dc == None:
    raise Exception("Incorrect source or destination data connection specified")

  dsl = ds_models.DataSet.objects.filter(id__in=dslist)
  for ds in dsl:
    if ds.dataconnection.id == src_dc.id:
      ds.dataconnection = dest_dc
      ds.save()
      print('Dataconnection for dataset %d:%s is now %d:%s' % (ds.id, ds.dataset_name, dest_dc.id, dest_dc.dataconnection_name))
    else:
      print('Dataset %d:%s does not belong to the %d:%s dataconnection' % \
            (ds.id, ds.dataset_name, src_dc.id, src_dc.dataconnection_name))

def parse():
  prog = os.path.basename(sys.argv[0])
  # parser = optparse.OptionParser(conflict_handler="resolve")
  parser = optparse.OptionParser()

  listgroup = optparse.OptionGroup(parser, "List Datasets",
                                   "%s -l [--noheaders] [--listunused]" % prog)
  listgroup.add_option('-l', '--list', dest="list",
                       action="store_true", default=False,
                       help="List datasets")
  listgroup.add_option('--noheaders', dest="headers",
                       action="store_false", default=True,
                       help="Don't print headers")
  listgroup.add_option('--listunused', dest="listunused",
                       action="store_true", default=False,
                       help="List only unused Datasets")

  creategroup = optparse.OptionGroup(parser, "Create Dataset",
                                     "%s -c --tablename=TABLENAME --dsname=DSNAME [ --dcid=DATACONNECTION_ID ] [--populatecolumns]" % prog)
  creategroup.add_option('-c', '--create', dest="create",
                         action="store_true", default=False,
                         help="Create a dataset")
  creategroup.add_option('--tablename', dest="tablename",
                         help="tablename to use for dataset creation",
                         default=None)
  creategroup.add_option('--dsname', dest="dsname",
                         help="Dataset name to use for creation", default='')
  creategroup.add_option('--dcid', dest="dcid",
                         help="Dataconnection id to use for creation",
                         default='')
  creategroup.add_option('--populatecolumns', dest="populate",
                         action="store_true", default=False,
                         help="Populate dataset column information from table on data connection")


  refreshgroup = optparse.OptionGroup(parser, "Refresh Datasets",
                                     "%s -r [ID1] [ID2] ... " % prog)
  refreshgroup.add_option('-r', '--refresh', dest="refresh",
                         action="store_true", default=False,
                         help="Refresh datasets: arguments are dataset ids")

  deletegroup = optparse.OptionGroup(parser, "Delete Datasets",
                                     "%s -d [--force] [--deleteunused] [ID1] [ID2] ... " % prog)
  deletegroup.add_option('-d', '--delete', dest="delete",
                         action="store_true", default=False,
                         help="Delete datasets: arguments are dataset ids")
  deletegroup.add_option('--force', dest="force",
                         action="store_true", default=False,
                         help="Force deletion even if currently in use. Dangerous!")
  deletegroup.add_option('--deleteunused', dest="deleteunused",
                         action="store_true", default=False,
                         help="Delete all unused datasets")

  exportgroup = optparse.OptionGroup(parser, "Export Datasets",
                                     "%s -e FILE [--populatecolumns] [ID1] [ID2] ... " % prog)
  exportgroup.add_option('-e', '--export', dest="exportfile",
                         metavar="FILE",
                         action="store", default=None,
                         help="Export datasets: arguments are dataset ids")
  exportgroup.add_option('--populatemissing', dest="populate",
                         action="store_true", default=False,
                         help="Populate dataset column information if not present")

  importgroup = optparse.OptionGroup(parser, "Import Datasets",
                 "%s -i FILE [ --dcid=DATACONNECTION_ID ] [ --dcmap=OLDDCID:NEWDCID ]" % prog)
  importgroup.add_option('-i', '--import', dest="importfile",
                         metavar="FILE",
                         action="store", default=None,
                         help="Import datasets")
  importgroup.add_option('--dcmap',
                         action="append", dest="dcmaps",
                         help="oldid:newid for dataconnection modifications. You can specify this option multiple times",
                         default=None)
  # try:
  #   importgroup.add_option('--dcid', dest="dcid",
  #                          help="Dataconnection id to use for creation", default='')
  # except optparse.OptionConflictError:
  #   pass

  switchgroup = optparse.OptionGroup(parser, "Switch Dataset's DataConnections",
     "%s -s [ --dcswitchids=SRC_DC_ID:DST_DC_ID ]" \
     "[ --dcswitchnames=SRC_DC_NAME:DST_DC_NAME] [DSID1] [DSID2] ..." % prog)
  switchgroup.add_option('-s', '--switchdataconn', dest="switch",
                        action="store_true", default=False,
                        help="Switch dataconnection IDS for specified datasets")
  switchgroup.add_option('--dcswitchids', dest="dcswitchids",
                        help="Specify src and destination dataconnection IDs",
                        default=None)
  switchgroup.add_option('--dcswitchnames', dest="dcswitchnames",
                        help="Specify src and destination dataconnection names",
                        default=None)

  parser.add_option_group(listgroup)
  parser.add_option_group(creategroup)
  parser.add_option_group(refreshgroup)
  parser.add_option_group(deletegroup)
  parser.add_option_group(exportgroup)
  parser.add_option_group(importgroup)
  parser.add_option_group(switchgroup)

  return parser


def main():
  parser = parse()
  options, args = parser.parse_args()

  if not 'DJANGO_SETTINGS_MODULE' in os.environ:
    print("error: Please set the 'DJANGO_SETTINGS_MODULE' environment variable")
    exit(1)

  # If the settings file has a relative path for the db we need to find it
  # relative to the settings module not this file
  dsmname=os.environ['DJANGO_SETTINGS_MODULE']
  #print "Using: DJANGO_SETTINGS_MODULE=%s" % dsmname
  dsm = __import__(dsmname, fromlist=[''])
  if dsm.DATABASES['default']['ENGINE'].find("sqlite") >= 0:
    dbname = dsm.DATABASES['default']['NAME']
    if os.path.isabs(dbname):
      #print "Using: db = %s" % dbname
      from django.conf import settings
    else:
      #print dsm.__file__
      newdbname = os.path.normpath(os.path.join(os.path.dirname(dsm.__file__),
                                                "..",
                                                dsm.DATABASES['default']['NAME']))
      from django.conf import settings
      settings.DATABASES['default']['NAME'] = newdbname
      from django.db import connection
      connection.close()
      #print "Using: db = %s" % newdbname


  # List
  if options.list:
    dataset_list(headers=options.headers, listunused=options.listunused)
    exit(0)

  # Create
  if options.create:
    if not options.tablename:
      raise Exception("Table name must be specified")
    if not options.dsname:
      raise Exception("Dataset name must be specified")
    dataset = dataset_create(dsname=options.dsname, tablename=options.tablename,
                             dcid=options.dcid, populate=options.populate)
    print(dataset.id)
    exit(0)

  # Refresh
  if options.refresh:
    try:
      dslist = [int(a) for a in args]
    except:
      parser.print_help()
      exit(1)
    dataset_refresh(dslist)
    exit(0)

  # Delete
  if options.delete:
    try:
      dslist = [int(a) for a in args]
    except:
      parser.print_help()
      exit(1)
    dataset_delete(dslist, force=options.force,
                   deleteunused=options.deleteunused)
    exit(0)

  # Export
  if options.exportfile:
    if not options.exportfile:
      raise Exception("Export filename must be specified")
    try:
      dslist = [int(a) for a in args]
    except:
      parser.print_help()
      exit(1)
    dataset_export(exportfile=options.exportfile, dslist=dslist, populate=options.populate)
    exit(0)

  # Import
  if options.importfile:
    dcmaps = {}
    if options.dcmaps:
      for dcmap in options.dcmaps:
        old_id, sep, new_id = dcmap.partition(":")
        old_id = int(old_id)
        new_id = int(new_id)
        dcmaps[old_id] = new_id
    if not options.importfile:
      raise Exception("Import filename must be specified")
    dataset_import(importfile=options.importfile, dcid=options.dcid,
                   dcmaps=dcmaps)
    exit(0)

  # Switch data connections
  if options.switch:
    if not options.dcswitchids and not options.dcswitchnames:
      raise Exception("Either of dcswitchids or dcswitchnames must be \
                      specified")
    try:
      dslist = [int(a) for a in args]
    except:
      parser.print_help()
      exit(1)
    dcswitchids = None
    dcswitchnames = None
    if options.dcswitchids:
      srcdcid, sep, dstdcid = options.dcswitchids.partition(":")
      dcswitchids = int(srcdcid), int(dstdcid)
    if options.dcswitchnames:
      srcdcname, sep, dstdcname = options.dcswitchnames.partition(":")
      dcswitchnames = (srcdcname, dstdcname)

    dataset_switchdataconnection(dslist,
                   dcswitchids=dcswitchids,
                   dcswitchnames=dcswitchnames)
    exit(0)

  # no action specified
  parser.print_help()
  exit(1)


if __name__ == "__main__":
  main()
