#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Show roles for a particular dataset

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

import os
import sys
import json
import optparse
import traceback

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

# Turn off the logging in the django parts (like hs2conn.py)
import logging
logging.disable(logging.INFO)

# Filter warnings
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

from datasets.models import DataSet
from rbac.models import Role
from rbac.permdefs import PERMDEFS

def fatal(text):
  """ Fatal error """
  estr = traceback.format_exc()
  if not estr.startswith("None"):
    print(estr)
  if text:
    print(text)
  sys.exit(-1)

def parse():

  parser = optparse.OptionParser()

  parser.add_option('--dsid', dest='dsid',
    help='Specify dataset ID')
  parser.add_option('--dsname', dest='dsname',
    help='Specify dataset name (case insensitive')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if len(args) > 0:
    fatal('Extra options: %s' % str(args))

  return options

def show_roles_for_ds(roles, dcid, dsid):

  dcid = str(dcid)
  dsid = str(dsid)
  for r in roles:
    privs = []
    perms = set()
    for p in r.privList():
      if p.get('ptype') == 'dataset' and \
        ((p.get('dcid') == '-1') or \
         (p.get('dcid') == dcid and '-1' in p.get('dslist')) or \
         (p.get('dcid') == dcid and dsid in p.get('dslist'))):
        privs.append(p)
        perms.update(p.get('perms'))

    if perms:
      print("Role %s (%s) : %s" % \
        (r.id, r.name, ', '.join(PERMDEFS.get(p, p) for p in perms)))

def main():
  options = parse()

  if options.dsid:
    dslist = DataSet.objects.filter(id=int(dsid))
    if len(dslist) == 0:
      fatal("Dataset ID %s not found" % options.dsid)

  elif options.dsname:
    dslist = DataSet.objects.filter(dataset_name__iexact=options.dsname)
    if len(dslist) == 0:
      fatal("Dataset name '%s' not found" % options.dsname)
    if len(dslist) > 1:
      fatal("Multple datasets with name '%s' found (IDs %s), please specify with --dsid" % \
            (options.dsname, ', '.join(str(ds.id) for ds in dslist)))

  ds = dslist[0]
  roles = Role.objects.all()
  show_roles_for_ds(roles, ds.dataconnection_id, ds.id)

if __name__ == "__main__":
  main()

