# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import os
from django.conf import settings
from reports.ff_models import AppFeatureFlipper
from util.ff_default_flippers import DEFAULT_FLIPPERS
import logging

FLIPPING_LEVELS = {
    'DB': 2,  # User-visible features
    'SETTINGS': 3,  # So that the user can overwrite default value for hidden feature
    'URL': 4,  # For embedded use-cases that might be different from the full value
}

LOGGER = logging.getLogger('arcweb')


class FeatureFlipper:
    class __FeatureFlipper:
        def __init__(self):
            self.__url_flippers = []
            self.__settings_flippers = settings.FEATURE_FLIPPERS
            self.__db_flippers = self.__get_DB_flippers()
            self.__default_flippers = DEFAULT_FLIPPERS

            self.flippers = self.__get_all_flippers()
            LOGGER.info(f'Flipper re-instantiated')

        def get_flipper_value(self, feature_name):
            flipper = next(filter(lambda x: x.id == feature_name, self.flippers), None)

            if flipper:
                return flipper.value

            LOGGER.warning(f'Flipper does not exist: {feature_name}')

        def set_flipper(self, flipper_id, flipper_value, flipper_level):
            LOGGER.info(f'Flipper set {flipper_id} to {flipper_value}')
            flipper = next(filter(lambda x: x.id == flipper_id, self.flippers), None)

            if not flipper:
                LOGGER.warning(f'Flipper does not exist: {flipper_id}')
                return None

            flipper_level_num = FLIPPING_LEVELS[flipper_level]
            is_not_more_than_default = flipper_level_num <= flipper.get_default_flipper_level_num()

            if is_not_more_than_default and flipper_level_num >= flipper.current_level_num:
                flipper.value = flipper_value
                flipper.current_level_num = flipper_level_num

            if flipper_level == 'DB':
                (f, _) = AppFeatureFlipper.objects.get_or_create(id=flipper_id)

                if f.value == flipper_value:
                    return None

                f.value = flipper_value
                f.save()
            elif flipper_level == 'SETTINGS' and flipper.value != flipper_value:
                try:
                    with open(f'{os.environ.get("BASEDIR")}/settings_local.py', 'a') as site_settings:
                        site_settings.write("\n\nFEATURE_FLIPPERS.append({'id': '%s',  'value': %s})" % (flipper_id, flipper_value))
                except Exception as e:
                    LOGGER.warning(f'Error writing to setting file: {str(e)}')

        def set_url_flippers(self, url_flippers):
            self.__url_flippers = url_flippers
            for flipper in url_flippers:
                self.set_flipper(flipper['id'], flipper['value'], 'URL')

        def __get_all_flippers(self):
            flippers = []
            for flipper_name in self.__get_flippers_names():
                flipper = self.__pick_flipper(flipper_name)
                if flipper:
                    flippers.append(flipper)

            return flippers

        def __get_DB_flippers(self):
            return list(AppFeatureFlipper.objects.values())

        @staticmethod
        def __get_flipper(flippers_list, flipper_name):
            flipper = list(
                filter(lambda flipper: flipper['id'] == flipper_name, flippers_list)
            )
            if flipper:
                return flipper[-1]

            return None

        def __get_flippers_names(self):
            all_flippers_names = set(self.__flippers_to_list(self.__url_flippers))
            all_flippers_names.update(self.__flippers_to_list(self.__settings_flippers))
            all_flippers_names.update(self.__flippers_to_list(self.__db_flippers))
            all_flippers_names.update(self.__flippers_to_list(self.__default_flippers))

            return sorted(list(all_flippers_names))

        @staticmethod
        def __flippers_to_list(features_list):
            return [f['id'] for f in features_list]

        def __pick_flipper(self, flipper_name):
            default_flipper = self.__get_flipper(self.__default_flippers, flipper_name)
            if not default_flipper:
                return None

            flippers = [
                self.__db_flippers,
                self.__settings_flippers,
                self.__url_flippers,
            ]

            default_flipper = Flipper(default_flipper)
            i = default_flipper.get_default_flipper_level_num()
            while i > 1:
                flipper = self.__get_flipper(flippers[i - 2], flipper_name)
                if flipper:
                    flipper = Flipper(flipper)
                    flipper.add_flipper_parameters(default_flipper)
                    flipper.current_level_num = i
                    return flipper

                i -= 1

            return default_flipper

    instance = None

    def __new__(cls):
        if FeatureFlipper.instance is None:
            FeatureFlipper.instance = FeatureFlipper.__FeatureFlipper()

        return FeatureFlipper.instance


class Flipper:
    def __init__(self, flipper_dict):
        self.id = flipper_dict['id']
        self.value = flipper_dict['value']
        self.level = flipper_dict.get('level')
        self.current_level_num = 1
        self.label = flipper_dict.get('label', '')
        self.info = flipper_dict.get('info', '')
        self.dependents = flipper_dict.get('dependents', [])
        self.tab = flipper_dict.get('tab', '')

    def get_default_flipper_level_num(self):
        if self.level:
            return FLIPPING_LEVELS[self.level]

        return 1

    def add_flipper_parameters(self, default_flipper):
        self.label = default_flipper.label
        self.info = default_flipper.info
        self.level = default_flipper.level
        self.dependents = default_flipper.dependents
        self.tab = default_flipper.tab
