# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import functools
import time
import json
import threading
from django.http import HttpResponse
import logging

from arcwebbase import access_info
from util.safethreadpool import DjangoSafeThreadPoolExecutor

from util.perfcounters import APPROX_COUNTERS

LOGGER = logging.getLogger('arcweb')

calls_counter = APPROX_COUNTERS.get('http_async.calls')
retries_counter = APPROX_COUNTERS.get('http_async.retries')
tasks_counter = APPROX_COUNTERS.get('http_async.tasks')

class Async(object):
  """Make the django view async"""

  def __init__(self, name=None, concurrency=10, optimistic_return_check_seconds=0,
      buckets=None, request_timeout=120):
    self.name = name
    self.buckets = buckets
    self.concurrency = concurrency
    self.request_timeout = request_timeout
    self.optimistic_return_check_seconds = optimistic_return_check_seconds
    self.execPool = None
    self.lock = threading.Lock()
    self.funcsLen = 0
    self.funcs = {}

  def cullRequests(self):
    toDelete = []
    with self.lock:
      t1 = time.time()
      for k, v in self.funcs.items():
        (future, userid, t2) = v
        if abs(t2 - t1) > self.request_timeout:
          toDelete.append(k)

      for k in toDelete:
        del self.funcs[k]
        self.funcsLen -= 1

    for k in toDelete:
      LOGGER.info("Culled async request %s", k)

  def handleAsyncResults(self, request):
    if request.method != 'POST' and request.method != 'GET':
      raise ValueError("Invalid call")

    a = request.POST.get('async')
    if not a:
      return None

    global retries_counter
    retries_counter += 1
    asyncResultCall = json.loads(a)

    future = None
    futureId = asyncResultCall.get('id')
    if futureId:
      with self.lock:
        t = self.funcs.get(futureId)
        if t:
          (future, userid, _) = t
          self.funcs[futureId] = (future, userid, time.time())

      if t:
        (future, userid, _) = t

    # Note that if the id does not exist, or if it exists but for
    # a different user, we return exactly the same message. The intent
    # here is that we don't let an attacker know that a particular id
    # exists, but is for a different user
    if not future or (future and userid != request.user.id):
      d = {
        "type": "async",
        "error": "notfound"
      }

      return HttpResponse(json.dumps(d), content_type="text/json")

    if not future.done():
      d = {
        "type": "async",
        "id": futureId,
      }

      return HttpResponse(json.dumps(d), content_type="text/json")

    with self.lock:
      del self.funcs[futureId]
      self.funcsLen -= 1

    try:
      val = future.result()
    except Exception as e:
      LOGGER.exception(e)

      d = {
        "type": "async",
        "error": str(e)
      }

      return HttpResponse(json.dumps(d), content_type="text/json")

    return val

  def __call__(self, func):
    def inner(request, *args, **kwargs):
      global calls_counter
      calls_counter += 1

      # Because of django startup potentially happening via daemonize etc,
      # we create the actual threadpool at execution time as opposed to when
      # this instance is created
      if not self.execPool:
        with self.lock:
          if not self.execPool:
            self.execPool = DjangoSafeThreadPoolExecutor(name=self.name,
                max_workers=self.concurrency, buckets=self.buckets)

      val = self.handleAsyncResults(request)
      if val:
        return val

      global tasks_counter
      tasks_counter += 1

      schedulingThreadName = threading.current_thread().name

      @access_info.store_access_info
      def logContext():
        LOGGER.info("async request from '%s' %s %s",
            schedulingThreadName, request.method, request.build_absolute_uri())
        return func(request, *args, **kwargs)

      future = self.execPool.submit(logContext)
      futureId = id(future)

      cullRequests = False

      with self.lock:
        self.funcs[futureId] = (future, request.user.id, time.time())
        self.funcsLen += 1
        if self.funcsLen >= 100 and self.funcsLen % 100 == 0:
          cullRequests = True

      if cullRequests:
        self.cullRequests()

      if self.optimistic_return_check_seconds:
        # If results are expected to come in quickly, then just wait here and
        # return results immediately
        t1 = time.time()
        while abs(time.time() - t1) < self.optimistic_return_check_seconds:
          if future.done():
            val = future.result()
            return val

          time.sleep(0.1)

      d = {
        "type": "async",
        "id": futureId,
      }

      return HttpResponse(json.dumps(d), content_type="text/json")

    return inner

