# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import optparse
import os
import shutil
import sys
import time
import traceback
import gzip

from datetime import datetime
from tzlocal import get_localzone
from datasets.models import DataConnection
from dataconn.thrift_hs2.TCLIService.ttypes import TGlogServerLogReq
from dataconn.thrift_hs2.TCLIService.LogConnections.ttypes import TLogApplication
from dataconn.thrift_hs2.TCLIService.LogConnections.ttypes import TServerLogSharedReq
from dataconn.thrift_hs2.TCLIService.LogConnections.ttypes import TGlogServerLogInternalReq
from dataconn.thrift_hs2.TCLIService.Logging.ttypes import TLogLevel

from dataconn.dataconn import get_dataconn

class MaxBytesWrittenException(Exception):
  pass

def parse(options):
  default_options = {
    'duration': '3600',
    'dcname': 'default',
    'nodes': '',
    'levels': 'INFO, WARNING, ERROR, FATAL',
    'processes': 'CATALOGD, STATESTORED, ARCENGINED',
    'coordinators_only': '',
    'max_bytes_to_zip': '2000000000', # 2GB
    'buffer_size': '10000000',  # 10MB
    'header_size': '100000', # 100kB
    'host': 'localhost',
    'port': '21050'
  }

  if not any(options):
    parser = optparse.OptionParser(usage="")
    parser.add_option('--username', '-u', dest='username',
                      help='username')
    parser.add_option('--start_time', '-s', dest='start_time',
                      help='Specify start time of data (YYYY-MM-DD HH:MM:SS)')
    parser.add_option('--duration', '-d',  dest='duration', default=default_options.get('duration'),
                      help='Specify duration of data in seconds to fetch')
    parser.add_option('--dcname', dest='dcname', default=default_options.get('dcname'),
                      help='Data connection name')
    parser.add_option('--nodes', '-n', dest='nodes', default=default_options.get('nodes'),
                      help='Comma-delimited list of data nodes to fetch data')
    parser.add_option('--coordinators_only', '-c', dest='coordinators_only', default=default_options.get('coordinators_only'),
                      action='store_true',
                      help='Fetch from coordinator nodes only')
    parser.add_option('--log_levels', '-l', dest='levels', default=default_options.get('levels'),
                      help='List of log levels to include')
    parser.add_option('--processes', '-p', dest='processes', default=default_options.get('processes'),
                      help='List of processes to include')
    parser.add_option('--max_bytes_to_zip', '-m', dest='max_bytes_to_zip', default=default_options.get('max_bytes_to_zip'),
                      help='Maximum total bytes to zip (approximate)')
    parser.add_option('--query_id', '-q', dest='query_id', default='',
                      help=optparse.SUPPRESS_HELP)
    parser.add_option('--buffer_size', '-b', dest='buffer_size', default=default_options.get('buffer_size'),
                      help=optparse.SUPPRESS_HELP)
    parser.add_option('--header_size', dest='header_size', default=default_options.get('header_size'),
                      help=optparse.SUPPRESS_HELP)
    parser.add_option('--output_dir', '-o', dest='output_dir',
                      help='output directory to write data')
    parser.add_option('--usage', dest='usage', action='store_true',
                      help='Display usage text and exit')

    for option in parser.option_list:
      if option.default != ("NO", "DEFAULT") and option.help != optparse.SUPPRESS_HELP:
        option.help += (" " if option.help else "") + "[default: %default]"

    options, args = parser.parse_args()

    if options.usage:
      parser.print_help()
      sys.exit(0)

    if len(args) > 0:
      parser.error('Extra options: %s' % str(args))

    options = vars(options)

    # Adjust to handle timezone -- avoiding dateutil because of dst issues
    pattern = '%Y-%m-%d %H:%M:%S'
    local_tz = get_localzone()
    local_time = datetime.strptime(options.get('start_time'), pattern).astimezone(local_tz)
    options['start_time'] = local_time
  else:
    for key, value in default_options.items():
      if options.get(key) is None:
        options[key] = value

  return options

def getInitialRequest(options, log_level, app):
  shared_req = TServerLogSharedReq()
  internal_req = TGlogServerLogInternalReq()
  server_log_request = TGlogServerLogReq()
  internal_req.logLevel = log_level
  shared_req.timestamp = options.get('start_time').timestamp()
  shared_req.duration = int(options.get('duration'))
  if options.get('nodes') != "":
    shared_req.nodes = [node.strip() for node in options.get('nodes').split(',')]
  if options.get('coordinators_only'):
    shared_req.coordinators_only = True

  internal_req.app = app
  internal_req.sharedReq = shared_req
  server_log_request.internalReq = internal_req

  return server_log_request

def writeLog(options, file_bytes_written, dc, log_request, files, app):
  resp = None
  log_request.internalReq.sharedReq.maxBytes = int(options.get('header_size'))

  file_count = 1
  prev_file_name = ""
  tar_dir_name = 'tar_dir'
  duration_in_minutes = str(int(options.get('duration', '0')) / 60)
  timestamp_warning = 'This log file covers the timeframe from ' + options.get('start_time').isoformat() + ' UTC to '\
    + duration_in_minutes + ' minutes later. Note that the line level timestamps are in server time zone.\n\n'
  while resp is None or resp.sharedResp.continuationToken is not None:
    resp = dc.hs2call(options.get('username'), 'GetGlogServerLogs', log_request, False)
    if resp.sharedResp.fileName is None:
      break

    log_request.internalReq.sharedReq.maxBytes = int(options.get('buffer_size'))
    if resp.sharedResp.errorMessage != None:
      resp.sharedResp.continuationToken = None
    else:
      is_arcengine = TLogApplication._VALUES_TO_NAMES[app] == 'ARCENGINED'
      if is_arcengine:
        inner_dir_name = os.path.join(options.get('output_dir'), tar_dir_name, 'arcengined', resp.sharedResp.nodeName)
      else:
        inner_dir_name = os.path.join(options.get('output_dir'), tar_dir_name, resp.sharedResp.nodeName)

      os.makedirs(inner_dir_name, exist_ok=True)

      if resp.sharedResp.fileName != prev_file_name:
        suffix = ""
        file_count = 1
      else:
        suffix = '.%s' % str(file_count)
        file_count += 1

      prev_file_name = resp.sharedResp.fileName
      suffixed_file_name = '%s%s.gz' % (os.path.basename(resp.sharedResp.fileName), suffix)
      if is_arcengine:
        relative_path = os.path.join('arcengined', resp.sharedResp.nodeName, suffixed_file_name)
      else:
        relative_path = os.path.join(resp.sharedResp.nodeName, suffixed_file_name)
      output_file = os.path.join(options.get('output_dir'), tar_dir_name, relative_path)
      f = gzip.open(output_file, 'wb')

      if len(resp.sharedResp.logData) != 0:
        if 'header.txt' not in prev_file_name:
          f.write(bytes(timestamp_warning, 'utf8'))
        f.write(bytes(resp.sharedResp.logData, 'utf8'))

      f.close()
      files[relative_path] = f
      file_bytes_written += os.path.getsize(output_file)
      if file_bytes_written >= int(options.get('max_bytes_to_zip')):
        raise MaxBytesWrittenException()

      log_request.internalReq.sharedReq.continuationToken = resp.sharedResp.continuationToken

  return file_bytes_written

def fetchData(options, app, file_bytes_written, dc, files):
  levels = [level.strip().upper() for level in options.get('levels').split(',')]
  # range values represent TLogLevel::INFO to TLogLevel::ERROR
  for value in range(3,7):
    if TLogLevel._VALUES_TO_NAMES[value] in levels:
      request = getInitialRequest(options, value, app)
      file_bytes_written += writeLog(options, file_bytes_written, dc, request, files, app)

  return file_bytes_written

def main(dc=None, options=None, arcviz=False):
  if options is None:
    options = {}

  options = parse(options)

  maxBytesHit = False
  tar_dir_name = 'tar_dir'

  if dc is None:
    dcName = options.get('dcname')
    dcl = DataConnection.objects.filter(dataconnection_name=dcName)

    if len(dcl) == 0:
      print('Unable to load data connection %s' % dcName)
      return

    dc = get_dataconn(dcl[0].id)

  files = {}

  try:
    processes = [proc.strip().upper() for proc in options.get('processes').split(',')]
    dir_path = os.path.join(options.get('output_dir'), tar_dir_name)
    os.makedirs(dir_path, exist_ok=True)
    file_bytes_written = 0

    for app in ['CATALOGD', 'STATESTORED', 'ARCENGINED']:
      if app in processes:
        file_bytes_written += fetchData(options, TLogApplication._NAMES_TO_VALUES[app], file_bytes_written, dc, files)

  except MaxBytesWrittenException:
    maxBytesHit = True
    print("Hit maximum bytes")
  except Exception:
    if arcviz:
      raise
    else:
      traceback.print_exc()
      sys.exit(1)

  if not arcviz:
    archivename = 'logs.%s.tar.gz' % str(int(time.time()))
    archivepath = os.path.join(options.get('output_dir'), archivename)
    tmp_dir_path = os.path.join(options.get('output_dir'), tar_dir_name)
    shutil.make_archive(archivepath, 'tar', root_dir=options.get('output_dir'), base_dir=tar_dir_name)
    shutil.rmtree(tmp_dir_path)

  return {'files': files, 'maxBytesHit': maxBytesHit}

