# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import os
import sys
import shutil

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)

from django.conf import settings
import reports.models
from reports import views_appgroups
from datasets.models import DataConnection

import re
import json

west_segment = '{"name":"West","group":"Regions","filters":["[state] in (\\"AZ\\", \\"CA\\", \\"CO\\", \\"ID\\", \\"MO\\", \\"NM\\", \\"NV\\", \\"OR\\", \\"UT\\", \\"WA\\", \\"WY\\")"],"entities":[],"id":1,"dataset_id":%s}'
midwest_segment = '{"name":"Midwest","group":"Regions","filters":["[state] in (\\"IA\\", \\"IL\\", \\"IN\\", \\"KS\\", \\"MI\\", \\"ND\\", \\"NE\\", \\"OH\\", \\"SD\\", \\"WI\\")"],"entities":[],"dataset_id":%s}'
south_segment = '{"name":"South","group":"Regions","filters":["[state] in (\\"AK\\", \\"AL\\", \\"DE\\", \\"FL\\", \\"GA\\", \\"KY\\", \\"MA\\", \\"MS\\", \\"NC\\", \\"OK\\", \\"SC\\", \\"TN\\", \\"TX\\", \\"VA\\", \\"WV\\")"],"entities":[],"id":3,"dataset_id":%s}'
northeast_segment = '{"name":"Northeast","group":"Regions","filters":["[state] in (\\"CT\\", \\"MA\\", \\"MN\\", \\"NH\\", \\"NJ\\", \\"NY\\", \\"PA\\", \\"RI\\", \\"VT\\")"],"entities":[],"dataset_id":%s}'

DEFAULT_SQLITE_BASE = "/var/lib/arcadia"

def _get_samples_path():
  samplessqlite = os.path.join("samples", "samples.sqlite")

  app_path = os.path.dirname(__file__)
  samplefilepath_src = os.path.join(app_path, "..", samplessqlite)
  if os.path.isfile(samplefilepath_src):
    return samplefilepath_src

  if 'ARCADIA_DATA_PATH' in os.environ:
    # ARCADIA_DATA_PATH env is set for CDH installs
    samplefilepath_dst = os.path.join(os.environ['ARCADIA_DATA_PATH'],
                                      samplessqlite)
  elif os.path.exists(DEFAULT_SQLITE_BASE):
    samplefilepath_dst = os.path.join(DEFAULT_SQLITE_BASE, samplessqlite)
  else:
    samplefilepath_dst = os.path.join(
      os.path.dirname(os.path.abspath(settings.DATABASES['default']['NAME'])), \
      samplessqlite)

  dbpath = os.path.dirname(samplefilepath_dst)
  if not os.path.exists(dbpath):
    os.makedirs(dbpath)

  if getattr(sys, 'frozen', False):
    app_path = os.path.dirname(sys.executable)
    samplefilepath_src = os.path.join(app_path, samplessqlite)
  else:
    app_path = os.path.dirname(__file__)
    samplefilepath_src = os.path.join(app_path, "..", samplessqlite)

  print("Copying sample files from %s to %s" % (samplefilepath_src, samplefilepath_dst))
  try:
    shutil.copy(samplefilepath_src, dbpath)
  except Exception as e:
    print("Could not copy samples files from %s to %s" % (samplefilepath_src, samplefilepath_dst))
    print("Samples connection and apps will not work")
    samplefilepath_src = samplefilepath_dst
    pass

  return samplefilepath_dst

def create_sample_dataconnection():
  print('Creating a "samples" dataconnection')
  dcl = DataConnection.objects.filter(dataconnection_name='samples')
  if len(dcl) > 0:
    print('Data connection "samples" already exists')
    return None

  samplefilepath = _get_samples_path()

  dcinfo = {}
  dcinfo['PARAMS'] = {}
  dcinfo['PARAMS']['FILE'] = samplefilepath

  dc = DataConnection(
      dataconnection_type='sqlite',
      dataconnection_name='samples',
      dataconnection_description='auto-created samples',
      dataconnection_info = json.dumps(dcinfo))
  dc.save()
  return dc.id

def create_sample_group():
  print('Looking for and creating a "samples" group')
  return views_appgroups.appgroup_create_sample()

def migrate_dashboard_clicks():
  print("Migrating dashboard clicks")

  appid_re = re.compile('^uri-\d+-widget-(\d+)$')
  rl = reports.models.Report.objects.all()
  all_apps = {}
  dashboards_to_migrate = {}
  for r in rl:
    all_apps[r.id] = r
    if r.report_type == "dashboard":
      d = json.loads(r.report_data)
      if 'click_behavior' in d:
        # Does not need migration
        continue

      dashboard_apps = []
      for base_app in d['dashboard_widgets']:
        dashboard_apps.append(int(appid_re.match(base_app['id']).group(1)))
      dashboards_to_migrate[r.id] = (r, dashboard_apps)

  for k, v in dashboards_to_migrate.items():
    r, dashboard_apps = v
    print(k, r, dashboard_apps)

    dashboard_clicks = {}
    for appid in dashboard_apps:
      app_r = all_apps.get(appid)
      if not app_r:
        continue
      d = json.loads(app_r.report_data)
      app_click_behaviors = d['report_data'].get('click_behaviors')
      if not app_click_behaviors:
        continue
      for marker_name in app_click_behaviors:
        dashboard_clicks[appid] = {
          marker_name: True
        }
    print('before', r.report_data)
    db_d = json.loads(r.report_data)
    db_d['click_behavior'] = {
      'reports': dashboard_clicks
    }
    r.report_data = json.dumps(db_d)
    print('after', r.report_data)
    r.save()

def main():
  create_sample_dataconnection()
  create_sample_group()
  migrate_dashboard_clicks()
 
if __name__ == "__main__":
  main()

