# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import threading
import time
import datetime

import logging

LOGGER = logging.getLogger("arcweb")

#
# Note that for performance reasons, these perf counters are approximate,
# not exact.
# We would need to lock the _value modifiers below to make this exact
#

class _ApproxAtomicLong(object):
  def __init__(self, value=0):
    self._value = value

  @property
  def value(self):
    return self._value

  @value.setter
  def value(self, new):
    self._value = new

  def __iadd__(self, inc):
    self._value += inc
    return self

  def __isub__(self, dec):
    self._value -= dec
    return self

class _PerfCounters(object):
  def __init__(self):
    self.__lock = threading.Lock()
    self.__counters = {}
    self.__timeSeriesCounters = {}
    self.__histograms = {}

  def get(self, name):
    c = self.__counters.get(name)
    if c:
      return c

    l = _ApproxAtomicLong(0)

    with self.__lock:
      c = self.__counters.get(name)
      if not c:
        self.__counters[name] = l
        c = l

    return c

  def getAllCounters(self):
    d = {}

    with self.__lock:
      for k, v in self.__counters.items():
        d[k] = v.value

    return d

  def getTimeSeriesCounter(self, name):
    with self.__lock:
      return self.__timeSeriesCounters.get(name)

  def setTimeSeriesCounter(self, name, c):
    with self.__lock:
      self.__timeSeriesCounters[name] = c

  def getAllTimeSeriesCounters(self):
    d = {}

    with self.__lock:
      for k, v in self.__timeSeriesCounters.items():
        d[k] = v

    for k in d:
      d[k] = d[k].data()

    return d

  def getHistogram(self, name=None, buckets=None):
    h = self.__histograms.get(name)
    if h:
      return h

    newh = Histogram(buckets=buckets)

    with self.__lock:
      h = self.__histograms.get(name)
      if h:
        return h

      self.__histograms[name] = newh
      return newh

  def getAllHistograms(self):
    d = {}
    with self.__lock:
      for k, v in self.__histograms.items():
        d[k] = v

    for k in d:
      d[k] = d[k].histogram()

    return d

APPROX_COUNTERS = _PerfCounters()

#
# This class is used to track counters over time: for example, to track the
# number of http requests we are getting per minute, and to do it for a few hours.
#
# We do this by maintaining an array of counters, and tracking the location which
# tracks the current time bucket. As the bucket needs to change, we use the next
# bucket (in a rolling manner). We protect bucket position changes with a lock
# for sane retrieval, but the actual counters are approximate in multi-threaded
# environments, like other counters in this file.
#
class TimeSeriesCounter(object):
  def __init__(self, name, timePeriodSeconds=60, timePeriodRetentionSeconds=60*60*24):
    self._name = name
    self._timePeriodSeconds = timePeriodSeconds
    self._timePeriodRetentionSeconds = timePeriodRetentionSeconds
    self._max = int(timePeriodRetentionSeconds / timePeriodSeconds)
    self._values = [0] * self._max

    # self._lock is used for protecting changes to _pos and _posTime.
    # It is never used to protect self._values, which means all counts
    #   are approximate
    self._lock = threading.Lock()
    self._pos = 0
    self._posTime = round(time.time() / self._timePeriodSeconds)
    self._rolloverCount = 0
    self._rolloverTotal = 0
    APPROX_COUNTERS.setTimeSeriesCounter(self._name, self)

  def convertTs(self, tsp):
    return datetime.datetime.fromtimestamp(tsp * self._timePeriodSeconds) \
        .strftime("%Y-%m-%d %H:%M:%S.%f")

  def data(self):
    with self._lock:
      return {
        'pos': self._pos,
        'posTime': self.convertTs(self._posTime),
        'values': self._values[:],
        'rolloverCount': self._rolloverCount,
        'rolloverTotal': self._rolloverTotal,
      }

  def __iadd__(self, inc):
    t = round(time.time() / self._timePeriodSeconds)

    # Check if we are still in the same time period
    if t != self._posTime:
      with self._lock:
        diff = t - self._posTime
        if diff != 0:
          p = self._pos

          if self._values[p] > 0:
            LOGGER.info("ts %s %s %s", self._name, self.convertTs(self._posTime), self._values[p])

          for i in range(1, diff + 1):
            p = (p + 1) % self._max
            self._rolloverTotal += self._values[p]
            self._values[p] = 0
            if p == 0:
              self._rolloverCount += 1

          self._pos = p
          self._posTime = t

    self._values[self._pos] += inc
    return self

class Histogram(object):
  def __init__(self, buckets=None):
    self._histogram = {}
    self._buckets = buckets[:]
    for b in self._buckets:
      self._histogram[b] = 0

  def histogram(self):
    return self._histogram

  def getBucket(self, t):
    for b in self._buckets:
      if t < b:
        return b

    return self._buckets[-1]

  def addToBucket(self, bucket):
    self._histogram[bucket] += 1
