# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import inspect
import time
import pprint
from functools import wraps
from django.utils.termcolors import colorize
from django.db import connection

import pprint
import logging

LOGGER = logging.getLogger('arcweb')

def perf_wrapper(count_queries=True,
                 show_queries=False,
                 skip_if_no_queries=False,
                 repeat=1,
                 fg_color="yellow",
                 bg_color="black"):
  """
  This is a decorator function that you can use to check the timing
  and Django DB access of a function.  It will print elapsed time
  as well as total query time and number of queries.
  To use it simply put "@perf_wrapper()" above the function def
  :param count_queries: If true count queries to Django DB
  :param show_queries: If true print a JSON string with all the queries
  :param skip_if_no_queries: If True don't print anything if the function had
                             no queries
  :param repeat: Run the function multiple times and print a summary.
                 Do *not* use this for functions that change state
                 i.e. write to the database.
  :param fg_color: foreground color - see django.utils.termcolors
  :param bg_color: background color - see django.utils.termcolors
  """
  if show_queries:
    import json

  def outer(func):
    @wraps(func)
    def inner(*args, **kwargs):
      starttime = []
      startquerycount = []
      endtime = []
      querytime = []
      endquerycount = []
      elapsedtime = []
      delta_query_count = []
      total_starttime = time.time()
      # Run func and collect data
      for count in range(0, repeat):
        starttime.append(time.time())
        if count_queries:
          startquerycount.append(len(connection.queries))
        func_returnval = func(*args, **kwargs)  # execute func
        endtime.append(time.time())
        elapsedtime.append(endtime[count] - starttime[count])
        if count_queries:
          endquerycount.append(len(connection.queries))
          delta_query_count.append(
            endquerycount[count] - startquerycount[count])
          querytime.append(sum((float(q['time']) for q in
                                connection.queries[
                                startquerycount[count]:endquerycount[count]])))
        # Print data
        if count_queries:
          if not skip_if_no_queries or total_delta_query_count > 0:
            print("Starttime: %s Endtime: %s" % (starttime[count], endtime[count]))
            print((colorize("%s()%s Time: %.4f  Query Time: %.4f   # Queries: %d" %
                           (func.__name__, "#%02d" % count if repeat > 1 else "",
                            elapsedtime[count], querytime[count],
                            delta_query_count[count]),
                           fg=fg_color, bg=bg_color)))
        else:
          print((colorize("%s()%s Time: %.4f" %
                         (func.__name__, "#%02d" % count if repeat > 1 else "",
                          elapsedtime[count]), fg=fg_color, bg=bg_color)))

      # Calculate totals
      total_endtime = time.time()
      total_elapsedtime = total_endtime - total_starttime
      if count_queries:
        total_delta_query_count = sum(delta_query_count)
        total_querytime = sum(querytime)

      # print totals
      if repeat > 1:
        if count_queries:
          print((
          colorize("%s() Totals: Time: %.4f  Query Time: %.4f   # Queries: %d" %
                   (func.__name__,
                    total_elapsedtime, total_querytime,
                    total_delta_query_count),
                   fg=fg_color, bg=bg_color, opts=('bold',))))
          print((
          colorize("%s() Average: Time: %.4f  Query Time: %.4f   # Queries: %d" %
                   (func.__name__,
                    total_elapsedtime / repeat, total_querytime / repeat,
                    total_delta_query_count / repeat),
                   fg=fg_color, bg=bg_color, opts=('bold',))))
        else:
          print((colorize("%s() Total Time: %.4f" %
                         (func.__name__, total_elapsedtime),
                         fg=fg_color, bg=bg_color, opts=('bold',))))
          print((colorize("%s() Average Time: %.4f" %
                         (func.__name__, total_elapsedtime / repeat),
                         fg=fg_color, bg=bg_color, opts=('bold',))))
      if show_queries:
        print((json.dumps(
          connection.queries[startquerycount[count]:endquerycount[count]])))

      return func_returnval

    return inner

  return outer


def perf_mark(label="",
              count_queries=True,
              show_queries=False,
              fg_color="magenta",
              bg_color="black"):
  """
  This is a simple way of instrumenting/timing sections of code.  Simply
  call perf_mark() before and after the section you are interested in.
  Subsequent calls to perf_mark() will give stats from the first the call plus
  delta stats from the previous call.
  :param label: Optional string to print to help identify the mark
  :param count_queries: Count the Django DB queries
                        *must* either be set or unset for all calls in function
  :param show_queries: Show the DB queries since last mark
  :param fg_color: foreground color - see django.utils.termcolors
  :param bg_color: background color - see django.utils.termcolors
  """
  if show_queries:
    import json
  frame = inspect.currentframe()
  caller = frame.f_back  # find our caller
  callerinfo = inspect.getframeinfo(caller)  # get caller info
  # tricky: store our data in the callers namespace
  perfinfo = caller.f_locals.setdefault("_arc_perfinfo",
                                        {"count": 0, "time": [], "lineno": [],
                                         "querycount": [], "querytime": [],
                                         "totalquerytime": 0.0})
  count = perfinfo['count']
  lastcount = max(count - 1, 0)
  perfinfo['time'].append(time.time())
  perfinfo['lineno'].append(callerinfo[1])
  if count_queries:
    perfinfo['querycount'].append(len(connection.queries))
    perfinfo['querytime'].append(sum((float(q['time']) for q in
                                      connection.queries[
                                      perfinfo['querycount'][lastcount]:
                                      perfinfo['querycount'][count]])))
    perfinfo['totalquerytime'] = perfinfo['totalquerytime'] + \
                                 perfinfo['querytime'][count]
  perfinfo['count'] = perfinfo['count'] + 1
  if count == 0:
    label = label if label else "** START **"
  LOGGER.debug(colorize(
    "%-40s %s %s" % ("%s[%s]#%02d %s" % (callerinfo[2], callerinfo[1], count, label),
                     "Time: %.4f +%.4f" % (
                       # perfinfo['time'][count] - perfinfo['time'][0],
                       # perfinfo['time'][count] - perfinfo['time'][lastcount]),
                       perfinfo['time'][len(perfinfo['time'])-1] - perfinfo['time'][0],
                       perfinfo['time'][len(perfinfo['time'])-1] - perfinfo['time'][lastcount]),
                     " Query Time %.4f +%.4f  Queries: %i +%i" % (
                       perfinfo['totalquerytime'], perfinfo['querytime'][count],
                       perfinfo['querycount'][count] - perfinfo['querycount'][
                         0],
                       perfinfo['querycount'][count] - perfinfo['querycount'][
                         lastcount])
                     if count_queries else "",
                     ), fg=fg_color, bg=bg_color))
  if show_queries and count_queries:
    for q in connection.queries[perfinfo['querycount'][lastcount]:perfinfo['querycount'][count]]:
      LOGGER.debug('Time:  %s' % q.get('time', 0))
      LOGGER.debug('Query: %s' % q.get('sql', '??'))
