#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Utility to do some basic permission tests

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

USAGE = "Test program"

import os
import sys
import json
import optparse
import datetime
import traceback

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)
sys.path.append("%s/../vizlib" % THISDIR)

from datasets.models import DataSet, DataConnection
from rbac.permdefs import PERMDEFS
from django.contrib.contenttypes.models import ContentType
from django.db.models.base import ObjectDoesNotExist
from django.contrib.auth.models import User, Group, Permission
from reports.views import report_get
from datasets.views import _dataset_get

def fatal(text):
  """ Fatal error """
  estr = traceback.format_exc()
  if not estr.startswith("None"):
    print(estr)
  if text:
    print(text)
  sys.exit(-1)

def parse():

  parser = optparse.OptionParser(usage="")

  parser.add_option('--username', dest='username', default='admin',
    help='Perform data operations as this user')

  parser.add_option('--groupname', dest='groupname',
    help='Specify group name for testing')
  
  parser.add_option('--dcid', dest='dcid', default=1, type='int',
    help='Specify data connection ID')

  parser.add_option('--perm', dest='perm',
    help='Specify permission to test for (one of %s)' % list(PERMDEFS.keys()))

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if len(args) > 0:
    fatal('Extra options: %s' % str(args))

  user = None
  try:
    user = User.objects.get(username=options.username)
  except ObjectDoesNotExist as e:
    sys.exc_clear()
  if not user:
    fatal("Username '%s' not found" % options.username)
  options.user = user

  dc = None
  try:
    dc = DataConnection.objects.get(id=options.dcid)
  except ObjectDoesNotExist as e:
    sys.exc_clear()
  if not dc:
    fatal("Data connection ID '%d' not found" % options.dcid)
  options.dc = dc

  group = None
  if options.groupname:
    try:
      group = Group.objects.get(name=options.groupname)
    except ObjectDoesNotExist as e:
      sys.exc_clear()
    if not group:
      fatal("Groupname '%s' not found" % options.groupname)
  options.group = group

  if options.perm:
    if options.perm not in list(PERMDEFS.keys()):
      fatal("Permission must be one of %s, not '%s'" % (list(PERMDEFS.keys()), options.perm))

  return options

def main():

  options = parse()

  user = options.user
  dcid = options.dcid
  perm = options.perm

  if perm:
    dc = options.dc
    allowed = dc.check_permission(options.user, perm)
    print("Permission '%s' %s for user '%s' on Data connection %d" % \
            (perm, 'GRANTED' if allowed else 'DENIED', options.username, dcid))
    sys.exit();

  perm = 'test'
  dataconn_ct = ContentType.objects.get_for_model(DataConnection)
  dataset_ct = ContentType.objects.get_for_model(DataSet)

  print("For User: ")
  print("Content_type app_label", type(options.user)._meta.app_label)

  print("For DataConnection: ")
  print("Content_type id: ", dataconn_ct.id)
  print("Content_type app_label", DataConnection._meta.app_label)

  print("For DataSet: ")
  print("Content_type id: ", dataset_ct.id)
  print("Content_type app_label", DataSet._meta.app_label)

  perm_code_name = "%s_%s_%s" % ("can", perm, dcid)
  perm_name = "%s %s %s" % ("Can", perm, dcid)
  print("Create test permission: '%s' ..." % perm_code_name)

  try:
    permission = Permission.objects.get(codename=perm_code_name, content_type_id=dataconn_ct.id)
    print("Permission already exists")
  except Permission.DoesNotExist:
    print("Permission not found...creating")
    permission = Permission.objects.create(codename=perm_code_name,
                                           name=perm_name,
                                           content_type=dataconn_ct)
    permission.save()


  app_label = DataSet._meta.app_label
  print("Testing access for app_label: %s" % app_label)

  if user.has_perm(app_label + "." + perm_code_name):
    print("Granted")
  else:
    print("Not granted")
    print("Adding permission")
    user.user_permissions.add(permission)
    user = User.objects.get(username=options.username)
    if user.has_perm(app_label + "." + perm_code_name):
      print("Ok, now granted")
    else:
      print("Hmmm...still not granted")

  print("Get dc permissions")
  pl = Permission.objects.filter(codename__contains='can_test_')
  for p in pl:
    print("  id = ", p.id)
    print("  name = ", p.name)
    print("  content_type_id = ", p.content_type_id)
    print("  codename = ", p.codename)

  print("pl = %s" % str(pl))
      
  ul = User.objects.filter(user_permissions__codename__regex=r'can_te.*_1')
  print("ul = %s" % str(ul))

  gl = Group.objects.filter(permissions__codename__regex=r'can_readwrite_.*')
  print("gl = %s" % str(gl))

  pl = Permission.objects.filter(user__username=options.username, codename__regex=r'can_te.*_1')
  print("pl = %s" % str(pl))


  print("Removing permission...")
  user.user_permissions.remove(pl[0])
  print("Done...")
  ul = User.objects.filter(user_permissions__codename__regex=r'can_te.*_1')
  print("ul = %s" % str(ul))

  ul = User.objects.filter(username__in=None)
  print("Final ul = %s" % str(ul))
 
  sys.exit(0)

if __name__ == "__main__":
  main()

