#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Utility to do some basic permission tests

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

USAGE = "Test program"

import os
import sys
import json
import optparse
import datetime
import traceback

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)
sys.path.append("%s/../vizlib" % THISDIR)

from datasets.models import DataSet, DataConnection
from django.db.models.base import ObjectDoesNotExist
from django.contrib.auth.models import User, Group, Permission
from django.conf import settings
from rbac.permdefs import PERMDEFS, PermissionException
from rbac.permcache import RBAC
from rbac.migrateperms import migratePerms, defaultRoles
from rbac.models import Role

def fatal(text):
  """ Fatal error """
  estr = traceback.format_exc()
  if not estr.startswith("None"):
    print(estr)
  if text:
    print(text)
  sys.exit(-1)

def parse():

  parser = optparse.OptionParser(usage="")

  parser.add_option('--username', dest='username', default='admin',
    help='Perform data operations as this user')

  parser.add_option('--groupname', dest='groupname',
    help='Specify group name for testing')
  
  parser.add_option('--dcname', dest='dcname',
    help='Specify data connection name')

  parser.add_option('--dsname', dest='dsname',
    help='Specify dataset name');

  parser.add_option('--perm', dest='perm',
    help='Specify permission to test for (one of %s)' % list(PERMDEFS.keys()))

  parser.add_option('--migrate', dest='migrate', action='store_true',
    help='Run migration algorithm')

  parser.add_option('--defroles', dest='defroles', action='store_true',
    help='Setup default roles')

  parser.add_option('--reset', dest='reset', action='store_true',
    help='Delete all existing roles before migrate and/or default creation')

  parser.add_option('--dryrun', dest='dryrun', action='store_true',
    help='For migrate, just print roles that would be created')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if len(args) > 0:
    fatal('Extra options: %s' % str(args))

  user = None
  try:
    user = User.objects.get(username=options.username)
  except ObjectDoesNotExist as e:
    sys.exc_clear()
  if not user:
    fatal("Username '%s' not found" % options.username)
  options.user = user

  dc = None
  if options.dcname:
    try:
      dc = DataConnection.objects.get(dataconnection_name=options.dcname)
    except ObjectDoesNotExist as e:
      sys.exc_clear()
    if not dc:
      fatal("Data connection '%s' not found" % options.dcname)
  options.dc = dc

  ds = None
  if options.dsname:
    try:
      if dc:
        ds = DataSet.objects.get(dataset_name=options.dsname,dataconnection_id=dc.id)
      else:
        ds = DataSet.objects.get(dataset_name=options.dsname)
    except ObjectDoesNotExist as e:
      sys.exc_clear()
    if not ds:
      fatal("Dataset '%s' not found" % options.dsname)
  options.ds = ds

  group = None
  if options.groupname:
    try:
      group = Group.objects.get(name=options.groupname)
    except ObjectDoesNotExist as e:
      sys.exc_clear()
    if not group:
      fatal("Groupname '%s' not found" % options.groupname)
  options.group = group

  if options.perm:
    if options.perm not in list(PERMDEFS.keys()):
      fatal("Permission must be one of %s, not '%s'" % (list(PERMDEFS.keys()), options.perm))

  return options

def main():

  options = parse()

  if options.migrate or options.defroles:
    if options.reset:
      Role.objects.all().delete()

    if options.migrate:
      migratePerms(options.dryrun)

    if options.defroles:
      defaultRoles()

    sys.exit(0)

  if options.reset:
    fatal("--reset is only valid with --defroles or --migrate")

  user = options.user
  perm = options.perm
  obj = None
  name = 'System'
  if options.ds:
    obj = options.ds
    name = 'DataSet ' + obj.dataset_name
  elif options.dc:
    obj = options.dc
    name = 'DataConnection ' + obj.dataconnection_name

  if not perm:
    perms = RBAC.get_user_obj_perms(user, obj)
    print("Perms for %s on %s: %s" % (user.username, name, perms))
  else:
    try:
      RBAC.check(user, perm, obj)
      print("User %s has permission %s on %s" % (user.username, perm, name))
    except PermissionException as e:
      print("User %s does not have permission %s on %s: %s" % \
            (user.username, perm, name, str(e)))

  sys.exit();

if __name__ == "__main__":
  main()

