# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from collections import defaultdict
import os
import json
from datetime import datetime
import os
import json
from usage.models import UsageDefaults
from reports.views_settings import get_settings
import vizlib.querylog as querylog
import logging
from datasets.models import DataSet
from util.ff_flipper import FeatureFlipper

LOGGER = logging.getLogger('arcweb')

class QueryLogSummarizer:
    def __init__(self):
      self.summary_data: UsageDefaults = [{
        'version': 1,
        'time': {'Hour': 0, 'Day': 0, 'Month': 0, 'Year': 0},
        'domain': '',
        'connectionTypes': defaultdict(int),
        'visualTypes': defaultdict(int),
        'datasetTypes': {'numberOfTables': defaultdict(int), 'sql': 0, 'table': 0},
        'numberUsers': 0,
        'userTotalTime': defaultdict(int),
        'non_visual_page': {},
        'non_visual_page_load_sec': {},
        'errors': {'visual': 0, 'query': 0, 'python': 0, 'js': 0},
        'query_load_sec': {1: 0, 5: 0, 10: 0, 30: 0, 60: 0, 120: 0},
        'features': {
          "featureFlippers": {},
          "completions": {},
          "embeddings": {},
        }
      }]

      self.current_datetime = datetime.now()
      self.unique_user_ids = set()
      self.unique_dataset_ids = set()
      self.sql_keywords = ['SELECT', 'INSERT', 'UPDATE', 'DELETE', 'FROM', 'WHERE', 'JOIN', 'GROUP BY', 'ORDER BY']
      self.temp_component_id = None

    @staticmethod
    def get_dataset_info(dataset_id):
      try:
        dataset_row = DataSet.objects.filter(id=dataset_id).values('dataset_detail', 'dataset_info').first()

        return dataset_row
      except DataSet.DoesNotExist:
        return None

    def summarize(self, queryDicts, user_sessions):
      self._set_time()
      self._set_domain()
      self.collect_profile_settings_data()

      if user_sessions:
        self._process_user_sessions(user_sessions)

      for date_time_key in queryDicts.keys():
        queryDictList = queryDicts[date_time_key]

        for queryDict in queryDictList:
          self._process_query_dict(queryDict)

      return self.summary_data

    def _set_time(self):
        self.summary_data[0]['time'] = {
          'Hour': self.current_datetime.hour,
          'Day': self.current_datetime.day,
          'Month': self.current_datetime.month,
          'Year': self.current_datetime.year,
        }

    def _set_domain(self):
      self.summary_data[0]['domain'] = os.environ['CDSW_DOMAIN']

    def _process_user_sessions(self, user_sessions):
      for user_id, session_duration in user_sessions:
        user_id_str = str(user_id)
        self.summary_data[0]['userTotalTime'][user_id_str] += session_duration

    def _process_query_dict(self, queryDict):
      connection_type = queryDict.get('dataconnection_type')
      connection_id = str(queryDict.get('dataconnection'))
      if connection_type and connection_type:
        if connection_type not in self.summary_data[0]['connectionTypes']:
          self.summary_data[0]['connectionTypes'][connection_type] = {}
        self.summary_data[0]['connectionTypes'][connection_type][connection_id] = \
          self.summary_data[0]['connectionTypes'][connection_type].get(connection_id, 0) + 1

      # visualTypes
      visual_type = queryDict.get('visual_type')
      if visual_type:
        self.summary_data[0]['visualTypes'][visual_type] += 1

      # datasetTypes
      dataset_id = queryDict.get('dataset')

      if dataset_id and dataset_id not in self.unique_dataset_ids:
        self.unique_dataset_ids.add(dataset_id)

        dataset_row = self.get_dataset_info(dataset_id)

        if dataset_row:
          if dataset_row['dataset_detail'] == 'Created from SQL':
            self.summary_data[0]['datasetTypes']['sql'] += 1
          else:
            self.summary_data[0]['datasetTypes']['table'] += 1

          dataset_info_list = json.loads(dataset_row['dataset_info'])
          tablename_count = 0

          for dataset_info in dataset_info_list:
            if 'tablename' in dataset_info:
              tablename_count += 1

            current_tablename_count = self.summary_data[0]['datasetTypes']['numberOfTables'].get(str(tablename_count), 0)

            self.summary_data[0]['datasetTypes']['numberOfTables'][str(tablename_count)] = current_tablename_count + 1

            if str(tablename_count) not in self.summary_data[0]['datasetTypes']['numberOfTables']:
              self.summary_data[0]['datasetTypes']['numberOfTables'][str(tablename_count)] = 1

      # numberUsers
      user_id = queryDict.get('user_id')
      if user_id and user_id not in self.unique_user_ids:
        self.summary_data[0]['numberUsers'] += 1
        self.unique_user_ids.add(user_id)

      # errors
      component_id = queryDict.get('component_id')
      query = queryDict.get('query')
      error = queryDict.get('error')

      if error:
        if self.temp_component_id is None:
          self.temp_component_id = component_id
          self.summary_data[0]['errors']['python'] += 1
          if any(keyword in query.upper() for keyword in self.sql_keywords):
            self.summary_data[0]['errors']['query'] += 1
        else:
          if self.temp_component_id != component_id:
            self.temp_component_id = component_id
            self.summary_data[0]['errors']['python'] += 1
            if any(keyword in query.upper() for keyword in self.sql_keywords):
              self.summary_data[0]['errors']['query'] += 1

      # query_load_sec
      runtime_str = queryDict['runtime']
      runtime = float(runtime_str) if runtime_str else 0.0
      runtime_ranges = [(1, 5), (5, 10), (10, 30), (30, 60), (60, 120)]
      query_load_sec_keys = [1, 5, 10, 30, 60, 120]

      for idx, (lower, upper) in enumerate(runtime_ranges):
        if lower <= runtime <= upper:
          self.summary_data[0]['query_load_sec'][query_load_sec_keys[idx]] += 1
          break

    def collect_profile_settings_data(self):
      feature_flipper = FeatureFlipper()
      all_flippers = {f.id: f.value for f in feature_flipper.flippers}

      def get_profile_settings(setting_key):
        settings = get_settings(True, setting_key)
        setting_value = json.loads(settings[0]["setting_value"])
        return setting_value.get("profileSettings", {})

      completions_profile_settings = get_profile_settings('ai_completion_settings')
      embeddings_profile_settings = get_profile_settings('ai_embeddings_settings')

      completions = self.process_profile_settings(profile_settings=completions_profile_settings, engine_key="completion_engine", stream_key="completion_stream")
      embeddings = self.process_profile_settings(profile_settings=embeddings_profile_settings, engine_key="embeddings_engine")

      self.summary_data[0]['features'] = {
        "featureFlippers": all_flippers,
        "completions": completions,
        "embeddings": embeddings
      }

    def process_profile_settings(self, profile_settings, engine_key, stream_key=None):
      result = {}

      if not isinstance(profile_settings, dict):
        print(f"Error: Expected a dictionary, but got {type(profile_settings)}")
        return result

      for profile_id, profile_data in profile_settings.items():
        values = profile_data.get("values", {})
        engine = values.get(engine_key)

        if engine:
          if engine not in result:
            result[engine] = {"count": 0}

          result[engine]["count"] += 1

          if values.get(stream_key) is True:
            if "stream" not in result[engine]:
              result[engine]["stream"] = 0
            result[engine]["stream"] += 1

      return result
