#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Utility to exercise the arcengine view recommendation path from input query

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

USAGE = """
   Use this utility to exercise the recommendation engine.
   e.g., ./arcviz util recommendation --dcname=<data connection name> --file=<path to the file containing input queries>
"""

import os
import sys
import json
import pprint
import getpass
import optparse
import string
import traceback

from dataconn.data_api_cmds import SPECIAL_CMDS
from optparse import OptionValueError
from vizlib.dataset_request import DatasetRequest

import warnings
with warnings.catch_warnings():
  warnings.filterwarnings("ignore", category=DeprecationWarning)

from django.db.models.base import ObjectDoesNotExist
from django.contrib.auth.models import User

from datasets.models import DataConnection
from dataconn.dataconn import get_dataconn
from dataconn.connections_types import DATACONN_ARCENGINE, isArcEngineBased

from dataconn.thrift_hs2.TCLIService.ttypes import TGetAnalyticalViewsRecommendationsReq
from vizlib.accel import get_existing_tables
# Turn off the logging in the django parts (like hs2conn.py)
import logging
logging.disable(logging.DEBUG)

LINE_SEPARATOR_COUNT = 150


def read_queries(query_file):
  f = open(query_file, 'r')
  queries = []
  s = ""
  for line in f:
    if not str(line).startswith('--'):
      s += str(line).strip('\n').strip() + ' '
    if s.strip().endswith(';'):
      queries.append(s)
      s = ''
  return queries

def parse():
  parser = optparse.OptionParser(usage="")
  parser.add_option(
      '--query',
      dest='queries',
      action='append',
      help='Specify a query to recommend (can specify more than one)')
  parser.add_option('--file', dest='filename',
                    help='Read queries from the FILE', metavar="FILE")
  parser.add_option('--dcname', dest='dcname', default='default',
                    help='Data connection name')
  parser.add_option('--user', dest='username', default='admin',
                    help='Perform data operations as this user')
  parser.add_option('--usage', dest='usage', action='store_true',
                    help='Display usage text and exit')
  parser.add_option(
      '-r',
      '--recommendation_only',
      dest='only_recommendations',
      action='store_true',
      help='Display only the create statements for the recommended aviews')
  parser.add_option(
      '-v',
      '--verbose',
      dest='verbose',
      action='store_true',
      help='Display additional information about the recommended aviews')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if options.usage:
    print (USAGE)
    parser.print_help()
    sys.exit(0)

  if len(args) > 0:
    parser.error('Extra options: %s' % str(args))

  if not options.queries and not options.filename:
    parser.error('No queries specified')
  elif options.filename:
    if not options.queries:
      options.queries = read_queries(options.filename)
    else:
      options.queries += read_queries(options.filename)

  user = None
  try:
    user = User.objects.get(username=options.username)
  except ObjectDoesNotExist as e:
    raise Exception("Username '%s' not found" % options.username)
  options.userid = user.pk
  options.user = user

  # Get data connection
  dcl = DataConnection.objects.filter(dataconnection_name=options.dcname)
  err = ' name %s' % options.dcname

  if len(dcl) == 0:
    parser.error("Unable to load data connection %s" % err)

  dc = dcl[0]
  if not isArcEngineBased(dc.dataconnection_type):
    parser.error("Dataconnection must be %s not %s" %
          (DATACONN_ARCENGINE, dc.dataconnection_type))

  options.dataconn = dc
  return options

# Get the existing analytical views in the system
def get_existing_aviews(dataconn, options):
  existing_view_data = {}
  get_existing_tables({'type': SPECIAL_CMDS['SHOW_AVIEWS']}, existing_view_data, options.user, dataconn, True)
  existing_view_names = list(existing_view_data.keys())
  return existing_view_names

# Find an available view name


def get_generated_name(existing_view_names, generated_names, name):
  fullname = ""
  vnum = 0
  while True:
    vnum = vnum + 1
    fullname = name[: (name.rfind('_') + 1)] + str(vnum)
    if fullname not in existing_view_names and fullname not in generated_names:
      break
  return str(fullname)


def aview_name_to_recommendation_map(output, existing_view_names):
  output = json.loads(output)
  aviewname_to_recommendations = {}
  recommendation_to_name = {}
  recommendations = set()
  no_recommendation = False
  generated_names = []
  for recom in output:
    data = json.loads(recom[2])
    for i in range(0, len(data)):
      if data[i]["query"]["recommendation_status"] == "RECOMMENDATION" and data[i]["aview"]["definition"] not in recommendations:
        recommendations.add(data[i]["aview"]["definition"])
        name = data[i]["aview"]["name"]
        generated_name = get_generated_name(existing_view_names, generated_names, name)
        generated_names.append(generated_name)
        recommendation_to_name[data[i]["aview"]["definition"]] = generated_name
        create_statement = data[i]["aview"]["definition"].replace(
            name, generated_name)
        aviewname_to_recommendations[generated_name] = create_statement
      elif data[i]["query"]["recommendation_status"] in ("NO_RECOMMENDATION", "UNKNOWN"):
        no_recommendation = True
  return (
      aviewname_to_recommendations,
      recommendation_to_name,
      no_recommendation)


def get_input_queries(output):
  output = json.loads(output)
  queries = []
  for recom in output:
    queries.append(recom[0].strip('"'))
  return queries

def get_query_object_map(output):
  output = json.loads(output)
  query_objects = {}
  for recom in output:
    query_objects[recom[0].strip('"')] = recom[1]
  return query_objects

def print_query_objects(query_objects, query):
  compute_stat_required = []
  for i in range(0, len(query_objects[query])):
    if query_objects[query][i]["is_an_lview"]:
      print (
          "      -- LV: " +
          query_objects[query][i]["name"] +
          " is a logical view on top of " +
          ", ".join(
              query_objects[query][i]["lview_basetables"]))
    else:
      print (
          "      -- BT: " +
          query_objects[query][i]["name"] +
          " is a base table in " +
          query_objects[query][i]["file_format"] +
          " format.")
    offset = "             "
    stats = []
    size_unit = ""
    if int(query_objects[query][i]["row_size"]) >= 0:
      size_unit = " B"
    else:
      compute_stat_required.append(query_objects[query][i]["name"])
    stats.append("Row Count: " + str(query_objects[query][i]["row_count"]))
    stats.append(
        "Row Size: " +
        str(
            query_objects[query][i]["row_size"]) +
        size_unit)
    stats.append(
        "In Memory Size: " +
        str(
            query_objects[query][i]["in_mem_size"]) +
        size_unit)
    stats.append(
        "On Disk Size: " +
        str(
            query_objects[query][i]["size_on_disk"]) +
        " B")
    print (offset + " | ".join(stats))
  return compute_stat_required


def get_query_to_recommendation_object_map(output):
  output = json.loads(output)
  recommendation_objects = {}
  for recom in output:
    recommendation_objects[recom[0].strip('"')] = recom[2]
  return recommendation_objects


def print_recommendation_objects_verbose(
        recommendation_objects,
        query,
        recommendation_to_name):
  compute_stat_required = []
  offset = "             "
  r = json.loads(recommendation_objects[query])
  for i in range(0, len(r)):
    if recommendation_objects[query][i]["query"]["recommendation_status"] == "NO_RECOMMENDATION":
      print(
          "      -- No recommendation: " +
          recommendation_objects[query][i]["query"]["reason_for_no_recommendation"])
      print(
          offset +
          "Query Block: " +
          str(
              recommendation_objects[query][i]["query"]["text"]))
    else:
      if recommendation_objects[query][i]["query"]["recommendation_status"] in (
              "AWAITING_REFRESH", "ROUTED"):
        print("      -- Query routes to an existing analytical view: " +
              recommendation_objects[query][i]["aview"]["name"])
        if int(recommendation_objects[query][i]["aview"]["row_size"]) <= 0:
          compute_stat_required.append(
            recommendation_objects[query][i]["aview"]["name"])
      else:
        print("      -- Recommended analytical view: " +
              recommendation_to_name[recommendation_objects[query][i]["aview"]["definition"]])
      stats = []
      size_unit = ""
      if int(recommendation_objects[query][i]["aview"]["row_size"]) > 0:
        size_unit = " B"
      stats.append(
          "Row Count: " +
          str(
              recommendation_objects[query][i]["aview"]["row_count"]))
      stats.append(
          "Row Size: " +
          str(
              recommendation_objects[query][i]["aview"]["row_size"]) +
          size_unit)
      stats.append(
          "In Memory Size: " +
          str(
              recommendation_objects[query][i]["aview"]["in_mem_size"]) +
          size_unit)
      if recommendation_objects[query][i]["query"]["recommendation_status"] in (
              "AWAITING_REFRESH", "ROUTED"):
        stats.append(
            "On Disk Size: " +
            str(
                recommendation_objects[query][i]["aview"]["size_on_disk"]) +
            " B")
      stats.append(
          "Partitioned: " +
          str(
              recommendation_objects[query][i]["aview"]["incrementally_refreshable"]))
      print (offset + " | ".join(stats))
      print(
          offset +
          "Query Block: " +
          str(
              recommendation_objects[query][i]["query"]["text"]))
  return compute_stat_required


def print_recommendation_objects(
        recommendation_objects,
        query,
        recommendation_to_name):

  recom = []
  no_recom = []
  wait_recom = []
  r = json.loads(recommendation_objects[query])
  for i in range(0, len(r)):
    if r[i]["query"]["recommendation_status"] == "NO_RECOMMENDATION":
      no_recom.append(
          r[i]["query"]["recommendation_status"])
    elif r[i]["query"]["recommendation_status"] in ("AWAITING_REFRESH", "ROUTED"):
      wait_recom.append(r[i]["aview"]["name"])
    else:
      recom.append(
          recommendation_to_name[r[i]["aview"]["definition"]])
  if len(recom) > 0:
    print ("  -- RECOMMENDATION: " + " | ".join(recom))
  elif len(no_recom) > 0:
    print ("  -- NO RECOMMENDATION")
  if len(wait_recom) > 0:
    print ("  -- ROUTES TO: " + " | ".join(wait_recom))


def print_alert(compute_stats_required):
  print ('-' * LINE_SEPARATOR_COUNT)
  str = ", ".join(set(compute_stats_required))
  print ('ATTENTION: Table metadata information is missing for {} because compute ' 
         'statistics has not been run. Contact your database administrator to '
         'schedule this to happen automatically. In the interim, you can run it '
         'yourself once manually.'.format(str))


def get_query_to_recommendation_map(output):
  query_to_recommendation = {}
  for recom in output:
    query_to_recommendation[recom[0]] = recom[2]
  return query_to_recommendation


def get_query_to_id_map(queries):
  queries_to_id = {}
  querycounter = 0
  for query in queries:
    querycounter += 1
    queries_to_id[query.strip('"')] = str(querycounter)
  return queries_to_id


def get_recommendation_to_query(output, queries):
  recommendation_to_query = {}
  for recom in output:
    data = recom[2]
    values = list(queries[recom[0].strip('"')])
    if data[0]["aview"]["definition"] in recommendation_to_query:
      values = recommendation_to_query.get(
          data[0]["aview"]["definition"]) + values
    recommendation_to_query[data[0]["aview"]["definition"]] = values
  return recommendation_to_query


def print_aview_defs(name_to_recommendations, only_recommendations):
  recoms = []
  if not only_recommendations:
    print ('-' * LINE_SEPARATOR_COUNT)
    print ("Recommended Analytival Views: ")
    print ("")
  for key, value in name_to_recommendations.items():
    str = ""
    if not only_recommendations:
      str += "-- " + key + " : \n   "
    str += value + ";"
    recoms.append(str)
  recoms.sort()
  print ("\n".join(recoms))
  print ("")

def render_output(vres, existing_aviews, verbose, only_recommendations):
  no_recommendation = False
  maps = aview_name_to_recommendation_map(vres[2], existing_aviews)
  name_to_recommendations = maps[0]
  recommendation_to_name = maps[1]
  no_recommendation = maps[2]
  #recommendations = get_recommendations(vres[2], no_recommendation)
  queries = get_input_queries(vres[2])
  query_objects = get_query_object_map(vres[2])
  recommendation_objects = get_query_to_recommendation_object_map(vres[2])
  print_aview_defs(name_to_recommendations, only_recommendations)
  if not only_recommendations:
    compute_stats_required = []
    print ('-' * LINE_SEPARATOR_COUNT)
    print ("")
    print ("Recommendation Details: ")
    counter = 0
    for query in queries:
      counter += 1
      print("")
      print("  -- Q" + str(counter) + ": " + query)
      if verbose:
        compute_stats_required += print_query_objects(query_objects, query)
        compute_stats_required += print_recommendation_objects_verbose(
            recommendation_objects, query, recommendation_to_name)
      else:
        print_recommendation_objects(
            recommendation_objects, query, recommendation_to_name)
    if len(compute_stats_required) > 0:
      print_alert(compute_stats_required)
    print ('-' * LINE_SEPARATOR_COUNT)

def main():
  options = parse()
  # Initialize all the datasets
  dc = get_dataconn(options.dataconn.id)
  vreq = TGetAnalyticalViewsRecommendationsReq(queries=options.queries)
  vres = dc.hs2call(options.username,
                    'GetAnalyticalViewsRecommendations', vreq, True)
  existing_aviews = get_existing_aviews(dc, options)
  render_output(vres, existing_aviews, options.verbose, options.only_recommendations)

if __name__ == "__main__":
  main()
