#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Try view recommendation on one or more apps/vizzes and report errors

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

import os
import re
import sys
import json
import pprint
import getpass
import optparse
import traceback

import time
from datetime import datetime, timedelta

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)
sys.path.append("%s/../vizlib" % THISDIR)

from django.conf import settings

from vizlib.accel import viewrec_accel_info, App, Apps, Shelf

from django.db.models.base import ObjectDoesNotExist
from django.contrib.auth.models import User

from datasets.models import DataSet, DataConnection
from dataconn.dataconn import get_dataconn
from dataconn.connections_types import DATACONN_ARCENGINE
from dataset import ModelDataset, getDatasetInfo
from vizlib.dataset_request import DatasetRequest
from vizlib.accel import viewrec_checkapps

from reports.models import Report

def fatal(text):
  """ Fatal error """
  estr = traceback.format_exc()
  if not estr.startswith("None"):
    print(estr)
  if text:
    print(text)
  sys.exit(-1)

def parse():

  parser = optparse.OptionParser(usage="")

  parser.add_option('--appid', dest='appids', type='int', action='append',
    help='App ID')

  parser.add_option('--alldb', dest='alldb', action='store_true',
    help='Process all apps')

  parser.add_option('--allviz', dest='allviz', action='store_true',
    help='Process all individual vizuals')

  parser.add_option('--allqueries', dest='allqueries', action='store_true',
    help='Show queries for succesful recommendations')

  parser.add_option('--showviews', dest='showviews', action='store_true',
    help='Show view definitions for recommendations')

  parser.add_option('--showids', dest='showids', action='store_true',
    help='Show the ID lists for each error type')

  parser.add_option('--noagg', dest='noagg', action='store_true',
    help='Show apps and queries that fail due to no aggs/grouping/distinct')

  parser.add_option('--verbose', dest='verbose', action='store_true',
    help='Enable django/python logging output')

  parser.add_option('--user', dest='username', default='admin',
    help='Perform data operations as this user')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if len(args) > 0:
    fatal('Extra options: %s' % str(args))

  # Turn off the logging in the django parts (like hs2conn.py)
  if not options.verbose:
    import logging
    logging.disable(logging.DEBUG)

  user = None
  try:
    user = User.objects.get(username=options.username)
  except ObjectDoesNotExist as e:
    sys.exc_clear()
  if not user:
    fatal("Username '%s' not found" % options.username)
  options.user = user

  return options

def main():

  options = parse()

  apptable = dict([(app.id, app) for app in Report.objects.all()])

  if options.appids is None:
    options.appids = []

  if options.alldb:
    options.appids.extend(app.id for app in list(apptable.values()) if app.report_type == 'dashboard')

  if options.allviz:
    options.appids.extend(app.id for app in list(apptable.values()) if app.report_type not in ['dashboard', 'picklist'])

  if not options.appids:
    fatal("No App IDs specified")

  appids = {}
  vizids = set(options.appids)
  DB_RE = re.compile("^uri-(\d+)-widget-(\d+)$")

  for id in set(options.appids):
    app = apptable[id]
    if app.report_type == "dashboard":
      idlist = [id]
      for wid in app.all_report_ids():
        if wid in apptable:
          idlist.append(wid)
      appids[app.id] = idlist
      vizids.difference_update(idlist)

  vizlist = list(vizids)
  vizlist.sort()
  applist = list(appids.keys())
  applist.sort()


  # List of errors to truncate for histogram index
  error_trunc = [
    'Syntax error in line 1',
    'Unable to substitute parameter',
    'GROUP BY expression must not contain aggregate functions',
    'select list expression not produced by aggregation output',
    'Analytical views may not contain non-deterministic function',
  ]

  # Do the apps, one at a time
  appcount = 0
  vizcount = 0
  errcount = 0
  noacount = 0
  histogram = {}
  histoapps = {}
  for id in applist + vizlist:
    app = apptable[id]
    if app.report_type == "dashboard":
      idlist = appids[id]
      appcount += 1
      vizcount += len(idlist) - 1
      print("Test app %s - process %s" % (id, idlist))
    else:
      idlist = [id]
      vizcount += 1
      print("Test viz %s" % id)

    rsp = viewrec_checkapps(options.user, idlist)
    appstat = rsp['appstat']
    viewdefs = dict([(v.get('name'), v.get('select')) for v in rsp['viewdefs']])
    for id in idlist:
      stat = appstat[id]
      assert stat, "No status for app %s" % id
      if apptable[id].report_type == "dashboard":
        assert stat['reason'] == "Must accelerate individual visuals"
      else:
        if stat['status'] == 'RECOMMENDATION':
          if options.allqueries:
            print("  App %s: %s" % (id, stat['status']))
            print("  Query: %s" % stat['query'])
            if options.showviews:
              print("  AView: %s" % viewdefs[stat['viewname']])
        else:
          reason = stat['reason']
          for t in error_trunc:
            if reason.startswith(t):
              reason = reason[0:len(t)]
              break
          histo = "%s - %s" % (stat['status'], reason)
          if histo not in histogram:
            histogram[histo] = 0
            histoapps[histo] = set()
          histogram[histo] += 1
          histoapps[histo].add(id)
          errcount += 1
          noagg = (stat['status'] == 'NO_RECOMMENDATION' and \
                   stat['reason'] == 'The query does not have any aggregates, distinct or grouping in it') or \
                  (stat['status'] == 'FAILED' and \
                   stat['reason'] == 'No aggregates selected')
          if noagg:
            noacount += 1
          if not noagg or options.noagg:
            print("  App %s: %s - %s" % (id, stat['status'], stat['reason']))
            print("  Query: %s" % stat['query'])

  print("Total apps: %s" % appcount)
  print("  Total visuals:  %s" % vizcount)
  print("  Recommenations: %s" % (vizcount - errcount))
  print("  No aggregates:  %s" % noacount)
  print("  Other errors:   %s" % (errcount - noacount))
  if errcount > 0:
    errs = [(val, key) for (key, val) in list(histogram.items())]
    errs.sort()
    for (val, key) in errs:
      print("    %3s : %s" % (val, key))
      if options.showids:
        idlist = list(histoapps[key])
        idlist.sort()
        print("        : %s" % idlist)

if __name__ == "__main__":
  main()

