# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
# In the previous CDV versions we allowed parenthesis as display name (alias) for the datasets' fields
# But with recent implementations it causes errors with filters
# The script attached here tries to find fields with has parenthesis in display name and remove the parenthesis
# For example, field with display name 'height (cm)' will be 'height - cm'

# To run it on a CDV instance running within CDSW, use:
# - export PYTHONPATH=/home/cdsw/.arc:/opt/vizapps/venv/lib/python3.6/site-packages/arcweb/:$PYTHONPATH
# - export BASEDIR=/home/cdsw/.arc
# - export VENV_PATH=/opt/vizapps/venv
# - export DJANGO_SETTINGS_MODULE=settings_local
# - export VIZ_PORT=8100
# - export CDV_MODE=cdsw
# - export PYTHONPATH=${PYTHONPATH}:${BASEDIR}
# - export DIR=/opt/vizapps/tools/arcviz
# - ${DIR}/arcviz util remove_columns_parenthesis
# Explore Mode
# - export BASEDIR=/home/cdsw/.explore
# - export PYTHONPATH=${PYTHONPATH}:${BASEDIR}
# Legacy Connection
# - export VENV_PATH=/opt/vizapps/legacy/venv

# Output looks like this
# Earthquake Data January 2019 with id: 136 has these affected fields:
#
# updated (date)
#
# Cereals with id: 140 has these affected fields:
#
# cereal (name)
# Done
#
# Examples of removing parenthesis from datasets 1 2 3 vs all datasets
# Removing from datasets 1 2 3
# ${DIR}/arcviz util remove_columns_parenthesis --remove --datasets 1 2 3
# Removing from all datasets
# ${DIR}/arcviz util remove_columns_parenthesis --remove --datasets 0


import re
import json
import argparse
from datasets.models import DataSet
from datasets.views import dataset_update_info

def remove_columns_parenthesis(dataset, show_info_only=False):
    dataset_info = json.loads(dataset.dataset_info)
    columns = dataset_info[0]['columns']

    cols_positions_with_match = get_parenthesis_aliased_columns_positions(columns)

    if cols_positions_with_match and show_info_only:
        print(f'\n{dataset.dataset_name} with id: {dataset.id} has these affected fields:\n')
        for (i, _) in cols_positions_with_match:
            print(columns[i]['alias'])

    if not show_info_only:
        for (i, match) in cols_positions_with_match:
            new_alias = overwrite_column_aliases(match)
            print(f'Changing the alias {columns[i]["alias"]} to {new_alias}')
            columns[i]['alias'] = new_alias

        if cols_positions_with_match:
            dataset_update_info(dataset, json.dumps(dataset_info))

    return len(cols_positions_with_match) > 0

def get_parenthesis_aliased_columns_positions(columns):
    cols_positions_with_match = []
    for i in range(len(columns)):
        alias = columns[i].get('alias')
        if alias:
            pattern = re.compile(r'(.*)\((.*)\)(.*)')
            matched = pattern.match(alias)
            if matched:
                cols_positions_with_match.append((i, matched))

    return cols_positions_with_match


def overwrite_column_aliases(alias_with_parenthesis):
    column_name = alias_with_parenthesis.group(1).strip()

    alias = f'{column_name} - {alias_with_parenthesis.group(2)}'

    if alias_with_parenthesis.group(3) != '':
        alias += f'-{alias_with_parenthesis.group(3)}'

    return alias

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--remove', action='store_true', help='Removing parentheses from datasets')
    parser.add_argument('--datasets', type=int, dest='datasets_to_process', nargs='+', help='Removing parentheses from specified datasets\' ids')
    args = parser.parse_args()

    no_affected_datasets = True
    affected_datasets = []

    datasets = DataSet.objects.all()
    for dataset in datasets:
        is_dataset_affected = remove_columns_parenthesis(dataset, True)
        if is_dataset_affected:
            no_affected_datasets = False
            affected_datasets.append(dataset)

    if no_affected_datasets:
        print('There are no affected datasets')
        return

    if args.remove:
        datasets_to_process = args.datasets_to_process if args.datasets_to_process else []

        if len(datasets_to_process) == 1 and  datasets_to_process[0] == 0:
            print('Removing parenthesis from all datasets')
            for dataset in affected_datasets:
                remove_columns_parenthesis(dataset)
            
            print('Done')
            return

        for dataset in affected_datasets:
            if dataset.id in datasets_to_process:
                print(f'Removing parenthesis from {dataset.dataset_name}')
                remove_columns_parenthesis(dataset)

    else:
        print('You can use --remove option to make changes')

    print('Done')

if __name__ == '__main__':
    import django
    django.setup()

    main()
