# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
# Purpose
  # This script aims to clean up redundant dataset privileges for roles.
  #
  # There are four 'types' of dataset privileges
  # 1. Allows permissions for all data connections and all datasets
  # 2. Allows permissions for all datasets on a specific data connection
  # 3. Allows permissions for multiple datasets on a specific data connection
  # 4. Allows permissions for one dataset
  #
  # If a privilege allows the same or more permissions as another privilege for all data connections and datasets all
  # all their other privileges that allows the same or less permissions are redundant
  #
  # If a privilege allows the same or more permissions as another privilege for all datasets on a specific data connection
  # all their other privileges that allows the same or less permissions for datasets belonging to that data connection are redundant
  #
  # If a privilege allows the same or more permissions as another privilege for multiple datasets on a specific data connection
  # all their other privileges that allows the same or less permissions for the selected dataset are redundant
  #
  # Redundant privileges are removed or in the case of privileges that allows multiple datasets
  # the selected dataset is removed from the datasets for that privilege

# To run it on a CDV instance running within CDSW, use:
  # export PYTHONPATH=/home/cdsw/.arc:/opt/vizapps/venv/lib/python3.9/site-packages/arcweb/:$PYTHONPATH
  # export BASEDIR=/home/cdsw/.arc
  # export VENV_PATH=/opt/vizapps/venv
  # export DJANGO_SETTINGS_MODULE=settings_local
  # export VIZ_PORT=8100
  # export CDV_MODE=cdsw
  # export PYTHONPATH=${PYTHONPATH}:${BASEDIR}
  # export DIR=/opt/vizapps/tools/arcviz
  # ${DIR}/arcviz util remove_redundant_dataset_privileges
# Explore Mode
  # export BASEDIR=/home/cdsw/.explore
  # export PYTHONPATH=${PYTHONPATH}:${BASEDIR}
# Legacy Connection
  # export VENV_PATH=/opt/vizapps/legacy/venv

# Environmental variables to control the behaviour
  # REMOVE_REDUNDANT_PRIVILEGES - set to true to make actual changes (otherwise it will just print the changes it would do)
  # SELECTED_DATASET_IDS - comma separated list of dataset ids to process the privileges by, if not specified all datasets will be selected

from textwrap import indent
import sys
import os
import json
from copy import deepcopy
from datasets.models import DataSet
from rbac.models import Role

from typing import List, TypedDict, Optional, Dict, Set, Literal

class Privilege(TypedDict):
  ptype: str
  dcid: str
  dslist: List[str]
  perms: List[str]

class MultipleDatasetsPrivilege(Privilege):
  id: int

class CategorizedPrivileges(TypedDict):
  not_affected_privileges: List[Privilege]
  all_connections_privilege: Optional[Privilege]
  all_datasets_privilege: Optional[Privilege]
  specific_dataset_privilege: Optional[Privilege]
  multiple_datasets_privileges: List[MultipleDatasetsPrivilege]

class Changes(TypedDict):
  all_datasets_privilege: Optional[bool]
  specific_dataset_privilege: Optional[bool]
  multiple_datasets_privileges: Optional[Dict[str, Set[str]]]

class AdditionalCondition(TypedDict):
  method: Literal['not-equal', 'superset']
  target: Privilege

def get_categorized_privileges(role: Role, dataset: DataSet) -> CategorizedPrivileges:
    not_affected_privileges: List[Privilege] = []
    # There CAN'T be multiple privileges of these types
    all_connections_privilege: Optional[Privilege] = None
    all_datasets_privilege: Optional[Privilege] = None
    specific_dataset_privilege: Optional[Privilege] = None
    # There CAN be multiple privileges of these type
    multiple_datasets_privileges: List[MultipleDatasetsPrivilege] = []
    id_for_multiple_datasets: int = -1

    privileges: List[Privilege] = json.loads(role.privileges)

    for privilege in privileges:
      # If the privilege type is not dataset, leave it as is
      if privilege['ptype'] != 'dataset':
        not_affected_privileges.append(privilege)
        continue

      # If not every connection is allowed and it is not the selected connection, leave it as is
      connection_id = privilege['dcid']
      if connection_id != '-1' and connection_id != str(dataset.dataconnection.id):
        not_affected_privileges.append(privilege)
        continue

      # If not every dataset is allowed and it does not include the selected dataset, leave it as is
      dataset_ids = privilege['dslist']
      if '-1' not in dataset_ids and str(dataset.id) not in dataset_ids:
        not_affected_privileges.append(privilege)
        continue

      # Group privileges by type
      if connection_id == '-1':
        all_connections_privilege = privilege
        continue

      if '-1' in dataset_ids:
        all_datasets_privilege = privilege
        continue

      if len(dataset_ids) == 1:
        specific_dataset_privilege = privilege
      else:
        id_for_multiple_datasets += 1
        privilege['id'] = id_for_multiple_datasets
        multiple_datasets_privileges.append(privilege)

    return {
      'not_affected_privileges': not_affected_privileges,
      'all_connections_privilege': all_connections_privilege,
      'all_datasets_privilege': all_datasets_privilege,
      'specific_dataset_privilege': specific_dataset_privilege,
      'multiple_datasets_privileges': multiple_datasets_privileges
    }

def get_changes(categorized_privileges: CategorizedPrivileges, dataset: DataSet) -> Changes:
  changes: Changes = {}

  multiple_datasets_privilege_ids_where_dataset_should_be_removed: List[str] = []

  def get_affected_privilege_ids_for_multiple_datasets_privileges(
    multiple_datasets_privileges: List[MultipleDatasetsPrivilege],
    dataset: DataSet,
    additional_condition: Optional[AdditionalCondition] = None
  ) -> List[str]:
    affected_ids = []

    for multiple_datasets_privilege in multiple_datasets_privileges:
      dataset_id = str(dataset.id)
      includes_dataset = dataset_id in multiple_datasets_privilege['dslist']

      is_additional_condition_truthy = True
      if additional_condition != None:
        if additional_condition['method'] == 'not-equal':
          is_additional_condition_truthy = additional_condition['target'] != multiple_datasets_privilege
        elif additional_condition['method'] == 'superset':
          is_additional_condition_truthy = additional_condition['target'].issuperset(set(multiple_datasets_privilege['perms']))

      if includes_dataset and is_additional_condition_truthy:
        affected_ids.append(multiple_datasets_privilege['id'])

    return affected_ids

  # all_connections & all_datasets
    # keep it
    # remove all_datasets that has same or less permissions
    # remove single dataset that has same or less permissions
    # remove the dataset_id from multiple_datasets that has same or less permissions
  if categorized_privileges['all_connections_privilege']:
    all_connections_permissions = set(categorized_privileges['all_connections_privilege']['perms'])

    if categorized_privileges['all_datasets_privilege'] and all_connections_permissions.issuperset(set(categorized_privileges['all_datasets_privilege']['perms'])):
      changes['all_datasets_privilege'] = True

    if categorized_privileges['specific_dataset_privilege'] and all_connections_permissions.issuperset(set(categorized_privileges['specific_dataset_privilege']['perms'])):
      changes['specific_dataset_privilege'] = True

    multiple_datasets_privilege_ids_where_dataset_should_be_removed.extend(
      get_affected_privilege_ids_for_multiple_datasets_privileges(
        multiple_datasets_privileges=categorized_privileges['multiple_datasets_privileges'],
        dataset=dataset,
        additional_condition={
          'target': all_connections_permissions,
          'method': 'superset',
        }
      )
    )

  # all_datasets
    # keep it
    # remove single dataset that has same or less permissions
    # remove the dataset_id from multiple_datasets that has same or less permissions
  if changes.get('all_datasets_privilege') == None and categorized_privileges['all_datasets_privilege']:
    all_connections_permissions = set(categorized_privileges['all_datasets_privilege']['perms'])

    if changes.get('specific_dataset_privilege') == None and categorized_privileges['specific_dataset_privilege'] and all_connections_permissions.issuperset(set(categorized_privileges['specific_dataset_privilege']['perms'])):
      changes['specific_dataset_privilege'] = True

    multiple_datasets_privilege_ids_where_dataset_should_be_removed.extend(
      get_affected_privilege_ids_for_multiple_datasets_privileges(
        multiple_datasets_privileges=categorized_privileges['multiple_datasets_privileges'],
        dataset=dataset,
        additional_condition={
          'target': all_connections_permissions,
          'method': 'superset',
        }
      )
    )

  # single_dataset
    # is there at least one multiple_datasets privilege that has more permissions?
      # yes
        # remove it
      # no
        # keep it
        # remove the dataset_id from all the multiple_datasets
  if changes.get('specific_dataset_privilege') == None and categorized_privileges['specific_dataset_privilege']:
    all_connections_permissions = set(categorized_privileges['specific_dataset_privilege']['perms'])

    for multiple_datasets_privilege in categorized_privileges['multiple_datasets_privileges']:
      multiple_datasets_permissions = set(multiple_datasets_privilege['perms'])
      if multiple_datasets_permissions.issuperset(all_connections_permissions) and len(all_connections_permissions) < len(multiple_datasets_permissions):
        changes['specific_dataset_privilege'] = True
        break

  # There WEREN'T any multiple_dataset privilege that had more permission
  if changes.get('specific_dataset_privilege') == None and categorized_privileges['specific_dataset_privilege']:
    multiple_datasets_privilege_ids_where_dataset_should_be_removed.extend(
      get_affected_privilege_ids_for_multiple_datasets_privileges(
        multiple_datasets_privileges=categorized_privileges['multiple_datasets_privileges'],
        dataset=dataset,
      )
    )

  # There WAS at least one multiple_dataset privilege that had more permission
  # multiple_datasets
    # are there multiple privileges with multiple_datasets?
      # yes
        # keep the one with the most permissions
        # remove the dataset_id from all the other multiple datasets
      # no
        # keep it
  elif len(categorized_privileges['multiple_datasets_privileges']) > 1:
    privilege_with_most_permissions: Optional[MultipleDatasetsPrivilege] = None
    for multiple_datasets_privilege in categorized_privileges['multiple_datasets_privileges']:
      if privilege_with_most_permissions == None:
        privilege_with_most_permissions = multiple_datasets_privilege

      if set(multiple_datasets_privilege['perms']).issuperset(set(privilege_with_most_permissions['perms'])):
        privilege_with_most_permissions = multiple_datasets_privilege

    multiple_datasets_privilege_ids_where_dataset_should_be_removed.extend(
      get_affected_privilege_ids_for_multiple_datasets_privileges(
        multiple_datasets_privileges=categorized_privileges['multiple_datasets_privileges'],
        dataset=dataset,
        additional_condition={
          'target': privilege_with_most_permissions,
          'method': 'not-equal',
        }
      )
    )

  for privilege_id in multiple_datasets_privilege_ids_where_dataset_should_be_removed:
      if changes.get('multiple_datasets_privileges') == None:
        changes['multiple_datasets_privileges'] = {}

      dataset_id = str(dataset.id)
      if changes['multiple_datasets_privileges'].get(privilege_id):
        changes['multiple_datasets_privileges'][privilege_id].add(dataset_id)
      else:
        changes['multiple_datasets_privileges'][privilege_id] = set([dataset_id])

  return changes

def get_new_privileges(categorized_privileges: CategorizedPrivileges, changes: Changes) -> List[Privilege]:
  categorized_privileges = deepcopy(categorized_privileges)

  if changes.get('all_datasets_privilege'):
    categorized_privileges['all_datasets_privilege'] = None

  if changes.get('specific_dataset_privilege'):
    categorized_privileges['specific_dataset_privilege'] = None

  if changes.get('multiple_datasets_privileges') != None:
    for multiple_datasets_privilege_id in changes['multiple_datasets_privileges']:
      for multiple_datasets_privilege in categorized_privileges['multiple_datasets_privileges']:
        if multiple_datasets_privilege_id == multiple_datasets_privilege['id']:
          removed_dataset_ids = changes['multiple_datasets_privileges'][multiple_datasets_privilege_id]
          multiple_datasets_privilege['dslist'] = list(filter(lambda id: id not in removed_dataset_ids, multiple_datasets_privilege['dslist']))

  new_privileges = categorized_privileges['not_affected_privileges']

  if categorized_privileges['all_connections_privilege']:
    new_privileges.append(categorized_privileges['all_connections_privilege'])

  if categorized_privileges['all_datasets_privilege']:
    new_privileges.append(categorized_privileges['all_datasets_privilege'])

  if categorized_privileges['specific_dataset_privilege']:
    new_privileges.append(categorized_privileges['specific_dataset_privilege'])

  if categorized_privileges['multiple_datasets_privileges']:
    for multiple_datasets_privileges in categorized_privileges['multiple_datasets_privileges']:
      multiple_datasets_privileges.pop('id')

    new_privileges.extend(categorized_privileges['multiple_datasets_privileges'])

  return new_privileges

def get_privilege_changes_based_on_dataset_by_role_id(
  datasets: List[DataSet],
  roles: List[Role],
) -> Dict[str, Dict[str, List[Changes]]]:
  # A dictionary containing the required privilege changes for each role by each dataset
  # Example:
  # For he role with the id of 1,
  # the privilege that only applies to the dataset with id of 14
  # should be removed
  # {
  #   1: {
  #     14: {
  #       'specific_privilege': True
  #     }
  #   }
  # }
  privilege_changes_by_dataset_id_by_role_id: Dict[str, Dict[str, List[Changes]]] = {}

  for selected_dataset in datasets:
    print(f'Dataset (name - id) : {selected_dataset.dataset_name} - {selected_dataset.id}')

    for role in roles:
      categorized_privileges = get_categorized_privileges(role=role, dataset=selected_dataset)

      changes = get_changes(categorized_privileges=categorized_privileges, dataset=selected_dataset)
      has_changes = len(changes.keys()) > 0

      if has_changes:
        # Register changes
        if privilege_changes_by_dataset_id_by_role_id.get(role.id):
          privilege_changes_by_dataset_id_by_role_id[role.id][selected_dataset.id] = changes
        else:
          privilege_changes_by_dataset_id_by_role_id[role.id] = { selected_dataset.id: changes }

        # Display expected changes to the user
        new_privileges = get_new_privileges(categorized_privileges=categorized_privileges, changes=changes)
        print(indent(f'Role: {role.name}', '  '))
        print(indent('Before:', '    '))
        print(indent(json.dumps(json.loads(role.privileges), indent=2), '      '))
        print(indent('After:', '    '))
        print(indent(json.dumps(new_privileges, indent=2), '      '))

  return privilege_changes_by_dataset_id_by_role_id

def get_unsaved_updated_roles(roles: List[Role], datasets: List[DataSet]) -> List[Role]:
  # Calculate what kind of privilege changes are needed based on each role and dataset
  privilege_changes_based_on_dataset_by_role_id = get_privilege_changes_based_on_dataset_by_role_id(
    roles=roles,
    datasets=datasets,
  )

  # Apply privilege changes to the roles to create the new roles
  new_roles: List[Role] = []

  datasets_by_id: Dict[str, DataSet] = {}
  for dataset in datasets:
    datasets_by_id[dataset.id] = dataset

  roles_by_id: Dict[str, Role] = {}
  for role in roles:
    roles_by_id[role.id] = role

  for role_id, changes_by_dataset_id in privilege_changes_based_on_dataset_by_role_id.items():
    role = roles_by_id[role_id]

    for dataset_id, changes_for_dataset in changes_by_dataset_id.items():
      dataset = datasets_by_id[dataset_id]

      categorized_privileges = get_categorized_privileges(role=role, dataset=dataset)
      new_privileges = get_new_privileges(categorized_privileges=categorized_privileges, changes=changes_for_dataset)
      role.privileges = json.dumps(new_privileges)

    new_roles.append(role)

  return new_roles

def main():
  # Get environmental variables
  remove_redundant_privileges = os.getenv('REMOVE_REDUNDANT_PRIVILEGES', '') == 'true'
  selected_dataset_ids = os.getenv('SELECTED_DATASET_IDS', '*')

  # Get datasets
  if selected_dataset_ids.strip() != '*':
    try:
      selected_dataset_ids = selected_dataset_ids.split(',')

      selected_dataset_ids = [int(id) for id in selected_dataset_ids]
    except ValueError:
      print('Error: Please provide SELECTED_DATASET_IDS as comma separated list, i.e: 2,3,4')
      sys.exit(1)

  if selected_dataset_ids[0] == '*':
    datasets = DataSet.objects.all()
    print('Calculating for every dataset, to specify datasets set SELECTED_DATASET_IDS environmental variable as a comma separated list of dataset ids')
  else:
    datasets = DataSet.objects.filter(id__in=selected_dataset_ids)

  # Get roles
  roles = Role.objects.all()

  # Get new roles
  unsaved_updated_roles = get_unsaved_updated_roles(roles, datasets)

  # Commit changes if REMOVE_REDUNDANT_PRIVILEGES is enabled
  if remove_redundant_privileges:
    for new_role in unsaved_updated_roles:
      new_role.save(update_fields=['privileges'])

  else :
    print('Changes are not applied, to apply the changes set REMOVE_REDUNDANT_PRIVILEGES environmental variable to true.')

if __name__ == '__main__':
  main()