#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Role update utility.

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

USAGE = """

The role_tool allows command line update operations for roles.  General form:

  python util/role_tool.py <action> <role_name> [options]

  or

  python manage.py util role <action> <role_name> [options]

Where action is one of:

  create - Create a new role
  update - Update an existing role
  delete - Delete an existing role
  list - List existing role summary data
  showcreate - Show how an existing role would be created with this tool.

The showcreate action is useful for seeing how an existing role would be
fully specified with this tool.

The create/update actions allow full access to the role: top level attributes,
role membership, and the individual privilege rows:

Top level attributes:
  
  --desc <description>  Role description text
  --rename <new name>   Set new role name (update only)

Role membership:

  For role users, add or delete members or specify the exact membership:

    --addusers <user list>  Add the specified users to the role
    --delusers <user list>  Remove the specified users from the role
    --setusers <user list>  Set the role users to the specified list

    The analogous commands exist for groups.  Note that user and group names are
    case-sensitive and the lists are comma-seperated.  For example to remove user
    'User1' and set the groups to contain exactly 'Data Team' and 'App Devs':

      --delusers User1 --setgroups 'Data Team,App Devs'

Permissions:

  Role permissions are made up of individual 'privilage rows', each row having
  having one of three types (System, Connection, and Dataset) and corresponding
  set of permissions for each type.

  Operations allowed on each privilege row are:

    --addpriv <specifier>  Add a new privilege row
    --delpriv <specifier>  Delete existing privilege row
    --modpriv <specifier>  Update existing privilege row

  The specifier format depends on the type of privilege:

    System     - No specifer--only on system privilege row exists in a role
    Dataconn   - Specify a list of data connection IDs: --dclist <list>
    Dataset    - Connection ID and list of dataset IDs: --dcid <ID> --dslist <list>

  Note: For both connection and dataset IDs, the special value "-1" is used to
  indicate 'All connections' or 'All datasets'.

  Like users and groups, the permissions for each privilege row are given with
  a comma separated list.

  The --modpriv operation will create the privilege row if it doesn't already
  exist and allows permissions to be added, dropped, or specifically set:

    --addperms <list>  Add specified permissions to the privilege row
    --delperms <list>  Remove the specified permissions to the privilege row
    --setperms <list>  Set the permissions for the privilege row to the specified list

  The --addpriv operation only permits permissions to be added with --addperms.

  The --delpriv operation does not permit any permission updates as it just removes
  the specified privilege row.

  System level permissions:

PERMDEFS_system

    Example: Set the system level permissions for the role to allow 'View activity logs' and 
    'Manage styles and settings':

      --modpriv system --setperms sys_styles,sys_viewlogs

  Role level permissions:

PERMDEFS_role

    Example: Add ability to grant dataset 'View visuals and dashboards' to role IDs 2,3:

      --addpriv role --rolelist 2,3 --addperms rgds_appview

  Connection level permissions:

PERMDEFS_dataconn

    Example: Add ability to import data on connection 1 to the role:

      --addpriv dataconn --dclist 1 --addperms dc_upload

  Dataset level permissions:

PERMDEFS_dataset

    Example: Add a privilege row to enable 'View visuals and dashboards' on every dataset:

      --addpriv dataset --dcid -1 --dslist -1 --addperms ds_appview

"""

import os
import sys
import json
import getpass
import optparse
import traceback

from optparse import OptionValueError

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)
sys.path.append("%s/../vizlib" % THISDIR)

from django.db.models.base import ObjectDoesNotExist
from rbac.models import Role, keypriv, privkey
from rbac.permdefs import PERMLISTS, PERMTYPES, PERMDEFS

# Turn off the logging in the django parts (like hs2conn.py)
import logging
logging.disable(logging.DEBUG)

ACTIONS = ['create', 'update', 'delete', 'list', 'showcreate']

def fatal(text):
  """ Fatal error """
  estr = traceback.format_exc()
  if not estr.startswith("None"):
    print(estr)
  if text:
    print(text)
  sys.exit(-1)

def parse_int(opt, val):
  try:
    return str(int(val))
  except:
    raise OptionValueError("--%s must be integer, not %s" % (opt, val))

def parse_ints(opt, val):
  try:
    return [str(int(i)) for i in val.split(',')]
  except:
    raise OptionValueError("--%s must be list of integers, not %s" % (opt, val))

# Validate options for the most recent privilege op
def checkop(op):

  def canthave(op, options):
    for option in options:
      if op.get(option) is not None:
        raise OptionValueError('Option --%s not allowed for privilege of type %s' % (option, op.get('ptype')))

  def musthave(op, options):
    for option in options:
      if op.get(option) is None:
        raise OptionValueError('Option --%s is required for privilege of type %s' % (option, op.get('ptype')))

  ptype = op.get('ptype')
  if ptype == 'system':
    canthave(op, ['dcid', 'dclist', 'dslist', 'rolelist'])
  elif ptype == 'dataconn':
    canthave(op, ['dcid', 'dslist', 'rolelist'])
    musthave(op, ['dclist'])  
  elif ptype == 'dataset':
    canthave(op, ['dclist', 'rolelist'])
    musthave(op, ['dcid', 'dslist'])
  elif ptype == 'role':
    canthave(op, ['dcid', 'dclist', 'dslist'])
    musthave(op, ['rolelist'])

# Start a new option structure
def newop(option, opt, val, parser):
  if not val:
    raise OptionValueError("%s requires privilege type parameter" % opt)
  if val not in PERMTYPES:
    raise OptionValueError("Privilege type must be one of %s not %s" % (PERMTYPES, val))
  if len(parser.ops) > 0:
    checkop(parser.ops[-1])
  op = {"op":opt[2:], "ptype":val}
  parser.ops.append(op)

# Process an parameter for the current operation
def param(option, opt, val, parser):
  if len(parser.ops) == 0:
    raise OptionValueError("%s requires privilege op specified via --addpriv, --delpriv, --modpriv" % opt)
  priv = parser.ops[-1]
  opt = opt[2:].lower()
  if opt == 'dcid':
    priv[opt] = parse_int(opt, val)
  elif opt in ['dclist', 'dslist', 'rolelist']:
    priv[opt] = parse_ints(opt, val)
  elif opt in ['addperms', 'delperms', 'setperms']:
    if priv.get('op') == 'addpriv' and opt in ['delperms']:
      raise OptionValueError('Option --delperms not allowed with --addpriv')
    if priv.get('opt') is None:
      priv[opt] = set()
    if val:
      for perm in [s.strip() for s in val.split(',')]:
        plist = PERMLISTS[priv.get('ptype')]
        if perm.lower() not in plist:
          raise OptionValueError("Invalid permission %s in list '%s'.  Must be one of %s" % (perm, val, plist))
        priv[opt].add(perm.lower())
  else:
    fatal("Unknown column operation parameter: --%s" % opt)

# Process a list of user/group names
def namelist(option, opt, val, parser):
  names = getattr(parser, opt[2:], None)
  if names is None:
    names = set()
    setattr(parser, opt[2:], names)
  if val:
    names.update([s.strip() for s in val.split(',')])

def parse():

  parser = optparse.OptionParser(usage="")
  parser.ops = []

  parser.add_option('--usage', dest='usage', action='store_true',
    help='Display usage text and exit')

  parser.add_option('--desc', dest='desc', help='Set role description text')
  parser.add_option('--rename', dest='rename', help='Set new role name')

  # These options handle updates to users and groups lists
  for (act, help) in [('add', ['Add', 'to']), ('del', ['Delete', 'from']), ('set', ['Set the exact', 'for'])]:
    for items in ['users', 'groups']:
      parser.add_option('--%s%s' % (act, items), type='string', action='callback', callback=namelist,
                        help='%s %s %s the role' % (help[0], items, help[1]))

  # These options start a new privilege operation structure being constructed
  for (act, help) in [('mod', 'Update an existing privilege'),
                      ('add', 'Add a new privilege'),
                      ('del', 'Delete an existing privilege')]:
    parser.add_option('--%spriv' % act, type='string', action='callback', callback=newop, help=help)

  # Parameters which are applied to the current privilege operations
  parser.add_option('--dcid', type='string', action='callback', callback=param,
    help='Specify the data connection ID for a dataset type privilege')
  parser.add_option('--dclist', type='string', action='callback', callback=param,
    help='Specify list of data connection IDs for dataconn type privilege')
  parser.add_option('--dslist', type='string', action='callback', callback=param,
    help='Specify list of dataset IDs for dataset type privilege')
  parser.add_option('--rolelist', type='string', action='callback', callback=param,
    help='Specify list of role IDs for role type privilege')

  # Permission options for the current privilege operation
  for (act, help) in [('set', 'Set permissions for the current privilege'),
                      ('add', 'Add permissions to current privilege'),
                      ('del', 'Delete permissions from current privilege')]:
    parser.add_option('--%sperms' % act, type='string', action='callback', callback=param, help=help)

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  # Fixup usage text with actual permission definitions
  newuse = []
  prefix = 'PERMDEFS_'
  for l in USAGE.split('\n'):
    if l.startswith(prefix):
      ptype = l[len(prefix):]
      for perm in PERMLISTS[ptype]:
        newuse.append('    %-14s - %s' % (perm, PERMDEFS[perm]))
    else:
      newuse.append(l)

  if options.usage:
    print(('\n'.join(newuse)))
    parser.print_help()
    sys.exit(0)

  if len(args) == 0:
    fatal('Role action not specified')

  action = args[0]
  if action.lower() not in ACTIONS:
    fatal('Role action must be one of %s, not %s' % (action, ACTIONS))
  options.action = action.lower()

  options.rolename = None
  if action == 'list':
    if len(args) > 1:
      fatal('Extra arguments: %s' % args[1:])
  else:
    if len(args) == 1:
      fatal('Role name not specified')
    options.rolename = args[1]
    if len(args) > 2:
      fatal('Extra arguments: %s' % args[2:])
    if options.action in ['update', 'delete', 'showcreate']:
      try:
        options.role = Role.objects.get(name=options.rolename)
      except ObjectDoesNotExist:
        fatal("Role name '%s' not found" % options.rolename)
    if options.action == 'create':
      try:
        options.role = Role.objects.get(name=options.rolename)
        fatal("Role name '%s' already exists" % options.rolename)
      except ObjectDoesNotExist:
        pass
      options.role = Role(name=options.rolename)

  if options.desc:
    if options.action not in ['create', 'update']:
      fatal('--desc option only valid for create/update actions')

  if options.rename:
    if options.action != 'update':
      fatal('--rename option only value for rename action')
    try:
      Role.objects.get(name=options.rename)
      fatal("Unable to rename to '%s': role already exists" % option.rename)
    except ObjectDoesNotExist:
      pass
  
  if len(parser.ops) > 0:
    checkop(parser.ops[-1])

  return (options, parser)

def process_members(role, parser):

  users = getattr(parser, 'setusers', None)
  if users is None:
    addset = getattr(parser, 'addusers', set())
    delset = getattr(parser, 'delusers', set())
    users = set(json.loads(role.users or '[]'))
    users.difference_update(delset)
    users.update(addset)
  role.users = json.dumps(list(users))

  groups = getattr(parser, 'setgroups', None)
  if groups is None:
    addset = getattr(parser, 'addgroups', set())
    delset = getattr(parser, 'delgroups', set())
    groups = set(json.loads(role.groups or '[]'))
    groups.difference_update(delset)
    groups.update(addset)
  role.groups = json.dumps(list(groups))

def process_ops(role, ops):
  privmap = {}
  for p in json.loads(role.privileges or '[]'):
    key = privkey(p)
    if privmap.get(key) is None:
      privmap[key] = set()
    privmap[key].update(p.get('perms'))

  for op in ops:
    key = privkey(op)
    opcode = op.get('op')

    if opcode == 'delpriv':
      try:
        del privmap[key]
      except KeyError:
        pass

    if opcode in ['addpriv', 'modpriv']:
      perms = op.get('setperms', None)
      if perms is None:
        addset = op.get('addperms', set())
        delset = op.get('delperms', set())
        perms = privmap.get(key, set())
        perms.difference_update(delset)
        perms.update(addset)

      # Update for forced permissions
      forced = set()
      for p in perms:
        forced.update(PERMLISTS.get(p, []))
      perms.update(forced)
      privmap[key] = perms

  privs = []
  for key in sorted(list(privmap.keys()), key=lambda k: PERMTYPES.index(k[0])):
    perms = privmap[key]
    if perms:
      priv = keypriv(key)
      priv['perms'] = list(sorted(privmap[key]))
      privs.append(priv)

  role.privileges = json.dumps(privs)

def showcreate(role):
  print("role_tool create '%s' \\" % role.name)
  print("  --desc '%s' \\" % role.description)
  print("  --setusers '%s' \\" % ','.join(json.loads(role.users or '[]')))
  print("  --setgroups '%s'" % ','.join(json.loads(role.groups or '[]')), end=' ')
  for priv in json.loads(role.privileges or '[]'):
    print("\\")
    ptype = priv.get('ptype')
    parts = ['  --addpriv', ptype]
    if ptype == 'role':
      parts.append('--rolelist')
      parts.append("'%s'" % ','.join(priv['rolelist']))
    if ptype == 'dataconn':
      parts.append('--dclist')
      parts.append("'%s'" % ','.join(priv['dclist']))
    elif ptype == 'dataset':
      parts.append('--dcid')
      parts.append(priv['dcid'])
      parts.append('--dslist')
      parts.append("'%s'" % ','.join(priv['dslist']))
    parts.append('--addperms')
    parts.append("'%s'" % ','.join(priv['perms']))
    print(" ".join(parts), end=' ')

  print()

def main():
  options, parser = parse()

  if options.action in ['create', 'update']:
    role = options.role
    process_members(role, parser)
    process_ops(role, parser.ops)
    if options.desc:
      role.description = options.desc
    if options.rename:
      role.name = options.rename
    role.save()

  if options.action == 'delete':
    options.role.delete()

  if options.action == 'showcreate':
    showcreate(options.role)

  if options.action == 'list':
    for r in Role.objects.all():
      print("%s %s" % (r.id, r.name))

if __name__ == "__main__":
  main()

