# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.db import close_old_connections

from concurrent.futures import ThreadPoolExecutor
import threading
import logging
import time

import util.perfcounters as perfcounters
from util.perfcounters import APPROX_COUNTERS

LOGGER = logging.getLogger("arcweb")

#
# When we run async jobs within django, we frequently need to access the
#   django backend db by accessing models. Access to the db happens via
#   backend database connections that have to be cleaned up by using calls
#   to close_old_connections(). The normal webserver threads call this
#   via the Django signal mechanism, but threads have to call this explicitly.
#
# To help with this, we create a safe DjangoSafeThreadPoolExecutor below which
#   should be used instead of ThreadPoolExecutor instances directly.
#
class DjangoSafeThreadPoolExecutor(object):
  def __init__(self, name=None, max_workers=None, buckets=None):
    self._lock = threading.Lock()
    self._name = name
    self._execPool = ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix=name)
    self._runCount = APPROX_COUNTERS.get(f"pool.{name}.concurrency")

    if not buckets:
      buckets = [1, 5, 10, 30, 31]

    self._hist = perfcounters.APPROX_COUNTERS.getHistogram(name=f"pool.{name}",
        buckets=buckets)

  def submit(self, func, *args, **kwargs):
    return self._execPool.submit(self.closeAndRun, func, *args, **kwargs)

  def closeAndRun(self, func, *args, **kwargs):
    try:
      close_old_connections()
    except Exception as e:
      LOGGER.exception(e)

    t1 = time.time()
    try:
      with self._lock:
        self._runCount += 1

      return func(*args, **kwargs)
    finally:
      diff = time.time() - t1
      bucket = self._hist.getBucket(diff)
      with self._lock:
        self._runCount -= 1
        self._hist.addToBucket(bucket)
