# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import io

from .json_utils import json_dumps

def StreamingResponseIfPossible(response):
  from django.conf import settings
  from django.http import StreamingHttpResponse

  # Within the response structure, the individual response.rows data is possibly a large one
  # Do not make another big copy of this string by calling json_dumps() on response.
  # Instead, use the StreamingJsonStr class to use this in the context of a
  # StreamingHttpResponse return from django. See special handling in StreamingJsonStr
  # to see special handling of the rows to optimize for memory usage there.
  if getattr(settings, 'CLEARTEXT_DSREQ', True):
    sj = StreamingJsonStr(response)
    return StreamingHttpResponse(sj, content_type='application/json')

  s = json_dumps(response)
  s = obfuscate(s)
  return HttpResponse(s, content_type='application/octet-stream')

class StreamingJsonStr(object):
  def __init__(self, o):
    self._o = o
    self._item_sep = ","
    self._map_sep = ":"

  def __iter__(self):
    # Iterate through the json sections, and return as many strings as
    # make sense so that chunks are < chunkMax (10K) as far as possible
    #
    # We want to do this instead of returning every single chunk as
    # it gets to us from json_gen() because we do not want to send out
    # tiny messages like "{" over the wire - it is more optimal to combine
    # them here
    #

    curr = io.StringIO()
    chunkMax = 10240
    for x in self.json_gen(self._o):
      l = len(x)

      if curr.tell() + l > chunkMax:
        s = curr.getvalue()
        curr.close()
        yield s

        curr = io.StringIO()
        yield x
        continue

      curr.write(x)

    if curr.tell() > 0:
      s = curr.getvalue()
      curr.close()
      yield s

  def json_gen(self, o):
    if type(o) == type([]):
      yield "["
      first = True
      for i in o:
        if first:
          first = False
        else:
          yield self._item_sep

        for x in self.json_gen(i):
          yield x

      yield "]"
      return

    if type(o) == type({}):
      yield "{"
      first = True
      for k in o:
        if first:
          first = False
        else:
          yield self._item_sep

        yield json_dumps(str(k)) + self._map_sep

        if k == "rows" and type(o[k]) == type(''):
          # Special handling for rows, which is the large dataset
          # that is being returned here. The reason we are doing all
          # of the special handling of this response to the client
          # is the line of code below, where we do not have to re-process
          # this large string, and instead can insert it into the output
          # network stream as-is
          yield o[k]
        else:
          for x in self.json_gen(o[k]):
            yield x

      yield "}"
      return

    yield json_dumps(o)

