# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from util import cml_models

import unittest

class TestCmlModel(unittest.TestCase):
  def test_embedding_response_transformer(self):
    body = {
        'response': [1.1, 2.2]
    }

    body2 = cml_models.embedding_response_transformer(body)

    self.assertEquals(
        body2, {'data': [{'embedding': [1.1, 2.2]}]})

  def test_completion_request_transformer_no_history(self):
    headers = {
        'a': 'b',
    }

    payload = {
        'request': {
            'prompt': [
                {
                    'content': 'SM1',
                    'role': 'system'
                },
                {
                    'content': 'AM1',
                    'role': 'assistant'
                },
                {
                    'content': 'UM1',
                    'role': 'user'
                },
            ]
        }
    }

    headers2, payload2 = cml_models.completion_request_transformer(
        headers, payload)

    # Headers should remain unchanged
    self.assertEquals(headers2, headers)

    EXPECTED_MSG_1 = "<s>[INST]<<SYS>>\nSM1\n<</SYS>>[/INST]\n\nAM1</s><s>[INST]UM1[/INST]"

    self.assertEquals(
        payload2,
        {
            'request': {
                'prompt': EXPECTED_MSG_1
            }
        }
    )

  def test_completion_request_transformer(self):
    self.maxDiff = None
    headers = {
        'a': 'b',
    }

    payload = {
        'request': {
            'prompt': [
                {
                    'content': 'SM1',
                    'role': 'system'
                },
                {
                    'content': 'AM1',
                    'role': 'assistant'
                },
                {
                    'content': 'UM1',
                    'role': 'user'
                },
                {
                    'content': 'AM2',
                    'role': 'assistant'
                },
                {
                    'content': 'UM2',
                    'role': 'user'
                },
            ]
        }
    }

    headers2, payload2 = cml_models.completion_request_transformer(
        headers, payload)

    # Headers should remain unchanged
    self.assertEquals(headers2, headers)

    system_prompt_section = "<<SYS>>\nSM1\n<</SYS>>"
    chatbot_first_line_section = "\n\nAM1"

    formatted_message = f"<s>[INST]{system_prompt_section}[/INST]{chatbot_first_line_section}</s>"

    messages = payload['request']['prompt']
    assistant_messages = [m['content'] for m in messages if m['role'] == 'assistant']
    user_messages = [m['content'] for m in messages if m['role'] == 'user']

    chat_history = list(zip(user_messages[:-1], assistant_messages[1:]))

    for user_msg, model_answer in chat_history:
        formatted_message += f"<s>[INST]{user_msg}[/INST]{model_answer}</s>"

    last_question = f"<s>[INST]{user_messages[-1]}[/INST]"
    formatted_message += last_question

    EXPECTED_MSG_2 = formatted_message

    self.assertEquals(
        payload2,
        {
            'request': {
                'prompt': EXPECTED_MSG_2
            }
        }
    )

  def test_completion_response_transformer(self):
    body = {
        'response': '[INST]A[/INST]B[INST]C[/INST]D'
    }

    body2 = cml_models.completion_response_transformer(body)

    # only the part after the last [/INST] is returned
    self.assertEquals(
        body2, {'choices': [{'message': {'content': 'D'}}]})

  def test_completion_response_transformer_no_closing_inst(self):
    body = {
        'response': '[INST]A'
    }

    body2 = cml_models.completion_response_transformer(body)

    self.assertEquals(
        body2, {'choices': [{'message': {'content': '[INST]A'}}]})
