# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.test import TestCase
from unittest.mock import mock_open, patch

from util.ff_flipper import FeatureFlipper
from reports.ff_models import AppFeatureFlipper

class FeatureFlipperTest(TestCase):
    def setUp(self):
        FeatureFlipper.instance = None

    def tearDown(self):
        FeatureFlipper.instance = None

    @patch('reports.ff_models.AppFeatureFlipper.objects')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'label': 'a', 'info': None, 'value': True, 'level': 'DB'}])
    def test_when_defatult_and_db_should_choose_db_with_same_level(self, flippers_table):
        # Given
        flippers_1 = [{'id': 'a', 'value': False, 'level': 'DB'}]
        flippers_table.values.return_value = flippers_1
        # When
        ret = FeatureFlipper().get_flipper_value('a')
        # Then
        self.assertFalse(ret)

    @patch('util.ff_flipper.settings')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'label': '', 'value': True, 'level': 'URL'}])
    def test_when_url_and_settings_should_choose_url_with_same_level(self, settings):
        # Given
        flippers_1 = [{'id': 'a', 'value': False, 'level': 'URL'}]
        flippers_2 = [{'id': 'a', 'value': True, 'level': 'URL'}]
        settings.FEATURE_FLIPPERS = flippers_2
        ff = FeatureFlipper()
        ff.set_url_flippers(flippers_1)
        # When
        ret = FeatureFlipper().get_flipper_value('a')
        # Then
        self.assertFalse(ret)

    @patch('util.ff_flipper.settings')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'label': '', 'value': False, 'level': 'SETTINGS'}])
    def test_when_url_and_settings_should_choose_settings_with_different_level(self, settings):
        # Given
        flippers_1 = [{'id': 'a', 'value': False, 'level': 'URL'}]
        flippers_2 = [{'id': 'a', 'value': True, 'level': 'SETTINGS'}]
        settings.FEATURE_FLIPPERS = flippers_2
        ff = FeatureFlipper()
        ff.set_url_flippers(flippers_1)
        # When
        ret = ff.get_flipper_value('a')
        # Then
        self.assertTrue(ret)

    @patch('reports.ff_models.AppFeatureFlipper.objects')
    @patch('util.ff_flipper.settings')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'label': '', 'value': True, 'level': 'SETTINGS'}])
    def test_when_settings_and_db_should_choose_settings_with_same_level(self, settings, flippers_table):
        # Given
        flippers_1 = [{'id': 'a', 'value': False, 'level': 'SETTINGS'}]
        flippers_2 = [{'id': 'a', 'value': True, 'level': 'SETTINGS'}]
        settings.FEATURE_FLIPPERS = flippers_1
        flippers_table.values.return_value = flippers_2
        # When
        ret = FeatureFlipper().get_flipper_value('a')
        # Then
        self.assertFalse(ret)

    @patch('reports.ff_models.AppFeatureFlipper.objects')
    @patch('util.ff_flipper.settings')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'label': '', 'value': False, 'level': 'DB'}])
    def test_when_settings_and_db_should_choose_db_with_different_level(self, settings, flippers_table):
        # Given
        flippers_1 = [{'id': 'a', 'value': False, 'level': 'SETTINGS'}]
        flippers_2 = [{'id': 'a', 'value': True, 'level': 'DB'}]
        settings.FEATURE_FLIPPERS = flippers_1
        flippers_table.values.return_value = flippers_2
        # When
        ret = FeatureFlipper().get_flipper_value('a')
        # Then
        self.assertTrue(ret)

    @patch('reports.ff_models.AppFeatureFlipper.objects')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'label': 'a', 'info': 'extra help', 'value': True, 'level': 'DB'}])
    def test_when_defatult_and_db_should_choose_db_and_add_info_n_label(self, flippers_table):
        # Given
        flippers_1 = [{'id': 'a', 'value': False, 'level': 'DB'}]
        flippers_table.values.return_value = flippers_1
        # When
        flipper = FeatureFlipper().flippers[0]
        # Then
        self.assertEqual(flipper.label, 'a')
        self.assertEqual(flipper.info, 'extra help')

    @patch('reports.ff_models.AppFeatureFlipper.objects')
    @patch('util.ff_flipper.settings')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'label': 'a', 'info': None, 'value': False, 'level': 'URL'}])
    def test_when_all_and_level_4_should_choose_url(self, settings, flippers_table):
        # Given
        settings.FEATURE_FLIPPERS = [{'id': 'a', 'value': False, 'level': 'URL'}]
        flippers_table.values.return_value = [{'id': 'a', 'value': False, 'level': 'URL'}]
        ff = FeatureFlipper()
        ff.set_url_flippers([{'id': 'a', 'value': True, 'level': 'URL'}])
        # When
        ret = ff.get_flipper_value('a')
        # Then
        self.assertTrue(ret)

    @patch('reports.ff_models.AppFeatureFlipper.objects')
    @patch('util.ff_flipper.settings')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'label': 'a', 'info': None, 'value': False, 'level': 'SETTINGS'}])
    def test_when_all_and_level_3_should_choose_settings(self, settings, flippers_table):
        # Given
        settings.FEATURE_FLIPPERS = [{'id': 'a', 'value': True, 'level': 'SETTINGS'}]
        flippers_table.values.return_value = [{'id': 'a', 'value': False, 'level': 'URL'}]
        ff = FeatureFlipper()
        ff.set_url_flippers([{'id': 'a', 'value': False, 'level': 'URL'}])
        # When
        ret = ff.get_flipper_value('a')
        # Then
        self.assertTrue(ret)

    @patch('reports.ff_models.AppFeatureFlipper.objects')
    @patch('util.ff_flipper.settings')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'label': 'a', 'info': None, 'value': False, 'level': 'DB'}])
    def test_when_all_and_level_2_should_choose_db(self, settings, flippers_table):
        # Given
        settings.FEATURE_FLIPPERS = [{'id': 'a', 'value': False, 'level': 'URL'}]
        flippers_table.values.return_value = [{'id': 'a', 'value': True, 'level': 'URL'}]
        ff = FeatureFlipper()
        ff.set_url_flippers([{'id': 'a', 'value': False, 'level': 'URL'}])
        # When
        ret = ff.get_flipper_value('a')
        # Then
        self.assertTrue(ret)

    @patch('builtins.open', new_callable=mock_open)
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'value': False, 'level': 'DB'}])
    def test_when_change_n_more_than_default_level_should_not_change_flipper_value(self, mock_file):
        # Given
        ff = FeatureFlipper()
        # When
        ff.set_flipper('a', True, 'SETTINGS')
        # Then
        self.assertFalse(ff.get_flipper_value('a'))
        self.assertEqual(len(AppFeatureFlipper.objects.all()), 0)
        # Settings file will be written regardless
        mock_file.assert_called_with(f'None/settings_local.py', 'a')

    @patch('util.ff_flipper.settings')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'value': True, 'level': 'URL'}])
    def test_when_change_n_lower_than_current_level_should_not_change_flipper_value(self, settings):
        # Given
        settings.FEATURE_FLIPPERS = [{'id': 'a', 'value': True}]
        ff = FeatureFlipper()
        # When
        ff.set_flipper('a', False, 'DB')
        # Then
        self.assertTrue(ff.get_flipper_value('a'))
        self.assertEqual(len(AppFeatureFlipper.objects.all()), 1)

    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'value': False, 'level': 'SETTINGS'}])
    def test_when_change_n_higher_level_should_change_flipper_value(self):
        # Given
        ff = FeatureFlipper()
        # When
        ff.set_flipper('a', True, 'DB')
        # Then
        self.assertTrue(ff.get_flipper_value('a'))
        flipper = AppFeatureFlipper.objects.values()[0]
        self.assertTrue(flipper['value'])

    @patch('builtins.open', new_callable=mock_open)
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'value': False, 'level': 'URL'}])
    def test_when_change_2_times_should_change_flipper_value(self, mock_file):
        # Given
        ff = FeatureFlipper()
        # When
        ff.set_flipper('a', True, 'URL')
        ff.set_flipper('a', False, 'SETTINGS')
        # Then
        self.assertTrue(ff.get_flipper_value('a'))
        # Settings file will be written regardless
        mock_file.assert_called_with(f'None/settings_local.py', 'a')


    @patch('util.ff_flipper.settings')
    @patch('util.ff_flipper.DEFAULT_FLIPPERS', [{'id': 'a', 'value': False, 'level': 'SETTINGS'}])
    def test_when_2_occurs_in_same_level_should_choose_last_occur(self, settings):
        # Given & When
        settings.FEATURE_FLIPPERS = [{'id': 'a', 'value': False}, {'id': 'a', 'value': True}]
        ff = FeatureFlipper()
        # Then
        self.assertTrue(ff.get_flipper_value('a'))