# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import time

from django.test import TestCase
from django.http import HttpResponse

from . import http_async
from arcwebbase import access_info
from concurrent.futures import ThreadPoolExecutor
from unittest import mock

class MockUser(object):
  def __init__(self):
    self.id = "user"

class MockRequest(object):
  def __init__(self):
    self.method = "POST"

    self.POST = {}

    self.user = MockUser()

  def build_absolute_uri(self):
    return "testuri"

# Note that we are expecting that by waiting for 5 seconds
# (see optimistic_return_check_seconds being set to 5 in __init__),
# we will get immediate results without async. For the alternative
# when optimistic_return_check_seconds is set to 0, see AlwaysAsync below
class OptimisticAsync(TestCase):
  def setUp(self):
    self.asyncCalls = http_async.Async(concurrency=5,
        optimistic_return_check_seconds=5)

  def test_single_call(self):
    @self.asyncCalls
    def c(request):
      return HttpResponse(json.dumps("ok"), content_type="text/json")

    request = MockRequest()
    resp = c(request)
    j = json.loads(resp.content)
    self.assertEqual(j, "ok")

  def tearDown(self):
    # make sure nothing remains in the main thread
    access_info.clear()

  def test_access_info_propagation(self):
    async_task_access_info = None

    @self.asyncCalls
    def c(request):
      nonlocal async_task_access_info
      async_task_access_info = access_info.get()
      return HttpResponse(json.dumps("ok"), content_type="text/json")

    request = MockRequest()

    # store access info for the current (main) thread. This should
    # be propagated to the async worker thread
    access_info.store(access_info.AccessInfo(username='http_async_user1'))

    c(request)
    self.assertEquals(async_task_access_info.username, 'http_async_user1')


class CullChecks(TestCase):
  def setUp(self):
    self.asyncCalls = http_async.Async(concurrency=5,
        optimistic_return_check_seconds=0, request_timeout=5)

  def test_cull(self):
    sleep = True

    @self.asyncCalls
    def c(request):
      while sleep:
        time.sleep(0.1)
      return HttpResponse(json.dumps("ok"), content_type="text/json")

    asyncs = []

    for i in range(0, 199):
      request = MockRequest()
      resp = c(request)
      j = json.loads(resp.content)
      type = j["type"]
      id = j["id"]
      self.assertEqual(type, "async")
      asyncs.append(j)

    # for 6 seconds, send down requests for all even numbered requests
    # Note that request_timeout is set to 10 seconds, so all the odd numbered requests
    #   should be timed out now
    for i in range(0, 6):
      time.sleep(1)
      for idx, j in enumerate(asyncs):
        if idx % 2 == 0:
          type = j["type"]
          id = j["id"]

          request = MockRequest()
          request.POST["async"] = json.dumps({
            "type": "async",
            "id": id
          })
          resp = c(request)
          j = json.loads(resp.content)
          self.assertEqual(j["type"], "async")
          self.assertEqual(j["id"], id)

    new_asyncs = []
    # Send down 1 more requests to trigger a cull (it triggers at every 100 additions)
    for i in range(0, 1):
      request = MockRequest()
      resp = c(request)
      j = json.loads(resp.content)
      type = j["type"]
      id = j["id"]
      self.assertEqual(type, "async")
      new_asyncs.append(j)

    # we should not have errors for all the odd requests
    for idx, j in enumerate(asyncs):
      type = j["type"]
      id = j["id"]

      request = MockRequest()
      request.POST["async"] = json.dumps({
        "type": "async",
        "id": id
      })
      resp = c(request)
      j = json.loads(resp.content)

      if idx % 2 == 0:
        self.assertEqual(j["type"], "async")
        self.assertEqual(j["id"], id)
      else:
        self.assertEqual(j["type"], "async")
        self.assertEqual(j["error"], "notfound")

    sleep = False

class AsyncErrors(TestCase):
  def setUp(self):
    self.asyncCalls = http_async.Async(concurrency=5,
        optimistic_return_check_seconds=1)

  def test_DELETE_PUT_exception(self):
    @self.asyncCalls
    def c(request):
      return HttpResponse(json.dumps("ok"), content_type="text/json")

    request = MockRequest()
    request.method = "PUT"
    with self.assertRaises(ValueError) as err:
      resp = c(request)

    request.method = "DELETE"
    with self.assertRaises(ValueError) as err:
      resp = c(request)

    request.method = "abc"
    with self.assertRaises(ValueError) as err:
      resp = c(request)

  def test_wrong_user_wrong_id(self):
    sleep = True

    # user1 sends the first request and user2 gets results back
    # this should not be permitted
    @self.asyncCalls
    def c(request):
      while sleep:
        time.sleep(0.1)
      return HttpResponse(json.dumps("ok"), content_type="text/json")

    request = MockRequest()
    resp = c(request)
    j = json.loads(resp.content)
    type = j["type"]
    id = j["id"]
    self.assertEqual(type, "async")

    # send in the wrong user, get an error
    request = MockRequest()
    request.user.id = "otheruser"
    request.POST["async"] = json.dumps({
      "type": "async",
      "id": id
    })
    resp = c(request)
    j = json.loads(resp.content)
    self.assertEqual(j["type"], "async")
    self.assertEqual(j["error"], "notfound")

    # send in the wrong id, get an error
    request = MockRequest()
    request.POST["async"] = json.dumps({
      "type": "async",
      "id": id + 1
    })
    resp = c(request)
    j = json.loads(resp.content)
    self.assertEqual(j["type"], "async")
    self.assertEqual(j["error"], "notfound")

    sleep = False
    # verify the right user/id gets back results eventually
    for i in range(0, 20):
      request = MockRequest()
      request.POST["async"] = json.dumps({
        "type": "async",
        "id": id
      })
      resp = c(request)
      j = json.loads(resp.content)
      if j == "ok":
        return

      self.assertEqual(j["type"], "async")
      self.assertEqual(j["id"], id)
      time.sleep(0.1)

class AlwaysAsync(TestCase):
  def setUp(self):
    self.asyncCalls = http_async.Async(concurrency=5)

  def test_single_call(self):
    @self.asyncCalls
    def c(request):
      return HttpResponse(json.dumps("ok"), content_type="text/json")

    request = MockRequest()
    resp = c(request)
    j = json.loads(resp.content)
    type = j["type"]
    id = j["id"]
    self.assertEqual(type, "async")

    # Give the async call time to complete
    time.sleep(0.1)

    request = MockRequest()
    request.POST["async"] = json.dumps({
      "type": "async",
      "id": id
    })
    resp = c(request)
    j = json.loads(resp.content)
    self.assertEqual(j, "ok")

  def test_single_exception(self):
    @self.asyncCalls
    def c(request):
      raise Exception("error in call")

    request = MockRequest()
    resp = c(request)
    j = json.loads(resp.content)
    type = j["type"]
    id = j["id"]
    self.assertEqual(type, "async")

    # Give the async call time to complete
    time.sleep(0.1)

    request = MockRequest()
    request.POST["async"] = json.dumps({
      "type": "async",
      "id": id
    })
    resp = c(request)
    j = json.loads(resp.content)
    self.assertEqual(j["type"], "async")
    self.assertEqual(j["error"], "error in call")

  def test_concurrent_calls(self):
    @self.asyncCalls
    def c(request):
      time.sleep(0.1)
      return HttpResponse(json.dumps("ok"), content_type="text/json")

    responsel = []

    for i in range(0, 25):
      request = MockRequest()
      response = c(request)
      responsel.append(response)

    while len(responsel) > 0:
      time.sleep(1)
      pre = responsel
      responsel = []
      for response in pre:
        j = json.loads(response.content)
        if j == "ok":
          continue

        type = j["type"]
        self.assertEqual(type, "async")
        id = j["id"]

        request = MockRequest()
        request.POST["async"] = json.dumps({
          "type": "async",
          "id": id
        })
        response = c(request)
        responsel.append(response)

