# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from django.test import TestCase

from util.init_cml import create_cml_workflow_groups, check_cml_syncing_and_init_groups, createGuestGroup
from util.ff_flipper import FeatureFlipper
from django.contrib.auth.models import Group
from rbac.roles_constants import DEFAULT_USER_GROUP
from dataconn.cml_dependencies import VIZCML_RW, VIZCML_RO
from unittest.mock import patch, Mock, mock_open


class CmlInitTest(TestCase):
    def test_create_cml_workflow_groups(self):
        # Given & When
        create_cml_workflow_groups()
        # Then
        vizcml_rw = Group.objects.filter(name=VIZCML_RW).first()
        vizcml_ro = Group.objects.filter(name=VIZCML_RO).first()
        self.assertIsNotNone(vizcml_rw)
        self.assertIsNotNone(vizcml_ro)

    def test_create_guest_group(self):
        # Given & When
        createGuestGroup()
        # Then
        viz_guest_group = Group.objects.filter(name=DEFAULT_USER_GROUP).first()
        self.assertIsNotNone(viz_guest_group)

    @patch('builtins.open', new_callable=mock_open)
    @patch('dataconn.cml_connection_importer.get_cmldata')
    def test_when_no_cml_api_found_should_not_create_cml_groups_and_create_guest_group(self, cml_api, mock_logging):
        # Given
        FeatureFlipper().set_flipper('enable_cml_sync', True, 'SETTINGS')
        cml_api.side_effect = FeatureFlipper().set_flipper('enable_cml_sync', False, 'SETTINGS')
        cml_api.return_value = None
        # When
        check_cml_syncing_and_init_groups(True)
        # Then
        vizcml_rw = Group.objects.filter(name=VIZCML_RW).first()
        vizcml_ro = Group.objects.filter(name=VIZCML_RO).first()
        viz_guest_group = Group.objects.filter(name=DEFAULT_USER_GROUP).first()

        self.assertIsNone(vizcml_rw)
        self.assertIsNone(vizcml_ro)
        self.assertIsNotNone(viz_guest_group)

    @patch('builtins.open', new_callable=mock_open)
    @patch('dataconn.cml_connection_importer.get_cmldata')
    def test_when_cml_api_found_should_create_cml_groups_and_not_guest_group(self, get_cmldata, mock_logging):
        # Given
        FeatureFlipper().set_flipper('enable_cml_sync', True, 'SETTINGS')
        cmldata = Mock()
        cmldata.get_hadoop_username.return_value = ''
        cmldata.get_workload_password.return_value = None
        cmldata.get_project_dataconnections.return_value = []

        get_cmldata.return_value = cmldata
        # When
        check_cml_syncing_and_init_groups(True)
        # Then
        vizcml_rw = Group.objects.filter(name=VIZCML_RW).first()
        vizcml_ro = Group.objects.filter(name=VIZCML_RO).first()
        viz_guest_group = Group.objects.filter(name=DEFAULT_USER_GROUP).first()
        self.assertIsNotNone(vizcml_rw)
        self.assertIsNotNone(vizcml_ro)
        self.assertIsNone(viz_guest_group)

    @patch('builtins.open', new_callable=mock_open)
    @patch('dataconn.cml_connection_importer.get_cmldata')
    def test_when_no_cml_api_found_not_initialized_should_not_create_cml_groups(self, cml_api, mock_logging):
        # Given
        FeatureFlipper().set_flipper('enable_cml_sync', True, 'SETTINGS')
        cml_api.side_effect = FeatureFlipper().set_flipper('enable_cml_sync', False, 'SETTINGS')
        cml_api.return_value = None
        # When
        check_cml_syncing_and_init_groups(False)
        # Then
        vizcml_rw = Group.objects.filter(name=VIZCML_RW).first()
        vizcml_ro = Group.objects.filter(name=VIZCML_RO).first()
        viz_guest_group = Group.objects.filter(name=DEFAULT_USER_GROUP).first()

        self.assertIsNone(vizcml_rw)
        self.assertIsNone(vizcml_ro)
        self.assertIsNone(viz_guest_group)
