# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
from django.test import TestCase
from util.remove_redundant_dataset_privileges import get_unsaved_updated_roles
from datasets.models import DataSet
from datasets.models import DataConnection
from rbac.models import Role
from unittest import skip

# This is needed because the perms order is nondeterministic
def sortPerms(privileges):
  for privilege in privileges:
    privilege["perms"] = sorted(privilege["perms"])

class RemoveRedundantPrivilegesTest(TestCase):
  def setUp(self):
    self.base_role = Role.objects.create(
      privileges=json.dumps([{'ptype':'system','perms':[]}]),
    )

    self.data_connection1 = DataConnection.objects.create(dataconnection_name='Samples')
    self.dataset1 = DataSet.objects.create(
      id=11,
      dataset_name="Cereals",
      dataconnection=self.data_connection1,
      dataset_detail="main.cereals",
      dataset_type="singletable",
    )

  # Should not change anything
  # When connection id does not match the selected dataset's
  def test_different_data_connection(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "2",
            "dslist": ["-1"],
            "perms": ["ds_appedit"]
          },
          {
            "ptype": "dataset",
            "dcid": "2",
            "dslist": ["12"],
            "perms": ["ds_appedit"]
          }
        ])
      )
    ]

    datasets = [self.dataset1]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    self.assertEqual([], new_roles)

  # Should not change anything
  # When there is no privilege with the selected datset
  def test_different_datasets(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["12"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["12", "13"],
            "perms": ["ds_appview"]
          }
        ])
      )
    ]

    datasets = [self.dataset1]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    self.assertEqual([], new_roles)

  # Should keep
  #   the "all dataconnections and all datasets" privilege
  #   the "multiple datasets" privilege,
  #     but it should remove the selected dataset from it
  # and remove the other dataset type privileges
  def test_all_data_connections_with_most_permissions(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "-1",
            "dslist": ["-1"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "12", "13"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "14", "15"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["-1"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11"],
            "perms": ["ds_appview"]
          }
        ])
      )
    ]

    datasets = [self.dataset1]

    expected_privileges = [
      {
        "ptype": "system",
        "perms": []
      },
      {
          "ptype": "dataset",
          "dcid": "-1",
          "dslist": ["-1"],
          "perms": ["ds_appedit", "ds_appview"]
        },
        {
          "ptype": "dataset",
          "dcid": "1",
          "dslist": ["12", "13"],
          "perms": ["ds_appview"]
        },
        {
          "ptype": "dataset",
          "dcid": "1",
          "dslist": ["14", "15"],
          "perms": ["ds_appview"]
        },
    ]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    new_privileges = json.loads(new_roles[0].privileges)
    sortPerms(new_privileges)

    self.assertEqual(expected_privileges, new_privileges)

  # Should keep
  #   the "all dataconnections and all datasets" privilege
  #   the "all datasets" privilege (because it has more permission)
  #   the "multiple datasets" privilege,
  #     but it should remove the selected dataset from it
  # and remove the other dataset type privileges
  def test_all_data_connections_without_most_permissions(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "-1",
            "dslist": ["-1"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "12", "13"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "14", "15"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["-1"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11"],
            "perms": ["ds_appedit", "ds_appview"]
          }
        ])
      )
    ]

    datasets = [self.dataset1]

    expected_privileges = [
      {
        "ptype": "system",
        "perms": []
      },
      {
        "ptype": "dataset",
        "dcid": "-1",
        "dslist": ["-1"],
        "perms": ["ds_appview"]
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["-1"],
        "perms": ['ds_appedit', "ds_appview"]
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["12", "13"],
        "perms": ["ds_appedit", "ds_appview"]
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["14", "15"],
        "perms": ["ds_appedit", "ds_appview"]
      }
    ]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    new_privileges = json.loads(new_roles[0].privileges)
    sortPerms(new_privileges)

    self.assertEqual(expected_privileges, new_privileges)

  # Should keep
  #   the "all datasets" privilege
  #   the "multiple datasets" privilege,
  #     but it should remove the selected dataset from it
  # and remove the other dataset type privileges
  def test_all_datasets_with_most_permissions(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["-1"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "12", "13"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "14", "15"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11"],
            "perms": ["ds_appview"]
          }
        ])
      )
    ]

    datasets = [self.dataset1]

    expected_privileges = [
      {
        "ptype": "system",
        "perms": []
      },
      {
          "ptype": "dataset",
          "dcid": "1",
          "dslist": ["-1"],
          "perms": ["ds_appedit", "ds_appview"]
        },
        {
          "ptype": "dataset",
          "dcid": "1",
          "dslist": ["12", "13"],
          "perms": ["ds_appedit", "ds_appview"]
        },
        {
          "ptype": "dataset",
          "dcid": "1",
          "dslist": ["14", "15"],
          "perms": ["ds_appview"]
        },
    ]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    new_privileges = json.loads(new_roles[0].privileges)
    sortPerms(new_privileges)

    self.assertEqual(expected_privileges, new_privileges)

  # Should keep
  #   "all datasets" privilege
  #   "multiple datasets" privilege,
  #     but it should remove the selected dataset from it
  #   "specific dataset" privilege (because it has more permissions)
  def test_all_datasets_without_most_permissions(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["-1"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "12", "13"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "14", "15"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11"],
            "perms": ["ds_appedit", "ds_appview"]
          }
        ])
      )
    ]

    datasets = [self.dataset1]

    expected_privileges = [
      {
        "ptype": "system",
        "perms": []
      },
      {
          "ptype": "dataset",
          "dcid": "1",
          "dslist": ["-1"],
          "perms": ["ds_appview"]
        },
        {
          "ptype": "dataset",
          "dcid": "1",
          "dslist": ["11"],
          "perms": ["ds_appedit", "ds_appview"]
        },
        {
          "ptype": "dataset",
          "dcid": "1",
          "dslist": ["12", "13"],
          "perms": ["ds_appview"]
        },
        {
          "ptype": "dataset",
          "dcid": "1",
          "dslist": ["14", "15"],
          "perms": ["ds_appview"]
        },
    ]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    new_privileges = json.loads(new_roles[0].privileges)
    sortPerms(new_privileges)

    self.assertEqual(expected_privileges, new_privileges)

  # Should keep
  #   the "multiple datasets" privilege,
  #     because it has more permissions
  # and remove "single dataset" privilege
  def test_multiple_datasets_with_most_permissions(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "12", "13"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11"],
            "perms": ["ds_appview"]
          }
        ])
      )
    ]

    datasets = [self.dataset1]

    expected_privileges = [
      {
        "ptype": "system",
        "perms": []
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["11", "12", "13"],
        "perms": ["ds_appedit", "ds_appview"]
      }
    ]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    new_privileges = json.loads(new_roles[0].privileges)
    sortPerms(new_privileges)

    self.assertEqual(expected_privileges, new_privileges)

    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "12", "13"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11"],
            "perms": ["ds_appview"]
          }
        ])
      )
    ]

    expected_privileges = [
      {
        "ptype": "system",
        "perms": []
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["11", "12", "13"],
        "perms": ["ds_appedit", "ds_appview"]
      }
    ]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    new_privileges = json.loads(new_roles[0].privileges)
    sortPerms(new_privileges)

    self.assertEqual(expected_privileges, new_privileges)

  # Should keep
  #   both privileges
  #   but remove the dataset from the "multiple datasets" privilege
  def test_specific_dataset_with_most_permissions(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "12", "13"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11"],
            "perms": ["ds_appedit", "ds_appview"]
          }
        ])
      )
    ]

    datasets = [self.dataset1]

    expected_privileges = [
      {
        "ptype": "system",
        "perms": []
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["11"],
        "perms": ["ds_appedit", "ds_appview"]
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["12", "13"],
        "perms": ["ds_appview"]
      }
    ]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    new_privileges = json.loads(new_roles[0].privileges)
    sortPerms(new_privileges)

    self.assertEqual(expected_privileges, new_privileges)

  # Should keep
  #   both privileges
  #   but remove the dataset from the "multiple datasets" privilege
  def test_two_multiple_datasets(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "12", "13"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "14", "15"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11"],
            "perms": ["ds_appview"]
          }
        ])
      )
    ]

    datasets = [self.dataset1]

    expected_privileges = [
      {
        "ptype": "system",
        "perms": []
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["11", "12", "13"],
        "perms": ["ds_appedit", "ds_appview"]
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["14", "15"],
        "perms": ["ds_appview"]
      }
    ]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    new_privileges = json.loads(new_roles[0].privileges)
    sortPerms(new_privileges)

    self.assertEqual(expected_privileges, new_privileges)

  # Should keep
  #   both privileges
  #   but remove the dataset from the "multiple datasets" privilege
  def test_two_multiple_datasets_2(self):
    roles = [
      Role.objects.create(
        privileges=json.dumps([
         {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11", "12", "13"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["11"],
            "perms": ["ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["14", "15"],
            "perms": ["ds_appedit", "ds_appview"]
          },
          {
            "ptype": "dataset",
            "dcid": "1",
            "dslist": ["14"],
            "perms": ["ds_appview"]
          }
        ])
      )
    ]

    dataset2 = DataSet.objects.create(
      id=14,
      dataset_name="Pyramids",
      dataconnection=self.data_connection1,
      dataset_type="singletable",
      dataset_detail="main.pyramids"
    )
    datasets = [self.dataset1, dataset2]

    expected_privileges = [
      {
        "ptype": "system",
        "perms": []
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["11", "12", "13"],
        "perms": ["ds_appedit", "ds_appview"]
      },
      {
        "ptype": "dataset",
        "dcid": "1",
        "dslist": ["14", "15"],
        "perms": ["ds_appedit", "ds_appview"]
      },
    ]

    new_roles = get_unsaved_updated_roles(roles, datasets)
    new_privileges = json.loads(new_roles[0].privileges)
    sortPerms(new_privileges)

    self.assertEqual(expected_privileges, new_privileges)