# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from apps.tests import tests_views_visuals_page
from datasets.models import DataSet
from django.contrib.auth.models import User
from rbac.models import Role
from rbac.permcache import groups_for_user
from util.ff_flipper import FeatureFlipper
from util.get_dataset_ids_accessible_by_user import get_dataset_ids_accessible_by_user, get_dataset_ids_managable_by_user, is_dataset_superuser
from vizlib.tests import test_report_operations
from vizlib.tests.test_report_operations import make_user, create_client
from vizlib.tests.test_sqlrun import Common
from vizlib.tests.util import run_with_db_query_check



import django.test
import json

def grant_explore_on_dataconnection(user: User, dc_id: int):
  role = Role(name=f'explore_on_{dc_id}')
  privs = [
    {
      'ptype': 'dataconn',
      'dclist': [str(dc_id)],
      'perms': ['dc_explore'],
    },
  ]
  role.privileges = json.dumps(privs)
  # add_users automatically saves:
  role.add_users([user.username])

class TestDatasetAccess(django.test.TestCase):

  def setUp(self):
    super().setUp()
    util = test_report_operations.TestDashboardOperations()
    self.util = util

    # create admin
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    self.admin = admin
    self.admin_client = admin_client

    # create user1
    user1 = make_user('user1', is_superuser=False)
    user1_client = create_client(user1)
    self.user1 = user1
    self.user1_client = user1_client

    dc_id1 = util.create_dataconnection(admin_client)
    self.dc_id1 = dc_id1

    ds1_id = util.create_dataset(admin_client, dc_id1, 'ds1')
    self.ds1_id = ds1_id

  def tearDown(self):
      FeatureFlipper.instance = None

  def do_run_action(self, user: User, action: callable, expected_query_count: int=None):

    # groups are cached on the user object. Clear them:
    if hasattr(user, '_arcadia_groupset'):
      del user._arcadia_groupset

    result, _ = run_with_db_query_check(
      action,
      asserter=self,
      expected_query_count=expected_query_count
    )

    return result

  def get_dataset_ids(self, user: User, expected_query_count: int=None):
    return self.do_run_action (
      user=user,
      action=lambda: get_dataset_ids_accessible_by_user(user),
      expected_query_count=expected_query_count
    )

  def get_dataset_ids_with_manage_permission(self, user: User, expected_query_count: int=None):
    return self.do_run_action (
      user=user,
      action=lambda: get_dataset_ids_managable_by_user(user),
      expected_query_count=expected_query_count
    )

  def test_admin(self):
    self.assertTrue(is_dataset_superuser(self.admin))
    self.assertIsNone(self.get_dataset_ids(self.admin, expected_query_count=0))
    self.assertIsNone(self.get_dataset_ids_with_manage_permission(self.admin, expected_query_count=0))

  def test_user(self):
    self.assertFalse(is_dataset_superuser(self.user1))

    expected_ds_ids = [] # user has no visible datasets:
    actual_ds_ids = self.get_dataset_ids(self.user1, expected_query_count=3)
    self.assertEqual(actual_ds_ids, expected_ds_ids)

    # grant permissions on a dataconnection so that the user can create datasets:
    dc_id2 = self.util.create_dataconnection(self.admin_client, 'DC 2')
    grant_explore_on_dataconnection(self.user1, dc_id2)

    # create a dataset:
    ds_1 = self.util.create_dataset(self.user1_client, dc_id2, 'User-created dataset - 1')

    expected_ds_ids = [ds_1]
    actual_ds_ids = self.get_dataset_ids(self.user1, expected_query_count=3)
    self.assertEqual(actual_ds_ids, expected_ds_ids)

    # create a second dataset to see if query count is constant
    ds_2 = self.util.create_dataset(self.user1_client, dc_id2, 'User-created dataset - 2')
    expected_ds_ids = {ds_1, ds_2}
    actual_ds_ids = self.get_dataset_ids(self.user1, expected_query_count=3)
    self.assertEqual(len(actual_ds_ids), 2)
    self.assertEqual(set(actual_ds_ids), expected_ds_ids)

    # grant access to the admin-created dataset
    tests_views_visuals_page.grant_read_access_on_dataset(self.user1, self.dc_id1, self.ds1_id)
    expected_ds_ids = {ds_1, ds_2, self.ds1_id}
    actual_ds_ids = self.get_dataset_ids(self.user1, expected_query_count=3)
    self.assertEqual(len(actual_ds_ids), 3)
    self.assertEqual(set(actual_ds_ids), expected_ds_ids)

  def test_user_datasets_with_manage_permission(self):
    self.assertFalse(is_dataset_superuser(self.user1))

    expected_ds_ids = [] # user has no visible datasets:
    actual_ds_ids = self.get_dataset_ids_with_manage_permission(self.user1, expected_query_count=3)
    self.assertEqual(actual_ds_ids, expected_ds_ids)

    # grant permissions on a dataconnection so that the user can create datasets:
    dc_id2 = self.util.create_dataconnection(self.admin_client, 'DC 2')
    grant_explore_on_dataconnection(self.user1, dc_id2)

    # create a dataset:
    ds_1 = self.util.create_dataset(self.user1_client, dc_id2, 'User-created dataset - 1')

    expected_ds_ids = [ds_1]
    actual_ds_ids = self.get_dataset_ids_with_manage_permission(self.user1, expected_query_count=3)
    self.assertEqual(actual_ds_ids, expected_ds_ids)

    # create a second dataset to see if query count is constant
    ds_2 = self.util.create_dataset(self.user1_client, dc_id2, 'User-created dataset - 2')
    expected_ds_ids = {ds_1, ds_2}
    actual_ds_ids = self.get_dataset_ids_with_manage_permission(self.user1, expected_query_count=3)
    self.assertEqual(len(actual_ds_ids), 2)
    self.assertEqual(set(actual_ds_ids), expected_ds_ids)

    # grant read access to the admin-created dataset
    tests_views_visuals_page.grant_read_access_on_dataset(self.user1, self.dc_id1, self.ds1_id)
    expected_ds_ids = {ds_1, ds_2}
    actual_ds_ids = self.get_dataset_ids_with_manage_permission(self.user1, expected_query_count=3)
    self.assertEqual(len(actual_ds_ids), 2) # same as above, no manage permission on self.ds1_id
    self.assertEqual(set(actual_ds_ids), expected_ds_ids)

    # grant manage access to the admin-created dataset (this is full access, ds_manage implies the other two perms)
    tests_views_visuals_page.grant_full_access_on_dataset(self.user1, self.dc_id1, self.ds1_id)
    expected_ds_ids = {ds_1, ds_2, self.ds1_id}
    actual_ds_ids = self.get_dataset_ids_with_manage_permission(self.user1, expected_query_count=3)
    self.assertEqual(len(actual_ds_ids), 3)
    self.assertEqual(set(actual_ds_ids), expected_ds_ids)

  def test_versioning_effect(self):

    tests_views_visuals_page.grant_read_access_on_dataset(self.user1, self.dc_id1, self.ds1_id)
    actual_ds_ids = self.get_dataset_ids(self.user1, expected_query_count=3)
    self.assertEqual(set(actual_ds_ids), {self.ds1_id})

    from datasets.test_views import enable_dataset_versioning
    enable_dataset_versioning()

    ds1 = DataSet.objects.get(id=self.ds1_id)
    sqlrun_common = Common()

    # with versioning enabled, creating segments creates dataset versions
    self.s1 = sqlrun_common.make_segment(
      self.admin_client, ds1, 's1', '[A] in ("doug", "veronica")')
    self.s11 = sqlrun_common.make_segment(
      self.admin_client, ds1, 's11', '[AA] in ("dougA", "veronicaA")')

    # dataset id count must not change
    actual_ds_ids = self.get_dataset_ids(self.user1, expected_query_count=3)
    self.assertEqual(set(actual_ds_ids), {self.ds1_id})