# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from collections import defaultdict
from django.db.models.manager import BaseManager
from instant import instant
from util.get_sample_dashboards_query import get_sample_dashboards_query
from vizlib.tests.test_report_operations import make_user
from vizlib.tests.util import run_with_db_query_check
from workspace.models import Workspace

import django.test


class TestSampleDashboardsQuery(django.test.TestCase):

  def setUp(self):
    super().setUp()

  def run_with_check(self, action: callable, expected_query_count: int):
    result, _ = run_with_db_query_check(
      lambda: action(),
      asserter=self,
      expected_query_count=expected_query_count
    )
    return result

  def createSamples(self):
    make_user(instant.ADMIN_USERNAME, is_superuser=True)
    instant.createSamples()

  def test_error_case(self):
    with self.assertRaises(Exception):
      get_sample_dashboards_query(None, None)

  def test_unfiltered(self):
    self.createSamples()

    query = self.run_with_check(
      lambda: get_sample_dashboards_query(None, None), expected_query_count=4)
    samples = self.run_with_check(lambda: list(query), expected_query_count=1)
    self.assertEquals(len(samples), 13)  # there are 13 sample dashboards

  def test_filtered(self):
    self.createSamples()

    query = self.run_with_check(
      lambda: get_sample_dashboards_query(None, None), expected_query_count=4)
    samples = self.run_with_check(lambda: list(query), expected_query_count=1)

    dataset_to_dashbboards = defaultdict(list)
    for s in samples:
      dataset_to_dashbboards[s.dataset_id].append(s.id)

    # find the dataset with most dashboards
    ds_id, ds_report_ids = sorted(dataset_to_dashbboards.items(
    ), key=lambda item: len(item[1]), reverse=True)[0]

    self.assertGreater(len(ds_report_ids), 1)
    self.assertLess(len(ds_report_ids), len(samples))

    sample_1 = [s for s in samples if s.id == ds_report_ids[0]][0]
    sample_2 = [s for s in samples if s.id not in ds_report_ids][0]

    # assign some of the samples to a different workspace
    ws2 = Workspace.objects.create(name='ws2')
    sample_1.workspace = ws2
    sample_1.save()

    sample_2.workspace = ws2
    sample_2.save()

    # filter by dataset id:
    query = self.run_with_check(lambda: get_sample_dashboards_query(dataset_ids_accessible_by_user=[
                                ds_id], workspace_ids_accessible_by_user=None), expected_query_count=4)
    samples = self.run_with_check(lambda: list(query), expected_query_count=1)

    self.assertEquals(len(samples), len(ds_report_ids))

    # filter by workspace id:
    query = self.run_with_check(lambda: get_sample_dashboards_query(
      dataset_ids_accessible_by_user=None, workspace_ids_accessible_by_user=[ws2.id]), expected_query_count=4)
    samples = self.run_with_check(lambda: list(query), expected_query_count=1)

    self.assertEquals({s.id for s in samples}, {sample_1.id, sample_2.id})

    # filter by workspace and dataset:
    query = self.run_with_check(lambda: get_sample_dashboards_query(dataset_ids_accessible_by_user=[
                                ds_id], workspace_ids_accessible_by_user=[ws2.id]), expected_query_count=4)
    samples = self.run_with_check(lambda: list(query), expected_query_count=1)

    self.assertEquals([s.id for s in samples], [sample_1.id])
