# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from django.contrib.auth.models import User
from rbac.models import Role
from util.get_workspace_ids_accessible_by_user import get_workspace_ids_accessible_by_user, is_workspace_superuser
from vizlib.tests.test_report_operations import make_user
from vizlib.tests.util import run_with_db_query_check
from workspace.models import WSACL_Level, WSACL_Member, Workspace


import django.test
import json


def grant_sys_misc(user: User):
  role = Role(name=f'sys_misc for user {user.username}')
  privs = [
    {
      'ptype': 'system',
      'perms': ['sys_misc'],
    },
  ]
  role.privileges = json.dumps(privs)
  # add_users automatically saves:
  role.add_users([user.username])


def grant_view_permission_on_workspace(ws: Workspace, user: User):
  acl = [[WSACL_Member.USER, WSACL_Level.VIEW, user.username]]
  ws.acl_obj = acl
  ws.save()


class TestWorkspaceAccess(django.test.TestCase):

  def setUp(self):
    super().setUp()

    # create admin
    admin = make_user('admin', is_superuser=True)
    self.admin = admin

    # create user1
    user1 = make_user('user1', is_superuser=False)
    self.user1 = user1

  def get_workspace_ids(self, user: User, expected_query_count: int = None):

    # groups are cached on the user object. Clear them:
    if hasattr(user, '_arcadia_groupset'):
      del user._arcadia_groupset

    result, _ = run_with_db_query_check(
      lambda: get_workspace_ids_accessible_by_user(user),
      asserter=self,
      expected_query_count=expected_query_count
    )
    return result

  def test_admin(self):
    self.assertTrue(is_workspace_superuser(self.admin))
    self.assertIsNone(self.get_workspace_ids(
      self.admin, expected_query_count=0))

  def test_user(self):
    # we have the public workspace
    self.assertEqual(Workspace.objects.count(), 1)
    public_ws_id = Workspace.objects.values_list('id', flat=True)[0]

    self.assertFalse(is_workspace_superuser(self.user1))

    actual_ds_ids = self.get_workspace_ids(self.user1, expected_query_count=3)
    # we have access to the public workspace
    self.assertEqual(actual_ds_ids, [public_ws_id])

    ws2 = Workspace.objects.create(name='ws2')
    self.assertEqual(Workspace.objects.count(), 2)

    actual_ds_ids = self.get_workspace_ids(self.user1, expected_query_count=3)
    # no access to the new workspace
    self.assertEqual(actual_ds_ids, [public_ws_id])

    grant_view_permission_on_workspace(ws2, self.user1)
    actual_ds_ids = self.get_workspace_ids(self.user1, expected_query_count=3)
    self.assertEqual(set(actual_ds_ids), {public_ws_id, ws2.id})

  def test_workspace_superuser(self):
    self.assertFalse(is_workspace_superuser(self.user1))

    grant_sys_misc(self.user1)
    self.assertTrue(is_workspace_superuser(self.user1))

    actual_ds_ids = self.get_workspace_ids(
      self.user1, expected_query_count=2)  # permissions need to be retrieved
    self.assertIsNone(actual_ds_ids)  # unlimited access to workspaces
