#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Create users from CSV file

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

import os
import os.path
import sys
import re
import json
import optparse
from pprint import pprint
import traceback
# from django.core.management import execute_from_command_line, call_command
from messytables import CSVTableSet, type_guess, \
  types_processor, headers_guess, headers_processor, \
  offset_processor, any_tableset, StringType
from django.conf import settings
from django_auth_ldap.config import _LDAPConfig, LDAPSearch
import ldap
from ldap.controls import SimplePagedResultsControl

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)



import warnings
with warnings.catch_warnings():
  warnings.filterwarnings("ignore", category=DeprecationWarning)


class LDAPSearchExtended(LDAPSearch):

    def __init__(self, search_obj):
        self.base_dn = search_obj.base_dn
        self.scope = search_obj.scope
        self.filterstr = search_obj.filterstr
        self.attrlist = search_obj.attrlist

    # Extended search method based on the original .execute method from LDAPSearch
    def execute(self, connection, filterargs=(), escape=True, page_size=200):
        """
            Executes the search on the given connection (an LDAPObject). filterargs
            is an object that will be used for expansion of the filter string.
            If escape is True, values in filterargs will be escaped.
            The python-ldap library returns utf8-encoded strings. For the sake of
            sanity, this method will decode all result strings and return them as
            Unicode.
            """
        page_control = SimplePagedResultsControl(True, size=page_size, cookie='')
        print("LDAP Pagination page size set to %s" % str(page_size))

        if escape:
            filterargs = self._escape_filterargs(filterargs)

        try:
            filterstr = self.filterstr % filterargs

            response = connection.search_ext(self.base_dn,
                                             self.scope,
                                             filterstr,
                                             [],
                                             serverctrls=[page_control])

            results = []
            pages = 0
            while True:
                print("Processing LDAP Results page: %s " % str(pages))
                pages += 1
                rtype, rdata, rmsgid, serverctrls = connection.result3(response)
                results.extend(rdata)
                controls = [control for control in serverctrls
                            if control.controlType == SimplePagedResultsControl.controlType]
                if not controls:
                    print('The server ignores RFC 2696 control. Ending paginated search.')
                    break
                if not controls[0].cookie:
                    print("No more pages. Ending paginated search.")
                    break
                page_control.cookie = controls[0].cookie
                response = connection.search_ext(self.base_dn,
                                                 self.scope,
                                                 filterstr,
                                                 [],
                                                 serverctrls=[page_control])

            print("Total users from LDAP Group searched: %s " % str(len(results)))

        except ldap.LDAPError as e:
            results = []
            print("search_s('{}', {}, '{}') raised {}".format(self.base_dn, self.scope, filterstr, pprint.pformat(e)))

        return self._process_results(results)


def _populate_user_from_attributes(ldapuser, attribute_map):
  user = {}
  user["DN"] = ldapuser[0]
  for field, attr in attribute_map.items():
    try:
      value = ldapuser[1][attr][0]
    except LookupError:
      pass
    else:
      user[field] = value
  return user


def ldap_create(ldapgroup, userprofileattrs=None, dryrun=True, paginate=False):
  from django.contrib.auth.models import User, Group
  from django.utils.encoding import force_str

  options = getattr(settings, 'AUTH_LDAP_GLOBAL_OPTIONS', None)

  uri = getattr(settings, 'AUTH_LDAP_SERVER_URI', None)

  # Setup ldap connection
  _ldap = _LDAPConfig.get_ldap(options)
  _connection = _ldap.initialize(uri, bytes_mode=False)

  for opt, value in settings.AUTH_LDAP_CONNECTION_OPTIONS.items():
    _connection.set_option(opt, value)

  if getattr(settings, 'AUTH_LDAP_START_TLS', False):
    _connection.start_tls_s()

  _connection.simple_bind_s(force_str(settings.AUTH_LDAP_BIND_DN),
                            force_str(settings.AUTH_LDAP_BIND_PASSWORD))

  # Default to AUTH_LDAP_REQUIRE_GROUP if no group DN given
  if ldapgroup == "":
    ldapgroup = settings.AUTH_LDAP_REQUIRE_GROUP

  # Base search from AUTH_LDAP settings, but update to search for all users
  _user_search = settings.AUTH_LDAP_USER_SEARCH
  user_attr = re.match(r'.*\((.+)=%\(user\)', _user_search.filterstr)[1]
  object_class = re.search(r'objectClass=(.+)\)\(', _user_search.filterstr)
  if object_class:
    object_class = object_class[1]
    _user_search.filterstr = "(objectClass=%s)" % object_class
  else:
    if user_attr.lower() == "uid":
      object_class = "posixaccount"
      _user_search.filterstr = "(objectClass=posixaccount)"
    elif user_attr.lower() == "samaccountname":
      object_class = "user"
      _user_search.filterstr = "(objectClass=user)"
    else:
      object_class = "user"
      _user_search.filterstr = "(|(objectClass=posixaccount)(objectClass=user))"
  # _user_search.filterstr = "(objectClass=%s)" % object_class

  if object_class.lower() == "posixaccount":
    # OpenLDAP does not support memberof queries by default
    # so we will lookup the group and find the memberuids
    _gsearch = LDAPSearch(ldapgroup, ldap.SCOPE_BASE)
    # run the query on the LDAP server
    _result = _gsearch.execute(_connection)
    usernames = _result[0][1]["memberuid"]
    filterstr = '(|%s)' % "".join(["(uid=%s)" % u for u in usernames])
    # Add users to filter
    _search = _user_search.search_with_additional_term_string(filterstr)
  else:
    # Add group DN to filter
    _search = _user_search.search_with_additional_terms({"memberOf": ldapgroup})

  # run the query on the LDAP server
  #   If pagination option used, then user our LDAPSearchExtended super class
  #   to extend the search and page through results
  if paginate:
    _search_ext = LDAPSearchExtended(_search)
    _result = _search_ext.execute(_connection)
  #   Else, run the LDAP user search normally
  else:
    _result = _search.execute(_connection)

  # Create a list of users based of the AUTH_LDAP_USER_ATTR_MAP mapping
  users = []
  attribute_map = {'username': user_attr}
  attribute_map.update(settings.AUTH_LDAP_USER_ATTR_MAP)
  for ldapuser in _result:
    users.append(_populate_user_from_attributes(ldapuser, attribute_map))

  # Get all current system users
  system_users = User.objects.all().values_list("username", flat=True)

  # Find new users
  ldap_usernames = set([u["username"] for u in users])
  # common_users = ldap_usernames.intersection(system_users)
  new_usernames = ldap_usernames.difference(system_users)
  new_users = [u for u in users if u["username"] in new_usernames]

  if dryrun:
    print()
    print('Users to be added from LDAP group: "%s"' % ldapgroup)
    if not new_users:
      print("No new users found.")
    for u in new_users:
      print('%s = "%s"' % (u["username"], u["DN"]))
  else:
    print()
    print('Adding Users from LDAP group: "%s"' % ldapgroup)
    for u in new_users:
      print('%s = "%s"' % (u["username"], u["DN"]))

      # Check for user exists but with incorrect casing, fixup as needed
      ul = User.objects.filter(username__iexact=u["username"])
      if len(ul) == 1:
        user = ul[0]
        new_name = u["username"]
        if user.username != new_name:
          print('%s -> %s' % (user.username, new_name))
          rename_user(user, new_name)
        continue

      new_user = User()
      # Set LDAP user with unusable password by default
      new_user.set_unusable_password()
      for k, v in u.items():
        setattr(new_user, k, v)
      new_user.save()
      if userprofileattrs:
        for a in userprofileattrs:
          k, v = a.split(":", maxsplit=1)
          print('  %s = "%s"' % (k, v))
          new_user.userprofile.setkey(k, v)
          new_user.save()


def rename_user(user, new_name):
  from rbac.models import Role
  from workspace.models import Workspace, WSACL_Member

  old_name = user.username
  user.username = new_name
  user.save()

  for r in Role.objects.all():
    if old_name in r.userSet():
      r.delete_users([old_name])
      r.add_users([new_name])

  for w in Workspace.objects.all():
    changed = False
    acl_obj = w.acl_obj
    for acl in acl_obj:
      if acl[0] == WSACL_Member.USER and acl[2] == old_name:
        acl[2] = new_name
        changed = True
      if changed:
        w.acl_obj = acl_obj
        w.save()

def user_attr(username=None, getattr=None, setattr=None, delattr=None):
  from django.contrib.auth.models import User, Group

  if not username:
    print("No username specified for user profile attribute operation")
    sys.exit(1)
  user = User.objects.get(username=username)
  print("User: %s" % user.username)
  print("=" * 30)
  # Delete attributes
  if delattr:
    attrs = user.userprofile.user_data_obj
    for a in delattr:
      attrs.pop(a, None)
    user.userprofile.user_data_obj = attrs
    user.userprofile.save()
  # Set attributes
  if setattr:
    for a in setattr:
      k, v = a.split(":")
      user.userprofile.setkey(k, v)
  # Get attributes
  attrs = user.userprofile.user_data_obj
  if getattr is not None:
    if getattr == [""]:
      getattr = attrs.keys()
    for a in getattr:
      print("%s: %s" % (a, attrs.get(a, "undefined")))


def create_user(username, firstname='', lastname='', email='', password='', groups=None, allgroups=None,
                overwrite=False):
  from django.contrib.auth.models import User, Group
  groups = groups or []
  allgroups = allgroups or {}
  if not overwrite and User.objects.filter(username=username).exists():
    print("error: user '%s' exists" % username)
    exit(1)
  u, newuser = User.objects.get_or_create(username=username)
  if firstname:
    u.first_name = firstname
  if lastname:
    u.last_name = lastname
  if email:
    u.email = email
  if password and password is not '':
    u.set_password(password)
  else:
    u.set_unusable_password()
  if groups:
    u.groups.clear()
    u.save()
    for g in groups:
      u.groups.add(getattr(allgroups, g, Group.objects.get_or_create(name=g)[0]))
  u.save()
  if newuser:
    print("Created user: %s: %s" % (u.username, u.get_full_name()))
  else:
    print("Modified user: %s: %s" % (u.username, u.get_full_name()))


def delete_user(username):
  from django.contrib.auth.models import User, Group
  print("Deleting user '%s'" % username)
  User.objects.filter(username=username).delete()


def list_users(csv=False):
  from django.contrib.auth.models import User, Group
  users = User.objects.all()
  print()
  print("===================================================================")
  if csv:
    print("username,firstname,lastname,email,password,groups")
    for u in users:
      groups = '"' + ", ".join(u.groups.values_list('name', flat=True)) + '"' if u.groups.count() > 0 else ''
      print("%s,%s,%s,%s,,%s" % (u.username, u.first_name, u.last_name, u.email, groups))
  else:
    print("Users:")
    for u in users:
      print("%s: %s <%s>" % (u.username, u.get_full_name(), u.email))


def bulk_create(filename, overwrite=False, delete_extra=False):
  from django.contrib.auth.models import User, Group
  try:
    fh = open(filename, 'rb')
  except:
    print("error:  error opening bulkfile: '%s'" % filename)
    exit(1)
  # Load a file object:
  # table_set = CSVTableSet(fh)
  table_set = any_tableset(fh)

  # A table set is a collection of tables:
  row_set = table_set.tables[0]

  # Setup headers and check if they match input file
  headers_ok = True
  offset, fileheaders = headers_guess(row_set.sample, tolerance=5)
  headers = ['username', 'firstname', 'lastname', 'email', 'password', 'groups']
  if len(headers) is not len(fileheaders):
    headers_ok = False
  else:
    for (i, h) in enumerate(headers):
      if str(h.strip().lower()) != str(fileheaders[i].strip().lower()):
        headers_ok = False
        break
  if not headers_ok:
    print("ERROR: Incorrect headers in input file")
    print("ERROR: expected headers are 'username', 'firstname', 'lastname', 'email', 'password', 'groups'")
    exit(1)

  row_set.register_processor(headers_processor(headers))

  # add one to begin with content, not the header:
  row_set.register_processor(offset_processor(offset + 1))

  # set column types:
  types = [StringType, StringType, StringType, StringType, StringType, StringType]
  row_set.register_processor(types_processor(types))

  # now run some operation on the data:
  groupset = set()
  users = {}
  for row in row_set:
    row_dict = {}
    # print
    for cell in row:
      # print cell,
      if cell.column == "groups":
        row_dict[str(cell.column).strip()] = [str(g).strip() for g in cell.value.split(',') if
                                              str(g).strip() is not '']
        groupset.update(row_dict[cell.column])
      else:
        row_dict[str(cell.column).strip()] = str(cell.value).strip()
    users[row_dict['username']] = row_dict
  # print

  current_users_qs = User.objects.all()
  duplicate_users = []
  new_users = []
  for u in users:
    if current_users_qs.filter(username__iexact=u).exists():
      duplicate_users.append(u)
    else:
      new_users.append(u)

  if not overwrite and len(duplicate_users) > 0:
    # Error duplicate users
    print("ERROR: Users already exist")
    print("Users: %s" % ", ".join(duplicate_users))
    exit(1)

  delete_users_qs = current_users_qs.exclude(username__in=new_users + duplicate_users)

  groups = {}
  for g in groupset:
    groups[g], newgroup = Group.objects.get_or_create(name=g)

  for u in list(users.values()):
    create_user(allgroups=groups, overwrite=overwrite, **u)

  if delete_extra:
    delete_users_qs.exclude(is_superuser=1).delete()


def parseargs():
  usage = """
  %prog [options]

  General Options:
  -h, --help              show this help message and exit

  List Users:
  --list                  list users
  --csv                   output in CSV format

  Single User Creation Options:
  --username=USERNAME     username
  --firstname=FIRST_NAME  firstname
  --lastname=LAST_NAME    lastname
  --email=EMAIL           email address
  --password=PASSWORD     password
  --group=GROUP           group (can be used multiple times)
  --overwrite             Overwrite existing user
  Note:  '--username' is required.  The other options are optional.

  Bulk User creation Options:
  --bulkfile=BULKFILE     CSV file for bulk creation
  --overwrite             Overwrite any existing user
  --delete-extra          Delete any existing users that are not in BULKFILE
  Note: CSV files should have the following column headers:
  'username', 'firstname', 'lastname', 'email', 'password', 'groups'

  LDAP Group User creation Options:
    Notes: LDAP Authentication must already be configured
           Use "" for LDAP_GROUP_DN to use AUTH_LDAP_REQUIRE_GROUP from settings
  --ldapgroup=LDAP_GROUP_DN     The DN of a LDAP group to add users from
  --ldap-dry-run                Print the users that would be added, but don't add them
  --add-attribute="ATTR:VALUE"  Add an attribute to the UserProfile

  User Profile Attribute Options
  --attruser=USERNAME            username for attribute operations
  --set-attribute="ATTR:VALUE"   set attribute
  --get-attribute="ATTR"         get attribute - leave blank to list all attributes
  --delete-attribute="ATTR"      remove attribute
  
  Delete User Options:
  --delete=USERNAME      username to delete

  """

  parser = optparse.OptionParser(usage=usage)

  parser.add_option('--list', dest='list', action="store_true",
                    help='list')
  parser.add_option('--csv', dest='csv', action="store_true",
                    help='csv', default=False)
  parser.add_option('--username', dest='username',
                    help='username')
  parser.add_option('--firstname', dest='firstname',
                    help='Firstname')
  parser.add_option('--lastname', dest='lastname',
                    help='Lastname')
  parser.add_option('--email', dest='email',
                    help='email address')
  parser.add_option('--password', dest='password',
                    help='password')
  parser.add_option('--group', dest='groups', action='append',
                    help='group')
  parser.add_option('--bulkfile', dest="bulkfile",
                    help='CSV file for bulk creation')
  parser.add_option('--overwrite', dest='overwrite', action="store_true",
                    help='overwite', default=False)
  parser.add_option('--delete-extra', dest='delete_extra', action="store_true",
                    help='delete-extra', default=False)
  parser.add_option('--ldapgroup', dest="ldapgroup",
                    help='LDAP group DN to create users from group')
  parser.add_option('--add-attribute', dest="userprofileattrs", action="append",
                    help='Add an attribute to the UserProfile in the form "attribute:value"')
  parser.add_option('--ldap-dry-run', dest='ldapdryrun', action="store_true",
                    help="Print the users that would be added, but don't add them", default=False)
  parser.add_option('--ldap-paginate', dest='ldappaginate', action="store_true",
                    help="Use pagination when attempting to search and create users from an LDAP group with over 1000 members.",
                    default=False)
  parser.add_option('--attruser', dest="attruser",
                    help='username for attribute operations')
  parser.add_option('--set-attribute', dest="setattr", action="append",
                    help='set attribute in the UserProfile in the form "attribute:value"')
  parser.add_option('--get-attribute', dest="getattr", action="append",
                    help="get attribute - leave blank to list all attributes")
  parser.add_option('--delete-attribute', dest="delattr", action="append",
                    help="remove attribute")
  parser.add_option('--delete', dest='delete',
                    help='delete')

  # for option in parser.option_list:
  #   if option.default != ("NO", "DEFAULT"):
  #     option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  actioncount = (options.list is not None) + \
                (options.username is not None) + \
                (options.bulkfile is not None) + \
                (options.delete is not None) + \
                (options.ldapgroup is not None) + \
                (options.attruser is not None)

  if actioncount > 1:
    parser.error(
      "options --list, --username, --bulkfile, --ldapgroup, --userattr and --delete are mutually exclusive. Choose one")

  if actioncount < 1:
    parser.error("--list, --username, --bulkfile, or --delete is required")

  if options.bulkfile:
    return (options, "bulk")
  if options.username:
    return (options, "single")
  if options.delete:
    return (options, "delete")
  if options.list:
    return (options, "list")
  if options.ldapgroup is not None:
    return (options, "ldap")
  if options.attruser:
    return (options, "attruser")


def main():
  options, action = parseargs()

  if not 'DJANGO_SETTINGS_MODULE' in os.environ:
    print("error: Please set the 'DJANGO_SETTINGS_MODULE' environment variable")
    exit(1)

  # If the settings file has a relative path for the db we need to find it
  # relative to the settings module not this file
  dsmname = os.environ['DJANGO_SETTINGS_MODULE']
  print("Using: DJANGO_SETTINGS_MODULE=%s" % dsmname)
  dsm = __import__(dsmname, fromlist=[''])
  if dsm.DATABASES['default']['ENGINE'].find("sqlite") >= 0:
    dbname = dsm.DATABASES['default']['NAME']
    if os.path.isabs(dbname):
      print("Using: db = %s" % dbname)
      from django.conf import settings
    else:
      print(dsm.__file__)
      newdbname = os.path.normpath(os.path.join(os.path.dirname(dsm.__file__),
                                                "..",
                                                dsm.DATABASES['default']['NAME']))
      from django.conf import settings
      settings.DATABASES['default']['NAME'] = newdbname
      from django.db import connection
      connection.close()
      print("Using: db = %s" % newdbname)

  import django
  django.setup()

  if action == "single":
    # create_user(username=options.username)
    # create_user(**dict(options))
    create_user(username=options.username,
                firstname=options.firstname,
                lastname=options.lastname,
                email=options.email,
                password=options.password,
                groups=options.groups,
                overwrite=options.overwrite)
  if action == "bulk":
    bulk_create(filename=options.bulkfile,
                overwrite=options.overwrite,
                delete_extra=options.delete_extra)
  if action == "delete":
    delete_user(options.delete)
  if action == "list":
    list_users(csv=options.csv)
  if action == "ldap":
    ldap_create(ldapgroup=options.ldapgroup,
                userprofileattrs=options.userprofileattrs,
                dryrun=options.ldapdryrun,
                paginate=options.ldappaginate)
  if action == "attruser":
    user_attr(options.attruser, options.getattr, options.setattr, options.delattr)


if __name__ == "__main__":
  main()