#!/usr/bin/env python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
  Utility to exercise the arcengine view recommendation path

  Importing of the django models stuff requires the settings path in the environment,
  something like:

    export DJANGO_SETTINGS_MODULE='arcwebbase.settings_local_dev'

"""

USAGE = """

  Put some usage text here

"""

import os
import sys
import json
import pprint
import getpass
import optparse
import traceback

from .appshow import Apps

from optparse import OptionValueError

THISDIR = os.path.dirname(os.path.realpath(__file__))
sys.path.append("%s/.." % THISDIR)
sys.path.append("%s/../vizlib" % THISDIR)

from django.db.models.base import ObjectDoesNotExist
from django.contrib.auth.models import User

from datasets.models import DataSet, DataConnection
from dataconn.dataconn import get_dataconn
from dataconn.connections_types import isArcEngineBased
from datasets.utils import get_default_dataconnection
from dataset import ModelDataset, getDatasetInfo
from vizlib.dataset_request import DatasetRequest

from reports.models import Report

from dataconn.thrift_hs2.TCLIService.ttypes import TGetAnalyticalViewsRecommendationsReq

# Turn off the logging in the django parts (like hs2conn.py)
import logging
logging.disable(logging.DEBUG)

def fatal(text):
  """ Fatal error """
  estr = traceback.format_exc()
  if not estr.startswith("None"):
    print(estr)
  if text:
    print(text)
  sys.exit(-1)

def parse():

  parser = optparse.OptionParser(usage="")

  parser.add_option('--appid', dest='appids', type='int', action='append',
    help='App ID')
  parser.add_option('--user', dest='username', default='admin',
    help='Perform data operations as this user')
  parser.add_option('--usage', dest='usage', action='store_true',
    help='Display usage text and exit')
  parser.add_option('--dryrun', dest='dryrun', action='store_true',
    help='Do operations but do not create the recommended views')
  parser.add_option('--verbose', dest='verbose', action='store_true',
    help='Set verbose output mode')
  parser.add_option('--skipviews', dest='skipviews', action='store_true',
    help='Do no ask for view recommendations from arcengine')

  for option in parser.option_list:
    if option.default != ("NO", "DEFAULT"):
      option.help += (" " if option.help else "") + "[default: %default]"

  options, args = parser.parse_args()

  if options.usage:
    print(USAGE)
    parser.print_help()
    sys.exit(0)

  if len(args) > 0:
    fatal('Extra options: %s' % str(args))

  user = None
  try:
    user = User.objects.get(username=options.username)
  except ObjectDoesNotExist as e:
    sys.exc_clear()
  if not user:
    fatal("Username '%s' not found" % options.username)
  options.userid = user.pk
  options.user = user

  # Load the listed apps, expand the dashboards
  if not len(options.appids):
    fatal("No App IDs specified")
  options.apps = Apps()
  for appid in options.appids:
    options.apps.add(appid)

  # All apps must be to the same arcengine dataconnection
  datasets = {}
  dataconn = None

  for app in options.apps.apps():
    dsid = app.dataset_id
    if not datasets.get(dsid):
      dl = DataSet.objects.filter(id=dsid)
      if len(dl) != 1:
        fatal("Unable to load dataset id %s for app %s" % (app.id, dsid))
      ds = dl[0]
      datasets[dsid] = ds

      dcid = ds.dataconnection_id
      if not dcid:
        dcid = get_default_dataconnection().id

      dcl = DataConnection.objects.filter(id=dcid);
      if len(dcl) == 0:
        fatal("Unable to load data connection id %s for app %s" % (app.id, dcid))

      dc = dcl[0];
      if dataconn is None:
        dataconn = dc
        if not isArcEngineBased(dc.dataconnection_type):
          fatal("App %s names invalid data connection type: %s" %
                (app.id, dc.dataconnection_type))

      elif dataconn.id != dc.id:
        fatal("App %s names dataconnection %s (%s) when we are already using %s (%s)",
              (app.id, dc.id, dc.dataconnection_name, dataconn.id, dataconn.dataconnection_name))

  options.datasets = datasets
  options.dataconn = dataconn

  return options

def main():

  options = parse()

  # Initialize all the datasets
  dc = get_dataconn(options.dataconn.id)
  dstab = {}
  for ds in list(options.datasets.values()):
    if not ds.dataset_info:
      try:
        ds.dataset_info = getDatasetInfo(ds.dataset_detail,
                                         options.dataconn.id, options.username)
        print("Loaded columns for table %s" % ds.dataset_detail)

      except Exception as e:
        fatal("Unable to fill in column details for table: %s" % ds.dataset_detail)


    dstab[ds.id] = ModelDataset(ds.dataset_info, dc.properties())

  # At this point, we have the data connection, all datasets loaded and modeled
  # and all apps loaded.

  # Probably want to organize apps by dataset but for now, just get all queries
  queries = []
  for app in options.apps.apps():

    dsreq = app.getDsreq()

    if options.verbose:
      print("App %d name '%s' type '%s'" % (app.id(), app.report_title(), app.report_type()))
      app.pprint()
      print("Dsreq: ")
      app.pprinter.pprint(dsreq);

    dataset_request = DatasetRequest(json.dumps(dsreq))
    build_request = dc.build(options.username, options.userid, None,
                               dataset_request=dataset_request)

    print("App %s\n%s" % (app.dbapp.id, build_request.query))
    queries.append("/* %s */ %s" % (app.dbapp.id, build_request.query))

  if not options.skipviews:
    print("Getting view recommendations...")
    vreq = TGetAnalyticalViewsRecommendationsReq(queries=queries)
    vres = dc.hs2call(options.username,
                      'GetAnalyticalViewsRecommendations', vreq, True)

    print("Did it:")
    for row in vres[2]:
      qtext= row[0]
      print(("Query:   %s" % qtext))
      if len(row) > 3:
        print((" Recommendation Info:\n%s" % pprint.pformat(dict(list(zip(vres[0], row))), indent=2)))
      else:
        objects = json.loads(row[1])
        recs = json.loads(row[2])
        print(("  Objects:\n%s" % pprint.pformat(objects, indent=2)))
        print(("  Recommendations:\n%s" % pprint.pformat(recs, indent=2)))

if __name__ == "__main__":
  main()

