# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
#
# This module privides the interface to the arcengine view recommendation feature
#

import re
import json
import time
import pprint
import logging
import sqlparse

from itertools import chain

from datetime import datetime, timedelta

from django.conf import settings
from django.http import HttpResponse
from django.db.models.base import ObjectDoesNotExist
from django.contrib.auth.models import User
from django.core.cache import cache

from reports.models import Report
from datasets.models import DataSet, DataConnection
from dataconn.dataconn import get_dataconn
from dataconn.connections_types import DATACONN_ARCENGINE
from .dataset import ModelDataset, getDatasetInfo
from vizlib.dataset_request import DatasetRequest
from rbac.permdefs import PermissionException

from vizlib.querylog import ComponentType, ActionType, QueryState, LogReader

from dataconn.thrift_hs2.TCLIService.ttypes import TGetAnalyticalViewsRecommendationsReq

from dataconn.data_api_cmds import SPECIAL_CMDS

LOGGER = logging.getLogger('arcweb')

"""
Takes a list of app IDs to accelerate
  Must be individual apps, not dashboards
  Requires ds_manage on each dataset, dc_aviews on each connection

  response format:

    viewdefs:
      list of recommended view structures
    appstat:
      list of app status structures

  status is one of:
    Dataset not connected to arcengine
    Insufficient privileges
    Unable to resolve parameters
    Already accelerated by views: v1, v2, ...
    Accelerated by views needing refresh: v1, v2, ...
    Unable to accelerate for reason: <from arcengine>

Hard parts:

  Parameter resolution & handling of complex parameters

  Showing usage of each view -- will need to enhance the logging
  to include a 'views used' column

"""

# Helper function to format a query but only below some reasonable length.
def sqlformat(query, **kwargs):
  if len(query) > getattr(settings, 'SQLPARSE_FORMAT_LIMIT', 10000):
    return query
  return sqlparse.format(query, **kwargs)

def viewrec_request(request):
  from reports.views import _get_report_data

  if request.method != 'POST':
    return HttpResponse('Invalid request', status=500)

  appids = request.POST.get('appids')
  if appids is None:
    return HttpResponse('Invalid request', status=500)
  appids = set(int(id) for id in json.loads(appids))
  reports = Report.objects.filter(id__in=appids)
  moreids = set([])

  rdtable = {}
  for r in reports:
    rd = _get_report_data(r, request.user,
                          include_data=True)
    if 'ds_appview' not in rd['perm']:
      continue

    if rd['report_type'] == 'dashboard':
      moreids.update(rd['visuals'])
      moreids.update(rd['pickers'])

    rdtable[r.id] = rd

  # Add in the visuals and pickers from any named dashboards
  if moreids.difference(appids):
    reports = Report.objects.filter(id__in=moreids.difference(appids))
    for r in reports:
      rd = _get_report_data(r, request.user,
                            include_data=True)
      if 'ds_appview' not in rd['perm']:
        continue

      rdtable[r.id] = rd
  reports = [rdtable[id] for id in sorted(rdtable.keys())]

  ci = AppInfo(request.user, rdtable.keys())
  # ci = AppInfo(request.user, json.loads(request.POST.get('appids', [])))
  ci.load_reports()
  apps_by_conn = ci.apps_by_connection()

  for dcid, apps in list(apps_by_conn.items()):
    ci.recommend_app_set(apps, dcid)

  rsp = {'appstat':ci.appstat, 'viewdefs':ci.viewdefs, 'old_style':ci.old_style,
         'reports': reports, 'dc_info':ci.dc_info}
  return HttpResponse(json.dumps(rsp), content_type="application/json")

# Get recommendations for a set of query statements
def viewrec_request_sql(request):

  if request.method != 'POST':
    return HttpResponse('Invalid request', status=500)

  qdict = json.loads(request.POST.get('qdict'))
  qdict = dict((int(id), query) for (id, query) in qdict.items())
  dcid = int(request.POST.get('dcid'))

  django_dc = DataConnection.objects.get(id=dcid)
  django_dc.checkPerm(request.user, 'dc_aviews')

  if not qdict or not dcid:
    return HttpResponse('Invalid request', status=500)

  ci = AppInfo(request.user)
  apps = ci.load_queries(qdict, dcid)
  ci.recommend_app_set(apps, dcid)
  rsp = {'appstat':ci.appstat, 'viewdefs':ci.viewdefs, 'old_style':ci.old_style,
      'dc_info':ci.dc_info}
  return HttpResponse(json.dumps(rsp), content_type="application/json")

# Store a list of queries for acceleration
def viewrec_request_queries(request):
  if request.method != 'POST':
    return HttpResponse('Invalid request', status=500)

  session_key = request.session.session_key
  query_data = {}
  query_data['queries'] = request.POST.get('queries')
  query_data['dcId'] = request.POST.get('dcId')
  query_data['fromDirect'] = request.POST.get('fromDirect')
  state_data = request.POST.get('state')
  timestamp = time.time()
  query_key = 'accel-queries-%s-%s' % (session_key, timestamp)
  state_key = 'table-state-%s-%s' % (session_key, timestamp)
  cache.set(query_key, query_data)
  cache.set(state_key, state_data)
  return HttpResponse(timestamp, status=200)

# For a list of apps check for current view usage by generating query for
# each app and using 'explain' to see if maps to a view.
def viewrec_checkapps(user, idlist):

  ci = AppInfo(user, idlist)
  ci.load_reports()
  apps_by_conn = ci.apps_by_connection()
  for dcid, apps in list(apps_by_conn.items()):
    ci.recommend_app_set(apps, dcid)

  rsp = {'appstat':ci.appstat, 'viewdefs':ci.viewdefs}
  return rsp

def viewrec_accel_info(idlist, days):
  """
    Walk through the log to gather execution data for the set of input apps
    For each type of execution, want count and total time
      Cached
      Cache check
      No rewrite
      Rewrite
    Should get up to last N executions
    Do we stop when app was last updated (yes unless query has not changed)
  """

  data = {}
  for id in idlist:
    data[id] = {
      'cached_count':0,
      'cached_seconds':0,
      'rewrite_count':0,
      'rewrite_seconds':0,
      'normal_count':0,
      'normal_seconds':0,
    }

  stop = datetime.fromtimestamp(time.time()) - timedelta(days)
  lr = LogReader()
  lines = 0
  for r in lr.read_json_backwards():
    ts = datetime.strptime(r.get('ts'), '%Y-%m-%d %H:%M:%S.%f')
    if ts < stop:
      break
    if r.get('component_type') == ComponentType.APP and \
       r.get('query_state') == QueryState.COMPLETE and \
       r.get('action') == ActionType.READ and \
       r.get('component_id') and \
       int(r.get('component_id')) in idlist:

      id = int(r.get('component_id'))
      rec = data[id]

      if r.get('cache'):
        key = 'cached'
      elif r.get('rewrite'):
        key = 'rewrite'
      else:
        key = 'normal'
      cnt = key + '_count'
      sec = key + '_seconds'
      rec[cnt] = rec[cnt] + 1
      rec[sec] = rec[sec] + float(r.get('runtime', 0.0))

      if key == 'rewrite' or key == 'normal':
        if rec.get('last_exec_mode') is None:
          rec['last_exec_mode'] = key
          rec['last_exec_ts'] = r.get('ts')
        last = 'last_' + key
        if rec.get(last) is None:
          rec[last + '_ts'] = r.get('ts')
          rec[last + '_runtime'] = r.get('runtime')

  return data
"""
          ('ts', self.ts.strftime('%Y-%m-%d %H:%M:%S.%f')),
          ('action', self.action),
          ('component_type', self.component_type),
          ('component_id', self.component_id),
          ('action_id', self.action_id),
          ('query_state', self.query_state),
          ('username', self.username),
          ('ip', self.client_ip),
          ('visual', self.visual_id),
          ('app', self.app_id),
          ('dataset', self.dataset_id),
          ('dataconnection', self.dataconnection_id),
          ('dataconnection_type', self.dataconnection_type),
          ('sync', self.sync),
          ('cache', self.served_from_cache),
          ('cache_check', self.served_from_cache_after_data_check),
          ('runtime', self.runtime),
          ('query', self.query)
  return {'id':id, 'status':'ok'}
"""

# Helper function to extract the view definition from the returned data
def get_vdef(text):
  try:
    m = re.match(r'.*(\s+STORED\s+AS\s+PARQUET\s+AS\s+)(.*)', text, re.DOTALL + re.IGNORECASE)
  except Exception as e:
    m = None
  if not m:
    if text is not None:
      LOGGER.error('Could not extract aview defintion from: %s' % text)

    return ''

  return m.groups()[1]

def get_existing_tables(datarequest_params, table_dict, user, dataconn, aviews=False, return_raw=False):
  # To retrieve and parse a list of tables
  # If aviews is true, the request is SHOW ANALYTICAL VIEWS
  # otherwise just SHOW TABLES
  dataset_request = DatasetRequest(json.dumps(datarequest_params), None)

  build_request = dataconn.build(user.username, user.id,
                                 dataset_request=dataset_request)
  rsp = dataconn.execute(build_request)

  hdrs = rsp.get('colnames')
  rows_parsed = json.loads(rsp.get('rows', '[]'))
  list_hdrs = ['aggregates', 'dimensions']
  for row in rows_parsed:
    if aviews:
      if len(hdrs) == 3:
        # starting arcengine v2.12.1 row is a list of jsonified aview information
        parsed_row = dict([(hdrs[i], json.loads(row[i])) for i in range(len(hdrs))])
        t = parsed_row.get('Metadata')
        if t.get('error_message'):
          t['state'] = 'INVALID'
          t['dimensions'] = []
          t['aggregates'] = []
          t['base_table_name'] = 'Error.not_available'
        select_substring = get_vdef(t.get('definition'))
        t['select'] = sqlformat(select_substring, reindent=True, keyword_case='upper', wrap_after=5)
        t['basetables'] = parsed_row.get('Objects') or []
        table_dict[parsed_row.get('Aview')] = t
      else:
        # row is a list of aview information described by hdrs
        t = dict([(hdrs[i], row[i]) for i in range(len(hdrs))])
        for hdr in list_hdrs:
          if t.get(hdr) == 'N/A':
            t[hdr] = []
          else:
            t[hdr] = json.loads(t.get(hdr)) or []
        table_dict[t.get('name')] = t
    else:
      # row is a one element list with the table name
      table_dict[row[0]] = True

  if return_raw:
    return rsp

# --------------- Control structure and functions for apps -> queries operations -------------
class AppInfo(object):
  def __init__(self, user, idlist=None):
    self.user = user
    if idlist is not None:
      self.appids = set(int (id) for id in idlist)
    self.appstat = {}
    self.viewdefs = []
    self.dataconns = {}
    self.datasets = {}
    self.reports = {}
    self.old_style = None
    self.dc_info = {}

  # Store a status record for an app
  def set_status(self, id, status, viewname='', reason='', query=''):
    view_stat_dict = {'id': id, 'status': status, 'viewname': viewname,
                      'reason': reason, 'query': query}
    if id in self.appstat:
      self.appstat[id].get('stats').append(view_stat_dict)
    else:
      self.appstat[id] = {'query': query, 'stats': [view_stat_dict]}

  def set_appstat(self, id, stat_dict_list, query):
    self.appstat[id] = {
      'query': query,
      'stats': stat_dict_list
    }

  def set_failed(self, id, reason, query=''):
    self.set_status(id, 'FAILED', reason=reason, query=query)

  def get_status(self, id):
    return self.appstat.get(id)

  # Check that the dataset for an app/query is valid
  def checkds(self, id, dsid):
    ds = self.datasets.get(dsid)
    if ds is None:
      try:
        ds = DataSet.objects.get(id=dsid)
      except DataSet.DoesNotExist:
        self.set_failed(id, 'Dataset not found')
        return None
      try:
        ds.checkPerm(self.user, 'ds_manage')
      except PermissionException:
        self.set_failed(id, 'Insufficient dataset permissions')
        return None
      self.datasets[ds.id] = ds
    return ds

  # Check that the data connection for the app/query id is valid
  def checkdc(self, id, dcid):
    dc = self.dataconns.get(dcid)
    if dc is None:
      try:
        dc = DataConnection.objects.get(id=dcid)
      except DataConnection.DoesNotExist:
        self.set_failed(id, 'Data connection not found')
        return None
      if dc.dataconnection_type != DATACONN_ARCENGINE:
        self.set_failed(id, 'Data connection must be of type Arcadia (legacy)')
        return None
      try:
        dc.checkPerm(self.user, 'dc_aviews')
      except PermissionException:
        self.set_failed(id, 'Insufficient data connection permissions')
        return None
      self.dataconns[dc.id] = dc
    return dc

  # Get some data connection information to be returned to the UI side
  def add_dcinfo(self, dc, dataconn):
    if self.dc_info.get(dc.id) is not None:
      return
    info = {}
    locs = json.loads(dc.dataconnection_info or '{}').get('PARAMS', {}).get('LOCATIONS', {})
    info['LOCATIONS'] = locs

    dataset_request = DatasetRequest(json.dumps({'type': SPECIAL_CMDS['SHOW_DATABASES']}), None)
    build_request = dataconn.build(self.user.username, self.user.id, dataset_request=dataset_request)
    rsp = dataconn.execute(build_request)

    rows_parsed = json.loads(rsp.get('rows', '[]'))
    info['DATABASES'] = [r[0] for r in rows_parsed]
    self.dc_info[dc.id] = info

  # Load the reports and validate permissions + connection type
  def load_reports(self):
    replist = Report.objects.filter(id__in=self.appids)
    self.reports = dict([(r.id, r) for r in replist])
    for id in self.appids.difference(list(self.reports.keys())):
      self.set_failed(id, 'App not found')

    for r in list(self.reports.values()):
      if r.report_type == 'dashboard':
        self.set_failed(r.id, 'Must accelerate individual visuals')
        continue
      if r.report_type in ['flow', 'funnel']:
        self.set_failed(r.id, 'Visual type %s not currently supported' % r.report_type)
        continue

      ds = self.checkds(r.id, r.dataset_id)
      if ds is not None:
        self.checkdc(r.id, ds.dataconnection_id)

    # Remove app IDs that have produced an error
    self.input_appids = self.appids
    self.appids = self.input_appids.difference(list(self.appstat.keys()))

  # Initialize for raw query acceleration
  def load_queries(self, qdict, dcid):
    self.appids = set(qdict.keys())
    apps = Apps()
    for id, query in qdict.items():
      apps.addQuery(id, query)
      self.checkdc(id, dcid)

    return apps

  # Process apps by data connection to build queries for each app
  def apps_by_connection(self):
    apps_by_dc = {}
    for dc in list(self.dataconns.values()):
      apps = Apps()
      apps_by_dc[dc.id] = apps
      for r in list(self.reports.values()):
        if r.id in self.appids and r.dataset.dataconnection_id == dc.id:
          apps.add(r.id, r)

      # Walk the apps to get the dataset scope parameters
      ds_scope = {}
      for app in apps.apps():
        ds_colname = app.getDatasetScopeColumn()
        if ds_colname:
          dsid = app.dataset_id
          if ds_scope.get(dsid) is None:
            ds_scope[dsid] = []
          if ds_colname not in ds_scope[dsid]:
            ds_scope[dsid].append(ds_colname)

      # Now get the dsreq's for each app
      for app in apps.apps():

        if app.isParameterApp():
          self.set_failed(app.id, 'No acceleration for parameter components')
          continue

        try:
          app.dsreq = app.getDsreq(expand_params=False, filter_dims=True)
        except Exception as e:
          LOGGER.exception('Error building query for %s' % app.id)
          self.set_failed(app.id, 'Error building query: %s' % str(e))
          continue

        # For now we can just add any dataset scope columns into the dimensions list

        # TODO: To fully support dataset scope parameters in all modes (not just
        # for recommendation path) need to get the column type from the dataset
        # and process as a filter element (i.e. generate a filter expression based
        # on the parameters and column type -- see arc_reports_base.js/datasetFiltersCb()

        if not app.reportIsFilter() and not app.user_settings.get('explicit_scope') == 'true':
          for col in ds_scope.get(app.dataset_id, []):
            expr = '[' + col + '] as `' + col + '`'
            if not any(d.get('expr') == expr for d in app.dsreq['dimensions']):
              app.dsreq['dimensions'].append({'expr':expr, 'type':'SIMPLE', 'alias':col})

      # Parameter resolution
      """
        Need to process parameter scope -- dataset level, app level

          Kind of hard--actual query for the visual depends on the context
          in which it runs (i.e. which app or dataset level filters it
          will pick up).

        if an app has 'click behavior',
          add the (expression alias : expression text) parameter for the target app

        if an app is a picker type
          add the (expression name, expression) to the global parameters
          (Need to process the picklist here
           Also the 'emit multiple' setting--easier actually)

      """

      # Here if a parameter is not substituted, this is an error for the app

      dataconn = get_dataconn(dc.id)
      for app in apps.apps():
        app.query = None
        if self.appstat.get(app.id):
          continue

        try:
          dsreq = app.dsreq
          for elem in dsreq['dimensions']:
            if elem['type'] == 'SIMPLE':
              elem['expr'] = app.substitute(elem['expr'])
          for elem in dsreq['aggregates']:
            elem['expr'] = app.substitute(elem['expr'])
          dsreq['filters'] = [app.substitute(f) for f in dsreq['filters']]
          dsreq['having'] = [app.substitute(h) for h in dsreq['having']]
        except ParameterNotFound as e:
          app.status = str(e)
          self.set_failed(app.id, app.status)
          continue

        try:
          dataset_request = DatasetRequest(json.dumps(app.dsreq))
          build_request = dataconn.build(self.user.username, self.user.id, None,
                                         dataset_request=dataset_request,
                                         comment='/* %s */' % app.id)
        except Exception as e:
          LOGGER.exception('Error building query')
          app.status = "Error building query: %s" % str(e)
          self.set_failed(app.id, app.status)
          continue

        app.query = build_request.comment + ' ' + build_request.query

    # Return apps grouped by data connection
    return apps_by_dc

  # Get recommendations for one set of apps for a single connection
  def recommend_app_set(self, apps, dcid):
    def mysplit(s):
      if s and isinstance(s, str) and s.lower() != 'null':
        s = json.loads(s)
        # in the old style sometimes the result comes as tuple
      if isinstance(s, tuple) and \
              isinstance(s[0], list):
        s = s[0]

      return s or []

    def get_aview_dict(name, vinfo, dataset_id):
      from reports.views import columns_internal
      raw_dimensions = vinfo.get('dimensions')
      if isinstance(raw_dimensions, list) and (len(raw_dimensions) == 0 or isinstance(raw_dimensions[0], dict)):
        dimensions = [x.get('name') for x in raw_dimensions]
        dimension_cardinalities = [x.get('cardinality') for x in raw_dimensions]
      else:
        dimensions = mysplit(raw_dimensions)
        dimension_cardinalities = mysplit(vinfo.get('dimension_cardinalities'))

      definition = vinfo.get('definition')
      if vinfo.get('select'):
        select = sqlformat(vinfo.get('select'), reindent=True, keyword_case='upper', wrap_after=5)
      elif definition:
        select_substring = get_vdef(definition)
        select = sqlformat(select_substring, reindent=True, keyword_case='upper', wrap_after=5)
      else:
        select = None

      full_sql = vinfo.get('full_sql', 'N/A')
      if full_sql == 'N/A' and definition:
        full_sql = definition

      basetables = vinfo.get('basetables', [])
      for basetable in basetables:
        cols = columns_internal(self.user.username, basetable.get('name'), dataconnection_id=dc.id)
        basetable['partitions'] = cols.get('partitions', [])
      base_tablename = vinfo.get('base_table_name')
      base_lvname = vinfo.get('base_lview_name')
      base = base_tablename or base_lvname

      return {
        'dc_id': dc.id,
        'dc_name': dc.dataconnection_name,
        'dc_type': dc.dataconnection_type,
        'ds_id': dataset_id,
        'ds_name': 'N/A' if dataset_id == -1 else self.datasets[dataset_id].dataset_name,
        'select': select,
        'full_sql': full_sql,
        'base': base,
        'state': vinfo.get('state'),
        'name': name,
        'view_num': len(self.viewdefs) + 1,
        'aggregates': mysplit(vinfo.get('aggregates')),
        'expensive_aggregates': mysplit(vinfo.get('expensive_aggregates')),
        'partitions': mysplit(vinfo.get('partitioning_columns')),
        'dimension_cardinalities': dimension_cardinalities,
        'dimensions': dimensions,
        'base_lview_name': base_lvname,
        'base_table_name': base_tablename,
        'file_format': vinfo.get('file_format'),
        'size_on_disk': vinfo.get('size_on_disk'),
        'in_mem_size': vinfo.get('in_mem_size'),
        'row_size': vinfo.get('row_size'),
        'row_count': vinfo.get('row_count'),
        'definition': definition,
        'incrementally_refreshable': vinfo.get('incrementally_refreshable'),
        'basetables': basetables
      }

    def make_aview_stat_dict(id, status, viewname='', reason='', query=''):
      return {'id': id, 'status': status, 'viewname': viewname,
                      'reason': reason, 'query': query}

    def update_dict(initial_dict, revised_dict):
      #Update values that are None
      filtered_initial = [(k, v) for (k, v) in list(initial_dict.items()) if v is not None]
      filtered_update = [(k, v) for (k, v) in list(revised_dict.items()) if v is not None]
      return dict(chain(filtered_initial, filtered_update))

    def get_aview_list(r):
      app_id_regex = r'/\*\s*(\d+)\s*\*/'
      if self.old_style:
        view_dict = dict([(hdrs[i].lower(), r[i]) for i in range(min(len(hdrs), len(r)))])
        view_dict['partitioning_columns'] = [] if view_dict['partitioning columns'] == 'NULL' else mysplit(view_dict['partitioning columns'])
        view_dict['dimensions'] = [] if view_dict['dimensions'] == 'NULL' else mysplit(view_dict['dimensions'])
        view_dict['aggregates'] = [] if view_dict['aggregates'] == 'NULL' else mysplit(view_dict['aggregates'])
        view_dict['reason'] = view_dict.get('reason for no recommendation')

        app_id = None
        match = re.search(app_id_regex, view_dict.get('query'))
        if match:
          app_id = int(match.group(1))
        return [view_dict], app_id
      else:
        res = []
        match = re.search(app_id_regex, r[0])
        if match:
          app_id = int(match.group(1))
        else:
          raise Exception('cannot extract app Id from query')

        recs = json.loads(r[2])
        for rec in recs:
          query = sqlformat(rec['query']['text'], reindent=True, keyword_case='upper', wrap_after=5)
          if 'aview' in rec:
            d = rec['aview']
            d.update({'status': rec['query']['recommendation_status'],
                      'query': query})
            res.append(d)
          else:
            res.append({'status': rec['query']['recommendation_status'],
                        'query': query,
                        'reason': rec['query']['reason_for_no_recommendation']
                        })
        return res, app_id

    app_table = dict((app.id, app) for app in apps.apps() if app.query)
    view_table = {}

    dc = self.dataconns[dcid]
    dataconn = get_dataconn(dc.id)

    # Get the existing analytical views in the system
    existing_views = {}
    get_existing_tables({'type': SPECIAL_CMDS['SHOW_AVIEWS']}, existing_views, self.user, dataconn, True)
    self.add_dcinfo(dc, dataconn)

    for (id, app) in list(app_table.items()):
      LOGGER.debug("Rec UI app %s : %s" % (id, app.query))

    try:
      vreq = TGetAnalyticalViewsRecommendationsReq(queries=[app.query for app in list(app_table.values())])
      vres = dataconn.hs2call(self.user.username,
                              'GetAnalyticalViewsRecommendations', vreq, True)
    except Exception as e:
      msg = str(e)
      if msg == "Connection reset: Invalid method name: 'GetAnalyticalViewsRecommendations'":
        msg = 'This Arcengine version does not support recommendation requests'
      for id in [app.id for app in apps.apps()]:
        self.set_failed(id, msg)
      return

    hdrs = vres[0]
    rows_parsed = json.loads(vres[2])
    if len(hdrs) > 3:
      self.old_style = True
    else:
      self.old_style = False

    existing_tables = {}

    for row in rows_parsed:
      qtext = row[0]
      if self.old_style:
        LOGGER.debug("Rec UI response:\nQuery:\n%s\nRecommendation Info:\n%s" % \
                     (qtext, pprint.pformat(dict(list(zip(hdrs, row))), indent=2)))
      else:
        objects = json.loads(row[1])
        recs = json.loads(row[2])
        LOGGER.debug("Rec UI response:\nQuery:\n%s\nObjects:\n%s\nRecommendations:\n%s" % \
                   (qtext, pprint.pformat(objects, indent=2), pprint.pformat(recs, indent=2)))

    # First grab the view definitions
    for i in range(len(rows_parsed)):
      aview_recs, app_id = get_aview_list(rows_parsed[i])
      # Note that we assign the view to dataset for the first app for this view
      app = app_table[app_id]
      for aview in aview_recs:
        basetables = [] if self.old_style else mysplit(rows_parsed[i][1])
        if aview.get('status') == 'NO_RECOMMENDATION':
          continue
        elif aview.get('status') in ['ROUTED', 'AWAITING_REFRESH']:
          # Make sure we include view definitions for existing views
          #  # need to be fixed when Rachit make the changes to get the proper name
          vname = aview.get('name') or aview.get('reason')
          if vname not in view_table and vname in existing_views:
            view_table[vname] = get_aview_dict(vname,
                                               existing_views[vname],
                                               app.dataset_id)
            aview['basetables'] = basetables
            view_table[vname] = update_dict(view_table[vname], get_aview_dict(vname, aview, app.dataset_id))
            self.viewdefs.append(view_table[vname])
          continue

        vfull = aview.get('definition')
        vdef = get_vdef(vfull)
        if vdef not in view_table:

          m = re.match(r'CREATE ANALYTICAL VIEW (\S+)\.(\S+) .* FROM (\S+)( |$)', vfull)
          table = m.groups()[2]
          dbname = m.groups()[0].replace('`', '')
          avorig = m.groups()[1]

          # Find an available view name
          vnum = 0
          if app.dataset_id == -1:
            ds_text = re.sub(r'`', '', table).replace('.', '_')
          else:
            ds_text = 'ds_%s' % app.dataset_id
          while True:
            if existing_tables.get(dbname) is None:
              # Get all existing tables on that database to avoid name collisions
              existing_tables[dbname] = {}
              get_existing_tables({'type': SPECIAL_CMDS['SHOW_TABLES'], 'dbname': dbname}, existing_tables[dbname], self.user, dataconn)
            vnum = vnum + 1
            avname = 'av_%s_%s' % (vnum, ds_text)
            fullname = '%s.%s' % (dbname, avname)
            if existing_tables.get(dbname).get(avname) is None and \
                            fullname not in [vd['name'] for vd in self.viewdefs]:
              break

          aview.update({
            'select': vdef,
            'base': re.sub(r'`', '', table),
            'full_sql': vfull.split('`%s`.%s' % (dbname, avorig))[1],
            'state': 'RECOMMENDED',
            'basetables': basetables
          })
          view_table[vdef] = get_aview_dict(
            fullname, aview, app.dataset_id)
          self.viewdefs.append(view_table[vdef])

    filtered_reasons = [
      'Recommendations are only available for legs of a Union Statement.',
      'Currently recommendation engine does not provide recommendations for complex queries.',
      'The query must have exactly one table reference in it to get a recommendation.',
      'This query contains joins. Please consider defining appropriate foreign key - primary key relationships in order to accelerate the query.'
    ]

    # Now process each row
    for r in rows_parsed:
      aview_recs, app_id = get_aview_list(r)
      qtext = re.sub(r'/\* \d+ \*/ ', '', r[0])[1:-1]
      qtext = re.sub(r'\\n', ' ', qtext)
      qtext = sqlformat(qtext, reindent=True, keyword_case='upper', wrap_after=5)
      responses = []
      filtered_responses = []
      for aview in aview_recs:
        query = aview.get('query')
        status = aview.get('status')
        reason = aview.get('reason')
        if status in ['ROUTED', 'AWAITING_REFRESH']:
          viewname = aview.get('name') or reason
          responses.append(make_aview_stat_dict(app_id, status, viewname=viewname, query=query))
        elif status == 'NO_RECOMMENDATION':
          curr_aview = make_aview_stat_dict(app_id, status, reason=reason, query=query)
          if reason in filtered_reasons:
            filtered_responses.append(curr_aview)
          else:
            responses.append(curr_aview)
        elif status == 'RECOMMENDATION':
          vdef = get_vdef(aview.get('definition'))
          responses.append(make_aview_stat_dict(app_id, status, viewname=view_table[vdef]['name'], query=query))
        else:
          responses.append(make_aview_stat_dict(app_id, 'FAILED', query=query,
                          reason='Unknown recommendation status: %s' % status))

      if responses:
        self.set_appstat(app_id, responses, qtext)
      else:
        self.set_appstat(app_id, filtered_responses, qtext)

    # Ensure we have a status for each app requested
    for id in [app.id for app in apps.apps()]:
      if self.get_status(id) is None:
        self.set_failed(id, 'No recommendation response received')

  # Check each app in the set to see if it currently uses a view
  def check_app_set(self, apps, dcid):

    dc = self.dataconns[dcid]
    dataconn = get_dataconn(dc.id)

    for app in apps.apps():
      if not app.query:
        continue

      try:
        rsp = dataconn.sync_runquery('explain ' + app.query, username=self.user.username)
        # TODO: Would like to get view names here, for now we just determine if a
        #       rewrite happened.
        rewrite = False
        rows_parsed = json.loads(rsp.get('rows', '[]'))
        for r in rows_parsed:
          if r[0].startswith('REWR :'):
            rewrite = True
            break
        self.set_status(app.id, 'ROUTED' if rewrite else 'NOT_ROUTED')

      except Exception as e:
        self.set_failed(app.id, 'Error checking for rewrite: %s' % str(e))

# ---------- Support for constructing dsreq from viz ID and params --------------
def check_dsreq(request, req_params, d1, viz_params, conn, django_ds, dfunc):

  d1 = d1 or {}

  report_id = int(req_params.get('report_id', 0))
  dashboard_id = int(req_params.get('parent_report_id', 0))
  dataset_id = django_ds.id

  def pf(s):
    return pprint.pformat(s, indent=2, width=200)

  LOGGER.info('Checking viz %s dashboard %s dataset %s' % (report_id, dashboard_id, dataset_id))

  # Build the datarequest from the report
  report = Report.objects.get(id=report_id)
  kv = dfunc((p['key'], p['value']) for p in viz_params if p.get('value'))
  app = App(report, kv)
  app.phase = req_params.get('phase')
  d2 = app.getDsreq()

  # Now add in any dataset scope filters
  ds = ModelDataset(django_ds.dataset_info, conn.properties())
  pdict = dict((p.get('key'), p) for p in viz_params)
  if report.report_type != 'daterange' and not app.user_settings.get('explicit_scope') == 'true':

    # picklist_filter_columns handling
    # TODO -- I think these can be expressions
    pfc = None
    if report.report_type in ['picklist', 'range']:
      pfc = app.user_settings.get('picklist_filter_columns', '').split(',')
      pfc = [c.strip(' []') for c in pfc if c]

    for p in viz_params:
      scope = p.get('scope', {})
      if scope.get('scopeType') == 'global' or \
         (scope.get('scopeType') == 'dataset' and scope.get('datasetId') == dataset_id):
        key = p.get('key')
        if any(key.endswith('.' + s) for s in ['exclude', 'end', 'include_nulls',
                                               'alias', 'config', 'ignore']):
          continue

        if pfc is not None:
          prefix = key.split('.')[0]
          if prefix not in pfc:
            continue
          else:
            pfc.remove(prefix)

        if key.endswith('.start'):
          key = key[:-len('.start')]
          start = p.get('value')
          end = pdict.get(key + '.end').get('value')
          if p.get('valuetype') == 'NUMERIC':
            if (start and end):
              expr = '([%s] >= %s AND [%s] <= %s)' % (key, start, key, end)
            elif end:
              expr = '([%s] <= %s)' % (key, end)
            else:
              expr = '([%s] >= %s)' % (key, start)
          else:
            expr = "[%s] between '%s' and '%s'" % (key, start, end)
          if pdict.get(key + '.include_nulls') and pdict.get(key + '.include_nulls').get('value') == 'true':
            expr = '((%s) or ([%s] is NULL))' % (expr, key)

        elif key.endswith('.data'):
          key = key[:-len('.data')]
          inout = pdict.get(key + '.exclude', {}).get('value', 'in')
          expr = filterExpr(key, p.get('valuetype', 'STRING'), p.get('value_list', []), inout)

        else:
          enquote = "'%s'" if p.get('valuetype') == 'STRING' else '%s'
          if p.get('value_list'):
            vtext = ', '.join(enquote % v for v in p.get('value_list'))
            expr = "[%s] %s (%s)" % (key, 'in', vtext)
          else:
            expr = '[%s] = %s' % (key, enquote % p.get('value'))

        if ds.getColumn(key):
          d2['filters'].append(expr)

    if pfc is not None:
      for c in pfc:
        d2['filters'].append('(1 = 1)')

  # Seems like actual UI code doesn't use alias
  for d in d2.get('dimensions'):
    d.pop('alias', None)
  for a in d2.get('aggregates'):
    a.pop('alias', None)

  # Remove any empty keys
  for key in list(d2.keys()):
    if not d2[key]:
      d2.pop(key)

  if 'limit' in d1:
    d1['limit'] = str(d1['limit'])

  if 'limit' in d2:
    d2['limit'] = str(d2['limit'])

  for k in ['having', 'filters']:
    for d in [d1, d2]:
      if k in d:
        d[k].sort()

  errtxt = []
  if d1 and d1 != d2:
    errtxt.append('dsreq mismatch for viz %s' % report_id)
    errtxt.append('d1 = \n%s' % pf(d1))
    errtxt.append('d2 = \n%s' % pf(d2))
    errtxt.append('viz_params = \n%s' % pf(viz_params))
    errtxt.append('kv = \n%s' % pf(kv))

    for k, v1 in d1.items():
      if k not in d2:
        errtxt.append('Key %s not in d2' % k)
        continue
      v2 = d2[k]
      if v1 == v2:
        continue
      elif type(v1) != type(v2):
        errtxt.append('%s entries of different types (%s != %s)' % (k, type(v1), type(v2)))
      elif type(v1) == type([]) and len(v1) != len(v2):
        errtxt.append('List %s different lengths (%s != %s)' % (k, len(v1), len(v2)))
      elif k in ['filters', 'having']:
        for i in range(len(v1)):
          if v1[i] != v2[i]:
            errtxt.append('%s expr %s mismatch (%s != %s)' % (k, i, v1[i], v2[i]))
      elif k in ['dimensions', 'aggregates']:
        for i in range(len(v1)):
          if v1[i] != v2[i]:
            errtxt.append('%s entry %s mismatch:\n%s\n%s' % (k, i, pf(v1[i]), pf(v2[i])))
      else:
        errtxt.append('%s mismatch:\nv1:\n%s\nv2:\n%s' % (k, pf(v1), pf(v2)))

    k2 = set(d2.keys()) - set(d1.keys())
    if k2:
      errtxt.append('Keys %s not in d1' % k2)

    LOGGER.warning('\n'.join(errtxt))

  return d2

# ----------------- Support for getting queries from django reports (apps) objects ----------------

# Parameter not found exception
class ParameterNotFound(Exception):
  def __init__(self, p):
    self.parm = p

  def __str__(self):
    return "Unable to substitute parameter: %s" % self.parm

# Process a string and do replacements for <<parameter:default>> (see arc_params.js)
def parameterReplace(s, params, raise_on_error=True):
  params = params or {}
  m = re.findall('<<.+?>>', s)
  for p in m:
    sp = s.split(p, 1)
    kv = p[2:-2].partition(':')
    middle = params.get(kv[0], p if kv[1] != ':' else kv[2])
    if raise_on_error and middle == p:
      raise ParameterNotFound(p)
    s = sp[0] + middle + sp[1]
  return s

class Apps(object):
  """ Class to hold list of apps and expand dashboards """
  DB_RE = re.compile("^uri-(\d+)-widget-(\d+)$")

  def __init__(self, params=None):
    self.app_dict = {}
    self.processed = set()
    self.params = params or {}

  def add(self, appid, dbapp=None):
    if appid in self.processed:
      return
    self.processed.add(appid)

    if dbapp is None:
      rl = Report.objects.filter(id = appid)
      if len(rl) != 1:
        raise Exception("Unable to load App ID %s" % appid)
      dbapp = rl[0]

    if dbapp.report_type != 'dashboard':
      self.app_dict[dbapp.id] = App(dbapp, self.params)
    else:
      d = json.loads(dbapp.report_data)
      for w in dbapp.all_report_ids():
        self.add(w)

  def apps(self):
    return list(self.app_dict.values())
  # Add a raw query app

  def addQuery(self, id, query):
    self.app_dict[id] = QueryApp(id, query)

# This is a simple app-like class for doing raw query acceleration
class QueryApp(object):

  def __init__(self, id, query):
    self.status = None
    self.id = id
    self.query = ('/* %s */ ' % id) + query
    self.dataset_id = -1

# See if expression already contains an alias
def splitExprAndAlias(expr):
  if expr:
    m = re.match(r"""^(.*)(\s+as\s+)([a-zA-Z0-9_<>]+)+$""", expr, re.I) or \
        re.match(r"""^(.*)\s+as\s+(['"`]).+?\2$""", expr, re.I) or \
        re.match(r"""^(.*)\s+as\s+(<<.+?>>)$""", expr, re.I)
    if m:
      return (m.groups()[0], m.groups()[1])
  return (expr, None)

# Build a filter expression for a col name and list of values
def filterExpr(fname, ftype, fdata, inout):

  # TODO: Is the the best place to handle empty filter data?
  text = '(1 = 1)'

  values = ['%s' % d for d in fdata if d not in ['NULL', "'NULL'"]]
  if values:
    if ftype != 'STRING':
      vtext = ', '.join(values)
    else:
      vtext = ', '.join("'%s'" % re.sub(r"""([^\\])(['])""", r'\1\\\2', v) for v in values)
    text = "[%s] %s (%s)" % (fname, inout, vtext)

  # NULL handling like arc_misc.js/arc.misc.whereClauseFromValues()
  if 'NULL' in fdata or "'NULL'" in fdata:
    if inout == 'not in':
      nulljoin = ' AND '
      nulltext = '[%s] is not NULL' % fname
    else:
      nulljoin = ' OR '
      nulltext = '[%s] is NULL' % fname
    if values:
      text = '(%s%s%s)' % (text, nulljoin, nulltext)
    else:
      text = nulltext
  return text

class App(object):

  def __init__(self, dbapp, params=None):
    self.dbapp = dbapp
    self.id = dbapp.id
    self.status = None
    self.phase = None
    self.appdata = json.loads(dbapp.report_data).get('report_data')
    self.user_settings = self.appdata.get('user_settings', {})
    self.dataset_id = dbapp.dataset_id

    self.pprinter = pprint.PrettyPrinter(indent=2)
    self.shelves = []
    for sh in self.appdata.get('core', {}).get('shelves', []):
      name = sh.get('name')
      if name != 'drill_shelf':
        self.shelves.append(Shelf(self, sh, self.appdata.get(name, [])))

    self.params = {} if params is None else params
    self.addDefaultParams(self.report_title())
    self.addDefaultParams(self.report_subtitle())
    for sh in self.shelves:
      for elem in sh.elements:
        self.addDefaultParams(elem.get('custom_expr', ''))

  # Pluck default parameter values from input string
  def addDefaultParams(self, s):
    m = re.findall('<<.+?>>', s)
    # js side seems to pick up ending default on a line rather than first
    m.reverse()
    for p in m:
      kv = p[2:-2].partition(':')
      if kv[1] == ':' and kv[2]:
        self.params.setdefault(kv[0], kv[2])

  # Perform arcviz type parameter substitution (see arc_params.js)
  def substitute(self, s):
    return parameterReplace(s, self.params, raise_on_error=False)

  def report_title(self):
    return self.appdata.get('report_title', '')

  def report_subtitle(self):
    return self.appdata.get('report_subtitle', '')

  def id(self):
    return self.dbapp.id

  def report_type(self):
    return self.dbapp.report_type

  def reportIsFilter(self):
    return self.report_type() in ['daterange', 'picklist', 'range']

  def pprint(self):
    self.pprinter.pprint(self.appdata)

  # For picker types with dataset scope, return the base column
  def getDatasetScopeColumn(self):
    col = None
    if self.report_type() == 'daterange':
      if self.user_settings.get('filter_dataset_scope') == "true":
        col = self.user_settings.get('daterange_base_field')

    elif self.report_type() in ['picklist', 'range']:
      if self.user_settings.get('filter_dataset_scope') == "true":
        col = self.user_settings.get('picklist_base_field')

    return col

  # True if app is parameter type (no query is build for these)
  def isParameterApp(self):
    return self.report_type() == 'picklist' and not self.user_settings.get('picklist_base_field')

  # True if this expression is an 'optional dim/agg' -- i.e it has a parameter that doesn't map
  # but has a default of ''.  See arc_reports_base.js/isShelfExpressionEmpty()
  #
  # Note that this is difficult for recommendation path as we don't know what columns
  # the optional item could refer to.
  def exprIsOptional(self, expr):
    (expr, unused) = splitExprAndAlias(expr)
    if not expr:
      return True

    # Substitute params but if no default and no param, can't be optional dim
    try:
      expr = self.substitute(expr)
    except:
      return False
    if not expr:
      return True

    # Like the js side, this finds stuff like 'max([<<dim1:>>])'
    for empty in ["[]", "['']", "('')"]:
      if expr.find(empty) != -1:
        return True
    return False

  def getDsreq(self, expand_params=True, filter_dims=False):
    dsreq = {"version": 1,
             "type": 'SQL',
             "dimensions": [],
             "aggregates": [],
             "filters": [],
             "having": [],
             "global_segments": [],
             "limit": str(Report.get_limit(self.appdata, self.report_type())),
             "dataset_id": self.dataset_id}
    self.dsreq = dsreq

    dd = self.appdata.get('report_derived_data')
    if dd:
      # Allow for parameterized DD definitions
      for di in dd:
        d = di.get('dsreq')
        for k in ['aggregates', 'dimensions']:
          for i in d.get(k, []):
            i['expr'] = self.substitute(i['expr'])
        for k in ['filters', 'having']:
          if k in d:
            d[k] = [self.substitute(e) for e in d[k]]

      dsreq['derived_data'] = dd

    # picker apps are different from shelf based apps
    if self.report_type() == 'daterange':
      col = self.user_settings.get('daterange_base_field')
      dsreq['aggregates'].append({'expr': 'min([%s]) as "min"' % col})
      dsreq['aggregates'].append({'expr': 'max([%s]) as "max"' % col})
      return dsreq

    if self.report_type() == 'picklist':
      col = self.user_settings.get('picklist_base_field')
      if self.exprIsOptional(col):
        raise Exception('No column specified on filter')
      agg = {'expr': self.substitute("[%s] as '%s'" % (col, col))}
      order = self.user_settings.get('filter_options_sort_order', 'ASCENDING')
      agg['order'] = {'asc': order == 'ASCENDING'}
      dsreq['aggregates'].append(agg)
      dsreq['limit'] = int(self.user_settings.get('picklist_itemcount', 1000)) + 1
      dsreq['distinct'] = True

      if self.phase and self.phase.get('phase') == 'search_text':
        dsreq['limit'] = int(self.user_settings.get('picklist_searchcount', 1000))

        # TODO: Should do data connection specific cast and lower case
        dsreq['filters'].append(" lower(cast([%s] as string)) LIKE '%%%s%%'" % \
                                (col, self.phase.get('search')))

      return dsreq

    if self.report_type() == 'range':
      col = self.user_settings.get('picklist_base_field')
      dsreq['aggregates'].append({'expr': 'min([%s])' % col})
      dsreq['aggregates'].append({'expr': 'max([%s])' % col})
      return dsreq

    # NULL order
    nulls_order = self.user_settings.get('nulls_order', 'Default')
    if nulls_order != 'Default':
      dsreq['nulls_order'] = nulls_order

    # Include global segments if needed
    gsegs = self.appdata.get('selected_segments', [])
    if len(gsegs):
      seglist = [{'seg_id':s.get('id'), 'negate':s.get('negate')} for s in gsegs]
      dsreq['global_segments'].append({'type':'SEGMENT', 'segments':seglist})

    for sh in self.shelves:
      for i in range(0, len(sh.elements)):
        agg, txt = sh.getExpr(i, dsreq, expand_params=expand_params, filter_dims=filter_dims)

    return dsreq

class Shelf(object):

  def __init__(self, app, desc, elements):
    self.app = app
    self.desc = desc
    self.elements = elements

  # A hack copied from arc_report_shelves.js to see if an expression
  # contains what we assume is an aggregate function.
  # TODO: The js side has gotten much more complex for this check
  def expressionIsAggregate(self, expr):
    expr = expr.lower()
    m = re.findall("""("[^"]+"|'[^']+'|[\w]+)""", expr)
    aggs = ['sum','count','max','min','avg','ndv','partial_ndv','final_ndv','having']
    return sum(w in aggs for w in m) > 0


  # Get the text for a shelf element
  # Mainly from arc_report_shelves.js/getExpression()

  # Awkward part is that we don't currently accurately track if the
  # expression has an aggregation or what is the data type of the
  # expression.

  def getExpr(self, num, dsreq=None, expand_params=True, filter_dims=False):
    elem = self.elements[num]
    expr = elem.get('custom_expr')
    colname = elem.get('dataset_colname')
    coltype = elem.get('dataset_coltype')
    alias = elem.get('col_alias')
    expr_hasagg = elem.get('expr_hasagg')
    fn_type = elem.get('fn_type')
    fn_name = elem.get('fn_name')
    fn_cast = elem.get('fn_cast')
    order = elem.get('order')
    order_limit = elem.get('order_limit')
    filter_column = elem.get('filter_column')
    filter_data = elem.get('filter_data')
    filter_exclude = elem.get('filter_exclude')

    # Here like in arc_report_shelves.js/element.custom_expr() if the expression
    # has an alias, then we force that one.
    (sexpr, salias) = splitExprAndAlias(expr)
    if salias:
      expr = sexpr
      alias = salias

    if self.desc.get('name','').startswith('filters') and filter_column:
      # Need some work here...see arc_reports_base.js/addFiltersToRb() (which seems
      # pretty odd, need more thought)
      if expr:
        text = expr
        agg = False
        # Can't put custom filter expressions on the dimension shelf
        filter_dims = False
      else:
        agg = False
        if filter_data and len(filter_data) > 0:
          inout = 'not in' if filter_exclude else 'in'
          text = filterExpr(filter_column, coltype, filter_data, inout)

        else:
          if not expand_params and filter_dims and dsreq:
            text = '[%s] as `%s`' % (filter_column, filter_column)
            dsreq['dimensions'].append({'expr':text, 'type':'SIMPLE', 'alias':filter_column})
          else:
            key = None
            if expand_params:
              key = self.app.params.get(filter_column)
            if not key:
              text = '(1 = 1)'
            else:
              if coltype == 'STRING':
                key = '"' + key + '"'
              text = "[%s] %s (%s)" % (filter_column, 'in', key)

      if expand_params:
        text = self.app.substitute(text)
      if dsreq and not filter_dims and text != '(1 = 1)':
        agg = agg or self.expressionIsAggregate(text)
        dsreq['having' if agg else 'filters'].append(text)

    else:

      # From arc.report.columnTypeEnum
      agg = expr_hasagg or (self.desc.get('column_type') == 2)
      if expr:
        text = expr;
        agg = agg or (self.app.report_type() not in ['singlevalue', 'kpi'] and self.expressionIsAggregate(text))

      elif colname == 'arc_segment':
        agg = False
        text = colname;

      else:
        text = "[%s]" % colname
        if fn_cast:
          alias = alias or colname
          text = "cast(%s as %s)" % (text, fn_cast)
        if fn_name:
          # From arc_reports.js/shelfElementPostProcess()
          # TODO: Need to generalize the shelf element processing to capture
          #       the full functionality of that routine.
          if fn_name == 'countdistinct':
            text = 'count(distinct (' + text + '))'
            agg = True
          else:
            text = "%s(%s)" % (fn_name, text)
            if fn_cast:
              alias = "%s(%s)" % (fn_name, alias)
          # From arc.report.fn_types
          agg = agg or (fn_type == 1)

      # Pick a character to quote the alias with
      q = "'"
      if not alias:
        if q in text:
          q = '"'
          if q in text:
            q = '`'
            if q in text:
              q = ''
        alias = text.replace('[','').replace(']','')

      text = text + " as " + q + alias + q

      if expand_params:
        text = self.app.substitute(text)

      if dsreq and not self.app.exprIsOptional(text):
        dselem = {'alias':alias}
        if order:
          if order_limit:
            dsreq['limit'] = order_limit
          dselem['order'] = {'asc':order.get('ascending',False),
                             'pri':order.get('priority', 1)}
        if agg:
          dselem['expr'] = text
          dsreq['aggregates'].append(dselem)
        else:
          if text == 'arc_segment as `Segment`':
            dselem['type'] = 'SEGMENT'
            del dselem['alias']
            dselem['segments'] = [{'seg_id':s.get('id'),
                                   'negate':s.get('negate',False)} for s in (filter_data or [])]
          else:
            dselem['type'] = 'SIMPLE'
            dselem['expr'] = text
          dsreq['dimensions'].append(dselem)

    return agg, text

