# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"File-based cache backend"

#
# Note that this is a copy of django's cache/backends/filebased.py
# with a change for the culling rate
#

import hashlib
import os
import shutil
import time
import _thread
try:
    from django.utils.six.moves import cPickle as pickle
except ImportError:
    import pickle

from django.core.cache.backends.base import BaseCache, DEFAULT_TIMEOUT
from django.utils.encoding import force_bytes

from django.conf import settings
import threading
import logging

CULL_LOCK = threading.Lock()

# Initialize to current time so that cli invocations do not end up running a cache
# cull check at startup
LAST_CULL = time.time()

LOGGER = logging.getLogger('arcweb')

class FileBasedCache(BaseCache):
    def __init__(self, dir, params):
        BaseCache.__init__(self, params)

        # Do not cull more that once every 4 hours
        options = params.get('OPTIONS', {})
        self._cull_mintime = int(params.get('cull_mintime',
            options.get('CULL_MINTIME', 3600 * 4)))

        self._dir = dir
        if not os.path.exists(self._dir):
            self._createdir(self._dir)

    def add(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        if self.has_key(key, version=version):
            return False

        self.set(key, value, timeout, version=version)
        return True

    def get(self, key, default=None, version=None):
        key = self.make_key(key, version=version)
        # self.validate_key(key)

        fname = self._key_to_file(key)
        try:
            with open(fname, 'rb') as f:
                exp = pickle.load(f)
                now = time.time()
                if exp is not None and exp < now:
                    self._delete(fname)
                else:
                    return pickle.load(f)
        except:
            # If we cannot read from the cache for ANY reason, return
            # the default value
            pass

        return default

    def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_key(key, version=version)
        # self.validate_key(key)

        fname = self._key_to_file(key)
        dirname = os.path.dirname(fname)

        if timeout == DEFAULT_TIMEOUT:
            timeout = self.default_timeout

        self._cull()

        try:
            if not os.path.exists(dirname):
                self._createdir(dirname)

            tempfile = '%s_%s_%s' % (fname, os.getpid(), _thread.get_ident())
            with open(tempfile, 'wb') as f:
                expiry = None if timeout is None else time.time() + timeout
                pickle.dump(expiry, f, pickle.HIGHEST_PROTOCOL)
                pickle.dump(value, f, pickle.HIGHEST_PROTOCOL)
            shutil.move(tempfile, fname)
        except:
            # If we cannot set a cache entry for ANY reason,
            # log it and move on
            LOGGER.exception("Cannot set cache entry")
            pass

    def delete(self, key, version=None):
        key = self.make_key(key, version=version)
        # self.validate_key(key)
        try:
            self._delete(self._key_to_file(key))
        except (IOError, OSError):
            pass

    def _delete(self, fname):
        os.remove(fname)
        try:
            # Remove the 2 subdirs if they're empty
            dirname = os.path.dirname(fname)
            os.rmdir(dirname)
            os.rmdir(os.path.dirname(dirname))
        except (IOError, OSError):
            pass

    def has_key(self, key, version=None):
        key = self.make_key(key, version=version)
        # self.validate_key(key)
        fname = self._key_to_file(key)
        try:
            with open(fname, 'rb') as f:
                exp = pickle.load(f)
            now = time.time()
            if exp < now:
                self._delete(fname)
                return False
            else:
                return True
        except:
            # If we cannot check cache contents for ANY reason,
            # treat it as similar to not having a cache key
            return False

    def _cull(self):

        if not getattr(settings, 'IS_MASTER', True):
          return

        global LAST_CULL
        with CULL_LOCK:
          start = time.time()
          if start - LAST_CULL < self._cull_mintime:
            return
          else:
            LAST_CULL = start

        num_entries = int(self._num_entries)
        LOGGER.debug('Cache cull check: counted %d entries in %.2f seconds.' % (num_entries, time.time() - start))
        if num_entries < self._max_entries:
            return

        start = time.time()
        try:
            filelist = sorted(os.listdir(self._dir))
        except:
            # If we cannot interact with a particular directory
            # log the error and move on
            LOGGER.exception("Cannot enumarate cache directory")
            return

        if self._cull_frequency == 0:
            doomed = filelist
        else:
            doomed = [os.path.join(self._dir, k) for (i, k) in enumerate(filelist) if i % self._cull_frequency == 0]

        deleted = 0
        for topdir in doomed:
            try:
                for root, _, files in os.walk(topdir):
                    deleted += len(files)
                    for f in files:
                        self._delete(os.path.join(root, f))
            except:
                # If we cannot iterate a directory or delete a file for ANY reason,
                # ignore the error and move on
                LOGGER.exception("Cannot cull cache entries")
                pass

        LOGGER.debug('Cull complete: deleted %d entries in %.2f seconds.' % (deleted, time.time() - start))

    def _createdir(self, dir):
        try:
            os.makedirs(dir)
        except OSError:
            if not os.path.isdir(self._dir):
                raise EnvironmentError("Cache directory '%s' does not exist and could not be created'" % dir)

    def _key_to_file(self, key):
        """
        Convert the filename into an md5 string. We'll turn the first couple
        bits of the path into directory prefixes to be nice to filesystems
        that have problems with large numbers of files in a directory.

        Thus, a cache key of "foo" gets turnned into a file named
        ``{cache-dir}ac/bd/18db4cc2f85cedef654fccc4a4d8``.
        """
        path = hashlib.md5(force_bytes(key)).hexdigest()
        path = os.path.join(path[:2], path[2:4], path[4:])
        return os.path.join(self._dir, path)

    def _get_num_entries(self):
        count = 0
        for _,_,files in os.walk(self._dir):
            count += len(files)
        return count
    _num_entries = property(_get_num_entries)

    def clear(self):
        try:
            shutil.rmtree(self._dir)
        except:
            # If we cannot clear the cache for ANY reason,
            # log the error and move on
            LOGGER.exception("Cannot clear cache entry")
            pass

# For backwards compatibility
class CacheClass(FileBasedCache):
    pass
