# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json, sys, re

from django.conf import settings
DD_USE_SUBQUERY = getattr(settings, 'DD_USE_SUBQUERY', False)

import logging

from dataconn import arctypes
from functools import cmp_to_key
import vizlib.util

LOGGER = logging.getLogger("arcweb")

COLUMN_RE = re.compile(r'(.*?)\[(.+?)\](.*)', re.DOTALL)

class _LowercaseDict(dict):
  """lowercase dictionary for case-insensitive column name matching"""
  # note: keys can only be strings!
  def __init__(self, kvs=None):
    if kvs is not None:
      return dict.__init__(self, [(k.lower(), v) for (k, v) in kvs])
    else:
      return dict.__init__(self)
  def __getitem__(self, name):
    return dict.__getitem__(self, name.lower())
  def get(self, name, default=None):
    return dict.get(self, name.lower(), default)
  def __setitem__(self, name, value):
    dict.__setitem__(self, name.lower(), value)
  def __contains__(self, name):
    return dict.__contains__(self, name.lower())
  def __delitem__(self, name):
    return dict.__delitem__(self, name.lower())
  def setdefault(self, name, value):
    if not self.__contains__(name):
      self.__setitem__(name, value)
    return self.__getitem__(name)

class ModelDataset(object):
  """ Represent a dataset object (Django) """

  def __init__(self, jsoninp, properties, params=None,
      dsreqRewriter=None):
    self.properties = properties
    self._dsreqRewriter = dsreqRewriter
    self.params = params
    self.quote_alias = properties.quote_alias
    self.quote = properties.quote
    self.quote_end = properties.quote_end
    self.quote_escape_character = properties.quote_escape_character
    self.sample_stmt = ''
    inp = json.loads(jsoninp)

    if self._dsreqRewriter:
      inp = self._dsreqRewriter.getExtractDataset()

    self.tables = []
    for t in inp:
      self.tables.append(ModelDataset.Table(self, t, params=params))

    self.views = dict()

    cull_types = getattr(settings, 'JOIN_CULL_LIST', ['left outer join', 'inner join'])
    for t in self.tables:  # second pass; compute distance from root
      if t.isRoot():
        self.join_cull = t.join_cull
      if t.join_type.lower() not in cull_types:
        self.join_cull = False

    self.dsuhash = _LowercaseDict()

    for i, t in enumerate(self.tables):
      if self.properties and self.properties.supports_table_aliases:
        # if not t.alias: SG: always use TA_ alias. overide existing alias
        t.alias = "TA_%d" % i
      for c in t.columns:
        self.addColumn(c)
      t.initComplexColumns()

  # Helper function for sorting a set of tables for join--first on distance to root,
  # then on table alias but put native tables before  DD 'views'
  @classmethod
  def cmptab(cls, t1, t2):
    if t1.distance != t2.distance:
      return (t1.distance > t2.distance) - (t1.distance < t2.distance)
    if t1.alias.startswith('TA') == t2.alias.startswith('TA'):
      return (t1.alias > t2.alias) - (t1.alias < t2.alias)
    if t1.alias.startswith('TA'):
      return -1
    return 1

  def processExprViaRewriterParams(self, expressionStr):
    if not self._dsreqRewriter:
      return expressionStr

    return self._dsreqRewriter.processExpression(expressionStr)

  def addColumn(self, col):
    name = col.attr.get('alias') or col.attr.get('name')
    if name not in self.dsuhash:
      self.dsuhash[name] = col

  def getColumn(self, colname):
    return self.dsuhash.get(colname)

  # Table lookup--a little awkward because the table can be in the dataset more than
  # once in which case only the alias is unique.  So we look for the alias first, then
  # for the native tablename
  def getTable(self, tabname):
    assert tabname
    for t in self.tables:
      if t.alias == tabname:
        return t
    for t in self.tables:
      if t.tablename == tabname:
        return t
    raise Exception("Table %s not found in dataset" % tabname)

  def addTable(self, table):
    self.tables.append(table)

  def enquote(self, s):
    if not self.quote:
      return s

    return self.quote + \
           s.replace(self.quote, self.quote_escape_character + self.quote) + \
           self.quote_end

  def enquote_as(self, s):
    if self.properties and not self.properties.alias_supports_spaces:
      s = s.replace(" ", "_")

    qs = self.quote
    qe = self.quote_end

    if not qs:
      return s

    if qs == "\001":
      qs = '"'
      qe = '"'

    return qs + \
           s.replace(qs, self.quote_escape_character + qs) + \
           qe

  def get_expr(self, tablename, alias):
    quoted_table = self.quote_tablename(tablename)

    if self.properties and not self.properties.supports_table_aliases:
      alias = ' '
    else:
      alias = ' ' + alias + ' '

    if self.properties.__class__.__name__ in ('ArcEngineConnProperties',
                                              'ArcEngineV2ConnProperties',
                                              'ImpalaConnProperties',
                                              'ImpylaConnProperties'):
      return quoted_table + alias + self.sample_stmt
    else:
      return quoted_table + ' ' + self.sample_stmt + alias

  def outerJoinMinimumSet(self, tables):
    tableset = set(tables)

    # Helper function to get tables at a particular level
    def atlevel(level):
      return [t for t in tableset if t.distance == level]

    # find min/max level
    min_level = sys.maxsize
    max_level = -1
    for t in tables:
      min_level = min(min_level, t.distance)
      max_level = max(max_level, t.distance)

    # add parents for all levels but the lowest
    lev = max_level
    while lev > min_level:
      tableset.update(self.getTable(t.parent_tablename) for t in atlevel(lev))
      lev -= 1

    # now continue towards root until we have a single table at that level
    while lev >= 0 and len(atlevel(lev)) > 1:
      tableset.update(self.getTable(t.parent_tablename) for t in atlevel(lev))
      lev -= 1

    tablelist = sorted(list(tableset), key=cmp_to_key(self.cmptab))

    expr = ""
    for i, t in enumerate(tablelist):
      if i == 0:
        expr = self.get_expr(t.tablename, t.alias)
      else:
        expr += t.toSQL()
    return expr

  def quote_tablename(self, fq_table_name, non_alphanum=False,
                            ignore_unavailable_tables=False):
    if hasattr(self.properties, 'quote_tablename'):
      return self.properties.quote_tablename(fq_table_name)

    t = None
    if ignore_unavailable_tables == False:
      t = self.getTable(fq_table_name)
    # If table has a select statement, don't do any quoting
    if t and t.select:
      return '(%s)' % t.select
    # Or if it is already a select statement, don't do any quoting
    if fq_table_name.startswith('('):
      return fq_table_name
    fields = fq_table_name.split(".")
    if len(fields) == 1:
      return self.enquote(fq_table_name)
    else:
      if non_alphanum:
        tmp = re.sub(r'\W', '_', fields[-1])
        fields[-1] = tmp
      return ".".join([self.enquote(field) for field in fields])

  def joinComplexColumns(self, ftxt, columns):
    for c in columns:
      t = c.table
      expr = t.alias + '.' + c.attr.get('name') + ' ' + c.getComplexColAlias()
      ftxt += " INNER JOIN " + expr
    return ftxt

  def outerJoinAll(self, tables=None):
    dstables = set(t for t in self.tables if not t.fromView)
    if tables is not None:
      alltables = list(dstables | set(tables))
    else:
      alltables = list(dstables)
    return self.outerJoinMinimumSet(alltables)

  def outerJoin(self, tables):  # TODO: when API allows generalize to any type of join
    if self.join_cull:
      if len(tables) == 1:  # nothing to join if only 1 table in set
        alias = list(tables)[0].alias if hasattr(list(tables)[0], 'alias') else ''
        return self.get_expr(list(tables)[0].tablename, alias)
      else:
        return self.outerJoinMinimumSet(tables)
    else:
      return self.outerJoinAll(tables=tables)

  class Column(object):
    def __init__(self, tab, attr):
      self.table = tab
      self.attr = attr
      self.ccalias = None

    # Only base columns have a name -- everything else uses 'alias'
    def isBaseColumn(self):
      return bool(self.attr.get('name'))

    def isComplexColumn(self):
      return (self.attr.get('type', '').lower() in ['map', 'array', 'struct'])

    def hasComplexParent(self):
      pn = self.attr.get('parent')
      parent_exist_in_table = self.table.bchash.get(self.attr.get('parent'), False)
      return bool(pn) and parent_exist_in_table

    def needsParentJoin(self):
      col = self

      parent = self.table.bchash.get(col.attr.get('parent'))
      while parent is not None:
        is_parent_struct = parent.attr.get('type').lower() == 'struct'
        is_column_map_or_array = col.attr.get('type').lower() in ['map', 'array']
        if not is_column_map_or_array and not is_parent_struct:
          return parent

        col = parent
        parent = self.table.bchash.get(col.attr.get('parent'))

      return False

    def setComplexColAlias(self, ccalias):
      self.ccalias = ccalias

    def getComplexColAlias(self):
      return self.ccalias

    def getSelectName(self):
      pc = self.needsParentJoin()
      if pc:
        name = self.attr.get('name')[len(pc.attr.get('name')):]
        return pc.getComplexColAlias() + name
      else:
        pc = self.table.bchash.get(self.attr.get('parent'))
        return pc.table.alias + '.' + self.attr.get('name')

  class View(object):
    def __init__(self, dataset, label=None, alias=None, sql=None, columns=None, parent_tablename=None, dims=None,
                 join_exprs=None, view_dependents=None):
      self.dataset = dataset
      self.label = label
      self.sql = sql
      self.view_dependents = view_dependents or set()

      tabdef = {
        'fromView': True,
        'tablename': alias,
        'table_alias': alias,
        'parent_tablename': parent_tablename,
        'columns': columns or [],
        'join_type': 'INNER JOIN',
        'select': ('(%s)' % sql) if DD_USE_SUBQUERY else None,
        'join_exprs': join_exprs,
      }

      self.table = dataset.Table(dataset, inp=tabdef)
      self.dataset.addTable(self.table)

      if dims is not None:
        for key in list(dims.keys()):
          self.table.join_pairs.append(["["+dims[key]+"]", key])

      for c in self.table.columns:
        self.dataset.addColumn(c)

  class Table(object):
    def __init__(self, ds, inp, params=None):
      self.dataset = ds
      self.tablename = inp.get('tablename') or inp.get('lvname')
      self.select = inp.get('select')
      self.alias = inp.get('table_alias')
      self.uname = (self.alias if self.alias is not None else self.tablename)
      self.parent_tablename = inp.get('parent_tablename')
      self.join_pairs = inp.get('join_pairs', [])
      self.join_exprs = inp.get('join_exprs', [])
      self.join_type = inp.get('join_type', 'LEFT OUTER JOIN')
      self.join_cull = inp.get('enable_join_elim', True)
      self.columns = [ModelDataset.Column(self, c) for c in inp.get('columns', [])]
      self.fromView = inp.get('fromView', False)
      # We need to initialize bchash here in case we don't call the initComplexColumns function
      # To avoid runtime errors
      self.bchash = {}

      # Here we support the 'parameterized sql statement as table' feature by
      # doing the parameter substitution on the sql statement that defines
      # the table.
      self.params = params
      if self.select:
        from vizlib.accel import parameterReplace
        self.select = parameterReplace(self.select, self.params)

      if self.parent_tablename is None:
        self.distance = 0
      else:
        self.distance = self.dataset.getTable(self.parent_tablename).distance + 1

    # Set up table aliases for the complex columns within this table
    def initComplexColumns(self):
      self.bchash = dict((c.attr.get('name'), c) for c in self.columns if c.attr.get('name'))
      self.ccindex = 0
      for c in self.columns:
        if c.isComplexColumn():
          c.setComplexColAlias(self.alias + ('_CC_%d' % self.ccindex))
          self.ccindex += 1

    def isRoot(self):
      return self.parent_tablename is None

    def enquote(self, s):
      return self.dataset.enquote(s)

    def toSQL(self):

      if self.isRoot():
        return self.dataset.quote_tablename(self.tablename) + ' ' + \
            (self.alias or '')

      # See if column expression expands, otherwise use old style
      def predcol(cname, tname):
        expression = self.dataset.process_expr(cname)
        if expression == cname:
          expression = tname + '.' + self.enquote(cname)
        return expression

      expr = ' ' + self.join_type + ' ' + \
             self.dataset.quote_tablename(self.tablename) + \
             ' ' + (self.alias or '')
      preds = []
      pn = self.dataset.getTable(self.parent_tablename).alias
      for val in self.join_pairs:
        preds.append(predcol(val[0], pn) + ' = ' + predcol(val[1], self.alias))
      for val in self.join_exprs:
        preds.append(self.dataset.process_expr(val))

      if preds:
        expr += " ON ( " + " and ".join(preds) + " ) "

      return expr

  def process_expr(self, expr, columns=None, alias=False):
    originalExpr = expr
    expr = self.processExprViaRewriterParams(expr)
    frags = []
    coltrack = {}
    m = re.match(COLUMN_RE, expr)
    col = None
    counter = 0
    counter_limit = getattr(settings, 'EXPRESSION_EXPANSION_LIMIT', 0)

    while m:
      counter += 1
      if counter_limit and counter > counter_limit:
        raise Exception('Exression expansion limit of %s exceeded: %s' %
                        (counter_limit, originalExpr))

      if m.group(1):
        # strip out the agg function, we deal with the closing paren below
        if self.properties.__class__.__name__ in ('CassandraConnProperties') \
            and m.group(1).startswith('agg('):
         pass
        else:
          frags.append(m.group(1))
      col = self.dsuhash.get(m.group(2))
      if col:
        if columns is not None and col.isBaseColumn():
          columns.add(col)
        if not alias:

          ctxt = col.attr.get('expr')
          key = m.group(2)

          if ctxt:
            # When we expand a column, the remaining text must be
            # shorter than the last time we expanded this column
            if coltrack.get(key) is None:
              coltrack[key] = m.group(3)
            elif len(coltrack[key]) <= len(m.group(3)) and m.group(3).endswith(coltrack[key]):
              raise Exception("Circular attribute expansion detected for [%s] -> %s" % (key, ctxt))
            coltrack[key] = m.group(3)

        else:
          ctxt = self.enquote(col.attr.get('alias') or col.attr.get('name'))

        if not ctxt:
          # Can't quote column names that are complex type references
          if col.hasComplexParent():
            ctxt = col.getSelectName()
          else:
            # Cassandra doesn't support table aliases nor does it like fully qualified
            # table names in the select list
            if self.properties and self.properties.supports_table_aliases:
              ctxt = col.table.alias + '.' + self.enquote(col.attr.get('name') or col.attr.get('alias'))
            else:
              ctxt = self.enquote(col.attr.get('name') or col.attr.get('alias'))

          # This is where we have expanded a dataset column name into a table column reference
          # so here we apply the cast operation if it is present in the dataset.
          if col.attr.get('cast'):
            casttype = col.attr.get('cast').upper()
            if casttype == arctypes.ARC_DATATYPE_STRING:
              ctxt = '(CAST(%s as %s))' % (ctxt, self.properties.string_type)
            elif casttype == arctypes.ARC_DATATYPE_BOOLEAN:
              ctxt = '(CAST(%s as %s))' % (ctxt, self.properties.boolean_type)
            elif casttype == arctypes.ARC_DATATYPE_TIMESTAMP:
              ctxt = '(CAST(%s as %s))' % (ctxt, self.properties.timestamp_type)
            elif casttype == arctypes.ARC_DATATYPE_INT:
              ctxt = '(CAST(%s as %s))' % (ctxt, self.properties.integer_type)
            elif casttype == arctypes.ARC_DATATYPE_REAL:
              ctxt = '(CAST(%s as %s))' % (ctxt, self.properties.real_type)
            else:
              ctxt = '(CAST(%s as %s))' % (ctxt, casttype)

        if m.group(1) and m.group(3):
          #Cassandra can't handle double parens
          if self.properties.__class__.__name__ in ('CassandraConnProperties'):
            if m.group(1).startswith('agg('):
              # strip out the closing paren ')' from the agg functions
              expr = ctxt + m.group(3)[1:]
            else:
              expr = ctxt + m.group(3)
          else:
            expr = '(' + ctxt + ')' + m.group(3)
        elif m.group(3):
          expr = ctxt + m.group(3)
        else:
          expr = ctxt

      # Here the text between the brackets did not match a column name, so
      # we put the brackets back and put it into the processed fragments list.
      else:
        if getattr(settings, 'LEGACY_BRACKET_HANDLING', False):
          # Legacy mode strips the brackets and iterates.
          expr = m.group(2) + m.group(3)
        else:
          frags.append('[' + m.group(2) + ']')
          expr = m.group(3)
      m = re.match(COLUMN_RE, expr)

    # Permit parameters in column expressions
    # Substitute parameters here
    from vizlib.accel import parameterReplace

    r = ''.join(frags) + expr
    r = parameterReplace(r, self.params, raise_on_error=False)

    return r

  def allColumns(self, dataset, sb=None, dscols=None):
    if sb is None:
      from .select_builder import BaseSelectBuilder

      sb = BaseSelectBuilder(dataset)
    cols = []
    for t in self.tables:
      sb.tables.add(t)
      for c in t.columns:
        if sb.hidden_columns and c.attr.get('hidden'):
          continue
        if c.isComplexColumn():
          continue
        if not c.attr.get('exprhasagg'):
          #if not c.attr.get('name') and c.attr.get('expr'):
          alias = c.attr.get('alias') or c.attr.get('name')
          expr = '[' + alias + '] as ' + self.enquote_as(alias)
          if dscols is not None:
            lv_name = re.sub(r'\W', '_', alias.lower())
            expr = '[' + alias + '] as ' + self.enquote_as(lv_name)
            col_attrs = dict(c.attr)
            col_attrs['name'] = lv_name
            col_attrs['alias'] = alias
            # Remove column attributes that we don't want to copy
            # For an LV, there is no concept of complex type or parent column
            for key in ['parent']:
              if key in col_attrs:
                del col_attrs[key]
            dscols.append(col_attrs)
          cols.append(expr)
    return cols

  def addAllToSb(self, sb, dscols=None):
    cols = self.allColumns(sb.dataset, sb=sb, dscols=dscols)
    list(map(lambda x: sb.addAggregate(self.process_expr(x, columns=sb.columns)), cols))

class ModelEvents(object):
  """ Represent an Events object (Django) """

  def __init__(self, dataset):
    from datasets.models import Event

    self.events = Event.objects.filter(dataset__in=dataset)


class ModelSegment(object):
  """ Represent a segment object (Django) """

  def __init__(self, jsoninp):
    self.data = None
    try:
      self.data = json.loads(jsoninp)
    except Exception as e:
      raise Exception("Unable to parse dataset: " + str(e))


# True if input type is a 'Measure' type
def column_is_measure(coltype):
  return str(coltype).upper() in ['INT', 'INTEGER', 'BIGINT', 'REAL', 'FLOAT', 'DOUBLE']


# function for getting dataset_info when absent (for single table)
# UI should always store dataset_info but older versions don't (for single table)
# this is for backwards compatibility
# example: [{"tablename":"default.census_pop",
# "columns":[{"name":"year","type":"INT","isdim":false,"alias":"year"},
# {"name":"state","type":"STRING","isdim":true,"alias":"state"},
# {"name":"population","type":"BIGINT","isdim":false,"alias":"population"}
#                      ]
#           }]
def getDatasetInfo(tableName, dc_id, username, select=None, nice_names=False):

  # e.g. {"info": ["Tables: "], "colnames": ["year", "state", "population"],
  #                             "coltypes": ["INT", "STRING", "BIGINT"],
  if select is None:
    from reports.views import columns_internal
    content = columns_internal(username, tableName, dataconnection_id=dc_id, no_cache=True)
  else:
    from dataconn.dataconn import get_dataconn
    from dataconn.connections_types import DATACONN_HIVE, DATACONN_HYVE
    dc = get_dataconn(dc_id)
    query = 'select * from (%s) `%s` limit 0' % (select, tableName)
    content = dc.sync_runquery(query, username=username)
    if dc.connection_type == DATACONN_HIVE or dc.connection_type == DATACONN_HYVE:
      content['colnames'] = [n[n.rfind('.') + 1:] for n in content['colnames']]

  colnames = content["colnames"]
  coltypes = content["coltypes"]
  partitions = content.get("partitions", {})
  hive_comments = content.get("hive_comments")
  data = dict()
  data["tablename"] = tableName
  data["columns"] = list()
  for idx, colname in enumerate(colnames):
    coltype = coltypes[idx]
    tmp2 = dict()
    tmp2["name"] = colname
    tmp2["type"] = coltype
    tmp2["isdim"] = not column_is_measure(coltype)
    tmp2["alias"] = tmp2["name"]
    if hive_comments and hive_comments[idx]:
      tmp2["hivecomment"] = hive_comments[idx]
    if colname in partitions:
      tmp2["ispart"] = True
      tmp2["isdim"] = True
    data["columns"].append(tmp2)

  # Build the parents for complex types
  coldict = dict((c["name"], c) for c in data["columns"])
  for c in data["columns"]:
    name = c["name"]
    rdot = name.rfind('.')
    if rdot != -1:
      c["parent"] = name[0:rdot]
      c["isdim"] = True
      coldict[c["parent"]]["isdim"] = True

  if nice_names:
    for c in data["columns"]:
      c["alias"] = niceName(c["alias"])

  return json.dumps([data])

# Convert a column name to nice_name format
def niceName(name):
  if not name:
    return name

  # For complex types, strip everything before the final '.'
  # TODO: Better only do this for columns with a 'parent' attribute.
  dot = name.rfind('.')
  if dot != -1 and dot != len(name) - 1:
    name = name[dot + 1:]

  name = name.replace('_', ' ')
  pos = 0
  while True:
    name = name[:pos] + name[pos].upper() + name[pos + 1:]
    pos = name.find(' ', pos) + 1
    if pos == 0 or pos >= len(name):
      break

  return name

# Return true if dataset column aliases have been Title Cased (my_col -> My Col)
def datasetHasNiceNames(ds):
  info = json.loads(ds.dataset_info)
  nice = 0
  total = 0
  for table in info:
    for col in table.get('columns'):
      name = col.get('name')
      if name:
        total +=1;
        alias = col.get('alias') or name
        nice += alias == niceName(alias)

  return nice * 2 > total

# Update on dataset table columns from current table
# Returns True if any changes made
def updateDatasetTableColumns(ds, table, username, nice_names):

  data = getDatasetInfo(table.get('tablename'), ds.dataconnection_id, username,
                        select=table.get('select'), nice_names=nice_names)
  data = json.loads(data)[0]

  newcols = dict((col.get('name'), col) for col in data.get('columns'))
  oldcols = table.get('columns')

  attrs = ['type', 'hivecomment', 'parent']
  changed = False
  dropcols = []

  for oldcol in oldcols:
    # Skip calculated fields
    name = oldcol.get('name')
    if not name:
      continue

    # Check column not in table anymore
    newcol = newcols.get(name)
    if not newcol:
      dropcols.append(oldcol)
      continue

    # Column is not new but might have changed
    del newcols[name]
    if sum(oldcol.get(attr) != newcol.get(attr) for attr in attrs):
      changed = True
      for attr in attrs + ['isdim']:
        if attr in newcol:
          oldcol[attr] = newcol[attr]
        elif attr in oldcol:
          del oldcol[attr]
      if oldcol['type'] == oldcol.get('cast'):
        del oldcol['cast']

  # Add in any new columns
  if newcols:
    changed = True
    oldcols.extend(list(newcols.values()))

  # Remove any dropped columns
  if dropcols:
    changed = True
    for col in dropcols:
      oldcols.remove(col)

  return changed

# This performs the same function as arc_datasets.js/Dataset.update_tables()
# Eventually we should move that function to use this logic.
def refreshDatasetColumns(ds, username):
  changed = False
  if not ds.dataset_info:
    return changed

  info = json.loads(ds.dataset_info)
  nice_names = datasetHasNiceNames(ds)
  for table in info:
    changed |= updateDatasetTableColumns(ds, table, username, nice_names)

  if changed:
    ds.update_user = username
    # Avoid circular import
    from datasets.views import dataset_update_info
    dataset_update_info(ds, json.dumps(info))

  return changed
