# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
from .dataset import ModelDataset
from vizlib.select_builder import BaseSelectBuilder
from dataconn.dataconn import get_dataconn


class DjangoDatasetLV(object):
  """
  This object represents a DataSet.dataset_lvinfo json string.
  Use it to parse an existing LV json info.
  """
  def __init__(self, lvjson):
    try:
      # lvjson comes from DataSet.dataset_lvinfo
      inp = json.loads(lvjson)[0]
    except:
      raise
    self.lvname = inp.get('lvname') # name for LV
    self.lvinfo = inp.get('columns') # list of tables and columns for LV (as with dataset_info in datasets/models.py)
    self.lvquery = inp.get('lvquery') # query/definition for LV

def dataset_update_lvinfo(ds, user_name, lvname=None, new_ds_info=None, old_ds_info=None):
  """
  Call this method to update a DataSet's lvinfo prior to saving it.
  Method returns ds unsaved, so user is responsible for persisting ds.
  4 cases for updating lvinfo
  - create new LV (CREATE)
  - remove associated LV (DROP)
  - only rename existing LV (ALTER)
  - change existing LV's definition and/or name (ALTER)
  :param ds: input Dataset object
  :param user_name:
  :param lvname: name of LV
  :param new_ds_info: ds_info for building a new LV
  :param old_ds_info: existing ds_info
  :return: True/False (whether LV was created/updated)
  """

  # Dataset may already be associated with an LV
  old_lvinfo_obj = None if not ds.dataset_lvinfo else DjangoDatasetLV(ds.dataset_lvinfo)

  # ds_info contains the set of tables and columns required for creating a LV
  ds_info = None if old_lvinfo_obj is not None else old_ds_info
  # if new ds_info is specified and different from old ds_info set that as ds_info for building LV
  if new_ds_info and _compare_infos(old_ds_info, new_ds_info):
    ds_info = new_ds_info

  if not old_lvinfo_obj:
  # No previously created associated Dataset LV exists
  # 1a. return if no lvname is provided
  # 1b. create new one if lvname is provided
    if not lvname:
      return False
    else:
      return _create_lv(ds, user_name, lvname, ds_info) if ds_info is not None else False
  else:
  # Previously created associated Dataset LV exists
  # 2. Drop it if lvname is 'DELETE'
  # 3. Rename it if nothing has changed other than its name
  # 4. Alter it if its definition has changed
    if lvname == "DELETE":
      return _drop_lv(ds, user_name, old_lvinfo_obj.lvname)
    if ds_info is None:
      return _alter_lv(ds, user_name, old_lvinfo_obj.lvname, lvname)
    else:
      return _alter_lv(ds, user_name, old_lvinfo_obj.lvname, lvname, ds_info)

def _compare_infos(old_ds_info, new_ds_info):
  """
  Compare old and new Dataset infos after removing "enable_join_elim" entry.
  This is a simple comparison of serialized dictionary values. We are relying on json.dumps library
  and that is preserves the order of elements in the dictionary, which may not be the case.
  In the future we should use instead an OrderedDictionary and a deep object comparison.
  Also we should not use json.dumps serialization in that case since it
  is known to break an OrderedDictionary's order (http://bugs.python.org/issue6105).
  :param old_ds_info: unicode String of previously used Dataset info
  :param new_ds_info: unicode String of new Dataset info
  :return: True when different and False when unchanged
  """
  if old_ds_info is None or new_ds_info is None:
    return True
  old_obj = json.loads(old_ds_info) # list of tables/columns
  new_obj = json.loads(new_ds_info) # list of tables/columns
  # filter out "enable_join_elim" entries and compare old and new infos
  for d in old_obj:
    if 'enable_join_elim' in d:
      del d['enable_join_elim']
  for d in new_obj:
    if 'enable_join_elim' in d:
      del d['enable_join_elim']
  if json.dumps(old_obj) != json.dumps(new_obj):
    return True
  return False

def _create_lv(ds, user_name, lvname, ds_info):
  dscols = []
  sub_query = _build_query(ds.dataconnection.id, ds_info, dscols)
  query = "CREATE VIEW " + enquote(lvname) + " AS " + sub_query
  _execute(ds, query, user_name)
  lv_info = {'lvname' : lvname,
             'columns' : dscols,
             'lvquery' : sub_query}
  ds.dataset_lvinfo = json.dumps([lv_info])
  return True

def _drop_lv(ds, user_name, lvname):
  query = "DROP VIEW IF EXISTS " + enquote(lvname)
  _execute(ds, query, user_name)
  ds.dataset_lvinfo = ""
  return True

def _alter_lv(ds, user_name, old_name, new_name, ds_info=None):
  lv_info = DjangoDatasetLV(ds.dataset_lvinfo)
  sub_query = None
  dscols = []
  updated = False
  if old_name != new_name:
    query = "ALTER VIEW " + enquote(old_name) + " RENAME TO " + enquote(new_name)
    _execute(ds, query, user_name)
    updated = True
  if ds_info is not None:
    sub_query = _build_query(ds.dataconnection.id, ds_info, dscols)
    query = "ALTER VIEW " + enquote(new_name) + " AS " + sub_query
    _execute(ds, query, user_name)
    updated = True
  lv_info.lvname = new_name
  lv_info.lvinfo = ds_info if ds_info is not None else lv_info.lvinfo
  lv_info.lvquery = sub_query if sub_query is not None else lv_info.lvquery
  lv_info = {'lvname': lv_info.lvname,
             'columns': dscols if dscols else lv_info.lvinfo,
             'lvquery': lv_info.lvquery}
  ds.dataset_lvinfo = json.dumps([lv_info])
  return updated

def _build_query(dc_id, ds_info, dscols):
  dc = get_dataconn(dc_id)
  dataset = ModelDataset(ds_info, properties=dc.properties())
  sb = BaseSelectBuilder(dataset, properties=dc.properties())
  dataset.addAllToSb(sb, dscols=dscols)
  return sb.getsql(nolimit=True)

def _execute(ds, query, user_name):
  dc = get_dataconn(ds.dataconnection.id)
  try:
    dc.sync_runquery(query=query, username=user_name)
  except:
    raise

def enquote(name):
  if not name:
    return None
  splits = name.split('.')
  if len(splits) > 1:
    return '`' + splits[0] + '`.`' + ''.join(splits[1:]) + '`'
  else:
    return name
