# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import logging
from django.utils.html import escape

from datasets.models import Segment
from datasets.models import DataSet
from .dataset import ModelSegment

from django.conf import settings

# create logger
LOGGER = logging.getLogger("arcweb")

REQUEST_TYPES = set()
from dataconn.data_api_cmds import NORM_CMDS, SPECIAL_CMDS, VALIDATE_CMDS
for n in list(NORM_CMDS.values()):
  REQUEST_TYPES.add(n)
for n in list(SPECIAL_CMDS.values()):
  REQUEST_TYPES.add(n)
for n in VALIDATE_CMDS.values():
  REQUEST_TYPES.add(n)

class DatasetRequest(object):
  """ Represent a dataset request """

  def __init__(self, dsreq, ds=None, params=None, django_ds=None):
    self.dataset = ds # Allow operation with dataset already loaded
    self.django_ds = django_ds # Keep ahold of the django dataset if already loaded
    self.params = params # Save input viz params for later ModelDataset
    """ Parse the request, ensure we have the dataset info """
    try:
      data = dsreq if type(dsreq) == type({}) else json.loads(dsreq)
      self.dstype = data.get('type', None)
      if self.dstype is None:
        LOGGER.error('request type is required')
        raise ValueError('request type is required')
      if not REQUEST_TYPES.__contains__(self.dstype):
        msg = "Invalid request_type: '%s'" % escape(self.dstype)
        raise ValueError(msg)
      if self.dstype == NORM_CMDS['SOLR']:
        self.dsreq = SolrRequest(data)
      elif self.dstype == NORM_CMDS['HTTP']:
        self.dsreq = HttpRequest(data)
      elif self.dstype == NORM_CMDS['SQL']:
        self.dsreq = SBrequest(data)
      elif self.dstype == NORM_CMDS['MR']:
        self.dsreq = MRrequest(data)
      elif self.dstype == NORM_CMDS['COLUMN_STATS']:
        self.dsreq = CLrequest(data)
      elif self.dstype == NORM_CMDS['getrows']:
        self.dsreq = RowsRequest(data)
      elif self.dstype == NORM_CMDS['CREATEAVIEW']:
        self.dsreq = CreateAViewRequest(data)
      elif self.dstype == NORM_CMDS['SAVE_TABLE']:
        self.dsreq = SaveTableRequest(data)
      elif self.dstype == SPECIAL_CMDS['SHOW_DATABASES']:
        self.dsreq = DBrequest(data)
      elif self.dstype == SPECIAL_CMDS['SHOW_TABLES']:
        self.dsreq = TArequest(data)
      elif self.dstype == SPECIAL_CMDS['SHOW_AVIEWS']:
        self.dsreq = ShowAViewRequest(data)
      elif self.dstype == SPECIAL_CMDS['REFRESH_AVIEWS']:
        self.dsreq = RefreshAViewRequest(data)
      elif self.dstype == SPECIAL_CMDS['DROP_AVIEWS']:
        self.dsreq = DropAViewRequest(data)
      elif self.dstype == SPECIAL_CMDS['INVALIDATE_METADATA']:
        self.dsreq = InvalidateMetadataRequest(data)
      elif self.dstype == SPECIAL_CMDS['COMPUTE_STATS']:
        self.dsreq = ComputeStatsRequest(data)
      elif self.dstype in list(VALIDATE_CMDS.values()):
        self.dsreq = ValidatorRequest(data)
      elif self.dstype == SPECIAL_CMDS['DATA_IMPORT']:
        self.dsreq = DataImportRequest(data)
      else:
        LOGGER.error('unknown request type: %s' % self.dstype)
        raise ValueError('unknown request type: %s' % self.dstype)
    except Exception as e:
      LOGGER.error(str(e))
      raise Exception("Unable to parse request: " + str(e))

  # Validate ourself based on user dataconnection and/or dataset permissions
  def validateAccess(self, user, dc, ds):

    if self.dstype == SPECIAL_CMDS['SHOW_TABLES'] or \
       self.dstype == SPECIAL_CMDS['SHOW_DATABASES'] or \
       self.dstype == SPECIAL_CMDS['COMPUTE_STATS'] or \
       self.dstype == SPECIAL_CMDS['INVALIDATE_METADATA'] or \
       self.dstype == NORM_CMDS['getrows']:
      dc.checkPerm(user, 'dc_explore')

    elif self.dstype == SPECIAL_CMDS['DATA_IMPORT'] or \
         self.dstype == NORM_CMDS['SAVE_TABLE']:
      dc.checkPerm(user, 'dc_upload')

    elif self.dstype in \
        [SPECIAL_CMDS['REFRESH_AVIEWS'], SPECIAL_CMDS['DROP_AVIEWS'],
         SPECIAL_CMDS['SHOW_AVIEWS'], NORM_CMDS['CREATEAVIEW']]:
      dc.checkPerm(user, 'dc_aviews')

    if ds:
      ds.checkPerm(user, 'ds_appview')

  class DatasetDerivedData(object):
    def __init__(self, inp):
      self.label = inp.get('label')
      self.dsreq = inp.get('dsreq')

  class DatasetOrder(object):
    def __init__(self, inp):
      self.ascending = inp.get('asc', False)
      self.priority = inp.get('pri')

  class DatasetSegment(object):
    def __init__(self, inp, params=None):
      self.id = inp['seg_id']
      self.negate = inp.get('negate')
      self.params = params
      from sqlrun.jsonselect_cache import jsonselect_get_segment
      self.segment = ModelSegment(jsonselect_get_segment(self.id).segment_data)

    def getexpr(self):
      entities = self.segment.data.get('entities')
      txt = "(" + " AND ".join(self.segment.data['filters']) + ")"
      if not len(entities):
        if self.params is not None:
          from vizlib.accel import parameterReplace
          txt = parameterReplace(txt, self.params, raise_on_error=False)
        if self.negate:
          txt = "(NOT " + txt + ")"
      return txt

    def getgroup(self):
      gname = self.segment.data['group'] or 'nogroup'
      return gname

    def addEntityToSb(self, entity, sb):
      tname = "arc_u%d" % sb.entity_alias
      sb.entity_alias += 1
      from .select_builder import BaseSelectBuilder, HiveSelectBuilder, \
        OracleSelectBuilder, CassandraSelectBuilder
      if type(sb) is HiveSelectBuilder:
        esb = HiveSelectBuilder(sb.dataset)
      elif type(sb) is OracleSelectBuilder:
        esb = OracleSelectBuilder(sb.dataset)
      elif type(sb) is CassandraSelectBuilder:
        esb = CassandraSelectBuilder(sb.dataset)
      else:
        esb = BaseSelectBuilder(sb.dataset)
      esb.indent = sb.indent + "  "
      esb.addFilter(self.getexpr())
      esb.entity = True  # turn flag on
      pairs = []
      for e in entity:
        e = sb.dataset.process_expr(e, columns=esb.columns)
        cname = "%s_%d" % (tname, len(pairs))
        pairs.append((e, cname))
        esb.addDimension(e + ' as "' + cname + '"')
      sb.addTables(esb.tables)
      if not self.negate:
        sb.addJoin("%s(%s) %s ON %s" %
                   (esb.indent, esb.getsql(), tname, " AND ".join("%s=%s" % (e, a) for (e, a) in pairs)))
      else:
        sb.addJoin("%s(%s) %s ON %s " %
                   (esb.indent, esb.getsql(), tname, " AND ".join("%s=%s" % (e, a) for (e, a) in pairs)),
                   join_type="left outer join")
        sb.addFilter(" AND ".join("%s is null" % a for (e, a) in pairs))
      esb.entity = False  # turn flag off


    def addToSb(self, sb):
      entities = self.segment.data.get('entities')
      if not len(entities):
        sb.addFilter(self.getexpr())
      else:
        self.addEntityToSb(entities, sb)

  class DatasetGlobalSegment(object):
    def __init__(self, inp, params=None):
      t = inp.get('type')
      if t is None or t != 'SEGMENT':
        raise ValueError("global segment type must be 'SEGMENT'")
      self.type = t
      self.segments = [DatasetRequest.DatasetSegment(s, params) for s in inp.get('segments', [])]
      self.groups = {}
      for s in self.segments:
        g = s.getgroup()
        if not self.groups.get(g):
          self.groups[g] = []
        self.groups[g].append(s)

    @staticmethod
    def hasEntities(s):
      if len(s.segment.data.get('entities')):
        return True
      else:
        return False

    def applyGroupToSb(self, group, sb):
      # cannot mix entity and non entity segments in same group!
      # cannot mix negeated and non-negated segments in same group!
      ents = False
      applied = []
      ids = []
      for i, g in enumerate(group):
        if ids.__contains__(g.id):
          index = applied.index(g.id)
          if applied[index].negate != g.negate:
            LOGGER.debug('Error: Cannot mix negated and non negated segments')
            raise ValueError('Error: Cannot mix negated and non negated segments')
        if self.hasEntities(g):  # entity segment
          if not ents and len(ids):  # list contains non-entity segments
            if not getattr(settings, 'DJANGO_TEST_ENV', False):
              LOGGER.debug('Error: Cannot mix segments with entities and without')
            raise ValueError('Error: Cannot mix segments with entities and without')
          elif (not ents and not len(ids)) or ents:  # adding entity segment in list
            ents = True
            ids.append(g.id)
            applied.append(g)
        else:  # non-entity segment
          if ents:  # list contains entity segments
            if not getattr(settings, 'DJANGO_TEST_ENV', False):
              LOGGER.debug('Error: Cannot mix segments with entities and without')
            raise ValueError('Error: Cannot mix segments with entities and without')
          else:
            ids.append(g.id)
            applied.append(g)
      if not ents:
        filters = ""
        for i, s in enumerate(applied):
          filters += s.getexpr()
          if i < len(applied) - 1:
            filters += " OR "
        if len(applied) > 1:
          filters = "(" + filters + ")"
        sb.addFilter(filters)
      else:
        for s in applied:
          ent = s.segment.data.get('entities')
          s.addEntityToSb(ent, sb)

    def addToSb(self, sb):
      for group in list(self.groups.keys()):
        self.applyGroupToSb(self.groups[group], sb)

  class FilterAssociationSegment(object):
    def __init__(self, inp, params=None):
      self.segments = DatasetRequest.DatasetGlobalSegment(inp, params=params)


  class DatasetDimension(object):
    def __init__(self, inp):
      t = inp.get('type', 'SIMPLE')
      if t not in ['SIMPLE', 'SEGMENT']:
        raise Exception("Unknown dimension type: " + t)
      self.type = t
      self.expr = removeNewlines(inp.get('expr'))
      self.label = inp.get('label')
      self.alias = inp.get('alias')
      self.segments = [DatasetRequest.DatasetSegment(s) for s in inp.get('segments', [])]
      o = inp.get('order')
      self.order = None if o is None else DatasetRequest.DatasetOrder(o)

  class DatasetAggregate(object):
    def __init__(self, inp):
      self.expr = removeNewlines(inp.get('expr'))
      self.label = inp.get('label')
      o = inp.get('order')
      self.order = None if o is None else DatasetRequest.DatasetOrder(o)

  class DatasetEvent(object):
    def __init__(self, inp):
      t = inp.get('type')
      if t is not None and t not in ['flow', 'funnel']:
        raise ValueError('data_api error: event type can be either flow or funnel')
      self.type = t
      self.events = inp.get('events')


class Request(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.dataset_id = data.get('dataset_id', None)
    self.type = data.get('type', None)
    self.aggregates = [DatasetRequest.DatasetAggregate(a) for a in data.get('aggregates', [])]
    self.dimensions = [DatasetRequest.DatasetDimension(d) for d in data.get('dimensions', [])]
    self.partitions = [DatasetRequest.DatasetDimension(d) for d in data.get('partitions', [])]
    # derived data are passed in json dsreq in order in which they are built
    # view hierarchy is based on derived data ordering
    self.derived_data = [DatasetRequest.DatasetDerivedData(d) for d in data.get('derived_data', []) if d is not None]
    self.global_segments = [DatasetRequest.DatasetGlobalSegment(s) for s in data.get('global_segments', [])]
    self.filters = data.get('filters', [])
    self.outerFilters = data.get('outerFilters', [])
    self.having = data.get('having', [])
    self.limit = data.get('limit', None)
    self.nulls_order = data.get('nulls_order', "Default")
    self.distinct = data.get('distinct', None)
    self.query = None
    self.ignore_alias = False
    self.ksqlStreamPosition = data.get('ksqlStreamPosition', '')
    self.dsreqRewriter = data.get('dsreqRewriter')

class MRrequest(Request):
  def __init__(self, data):
    super(MRrequest, self).__init__(data)
    self.request_type = 'MR'
    self.time = data.get('time', [])
    self.session = data.get('session', [])
    self.events = [DatasetRequest.DatasetEvent(e) for e in data.get('events', [])]
    if not len(self.time) or not len(self.session) or not len(self.events):
      raise ValueError('MRrequest requires "time", "session" and "events" fields')


class SBrequest(Request):
  def __init__(self, data):
    super(SBrequest, self).__init__(data)
    self.request_type = NORM_CMDS['SQL']

class ShowAViewRequest(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.dataset_id = None
    self.request_type = '.show_aviews'
    self.query = '.show_aviews'

class RefreshAViewRequest(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.dataset_id = None
    self.request_type = '.refresh_aviews'
    self.query = '.refresh_aviews'
    self.viewname = data.get('table_name', 'unknown')

class DropAViewRequest(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.dataset_id = None
    self.request_type = SPECIAL_CMDS['DROP_AVIEWS']
    self.query = '.drop_aviews'
    self.viewname = data.get('table_name', 'unknown')

class InvalidateMetadataRequest(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.dataset_id = data.get('dataset_id', None)
    self.table_name = data.get('table_name', None)
    self.request_type = '.invalidate_metadata'
    self.query = '.invalidate_metadata'

class ComputeStatsRequest(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.dataset_id = data.get('dataset_id', None)
    self.table_name = data.get('table_name', None)
    self.request_type = '.compute_stats'
    self.query = '.compute_stats'

class TArequest(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.dataset_id = None
    self.request_type = '.tables'
    self.dbname = data.get('dbname', '')
    self.query = '.tables %s' % (self.dbname)

class DBrequest(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.dataset_id = None
    self.request_type = '.databases'
    self.query = '.databases'


class CLrequest(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.dataset_id = data.get('dataset_id', None)
    self.request_type = 'columnstats'
    self.query = '%s'

class RowsRequest(Request):
  def __init__(self, data):
    super(RowsRequest, self).__init__(data)
    self.request_type = 'getrows'
    self.table_name = data.get('table_name', None)
    self.limit = data.get('limit', 10)
    self.query = ''

class SaveTableRequest(Request):
  def __init__(self, data):
    super(SaveTableRequest, self).__init__(data)
    self.request_type = NORM_CMDS['SAVE_TABLE']
    self.table_name = data.get('table_name', None)
    self.materialized = data.get('materialized', None)
    self.materializedType = data.get('materializedType', None)
    self.table_format = data.get('table_format', None)
    self.query = data.get('query', None)
    self.limit = data.get('limit', None)
    self.ignore_alias = True

class CreateAViewRequest(Request):
  def __init__(self, data):
    super(CreateAViewRequest, self).__init__(data)
    self.request_type = NORM_CMDS['CREATEAVIEW']
    self.table_name = data.get('table_name', None)
    self.materialized = data.get('materialized', None)
    self.table_format = data.get('table_format', None)
    self.query = data.get('query', None)
    self.use_assoc_lv = data.get('use_assoc_lv', True)
    self.view_location = data.get('view_location', None)
    # Request.__init__ is setting up a default limit. Override it here
    self.limit = data.get('limit', None)
    self.ignore_alias = True

class ValidatorRequest(Request):
  def __init__(self, data):
    super(ValidatorRequest, self).__init__(data)
    self.request_type = '.validate'
    self.limit = 0
    self.expr = removeNewlines(data.get('validate', None))
    self.query = ''
    self.alias = data.get('alias', None)


class DataImportRequest(object):
  def __init__(self, data):
    self.version = data.get('version', None)
    self.request_type = '.data_import'
    self.connection_id = data.get('connection_id', None)
    self.dataset_id = None
    self.table_name = data.get('table_name', None)
    self.dbname = data.get('dbname', None)
    self.limit = data.get('limit', 1000)


class HttpRequest(object):
  def __init__(self, data):
    self.version = data.get('version')
    self.data = data
    self.query = None

class SolrRequest(object):
  def __init__(self, data):
    self.version = data.get('version')
    self.connection_id = data.get('connection_id')
    self.dataset_id = data.get('dataset_id', None)
    self.aggregates = [DatasetRequest.DatasetAggregate(a) for a in data.get('aggregates', [])]
    self.dimensions = [DatasetRequest.DatasetDimension(d) for d in data.get('dimensions', [])]
    self.filters = data.get('filters', [])
    self.limit = data.get('limit')
    self.highlighting = data.get('highlighting')
    self.solr_query = data.get('solr_query', ['*:*'])
    self.distinct = data.get('distinct', None)
    self.request_type = NORM_CMDS['SQL']
    self.query = None
    self.autotime_override = data.get('autotime_override', 'AUTO')

def removeNewlines(s):
  if s is None:
    return None
  return s.replace('\n', ' ')
