# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from vizlib.dataset import COLUMN_RE
import re

_ALIAS_RE = re.compile(r'^(.*?) as ([\'\"].+[\'\"])$', re.DOTALL | re.I)
_SIMPLE_FN_RE = re.compile(r'\s*([a-z]+)\s*\(\s*(.+)\s*\)', re.DOTALL | re.I)
_DISTINCT_RE = re.compile(r'distinct\s+(.+)\s*', re.DOTALL | re.I)

_SUM_RE = re.compile(r'^(?P<pre>.*)(sum)\s*\(\s*(?P<arg>\(*[^\(\)]+\)*?)\s*\)(?P<post>.*)$', re.DOTALL | re.I)
_SUM_REMAINING_VALID = re.compile(r'^[\s\(\)\+\-\/\*\d]*$')

#
# Given an expression like "[channel] as 'channel'", this function returns "[channel]"
#   i.e. it strips out the alias from the expression
#
def stripAlias(s):
  return Expression(s).baseExpr()

def unbracketColumn(s, pre='[', post=']'):
  if not s:
    return s

  sStripped = s.strip()
  if sStripped and pre and post and sStripped[0] == pre and sStripped[-1] == post:
    return unbracketColumn(sStripped[1:len(s)-1], pre=pre, post=post)

  return s

class Expression:
  def __init__(self, expr):
    self._expr = expr.strip()
    self._originalExpr = self._expr
    self._computeComponents()

  @classmethod
  def findColumns(cls, expressionStr):
    m = re.match(COLUMN_RE, expressionStr)
    while m:
      yield m.group(2)
      expressionStr = m.group(3)
      m = re.match(COLUMN_RE, expressionStr)

  @classmethod
  def buildSimpleFnStr(cls, fnName, fnArg):
    return f'{fnName}({fnArg})'

  # Apply the aggregation in 'aggExprStr' to 'alias'. Return the resulting expression string
  @classmethod
  def applyAggToAlias(cls, aggExprStr, alias):
    expr = Expression(aggExprStr)
    if not expr.simpleFn():
      return alias

    expr.setSimpleFnArg(f'[{alias}]')
    return expr.baseExpr()

  def simpleFnComponents(self):
    return (self.simpleFn(), self.simpleFnArg())

  def components(self):
    return {
      'baseExpr': self._baseExpr,
      'alias': self._alias,
      'simpleFn': self._simpleFn,
      'simpleFnArg': self._simpleFnArg,
    }

  def fullExpressionStr(self):
    return self._expr

  def baseExpr(self):
    return self._baseExpr

  def setBaseExpr(self, baseExpr):
    self._baseExpr = baseExpr
    self._computeExpr()

  def alias(self):
    return self._alias

  def setAlias(self, alias):
    self._alias = alias
    self._computeExpr()

  def simpleFn(self):
    return self._simpleFn

  def simpleFnArg(self):
    return self._simpleFnArg

  def setSimpleFnArg(self, col):
    self._simpleFnArg = col
    self._baseExpr = Expression.buildSimpleFnStr(self._simpleFn, self._simpleFnArg)
    self._computeExpr()

  def sumsOnly(self):
    sumCount = 0
    parts = [self._baseExpr]

    while True:
      newParts = []
      for part in parts:
        if type(part) == Expression:
          newParts.append(part)
          continue

        m = _SUM_RE.search(part)
        if not m:
          newParts.append(part)
          continue

        sumCount += 1
        newParts.extend([m.group('pre'),
                         Expression(Expression.buildSimpleFnStr('sum', m.group('arg'))),
                         m.group('post')])

      if len(parts) == len(newParts):
        break

      parts = newParts

    if not sumCount:
      return []

    for part in parts:
      if type(part) == Expression:
        continue

      m = _SUM_REMAINING_VALID.match(part)
      if not m:
        return []

    return parts

  def _computeExpr(self):
    if self._alias:
      self._expr = f'{self._baseExpr} as {self._alias}'
      return

    self._expr = self._baseExpr

  def _computeComponents(self):
    self._computeAlias()
    self._computeSimpleFn()
    self._computeCountDistinct()

  def _computeAlias(self):
    m = _ALIAS_RE.match(self._expr)
    if m:
      self._baseExpr = m.group(1)
      self._alias = m.group(2)
      return

    self._baseExpr = self._expr
    self._alias = None

  def _computeSimpleFn(self):
    m = _SIMPLE_FN_RE.match(self._baseExpr)
    if m:
      self._simpleFn = m.group(1)
      self._simpleFnArg = m.group(2)
      return

    self._simpleFn = None
    self._simpleFnArg = None

  def _computeCountDistinct(self):
    if self._simpleFn != 'count':
      self._isCountDistinct = False
      self._countDistinctArg = None
      return

    m = _DISTINCT_RE.match(self._simpleFnArg)
    if m:
      self._isCountDistinct = True
      self._countDistinctArg = m.group(1)
      return

    self._isCountDistinct = False
    self._countDistinctArg = None

  def isCountDistinct(self):
    return self._isCountDistinct

  def countDistinctArg(self):
    return self._countDistinctArg

  def setCountDistinctArg(self, col):
    self.setSimpleFnArg(f'distinct {col}')

