# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import time
import threading
import logging

LOGGER = logging.getLogger("arcweb")

# Reclaim handles after we track 1K handles
_MAXSIZE_BEFORE_CHECK_RECLAIM = 1024

# Reclaim handles after 10 minutes without a client working on it
_HANDLE_RECLAIM_TIME = 10 * 60

class Handles(object):
  def __init__(self):
    self.lock = threading.Lock()
    self.handles = {}
    self._cleanup_count = 0
    self._handle_reclaim_time = _HANDLE_RECLAIM_TIME

  def set_handle_reclaim_time(self, reclaim_time):
    self._handle_reclaim_time = reclaim_time

  def handles_count(self):
    with self.lock:
      return len(self.handles)

  def get_keys(self):
    with self.lock:
      return list(self.handles.keys())

  def get_values(self):
    with self.lock:
      return list(self.handles.values())

  def set_data(self, handle, data):
    do_cleanup = False
    with self.lock:
      if data is None:
        if handle in self.handles:
          del self.handles[handle]
      else:
        self.handles[handle] = (time.time(), data)
      if len(self.handles) > _MAXSIZE_BEFORE_CHECK_RECLAIM:
        do_cleanup = True

    if do_cleanup:
      self.cleanup_handles()

  def get_data(self, handle):
    with self.lock:
      d = self.handles.get(handle)
      if d:
        (tm, data) = d

        # Update the last time we were asked about this handle
        self.handles[handle] = (time.time(), data)

        return data
      return None

  def cleanup_count(self):
    """Used in unit tests"""
    return self._cleanup_count

  # Called to cleanup stale handles
  def cleanup_handles(self):
    self._cleanup_count += 1
    with self.lock:
      to_delete = []
      # First, iterate over all elements, and pick the ones to reclaim
      for k, v in self.handles.items():
        tm, v = v
        if abs(time.time() - tm) >= self._handle_reclaim_time:
          to_delete.append(k)

      # Then, delete the ones that have not been tracked for
      # self._handle_reclaim_time
      for k in to_delete:
        # LOGGER.info('Deleting query handle %s', k)
        del self.handles[k]

