# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
#
# This module helps track long-running tasks. For each long-running task
# in the server, create a LongRunningTask object. Specify progress with
# the .currstep() and .totalsteps() function. Specify completion with the
# .done() function.
#

import datetime
import threading

class TaskType:
  # Task Types
  REFRESH_AVIEW = 1

_TASKS = []
_TASKS_LOCK = threading.Lock()

def getTasks():
  with _TASKS_LOCK:
    d = [t.getState() for t in _TASKS]
  return d

class LongRunningTask(object):
  def __init__(self, tasktype=None, taskname=None, connectionid=None,
      taskid=None, totalsteps=1, cancelcb=None):
    self._tasktype = tasktype
    self._taskname = taskname
    self._taskid = taskid
    self._starttime = datetime.datetime.utcnow()
    self._endtime = None
    self._currstep = 0
    self._totalsteps = totalsteps
    self._connectionid = connectionid
    self._cancelcb = cancelcb
    self._error = ''
    with _TASKS_LOCK:
      _TASKS.append(self)

  def currstep(self, currstep):
    self._currstep = currstep

  def totalstaps(self, totalstaps):
    self._totalstaps = totalstaps

  def done(self, error=None):
    self._endtime = datetime.datetime.utcnow()
    self._currstep = self._totalsteps
    self._error = error
    td = self._endtime - self._starttime
    ts = (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6
    self.runtime = "%.3f" % ts
    with _TASKS_LOCK:
      _TASKS.remove(self)

  def getState(self):
    if self._endtime:
      endtime = self._endtime.strftime('%Y-%m-%d %H:%M:%S.%f')
    else:
      endtime = ''

    return {
      'tasktype': self._tasktype,
      'taskname': self._taskname,
      'starttime': self._starttime.strftime('%Y-%m-%d %H:%M:%S.%f'),
      'endtime': endtime,
      'currstep': self._currstep,
      'totalsteps': self._totalsteps,
      'error': self._error,
      'connectionid': self._connectionid
    }

