# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""Miscellaneous functions"""

import os
import logging

import urllib.request, urllib.parse, urllib.error
import re
from django.http import HttpResponse

from dataconn import dataconn
from . import workq
from . import handles
from vizlib.arcprof import ArcProf
from django.conf import settings
from django.core.cache import cache

import threading
import json
import time
import filelock

from apps.models import APIKey
from apps.models import lookup_apikey
from django.contrib.auth.decorators import login_required

_HANDLES = handles.Handles()

# create logger
LOGGER = logging.getLogger("arcweb")

_ASYNC_RUNNER = dict[int, workq.AsyncRunner]()
_ASYNC_RUNNER_LOCK = threading.Lock()

def get_async_runner(connectionId):
  return _ASYNC_RUNNER[connectionId]

def initAsyncConnectionThreadpool(connectionId):
  __initAsyncConnectionThreadpool(connectionId)

def __initAsyncConnectionThreadpool(connectionId):
  # In Ambari installations, we run cherrypy with daemonize=1. That leads to
  # importing this module pre-daemonizing, and then forking without threads.
  # This in turn means we cannot initialize _ASYNC_RUNNER at import time
  # and instead have to do it at a later place, protected by _ASYNC_RUNNER_LOCK

  global _ASYNC_RUNNER

  # This is the concurrency of threads that are running queries on connections
  if _ASYNC_RUNNER.get(connectionId) is None:
    with _ASYNC_RUNNER_LOCK:
      if _ASYNC_RUNNER.get(connectionId) is None:
        _ASYNC_RUNNER[connectionId] = workq.AsyncRunner(
            threadcount=dataconn.getDataConnConcurrency(connectionId),
            name=f"conn_{connectionId}.async",
            queryTimeoutSeconds=(dataconn.getDataConnQueryTimeoutMinutes(connectionId)*60),
            categoryConcurrency=dataconn.getDataConnConcurrencyUser(connectionId))

def reinitAsyncConnectionThreadpool(connectionId):
  global _ASYNC_RUNNER

  with _ASYNC_RUNNER_LOCK:
    if _ASYNC_RUNNER.get(connectionId) is None:
      LOGGER.info("DID NOT FIND %s %s", connectionId, type(connectionId))
      return

    _ASYNC_RUNNER[connectionId].setParams(
        threadcount=dataconn.getDataConnConcurrency(connectionId),
        queryTimeoutSeconds=(dataconn.getDataConnQueryTimeoutMinutes(connectionId)*60),
        categoryConcurrency=dataconn.getDataConnConcurrencyUser(connectionId))

def global_lock(name, timeout=10.0):
  lockname = os.path.join(settings.FILELOCK_PATH, name)
  return filelock.FileLock(lockname, timeout=timeout)

# Get the source IP accounting for the forwarding by caddy
def get_request_source_ip(request):
  remote_address = request.META.get('HTTP_X_FORWARDED_FOR', request.META.get('REMOTE_ADDR', ''))
  if remote_address:
    # there might be multiple addresses, separated by commas.
    # return only the first one:
    remote_address = remote_address.split(',')[0]
  return remote_address

# ------------------------------------------------------------------------------

def asyncquery_cancel(qhandle, username):
  """Cancel an async query"""

  connection_id = handle_to_connection_id(qhandle, username, 'asyncquery_cancel')

  conn_properties = dataconn.get_dataconn_properties(connection_id=connection_id)
  if conn_properties.supports_async:
    conn = dataconn.get_dataconn(connection_id=connection_id)
    try:
      return conn.async_cancel(qhandle)
    finally:
      conn.close()

  status = _ASYNC_RUNNER[connection_id].markCancelled(qhandle)
  return ''

def asyncquery_status(qhandle, username=None, profile=None):
  """Get the status of an async query"""
  connection_id = handle_to_connection_id(qhandle, username, 'asyncquery_status')
  __initAsyncConnectionThreadpool(connection_id)

  conn_properties = dataconn.get_dataconn_properties(connection_id=connection_id)
  if conn_properties.supports_async:
    conn = dataconn.get_dataconn(connection_id=connection_id)
    try:
      return conn.async_status(qhandle, username=username, profile=profile)
    finally:
      conn.close()

  runner = _ASYNC_RUNNER[connection_id].getRunner(qhandle)

  if runner:
    status = runner.status()
    if status == 'Running':
      progress, total = runner.progress()
      response = {"status": "Running", "progress": "%s" % progress, "total": "%s" % total}
      pr = runner.progress_runner()
      if hasattr(pr, 'get_chunks') and callable(pr.get_chunks):
        chunks = pr.get_chunks()

        if chunks:
          response['chunk_text'] = "".join(chunks)
          pr.reset_chunks()

      return response
    else:
      return {"status": "%s" % status}
  else:
    return {"status": "Failed to locate tracking info"}


def handle_to_connection_id(qhandle, username, f=None):
  data = _HANDLES.get_data(qhandle)
  if not data:
    LOGGER.info("%s: Could not find query handle %s" % (f, qhandle))
    raise Exception("Could not find query handle")

  if data.get('username') != username:
    LOGGER.info("username mismatch for query handle %s (%s != %s)" % \
        (qhandle, data.get('username'), username))
    raise Exception("Could not find query handle");

  return data['connection_id']

def add_handle_data(qhandle, data):
  _HANDLES.set_data(qhandle, data)

def get_handle_data(qhandle, username):
  data = _HANDLES.get_data(qhandle)
  if data.get('username') != username:
    LOGGER.info("username mismatch for query handle %s (%s != %s)" % \
        (qhandle, data.get('username'), username))
    raise Exception("Could not find query handle");

  return data

def asyncquery_results(qhandle, username=None, profile=None):
  """Get the results of an async query"""
  connection_id = handle_to_connection_id(qhandle, username, 'asyncquery_results')
  __initAsyncConnectionThreadpool(connection_id)

  conn_properties = dataconn.get_dataconn_properties(connection_id=connection_id)
  if conn_properties.supports_async:
    conn = dataconn.get_dataconn(connection_id=connection_id)
    try:
      results = conn.async_results(qhandle,
                                username=username,
                                profile=profile)
      return (results, None)
    finally:
      conn.close()

  runner = _ASYNC_RUNNER[connection_id].getRunner(qhandle)
  if runner:
    return (runner.results(), runner.profile())
  else:
    return ('Error: tracking info deleted',)

def scale_to_int(scale):
  """
  Return from a scale string to a number
  """

  if scale == 'K':
    return 1024
  if scale == 'M':
    return 1024 * 1024
  if scale == 'G':
    return 1024 * 1024 * 1024
  if scale == 'T':
    return 1024 * 1024 * 1024 * 1024
  if scale == 'E':
    return 1024 * 1024 * 1024 * 1024 * 1024

  raise Exception("unknown scale %s" % scale)

class SqlProgressRunner(workq.ProgressRunner):
  def __init__(self, connection_id=-1, query=None, comment=None,
               username=None, sample_pct=None, profile=None):
    self._connection_id = connection_id
    self._query = query
    self._comment = comment
    self._results = "Error: could not get results"
    self._completed = 0
    self._username = username
    self._sample_pct = sample_pct
    self._profile = profile or ArcProf('sqlrunner', username=username)
    self._cancelCb = None
    self._interrupted = False
    self._querytracker = self._profile.querytracker()

  def toDisplayDict(self):
    return {
      'querytracker': self._querytracker,
      'username': self._username,
      'connection': self._connection_id,
      'query': self._query,
    }

  def __str__(self):
    return json.dumps(self.toDisplayDict())

  def __repr__(self):
    return json.dumps(self.toDisplayDict())

  # Override
  def profile(self):
    return self._profile

  # Override
  def admissionControlCategory(self):
    if not self._username:
      return None

    return f"User: {self._username}"

  # Override
  def interrupt(self):
    self._interrupted = True

    if not self._cancelCb:
      LOGGER.info(f"querytracker [{self._querytracker}]. interrupt called "
          "but cancelation function is not available")
      self._results = 'Error: Query has been cancelled by user'
      return

    LOGGER.info(f"querytracker [{self._querytracker}]. interrupt called. "
        f"Canceling query {self._query}")
    try:
      self._cancelCb()
    except:
      LOGGER.exception("Error canceling query")

  def setCancelCb(self, cancelCb):
    self._cancelCb = cancelCb

  # Override
  def run(self):
    """Start running the task"""

    if self._interrupted:
      self._completed = 1
      self._results = 'Error: Query has been cancelled by user'
      return

    t1 = time.time()
    try:
      # Note that for some databases like SQLite, we have to get connection
      # objects in the thread that is using it
      self._cancelCb = None
      conn = dataconn.get_dataconn(connection_id=self._connection_id)
      try:
        self._results = conn.sync_runquery(self._query,
                                           comment=self._comment,
                                           username=self._username,
                                           sample_pct=self._sample_pct,
                                           profile=self._profile,
                                           setCancelCbFn=self.setCancelCb)
        self._completed = 1
      finally:
        conn.close()
    except Exception as e:
      error_message = str(e)
      LOGGER.exception(error_message)
      self._results = 'Error: %s' % error_message
    finally:
      LOGGER.info("querytracker [%s]. SqlProgressRunner complete. duration: %.3f",
          self._querytracker, time.time() - t1)
      self._completed = 1
      self._cancelCb = None

  def profile(self):
    return self._profile

  # Override
  def progress(self):
    """Returns a tuple that indicates (steps_completed, total_steps)"""
    return (self._completed, 1)

  # Override
  def results(self):
    """Returns results of the function"""
    return self._results

def asyncrunner_run(connection_id=-1, query=None, comment=None,
                    username=None, sample_pct=None, profile=None):

  __initAsyncConnectionThreadpool(connection_id)

  runner = SqlProgressRunner(connection_id=connection_id, query=query,
                             comment=comment,
                             username=username,
                             sample_pct=sample_pct,
                             profile=profile)
  qhandle = _ASYNC_RUNNER[connection_id].queue(runner)
  return {"qhandle": "%s" % qhandle}

def asyncrunner_connection_ids():
  with _ASYNC_RUNNER_LOCK:
    return _ASYNC_RUNNER.keys()

def asyncrunner_qt_cancel(user=None, connection_id=None, querytracker=None,
    allusers=None):

  if not user or not connection_id:
    return 0

  with _ASYNC_RUNNER_LOCK:
    asyncRunner = _ASYNC_RUNNER.get(connection_id)
    if asyncRunner is None:
      return 0

  matchUsername = user.username
  if user.is_superuser and allusers:
    matchUsername = None

  return asyncRunner.cancelQueryViaQueryTracker(
      matchUsername=matchUsername,
      querytracker=querytracker)

def asyncrunner_reset_pending_work(connection_id):
  with _ASYNC_RUNNER_LOCK:
    asyncRunner = _ASYNC_RUNNER.get(connection_id)
    if asyncRunner is None:
      return

  asyncRunner.requePendingWork()

def asyncrunner_stop_tracking(qhandle=None, username=None):
  connection_id = handle_to_connection_id(qhandle, username, 'asyncquery_stop_tracking')
  if _ASYNC_RUNNER.get(connection_id):
    _ASYNC_RUNNER[connection_id].stop_tracking(qhandle)

def check_login_or_use_apikey(request, api_key_auth_only=None):
  if api_key_auth_only is None:
    api_key_auth_only = getattr(settings, "ADMINAPI_REQUIRE_APIKEY", False)

  if api_key_auth_only or not request.user.is_authenticated:
    try:
      match = re.match(r"apikey\s+(\w+)", request.META["HTTP_AUTHORIZATION"])
      apikeystring = match.group(1)
      apikey = lookup_apikey(apikeystring)
    except:
      LOGGER.debug("Authentication failed")
      return HttpResponse('Unauthenticated user', status=401)
    if not apikey or not apikey.valid(apikeystring):
      LOGGER.debug("APIKEY authentication failed")
      return HttpResponse('Unauthenticated user', status=401)

    LOGGER.debug("apikey auth succeeded: %s" % apikey.apikey_hash[0:5])

    request.user = apikey.user
    return None

# A class to encapsulate logic to track updates to a timestamp in the cache
# Note that upon True return from is_changed(), the client MUST update the
# relevant data.
class ChangeChecker(object):
  def __init__(self, name, check_time=0):
    self.name = name
    self.last_check = 0
    self.last_timestamp = 0
    self.check_time = check_time

  # Return True if data needs to be re-loaded
  def is_changed(self):
    if self.check_time:
      now = time.time()
      if (now - self.last_check) < self.check_time:
        return False
      self.last_check = now

    timestamp = cache.get(self.name)
    if timestamp == self.last_timestamp:
      return False

    self.last_timestamp = timestamp
    return True

  # Update cache change flag to indicate data has changed
  def set_changed(self):
    cache.set(self.name, time.time())

# Run the character_code / string_position scrambling
def scramble(s, posenc=0x0D, chrenc=0x5A):
  if type(s) == type(b''):
    return ''.join(chr(s[i ^ posenc] ^ chrenc) for i in range(len(s)))
  else:
    return ''.join(chr(ord(s[i ^ posenc]) ^ chrenc) for i in range(len(s)))

# Invert the obfuscation function
def unobfuscate(s):
  return scramble(s)

# Obfuscate a string
def obfuscate(s):
  return scramble(s + ' ' * (16 - (len(s) % 16)))
