# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json, logging

from .select_builder import BaseSelectBuilder
from .dataset import ModelEvents

# create logger
LOGGER = logging.getLogger("arcweb")


class MRBuilder(object):
  """This class represent a Match Recognizer SQL builder."""

  def __init__(self, dataset, mrrequest=None):
    self.dataset = dataset
    self.sb = BaseSelectBuilder(dataset)
    self.type = None
    self.time = []
    self.session = []
    self.events = []
    self.defines = []
    self.measures = []
    self.pattern = []
    self.filters = []
    self.having = []
    self.global_segments = []
    self.aggregate = None
    self.limit = None
    if mrrequest is not None:
      from .dataset_request import MRrequest

      if type(mrrequest) is not MRrequest:
        raise ValueError('Invalid MRBuilder request type!')
      else:
        from datasets.models import DataSet

        self.aggregate = self.dataset.process_expr(mrrequest.aggregates[-1].expr,
            columns=self.sb.columns, alias=True)
        self.datasetEvents = ModelEvents(DataSet.objects.filter(id=mrrequest.dataset_id))
        self.filters = mrrequest.filters
        self.having = mrrequest.having
        self.global_segments = mrrequest.global_segments
        self.time = mrrequest.time
        self.session = mrrequest.session
        self.events = mrrequest.events
        self.limit = mrrequest.limit

  def path(self, events, datasetEvents):
    dsEvent = events[0]  # we support one event type at a time, expect only 1 element
    self.type = dsEvent.type
    self.events = dsEvent.events
    for i, e in enumerate(self.events):
      obj = datasetEvents.events.get(id=e.get('id'))
      e['name'] = obj.event_name
      data = json.loads(obj.event_data)
      e['data'] = [x.replace(']', '').replace('[', '') for x in data.get('filters', [])]
      e['display'] = data.get('display')
      self.defines.append(chr(65 + i))
      self.defines.append(' as (')
      self.defines.append(') AND ('.join(e['data']))
      self.defines.append(')')
      if i < len(self.events) - 1:
        self.defines.append(',\n\t       ')
      if self.type == 'flow':
        self.flow(i, len(self.events), e['min'], e['max'], e['display'])
      elif self.type == 'funnel':
        if i == 0:
          concats = list()
        self.funnel(i, len(self.events), e['min'], e['max'], e['display'], e['data'], concats)

  def flow(self, i, length, minimum, maximum, display):
    if minimum > maximum:
      raise ValueError('Event min range cannot be larger than max')
    self.pattern.append(chr(65 + i))
    if minimum != 1 or maximum != 1:
      self.pattern.append('{')
      self.pattern.append(str(minimum))
      self.pattern.append(',')
      self.pattern.append(str(maximum))
      self.pattern.append('} ')
    elif i < length - 1:
      self.pattern.append(' ')
    if i == 0:
      self.measures.append('CONCAT_WS(\',\' , ')
    if display[0] == '\'' and display[-1] == '\'':
      self.measures.append(display)
    else:
      self.measures.append('GROUP_CONCAT(')
      self.measures.append(chr(65 + i))
      self.measures.append('@')
      self.measures.append(display)
      self.measures.append(', \',\')')
    if i < length - 1:
      self.measures.append(', ')
    else:
      self.measures.append(') as path ')

  @staticmethod
  def concat_expr(i, display):
    tmp = list()
    tmp.append('concat_ws(\'\', \'ARC:FUNNEL:\', ')
    if display[0] == '\'' and display[-1] == '\'':
      tmp.append(display)
      tmp.append(')')
    else:
      tmp.append(chr(65 + i))
      tmp.append('@')
      tmp.append(display)
      tmp.append(')')
    return ''.join(tmp)

  def funnel(self, i, length, minimum, maximum, display, data, concats):
    if minimum > maximum:
      raise ValueError('Event min range cannot be larger than max')
    partition_column = self.dataset.process_expr(self.session[0],
        columns=self.sb.columns, alias=True)
    prefix = ' concat_ws(\',\', '
    if i == 0:
      self.pattern.append('^ZA*? A+ ')
      self.measures.append(' case ')
      concats.append(self.concat_expr(i, display))
    elif i <= length - 1:
      self.pattern.append('Z')
      self.pattern.append(chr(65 + i))
      self.pattern.append('* ')
      self.pattern.append(chr(65 + i))
      self.measures.append('\n\t          when (')
      self.measures.append(chr(65 + i))
      self.measures.append('@')
      self.measures.append(partition_column)
      self.measures.append(' is null) then ')
      self.measures.extend(prefix)
      self.measures.extend(', '.join(concats[0:i]))
      self.measures.append(', \'Drop Off\' ')
      concats.append(self.concat_expr(i, display))
      self.measures.append(')')
      if i == length - 1:
        self.pattern.append('*')
        self.measures.append('\n\t          else ')
        self.measures.extend(prefix)
        self.measures.extend(', '.join(concats))
        self.measures.append(') end as path, ')
        self.measures.append('\n\t          case\n\t          when (')
        self.measures.append(chr(65 + i))
        self.measures.append('@')
        self.measures.append(partition_column)
        self.measures.append(' is null) then 2 else 1 end as arc_path_priority')
      else:
        self.pattern.append('* ')
    if i == length - 1:
      self.defines.append(',\n\t       ')
    self.defines.append('Z')
    self.defines.append(chr(65 + i))
    self.defines.append(' as (!((')
    self.defines.append(') AND ('.join(data))
    self.defines.append(')))')
    if i != length - 1:
      self.defines.append(',\n\t       ')

  def addFilters(self, filters):
    self.filters = filters

  def topK(self, limit):
    self.limit = limit

  def addDefine(self):
    for i, e in enumerate(self.events):
      self.defines.append(chr(i).upper())
      self.defines.append(' as ')
      self.defines.append(' AND '.join(e.data.filters))

  def sqlPrefix(self):
    sb = BaseSelectBuilder(self.dataset)
    self.dataset.addAllToSb(sb)
    list(map(lambda x: sb.addFilter(x), self.filters))
    list(map(lambda x: x.addToSb(sb), self.global_segments))
    # list join is faster than string concatenation
    if self.type == 'flow':
      prefix = list("SELECT * FROM \n(\n SELECT path, ")
    elif self.type == 'funnel':
      prefix = list("SELECT * FROM \n(\n SELECT path, arc_path_priority, ")
    else:
      raise ValueError('MR type can either be flow or funnel')
    # last aggregate is used as a path_frequency, table alias is mr_tmp_1
    prefix.append(self.aggregate)
    prefix.append(' AS path_frequency')
    prefix.append('\n FROM MATCH_RECOGNIZE')
    prefix.append('\n\t(')
    prefix.append(sb.getsql())
    prefix.append('\n\t) ')
    prefix.append('\n\tINPUT (\n\tPARTITION BY ')
    prefix.append(', '.join(self.dataset.process_expr(s, alias=True) for s in self.session))
    prefix.append('\n\tORDER BY ')
    prefix.append(', '.join(self.dataset.process_expr(s, alias=True) for s in self.time))
    prefix.append('\n\tMEASURES ')
    prefix.extend(self.measures)
    prefix.append('\n\tONE_ROW_PER_MATCH ')
    prefix.append('\n\tPATTERN (\'')
    prefix.extend(''.join(self.pattern))
    prefix.append('\')')
    prefix.append('\n\tDEFINE ')
    prefix.extend(self.defines)
    prefix.append('\n\t) mr_tmp_1 GROUP BY 1')
    if self.type == 'funnel':
      prefix.append(', 2')
    prefix.append('\n) mr_tmp_2')
    prefix.append('\nORDER BY path_frequency DESC')
    if self.limit:
      prefix.append('\nLIMIT ')
      prefix.append(str(self.limit))
    return ''.join(prefix)

  def build(self, ds_req=None):
    self.path(self.events, self.datasetEvents)
    return self.sqlPrefix()
