# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
import logging, re
import os
from dataconn.data_api_cmds import SPECIAL_CMDS
from urllib.parse import quote

# create logger
LOGGER = logging.getLogger("arcweb")

FILTER_RE = re.compile('\[(.+?)\]\s+(.+?)\s+\((.+?)\)')
TIMERANGE_RE = re.compile('\[(.+?)\]\s+between\s+[\"\'](\S+)[\"\'] and [\"\'](\S+)[\"\']')
NUMRANGE_RE = re.compile('\[(.+?)\]\s*([!=<>]+?)\s*([\d\.]+)\s*')

SOLR_DATE_RE = re.compile('\d+-\d+-\d+T\d+:\d+:\d+.?\d*Z')

ALWAYS_SEND_JSON_QUERY = os.environ.get('SOLR_ALWAYS_JSON','0') == '1'

class NoSqlBuilder(object):
  """
  This class represents a noSQL builder.
  """

  def __init__(self, dataset, ds_req=None, properties=None):
    self.dataset = dataset
    self.properties = dataset.properties if properties is None else properties
    self.ds_req = ds_req

  def addDimension(self, dim_expr):
    raise Exception('addDimension not implemented')

  def addAggregate(self, agg_expr):
    raise Exception('addAggregate not implemented')

  def build(self, dsreq=None):
    raise Exception('buildQuery not implemented')

  def addFilter(self, fil_expr):
    raise Exception('addFilter not implemented')


class SolrBuilder(NoSqlBuilder):
  def __init__(self, dataset, ds_req=None, properties=None):
    super(SolrBuilder, self).__init__(dataset, ds_req=ds_req, properties=properties)
    self.collection = ''.join(dataset.tables[0].tablename.split('.')[1:]) if dataset else None
    self.dimensions = []
    self.aggregates = []
    self.filters = []
    self.order = []
    self.solr_url = None
    self.limit = ds_req.limit if hasattr(ds_req, 'limit') else 10000
    self.solr_query = ds_req.solr_query if hasattr(ds_req, 'solr_query') else None
    self.highlighting = ds_req.highlighting if hasattr(ds_req, 'highlighting') else False
    self.isSegmented = False
    self.autotime_override = ds_req.autotime_override if hasattr(ds_req, 'autotime_override') else None
    self.dsuhash = dataset.dsuhash if dataset and dataset.dsuhash else {}

    if hasattr(ds_req, 'dimensions'):
      for idx, d in enumerate(ds_req.dimensions):
        self.addDimension(d.expr, d.order, d.type, idx)
    if hasattr(ds_req, 'aggregates'):
      for idx, a in enumerate(ds_req.aggregates):
        if SolrBuilder._extract_aggregation_type(a.expr) == 'agg':
          self.addDimension(a.expr, a.order, 'SIMPLE', idx)
        else:
          self.addAggregate(a.expr, a.order)
    if hasattr(ds_req, 'filters'):
      for f in ds_req.filters:
        self.addFilter(f)
    self.distinct = ds_req.distinct if hasattr(ds_req, 'distinct') else None

  #TODO: Improve this to use regexes
  #count(*) is same as sum(1)
  @staticmethod
  def _extract_aggregation_type(expr):
    agg_type = None
    right = expr.find("(")
    if right > 0:
      agg_type = expr[0:right]
    if agg_type == 'count':
      if expr[right+1:right+10] == 'distinct ' or \
         expr[right+1:right+10] == 'distinct(':
        agg_type = 'countDistinct'
    elif agg_type == 'sum':
      if expr[right+1:right+3] == '1)':
        agg_type = 'recordcount'
    elif agg_type == 'avg':
       agg_type = 'mean'
    return agg_type

  @staticmethod
  def _remove_bracket(expr):
    left = expr.find("[")
    right = expr.find("]")
    if left > -1 and right > 0:
      expr = expr[left + 1:right]
    return expr

  def addDimension(self, dim_expr, order=None, dimType=None, dimIdx = None):
    if dimType == 'SEGMENT':
      self.isSegmented = True
      self.segmentIndex = dimIdx
      return

    dim_expr = SolrBuilder._remove_bracket(dim_expr)
    dim_expr = self._param_to_colname(dim_expr)
    # a dimension is treated as a facet in SOLR
    self.dimensions.append(dim_expr)
    if order:
      self.order.append((dim_expr, order))

  def addAggregate(self, agg_expr, order=None):
    agg_type = SolrBuilder._extract_aggregation_type(agg_expr)
    agg_field = SolrBuilder._remove_bracket(agg_expr)
    agg_field = self._param_to_colname(agg_field)
    self.aggregates.append((agg_field, agg_type))
    if order:
      self.order.append((agg_field, order))

  def _param_to_colname(self, key):
    # This is because when we send parameters, we send them all as lower case,
    # which enters into our fq parameter and breaks it. We get around this in
    # the SQL queries by doing aliasing, but we can't do aliasing for the fields
    # in a solr collection. Instead, we the lowercase dict to lookup the name
    # of the column in solr
    column = self.dsuhash.get(key)
    name = None
    basecol = None
    if column:
      name = column.attr['name'] if 'name' in column.attr else None
      basecol = column.attr['basecol'] if 'basecol' in column.attr else None

    return name or basecol or key

  def addFilter(self, fil_expr):
    # Converts SQL picklist filter expressions into a solr expression.
    # For dates, adding the expression onto fq doesn't work, so we need to detect
    # if dates are being used and then add the resulting expression to the solr
    # search query instead.
    m = re.match(FILTER_RE, fil_expr)
    if m and len(m.groups()) == 3:
      match_is_date = False
      key = m.group(1)
      vals_string = m.group(3)
      vals = vals_string.split(',')
      values = []
      for val in vals:
        val = val.strip()
        if val[0] == "'" or val[0] == '"':
          val = val[1:]
        if val[-1] == "'" or val[-1] == '"':
          val = val[:-1]
        if re.match(SOLR_DATE_RE, val):
          match_is_date = True
        values.append(val if val[-1] == '*' else '\"' + val + '\"')

      value = "(" + ' OR '.join(values) + ")" if len(values) > 1 else values[0]
      operator = m.group(2)
      negate = "!" if 'not' in str(operator) else ""

      if operator.lower() == 'knn':
        if self.solr_query is None:
          self.solr_query = []

        vector_column_name = self._param_to_colname(key)
        knn_query = '{!knn f=%s topK=%s}[%s]'%(vector_column_name, self.limit, vals_string)
        self.solr_query.append(knn_query)
        return
      else:
        param_to_add = negate + self._param_to_colname(key) + ":" + value

      if match_is_date:
        if self.solr_query:
          self.solr_query.append(param_to_add)
        else:
          self.solr_query = [param_to_add]
      else:
        self.filters.append(param_to_add)
      return

    # Converts SQL numeric range filter expressions into a solr expression.
    # i.e. converts ([colname] >= value) to colname:[value TO *] or
    # ([colname] >= value1) AND ([colname] <= value2) to colname:[value1 TO value2]
    m = re.findall(NUMRANGE_RE, fil_expr)
    if m:
      cols_to_filt = {}
      for match in m:
        col_name = self._param_to_colname(match[0].strip())

        if col_name not in cols_to_filt:
          cols_to_filt[col_name] = {
            'min_val': '*',
            'min_bracket': '[',
            'max_val': '*',
            'max_bracket': ']',
            'equals': None,
            'no_equals': []
          }

        s = match[1].strip()
        if s.find('>') >= 0 and s.find('<') == -1 and s.find('!') == -1:
          cols_to_filt[col_name]['min_val'] = match[2]
          cols_to_filt[col_name]['min_bracket'] = '{' if s.find('=') == -1 else '['
        elif s.find('<') >= 0 and s.find('>') == -1 and s.find('!') == -1:
          cols_to_filt[col_name]['max_val'] = match[2]
          cols_to_filt[col_name]['max_bracket'] = '}' if s.find('=') == -1 else ']'
        elif s.find('=') >= 0 and s.find('<') == -1 and s.find('>') == -1:
          if s.find('!') != -1 and s.find('!') < s.find('='):
            cols_to_filt[col_name]['no_equals'].append(match[2])
          else:
            cols_to_filt[col_name]['equals'] = match[2]

      for col in list(cols_to_filt.keys()):
        params = cols_to_filt[col]
        self.filters.append(col + ':' + params['min_bracket'] + params['min_val'] +
          ' TO ' + params['max_val'] + params['max_bracket'])

        if params['equals']:
          self.filters.append(col + ':' + params['equals'])

        no_equals = params['no_equals']
        if len(no_equals) > 0:
          val = '(' + ' OR '.join(no_equals) + ')' if len(no_equals) > 1 else no_equals[0]
          self.filters.append('-' + col + ':' + val)

    # Converts SQL date range filter expressions into solr range filters.
    # i.e. [colname] between date1 and date2 gets converted to colname:[date1 TO date2]
    m = re.match(TIMERANGE_RE, fil_expr)
    if m:
      col_name = self._param_to_colname(m.group(1))
      self.filters.append(col_name + ":[" + m.group(2) + ' TO ' + m.group(3) + ']')
      return

  def _kvsToUrl(self, kvs):
    return '&'.join([list(x.keys())[0] + "=" + str(list(x.values())[0]) for x in kvs])

  def _baseFilters(self, kvs, urlencode=False):
    if self.isSegmented:
      return
    if urlencode:
      kvs.append({'q': quote(' AND '.join(self.solr_query) if self.solr_query else '*:*', safe='*:')})
    else:
      kvs.append({'q': ' AND '.join(self.solr_query) if self.solr_query else '*:*'})


  def _facetFilters(self, kvs):
    if self.filters:
      for fil in self.filters:
        kvs.append({'fq': fil})

  def _isFacetingQuery(self):
    return True if self.aggregates else False

  def params(self, as_url=True):
    kvs = list()
    # kvs.append({'indent': 'on'})
    kvs.append({'wt': self.properties.property_map['wt']})
    self._baseFilters(kvs, urlencode=as_url)

    if self._isFacetingQuery():
      # If aggregates exist, this is a faceted search.
      # Use the dimensions as facets

      # If aggregation other than record count, use stats to get
      # sum/min/max/avg/countDistinct/stddev/missing/sumOfSquares(field)
      kvs.append({'rows': 0})
      kvs.append({'stats':'true'})
      agg_dict = {}
      for field, agg in self.aggregates:
        if not re.match("sum\(1\)\sas\s'.+'", field):
          if field in agg_dict:
            agg_dict[field] = agg_dict[field] + [agg]
          else:
            agg_dict[field] = [agg]

      for field, agg_list in agg_dict.items():
        agg_str = ''
        for agg in agg_list:
          agg_str = f'{agg_str} {agg}=true'
        kvs.append({'stats.field': f'{{!tag=piv {agg_str} }}{field}'})
      if self.dimensions:
        kvs.append({'rows': 0})
        kvs.append({'facet': 'on'})
        dims_str = ','.join(self.dimensions)
        kvs.append({'facet.pivot': f'{{!stats=piv}}{dims_str}'})
        kvs.append({'facet.limit': self.limit})
      if self.order:
        order = self.order[0][1]
        if order.ascending:
          kvs.append({'facet.sort': 'count,asc'})
      else:
        kvs.append({'rows': self.limit})
    else:
      # If no aggregates exist, dimensions can be used to only return a subset
      # of fields
      if self.dimensions:
        for d in self.dimensions:
          kvs.append({'fl': d})
        if self.order:
          sort_order = sorted(self.order, key=lambda tup: tup[1].priority)
          sort_value = []
          for o in sort_order:
            sort_value.append(o[0] + (' asc' if o[1].ascending else ' desc'))
          kvs.append({'sort': ','.join(sort_value)})
      if self.highlighting:
        kvs.append({'fl': 'id'})
        kvs.append({'hl':'true'})
        kvs.append({'hl.simple.pre':'<mark>'})
        kvs.append({'hl.simple.post':'</mark>'})
        kvs.append({'hl.fl':'*'})

      kvs.append({'rows': self.limit})

    self._facetFilters(kvs)
    if as_url:
      return self._kvsToUrl(kvs)
    else:
      result = {}
      for kv in kvs:
        for key, value in kv.items():
          if not key in result:
            result[key] = value
          elif key == 'rows':
            # rows is sometimes set multiple times
            # however, it can have only a single value.
            # let the last value win:
            result[key] = value
          else:
            existing_val = result[key]
            if type(existing_val) == list:
              existing_val.append(value)
            else:
              existing_val = [existing_val, value]
              result[key] = existing_val

      return result

  def build(self, dsreq=None):
    if dsreq.query and (dsreq.query == SPECIAL_CMDS['SHOW_DATABASES'] or
                          dsreq.query.startswith(SPECIAL_CMDS['SHOW_TABLES'])):
      return dsreq.query

    if dsreq.request_type == 'getrows':
      if dsreq.table_name:
        collectionName = dsreq.table_name.split('.')[1]
      elif self.collection:
        collectionName = self.collection
      else:
        raise AssertionError("no tablename found")
      self.limit = 100
      return self.properties.property_map['baseurl'] + collectionName + '/select?' + self.params()

    if self.solr_query is not None and len(self.solr_query) == 1:
      q = self.solr_query[0]
      if q:
        if q.startswith('columnstats'):
          c = q[len('columnstats') + 1:]
          return 'columnstats %s' % c
        elif q.startswith('computeautotime'):
          c = q[len('computeautotime') + 1:]
          return 'computeautotime %s' % c

    if self.aggregates and self.dimensions:
      if len(self.dimensions) == 1:
        dimName = self.dimensions[0]
        c = self.dataset.getColumn(dimName)
        if c:
          if c.attr.get('type') == 'timestamp' and self.autotime_override:
            kvs = list()
            self._baseFilters(kvs)
            self._facetFilters(kvs)
            return 'autotime %s' % self._kvsToUrl(kvs)

    simple_req = self.properties.property_map['baseurl'] + str(self.collection) + '/select?' + self.params()
    if ALWAYS_SEND_JSON_QUERY or quote('!knn ') in simple_req:
      req = {
        'url': self.properties.property_map['baseurl'] + str(self.collection) + '/select',
        'body': {
          'params': self.params(as_url=False)
        }
      }
      return 'json:'+json.dumps(req)
    else:
      return simple_req

