# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging

# create logger
LOGGER = logging.getLogger("arcweb")

from .dataset import ModelDataset
from .dataset import getDatasetInfo
from .filter_association import UserFilter
from datasets.models import DataSet
from .select_builder import BaseSelectBuilder, HiveSelectBuilder, AnalyticalViewBuilder
from .select_builder import BaseSelectBuilder, HiveSelectBuilder, AnalyticalViewBuilder, OracleSelectBuilder, \
    CassandraSelectBuilder, SqliteSelectBuilder
from .nosql_builder import SolrBuilder
from .mr_builder import MRBuilder
from .stats_builder import StatsBuilder
from dataconn.data_api_cmds import SPECIAL_CMDS, NORM_CMDS, VALIDATE_CMDS
from dataconn.connections_types import DATACONN_HIVE, DATACONN_MONGODB, \
    DATACONN_SPARKSQL, DATACONN_DRILL, DATACONN_ORACLE, DATACONN_CASSANDRA, \
    DATACONN_HYVE, DATACONN_SQLITE

from django.conf import settings

class QueryBuilder(object):
  """This class is used for building a query from a data_api JSON object (dsreq)"""

  def __init__(self, dataset_request, connection, qcntl=None):
    self.dataset_request = dataset_request
    self.dstype = self.dataset_request.dstype
    self.qcntl = qcntl
    self.builder = None
    self.connection = connection
    self.connection_properties = connection.properties()

  def buildQuery(self, username, userid, query, sample_pct=None):
    """Build query using a Django request"""

    sbrequest = self.dataset_request.dsreq
    dataset_id = sbrequest.dataset_id
    dsreqRewriter = getattr(sbrequest, 'dsreqRewriter', None)
    params = self.dataset_request.params
    ds = self.dataset_request.dataset  # Dataset

    # Lookup the FA here so we can apply to derived data queries
    user_filter = None
    if dataset_id is not None:
      user_filter = UserFilter(userid, dataset_id)

    if ds is None and dataset_id is not None:
      django_ds = self.dataset_request.django_ds

      if not django_ds:
        dl = DataSet.objects.filter(id=dataset_id)
        if len(dl) == 0:
          msg = "Failed to load dataset %s" % dataset_id
          LOGGER.error(msg)
          raise Exception(msg)
        django_ds = dl[0]

      # Get a handle to the data connection related to this dataset
      dc = django_ds.dataconnection
      if not dc:
        from datasets.utils import get_default_dataconnection
        # Deal with older datasets that we are migrating over
        dc = get_default_dataconnection()

      # dataset_info is not always present in the Dataset Django model
      if hasattr(sbrequest, 'use_assoc_lv') and not sbrequest.use_assoc_lv:
        dataset_info = django_ds.dataset_info
      else:
        dataset_info = django_ds.dataset_lvinfo or django_ds.dataset_info
      if not dataset_info:
        dataset_info = getDatasetInfo(django_ds.dataset_detail, dc.id, username)
        ds = ModelDataset(dataset_info, self.connection_properties,
            dsreqRewriter=dsreqRewriter)
        # Save the info so that we don't have to build it each time.
        django_ds.dataset_info = dataset_info
        django_ds.save()
      else:
        ds = ModelDataset(dataset_info, self.connection_properties, params=params,
            dsreqRewriter=dsreqRewriter)

      self.dataset_request.dataset = ds

    if str(self.connection.connection_type).upper() == NORM_CMDS['SOLR']:
      self.builder = SolrBuilder(ds, sbrequest, properties = self.connection.properties())
    elif self.dstype == NORM_CMDS['SQL']:
      try:
        bParams = {'properties':self.connection_properties, 'user_filter':user_filter, 'qcntl':self.qcntl}
        if self.connection.connection_type in [DATACONN_HIVE, DATACONN_HYVE, DATACONN_MONGODB, \
              DATACONN_SPARKSQL, DATACONN_DRILL]:
          self.builder = HiveSelectBuilder(ds, sbrequest, **bParams)
        elif self.connection.connection_type is DATACONN_ORACLE:
          self.builder = OracleSelectBuilder(ds, sbrequest, **bParams)
        elif self.connection.connection_type is DATACONN_CASSANDRA:
          self.builder = CassandraSelectBuilder(ds, sbrequest, **bParams)
        elif self.connection.connection_type is DATACONN_SQLITE:
          self.builder = SqliteSelectBuilder(ds, sbrequest, **bParams)
        else:
          self.builder = BaseSelectBuilder(ds, sbrequest, **bParams)
        self.builder.sample_pct = sample_pct
      except ValueError as e:
        if not getattr(settings, 'DJANGO_TEST_ENV', False):
          LOGGER.exception(e)
        raise ValueError(e)
    elif self.dstype in [NORM_CMDS['CREATEAVIEW'], NORM_CMDS['SAVE_TABLE']]:
      self.builder = AnalyticalViewBuilder(ds, sbrequest,
          properties = self.connection_properties,
          user=username)
    elif self.dstype == NORM_CMDS['MR']:
      try:
        self.builder = MRBuilder(ds, sbrequest)
      except ValueError as e:
        LOGGER.error(e)
        raise ValueError(e)
    elif self.dstype in [SPECIAL_CMDS['SHOW_TABLES'], SPECIAL_CMDS['SHOW_AVIEWS'],
        SPECIAL_CMDS['SHOW_DATABASES'], NORM_CMDS['COLUMN_STATS'],
        NORM_CMDS['getrows'], SPECIAL_CMDS['DATA_IMPORT'], VALIDATE_CMDS['join'],
        SPECIAL_CMDS['REFRESH_AVIEWS'], SPECIAL_CMDS['DROP_AVIEWS'],
        SPECIAL_CMDS['INVALIDATE_METADATA'], SPECIAL_CMDS['COMPUTE_STATS']]:
      if self.dstype == VALIDATE_CMDS['join']:
        prefix = self.dstype
      else:
        prefix = None
      try:
        self.builder = StatsBuilder(ds, sbrequest, query,
            conn_type=self.connection.connection_type,
            properties=self.connection_properties,
            prefix=prefix)
      except ValueError as e:
        LOGGER.error(e)
        raise ValueError(e)
    elif self.dstype in list(VALIDATE_CMDS.values()):
      try:
        # Because of python module dependencies, we have to include this here,
        # and not at the top of the file
        from .validator_builder import ExpressionValidator
        self.builder = ExpressionValidator(self.dstype, ds, conn_type=self.connection.connection_type)
      except ValueError as e:
        LOGGER.error(e)
        raise ValueError(e)

    if self.builder is None:
      LOGGER.error('could not instantiate builder')
      raise ValueError('could not instantiate builder')

    # Apply filter associations
    if user_filter is not None:
      user_filter.applyFA(self.builder, self.qcntl)

    sql = self.builder.build(sbrequest)
    return sql

