# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import time
import datetime
import uuid
import os
import os.path
import logging
import json
import requests
import threading
from requests.auth import HTTPBasicAuth

from util.safethreadpool import DjangoSafeThreadPoolExecutor

try:
  from collections import OrderedDict
except ImportError:
  from ordereddict import OrderedDict

import vizlib.handles

from arcwebbase import access_info
from django.conf import settings

from util.perfcounters import APPROX_COUNTERS
from vizlib.misc import get_request_source_ip

from .querylog_helper import QueryState, ComponentType, ActionType, ABANDONED_TIME

from usage.views_helper import to_usage_logentry_model
from usage.models import LogEntry

LOGGER = logging.getLogger("arcweb")

_TRACKED_QUERIES = vizlib.handles.Handles()

accessTrackerLogsTotal = APPROX_COUNTERS.get('accesstracker.total')
accessTrackerLogsError = APPROX_COUNTERS.get('accesstracker.error')
accessTrackerLogsIgnored = APPROX_COUNTERS.get('accesstracker.ignored')
accessTrackerLogsQueued = APPROX_COUNTERS.get('accesstracker.queued')

def mark_queries_abandoned():
  _TRACKED_QUERIES.cleanup_handles()

  for data in _TRACKED_QUERIES.get_values():
    (tm, log) = data

    if log.query_state in [QueryState.STARTED, QueryState.QUEUED] and \
       ((time.time() - tm) > ABANDONED_TIME):
      log.logQueryAbandoned()


def mark_query_canceled(qhandle):
  log = _TRACKED_QUERIES.get_data(qhandle)
  if log is None:
    return

  if log.query_state in [QueryState.STARTED, QueryState.QUEUED, QueryState.ABANDONED]:
    log.logQueryCanceled()


def read_ongoing_queries():
  data = []
  mark_queries_abandoned()

  for d in _TRACKED_QUERIES.get_values():
    (_, v) = d
    if v.queryComplete():
      continue
    d = v.fetchLogJsonRaw()
    data.append(d)

  data.sort(key=lambda x: x['ts'], reverse=True)
  return data

def auditLogExists():
  return settings.AUDIT_LOGFILE is not None

# list of functions that might redact the queriy strings
REDACTORS = []

def redact_query(query: str):
  for redactor in REDACTORS:
    redacted, redacted_query = redactor(query)
    if redacted:
      return redacted_query

  # no redactor modified the query
  return query

class LogReader(object):
  def __init__(self, filename=None, fileindex=None):
    if filename:
      self.filename = filename
    else:
      self.filename = settings.AUDIT_LOGFILE
      if fileindex is not None:
        self.filename = "%s.%s" % (self.filename, fileindex)

  def _is_json_start(self, data):
    return data.startswith("{")

  def file_exists(self):
    if not self.filename:
      return False

    return os.path.isfile(self.filename)

  def read_json_backwards(self):
    if not self.filename:
      return

    if not os.path.isfile(self.filename):
      return

    with open(self.filename) as fh:
      collected_lines = []
      for line in self._reversed_lines(fh):
        if not self._is_json_start(line):
          collected_lines.append(line)
          continue

        # We know 'line' is the start of a json structure
        if collected_lines:
          collected_lines.append(line)
          collected_lines.reverse()
          data = "\n".join(collected_lines)
          collected_lines = []
        else:
          # Typical case: a single line has the entire json structure
          # Optimize for it
          data = line
        try:
          yield json.loads(data)
        except Exception:
          pass

  def markStartedQueryCanceled(self, j):
    s = j.get('query_state')
    if s == QueryState.STARTED:
      j['query_state'] = QueryState.CANCELED

  def _reversed_lines(self, fh):
    "Generate the lines of file in reverse order."
    part = ''
    for block in self._reversed_blocks(fh):
      end = len(block)
      while end > 0:
        pos = block.rfind('\n', 0, end)
        if pos != -1:
          yield block[pos + 1:end] + part
          part = ''
        else:
          part = block[0:end] + part
        end = pos
    if part:
      yield part

  def _reversed_blocks(self, fh, blocksize=4096):
    "Generate blocks of file's contents in reverse order."
    fh.seek(0, os.SEEK_END)
    here = fh.tell()
    while 0 < here:
      delta = min(blocksize, here)
      here -= delta
      fh.seek(here, os.SEEK_SET)
      yield fh.read(delta)

#
# Access tracker for dashboards & applications. Note that these updates are attempted,
#   but not guaranteed.
#
# This class manages a threadpool (self._execPool) which it uses to post the actual updates
#   so as to not slow down any sync processing.
#
class AccessTracker():
  ACCESS_TRACKER_COMPONENTS = set([ComponentType.APPGROUP, ComponentType.APP])

  def __init__(self):
    self._lock = threading.Lock()
    self._execPool = None

  def track(self, log):
    from reports.views_settings import setting_get_single

    global accessTrackerLogsIgnored, accessTrackerLogsError

    if not self._execPool:
      with self._lock:
        if not self._execPool:
          self._execPool = DjangoSafeThreadPoolExecutor(name="tracker",
              max_workers=1)

    accessTrackerEnabled = setting_get_single("access_tracker_enabled") == "true"
    if not accessTrackerEnabled:
      accessTrackerLogsIgnored += 1
      return

    try:
      self._track_internal(log)
    except Exception as e:
      accessTrackerLogsError += 1
      LOGGER.exception(e)

  def _track_internal(self, log):
    from reports.models import Report, AppGroup
    from reports.views_settings import setting_get_single
    import reports.views_appgroups

    global accessTrackerLogsIgnored, accessTrackerLogsQueued

    componentType = log.component_type
    if componentType not in AccessTracker.ACCESS_TRACKER_COMPONENTS:
      accessTrackerLogsIgnored += 1
      LOGGER.info("Not tracking component %s", componentType)
      return

    trackerUrl = setting_get_single('access_tracker_url')
    if not trackerUrl:
      accessTrackerLogsIgnored += 1
      LOGGER.info("Not tracking because no good url")
      return

    sourceSystem = setting_get_single('access_tracker_source_id')
    sourceUrl = setting_get_single('access_tracker_source_url')

    action = log.action
    componentId = log.component_id
    username = log.username

    delete = (action == ActionType.DELETE)

    if componentType == ComponentType.APPGROUP:
      objectType = "dashboard"
      viewUrl = f"{sourceUrl}/arc/apps/appgroup/{componentId}"
      editUrl = f"{sourceUrl}/arc/apps/appedit/{componentId}"
      ag = AppGroup.objects.get(id=componentId)
      title = ag.group_name
      subtitle = None
      thumbnailUrl = f"{sourceUrl}/arc/apps/snapshot_ag/{componentId}"
    else:
      objectType = "dashboard"
      viewUrl = f"{sourceUrl}/arc/apps/app/{componentId}"
      editUrl = f"{sourceUrl}/arc/apps/builder/{componentId}"
      thumbnailUrl = f"{sourceUrl}/arc/apps/snapshot/{componentId}"
      report = Report.objects.get(id=componentId)
      j = json.loads(report.report_data)
      title = j.get('report_title', '')
      subtitle = j.get('report_subtitle', '')

    data = {
      "entity": {
        "typeName": "c1cstudio_object",
        "attributes": {
          "userName": username,
          "description": title,
          "displayName": title,
          "editUrl": editUrl,
          "viewUrl": viewUrl,
          "userDescription": subtitle or title,
          "name": title,
          "objectId": f"{componentId}",
          "objectType": objectType,
          "qualifiedName": f"dataviz:{sourceSystem}:{objectType}:{componentId}",
          "sourceSystem": sourceSystem,
          "lastAccessTime": datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
          "thumbnail": {
            "imageUrl": thumbnailUrl,
            "icon": {
              "iconType": "font-awesome",
              "iconSource": "chart-bar"
            },
            "title": title,
          },
        },
      },
    }

    accessTrackerLogsQueued += 1
    self._execPool.submit(self._sendLog, data)

  def _sendLog(self, data):
    from reports.views_settings import setting_get_single
    from jobs.views_encryption import decrypt_secret

    global accessTrackerLogsTotal, accessTrackerLogsError

    trackerUrl = setting_get_single('access_tracker_url')
    trackerUser = setting_get_single('access_tracker_user')
    trackerPassword = decrypt_secret(**json.loads(setting_get_single(
        'access_tracker_password')))

    accessTrackerLogsTotal += 1
    try:
      if trackerUser:
        x = requests.post(trackerUrl, json=data,
                auth=HTTPBasicAuth(trackerUser, trackerPassword))
      else:
        x = requests.post(trackerUrl, json=data)
      if x.status_code != 200:
        LOGGER.info("Error tracking URL: %s %s", x, x.text)
        accessTrackerLogsError += 1
    except Exception as e:
      accessTrackerLogsError += 1
      LOGGER.exception(e)
      pass

ACCESS_TRACKER = AccessTracker()

def logAction(**args):
  if not settings.AUDIT_LOGGER:
    return

  _populate_access_info_fields(args)
  log = _ActionLog(**args)
  ACCESS_TRACKER.track(log)

  log.logJson()

def _populate_access_info_fields(args: dict):
  '''
  If the provided dictionary is missing access information, then
  try to fill them from the thread local storage
  '''
  ai = access_info.get()
  args['username'] = args.get('username') or ai.username
  args['user_id'] = args.get('user_id') or ai.user_id
  args['client_ip'] = args.get('client_ip') or ai.client_ip

def _populate_default_query_fields(args):
  if not 'action' in args:
    args['action'] = ActionType.READ
  if not 'component_type' in args:
    args['component_type'] = ComponentType.APP
  if not 'component_id' in args:
    args['component_id'] = args.get('visual_id') or args.get('app_id') or ''

class QueryLog(object):
  def __init__(self, **args):
    if not settings.AUDIT_LOGGER:
      self._ql = _NullQueryLog()
      return
    _populate_default_query_fields(args)
    _populate_access_info_fields(args)
    self._ql = _ActionLog(**args)
    self._ql.logQueryStart()

  def __enter__(self):
    return self._ql

  def __exit__(self, etype, value, tb):
    if etype:
      self._ql.logQueryEnd(error=True, error_detail=str(value))
    else:
      self._ql.logQueryEnd()

  def isNull(self):
    return False

def querylog_create(**args):
  if settings.AUDIT_LOGGER:
    _populate_default_query_fields(args)
    _populate_access_info_fields(args)
    return _ActionLog(**args)
  return _NullQueryLog()

def actionlog_create(**args):
  if settings.AUDIT_LOGGER:
    _populate_access_info_fields(args)
    return _ActionLog(**args)
  return _NullQueryLog()

def querylog_fetch(qhandle):
  ql = _TRACKED_QUERIES.get_data(qhandle)
  if ql is None:
    return _NullQueryLog()
  return ql

class _ActionLog:
  def __init__(self, username='', visual_id='', app_id='',
      dataset_id='', dataconnection_id='', dataconnection_type='',
      dataconnection_fbid='', appgroup_id='',
      query='', sync=False, client_ip='', action='',
      component_type='', component_id='', error_detail='', query_state=0,
      visual_type='', visual_name='', dataset_name='', dataconnection_name='',
      appgroup_name='', app_name='', user_id='', rewrite_target='', request=None):
    self.action_id = str(uuid.uuid4())
    self.username = username
    self.visual_id = visual_id
    self.app_id = app_id
    self.appgroup_id = appgroup_id
    self.dataset_id = dataset_id
    self.dataconnection_id = dataconnection_id
    self.dataconnection_fbid = dataconnection_fbid
    self.dataconnection_type = dataconnection_type
    self.served_from_cache = False
    self.served_from_cache_after_data_check = False
    self.query_state = query_state
    self.query = query
    self.rewrite = ''
    self.runtime = ''
    self.start_time = datetime.datetime.utcnow()
    self.ts = self.start_time
    self.sync = sync
    self.client_ip = client_ip
    self.rewrite_target = rewrite_target
    self.action = action
    self.component_type = component_type
    self.component_id = component_id
    self.query_error = False
    self.error_detail = error_detail
    self.process_id = settings.PERSISTENT_PROCESS_ID
    self.visual_type = visual_type
    self.visual_name = visual_name
    self.dataset_name = dataset_name
    self.dataconnection_name = dataconnection_name
    self.appgroup_name = appgroup_name
    self.app_name = app_name
    self.user_id = user_id
    if request:
      if not self.username:
        self.username = request.user.username
      if not self.user_id:
        self.user_id = request.user.id
      if not self.client_ip:
        self.client_ip = get_request_source_ip(request)

  def usage_track(self, log_json):
    logEntry = to_usage_logentry_model(log_json)
    LogEntry.new_entries.append(logEntry)

  def track(self, qhandle):
    _TRACKED_QUERIES.set_data(qhandle, self)

  def untrack(self, qhandle):
    _TRACKED_QUERIES.set_data(qhandle, None)

  def isNull(self):
    return False

  def _csv_sanitize_query(self, query):
    q = query.replace("\n", " ")
    q = q.replace('"', '""')
    q = '"' + q + '"'
    return q

  def logQueryQueued(self):
    self.query_state = QueryState.QUEUED
    self.ts = datetime.datetime.utcnow()
    self.logJson()

  def logQueryStart(self):
    self.query_state = QueryState.STARTED
    self.ts = datetime.datetime.utcnow()
    self.logJson()

  def logQueryAbandoned(self):
    self.query_state = QueryState.ABANDONED
    self.ts = datetime.datetime.utcnow()
    self.logJson()

  def logQueryCanceled(self):
    self.query_state = QueryState.CANCELED
    self.ts = datetime.datetime.utcnow()
    self.logJson()

  def logQueryEnd(self, error=False, cache=False, cache_check=False, rewrite='', error_detail='',
                  dataconnection_fbid=''):
    self.query_state = QueryState.COMPLETE
    self.ts = datetime.datetime.utcnow()
    td = self.ts - self.start_time
    ts = float((td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6)) / 10**6
    self.runtime = "%.6f" % ts
    self.rewrite = rewrite
    self.untrack(self.action_id)
    if cache:
      self.served_from_cache = True
    if cache_check:
      self.served_from_cache = True
      self.served_from_cache_after_data_check = True
    if error:
      self.query_error = True
      self.error_detail = error_detail
    if dataconnection_fbid:
      self.dataconnection_fbid = dataconnection_fbid
    self.logJson()

  def logJson(self):
    log_json = self.fetchLogJsonRaw()
    self.usage_track(log_json)
    settings.AUDIT_LOGGER.info('', extra={'audit_record':log_json})

  def queryComplete(self):
    return self.query_state == QueryState.COMPLETE

  def fetchLogJsonRaw(self):
    return OrderedDict([
          ('ts', self.ts.strftime('%Y-%m-%d %H:%M:%S.%f')),
          ('action', self.action),
          ('action_name', ActionType.getStr(self.action)),
          ('component_type', self.component_type),
          ('component_type_name', ComponentType.getStr(self.component_type)),
          ('component_id', self.component_id),
          ('action_id', self.action_id),
          ('query_state', self.query_state),
          ('query_state_name', QueryState.getStr(self.query_state)),
          ('username', self.username),
          ('ip', self.client_ip),
          ('visual', self.visual_id),
          ('app', self.app_id),
          ('appgroup', self.appgroup_id),
          ('dataset', self.dataset_id),
          ('dataconnection', self.dataconnection_id),
          ('dataconnection_fbid', self.dataconnection_fbid),
          ('dataconnection_type', self.dataconnection_type),
          ('sync', self.sync),
          ('cache', self.served_from_cache),
          ('cache_check', self.served_from_cache_after_data_check),
          ('runtime', self.runtime),
          ('rewrite_target', self.rewrite_target),
          ('query', redact_query(self.query)),
          ('rewrite', self.rewrite),
          ('error', self.query_error),
          ('error_detail', self.error_detail),
          ('process_id', self.process_id),
          ('visual_type', self.visual_type),
          ('visual_name', self.visual_name),
          ('dataset_name', self.dataset_name),
          ('dataconnection_name', self.dataconnection_name),
          ('appgroup_name', self.appgroup_name),
          ('dashboard_name', self.app_name),
          ('user_id', self.user_id),
        ])

class _NullQueryLog(_ActionLog):

  def isNull(self):
    return True

  def track(self, qhandle):
    pass

  def untrack(self, qhandle):
    pass

  def logQueryQueued(self):
    pass

  def logQueryStart(self):
    pass

  def logQueryEnd(self, **args):
    pass