# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging, re

from django.conf import settings
from dataconn.connections_types import DATACONN_HIVE, DATACONN_HYVE
from vizlib.dataset import ModelDataset
from functools import cmp_to_key

from vizlib.cviz_remotecall import cviz_rest_expr
import vizlib.expression

DD_USE_SUBQUERY = getattr(settings, 'DD_USE_SUBQUERY', False)

# create logger
LOGGER = logging.getLogger("arcweb")

SEGMENT_RE = re.compile(r'(.*?) as Segment$', re.DOTALL)

class BaseSelectBuilder(object):
  """
  This class represents a SQL select builder.
     Dimensions (self.dimensions) are grouped by and are either simple columns
       or segment names.
     Aggregates (self.aggregates) are aggregate expressions and are not grouped
       by.
     Filters (self.filters) are WHERE clauses.
     Outer Filters (self.outerFilters) lead to the base SQL being put in a
       subselect and then where clauses being applied to that subselect
     Having (self.having) are HAVING clauses.
     Order (self.order) represents ORDER BY columns
     Limit (self.limit) represents a LIMIT clause
     Tables (self.tables) are the set of tables in the FROM clause
  """

  def __init__(self, dataset, sbrequest=None, properties=None, user_filter=None, qcntl=None):
    if properties is None:
      properties = dataset.properties

    self.dataset = dataset
    self.user_filter = user_filter
    self.qcntl = qcntl
    self.dimensions = []
    self.partitions = []
    self.aggregates = []
    self.joins = []
    self.limit = None
    self.order = []
    self.tables = set()
    self.columns = set()
    self.complex_columns = set()
    self.filters = []
    self.outerFilters = []
    self.filter_columns = []
    self.having = []
    self.indent = "\n  "
    self.entity = False
    self.distinct = None
    self.entity_alias = 10
    self.ignore_alias = False

    self.dimension_exclude_list = list()
    self.viewcounter = 0
    self.derived_data = {}
    self.nulls_order = "Default"
    self.hidden_columns = True
    self.properties = properties
    self.sample_pct = None
    self.conn_type = dataset.properties.conn_type

    self.sanitize_alias = False
    if not properties.alias_supports_spaces:
      self.sanitize_alias = True

    if properties.alias_supports_backticks:
      self.alias_backtick = properties.quote_alias
    else:
      self.alias_backtick = ""

    if sbrequest is not None:
      from .dataset_request import SBrequest, RowsRequest, ValidatorRequest, \
                                  CreateAViewRequest, SaveTableRequest

      if type(sbrequest) is not SBrequest \
          and type(sbrequest) is not RowsRequest \
          and type(sbrequest) is not CreateAViewRequest \
          and type(sbrequest) is not SaveTableRequest \
          and type(sbrequest) is not ValidatorRequest:
        raise ValueError('Invalid SelectBuilder request type!')
      else:
        for d in sbrequest.derived_data:
          # create views from ordered dictionary
          view = self.createView(view_label=d.label, view_dsreq=d.dsreq)
          if view is not None:
            self.derived_data[view.table.alias] = view
        for d in sbrequest.dimensions:
          self.addDimension(d.expr or 'SEGMENT', d.order)
        for d in sbrequest.partitions:
          self.addPartition(d.expr, d.order)
        for a in sbrequest.aggregates:
          self.addAggregate(a.expr, a.order)
        for f in sbrequest.filters:
          self.addFilter(f)
        for f in sbrequest.outerFilters:
          self.addOuterFilter(f)
        for f in sbrequest.having:
          self.addHaving(f)
        for gg in sbrequest.global_segments:
          gg.addToSb(self)
        self.limit = sbrequest.limit
        self.distinct = sbrequest.distinct
        self.ignore_alias = sbrequest.ignore_alias
        self.nulls_order = sbrequest.nulls_order

  def isEmpty(self):
    if not len(self.dimensions) and not len(self.aggregates) \
        and not len(self.partitions):
      return True
    else:
      return False

  def isMissingTables(self):
    if not self.tables and not self.dataset.views:
      return True
    else:
      return False

  def clone(self):
    bParams = {'properties':self.properties, 'user_filter':self.user_filter, 'qcntl':self.qcntl}
    if type(self) is HiveSelectBuilder:
      newsb = HiveSelectBuilder(self.dataset, **bParams)
    elif type(self) is OracleSelectBuilder:
      newsb = OracleSelectBuilder(self.dataset, **bParams)
    elif type(self) is CassandraSelectBuilder:
      newsb = CassandraSelectBuilder(self.dataset, **bParams)
    else:
      newsb = BaseSelectBuilder(self.dataset, **bParams)
    newsb.dimensions = self.dimensions[:]
    newsb.partitions = self.partitions[:]
    newsb.aggregates = self.aggregates[:]
    newsb.joins = self.joins[:]
    newsb.limit = self.limit
    newsb.order = self.order[:]
    newsb.tables = self.tables.copy()
    newsb.columns = self.columns.copy()
    newsb.complex_columns = self.complex_columns.copy()
    newsb.filters = self.filters[:]
    newsb.having = self.having[:]
    newsb.indent = self.indent
    newsb.entity = self.entity
    newsb.distinct = self.distinct
    newsb.entity_alias = self.entity_alias
    newsb.dimension_exclude_list = self.dimension_exclude_list[:]
    newsb.alias_backtick = self.alias_backtick
    newsb.viewcounter = self.viewcounter
    newsb.derived_data = self.derived_data
    newsb.hidden_columns = self.hidden_columns
    return newsb

  def createView(self, view_label, view_dsreq, sbrequest=None):
    """
    :param view_label: label of view to be created
    :param view_dsreq: dsreq json object required to build view
    :param sbrequest: select builder request object built out of dsreq
    :return: returns create View
    """
    if sbrequest is None:
      from .dataset_request import SBrequest  # TODO add MRrequest
      sbrequest = SBrequest(view_dsreq)

    # Note that the js side also has a notion of DD parent/child which we should probably
    # move here through some sort of validation.

    # Special case is expression only where we have a single dim/agg.  In this
    # case no table is created, just a new column which we glue to the base table
    if len(sbrequest.aggregates) + len(sbrequest.dimensions) == 1 and \
       len(sbrequest.filters) == 0:

      if len(sbrequest.aggregates) == 1:
        isagg = True
        expr = sbrequest.aggregates[0].expr
      else:
        isagg = False
        expr = sbrequest.dimensions[0].expr

      slist = self.selectlist([expr])
      col = self.dataset.Column(self.dataset.tables[0],
                                {'alias': self.dequote_alias(slist[1][0]),
                                 'expr': slist[0][0],
                                 'exprhasagg': isagg})
      self.dataset.addColumn(col)
      return None

    # naming matters because it affects the ordering in which views are used in the SQL expression
    # views are ordered and appear in SQL as follows T1, T2, T3, ... TN
    self.viewcounter += 1
    view_alias = "T" + str(self.viewcounter)

    join_exprs = list()
    allcols = set()

    (dims_exprs, dims_alias) = self.selectlist([d.expr for d in sbrequest.dimensions])
    dims_alias = list(map(self.dequote_alias, dims_alias))

    dims_set = set()
    for (i, d) in enumerate(sbrequest.dimensions):
      expr = self.dataset.process_expr(d.expr, columns=dims_set)

      # Don't add window function to join
      if expr is None or re.search('\s+over\s*\(', expr.lower()):
        continue

      # Use demensions for join expressions
      join_exprs.append('%s = [%s]' % (dims_exprs[i], dims_alias[i]))

    allcols.update(dims_set)

    aggs_alias = self.selectlist([a.expr for a in sbrequest.aggregates])[1]
    aggs_alias = list(map(self.dequote_alias, aggs_alias))

    aggs_set = set()
    for a in sbrequest.aggregates:
      self.dataset.process_expr(a.expr, columns=aggs_set)
    allcols.update(aggs_set)

    filt_set = set()
    for f in sbrequest.filters:
      self.dataset.process_expr(f, columns=filt_set)
    allcols.update(filt_set)

    # Build the list of dependent views based on columns named in this view's expressions
    view_dependents = set()
    for c in allcols:
      alias = c.table.alias
      view = self.derived_data.get(alias)
      if view:
        view_dependents.add(alias)
        view_dependents.update(view.view_dependents)

    # Select the parent table from the farthest table named in the dimensions list
    if dims_set:
      dimlist = sorted(list(set(c.table for c in dims_set)),
                       key=cmp_to_key(ModelDataset.cmptab))
      parent = dimlist[-1].alias
    else:
      parent = self.dataset.tables[0].alias

    if type(self) is HiveSelectBuilder:
      builder = HiveSelectBuilder(self.dataset, sbrequest=sbrequest, properties=self.properties)
    elif type(self) is OracleSelectBuilder:
      builder = OracleSelectBuilder(self.dataset, sbrequest=sbrequest, properties=self.properties)
    elif type(self) is CassandraSelectBuilder:
      builder = CassandraSelectBuilder(self.dataset, sbrequest=sbrequest, properties=self.properties)
    else:
      builder = BaseSelectBuilder(self.dataset, sbrequest=sbrequest, properties=self.properties)

    # Apply FA to the DD sub-query here
    if self.user_filter is not None:
      self.user_filter.applyFA(builder, self.qcntl)

    view_query = builder.getsql()
    sql = list()
    if DD_USE_SUBQUERY:
      sql.append(view_query)
    else:
      sql.append(view_alias)
      sql.append("AS \n(")
      sql.append(view_query)
      sql.append("\n)")

    columns = [{'name':c} for c in dims_alias + aggs_alias]
    view = self.dataset.View(dataset=self.dataset,
                             label=view_label,
                             alias=view_alias,
                             sql=' '.join(sql),
                             parent_tablename=parent,
                             join_exprs=join_exprs,
                             columns=columns,
                             view_dependents=view_dependents)

    return view

  def addDimension(self, expr, order=None):
    expr = self.dataset.process_expr(expr, self.columns)
    self.dimensions.append(expr)
    if order:
      self.order.append((len(self.dimensions), order))

  def addAggregate(self, expr, order=None, alias=False):
    expr = self.dataset.process_expr(expr, self.columns, alias=alias)
    self.aggregates.append(expr)
    if order:
      self.order.append((len(self.dimensions) + len(self.aggregates), order))

  def addPartition(self, expr, order=None):
    expr = self.dataset.process_expr(expr, self.columns)
    self.partitions.append(expr)
    if order:
      self.order.append((len(self.dimensions) + len(self.aggregates) + \
          len(self.partitions), order))

  def addFilter(self, expr, addToColumns=True):
    columns = set()
    expr = self.dataset.process_expr(expr, columns, alias=False)
    self.filters.append(expr)
    if addToColumns:
      self.columns.update(columns)

  def addFilterColumns(self, columns):
    for col in columns:
      self.filter_columns.append('%s.%s' % (col.table.tablename.lower(),
                                            col.attr.get('name').lower()))

  def addOuterFilter(self, expr):
    if expr is not None:
      self.outerFilters.append(expr)

  def addHaving(self, expr):
    expr = self.dataset.process_expr(expr, self.columns)
    self.having.append(expr)

  def addJoin(self, expr, join_type="join"):
    self.joins.append((expr, join_type))

  def addTables(self, tables):
    self.tables.update(tables)

  def prefix(self, nolimit=False):
    self.tables.update(c.table for c in self.columns)
    if self.dataset.join_cull and not self.columns:
      # If the query refers to no columns (for example sum(1)),
      # then we just add in the first table
      self.tables.update([self.dataset.tables[0]])

    # Add in the referenced DD views and their required views
    for t in self.tables:
      view = self.derived_data.get(t.alias)
      if view:
        self.dataset.views[t.alias] = view
        for v in view.view_dependents:
          self.dataset.views[v] = self.derived_data.get(v)

    if self.isEmpty():  # no dimensions or aggregates
      self.dataset.addAllToSb(self)
    elif self.isMissingTables():  # missing FROM clause
      self.tables = set(self.dataset.tables)
    if self.distinct:
      # In case dimension(s) and aggregate(s) are both present in the request
      # group by will automatically be added (see getsql relevant lines).
      # Select distinct can't be/shouldn't be use together with grouped by.
      # To avoid sending invalid query, the code returns the error earlier
      groupby_would_be_applied = self.dimensions and len(self.dimensions) \
        and self.aggregates and len(self.aggregates)
      if groupby_would_be_applied:
        raise ValueError('cannot combine SELECT DISTINCT with GROUP BY')
      sql = self.selectlist_tosql(self.dimensions + self.aggregates, distinct=True, nolimit=nolimit)
    else:
      sql = self.selectlist_tosql(self.dimensions + self.aggregates + \
          self.partitions, nolimit=nolimit)

    for c in self.columns:
      pc = c.needsParentJoin()
      if pc:
        # we add here the map and array types so we can join on them, but the columns with
        # type struct is not added to complex_columns even though it is a complex type
        self.complex_columns.add(pc)

    return sql

  def selectlist_tosql(self, mylist, distinct=False, nolimit=False):
    sellist, sellist_alias = self.selectlist(mylist)
    sql = list()
    sql.append('SELECT ')

    topk = ''
    if self.properties and self.properties.supports_top:
      if not nolimit and self.limit is not None and int(self.limit) >= 0:
        if self.properties.supports_top_brackets:
          topk = "TOP(%s) " % self.limit
          self.limit = None
        else:
          topk = "TOP %s " % self.limit
          self.limit = None

    if distinct:
      sql.append('DISTINCT ')
      sql.append(topk)
      for idx, item in enumerate(sellist):
        sql.append(item)
        if idx < len(sellist) - 1:
          sql.append(', ')
    else:
      sql.append(topk)
      for idx, item in enumerate(sellist):
        is_dim, e = cviz_rest_expr(item)
        if e:
          item = '"Remote call"'

        if self.ignore_alias or not self.properties.sb_specify_alias:
          sql.append(item)
        else:
          sql.append(item + ' as ' + sellist_alias[idx])
        if idx < len(sellist) - 1:
          sql.append(', ')
    return ''.join(sql)


  @staticmethod
  def sanitize_alias(alias):
    tmp = list(alias)
    for i, c in enumerate(tmp):
      if not c.isalpha() and not c.isdigit() and c != '_':
        tmp[i] = '_'
    return ''.join(tmp)

  def selectlist(self, mylist, index_segment=False):
    # TODO: Really the result column names (aliases) for the dimensions and aggregrates
    # should be specified from the client and available directly from the select_builder
    sellist = list()
    sellist_alias = list()
    for s in mylist:
      segment = False
      if s != 'SEGMENT':
        idx = -1
        if 'as Segment' in s:
          segment = True
          idx = s.rfind(' as ')
        if idx == -1:
          idx = s.rfind(' as \'')
        if idx == -1:
          idx = s.rfind(' as `')
        if idx == -1:
          idx = s.rfind(' as "')
        if idx == -1:
          sellist.append(str(s))
          alias = s
        else:
          sellist.append(str(s[:idx]))
          alias = s[(idx + 4):]
        if segment:
          sellist_alias.append(str(alias))
        elif self.alias_backtick:
          if self.alias_backtick in alias:
            tmp = list(str(alias).replace(self.alias_backtick, '\''))
            if tmp[0] == '\'':
              tmp[0] = self.alias_backtick
            else:
              tmp.insert(0, self.alias_backtick)
            if tmp[-1] == '\'':
              tmp.pop()
              tmp.append(self.alias_backtick)
            else:
              tmp.append(self.alias_backtick)
            sellist_alias.append(''.join(tmp))
          else:
            if alias[0] == alias[-1] and alias[0] in ["'", '"', "`"]:
              sellist_alias.append(self.alias_backtick + alias[1:-1] + self.alias_backtick)
            else:
              sellist_alias.append(self.alias_backtick + alias + self.alias_backtick)
        else:
          if alias[0] == alias[-1] and alias[0] in ["'", '"', "`"]:
            sellist_alias.append(self.enquote_alias(str(alias)[1:-1]))
          else:
            sellist_alias.append(self.enquote_alias(str(alias)))
      elif index_segment:
        sellist.append('Segment')
        sellist_alias.append('Segment')
    if self.sanitize_alias and self.conn_type not in [DATACONN_HYVE, DATACONN_HIVE]:
      for i, alias in enumerate(sellist_alias):
        sellist_alias[i] = BaseSelectBuilder.sanitize_alias(alias)

    return sellist, sellist_alias

  def enquote_alias(self, alias):
    if self.properties and not self.properties.alias_supports_spaces:
      alias = alias.replace(" ", "_")

    if self.properties:
      alias = self.properties.alias_transform(alias)

    if alias is None:
      return None

    return self.dataset.quote_alias + \
           alias.replace(self.dataset.quote_alias, self.dataset.quote_escape_character + self.dataset.quote_alias) + \
           self.dataset.quote_alias

  def dequote_alias(self, alias):
    return alias[1:-1]

  def order_clause_list(self, sellist_alias=None):
    if self.properties.supports_nulls_order and self.nulls_order != "Default":
      nulls_order = " NULLS " + str(self.nulls_order).upper()
    else:
      nulls_order = ""
    if sellist_alias:
      olist = [(o.priority, sellist_alias[i - 1] +
                (" ASC" if o.ascending else " DESC") + nulls_order) for (i, o) in self.order]
    else:
      olist = [(o.priority, str(i) +
                (" ASC" if o.ascending else " DESC") + nulls_order) for (i, o) in self.order]
    return olist

  def strip_alias(self, s):
    m = SEGMENT_RE.search(s)
    if m:
      return None

    return vizlib.expression.stripAlias(s)

  def get_knn_sql(self, nolimit=False):
    """
    Should be overridden in subclasses
    """
    return None


  def getsql(self, nolimit=False):
    sql = self.get_knn_sql(nolimit)
    if sql:
      return sql

    sql = self.prefix(nolimit=nolimit)
    if self.sample_pct not in (None, ''):
      self.dataset.sample_stmt = "TABLESAMPLE SYSTEM(%s)" % self.sample_pct

    ftxt = self.dataset.outerJoin(self.tables)
    for join in self.joins:
      ftxt += " " + join[1] + " " + join[0]
    ftxt = self.dataset.joinComplexColumns(ftxt, self.complex_columns)

    sql += self.indent + "FROM " + ftxt

    if len(self.filters):
      sql += self.indent + "WHERE (" + ") AND (".join(self.filters) + ")"

    if (self.entity and len(self.dimensions)) or \
        (len(self.aggregates) and (len(self.dimensions) or len(self.partitions))):

      # Partitions are used for CREATE ANALYTICAL VIEW statements because
      # partition columns have to be at the end of the select list. Special
      # handling for that is below
      if len(self.partitions) > 0 and len(self.dimensions) > 0:
        extra_gby = ", "
      else:
        extra_gby = ""

      if not self.properties or self.properties.supports_groupby_colindex and self.conn_type not in [DATACONN_HYVE, DATACONN_HIVE]:
        sql += self.indent + "GROUP BY " + \
            ", ".join(str(x) for x in range(1, len(self.dimensions) + 1)) + \
            extra_gby + \
            ", ".join(str(x + len(self.dimensions) + len(self.aggregates) + 1) \
                for x in range(0, len(self.partitions)))
      else:
        gby_cols = [x for x in [self.strip_alias(x) for x in self.dimensions] if x is not None]
        if len(gby_cols) > 0:
          sql += self.indent + "GROUP BY " + ", ".join(gby_cols)

    if len(self.having):
      sql += self.indent + "HAVING (" + " AND ".join(self.having) + ")"

    if not self.properties or self.properties.supports_orderby:
      if len(self.order):
        olist = self.order_clause_list()
        olist.sort()
        sql += self.indent + "ORDER BY " + ", ".join(txt for (pos, txt) in olist)

    if self.distinct and not len(self.order):
      sql += self.indent + "ORDER BY 1 "

    if self.outerFilters:
      sql = "SELECT * FROM (" + sql + ") arc_outer WHERE " + \
          " AND ".join(self.outerFilters)

    if not nolimit and self.limit is not None and int(self.limit) >= 0:
      if not self.properties or self.properties.supports_limit:
        sql += self.indent + "LIMIT " + str(self.limit)
      elif self.properties.supports_limit_sample:
        sql += self.indent + "SAMPLE " + str(self.limit)

    return sql

  def orderby_limit_sql(self):
    sql = ' '
    if len(self.order):
      if type(self) is not HiveSelectBuilder:
        olist = self.order_clause_list()
      else:
        if len(self.dimensions) + len(self.aggregates) > len(self.dimension_exclude_list):
          sellist_alias = self.selectlist(self.dimensions + self.aggregates, index_segment=True)[1]
          olist = self.order_clause_list(sellist_alias)
      olist.sort()
      sql += self.indent + "ORDER BY " + ", ".join(txt for (pos, txt) in olist)
    if self.distinct:
      sql += self.indent + "ORDER BY 1 "
    return sql

  def supports_order_outside_union(self):
    return self.properties and self.properties.supports_order_outside_union

  def reset_order_limit(self):
    self.limit = None
    self.order = []

  def limit_union(self, outside=True, nolimit=False):
    if self.properties and not self.properties.supports_limit:
      return ""

    if not outside or nolimit:
      return ""

    return (self.indent + "LIMIT " + str(self.limit)) if self.limit is not None and int(self.limit) >= 0 else ""

  def segment_data_as_dimension(self, negate, name):
    if self.properties.supports_groupby_literal:
      name = (self.properties.quote_string_constant + "not " if negate else self.properties.quote_string_constant) + \
             name + self.properties.quote_string_constant + " as Segment"
      return name

    if negate:
      return 'max(' + self.properties.quote_string_constant + "not " + name + self.properties.quote_string_constant + ') as Segment'

    return 'max(' + self.properties.quote_string_constant + name + self.properties.quote_string_constant + ') as Segment'

  def segment_as_dimension(self, s):
    return self.segment_data_as_dimension(s.negate, s.segment.data['name'])

  def build_union(self, ds_req):
    txt = []
    segidx = self.dimensions.index('SEGMENT')
    segdim = ds_req.dimensions[segidx]
    self.dimension_exclude_list.append(segidx)

    if self.supports_order_outside_union():
      orderby_clause = self.orderby_limit_sql()

    if not len(segdim.segments):
      newsb = self.clone()
      if self.supports_order_outside_union():
        newsb.reset_order_limit()
      newsb.dimensions[segidx] = self.segment_data_as_dimension(False, 'all')
      txt.append(newsb.getsql(nolimit=(True if self.properties.supports_limit_outside_union else False)))

    else:
      for s in segdim.segments:
        newsb = self.clone()
        if self.supports_order_outside_union():
          newsb.reset_order_limit()
        newsb.dimensions[segidx] = self.segment_as_dimension(s)
        s.addToSb(newsb)
        txt.append(newsb.getsql(nolimit=(True if self.properties.supports_limit_outside_union else False)))

    if type(self) is HiveSelectBuilder:
      sellist_alias = self.selectlist(self.dimensions + self.aggregates)[1]
      sel = list()
      for alias in sellist_alias:
        sel.append('TA.' + alias + ' as ' + alias)
      sql = 'SELECT Segment' + (', ' if len(sel) else '') + ', '.join(sel) + \
            self.indent + 'FROM ' + self.indent + '(' + "\nUNION ALL\n".join(txt) + ") TA"
      if self.supports_order_outside_union():
        sql = "SELECT * from (" + sql + self.indent + ") TA " + orderby_clause + self.limit_union()
      return sql + self.limit_union(outside=self.properties.supports_limit_outside_union)
    else:
      if self.supports_order_outside_union():
        cmd = "SELECT * from (" + "\nUNION ALL\n".join(txt) + ") TA " + orderby_clause + self.limit_union()
        return cmd
      sql = "\nUNION ALL\n".join(txt)
      return sql + self.limit_union(outside=self.properties.supports_limit_outside_union)

  def build(self, ds_req):
    if 'SEGMENT' not in self.dimensions:
      sql = self.getsql()
    else:
      sql = self.build_union(ds_req)

    if not self.dataset.views or DD_USE_SUBQUERY:
      return sql
    view_sql = [self.dataset.views[key].sql for key in sorted(self.dataset.views.keys())]
    return "WITH " + ',\n'.join(view_sql) + " " + sql

class AnalyticalViewBuilder(BaseSelectBuilder):
  def __init__(self, dataset, sbrequest=None, properties=None, user=""):
    self.user = user
    BaseSelectBuilder.__init__(self, dataset, sbrequest=sbrequest,
        properties=properties)

  def build(self, ds_req):
    # if materialized exist build regular view/table instead of analytical view
    table_format = ""
    partition_str = ""
    info = ""
    query = None
    if ds_req.materialized is not None:
      query = ds_req.query
      self.ignore_alias = False
      self.sanitize_alias = True
      if ds_req.materializedType == 'ksql_stream':
        table_view = "STREAM"
      elif ds_req.materializedType == 'ksql_table':
        table_view = "TABLE"
      elif ds_req.materialized:
        table_view = "TABLE"
        if self.properties.supports_tableproperties:
          info = "TBLPROPERTIES('arc_saveastableby'='%s')" % self.user
        if ds_req.table_format:
          table_format = "STORED AS %s" % ds_req.table_format
      else:
        table_view = "VIEW"
    else:
      table_view = "ANALYTICAL VIEW"
      table_format = "STORED AS PARQUET"
      location = getattr(ds_req, 'view_location', None)
      if location:
        table_format += " LOCATION '%s' " % location
      self.ignore_alias = False
      self.sanitize_alias = True
      if ds_req.partitions:
        sellist, aliases = self.selectlist([p.expr for p in ds_req.partitions])
        partition_str = "PARTITIONED BY (" + ", ".join([a.lower() for a in aliases]) + ")"
    if query is not None:
      s = query
    else:
      s = BaseSelectBuilder.build(self, ds_req)
    s = "CREATE %s %s %s %s %s AS %s" % \
        (table_view,
         self.dataset.quote_tablename(ds_req.table_name, non_alphanum=True, ignore_unavailable_tables=True),
         partition_str, table_format, info, s)
    return s

class HiveSelectBuilder(BaseSelectBuilder):
  def __init__(self, dataset, sbrequest=None, properties=None, **kwargs):
    from dataconn.dataconn_hive import HiveConnProperties

    if properties is None:
      properties = HiveConnProperties()
    super(HiveSelectBuilder, self).__init__(dataset, sbrequest=sbrequest,
                                            properties=properties, **kwargs)

  def getsql(self, nolimit=False):
    sql = self.prefix(nolimit=nolimit)

    sellist, sellist_alias = self.selectlist(self.dimensions + self.aggregates)
    if self.sample_pct not in (None, ''):
      self.dataset.sample_stmt = "TABLESAMPLE (%s PERCENT)" % self.sample_pct

    ftxt = self.dataset.outerJoin(self.tables)
    for join in self.joins:
      ftxt += " " + join[1] + " " + join[0]
    sql += self.indent + "FROM " + ftxt
    if len(self.filters):
      sql += self.indent + "WHERE (" + ") AND (".join(self.filters) + ")"
    if (self.entity and len(self.dimensions)) or (len(self.aggregates) and len(self.dimensions)):
      if len(self.dimensions) > len(self.dimension_exclude_list):
        sql += self.indent + "GROUP BY " + ", ".join(
          sellist[x] for x in range(0, len(self.dimensions)) if x not in self.dimension_exclude_list)
    if len(self.having):
      sql += self.indent + "HAVING (" + " AND ".join(self.having) + ")"
    if len(self.order):
      if len(self.dimensions) + len(self.aggregates) > len(self.dimension_exclude_list):
        if self.distinct:
          sql += self.indent + "ORDER BY 1 "
        else:
          olist = self.order_clause_list(sellist_alias)
          olist.sort()
          sql += self.indent + "ORDER BY " + ", ".join(txt for (pos, txt) in olist)
    if not nolimit and self.limit is not None and int(self.limit) >= 0:
      sql += self.indent + "LIMIT " + str(self.limit)
    return sql


class OracleSelectBuilder(BaseSelectBuilder):
  def __init__(self, dataset, sbrequest=None, properties=None, **kwargs):
    from dataconn.dataconn_oracle import OracleConnProperties

    if properties is None:
      properties = OracleConnProperties()
    super(OracleSelectBuilder, self).__init__(dataset, sbrequest=sbrequest,
                                            properties=properties, **kwargs)

  def getsql(self, nolimit=False):
    sql = self.prefix(nolimit=nolimit)

    sellist, sellist_alias = self.selectlist(self.dimensions + self.aggregates)

    ftxt = self.dataset.outerJoin(self.tables)
    for join in self.joins:
      ftxt += " " + join[1] + " " + join[0]
    sql += self.indent + "FROM " + ftxt
    if len(self.filters):
      sql += self.indent + "WHERE (" + ") AND (".join(self.filters) + ")"
    if not len(self.dimensions) and not nolimit and self.limit is not None and int(self.limit) >= 0:
      if len(self.filters):
        sql += self.indent + "AND ROWNUM <= " + str(self.limit)
      else:
        if self.distinct:
          sql = "SELECT * FROM \n(" + sql + "\n) WHERE ROWNUM <= " + str(self.limit)
        else:
          sql += self.indent + "WHERE ROWNUM <= " + str(self.limit)
    if (self.entity and len(self.dimensions)) or (len(self.aggregates) and len(self.dimensions)):
      if len(self.dimensions) > len(self.dimension_exclude_list):
        sql += self.indent + "GROUP BY " + ", ".join(
          sellist[x] for x in range(0, len(self.dimensions)) if x not in self.dimension_exclude_list)
    if len(self.having):
      sql += self.indent + "HAVING (" + "AND ".join(self.having) + ")"
    if len(self.order):
      if len(self.dimensions) + len(self.aggregates) > len(self.dimension_exclude_list):
        olist = self.order_clause_list(sellist_alias)
        olist.sort()
        sql += self.indent + "ORDER BY " + ", ".join(txt for (pos, txt) in olist)
    if len(self.dimensions) and not nolimit and self.limit is not None and int(self.limit) >= 0:
      sql = "SELECT * FROM \n(" + sql + "\n) WHERE ROWNUM <=" + str(self.limit)
    return sql

  def enquote_alias(self, alias):
    if self.properties and not self.properties.alias_supports_spaces:
      alias = alias.replace(" ", "_")

    chopped = alias if len(alias) <= 30 else alias[0:25] + "..." + alias[-1]
    return self.dataset.quote_alias + \
           chopped.replace(self.dataset.quote, self.dataset.quote_escape_character) + \
           self.dataset.quote_alias

class SqliteSelectBuilder(BaseSelectBuilder):
  def __init__(self, dataset, sbrequest=None, properties=None, **kwargs):
    super(SqliteSelectBuilder, self).__init__(dataset, sbrequest=sbrequest,
                                            properties=properties, **kwargs)

  def get_knn_sql(self, nolimit=False):
    """
    temporary hack to get KNN searches work
    Look for filters with text like 'TA_0."vector" knn (0,0,0,2)'
    """

    for f in self.filters:
      filter_parts = f.split(' ')
      if len(filter_parts) == 3 and filter_parts[1].lower() == 'knn':
        # this is a KNN search:
        json_vector=f"[{filter_parts[2][1:-1]}]"

        sql = self.prefix(nolimit=nolimit)
        ftxt = self.dataset.outerJoin(self.tables)
        for join in self.joins:
          ftxt += " " + join[1] + " " + join[0]
        ftxt = self.dataset.joinComplexColumns(ftxt, self.complex_columns)

        sql += " FROM " + ftxt
        sql += f' ORDER BY vector_distance({filter_parts[0]},"{json_vector}") ASC LIMIT {self.limit}'
        return sql

class CassandraSelectBuilder(BaseSelectBuilder):
  def __init__(self, dataset, sbrequest=None, properties=None, **kwargs):
    from dataconn.dataconn_cassandra import CassandraConnProperties

    if properties is None:
      properties = CassandraConnProperties()
    super(CassandraSelectBuilder, self).__init__(dataset, sbrequest=sbrequest,
                                            properties=properties, **kwargs)

  def getsql(self, nolimit=False):
    sql = self.prefix(nolimit=nolimit)

    sellist, sellist_alias = self.selectlist(self.dimensions + self.aggregates)
    ftxt = self.dataset.outerJoin(self.tables)
    for join in self.joins:
      ftxt += " " + join[1] + " " + join[0]
    sql += self.indent + "FROM " + ftxt
    if len(self.filters):
      sql += self.indent + "WHERE (" + ") AND (".join(self.filters) + ")"
    #if (self.entity and len(self.dimensions)) or (len(self.aggregates) and len(self.dimensions)):
    #  if len(self.dimensions) > len(self.dimension_exclude_list):
    #    sql += self.indent + "GROUP BY " + ", ".join(
    #      sellist[x] for x in range(0, len(self.dimensions)) if x not in self.dimension_exclude_list)
    if len(self.having):
      sql += self.indent + "HAVING (" + " AND ".join(self.having) + ")"
    if len(self.order):
      if len(self.dimensions) + len(self.aggregates) > len(self.dimension_exclude_list):
        olist = self.order_clause_list(sellist_alias)
        olist.sort()
        sql += self.indent + "ORDER BY " + ", ".join(txt for (pos, txt) in olist)
    if not nolimit and self.limit is not None and int(self.limit) >= 0:
      sql += self.indent + "LIMIT " + str(self.limit)
    if len(self.filters):
      sql += self.indent + "ALLOW FILTERING"
    return sql
