# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
These tests provide a base class that help test SQL backends.
"""

from django.test import TestCase, RequestFactory
from django.contrib.auth.models import User
from datasets.models import DataSet, DataConnection
from django.test.client import Client

from vizlib.query_builder import QueryBuilder
from vizlib.dataset_request import DatasetRequest
from dataconn.dataconn import get_dataconn, del_dataconn
from dataconn import arctypes

import json

# start client
c = Client()

EXPECTED_DATA_FORMAT = {
          'colnames': [],
          'coltypes': [],
          'rows': [[]],
          'rowcount': 0,
          'info': ['Query:%s' % 'select * from test'],
          '_query_warnings': {},
        }

EXPECTED_DATA_FORMAT_WITH_PARTITIONS = {'partitions': []}
EXPECTED_DATA_FORMAT_WITH_PARTITIONS.update(EXPECTED_DATA_FORMAT)

EXPECTED_DATA_LENGTH = len(EXPECTED_DATA_FORMAT)
EXPECTED_DATA_WITH_PARTITIONS_LENGTH = len(EXPECTED_DATA_FORMAT_WITH_PARTITIONS)

class MockCursor(object):
  """
  Permit mocking out a cursor object. See usage in test_dataconn_*.py.

  An instance of this class is setup as the 'cursor' in a DataEngine
  object, and then is initialized with expected SQL queries (which
  it verifies), and expected metadata and results (which it returns).

  The calling test code verifies that the returned metadata and results
  translate to appropriate returns from the DataEngine interfaces.
  """

  def __init__(self, test_obj):
    self.test_obj = test_obj
    self.descriptions = []
    self.queries = []
    self.results = []
    self.rowcount = 0
    self.execute_idx = -1

  def close(self):
    pass

  def add_description(self, description):
    self.descriptions.append(description)

  def add_query(self, query):
    self.queries.append(query)

  def add_results(self, result):
    self.results.append(result)

  def verifySQL(self, str1, str2):
    msg = '\nactual:\n' + str1 + '\ngolden:\n' + str2
    str1 = ' '.join(str1.split())  # remove all extra whitespace characters
    str2 = ' '.join(str2.split())  # remove all extra whitespace characters
    self.test_obj.assertEqual(str1, str2, msg=msg)

  def execute(self, *args, **kwargs):
    self.test_obj.assertEqual(len(args), 1)
    self.test_obj.assertEqual(len(kwargs), 0)
    q = self.queries.pop(0)
    self.verifySQL(args[0], q)
    self.execute_idx += 1

  def execute_async(self, *args, **kwargs):
    self.test_obj.assertEqual(len(args), 1)
    self.test_obj.assertEqual(len(kwargs), 0)
    q = self.queries.pop(0)
    self.verifySQL(args[0], q)
    self.execute_idx += 1

  def is_executing(self):
    return False

  def fetchmany(self, size=None):
    r = self.fetchone()
    if not r:
      return r

    return [r]

  def fetchone(self):
    r = self.results[self.execute_idx]
    if not r:
      return None
    return r.pop(0)

  def fetchall(self):
    return self.results.pop(0)

  @property
  def description(self):
    return self.descriptions[self.execute_idx]

class MockBuildRequest(object):
  def __init__(self):
    self.qcntl = None
    self.sample_pct = None
    self.comment = None

class UserAdmin(object):
  def __init__(self):
    self.username = 'admin'
    self.user = User.objects.create_superuser(self.username, '', 'arcadia')
    self.id = self.user.id
    # print 'created admin user...'

  @staticmethod
  def is_superuser(self):
    return True


class UserNonAdmin(object):
  def __init__(self, dataset_id=None, username=None, perm='readonly', filterassociation_id='None'):
    self.username = username
    if username is None:
      self.username = 'non_admin'
      self.user = User.objects.create(username=self.username, password='arcadia')
    else:
      self.user = User.objects.create(username=self.username, password='arcadia')
    self.id = self.user.id
    # print 'created user ' + self.username + '...'
    if perm not in ['readwrite', 'readonly']:
      perm = 'readonly'
    if dataset_id is not None:  # set permission for user
      url = '/datasets/userperms/' + str(dataset_id)
      post = {
              'username': self.username,
              'perm': perm,
              'filterassociation_id': filterassociation_id
             }
      response = c.post(url, data=post, follow=True)
      if response.status_code != 200:
        print(str(response))
        print(str(response.content))
        raise Exception('could not set permissions: %s %s' % (response.content, post))
      # print 'set permission ' + perm + ' to user ' + self.username + '...'
    else:
      # print 'no permissions set for user ' + self.username + '...'
      pass

  @staticmethod
  def is_superuser(self):
    return False


# since there is no real backend we cannot use client to create dataset as we must specify
# table columns in dataset_info
class CensusData(object):
  def __init__(self, dc):
    data_info = [{"tablename": "default.census_pop",
                  "columns": [{"name": "year", "type": "INT", "isdim": False, "alias": "year"},
                              {"name": "state", "type": "STRING", "isdim": True, "alias": "state"},
                              {"name": "population", "type": "BIGINT", "isdim": False, "alias": "population"}]}]
    self.dataset = DataSet(dataconnection=dc,
                           dataset_name='census',
                           dataset_info=json.dumps(data_info),
                           dataset_type='singletable',
                           dataset_detail='default.census_pop')
    self.dataset.save()
    # print 'created census dataset...'


class CensusDataCast(object):
  def __init__(self, dc):
    data_info = [{"tablename":"default.census_pop",
                  "columns":[{"name":"year","type":"INT","isdim":False,"alias":"year","dflagg":"sum","cast":"STRING"},
                             {"name":"state","type":"STRING","isdim":True,"alias":"state"},
                             {"name":"population","type":"BIGINT","isdim":False,"alias":"population"}]}]
    self.dataset = DataSet(dataconnection=dc,
                           dataset_name='census',
                           dataset_info=json.dumps(data_info),
                           dataset_type='singletable',
                           dataset_detail='default.census_pop')
    self.dataset.save()

class CensusDataCast2(object):
  def __init__(self, dc):
    data_info = [{"tablename":"default.census_pop",
                  "columns":[{"name":"year","type":"INT","isdim":False,"alias":"year","dflagg":"sum","cast":"STRING"},
                             {"name":"field_1","type":"STRING","isdim":True,"alias":"bool field", "cast":"BOOLEAN"},
                             {"name":"field_2","type":"STRING","isdim":True,"alias":"time field", "cast":"TIMESTAMP"},
                             {"name":"field_3","type":"STRING","isdim":True,"alias":"integer field", "cast":"INT"},
                             {"name":"field_4","type":"STRING","isdim":True,"alias":"float field", "cast":"REAL"}
                            ]
                }]
    self.dataset = DataSet(dataconnection=dc,
                           dataset_name='census',
                           dataset_info=json.dumps(data_info),
                           dataset_type='singletable',
                           dataset_detail='default.census_pop')
    self.dataset.save()


class CensusDataLV(object):
  def __init__(self, dc):
    data_info = [{"tablename": "default.census_pop",
                  "columns": [{"name": "year", "type": "INT", "isdim": False, "alias": "year"},
                              {"name": "state", "type": "STRING", "isdim": True, "alias": "state"},
                              {"name": "population", "type": "BIGINT", "isdim": False, "alias": "population"}]}]
    data_lvinfo = [{"lvname": "default.census_lv",
                    "columns": [{"alias": "year", "type": "INT", "name": "year", "isdim": False},
                                {"alias": "state", "type": "STRING", "name": "state", "isdim": True},
                                {"alias": "population", "type": "BIGINT", "name": "population", "isdim": False}],
                    "lvquery": "SELECT TA_0.`year` as `year`, " + \
                               "TA_0.`state` as `state`, TA_0.`population` as `population`\n  " + \
                               "FROM `default`.`census_pop` TA_0"}]
    self.dataset = DataSet(dataconnection=dc,
                           dataset_name='census',
                           dataset_info=json.dumps(data_info),
                           dataset_lvinfo=json.dumps(data_lvinfo),
                           dataset_type='singletable',
                           dataset_detail='default.census_pop')
    self.dataset.save()


class GroupTest(object):
  def __init__(self, dc):
    data_info = [{"tablename": "default.grouptest",
                  "columns": [{"name": "value", "type": "INT", "isdim": False, "alias": "value"},
                              {"name": "a", "type": "STRING", "isdim": True, "alias": "a"}]}]
    self.dataset = DataSet(dataconnection=dc,
                           dataset_name='grouptest',
                           dataset_info=json.dumps(data_info),
                           dataset_type='singletable',
                           dataset_detail='default.grouptest')
    self.dataset.save()

class Fact2DimsData(object):
  def __init__(self, dc):
    data_info = [{"tablename": "default.jvfact",
                  "columns": [{"name": "rownum", "type": "INT", "isdim": False, "alias": "rownum"},
                              {"name": "id", "type": "INT", "isdim": False, "alias": "id"},
                              {"name": "v1", "type": "INT", "isdim": False, "alias": "v1"},
                              {"name": "v2", "type": "DOUBLE", "isdim": False, "alias": "v2"},
                              {"name": "name", "type": "STRING", "isdim": True, "alias": "name"},
                              {"name": "year", "type": "INT", "isdim": False, "alias": "year"},
                              {"name": "month", "type": "INT", "isdim": False, "alias": "month"},
                              {"name": "day", "type": "INT", "isdim": False, "alias": "day"}]},
                 {"tablename": "default.jvdim", "parent_tablename": "default.jvfact", "join_pairs": [["name", "name"]],
                  "columns": [{"name": "name", "type": "STRING", "isdim": True, "alias": "jvdim name"},
                              {"name": "age", "type": "INT", "isdim": False, "alias": "age"},
                              {"name": "sex", "type": "STRING", "isdim": True, "alias": "sex"}]},
                 {"tablename": "default.jvdimp1", "parent_tablename": "default.jvfact",
                  "join_pairs": [["name", "name"]],
                  "columns": [{"name": "name", "type": "STRING", "isdim": True, "alias": "jvdimp1 name"},
                              {"name": "age", "type": "INT", "isdim": False, "alias": "jvdimp1 age"},
                              {"name": "gender", "type": "STRING", "isdim": True, "alias": "gender"},
                              {"name": "yr", "type": "INT", "isdim": False, "alias": "yr"}]}]
    self.dataset = DataSet(dataconnection=dc,
                           dataset_name='jvfact',
                           dataset_info=json.dumps(data_info),
                           dataset_type='multipletables',
                           dataset_detail='default.jvfact')
    self.dataset.save()
    # print 'created jvfact (fact + 2 dims) dataset'


class Fac3DimsData(object):
  def __init__(self, dc):
    data_info = [{"tablename": "default.jvfact",
                  "columns": [{"name": "rownum", "type": "INT", "isdim": False, "alias": "rownum"},
                              {"name": "id", "type": "INT", "isdim": False, "alias": "id"},
                              {"name": "v1", "type": "INT", "isdim": False, "alias": "v1"},
                              {"name": "v2", "type": "DOUBLE", "isdim": False, "alias": "v2"},
                              {"name": "name", "type": "STRING", "isdim": True, "alias": "name"},
                              {"name": "year", "type": "INT", "isdim": False, "alias": "year"},
                              {"name": "month", "type": "INT", "isdim": False, "alias": "month"},
                              {"name": "day", "type": "INT", "isdim": False, "alias": "day"}]},
                 {"tablename": "default.jvdim", "parent_tablename": "default.jvfact", "join_pairs": [["name", "name"]],
                  "columns": [{"name": "name", "type": "STRING", "isdim": True, "alias": "jvdim name"},
                              {"name": "age", "type": "INT", "isdim": False, "alias": "age"},
                              {"name": "sex", "type": "STRING", "isdim": True, "alias": "sex"}]},
                 {"tablename": "default.jvdimp1", "parent_tablename": "default.jvfact",
                  "join_pairs": [["name", "name"]],
                  "columns": [{"name": "name", "type": "STRING", "isdim": True, "alias": "jvdimp1 name"},
                              {"name": "age", "type": "INT", "isdim": False, "alias": "jvdimp1 age"},
                              {"name": "gender", "type": "STRING", "isdim": True, "alias": "gender"},
                              {"name": "yr", "type": "INT", "isdim": False, "alias": "yr"}]},
                 {"tablename": "default.jvdim1", "table_alias": "jvdim_1", "parent_tablename": "default.jvdimp1",
                  "join_pairs": [["name", "name"]],
                  "columns": [{"name": "name", "type": "STRING", "isdim": True, "alias": "jvdim_1 name"},
                              {"name": "age", "type": "INT", "isdim": False, "alias": "jvdim_1 age"},
                              {"name": "sex", "type": "STRING", "isdim": True, "alias": "jvdim_1 sex"}]}]
    self.dataset = DataSet(dataconnection=dc,
                           dataset_name='jvfact2',
                           dataset_info=json.dumps(data_info),
                           dataset_type='multipletables',
                           dataset_detail='default.jvfact')
    self.dataset.save()
    # print 'created jvfact2 (fact + 2+1 dims) dataset'


class MobileEvents(object):
  def __init__(self, dc):
    data_info = [{"tablename": "appboy.mobile_events",
                  "columns": [{"name": "event_id_hi", "type": "BIGINT", "isdim": False, "alias": "event_id_hi"},
                              {"name": "event_id_lo", "type": "BIGINT", "isdim": False, "alias": "event_id_lo"},
                              {"name": "app_id", "type": "STRING", "isdim": True, "alias": "app_id"},
                              {"name": "app_instance_id", "type": "STRING", "isdim": True, "alias": "app_instance_id"},
                              {"name": "name", "type": "STRING", "isdim": True, "alias": "name"},
                              {"name": "time", "type": "TIMESTAMP", "isdim": True, "alias": "time"},
                              {"name": "user_id", "type": "STRING", "isdim": True, "alias": "user_id", "hidden": False},
                              {"name": "", "type": "STRING", "isdim": True, "alias": "Copy_of_user_id", "hidden": False,
                               "expr": "hex([user_id])", "basecol": "user_id", "dflagg": "sum", "ispart": False,
                               "exprhasagg": False},
                              {"name": "device_id", "type": "STRING", "isdim": True, "alias": "device_id"},
                              {"name": "", "type": "STRING", "isdim": True, "alias": "Copy_of_device_id",
                               "expr": "hex([device_id])", "basecol": "device_id", "dflagg": "sum", "ispart": False,
                               "exprhasagg": False, "hidden": False},
                              {"name": "parent_event_id_hi", "type": "BIGINT", "isdim": False,
                               "alias": "parent_event_id_hi"},
                              {"name": "parent_event_id_lo", "type": "BIGINT", "isdim": False,
                               "alias": "parent_event_id_lo"},
                              {"name": "platform", "type": "STRING", "isdim": True, "alias": "platform"},
                              {"name": "sdk_version", "type": "STRING", "isdim": True, "alias": "sdk_version"},
                              {"name": "package_name", "type": "STRING", "isdim": True, "alias": "package_name"},
                              {"name": "version_name", "type": "STRING", "isdim": True, "alias": "version_name"},
                              {"name": "version_code", "type": "INT", "isdim": False, "alias": "version_code"},
                              {"name": "year", "type": "INT", "isdim": False, "alias": "year"},
                              {"name": "month", "type": "INT", "isdim": False, "alias": "month"},
                              {"name": "day", "type": "INT", "isdim": False, "alias": "day"},
                              {"name": "datestring", "type": "STRING", "isdim": True, "alias": "datestring"}]}]
    self.dataset = DataSet(dataconnection=dc,
                           dataset_name='mobile_events',
                           dataset_info=json.dumps(data_info),
                           dataset_type='singletable',
                           dataset_detail='appboy.mobile_events')
    self.dataset.save()
    # print 'created mobile_events dataset'


class SegmentData(object):
  def __init__(self, dataset=None):
    if dataset is None:
      raise ValueError('DataSet object is required')
    self._dataset = dataset
    self.segment_id = None  # gets assigned after createSegment method is called
    self.segment_name = None  # gets assigned after createSegment method is called

  def createSegment(self, segment_name=None, segment_group="", filters=None, entities=None):
    filters = filters or []
    entities = entities or []
    if segment_name is None or not len(filters):
      raise ValueError('segment_name and segment_data are required')
    self.segment_name = segment_name
    segment_data = {'name': segment_name,
                    'group': segment_group,
                    'filters': filters,
                    'entities': entities,
                    'dataset_id': self._dataset.dataset.id}
    url = '/datasets/segment'
    post = {'name': segment_name,
            'data': json.dumps(segment_data),
            'dataset_id': self._dataset.dataset.id}
    response = c.post(url, post, follow=True)
    if response.status_code != 200:
      raise Exception('could not create segment: ' + response.status_code)
    else:
      # print 'created segment ' + segment_name + ' on dataset ' + self._dataset.dataset.dataset_name + '...'
      pass
    self.segment_id = json.loads(response.content)[0].get('id')


class EventData(object):
  def __init__(self, dataset=None):
    if dataset is None:
      raise ValueError('DataSet object is required')
    self._dataset = dataset
    self.event_id = None  # gets assigned after createEvent method is called
    self.event_name = None  # gets assigned after createEvent method is called

  def createEvent(self, event_name=None, filters=None, display=None):
    filters = filters or []
    if event_name is None or not len(filters) or display is None:
      raise ValueError('event_name, display and filters are required')
    event_data = {'filters': filters,
                  'display': display}
    url = '/datasets/event'
    post = {'name': event_name,
            'data': json.dumps(event_data),
            'dataset_id': self._dataset.dataset.id}
    response = c.post(url, post, follow=True)
    if response.status_code != 200:
      raise Exception('could not create segment: ' + response.status_code)
    else:
      # print 'created event ' + event_name + ' on dataset ' + self._dataset.dataset.dataset_name + '...'
      pass
    self.event_id = json.loads(response.content)[0].get('id')


def addFilterAssociation(dataset=None, segment_list=None, name="", user_name=None):
  segment_list = segment_list or []
  if dataset is None:
    raise ValueError('DataSet is required')
  if not len(segment_list):
    raise ValueError('List of segments is required')
  if user_name is None:
    raise ValueError('username is required')
  url = '/datasets/filterassociation'
  # for now we are only adding segments with negate = False
  if not name:
    name = segment_list[0].segment_name
  data = {'applied_segments': []}
  for s in segment_list:
    data['applied_segments'].append({'id': s.segment_id, 'name': s.segment_name, 'negate': False})
  post = {'name': name,
          'dataset_id': dataset.dataset.id,
          'data': json.dumps(data)}
  response = c.post(url, post, follow=True)
  if response.status_code != 200:
    raise Exception('could not create filterassociation: ' + str(response.status_code))
  fa_id = json.loads(response.content)[0].get('id')
  url = '/datasets/userperms/' + str(dataset.dataset.id)
  post = {'username': user_name,
          'perm': 'readonly',
          'filterassociation_id': fa_id}
  c.post(url, post, follow=True)
  # print 'added filterassociation ' + name + ' to user ' + user_name


class Test_Base_Data_Connection(TestCase):
  def __init__(self, *args, **kwargs):
    super(Test_Base_Data_Connection, self).__init__(*args, **kwargs)

  # initial unit test setup
  def setUp(self):
    # print '*********** UNIT TESTS **************'
    # print 'SETTING UP STARTED...'

    if not hasattr(self, 'dataconnection_type'):
      return

    # create dataconnection
    print("Creating '%s' data connection" % self.dataconnection_type)
    dc = DataConnection(dataconnection_type=self.dataconnection_type,
      dataconnection_info = '{"PARAMS": {"FOR_TESTING": "true", "DATABASE": "tmp", "FILE": "/tmp/sqlite3_test_file", "HOST": "unknown", "PORT": 99, "AUTH": "NoSasl"}}')
    dc.save()
    self.dc = dc

    # create datasets
    census = CensusData(dc)
    self.censusdata_id = census.dataset.id
    censusCast = CensusDataCast(dc)
    self.censusdataCast_id = censusCast.dataset.id
    censusCast2 = CensusDataCast2(dc)
    self.censusdataCast2_id = censusCast2.dataset.id
    censusLV = CensusDataLV(dc)
    self.censusdataLV_id = censusLV.dataset.id
    jvfact = Fact2DimsData(dc)
    self.jvfactdata_id = jvfact.dataset.id
    grouptest = GroupTest(dc)
    self.grouptest_id = grouptest.dataset.id
    jvfact2 = Fac3DimsData(dc)
    self.jvfactdata2_id = jvfact2.dataset.id
    mobile_events = MobileEvents(dc)
    self.mobile_events_id = mobile_events.dataset.id

    # create users
    self.admin = UserAdmin()
    self.superuser_id = self.admin.id
    # login admin user to create segments
    if c.login(username=self.admin.username, password='arcadia'):
      pass
    else:
      raise Exception('admin user could not be authenticated')
    self.nonadmin = UserNonAdmin(dataset_id=self.censusdata_id, username='nonadmin')
    self.user_id = self.nonadmin.id
    # get request object
    factory = RequestFactory()
    # Create an instance of a GET request.
    self.request = factory.get('/apps/')
    # create segments
    self.segment1 = CaSegment(dataset=census)
    self.segment2 = FourStatesSegment(dataset=census)
    self.segment3 = CAentityYearPopL10M(dataset=census)
    self.segment4 = NYgroupASegment(dataset=census)
    self.segment5 = GAgroupASegment(dataset=census)
    self.segment6 = FemaleNameAndAge(dataset=jvfact)
    self.segment7 = Dailyme(dataset=mobile_events)
    self.event1 = InAppPurchase(dataset=mobile_events)
    self.event2 = StandardAppEvents(dataset=mobile_events)
    self.event3 = AllStandardAppEvents(dataset=mobile_events)
    self.event4 = SlideUpImpressionEvent(dataset=mobile_events)
    self.event5 = SlideUpClickedEvent(dataset=mobile_events)
    self.event6 = SessionStarted(dataset=mobile_events)
    addFilterAssociation(dataset=census, segment_list=[self.segment1], name=self.segment1.segment_name,
                         user_name=self.nonadmin.username)
    # print 'SETTING UP ENDED...'
    # print 'UNIT TESTS STARTED...'

  # called after unit tests are executed
  def tearDown(self):
    if hasattr(self, 'dc'):
      del_dataconn(self.dc.id)
    # print 'UNIT TESTS ENDED...'
    # print '*************************************'

  def generateSql(self, dsjson, request):
    dc = get_dataconn(self.dc.id)
    ds_request = DatasetRequest(json.dumps(dsjson))
    builder = QueryBuilder(ds_request, dc)
    query = builder.buildQuery(request.user.username, request.user.id, None)
    return dc.swap_quotes(query)

  def generateValidateSql(self, dsjson, request):
    dc = get_dataconn(self.dc.id)
    ds_request = DatasetRequest(json.dumps(dsjson))
    builder = QueryBuilder(ds_request, dc)
    builder.buildQuery(request.user.username, request.user.id, None)
    query = builder.builder.expr_hasAggregate(dsjson['validate'],
        builder.dataset_request.dataset)
    query = dc.swap_quotes(query)
    return query

  def assertStrings(self, str1, str2):
    msg = '\nactual:\n' + str1 + '\ngolden:\n' + str2
    str1 = ''.join(str1.split())  # remove all whitespace characters
    str2 = ''.join(str2.split())  # remove all whitespace characters
    self.assertEqual(str1, str2, msg=msg)

  def assertStringArrays(self, strl1, strl2):
    self.assertEqual(len(strl1), len(strl2))
    for i in range(0, len(strl1)):
      self.assertStrings(strl1[i], strl2[i])

# Write your segments below
class FourStatesSegment(SegmentData):
  def __init__(self, dataset=None):
    super(FourStatesSegment, self).__init__(dataset=dataset)
    super(FourStatesSegment, self).createSegment(segment_name='4states',
                                                 filters=["[state] in (\"DC\", \"DE\", \"FL\", \"GA\")"])


class CaSegment(SegmentData):
  def __init__(self, dataset=None):
    super(CaSegment, self).__init__(dataset=dataset)
    super(CaSegment, self).createSegment(segment_name='ca',
                                         filters=["[state] in (\"CA\")"])


class CAentityYearPopL10M(SegmentData):
  def __init__(self, dataset=None):
    super(CAentityYearPopL10M, self).__init__(dataset=dataset)
    super(CAentityYearPopL10M, self).createSegment(segment_name='CA_year_popL10M',
                                                   filters=["[state] in (\"CA\")", "[population] > 10000000"],
                                                   entities=["[population]", "[year]"])


class NYgroupASegment(SegmentData):
  def __init__(self, dataset=None):
    super(NYgroupASegment, self).__init__(dataset=dataset)
    super(NYgroupASegment, self).createSegment(segment_name='ny_groupA',
                                               segment_group='A',
                                               filters=["[state] in (\"NY\")"])


class GAgroupASegment(SegmentData):
  def __init__(self, dataset=None):
    super(GAgroupASegment, self).__init__(dataset=dataset)
    super(GAgroupASegment, self).createSegment(segment_name='ga_groupA',
                                               segment_group='A',
                                               filters=["[state] in (\"GA\")"])


class FemaleNameAndAge(SegmentData):
  def __init__(self, dataset=None):
    super(FemaleNameAndAge, self).__init__(dataset=dataset)
    super(FemaleNameAndAge, self).createSegment(segment_name='FemaleNameAndAgeEntity',
                                                filters=["[gender] in (\"F\")"],
                                                entities=["[age]", "[jvfact.name]"])


class Dailyme(SegmentData):
  def __init__(self, dataset=None):
    super(Dailyme, self).__init__(dataset=dataset)
    super(Dailyme, self).createSegment(segment_name='Dailyme',
                                       filters=["app_id='53163ecd0daac1a107000061'"],
                                       segment_group='Customers')


# Write your events below
class InAppPurchase(EventData):
  def __init__(self, dataset=None):
    super(InAppPurchase, self).__init__(dataset=dataset)
    super(InAppPurchase, self).createEvent(event_name='InAppPurchase',
                                           filters=["name = '$iap'"],
                                           display='\'InAppPurchase\'')


class StandardAppEvents(EventData):
  def __init__(self, dataset=None):
    super(StandardAppEvents, self).__init__(dataset=dataset)
    super(StandardAppEvents, self).createEvent(event_name='StandardAppEvents',
                                               filters=["name like \'$%\'",
                                                        "name != \'$session_started\'",
                                                        "name != \'$session_ended\'",
                                                        "name != \'$iap\'"],
                                               display='name')


class AllStandardAppEvents(EventData):
  def __init__(self, dataset=None):
    super(AllStandardAppEvents, self).__init__(dataset=dataset)
    super(AllStandardAppEvents, self).createEvent(event_name='AllStandardAppEvents',
                                                  filters=["name like \'$%\'"],
                                                  display='name')


class SlideUpImpressionEvent(EventData):
  def __init__(self, dataset=None):
    super(SlideUpImpressionEvent, self).__init__(dataset=dataset)
    super(SlideUpImpressionEvent, self).createEvent(event_name='SlideUpImpression',
                                                    filters=["name = \'$slideup_impression\'"],
                                                    display='\'SlideUp Impression\'')


class SlideUpClickedEvent(EventData):
  def __init__(self, dataset=None):
    super(SlideUpClickedEvent, self).__init__(dataset=dataset)
    super(SlideUpClickedEvent, self).createEvent(event_name='SlideUpClickedEvent',
                                                 filters=["name = \'$slideup_clicked\'"],
                                                 display='\'SlideUp Clicked\'')


class SessionStarted(EventData):
  def __init__(self, dataset=None):
    super(SessionStarted, self).__init__(dataset=dataset)
    super(SessionStarted, self).createEvent(event_name='SessionStarted',
                                            filters=["name = \'$session_started\'"],
                                            display='\'Session Started\'')

