# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
These tests verify SQL generation for the various backends.

We know the SQL we want to generate for the various backends, given
the dimensions, measures, segments, filter associations, and dataset
definitions that are created against the data connection. We use these
unit tests in order to confirm that the SQL generation remains as
expected.
"""

from . import test_base_dataconn
import re

MYSQL_TYPES = set(["mysql", "mariadb"])
PG_TYPES = set(["postgres", "redshift", "snowflake"])

class TestBuild_sql(test_base_dataconn.Test_Base_Data_Connection):
  def __init__(self, *args, **kwargs):
    super(TestBuild_sql, self).__init__(*args, **kwargs)

    """
    Expected outputs for extensions. Extension dataconnection tests are expected to populate this dictionary
    with expected SQL queries using the test method names as keys.
    """
    self.ext_outputs = {}

  # this method contains all unit tests to be executed
  def test_build_sql(self):
    if not hasattr(self, 'dataconnection_type'):
      return

    self.expr_hasaggr()
    self.select_from_dataset_cast1()
    self.select_from_dataset_cast2()
    self.select_from_lv()
    self.no_dims_no_aggrs1()
    self.no_dims_no_aggrs2()
    self.segmentShelfNoDataWglobalSegs()
    self.segmentsUnionAll()
    self.simple_dims_and_aggrs()
    self.nulls_order_first()
    self.nulls_order_last()
    self.entity_year_pop_forCA_popL10M()
    self.ca_global_seg()
    self.ca_global_seg_neg()
    self.ca_global_entityseg_neg_withfilter()
    self.global_segs_samegroup()
    self.global_segsdifgroup()
    self.global_segs_samegroup_error()
    self.multitable_nodimsnoaggrs()
    self.multitable_nofactselection1()
    self.multitable_nofactselection2()
    self.multitable_nofactselection3()
    self.multitable_nofactselection4()
    self.inapp_purchase_flow1()
    self.inapp_purchase_flow2()
    self.inapp_purchase_flow3()
    self.inapp_purchase_funnel1()
    self.SlideUpFunnel()

  def validateSQLGeneration(self, dsjson, validate=False, user=None):

    # attempt to support extension data connection tests.
    # the test function calling this function has to be the last but one
    # in the call stack:
    import traceback
    function_names = [s.name for s in traceback.extract_stack()]
    test_function = function_names[-2]
    print('test_function', test_function)

    if not user:
      user = self.admin

    self.request.user = user
    if validate:
      sql = self.generateValidateSql(dsjson, request=self.request)
    else:
      sql = self.generateSql(dsjson, request=self.request)

    if 'druid' in self.expected_outputs and not 'duckdb' in self.expected_outputs:
      self.expected_outputs["duckdb"] = self.expected_outputs["druid"] \
          .replace("STRING", "VARCHAR") \
          .replace("LIMIT 1)", "LIMIT 0)")

    if self.dataconnection_type in self.expected_outputs:
      self.assertStrings(sql, self.expected_outputs[self.dataconnection_type])
      return

    if self.dataconnection_type == "arcengine_v2" and self.expected_outputs.get("arcengine"):
      self.assertStrings(sql, self.expected_outputs["arcengine"])
      return

    if self.dataconnection_type == "impyla" and self.expected_outputs.get("impala"):
      self.assertStrings(sql, self.expected_outputs["impala"])
      return

    if self.dataconnection_type == "hyve" and self.expected_outputs.get("hive"):
      self.assertStrings(sql, self.expected_outputs["hive"])
      return

    if self.dataconnection_type in MYSQL_TYPES and self.expected_outputs.get("mysql"):
      self.assertStrings(sql, self.expected_outputs["mysql"])
      return

    if self.dataconnection_type in PG_TYPES and self.expected_outputs.get("postgres"):
      self.assertStrings(sql, self.expected_outputs["postgres"])
      return

    if test_function in self.ext_outputs:
      self.assertStrings(sql, self.ext_outputs[test_function])
      return

    self.assertStrings(sql, self.expected_outputs["default"])

  # write your unit tests below
  def expr_hasaggr(self):
    # print 'test: expression has aggregate'
    dsjson = {"version":1,
              "type":".validate-column",
              "time":[],
              "session":[],
              "dimensions":[],
              "aggregates":[],
              "filters":[],
              "having":[],
              "global_segments":[],
              "events":[],
              "distinct":False,
              "dataset_id": self.censusdata_id,
              "validate":"max([year])"}

    self.expected_outputs = {}
    self.expected_outputs["default"] = """
    EXPLAIN SELECT 1, max((`year`))
    FROM (SELECT TA_0.`year` as `year`, TA_0.`state` as `state`, TA_0.`population` as `population`, max((TA_0.`year`)) as `max((TA_0.'year'))`
    FROM `default`.`census_pop` TA_0
    LIMIT 0) TA_0 GROUP BY 1, 2
    """

    self.expected_outputs["hive"] = """
    EXPLAIN SELECT 1, max((`year`))
    FROM (SELECT TA_0.`year` as `year`, TA_0.`state` as `state`, TA_0.`population` as `population`, max((TA_0.`year`)) as `max((TA_0.'year'))`
    FROM `default`.`census_pop`  TA_0
    LIMIT 0) TA_0 GROUP BY  max((`year`))
    """

    self.expected_outputs["mysql"] = """
    SELECT max((`year`))
    FROM (SELECT TA_0.`year` as 'year', TA_0.`state` as 'state', TA_0.`population` as 'population', max((TA_0.`year`)) as 'max((TA_0.`year`))'
    FROM `default`.`census_pop` TA_0
    LIMIT 0) TA_0 GROUP BY 1
    """

    self.expected_outputs["mssqlserver"] = """
    SELECT max(([year]))
    FROM (SELECT TOP(0) TA_0.[year] as 'year', TA_0.[state] as 'state', TA_0.[population] as 'population', max((TA_0.[year])) as 'max((TA_0.[year]))'
    FROM [default].[census_pop] TA_0) TA_0 GROUP BY max([year])
    """

    self.expected_outputs["teradata"] = """
    SELECT max(("year"))
    FROM (SELECT TOP 1 TA_0."year" as "year", TA_0."state" as "state", TA_0."population" as "population", max((TA_0."year")) as "max((TA_0.""year""))"
    FROM "default"."census_pop" TA_0) TA_0 GROUP BY 1
    """

    self.expected_outputs["postgres"] = """
    SELECT max(("year"))
    FROM (SELECT TA_0."year" as "year", TA_0."state" as "state", TA_0."population" as "population"
    FROM "default"."census_pop" TA_0
    LIMIT 0) TA_0 GROUP BY 1
    """

    self.expected_outputs["sqlite"] = """
    SELECT max((`year`))
    FROM (SELECT TA_0.`year` as `year`, TA_0.`state` as `state`, TA_0.`population` as `population`, max((TA_0.`year`)) as `max((TA_0.'year'))`
    FROM `default`.`census_pop` TA_0
    LIMIT 1) TA_0 GROUP BY 1
    """

    self.expected_outputs["druid"] = """
    SELECT max(("year"))
    FROM (SELECT TA_0."year" as "year", TA_0."state" as "state", TA_0."population" as "population", max((TA_0."year")) as "max__TA_0__year___"
    FROM "default"."census_pop"  TA_0
    LIMIT 1) TA_0 GROUP BY 1
    """

    self.validateSQLGeneration(dsjson, validate=True)

  def select_from_dataset_cast1(self):
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [],
              "aggregates": [],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.censusdataCast_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = """
      SELECT (CAST (TA_0.`year` as STRING)) as `year`, TA_0.`state` as `state`, TA_0.`population` as `population`
        FROM `default`.`census_pop` TA_0
        LIMIT 100
    """

    self.expected_outputs["sqlite"] = """
      SELECT (CAST (TA_0.`year` as TEXT)) as `year`, TA_0.`state` as `state`, TA_0.`population` as `population`
        FROM `default`.`census_pop` TA_0
        LIMIT 100
    """

    self.expected_outputs["mysql"] = """
      SELECT (CAST(TA_0.`year` as CHAR)) as 'year', TA_0.`state` as 'state', TA_0.`population` as 'population'
        FROM `default`.`census_pop` TA_0
        LIMIT 100
      """

    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) (CAST (TA_0.[year] as VARCHAR(MAX))) as 'year', TA_0.[state] as 'state', TA_0.[population] as 'population'
    FROM [default].[census_pop] TA_0
    """

    self.expected_outputs["teradata"] = """
    SELECT TOP 100 (CAST (TA_0."year" as VARCHAR(1024))) as "year", TA_0."state" as "state", TA_0."population" as "population"
    FROM "default"."census_pop" TA_0
    """

    self.expected_outputs["postgres"] = """
      SELECT (CAST (TA_0."year" as TEXT)) as "year", TA_0."state" as "state", TA_0."population" as "population"
        FROM "default"."census_pop" TA_0
        LIMIT 100
    """

    self.expected_outputs["druid"] = self.expected_outputs["postgres"].replace("TEXT", "STRING")

    self.validateSQLGeneration(dsjson)

  def select_from_dataset_cast2(self):
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[year] as 'year'"},
                             {"type": "SIMPLE", "expr": "[state] as 'state'"}],
              "aggregates": [{"expr": "sum([population]) as 'sum(population)'"}],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.censusdataCast_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = """
      SELECT (CAST (TA_0.`year` as STRING)) as `year`, TA_0.`state` as `state`, sum((TA_0.`population`)) as `sum(population)`
        FROM `default`.`census_pop` TA_0
        GROUP BY 1, 2
        LIMIT 100
    """
    self.expected_outputs["sqlite"] = """
      SELECT (CAST (TA_0.`year` as TEXT)) as `year`, TA_0.`state` as `state`, sum((TA_0.`population`)) as `sum(population)`
        FROM `default`.`census_pop` TA_0
        GROUP BY 1, 2
        LIMIT 100
    """
    self.expected_outputs["hive"] = """
      SELECT (CAST (TA_0.`year` as STRING)) as `year`, TA_0.`state` as `state`, sum((TA_0.`population`)) as `sum(population)`
        FROM `default`.`census_pop` TA_0
        GROUP BY (CAST (TA_0.`year` as STRING)), TA_0.`state`
        LIMIT 100
    """
    self.expected_outputs["mysql"] = """
      SELECT (CAST(TA_0.`year` as CHAR)) as 'year', TA_0.`state` as 'state', sum((TA_0.`population`)) as 'sum(population)'
        FROM `default`.`census_pop` TA_0
        GROUP BY 1, 2
        LIMIT 100
      """

    self.expected_outputs["mssqlserver"] = """
      SELECT TOP(100) (CAST (TA_0.[year] as VARCHAR(MAX))) as 'year', TA_0.[state] as 'state', sum((TA_0.[population])) as 'sum(population)'
        FROM [default].[census_pop] TA_0
        GROUP BY (CAST (TA_0.[year] as VARCHAR(MAX))), TA_0.[state]
    """

    self.expected_outputs["teradata"] = """
      SELECT TOP 100 (CAST (TA_0."year" as VARCHAR(1024))) as "year", TA_0."state" as "state", sum((TA_0."population")) as "sum(population)"
        FROM "default"."census_pop" TA_0
        GROUP BY 1, 2
    """

    self.expected_outputs["postgres"] = """
      SELECT (CAST (TA_0."year" as TEXT)) as "year", TA_0."state" as "state", sum((TA_0."population")) as "sum(population)"
        FROM "default"."census_pop" TA_0
        GROUP BY 1, 2
        LIMIT 100
    """

    self.expected_outputs["snowflake"] = """
      SELECT (CAST (TA_0."year" as TEXT)) as "year", TA_0."state" as "state", sum((TA_0."population")) as "sum_population_"
        FROM "default"."census_pop" TA_0
        GROUP BY 1, 2
        LIMIT 100
    """

    self.expected_outputs["druid"] = self.expected_outputs["snowflake"].replace("TEXT", "STRING")

    self.validateSQLGeneration(dsjson)

  def select_from_lv(self):
    if self.dataconnection_type not in ["arcengine", "arcengine_v2"]:
      return

    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [],
              "aggregates": [],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.censusdataLV_id}

    self.expected_outputs = {}
    self.expected_outputs["arcengine"] = """
      SELECT TA_0.`year` as `year`, TA_0.`state` as `state`, TA_0.`population` as `population`
        FROM `default`.`census_lv` TA_0
        LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def no_dims_no_aggrs1(self):
    # print 'test: admin - no dimensions, no aggregates, no global segments'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [],
              "aggregates": [],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.censusdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`year` as `year`, " + \
                      "TA_0.`state` as `state`, " + \
                      "TA_0.`population` as `population`" + \
                      "\n FROM `default`.`census_pop` TA_0\n LIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_0.`year` as 'year', " + \
                      "TA_0.`state` as 'state', " + \
                      "TA_0.`population` as 'population'" + \
                      "\n FROM `default`.`census_pop` TA_0\n LIMIT 100"
    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[year] as 'year', TA_0.[state] as 'state', TA_0.[population] as 'population'
    FROM [default].[census_pop] TA_0
    """
    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."year" as "year", TA_0."state" as "state", TA_0."population" as "population"
    FROM "default"."census_pop" TA_0
    """

    self.expected_outputs["postgres"] = """
      SELECT TA_0."year" as "year", TA_0."state" as "state", TA_0."population" as "population"
        FROM "default"."census_pop" TA_0
        LIMIT 100
    """

    self.expected_outputs["druid"] = self.expected_outputs["postgres"]

    self.validateSQLGeneration(dsjson)

  def no_dims_no_aggrs2(self):
    # print 'test: no dimensions, no aggregates, no global segments, w/ association filters'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [],
              "aggregates": [],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.censusdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`year` as `year`, " + \
                      "TA_0.`state` as `state`, " + \
                      "TA_0.`population` as `population`" + \
                      "\n FROM `default`.`census_pop` TA_0" + \
                      "\nWHERE (((TA_0.`state`) in (\"CA\"))) \n LIMIT 100"

    self.expected_outputs["mysql"] = "SELECT TA_0.`year` as 'year', " + \
                      "TA_0.`state` as 'state', " + \
                      "TA_0.`population` as 'population'" + \
                      "\n FROM `default`.`census_pop` TA_0" + \
                      "\nWHERE (((TA_0.`state`) in (\"CA\"))) \n LIMIT 100"

    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[year] as 'year', TA_0.[state] as 'state', TA_0.[population] as 'population'
    FROM [default].[census_pop] TA_0
    WHERE (((TA_0.[state]) in ("CA")))
    """

    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."year" as "year", TA_0."state" as "state", TA_0."population" as "population"
    FROM "default"."census_pop" TA_0
    WHERE (((TA_0."state") in ("CA")))
    """

    self.expected_outputs["postgres"] = """
      SELECT TA_0."year" as "year", TA_0."state" as "state", TA_0."population" as "population"
        FROM "default"."census_pop" TA_0
        WHERE (((TA_0."state") in ("CA")))
        LIMIT 100
    """

    self.expected_outputs["druid"] = self.expected_outputs["postgres"]

    self.validateSQLGeneration(dsjson, user=self.nonadmin)

  def simple_dims_and_aggrs(self):
    # print 'test: simple test with dimensions and aggregates'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[state] as 'state'"},
                             {"type": "SIMPLE", "expr": "[year] as 'year'"}],
              "aggregates": [{"expr": "sum([population]) as 'sum(population)'"}],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.censusdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`state` as `state`, TA_0.`year` as `year`, " + \
                      "sum((TA_0.`population`)) as `sum(population)`" + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nGROUP BY 1, 2" + \
                      "\nLIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_0.`state` as 'state', TA_0.`year` as 'year', " + \
                      "sum((TA_0.`population`)) as 'sum(population)'" + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nGROUP BY 1, 2" + \
                      "\nLIMIT 100"
    self.expected_outputs["hive"] = "SELECT TA_0.`state` as `state`, TA_0.`year` as `year`, " + \
                      "sum((TA_0.`population`)) as `sum(population)`" + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nGROUP BY TA_0.`state`, TA_0.`year`" + \
                      "\nLIMIT 100"

    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[state] as 'state', TA_0.[year] as 'year', sum((TA_0.[population])) as 'sum(population)'
     FROM [default].[census_pop] TA_0
     GROUP BY TA_0.[state], TA_0.[year]
    """

    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum(population)"
     FROM "default"."census_pop" TA_0
     GROUP BY 1, 2
    """

    self.expected_outputs["postgres"] = """
      SELECT TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum(population)"
        FROM "default"."census_pop" TA_0
        GROUP BY 1, 2
        LIMIT 100
    """

    self.expected_outputs["snowflake"] = """
      SELECT TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum_population_"
        FROM "default"."census_pop" TA_0
        GROUP BY 1, 2
        LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum_population_"
    FROM "default"."census_pop"  TA_0
    GROUP BY 1, 2
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def nulls_order_first(self):
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[state] as 'state'", "order": {"asc": False, "pri": 1}},
                             {"type": "SIMPLE", "expr": "[year] as 'year'", "order": {"asc": True, "pri": 2}}],
              "aggregates": [{"expr": "sum([population]) as 'sum(population)'"}],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.censusdata_id,
              "nulls_order": "First"}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`state` as `state`, TA_0.`year` as `year`, " + \
                      "sum((TA_0.`population`)) as `sum(population)`" + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nGROUP BY 1, 2" + \
                      "\nORDER BY 1 DESC NULLS FIRST, 2 ASC NULLS FIRST" + \
                      "\nLIMIT 100"

    self.expected_outputs["sqlite"] = "SELECT TA_0.`state` as `state`, TA_0.`year` as `year`, " + \
                             "sum((TA_0.`population`)) as `sum(population)`" + \
                             "\nFROM `default`.`census_pop` TA_0" + \
                             "\nGROUP BY 1, 2" + \
                             "\nORDER BY 1 DESC, 2 ASC" + \
                             "\nLIMIT 100"

    self.expected_outputs["mysql"] = "SELECT TA_0.`state` as 'state', TA_0.`year` as 'year', " + \
                            "sum((TA_0.`population`)) as 'sum(population)'" + \
                            "\nFROM `default`.`census_pop` TA_0" + \
                            "\nGROUP BY 1, 2" + \
                            "\nORDER BY 1 DESC, 2 ASC" + \
                            "\nLIMIT 100"
    self.expected_outputs["hive"] = "SELECT TA_0.`state` as `state`, TA_0.`year` as `year`, " + \
                           "sum((TA_0.`population`)) as `sum(population)`" + \
                           "\nFROM `default`.`census_pop` TA_0" + \
                           "\nGROUP BY TA_0.`state`, TA_0.`year`" + \
                           "\nORDER BY `state` DESC, `year` ASC" + \
                           "\nLIMIT 100"

    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[state] as 'state', TA_0.[year] as 'year', sum((TA_0.[population])) as 'sum(population)'
      FROM [default].[census_pop] TA_0
      GROUP BY TA_0.[state], TA_0.[year]
      ORDER BY 1 DESC, 2 ASC
    """

    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum(population)"
      FROM "default"."census_pop" TA_0
      GROUP BY 1, 2
      ORDER BY 1 DESC, 2 ASC
    """

    self.expected_outputs["postgres"] = """
      SELECT TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum(population)"
        FROM "default"."census_pop" TA_0
        GROUP BY 1, 2
        ORDER BY 1 DESC NULLS FIRST, 2 ASC NULLS FIRST
        LIMIT 100
    """

    self.expected_outputs["snowflake"] = """
      SELECT TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum_population_"
        FROM "default"."census_pop" TA_0
        GROUP BY 1, 2
        ORDER BY 1 DESC NULLS FIRST, 2 ASC NULLS FIRST
        LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum_population_"
    FROM "default"."census_pop"  TA_0
    GROUP BY 1, 2
    ORDER BY 1 DESC, 2 ASC
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def nulls_order_last(self):
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[state] as 'state'", "order": {"asc": False, "pri": 1}},
                             {"type": "SIMPLE", "expr": "[year] as 'year'", "order": {"asc": True, "pri": 2}}],
              "aggregates": [{"expr": "sum([population]) as 'sum(population)'"}],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.censusdata_id,
              "nulls_order": "Last"}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`state` as `state`, TA_0.`year` as `year`, " + \
                      "sum((TA_0.`population`)) as `sum(population)`" + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nGROUP BY 1, 2" + \
                      "\nORDER BY 1 DESC NULLS LAST, 2 ASC NULLS LAST" + \
                      "\nLIMIT 100"

    self.expected_outputs["sqlite"] = "SELECT TA_0.`state` as `state`, TA_0.`year` as `year`, " + \
                             "sum((TA_0.`population`)) as `sum(population)`" + \
                             "\nFROM `default`.`census_pop` TA_0" + \
                             "\nGROUP BY 1, 2" + \
                             "\nORDER BY 1 DESC, 2 ASC" + \
                             "\nLIMIT 100"

    self.expected_outputs["mysql"] = "SELECT TA_0.`state` as 'state', TA_0.`year` as 'year', " + \
                          "sum((TA_0.`population`)) as 'sum(population)'" + \
                          "\nFROM `default`.`census_pop` TA_0" + \
                          "\nGROUP BY 1, 2" + \
                          "\nORDER BY 1 DESC, 2 ASC" + \
                          "\nLIMIT 100"

    self.expected_outputs["hive"] = "SELECT TA_0.`state` as `state`, TA_0.`year` as `year`, " + \
                           "sum((TA_0.`population`)) as `sum(population)`" + \
                           "\nFROM `default`.`census_pop` TA_0" + \
                           "\nGROUP BY TA_0.`state`, TA_0.`year`" + \
                           "\nORDER BY `state` DESC, `year` ASC" + \
                           "\nLIMIT 100"

    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[state] as 'state', TA_0.[year] as 'year', sum((TA_0.[population])) as 'sum(population)'
      FROM [default].[census_pop] TA_0
      GROUP BY TA_0.[state], TA_0.[year]
      ORDER BY 1 DESC, 2 ASC
    """

    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum(population)"
      FROM "default"."census_pop" TA_0
      GROUP BY 1, 2
      ORDER BY 1 DESC, 2 ASC
    """

    self.expected_outputs["postgres"] = """
      SELECT TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum(population)"
        FROM "default"."census_pop" TA_0
        GROUP BY 1, 2
        ORDER BY 1 DESC NULLS LAST, 2 ASC NULLS LAST
        LIMIT 100
    """

    self.expected_outputs["snowflake"] = """
      SELECT TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum_population_"
        FROM "default"."census_pop" TA_0
        GROUP BY 1, 2
        ORDER BY 1 DESC NULLS LAST, 2 ASC NULLS LAST
        LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_0."state" as "state", TA_0."year" as "year", sum((TA_0."population")) as "sum_population_"
    FROM "default"."census_pop"  TA_0
    GROUP BY 1, 2
    ORDER BY 1 DESC, 2 ASC
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def segmentShelfNoDataWglobalSegs(self):
    # print 'test: test using a Segment in the self that is empty but there are clicked global segments'
    # json post request object
    dsjson = {"version": 1, "type": "SQL", "time": [], "session": [],
              "dimensions": [{"type": "SEGMENT", "segments": []}, {"type": "SIMPLE", "expr": "[year] as 'year'"},
                             {"type": "SIMPLE", "expr": "[state] as 'state'"}],
              "aggregates": [{"expr": "avg([population]) as 'avg(population)'"}], "filters": [], "having": [],
              "global_segments": [{"type": "SEGMENT",
                                   "segments": [{"seg_id": self.segment1.segment_id, "negate": False},
                                                {"seg_id": self.segment2.segment_id, "negate": False},
                                   ]}],
              "events": [], "distinct": False, "dataset_id": self.censusdata_id, "limit": 100000}

    self.expected_outputs = {}

    self.expected_outputs["default"] = """
    SELECT * from (SELECT "all" as Segment, TA_0.`year` as `year`, TA_0.`state` as `state`, avg((TA_0.`population`)) as `avg(population)`
      FROM `default`.`census_pop` TA_0
      WHERE ((((TA_0.`state`) in ("CA")) OR ((TA_0.`state`) in ("DC", "DE", "FL", "GA"))))
      GROUP BY 1, 2, 3) TA
      LIMIT 100000
    """

    self.expected_outputs["hive"] = """
    SELECT * from (SELECT Segment, TA.`year` as `year`, TA.`state` as `state`, TA.`avg(population)` as `avg(population)`
      FROM
      (SELECT "all" as Segment, TA_0.`year` as `year`, TA_0.`state` as `state`, avg((TA_0.`population`)) as `avg(population)`
      FROM `default`.`census_pop`  TA_0
      WHERE ((((TA_0.`state`) in ("CA")) OR ((TA_0.`state`) in ("DC", "DE", "FL", "GA"))))
      GROUP BY TA_0.`year`, TA_0.`state`) TA
      ) TA
      LIMIT 100000
    """

    self.expected_outputs["mysql"] = """
    SELECT * from (SELECT "all" as Segment, TA_0.`year` as 'year', TA_0.`state` as 'state', avg((TA_0.`population`)) as 'avg(population)'
      FROM `default`.`census_pop`  TA_0
      WHERE ((((TA_0.`state`) in ("CA")) OR ((TA_0.`state`) in ("DC", "DE", "FL", "GA"))))
      GROUP BY 1, 2, 3) TA
      LIMIT 100000
    """

    self.expected_outputs["sqlite"] = """
    SELECT * from (SELECT "all" as Segment, TA_0.`year` as `year`, TA_0.`state` as `state`, avg((TA_0.`population`)) as `avg(population)`
      FROM `default`.`census_pop`  TA_0
      WHERE ((((TA_0.`state`) in ("CA")) OR ((TA_0.`state`) in ("DC", "DE", "FL", "GA"))))
      GROUP BY 1, 2, 3) TA
      LIMIT 100000
    """

    self.expected_outputs["mssqlserver"] = """
    SELECT * from (SELECT max('all') as Segment, TA_0.[year] as 'year', TA_0.[state] as 'state', avg((TA_0.[population])) as 'avg(population)'
      FROM [default].[census_pop]  TA_0
      WHERE ((((TA_0.[state]) in ("CA")) OR ((TA_0.[state]) in ("DC", "DE", "FL", "GA"))))
      GROUP BY TA_0.[year], TA_0.[state]) TA
    """

    self.expected_outputs["teradata"] = """
    SELECT * from (SELECT 'all' as Segment, TA_0."year" as "year", TA_0."state" as "state", avg((TA_0."population")) as "avg(population)"
      FROM "default"."census_pop"  TA_0
      WHERE ((((TA_0."state") in ("CA")) OR ((TA_0."state") in ("DC", "DE", "FL", "GA"))))
      GROUP BY 1, 2, 3) TA
    """

    self.expected_outputs["postgres"] = """
    SELECT * from (SELECT 'all' as Segment, TA_0."year" as "year", TA_0."state" as "state", avg((TA_0."population")) as "avg(population)"
      FROM "default"."census_pop"  TA_0
      WHERE ((((TA_0."state") in ("CA")) OR ((TA_0."state") in ("DC", "DE", "FL", "GA"))))
      GROUP BY 1, 2, 3) TA
      LIMIT 100000
    """

    self.expected_outputs["snowflake"] = """
    SELECT * from (SELECT 'all' as Segment, TA_0."year" as "year", TA_0."state" as "state", avg((TA_0."population")) as "avg_population_"
      FROM "default"."census_pop"  TA_0
      WHERE ((((TA_0."state") in ("CA")) OR ((TA_0."state") in ("DC", "DE", "FL", "GA"))))
      GROUP BY 1, 2, 3) TA
      LIMIT 100000
    """

    self.expected_outputs["druid"] = """
    SELECT * from (SELECT "all" as Segment, TA_0."year" as "year", TA_0."state" as "state", avg((TA_0."population")) as "avg_population_"
    FROM "default"."census_pop"  TA_0
    WHERE ((((TA_0."state") in ("CA")) OR ((TA_0."state") in ("DC", "DE", "FL", "GA"))))
    GROUP BY 1, 2, 3) TA
    LIMIT 100000
    """

    self.validateSQLGeneration(dsjson)

  def segmentsUnionAll(self):
    dsjson = {
      "version": 1,
      "type": 'SQL',
      "dimensions":[{"type":"SEGMENT",
                     "segments":[{"seg_id":self.segment1.segment_id,"negate":False},
                                 {"seg_id":self.segment2.segment_id,"negate":True}]},
                    {"type":"SIMPLE","expr":"[state] as 'state'"}],
      "aggregates": [{"expr": "sum(1) as 'Record Count'"}],
      "filters": [],
      "global_segments": [{"type": "SEGMENT", "segments": []}],
      "distinct": False, "dataset_id": self.censusdata_id, "limit": 100000
    }

    self.expected_outputs = {}
    self.expected_outputs["default"] = """
    SELECT * from (SELECT "ca" as Segment, TA_0.`state` as `state`, sum(1) as `Record Count`
      FROM `default`.`census_pop` TA_0
      WHERE (((TA_0.`state`) in ("CA")))
      GROUP BY 1, 2
    UNION ALL
    SELECT "not 4states" as Segment, TA_0.`state` as `state`, sum(1) as `Record Count`
      FROM `default`.`census_pop` TA_0
      WHERE ((NOT ((TA_0.`state`) in ("DC", "DE", "FL", "GA"))))
      GROUP BY 1, 2) TA
    LIMIT 100000
    """
    self.expected_outputs["hive"] = """
    SELECT * from (SELECT Segment, TA.`state` as `state`, TA.`Record Count` as `Record Count`
    FROM
    (SELECT "ca" as Segment, TA_0.`state` as `state`, sum(1) as `Record Count`
      FROM `default`.`census_pop` TA_0
      WHERE (((TA_0.`state`) in ("CA")))
      GROUP BY TA_0.`state`
    UNION ALL
    SELECT "not 4states" as Segment, TA_0.`state` as `state`, sum(1) as `Record Count`
      FROM `default`.`census_pop` TA_0
      WHERE ((NOT ((TA_0.`state`) in ("DC", "DE", "FL", "GA"))))
      GROUP BY TA_0.`state`) TA
    ) TA
    LIMIT 100000
    """
    self.expected_outputs["mysql"] = """
    SELECT * from (
    SELECT "ca" as Segment, TA_0.`state` as 'state', sum(1) as 'Record Count'
    FROM `default`.`census_pop` TA_0
    WHERE (((TA_0.`state`) in ("CA")))
    GROUP BY 1, 2
    UNION ALL
    SELECT "not 4states" as Segment, TA_0.`state` as 'state', sum(1) as 'Record Count'
    FROM `default`.`census_pop` TA_0
    WHERE ((NOT ((TA_0.`state`) in ("DC", "DE", "FL", "GA"))))
    GROUP BY 1, 2 ) TA
    LIMIT 100000
    """
    self.expected_outputs["sqlite"] = """
    SELECT * from (SELECT "ca" as Segment, TA_0.`state` as `state`, sum(1) as `Record Count`
    FROM `default`.`census_pop` TA_0
    WHERE (((TA_0.`state`) in ("CA")))
    GROUP BY 1, 2
    UNION ALL
    SELECT "not 4states" as Segment, TA_0.`state` as `state`, sum(1) as `Record Count`
    FROM `default`.`census_pop` TA_0
    WHERE ((NOT ((TA_0.`state`) in ("DC", "DE", "FL", "GA"))))
    GROUP BY 1, 2) TA
    LIMIT 100000
    """

    self.expected_outputs["mssqlserver"] = """
    SELECT * from (SELECT max('ca') as Segment, TA_0.[state] as 'state', sum(1) as 'Record Count'
      FROM [default].[census_pop] TA_0
      WHERE (((TA_0.[state]) in ("CA")))
      GROUP BY TA_0.[state]
    UNION ALL
    SELECT max('not 4states') as Segment, TA_0.[state] as 'state', sum(1) as 'Record Count'
      FROM [default].[census_pop] TA_0
      WHERE ((NOT ((TA_0.[state]) in ("DC", "DE", "FL", "GA"))))
      GROUP BY TA_0.[state]) TA
    """

    self.expected_outputs["teradata"] = """
    SELECT * from (SELECT 'ca' as Segment, TA_0."state" as "state", sum(1) as "Record Count"
      FROM "default"."census_pop" TA_0
      WHERE (((TA_0."state") in ("CA")))
      GROUP BY 1, 2
    UNION ALL
    SELECT 'not 4states' as Segment, TA_0."state" as "state", sum(1) as "Record Count"
      FROM "default"."census_pop" TA_0
      WHERE ((NOT ((TA_0."state") in ("DC", "DE", "FL", "GA"))))
      GROUP BY 1, 2) TA
    """

    self.expected_outputs["postgres"] = """
    SELECT * from ( SELECT 'ca' as Segment, TA_0."state" as "state", sum(1) as "Record Count"
    FROM "default"."census_pop" TA_0
    WHERE (((TA_0."state") in ("CA")))
    GROUP BY 1, 2
    UNION ALL
    SELECT 'not 4states' as Segment, TA_0."state" as "state", sum(1) as "Record Count"
    FROM "default"."census_pop" TA_0
    WHERE ((NOT ((TA_0."state") in ("DC", "DE", "FL", "GA"))))
    GROUP BY 1, 2 ) TA
    LIMIT 100000
    """

    self.expected_outputs["druid"] = """
    SELECT * from (SELECT "ca" as Segment, TA_0."state" as "state", sum(1) as "Record Count"
    FROM "default"."census_pop"  TA_0
    WHERE (((TA_0."state") in ("CA")))
    GROUP BY 1, 2
    UNION ALL
    SELECT "not 4states" as Segment, TA_0."state" as "state", sum(1) as "Record Count"
    FROM "default"."census_pop"  TA_0
    WHERE ((NOT ((TA_0."state") in ("DC", "DE", "FL", "GA"))))
    GROUP BY 1, 2) TA
    LIMIT 100000
    """

    self.validateSQLGeneration(dsjson)

  def entity_year_pop_forCA_popL10M(self):
    # print 'test: entity segment year and pop of CA for pop > 10M'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SEGMENT", "segments": [{"seg_id": self.segment3.segment_id, "negate": False}]},
                             {"type": "SIMPLE", "expr": "[state] as 'state'"},
                             {"type": "SIMPLE", "expr": "[year] as 'year'", "order": {"asc": False, "pri": 1}},
                             {"type": "SIMPLE", "expr": "[population] as 'population'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}], "limit": 100, "dataset_id": self.censusdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT * from ( SELECT \"CA_year_popL10M\" as Segment, " + \
                      "TA_0.`state` as `state`, " + \
                      "TA_0.`year` as `year`, " + \
                      "TA_0.`population` as `population`, " + \
                      "sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0 join " + \
                      "(SELECT TA_0.`population` as `arc_u10_0`, " + \
                      "TA_0.`year` as `arc_u10_1` " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "WHERE (((TA_0.`state`) in (\"CA\") " + \
                      "AND (TA_0.`population`) > 10000000)) " + \
                      "GROUP BY 1, 2) arc_u10 " + \
                      "ON TA_0.`population`=arc_u10_0 " + \
                      "AND TA_0.`year`=arc_u10_1 " + \
                      "\nGROUP BY 1, 2, 3, 4 ) TA" + \
                      "\nORDER BY 3 DESC " + \
                      "\nLIMIT 100"
    self.expected_outputs["hive"] = """
        SELECT * from (SELECT Segment, TA.`state` as `state`, TA.`year` as `year`, TA.`population` as `population`, TA.`Record Count` as `Record Count`
          FROM
          (SELECT "CA_year_popL10M" as Segment, TA_0.`state` as `state`, TA_0.`year` as `year`, TA_0.`population` as `population`, sum(1) as `Record Count`
          FROM `default`.`census_pop` TA_0 join
            (SELECT TA_0.`population` as `arc_u10_0`, TA_0.`year` as `arc_u10_1`
            FROM `default`.`census_pop` TA_0
            WHERE (((TA_0.`state`) in ("CA") AND (TA_0.`population`) > 10000000))
            GROUP BY TA_0.`population`, TA_0.`year`) arc_u10 ON TA_0.`population`=arc_u10_0 AND TA_0.`year`=arc_u10_1
          GROUP BY TA_0.`state`, TA_0.`year`, TA_0.`population`) TA
          ) TA
          ORDER BY `year` DESC
          LIMIT 100
    """
    self.expected_outputs["mysql"] = """
        SELECT * from (
        SELECT "CA_year_popL10M" as Segment, TA_0.`state` as 'state', TA_0.`year` as 'year', TA_0.`population` as 'population', sum(1) as 'Record Count'
          FROM `default`.`census_pop` TA_0 join
            (SELECT TA_0.`population` as 'arc_u10_0', TA_0.`year` as 'arc_u10_1'
            FROM `default`.`census_pop` TA_0
            WHERE (((TA_0.`state`) in ("CA") AND (TA_0.`population`) > 10000000))
            GROUP BY 1, 2) arc_u10 ON TA_0.`population`=arc_u10_0 AND TA_0.`year`=arc_u10_1
          GROUP BY 1, 2, 3, 4 ) TA
          ORDER BY 3 DESC
          LIMIT 100
    """
    self.expected_outputs["sqlite"] = """
        SELECT * from (SELECT "CA_year_popL10M" as Segment, TA_0.`state` as `state`, TA_0.`year` as `year`, TA_0.`population` as `population`, sum(1) as `Record Count`
          FROM `default`.`census_pop` TA_0 join
            (SELECT TA_0.`population` as `arc_u10_0`, TA_0.`year` as `arc_u10_1`
            FROM `default`.`census_pop` TA_0
            WHERE (((TA_0.`state`) in ("CA") AND (TA_0.`population`) > 10000000))
            GROUP BY 1, 2) arc_u10 ON TA_0.`population`=arc_u10_0 AND TA_0.`year`=arc_u10_1
          GROUP BY 1, 2, 3, 4) TA
          ORDER BY 3 DESC
          LIMIT 100
    """

    self.expected_outputs["mssqlserver"] = """
    SELECT * from (SELECT max('CA_year_popL10M') as Segment, TA_0.[state] as 'state', TA_0.[year] as 'year', TA_0.[population] as 'population', sum(1) as 'Record Count'
      FROM [default].[census_pop] TA_0 join
        (SELECT TA_0.[population] as 'arc_u10_0', TA_0.[year] as 'arc_u10_1'
        FROM [default].[census_pop] TA_0
        WHERE (((TA_0.[state]) in ("CA") AND (TA_0.[population]) > 10000000))
        GROUP BY TA_0.[population], TA_0.[year]) arc_u10 ON TA_0.[population]=arc_u10_0 AND TA_0.[year]=arc_u10_1
      GROUP BY TA_0.[state], TA_0.[year], TA_0.[population]) TA
      ORDER BY 3 DESC
    """

    self.expected_outputs["teradata"] = """
    SELECT * from (SELECT 'CA_year_popL10M' as Segment, TA_0."state" as "state", TA_0."year" as "year", TA_0."population" as "population", sum(1) as "Record Count"
      FROM "default"."census_pop" TA_0 join
        (SELECT TA_0."population" as "arc_u10_0", TA_0."year" as "arc_u10_1"
        FROM "default"."census_pop" TA_0
        WHERE (((TA_0."state") in ("CA") AND (TA_0."population") > 10000000))
        GROUP BY 1, 2) arc_u10 ON TA_0."population"=arc_u10_0 AND TA_0."year"=arc_u10_1
      GROUP BY 1, 2, 3, 4) TA
      ORDER BY 3 DESC
    """

    self.expected_outputs["postgres"] = """
      SELECT * from (
      SELECT 'CA_year_popL10M' as Segment, TA_0."state" as "state", TA_0."year" as "year", TA_0."population" as "population", sum(1) as "Record Count"
        FROM "default"."census_pop" TA_0 join
          (SELECT TA_0."population" as "arc_u10_0", TA_0."year" as "arc_u10_1"
          FROM "default"."census_pop" TA_0
          WHERE (((TA_0."state") in ("CA") AND (TA_0."population") > 10000000))
          GROUP BY 1, 2) arc_u10 ON TA_0."population"=arc_u10_0 AND TA_0."year"=arc_u10_1
        GROUP BY 1, 2, 3, 4 ) TA
        ORDER BY 3 DESC
        LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT * from (SELECT "CA_year_popL10M" as Segment, TA_0."state" as "state", TA_0."year" as "year", TA_0."population" as "population", sum(1) as "Record Count"
    FROM "default"."census_pop"  TA_0  join
      (SELECT TA_0."population" as "arc_u10_0", TA_0."year" as "arc_u10_1"
      FROM "default"."census_pop"  TA_0
      WHERE (((TA_0."state") in ("CA") AND (TA_0."population") > 10000000))
      GROUP BY 1, 2) arc_u10 ON TA_0."population"=arc_u10_0 AND TA_0."year"=arc_u10_1
    GROUP BY 1, 2, 3, 4) TA
    ORDER BY 3 DESC
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def ca_global_seg(self):
    # print 'test: global segment CA'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[state] as 'state'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [
                {"type": "SEGMENT", "segments": [{"seg_id": self.segment1.segment_id, "negate": False}]}], "limit": 100,
              "dataset_id": self.censusdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`state` as `state`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE (((TA_0.`state`) in (\"CA\"))) " + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_0.`state` as 'state', sum(1) as 'Record Count' " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE (((TA_0.`state`) in (\"CA\"))) " + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["hive"] = "SELECT TA_0.`state` as `state`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE (((TA_0.`state`) in (\"CA\"))) " + \
                      "\nGROUP BY TA_0.`state` " + \
                      "\nLIMIT 100"
    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[state] as 'state', sum(1) as 'Record Count'
      FROM [default].[census_pop] TA_0
      WHERE (((TA_0.[state]) in ("CA")))
      GROUP BY TA_0.[state]
    """
    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."state" as "state", sum(1) as "Record Count"
      FROM "default"."census_pop" TA_0
      WHERE (((TA_0."state") in ("CA")))
      GROUP BY 1
    """
    self.expected_outputs["postgres"] = """
      SELECT TA_0."state" as "state", sum(1) as "Record Count"
        FROM "default"."census_pop" TA_0
        WHERE (((TA_0."state") in ("CA")))
        GROUP BY 1
        LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_0."state" as "state", sum(1) as "Record Count"
    FROM "default"."census_pop"  TA_0
    WHERE (((TA_0."state") in ("CA")))
    GROUP BY 1
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def ca_global_seg_neg(self):
    # print 'test: negated global segment CA'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[state] as 'state'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [
                {"type": "SEGMENT", "segments": [{"seg_id": self.segment1.segment_id, "negate": True}]}], "limit": 100,
              "dataset_id": self.censusdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`state` as `state`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE ((NOT ((TA_0.`state`) in (\"CA\")))) " + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_0.`state` as 'state', sum(1) as 'Record Count' " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE ((NOT ((TA_0.`state`) in (\"CA\")))) " + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["hive"] = "SELECT TA_0.`state` as `state`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE ((NOT ((TA_0.`state`) in (\"CA\")))) " + \
                      "\nGROUP BY TA_0.`state` " + \
                      "\nLIMIT 100"
    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[state] as 'state', sum(1) as 'Record Count'
      FROM [default].[census_pop] TA_0
      WHERE ((NOT ((TA_0.[state]) in ("CA"))))
      GROUP BY TA_0.[state]
    """
    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."state" as "state", sum(1) as "Record Count"
      FROM "default"."census_pop" TA_0
      WHERE ((NOT ((TA_0."state") in ("CA"))))
      GROUP BY 1
    """
    self.expected_outputs["postgres"] = """
      SELECT TA_0."state" as "state", sum(1) as "Record Count"
        FROM "default"."census_pop" TA_0
        WHERE ((NOT ((TA_0."state") in ("CA"))))
        GROUP BY 1
        LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_0."state" as "state", sum(1) as "Record Count"
    FROM "default"."census_pop"  TA_0
    WHERE ((NOT ((TA_0."state") in ("CA"))))
    GROUP BY 1
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def ca_global_entityseg_neg_withfilter(self):
    # print 'test: negated global entity segment CA'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[state] as 'state'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}],
              "filters": ["[year] in ('2000')"],
              "global_segments": [
                {"type": "SEGMENT", "segments": [{"seg_id": self.segment3.segment_id, "negate": True}]}], "limit": 100,
              "dataset_id": self.censusdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`state` as `state`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0 left outer join" + \
                      "\n(SELECT TA_0.`population` as `arc_u10_0`, TA_0.`year` as `arc_u10_1`" + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE (((TA_0.`state`) in (\"CA\") AND (TA_0.`population`) > 10000000)) " + \
                      "\nGROUP BY 1, 2) arc_u10 " + \
                      "ON TA_0.`population`=arc_u10_0 AND TA_0.`year`=arc_u10_1 " + \
                      "WHERE (TA_0.`year` in ('2000')) AND (arc_u10_0 is null AND arc_u10_1 is null)" + \
                      "\nGROUP BY 1" + \
                      "\nLIMIT 100"
    self.expected_outputs["mysql"] = """
        SELECT TA_0.`state` as 'state', sum(1) as 'Record Count'
          FROM `default`.`census_pop` TA_0 left outer join
            (SELECT TA_0.`population` as 'arc_u10_0', TA_0.`year` as 'arc_u10_1'
            FROM `default`.`census_pop` TA_0
            WHERE (((TA_0.`state`) in ("CA") AND (TA_0.`population`) > 10000000))
            GROUP BY 1, 2) arc_u10 ON TA_0.`population`=arc_u10_0 AND TA_0.`year`=arc_u10_1
          WHERE (TA_0.`year` in ('2000')) AND (arc_u10_0 is null AND arc_u10_1 is null)
          GROUP BY 1
          LIMIT 100
    """
    self.expected_outputs["hive"] = """
        SELECT TA_0.`state` as `state`, sum(1) as `Record Count`
          FROM `default`.`census_pop` TA_0 left outer join
            (SELECT TA_0.`population` as `arc_u10_0`, TA_0.`year` as `arc_u10_1`
            FROM `default`.`census_pop` TA_0
            WHERE (((TA_0.`state`) in ("CA") AND (TA_0.`population`) > 10000000))
            GROUP BY TA_0.`population`, TA_0.`year`) arc_u10 ON TA_0.`population`=arc_u10_0 AND TA_0.`year`=arc_u10_1
          WHERE (TA_0.`year` in ('2000')) AND (arc_u10_0 is null AND arc_u10_1 is null)
          GROUP BY TA_0.`state`
          LIMIT 100
    """
    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[state] as 'state', sum(1) as 'Record Count'
      FROM [default].[census_pop] TA_0 left outer join
        (SELECT TA_0.[population] as 'arc_u10_0', TA_0.[year] as 'arc_u10_1'
        FROM [default].[census_pop] TA_0
        WHERE (((TA_0.[state]) in ("CA") AND (TA_0.[population]) > 10000000))
        GROUP BY TA_0.[population], TA_0.[year]) arc_u10 ON TA_0.[population]=arc_u10_0 AND TA_0.[year]=arc_u10_1
      WHERE (TA_0.[year] in ('2000')) AND (arc_u10_0 is null AND arc_u10_1 is null)
      GROUP BY TA_0.[state]
    """
    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."state" as "state", sum(1) as "Record Count"
      FROM "default"."census_pop" TA_0 left outer join
        (SELECT TA_0."population" as "arc_u10_0", TA_0."year" as "arc_u10_1"
        FROM "default"."census_pop" TA_0
        WHERE (((TA_0."state") in ("CA") AND (TA_0."population") > 10000000))
        GROUP BY 1, 2) arc_u10 ON TA_0."population"=arc_u10_0 AND TA_0."year"=arc_u10_1
      WHERE (TA_0."year" in ('2000')) AND (arc_u10_0 is null AND arc_u10_1 is null)
      GROUP BY 1
    """
    self.expected_outputs["postgres"] = """
      SELECT TA_0."state" as "state", sum(1) as "Record Count"
        FROM "default"."census_pop" TA_0 left outer join
          (SELECT TA_0."population" as "arc_u10_0", TA_0."year" as "arc_u10_1"
          FROM "default"."census_pop" TA_0
          WHERE (((TA_0."state") in ("CA") AND (TA_0."population") > 10000000))
          GROUP BY 1, 2) arc_u10 ON TA_0."population"=arc_u10_0 AND TA_0."year"=arc_u10_1
        WHERE (TA_0."year" in ('2000')) AND (arc_u10_0 is null AND arc_u10_1 is null)
        GROUP BY 1
        LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_0."state" as "state", sum(1) as "Record Count"
    FROM "default"."census_pop"  TA_0  left outer join
      (SELECT TA_0."population" as "arc_u10_0", TA_0."year" as "arc_u10_1"
      FROM "default"."census_pop"  TA_0
      WHERE (((TA_0."state") in ("CA") AND (TA_0."population") > 10000000))
      GROUP BY 1, 2) arc_u10 ON TA_0."population"=arc_u10_0 AND TA_0."year"=arc_u10_1
    WHERE (TA_0."year" in ('2000')) AND (arc_u10_0 is null AND arc_u10_1 is null)
    GROUP BY 1
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def global_segs_samegroup(self):
    # print 'test: 2 global segments in same group'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[state] as 'state'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [
                {"type": "SEGMENT", "segments": [{"seg_id": self.segment4.segment_id, "negate": False},
                                                 {"seg_id": self.segment5.segment_id, "negate": False}]}], "limit": 100,
              "dataset_id": self.censusdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`state` as `state`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE ((((TA_0.`state`) in (\"NY\")) OR ((TA_0.`state`) in (\"GA\")))) " + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_0.`state` as 'state', sum(1) as 'Record Count' " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE ((((TA_0.`state`) in (\"NY\")) OR ((TA_0.`state`) in (\"GA\")))) " + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["hive"] = "SELECT TA_0.`state` as `state`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE ((((TA_0.`state`) in (\"NY\")) OR ((TA_0.`state`) in (\"GA\")))) " + \
                      "\nGROUP BY TA_0.`state` " + \
                      "\nLIMIT 100"

    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[state] as 'state', sum(1) as 'Record Count'
      FROM [default].[census_pop] TA_0
      WHERE ((((TA_0.[state]) in ("NY")) OR ((TA_0.[state]) in ("GA"))))
      GROUP BY TA_0.[state]
    """

    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."state" as "state", sum(1) as "Record Count"
      FROM "default"."census_pop" TA_0
      WHERE ((((TA_0."state") in ("NY")) OR ((TA_0."state") in ("GA"))))
      GROUP BY 1
    """

    self.expected_outputs["postgres"] = """
      SELECT TA_0."state" as "state", sum(1) as "Record Count"
        FROM "default"."census_pop" TA_0
        WHERE ((((TA_0."state") in ("NY")) OR ((TA_0."state") in ("GA"))))
        GROUP BY 1
        LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_0."state" as "state", sum(1) as "Record Count"
    FROM "default"."census_pop"  TA_0
    WHERE ((((TA_0."state") in ("NY")) OR ((TA_0."state") in ("GA"))))
    GROUP BY 1
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def global_segsdifgroup(self):
    # print 'test: 2 global segments in different groups'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[state] as 'state'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [
                {"type": "SEGMENT", "segments": [{"seg_id": self.segment1.segment_id, "negate": False},
                                                 {"seg_id": self.segment4.segment_id, "negate": False}]}], "limit": 100,
              "dataset_id": self.censusdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`state` as `state`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE (((TA_0.`state`) in (\"CA\"))) AND (((TA_0.`state`) in (\"NY\"))) " + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_0.`state` as 'state', sum(1) as 'Record Count' " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE (((TA_0.`state`) in (\"CA\"))) AND (((TA_0.`state`) in (\"NY\"))) " + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["hive"] = "SELECT TA_0.`state` as `state`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`census_pop` TA_0" + \
                      "\nWHERE (((TA_0.`state`) in (\"CA\"))) AND (((TA_0.`state`) in (\"NY\"))) " + \
                      "\nGROUP BY TA_0.`state` " + \
                      "\nLIMIT 100"
    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[state] as 'state', sum(1) as 'Record Count'
      FROM [default].[census_pop] TA_0
      WHERE (((TA_0.[state]) in ("CA"))) AND (((TA_0.[state]) in ("NY")))
      GROUP BY TA_0.[state]
    """
    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."state" as "state", sum(1) as "Record Count"
      FROM "default"."census_pop" TA_0
      WHERE (((TA_0."state") in ("CA"))) AND (((TA_0."state") in ("NY")))
      GROUP BY 1
    """
    self.expected_outputs["postgres"] = """
      SELECT TA_0."state" as "state", sum(1) as "Record Count"
        FROM "default"."census_pop" TA_0
        WHERE (((TA_0."state") in ("CA"))) AND (((TA_0."state") in ("NY")))
        GROUP BY 1
        LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_0."state" as "state", sum(1) as "Record Count"
    FROM "default"."census_pop"  TA_0
    WHERE (((TA_0."state") in ("CA"))) AND (((TA_0."state") in ("NY")))
    GROUP BY 1
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def global_segs_samegroup_error(self):
    with self.assertRaises(ValueError):
      # print 'test: 2 global segments in same group 1 entity and 1 non-entity raising exception'
      # json post request object
      dsjson = {"version": 1,
                "type": 'SQL',
                "dimensions": [{"type": "SIMPLE", "expr": "[state] as 'state'"}],
                "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
                "global_segments": [
                  {"type": "SEGMENT", "segments": [{"seg_id": self.segment1.segment_id, "negate": False},
                                                   {"seg_id": self.segment3.segment_id, "negate": False}]}],
                "limit": 100,
                "dataset_id": self.censusdata_id}
      self.request.user = self.admin
      self.generateSql(dsjson, request=self.request)

  def multitable_nodimsnoaggrs(self):
    # print 'test: multitable test of outer joins'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [],
              "aggregates": [],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.jvfactdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_0.`rownum` as `rownum`, TA_0.`id` as `id`, " + \
                      "TA_0.`v1` as `v1`, TA_0.`v2` as `v2`, TA_0.`name` as `name`, " + \
                      "TA_0.`year` as `year`, TA_0.`month` as `month`, TA_0.`day` as `day`, " + \
                      "TA_1.`name` as `jvdim name`, TA_1.`age` as `age`, TA_1.`sex` as `sex`, " + \
                      "TA_2.`name` as `jvdimp1 name`, TA_2.`age` as `jvdimp1 age`, " + \
                      "TA_2.`gender` as `gender`, TA_2.`yr` as `yr` " + \
                      "\n  FROM `default`.`jvfact` TA_0 LEFT OUTER JOIN `default`.`jvdim` TA_1 " + \
                      "ON ( TA_0.`name` = TA_1.`name` ) " + \
                      "LEFT OUTER JOIN `default`.`jvdimp1` TA_2 ON ( TA_0.`name` = TA_2.`name` ) " + \
                      "\n  LIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_0.`rownum` as 'rownum', TA_0.`id` as 'id', " + \
                      "TA_0.`v1` as 'v1', TA_0.`v2` as 'v2', TA_0.`name` as 'name', " + \
                      "TA_0.`year` as 'year', TA_0.`month` as 'month', TA_0.`day` as 'day', " + \
                      "TA_1.`name` as 'jvdim name', TA_1.`age` as 'age', TA_1.`sex` as 'sex', " + \
                      "TA_2.`name` as 'jvdimp1 name', TA_2.`age` as 'jvdimp1 age', " + \
                      "TA_2.`gender` as 'gender', TA_2.`yr` as 'yr' " + \
                      "\n  FROM `default`.`jvfact` TA_0 LEFT OUTER JOIN `default`.`jvdim` TA_1 " + \
                      "ON ( TA_0.`name` = TA_1.`name` ) " + \
                      "LEFT OUTER JOIN `default`.`jvdimp1` TA_2 ON ( TA_0.`name` = TA_2.`name` ) " + \
                      "\n  LIMIT 100"
    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_0.[rownum] as 'rownum', TA_0.[id] as 'id', TA_0.[v1] as 'v1', TA_0.[v2] as 'v2', TA_0.[name] as 'name', TA_0.[year] as 'year', TA_0.[month] as 'month', TA_0.[day] as 'day', TA_1.[name] as 'jvdim name', TA_1.[age] as 'age', TA_1.[sex] as 'sex', TA_2.[name] as 'jvdimp1 name', TA_2.[age] as 'jvdimp1 age', TA_2.[gender] as 'gender', TA_2.[yr] as 'yr'
    FROM [default].[jvfact] TA_0 LEFT OUTER JOIN [default].[jvdim] TA_1 ON ( TA_0.[name] = TA_1.[name] )  LEFT OUTER JOIN [default].[jvdimp1] TA_2 ON ( TA_0.[name] = TA_2.[name] )
    """
    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_0."rownum" as "rownum", TA_0."id" as "id", TA_0."v1" as "v1", TA_0."v2" as "v2", TA_0."name" as "name", TA_0."year" as "year", TA_0."month" as "month", TA_0."day" as "day", TA_1."name" as "jvdim name", TA_1."age" as "age", TA_1."sex" as "sex", TA_2."name" as "jvdimp1 name", TA_2."age" as "jvdimp1 age", TA_2."gender" as "gender", TA_2."yr" as "yr"
    FROM "default"."jvfact" TA_0 LEFT OUTER JOIN "default"."jvdim" TA_1 ON ( TA_0."name" = TA_1."name" )  LEFT OUTER JOIN "default"."jvdimp1" TA_2 ON ( TA_0."name" = TA_2."name" )
    """
    self.expected_outputs["postgres"] = """
      SELECT TA_0."rownum" as "rownum", TA_0."id" as "id", TA_0."v1" as "v1", TA_0."v2" as "v2", TA_0."name" as "name", TA_0."year" as "year", TA_0."month" as "month", TA_0."day" as "day", TA_1."name" as "jvdim name", TA_1."age" as "age", TA_1."sex" as "sex", TA_2."name" as "jvdimp1 name", TA_2."age" as "jvdimp1 age", TA_2."gender" as "gender", TA_2."yr" as "yr"
      FROM "default"."jvfact" TA_0 LEFT OUTER JOIN "default"."jvdim" TA_1 ON ( TA_0."name" = TA_1."name" )  LEFT OUTER JOIN "default"."jvdimp1" TA_2 ON ( TA_0."name" = TA_2."name" )
      LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_0."rownum" as "rownum", TA_0."id" as "id", TA_0."v1" as "v1", TA_0."v2" as "v2", TA_0."name" as "name", TA_0."year" as "year", TA_0."month" as "month", TA_0."day" as "day", TA_1."name" as "jvdim name", TA_1."age" as "age", TA_1."sex" as "sex", TA_2."name" as "jvdimp1 name", TA_2."age" as "jvdimp1 age", TA_2."gender" as "gender", TA_2."yr" as "yr"
    FROM "default"."jvfact"  TA_0  LEFT OUTER JOIN "default"."jvdim" TA_1 ON ( TA_0."name" = TA_1."name" )  LEFT OUTER JOIN "default"."jvdimp1" TA_2 ON ( TA_0."name" = TA_2."name" )
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def multitable_nofactselection1(self):
    # print 'test: multitable using only 1 dim table'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[sex] as 'sex'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}], "limit": 100, "dataset_id": self.jvfactdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_1.`sex` as `sex`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`jvdim` TA_1" + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_1.`sex` as 'sex', sum(1) as 'Record Count' " + \
                      "\nFROM `default`.`jvdim` TA_1" + \
                      "\nGROUP BY 1 " + \
                      "\nLIMIT 100"
    self.expected_outputs["hive"] = "SELECT TA_1.`sex` as `sex`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`jvdim` TA_1" + \
                      "\nGROUP BY TA_1.`sex` " + \
                      "\nLIMIT 100"
    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_1.[sex] as 'sex', sum(1) as 'Record Count'
      FROM [default].[jvdim] TA_1
      GROUP BY TA_1.[sex]
    """
    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_1."sex" as "sex", sum(1) as "Record Count"
      FROM "default"."jvdim" TA_1
      GROUP BY 1
    """
    self.expected_outputs["postgres"] = """
      SELECT TA_1."sex" as "sex", sum(1) as "Record Count"
      FROM "default"."jvdim" TA_1
      GROUP BY 1
      LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_1."sex" as "sex", sum(1) as "Record Count"
    FROM "default"."jvdim"  TA_1
    GROUP BY 1
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def multitable_nofactselection2(self):
    # print 'test: multitable using 2 dims with same parent fact'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[sex] as 'sex'"},
                             {"type": "SIMPLE", "expr": "[gender] as 'gender'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}], "limit": 100, "dataset_id": self.jvfactdata_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_1.`sex` as `sex`, TA_2.`gender` as `gender`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`jvfact` TA_0 LEFT OUTER JOIN `default`.`jvdim` TA_1 " + \
                      "ON ( TA_0.`name` = TA_1.`name` )  LEFT OUTER JOIN `default`.`jvdimp1` TA_2 ON " + \
                      "( TA_0.`name` = TA_2.`name` ) " + \
                      "\nGROUP BY 1, 2 " + \
                      "\nLIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_1.`sex` as 'sex', TA_2.`gender` as 'gender', sum(1) as 'Record Count' " + \
                      "\nFROM `default`.`jvfact` TA_0 LEFT OUTER JOIN `default`.`jvdim` TA_1 " + \
                      "ON ( TA_0.`name` = TA_1.`name` )  LEFT OUTER JOIN `default`.`jvdimp1` TA_2 ON " + \
                      "( TA_0.`name` = TA_2.`name` ) " + \
                      "\nGROUP BY 1, 2 " + \
                      "\nLIMIT 100"
    self.expected_outputs["hive"] = "SELECT TA_1.`sex` as `sex`, TA_2.`gender` as `gender`, sum(1) as `Record Count` " + \
                      "\nFROM `default`.`jvfact` TA_0 LEFT OUTER JOIN `default`.`jvdim` TA_1 " + \
                      "ON ( TA_0.`name` = TA_1.`name` )  LEFT OUTER JOIN `default`.`jvdimp1` TA_2 ON " + \
                      "( TA_0.`name` = TA_2.`name` ) " + \
                      "\nGROUP BY TA_1.`sex`, TA_2.`gender` " + \
                      "\nLIMIT 100"
    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_1.[sex] as 'sex', TA_2.[gender] as 'gender', sum(1) as 'Record Count'
      FROM [default].[jvfact] TA_0 LEFT OUTER JOIN [default].[jvdim] TA_1 ON ( TA_0.[name] = TA_1.[name] )  LEFT OUTER JOIN [default].[jvdimp1] TA_2 ON ( TA_0.[name] = TA_2.[name] )
      GROUP BY TA_1.[sex], TA_2.[gender]
    """
    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_1."sex" as "sex", TA_2."gender" as "gender", sum(1) as "Record Count"
      FROM "default"."jvfact" TA_0 LEFT OUTER JOIN "default"."jvdim" TA_1 ON ( TA_0."name" = TA_1."name" )  LEFT OUTER JOIN "default"."jvdimp1" TA_2 ON ( TA_0."name" = TA_2."name" )
      GROUP BY 1, 2
    """

    self.expected_outputs["postgres"] = """
      SELECT TA_1."sex" as "sex", TA_2."gender" as "gender", sum(1) as "Record Count"
      FROM "default"."jvfact" TA_0 LEFT OUTER JOIN "default"."jvdim" TA_1 ON ( TA_0."name" = TA_1."name" )  LEFT OUTER JOIN "default"."jvdimp1" TA_2 ON ( TA_0."name" = TA_2."name" )
      GROUP BY 1, 2
      LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_1."sex" as "sex", TA_2."gender" as "gender", sum(1) as "Record Count"
    FROM "default"."jvfact"  TA_0  LEFT OUTER JOIN "default"."jvdim" TA_1 ON ( TA_0."name" = TA_1."name" )  LEFT OUTER JOIN "default"."jvdimp1" TA_2 ON ( TA_0."name" = TA_2."name" )
    GROUP BY 1, 2
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def multitable_entity1(self):
    # print 'test: multitable using entity segment with fact and dim columns'
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SEGMENT", "segments": [{"seg_id": 10, "negate": False}]},
                             {"type": "SIMPLE", "expr": "[name] as 'name'"},
                             {"type": "SIMPLE", "expr": "[sex] as 'sex'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}], "limit": 100, "dataset_id": self.jvfactdata_id}
    expected_output = ""

  def multitable_nofactselection3(self):
    # print 'test: multitable using 2 dims with different parent fact'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[jvdim_1 sex] as 'jvdim_1 sex'"},
                             {"type": "SIMPLE", "expr": "[jvdimp1 name] as 'jvdimp1 name'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}], "limit": 100, "dataset_id": self.jvfactdata2_id}

    self.expected_outputs = {}
    self.expected_outputs["default"] = "SELECT TA_3.`sex` as `jvdim_1 sex`, TA_2.`name` as `jvdimp1 name`, " + \
                      "sum(1) as `Record Count` " + \
                      "\nFROM `default`.`jvdimp1` TA_2 " + \
                      "LEFT OUTER JOIN `default`.`jvdim1` TA_3 " + \
                      "ON ( TA_2.`name` = TA_3.`name` ) " + \
                      "\nGROUP BY 1, 2 " + \
                      "\nLIMIT 100"
    self.expected_outputs["mysql"] = "SELECT TA_3.`sex` as 'jvdim_1 sex', TA_2.`name` as 'jvdimp1 name', " + \
                      "sum(1) as 'Record Count' " + \
                      "\nFROM `default`.`jvdimp1` TA_2 " + \
                      "LEFT OUTER JOIN `default`.`jvdim1` TA_3 " + \
                      "ON ( TA_2.`name` = TA_3.`name` ) " + \
                      "\nGROUP BY 1, 2 " + \
                      "\nLIMIT 100"
    self.expected_outputs["hive"] = "SELECT TA_3.`sex` as `jvdim_1 sex`, TA_2.`name` as `jvdimp1 name`, " + \
                      "sum(1) as `Record Count` " + \
                      "\nFROM `default`.`jvdimp1` TA_2 " + \
                      "LEFT OUTER JOIN `default`.`jvdim1` TA_3 " + \
                      "ON ( TA_2.`name` = TA_3.`name` ) " + \
                      "\nGROUP BY TA_3.`sex`, TA_2.`name` " + \
                      "\nLIMIT 100"
    self.expected_outputs["mssqlserver"] = """
    SELECT TOP(100) TA_3.[sex] as 'jvdim_1 sex', TA_2.[name] as 'jvdimp1 name', sum(1) as 'Record Count'
    FROM [default].[jvdimp1] TA_2 LEFT OUTER JOIN [default].[jvdim1] TA_3 ON ( TA_2.[name] = TA_3.[name] )
    GROUP BY TA_3.[sex], TA_2.[name]
    """
    self.expected_outputs["teradata"] = """
    SELECT TOP 100 TA_3."sex" as "jvdim_1 sex", TA_2."name" as "jvdimp1 name", sum(1) as "Record Count"
    FROM "default"."jvdimp1" TA_2 LEFT OUTER JOIN "default"."jvdim1" TA_3 ON ( TA_2."name" = TA_3."name" )
    GROUP BY 1, 2
    """

    self.expected_outputs["postgres"] = """
      SELECT TA_3."sex" as "jvdim_1 sex", TA_2."name" as "jvdimp1 name", sum(1) as "Record Count"
      FROM "default"."jvdimp1" TA_2 LEFT OUTER JOIN "default"."jvdim1" TA_3 ON ( TA_2."name" = TA_3."name" )
      GROUP BY 1, 2
      LIMIT 100
    """

    self.expected_outputs["druid"] = """
    SELECT TA_3."sex" as "jvdim_1 sex", TA_2."name" as "jvdimp1 name", sum(1) as "Record Count"
    FROM "default"."jvdimp1"  TA_2  LEFT OUTER JOIN "default"."jvdim1" TA_3 ON ( TA_2."name" = TA_3."name" )
    GROUP BY 1, 2
    LIMIT 100
    """

    self.validateSQLGeneration(dsjson)

  def multitable_nofactselection4(self):
    if self.dataconnection_type not in ["arcengine", "arcengine_v2"]:
      return

    # print 'test: validate consistent join order with different dim order'
    # json post request object
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[sex] as 'sex'"},
                             {"type": "SIMPLE", "expr": "[jvdimp1 name] as 'jvdimp1 name'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}], "limit": 100, "dataset_id": self.jvfactdata2_id}

    expected_output = "SELECT TA_1.`sex` as `sex`, TA_2.`name` as `jvdimp1 name`, " + \
                      "sum(1) as `Record Count` " + \
                      "\nFROM `default`.`jvfact` TA_0 " + \
                      "LEFT OUTER JOIN `default`.`jvdim` TA_1 " + \
                      "ON ( TA_0.`name` = TA_1.`name`) " + \
                      "LEFT OUTER JOIN `default`.`jvdimp1` TA_2 " + \
                      "ON ( TA_0.`name` = TA_2.`name` ) " + \
                      "\nGROUP BY 1, 2 " + \
                      "\nLIMIT 100"
    dsjson_sw = {"version": 1,
              "type": 'SQL',
              "dimensions": [{"type": "SIMPLE", "expr": "[jvdimp1 name] as 'jvdimp1 name'"},
                             {"type": "SIMPLE", "expr": "[sex] as 'sex'"}],
              "aggregates": [{"expr": "sum(1) as 'Record Count'"}], "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}], "limit": 100, "dataset_id": self.jvfactdata2_id}
    expected_output_sw = "SELECT TA_2.`name` as `jvdimp1 name`, TA_1.`sex` as `sex`, " + \
                      "sum(1) as `Record Count` " + \
                      "\nFROM `default`.`jvfact` TA_0 " + \
                      "LEFT OUTER JOIN `default`.`jvdim` TA_1 " + \
                      "ON ( TA_0.`name` = TA_1.`name`) " + \
                      "LEFT OUTER JOIN `default`.`jvdimp1` TA_2 " + \
                      "ON ( TA_0.`name` = TA_2.`name` ) " + \
                      "\nGROUP BY 1, 2 " + \
                      "\nLIMIT 100"

    self.request.user = self.admin
    sql = self.generateSql(dsjson, request=self.request)
    self.assertStrings(sql, expected_output)
    sql = self.generateSql(dsjson_sw, request=self.request)
    self.assertStrings(sql, expected_output_sw)

  def inapp_purchase_flow1(self):
    if self.dataconnection_type not in ["arcengine", "arcengine_v2"]:
      return

    # print 'test: MR flow for in app events 1 of 3'
    # json post request object
    dsjson = {"version": 1,
              "type": 'MR',
              "time": ["[time]"],
              "session": ["[app_id]", "[user_id]"],
              "dimensions": [],
              "aggregates": [{"expr": "sum(1)"}],
              "filters": ["`month` in (6)"],
              "having": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "events": [{"type": "flow", "events": [{"id": self.event2.event_id, "min": 1, "max": 1, "goal": False},
                                                     {"id": self.event1.event_id, "min": 1, "max": 1, "goal": False}]}],
              "dataset_id": self.mobile_events_id,
              "limit": 20}
    self.request.user = self.admin
    sql = self.generateSql(dsjson, request=self.request)
    expected_output = "SELECT * FROM \n(\n SELECT path , sum(1) AS path_frequency \nFROM " + \
                      "MATCH_RECOGNIZE \n(SELECT TA_0.`event_id_hi` as `event_id_hi`, " + \
                      "TA_0.`event_id_lo` as `event_id_lo`, " + \
                      "TA_0.`app_id` as `app_id`, " + \
                      "TA_0.`app_instance_id` as `app_instance_id`, " + \
                      "TA_0.`name` as `name`, TA_0.`time` as `time`, " + \
                      "TA_0.`user_id` as `user_id`, hex((TA_0.`user_id`)) as `Copy_of_user_id`, " + \
                      "TA_0.`device_id` as `device_id`, hex((TA_0.`device_id`)) as `Copy_of_device_id`, " + \
                      "TA_0.`parent_event_id_hi` as `parent_event_id_hi`, " + \
                      "TA_0.`parent_event_id_lo` as `parent_event_id_lo`, " + \
                      "TA_0.`platform` as `platform`, " + \
                      "TA_0.`sdk_version` as `sdk_version`, " + \
                      "TA_0.`package_name` as `package_name`, " + \
                      "TA_0.`version_name` as `version_name`, " + \
                      "TA_0.`version_code` as `version_code`, " + \
                      "TA_0.`year` as `year`, TA_0.`month` as `month`, " + \
                      "TA_0.`day` as `day`, TA_0.`datestring` as `datestring`" + \
                      "\nFROM `appboy`.`mobile_events` TA_0" + \
                      "\nWHERE (`month` in (6)) \n) \nINPUT " + \
                      "(\nPARTITION BY `app_id` , `user_id` " + \
                      "\nORDER BY `time` " + \
                      "\nMEASURES CONCAT_WS(',' , GROUP_CONCAT(A@name, ',') , 'In App Purchase' ) as path " + \
                      "\nONE_ROW_PER_MATCH \nPATTERN ('A B') " + \
                      "\nDEFINE A as (name like '$%') AND (name != '$session_started') " + \
                      "AND (name != '$session_ended') AND (name != '$iap') , " + \
                      "\nB as (name = '$iap') \n) mr_tmp_1 GROUP BY 1 \n) mr_tmp_2 " + \
                      "\nORDER BY path_frequency DESC \nLIMIT 20"
    self.assertStrings(sql, expected_output)


  def inapp_purchase_flow2(self):
    if self.dataconnection_type not in ["arcengine", "arcengine_v2"]:
      return

    # print 'test: MR flow for in app events 2 of 3'
    # json post request object
    dsjson = {"version": 1,
              "type": 'MR',
              "time": ["[time]"],
              "session": ["[app_id]", "[user_id]"],
              "dimensions": [],
              "aggregates": [{"expr": "sum(1)"}],
              "filters": ["`month` in (6)"],
              "having": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "events": [{"type": "flow", "events": [{"id": self.event2.event_id, "min": 2, "max": 3, "goal": False},
                                                     {"id": self.event1.event_id, "min": 1, "max": 1, "goal": False}]}],
              "dataset_id": self.mobile_events_id,
              "limit": 20}

    self.request.user = self.admin
    sql = self.generateSql(dsjson, request=self.request)
    expected_output = "SELECT * FROM \n(\n SELECT path , sum(1) AS path_frequency \nFROM " + \
                      "MATCH_RECOGNIZE \n(SELECT TA_0.`event_id_hi` as `event_id_hi`, " + \
                      "TA_0.`event_id_lo` as `event_id_lo`, " + \
                      "TA_0.`app_id` as `app_id`, " + \
                      "TA_0.`app_instance_id` as `app_instance_id`, " + \
                      "TA_0.`name` as `name`, TA_0.`time` as `time`, " + \
                      "TA_0.`user_id` as `user_id`, hex((TA_0.`user_id`)) as `Copy_of_user_id`, " + \
                      "TA_0.`device_id` as `device_id`, hex((TA_0.`device_id`)) as `Copy_of_device_id`, " + \
                      "TA_0.`parent_event_id_hi` as `parent_event_id_hi`, " + \
                      "TA_0.`parent_event_id_lo` as `parent_event_id_lo`, " + \
                      "TA_0.`platform` as `platform`, " + \
                      "TA_0.`sdk_version` as `sdk_version`, " + \
                      "TA_0.`package_name` as `package_name`, " + \
                      "TA_0.`version_name` as `version_name`, " + \
                      "TA_0.`version_code` as `version_code`, " + \
                      "TA_0.`year` as `year`, TA_0.`month` as `month`, " + \
                      "TA_0.`day` as `day`, TA_0.`datestring` as `datestring`" + \
                      "\nFROM `appboy`.`mobile_events` TA_0" + \
                      "\nWHERE (`month` in (6)) \n) \nINPUT " + \
                      "(\nPARTITION BY `app_id` , `user_id` " + \
                      "\nORDER BY `time` " + \
                      "\nMEASURES CONCAT_WS(',' , GROUP_CONCAT(A@name, ',') , 'In App Purchase' ) as path " + \
                      "\nONE_ROW_PER_MATCH \nPATTERN ('A{2,3} B') " + \
                      "\nDEFINE A as (name like '$%') AND (name != '$session_started') " + \
                      "AND (name != '$session_ended') AND (name != '$iap') , " + \
                      "\nB as (name = '$iap') \n) mr_tmp_1 GROUP BY 1 \n) mr_tmp_2 " + \
                      "\nORDER BY path_frequency DESC \nLIMIT 20"
    self.assertStrings(sql, expected_output)

  def inapp_purchase_flow3(self):
    if self.dataconnection_type not in ["arcengine", "arcengine_v2"]:
      return

    # print 'test: MR flow for in app events 3 of 3'
    dsjson = {"version": 1, "type": "MR", "time": ["[time]"], "session": ["[app_id]", "[user_id]"], "dimensions": [],
              "aggregates": [{"expr": "sum(1)"}],
              "filters": ["[month] in (6)", "name like '$%'", "name not in ('$session_started', '$session_ended')"],
              "having": [], "global_segments": [
      {"type": "SEGMENT", "segments": [{"seg_id": self.segment7.segment_id, "negate": False}]}],
              "events": [{"type": "flow", "events": [{"id": self.event3.event_id, "min": 1, "max": 1, "goal": False},
                                                     {"id": self.event1.event_id, "min": 1, "max": 1, "goal": True}]}],
              "distinct": False,
              "dataset_id": self.mobile_events_id, "limit": 30}

    self.request.user = self.admin
    sql = self.generateSql(dsjson, request=self.request)
    expected_output = "SELECT * FROM \n(\n SELECT path , sum(1) AS path_frequency \nFROM " + \
                      "MATCH_RECOGNIZE \n(SELECT TA_0.`event_id_hi` as `event_id_hi`, " + \
                      "TA_0.`event_id_lo` as `event_id_lo`, " + \
                      "TA_0.`app_id` as `app_id`, " + \
                      "TA_0.`app_instance_id` as `app_instance_id`, " + \
                      "TA_0.`name` as `name`, TA_0.`time` as `time`, " + \
                      "TA_0.`user_id` as `user_id`, hex((TA_0.`user_id`)) as `Copy_of_user_id`, " + \
                      "TA_0.`device_id` as `device_id`, hex((TA_0.`device_id`)) as `Copy_of_device_id`, " + \
                      "TA_0.`parent_event_id_hi` as `parent_event_id_hi`, " + \
                      "TA_0.`parent_event_id_lo` as `parent_event_id_lo`, " + \
                      "TA_0.`platform` as `platform`, " + \
                      "TA_0.`sdk_version` as `sdk_version`, " + \
                      "TA_0.`package_name` as `package_name`, " + \
                      "TA_0.`version_name` as `version_name`, " + \
                      "TA_0.`version_code` as `version_code`, " + \
                      "TA_0.`year` as `year`, TA_0.`month` as `month`, " + \
                      "TA_0.`day` as `day`, TA_0.`datestring` as `datestring`" + \
                      "\nFROM `appboy`.`mobile_events` TA_0" + \
                      "\nWHERE (TA_0.`month` in (6)) AND " + \
                      "(name like '$%') AND (name not in ('$session_started', '$session_ended')) " + \
                      "AND ((app_id='53163ecd0daac1a107000061')) \n) \nINPUT " + \
                      "(\nPARTITION BY `app_id` , `user_id` " + \
                      "\nORDER BY `time` " + \
                      "\nMEASURES CONCAT_WS(',' , GROUP_CONCAT(A@name, ',') , 'In App Purchase' ) as path " + \
                      "\nONE_ROW_PER_MATCH \nPATTERN ('A B') " + \
                      "\nDEFINE A as (name like '$%'), " + \
                      "\nB as (name = '$iap') \n) mr_tmp_1 GROUP BY 1 \n) mr_tmp_2 " + \
                      "\nORDER BY path_frequency DESC \nLIMIT 30"
    self.assertStrings(sql, expected_output)


  def inapp_purchase_funnel1(self):
    if self.dataconnection_type not in ["arcengine", "arcengine_v2"]:
      return

    # print 'test: MR funnel for in app events 1 of 2'
    # json post request object
    dsjson = {"version": 1,
              "type": 'MR',
              "time": ["[time]"],
              "session": ["[app_id]", "[user_id]"],
              "dimensions": [],
              "aggregates": [{"expr": "sum(1)"}],
              "filters": ["`month` in (6)"],
              "having": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "events": [{"type": "funnel", "events": [{"id": self.event2.event_id, "min": 1, "max": 1, "goal": False},
                                                       {"id": self.event1.event_id, "min": 1, "max": 1,
                                                        "goal": False}]}],
              "dataset_id": self.mobile_events_id,
              "limit": 20}

    self.request.user = self.admin
    sql = self.generateSql(dsjson, request=self.request)
    expected_output = "SELECT * FROM \n(\n SELECT path, arc_path_priority, sum(1) AS path_frequency \nFROM " + \
                      "MATCH_RECOGNIZE \n(SELECT TA_0.`event_id_hi` as `event_id_hi`, " + \
                      "TA_0.`event_id_lo` as `event_id_lo`, " + \
                      "TA_0.`app_id` as `app_id`, " + \
                      "TA_0.`app_instance_id` as `app_instance_id`, " + \
                      "TA_0.`name` as `name`, TA_0.`time` as `time`, " + \
                      "TA_0.`user_id` as `user_id`, hex((TA_0.`user_id`)) as `Copy_of_user_id`, " + \
                      "TA_0.`device_id` as `device_id`, hex((TA_0.`device_id`)) as `Copy_of_device_id`, " + \
                      "TA_0.`parent_event_id_hi` as `parent_event_id_hi`, " + \
                      "TA_0.`parent_event_id_lo` as `parent_event_id_lo`, " + \
                      "TA_0.`platform` as `platform`, " + \
                      "TA_0.`sdk_version` as `sdk_version`, " + \
                      "TA_0.`package_name` as `package_name`, " + \
                      "TA_0.`version_name` as `version_name`, " + \
                      "TA_0.`version_code` as `version_code`, " + \
                      "TA_0.`year` as `year`, TA_0.`month` as `month`, " + \
                      "TA_0.`day` as `day`, TA_0.`datestring` as `datestring`" + \
                      "\nFROM `appboy`.`mobile_events` TA_0" + \
                      "\nWHERE (`month` in (6)) \n) \nINPUT " + \
                      "(\nPARTITION BY `app_id` , `user_id` " + \
                      "\nORDER BY `time` " + \
                      "\nMEASURES case \nwhen (B@`app_id` is null) " + \
                      "then concat_ws(',', concat_ws('', 'ARC:FUNNEL:', A@name) , 'Drop Off' ) " + \
                      "\nelse concat_ws(',', concat_ws('', 'ARC:FUNNEL:', A@name) , " + \
                      "concat_ws('', 'ARC:FUNNEL:', 'In App Purchase') ) end as path, " + \
                      "case \nwhen (B@`app_id` is null) then 2 else 1 end as arc_path_priority " + \
                      "\nONE_ROW_PER_MATCH \nPATTERN  ('^ZA*? A+ ZB* B*') " + \
                      "\nDEFINE A as (name like '$%') AND (name != '$session_started') " + \
                      "AND (name != '$session_ended') AND (name != '$iap'), " + \
                      "\nZA as (!((name like '$%') AND (name != '$session_started') AND " + \
                      "(name != '$session_ended') AND (name != '$iap'))) , " + \
                      "\nB as (name = '$iap'), " + \
                      "\nZB as (!((name = '$iap'))) " + \
                      "\n) mr_tmp_1 GROUP BY 1 , 2 \n) mr_tmp_2 " + \
                      "\nORDER BY path_frequency DESC \nLIMIT 20"
    self.assertStrings(sql, expected_output)

  def SlideUpFunnel(self):
    if self.dataconnection_type not in ["arcengine", "arcengine_v2"]:
      return

    # print 'test: MR funnel for SlideUp Impressions'
    # json post request object
    dsjson = {"version": 1, "type": "MR", "time": ["[time]"], "session": ["[app_id]", "[user_id]"], "dimensions": [],
              "aggregates": [{"expr": "sum(1)"}], "filters": ["app_id = '52726ac288282d045b00002a'"], "having": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "events": [{"type": "funnel",
                          "events": [
                            {"id": self.event4.event_id, "min": 1, "max": 1, "goal": False},
                            {"id": self.event5.event_id, "min": 1, "max": 1, "goal": False},
                            {"id": self.event6.event_id, "min": 1, "max": 1, "goal": False}]}],
              "distinct": False, "dataset_id": self.mobile_events_id, "limit": 30}
    self.request.user = self.admin
    sql = self.generateSql(dsjson, request=self.request)
    expected_output = "SELECT * FROM \n(SELECT path, arc_path_priority, sum(1) AS path_frequency" + \
                      "\nFROM MATCH_RECOGNIZE \n(SELECT TA_0.`event_id_hi` as `event_id_hi`, " + \
                      "TA_0.`event_id_lo` as `event_id_lo`, " + \
                      "TA_0.`app_id` as `app_id`, " + \
                      "TA_0.`app_instance_id` as `app_instance_id`, " + \
                      "TA_0.`name` as `name`, TA_0.`time` as `time`, " + \
                      "TA_0.`user_id` as `user_id`, hex((TA_0.`user_id`)) as `Copy_of_user_id`, " + \
                      "TA_0.`device_id` as `device_id`, hex((TA_0.`device_id`)) as `Copy_of_device_id`, " + \
                      "TA_0.`parent_event_id_hi` as `parent_event_id_hi`, " + \
                      "TA_0.`parent_event_id_lo` as `parent_event_id_lo`, " + \
                      "TA_0.`platform` as `platform`, " + \
                      "TA_0.`sdk_version` as `sdk_version`, " + \
                      "TA_0.`package_name` as `package_name`, " + \
                      "TA_0.`version_name` as `version_name`, " + \
                      "TA_0.`version_code` as `version_code`, " + \
                      "TA_0.`year` as `year`, TA_0.`month` as `month`, " + \
                      "TA_0.`day` as `day`, TA_0.`datestring` as `datestring`" + \
                      "\nFROM `appboy`.`mobile_events` TA_0" + \
                      "\nWHERE (app_id = '52726ac288282d045b00002a')" + \
	                    "\n) \nINPUT (" + \
	                    "\nPARTITION BY `app_id`, `user_id`" + \
	                    "\nORDER BY `time`" + \
	                    "\nMEASURES  case" + \
	                    "\nwhen (B@`app_id` is null) then  " + \
                      "concat_ws(',', concat_ws('', 'ARC:FUNNEL:', 'SlideUp Impression'), 'Drop Off' ) " + \
	                    "\nwhen (C@`app_id` is null) then  " + \
                      "concat_ws(',', concat_ws('', 'ARC:FUNNEL:', 'SlideUp Impression'), " + \
                      "concat_ws('', 'ARC:FUNNEL:', 'SlideUp Clicked'), 'Drop Off' ) " + \
	                    "\nelse  concat_ws(',', concat_ws('', 'ARC:FUNNEL:', 'SlideUp Impression'), " + \
                      "concat_ws('', 'ARC:FUNNEL:', 'SlideUp Clicked'), " + \
                      "concat_ws('', 'ARC:FUNNEL:', 'Session Started')) end as path, " + \
	                    "\ncase \nwhen (C@`app_id` is null) then 2 else 1 end as arc_path_priority" + \
	                    "\nONE_ROW_PER_MATCH" + \
	                    "\nPATTERN ('^ZA*? A+ ZB* B* ZC* C*')" + \
	                    "\nDEFINE A as (name = '$slideup_impression'), " + \
	                    "\nZA as (!((name = '$slideup_impression'))), " + \
	                    "\nB as (name = '$slideup_clicked'), " + \
	                    "\nZB as (!((name = '$slideup_clicked'))), " + \
	                    "\nC as (name = '$session_started'), " + \
	                    "\nZC as (!((name = '$session_started'))) " + \
	                    "\n) mr_tmp_1 GROUP BY 1, 2 " + \
                      "\n) mr_tmp_2 " + \
                      "\nORDER BY path_frequency DESC " + \
                      "\nLIMIT 30"
    self.assertStrings(sql, expected_output)

class Test_BuildSql_ArcEngine(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'arcengine'
    super(Test_BuildSql_ArcEngine, self).__init__(*args, **kwargs)

class Test_BuildSql_ArcEngineV2(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'arcengine_v2'
    super(Test_BuildSql_ArcEngineV2, self).__init__(*args, **kwargs)

class Test_BuildSql_Impala(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'impala'
    super(Test_BuildSql_Impala, self).__init__(*args, **kwargs)

class Test_BuildSql_Impyla(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'impyla'
    super(Test_BuildSql_Impyla, self).__init__(*args, **kwargs)

class Test_BuildSql_Hive(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'hive'
    super(Test_BuildSql_Hive, self).__init__(*args, **kwargs)

class Test_BuildSql_Hyve(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'hyve'
    super(Test_BuildSql_Hyve, self).__init__(*args, **kwargs)

class Test_BuildSql_Sqlite(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'sqlite'
    super(Test_BuildSql_Sqlite, self).__init__(*args, **kwargs)

class Test_BuildSql_Mysql(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'mysql'
    super(Test_BuildSql_Mysql, self).__init__(*args, **kwargs)

  def test_cast_syntax(self):
    dsjson = {"version": 1,
              "type": 'SQL',
              "dimensions": [],
              "aggregates": [],
              "filters": [],
              "global_segments": [{"type": "SEGMENT", "segments": []}],
              "limit": 100,
              "dataset_id": self.censusdataCast2_id}

    self.request.user = self.admin
    sql = self.generateSql(dsjson, request=self.request)
    spaces_after_CAST = re.compile('CAST( *)\(')
    spaces = spaces_after_CAST.findall(sql)

    # there must be no spaces after CAST:
    if len(''.join(spaces)):
      self.fail('MySQL does not tolerate spaces after CAST: '+sql)

    expected_sql = '''
      SELECT
        (CAST(TA_0.`year` as CHAR)) as 'year',
        (CAST(TA_0.`field_1` as BOOLEAN)) as 'bool field',
        (CAST(TA_0.`field_2` as DATETIME)) as 'time field',
        (CAST(TA_0.`field_3` as SIGNED INTEGER)) as 'integer field',
        (CAST(TA_0.`field_4` as REAL)) as 'float field'
        FROM `default`.`census_pop`  TA_0
        LIMIT 100
    '''
    self.assertStrings(sql, expected_sql)


class Test_BuildSql_MariaDB(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'mariadb'
    super(Test_BuildSql_MariaDB, self).__init__(*args, **kwargs)

class Test_BuildSql_Postgres(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'postgres'
    super(Test_BuildSql_Postgres, self).__init__(*args, **kwargs)

class Test_BuildSql_Redshift(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'redshift'
    super(Test_BuildSql_Redshift, self).__init__(*args, **kwargs)

class Test_BuildSql_Mssql(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'mssqlserver'
    super(Test_BuildSql_Mssql, self).__init__(*args, **kwargs)

class Test_BuildSql_Teradata(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'teradata'
    super(Test_BuildSql_Teradata, self).__init__(*args, **kwargs)

class Test_BuildSql_Snowflake(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'snowflake'
    super(Test_BuildSql_Snowflake, self).__init__(*args, **kwargs)

class Test_BuildSql_Druid(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'druid'
    super(Test_BuildSql_Druid, self).__init__(*args, **kwargs)

class Test_BuildSql_duckdb(TestBuild_sql):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'duckdb'
    super(Test_BuildSql_duckdb, self).__init__(*args, **kwargs)

