# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
These tests verify the arcengine_v2 DataEngine interface.

We mock out the cursor into the backend, and verify that the
appropriate SQL command is sent it by the backend for SHOW_DATABASES,
SHOW_TABLES, table_columns(), and simple SQL execution. Note that
the actual SQL-generation tests are elsewhere (in test_build_sql.py)
"""

from dataconn import dataconn_bigquery

from dataconn.dataconn import get_dataconn, get_django_dataconn
from django.conf import settings
from unittest import mock
from vizlib.querylog_helper import ComponentType
from vizlib.tests.test_querylog import get_logged_field

from . import test_base_dataconn
from .test_base_dataconn import MockCursor, MockBuildRequest

import json


class Test_DataConnection(test_base_dataconn.Test_Base_Data_Connection):
  def __init__(self, *args, **kwargs):
    super(Test_DataConnection, self).__init__(*args, **kwargs)


class Test_DataConnection_BigQuery(Test_DataConnection):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'bigquery'
    Test_DataConnection.__init__(self, *args, **kwargs)

  def setUp(self):
    super().setUp()

    # bigquery is a package we do not have in requirements.txt. Mock it:
    self.bigquery = getattr(dataconn_bigquery, 'bigquery', None)
    if self.bigquery is None:
      dataconn_bigquery.bigquery = mock.MagicMock()

  def tearDown(self):
    # remove the mocked bigquery:
    if self.bigquery is None:
      del dataconn_bigquery.bigquery
    super().tearDown()

  def init_mockcursor(self):
    # we need to add the PASSWORD to the DB settings first:
    django_dc = get_django_dataconn(self.dc.id)
    dc_info = json.loads(django_dc.dataconnection_info)
    dc_info['PARAMS']['PASSWORD'] = 'test_password'
    django_dc.dataconnection_info = json.dumps(dc_info)
    django_dc.save()

    self.dataconn = get_dataconn(self.dc.id)
    self.mock_cursor = MockCursor(self)
    self.dataconn.cursor = self.mock_cursor
    self.build_request = MockBuildRequest()
    self.build_request.username = 'abc'
    self.build_request.connection = self.dataconn

  @mock.patch.object(settings, 'AUDIT_LOGGER')
  def test_audit_logging(self, logger: mock.Mock):
    self.init_mockcursor()

    logger.info.reset_mock()
    # _runq_logged is a private method that is never used
    self.dataconn._runq_logged('select * from main.table_1')

    self.assertEquals(get_logged_field(logger.info, 'query'),
                      'select * from main.table_1')
    self.assertEquals(get_logged_field(
      logger.info, 'component_type'), ComponentType.DATACONNECTION)
