# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
These tests verify the druid DataEngine interface.

We mock out the cursor into the backend, and verify that the
appropriate SQL command is sent it by the backend for SHOW_DATABASES,
SHOW_TABLES, table_columns(), and simple SQL execution. Note that
the actual SQL-generation tests are elsewhere (in test_build_sql.py)
"""

import json
import collections

from dataconn.data_api_cmds import SPECIAL_CMDS
from dataconn.dataconn import get_dataconn
from util.json_utils import json_dumps

from . import test_base_dataconn
from .test_base_dataconn import MockCursor, MockBuildRequest, EXPECTED_DATA_LENGTH, EXPECTED_DATA_WITH_PARTITIONS_LENGTH

class Test_DataConnection(test_base_dataconn.Test_Base_Data_Connection):
  def __init__(self, *args, **kwargs):
    super(Test_DataConnection, self).__init__(*args, **kwargs)

class Test_DataConnection_Druid(Test_DataConnection):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'druid'
    Test_DataConnection.__init__(self, *args, **kwargs)

  def init_mockcursor(self):
    self.dataconn = get_dataconn(self.dc.id)
    self.mock_cursor = MockCursor(self)
    self.dataconn.cursor = self.mock_cursor
    self.build_request = MockBuildRequest()
    self.build_request.username = 'abc'
    self.build_request.connection = self.dataconn

  def test_databases(self):
    self.init_mockcursor()

    Column = collections.namedtuple('Column', ['name', 'type_code', 'display_size', 'internal_size', 'precision', 'scale', 'null_ok'])
    t = Column('table_schema', 1043, None, 65535, None, None, None)

    self.mock_cursor.add_description((t,))
    self.mock_cursor.add_query("""
    SELECT
          distinct TABLE_SCHEMA
        FROM INFORMATION_SCHEMA.TABLES
        WHERE
          TABLE_SCHEMA != 'INFORMATION_SCHEMA'
          AND
          TABLE_SCHEMA != 'sys'
        ORDER BY 1
    """)
    result = [
        ['public'],
        ['prod']
    ]
    result_copy = result[:]
    self.mock_cursor.add_results(result)

    self.build_request.query = SPECIAL_CMDS['SHOW_DATABASES']
    out = self.dataconn.execute(self.build_request)
    self.assertEqual(len(out), EXPECTED_DATA_LENGTH)
    self.assertTrue('coltypes' in out)
    self.assertEqual(out['coltypes'], ["STRING"])

    self.assertTrue('colnames' in out)
    self.assertEqual(out['colnames'], ["table_schema"])

    self.assertTrue('rows' in out)
    self.assertEqual(out['rows'], json_dumps([["public"], ["prod"]]))

  def test_tables(self):
    self.init_mockcursor()

    self.mock_cursor.add_description(
        (('name', None, None, None, None, None, None),))
    self.mock_cursor.add_query(
        """
        SELECT
          distinct TABLE_NAME
        FROM INFORMATION_SCHEMA.TABLES
        WHERE
          TABLE_SCHEMA = 'main'
        ORDER BY 1
        """)
    self.mock_cursor.add_results([
        ['census_pop'],
        ['cereals'],
        ['gh'],
        ['iris'],
        ['masoud'],
        ['trips'],
        ['trips_detail'],
        ['us_counties'],
        ['world_life_expectancy']
    ])

    self.build_request.query = SPECIAL_CMDS['SHOW_TABLES'] + " main"
    out = self.dataconn.execute(self.build_request)
    self.assertEqual(len(out), EXPECTED_DATA_LENGTH)
    self.assertTrue('coltypes' in out)
    self.assertEqual(out['coltypes'], ["STRING"])

    self.assertTrue('colnames' in out)
    self.assertEqual(out['colnames'], ["name"])

    self.assertTrue('rows' in out)
    self.assertEqual(out['rows'],
        json_dumps([['census_pop'], ['cereals'], ['gh'], ['iris'], ['masoud'],
         ['trips'], ['trips_detail'], ['us_counties'],
         ['world_life_expectancy']]))

  def test_table_columns(self):
    self.init_mockcursor()

    self.mock_cursor.add_description(
        (
          ('id', 2, None, None, None, None, None),
          ('sepal_length', 3, None, None, None, None, None),
          ('species', 1, None, None, None, None, None),
        ))
    self.mock_cursor.add_query('select * from "prod"."events" limit 1')
    self.mock_cursor.add_results([])

    out = self.dataconn.table_columns('prod.events')
    self.assertEqual(len(out), EXPECTED_DATA_WITH_PARTITIONS_LENGTH)
    self.assertTrue('coltypes' in out)
    self.assertEqual(out['coltypes'], ["BIGINT", "DOUBLE", "STRING"])

    self.assertTrue('colnames' in out)
    self.assertEqual(out['colnames'], ["id", "sepal_length", "species"])

    self.assertTrue('partitions' in out)
    self.assertEqual(out['partitions'], [])

  def test_query_execution(self):
    self.init_mockcursor()

    self.mock_cursor.add_description((
        ('species', None, None, None, None, None, None),
        ('Record Count', None, None, None, None, None, None)))
    query = 'select species, sum(1) as "Record Count" from iris group by 1'
    self.mock_cursor.add_query(query)
    result = [
      ['setosa', 30],
      ['versicolor', 40],
      ['virginica', 50]
    ]
    result_copy = result[:]
    self.mock_cursor.add_results(result)

    self.build_request.query = query
    out = self.dataconn.execute(self.build_request)
    self.assertEqual(len(out), EXPECTED_DATA_LENGTH)
    self.assertTrue('coltypes' in out)
    self.assertEqual(out['coltypes'], ["STRING", "BIGINT"])

    self.assertTrue('colnames' in out)
    self.assertEqual(out['colnames'], ["species", "Record Count"])

    self.assertTrue('rows' in out)
    self.assertEqual(out['rows'], json_dumps(result_copy))

