# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
These tests verify the redshift DataEngine interface.

We mock out the cursor into the backend, and verify that the
appropriate SQL command is sent it by the backend for SHOW_DATABASES,
SHOW_TABLES, table_columns(), and simple SQL execution. Note that
the actual SQL-generation tests are elsewhere (in test_build_sql.py)
"""

import json
import collections

from dataconn.data_api_cmds import SPECIAL_CMDS
from dataconn.dataconn import get_dataconn
from util.json_utils import json_dumps

from . import test_base_dataconn
from .test_base_dataconn import MockCursor, MockBuildRequest, EXPECTED_DATA_LENGTH, EXPECTED_DATA_WITH_PARTITIONS_LENGTH

class Test_DataConnection(test_base_dataconn.Test_Base_Data_Connection):
  def __init__(self, *args, **kwargs):
    super(Test_DataConnection, self).__init__(*args, **kwargs)

class Test_DataConnection_Redshift(Test_DataConnection):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'redshift'
    Test_DataConnection.__init__(self, *args, **kwargs)

  def init_mockcursor(self):
    self.dataconn = get_dataconn(self.dc.id)
    self.mock_cursor = MockCursor(self)
    self.dataconn.cursor = self.mock_cursor
    self.build_request = MockBuildRequest()
    self.build_request.username = 'abc'
    self.build_request.connection = self.dataconn
    self.build_request.sample_pct = 40
    # The comment is empty because we don't support comments for this dc
    self.build_request.comment = ''

  def test_databases(self):
    self.init_mockcursor()

    Column = collections.namedtuple('Column', ['name', 'type_code', 'display_size', 'internal_size', 'precision', 'scale', 'null_ok'])
    t = Column('table_schema', 1043, None, 65535, None, None, None)

    self.mock_cursor.add_description((t,))
    self.mock_cursor.add_query(
        """SELECT distinct table_schema FROM information_schema.tables WHERE table_schema NOT IN ('pg_catalog', 'information_schema')""")
    result = [
        ['public'],
        ['prod']
    ]
    result_copy = result[:]
    self.mock_cursor.add_results(result)
    self.build_request.query = SPECIAL_CMDS['SHOW_DATABASES']

    out = self.dataconn.execute(self.build_request)

    self.assertEqual(len(out), EXPECTED_DATA_LENGTH)
    self.assertTrue('coltypes' in out)
    self.assertEqual(out['coltypes'], ["STRING"])

    self.assertTrue('colnames' in out)
    self.assertEqual(out['colnames'], ["table_schema"])

    self.assertTrue('rows' in out)
    self.assertEqual(out['rows'], json_dumps(result_copy))

  def test_tables(self):
    self.init_mockcursor()

    Column = collections.namedtuple('Column', ['name', 'type_code', 'display_size', 'internal_size', 'precision', 'scale', 'null_ok'])
    t = Column('table_name', 1043, None, 65535, None, None, None)
    self.mock_cursor.add_description((t,))
    self.mock_cursor.add_query(
        """SELECT table_name FROM information_schema.tables WHERE table_schema IN ('public')""")
    result = [
        ['census_pop'],
        ['testdb']
    ]
    self.mock_cursor.add_results(result)
    result_copy = result[:]

    self.build_request.query = SPECIAL_CMDS['SHOW_TABLES'] + " public"
    out = self.dataconn.execute(self.build_request)

    self.assertEqual(len(out), EXPECTED_DATA_LENGTH)
    self.assertTrue('coltypes' in out)
    self.assertEqual(out['coltypes'], ["STRING"])

    self.assertTrue('colnames' in out)
    self.assertEqual(out['colnames'], ["table_name"])

    self.assertTrue('rows' in out)
    self.assertEqual(out['rows'], json_dumps(result_copy))

  def test_table_columns(self):
    self.init_mockcursor()

    Column = collections.namedtuple('Column', ['name', 'type_code', 'display_size', 'internal_size', 'precision', 'scale', 'null_ok'])
    self.mock_cursor.add_description((
      Column('salesid', 23, None, 4, None, None, None),
      Column('listid', 23, None, 4, None, None, None),
      Column('qtysold', 21, None, 4, None, None, None),
      Column('pricepaid', 1700, None, 4, None, None, None),
      Column('price', 700, None, 4, None, None, None),
      Column('saletime', 1114, None, 4, None, None, None),
      Column('promotion', 16, None, 4, None, None, None),
    ))
    self.mock_cursor.add_query('select * from "public"."sales" limit 0')
    self.mock_cursor.add_results([])

    out = self.dataconn.table_columns('public.sales')

    self.assertEqual(len(out), EXPECTED_DATA_WITH_PARTITIONS_LENGTH)

    self.assertTrue('coltypes' in out)
    self.assertEqual(out['coltypes'], ["BIGINT", "BIGINT", "BIGINT", "STRING", "DOUBLE", "TIMESTAMP", "BOOLEAN"])

    self.assertTrue('colnames' in out)
    self.assertEqual(out['colnames'], ['salesid', 'listid', 'qtysold', 'pricepaid', 'price', 'saletime', 'promotion'])

    self.assertTrue('partitions' in out)
    self.assertEqual(out['partitions'], [])

    self.assertTrue('rows' in out)
    self.assertEqual(out['rows'], json_dumps([]))

  def test_query_execution(self):
    self.init_mockcursor()

    Column = collections.namedtuple('Column', ['name', 'type_code', 'display_size', 'internal_size', 'precision', 'scale', 'null_ok'])
    self.mock_cursor.add_description((
      Column('state', 1043, None, 256, None, None, None),
      Column('record count', 20, None, 8, None, None, None)
    ))
    query = "select state, sum(1) as 'Record Count' from public.census_pop group by 1"
    self.mock_cursor.add_query(query)
    result = [
      ['CA', 30],
      ['AL', 40],
      ['AZ', 50]
    ]
    result_copy = result[:]
    self.mock_cursor.add_results(result)

    self.build_request.query = query
    out = self.dataconn.execute(self.build_request)
    self.assertEqual(len(out), EXPECTED_DATA_LENGTH)
    self.assertTrue('coltypes' in out)
    self.assertEqual(out['coltypes'], ["STRING", "BIGINT"])

    self.assertTrue('colnames' in out)
    self.assertEqual(out['colnames'], ["state", "record count"])

    self.assertTrue('rows' in out)
    self.assertEqual(out['rows'], json_dumps(result_copy))

