# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
These tests verify the Teradata DataEngine interface.

We mock out the cursor into the backend, and verify that the
appropriate SQL command is sent it by the backend for SHOW_DATABASES,
SHOW_TABLES, table_columns(), and simple SQL execution. Note that
the actual SQL-generation tests are elsewhere (in test_build_sql.py)
"""

import json

from dataconn.data_api_cmds import SPECIAL_CMDS
from dataconn.dataconn import get_dataconn
from util.json_utils import json_dumps

from . import test_base_dataconn
from .test_base_dataconn import MockCursor, MockBuildRequest, EXPECTED_DATA_LENGTH

class Test_DataConnection(test_base_dataconn.Test_Base_Data_Connection):
  def __init__(self, *args, **kwargs):
    super(Test_DataConnection, self).__init__(*args, **kwargs)

class Test_DataConnection_Teradata(Test_DataConnection):
  def __init__(self, *args, **kwargs):
    self.dataconnection_type = 'teradata'
    Test_DataConnection.__init__(self, *args, **kwargs)

  def init_mockcursor(self):
    self.dataconn = get_dataconn(self.dc.id)
    self.mock_cursor = MockCursor(self)
    self.dataconn.cursor = self.mock_cursor
    self.build_request = MockBuildRequest()
    self.build_request.username = 'abc'
    self.build_request.connection = self.dataconn
    self.build_request.sample_pct = 40
    # The comment is empty because we don't support comments for this dc
    self.build_request.comment = ''

  def test_databases(self):
    self.init_mockcursor()

    self.mock_cursor.add_description(
        [('name', type(''), None, None, None, None, 0, 1)])
    self.mock_cursor.add_query(
        """SELECT DISTINCT "databasename" FROM dbc.tablesv order by databasename asc""")
    result = [
        ['information_schema'],
        ['default'],
        ['larry'],
        ['mysql'],
        ['sushil'],
        ['test']
    ]
    result_copy = result[:]
    self.mock_cursor.add_results(result)
    self.build_request.query = SPECIAL_CMDS['SHOW_DATABASES']

    d = self.dataconn.execute(self.build_request)

    self.assertEqual(len(d), EXPECTED_DATA_LENGTH)
    self.assertTrue('coltypes' in d)
    self.assertEqual(d['coltypes'], ["STRING"])

    self.assertTrue('colnames' in d)
    self.assertEqual(d['colnames'], ["name"])

    self.assertTrue('rows' in d)
    self.assertEqual(d['rows'], json_dumps(result_copy))

  def test_tables(self):
    self.init_mockcursor()

    self.mock_cursor.add_description(
        [('Tables_in_larry', type(''), None, None, None, None, 0, 1)])
    self.mock_cursor.add_query(
        """SELECT tablename FROM dbc.tablesv where tablekind in ('T', 'V') and databasename='larry' order by tablename asc""")
    result = [
        ['census_pop'],
        ['testdb']
    ]
    self.mock_cursor.add_results(result)
    result_copy = result[:]

    self.build_request.query = SPECIAL_CMDS['SHOW_TABLES'] + " larry"
    d = self.dataconn.execute(self.build_request)

    self.assertEqual(len(d), EXPECTED_DATA_LENGTH)
    self.assertTrue('coltypes' in d)
    self.assertEqual(d['coltypes'], ["STRING"])

    self.assertTrue('colnames' in d)
    self.assertEqual(d['colnames'], ["Tables_in_larry"])

    self.assertTrue('rows' in d)
    self.assertEqual(d['rows'], json_dumps(result_copy))

  def test_table_columns(self):
    self.init_mockcursor()

    self.mock_cursor.add_description(
        [('year', type(''), None, None, None, None, 1, 16),
         ('state', type(''), None, None, None, None, 1, 16),
         ('population', type(3.2), None, None, None, None, 1, 16)])
    self.mock_cursor.add_query('select * from "larry"."census_pop" sample 0')
    self.mock_cursor.add_results([])

    d = self.dataconn.table_columns('larry.census_pop')

    self.assertEqual(len(d), EXPECTED_DATA_LENGTH +1 )

    self.assertTrue('coltypes' in d)
    self.assertEqual(d['coltypes'], ["STRING", "STRING", "REAL"])

    self.assertTrue('colnames' in d)
    self.assertEqual(d['colnames'], ["year", "state", "population"])

    self.assertTrue('partitions' in d)
    self.assertEqual(d['partitions'], [])

    self.assertTrue('rows' in d)
    self.assertEqual(d['rows'], json_dumps([]))

  def test_query_execution(self):
    self.init_mockcursor()

    self.mock_cursor.add_description([
        ('state', type(''), None, None, None, None, 1, 32784),
        ('Record Count', type(1), None, None, None, None, 1, 0)])
    query = "select state, sum(1) as 'Record Count' from larry.census_pop group by 1"
    self.mock_cursor.add_query(query)
    result = [
      ['CA', 30],
      ['AL', 40],
      ['AZ', 50]
    ]
    result_copy = result[:]
    self.mock_cursor.add_results(result)

    self.build_request.query = query
    d = self.dataconn.execute(self.build_request)
    print(d)
    self.assertEqual(len(d), EXPECTED_DATA_LENGTH)
    self.assertTrue('coltypes' in d)
    self.assertEqual(d['coltypes'], ["STRING", "INT"])

    self.assertTrue('colnames' in d)
    self.assertEqual(d['colnames'], ["state", "Record Count"])

    self.assertTrue('rows' in d)
    self.assertEqual(d['rows'], json_dumps(result_copy))

