# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
from http import client
import json
import django.db
import django.test
from django.core.serializers import serialize


from django.test.client import Client
from datasets.models import DataConnection, DataSet
from rbac.permcache import RBAC
from reports.test_views import make_ds_viewer_role
from .test_report_operations import make_user, create_client
from django.urls import reverse
from . import test_report_operations
from vizlib.tests.util import run_with_db_query_check


class TestDatasetQueryCountBase(django.test.TestCase):
  def common_setup(self):
    util = test_report_operations.TestDashboardOperations()
    self.util = util
    # crete admin
    admin = make_user('admin', is_superuser=True)
    admin_client = create_client(admin)
    self.admin_client = admin_client
    # create normal user
    user1 = make_user('user1', is_superuser=False)
    user1_client = create_client(user1)
    self.user1_client = user1_client
    # create sqlite connection and dataset
    dc_id1 = self.create_dataconnection(admin_client, 'sqlite')
    self.dc_id1 = dc_id1
    ds1_id = util.create_dataset(admin_client, dc_id1, 'ds1')
    util.create_dataset(admin_client, dc_id1, 'ds11', dataset_type='multitable')
    self.ds1_id = ds1_id
    # create impala connection and dataset
    dc_id2 = self.create_dataconnection(admin_client, 'impala')
    ds2_id = util.create_dataset(admin_client, dc_id2, 'ds2', dataset_type='singletable-sql')
    util.create_dataset(admin_client, dc_id2, 'ds22')
    self.ds2_id = ds2_id
    # create hive connection and dataset
    dc_id3 = self.create_dataconnection(admin_client, 'hive')
    ds3_id = util.create_dataset(admin_client, dc_id3, 'ds3', dataset_type='multitable')
    util.create_dataset(admin_client, dc_id3, 'ds33', dataset_type='singletable-sql')
    self.ds3_id = ds3_id

    for dc_id, ds_id in [(dc_id1, ds1_id), (dc_id2, ds2_id), (dc_id3, ds3_id)]:
      ds_view_role = make_ds_viewer_role(
      f'ds_viewer_{ds_id}', dc_id, [ds_id])
      ds_view_role.add_users([user1.username])
      ds_view_role.save()

  def create_dataconnection(self, admin_client, connection_type):
    impala_payload = {
      'dataconnection_type': 'impyla',
      'dataconnection_info': '{}',
      'dataconnection_name': 'Impala Judit',
    }
    sqlite_payload = {
      'dataconnection_type': 'sqlite',
      'dataconnection_info': '{}',
      'dataconnection_name': 'samples',
    }
    hive_payload = {
      'dataconnection_type': 'hive',
      'dataconnection_info': '{}',
      'dataconnection_name': 'Hive Judit',
    }
    if connection_type == 'impala':
      dc_payload = impala_payload
    elif connection_type == 'hive':
      dc_payload = hive_payload
    else:
      dc_payload = sqlite_payload

    return self.util.create_dataconnection_with_payload(admin_client, dc_payload)

  def setUp(self):
    self.common_setup()

class TestDatasetQueryCount(TestDatasetQueryCountBase):
  def setUp(self):
    super().common_setup()

  def test_datasets_query_logging(self):
    admin_client = self.admin_client
    user1_client = self.user1_client

    self.verify_datasets_summary_query_count(admin_client, 4, 6)
    self.verify_datasets_summary_query_count(user1_client, 7, 3)

    self.verify_datasets_summary_query_count(admin_client, 4, 2, dc_id=self.dc_id1)
    self.verify_datasets_summary_query_count(user1_client, 7, 1, dc_id=self.dc_id1)

    self.verify_datasets_all_query_count(admin_client, 4, 6)
    self.verify_datasets_all_query_count(user1_client, 7, 3)

    self.verify_datasets_all_query_count(admin_client, 4, 2, dc_id=self.dc_id1)
    self.verify_datasets_all_query_count(user1_client, 7, 1, dc_id=self.dc_id1)

  def verify_datasets_summary_query_count(self, client: Client, expected_count: int, expected_ds_count: int, dc_id=None):
    datasets, actual_query_count = run_with_db_query_check(
      lambda: self.fetch_datasets_summary(client, dc_id))
    self.assertEquals(actual_query_count, expected_count)
    self.assertEquals(len(datasets), expected_ds_count)

  def verify_datasets_all_query_count(self, client: Client, expected_count: int, expected_ds_count: int, dc_id=None):
    datasets, actual_query_count = run_with_db_query_check(
      lambda: self.fetch_datasets_all(client, dc_id))
    self.assertEquals(actual_query_count, expected_count)
    self.assertEquals(len(datasets), expected_ds_count)

  def fetch_datasets_summary(self, client: Client, dc_id: int):
    url = reverse('datasets-dataset')
    data = {}
    if dc_id:
      data['dataconnection_id'] = dc_id

    response = client.get(url + '?summary=true', data)
    report_list = response.json()
    self.assertEquals(response.status_code, 200)
    return report_list

  def fetch_datasets_all(self, client: Client, dc_id: int):
    url = reverse('datasets-dataset')
    data = {}
    if dc_id:
      data['dataconnection_id'] = dc_id

    response = client.get(url, data)
    report_list = response.json()
    self.assertEquals(response.status_code, 200)
    return report_list

class DataSetTests(django.test.TestCase):

    def setUp(self):
        admin = make_user('admin', is_superuser=True)
        admin_client = create_client(admin)

        dc = DataConnection.objects.create(dataconnection_type='sqlite')

        self.admin_client = admin_client
        # Create some sample data
        self.data1 = DataSet.objects.create(
          dataset_name="name1",
          dataconnection=dc,
          dataset_type="singletable",
          dataset_detail="main.cereals",
          dataset_description="",
          create_user="vizapps_admin",
          update_user="vizapps_admin",
          version_name="v_name1",
          dataset_tablenames='["main.cereals"]',
        )
        self.data2 = DataSet.objects.create(
          dataset_name="name2",
          dataconnection=dc,
          dataset_type="singletable",
          dataset_detail="main.cereals",
          dataset_description="",
          create_user="vizapps_admin",
          update_user="vizapps_admin",
          version_name="v_name2",
          version_group_id=self.data1.id,
          dataset_tablenames='["main.cereals"]',
        )


    def test_swap_records(self):
        url = reverse('swap_records')
        data = {'master_id': self.data1.id, 'version_id': self.data2.version_id}
        print("sajt",url)
        response = self.admin_client.post(url, data)
        ds = DataSet.objects.get(id=self.data1.id)
        ds2 = DataSet.objects.get(id=self.data2.id)
        print(response)
        print(ds2.version_group_id)


        self.data1.refresh_from_db()
        self.data2.refresh_from_db()

        self.assertEqual(self.data1.dataset_name, "name2")
        self.assertEqual(self.data2.dataset_name, "name1")


        self.assertEqual(response.status_code, 200)







