# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
ETAG Unit tests:

Currently we just have a test for ARC-7876 to ensure a group change updates the etag

But we create lots more metadata objects to set up for adding more tests.

"""

import sys
import pdb
import time
import json
import unittest

from django.conf import settings
from django.contrib.auth.models import User, Group
from django.http import HttpRequest

from datasets.models import DataSet, DataConnection
from workspace.models import Workspace, WSACL_Member, WSACL_Level
from reports.models import Report, AppSetting
from reports.views_appgroups import appgroup_starred
from rbac.models import Role

from django.test.client import Client
from django.conf import settings

def get_etag(rsp):
  return rsp.headers.get('etag')

# Logging is handy for debugging, but currently not used
import logging
def _getlog(name=''):
  settings.ACCESS_LOG = True
  lg = logging.getLogger(name)
  lg.setLevel(logging.DEBUG)
  lh = logging.StreamHandler()
  lh.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(process)d %(threadName)s %(message)s'))
  lg.addHandler(lh)
  return lg

class TestBasic(unittest.TestCase):

  @classmethod
  def setUpClass(cls):
    AppSetting.objects.create(setting_key='registered', setting_value='True')
#   _getlog('arcweb')

  @classmethod
  def tearDownClass(cls):
    pass

  def setUp(self):
    self.objects = type('', (), {})
    o = self.objects

    def make_user(username, is_superuser=False):
      u = User.objects.create(username=username, is_superuser=is_superuser)
      u.set_password(username)
      u.save()
      request = HttpRequest()
      request.user = u
      appgroup_starred(request)
      return u

    o.ua = make_user('admin', is_superuser=True)
    o.u1 = make_user('user1')
    o.u2 = make_user('user2')

    o.g1 = Group.objects.create(name='group1')
    o.g2 = Group.objects.create(name='group2')

    o.dc1 = DataConnection.objects.create(dataconnection_name='dataconn1')
    o.dc2 = DataConnection.objects.create(dataconnection_name='dataconn2')
    o.ds1 = DataSet.objects.create(dataset_name='dataset1', dataconnection=o.dc1,
        dataset_detail='test', dataset_type='test')
    o.ds2 = DataSet.objects.create(dataset_name='dataset2', dataconnection=o.dc2,
        dataset_detail='test', dataset_type='test')

    o.u1.groups.add(o.g1)
    o.u2.groups.add(o.g2)
    o.u1.save()
    o.u2.save()

    o.r1 = Role.objects.create(name='role1',
                               privileges=json.dumps([{'ptype':'system','perms':['sys_editperm']}]))
    o.r2 = Role.objects.create(name='role2',
                               privileges=json.dumps([{'ptype':'system','perms':['sys_editconn']}]))

    o.ca = Client()
    o.ca.login(username=o.ua.username, password=o.ua.username)
    o.cu1 = Client()
    o.cu1.login(username=o.u1.username, password=o.u1.username)

  def tearDown(self):
    self.objects.ca.logout()
    self.objects.cu1.logout()
    del self.objects.ca
    del self.objects.cu1
    d = self.objects.__dict__
    for name in d.keys():
      if not name.startswith('__'):
        d[name].delete()

  def test_1(self):
    # Perms from user
    o = self.objects
    rsp = o.ca.get('/reports/report')
    self.assertEqual(rsp.status_code, 200)
    etag = get_etag(rsp)
    rsp = o.ca.get('/reports/report', HTTP_IF_NONE_MATCH=etag)
    self.assertEqual(rsp.status_code, 304)

    # Change in groups invalidates etag
    o.ua.groups.add(o.g1)
    rsp = o.ca.get('/reports/report', HTTP_IF_NONE_MATCH=etag)
    self.assertEqual(rsp.status_code, 200)

    # Validate new etag
    etag = get_etag(rsp)
    rsp = o.ca.get('/reports/report', HTTP_IF_NONE_MATCH=etag)
    self.assertEqual(rsp.status_code, 304)

if __name__ == '__main__':
  unittest.main()
