#!/usr/bin/python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import unittest

from vizlib.expression import Expression, unbracketColumn, stripAlias

class TestExpression(unittest.TestCase):
  def test_stripAlias(self):
    self.assertEquals(stripAlias('[channel] as "channel"'), '[channel]')
    self.assertEquals(stripAlias('sum([channel]) as "sum_channel"'), 'sum([channel])')
    self.assertEquals(stripAlias("[channel] as 'channel'"), '[channel]')
    self.assertEquals(stripAlias("sum([channel]) as 'sum_channel'"), 'sum([channel])')
    self.assertEquals(stripAlias('sum([channel]) as sum_channel'), 'sum([channel]) as sum_channel')

  def test_unbracketColumn(self):
    self.assertEquals(unbracketColumn('abc'), 'abc')
    self.assertEquals(unbracketColumn('[def]'), 'def')
    self.assertEquals(unbracketColumn('[[ghi]'), '[ghi')
    self.assertEquals(unbracketColumn('[xyz'), '[xyz')
    self.assertEquals(unbracketColumn('xyz]'), 'xyz]')

  def test_columns(self):
    cols = list(Expression.findColumns('[abc]'))
    self.assertEquals(cols, ['abc'])

    cols = list(Expression.findColumns(' xy ( [abc] ) z '))
    self.assertEquals(cols, ['abc'])

    cols = list(Expression.findColumns('[abc][def][xyz]'))
    self.assertEquals(cols, ['abc', 'def', 'xyz'])

    cols = list(Expression.findColumns('concat([abc], 3, [def], [xyz])'))
    self.assertEquals(cols, ['abc', 'def', 'xyz'])

    cols = list(Expression.findColumns(' xy ( [abc] ) z '))
    self.assertEquals(cols, ['abc'])

    cols = list(Expression.findColumns('sum([abc])'))
    self.assertEquals(cols, ['abc'])

  def test_dimension(self):
    e = Expression('somecol')
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': 'somecol',
      'alias': None,
      'simpleFn': None,
      'simpleFnArg': None
    })
    self.assertFalse(e.isCountDistinct())

  def test_dimension_squarebrackets(self):
    e = Expression('[somecol]')
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': '[somecol]',
      'alias': None,
      'simpleFn': None,
      'simpleFnArg': None
    })

  def test_dimension_alias(self):
    e = Expression("somecol as 'some alias'")
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': 'somecol',
      'alias': "'some alias'",
      'simpleFn': None,
      'simpleFnArg': None
    })

    e = Expression("[somecol] as 'some alias'")
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': '[somecol]',
      'alias': "'some alias'",
      'simpleFn': None,
      'simpleFnArg': None
    })

  def test_alias_singlequotes(self):
    e = Expression("somefunction(somearg) as 'somealias'")
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': 'somefunction(somearg)',
      'alias': "'somealias'",
      'simpleFn': 'somefunction',
      'simpleFnArg': 'somearg'
    })

    e = Expression("sum(1) as 'record_count'")
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': 'sum(1)',
      'alias': "'record_count'",
      'simpleFn': 'sum',
      'simpleFnArg': '1'
    })

  def test_alias_doublequotes(self):
    e = Expression('somefunction([somearg]) as "somealias"')
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': 'somefunction([somearg])',
      'alias': '"somealias"',
      'simpleFn': 'somefunction',
      'simpleFnArg': '[somearg]'
    })

  def test_spaces(self):
    e = Expression("somefunction(some arg) as 'some alias'")
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': 'somefunction(some arg)',
      'alias': "'some alias'",
      'simpleFn': 'somefunction',
      'simpleFnArg': 'some arg'
    })

  def test_spaces_in_function(self):
    e = Expression("some function([some_arg]) as 'some alias'")
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': 'some function([some_arg])',
      'alias': "'some alias'",
      'simpleFn': None,
      'simpleFnArg': None
    })

  def test_count_distinct(self):
    e = Expression("count(*) as 'rec_count'")
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': 'count(*)',
      'alias': "'rec_count'",
      'simpleFn': 'count',
      'simpleFnArg': '*'
    })

    self.assertFalse(e.isCountDistinct())

    e = Expression("count(distinct [arg]) as 'count_arg'")
    self.assertTrue(e is not None)
    self.assertEquals(e.components(), {
      'baseExpr': 'count(distinct [arg])',
      'alias': "'count_arg'",
      'simpleFn': 'count',
      'simpleFnArg': 'distinct [arg]'
    })

    self.assertTrue(e.isCountDistinct())


