# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import django.test
from unittest.mock import Mock

from dataconn import connections_types
from vizlib.dsreq import DsreqCreator
from vizlib import nosql_builder
from vizlib.nosql_builder import SolrBuilder

from extracts.tests import tests_rewrites

SOLR = connections_types.DATACONN_SOLR
class Expando(object):
  pass

class SolrBuilderTestBase(django.test.TestCase):
  def setUp(self):
    super().setUp()
    # prevent diff truncation:
    self.maxDiff = None

    self.old_json_settings = nosql_builder.ALWAYS_SEND_JSON_QUERY
    nosql_builder.ALWAYS_SEND_JSON_QUERY = True

    self.helper = self.create_helper()

  def tearDown(self):
    nosql_builder.ALWAYS_SEND_JSON_QUERY = self.old_json_settings
    super().tearDown()

  def create_helper(self):
    helper = tests_rewrites.Test_Rewrites()
    helper.property_map = {
        'baseurl': 'http://${SOLR_HOST}/api/collections/',
        'wt': 'json'
    }

    return helper

class TestSolrBuilderBasic(SolrBuilderTestBase):

  def mock_dataset(self):
    dataset = Mock()
    dataset.properties = {}
    table = Mock()
    table.tablename = 'testDb.testTable'
    dataset.tables = [table]
    dataset.dsuhash = {}

    return dataset

  def mock_column(self, **kwargs):
    column = Mock()
    column.attr = dict(kwargs)
    return column

  def test_alias_lookup(self):
    """
    Column name lookup based on alias must not throw exceptions.
    If the lookup fails, return the input
    """
    dataset = self.mock_dataset()

    dataset.dsuhash = {
      'Alias B': self.mock_column(name='b', basecol='base_b'),
      'Alias C': self.mock_column(basecol='c'),
      'Alias D': self.mock_column(),
    }
    solr_builder = SolrBuilder(dataset)

    # it is OK to ask for non-existing column:
    self.assertEquals(solr_builder._param_to_colname('a'), 'a')

    # it is OK to ask for a column with name:
    self.assertEquals(solr_builder._param_to_colname('Alias B'), 'b')

    # it is OK to ask for a column with a base name:
    self.assertEquals(solr_builder._param_to_colname('Alias C'), 'c')

    # it is OK to ask for a column with no info:
    self.assertEquals(solr_builder._param_to_colname('Alias D'), 'Alias D')

class TestSolrBuilderKNN(SolrBuilderTestBase):
  """
  Steps to create a collection with a vector field:

export SOLR_HOST=viz-team-cdpdc-4.cc-os.cloudera.com:8983

# step 1: create config set
curl -k 'http://${SOLR_HOST}/solr/admin/configs?action=CREATE&name=ampCollectionConfigSet&baseConfigSet=_default&configSetProp.immutable=false&wt=xml&omitHeader=true'

# step 2: create collection using the config set
curl -k -X POST http://${SOLR_HOST}/api/collections -H 'Content-Type: application/json' -d '
  {
    "name": "ampCollection",
    "config": "ampCollectionConfigSet",
    "numShards": 1
  }
'

# step 3: add vector field type
curl -k -X POST -H 'Content-type:application/json' \
  --data-binary '{
    "add-field-type" : {
      "name":"knn_vector",
      "class":"solr.DenseVectorField",
      "vectorDimension":"5",
      "similarityFunction":"euclidean",
      "knnAlgorithm":"hnsw",
      "hnswMaxConnections":"10",
      "hnswBeamWidth":"40"
    }
  }' \
  http://${SOLR_HOST}/solr/ampCollection/schema

# step 4: define the fields for the collection
curl -X POST \
  --url http://${SOLR_HOST}/api/collections/ampCollection/schema \
  --header 'Content-Type: application/json' \
  --data '{
    "add-field": [
    {"name": "data", "type": "string", "multiValued": false},
    {"name": "meta_information", "type": "string", "multiValued": false},
    {"name": "embedding", "type": "knn_vector", "indexed":true, "stored":true}
    ]
  }'

# create data:
curl -X POST \
  --url http://${SOLR_HOST}/api/collections/ampCollection/update \
  --header 'Content-Type: application/json' \
  --data '{
    "data": "some data 4",
    "meta_information": "Document URL",
    "embedding":[1,2,3,4,50]
  }'
  """

  def setUp(self):
    super().setUp()
    self.ds = self.createDataset()

  def createDataset(self):

    columns = [
      {
        "name": "data",
        "alias": "My Data"
      },
      {
        "name": "embedding",
        "alias": "Embedding Vector"
      },
    ]

    dc = self.helper.make_dataconnection()
    dc.dataconnection_type = SOLR
    ds = self.helper.make_dataset(dc, 'testdb.ampCollection', columns=columns)
    return ds

  def test_KNN(self):

    # test KNN search with extra fields selected. Use field aliases.
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[My Data] as 'A'") \
      .addFilter("[Embedding Vector] knn (0,0,0,2)") \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/ampCollection/select", "body": {"params": {"wt": "json", "q": "{!knn f=embedding topK=6}[0,0,0,2]", "fl": "data", "rows": 6}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)


class TestSolrBuilderQueries(SolrBuilderTestBase):
  """
  Queries were created/captured on a dataset based on the HR data (because it contains many field types):
  https://github.infra.cloudera.com/Sense/viz/tree/adb0b7876bd98cf4d21c06a5de665266805eff47/tests/solr/data_solr_only
  """

  def setUp(self):
    super().setUp()
    self.ds = self.createDataset()

  def createDataset(self):
    columns = [
      {
        "name": "Employee_Name",
        "alias": "Employee Name"
      },
      {
        "name": "Salary",
        "alias": "Gross Salary"
      },
      {
        "name": "EmpSatisfaction",
        "alias": "Emp Satisfaction"
      },
      {
        "name": "Position",
        "alias": "Emp Position"
      },
      {
        "name": "Absences",
        "alias": "Number of Absences"
      },
      {
        "name": "DateofHire",
        "alias": "Date of Hire"
      },
      {
        "name": "EmploymentStatus",
        "alias": "Employment Status"
      },
    ]
    dc = self.helper.make_dataconnection()
    dc.dataconnection_type = SOLR
    ds = self.helper.make_dataset(dc, 'testdb.hr_dataset', columns=columns)
    return ds

  def test_dims_asc(self):

    # Dimensions only, order ascending:
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employee Name] as 'A'", orderby={'asc': True}) \
      .addDimension("[Employment Status] as 'B'") \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "fl": ["Employee_Name", "EmploymentStatus"], "sort": "Employee_Name asc", "rows": 6}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_dims_desc(self):
    # Dimensions only, order descending:
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employee Name] as 'A'", orderby={'asc': False}) \
      .addDimension("[Employment Status] as 'B'") \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "fl": ["Employee_Name", "EmploymentStatus"], "sort": "Employee_Name desc", "rows": 6}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_unsorted(self):
    # test record count, unsorted
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'") \
      .addAggregate("sum(1) as 'Record Count'") \
      .setLimit(6) \
      .getdsreq()

    # CHECK: rows should be probaly 0 here (as we have facet.limit)
    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 6, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_asc_by_dim(self):

    # test record count, ascending by dimension:
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .setLimit(6) \
      .getdsreq()

    # CHECK: why does this work? We want sorting on Employment status, not count
    # probaly "facet.sort": "count,asc" is not a valid setting. Valid values are "count" that means descending ordering by count, "index" that means ascending ordering by the facet field value.
    # an invalid setting (like "facet.sort": "aaa") seems to trigger index ordering.
    # to have ordering work correctly, one should probably switch to the JSON facet API
    #
    # E.g the following works with any of "index asc", "index desc", "count asc", "count desc" filter values, the last one being the default:
    """
    curl http://${SOLR_HOST}/solr/hr_dataset/select -d '{
      "query": "*:*",
      "limit": 0,
      "facet": {
        "categories" : {
          "type": "terms",
          "field": "EmploymentStatus",
          "sort": "index asc"
        }
      }
    }'
    """

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc"}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_desc_by_dim(self):

    # test record count, ascending by dimension:
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': False}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .setLimit(6) \
      .getdsreq()

    # CHECK: sorting DOES NOT WORK
    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_asc_by_agg(self):

    # test record count, ascending by dimension:
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'") \
      .addAggregate("sum(1) as 'Record Count'", orderby={'asc': True}) \
      .setLimit(6) \
      .getdsreq()

    # CHECK: Totally wrong, emits exactly the same query as when sorting by the dimension
    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc"}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_desc_by_agg(self):

    # test record count, ascending by dimension:
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'") \
      .addAggregate("sum(1) as 'Record Count'", orderby={'asc': False}) \
      .setLimit(6) \
      .getdsreq()

    # CHECK: sorting DOES NOT WORK, same query as when asking for dimension sorting
    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_aggregates(self):

    # test record count, ascending by dimension:
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'") \
      .addAggregate("sum([Gross Salary]) as 'B'", orderby={'asc': False}) \
      .addAggregate("avg([Emp Satisfaction]) as 'B'") \
      .addAggregate("min([Number of Absences]) as 'B'") \
      .addAggregate("max([Number of Absences]) as 'B'") \
      .addAggregate("min([Date of Hire]) as 'B'") \
      .addAggregate("max([Date of Hire]) as 'B'") \
      .setLimit(6) \
      .getdsreq()

    # CHECK: sorting is done on the dimension, not the requested aggregate
    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "stats.field": ["{!tag=piv  sum=true }Salary", "{!tag=piv  mean=true }EmpSatisfaction", "{!tag=piv  min=true max=true }Absences", "{!tag=piv  min=true max=true }DateofHire"], "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_with_filter_from_picklist(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter("[Emp Position] in ('Area Sales Manager','Production Technician I')") \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc", "fq": "Position:(\\"Area Sales Manager\\" OR \\"Production Technician I\\")"}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_with_filter_from_picklist_negated(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter("[Emp Position] not in ('Area Sales Manager','Production Technician I')") \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc", "fq": "!Position:(\\"Area Sales Manager\\" OR \\"Production Technician I\\")"}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_with_filter_numeric_equality(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter("[Emp Satisfaction] = 5") \
      .setLimit(6) \
      .getdsreq()

    # CHECK: "EmpSatisfaction:[* TO *]" is probably superfluous
    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc", "fq": ["EmpSatisfaction:[* TO *]", "EmpSatisfaction:5"]}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_with_filter_numeric_equality_negated(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter("[Emp Satisfaction] != 5") \
      .setLimit(6) \
      .getdsreq()

    # CHECK: "EmpSatisfaction:[* TO *]" is probably superfluous
    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc", "fq": ["EmpSatisfaction:[* TO *]", "-EmpSatisfaction:5"]}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_with_filter_numeric_range(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter("[Emp Satisfaction] > 2 AND [Emp Satisfaction] <= 4") \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc", "fq": "EmpSatisfaction:{2 TO 4]"}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_with_filter_time_range(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter('[Date of Hire] between "2015-01-01T00:00:00Z" and "2015-12-31T00:00:00Z"') \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc", "fq": "DateofHire:[2015-01-01T00:00:00Z TO 2015-12-31T00:00:00Z]"}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_with_filter_time_range_with_single_quotes(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter("[Date of Hire] between '2015-01-01T00:00:00Z' and '2015-12-31T00:00:00Z'") \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc", "fq": "DateofHire:[2015-01-01T00:00:00Z TO 2015-12-31T00:00:00Z]"}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def xtest_record_count_with_filter_time_range_including_nulls(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter('(([Date of Hire] between "2015-01-01T00:00:00Z" and "2015-12-31T00:00:00Z") or ([Date of Hire] is NULL))') \
      .setLimit(6) \
      .getdsreq()

    # CHECK: does not work, the daterange filter is not recognized because if the "or"
    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc", "fq": "DateofHire:[2015-01-01T00:00:00Z TO 2015-12-31T00:00:00Z]"}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_record_count_with_filter_from_picklist_with_dates(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'", orderby={'asc': True}) \
      .addAggregate("sum(1) as 'Record Count'") \
      .addFilter('[Date of Hire] in ("2015-01-05T00:00:00Z", "2015-02-16T00:00:00Z")') \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "DateofHire:(\\"2015-01-05T00:00:00Z\\" OR \\"2015-02-16T00:00:00Z\\")", "rows": 0, "stats": "true", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6, "facet.sort": "count,asc"}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)

  def test_distinct_count(self):

    # test record count, ascending by dimension, filter by another dimension
    dsreq = DsreqCreator() \
      .datasetId(4001) \
      .addDimension("[Employment Status] as 'A'") \
      .addAggregate("count(distinct([Emp Satisfaction]) as 'B'") \
      .setLimit(6) \
      .getdsreq()

    expected_query='json:{"url": "http://${SOLR_HOST}/api/collections/hr_dataset/select", "body": {"params": {"wt": "json", "q": "*:*", "rows": 6, "stats": "true", "stats.field": "{!tag=piv  countDistinct=true }EmpSatisfaction", "facet": "on", "facet.pivot": "{!stats=piv}EmploymentStatus", "facet.limit": 6}}}'
    build_request = self.helper.create_build_request(dsreq, ds=self.ds)

    self.assertEqual(build_request.query, expected_query)