#!/usr/bin/python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from typing import List
import unittest
import os
import sys
import json
from django.conf import settings
from django.core.exceptions import ValidationError
import django.test
from datasets.models import DataConnection, DataSet

from reports.models import Report
import reports.views
import reports.views_appgroups

from rbac.models import Role

import datasets.views
import apps.views
import datasets.views_connections

from arcwebbase import access_info
from django.contrib.auth.models import User
from unittest import mock
from util.request import create_request
from vizlib import querylog
from vizlib.querylog import ActionType, ComponentType

THISDIR = os.path.dirname(os.path.realpath(__file__))
SHORT_PASSWORD_MSG = 'Password must be at least 5 characters'

sys.path.append("%s/.." % THISDIR)

class MockGroup(object):
  def __init__(self, name=None):
    self.name = name

  def values_list(self, attr, flat=False):
    return []

class MockUser(object):
  def __init__(self, username=None, is_superuser=True, id=1):
    self.username = username
    self.is_superuser = is_superuser
    self.id = id
    self.groups = MockGroup()
    self.is_authenticated = True

  def has_perm(self, perm):
    return True

class MockRequest(object):
  def __init__(self, post=None, user=None, method=None, body=None, content_type=None):
    self.GET = {}
    self.POST = {}
    self.body = {}
    self.META = {}
    self.session = type('', (), {})() # create a type that we can assign attributes to
    self.session.session_key = "1234"
    # session.cycle_key is used by the update_session_auth_hash method internally
    # update_session_auth_hash is used to update the session after the user changes it's password
    self.session.cycle_key = lambda: "12345"
    self.user = type('', (), {})() # create a type that we can assign attributes to
    self.user.id = 1
    self.content_type = content_type

    if post is not None:
      self.POST = post
    if user:
      self.user = user
    if method:
      self.method = method
    if body is not None:
      self.body=body

def make_dataset_manager_role(dc_id: str, ds_ids: List[str]):
  role = Role()
  role.name = 'mixed_ds_perms'
  role.privileges = json.dumps(
      [
          # no system permissions:
          {'ptype': 'system', 'perms': []},
          # full rights on all dataconnections:
          {
              'ptype': 'dataconn',
              'dclist': ['-1'],
              'perms': ['dc_upload', 'dc_explore', 'dc_aviews'],
          },
          # full rights on the specified datasets
          {
              'ptype': 'dataset',
              'dcid': str(dc_id),
              'dslist': [str(ds_id) for ds_id in ds_ids],
              'perms': ['ds_manage', 'ds_appview', 'ds_appedit'],
          },
      ]
  )
  role.save()
  return role

class TestHandles(django.test.TestCase):
  dataconnection_id = None
  dataset_id = None

  def setUp(self):
    self.saved_audit_logger = settings.AUDIT_LOGGER
    settings.AUDIT_LOGGER = self
    self.messages = []
    self.admin_username = 'boss'
    self.adminuser = MockUser(username=self.admin_username, is_superuser=True)
    self.dataset_manager = MockUser(username='dataset_manager', is_superuser=False)

  def info(self, msg, extra):
    self.messages.append(extra.get('audit_record'))

  def _attempt_user(self):
    self.messages = []
    fakeadmin = MockUser(username='shortpass', is_superuser=True)
    post = {
      'password': 'pwpw',
      'is_superuser': 'true'
    }
    request = MockRequest(post=post, user=fakeadmin, method='POST')
    with self.assertRaises(ValidationError):
      try:
        apps.views.users(request, self.admin_username)
      except ValidationError as v:
        self.assertEqual(v.messages, [SHORT_PASSWORD_MSG])
        raise


  def _create_user(self):
    self.messages = []
    fakeadmin = MockUser(username='fakeboss', is_superuser=True)
    post = {
      'password': 'pwpwp',
      'is_superuser': 'true'
    }
    request = MockRequest(post=post, user=fakeadmin, method='POST')
    apps.views.users(request, self.admin_username)
    self.assertEqual(len(self.messages), 1)

    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.CREATE)
    self.assertEqual(m['component_type'], ComponentType.USER)
    self.assertEqual(m['component_id'], fakeadmin.id)
    self.assertEqual(m['username'], 'fakeboss')

  def _modify_user(self):
    self.messages = []
    body = 'old_password=pwpwp&new_password=wpwpws&confirm_password=wpwpws'
    request = MockRequest(body=body, user=self.adminuser, method='PUT')
    apps.views.users(request, self.admin_username)
    self.assertEqual(len(self.messages), 1)

    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.UPDATE)
    self.assertEqual(m['component_type'], ComponentType.USER)
    self.assertEqual(m['component_id'], request.user.id)
    self.assertEqual(m['username'], self.admin_username)

  def _delete_user(self):
    self.messages = []
    request = MockRequest(user=self.adminuser, method='DELETE')
    apps.views.users(request, username=self.admin_username)
    self.assertEqual(len(self.messages), 1)

    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.DELETE)
    self.assertEqual(m['component_type'], ComponentType.USER)
    self.assertEqual(m['component_id'], request.user.id)
    self.assertEqual(m['username'], self.admin_username)

  def _create_dataconnection(self):
    self.messages = []
    post = {
      'dataconnection_type': 'arcengine',
      'dataconnection_info': '{"PARAMS":{"HOST":"10.0.0.211","PORT":"31050"}}',
      'dataconnection_name': 'abc',
    }
    request = MockRequest(post=post, user=self.adminuser, method='POST')
    datasets.views_connections.dataconnection(request)
    self.assertEqual(len(self.messages), 1)
    self.dataconnection_id = DataConnection.objects.get(dataconnection_name='abc').id

    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.CREATE)
    self.assertEqual(m['component_type'], ComponentType.DATACONNECTION)
    self.assertEqual(m['component_id'], self.dataconnection_id)
    self.assertEqual(m['username'], self.admin_username)

  def _rename_dataconnection(self):
    self.messages = []
    body = 'dataconnection_type=arcengine&dataconnection_name=abc&dataconnection_info={"PARAMS":{"HOST":"10.0.0.211","PORT":"31050"}}'
    request = MockRequest(post='', body=body, user=self.adminuser, method='PUT')
    datasets.views_connections.dataconnection(request, dataconnection_id=self.dataconnection_id)
    self.assertEqual(len(self.messages), 1)

    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.UPDATE)
    self.assertEqual(m['component_type'], ComponentType.DATACONNECTION)
    self.assertEqual(m['component_id'], self.dataconnection_id)
    self.assertEqual(m['username'], self.admin_username)

  def _delete_dataconnection(self):
    self.messages = []
    request = MockRequest(user=self.adminuser, method='DELETE')
    datasets.views_connections.dataconnection(request, dataconnection_id=self.dataconnection_id)
    self.assertEqual(len(self.messages), 1)

    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.DELETE)
    self.assertEqual(m['component_type'], ComponentType.DATACONNECTION)
    self.assertEqual(m['component_id'], self.dataconnection_id)
    self.assertEqual(m['username'], self.admin_username)

  def _create_dataset(self, dataset_name):
    self.messages = []
    post = {
      'dataconnection_id': self.dataconnection_id,
      'dataset_name': dataset_name,
      'dataset_type': 'singletable',
      'dataset_info': '{}',
      'dataset_detail': '{}'
    }
    request = MockRequest(post=post, user=self.adminuser, method='POST')
    datasets.views.dataset(request)
    self.assertEqual(len(self.messages), 1)
    dataset_id = DataSet.objects.get(dataset_name=dataset_name).id
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.CREATE)
    self.assertEqual(m['component_type'], ComponentType.DATASET)
    self.assertEqual(m['component_id'], dataset_id)
    self.assertEqual(m['username'], self.admin_username)
    return dataset_id

  def _rename_dataset(self):
    self.messages = []
    dataset_info = '[{"tablename":"default.census_pop",' \
                   '"columns":[{"name":"year","type":"INT","isdim":false,"alias":"year"},' \
                   '{"name":"state","type":"STRING","isdim":true,"alias":"state"},' \
                   '{"name":"population","type":"BIGINT","isdim":false,"alias":"population"}]}]'
    body = "dataset_name=ab&dataset_type=singletable&dataset_info=%s&dataset_detail=detail" % dataset_info
    request = MockRequest(body=body, user=self.adminuser, method='PUT', content_type='')
    datasets.views.dataset(request, dataset_id=self.dataset_id)
    self.assertEqual(len(self.messages), 1)
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.UPDATE)
    self.assertEqual(m['component_type'], ComponentType.DATASET)
    self.assertEqual(m['component_id'], self.dataset_id)
    self.assertEqual(m['username'], self.admin_username)

  def _delete_dataset(self):
    self.messages = []
    request = MockRequest(user=self.adminuser, method='DELETE', body='')
    datasets.views.dataset(request, dataset_id=self.dataset_id)
    self.assertEqual(len(self.messages), 1)
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.DELETE)
    self.assertEqual(m['component_type'], ComponentType.DATASET)
    self.assertEqual(m['component_id'], self.dataset_id)
    self.assertEqual(m['username'], self.admin_username)

  def _delete_dataset4(self):
    """
    dataset 4 has a dashboard and and one visual (visual_6).
    When deleted, the dashboard is not deleted because it contains another visual that belongs
    to dataset 3. The dashboard is reparented to dataset 3.
    """
    expected_visual_audits ={
      self.visual_5:{
        'visual': self.visual_5,
        'app': self.dashboard_id3,
      }
    }
    self._do_delete_dataset(user=self.dataset_manager, dataset_id=self.dataset_id4, expected_extra_audits=expected_visual_audits)

    # check dataset updating:
    dashboard_3: Report = Report.objects.get(id=self.dashboard_id3)
    dashboard_3_dataset: DataSet = dashboard_3.dataset
    self.assertEquals(dashboard_3_dataset.id, self.dataset_id3)

  def _delete_dataset3(self):
    """
    dataset 3 did not have any dashboards crreated explicitly,
    but got one after dataset 4 was removed
    """
    expected_visual_audits ={
      self.visual_6:{
        'visual': self.visual_6,
        'app': self.dashboard_id3,
      },
      self.dashboard_id3:{
        'visual': '',
        'app': self.dashboard_id3,
      }
    }
    self._do_delete_dataset(user=self.dataset_manager, dataset_id=self.dataset_id3, expected_extra_audits=expected_visual_audits)

  def _delete_dataset2(self):

    expected_visual_audits ={
      self.visual_3:{
        'visual': self.visual_3,
        'app': self.dashboard_id,
      },
      self.visual_4:{
        'visual': self.visual_4,
        'app': self.dashboard_id2,
      },
      self.dashboard_id2:{
        'visual': '',
        'app': self.dashboard_id2,
      }
    }
    self._do_delete_dataset(user=self.adminuser, dataset_id=self.dataset_id2, expected_extra_audits=expected_visual_audits)

  def _do_delete_dataset(self, user: MockUser, dataset_id, expected_extra_audits):
    """
    Tests auditing of dependent objects (dashboard+visuals)
    """
    self.messages = []
    request = MockRequest(user=user, method='DELETE', body='')
    datasets.views.dataset(request, dataset_id=dataset_id)

    expected_audits = 1 + len(expected_extra_audits) # 1 dataset, + visuals and dashboards
    self.assertEqual(len(self.messages), expected_audits)

    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.DELETE)
    self.assertEqual(m['component_type'], ComponentType.DATASET)
    self.assertEqual(m['component_id'], dataset_id)
    self.assertEqual(m['username'], user.username)

    app_messages = sorted(self.messages[1:], key=lambda m: m['component_id'])
    expected_ids = sorted(expected_extra_audits.keys())

    for m, expected_id in zip (app_messages, expected_ids):
      self.assertEqual(m['action'], ActionType.DELETE)
      self.assertEqual(m['component_type'], ComponentType.APP)
      self.assertEqual(m['component_id'], expected_id)
      self.assertEqual(m['username'], user.username)
      self.assertEqual(m['visual'], expected_extra_audits[expected_id]['visual'])
      self.assertEqual(m['app'], expected_extra_audits[expected_id]['app'])

  def _create_app(self, dataset_id=None):
    if dataset_id == None:
      dataset_id = self.dataset_id

    self.messages = []
    post = {
      'workspace_id' : 1,
      'visual_type': 'table'
    }
    request = MockRequest(post=post, user=MockUser(username=self.admin_username))
    response = reports.views.create(request, dataset_id=dataset_id)
    report_id = json.loads(response.content)[0]['id']
    self.assertEqual(len(self.messages), 1)
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.CREATE)
    self.assertEqual(m['component_type'], ComponentType.APP)
    self.assertEqual(m['component_id'], report_id)
    self.assertEqual(m['username'], self.admin_username)

    return report_id

  def _rename_app(self):
    self.messages = []
    post = {
      'workspace_id': 1,
      'dataset_id': self.dataset_id,
      'report_description': 'renamed app',
      'report_type': 'table',
      'report_data': '{}'
    }
    request = MockRequest(post=post, user=MockUser(username=self.admin_username), method='POST')
    reports.views.report(request, report_id=1)
    self.assertEqual(len(self.messages), 1)
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.UPDATE)
    self.assertEqual(m['component_type'], ComponentType.APP)
    self.assertEqual(m['component_id'], 1)
    self.assertEqual(m['username'], self.admin_username)

  def _delete_app(self):
    self.messages = []
    request = MockRequest(user=MockUser(username=self.admin_username), method='DELETE')
    reports.views.report(request, report_id=1)
    # self.assertTrue(len(self.messages) >= 1)
    self.assertEqual(len(self.messages), 1)
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.DELETE)
    self.assertEqual(m['component_type'], ComponentType.APP)
    self.assertEqual(m['component_id'], 1)
    self.assertEqual(m['username'], self.admin_username)

  def _create_appgroup(self):
    self.messages = []
    post = {
      'data': json.dumps(
        [{
          'group_id': -1,
          'group_name': 'ag1',
          'group_subgroup_order': 1,
          'group_options': '',
        }]
      )
    }

    new_user = User.objects.get_or_create(username='new_user')[0]
    request = MockRequest(post=post, user=new_user, method='POST')
    response = reports.views_appgroups.appgroup(request)
    appGroup = json.loads(response.content)
    self.appgroupId = appGroup['saved_app_id']
    self.assertEqual(len(self.messages), 1)
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.CREATE)
    self.assertEqual(m['component_type'], ComponentType.APPGROUP)
    self.assertEqual(m['component_id'], self.appgroupId)
    self.assertEqual(m['username'], new_user.username)

  def _rename_appgroup(self):
    self.messages = []
    post = {
      'data': json.dumps(
        [{
          'group_id': self.appgroupId,
          'group_name': 'ag2',
          'group_subgroup_order': 1,
          'group_options': '',
        }]
      )
    }

    request = MockRequest(post=post, user=MockUser(username=self.admin_username), method='POST')
    reports.views_appgroups.appgroup(request)
    self.assertEqual(len(self.messages), 1)
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.UPDATE)
    self.assertEqual(m['component_type'], ComponentType.APPGROUP)
    self.assertEqual(m['component_id'], self.appgroupId)
    self.assertEqual(m['username'], self.admin_username)

  def _delete_appgroup(self):
    self.messages = []
    request = MockRequest(user=MockUser(username=self.admin_username), method='DELETE')
    reports.views_appgroups.appgroup(request, appgroup_id=self.appgroupId)
    self.assertTrue(len(self.messages) >= 1)
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.DELETE)
    self.assertEqual(m['component_type'], ComponentType.APPGROUP)
    self.assertEqual(m['component_id'], self.appgroupId)
    self.assertEqual(m['username'], self.admin_username)

  def _create_mixed_dashboard(self):
    """
    a dashboard with mixed datasets
    visual_3 will be removed when dataset_id2 is removed
    """
    self.visual_1 = self._create_app(self.dataset_id)
    self.visual_2 = self._create_app(self.dataset_id)
    self.visual_3 = self._create_app(self.dataset_id2)
    self.dashboard_id = self._do_create_dashboard(self.dataset_id2, [self.visual_1, self.visual_2, self.visual_3])

  def _create_dashboard2(self):
    # homogenous dashboard
    self.visual_4 = self._create_app(self.dataset_id2)
    self.dashboard_id2 = self._do_create_dashboard(self.dataset_id2, [self.visual_4])

  def _create_dashboard3(self):
    # another mixed-dataset dashboard
    self.visual_5 = self._create_app(self.dataset_id4)
    self.visual_6 = self._create_app(self.dataset_id3)
    self.dashboard_id3 = self._do_create_dashboard(self.dataset_id4, [self.visual_5, self.visual_6])

  def _do_create_dashboard(self, dataset_id, visuals):

    self.messages = []
    post = {
      'dashboard_dataset_id': dataset_id,
      'unshared_visuals': json.dumps(visuals),
      'dashboard_data': json.dumps(
        {
          'dashboard_sheets':[
            {
              'visual_widgets':[ {'id': 'uri-1-widget-%s' % visual_id } for visual_id in visuals]
            }
          ]
        }
      )
    }

    request = MockRequest(post=post, user=self.adminuser, method='POST')
    response = reports.views.savedashboard(request)
    dashboard = json.loads(response.content)
    dashboard_id = dashboard['id']

    self.assertEqual(len(self.messages), 1)
    m = self.messages[0]
    self.assertEqual(m['action'], ActionType.CREATE)
    self.assertEqual(m['component_type'], ComponentType.APP)
    self.assertEqual(m['component_id'], dashboard_id)
    self.assertEqual(m['username'], self.admin_username)
    return dashboard_id

  def _delete_dashboard(self):
    self.messages = []
    request = MockRequest(user=MockUser(username=self.admin_username), method='DELETE')
    reports.views.report(request, report_id=self.dashboard_id)

    # visuals deleted together with the dashboard should be also audited
    self.assertEqual(len(self.messages), 3)

    expected_ids = [self.visual_1, self.visual_2, self.dashboard_id]

    for m, expected_id in zip (self.messages, expected_ids):
      self.assertEqual(m['action'], ActionType.DELETE)
      self.assertEqual(m['component_type'], ComponentType.APP)
      self.assertEqual(m['component_id'], expected_id)
      self.assertEqual(m['username'], self.admin_username)
      self.assertEqual(m['app'], self.dashboard_id)
      if expected_id == self.dashboard_id:
        self.assertEqual(m['visual'], '')
      else:
        self.assertEqual(m['visual'], m['component_id'])

  def _empower_dataset_manager(self):
    role = make_dataset_manager_role(self.dataconnection_id,[self.dataset_id3, self.dataset_id4])
    role.add_users([self.dataset_manager.username])
    role.save()


  def test_actions(self):
    self.messages = []

    # First, create objects
    self._attempt_user()
    self._create_user()
    self._create_dataconnection()
    self.dataset_id = self._create_dataset('ds1')
    self.dataset_id2 = self._create_dataset('ds2')
    self.dataset_id3 = self._create_dataset('ds3')
    self.dataset_id4 = self._create_dataset('ds4')
    self._empower_dataset_manager()
    self._create_app()
    self._create_appgroup()
    self._create_mixed_dashboard()
    self._create_dashboard2()
    self._create_dashboard3()

    # Next, test modifications
    self._modify_user()
    self._rename_dataconnection()
    self._rename_dataset()
    self._rename_app()
    self._rename_appgroup()

    # Finally, test deletions
    self._delete_dataset4()
    self._delete_dataset3()
    self._delete_dataset2()
    self._delete_dashboard()
    self._delete_appgroup()
    self._delete_app()
    self._delete_dataset()
    self._delete_dataconnection()
    self._delete_user()

  def tearDown(self):
    settings.AUDIT_LOGGER = self.saved_audit_logger

class TestQueryRedacting(django.test.TestCase):
  def setUp(self):
    self.old_redactors = querylog.REDACTORS.copy()
    querylog.REDACTORS.clear()

  def tearDown(self):
    querylog.REDACTORS.extend(self.old_redactors)

  def get_query_from_call(self, fn: mock.Mock):
    return get_logged_field(fn, 'query')

  @mock.patch.object(settings,'AUDIT_LOGGER')
  def test_redacting(self, logger: mock.Mock):

    def redactor(query):
      if query == 'A':
        return True, 'REDACTED'
      else:
        return False, None

    querylog.REDACTORS.append(redactor)

    querylog.logAction(query='A')
    logger.info.assert_called_once()
    logged_query = self.get_query_from_call(logger.info)

    self.assertEquals(logged_query, 'REDACTED')

    logger.info.reset_mock()
    querylog.logAction(query='B')
    logger.info.assert_called_once()
    logged_query = self.get_query_from_call(logger.info)
    self.assertEquals(logged_query, 'B') # should stay unchanged

def get_logged_field(audit_logger_mock: mock.Mock, field_name: str):
  """
  returns the specified field of the audit long intercepted by the provided mock
  """
  logger_extra_args = audit_logger_mock.call_args_list[0].kwargs['extra']
  field_value = logger_extra_args['audit_record'][field_name]
  return field_value


class TestRequestUse(django.test.TestCase):
  '''
  Test how we use the "request" object passed to the audit logger
  '''

  @mock.patch.object(settings,'AUDIT_LOGGER')
  def test_user_and_ip_logging(self, logger: mock.Mock):

    logger.info.reset_mock()
    querylog.logAction(query='A')
    self.assertEquals(get_logged_field(logger.info, 'username'), '')
    self.assertEquals(get_logged_field(logger.info, 'user_id'), '')
    self.assertEquals(get_logged_field(logger.info, 'ip'), '')

    # request can be used to provide username and ip:
    logger.info.reset_mock()
    querylog.logAction(query='A', request=create_request(MockUser('user1', id=1), 'ip1'))
    self.assertEquals(get_logged_field(logger.info, 'username'), 'user1')
    self.assertEquals(get_logged_field(logger.info, 'user_id'), 1)
    self.assertEquals(get_logged_field(logger.info, 'ip'), 'ip1')

    # values passed explicitly have higher priority than the ones packed into request:
    logger.info.reset_mock()
    querylog.logAction(
      query='A',
      request=create_request(MockUser('user1', id=1), 'ip1'),
      username='user2',
      user_id=2,
      client_ip='ip2'
    )
    self.assertEquals(get_logged_field(logger.info, 'username'), 'user2')
    self.assertEquals(get_logged_field(logger.info, 'user_id'), 2)
    self.assertEquals(get_logged_field(logger.info, 'ip'), 'ip2')

class TestAccessInfoUse(django.test.TestCase):

  def tearDown(self):
    access_info.clear()

  @mock.patch.object(settings,'AUDIT_LOGGER')
  def test_access_info_use(self, logger: mock.Mock):
    '''
    if the logging action does not contain user and client_ip, the access_info storage should be consulted
    '''

    access_info.store(access_info.AccessInfo(username='user_a', user_id=111, client_ip='1.2.3.4'))

    logger.info.reset_mock()
    querylog.logAction(query='A')
    self.assertEquals(get_logged_field(logger.info, 'username'), 'user_a')
    self.assertEquals(get_logged_field(logger.info, 'user_id'), 111)
    self.assertEquals(get_logged_field(logger.info, 'ip'), '1.2.3.4')

    log = querylog.querylog_create(query='A')
    self.assertEquals(log.username, 'user_a')
    self.assertEquals(log.user_id, 111)
    self.assertEquals(log.client_ip, '1.2.3.4')

    log = querylog.actionlog_create(query='A')
    self.assertEquals(log.username, 'user_a')
    self.assertEquals(log.user_id, 111)
    self.assertEquals(log.client_ip, '1.2.3.4')

    with querylog.QueryLog(query='A') as log:
      self.assertEquals(log.username, 'user_a')
      self.assertEquals(log.user_id, 111)
      self.assertEquals(log.client_ip, '1.2.3.4')

if __name__ == '__main__':
  unittest.main()
