# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
RBAC Unit tests:

Basic tests:
  Perms from user
  Perms from group
  Multiple perms from multiple roles
  Dataset, Connection and System types
  Multiple perms from privilege lines
  Admin user has all permissions
  Everyone group contains all users
  Verify -1 = 'All Connections' and/or 'All Datasets'
  Verify 'requre_all_groups' function
  Verify Everyone group works as part of 'require_all_groups'

  Ensure dataset creator has full access (when still has explore on dataconn)
  Ensure dataconn creator has full access (when still has system editconn)

System level:
  views method
  Role update resets perm cache
  Dataset, Dataconnection object level calls
  Dataset, Dataconnection checkperm / exceptions
  Dataset, Dataconnection deletes are picked up in roles

Sentry tests:
  Basic sentry connection
  Sentry roles propagates with groups
  Sentry group change is propagated
  Sentry role delete removes sentry groups but leaves role
  Reconnect afer exception from server
  Verify connects with server version 1 and 2
  Verify server error is reported
  Ensure SENTRY['GROUP_NAME_IGNORE_CASE'] setting works

For the sentry tests, we mock the sentry server side to drive
new sentry roles and membership

Workspace tests:
  Public workspace has edit, view for everyone
  Empty ACL no permissions
  Validate ACL entry types (user/group, each level)

"""

import sys
import pdb
import time
import json
import unittest

from django.conf import settings
from django.contrib.auth.models import User, Group
from dataconn.sentryif import SentryIF
from rbac.permcache import RBAC, groups_for_user
from rbac.permdefs import PERMDEFS, PERMLISTS, PermissionException
from datasets.models import DataSet, DataConnection
from workspace.models import Workspace, WSACL_Member, WSACL_Level
from rbac.models import Role

# pylint: disable=W0401,W0614
import dataconn
sys.path.extend(dataconn.__path__)

import dataconn.thrift_sentry
sys.path.extend(dataconn.thrift_sentry.__path__)
from dataconn.thrift_sentry.sentry_policy_service import SentryPolicyService
from dataconn.thrift_sentry.sentry_policy_service.ttypes import *
from dataconn.thrift_sentry.sentry_common_service.constants import *
# pylint: enable=W0401,W0614

class MockLogger(object):
  def __init__(self):
    self.reset()

  def reset(self):
    self.lines = []

  def log(self, lvl, msg):
    self.lines.append((msg, lvl))

  def messages(self, reset=True):
    text = [t[0] for t in self.lines]
    if reset:
      self.reset()
    return text

class MockSentryPolicyService(SentryPolicyService.Iface):
  """ Move Sentry server class """

  def __init__(self):
    self.roles = []
    self.version = TSENTRY_SERVICE_V1
    self.message = None

  def config(self, roles, version=TSENTRY_SERVICE_V1, message=None):
    self.roles = roles
    self.version = version
    self.message = message

  def list_sentry_roles_by_group(self, req):
    if self.roles is None:
      raise Exception("Forced exception")

    if self.message is not None:
      rsp = TListSentryRolesResponse(
              status=TSentryResponseStatus(value=TSENTRY_STATUS_RUNTIME_ERROR,
                                           message=self.message))
      return rsp

    if req.protocol_version != self.version:
      rsp = TListSentryRolesResponse(
              status=TSentryResponseStatus(value=TSENTRY_STATUS_THRIFT_VERSION_MISMATCH,
                                           message="Interface version mismatch"))
      return rsp

    roles = []
    for r in self.roles:
      sr = TSentryRole(roleName=r[0],
                       grantorPrincipal='',
                       groups=[TSentryGroup(groupName=g) for g in r[1]])
      roles.append(sr)


    rsp = TListSentryRolesResponse(status=TSentryResponseStatus(value=TSENTRY_STATUS_OK),
                                   roles=roles)
    return rsp

class TestBasic(unittest.TestCase):

  #@classmethod
  #def setUpClass(cls):
  #  RBAC.pollstart(polltime=5)

  #@classmethod
  #def tearDownClass(cls):
  #  RBAC.pollstop()

  def setUp(self):
    self.objects = type('', (), {})
    o = self.objects

    o.ua = User.objects.create(username='admin', is_superuser=True)
    o.u1 = User.objects.create(username='user1')
    o.u2 = User.objects.create(username='user2')
    o.u3 = User.objects.create(username='user3')

    o.g1 = Group.objects.create(name='group1')
    o.g2 = Group.objects.create(name='group2')

    o.dc1 = DataConnection.objects.create(dataconnection_name='dataconn1', dataconnection_type = 'sqlite')
    o.dc2 = DataConnection.objects.create(dataconnection_name='dataconn2', dataconnection_type = 'sqlite')
    o.ds1 = DataSet.objects.create(dataset_name='dataset1', dataconnection=o.dc1,
        dataset_detail='test', dataset_type='test')
    o.ds2 = DataSet.objects.create(dataset_name='dataset2', dataconnection=o.dc2,
        dataset_detail='test', dataset_type='test')

    o.u1.groups.add(o.g1)
    o.u2.groups.add(o.g2)
    o.u1.save()
    o.u2.save()

    o.r1 = Role.objects.create(name='role1',
                               privileges=json.dumps([{'ptype':'system','perms':['sys_editperm']}]))
    o.r2 = Role.objects.create(name='role2',
                               privileges=json.dumps([{'ptype':'system','perms':['sys_editconn']}]))
    RBAC.pollstart(polltime=3)

  def tearDown(self):
    RBAC.pollstop()
    Role.objects.all().delete()
    DataSet.objects.all().delete()
    DataConnection.objects.all().delete()
    Group.objects.all().delete()
    User.objects.all().delete()

  # Check a list of strings vs a list of ints
  def oneset(self, lstr, lint):
    self.assertEqual(set(int(s) for s in lstr), set(lint))

  # Wait for a check not to assert
  def waitfor(self, func, limit=1):
    RBAC.pollwait(limit)
    start = time.time()
    while True:
      try:
        func()
        return
      except:
        if (time.time() - start) > limit:
          raise
      time.sleep(1)

  def test_1(self):
    # Perms from user
    o = self.objects
    self.assertFalse(o.r1.user_has_role(o.u1))
    o.r1.add_users(['user1','user2'])
    self.assertTrue(o.r1.user_has_role(o.u1))
    self.assertTrue(o.r1.user_has_role(o.u2))

  def test_2(self):
    # Perms from groups
    o = self.objects
    self.assertFalse(o.r1.user_has_role(o.u1))
    o.r1.add_groups(['group1', 'group2', 'othergroup'])
    self.assertTrue(o.r1.user_has_role(o.u1))
    self.assertTrue(o.r1.user_has_role(o.u2))
    self.assertFalse(o.r1.user_has_role(o.u3))

    o.r1.require_all_groups = True
    o.r1.save()
    self.assertFalse(o.r1.user_has_role(o.u1))
    o.u1.groups.add(o.g2)
    o.u1.save()
    if hasattr(o.u1, '_arcadia_groupset'):
      del o.u1._arcadia_groupset
    self.assertFalse(o.r1.user_has_role(o.u1))
    o.r1.groups = json.dumps(['group1', 'group2'])
    o.r1.save()
    self.assertTrue(o.r1.user_has_role(o.u1))
    o.r1.groups = json.dumps(['group1', 'group2', settings.EVERYONE_GROUP])
    o.r1.save()
    self.assertTrue(o.r1.user_has_role(o.u1))
    o.r1.groups = json.dumps(['group1', 'group2', settings.EVERYONE_GROUP, 'othergroup'])
    o.r1.save()
    self.assertFalse(o.r1.user_has_role(o.u1))
    o.u1.groups.remove(o.g2)
    o.u1.save()
    if hasattr(o.u1, '_arcadia_groupset'):
      del o.u1._arcadia_groupset
    self.assertFalse(o.r1.user_has_role(o.u1))
    o.r1.groups = json.dumps(['group2', settings.EVERYONE_GROUP])
    o.r1.save()
    self.assertFalse(o.r1.user_has_role(o.u1))
    o.r1.require_all_groups = False
    o.r1.save()
    self.assertTrue(o.r1.user_has_role(o.u1))

  def test_3(self):
    # Multiple perms from multiple roles
    o = self.objects
    self.assertFalse(o.r1.user_has_role(o.u1))
    o.r1.add_users(['user1','user3'])
    o.r2.add_groups(['group1', 'group2', 'othergroup'])

    self.assertEqual(RBAC.get_user_obj_perms(o.u1), set(['sys_editperm', 'sys_editconn', 'sys_viewroles']))
    self.assertEqual(RBAC.get_user_obj_perms(o.u2), set(['sys_editconn']))

  def test_4(self):
    # Dataset perms
    o = self.objects
    self.assertEqual(RBAC.get_user_obj_perms(o.u1, o.ds1), set([]))
    o.r1.add_priv('dataset', ['ds_appedit', 'ds_appview'],
                  dcid=o.dc1.id, dslist=[o.ds1.id])
    o.r1.add_users([o.u1.username])
    self.assertEqual(RBAC.get_user_obj_perms(o.u1, o.ds1), set(['ds_appedit', 'ds_appview']))
    self.assertEqual(RBAC.get_user_obj_perms(o.u2, o.ds1), set([]))
    self.assertEqual(RBAC.get_user_obj_perms(o.u1, o.ds2), set([]))

    # Data connection perm (+ forced permission)
    o.r1.add_priv('dataconn', ['dc_aviews'], dclist=[o.dc1.id])
    self.assertEqual(RBAC.get_user_obj_perms(o.u1, o.dc1), set(['dc_aviews', 'dc_explore']))
    self.assertEqual(RBAC.get_user_obj_perms(o.u2, o.dc1), set([]))
    self.assertEqual(RBAC.get_user_obj_perms(o.u1, o.dc2), set([]))

    # Multple perms from multiple privilege lines
    # -1 for 'All connections'
    o.r1.add_priv('dataconn', ['dc_upload'], dclist=[-1])
    self.assertEqual(RBAC.get_user_obj_perms(o.u1, o.dc1),
                     set(['dc_aviews', 'dc_upload', 'dc_explore']))

    # Everyone group (brings user3 into the role)
    self.assertEqual(RBAC.get_user_obj_perms(o.u3, o.dc1), set([]))
    o.r1.add_groups([settings.EVERYONE_GROUP])
    self.assertEqual(RBAC.get_user_obj_perms(o.u3, o.dc1),
                     set(['dc_aviews', 'dc_upload', 'dc_explore']))

    # -1 = 'All datasets'
    self.assertEqual(RBAC.get_user_obj_perms(o.u1, o.ds1), set(['ds_appedit', 'ds_appview']))
    o.r1.add_priv('dataset', ['ds_manage'], dcid=o.dc1.id, dslist=[-1])
    self.assertEqual(RBAC.get_user_obj_perms(o.u1, o.ds1),
                     set(['ds_appedit', 'ds_appview', 'ds_manage']))

    # -1 = 'All connections / All datasets' (+ forced)
    self.assertEqual(RBAC.get_user_obj_perms(o.u3, o.ds2), set([]))
    o.r1.add_priv('dataset', ['ds_manage'], dcid=-1, dslist=[-1])
    self.assertEqual(RBAC.get_user_obj_perms(o.u3, o.ds2), set(['ds_manage', 'ds_appedit', 'ds_appview']))

    # Admin user has all permissions
    self.assertEqual(RBAC.get_user_obj_perms(o.ua), set(PERMLISTS['system']))
    self.assertEqual(RBAC.get_user_obj_perms(o.ua, o.ds1), set(PERMLISTS['dataset']))
    self.assertEqual(RBAC.get_user_obj_perms(o.ua, o.dc1), set(PERMLISTS['dataconn']))

  def test_5(self):
    # Object level perms
    o = self.objects
    self.assertEqual(RBAC.get_user_obj_perms(o.u1, o.ds1), set([]))
    o.r1.add_priv('dataconn', ['dc_explore'], dclist=[o.dc1.id])
    o.r1.add_priv('dataset', ['ds_appedit', 'ds_appview'],
                  dcid=o.dc1.id, dslist=[o.ds1.id])
    o.r1.add_users([o.u1.username])

    self.assertEqual(o.dc1.getPerms(o.u1), set(['dc_explore']))
    self.assertEqual(o.ds1.getPerms(o.u1), set(['ds_appedit', 'ds_appview']))

    o.dc1.checkPerm(o.u1, 'dc_explore')
    o.ds1.checkPerm(o.u1, 'ds_appedit')

    exc = None
    try:
      o.dc1.checkPerm(o.u2, 'dc_explore')
    except PermissionException as e:
      exc = e
    self.assertTrue(isinstance(exc, PermissionException))

    exc = None
    try:
      o.ds1.checkPerm(o.u2, 'ds_appedit')
    except PermissionException as e:
      exc = e
    self.assertTrue(isinstance(exc, PermissionException))


  def test_6(self):
    o = self.objects
    o.r1.add_priv('dataconn', ['dc_explore'], dclist=[o.dc1.id, o.dc2.id])
    self.oneset(o.r1.privList()[1].get('dclist'), [o.dc1.id, o.dc2.id])

    # Deleted connection gets picked up in role
    o.dc2.delete()
    del o.dc2
    def check1():
      o.r1 = Role.objects.get(id=o.r1.id)
      self.oneset(o.r1.privList()[1].get('dclist'), [o.dc1.id])
    self.waitfor(check1, 15)

    # Next delete removes privilege entirely
    o.dc1.delete()
    del o.dc1
    def check2():
      o.r1 = Role.objects.get(id=o.r1.id)
      self.assertEqual(len(o.r1.privList()), 1)
    self.waitfor(check2, 15)

  def test_7(self):
    o = self.objects
    o.ds1a = DataSet.objects.create(dataset_name='dataset1a', dataconnection=o.dc1,
        dataset_detail='testa', dataset_type='testa')
    o.r1.add_priv('dataset', ['ds_appedit', 'ds_appview'],
                  dcid=o.dc1.id, dslist=[o.ds1.id, o.ds1a.id])
    self.oneset(o.r1.privList()[1].get('dslist'), [o.ds1.id, o.ds1a.id])

    # Changed data connection creates new privilege role
    o.ds1a.dataconnection_id = o.dc2.id
    o.ds1a.save()
    def check1():
      o.r1 = Role.objects.get(id=o.r1.id)
      self.assertEqual(len(o.r1.privList()), 3)
    self.waitfor(check1, 15)

    # Deleted dataset gets picked up in role
    o.ds1a.delete()
    del o.ds1a
    def check2():
      o.r1 = Role.objects.get(id=o.r1.id)
      self.assertEqual(len(o.r1.privList()), 2)
      self.oneset(o.r1.privList()[1].get('dslist'), [o.ds1.id])
    self.waitfor(check2, 15)

    # Next delete removes privilege entirely
    o.ds1.delete()
    del o.ds1
    def check2():
      o.r1 = Role.objects.get(id=o.r1.id)
      self.assertEqual(len(o.r1.privList()), 1)
    self.waitfor(check2, 15)

  def test_8(self):
    o = self.objects
    o.r1.add_priv('dataset', ['ds_appedit', 'ds_appview'],
                  dcid=o.dc1.id, dslist=[o.ds1.id, -1])
    self.oneset(o.r1.privList()[1].get('dslist'), [o.ds1.id, -1])

    # Deleted connection removes dataset privilege entirely
    o.dc1.delete()
    del o.dc1
    def check1():
      o.r1 = Role.objects.get(id=o.r1.id)
      self.assertEqual(len(o.r1.privList()), 1)
    self.waitfor(check1, 15)

  def test_9(self):
    o = self.objects
    o.r3 = Role.objects.create(name='role2',
                               privileges=json.dumps([{'ptype':'system','perms':['sys_editconn']}]))
    o.r1.add_priv('role', ['rgds_manage'], rolelist=[o.r2.id, o.r3.id])
    self.oneset(o.r1.privList()[1].get('rolelist'), [o.r2.id, o.r3.id])

    # Deleted role gets picked up in role
    def delete_r3():
      o.r3.delete()
    self.waitfor(delete_r3, 15)
    del o.r3
    def check1():
      o.r1 = Role.objects.get(id=o.r1.id)
      self.oneset(o.r1.privList()[1].get('rolelist'), [o.r2.id])
    self.waitfor(check1, 15)

    # Next delete removes privilege entirely
    def delete_r2():
      o.r2.delete()
    self.waitfor(delete_r2, 15)
    del o.r2
    def check2():
      o.r1 = Role.objects.get(id=o.r1.id)
      self.assertEqual(len(o.r1.privList()), 1)
    self.waitfor(check2, 15)

  def test_a1(self):
    o = self.objects
    u = o.u1
    ds = o.ds1

    # No perms on dataset
    self.assertEqual(ds.getPerms(u), set())

    # Add ds_appview
    o.r1.add_users([u.username])
    o.r1.add_priv('dataset', ['ds_appview'], dcid=ds.dataconnection_id, dslist=[ds.id])
    self.assertEqual(ds.getPerms(u), set(['ds_appview']))

    # Make user the dataset creator
    ds.create_user = u.username

    # No change in perms
    self.assertEqual(ds.getPerms(u), set(['ds_appview']))

    # Now give explore on the data connection
    o.r2.add_users([u.username])
    o.r2.add_priv('dataconn', ['dc_explore'], dclist=[ds.dataconnection.id])

    # Should have all perms
    self.assertEqual(ds.getPerms(u), RBAC.all_perms_for_obj(ds))

    # Now remove explore and verify back to single perm
    o.r2.delete_users([u.username])
    self.assertEqual(ds.getPerms(u), set(['ds_appview']))


  def test_a2(self):
    o = self.objects
    u = o.u1
    dc = o.dc1

    # No perms on dataset
    self.assertEqual(dc.getPerms(u), set())

    # Add ds_appview
    o.r1.add_users([u.username])
    o.r1.add_priv('dataconn', ['dc_explore'], dclist=[dc.id])
    self.assertEqual(dc.getPerms(u), set(['dc_explore']))

    # Make user the data connection creator
    dc.create_user = u.username

    # No change in perms
    self.assertEqual(dc.getPerms(u), set(['dc_explore']))

    # Now give system level manage data connections
    o.r2.add_users([u.username])
    o.r2.add_priv('system', ['sys_editconn'])

    # Should have all perms
    self.assertEqual(dc.getPerms(u), RBAC.all_perms_for_obj(dc))

    # Now remove connection edit and verify back to single perm
    o.r2.delete_users([u.username])
    self.assertEqual(dc.getPerms(u), set(['dc_explore']))

  # Workspace tests
  def test_ws1(self):
    ws1 = Workspace.objects.get(id=1)
    ws2 = Workspace.objects.create(name='ws2')
    u1 = User.objects.get(username='user1')
    u2 = User.objects.get(username='user2')
    u3 = User.objects.get(username='user3')
    g1 = Group.objects.get(name='group1')
    g2 = Group.objects.get(name='group2')

    self.assertEqual(ws1.getPerms(u1), set(['ws_appedit', 'ws_appview']))
    self.assertEqual(ws2.getPerms(u1), set())

    acl = [[WSACL_Member.USER, WSACL_Level.VIEW, u1.username]]
    ws2.acl_obj = acl
    ws2.save()
    self.assertEqual(ws2.getPerms(u1), set(['ws_appview']))

    acl.append([WSACL_Member.GROUP, WSACL_Level.EDIT, g2.name])
    ws2.acl_obj = acl
    ws2.save()
    self.assertEqual(ws2.getPerms(u1), set(['ws_appview']))
    self.assertEqual(ws2.getPerms(u2), set(['ws_appedit', 'ws_appview']))

    dsp = set(['ds_manage', 'ds_appedit', 'ds_appview'])
    self.assertEqual(ws2.maskPerms(u1, dsp), set(['ds_manage', 'ds_appview']))
    self.assertEqual(ws2.maskPerms(u2, dsp), set(['ds_manage', 'ds_appedit', 'ds_appview']))
    self.assertEqual(ws2.maskPerms(u3, dsp), set(['ds_manage']))


class TestSentryIF(unittest.TestCase):

  @classmethod
  def setUpClass(cls):
    # Set up sentry
    cls.client = MockSentryPolicyService()
    cls.logger = MockLogger()
    cls.conf = {
      "HOST":'host',
      "AUTH":'auth',
      "POLL_FREQUENCY":'1',
    }
    cls.sif = SentryIF(cls.conf, cls.logger, cls.client)

  # Wait up to some limit for log entries to appear
  def waitlog(self, limit=5, msgs=None):
    if msgs is None:
      time.sleep(limit)
      return
    start = time.time()
    now = start
    found = False
    log = ''
    while not found and ((now - start) < limit):
      time.sleep(.1)
      now = time.time()
      found = True
      log += "\n".join(self.logger.messages())
      for m in msgs:
        if log.find(m) == -1:
          found = False
          break

    self.assertTrue(found,
                    "Timeout waiting for log:\n\n%s\n\nTo Contain\n\n  %s" %
                    (log, "\n  ".join(msgs)))

  def test_1(self):
    self.sif.activate()
    self.waitlog(10, ['Connected!'])

  def test_s2(self):
    self.client.config([('sentry_role1', ['group1','group2']),
                        ('sentry_role2', ['group3','group4'])])
    self.waitlog(10, ["Created role 'sentry_role1'",
                     "Created role 'sentry_role2'"])

  def test_s3(self):
    self.client.config([('sentry_role1', ['group1']),
                        ('sentry_role2', ['group3','group5'])])
    self.waitlog(10, ["Updated role 'sentry_role1'",
                     "Updated role 'sentry_role2'"])

    r1 = Role.objects.get(name='sentry_role1')
    r2 = Role.objects.get(name='sentry_role2')

    self.assertEqual(r1.sentryGroupSet(), set(['group1']))
    # print('here', r2.sentryGroupSet())
    self.assertEqual(r2.sentryGroupSet(), set(['group3', 'group5']))

  def test_s4(self):
    self.client.config([('sentry_role1', ['group1'])])
    self.waitlog(10, ["Clearing groups"])

    r1 = Role.objects.get(name='sentry_role1')
    r2 = Role.objects.get(name='sentry_role2')

    self.assertEqual(r1.sentryGroupSet(), set(['group1']))
    self.assertEqual(r2.sentryGroupSet(), set([]))

  def test_s5(self):
    self.client.config(None)
    self.waitlog(4, ["Exception"])

    self.client.config([], version=TSENTRY_SERVICE_V2, message="Forced Error")
    self.waitlog(4, ["Forced Error"])

    self.client.config([('sentry_role1', ['group1'])], version=TSENTRY_SERVICE_V2)
    self.waitlog(4, ["Connected!"])

    r1 = Role.objects.get(name='sentry_role1')
    r2 = Role.objects.get(name='sentry_role2')

    self.assertEqual(r1.sentryGroupSet(), set(['group1']))
    self.assertEqual(r2.sentryGroupSet(), set([]))

  def test_s6(self):
    u1 = User.objects.create(username='user1')
    u2 = User.objects.create(username='user2')

    g1 = Group.objects.create(name='group1')
    g2 = Group.objects.create(name='GROUP2')

    u1.groups.add(g1)
    u2.groups.add(g2)

    self.client.config([('sentry_role1', ['group1', 'group3']),
                        ('sentry_role2', ['group2','group1'])])
    self.waitlog(10, ["Updated role 'sentry_role1'",
                     "Updated role 'sentry_role2'"])

    r1 = Role.objects.get(name='sentry_role1')
    r2 = Role.objects.get(name='sentry_role2')

    self.assertEqual(r1.sentryGroupSet(), set(['group1', 'group3']))
    self.assertEqual(r2.sentryGroupSet(), set(['group2', 'group1']))

    RBAC.reset()
    self.assertEqual(set(['sentry_role1', 'sentry_role2']), set(r.name for r in RBAC.roles_for_user(u1)))
    self.assertEqual(set([]), set(r.name for r in RBAC.roles_for_user(u2)))

    RBAC.reset()
    settings.SENTRY = getattr(settings, 'SENTRY', {})
    settings.SENTRY['GROUP_NAME_IGNORE_CASE'] = True
    self.assertEqual(set(['sentry_role1', 'sentry_role2']), set(r.name for r in RBAC.roles_for_user(u1)))
    self.assertEqual(set(['sentry_role2']), set(r.name for r in RBAC.roles_for_user(u2)))

if __name__ == '__main__':
  # unittest.main()
  pass
