# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import unittest

import json

from reports.models import Report
from reports.models import ReportUser
from reports.models import ReportImage
from reports.models import ReportUsage
from django.core.exceptions import PermissionDenied

# TODO: linked deletion from dash when in multiple sheets (w/ model update)

_test_objects = []
DEPENDENT_CLASSES = [ReportImage, ReportUsage, ReportUser]

def Dashboard(*sheets):
    ds = []
    svd = []
    for sheet in sheets:
        cw = []
        vw = []
        for viz in sheet:
           if viz.report_type == 'visual':
               vw.append({'id': 'uri-{0}-widget-{1}'.format(len(vw), viz.id)})
           else:
               cw.append(viz.id)

           if viz.category() == 'linked':
               svd.append(viz.id)


        ds.append({'control_widgets': cw, 'visual_widgets': vw})

    config = {
        'report_type': 'dashboard',
        'report_data': json.dumps({'dashboard_sheets': ds}),
        'shared_visual_dashboards': json.dumps(svd),
    }

    dash = Report.objects.create(**config)
    create_dependents(dash)

    for sheet in sheets:
        for viz in sheet:
            if viz.category() == 'linked':
                s = viz.shared_dashboards()
                s.append(dash.id)
                viz.shared_dashboards(s)
                viz.save()

    _test_objects.append(dash)
    return dash


def LinkedVisual():
    config = {
        'report_type': 'visual',
        'report_data': json.dumps({'report_data': {}}),
        'shared_visual_dashboards': '[]'
    }

    linked = Report.objects.create(**config)
    create_dependents(linked)
    _test_objects.append(linked)
    return linked


def UnlinkedVisual():
    config = {
        'report_type': 'visual',
        'report_data': json.dumps({'report_data': {}}),
    }

    unlinked = Report.objects.create(**config)
    create_dependents(unlinked)
    _test_objects.append(unlinked)
    return unlinked


def Picker():
    config = {
        'report_type': 'picker',
        'report_data': json.dumps({'report_data': {}}),
    }

    picker = Report.objects.create(**config)
    create_dependents(picker)
    _test_objects.append(picker)
    return picker


def ensure_delete(*args):
    _flatten = lambda lst: [item for sublist in lst for item in sublist]
    for report in _flatten([arg if hasattr(arg, '__iter__') else [arg] for arg in args]):
        if report.id is not None:
            try:
                report.report_type = 'table'
                report.shared_visual_dashboards = None
                report.delete()
            except:
                pass


def create_dependents(report):
    for cls in DEPENDENT_CLASSES:
        r = cls.objects.create(report_id=report.id)
        _test_objects.append(r)


def refresh(obj):
    if obj.id:
        return Report.objects.filter(id=obj.id).first()


class TestDashboardAllReportIds(unittest.TestCase):

    def tearDown(self):
        ensure_delete(*_test_objects)

    def test_empty(self):
        empty = Dashboard()
        ids = empty.all_report_ids()
        self.assertTrue(ids == [])

    def test_single_sheet(self):
        s1 = [LinkedVisual(), UnlinkedVisual(), Picker()]
        ss_dash = Dashboard(s1)
        ids = ss_dash.all_report_ids()
        self.assertTrue(set(ids) == set([r.id for r in s1]))

    def test_multi_sheet(self):
        s1 = [LinkedVisual(), UnlinkedVisual(), Picker()]
        s2 = [LinkedVisual(), UnlinkedVisual(), Picker()]
        ms_dash = Dashboard(s1, s2)
        ids = ms_dash.all_report_ids()
        self.assertTrue(set(ids) == set([r.id for r in s1 + s2]))


class TestPermitCategory(unittest.TestCase):

    def tearDown(self):
        ensure_delete(*_test_objects)

    def test_dashboard(self):
        dash = Dashboard()
        self.assertRaises(TypeError, dash.is_unused)

    def test_linked(self):
        linked = LinkedVisual()
        self.assertRaises(TypeError, linked.all_report_ids)
        self.assertRaises(TypeError, linked.all_sheets)
        self.assertRaises(TypeError, linked.remove_visual)

    def test_unlinked(self):
        unlinked = UnlinkedVisual()
        self.assertRaises(TypeError, unlinked.shared_dashboards)
        self.assertRaises(TypeError, unlinked.is_unused)
        self.assertRaises(TypeError, unlinked.remove_shared)
        self.assertRaises(TypeError, unlinked.all_report_ids)
        self.assertRaises(TypeError, unlinked.all_sheets)
        self.assertRaises(TypeError, unlinked.remove_visual)


class TestRemoveShared(unittest.TestCase):

    def tearDown(self):
        ensure_delete(*_test_objects)

    def test_removes_single_sheet_dashboard(self):
        linked = LinkedVisual()
        dash = Dashboard([linked])
        linked.remove_shared(dash)
        self.assertTrue(linked.shared_dashboards() == [])

    def test_removes_multi_sheet_dashboard(self):
        linked = LinkedVisual()
        other_linked = LinkedVisual()
        dash = Dashboard([other_linked], [linked])
        self.assertTrue(dash.id in linked.shared_dashboards())
        linked.remove_shared(dash)
        self.assertFalse(dash.id in linked.shared_dashboards())

    def test_does_nothing_when_dashboard_empty(self):
        dash = Dashboard()
        linked = LinkedVisual()
        self.assertTrue(linked.shared_dashboards() == [])
        linked.remove_shared(dash)
        self.assertTrue(linked.shared_dashboards() == [])

    def test_does_nothing_when_does_not_contain_passed_dashboard(self):
        linked = LinkedVisual()
        other_linked = LinkedVisual()
        other_dash = Dashboard([other_linked])
        dash = Dashboard([linked])
        self.assertTrue(dash.id in linked.shared_dashboards())
        linked.remove_shared(other_dash)
        self.assertTrue(dash.id in linked.shared_dashboards())

    def test_handles_same_linked_viz_in_multiple_sheets(self):
        linked = LinkedVisual()
        dash = Dashboard([linked], [linked])
        self.assertTrue(dash.id in linked.shared_dashboards())
        linked.remove_shared(dash)
        self.assertTrue(dash.id not in linked.shared_dashboards())


class TestDashboardRemoveVisual(unittest.TestCase):

    def tearDown(self):
        ensure_delete(*_test_objects)

    def test_removes_linked_visual_from_widgets_and_shared(self):
        l1 = LinkedVisual()
        l2 = LinkedVisual()
        dash = Dashboard([l1], [l2])
        dash.remove_visual(l1)
        dash.remove_visual(l2)
        self.assertTrue(l1.id not in dash.all_report_ids())
        self.assertTrue(l1.id not in dash.shared_dashboards())
        self.assertTrue(l2.id not in dash.all_report_ids())
        self.assertTrue(l2.id not in dash.shared_dashboards())

    def test_removes_unlinked_from_widgets(self):
        u1 = UnlinkedVisual()
        u2 = UnlinkedVisual()
        dash = Dashboard([u1], [u2])


class TestReportDeletionBehavior(unittest.TestCase):

    def tearDown(self):
        ensure_delete(*_test_objects)

    def test_prevents_deletion_of_linked_viz_when_in_dashboard(self):
        linked = LinkedVisual()
        dash = Dashboard([linked])
        self.assertRaises(Exception, linked.delete)

    def test_allows_linked_viz_delete_after_containing_dashboard_delete(self):
        linked = LinkedVisual()
        dash = Dashboard([linked])
        self.assertRaises(Exception, linked.delete)
        dash.delete()
        try:
            refresh(linked).delete()
        except PermissionDenied as err:
            self.fail(err.message)

    # TODO upgrade python so this can be an expected failure decorator
    # def test_linked_deletion_from_dash_removes_reference_preserves_linked(self):
    #     l1 = LinkedVisual()
    #     l2 = LinkedVisual()
    #     dash1 = Dashboard([l1])
    #     dash2 = Dashboard([l1], [l2])
    #     dash3 = Dashboard([l1, l2], [l1, l2])
    #     l1.delete(dash_id=dash1.id)
    #     self.assertTrue(l1.id is not None, 'single sheet')
    #     l2.delete(dash_id=dash2.id)
    #     self.assertTrue(l1.id is not None, 'simple multi sheet')
    #     self.assertTrue(l2.id is not None, 'simple multi sheet')
    #     self.assertTrue(l1.id in dash2.all_report_ids(), 'simple multi sheet')
    #     self.assertTrue(l2.id not in dash2.all_report_ids(), 'simple multi sheet')
    #     # probably fail
    #     l1.delete(dash_id=dash3.id)
    #     self.assertTrue(l1.id is not None, 'complex multi sheet')
    #     self.assertTrue(l2.id is not None, 'complex multi sheet')
    #     self.assertTrue(dash3.all_report_ids().count(l1.id) == 1, 'complex multi sheet')
    #     self.assertTrue(dash3.all_report_ids().count(l2.id) == 2, 'complex multi sheet')

    def test_dashboard_delete_deletes_all_member_unlinked(self):
        ul1 = UnlinkedVisual()
        ul2 = UnlinkedVisual()
        ul3 = UnlinkedVisual()
        dash1 = Dashboard([ul1])
        dash2 = Dashboard([ul2], [ul3])
        dash1.delete()
        dash2.delete()
        self.assertTrue(refresh(ul1) is None, 'single sheet dashboard')
        self.assertTrue(refresh(ul3) is None, 'multi sheet dashboard')
        self.assertTrue(refresh(ul2) is None, 'multi sheet dashboard')

    def test_dashboard_delete_preserves_linked_visuals(self):
        l1 = LinkedVisual()
        l2 = LinkedVisual()
        dash1 = Dashboard([l1])
        dash2 = Dashboard([l1], [l2])
        dash3 = Dashboard([l1, l2], [l1, l2])
        dash1.delete()
        self.assertTrue(refresh(l1) is not None, 'single sheet')
        dash2.delete()
        self.assertTrue(refresh(l1) is not None, 'simple multi sheet')
        self.assertTrue(refresh(l2) is not None, 'simple multi sheet')
        dash3.delete()
        self.assertTrue(refresh(l1) is not None, 'complex multi sheet')
        self.assertTrue(refresh(l2) is not None, 'complex multi sheet')

    def test_dashboard_deletion_removes_reference_from_linked_viz(self):
        l1 = LinkedVisual()
        l2 = LinkedVisual()
        dash1 = Dashboard([l1])
        dash2 = Dashboard([l1], [l2])
        dash3 = Dashboard([l1, l2], [l1, l2])
        id1 = dash1.id
        id2 = dash2.id
        id3 = dash3.id
        dash1.delete()
        self.assertTrue(id1 not in refresh(l1).shared_dashboards(), 'single sheet')
        dash2.delete()
        self.assertTrue(id2 not in refresh(l1).shared_dashboards(), 'simple multi sheet')
        self.assertTrue(id2 not in refresh(l2).shared_dashboards(), 'simple multi sheet')
        dash3.delete()
        self.assertTrue(id3 not in refresh(l1).shared_dashboards(), 'complex multi sheet')
        self.assertTrue(id3 not in refresh(l2).shared_dashboards(), 'complex multi sheet')

    def test_unlinked_visual_deletion_removes_reference_from_dashboard(self):
        u1 = UnlinkedVisual()
        u2 = UnlinkedVisual()
        u3 = UnlinkedVisual()
        dash1 = Dashboard([u1])
        dash2 = Dashboard([u2], [u3])
        id1 = u1.id
        id3 = u3.id
        u1.delete(dash_id=dash1.id)
        u3.delete(dash_id=dash2.id)
        self.assertFalse(id1 in refresh(dash1).all_report_ids(), 'single sheet dashboard')
        self.assertFalse(id3 in refresh(dash2).all_report_ids(), 'multi sheet dashboard')

    def test_picker_deletion_removes_reference_from_dashboard(self):
        u1 = Picker()
        u2 = Picker()
        u3 = Picker()
        dash1 = Dashboard([u1])
        dash2 = Dashboard([u2], [u3])
        id1 = u1.id
        id3 = u3.id
        u1.delete(dash_id=dash1.id)
        u3.delete(dash_id=dash2.id)
        self.assertFalse(id1 in refresh(dash1).all_report_ids(), 'single sheet dashboard')
        self.assertFalse(id3 in refresh(dash2).all_report_ids(), 'multi sheet dashboard')

    def test_delete_removes_all_dependent_objects(self):
        lst = [Picker(), Dashboard(), LinkedVisual(), UnlinkedVisual()]
        ids = [r.id for r in lst]

        for report in lst:
            report.delete()

        for cls in DEPENDENT_CLASSES:
            self.assertTrue(cls.objects.filter(id__in=ids).first() is None)


if __name__ == '__main__':
    unittest.main()
