# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import json
from typing import List
import django.test
from django.test.client import Client
from django.contrib.auth.models import User
from util.ff_flipper import FeatureFlipper
from django.urls import reverse
from reports.models import Report
from urllib.parse import urlencode

from rbac.models import Role


def create_client(user: User):
    username = user.username
    client = Client()
    client.login(username=username, password=username)
    return client


def make_user(username, is_superuser=False) -> User:
    u = User()
    u.username = username
    u.set_password(username)
    u.is_superuser = is_superuser
    u.save()
    return u


def dataset_payload(ds_name, dc_id, dataset_type='singletable'):

    table_1 = 'default.table_1'
    tables = [{'tablename': table_1, 'columns': [{'name': 'a', 'type': 'STRING'}]}]

    tables_sql = [{
                    'tablename': 'text.sql',
                    'select': 'select * from main.abc',
                    'columns': [
                        {'name': 'col1', 'alias': 'col1', 'type': 'STRING', 'isdim': True},
                        {'name': 'col2', 'alias': 'col2', 'type': 'REAL', 'isdim': False}
                    ]
                }]

    if dataset_type == 'singletable':
        dataset_detail = table_1
    elif dataset_type == 'multitable':
        dataset_detail = table_1
        tables.append({
            'tablename': 'default.table_2',
            'columns': [{'name': 'b', 'type': 'INT'}]
        })
    elif dataset_type == 'singletable-sql':
        dataset_type = 'singletable'
        dataset_detail = 'Created from SQL'
        tables = tables_sql
    elif dataset_type == 'multitable-sql':
        dataset_type = 'multitable'
        dataset_detail = 'Created from SQL'
        tables = tables_sql
        tables.append({
            'tablename': 'default.table_2',
            'columns': [{'name': 'b', 'type': 'INT'}]
        })
    else:
        raise Exception(f'unknown dataset type "{dataset_type}"')


    payload = {
        'dataconnection_id': dc_id,
        'dataset_name': ds_name,
        'dataset_type': dataset_type,
        'dataset_info': json.dumps(tables),
        'dataset_detail': dataset_detail,
    }

    return payload

def make_mixed_perms_role(dc_id: str, ds_manage_id: str, ds_view_id: str):
    role = Role()
    role.name = 'mixed_ds_perms'
    role.privileges = json.dumps(
        [
            # no system permissions:
            {'ptype': 'system', 'perms': []},
            # full rights on all dataconnections:
            {
                'ptype': 'dataconn',
                'dclist': ['-1'],
                'perms': ['dc_upload', 'dc_explore', 'dc_aviews'],
            },
            # view rights for one dataset
            {
                'ptype': 'dataset',
                'dcid': str(dc_id),
                'dslist': [str(ds_view_id)],
                'perms': ['ds_appview'],
            },
            # full rights for the other
            {
                'ptype': 'dataset',
                'dcid': str(dc_id),
                'dslist': [str(ds_manage_id)],
                'perms': ['ds_manage', 'ds_appview', 'ds_appedit'],
            },
        ]
    )
    role.save()
    return role


class TestDashboardOperations(django.test.TestCase):
    def setUp(self):

        admin = make_user('admin', is_superuser=True)
        admin_client = create_client(admin)
        self.admin_client = admin_client

        user = make_user('user1', is_superuser=False)
        user_client = create_client(user)
        self.user = user
        self.user_client = user_client

        dc_id = self.create_dataconnection(admin_client)

        ds1_id = self.create_dataset(admin_client, dc_id, 'ds1')
        ds2_id = self.create_dataset(admin_client, dc_id, 'ds2')

        self.setup_user_role(user, dc_id, ds1_id, ds2_id)

        visual_1 = self._create_visual(ds1_id)
        visual_11 = self._create_visual(ds1_id)
        visual_2 = self._create_visual(ds2_id)

        # dashboard referencing only dataset 1
        db_homogenous_id = self._create_dashboard(
            dataset_id=ds1_id, visuals=[visual_11]
        )
        self._move_to_public_workspace(dashboard_id=db_homogenous_id)
        self.db_homogenous_id = db_homogenous_id

        # dashboard referencing both dataset 1 and 2
        db_mixed_id = self._create_dashboard(
            dataset_id=ds1_id, visuals=[visual_1, visual_2]
        )
        self._move_to_public_workspace(dashboard_id=db_mixed_id)
        self.db_mixed_id = db_mixed_id

    def test_operations(self):
        user = self.user
        user_client = self.user_client
        db_homogenous_id = self.db_homogenous_id
        db_mixed_id = self.db_mixed_id

        # test dashboard cloning:
        self._test_clone(user, user_client, db_homogenous_id)

        # test delete: deleting the homogenous report should succeed:
        response = user_client.delete(f'/reports/report/{db_homogenous_id}')
        self.assertEquals(response.status_code, 200)

        # test delete: the user does not have the rights to delete visual_2
        # as that belongs to dataset 2 for which he only has view permissions
        response = user_client.delete(f'/reports/report/{db_mixed_id}')
        self.assertNotEquals(response.status_code, 200)

    def test_clone_badge(self):
        user = self.user
        user_client = self.user_client
        src_db_id = self.db_homogenous_id

        FeatureFlipper().set_flipper('enable_clone_info_badge', True, 'DB')
        src_data_1 = user_client.get(f'/reports/report/{src_db_id}?getcloneinfo=true').json()[0]
        self.assertEquals(src_data_1['descendants_count'], 0)
        self.assertFalse(src_data_1['is_clone'])

        # perform cloning
        cloned1_db_id = self._clone_dashboard(user_client, src_db_id)

        cloned1_data_1 = user_client.get(f'/reports/report/{cloned1_db_id}?getcloneinfo=true').json()[0]
        self.assertEquals(cloned1_data_1['descendants_count'], 0)
        self.assertTrue(cloned1_data_1['is_clone'])

        src_data_2 = user_client.get(f'/reports/report/{src_db_id}?getcloneinfo=true').json()[0]
        self.assertEquals(src_data_2['descendants_count'], 1)
        self.assertFalse(src_data_2['is_clone'])

        # clone the clone:
        cloned2_db_id = self._clone_dashboard(user_client, cloned1_db_id)
        cloned2_data_1 = user_client.get(f'/reports/report/{cloned2_db_id}?getcloneinfo=true').json()[0]
        self.assertEquals(cloned2_data_1['descendants_count'], 0)
        self.assertTrue(cloned2_data_1['is_clone'])

        # check the parent:
        cloned1_data_2 = user_client.get(f'/reports/report/{cloned1_db_id}?getcloneinfo=true').json()[0]
        self.assertEquals(cloned1_data_2['descendants_count'], 1)
        self.assertTrue(cloned1_data_2['is_clone'])

    def _create_visual(self, dataset_id):
        return self.create_visual(self.admin_client, dataset_id)

    def create_visual(self, client: Client, dataset_id, visual_type='table') -> int:

        payload = {'workspace_id': 1, 'visual_type': visual_type}
        response = client.post(f'/reports/create/{dataset_id}', payload)
        self.assertEquals(response.status_code, 200)
        visual_id = response.json()[0]['id']

        return visual_id

    def _create_dashboard(self, dataset_id, visuals: List[int]):
        return self.create_dashboard(self.admin_client, dataset_id, visuals)

    def create_dashboard(self, client: Client, dataset_id, visuals: List[int], report_name='Untitled') -> int:

        payload = {
            'dashboard_name': report_name,
            'dashboard_dataset_id': dataset_id,
            'unshared_visuals': json.dumps(visuals),
            'dashboard_data': json.dumps(
                {
                    'dashboard_sheets': [
                        {
                            'visual_widgets': [
                                {'id': 'uri-1-widget-%s' % visual_id}
                                for visual_id in visuals
                            ]
                        }
                    ]
                }
            ),
        }

        response = client.post('/reports/savedashboard', payload)
        self.assertEquals(response.status_code, 200)
        dashboard_id = response.json()['id']
        return dashboard_id

    def _clone_dashboard(self, client: Client, db_id, changed_visual_data=[]):
        db: Report = Report.objects.get(id=db_id)
        (visuals, _) = db.get_visuals_in_dashboard()

        clone_data = {
            "parent_dashboard_id":db_id,
            "shared_visuals_to_clone":[],
            "not_saved_report_data_of_original_dashboard": changed_visual_data
        }
        payload = {
            'dashboard_dataset_id': db.dataset.id,
            'unshared_visuals': json.dumps(visuals),
            'dashboard_data': json.dumps(
                {
                    'dashboard_sheets': [
                        {
                            'visual_widgets': [
                                {'id': 'uri-1-widget-%s' % visual_id}
                                for visual_id in visuals
                            ]
                        }
                    ]
                }
            ),
            'clone_data': json.dumps(clone_data)
        }

        response = client.post('/reports/savedashboard', payload)
        self.assertEquals(response.status_code, 200)
        dashboard_id = response.json()['id']
        return dashboard_id

    def create_unsaved_report_data(self, dashboard_id: int, report: Report, report_description: str):
        return {
            "report_data": json.dumps(
                {
                    "report_data": {
                        "report_title":"new report title",
                        "dashboard_id": dashboard_id,
                    }
                }),
            "dataset_id": report.dataset.id,
            "report_type": "table",
            "report_id":  report.id,
            "report_description":report_description,
            "report_search_text":None,
        }

    def _test_clone(self, user: User, user_client: Client, db_id):
        (visualIds, _) = Report.objects.get(id=db_id).get_visuals_in_dashboard()
        self.assertEquals(visualIds, [2])
        originalVisual =  Report.objects.get(id=visualIds[0])
        new_report_description = "not_saved_report_description"

        unsaved_report_data = self.create_unsaved_report_data(db_id, originalVisual, new_report_description)

        clone_db_id = self._clone_dashboard(user_client, db_id, [unsaved_report_data])
        clone_db: Report = Report.objects.get(id=clone_db_id)
        (visuals, _) = clone_db.get_visuals_in_dashboard()
        self.assertEquals(len(visuals), 1)

        clone_visual: Report = Report.objects.get(id=visuals[0])
        new_report_data_after_clone =  '{"report_data": {"report_title":"new report title", "dashboard_id":' + str(clone_db.id) + '}}'
        self.assertNotEqual(clone_visual.id, originalVisual.id)
        # The user can request cloning an dashboard with unsaved modification like color change etc.
        # These changes should be reflected in the the cloned visual and keep the original visual untouched!
        self.assertNotEqual(json.loads(originalVisual.report_data), json.loads(new_report_data_after_clone))
        self.assertEquals(json.loads(clone_visual.report_data), json.loads(new_report_data_after_clone))
        self.assertEquals(clone_visual.report_description, new_report_description)
        self.assertNotEqual(originalVisual.report_description, new_report_description)
        self.assertEquals(clone_visual.create_user, user.username)
        self.assertEquals(clone_visual.update_user, user.username)

    def setup_user_role(self, user, dc_id, ds1_id, ds2_id):
        """
        Creates a role that has manage permissions for ds1 and view apps permission only for ds2
        The provided user gets assigned this role.
        """
        mixed_perms_role = make_mixed_perms_role(
            dc_id=dc_id, ds_manage_id=ds1_id, ds_view_id=ds2_id
        )
        mixed_perms_role.add_users([user.username])
        mixed_perms_role.save()

    def _move_to_public_workspace(self, dashboard_id):
        self.move_to_public_workspace(self.admin_client, dashboard_id)

    def move_to_public_workspace(self, client: Client, dashboard_id):
        payload = {
            'dashboards_info': json.dumps([{'report_id': dashboard_id}]),
            'current_workspace_id': 1,
        }
        response = client.post('/reports/move', payload)
        self.assertEquals(response.status_code, 200)
        r = response.json()[0]
        self.assertEquals(r['report_id'], dashboard_id)
        self.assertEquals(r['reports_failed'], [])

    def create_dataset(self, admin_client, dc_id, ds_name, dataset_type='singletable') -> int:
        url = reverse('datasets-dataset')
        ds1_payload = dataset_payload(ds_name=ds_name, dc_id=dc_id, dataset_type=dataset_type)
        response = admin_client.post(url, ds1_payload)
        self.assertEquals(response.status_code, 200)
        ds1_id = response.json()[0]['id']
        return ds1_id

    def create_dataconnection(self, admin_client, connection_name='abc'):
        dc_payload = {
            'dataconnection_type': 'arcengine',
            'dataconnection_info': '{"PARAMS":{"HOST":"10.0.0.211","PORT":"31050"}}',
            'dataconnection_name': connection_name,
        }
        return self.create_dataconnection_with_payload(admin_client, dc_payload)

    def create_dataconnection_with_payload(self, admin_client, dc_payload) -> int:
        url = reverse('datasets-dataconnection')
        response = admin_client.post(url, dc_payload)
        self.assertEquals(response.status_code, 200)
        dc_id = response.json()['id']
        return dc_id

    def edit_dataconnection_with_payload(self, admin_client, dc_payload, dataconnection_id) -> int:
        url = reverse('datasets-dataconnection-id', args=[dataconnection_id])
        data = urlencode(dc_payload)
        response = admin_client.put(url, data, content_type='application/x-www-form-urlencoded')
        self.assertEquals(response.status_code, 200)
        return response.json()
