# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
"""
First pass at a set of tests for the arcviz query generation / execution path

This requires live back-end data connections, as defined in the settings file

The standard unit test file doesn't have any back end definitions so all tests
get skipped.

I run with:

(arcviz_3.6.2) doug@Dougs-MacBook-Pro:~/arcdev/arcviz/arcweb$ python manage.py test --settings arcwebbase.settings_test_sqlrun --failfast vizlib.tests.test_sqlrun

Test areas:

   Results checking

   Cache testing:
     Arcengine recheck when table changes
     Cache invalidate at dataset and connection level
     Other dataset / connection updates do not invalidate cache

  Fallback checks
    Basic fallback on no AV
    AV no fallback
    Cache recheck valid on fallback path

  FA checks:
    Old and new arcengine
    Other engine
    Applied for user not admin
    Cache results include FA

  SQL Injection tests
    Table outside dataset access
    For Simple, LV, Dataset LV, Joined dataset

"""

import sys
import pdb
import time
import json
import os
import pprint
import urllib
import typing
import unittest
import logging

from django.conf import settings
from django.core.cache import cache
from django.http import HttpRequest, HttpResponse

import arcwebbase.settings_test_sqlrun as test_settings

RETENTION = getattr(settings, 'RETENTION', test_settings.RETENTION)
TEMPFILE = getattr(settings, 'TEMPFILE', test_settings.TEMPFILE)
DO_LOGGING = getattr(settings, 'DO_LOGGING', test_settings.DO_LOGGING)

from django.contrib.auth.models import User, Group
from datasets.models import DataSet, DataConnection, Segment, FilterAssociation
from dataconn.dataconn import get_dataconn
from dataconn.hs2conn import supports_extmeta
from datasources.dataimport import DataImport

from reports.models import Report, AppSetting
from reports.views_appgroups import appgroup_starred
from rbac.models import Role

import django.test
from django.test.client import Client
from vizlib.dataset import refreshDatasetColumns

# For debugging -- pprint with optional title
def pp(*args):
  fmt = pprint.pformat(args[-1], indent=2)
  if len(args) == 2:
    fmt = args[0] + '\n' + fmt
  print(fmt)

# --------------------------- Helper functions not tied to any class

def local_sqlite_db_path():
  return os.path.join(os.path.dirname(__file__),'test_sqlrun.sqlite')

LOCAL_TEST_DATACONNECTIONS = {
  'sqlite': {
    'dataconnection_type': 'sqlite',
    'dataconnection_info': {'PARAMS' : {
      'DATABASE': 'main',
      'FILE': local_sqlite_db_path(),
      'CACHE': {'ENABLED': 1, 'RETENTION': RETENTION},
    }}
  }
}

def get_dc_defs():
  dc_defs = getattr(settings, 'TEST_DATACONNECTIONS', {})
  return dc_defs | LOCAL_TEST_DATACONNECTIONS

def havedc(name):
  dc_defs = get_dc_defs()
  return dc_defs.get(name) is not None

DCS = {'init':True}
def get_djangodc(name):
  if DCS.get('init'):
    del DCS['init']
    for (dcname, params) in list(get_dc_defs().items()):
      params['dataconnection_name'] = dcname
      params['dataconnection_info'] = json.dumps(params['dataconnection_info'])
      DCS[dcname] = DataConnection.objects.create(**params)

  return DCS.get(name)

def get_dc(name) -> typing.Tuple[typing.Any, DataConnection]:
  django_dc = get_djangodc(name)
  if not django_dc:
    return None

  return (get_dataconn(django_dc.id, django_dc), django_dc)

def retention_pause(dc, factor=1):
  info = json.loads(dc.dataconnection_info)
  pause = info['PARAMS']['CACHE']['RETENTION']
  time.sleep(pause * factor)

# Run a list of dsreq structs
def run_dslist(client: Client, dslist: list[dict], wait_for_result=True, **kwargs):

  try:
    show = kwargs.pop('show')
  except KeyError:
    show = False

  data = []
  for dsreq in dslist:
    params = {'cache_read':'true', 'parent_report_id':10, 'async':'true'}
    params.update(kwargs)
    params['dsreq'] = json.dumps(dsreq)
    data.append({'params':params})
  rsp = client.post('/sqlrun/jsonselect_parallel', data={'data':json.dumps(data)})
  assert(rsp.status_code == 200)

  retlist = get_json_response(rsp)

  if not wait_for_result:
    if show:
      pp(retlist)
    return retlist

  while True:
    qhandles = [r.get('response', {}).get('qhandle') for r in retlist]
    running = [q for q in qhandles if q is not None]
    if not running:
      break
    pnum = running[0].split('.')[0]
    time.sleep(0.25)
    running_qhandles = ','.join(running)
    rsp = client.post('/sqlrun/%s/asyncstatus' % pnum, data={'qhandles': running_qhandles})
    assert(rsp.status_code == 200)
    sresult = get_json_response(rsp)
    assert(len(sresult) == len(running))
    for (handle, status) in zip(running, sresult):
      pos = qhandles.index(handle)
      res = retlist[pos]
      rsp = res['response']

      res.get('profile', []).extend(status.get('profile', []))

      if status.get('error'):
        res['error'] = status.get('error')
        del rsp['qhandle']

      if 'response' in status:
        rsp.update(status['response'])
        if status['response'].get('status') is None:
          del res['response']['qhandle']

  # All done, display if requested, then return list of response structures
  if show:
    pp(retlist)
  return retlist

def get_json_response(rsp):
  if hasattr(rsp, 'streaming_content'):
    return json.loads(b''.join(rsp.streaming_content).decode('utf-8'))
  else:
    return json.loads(rsp.content)

# Stream a single dsreq into a file
def stream_dsreq(client, dsreq, filename, **kwargs):

  params = kwargs.copy()
  params['csvdownload'] = 'true'
  params['csvfilename'] = filename
  params['async'] = 'false'
  params['language_code'] = 'en_US'
  params['dsreq'] = json.dumps(dsreq)

  rsp = client.post('/sqlrun/csvdownloadStore', data=params)
  assert(rsp.status_code == 200)

  rsp = client.get('/sqlrun/jsonselect?csvdownload=true', stream=True)
  assert(rsp.status_code == 200)
  save_response_to_file(rsp, filename)

def save_response_to_file(rsp: HttpRequest, filename: str):
  with open(filename, 'wb') as f:
    if hasattr(rsp, 'content'):
      f.write(rsp.content)
    else:
      for chunk in rsp.streaming_content:
        if chunk:
          f.write(chunk)

# Return error from response structure
def rsp_error(rsp):
  return rsp.get('error')

# Return True if response passed (i.e. no error and have a response struct)
# Probably should validate rows, cols, etc.
def rsp_passed(rsp):
  return rsp_error(rsp) is None and rsp.get('response') is not None

# Update fallback for a connection
def dc_fallback(dc, fdc):
  info = json.loads(dc.dataconnection_info)
  info['PARAMS']['FALLBACK'] = {} if fdc is None else {'CONNECTION_ID':fdc.id}
  dc.dataconnection_info = json.dumps(info)
  dc.save()

# Update dataconnection cache settings
def dc_cache(dc, cs):
  info = json.loads(dc.dataconnection_info)
  info['PARAMS']['CACHE'] = cs
  dc.dataconnection_info = json.dumps(info)
  dc.save()

# Update dataset cache settings
def ds_cache(ds, cs):
  settings = json.loads(ds.dataset_settings or '{}')
  settings['CACHE'] = cs
  ds.dataset_settings = json.dumps(settings)
  ds.save()

# Get type of cache access expected once retention expires
def get_dc_recheck(dc):
  return 'recheck' if dc.dataconnection_type == 'arcengine' or \
                      dc.dataconnection_type == 'arcengine_v2' else 'engine'

# Run a query against a data connection
def run_query(dc, query):
  conn = get_dataconn(dc.id)
  return conn.sync_runquery(query)

# Refresh analytical view (if given--allows avoiding a bunch of ifs in mainline test)
def refresh_aview(dc, avname):
  if avname:
    run_query(dc, 'refresh analytical view %s' % avname)

def drop_aview(dc, avname):
  if avname:
    run_query(dc, 'drop analytical view if exists %s' % avname)

# Create and refresh analytical view (if given)
def make_aview(dc, avname, avtext):
  if avname:
    drop_aview(dc, avname)
    run_query(dc, 'create analytical view %s stored as parquet as %s' % (avname, avtext))
    refresh_aview(dc, avname)

# Return True if two responses are the same.
# Note that fallback is a problem if the column name casing is not preserved.
def responses_equal(r1, r2, ordered=False):

  r1 = r1.get('response')
  r2 = r2.get('response')

  if r1.get('colnames') != r2.get('colnames') or \
     r1.get('coltypes') != r2.get('coltypes') or \
     len(r1.get('rows')) != len(r2.get('rows')):
    return False

  if ordered:
    return r1.get('rows') == r2.get('rows')

  r1rows = r1.get('rows')
  r2rows = r2.get('rows')[:]
  for r in r1rows:
    idx = r2rows.index(r)
    if idx == -1:
      return False
    del r2rows[idx]

  return True

# Return True if downloaded file matches the expected response
# Note that fallback is a problem if the column name casing is not preserved.
def download_equal(r1, filename, ordered=False):

  r1 = r1.get('response')

  r1colnames = ','.join(r1.get('colnames'))
  r1rows = [','.join([str(ri) for ri in r]) for r in r1.get('rows')]

  rows = open(filename).read().split('\n')
  r2colnames = rows[0]

  # Drop the column names and the weird empty line at the end
  r2rows = rows[1:]
  if r2rows[-1] == '':
    r2rows = r2rows[:-1]

  if r1colnames != r2colnames or \
     len(r1rows) != len(r2rows):
    return False

  if ordered:
    return r1rows == r2rows

  for r in r1rows:
    idx = r2rows.index(r)
    if idx == -1:
      return False
    del r2rows[idx]

  return True

def create_sqlite_db():
  import csv
  import sqlite3

  db_file = local_sqlite_db_path()
  with sqlite3.connect(db_file) as conn:

      cur = conn.cursor()
      cur.execute('drop table if exists arcadia_grouptest')
      cur.execute('create table arcadia_grouptest (a text, b text, p1 text, p2 int, value int, c int, d int)')

      source_csv = os.path.join(os.path.dirname(__file__),'test_sqlrun_grouptest.csv')
      with open(source_csv) as csv_file:
          csv_content = csv.reader(csv_file)
          next(csv_content) # skip the header line
          cur.executemany('insert into arcadia_grouptest values (?,?,?,?,?,?,?)', csv_content)

      conn.commit()

# Base class for functions used across data connections
class Common(django.test.TransactionTestCase):
  serialized_rollback = True

  @classmethod
  def make_user(cls, username, is_superuser=False):
    (u, created) = User.objects.get_or_create(username=username)
    u.set_password(username)
    u.is_superuser = is_superuser
    u.save()
    request = HttpRequest()
    request.user = u
    appgroup_starred(request)
    return u

  @classmethod
  def setUpClass(cls):
    create_sqlite_db()

    from reports.views_settings import initialize_settings
    initialize_settings(getattr(settings, 'SITE_SETTING_DEFAULTS', {}))

    (registered, created) = AppSetting.objects.get_or_create(setting_key='registered', setting_value='True')
    if created and DO_LOGGING:
      settings.ACCESS_LOG = True
      lg = logging.getLogger('arcweb')
      lg.propagate = False
      lg.setLevel(logging.DEBUG)
      lh = logging.StreamHandler()
      lh.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(process)d %(threadName)s %(message)s'))
      lg.addHandler(lh)

    o = type('', (), {})
    cls.objects = o

    o.ua = cls.make_user('admin', is_superuser=True)
    o.u1 = cls.make_user('user1')
    o.u2 = cls.make_user('user2')

    (o.g1, created) = Group.objects.get_or_create(name='group1')
    (o.g2, created) = Group.objects.get_or_create(name='group2')

    o.u1.groups.clear()
    o.u1.groups.add(o.g1)
    o.u2.groups.clear()
    o.u2.groups.add(o.g2)

    o.r1 = Role.objects.create(name='role1',
                               privileges=json.dumps([{'ptype':'system','perms':['sys_editperm']}]))
    o.r2 = Role.objects.create(name='role2',
                               privileges=json.dumps([{'ptype':'system','perms':['sys_editconn']}]))

  #   o.v1 = Report.objects.create(report_name='v1', dataset=o.ds1, report_type='table', report_data='[]')

    o.ca = Client()
    o.ca.login(username=o.ua.username, password=o.ua.username)
    o.cu1 = Client()
    o.cu1.login(username=o.u1.username, password=o.u1.username)
    o.cu2 = Client()
    o.cu2.login(username=o.u2.username, password=o.u2.username)

  @classmethod
  def tearDownClass(cls):
    os.remove(local_sqlite_db_path())

    o = cls.objects
    o.ca.logout()
    o.cu1.logout()
    o.cu2.logout()
    del o.ca
    del o.cu1
    del o.cu2
    d = o.__dict__
    for name in d.keys():
      if not name.startswith('__'):
        d[name].delete()

    # recreate the test DB because with TransactionTestCase all tables get truncated after each test case:
    from django.db import connection
    connection._connections['default'].creation.create_test_db(autoclobber=True)

  # Create a dataset for connection on the given table
  def make_dataset(self, django_dc, tablename, select=None):
    table = {"tablename": tablename, "columns": []}
    if select:
      table['select'] = select
    ds = DataSet.objects.create(dataset_name=tablename, dataconnection=django_dc,
          dataset_detail='default.grouptest', dataset_type='singletable',
          dataset_info=json.dumps([table]))

    self.assertTrue(refreshDatasetColumns(ds, 'admin'))
    return ds

  # Create a segment for the given dataset (seems easier just to issue the request...)
  def make_segment(self, client, ds, segname, segexpr):
    filters = [segexpr]
    if isinstance(segexpr, list):
      # the argument is a list, use it directly
      filters = segexpr

    segdata = {'group':'', 'filters': filters, 'entities': []}
    rsp = client.post('/datasets/segment', data={'dataset_id':ds.id, 'name':segname, 'data':json.dumps(segdata)})
    self.assertTrue(rsp.status_code == 200)
    return json.loads(rsp.content.decode())[0]

  # Create a filter association
  def make_fa(self, client, ds, faname, segid, users=None, groups=None):
    data = {'groups': json.dumps(groups or []),
            'users': json.dumps(users or []),
            'name': faname,
            'dataset_id': ds.id,
            'data': json.dumps({'applied_segments': [{'id':segid}]})}

    rsp = client.post('/datasets/filterassociation', data=data);
    self.assertTrue(rsp.status_code == 200)
    return json.loads(rsp.content.decode())[0]

  # Update users/groups for a filter association
  def update_fa(self, client, fa, users=None, groups=None):
    data = {'groups': json.dumps(fa.get('groups') if groups is None else groups),
            'users': json.dumps(fa.get('users') if users is None else users),
            'name': fa.get('name'),
            'data': fa.get('data')}

    urlparms = urllib.parse.urlencode(data)
    rsp = client.put('/datasets/filterassociation/%s' % fa.get('id'), data=urlparms)
    self.assertTrue(rsp.status_code == 200)
    return json.loads(rsp.content.decode())[0]

  # dsreq based on grouptest dataset
  def gt_dsreq1(self, ds):
    dsreq = {
       'type': 'SQL',
       'version': 1,
       'dimensions': [{'type':'SIMPLE', 'expr':"[A] as 'A'"}, {'type':'SIMPLE', 'expr':"[B] as 'B'"}],
       'aggregates': [{'expr':"sum([Value]) as 'sum(Value)'"}],
       'limit': 100}
    dsreq['dataset_id'] = ds.id
    return dsreq

  def gt_dsreq2(self, ds):
    dsreq = {
       'type': 'SQL',
       'version': 1,
       'dimensions': [{'type':'SIMPLE', 'expr':"[A] as 'A'"}, {'type':'SIMPLE', 'expr':"[B] as 'B'"}],
       'aggregates': [{'expr':"max([Value]) as 'max(Value)'"}],
       'limit': 100}
    dsreq['dataset_id'] = ds.id
    return dsreq

  # Get a dsreq that will fail
  def gt_dsreq_err1(self, ds):
    dsreq = {
       'type': 'SQL',
       'version': 1,
       'dimensions': [{'type':'SIMPLE', 'expr':"[A] + fail as 'A'"}, {'type':'SIMPLE', 'expr':"[B] as 'B'"}],
       'aggregates': [{'expr':"max([Value]) as 'max(Value)'"}],
       'limit': 100}
    dsreq['dataset_id'] = ds.id
    return dsreq

  def gt_dsreq_err2(self, ds):
    dsreq = {
       'type': 'SQL',
       'version': 1,
       'dimensions': [{'type':'SIMPLE', 'expr':"[A] as 'A'"}, {'type':'SIMPLE', 'expr':"[B] error as 'B'"}],
       'aggregates': [{'expr':"max([Value]) as 'max(Value)'"}],
       'limit': 100}
    dsreq['dataset_id'] = ds.id
    return dsreq

  # Get a dsreq with a derived data component.  Note that this dsreq is chosen specifically as
  # we will get a different result when an FA based on [A] is applied to the DD query (with the
  # current grouptest table data of course).
  def gt_dsreq_dd(self, ds):
    ddreq = {
       'type': 'SQL',
       'version': 1,
       'dimensions': [{'type':'SIMPLE', 'expr':"[A] as 'A'"},],
       'aggregates': [{'expr':"min([Value]) as 'MINV'"}],
       'limit': 100}
    dsreq = {
       'type': 'SQL',
       'version': 1,
       'dimensions': [{'type':'SIMPLE', 'expr':"[A] as 'A'"}],
       'aggregates': [{'expr':"min([MINV]) as 'OUT_MINV'"}],
       'derived_data' : [{'label': 'DD_1', 'dsreq' : ddreq}],
       'limit': 100}
    dsreq['dataset_id'] = ds.id
    return dsreq

  # AV statement for dsreq based on grouptest dataset
  # Could actually issue the dsreq and pull the query from the response
  def gt_dsreq1_av(self, ds):
    tname = json.loads(ds.dataset_info)[0].get('tablename')
    return 'select a,b,sum(value) from %s group by 1,2' % tname

  # Run dsreq (or list of dsreqs) and ensure expected execution as indindicated by profile
  def dsreq_check(self, dsreqs, *checks, **kwargs):
    if 'client' in kwargs:
      client = kwargs.pop('client')
    else:
      client = self.__class__.objects.ca
    dflargs = {'cache_read':'true', 'parent_report_id':10, 'async':'true'}
    dflargs.update(kwargs)
    reqlist = dsreqs if type(dsreqs) == type([]) else [dsreqs]
    rsplist = run_dslist(client, reqlist, **dflargs)
    for dsrsp in rsplist:

      if 'error' not in checks:
        self.assertTrue(rsp_passed(dsrsp))

      profkeys = [p[0] for p in dsrsp.get('profile', [])]
      for check in checks:
        if check == 'cache':
          self.assertTrue('cache.fetch' in profkeys)
        elif check == 'nocache':
          self.assertTrue('cache.fetch' not in profkeys)
        elif check == 'recheck':
          self.assertTrue('cache.tsupdate.1' in profkeys or 'cache.tsupdate.2' in profkeys)
        elif check == 'engine':
          self.assertTrue('cache.post' in profkeys or 'async_results.cachepost' in profkeys)
        elif check == 'fallback':
          self.assertTrue(any(p.startswith('fallback to connection') for p in profkeys))
        elif check == 'nofallback':
          self.assertFalse(any(p.startswith('fallback to connection') for p in profkeys))
        elif check == 'error':
          self.assertTrue(rsp_error(dsrsp))
        else:
          self.assertTrue(False, 'Unknown dsrsp check: %s' % check)
    return rsplist if len(rsplist) > 1 else rsplist[0]

  # Basic cache test for a connection (via input dataset)
  # Here we are assuming the grouptest dataset but should construct this from actual dataset columns
  # First access not cached
  # Next access cached
  # Wait for cache recheck time
  # Next access rechecked or from engine
  # Next access from cache
  def cache_basic(self, ds):
    client = self.__class__.objects.ca
    dsreq = self.gt_dsreq1(ds)

    cache.clear()
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    retention_pause(ds.dataconnection)
    self.dsreq_check(dsreq, get_dc_recheck(ds.dataconnection))
    self.dsreq_check(dsreq, 'cache')

  # Check connect and dataset cache settings
  def cache_config(self, ds):
    dsreq = self.gt_dsreq1(ds)

    # Init
    cache.clear()
    dc = ds.dataconnection
    recheck = get_dc_recheck(dc)

    dc_cache(dc, {'ENABLED':1, 'RETENTION':RETENTION})
    ds_cache(ds, {'OVERRIDE':0})
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # data connection cache reset
    client = self.__class__.objects.ca
    rsp = client.post('/datasets/dataconnection/cache_reset/%s' % ds.dataconnection_id)
    assert(rsp.status_code == 200)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # Connection settings - enable and retention
    dc_cache(dc, {'ENABLED':0})
    self.dsreq_check(dsreq, 'engine')
    dc_cache(dc, {'ENABLED':1, 'RETENTION':RETENTION * 2})
    self.dsreq_check(dsreq, 'cache')
    retention_pause(dc, 0.5)
    self.dsreq_check(dsreq, 'cache')
    retention_pause(dc)
    self.dsreq_check(dsreq, recheck)
    self.dsreq_check(dsreq, 'cache')

    # Dataset cache reset
    rsp = client.post('/datasets/dataset/cache_reset/%s' % ds.id)
    assert(rsp.status_code == 200)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # Dataset settings -- override + enable and retention
    ds_cache(ds, {'OVERRIDE':1, 'ENABLED':0})
    self.dsreq_check(dsreq, 'engine')
    ds_cache(ds, {'OVERRIDE':1, 'ENABLED':1, 'RETENTION':RETENTION * 2})
    self.dsreq_check(dsreq, 'cache')
    retention_pause(dc, 0.5)
    self.dsreq_check(dsreq, 'cache')
    retention_pause(dc)
    self.dsreq_check(dsreq, recheck)
    self.dsreq_check(dsreq, 'cache')

    # Put settings back
    dc_cache(dc, {'ENABLED':1, 'RETENTION':RETENTION})
    ds_cache(ds, {'OVERRIDE':0})

  # Ensure cache recheck detects table changes
  # Should do:
  #   Table definition change
  #   (Not yet) Data table file updated (could do by updating via hive)
  #   Data added to table
  #   Logical view updated
  #   Data table under logical view data added
  #   Data table under logical view updated
  def cache_tablecheck(self, dc, src_tab, tgt_tab, tgt_lview, tgt_aview=None):

    # Set up a target table we can mess with
    run_query(dc, 'drop table if exists %s' % tgt_tab)
    run_query(dc, 'create table %s as select * from %s' % (tgt_tab, src_tab))
    ds = self.make_dataset(dc, tgt_tab)

    avtext = self.gt_dsreq1_av(ds)
    make_aview(dc, tgt_aview, avtext)
    cache.clear()

    dsreq = self.gt_dsreq1(ds)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # Table data change invalidates cache
    run_query(dc, 'insert overwrite %s select * from %s' % (tgt_tab, src_tab))
    refresh_aview(dc, tgt_aview)
    retention_pause(dc)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # Table definition change invalidates cache
    run_query(dc, 'alter table %s change column a a int' % tgt_tab)
    run_query(dc, 'alter table %s change column a a string' % tgt_tab)
    make_aview(dc, tgt_aview, avtext)
    retention_pause(dc)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # Set up a target view we can mess with
    run_query(dc, 'drop view if exists %s' % tgt_lview)
    run_query(dc, 'create view %s as select * from %s' % (tgt_lview, tgt_tab))
    ds = self.make_dataset(dc, tgt_lview)
    dsreq = self.gt_dsreq1(ds)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # Base table data change invalidates cache from logical view
    run_query(dc, 'insert overwrite %s select * from %s' % (tgt_tab, src_tab))
    make_aview(dc, tgt_aview, avtext)
    retention_pause(dc)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # Base table definition change invalidates cache
    run_query(dc, 'alter table %s change column a a int' % tgt_tab)
    run_query(dc, 'alter table %s change column a a string' % tgt_tab)
    make_aview(dc, tgt_aview, avtext)
    retention_pause(dc)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # Change view invalidates cache
    run_query(dc, 'alter view %s as select * from %s' % (tgt_lview, tgt_tab))
    retention_pause(dc)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    # Clean up temp tables
    run_query(dc, 'drop table if exists %s' % tgt_tab)
    run_query(dc, 'drop view if exists %s' % tgt_lview)
    drop_aview(dc, tgt_aview)

  # Test request parameters
  # Just 'cache_read' right now
  def params_basic(self, ds):
    client = self.__class__.objects.ca
    dsreq = self.gt_dsreq1(ds)
    cache.clear()
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')
    self.dsreq_check(dsreq, 'nocache', cache_read='false')

  def fallback_cache(self, dc, fdc, src_tab, tgt_tab, tgt_lview, tgt_aview):
    # Set up a target table we can mess with
    run_query(dc, 'drop table if exists %s' % tgt_tab)
    run_query(dc, 'create table %s as select * from %s' % (tgt_tab, src_tab))
    drop_aview(dc, tgt_aview)
    ds = self.make_dataset(dc, tgt_tab)
    dsreq = self.gt_dsreq1(ds)
    self.dsreq_check(dsreq, 'engine')
    self.dsreq_check(dsreq, 'cache')

    dc_fallback(dc, fdc)
    cache.clear()
    self.dsreq_check(dsreq, 'nocache', 'fallback')

    make_aview(dc, tgt_aview, self.gt_dsreq1_av(ds))

    cache.clear()
    self.dsreq_check(dsreq, 'nocache', 'nofallback')
    self.dsreq_check(dsreq, 'cache')
    retention_pause(dc)
    self.dsreq_check(dsreq, 'recheck')

    # Cache table checks should all pass with fallback, both with and without an AV
    self.cache_tablecheck(dc, src_tab, tgt_tab, tgt_lview)
    self.cache_tablecheck(dc, src_tab, tgt_tab, tgt_lview, tgt_aview)

    # Disable fallback
    dc_fallback(dc, None)

  # Issue a known bad request and ensure error response
  def fail_check(self, ds):
    dsreq = self.gt_dsreq_err1(ds)
    self.dsreq_check(dsreq, 'error')

  # Run some csv download tests
  def download_check(self, ds, client=None):
    if client is None:
      client = self.__class__.objects.ca
    cache.clear()
    dsreq = self.gt_dsreq1(ds)
    filename = TEMPFILE
    cache.clear()
    stream_dsreq(client, dsreq, filename)

    rsp1 = self.dsreq_check(dsreq, 'cache')

    self.assertTrue(download_equal(rsp1, filename))

    rsp2 = self.dsreq_check(dsreq, 'cache')
    stream_dsreq(client, dsreq, filename)
    self.assertTrue(download_equal(rsp1, filename))
    self.assertTrue(responses_equal(rsp1, rsp2))

  # Basic FA test
  def fa_basic(self, ds, cache, viz_async):
    o = self.__class__.objects
    s1 = self.make_segment(o.ca, ds, 's1', '[A] in ("doug", "veronica")')
    s2 = self.make_segment(o.ca, ds, 's2', '[A] in ("alan", "sean", "amanda")')
    fa1 = self.make_fa(o.ca, ds, 'fa_s1', s1.get('id'), users=[o.u1.id])
    fa2 = self.make_fa(o.ca, ds, 'fa_s2', s2.get('id'), users=[o.u2.id])

    if cache:
      checks = ['cache']
      kwargs = {'cache_read':'true'}
    else:
      checks = ['nocache']
      kwargs = {'cache_read':'false'}

    kwargs['async'] = 'true' if viz_async else 'false'

    filename = TEMPFILE

    settings.FAIL_SAFE = True
    settings.FAIL_SAFE_CONFIG = { ds.id : ['default.grouptest.a'] }

    # Ensure these regular users can access the dataset
    o.r1.add_users([o.u1.username, o.u2.username])
    o.r1.add_priv('dataset', ['ds_appview'], dcid=ds.dataconnection_id, dslist=[ds.id])

    # Issue request as master with s1 as global segment and request as u1 -- same result
    dsreq = self.gt_dsreq1(ds)
    u1rsp = self.dsreq_check(dsreq, *checks, client=o.cu1, **kwargs)

    uareq = self.gt_dsreq1(ds)
    uareq['global_segments'] = [{'type':'SEGMENT', 'segments':[{'seg_id':s1.get('id'), 'negate':False}]}]
    uarsp = self.dsreq_check(uareq, *checks, client=o.ca, **kwargs)
    self.assertTrue(responses_equal(u1rsp, uarsp))

    # Also ensure download csv respects the FA
    stream_dsreq(o.cu1, dsreq, filename)
    self.assertTrue(download_equal(u1rsp, filename))

    # Same check for different user
    u2rsp = self.dsreq_check(dsreq, *checks, client=o.cu2, **kwargs)

    uareq = self.gt_dsreq1(ds)
    uareq['global_segments'] = [{'type':'SEGMENT', 'segments':[{'seg_id':s2.get('id'), 'negate':False}]}]
    uarsp = self.dsreq_check(uareq, *checks, client=o.ca, **kwargs)
    self.assertTrue(responses_equal(u2rsp, uarsp))

    stream_dsreq(o.cu2, dsreq, filename)
    self.assertTrue(download_equal(u2rsp, filename))

    # Sanity check -- ensure different users got different results
    self.assertFalse(responses_equal(u1rsp, u2rsp))

    # Ensure when user1 has both user and group FA, that the user FA take precedence
    fa2 = self.update_fa(o.ca, fa2, users=[o.u2.id], groups=[o.g1.id])
    u1rsp_g1 = self.dsreq_check(dsreq, *checks, client=o.cu1, **kwargs)
    self.assertTrue(responses_equal(u1rsp, u1rsp_g1))

    # Now dropping user specification from fa1 should make fa2 apply to user1
    fa1 = self.update_fa(o.ca, fa1, users=[])
    u1rsp_g2 = self.dsreq_check(dsreq, *checks, client=o.cu1, **kwargs)
    self.assertTrue(responses_equal(u2rsp, u1rsp_g2))

    # Ensure FA is applied to derived data query as well as main query
    gs = [{'type':'SEGMENT', 'segments':[{'seg_id':s2.get('id'), 'negate':False}]}]
    dd_uareq = self.gt_dsreq_dd(ds)
    dd_uareq['global_segments'] = gs
    dd_uareq['derived_data'][0]['dsreq']['global_segments'] = gs
    dd_uarsp = self.dsreq_check(dd_uareq, *checks, client=o.ca, **kwargs)

    dd_u1req = self.gt_dsreq_dd(ds)
    dd_u1rsp = self.dsreq_check(dd_u1req, *checks, client=o.cu1, **kwargs)
    self.assertTrue(responses_equal(dd_u1rsp, dd_uarsp))

    # Finally, dropping the group from fa2 should trip the error catcher for user1
    fa2 = self.update_fa(o.ca, fa2, users=[o.u2.id], groups=[])
    self.dsreq_check(dsreq, 'error', client=o.cu1)

    # Ensure a query with a filter applied but not via FA also trips the error catcher
    dsreq['global_segments'] = [{'type':'SEGMENT', 'segments':[{'seg_id':s1.get('id'), 'negate':False}]}]
    self.dsreq_check(dsreq, 'error', client=o.cu1)

    # Restore setting
    settings.FAIL_SAFE = False

  # Run the FA tests with all combinations of caching and async
  def fa_tests(self, dc, ds=None):
    # Needs longer retention time so we don't get cache rechecks when not expected
    dc_cache(dc, {'ENABLED':1, 'RETENTION':20})
    if not ds:
      ds = self.make_dataset(dc, 'default.grouptest')
    cache.clear()
    self.fa_basic(ds, cache=False, viz_async=True)
    self.fa_basic(ds, cache=True, viz_async=True)
    cache.clear()
    self.fa_basic(ds, cache=False, viz_async=False)
    self.fa_basic(ds, cache=True, viz_async=False)

  # FA test with fallback
  def fa_tests_fb(self, dc, fdc):
    dc_fallback(dc, fdc)
    ds = self.make_dataset(dc, 'default.grouptest')
    dsreq = self.gt_dsreq1(ds)

    # Ensure FA tests pass with fallback enabled and fallback
    self.dsreq_check(dsreq, 'nocache', 'fallback')
    self.fa_tests(dc, ds)

    # Now FA tests with fallback enabled but AV present
    make_aview(dc, 'default.gt_unittest_lv', self.gt_dsreq1_av(ds))
    cache.clear()
    self.dsreq_check(dsreq, 'nocache', 'nofallback')
    self.fa_tests(dc, ds)
    drop_aview(dc, 'default.gt_unittest_lv')
    dc_fallback(dc, None)

  # Test dataset from SQL
  def ds_from_sql(self, django_dc, tablename='default.grouptest'):
    ds = self.make_dataset(django_dc, tablename)
    dsreq = self.gt_dsreq1(ds)
    ds_sql = self.make_dataset(django_dc, 'grouptest_from_SQL',
                               select='select * from %s where a in ("doug", "veronica")' % tablename)
    dsreq_sql = self.gt_dsreq1(ds_sql)

    cache.clear()
    res = self.dsreq_check(dsreq, 'nocache')
    res_sql = self.dsreq_check(dsreq_sql, 'nocache')
    self.assertTrue(len(res['response']['rows']) > len(res_sql['response']['rows']))
    res['response']['rows'] = [r for r in res['response']['rows'] if r[0] in ['doug', 'veronica']]
    self.assertTrue(responses_equal(res, res_sql))

@unittest.skipIf(not havedc('sqlite'), 'Sqlite backend not configured')
class TestSqlite(Common):

  def setUp(self):
    super().setUp()
    self.test_csv_input = 'test_input.csv'

  def tearDown(self):
    super().tearDown()
    if os.path.isfile(self.test_csv_input):
      os.remove(self.test_csv_input)

  def test_1(self):
    (dc, django_dc) = get_dc('sqlite')

    ds = self.make_dataset(django_dc, 'main.arcadia_grouptest')
    self.fail_check(ds)
    self.cache_basic(ds)
    self.download_check(ds)
    self.ds_from_sql(django_dc, 'main.arcadia_grouptest')
    self._test_csv_import()

  def test_vector_distance(self):
    (dc, _) = get_dc('sqlite')
    connection = dc.get_connection()
    cursor = connection.cursor()
    cursor.execute('select vector_distance("[2,3]","[-1,-1]")')
    rows = cursor.fetchall()
    distance = float(rows[0][0])
    self.assertEquals(distance, 5)

  def _test_csv_import(self):
    (dc, django_dc) = get_dc('sqlite')

    # create csv input
    with open(self.test_csv_input, mode='wt') as csv_file:
      csv_file.writelines(['col1, col2, col3\n'])
      csv_file.writelines(['2, "a", 1.2\n'])

    # import the CSV file:
    di=DataImport(uploadedfile=self.test_csv_input,
                  dataconnection_id=django_dc.id,
                  data={
                    'datasource_tablename': 'test_table'
                    }
                  )
    di.reset_file()
    di.create_table({})
    di.insert_data(batch_count=settings.IMPORT_BATCH_INSERT)

    # quuery the database:
    connection = dc.get_connection()
    cursor = connection.cursor()
    cursor.execute('select * from test_table')
    rows = cursor.fetchall()
    self.assertEquals(len(rows), 1)
    self.assertEquals(rows[0], (2,'a',1.2))

    column_info = dc.table_columns('test_table')
    self.assertEquals(column_info['colnames'], ['col1','col2','col3'])
    self.assertEquals(column_info['coltypes'], ['INT', 'STRING', 'DOUBLE'])


@unittest.skipIf(not havedc('mysql'), 'MySQL backend not configured')
class TestMySQL(Common):
  def test_1(self):
    (dc, django_dc) = get_dc('mysql')

    ds = self.make_dataset(django_dc, 'arcadia.grouptest')
    self.fail_check(ds)
    self.cache_basic(ds)
    self.download_check(ds)
    self.ds_from_sql(django_dc, 'arcadia.grouptest')

@unittest.skipIf(not havedc('hive'), 'Hive backend not configured')
class TestHive(Common):
  def test_1(self):
    (dc, django_dc) = get_dc('hive')
    self.assertEquals(dc.srv_name(), 'Apache Hive')
    self.assertFalse(supports_extmeta(dc.srv_name(), dc.srv_version()))

    ds = self.make_dataset(django_dc, 'default.grouptest')
    self.fail_check(ds)
    self.cache_basic(ds)
    self.download_check(ds)
    self.ds_from_sql(django_dc)

@unittest.skipIf(not havedc('arcengine'), 'Arcengine backend not configured')
class TestArcengine(Common):

  def test_1(self):
    (dc, django_dc) = get_dc('arcengine')
    self.assertEquals(dc.srv_name(), 'Arcengine')
    self.assertTrue(supports_extmeta(dc.srv_name(), dc.srv_version()))

    ds = self.make_dataset(django_dc, 'default.grouptest')
    self.fail_check(ds)
    self.cache_basic(ds)
    self.download_check(ds)

    # For first connection, ensure 'cache_read' flag is respected
    self.params_basic(ds)
    self.cache_config(ds)

  def test_2(self):
    (dc, django_dc) = get_dc('arcengine')
    self.cache_tablecheck(django_dc, 'default.grouptest', 'default.gt_unittest', 'default.gt_unittest_lv')

  # Test case for invalid query handle -- with short timeout issue two requests immediately following invalida metadata
  def test_3(self):
    (conn, dc) = get_dc('arcengine')
    client = self.__class__.objects.ca
    save_info = json.loads(dc.dataconnection_info)
    info = json.loads(dc.dataconnection_info)
    info['PARAMS'].update({
      'CACHE':{'ENABLED': 0},
      'IMPERSONATION': True,
      'TIMEOUT': 3,
      'SESSION_TIMEOUT': 3,
    })
    dc.dataconnection_info = json.dumps(info)
    dc.save()

    ds = self.make_dataset(dc, 'default.grouptest')
    dsreqs = [self.gt_dsreq1(ds), self.gt_dsreq2(ds)]

    run_query(dc, 'invalidate metadata')
    dsrsps = run_dslist(client, dsreqs)

    dc.dataconnection_info = json.dumps(save_info)
    dc.save()

    self.assertTrue(rsp_passed(dsrsps[0]))
    self.assertTrue(rsp_passed(dsrsps[1]))

  # FA tests to arcengine
  def test_4(self):
    (conn, dc) = get_dc('arcengine')
    self.fa_tests(dc)

  def test_5(self):
    (conn, dc) = get_dc('arcengine')
    self.ds_from_sql(dc)
    settings.ARCENGINE_VALIDATE_ACCESS = True
    self.ds_from_sql(dc)

  # Some validation when requests get queued due to too many outstanding requests
  def test_59(self):
    (conn, dc) = get_dc('arcengine')
    settings.HS2_QUEUE_SETTINGS = {'max_connections':1,
                                   'max_requests':1,
                                   'max_reqperconn':1}

    # Update connection to force re-load with above settings overrides
    dc_cache(dc, {'ENABLED':1, 'RETENTION':RETENTION})

    ds = self.make_dataset(dc, 'default.grouptest')
    reqlist = [self.gt_dsreq1(ds), self.gt_dsreq2(ds)]
    cache.clear()
    self.dsreq_check(reqlist, 'engine')
    rsplist1 = self.dsreq_check(reqlist, 'cache')
    retention_pause(dc)

    # Currently, cache recheck on the query queued path does not work
    # rsplist2 = self.dsreq_check(reqlist, 'recheck')

    # But we can at least make sure the results match
    rsplist2 = self.dsreq_check(reqlist)
    for r1, r2 in zip(rsplist1, rsplist2):
      self.assertTrue(responses_equal(r1, r2))

    # Ensure no issues on sync mode with queueing
    rsplist2 = self.dsreq_check(reqlist, viz_async='false')
    for r1, r2 in zip(rsplist1, rsplist2):
      self.assertTrue(responses_equal(r1, r2))

    # Make sure that errors returned on queued requests
    reqlist = [self.gt_dsreq_err1(ds), self.gt_dsreq_err2(ds)]
    self.dsreq_check(reqlist, 'error')

    # Put the settings back
    del settings.HS2_QUEUE_SETTINGS
    dc_cache(dc, {'ENABLED':1, 'RETENTION':RETENTION})

@unittest.skipIf(not havedc('arcengine') or not havedc('impala_arcengine'),
                 'Arcengine or fallback backend not configured')
class TestArcengineFallback(Common):

  # Basic fallback test
  def test_1(self):
    (dc, django_dc) = get_dc('arcengine')
    (fdc, django_fdc) = get_dc('impala_arcengine')
    self.fallback_cache(django_dc, django_fdc, 'default.grouptest', 'default.gt_unittest', 'default.gt_unittest_lv', 'default.gt_unittest_av')

  # Filter association tests with fallback
  def test_2(self):
    (conn, dc) = get_dc('arcengine')
    (fconn, fdc) = get_dc('impala_arcengine')
    self.fa_tests_fb(dc, fdc)

@unittest.skipIf(not havedc('old_arcengine'), 'Old arcengine backend not configured')
class TestOldArcengine(Common):

  def test_1(self):
    (dc, django_dc) = get_dc('old_arcengine')
    self.assertEquals(dc.srv_name(), 'Arcengine')
    self.assertFalse(supports_extmeta(dc.srv_name(), dc.srv_version()))

    ds = self.make_dataset(django_dc, 'default.grouptest')
    self.fail_check(ds)
    self.cache_basic(ds)
    self.download_check(ds)

  # FA test to old arcengine
  def test_2(self, ds=None):
    (conn, dc) = get_dc('old_arcengine')
    # Needs longer retention time due to length of test
    self.fa_tests(dc)

@unittest.skipIf(not havedc('old_arcengine') or not havedc('impala_old_arcengine'),
                 'Old arcengine or fallback backend not configured')
class TestOldArcengineFallback(Common):

  def test_1(self):
    (dc, django_dc) = get_dc('old_arcengine')
    (fdc, django_fdc) = get_dc('impala_old_arcengine')
    self.fallback_cache(django_dc, django_fdc, 'default.grouptest', 'default.gt_unittest', 'default.gt_unittest_lv', 'default.gt_unittest_av')

  # FA test to old arcengine with fallback
  def test_2(self):
    (conn, dc) = get_dc('old_arcengine')
    (fconn, fdc) = get_dc('impala_old_arcengine')
    # Needs longer retention time due to length of test
    self.fa_tests_fb(dc, fdc)

if __name__ == '__main__':
  unittest.main()
