#!/usr/bin/python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import unittest
import time
import random
import json

import vizlib.util
import cli.core.threadpool

class TestIdAllocator(unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_simple(self):
    t1 = int(time.time()) + 1
    idAllocator = vizlib.util.IdAllocator()
    id1 = idAllocator.allocateId()
    t2 = t1 + 1

    self.assertTrue(int(id1) >= int(t1))
    self.assertTrue(int(id1) <= int(t2))

    id2 = idAllocator.allocateId()
    self.assertEqual(int(id1) + 1, int(id2))

    id3 = idAllocator.allocateId()
    self.assertEqual(int(id1) + 2, int(id3))

  def test_parallel(self):
    idAllocator = vizlib.util.IdAllocator()

    workq = cli.core.threadpool.WorkQ()
    for i in range(0, 50):
      workq.add_work(cli.core.threadpool.Runner(idAllocator.allocateId))

    threadpool = cli.core.threadpool.ThreadPool(maxlen=10)

    id1 = idAllocator.allocateId()
    threadpool.run_workq(workq)
    id2 = idAllocator.allocateId()

    self.assertEqual(int(id2) - int(id1), 51)

class TestIdKeepLargestItemsList(unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_simple(self):
    kl = vizlib.util.KeepLargestItemsList(maxCount=5)

    for i in range(0, 50):
      kl.append(i)

    self.assertEqual(kl.getCopy(), [45, 46, 47, 48, 49])

    for i in range(0, 20):
      kl.append(i)
    self.assertEqual(kl.getCopy(), [45, 46, 47, 48, 49])

    kl.append(47.5)
    self.assertEqual(kl.getCopy(), [46, 47, 47.5, 48, 49])

    kl.append(61)
    self.assertEqual(kl.getCopy(), [47, 47.5, 48, 49, 61])

  def test_parallel(self):
    kl = vizlib.util.KeepLargestItemsList(maxCount=5)
    workq = cli.core.threadpool.WorkQ()
    nums = list(range(0, 500))
    random.shuffle(nums)

    workq = cli.core.threadpool.WorkQ()
    for i in nums:
      workq.add_work(cli.core.threadpool.Runner(kl.append, (i, )))

    threadpool = cli.core.threadpool.ThreadPool(maxlen=20)

    threadpool.run_workq(workq)

    self.assertEqual(kl.getCopy(), [495, 496, 497, 498, 499])

class TestSingleRequestTracker(unittest.TestCase):
  def test_simple(self):
    testcase = self

    class Tracker:
      def __init__(self, requestData, profile, startTime):
        self.requestData = requestData
        self.profile = profile
        self.startTime = startTime
        self.t1 = time.time()
        self.startCalled = 0
        self.stopCalled = 0

      def startRequest(self, requestData, profile, startTime):
        testcase.assertEqual(self.requestData, requestData)
        testcase.assertEqual(self.profile, profile)
        testcase.assertEqual(self.startTime, startTime)
        testcase.assertTrue(startTime <= self.t1)
        self.startCalled += 1
 
      def stopRequest(self, requestData, profile, startTime, runtime):
        testcase.assertEqual(self.requestData, requestData)
        testcase.assertEqual(self.profile, profile)
        testcase.assertEqual(self.startTime, startTime)
        testcase.assertTrue(runtime > 0)
        testcase.assertTrue(time.time() - self.startTime >= runtime)
        self.stopCalled += 1

    requestData = 1
    profile = 2
    startTime = time.time()
    tracker = Tracker(requestData, profile, startTime)

    with vizlib.util._SingleRequestTracker(tracker, requestData, profile, startTime):
      pass

    testcase.assertEqual(tracker.startCalled, 1)
    testcase.assertEqual(tracker.stopCalled, 1)

class MockProfile:
  def __init__(self, id):
    self._id = id

  def username(self):
    return f"user_{self._id}"

  def querytracker(self):
    return f"querytracker_{self._id}"

class MockRunner:
  def __init__(self, id):
    self._id = id

  def admissionControlCategory(self):
    return f"admission_{self._id}"

class MockRequestData:
  def __init__(self, id):
    self._id = id
    self._runner = MockRunner(id)

  def runner(self):
    return self._runner

  def __str__(self):
    return f"str_{self._id}"

  def __repr__(self):
    return f"repr_{self._id}"

class TestRequestStatsTracker(unittest.TestCase):
  def test_simple(self):

    # So we can see any errors during asserts in full
    self.maxDiff = None

    stats = vizlib.util.RequestStatsTracker('testtracker', capacity=100, perUserCapacity=50)

    data = {}
    stats.updateRequestStats(data, user=None)
    self.assertEqual(data, {
      "testtracker": {
        "active": 0,
        "capacity": 100,
        "perUserCapacity": 50,
        "pendingCount": 0,
        "total": 0,
        "removalErrors": 0,
        "longest": [],
        "curr": [],
        "pending": []
      }
    })

    profile2 = MockProfile(2)
    stats.startRequest(1, profile2, 3)
    data = {}
    stats.updateRequestStats(data, user=None)
    data["testtracker"]["curr"][0]["runtimeOngoing"] = "r0"
    self.assertEqual(data, {
      "testtracker": {
        "active": 1,
        "capacity": 100,
        "perUserCapacity": 50,
        "pendingCount": 0,
        "total": 1,
        "removalErrors": 0,
        "longest": [],
        "curr": [
          {
            "runtimeOngoing": "r0",
            "startTime": "1970-01-01T00:00:03",
            "request": "1",
            "username": "user_2",
            "querytracker": "querytracker_2"
          }
        ],
        "pending": []
      }
    })

    profile22 = MockProfile(22)
    stats.startRequest(11, profile22, 33)
    data = {}
    stats.updateRequestStats(data, user=None)
    data["testtracker"]["curr"][0]["runtimeOngoing"] = "r0"
    data["testtracker"]["curr"][1]["runtimeOngoing"] = "r1"
    self.assertEqual(data, {
      "testtracker": {
        "active": 2,
        "capacity": 100,
        "perUserCapacity": 50,
        "pendingCount": 0,
        "total": 2,
        "removalErrors": 0,
        "longest": [],
        "curr": [
          {
            "runtimeOngoing": "r0",
            "startTime": "1970-01-01T00:00:03",
            "request": "1",
            "username": "user_2",
            "querytracker": "querytracker_2"
          },
          {
            "runtimeOngoing": "r1",
            "startTime": "1970-01-01T00:00:33",
            "request": "11",
            "username": "user_22",
            "querytracker": "querytracker_22"
          }
        ],
        "pending": []
      }
    })

    stats.stopRequest(1, profile2, 3, 4)
    data = {}
    stats.updateRequestStats(data, user=None)
    data["testtracker"]["curr"][0]["runtimeOngoing"] = "r1"
    self.assertEqual(data, {
      "testtracker": {
        "active": 1,
        "capacity": 100,
        "perUserCapacity": 50,
        "pendingCount": 0,
        "total": 2,
        "removalErrors": 0,
        "longest": [],
        "curr": [
          {
            "runtimeOngoing": "r1",
            "startTime": "1970-01-01T00:00:33",
            "request": "11",
            "username": "user_22",
            "querytracker": "querytracker_22"
          }
        ],
        "pending": []
      }
    })

    stats.stopRequest(11, profile22, 33, 44)
    data = {}
    stats.updateRequestStats(data, user=None)
    self.assertEqual(data, {
      "testtracker": {
        "active": 0,
        "capacity": 100,
        "perUserCapacity": 50,
        "pendingCount": 0,
        "total": 2,
        "removalErrors": 0,
        'longest': [{
          'querytracker': 'querytracker_22',
          'request': '11',
          'runtime': '0:00:44',
          'startTime': '1970-01-01T00:00:33',
          'username': 'user_22'
        }],
        "curr": [],
        "pending": []
      }
    })

    stats.startRequest(1, profile2, 3)
    data = {}
    stats.updateRequestStats(data, user=None)
    data["testtracker"]["curr"][0]["runtimeOngoing"] = "r0"
    self.assertEqual(data, {
      "testtracker": {
        "active": 1,
        "capacity": 100,
        "perUserCapacity": 50,
        "pendingCount": 0,
        "total": 3,
        "removalErrors": 0,
        'longest': [{
          'querytracker': 'querytracker_22',
          'request': '11',
          'runtime': '0:00:44',
          'startTime': '1970-01-01T00:00:33',
          'username': 'user_22'
        }],
        "curr": [
          {
            "runtimeOngoing": "r0",
            "startTime": "1970-01-01T00:00:03",
            "request": "1",
            "username": "user_2",
            "querytracker": "querytracker_2"
          }
        ],
        "pending": []
      }
    })

    requestData = MockRequestData(11)
    stats.enterAdmissionControl(requestData, profile22, 33)
    data = {}
    stats.updateRequestStats(data, user=None)
    data["testtracker"]["curr"][0]["runtimeOngoing"] = "r0"
    data["testtracker"]["pending"][0]["runtimeOngoing"] = "r1"
    self.assertEqual(data, {
      "testtracker": {
        "active": 1,
        "capacity": 100,
        "perUserCapacity": 50,
        "pendingCount": 1,
        "total": 3,
        "removalErrors": 0,
        'longest': [{
          'querytracker': 'querytracker_22',
          'request': '11',
          'runtime': '0:00:44',
          'startTime': '1970-01-01T00:00:33',
          'username': 'user_22'
        }],
        "curr": [
          {
            "runtimeOngoing": "r0",
            "startTime": "1970-01-01T00:00:03",
            "request": "1",
            "username": "user_2",
            "querytracker": "querytracker_2"
          }
        ],
        'pending': [{
          'querytracker': 'querytracker_22',
          'queue': 'admission_11',
          'request': 'str_11',
          'runtimeOngoing': 'r1',
          'startTime': '1970-01-01T00:00:33',
          'username': 'user_22'
        }]
      }
    })

    stats.exitAdmissionControl(requestData, profile22, 33)
    data = {}
    stats.updateRequestStats(data, user=None)
    data["testtracker"]["curr"][0]["runtimeOngoing"] = "r0"
    self.assertEqual(data, {
      "testtracker": {
        "active": 1,
        "capacity": 100,
        "perUserCapacity": 50,
        "pendingCount": 0,
        "total": 3,
        "removalErrors": 0,
        'longest': [{
          'querytracker': 'querytracker_22',
          'request': '11',
          'runtime': '0:00:44',
          'startTime': '1970-01-01T00:00:33',
          'username': 'user_22'
        }],
        "curr": [
          {
            "runtimeOngoing": "r0",
            "startTime": "1970-01-01T00:00:03",
            "request": "1",
            "username": "user_2",
            "querytracker": "querytracker_2"
          }
        ],
        'pending': []
      }
    })

  def test_parallel(self):
    # So we can see any errors during asserts in full
    self.maxDiff = None

    stats = vizlib.util.RequestStatsTracker('testtracker', capacity=100, perUserCapacity=50)

    def startStop(idx):
      requestData = MockRequestData(idx)
      profile = MockProfile(idx)

      if idx < 480:
        for i in range(0, 100):
          stats.startRequest(idx, profile, idx)
          stats.stopRequest(idx, profile, idx, idx)
          stats.enterAdmissionControl(requestData, profile, idx)
          stats.exitAdmissionControl(requestData, profile, idx)

      if idx >= 480:
        stats.startRequest(idx, profile, idx)
        stats.stopRequest(idx, profile, idx, idx)

      if idx % 100 == 0:
        stats.startRequest(idx, profile, idx)

      if idx % 99 == 0:
        stats.enterAdmissionControl(requestData, profile, idx)

    workq = cli.core.threadpool.WorkQ()
    for idx in range(0, 500):
      workq.add_work(cli.core.threadpool.Runner(startStop, (idx, )))

    threadpool = cli.core.threadpool.ThreadPool(maxlen=20)

    threadpool.run_workq(workq)

    data = {}
    stats.updateRequestStats(data, user=None)
    self.assertEqual(len(data["testtracker"]["longest"]), 20)
    self.assertEqual(data["testtracker"]["longest"][0]["querytracker"], "querytracker_480")
    self.assertEqual(data["testtracker"]["longest"][19]["querytracker"], "querytracker_499")

    data["testtracker"]["longest"] = []
    for d in data["testtracker"]["curr"]:
      d["runtimeOngoing"] = -1

    for d in data["testtracker"]["pending"]:
      d["runtimeOngoing"] = -1

    self.assertEqual(data, {
      "testtracker": {
        "active": 5,
        "capacity": 100,
        "perUserCapacity": 50,
        "pendingCount": 6,
        "total": 48025,
        "removalErrors": 0,
        "longest": [],
        'curr': [{'querytracker': 'querytracker_0',
                  'request': '0',
                  'runtimeOngoing': -1,
                  'startTime': '1970-01-01T00:00:00',
                  'username': 'user_0'},
                 {'querytracker': 'querytracker_100',
                  'request': '100',
                  'runtimeOngoing': -1,
                  'startTime': '1970-01-01T00:01:40',
                  'username': 'user_100'},
                 {'querytracker': 'querytracker_200',
                  'request': '200',
                  'runtimeOngoing': -1,
                  'startTime': '1970-01-01T00:03:20',
                  'username': 'user_200'},
                 {'querytracker': 'querytracker_300',
                  'request': '300',
                  'runtimeOngoing': -1,
                  'startTime': '1970-01-01T00:05:00',
                  'username': 'user_300'},
                 {'querytracker': 'querytracker_400',
                  'request': '400',
                  'runtimeOngoing': -1,
                  'startTime': '1970-01-01T00:06:40',
                  'username': 'user_400'}],
        'pending': [{'querytracker': 'querytracker_0',
                     'queue': 'admission_0',
                     'request': 'str_0',
                     'runtimeOngoing': -1,
                     'startTime': '1970-01-01T00:00:00',
                     'username': 'user_0'},
                    {'querytracker': 'querytracker_99',
                     'queue': 'admission_99',
                     'request': 'str_99',
                     'runtimeOngoing': -1,
                     'startTime': '1970-01-01T00:01:39',
                     'username': 'user_99'},
                    {'querytracker': 'querytracker_198',
                     'queue': 'admission_198',
                     'request': 'str_198',
                     'runtimeOngoing': -1,
                     'startTime': '1970-01-01T00:03:18',
                     'username': 'user_198'},
                    {'querytracker': 'querytracker_297',
                     'queue': 'admission_297',
                     'request': 'str_297',
                     'runtimeOngoing': -1,
                     'startTime': '1970-01-01T00:04:57',
                     'username': 'user_297'},
                    {'querytracker': 'querytracker_396',
                     'queue': 'admission_396',
                     'request': 'str_396',
                     'runtimeOngoing': -1,
                     'startTime': '1970-01-01T00:06:36',
                     'username': 'user_396'},
                    {'querytracker': 'querytracker_495',
                     'queue': 'admission_495',
                     'request': 'str_495',
                     'runtimeOngoing': -1,
                     'startTime': '1970-01-01T00:08:15',
                     'username': 'user_495'}],
      }
    })

