#!/usr/bin/python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import unittest
import os
import sys
import time
import threading

from arcwebbase import access_info
from util.perfcounters import APPROX_COUNTERS
from unittest import mock
from vizlib import util

THISDIR = os.path.dirname(os.path.realpath(__file__))

sys.path.append("%s/.." % THISDIR)

import workq

def _logit(s):
  pass
  # print(s)

class ConditionalSingleWorker(workq.ProgressRunner):
  def __init__(self, idx, lock, cv, notifyCb, category=None):
    self.idx = idx
    self.lock = lock
    self.cv = cv
    self.notifyCb = notifyCb
    self.category = category
    self.completed = False

  def admissionControlCategory(self):
    return self.category

  def unpause(self):
    with self.lock:
      self.cv.notifyAll()

  def _pauseLocked(self):
    self.cv.wait()

  def run(self):
    with self.lock:
      # Let the controlling process know we grabbed the lock, and are waiting
      self.notifyCb()

      self._pauseLocked()

    self.completed = True

  def progress(self):
    if self.completed:
      return (2, 2)

    return (1, 2)

  def results(self):
    return f'done_{self.idx}'

class ConditionalMultiStepWorker(workq.ProgressRunner):
  def __init__(self, phase_count, lock, cv, notifyCb):
    self.phase_count = phase_count
    self.completed = 0
    self.create_time = time.time()
    self.lock = lock
    self.cv = cv
    self.notifyCb = notifyCb

  def unpause(self):
    with self.lock:
      self.cv.notifyAll()

  def _pauseLocked(self):
    self.cv.wait()

  # Note that you get 1 notification cb for each phase,
  # then 1 more when all stages are complete,
  # then 1 more just before exit.
  def run(self):
    self.start_time = time.time()
    for i in range(0, self.phase_count):
      with self.lock:
        _logit(('notifying', i, '/', self.phase_count))
        self.notifyCb()
        self._pauseLocked()
        self.completed = i + 1
        _logit(('waking up', i+1, '/', self.phase_count))

    with self.lock:
      _logit(('nofifying', 'complete'))
      self.notifyCb()
      self._pauseLocked()
      _logit(('waking up', 'complete'))

    self.end_time = time.time()
    self.duration = self.end_time - self.start_time
    with self.lock:
      self.notifyCb()

  def progress(self):
    return (self.completed, self.phase_count)

  def results(self):
    return 'done'

class TestConcurrency(unittest.TestCase):
  def setUp(self):
    self.notifyCount = 0
    self.threadcount = 10
    self.categoryConcurrency = 5

    self.lock = threading.Lock()
    self.cv = threading.Condition(self.lock)
    self.asyncRunner = workq.AsyncRunner(threadcount=self.threadcount, name="test",
        categoryConcurrency=self.categoryConcurrency)

  def notifyCb(self):
    self.notifyCount += 1

  def test_1_simple(self):
    items = []
    for idx in range(0, 50):
      items.append(ConditionalSingleWorker(idx, self.lock, self.cv, self.notifyCb))

    handles = []
    for item in items:
      handles.append(self.asyncRunner.queue(item))

    for idx in range(0, 5):
      while self.notifyCount != (idx + 1) * self.threadcount:
        time.sleep(0.01)

      cnt = self.notifyCount
      items[0].unpause()

      for i in range(0, cnt):
        while self.asyncRunner.status(handles[i]) != workq.Runner.DONE:
          time.sleep(0.01)

        self.assertEqual(items[i].progress(), (2, 2))
        self.assertEqual(items[i].results(), f"done_{i}")

  def test_3_manyconcurrent(self):
    for changingCategoryConcurrency in [1, 3, 5, 7, 9, 10]:
      self.categoryConcurrency = changingCategoryConcurrency
      self.asyncRunner.setParams(threadcount=self.threadcount,
          categoryConcurrency=self.categoryConcurrency)
      self.notifyCount = 0

      items = []
      for categoryIdx in range(65, 75):
        for idx in range(0, 25):
          items.append(ConditionalSingleWorker(idx*100 + categoryIdx,
              self.lock, self.cv, self.notifyCb, category=f"category{chr(categoryIdx)}"))

      self.assertEqual(len(items), 250)

      handles = []
      for item in items:
        handles.append(self.asyncRunner.queue(item))

      for iterationIdx in range(0, 25):
        # Wait for self.threadcount items to norify us that they are waiting
        while self.notifyCount != (iterationIdx + 1) * self.threadcount:
          time.sleep(0.01)

        # Wake all self.threadcount items up so they complete
        cnt = self.notifyCount
        items[0].unpause()

        # Wait for self.threadcount items to be done
        while True:
          doneCnt = 0
          for item in items:
            if item.progress() == (2, 2):
              doneCnt += 1

          self.assertTrue(doneCnt <= self.threadcount)
          if doneCnt == self.threadcount:
            break

          time.sleep(0.01)

        # Remove those self.threadcount items from the list
        doneItems = []
        newItems = []
        for item in items:
          if item.progress() == (2, 2):
            doneItems.append(item)
          else:
            newItems.append(item)

        items = newItems
        self.assertEqual(len(doneItems), self.threadcount)

        # for the items that are done, verify that no more than self.categoryConcurrency
        # were from the same category
        categoryConcurrencies = {}
        for doneItem in doneItems:
          category = doneItem.admissionControlCategory()
          categoryConcurrencies[category] = categoryConcurrencies.get(category, 0) + 1

        for k in categoryConcurrencies:
          self.assertTrue(categoryConcurrencies[k] <= self.categoryConcurrency)

  # Add in 10 items of a category A, and 40 items of no category
  # Confirm that the first 5 items get done for A before the other items for those
  #   items get postponed. The remaining items only get done after the 40 non-category
  #   items get done (because they get added back at the end of the list)
  def test_2_concurrency(self):
    items = []
    for idx in range(0, 50):
      if idx < 10:
        items.append(ConditionalSingleWorker(idx, self.lock, self.cv, self.notifyCb,
            category="categoryA"))
        continue

      items.append(ConditionalSingleWorker(idx, self.lock, self.cv, self.notifyCb))

    handles = []
    for item in items:
      handles.append(self.asyncRunner.queue(item))

    for idx in range(0, 5):
      while self.notifyCount != (idx + 1) * self.threadcount:
        time.sleep(0.01)

      cnt = self.notifyCount
      items[0].unpause()

      for i in range(0, self.categoryConcurrency):
        while self.asyncRunner.status(handles[i]) != workq.Runner.DONE:
          time.sleep(0.01)

        self.assertEqual(items[i].progress(), (2, 2))

      for i in range(self.categoryConcurrency, 10):
        if idx < 4:
          self.assertEqual(items[i].progress(), (1, 2))
        else:
          while self.asyncRunner.status(handles[i]) != workq.Runner.DONE:
            time.sleep(0.01)

          self.assertEqual(items[i].progress(), (2, 2))

      for i in range(10, 10 + cnt - self.categoryConcurrency):
        if i >= len(items):
          break

        while self.asyncRunner.status(handles[i]) != workq.Runner.DONE:
          time.sleep(0.01)

        self.assertEqual(items[i].progress(), (2, 2))

def waitForEquals(test: unittest.TestCase, func, value2, message, timeout=1):
  start = time.time()
  counter = 0
  while not func() == value2:
    # print('waitForEquals count %s'%counter)
    counter += 1
    time.sleep(0.1)
    if time.time() - start > timeout:
      break

  test.assertEquals(func(), value2, message)

class TestConcurrencyWithCancellation(unittest.TestCase):
  def setUp(self):
    self.notifyCount = 0

  def notifyCb(self):
    self.notifyCount += 1

  # Concurrency is 2, category concurrency is 1.
  #  - add 3 items of a category A. One worker starts working on the first,
  #    the other two will be marked as userpending.
  #  - cancel the two pending queries
  #  - verify that the pending queries are removed from the storage
  def test_1(self):
    lock = threading.Lock()
    cv = threading.Condition(lock)
    asyncRunner = workq.AsyncRunner(
      name='test',
      threadcount=2,
      categoryConcurrency=1)

    w1 = ConditionalSingleWorker(1, lock, cv, self.notifyCb, category='A')
    asyncRunner.queue(w1)

    waitForEquals(self, lambda: self.notifyCount, 1, 'one worker running')

    w2 = ConditionalSingleWorker(2, lock, cv, self.notifyCb, category='A')
    w3 = ConditionalSingleWorker(3, lock, cv, self.notifyCb, category='A')

    handle2 = asyncRunner.queue(w2)
    handle3 = asyncRunner.queue(w3)

    waitForEquals(self, lambda: APPROX_COUNTERS.get("test.workq.userpending").value, 2, 'two items pending in the user queue')

    # cancel the items in the user queue
    asyncRunner.markCancelled(handle2)
    asyncRunner.markCancelled(handle3)

    # let the first thread proceed. Since it worked on category A, it is responsible
    # for clearing the queued tasks for that category
    w1.unpause()

    # userpending items should clear up
    waitForEquals(self, lambda: APPROX_COUNTERS.get("test.workq.userpending").value, 0, 'no items in the user queue')



class TestWorkQ(unittest.TestCase):
  def setUp(self):
    self.setup_time = time.time()
    self.async_runner = workq.AsyncRunner(threadcount=2, name="test")
    self.notify_lock = threading.Lock()
    self.notify_count = 0
    self.lock = threading.Lock()
    self.cv = threading.Condition(self.lock)

  def tearDown(self):
    self.async_runner.cancel()

  def notifyCb(self):
    with self.notify_lock:
      self.notify_count += 1

  def test_1_single(self):
    # Run a single request through its paces
    s1 = ConditionalMultiStepWorker(2, self.lock, self.cv, self.notifyCb)

    # Queue up the request
    # Confirm its handle, and that it is in stage 0
    # Hold the lock, so it can't move ahead yet
    with self.lock:
      h1 = self.async_runner.queue(s1)

      self.assertTrue(self.async_runner.status(h1) in \
          [workq.Runner.QUEUED, workq.Runner.RUNNING])
      self.assertEqual(self.async_runner.progress(h1), (0, 2))

    # Let it move into stage 1, and verify status
    while self.notify_count != 1:
      time.sleep(0.01)
    self.assertEqual(self.async_runner.progress(h1), (0, 2))

    s1.unpause()
    while self.notify_count != 2:
      time.sleep(0.01)
    self.assertEqual(self.async_runner.status(h1), workq.Runner.RUNNING)
    self.assertEqual(self.async_runner.progress(h1), (1, 2))

    # Let it move into stage 2, and verify status
    s1.unpause()
    while self.notify_count != 3:
      time.sleep(0.01)
    self.assertEqual(self.async_runner.status(h1), workq.Runner.RUNNING)
    self.assertEqual(self.async_runner.progress(h1), (2, 2))

    # Wait for completion, and verify status
    s1.unpause()
    while self.notify_count != 4:
      time.sleep(0.01)

    # Some short time after notify_count, the thread will be DONE
    while self.async_runner.status(h1) != workq.Runner.DONE:
      time.sleep(0.01)

    self.assertEqual(self.async_runner.status(h1), workq.Runner.DONE)
    self.assertEqual(self.async_runner.progress(h1), (2, 2))

    self.async_runner.stop_tracking(h1)
    self.assertEqual(self.async_runner.status(h1), None)

  def test_2_threadcount(self):
    # Run two requests (same as threadcount) through their paces
    s1 = ConditionalMultiStepWorker(2, self.lock, self.cv, self.notifyCb)
    s2 = ConditionalMultiStepWorker(2, self.lock, self.cv, self.notifyCb)

    # Queue up the requests
    # Confirm handles, and that they are in stage 0
    with self.lock:
      h1 = self.async_runner.queue(s1)
      h2 = self.async_runner.queue(s2)
      self.assertTrue(int(h2) > int(h1))

      for h in [h1, h2]:
        self.assertTrue(self.async_runner.status(h) in \
            [workq.Runner.QUEUED, workq.Runner.RUNNING])
        self.assertEqual(self.async_runner.progress(h), (0, 2))

    # Let them move into stage 1, and verify status
    while self.notify_count != 2:
      time.sleep(0.01)
    s1.unpause()
    while self.notify_count != 4:
      time.sleep(0.01)
    for h in [h1, h2]:
      self.assertEqual(self.async_runner.status(h), workq.Runner.RUNNING)
      self.assertEqual(self.async_runner.progress(h), (1, 2))

    # Let them move into stage 2, and verify status
    s1.unpause()
    while self.notify_count != 6:
      time.sleep(0.01)
    for h in [h1, h2]:
      self.assertEqual(self.async_runner.status(h), workq.Runner.RUNNING)
      self.assertEqual(self.async_runner.progress(h), (2, 2))

    # Wait for completion, and verify status
    with self.lock:
      self.cv.notifyAll()

    for h in [h1, h2]:
      while self.async_runner.status(h) != workq.Runner.DONE:
        time.sleep(0.01)

    for h in [h1, h2]:
      self.assertEqual(self.async_runner.status(h), workq.Runner.DONE)
      self.assertEqual(self.async_runner.progress(h), (2, 2))

    for h in [h1, h2]:
      self.async_runner.stop_tracking(h)
      self.assertEqual(self.async_runner.status(h), None)

  def test_3_2x_threadcount(self):
    # Run four requests (2x threadcount) through their paces
    s1 = ConditionalMultiStepWorker(2, self.lock, self.cv, self.notifyCb)
    s2 = ConditionalMultiStepWorker(2, self.lock, self.cv, self.notifyCb)
    s3 = ConditionalMultiStepWorker(2, self.lock, self.cv, self.notifyCb)
    s4 = ConditionalMultiStepWorker(2, self.lock, self.cv, self.notifyCb)

    # Queue up the requests
    # Confirm handles, and that they are in stage 0
    with self.lock:
      h1 = self.async_runner.queue(s1)
      h2 = self.async_runner.queue(s2)
      h3 = self.async_runner.queue(s3)
      h4 = self.async_runner.queue(s4)
      self.assertTrue(int(h2) > int(h1))
      self.assertTrue(int(h3) > int(h2))
      self.assertTrue(int(h4) > int(h3))

      for h in [h1, h2]:
        self.assertTrue(self.async_runner.status(h) in \
            [workq.Runner.QUEUED, workq.Runner.RUNNING])
        self.assertEqual(self.async_runner.progress(h), (0, 2))
      for h in [h3, h4]:
        self.assertTrue(self.async_runner.status(h) in \
            [workq.Runner.QUEUED])
        self.assertEqual(self.async_runner.progress(h), (0, 2))

    # Let them move into stage 1, and verify status
    while self.notify_count != 2:
      time.sleep(0.01)
    with self.lock:
      self.cv.notifyAll()
    while self.notify_count != 4:
      time.sleep(0.01)
    for h in [h1, h2]:
      self.assertEqual(self.async_runner.status(h), workq.Runner.RUNNING)
      self.assertEqual(self.async_runner.progress(h), (1, 2))
    for h in [h3, h4]:
      self.assertTrue(self.async_runner.status(h) in \
          [workq.Runner.QUEUED])
      self.assertEqual(self.async_runner.progress(h), (0, 2))

    # Let them move into stage 2, and verify status
    with self.lock:
      self.cv.notifyAll()
    while self.notify_count != 6:
      time.sleep(0.01)
    with self.lock:
      self.cv.notifyAll()
    for h in [h1, h2]:
      self.assertEqual(self.async_runner.status(h), workq.Runner.RUNNING)
      self.assertEqual(self.async_runner.progress(h), (2, 2))
    for h in [h3, h4]:
      self.assertTrue(self.async_runner.status(h) in \
          [workq.Runner.QUEUED])
      self.assertEqual(self.async_runner.progress(h), (0, 2))

    # Wait for completion, and verify status
    while self.notify_count != 10:
      time.sleep(0.01)
    for h in [h1, h2]:
      while self.async_runner.status(h) != workq.Runner.DONE:
        time.sleep(0.01)

    for h in [h1, h2]:
      self.assertEqual(self.async_runner.status(h), workq.Runner.DONE)
      self.assertEqual(self.async_runner.progress(h), (2, 2))
    for h in [h3, h4]:
      self.assertTrue(self.async_runner.status(h) in \
          [workq.Runner.QUEUED, workq.Runner.RUNNING])
      self.assertEqual(self.async_runner.progress(h), (0, 2))

    # Let them move into stage 1, and verify status
    with self.lock:
      self.cv.notifyAll()
    while self.notify_count != 12:
      time.sleep(0.01)
    for h in [h3, h4]:
      self.assertEqual(self.async_runner.status(h), workq.Runner.RUNNING)
      self.assertEqual(self.async_runner.progress(h), (1, 2))

    # Let them move into stage 2, and verify status
    with self.lock:
      self.cv.notifyAll()
    while self.notify_count != 14:
      time.sleep(0.01)
    with self.lock:
      self.cv.notifyAll()
    for h in [h3, h4]:
      self.assertEqual(self.async_runner.status(h), workq.Runner.RUNNING)
      self.assertEqual(self.async_runner.progress(h), (2, 2))

    # Wait for completion, and verify status
    while self.notify_count != 16:
      time.sleep(0.01)
    for h in [h1, h2]:
      while self.async_runner.status(h) != workq.Runner.DONE:
        time.sleep(0.01)

    for h in [h1, h2, h3, h4]:
      self.assertEqual(self.async_runner.status(h), workq.Runner.DONE)
      self.assertEqual(self.async_runner.progress(h), (2, 2))

interrupts = APPROX_COUNTERS.get('test.interrupts')

class InterruptOnlyWorker(workq.ProgressRunner):
  def __init__(self, idx):
    self.idx = idx
    self.interrupted = False

  def interrupt(self):
    self.interrupted = True

  def run(self):
    while True:
      if self.interrupted:
        break

      time.sleep(0.001)

    global interrupts
    interrupts += 1

  def progress(self):
    if self.interrupted:
      return (2, 2)

    return (1, 2)

  def results(self):
    return f'done_{self.idx}'

class MockThread:
  def __init__(self, id):
    self._id = id
    self.calls = 0

  def interruptIfTimedOut(self, queryTimeoutSeconds=None):
    self.calls += 1

class MockAsyncRunner:
  def __init__(self, threadcount=3):
    self.name = 'MockAsyncRunner'
    self._threadcount = threadcount
    self.threads = []
    for i in range(0, self._threadcount):
      self.threads.append(MockThread(i))

    self.calls = 0

  def getTimedOutThreads(self, queryTimeoutSeconds=None):
    self.calls += 1
    return [self.threads[1], self.threads[2]]

  def cleanupTrackers(self):
    pass

class TestConcurrencyUpdate(unittest.TestCase):
  def get_usage_info(self):
    storage = {}
    util.updateRequestStats(storage)
    queue_data = storage[f'{self.name}.workq']
    return queue_data

  def test_request_stats_update(self):
    name = 'TestConcurrencyUpdate'
    self.name = name
    runner = workq.AsyncRunner(name=name, threadcount=19, categoryConcurrency=13)

    queue_data = self.get_usage_info()
    self.assertEqual(queue_data['capacity'], 19)
    self.assertEqual(queue_data['perUserCapacity'], 13)

    runner.setParams(threadcount=19, categoryConcurrency=14)
    queue_data = self.get_usage_info()
    self.assertEqual(queue_data['capacity'], 19)
    self.assertEqual(queue_data['perUserCapacity'], 14)

    runner.setParams(threadcount=20, categoryConcurrency=14)
    queue_data = self.get_usage_info()
    self.assertEqual(queue_data['capacity'], 20)
    self.assertEqual(queue_data['perUserCapacity'], 14)

    runner.setParams(threadcount=21, categoryConcurrency=15)
    queue_data = self.get_usage_info()
    self.assertEqual(queue_data['capacity'], 21)
    self.assertEqual(queue_data['perUserCapacity'], 15)

class TestQueryCleanupThread(unittest.TestCase):
  def setUp(self):
    workq.DEFAULT_CLEANUP_DELAY = 0.001
    workq.CLEANUP_QUERY_TRACKERS_GRACE_PERIOD_SECONDS = 1

  def test_1_simple(self):
    asyncRunner = MockAsyncRunner()
    queryTimeoutSeconds = 0.03
    delay = 0.01

    creates = APPROX_COUNTERS.get('workq.querycleanup.create')
    cancels = APPROX_COUNTERS.get('workq.querycleanup.cancels')

    cnt = creates.value
    queryCleanup = workq._QueryCleanupThread(asyncRunner, queryTimeoutSeconds, delay)
    self.assertEqual(cnt + 1, creates.value)

    # Let a few calls happen
    while asyncRunner.calls < 5:
      time.sleep(0.01)

    # Now cancel
    cnt = cancels.value
    queryCleanup.cancel()
    while True:
      if (cnt + 1 == cancels.value):
        break

      time.sleep(0.01)

    self.assertEqual(cnt + 1, cancels.value)

    # Confirm no additional calls happen
    runnerCalls = asyncRunner.calls
    time.sleep(0.02)
    self.assertEqual(runnerCalls, asyncRunner.calls)

    threads = asyncRunner.threads
    self.assertEqual(threads[0].calls, 0)
    self.assertEqual(threads[1].calls, asyncRunner.calls)
    self.assertEqual(threads[2].calls, asyncRunner.calls)

  def test_2_setparams(self):
    creates = APPROX_COUNTERS.get('workq.querycleanup.create')
    cancels = APPROX_COUNTERS.get('workq.querycleanup.cancels')

    createsCount = creates.value
    cancelsCount = cancels.value

    asyncRunner = workq.AsyncRunner(threadcount=2, name="test")
    self.assertEqual(createsCount, creates.value)
    self.assertEqual(cancelsCount, cancels.value)
    self.assertEqual([], asyncRunner.getTimedOutThreads())

    asyncRunner = workq.AsyncRunner(threadcount=2, name="test", queryTimeoutSeconds=0.03)
    self.assertEqual(createsCount + 1, creates.value)
    self.assertEqual(cancelsCount, cancels.value)
    self.assertEqual([], asyncRunner.getTimedOutThreads())

    asyncRunner.setParams(threadcount=2, queryTimeoutSeconds=0.04)
    self.assertEqual(createsCount + 2, creates.value)
    while cancelsCount + 1 != cancels.value:
      time.sleep(0.01)
    self.assertEqual(cancelsCount + 1, cancels.value)
    self.assertEqual([], asyncRunner.getTimedOutThreads())

    asyncRunner.setParams(threadcount=2, queryTimeoutSeconds=0.03)
    self.assertEqual(createsCount + 3, creates.value)
    while cancelsCount + 2 != cancels.value:
      time.sleep(0.01)
    self.assertEqual(cancelsCount + 2, cancels.value)
    self.assertEqual([], asyncRunner.getTimedOutThreads())

  def test_3_interrupt(self):
    global interrupts
    interruptsCount = interrupts.value

    itemCount = 10
    items = []
    for idx in range(0, itemCount):
      items.append(InterruptOnlyWorker(idx))

    self.assertEqual(interruptsCount, interrupts.value)

    creates = APPROX_COUNTERS.get('workq.querycleanup.create')
    cancels = APPROX_COUNTERS.get('workq.querycleanup.cancels')
    markInterrupted = APPROX_COUNTERS.get('workq.markInterrupted')

    createsCount = creates.value
    cancelsCount = cancels.value

    asyncRunner = workq.AsyncRunner(threadcount=2, name="test", queryTimeoutSeconds=0.003)

    markInterruptedCount = markInterrupted.value
    handles = []
    for item in items:
      handle = asyncRunner.queue(item)
      handles.append(handle)

    self.assertEqual(createsCount + 1, creates.value)
    self.assertEqual(cancelsCount, cancels.value)

    while True:
      newInterruptedCount = markInterrupted.value - markInterruptedCount
      if interruptsCount + itemCount == interrupts.value + newInterruptedCount:
        break

      time.sleep(0.001)

    newInterruptedCount = markInterrupted.value - markInterruptedCount
    self.assertEqual(interruptsCount + itemCount, interrupts.value + newInterruptedCount)

    # after a while, the results must all turn into None (trackers are removed by the cleanup thread)
    waitForEquals(self,
                  lambda: list(map(asyncRunner.results, handles)).count(None),
                  len(handles),
                  'handle results are all None',
                  timeout=workq.CLEANUP_QUERY_TRACKERS_GRACE_PERIOD_SECONDS+1)

class TestTrackerIdAllocator(unittest.TestCase):
  def test_id_allocation_unique(self):
    runner_1 = workq.AsyncRunner()
    runner_2 = workq.AsyncRunner()

    handle_count = 3
    handles_1 = {runner_1.queue(workq.ProgressRunner()) for _ in range(handle_count)}
    handles_2 = {runner_2.queue(workq.ProgressRunner()) for _ in range(handle_count)}

    # handles are unique:
    self.assertEquals(len(handles_1), handle_count)
    self.assertEquals(len(handles_2), handle_count)

    # the handles must have no intersection:
    self.assertEquals(handles_1.intersection(handles_2), set())

class TestAccessInfoUsage(unittest.TestCase):
  def tearDown(self):
    return access_info.clear()

  def test_access_info_propagation(self):

    runner = workq.AsyncRunner()

    ai = access_info.AccessInfo(username='workq_test_user')
    access_info.store(ai)

    handle = runner.queue(workq.ProgressRunner())

    storeSpy = mock.MagicMock(wraps=access_info.store)
    clearSpy = mock.MagicMock(wraps=access_info.clear)
    with mock.patch.object(access_info, 'store', storeSpy):
      with mock.patch.object(access_info, 'clear', clearSpy):

        # wait for processing:
        waitForEquals(self, lambda: runner.status(handle), workq.Runner.DONE, 'Task done')

        storeSpy.assert_called_once()
        clearSpy.assert_called_once()

        stored_access_info = storeSpy.call_args.args[0]
        self.assertEquals(stored_access_info.username, 'workq_test_user')


if __name__ == '__main__':
  unittest.main()

