#!/usr/bin/python
# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

from django.test.utils import CaptureQueriesContext
from rbac.models import RBAC
from typing import Callable, TypeVar
from unittest import mock

import django.db
import os
import unittest

import logging

LOG = logging.getLogger(__name__)

"""
This is the query loading all roles.
Role caching is complicated. Right now we might opt to ignore these queries as
they can appear and disappear randomly.
"""
RBAC_ROLE_LOAD_QUERY = 'SELECT "rbac_role"."id", "rbac_role"."name", "rbac_role"."description", "rbac_role"."privileges", "rbac_role"."users", "rbac_role"."groups", "rbac_role"."sentry_groups", "rbac_role"."require_all_groups" FROM "rbac_role"'

PRINT_DB_QUERIES = os.environ.get('PRINT_DB_QUERIES', 'false') == 'true'
IGNORE_QUERY_COUNT_MISMATCH = os.environ.get('IGNORE_QUERY_COUNT_MISMATCH', 'false') == 'true'

T = TypeVar('T')


def run_with_db_query_check(function: Callable[[], T], expected_query_count:int=None, asserter:unittest.TestCase=None) -> tuple[T, int]:
  django.db.reset_queries()
  # queries are collected in
  # django.db.connection.queries

  RBAC.reset()  # permissions are cached. If not cleared, can lead to a lower DB access count

  with CaptureQueriesContext(django.db.connection) as query_context:
    try:
      # if tests run slowly, roles might get reloaded. Try to prevent this:
      with mock.patch.object(RBAC.change_checker, 'is_changed', return_value=False):
        return_value = function()
    finally:
      print_db_queries(query_context)
  query_count = len(query_context.captured_queries)

  if expected_query_count is not None and expected_query_count != query_count:
    message = f'DB Query count mismatch: expected {expected_query_count}, actual: ' + db_queries_as_string(query_context.captured_queries)
    if IGNORE_QUERY_COUNT_MISMATCH:
      LOG.warning(message)
    else:
      asserter.assertEqual(query_count, expected_query_count, message)

  return return_value, len(query_context.captured_queries)


def run_with_db_query_check2(function: Callable[[], T]) -> T:
  """
  Return the result alone, without the query count
  """
  result, _ = run_with_db_query_check(function)
  return result


def db_queries_as_string(queries: list[dict]):
  result = f'DB queries count {len(queries)}'
  for c in queries:
    result += f"\n  {c['sql']}"
  result += '\n'
  return result


def print_db_queries(queryContext: CaptureQueriesContext):
  if not PRINT_DB_QUERIES:
    return

  print(db_queries_as_string(queryContext.captured_queries))


def get_db_queries(exclude_rbac_role_loads=False):
  queries = [query['sql'] for query in django.db.connection.queries]

  result = queries
  if exclude_rbac_role_loads:
    result = [query for query in queries if not query == RBAC_ROLE_LOAD_QUERY]

  return result
