# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import threading
import bisect
import time
import logging
import json
import re

from datetime import timedelta

LOGGER = logging.getLogger("arcweb")

class IdAllocator:
  def __init__(self, startId=None):
    self._id = startId or int(time.time())
    self._lock = threading.Lock()

  def allocateId(self):
    with self._lock:
      self._id += 1
      return str(self._id)

class KeepLargestItemsList:
  def __init__(self, maxCount=20):
    self.__maxCount = maxCount
    self.__data = []
    self.__lock = threading.Lock()
    self.__isFull = False

  def append(self, d):
    if self.__isFull and d < self.__data[0]:
      return

    with self.__lock:
      if not self.__isFull:
        if len(self.__data) < self.__maxCount:
          bisect.insort(self.__data, d)
          return
        else:
          self.__isFull = True

      bisect.insort(self.__data, d)
      self.__data = self.__data[1:]

  def getCopy(self):
    with self.__lock:
      return self.__data[:]

ALL_REQUEST_STATS_TRACKERS = set['RequestStatsTracker']()
ALL_REQUEST_STATS_TRACKERS_LOCK = threading.Lock()

def updateRequestStats(d, user=None):
  with ALL_REQUEST_STATS_TRACKERS_LOCK:
    trackersCopy = list(ALL_REQUEST_STATS_TRACKERS)

  for t in trackersCopy:
    t.updateRequestStats(d, user=user)

class _SingleRequestTracker:
  def __init__(self, tracker, requestData, profile, startTime):
    self._tracker = tracker
    self._requestData = requestData
    self._profile = profile
    self._startTime = startTime

  def __enter__(self):
    self._tracker.startRequest(self._requestData, self._profile, self._startTime)

  def __exit__(self, type, value, traceback):
    self._tracker.stopRequest(self._requestData, self._profile, self._startTime,
        time.time() - self._startTime)

class RequestStatsTracker:
  def __init__(self, name, capacity=None, perUserCapacity=None):
    self._name = name
    self._capacity = capacity
    self._perUserCapacity = perUserCapacity
    self._admissionControlRequests = set()
    self._currentRequests = set()
    self._longestRequests = KeepLargestItemsList(maxCount=20)
    self._removalErrors = 0

    self._total = 0
    self._lock = threading.Lock()

    # Log longest requests per hour
    self._longestTimePeriodSeconds = 60 * 60

    # Log any requests greater than this time
    self._longestLogThresholdSeconds = 30

    self._currentTimePeriod = int(time.time() / self._longestTimePeriodSeconds)

    with ALL_REQUEST_STATS_TRACKERS_LOCK:
      ALL_REQUEST_STATS_TRACKERS.add(self)

  def updateCapacity(self, capacity, perUserCapacity):
    self._capacity = capacity
    self._perUserCapacity = perUserCapacity

  def tracker(self, requestData, profile, startTime=None):
    if not startTime:
      startTime = time.time()

    return _SingleRequestTracker(self, requestData, profile, startTime)

  def enterAdmissionControl(self, requestData, profile, startTime):
    with self._lock:
      self._admissionControlRequests.add((startTime, requestData, profile))

  def exitAdmissionControl(self, requestData, profile, startTime):
    with self._lock:
      try:
        self._admissionControlRequests.remove((startTime, requestData, profile))
      except:
        self._removalErrors += 1
        LOGGER.info('Pending Removal failure %s %s %s', startTime, requestData, profile)

  def startRequest(self, requestData, profile, startTime):
    # Check if periodic logs about the longest requests have to be sent out
    timePeriod = int(time.time() / self._longestTimePeriodSeconds)
    droppedLongestRequests = []
    if timePeriod != self._currentTimePeriod:
      with self._lock:
        if timePeriod != self._currentTimePeriod:
          droppedLongestRequests = self._longestRequests.getCopy()
          self._currentTimePeriod = timePeriod
          self._longestRequests = KeepLargestItemsList(maxCount=20)

    with self._lock:
      self._total += 1
      self._currentRequests.add((startTime, str(requestData), profile))

    if droppedLongestRequests:
      for idx, r in enumerate(droppedLongestRequests):
        (runtime, startTime, requestData, profile) = r
        LOGGER.info(f'Slow request. querytracker [{profile.querytracker()}] '
            f'{self._name}.longest.{idx} '
            f'runtime={str(timedelta(seconds=runtime))} '
            f"startTime={time.strftime('%Y-%m-%dT%H:%M:%S', time.localtime(startTime))} "
            f'request={str(requestData)}')

  def stopRequest(self, requestData, profile, startTime, runtime):
    with self._lock:
      if runtime > self._longestLogThresholdSeconds:
        self._longestRequests.append((runtime, startTime, requestData, profile))

      try:
        self._currentRequests.remove((startTime, str(requestData), profile))
      except:
        self._removalErrors += 1
        LOGGER.info('Removal failure %s %s', startTime, requestData)

  def __requestToDict(self, r, username):
    (runtime, startTime, requestData, profile) = r
    if not username or username == profile.username():
      return {
        "runtime": str(timedelta(seconds=runtime)),
        "startTime": time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime(startTime)),
        "request": str(requestData),
        "username": profile.username(),
        "querytracker": profile.querytracker()
      }

    return None

  def updateRequestStats(self, counters, user=None):
    with self._lock:
      currentRequests = list(self._currentRequests)
      admissionControlRequests = list(self._admissionControlRequests)
      longestRequests = self._longestRequests.getCopy()

    currentRequests.sort()
    admissionControlRequests.sort()

    s = {}
    counters[self._name] = s

    s["active"] = len(currentRequests)
    if self._capacity is not None:
      s["capacity"] = self._capacity
    if self._perUserCapacity is not None:
      s["perUserCapacity"] = self._perUserCapacity

    s["pendingCount"] = len(admissionControlRequests)
    s["total"] = self._total
    s["removalErrors"] = self._removalErrors

    if user:
      username = user.username
    else:
      username = None

    sLongest = []
    for idx, r in enumerate(longestRequests):
      d = self.__requestToDict(r, username)
      if d:
        sLongest.append(d)

    s["longest"] = sLongest

    t = time.time()

    sCurr = []
    for idx, r in enumerate(currentRequests):
      (startTime, requestData, profile) = r
      if not username or username == profile.username():
        d = {
          "runtimeOngoing": str(timedelta(seconds=int(t - startTime))),
          "startTime": time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime(startTime)),
          "request": str(requestData),
          "username": profile.username(),
          "querytracker": profile.querytracker()
        }

        sCurr.append(d)

    s["curr"] = sCurr

    sPending = []
    for idx, r in enumerate(admissionControlRequests):
      (startTime, requestData, profile) = r

      runner = requestData.runner()
      category = runner.admissionControlCategory()

      if not username or username == profile.username():
        d = {
          "runtimeOngoing": str(timedelta(seconds=int(t - startTime))),
          "startTime": time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime(startTime)),
          "queue": category,
          "request": str(requestData),
          "username": profile.username(),
          "querytracker": profile.querytracker()
        }

        sPending.append(d)

    s["pending"] = sPending

