# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.
import logging
from .select_builder import BaseSelectBuilder, HiveSelectBuilder, OracleSelectBuilder, CassandraSelectBuilder
import dataconn.dataconn
from dataconn.connections_types import DATACONN_HIVE, DATACONN_IMPALA, DATACONN_ARCENGINE, \
  DATACONN_SPARKSQL, DATACONN_ORACLE, DATACONN_DRILL, DATACONN_PG, DATACONN_REDSHIFT, \
  DATACONN_TERADATA, DATACONN_SNOWFLAKE, DATACONN_MSSQL, DATACONN_CASSANDRA, DATACONN_IMPYLA, \
  DATACONN_HYVE, DATACONN_ARCENGINE_V2
import re

LOGGER = logging.getLogger("arcweb")

class ExpressionValidator(object):
  def __init__(self, val_type, dataset=None, conn_type=None):
    self.val_type = val_type
    self.dataset = dataset
    self.conn_type = conn_type
    self.properties = dataconn.dataconn.get_dataconn_type_properties(
        connection_type=self.conn_type)

  def build(self, ds_req):
    return "%s %s" % (self.val_type, ds_req.expr)

  def expr_filter(self, expr, dataset, limit=0, derived_data=None, expr_alias=None, add_expr=False):
    return self.expr_header(expr, dataset, 'filter', limit=limit, derived_data=derived_data,
        expr_alias=expr_alias, add_expr=False, check_filter=True)

  def expr_header(self, expr, dataset, type, limit=None, derived_data=None, expr_alias=None, add_expr=False,
      check_filter=False):

    if limit is None:
      if self.properties.supports_limit_0:
        limit = 0
      else:
        limit = 1

    if type == 'query':
      if self.conn_type == DATACONN_ORACLE:
        expr = 'SELECT * FROM (%s) T1 WHERE ROWNUM <= 1' % expr
      elif self.conn_type == DATACONN_MSSQL:
        expr = 'SELECT TOP(1) * FROM (%s) T1' % expr
      else:
        if self.properties.supports_limit_0:
          limit_str = "LIMIT 0"
        else:
          limit_str = "LIMIT 1"

        expr = re.sub(r'\bLIMIT\b\s*\d+', limit_str, expr, 0, re.IGNORECASE)
        if re.search(r'\blimit\b', expr, re.IGNORECASE) is None:
          expr = expr + ' ' + limit_str

      return expr, [], ['', '', '']

    if self.conn_type is not None and self.conn_type in [DATACONN_HIVE, DATACONN_HYVE, DATACONN_DRILL]:
      outer = HiveSelectBuilder(dataset, properties=self.properties)
    elif self.conn_type is not None and self.conn_type == DATACONN_ORACLE:
      outer = OracleSelectBuilder(dataset, properties=self.properties)
    elif self.conn_type is not None and self.conn_type == DATACONN_CASSANDRA:
      outer = CassandraSelectBuilder(dataset, properties=self.properties)
    else:
      outer = BaseSelectBuilder(dataset, properties=self.properties)

    inner = outer.clone()
    inner.limit = limit
    inner.hidden_columns = False
    inner.dataset.addAllToSb(inner)
    alias = True
    if derived_data is not None:
      for d in derived_data:
        view = outer.createView(view_label=d.label, view_dsreq=d.dsreq)
        if view is not None:
          outer.derived_data[view.table.alias] = view
        else:
          alias = False
    outer.ignore_alias = True
    if check_filter:
      outer.addAggregate('*')
      outer.limit = 1
      outer.addFilter(expr, addToColumns=False)
    elif type == 'column':
      outer.addAggregate(expr, alias=alias)

    # Ensure the expression doesn't contain a circular reference or other error
    outer.dataset.process_expr(expr)

    view_sql = list()
    if outer.dataset.views:
      for view in list(outer.dataset.views.values()):
        inner.aggregates.append(view.table.alias + '.' +
                                inner.dataset.enquote(view.label)
                                + ' as '
                                + inner.dataset.enquote(view.label))
        inner.tables.append(view.table)
        view_sql.append(view.sql)

    processed_expr = inner.dataset.process_expr(expr)

    if add_expr:
      inner.addAggregate(processed_expr, alias=alias)

    with_sql = ("WITH " if len(view_sql) else "") + ',\n'.join(view_sql)
    outer_sql = outer.getsql()
    outer_select = outer_sql.split('FROM')[0] + ' FROM '
    inner_query = '(' + inner.getsql() + ') TA_0'
    if not self.properties.supports_subselect_alias:
      # no alias for subselect:
      inner_query = '(' + inner.getsql() + ')'

    if expr_alias:
      inner_query = inner_query.replace('[{0}]'.format(expr_alias), '({0})'.format(processed_expr))

    header_query = with_sql + '\n' + outer_select + inner_query

    if type == 'filter':
      outer_where = ' WHERE ' + outer_sql.split('WHERE')[1]
      header_query = header_query + outer_where

    return header_query, outer.columns, [with_sql, inner_query, outer_select]

  # Generates query that will cause specific error that indicates the query is an aggregate
  # At the same time, test that query uses only columns that exist in table,
  # Not columns defined in frontend
  def expr_hasAggregate(self, expr, dataset, derived_data=None, expr_alias=None):
    with_sql = str(self.expr_header(expr, dataset, 'column', expr_alias=expr_alias)[2][0])
    if self.conn_type == DATACONN_DRILL:
      inner_query = str(self.expr_header(expr, dataset, 'column', derived_data=derived_data, limit=0, expr_alias=expr_alias,
                                         add_expr=True)[2][1])
    elif self.conn_type in [DATACONN_PG, DATACONN_REDSHIFT, DATACONN_SNOWFLAKE]:
      inner_query = str(self.expr_header(expr, dataset, 'column', derived_data=derived_data, expr_alias=expr_alias)[2][1])
    else:
      # Test that columns exist
      inner_query = str(self.expr_header(expr, dataset, 'column', derived_data=derived_data, expr_alias=expr_alias,
                                         add_expr=True)[2][1])
    outer_query = str(self.expr_header(expr, dataset, 'column', derived_data=derived_data, expr_alias=expr_alias,
                                       add_expr=True)[2][2])
    if self.conn_type == DATACONN_DRILL:
      return with_sql + '\n' + outer_query + inner_query

    gby_expr = ''
    if self.conn_type in [DATACONN_HIVE, DATACONN_IMPALA, DATACONN_ARCENGINE, \
        DATACONN_SPARKSQL, DATACONN_IMPYLA, DATACONN_HYVE, DATACONN_ARCENGINE_V2]:
      gby_expr = '1, 2'
      if with_sql:
        with_sql = 'EXPLAIN ' + with_sql
        outer_query = outer_query.replace('SELECT', 'SELECT 1,', 1)
      else:
        outer_query = outer_query.replace('SELECT', 'EXPLAIN SELECT 1,', 1)

    # Apache Hive 1.1.0-cdh5.13.1 and SparkSQL pre-2.3:
    #   the outer group by index does not work well
    #   and returns error complaining about column not being in group by
    #   Thus, add hack to get the expression and use in group by
    # Alternatively, could do dataset.process_expr(expr),
    #   but that'll give inner expression rather than alias
    if self.conn_type in [DATACONN_SPARKSQL, DATACONN_HIVE, DATACONN_HYVE]:
      first_column = re.search(r"SELECT 1,(.*)", outer_query)
      gby_expr = first_column.group(1)
    elif self.properties.supports_groupby_colindex:
      gby_expr = gby_expr or '1'
    else:
      gby_expr = expr

    return with_sql + '\n' + outer_query + inner_query + ' GROUP BY ' + gby_expr

  def addFilterColumns(self, columns):
    pass

  def addFilter(self, expr, addToColumns=True):
    pass
