# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

"""
Provides an async workq with status reporting

AsyncRunner instances are created with a thread count that defines the max
parallelism.

Work is queued up by calling add_work with an object that implements the
interface in ProgressRunner. A handle is returned and can be used to check
for status later on.

In the context of async db access, we use the following procedure:
. When this process starts up, we create an AsyncRunner object for each
  data connection in play. Per-connection parallelism is setup based on
  settings.
. As SQL requests come in, we call add_work with a SqlProgressRunner object.
  A handle ID is returned back to the caller.
. The handle is used later to get progress and results.
"""

import threading
from collections import deque
import time
import logging
import traceback
import json

from arcwebbase import access_info
from util.perfcounters import APPROX_COUNTERS
import vizlib.util
from vizlib.arcprof import ArcProf

from django.db import close_old_connections
from django.conf import settings

LOGGER = logging.getLogger('arcweb')

WORKQ_ERROR_COUNTER = APPROX_COUNTERS.get('workq.error')
WORKQ_MARK_INTERRUPTED = APPROX_COUNTERS.get('workq.markInterrupted')

HANDLE_ID_ALLOCATOR = vizlib.util.IdAllocator()

class RequestData:
  def __init__(self, startTime: float, runner: 'Runner'):
    self._startTime = startTime
    self._runner = runner

  def startTime(self):
    return self._startTime

  def runner(self):
    return self._runner

  def __str__(self):
    return json.dumps({
      'startTime': self._startTime,
      'runner': self._runner.toDisplayDict()
    })

  def __repr__(self):
    return json.dumps({
      'startTime': self._startTime,
      'runner': self._runner.toDisplayDict()
    })

class Runner:
  """
  A runner object to add to a threadpool work queue
  """

  QUEUED = 'Queued'
  RUNNING = 'Running'
  DONE = 'Finished'

  def __init__(self, progress_runner: 'ProgressRunner', queryTimeoutSeconds=None):
    self.__progress_runner = progress_runner
    self.set_status(Runner.QUEUED)
    self.__results = None
    self.__exception = None
    self.__cancelled = False
    self.__startTime = time.time()
    self.__queryTimeoutSeconds = queryTimeoutSeconds
    self.__access_info = None

  def setAccessInfo(self, access_info):
    self.__access_info = access_info

  def progress_runner(self):
    return self.__progress_runner

  def toDisplayDict(self):
    return self.__progress_runner.toDisplayDict()

  def __str__(self):
    return str(self.__progress_runner)

  def __repr__(self):
    return repr(self.__progress_runner)

  def startTime(self):
    return self.__startTime

  def queryTimeoutSeconds(self):
    return self.__queryTimeoutSeconds

  def admissionControlCategory(self):
    return self.__progress_runner.admissionControlCategory()

  def interrupt(self):
    try:
      self.__progress_runner.interrupt()
    except Exception as e:
      LOGGER.exception("Exception from call to interrupt")

  def markAsInterrupted(self):
    self.__exception = Exception("Connection timeout triggered - not running query")
    self.set_status(Runner.DONE)

  def markAsCancelled(self):
    self.__exception = Exception("Cancelled by user")
    self.set_status(Runner.DONE)

  def execute_runner(self):
    """
    Run the function associated with this object
    """
    global WORKQ_ERROR_COUNTER

    close_old_connections()
    self.set_status(Runner.RUNNING)
    try:
      access_info.store(self.__access_info)
      self.__results = self.__progress_runner.run()
    except Exception as e:
      WORKQ_ERROR_COUNTER += 1
      self.__exception = e
      LOGGER.exception("Exception from workq task")

    access_info.clear()
    self.set_status(Runner.DONE)
    close_old_connections()

  def profile(self):
    """Accessor"""
    try:
      return self.__progress_runner.profile()
    except:
      return None

  def results(self):
    """Accessor"""
    return self.__progress_runner.results()

  def exception(self):
    """Accessor"""
    return self.__exception

  def status(self) -> str:
    return self.__status

  def set_status(self, status: str):
    self.__status = status

  def progress(self):
    return self.__progress_runner.progress()

  def isCancelled(self):
    return self.__cancelled

  def markCancelled(self):
    if not self.__cancelled:
      self.__cancelled = True
      self.interrupt()
      self.set_status(Runner.DONE)

# pylint: disable=R0924
class _WorkQ:
  """
  A threadsafe work queue object. Callers add work items to this object using
  add_work. During threaded execution, get_work gets the next item of work to
  execute
  """

  def __init__(self, name=None, capacity=None, categoryConcurrency=None):
    self._name = name
    self._categoryConcurrency = categoryConcurrency

    self.__work_q = deque()
    self.lock = threading.Lock()
    self.cv = threading.Condition(self.lock)
    self.hist = APPROX_COUNTERS.getHistogram(name=f"{name}.workq",
        buckets=[1, 5, 10, 30, 60, 120, 180, 240, 300, 301])
    self.canceled = False

    self._categoryConcurrencyTracker = dict[str, int]()
    self._categoryLock = threading.Lock()

    self._pendingWorkLock = threading.Lock()
    self._pendingWork = dict[str,list[RequestData]]()

    self._workqPending = APPROX_COUNTERS.get(f'{name}.workq.pending')
    self._workqUserPending = APPROX_COUNTERS.get(f'{name}.workq.userpending')
    self._workqUserSleep = APPROX_COUNTERS.get(f'{name}.workq.usersleep')
    self._workqUserWakeup = APPROX_COUNTERS.get(f'{name}.workq.userwakeup')
    self._workqStats = vizlib.util.RequestStatsTracker(f'{name}.workq',
        capacity=capacity, perUserCapacity=categoryConcurrency)

  def workqStats(self):
    return self._workqStats

  def _lockedIteratorPendingWork(self, category:str=None):
    if not category:
      for category in self._pendingWork.keys():
        for pendingItem in self._pendingWork[category]:
          yield (pendingItem, category)

      return

    for pendingItem in self._pendingWork.get(category, []):
      yield (pendingItem, category)

  def requePendingWork(self, category=None, requeCount=None):
    # Put some queries back on the workq to be evaluated
    with self._pendingWorkLock:
      if category and not self._pendingWork.get(category):
        return

      newPendingWork = dict[str,list[RequestData]]()
      count = 0
      for requestData, rCategory in self._lockedIteratorPendingWork(category=category):
        if requeCount and count >= requeCount:
          if not newPendingWork.get(rCategory):
            newPendingWork[rCategory] =[]

          newPendingWork[rCategory].append(requestData)
          continue

        delayedRunner = requestData.runner()
        startTime = delayedRunner.startTime()
        if not category or category == delayedRunner.admissionControlCategory():
          self._workqUserWakeup += 1

          self._workqStats.exitAdmissionControl(requestData, delayedRunner.profile(),
              startTime)

          self.add_work(delayedRunner)
          count += 1
          self._workqUserPending.value -= 1
        else:
          if newPendingWork.get(rCategory):
            newPendingWork[rCategory].append(requestData)
          else:
            newPendingWork[rCategory] = [requestData]

      if category:
        self._pendingWork[category] = newPendingWork.get(category, [])
      else:
        self._pendingWork = newPendingWork

  def setCategoryConcurrency(self, categoryConcurrency):
    self._categoryConcurrency = categoryConcurrency
    self.requePendingWork()

  def name(self):
    return self._name

  def markWorkStartedOrMakePending(self, requestData: RequestData):
    runner = requestData.runner()
    startTime = requestData.startTime()

    start = True

    try:
      if not self._categoryConcurrency:
        return start

      category = runner.admissionControlCategory()
      if not category:
        return start

      with self._pendingWorkLock:
        with self._categoryLock:
          cnt = self._categoryConcurrencyTracker.get(category, 0)
          if cnt < self._categoryConcurrency:
            self._categoryConcurrencyTracker[category] = cnt + 1
            return start

        start = False
        if not self._pendingWork.get(category):
          self._pendingWork[category] = []

        self._pendingWork[category].append(requestData)

        self._workqStats.enterAdmissionControl(requestData,
            runner.profile(), startTime)

        self._workqUserPending += 1
        self._workqUserSleep += 1

    finally:
      if start:
        self.statsStart(requestData)

    return start

  def markWorkEnded(self, requestData: RequestData, runtime):
    b = self.hist.getBucket(runtime)
    self.hist.addToBucket(b)

    runner = requestData.runner()
    self.statsStop(requestData)

    if not self._categoryConcurrency:
      return

    category = runner.admissionControlCategory()
    if not category:
      return

    with self._categoryLock:
      self._categoryConcurrencyTracker[category] -= 1

    self.requePendingWork(category=category, requeCount=1)

  def statsStart(self, requestData: RequestData):
    self._workqStats.startRequest(requestData, requestData.runner().profile(),
                                  requestData.startTime())

  def statsStop(self, requestData: RequestData):
    startTime = requestData.startTime()
    self._workqStats.stopRequest(requestData, requestData.runner().profile(),
        startTime, time.time() - startTime)

  def cancel(self):
    with self.lock:
      self.canceled = True
      self.cv.notifyAll()

  # Note that we are not using self.cv for this
  # self.cv is used for a producer-consumer queue and having non-producers
  #   wait would mean that we need to use .notifyAll() instead of .notify()
  #   and that is way more inefficient
  def wait_for_empty(self):
    self.lock.acquire()
    while len(self.__work_q):
      self.lock.release()
      time.sleep(1)
      self.lock.acquire()

  def get_work(self) -> RequestData:
    """
    Blocking function to get the next work element. Returns None if this workq
    has been terminated
    """

    with self.lock:
      while not len(self.__work_q):
        if self.canceled:
          return None
        self.cv.wait()

      if self.canceled:
        return None

      requestData: RequestData = self.__work_q.popleft()
      runner = None
      if requestData is None:
        LOGGER.info("workq: empty requestdata")
      else:
        self._workqPending -= 1
        runner = requestData.runner()
        self._workqStats.exitAdmissionControl(requestData,
            runner.profile(), runner.startTime())

      return requestData

  def createRequestData(self, runner: 'Runner'):
    startTime = runner.startTime()
    return RequestData(startTime, runner)

  def add_work(self, runner: 'Runner'):
    """Add work to the workq in a threadsafe way"""
    requestData = self.createRequestData(runner)

    self._workqStats.enterAdmissionControl(requestData,
        runner.profile(), requestData.startTime())

    with self.lock:
      self.__work_q.append(requestData)
      self._workqPending += 1
      self.cv.notify()

  def __len__(self):
    """Return the number of elements in this workq"""
    with self.lock:
      return len(self.__work_q)

class _WorkQThread(threading.Thread):
  """
  Thread-derived object for running tasks in parallel
  """

  def __init__(self, asyncRunner: 'AsyncRunner', name=None):
    threading.Thread.__init__(self, name=name)
    self.daemon = True
    self._asyncRunner = asyncRunner
    self.startTime = None
    self._cancelled = False
    self._runner = None

    workqName = self._asyncRunner.workq().name()

  def activeStartTime(self):
    if self._cancelled:
      return None

    return self.startTime

  def cancelThread(self):
    self._cancelled = True

  def interruptIfTimedOut(self, queryTimeoutSeconds=None):
    s = self.startTime
    if queryTimeoutSeconds and s and ((time.time() - s) > queryTimeoutSeconds):
      LOGGER.info("Canceling runner")
      self._runner.interrupt()

  def run(self):
    """Function runner for this thread object"""

    workq = self._asyncRunner.workq()

    while True:
      if self._cancelled:
        break

      requestData = workq.get_work()
      if requestData == None:
        return

      startTime = requestData.startTime()
      self._runner = requestData.runner()

      t1 = time.time()
      runit = workq.markWorkStartedOrMakePending(requestData)
      if not runit:
        continue

      try:
        if self._runner.isCancelled():
          self._runner.markAsCancelled()
          continue

        if self._runner.queryTimeoutSeconds() and (t1 - startTime) > self._runner.queryTimeoutSeconds():
          global WORKQ_MARK_INTERRUPTED
          WORKQ_MARK_INTERRUPTED += 1
          self._runner.markAsInterrupted()
          continue

        self.startTime = startTime
        self._runner.execute_runner()
      finally:
        workq.markWorkEnded(requestData, time.time() - t1)
        self.startTime = None
        self._runner = None

QUERY_CLEANUP_CREATES = APPROX_COUNTERS.get('workq.querycleanup.create')
QUERY_CLEANUP_CANCELS = APPROX_COUNTERS.get('workq.querycleanup.cancels')

DEFAULT_CLEANUP_DELAY = 10

CLEANUP_QUERY_TRACKERS_GRACE_PERIOD_SECONDS = settings.CLEANUP_QUERY_TRACKERS_GRACE_PERIOD_SECONDS
CLEANUP_QUERY_TRACKERS=settings.CLEANUP_QUERY_TRACKERS

class _QueryCleanupThread:
  def __init__(self, asyncRunner: 'AsyncRunner', queryTimeoutSeconds, delay=None):
    if not delay:
      delay = DEFAULT_CLEANUP_DELAY

    self._asyncRunner = asyncRunner
    self._queryTimeoutSeconds = queryTimeoutSeconds
    self._delay = delay
    self._canceled = False
    self._cleanupThread = threading.Thread(target=self.cleanupTimedOutThreads,
                                           name=f'QueryCleanup-{asyncRunner.name}')
    self._cleanupThread.daemon = True
    self._cleanupThread.start()

    global QUERY_CLEANUP_CREATES
    QUERY_CLEANUP_CREATES += 1

  def cancel(self):
    self._canceled = True

  def cleanupTimedOutThreads(self):
    while True:
      time.sleep(self._delay)

      if self._canceled:
        global QUERY_CLEANUP_CANCELS
        QUERY_CLEANUP_CANCELS += 1

        return

      toCancel = self._asyncRunner.getTimedOutThreads(
          queryTimeoutSeconds=self._queryTimeoutSeconds)

      for thread in toCancel:
        try:
          thread.interruptIfTimedOut(queryTimeoutSeconds=self._queryTimeoutSeconds)
        except:
          LOGGER.exception("Exception attempting to interrupt thread")

      self.cleanupTimedOutTrackers()

  def cleanupTimedOutTrackers(self):
    if not CLEANUP_QUERY_TRACKERS:
      return

    self._asyncRunner.cleanupTrackers()

class AsyncRunner:
  def __init__(self, threadcount=10, name=None, queryTimeoutSeconds=None,
      categoryConcurrency=None):

    self.name = name

    # Start with 0 so we create threads in the call to setParams below
    self._threadcount = 0
    self._queryTimeoutSeconds = 0

    self._workq = _WorkQ(name=name, capacity=threadcount,
        categoryConcurrency=categoryConcurrency)

    # Protect changes to self._threads and self._queryCleanupThread
    self._threadLock = threading.Lock()
    self._threads = list[_WorkQThread]()
    self._queryCleanupThread = None

    self._lock = threading.Lock()

    self._workqTracked = APPROX_COUNTERS.get(f'{name}.workq.tracked')

    self.tracker = dict[str,Runner]()

    self.setParams(threadcount=threadcount, queryTimeoutSeconds=queryTimeoutSeconds,
        categoryConcurrency=categoryConcurrency)

  def queryTimeoutSeconds(self):
    return self._queryTimeoutSeconds

  def getTimedOutThreads(self, queryTimeoutSeconds=None):
    if not queryTimeoutSeconds:
      return []

    threads = list[_WorkQThread]()

    with self._threadLock:
      for t in self._threads:
        s = t.activeStartTime()
        if not s:
          continue

        if (time.time() - s) > queryTimeoutSeconds:
          threads.append(t)

    return threads

  def setParams(self, threadcount=10, queryTimeoutSeconds=None, categoryConcurrency=None):
    self._workq.setCategoryConcurrency(categoryConcurrency)
    self._workq.workqStats().updateCapacity(capacity=threadcount, perUserCapacity=categoryConcurrency)

    APPROX_COUNTERS.get(f'{self.name}.workq.capacity').value = threadcount
    if categoryConcurrency:
      APPROX_COUNTERS.get(f'{self.name}.workq.capacity-per-user').value = categoryConcurrency

    with self._threadLock:
      if queryTimeoutSeconds and self._queryTimeoutSeconds != queryTimeoutSeconds:
        if self._queryCleanupThread:
          self._queryCleanupThread.cancel()

        self._queryCleanupThread = _QueryCleanupThread(self, queryTimeoutSeconds)
        self._queryTimeoutSeconds = queryTimeoutSeconds

      if threadcount > self._threadcount:
        if self.name:
          name = f"Thread-{self.name}"
        else:
          name = "Thread-Async"

        for i in range(0, threadcount - self._threadcount):
          t = _WorkQThread(self, name=f"{name}-{self._threadcount+i}")
          self._threads.append(t)
          t.start()

      elif threadcount < self._threadcount:
        toCancel = self._threads[threadcount:]
        self._threads = self._threads[:threadcount]

        for t in toCancel:
          t.cancelThread()

      self._threadcount = threadcount

  def workq(self):
    return self._workq

  def requePendingWork(self, category=None, requeCount=None):
    self._workq.requePendingWork(category=category, requeCount=requeCount)

  def queue_untracked(self, progress_runner: 'ProgressRunner') -> Runner:
    """
    Submit an untracked run for tasks that do not use the tracker for reporting status,
    like emailing and snapshots. The returned Runner can be used to monitor the status and
    retrieve the results, if needed.
    """
    runner = Runner(progress_runner, queryTimeoutSeconds=self._queryTimeoutSeconds)
    runner.setAccessInfo(access_info.get())
    self._workq.add_work(runner)
    return runner

  def queue(self, progress_runner: 'ProgressRunner') -> str:
    runner = self.queue_untracked(progress_runner)
    handle = HANDLE_ID_ALLOCATOR.allocateId()
    self.start_tracking(handle, runner)
    return handle

  def start_tracking(self, handle, runner: Runner):
    with self._lock:
      self.tracker[handle] = runner
      self._workqTracked += 1

  def stop_tracking(self, handle):
    with self._lock:
      if handle in self.tracker:
        del self.tracker[handle]
        self._workqTracked -= 1

  def wait_for_empty(self):
    self._workq.wait_for_empty()

  def wait_for_done(self):
    while True:
      done = True
      for t in self._threads:
        if t.is_alive():
          done = False
          break
      if done:
        return
      time.sleep(0.1)

  def cancel(self):
    self._workq.cancel()

  def getRunner(self, handle):
    with self._lock:
      return self.tracker.get(handle)

  def results(self, handle):
    with self._lock:
      r = self.tracker.get(handle)

    if not r:
      return None

    return r.results()

  def status(self, handle):
    with self._lock:
      r = self.tracker.get(handle)

    if not r:
      return None

    return r.status()

  def progress(self, handle):
    with self._lock:
      r = self.tracker.get(handle)

    if not r:
      return None

    return r.progress()

  def cancelQueryViaQueryTracker(self, matchUsername=None, querytracker=None):
    runner = None

    toCancel = []

    with self._lock:
      for runner in self.tracker.values():
        if not querytracker or querytracker == runner.profile().querytracker():
          toCancel.append(runner)

    if not toCancel:
      return 0

    cnt = 0
    for runner in toCancel:
      if matchUsername:
        if runner.profile().username() != matchUsername:

          # Log if querytracker was specified by username does not match
          if querytracker:
            LOGGER.info("Cancelation attempted, but username is mismatched (%s, %s)",
                runner.profile().username(), matchUsername)

          continue

      cnt += 1
      runner.markCancelled()

    return cnt

  def cleanupTrackers(self):

    with self._lock:
      now = time.time()
      expiredHandles = list[str]()
      for handle, runner in self.tracker.items():
        timeout_seconds = runner.queryTimeoutSeconds() + CLEANUP_QUERY_TRACKERS_GRACE_PERIOD_SECONDS
        runTime = now - runner.startTime()
        if runTime > timeout_seconds:
          expiredHandles.append(handle)

    for handle in expiredHandles:
      self.stop_tracking(handle)

  def markCancelled(self, handle):
    with self._lock:
      r = self.tracker.get(handle)

    if not r:
      LOGGER.info("Nothing to cancel: %s", handle)
      return False

    r.markCancelled()
    return True

class ProgressRunner:
  """
  An interface that knows how to run a long-running task and report back on
  progress.
  """
  def run(self):
    """Start running the task"""
    pass

  def progress(self):
    """Returns a tuple that indicates (steps_completed, total_steps)"""
    pass

  def results(self):
    """Returns results of the function"""
    pass

  def interrupt(self):
    """Interrupts & cancels long-running work"""
    LOGGER.info("interrupt called but is not implemented")
    pass

  def admissionControlCategory(self) -> str:
    """
    If these objects get queued up into a workq that tracks max concurrency
    by category, subclasses can return the category for work by implementing
    this method.
    """
    return None

  def toDisplayDict(self):
    """
    Used to display information about this class - this is expected to be
    over-ridden
    """
    return {
      'class': str(type(self))
    }

  def profile(self):
    """
    Return a profile object used to log progress. Subclasses of this object
    are expected to override this implementation.
    """
    return None

class JobProgressRunner(ProgressRunner):
  def __init__(self, job, username):
    self.__job = job
    self.__profile = ArcProf("JobProgressRunner", username=username)

  def run(self):
    return self.__job.run()

  def progress(self):
    return self.__job.progress()

  def results(self):
    return self.__job.results()

  def interrupt(self):
    LOGGER.info("interrupt called but is not implemented")

  def admissionControlCategory(self):
    return None

  def toDisplayDict(self):
    return {
      'class': str(type(self.__job))
    }

  def profile(self):
    return self.__profile

