# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

# Models for workspaces

from django.db import models
from django.conf import settings

from django.db.models.signals import post_save, post_delete, pre_delete
from django.dispatch import receiver
from django.contrib.auth.models import User

import json
import logging

from rbac.permdefs import PermissionException
from arcwebbase.validator import validateNoHtml, validateStringNoHtml

# Sync with workspaceInfo in arc_workspace.js
class WSACL_Member:
  USER = 1
  GROUP = 2

class WSACL_Level:
  VIEW = 1
  EDIT = 2
  MANAGE = 3

# Validate one acl entry
def validate_acl_entry(a):
  if type(a) != type([]) or len(a) != 3:
    raise Exception("Invalid workspace ACL entry format")
  if a[0] not in [WSACL_Member.USER, WSACL_Member.GROUP]:
    raise Exception("Invalid workspace ACL entry member type")
  if a[1] not in [WSACL_Level.VIEW, WSACL_Level.EDIT, WSACL_Level.MANAGE]:
    raise Exception("Invalid workspace ACL entry level type")
  if not a[2] or type(a[2]) not in [type(''), type('')]:
    raise Exception("Invalid workspace ACL entry user/group name")

# Validate acl object
def validate_acl_obj(acl):
  if type(acl) != type([]):
    raise Exception("Invalid workspace ACL format")
  for a in acl:
    validate_acl_entry(a)

# Validate acl json
def validate_acl_json(acl_json):
  if type(acl_json) not in [type(''), type('')]:
    raise Exception("Invalid workspace ACL format")
  try:
    acl = json.loads(acl_json)
  except Exception as e:
    raise Exception("Invalid workspace ACL format")
  validate_acl_obj(acl)

# Model for workspace
class Workspace(models.Model):
  """Workspace model"""

  def save(self, *args, **kwargs):
    validateNoHtml(self)
    super(Workspace, self).save(*args, **kwargs)

  def delete(self, *args, **kwargs):
    if not self.editable:
      raise Exception("Can not delete the public or private workspace")
    super(Workspace, self).delete(*args, **kwargs)

  # name
  name = models.CharField(max_length=1000, null=True, blank=True,
      validators=[validateStringNoHtml])

  # description
  description = models.CharField(max_length=1000, null=True, blank=True,
      validators=[validateStringNoHtml])

  # Is false for public and private workspaces
  editable = models.BooleanField(default=True)

  # User when this is a system created private workspace for a user
  private_user = models.OneToOneField(User, null=True, blank=True, on_delete=models.SET_NULL)

  # json formatted ACL
  # TODO -- Validate the acl
  acl = models.TextField(max_length=32*1024, null=True, blank=True, default='[]',
      validators=[validate_acl_json])

  # Create time, user
  create_time = models.DateTimeField(auto_now_add=True, null=True, blank=True)
  create_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  # Update time, user
  update_time = models.DateTimeField(auto_now=True, null=True, blank=True)
  update_user = models.CharField(max_length=200, blank=True,
      validators=[validateStringNoHtml])

  @property
  def acl_obj(self):
    return json.loads(self.acl)

  @acl_obj.setter
  def acl_obj(self, data):
    self.acl = json.dumps(data)

  # Get the permissions for the specified user on this workspace
  # Note: only uses workspace definition, doesn't take into account if user is superuser
  #       Used for getPerms => get_user_obj_perms
  def perms_for_user(self, user, groups=None):
    name = user.username
    if groups is None:
      from rbac.permcache import groups_for_user
      groups = groups_for_user(user)

    # Organize the acl into sets for each (type, level)
    pmap = {}
    for a in self.acl_obj:
      key = (a[0], a[1])
      pmap[key] = pmap.get(key, set()) | set([a[2]])

    # Helper function to check user/group present at the current level
    def check_level(level):
      lvl_names = pmap.get((WSACL_Member.USER, level), set())
      lvl_groups = pmap.get((WSACL_Member.GROUP, level), set())
      return name in lvl_names or \
             settings.EVERYONE_GROUP in lvl_groups or \
             groups.intersection(lvl_groups)

    # Return permissions based on first level the user shows up in
    if check_level(WSACL_Level.MANAGE):
      return set(['ws_manage', 'ws_appedit', 'ws_appview'])
    if check_level(WSACL_Level.EDIT):
      return set(['ws_appedit', 'ws_appview'])
    if check_level(WSACL_Level.VIEW):
      return set(['ws_appview'])
    return set()

  # Permission type for objects of this class
  @staticmethod
  def perm_type():
    return 'workspace'

  def is_public(self):
    return self.name == 'Public'

  # Check if user has given permission on workspace
  def checkPerm(self, user, perm):
    from rbac.permcache import RBAC
    RBAC.checkPerm(user, perm, self)

  # Get all permissions a user has on this workspace
  def getPerms(self, user):
    from rbac.permcache import RBAC
    return RBAC.get_user_obj_perms(user, self)

  # Remove permissions from input set based on workspace permissions
  def maskPerms(self, user, input):
    myperms = self.getPerms(user)
    if 'ws_appedit' in myperms:
      mask = set()
    elif 'ws_appview' in myperms:
      mask = set(['ds_appedit'])
    else:
      mask = set(['ds_appedit', 'ds_appview'])

    return input.difference(mask)

# Build description text for user's private ws
def private_ws_description(id, username):
  return '%s private workspace' % username

# Build the private workspace values for a user
def private_ws_defaults(user):
  return {
    'name':'Private',
    'editable':False,
    'private_user_id':user.id,
    'description':private_ws_description(user.id, user.username),
    'acl':json.dumps([[WSACL_Member.USER, WSACL_Level.EDIT, user.username]])}

# Get the private workspace for a user, create if needed
def get_or_create_private_ws(user):
  # First, try and return the workspace with the right id
  try:
    return Workspace.objects.get(private_user_id=user.id)
  except Workspace.DoesNotExist:
    pass

  # Try and create a new one. Note that multiple threads/processes may be trying this
  # and this may fail
  try:
    (ws, created) = Workspace.objects.get_or_create(private_user_id=user.id,
                                                    defaults=private_ws_defaults(user))
    return ws
  except:
    pass

  # Try one more time
  try:
    (ws, created) = Workspace.objects.get_or_create(private_user_id=user.id,
                                                    defaults=private_ws_defaults(user))
    return ws
  except:
    pass

  # Final attempt at returning the created object
  return Workspace.objects.get(private_user_id=user.id)

def get_default_workspace(user, ws_perm='ws_appedit'):
  LOGGER = logging.getLogger("arcweb")
  DEFAULT_WORKSPACE = settings.DEFAULT_WORKSPACE

  if settings.DEFAULT_WORKSPACE:
    # First, try and return the workspace with the right id
    try:
      workspace = Workspace.objects.get(id=DEFAULT_WORKSPACE)
    except:
      LOGGER.warning(f'No workspace with id: {DEFAULT_WORKSPACE}')

    # Then, try and return the workspace with the right name
    try:
      workspace = Workspace.objects.get(name=DEFAULT_WORKSPACE)
    except:
      LOGGER.warning(f'No workspace with name: {DEFAULT_WORKSPACE}')

    # Now check that the user has the specified permission on the workspace
    try:
      workspace.checkPerm(user, ws_perm)
      return workspace
    except PermissionException:
      LOGGER.warning(f'No premission for user: {user} to the workspace: {workspace.name}')

  return get_or_create_private_ws(user)

# Update a user's private workspace when the user has been deleted.
def workspace_user_delete(id, username):
  try:
    ws = Workspace.objects.get(private_user_id=id)
    ws.editable = True
    ws.private_user_id = None
    ws.description= '%s (deleted) private workspace (User ID=%s)' % (username, id)
    ws.name = '%s private workspace' % (username)
    ws.acl_obj = []
    ws.save()
  except Workspace.DoesNotExist:
    pass

# Update a user's private workspace if the user has been renamed.
def workspace_user_update(id, username):
  try:
    ws = Workspace.objects.get(private_user_id=id)
    desc = private_ws_description(id, username)
    if ws.description != desc:
      ws.description= desc
      ws.save()
  except:
    pass

# Reset permission cache on any workspace update
@receiver(post_save, sender=Workspace)
def workspace_post_save(sender, instance, **kwargs):
  if isinstance(instance, Workspace):
    from rbac.permcache import RBAC
    RBAC.reset()

@receiver(post_delete, sender=Workspace)
def workspace_post_delete(sender, instance, **kwargs):
  if isinstance(instance, Workspace):
    from rbac.permcache import RBAC
    RBAC.reset()

# Fixup user's private workspace in case user name changes
@receiver(post_save, sender=User)
def workspace_post_user_save(sender, instance, **kwargs):
  if isinstance(instance, User):
    workspace_user_update(instance.id, instance.username)

# Fixup user's private workspace when user is deleted
@receiver(pre_delete, sender=User)
def workspace_pre_user_delete(sender, instance, **kwargs):
  if isinstance(instance, User):
    workspace_user_delete(instance.id, instance.username)

