# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#  
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#  
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

# Workspace url support

from django.http import HttpResponse

from django.contrib.auth.decorators import login_required
from django.db.models import Q


from workspace.models import Workspace, validate_acl_obj, private_ws_defaults
from reports.models import Report
from rbac.permcache import RBAC

from django.views.decorators.http import etag
from arcwebbase.etag_caching import etag_metadata_func

import json

from vizlib.querylog import logAction, ComponentType, ActionType

def _auditWorkspace(ws, action, request):
  logAction(action=action, component_type=ComponentType.WORKSPACE,
      component_id=ws.id, request=request)

def update_private_name(ws, user, wsobj=None):
  wsobj = wsobj or {
    'desc': ws.description,
    'name': ws.name,
  }

  # For admin user getting access to a different user's private ws,
  # we swap the ws and description (which contains the user name).
  # This helps the admin user can find a particular user's private ws.
  # TODO: Maybe this is better done in the UI?
  if (user.is_superuser or ('sys_misc' in RBAC.get_user_obj_perms(user))) and \
      not ws.editable and \
      ws.name != 'Public' and \
      ws.private_user_id != user.id:
    wsobj['name'], wsobj['desc'] = wsobj['desc'], wsobj['name']

  return wsobj

@login_required
@etag(etag_metadata_func)
def workspace(request, workspace_id=None):
  """
  REST endpoint for workspace access
  """
  return workspace_internal(request, workspace_id)

def workspace_internal(request, workspace_id=None, filter=None):
  """
  Generic interface to front end for one workspace:
    {
      name: <particular workspace name>
      desc: <workspace description>
      id: <id for this workspace>
      acl: [<list of privileges for this workspace>]
    }

  The acl is a list of entries indicating (user/group, name, permission)
  """
  def validate_workspace(id):
    if not id:
      raise Exception("No workspace id specified")
    wslist = Workspace.objects.filter(id=int(id))
    if len(wslist) == 0:
      raise Exception("Workspace id not found")
    return wslist[0]

  if request.method == 'GET':
    if workspace_id is not None:
      wslist = [validate_workspace(workspace_id)]
    else:
      wslist = Workspace.objects.all()
      if filter:
        wslist = wslist.filter(**filter)
      wslist = list(wslist)

      # Create user now so "Private" workspace appears in "Move workspace" popover
      ws = next((ws for ws in wslist if ws.private_user_id == request.user.id), None)
      if not ws and not filter:
        wslist.append(Workspace(**private_ws_defaults(request.user)))

    result = []

    for ws in wslist:
      perms = ws.getPerms(request.user)

      if 'ws_appview' in perms or ('sys_misc' in RBAC.get_user_obj_perms(request.user)):
        wsobj = to_response_obj(ws, perms)

        # Get permissions requesting user has on this ws
        # But even admin users can not manage system managed workspaces
        if not ws.is_public() and not ws.editable and 'ws_manage' in perms:
          perms.remove('ws_manage')

        if not filter:
          update_private_name(ws, request.user, wsobj)

        result.append(wsobj)

    return HttpResponse(json.dumps(result))

  if request.method == 'PUT':
    return HttpResponse("Invalid request", status=500)

  if request.method == 'POST':
    rsp = {}
    data = json.loads(request.POST.get("data"))
    for wd in data:
      name = wd.get('name')
      if not name:
        return HttpResponse("No workspace name specified", status=500)
      if name.lower() == 'private':
        return HttpResponse("Workspace name '%s' reserved for system use" % name, status=500)
      wl = Workspace.objects.filter(name=name)
      if wd.get('id') is None:
        RBAC.checkPerm(request.user, 'sys_workspace')
        action_type = ActionType.CREATE
        if len(wl) > 0:
          return HttpResponse("Workspace name '%s' already in use" % name, status=500)
        workspace = Workspace()
        workspace.create_user = request.user.username
      else:
        action_type = ActionType.UPDATE
        workspace = Workspace.objects.get(id=wd['id'])
        if not workspace.editable and not workspace.is_public():
          return HttpResponse("Not permitted to change system managed workspace", status=500)
        if len(wl) > 1 or (len(wl) == 1 and wl[0].id != workspace.id):
          return HttpResponse("Workspace name '%s' already in use" % name, status=500)
        workspace.checkPerm(request.user, 'ws_manage')

      # Only allow non-public workspaces to have name and description edited
      if not workspace.is_public():
        workspace.name = name
        workspace.description = wd.get('desc', '')

      # Validate the acl
      acl = wd.get('acl', [])
      try:
        validate_acl_obj(acl)
      except Exception as e:
        return HttpResponse("Error updating workspace '%s':" % str(e), status=500)

      workspace.acl_obj = acl
      workspace.update_user = request.user.username
      workspace.save()
      _auditWorkspace(workspace, action_type, request)
      wd['id'] = workspace.id
      wd['editable'] = workspace.editable
      wd['private_user_id'] = workspace.private_user_id
      wd['create_time'] = str(workspace.create_time)
      wd['create_user'] = workspace.create_user
      wd['update_time'] = str(workspace.update_time)
      wd['update_user'] = workspace.update_user
      wd['perms'] = list(workspace.getPerms(request.user))
    rsp['status'] = "success"
    rsp['data'] = data
    return HttpResponse(json.dumps(rsp), content_type="application/json")

  if request.method == 'DELETE':
    if workspace_id:
      workspace = validate_workspace(workspace_id)
      workspace.checkPerm(request.user, 'ws_manage')
      rl = Report.objects.filter(workspace_id=workspace_id) \
                         .filter(Q(shared_visual_dashboards__isnull=False) | Q(report_type='dashboard'))
      if len(rl):
        return HttpResponse("Unable to remove non-empty workspace", status=500)
      _auditWorkspace(workspace, ActionType.DELETE, request)
      workspace.delete()
      return HttpResponse('{"status": "success"}', content_type="application/json")
    else:
      return HttpResponse("No workspace ID specified", status=500)

  return HttpResponse('Invalid request method', status=500)

def to_response_obj(workspace, perms):
  return {
          'id': workspace.id,
          'name': workspace.name,
          'desc': workspace.description,
          'acl' : workspace.acl_obj,
          'editable': workspace.editable,
          'private_user_id': workspace.private_user_id,
          'create_time': 'None' if workspace.create_time is None else workspace.create_time.isoformat(),
          'create_user': workspace.create_user,
          'update_time': 'None' if workspace.update_time is None else workspace.update_time.isoformat(),
          'update_user': workspace.update_user,
          'perms': list(perms)
         }
