# Frontend package
This package along with the arc package serves as the frontend component of our project.

On one side, it includes code from the arc package, which has been rewritten in React + TypeScript or, if that is not possible, in TypeScript only. On the other side, we've incorporating newer features into this repository.

# Table of contents
1. [Usage](#usage)
2. [Folder structure](#folder-structure)
3. [Built files in order of inclusion](#built-files-in-order-of-inclusion)
4. [Formating](#formating)
5. [FAQ](#faq)

## Usage
Run `scons --help` in your terminal and look for the Frontend commands in the output.

## Folder structure
```bash
┌── apis - "contains files dealing with HTTP requests"
│    └── get-datasets.ts
├── arc - "contains parts of the old codebase that was rewritten into Typescript"
├── assets - "contains static assets like images and fonts"
├── components - "contains React components"
│     └── component-name
│           ├── ComponentName.tsx
│           ├── ComponentName.test.tsx
│           └── ComponentName.module.scss
├── constants - "contains constant variables"
│     └── constant-one.ts
├── hooks - "contains React hooks"
│     └── use-hook.ts
├── pages - "contains code for a specific React page"
│    └── page
│          ├── apis - "api calls used only on this page"
│          ├── components - "components used only on this page"
│          ├── constants - "constants used only on this page"
│          ├── hooks - "hooks used only on this page"
│          ├── stores - "stores used only on this page"
│          ├── types - "types used only on this page"
│          ├── utils - "util functions used only on this page"
│          └── index.ts - "page's entry point and also exposed to global window object"
├── scss - "contains global scss"
│     └── scss-one.scss
├── services - "a service is an object/class with a narrow purpose, that is not related to the view"
│     └── service-name
│           └── index.ts - "service exposed to global window object"
├── stores - "contains stores that are dealing with global state management"
│     └── use-store.ts
├── types - "contains typescript types"
│     └── report.type.ts
└── utils - "contains utility functions"
      └── util-function.ts
```

## Built files in order of inclusion
![Built files in order of inclusion](readme-images/built-files-in-order-of-inclusion.png)
## Formating

Autoformatting for TypeScript and TypeScriptReact files is enabled by default via the `.vscode/settings.json`.
By default, it will format the code every time you save the file.
You will also have to install the recommended extensions, namely these two:

  1. Eslint - dbaeumer.vscode-eslint
  2. Prettier - esbenp.prettier-vscode

If, for some reason, you would like to disable it, you can override these setting by changing your VSCode user settings.

## FAQ
1. [Where should I put my code?](#where-should-i-put-my-code)
2. [How can I access my code in Django templates or in the arc package?](#how-can-i-access-my-code-in-django-templates-or-in-the-arc-package)
3. [How can I send events between the arc package and the frontend package?](#how-can-i-send-events-between-the-arc-package-and-the-frontend-package)
4. [How can I add a new React page?](#how-can-i-add-a-new-react-page)

### Where should I put my code?
```mermaid
  flowchart LR
    id1(Where should I put my code?) --> id2{Is it old code moved into typescript?}

    id2 -- Yes --> id3(frontend/src/arc)

    id2 -- No --> id4{Is it used only in one React page?}

    id4 -- Yes --> id5(frontend/src/pages/page-name)

    id4 -- No --> id6(frontend/src/component-type)
```

### How can I access my code in Django templates or in the arc package?
  You have to expose your code on the window object, using the `exposeToGlobal` function.

  Example:
  ```js
  import { exposeToGlobal } from 'utils/expose-to-global'

  exposeToGlobal(
    // The object to be merged into
    {
      myFunction
    },
    // the property of the window.arc object
    'frontend'
  );
  ```
  ```mermaid
    flowchart LR
      id1(How should I expose my code?) --> id2{Does it have to be loaded before everything else?}

      id2 -- Yes --> id3(frontend/webpack/entries/init.ts)

      id2 -- No --> id4{Is it a React page?}

      id4 -- Yes --> id5(frontend/src/pages/page-name/index.ts)

      id4 -- No --> id6{Is it only used in the frontend codebase?}

      id6 -- Yes --> id7(No need to expose it, just import it in the file you are using it)

      id6 -- No --> id8(frontend/webpack/entries/arc-dependency.ts)
  ```
  After that you can access your function as `window.arc.frontend.myFunction`.

### How can I send events between the arc package and the frontend package?
#### Frontend example:
```tsx
import { useCallback } from 'react';
import { legacyBridge, useLegacyBridge } from 'services/legacy-bridge';

...

// Listen and react to specific events (it is important to use useCallback)
const onEvent = useCallback(() => {
  // Do something when the event is dispatched
}, [])

useLegacyBridge('event-type', onEvent)
...

// Dispatch events
const onChange = () => {
  legacyBridge.events.next({ type: 'calculated-meaning-of-life', payload: { value: 42 } });
}
```

#### Arc example
```js
// Emit modal-closed event when the modal is hidden
$("#modal").on("hidden.bs.modal", () => {
  arc.frontend.legacyBridge.events.next({ type: 'modal-closed' });
});

...

const { events, ofType } = arc.frontend.legacyBridge;

// Listen to form-submitted events
const subscription = events
  .pipe(ofType('calculated-meaning-of-life'))
  .subscribe(
    // Get the payload of the event and do something with it when the event is dispatched
    ({ value }) => {
      $('#meaning-of-life-container').html(`<h1>${value}</h1`);
    }
  )

...

// Don't forget to unsubscribe from the subscription when you no longer need it
subscription.unsubscribe();
```

### How can I add a new React page?

You have to follow these steps:
<details>
  <summary>
    1. Create and expose your page's React component
  </summary>

  - `frontend/src/pages/page-name/PageName.tsx`
  - ```tsx
    import React from 'react';

    type PageNameProps = {
      propKey: string;
    };

    export const PageName = ({ propKey }: PageNameProps): JSX.Element => {
      return <h1>{propKey}</h1>
    }
    ```
  - `frontend/src/pages/page-name/index.ts`
  - ```ts
    import { exposeToGlobal } from 'utils/expose-to-global';
    import { PageName } from './PageName';

    exposeToGlobal({ pages: { PageName } }, 'frontend');
    ```
</details>

<details>
  <summary>2. Configure webpack to bundle your page</summary>

  - `frontend/webpack/webpack.common.ts`
  - ```ts
    ...
    export const commonWebpackConfiguration: Configuration = {
      entry: {
        ...
        // Pages
        'page-name': './src/pages/page-name/index.ts'
      },
      ...
    }
    ```
</details>

<details>
  <summary>3. Set up a Django endpoint to serve your page</summary>

  1. Create your page's template
  - `arcweb/apps/templates/page_name.html`
  - ```html
    {% extends "apps_skeleton_react.html" %}

    {% block load_script %}
      <link rel="stylesheet" type="text/css" href="{{ javascript_directory }}/frontend/page-name.css?v={{ static_version }}"></link>
      <script src="{{ javascript_directory }}/frontend/page-name.js?v={{ static_version }}"></script>
    {% endblock %}

    {% block on_load %}
      arc.frontend.renderReactComponent(
        'root',
        arc.frontend.pages.PageName,
        {
          propKey: "prop value",
        }
      )
    {% endblock %}
    ```
  2. Create your view
  - `arcweb/apps/views.py`
  - ```python
    from django.shortcuts import render
    from django.contrib.auth.decorators import login_required
    ...
    @login_required
    def page_name(request):
      context = { "key": "value" }

      return render(request, 'page_name.html', context)
    ```
  3. Create your url
  - `arcweb/apps/urls.py`
  - ```python
    from django.urls import re_path
    from apps import views

    urlpatterns = [
      re_path(r'^page_name$', views.page_name, name="page-name")
      ...
    ]
    ```
</details>


<details>
  <summary>4. Expose your page's url</summary>

  1. Pass your url from the backend to the frontend
- `arcweb/apps/templates/partial_frontend_globals.html`
  - ```html
      <script>
        arc.frontend.globals = {
          urls: {
            pageNameUrl: "{% url 'page-name' %}"
            ...
          }
          ...
        }
      </script>
    ```
  2. Update typescript about the changes
  - `frontend/src/services/globals/globals.type.ts`
  - ```ts
    export type Globals = {
      urls: {
        pageNameUrl: string;
        ...
      }
      ...
    }
    ```
</details>
