import { getGlobalArc } from 'utils/get-global-arc';
import { vizHttp } from 'services/viz-http';

import type {
  DataConnection,
  DataConnectionType,
} from 'types/data-connection.type';

export const dataconnectionQueryToken = 'dataconnection';

type BackendDataConnection = {
  id: number;
  name: string;

  type: string;
  typeName: string;

  created: string;
  created_by: string;
  updated: string;
  updated_by: string;

  info: string;
  description: string;

  cmlId: null | number;
  cmlStatus: string;

  isPasswordSet: boolean;
  supportsImportData: boolean;

  lastUsed: string;

  perms: string[];
};

const arc = getGlobalArc();

const transformBackendDataConnectionToDataConnection = ({
  id,
  name,
  type,
  typeName,
  info,
  cmlStatus,
  cmlId,
  isPasswordSet,
  supportsImportData,
}: BackendDataConnection): DataConnection | null => {
  try {
    const parsedInfo = JSON.parse(info) as DataConnection['info'];
    const parsedCmlStatus = JSON.parse(cmlStatus) as {
      isAvailable?: boolean;
    };

    const transformedConnection = {
      id,
      info: parsedInfo,
      isPasswordSet,
      name,
      type: type as DataConnectionType,
      typeName,
      cml: {
        id: cmlId,
        isAvailable: Boolean(parsedCmlStatus.isAvailable),
      },
      supportsImportData,
    };

    return transformedConnection;
  } catch (error) {
    console.error(error);

    return null;
  }
};

export const getAllDataConnections = async (
  limit?: number
): Promise<DataConnection[]> => {
  const params: Record<string, string> = {
    get_type: 'all',
    order_by: 'last_used',
    import_data: String(true),
  };
  if (limit !== undefined) {
    params.limit = String(limit);
  }

  const { data } = await vizHttp.get<BackendDataConnection[]>({
    url: arc.frontend.globals.urls.dataConnectionUrl,
    params,
    isOldEndpoint: true,
  });

  const transformedDataConnections = data.map(
    transformBackendDataConnectionToDataConnection
  );

  return transformedDataConnections.filter(
    (dataConnection): dataConnection is DataConnection =>
      dataConnection !== null
  );
};
