import { getCompletion } from './get-completion';
import { OverflowPolicy } from 'types/chatbot-message.type';

afterEach(() => {
  jest.clearAllMocks();
});

describe('getCompletion', () => {
  const sourceInfo = {
    filteredRows: [['row1', 'row2']],
    aliases: ['alias1'],
    contextPrompt: 'Use the following sources...',
  };
  const previousMessages = [{ role: 'user', content: 'Previous message' }];
  const question = 'What is the capital of France?';
  const settings = {
    completionProfile: 'open_ai',
    paramSelection: 'all',
    questionPromptPrefix: 'Q: <<question>>',
    completionMaxTokens: '512',
    overflowPolicy: 'throw-error' as OverflowPolicy,
  };
  const chunkCb = jest.fn();

  it('should return the completion response content', async () => {
    const sendHttpReq = jest.fn().mockResolvedValue({
      response: {
        body: {
          choices: [{ message: { content: 'Paris' } }],
        },
      },
    });

    const completion = await getCompletion(
      sendHttpReq,
      chunkCb,
      sourceInfo,
      previousMessages,
      question,
      settings
    );

    expect(completion).toBe('Paris');
  });

  it('should throw an error if API call fails', async () => {
    const sendHttpReq = jest
      .fn()
      .mockRejectedValue(new Error('API call failed'));

    await expect(
      getCompletion(
        sendHttpReq,
        chunkCb,
        sourceInfo,
        previousMessages,
        question,
        settings
      )
    ).rejects.toThrow('API call failed');
  });

  it('should set user message content to question when questionPromptPrefix is not defined', async () => {
    const modifiedSettings = {
      ...settings,
      questionPromptPrefix: '',
    };

    const sendHttpReq = jest.fn().mockResolvedValue({
      response: {
        body: {
          choices: [{ message: { content: 'Some response content' } }],
        },
      },
    });

    await getCompletion(
      sendHttpReq,
      chunkCb,
      sourceInfo,
      previousMessages,
      question,
      modifiedSettings
    );

    expect(sendHttpReq).toHaveBeenCalledWith(
      expect.objectContaining({
        body: expect.objectContaining({
          upstreamReqBody: expect.objectContaining({
            input: expect.arrayContaining([
              expect.objectContaining({ role: 'user', content: question }),
            ]),
          }),
        }),
      }),
      chunkCb
    );
  });
});
