import { Column, Dataset } from 'arc/types/dataset.type';
import {
  runSqlQuery,
  SqlQueryResponse,
} from 'components/sql-explorer/util/util';
import { getGlobalArc } from 'utils/get-global-arc';
import {
  DimensionBaseStatistics,
  MeasureBaseStatistics,
  MeasureStatistics,
  DimensionStatistics,
} from '../types/profile-statistic.types';
import { SelectBuilder } from 'arc/types/arc.type';

const arc = getGlobalArc();
// Sometimes the end result of a query is returned as string  or vica-versa rather than number (depends on the db)
type PossibleResultType =
  | number
  | string
  | null
  | 'null'
  | 'NULL'
  | undefined
  | 'None';

const castToFloat = (value: PossibleResultType) => {
  switch (typeof value) {
    case 'number':
      return value;
    case 'string':
      const parsedValue = Number.parseFloat(value);
      return Number.isNaN(parsedValue) ? 0 : parsedValue;
    default: // undefined, null
      return 0;
  }
};

const castToString = (value: PossibleResultType) => {
  switch (typeof value) {
    case 'number':
      return value.toString();
    case 'string':
      return value;
    default:
      return 'undefined';
  }
};

const isSqlLiteConn = (dataset: Dataset) =>
  dataset.dataconnection_type() === 'sqlite';

const isClouderaConnection = (dataset: Dataset) =>
  ['impyla', 'impala', 'hive', 'hyve'].includes(dataset.dataconnection_type());

const isTimeStampType = (column: Column) =>
  column.coltype().toUpperCase() === 'TIMESTAMP';

function numberOfZeroes(expr: string): string {
  return `sum(case when ${expr}=0 then 1 else 0 end)`;
}

function numberOfEmptyString(
  expression: string,
  dataset: Dataset,
  column: Column
): string {
  return `sum(case when ${castTimeStampStringIfNeeded(
    expression,
    dataset,
    column
  )}='' then 1 else 0 end)`;
}

function convertStringNullToNULL(expr: string): string {
  return `case when ${expr} IS 'NULL' then NULL else ${expr} end`;
}

function varianceCalculator(expr: string, dataset: Dataset) {
  const tempSelectBuilder = arc.selectBuilder();
  tempSelectBuilder.dataset(dataset);
  tempSelectBuilder.addProjection(`avg(${expr})`);
  const averageCalculatorSnippet = tempSelectBuilder.getSql();
  if (isClouderaConnection(dataset)) {
    return `VARIANCE(${expr})`;
  }
  return `SUM((${expr}-(${averageCalculatorSnippet}))*(${expr}-(${averageCalculatorSnippet})) ) / (COUNT(${expr})-1)`;
}

const castTimeStampStringIfNeeded = (
  expr: string,
  dataset: Dataset,
  column: Column
) => {
  const isMeasure = dataset
    .tables()
    .flatMap((table) => table.measuresOnly())
    .filter((c) => !c.isHierarchy() && !c.isComplexType())
    .includes(column);
  const castToString =
    isTimeStampType(column) &&
    ['impyla', 'impala', 'hive', 'hyve', 'postgres'].includes(
      dataset.dataconnection_type()
    ) &&
    !isMeasure;

  return castToString
    ? `cast(${expr} as ${
        dataset.dataconnection_type() === 'postgres' ? 'char' : 'string'
      })`
    : expr;
};
function numNulls(expr: string, dataset: Dataset, column: Column): string {
  const isMeasure = dataset
    .tables()
    .flatMap((table) => table.measuresOnly())
    .filter((c) => !c.isHierarchy() && !c.isComplexType())
    .includes(column);
  let expression = castTimeStampStringIfNeeded(expr, dataset, column);
  const includeStringNull = isSqlLiteConn(dataset) && isMeasure;
  return `sum(case when ${expression} IS NULL ${
    includeStringNull ? 'or "NULL"' : ''
  } then 1 else 0 end)`;
}

function longestWordLength(
  expression: string,
  dataset: Dataset,
  column: Column
): string {
  return `max(length(${castTimeStampStringIfNeeded(
    expression,
    dataset,
    column
  )}))`;
}

function shortestWordLength(
  expression: string,
  dataset: Dataset,
  column: Column
): string {
  return `min(length(${castTimeStampStringIfNeeded(
    expression,
    dataset,
    column
  )}))`;
}

const avg = (colname: string) => `avg(${colname})`;
const min = (colname: string) => `min(${colname})`;
const max = (colname: string) => `max(${colname})`;
const count = (colname: string) => `count(${colname})`;

function countDistinctFn(expression: string, dataset: Dataset): string {
  return arc.dataconnection.approximateOrExactCount(
    dataset.dataconnection_type(),
    expression
  );
}

function getExpression(
  column: Column,
  dataset: Dataset,
  sb: SelectBuilder
): string {
  if (!column.build_dsobject().basecol) {
    let exp = arc.dataconnection.types.brackets(
      column.build_dsobject().alias ?? column.build_dsobject().name
    );
    exp = sb.addTablenamesFromExpression(exp);
    return exp;
  } else {
    return dataset.colnames_from_expr(column.build_dsobject().expr).sqltext;
  }
}

export const getTextBaseStatistics = (
  column: Column,
  dataset: Dataset
): Promise<DimensionBaseStatistics> => {
  return new Promise((resolve, reject) => {
    const sb = arc.selectBuilder();
    sb.dataset(dataset);
    const expression = getExpression(column, dataset, sb);

    sb.addProjection(min(expression)); // Question: Should we exclude 'NULL' here as well?
    sb.addProjection(max(expression)); // Question: Should we exclude 'NULL' here as well?
    sb.addProjection(countDistinctFn(expression, dataset));
    sb.addProjection(numNulls(expression, dataset, column));
    sb.addProjection(count(expression)); // TODO check on Every connection is it the right one?
    sb.addProjection(longestWordLength(expression, dataset, column));
    sb.addProjection(shortestWordLength(expression, dataset, column));
    sb.addProjection(numberOfEmptyString(expression, dataset, column));

    runSqlQuery(
      sb.getSql(),
      Number(dataset.dataconnection_id()),
      (result: SqlQueryResponse<PossibleResultType>) => {
        const [
          min,
          max,
          distinctCount,
          numberOfNulls,
          recordCount,
          longestWord,
          shortestWord,
          numberOfEmpty,
        ] = result.rows[0];
        const dimensionProps: DimensionBaseStatistics = {
          min: castToString(min),
          max: castToString(max),
          distinctCount: castToFloat(distinctCount),
          numberOfNulls: castToFloat(numberOfNulls),
          recordCount: castToFloat(recordCount),
          longestWord: castToString(longestWord),
          shortestWord: castToString(shortestWord),
          numberOfEmpty: castToFloat(numberOfEmpty),
          fieldType: column.coltype(),
          name: column.build_dsobject().alias,
          duplicatedRows: castToFloat(recordCount) - castToFloat(distinctCount),
        };
        return resolve(dimensionProps);
      },
      (error: Error | string) => {
        return reject(error);
      }
    );
  });
};

// Get most or least frequent value
function getOutlierFrequencyValue<T>(
  dataset: Dataset,
  column: Column,
  arrange: 'asc' | 'desc',
  castFn: (value: PossibleResultType) => T
): Promise<{ mostFrequentValue: T } | { leastFrequentValue: T }> {
  return new Promise((resolve, reject) => {
    const sb = arc.selectBuilder();
    sb.dataset(dataset);
    const expression = getExpression(column, dataset, sb);
    sb.addOrderBy(`count(${expression}) ${arrange}`);
    sb.addGroup(expression);
    sb.limit(1);

    runSqlQuery(
      sb.getSql(),
      Number(dataset.dataconnection_id()),
      (result: SqlQueryResponse<PossibleResultType>) => {
        const resultRow = result.rows?.[0];
        const value = castFn(resultRow?.[0]);
        return resolve(
          arrange === 'desc'
            ? { mostFrequentValue: value }
            : { leastFrequentValue: value }
        );
      },
      (error: Error | string) => {
        return reject(error);
      }
    );
  });
}

export const getTextColumnStatistics = (
  column: Column,
  dataset: Dataset
): Promise<DimensionStatistics> => {
  return Promise.allSettled([
    getTextBaseStatistics(column, dataset),
    getOutlierFrequencyValue(dataset, column, 'desc', castToString),
    getOutlierFrequencyValue(dataset, column, 'asc', castToString),
  ]).then((results) => {
    let statistic: DimensionStatistics = {
      min: '',
      max: '',
      shortestWord: '',
      longestWord: '',
      numberOfEmpty: -1,
      duplicatedRows: -1,
      name: '',
      numberOfNulls: -1,
      recordCount: -1,
      distinctCount: -1,
      mostFrequentValue: '',
      leastFrequentValue: '',
      fieldType: 'UNKNOWN',
    };

    if (results[0].status === 'fulfilled') {
      statistic = { ...statistic, ...results[0].value };
    } else {
      throw results[0].reason;
    }
    if (results[1].status === 'fulfilled') {
      statistic = { ...statistic, ...results[1].value };
    } else {
      statistic.mostFrequentValue = new Error(
        `Failed to retrieve it: ${results[1].reason}`
      );
    }
    if (results[2].status === 'fulfilled') {
      statistic = { ...statistic, ...results[2].value };
    } else {
      statistic.leastFrequentValue = new Error(
        `Failed to retrieve it: ${results[2].reason}`
      );
    }

    return statistic;
  });
};

/*
  This was necessary to handle scenarios, to handle when the database stores it as a string,
  however we need numeric values.
*/

export function castToReal(expr: string, dataset: Dataset): string {
  const dctype = dataset.dataconnection_type();
  const dccast = arc.dataconnection.types[dctype]().cast_real;
  return `cast(${expr} as ${dccast})`;
}

export const getMeasureColumnStatistics = (
  column: Column,
  dataset: Dataset
): Promise<MeasureStatistics> => {
  return Promise.allSettled([
    getMeasureBaseStatistics(column, dataset),
    getOutlierFrequencyValue(
      dataset,
      column,
      'desc' /* most frequent */,
      castToFloat
    ),
    getOutlierFrequencyValue(
      dataset,
      column,
      'asc' /* least frequent*/,
      castToFloat
    ),
  ]).then((results) => {
    let statistic: MeasureStatistics = {
      min: -1,
      max: -1,
      avg: -1,
      variance: -1,
      numberOfZeroes: -1,
      duplicatedRows: -1,
      name: '',
      numberOfNulls: -1,
      recordCount: -1,
      distinctCount: -1,
      mostFrequentValue: '',
      leastFrequentValue: '',
      fieldType: 'UNKNOWN',
    };

    // The reason I am not using Promise.all, because the basic statistics works more or less fine
    // However if only the most or least frequent value fails (timed out etc), I don't want the rest to stop.
    if (results[0].status === 'fulfilled') {
      statistic = { ...statistic, ...results[0].value };
    } else {
      throw results[0].reason;
    }
    if (results[1].status === 'fulfilled') {
      statistic = { ...statistic, ...results[1].value };
    } else {
      statistic.mostFrequentValue = new Error(
        `Failed to retrieve it: ${results[1].reason}`
      );
    }
    if (results[2].status === 'fulfilled') {
      statistic = { ...statistic, ...results[2].value };
    } else {
      statistic.leastFrequentValue = new Error(
        `Failed to retrieve it: ${results[2].reason}`
      );
    }
    return statistic;
  });
};

export const getMeasureBaseStatistics = (
  column: Column,
  dataset: Dataset
): Promise<MeasureBaseStatistics> => {
  return new Promise((resolve, reject) => {
    const baseStatSelectbuilder = arc.selectBuilder();
    baseStatSelectbuilder.limit(100);
    baseStatSelectbuilder.dataset(dataset);
    let expression = getExpression(column, dataset, baseStatSelectbuilder);

    if (isTimeStampType(column)) {
      expression = castToReal(expression as string, dataset);
    }

    const isSqlLite = isSqlLiteConn(dataset);

    if (isSqlLite) {
      expression = convertStringNullToNULL(expression);
    }

    baseStatSelectbuilder.addProjection(min(expression));
    baseStatSelectbuilder.addProjection(max(expression));
    baseStatSelectbuilder.addProjection(countDistinctFn(expression, dataset));
    baseStatSelectbuilder.addProjection(numNulls(expression, dataset, column));
    baseStatSelectbuilder.addProjection(avg(expression));
    baseStatSelectbuilder.addProjection(count(expression));
    baseStatSelectbuilder.addProjection(numberOfZeroes(expression));
    baseStatSelectbuilder.addProjection(
      varianceCalculator(expression, dataset)
    );

    runSqlQuery(
      baseStatSelectbuilder.getSql(),
      Number(dataset.dataconnection_id()),
      (result: SqlQueryResponse<PossibleResultType>) => {
        const [
          min,
          max,
          distinctCount,
          numberOfNulls,
          avg,
          recordCount,
          numberOfZeroes,
          variance,
        ] = result.rows[0].map((item) => castToFloat(item));

        const measureProps: MeasureBaseStatistics = {
          min,
          max,
          distinctCount,
          numberOfNulls,
          avg,
          recordCount,
          numberOfZeroes,
          variance,
          fieldType: column.coltype(),
          name: column.build_dsobject().alias,
          duplicatedRows: recordCount - distinctCount,
        };
        return resolve(measureProps);
      },
      (error: Error | string) => {
        return reject(error);
      }
    );
  });
};
