import { useQuery, useQueryClient } from '@tanstack/react-query';
import { QueryKeyBase } from 'apis/query-keys';
import { vizHttp } from 'services/viz-http';
import { JobStatus } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import { useLastLogQuery } from './logs/get-logs';

const arc = getGlobalArc();

type BackendJobStatus = {
  status: JobStatus;
};

export const getJobStatus = async (id: number): Promise<JobStatus> => {
  const { data } = await vizHttp.get<BackendJobStatus>({
    url: arc.frontend.globals.urls.jobGetStatusUrl.replace(
      'schedule_id',
      String(id)
    ),
    isOldEndpoint: false,
  });

  return data.status;
};

/**
 * Periodically checks the status of the currently queued/running job.
 * Invalidates the job and logs queries when the status has changed.
 *
 * @param jobId The id of the job to check
 * @param interval The refetch interval in ms
 */
export const usePeriodicalJobStatusCheck = (
  jobId: number,
  interval: number
) => {
  const queryClient = useQueryClient();

  const { data: lastLog } = useLastLogQuery(jobId);

  useQuery({
    queryKey: [QueryKeyBase.JobStatus, jobId],
    queryFn: () =>
      getJobStatus(jobId).then((status) => {
        if (lastLog?.status !== status) {
          queryClient.invalidateQueries([QueryKeyBase.Jobs, jobId]);
          queryClient.invalidateQueries([QueryKeyBase.JobLogs, jobId]);
        }

        return status;
      }),
    enabled:
      !!lastLog &&
      [JobStatus.Running, JobStatus.Queued].includes(lastLog?.status),
    refetchInterval: interval,
    staleTime: interval,
  });
};
