import { useQuery } from '@tanstack/react-query';
import {
  ColumnFiltersState,
  PaginationState,
  SortingState,
} from '@tanstack/react-table';
import { QueryKeyBase } from 'apis/query-keys';
import { vizHttp } from 'services/viz-http';
import { Paginated, PaginationOptions, SortingOptions } from 'types/api.type';
import { Job, JobStatus, JobType, SchedulingStatus } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import { BackendJob, parseJob } from './parse-job';

const arc = getGlobalArc();

type BackendJobs = {
  jobs: BackendJob[];
  total: number;
};

type FilteringOptions = {
  filters?: {
    state?: (JobStatus | SchedulingStatus)[];
    isScheduled?: boolean;
    owner?: number[];
    type?: JobType[];
    lastFinishedTime?: string;
    search?: string;
    thresholdOnly?: boolean;
  };
};

const parseToBackendColumn = (key: string) => {
  switch (key) {
    case 'id':
      return 'id';
    case 'name':
      return 'job_name';
    case 'updateTime':
      return 'update_time';
    case 'state':
      return 'last_status';
    case 'scheduleGroup_id':
      return 'schedule_group_id';
    case 'lastFinishedTime':
      return 'last_finished_time';
    default:
      return key;
  }
};

export const getJobs = async ({
  page,
  pageSize,
  sort,
  order,
  filters: {
    state,
    isScheduled,
    owner,
    lastFinishedTime,
    type,
    search,
    thresholdOnly,
  } = {},
}: PaginationOptions & SortingOptions & FilteringOptions = {}): Promise<
  Paginated<Job>
> => {
  const sortByColumn =
    sort && `${order === 'desc' ? '-' : ''}${parseToBackendColumn(sort)}`;

  const { data } = await vizHttp.get<BackendJobs>({
    url: arc.frontend.globals.urls.jobsApiUrl,
    isOldEndpoint: false,
    params: {
      ...(page && { page: String(page) }),
      ...(pageSize && { page_size: String(pageSize) }),
      ...(sortByColumn && { sort: sortByColumn }),
      ...(search && { search }),
      ...(isScheduled !== undefined && { is_scheduled: String(isScheduled) }),
      ...(state && { status: state.join(',') }),
      ...(owner && { owner: owner.join(',') }),
      ...(type?.length === 1 && {
        is_extract: String(type[0] === JobType.DataExtract),
      }),
      ...(lastFinishedTime?.[0] && {
        last_finished_time_start: lastFinishedTime[0],
      }),
      ...(lastFinishedTime?.[1] && {
        last_finished_time_end: lastFinishedTime[1],
      }),
      ...(thresholdOnly && {
        threshold_only: String(thresholdOnly),
      }),
    },
  });

  const jobs = data.jobs.map((job) => parseJob(job));

  return {
    data: jobs,
    total: data.total,
  };
};

type useGetJobsQueryProps = {
  pagination: PaginationState;
  sorting: SortingState;
  columnFilters: ColumnFiltersState;
  globalFilter: string;
  thresholdOnly: boolean;
};

export const useGetJobsQuery = ({
  pagination,
  sorting,
  columnFilters,
  globalFilter,
  thresholdOnly,
}: useGetJobsQueryProps) => {
  const params = {
    page: pagination.pageIndex + 1,
    pageSize: pagination.pageSize,
    sort: sorting[0]?.id,
    order: sorting[0]?.desc ? ('desc' as const) : ('asc' as const),
    filters: {
      ...Object.fromEntries(columnFilters.map(({ id, value }) => [id, value])),
      isScheduled: true,
      search: globalFilter,
      thresholdOnly,
    },
  };

  const query = useQuery({
    queryKey: [QueryKeyBase.Jobs, params],
    staleTime: 300000, // 5 minutes
    queryFn: () => getJobs(params),
  });

  return query;
};
