import { useQuery } from '@tanstack/react-query';
import {
  ColumnFiltersState,
  PaginationState,
  SortingState,
} from '@tanstack/react-table';
import { QueryKeyBase } from 'apis/query-keys';
import { vizHttp } from 'services/viz-http';
import { Paginated, PaginationOptions, SortingOptions } from 'types/api.type';
import { JobLog } from 'types/job-log.type';
import { JobStatus } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import { parseJobLog } from './parse-log';

const arc = getGlobalArc();

type BackendLogs = {
  logs: {
    contents_archive: string;
    id: number;
    status: JobStatus;
    schedule_id: number;
    start_time: string;
    end_time: string;
    task_list: string;
    error_message: string;
  }[];
  total: number;
};

const parseJobLogs = (backendLogs: BackendLogs): JobLog[] => {
  return backendLogs.logs.map((backendLog) => parseJobLog(backendLog));
};

const parseToBackendColumn = (key: string) => {
  switch (key) {
    case 'id':
      return 'id';
    case 'startTime':
      return 'start_time';
    case 'endTime':
      return 'end_time';
    case 'status':
      return 'status';
  }
};

type FilteringOptions = {
  filters?: {
    status?: JobStatus[];
    startTime?: string;
    endTime?: string;
  };
};

export const getJobLogs = async (
  jobId: number,
  {
    page,
    pageSize,
    sort,
    order,
    filters: { status, startTime, endTime } = {},
  }: PaginationOptions & SortingOptions & FilteringOptions = {}
): Promise<Paginated<JobLog>> => {
  const sortByColumn =
    sort && `${order === 'desc' ? '-' : ''}${parseToBackendColumn(sort)}`;

  const { data } = await vizHttp.get<BackendLogs>({
    url: arc.frontend.globals.urls.jobGetLogsUrl.replace(
      'schedule_id',
      String(jobId)
    ),
    isOldEndpoint: false,
    params: {
      ...(page && { page: String(page) }),
      ...(pageSize && { page_size: String(pageSize) }),
      ...(sortByColumn && { sort: sortByColumn }),
      ...(status && { status: status.join(',') }),
      ...(startTime && { start_time: startTime }),
      ...(endTime && { end_time: endTime }),
    },
  });

  return {
    data: parseJobLogs(data),
    total: data.total,
  };
};

type useGetLogsQueryProps = {
  jobId: number;
  pagination: PaginationState;
  sorting: SortingState;
  columnFilters: ColumnFiltersState;
};

export const useGetLogsQuery = ({
  jobId,
  pagination,
  sorting,
  columnFilters,
}: useGetLogsQueryProps) => {
  const params = {
    page: pagination.pageIndex + 1,
    pageSize: pagination.pageSize,
    sort: sorting[0]?.id,
    order: sorting[0]?.desc ? ('desc' as const) : ('asc' as const),
    filters: {
      ...Object.fromEntries(columnFilters.map(({ id, value }) => [id, value])),
    },
  };

  const query = useQuery({
    queryKey: [QueryKeyBase.JobLogs, jobId, params],
    staleTime: 300000, // 5 minutes
    queryFn: () => getJobLogs(jobId, params),
    keepPreviousData: true,
  });

  return query;
};

export const useLastLogQuery = (jobId: number) => {
  const query = useQuery({
    queryKey: [QueryKeyBase.JobLogs, jobId, { page: 1, pageSize: 1 }],
    staleTime: 300000, // 5 minutes
    queryFn: () => getJobLogs(jobId, { page: 1, pageSize: 1 }),
    select: (data) => data.data[0],
  });

  return query;
};
