import {
  DataCopyTarget,
  EmailTarget,
  JobStatus,
  SnapshotTarget,
} from 'types/job.type';
import {
  BackendTargetByTargetType,
  parseDataCopyTarget,
  parseEmailTarget,
  parseSnapshotTarget,
} from '../utils';
import { JobLog, JobLogTask } from 'types/job-log.type';
import moment from 'moment';

type BackendRunParameters = {
  creator_id?: number;
  creator_username?: string;
  owner_id?: number;
  owner_username?: string;
  skip_next_targets: boolean;

  // Present if the run was triggered manually
  force_run?: boolean;
  editor_id?: number;
  editor_username?: string;

  // Present if the run failed
  error?: string;
  last_target?: string;

  // Present if the target is an extract and the run failed or finished
  extract?: unknown;
} & Partial<BackendTargetByTargetType['E']>;

type BackendJobLogTask = {
  message: string | unknown;
  date?: string;
};

type BackendLogTaskList = {
  tasks: BackendJobLogTask[];
};

type BackendContentsArchive = {
  run_parameters: BackendRunParameters;
  0?: BackendTargetByTargetType['S'];
  1?: BackendTargetByTargetType['S'];
};

export type BackendLog = {
  contents_archive: string;
  id: number;
  status: JobStatus;
  schedule_id: number;
  start_time: string;
  end_time: string;
  task_list: string;
  error_message: string;
};

const parseTaskList = (backendTaskList: string): JobLogTask[] => {
  const taskList: BackendLogTaskList = JSON.parse(backendTaskList);

  return taskList.tasks.map(({ date, message }) => ({
    date: date,
    message: typeof message === 'string' ? message : JSON.stringify(message),
  }));
};

export const parseJobLog = (backendLog: BackendLog): JobLog => {
  const contentArchive: BackendContentsArchive = JSON.parse(
    backendLog.contents_archive
  );

  let emailTarget: EmailTarget | undefined;
  try {
    // the contentArchive does not have a set structure, sometimes it contains an email target
    // a visual url is required for parsing the email target
    if (contentArchive.run_parameters.visual_url) {
      emailTarget = parseEmailTarget([
        {
          fields: {
            target_type: 'E',
            target_data: JSON.stringify(contentArchive.run_parameters),
          },
        },
      ]);
    }
  } catch (error) {
    console.error('Unable to parse log email target: ', error);
  }

  let snapshotTarget: SnapshotTarget | undefined;
  try {
    snapshotTarget = parseSnapshotTarget([
      {
        fields: {
          target_type: 'S',
          target_data: JSON.stringify(
            contentArchive['0'] ?? contentArchive['1']
          ),
        },
      },
    ]);
  } catch (error) {
    console.error('Unable to parse log snapshot target: ', error);
  }

  let dataCopyTarget: DataCopyTarget | undefined;
  try {
    if (contentArchive.run_parameters.extract) {
      dataCopyTarget = parseDataCopyTarget([
        {
          fields: {
            target_type: 'D',
            target_data: JSON.stringify(
              contentArchive['0'] ?? contentArchive['1']
            ),
          },
        },
      ]);
    }
  } catch (error) {
    console.error('Unable to parse log data copy target: ', error);
  }

  return {
    id: backendLog.id,
    status: backendLog.status,
    jobId: backendLog.schedule_id,
    ownerUsername: contentArchive.run_parameters.owner_username,
    editorUsername: contentArchive.run_parameters.editor_username,
    startTime: backendLog.start_time,
    endTime: backendLog.end_time,
    totalRunTime: moment(backendLog.end_time).diff(backendLog.start_time),
    taskList: parseTaskList(backendLog.task_list),
    errorMessage: backendLog.error_message,
    triggeredManually: contentArchive.run_parameters.force_run ?? false,
    triggeredByUsername: contentArchive.run_parameters.editor_username,
    dataCopyTarget,
    emailTarget,
    snapshotTarget,
  };
};
