import { Job, JobStatus, JobType, SchedulingStatus } from 'types/job.type';
import {
  parseDataCopyTarget,
  parseEmailTarget,
  parseRefreshTarget,
  parseSnapshotTarget,
  parseThresholdTarget,
} from './utils';

export type BackendJob = {
  creator_id?: number;
  creator_username?: string;
  job_name: string;
  extract_id?: number | null;
  last_finished_time?: string;
  last_started_time?: string;
  owner_id?: number;
  owner_username?: string;
  schedule_group_id?: number;
  schedule_group?: {
    id: number;
    schedule_name: string;
    run_interval: string;
    next_run_time: string;
    jobschedule_set: {
      model: string;
      pk: number;
      fields: {
        job_name: string;
      };
    }[];
  };
  schedule_id: number;
  scheduler_status: JobStatus | SchedulingStatus.Paused;
  targets?: {
    fields: {
      target_type: string;
      minimum_time_interval?: number;
      target_data: string;
    };
  }[];
};

export const parseJob = (backendJob: BackendJob): Job => {
  const {
    schedule_id,
    job_name,
    scheduler_status,
    targets,
    schedule_group_id,
    schedule_group,
    owner_username,
    last_finished_time,
    extract_id,
  } = backendJob;

  const snapshotTarget = parseSnapshotTarget(targets);
  const emailTarget = parseEmailTarget(targets);
  const dataCopyTarget = parseDataCopyTarget(targets);
  const thresholdTarget = parseThresholdTarget(targets);
  const refreshTarget = parseRefreshTarget(targets);

  const scheduleGroupId = schedule_group?.id || schedule_group_id;
  const scheduleGroup = scheduleGroupId
    ? {
        id: scheduleGroupId,
        name: schedule_group?.schedule_name,
      }
    : undefined;

  const baseJob = {
    id: schedule_id,
    name: job_name,
    isSchedulingPaused: scheduler_status === SchedulingStatus.Paused,
    owner: owner_username,
    lastStatus:
      scheduler_status === SchedulingStatus.Paused
        ? JobStatus.Unknown
        : scheduler_status,
    lastFinishedTime: last_finished_time,
    scheduleGroup,
  };

  if (dataCopyTarget && extract_id) {
    return {
      ...baseJob,
      dataCopyTarget,
      extractId: extract_id,
      type: JobType.DataExtract,
    };
  }

  if (emailTarget) {
    return {
      ...baseJob,
      type: JobType.Email,
      thresholdTarget,
      emailTarget,
      snapshotTarget,
    };
  }

  if (refreshTarget) {
    return {
      ...baseJob,
      type: JobType.Refresh,
      refreshTarget,
    };
  }

  return {
    ...baseJob,
    type: JobType.Other,
  };
};
