import { useQueryClient } from '@tanstack/react-query';
import { QueryKeyBase } from 'apis/query-keys';
import { AxiosError } from 'axios';
import { useMutationWithNotification } from 'hooks/use-mutation-with-notification';
import { vizHttp } from 'services/viz-http';
import { getGlobalArc } from 'utils/get-global-arc';

const arc = getGlobalArc();

type CreateViewRefreshJobParams = {
  name: string;
  errorTo: string[];
  scheduleGroupId?: number;
  scheduleGroup?: { name: string; interval: string } | null;
  dataConnectionId: number;
  viewName: string;
};

export const createViewRefreshJob = async (
  params: CreateViewRefreshJobParams
) => {
  const result = await vizHttp.post<{ schedule_id: number; next_run: string }>({
    url: arc.frontend.globals.urls.jobViewRefreshUrl,
    headers: {
      'X-CSRFTOKEN': arc.frontend.globals.authentication.csrfToken,
    },
    isOldEndpoint: true,
    data: {
      opcode: 'schedule',
      formValues: JSON.stringify({
        job_name: params.name,
        error_to_addresses: params.errorTo,
        schedule_group: params.scheduleGroupId,
        schedule_group_name: params.scheduleGroup?.name,
        run_interval: params.scheduleGroup?.interval,
        dcid: params.dataConnectionId,
        view: params.viewName,
      }),
    },
  });

  return result.data;
};

export const useCreateViewRefreshJobMutation = () => {
  const queryClient = useQueryClient();
  const mutation = useMutationWithNotification<
    Awaited<ReturnType<typeof createViewRefreshJob>>,
    AxiosError<string>,
    CreateViewRefreshJobParams
  >(
    {
      mutationFn: createViewRefreshJob,
      onSuccess: (_, { scheduleGroup }) => {
        queryClient.invalidateQueries([QueryKeyBase.Jobs]);

        if (scheduleGroup?.name && scheduleGroup?.interval) {
          queryClient.invalidateQueries([QueryKeyBase.ScheduleGroups]);
        }
      },
    },
    {
      successMessage: 'Successfully scheduled analytical view refresh',
    }
  );

  return mutation;
};
