/*
Relating HTML: visual_edit_sql.html, it is currently used only
in the ML(machine learning) explore workflow.
 */

import { getGlobalArc } from 'utils/get-global-arc';
import {
  addLimitToSQLQuery,
  moveCursorToEnd,
  setCommandEnterShortcutOnExpressionEditor,
  activeAlertContainer,
  deactiveAlertContainer,
} from '../misc/misc';
import type { EditorWrapper } from '../types/editor-wrapper.type';
import type { Dataset } from '../types/dataset.type';
import type { Report } from '../types/report.type';

const arc = getGlobalArc();
const SQLEDITOR_MODAL_ID = '.dataset-sql-editor';

export const datasetEditor = (selectedReport: Report): { open: () => void } => {
  const sqlEditorContainerId = '#sql-super-modal';
  const $modal = $(sqlEditorContainerId);
  const createExpressionEditor = (): EditorWrapper => {
    const $sel = $(`${sqlEditorContainerId} .rawcommand-holder`);
    $sel.find('.rawsql-expr-editor-container').remove();

    $sel.append($('#rawsql-textbox-template').html());
    const exprId = `expr-editor-${arc.misc.uniqueId()}`;

    $sel.find('.expression-field').attr('id', exprId);

    const exprEditor = arc.misc.expressionEditor(
      undefined,
      undefined,
      exprId,
      $sel.find('.rawsql-expr-editor-container'),
      undefined,
      true // enable resize
    );
    return exprEditor;
  };

  const getDatasetNameEditor = (): JQuery<HTMLDivElement> =>
    $<HTMLDivElement>(`${SQLEDITOR_MODAL_ID} div#dataset-name-holder`);

  const modal: {
    identifier: string;
    expressionEditor: EditorWrapper;
    datasetNameEditor: JQuery<HTMLDivElement>;
    selectedVisual: Report;
    originalDataset: Dataset;
    alterAllVisual: boolean;
  } = {
    identifier: SQLEDITOR_MODAL_ID,
    expressionEditor: createExpressionEditor(),
    datasetNameEditor: getDatasetNameEditor(),
    originalDataset: selectedReport.dataset(),
    selectedVisual: selectedReport,
    alterAllVisual: true,
  };

  const $limitCheckbox = $(`${modal.identifier} #add-limit`);

  function getQueryFromEditor(withLimit = true) {
    const rawQuery = modal.expressionEditor
      ?.aceEditor()
      .getSession()
      .getValue();
    if (!rawQuery) {
      return '';
    }
    let query = rawQuery;
    if (withLimit && $limitCheckbox.prop('checked')) {
      query = addLimitToSQLQuery(rawQuery, 100);
    }
    return query;
  }

  function getQueryFromDatasetInfo(dataset: Dataset) {
    // I suspect only one element exisit within dataset_info in case of SQL dataset
    return JSON.parse(dataset.dataset_info())[0].select;
  }

  function setQueryInExpressionEditor(query: string) {
    modal.expressionEditor?.aceEditor()?.getSession().setValue(query);
  }

  function enableApplyChangesButton() {
    const oldName = modal.originalDataset.dataset_name();
    const newName = modal.datasetNameEditor.text().trim();
    const isDataSetNameChanged = !newName ? false : oldName !== newName;
    const isSqlQueryChanged =
      getQueryFromEditor(false) !==
      getQueryFromDatasetInfo(modal.originalDataset);
    $(`${modal.identifier} #apply-button`).attr(
      'disabled',
      isSqlQueryChanged || isDataSetNameChanged ? null : 'true'
    );
  }

  function setupDatasetNameEditor() {
    modal.datasetNameEditor.text(modal.originalDataset.dataset_name());
    modal.datasetNameEditor.offOn(
      'keypress',
      function loseFocus(this: JQuery<HTMLDivElement>, event: KeyboardEvent) {
        if (event.key === 'Enter') {
          $(this).trigger('blur');
        }
      }
    );

    modal.datasetNameEditor.offOn('keyup', () => {
      enableApplyChangesButton();
    });

    modal.datasetNameEditor.offOn(
      'focusout',
      function fixEmptyName(this: JQuery<HTMLDivElement>) {
        // do not allow empty name!
        enableApplyChangesButton();
        if ($(this).text().trim() === '') {
          $(this).text(modal.originalDataset.dataset_name());
        }

        $(this).removeAttr('contenteditable');
        $(this).scrollLeft(0);
      }
    );

    return modal.datasetNameEditor;
  }

  function setupExpressionEditor() {
    setCommandEnterShortcutOnExpressionEditor(modal.expressionEditor, () => {});

    modal.expressionEditor
      ?.aceEditor()
      .getSession()
      .on('change', () => {
        enableApplyChangesButton();
      });

    const originalQuery = getQueryFromDatasetInfo(modal.originalDataset);
    setQueryInExpressionEditor(originalQuery);
  }

  function enableToEditDatasetName() {
    if (modal.datasetNameEditor) {
      modal.datasetNameEditor.attr('contenteditable', 'true');
      moveCursorToEnd(modal.datasetNameEditor); // this method also focus on it
      // somehow this scrollLeft set to the end not exactly works as expected
      // => multiple by 100 will do the trick
      modal.datasetNameEditor.scrollLeft(
        modal.datasetNameEditor.text().length * 100
      );
    }
  }

  function setupDefaultScope() {
    modal.alterAllVisual = true;
    $('#apply-to-options input#all-visual-option').prop('checked', true);
  }

  function resetForm() {
    modal.datasetNameEditor?.off();
    modal.datasetNameEditor?.text('');
    setQueryInExpressionEditor('');
    $limitCheckbox.prop('checked', true);
    $(`${modal.identifier} button#apply-button`).off();
    $(`${modal.identifier} #rawcommand-output-title .title`).empty();
    $(`${modal.identifier} #rawcommand-output`).empty();
    $(`${modal.identifier} #close-modal`).off();
    deactiveAlertContainer($modal);
  }

  function closeModal() {
    resetForm();
    $modal.fadeOut();
  }

  function setupModalClosingElements() {
    $(`${modal.identifier} #close-modal`).offOn('click', closeModal);
  }

  function setupDatasetDetailsPageLink() {
    // In the  ML workflow , CDV will be in an iframe where open in a new tab loose the point
    const url = `${
      arc.frontend.globals.urls.datasetEditUrl
    }/${modal.originalDataset.dataset_id()}`;
    $(`${modal.identifier} a#datasetpage-link`).offOn('click', () =>
      arc.misc.openInNewWindow(url)
    );
  }

  function setupEditDatasetNameLink() {
    $(`${modal.identifier} a#edit-dsname-icon`).offOn('click', () => {
      enableToEditDatasetName();
    });
  }

  function getVisualTitle() {
    return $(`#${modal.selectedVisual.title_elem()}`).text();
  }

  function setupVisualDetails() {
    const visualName = getVisualTitle();
    $(`${modal.identifier} #visual-name-holder`).text(
      visualName === '' ? 'untitled' : visualName
    );
    $(`${modal.identifier} #visual-id-holder`).text(
      `(ID: ${modal.selectedVisual.report_id()})`
    );
  }

  const $requestContainer = $('#modal-request-actions');
  const $runSqlButton = $requestContainer.find('#rawcommand-run');
  const $cancelRequestButton = $requestContainer.find('#rawcommand-cancel');

  const $cmdTitleHolder = $('#rawcommand-output-title');
  const $title = $cmdTitleHolder.find('.title');
  const $tableContainer = $('#rawcommand-output');

  const runQuery = () => {
    $title.empty();
    $tableContainer.empty();

    const onError = (error: string) => {
      // error handled by the query runner anyway no need to add extra error container :(
      arc.misc.exChangeElementsVisibility({
        elementToShow: $runSqlButton,
        elementToHide: $cancelRequestButton,
      });
    };

    const onSuccess = (response: { colnames: string[]; rows: any[] }) => {
      if (response.colnames.length === 0) {
        if (response.rows && response.rows.length > 0) {
          response.colnames = [];
          response.rows[0].forEach(function () {
            response.colnames.push('');
          });
        }
      }

      if (response.rows.length > 0) {
        arc.misc.plotTable(response.rows, response.colnames, $tableContainer);
      } else {
        $tableContainer.html(
          `<div style='text-align:center'>No data availale</div>`
        );
      }

      arc.misc.exChangeElementsVisibility({
        elementToShow: $runSqlButton,
        elementToHide: $cancelRequestButton,
      });
    };

    arc.misc.exChangeElementsVisibility({
      elementToShow: $cancelRequestButton,
      elementToHide: $runSqlButton,
    });
    const query = arc.data
      .queryCmd()
      .query(getQueryFromEditor())
      .dataconnection_id(
        Number(modal.originalDataset.dataconnection_id()) ?? -1
      )
      .cache_read(false);
    const baseQuery = arc.data.runQuery(
      undefined,
      onSuccess,
      onError,
      /* lib */ undefined,
      function errorreportingcb(qstate, msg, error) {
        onError(error);
        arc.report.alertMessage(
          'Failed to draw explore visuals',
          `${msg}: ${error}`,
          { alertContainerSelector: '#sql-super-modal' }
        );
      },
      /* timeoutSeconds */ undefined,
      /* timeoutCb */ undefined,
      query,
      /* loader */ undefined
    );

    $cancelRequestButton.offOn('click.cancel', () => {
      baseQuery.qstate.cancel();
      $tableContainer.html(`<h5>Error<h5><pre>Cancelled</pre>`);
      arc.misc.exChangeElementsVisibility({
        elementToShow: $runSqlButton,
        elementToHide: $cancelRequestButton,
      });
    });
  };

  async function setupRunQueryButton() {
    $(`${modal.identifier} button#rawcommand-run`).offOn('click', runQuery);
  }

  function setupModificationScopeCheckboxes() {
    const checkBoxToAll = $(`${modal.identifier} input#all-visual-option`);
    checkBoxToAll.prop('checked', true);
    checkBoxToAll.offOn('click', () => {
      modal.alterAllVisual = true;
    });

    const checkBoxToSelected = $(
      `${modal.identifier} input#selected-visual-option`
    );
    checkBoxToSelected.prop('checked', false);
    checkBoxToSelected.offOn('click', () => {
      modal.alterAllVisual = false;
    });
  }

  function initDatasetFromDB(datasetId: string | number): Promise<Dataset> {
    return new Promise<Dataset>((resolve, reject) => {
      // get dataset object from server
      arc.globals.dashboard.fetchDataset(
        { dataset_id: datasetId },
        (_config, datasetData, columninfo) => {
          const newDataset = arc.datasets
            .Dataset()
            .dataset_id(datasetId)
            // updates the dataset's UI represenatation
            .init_from_fetch(datasetData, columninfo);
          resolve(newDataset);
        },
        () => reject(new Error('Failed tp fetch dataset from server'))
      );
    });
  }

  const updateTables = (dataset: Dataset) =>
    new Promise((resolve, reject) => {
      dataset.update_tables((_, error) =>
        error ? reject(error) : resolve(undefined)
      );
    });

  const saveDataset = (dataset: Dataset) =>
    new Promise((resolve, reject) =>
      dataset.save((error: string | Error) =>
        error
          ? reject(new Error(`Failed to save dataset changes: ${error}`))
          : resolve(undefined)
      )
    );

  const getUpdateParameters = (
    datasetId: string | number,
    shouldApplyChangesOnNewDataset: boolean
  ): { name: string; cmd: string } => {
    const nameOnForm = modal.datasetNameEditor?.text().trim();
    const originalName = modal.originalDataset.dataset_name();
    let name = !nameOnForm ? originalName : nameOnForm;
    if (shouldApplyChangesOnNewDataset && name === originalName) {
      name = `${originalName} (${datasetId})`;
    }

    const cmd = getQueryFromEditor(false);
    return { name, cmd };
  };

  const updateDatasetLocally = async (
    dataset: Dataset,
    updateParameter: { name: string; cmd: string }
  ): Promise<Dataset> => {
    dataset.tables()[0].table_select(updateParameter.cmd);
    dataset.table_select(updateParameter.cmd);
    dataset.dataset_name(updateParameter.name);
    await updateTables(dataset);
    return dataset;
  };

  async function updateDatasetInDB(dataset: Dataset): Promise<Dataset> {
    await saveDataset(dataset);
    try {
      const newDataset = await initDatasetFromDB(dataset.dataset_id());
      arc.misc.notify('The SQL sucessfully saved');
      return newDataset;
    } catch (error) {
      throw new Error(`Failed to fetch the updated dataset: ${error}`);
    }
  }

  const unscopedVisualsDependOnDataset = async () => {
    // if allVisual -> any other dashbard uses it
    // if selected visual -> any other dashbaord OR more than the selected visual in the current
    const { dashboardIds } = await arc.datasets.getAssociatedDashboards(
      modal.originalDataset.dataset_id()
    );
    // if the dashboard is not saved and dataset used by only the unsaved dashboard
    // no ids-t will be returned
    const isDatasetUsedOnlyByThisDashboard =
      dashboardIds.length === 0 ||
      (dashboardIds.length === 1 &&
        dashboardIds[0] === arc.globals.dashboard.dashboard_id());

    if (!isDatasetUsedOnlyByThisDashboard) {
      return true;
    }

    if (!modal.alterAllVisual) {
      const reports = arc.globals.dashboard.reports().getAllReports();
      const isDatasetUsedByOtherVisualInCurrentDashboard = reports.some(
        (_report) =>
          _report?.dataset()?.dataset_id() ===
            modal.originalDataset.dataset_id() &&
          _report.report_id() !== modal.selectedVisual.report_id()
      );

      return isDatasetUsedByOtherVisualInCurrentDashboard;
    }

    return false;
  };

  const updateVisualLocally = (report: Report, dataset: Dataset) => {
    report.dataset_id(dataset.dataset_id());
    report.dataset(dataset);
    report.refresh(true);
  };

  const updateVisualsInDB = async (
    reports: Report[],
    dataset: Dataset
  ): Promise<string[]> => {
    const reportsIds = reports.map((visual) => visual.report_id());
    const datasetId = dataset.dataset_id();
    const hasDatasetIdChanged =
      datasetId !== modal.originalDataset.dataset_id();

    if (hasDatasetIdChanged) {
      return arc.globals.dashboard.patchDatasetRefOfReports({
        dataset_id: dataset.dataset_id(),
        report_ids: reportsIds,
      });
    }
    return reportsIds;
  };

  const updateVisuals = async (dataset: Dataset): Promise<void> => {
    try {
      const visuals = modal.alterAllVisual
        ? arc.globals.dashboard
            .reports()
            .getAllReports()
            .filter(
              (visual) =>
                visual.dataset_id() === modal.originalDataset.dataset_id()
            )
        : [modal.selectedVisual];

      await updateVisualsInDB(visuals, dataset);
      visuals.forEach((visual) => updateVisualLocally(visual, dataset));
    } catch (error) {
      throw new Error(`Failed to update visual(s): ${error}`);
    }
  };

  const updateDashboardSideBarsDatasetList = (dataset: Dataset) => {
    const dataconnectionOfAddFilterSidebar = arc.globals.dashboard
      .dashboard_filter_controls()
      .data_connection()
      .find((dc) => dc.dataconnection_id() === dataset.dataconnection_id());

    if (dataconnectionOfAddFilterSidebar) {
      const indexOfDataset = dataconnectionOfAddFilterSidebar
        ?.datasets()
        .findIndex((ds) => ds.dataset_id() === dataset.dataset_id());
      if (indexOfDataset !== -1) {
        dataconnectionOfAddFilterSidebar.datasets()[indexOfDataset] = dataset;
      } else {
        dataconnectionOfAddFilterSidebar.datasets().push(dataset);
      }
    }

    const dataConnectionOfAddVisualSideBar = arc.globals.dashboard
      .data_connections()
      .find((dc) => dc.dataconnection_id() === dataset.dataconnection_id());

    if (dataConnectionOfAddVisualSideBar) {
      const indexOfDataset = dataConnectionOfAddVisualSideBar
        .datasets()
        .findIndex((ds) => ds.dataset_id() === dataset.dataset_id());
      if (indexOfDataset !== -1) {
        dataConnectionOfAddVisualSideBar.datasets()[indexOfDataset] = dataset;
      } else {
        arc.globals.dashboard
          .data_connections_hash()
          [dataset.dataconnection_id()].add_dataset(dataset);
      }
    }

    arc.globals.dashboard.paintVisualSidebarDatasetsOfConn(
      dataset.dataconnection_id()
    );
    arc.globals.dashboard.dashboard_filter_controls().paint_datasets();
  };

  const updateApplication = async (
    updatedDataset: Dataset,
    appliedOnNewDataset: boolean
  ): Promise<void> => {
    const datasetHash = arc.globals.dashboard.dataset_hash();
    datasetHash[updatedDataset.dataset_id()] = updatedDataset;

    await updateVisuals(updatedDataset);
    updateDashboardSideBarsDatasetList(updatedDataset);
    const successMessage = appliedOnNewDataset
      ? `The SQL saved into a new dataset (name: "${modal.selectedVisual
          .dataset()
          .dataset_name()}"), and the effected visual(s) updated`
      : 'Dataset is saved and related visual(s) updated';

    arc.misc.notify(successMessage);
  };

  const onConfirm = async (
    shouldApplyModificationOnNewDataset: boolean
  ): Promise<void> => {
    try {
      let datasetToUpdate = modal.originalDataset;
      if (shouldApplyModificationOnNewDataset) {
        const clonedDatasetId =
          await modal.originalDataset.cloneDatasetOnServer();
        // As the cloned dataset returns only the ID,
        // we need to get the dataset object from the server
        datasetToUpdate = await initDatasetFromDB(clonedDatasetId);
      }

      const updateParameters = getUpdateParameters(
        datasetToUpdate.dataset_id(),
        shouldApplyModificationOnNewDataset
      );
      const updatedDataset = await updateDatasetLocally(
        datasetToUpdate,
        updateParameters
      );
      const savedDataset = await updateDatasetInDB(updatedDataset);
      await updateApplication(
        savedDataset,
        shouldApplyModificationOnNewDataset
      );
      closeModal();
    } catch (error) {
      const visualRepresentation = getVisualTitle()
        ? `visual (title: ${getVisualTitle()}`
        : `visual (id: ${modal.selectedVisual.report_id()})`;
      console.error(error);
      arc.report.alertMessageSmall(
        `Failed to save the dataset's changes of ${visualRepresentation}`,
        'alert-danger'
      );
    }
  };

  const openConfirmationDialog = async () => {
    try {
      const shouldApplyChangesOnNewDataset =
        await unscopedVisualsDependOnDataset();
      const defaultConfirmText = modal.alterAllVisual
        ? 'Changes will apply to all visuals that use this dataset and are irreversible. Apply to continue.'
        : 'Applying the changes is irreversible! Apply to continue.';
      const applyOnNewDatasetConfirmationText = modal.alterAllVisual
        ? `Changes will apply to all visuals that use this dataset and are irreversible.
           Note: a new dataset will be created to avoid affecting other dashboards. Apply to continue.`
        : 'Note: a new dataset will be created for this visual only. Apply to continue.';

      const { askUser } = arc.misc.simpleUserConfirmation({
        confirmTitle: 'Confirmation',
        confirmQus: shouldApplyChangesOnNewDataset
          ? applyOnNewDatasetConfirmationText
          : defaultConfirmText,
        confirmText: 'Apply',
      });
      askUser(() => {
        onConfirm(shouldApplyChangesOnNewDataset);
      });
    } catch (error) {
      console.error(error);
      arc.report.alertMessageSmall(
        "Failed to save the dataset's changes",
        'alert-danger'
      );
    }
  };

  function setupApplyButton() {
    $(`${modal.identifier} button#apply-button`).offOn(
      'click',
      openConfirmationDialog
    );
  }

  function setupModal() {
    setupExpressionEditor();
    setupDatasetNameEditor();
    setupVisualDetails();
    setupDefaultScope();
    setupModalClosingElements();
    setupDatasetDetailsPageLink();
    setupEditDatasetNameLink();
    setupRunQueryButton();
    setupModificationScopeCheckboxes();
    setupApplyButton();
    activeAlertContainer($modal);
  }

  const openDatasetSqlEditorModal = (): void => {
    $modal.fadeIn();
    setupModal();
  };

  return {
    open: openDatasetSqlEditorModal,
  };
};
