import type { Dataset } from '../types/dataset.type';
import type { EditorWrapper } from '../types/editor-wrapper.type';
import type { Report } from '../types/report.type';
import type { Setting } from '../types/setting.type';
import type { ExpressionFilter } from '../types/expression-filter.type';
import { getGlobalArc } from 'utils/get-global-arc';
import * as d3 from 'd3';

const arc = getGlobalArc();

const getExpressionEditorTemplate = (expressionFieldId: string) => `
<div class="edit-expression-body">
  <div class="edit-expression-textbox">
    <div class="form-group">
        <div id="${expressionFieldId}" class="form-control expression-field arc-no-dragging expression-data"></div>
    </div>

    <div class="pull-right edit-expression-autocomplete"></div>

    <button type="button" class="btn btn-secondary btn-sm viz-validate-btn expression-validate">Validate Expression</button>

    <div arcid="validate-spinner" style="display:none;">
      <i class="fa fa-spinner fa-spin arc-spinner"></i>
    </div>
  </div>

  <div class="edit-expression-functions-lists"></div>

  <div class="edit-expression-fields-lists"></div>
</div>`;

export const expression = (): Setting => {
  const setting = arc.visual
    .setting()
    .id('expression')
    .tab('Expression')
    .type(arc.visual.settingTypeEnum.EXPRESSION)
    .add_name('EXPRESSION')
    .value('');

  let selectedFilter: ExpressionFilter;
  let expressionEditor: EditorWrapper;

  const fetchFn = (): { expr: string } => {
    const value = expressionEditor.value();

    if (selectedFilter.getExpressionForEdit() === value) {
      return { expr: value };
    }

    selectedFilter.clearCustomExpr();
    selectedFilter.custom_expr(value);

    // Ensure we don't set alias or description for filter elements
    if (selectedFilter.filter_column()) {
      selectedFilter.col_alias(undefined);
      selectedFilter.col_desc(undefined);

      // If the expression is exactly the filter column name with brackets
      // then clear out the custom expression so that we get the special
      // (1=1) behavior when no parameter is present matching the column name
      if (value === `[${selectedFilter.filter_column()}]`) {
        selectedFilter.clearCustomExpr();

        // Allow users to change custom expression when
        // filter items were created in arcviz versions before release_3.3.0
        selectedFilter.filter_data([]);
      }
    }

    // Clear out other function selections in the menu
    selectedFilter.fn_name(null).fn_type(null).fn_cast(null);

    // Part of resetDateFunction
    if (selectedFilter.shelf_date_function()) {
      selectedFilter.shelf_date_function(null);
      selectedFilter.display_format(null).col_alias('');
    }

    return { expr: value };
  };

  const getOnValidate =
    (report: Report, dataset: Dataset) =>
    (e: MouseEvent, alternativeSuccessCb: () => void) => {
      if (!e.target) {
        return;
      }

      // Remove any existing alert text
      const $parent = $(e.target).parent();
      $parent.find('.alert').remove();
      $parent.find('[arcid=validate-spinner]').show();

      let value = expressionEditor.value();

      // Do parameter substitution & split out the "as 'alias'" part if present
      const valueBeforeSubstitution = value;
      value = report.viz().substituteParams(value);
      const { expr } = arc.misc.splitExprAndAlias(value);

      const validateExpression = (params: { [key: string]: unknown }) => {
        const { err, info } = params;

        $parent.find('.alert').remove();
        $parent.find('[arcid=validate-spinner]').hide();

        const loc = $parent.find('.expression-validate');
        if (err) {
          loc.after(
            arc.clone_template('col-val-failed', { errtext: err }).html()
          );
        } else if (info) {
          loc.after(arc.clone_template('col-val-info', { info }).html());
        } else if (alternativeSuccessCb) {
          alternativeSuccessCb();
        } else {
          loc.after(arc.clone_template('col-val-passed').html());
        }
      };

      // When the expression is evaluated to be an empty string after
      // substitution, don't do full validation and instead immediately call
      // validation success callback
      if (
        expr !== valueBeforeSubstitution &&
        report.viz().isShelfExpressionEmpty(expr)
      ) {
        validateExpression({ type: 'STRING', hasagg: false, ispart: false });
      } else if (arc.misc.isWindowFunction(expr)) {
        validateExpression({
          type: 'STRING',
          hasagg: false,
          ispart: false,
          info:
            'Analytic functions cannot be properly validated. Please refresh visual to see ' +
            'any error messages.',
        });
      } else {
        arc.validate_expression(expr, dataset, validateExpression, 'column');
      }
    };

  const paintFn = (
    containerWrapper: d3.Selection<HTMLElement, unknown, HTMLElement, unknown>
  ) => {
    const containerId = setting.div_id();
    containerWrapper.append('div').attr('id', containerId);
    const $container = $(`#${containerId}`);

    const $selectedFilter = $('.selected-list > li.selected .selection_field');
    const d3SelectedFilter = d3.selectAll($selectedFilter.toArray());
    selectedFilter = d3SelectedFilter.datum() as ExpressionFilter;

    const expressionFieldId = `${containerId}-expression-field`;
    const expressionEditorTemplate =
      getExpressionEditorTemplate(expressionFieldId);
    $container.append(expressionEditorTemplate);

    const report = arc.report.getPrimaryReport();
    const dataset = report.dataset();

    const $validateButton = $container.find('.expression-validate');
    $validateButton.parent().find('.alert').remove();
    $validateButton.offOn('click', getOnValidate(report, dataset));

    expressionEditor = arc.misc.expressionEditor(
      dataset,
      undefined,
      expressionFieldId,
      $container,
      { includeCalculated: true },
      true // enable resize
    );
    expressionEditor.value(selectedFilter.getExpressionForEdit());
  };

  setting.paint_fn(paintFn).fetch_fn(fetchFn);

  return setting;
};
