import moment, { MomentInput } from 'moment';
import { getGlobalArc } from 'utils/get-global-arc';
import type { EditorWrapper } from '../types/editor-wrapper.type';
import { format } from 'sql-formatter';

const arc = getGlobalArc();
export const moveCursorToEnd = (
  $el: JQuery<HTMLElement>,
  cssSelector?: string
): void => {
  $el.trigger('focus');
  const el = cssSelector ? $el.find(cssSelector)[0] : $el[0];
  if (
    typeof window.getSelection !== 'undefined' &&
    typeof document.createRange !== 'undefined'
  ) {
    const range = document.createRange();
    range.selectNodeContents(el);
    range.collapse(false);
    const sel = window.getSelection();
    if (sel) {
      sel.removeAllRanges();
      sel.addRange(range);
    }
    // @ts-ignore: It's IE support: no createTextRange in the HTML5 API for modern browsers
  } else if (typeof document.body.createTextRange !== 'undefined') {
    // @ts-ignore: Same as above
    const textRange = document.body.createTextRange();
    textRange.moveToElementText(el);
    textRange.collapse(false);
    textRange.select();
  }
};

export const addLimitToSQLQuery = (query: string, limit: number): string => {
  if (!query) {
    throw new Error(
      `The provided query (${query}) is invalid. Please, provide a valid SELECT query.`
    );
  }

  const segments: Array<string> = query.trim().split(/\s+/);
  const lowercaseSegments = segments.map((s) => s.toLowerCase());
  let isOffsetAtTheEnd = false;
  let containsLimitNearTheEnd = false;

  const isLowerCase = lowercaseSegments[0] === segments[0];
  const startsWithSelect = lowercaseSegments[0] === 'select';
  const startsWithWith = lowercaseSegments[0] === 'with';
  const containsSelectKeyword = lowercaseSegments.includes('select');
  if (startsWithSelect || (startsWithWith && containsSelectKeyword)) {
    /*
  I am trying to catch the most common scenarios. It is hard to prepare
  for all the possible ones without parsing properly the SQL.
  Usually limit, offset are the last items on the end, but in different connections it can vary.
  */

    if (segments.length > 5) {
      containsLimitNearTheEnd = lowercaseSegments
        .slice(lowercaseSegments.length - 5)
        .includes('limit');
      isOffsetAtTheEnd = lowercaseSegments
        .slice(lowercaseSegments.length - 3)
        .includes('offset');
    }

    const limitStatement = isLowerCase ? `limit ${limit}` : `LIMIT ${limit}`;
    if (!containsLimitNearTheEnd) {
      if (!isOffsetAtTheEnd) {
        segments.push(limitStatement);
      } else {
        const lastOffsetIndex = lowercaseSegments.lastIndexOf('offset');
        segments.splice(lastOffsetIndex, 0, limitStatement);
      }
    }
    return segments.join(' ');
  }

  return query;
};

export const setCommandEnterShortcutOnExpressionEditor = (
  expressionEditor: EditorWrapper,
  callback: () => void
): void => {
  expressionEditor.aceEditor().commands.addCommand({
    name: 'runCommandQuery',
    bindKey: { win: 'Ctrl-Enter', mac: 'Command-Enter' },
    exec() {
      callback();
    },
  });
};

/**
 * In case CDV is embedded (in iframe currently) opening link in new tab
 * leads to losing the embedding CML context. The solution is to replace the link's default domain
 * in this scenario.
 */
function setDomainToEmbeddingDomain(url: string) {
  const embedUrlPrefix = arc.frontend.globals.settings.embedUrlPrefix;
  if (window.location !== window.parent.location && embedUrlPrefix) {
    url = url.replace(window.location.origin, `${embedUrlPrefix}`);
    if (!url?.includes(embedUrlPrefix)) {
      url = `${embedUrlPrefix}${encodeURIComponent(url)}`;
    }
  }

  return url;
}

/**
 * In case CDV is embedded (in iframe currently) opening link in new tab
 * leads to losing the embedding CML context. The solution is to extend the links which opens new tab
 * or window with the absolute path in this scenario
 */
export function getBlankTargetUrl(url: string, target: string = '_blank') {
  if (arc.frontend.globals.settings.embedUrlPrefix && target === '_blank') {
    url = setDomainToEmbeddingDomain(url);
  }

  return url;
}

export function openInNewWindow(url: string, urlTarget = '_blank'): void {
  let openUrl = url;

  // If in CMLExplore iframe, we should pre-pend parent url
  if (
    arc.frontend.featureFlipper.getValueById('enable_explore_style') &&
    window.location !== window.parent.location &&
    urlTarget !== '_self'
  ) {
    openUrl = getBlankTargetUrl(url);
  }

  window.open(openUrl, urlTarget);
}

export const howToSetWorkloadPasswordText = `
<div class="%POPOVER_ID%">
  <ol>
    <li>In the control panel, add a workload password on your profile page</li>
    <li>In your Cloudera AI User Settings add the password to "WORKLOAD_PASSWORD" under the "Environment Variables" tab</li>
    <li>Restart the %PRODUCT_NAME% application</li>
  </ol>
</div>`;

export const activeAlertContainer = ($element: JQuery<HTMLElement>) => {
  resetAlerts($element);
  $element.find('.alert-container').addClass('active');
};

export const deactiveAlertContainer = ($element: JQuery<HTMLElement>) => {
  resetAlerts($element);
  $element.find('.alert-container').removeClass('active');
};

export const handleAlertContainerActivationOnModal = (
  $element: JQuery<HTMLElement>
) => {
  $element.offOn('show.bs.modal', () => {
    activeAlertContainer($element);
  });

  $element.offOn('hide.bs.modal', () => {
    deactiveAlertContainer($element);
  });
};

export const resetAlerts = function ($element?: JQuery<HTMLElement>) {
  const $alert = $element
    ? $element.find(`.alert-container.active`)
    : $(`.alert-container.active`);
  $alert.hide();
  $alert.find('.alert-number-errors').text(0);
  $alert.find('.alert-plural').empty();
  $alert.find('.alert-error-messages').empty();
};

export const copyToClipboard = async (text: string) => {
  try {
    return await navigator.clipboard.writeText(text);
  } catch {
    const el = document.createElement('textarea');
    el.value = text;
    el.setAttribute('readonly', '');

    el.style.position = 'absolute';
    el.style.left = '-9999px';

    document.body.appendChild(el);
    el.select();
    document.execCommand('copy');

    document.body.removeChild(el);
  }
};

type EscapedTags = {
  '&': string;
  '<': string;
  '>': string;
  '"': string;
  "'": string;
  ';': string;
  '(': string;
  ')': string;
  '[': string;
  ']': string;
};

const esc: EscapedTags = {
  '&': '&amp;',
  '<': '&lt;',
  '>': '&gt;',
  '"': '&quot;',
  "'": '&#39;',
  ';': '&#59;',
  '(': '&#40;',
  ')': '&#41;',
  '[': '&#91;',
  ']': '&#93;',
};

export const escapeHTML = (
  text: string,
  tags: Record<string, string> = esc
): string => {
  if (typeof text !== 'string') {
    return text;
  }

  let escaped: string[] = [];

  for (let i = 0; i < text.length; i++) {
    const char = text[i];
    escaped.push(tags[char] || char);
  }

  return escaped.join('');
};

export const formattingQuery = (
  text: string,
  language:
    | 'mysql'
    | 'postgresql'
    | 'sqlite'
    | 'mariadb'
    | 'snowflake'
    | 'bigquery'
    | 'hive'
    | 'redshift'
    | 'spark'
    | 'plsql'
    | 'transactsql'
    | 'trino'
    | 'tsql' = 'mysql'
) => {
  try {
    const parsedJson = JSON.parse(text);
    return JSON.stringify(parsedJson, null, 2);
  } catch (error) {
    try {
      return format(text, { language: language });
    } catch {
      return text;
    }
  }
};

export const isBrowserInternetExplorer = () => {
  // https://www.geeksforgeeks.org/how-to-detect-the-user-browser-safari-chrome-ie-firefox-and-opera-using-javascript/
  const ua = window.navigator.userAgent;
  const isIeOld = ua.includes('MSIE');
  const isIE11 = !!ua.match(/Trident.*rv:11\./);
  return isIeOld || isIE11;
};

export const formatDate = (dateTimeInput: MomentInput): string => {
  const date = moment(dateTimeInput);
  if (!date.isValid()) {
    throw new Error('Invalid date');
  }
  return date.format('MMM DD YYYY, h:mm A');
};

export const determineDatabaseInQuery = (query?: string) => {
  if (!query) {
    return;
  }

  const regexNormal = /from\s+([\w\d]+)\./i;
  const regexForSpacedColumnNames = /from\s+['"`]([ \w\d]+)\./i;
  const database =
    regexNormal.exec(query)?.[1] || regexForSpacedColumnNames.exec(query)?.[1];
  return database;
};

export const directAccessTypes = [
  'arcengine',
  'arcengine_v2',
  'impala',
  'impyla',
  'hive',
  'hyve',
  'sqlite',
  'vertica',
  'ksql',
  'snowflake',
  'duckdb',
  'druid',
  'mysql',
  'mariadb',
  'postgres',
  'oracle',
];

const DATE_REGEX =
  /(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3})\d{0,3}([-]\d{2}:\d{2})/;

export const fromNow = (
  dateString: string,
  customTxt = 'Invalid',
  alwaysInPast = true
) => {
  if (dateString) {
    const truncateMicrosecs = dateString.replace(DATE_REGEX, '$1$2');
    const m = moment(truncateMicrosecs, 'YYYY-MM-DDTHH:mm:ss.SSSZZ');
    const isValidDate = m.isValid();

    // If the date is always supposed to be in past, and the saved date is somehow
    // in the future, just assume it is a couple seconds off
    const currentDate = new Date();
    if (alwaysInPast && isValidDate && m.toDate() > currentDate) {
      const timeDiff = (m.toDate().getTime() - currentDate.getTime()) / 1000;
      console.log(
        'Timezone or computer/server issue. ' +
          `Off by ${timeDiff} seconds (~${Math.ceil(timeDiff / 3600)} hour(s))`
      );
      return 'a few seconds ago';
    }
    return isValidDate ? m.fromNow() : customTxt;
  } else {
    return customTxt;
  }
};
