import { getGlobalArc } from 'utils/get-global-arc';
import { openInNewWindow } from './misc';

const arc = getGlobalArc();

declare global {
  interface Window {
    SpeechRecognition: SpeechRecognition;
    webkitSpeechRecognition?: SpeechRecognition;
  }
}

interface speechGrammar {
  src: string;
  weight: number;
}

/*
  This function expects an $element in the format:
    <div>
      <span class="arc-speech-trigger"><span class="fa fa-microphone"></span></span>
      <div class="arc-speech-text"></div>
    </div>
   */
export function setupSpeechDetection(
  $element: JQuery<HTMLElement>,
  endCb: (finalText?: string) => void,
  disabledCb?: () => void,
  grammar?: speechGrammar[]
): (() => void) | void {
  const $trigger = $element.find('.arc-speech-trigger');
  if (!arc.frontend.featureFlipper.getValueById('enable_speech_recognition')) {
    $trigger.hide();
    return;
  }

  function setupUnsupportedTooltip() {
    if (!$trigger.data('bs.tooltip')) {
      $trigger.attr('disabled', 'disabled');
      $trigger.tooltip({
        title: 'Current browser does not support speech recognition',
        delay: {
          show: 200,
        },
      });

      if (disabledCb) {
        disabledCb();
      }
    }
  }

  let finalTranscript = '';
  let isRecognizing = false;
  let tooltip = '';
  let startTime = new Date();
  let recognizer: SpeechRecognition | null = null;
  const $speech = $element.find('.arc-speech-text');

  const resultHandler = (speechEvent: SpeechRecognitionEvent) => {
    if (typeof speechEvent.results === 'undefined') {
      recognizer?.stop();
      setupUnsupportedTooltip();
      return;
    }

    let transcript = '';
    let isFinal = false;
    for (let idx = 0; idx < speechEvent.results.length; idx += 1) {
      const result = speechEvent.results[idx];
      transcript += result[0].transcript;
      isFinal = isFinal || result.isFinal;
    }

    if ($speech.is('input')) {
      $speech.val(transcript);
    } else {
      $speech.text(transcript);
    }

    if (isFinal) {
      finalTranscript = transcript;
    }
  };

  const errorHandler = (speechEvent: SpeechRecognitionErrorEvent) => {
    isRecognizing = false;
    $trigger.addClass('speech-error');
    if (speechEvent.error === 'no-speech') {
      tooltip =
        'No speech detected; you might need to adjust your microphone settings';
    } else if (speechEvent.error === 'audio-capture') {
      tooltip = 'No microphone found; ensure that a microphone is installed';
    } else if (speechEvent.error === 'not-allowed') {
      if (speechEvent.timeStamp - startTime.valueOf() < 100) {
        tooltip = 'Permission to use microphone was blocked';
      } else {
        tooltip = 'Permission to use microphone was denied';
      }
    }

    $trigger.attr('data-original-title', tooltip);
  };

  const endHandler = () => {
    $trigger.removeClass('active-speech');
    $speech.removeClass('speech-in-progress');
    endCb(finalTranscript);
    isRecognizing = false;
    finalTranscript = '';
    tooltip = '';
  };

  window.SpeechRecognition =
    window.SpeechRecognition || window.webkitSpeechRecognition;
  if (SpeechRecognition) {
    recognizer = new SpeechRecognition();

    // Set up initial settings
    recognizer.interimResults = true;
    recognizer.lang = 'en-US';

    // Set up grammars
    if (grammar && SpeechGrammarList) {
      const grammarList = new SpeechGrammarList();
      grammar.forEach((list: speechGrammar) => {
        grammarList.addFromString(list.src, list.weight);
      });

      recognizer.grammars = grammarList;
    }

    // Set up speech recognition
    // Put results in .arc-speech-text
    recognizer.addEventListener('result', resultHandler);

    // Show tooltip and error
    // https://www.google.com/intl/en/chrome/demos/speech.html

    recognizer.onerror = errorHandler;

    if (!$trigger.data('bs.tooltip')) {
      $trigger.tooltip({
        delay: {
          show: 200,
        },
      });
    }

    // Allow user to press trigger again
    recognizer.onend = endHandler;

    // Start recognizing speech
    $trigger.off('click.triggerSpeech').on('click.triggerSpeech', (e) => {
      e.preventDefault();
      e.stopPropagation();
      if (isRecognizing) return;
      isRecognizing = true;
      $trigger.addClass('active-speech');

      $trigger.removeClass('speech-error');
      $trigger.tooltip('hide');
      $trigger.removeAttr('data-original-title');

      $speech.addClass('speech-in-progress');
      recognizer?.start();
      startTime = new Date();
    });
  } else {
    setupUnsupportedTooltip();
  }

  return () => {
    if (recognizer) {
      recognizer.stop();
      recognizer.removeEventListener('result', resultHandler);
      recognizer.onerror = null;
      recognizer.onend = null;
    }
  };
}

function runGlobalSearch(searchText: string) {
  const searchTextLower = searchText.toLowerCase().trim();
  if (!searchTextLower) return;

  const isCreate = searchTextLower.startsWith('create');
  const isFind = searchTextLower.startsWith('find');
  const APPGROUP_ALL = -1;

  // Take to create app, connection, dataset, or dashboard
  if (isCreate && searchText.includes('app')) {
    openInNewWindow(arc.frontend.globals.urls.appGroupEditUrl);
  } else if (isCreate && searchTextLower.includes('connection')) {
    openInNewWindow(`${arc.frontend.globals.urls.dataUrl}#connection-modal`);
  } else if (isCreate && searchTextLower.includes('dataset')) {
    openInNewWindow(`${arc.frontend.globals.urls.dataUrl}#dataset-modal`);
  } else if (isCreate && searchTextLower.includes('dash')) {
    openInNewWindow(
      `${arc.frontend.globals.urls.reportBuilderUrl}/0?report_type=dashboard`
    );

    // Take to sql page
  } else if (
    (isCreate && searchTextLower.includes('sql')) ||
    searchTextLower.includes('run sql')
  ) {
    openInNewWindow(`${arc.frontend.globals.urls.sqlUrl}#sql-tab`);

    // Take to search within CDV
  } else if (isFind && searchTextLower.includes('app')) {
    const searchString = decodeURIComponent(
      searchText
        .replace(/^find/, '')
        .replace(/app(lication)?s?/, '')
        .trim()
    );
    openInNewWindow(
      `${arc.frontend.globals.urls.appsUrl}?displaymode=apps&setappgroup=${APPGROUP_ALL}&searchbox=${searchString}`
    );
  } else if (isFind && searchTextLower.includes('dataset')) {
    const searchString = decodeURIComponent(
      searchText
        .replace(/^find/, '')
        .replace(/data(set)?s?/, '')
        .trim()
    );
    openInNewWindow(
      `${arc.frontend.globals.urls.dataUrl}?dataconnection_id=0&searchbox=${searchString}`
    );
  } else if (isFind && searchTextLower.includes('dash')) {
    const searchString = decodeURIComponent(
      searchText
        .replace(/^find/, '')
        .replace(/dash(board)?s?/, '')
        .trim()
    );
    openInNewWindow(
      `${arc.frontend.globals.urls.appsUrl}?displaymode=dashboards&setappgroup=${APPGROUP_ALL}&searchbox=${searchString}`
    );

    // Take to doc search
  } else {
    let searchItem = searchTextLower.replace(
      /^search\s+[docs|documentation]?\s*[for|about]?/i,
      ''
    );

    searchItem = searchItem.replace(/\s+/g, '+');
    openInNewWindow(
      `https://www.google.com/search?q=${encodeURIComponent(
        searchItem
      )}+site%3A${encodeURIComponent(
        arc.frontend.globals.documentation.baseUrl
      )}`
    );
  }
}

export function setupNavbarSearch($searchElement: JQuery<HTMLElement>): void {
  const $searchInput = $searchElement.find('input');
  $searchElement
    .find('.pop-search-remove')
    .off('click.clearField')
    .on('click.clearField', (e) => {
      e.stopPropagation();
      e.preventDefault();
      $searchInput.val('');
    });

  $searchInput
    .off('keypress.triggerSearch')
    .on('keypress.triggerSearch', (e) => {
      if (e.key === 'Enter') {
        e.stopPropagation();
        e.preventDefault();
        const value = $searchInput.val();
        if (typeof value === 'string') {
          runGlobalSearch(value);
        }
      }
    });

  setupSpeechDetection($searchElement, (finalText) => {
    finalText = (finalText || '').toLowerCase().trim();
    runGlobalSearch(finalText);
  });
}
