import * as React from 'react';
import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import {
  mockQueryLogs,
  mockDataconnections,
  mockSummaryDatasets,
} from './utils/mocks';
import { getQueryLogs } from 'apis/get-query-logs';
import { getAllDataConnections } from 'apis/get-all-data-connections';
import { getAllDatasets } from 'apis/get-all-datasets';
import { ActivityLogs } from './ActivityLogs';
import { frontendDownloadCSV } from 'utils/frontend-download-csv';
import { generateCsvData } from './utils/activity-logs-csv-data-generator';

configure({ testIdAttribute: 'data-test-id' });

const queryClient = new QueryClient({
  defaultOptions: {
    queries: {
      retry: false,
    },
  },
});

const mockAddNotification = jest.fn();

jest.mock('../../utils/frontend-download-csv.ts', () => ({
  frontendDownloadCSV: jest.fn(),
}));

jest.mock(
  '../../../src/components/activity-logs/utils/activity-logs-csv-data-generator.ts',
  () => ({
    generateCsvData: jest.fn(),
  })
);

jest.mock('../../../src/stores/use-notification-store.ts', () => ({
  useNotificationStore: () => ({ addNotification: mockAddNotification }),
}));

jest.mock('../../apis/get-query-logs.ts', () => ({
  getQueryLogs: jest.fn(),
}));

jest.mock('../../apis/get-all-data-connections.ts', () => ({
  getAllDataConnections: jest.fn(),
}));

jest.mock('../../apis/get-all-datasets.ts', () => ({
  getAllDatasets: jest.fn(),
}));

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: {
          viewDashboardUrl: 'customUrl',
        },
        settings: {
          isArcengineDataconnectionEnable: true,
        },
      },
      featureFlipper: {
        getValueById: jest.fn(),
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

jest.mock('../../components/navigation-bar/NavigationBar', () => ({
  NavigationBar: () => null,
}));

jest.mock('../../components/settings-navigation/SettingsNavigation', () => ({
  SettingsNavigation: () => null,
}));

beforeEach(() => {
  jest.mocked(getQueryLogs).mockResolvedValue(mockQueryLogs);
  jest.mocked(getAllDataConnections).mockResolvedValue(mockDataconnections);
  jest.mocked(getAllDatasets).mockResolvedValue(mockSummaryDatasets);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('Activity Logs component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ActivityLogs />
      </QueryClientProvider>
    );
  });

  it('should display the activity logs table', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ActivityLogs />
      </QueryClientProvider>
    );

    const stateColumnHeader = screen.getByRole('button', { name: 'State' });
    expect(stateColumnHeader).toBeInTheDocument();

    const timeColumnHeader = screen.getByRole('button', { name: 'Time' });
    expect(timeColumnHeader).toBeInTheDocument();

    const userColumnHeader = screen.getByRole('button', { name: 'User' });
    expect(userColumnHeader).toBeInTheDocument();

    const ipColumnHeader = screen.getByRole('button', { name: 'IP' });
    expect(ipColumnHeader).toBeInTheDocument();

    const actionColumnHeader = screen.getByRole('button', { name: 'Action' });
    expect(actionColumnHeader).toBeInTheDocument();

    const componentTypeColumnHeader = screen.getByRole('button', {
      name: 'Component Type',
    });
    expect(componentTypeColumnHeader).toBeInTheDocument();

    const componentIDColumnHeader = screen.getByRole('button', {
      name: 'Component ID',
    });
    expect(componentIDColumnHeader).toBeInTheDocument();

    const visualColumnHeader = screen.getByRole('button', { name: 'Visual' });
    expect(visualColumnHeader).toBeInTheDocument();

    const dashboardColumnHeader = screen.getByRole('button', {
      name: 'Dashboard',
    });
    expect(dashboardColumnHeader).toBeInTheDocument();

    const datasetColumnHeader = screen.getByRole('button', { name: 'Dataset' });
    expect(datasetColumnHeader).toBeInTheDocument();

    const connectionColumnHeader = screen.getByRole('button', {
      name: 'Connection',
    });
    expect(connectionColumnHeader).toBeInTheDocument();

    const typeColumnHeader = screen.getByRole('button', { name: 'Type' });
    expect(typeColumnHeader).toBeInTheDocument();

    const fallbackColumnHeader = screen.getByRole('button', {
      name: 'Fallback',
    });
    expect(fallbackColumnHeader).toBeInTheDocument();

    const cacheColumnHeader = screen.getByRole('button', { name: 'Cache' });
    expect(cacheColumnHeader).toBeInTheDocument();

    const cacheCheckColumnHeader = screen.getByRole('button', {
      name: 'Cache Check',
    });
    expect(cacheCheckColumnHeader).toBeInTheDocument();

    const extractUsedColumnHeader = screen.getByRole('button', {
      name: 'Extract Used',
    });
    expect(extractUsedColumnHeader).toBeInTheDocument();

    const runtimeColumnHeader = screen.getByRole('button', { name: 'Runtime' });
    expect(runtimeColumnHeader).toBeInTheDocument();

    const queryColumnHeader = screen.getByRole('button', { name: 'Query' });
    expect(queryColumnHeader).toBeInTheDocument();
  });
});

describe('Auto-refetch checkbox', () => {
  beforeEach(() => {
    jest.clearAllMocks();
  });

  it('should refetch the queryLogs on auto-refetch checkbox', async () => {
    // await act(async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ActivityLogs />
      </QueryClientProvider>
    );
    // });

    const autoRefetchCheckbox = screen.getByRole('checkbox', {
      name: 'Auto-refresh queries (every 60 seconds)',
    });

    expect(autoRefetchCheckbox).toBeInTheDocument();

    // await act(async () => {
    fireEvent.click(autoRefetchCheckbox);
    // });

    expect(autoRefetchCheckbox).toBeChecked();

    await waitFor(() => {
      expect(getQueryLogs).toHaveBeenCalledWith('500', null, '', false, false);
    });

    await waitFor(() => {
      expect(getQueryLogs).toHaveBeenCalledTimes(1);
    });
  });
});

describe('Edits_only checkox', () => {
  it('should send a request out with the "edits_only=true" query string', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ActivityLogs />
      </QueryClientProvider>
    );

    const editsOnlyCheckbox = screen.getByRole('checkbox', {
      name: 'Only display logs that change state (no views)',
    });

    expect(editsOnlyCheckbox).toBeInTheDocument();

    jest.clearAllMocks();

    fireEvent.click(editsOnlyCheckbox);

    expect(editsOnlyCheckbox).toBeChecked();

    await waitFor(() => {
      expect(getQueryLogs).toHaveBeenCalledWith('500', null, '', true, false);
    });

    await waitFor(() => {
      expect(getQueryLogs).toHaveBeenCalledTimes(1);
    });
  });
});

describe('Views_only checkbox', () => {
  it('should send a request out with the "visuals_only=true" query string', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ActivityLogs />
      </QueryClientProvider>
    );

    const viewsOnlyCheckbox = screen.getByRole('checkbox', {
      name: 'Only display queries generated by visuals',
    });

    expect(viewsOnlyCheckbox).toBeInTheDocument();

    jest.clearAllMocks();

    fireEvent.click(viewsOnlyCheckbox);

    expect(viewsOnlyCheckbox).toBeChecked();

    await waitFor(() => {
      expect(getQueryLogs).toHaveBeenCalledWith('500', null, '', false, true);
    });

    await waitFor(() => {
      expect(getQueryLogs).toHaveBeenCalledTimes(1);
    });
  });
});

describe('Query Count input', () => {
  it('should decrement the query count when minus button is clicked', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ActivityLogs />
      </QueryClientProvider>
    );

    const queriesInput = screen.getByTestId(
      'query-count-input'
    ) as HTMLInputElement;

    expect(queriesInput.value).toBe('500');

    const minusButton = screen.getByTestId('query-count-minus');

    jest.clearAllMocks();

    fireEvent.click(minusButton);

    await waitFor(() => {
      expect(queriesInput.value).toBe('499');
    });

    await waitFor(() => {
      expect(getQueryLogs).toHaveBeenCalledTimes(1);
    });
  });

  it('should increment the query count when plus button is clicked', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ActivityLogs />
      </QueryClientProvider>
    );

    const queriesInput = screen.getByTestId(
      'query-count-input'
    ) as HTMLInputElement;

    expect(queriesInput.value).toBe('500');

    const plusButton = screen.getByTestId('query-count-plus');

    jest.clearAllMocks();

    fireEvent.click(plusButton);

    await waitFor(() => {
      expect(queriesInput.value).toBe('501');
    });

    await waitFor(() => {
      expect(getQueryLogs).toHaveBeenCalledTimes(1);
    });
  });

  it('should update the query count when typing into the queriesInput field', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ActivityLogs />
      </QueryClientProvider>
    );

    const queriesInput = screen.getByTestId(
      'query-count-input'
    ) as HTMLInputElement;

    expect(queriesInput.value).toBe('500');

    jest.clearAllMocks();

    fireEvent.change(queriesInput, { target: { value: '10' } });

    await waitFor(() => {
      expect(queriesInput.value).toBe('10');
    });

    await waitFor(() => {
      expect(getQueryLogs).toHaveBeenCalledTimes(1);
    });
  });
});

describe('Activity Logs Download button', () => {
  it('should call frontendDownloadCSV when the download button is clicked', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ActivityLogs />
      </QueryClientProvider>
    );

    const downloadButton = screen.getByText('DOWNLOAD');

    jest.clearAllMocks();

    fireEvent.click(downloadButton);

    const csvData = generateCsvData(mockQueryLogs);

    await waitFor(() => {
      expect(frontendDownloadCSV).toHaveBeenCalledWith(
        csvData,
        expect.stringContaining('Activities')
      );
    });

    expect(frontendDownloadCSV).toHaveBeenCalledTimes(1);
  });
});
