import React, { useCallback, useEffect, useMemo, useState } from 'react';
import { getQueryLogs, getQueryLogsQueryToken } from 'apis/get-query-logs';
import { useQuery } from '@tanstack/react-query';
import { useNotificationStore } from 'stores/use-notification-store';
import { Col, Form, Row } from 'react-bootstrap';
import moment from 'moment';
import RefreshIcon from '@cloudera-internal/cuix-core/icons/svg/ic-refresh.svg';
import { VizButton } from 'components/viz-button/VizButton';
import {
  getCoreRowModel,
  getSortedRowModel,
  getPaginationRowModel,
} from '@tanstack/react-table';
import { frontendDownloadCSV } from 'utils/frontend-download-csv';
import { ClouderaTable } from 'components/cloudera-table';
import { SearchBar } from 'pages/visuals-page/components/search-bar/SearchBar';

import { getGlobalArc } from 'utils/get-global-arc';
import { generateCsvData } from './utils/activity-logs-csv-data-generator';
import { ColumnConfig } from './components/ColumnConfig';
import { activityLogsStatistics } from './utils/activity-logs-statistics';
import {
  dataconnectionQueryToken,
  getAllDataConnections,
} from 'apis/get-all-data-connections';
import { datasetsQueryToken, getAllDatasets } from 'apis/get-all-datasets';
import { legacyBridge } from 'services/legacy-bridge';
import classes from './ActivityLogs.module.scss';
import classNames from 'classnames';
import { useTable } from 'hooks/use-table';

const arc = getGlobalArc();
const isArcengineEnabled =
  arc.frontend.globals.settings.isArcengineDataconnectionEnable;
const isSiteEnabledExtracts = arc.frontend.featureFlipper.getValueById(
  'site_enable_extracts'
);

const fetchFrequencyInSeconds = 60;
const DEFAULT_QUERY_COUNT = '500';

export const ActivityLogs = () => {
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));
  const [queryCount, setQueryCount] = useState(DEFAULT_QUERY_COUNT);
  const [debouncedQueryCount, setDebouncedQueryCount] =
    useState(DEFAULT_QUERY_COUNT);
  const [debouncedQueryCountTimeout, setDebouncedQueryCountTimeout] =
    useState<NodeJS.Timeout | null>(null);
  const setDebouncedQueryCountAfterTimeout = useCallback(
    (value: string) => {
      if (debouncedQueryCountTimeout) clearTimeout(debouncedQueryCountTimeout);

      const timeout = setTimeout(() => setDebouncedQueryCount(value), 1000);

      setDebouncedQueryCountTimeout(timeout);
    },
    [debouncedQueryCountTimeout]
  );

  const [isAutoRefreshEnabled, setIsAutoRefreshEnabled] = useState(false);
  const [globalFilter, setGlobalFilter] = useState('');
  const [isEditsOnly, setIsEditsOnly] = useState(false);
  const [isVisualsOnly, setIsVisualsOnly] = useState(false);
  const [visiblePopoverId, setVisiblePopoverId] = useState('');

  const [showStats, setShowStats] = useState(false);
  const {
    data: queryLogs,
    isFetching: isQueryLogsLoading,
    dataUpdatedAt: lastFetchedTime,
    refetch,
  } = useQuery({
    queryKey: [
      getQueryLogsQueryToken,
      debouncedQueryCount,
      globalFilter,
      isEditsOnly,
      isVisualsOnly,
    ],
    queryFn: () =>
      getQueryLogs(
        debouncedQueryCount,
        null,
        globalFilter,
        isEditsOnly,
        isVisualsOnly
      ),
    initialData: [],
    enabled: debouncedQueryCount !== '' && debouncedQueryCount !== '0',
    onError: () =>
      addNotification(
        {
          variant: 'danger',
          message: 'Failed to load activity logs',
        },
        { delay: 3000 }
      ),
    refetchInterval: isAutoRefreshEnabled
      ? fetchFrequencyInSeconds * 1000
      : false,
  });

  const { data: allDataConnections } = useQuery({
    queryKey: [dataconnectionQueryToken],
    queryFn: () => getAllDataConnections(),
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Error fetching data connections.',
        },
        { delay: 3000 }
      );
    },
    initialData: [],
    select: (dataconnections) => {
      return dataconnections.map(({ id, name, type, typeName }) => ({
        id,
        name,
        type,
        typeName,
      }));
    },
  });

  const { data: allDatasets } = useQuery({
    queryKey: [datasetsQueryToken],
    queryFn: () => getAllDatasets({ isSummary: false }),
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Error fetching datasets.',
        },
        { delay: 3000 }
      );
    },
    initialData: [],
    select: (datasets) => {
      const allDatasets: {
        id: number;
        name: string;
        dc_id: number;
        dc_name: string;
      }[] = [];

      datasets.forEach(({ id, name, dc_id, dc_name }) => {
        allDatasets.push({ id, name, dc_id, dc_name });
      });

      return allDatasets;
    },
  });

  const summaryActivityLogsStats = useMemo(() => {
    return activityLogsStatistics(queryLogs, allDataConnections, allDatasets);
  }, [queryLogs, allDataConnections, allDatasets]);

  const lastFetchedFirst = moment
    .utc(summaryActivityLogsStats?.firsttime)
    .local()
    .format('YYYY-MM-DD HH:mm');
  const lastFetchedLast = moment
    .utc(summaryActivityLogsStats?.lasttime)
    .local()
    .format('YYYY-MM-DD HH:mm');

  const handleTogglePopover = useCallback((id: string) => {
    setVisiblePopoverId(id);
  }, []);

  const csvData = useMemo(() => {
    return generateCsvData(queryLogs || []);
  }, [queryLogs]);

  const columns = useMemo(() => {
    return ColumnConfig(
      visiblePopoverId,
      handleTogglePopover,
      allDataConnections
    );
  }, [visiblePopoverId, handleTogglePopover, allDataConnections]);

  const table = useTable({
    data: queryLogs,
    columns,
    getCoreRowModel: getCoreRowModel(),
    getSortedRowModel: getSortedRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    onGlobalFilterChange: setGlobalFilter,
    state: {
      globalFilter,
      columnVisibility: {
        dataconnection: false,
        error: false,
        error_detail: false,
        dataconnection_fbid: isArcengineEnabled,
        cache_check: isArcengineEnabled,
        rewrite_target: isSiteEnabledExtracts || isArcengineEnabled,
      },
    },
  });

  const handleQueryCountChange = (
    event: React.ChangeEvent<HTMLInputElement>
  ) => {
    const newCount = event.target.value;
    setQueryCount(newCount);
    setDebouncedQueryCountAfterTimeout(newCount);
  };

  useEffect(() => {
    if (summaryActivityLogsStats) {
      legacyBridge.events.next({
        type: 'send-activity-logs-stats',
        payload: {
          statistics: summaryActivityLogsStats,
          allDataConnection: allDataConnections,
        },
      });
    }
  }, [summaryActivityLogsStats, allDataConnections]);

  return (
    <>
      <div className="d-flex justify-content-between">
        <div>
          <Form.Check
            type="checkbox"
            id="autoQueries"
            label={`Auto-refresh queries (every ${fetchFrequencyInSeconds} seconds)`}
            checked={isAutoRefreshEnabled}
            onChange={(e) => setIsAutoRefreshEnabled(e.target.checked)}
          />
          <Form.Check
            type="checkbox"
            id="displayLogsChangeState"
            label="Only display logs that change state (no views)"
            checked={isEditsOnly}
            onChange={(e) => setIsEditsOnly(e.target.checked)}
          />
          <Form.Check
            type="checkbox"
            id="displayQueriesVisuals"
            label="Only display queries generated by visuals"
            checked={isVisualsOnly}
            onChange={(e) => setIsVisualsOnly(e.target.checked)}
          />
        </div>
        <div>
          <div>
            Last fetch:{' '}
            {moment.utc(lastFetchedTime).local().format('YYYY-MM-DD HH:mm')}
          </div>
          <Form.Group
            controlId="fetchedQueriesCount"
            className="d-flex align-items-center mb-0"
          >
            <Form.Label className="me-2 mb-0 mr-1">Queries:</Form.Label>
            <VizButton
              size="small"
              color="secondary"
              onClick={() =>
                setQueryCount((prev) => {
                  const newCount = String(Math.max(Number(prev) - 1, 1));
                  setDebouncedQueryCountAfterTimeout(newCount);
                  return newCount;
                })
              }
              className="py-1 px-2"
              testId="query-count-minus"
            >
              <i className="fa fa-minus"></i>
            </VizButton>

            <Form.Control
              type="text"
              value={queryCount}
              onChange={handleQueryCountChange}
              className={classes.queryCountInput}
              data-test-id="query-count-input"
            />

            <VizButton
              size="small"
              color="secondary"
              onClick={() =>
                setQueryCount((prev) => {
                  const newCount = String(Number(prev) + 1);
                  setDebouncedQueryCountAfterTimeout(newCount);
                  return newCount;
                })
              }
              className="py-1 px-2"
              testId="query-count-plus"
            >
              <i className="fa fa-plus"></i>
            </VizButton>
          </Form.Group>
          <div>
            Time Span: {lastFetchedFirst} to {lastFetchedLast}
          </div>
        </div>
      </div>
      <div className="d-flex mb-4 gap-1">
        <VizButton
          size="medium"
          color="secondary"
          onClick={() => setShowStats((prevShowStats) => !prevShowStats)}
        >
          {showStats ? 'HIDE STATS' : 'SHOW STATS'}
        </VizButton>
        <VizButton size="medium" color="secondary" onClick={() => refetch()}>
          <RefreshIcon /> REFRESH
        </VizButton>
      </div>

      <div className="d-flex mb-4">
        <div
          className={classNames(classes.activitiesContainer, {
            [classes.active]: showStats,
          })}
        >
          <div className={classes.vizAppsQueriesWide} id="bar-by-app"></div>
          <div className={classes.vizAppsQueriesWide} id="bar-by-action"></div>
          <div
            className={classes.vizAppsQueriesWide}
            id="bar-by-cache-statistic"
          ></div>
          <div className={classes.vizAppsQueriesNarrow} id="bar-by-user"></div>
          <div className={classes.vizAppsQueriesWide} id="bar-by-dataset"></div>
          <div className={classes.vizAppsQueriesWide} id="bar-by-ip"></div>
          <div
            className={classes.vizAppsQueriesNarrow}
            id="bar-by-connection"
          ></div>
          <div
            className={classes.vizAppsQueriesNarrow}
            id="bar-by-runtime"
          ></div>
        </div>
      </div>

      <Row className="mb-3">
        <Col md={11}>
          <SearchBar
            id="search-query"
            placeholder="Search"
            testId="searchbox-query"
            searchTerm={globalFilter}
            setSearchTerm={setGlobalFilter}
            debounce={500}
          />
        </Col>
        <Col md={1} className="text-right">
          <VizButton
            size="medium"
            color="secondary"
            onClick={() =>
              frontendDownloadCSV(
                csvData,
                `Activities ${new Date()
                  .toISOString()
                  .split('T')[0]
                  .replace(/-/g, '')}`
              )
            }
          >
            DOWNLOAD
          </VizButton>
        </Col>
      </Row>
      <ClouderaTable
        isLoading={isQueryLogsLoading}
        table={table}
        className="query-logs-table"
        data-test-id="queryLogsTable"
      />
    </>
  );
};
