import React from 'react';
import moment from 'moment';
import ErrorPopover from './ErrorPopover';
import { ActivityLogsStatusIndicator } from './ActivityLogsStatusIndicator';
import {
  ActivityLogActions,
  ActivityLogComponentTypes,
} from '../types/activity-logs.type';
import { getBlankTargetUrl } from 'arc/misc/misc';
import QueryPopover from './QueryPopover';

import { createColumnHelper } from '@tanstack/react-table';

import { getGlobalArc } from 'utils/get-global-arc';
import { QueryLog } from 'apis/get-query-logs';
import { DataConnectionType } from 'types/data-connection.type';
import classes from '../ActivityLogs.module.scss';

const arc = getGlobalArc();

const columnHelper = createColumnHelper<QueryLog>();

export const ColumnConfig = (
  visiblePopoverId: string,
  handleTogglePopover: (id: string) => void,
  allDataConnections: {
    id: number;
    name: string;
    type: DataConnectionType;
    typeName: string;
  }[]
) => {
  const typeNameMapping = allDataConnections.reduce((acc, connection) => {
    acc[connection.type] = connection.typeName;
    return acc;
  }, {} as Record<DataConnectionType, string>);

  return [
    columnHelper.accessor('query_state', {
      header: () => 'State',
      cell: (info) => {
        const error = Boolean(info.row.getValue('error'));

        return (
          <div>
            {error ? (
              <ErrorPopover
                rowId={info.row.id}
                errorDetail={info.row.getValue('error_detail')}
                visiblePopoverId={visiblePopoverId}
                onToggle={handleTogglePopover}
              />
            ) : (
              <ActivityLogsStatusIndicator
                status={info.getValue()}
                isAnimated
              />
            )}
          </div>
        );
      },
    }),
    columnHelper.accessor('ts', {
      header: () => 'Time',
      cell: (info) => {
        const ts = moment
          .utc(info.getValue())
          .local()
          .format('YYYY-MM-DD HH:mm');

        return <span className={classes.noWrap}>{ts}</span>;
      },
    }),
    columnHelper.accessor('username', {
      header: () => 'User',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('ip', {
      header: () => 'IP',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('action', {
      header: () => 'Action',
      cell: (info) => ActivityLogActions[info.getValue()],
    }),
    columnHelper.accessor('component_type', {
      header: () => 'Component Type',
      cell: (info) => ActivityLogComponentTypes[info.getValue()],
    }),
    columnHelper.accessor('component_id', {
      header: () => 'Component ID',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('visual', {
      header: () => 'Visual',
      cell: (info) => {
        const visualId = info.getValue();
        return (
          <a
            href={getBlankTargetUrl(
              `${arc.frontend.globals.urls.viewDashboardUrl}/${visualId}`
            )}
            target="_blank"
            rel="noreferrer"
          >
            {visualId}
          </a>
        );
      },
    }),
    columnHelper.accessor('app', {
      header: () => 'Dashboard',
      cell: (info) => {
        const dashboardId = info.getValue();
        return (
          <a
            href={getBlankTargetUrl(
              `${arc.frontend.globals.urls.viewDashboardUrl}/${dashboardId}`
            )}
            target="_blank"
            rel="noreferrer"
          >
            {dashboardId}
          </a>
        );
      },
    }),
    columnHelper.accessor('dataset', {
      header: () => 'Dataset',
      cell: (info) => {
        const datasetId = info.getValue();
        return (
          <a
            href={getBlankTargetUrl(
              `${arc.frontend.globals.urls.datasetEditUrl}/${datasetId}`
            )}
            target="_blank"
            rel="noreferrer"
          >
            {datasetId}
          </a>
        );
      },
    }),
    columnHelper.accessor('dataconnection', {
      header: () => 'Dataconnection',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('dataconnection_name', {
      header: () => 'Connection',
      cell: (info) => {
        const dataconnectionId = info.row.getValue('dataconnection');
        return (
          <a
            href={getBlankTargetUrl(
              `${arc.frontend.globals.urls.dataUrl}?dataconnection_id=${dataconnectionId}`
            )}
            target="_blank"
            rel="noreferrer"
          >
            {info.getValue()}
          </a>
        );
      },
    }),
    columnHelper.accessor('dataconnection_type', {
      header: () => 'Type',
      cell: (info) => {
        const dataconnectionType = info.getValue();
        return typeNameMapping[dataconnectionType] || dataconnectionType;
      },
    }),
    columnHelper.accessor('dataconnection_fbid', {
      header: () => 'Fallback',
      cell: (info) => {
        const fallbackId = info.getValue();
        if (fallbackId) {
          return (
            <a
              href={getBlankTargetUrl(
                `${arc.frontend.globals.urls.dataUrl}?dataconnection_id=${fallbackId}`
              )}
              target="_blank"
              rel="noreferrer"
            >
              {info.getValue()}
            </a>
          );
        }
        return 'N/A';
      },
    }),
    columnHelper.accessor('cache', {
      header: () => 'Cache',
      cell: (info) => {
        const cache = info.getValue() ? 'true' : 'false';
        return cache;
      },
    }),
    columnHelper.accessor('cache_check', {
      header: () => 'Cache Check',
      cell: (info) => {
        const cacheCheck = info.getValue() ? 'true' : 'false';
        return cacheCheck;
      },
    }),
    columnHelper.accessor('rewrite_target', {
      header: () => 'Extract Used',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('runtime', {
      header: () => 'Runtime',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('error', {
      header: () => 'Error',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('error_detail', {
      header: () => 'Error Detail',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('query', {
      header: () => 'Query',
      cell: (info) => {
        const queryText = info.getValue();
        return queryText ? (
          <QueryPopover
            rowId={info.row.id}
            visiblePopoverId={visiblePopoverId}
            onToggle={handleTogglePopover}
            queryText={queryText}
          />
        ) : (
          ''
        );
      },
    }),
  ];
};
