import { QueryLog } from 'apis/get-query-logs';
import moment from 'moment';
import {
  ActivityLogActions,
  ActivityLogComponentTypes,
  Statistics,
} from '../types/activity-logs.type';
import { DataConnectionType } from 'types/data-connection.type';

const timeIntervals = {
  s2: '0-2 s',
  s5: '2-5 s',
  s10: '5-10 s',
  s30: '10-30 s',
  s60: '30-60 s',
  s300: '60-300 s',
  slarge: '>300 s',
};

const metricColumnIdx = 1;

export const activityLogsStatistics = (
  queryLogs: QueryLog[],
  allDataConnections: {
    id: number;
    name: string;
    type: DataConnectionType;
    typeName: string;
  }[],
  allDatasets: { id: number; name: string; dc_id: number; dc_name: string }[]
): Statistics => {
  let statistics: Statistics = {
    aggAction: {},
    aggUsername: {},
    aggIp: {},
    aggConnection: {},
    aggVisual: {},
    aggApp: {},
    aggDataset: {},
    aggRuntime: Object.fromEntries(
      Object.values(timeIntervals).map((value) => [value, 0])
    ),
    aggCache: {
      Yes: 0,
      No: 0,
    },
    firsttime: undefined,
    lasttime: undefined,
    aggCacheStat: {},
  };

  const computeStats = (query: QueryLog) => {
    if (query.runtime && !isNaN(Number(query.runtime))) {
      const runtime = Number(query.runtime);
      if (runtime <= 2) statistics.aggRuntime[timeIntervals.s2] += 1;
      else if (runtime <= 5) statistics.aggRuntime[timeIntervals.s5] += 1;
      else if (runtime <= 10) statistics.aggRuntime[timeIntervals.s10] += 1;
      else if (runtime <= 30) statistics.aggRuntime[timeIntervals.s30] += 1;
      else if (runtime <= 60) statistics.aggRuntime[timeIntervals.s60] += 1;
      else if (runtime <= 300) statistics.aggRuntime[timeIntervals.s300] += 1;
      else statistics.aggRuntime[timeIntervals.slarge] += 1;
    }

    if (query.action) {
      const actionKey = [
        ActivityLogActions[query.action],
        ActivityLogComponentTypes[query.component_type],
      ].join(',');

      if (actionKey in statistics.aggAction) {
        statistics.aggAction[actionKey][metricColumnIdx] += 1;
      } else {
        statistics.aggAction[actionKey] = [
          ActivityLogActions[query.action],
          1,
          ActivityLogComponentTypes[query.component_type],
        ];
      }
    }

    if (query.username) {
      statistics.aggUsername[query.username] =
        (statistics.aggUsername[query.username] || 0) + 1;
    }

    if (query.ip) {
      statistics.aggIp[query.ip] = (statistics.aggIp[query.ip] || 0) + 1;
    }

    const connectionId = query.dataconnection;
    const connectionNameOrId =
      allDataConnections.find((connection) => connection.id === connectionId)
        ?.name || connectionId;

    if (query.dataconnection) {
      statistics.aggConnection[connectionNameOrId] =
        (statistics.aggConnection[connectionNameOrId] || 0) + 1;
    }

    if (query.cache) statistics.aggCache.Yes += 1;
    else statistics.aggCache.No += 1;

    if (
      query.cache !== undefined &&
      connectionNameOrId &&
      connectionNameOrId !== ''
    ) {
      const fromCacheId = `${connectionNameOrId},from cache`;
      const fromBackendId = `${connectionNameOrId},from backend`;
      const cacheStatId = query.cache ? fromCacheId : fromBackendId;

      if (
        !statistics.aggCacheStat[fromBackendId] &&
        !statistics.aggCacheStat[fromCacheId]
      ) {
        statistics.aggCacheStat[fromCacheId] = [connectionNameOrId, 0, 'Cache'];
        statistics.aggCacheStat[fromBackendId] = [
          connectionNameOrId,
          0,
          'Backend',
        ];
      }

      statistics.aggCacheStat[cacheStatId][1] += 1;
    }

    let dsid;
    if (query.dataset) {
      dsid = query.dataset;

      if (query.dataset in statistics.aggDataset) {
        statistics.aggDataset[query.dataset][metricColumnIdx] += 1;
      } else {
        statistics.aggDataset[query.dataset] = [
          dsid,
          1,
          dsid,
          allDatasets.find((ds) => ds.id === dsid)?.name || '',
        ];
      }
    }

    if (query.app) {
      const key = `${query.app},${query.username}`;

      if (key in statistics.aggApp) {
        statistics.aggApp[key][metricColumnIdx] += 1;
      } else {
        statistics.aggApp[key] = [
          query.app,
          1,
          query.username,
          dsid,
          allDatasets.find((ds) => ds.id === dsid)?.name || '',
        ];
      }
    }

    if (query.visual) {
      if (query.visual in statistics.aggVisual) {
        statistics.aggVisual[query.visual][metricColumnIdx] += 1;
      } else {
        statistics.aggVisual[query.visual] = [
          Number(query.visual),
          1,
          dsid,
          allDatasets.find((ds) => ds.id === dsid)?.name || '',
        ];
      }
    }

    const queryTimestamp = moment.utc(query.ts);
    if (queryTimestamp.isValid()) {
      const formattedTimestamp =
        queryTimestamp.format('YYYY-MM-DD HH:mm:ss.SSS') + '000';
      statistics.firsttime = formattedTimestamp;
      if (!statistics.lasttime) {
        statistics.lasttime = formattedTimestamp;
      }
    }
  };

  queryLogs.forEach(computeStats);

  return statistics;
};
