import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import * as getScheduleGroups from 'apis/jobs/schedule-groups/get-schedule-groups';
import * as createViewRefreshJob from 'apis/jobs/view-refresh/create-view-refresh-job';
import { mockScheduleGroups } from 'pages/job-page/utils/mocks';
import * as React from 'react';
import { AnalyticalJobModal } from './AnalyticalJobModal';

window.arc.frontend = {
  globals: {
    settings: { emailHostUser: 'test@test.com' },
    emailAddresses: '[]',
  },
} as any;

jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
}));

configure({ testIdAttribute: 'data-test-id' });

const DOWN_ARROW = { keyCode: 40 };

beforeEach(() => {
  jest
    .spyOn(createViewRefreshJob, 'createViewRefreshJob')
    .mockImplementation(jest.fn());
  jest
    .spyOn(getScheduleGroups, 'getScheduleGroups')
    .mockResolvedValue(mockScheduleGroups);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('AnalyticalJobModal component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <AnalyticalJobModal viewName="Test View" dataConnectionId={1} />
      </QueryClientProvider>
    );
  });

  describe('when the create interval option is selected in Schedule', () => {
    it('should display two new inputs for the new interval', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <AnalyticalJobModal viewName="Test View" dataConnectionId={1} />
        </QueryClientProvider>
      );

      fireEvent.keyDown(screen.getByLabelText('Schedule'), DOWN_ARROW);
      await waitFor(() => {
        expect(screen.getByText('Create Interval')).toBeInTheDocument();
      });
      fireEvent.click(screen.getByText('Create Interval'));

      await waitFor(() => {
        expect(screen.getByTestId('intervalName')).toBeInTheDocument();
      });
      expect(screen.getByTestId('dropdown-input-minutes')).toBeInTheDocument();
    });

    it('should not save the job with create interval selected and missing interval data', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <AnalyticalJobModal viewName="Test View" dataConnectionId={1} />
        </QueryClientProvider>
      );

      fireEvent.change(screen.getByLabelText('Name'), {
        target: { value: 'Job name' },
      });

      fireEvent.keyDown(screen.getByLabelText('Schedule'), DOWN_ARROW);
      await waitFor(() => {
        expect(screen.getByText('Create Interval')).toBeInTheDocument();
      });
      fireEvent.click(screen.getByText('Create Interval'));

      await waitFor(() => {
        expect(screen.getByTestId('intervalName')).toBeInTheDocument();
      });
      expect(screen.getByTestId('dropdown-input-minutes')).toBeInTheDocument();

      fireEvent.click(screen.getByRole('button', { name: 'Save' }));

      await waitFor(() => {
        expect(screen.getByTestId('intervalError')).toBeInTheDocument();
      });
      expect(screen.getByTestId('intervalNameError')).toBeInTheDocument();
    });
  });

  describe('when the cancel button is pressed', () => {
    it('closes the overlay without saving the values', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <AnalyticalJobModal viewName="Test View" dataConnectionId={1} />
        </QueryClientProvider>
      );
      fireEvent.click(screen.getByRole('button', { name: 'Cancel' }));
      expect(
        jest.mocked(createViewRefreshJob.createViewRefreshJob)
      ).not.toHaveBeenCalled();
    });
  });

  describe('when the save button is pressed', () => {
    describe('and the form contains a new interval', () => {
      it('saves the job and closes the overlay', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <AnalyticalJobModal viewName="Test View" dataConnectionId={1} />
          </QueryClientProvider>
        );

        fireEvent.change(screen.getByLabelText('Name'), {
          target: { value: 'Job Name' },
        });

        fireEvent.keyDown(screen.getByLabelText('Schedule'), DOWN_ARROW);
        await waitFor(() => {
          expect(screen.getByText('Create Interval')).toBeInTheDocument();
        });
        fireEvent.click(screen.getByText('Create Interval'));
        await waitFor(() => {
          expect(screen.getByTestId('intervalName')).toBeInTheDocument();
        });
        fireEvent.change(screen.getByTestId('intervalName'), {
          target: { value: 'test' },
        });

        fireEvent.click(screen.getByTestId('dropdown-input-minutes'));
        await waitFor(() => {
          expect(screen.getByTestId('popover-0')).toBeInTheDocument();
        });
        fireEvent.click(screen.getByTestId('popover-0'));

        fireEvent.click(screen.getByRole('button', { name: 'Save' }));

        await waitFor(() => {
          expect(
            jest.mocked(createViewRefreshJob.createViewRefreshJob)
          ).toHaveBeenCalledWith({
            name: 'Job Name',
            errorTo: [],
            scheduleGroup: { interval: '0,15 * * * *', name: 'test' },
            scheduleGroupId: -2,
            dataConnectionId: 1,
            viewName: 'Test View',
          });
        });
      });
    });

    describe('and the form contains an existing interval', () => {
      it('saves the job and closes the overlay', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <AnalyticalJobModal viewName="Test View" dataConnectionId={1} />
          </QueryClientProvider>
        );

        fireEvent.change(screen.getByLabelText('Name'), {
          target: { value: 'Job Name' },
        });

        fireEvent.keyDown(screen.getByLabelText('Schedule'), DOWN_ARROW);
        await waitFor(() => {
          expect(screen.getByText('Hourly')).toBeInTheDocument();
        });
        fireEvent.click(screen.getByText('Hourly'));

        fireEvent.click(screen.getByRole('button', { name: 'Save' }));

        await waitFor(() => {
          expect(
            jest.mocked(createViewRefreshJob.createViewRefreshJob)
          ).toHaveBeenCalledWith({
            name: 'Job Name',
            errorTo: [],
            scheduleGroupId: 1,
            dataConnectionId: 1,
            viewName: 'Test View',
          });
        });
      });
    });

    describe('and the form contains invalid data', () => {
      it('shows an error below the incorrect fields', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <AnalyticalJobModal viewName="Test View" dataConnectionId={1} />
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('button', { name: 'Save' }));

        const nameError = screen.getByTestId('nameError');
        await waitFor(() => {
          expect(nameError).toHaveTextContent('Name is a required field');
        });

        expect(
          jest.mocked(createViewRefreshJob.createViewRefreshJob)
        ).not.toHaveBeenCalled();
      });
    });
  });
});
