import HelpIcon from '@cloudera-internal/cuix-core/icons/svg/ic-help.svg';
import { yupResolver } from '@hookform/resolvers/yup';
import { useGetScheduleGroupsQuery } from 'apis/jobs/schedule-groups/get-schedule-groups';
import { useCreateViewRefreshJobMutation } from 'apis/jobs/view-refresh/create-view-refresh-job';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { useClouderaDialogContext } from 'components/cloudera-dialog/ClouderaDialog';
import { Cron } from 'components/cron/Cron';
import { Notification } from 'components/notification/Notification';
import { TokenInput } from 'components/token-input/TokenInput';
import { VizTooltip } from 'components/VizTooltip';
import cronstrue from 'cronstrue';
import moment from 'moment';
import React, { useCallback, useMemo } from 'react';
import { Form, Modal, Popover } from 'react-bootstrap';
import { Controller, useForm } from 'react-hook-form';
import Select from 'react-select';
import classes from '../email-job-modal/EmailJobModal.module.scss';
import {
  AnalyticalJobFormSchema,
  analyticalJobFormSchema,
} from './analytical-job-form-schema';
import { getEmailAddressHints } from 'utils/get-email-address-hints';

type AnalyticalJobModalProps = { viewName: string; dataConnectionId: number };

/**
 * Modal for creating scheduled jobs for refreshing analytical views
 */
export const AnalyticalJobModal = ({
  viewName,
  dataConnectionId,
}: AnalyticalJobModalProps): JSX.Element => {
  const { reset: resetDialog } = useClouderaDialogContext();

  const emailAddressOptions = useMemo(
    () =>
      getEmailAddressHints().map((email) => ({
        label: email,
        value: email,
      })),
    []
  );

  const form = useForm({
    resolver: yupResolver(analyticalJobFormSchema),
    defaultValues: {
      errorTo: [],
    },
  });

  const { register, handleSubmit, control, clearErrors, formState } = form;
  const scheduleGroupId = form.watch('scheduleGroupId');

  const { data: scheduleGroupOptions = [] } = useGetScheduleGroupsQuery({
    select: (scheduleGroups) => {
      const scheduleOptions = scheduleGroups.map(
        ({ id, name, runInterval }) => ({
          value: id,
          label: name,
          runInterval,
        })
      );

      return [
        ...scheduleOptions,
        { value: -2, label: 'Create Interval', runInterval: '' },
      ];
    },
  });

  const {
    mutateAsync: createMutation,
    reset: resetCreate,
    error: createError,
  } = useCreateViewRefreshJobMutation();

  const resetErrors = useCallback(() => {
    resetCreate();
    clearErrors('root');
  }, [clearErrors, resetCreate]);

  const onSubmit = useCallback(
    async (formFields: AnalyticalJobFormSchema) => {
      await createMutation({ ...formFields, dataConnectionId, viewName });
      resetDialog();
    },
    [createMutation, dataConnectionId, resetDialog, viewName]
  );

  return (
    <>
      <Modal.Header closeButton>
        <Modal.Title>Schedule refresh for {viewName}</Modal.Title>
      </Modal.Header>

      <Modal.Body>
        <Notification
          title="Error"
          message={
            createError?.response?.data ?? formState.errors.root?.message ?? ''
          }
          variant="danger"
          show={!!createError || !!formState.errors.root}
          dismissible
          onClose={resetErrors}
        />
        <Form role="form" onSubmit={handleSubmit(onSubmit)} id="emailModalForm">
          <Form.Group controlId="name">
            <Form.Label className="required-label">Name</Form.Label>
            <Form.Control
              {...register('name', { shouldUnregister: true })}
              isInvalid={!!formState.errors.name}
              data-test-id="jobName"
            />

            <Form.Control.Feedback type="invalid" data-test-id="nameError">
              {formState.errors.name?.message}
            </Form.Control.Feedback>
          </Form.Group>

          <Form.Group controlId="scheduleGroupId">
            <Form.Label className="required-label">Schedule</Form.Label>
            <Controller
              name="scheduleGroupId"
              control={control}
              shouldUnregister
              render={({
                field: { onChange, value, onBlur, name },
                fieldState: { error, invalid },
              }) => {
                const scheduleGroup = scheduleGroupOptions.find(
                  (option) => option.value === value
                );

                return (
                  <>
                    <Select
                      options={scheduleGroupOptions}
                      value={scheduleGroup}
                      onChange={(event) => onChange(event?.value)}
                      onBlur={onBlur}
                      name={name}
                      isClearable
                      maxMenuHeight={150}
                      placeholder="No Scheduling"
                      inputId="scheduleGroupId"
                      id="editJobSchedule"
                      classNamePrefix="react-select"
                      className={invalid ? 'is-invalid' : ''}
                    />
                    {!!scheduleGroup?.runInterval && (
                      <Form.Text className="mt-1">
                        {cronstrue.toString(scheduleGroup.runInterval)} (UTC)
                      </Form.Text>
                    )}
                    <Form.Control.Feedback
                      type="invalid"
                      data-test-id="scheduleError"
                    >
                      {error?.message}
                    </Form.Control.Feedback>
                  </>
                );
              }}
            />
          </Form.Group>

          {scheduleGroupId === -2 && (
            <div className="ml-4">
              <Form.Group controlId="scheduleGroup.name">
                <Form.Label className="required-label">
                  Interval Name
                </Form.Label>
                <Form.Control
                  {...register('scheduleGroup.name', {
                    shouldUnregister: true,
                  })}
                  isInvalid={!!formState.errors.scheduleGroup?.name}
                  data-test-id="intervalName"
                />

                <Form.Control.Feedback
                  type="invalid"
                  data-test-id="intervalNameError"
                >
                  {formState.errors.scheduleGroup?.name?.message}
                </Form.Control.Feedback>
              </Form.Group>
              <Form.Group controlId="scheduleGroup.interval">
                <Form.Label className="required-label">
                  Interval (in UTC)
                </Form.Label>
                <Controller
                  name="scheduleGroup.interval"
                  control={control}
                  shouldUnregister
                  defaultValue="15 * * * *"
                  render={({
                    field: { onChange, value },
                    fieldState: { error, invalid },
                  }) => (
                    <>
                      <Cron
                        value={value}
                        onChange={onChange}
                        className={invalid ? 'is-invalid' : ''}
                      />

                      <Form.Control.Feedback
                        type="invalid"
                        data-test-id="intervalError"
                      >
                        {error?.message}
                      </Form.Control.Feedback>
                    </>
                  )}
                />

                <p className={classes.timeIndicator}>
                  UTC: {moment.utc(Date.now()).format('YYYY-MM-DD HH:mm')}
                  <br />
                  Local Time:{' '}
                  {moment.utc(Date.now()).local().format('YYYY-MM-DD HH:mm')}
                </p>
              </Form.Group>
            </div>
          )}
          <Form.Group controlId="emailErrorTo">
            <Form.Label className="d-flex align-items-center gap-1">
              Error Notification Emails To
              <VizTooltip
                overlay={
                  <Popover id="emailErrorToTooltip">
                    <Popover.Content>
                      Enter email addresses here that we will send error
                      notifications to
                    </Popover.Content>
                  </Popover>
                }
              >
                <HelpIcon color="#2e89c5" />
              </VizTooltip>
            </Form.Label>
            <Controller
              name="errorTo"
              control={control}
              render={({ field }) => (
                <TokenInput
                  {...field}
                  id="emailErrorTo"
                  isInvalid={!!formState.errors?.errorTo}
                  options={emailAddressOptions}
                />
              )}
            />
            <Form.Control.Feedback type="invalid">
              {formState.errors.errorTo?.message ??
                formState.errors.errorTo?.find?.((item) => item?.message)
                  ?.message}
            </Form.Control.Feedback>
          </Form.Group>
        </Form>
      </Modal.Body>
      <Modal.Footer>
        <ClouderaButton
          type="button"
          disabled={formState.isSubmitting}
          onClick={resetDialog}
        >
          Cancel
        </ClouderaButton>
        <ClouderaButton
          disabled={formState.isSubmitting}
          color="primary"
          type="submit"
          form="emailModalForm"
          testId="submitEmailModalForm"
        >
          Save
        </ClouderaButton>
      </Modal.Footer>
    </>
  );
};
