import '@testing-library/jest-dom';
import { configure, fireEvent, render, screen } from '@testing-library/react';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import * as React from 'react';
import { BulkActionButtonType, BulkActionModal } from './BulkActionModal';

configure({ testIdAttribute: 'data-test-id' });

const onConfirm = jest.fn();
const mockedItems = [
  { id: 1, name: 'item-1' },
  { id: 2, name: 'item-2' },
];

afterEach(() => {
  jest.resetAllMocks();
});

describe('BulkActionModal component', () => {
  it('renders without error', () => {
    render(
      <ClouderaDialog>
        <ClouderaDialogTrigger>
          <button>Open</button>
        </ClouderaDialogTrigger>
        <BulkActionModal
          title="Test"
          items={mockedItems}
          onConfirm={onConfirm}
        />
      </ClouderaDialog>
    );

    fireEvent.click(screen.getByRole('button', { name: 'Open' }));
  });

  it('can be triggered by opening it', async () => {
    render(
      <ClouderaDialog>
        <ClouderaDialogTrigger>
          <button>Open</button>
        </ClouderaDialogTrigger>
        <BulkActionModal
          title="Test Modal"
          items={mockedItems}
          onConfirm={onConfirm}
          actionType={BulkActionButtonType.Pause}
        />
      </ClouderaDialog>
    );

    expect(screen.queryByText('Test Modal')).not.toBeInTheDocument();
    fireEvent.click(screen.getByRole('button', { name: 'Open' }));
    expect(screen.getByText('Test Modal')).toBeInTheDocument();
  });

  describe('when the Cancel button is pressed', () => {
    it('closes the modal without calling onConfirm', async () => {
      render(
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <button>Open</button>
          </ClouderaDialogTrigger>
          <BulkActionModal
            title="Test Modal"
            items={mockedItems}
            onConfirm={onConfirm}
            actionType={BulkActionButtonType.Resume}
          />
        </ClouderaDialog>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Open' }));

      expect(screen.getByText('Test Modal')).toBeInTheDocument();
      fireEvent.click(screen.getByRole('button', { name: 'Cancel' }));
      expect(onConfirm).toHaveBeenCalledTimes(0);
    });
  });

  describe('when the main button is pressed', () => {
    it('closes the modal and calls onConfirm', async () => {
      render(
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <button>Open</button>
          </ClouderaDialogTrigger>
          <BulkActionModal
            title={'Test Modal'}
            items={mockedItems}
            onConfirm={onConfirm}
            actionType={BulkActionButtonType.Delete}
          />
        </ClouderaDialog>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Open' }));

      expect(screen.getByText('Test Modal')).toBeInTheDocument();
      fireEvent.click(screen.getByRole('button', { name: 'Delete' }));
      expect(onConfirm).toHaveBeenCalledWith([1, 2]);
    });

    describe('and only some items are selected', () => {
      it('performs the action to just those items', async () => {
        render(
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <button>Open</button>
            </ClouderaDialogTrigger>
            <BulkActionModal
              title={'Test Modal'}
              items={mockedItems}
              onConfirm={onConfirm}
              actionType={BulkActionButtonType.Delete}
            />
          </ClouderaDialog>
        );

        fireEvent.click(screen.getByRole('button', { name: 'Open' }));

        expect(screen.getByText('Test Modal')).toBeInTheDocument();
        fireEvent.click(
          screen.getAllByRole('checkbox', { name: 'Select Item' })[0]
        );
        fireEvent.click(screen.getByRole('button', { name: 'Delete' }));
        expect(onConfirm).toHaveBeenCalledWith([2]);
      });
    });
  });

  describe('when the disableSelection prop is set', () => {
    it('does not allow deselecting items', async () => {
      render(
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <button>Open</button>
          </ClouderaDialogTrigger>
          <BulkActionModal
            title={'Test Modal'}
            items={mockedItems}
            onConfirm={onConfirm}
            actionType={BulkActionButtonType.Delete}
            disableSelection
          />
        </ClouderaDialog>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Open' }));

      expect(screen.getByText('Test Modal')).toBeInTheDocument();

      expect(
        screen.queryByRole('checkbox', { name: 'Select Item' })
      ).not.toBeInTheDocument();
    });

    it('performs the action to all of the listed items', async () => {
      render(
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <button>Open</button>
          </ClouderaDialogTrigger>
          <BulkActionModal
            title={'Test Modal'}
            items={mockedItems}
            onConfirm={onConfirm}
            actionType={BulkActionButtonType.Delete}
            disableSelection
          />
        </ClouderaDialog>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Open' }));

      expect(screen.getByText('Test Modal')).toBeInTheDocument();

      fireEvent.click(screen.getByRole('button', { name: 'Delete' }));
      expect(onConfirm).toHaveBeenCalledWith([1, 2]);
    });
  });
});
