import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import PauseIcon from '@cloudera-internal/cuix-core/icons/svg/ic-pause.svg';
import ResumeIcon from '@cloudera-internal/cuix-core/icons/svg/ic-play.svg';
import RunNowIcon from '@cloudera-internal/cuix-core/icons/svg/ic-refresh.svg';
import StopIcon from '@cloudera-internal/cuix-core/icons/svg/ic-stop.svg';
import { createColumnHelper, getCoreRowModel } from '@tanstack/react-table';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import {
  ClouderaDialogBody,
  ClouderaDialogContent,
  ClouderaDialogFooter,
  ClouderaDialogHeader,
  ClouderaDialogTitle,
  useClouderaDialogContext,
} from 'components/cloudera-dialog/ClouderaDialog';
import { ClouderaTable } from 'components/cloudera-table';
import { useTable } from 'hooks/use-table';
import React, { useCallback, useEffect, useMemo } from 'react';
import { FormCheck } from 'react-bootstrap';
const columnHelper = createColumnHelper<{ id: number; name: string }>();

export enum BulkActionButtonType {
  Delete = 'delete',
  Cancel = 'cancel',
  Run = 'run',
  Resume = 'resume',
  Pause = 'pause',
}

const columns = [
  columnHelper.display({
    id: 'select',
    header: ({ table }) => (
      <FormCheck
        aria-label="Select All"
        checked={table.getIsAllPageRowsSelected()}
        onChange={table.getToggleAllPageRowsSelectedHandler()}
      />
    ),
    cell: ({ row }) => (
      <FormCheck
        aria-label="Select Item"
        checked={row.getIsSelected()}
        onChange={row.getToggleSelectedHandler()}
      />
    ),
  }),
  columnHelper.accessor('id', {
    header: 'ID',
    cell: (info) => info.getValue(),
  }),
  columnHelper.accessor('name', {
    header: 'Name',
    cell: (info) => info.getValue(),
  }),
];

type BulkActionModalProps = {
  title: string;
  description?: string;
  items: { id: number; name: string }[];
  onConfirm: (ids: number[]) => void;
  actionType?: BulkActionButtonType;
  disableSelection?: boolean;
  testId?: string;
};

/**
 * A simple confirmation modal for confirming actions to be performed on multiple items.
 * The items are displayed in a table and they can be selected/deselected from the list of items the action will be performed on.
 * @param title The title of the modal
 * @param description Optional additional text
 * @param items the list of items the action will be performed on
 * @param onConfirm the action to be performed
 * @param actionType customizes the primary button to match the action
 * @param disableSelection disables removing elements from the items list by selection
 * @param testId data test id
 */
export const BulkActionModal = ({
  title,
  description,
  items,
  onConfirm,
  actionType,
  disableSelection = false,
  testId = 'bulkModal',
}: BulkActionModalProps) => {
  const { reset } = useClouderaDialogContext();
  const table = useTable({
    data: items,
    columns,
    getCoreRowModel: getCoreRowModel(),
    initialState: {
      columnVisibility: {
        select: !disableSelection,
      },
    },
  });

  useEffect(() => {
    table.toggleAllRowsSelected(true);
  }, [table, items]);

  const handleConfirm = useCallback(() => {
    onConfirm(
      table.getSelectedRowModel().rows.map((row) => row.getValue('id'))
    );
    reset();
  }, [onConfirm, reset, table]);

  const isSomeSelected =
    table.getIsAllRowsSelected() || table.getIsSomeRowsSelected();

  const actionButton = useMemo(() => {
    switch (actionType) {
      case BulkActionButtonType.Delete:
        return (
          <ClouderaButton
            color="danger"
            icon={<DeleteIcon />}
            onClick={handleConfirm}
            testId="deleteConfirm"
            disabled={!isSomeSelected}
          >
            Delete
          </ClouderaButton>
        );
      case BulkActionButtonType.Run:
        return (
          <ClouderaButton
            color="primary"
            icon={<RunNowIcon />}
            onClick={handleConfirm}
            testId="runConfirm"
            disabled={!isSomeSelected}
          >
            Run
          </ClouderaButton>
        );
      case BulkActionButtonType.Cancel:
        return (
          <ClouderaButton
            color="primary"
            icon={<StopIcon />}
            onClick={handleConfirm}
            testId="cancelConfirm"
            disabled={!isSomeSelected}
          >
            Cancel Run
          </ClouderaButton>
        );
      case BulkActionButtonType.Resume:
        return (
          <ClouderaButton
            color="primary"
            icon={<ResumeIcon />}
            onClick={handleConfirm}
            testId="resumeConfirm"
            disabled={!isSomeSelected}
          >
            Resume
          </ClouderaButton>
        );
      case BulkActionButtonType.Pause:
        return (
          <ClouderaButton
            color="primary"
            icon={<PauseIcon />}
            onClick={handleConfirm}
            testId="pauseConfirm"
            disabled={!isSomeSelected}
          >
            Pause
          </ClouderaButton>
        );
      default:
        <ClouderaButton
          color="primary"
          onClick={handleConfirm}
          disabled={!isSomeSelected}
        >
          Confirm
        </ClouderaButton>;
    }
  }, [actionType, handleConfirm, isSomeSelected]);

  return (
    <ClouderaDialogContent testId={testId}>
      <ClouderaDialogHeader closeButton>
        <ClouderaDialogTitle>{title}</ClouderaDialogTitle>
      </ClouderaDialogHeader>

      <ClouderaDialogBody>
        {description}
        <ClouderaTable table={table} className="mt-1" />
      </ClouderaDialogBody>

      <ClouderaDialogFooter>
        <ClouderaButton onClick={reset}>Cancel</ClouderaButton>
        {actionButton}
      </ClouderaDialogFooter>
    </ClouderaDialogContent>
  );
};
