import React from 'react';
import {
  render,
  screen,
  fireEvent,
  waitFor,
  within,
} from '@testing-library/react';
import '@testing-library/jest-dom';
import ChangeAiProfileModal, {
  ChangeAiProfileModalProps,
} from './ChangeAiProfileModal';
import { legacyBridge } from 'services/legacy-bridge';

jest.mock('services/legacy-bridge', () => ({
  legacyBridge: {
    events: {
      next: jest.fn(),
    },
  },
}));

jest.mock('pages/site-settings-page/constants/setting-constants', () => ({
  SITE_SETTING_LABELS: {
    ai_engine: 'Engine',
    ai_api_key: 'API Key',
  },
}));

const defaultProps: ChangeAiProfileModalProps = {
  modalID: 'testModalID',
  profileID: 'embeddings',
  profiles: {
    default: {
      label: 'Default Profile',
      values: { embeddings_api_key: '', embeddings_engine: 'open_ai' },
    },
    'new-profile': {
      label: 'New Profile',
      values: { embeddings_api_key: 'new-api-key', embeddings_engine: 'llama' },
    },
  },
  currentProfile: 'default',
};

describe('ChangeAiProfileModal', () => {
  afterEach(() => {
    jest.clearAllMocks();
  });

  test('renders correctly', () => {
    render(<ChangeAiProfileModal {...defaultProps} />);

    expect(screen.getByText('Change Profile')).toBeInTheDocument();
    expect(screen.getByText('Select Profile:')).toBeInTheDocument();
    expect(screen.getByText('Save')).toBeInTheDocument();
    expect(screen.getByText('Cancel')).toBeInTheDocument();
  });

  test('dropdown displays profile options', async () => {
    render(<ChangeAiProfileModal {...defaultProps} />);

    const selectElement = screen.getByLabelText('Select Profile:');
    fireEvent.keyDown(selectElement, { key: 'ArrowDown' });

    const dropdownMenu = await screen.findByRole('listbox');

    expect(
      within(dropdownMenu).getByText('Default Profile')
    ).toBeInTheDocument();
    expect(within(dropdownMenu).getByText('New Profile')).toBeInTheDocument();
  });

  test('selecting a profile updates the state', async () => {
    render(<ChangeAiProfileModal {...defaultProps} />);

    const selectElement = screen.getByLabelText('Select Profile:');
    fireEvent.keyDown(selectElement, { key: 'ArrowDown' });
    fireEvent.click(await screen.findByText('New Profile'));

    expect(await screen.findByText('New Profile')).toBeInTheDocument();
  });

  test('displays API warning if API key is not provided', () => {
    render(<ChangeAiProfileModal {...defaultProps} />);

    expect(screen.getByText('API Key Not Provided')).toBeInTheDocument();
  });

  test('does not display API warning if API key is provided', () => {
    render(
      <ChangeAiProfileModal {...defaultProps} currentProfile="new-profile" />
    );

    expect(screen.queryByText('API Key Not Provided')).not.toBeInTheDocument();
  });

  test('save button calls legacyBridge with correct payload', async () => {
    render(<ChangeAiProfileModal {...defaultProps} />);

    fireEvent.click(screen.getByText('Save'));

    await waitFor(() => {
      expect(legacyBridge.events.next).toHaveBeenCalledWith({
        type: 'testModalID',
        payload: {
          label: 'Default Profile',
          value: 'default',
        },
      });
    });
  });

  test('save button closes the modal', async () => {
    render(<ChangeAiProfileModal {...defaultProps} />);

    fireEvent.click(screen.getByText('Save'));

    await waitFor(() => {
      expect(screen.queryByText('Change Profile')).not.toBeInTheDocument();
    });
  });

  test('cancel button closes the modal', async () => {
    render(<ChangeAiProfileModal {...defaultProps} />);

    fireEvent.click(screen.getByText('Cancel'));

    await waitFor(() => {
      expect(screen.queryByText('Change Profile')).not.toBeInTheDocument();
    });
  });

  test('renders profile details correctly', () => {
    render(<ChangeAiProfileModal {...defaultProps} />);

    expect(screen.getByText('ENGINE')).toBeInTheDocument();
    expect(screen.getByText('OpenAI')).toBeInTheDocument();
  });

  test('switching profiles updates profile details', async () => {
    render(<ChangeAiProfileModal {...defaultProps} />);

    const selectElement = screen.getByLabelText('Select Profile:');
    fireEvent.keyDown(selectElement, { key: 'ArrowDown' });
    fireEvent.click(await screen.findByText('New Profile'));

    await waitFor(() => {
      expect(screen.getByText('Cloudera AI Hosted MiniLM')).toBeInTheDocument();
    });
    expect(screen.queryByText('OpenAI')).not.toBeInTheDocument();
  });
});
