import React, { useCallback, useEffect, useState } from 'react';
import Modal from 'react-bootstrap/Modal';
import { VizButton } from 'components/viz-button/VizButton';
import {
  AiEngine,
  AiTab,
  ProfileOption,
  ProfileSettings,
} from 'pages/site-settings-page/types/ai-settings.type';
import { Alert, Form } from 'react-bootstrap';
import Select, { SingleValue } from 'react-select';
import { SITE_SETTING_LABELS } from 'pages/site-settings-page/constants/setting-constants';
import classes from './ChangeAiProfileModal.module.scss';
import { legacyBridge } from 'services/legacy-bridge';
import { getEngineLabel } from 'pages/site-settings-page/utils/setting-utils';

export type ChangeAiProfileModalProps = {
  modalID: string;
  profileID: AiTab;
  profiles: ProfileSettings;
  currentProfile: string;
};

const ChangeAiProfileModal = ({
  modalID,
  profileID,
  profiles,
  currentProfile,
}: ChangeAiProfileModalProps) => {
  const [isOpen, setIsOpen] = useState(true);
  const [activeProfile, setActiveProfile] = useState<ProfileOption>({
    label: profiles[currentProfile].label,
    value: currentProfile,
  });
  const [showApiWarning, setShowApiWarning] = useState(false);

  const showApiKeyWarning = useCallback(
    (profile: ProfileOption) => {
      const settingValues = profiles[profile.value].values;
      const cmlAuthModeKey = `${profileID}_cml_auth_mode`;
      const isAuthModeJWT =
        cmlAuthModeKey in settingValues
          ? settingValues[cmlAuthModeKey] === 'jwt'
          : false;

      if (isAuthModeJWT) return false;

      return settingValues[`${profileID}_engine`] === 'bedrock'
        ? !settingValues[`${profileID}_bedrock_credentials`]
        : !settingValues[`${profileID}_api_key`];
    },
    [profileID, profiles]
  );

  useEffect(() => {
    const warnApiKey = showApiKeyWarning(activeProfile);

    setShowApiWarning(warnApiKey);
  }, [activeProfile, showApiKeyWarning]);

  const handleSubmit = (e: React.FormEvent<HTMLFormElement>) => {
    e.preventDefault();
    legacyBridge.events.next({
      type: modalID,
      payload: activeProfile,
    });
    setIsOpen(false);
  };

  const dropdownOptions = Object.entries(profiles).map(
    ([value, { label }]) => ({ label, value })
  );

  const handleProfileChange = (profile: SingleValue<ProfileOption>) => {
    if (profile) {
      const warnApiKey = showApiKeyWarning(profile);

      setShowApiWarning(warnApiKey);
      setActiveProfile(profile);
    }
  };

  const renderBody = () => {
    const settingValues = profiles[activeProfile.value].values;

    const hiddenItems = [
      `${profileID}_api_key`,
      `${profileID}_auth_key`,
      `${profileID}_bedrock_credentials`,
      `${profileID}_custom_engine_settings`,
      `${profileID}_bedrock_custom_transformers`,
    ];

    const entriesWithLabels = Object.entries(settingValues)
      .filter(([key, _]) => !hiddenItems.includes(key))
      .map(([k, v]) => {
        const key = k.replace(profileID, 'ai');
        const label = SITE_SETTING_LABELS[key].toUpperCase();
        let value = v;

        if (key === 'ai_engine') {
          value = getEngineLabel(v as AiEngine, profileID);
        } else if (typeof v === 'boolean') {
          value = v ? 'Enabled' : 'Disabled';
        }

        return { value, label };
      });

    const sortedEntries = entriesWithLabels.sort((a, b) =>
      a.label.localeCompare(b.label)
    );

    return sortedEntries.map(({ value, label }) => {
      return (
        <div key={`${profileID}-${label}`}>
          <span className={classes.label}>{label}</span>
          <p>{value}</p>
        </div>
      );
    });
  };

  return (
    <Modal show={isOpen} onHide={() => setIsOpen(false)}>
      <Modal.Header closeButton>
        <Modal.Title>Change Profile</Modal.Title>
      </Modal.Header>

      <form onSubmit={handleSubmit}>
        <Modal.Body>
          <Form.Label data-test-id="setHomepageLabel" id="profileSelectLabel">
            Select Profile:
          </Form.Label>
          <Select
            name="profileSelect"
            inputId="profileSelect"
            aria-labelledby="profileSelectLabel"
            isSearchable
            options={dropdownOptions}
            value={activeProfile}
            onChange={handleProfileChange}
          />
          {showApiWarning && (
            <Alert className="mt-3" variant="warning">
              API Key Not Provided
            </Alert>
          )}
          <div className={`my-3 p-3 ${classes.profileDetails}`}>
            {renderBody()}
          </div>
        </Modal.Body>

        <Modal.Footer>
          <VizButton color="primary" size="medium" type="submit">
            Save
          </VizButton>
          <VizButton
            color="secondary"
            size="medium"
            onClick={() => setIsOpen(false)}
          >
            Cancel
          </VizButton>
        </Modal.Footer>
      </form>
    </Modal>
  );
};

export default ChangeAiProfileModal;
