import React from 'react';
import {
  fireEvent,
  queryByAttribute,
  render,
  screen,
} from '@testing-library/react';
import { Chatbot } from './Chatbot';
import { useChatbotMessaging } from '../../hooks/use-chatbot-messaging';
import {
  OverflowPolicy,
  ParamSelection,
  ParamSource,
} from 'types/chatbot-message.type';
import '@testing-library/jest-dom';

jest.mock('react-markdown', () => (props) => <div>{props.children}</div>);
// There is a known parsing issue with react-markdown so it needs to be mocked -> https://github.com/remarkjs/react-markdown/issues/635

jest.mock('../../hooks/use-chatbot-messaging', () => ({
  useChatbotMessaging: jest.fn(),
}));

jest.mock('react-intersection-observer', () => ({
  useInView: () => ({
    ref: jest.fn(),
    inView: true,
  }),
}));

jest.mock('arc/misc/speech-detection', () => ({
  setupSpeechDetection: jest.fn(),
}));

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      featureFlipper: {
        getValueById: jest.fn((id) => id === 'enable_speech_recognition'),
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

const mockJQuery = jest.fn().mockImplementation(() => {
  return {
    length: 1,
    find: jest.fn().mockReturnThis(),
  };
});

const defaultProps = {
  sendDsReq: jest.fn(() => Promise.resolve({ rows: [], colnames: [] })),
  sendHttpReq: jest.fn(() => Promise.resolve({})),
  getEmbeddingSourceInfo: jest.fn(() => ({
    filteredRows: [],
    aliases: [],
    contextPrompt: '',
  })),
  getTooltips: jest.fn(() => []),
  setDownloadInfo: jest.fn(),
  alertMessage: jest.fn(),
  addParams: jest.fn(),
  reportId: 1,
  datasetId: 1,
  settings: {
    welcomeMessage: 'Welcome to the chat!',
    embeddingProfile: 'default',
    completionProfile: 'default',
    isMarkdown: false,
    embeddingMaxTokens: '1000',
    completionMaxTokens: '4000',
    overflowPolicy: 'throw-error' as OverflowPolicy,
    paramSelection: 'all' as ParamSelection,
    paramSource: 'all' as ParamSource,
    questionPromptPrefix: '<<question>>',
    fontSize: '16px',
    saveConversation: false,
  },
  savedConversation: null,
  needSave: () => {},
};

const defaultMockValues = {
  messages: [
    {
      type: 'welcome',
      id: '1',
      text: 'Welcome to the chat!',
      sender: 'bot',
      ts: new Date(),
    },
  ],
  input: '',
  setInput: jest.fn(),
  sendMessage: jest.fn(),
  resetMessages: jest.fn(),
  loadingStatus: 'None',
  setTooltips: jest.fn(),
  hideTooltips: jest.fn(),
  tooltips: null,
};

describe('Chatbot Component', () => {
  beforeEach(() => {
    jest.clearAllMocks();
    (global as any).$ = mockJQuery;
  });

  test('renders initial messages', () => {
    (useChatbotMessaging as jest.Mock).mockReturnValue(defaultMockValues);
    render(<Chatbot {...defaultProps} />);

    expect(screen.getByText('Welcome to the chat!')).toBeInTheDocument();
  });

  test('displays multiple messages correctly', () => {
    const mockMessages = [
      {
        type: 'question',
        id: '1',
        text: 'Question 1',
        sender: 'user',
        ts: new Date(),
        tooltips: [],
      },
      {
        type: 'answer',
        id: '2',
        text: 'Answer 1',
        sender: 'bot',
        ts: new Date(),
        tooltips: [],
      },
      {
        type: 'question',
        id: '3',
        text: 'Question 2',
        sender: 'user',
        ts: new Date(),
        tooltips: [],
      },
      {
        type: 'answer',
        id: '4',
        text: 'Answer 2',
        sender: 'bot',
        ts: new Date(),
        tooltips: [],
      },
    ];

    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      messages: mockMessages,
    });

    render(<Chatbot {...defaultProps} />);

    mockMessages.forEach((message) => {
      expect(screen.getByText(message.text)).toBeInTheDocument();
    });
  });

  test('shows ChatbotLoadingIndicator during Embedding request', () => {
    const mockValues = { ...defaultMockValues, loadingStatus: 'Embedding' };
    (useChatbotMessaging as jest.Mock).mockReturnValue(mockValues);

    const { container } = render(<Chatbot {...defaultProps} />);

    const indicator = queryByAttribute(
      'data-test-id',
      container,
      'chatbot-loading-indicator'
    );

    const message = queryByAttribute(
      'data-test-id',
      container,
      'chatbot-loading-message'
    );

    expect(indicator).toBeInTheDocument();

    expect(message).toBeInTheDocument();
    expect(message?.textContent).toMatch(/Processing Embedding Request/);
  });

  test('shows ChatbotLoadingIndicator during Completion request', () => {
    const mockValues = { ...defaultMockValues, loadingStatus: 'Completion' };
    (useChatbotMessaging as jest.Mock).mockReturnValue(mockValues);

    const { container } = render(<Chatbot {...defaultProps} />);

    const indicator = queryByAttribute(
      'data-test-id',
      container,
      'chatbot-loading-indicator'
    );

    const message = queryByAttribute(
      'data-test-id',
      container,
      'chatbot-loading-message'
    );

    expect(indicator).toBeInTheDocument();

    expect(message).toBeInTheDocument();
    expect(message?.textContent).toMatch(/Processing Completion Request/);
  });

  test('shows ChatbotLoadingIndicator during Vector DB search', () => {
    const mockValues = { ...defaultMockValues, loadingStatus: 'Vector' };
    (useChatbotMessaging as jest.Mock).mockReturnValue(mockValues);

    const { container } = render(<Chatbot {...defaultProps} />);

    const indicator = queryByAttribute(
      'data-test-id',
      container,
      'chatbot-loading-indicator'
    );

    const message = queryByAttribute(
      'data-test-id',
      container,
      'chatbot-loading-message'
    );

    expect(indicator).toBeInTheDocument();

    expect(message).toBeInTheDocument();
    expect(message?.textContent).toMatch(/Performing Vector Search/);
  });

  test('does not show ChatbotLoadingIndicator when no requests', () => {
    (useChatbotMessaging as jest.Mock).mockReturnValue(defaultMockValues);

    const { container } = render(<Chatbot {...defaultProps} />);

    const indicator = queryByAttribute(
      'data-test-id',
      container,
      'chatbot-loading-indicator'
    );

    expect(indicator).not.toBeInTheDocument();
  });

  test('updates input field on typing', () => {
    const setInputMock = jest.fn();
    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      setInput: setInputMock,
    });

    render(<Chatbot {...defaultProps} />);

    const inputField = screen.getByPlaceholderText('Send a message');
    fireEvent.change(inputField, { target: { value: 'Hello, world!' } });

    expect(setInputMock).toHaveBeenCalledWith('Hello, world!');
  });

  test('sends a message and clears the input field', () => {
    let inputState = '';
    const setInputMock = jest.fn((newInput) => (inputState = newInput));
    const sendMessageMock = jest.fn();
    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      input: inputState,
      setInput: setInputMock,
      sendMessage: sendMessageMock,
    });

    render(<Chatbot {...defaultProps} />);

    const inputField = screen.getByPlaceholderText('Send a message');
    fireEvent.change(inputField, { target: { value: 'Hi there!' } });

    const sendButton = screen.getByText('Send');
    fireEvent.click(sendButton);

    expect(sendMessageMock).toHaveBeenCalled();
    expect(inputField).toHaveValue('');
  });

  test('clears messages when clear button is clicked', () => {
    const resetMessagesMock = jest.fn();
    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      resetMessages: resetMessagesMock,
    });

    render(<Chatbot {...defaultProps} />);

    const clearButton = screen.getByText('Clear');
    fireEvent.click(clearButton);

    expect(resetMessagesMock).toHaveBeenCalled();
  });

  test('displays an error message correctly', () => {
    const errorMessage = 'Something went wrong, please try again later.';
    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      messages: [
        {
          type: 'error',
          id: 'error1',
          text: errorMessage,
          sender: 'bot',
          ts: new Date(),
          hasError: true,
        },
      ],
    });

    render(<Chatbot {...defaultProps} />);

    expect(screen.getByText(errorMessage)).toBeInTheDocument();
  });

  test('displays tooltips correctly for an answer message', () => {
    const tooltipsMock = [{ alias: 'example', value: 'Tooltip content' }];
    const setTooltipsMock = jest.fn();
    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      setTooltips: setTooltipsMock,
      messages: [
        {
          type: 'answer',
          id: 'answer1',
          text: 'Answer text',
          sender: 'bot',
          ts: new Date(),
          tooltips: tooltipsMock,
        },
      ],
    });

    const { container } = render(<Chatbot {...defaultProps} />);

    const tooltipIcon = queryByAttribute(
      'data-test-id',
      container,
      'tooltip-icon'
    );

    if (!tooltipIcon) {
      throw new Error('Tooltip icon not found in the document');
    }

    fireEvent.click(tooltipIcon);
    expect(setTooltipsMock).toHaveBeenCalledWith(tooltipsMock);
  });

  test('clicking retry icon sends message with errorId', () => {
    const errorId = 'error-message-id';
    const sendMessageMock = jest.fn();
    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      sendMessage: sendMessageMock,
      messages: [
        {
          type: 'answer',
          id: 'answer1',
          text: 'Error occurred',
          sender: 'bot',
          ts: new Date(),
          hasError: true,
          errorId: errorId,
          tooltips: [],
        },
      ],
    });

    const { container } = render(<Chatbot {...defaultProps} />);
    const retryIcon = queryByAttribute('data-test-id', container, 'retry-icon');

    if (!retryIcon) {
      throw new Error('Retry icon not found in the document');
    }

    fireEvent.click(retryIcon);
    expect(sendMessageMock).toHaveBeenCalledWith(errorId);
  });

  test('pressing Enter sends a message', () => {
    const sendMessageMock = jest.fn();
    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      input: 'Some message',
      sendMessage: sendMessageMock,
    });

    render(<Chatbot {...defaultProps} />);

    const inputField = screen.getByPlaceholderText('Send a message');
    fireEvent.keyDown(inputField, { key: 'Enter', code: 'Enter' });

    expect(sendMessageMock).toHaveBeenCalled();
  });

  test('displays tooltip content correctly', () => {
    const tooltipsMock = [
      { alias: 'example1', value: 'Tooltip content 1' },
      { alias: 'example2', value: 'Tooltip content 2' },
    ];
    const hideTooltipsMock = jest.fn();
    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      tooltips: tooltipsMock,
      hideTooltips: hideTooltipsMock,
    });

    render(<Chatbot {...defaultProps} />);

    tooltipsMock.forEach((tooltip) => {
      Object.entries(tooltip).forEach(([alias, value]) => {
        expect(screen.getByText(`${alias}: ${value}`)).toBeInTheDocument();
      });
    });
  });

  test('closes tooltips when close button is clicked', () => {
    const tooltipsMock = [{ alias: 'example', value: 'Tooltip content' }];
    const hideTooltipsMock = jest.fn();
    (useChatbotMessaging as jest.Mock).mockReturnValue({
      ...defaultMockValues,
      tooltips: tooltipsMock,
      hideTooltips: hideTooltipsMock,
    });

    const { container } = render(<Chatbot {...defaultProps} />);
    const tooltipCloseButton = queryByAttribute(
      'data-test-id',
      container,
      'tooltip-close-button'
    );

    if (!tooltipCloseButton) {
      throw new Error('Tooltip close button not found in the document');
    }

    render(<Chatbot {...defaultProps} />);

    fireEvent.click(tooltipCloseButton);

    expect(hideTooltipsMock).toHaveBeenCalled();
  });

  test('displays microphone icon when speech recognition is enabled', () => {
    const { container } = render(<Chatbot {...defaultProps} />);

    const mic = queryByAttribute(
      'data-test-id',
      container,
      'chatbot-microphone'
    );

    expect(mic).toBeInTheDocument();
  });
});
