import React, { useEffect, useRef, useCallback, useState } from 'react';
import classes from './Chatbot.module.scss';
import classNames from 'classnames';
import Markdown from './Markdown';
import { Typewriter } from 'components/typewriter/Typewriter';
import { getGlobalArc } from 'utils/get-global-arc';
import moment from 'moment';
import { ChatbotLoadingIndicator } from 'components/chatbot-loading-indicator/ChatbotLoadingIndicator';
import { useChatbotMessaging } from 'hooks/use-chatbot-messaging';
import { CopyToClipboard } from 'components/copy-to-clipboard/CopyToClipboard';
import {
  DownloadInfo,
  OverflowPolicy,
  ParamSource,
  Message,
  ParamSelection,
  AddParamArg,
} from 'types/chatbot-message.type';
import { setupSpeechDetection } from 'arc/misc/speech-detection';
import { useInView } from 'react-intersection-observer';

const arc = getGlobalArc();
const isSpeechRecognitionEnabled = arc.frontend.featureFlipper.getValueById(
  'enable_speech_recognition'
);

type ChatbotProps = {
  sendDsReq: (
    embeddings: number[]
  ) => Promise<{ rows: string[][]; colnames: string[] }>;
  sendHttpReq: (proxyReq: any) => Promise<any>;
  getEmbeddingSourceInfo: (rows: string[][]) => {
    filteredRows: string[][];
    aliases: string[];
    contextPrompt: string;
  };
  getTooltips: (rows: string[][]) => Record<string, string>[];
  setDownloadInfo: (downloadInfo: DownloadInfo) => void;
  alertMessage: (errorMsg: string) => void;
  addParams: (params: AddParamArg[]) => void;
  needSave: () => void;
  reportId: number;
  datasetId: number;
  savedConversation: Message[] | null;
  settings: {
    welcomeMessage: string;
    embeddingProfile: string;
    completionProfile: string;
    isMarkdown: boolean;
    embeddingMaxTokens: string;
    completionMaxTokens: string;
    overflowPolicy: OverflowPolicy;
    paramSelection: ParamSelection;
    paramSource: ParamSource;
    questionPromptPrefix: string;
    fontSize: string;
    saveConversation: boolean;
  };
};

export const Chatbot = ({
  sendDsReq,
  sendHttpReq,
  getEmbeddingSourceInfo,
  getTooltips,
  setDownloadInfo,
  alertMessage,
  addParams,
  needSave,
  reportId,
  datasetId,
  savedConversation,
  settings,
}: ChatbotProps) => {
  const chatWindowRef = useRef<HTMLDivElement>(null);
  const [animationInProgress, setAnimationInProgress] = useState(false);
  const { ref, inView } = useInView();

  const scrollToBottom = useCallback(() => {
    if (chatWindowRef.current) {
      chatWindowRef.current.scrollTop = chatWindowRef.current.scrollHeight;
    }
  }, []);

  const {
    input,
    setInput,
    messages,
    sendMessage,
    resetMessages,
    loadingStatus,
    setTooltips,
    hideTooltips,
    tooltips,
  } = useChatbotMessaging({
    getEmbeddingSourceInfo,
    getTooltips,
    setDownloadInfo,
    sendDsReq,
    alertMessage,
    addParams,
    needSave,
    reportId,
    datasetId,
    savedConversation,
    sendHttpReq,
    settings,
  });

  // Always scroll if loading status or messages changed
  useEffect(() => {
    scrollToBottom();
  }, [loadingStatus, messages, scrollToBottom]);

  // Conditionally scroll if animation is in progress but user is not at bottom
  useEffect(() => {
    if (animationInProgress && !inView) {
      scrollToBottom();
    }
  }, [animationInProgress, inView, scrollToBottom]);

  useEffect(() => {
    const $element = $(`#ai-chat-${reportId}`);
    let cleanupFunction;

    if ($element.length) {
      const endCb = (finalText?: string) => {
        finalText = (finalText || '').toLowerCase().trim();
        setInput(finalText);
        $element.find('.arc-speech-text').trigger('focus');
      };

      const disabledCb = () => {
        alertMessage('Speech detection is not supported or enabled.');
      };

      cleanupFunction = setupSpeechDetection($element, endCb, disabledCb);
    }

    return () => {
      if (cleanupFunction) {
        cleanupFunction();
      }
    };
  }, [alertMessage, reportId, setInput]);

  return (
    <div
      id={`ai-chat-${reportId}`}
      className={classes.container}
      style={{ fontSize: settings.fontSize }}
    >
      <div ref={chatWindowRef} className={classes.chatWindow}>
        {messages.map((message, index) => (
          <div key={index} className={classes.message}>
            <div
              className={classNames('d-flex w-100', classes[message.sender])}
            >
              <div
                className={classNames(classes.bubble, {
                  [classes.error]:
                    message.type === 'answer' && message.hasError,
                })}
              >
                {message.type === 'answer' ? (
                  message.shouldAnimate ? (
                    <Typewriter
                      text={message.text}
                      isMarkdown={settings.isMarkdown}
                      onAnimationStateChange={setAnimationInProgress}
                    />
                  ) : settings.isMarkdown ? (
                    <Markdown markdownText={message.text} />
                  ) : (
                    <p>{message.text}</p>
                  )
                ) : (
                  <p>{message.text}</p>
                )}
              </div>

              {message.type === 'answer' && message.tooltips.length > 0 && (
                <span
                  data-test-id="tooltip-icon"
                  className={classNames(
                    classes.icon,
                    classes.infoIcon,
                    'info-icon'
                  )}
                  onClick={() => {
                    if (message.tooltips) {
                      setTooltips(message.tooltips);
                    }
                  }}
                />
              )}

              {message.type === 'answer' && !message.hasError && (
                <CopyToClipboard text={message.text} />
              )}

              {message.type === 'answer' && message.hasError && (
                <span
                  data-test-id="retry-icon"
                  className={classNames(
                    classes.icon,
                    classes.retryIcon,
                    'fa fa-repeat',
                    loadingStatus !== 'None' ? 'd-none' : ''
                  )}
                  onClick={() => sendMessage(message.errorId)}
                />
              )}
            </div>

            {message.ts && (
              <div
                className={`${classes.timestamp} ${classes[message.sender]}`}
              >
                {moment(message.ts).format('ddd HH:mm')}
              </div>
            )}
          </div>
        ))}
        <div ref={ref} />
      </div>

      {tooltips && (
        <div className={classes.sources}>
          <div className={classes.closeButton}>
            <i
              data-test-id="tooltip-close-button"
              className="fa fa-times-circle"
              onClick={hideTooltips}
            />
          </div>
          Similarity search results:
          {tooltips.map((tooltip, i) => (
            <div key={i} className={classes.tooltipCard}>
              {Object.entries(tooltip).map(([alias, value]) => (
                <p key={alias}>
                  {alias}: {value}
                </p>
              ))}
            </div>
          ))}
        </div>
      )}

      <div className="d-flex flex-column">
        <ChatbotLoadingIndicator loadingStatus={loadingStatus} />
        <div className={classes.inputWrp}>
          <div className={classes.inputSection}>
            <input
              type="text"
              value={input}
              disabled={loadingStatus !== 'None'}
              className={classNames(
                isSpeechRecognitionEnabled ? classes.pr30 : '',
                loadingStatus !== 'None' ? classes.zIndex5 : '',
                'arc-speech-text'
              )}
              autoFocus
              placeholder="Send a message"
              onChange={(e) => setInput(e.target.value)}
              onKeyDown={(e) => {
                if (e.key === 'Enter' && input.trim()) {
                  sendMessage();
                }
              }}
            />
            {isSpeechRecognitionEnabled && (
              <div
                data-test-id="chatbot-microphone"
                className={classNames(
                  classes.speechRecognition,
                  loadingStatus !== 'None' ? classes.disabledMic : '',
                  'arc-speech-trigger'
                )}
              >
                <i className="fa fa-microphone" />
              </div>
            )}
          </div>
          <button
            onClick={() => sendMessage()}
            disabled={loadingStatus !== 'None'}
            className="btn btn-primary mr-2"
          >
            Send
          </button>

          <button
            onClick={resetMessages}
            disabled={loadingStatus !== 'None'}
            className="btn btn-secondary"
          >
            Clear
          </button>
        </div>
      </div>
    </div>
  );
};
