import { useToggle } from 'hooks/use-toggle';
import React, {
  PropsWithChildren,
  createContext,
  useCallback,
  useContext,
  useMemo,
} from 'react';
import { Modal } from 'react-bootstrap';

type DialogContextValue = {
  isOpen: boolean;
  toggle: () => void;
  reset: () => void;
};

const DialogContext = createContext<DialogContextValue | null>({
  isOpen: false,
  toggle: () => {},
  reset: () => {},
});

export const useClouderaDialogContext = () => {
  const context = useContext(DialogContext);

  if (!context) {
    throw new Error(
      'useDialogContext must be used within a DialogContext Provider'
    );
  }

  return context;
};

/**
 * Wrapper for the bootstrap Modal component, which can be controlled by props or uncontrolled (manages its own state).
 * @param isOpen when the component is controlled by props, it indicates the open state
 * @param onChange when the component is controlled by props, it modifies the open state
 *
 * @example Uncontrolled
 * ```jsx
 * <ClouderaDialog>
 *   <ClouderaDialogTrigger>
 *     <Button>Show</Button>
 *   </ClouderaDialogTrigger>
 *   <ClouderaDialogContent>
 *     <ClouderaDialogHeader closeButton>
 *       <ClouderaDialogTitle>Cloudera dialog</ClouderaDialogTitle>
 *     </ClouderaDialogHeader>
 *     <ClouderaDialogBody>
 *       This is a Cloudera Dialog!
 *     </ClouderaDialogBody>
 *   </ClouderaDialogContent>
 * </ClouderaDialog>
 * ```
 * @example Controlled
 * ```jsx
 * const [isOpen, setIsOpen] = useState(false);
 *
 * <ClouderaDialog isOpen={isOpen} onChange={setIsOpen}>
 *   <ClouderaDialogContent>
 *     <ClouderaDialogHeader closeButton>
 *       <ClouderaDialogTitle>Cloudera dialog</ClouderaDialogTitle>
 *     </ClouderaDialogHeader>
 *     <ClouderaDialogBody>
 *       This is a Cloudera Dialog!
 *     </ClouderaDialogBody>
 *   </ClouderaDialogContent>
 * </ClouderaDialog>
 * ```
 */
export const ClouderaDialog = ({
  isOpen: isOpenControlled,
  onChange,
  children,
}: PropsWithChildren<{
  isOpen?: boolean;
  onChange?: (open: boolean) => void;
}>) => {
  const [isOpen, toggle, reset] = useToggle();

  const value = useMemo(() => {
    return {
      isOpen: isOpenControlled ?? isOpen,
      toggle: onChange ? () => onChange(!(isOpenControlled ?? isOpen)) : toggle,
      reset: onChange ? () => onChange(false) : reset,
    };
  }, [isOpen, isOpenControlled, onChange, reset, toggle]);

  return (
    <DialogContext.Provider value={value}>{children}</DialogContext.Provider>
  );
};

export const ClouderaDialogTrigger = ({ children }: PropsWithChildren) => {
  const { toggle } = useClouderaDialogContext();

  return (
    <div
      onClick={(e) => {
        // open the modal only if the click event originated from the children
        // handles the case when the trigger is disabled
        if (e.currentTarget !== e.target) {
          toggle();
        }
      }}
    >
      {children}
    </div>
  );
};

type ClouderaDialogContentProps = {
  size?: 'sm' | 'lg' | 'xl';
  scrollable?: boolean;
  onClose?: () => void;
  testId?: string;
};

export const ClouderaDialogContent = ({
  children,
  size,
  scrollable,
  onClose,
  testId,
}: PropsWithChildren<ClouderaDialogContentProps>) => {
  const { isOpen, reset } = useClouderaDialogContext();

  const handleClose = useCallback(() => {
    reset();
    onClose?.();
  }, [onClose, reset]);

  return (
    <Modal
      show={isOpen}
      onHide={handleClose}
      size={size}
      scrollable={scrollable}
      data-test-id={testId}
    >
      {children}
    </Modal>
  );
};

export const ClouderaDialogHeader = Modal.Header;
export const ClouderaDialogTitle = Modal.Title;
export const ClouderaDialogBody = Modal.Body;
export const ClouderaDialogFooter = Modal.Footer;
