import DropdownIcon from '@cloudera-internal/cuix-core/icons/svg/ic-dropdown.svg';
import SearchIcon from '@cloudera-internal/cuix-core/icons/svg/ic-search.svg';
import Fuse from 'fuse.js';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import React, { useCallback, useMemo, useState } from 'react';
import { Form, InputGroup, OverlayTrigger, Popover } from 'react-bootstrap';
import classes from './ClouderaGroupedListFilter.module.scss';
import classNames from 'classnames';

interface TOptionGrouped<TValue> {
  group: string | number;
  value: TValue;
  label: React.ReactNode;
  stringLabel?: string;
  id: string;
}

interface ClouderaGroupedListFilterProps<TValue>
  extends Omit<
    React.ComponentProps<typeof ClouderaButton>,
    'value' | 'onChange'
  > {
  value: TValue[];
  onChange: (value: TValue[]) => void;
  options: TOptionGrouped<TValue>[];
  label: string;
  confirmChange?: boolean;
}

const groupOptions = <TValue extends unknown>(
  options: TOptionGrouped<TValue>[]
): Map<string | number | undefined, TOptionGrouped<TValue>[]> => {
  const groupedOptions = new Map<
    string | number | undefined,
    TOptionGrouped<TValue>[]
  >();

  options.forEach((option) => {
    const group = option.group;
    const currentValue = groupedOptions.get(group) ?? [];
    currentValue.push(option);
    groupedOptions.set(group, currentValue);
  });

  return groupedOptions;
};

/**
 * A popover style list filter with search functionality
 * @param options a list of filter options
 * @param label the text to be displayed on the filter button
 * @param id unique identifier of the filter element
 * @param value a list of the currently selected options
 * @param onChange callback invoked when the currently selected options change
 * @param confirmChange all the selected filters will be applied at once, when the Apply button is pressed, @default true
 */
export const ClouderaGroupedListFilter = <TValue extends unknown>({
  value: appliedFilterValues,
  onChange,
  id,
  options,
  label,
  confirmChange = false,
  ...props
}: ClouderaGroupedListFilterProps<TValue>): JSX.Element => {
  const [searchTerm, setSearchTerm] = useState('');
  const [showFilter, setShowFilter] = useState(false);
  const [draftFilterValues, setDraftFilterValues] =
    useState(appliedFilterValues);

  const appliedOptions = options.filter((option) =>
    appliedFilterValues.includes(option.value)
  );

  const groupedOptions = useMemo(() => groupOptions(options), [options]);
  const groupKeys = [...groupedOptions.keys()];
  const displayedGroupedOptions = useMemo(() => {
    if (!searchTerm) {
      return groupedOptions;
    }

    const fuse = new Fuse(options, {
      keys: ['stringLabel', 'label'],
      threshold: 0.4,
    });
    const ungroupedResult = fuse
      .search(searchTerm)
      ?.map((result) => result.item);

    const groupedResult = groupOptions(ungroupedResult);
    return groupedResult;
  }, [groupedOptions, searchTerm, options]);

  const toggle = useCallback(
    (show: boolean) => {
      if (show) {
        setSearchTerm('');
        setDraftFilterValues(appliedFilterValues);
      }

      setShowFilter(show);
    },
    [appliedFilterValues]
  );

  const handleAll = () => {
    if (confirmChange) {
      setDraftFilterValues(options.map((option) => option.value));
    } else {
      setDraftFilterValues(options.map((option) => option.value));
      onChange(options.map((option) => option.value));
    }
  };

  const handleClear = useCallback(() => {
    confirmChange ? setDraftFilterValues([]) : onChange([]);
    setDraftFilterValues([]);
    setSearchTerm('');
  }, [onChange, confirmChange]);

  const handleApply = useCallback(() => {
    onChange(draftFilterValues);
    setShowFilter(false);
  }, [draftFilterValues, onChange]);

  const handleSelectOption = useCallback(
    (option: TValue) => {
      // apply changes immediately on the column
      if (!confirmChange) {
        if (appliedFilterValues.includes(option)) {
          onChange(appliedFilterValues.filter((item) => item !== option));
        } else {
          onChange([...appliedFilterValues, option]);
        }
      }

      if (draftFilterValues.includes(option)) {
        setDraftFilterValues((filters) =>
          filters.filter((filter) => filter !== option)
        );
      } else {
        setDraftFilterValues((filters) => [...filters, option]);
      }
    },
    [confirmChange, draftFilterValues, appliedFilterValues, onChange]
  );

  const handleGroupSelection = useCallback(
    (groupName?: string | number) => {
      if (!groupName) {
        return;
      }
      const options = groupedOptions.get(groupName);
      const values = options?.map((o) => o.value) ?? [];
      const allGroupValueChecked = options?.every((option) =>
        draftFilterValues.includes(option.value)
      );
      // apply changes immediately on the column

      if (!confirmChange) {
        if (allGroupValueChecked) {
          onChange(
            appliedFilterValues.filter((item) => !values?.includes(item))
          );
        } else {
          onChange([...new Set([...appliedFilterValues, ...values])]);
        }
      }

      if (allGroupValueChecked) {
        setDraftFilterValues((filters) =>
          filters.filter((filter) => !values?.includes(filter))
        );
      } else {
        setDraftFilterValues((filters) => [
          ...new Set([...filters, ...values]),
        ]);
      }
    },
    [
      confirmChange,
      draftFilterValues,
      appliedFilterValues,
      groupedOptions,
      onChange,
    ]
  );

  const noResultAfterSearch = ![...displayedGroupedOptions?.values()].flat()
    .length;

  const popover = (
    <Popover id={`popover-${id}`} className={classes.popover}>
      <Popover.Title className={classes.popoverTitle}>
        <div className="d-flex justify-content-between mb-2">
          <h5>{label}</h5>
          <ClouderaButton
            variant="link"
            onClick={handleClear}
            disabled={!draftFilterValues.length}
          >
            Clear
          </ClouderaButton>
        </div>
        <InputGroup>
          <InputGroup.Prepend>
            <InputGroup.Text>
              <SearchIcon />
            </InputGroup.Text>
          </InputGroup.Prepend>
          <Form.Control
            type="text"
            aria-label="Search"
            placeholder="Search"
            value={searchTerm}
            onChange={(e) => setSearchTerm(e.target.value)}
          />
        </InputGroup>
      </Popover.Title>
      {groupKeys.length > 1 && (
        <Form.Group key="All" className={classes.option}>
          <Form.Check
            checked={draftFilterValues.length === options.length}
            onChange={
              draftFilterValues.length === options.length
                ? handleClear
                : handleAll
            }
          ></Form.Check>
          <Form.Label>{'All'}</Form.Label>
        </Form.Group>
      )}
      <div className={classNames(classes.roundedScrollBar, classes.maxHeight)}>
        {groupKeys.map((groupName) => {
          const options = displayedGroupedOptions.get(groupName);
          const allGroupValueChecked = options?.every((option) =>
            draftFilterValues.includes(option.value)
          );
          return (
            options?.length && (
              <div className={classes.group} key={groupName}>
                <div className={classes.groupName}>
                  {/* {groupName} */}
                  <Form.Group key="groupName" className={classes.option}>
                    <Form.Check
                      className="pl-2"
                      checked={allGroupValueChecked}
                      onChange={() => handleGroupSelection(groupName)}
                    ></Form.Check>
                    <Form.Label>{groupName}</Form.Label>
                  </Form.Group>
                </div>
                {options.map((option) => (
                  <Form.Group
                    key={option.id}
                    className={classes.option}
                    controlId={`option-${option.id}`}
                    data-test-id={`option-${option.id}`}
                  >
                    <Form.Check
                      checked={draftFilterValues?.includes(option.value)}
                      onChange={() => {
                        handleSelectOption(option.value);
                      }}
                    ></Form.Check>
                    <Form.Label>{option.label}</Form.Label>
                  </Form.Group>
                ))}
              </div>
            )
          );
        })}
        {noResultAfterSearch && <div className="text-center">No results.</div>}
        {/* no option and no result */}
      </div>
      {confirmChange && (
        <Popover.Title className={classes.popoverFooter}>
          <ClouderaButton color="primary" onClick={handleApply}>
            Apply
          </ClouderaButton>
          <ClouderaButton variant="link" onClick={() => toggle(false)}>
            Cancel
          </ClouderaButton>
        </Popover.Title>
      )}
    </Popover>
  );

  return (
    <OverlayTrigger
      trigger="click"
      placement="bottom-start"
      rootClose
      overlay={popover}
      show={showFilter}
      onToggle={toggle}
    >
      <ClouderaButton
        color="secondary"
        variant="pill"
        size="sm"
        icon={<DropdownIcon />}
        iconPosition="right"
        testId={`filter-${id}`}
        {...props}
      >
        {label}
        {!!appliedOptions.length &&
          `: ${appliedOptions[0].stringLabel ?? appliedOptions[0].label}`}
        {appliedOptions.length > 1 && ` +${appliedOptions.length - 1}`}
      </ClouderaButton>
    </OverlayTrigger>
  );
};
