import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import * as React from 'react';
import { ClouderaListFilter } from './ClouderaListFilter';

configure({ testIdAttribute: 'data-test-id' });

const mockOptions = Array.from({ length: 3 }, (_, i) => ({
  value: `category-${i % 3}`,
  label: `category-${i % 3}`,
}));

describe('ClouderaListFilter component', () => {
  it('renders without error', () => {
    render(
      <ClouderaListFilter
        id={'category'}
        label="Category"
        options={mockOptions}
        value={[]}
        onChange={() => {}}
      />
    );
  });

  it('displays a list of options upon opening it', async () => {
    const mockOnChange = jest.fn();

    render(
      <ClouderaListFilter
        id="category"
        label="Category"
        options={mockOptions}
        value={[]}
        onChange={mockOnChange}
      />
    );

    fireEvent.click(screen.getByRole('button', { name: 'Category' }));

    await waitFor(() => {
      expect(screen.getByTestId('option-category-0')).toBeInTheDocument();
    });
    expect(screen.getByTestId('option-category-1')).toBeInTheDocument();
    expect(screen.getByTestId('option-category-2')).toBeInTheDocument();
  });

  it('the options can be selected', async () => {
    const mockOnChange = jest.fn();

    render(
      <ClouderaListFilter
        id={'category'}
        label="Category"
        options={mockOptions}
        value={[]}
        onChange={mockOnChange}
      />
    );

    fireEvent.click(screen.getByRole('button', { name: 'Category' }));

    await waitFor(() => {
      expect(screen.getByTestId('option-category-0')).toBeInTheDocument();
    });
    expect(
      screen.getByRole('checkbox', { name: 'category-1' })
    ).not.toBeChecked();

    fireEvent.click(screen.getByRole('checkbox', { name: 'category-1' }));
    expect(screen.getByRole('checkbox', { name: 'category-1' })).toBeChecked();
    expect(mockOnChange).toHaveBeenCalledWith(['category-1']);
  });

  it('the options can be de-selected', async () => {
    const mockOnChange = jest.fn();

    render(
      <ClouderaListFilter
        id={'category'}
        label="Category"
        options={mockOptions}
        value={['category-1']}
        onChange={mockOnChange}
      />
    );

    fireEvent.click(screen.getByTestId('filter-category'));

    await waitFor(() => {
      expect(screen.getByTestId('option-category-0')).toBeInTheDocument();
    });

    expect(screen.getByRole('checkbox', { name: 'category-1' })).toBeChecked();

    fireEvent.click(screen.getByRole('checkbox', { name: 'category-1' }));
    expect(
      screen.getByRole('checkbox', { name: 'category-1' })
    ).not.toBeChecked();
    expect(mockOnChange).toHaveBeenCalledWith([]);
  });

  it('the options can be searched', async () => {
    const mockOnChange = jest.fn();

    render(
      <ClouderaListFilter
        id={'category'}
        label="Category"
        options={[
          { value: 'first', label: 'first' },
          { value: 'second', label: 'second' },
        ]}
        value={[]}
        onChange={mockOnChange}
      />
    );

    fireEvent.click(screen.getByRole('button', { name: 'Category' }));

    await waitFor(() => {
      expect(screen.getByTestId('option-first')).toBeInTheDocument();
    });
    expect(screen.getByTestId('option-second')).toBeInTheDocument();

    fireEvent.change(screen.getByRole('textbox', { name: 'Search' }), {
      target: { value: 'firs' },
    });

    await waitFor(() => {
      expect(screen.getByTestId('option-first')).toBeInTheDocument();
    });
    expect(screen.queryByTestId('option-second')).not.toBeInTheDocument();
  });

  it('lets the user clear the filters', async () => {
    const mockOnChange = jest.fn();

    render(
      <ClouderaListFilter
        id={'category'}
        label="Category"
        options={mockOptions}
        value={['category-1']}
        onChange={mockOnChange}
      />
    );

    fireEvent.click(screen.getByTestId('filter-category'));
    fireEvent.click(screen.getByRole('button', { name: 'Clear' }));

    await waitFor(() => {
      expect(mockOnChange).toHaveBeenCalledWith([]);
    });
  });

  describe('when apply confirmation is not needed', () => {
    it('allows an option to be selected automatically', async () => {
      const mockOnChange = jest.fn();

      render(
        <ClouderaListFilter
          id={'category'}
          label="Category"
          options={mockOptions}
          value={[]}
          onChange={mockOnChange}
        />
      );

      fireEvent.click(screen.getByTestId('filter-category'));

      await waitFor(() => {
        expect(screen.getByTestId('option-category-0')).toBeInTheDocument();
      });

      fireEvent.click(screen.getByRole('checkbox', { name: 'category-1' }));
      expect(mockOnChange).toHaveBeenCalledWith(['category-1']);
    });
  });

  describe('when apply confirmation is needed', () => {
    it('the selected option will not be applied automatically', async () => {
      const mockOnChange = jest.fn();

      render(
        <ClouderaListFilter
          id={'category'}
          label="Category"
          options={mockOptions}
          value={[]}
          onChange={mockOnChange}
          confirmChange
        />
      );

      fireEvent.click(screen.getByTestId('filter-category'));

      await waitFor(() => {
        expect(screen.getByTestId('option-category-0')).toBeInTheDocument();
      });

      fireEvent.click(screen.getByRole('checkbox', { name: 'category-1' }));
      expect(
        screen.getByRole('checkbox', { name: 'category-1' })
      ).toBeChecked();
      expect(mockOnChange).not.toHaveBeenCalled();
    });

    describe('when clicking the apply button', () => {
      it('the selected option will be applied', async () => {
        const mockOnChange = jest.fn();

        render(
          <ClouderaListFilter
            id={'category'}
            label="Category"
            options={mockOptions}
            value={[]}
            onChange={mockOnChange}
            confirmChange
          />
        );

        fireEvent.click(screen.getByTestId('filter-category'));

        await waitFor(() => {
          expect(screen.getByTestId('option-category-0')).toBeInTheDocument();
        });

        fireEvent.click(screen.getByRole('checkbox', { name: 'category-1' }));
        expect(
          screen.getByRole('checkbox', { name: 'category-1' })
        ).toBeChecked();
        expect(mockOnChange).not.toHaveBeenCalled();

        fireEvent.click(screen.getByRole('button', { name: 'Apply' }));

        expect(mockOnChange).toHaveBeenCalledWith(['category-1']);
      });
    });
    describe('when clicking the cancel button', () => {
      it('the selected option will not be applied', async () => {
        const mockOnChange = jest.fn();

        render(
          <ClouderaListFilter
            id={'category'}
            label="Category"
            options={mockOptions}
            value={[]}
            onChange={mockOnChange}
            confirmChange
          />
        );

        fireEvent.click(screen.getByTestId('filter-category'));

        await waitFor(() => {
          expect(screen.getByTestId('option-category-0')).toBeInTheDocument();
        });

        fireEvent.click(screen.getByRole('checkbox', { name: 'category-1' }));
        expect(
          screen.getByRole('checkbox', { name: 'category-1' })
        ).toBeChecked();
        expect(mockOnChange).not.toHaveBeenCalled();

        fireEvent.click(screen.getByRole('button', { name: 'Cancel' }));

        expect(mockOnChange).not.toHaveBeenCalled();
      });
    });
  });
});
