import DropdownIcon from '@cloudera-internal/cuix-core/icons/svg/ic-dropdown.svg';
import SearchIcon from '@cloudera-internal/cuix-core/icons/svg/ic-search.svg';
import Fuse from 'fuse.js';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import React, { memo, useCallback, useEffect, useMemo, useState } from 'react';
import { Form, InputGroup, OverlayTrigger, Popover } from 'react-bootstrap';
import classes from './ClouderaListFilter.module.scss';

interface ClouderaListFilterProps<TValue>
  extends Omit<
    React.ComponentProps<typeof ClouderaButton>,
    'value' | 'onChange'
  > {
  value: TValue[];
  onChange: (value: TValue[]) => void;
  id: string;
  options: { value: TValue; label: React.ReactNode; stringLabel?: string }[];
  label: string;
  confirmChange?: boolean;
}

//Support generic prop types with memo (https://stackoverflow.com/a/60386939)
const typedMemo: <T>(c: T) => T = memo;

/**
 * A popover style list filter with search functionality
 * @param options a list of filter options
 * @param label the text to be displayed on the filter button
 * @param id unique identifier of the filter element
 * @param value a list of the currently selected options
 * @param onChange callback invoked when the currently selected options change
 * @param confirmChange all the selected filters will be applied at once, when the Apply button is pressed, @default true
 */
export const ClouderaListFilter = typedMemo(
  <TValue extends unknown>({
    value: appliedFilterValues,
    onChange,
    id,
    options,
    label,
    confirmChange = false,
    ...props
  }: ClouderaListFilterProps<TValue>): JSX.Element => {
    const [searchTerm, setSearchTerm] = useState('');
    const [showFilter, setShowFilter] = useState(false);
    const [draftFilterValues, setDraftFilterValues] =
      useState(appliedFilterValues);

    useEffect(() => {
      setDraftFilterValues(appliedFilterValues);
    }, [appliedFilterValues]);

    const appliedOptions = options.filter((option) =>
      appliedFilterValues.includes(option.value)
    );

    const displayedOptions = useMemo(() => {
      if (!searchTerm) {
        return options;
      }

      const fuse = new Fuse(options, { keys: ['stringLabel', 'label'] });
      return fuse.search(searchTerm)?.map((result) => result.item);
    }, [options, searchTerm]);

    const toggle = useCallback(
      (show: boolean) => {
        if (show) {
          setSearchTerm('');
          setDraftFilterValues(appliedFilterValues);
        }

        setShowFilter(show);
      },
      [appliedFilterValues]
    );

    const handleClear = useCallback(() => {
      onChange([]);
      setDraftFilterValues([]);
      setSearchTerm('');
    }, [onChange]);

    const handleApply = useCallback(() => {
      onChange(draftFilterValues);
      setShowFilter(false);
    }, [draftFilterValues, onChange]);

    const handleSelectOption = useCallback(
      (option: TValue) => {
        // apply changes immediately on the column
        if (!confirmChange) {
          if (appliedFilterValues.includes(option)) {
            onChange(appliedFilterValues.filter((item) => item !== option));
          } else {
            onChange([...appliedFilterValues, option]);
          }
        }

        if (draftFilterValues.includes(option)) {
          setDraftFilterValues((filters) =>
            filters.filter((filter) => filter !== option)
          );
        } else {
          setDraftFilterValues((filters) => [...filters, option]);
        }
      },
      [confirmChange, draftFilterValues, appliedFilterValues, onChange]
    );

    const popover = (
      <Popover id={`popover-${id}`} className={classes.popover}>
        <Popover.Title className={classes.popoverTitle}>
          <div className="d-flex justify-content-between mb-2">
            <h5>{label}</h5>
            <ClouderaButton
              variant="link"
              onClick={handleClear}
              disabled={!draftFilterValues.length}
            >
              Clear
            </ClouderaButton>
          </div>
          <InputGroup>
            <InputGroup.Prepend>
              <InputGroup.Text>
                <SearchIcon />
              </InputGroup.Text>
            </InputGroup.Prepend>
            <Form.Control
              type="text"
              aria-label="Search"
              placeholder="Search"
              value={searchTerm}
              onChange={(e) => setSearchTerm(e.target.value)}
            />
          </InputGroup>
        </Popover.Title>

        <Popover.Content className={classes.popoverContent}>
          {displayedOptions.map((option) => (
            <Form.Group
              key={String(option.value)}
              className={classes.filterOption}
              controlId={`option-${String(option.value)}`}
              data-test-id={`option-${String(option.value)}`}
            >
              <Form.Check
                checked={draftFilterValues?.includes(option.value)}
                onChange={() => handleSelectOption(option.value)}
              ></Form.Check>
              <Form.Label>{option.label}</Form.Label>
            </Form.Group>
          ))}
          {!displayedOptions.length && (
            <div className="text-center">No results.</div>
          )}
        </Popover.Content>
        {confirmChange && (
          <Popover.Title className={classes.popoverFooter}>
            <ClouderaButton color="primary" onClick={handleApply}>
              Apply
            </ClouderaButton>
            <ClouderaButton variant="link" onClick={() => toggle(false)}>
              Cancel
            </ClouderaButton>
          </Popover.Title>
        )}
      </Popover>
    );

    return (
      <OverlayTrigger
        trigger="click"
        placement="bottom-start"
        rootClose
        overlay={popover}
        show={showFilter}
        onToggle={toggle}
      >
        <ClouderaButton
          type="button"
          color="secondary"
          variant="pill"
          size="sm"
          icon={<DropdownIcon />}
          iconPosition="right"
          testId={`filter-${id}`}
          {...props}
        >
          {label}
          {!!appliedOptions.length &&
            `: ${appliedOptions[0].stringLabel ?? appliedOptions[0].label}`}
          {appliedOptions.length > 1 && ` +${appliedOptions.length - 1}`}
        </ClouderaButton>
      </OverlayTrigger>
    );
  }
);
