import {
  createColumnHelper,
  createTable,
  getCoreRowModel,
  getFilteredRowModel,
  getPaginationRowModel,
  getSortedRowModel,
} from '@tanstack/react-table';
import '@testing-library/jest-dom';
import { configure, fireEvent, render, screen } from '@testing-library/react';
import * as React from 'react';
import { ClouderaTable } from './ClouderaTable';

configure({ testIdAttribute: 'data-test-id' });

type MockDataType = {
  id: number;
  name: string;
  category: string;
};

const columnHelper = createColumnHelper<MockDataType>();

const mockData = Array.from({ length: 30 }, (_, i) => ({
  id: i,
  name: `name ${i}`,
  category: `category ${i % 3}`,
})) satisfies MockDataType[];

const mockColumns = [
  columnHelper.accessor('id', {
    header: 'Id',
    cell: (info) => info.getValue(),
  }),
  columnHelper.accessor('name', {
    header: 'Name',
    cell: (info) => info.getValue(),
  }),
  columnHelper.accessor('category', {
    header: 'Category',
    cell: (info) => info.getValue(),
  }),
];

describe('ClouderaTable component', () => {
  const tableOptions = {
    data: mockData,
    columns: mockColumns,
    getCoreRowModel: getCoreRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    getSortedRowModel: getSortedRowModel(),
    getFilteredRowModel: getFilteredRowModel(),
    state: {
      sorting: [],
      columnFilters: [],
      pagination: {
        pageIndex: 0,
        pageSize: 10,
      },
      columnPinning: {
        left: [],
        right: [],
      },
    },
    onStateChange: jest.fn(),
    renderFallbackValue: null,
  };

  it('renders without error', () => {
    const table = createTable(tableOptions);

    render(<ClouderaTable table={table} />);
  });

  it('displays a loading indicator when the data is still loading', () => {
    const table = createTable(tableOptions);

    render(<ClouderaTable table={table} isLoading />);

    expect(screen.getByTestId('vizLoadingSpinner')).toBeVisible();
  });

  describe('the table should handle pagination', () => {
    it('should switch between the pages using the buttons from the footer', () => {
      const table = createTable({
        ...tableOptions,
        state: {
          ...tableOptions.state,
          pagination: { pageIndex: 1, pageSize: 10 },
        },
      });

      const mockedFirstPage = jest.spyOn(table, 'firstPage');
      const mockedPreviousPage = jest.spyOn(table, 'previousPage');
      const mockedNextPage = jest.spyOn(table, 'nextPage');
      const mockedLastPage = jest.spyOn(table, 'nextPage');

      render(<ClouderaTable table={table} />);

      expect(screen.getByTestId('pageElementIndicator')).toHaveTextContent(
        '11'
      );
      expect(screen.getByLabelText('Rows per page')).toHaveValue('10');

      fireEvent.click(screen.getByTestId('firstPage'));
      expect(mockedFirstPage).toHaveBeenCalled();

      fireEvent.click(screen.getByTestId('nextPage'));
      expect(mockedNextPage).toHaveBeenCalled();

      fireEvent.click(screen.getByTestId('previousPage'));
      expect(mockedPreviousPage).toHaveBeenCalled();

      fireEvent.click(screen.getByTestId('lastPage'));
      expect(mockedLastPage).toHaveBeenCalled();
    });

    it('should disable first and previous page buttons on the first page', () => {
      const table = createTable({
        ...tableOptions,
        state: {
          ...tableOptions.state,
          pagination: { pageIndex: 0, pageSize: 10 },
        },
      });

      const mockedFirstPage = jest.spyOn(table, 'firstPage');
      const mockedPreviousPage = jest.spyOn(table, 'previousPage');
      const mockedNextPage = jest.spyOn(table, 'nextPage');
      const mockedLastPage = jest.spyOn(table, 'nextPage');

      render(<ClouderaTable table={table} />);

      expect(screen.getByTestId('firstPage')).toBeDisabled();
      fireEvent.click(screen.getByTestId('firstPage'));
      expect(mockedFirstPage).not.toHaveBeenCalled();

      expect(screen.getByTestId('previousPage')).toBeDisabled();
      fireEvent.click(screen.getByTestId('previousPage'));
      expect(mockedPreviousPage).not.toHaveBeenCalled();

      fireEvent.click(screen.getByTestId('nextPage'));
      expect(mockedNextPage).toHaveBeenCalled();

      fireEvent.click(screen.getByTestId('lastPage'));
      expect(mockedLastPage).toHaveBeenCalled();
    });

    it('should disable next and last page buttons on the last page', () => {
      const table = createTable({
        ...tableOptions,
        state: {
          ...tableOptions.state,
          pagination: { pageIndex: 2, pageSize: 10 },
        },
      });

      const mockedFirstPage = jest.spyOn(table, 'firstPage');
      const mockedPreviousPage = jest.spyOn(table, 'previousPage');
      const mockedNextPage = jest.spyOn(table, 'nextPage');
      const mockedLastPage = jest.spyOn(table, 'nextPage');

      render(<ClouderaTable table={table} />);

      fireEvent.click(screen.getByTestId('firstPage'));
      expect(mockedFirstPage).toHaveBeenCalled();

      fireEvent.click(screen.getByTestId('previousPage'));
      expect(mockedPreviousPage).toHaveBeenCalled();

      expect(screen.getByTestId('nextPage')).toBeDisabled();
      fireEvent.click(screen.getByTestId('nextPage'));
      expect(mockedNextPage).not.toHaveBeenCalled();

      expect(screen.getByTestId('lastPage')).toBeDisabled();
      fireEvent.click(screen.getByTestId('lastPage'));
      expect(mockedLastPage).not.toHaveBeenCalled();
    });
  });

  describe('the table should handle sorting', () => {
    it('should display the sorting icon next to the sorted column', () => {
      const table = createTable({
        ...tableOptions,
        state: {
          ...tableOptions.state,
          sorting: [{ id: 'category', desc: true }],
        },
      });

      const categoryColumn = table.getColumn('category');

      const mockedToggleSorting = categoryColumn
        ? jest.spyOn(categoryColumn, 'getToggleSortingHandler')
        : undefined;

      render(<ClouderaTable table={table} />);

      const sortCategoryButton = screen.getByTestId('sortCategory');
      expect(sortCategoryButton).toBeEnabled();
      fireEvent.click(sortCategoryButton);
      expect(mockedToggleSorting).toHaveBeenCalled();
    });
  });
});
