import PageFirstIcon from '@cloudera-internal/cuix-core/icons/svg/ic-page-first.svg';
import PageLastIcon from '@cloudera-internal/cuix-core/icons/svg/ic-page-last.svg';
import PageNextIcon from '@cloudera-internal/cuix-core/icons/svg/ic-page-next.svg';
import PagePreviousIcon from '@cloudera-internal/cuix-core/icons/svg/ic-page-previous.svg';
import SortAscendingIcon from '@cloudera-internal/cuix-core/icons/svg/ic-sort-ascending.svg';
import SortDescendingIcon from '@cloudera-internal/cuix-core/icons/svg/ic-sort-descending.svg';
import SortIcon from '@cloudera-internal/cuix-core/icons/svg/ic-sort.svg';

import { flexRender } from '@tanstack/react-table';
import React from 'react';
import { ButtonGroup, Form, Table } from 'react-bootstrap';
import classes from './ClouderaTable.module.scss';
import { VizSpinner } from 'components/viz-spinner/VizSpinner';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';

import { Table as TanstackTable } from '@tanstack/react-table';
import classNames from 'classnames';

type ClouderaTableProps<TData> = React.ComponentProps<typeof Table> & {
  table: TanstackTable<TData>;
  hasStaleData?: boolean;
  isLoading?: boolean;
  noDataText?: string;
};

/**
 * Custom table implementation wrapped around Tanstack Table.
 * Supports: pagination, filtering, sorting, selection etc.
 * @param table a Tanstack Table config, including columns, data and options, such as sorting, pagination and filtering
 * @param isLoading indicates whether the data for the table is still loading
 * @extends the list of params supported by Tanstack Table, such as className
 */
export const ClouderaTable = <TData extends unknown>({
  table,
  isLoading = false,
  noDataText = 'No data.',
  hasStaleData = false,
  className,
  ...props
}: ClouderaTableProps<TData>): JSX.Element => {
  const isPaginationEnabled =
    !!table.options.getPaginationRowModel || table.options.manualPagination;
  const isSortingEnabled =
    !!table.options.getSortedRowModel || table.options.manualSorting;

  const getAriaSortForColumn = (order: string | boolean) => {
    switch (order) {
      case 'asc':
        return 'ascending';
      case 'desc':
        return 'descending';
      default:
        return 'none';
    }
  };

  const getSortingIconForColumn = (order: string | boolean): JSX.Element => {
    switch (order) {
      case 'asc':
        return <SortAscendingIcon className="flex-shrink-0" />;
      case 'desc':
        return <SortDescendingIcon className="flex-shrink-0" />;
      default:
        return <SortIcon className="flex-shrink-0" />;
    }
  };

  return (
    <Table
      responsive
      className={classNames(classes.table, className)}
      {...props}
    >
      <thead>
        <tr>
          {table.getFlatHeaders().map((header) => (
            <th
              className="no-wrap"
              key={header.id}
              aria-sort={getAriaSortForColumn(header.column.getIsSorted())}
            >
              {isSortingEnabled && header.column.getCanSort() ? (
                <button
                  className="p-0"
                  onClick={header.column.getToggleSortingHandler()}
                  data-test-id={`sort${header.column.columnDef.header}`}
                  type="button"
                >
                  {flexRender(
                    header.column.columnDef.header,
                    header.getContext()
                  )}
                  {getSortingIconForColumn(header.column.getIsSorted())}
                </button>
              ) : (
                flexRender(header.column.columnDef.header, header.getContext())
              )}
            </th>
          ))}
        </tr>
      </thead>

      <tbody className={hasStaleData ? classes.staleData : ''}>
        {isLoading ? (
          <tr>
            <td
              colSpan={table.getVisibleLeafColumns().length}
              className="text-center text-lg"
            >
              <VizSpinner />
            </td>
          </tr>
        ) : table.getRowCount() === 0 ? (
          <tr>
            <td
              colSpan={table.getVisibleLeafColumns().length}
              className="text-center text-muted"
            >
              {noDataText}
            </td>
          </tr>
        ) : (
          table.getRowModel().rows.map((row) => (
            <tr
              key={row.id}
              className={`${
                row.getCanSelect() ? '' : `${classes.notSelectable} disabled`
              }`}
            >
              {row.getVisibleCells().map((cell) => (
                <td key={cell.id}>
                  {flexRender(cell.column.columnDef.cell, cell.getContext())}
                </td>
              ))}
            </tr>
          ))
        )}
      </tbody>

      {isPaginationEnabled && (
        <tfoot>
          <tr>
            <td colSpan={table.getVisibleLeafColumns().length}>
              <div className="d-flex justify-content-end align-items-center gap-12">
                <Form.Group className="d-flex align-items-baseline mb-0 gap-2">
                  <Form.Label className="shrink-0">Rows per page:</Form.Label>
                  <Form.Control
                    as="select"
                    aria-label="Rows per page"
                    value={table.getState().pagination.pageSize}
                    onChange={(e) => table.setPageSize(Number(e.target.value))}
                  >
                    {[5, 10, 25, 50, 100].map((option) => (
                      <option key={option}>{option}</option>
                    ))}
                  </Form.Control>
                </Form.Group>

                {!isLoading && (
                  <div data-test-id="pageElementIndicator">
                    {table.getState().pagination.pageIndex *
                      table.getState().pagination.pageSize +
                      Math.min(1, table.getRowCount())}
                    &ndash;
                    {Math.min(
                      (table.getState().pagination.pageIndex + 1) *
                        table.getState().pagination.pageSize,
                      table.getRowCount()
                    )}{' '}
                    of {table.getRowCount()}
                  </div>
                )}

                <ButtonGroup>
                  <ClouderaButton
                    icon={<PageFirstIcon />}
                    disabled={!table.getCanPreviousPage()}
                    onClick={table.firstPage}
                    variant="ghost"
                    testId="firstPage"
                    type="button"
                  />
                  <ClouderaButton
                    icon={<PagePreviousIcon />}
                    disabled={!table.getCanPreviousPage()}
                    onClick={table.previousPage}
                    variant="ghost"
                    testId="previousPage"
                    type="button"
                  />
                  <ClouderaButton
                    icon={<PageNextIcon />}
                    disabled={!table.getCanNextPage()}
                    onClick={table.nextPage}
                    variant="ghost"
                    testId="nextPage"
                    type="button"
                  />
                  <ClouderaButton
                    icon={<PageLastIcon />}
                    disabled={!table.getCanNextPage()}
                    onClick={table.lastPage}
                    variant="ghost"
                    testId="lastPage"
                    type="button"
                  />
                </ButtonGroup>
              </div>
            </td>
          </tr>
        </tfoot>
      )}
    </Table>
  );
};
