import {
  createColumnHelper,
  createTable,
  getCoreRowModel,
  getFilteredRowModel,
} from '@tanstack/react-table';
import '@testing-library/jest-dom';
import { configure, fireEvent, render, screen } from '@testing-library/react';
import * as React from 'react';
import { ClouderaTableDateFilter } from './ClouderaTableDateFilter';

configure({ testIdAttribute: 'data-test-id' });

type MockDataType = {
  id: number;
  startDate: string;
};

const columnHelper = createColumnHelper<MockDataType>();

const mockData = Array.from({ length: 10 }, (_, i) => ({
  id: i,
  startDate: new Date().toString(),
})) satisfies MockDataType[];

const mockColumns = [
  columnHelper.accessor('id', {
    header: 'Id',
    cell: (info) => info.getValue(),
  }),
  columnHelper.accessor('startDate', {
    header: 'Start Date',
    cell: (info) => info.getValue(),
  }),
];

describe('ClouderaTableDateFilter component', () => {
  const tableOptions = {
    data: mockData,
    columns: mockColumns,
    getCoreRowModel: getCoreRowModel(),
    getFilteredRowModel: getFilteredRowModel(),
    state: {
      columnFilters: [],
      columnPinning: {
        left: [],
        right: [],
      },
    },
    onStateChange: jest.fn(),
    renderFallbackValue: null,
  };

  it('renders without error', () => {
    const table = createTable(tableOptions);

    render(
      <ClouderaTableDateFilter
        column={table.getColumn('startDate')!}
        label="Start Date"
      />
    );
  });

  it('lets you select a date and sets it as a filter', async () => {
    const table = createTable(tableOptions);
    const mockedSetFilterValue = jest.spyOn(
      table.getColumn('startDate')!,
      'setFilterValue'
    );

    render(
      <ClouderaTableDateFilter
        column={table.getColumn('startDate')!}
        label="Start Date"
      />
    );

    const dateFilter = screen.getByLabelText('Filter Start Date');
    fireEvent.change(dateFilter, {
      target: { value: '2024-06-05' },
    });
    expect(mockedSetFilterValue).toHaveBeenCalledWith('2024-06-05');
  });

  it('displays the column filter on the button', () => {
    const table = createTable({
      ...tableOptions,
      state: {
        ...tableOptions.state,
        columnFilters: [{ id: 'startDate', value: '2024-06-05' }],
      },
    });

    render(
      <ClouderaTableDateFilter
        column={table.getColumn('startDate')!}
        label="Start Date"
      />
    );

    expect(screen.getByTestId('date-filter-startDate')).toHaveTextContent(
      '2024-06-05'
    );
  });
});
