import { Column as TanstackColumn } from '@tanstack/react-table';
import { ClouderaDateFilter } from 'components/cloudera-filter/ClouderaDateFilter';
import moment from 'moment';
import React, { useCallback } from 'react';

interface ClouderaTableDateFilterProps<TData, TValue>
  extends Omit<
    React.ComponentProps<typeof ClouderaDateFilter>,
    'id' | 'onChange' | 'value'
  > {
  column: TanstackColumn<TData, TValue>;
  label: string;
  includeTime?: boolean;
}

/**
 * A wrapper around ClouderaDateFilter, to be used alongside the ClouderaTable component.
 * It includes two filters for a minimum and a maximum date.
 * @param column a Tanstack Table column definition for the data that should be filtered
 * @param label the text to be displayed on the filter button
 * @param includeTime changes the input type from date to date-time
 */
export const ClouderaTableDateRangeFilter = <
  TData extends unknown,
  TValue extends unknown
>({
  column,
  label,
  ...props
}: ClouderaTableDateFilterProps<TData, TValue>): JSX.Element => {
  const [startValue = '', endValue = ''] = (column.getFilterValue() as
    | [string | undefined, string | undefined]
    | undefined) ?? ['', ''];

  const handleStartChange = useCallback(
    (value: string | undefined) =>
      column.setFilterValue([value, column.getFilterValue()?.[1]]),
    [column]
  );

  const handleEndChange = useCallback(
    (value: string | undefined) =>
      column.setFilterValue([
        column.getFilterValue()?.[0],
        props.includeTime ? value : moment(value).endOf('day').toISOString(),
      ]),
    [column, props.includeTime]
  );

  return (
    <>
      <ClouderaDateFilter
        value={startValue}
        onChange={handleStartChange}
        id={column.id}
        label={`${label} Start`}
        {...props}
      />
      <ClouderaDateFilter
        value={endValue}
        onChange={handleEndChange}
        id={column.id}
        label={`${label} End`}
        {...props}
      />
    </>
  );
};
