import {
  createColumnHelper,
  createTable,
  getCoreRowModel,
  getFilteredRowModel,
} from '@tanstack/react-table';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import * as React from 'react';
import { ClouderaTableListFilter } from './ClouderaTableListFilter';

configure({ testIdAttribute: 'data-test-id' });

type MockDataType = {
  id: number;
  name: string;
  category: string;
};

const columnHelper = createColumnHelper<MockDataType>();

const mockData = Array.from({ length: 30 }, (_, i) => ({
  id: i,
  name: `name-${i}`,
  category: `category-${i % 3}`,
})) satisfies MockDataType[];

const mockColumns = [
  columnHelper.accessor('id', {
    header: 'Id',
    cell: (info) => info.getValue(),
  }),
  columnHelper.accessor('name', {
    header: 'Name',
    cell: (info) => info.getValue(),
  }),
  columnHelper.accessor('category', {
    header: 'Category',
    cell: (info) => info.getValue(),
  }),
];

const mockOptions = Array.from({ length: 3 }, (_, i) => ({
  value: `category-${i % 3}`,
  label: `category-${i % 3}`,
}));

describe('ClouderaTableListFilter component', () => {
  const tableOptions = {
    data: mockData,
    columns: mockColumns,
    getCoreRowModel: getCoreRowModel(),
    getFilteredRowModel: getFilteredRowModel(),
    state: {
      columnFilters: [],
      columnPinning: {
        left: [],
        right: [],
      },
    },
    onStateChange: jest.fn(),
    renderFallbackValue: null,
  };

  it('renders without error', () => {
    const table = createTable(tableOptions);

    render(
      <ClouderaTableListFilter
        column={table.getColumn('category')!}
        label="Category"
        options={mockOptions}
      />
    );
  });

  it('displays a list of filter options that are selectable and clearable', async () => {
    const table = createTable(tableOptions);
    const mockedSetFilterValue = jest.spyOn(
      table.getColumn('category')!,
      'setFilterValue'
    );

    render(
      <ClouderaTableListFilter
        column={table.getColumn('category')!}
        label="Category"
        options={mockOptions}
      />
    );

    fireEvent.click(screen.getByRole('button', { name: 'Category' }));

    await waitFor(() => {
      expect(screen.getByTestId('option-category-0')).toBeInTheDocument();
    });
    expect(screen.getByTestId('option-category-1')).toBeInTheDocument();
    expect(screen.getByTestId('option-category-2')).toBeInTheDocument();

    fireEvent.click(screen.getByRole('checkbox', { name: 'category-1' }));
    expect(mockedSetFilterValue).toHaveBeenCalledWith(['category-1']);
  });

  it('lets the user clear the filters', async () => {
    const table = createTable({
      ...tableOptions,
      state: {
        ...tableOptions.state,
        columnFilters: [{ id: 'category', value: ['category-1'] }],
      },
    });
    const mockedSetFilterValue = jest.spyOn(
      table.getColumn('category')!,
      'setFilterValue'
    );

    render(
      <ClouderaTableListFilter
        column={table.getColumn('category')!}
        label="Category"
        options={mockOptions}
      />
    );

    fireEvent.click(screen.getByTestId('filter-category'));
    fireEvent.click(screen.getByRole('button', { name: 'Clear' }));

    await waitFor(() => {
      expect(mockedSetFilterValue).toHaveBeenCalledWith([]);
    });
  });
});
