import React, { useCallback, useState } from 'react';
import Modal from 'react-bootstrap/Modal';
import Form from 'react-bootstrap/Form';
import { useLegacyBridge } from 'services/legacy-bridge';
import { openInNewWindow } from 'arc/misc/misc';
import { useMutation } from '@tanstack/react-query';
import { useForm, useWatch } from 'react-hook-form';
import { Dataset } from 'arc/types/dataset.type';
import { createDatasetFromTable } from 'apis/create-dataset';
import { LoadingIndicatorActionButton } from 'components/loading-indicator-action-button/LoadingIndicatorActionButton';
import { VizButton } from 'components/viz-button/VizButton';
import { getGlobalArc } from 'utils/get-global-arc';
import { transformUnknownError } from 'utils/transformUnknownError';
import { Notification } from 'components/notification/Notification';

type ModalProps = {
  connection: { id: number; type: string };
  type: string;
  databaseName: string;
  tableName: string;
  onSuccess?: (param?: Dataset) => void;
  onFailure?: (errorMsg: string) => void;
};

export const CreateDatasetFromTableModal = (): JSX.Element | null => {
  const { register, control, setValue } = useForm();
  const title = useWatch({ control, name: 'title' });
  const [error, setError] = useState<Error | null>(null);
  const [modalProps, setModalProps] = useState<ModalProps | undefined>();
  const isFormFilled = !!modalProps && !!title?.length;
  const resetModal = () => {
    setModalProps(undefined);
    setError(null);
    setValue('title', undefined);
  };
  const onSuccess = (dataset: Dataset) => {
    if (modalProps?.onSuccess) {
      modalProps.onSuccess(dataset);
      resetModal();
    } else {
      openInNewWindow(
        `${
          getGlobalArc().frontend.globals.urls.datasetEditUrl
        }/${dataset?.dataset_id()}`
      );
    }
  };

  const { mutateAsync: createDataset, isLoading } = useMutation({
    mutationFn: async () =>
      modalProps &&
      isFormFilled &&
      (await createDatasetFromTable(
        modalProps.connection?.id,
        modalProps?.type,
        title,
        modalProps.databaseName,
        modalProps.tableName
      )),
    onSuccess: (data, variables, context) => {
      data && onSuccess(data);
      resetModal();
    },
    onError: (error, variables, context) => {
      const e = transformUnknownError(error, 'Failed to create dataset');
      setError(e);
      modalProps?.onFailure?.(e.message);
    },
  });

  const onConnectionIdChange = useCallback(
    ({ payload }: { payload?: ModalProps }) => {
      if (payload) {
        setModalProps(payload);
      }
    },

    [setModalProps]
  );

  useLegacyBridge('create-dataset-from-table-modal', onConnectionIdChange);

  return modalProps === undefined ? null : (
    <Modal
      show={!!modalProps}
      size={'sm'}
      data-test-id={'dataset-create-from-table-modal'}
    >
      <Modal.Header>
        <Modal.Title>
          {`New Dataset from table: ${modalProps.databaseName}.${modalProps.tableName}`}
        </Modal.Title>
      </Modal.Header>
      <Form className="mt-2 mb-3">
        <Modal.Body>
          <Notification
            title="Error"
            message={error?.message || ''}
            variant="danger"
            show={!!error}
            dismissible
            onClose={() => {
              setError(null);
            }}
          />
          <Form.Group>
            <Form.Label htmlFor={'title'} className="mt-1">
              Dataset Title<span className="required">*</span>
            </Form.Label>
            <Form.Control
              id={'title'}
              type={'text'}
              placeholder={'Enter a title'}
              required={true}
              {...register('title')}
            ></Form.Control>
          </Form.Group>
          <br></br>
        </Modal.Body>
        <Modal.Footer>
          <VizButton
            color="secondary"
            size="medium"
            onClick={() => {
              resetModal();
            }}
            testId="close-modal"
          >
            Close
          </VizButton>
          <LoadingIndicatorActionButton
            onClick={createDataset}
            isValid={isFormFilled}
            isSaving={isLoading}
            label="Create"
          ></LoadingIndicatorActionButton>
        </Modal.Footer>
      </Form>
    </Modal>
  );
};
