import {
  act,
  render,
  screen,
  fireEvent,
  waitFor,
  configure,
} from '@testing-library/react';
import { CreateDatasetModal } from './CreateDatasetModal';
import React from 'react';
import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import { Dataset } from 'arc/types/dataset.type';
import { legacyBridge } from 'services/legacy-bridge';
import '@testing-library/jest-dom';
import {
  getDatabases,
  getDatatables,
} from 'apis/get-database-and-datatable-list';
import { createDatasetFromTable } from 'apis/create-dataset';

configure({ testIdAttribute: 'data-test-id' });

jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

jest.mock('../../apis/get-database-and-datatable-list.ts', () => ({
  getDatabases: jest.fn(),
  getDatatables: jest.fn(),
}));

jest.mock('../../apis/create-dataset.ts', () => ({
  createDatasetFromTable: jest.fn(),
}));

jest.mock('arc/misc/misc.ts', () => ({
  determineDatabaseInQuery: jest.fn(),
}));

const mockDatabaseList = ['main', 'database2', 'database3'];
const mockTableList = ['table1', 'table2', 'table3'];
const connectionId = 111;
const connectionType = 'sqlite';
const testDatabaseName = mockDatabaseList[0];
const testDataTableName = mockTableList[0];
const testPayload = {
  initialData: {
    database: testDatabaseName,
    table: testDataTableName,
    sourceType: 'table',
  },

  connection: {
    id: connectionId,
    type: connectionType,
  },
};
const sourceSelectorTestId = 'dataset-source-type-select';
const databaseSelectorTestId = 'selected-database-selector';
const tableSelectorTestId = 'selected-table-selector';

beforeEach(() => {
  jest.mocked(getDatabases).mockResolvedValue(mockDatabaseList);
  jest.mocked(getDatatables).mockResolvedValue(mockTableList);
});

describe('CreateDatasetModal component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <CreateDatasetModal />
      </QueryClientProvider>
    );
  });

  it('Opens the modal and loads the initial data and state', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <CreateDatasetModal />
      </QueryClientProvider>
    );

    act(() => {
      legacyBridge.events.next({
        type: 'open-create-dataset-modal',
        payload: testPayload,
      });
    });

    const modal = await screen.findByText('Create New Dataset');
    expect(modal).toBeVisible();
    expect(screen.getByPlaceholderText('Enter a title')).toBeInTheDocument();
    expect(screen.getByTestId(sourceSelectorTestId)).toHaveTextContent('Table');
    expect(screen.getByTestId(databaseSelectorTestId)).toHaveTextContent(
      'Select Database'
    );
    expect(screen.getByTestId(tableSelectorTestId)).toHaveTextContent(
      'Select Table'
    );
    expect(screen.getByText('Create')).toBeDisabled();
  });

  it('Fills the modal with correct initial data, when initiated to create dataset from table', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <CreateDatasetModal />
      </QueryClientProvider>
    );

    act(() => {
      legacyBridge.events.next({
        type: 'open-create-dataset-modal',
        payload: testPayload,
      });
    });

    const modal = await screen.findByText('Create New Dataset');
    expect(modal).toBeVisible();
    expect(screen.getByTestId(sourceSelectorTestId)).toHaveTextContent('Table');
    expect(screen.getByTestId(databaseSelectorTestId)).toHaveTextContent(
      testDatabaseName
    );
    expect(screen.getByTestId(tableSelectorTestId)).toHaveTextContent(
      testDataTableName
    );
  });

  it('Enables the Save, if the necessary information (title, db, tablename) is filled', async () => {
    const datasetName = 'Dataset Name';
    const mockSuccessCallback = jest.fn((dataset: Dataset) => {
      /* Do nothing */
    });
    jest.mocked(createDatasetFromTable).mockResolvedValue({} as Dataset);

    render(
      <QueryClientProvider client={new QueryClient()}>
        <CreateDatasetModal />
      </QueryClientProvider>
    );

    act(() => {
      legacyBridge.events.next({
        type: 'open-create-dataset-modal',
        payload: {
          ...testPayload,
          onSuccess: mockSuccessCallback,
        },
      });
    });

    const modal = await screen.findByText('Create New Dataset');
    expect(modal).toBeVisible();
    const createBtn = screen.getByText('Create');
    expect(createBtn).toBeDisabled();
    const titleInput = screen.getByTestId('dataset-title-input');

    fireEvent.change(titleInput, { target: { value: datasetName } });

    await waitFor(async () => {
      expect(titleInput).toHaveValue(datasetName);
    });
    expect(createBtn).toBeEnabled();
    fireEvent.click(createBtn);

    await waitFor(() => {
      expect(createDatasetFromTable).toHaveBeenCalledWith(
        connectionId,
        connectionType,
        datasetName,
        testDatabaseName,
        testDataTableName
      );
    });
    expect(mockSuccessCallback).toHaveBeenCalled();
  });
});
