import React, { useCallback, useEffect, useRef, useState } from 'react';
import Modal from 'react-bootstrap/Modal';
import Form from 'react-bootstrap/Form';
import { useLegacyBridge } from 'services/legacy-bridge';
import { InputGroup } from 'react-bootstrap';
import { determineDatabaseInQuery, openInNewWindow } from 'arc/misc/misc';

import { Controller, FormProvider, useForm } from 'react-hook-form';
import { useMutation, useQuery } from '@tanstack/react-query';
import { Dataset } from 'arc/types/dataset.type';
import styles from './CreateDatasetModal.module.scss';
import Select from 'react-select';
import {
  createDatasetFromTable,
  createDatasetFromSql,
} from 'apis/create-dataset';
import {
  getDatabases,
  getDatatables,
} from 'apis/get-database-and-datatable-list';
import { LoadingIndicatorActionButton } from 'components/loading-indicator-action-button/LoadingIndicatorActionButton';
import { SQLExplorer } from 'components/sql-explorer/SQLExplorer';
import { VizButton } from 'components/viz-button/VizButton';
import { getGlobalArc } from 'utils/get-global-arc';
import { transformUnknownError } from 'utils/transformUnknownError';
import { Spinner } from 'components/spinner/Spinner';
import { FromTableFormSection } from './FromTableForm/FromTableForm';
import { Notification } from 'components/notification/Notification';
import { ConnectionBase as Connection } from 'types/data-connection.type';

const sourceOption = [
  { value: 'table', label: 'From Table' },
  { value: 'sql', label: 'SQL' },
];

type FormFields = {
  title: string;
  sourceType: string;
  selectedDatabase: string;
  query?: string;
  selectedTable?: string;
};

type InitialData = {
  sourceType?: 'table' | 'sql';
  database?: string;
  table?: string;
  query?: string;
};

type ModalProps = {
  sourceType: 'table' | 'sql';
  connection: Connection;
  initialData?: InitialData;
  onSuccess?: (param?: Dataset) => void;
  onFailure?: (errorMsg: string) => void;
};

const CreateDatasetModal = (): JSX.Element | null => {
  // This helps to initialize the tableList selector - use the initial value only once
  // Set the initial table value only once after the first tablelist load happened.
  const isTableInitializationInProgressRef = useRef(true);

  // Helps to read from the nested expressionEditor withour prop/state drilling
  const [modalProps, setModalProps] = useState<ModalProps | undefined>();
  const [error, setError] = useState<Error>();

  const methods = useForm<FormFields>();
  const {
    watch,
    control,
    reset,
    formState: { errors, isValid },
    handleSubmit,
    setValue,
  } = methods;

  const source = watch('sourceType');
  const selectedDatabase = watch('selectedDatabase');

  const resetModal = () => {
    setModalProps(undefined);
    setError(undefined);
    isTableInitializationInProgressRef.current = true;
    reset({
      title: '',
      sourceType: 'table',
      selectedDatabase: undefined,
      selectedTable: undefined,
    });
  };

  const {
    data: databaseList,
    error: databaseLoadError,
    isFetching: isDatabaseListLoading,
  } = useQuery({
    queryKey: ['databaseList', modalProps?.connection?.id],
    queryFn: () => getDatabases(modalProps?.connection?.id),
    enabled: modalProps?.connection?.id !== undefined,
    onError(err) {
      setError(transformUnknownError(err, 'Failed to load databases'));
    },
  });

  const {
    data: tableList,
    error: tableLoadError,
    isFetching: isTableListLoading,
  } = useQuery({
    queryKey: ['tableList', modalProps?.connection.id, selectedDatabase],
    queryFn: () => getDatatables(selectedDatabase, modalProps?.connection.id),
    onError(err) {
      setError(transformUnknownError(err, 'Failed to load tables'));
    },
    onSuccess() {
      setError(undefined);
    },
    enabled:
      selectedDatabase !== undefined && modalProps?.connection.id !== undefined,
  });

  const onSuccess = (dataset: Dataset) => {
    if (modalProps?.onSuccess) {
      modalProps.onSuccess(dataset);
    } else {
      openInNewWindow(
        `${
          getGlobalArc().frontend.globals.urls.datasetEditUrl
        }/${dataset?.dataset_id()}`
      );
    }
    resetModal();
  };

  const { mutateAsync: mutateDataset, isLoading } = useMutation<
    Dataset | undefined,
    unknown,
    FormFields
  >({
    mutationFn: async (fieldValues) =>
      modalProps?.connection &&
      (await (source === 'table'
        ? createDatasetFromTable(
            modalProps.connection.id,
            modalProps.connection.type,
            fieldValues.title,
            fieldValues.selectedDatabase!,
            fieldValues.selectedTable!
          )
        : createDatasetFromSql(
            modalProps.connection.id,
            modalProps.connection.type,
            fieldValues.title,
            fieldValues.query!.replace(/;$/, '')
          ))),
    onSuccess: (data, variables, context) => data && onSuccess(data),
    onError: (error, variables, context) => {
      const e = transformUnknownError(error, 'Failed to create dataset');
      setError(e);
      modalProps?.onFailure?.(e.message);
    },
  });

  useEffect(() => {
    const setTable = () => {
      if (tableList && !isTableInitializationInProgressRef.current) {
        setValue(
          'selectedTable',
          tableList.length === 1 ? tableList[0] : undefined
        );
      }

      if (tableList && isTableInitializationInProgressRef.current) {
        isTableInitializationInProgressRef.current = false;
      }
    };
    setTable();
  }, [setValue, tableList]);

  const onConnectionIdChange = useCallback(
    ({ payload }: { payload?: ModalProps }) => {
      if (payload) {
        if (payload.initialData?.query) {
          const defaultDatabase = determineDatabaseInQuery(
            payload.initialData.query
          );
          payload.initialData.database = defaultDatabase;
        }

        setModalProps(payload);
        reset({
          title: '',
          sourceType: payload.initialData?.sourceType ?? 'table',
          selectedDatabase: payload.initialData?.database,
          selectedTable: payload.initialData?.table,
        });
      }
    },
    [setModalProps, reset]
  );
  useLegacyBridge('open-create-dataset-modal', onConnectionIdChange);

  const props = {
    databaseList:
      databaseList ??
      (modalProps?.initialData?.database
        ? [modalProps.initialData.database]
        : []),
    tableList,
    isTableListLoading,
    isDatabaseListLoading,
    databaseLoadError,
    tableLoadError,
    connection: modalProps?.connection,
    // Propagate the default values at first!
    defaultDatabase: isTableInitializationInProgressRef.current
      ? modalProps?.initialData?.database
      : undefined,
    defaultTable: isTableInitializationInProgressRef.current
      ? modalProps?.initialData?.table
      : undefined,
    control,
  };
  return modalProps === undefined ? null : (
    <Modal
      show={!!modalProps}
      size={`${source === 'sql' ? 'lg' : 'sm'}`}
      data-test-id="dataset-create-modal"
      dialogClassName={styles.maxWidthTransition}
    >
      <Modal.Header>
        <Modal.Title>Create New Dataset</Modal.Title>
      </Modal.Header>
      <FormProvider {...methods}>
        <Form onSubmit={handleSubmit((data) => mutateDataset(data))}>
          <Modal.Body>
            <Notification
              title="Error"
              message={
                (tableLoadError as Error)?.message ||
                (databaseLoadError as Error)?.message ||
                error?.message ||
                ''
              }
              variant="danger"
              show={!!error && !isTableListLoading && !isDatabaseListLoading}
              dismissible
              onClose={() => setError(undefined)}
            />

            <div>
              Create a dataset from data on this connection. You need to create
              a dataset before you can create dashboards or apps.
            </div>
            <br></br>
            <Controller
              control={control}
              rules={{
                required: true,
              }}
              render={({ field: { onChange, onBlur, value } }) => (
                <Form.Group data-test-id="title">
                  <Form.Label htmlFor={'title'}>
                    Dataset Title<span className="required">*</span>
                  </Form.Label>

                  <InputGroup>
                    <Form.Control
                      id={'title'}
                      data-test-id={'dataset-title-input'}
                      type={'text'}
                      placeholder={'Enter a title'}
                      onBlur={onBlur}
                      onChange={onChange}
                      value={value}
                      isInvalid={!!errors.title}
                    ></Form.Control>
                    <Form.Control.Feedback type="invalid">
                      {errors?.title?.type === 'required' &&
                        'This field is required'}
                    </Form.Control.Feedback>
                  </InputGroup>
                </Form.Group>
              )}
              name="title"
            />

            <br></br>

            <Controller
              control={control}
              render={({ field: { onChange, value } }) => (
                <Form.Group data-test-id="dataset-source-type-select">
                  <Form.Label htmlFor={'sourceType'}>Dataset Source</Form.Label>
                  <Select
                    options={sourceOption}
                    defaultValue={sourceOption.find((x) => x.value === value)}
                    id={'sourceType'}
                    onChange={(e) => {
                      onChange(e?.value);
                    }}
                  />
                </Form.Group>
              )}
              name="sourceType"
            />

            <br></br>
            {source === 'table' && (
              <FromTableFormSection {...props}></FromTableFormSection>
            )}
            {source === 'sql' && (
              <Controller
                control={control}
                rules={{ required: true }}
                defaultValue={modalProps?.initialData?.query}
                render={({ field: { onChange, value } }) => (
                  <SQLExplorer {...props} query={value} onChange={onChange} />
                )}
                name="query"
              />
            )}
          </Modal.Body>

          <Modal.Footer>
            <div className={`position-absolute ${styles.left0}`}>
              {(isTableListLoading || isDatabaseListLoading) && <Spinner />}
            </div>
            <VizButton
              color="secondary"
              size="medium"
              onClick={() => {
                resetModal();
              }}
              testId="close-modal"
            >
              Close
            </VizButton>
            <LoadingIndicatorActionButton
              type="submit"
              isValid={isValid}
              isSaving={isLoading}
              label="Create"
            ></LoadingIndicatorActionButton>
          </Modal.Footer>
        </Form>
      </FormProvider>
    </Modal>
  );
};

export { CreateDatasetModal };
