import { ToggleDropdownInput } from 'components/toggle-dropdown-input/ToggleDropdownInput';
import React, { useCallback, useEffect, useMemo, useState } from 'react';
import { CronPartInput } from './CronPartInput';
import { CronPeriod, CronUnitType, periodOptions } from './constants';
import {
  getCronStringFromValues,
  getPeriodFromCronParts,
  parseCronString,
} from './converter';
import classNames from 'classnames';

interface CronProps {
  value: string;
  onChange: (value: string) => void;
  className?: string;
}

export const Cron = ({ value, onChange, className }: CronProps) => {
  const [period, setPeriod] = useState(CronPeriod.Minute);
  const parts = useMemo(() => parseCronString(value), [value]);

  const onPartChange = useCallback(
    (part: number[], unit: CronUnitType) => {
      const newParts = { ...parts, [unit]: part };

      const cronString = getCronStringFromValues(
        period,
        newParts.months,
        newParts.monthDays,
        newParts.weekDays,
        newParts.hours,
        newParts.minutes
      );

      onChange(cronString);
    },
    [onChange, parts, period]
  );

  const onPeriodChange = useCallback(
    (newPeriod: CronPeriod) => {
      const cronString = getCronStringFromValues(
        newPeriod,
        parts.months,
        parts.monthDays,
        parts.weekDays,
        parts.hours,
        parts.minutes
      );

      onChange(cronString);
      setPeriod(newPeriod);
    },
    [onChange, parts]
  );

  /**
   * Update the period when the cron value changes
   */
  useEffect(() => {
    const newPeriod = getPeriodFromCronParts([
      parts.minutes,
      parts.hours,
      parts.monthDays,
      parts.months,
      parts.weekDays,
    ]);

    const periods = Object.values(CronPeriod);
    const oldPeriodIndex = periods.findIndex((item) => item === period);
    const newPeriodIndex = periods.findIndex((item) => item === newPeriod);

    // Only update if the new period is a broader category than the old one
    if (oldPeriodIndex < newPeriodIndex) {
      setPeriod(newPeriod);
    }
  }, [parts, period]);

  return (
    <div
      className={classNames(
        'd-flex gap-2 flex-wrap align-items-baseline',
        className
      )}
    >
      <div>Every</div>
      <ToggleDropdownInput
        id="period"
        label={period}
        options={periodOptions}
        value={period}
        onChange={onPeriodChange}
        isMulti={false}
      />
      {[CronPeriod.Year].includes(period) && (
        <>
          <div>in</div>
          <CronPartInput
            unitType={CronUnitType.Months}
            value={parts[CronUnitType.Months]}
            onChange={onPartChange}
          />
        </>
      )}
      {[CronPeriod.Year, CronPeriod.Month].includes(period) && (
        <>
          <div>on</div>
          <CronPartInput
            unitType={CronUnitType.MonthDays}
            value={parts[CronUnitType.MonthDays]}
            onChange={onPartChange}
          />
        </>
      )}
      {[CronPeriod.Year, CronPeriod.Month, CronPeriod.Week].includes(
        period
      ) && (
        <>
          <div>{period === CronPeriod.Week ? 'on' : 'and'}</div>
          <CronPartInput
            unitType={CronUnitType.WeekDays}
            value={parts[CronUnitType.WeekDays]}
            onChange={onPartChange}
          />
        </>
      )}
      {![CronPeriod.Hour, CronPeriod.Minute].includes(period) && (
        <>
          <div>at</div>
          <CronPartInput
            unitType={CronUnitType.Hours}
            value={parts[CronUnitType.Hours]}
            onChange={onPartChange}
          />
        </>
      )}
      {![CronPeriod.Minute].includes(period) && (
        <>
          <div>{period === CronPeriod.Hour ? 'at' : ':'}</div>
          <CronPartInput
            unitType={CronUnitType.Minutes}
            value={parts[CronUnitType.Minutes]}
            onChange={onPartChange}
          />
          <div>minute(s) UTC</div>
        </>
      )}
    </div>
  );
};
