import { CronPeriod, UNITS } from './constants';
import {
  getCronStringFromValues,
  getPeriodFromCronParts,
  parseCronString,
  partToString,
} from './converter';

describe('getCronStringFromValues() function', () => {
  it.each([
    [CronPeriod.Day, [], [], [], [], [], '* * * * *'], // empty lists
    [
      CronPeriod.Day,
      undefined,
      undefined,
      undefined,
      undefined,
      undefined,
      '* * * * *',
    ], // no data
    [CronPeriod.Day, [1], [2], [3], [4], [5], '5 4 * * *'], // period is day, rest should be ignored
    [CronPeriod.Year, [1], [2], [3], [4], [5], '5 4 2 1 3'], // period is year, everything should be taken into consideration
    [CronPeriod.Day, [], [], [], [1, 2, 3, 4], [1, 2, 10], '1-2,10 1-4 * * *'], // range values
    [CronPeriod.Year, [1, 4, 7, 10], [], [], [], [], '* * * */3 *'], // quarterly
  ])(
    'returns the correct cron string based on the period values - %#',
    (period, months, monthDays, weekDays, hours, minutes, result) => {
      expect(
        getCronStringFromValues(
          period,
          months,
          monthDays,
          weekDays,
          hours,
          minutes
        )
      ).toBe(result);
    }
  );
});

describe('partToString() function', () => {
  it.each([
    [[1, 3], UNITS[0], false, '1,3'], // simple
    [[1, 2, 3, 5], UNITS[1], false, '1-3,5'], // range
    [[1, 4, 7], UNITS[3], false, '1-7/3'], // month interval
    [[1, 3], UNITS[0], true, '01, 03'], // simple humanized
    [[1, 2, 3, 5], UNITS[0], true, '01-03, 05'], // range humanized
    [[1, 4, 7], UNITS[3], true, 'Jan-Jul/3'], // month interval humanized
  ])(
    'returns the correct part string based on the parts - %#',
    (cronPart, unit, humanize, result) => {
      expect(partToString(cronPart, unit, humanize)).toBe(result);
    }
  );
});

describe('getPeriodFromCronParts() function', () => {
  it.each([
    [[[1], [2], [3], [4], [5]], CronPeriod.Year],
    [[[1], [2], [3], [], [5]], CronPeriod.Month],
    [[[1], [2], [], [], [5]], CronPeriod.Week],
    [[[1], [2], [], [], []], CronPeriod.Day],
    [[[1], [], [], [], []], CronPeriod.Hour],
    [[[], [], [], [], []], CronPeriod.Minute],
  ])(
    'returns the correct period based on the parts - %#',
    (cronParts, result) => {
      expect(getPeriodFromCronParts(cronParts)).toBe(result);
    }
  );
});

describe('parseCronString() function', () => {
  it.each([
    [
      '* * * * *',
      { minutes: [], hours: [], monthDays: [], months: [], weekDays: [] },
    ], // every minute
    [
      '5 4 * * *',
      { minutes: [5], hours: [4], monthDays: [], months: [], weekDays: [] },
    ], // period is day
    [
      '5 4 2 1 3',
      {
        minutes: [5],
        hours: [4],
        monthDays: [2],
        months: [1],
        weekDays: [3],
      },
    ], // period is year
    [
      '1-2,10 1-4 * * *',
      {
        minutes: [1, 2, 10],
        hours: [1, 2, 3, 4],
        monthDays: [],
        months: [],
        weekDays: [],
      },
    ], // range values
    [
      '* * * */3 *',
      {
        minutes: [],
        hours: [],
        monthDays: [],
        months: [1, 4, 7, 10],
        weekDays: [],
      },
    ], // quarterly
    [
      '* * * * 7',
      { minutes: [], hours: [], monthDays: [], months: [], weekDays: [0] },
    ], // sunday
  ])(
    'returns the correct period based on the parts - %s',
    (cronString, result) => {
      expect(parseCronString(cronString)).toEqual(result);
    }
  );

  it.each([
    '* * * * * * * *', // too many parts
    '* * * */2/3 *', // invalid step
    '1/0 * * * *', // invalid step
    '* * * 100 *', // value out of range
    'x-y * * * *', // value not number
    '3-1 * * * *', // incorrect min-max values
    '1-2-3 * * * *', // invalid range
  ])('throws error if input is not a valid cron string', (cronString) => {
    expect(() => parseCronString(cronString)).toThrow();
  });
});
