import { CronPeriod, CronUnitType, UNITS } from './constants';
import { Unit } from './constants';

/**
 * Creates an array of integers from start to end, inclusive
 */
function range(start: number, end: number) {
  const array: number[] = [];

  for (let i = start; i <= end; i++) {
    array.push(i);
  }

  return array;
}

/**
 * Sorts an array of numbers
 */
function sort(array: number[]) {
  array.sort(function (a, b) {
    return a - b;
  });

  return array;
}

/**
 * Get cron string from values
 */
export function getCronStringFromValues(
  period: CronPeriod,
  months: number[] | undefined,
  monthDays: number[] | undefined,
  weekDays: number[] | undefined,
  hours: number[] | undefined,
  minutes: number[] | undefined
) {
  const newMonths = period === CronPeriod.Year && months ? months : [];
  const newMonthDays =
    (period === CronPeriod.Year || period === CronPeriod.Month) && monthDays
      ? monthDays
      : [];
  const newWeekDays =
    (period === CronPeriod.Year ||
      period === CronPeriod.Month ||
      period === CronPeriod.Week) &&
    weekDays
      ? weekDays
      : [];
  const newHours =
    period !== CronPeriod.Minute && period !== CronPeriod.Hour && hours
      ? hours
      : [];
  const newMinutes = period !== CronPeriod.Minute && minutes ? minutes : [];

  const parsedArray = parseCronArray([
    newMinutes,
    newHours,
    newMonthDays,
    newMonths,
    newWeekDays,
  ]);

  return parsedArray.join(' ');
}

/**
 * Returns the cron part array as a string.
 */
export function partToString(cronPart: number[], unit: Unit, humanize = false) {
  let stringifiedPart = '';

  if (isFull(cronPart, unit) || cronPart.length === 0) {
    stringifiedPart = '*';
  } else {
    const step = getStep(cronPart);

    if (step && isInterval(cronPart, step)) {
      if (isFullInterval(cronPart, unit, step)) {
        stringifiedPart = `*/${step}`;
      } else {
        stringifiedPart = `${formatValue(
          getMin(cronPart),
          unit,
          humanize
        )}-${formatValue(getMax(cronPart), unit, humanize)}/${step}`;
      }
    } else {
      stringifiedPart = toRanges(cronPart)
        .map((range: number | number[]) => {
          if (Array.isArray(range)) {
            return `${formatValue(range[0], unit, humanize)}-${formatValue(
              range[1],
              unit,
              humanize
            )}`;
          }

          return formatValue(range, unit, humanize);
        })
        .join(humanize ? ', ' : ',');
    }
  }
  return stringifiedPart;
}

/**
 * Format the value
 */
function formatValue(value: number, unit: Unit, humanize = false) {
  let cronPartString = value.toString();
  const { type } = unit;
  const needLeadingZero =
    humanize && [CronUnitType.Hours, CronUnitType.Minutes].includes(type);

  if (humanize && [CronUnitType.WeekDays, CronUnitType.Months].includes(type)) {
    cronPartString =
      unit.options.find((item) => item.value === value)?.label?.slice(0, 3) ??
      cronPartString;
  }

  if (value < 10 && needLeadingZero) {
    cronPartString = cronPartString.padStart(2, '0');
  }

  return cronPartString;
}

/**
 * Validates a range of positive integers
 */
function parsePartArray(arr: number[], unit: Unit) {
  const values = sort(Array.from(new Set(fixSunday(arr, unit))));

  if (values.length === 0) {
    return values;
  }

  const value = outOfRange(values, unit);

  if (typeof value !== 'undefined') {
    throw new Error(`Value "${value}" out of range for ${unit.type}`);
  }

  return values;
}

/**
 * Parses a 2-dimensional array of integers as a cron schedule
 */
function parseCronArray(cronArr: number[][]) {
  return cronArr.map((partArr, idx) => {
    const unit = UNITS[idx];
    const parsedArray = parsePartArray(partArr, unit);

    return partToString(parsedArray, unit);
  });
}

/**
 * Find the period from cron parts
 */
export function getPeriodFromCronParts(cronParts: number[][]): CronPeriod {
  if (cronParts[3].length > 0) {
    return CronPeriod.Year;
  } else if (cronParts[2].length > 0) {
    return CronPeriod.Month;
  } else if (cronParts[4].length > 0) {
    return CronPeriod.Week;
  } else if (cronParts[1].length > 0) {
    return CronPeriod.Day;
  } else if (cronParts[0].length > 0) {
    return CronPeriod.Hour;
  }
  return CronPeriod.Minute;
}

/**
 * Parses a cron string to an object of parts
 */
export function parseCronString(str: string) {
  const parts = str.replace(/\s+/g, ' ').trim().split(' ');

  if (parts.length === 5) {
    return parts.reduce((acc, partStr, idx) => {
      acc[UNITS[idx].type] = parsePartString(partStr, UNITS[idx]);
      return acc;
    }, {} as Record<CronUnitType, number[]>);
  }

  throw new Error('Invalid cron string format');
}

/**
 * Parses a string as a range of positive integers
 */
export function parsePartString(str: string, unit: Unit) {
  if (str === '*' || str === '*/1') {
    return [];
  }

  const values = sort(
    Array.from(
      new Set(
        fixSunday(
          str
            .split(',')
            .map((value) => {
              const valueParts = value.split('/');

              if (valueParts.length > 2) {
                throw new Error(`Invalid value "${str} for "${unit.type}"`);
              }

              let parsedValues: number[];
              const left = valueParts[0];
              const right = valueParts[1];

              if (left === '*') {
                parsedValues = range(unit.min, unit.max);
              } else {
                parsedValues = parseRange(left, str, unit);
              }

              const step = parseStep(right, unit);
              const intervalValues = applyInterval(parsedValues, step);

              return intervalValues;
            })
            .flat(),
          unit
        )
      )
    )
  );

  const value = outOfRange(values, unit);

  if (typeof value !== 'undefined') {
    throw new Error(`Value "${value}" out of range for ${unit.type}`);
  }

  // Prevent to return full array
  // If all values are selected we don't want any selection visible
  if (values.length === unit.total) {
    return [];
  }

  return values;
}

/**
 * Replace all 7 with 0 as Sunday can be represented by both
 */
function fixSunday(values: number[], unit: Unit) {
  if (unit.type === CronUnitType.WeekDays) {
    values = values.map(function (value) {
      if (value === 7) {
        return 0;
      }

      return value;
    });
  }

  return values;
}

/**
 * Parses a range string
 */
function parseRange(rangeStr: string, context: string, unit: Unit) {
  const subparts = rangeStr.split('-');

  if (subparts.length === 1) {
    const value = Number(subparts[0]);

    if (isNaN(value)) {
      throw new Error(`Invalid value "${context}" for ${unit.type}`);
    }

    return [value];
  } else if (subparts.length === 2) {
    const minValue = Number(subparts[0]);
    const maxValue = Number(subparts[1]);

    if (isNaN(minValue) || isNaN(maxValue)) {
      throw new Error(`Invalid value "${context}" for ${unit.type}`);
    }

    // Fix to allow equal min and max range values
    // cf: https://github.com/roccivic/cron-converter/pull/15
    if (maxValue < minValue) {
      throw new Error(
        `Max range is less than min range in "${rangeStr}" for ${unit.type}`
      );
    }

    return range(minValue, maxValue);
  } else {
    throw new Error(`Invalid value "${rangeStr}" for ${unit.type}`);
  }
}

/**
 * Finds an element from values that is outside of the range of unit
 */
function outOfRange(values: number[], unit: Unit) {
  const first = values[0];
  const last = values[values.length - 1];

  if (first < unit.min) {
    return first;
  } else if (last > unit.max) {
    return last;
  }

  return;
}

/**
 * Parses the step from a part string
 */
function parseStep(step: string, unit: Unit) {
  if (typeof step !== 'undefined') {
    const parsedStep = Number(step);

    if (isNaN(parsedStep) || parsedStep < 1) {
      throw new Error(`Invalid interval step value "${step}" for ${unit.type}`);
    }

    return parsedStep;
  }
}

/**
 * Applies an interval step to a collection of values
 */
function applyInterval(values: number[], step?: number) {
  if (step) {
    const minVal = values[0];

    values = values.filter((value) => {
      return value % step === minVal % step || value === minVal;
    });
  }

  return values;
}

/**
 * Returns true if range has all the values of the unit
 */
function isFull(values: number[], unit: Unit) {
  return values.length === unit.max - unit.min + 1;
}

/**
 * Returns the difference between first and second elements in the range
 */
function getStep(values: number[]) {
  if (values.length > 2) {
    const step = values[1] - values[0];

    if (step > 1) {
      return step;
    }
  }
}

/**
 * Returns true if the range can be represented as an interval
 */
function isInterval(values: number[], step: number) {
  for (let i = 1; i < values.length; i++) {
    const prev = values[i - 1];
    const value = values[i];

    if (value - prev !== step) {
      return false;
    }
  }

  return true;
}

/**
 * Returns true if the range contains all the interval values
 */
function isFullInterval(values: number[], unit: Unit, step: number) {
  const min = getMin(values);
  const max = getMax(values);
  const haveAllValues = values.length === (max - min) / step + 1;

  if (min === unit.min && max + step > unit.max && haveAllValues) {
    return true;
  }

  return false;
}

/**
 * Returns the smallest value in the range
 */
function getMin(values: number[]) {
  return values[0];
}

/**
 * Returns the largest value in the range
 */
function getMax(values: number[]) {
  return values[values.length - 1];
}

/**
 * Returns the range as an array of ranges
 * defined as arrays of positive integers
 */
function toRanges(values: number[]) {
  const ranges: (number[] | number)[] = [];
  let startPart: number | null = null;

  values.forEach((value, index, self) => {
    if (value !== self[index + 1] - 1) {
      if (startPart !== null) {
        ranges.push([startPart, value]);
        startPart = null;
      } else {
        ranges.push(value);
      }
    } else if (startPart === null) {
      startPart = value;
    }
  });

  return ranges;
}
