import * as React from 'react';
import '@testing-library/jest-dom';
import {
  render,
  screen,
  fireEvent,
  waitFor,
  configure,
} from '@testing-library/react';
import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import { DataConnectionModal } from './DataConnectionModal';
import { getAllDataConnections } from 'apis/get-all-data-connections';
import { createOrUpdateConnection } from 'apis/create-or-update-data-connection';
import { allConnectionTypesMock, mockDataconnections } from './utils/mocks';
import {
  DataConnectionInfo,
  DataConnectionType,
} from '../../types/data-connection.type';

jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
}));

configure({ testIdAttribute: 'data-test-id' });

const mockAddNotification = jest.fn();
jest.mock('../../../src/stores/use-notification-store.ts', () => ({
  useNotificationStore: () => ({ addNotification: mockAddNotification }),
}));

jest.mock('apis/get-all-data-connections.ts', () => ({
  getAllDataConnections: jest.fn(),
}));

jest.mock('../../apis/create-or-update-data-connection', () => ({
  createOrUpdateConnection: jest.fn(),
}));

jest.mock('utils/get-global-arc', () => {
  const arcMock = {
    frontend: {
      globals: {
        urls: {
          viewDashboardUrl: 'customUrl',
        },
        settings: {
          dataConnectionConcurrencyDefault: 100,
          dataConnectionConcurrencyUserDefault: 5,
          dataConnectionQueryTimeoutDefault: 10,
          uploadMaxrowsDefault: 1000000,
        },
      },
      legacyBridge: {
        events: {
          next: jest.fn(),
        },
      },
    },
    misc: {
      getBlankTargetUrl: jest.fn().mockReturnValue('customUrl'),
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

beforeEach(() => {
  jest.mocked(getAllDataConnections).mockResolvedValue(mockDataconnections);
});

afterEach(() => {
  jest.resetAllMocks();
});

const queryClient = new QueryClient();

const renderModalAndTestConnection = async ({
  connectionId,
  connectionType,
  connectionName,
  connectionTypeName,
  dataconnectionParams,
  testIdMappings,
}: {
  connectionId: number;
  connectionType: DataConnectionType;
  connectionName: string;
  connectionTypeName: string;
  dataconnectionParams: DataConnectionInfo['PARAMS'];
  testIdMappings: Record<string, string>;
}) => {
  const defaultProps = {
    isShow: true,
    dataconnectionId: connectionId,
    associatedDatasetLength: 0,
    allConnectionTypes: allConnectionTypesMock,
    modalType: 'existing' as const,
    onClose: jest.fn(),
  };

  const dataConnectionData = {
    dataconnectionData: {
      id: connectionId,
      type: connectionType,
      name: connectionName,
      info: { PARAMS: dataconnectionParams },
      typeName: connectionTypeName,
      password: undefined,
    },
    shouldValidate: 'true',
  };

  render(
    <QueryClientProvider client={queryClient}>
      <DataConnectionModal {...defaultProps} />
    </QueryClientProvider>
  );

  // Ensure the data connection mock function is called
  await waitFor(() => {
    expect(getAllDataConnections).toHaveBeenCalled();
  });

  // Validate form fields
  await waitFor(() => {
    expect(screen.getByTestId('connectionType')).toHaveTextContent(
      dataConnectionData.dataconnectionData.typeName
    );
  });

  await waitFor(() => {
    expect(screen.getByTestId('connectionName')).toHaveValue(connectionName);
  });

  for (const [testId, expectedValue] of Object.entries(testIdMappings)) {
    await waitFor(() => {
      expect(screen.getByTestId(testId)).toHaveValue(expectedValue);
    });
  }

  fireEvent.click(screen.getByRole('button', { name: 'Test' }));

  // I chose to use toHaveBeenCalledWith(expectedObject) because it compares
  // the entire structure of expectedObject and receivedObject, while
  // expect.objectContaining({ ... }) only checks for required fields, ignoring extra properties.
  await waitFor(() => {
    expect(createOrUpdateConnection).toHaveBeenCalledWith(dataConnectionData);
  });

  await waitFor(() => {
    expect(createOrUpdateConnection).toHaveBeenCalledTimes(1);
  });
};

describe('Data Connection Modal component', () => {
  const defaultProps = {
    isShow: true,
    dataconnectionId: null,
    associatedDatasetLength: 0,
    allConnectionTypes: allConnectionTypesMock,
    modalType: 'new' as const,
    onClose: jest.fn(),
  };

  it('renders modal correctly', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <DataConnectionModal {...defaultProps} />
      </QueryClientProvider>
    );
  });

  it('switches tabs correctly', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <DataConnectionModal {...defaultProps} />
      </QueryClientProvider>
    );

    const advancedTab = screen.getByRole('tab', { name: /Advanced/i });
    const parametersTab = screen.getByRole('tab', { name: /Parameters/i });
    const dataTab = screen.getByRole('tab', { name: /Data/i });

    fireEvent.click(advancedTab);

    await waitFor(() => {
      expect(advancedTab).toHaveAttribute('aria-selected', 'true');
    });

    fireEvent.click(parametersTab);

    await waitFor(() => {
      expect(parametersTab).toHaveAttribute('aria-selected', 'true');
    });

    fireEvent.click(dataTab);

    await waitFor(() => {
      expect(dataTab).toHaveAttribute('aria-selected', 'true');
    });
  });

  it('validates form fields before submission', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <DataConnectionModal {...defaultProps} />
      </QueryClientProvider>
    );

    const submitButton = screen.getByRole('button', { name: 'Connect' });
    expect(submitButton).not.toBeDisabled();
    fireEvent.click(submitButton);

    const connectionNameError = screen.getByTestId('connectionNameError');
    await waitFor(() => {
      expect(connectionNameError).toHaveTextContent(
        'Connection name is required'
      );
    });
    expect(jest.mocked(createOrUpdateConnection)).not.toHaveBeenCalled();
  });

  it('renders new data connection modal correctly', async () => {
    const defaultProps = {
      isShow: true,
      dataconnectionId: null,
      associatedDatasetLength: 0,
      allConnectionTypes: allConnectionTypesMock,
      modalType: 'new' as const,
      onClose: jest.fn(),
    };

    render(
      <QueryClientProvider client={queryClient}>
        <DataConnectionModal {...defaultProps} />
      </QueryClientProvider>
    );

    // Basic tab
    expect(screen.getByTestId('connectionType')).toHaveTextContent(
      'Cloudera Data Warehouse Impala'
    );
    expect(screen.getByTestId('impyla:cdw-impyla.basic.port')).toHaveValue(
      '443'
    );

    // Advanced tab
    // open Advanced tab
    const advancedTab = screen.getByRole('tab', { name: 'Advanced' });
    fireEvent.click(advancedTab);

    await waitFor(() => {
      expect(advancedTab).toHaveAttribute('aria-selected', 'true');
    });
    // Connection mode
    const binaryConnMode = screen.getByRole('radio', { name: /binary/i });
    const httpConnMode = screen.getByRole('radio', { name: /http/i });

    expect(binaryConnMode).not.toBeChecked();
    expect(httpConnMode).toBeChecked();
    // HTTP Path
    expect(
      screen.getByTestId('impyla:cdw-impyla.advanced.http_path')
    ).toBeInTheDocument();

    // Socket type
    const normalSocketType = screen.getByLabelText('Normal');
    const sslSocketType = screen.getByLabelText('SSL');
    const sslCertSocketType = screen.getByLabelText('SSL with certificate');

    expect(normalSocketType).not.toBeChecked();
    expect(sslSocketType).toBeChecked();
    expect(sslCertSocketType).not.toBeChecked();

    // Authentication mode
    const noSaslAuthMode = screen.getByLabelText('NoSasl');
    const plainAuthMode = screen.getByLabelText('Plain');
    const ldapAuthMode = screen.getByLabelText('LDAP');
    const kerberosAuthMode = screen.getByLabelText('Kerberos');

    expect(noSaslAuthMode).not.toBeChecked();
    expect(plainAuthMode).not.toBeChecked();
    expect(ldapAuthMode).toBeChecked();
    expect(kerberosAuthMode).not.toBeChecked();

    // Socket timeout
    expect(
      screen.getByTestId('impyla:cdw-impyla.advanced.socket_timeout')
    ).toHaveValue('60');

    // Impersonation
    const impersonationCheckbox = screen.getByTestId(
      'impyla:cdw-impyla.advanced.impersonation'
    );
    expect(impersonationCheckbox).not.toBeChecked();

    // Trusted Impersonation
    const trustedImpersonationCheckbox = screen.getByTestId(
      'impyla:cdw-impyla.advanced.trusted_impersonation'
    );
    expect(trustedImpersonationCheckbox).not.toBeChecked();

    // User dependent cache
    const userDependentCacheCheckbox = screen.getByTestId(
      'impyla:cdw-impyla.advanced.user_dependent_cache'
    );
    expect(userDependentCacheCheckbox).toBeDisabled();

    // Data tab
    // open the Data tab
    const dataTab = screen.getByRole('tab', { name: /Data/i });
    fireEvent.click(dataTab);

    await waitFor(() => {
      expect(dataTab).toHaveAttribute('aria-selected', 'true');
    });

    // Concurrency
    const concurrency = screen.getByTestId(
      'impyla:cdw-impyla.data.concurrency'
    );
    expect(concurrency).toHaveValue('100');

    // Concurrency per user
    const concurrencyPerUser = screen.getByTestId(
      'impyla:cdw-impyla.data.concurrency_per_person'
    );
    expect(concurrencyPerUser).toHaveValue('5');

    // Query timeout
    const queryTimeout = screen.getByTestId(
      'impyla:cdw-impyla.data.query_timeout'
    );
    expect(queryTimeout).toHaveValue('10');

    // Query loading warning
    const queryLoadingWarning = screen.getByTestId(
      'impyla:cdw-impyla.data.query_loading_warning'
    );
    expect(queryLoadingWarning).toHaveValue('20');

    // Row upload limit
    const rowUploadLimit = screen.getByTestId(
      'impyla:cdw-impyla.data.row_upload_limit'
    );
    expect(rowUploadLimit).toHaveValue('1000000');

    // Result cache
    const resultCache = screen.getByTestId(
      'impyla:cdw-impyla.data.result_cache'
    );
    expect(resultCache).not.toBeChecked();

    // Cache retention
    const cacheRetention = screen.getByTestId(
      'impyla:cdw-impyla.data.cache_retention_timeout'
    );
    expect(cacheRetention).toHaveValue('86400');
  });

  it('render edit dataconnection modal correctly - PostgreSQL', async () => {
    const defaultProps = {
      isShow: true,
      dataconnectionId: 5,
      associatedDatasetLength: 0,
      allConnectionTypes: allConnectionTypesMock,
      modalType: 'existing' as const,
      onClose: jest.fn(),
    };

    render(
      <QueryClientProvider client={queryClient}>
        <DataConnectionModal {...defaultProps} />
      </QueryClientProvider>
    );

    await waitFor(() => {
      // Basic tab
      expect(screen.getByTestId('connectionType')).toHaveTextContent(
        'PostgreSQL'
      );
    });
    await waitFor(() => {
      expect(screen.getByTestId('connectionName')).toHaveValue(
        'test_postgresql'
      );
    });
    await waitFor(() => {
      expect(screen.getByTestId('postgres.basic.hostname')).toHaveValue(
        'viz-team-6.cc-os.cloudera.com'
      );
    });
    await waitFor(() => {
      expect(screen.getByTestId('postgres.basic.port')).toHaveValue('5432');
    });
    await waitFor(() => {
      expect(screen.getByTestId('postgres.basic.database')).toHaveValue(
        'viz_test_db'
      );
    });
    await waitFor(() => {
      expect(screen.getByTestId('postgres.basic.username')).toHaveValue('viz');
    });

    // Advanced tab
    // open Advanced tab
    const advancedTab = screen.getByRole('tab', { name: 'Advanced' });
    fireEvent.click(advancedTab);

    await waitFor(() => {
      expect(advancedTab).toHaveAttribute('aria-selected', 'true');
    });

    // Socket type
    await waitFor(() => {
      const normalSocketType = screen.getByLabelText('Normal');
      expect(normalSocketType).toBeChecked();
    });
    await waitFor(() => {
      const sslSocketType = screen.getByLabelText('SSL');
      expect(sslSocketType).not.toBeChecked();
    });
    await waitFor(() => {
      const sslCertSocketType = screen.getByLabelText('SSL with certificate');
      expect(sslCertSocketType).not.toBeChecked();
    });

    // Data tab
    // open the Data tab
    const dataTab = screen.getByRole('tab', { name: /Data/i });
    fireEvent.click(dataTab);

    await waitFor(() => {
      expect(dataTab).toHaveAttribute('aria-selected', 'true');
    });

    await waitFor(() => {
      // Concurrency
      const concurrency = screen.getByTestId('postgres.data.concurrency');
      expect(concurrency).toHaveValue('100');
    });
    await waitFor(() => {
      // Concurrency per user
      const concurrencyPerUser = screen.getByTestId(
        'postgres.data.concurrency_per_person'
      );
      expect(concurrencyPerUser).toHaveValue('5');
    });
    await waitFor(() => {
      // Query timeout
      const queryTimeout = screen.getByTestId('postgres.data.query_timeout');
      expect(queryTimeout).toHaveValue('10');
    });
    await waitFor(() => {
      // Query loading warning
      const queryLoadingWarning = screen.getByTestId(
        'postgres.data.query_loading_warning'
      );
      expect(queryLoadingWarning).toHaveValue('20');
    });
    await waitFor(() => {
      // Row upload limit
      const rowUploadLimit = screen.getByTestId(
        'postgres.data.row_upload_limit'
      );
      expect(rowUploadLimit).toHaveValue('1000000');
    });
    await waitFor(() => {
      // Result cache
      const resultCache = screen.getByTestId('postgres.data.result_cache');
      expect(resultCache).not.toBeChecked();
    });
    await waitFor(() => {
      // Cache retention
      const cacheRetention = screen.getByTestId(
        'postgres.data.cache_retention_timeout'
      );
      expect(cacheRetention).toHaveValue('86400');
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - PostgreSQL', async () => {
    await renderModalAndTestConnection({
      connectionId: 5,
      connectionType: 'postgres',
      connectionName: 'test_postgresql',
      connectionTypeName: 'PostgreSQL',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        //   ACCOUNT: '',
        //   DRIVER: '',
        //   DSN: '',
        //   Authentication: 'No Authentication',
        //   ARCENGINE_HISTORY_ETL: {
        //     ENABLED: 0,
        //     TARGET_TABLE: 'default.history_table',
        //     JOB_NAME: 'Connection samples history table ETL',
        //     SCHEDULE: 'Weekly',
        //   },
        ARCENGINE_HISTORY_ETL: {},
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        CONNECTIONNAME_EXTENDED: '',
        DATABASE: 'viz_test_db',
        HOST: 'viz-team-6.cc-os.cloudera.com',
        PORT: '5432',
        QUERY_LOADING_WARNING_SECONDS: 20,
        QUERY_TIMEOUT: 10,
        SOCK: 'normal',
        SSLCERT_NAME_MISMATCH: false,
        UPLOAD_MAXROWS: 1000000,
        USERNAME: 'viz',
      },
      testIdMappings: {
        'postgres.basic.hostname': 'viz-team-6.cc-os.cloudera.com',
        'postgres.basic.port': '5432',
        'postgres.basic.database': 'viz_test_db',
        'postgres.basic.username': 'viz',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - SQLite', async () => {
    await renderModalAndTestConnection({
      connectionId: 8,
      connectionType: 'sqlite',
      connectionName: 'test_sqlite',
      connectionTypeName: 'SQLite',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: 'default.history_table',
        //   JOB_NAME: 'Connection samples history table ETL',
        //   SCHEDULE: 'Weekly',
        // },
        CONNECTIONNAME_EXTENDED: '',
        FILE: '/home/cdsw/data.sqlite',
        CONCURRENCY: 10,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 48,
        QUERY_LOADING_WARNING_SECONDS: 25,
        UPLOAD_MAXROWS: 10000,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'sqlite.basic.file_name': '/home/cdsw/data.sqlite',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - MySQL', async () => {
    await renderModalAndTestConnection({
      connectionId: 9,
      connectionType: 'mysql',
      connectionName: 'test_mysql',
      connectionTypeName: 'MySQL',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ACCOUNT: '',
        // DRIVER: '',
        // DSN: '',
        // Authentication: 'No Authentication',
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: 'default.history_table',
        //   JOB_NAME: 'Connection samples history table ETL',
        //   SCHEDULE: 'Weekly',
        // },
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 10,
        CONNECTIONNAME_EXTENDED: '',
        HOST: 'viz-team-6.cc-os.cloudera.com',
        PORT: '3311',
        DATABASE: '',
        USERNAME: 'root',
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 10000,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'mysql.basic.hostname': 'viz-team-6.cc-os.cloudera.com',
        'mysql.basic.port': '3311',
        'mysql.basic.username': 'root',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Maria MySQL', async () => {
    await renderModalAndTestConnection({
      connectionId: 10,
      connectionType: 'mysql',
      connectionName: 'test_maria',
      connectionTypeName: 'MySQL',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ACCOUNT: '',
        // DRIVER: '',
        // DSN: '',
        // Authentication: 'No Authentication',
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: 'default.history_table',
        //   JOB_NAME: 'Connection samples history table ETL',
        //   SCHEDULE: 'Weekly',
        // },
        ARCENGINE_HISTORY_ETL: {},
        CACHE: {
          ENABLED: 0,
          RETENTION: 86400,
        },
        CONNECTIONNAME_EXTENDED: '',
        DATABASE: '',
        HOST: 'viz-team-6.cc-os.cloudera.com',
        PORT: '3310',
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 10,
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 1000000,
        USERNAME: 'root',
      },
      testIdMappings: {
        'mysql.basic.hostname': 'viz-team-6.cc-os.cloudera.com',
        'mysql.basic.port': '3310',
        'mysql.basic.username': 'root',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - MariaDB', async () => {
    await renderModalAndTestConnection({
      connectionId: 12,
      connectionType: 'mariadb',
      connectionName: 'test_mariadb',
      connectionTypeName: 'MariaDB',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ACCOUNT: '',
        // DRIVER: '',
        // DSN: '',
        // Authentication: 'No Authentication',
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: 'default.history_table',
        //   JOB_NAME: 'Connection samples history table ETL',
        //   SCHEDULE: 'Weekly',
        // },
        CONNECTIONNAME_EXTENDED: '',
        HOST: 'viz-team-6.cc-os.cloudera.com',
        PORT: '3310',
        DATABASE: '',
        USERNAME: 'root',
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 10,
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 10000,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'mariadb.basic.hostname': 'viz-team-6.cc-os.cloudera.com',
        'mariadb.basic.port': '3310',
        'mariadb.basic.username': 'root',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Solr', async () => {
    await renderModalAndTestConnection({
      connectionId: 11,
      connectionType: 'solr',
      connectionName: 'test_solr',
      connectionTypeName: 'Solr',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ACCOUNT: '',
        // DATABASE: '',
        // DRIVER: '',
        // DSN: '',
        // Authentication: 'No Authentication',
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: 'default.history_table',
        //   JOB_NAME: 'Connection samples history table ETL',
        //   SCHEDULE: 'Weekly',
        // },
        CONNECTIONNAME_EXTENDED: '',
        HOST: 'viz-team-5.cc-os.cloudera.com',
        PORT: '8983',
        SOLRTYPE: 'cloud',
        SOLR_JOINCHAR: ',',
        USERNAME: 'vizapps_admin',
        SOCK: 'normal',
        SSLCERT_NAME_MISMATCH: false,
        AUTH: 'nosasl',
        KERBSERV: '',
        SOCKET_TIMEOUT: 0,
        IMPERSONATION: false,
        IMPERSONATION_TRUSTED: false,
        SETTINGS: {},
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 10,
        UPLOAD_MAXROWS: 1000000,
        QUERY_LOADING_WARNING_SECONDS: 20,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'solr.basic.hostname': 'viz-team-5.cc-os.cloudera.com',
        'solr.basic.port': '8983',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Snowflake', async () => {
    await renderModalAndTestConnection({
      connectionId: 13,
      connectionType: 'snowflake',
      connectionName: 'test_snowflake',
      connectionTypeName: 'Snowflake',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // HOST: '',
        // DRIVER: '',
        // DSN: '',
        // Authentication: 'No Authentication',
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: 'default.history_table',
        //   JOB_NAME: 'Connection samples history table ETL',
        //   SCHEDULE: 'Weekly',
        // },
        CONNECTIONNAME_EXTENDED: '',
        WAREHOUSE: 'compute_wh',
        DATABASE: 'viz_sample_data',
        USERNAME: 'viz_team',
        ROLE: '',
        ACCOUNT: 'sx41684.us-east-2.aws',
        SETTINGS: {},
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 10,
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 10000,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'snowflake.basic.account_name': 'sx41684.us-east-2.aws',
        'snowflake.basic.warehouse': 'compute_wh',
        'snowflake.basic.database': 'viz_sample_data',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Spark SQL', async () => {
    await renderModalAndTestConnection({
      connectionId: 14,
      connectionType: 'sparksql',
      connectionName: 'test_spark_sql',
      connectionTypeName: 'Spark SQL',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ACCOUNT: '',
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: '',
        //   JOB_NAME: '',
        //   SCHEDULE: 'Weekly',
        // },
        CONNECTIONNAME_EXTENDED: '',
        HOST: 'szhou-datavizimpala-1.vpc.cloudera.com',
        PORT: '10000',
        USERNAME: 'admin',
        MODE: 'binary',
        HS2_HTTP_SQLPATH: '',
        HS2_HTTP_ACCESS_TOKEN: '',
        SOCK: 'normal',
        SSLCERT_NAME_MISMATCH: false,
        AUTH: 'plain',
        KERBSERV: '',
        SOCKET_TIMEOUT: 60,
        IMPERSONATION: false,
        IMPERSONATION_TRUSTED: false,
        SETTINGS: {},
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 10,
        UPLOAD_MAXROWS: 1000000,
        QUERY_LOADING_WARNING_SECONDS: 20,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'sparksql.basic.hostname': 'szhou-datavizimpala-1.vpc.cloudera.com',
        'sparksql.basic.port': '10000',
        'sparksql.basic.username': 'admin',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Trino', async () => {
    await renderModalAndTestConnection({
      connectionId: 15,
      connectionType: 'trino',
      connectionName: 'test_trino',
      connectionTypeName: 'Trino',
      dataconnectionParams: {
        CONNECTIONNAME_EXTENDED: '',
        HOST: '10.96.131.181',
        PORT: '8080',
        USERNAME: 'admin',
        SOCK: 'normal',
        SSLCERT_NAME_MISMATCH: false,
        AUTH: 'nosasl',
        KERBSERV: '',
        SOCKET_TIMEOUT: 60,
        IMPERSONATION: false,
        IMPERSONATION_TRUSTED: false,
        USER_DEPENDENT_CACHE: false,
        CONCURRENCY: 22,
        CONCURRENCY_USER: 22,
        QUERY_TIMEOUT: 22,
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 10000,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'trino.basic.hostname': '10.96.131.181',
        'trino.basic.port': '8080',
        'trino.basic.username': 'admin',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Phoenix', async () => {
    await renderModalAndTestConnection({
      connectionId: 16,
      connectionType: 'phoenix',
      connectionName: 'test_phoenix',
      connectionTypeName: 'Phoenix',
      dataconnectionParams: {
        CONNECTIONNAME_EXTENDED: '',
        HOST: 'viz-team-1.cc-os.cloudera.com',
        PORT: '8765',
        USERNAME: '',
        MODE: 'http',
        HS2_HTTP_SQLPATH: '',
        SOCK: 'sslcert',
        SSLCERT_NAME_MISMATCH: true,
        AUTH: 'kerberos',
        KERBSERV: '',
        SOCKET_TIMEOUT: 60,
        IMPERSONATION: false,
        IMPERSONATION_TRUSTED: false,
        USER_DEPENDENT_CACHE: false,
        CONCURRENCY: 22,
        CONCURRENCY_USER: 22,
        QUERY_TIMEOUT: 22,
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 10000,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'phoenix.basic.hostname': 'viz-team-1.cc-os.cloudera.com',
        'phoenix.basic.port': '8765',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Druid', async () => {
    await renderModalAndTestConnection({
      connectionId: 17,
      connectionType: 'druid',
      connectionName: 'test_druid',
      connectionTypeName: 'Druid',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ACCOUNT: '',
        // DRIVER: '',
        // DSN: '',
        // Authentication: 'No Authentication',
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: '',
        //   JOB_NAME: '',
        //   SCHEDULE: 'Weekly',
        // },
        CONNECTIONNAME_EXTENDED: '',
        HOST: 'viz-team-6.cc-os.cloudera.com',
        PORT: '3311',
        DATABASE: '',
        USERNAME: '',
        SOCK: 'normal',
        SSLCERT_NAME_MISMATCH: false,
        CONCURRENCY: 22,
        CONCURRENCY_USER: 22,
        QUERY_TIMEOUT: 22,
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 1000000,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'druid.basic.hostname': 'viz-team-6.cc-os.cloudera.com',
        'druid.basic.port': '3311',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Impala', async () => {
    await renderModalAndTestConnection({
      connectionId: 3,
      connectionType: 'impyla',
      connectionName: 'test_impala',
      connectionTypeName: 'Impala',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ACCOUNT: '',
        // HS2_HTTP_ACCESS_TOKEN: '',
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: '',
        //   JOB_NAME: '',
        //   SCHEDULE: 'Weekly',
        // },
        AUTH: 'nosasl',
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        CONNECTIONNAME_EXTENDED: '',
        MODE: 'binary',
        PORT: '21050',
        QUERY_LOADING_WARNING_SECONDS: 20,
        QUERY_TIMEOUT: 10,
        SOCK: 'normal',
        SOCKET_TIMEOUT: 60,
        SSLCERT_NAME_MISMATCH: false,
        UPLOAD_MAXROWS: 1000000,
        USERNAME: '',
        APP_NAME: 'viz',
        ARCENGINE_HISTORY_ETL: {},
        HOST: 'szhou-datavizimpala-3.vpc.cloudera.com',
        HS2_HTTP_SQLPATH: '',
        IMPERSONATION: false,
        IMPERSONATION_TRUSTED: false,
        KERBSERV: '',
        SETTINGS: {},
        USER_DEPENDENT_CACHE: false,
      },
      testIdMappings: {
        'impyla.basic.hostname': 'szhou-datavizimpala-3.vpc.cloudera.com',
        'impyla.basic.port': '21050',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Hive', async () => {
    await renderModalAndTestConnection({
      connectionId: 1,
      connectionType: 'hyve',
      connectionName: 'test_hive',
      connectionTypeName: 'Hive',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ACCOUNT: '',
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: 'default.history_table',
        //   JOB_NAME: 'Connection Hive Judit history table ETL',
        //   SCHEDULE: 'Weekly',
        // },
        CONNECTIONNAME_EXTENDED: '',
        HOST: 'szhou-datavizimpala-1.vpc.cloudera.com',
        PORT: '10000',
        USERNAME: 'admin',
        MODE: 'binary',
        HS2_HTTP_SQLPATH: 'gateway/cdp-proxy-api/hive',
        HS2_HTTP_ACCESS_TOKEN: '',
        SOCK: 'normal',
        SSLCERT_NAME_MISMATCH: true,
        AUTH: 'plain',
        KERBSERV: '',
        SOCKET_TIMEOUT: 60,
        IMPERSONATION: false,
        IMPERSONATION_TRUSTED: false,
        USER_DEPENDENT_CACHE: false,
        APP_NAME: 'viz134',
        SETTINGS: {
          'hive.server2.async.exec.async.compile': 'false',
          ffg: 'juyu',
        },
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 10,
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 1000000,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'hyve.basic.hostname': 'szhou-datavizimpala-1.vpc.cloudera.com',
        'hyve.basic.port': '10000',
        'hyve.basic.username': 'admin',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Cloudera Data Warehouse Impala', async () => {
    await renderModalAndTestConnection({
      connectionId: 7,
      connectionType: 'impyla',
      connectionName: 'test_cdw_impala',
      connectionTypeName: 'Cloudera Data Warehouse Impala',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: '',
        //   JOB_NAME: '',
        //   SCHEDULE: 'Weekly',
        // },
        CONNECTIONNAME_EXTENDED: 'cdw-impyla',
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 10,
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 1000000,
        APP_NAME: 'viz',
        ACCOUNT: undefined,
        HOST: 'coordinator.impala-1739180273-9kcj.svc.cluster.local',
        PORT: '28000',
        USERNAME: 'srv_env-fwkk6k',
        SETTINGS: {},
        SOCKET_TIMEOUT: 600,
        MODE: 'http',
        HS2_HTTP_SQLPATH: '',
        HS2_HTTP_ACCESS_TOKEN: undefined,
        IMPERSONATION: false,
        IMPERSONATION_TRUSTED: true,
        AUTH: 'ldap',
        SOCK: 'normal',
        KERBSERV: '',
        SSLCERT_NAME_MISMATCH: false,
        CACHE: { ENABLED: 0, RETENTION: 0 },
        ARCENGINE_HISTORY_ETL: {},
        USER_DEPENDENT_CACHE: false,
      },
      testIdMappings: {
        'impyla:cdw-impyla.basic.hostname':
          'coordinator.impala-1739180273-9kcj.svc.cluster.local',
        'impyla:cdw-impyla.basic.port': '28000',
      },
    });
  });

  it('check the checkOrUpdateConnection receive the correct data - Cloudera Data Warehouse Hive', async () => {
    await renderModalAndTestConnection({
      connectionId: 6,
      connectionType: 'hyve',
      connectionName: 'test_cdw_hive',
      connectionTypeName: 'Cloudera Data Warehouse Hive',
      // Commented-out key-value pairs from the 7.2.9 payload
      dataconnectionParams: {
        // ARCENGINE_HISTORY_ETL: {
        //   ENABLED: 0,
        //   TARGET_TABLE: '',
        //   JOB_NAME: '',
        //   SCHEDULE: 'Weekly',
        // },
        CONNECTIONNAME_EXTENDED: 'cdw-hyve',
        HOST: 'hiveserver2-service.compute-1736820433-r9kl.svc.cluster.local',
        PORT: '80',
        USERNAME: 'hive',
        MODE: 'http',
        HS2_HTTP_SQLPATH: 'cliservice',
        HS2_HTTP_ACCESS_TOKEN: '',
        SOCK: 'normal',
        AUTH: 'ldap',
        SOCKET_TIMEOUT: 600,
        IMPERSONATION: false,
        IMPERSONATION_TRUSTED: true,
        USER_DEPENDENT_CACHE: false,
        APP_NAME: 'viz',
        SETTINGS: { 'hive.server2.async.exec.async.compile': 'false' },
        CONCURRENCY: 100,
        CONCURRENCY_USER: 5,
        QUERY_TIMEOUT: 10,
        QUERY_LOADING_WARNING_SECONDS: 20,
        UPLOAD_MAXROWS: 1000000,
        CACHE: { ENABLED: 0, RETENTION: 86400 },
        ARCENGINE_HISTORY_ETL: {},
      },
      testIdMappings: {
        'hyve:cdw-hyve.basic.hostname':
          'hiveserver2-service.compute-1736820433-r9kl.svc.cluster.local',
        'hyve:cdw-hyve.basic.port': '80',
      },
    });
  });
});
