import React, { useMemo, useState } from 'react';
import {
  Col,
  Container,
  Form,
  FormControl,
  Modal,
  Popover,
  Row,
  Tab,
  Tabs,
} from 'react-bootstrap';
import Select from 'react-select';
import classNames from 'classnames';
import classes from './DataConnectionModal.module.scss';
import { getGlobalArc } from '../../utils/get-global-arc';
import { AxiosError } from 'axios';
import { Controller, FormProvider, useForm } from 'react-hook-form';
import { useMutation, useQuery } from '@tanstack/react-query';

import { DATACONN_CONFIG } from './constants/dataconnection-constants';
import { VizButton } from 'components/viz-button/VizButton';

import {
  dataconnectionQueryToken,
  getAllDataConnections,
} from 'apis/get-all-data-connections';
import {
  validateOrMaybecreateOrUpdateConnectionQueryToken,
  createOrUpdateConnection,
} from 'apis/create-or-update-data-connection';
import {
  ConnectionType,
  dataConnectionDefaultValues,
  dataConnectionSchema,
} from './utils/dataconnection-schema';

import { useNotificationStore } from 'stores/use-notification-store';

import {
  transformFormValuesToDataConnection,
  transformConnectionDataToFormValues,
} from './utils/dataconnections-transformers';
import { DataconnectionBasicTab } from './components/DataconnectionBasicTab';
import { DataconnectionAdvancedTab } from './components/DataconnectionAdvancedTab';
import { DataconnectionParametersTab } from './components/DataconnectionParametersTab';
import { DataconnectionDataTab } from './components/DataconnectionDataTab';
import { DataconnectionHistoryTab } from './components/DataconnectionHistoryTab';
import { DataconnectionStorageLocationsTab } from './components/DataconnectionStorageTab';
import { VizTooltip } from 'components/VizTooltip';
import { VizSpinner } from 'components/viz-spinner/VizSpinner';
import { Notification } from 'components/notification/Notification';
import { deleteDataconnection } from 'apis/delete-data-connection';
import { cloneDataconnection } from 'apis/clone-data-connection';
import { getCDWWarehouses, warehouseQueryToken } from 'apis/get-cdw-warehouses';
import ErrorIcon from '@cloudera-internal/cuix-core/icons/svg/ic-error.svg';
import { yupResolver } from '@hookform/resolvers/yup';
import { getDefaultValues } from './utils/dataconnection-utils';

const arc = getGlobalArc();

type DataConnectionModalProps = {
  isShow: boolean;
  dataconnectionId: number | null;
  associatedDatasetLength: number;
  allConnectionTypes: { [Key in ConnectionType]: string };
  modalType: 'new' | 'existing';
  onClose: () => void;
};

const CDV_MODE = arc.frontend.globals.settings.cdvMode;
const isCdwPrivate = arc.frontend.globals.cdwEnvironment === 'private';
const legacyImpalaEnabled = arc.frontend.globals.settings.isLegacyImpalaEnabled;
const legacyHiveEnabled = arc.frontend.globals.settings.isLegacyHiveEnabled;

export const DataConnectionModal = ({
  isShow,
  dataconnectionId,
  associatedDatasetLength,
  allConnectionTypes,
  modalType,
  onClose,
}: DataConnectionModalProps): JSX.Element => {
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));
  const [activeTab, setActiveTab] = useState('Basic');
  const [response, setResponse] = useState('');

  const { data: dataconnections, refetch } = useQuery({
    queryKey: [dataconnectionQueryToken],
    queryFn: () => getAllDataConnections(),
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Error fetching data connections.',
        },
        { delay: 3000 }
      );
    },
    initialData: [],
  });

  const existingConnectionData = useMemo(
    () =>
      isShow
        ? dataconnections.find(({ id }) => id === dataconnectionId)
        : undefined,
    [dataconnectionId, dataconnections, isShow]
  );

  const methods = useForm({
    mode: 'onTouched',
    values:
      modalType === 'existing' && existingConnectionData
        ? {
            ...dataConnectionDefaultValues,
            ...transformConnectionDataToFormValues(existingConnectionData),
          }
        : dataConnectionDefaultValues,
    resolver: yupResolver(dataConnectionSchema),
  });

  const connectionType = methods.watch('connectionType');
  const connectionName = methods.watch('connectionName');

  const selectedCdwOption = useMemo(() => {
    if (connectionType?.value === 'impyla:cdw-impyla') return 'impala';
    if (connectionType?.value === 'hyve:cdw-hyve') return 'hive';
    return '';
  }, [connectionType]);

  const { data: cdwWareHouses } = useQuery({
    queryKey: [warehouseQueryToken, selectedCdwOption],
    queryFn: () => getCDWWarehouses({ type: selectedCdwOption }),
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Error fetching data connections.',
        },
        { delay: 3000 }
      );
    },
    enabled: CDV_MODE === 'cdw' && !!selectedCdwOption,
    initialData: [],
  });

  const cdwHousesDataOptions = useMemo<
    { value: string; label: string }[]
  >(() => {
    return [
      {
        value: '',
        label:
          cdwWareHouses.length === 0
            ? 'No Virtual Warehouses'
            : 'Select a Virtual Warehouse URL',
      },
      ...cdwWareHouses.map(({ namespace }) => ({
        value: namespace,
        label: namespace,
      })),
    ];
  }, [cdwWareHouses]);

  // arcengine and arcengine_v2 fallback dropdown
  const fallbackDataconnections = useMemo<
    { value: string; label: string }[]
  >(() => {
    const fallbackTypes = ['impyla', 'hyve', 'arcengine', 'arcengine_v2'];
    return dataconnections
      .filter(({ type }) => fallbackTypes.includes(type))
      .map(({ id, name }) => ({
        value: id.toString(),
        label: name,
      }));
  }, [dataconnections]);

  const selectedConnectionConfig = useMemo(() => {
    return DATACONN_CONFIG[connectionType?.value];
  }, [connectionType?.value]);

  const options = useMemo(() => {
    const additionalOptions: {
      value: ConnectionType;
      label: string;
      disabled: boolean;
    }[] = [
      {
        value: 'impyla:cdw-impyla',
        label: 'Cloudera Data Warehouse Impala',
        disabled: false,
      },
      { value: 'impyla', label: 'Impala', disabled: false },
    ];

    if (legacyImpalaEnabled) {
      additionalOptions.push({
        value: 'impala',
        label: 'Impala (Legacy)',
        disabled: false,
      });
    }

    additionalOptions.push(
      {
        value: 'hyve:cdw-hyve',
        label: 'Cloudera Data Warehouse Hive',
        disabled: false,
      },
      { value: 'hyve', label: 'Hive', disabled: false }
    );

    if (legacyHiveEnabled) {
      additionalOptions.push({
        value: 'hive',
        label: 'Hive (Legacy)',
        disabled: false,
      });
    }

    const defaultOptions = Object.entries(allConnectionTypes)
      .map(([key, value]) => ({
        value: key as ConnectionType,
        label: value,
        disabled: CDV_MODE === 'cdw' && (key === 'duckdb' || key === 'sqlite'),
      }))
      .filter(
        (option) => !['impyla', 'hyve', 'impala', 'hive'].includes(option.value)
      )
      .concat([
        {
          value: 'postgres:ssb',
          label: 'SQL Stream Builder (Tech Preview)',
          disabled: false,
        },
      ])
      .sort((a, b) => a.label.localeCompare(b.label));

    return [...additionalOptions, ...defaultOptions];
  }, [allConnectionTypes]);

  const onSubmit = methods.handleSubmit((values) => {
    setResponse('');
    const dataconnectionData = transformFormValuesToDataConnection({
      dataconnectionId: dataconnectionId ?? undefined,
      connectionName: values.connectionName,
      connectionType: values.connectionType,
      values,
    });

    handleValidationOrMaybeCreation.mutate({
      shouldCreate: true,
      dataconnectionData,
      shouldValidate: 'true',
    });
  });

  const handleDeleteDataconnection = useMutation({
    mutationFn: () => {
      if (dataconnectionId !== null) {
        return deleteDataconnection({ id: dataconnectionId });
      }
      return Promise.reject(new Error('No valid dataconnectionId'));
    },
    onSuccess: () => {
      addNotification(
        {
          variant: 'success',
          message: `Data connection with ID ${dataconnectionId} has been deleted`,
        },
        { delay: 2000 }
      );
      arc.apps_data.fetch_and_draw(() => '');
      refetch();
      handleClose();
    },
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Failed to delete data connection',
        },
        { delay: 3000 }
      );
    },
  });

  const handleCloneDataconnection = useMutation({
    mutationFn: () => {
      if (dataconnectionId !== null) {
        return cloneDataconnection({ id: dataconnectionId });
      }
      return Promise.reject(new Error('No valid dataconnectionId'));
    },
    onSuccess: (clonedDataConnection) => {
      addNotification(
        {
          variant: 'success',
          message: `Successful cloned data connection!`,
        },
        { delay: 2000 }
      );

      const dataconnectionData = {
        id: clonedDataConnection.id ?? 0,
        name: clonedDataConnection.name,
      };
      arc.apps_data.set_active_dataconnection(dataconnectionData);

      arc.apps_data.fetch_and_draw(() => '');
      refetch();
      handleClose();
    },
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Failed to clone data connection',
        },
        { delay: 3000 }
      );
    },
  });

  const handleValidationOrMaybeCreation = useMutation({
    mutationKey: [validateOrMaybecreateOrUpdateConnectionQueryToken],
    mutationFn: ({
      shouldCreate,
      dataconnectionData,
      shouldValidate,
    }: {
      shouldCreate: boolean;
      dataconnectionData;
      shouldValidate: string;
    }) =>
      createOrUpdateConnection({
        dataconnectionData,
        shouldValidate,
      }),
    onSuccess: async (
      response,
      { dataconnectionData, shouldCreate, shouldValidate }
    ) => {
      setResponse(response?.status ?? 'Unknown Status');
      if (shouldCreate) {
        try {
          const returnedData = await createOrUpdateConnection({
            dataconnectionData,
            shouldValidate: 'false',
          });

          arc.apps_data.set_active_dataconnection({
            id: returnedData.id,
            name: returnedData.name,
          });

          arc.apps_data.fetch_and_draw(() => dataconnectionData.name);
          refetch();
          addNotification(
            {
              variant: 'success',
              message: `Successful created ${dataconnectionData.name} data connection!`,
            },
            { delay: 2000 }
          );

          handleClose();
        } catch (error) {
          addNotification(
            {
              variant: 'danger',
              message: `Failed to create data connection, ${
                (error as AxiosError)?.response?.data
              }`,
            },
            { delay: 3000 }
          );
        }
      }
    },
    onError: (
      error: AxiosError,
      { shouldCreate, dataconnectionData, shouldValidate }
    ) => {
      const errorResponse = error?.response?.data as string;
      const isAppNameOrParameterProblem = errorResponse.includes(
        'It is not in list of params that are allowed to be modified at runtime'
      );

      if (!isAppNameOrParameterProblem) {
        setResponse(errorResponse);
      } else {
        if (
          dataconnectionData.type !== 'hive' &&
          dataconnectionData.type !== 'hyve'
        )
          return;

        let parameters = dataconnectionData?.info?.PARAMS?.SETTINGS;
        const parameterKeys = Object.keys(parameters ?? {});
        const hasErrorWithParameters = parameterKeys.some((paramName) =>
          errorResponse.includes(paramName)
        );

        let newData = { ...dataconnectionData };
        let updatedItem = '';
        if (errorResponse.includes('applicationName')) {
          newData = {
            ...dataconnectionData,
            info: {
              ...dataconnectionData.info,
              PARAMS: {
                ...dataconnectionData.info.PARAMS,
                APP_NAME: '',
              },
            },
          };
          updatedItem = 'application name';
        } else if (hasErrorWithParameters) {
          newData = {
            ...dataconnectionData,
            info: {
              ...dataconnectionData.info,
              PARAMS: {
                ...dataconnectionData.info.PARAMS,
                SETTINGS: {},
              },
            },
          };
          updatedItem = 'parameters';
        }

        handleValidationOrMaybeCreation.mutate({
          shouldCreate,
          dataconnectionData: newData,
          shouldValidate,
        });

        setResponse(`Updating ${updatedItem} and retrying...`);
      }
    },
  });

  const collectDataforValidationOrMaybeCreation = async ({ shouldCreate }) => {
    setResponse('');
    if (!(await methods.trigger())) return;

    const formValues = await dataConnectionSchema.validate(methods.getValues());

    const dataconnectionData = transformFormValuesToDataConnection({
      dataconnectionId: dataconnectionId ?? undefined,
      connectionName,
      connectionType,
      values: formValues,
    });

    handleValidationOrMaybeCreation.mutate({
      shouldCreate,
      dataconnectionData,
      shouldValidate: 'true',
    });
  };

  const hasTabErrors = (tabName: string): boolean => {
    return (
      Object.keys(
        methods.formState.errors[connectionType?.value]?.[tabName] || {}
      ).length > 0
    );
  };

  const handleClose = () => {
    methods.reset(dataConnectionDefaultValues);
    setActiveTab('Basic');
    setResponse('');
    dataconnectionId = 0;
    onClose();
  };

  return (
    <Modal show={isShow} onHide={handleClose} data-test-id="connection-modal">
      <FormProvider {...methods}>
        <Form onSubmit={onSubmit}>
          <Modal.Header closeButton>
            <Modal.Title>
              {modalType === 'new'
                ? 'Create New Data Connection'
                : 'Edit Data Connection'}
            </Modal.Title>
          </Modal.Header>
          <Modal.Body>
            <Container>
              <Form.Group as={Row} className="mb-3" controlId="formConnType">
                <Form.Label column sm="3">
                  Connection type
                </Form.Label>
                <Col sm="9">
                  <Controller
                    name="connectionType"
                    control={methods.control}
                    render={({ field }) => (
                      <div data-test-id="connectionType">
                        <Select
                          {...field}
                          id={'connectionType'}
                          options={options}
                          isSearchable
                          placeholder="Select a connection type"
                          onChange={(option) => {
                            methods.reset({
                              ...dataConnectionDefaultValues,
                              connectionType: {
                                value: option?.value,
                                label: option?.label,
                                disabled: option?.disabled,
                              },
                            });
                            setActiveTab('Basic');
                          }}
                          getOptionLabel={(option) =>
                            option.disabled
                              ? `${option.label} (Not supported in Cloudera Data Warehouse)`
                              : option.label
                          }
                          isOptionDisabled={(option) => option.disabled}
                        />
                      </div>
                    )}
                  />
                </Col>
              </Form.Group>
              <Form.Group
                as={Row}
                className={classNames('mb-3', {
                  [classes.bottomDivider]: CDV_MODE !== 'cdw',
                })}
                controlId="formConnName"
              >
                <Form.Label column sm="3">
                  Connection name
                </Form.Label>
                <Col sm="9">
                  <FormControl
                    type="text"
                    data-test-id="connectionName"
                    {...methods.register('connectionName')}
                    isInvalid={!!methods.formState.errors.connectionName}
                  />
                  <Form.Control.Feedback
                    type="invalid"
                    className="text-right"
                    data-test-id="connectionNameError"
                  >
                    {methods.formState.errors.connectionName?.message}
                  </Form.Control.Feedback>
                </Col>
              </Form.Group>
              {CDV_MODE === 'cdw' &&
                (connectionType.value === 'impyla:cdw-impyla' ||
                  connectionType.value === 'hyve:cdw-hyve') && (
                  <Form.Group
                    as={Row}
                    className={classNames('mb-3', {
                      [classes.bottomDivider]: CDV_MODE === 'cdw',
                    })}
                    controlId="formCDWWarehouse"
                  >
                    <Form.Label column sm="3">
                      Cloudera Data Warehouse Virtual Warehouse{' '}
                      {cdwWareHouses.length === 0 && (
                        <VizTooltip
                          placement="top"
                          overlay={
                            <Popover id={`cdwWarehouse-tooltip`}>
                              <span className="d-flex p-3">
                                No available Cloudera Data Warehouse Virtual
                                Warehouse within this environment
                              </span>
                            </Popover>
                          }
                        >
                          <span
                            className={classNames(
                              classes.infoIcon,
                              'fa fa-info-circle'
                            )}
                          />
                        </VizTooltip>
                      )}
                    </Form.Label>
                    <Col sm="9">
                      <Controller
                        name="cdwWarehouse"
                        control={methods.control}
                        render={({ field }) => (
                          <Select
                            {...field}
                            id="cdwWarehouse"
                            options={cdwHousesDataOptions}
                            defaultValue={cdwHousesDataOptions[0]}
                            isDisabled={cdwHousesDataOptions.length === 1}
                            isSearchable
                            value={cdwHousesDataOptions.find(
                              (option) =>
                                'value' in field.value &&
                                option?.value === field.value.value
                            )}
                            onChange={(option) => {
                              const selectedWarehouse = cdwWareHouses.find(
                                (item) => item.namespace === option?.value
                              );

                              if (selectedWarehouse) {
                                const newDefaultValues = getDefaultValues({
                                  connectionType,
                                  connectionName,
                                  selectedWarehouse,
                                  isCdwPrivate,
                                });

                                methods.reset(newDefaultValues);
                              }

                              methods.setValue('cdwWarehouse', {
                                value: option?.value ?? '',
                                label: option?.label ?? '',
                              });
                            }}
                          />
                        )}
                      />
                    </Col>
                  </Form.Group>
                )}
              <Tabs
                activeKey={activeTab}
                onSelect={(tab) => setActiveTab(tab || 'Basic')}
                id="data-connection-tabs"
              >
                {selectedConnectionConfig?.basic && (
                  <Tab
                    key={`basic-${connectionType.value}`}
                    eventKey="Basic"
                    title={
                      <span className="d-flex align-items-center">
                        Basic{' '}
                        {hasTabErrors('basic') && (
                          <ErrorIcon
                            className={classNames('ml-1', classes.redError)}
                          />
                        )}
                      </span>
                    }
                  >
                    <DataconnectionBasicTab
                      basicFields={selectedConnectionConfig.basic}
                      connectionType={connectionType?.value}
                    />
                  </Tab>
                )}
                {selectedConnectionConfig?.advanced && (
                  <Tab
                    key={`advanced-${connectionType.value}`}
                    eventKey="Advanced"
                    title={
                      <span className="d-flex align-items-center">
                        Advanced{' '}
                        {hasTabErrors('advanced') && (
                          <ErrorIcon
                            className={classNames('ml-1', classes.redError)}
                          />
                        )}
                      </span>
                    }
                  >
                    <DataconnectionAdvancedTab
                      advancedFields={selectedConnectionConfig.advanced}
                      connectionType={connectionType?.value}
                      fallbackDataconnections={fallbackDataconnections}
                    />
                  </Tab>
                )}
                {!!selectedConnectionConfig?.parameters && (
                  <Tab
                    key={`parameters-${connectionType.value}`}
                    eventKey="Parameters"
                    title={
                      <span className="d-flex align-items-center">
                        Parameters{' '}
                        {hasTabErrors('parameters') && (
                          <ErrorIcon
                            className={classNames('ml-1', classes.redError)}
                          />
                        )}
                      </span>
                    }
                  >
                    <DataconnectionParametersTab
                      connectionType={connectionType?.value}
                    />
                  </Tab>
                )}
                {selectedConnectionConfig?.data && (
                  <Tab
                    eventKey="Data"
                    key={`data-${connectionType.value}`}
                    title={
                      <span className="d-flex align-items-center">
                        Data{' '}
                        {hasTabErrors('data') && (
                          <ErrorIcon
                            className={classNames('ml-1', classes.redError)}
                          />
                        )}
                      </span>
                    }
                  >
                    <DataconnectionDataTab
                      dataFields={selectedConnectionConfig.data}
                      connectionType={connectionType?.value}
                    />
                  </Tab>
                )}
                {selectedConnectionConfig?.history && (
                  <Tab eventKey="History" title="History">
                    <DataconnectionHistoryTab
                      historyFields={selectedConnectionConfig.history}
                      connectionType={connectionType?.value}
                    />
                  </Tab>
                )}
                {selectedConnectionConfig?.storageLocations && (
                  <Tab
                    key={`storage-${connectionType.value}`}
                    eventKey="Storage Location"
                    title={
                      <span className="d-flex align-items-center">
                        Storage Location{' '}
                        {hasTabErrors('storageLocations') && (
                          <ErrorIcon
                            className={classNames('ml-1', classes.redError)}
                          />
                        )}
                      </span>
                    }
                  >
                    <DataconnectionStorageLocationsTab
                      connectionType={connectionType?.value}
                    />
                  </Tab>
                )}
              </Tabs>
            </Container>
          </Modal.Body>
          <Modal.Footer>
            {handleValidationOrMaybeCreation.isLoading && (
              <div className="d-flex flex-row justify-content-center w-100 text-lg">
                <VizSpinner />
              </div>
            )}
            <div className="d-flex flex-column w-100">
              {response !== '' && (
                <Notification
                  message={
                    response === 'success'
                      ? 'Connection verified'
                      : `Error: ${response}`
                  }
                  dismissible
                  useHtml
                  variant={response === 'success' ? 'success' : 'danger'}
                  onClose={() => {
                    setResponse('');
                  }}
                />
              )}
            </div>
            <div className="d-flex justify-content-between w-100">
              <VizButton
                color="secondary"
                size="medium"
                onClick={() =>
                  collectDataforValidationOrMaybeCreation({
                    shouldCreate: false,
                  })
                }
                type="button"
                disabled={handleValidationOrMaybeCreation.isLoading}
                testId="dataconnection-validate"
              >
                Test
              </VizButton>
              <div className="d-flex gap-1">
                {modalType === 'existing' && (
                  <div className="d-flex gap-1">
                    <VizButton
                      color="secondary"
                      size="medium"
                      onClick={() => {
                        handleCloneDataconnection.mutate();
                      }}
                      type="button"
                    >
                      Clone
                    </VizButton>
                    <VizTooltip
                      placement="top"
                      overlay={
                        <Popover id="delete-connection-tooltip">
                          <span className="d-flex p-3">
                            You must delete all associated datasets before you
                            can delete this connection
                          </span>
                        </Popover>
                      }
                    >
                      <div>
                        <VizButton
                          color="secondary"
                          size="medium"
                          onClick={() => {
                            handleDeleteDataconnection.mutate();
                          }}
                          disabled={associatedDatasetLength > 0}
                          type="button"
                        >
                          Delete connection
                        </VizButton>
                      </div>
                    </VizTooltip>
                  </div>
                )}
                <VizButton
                  color="primary"
                  size="medium"
                  type="submit"
                  disabled={Object.keys(methods.formState.errors).length > 0}
                  testId="dataconnection-create-commit"
                >
                  {modalType === 'existing' ? 'Save' : 'Connect'}
                </VizButton>
              </div>
            </div>
          </Modal.Footer>
        </Form>
      </FormProvider>
    </Modal>
  );
};
