import React from 'react';
import { useFormContext } from 'react-hook-form';
import {
  ConnectionType,
  DataConnectionSchemaType,
} from '../utils/dataconnection-schema';
import {
  checkHasAdvancedTab,
  checkHasAuthenticationMode,
  checkHasConnectionMode,
  checkHasFallBack,
  checkHasSocketType,
} from '../utils/dataconnection-type-guards';
import { DataConnectionConfig } from '../constants/dataconnection-constants';
import { DataconnectionInput } from './DataconnectionInput';
import { DataconnectionCheckbox } from './DataconnectionCheckbox';
import { DataconnectionRadio } from './DataconnectionRadio';
import { DataconnectionDropdown } from './DataconnectionDropdown';

type DataconnectionAdvancedTabProps = {
  advancedFields: DataConnectionConfig[keyof DataConnectionConfig]['advanced'];
  connectionType: ConnectionType;
  fallbackDataconnections: {
    value: string;
    label: string;
  }[];
};

export const DataconnectionAdvancedTab = ({
  advancedFields,
  connectionType,
  fallbackDataconnections,
}: DataconnectionAdvancedTabProps): JSX.Element | null => {
  const { watch } = useFormContext<DataConnectionSchemaType>();

  if (!advancedFields) return null;

  let socketType: string | undefined;
  if (checkHasSocketType(connectionType)) {
    socketType = watch(`${connectionType}.advanced.socket_type`);
  }

  let connectionMode: string | undefined;
  if (checkHasConnectionMode(connectionType)) {
    connectionMode = watch(`${connectionType}.advanced.connection_mode`);
  }

  let authenticationMode: string | undefined;
  let impersonation: boolean | undefined;
  let trustedImpersonation: boolean | undefined;
  if (checkHasAuthenticationMode(connectionType)) {
    authenticationMode = watch(
      `${connectionType}.advanced.authentication_mode`
    );
    impersonation = watch(`${connectionType}.advanced.impersonation`);
    trustedImpersonation = watch(
      `${connectionType}.advanced.trusted_impersonation`
    );
  }

  return (
    <>
      {Object.keys(advancedFields).map((key) => {
        if (!checkHasAdvancedTab(connectionType)) return null;
        const fieldId = `${connectionType}.advanced.${
          key as keyof typeof advancedFields
        }` as const;

        const { config } = advancedFields[key];

        if (
          (key === 'kerberos_service_name' &&
            authenticationMode !== 'kerberos') ||
          (key === 'http_path' && connectionMode !== 'http') ||
          (key === 'hostname_mismatch' && socketType !== 'sslcert')
        ) {
          return null;
        }

        switch (config.type) {
          case 'input':
            return (
              <DataconnectionInput
                key={key}
                fieldId={fieldId}
                config={config}
              />
            );
          case 'checkbox':
            return (
              <DataconnectionCheckbox
                key={key}
                config={config}
                fieldId={fieldId}
                impersonation={impersonation}
                trustedImpersonation={trustedImpersonation}
              />
            );
          case 'radio':
            return (
              config.options && (
                <DataconnectionRadio
                  key={key}
                  config={config}
                  fieldId={fieldId}
                />
              )
            );
          case 'dropdown':
            return (
              checkHasFallBack(connectionType) &&
              fallbackDataconnections && (
                <DataconnectionDropdown
                  key={key}
                  config={config}
                  fieldId={fieldId}
                  options={fallbackDataconnections}
                  placeholder={'Select a fallback connection'}
                />
              )
            );
          default:
            return null;
        }
      })}
    </>
  );
};
