import React from 'react';
import classNames from 'classnames';
import classes from '../DataConnectionModal.module.scss';
import type { DataConnectionConfig } from '../constants/dataconnection-constants';
import { ConnectionType } from '../utils/dataconnection-schema';
import {
  checkHasSolrTypeAndJoinChar,
  checkHasAuthentication,
} from '../utils/dataconnection-type-guards';
import { DataconnectionInput } from './DataconnectionInput';
import { DataconnectionRadio } from './DataconnectionRadio';
import { DataconnectionTextarea } from './DataconnectionTextarea';
import { DataconnectionPassword } from './DataconnectionPassword';

type DataconnectionBasicTabProps = {
  basicFields: DataConnectionConfig[keyof DataConnectionConfig]['basic'];
  connectionType: ConnectionType;
};

const credentialFields = [
  'username',
  'password',
  'role',
  'access_key_id',
  'secret_access_key',
];

export const DataconnectionBasicTab = ({
  basicFields,
  connectionType,
}: DataconnectionBasicTabProps): JSX.Element | null => {
  const credentialFieldKeys = Object.keys(basicFields).filter((key) =>
    credentialFields.includes(key)
  );
  const nonCredentialFieldKeys = Object.keys(basicFields).filter(
    (key) => !credentialFields.includes(key)
  );

  return (
    <>
      {nonCredentialFieldKeys.map((key) => {
        const fieldId = `${connectionType}.basic.${
          key as keyof typeof basicFields
        }` as const;
        const { config } = basicFields[key];

        switch (config.type) {
          case 'input':
            return (
              <DataconnectionInput
                key={key}
                config={config}
                fieldId={fieldId}
              />
            );
          case 'radio':
            if (
              config.options &&
              (checkHasSolrTypeAndJoinChar(connectionType) ||
                checkHasAuthentication(connectionType))
            ) {
              return (
                <DataconnectionRadio
                  key={key}
                  config={config}
                  fieldId={fieldId}
                />
              );
            }
            return null;
          case 'textarea':
            return (
              <DataconnectionTextarea
                key={key}
                config={config}
                fieldId={fieldId}
              />
            );
          default:
            return null;
        }
      })}
      {credentialFieldKeys.length > 0 && (
        <>
          <h5 className={classNames('mb-3', classes.topDivider)}>
            Credentials
          </h5>
          {credentialFieldKeys.map((key) => {
            const fieldId = `${connectionType}.basic.${
              key as keyof typeof basicFields
            }` as const;
            const { config } = basicFields[key];

            switch (config.type) {
              case 'input':
                return (
                  <DataconnectionInput
                    key={`credential-${key}`}
                    config={config}
                    fieldId={fieldId}
                  />
                );
              case 'password':
                return (
                  <DataconnectionPassword
                    key={`credential-${key}`}
                    config={config}
                    fieldId={fieldId}
                  />
                );
              default:
                return null;
            }
          })}
        </>
      )}
    </>
  );
};
