import * as yup from 'yup';
import { getGlobalArc } from 'utils/get-global-arc';
import { BaseDataConnectionType } from 'types/data-connection.type';
const arc = getGlobalArc();

const defaultConcurrency =
  arc.frontend.globals.settings.dataConnectionConcurrencyDefault;
const defaultConcurrencyPerUser =
  arc.frontend.globals.settings.dataConnectionConcurrencyUserDefault;
const defaultQueryTimeout =
  arc.frontend.globals.settings.dataConnectionQueryTimeoutDefault;
const defaultRowUploadLimit =
  arc.frontend.globals.settings.uploadMaxrowsDefault;

export type ConnectionType =
  | BaseDataConnectionType
  | 'hyve:cdw-hyve'
  | 'impyla:cdw-impyla'
  | 'postgres:ssb';

const nonConnectionSpecificValues = {
  connectionType: {
    value: 'impyla:cdw-impyla' as const,
    label: 'Cloudera Data Warehouse Impala',
    disabled: false,
  },
  connectionName: '',
  cdwWarehouse: {
    value: '',
    label: '',
  },
};

const dataConcurrencySchema = yup
  .number()
  .typeError('Concurrency is required and must be a number');
const dataConcurrencyPerPersonSchema = yup
  .number()
  .typeError('Concurrency Per Person is required and must be a number');
const dataQueryTimeoutSchema = yup
  .number()
  .typeError('Query Timeout is required and must be a number');
const dataQueryLoadingWarningSchema = yup
  .number()
  .typeError('Query Loading Warning is required and must be a number');
const dataRowUploadLimitSchema = yup
  .number()
  .typeError('Row Upload Limit is required and must be a number');
const dataResultCacheSchema = yup.boolean();
const dataCacheRetentionTimeoutSchema = yup.number();
const sharedDataSchema = yup.object({
  concurrency: dataConcurrencySchema,
  concurrency_per_person: dataConcurrencyPerPersonSchema,
  query_timeout: dataQueryTimeoutSchema,
  query_loading_warning: dataQueryLoadingWarningSchema,
  row_upload_limit: dataRowUploadLimitSchema,
  result_cache: dataResultCacheSchema,
  cache_retention_timeout: dataCacheRetentionTimeoutSchema,
});

export const dataConnectionSchema = yup.object({
  connectionType: yup
    .object({
      value: yup
        .string()
        .oneOf([
          'arcengine',
          'arcengine_v2',
          'aster',
          'athena',
          'bigquery',
          'cassandra',
          'denodo',
          'drill',
          'druid',
          'duckdb',
          'elasticsearch',
          'hyve',
          'hive',
          'impyla',
          'impala',
          'ksql',
          'mariadb',
          'mssqlserver',
          'mysql',
          'oracle',
          'postgres',
          'redshift',
          'snowflake',
          'solr',
          'sparksql',
          'sqlite',
          'teradata',
          'vertica',
          'hyve:cdw-hyve',
          'impyla:cdw-impyla',
          'postgres:ssb',
          'trino',
          'phoenix',
        ])
        .required(),
      label: yup.string().required(),
      disabled: yup.boolean().required(),
    })
    .required(),
  connectionName: yup.string().required('Connection name is required'),
  cdwWarehouse: yup.object({
    value: yup.string().when('connectionType', {
      is: (connectionType?: { value: ConnectionType }) =>
        connectionType?.value === 'impyla:cdw-impyla' ||
        connectionType?.value === 'hyve:cdw-hyve',
      then: (schema) => schema.required(),
    }),
    label: yup.string().when('connectionType', {
      is: (connectionType?: { value: ConnectionType }) =>
        connectionType?.value === 'impyla:cdw-impyla' ||
        connectionType?.value === 'hyve:cdw-hyve',
      then: (schema) => schema.required(),
    }),
  }),
  'impyla:cdw-impyla': yup.object({
    basic: yup
      .object({
        account_name: yup.string(),
        file_name: yup.string(),
        authentication: yup.string(),
        dsn: yup.string(),
        hostname: yup.string().when('connectionType', {
          is: (connectionType?: { value: ConnectionType }) =>
            connectionType?.value === 'impyla:cdw-impyla',
          then: (schema) => schema.required('Hostname is required'),
        }),
        port: yup.string().when('connectionType', {
          is: (connectionType?: { value: ConnectionType }) =>
            connectionType?.value === 'impyla:cdw-impyla',
          then: (schema) => schema.required('Port is required'),
        }),
        username: yup.string(),
        password: yup.string(),
      })
      .required(),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      socket_timeout: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      application_name: yup.string(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  impyla: yup.object({
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'impyla',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'impyla',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      socket_timeout: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      application_name: yup.string(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  impala: yup.object({
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'impyla',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'impyla',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      socket_timeout: yup.string(),
      application_name: yup.string(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  'hyve:cdw-hyve': yup.object({
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'hyve:cdw-hyve',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'hyve:cdw-hyve',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      socket_timeout: yup.string(),
      application_name: yup.string(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  hyve: yup.object({
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'hyve',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'hyve',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      socket_timeout: yup.string(),
      application_name: yup.string(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  hive: yup.object({
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'hive',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'hive',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      socket_timeout: yup.string(),
      application_name: yup.string(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  athena: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string(),
      port: yup.string(),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      region_name: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'athena',
        then: (schema) => schema.required('Region Name is required'),
      }),
      s3_staging_directory: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'athena',
        then: (schema) => schema.required('S3 Staging Directory is required'),
      }),
      access_key_id: yup.string(),
      secret_access_key: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  aster: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'aster',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'aster',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  cassandra: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'cassandra',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'cassandra',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  arcengine: yup.object({
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'arcengine',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'arcengine',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      query_timeout: yup.number(),
      session_timeout: yup.string(),
      socket_timeout: yup.string(),
      queue_depth: yup.string(),
      fallback: yup.string(),
    }),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
    history: yup.object({
      history_etl: yup.boolean(),
      target_table: yup.string(),
      job_name: yup.string(),
      schedule: yup.string(),
    }),
    storageLocations: yup
      .array()
      .of(
        yup.object({
          name: yup.string().min(1, 'Location name cannot be empty').required(),
          value: yup
            .string()
            .min(1, 'Location string cannot be empty')
            .required(),
        })
      )
      .required(),
  }),
  arcengine_v2: yup.object({
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'arcengine_v2',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'arcengine_v2',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      socket_timeout: yup.string(),
      fallback: yup.string(),
    }),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
    history: yup.object({
      history_etl: yup.boolean(),
      target_table: yup.string(),
      job_name: yup.string(),
      schedule: yup.string(),
    }),
    storageLocations: yup
      .array()
      .of(
        yup.object({
          name: yup.string().min(1, 'Location name cannot be empty').required(),
          value: yup
            .string()
            .min(1, 'Location string cannot be empty')
            .required(),
        })
      )
      .required(),
  }),
  ksql: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'ksql',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'ksql',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  denodo: yup.object({
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'denodo',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'denodo',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  duckdb: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string(),
      port: yup.string(),
      account_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      file_name: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'duckdb',
        then: (schema) => schema.required('File name is required'),
      }),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  druid: yup.object({
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'druid',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'druid',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  drill: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'drill',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'drill',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  elasticsearch: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'elasticsearch',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'elasticsearch',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  bigquery: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string(),
      port: yup.string(),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      service_account_key: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'bigquery',
        then: (schema) => schema.required('Service Account Key is required'),
      }),
      billing_project: yup.string(),
      data_project: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  mariadb: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'mariadb',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'mariadb',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  mssqlserver: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'mssqlserver',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'mssqlserver',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  mysql: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'mysql',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'mysql',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  oracle: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'oracle',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'oracle',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  postgres: yup.object({
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'postgres',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'postgres',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  snowflake: yup.object({
    basic: yup.object({
      hostname: yup.string(),
      port: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      account_name: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'snowflake',
        then: (schema) => schema.required('Account Name is required'),
      }),
      warehouse: yup.string(),
      database: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
      role: yup.string().nullable(),
    }),
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  solr: yup.object({
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'solr',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'solr',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      solrType: yup.string(),
      array_join_char: yup.string().required('Join character is required'),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      socket_timeout: yup.string(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  sparksql: yup.object({
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'sparksql',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'sparksql',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      socket_timeout: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().of(
      yup.object({
        name: yup.string().min(1, 'Parameter name cannot be empty').required(),
        value: yup
          .string()
          .min(1, 'Parameter value cannot be empty')
          .required(),
      })
    ),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  sqlite: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      file_name: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'sqlite',
        then: (schema) => schema.required('File name is required'),
      }),
      authentication: yup.string(),
      account_name: yup.string(),
      dsn: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  'postgres:ssb': yup.object({
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'postgres:ssb',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'postgres:ssb',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string().nullable(),
      password: yup.string().nullable(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      //
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  redshift: yup.object({
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'redshift',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'redshift',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      access_token: yup.string(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  teradata: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      dsn: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'teradata',
        then: (schema) => schema.required('DSN is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  vertica: yup.object({
    advanced: yup.object().optional(),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    basic: yup.object({
      hostname: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'vertica',
        then: (schema) => schema.required('Hostname is required'),
      }),
      port: yup.string().when('connectionType', {
        is: (connectionType?: { value: ConnectionType }) =>
          connectionType?.value === 'vertica',
        then: (schema) => schema.required('Port is required'),
      }),
      account_name: yup.string(),
      file_name: yup.string(),
      authentication: yup.string(),
      dsn: yup.string(),
      database: yup.string(),
      username: yup.string(),
      password: yup.string(),
    }),
    data: sharedDataSchema.concat(
      yup.object({
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
      })
    ),
  }),
  trino: yup.object({
    basic: yup
      .object({
        account_name: yup.string(),
        file_name: yup.string(),
        authentication: yup.string(),
        dsn: yup.string(),
        hostname: yup.string().when('connectionType', {
          is: (connectionType?: { value: ConnectionType }) =>
            connectionType?.value === 'trino',
          then: (schema) => schema.required('Hostname is required'),
        }),
        port: yup.string().when('connectionType', {
          is: (connectionType?: { value: ConnectionType }) =>
            connectionType?.value === 'trino',
          then: (schema) => schema.required('Port is required'),
        }),
        username: yup.string(),
        password: yup.string(),
      })
      .required(),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      socket_timeout: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      application_name: yup.string(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
  phoenix: yup.object({
    basic: yup
      .object({
        account_name: yup.string(),
        file_name: yup.string(),
        authentication: yup.string(),
        dsn: yup.string(),
        hostname: yup.string().when('connectionType', {
          is: (connectionType?: { value: ConnectionType }) =>
            connectionType?.value === 'phoenix',
          then: (schema) => schema.required('Hostname is required'),
        }),
        port: yup.string().when('connectionType', {
          is: (connectionType?: { value: ConnectionType }) =>
            connectionType?.value === 'phoenix',
          then: (schema) => schema.required('Port is required'),
        }),
        username: yup.string(),
        password: yup.string(),
      })
      .required(),
    advanced: yup.object({
      connection_mode: yup.string(),
      http_path: yup.string(),
      socket_type: yup.string(),
      hostname_mismatch: yup.boolean(),
      authentication_mode: yup.string(),
      kerberos_service_name: yup.string(),
      socket_timeout: yup.string(),
      impersonation: yup.boolean(),
      trusted_impersonation: yup.boolean(),
      user_dependent_cache: yup.boolean(),
      application_name: yup.string(),
    }),
    history: yup.object().optional(),
    storageLocations: yup.array().optional(),
    parameters: yup.array().optional(),
    data: sharedDataSchema.concat(
      yup.object({
        concurrency: dataConcurrencySchema.required(),
        concurrency_per_person: dataConcurrencyPerPersonSchema.required(),
        query_timeout: dataQueryTimeoutSchema.required(),
        query_loading_warning: dataQueryLoadingWarningSchema.required(),
        row_upload_limit: dataRowUploadLimitSchema.required(),
      })
    ),
  }),
});

export type DataConnectionSchemaType = yup.InferType<
  typeof dataConnectionSchema
>;

export const dataConnectionDefaultValues: DataConnectionSchemaType = {
  ...nonConnectionSpecificValues,
  'impyla:cdw-impyla': {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '443',
      username: '',
      password: '',
    },
    advanced: {
      connection_mode: 'http',
      http_path: '',
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'ldap',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      application_name: 'viz',
    },
    parameters: [] as { name: string; value: string }[],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  impyla: {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '21050',
      username: '',
      password: '',
    },
    advanced: {
      connection_mode: 'binary',
      http_path: '',
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'ldap',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      application_name: 'viz',
    },
    parameters: [] as { name: string; value: string }[],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  impala: {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '21050',
      username: '',
      password: '',
    },
    advanced: {
      // connection_mode: 'binary',
      // http_path: '',
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'ldap',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      application_name: 'viz',
    },
    parameters: [] as { name: string; value: string }[],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  'hyve:cdw-hyve': {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '443',
      username: '',
      password: '',
    },
    advanced: {
      connection_mode: 'http',
      http_path: '',
      access_token: '',
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'ldap',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      application_name: 'viz',
    },
    parameters: [
      {
        name: 'hive.server2.async.exec.async.compile',
        value: 'false',
      },
    ],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  hyve: {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '10000',
      username: '',
      password: '',
    },
    advanced: {
      connection_mode: 'http',
      http_path: '',
      access_token: '',
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'plain',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      application_name: 'viz',
    },
    parameters: [
      {
        name: 'hive.server2.async.exec.async.compile',
        value: 'false',
      },
    ],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  hive: {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '10000',
      username: '',
      password: '',
    },
    advanced: {
      connection_mode: 'http',
      http_path: '',
      access_token: '',
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'plain',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      application_name: 'viz',
    },
    parameters: [
      {
        name: 'hive.server2.async.exec.async.compile',
        value: 'false',
      },
    ],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  athena: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      region_name: '',
      s3_staging_directory: '',
      access_key_id: '',
      secret_access_key: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: true,
      cache_retention_timeout: 1800,
    },
  },
  cassandra: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '9042',
      database: '',
      username: '',
      password: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  arcengine: {
    basic: {
      hostname: '',
      port: '31050',
      username: '',
      password: '',
    },
    advanced: {
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'nosasl',
      kerberos_service_name: '',
      query_timeout: 60,
      session_timeout: '60',
      socket_timeout: '60',
      queue_depth: '',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      fallback: '',
    },
    parameters: [] as { name: string; value: string }[],
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: true,
      cache_retention_timeout: 300,
    },
    history: {
      history_etl: false,
      target_table: '',
      job_name: '',
      schedule: 'weekly',
    },
    storageLocations: [] as { name: string; value: string }[],
  },
  arcengine_v2: {
    basic: {
      hostname: '',
      port: '31050',
      username: '',
      password: '',
    },
    advanced: {
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'nosasl',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      fallback: '',
    },
    parameters: [] as { name: string; value: string }[],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: true,
      cache_retention_timeout: 300,
    },
    history: {
      history_etl: false,
      target_table: '',
      job_name: '',
      schedule: 'weekly',
    },
    storageLocations: [] as { name: string; value: string }[],
  },
  aster: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '2406',
      database: '',
      dsn: '',
      authentication: 'No Authentication',
      username: '',
      password: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  ksql: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '9098',
      username: '',
      password: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  denodo: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '9996',
      database: '',
      username: '',
      password: '',
    },
    advanced: {
      socket_type: 'ssl',
      hostname_mismatch: false,
    },
    data: {
      row_upload_limit: undefined,
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  duckdb: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      file_name: '',
    },
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  druid: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '0',
      database: '',
      username: '',
      password: '',
    },
    advanced: {
      socket_type: 'ssl',
      hostname_mismatch: false,
    },
    data: {
      row_upload_limit: defaultRowUploadLimit,
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  drill: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '31010',
      database: '',
      dsn: '',
      authentication: 'No Authentication',
      username: '',
      password: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  elasticsearch: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '9200',
      username: '',
      password: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: true,
      cache_retention_timeout: 86400,
    },
  },
  bigquery: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      service_account_key: '',
      billing_project: '',
      data_project: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  mariadb: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '3306',
      database: '',
      username: '',
      password: '',
    },
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  mssqlserver: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '1433',
      database: '',
      username: '',
      password: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  mysql: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '3306',
      database: '',
      username: '',
      password: '',
    },
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  oracle: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '1521',
      database: '',
      username: '',
      password: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  postgres: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '5432',
      database: '',
      username: '',
      password: '',
    },
    advanced: {
      socket_type: 'ssl',
      hostname_mismatch: false,
    },
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  snowflake: {
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      account_name: '',
      warehouse: '',
      database: '',
      username: '',
      password: '',
      role: '',
    },
    parameters: [] as { name: string; value: string }[],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  solr: {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '443',
      solrType: 'cloud',
      array_join_char: ',',
      username: '',
      password: '',
    },
    advanced: {
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'nosasl',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
    },
    parameters: [] as { name: string; value: string }[],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  sparksql: {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '10000',
      username: '',
      password: '',
    },
    advanced: {
      connection_mode: 'http',
      http_path: '',
      access_token: '',
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'ssl',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
    },
    parameters: [] as { name: string; value: string }[],
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  'postgres:ssb': {
    parameters: undefined,
    storageLocations: undefined,
    basic: {
      hostname: '',
      port: '5432',
      database: '',
      username: '',
      password: '',
    },
    advanced: {
      socket_type: 'ssl',
      hostname_mismatch: false,
    },
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  sqlite: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      file_name: '',
    },
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  redshift: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '5439',
      database: '',
      username: '',
      password: '',
    },
    advanced: {
      socket_type: 'ssl',
      hostname_mismatch: false,
    },
    data: {
      row_upload_limit: undefined,
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  teradata: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      dsn: '',
      username: '',
      password: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  vertica: {
    parameters: undefined,
    storageLocations: undefined,
    history: undefined,
    advanced: undefined,
    basic: {
      hostname: '',
      port: '5433',
      database: '',
      username: '',
      password: '',
    },
    data: {
      concurrency_per_person: undefined,
      query_timeout: undefined,
      row_upload_limit: undefined,
      concurrency: undefined,
      query_loading_warning: 20,
      result_cache: true,
      cache_retention_timeout: 1800,
    },
  },
  trino: {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '8080',
      username: '',
      password: '',
    },
    advanced: {
      connection_mode: 'binary',
      http_path: undefined,
      socket_type: 'normal',
      hostname_mismatch: false,
      authentication_mode: 'nosasl',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      application_name: undefined,
    },
    parameters: undefined,
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
  phoenix: {
    storageLocations: undefined,
    history: undefined,
    basic: {
      hostname: '',
      port: '8765',
      username: '',
      password: '',
    },
    advanced: {
      connection_mode: 'http',
      http_path: '',
      socket_type: 'ssl',
      hostname_mismatch: false,
      authentication_mode: 'plain',
      kerberos_service_name: '',
      socket_timeout: '60',
      impersonation: false,
      trusted_impersonation: false,
      user_dependent_cache: false,
      application_name: 'viz',
    },
    parameters: undefined,
    data: {
      concurrency: defaultConcurrency,
      concurrency_per_person: defaultConcurrencyPerUser,
      query_timeout: defaultQueryTimeout,
      query_loading_warning: 20,
      row_upload_limit: defaultRowUploadLimit,
      result_cache: false,
      cache_retention_timeout: 86400,
    },
  },
};
