import { DataConnection, DataConnectionType } from 'types/data-connection.type';
import {
  ConnectionType,
  DataConnectionSchemaType,
  dataConnectionDefaultValues,
} from './dataconnection-schema';
import {
  checkHasPassword,
  checkHasHostnameAndPort,
  checkHasWarehouse,
  checkHasSolrTypeAndJoinChar,
  checkHasDatabase,
  checkHasFileName,
  checkHasConnectionMode,
  checkHasHTTPPath,
  checkHasAccessToken,
  checkHasSocketType,
  checkHasSessionTimeOut,
  checkHasAuthenticationMode,
  checkHasSocketTimeout,
  checkHasAppName,
  checkHasFallBack,
  checkHasParameters,
  checkHasConcurrencyAndConcurrencyUser,
  checkHasQueryTimeout,
  checkHasRowUploadLimit,
  checkHasAuthentication,
  checkHasDSN,
  checkHasArcengineHistoryETL,
  checkHasStorageLoctions,
} from './dataconnection-type-guards';

export const transformConnectionDataToFormValues = (
  connectionData: DataConnection
) => {
  const parametersAndStorageLocationsConverter = (
    obj: Record<string, string>
  ) =>
    Object.entries(obj).map(([key, value]) => ({
      name: key,
      value: value,
    }));

  if (connectionData.type === 'odbc' || connectionData.type === 'mongodb')
    throw new Error(`NOT A VALID TYPE: ${connectionData.type}`);

  let connectionType: ConnectionType = connectionData.type;
  let connectionLabel = connectionData.typeName;

  if (connectionData.info?.PARAMS?.CONNECTIONNAME_EXTENDED === 'cdw-impyla') {
    connectionType = 'impyla:cdw-impyla';
    connectionLabel = 'Cloudera Data Warehouse Impala';
  }

  if (connectionData.info?.PARAMS?.CONNECTIONNAME_EXTENDED === 'cdw-hyve') {
    connectionType = 'hyve:cdw-hyve';
    connectionLabel = 'Cloudera Data Warehouse Hive';
  }

  const connectionTypeOption = {
    value: connectionType,
    label: connectionLabel,
    disabled: false,
  };

  const connectionSpecificValues = {
    basic: {
      hostname: connectionData?.info?.PARAMS?.HOST,
      port: connectionData?.info?.PARAMS?.PORT,
      warehouse: connectionData?.info?.PARAMS?.WAREHOUSE,
      solrType: connectionData?.info?.PARAMS?.SOLRTYPE,
      array_join_char: connectionData?.info?.PARAMS?.SOLR_JOINCHAR,
      database: connectionData?.info?.PARAMS?.DATABASE,
      username: connectionData?.info?.PARAMS?.USERNAME,
      password: connectionData?.isPasswordSet ? '********' : '',
      role: connectionData?.info?.PARAMS?.ROLE,
      account_name: connectionData?.info?.PARAMS?.ACCOUNT,
      file_name: connectionData?.info?.PARAMS?.FILE,
      authentication: connectionData?.info?.PARAMS?.AUTHENTICATION,
      dsn: connectionData?.info?.PARAMS?.DSN,
    },
    advanced: {
      connection_mode: connectionData?.info?.PARAMS?.MODE,
      http_path: connectionData?.info?.PARAMS?.HS2_HTTP_SQLPATH,
      access_token: connectionData?.info?.PARAMS?.HS2_HTTP_ACCESS_TOKEN,
      socket_type: connectionData?.info?.PARAMS?.SOCK,
      query_timeout: connectionData?.info?.PARAMS?.QUERY_TIMEOUT,
      session_timeout: connectionData?.info?.PARAMS?.SESSION_TIMEOUT,
      queue_depth: connectionData?.info?.PARAMS?.QUEUE_DEPTH,
      hostname_mismatch: connectionData?.info?.PARAMS?.SSLCERT_NAME_MISMATCH,
      authentication_mode: connectionData?.info?.PARAMS?.AUTH,
      kerberos_service_name: connectionData?.info?.PARAMS?.KERBSERV,
      socket_timeout: connectionData?.info?.PARAMS?.SOCKET_TIMEOUT,
      impersonation: connectionData?.info?.PARAMS?.IMPERSONATION,
      trusted_impersonation:
        connectionData?.info?.PARAMS?.IMPERSONATION_TRUSTED,
      user_dependent_cache: connectionData?.info?.PARAMS?.USER_DEPENDENT_CACHE,
      application_name: connectionData?.info?.PARAMS?.APP_NAME,
      fallback: connectionData?.info?.PARAMS?.FALLBACK,
    },
    parameters: connectionData.info.PARAMS?.SETTINGS
      ? parametersAndStorageLocationsConverter(
          connectionData?.info?.PARAMS?.SETTINGS
        )
      : undefined,
    data: {
      concurrency:
        connectionData?.info?.PARAMS?.CONCURRENCY !== undefined
          ? +connectionData?.info?.PARAMS?.CONCURRENCY
          : dataConnectionDefaultValues[connectionData.type].data.concurrency,
      concurrency_per_person:
        connectionData?.info?.PARAMS?.CONCURRENCY_USER !== undefined
          ? +connectionData?.info?.PARAMS?.CONCURRENCY_USER
          : dataConnectionDefaultValues[connectionData.type].data
              .concurrency_per_person,
      query_timeout:
        connectionData?.info?.PARAMS?.QUERY_TIMEOUT !== undefined
          ? +connectionData?.info?.PARAMS?.QUERY_TIMEOUT
          : dataConnectionDefaultValues[connectionData.type].data.query_timeout,
      query_loading_warning:
        connectionData?.info?.PARAMS?.QUERY_LOADING_WARNING_SECONDS !==
        undefined
          ? +connectionData?.info?.PARAMS?.QUERY_LOADING_WARNING_SECONDS
          : dataConnectionDefaultValues[connectionData.type].data
              .query_loading_warning,
      row_upload_limit:
        connectionData?.info?.PARAMS?.UPLOAD_MAXROWS !== undefined
          ? +connectionData?.info?.PARAMS?.UPLOAD_MAXROWS
          : dataConnectionDefaultValues[connectionData.type].data
              .row_upload_limit,
      result_cache: connectionData.info.PARAMS?.CACHE?.ENABLED === 1 || false,
      cache_retention_timeout:
        connectionData.info.PARAMS?.CACHE?.RETENTION || 0,
    },
    history: {
      history_etl:
        connectionData.info.PARAMS?.ARCENGINE_HISTORY_ETL?.ENABLED === 1 ||
        false,
      target_table:
        connectionData.info.PARAMS?.ARCENGINE_HISTORY_ETL?.TARGET_TABLE,
      job_name: connectionData.info.PARAMS?.ARCENGINE_HISTORY_ETL?.JOB_NAME,
      schedule: connectionData.info.PARAMS?.ARCENGINE_HISTORY_ETL?.SCHEDULE,
    },
    storage: connectionData.info.PARAMS?.LOCATIONS
      ? parametersAndStorageLocationsConverter(
          connectionData?.info?.PARAMS?.LOCATIONS
        )
      : undefined,
  };

  return {
    connectionId: connectionData.id,
    connectionType: connectionTypeOption,
    connectionName: connectionData.name,

    [connectionType]: connectionSpecificValues,
  };
};

export const transformFormValuesToDataConnection = <
  Type extends ConnectionType
>(dataConnection: {
  dataconnectionId?: number;
  connectionName: string;
  connectionType: { label: string; value: Type };
  values: DataConnectionSchemaType;
}) => {
  const parametersAndStorageLocationsReverseConverter = (
    array: { name: string; value: string }[]
  ) =>
    array.reduce((acc, { name, value }) => {
      acc[name] = value;
      return acc;
    }, {});

  let password: string | undefined = undefined;
  let USERNAME: string | undefined = undefined;
  if (checkHasPassword(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    password =
      connectionSpecificValues?.basic?.password !== '********'
        ? connectionSpecificValues?.basic?.password || undefined
        : undefined;
    USERNAME = connectionSpecificValues?.basic?.username ?? undefined;
  }

  let HOST: string | undefined = undefined;
  let PORT: string | undefined = undefined;
  if (checkHasHostnameAndPort(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    HOST = connectionSpecificValues?.basic?.hostname;
    PORT = connectionSpecificValues?.basic?.port;
  }

  let WAREHOUSE: string | undefined = undefined;
  let ROLE: string | undefined = undefined;
  let ACCOUNT: string | undefined = undefined;
  if (checkHasWarehouse(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    WAREHOUSE = connectionSpecificValues?.basic?.warehouse;
    ROLE = connectionSpecificValues?.basic?.role ?? undefined;
    ACCOUNT = connectionSpecificValues?.basic?.account_name;
  }

  let SOLRTYPE: string | undefined = undefined;
  let SOLR_JOINCHAR: string | undefined = undefined;

  if (checkHasSolrTypeAndJoinChar(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    SOLRTYPE = connectionSpecificValues?.basic?.solrType;
    SOLR_JOINCHAR = connectionSpecificValues?.basic?.array_join_char;
  }

  let DATABASE: string | undefined = undefined;
  if (checkHasDatabase(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    DATABASE = connectionSpecificValues?.basic?.database;
  }

  let FILE: string | undefined = undefined;
  if (checkHasFileName(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    FILE = connectionSpecificValues?.basic?.file_name;
  }

  let MODE: string | undefined = undefined;
  if (checkHasConnectionMode(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    MODE = connectionSpecificValues?.advanced?.connection_mode;
  }

  let HS2_HTTP_SQLPATH: string | undefined = undefined;
  if (checkHasHTTPPath(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    HS2_HTTP_SQLPATH = connectionSpecificValues?.advanced?.http_path;
  }

  let HS2_HTTP_ACCESS_TOKEN: string | undefined = undefined;
  if (checkHasAccessToken(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    HS2_HTTP_ACCESS_TOKEN = connectionSpecificValues?.advanced?.access_token;
  }

  let SOCK: string | undefined = undefined;
  let SSLCERT_NAME_MISMATCH: boolean | undefined = undefined;
  if (checkHasSocketType(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    SOCK = connectionSpecificValues?.advanced?.socket_type;
    SSLCERT_NAME_MISMATCH =
      connectionSpecificValues?.advanced?.hostname_mismatch;
  }

  let SESSION_TIMEOUT: string | undefined = undefined;
  let QUEUE_DEPTH: string | undefined = undefined;
  if (checkHasSessionTimeOut(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    SESSION_TIMEOUT = connectionSpecificValues?.advanced?.session_timeout;
    QUEUE_DEPTH = connectionSpecificValues?.advanced?.queue_depth;
  }

  let AUTH: string | undefined = undefined;
  let KERBSERV: string | undefined = undefined;
  let IMPERSONATION: boolean | undefined = undefined;
  let IMPERSONATION_TRUSTED: boolean | undefined = undefined;
  let USER_DEPENDENT_CACHE: boolean | undefined = undefined;
  if (checkHasAuthenticationMode(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    AUTH = connectionSpecificValues?.advanced?.authentication_mode;
    KERBSERV = connectionSpecificValues?.advanced?.kerberos_service_name;
    IMPERSONATION = connectionSpecificValues?.advanced?.impersonation;
    IMPERSONATION_TRUSTED =
      connectionSpecificValues?.advanced?.trusted_impersonation;
    USER_DEPENDENT_CACHE =
      connectionSpecificValues?.advanced?.user_dependent_cache;
  }

  let SOCKET_TIMEOUT: number | undefined = undefined;
  if (checkHasSocketTimeout(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    SOCKET_TIMEOUT =
      connectionSpecificValues?.advanced?.socket_timeout !== undefined
        ? +connectionSpecificValues?.advanced?.socket_timeout
        : undefined;
  }

  let APP_NAME: string | undefined = undefined;
  if (checkHasAppName(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    APP_NAME = connectionSpecificValues?.advanced?.application_name;
  }

  let FALLBACK: string | undefined = undefined;
  if (checkHasFallBack(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    FALLBACK = connectionSpecificValues?.advanced?.fallback;
  }

  let SETTINGS: Record<string, string> | undefined = undefined;
  if (checkHasParameters(dataConnection.connectionType.value)) {
    const parameters =
      dataConnection.values[dataConnection.connectionType.value].parameters;
    SETTINGS = parameters
      ? parametersAndStorageLocationsReverseConverter(parameters)
      : undefined;
  }

  let CONCURRENCY: number | undefined = undefined;
  let CONCURRENCY_USER: number | undefined = undefined;
  if (
    checkHasConcurrencyAndConcurrencyUser(dataConnection.connectionType.value)
  ) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    CONCURRENCY = connectionSpecificValues?.data?.concurrency;
    CONCURRENCY_USER = connectionSpecificValues?.data?.concurrency_per_person;
  }

  let QUERY_TIMEOUT: number | undefined = undefined;
  if (checkHasQueryTimeout(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    QUERY_TIMEOUT = connectionSpecificValues?.data?.query_timeout;
  }

  let UPLOAD_MAXROWS: number | undefined = undefined;
  if (checkHasRowUploadLimit(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    UPLOAD_MAXROWS = connectionSpecificValues?.data?.row_upload_limit;
  }

  let AUTHENTICATION: string | undefined = undefined;
  if (checkHasAuthentication(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    AUTHENTICATION = connectionSpecificValues?.basic?.authentication;
  }

  let DSN: string | undefined = undefined;
  if (checkHasDSN(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    DSN = connectionSpecificValues?.basic?.dsn;
  }

  let ENABLED: number | undefined = undefined;
  let TARGET_TABLE: string | undefined = undefined;
  let JOB_NAME: string | undefined = undefined;
  let SCHEDULE: string | undefined = undefined;

  if (checkHasArcengineHistoryETL(dataConnection.connectionType.value)) {
    const connectionSpecificValues =
      dataConnection.values[dataConnection.connectionType.value];
    ENABLED = connectionSpecificValues?.history?.history_etl ? 1 : 0;
    TARGET_TABLE = connectionSpecificValues?.history?.target_table;
    JOB_NAME = connectionSpecificValues?.history?.job_name;
    SCHEDULE = connectionSpecificValues?.history?.schedule;
  }

  let LOCATIONS: Record<string, string> | undefined = undefined;
  if (checkHasStorageLoctions(dataConnection.connectionType.value)) {
    const storageLocations =
      dataConnection.values[dataConnection.connectionType.value]
        ?.storageLocations;
    LOCATIONS = storageLocations
      ? parametersAndStorageLocationsReverseConverter(storageLocations)
      : undefined;
  }

  const QUERY_LOADING_WARNING_SECONDS =
    dataConnection.values[dataConnection.connectionType.value].data
      ?.query_loading_warning;

  const connectionData =
    dataConnection.values[dataConnection.connectionType.value];

  const CACHE = {
    ENABLED: connectionData.data?.result_cache ? 1 : 0,
    RETENTION: connectionData.data?.cache_retention_timeout || 0,
  };

  let connectionType: DataConnectionType;
  let CONNECTIONNAME_EXTENDED = '';

  if (dataConnection.connectionType.value === 'impyla:cdw-impyla') {
    connectionType = 'impyla';
    CONNECTIONNAME_EXTENDED = 'cdw-impyla';
  } else if (dataConnection.connectionType.value === 'hyve:cdw-hyve') {
    connectionType = 'hyve';
    CONNECTIONNAME_EXTENDED = 'cdw-hyve';
  } else if (dataConnection.connectionType.value === 'postgres:ssb') {
    connectionType = 'postgres';
  } else {
    connectionType = dataConnection.connectionType.value;
  }

  return {
    id: dataConnection.dataconnectionId,
    type: connectionType,
    typeName: dataConnection.connectionType.label,
    name: dataConnection.connectionName,
    password,

    info: {
      PARAMS: {
        CONNECTIONNAME_EXTENDED,
        HOST,
        PORT,
        WAREHOUSE,
        SOLRTYPE,
        SOLR_JOINCHAR,
        DATABASE,
        USERNAME,
        ROLE,
        ACCOUNT,
        FILE,
        AUTHENTICATION,
        DSN,

        MODE,
        HS2_HTTP_SQLPATH,
        HS2_HTTP_ACCESS_TOKEN,
        SOCK,
        SESSION_TIMEOUT,
        QUEUE_DEPTH,
        SSLCERT_NAME_MISMATCH,
        AUTH,
        KERBSERV,
        SOCKET_TIMEOUT,
        IMPERSONATION,
        IMPERSONATION_TRUSTED,
        USER_DEPENDENT_CACHE,
        APP_NAME,
        FALLBACK,

        SETTINGS,

        CONCURRENCY,
        CONCURRENCY_USER,
        QUERY_TIMEOUT,
        QUERY_LOADING_WARNING_SECONDS,
        UPLOAD_MAXROWS,
        CACHE,

        ARCENGINE_HISTORY_ETL: {
          ENABLED,
          TARGET_TABLE,
          JOB_NAME,
          SCHEDULE,
        },

        LOCATIONS,
      },
    },
  };
};
