import styles from './Card.module.scss';
import React, { useEffect } from 'react';
import { Column, Dataset } from 'arc/types/dataset.type';
import { useQuery } from '@tanstack/react-query';
import { Spinner } from 'components/spinner/Spinner';
import { getGlobalArc } from 'utils/get-global-arc';
import { VizTooltip } from 'components/VizTooltip';
import { Tooltip } from 'react-bootstrap';
import { Notification } from 'components/notification/Notification';
import RefreshIcon from '@cloudera-internal/cuix-core/icons/react/RefreshIcon';
import { getTextColumnStatistics } from 'apis/get-profile-data';
import { ArcVisualContainer } from 'components/visual-container/VisualContainer';
import { FieldDetail } from 'components/field-detail/FieldDetail';
import { removeNonIdCompatibleChars } from 'utils/remove-non-id-compatible-chars';
import { DatasetColumnPill } from 'components/dataset-column-pill/DatasetColumnPill';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { DimensionStatistics } from 'types/profile-statistic.types';
import classNames from 'classnames';

const arc = getGlobalArc();

export const DimensionCard = ({
  column,
  dataset,
  id,
}: {
  column: Column;
  dataset: Dataset;
  id: string;
}): JSX.Element | null => {
  const columnName = column.display_name();
  const connectionId = dataset.dataconnection_id();
  const chartDivId = `dimensionChartDiv-${removeNonIdCompatibleChars(id)}`;

  const {
    data: result,
    isLoading,
    error,
    refetch,
  } = useQuery<DimensionStatistics, Error>({
    staleTime: 5 * 3600,
    queryKey: [columnName, connectionId, 'dimension'],
    queryFn: () => column && getTextColumnStatistics(column, dataset),
  });

  useEffect(() => {
    arc.dataprofileHelper.plotTopFrequentValueVisual(
      dataset,
      column,
      chartDivId
    );
  }, [result, dataset, column, chartDivId]);

  return (
    <div className={`${styles.card} ${styles.cardString}`}>
      <div id={id} className={styles.header}>
        <DatasetColumnPill
          label={column.build_dsobject().alias}
          variant="dimension"
        />
        <div className={styles.type}>
          {/* In the Figma design a string representation of the type was made however Sherry suggested the icon and seems nice */}
          <VizTooltip
            placement="top"
            overlay={
              <Tooltip id={`tooltip-type-${id}`}>
                Type: {column.build_dsobject().type}
              </Tooltip>
            }
          >
            <div>
              <span
                className={classNames('dim-icon-type', column.coltype_class())}
              ></span>
            </div>
          </VizTooltip>
        </div>
      </div>
      {isLoading && <Spinner />}
      <div className={styles.detailsContainer}>
        {!!error && !isLoading && (
          <>
            <Notification
              title="Error"
              useHtml
              message={error.message}
              variant="danger"
              show={!!error}
            />
            <div className={styles.retry}>
              <ClouderaButton variant="link" onClick={() => refetch()}>
                <RefreshIcon fontSize="larger" className="mr-1" />
                Try Again
              </ClouderaButton>
            </div>
          </>
        )}
        {result && (
          <>
            <div className="d-flex justify-content-center">
              <Notification
                message={'No data is available'}
                variant="info"
                show={!result.recordCount}
              />
            </div>
            {!!result.recordCount && (
              <>
                <FieldDetail
                  label="Most frequent"
                  value={result.mostFrequentValue}
                  styleClass={{ valueClass: 'w-60' }}
                />
                <FieldDetail
                  label="Least frequent"
                  value={result.leastFrequentValue}
                  styleClass={{ valueClass: 'w-50' }}
                />
                <FieldDetail
                  label="Number of Empty string"
                  value={result.numberOfEmpty}
                  percentage={(result.numberOfEmpty / result.recordCount) * 100}
                />
                <FieldDetail
                  label="Empty cells"
                  value={result.numberOfNulls}
                  percentage={(result.numberOfNulls / result.recordCount) * 100}
                />
                <FieldDetail label="Record count" value={result.recordCount} />
                <FieldDetail
                  label="Duplicated rows"
                  value={result.duplicatedRows}
                  percentage={
                    (result.duplicatedRows / result.recordCount) * 100
                  }
                />
                <FieldDetail
                  label="Distinct count"
                  value={result.distinctCount}
                  percentage={(result.distinctCount / result.recordCount) * 100}
                />
                <FieldDetail label="Longest word" value={result.longestWord} />
                <FieldDetail
                  label="Shortest word"
                  value={result.shortestWord}
                />
                <FieldDetail
                  label="Min"
                  value={result.min}
                  styleClass={{
                    labelClass: styles.uppercase,
                    valueClass: 'w-80',
                  }}
                />
                <FieldDetail
                  label="Max"
                  value={result.max}
                  styleClass={{
                    labelClass: styles.uppercase,
                    valueClass: 'w-80',
                  }}
                />
              </>
            )}
          </>
        )}
      </div>

      <div className="dropdown-divider" role="separator"></div>
      {!error && !isLoading && <ArcVisualContainer id={chartDivId} />}
    </div>
  );
};
