import styles from './Card.module.scss';
import React, { useEffect, useState } from 'react';
import { Column, Dataset } from 'arc/types/dataset.type';
import { useQuery } from '@tanstack/react-query';
import { Spinner } from 'components/spinner/Spinner';
import { getGlobalArc } from 'utils/get-global-arc';
import { VizTooltip } from 'components/VizTooltip';
import { Tooltip } from 'react-bootstrap';
import { Notification } from 'components/notification/Notification';
import RefreshIcon from '@cloudera-internal/cuix-core/icons/react/RefreshIcon';
import { getMeasureColumnStatistics } from 'apis/get-profile-data';
import { ArcVisualContainer } from 'components/visual-container/VisualContainer';
import { FieldDetail } from 'components/field-detail/FieldDetail';
import { removeNonIdCompatibleChars } from 'utils/remove-non-id-compatible-chars';
import { DatasetColumnPill } from 'components/dataset-column-pill/DatasetColumnPill';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { MeasureStatistics } from 'types/profile-statistic.types';
import classNames from 'classnames';

const arc = getGlobalArc();
export const MeasureCard = ({
  column,
  dataset,
  id,
}: {
  column: Column;
  dataset: Dataset;
  id: string;
}): JSX.Element | null => {
  const columnName = column.display_name();
  const connectionId = dataset.dataconnection_id();
  const chartDivId = `measureChartDiv-${removeNonIdCompatibleChars(id)}`;
  const [histogramRequestError, setHistogramError] = useState<{
    msg: string;
    detail: string;
  } | null>(null);

  const {
    data: result,
    isLoading,
    error,
    refetch,
  } = useQuery<MeasureStatistics, Error>({
    staleTime: 5 * 3600,
    queryKey: [columnName, connectionId, 'measureStats'],
    queryFn: () => getMeasureColumnStatistics(column, dataset),
  });

  useEffect(() => {
    if (result) {
      arc.dataprofileHelper.plotProfileHistogram(
        dataset,
        column,
        chartDivId,
        [result.min, result.max],
        (msg, detail) => {
          setHistogramError({ msg, detail });
        }
      );
    }
  }, [result, dataset, column, chartDivId]);
  return (
    <div className={`${styles.card}`} id={id}>
      <div id={id} className={styles.header}>
        <DatasetColumnPill
          label={column.build_dsobject().alias}
          variant="measure"
        />
        <div className={styles.type}>
          <VizTooltip
            placement="top"
            overlay={
              <Tooltip id={`tooltip-type-${id}`}>
                Type: {column.build_dsobject().type}
              </Tooltip>
            }
          >
            <div>
              <span
                className={classNames('dim-icon-type', column.coltype_class())}
              ></span>
            </div>
          </VizTooltip>
        </div>
      </div>
      {isLoading && <Spinner />}
      <div className={styles.detailsContainer}>
        {!!error && !isLoading && (
          <>
            <Notification
              title="Error"
              useHtml
              message={error.message}
              variant="danger"
              show={!!error}
            />
            <div className={styles.retry}>
              <ClouderaButton variant="link" onClick={() => refetch()}>
                <RefreshIcon fontSize="larger" className="mr-1" />
                Try Again
              </ClouderaButton>
            </div>
          </>
        )}
        {!isLoading && result && !error && (
          <>
            <div className="d-flex justify-content-center">
              <Notification
                message={'No data is available'}
                variant="info"
                show={!result.recordCount}
              />
            </div>
            {!!result.recordCount && (
              <>
                <FieldDetail
                  label="Min"
                  value={result.min}
                  styleClass={{ labelClass: styles.uppercase }}
                />
                <FieldDetail
                  label="Max"
                  value={result.max}
                  styleClass={{ labelClass: styles.uppercase }}
                />
                <FieldDetail
                  label="Avg"
                  value={result.avg?.toFixed(
                    Number.isInteger(result.avg) ? 0 : 2
                  )}
                />
                <FieldDetail
                  label="Standard Dev"
                  value={Math.sqrt(result.variance)?.toFixed(2)}
                />
                <FieldDetail
                  label="Number of Zeroes"
                  value={result.numberOfZeroes}
                  percentage={
                    (result.numberOfZeroes / result.recordCount) * 100
                  }
                />
                <FieldDetail
                  label="Empty cells"
                  value={result.numberOfNulls}
                  percentage={(result.numberOfNulls / result.recordCount) * 100}
                />
                <FieldDetail label="Record count" value={result.recordCount} />
                <FieldDetail
                  label="Duplicated rows"
                  value={result.duplicatedRows}
                  percentage={
                    (result.duplicatedRows / result.recordCount) * 100
                  }
                />
                <FieldDetail
                  label="Distinct count"
                  value={result.distinctCount}
                  percentage={(result.distinctCount / result.recordCount) * 100}
                />
                <FieldDetail
                  label="Most frequent"
                  value={result.mostFrequentValue}
                />
                <FieldDetail
                  label="Least frequent"
                  value={result.leastFrequentValue}
                />
              </>
            )}
          </>
        )}
      </div>

      <div className="dropdown-divider" role="separator"></div>
      <Notification
        title="Error"
        message={histogramRequestError?.msg ?? ''}
        expandedMessage={histogramRequestError?.detail ?? ''}
        variant="danger"
        show={!!histogramRequestError}
      />
      {!error && !histogramRequestError && !isLoading && (
        <ArcVisualContainer id={chartDivId} />
      )}
    </div>
  );
};
