import React, { useEffect, useMemo, useState } from 'react';
import { Column, Dataset } from 'arc/types/dataset.type';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import styles from './DatasetProfileTab.module.scss';
import { ClouderaGroupedListFilter } from 'components/cloudera-filter/ClouderaGroupedListFilter';
import { isEqual } from 'lodash-es';
import { DimensionCard } from 'components/data-profile-cards/DimensionCard';
import { MeasureCard } from 'components/data-profile-cards/MeasureCard';
import { DataProfileSelection } from 'types/data-profile-selection.types';
import { getGlobalArc } from 'utils/get-global-arc';
import { ColumnLabel } from 'components/column-label/ColumnLabel';
import { DatasetColumnPill } from 'components/dataset-column-pill/DatasetColumnPill';
import { directAccessTypes } from 'arc/misc/misc';

const arc = getGlobalArc();
export const getCardId = (column: Column) =>
  `${column.table().tablename()} ${column.display_name()}`;
export const scrollIntoCard = (column: Column) => {
  const element = document.getElementById(getCardId(column));
  element?.scrollIntoView({ behavior: 'smooth' });
};
export const removeColumn = (
  dataProfileSelection: DataProfileSelection,
  column: Column
) => {
  const newValue = {
    ...dataProfileSelection,
    dimensions: dataProfileSelection.dimensions.filter(
      (elem) => elem !== column
    ),
    measures: dataProfileSelection.measures.filter((elem) => elem !== column),
  } as DataProfileSelection;

  return newValue;
};

export const getDimensionColumns = (dataset: Dataset) =>
  dataset
    .tables()
    .flatMap((tables) => tables.dimensionsAndHierarchies())
    .filter((c) => !c.isHierarchy() && !c.isComplexType());

export const getMeasureColumns = (dataset: Dataset) =>
  dataset
    .tables()
    .flatMap((table) => table.measuresOnly())
    .filter((c) => !c.isHierarchy() && !c.isComplexType());

export const convertColumnsToGroupedList = (columns: Column[]) => {
  return (
    columns.map((d) => {
      return {
        value: d,
        label: ColumnLabel(d.display_name(), d.coltype_class()),
        stringLabel: d.display_name(),
        group: arc.datasets.drop_dbname(d.table().tablename()),
        id: arc.datasets.drop_dbname(d.table().tablename()) + d.display_name(),
      };
    }) ?? []
  );
};

const emptySelection = {
  dimensions: [],
  measures: [],
};
export const DatasetProfileTab = ({
  dataset,
  title,
}: {
  dataset?: Dataset;
  title: string;
}): JSX.Element | null => {
  const [draftDataProfileSelection, setDraftDataProfileSelection] =
    useState<DataProfileSelection>();
  const [confirmedDataProfileSelection, setConfirmedDataProfileSelection] =
    useState<DataProfileSelection>();

  useEffect(() => {
    if (dataset) {
      setConfirmedDataProfileSelection({ ...emptySelection, dataset });
      setDraftDataProfileSelection({ ...emptySelection, dataset });
    }
  }, [dataset]);

  const dimensionList = useMemo(
    () => (dataset ? getDimensionColumns(dataset) : []),
    [dataset]
  );
  const measuresList = useMemo(
    () => (dataset ? getMeasureColumns(dataset) : []),
    [dataset]
  );

  const removePillValue = (column: Column) => {
    // Remove Pill should take effect both on draft and already drawn values

    setDraftDataProfileSelection((prev) => prev && removeColumn(prev, column));
    setConfirmedDataProfileSelection(
      (prev) => prev && removeColumn(prev, column)
    );
  };

  const isAnySelected =
    draftDataProfileSelection?.dimensions.length ||
    draftDataProfileSelection?.measures.length;

  const noNewItem = isEqual(
    draftDataProfileSelection,
    confirmedDataProfileSelection
  );
  if (!dataset) {
    return null;
  }
  return (
    <>
      <div className="mr-3 ml-3">
        <div>
          <div className="d-flex flex-row align-items-center justify-content-between">
            <h3 className="data-section-header">{title}</h3>
          </div>
          <div
            className={
              'd-flex flex-row align-content-center align-items-center justify-content-between'
            }
          >
            {!directAccessTypes.includes(dataset.dataconnection_type()) && (
              <div className="text-center m-5 w-100">
                This feature is not enabled on this connection type
              </div>
            )}
            {directAccessTypes.includes(dataset.dataconnection_type()) && (
              <>
                <div className="d-flex flex-row mt-3 mb-3 align-items-center">
                  <div className={`mr-3 ${styles.title}`}>
                    Select fields to Profile
                  </div>
                  <div className={'mr-1'}>
                    <ClouderaGroupedListFilter
                      value={draftDataProfileSelection?.dimensions ?? []}
                      options={convertColumnsToGroupedList(dimensionList)}
                      onChange={(selectedDimensions) => {
                        setDraftDataProfileSelection((prev) => {
                          return (
                            prev && {
                              ...prev,
                              dimensions: selectedDimensions,
                            }
                          );
                        });
                      }}
                      label={'Dimension'}
                    />
                  </div>
                  <div className={'mr-1'}>
                    <ClouderaGroupedListFilter
                      value={draftDataProfileSelection?.measures ?? []}
                      options={convertColumnsToGroupedList(measuresList)}
                      onChange={(selectedMeasures) => {
                        setDraftDataProfileSelection((prev) => {
                          return (
                            prev && {
                              ...prev,
                              measures: selectedMeasures,
                            }
                          );
                        });
                      }}
                      label={'Measures'}
                    />
                  </div>
                </div>
                <div>
                  <ClouderaButton
                    color="primary"
                    onClick={() => {
                      setConfirmedDataProfileSelection(
                        draftDataProfileSelection
                      );
                    }}
                    disabled={!isAnySelected || noNewItem}
                  >
                    <div className="d-flex">
                      <span className={'explore-icon mr-1 bg-white'}></span>
                      Explore
                    </div>
                  </ClouderaButton>
                </div>
              </>
            )}
          </div>

          <div className="d-flex mt-3 mb-3 flex-wrap">
            {draftDataProfileSelection?.dimensions.map((dimension) => (
              <DatasetColumnPill
                key={`pill-${getCardId(dimension)}`}
                label={dimension.build_dsobject().alias}
                variant="dimension"
                onClick={() => {
                  scrollIntoCard(dimension);
                }}
                onDelete={() => removePillValue(dimension)}
              />
            ))}
            {draftDataProfileSelection?.measures.map((measure) => (
              <DatasetColumnPill
                key={`pill-${getCardId(measure)}`}
                label={measure.build_dsobject().alias}
                variant="measure"
                onClick={() => {
                  scrollIntoCard(measure);
                }}
                onDelete={() => removePillValue(measure)}
              />
            ))}
          </div>
        </div>
        <div className="mt-3 mb-3">
          <div className={'d-flex flex-wrap justify-content-start'}>
            {confirmedDataProfileSelection?.dimensions.map((item) => (
              <DimensionCard
                id={getCardId(item)}
                key={getCardId(item)}
                column={item}
                dataset={confirmedDataProfileSelection.dataset}
              />
            ))}
            {confirmedDataProfileSelection?.measures.map((item) => (
              <MeasureCard
                id={getCardId(item)}
                key={getCardId(item)}
                column={item}
                dataset={confirmedDataProfileSelection.dataset}
              />
            ))}
          </div>
        </div>
      </div>
    </>
  );
};

/*

SELECT order_date_year, COUNT(*) AS count
FROM main.superstore_sales
GROUP BY order_date_year


order_date_year 	count 
2009	            2153
2010	            2142
2011	            2002
2012	            2102


SELECT order_date_year, COUNT(*) AS count, max(order_date_year) as max, min(order_date_year) AS min, avg(order_date_year) AS avg
FROM main.superstore_sales
GROUP BY order_date_year;


order_date_year 	count 	max 	min 	avg 
2009	            2153	2009	2009	2009
2010            	2142	2010	2010	2010
2011	            2002	2011	2011	2011
2012	            2102	2012	2012	2012


SELECT max(order_date_year) as max, min(order_date_year) AS min, avg(order_date_year) AS avg, sum(case when order_date_year IS NULL then 1 else 0 end) as nulls, count(distinct order_date_year) as distinct_0
FROM main.superstore_sales
max 	min 	avg 	nulls 	distinct_0 
2012	2009	2010.4825574473152	0	4
*/
