import '@testing-library/jest-dom';
import { fireEvent, render, screen } from '@testing-library/react';

import { DatasetActionsDropdown } from './DatasetActionsDropdown';
import * as React from 'react';

jest.mock('stores/use-notification-store.ts', () => ({
  useNotificationStore: () => ({
    addNotification: jest.fn(),
  }),
}));

jest.mock('components/dataset-delete-modal/DatasetDeleteModal', () => ({
  DatasetDeleteModal: () => <div>Delete Dataset</div>,
}));

afterEach(() => {
  jest.resetAllMocks();
  jest.restoreAllMocks();
});

describe('DatasetActionsDropdown', () => {
  const datasetPropsAdminUser = {
    id: 123,
    name: 'Test Dataset 1',
    hasEditPerm: true,
  };
  const datasetPropsNonAdminUser = {
    id: 456,
    name: 'Test Dataset 2',
    hasEditPerm: false,
  };

  it('renders the dropdown with Actions button', async () => {
    render(<DatasetActionsDropdown {...datasetPropsAdminUser} />);

    const actionsDropdown = await screen.findByRole('button', {
      name: 'Actions',
    });
    expect(actionsDropdown).toBeInTheDocument();
    expect(actionsDropdown).toHaveTextContent('Actions');
  });

  it('displays the delete modal when delete button is clicked', async () => {
    render(<DatasetActionsDropdown {...datasetPropsAdminUser} />);

    const actionsDropdown = await screen.findByRole('button', {
      name: 'Actions',
    });
    fireEvent.click(actionsDropdown);

    const deleteButton = await screen.findByRole('button', {
      name: 'Delete',
    });
    fireEvent.click(deleteButton);

    expect(screen.getByText('Delete Dataset')).toBeInTheDocument();
  });

  it('disables the dropdown when user does not have edit permissions', async () => {
    render(<DatasetActionsDropdown {...datasetPropsNonAdminUser} />);

    const actionsDropdown = await screen.findByRole('button', {
      name: 'Actions',
    });
    expect(actionsDropdown).toBeDisabled();
  });

  it('displays a tooltip when user does not have edit permissions and hovers over the dropdown', async () => {
    render(<DatasetActionsDropdown {...datasetPropsNonAdminUser} />);

    const actionsDropdown = await screen.findByRole('button', {
      name: 'Actions',
    });

    fireEvent.mouseOver(actionsDropdown);

    const tooltip = await screen.findByText(
      'You do not have permission to edit this dataset'
    );

    expect(tooltip).toBeInTheDocument();
    expect(actionsDropdown).toBeDisabled();
  });
});
