import React, { useState, useCallback } from 'react';

import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import CloneIcon from '@cloudera-internal/cuix-core/icons/svg/ic-copy-clipboard.svg';
import ArrowDown from '@cloudera-internal/cuix-core/icons/svg/ic-dropdown.svg';

import { getGlobalArc } from 'utils/get-global-arc';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { Dropdown, Popover } from 'react-bootstrap';
import { VizTooltip } from 'components/VizTooltip';
import { DatasetDeleteModal } from 'components/dataset-delete-modal/DatasetDeleteModal';
import { useNotificationStore } from 'stores/use-notification-store';
import { useLegacyBridge } from 'services/legacy-bridge';
import classes from './DatasetActionsDropdown.module.scss';

const arc = getGlobalArc();

export type DatasetActionsDropdownProps = {
  id: number;
  name: string;
  hasEditPerm: boolean;
};

export const DatasetActionsDropdown = ({
  id,
  name,
  hasEditPerm,
}: DatasetActionsDropdownProps) => {
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));
  const [isDeleteModalOpen, setIsDeleteModalOpen] = useState(false);

  const handleCloneSuccess = useCallback(
    (newDatasetId: string) => {
      addNotification({
        variant: 'success',
        message: 'Cloned dataset successfully. Loading...',
      });
      setTimeout(() => {
        window.location.assign(`${newDatasetId}`);
      }, 1000);
    },
    [addNotification]
  );

  const handleCloneDataset = useCallback(() => {
    if (hasEditPerm) {
      arc.datasets
        .Dataset()
        .dataset_id(id)
        .cloneDatasetOnServer()
        .then(handleCloneSuccess)
        .catch((error) =>
          addNotification({ variant: 'danger', message: error.message })
        );
    }
  }, [id, hasEditPerm, addNotification, handleCloneSuccess]);

  useLegacyBridge(
    'dataset-delete-modal',
    useCallback(
      ({
        payload,
      }: {
        payload?: { id: number; name: string; hasEditPerm: boolean };
      }) => {
        if (!payload) return;
        setIsDeleteModalOpen(true);
      },
      []
    )
  );

  return (
    <>
      <VizTooltip
        overlay={
          <Popover id="dataset-actions-tooltip">
            {!hasEditPerm && (
              <Popover.Content>
                You do not have permission to edit this dataset
              </Popover.Content>
            )}
          </Popover>
        }
      >
        <Dropdown>
          <div className={classes.actions} data-test-id="actions-dropdown">
            <Dropdown.Toggle
              as={ClouderaButton}
              disabled={!hasEditPerm}
              icon={<ArrowDown />}
              iconPosition="right"
            >
              Actions
            </Dropdown.Toggle>
          </div>

          <Dropdown.Menu>
            <Dropdown.Item
              className={classes.dropdownItem}
              onClick={handleCloneDataset}
            >
              <CloneIcon className="mr-2" />
              Clone Dataset
            </Dropdown.Item>
            <Dropdown.Item
              className={classes.dropdownItem}
              onClick={() => setIsDeleteModalOpen(true)}
              disabled={!hasEditPerm}
            >
              <DeleteIcon className="mr-2" />
              Delete
            </Dropdown.Item>
          </Dropdown.Menu>
        </Dropdown>
      </VizTooltip>
      {hasEditPerm && (
        <DatasetDeleteModal
          dataset={{ id, name, hasEditPerm }}
          isShow={isDeleteModalOpen}
          onClose={() => setIsDeleteModalOpen(false)}
          onSuccess={() =>
            setTimeout(() => {
              window.location.assign(arc.frontend.globals.urls.dataUrl);
            }, 1000)
          }
        />
      )}
    </>
  );
};
