import React, { useMemo } from 'react';
import classes from '../dataset-delete-modal/DatasetDeleteModal.module.scss';
import { Badge, Tooltip } from 'react-bootstrap';
import { VizTooltip } from 'components/VizTooltip';
import type { PossibleDeletedReport } from 'apis/get-possible-deleted-reports';
import { formatDate, formatTimeDifference } from 'utils/date-formatter';
import { getGlobalArc } from 'utils/get-global-arc';
import {
  createColumnHelper,
  getCoreRowModel,
  getPaginationRowModel,
  getSortedRowModel,
} from '@tanstack/react-table';
import { ClouderaTable } from 'components/cloudera-table';
import { getBlankTargetUrl } from 'arc/misc/misc';
import { useTable } from 'hooks/use-table';

const arc = getGlobalArc();

type DatasetDeleteModalDashboardsTableProp = {
  possibleDeletedReports: PossibleDeletedReport[];
  isPossibleDeletedReportsLoading: boolean;
};

const columnHelper = createColumnHelper<PossibleDeletedReport>();

export const DatasetDeleteModalDashboardsTable = ({
  possibleDeletedReports,
  isPossibleDeletedReportsLoading,
}: DatasetDeleteModalDashboardsTableProp) => {
  const columns = useMemo(
    () => [
      columnHelper.accessor('img', {
        header: () => '',
        cell: (info) => {
          const reportImg: string = info.getValue();
          const reportId: number = info.row.getValue('id');
          return reportImg ? (
            <div className={classes.thumbnailWrapper}>
              <a
                href={getBlankTargetUrl(
                  `${arc.frontend.globals.urls.viewDashboardUrl}/${reportId}`
                )}
                target="_blank"
                rel="noreferrer"
              >
                <img
                  src={info.getValue()}
                  alt="report thumbnail"
                  className={classes.reportThumbnailImg}
                />
              </a>
              <div className={classes.dashboardObjectType}></div>
            </div>
          ) : (
            <div>
              <VizTooltip
                placement="top"
                trigger="focus"
                overlay={
                  <Tooltip id="restricted-mode">
                    Information of this visual is not visible as it lives under
                    a workspace you don’t have access to
                  </Tooltip>
                }
              >
                <Badge pill className={classes.restrictedBadge} as="button">
                  Restricted
                </Badge>
              </VizTooltip>
            </div>
          );
        },
        enableSorting: false,
      }),
      columnHelper.accessor('name', {
        header: () => 'Title',
        cell: (info) => (
          <a
            href={getBlankTargetUrl(
              `${
                arc.frontend.globals.urls.viewDashboardUrl
              }/${info.row.getValue('id')}`
            )}
            target="_blank"
            rel="noreferrer"
          >
            <span>{info.getValue()}</span>
          </a>
        ),
      }),
      columnHelper.accessor('id', {
        header: () => 'ID',
        cell: (info) => info.getValue(),
      }),
      columnHelper.accessor('visuals', {
        header: () => 'Related Visuals',
        cell: (info) => {
          const visuals: number[] = info.getValue();
          return (
            <span data-test-id="applist">
              {visuals?.map((visualId, index) => (
                <React.Fragment key={visualId}>
                  <a
                    href={getBlankTargetUrl(
                      `${
                        arc.frontend.globals.urls.viewDashboardUrl
                      }/${info.row.getValue('id')}`
                    )}
                    target="_blank"
                    rel="noreferrer"
                  >
                    {visualId}
                  </a>
                  {index !== visuals.length - 1 && ', '}
                </React.Fragment>
              ))}
            </span>
          );
        },
        enableSorting: false,
      }),
      columnHelper.accessor('created', {
        header: () => 'Created',
        cell: (info) => {
          const formattedDate = formatDate(info.getValue());
          return (
            <>
              <span
                className={classes.originalDate}
                data-created={info.getValue()}
              ></span>
              <span>{formattedDate}</span>
            </>
          );
        },
      }),
      columnHelper.accessor('updated', {
        header: () => 'Last Updated',
        cell: (info) => {
          const formattedDate = formatTimeDifference(info.getValue());
          return (
            <>
              <span title={formatDate(info.getValue())}>
                <span className={classes.originalDate}>{info.getValue()}</span>
                <span>{formattedDate}</span>
              </span>
            </>
          );
        },
      }),
      columnHelper.accessor('updated_by', {
        header: () => 'Modified By',
        cell: (info) => <span>{info.getValue()}</span>,
      }),
      columnHelper.accessor('workspace_name', {
        header: () => 'Workspace',
        cell: (info) => {
          const workspace = info.getValue();
          return workspace === 'Private' ? (
            <span>{workspace}</span>
          ) : (
            <a
              href={getBlankTargetUrl(
                `${arc.frontend.globals.urls.appsUrl}?group=${workspace}&grouptype=workspaces`
              )}
              target="_blank"
              rel="noreferrer"
            >
              {info.getValue()}
            </a>
          );
        },
      }),
    ],
    []
  );

  const data = useMemo(() => {
    return possibleDeletedReports?.filter(
      (report) => report.type === 'dashboard'
    );
  }, [possibleDeletedReports]);

  const table = useTable({
    data,
    columns,
    getCoreRowModel: getCoreRowModel(),
    getSortedRowModel: getSortedRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    initialState: {
      pagination: {
        pageIndex: 0,
        pageSize: 5,
      },
    },
  });

  return (
    <div className={classes.datasetDeleteModalTable}>
      <ClouderaTable
        isLoading={isPossibleDeletedReportsLoading}
        table={table}
        data-test-id="datasetDeleteModalDashboardsTable"
      />
    </div>
  );
};
