import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  act,
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import { getPossibleDeletedReports } from 'apis/get-possible-deleted-reports';
import { mockPossibleDeletedReport } from './utils/mocks';
import * as React from 'react';
import { DatasetDeleteModal } from './DatasetDeleteModal';
import { deleteDataset } from 'apis/delete-dataset';
import { getGlobalArc } from 'utils/get-global-arc';
import { DatasetDeleteModalDashboardsTable } from 'components/dataset-delete-modal-dashboards-table/DatasetDeleteModalDashboardsTable';
import { DatasetDeleteModalVisualsTable } from 'components/dataset-delete-modal-visuals-table/DatasetDeleteModalVisualsTable';

configure({ testIdAttribute: 'data-test-id' });

const mockAddNotification = jest.fn();
jest.mock('../../../src/stores/use-notification-store.ts', () => ({
  useNotificationStore: () => ({ addNotification: mockAddNotification }),
}));

jest.mock('../../apis/get-possible-deleted-reports', () => ({
  getPossibleDeletedReports: jest.fn(),
}));
jest.mock('../../apis/delete-dataset', () => ({
  deleteDataset: jest.fn(),
}));
jest.mock('arc/misc/misc', () => ({
  getBlankTargetUrl: (url: string) => url,
}));

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: {
          viewDashboardUrl: 'customUrl',
        },
      },
      legacyBridge: {
        events: {
          next: jest.fn(),
        },
      },
    },
    misc: {
      getBlankTargetUrl: jest.fn().mockReturnValue('customUrl'),
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

const arc = getGlobalArc();

const mockDataUrl = arc.frontend.globals.urls.dataUrl;

beforeEach(() => {
  jest
    .mocked(getPossibleDeletedReports)
    .mockResolvedValue(mockPossibleDeletedReport);

  jest.spyOn(window, 'location', 'get').mockReturnValue({
    ...window.location,
    pathname: mockDataUrl,
  });
});

afterEach(() => {
  jest.resetAllMocks();
  jest.restoreAllMocks();
});

describe('Dataset Delete Modal component', () => {
  const datasetDeleteModalProps = {
    dataset: { id: 123, name: 'Test Dataset', hasEditPerm: true },
    isShow: true,
    onClose: () => {},
    onSuccess: () => {},
  };

  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <DatasetDeleteModal {...datasetDeleteModalProps} />
      </QueryClientProvider>
    );
  });

  it('should open the modal', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <DatasetDeleteModal {...datasetDeleteModalProps} />
      </QueryClientProvider>
    );

    await act(async () => {
      arc.frontend.legacyBridge.events.next({
        type: 'dataset-delete-modal',
        payload: { id: 1, name: 'Test Dataset' },
      });
    });

    const datasetDeleteModalHeader = await screen.findByText('Delete Dataset');
    expect(datasetDeleteModalHeader).toBeInTheDocument();
  });

  it('should render the child components without error', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <DatasetDeleteModalDashboardsTable
          possibleDeletedReports={mockPossibleDeletedReport}
          isPossibleDeletedReportsLoading={false}
        />
        <DatasetDeleteModalVisualsTable
          possibleDeletedReports={mockPossibleDeletedReport}
          isPossibleDeletedReportsLoading={false}
        />
      </QueryClientProvider>
    );
  });

  it('should delete the dataset and call addNotification when URL matches dataUrl', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <DatasetDeleteModal {...datasetDeleteModalProps} />
      </QueryClientProvider>
    );

    await act(async () => {
      arc.frontend.legacyBridge.events.next({
        type: 'dataset-delete-modal',
        payload: { id: 1, name: 'Test Dataset 1' },
      });
    });

    const deleteButton = await screen.findByRole('button', {
      name: 'Delete',
    });
    fireEvent.click(deleteButton);

    const input = await screen.findByRole('textbox', { name: /confirm/ });
    fireEvent.change(input, { target: { value: 'DELETE' } });

    fireEvent.click(screen.getByRole('button', { name: 'Delete' }));

    await waitFor(() => expect(deleteDataset).toHaveBeenCalled());
    expect(mockAddNotification).toHaveBeenCalled();
  });
});
