import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import React, { useState } from 'react';
import {
  Badge,
  FormControl,
  FormGroup,
  FormLabel,
  Modal,
} from 'react-bootstrap';
import { legacyBridge } from 'services/legacy-bridge';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import { useMutation, useQuery } from '@tanstack/react-query';
import {
  getPossibleDeletedReports,
  possibleDeletedReportsQueryToken,
} from 'apis/get-possible-deleted-reports';
import { deleteDataset } from 'apis/delete-dataset';
import { useNotificationStore } from 'stores/use-notification-store';
import classes from './DatasetDeleteModal.module.scss';
import classNames from 'classnames';
import { DatasetDeleteModalDashboardsTable } from 'components/dataset-delete-modal-dashboards-table/DatasetDeleteModalDashboardsTable';
import { DatasetDeleteModalVisualsTable } from 'components/dataset-delete-modal-visuals-table/DatasetDeleteModalVisualsTable';
import { DatasetActionsDropdownProps } from 'components/dataset-actions-dropdown/DatasetActionsDropdown';

type DatasetDeleteModalProps = {
  dataset: DatasetActionsDropdownProps;
  isShow: boolean;
  onClose: () => void;
  onSuccess?: () => void;
};

export const DatasetDeleteModal = ({
  dataset,
  isShow,
  onClose,
  onSuccess,
}: DatasetDeleteModalProps) => {
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));
  const [inputValue, setInputValue] = useState('');

  const isValid = inputValue.toLowerCase() === 'delete';

  const {
    data: possibleDeletedReports,
    isLoading: isPossibleDeletedReportsLoading,
  } = useQuery({
    queryKey: [possibleDeletedReportsQueryToken, dataset.id],
    queryFn: () => getPossibleDeletedReports({ id: dataset.id }),
    onError: () => {
      addNotification(
        {
          id: 'roles-danger',
          variant: 'danger',
          message: 'Error fetching specific roles.',
        },
        { delay: 3000 }
      );
    },
  });

  const handleDeleteDataset = useMutation({
    mutationFn: () => deleteDataset({ id: dataset.id }),
    onSuccess: () => {
      addNotification(
        {
          variant: 'success',
          message: `Dataset with ID ${dataset.id} has been deleted`,
        },
        { delay: 2000 }
      );
      if (onSuccess) onSuccess();
      legacyBridge.events.next({
        type: 'successfully-deleted-dataset',
      });
      onClose();
    },
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Failed to delete dataset',
        },
        { delay: 3000 }
      );
    },
  });

  return (
    <Modal
      data-test-id="delete-dataset-modal"
      show={isShow}
      onHide={onClose}
      className={classes.datasetDeleteModal}
      dialogClassName={classes.datasetDeleteModalDialog}
    >
      <Modal.Header closeButton>
        <Modal.Title>Delete Dataset</Modal.Title>
      </Modal.Header>

      <Modal.Body>
        <div className={classNames(classes.datasetDeleteHeader, 'pt-3 pb-3')}>
          Dataset <span className="font-weight-bold">{dataset.name}</span> will
          be deleted
        </div>
        <div
          className={classNames(
            classes.dashboardsCategoryHeader,
            'font-weight-bold'
          )}
        >
          Dashboards{' '}
          <Badge pill variant="secondary" className={classes.badge} as="button">
            {
              possibleDeletedReports?.filter(
                (report) => report.type === 'dashboard'
              ).length
            }
          </Badge>
        </div>
        <div className={classNames(classes.dashboardsCategoryDesc, 'pb-3')}>
          The following dashboards related to this dataset will also be deleted:
        </div>

        <DatasetDeleteModalDashboardsTable
          possibleDeletedReports={possibleDeletedReports ?? []}
          isPossibleDeletedReportsLoading={isPossibleDeletedReportsLoading}
        />

        <div className="font-weight-bold">
          Visuals{' '}
          <Badge pill variant="secondary" className={classes.badge} as="button">
            {
              possibleDeletedReports?.filter(
                (report) => report.type !== 'dashboard'
              ).length
            }
          </Badge>
        </div>
        <div className={classNames(classes.visualsCategoryDesc, 'pb-3')}>
          The following visuals related to this dataset will also be deleted:
        </div>

        <DatasetDeleteModalVisualsTable
          possibleDeletedReports={possibleDeletedReports ?? []}
          isPossibleDeletedReportsLoading={isPossibleDeletedReportsLoading}
        />

        <FormGroup controlId="confirmation-input">
          <FormLabel>
            To confirm, type the word DELETE in the field below and then click
            the Delete button.
          </FormLabel>

          <FormControl
            value={inputValue}
            onChange={(event) => {
              setInputValue(event.target.value);
            }}
            data-test-id="viz-delete-dataset-confirm"
          />
        </FormGroup>
      </Modal.Body>
      <Modal.Footer>
        <ClouderaButton onClick={onClose}>Cancel</ClouderaButton>

        <ClouderaButton
          color="danger"
          icon={<Delete />}
          disabled={!isValid}
          onClick={() => {
            if (isValid) {
              handleDeleteDataset.mutate();
            }
          }}
          testId="delete-confirmation-yes"
        >
          Delete
        </ClouderaButton>
      </Modal.Footer>
    </Modal>
  );
};
