import React, { useCallback, useState } from 'react';
import Modal from 'react-bootstrap/Modal';
import Form from 'react-bootstrap/Form';
import { useLegacyBridge } from 'services/legacy-bridge';
import { Controller, FormProvider, useForm } from 'react-hook-form';
import { Dataset } from 'arc/types/dataset.type';
import styles from './EditDatasetModal.module.scss';
import { useQuery } from '@tanstack/react-query';
import {
  getDatabases,
  getDatatables,
} from 'apis/get-database-and-datatable-list';
import { SQLExplorer } from 'components/sql-explorer/SQLExplorer';
import { VizButton } from 'components/viz-button/VizButton';
import { transformUnknownError } from 'utils/transformUnknownError';
import { Spinner } from 'components/spinner/Spinner';
import { LoadingIndicatorActionButton } from 'components/loading-indicator-action-button/LoadingIndicatorActionButton';
import { Notification } from 'components/notification/Notification';
import { determineDatabaseInQuery } from 'arc/misc/misc';
import { DataConnectionType } from 'types/data-connection.type';

type FormFields = {
  query: string;
  selectedDatabase?: string;
};

type ModalProps = {
  dataset: Dataset;
  onSuccess?: () => void;
  onFailure?: (errorMsg: string) => void;
};

export const EditDatasetSqlModal = (): JSX.Element | null => {
  const [modalProps, setModalProps] = useState<ModalProps | undefined>();
  const [error, setError] = useState<Error>();

  const methods = useForm<FormFields>();
  const selectedDatabase = methods.watch('selectedDatabase');

  const resetModal = () => {
    setModalProps(undefined);
    methods.reset({ selectedDatabase });
    setError(undefined);
  };

  const onDatasetSQLUpdate = (fieldValues: FormFields) => {
    if (modalProps?.dataset) {
      const datasetQuery = fieldValues.query.replace(/;$/, '');
      modalProps.dataset.tables()[0].table_select(datasetQuery);
      return new Promise<void>((resolve) => {
        modalProps.dataset.table_select(datasetQuery).save(
          (error, dataset) => {
            if (!error) {
              modalProps?.onSuccess?.();
              resetModal();
              resolve();
            }
          },
          (error) => {
            if (error) {
              setError(
                transformUnknownError(error, 'Failed to update database')
              );
              resolve();
            }
          }
        );
      });
    }
  };

  const {
    data: databaseList,
    error: databaseLoadError,
    isFetching: isDatabaseListLoading,
  } = useQuery({
    queryKey: ['databaseList', modalProps?.dataset?.dataconnection_id()],
    queryFn: () => getDatabases(modalProps?.dataset?.dataconnection_id()),
    enabled: modalProps?.dataset?.dataconnection_id() !== undefined,
    onError(err) {
      setError(transformUnknownError(err, 'Failed to load databases'));
    },
  });

  const {
    data: tableList,
    error: tableLoadError,
    isFetching: isTableListLoading,
  } = useQuery({
    queryKey: [
      'tableList',
      modalProps?.dataset?.dataconnection_id(),
      selectedDatabase,
    ],
    queryFn: async () => {
      return await getDatatables(
        selectedDatabase,
        modalProps?.dataset?.dataconnection_id()
      );
    },
    onError(err) {
      setError(transformUnknownError(err, 'Failed to load tables'));
    },
    enabled:
      !!selectedDatabase?.length &&
      modalProps?.dataset?.dataconnection_id() !== undefined,
  });

  const onOpen = useCallback(
    ({ payload }: { payload?: ModalProps }) => {
      if (payload) {
        setModalProps(payload);
        // Find the first database mentioned in the query and load it's information
        // for  a better, seamless UX experience
        methods.reset({
          selectedDatabase: determineDatabaseInQuery(
            payload.dataset?.tables()[0].table_select()
          ),
        });
      }
    },
    [methods]
  );

  useLegacyBridge('open-edit-dataset-sql-modal', onOpen);
  const defaultDatabase = determineDatabaseInQuery(
    modalProps?.dataset?.tables()[0].table_select()
  );
  return modalProps === undefined ? null : (
    <Modal
      show={!!modalProps}
      onHide={() => {
        resetModal();
      }}
      size="lg"
    >
      <Modal.Header>
        <Modal.Title>Edit Dataset SQL</Modal.Title>
      </Modal.Header>
      <FormProvider {...methods}>
        <Form onSubmit={methods.handleSubmit(onDatasetSQLUpdate)}>
          <Modal.Body>
            <Notification
              title="Error"
              message={error?.message || ''}
              variant="danger"
              show={!!error}
              dismissible
              onClose={() => setError(undefined)}
            />
            <br></br>
            <Controller
              control={methods.control}
              rules={{
                required: true,
                validate: (value) =>
                  value.trim() !==
                  modalProps?.dataset?.tables()[0].table_select().trim(),
              }}
              defaultValue={modalProps.dataset?.tables()[0].table_select()}
              render={({ field: { onChange, value } }) => (
                <SQLExplorer
                  connection={{
                    id: Number(modalProps.dataset.dataconnection_id()) ?? -1,
                    type: modalProps.dataset.dataconnection_type() as DataConnectionType,
                  }}
                  databaseList={
                    databaseList ?? (defaultDatabase ? [defaultDatabase] : [])
                  }
                  tableList={tableList}
                  isTableListLoading={isTableListLoading}
                  isDatabaseListLoading={isDatabaseListLoading}
                  tableLoadError={tableLoadError}
                  databaseLoadError={databaseLoadError}
                  query={value}
                  onChange={onChange}
                />
              )}
              name="query"
            />
          </Modal.Body>

          <Modal.Footer>
            <div className={`position-absolute ${styles.left0}`}>
              {(isTableListLoading || isDatabaseListLoading) && <Spinner />}
            </div>
            <VizButton
              color="secondary"
              size="medium"
              onClick={() => {
                resetModal();
              }}
              testId="close"
            >
              Close
            </VizButton>

            <LoadingIndicatorActionButton
              type="submit"
              isSaving={methods.formState.isSubmitting}
              label="Save"
              isValid={methods.formState.isValid}
            ></LoadingIndicatorActionButton>
          </Modal.Footer>
        </Form>
      </FormProvider>
    </Modal>
  );
};
