import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import { editJob } from 'apis/jobs/edit-job';
import { getEmailTemplates } from 'apis/jobs/email-templates/get-email-templates';
import * as getScheduleGroups from 'apis/jobs/schedule-groups/get-schedule-groups';
import * as React from 'react';
import { JobThresholdTriggerType } from 'types/job.type';

import {
  mockEmailJob,
  mockEmailTemplateOptions,
  mockScheduleGroups,
} from 'pages/job-page/utils/mocks';
import { EmailJobModal } from './EmailJobModal';

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: { previewCustomEmailUrl: 'customUrl' },
        settings: {
          emailHostUser: 'test@test.com',
          uploadMaxrowsDefault: 1000000,
        },
        authentication: { username: 'user' },
        emailAddresses: '[]',
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

jest.mock('apis/jobs/edit-job', () => ({
  editJob: jest.fn(),
}));
jest.mock('apis/jobs/email-templates/get-email-templates', () => ({
  getEmailTemplates: jest.fn(),
}));
jest.mock('arc/misc/misc', () => ({
  getBlankTargetUrl: (url: string) => url,
}));
jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

configure({ testIdAttribute: 'data-test-id' });

beforeEach(() => {
  jest
    .spyOn(getScheduleGroups, 'getScheduleGroups')
    .mockResolvedValue(mockScheduleGroups);
  jest.mocked(getEmailTemplates).mockResolvedValue(mockEmailTemplateOptions);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('EmailJobModal component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <EmailJobModal job={mockEmailJob} />
      </QueryClientProvider>
    );
  });

  it('populates the input fields with the current values', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <EmailJobModal job={mockEmailJob} />
      </QueryClientProvider>
    );

    expect(screen.getByLabelText('To')).toBeInTheDocument();
    expect(screen.getByLabelText('CC')).toBeInTheDocument();
    expect(
      screen.getByLabelText('Error Notification Emails To')
    ).toBeInTheDocument();
    expect(screen.getByLabelText('Reply To')).toBeInTheDocument();
    expect(screen.getByLabelText('Email Template')).toBeInTheDocument();
    expect(screen.getByLabelText('Subject')).toBeInTheDocument();
    expect(screen.getByLabelText('Message')).toBeInTheDocument();

    fireEvent.click(screen.getByRole('tab', { name: 'Schedule Details' }));
    expect(screen.getByLabelText('Name')).toBeInTheDocument();
    expect(screen.getByLabelText('Schedule')).toBeInTheDocument();

    await waitFor(() => {
      expect(screen.getByRole('form')).toHaveFormValues({
        name: 'Mock Job',
        'emailTarget.templateId': '1',
        'emailTarget.to': ['test@test.com', 'test2@test.com'],
        'emailTarget.cc': '',
        'emailTarget.errorTo': '',
        'emailTarget.replyTo': '',
      });
    });
  });

  describe('when on the Schedule Details Tab', () => {
    it('displays the snapshot information in disabled fields', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <EmailJobModal job={mockEmailJob} />
        </QueryClientProvider>
      );

      expect(
        screen.getByRole('checkbox', {
          name: 'Include page parameters',
          checked: false,
        })
      ).toBeDisabled();
      expect(
        screen.getByRole('checkbox', {
          name: 'Use URL alias',
          checked: false,
        })
      ).toBeDisabled();
      expect(
        screen.getByRole('checkbox', {
          name: 'Embedded Image',
          checked: true,
        })
      ).toBeDisabled();
      expect(
        screen.getByRole('checkbox', {
          name: 'PNG',
          checked: false,
        })
      ).toBeDisabled();
      expect(
        screen.getByRole('checkbox', {
          name: 'PDF',
          checked: true,
        })
      ).toBeDisabled();
      expect(
        screen.getByRole('checkbox', {
          name: 'XLSX',
          checked: false,
        })
      ).toBeDisabled();
      expect(
        screen.getByRole('checkbox', {
          name: 'CSV',
          checked: false,
        })
      ).toBeDisabled();
    });

    describe('when the selected schedule is Create Interval', () => {
      it('should display two new inputs for the new interval', async () => {
        const DOWN_ARROW = { keyCode: 40 };

        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('tab', { name: 'Schedule Details' }));
        fireEvent.keyDown(screen.getByLabelText('Schedule'), DOWN_ARROW);
        await waitFor(() => {
          expect(screen.getByText('Create Interval')).toBeInTheDocument();
        });
        fireEvent.click(screen.getByText('Create Interval'));

        await waitFor(() => {
          expect(screen.getByTestId('intervalName')).toBeInTheDocument();
        });
        expect(
          screen.getByTestId('dropdown-input-minutes')
        ).toBeInTheDocument();
      });

      it('should not save the job with create interval selected and missing interval data', async () => {
        const DOWN_ARROW = { keyCode: 40 };

        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('tab', { name: 'Schedule Details' }));
        fireEvent.keyDown(screen.getByLabelText('Schedule'), DOWN_ARROW);
        await waitFor(() => {
          expect(screen.getByText('Create Interval')).toBeInTheDocument();
        });
        fireEvent.click(screen.getByText('Create Interval'));

        await waitFor(() => {
          expect(screen.getByTestId('intervalName')).toBeInTheDocument();
        });
        expect(
          screen.getByTestId('dropdown-input-minutes')
        ).toBeInTheDocument();

        fireEvent.click(screen.getByRole('button', { name: 'Save' }));

        await waitFor(() => {
          expect(screen.getByTestId('intervalError')).toBeInTheDocument();
        });
        expect(screen.getByTestId('intervalNameError')).toBeInTheDocument();
      });
    });
  });

  describe('when on the Email Details Tab', () => {
    describe('when clicking on the Preview Email Template link', () => {
      it('should open a new window and show the template', () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        const previewLink = screen.getByRole('link', {
          name: 'Preview Email Template',
        });
        expect(previewLink).toHaveAttribute('href', 'customUrl/1');
        expect(previewLink).toHaveAttribute('target', '_blank');
      });
    });

    describe('when the Show Job Parameters button is pressed', () => {
      it('should show the job parameters that can be added to the email message', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByTestId('showJobParameters'));
        await waitFor(() =>
          expect(
            screen.getByRole('button', {
              name: 'Job Parameter Current Username',
            })
          ).toBeEnabled()
        );

        fireEvent.click(
          screen.getByRole('button', { name: 'Job Parameter Relative Link' })
        );
        await waitFor(() =>
          expect(screen.getByRole('form')).toHaveFormValues({
            name: 'Mock Job',
            'emailTarget.templateId': '1',
            'emailTarget.to': ['test@test.com', 'test2@test.com'],
            'emailTarget.cc': '',
            'emailTarget.errorTo': '',
            'emailTarget.replyTo': '',
            'emailTarget.message': 'Default message<<relative_url>>',
          })
        );
      });
    });

    describe('when the Show Page Parameters button is pressed', () => {
      it('should show the url parameters which can edited', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        const showPageParams = screen.getByTestId('showPageParameters');
        fireEvent.click(showPageParams);
        await waitFor(() =>
          expect(showPageParams).toHaveTextContent('Hide Page Parameters')
        );
        expect(
          screen.getByLabelText('Edit Page Parameters')
        ).toBeInTheDocument();
      });
    });

    describe('when clicking on the Visual Id', () => {
      it('should open a new window and show the dashboard/app with the visual', () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        const visualLink = screen.getByRole('link', { name: 'Report ID' });
        expect(visualLink).toHaveAttribute('href', 'url');
        expect(visualLink).toHaveAttribute('target', '_blank');
      });
    });
  });

  describe('when on the Trigger Conditions Tab', () => {
    it('displays the threshold related inputs instead of the email related ones', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <EmailJobModal job={mockEmailJob} />
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByRole('tab', { name: 'Threshold Details' }));

      expect(screen.getByLabelText('Trigger Visual ID')).toBeInTheDocument();
      expect(
        screen.getByRole('radio', {
          name: 'Threshold met',
        })
      ).toBeChecked();
      expect(
        screen.getByRole('radio', {
          name: 'Value changed since last assessment',
        })
      ).not.toBeChecked();
      expect(screen.getAllByLabelText('Trigger Condition Value').length).toBe(
        2
      );
      expect(
        screen.getAllByLabelText('Trigger Condition Equality').length
      ).toBe(2);
    });

    describe('when clicking on the Add Condition button', () => {
      it('should display a new condition input row', () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('tab', { name: 'Threshold Details' }));
        fireEvent.click(screen.getByRole('button', { name: 'Add Condition' }));

        expect(screen.getAllByLabelText('Trigger Condition Value').length).toBe(
          3
        );
        expect(
          screen.getAllByLabelText('Trigger Condition Equality').length
        ).toBe(3);
      });
    });

    describe('when clicking on the - icon next to a condition', () => {
      it('should remove the condition', () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('tab', { name: 'Threshold Details' }));
        fireEvent.click(
          screen.getAllByRole('button', { name: 'Remove Condition' })[1]
        );

        expect(screen.getAllByLabelText('Trigger Condition Value').length).toBe(
          1
        );
        expect(
          screen.getAllByLabelText('Trigger Condition Equality').length
        ).toBe(1);
      });
    });
  });

  describe('when the cancel button is pressed', () => {
    it('closes the overlay without saving the values', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <EmailJobModal job={mockEmailJob} />
        </QueryClientProvider>
      );
      fireEvent.click(screen.getByRole('button', { name: 'Cancel' }));
      expect(jest.mocked(editJob)).not.toHaveBeenCalled();
    });
  });

  describe('when the save button is pressed', () => {
    describe('and the form contains valid data', () => {
      it('saves the job and closes the overlay', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('tab', { name: 'Schedule Details' }));
        fireEvent.change(screen.getByLabelText('Name'), {
          target: { value: 'Edited name' },
        });

        // go to the other tab and modify some fields there as well
        fireEvent.click(screen.getByRole('tab', { name: 'Threshold Details' }));
        fireEvent.change(screen.getByLabelText('Trigger Visual ID'), {
          target: { value: 5 },
        });
        fireEvent.click(
          screen.getByRole('radio', {
            name: 'Value changed since last assessment',
          })
        );

        fireEvent.click(screen.getByRole('button', { name: 'Add Condition' }));
        const conditionValues = screen.getAllByLabelText(
          'Trigger Condition Value'
        );
        const conditionEqualities = screen.getAllByLabelText(
          'Trigger Condition Equality'
        );
        fireEvent.change(conditionEqualities[2], {
          target: { value: '=' },
        });
        fireEvent.change(conditionValues[2], {
          target: { value: 6 },
        });

        fireEvent.click(screen.getByRole('button', { name: 'Save' }));

        await waitFor(() => {
          expect(jest.mocked(editJob)).toHaveBeenCalledWith(1, {
            emailTarget: {
              cc: [],
              templateId: 1,
              errorTo: [],
              message: 'Default message',
              replyTo: '',
              subject: 'Cloudera Data Visualization',
              to: ['test@test.com', 'test2@test.com'],
              urlParams: 'embed=true',
              includeCSV: false,
              includeEmbeddedImage: true,
              includePageParams: false,
              includePdf: true,
              includePng: false,
              includeXLSX: false,
              useUrlAlias: false,
              pdfOption: 'pdf-current',
              csvDownloadRows: 5000,
            },
            thresholdTarget: {
              minimumTimeInterval: 604800,
              triggerConditions: [
                { equality: '>', value: 5 },
                { equality: '<', value: 2 },
                { equality: '=', value: 6 },
              ],
              triggerId: 5,
              triggerType: JobThresholdTriggerType.ValueChanged,
            },
            name: 'Edited name',
            scheduleGroupId: 1,
            isCreate: false,
            isScheduled: true,
            isThreshold: true,
          });
        });
      });
    });

    describe('and the form contains invalid data', () => {
      it('shows an error below the incorrect fields', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <EmailJobModal job={mockEmailJob} />
          </QueryClientProvider>
        );

        fireEvent.change(screen.getByLabelText('Reply To'), {
          target: { value: 'test' },
        });

        fireEvent.click(screen.getByRole('tab', { name: 'Schedule Details' }));
        fireEvent.change(screen.getByLabelText('Name'), {
          target: { value: '' },
        });
        fireEvent.change(screen.getByLabelText('Subject'), {
          target: { value: '' },
        });

        fireEvent.click(screen.getByRole('button', { name: 'Save' }));

        const nameError = screen.getByTestId('nameError');
        const subjectError = screen.getByTestId('subjectError');
        const replyToError = screen.getByTestId('replyToError');
        await waitFor(() => {
          expect(nameError).toHaveTextContent('Name is a required field');
        });
        expect(subjectError).toHaveTextContent('Subject is a required field');
        expect(replyToError).toHaveTextContent('test is not a valid email');

        expect(jest.mocked(editJob)).not.toHaveBeenCalled();
      });

      describe('and the invalid input is modified', () => {
        it('no longer shows the error message', async () => {
          render(
            <QueryClientProvider client={new QueryClient()}>
              <EmailJobModal job={mockEmailJob} />
            </QueryClientProvider>
          );

          fireEvent.change(screen.getByLabelText('Name'), {
            target: { value: '' },
          });
          fireEvent.click(screen.getByRole('button', { name: 'Save' }));

          const nameError = await screen.findByTestId('nameError');
          await waitFor(() =>
            expect(nameError).toHaveTextContent('Name is a required field')
          );

          fireEvent.change(screen.getByLabelText('Name'), {
            target: { value: 'New Name' },
          });

          await waitFor(() => expect(nameError).toHaveTextContent(''));
        });
      });
    });
  });
});
