import ExpandIcon from '@cloudera-internal/cuix-core/icons/svg/ic-accordion-expand.svg';
import CollapseIcon from '@cloudera-internal/cuix-core/icons/svg/ic-collapse.svg';
import ErrorIcon from '@cloudera-internal/cuix-core/icons/svg/ic-error.svg';
import HelpIcon from '@cloudera-internal/cuix-core/icons/svg/ic-help.svg';
import { cdlRed500 } from '@cloudera-internal/cuix-core/variables';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { TokenInput } from 'components/token-input/TokenInput';
import { VizTooltip } from 'components/VizTooltip';
import { useToggle } from 'hooks/use-toggle';
import React, { useMemo } from 'react';
import { Accordion, Form, Popover } from 'react-bootstrap';
import { Controller, UseFormReturn } from 'react-hook-form';
import { getEmailAddressHints } from 'utils/get-email-address-hints';
import { EmailJobFormSchema } from '../email-job-form-schema';
import { EmailJobAliasSection } from './EmailJobAliasSection';

type EmailJobAdvancedSectionProps = {
  form: UseFormReturn<EmailJobFormSchema>;
  dashboardId?: number;
};

export const EmailJobAdvancedSection = ({
  form,
  dashboardId,
}: EmailJobAdvancedSectionProps): JSX.Element => {
  const [showPageParameters, togglePageParameters] = useToggle();
  const [showAdvanced, toggleShowAdvanced] = useToggle();
  const { control, register, formState, trigger } = form;
  const isCreate = form.watch('isCreate');

  const emailAddressOptions = useMemo(
    () =>
      getEmailAddressHints().map((email) => ({
        label: email,
        value: email,
      })),
    []
  );

  const hasError = useMemo(
    () =>
      formState.errors.emailTarget &&
      (formState.errors.emailTarget.errorTo ||
        formState.errors.emailTarget.replyTo ||
        formState.errors.emailTarget.useUrlAlias ||
        formState.errors.emailTarget.includePageParams),
    [formState.errors.emailTarget]
  );

  return (
    <Accordion activeKey={showAdvanced ? 'advanced' : ''}>
      <Accordion.Toggle
        as="p"
        className="mb-0 cursor-pointer d-flex align-items-center font-weight-bold"
        eventKey="advanced"
        data-test-id="advancedToggle"
        onClick={toggleShowAdvanced}
      >
        {showAdvanced ? (
          <CollapseIcon className="mr-2" />
        ) : (
          <ExpandIcon className="mr-2" />
        )}
        Advanced Settings
        {hasError && <ErrorIcon className="ml-2" color={cdlRed500} />}
      </Accordion.Toggle>
      <Accordion.Collapse eventKey="advanced">
        <div className="mt-3">
          <Form.Group controlId="emailErrorTo">
            <Form.Label className="d-flex align-items-center gap-1">
              Error Notification Emails To
              <VizTooltip
                overlay={
                  <Popover id="emailErrorToTooltip">
                    <Popover.Content>
                      Enter email addresses here that we will send error
                      notifications to
                    </Popover.Content>
                  </Popover>
                }
              >
                <HelpIcon color="#2e89c5" />
              </VizTooltip>
            </Form.Label>
            <Controller
              name="emailTarget.errorTo"
              control={control}
              render={({ field: { onBlur, ...field }, fieldState }) => (
                <TokenInput
                  {...field}
                  onBlur={() => {
                    if (fieldState.isDirty || fieldState.invalid) {
                      trigger(field.name);
                    }
                    onBlur();
                  }}
                  id="emailErrorTo"
                  isInvalid={!!formState.errors.emailTarget?.errorTo}
                  options={emailAddressOptions}
                />
              )}
            />
            <Form.Control.Feedback type="invalid">
              {formState.errors.emailTarget?.errorTo?.message ??
                formState.errors.emailTarget?.errorTo?.find?.(
                  (item) => item?.message
                )?.message}
            </Form.Control.Feedback>
          </Form.Group>
          <Form.Group controlId="emailReplyTo">
            <Form.Label className="d-flex align-items-center gap-1">
              Reply To
              <VizTooltip
                overlay={
                  <Popover id="emailErrorToTooltip">
                    <Popover.Content>
                      Enter an email address if you would like responses to go
                      to a different email address than sender
                    </Popover.Content>
                  </Popover>
                }
              >
                <HelpIcon color="#2e89c5" />
              </VizTooltip>
            </Form.Label>
            <Form.Control
              {...register('emailTarget.replyTo')}
              isInvalid={!!formState.errors?.emailTarget?.replyTo}
            />
            <Form.Control.Feedback type="invalid" data-test-id="replyToError">
              {formState.errors.emailTarget?.replyTo?.message}
            </Form.Control.Feedback>
          </Form.Group>

          <EmailJobAliasSection form={form} dashboardId={dashboardId} />

          <Form.Group>
            <Form.Label className="d-flex align-items-center gap-1">
              Page Parameters
              <VizTooltip
                overlay={
                  <Popover id="pageParamsTooltip">
                    <Popover.Content>
                      Changing the parameters of the snapshots after saving is
                      not available. As pasted in the message however it is
                      editable
                    </Popover.Content>
                  </Popover>
                }
              >
                <HelpIcon color="#2e89c5" />
              </VizTooltip>
            </Form.Label>
            <Form.Check
              disabled={!isCreate}
              id="includePageParams"
              label="Include page parameters"
              {...register('emailTarget.includePageParams', {
                onChange: (e) => {
                  if (e.target.checked) {
                    form.resetField('emailTarget.urlParams');
                  } else {
                    const params = new URLSearchParams(
                      form.formState.defaultValues?.emailTarget?.urlParams ?? ''
                    );

                    form.setValue(
                      'emailTarget.urlParams',
                      new URLSearchParams({
                        embed: params.get('embed') ?? '',
                        sheet: params.get('sheet') ?? '',
                      }).toString()
                    );
                  }
                },
              })}
            />
            <ClouderaButton
              type="button"
              variant="link"
              onClick={togglePageParameters}
              testId="showPageParameters"
            >
              {showPageParameters ? 'Hide' : 'Show'} Page Parameters
            </ClouderaButton>
            {showPageParameters && (
              <Form.Control
                className="mt-2"
                as="textarea"
                aria-label="Edit Page Parameters"
                rows={3}
                {...register('emailTarget.urlParams')}
              />
            )}
          </Form.Group>
        </div>
      </Accordion.Collapse>
    </Accordion>
  );
};
