import HelpIcon from '@cloudera-internal/cuix-core/icons/svg/ic-help.svg';
import { getBlankTargetUrl } from 'arc/misc/misc';
import { ClouderaLink } from 'components/cloudera-link/ClouderaLink';
import { TokenInput } from 'components/token-input/TokenInput';
import { VizTooltip } from 'components/VizTooltip';
import React, { useMemo } from 'react';
import { Col, Form, Popover } from 'react-bootstrap';
import { Controller, UseFormReturn } from 'react-hook-form';
import { getGlobalArc } from 'utils/get-global-arc';
import { EmailJobFormSchema } from '../email-job-form-schema';
import classes from '../EmailJobModal.module.scss';
import { EmailJobAdvancedSection } from './EmailJobAdvancedSection';
import { EmailJobMessageSection } from './EmailJobMessageSection';
import { getEmailAddressHints } from 'utils/get-email-address-hints';

const arc = getGlobalArc();

type EmailJobEmailTabProps = {
  reportId: number;
  reportType: 'visual' | 'sheet';
  reportUrl: string;
  form: UseFormReturn<EmailJobFormSchema>;
  dashboardId?: number;
  disableXlsxAndCsv?: boolean;
};

export const EmailJobEmailTab = ({
  reportId,
  reportType,
  reportUrl,
  dashboardId,
  disableXlsxAndCsv,
  form,
}: EmailJobEmailTabProps): JSX.Element => {
  const { control, formState, register, trigger } = form;

  const isCreate = form.watch('isCreate');
  const includeXLSX = form.watch('emailTarget.includeXLSX');
  const includeCSV = form.watch('emailTarget.includeCSV');
  const includePdf = form.watch('emailTarget.includePdf');

  const emailAddressOptions = useMemo(
    () =>
      getEmailAddressHints().map((email) => ({
        label: email,
        value: email,
      })),
    []
  );

  return (
    <>
      {!isCreate && (
        <Form.Group>
          <Form.Label id="visualIdLabel">Report ID</Form.Label>
          <ClouderaLink
            aria-labelledby="visualIdLabel"
            href={getBlankTargetUrl(reportUrl)}
            target="_blank"
            className="d-block w-fit"
          >
            {reportId}
          </ClouderaLink>
        </Form.Group>
      )}

      <Form.Row>
        <Form.Group as={Col} controlId="emailTo">
          <Form.Label className="required-label">To</Form.Label>
          <Controller
            name="emailTarget.to"
            control={control}
            render={({ field: { onBlur, ...field }, fieldState }) => (
              <TokenInput
                {...field}
                onBlur={() => {
                  if (fieldState.isDirty || fieldState.invalid) {
                    trigger(field.name);
                  }
                  onBlur();
                }}
                id="emailTo"
                isInvalid={!!formState.errors.emailTarget?.to}
                options={emailAddressOptions}
              />
            )}
          />

          <Form.Control.Feedback type="invalid" data-test-id="emailToError">
            {formState.errors.emailTarget?.to?.message ??
              formState.errors.emailTarget?.to?.find?.((item) => item?.message)
                ?.message}
          </Form.Control.Feedback>
        </Form.Group>

        <Form.Group as={Col} controlId="emailCc">
          <Form.Label>CC</Form.Label>
          <Controller
            name="emailTarget.cc"
            control={control}
            render={({ field: { onBlur, ...field }, fieldState }) => (
              <TokenInput
                {...field}
                onBlur={() => {
                  if (fieldState.isDirty || fieldState.invalid) {
                    trigger(field.name);
                  }
                  onBlur();
                }}
                id="emailCc"
                isInvalid={!!formState.errors.emailTarget?.cc}
                options={emailAddressOptions}
              />
            )}
          />
          <Form.Control.Feedback type="invalid">
            {formState.errors.emailTarget?.cc?.message ??
              formState.errors.emailTarget?.cc?.find?.((item) => item?.message)
                ?.message}
          </Form.Control.Feedback>
        </Form.Group>
      </Form.Row>

      <Form.Group>
        <Form.Label>From</Form.Label>
        <Form.Control
          value={`${arc.frontend.globals.authentication.username}<${arc.frontend.globals.settings.emailHostUser}>`}
          disabled
        />
      </Form.Group>

      <Form.Group controlId="emailSubject">
        <Form.Label className="required-label">Subject</Form.Label>
        <Form.Control
          {...register('emailTarget.subject')}
          data-test-id="emailSubject"
          isInvalid={!!formState.errors.emailTarget?.subject}
        />
        <Form.Control.Feedback type="invalid" data-test-id="subjectError">
          {formState.errors.emailTarget?.subject?.message}
        </Form.Control.Feedback>
      </Form.Group>

      <EmailJobMessageSection form={form} />

      <Form.Group>
        <Form.Label className="d-flex align-items-center gap-1">
          Attach
          <VizTooltip
            overlay={
              <Popover id="attachTooltip">
                <Popover.Content>
                  Note that emails larger than 25MB might not be sent depending
                  on email server limits. Attachment type cannot be edited later
                </Popover.Content>
              </Popover>
            }
          >
            <HelpIcon color="#2e89c5" />
          </VizTooltip>
        </Form.Label>
        <div>
          <Form.Check
            inline
            disabled={!isCreate}
            id="attachEmbeddedImage"
            label="Embedded Image"
            {...register('emailTarget.includeEmbeddedImage')}
          />
          <Form.Check
            inline
            disabled={!isCreate}
            id="attachPng"
            label="PNG"
            {...register('emailTarget.includePng')}
          />
          <Form.Check
            inline
            disabled={!isCreate}
            id="attachPdf"
            label="PDF"
            {...register('emailTarget.includePdf')}
          />
          <Form.Check
            inline
            disabled={!isCreate || disableXlsxAndCsv}
            id="attachXlsx"
            label="XLSX"
            {...register('emailTarget.includeXLSX')}
          />
          <Form.Check
            inline
            disabled={!isCreate || disableXlsxAndCsv}
            id="attachCsv"
            label="CSV"
            {...register('emailTarget.includeCSV')}
          />
          <div className="d-flex flex-column gap-2">
            {(includeXLSX || includeCSV) && (
              <Form.Group>
                <div className="d-flex align-items-baseline gap-2">
                  <span>Limit:</span>
                  <Form.Control
                    type="number"
                    className={classes.limitInput}
                    aria-label="Download row limit"
                    defaultValue={5000}
                    disabled={!isCreate}
                    {...register('emailTarget.csvDownloadRows')}
                  />
                  <span>rows</span>
                </div>
                {formState.errors.emailTarget?.csvDownloadRows && (
                  <Form.Control.Feedback type="invalid" className="d-block">
                    {formState.errors.emailTarget?.csvDownloadRows?.message}
                  </Form.Control.Feedback>
                )}
              </Form.Group>
            )}
            {includePdf && reportType === 'sheet' && (
              <div className="d-flex align-items-baseline gap-2">
                <span className="shrink-0">Pdf option:</span>
                <Form.Control
                  className={classes.pdfOptionInput}
                  as="select"
                  aria-label="PDF Option"
                  disabled={!isCreate}
                  defaultValue="pdf-current"
                  {...register('emailTarget.pdfOption')}
                >
                  <option value="pdf-current">Current sheet</option>
                  <option value="pdf-one-viz">
                    One visual per page on current sheet
                  </option>
                  <option value="pdf-all-sheets">One sheet per page</option>
                </Form.Control>
              </div>
            )}
          </div>
        </div>
      </Form.Group>

      <hr />

      <EmailJobAdvancedSection form={form} dashboardId={dashboardId} />
    </>
  );
};
