import { useQuery } from '@tanstack/react-query';
import {
  getEmailTemplates,
  getEmailTemplatesQueryToken,
} from 'apis/jobs/email-templates/get-email-templates';
import { getBlankTargetUrl } from 'arc/misc/misc';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { ClouderaLink } from 'components/cloudera-link/ClouderaLink';
import { useToggle } from 'hooks/use-toggle';
import React, { useCallback, useRef } from 'react';
import { Col, Form, Row } from 'react-bootstrap';
import { Controller, UseFormReturn } from 'react-hook-form';
import Select from 'react-select';
import { getGlobalArc } from 'utils/get-global-arc';
import { EmailJobFormSchema } from '../email-job-form-schema';
import { JobParameterButton } from 'pages/job-page/components/job-parameter-button/JobParameterButton';

const arc = getGlobalArc();

type JobParam = {
  value: string;
  label: string;
};

const generalJobParams: JobParam[] = [
  { value: '<<cloudera_user_id>>', label: 'Current Username' },
  { value: '<<app_id>>', label: 'ID' },
  { value: '<<app_name>>', label: 'Name' },
  { value: '<<app_url>>', label: 'Link' },
  { value: '<<relative_url>>', label: 'Relative Link' },
  { value: '<<start_time>>', label: 'Job Start Time' },
  { value: '<<params>>', label: 'Page Parameters' },
  { value: '<<run_time>>', label: 'Total Job Run Time (as of time emailed)' },
];

const scheduledJobParams: JobParam[] = [
  { value: '<<next_run_time>>', label: 'Next Run Time' },
  { value: '<<run_interval>>', label: 'Interval Name' },
  { value: '<<run_interval_cron>>', label: 'Interval Cron' },
];

const thresholdJobParams: JobParam[] = [
  { value: '<<trigger_fullconditional>>', label: 'Full Trigger Conditional' },
  { value: '<<trigger_data_value>>', label: 'Data When Triggered' },
  { value: '<<trigger_difference>>', label: 'Change in Data When Triggered' },
  { value: '<<triggered_time>>', label: 'When Trigger Assessed' },
  { value: '<<trigger_id>>', label: 'Trigger ID' },
  { value: '<<trigger_name>>', label: 'Trigger Name' },
  { value: '<<minimum_time_interval>>', label: 'Email Limit' },
];

type EmailJobMessageSectionProps = {
  form: UseFormReturn<EmailJobFormSchema>;
};

export const EmailJobMessageSection = ({
  form,
}: EmailJobMessageSectionProps): JSX.Element => {
  const messageInputRef = useRef<HTMLTextAreaElement | null>(null);
  const { control, getValues, setValue, register } = form;
  const [showJobParameters, toggleJobParameters] = useToggle();
  const isScheduled = form.watch('isScheduled');
  const isThreshold = form.watch('isThreshold');

  const { data: emailTemplateOptions } = useQuery({
    queryKey: [getEmailTemplatesQueryToken],
    queryFn: getEmailTemplates,
    placeholderData: [],
    staleTime: 300000,
    select: (emailTemplates) =>
      emailTemplates.map(({ id, name, defaultMessage }) => ({
        value: id,
        label: name,
        defaultMessage,
      })),
  });

  const addParamToMessage = useCallback(
    (param: string) => {
      const start = messageInputRef.current?.selectionStart ?? 0;
      const end = messageInputRef.current?.selectionEnd ?? 0;
      const nextCursorPosition = start + param.length;
      const value = getValues('emailTarget.message') ?? '';

      // Replace the selected text with the param
      setValue(
        'emailTarget.message',
        `${value.slice(0, start)}${param}${value.slice(end)}`
      );

      // Restore focus and cursor position
      messageInputRef.current?.focus();
      messageInputRef.current?.setSelectionRange(
        nextCursorPosition,
        nextCursorPosition
      );
    },
    [getValues, setValue]
  );

  return (
    <>
      <Form.Group controlId="emailTemplateId">
        <Form.Label>Email Template</Form.Label>
        <Controller
          name="emailTarget.templateId"
          control={control}
          render={({ field: { onChange, value, onBlur, name } }) => {
            const template = emailTemplateOptions?.find(
              (option) => option.value === value
            );
            return (
              <>
                <Select
                  options={emailTemplateOptions}
                  value={template}
                  onChange={(event) => {
                    onChange(event?.value ?? '');
                    form.setValue(
                      'emailTarget.message',
                      emailTemplateOptions?.find(
                        (option) => option.value === event?.value
                      )?.defaultMessage ?? ''
                    );
                  }}
                  onBlur={onBlur}
                  name={name}
                  inputId="emailTemplateId"
                  isClearable
                  placeholder="Default Email Template"
                />

                <ClouderaLink
                  href={getBlankTargetUrl(
                    `${arc.frontend.globals.urls.previewCustomEmailUrl}${
                      value ? `/${value}` : ''
                    }`
                  )}
                  target="_blank"
                >
                  Preview Email Template
                </ClouderaLink>
              </>
            );
          }}
        />
      </Form.Group>
      <Form.Group controlId="emailMessage">
        <Form.Label>Message</Form.Label>
        <Form.Control
          as="textarea"
          data-test-id="emailMessage"
          placeholder="Message"
          rows={3}
          {...register('emailTarget.message')}
          ref={(instance: HTMLTextAreaElement | null) => {
            messageInputRef.current = instance;
            register('emailTarget.message').ref(instance);
          }}
        />

        <ClouderaButton
          type="button"
          variant="link"
          testId="showJobParameters"
          onClick={toggleJobParameters}
        >
          {showJobParameters ? 'Hide' : 'Show'} Job Parameters
        </ClouderaButton>

        {showJobParameters && (
          <div>
            <div>Click on following links to add parameter to message:</div>
            <Row xs={1} sm={3} md={3} className="row-gap-2 mt-2 mr-0">
              <Col>
                {generalJobParams.map(({ value, label }) => (
                  <JobParameterButton
                    key={value}
                    value={value}
                    label={label}
                    onSelect={addParamToMessage}
                  />
                ))}
              </Col>
              {isScheduled && (
                <Col>
                  {scheduledJobParams.map(({ value, label }) => (
                    <JobParameterButton
                      key={value}
                      value={value}
                      label={label}
                      onSelect={addParamToMessage}
                    />
                  ))}
                </Col>
              )}
              {isThreshold && (
                <Col>
                  {thresholdJobParams.map(({ value, label }) => (
                    <JobParameterButton
                      key={value}
                      value={value}
                      label={label}
                      onSelect={addParamToMessage}
                    />
                  ))}
                </Col>
              )}
            </Row>
          </div>
        )}
      </Form.Group>
    </>
  );
};
