import { useGetScheduleGroupsQuery } from 'apis/jobs/schedule-groups/get-schedule-groups';
import { Cron } from 'components/cron/Cron';
import cronstrue from 'cronstrue';
import moment from 'moment';
import React from 'react';
import { Form } from 'react-bootstrap';
import { Controller, UseFormReturn } from 'react-hook-form';
import Select from 'react-select';
import { EmailJobFormSchema } from '../email-job-form-schema';
import classes from '../EmailJobModal.module.scss';

type EmailJobScheduleTabProps = {
  form: UseFormReturn<EmailJobFormSchema>;
};

export const EmailJobScheduleTab = ({
  form,
}: EmailJobScheduleTabProps): JSX.Element => {
  const { control, formState, register } = form;

  const isCreate = form.watch('isCreate');
  const isScheduled = form.watch('isScheduled');
  const scheduleGroupId = form.watch('scheduleGroupId');

  const { data: scheduleGroupOptions = [] } = useGetScheduleGroupsQuery({
    select: (scheduleGroups) => {
      const scheduleOptions = scheduleGroups.map(
        ({ id, name, runInterval }) => ({
          value: id,
          label: name,
          runInterval,
        })
      );

      return [
        ...scheduleOptions,
        { value: -2, label: 'Create Interval', runInterval: '' },
      ];
    },
  });

  return (
    <>
      {(!isCreate || isScheduled) && (
        <Form.Group controlId="name">
          <Form.Label className="required-label">Name</Form.Label>
          <Form.Control
            {...register('name', { shouldUnregister: true })}
            isInvalid={!!formState.errors.name}
            data-test-id="jobName"
          />

          <Form.Control.Feedback type="invalid" data-test-id="nameError">
            {formState.errors.name?.message}
          </Form.Control.Feedback>
        </Form.Group>
      )}

      {(!isCreate || isScheduled) && (
        <Form.Group controlId="scheduleGroupId">
          <Form.Label className="required-label">Schedule</Form.Label>
          <Controller
            name="scheduleGroupId"
            control={control}
            shouldUnregister
            render={({
              field: { onChange, value, onBlur, name },
              fieldState: { error, invalid },
            }) => {
              const scheduleGroup = scheduleGroupOptions.find(
                (option) => option.value === value
              );

              return (
                <>
                  <Select
                    options={scheduleGroupOptions}
                    value={scheduleGroup}
                    onChange={(event) => onChange(event?.value)}
                    onBlur={onBlur}
                    name={name}
                    isClearable={isCreate}
                    placeholder="No Scheduling"
                    inputId="scheduleGroupId"
                    id="editJobSchedule"
                    classNamePrefix="react-select"
                    className={invalid ? 'is-invalid' : ''}
                  />
                  {!!scheduleGroup?.runInterval && (
                    <Form.Text className="mt-1">
                      {cronstrue.toString(scheduleGroup.runInterval)} (UTC)
                    </Form.Text>
                  )}
                  <Form.Control.Feedback
                    type="invalid"
                    data-test-id="scheduleError"
                  >
                    {error?.message}
                  </Form.Control.Feedback>
                </>
              );
            }}
          />
        </Form.Group>
      )}

      {scheduleGroupId === -2 && (
        <div className="ml-4">
          <Form.Group controlId="scheduleGroup.name">
            <Form.Label className="required-label">Interval Name</Form.Label>
            <Form.Control
              {...register('scheduleGroup.name', {
                shouldUnregister: true,
              })}
              isInvalid={!!formState.errors.scheduleGroup?.name}
              data-test-id="intervalName"
            />

            <Form.Control.Feedback
              type="invalid"
              data-test-id="intervalNameError"
            >
              {formState.errors.scheduleGroup?.name?.message}
            </Form.Control.Feedback>
          </Form.Group>
          <Form.Group controlId="scheduleGroup.interval">
            <Form.Label className="required-label">
              Interval (in UTC)
            </Form.Label>
            <Controller
              name="scheduleGroup.interval"
              control={control}
              shouldUnregister
              defaultValue="0 * * * *"
              render={({
                field: { onChange, value },
                fieldState: { error, invalid },
              }) => (
                <>
                  <Cron
                    value={value}
                    onChange={onChange}
                    className={invalid ? 'is-invalid' : ''}
                  />

                  <Form.Control.Feedback
                    type="invalid"
                    data-test-id="intervalError"
                  >
                    {error?.message}
                  </Form.Control.Feedback>
                </>
              )}
            />

            <p className={classes.timeIndicator}>
              UTC: {moment.utc(Date.now()).format('YYYY-MM-DD HH:mm')}
              <br />
              Local Time:{' '}
              {moment.utc(Date.now()).local().format('YYYY-MM-DD HH:mm')}
            </p>
          </Form.Group>
        </div>
      )}
    </>
  );
};
