import MinusCircleIcon from '@cloudera-internal/cuix-core/icons/svg/ic-minus-circle.svg';
import PlusIcon from '@cloudera-internal/cuix-core/icons/svg/ic-plus.svg';
import { useMutation } from '@tanstack/react-query';
import { createReport } from 'apis/create-report';
import { getBlankTargetUrl } from 'arc/misc/misc';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { ClouderaLink } from 'components/cloudera-link/ClouderaLink';
import { minimumTimeOptions } from 'pages/job-page/utils/get-minimum-time-interval';
import React from 'react';
import { Col, Form } from 'react-bootstrap';
import { Controller, UseFormReturn, useFieldArray } from 'react-hook-form';
import Select from 'react-select';
import { JobReport, JobThresholdTriggerType } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import { EmailJobFormSchema } from '../email-job-form-schema';

const arc = getGlobalArc();

type EmailJobThresholdTabProps = {
  form: UseFormReturn<EmailJobFormSchema>;
  report?: JobReport;
};

export const EmailJobThresholdTab = ({
  form,
  report,
}: EmailJobThresholdTabProps): JSX.Element => {
  const { control, formState, register, watch, setValue, getValues } = form;

  const { mutateAsync: createReportMutation } = useMutation({
    mutationFn: async () => {
      if (!report?.datasetId) {
        throw new Error('No dataset.');
      }
      return await createReport({
        datasetId: report.datasetId,
        title: `Trigger: ${getValues('name')}`.trim(),
        isThreshold: true,
      });
    },
    onSuccess: (response) => {
      setValue('thresholdTarget.triggerId', response.visualIds[0]);
      arc.misc.openInNewWindow(
        `${arc.frontend.globals.urls.reportBuilderUrl}/${response.dashboardId}`
      );
    },
    onError: () => form.setError('root', { message: 'Error creating report' }),
  });

  const {
    remove: removeCondition,
    append: appendCondition,
    fields: conditions,
  } = useFieldArray({
    control,
    shouldUnregister: true,
    name: 'thresholdTarget.triggerConditions',
  });

  const isCreate = watch('isCreate');
  const triggerId = watch('thresholdTarget.triggerId');

  return (
    <>
      <p>The threshold should be based on the value in a table visual</p>
      <Form.Group controlId="triggerId">
        <Form.Label className="required-label">Trigger Visual ID</Form.Label>
        <div className="d-flex gap-2">
          <Form.Control
            type="number"
            {...register('thresholdTarget.triggerId', {
              shouldUnregister: true,
            })}
            isInvalid={!!formState.errors.thresholdTarget?.triggerId}
          />
          {isCreate && (
            <ClouderaButton
              type="button"
              size="large"
              onClick={() => createReportMutation()}
              testId="createTriggerVisual"
            >
              Create Visual
            </ClouderaButton>
          )}
        </div>

        <Form.Text>
          ID of a table visual: first measure will be assessed.{' '}
          {!!triggerId && (
            <ClouderaLink
              href={getBlankTargetUrl(
                `${arc.frontend.globals.urls.viewDashboardUrl}/${triggerId}`
              )}
              target="_blank"
            >
              See visual
            </ClouderaLink>
          )}
        </Form.Text>

        <Form.Control.Feedback type="invalid" className="d-block">
          {formState.errors.thresholdTarget?.triggerId?.message}
        </Form.Control.Feedback>
      </Form.Group>

      <Form.Group>
        <Form.Label>Send Email Limit</Form.Label>
        <Controller
          name="thresholdTarget.minimumTimeInterval"
          control={control}
          shouldUnregister
          defaultValue={0}
          render={({ field: { onChange, value, onBlur, name } }) => (
            <Select
              options={minimumTimeOptions}
              value={minimumTimeOptions.find(
                (option) => option.value === value
              )}
              onChange={(event) => onChange(event?.value)}
              onBlur={onBlur}
              name={name}
              inputId="thresholdTarget.minimumTimeInterval"
              id="editMinimumTimeInterval"
              placeholder="Send Email Limit"
              aria-label="Send Email Limit"
            />
          )}
        />
      </Form.Group>

      <Form.Group>
        <Form.Label>Trigger Type</Form.Label>
        <div>
          <Form.Check
            inline
            type="radio"
            id="triggerTypeMet"
            label="Threshold met"
            value={JobThresholdTriggerType.ThresholdMet}
            defaultChecked
            {...register('thresholdTarget.triggerType', {
              shouldUnregister: true,
            })}
          />
          <Form.Check
            inline
            type="radio"
            id="triggerTypeChanged"
            label="Value changed since last assessment"
            value={JobThresholdTriggerType.ValueChanged}
            {...register('thresholdTarget.triggerType', {
              shouldUnregister: true,
            })}
          />
        </div>
      </Form.Group>

      <Form.Group>
        <Form.Label>Trigger Conditions</Form.Label>

        {conditions.map((field, index) => (
          <Form.Row key={field.id} className="d-flex align-items-center mb-2">
            <Col xs={2}>
              <Form.Control
                as="select"
                aria-label="Trigger Condition Equality"
                {...register(
                  `thresholdTarget.triggerConditions.${index}.equality`
                )}
              >
                <option value="<">{'<'}</option>
                <option value="<=">≤</option>
                <option value=">">{'>'}</option>
                <option value=">=">≥</option>
                <option value="=">=</option>
              </Form.Control>
            </Col>

            <Col xs={3}>
              <Form.Control
                type="number"
                aria-label="Trigger Condition Value"
                {...register(
                  `thresholdTarget.triggerConditions.${index}.value`
                )}
                isInvalid={
                  !!formState.errors.thresholdTarget?.triggerConditions?.[index]
                    ?.value
                }
              />
            </Col>

            <Col>
              <ClouderaButton
                aria-label="Remove Condition"
                testId={`remove-trigger-condition-button-${index}`}
                type="button"
                icon={<MinusCircleIcon />}
                variant="link"
                color="danger"
                onClick={() => removeCondition(index)}
              />
            </Col>

            <Col xs={12}>
              {!!formState.errors.thresholdTarget?.triggerConditions?.[index]
                ?.value && (
                <Form.Control.Feedback type="invalid" className="d-block">
                  {
                    formState.errors.thresholdTarget.triggerConditions?.[index]
                      ?.value?.message
                  }
                </Form.Control.Feedback>
              )}
            </Col>
          </Form.Row>
        ))}

        {!!formState.errors.thresholdTarget?.triggerConditions?.message && (
          <Form.Control.Feedback type="invalid" className="d-block">
            {formState.errors.thresholdTarget.triggerConditions.message}
          </Form.Control.Feedback>
        )}

        <ClouderaButton
          type="button"
          icon={<PlusIcon />}
          testId="add-trigger-condition-button"
          onClick={() => appendCondition({ equality: '<', value: NaN })}
        >
          Add Condition
        </ClouderaButton>
      </Form.Group>
    </>
  );
};
