import { intervalSchema } from 'pages/job-overview-page/components/interval-modal/interval-schema';
import { JobThresholdTriggerType } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import * as yup from 'yup';

const arc = getGlobalArc();

yup.setLocale({
  string: {
    email: ({ value }) => `${value} is not a valid email`,
  },
});

export const emailJobFormSchema = yup.object({
  isCreate: yup.boolean().required(),
  isScheduled: yup.boolean().required(),
  isThreshold: yup
    .boolean()
    .default(false)
    .when('isScheduled', {
      is: false,
      then: (schema) => schema.oneOf([false]),
    }),
  name: yup
    .string()
    .when(['isCreate', 'isScheduled'], {
      is: (isCreate: boolean, isScheduled: boolean) => !isCreate || isScheduled,
      then: (schema) => schema.required(),
    })
    .label('Name'),
  scheduleGroupId: yup
    .number()
    .when('isScheduled', {
      is: true,
      then: (schema) => schema.required(),
    })
    .transform((value) => (isNaN(value) ? undefined : value))
    .label('Schedule interval'),
  scheduleGroup: intervalSchema
    .notRequired()
    .default(undefined)
    .when('scheduleGroupId', {
      is: -2,
      then: (schema) => schema.required(),
    }),
  emailTarget: yup.object({
    subject: yup.string().required().label('Subject'),
    message: yup.string(),
    to: yup
      .array(yup.string().email().required())
      .min(1)
      .required()
      .label('To'),
    cc: yup.array(yup.string().email().required()).required().label('CC'),
    errorTo: yup.array(yup.string().email().required()).required(),
    replyTo: yup.string().email(),
    templateId: yup
      .number()
      .transform((value) => (isNaN(value) ? undefined : value)),
    urlParams: yup.string().default(''),
    useUrlAlias: yup.boolean().required(),
    alias: yup
      .string()
      .label('Alias')
      .when(['isCreate', 'useUrlAlias'], {
        is: (isCreate: boolean, useUrlAlias: boolean) =>
          isCreate && useUrlAlias,
        then: (schema) => schema.required().min(1),
      }),
    includeEmbeddedImage: yup.boolean().required(),
    includePng: yup.boolean().required(),
    includePdf: yup.boolean().required(),
    pdfOption: yup
      .string()
      .oneOf(['pdf-current', 'pdf-one-viz', 'pdf-all-sheets'])
      .when(['includePdf'], {
        is: true,
        then: (schema) => schema.required().default('pdf-current'),
      }),
    includePageParams: yup.boolean().required(),
    includeXLSX: yup.boolean().required(),
    includeCSV: yup.boolean().required(),
    csvDownloadRows: yup
      .number()
      .max(
        Number(arc.frontend.globals.settings.uploadMaxrowsDefault) || Infinity
      )
      .when(['includeCSV', 'includeXLSX'], {
        is: (includeXLSX: boolean, includeCSV: boolean) =>
          includeXLSX || includeCSV,
        then: (schema) => schema.required().default(5000),
      })
      .label('Limit'),
  }),
  thresholdTarget: yup
    .object({
      triggerId: yup
        .number()
        .typeError('Trigger id is a required field')
        .required()
        .label('Trigger id'),
      triggerType: yup
        .string()
        .oneOf(Object.values(JobThresholdTriggerType))
        .required(),
      triggerConditions: yup
        .array(
          yup.object({
            equality: yup.string().required(),
            value: yup
              .number()
              .required()
              .typeError(
                'The condition value must not be empty. Provide a value or remove the condition.'
              ),
          })
        )
        .required()
        .min(1)
        .label('Trigger Conditions'),
      minimumTimeInterval: yup.number().required(),
    })
    .notRequired()
    .default(undefined),
});

export type EmailJobFormSchema = yup.InferType<typeof emailJobFormSchema>;
