import { useQuery, useQueryClient } from '@tanstack/react-query';
import { getJobStatus } from 'apis/jobs/get-job-status';
import { getJobLogs } from 'apis/jobs/logs/get-logs';
import { QueryKeyBase } from 'apis/query-keys';
import { useCallback, useState } from 'react';
import { useNotificationStore } from 'stores/use-notification-store';
import { JobStatus } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';

const arc = getGlobalArc();

/**
 * Periodically checks the status of the currently queued/running job.
 * Displays the right notification on the dashboard.
 *
 * @param jobId The id of the job to check
 */
export const useEmailNowNotification = () => {
  const [jobId, setJobId] = useState<number>();
  const queryClient = useQueryClient();
  const { addNotification, removeNotification } = useNotificationStore(
    ({ addNotification, removeNotification }) => ({
      addNotification,
      removeNotification,
    })
  );

  useQuery({
    queryKey: [QueryKeyBase.JobStatus, jobId],
    queryFn: async () => {
      const status = await getJobStatus(jobId!);

      if (![JobStatus.Queued, JobStatus.Running].includes(status)) {
        setJobId(undefined);
        removeNotification('sendingEmail');

        if (status === JobStatus.Finished) {
          addNotification(
            { variant: 'success', message: 'Email successfully sent!' },
            { delay: 3000 }
          );
        } else if (status === JobStatus.Error) {
          const result = await getJobLogs(jobId!, { page: 1, pageSize: 1 });
          const taskList = result.data[0].taskList;
          const failedTask = `Failed when ${
            taskList[taskList.length - 1].message
          }`;
          const error = `Details ${result.data[0].errorMessage}`;

          // Place the error into the visual error container and notification center
          arc.report.alertMessage(
            'Failed to send email',
            [failedTask, error].join('<br/>'),
            { is_html: true }
          );

          // Popup notification
          addNotification(
            { variant: 'danger', message: 'Failed to send email' },
            { delay: 3000 }
          );
        }

        queryClient.invalidateQueries([QueryKeyBase.Jobs, jobId]);
        queryClient.invalidateQueries([QueryKeyBase.JobLogs, jobId]);
      }

      return null;
    },
    enabled: !!jobId,
    refetchInterval: 1500,
    staleTime: 1500,
  });

  const startCheckingJobStatus = useCallback(
    (jobId: number) => {
      setJobId(jobId);
      addNotification({
        id: 'sendingEmail',
        variant: 'info',
        title: 'Sending email',
        message: 'Email will be sent in a few seconds...',
      });
    },
    [addNotification]
  );

  return { startCheckingJobStatus };
};
