import classNames from 'classnames';
import React, { ReactElement, useCallback, useMemo, useState } from 'react';
import Modal from 'react-bootstrap/Modal';

import { getGlobalArc } from 'utils/get-global-arc';
import { VizButton } from 'components/viz-button/VizButton';
import { useImportDataModalStore } from 'stores/use-import-data-modal-store';
import { useLegacyBridge } from 'services/legacy-bridge';

import { VizSpinner } from 'components/viz-spinner/VizSpinner';
import { CsvImportTabContent } from '../csv-import-tab-content/CsvImportTabContent';
import { UrlImportTabContent } from '../url-import-tab-content/UrlImportTabContent';

type Tab = {
  id: string;
  image: ReactElement;
  title: string;
  content: ReactElement;
};

const arc = getGlobalArc();

const UPLOAD_MAXROWS_DEFAULT =
  arc.frontend.globals.settings.uploadMaxrowsDefault;
const UPLOAD_MAXROWS_MAX = arc.frontend.globals.settings.uploadMaxrowsMax;

export const ImportDataModal = (): JSX.Element => {
  let { open, connectionId, connectionMaxRowUpload, connectionType, close } =
    useImportDataModalStore(
      ({
        open,
        connectionId,
        close,
        connectionMaxRowUpload,
        connectionType,
      }) => ({
        open,
        connectionId,
        connectionMaxRowUpload,
        connectionType,
        close,
      })
    );

  let maxRowUpload = connectionMaxRowUpload ?? UPLOAD_MAXROWS_DEFAULT;

  if (maxRowUpload >= UPLOAD_MAXROWS_MAX) {
    maxRowUpload = UPLOAD_MAXROWS_MAX;
  }

  const [isValid, setValid] = useState(false);
  const tabs: Tab[] = useMemo(
    () => [
      {
        id: 'csv',
        image: (
          <img
            src={`${arc.frontend.globals.urls.staticUrl}/img/file_csv.png`}
            alt="csv file"
          />
        ),
        title: 'CSV',
        content: (
          <CsvImportTabContent
            connectionMaxRowUpload={maxRowUpload}
            setIsFormValid={(isFormValid) => setValid(isFormValid)}
            connectionType={connectionType ?? 'undefined'}
          />
        ),
      },
      {
        id: 'url',
        image: <i className="fa fa-external-link" />,
        title: 'URL',
        content: (
          <UrlImportTabContent
            setIsFormValid={(isFormValid) => setValid(isFormValid)}
          />
        ),
      },
    ],
    [maxRowUpload, connectionType]
  );

  const [selectedTabIndex, setSelectedTabIndex] = useState(0);
  const [isLoading, setIsLoading] = useState(false);
  const selectedTab = tabs[selectedTabIndex];

  const selectTab = (tabIndex: number): void => {
    setValid(false);
    setSelectedTabIndex(tabIndex);
  };

  const closeModal = () => {
    setValid(false);
    setIsLoading(false);
    selectTab(0);
    close();
  };

  const onConnectionIdChange = useCallback(
    ({
      payload,
    }: {
      payload?: {
        connectionId: number;
        connectionMaxRowUpload: number;
        connectionType: string;
      };
    }) => {
      if (payload) {
        open(payload);
      }
    },
    [open]
  );
  useLegacyBridge('open-import-data-modal', onConnectionIdChange);

  return (
    <Modal show={connectionId !== undefined} onHide={closeModal}>
      <Modal.Header>
        <Modal.Title>Add Data</Modal.Title>
      </Modal.Header>

      <form
        action={arc.frontend.globals.urls.dataImportUrl}
        method="post"
        encType="multipart/form-data"
        onSubmit={() => {
          setIsLoading(true);
        }}
      >
        <Modal.Body>
          <div className="d-flex">
            {tabs.map((tab, index) => (
              <div
                key={tab.id}
                className="form-group source-type"
                onClick={() => selectTab(index)}
              >
                <div
                  className={classNames('data-type-selector image-wrapper', {
                    'type-selected': tab === selectedTab,
                  })}
                >
                  <div className="image-container">{tab.image}</div>
                  <div className="type-title">{tab.title}</div>
                </div>
              </div>
            ))}
          </div>

          {selectedTab.content}

          {connectionId !== undefined && (
            <input
              type="hidden"
              name="datasource_active_dataconnid"
              value={connectionId.toString()}
            />
          )}

          <input
            name="csrfmiddlewaretoken"
            type="hidden"
            value={arc.frontend.globals.authentication.csrfToken}
          />
        </Modal.Body>

        <Modal.Footer>
          <VizButton color="secondary" size="medium" onClick={closeModal}>
            Close
          </VizButton>

          <VizButton
            color="primary"
            size="medium"
            disabled={!isValid || isLoading}
            type="submit"
          >
            {isLoading && <VizSpinner />}
            Get data
          </VizButton>
        </Modal.Footer>
      </form>
    </Modal>
  );
};
