import React, { useCallback, useEffect, useState } from 'react';
import Nav from 'react-bootstrap/Nav';
import Navbar from 'react-bootstrap/Navbar';
import ClouderaLogo from 'assets/logos/cloudera-logo-white.svg';

import { getGlobalArc } from 'utils/get-global-arc';

import classes from './NavigationBar.module.scss';
import { SearchBar } from '../search-bar/SearchBar';
import { SettingsMenu } from '../settings-menu/SettingsMenu';
import { HelpMenu } from '../help-menu/HelpMenu';
import { WhatSNewModal } from '../what-s-new-modal/WhatSNewModal';
import { UserMenu } from '../user-menu/UserMenu';
import { NaturalLanguageSearch } from '../natural-language-search/NaturalLanguageSearch';
import { Notifications } from '../notifications/Notifications';
import { NotifcationCenter } from '../notification-center/NotifcationCenter';
import { NotificationCenterModal } from '../notification-center-modal/NotificationCenterModal';
import { getNotifications } from 'apis/get-notifications';
import { NotificationInfo } from '../../types/notification-info.type';
import { readNotifications } from 'apis/read-notification';
import { VizNavLink } from '../viz-nav-link/VizNavLink';
import { pinNotification } from 'apis/pin-notification';
import { unPinNotification } from 'apis/unpin-notification';
import classNames from 'classnames';
import { uniqueId } from 'lodash-es';
import { logger } from 'services/logger';
import { legacyBridge } from 'services/legacy-bridge';

type NavigationBarProps = {
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
  customClass?: string;
  defaultSearchTerm?: string;
  onSearch?: (searchTerm: string) => void;
  activeTabs:
    | {
        main: string | null;
        sub?: string;
      }
    | undefined;
};

const arc = getGlobalArc();

const isExploreStyleEnabled = arc.frontend.featureFlipper.getValueById(
  'enable_explore_style'
);

const isC1CSubModeEnabled = arc.frontend.globals.settings.cdvSubMode === 'C1C';

const isClouderaStyleEnabled = arc.frontend.featureFlipper.getValueById(
  'enable_cloudera_style'
);

const isDirectAccessEnabled = arc.frontend.featureFlipper.getValueById(
  'enable_direct_access'
);

const isSiteSearchMenuEnabled = arc.frontend.featureFlipper.getValueById(
  'site_enable_search_menu'
);

const isSiteSearchEnabled =
  arc.frontend.featureFlipper.getValueById('site_enable_search');

const hasSQLTab =
  isClouderaStyleEnabled &&
  isDirectAccessEnabled &&
  arc.frontend.globals.canUserExploreAnyDataconnections;

const routes: {
  id: string;
  testId: string;
  text: string;
  path: string;
  hash?: string;
}[] = [
  {
    id: 'home',
    testId: 'topnav-home',
    text: 'Home',
    path: arc.frontend.globals.urls.homeUrl,
  },
  {
    id: 'visuals',
    testId: 'topnav-visuals',
    text: 'Visuals',
    path: arc.frontend.globals.urls.appsUrl,
  },
  {
    id: 'data',
    testId: 'topnav-data',
    text: isExploreStyleEnabled ? 'Datasets' : 'Data',
    path: arc.frontend.globals.urls.dataUrl,
  },
];

if (hasSQLTab) {
  routes.splice(1, 0, {
    id: 'sql',
    testId: 'topnav-sql',
    text: 'SQL',
    path: arc.frontend.globals.urls.sqlUrl,
    hash: 'sql-tab',
  });
}

const pagesWithSearchBar = [
  arc.frontend.globals.urls.sqlUrl,
  arc.frontend.globals.urls.dataUrl,
];

export const NavigationBar = ({
  isAlreadySeen,
  shouldShowModal,
  customClass,
  defaultSearchTerm,
  onSearch,
  activeTabs,
}: NavigationBarProps): JSX.Element | null => {
  const [isNotificationCenterOpen, setNotificationCenterIsOpen] =
    useState(false);

  const [notifications, setNotifications] = useState<NotificationInfo[]>([]);

  const [isLoading, setIsLoading] = useState(false);

  const [activeTab, setActiveTab] = useState<string | null>(
    activeTabs?.main ?? null
  );

  const fetchNotifications = useCallback(() => {
    setIsLoading(true);
    getNotifications()
      .then((v) => {
        setNotifications((current) => {
          const frontendErrors = current.filter(
            ({ source }) => source === 'frontend'
          );
          return [...frontendErrors, ...v].sort((a, b) => {
            const [bTimestamp] = b.timestamps.timestamps;
            const [aTimestamp] = a.timestamps.timestamps;
            return bTimestamp - aTimestamp;
          });
        });
        setIsLoading(false);
      })
      .catch((error) => {
        console.error(error);
        setIsLoading(false);
      });
  }, []);

  useEffect(fetchNotifications, [fetchNotifications]);

  useEffect(() => {
    const subscription = logger.logs$.subscribe(({ msg, stackTrace }) => {
      const timestamp = Date.now();
      const msgString = msg ? `: ${String(msg)}` : '';
      const title = `Frontend Error${msgString}`;
      setNotifications((current) => {
        const duplicate = current.find((notification) => {
          const hasSameTitle = notification.title === title;
          const hasSameBody = notification.body === stackTrace;
          return hasSameTitle && hasSameBody;
        });

        if (duplicate) {
          duplicate.timestamps.timestamps.unshift(timestamp);
          return [...current];
        } else {
          const frontendErrorNotification: NotificationInfo = {
            id: uniqueId(),
            type: 'error',
            title,
            body: stackTrace,
            status: 'U',
            timestamps: { timestamps: [timestamp] },
            source: 'frontend',
            hasErrorSync: false,
          };

          return [frontendErrorNotification, ...current];
        }
      });
    });

    return () => {
      subscription.unsubscribe();
    };
  }, []);

  useEffect(() => {
    if (activeTabs === undefined) {
      const subscription = legacyBridge.navigationBarEvents.subscribe(
        ({ type, payload }) => {
          if (type === 'set-active-navigation-tab') {
            setActiveTab(payload.activeTab);
          }
        }
      );

      return () => subscription.unsubscribe();
    }
  }, [activeTabs]);

  const togglePin = async (notificationId: string) => {
    await Promise.allSettled(
      notifications.map(async (notification) => {
        if (notificationId === notification.id) {
          if (notification.status === 'P') {
            if (notification.source === 'backend') {
              try {
                await unPinNotification(notification.id);
                notification = {
                  ...notification,
                  status: 'R',
                  hasErrorSync: false,
                };
              } catch (error) {
                console.log(error);
                notification = { ...notification, hasErrorSync: true };
              }
            } else {
              notification = { ...notification, status: 'R' };
            }
          } else {
            if (notification.source === 'backend') {
              try {
                await pinNotification(notification.id);
                notification = {
                  ...notification,
                  status: 'P',
                  hasErrorSync: false,
                };
              } catch (error) {
                console.log(error);
                notification = { ...notification, hasErrorSync: true };
              }
            } else {
              notification = { ...notification, status: 'P' };
            }
          }
        }
        return notification;
      })
    ).then((responses) => {
      const fulfilled = responses.filter(
        (resp) => resp.status === 'fulfilled'
      ) as PromiseFulfilledResult<NotificationInfo>[];
      const updatedNotifications = fulfilled.map((resp) => resp.value);
      setNotifications(updatedNotifications);
    });
  };

  const [shouldFilterUnread, setShouldFilterUnread] = useState(true);
  const [shouldFilterPinned, setShouldFilterPinned] = useState(false);

  let visibleNotifications = notifications;

  if (shouldFilterUnread) {
    visibleNotifications = notifications.filter(({ status }) => status === 'U');
  } else if (shouldFilterPinned) {
    visibleNotifications = notifications.filter(({ status }) => status === 'P');
  }

  const toggleFilterUnread = () => {
    setShouldFilterUnread((curr) => !curr);
  };

  const toggleFilterPinned = () => {
    setShouldFilterPinned((curr) => !curr);
  };

  if (
    isC1CSubModeEnabled ||
    new URLSearchParams(window.location.search).get('embed') === 'true'
  ) {
    return null;
  }

  return (
    <>
      <Navbar
        className={classNames(
          classes.navbar,
          isClouderaStyleEnabled ? classes.clouderaStyle : ''
        )}
        style={{ flexShrink: 0 }}
      />

      <Navbar
        className={classNames(
          classes.navbar,
          isClouderaStyleEnabled ? classes.clouderaStyle : '',
          customClass ?? ''
        )}
        expand="md"
        fixed="top"
        variant="dark"
      >
        {!isExploreStyleEnabled && (
          <Navbar.Brand
            className={classes.navbarBrand}
            href={arc.frontend.globals.urls.homeUrl}
          >
            <ClouderaLogo
              className={classes.logo}
              height={100}
              width={100}
              viewBox="164 0 100 100"
            />
          </Navbar.Brand>
        )}

        {!arc.frontend.globals.isMobileView && (
          <Navbar.Toggle className={classes.navbarToggle}>
            <i className={classNames(classes.toggleIcon, 'fa fa-bars')} />
          </Navbar.Toggle>
        )}

        <Navbar.Collapse className={classes.navbarCollapse}>
          <Nav className={classes.nav}>
            {routes.map(({ id, testId, text, path, hash }) => (
              <Nav.Link
                id={`topnav-${id}`}
                as={VizNavLink}
                key={id}
                eventKey={id}
                data-test-id={testId}
                href={hash ? `${path}#${hash}` : path}
                active={activeTab === id}
              >
                {text}
              </Nav.Link>
            ))}
          </Nav>

          {pagesWithSearchBar.some((path) =>
            window.location.pathname.startsWith(path)
          ) ? (
            <SearchBar defaultValue={defaultSearchTerm} onChange={onSearch} />
          ) : (
            <div className="mx-auto" />
          )}

          {isSiteSearchMenuEnabled && isSiteSearchEnabled && (
            <NaturalLanguageSearch />
          )}

          <NotifcationCenter
            openNotificationCenterModal={() =>
              setNotificationCenterIsOpen(true)
            }
            notifications={notifications}
          />

          <SettingsMenu
            isActive={activeTab === 'settings'}
            activeTab={activeTabs?.sub}
          />

          <HelpMenu />

          <UserMenu />
        </Navbar.Collapse>
      </Navbar>

      <WhatSNewModal
        isAlreadySeen={isAlreadySeen}
        shouldShowModal={shouldShowModal}
      />

      <NotificationCenterModal
        close={() => {
          setNotificationCenterIsOpen(false);

          const unreadNotificationIds = notifications
            .filter(
              ({ status, source }) => source === 'backend' && status === 'U'
            )
            .map(({ id }) => id);
          readNotifications(unreadNotificationIds);

          setNotifications((current) =>
            current.map((notification) =>
              notification.status === 'U'
                ? { ...notification, status: 'R' }
                : notification
            )
          );
        }}
        isOpen={isNotificationCenterOpen}
        notifications={visibleNotifications}
        toggleFilterUnread={toggleFilterUnread}
        isFiltered={shouldFilterUnread}
        toggleFilterPinned={toggleFilterPinned}
        isFilteredPinned={shouldFilterPinned}
        fetchNotifications={fetchNotifications}
        isLoading={isLoading}
        togglePin={togglePin}
      />

      <Notifications />
    </>
  );
};
