import React, { useState } from 'react';
import classNames from 'classnames';
import Modal from 'react-bootstrap/Modal';
import Nav from 'react-bootstrap/Nav';
import NavLink from 'react-bootstrap/NavLink';

import { VizButton } from 'components/viz-button/VizButton';

import classes from './NotificationCenterModal.module.scss';
import { NotificationInfo } from 'types/notification-info.type';
import { NotificationItem } from '../notification-item/NotificationItem';
import { InView } from 'react-intersection-observer';

type NotificationCenterModalProps = {
  isOpen: boolean;
  close: () => void;
  notifications: NotificationInfo[];
  toggleFilterUnread: () => void;
  isFiltered: boolean;
  toggleFilterPinned: () => void;
  isFilteredPinned: boolean;
  fetchNotifications: () => void;
  isLoading: boolean;
  togglePin: (id: string) => void;
};

export const NotificationCenterModal = ({
  isOpen,
  close,
  notifications,
  toggleFilterUnread,
  isFiltered,
  toggleFilterPinned,
  isFilteredPinned,
  fetchNotifications,
  isLoading,
  togglePin,
}: NotificationCenterModalProps) => {
  const [selectedTab, setSelectedTab] = useState<string>('all');

  const navItems = [
    {
      title: 'All',
      key: 'all',
      content: (
        <div>
          {notifications.map((notification) => (
            <InView key={notification.id}>
              {({ ref, inView }) => (
                <div ref={ref} className={classes.notificationItem}>
                  {inView && (
                    <NotificationItem
                      notification={notification}
                      togglePin={togglePin}
                    />
                  )}
                </div>
              )}
            </InView>
          ))}

          {notifications.length === 0 && (
            <div className={classes.noNotifications}>
              You currently don't have any{' '}
              {isFiltered ? <strong>unread</strong> : ''} notifications.
            </div>
          )}
        </div>
      ),
    },
  ];
  const selectedNavItem = navItems.find(
    (navItem) => navItem.key === selectedTab
  )!;

  const refreshOrLoadingIcon = isLoading ? (
    <i className="fa fa-spinner fa-spin loading-icon" />
  ) : (
    <i className="fa fa-refresh" aria-hidden="true" />
  );

  return (
    <Modal show={isOpen} onHide={close} scrollable={true} size="lg">
      <Modal.Header closeButton>
        <Modal.Title>Notification Center</Modal.Title>
      </Modal.Header>
      <div className={classes.optionsContainer}>
        <span style={{ fontSize: '1.2em' }}> Filters: </span>
        <div
          className={classes.iconWrapper}
          data-toggle={'tooltip'}
          title={
            isFiltered
              ? 'Show all notifications'
              : 'Only show unread notifications'
          }
          onClick={() => {
            if (isFilteredPinned) {
              toggleFilterPinned();
            }
            toggleFilterUnread();
          }}
        >
          <i
            className={classNames(
              classes.icon,
              { [classes.pinned]: isFiltered },
              'fa-inbox',
              'fa'
            )}
          />
        </div>
        <div
          className={classes.iconWrapper}
          data-toggle={'tooltip'}
          title={
            isFilteredPinned
              ? 'Show all notifications'
              : 'Only show pinned notifications'
          }
          onClick={() => {
            if (isFiltered) {
              toggleFilterUnread();
            }
            toggleFilterPinned();
          }}
        >
          <i
            className={classNames(
              classes.icon,
              { [classes.pinned]: isFilteredPinned },
              'fa-thumb-tack',
              'fa'
            )}
          />
        </div>
      </div>

      <Modal.Body>
        <Nav>
          {navItems.map(({ title, key }) => (
            <Nav.Item key={key}>
              <NavLink
                className={classes.tab}
                eventKey={key}
                active={key === selectedTab}
                onClick={() => setSelectedTab(key)}
              >
                {title}
              </NavLink>
            </Nav.Item>
          ))}
        </Nav>
        {selectedNavItem.content}
      </Modal.Body>
      <Modal.Footer className="justify-content-between">
        <VizButton color="secondary" size="medium" onClick={fetchNotifications}>
          {refreshOrLoadingIcon}
        </VizButton>

        <VizButton size="medium" color="primary" onClick={close}>
          Close
        </VizButton>
      </Modal.Footer>
    </Modal>
  );
};
