import classNames from 'classnames';
import { VizWrapperLink } from 'components/viz-wrapper-link/VizWrapperLink';
import { NotificationInfo } from 'types/notification-info.type';
import React, { useState } from 'react';
import classes from './NotificationItem.module.scss';
import moment from 'moment';
import Badge from 'react-bootstrap/Badge';
import { VizTooltip } from 'components/VizTooltip';
import Popover from 'react-bootstrap/Popover';

type NotificationItemProps = {
  notification: NotificationInfo;
  togglePin: (id: string) => void;
};

export const NotificationItem = ({
  notification: { timestamps, id, status, title, body, type, hasErrorSync },
  togglePin,
}: NotificationItemProps): JSX.Element => {
  const charLimit = 100;
  const hasTooManyChars = body.length > charLimit;

  const [isOpen, setIsOpen] = useState(false);

  const [isHovering, setIsHovering] = useState(false);

  const handleMouseOver = () => {
    setIsHovering(true);
  };

  const handleMouseOut = () => {
    setIsHovering(false);
  };

  const [latestTime] = timestamps.timestamps;

  let displayText = body;
  if (hasTooManyChars && !isOpen) {
    displayText = `${displayText.slice(0, charLimit).split('\n')[0]}...`;
  }

  let iconClass = '';

  if (type === 'warning') {
    iconClass = 'fa-exclamation-triangle';
  }

  if (type === 'info') {
    iconClass = 'fa-info-circle';
  }

  if (type === 'error') {
    iconClass = 'fa-times-circle';
  }

  let statusClass;

  if (status === 'R') {
    statusClass = classes.read;
  }

  if (status === 'P') {
    statusClass = classes.pinned;
  }

  return (
    <div
      className={classNames(classes.notification, statusClass)}
      key={id}
      onMouseOver={handleMouseOver}
      onMouseOut={handleMouseOut}
    >
      <div>
        <div className="d-flex">
          <div className={classes.iconWrapperType}>
            <i
              className={classNames(
                classes.iconType,
                { [classes.read]: status === 'R' },
                { [classes.pinned]: status === 'P' },
                classes[type],
                iconClass,
                'fa'
              )}
            />
          </div>
          <div>
            <div className={classes.date}>
              {moment(latestTime).format('LLL')}
              {timestamps.timestamps.length > 1 && (
                <VizTooltip
                  placement="bottom"
                  trigger="focus"
                  overlay={
                    <Popover id={id}>
                      <Popover.Content>
                        <div style={{ maxHeight: '200px', overflowY: 'auto' }}>
                          {timestamps.timestamps.map((ts) => (
                            <p key={ts}> {moment(ts).format('LLL')}</p>
                          ))}
                        </div>
                      </Popover.Content>
                    </Popover>
                  }
                >
                  <Badge
                    pill
                    variant="secondary"
                    style={{ marginLeft: '10px', borderWidth: '0' }}
                    as="button"
                  >
                    {' '}
                    {timestamps.timestamps.length}{' '}
                  </Badge>
                </VizTooltip>
              )}
            </div>
            <div className={classes.title}>{title}</div>
            <div className={classes.body}>
              <div className={classes.codeText}>{displayText}</div>

              {hasTooManyChars && (
                <VizWrapperLink
                  className="ml-2"
                  style={{ color: '#34a1eb' }}
                  onClick={() => {
                    setIsOpen((current) => !current);
                  }}
                >
                  {isOpen ? 'read less' : 'read more'}
                </VizWrapperLink>
              )}
            </div>
          </div>
        </div>
      </div>
      <div
        className="d-flex"
        data-toggle={'tooltip'}
        title={'Error in pin/unpin the notification'}
      >
        {hasErrorSync && (
          <div className={classes.iconWrapper}>
            <i
              className={classNames(
                classes.errorIcon,
                'fa-exclamation-circle',
                'fa'
              )}
            />
          </div>
        )}
        {(isHovering || status === 'P') && (
          <div
            className={classes.iconWrapper}
            data-toggle={'tooltip'}
            title={status !== 'P' ? 'Click to pin' : 'Click to unpin'}
            onClick={() => togglePin(id)}
          >
            <i
              className={classNames(
                classes.icon,
                { [classes.pinned]: status === 'P' },
                'fa-thumb-tack',
                iconClass,
                'fa'
              )}
            />
          </div>
        )}
      </div>
    </div>
  );
};
