import React, { useState } from 'react';
import Alert from 'react-bootstrap/Alert';

import ErrorIcon from '@cloudera-internal/cuix-core/icons/svg/ic-error.svg';
import WarningIcon from '@cloudera-internal/cuix-core/icons/svg/ic-warning.svg';
import SuccessIcon from '@cloudera-internal/cuix-core/icons/svg/ic-success.svg';
import InfoIcon from '@cloudera-internal/cuix-core/icons/svg/ic-info.svg';
import Close from '@cloudera-internal/cuix-core/icons/svg/ic-close.svg';
import ExpandIcon from '@cloudera-internal/cuix-core/icons/svg/ic-accordion-expand.svg';
import CollapseIcon from '@cloudera-internal/cuix-core/icons/svg/ic-collapse.svg';
import {
  cdlRed500,
  cdlBlue500,
  cdlGreen500,
  cdlOrange500,
} from '@cloudera-internal/cuix-core/variables';

import type { AlertVariant } from '../../types/alert-variant.type';
import classes from './Notification.module.scss';
import classNames from 'classnames';
import { Button } from 'react-bootstrap';

const variantToIcon = {
  danger: ErrorIcon,
  info: InfoIcon,
  success: SuccessIcon,
  warning: WarningIcon,
};

const variantToColor = {
  danger: cdlRed500,
  info: cdlBlue500,
  success: cdlGreen500,
  warning: cdlOrange500,
};

type NotificationProps = {
  message: string;
  variant: AlertVariant;
  title?: string;
  useHtml?: boolean;
  show?: boolean;
  onClose?: () => void;
  dismissible?: boolean;
  expandedMessage?: string;
};

/**
 * Displays a notification based on the different parameters.
 * It can be used for inline and popup (through addNotification() from the notification store) notifications.
 * @param message The alert message.
 * @param variant The type of alert: danger | warning | info | success.
 * @param title It is optional, regular inline alerts do not have one.
 * @param useHtml Determines wether the error message contains links and styling.
 * @param show Controls the visibility of the alert.
 * @param onClose Contains the actions needed for the dismissal of the notification.
 * @param dismissible Determines wether the notification should have a close button.
 * @param expandedMessage Contains some additional information, usually in case of errors. If present, the alert will have a 'More...' button.
 */
export const Notification = ({
  message,
  variant,
  title,
  useHtml = false,
  show = true,
  dismissible = false,
  onClose,
  expandedMessage,
}: NotificationProps): JSX.Element => {
  const [isExpanded, setIsExpanded] = useState(false);
  const Icon = variantToIcon[variant];

  return (
    <Alert
      variant={variant}
      show={show}
      className={classNames(classes.notification, !!title && classes.large)}
    >
      <Icon fontSize={title ? 24 : 20} color={variantToColor[variant]} />
      <div className={classes.content} data-test-id="notification-content">
        {title && (
          <p className={classes.title} data-test-id="notification-title">
            {title}
          </p>
        )}
        {useHtml ? (
          <p
            dangerouslySetInnerHTML={{ __html: message }}
            data-test-id="notification-message"
          />
        ) : (
          <p data-test-id="notification-message">{message}</p>
        )}

        {expandedMessage && (
          <Button
            variant="link"
            className={classes.linkButton}
            onClick={() => setIsExpanded((value) => !value)}
          >
            {isExpanded ? (
              <>
                Less Details
                <CollapseIcon className="ml-1" />
              </>
            ) : (
              <>
                More Details
                <ExpandIcon className="ml-1" />
              </>
            )}
          </Button>
        )}

        {isExpanded && expandedMessage && (
          <>
            <hr />
            {useHtml ? (
              <p dangerouslySetInnerHTML={{ __html: expandedMessage }} />
            ) : (
              <p>{expandedMessage}</p>
            )}
          </>
        )}
      </div>

      {dismissible && (
        <button
          className={classes.closeButton}
          type="button"
          aria-label="Close"
          onClick={onClose}
          data-test-id="notification-close"
        >
          <Close />
        </button>
      )}
    </Alert>
  );
};
