import React, { useCallback } from 'react';

import type { AlertVariant } from '../../types/alert-variant.type';
import { Notification } from 'components/notification/Notification';
import { useNotificationStore } from 'stores/use-notification-store';
import classes from './Notifications.module.scss';
import { useLegacyBridge } from 'services/legacy-bridge';

const variantToTitle: Record<AlertVariant, string> = {
  danger: 'Error',
  info: 'Info',
  success: 'Success',
  warning: 'Warning',
};

export const Notifications = (): JSX.Element => {
  const {
    notifications,
    addNotification,
    removeNotification,
    removeAllNotifications,
  } = useNotificationStore(
    ({
      notifications,
      addNotification,
      removeNotification,
      removeAllNotifications,
    }) => ({
      notifications,
      addNotification,
      removeNotification,
      removeAllNotifications,
    })
  );

  /**
   * Wrapper for addNotification(): enables adding notifications from legacy code
   */
  const addNotificationFromLegacy = useCallback(
    ({
      payload,
    }: {
      payload?: {
        id?: string;
        variant: AlertVariant;
        message: string;
        title?: string;
        delay?: number;
        useHtml?: boolean;
        onClose?: () => void;
        expandedMessage?: string;
        stackMessages?: boolean;
      };
    }) => {
      if (payload) {
        const {
          id,
          variant,
          message,
          title,
          delay = 4000,
          useHtml = false,
          onClose,
          stackMessages = false,
        } = payload;

        addNotification(
          { id, variant, message, title, useHtml, onClose },
          { delay, stackMessages }
        );
      }
    },
    [addNotification]
  );

  useLegacyBridge('add-notification', addNotificationFromLegacy);

  /**
   * Enables removing notifications from legacy code
   * If no title was specified, it clears the notification container
   * In this context, the title itself is unique
   */
  const hideNotificationFromLegacy = useCallback(
    ({ payload }: { payload?: { id?: string } }) => {
      if (payload?.id) {
        removeNotification(payload.id);
      } else {
        removeAllNotifications();
      }
    },
    [removeAllNotifications, removeNotification]
  );

  useLegacyBridge('remove-notification', hideNotificationFromLegacy);

  return (
    <div className={classes.notifications}>
      {notifications.map(
        ({
          id,
          message,
          title,
          variant,
          useHtml,
          onClose,
          expandedMessage,
        }) => (
          <Notification
            key={id}
            message={message}
            variant={variant}
            title={title ?? variantToTitle[variant]}
            useHtml={useHtml}
            onClose={() => {
              removeNotification(id);
              onClose?.();
            }}
            expandedMessage={expandedMessage}
            dismissible
          />
        )
      )}
    </div>
  );
};
