import { useQuery } from '@tanstack/react-query';
import {
  dataconnectionQueryToken,
  getAllDataConnections,
} from 'apis/get-all-data-connections';
import { getQueryLogs, getQueryLogsQueryToken } from 'apis/get-query-logs';
import { ColumnConfig } from 'components/activity-logs/components/ColumnConfig';
import React, { useCallback, useMemo, useState } from 'react';
import { useLegacyBridge } from 'services/legacy-bridge';
import { useNotificationStore } from 'stores/use-notification-store';
import { frontendDownloadCSV } from 'utils/frontend-download-csv';
import { getGlobalArc } from 'utils/get-global-arc';
import {
  getCoreRowModel,
  getPaginationRowModel,
  getSortedRowModel,
} from '@tanstack/react-table';
import { ClouderaTable } from 'components/cloudera-table';
import { VizButton } from 'components/viz-button/VizButton';
import { generateCsvData } from 'components/activity-logs/utils/activity-logs-csv-data-generator';
import { Form } from 'react-bootstrap';
import classes from '../activity-logs/ActivityLogs.module.scss';
import { useTable } from 'hooks/use-table';

const arc = getGlobalArc();
const isArcengineEnabled =
  arc.frontend.globals.settings.isArcengineDataconnectionEnable;
const isSiteEnabledExtracts = arc.frontend.featureFlipper.getValueById(
  'site_enable_extracts'
);
const DEFAULT_QUERY_COUNT = '10';

export const PerfUsageTable = () => {
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));
  const [dashboardId, setDashboardId] = useState<number | undefined>(undefined);
  const [visiblePopoverId, setVisiblePopoverId] = useState('');
  const [queryCount, setQueryCount] = useState(DEFAULT_QUERY_COUNT);
  const [debouncedQueryCount, setDebouncedQueryCount] =
    useState(DEFAULT_QUERY_COUNT);
  const [debouncedQueryCountTimeout, setDebouncedQueryCountTimeout] =
    useState<NodeJS.Timeout | null>(null);

  const setDebouncedQueryCountAfterTimeout = useCallback(
    (value: string) => {
      if (debouncedQueryCountTimeout) clearTimeout(debouncedQueryCountTimeout);

      const timeout = setTimeout(() => setDebouncedQueryCount(value), 500);

      setDebouncedQueryCountTimeout(timeout);
    },
    [debouncedQueryCountTimeout]
  );

  useLegacyBridge(
    'open-perf-usage-table',
    ({ payload }: { payload?: { dashboardId?: number } }) => {
      if (!payload) return;
      if (payload.dashboardId !== undefined) {
        setDashboardId(payload.dashboardId);

        if (payload.dashboardId === dashboardId) {
          refetch();
        }
      }
    }
  );

  const {
    data: queryLogs,
    isFetching: isQueryLogsLoading,
    refetch,
  } = useQuery({
    queryKey: [getQueryLogsQueryToken, debouncedQueryCount],
    queryFn: () => getQueryLogs(debouncedQueryCount, dashboardId),
    initialData: [],
    enabled: !!dashboardId,
    onError: () =>
      addNotification(
        {
          id: 'perf-usage-query-logs-danger',
          variant: 'danger',
          message: 'Failed to load activity logs',
        },
        { delay: 3000 }
      ),
  });

  const { data: allDataConnections } = useQuery({
    queryKey: [dataconnectionQueryToken],
    queryFn: () => getAllDataConnections(),
    enabled: !!dashboardId,
    onError: () => {
      addNotification(
        {
          id: 'perf-usage-connection-danger',
          variant: 'danger',
          message: 'Error fetching dataconnections.',
        },
        { delay: 3000 }
      );
    },
    initialData: [],
    select: (dataconnections) => {
      return dataconnections.map(({ id, name, type, typeName }) => ({
        id,
        name,
        type,
        typeName,
      }));
    },
  });

  const handleTogglePopover = useCallback((id: string) => {
    setVisiblePopoverId(id);
  }, []);

  const columns = useMemo(() => {
    return ColumnConfig(
      visiblePopoverId,
      handleTogglePopover,
      allDataConnections
    );
  }, [visiblePopoverId, handleTogglePopover, allDataConnections]);

  const table = useTable({
    data: queryLogs,
    columns,
    getCoreRowModel: getCoreRowModel(),
    getSortedRowModel: getSortedRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    state: {
      columnVisibility: {
        dataconnection: false,
        error: false,
        error_detail: false,
        dataconnection_fbid: isArcengineEnabled,
        cache_check: isArcengineEnabled,
        rewrite_target: isSiteEnabledExtracts || isArcengineEnabled,
      },
    },
  });

  const csvData = generateCsvData(queryLogs || []);

  const handleQueryCountChange = (
    event: React.ChangeEvent<HTMLInputElement>
  ) => {
    const newCount = event.target.value;
    setQueryCount(newCount);
    setDebouncedQueryCountAfterTimeout(newCount);
  };

  return (
    <>
      <div className="mb-3">
        <VizButton
          size="small"
          color="secondary"
          onClick={() =>
            frontendDownloadCSV(
              csvData,
              `Historical Logs ${new Date()
                .toISOString()
                .split('T')[0]
                .replace(/-/g, '')}`
            )
          }
        >
          DOWNLOAD USAGE LOGS
        </VizButton>
      </div>
      <div className="mb-3">
        <Form.Group
          controlId="fetchedQueriesCount"
          className="d-flex align-items-center mb-0"
        >
          <Form.Label className="me-2 mb-0 mr-1">Queries:</Form.Label>
          <VizButton
            size="small"
            color="secondary"
            onClick={() =>
              setQueryCount((prev) => {
                const newCount = String(Math.max(Number(prev) - 1, 1));
                setDebouncedQueryCountAfterTimeout(newCount);
                return newCount;
              })
            }
            className="py-1 px-2"
          >
            <i className="fa fa-minus"></i>
          </VizButton>

          <Form.Control
            type="text"
            value={queryCount}
            onChange={handleQueryCountChange}
            className={classes.queryCountInput}
          />

          <VizButton
            size="small"
            color="secondary"
            onClick={() =>
              setQueryCount((prev) => {
                const newCount = String(Number(prev) + 1);
                setDebouncedQueryCountAfterTimeout(newCount);
                return newCount;
              })
            }
            className="py-1 px-2"
          >
            <i className="fa fa-plus"></i>
          </VizButton>
        </Form.Group>
      </div>
      <ClouderaTable
        isLoading={isQueryLogsLoading}
        table={table}
        className="perf-usage-table"
        data-test-id="perfUsageTable"
      />
    </>
  );
};
