import moment from 'moment';
import React, { Fragment } from 'react';
import OverlayTrigger from 'react-bootstrap/OverlayTrigger';
import Popover from 'react-bootstrap/Popover';

import { getGlobalArc } from 'utils/get-global-arc';
import { DataExtractBadge } from 'components/data-extract-badge/DataExtractBadge';

import classes from './ReportTilePopover.module.scss';
import classNames from 'classnames';
import { DashboardDatasetInfo } from '../../pages/home/types/dashboard-dataset-info.type';

type ReportTilePopoverProps = {
  report: {
    id: number;
    title: string;
    updated: {
      date: string;
      by: string;
    };
    workspace: string;
  };
  datasetInfo?: DashboardDatasetInfo;
  sharedVisualDashboardIds?: string[];
  children: JSX.Element;
};

const arc = getGlobalArc();

const isSiteEnabledExtracts = arc.frontend.featureFlipper.getValueById(
  'site_enable_extracts'
);

const shouldHideWorkspaces =
  arc.frontend.featureFlipper.getValueById('hide_workspaces');

export const ReportTilePopover = ({
  report,
  datasetInfo,
  sharedVisualDashboardIds,
  children,
}: ReportTilePopoverProps): JSX.Element => {
  return (
    <OverlayTrigger
      trigger={['hover', 'focus']}
      placement="top"
      overlay={(props) => {
        const { id, title, updated, workspace } = report;

        const values: { key: string; value: JSX.Element | string | number }[] =
          [
            { key: 'ID', value: id },
            { key: 'Last updated', value: moment(updated.date).fromNow() },
            { key: 'Modified by', value: updated.by || ' ' },
          ];

        if (datasetInfo) {
          values.push({ key: 'Dataset', value: datasetInfo.name });

          if (isSiteEnabledExtracts) {
            const extractTags: JSX.Element[] = [];

            if (datasetInfo?.extract?.isSource) {
              extractTags.push(<DataExtractBadge type="source" />);
            }

            if (datasetInfo?.extract?.isTarget) {
              extractTags.push(<DataExtractBadge type="target" />);
            }

            if (extractTags.length > 0) {
              values.push({ key: 'Dataset Tags', value: <>{extractTags}</> });
            }
          }
        }

        if (!shouldHideWorkspaces) {
          values.push({ key: 'Workspace', value: workspace });
        }

        if (sharedVisualDashboardIds?.length) {
          values.push({
            key: `Used in Dashboards (${sharedVisualDashboardIds?.length})`,
            value: (
              <>
                {sharedVisualDashboardIds.map((id, index) => (
                  <Fragment key={id}>
                    <a
                      href={`${arc.frontend.globals.urls.viewDashboardUrl}/${id}`}
                    >
                      {id}
                    </a>{' '}
                    {index !== sharedVisualDashboardIds.length - 1 && (
                      <span>, </span>
                    )}
                  </Fragment>
                ))}
              </>
            ),
          });
        }
        return (
          <Popover id={id.toString()} {...props}>
            {title && (
              <Popover.Title className={classes.title}>{title}</Popover.Title>
            )}

            <Popover.Content className={classes.content}>
              {values.map(({ key, value }) => (
                <Fragment key={key}>
                  <div className={classes.key}>{key}: </div>

                  <div className={classNames(classes.key, classes.value)}>
                    {value}
                  </div>
                </Fragment>
              ))}
            </Popover.Content>
          </Popover>
        );
      }}
      delay={{ show: 500, hide: 200 }}
    >
      {children}
    </OverlayTrigger>
  );
};
