import React from 'react';

import { VizWrapperLink } from 'components/viz-wrapper-link/VizWrapperLink';

import classes from './ReportTile.module.scss';
import { ReportTileImage } from '../report-tile-image/ReportTileImage';
import { ReportTilePopover } from '../report-tile-popover/ReportTilePopover';
import classNames from 'classnames';
import { DashboardDatasetInfo } from '../../pages/home/types/dashboard-dataset-info.type';

export type ReportTileProps = {
  id: number;
  report: {
    id: number;
    type: string;
    baseImg: string;
    editUrl: string;
    title: string;
    url: string;
    permissions: string[];
    updated: {
      date: string;
      by: string;
    };
    workspace: string;
    // Available when dashboard
    img?: string;
    // Available when app
    rw?: boolean;
    // For linked visuals
    sharedVisualDashboardIds?: string[];
  };
  isSelectMode?: boolean;
  onSelect?: () => void;
  selected?: boolean;
  // Undefined for Apps
  datasetInfo?: DashboardDatasetInfo;
};

export const ReportTile = ({
  report,
  datasetInfo,
  onSelect,
  selected,
  isSelectMode,
}: ReportTileProps): JSX.Element => {
  const {
    id,
    type,
    baseImg,
    editUrl,
    title,
    url,
    permissions,
    sharedVisualDashboardIds,
  } = report;
  const isApp = type === 'app';

  const hasEditPermission =
    (type === 'dashboard' && permissions.includes('ds_appedit')) ||
    (isApp && report.rw);

  let reportClassName = '';

  if (type === 'app') {
    reportClassName = classes.appIcon;
  } else if (type === 'dashboard') {
    reportClassName = classes.dashboardIcon;
  } else {
    reportClassName = classes.visualIcon;
  }

  return (
    <div
      className={classes.relativeContainer}
      data-test-id={title}
      data-test-report-id={id}
    >
      <VizWrapperLink
        className="d-flex flex-column h-100 position-relative"
        href={url}
        target={isApp ? '_blank' : '_self'}
        onClick={(event) => {
          if (isSelectMode) {
            event.preventDefault();
            onSelect?.();
          }
        }}
        testId="grid-item-link"
      >
        {isApp && (
          <div className={classes.launchApp}>
            <span className="fa fa-external-link" /> LaunchApp
          </div>
        )}

        <div className={classes.imageContainer}>
          <ReportTileImage
            img={type === 'app' ? undefined : report.img}
            baseImg={baseImg}
            type={type}
            size="md"
          />
        </div>

        <ReportTilePopover
          report={report}
          datasetInfo={datasetInfo}
          sharedVisualDashboardIds={sharedVisualDashboardIds}
        >
          <div className={classNames(classes.title, 'text-truncate')}>
            {title}
          </div>
        </ReportTilePopover>
      </VizWrapperLink>

      {onSelect && (
        <i
          className={classNames('fa fa-check', classes.selectIcon, {
            [classes.selected]: selected,
          })}
          onClick={onSelect}
          data-test-id={selected ? 'btn-selected' : 'btn-unselected'}
        />
      )}

      <div
        className={classNames(
          classes.icon,
          classes.reportIcon,
          reportClassName
        )}
      />

      {hasEditPermission && (
        <VizWrapperLink
          className={classNames(classes.icon, classes.editIcon)}
          href={editUrl}
          onClick={(event) => {
            if (isSelectMode) {
              event.preventDefault();
              onSelect?.();
            }
          }}
          testId="edit-report-button"
        />
      )}
    </div>
  );
};
