import React, { useCallback, useEffect, useState } from 'react';
import Dropdown from 'react-bootstrap/Dropdown';

import classes from './SettingsMenu.module.scss';
import {
  getVisibleSettingsMenuDropdownItems,
  SettingsMenuItem,
} from './get-visible-settings-menu-dropdown-items';
import { VizNavDropdown } from '../viz-nav-dropdown/VizNavDropdown';
import classNames from 'classnames';
import { legacyBridge } from 'services/legacy-bridge';

const { visibleItems, visibleSiteAdministrationItems } =
  getVisibleSettingsMenuDropdownItems();

const hasVisibleDropdownItems =
  visibleItems.length + visibleSiteAdministrationItems.length > 0;

type SettingsMenuProps = {
  isActive: boolean;
  activeTab: string | undefined;
};

export const SettingsMenu = ({
  isActive,
  activeTab,
}: SettingsMenuProps): JSX.Element | null => {
  const [activeDropdownItem, setActiveDropdownItem] = useState<string | null>(
    activeTab ?? null
  );

  useEffect(() => {
    if (activeTab === undefined) {
      const subscription = legacyBridge.navigationBarEvents.subscribe(
        ({ type, payload }) => {
          if (
            type === 'set-active-navigation-tab' &&
            payload.activeTab === 'settings'
          ) {
            setActiveDropdownItem(payload?.activeDropdownItem ?? null);
          }
        }
      );

      return () => subscription.unsubscribe();
    }
  }, [activeTab]);

  const getDropdownItemElement = useCallback(
    ({ id, href, label, iconClass, testId }: SettingsMenuItem) => (
      <Dropdown.Item
        key={label}
        as="a"
        href={href}
        data-test-id={testId}
        className={id === activeDropdownItem ? classes.menuItemActive : ''}
      >
        <i className={classNames(classes.menuItemIcon, `fa ${iconClass}`)} />
        {label}
      </Dropdown.Item>
    ),
    [activeDropdownItem]
  );

  if (!hasVisibleDropdownItems) {
    return null;
  }

  return (
    <VizNavDropdown
      toggle={{
        id: 'settings-menu',
        testId: 'navbar-settings-dropdown-toggle',
        content: <i className="fa fa-cog" />,
        isActive,
      }}
      menu={{
        testId: 'navbar-settings-menu',
        content: (
          <>
            {visibleItems.length > 0 && (
              <>{visibleItems.map(getDropdownItemElement)}</>
            )}

            {visibleSiteAdministrationItems.length > 0 && (
              <>
                <Dropdown.Divider />

                <div className={classes.header}>Site Administration:</div>

                {visibleSiteAdministrationItems.map(getDropdownItemElement)}
              </>
            )}
          </>
        ),
      }}
    />
  );
};
