import { AceEditor } from 'arc/types/editor-wrapper.type';
import { VizButton } from 'components/viz-button/VizButton';
import React, { memo, useEffect, useRef, useState } from 'react';
import { getGlobalArc } from 'utils/get-global-arc';
import {
  getDatabaseListPlaceHolder,
  getTableListPlaceHolder,
} from 'utils/list-placeholder';
import styles from './SQLExplorer.module.scss';
import { ExpressionEditor } from './expression-editor/ExpressionEditor';
import { TableList } from './table-selector-container/table-selector/TableList';
import {
  SqlQueryResponse,
  getSelectedTextFromTheExpressionEditor,
  runSqlQuery,
} from './util/util';
import { Notification } from 'components/notification/Notification';
import { BaseQuery } from 'arc/types/baseQuery.type';
import { addLimitToSQLQuery } from 'arc/misc/misc';
import { Form, Popover } from 'react-bootstrap';
import InfoIcon from '@cloudera-internal/cuix-core/icons/svg/ic-info.svg';
import { cdlOrange500 } from '@cloudera-internal/cuix-core/variables';
import { VizTooltip } from 'components/VizTooltip';
import { ConnectionBase as Connection } from 'types/data-connection.type';
export interface ListProps {
  databaseList?: string[];
  tableList?: string[];
  defaultDatabase?: string;
  databaseLoadError?: Error | unknown;
  tableLoadError?: Error | unknown;
  isTableListLoading: boolean;
  isDatabaseListLoading: boolean;
}

export interface EditorProps {
  connection?: Connection;
  query?: string;
  onChange?: (text: string) => void;
}

const resultContainerId = 'modal-sample-result-table';
const displayQueryResultInTable = (
  sqlResult: SqlQueryResponse<number | string | Date | undefined | null>,
  id: string
) => {
  if (!sqlResult.rowcount) {
    return;
  }

  if (sqlResult?.colnames?.length === 0) {
    sqlResult.colnames = Array<string>(sqlResult.rowcount).fill('');
  }

  getGlobalArc().misc.drawTable(
    id,
    sqlResult.rows,
    sqlResult.colnames,
    [
      {
        key: 'table_wrap_column_contents',
        value: 'false',
      },
      {
        key: 'table_full_width',
        value: 'true',
      },
    ],
    { enablePagination: true, pageSize: 10, enableInfo: true }
  );
};

const SQLExplorer = memo(
  ({
    connection,
    databaseList,
    tableList,
    isTableListLoading,
    isDatabaseListLoading,
    databaseLoadError,
    tableLoadError,
    query,
    onChange,
  }: ListProps & EditorProps): JSX.Element => {
    const [queryResultHeightInPx, setQueryResultHeightInPx] = useState(50);
    const [isLoading, setIsLoading] = useState<boolean>(false);
    const [baseQuery, setBaseQuery] = useState<BaseQuery | null>(null);
    const [result, setResult] = useState<SqlQueryResponse<
      number | string | Date | undefined | null
    > | null>(null);
    const [error, setError] = useState<any>(null);
    const editorRef = useRef<AceEditor>();
    const addLimitRef = useRef<HTMLInputElement>(null);
    const resultTableContainerRef = useRef<HTMLDivElement>(null);

    const onTableNameClick = (database: string, table: string) => {
      if (editorRef.current) {
        editorRef.current
          .getSession()
          .insert(
            editorRef.current.getCursorPosition(),
            table ? `${database}.${table}` : ''
          );
        editorRef.current.focus();
      }
    };

    const runQuery = async () => {
      const editor = editorRef.current;
      if (editor) {
        let query = getSelectedTextFromTheExpressionEditor(editor);
        if (!query.length) {
          setError('No command specified');
          return;
        }
        setIsLoading(true);
        setError(null);
        setResult(null);
        if (addLimitRef.current?.checked) {
          query = addLimitToSQLQuery(query, 100);
        }
        editor.focus();

        const handleResult = (
          result: SqlQueryResponse<number | string | Date | undefined | null>
        ) => {
          setQueryResultHeightInPx(
            result.rowcount ? (Math.min(result.rowcount, 10) + 2) * 3 + 40 : 50
          );
          setResult(result);
          setIsLoading(false);
        };

        const errorReportCb = (error: Error | string) => {
          setError(new Error(error.toString()));
          setIsLoading(false);
        };

        const baseQuery = runSqlQuery(
          query,
          connection?.id ?? -1,
          handleResult,
          errorReportCb
        );

        setBaseQuery({ ...baseQuery });
      }
    };

    useEffect(() => {
      if (!isLoading && result) {
        displayQueryResultInTable(result, resultContainerId);
        resultTableContainerRef?.current?.scrollIntoView({
          behavior: 'smooth',
        });
      }
    }, [result, isLoading]);

    return (
      <div className={styles.mainContainer}>
        <TableList
          onTableNameClick={onTableNameClick}
          open={true}
          tableList={tableList}
          databaseList={databaseList}
          databaseListPlaceHolder={getDatabaseListPlaceHolder(
            isDatabaseListLoading,
            databaseLoadError
          )}
          tableListSelectorPlaceholder={getTableListPlaceHolder(
            isTableListLoading,
            isDatabaseListLoading, // TODO naming
            false,
            tableLoadError,
            databaseList,
            tableList
          )}
        ></TableList>
        <div className={`${styles.expressionEditorContainer} pl-2`}>
          <ExpressionEditor
            query={query}
            onChange={onChange}
            ref={editorRef}
          ></ExpressionEditor>
          <div className={styles.buttonContainer}>
            <div className="d-flex flex-row">
              <Form.Check
                ref={addLimitRef}
                type="checkbox"
                id="autoQueries"
                label={`Apply "100 LIMIT" clause`}
                defaultChecked={true}
                className="ml-1 mb-1 mr-0"
              />
              {addLimitRef.current?.checked && error && (
                <VizTooltip
                  placement="bottom"
                  trigger="hover"
                  delay={100}
                  overlay={
                    <Popover id="limit-info-popover">
                      {
                        <span className="d-flex p-3">
                          Try to re-run without enforcing LIMIT 100 clause
                        </span>
                      }
                    </Popover>
                  }
                >
                  <InfoIcon
                    color={cdlOrange500}
                    className={styles.warning}
                  ></InfoIcon>
                </VizTooltip>
              )}
            </div>
            {!isLoading && (
              <>
                <VizButton
                  color="secondary"
                  size="medium"
                  onClick={runQuery}
                  disabled={isLoading}
                  className="pt-2 pl-3 pr-3 pb-2 ml-1"
                >
                  <i className="fa fa-table mr-1"></i>
                  <span>{'Show Data'} </span>
                </VizButton>
              </>
            )}

            {isLoading && (
              <VizButton
                color="danger"
                size="medium"
                disabled={!baseQuery}
                onClick={() => {
                  baseQuery?.qstate.cancel();
                  setIsLoading(false);
                  setError(new Error('Cancelled'));
                }}
                className="pt-2 pl-3 pr-3 pb-2 ml-2"
              >
                <i className="fa fa-spinner fa-spin"></i>
                <span className="ml-1">Cancel</span>
              </VizButton>
            )}
          </div>
          <div>
            {!isLoading && result && (
              <div
                id={resultContainerId}
                ref={resultTableContainerRef}
                className="w-100 p-3 mt-3"
                style={{
                  minHeight: `${queryResultHeightInPx}px`,
                }}
              >
                {!result.rowcount && (
                  <Notification message="No data available" variant="success" />
                )}
              </div>
            )}
            {!isLoading && error && (
              <Notification
                message={error?.message ?? error}
                variant="danger"
              />
            )}
            {!isLoading && !result && !error && null}
          </div>
        </div>
      </div>
    );
  }
);

export { SQLExplorer };
