import { AceEditor } from 'arc/types/editor-wrapper.type';
import React, {
  forwardRef,
  memo,
  useEffect,
  useImperativeHandle,
  useMemo,
  useState,
} from 'react';
import { getGlobalArc } from 'utils/get-global-arc';
import classes from './ExpressionEditor.module.scss';

const arc = getGlobalArc();

export type ExpressionEditorProp = {
  query?: string;
  onChange?: (text: string) => void;
};
const ExpressionEditor = memo(
  forwardRef<AceEditor | undefined, ExpressionEditorProp>(
    ({ query = '', onChange }, ref): JSX.Element => {
      const exprId = useMemo(() => `expr-editor-${arc.misc.uniqueId()}`, []);
      const [editor, setEditor] = useState<AceEditor>();
      useImperativeHandle(ref, () => editor, [editor]);

      // Initialize the editor and attach it to the DOM.
      useEffect(() => {
        if (!exprId) {
          return;
        }

        const aceEditorObject = arc.misc.expressionEditor(
          undefined,
          undefined,
          exprId,
          $('#react-rawsql-expr-editor-container'),
          undefined,
          true
        );
        const aceEditor = aceEditorObject.aceEditor();
        setEditor(aceEditor);

        return () => aceEditorObject.destroy();
      }, [exprId]);

      // Add the change listener to the editor.
      useEffect(() => {
        if (!onChange || !editor) {
          return;
        }

        editor.getSession().on('change', () => {
          onChange(editor.getSession().getValue());
        });

        return () => editor.getSession().removeAllListeners('change');
      }, [editor, onChange]);

      // Add the initial query to the editor.
      useEffect(() => {
        if (editor && editor.getSession().getValue() !== query) {
          editor.getSession().setValue(query);
        }
      }, [editor, query]);

      return (
        <div className={`${classes.editorContainer} pl-2`}>
          <h5>Enter SQL below</h5>
          {exprId && (
            <div
              id="react-rawsql-expr-editor-container"
              className="edit-expression-textbox w-100 float-none"
              data-test-id="react-rawsql-expr-editor-container"
            >
              <div className="form-group">
                <div
                  className={`form-control expression-field arc-no-dragging mw-100 mh-100 ${classes.height300px}`}
                  id={exprId}
                ></div>
              </div>
              <div className="pull-right edit-expression-autocomplete"></div>
            </div>
          )}
        </div>
      );
    }
  )
);

export { ExpressionEditor };
