import React, { useCallback, useEffect, useState } from 'react';

import { useToggle } from 'hooks/use-toggle';
import { useLegacyBridge } from 'services/legacy-bridge';

import classes from './TemporaryPasswordField.module.scss';
import classNames from 'classnames';

interface DomElementWithValidation extends JQuery<HTMLElement> {
  validator: (cmd: string) => void;
}

// This component should be removed
// after the create and edit user form is moved into a React component
interface TemporaryPasswordFieldProps {
  name: string;
  label: string;
  placeholder?: string;
  autofocus?: boolean;
  validation: {
    minLength: number;
    hasNumber?: boolean;
    hasSymbol?: boolean;
  };
  domElementWithValidation?: DomElementWithValidation;
}

const TemporaryPasswordField = ({
  name,
  label,
  placeholder,
  autofocus,
  validation,
  domElementWithValidation = $(`#${name}`).closest(
    'form'
  ) as DomElementWithValidation,
}: TemporaryPasswordFieldProps): JSX.Element => {
  const [shouldBeValidated, setShouldBeValidated] = useState(true);
  const [isVisible, toggleVisibility, resetVisibility] = useToggle();

  useEffect(() => {
    // In the bootstrap-validator-0.11.9/validator.js the onInput method
    // checks weather the target is in the initial form's input lists,
    // decorated with [data-validation=true].
    // The react TemporaryPasswordField when it rerenders, the reference and/or the 'data-validation'
    // attributes changes.
    // When the reference/data-validate attr changes, the validation without this update
    // use the old initiated list of items to be checked for validation, and returns invalid result!
    domElementWithValidation?.validator('update');
  }, [shouldBeValidated, domElementWithValidation]);

  const onValidateChange = useCallback(
    ({ payload }: { payload?: { validate: boolean } }) => {
      setShouldBeValidated(!!payload?.validate);
    },
    [setShouldBeValidated]
  );

  // password-edit-reset-visibility | password-novel-reset-visibility
  useLegacyBridge(`${name}-reset-visibility`, resetVisibility);

  // password-edit-change-validate
  useLegacyBridge(`${name}-change-validate`, onValidateChange);

  const dynamicValidationAttributes: {
    'data-requirenumber'?: boolean;
    'data-requiresymbol'?: boolean;
  } = {};

  if (validation.hasNumber) {
    dynamicValidationAttributes['data-requirenumber'] = true;
  }

  if (validation.hasSymbol) {
    dynamicValidationAttributes['data-requiresymbol'] = true;
  }

  const showPasswordId = `${name}-showPasswordField`;

  return (
    <div className="form-row w-100 col">
      <div className="form-group col col-md-6">
        <label htmlFor={name}>{label}</label>
        <div className="input-group mb-0 w-100">
          <input
            id={name}
            type={isVisible ? 'text' : 'password'}
            className="form-control form-control-sm password-input form-control-sm form-element"
            placeholder={placeholder}
            autoFocus={autofocus}
            data-validate={shouldBeValidated}
            required
            data-minlength={validation.minLength}
            data-minlength-error={`Password must be at least ${validation.minLength} characters`}
            {...dynamicValidationAttributes}
            autoComplete="off"
          />
          <div
            className={classNames('input-group-append', classes.clickable)}
            id={showPasswordId}
            onClick={toggleVisibility}
          >
            <span className="input-group-text">
              <i className={`fa ${isVisible ? 'fa-eye' : 'fa-eye-slash'}`} />
            </span>
          </div>
        </div>

        <p className="help-block form-text with-errors" />
      </div>
    </div>
  );
};

export default TemporaryPasswordField;
