import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import * as React from 'react';
import { ToggleDropdownInput } from './ToggleDropdownInput';

configure({ testIdAttribute: 'data-test-id' });

const mockOptions = Array.from({ length: 3 }, (_, i) => ({
  value: `category-${i % 3}`,
  label: `category-${i % 3}`,
}));

describe('TokenDropdownInput component', () => {
  it('renders without error', () => {
    render(
      <ToggleDropdownInput
        id={'category'}
        label="Category"
        options={mockOptions}
        value={[]}
        onChange={() => {}}
        isMulti
      />
    );
  });

  it('displays a list of options upon opening it', async () => {
    const mockOnChange = jest.fn();

    render(
      <ToggleDropdownInput
        id="category"
        label="Category"
        options={mockOptions}
        value={[]}
        onChange={mockOnChange}
        isMulti={false}
      />
    );

    fireEvent.click(screen.getByRole('button', { name: 'Category' }));

    await waitFor(() => {
      expect(screen.getByTestId('popover-category-0')).toBeInTheDocument();
    });
    expect(screen.getByTestId('popover-category-1')).toBeInTheDocument();
    expect(screen.getByTestId('popover-category-2')).toBeInTheDocument();
  });

  it('options can be deselected', async () => {
    const mockOnChange = jest.fn();

    render(
      <ToggleDropdownInput
        id={'category'}
        label="Category"
        options={mockOptions}
        value={['category-0', 'category-1']}
        onChange={mockOnChange}
        isMulti
      />
    );

    fireEvent.click(screen.getByRole('button', { name: 'Category' }));
    await waitFor(() => {
      expect(screen.getByTestId('popover-category-0')).toBeInTheDocument();
    });

    // the 0 option is not there anymore
    fireEvent.click(screen.getByTestId('popover-category-0'));
    expect(mockOnChange).toHaveBeenCalledWith(['category-1']);
  });

  describe('when the isMulti prop is true', () => {
    it('multiple options can be selected', async () => {
      const mockOnChange = jest.fn();

      render(
        <ToggleDropdownInput
          id={'category'}
          label="Category"
          options={mockOptions}
          value={['category-0']}
          onChange={mockOnChange}
          isMulti
        />
      );

      fireEvent.click(screen.getByRole('button', { name: 'Category' }));
      await waitFor(() => {
        expect(screen.getByTestId('popover-category-1')).toBeInTheDocument();
      });

      fireEvent.click(screen.getByTestId('popover-category-1'));
      expect(mockOnChange).toHaveBeenCalledWith(['category-0', 'category-1']);
    });
  });

  describe('when the isMulti prop is false', () => {
    it('only one option can be selected', async () => {
      const mockOnChange = jest.fn();

      render(
        <ToggleDropdownInput
          id={'category'}
          label="Category"
          options={mockOptions}
          value={['category-0']}
          onChange={mockOnChange}
          isMulti={false}
        />
      );

      fireEvent.click(screen.getByRole('button', { name: 'Category' }));
      await waitFor(() => {
        expect(screen.getByTestId('popover-category-1')).toBeInTheDocument();
      });

      fireEvent.click(screen.getByTestId('popover-category-1'));
      expect(mockOnChange).toHaveBeenCalledWith('category-1');
    });
  });
});
