import DropdownIcon from '@cloudera-internal/cuix-core/icons/svg/ic-dropdown.svg';
import classNames from 'classnames';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import React, { memo, useCallback, useState } from 'react';
import { OverlayTrigger, Popover } from 'react-bootstrap';
import classes from './ToggleDropdownInput.module.scss';

type ToggleDropdownInputProps<TValue> = Omit<
  React.ComponentProps<typeof ClouderaButton>,
  'value' | 'onChange'
> & {
  id: string;
  options: {
    value: TValue;
    label: React.ReactNode;
    stringLabel?: string;
  }[];
  label: string;
  columnCount?: number;
} & (
    | {
        value: TValue;
        onChange: (value: TValue) => void;
        isMulti: false;
      }
    | {
        value: TValue[];
        onChange: (value: TValue[]) => void;
        isMulti: true;
      }
  );

//Support generic prop types with memo (https://stackoverflow.com/a/60386939)
const typedMemo: <T>(c: T) => T = memo;

/**
 * A popover dropdown style list input
 * @param options a list of filter options
 * @param label the text to be displayed on the filter button
 * @param id unique identifier of the filter element
 * @param value a list of the currently selected options or the selected option
 * @param onChange callback invoked when the selection changes
 * @param isMulti indicates wether the one or several values are accepted. Based on this, the value is a list or a single value
 */
export const ToggleDropdownInput = typedMemo(
  <TValue extends unknown>({
    value,
    onChange,
    id,
    options,
    label,
    isMulti,
    columnCount = 1,
    ...props
  }: ToggleDropdownInputProps<TValue>): JSX.Element => {
    const [showFilter, setShowFilter] = useState(false);

    const handleSelectOption = useCallback(
      (option: TValue) => {
        if (isMulti) {
          if (value.includes(option)) {
            onChange(value.filter((item) => item !== option));
          } else {
            onChange([...value, option]);
          }
        } else {
          onChange(option);
          setShowFilter(false);
        }
      },
      [isMulti, value, onChange]
    );

    const popover = (
      <Popover id={`popover-${id}`} className={classes.popover}>
        <Popover.Content
          className={classes.popoverContent}
          style={{ gridTemplateColumns: `repeat(${columnCount}, 1fr)` }}
        >
          {options.map((option) => {
            const isSelected = isMulti
              ? value.includes(option.value)
              : value === option.value;

            return (
              <button
                key={String(option.value)}
                type="button"
                data-test-id={`popover-${option.value}`}
                onClick={() => handleSelectOption(option.value)}
                className={classNames(
                  classes.popoverButton,
                  isSelected && classes.selected
                )}
              >
                {option.label}
              </button>
            );
          })}
        </Popover.Content>
      </Popover>
    );

    return (
      <OverlayTrigger
        trigger="click"
        placement="bottom-start"
        rootClose
        overlay={popover}
        show={showFilter}
        onToggle={setShowFilter}
      >
        <button
          type="button"
          className={classes.toggleButton}
          data-test-id={`dropdown-input-${id}`}
          {...props}
        >
          {label}
          <DropdownIcon className="ml-1" />
        </button>
      </OverlayTrigger>
    );
  }
);
