import React, { useState, useEffect, useRef } from 'react';
import Markdown from '../chatbot/Markdown';
import { tokenize } from 'utils/ai-response-utils';
import { Token } from 'types/api.type';

type TypewriterProps = {
  text: string;
  isMarkdown: boolean;
  onAnimationStateChange: (x: boolean) => void;
  speed?: number;
};

export const Typewriter: React.FC<TypewriterProps> = ({
  text,
  isMarkdown,
  onAnimationStateChange,
  speed = 35,
}) => {
  const [tokens, setTokens] = useState<Token[]>([]);
  const [displayedTokens, setDisplayedTokens] = useState<Token[]>([]);
  const [currentTokenIndex, setCurrentTokenIndex] = useState(0);
  const [prevText, setPrevText] = useState('');
  const [animating, setAnimating] = useState(false);

  const frameRequest = useRef<number | null>(null);
  const startTimeRef = useRef<number | null>(null);

  useEffect(() => {
    const newAnimating =
      currentTokenIndex > 0 && currentTokenIndex < tokens.length;
    if (newAnimating !== animating) {
      setAnimating(newAnimating);
      onAnimationStateChange(newAnimating);
    }
  }, [animating, currentTokenIndex, onAnimationStateChange, tokens.length]);

  useEffect(() => {
    if (text === prevText) return;

    if (text.startsWith(prevText) && text.length > prevText.length) {
      // Streaming: append new tokens
      const appended = text.slice(prevText.length);
      const newTokens = tokenize(appended, isMarkdown);
      setTokens((prev) => [...prev, ...newTokens]);
    } else {
      // Big change: skip animation for all
      const newTokens = tokenize(text, isMarkdown).map((t) => ({
        ...t,
        isImmediate: true,
      }));
      setTokens(newTokens);
      setDisplayedTokens(newTokens);
      setCurrentTokenIndex(newTokens.length);
    }
    setPrevText(text);
  }, [prevText, text, isMarkdown]);

  useEffect(() => {
    if (currentTokenIndex >= tokens.length) return;

    const currentToken = tokens[currentTokenIndex];
    if (!currentToken) return;

    if (currentToken.isImmediate) {
      setDisplayedTokens((prev) => [...prev, currentToken]);
      setCurrentTokenIndex((prev) => prev + 1);
      return;
    }

    function animateToken(timestamp: number) {
      if (startTimeRef.current == null) {
        startTimeRef.current = timestamp;
      }
      const elapsed = timestamp - startTimeRef.current;

      if (elapsed >= speed) {
        // We can display this token now
        setDisplayedTokens((prev) => [...prev, currentToken]);
        setCurrentTokenIndex((prev) => prev + 1);
        startTimeRef.current = null;
      } else {
        // Not enough time: keep waiting
        frameRequest.current = requestAnimationFrame(animateToken);
      }
    }

    // Kick off the animation
    frameRequest.current = requestAnimationFrame(animateToken);

    return () => {
      if (frameRequest.current) {
        cancelAnimationFrame(frameRequest.current);
        frameRequest.current = null;
      }
    };
  }, [currentTokenIndex, tokens, speed]);

  if (isMarkdown) {
    const content = displayedTokens.map((t) => t.content).join(' ');
    return <Markdown markdownText={content} />;
  }

  // If not markdown, render as plain text with line breaks and spacing
  return (
    <p>
      {displayedTokens.map((token, idx) => {
        if (token.content.includes('\n')) {
          const newlines = (token.content.match(/\n/g) || []).length;
          return (
            <React.Fragment key={idx}>
              {Array.from({ length: newlines }).map((_, i) => (
                <br key={i} />
              ))}
            </React.Fragment>
          );
        } else {
          return <span key={idx}>{token.content} </span>;
        }
      })}
    </p>
  );
};
