import React, { useMemo, SetStateAction, Dispatch, useEffect } from 'react';
import { useState, useCallback } from 'react';
import { useMutation, useQuery, useQueryClient } from '@tanstack/react-query';
import { legacyBridge, useLegacyBridge } from 'services/legacy-bridge';
import {
  DataSet,
  getDatasetVersions,
  datasetVersionQueryToken,
} from 'apis/get-dataset-versions';
import { changeDatasetVersion } from 'apis/change-dataset-version';
import { changeDatasetVersionName } from 'apis/change-dataset-version-name';
import { formatDate } from 'arc/misc/misc';
import { useNotificationStore } from 'stores/use-notification-store';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import classes from './VersionControlPage.module.scss';
import { VizButton } from 'components/viz-button/VizButton';
import {
  Button,
  FormControl,
  InputGroup,
  Popover,
  Table,
  Tooltip,
} from 'react-bootstrap';
import { VizTooltip } from 'components/VizTooltip';

type VersionControlPageProps = {
  masterId: string;
};

export const VersionControlPage = ({ masterId }: VersionControlPageProps) => {
  const queryClient = useQueryClient();

  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const { data: datasetVersions } = useQuery({
    queryKey: [datasetVersionQueryToken, masterId],
    queryFn: async () => {
      const versions = await getDatasetVersions({ id: masterId });
      return versions.sort((a, b) => b.version_id - a.version_id);
    },
    onError: () => {
      addNotification(
        {
          id: 'datasets-danger',
          variant: 'danger',
          message: 'Error fetching datasets.',
        },
        { delay: 2000 }
      );
    },
    initialData: [],
  });

  const { mutateAsync: changeVersion } = useMutation(changeDatasetVersion);
  const { mutateAsync: changeVersionName } = useMutation(
    changeDatasetVersionName
  );

  const onNewVersionCreate = useCallback(() => {
    queryClient.invalidateQueries({
      queryKey: [datasetVersionQueryToken, masterId],
      refetchType: 'active',
    });
  }, [masterId, queryClient]);

  useLegacyBridge('new-dataset-version-created', onNewVersionCreate);

  const [editingDataset, setEditingDataset] = useState<null | DataSet>(null);

  const saveChangedVersionName = async (
    versionName: string,
    selectedDataset: DataSet
  ) => {
    try {
      await changeVersionName({
        id: selectedDataset.id,
        version_name: versionName,
        dataset_name: selectedDataset.name,
        dataset_type: selectedDataset.type,
        dataset_detail: selectedDataset.detail,
        dataset_info: selectedDataset.info,
        is_named_version: true,
      });

      queryClient.invalidateQueries({
        queryKey: [datasetVersionQueryToken, masterId],
        refetchType: 'active',
      });

      addNotification(
        {
          id: 'version-name--change-success',
          variant: 'success',
          message: `Dataset name has been changed `,
        },
        { delay: 1500 }
      );

      setEditingDataset(null);

      legacyBridge.events.next({
        type: 'dataset-version-changed',
      });
    } catch (error) {
      addNotification(
        {
          id: 'version-name-danger',
          variant: 'danger',
          message: `Failed to change version name. ${error}`,
        },
        { delay: 3000 }
      );
    }
  };

  const revertToVersion = async (id: number) => {
    try {
      await changeVersion({
        master_id: masterId,
        version_id: id,
      });

      queryClient.invalidateQueries({
        queryKey: [datasetVersionQueryToken, masterId],
        refetchType: 'active',
      });

      addNotification(
        {
          id: 'version-change-success',
          variant: 'success',
          message: `Active dataset has been changed to dataset: ${
            datasetVersions.filter((ds) => ds.id === id)[0].version_name
          }`,
        },
        { delay: 1500 }
      );

      legacyBridge.events.next({
        type: 'dataset-version-changed',
      });
    } catch (error) {
      addNotification(
        {
          id: 'version-danger',
          variant: 'danger',
          message: `Failed to change version. ${error}`,
        },
        { delay: 3000 }
      );
    }
  };

  const activeDatasetVersion = useMemo(
    () =>
      datasetVersions.find(
        (datasetVersion) => datasetVersion.id.toString() === masterId
      ),
    [datasetVersions, masterId]
  );

  return (
    <div>
      <div className={classes.title}>
        <h5>
          <span className="text-muted">Dataset: </span>
          <span key={activeDatasetVersion?.id}>
            {activeDatasetVersion?.name}
          </span>
        </h5>
        <h3 className="mt-2 mp-3">Version Control</h3>
      </div>

      <div className={classes.mainSection}>
        <h5>This is the current version of this dataset</h5>
        {activeDatasetVersion && (
          <Table striped bordered hover>
            <thead>
              <tr>
                <th>Main Dataset Id</th>
                <th>Version Id</th>
                <th>
                  Version Name
                  <VizTooltip
                    placement="top"
                    overlay={
                      <Popover id="version_name_tooltip">
                        {
                          <span className="d-flex p-3">
                            To keep a version, you must name it; otherwise, it
                            will be removed according to the values set in the
                            Version Control Site settings.
                          </span>
                        }
                      </Popover>
                    }
                  >
                    <span className={classes.infoIcon} />
                  </VizTooltip>
                </th>
                <th>Dataset Name</th>
                <th>Create Date</th>
                <th>Created by</th>
                <th>Update Date</th>
                <th>Updated by</th>
              </tr>
            </thead>
            <tbody>
              <tr>
                <td>{activeDatasetVersion.id}</td>
                <td>{activeDatasetVersion.version_id}</td>
                <td>
                  <VersionNameColumn
                    onSave={saveChangedVersionName}
                    editingDataset={editingDataset}
                    setEditingDataset={setEditingDataset}
                    version={activeDatasetVersion}
                  />
                </td>
                <td>{activeDatasetVersion.name}</td>
                <td>{formatDate(activeDatasetVersion.created)}</td>
                <td>{activeDatasetVersion.created_by}</td>
                <td>{formatDate(activeDatasetVersion.updated)}</td>
                <td>{activeDatasetVersion.updated_by}</td>
              </tr>
            </tbody>
          </Table>
        )}

        <div className="py-5"></div>

        <h5>These are the previous versions of this dataset</h5>
        <Table striped bordered hover size="sm" responsive>
          <thead>
            <tr>
              <th>Version Id</th>
              <th>
                Version Name
                <VizTooltip
                  placement="top"
                  overlay={
                    <Popover id="version_name_tooltip">
                      {
                        <span className="d-flex p-3">
                          To keep a version, you must name it; otherwise, it
                          will be removed according to the values set in the
                          Version Control Site settings.
                        </span>
                      }
                    </Popover>
                  }
                >
                  <span className={classes.infoIcon} />
                </VizTooltip>
              </th>
              <th>Dataset Name</th>
              <th>Update Date</th>
              <th>Updated by</th>
              <th>Action</th>
            </tr>
          </thead>
          <tbody>
            {datasetVersions
              .filter(
                (datasetVersion) => datasetVersion.id.toString() !== masterId
              )
              .map((datasetVersion) => (
                <tr key={datasetVersion.id}>
                  <td>{datasetVersion.version_id}</td>
                  <td>
                    <VersionNameColumn
                      onSave={saveChangedVersionName}
                      editingDataset={editingDataset}
                      setEditingDataset={setEditingDataset}
                      version={datasetVersion}
                    />
                  </td>
                  <td>{datasetVersion.name}</td>
                  <td>{formatDate(datasetVersion.updated)}</td>
                  <td>{datasetVersion.updated_by}</td>
                  <VizTooltip
                    placement="auto"
                    overlay={
                      <Tooltip id="version_swap_tooltip">
                        Change Version
                      </Tooltip>
                    }
                  >
                    <td>
                      <VizButton
                        color="secondary"
                        size="medium"
                        onClick={() => revertToVersion(datasetVersion.id)}
                      >
                        <i className="fa fa-history" aria-hidden="true"></i>
                      </VizButton>
                    </td>
                  </VizTooltip>
                </tr>
              ))}
          </tbody>
        </Table>
      </div>
    </div>
  );
};

const VersionNameColumn = ({
  version,
  setEditingDataset,
  editingDataset,
  onSave,
}: {
  version: DataSet;
  editingDataset: DataSet | null;
  setEditingDataset: Dispatch<SetStateAction<DataSet | null>>;
  onSave: (versionName: string, editingDataset: DataSet) => Promise<void>;
}) => {
  const [versionName, setVersionName] = useState(version.version_name);

  useEffect(() => {
    setVersionName(version.version_name);
  }, [version.version_name]);

  if (editingDataset?.id === version.id) {
    return (
      <InputGroup className="w-50" size="sm">
        <FormControl
          onChange={(event) => setVersionName(event.target.value)}
          value={versionName}
        />

        <InputGroup.Append>
          <Button
            onClick={() => {
              setEditingDataset(null);
              setVersionName(version.version_name);
            }}
            variant="outline-secondary"
          >
            Cancel
          </Button>

          <VizTooltip
            overlay={
              <Popover id="version-name-change-alert">
                <span className="d-flex p-3">Version name did not change</span>
              </Popover>
            }
            disabled={versionName !== version.version_name}
          >
            <Button
              disabled={versionName === version.version_name}
              onClick={() => {
                if (versionName !== version.version_name) {
                  onSave(versionName, editingDataset);
                }
              }}
              variant="outline-primary"
            >
              Save
            </Button>
          </VizTooltip>
        </InputGroup.Append>
      </InputGroup>
    );
  }

  return (
    <div className="d-flex gap-2 align-items-center">
      <span>{version.version_name}</span>

      <ClouderaButton onClick={() => setEditingDataset(version)} variant="link">
        <i className="fa fa-pencil"></i>
      </ClouderaButton>
    </div>
  );
};

export default VersionControlPage;
