import React, { useEffect, useState } from 'react';
import Modal from 'react-bootstrap/Modal';

import { VizButton } from 'components/viz-button/VizButton';
import { VizLink } from 'components/viz-link/VizLink';

import classes from './WhatSNewModal.module.scss';
import { Release } from '../release/Release';
import { getReleaseConfigs } from './get-release-configs';
import { useWhatSNewModalStore } from 'stores/use-what-s-new-modal-store';
import { setWhatSNewModalSeen } from 'apis/set-what-s-new-modal-seen';
import { getGlobalArc } from 'utils/get-global-arc';

const arc = getGlobalArc();

type WhatSNewModalProps = {
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
};

export const WhatSNewModal = ({
  shouldShowModal,
  isAlreadySeen,
}: WhatSNewModalProps): JSX.Element | null => {
  const { isOpen, open, close } = useWhatSNewModalStore(
    ({ isOpen, open, close }) => ({
      isOpen,
      open,
      close,
    })
  );

  useEffect(() => {
    if (shouldShowModal && !isAlreadySeen) {
      open();
    } else {
      close();
    }
  }, [close, isAlreadySeen, open, shouldShowModal]);

  const releases = getReleaseConfigs();

  const [openAccordionIds, setOpenAccordionIds] = useState(() => {
    const [latestRelease] = releases;

    return [
      latestRelease.version,
      ...latestRelease.items.map(
        ({ title }) => `${latestRelease.version}-${title}`
      ),
    ];
  });

  const toggleAccordion = (selectedId: string) =>
    setOpenAccordionIds((ids) => {
      const isAlreadyOpen = openAccordionIds.includes(selectedId);

      return isAlreadyOpen
        ? ids.filter((id) => selectedId !== id)
        : [...ids, selectedId];
    });

  const onConfirm = () => {
    setWhatSNewModalSeen().then(close);
  };

  if (!isOpen) {
    return null;
  }

  return (
    <Modal show={isOpen} onHide={onConfirm} size="lg" backdrop="static">
      <Modal.Header closeButton>
        <Modal.Title className={classes.modalTitle}>What's New</Modal.Title>
      </Modal.Header>

      <Modal.Body className={classes.modalBody}>
        {releases.map((release) => (
          <Release
            key={release.version}
            release={release}
            openAccordionIds={openAccordionIds}
            toggle={toggleAccordion}
          />
        ))}
      </Modal.Body>

      <Modal.Footer className="justify-content-between">
        <VizLink
          href={arc.frontend.globals.documentation.releaseNotesUrl}
          target="_blank"
        >
          Read the Complete Release Notes
          <i className="fa fa-external-link ml-1" />
        </VizLink>

        <VizButton
          color="secondary"
          size="medium"
          onClick={onConfirm}
          testId="arc-gotit-btn"
        >
          GOT IT
        </VizButton>
      </Modal.Footer>
    </Modal>
  );
};
