import * as yup from 'yup';
import React from 'react';
import axios from 'axios';

import { PasswordField } from 'components/password-field/PasswordField';
import { getGlobalArc } from 'utils/get-global-arc';
import { useNotificationStore } from 'stores/use-notification-store';
import { changeOwnPassword } from 'apis/change-own-password';
import { FieldErrors, UseFormRegister } from 'react-hook-form/dist/types';

type ChangePasswordFormProps = {
  onSuccess?: () => void;
};

const initialValues: ChangePasswordFormState = {
  password: '',
  newPassword: '',
  confirmPassword: '',
};

const formFields = (
  register: UseFormRegister<ChangePasswordFormState>,
  errors: FieldErrors<ChangePasswordFormState>
) => (
  <>
    <PasswordField
      label="Current Password"
      placeholder="enter current password"
      hasToggleVisibility
      data-test-id="change-password-form-password"
      {...register('password')}
      error={errors['password']}
    />

    <PasswordField
      label="New Password"
      placeholder="enter new password"
      hasToggleVisibility
      data-test-id="change-password-form-new-password"
      {...register('newPassword')}
      error={errors['newPassword']}
    />

    <PasswordField
      label="Confirm Password"
      placeholder="enter new password again"
      hasToggleVisibility
      data-test-id="change-password-form-confirm-new-password"
      {...register('confirmPassword')}
      error={errors['confirmPassword']}
    />
  </>
);

const arc = getGlobalArc();

const getValidationSchema = () => {
  let newPasswordJSONSchema = yup
    .string()
    .required('New Password is required.')
    .min(
      arc.frontend.globals.authentication.passwordValidation.minLength,
      `Password must be at least ${arc.frontend.globals.authentication.passwordValidation.minLength} characters`
    );

  if (arc.frontend.globals.authentication.passwordValidation.hasSymbol) {
    newPasswordJSONSchema = newPasswordJSONSchema.matches(
      /\W+/,
      'Must include at least one symbol.'
    );
  }

  if (arc.frontend.globals.authentication.passwordValidation.hasNumber) {
    newPasswordJSONSchema = newPasswordJSONSchema.matches(
      /\d+/,
      'Must include at least one number.'
    );
  }

  return yup.object().shape({
    password: yup.string().required('Password is required.'),
    newPassword: newPasswordJSONSchema,
    confirmPassword: yup
      .string()
      .required('Confirm password is required.')
      .oneOf([yup.ref('newPassword')], 'Your passwords don’t match'),
  });
};

export type ChangePasswordFormState = yup.InferType<
  ReturnType<typeof getValidationSchema>
>;

export const useChangePasswordForm = ({
  onSuccess,
}: ChangePasswordFormProps) => {
  const { addNotification, removeNotification } = useNotificationStore(
    ({ addNotification, removeNotification }) => ({
      addNotification,
      removeNotification,
    })
  );

  const validationSchema = getValidationSchema();

  const onSubmit = async ({
    password,
    newPassword,
    confirmPassword,
  }: ChangePasswordFormState) => {
    try {
      await changeOwnPassword(password, newPassword, confirmPassword);

      addNotification({
        id: -1,
        variant: 'success',
        message: 'Password successfully changed',
      });

      setTimeout(() => {
        removeNotification(-1);
        onSuccess?.();
      }, 1500);
    } catch (error) {
      if (axios.isAxiosError(error)) {
        addNotification(
          {
            variant: 'danger',
            message: `Failed to change password. ${error.response?.data}`,
          },
          { delay: 3000 }
        );
      }

      throw error;
    }
  };

  return {
    initialValues,
    formFields,
    validationSchema,
    onSubmit,
  };
};
