import { getCompletion } from 'apis/get-completion';
import { getEmbeddings } from 'apis/get-embeddings';
import { useChatbotMessaging } from './use-chatbot-messaging';
import { act, renderHook } from '@testing-library/react';
import {
  Answer,
  OverflowPolicy,
  ParamSelection,
  ParamSource,
} from 'types/chatbot-message.type';

jest.mock('apis/get-completion');
jest.mock('apis/get-embeddings');

window.arc.misc = {
  sessionKeyGenerator: {
    chatBot: jest.fn(),
  },
} as any;

const savedMessages =
  '[{"type":"welcome","text":"Welcome to the chat!","sender":"bot","ts":"2024-02-27T15:09:34.551Z"},{"type":"question","text":"New message","sender":"user","ts":"2024-02-27T15:09:34.553Z"}]';

const input = {
  sendDsReq: jest.fn(() => Promise.resolve({ rows: [], colnames: [] })),
  sendHttpReq: jest.fn(() => Promise.resolve({})),
  getEmbeddingSourceInfo: jest.fn(),
  getTooltips: jest.fn(),
  setDownloadInfo: jest.fn(),
  alertMessage: jest.fn(),
  addParams: jest.fn(),
  needSave: jest.fn(),
  reportId: 1,
  datasetId: 1,
  savedConversation: null,
  settings: {
    welcomeMessage: 'Welcome to the chat!',
    embeddingProfile: 'default',
    completionProfile: 'default',
    isMarkdown: false,
    embeddingMaxTokens: '1000',
    completionMaxTokens: '4000',
    overflowPolicy: 'throw-error' as OverflowPolicy,
    paramSelection: 'all' as ParamSelection,
    paramSource: 'all' as ParamSource,
    questionPromptPrefix: '<<question>>',
    fontSize: '16px',
    saveConversation: false,
  },
};

beforeEach(() => {
  jest.mocked(getEmbeddings).mockResolvedValue([]);
  jest.mocked(getCompletion).mockResolvedValue('Mock answer');

  input.sendDsReq.mockResolvedValue({ rows: [], colnames: [] });
  window.arc.misc.sessionKeyGenerator.chatBot = jest
    .fn()
    .mockReturnValue('TEST-STORAGE-KEY');
});

afterEach(() => {
  window.sessionStorage.clear();
  jest.resetAllMocks();
});

describe('Test useChatbotMessaging hook', () => {
  test('it should load the saved messages from the session storage', () => {
    sessionStorage.setItem('TEST-STORAGE-KEY', savedMessages);
    const { result } = renderHook(useChatbotMessaging, { initialProps: input });

    expect(JSON.stringify(result.current.messages)).toBe(savedMessages);
  });

  describe('when the sendMessage function is invoked', () => {
    test('it should generate an answer', async () => {
      const { result } = renderHook(useChatbotMessaging, {
        initialProps: { ...input, savedConversation: [] },
      });
      // Insert a message into the input
      act(() => {
        result.current.setInput('New message');
      });

      // Send the message
      await act(() => result.current.sendMessage());

      expect(result.current.messages.length).toBe(2);
      expect(result.current.messages[0].type).toBe('question');
      expect(result.current.messages[0].text).toBe('New message');
      expect(result.current.messages[1].type).toBe('answer');
      expect(result.current.messages[1].text).toBe('Mock answer');
    });

    test('it should save successful conversation to session storage', async () => {
      const { result } = renderHook(useChatbotMessaging, {
        initialProps: input,
      });
      // Insert a message into the input
      act(() => {
        result.current.setInput('New message');
      });

      // Send the message
      await act(() => result.current.sendMessage());

      // setDownloadInfo is called if message sending succeeded:
      expect(input.setDownloadInfo).toHaveBeenCalled();

      // alerts should have not appeared
      expect(input.alertMessage).toHaveBeenCalledTimes(0);

      // check if the session storage contains the key
      expect(sessionStorage.getItem('TEST-STORAGE-KEY')).toBeTruthy();
    });

    describe('with a message that previously resulted in an error', () => {
      test('it should replace the error with an answer', async () => {
        const { result } = renderHook(useChatbotMessaging, {
          initialProps: {
            ...input,
            savedConversation: [
              {
                type: 'question',
                id: 'questionId',
                text: 'Test question',
                sender: 'user',
                ts: new Date(),
              },
              {
                type: 'answer',
                id: 'answerId',
                errorId: 'questionId',
                text: 'Test error message',
                sender: 'bot',
                ts: new Date(),
                tooltips: [],
                embeddings: [],
                hasError: true,
                shouldAnimate: true,
              },
            ],
          },
        });

        // Retry the error message
        await act(() => result.current.sendMessage('questionId'));

        // Check that the answer was regenerated for the correct question
        expect(getEmbeddings).toHaveBeenCalledWith(
          input.sendHttpReq,
          'Test question',
          input.settings.embeddingProfile,
          input.settings.embeddingMaxTokens
        );

        // Check that the error was replaced by the answer
        expect(result.current.messages.length).toBe(2);
        const answer = result.current.messages[1] as Answer;
        expect(answer.id).toBe('answerId');
        expect(answer.text).toBe('Mock answer');
        expect(answer.hasError).toBeFalsy();
      });
    });

    describe('and the answer generation results in an error', () => {
      test('it should generate an error answer', async () => {
        // answer generation will result in an error
        jest.mocked(getEmbeddings).mockRejectedValue('error');

        const { result } = renderHook(useChatbotMessaging, {
          initialProps: { ...input, savedConversation: [] },
        });
        // Insert a message into the input
        act(() => {
          result.current.setInput('New message');
        });

        // Send the message
        await act(() => result.current.sendMessage());

        expect(result.current.messages.length).toBe(2);
        expect(result.current.messages[0].type).toBe('question');
        expect(result.current.messages[0].text).toBe('New message');

        const answer = result.current.messages[1] as Answer;
        expect(answer.type).toBe('answer');
        expect(answer.text).toContain('Something went wrong');
        expect(answer.hasError).toBeTruthy();
      });
    });
  });

  describe('when the resetMessage function is invoked', () => {
    test('it should reset the conversation to the initial message', () => {
      const { result } = renderHook(useChatbotMessaging, {
        initialProps: {
          ...input,
          savedConversation: [
            {
              type: 'question',
              id: '1234',
              text: 'Test message',
              sender: 'user',
              ts: new Date(),
            },
          ],
        },
      });
      // Check if the test message is in the chat
      expect(result.current.messages[0].text).toBe('Test message');

      act(() => {
        result.current.resetMessages();
      });

      // Check if the conversation was successfully reset
      expect(result.current.messages.length).toBe(1);
      expect(result.current.messages[0].type).toBe('welcome');
    });
  });
});
