import { useCallback, useEffect, useMemo, useState } from 'react';
import { getCompletion } from 'apis/get-completion';
import { getEmbeddings } from 'apis/get-embeddings';
import { getGlobalArc } from 'utils/get-global-arc';
import ShortUniqueId from 'short-unique-id';
import {
  Answer,
  DownloadInfo,
  LoadingStatus,
  Message,
  OverflowPolicy,
  ParamSource,
  ParamSelection,
  AddParamArg,
} from 'types/chatbot-message.type';
import { finalizeAnswer } from 'utils/ai-response-utils';

const arc = getGlobalArc();

export const useChatbotMessaging = ({
  sendDsReq,
  sendHttpReq,
  getEmbeddingSourceInfo,
  getTooltips,
  setDownloadInfo,
  alertMessage,
  addParams,
  needSave,
  reportId,
  datasetId,
  savedConversation,
  settings,
}: {
  sendDsReq: (
    embeddings: number[]
  ) => Promise<{ rows: string[][]; colnames: string[] }>;
  sendHttpReq: (httpReq: any) => Promise<any>;
  getEmbeddingSourceInfo: (rows: string[][]) => {
    filteredRows: string[][];
    aliases: string[];
    contextPrompt: string;
  };
  getTooltips: (rows: string[][]) => Record<string, string>[];
  setDownloadInfo: (downloadInfo: DownloadInfo) => void;
  alertMessage: (errorMsg: string) => void;
  addParams: (params: AddParamArg[]) => void;
  needSave: () => void;
  reportId: number;
  datasetId: number;
  savedConversation: Message[] | null;
  settings: {
    welcomeMessage: string;
    embeddingProfile: string;
    completionProfile: string;
    embeddingMaxTokens: string;
    completionMaxTokens: string;
    overflowPolicy: OverflowPolicy;
    paramSelection: ParamSelection;
    paramSource: ParamSource;
    questionPromptPrefix: string;
    fontSize: string;
    saveConversation: boolean;
  };
}) => {
  const uuid = useMemo(() => new ShortUniqueId({ length: 10 }), []);

  const getInitialMessages = useCallback(
    () => [
      {
        type: 'welcome',
        id: uuid.randomUUID(),
        text: settings.welcomeMessage,
        sender: 'bot',
        ts: new Date(),
      } as const,
    ],
    [settings.welcomeMessage, uuid]
  );

  const sessionKey = arc.misc.sessionKeyGenerator.chatBot(reportId);

  const loadSavedMessages = useCallback(() => {
    const savedMessages = sessionStorage.getItem(sessionKey);

    if (savedMessages) {
      return JSON.parse(savedMessages);
    }

    if (savedConversation) {
      return savedConversation;
    }

    return getInitialMessages();
  }, [getInitialMessages, savedConversation, sessionKey]);

  const [input, setInput] = useState<string>('');

  const [messages, setMessages] = useState<Message[]>(loadSavedMessages);
  const [tooltips, setTooltips] = useState<Record<string, string>[] | null>(
    null
  );
  const [loadingStatus, setLoadingStatus] = useState<LoadingStatus>('None');

  useEffect(() => {
    if (
      settings.saveConversation &&
      messages.length !== (savedConversation ?? []).length
    ) {
      needSave();
    }
  }, [messages, needSave, savedConversation, settings.saveConversation]);

  useEffect(() => {
    setMessages((current) =>
      current.map((msg) => {
        if (msg.type === 'welcome') {
          return { ...msg, text: settings.welcomeMessage };
        }
        return msg;
      })
    );
  }, [settings.welcomeMessage]);

  const saveConversationToSessionStorage = useCallback(
    (messages: Message[]) => {
      const messagesWithoutEmbeddings = messages.map((msg) => {
        if (msg.type === 'answer') {
          const { embeddings, ...rest } = msg;
          (rest as Answer).shouldAnimate = false;
          return rest;
        }
        return msg;
      });

      sessionStorage.setItem(
        sessionKey,
        JSON.stringify(messagesWithoutEmbeddings)
      );
    },
    [sessionKey]
  );

  const sendMessage = useCallback(
    async (errorId?: string) => {
      if (!input && !errorId) {
        return;
      }

      const msgId = uuid.randomUUID();
      let newMessages = [...messages];
      let embeddings: number[] = [];
      let errorMsgObj: Answer | undefined;

      if (!errorId) {
        newMessages.push({
          type: 'question',
          id: msgId,
          text: input,
          sender: 'user',
          ts: new Date(),
        });

        setMessages(newMessages);
      }

      setInput('');

      try {
        let previousMessages = newMessages
          .slice(0, -1)
          .map(({ sender, text }) => ({
            role: sender === 'user' ? 'user' : 'assistant',
            content: text,
          }));

        let inputMsg = input;
        let downloadInfo: DownloadInfo;

        if (errorId) {
          const targetMsg = newMessages.find((msg) => msg.id === errorId);
          const targetIndex = newMessages.findIndex(
            (msg) => msg.id === targetMsg?.id
          );

          previousMessages = previousMessages.slice(0, targetIndex + 1);
          inputMsg = targetMsg?.text ?? '';
          const existingAnswerMsg = newMessages.find(
            (m) => m.type === 'answer' && m.errorId === errorId
          );

          if (existingAnswerMsg && existingAnswerMsg.type === 'answer') {
            embeddings = existingAnswerMsg.embeddings;
            errorMsgObj = existingAnswerMsg;
          }
        }

        if (!embeddings || !embeddings.length) {
          setLoadingStatus('Embedding');
          embeddings = await getEmbeddings(
            sendHttpReq,
            inputMsg,
            settings.embeddingProfile,
            settings.embeddingMaxTokens
          );
        }

        setLoadingStatus('Vector');
        const { rows, colnames } = await sendDsReq(embeddings);
        let relevantRows = rows;

        const convertParam = (updatedRows: string[][]): AddParamArg[] => {
          const includeAllParams = settings.paramSource === 'all';
          return colnames.map((name, i) => {
            const value_list = [
              ...new Set(
                includeAllParams
                  ? updatedRows.map((row) => row[i])
                  : [updatedRows[0][i]]
              ),
            ];

            return {
              key: includeAllParams ? `${name}.data` : name,
              value: value_list.join(','),
              value_list,
              source_visual_id: reportId,
              source_dataset_id: datasetId,
              valueType: includeAllParams ? 'multi' : 'single',
            };
          });
        };

        const chunkCb = (chunk: string) => {
          setMessages((currentMessages) => {
            const targetMessage =
              errorId && errorMsgObj
                ? errorMsgObj
                : currentMessages[currentMessages.length - 1];

            // If the target message is an answer, update it
            if (targetMessage && targetMessage.type === 'answer') {
              return currentMessages.map((msg) => {
                if (msg.id === targetMessage.id) {
                  // If we had an error message, remove hasError & reset
                  if (msg.type === 'answer' && msg.hasError) {
                    const { hasError, ...rest } = msg;
                    return {
                      ...rest,
                      text: chunk,
                      isChunk: true,
                    };
                  }
                  // Otherwise, keep appending the chunk
                  return { ...msg, text: msg.text + chunk };
                }

                // Not the target => return unchanged
                return msg;
              });
            } else {
              // If last msg is not an answer, we create a new answer
              const newAnswer: Answer = {
                type: 'answer',
                id: uuid.randomUUID(),
                text: chunk,
                sender: 'bot',
                tooltips: [],
                ts: new Date(),
                embeddings: [],
                isChunk: true,
                shouldAnimate: true,
              };
              return [...currentMessages, newAnswer];
            }
          });
        };

        setLoadingStatus('Completion');
        let answer = await getCompletion(
          sendHttpReq,
          chunkCb,
          getEmbeddingSourceInfo(relevantRows),
          previousMessages,
          inputMsg,
          settings
        );

        if (
          settings.paramSource !== 'none' &&
          settings.paramSelection === 'all'
        ) {
          addParams(convertParam(relevantRows));
        }

        // Finalize the answer by appending to the last chunked message if applicable
        setMessages((currentMessages) => {
          if (errorId) {
            // replace the error
            const updatedMessages = currentMessages.map((msg) => {
              if (msg.type === 'answer' && msg.errorId === errorId) {
                let finalText = msg.isChunk ? msg.text + answer : answer;

                // If 'used-by-llm' logic applies even in error scenario:
                const { finalText: newText, updatedRows } = finalizeAnswer({
                  answerText: finalText,
                  relevantRows,
                  settings: {
                    paramSource: settings.paramSource,
                    paramSelection: settings.paramSelection,
                  },
                  addParams,
                  convertParam,
                });
                relevantRows = updatedRows;
                finalText = newText;

                return {
                  ...msg,
                  text: finalText,
                  hasError: false,
                  isChunk: false,
                  tooltips: getTooltips(relevantRows),
                  embeddings,
                  shouldAnimate: true,
                };
              }
              return msg;
            });

            downloadInfo = { conversation: updatedMessages };
            setDownloadInfo(downloadInfo);
            saveConversationToSessionStorage(updatedMessages);
            return updatedMessages;
          }

          const lastMessage = currentMessages[currentMessages.length - 1];

          if (
            lastMessage &&
            lastMessage.type === 'answer' &&
            lastMessage.isChunk
          ) {
            // Append the final part to the last chunked answer
            const updatedMessages = currentMessages.map((msg, index) => {
              if (index === currentMessages.length - 1) {
                const fullAnswerText = msg.text + answer;

                const { finalText, updatedRows } = finalizeAnswer({
                  answerText: fullAnswerText,
                  relevantRows,
                  settings: {
                    paramSource: settings.paramSource,
                    paramSelection: settings.paramSelection,
                  },
                  addParams,
                  convertParam,
                });
                relevantRows = updatedRows;

                return {
                  ...msg,
                  text: finalText,
                  isChunk: false,
                  tooltips: getTooltips(relevantRows),
                  embeddings,
                };
              }
              return msg;
            });

            // Update downloadInfo and save to sessionStorage
            downloadInfo = { conversation: updatedMessages };
            setDownloadInfo(downloadInfo);
            saveConversationToSessionStorage(updatedMessages);

            return updatedMessages;
          } else {
            // No streaming scenario, or no partial chunks => just append an answer
            // If paramSource === 'used-by-llm', we do the same citation logic here
            let finalText = answer;

            const { finalText: newText, updatedRows } = finalizeAnswer({
              answerText: finalText,
              relevantRows,
              settings: {
                paramSource: settings.paramSource,
                paramSelection: settings.paramSelection,
              },
              addParams,
              convertParam,
            });
            relevantRows = updatedRows;

            const newAnswer: Answer = {
              type: 'answer',
              id: uuid.randomUUID(),
              text: newText,
              sender: 'bot',
              tooltips: getTooltips(relevantRows),
              ts: new Date(),
              embeddings,
              shouldAnimate: true,
            };
            const updated = [...currentMessages, newAnswer];

            // Update downloadInfo and save to sessionStorage
            downloadInfo = { conversation: updated };
            setDownloadInfo(downloadInfo);
            saveConversationToSessionStorage(updated);
            return updated;
          }
        });
      } catch (error) {
        const errorMessage =
          error instanceof Error ? error.message : String(error);

        alertMessage(errorMessage);

        if (!errorId) {
          const errorAnswer: Answer = {
            type: 'answer',
            id: uuid.randomUUID(),
            text: 'Something went wrong, please try again later.',
            sender: 'bot',
            ts: new Date(),
            hasError: true,
            tooltips: [],
            errorId: msgId,
            embeddings,
            shouldAnimate: true,
          };
          newMessages = [...newMessages, errorAnswer];
          saveConversationToSessionStorage(newMessages);
          setMessages(newMessages);
        }
      } finally {
        setLoadingStatus('None');
      }
    },
    [
      input,
      uuid,
      messages,
      sendDsReq,
      sendHttpReq,
      getEmbeddingSourceInfo,
      settings,
      reportId,
      datasetId,
      addParams,
      setDownloadInfo,
      saveConversationToSessionStorage,
      getTooltips,
      alertMessage,
    ]
  );

  const resetMessages = useCallback(() => {
    setMessages(getInitialMessages());
    hideTooltips();

    sessionStorage.removeItem(sessionKey);
  }, [getInitialMessages, sessionKey]);

  const hideTooltips = () => {
    setTooltips(null);
  };

  return {
    input,
    setInput,
    messages,
    sendMessage,
    resetMessages,
    loadingStatus,
    setTooltips,
    hideTooltips,
    tooltips,
  };
};
