import React, { useEffect, useState } from 'react';

import { getGlobalArc } from 'utils/get-global-arc';
import { NewAppButton } from 'components/new-app-button/NewAppButton';
import { NewDashboardButton } from 'components/new-dashboard-button/NewDashboardButton';
import { ChromeBrowserSuggestion } from 'components/chrome-browser-suggestion/ChromeBrowserSuggestion';
import { LastLogin } from 'components/last-login/LastLogin';
import { IncompatibleBrowserAlertModal } from 'components/incompatible-browser-alert-modal/IncompatibleBrowserAlertModal';
import { NavigationBar } from 'components/navigation-bar/NavigationBar';
import { vizCookie } from 'services/viz-cookie';

import classes from './HomePage.module.scss';
import { Blocks } from './components/blocks/Blocks';
import { GetStarted } from './components/get-started/GetStarted';
import { Overviews } from './components/overviews/Overviews';
import { Learn } from './components/learn/Learn';
import { NewSqlButton } from './components/new-sql-button/NewSqlButton';
import type { GetStartedState } from './types/get-started-state.type';
import { usePrefetchResources } from './hooks/use-prefetch-resources';
import { CreatedRecently } from './components/created-recently/CreatedRecently';
import { useReportCreationPermissions } from 'hooks/use-dataset-permissions';

type HomePageProps = {
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
};

const arc = getGlobalArc();

const isExploreStyleEnabled = arc.frontend.featureFlipper.getValueById(
  'enable_explore_style'
);

const getStartedCookieKey = 'getstarted';

const getStartedCookie =
  vizCookie.getJson<GetStartedState>(getStartedCookieKey);

const isCMLSyncEnabled =
  arc.frontend.featureFlipper.getValueById('enable_cml_sync');

export const HomePage = ({
  isAlreadySeen,
  shouldShowModal,
}: HomePageProps): JSX.Element => {
  usePrefetchResources();

  const { canCreateDashboard, canCreateApp } = useReportCreationPermissions();

  const [getStartedState, setGetStartedState] = useState<GetStartedState>({
    isOpen: !!(getStartedCookie?.isOpen ?? isCMLSyncEnabled),
    isVisible: !!getStartedCookie?.isVisible,
  });

  useEffect(() => {
    vizCookie.setJson(getStartedCookieKey, getStartedState);
  }, [getStartedState]);

  return (
    <div className={classes.container}>
      <IncompatibleBrowserAlertModal />
      <NavigationBar
        isAlreadySeen={isAlreadySeen}
        shouldShowModal={shouldShowModal}
        activeTabs={{
          main: 'home',
        }}
      />

      <div className={classes.main}>
        <div className={classes.leftSide}>
          {!isExploreStyleEnabled && <ChromeBrowserSuggestion />}

          <GetStarted
            getStartedState={getStartedState}
            setGetStartedState={setGetStartedState}
            productName={arc.frontend.globals.product.name}
            canCreateDashboard={canCreateDashboard}
            canCreateApp={canCreateApp}
          />

          <Overviews />

          <Blocks />
        </div>

        {!isExploreStyleEnabled && (
          <div className={classes.rightSide}>
            <div className={classes.rightSideTopSection}>
              {canCreateApp && (
                <div className={classes.buttons}>
                  <NewSqlButton />

                  {canCreateDashboard && (
                    <NewDashboardButton testId="home-new-dashboard" />
                  )}

                  <NewAppButton />
                </div>
              )}

              <CreatedRecently />

              <Learn
                forceOpenGetStartedAccordion={() => {
                  setGetStartedState({ isOpen: true, isVisible: true });
                }}
              />
            </div>

            <LastLogin />
          </div>
        )}
      </div>
    </div>
  );
};
