import React, { useEffect, useRef, useState } from 'react';

import { useResize } from 'hooks/use-resize';

import classes from './Block.module.scss';
import type { LoadingPlaceholderContainerProps } from '../loading-placeholder-container/LoadingPlaceholderContainer';
import type { LoadingStripesPlaceholderProps } from '../loading-stripes-placeholder/LoadingStripesPlaceholder';
import type { SqlQueryTileProps } from '../sql-query-tile/SqlQueryTile';
import type { DatasetTileProps } from '../dataset-tile/DatasetTile';
import type { DataConnectionTileProps } from '../data-connection-tile/DataConnectionTile';
import type { ReportTileProps } from '../../../../components/report-tile/ReportTile';

export type TileProps =
  | SqlQueryTileProps
  | DatasetTileProps
  | DataConnectionTileProps
  | ReportTileProps;

export type BlockProps<T extends TileProps> = {
  title: string;
  order: number;
  emptyContent?: JSX.Element | false;
  fetch: () => Promise<T[]>;
  Placeholder: (
    props: LoadingPlaceholderContainerProps | LoadingStripesPlaceholderProps
  ) => JSX.Element;
  tileConfig: {
    Component: (props: T) => JSX.Element;
    width: number;
    height: number;
  };
  modals?: JSX.Element;
};

const TILE_GAP = 15;

export const Block = <Tile extends TileProps>({
  fetch,
  tileConfig,
  title,
  order,
  emptyContent,
  Placeholder,
  modals,
}: BlockProps<Tile>): JSX.Element => {
  const [isLoading, setIsLoading] = useState(false);
  const [tiles, setTiles] = useState<Tile[]>([]);

  const tilesContainerRef = useRef<HTMLDivElement>(null);
  const [numberOfVisibleTiles, setNumberOfVisibleTiles] = useState(0);
  const visibleTiles = tiles.slice(0, numberOfVisibleTiles);

  useEffect(() => {
    setIsLoading(true);
    fetch()
      .then((tiles) => {
        setTiles(tiles);
        setIsLoading(false);
      })
      .catch((error) => {
        console.error(error);
        setIsLoading(false);
      });
  }, [fetch]);

  useResize(() => {
    if (tilesContainerRef.current?.offsetWidth) {
      setNumberOfVisibleTiles(
        Math.floor(
          tilesContainerRef.current.offsetWidth / (tileConfig.width + TILE_GAP)
        )
      );
    }
  });

  if (!isLoading && tiles.length === 0 && !emptyContent) {
    return <></>;
  }

  let tilesElement: JSX.Element;
  if (isLoading) {
    tilesElement = (
      <>
        {Array.from({ length: numberOfVisibleTiles }, (_, index) => (
          <Placeholder
            key={index}
            width={`${tileConfig.width}px`}
            height={`${tileConfig.height}px`}
          />
        ))}
      </>
    );
  } else {
    if (tiles.length === 0) {
      tilesElement = <div className={classes.emptyContent}>{emptyContent}</div>;
    } else {
      tilesElement = (
        <>
          {visibleTiles.map((tile) => (
            <div
              key={`${title}-${tile.id}${
                'report' in tile ? `-${tile.report.type}` : ''
              }`}
              className={classes.tile}
              style={{
                width: tileConfig.width,
                height: tileConfig.height,
              }}
            >
              <tileConfig.Component {...tile} />
            </div>
          ))}
        </>
      );
    }
  }

  return (
    <div style={{ order }}>
      <h4 className="mt-0">{title}</h4>

      <div
        ref={tilesContainerRef}
        className="d-flex align-items-center p-0 m-0 flex-wrap"
        style={{ gap: TILE_GAP }}
      >
        {tilesElement}
      </div>

      {modals}
    </div>
  );
};
