import React from 'react';

import { getGlobalArc } from 'utils/get-global-arc';
import { VizLink } from 'components/viz-link/VizLink';

import {
  ReportTile,
  ReportTileProps,
} from '../../../../components/report-tile/ReportTile';
import { LoadingImagePlaceholder } from '../loading-image-placeholder/LoadingImagePlaceholder';
import { LoadingStripesPlaceholder } from '../loading-stripes-placeholder/LoadingStripesPlaceholder';
import { DataConnectionTile } from '../data-connection-tile/DataConnectionTile';
import { DatasetTile } from '../dataset-tile/DatasetTile';
import { SqlQueryTile } from '../sql-query-tile/SqlQueryTile';
import { homePageApi } from 'pages/home/api';
import type { DataConnection } from 'types/data-connection.type';
import type { SqlQuery } from 'types/sql-query.type';
import { ImportDataModal } from 'components/import-data-modal/ImportDataModal';
import { BlockProps, TileProps } from '../block/Block';
import { HomepageDataset } from 'pages/home/types/home-page-dataset.type';
import { getAllDataConnections } from 'apis/get-all-data-connections';
import { getAllSqlQueries } from 'apis/get-all-sql-queries';

const arc = getGlobalArc();

const isExploreStyleEnabled = arc.frontend.featureFlipper.getValueById(
  'enable_explore_style'
);

const isClouderaStyleEnabled = arc.frontend.featureFlipper.getValueById(
  'enable_cloudera_style'
);

const canCreateConnection =
  arc.frontend.globals.permissions.system.includes('sys_editperm') &&
  !isExploreStyleEnabled;
const newConnectionLink = `${arc.frontend.globals.urls.dataUrl}#connection-modal`;

const reportBlockConfig = {
  Placeholder: LoadingImagePlaceholder,
  tileConfig: {
    Component: ReportTile,
    width: 182,
    height: 196,
  },
};

const nonReportBlockConfig = {
  width: 215,
  height: 112,
};

type BlockWithoutOrder<T extends TileProps> = Omit<BlockProps<T>, 'order'>;

const recentQueries: BlockWithoutOrder<SqlQuery> = {
  title: 'Recent Queries',
  fetch: getAllSqlQueries,
  tileConfig: {
    ...nonReportBlockConfig,
    Component: SqlQueryTile,
  },
  Placeholder: LoadingStripesPlaceholder,
  emptyContent: (
    <>
      No Queries have been run yet. Run a&nbsp;
      <VizLink href={`${arc.frontend.globals.urls.sqlUrl}#sql-tab`}>
        New SQL
      </VizLink>
      &nbsp;query.
    </>
  ),
};

const recentDataConnections: BlockWithoutOrder<DataConnection> = {
  title: 'Recent Connections',
  fetch: () => getAllDataConnections(30),
  tileConfig: {
    ...nonReportBlockConfig,
    Component: DataConnectionTile,
  },
  Placeholder: LoadingStripesPlaceholder,
  emptyContent: (
    <>
      No connections are available.{' '}
      {canCreateConnection && (
        <>
          Create a&nbsp;
          <VizLink href={newConnectionLink}>New Connection</VizLink>
        </>
      )}
    </>
  ),
  modals: <ImportDataModal />,
};

const recentDatasets: BlockWithoutOrder<HomepageDataset> = {
  title: 'Recent Datasets',
  fetch: homePageApi.getDatasets,
  tileConfig: {
    ...nonReportBlockConfig,
    Component: DatasetTile,
  },
  Placeholder: LoadingStripesPlaceholder,
  emptyContent: (
    <>
      No Datasets are available.{' '}
      {arc.frontend.globals.canUserExploreAnyDataconnections && (
        <>
          Create a&nbsp;
          <VizLink href={`${arc.frontend.globals.urls.dataUrl}#dataset-modal`}>
            New Dataset
          </VizLink>
        </>
      )}
    </>
  ),
};

const favorites: BlockWithoutOrder<ReportTileProps> = {
  ...reportBlockConfig,
  title: 'Favorites',
  fetch: homePageApi.getFavoriteDashboards,
};

const recentDashboardsAndApps: BlockWithoutOrder<ReportTileProps> = {
  ...reportBlockConfig,
  title: isExploreStyleEnabled
    ? 'Recent Dashboards and Apps'
    : 'Last Viewed by You',
  fetch: homePageApi.getLastViewedReportsByUser,
  emptyContent: isClouderaStyleEnabled && (
    <>
      No Dashboards have been created. Create a&nbsp;
      <VizLink
        href={`${arc.frontend.globals.urls.reportBuilderUrl}/0?report_type=dashboard`}
      >
        New Dashboard
      </VizLink>
      .
    </>
  ),
};

const recentlyCreatedByYou: BlockWithoutOrder<ReportTileProps> = {
  ...reportBlockConfig,
  title: 'Recently Created by You',
  fetch: homePageApi.getRecentlyCreatedDashboardsByUser,
};

const overallMostPopular: BlockWithoutOrder<ReportTileProps> = {
  ...reportBlockConfig,
  title: 'Overall Most Popular',
  fetch: homePageApi.getMostPopularReports,
};

const sampleDashboards: BlockWithoutOrder<ReportTileProps> = {
  ...reportBlockConfig,
  title: 'Samples',
  fetch: homePageApi.getSamples,
};

const blockConfigs = {
  recentQueries,
  recentDataConnections,
  recentDatasets,
  favorites,
  recentDashboardsAndApps,
  recentlyCreatedByYou,
  overallMostPopular,
  sampleDashboards,
};

type BlockName = keyof typeof blockConfigs;
let blockNameByOrder: Record<number, BlockName>;

if (isExploreStyleEnabled && isClouderaStyleEnabled) {
  blockNameByOrder = {
    1: 'recentQueries',
    2: 'recentDashboardsAndApps',
    3: 'recentDatasets',
    4: 'recentDataConnections',
    5: 'favorites',
    6: 'sampleDashboards',
  };

  if (!arc.frontend.globals.canUserExploreAnyDataconnections) {
    delete blockNameByOrder[1];
  }
} else if (!isExploreStyleEnabled && isClouderaStyleEnabled) {
  blockNameByOrder = {
    1: 'recentQueries',
    2: 'recentDataConnections',
    3: 'recentDatasets',
    4: 'favorites',
    5: 'recentDashboardsAndApps',
    6: 'recentlyCreatedByYou',
    7: 'overallMostPopular',
    8: 'sampleDashboards',
  };

  if (!arc.frontend.globals.canUserExploreAnyDataconnections) {
    delete blockNameByOrder[1];
  }
} else if (isExploreStyleEnabled && !isClouderaStyleEnabled) {
  blockNameByOrder = {
    1: 'recentDashboardsAndApps',
    2: 'favorites',
    3: 'sampleDashboards',
  };
} else {
  blockNameByOrder = {
    1: 'favorites',
    2: 'recentDashboardsAndApps',
    3: 'recentlyCreatedByYou',
    4: 'overallMostPopular',
    5: 'sampleDashboards',
  };
}

export const getBlocks = () =>
  Object.entries(blockNameByOrder).map(
    ([order, blockName]) =>
      ({
        ...blockConfigs[blockName]!,
        order: +order,
      } as BlockProps<TileProps>)
  );
