import React from 'react';

import { getGlobalArc } from 'utils/get-global-arc';
import { VizLink } from 'components/viz-link/VizLink';
import type { GetStartedItem } from 'pages/home/types/get-started-item.type';

import { ResponsiveTwoColumnLayout } from '../responsive-two-column-layout/ResponsiveTwoColumnLayout';
import classes from './get-get-started-items.module.scss';
import classNames from 'classnames';
import { getBlankTargetUrl } from 'arc/misc/misc';

const arc = getGlobalArc();

const isExploreStyleEnabled = arc.frontend.featureFlipper.getValueById(
  'enable_explore_style'
);

export const getGetStartedItems = ({
  productName,
  canCreateApp,
  canCreateDashboard,
}: {
  productName: string;
  canCreateApp: boolean;
  canCreateDashboard: boolean;
}): GetStartedItem[] => [
  {
    title: isExploreStyleEnabled ? 'Sync Connections' : 'Create a Connection',
    content: isExploreStyleEnabled ? (
      <ResponsiveTwoColumnLayout
        img={{
          src: `${arc.frontend.globals.urls.staticUrl}/img/get-started/data-connection-sync.png`,
          alt: 'Add workload password',
        }}
        text={
          <>
            <p>
              Data Connections may already be setup for you to use on the{' '}
              <VizLink href={arc.frontend.globals.urls.dataUrl}>
                datasets
              </VizLink>{' '}
              page.
            </p>

            <p>
              If they are not, you will need to set up your workspace password:
            </p>

            <ol>
              <li>On the Cloudera Management Console, select your username</li>
              <li>
                View your profile, and set a workload password using the "Set
                Workload Password" link
              </li>
              <li>In your ML Workspace, go to "User Settings" page</li>
              <li>
                Under "Environmental Variables" tab, set your
                "WORKLOAD_PASSWORD" variable to the same password from your
                Cloudera Management Console profile
              </li>
              <li>
                Restart the "Data Discovery and Visualization" application in
                Cloudera AI
              </li>
            </ol>
          </>
        }
      />
    ) : (
      <>
        Data Connections may already be setup for you to use. If you need to
        create a new connection,{' '}
        <VizLink href={arc.frontend.globals.urls.dataUrl}>click here</VizLink>.
      </>
    ),
    icon: (
      <i
        className={classNames(
          classes.icon,
          `fa ${isExploreStyleEnabled ? 'fa-link' : 'fa-refresh'}`
        )}
      />
    ),
  },
  {
    title: isExploreStyleEnabled ? 'Explore with SQL' : 'Create a Dataset',
    content: isExploreStyleEnabled ? (
      <ResponsiveTwoColumnLayout
        img={{
          src: `${arc.frontend.globals.urls.staticUrl}/img/get-started/sql.png`,
          alt: 'run sql queries',
        }}
        text={
          <p>
            Run queries on the{' '}
            <a href={`${arc.frontend.globals.urls.sqlUrl}#sql-tab`}>SQL</a> page
            and directly create datasets or dashboards from results
          </p>
        }
      />
    ) : (
      <ResponsiveTwoColumnLayout
        img={{
          src: `${arc.frontend.globals.urls.staticUrl}/img/get-started/dataset.png`,
          alt: 'create a new dataset',
        }}
        text={
          <>
            A Dataset provides access to specific tables in the data store.
            Datasets are defined on a specific connection. Once created, you
            then can create dashboards based on the dataset.
            <ol>
              <li>
                Click the <strong>Data</strong> tab on the main navigation bar.
              </li>

              <li>
                Click the <strong>Connection Explorer</strong> tab.
              </li>

              <li>Select the database containing the data.</li>

              <li>Select the table containing the data.</li>

              <li>
                Click the <strong>New dataset</strong> link.
              </li>
            </ol>
            <a href={arc.frontend.globals.urls.dataUrl}>
              <span className="fa fa-plus"></span> Create a New Dataset
            </a>
          </>
        }
      />
    ),
    icon: <i className={classNames(classes.icon, 'fa fa-cubes')} />,
  },
  {
    title: 'Import Data',
    content: (
      <ResponsiveTwoColumnLayout
        img={{
          src: `${arc.frontend.globals.urls.staticUrl}/img/get-started/importdata.png`,
          alt: 'create a new dataset',
        }}
        text={
          <>
            Add (additional) data to an existing connection and automatically
            create a table behind the scenes in the database is easy by
            importing the data in csv format.
            <ol>
              <li>
                Click the <strong>Add Data icon</strong> on a the Connection
                tile on Home page or go to the Data page and click the &nbsp;
                <strong> ADD DATA</strong> button. The "Add Data" modal should
                appear.
              </li>

              <li>
                Select the data source - <strong>CSV or URL</strong>.
              </li>

              <li>
                Provide the data's source - a csv file location, or an url.
              </li>

              <li>
                Press <strong>GET DATA</strong>
              </li>

              <li>
                On the data import surface set your desired import
                configurations: target database, the table name under you wish
                to find it later and other data related settings.
              </li>

              <li>
                Click the <strong>CONFIRM IMPORT</strong> button to finalize the
                process!
              </li>
            </ol>
            <a
              href={getBlankTargetUrl(arc.frontend.globals.urls.dataUrl)}
              target="_blank"
              rel="noreferrer"
            >
              <span className="fa fa-plus"></span> Import Data
            </a>
          </>
        }
      />
    ),
    icon: <i className={classNames(classes.icon, 'fa fa-align-justify')} />,
  },
  {
    title: 'Create a Dashboard',
    content: (
      <ResponsiveTwoColumnLayout
        img={{
          src: `${arc.frontend.globals.urls.staticUrl}/img/get-started/dashboard.png`,
          alt: 'create a new dashboard',
        }}
        text={
          <>
            Dashboards are a collection of visuals and can also include a
            variety of filter options. You can style dashboards to fit your
            branding.
            <ol>
              <li>
                Click the <strong>New Dashboard</strong> button on the main{' '}
                <strong>Home</strong> interface, in the top right corner on the
                side panel.
              </li>
              <li>
                Alternatively, on the{' '}
                <strong>{isExploreStyleEnabled ? 'Dataset' : 'Data'}</strong>{' '}
                interface, click the new dashboard icon next to the name of a
                particular dataset.
              </li>
              <li>
                Create your dashboard in the <strong>Dashboard Designer</strong>
                .
                <br />
                Add visuals by selecting them from the Visuals menu on the
                right.
                <ol type="a">
                  <li>
                    Add visuals by selecting them from the{' '}
                    <strong>Visuals</strong> menu on the right. Alternatively,
                    you can create a new visual by clicking the{' '}
                    <strong>New Visual</strong> button.
                    <br />
                    You cannot save an empty dashboard. You must add at least
                    one visual to a sheet before saving the dashboard.
                  </li>
                  <li>
                    Add a title for the dashboard, and an optional subtitle.
                  </li>
                  <li>
                    [Optional] You can customize <strong>Settings</strong>,{' '}
                    <strong>Styles</strong>, and <strong>Custom Styles</strong>{' '}
                    for the dashboard in the <strong>Dashboard Designer</strong>{' '}
                    menu.
                  </li>
                </ol>
              </li>
              <li>
                Click <strong>Save</strong>
              </li>
            </ol>
            {canCreateDashboard && (
              <VizLink
                href={`${arc.frontend.globals.urls.reportBuilderUrl}/0?report_type=dashboard`}
              >
                <span className="fa fa-plus"></span> Create a New Dashboard
              </VizLink>
            )}
          </>
        }
      />
    ),
    icon: <i className={classNames(classes.icon, 'fa fa-dashboard')} />,
  },
  {
    title: isExploreStyleEnabled ? 'Cloudera AI Notebook' : 'Create an App',
    content: isExploreStyleEnabled ? (
      <ResponsiveTwoColumnLayout
        img={{
          src: `${arc.frontend.globals.urls.staticUrl}/img/get-started/notebook.png`,
          alt: 'run sql queries',
        }}
        text={
          <p>
            Do further analysis of data through Cloudera Machine Learning
            session
          </p>
        }
      />
    ) : (
      <ResponsiveTwoColumnLayout
        img={{
          src: `${arc.frontend.globals.urls.staticUrl}/img/get-started/app.png`,
          alt: 'create a new app',
        }}
        text={
          <>
            An app is a standalone collection of dashboards. Apps are useful
            when you want users to only access the dashboards that are relevant
            to a particular group or function. They can be styled to fit your
            branding, including styling the top menu bar and adding your own
            logo.
            <br />
            The dashboards you want to include in an app must be created before
            creating the app.
            <ol>
              <li>
                Click the <strong>New App</strong> button on the main Home
                interface, in the top right corner on the side panel.
                Alternatively, click <strong>New App</strong> on the main{' '}
                <strong>Visuals</strong> interface, in the top left corner.
              </li>

              <li>
                Create your app in the <strong>App Designer</strong>.
                <ol type="a">
                  <li>
                    Enter a name for your app in the <strong>App Name</strong>{' '}
                    text box.
                  </li>

                  <li>In the highlighted empty tab, enter the tab title.</li>

                  <li>Add a dashboard to this tab.</li>
                </ol>
              </li>

              <li>
                Click <strong>Save</strong>.
              </li>
            </ol>
            {canCreateApp && (
              <VizLink href={arc.frontend.globals.urls.appGroupEditUrl}>
                <span className="fa fa-plus"></span> Create New App
              </VizLink>
            )}
          </>
        }
      />
    ),
    icon: isExploreStyleEnabled ? (
      <i className={classes.icon}>
        <svg
          viewBox="0 0 24 24"
          width="1em"
          height="1em"
          className="cdp-icon-cli"
        >
          <path
            d="M4.414 4L3 5.3l7.293 6.7L3 18.702 4.414 20l8.708-8-8.708-8zM13 20h8v-1.882h-8V20z"
            fill="white"
          ></path>
        </svg>
      </i>
    ) : (
      <i className={classNames(classes.icon, 'sym-app')} />
    ),
  },
  {
    title: "What's Next?",
    content: (
      <>
        Take a look at {productName}{' '}
        <VizLink
          href={arc.frontend.globals.documentation.url}
          target="_blank"
          rel="noreferrer"
        >
          documentation
        </VizLink>
        .
      </>
    ),
    icon: (
      <i className={classNames(classes.icon, 'fa fa-arrow-circle-right')} />
    ),
  },
];
