import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import * as deleteJobs from 'apis/jobs/delete-jobs';
import * as getJobs from 'apis/jobs/get-jobs';
import * as getLatestRuns from 'apis/jobs/get-latest-runs';
import * as pauseOrResume from 'apis/jobs/pause-or-resume-jobs';
import * as rerun from 'apis/jobs/rerun-jobs';
import * as getScheduleGroups from 'apis/jobs/schedule-groups/get-schedule-groups';
import * as getUsers from 'pages/visuals-page/apis/get-users';
import * as React from 'react';
import { JobOverviewPage } from './JobOverviewPage';

configure({ testIdAttribute: 'data-test-id' });

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: { previewCustomEmailUrl: 'customUrl' },
        settings: {
          emailHostUser: 'test@test.com',
          uploadMaxrowsDefault: 1000000,
        },
        emailAddresses: '[]',
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

jest.mock('components/navigation-bar/NavigationBar', () => ({
  NavigationBar: () => null,
}));
jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

beforeAll(() => {
  jest.spyOn(pauseOrResume, 'pauseOrResumeJobs').mockImplementation(jest.fn());
  jest.spyOn(rerun, 'rerunJobs').mockImplementation(jest.fn());
  jest.spyOn(deleteJobs, 'deleteJobs').mockImplementation(jest.fn());
});

beforeEach(() => {
  jest.spyOn(getScheduleGroups, 'getScheduleGroups').mockResolvedValue([]);
  jest.spyOn(getJobs, 'getJobs').mockResolvedValue({ data: [], total: 0 });
  jest.spyOn(getLatestRuns, 'getLatestRuns').mockResolvedValue([]);
  jest.spyOn(getUsers, 'getUsers').mockResolvedValue([]);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('JobOverviewPage component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <JobOverviewPage shouldShowModal={false} isAlreadySeen={false} />
      </QueryClientProvider>
    );
  });

  describe('has three tabs', () => {
    it('automatically displays the Runs tab', () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <JobOverviewPage shouldShowModal={false} isAlreadySeen={false} />
        </QueryClientProvider>
      );

      const runsTab = screen.getByTestId('runsTab');
      expect(runsTab).toBeInTheDocument();
      expect(screen.getByTestId('scheduledJobsTab')).toBeInTheDocument();
      expect(screen.getByTestId('intervalsTab')).toBeInTheDocument();

      expect(runsTab).toHaveClass('active');
    });

    it('switching between tabs should change the active table', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <JobOverviewPage shouldShowModal={false} isAlreadySeen={false} />
        </QueryClientProvider>
      );

      const runsTab = screen.getByRole('tab', { name: 'Runs (Latest)' });
      const jobsTab = screen.getByRole('tab', { name: 'Scheduled Jobs' });
      const intervalsTab = screen.getByRole('tab', { name: 'Intervals' });

      expect(runsTab).toHaveClass('active');
      expect(jobsTab).not.toHaveClass('active');
      expect(intervalsTab).not.toHaveClass('active');

      fireEvent.click(jobsTab);
      await waitFor(() => {
        expect(runsTab).not.toHaveClass('active');
      });
      expect(jobsTab).toHaveClass('active');
      expect(intervalsTab).not.toHaveClass('active');

      fireEvent.click(intervalsTab);
      expect(runsTab).not.toHaveClass('active');
      await waitFor(() => {
        expect(jobsTab).not.toHaveClass('active');
      });
      expect(intervalsTab).toHaveClass('active');
    });
  });
});
