import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import * as createScheduleGroup from 'apis/jobs/schedule-groups/create-schedule-group';
import * as editScheduleGroup from 'apis/jobs/schedule-groups/edit-schedule-group';
import * as React from 'react';
import { IntervalModal } from './IntervalModal';
import { mockScheduleGroups } from 'pages/job-page/utils/mocks';

jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

configure({ testIdAttribute: 'data-test-id' });

beforeAll(() => {
  jest
    .spyOn(createScheduleGroup, 'createScheduleGroup')
    .mockImplementation(jest.fn());

  jest
    .spyOn(editScheduleGroup, 'editScheduleGroup')
    .mockImplementation(jest.fn());
});

describe('IntervalModal component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <button>Interval</button>
          </ClouderaDialogTrigger>
          <IntervalModal />
        </ClouderaDialog>
      </QueryClientProvider>
    );

    fireEvent.click(screen.getByRole('button', { name: 'Interval' }));
  });

  describe('when it is rendered without a scheduleGroup prop', () => {
    it('opens the modal in create mode', () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <button>Interval</button>
            </ClouderaDialogTrigger>
            <IntervalModal />
          </ClouderaDialog>
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Interval' }));
      expect(screen.getByText('Create Interval')).toBeInTheDocument();
      expect(screen.queryByText('Edit Interval')).not.toBeInTheDocument();
    });

    describe('and the form is filled with the correct data', () => {
      it('saves the new interval', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <button>Interval</button>
              </ClouderaDialogTrigger>
              <IntervalModal />
            </ClouderaDialog>
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('button', { name: 'Interval' }));
        fireEvent.change(screen.getByTestId('intervalName'), {
          target: { value: 'test' },
        });

        fireEvent.click(screen.getByTestId('dropdown-input-minutes'));
        await waitFor(() => {
          expect(screen.getByTestId('popover-0')).toBeInTheDocument();
        });
        fireEvent.click(screen.getByTestId('popover-15'));

        fireEvent.click(screen.getByTestId('saveInterval'));
        await waitFor(() => {
          expect(createScheduleGroup.createScheduleGroup).toHaveBeenCalledWith({
            interval: '0,15 * * * *',
            name: 'test',
          });
        });
      });
    });
  });

  describe('when it is rendered with a scheduleGroup prop', () => {
    it('opens the modal in edit mode', () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <button>Interval</button>
            </ClouderaDialogTrigger>
            <IntervalModal scheduleGroup={mockScheduleGroups[0]} />
          </ClouderaDialog>
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Interval' }));
      expect(screen.getByText('Edit Interval')).toBeInTheDocument();
      expect(screen.queryByText('Create Interval')).not.toBeInTheDocument();
    });

    it('displays a notification about the affected jobs', () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <button>Interval</button>
            </ClouderaDialogTrigger>
            <IntervalModal scheduleGroup={mockScheduleGroups[0]} />
          </ClouderaDialog>
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Interval' }));
      expect(screen.getByTestId('notification-title')).toHaveTextContent(
        'Will affect the following jobs'
      );
      expect(screen.getByTestId('notification-message')).toHaveTextContent(
        'dependent'
      );
    });

    describe('and the form is filled with the correct data', () => {
      it('edits interval', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <button>Interval</button>
              </ClouderaDialogTrigger>
              <IntervalModal scheduleGroup={mockScheduleGroups[0]} />
            </ClouderaDialog>
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('button', { name: 'Interval' }));

        fireEvent.click(screen.getByTestId('dropdown-input-minutes'));
        await waitFor(() => {
          expect(screen.getByTestId('popover-0')).toBeInTheDocument();
        });
        fireEvent.click(screen.getByTestId('popover-15'));

        fireEvent.click(screen.getByTestId('saveInterval'));
        await waitFor(() => {
          expect(editScheduleGroup.editScheduleGroup).toHaveBeenCalledWith({
            interval: '0,15 * * * *',
            id: mockScheduleGroups[0].id,
            name: mockScheduleGroups[0].name,
          });
        });
      });
    });
  });

  describe('when the form is submitted without a title', () => {
    it('shows a name error', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <button>Interval</button>
            </ClouderaDialogTrigger>
            <IntervalModal />
          </ClouderaDialog>
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Interval' }));

      fireEvent.click(screen.getByTestId('dropdown-input-period'));
      await waitFor(() => {
        expect(screen.getByTestId('popover-day')).toBeInTheDocument();
      });
      fireEvent.click(screen.getByTestId('popover-day'));

      fireEvent.click(screen.getByTestId('dropdown-input-hours'));
      await waitFor(() => {
        expect(screen.getByTestId('popover-0')).toBeInTheDocument();
      });
      fireEvent.click(screen.getByTestId('popover-6'));

      fireEvent.click(screen.getByTestId('saveInterval'));
      await waitFor(() => {
        expect(screen.getByTestId('nameError')).toHaveTextContent(
          'Name is a required field'
        );
      });
    });
  });

  describe('when the form is submitted with an interval less than 10 minutes', () => {
    it('shows an interval error', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <button>Interval</button>
            </ClouderaDialogTrigger>
            <IntervalModal />
          </ClouderaDialog>
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Interval' }));

      fireEvent.click(screen.getByTestId('dropdown-input-minutes'));
      await waitFor(() => {
        expect(screen.getByTestId('popover-0')).toBeInTheDocument();
      });
      fireEvent.click(screen.getByTestId('popover-5'));

      fireEvent.click(screen.getByTestId('saveInterval'));
      await waitFor(() => {
        expect(screen.getByTestId('intervalError')).toHaveTextContent(
          'Do not schedule triggers less than 10 minutes apart'
        );
      });
    });
  });

  describe('when the Reset Interval button is pressed', () => {
    it('resets the original value', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <button>Interval</button>
            </ClouderaDialogTrigger>
            <IntervalModal />
          </ClouderaDialog>
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Interval' }));

      const minutesDropdown = screen.getByTestId('dropdown-input-minutes');
      fireEvent.click(minutesDropdown);
      await waitFor(() => {
        expect(screen.getByTestId('popover-0')).toBeInTheDocument();
      });
      fireEvent.click(screen.getByTestId('popover-15'));

      expect(minutesDropdown).toHaveTextContent('00, 15');
      fireEvent.click(screen.getByTestId('resetInterval'));
      expect(minutesDropdown).toHaveTextContent('00');
    });
  });
});
