import { yupResolver } from '@hookform/resolvers/yup';
import { useCreateScheduleGroupMutation } from 'apis/jobs/schedule-groups/create-schedule-group';
import { useEditScheduleGroupMutation } from 'apis/jobs/schedule-groups/edit-schedule-group';
import { isAxiosError } from 'axios';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import {
  ClouderaDialogBody,
  ClouderaDialogContent,
  ClouderaDialogFooter,
  ClouderaDialogHeader,
  ClouderaDialogTitle,
  useClouderaDialogContext,
} from 'components/cloudera-dialog/ClouderaDialog';
import { Notification } from 'components/notification/Notification';
import React, { useCallback } from 'react';
import { Form } from 'react-bootstrap';
import { Controller, useForm } from 'react-hook-form';
import { ScheduleGroup } from 'types/schedule-group.type';
import { IntervalSchema, intervalSchema } from './interval-schema';
import moment from 'moment';
import classes from './IntervalModal.module.scss';
import { Cron } from 'components/cron/Cron';

type CreateIntervalProps = {
  scheduleGroup?: ScheduleGroup;
};

/**
 * Modal for creating/editing schedule interval
 * @param scheduleGroup the interval to edit (if it's not passed, a new one will be created)
 */
export const IntervalModal = ({ scheduleGroup }: CreateIntervalProps) => {
  const { reset: resetDialog } = useClouderaDialogContext();

  const form = useForm({
    resolver: yupResolver(intervalSchema),
    values: {
      name: scheduleGroup?.name ?? '',
      interval: scheduleGroup?.runInterval ?? '0 * * * *',
    },
  });

  const {
    register,
    handleSubmit,
    control,
    formState,
    reset,
    resetField,
    setError,
  } = form;

  const { mutateAsync: createInterval } = useCreateScheduleGroupMutation();
  const { mutateAsync: editInterval } = useEditScheduleGroupMutation();

  const onClose = useCallback(() => {
    reset();
    resetDialog();
  }, [reset, resetDialog]);

  const onSave = useCallback(
    async (formFields: IntervalSchema) => {
      try {
        if (scheduleGroup) {
          await editInterval({
            id: scheduleGroup.id,
            name: formFields.name,
            interval: formFields.interval,
          });
        } else {
          await createInterval({
            name: formFields.name,
            interval: formFields.interval,
          });
        }

        onClose();
      } catch (error) {
        if (isAxiosError(error)) {
          if (error.response?.status === 400) {
            setError('name', {
              message: 'A schedule group with this name already exists.',
            });
          } else {
            setError('root', {
              message: 'Failed to update interval',
            });
          }
          return;
        }

        throw error;
      }
    },
    [createInterval, editInterval, onClose, scheduleGroup, setError]
  );

  const onResetInterval = useCallback(() => {
    resetField('interval');
  }, [resetField]);

  return (
    <ClouderaDialogContent testId="createIntervalModal">
      <ClouderaDialogHeader closeButton>
        <ClouderaDialogTitle>
          {scheduleGroup ? 'Edit' : 'Create'} Interval
        </ClouderaDialogTitle>
      </ClouderaDialogHeader>

      <Form
        role="form"
        onSubmit={handleSubmit((formFields) => onSave(formFields))}
        id="createIntervalForm"
      >
        <ClouderaDialogBody>
          {!!formState.errors.root?.message && (
            <Notification
              variant="danger"
              message={formState.errors.root.message}
            />
          )}
          <Form.Group controlId="name">
            <Form.Label className="required-label">Name</Form.Label>
            <Form.Control
              {...register('name')}
              isInvalid={!!formState.errors.name}
              data-test-id="intervalName"
            />

            <Form.Control.Feedback type="invalid" data-test-id="nameError">
              {formState.errors.name?.message}
            </Form.Control.Feedback>
          </Form.Group>

          <Form.Group controlId="interval">
            <Form.Label className="required-label">
              Interval (in UTC)
            </Form.Label>
            <Controller
              name="interval"
              control={control}
              render={({
                field: { onChange, value },
                fieldState: { error, invalid },
              }) => (
                <>
                  <Cron
                    value={value}
                    onChange={onChange}
                    className={invalid ? 'is-invalid' : undefined}
                  />

                  <Form.Control.Feedback
                    type="invalid"
                    data-test-id="intervalError"
                  >
                    {error?.message}
                  </Form.Control.Feedback>
                </>
              )}
            />

            <p className={classes.timeIndicator}>
              UTC: {moment.utc(Date.now()).format('YYYY-MM-DD HH:mm')}
              <br />
              Local Time:{' '}
              {moment.utc(Date.now()).local().format('YYYY-MM-DD HH:mm')}
            </p>
          </Form.Group>

          <ClouderaButton
            type="button"
            color="primary"
            size="sm"
            onClick={onResetInterval}
            testId="resetInterval"
          >
            Reset Interval
          </ClouderaButton>

          {!!scheduleGroup?.dependentJobs?.length && (
            <Notification
              variant="warning"
              title=" Will affect the following jobs"
              message={scheduleGroup.dependentJobs
                .map((job) => job.name)
                .join(', ')}
            />
          )}
        </ClouderaDialogBody>

        <ClouderaDialogFooter>
          <ClouderaButton type="button" onClick={onClose}>
            Cancel
          </ClouderaButton>
          <ClouderaButton
            color="primary"
            disabled={formState.isSubmitting}
            testId="saveInterval"
          >
            Save
          </ClouderaButton>
        </ClouderaDialogFooter>
      </Form>
    </ClouderaDialogContent>
  );
};
