import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import * as cancel from 'apis/jobs/cancel-job-run';
import * as deleteLogs from 'apis/jobs/logs/delete-logs';
import * as rerun from 'apis/jobs/rerun-jobs';
import * as React from 'react';

import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import {
  mockCompletedLog,
  mockEmailJob,
  mockRunningLog,
} from 'pages/job-page/utils/mocks';
import { LatestRunActions } from './LatestRunActions';

configure({ testIdAttribute: 'data-test-id' });

window.arc.frontend = {
  globals: { urls: { previewCustomEmailUrl: 'customUrl' } },
} as any;

jest.mock('arc/misc/misc', () => ({
  getBlankTargetUrl: (url: string) => url,
}));
jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

const mockCompletedRun = {
  job: mockEmailJob,
  log: mockCompletedLog,
};

const mockInProgressRun = {
  job: mockEmailJob,
  log: mockRunningLog,
};

beforeAll(() => {
  jest.spyOn(rerun, 'rerunJobs').mockImplementation(jest.fn());
  jest.spyOn(cancel, 'cancelJobRun').mockImplementation(jest.fn());
  jest.spyOn(deleteLogs, 'deleteLogs').mockImplementation(jest.fn());
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('LatestRunActions component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <LatestRunActions run={mockCompletedRun} />
      </QueryClientProvider>
    );
  });

  describe('when clicking on the Actions button for a completed run', () => {
    it('should open the dropdown and the actions should appear', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <LatestRunActions run={mockCompletedRun} />
        </QueryClientProvider>
      );
      fireEvent.click(screen.getByTestId('runRowActions'));

      expect(
        await screen.findByRole('button', { name: 'Details' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Run Now' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Delete' })
      ).toBeInTheDocument();
    });
  });

  describe('when clicking on the Actions button for a running asset', () => {
    it('should open the dropdown and the actions should appear', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <LatestRunActions run={mockInProgressRun} />
        </QueryClientProvider>
      );
      fireEvent.click(screen.getByTestId('runRowActions'));

      expect(
        await screen.findByRole('button', { name: 'Details' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Cancel' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Delete' })
      ).toBeInTheDocument();
    });
  });

  describe('when clicking on the Run Now button', () => {
    it('should display a confirmation modal for the specified job', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <LatestRunActions run={mockCompletedRun} />
        </QueryClientProvider>
      );
      fireEvent.click(screen.getByTestId('runRowActions'));

      fireEvent.click(screen.getByRole('button', { name: 'Run Now' }));
      fireEvent.click(screen.getByRole('button', { name: 'Run' }));
      await waitFor(() => {
        expect(jest.mocked(rerun.rerunJobs)).toHaveBeenCalledWith([1]);
      });
    });
  });

  describe('when clicking on the Cancel button', () => {
    it('should display a confirmation modal', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <LatestRunActions run={mockInProgressRun} />
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByTestId('runRowActions'));
      fireEvent.click(screen.getByRole('button', { name: 'Cancel' }));

      const cancelModalButton = await screen.findByTestId('cancelConfirm');
      fireEvent.click(cancelModalButton);
      await waitFor(() => {
        expect(jest.mocked(cancel.cancelJobRun)).toHaveBeenCalledWith({
          jobId: 73,
          logId: 2,
        });
      });
    });
  });

  describe('when clicking on the Delete button', () => {
    it('should display a confirmation modal for the specified job', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <LatestRunActions run={mockInProgressRun} />
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByTestId('runRowActions'));
      fireEvent.click(screen.getByRole('button', { name: 'Delete' }));

      fireEvent.click(screen.getByRole('button', { name: 'Delete' }));
      await waitFor(() => {
        expect(jest.mocked(deleteLogs.deleteLogs)).toHaveBeenCalledWith([2]);
      });
    });
  });
});
