import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import React, { memo, useMemo } from 'react';
import { Dropdown } from 'react-bootstrap';

import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import HelpIcon from '@cloudera-internal/cuix-core/icons/svg/ic-help.svg';
import MoreIcon from '@cloudera-internal/cuix-core/icons/svg/ic-more-vertical.svg';
import RunNowIcon from '@cloudera-internal/cuix-core/icons/svg/ic-refresh.svg';
import StopIcon from '@cloudera-internal/cuix-core/icons/svg/ic-stop.svg';
import { useCancelJobRunMutation } from 'apis/jobs/cancel-job-run';
import { useDeleteLogsMutation } from 'apis/jobs/logs/delete-logs';
import { useRerunJobsMutation } from 'apis/jobs/rerun-jobs';
import {
  BulkActionButtonType,
  BulkActionModal,
} from 'components/bulk-action-modal/BulkActionModal';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import { RunDetailsModal } from 'pages/job-page/components/run-details-modal/RunDetailsModal';
import { JobStatus, JobType, JobWithLog } from 'types/job.type';

type LatestRunActionsProps = {
  run: JobWithLog;
};

export const LatestRunActions = memo(({ run }: LatestRunActionsProps) => {
  const runMutation = useRerunJobsMutation();
  const deleteMutation = useDeleteLogsMutation();
  const cancelMutation = useCancelJobRunMutation();

  const jobs = useMemo(() => {
    return [{ id: run.job.id, name: run.job.name }];
  }, [run.job.id, run.job.name]);

  const logs = useMemo(() => {
    return [{ id: run.log.id, name: run.job.name }];
  }, [run.job.name, run.log.id]);

  return (
    <Dropdown>
      <Dropdown.Toggle
        as={ClouderaButton}
        icon={<MoreIcon />}
        testId="runRowActions"
        variant="link"
      />

      <Dropdown.Menu popperConfig={{ strategy: 'fixed' }}>
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <Dropdown.Item disabled={run.job.type === JobType.Other}>
              <HelpIcon className="mr-2" />
              Details
            </Dropdown.Item>
          </ClouderaDialogTrigger>
          <RunDetailsModal log={run.log} />
        </ClouderaDialog>

        {run.log.status === JobStatus.Running && (
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <Dropdown.Item>
                <StopIcon className="mr-2" />
                Cancel
              </Dropdown.Item>
            </ClouderaDialogTrigger>
            <BulkActionModal
              title="Cancel"
              onConfirm={() =>
                cancelMutation.mutateAsync({
                  jobId: run.log.jobId,
                  logId: run.log.id,
                })
              }
              actionType={BulkActionButtonType.Cancel}
              items={jobs}
              disableSelection
            />
          </ClouderaDialog>
        )}

        {run.log.status !== JobStatus.Running && (
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <Dropdown.Item>
                <RunNowIcon className="mr-2" />
                Run Now
              </Dropdown.Item>
            </ClouderaDialogTrigger>
            <BulkActionModal
              title="Run Now"
              onConfirm={runMutation.mutateAsync}
              actionType={BulkActionButtonType.Run}
              items={jobs}
              disableSelection
            />
          </ClouderaDialog>
        )}

        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <Dropdown.Item data-test-id="deleteRun">
              <DeleteIcon className="mr-2" />
              Delete
            </Dropdown.Item>
          </ClouderaDialogTrigger>
          <BulkActionModal
            title="Delete Run"
            description="Are you sure you want to delete this job run?"
            onConfirm={deleteMutation.mutateAsync}
            actionType={BulkActionButtonType.Delete}
            items={logs}
            disableSelection
          />
        </ClouderaDialog>
      </Dropdown.Menu>
    </Dropdown>
  );
});

LatestRunActions.displayName = 'LatestRunActions';
