import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
  within,
} from '@testing-library/react';
import * as cancel from 'apis/jobs/cancel-job-run';
import * as getLatestRuns from 'apis/jobs/get-latest-runs';
import * as deleteLogs from 'apis/jobs/logs/delete-logs';
import * as rerun from 'apis/jobs/rerun-jobs';
import * as getUsers from 'pages/visuals-page/apis/get-users';
import {
  mockCompletedLog,
  mockEmailJob,
  mockRunningLog,
} from 'pages/job-page/utils/mocks';
import * as React from 'react';
import { LatestRunsTable } from './LatestRunsTable';

configure({ testIdAttribute: 'data-test-id' });

window.arc.frontend = {
  globals: { urls: { previewCustomEmailUrl: 'customUrl' } },
} as any;

jest.mock('arc/misc/misc', () => ({
  getBlankTargetUrl: (url: string) => url,
}));

jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

const mockCompletedRun = {
  job: mockEmailJob,
  log: mockCompletedLog,
};

const mockInProgressRun = {
  job: { ...mockEmailJob, name: 'Mock Job 2', id: 12 },
  log: mockRunningLog,
};

beforeAll(() => {
  jest.spyOn(rerun, 'rerunJobs').mockImplementation(jest.fn());
  jest.spyOn(cancel, 'cancelJobRun').mockImplementation(jest.fn());
  jest.spyOn(deleteLogs, 'deleteLogs').mockImplementation(jest.fn());
});

beforeEach(() => {
  jest
    .spyOn(getLatestRuns, 'getLatestRuns')
    .mockResolvedValue([mockCompletedRun, mockInProgressRun]);
  jest.spyOn(getUsers, 'getUsers').mockResolvedValue([]);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('LatestRunTable component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <LatestRunsTable />
      </QueryClientProvider>
    );
  });

  it('should display the runs in a table', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <LatestRunsTable />
      </QueryClientProvider>
    );

    expect(screen.getByTestId('runsTable')).toBeInTheDocument();
  });

  describe('when selecting some rows', () => {
    describe('and clicking on the Delete button', () => {
      it('should display a confirmation modal for the bulk delete action', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <LatestRunsTable />
          </QueryClientProvider>
        );

        fireEvent.click(
          (await screen.findAllByRole('checkbox', { name: 'Select Log' }))[1]
        );

        const deleteButton = screen.getByTestId('deleteSelectedRuns');
        expect(deleteButton).not.toBeDisabled();

        fireEvent.click(deleteButton);
        expect(screen.getByText('Delete Runs')).toBeInTheDocument();
        const bulkDeleteModal = screen.getByTestId('bulkModal');
        expect(
          within(bulkDeleteModal).getByText('Mock Job 2')
        ).toBeInTheDocument();

        fireEvent.click(screen.getByTestId('deleteConfirm'));
        await waitFor(() => {
          expect(deleteLogs.deleteLogs).toHaveBeenCalledWith([2]);
        });
      });
    });
  });

  describe('when clicking on the Run button', () => {
    it('should display a confirmation modal for the specified jobs', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <LatestRunsTable />
        </QueryClientProvider>
      );

      fireEvent.click(
        (await screen.findAllByRole('checkbox', { name: 'Select Log' }))[1]
      );

      const runButton = screen.getByTestId('runSelected');
      expect(runButton).not.toBeDisabled();
      fireEvent.click(runButton);

      expect(screen.getByText('Run jobs')).toBeInTheDocument();
      const bulkDeleteModal = screen.getByTestId('bulkModal');
      expect(
        within(bulkDeleteModal).getByText('Mock Job 2')
      ).toBeInTheDocument();

      fireEvent.click(screen.getByTestId('runConfirm'));
      await waitFor(() => {
        expect(rerun.rerunJobs).toHaveBeenCalledWith([12]);
      });
    });
  });
});
