import BackIcon from '@cloudera-internal/cuix-core/icons/svg/ic-back.svg';
import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import RunNowIcon from '@cloudera-internal/cuix-core/icons/svg/ic-refresh.svg';
import { useQuery } from '@tanstack/react-query';
import {
  createColumnHelper,
  getCoreRowModel,
  getFilteredRowModel,
  getPaginationRowModel,
  getSortedRowModel,
} from '@tanstack/react-table';
import { useGetLatestRunsQuery } from 'apis/jobs/get-latest-runs';
import { useDeleteLogsMutation } from 'apis/jobs/logs/delete-logs';
import { useRerunJobsMutation } from 'apis/jobs/rerun-jobs';
import { VizTooltip } from 'components/VizTooltip';
import {
  BulkActionButtonType,
  BulkActionModal,
} from 'components/bulk-action-modal/BulkActionModal';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import { ClouderaLink } from 'components/cloudera-link/ClouderaLink';
import { ClouderaTable } from 'components/cloudera-table';
import { ClouderaTableListFilter } from 'components/cloudera-table/ClouderaTableListFilter';
import moment from 'moment';
import { StatusIndicator } from 'pages/job-page/components/status-indicator/StatusIndicator';
import { getUsers, usersQueryToken } from 'pages/visuals-page/apis/get-users';
import React, { useCallback, useMemo } from 'react';
import { FormCheck, Popover } from 'react-bootstrap';
import { JobStatus, JobType, JobWithLog } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import classes from '../../JobOverviewPage.module.scss';
import { LatestRunActions } from '../latest-run-actions/LatestRunActions';
import { useTable } from 'hooks/use-table';

const arc = getGlobalArc();
const columnHelper = createColumnHelper<JobWithLog>();

const statusOptions = Object.keys(JobStatus).map((status) => ({
  label: <StatusIndicator status={JobStatus[status]} />,
  stringLabel: status,
  value: JobStatus[status] as string,
}));

const typeOptions = Object.keys(JobType).map((type) => ({
  label: JobType[type] as string,
  value: JobType[type] as string,
}));

const intervalOptions = [
  {
    label: 'Adhoc',
    value: 'Adhoc',
  },
  {
    label: 'Scheduled',
    value: 'Scheduled',
  },
];

export const LatestRunsTable = (): JSX.Element => {
  const { data: runs, isFetching, isLoading } = useGetLatestRunsQuery();

  const { data: userOptions } = useQuery({
    queryKey: [usersQueryToken],
    queryFn: () => getUsers({ isSummary: true }),
    staleTime: 10 * 60 * 1000, // 10 minutes
    cacheTime: 30 * 60 * 1000, // 30 minutes
    select: (users) =>
      users.map((user) => ({
        value: user.username,
        label: user.username,
      })),
  });

  const { mutateAsync: deleteJobLogs } = useDeleteLogsMutation();

  const { mutateAsync: runJobs } = useRerunJobsMutation();

  const columns = useMemo(
    () => [
      columnHelper.display({
        id: 'select',
        header: ({ table }) => (
          <FormCheck
            aria-label="Select All"
            checked={table.getIsAllPageRowsSelected()}
            onChange={table.getToggleAllPageRowsSelectedHandler()}
          />
        ),
        cell: ({ row }) => (
          <FormCheck
            aria-label="Select Log"
            checked={row.getIsSelected()}
            onChange={row.getToggleSelectedHandler()}
          />
        ),
      }),
      columnHelper.accessor('log.status', {
        id: 'status',
        header: 'Status',
        cell: (info) => {
          const error = info.row.original.log.errorMessage;
          const taskList = info.row.original.log.taskList;

          return (
            <VizTooltip
              overlay={
                <Popover id="schedule-tooltip">
                  <Popover.Content>
                    <p>
                      <b>Error Message:</b> {error}
                    </p>
                    <p>
                      <b>Last Task:</b> {taskList[taskList.length - 1]?.message}
                    </p>
                  </Popover.Content>
                </Popover>
              }
              disabled={!error}
            >
              <span>
                <StatusIndicator status={info.getValue()} isAnimated />
              </span>
            </VizTooltip>
          );
        },
        filterFn: 'arrIncludesSome',
      }),
      columnHelper.accessor('job.id', {
        id: 'jobId',
        header: 'Job ID',
        cell: (info) => info.getValue(),
      }),
      columnHelper.accessor('log.id', {
        id: 'logId',
        header: 'Log ID',
        cell: (info) => info.getValue(),
      }),
      columnHelper.accessor('job.name', {
        id: 'name',
        header: 'Name',
        cell: (info) => (
          <ClouderaLink
            href={`${
              arc.frontend.globals.urls.jobDetailPageUrl
            }/${info.row.getValue('jobId')}`}
            testId="jobName"
            className={classes.jobNameLink}
            disabled={
              info.row.getValue('type') === JobType.Refresh ||
              info.row.getValue('type') === JobType.Other
            }
          >
            <VizTooltip
              overlay={
                <Popover id="job-name-tooltip-latest">
                  <Popover.Content>{info.getValue()}</Popover.Content>
                </Popover>
              }
            >
              <span className={classes.jobNameText}>{info.getValue()}</span>
            </VizTooltip>
          </ClouderaLink>
        ),
      }),
      columnHelper.accessor('job.owner', {
        id: 'owner',
        header: 'Owner',
        cell: (info) => <span data-test-id="jobOwner">{info.getValue()}</span>,
        filterFn: 'arrIncludesSome',
      }),

      columnHelper.accessor('job.type', {
        id: 'type',
        header: 'Type',
        cell: (info) => info.getValue(),
        filterFn: 'arrIncludesSome',
      }),
      columnHelper.accessor(
        (item) => (item.job.scheduleGroup ? 'Scheduled' : 'Adhoc'),
        {
          id: 'interval',
          header: 'Interval',
          cell: (info) => info.getValue(),
          filterFn: 'arrIncludesSome',
        }
      ),
      columnHelper.accessor('log.startTime', {
        id: 'startTime',
        header: 'Start Time (Local)',
        cell: (info) =>
          moment.utc(info.getValue()).local().format('YYYY-MM-DD HH:mm'),
        filterFn: (row, columnId, filterValue) =>
          new Date(row.getValue(columnId)) >= new Date(filterValue),
      }),
      columnHelper.accessor('log.totalRunTime', {
        header: 'Total Run Time',
        cell: (info) => moment.duration(info.getValue()).humanize(),
      }),

      columnHelper.display({
        id: 'actions',
        header: '',
        cell: (info) => <LatestRunActions run={info.row.original} />,
      }),
    ],
    []
  );

  const table = useTable({
    data: runs,
    columns,
    getCoreRowModel: getCoreRowModel(),
    getSortedRowModel: getSortedRowModel(),
    getFilteredRowModel: getFilteredRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    getRowId: (row) => String(row.log.id),
    initialState: {
      sorting: [{ id: 'startTime', desc: true }],
    },
  });

  const selectedLogs = table.getSelectedRowModel().rows.map((row) => ({
    id: row.getValue('logId') as number,
    name: row.getValue('name') as string,
  }));

  const selectedJobs = table.getSelectedRowModel().rows.map((row) => ({
    id: row.getValue('jobId') as number,
    name: row.getValue('name') as string,
  }));

  const handleBulkDelete = useCallback(
    (ids: number[]) => {
      deleteJobLogs(ids);
      table.resetRowSelection();
    },
    [deleteJobLogs, table]
  );

  const handleBulkRun = useCallback(
    (ids: number[]) => {
      runJobs(ids);
      table.resetRowSelection();
    },
    [runJobs, table]
  );

  return (
    <>
      <div className="d-flex flex-wrap justify-content-between align-items-end gap-2 my-3">
        <div className="d-flex flex-wrap gap-2 align-self-center">
          <ClouderaTableListFilter
            column={table.getColumn('status')!}
            options={statusOptions}
            label="Last Status"
            confirmChange
          />
          <ClouderaTableListFilter
            column={table.getColumn('owner')!}
            options={userOptions ?? []}
            label="Owner"
            confirmChange
          />
          <ClouderaTableListFilter
            column={table.getColumn('type')!}
            options={typeOptions}
            label="Type"
            confirmChange
          />
          <ClouderaTableListFilter
            column={table.getColumn('interval')!}
            options={intervalOptions}
            label="Interval"
            confirmChange
          />
          <ClouderaButton
            variant="link"
            icon={<BackIcon />}
            onClick={() => table.resetColumnFilters()}
            disabled={!table.getState().columnFilters.length}
          >
            Clear
          </ClouderaButton>
        </div>
        <div className="d-flex flex-wrap gap-2">
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <ClouderaButton
                icon={<DeleteIcon />}
                disabled={!Object.keys(table.getState().rowSelection).length}
                testId="deleteSelectedRuns"
              >
                Delete ({table.getSelectedRowModel().rows.length})
              </ClouderaButton>
            </ClouderaDialogTrigger>
            <BulkActionModal
              title="Delete Runs"
              description="Are you sure you want to delete the following runs?"
              items={selectedLogs}
              onConfirm={handleBulkDelete}
              actionType={BulkActionButtonType.Delete}
            />
          </ClouderaDialog>

          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <ClouderaButton
                icon={<RunNowIcon />}
                disabled={!table.getIsSomeRowsSelected()}
                testId="runSelected"
              >
                Run ({table.getSelectedRowModel().rows.length})
              </ClouderaButton>
            </ClouderaDialogTrigger>
            <BulkActionModal
              title="Run jobs"
              description="Are you sure you want to run the following jobs?"
              items={selectedJobs}
              onConfirm={handleBulkRun}
              actionType={BulkActionButtonType.Run}
            />
          </ClouderaDialog>
        </div>
      </div>

      <ClouderaTable
        isLoading={isLoading}
        hasStaleData={isFetching && !isLoading}
        table={table}
        className={classes.table}
        data-test-id="runsTable"
      />
    </>
  );
};
