import React, { useCallback, useMemo } from 'react';

import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import EditIcon from '@cloudera-internal/cuix-core/icons/svg/ic-edit.svg';
import PlusIcon from '@cloudera-internal/cuix-core/icons/svg/ic-plus-circle.svg';
import {
  createColumnHelper,
  getCoreRowModel,
  getFilteredRowModel,
  getPaginationRowModel,
  getSortedRowModel,
} from '@tanstack/react-table';
import { useDeleteScheduleGroupsMutation } from 'apis/jobs/schedule-groups/delete-schedule-group';
import { useGetScheduleGroupsQuery } from 'apis/jobs/schedule-groups/get-schedule-groups';
import { VizTooltip } from 'components/VizTooltip';
import {
  BulkActionButtonType,
  BulkActionModal,
} from 'components/bulk-action-modal/BulkActionModal';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import { ClouderaTable } from 'components/cloudera-table';
import { CollapsibleList } from 'components/collapsible-list/CollapsibleList';
import cronstrue from 'cronstrue';
import moment from 'moment';
import { SearchBar } from 'pages/visuals-page/components/search-bar/SearchBar';
import { FormCheck, Popover } from 'react-bootstrap';
import { ScheduleGroup } from 'types/schedule-group.type';
import classes from '../../JobOverviewPage.module.scss';
import { IntervalModal } from '../interval-modal/IntervalModal';
import { useTable } from 'hooks/use-table';

const columnHelper = createColumnHelper<ScheduleGroup>();

export const ScheduleGroupsTable = (): JSX.Element => {
  const {
    data: scheduleGroups = [],
    isLoading,
    isFetching,
  } = useGetScheduleGroupsQuery();

  const deleteMutation = useDeleteScheduleGroupsMutation();

  const columns = useMemo(
    () => [
      columnHelper.display({
        id: 'select',
        header: ({ table }) => (
          <FormCheck
            aria-label="Select All"
            checked={table.getIsAllPageRowsSelected()}
            onChange={table.getToggleAllPageRowsSelectedHandler()}
          />
        ),
        cell: ({ row }) => (
          <FormCheck
            aria-label="Select Interval"
            checked={row.getIsSelected()}
            onChange={row.getToggleSelectedHandler()}
          />
        ),
      }),
      columnHelper.accessor('id', {
        header: 'ID',
        cell: (info) => info.getValue(),
      }),
      columnHelper.accessor('name', {
        header: 'Name',
        cell: (info) => info.getValue(),
      }),
      columnHelper.accessor('runInterval', {
        header: 'Interval',
        cell: (info) => cronstrue.toString(info.getValue()),
      }),
      columnHelper.accessor('nextRunTime', {
        header: 'Next Run (Local)',
        cell: (info) =>
          moment.utc(info.getValue()).local().format('YYYY-MM-DD HH:mm'),
        enableSorting: false,
        enableGlobalFilter: false,
      }),
      columnHelper.accessor('dependentJobs', {
        header: 'Dependent Jobs',
        cell: (info) => {
          return (
            <CollapsibleList id="dependentJobs" threshold={2}>
              {info.getValue().map((job) => (
                <p key={job.id} className="my-0">
                  {job.name}
                </p>
              ))}
            </CollapsibleList>
          );
        },
        enableSorting: false,
      }),
      columnHelper.display({
        id: 'actions',
        header: '',
        cell: (info) => (
          <div className="d-flex gap-4">
            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <VizTooltip
                  overlay={
                    <Popover id="scheduleGroupEditTooltip">
                      <Popover.Content>Edit</Popover.Content>
                    </Popover>
                  }
                >
                  <ClouderaButton
                    aria-label="Edit"
                    variant="link"
                    onClick={() => {}}
                    icon={<EditIcon />}
                  />
                </VizTooltip>
              </ClouderaDialogTrigger>
              <IntervalModal scheduleGroup={info.row.original} />
            </ClouderaDialog>
          </div>
        ),
      }),
    ],
    []
  );

  const table = useTable({
    data: scheduleGroups,
    columns: columns,
    getCoreRowModel: getCoreRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    getSortedRowModel: getSortedRowModel(),
    getFilteredRowModel: getFilteredRowModel(),
    initialState: {
      sorting: [{ id: 'nextRunTime', desc: true }],
      globalFilter: '',
    },
  });

  const selectedItems = table.getSelectedRowModel().rows.map((row) => ({
    id: row.getValue('id') as number,
    name: row.getValue('name') as string,
  }));

  const handleBulkDelete = useCallback(
    (ids: number[]) => {
      deleteMutation.mutateAsync(ids);
      table.resetRowSelection();
    },
    [deleteMutation, table]
  );

  return (
    <>
      <div className="d-flex flex-wrap justify-content-between align-items-end gap-2 my-3">
        <div className="d-flex flex-wrap gap-2 align-self-center">
          <SearchBar
            id="searchIntervals"
            testId="searchIntervals"
            placeholder="Search Intervals"
            searchTerm={table.getState().globalFilter}
            setSearchTerm={table.setGlobalFilter}
            debounce={500}
          />
        </div>
        <div className="d-flex flex-wrap gap-2">
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <ClouderaButton
                icon={<DeleteIcon />}
                disabled={!Object.keys(table.getState().rowSelection).length}
                testId="bulkDeleteScheduleGroups"
              >
                Delete ({table.getSelectedRowModel().rows.length})
              </ClouderaButton>
            </ClouderaDialogTrigger>
            <BulkActionModal
              title="Delete Intervals"
              description="Are you sure you want to delete the following?"
              items={selectedItems}
              onConfirm={handleBulkDelete}
              actionType={BulkActionButtonType.Delete}
            />
          </ClouderaDialog>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <ClouderaButton
                color="primary"
                icon={<PlusIcon />}
                testId="createInterval"
              >
                Create Schedule Group
              </ClouderaButton>
            </ClouderaDialogTrigger>
            <IntervalModal />
          </ClouderaDialog>
        </div>
      </div>

      <ClouderaTable
        isLoading={isLoading}
        hasStaleData={isFetching && !isLoading}
        table={table}
        className={classes.table}
        data-test-id="scheduleGroupsTable"
      />
    </>
  );
};
