import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import * as deleteJobs from 'apis/jobs/delete-jobs';
import * as pauseOrResume from 'apis/jobs/pause-or-resume-jobs';
import * as rerun from 'apis/jobs/rerun-jobs';
import * as getScheduleGroups from 'apis/jobs/schedule-groups/get-schedule-groups';
import * as getUsers from 'pages/visuals-page/apis/get-users';
import * as React from 'react';

import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import * as getEmailTemplates from 'apis/jobs/email-templates/get-email-templates';
import { mockEmailJob, mockExtractJob } from 'pages/job-page/utils/mocks';
import { ScheduledJobActions } from './ScheduledJobActions';

configure({ testIdAttribute: 'data-test-id' });

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: { previewCustomEmailUrl: 'customUrl' },
        settings: {
          emailHostUser: 'test@test.com',
          uploadMaxrowsDefault: 1000000,
        },
        authentication: { username: 'user' },
        emailAddresses: '[]',
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

jest.mock('arc/misc/misc', () => ({
  getBlankTargetUrl: (url: string) => url,
}));
jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

beforeAll(() => {
  jest.spyOn(pauseOrResume, 'pauseOrResumeJobs').mockImplementation(jest.fn());
  jest.spyOn(rerun, 'rerunJobs').mockImplementation(jest.fn());
  jest.spyOn(deleteJobs, 'deleteJobs').mockImplementation(jest.fn());
});

beforeEach(() => {
  jest.spyOn(getUsers, 'getUsers').mockResolvedValue([]);
  jest.spyOn(getScheduleGroups, 'getScheduleGroups').mockResolvedValue([]);
  jest.spyOn(getEmailTemplates, 'getEmailTemplates').mockResolvedValue([]);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('ScheduledJobActions component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <ScheduledJobActions job={mockEmailJob} />
      </QueryClientProvider>
    );
  });

  describe('when clicking on the Actions button', () => {
    it('should open the dropdown and the actions should appear', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ScheduledJobActions job={mockEmailJob} />
        </QueryClientProvider>
      );
      fireEvent.click(screen.getByTestId('jobRowActions'));

      expect(
        await screen.findByRole('button', { name: 'Details' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Run Now' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Edit' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Delete' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Pause Scheduling' })
      ).toBeInTheDocument();
    });
  });

  describe('when clicking on the Edit button', () => {
    describe('and the job target is Extract', () => {
      it('the button should be disabled', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <ScheduledJobActions job={mockExtractJob} />
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByTestId('jobRowActions'));
        const editButton = await screen.findByRole('button', { name: 'Edit' });
        expect(editButton).toHaveClass('disabled');
        fireEvent.click(editButton);
        expect(screen.queryByText('Edit Job Details')).not.toBeInTheDocument();
      });
    });

    describe('and the job target is Email', () => {
      it('should open the Edit Email Modal', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <ScheduledJobActions job={mockEmailJob} />
          </QueryClientProvider>
        );
        fireEvent.click(screen.getByTestId('jobRowActions'));

        const editButton = await screen.findByRole('button', { name: 'Edit' });
        fireEvent.click(editButton);

        expect(screen.getByText('Edit Job Details')).toBeInTheDocument();
      });
    });
  });

  describe('when clicking on the Run Now button', () => {
    it('should display a confirmation modal for the specified job', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ScheduledJobActions job={mockEmailJob} />
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByTestId('jobRowActions'));
      fireEvent.click(screen.getByRole('button', { name: 'Run Now' }));

      fireEvent.click(screen.getByRole('button', { name: 'Run' }));
      await waitFor(() => {
        expect(jest.mocked(rerun.rerunJobs)).toHaveBeenCalledWith([1]);
      });
    });
  });

  describe('when clicking on the Pause Scheduling button', () => {
    it('should display a confirmation modal', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ScheduledJobActions job={mockEmailJob} />
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByTestId('jobRowActions'));
      fireEvent.click(screen.getByRole('button', { name: 'Pause Scheduling' }));

      fireEvent.click(screen.getByRole('button', { name: 'Pause' }));
      await waitFor(() => {
        expect(
          jest.mocked(pauseOrResume.pauseOrResumeJobs)
        ).toHaveBeenCalledWith([1]);
      });
    });
  });

  describe('when clicking on the Resume Scheduling button', () => {
    it('should display a confirmation modal', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ScheduledJobActions
            job={{ ...mockEmailJob, isSchedulingPaused: true }}
          />
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByTestId('jobRowActions'));
      fireEvent.click(
        screen.getByRole('button', { name: 'Resume Scheduling' })
      );

      fireEvent.click(screen.getByRole('button', { name: 'Resume' }));
      await waitFor(() => {
        expect(
          jest.mocked(pauseOrResume.pauseOrResumeJobs)
        ).toHaveBeenCalledWith([1]);
      });
    });
  });

  describe('when clicking on the Delete button', () => {
    it('should display a confirmation modal for the specified job', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ScheduledJobActions job={mockEmailJob} />
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByTestId('jobRowActions'));
      fireEvent.click(screen.getByRole('button', { name: 'Delete' }));

      expect(
        screen.getByText('Are you sure you want to delete this job?')
      ).toBeInTheDocument();
      fireEvent.click(screen.getByRole('button', { name: 'Delete' }));

      await waitFor(() => {
        expect(jest.mocked(deleteJobs.deleteJobs)).toHaveBeenCalledWith([1]);
      });
    });
  });
});
