import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import React, { memo, useMemo } from 'react';
import { Dropdown, Popover } from 'react-bootstrap';

import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import EditIcon from '@cloudera-internal/cuix-core/icons/svg/ic-edit.svg';
import HelpIcon from '@cloudera-internal/cuix-core/icons/svg/ic-help.svg';
import MoreIcon from '@cloudera-internal/cuix-core/icons/svg/ic-more-vertical.svg';
import PauseIcon from '@cloudera-internal/cuix-core/icons/svg/ic-pause.svg';
import ResumeIcon from '@cloudera-internal/cuix-core/icons/svg/ic-play.svg';
import RunNowIcon from '@cloudera-internal/cuix-core/icons/svg/ic-refresh.svg';
import UserIcon from '@cloudera-internal/cuix-core/icons/svg/ic-user.svg';
import { useDeleteJobsMutation } from 'apis/jobs/delete-jobs';
import { usePauseOrResumeJobsMutation } from 'apis/jobs/pause-or-resume-jobs';
import { useRerunJobsMutation } from 'apis/jobs/rerun-jobs';
import { VizTooltip } from 'components/VizTooltip';
import {
  BulkActionButtonType,
  BulkActionModal,
} from 'components/bulk-action-modal/BulkActionModal';
import {
  ClouderaDialog,
  ClouderaDialogContent,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import { EmailJobModal } from 'components/email-job-modal/EmailJobModal';
import { ChangeJobOwnerModal } from 'pages/job-page/components/change-job-owner-modal/ChangeJobOwnerModal';
import { RunDetailsModal } from 'pages/job-page/components/run-details-modal/RunDetailsModal';
import { Job, JobStatus, JobType } from 'types/job.type';

type ScheduledJobActionsProps = {
  job: Job;
};

export const ScheduledJobActions = memo(({ job }: ScheduledJobActionsProps) => {
  const deleteMutation = useDeleteJobsMutation();
  const runMutation = useRerunJobsMutation();
  const pauseOrResumeMutation = usePauseOrResumeJobsMutation();

  const actionItems = useMemo(() => {
    return [{ id: job.id, name: job.name }];
  }, [job.id, job.name]);

  return (
    <Dropdown>
      <Dropdown.Toggle
        as={ClouderaButton}
        icon={<MoreIcon />}
        testId="jobRowActions"
        variant="link"
      />

      <Dropdown.Menu popperConfig={{ strategy: 'fixed' }}>
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <Dropdown.Item
              data-test-id="details"
              disabled={job.type === JobType.Other}
            >
              <HelpIcon className="mr-2" />
              Details
            </Dropdown.Item>
          </ClouderaDialogTrigger>
          <RunDetailsModal job={job} />
        </ClouderaDialog>

        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <Dropdown.Item
              data-test-id="runNow"
              disabled={[JobStatus.Running, JobStatus.Queued].includes(
                job.lastStatus
              )}
            >
              <RunNowIcon className="mr-2" />
              Run Now
            </Dropdown.Item>
          </ClouderaDialogTrigger>
          <BulkActionModal
            title="Run Job"
            onConfirm={runMutation.mutateAsync}
            actionType={BulkActionButtonType.Run}
            items={actionItems}
            disableSelection
          />
        </ClouderaDialog>

        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <Dropdown.Item>
              {job.isSchedulingPaused ? (
                <>
                  <ResumeIcon className="mr-2" />
                  Resume Scheduling
                </>
              ) : (
                <>
                  <PauseIcon className="mr-2" />
                  Pause Scheduling
                </>
              )}
            </Dropdown.Item>
          </ClouderaDialogTrigger>
          <BulkActionModal
            title="Pause/Resume Scheduling"
            description="Are you sure you want to pause/resume this job?"
            onConfirm={pauseOrResumeMutation.mutateAsync}
            actionType={
              job.isSchedulingPaused
                ? BulkActionButtonType.Resume
                : BulkActionButtonType.Pause
            }
            items={actionItems}
            disableSelection
          />
        </ClouderaDialog>

        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <VizTooltip
              disabled={job.type === JobType.Email}
              overlay={
                <Popover id="job-edit-tooltip">
                  <Popover.Content>
                    Only email jobs can be edited from this page.
                  </Popover.Content>
                </Popover>
              }
            >
              <div>
                <Dropdown.Item
                  disabled={job.type !== JobType.Email}
                  data-test-id="edit"
                >
                  <EditIcon className="mr-2" />
                  Edit
                </Dropdown.Item>
              </div>
            </VizTooltip>
          </ClouderaDialogTrigger>
          {job.type === JobType.Email && (
            <ClouderaDialogContent size="lg" scrollable>
              <EmailJobModal job={job} />
            </ClouderaDialogContent>
          )}
        </ClouderaDialog>

        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <Dropdown.Item data-test-id="jobActionsChangeOwner">
              <UserIcon className="mr-2" />
              Change Owner
            </Dropdown.Item>
          </ClouderaDialogTrigger>
          <ChangeJobOwnerModal job={job} />
        </ClouderaDialog>

        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <Dropdown.Item>
              <DeleteIcon className="mr-2" />
              Delete
            </Dropdown.Item>
          </ClouderaDialogTrigger>
          <BulkActionModal
            title="Delete Job"
            description="Are you sure you want to delete this job?"
            onConfirm={deleteMutation.mutateAsync}
            actionType={BulkActionButtonType.Delete}
            items={actionItems}
            disableSelection
          />
        </ClouderaDialog>
      </Dropdown.Menu>
    </Dropdown>
  );
});

ScheduledJobActions.displayName = 'ScheduledJobActions';
