import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
  within,
} from '@testing-library/react';
import * as deleteJobs from 'apis/jobs/delete-jobs';
import * as getJobs from 'apis/jobs/get-jobs';
import * as pauseOrResume from 'apis/jobs/pause-or-resume-jobs';
import * as rerun from 'apis/jobs/rerun-jobs';
import * as getScheduleGroups from 'apis/jobs/schedule-groups/get-schedule-groups';
import * as getUsers from 'pages/visuals-page/apis/get-users';
import { mockEmailJob, mockScheduleGroups } from 'pages/job-page/utils/mocks';
import * as React from 'react';
import { ScheduledJobsTable } from './ScheduledJobsTable';

configure({ testIdAttribute: 'data-test-id' });

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: { previewCustomEmailUrl: 'customUrl' },
        settings: { uploadMaxrowsDefault: 1000000 },
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

jest.mock('arc/misc/misc', () => ({
  getBlankTargetUrl: (url: string) => url,
}));
jest.mock('components/navigation-bar/NavigationBar', () => ({
  NavigationBar: () => null,
}));
jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

beforeAll(() => {
  jest.spyOn(pauseOrResume, 'pauseOrResumeJobs').mockImplementation(jest.fn());
  jest.spyOn(rerun, 'rerunJobs').mockImplementation(jest.fn());
  jest.spyOn(deleteJobs, 'deleteJobs').mockImplementation(jest.fn());
});

beforeEach(() => {
  jest
    .spyOn(getScheduleGroups, 'getScheduleGroups')
    .mockResolvedValue(mockScheduleGroups);
  jest.spyOn(getJobs, 'getJobs').mockResolvedValue({
    data: [mockEmailJob, { ...mockEmailJob, id: 12, name: 'Mock Job 2' }],
    total: 2,
  });
  jest.spyOn(getUsers, 'getUsers').mockResolvedValue([]);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('ScheduledJobsTable component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <ScheduledJobsTable />
      </QueryClientProvider>
    );
  });

  describe('displays the runs in a table', () => {
    it('should display a confirmation modal for the specified job', () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ScheduledJobsTable />
        </QueryClientProvider>
      );

      expect(screen.getByTestId('scheduledJobsTable')).toBeInTheDocument();
    });
  });

  describe('when clicking on the Delete button', () => {
    it('should display a confirmation modal for the specified jobs', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ScheduledJobsTable />
        </QueryClientProvider>
      );

      fireEvent.click(
        (await screen.findAllByRole('checkbox', { name: 'Select Job' }))[1]
      );

      const deleteButton = screen.getByTestId('bulkDelete');
      expect(deleteButton).not.toBeDisabled();
      fireEvent.click(deleteButton);

      expect(screen.getByText('Delete Jobs')).toBeInTheDocument();
      const bulkModal = screen.getByTestId('bulkModal');
      expect(within(bulkModal).getByText('Mock Job 2')).toBeInTheDocument();

      fireEvent.click(screen.getByTestId('deleteConfirm'));
      await waitFor(() => {
        expect(deleteJobs.deleteJobs).toHaveBeenCalledWith([12]);
      });
    });
  });

  describe('when clicking on the Run button', () => {
    it('should display a confirmation modal for the specified jobs', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ScheduledJobsTable />
        </QueryClientProvider>
      );

      fireEvent.click(
        (await screen.findAllByRole('checkbox', { name: 'Select Job' }))[1]
      );

      const runButton = screen.getByTestId('bulkRun');
      expect(runButton).not.toBeDisabled();
      fireEvent.click(runButton);

      expect(screen.getByText('Run Jobs')).toBeInTheDocument();
      const bulkModal = screen.getByTestId('bulkModal');
      expect(within(bulkModal).getByText('Mock Job 2')).toBeInTheDocument();

      fireEvent.click(screen.getByTestId('runConfirm'));
      await waitFor(() => {
        expect(rerun.rerunJobs).toHaveBeenCalledWith([12]);
      });
    });
  });

  describe('when clicking on the Resume Scheduling button', () => {
    it('should display a confirmation modal for the specified jobs', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ScheduledJobsTable />
        </QueryClientProvider>
      );

      fireEvent.click(
        (await screen.findAllByRole('checkbox', { name: 'Select Job' }))[1]
      );

      const resumeButton = screen.getByTestId('bulkResume');
      expect(resumeButton).not.toBeDisabled();
      fireEvent.click(resumeButton);

      expect(screen.getByText('Resume Jobs')).toBeInTheDocument();
      const bulkModal = screen.getByTestId('bulkModal');
      expect(within(bulkModal).getByText('Mock Job 2')).toBeInTheDocument();

      fireEvent.click(screen.getByTestId('resumeConfirm'));
      await waitFor(() => {
        expect(pauseOrResume.pauseOrResumeJobs).toHaveBeenCalledWith(
          [12],
          true
        );
      });
    });
  });
});
