import BackIcon from '@cloudera-internal/cuix-core/icons/svg/ic-back.svg';
import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import ResumeIcon from '@cloudera-internal/cuix-core/icons/svg/ic-play.svg';
import RunNowIcon from '@cloudera-internal/cuix-core/icons/svg/ic-refresh.svg';
import { useQuery } from '@tanstack/react-query';
import {
  ColumnFiltersState,
  SortingState,
  createColumnHelper,
  getCoreRowModel,
} from '@tanstack/react-table';
import { useDeleteJobsMutation } from 'apis/jobs/delete-jobs';
import { useGetJobsQuery } from 'apis/jobs/get-jobs';
import { useResumeJobsMutation } from 'apis/jobs/pause-or-resume-jobs';
import { useRerunJobsMutation } from 'apis/jobs/rerun-jobs';
import { useGetScheduleGroupsQuery } from 'apis/jobs/schedule-groups/get-schedule-groups';
import { isAxiosError } from 'axios';
import { VizTooltip } from 'components/VizTooltip';
import {
  BulkActionButtonType,
  BulkActionModal,
} from 'components/bulk-action-modal/BulkActionModal';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import { ClouderaLink } from 'components/cloudera-link/ClouderaLink';
import { ClouderaTable } from 'components/cloudera-table';
import { ClouderaTableDateRangeFilter } from 'components/cloudera-table/ClouderaTableDateRangeFilter';
import { ClouderaTableListFilter } from 'components/cloudera-table/ClouderaTableListFilter';
import moment from 'moment';
import { ErrorPage } from 'pages/error-page/ErrorPage';
import { getErrorTypeByStatus } from 'pages/error-page/utils/get-error-type-by-status';
import { StatusIndicator } from 'pages/job-page/components/status-indicator/StatusIndicator';
import { getUsers, usersQueryToken } from 'pages/visuals-page/apis/get-users';
import { SearchBar } from 'pages/visuals-page/components/search-bar/SearchBar';
import React, { useCallback, useMemo, useState } from 'react';
import { Badge, FormCheck, Popover } from 'react-bootstrap';
import { Job, JobStatus, JobType, SchedulingStatus } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import { ScheduledJobActions } from '../scheduled-job-actions/ScheduledJobActions';
import { ToggleSwitchThreshold } from '../toggle-switch-threshold/ToggleSwitchThreshold';
import classes from '../../JobOverviewPage.module.scss';
import { useTable } from 'hooks/use-table';

const arc = getGlobalArc();

const typeOptions = Object.keys(JobType).map((type) => ({
  label: JobType[type],
  value: JobType[type],
}));

const statusOptions = Object.keys(JobStatus)
  .map((status) => ({
    label: <StatusIndicator status={JobStatus[status]} />,
    stringLabel: status,
    value: JobStatus[status],
  }))
  .concat({
    label: <StatusIndicator status={SchedulingStatus.Paused} />,
    stringLabel: 'Paused',
    value: SchedulingStatus.Paused,
  });

const columnHelper = createColumnHelper<Job>();

export const ScheduledJobsTable = (): JSX.Element => {
  const [pagination, setPagination] = useState({ pageIndex: 0, pageSize: 10 });
  const [sorting, setSorting] = useState<SortingState>([
    { id: 'updateTime', desc: true },
  ]);
  const [columnFilters, setColumnFilters] = useState<ColumnFiltersState>([]);
  const [globalFilter, setGlobalFilter] = useState<string>('');
  const [thresholdOnly, setThresholdOnly] = useState<boolean>(false);

  const {
    data: jobs,
    isError,
    isLoading,
    isFetching,
    error,
  } = useGetJobsQuery({
    pagination,
    sorting,
    columnFilters,
    globalFilter,
    thresholdOnly,
  });

  const { data: scheduleGroups = [] } = useGetScheduleGroupsQuery();

  const { data: userOptions } = useQuery({
    queryKey: [usersQueryToken],
    queryFn: () => getUsers({ isSummary: true }),
    staleTime: 10 * 60 * 1000, // 10 minutes
    cacheTime: 30 * 60 * 1000, // 30 minutes
    select: (users) =>
      users.map((user) => ({
        value: user.id,
        label: user.username,
      })),
  });

  const { mutateAsync: deleteJobs } = useDeleteJobsMutation();
  const { mutateAsync: resumeJobs } = useResumeJobsMutation();
  const { mutateAsync: rerunJobs } = useRerunJobsMutation();

  const columns = useMemo(
    () => [
      columnHelper.display({
        id: 'select',
        header: ({ table }) => (
          <FormCheck
            aria-label="Select All"
            checked={table.getIsAllPageRowsSelected()}
            onChange={table.getToggleAllPageRowsSelectedHandler()}
          />
        ),
        cell: ({ row }) => (
          <FormCheck
            aria-label="Select Job"
            checked={row.getIsSelected()}
            onChange={row.getToggleSelectedHandler()}
          />
        ),
      }),
      columnHelper.accessor(
        (item) =>
          item.isSchedulingPaused ? SchedulingStatus.Paused : item.lastStatus,
        {
          id: 'state',
          header: 'State',
          cell: (info) => (
            <StatusIndicator status={info.getValue()} isAnimated />
          ),
          filterFn: 'arrIncludesSome',
        }
      ),
      columnHelper.accessor('id', {
        header: 'ID',
        cell: (info) => info.getValue(),
      }),
      columnHelper.accessor('name', {
        header: 'Name',
        cell: (info) => (
          <ClouderaLink
            href={`${
              arc.frontend.globals.urls.jobDetailPageUrl
            }/${info.row.getValue('id')}`}
            testId="jobName"
            className={classes.jobNameLink}
            disabled={
              info.row.getValue('type') === JobType.Refresh ||
              info.row.getValue('type') === JobType.Other
            }
          >
            <VizTooltip
              overlay={
                <Popover id="job-name-tooltip-scheduled">
                  <Popover.Content>{info.getValue()}</Popover.Content>
                </Popover>
              }
            >
              <span className={classes.jobNameText}>{info.getValue()}</span>
            </VizTooltip>
            <span>
              {info.row.original.thresholdTarget && (
                <Badge pill className={classes.badge}>
                  Threshold
                </Badge>
              )}
            </span>
          </ClouderaLink>
        ),
      }),
      columnHelper.accessor('type', {
        header: 'Type',
        cell: (info) => info.getValue(),
        enableSorting: false,
        enableGlobalFilter: false,
      }),
      columnHelper.accessor('owner', {
        header: 'Owner',
        cell: (info) => <span data-test-id="jobOwner">{info.getValue()}</span>,
        enableGlobalFilter: false,
      }),
      columnHelper.accessor('scheduleGroup.id', {
        header: 'Schedule',
        enableGlobalFilter: false,
        cell: (info) => {
          const intervalId = info.getValue();
          const interval = scheduleGroups.find(
            (interval) => interval.id === intervalId
          );

          const nextRunTime = moment
            .utc(interval?.nextRunTime)
            .local()
            .format('YYYY-MM-DD HH:mm');

          return (
            <VizTooltip
              overlay={
                <Popover id="schedule-tooltip">
                  <Popover.Content>
                    {info.row.original.isSchedulingPaused
                      ? `Paused (would run at ${nextRunTime})`
                      : `Next Run: ${nextRunTime ?? '-'}`}
                  </Popover.Content>
                </Popover>
              }
            >
              <span>{interval?.name ?? intervalId}</span>
            </VizTooltip>
          );
        },
      }),
      columnHelper.accessor('lastFinishedTime', {
        header: 'Last Run (Local)',
        enableGlobalFilter: false,
        cell: (info) => {
          const lastRun = info.getValue();
          return lastRun
            ? moment.utc(lastRun).local().format('YYYY-MM-DD HH:mm')
            : '-';
        },
      }),
      columnHelper.display({
        id: 'actions',
        header: '',
        cell: (info) => <ScheduledJobActions job={info.row.original} />,
      }),
    ],
    [scheduleGroups]
  );

  const table = useTable({
    data: jobs?.data,
    columns,
    getCoreRowModel: getCoreRowModel(),
    manualPagination: true,
    manualSorting: true,
    manualFiltering: true,
    rowCount: jobs?.total,
    state: { pagination, sorting, columnFilters, globalFilter },
    onPaginationChange: setPagination,
    onSortingChange: setSorting,
    onColumnFiltersChange: setColumnFilters,
    onGlobalFilterChange: setGlobalFilter,
    getRowId: (row) => String(row.id),
  });

  const selectedItems = table.getSelectedRowModel().rows.map((row) => ({
    id: row.getValue('id') as number,
    name: row.getValue('name') as string,
  }));

  const handleBulkDelete = useCallback(
    (ids: number[]) => {
      deleteJobs(ids);
      table.resetRowSelection();
    },
    [deleteJobs, table]
  );

  const handleBulkRun = useCallback(
    (ids: number[]) => {
      rerunJobs(ids);
      table.resetRowSelection();
    },
    [rerunJobs, table]
  );

  const handleBulkResume = useCallback(
    (ids: number[]) => {
      resumeJobs(ids);
      table.resetRowSelection();
    },
    [resumeJobs, table]
  );

  const handleToggleThresholdOnly = useCallback(() => {
    setThresholdOnly((prev: boolean) => !prev);
  }, [setThresholdOnly]);

  if (isError) {
    const status = isAxiosError(error) ? error.response?.status : 500;
    return <ErrorPage type={getErrorTypeByStatus(status)} showNavBar />;
  }

  return (
    <>
      <div className="d-flex flex-wrap justify-content-between align-items-end gap-2 my-3">
        <div className="d-flex flex-wrap gap-2 align-self-center align-items-center">
          <SearchBar
            id="searchScheduledJobs"
            placeholder="Search Jobs"
            searchTerm={table.getState().globalFilter}
            setSearchTerm={table.setGlobalFilter}
            debounce={1000}
          />
          <ClouderaTableListFilter
            column={table.getColumn('state')!}
            options={statusOptions}
            label="State"
            confirmChange
          />
          <ClouderaTableListFilter
            column={table.getColumn('type')!}
            options={typeOptions}
            label="Type"
            confirmChange
          />
          <ClouderaTableListFilter
            column={table.getColumn('owner')!}
            options={userOptions ?? []}
            label="Owner"
            confirmChange
          />
          <ClouderaTableDateRangeFilter
            column={table.getColumn('lastFinishedTime')!}
            label="Last Run"
          />
          <ToggleSwitchThreshold
            thresholdOnly={thresholdOnly}
            onToggle={handleToggleThresholdOnly}
          />
          <ClouderaButton
            variant="link"
            icon={<BackIcon />}
            onClick={() => {
              table.resetColumnFilters();
              table.setGlobalFilter('');
              setThresholdOnly(false);
            }}
            disabled={
              !table.getState().columnFilters.length &&
              !table.getState().globalFilter &&
              !thresholdOnly
            }
          >
            Clear
          </ClouderaButton>
        </div>
        <div className="d-flex flex-wrap gap-2">
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <ClouderaButton
                icon={<DeleteIcon />}
                disabled={!Object.keys(table.getState().rowSelection).length}
                testId="bulkDelete"
              >
                Delete ({table.getSelectedRowModel().rows.length})
              </ClouderaButton>
            </ClouderaDialogTrigger>
            <BulkActionModal
              title="Delete Jobs"
              description="Are you sure you want to delete the following jobs?"
              items={selectedItems}
              onConfirm={handleBulkDelete}
              actionType={BulkActionButtonType.Delete}
            />
          </ClouderaDialog>

          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <ClouderaButton
                icon={<RunNowIcon />}
                disabled={!Object.keys(table.getState().rowSelection).length}
                testId="bulkRun"
              >
                Run ({table.getSelectedRowModel().rows.length})
              </ClouderaButton>
            </ClouderaDialogTrigger>
            <BulkActionModal
              title="Run Jobs"
              description="Are you sure you want to run the following jobs?"
              items={selectedItems}
              onConfirm={handleBulkRun}
              actionType={BulkActionButtonType.Run}
            />
          </ClouderaDialog>

          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <ClouderaButton
                icon={<ResumeIcon />}
                disabled={!table.getIsSomeRowsSelected()}
                testId="bulkResume"
              >
                Resume scheduling ({table.getSelectedRowModel().rows.length})
              </ClouderaButton>
            </ClouderaDialogTrigger>
            <BulkActionModal
              title="Resume Jobs"
              description="Are you sure you want to resume the following jobs? The active jobs from the selected list will be automatically skipped during the process."
              items={selectedItems}
              onConfirm={handleBulkResume}
              actionType={BulkActionButtonType.Resume}
            />
          </ClouderaDialog>
        </div>
      </div>
      <ClouderaTable
        isLoading={isLoading}
        hasStaleData={isFetching && !isLoading}
        table={table}
        className={classes.table}
        data-test-id="scheduledJobsTable"
      />
    </>
  );
};
