import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import { fireEvent, render, screen, waitFor } from '@testing-library/react';
import { getExtract } from 'apis/extracts/get-extract';
import * as cancelJobRun from 'apis/jobs/cancel-job-run';
import * as deleteJobs from 'apis/jobs/delete-jobs';
import { getEmailTemplates } from 'apis/jobs/email-templates/get-email-templates';
import * as getJob from 'apis/jobs/get-job';
import * as getJobStatus from 'apis/jobs/get-job-status';
import * as getJobLogs from 'apis/jobs/logs/get-logs';
import * as pauseOrResume from 'apis/jobs/pause-or-resume-jobs';
import * as rerun from 'apis/jobs/rerun-jobs';
import * as getScheduleGroups from 'apis/jobs/schedule-groups/get-schedule-groups';
import * as getUsers from 'pages/visuals-page/apis/get-users';
import * as React from 'react';
import { JobStatus } from 'types/job.type';
import { JobPage } from './JobPage';
import {
  mockCompletedLog,
  mockEmailJob,
  mockEmailTemplateOptions,
  mockExtract,
  mockExtractJob,
  mockRunningLog,
} from './utils/mocks';

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: {
          previewCustomEmailUrl: 'customUrl',
        },
        settings: {
          emailHostUser: 'test@test.com',
          uploadMaxrowsDefault: 1000000,
        },
        authentication: { username: 'user' },
        emailAddresses: '[]',
      },
      featureFlipper: {
        getValueById: jest.fn(),
      },
    },
    misc: { getBlankTargetUrl: jest.fn().mockReturnValue('customUrl') },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

const url = 'http://dummy.com';
Object.defineProperty(window, 'location', {
  value: {
    href: url,
  },
  writable: true,
});
jest.mock('apis/extracts/get-extract', () => ({
  getExtract: jest.fn(),
  getExtractQueryToken: 'getExtractQueryToken',
}));
jest.mock('apis/jobs/email-templates/get-email-templates', () => ({
  getEmailTemplates: jest.fn(),
  getEmailTemplatesQueryToken: 'getEmailTemplatesQueryToken',
}));
jest.mock(
  './components/change-job-owner-modal/ChangeJobOwnerModal.tsx',
  () => ({
    ChangeJobOwnerModal: () => null,
  })
);
jest.mock('../../components/navigation-bar/NavigationBar', () => ({
  NavigationBar: () => null,
}));
jest.mock('arc/misc/misc', () => ({
  getBlankTargetUrl: (url: string) => url,
}));
jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

beforeAll(() => {
  jest.spyOn(cancelJobRun, 'cancelJobRun').mockImplementation(jest.fn());
  jest.spyOn(pauseOrResume, 'pauseOrResumeJobs').mockImplementation(jest.fn());
  jest.spyOn(rerun, 'rerunJobs').mockImplementation(jest.fn());
  jest.spyOn(deleteJobs, 'deleteJobs').mockImplementation(jest.fn());
});

beforeEach(() => {
  jest.spyOn(getJob, 'getJob').mockResolvedValue(mockEmailJob);
  jest
    .spyOn(getJobLogs, 'getJobLogs')
    .mockResolvedValue({ data: [mockCompletedLog, mockRunningLog], total: 2 });
  jest
    .spyOn(getJobStatus, 'getJobStatus')
    .mockResolvedValue(JobStatus.Finished);
  jest.spyOn(getScheduleGroups, 'getScheduleGroups').mockResolvedValue([]);
  jest.mocked(getEmailTemplates).mockResolvedValue(mockEmailTemplateOptions);
  jest.mocked(getExtract).mockResolvedValue(mockExtract);
  jest.spyOn(getUsers, 'getUsers').mockResolvedValue([]);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('JobPage component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
      </QueryClientProvider>
    );
  });

  it('should display the schedule details', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
      </QueryClientProvider>
    );

    const scheduleDetailsCard = await screen.findByText('Job Details');
    expect(scheduleDetailsCard).toBeInTheDocument();
    expect(screen.getByLabelText('Name')).toHaveTextContent('Mock Job');
    expect(screen.getByLabelText('Schedule')).toHaveTextContent('Hourly');
    expect(screen.getByLabelText('Interval')).toHaveTextContent('Scheduled');
    expect(screen.getByLabelText('Type')).toHaveTextContent('Email');
    expect(screen.getByLabelText('Threshold')).toHaveTextContent('Yes');
    expect(screen.getByLabelText('Visual ID')).toHaveTextContent('1');
    expect(screen.getByLabelText('Visual URL')).toHaveTextContent('url');
    expect(screen.getByLabelText('Subject')).toHaveTextContent(
      'Cloudera Data Visualization'
    );
    expect(screen.getByLabelText('Message')).toHaveTextContent(
      'Default message'
    );
  });

  describe('when clicking on the Actions button', () => {
    it('should open the dropdown and the actions should appear', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
        </QueryClientProvider>
      );

      const actionsButton = await screen.findByRole('button', {
        name: 'Actions',
      });
      fireEvent.click(actionsButton);

      expect(
        await screen.findByRole('button', { name: 'Edit' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Run Now' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Delete' })
      ).toBeInTheDocument();
      expect(
        await screen.findByRole('button', { name: 'Pause Scheduling' })
      ).toBeInTheDocument();
    });
  });

  describe('when clicking on the Edit button', () => {
    describe('and the job target is Extract', () => {
      beforeEach(() => {
        jest.spyOn(getJob, 'getJob').mockResolvedValue(mockExtractJob);
      });
      it('should open the Edit Extract Modal', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
          </QueryClientProvider>
        );

        const actionsButton = await screen.findByRole('button', {
          name: 'Actions',
        });
        fireEvent.click(actionsButton);

        const editButton = await screen.findByRole('button', { name: 'Edit' });
        fireEvent.click(editButton);

        expect(screen.getByText('Edit Data Extract Job')).toBeInTheDocument();
      });
    });

    describe('and the job target is Email', () => {
      it('should open the Edit Email Modal', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
          </QueryClientProvider>
        );

        const actionsButton = await screen.findByRole('button', {
          name: 'Actions',
        });
        fireEvent.click(actionsButton);

        const editButton = await screen.findByRole('button', { name: 'Edit' });
        fireEvent.click(editButton);

        expect(screen.getByText('Edit Job Details')).toBeInTheDocument();
      });
    });
  });

  describe('when clicking on the Delete button', () => {
    it('should display the Delete Dialog and delete the job', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
        </QueryClientProvider>
      );

      const actionsButton = await screen.findByRole('button', {
        name: 'Actions',
      });
      fireEvent.click(actionsButton);
      fireEvent.click(screen.getByRole('button', { name: 'Delete' }));

      const input = await screen.findByRole('textbox', { name: /confirm/ });
      fireEvent.change(input, { target: { value: 'DELETE' } });

      fireEvent.click(screen.getByRole('button', { name: 'Delete' }));

      await waitFor(() => {
        expect(jest.mocked(deleteJobs.deleteJobs)).toHaveBeenCalledWith([1]);
      });
    });
  });

  describe('when the job is running', () => {
    beforeEach(() => {
      jest.spyOn(getJobLogs, 'getJobLogs').mockResolvedValue({
        data: [{ ...mockCompletedLog, status: JobStatus.Running }],
        total: 1,
      });
    });

    it('should display a Cancel button', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
        </QueryClientProvider>
      );

      const actionsButton = await screen.findByRole('button', {
        name: 'Actions',
      });
      fireEvent.click(actionsButton);

      const cancelButton = await screen.findByRole('button', {
        name: 'Cancel',
      });
      expect(cancelButton).toBeInTheDocument();
    });

    describe('when pressing the Cancel button', () => {
      it('should cancel the job and refetch it', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
          </QueryClientProvider>
        );

        const actionsButton = await screen.findByRole('button', {
          name: 'Actions',
        });
        fireEvent.click(actionsButton);

        const cancelButton = await screen.findByRole('button', {
          name: 'Cancel',
        });
        fireEvent.click(cancelButton);
        await waitFor(() => {
          expect(cancelJobRun.cancelJobRun).toHaveBeenCalledWith({
            jobId: 1,
            logId: 1,
          });
        });
        await waitFor(() => {
          expect(jest.spyOn(getJob, 'getJob')).toHaveBeenCalledTimes(1);
        });
      });
    });
  });

  describe('when the job is not running, a Run Now button should be displayed', () => {
    it('should display a Run Now button', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
        </QueryClientProvider>
      );

      const actionsButton = await screen.findByRole('button', {
        name: 'Actions',
      });
      fireEvent.click(actionsButton);

      const runNowButton = await screen.findByRole('button', {
        name: 'Run Now',
      });
      expect(runNowButton).toBeInTheDocument();
    });

    describe('when pressing the Run Now button', () => {
      it('should run the job', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
          </QueryClientProvider>
        );

        const actionsButton = await screen.findByRole('button', {
          name: 'Actions',
        });
        fireEvent.click(actionsButton);

        const runNowButton = await screen.findByRole('button', {
          name: 'Run Now',
        });
        fireEvent.click(runNowButton);
        await waitFor(() => {
          expect(jest.mocked(rerun.rerunJobs)).toHaveBeenCalledWith([1]);
        });
        await waitFor(() => {
          expect(jest.spyOn(getJob, 'getJob')).toHaveBeenCalledTimes(2);
        });
      });
    });
  });

  describe('when the Pause Scheduling button is pressed', () => {
    it('should change the job status to paused', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
        </QueryClientProvider>
      );

      const actionsButton = await screen.findByRole('button', {
        name: 'Actions',
      });
      fireEvent.click(actionsButton);

      const pauseButton = await screen.findByRole('button', {
        name: 'Pause Scheduling',
      });
      expect(pauseButton).toBeInTheDocument();

      fireEvent.click(pauseButton);
      await waitFor(() => {
        expect(
          jest.mocked(pauseOrResume.pauseOrResumeJobs)
        ).toHaveBeenCalledWith([1]);
      });
    });
  });

  describe('when the job is Paused, there should be a Resume Scheduling button', () => {
    beforeEach(() => {
      jest
        .spyOn(getJob, 'getJob')
        .mockResolvedValue({ ...mockEmailJob, isSchedulingPaused: true });
    });

    it('should change the job status from paused and should resume scheduling', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <JobPage shouldShowModal={false} isAlreadySeen={false} jobId="1" />
        </QueryClientProvider>
      );

      const actionsButton = await screen.findByRole('button', {
        name: 'Actions',
      });
      fireEvent.click(actionsButton);

      const resumeButton = await screen.findByRole('button', {
        name: 'Resume Scheduling',
      });
      expect(resumeButton).toBeInTheDocument();

      fireEvent.click(resumeButton);
      await waitFor(() => {
        expect(
          jest.mocked(pauseOrResume.pauseOrResumeJobs)
        ).toHaveBeenCalledWith([1]);
      });
    });
  });
});
