import React from 'react';

import ExpandIcon from '@cloudera-internal/cuix-core/icons/svg/ic-accordion-expand.svg';
import BackIcon from '@cloudera-internal/cuix-core/icons/svg/ic-back.svg';
import CollapseIcon from '@cloudera-internal/cuix-core/icons/svg/ic-collapse.svg';

import classes from './JobPage.module.scss';

import { useGetJobQuery } from 'apis/jobs/get-job';
import { usePeriodicalJobStatusCheck } from 'apis/jobs/get-job-status';
import { useLastLogQuery } from 'apis/jobs/logs/get-logs';
import { isAxiosError } from 'axios';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { NavigationBar } from 'components/navigation-bar/NavigationBar';
import { VizSpinner } from 'components/viz-spinner/VizSpinner';
import { useToggle } from 'hooks/use-toggle';
import { ErrorPage } from 'pages/error-page/ErrorPage';
import { Accordion, Card, Container } from 'react-bootstrap';
import { JobType } from 'types/job.type';
import { getErrorTypeByStatus } from '../error-page/utils/get-error-type-by-status';
import { EmailDetailsGrid } from './components/email-details-grid/EmailDetailsGrid';
import { ExtractDetailsGrid } from './components/extract-details-grid/ExtractDetailsGrid';
import { JobActionsDropdown } from './components/job-actions-dropdown/JobActionsDropdown';
import { PreviousRunsSection } from './components/previous-runs-section/PreviousRunsSection';
import { ScheduleDetailsGrid } from './components/schedule-details-grid/ScheduleDetailsGrid';
import { ThresholdDetailsGrid } from './components/threshold-details-grid/ThresholdDetailsGrid';

type JobPageProps = {
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
  jobId: string;
};

export const JobPage = ({
  shouldShowModal,
  isAlreadySeen,
  jobId: jobIdParam,
}: JobPageProps): JSX.Element => {
  const jobId = Number(jobIdParam);

  const [isTargetDetailsOpen, toggleTargetDetails] = useToggle(false);

  const { data: job, isError, error } = useGetJobQuery(jobId);
  const { data: lastLog } = useLastLogQuery(jobId);
  usePeriodicalJobStatusCheck(jobId, 5000);

  if (isError) {
    const status = isAxiosError(error) ? error.response?.status : 500;
    return <ErrorPage type={getErrorTypeByStatus(status)} showNavBar />;
  }

  return (
    <Container fluid className={classes.container}>
      <NavigationBar
        isAlreadySeen={isAlreadySeen}
        shouldShowModal={shouldShowModal}
        activeTabs={{
          main: 'settings',
          sub: 'navbar-settings-menu-job-status',
        }}
      />

      {job ? (
        <Container fluid className={classes.main}>
          <ClouderaButton
            icon={<BackIcon className="mr-1" />}
            variant="link"
            onClick={() => window.history.back()}
            testId="goBackToJobs"
          >
            Go back to Jobs
          </ClouderaButton>

          {/* There's a bug in this React Bootstrap version when rendering tabs inside cards: https://github.com/react-bootstrap/react-bootstrap/issues/4771#issuecomment-735006640 */}
          <div className="card my-3">
            <Card.Body>
              <div className="d-flex justify-content-between align-items-center">
                <div className="d-flex align-items-center grow overflow-hidden gap-2">
                  <h3>Job Details</h3>
                </div>

                <JobActionsDropdown job={job} lastLog={lastLog} />
              </div>

              <ScheduleDetailsGrid job={job} />
            </Card.Body>
          </div>

          <Accordion activeKey={isTargetDetailsOpen ? 'details' : ''}>
            <Card className="mb-3">
              <Card.Body>
                <Accordion.Toggle
                  as="h3"
                  className="mb-0 cursor-pointer d-flex align-items-center"
                  eventKey="details"
                  data-test-id="detailsToggle"
                  onClick={toggleTargetDetails}
                >
                  {isTargetDetailsOpen ? (
                    <CollapseIcon className="mr-2" />
                  ) : (
                    <ExpandIcon className="mr-2" />
                  )}
                  {job?.type === JobType.Email
                    ? 'Email Details'
                    : 'Data Extract Details'}
                </Accordion.Toggle>

                <Accordion.Collapse eventKey="details">
                  <div className="mt-3">
                    {job?.type === JobType.Email && !!job.emailTarget && (
                      <EmailDetailsGrid
                        emailTarget={job.emailTarget}
                        snapshotTarget={job.snapshotTarget}
                      />
                    )}

                    {!!job.thresholdTarget && (
                      <ThresholdDetailsGrid
                        thresholdTarget={job.thresholdTarget}
                      />
                    )}

                    {job?.type === JobType.DataExtract &&
                      !!job.dataCopyTarget && (
                        <ExtractDetailsGrid
                          dataCopyTarget={job.dataCopyTarget}
                        />
                      )}
                  </div>
                </Accordion.Collapse>
              </Card.Body>
            </Card>
          </Accordion>

          <PreviousRunsSection jobId={jobId} />
        </Container>
      ) : (
        <div className="flex text-center text-2xl">
          <VizSpinner />
        </div>
      )}
    </Container>
  );
};
