import { render, screen, fireEvent, waitFor } from '@testing-library/react';
import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import React from 'react';
import { ChangeJobOwnerModal } from './ChangeJobOwnerModal';
import { editJobOwner } from 'apis/jobs/edit-job-owner';
import * as getUsers from 'pages/visuals-page/apis/get-users';
import { useNotificationStore } from 'stores/use-notification-store';
import { mockEmailJob } from 'pages/job-page/utils/mocks';
import '@testing-library/jest-dom';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';

jest.mock('apis/jobs/edit-job-owner', () => ({
  editJobOwner: jest.fn(),
}));
jest.mock('stores/use-notification-store');

const mockUseNotificationStore = useNotificationStore as unknown as jest.Mock;

const queryClient = new QueryClient();

const mockUsers = [
  {
    date_joined: '2020-02-16T09:10:34.825676+00:00',
    first_name: '',
    full_name: '',
    id: 1,
    is_active: true,
    is_password_editable: true,
    is_superuser: true,
    last_login: 'N/A',
    last_name: '',
    profile: '{}',
    username: 'vizapps_admin',
  },
  {
    date_joined: '2024-05-29T13:18:49.404189+00:00',
    first_name: 'John',
    full_name: 'John Doe',
    id: 2,
    is_active: true,
    is_password_editable: true,
    is_superuser: true,
    last_login: 'N/A',
    last_name: 'Doe',
    profile: '{}',
    username: 'jondoe001',
  },
  {
    date_joined: '2022-06-16T09:10:34.825676+00:00',
    first_name: '',
    full_name: '',
    id: 3,
    is_active: true,
    is_password_editable: true,
    is_superuser: true,
    last_login: 'N/A',
    last_name: '',
    profile: '{}',
    username: 'Owner',
  },
];

beforeEach(() => {
  jest.spyOn(getUsers, 'getUsers').mockResolvedValue(mockUsers);
  mockUseNotificationStore.mockReturnValue({
    addNotification: jest.fn(),
  });
});

afterEach(() => {
  jest.resetAllMocks();
  jest.useRealTimers();
});

const renderComponent = () => {
  return render(
    <QueryClientProvider client={queryClient}>
      <ClouderaDialog>
        <ClouderaDialogTrigger>
          <button>Change Owner</button>
        </ClouderaDialogTrigger>
        <ChangeJobOwnerModal job={mockEmailJob} />
      </ClouderaDialog>
    </QueryClientProvider>
  );
};

describe('ChangeJobOwnerModal', () => {
  it('renders without error', () => {
    renderComponent();
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));

    expect(screen.getByRole('dialog')).toBeInTheDocument();
    expect(screen.getByText('Change Job Owner')).toBeInTheDocument();
    expect(screen.getByLabelText('Current Owner')).toHaveValue(
      mockEmailJob.owner
    );
    expect(screen.getByRole('button', { name: 'Save' })).not.toBeDisabled();
    expect(jest.mocked(getUsers.getUsers)).toHaveBeenCalledTimes(1);
  });

  it('does not render when isOpen is false and does not call getUsers API', () => {
    renderComponent();
    expect(screen.queryByRole('dialog')).not.toBeInTheDocument();
    expect(jest.mocked(getUsers.getUsers)).not.toHaveBeenCalled();
  });

  it('displays user options fetched from API', async () => {
    renderComponent();
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));

    const selectInput = screen.getByRole('combobox');
    fireEvent.mouseDown(selectInput);

    await waitFor(() => {
      expect(screen.getByText(mockUsers[0].username)).toBeInTheDocument();
    });
    await waitFor(() => {
      expect(screen.getByText(mockUsers[1].username)).toBeInTheDocument();
    });
  });

  it('displays error when form is submitted without selecting a new owner', async () => {
    renderComponent();
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));

    fireEvent.click(screen.getByRole('button', { name: 'Save' }));
    await waitFor(() => {
      expect(
        screen.getByText('New job owner is a required field')
      ).toBeInTheDocument();
    });
  });

  it('submits the form and calls API with correct data', async () => {
    renderComponent();
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));

    const selectInput = screen.getByRole('combobox');
    fireEvent.mouseDown(selectInput);
    fireEvent.click(screen.getByText(mockUsers[0].username));
    fireEvent.click(screen.getByRole('button', { name: 'Save' }));

    await waitFor(() => {
      expect(editJobOwner).toHaveBeenCalledWith(
        mockEmailJob.id,
        mockUsers[0].id
      );
    });
    await waitFor(() => {
      expect(mockUseNotificationStore().addNotification).toHaveBeenCalledWith(
        {
          message: 'Successfully changed job owner',
          variant: 'success',
        },
        { delay: 3000 }
      );
    });

    expect(jest.mocked(editJobOwner)).toHaveBeenCalledTimes(1);
  });

  it('displays "No owner" when there is no current job owner and allows user to choose any user from the selection', async () => {
    const mockJobWithoutOwner = { ...mockEmailJob, owner: undefined };
    render(
      <QueryClientProvider client={queryClient}>
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <button>Change Owner</button>
          </ClouderaDialogTrigger>
          <ChangeJobOwnerModal job={mockJobWithoutOwner} />
        </ClouderaDialog>
      </QueryClientProvider>
    );
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));

    expect(screen.getByLabelText('Current Owner')).toHaveValue('No owner');
    const selectInput = screen.getByRole('combobox');
    fireEvent.mouseDown(selectInput);

    await waitFor(() => {
      expect(screen.getByText(mockUsers[0].username)).toBeInTheDocument();
    });
    await waitFor(() => {
      expect(screen.getByText(mockUsers[1].username)).toBeInTheDocument();
    });

    fireEvent.click(screen.getByText(mockUsers[1].username));
    expect(screen.getByText(mockUsers[1].username)).toBeInTheDocument();
    fireEvent.click(screen.getByRole('button', { name: 'Save' }));
    await waitFor(() => {
      expect(editJobOwner).toHaveBeenCalledWith(
        mockEmailJob.id,
        mockUsers[1].id
      );
    });
    await waitFor(() => {
      expect(mockUseNotificationStore().addNotification).toHaveBeenCalledWith(
        {
          message: 'Successfully changed job owner',
          variant: 'success',
        },
        { delay: 3000 }
      );
    });

    expect(jest.mocked(editJobOwner)).toHaveBeenCalledTimes(1);
  });

  it('if there is a current job owner, it displays the name of the current owner correctly', async () => {
    renderComponent();
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));
    expect(screen.getByLabelText('Current Owner')).toHaveValue(
      mockEmailJob.owner
    );
  });

  it('displays correct alert message if saved with the same owner', async () => {
    renderComponent();
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));
    const selectInput = screen.getByRole('combobox');
    fireEvent.mouseDown(selectInput);
    fireEvent.click(screen.getByText(mockUsers[2].username));
    fireEvent.click(screen.getByRole('button', { name: 'Save' }));

    await waitFor(() => {
      expect(editJobOwner).toHaveBeenCalledWith(
        mockEmailJob.id,
        mockUsers[2].id
      );
    });
    await waitFor(() => {
      expect(mockUseNotificationStore().addNotification).toHaveBeenCalledWith(
        {
          message: 'No owner change made',
          variant: 'info',
        },
        { delay: 3000 }
      );
    });

    expect(jest.mocked(editJobOwner)).toHaveBeenCalledTimes(1);
  });

  it('if a job owner is changed once, the modal should display the name of the new job owner when reopened', async () => {
    const { rerender } = render(
      <QueryClientProvider client={queryClient}>
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <button>Change Owner</button>
          </ClouderaDialogTrigger>
          <ChangeJobOwnerModal job={mockEmailJob} />
        </ClouderaDialog>
      </QueryClientProvider>
    );
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));
    expect(screen.getByLabelText('Current Owner')).toHaveValue(
      mockEmailJob.owner
    );

    const selectInput = screen.getByRole('combobox');
    fireEvent.mouseDown(selectInput);
    fireEvent.click(screen.getByText(mockUsers[1].username));
    fireEvent.click(screen.getByRole('button', { name: 'Save' }));

    await waitFor(() => {
      expect(editJobOwner).toHaveBeenCalledWith(
        mockEmailJob.id,
        mockUsers[1].id
      );
    });
    await waitFor(() => {
      expect(mockUseNotificationStore().addNotification).toHaveBeenCalledWith(
        {
          message: 'Successfully changed job owner',
          variant: 'success',
        },
        { delay: 3000 }
      );
    });

    const updatedJob = { ...mockEmailJob, owner: mockUsers[1].username };
    rerender(
      <QueryClientProvider client={queryClient}>
        <ClouderaDialog>
          <ClouderaDialogTrigger>
            <button>Change Owner</button>
          </ClouderaDialogTrigger>
          <ChangeJobOwnerModal job={updatedJob} />
        </ClouderaDialog>
      </QueryClientProvider>
    );
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));

    expect(screen.getByLabelText('Current Owner')).toHaveValue(
      mockUsers[1].username
    );
  });

  it('does not send editJobOwner API query when modal is closed without selecting a new owner', () => {
    renderComponent();
    fireEvent.click(screen.getByRole('button', { name: 'Change Owner' }));
    fireEvent.click(screen.getByRole('button', { name: 'Cancel' }));
    expect(editJobOwner).not.toHaveBeenCalled();
  });
});
