import { yupResolver } from '@hookform/resolvers/yup';
import { useMutation, useQuery, useQueryClient } from '@tanstack/react-query';
import React, { useCallback, useEffect, useRef, useState } from 'react';

import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { Form, Modal } from 'react-bootstrap';
import Select, { GroupBase, SelectInstance, SingleValue } from 'react-select';
import { Job } from 'types/job.type';

import { editJobOwner } from 'apis/jobs/edit-job-owner';
import { QueryKeyBase } from 'apis/query-keys';
import {
  ClouderaDialogContent,
  useClouderaDialogContext,
} from 'components/cloudera-dialog/ClouderaDialog';
import { getUsers, usersQueryToken } from 'pages/visuals-page/apis/get-users';
import { Controller, useForm } from 'react-hook-form';
import { useNotificationStore } from 'stores/use-notification-store';
import * as yup from 'yup';
import classes from './ChangeJobOwnerModal.module.scss';

type ChangeJobOwnerModalProps = {
  job: Job;
};

type UserOption = {
  value: number;
  label: string;
};

const changeJobOwnerSchema = yup.object().shape({
  newJobOwner: yup
    .number()
    .nullable()
    .required('New job owner is a required field'),
});

type ChangeJobOwnerSchema = {
  newJobOwner: number | null;
};

export const ChangeJobOwnerModal = ({ job }: ChangeJobOwnerModalProps) => {
  const [newOwner, setNewOwner] = useState<string | null>(null);
  const [newOwnerId, setNewOwnerId] = useState<number | null>(null);
  const { reset: resetDialog, isOpen } = useClouderaDialogContext();
  const queryClient = useQueryClient();
  const selectRef = useRef<SelectInstance<
    UserOption,
    false,
    GroupBase<UserOption>
  > | null>(null);

  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const { data: userOptions } = useQuery({
    queryKey: [usersQueryToken],
    queryFn: () => getUsers({ isSummary: true }),
    staleTime: 10 * 60 * 1000, // 10 minutes
    cacheTime: 30 * 60 * 1000, // 30 minutes
    enabled: isOpen,
    select: (users) =>
      users.map((user) => ({
        value: user.id,
        label: user.username,
      })),
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Error fetching users.',
        },
        { delay: 3000 }
      );
    },
  });

  const changeOwnerMutation = useMutation({
    mutationFn: (newOwnerId: number) => editJobOwner(job.id, newOwnerId),
    onSuccess: () => {
      queryClient.invalidateQueries([QueryKeyBase.JobLogs]);
      queryClient.invalidateQueries([QueryKeyBase.Jobs]);

      if (newOwner === job.owner) {
        addNotification(
          {
            message: 'No owner change made',
            variant: 'info',
          },
          { delay: 3000 }
        );
      } else {
        addNotification(
          {
            message: 'Successfully changed job owner',
            variant: 'success',
          },
          { delay: 3000 }
        );
      }

      handleClose();
    },
    onError: () => {
      addNotification(
        {
          message: 'Error changing job owner. Please, try again!',
          variant: 'danger',
        },
        { delay: 3000 }
      );

      handleClose();
    },
  });

  const {
    control,
    handleSubmit,
    formState: { errors },
    reset,
    setValue,
    clearErrors,
  } = useForm<ChangeJobOwnerSchema>({
    resolver: yupResolver<ChangeJobOwnerSchema>(changeJobOwnerSchema),
    defaultValues: {
      newJobOwner: null,
    },
  });

  useEffect(() => {
    if (errors.newJobOwner) {
      selectRef.current?.focus();
    }
  }, [errors]);

  const onSubmit = () => {
    if (!newOwnerId) {
      return;
    }
    changeOwnerMutation.mutateAsync(newOwnerId);
  };

  const handleClose = useCallback(() => {
    resetDialog();
    reset();
    clearErrors();
    setValue('newJobOwner', null);
    setNewOwner(null);
    setNewOwnerId(null);
  }, [resetDialog, reset, clearErrors, setValue]);

  return (
    <ClouderaDialogContent onClose={handleClose} testId="changeJobOwnerModal">
      <Modal.Header closeButton>
        <Modal.Title>Change Job Owner</Modal.Title>
      </Modal.Header>

      <Modal.Body>
        <Form role="form" onSubmit={handleSubmit(onSubmit)}>
          <Form.Group className="d-flex">
            <Form.Label htmlFor="currentOwner" className="w-25">
              Current Owner
            </Form.Label>
            <Form.Control
              data-test-id="currentOwner"
              id="currentOwner"
              className="w-75"
              value={job.owner ? job.owner : 'No owner'}
              disabled
            />
          </Form.Group>

          <Form.Group className="d-flex">
            <Form.Label className={`${classes.requiredLabel} w-25`}>
              New job owner
            </Form.Label>

            <Controller
              name="newJobOwner"
              control={control}
              render={({ field: { onChange, onBlur, name, ref } }) => (
                <Select
                  ref={(select) => {
                    selectRef.current = select;
                    ref(select);
                  }}
                  className="w-75"
                  id="changeJobOwner"
                  options={userOptions}
                  onChange={(e: SingleValue<UserOption>) => {
                    onChange(e ? e.value : null);
                    setNewOwner(e ? e.label : null);
                    setNewOwnerId(e ? e.value : null);
                  }}
                  onBlur={onBlur}
                  name={name}
                  isClearable={false}
                />
              )}
            />
          </Form.Group>

          <div className="d-flex">
            <div className="w-25"></div>
            <div className="w-75 position-relative">
              {errors.newJobOwner && (
                <p className="text-danger mt-n3 pl-2 position-absolute">
                  <small>{errors.newJobOwner.message}</small>
                </p>
              )}
            </div>
          </div>
        </Form>
      </Modal.Body>

      <Modal.Footer>
        <ClouderaButton
          type="button"
          onClick={handleClose}
          testId="cancelChangeJobOwner"
        >
          Cancel
        </ClouderaButton>

        <ClouderaButton
          onClick={handleSubmit(onSubmit)}
          type="submit"
          disabled={changeOwnerMutation.isLoading}
          color="primary"
          testId="saveChangeJobOwner"
        >
          Save
        </ClouderaButton>
      </Modal.Footer>
    </ClouderaDialogContent>
  );
};
