import NewWindowIcon from '@cloudera-internal/cuix-core/icons/svg/ic-new-window.svg';
import { useQuery } from '@tanstack/react-query';
import { ClouderaLink } from 'components/cloudera-link/ClouderaLink';
import {
  getEmailTemplates,
  getEmailTemplatesQueryToken,
} from 'apis/jobs/email-templates/get-email-templates';
import React from 'react';
import { Col, Container, Row } from 'react-bootstrap';
import { useNotificationStore } from 'stores/use-notification-store';
import { EmailTarget, SnapshotTarget } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import { CollapsibleList } from '../../../../components/collapsible-list/CollapsibleList';
import classes from '../../JobPage.module.scss';
import { getBlankTargetUrl } from 'arc/misc/misc';
import { getPdfUploadFormat } from 'pages/job-page/utils/get-pdf-upload-format';

const arc = getGlobalArc();

type EmailDetailsGridProps = {
  emailTarget: EmailTarget;
  snapshotTarget: SnapshotTarget | undefined;
};

export const EmailDetailsGrid = ({
  emailTarget,
  snapshotTarget,
}: EmailDetailsGridProps): JSX.Element => {
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const { data: emailTemplateOptions, isError: emailTemplateError } = useQuery({
    queryKey: [getEmailTemplatesQueryToken],
    queryFn: getEmailTemplates,
    placeholderData: [],
    staleTime: 300000,
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Failed to retrieve email template options',
        },
        { delay: 3000 }
      );
    },
    select: (emailTemplates) => {
      let options = emailTemplates.map(({ id, name }) => ({
        value: id,
        label: name,
      }));

      return options;
    },
  });

  return (
    <Container fluid className={classes.detailContainer}>
      <Row xs={1} sm={2}>
        <Col>
          <label>VISUAL ID</label>
          <p aria-label="Visual ID">{emailTarget.id}</p>
        </Col>

        <Col>
          <label>VISUAL URL</label>
          <p aria-label="Visual URL">
            <ClouderaLink
              href={getBlankTargetUrl(emailTarget.visualURL)}
              target="_blank"
            >
              {emailTarget.visualURL} <NewWindowIcon />
            </ClouderaLink>
          </p>
        </Col>
      </Row>
      <Row xs={1} sm={2}>
        <Col>
          <label>TO</label>
          {!!emailTarget.toAddresses.length ? (
            <CollapsibleList id="emailTo" threshold={3}>
              {emailTarget.toAddresses.map((email) => (
                <p key={`to-${email}`} data-test-id={`to-${email}`}>
                  {email}
                </p>
              ))}
            </CollapsibleList>
          ) : (
            <p>-</p>
          )}
        </Col>
        <Col>
          <label>CC</label>
          {!!emailTarget.ccAddresses.length ? (
            <CollapsibleList id="emailCc" threshold={3}>
              {emailTarget.ccAddresses.map((email) => (
                <p key={`cc-${email}`} data-test-id={`cc-${email}`}>
                  {email}
                </p>
              ))}
            </CollapsibleList>
          ) : (
            <p>-</p>
          )}
        </Col>
      </Row>
      <Row xs={1} sm={2}>
        <Col>
          <label>ERROR TO</label>
          {!!emailTarget.errorToAddresses.length ? (
            <CollapsibleList
              id="emailErrorTo"
              threshold={3}
              data-test-id="error-to"
            >
              {emailTarget.errorToAddresses.map((email) => (
                <p key={`error-to-${email}`}>{email}</p>
              ))}
            </CollapsibleList>
          ) : (
            <p>-</p>
          )}
        </Col>

        <Col>
          <label>REPLY TO</label>
          <p aria-label="Reply To">{emailTarget.replyToAddress || '-'}</p>
        </Col>
      </Row>
      <Row xs={1}>
        <Col>
          <label>SUBJECT</label>
          <p aria-label="Subject"> {emailTarget.subject || '-'}</p>
        </Col>
      </Row>
      <Row xs={1}>
        <Col>
          <label>MESSAGE</label>
          <p aria-label="Message">{emailTarget.message || '-'}</p>
        </Col>
      </Row>
      <Row xs={1}>
        <Col>
          <label>EMAIL TEMPLATE</label>
          <p>
            <ClouderaLink
              href={getBlankTargetUrl(
                `${arc.frontend.globals.urls.previewCustomEmailUrl}${
                  emailTarget.emailTemplateId
                    ? `/${emailTarget.emailTemplateId}`
                    : ''
                }`
              )}
              target="_blank"
              testId="emailTemplateLink"
            >
              {!emailTemplateError
                ? emailTemplateOptions?.find(
                    (option) => option.value === emailTarget.emailTemplateId
                  )?.label ?? 'Default Email Template'
                : 'Failed to load email templates'}{' '}
              <NewWindowIcon />
            </ClouderaLink>
          </p>
        </Col>
      </Row>
      <Row xs={1} sm={2}>
        <Col>
          <label>ATTACHMENTS</label>
          <p aria-label="Attachments">
            {[
              emailTarget.includeEmbeddedImage && 'Embedded image',
              emailTarget.includePng && 'PNG',
              emailTarget.includePdf && 'PDF',
              snapshotTarget?.includeXLSX && 'XLSX',
              snapshotTarget?.includeCSV && 'CSV',
            ]
              .filter(Boolean)
              .join(', ') || '-'}
          </p>
        </Col>
        {snapshotTarget?.uploadFormat && (
          <Col>
            <label>PDF Option</label>
            <p aria-label="PDF Option">
              {getPdfUploadFormat(snapshotTarget.uploadFormat)}
            </p>
          </Col>
        )}
      </Row>
      <Row xs={1}>
        <Col>
          <label>PAGE PARAMETERS</label>
          <CollapsibleList
            threshold={0}
            id="pageParams"
            expandText="Show"
            collapseText="Hide"
          >
            <p>{emailTarget.urlParams || snapshotTarget?.paramString || '-'}</p>
          </CollapsibleList>
        </Col>
      </Row>
    </Container>
  );
};
