import { useQuery } from '@tanstack/react-query';
import { getExtract, getExtractQueryToken } from 'apis/extracts/get-extract';
import { getSourceDatasetLink } from 'pages/job-page/utils/get-source-dataset-link';
import React from 'react';
import { Col, Container, Row } from 'react-bootstrap';
import { DataCopyTarget } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import { CollapsibleList } from '../../../../components/collapsible-list/CollapsibleList';
import classes from '../../JobPage.module.scss';
import { Notification } from 'components/notification/Notification';
import { getBlankTargetUrl } from 'arc/misc/misc';

const arc = getGlobalArc();

type ExtractDetailsGridProps = {
  dataCopyTarget: DataCopyTarget;
};

export const ExtractDetailsGrid = ({
  dataCopyTarget,
}: ExtractDetailsGridProps): JSX.Element => {
  const { data: extract, isError: extractError } = useQuery({
    queryKey: [getExtractQueryToken, dataCopyTarget.id],
    queryFn: () => getExtract(dataCopyTarget.id),
  });

  if (!extract) {
    return (
      <Notification
        title="Error"
        message={'Failed to request extract details'}
        variant="danger"
        show={extractError}
        dismissible
      />
    );
  }

  return (
    <Container fluid className={classes.detailContainer}>
      <Row xs={1} sm={2} md={3}>
        <Col>
          <label>EXTRACT ID</label>
          <p aria-label="Extract Id">{dataCopyTarget.id}</p>
        </Col>
        <Col>
          <label>OWNER</label>
          <p aria-label="Extract Owner">{dataCopyTarget.username}</p>
        </Col>
        <Col>
          <label>REWRITE STATUS</label>
          <p aria-label="Rewrite Status">{extract.rewriteStatus}</p>
        </Col>
      </Row>
      <Row xs={1} sm={2} md={3}>
        <Col>
          <label>SOURCE DATASET</label>
          <p aria-label="Source Dataset">
            {dataCopyTarget.scrDatasetURL ? (
              <a
                href={getBlankTargetUrl(
                  getSourceDatasetLink(
                    dataCopyTarget.scrDatasetURL,
                    dataCopyTarget.id
                  )
                )}
              >
                {extract.srcDatasetName}
              </a>
            ) : (
              extract.srcDatasetName
            )}
          </p>
        </Col>
        <Col>
          <label>DIMENSIONS</label>
          <p aria-label="Dimensions">
            {extract.dimensions.length
              ? extract.dimensions.map((dim) => dim.alias).join(', ')
              : '-'}
          </p>
        </Col>
        <Col>
          <label>MEASURES</label>
          <p aria-label="Measures">
            {extract.measures.length
              ? extract.measures.map((measure) => measure.alias).join(', ')
              : '-'}
          </p>
        </Col>
      </Row>
      <Row xs={1} sm={2} md={3}>
        <Col>
          <label>TARGET DATA CONNECTION</label>
          <p aria-label="Target Data Connection">
            <a
              href={`${arc.frontend.globals.urls.dataUrl}?dataconnection_name=${extract.tgtDataConnectionName}`}
            >
              {extract.tgtDataConnectionName}
            </a>
          </p>
        </Col>
        <Col>
          <label>TARGET TABLE</label>
          <p aria-label="Target Table">{`${extract.tgtDbName}.${extract.tgtTableName}`}</p>
        </Col>
      </Row>
      <Row xs={1} sm={2} md={3}>
        <Col>
          <label>SEND EMAIL ON FAILURE</label>
          <p aria-label="Send Email on Failure">
            {extract.emailFail ? 'Yes' : 'No'}
          </p>
        </Col>
        <Col>
          <label>SEND EMAIL ON SUCCESS</label>
          <p aria-label="Send Email on Success">
            {extract.emailSuccess ? 'Yes' : 'No'}
          </p>
        </Col>
      </Row>
      <Row xs={1}>
        <Col>
          <label>EMAIL TO</label>
          {!!extract.emailTo?.length ? (
            <CollapsibleList id="emailTo" threshold={3}>
              {extract.emailTo.map((email) => (
                <p key={`to-${email}`} data-test-id={`extract-to-${email}`}>
                  {email}
                </p>
              ))}
            </CollapsibleList>
          ) : (
            <p>-</p>
          )}
        </Col>
      </Row>
      <Row xs={1}>
        <Col>
          <label>EMAIL SUBJECT</label>
          <p aria-label="Extract Email Subject">
            {extract.emailSubject ?? '-'}
          </p>
        </Col>
      </Row>
    </Container>
  );
};
