import React from 'react';

import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import ArrowDown from '@cloudera-internal/cuix-core/icons/svg/ic-dropdown.svg';
import EditIcon from '@cloudera-internal/cuix-core/icons/svg/ic-edit.svg';
import PauseIcon from '@cloudera-internal/cuix-core/icons/svg/ic-pause.svg';
import ResumeIcon from '@cloudera-internal/cuix-core/icons/svg/ic-play.svg';
import RefreshIcon from '@cloudera-internal/cuix-core/icons/svg/ic-refresh.svg';
import CancelIcon from '@cloudera-internal/cuix-core/icons/svg/ic-stop.svg';
import UserIcon from '@cloudera-internal/cuix-core/icons/svg/ic-user.svg';

import { getGlobalArc } from 'utils/get-global-arc';

import { useCancelJobRunMutation } from 'apis/jobs/cancel-job-run';
import { useDeleteJobsMutation } from 'apis/jobs/delete-jobs';
import { usePauseOrResumeJobsMutation } from 'apis/jobs/pause-or-resume-jobs';
import { useRerunJobsMutation } from 'apis/jobs/rerun-jobs';
import { VizTooltip } from 'components/VizTooltip';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import {
  ClouderaDialog,
  ClouderaDialogContent,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import { DeleteDialog } from 'components/delete-dialog/DeleteDialog';
import { useToggle } from 'hooks/use-toggle';
import { Dropdown, Popover } from 'react-bootstrap';
import { JobLog } from 'types/job-log.type';
import { Job, JobStatus, JobType } from 'types/job.type';
import { ChangeJobOwnerModal } from '../change-job-owner-modal/ChangeJobOwnerModal';
import { EditExtractModal } from '../edit-extract-modal/EditExtractModal';
import { EmailJobModal } from 'components/email-job-modal/EmailJobModal';

const arc = getGlobalArc();

type JobActionsDropdownProps = {
  job: Job;
  lastLog: JobLog | undefined;
};

export const JobActionsDropdown = ({
  job,
  lastLog,
}: JobActionsDropdownProps): JSX.Element => {
  const [isDeleteDialogOpen, toggleDeleteDialog, closeDeleteDialog] =
    useToggle();

  const deleteMutation = useDeleteJobsMutation();
  const runMutation = useRerunJobsMutation();
  const pauseOrResumeMutation = usePauseOrResumeJobsMutation();
  const cancelMutation = useCancelJobRunMutation();

  return (
    <>
      <DeleteDialog
        title="Delete Job"
        isOpen={isDeleteDialogOpen}
        close={closeDeleteDialog}
        onConfirm={async () => {
          await deleteMutation.mutateAsync([job.id]);
          window.location.href = arc.frontend.globals.urls.jobOverviewUrl;
        }}
        isLoading={deleteMutation.isLoading}
      />

      <Dropdown>
        <VizTooltip
          overlay={
            <Popover id="job-schedule-actions-tooltip">
              <Popover.Content>Job Details Are Loading</Popover.Content>
            </Popover>
          }
          disabled={!!job}
        >
          <div>
            <Dropdown.Toggle
              as={ClouderaButton}
              disabled={!job}
              icon={<ArrowDown />}
              iconPosition="right"
              testId="jobActions"
            >
              Actions
            </Dropdown.Toggle>
          </div>
        </VizTooltip>

        <Dropdown.Menu>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <Dropdown.Item disabled={!job} data-test-id="jobActionsEdit">
                <EditIcon className="mr-2" />
                Edit
              </Dropdown.Item>
            </ClouderaDialogTrigger>
            {job.type === JobType.DataExtract && (
              <EditExtractModal dataCopyTarget={job.dataCopyTarget!} />
            )}
            {job.type === JobType.Email && (
              <ClouderaDialogContent size="lg" scrollable>
                <EmailJobModal job={job} />
              </ClouderaDialogContent>
            )}
          </ClouderaDialog>

          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <Dropdown.Item data-test-id="jobActionsChangeOwner">
                <UserIcon className="mr-2" />
                Change Owner
              </Dropdown.Item>
            </ClouderaDialogTrigger>
            <ChangeJobOwnerModal job={job} />
          </ClouderaDialog>

          <VizTooltip
            overlay={
              <Popover id="delete-schedule-tooltip">
                <Popover.Content>
                  {deleteMutation.isLoading
                    ? 'Deleting in progress...'
                    : 'Delete the job schedule and all of the related runs and logs.'}
                </Popover.Content>
              </Popover>
            }
          >
            <div>
              <Dropdown.Item
                onClick={toggleDeleteDialog}
                disabled={deleteMutation.isLoading}
                data-test-id="jobActionsDelete"
              >
                <DeleteIcon className="mr-2" />
                Delete
              </Dropdown.Item>
            </div>
          </VizTooltip>

          {!!lastLog &&
          [JobStatus.Running, JobStatus.Queued].includes(lastLog.status) ? (
            <Dropdown.Item
              onClick={() =>
                cancelMutation.mutateAsync({ jobId: job.id, logId: lastLog.id })
              }
              disabled={cancelMutation.isLoading}
              data-test-id="jobActionsCancel"
            >
              <CancelIcon className="mr-2" />
              Cancel
            </Dropdown.Item>
          ) : (
            <VizTooltip
              disabled={!!job.owner}
              overlay={
                <Popover id="run-job-tooltip">
                  <Popover.Content>
                    {!job.owner &&
                      'Please assign an owner to this job before running'}
                  </Popover.Content>
                </Popover>
              }
            >
              <div>
                <Dropdown.Item
                  onClick={() => runMutation.mutateAsync([job.id])}
                  disabled={runMutation.isLoading || !job.owner}
                  data-test-id="jobActionsRun"
                >
                  <RefreshIcon className="mr-2" />
                  Run Now
                </Dropdown.Item>
              </div>
            </VizTooltip>
          )}

          {!!job.scheduleGroup && (
            <Dropdown.Item
              onClick={() => pauseOrResumeMutation.mutateAsync([job.id])}
              disabled={pauseOrResumeMutation.isLoading}
              data-test-id="jobActionsPauseResume"
            >
              {job.isSchedulingPaused ? (
                <>
                  <ResumeIcon />
                  Resume Scheduling
                </>
              ) : (
                <>
                  <PauseIcon />
                  Pause Scheduling
                </>
              )}
            </Dropdown.Item>
          )}
        </Dropdown.Menu>
      </Dropdown>
    </>
  );
};
