import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import * as cancelJobRun from 'apis/jobs/cancel-job-run';
import { getEmailTemplates } from 'apis/jobs/email-templates/get-email-templates';
import * as deleteLogs from 'apis/jobs/logs/delete-logs';
import * as getJobLogs from 'apis/jobs/logs/get-logs';
import {
  mockCompletedLog,
  mockEmailTemplateOptions,
  mockRunningLog,
} from 'pages/job-page/utils/mocks';
import * as React from 'react';
import { PreviousRunsSection } from './PreviousRunsSection';

window.arc.frontend = {
  globals: {
    urls: {
      previewCustomEmailUrl: 'customUrl',
    },
  },
  featureFlipper: {
    getValueById: jest.fn(),
  },
} as any;

configure({ testIdAttribute: 'data-test-id' });

const mockAddNotification = jest.fn();
jest.mock('../../../../stores/use-notification-store', () => ({
  useNotificationStore: () => ({ addNotification: mockAddNotification }),
}));
jest.mock('apis/jobs/email-templates/get-email-templates', () => ({
  getEmailTemplates: jest.fn(),
  getEmailTemplatesQueryToken: 'getEmailTemplatesQueryToken',
}));
jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

beforeEach(() => {
  jest
    .spyOn(getJobLogs, 'getJobLogs')
    .mockResolvedValue({ data: [mockCompletedLog, mockRunningLog], total: 2 });
  jest.spyOn(cancelJobRun, 'cancelJobRun').mockImplementation(jest.fn());
  jest.spyOn(deleteLogs, 'deleteLogs').mockImplementation(jest.fn());
  jest.mocked(getEmailTemplates).mockResolvedValue(mockEmailTemplateOptions);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('PreviousRuns component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <PreviousRunsSection jobId={1} />
      </QueryClientProvider>
    );
  });

  describe('the Delete button', () => {
    describe('when no logs are selected', () => {
      it('should be inactive', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <PreviousRunsSection jobId={1} />
          </QueryClientProvider>
        );

        const deleteButton = await screen.findByRole('button', {
          name: 'Delete (0)',
        });
        expect(deleteButton).toBeDisabled();
      });
    });

    describe('when one or more logs are selected', () => {
      it('should delete the logs and display a notification', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <PreviousRunsSection jobId={1} />
          </QueryClientProvider>
        );

        const logCheckboxes = await screen.findAllByRole('checkbox', {
          name: 'Select Log',
        });
        fireEvent.click(logCheckboxes[0]);

        const deleteButton = await screen.findByRole('button', {
          name: 'Delete (1)',
        });
        fireEvent.click(deleteButton);

        const input = await screen.findByRole('textbox', { name: /confirm/ });
        fireEvent.change(input, { target: { value: 'DELETE' } });

        fireEvent.click(screen.getByRole('button', { name: 'Delete' }));

        await waitFor(() => expect(deleteLogs.deleteLogs).toHaveBeenCalled());
        expect(mockAddNotification).toHaveBeenCalled();
      });
    });
  });

  describe('the Cancel Run button', () => {
    describe('when the job is not running', () => {
      it('should be inactive', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <PreviousRunsSection jobId={1} />
          </QueryClientProvider>
        );

        const cancelButtons = await screen.findAllByLabelText('Cancel Run');
        expect(cancelButtons[0]).toBeDisabled();
      });
    });

    describe('when the job is running', () => {
      it('should be active and cancel the run', async () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <PreviousRunsSection jobId={1} />
          </QueryClientProvider>
        );

        const cancelButtons = await screen.findAllByLabelText('Cancel Run');
        fireEvent.click(cancelButtons[1]);
        await waitFor(() =>
          expect(cancelJobRun.cancelJobRun).toHaveBeenCalledTimes(1)
        );
        expect(mockAddNotification).toHaveBeenCalled();
      });
    });
  });

  describe('the More Info button', () => {
    it('should open the Detailed Log overlay', async () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <PreviousRunsSection jobId={1} />
        </QueryClientProvider>
      );

      const runDetailsButtons = await screen.findAllByLabelText(
        'View Run Details'
      );
      expect(runDetailsButtons[0]).toBeInTheDocument();
      fireEvent.click(runDetailsButtons[0]);

      expect(screen.getByText('Run Details')).toBeInTheDocument();
    });
  });
});
