import BackIcon from '@cloudera-internal/cuix-core/icons/svg/ic-back.svg';
import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import HelpIcon from '@cloudera-internal/cuix-core/icons/svg/ic-help.svg';
import StopIcon from '@cloudera-internal/cuix-core/icons/svg/ic-stop.svg';
import {
  ColumnFiltersState,
  SortingState,
  createColumnHelper,
  getCoreRowModel,
} from '@tanstack/react-table';
import { useCancelJobRunMutation } from 'apis/jobs/cancel-job-run';
import { useDeleteLogsMutation } from 'apis/jobs/logs/delete-logs';
import { useGetLogsQuery } from 'apis/jobs/logs/get-logs';
import { VizTooltip } from 'components/VizTooltip';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import { ClouderaTable } from 'components/cloudera-table';
import { ClouderaTableDateFilter } from 'components/cloudera-table/ClouderaTableDateFilter';
import { ClouderaTableListFilter } from 'components/cloudera-table/ClouderaTableListFilter';
import { DeleteDialog } from 'components/delete-dialog/DeleteDialog';
import { useToggle } from 'hooks/use-toggle';
import moment from 'moment';
import React, { useMemo, useState } from 'react';
import { Card, FormCheck, Popover } from 'react-bootstrap';
import { JobLog } from 'types/job-log.type';
import { JobStatus } from 'types/job.type';
import { RunDetailsModal } from '../run-details-modal/RunDetailsModal';
import { StatusIndicator } from '../status-indicator/StatusIndicator';
import classes from './PreviousRunsSection.module.scss';
import { useTable } from 'hooks/use-table';

type PreviousRunsSectionProps = {
  jobId: number;
};

const columnHelper = createColumnHelper<JobLog>();

const statusOptions = Object.keys(JobStatus).map((status) => ({
  label: <StatusIndicator status={JobStatus[status]} />,
  stringLabel: status,
  value: JobStatus[status],
}));

export const PreviousRunsSection = ({ jobId }: PreviousRunsSectionProps) => {
  const [pagination, setPagination] = useState({ pageIndex: 0, pageSize: 10 });
  const [sorting, setSorting] = useState<SortingState>([
    { id: 'startTime', desc: true },
  ]);
  const [columnFilters, setColumnFilters] = useState<ColumnFiltersState>([]);

  const [deleteDialogIsOpen, toggleDeleteDialog, closeDeleteDialog] =
    useToggle();
  const { data: paginatedLogs } = useGetLogsQuery({
    jobId,
    pagination,
    sorting,
    columnFilters,
  });

  const { mutateAsync: deleteLogsMutateAsync, isLoading: deleteLogsLoading } =
    useDeleteLogsMutation(jobId);
  const { mutateAsync: cancelJobMutateAsync } = useCancelJobRunMutation();

  const columns = useMemo(
    () => [
      columnHelper.display({
        id: 'select',
        header: ({ table }) => (
          <FormCheck
            aria-label="Select All"
            checked={table.getIsAllPageRowsSelected()}
            onChange={table.getToggleAllPageRowsSelectedHandler()}
          />
        ),
        cell: ({ row }) => (
          <FormCheck
            aria-label="Select Log"
            checked={row.getIsSelected()}
            onChange={row.getToggleSelectedHandler()}
          />
        ),
      }),
      columnHelper.accessor('status', {
        header: 'Status',
        cell: (info) => <StatusIndicator status={info.getValue()} isAnimated />,
        filterFn: 'arrIncludesSome',
      }),
      columnHelper.accessor('id', {
        header: 'Log ID',
        cell: (info) => info.getValue(),
        filterFn: 'arrIncludesSome',
        enableSorting: false,
      }),
      columnHelper.accessor('startTime', {
        header: 'Start Time (Local)',
        cell: (info) =>
          moment.utc(info.getValue()).local().format('YYYY-MM-DD HH:mm'),
        filterFn: (row, columnId, filterValue) =>
          new Date(row.getValue(columnId)) >= new Date(filterValue),
      }),
      columnHelper.accessor('endTime', {
        header: 'End Time (Local)',
        cell: (info) =>
          moment.utc(info.getValue()).local().format('YYYY-MM-DD HH:mm'),
        filterFn: (row, columnId, filterValue) =>
          new Date(row.getValue(columnId)) <= new Date(filterValue),
      }),
      columnHelper.accessor('totalRunTime', {
        header: 'Total Run Time',
        cell: (info) => moment.duration(info.getValue()).humanize(),
        enableSorting: false,
      }),
      columnHelper.display({
        id: 'actions',
        header: '',
        cell: (info) => (
          <div className="d-flex gap-4">
            <VizTooltip
              overlay={
                <Popover id="job-cancel-run-tooltip">
                  <Popover.Content>Cancel Run</Popover.Content>
                </Popover>
              }
              disabled={
                ![JobStatus.Running, JobStatus.Queued].includes(
                  info.row.getValue('status')
                )
              }
            >
              <ClouderaButton
                aria-label="Cancel Run"
                color="danger"
                variant="link"
                icon={<StopIcon />}
                onClick={() =>
                  cancelJobMutateAsync({
                    jobId,
                    logId: info.row.getValue('id'),
                  })
                }
                disabled={
                  ![JobStatus.Running, JobStatus.Queued].includes(
                    info.row.getValue('status')
                  )
                }
              />
            </VizTooltip>
            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <VizTooltip
                  overlay={
                    <Popover id="job-run-details-tooltip">
                      <Popover.Content>View Run Details</Popover.Content>
                    </Popover>
                  }
                >
                  <ClouderaButton
                    aria-label="View Run Details"
                    variant="link"
                    icon={<HelpIcon />}
                  />
                </VizTooltip>
              </ClouderaDialogTrigger>
              <RunDetailsModal log={info.row.original} />
            </ClouderaDialog>
          </div>
        ),
      }),
    ],
    [cancelJobMutateAsync, jobId]
  );

  const table = useTable({
    data: paginatedLogs?.data,
    columns,
    getCoreRowModel: getCoreRowModel(),
    manualPagination: true,
    manualSorting: true,
    manualFiltering: true,
    rowCount: paginatedLogs?.total,
    state: { pagination, sorting, columnFilters },
    onPaginationChange: setPagination,
    onSortingChange: setSorting,
    onColumnFiltersChange: setColumnFilters,
    getRowId: (row) => String(row.id),
  });

  return (
    // There's a bug in this React Bootstrap version when rendering tabs inside cards: https://github.com/react-bootstrap/react-bootstrap/issues/4771#issuecomment-735006640
    <div className="card my-3">
      <Card.Body>
        <h3>Previous Runs</h3>

        <div className="d-flex justify-content-between align-items-end gap-2 my-3">
          <div className="d-flex flex-wrap gap-2 align-self-center">
            <ClouderaTableListFilter
              column={table.getColumn('status')!}
              options={statusOptions}
              label="Status"
              confirmChange
            />
            <ClouderaTableDateFilter
              column={table.getColumn('startTime')!}
              label="Start Time"
              includeTime
            />
            <ClouderaTableDateFilter
              column={table.getColumn('endTime')!}
              label="End Time"
              includeTime
            />
            <ClouderaButton
              variant="link"
              icon={<BackIcon />}
              onClick={() => table.resetColumnFilters()}
              disabled={!table.getState().columnFilters.length}
            >
              Clear
            </ClouderaButton>
          </div>

          <ClouderaButton
            icon={<DeleteIcon />}
            disabled={!Object.keys(table.getState().rowSelection).length}
            onClick={toggleDeleteDialog}
            testId="deleteJobLogs"
          >
            Delete ({Object.keys(table.getState().rowSelection).length})
          </ClouderaButton>
        </div>

        <ClouderaTable
          isLoading={!paginatedLogs}
          table={table}
          className={classes.table}
          data-test-id="previousRunsTable"
        />

        <DeleteDialog
          title="Delete Log(s)"
          isOpen={deleteDialogIsOpen}
          close={closeDeleteDialog}
          onConfirm={async () => {
            await deleteLogsMutateAsync(
              Object.keys(table.getState().rowSelection).map(Number)
            );
            table.resetRowSelection();
            closeDeleteDialog();
          }}
          isLoading={deleteLogsLoading}
        />
      </Card.Body>
    </div>
  );
};
