import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import { configure, fireEvent, render, screen } from '@testing-library/react';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import {
  mockCompletedLog,
  mockEmailLog,
  mockExtract,
  mockExtractLog,
  mockRunningLog,
} from 'pages/job-page/utils/mocks';
import * as React from 'react';
import { RunDetailsModal } from './RunDetailsModal';
import { getExtract } from 'apis/extracts/get-extract';
import { getEmailTemplates } from 'apis/jobs/email-templates/get-email-templates';

configure({ testIdAttribute: 'data-test-id' });

jest.mock('apis/jobs/email-templates/get-email-templates', () => ({
  getEmailTemplates: jest.fn(),
  getEmailTemplatesQueryToken: 'getEmailTemplatesQueryToken',
}));
jest.mock('apis/extracts/get-extract', () => ({
  getExtract: jest.fn(),
  getExtractQueryToken: 'getExtractQueryToken',
}));
jest.mock('@cloudera-internal/cuix-core/variables', () => ({
  cdlRed500: '#fff',
  cdlBlue500: '#fff',
  cdlGreen500: '#fff',
  cdlOrange500: '#fff',
}));

window.arc.frontend = {
  globals: {
    urls: {
      previewCustomEmailUrl: 'customUrl',
    },
  },
} as any;
jest.mock('arc/misc/misc', () => ({
  getBlankTargetUrl: (url: string) => url,
}));

beforeEach(() => {
  jest.mocked(getExtract).mockResolvedValue(mockExtract);
  jest.mocked(getEmailTemplates).mockResolvedValue([]);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('RunDetailsModal component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <RunDetailsModal />
      </QueryClientProvider>
    );
  });

  describe('when there is no log', () => {
    it('displays a no data available message', () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <button>Open</button>
            </ClouderaDialogTrigger>
            <RunDetailsModal />
          </ClouderaDialog>
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Open' }));

      expect(
        screen.getByText('The details for this asset are not available')
      ).toBeInTheDocument();
    });
  });

  describe('when the log details are available', () => {
    describe('and the log has a running status', () => {
      it('should display the log details', () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <button>Open</button>
              </ClouderaDialogTrigger>
              <RunDetailsModal log={mockRunningLog} />
            </ClouderaDialog>
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('button', { name: 'Open' }));

        expect(screen.getByLabelText('Log Id')).toHaveTextContent('2');
        expect(screen.getByLabelText('Job Id')).toHaveTextContent('73');
        expect(screen.getByLabelText('Owner username')).toHaveTextContent(
          'vizapps_admin'
        );
        expect(screen.getByLabelText('Triggered By')).toHaveTextContent(
          'vizapps_admin'
        );
      });

      it('should not display a notification about incomplete information', () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <button>Open</button>
              </ClouderaDialogTrigger>
              <RunDetailsModal log={mockRunningLog} />
            </ClouderaDialog>
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('button', { name: 'Open' }));

        expect(
          screen.queryByTestId('notification-content')
        ).not.toBeInTheDocument();
      });
    });

    describe('and the log has a completed status', () => {
      it('should display the email target if it is available', () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <button>Open</button>
              </ClouderaDialogTrigger>
              <RunDetailsModal log={mockEmailLog} />
            </ClouderaDialog>
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('button', { name: 'Open' }));

        expect(
          screen.getByRole('tab', { name: 'Email Details' })
        ).toBeInTheDocument();
      });

      it('should display the extract target if it is available', () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <button>Open</button>
              </ClouderaDialogTrigger>
              <RunDetailsModal log={mockExtractLog} />
            </ClouderaDialog>
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('button', { name: 'Open' }));

        expect(
          screen.getByRole('tab', { name: 'Extract Details' })
        ).toBeInTheDocument();
      });

      it('should display a notification about incomplete information if there is no email target', () => {
        render(
          <QueryClientProvider client={new QueryClient()}>
            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <button>Open</button>
              </ClouderaDialogTrigger>
              <RunDetailsModal log={mockCompletedLog} />
            </ClouderaDialog>
          </QueryClientProvider>
        );

        fireEvent.click(screen.getByRole('button', { name: 'Open' }));

        expect(screen.getByTestId('notification-content')).toBeInTheDocument();
      });
    });

    it('should display the task list in a table', () => {
      render(
        <QueryClientProvider client={new QueryClient()}>
          <ClouderaDialog>
            <ClouderaDialogTrigger>
              <button>Open</button>
            </ClouderaDialogTrigger>
            <RunDetailsModal log={mockCompletedLog} />
          </ClouderaDialog>
        </QueryClientProvider>
      );

      fireEvent.click(screen.getByRole('button', { name: 'Open' }));

      expect(screen.getByTestId('tasksTable')).toBeInTheDocument();
      expect(screen.getByText('Get data')).toBeInTheDocument();
    });
  });
});
