import {
  createColumnHelper,
  getCoreRowModel,
  getPaginationRowModel,
  getSortedRowModel,
} from '@tanstack/react-table';
import { ClouderaTable } from 'components/cloudera-table';
import { useTable } from 'hooks/use-table';
import moment from 'moment';
import React, { useMemo } from 'react';
import { JobLogTask } from 'types/job-log.type';

type RunDetailsTasksTableProps = {
  tasks: JobLogTask[];
};

const columnHelper = createColumnHelper<
  JobLogTask & { differenceMs: number }
>();

export const RunDetailsTasksTable = ({
  tasks,
}: RunDetailsTasksTableProps): JSX.Element => {
  const taskList = useMemo(
    () =>
      tasks.map((task, i, arr) => ({
        ...task,
        differenceMs:
          task.date && arr[i - 1]?.date
            ? moment(task.date).diff(moment(arr[i - 1].date), 'ms')
            : 0,
      })),
    [tasks]
  );

  const columns = useMemo(
    () => [
      columnHelper.accessor('message', {
        header: () => 'Message',
        cell: (info) => info.getValue(),
      }),
      columnHelper.accessor('date', {
        header: () => 'Date',
        cell: (info) =>
          info.getValue()
            ? moment
                .utc(info.getValue())
                .local()
                .format('YYYY-MM-DD HH:mm:ss.SSS')
            : '-',
      }),
      columnHelper.accessor('differenceMs', {
        header: () => 'Difference',
        cell: (info) => `${moment.duration(info.getValue()).asSeconds()} s`,
      }),
    ],
    []
  );

  const table = useTable({
    data: taskList,
    columns,
    getCoreRowModel: getCoreRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    getSortedRowModel: getSortedRowModel(),
    initialState: {
      pagination: {
        pageIndex: 0,
        pageSize: 10,
      },
      sorting: [{ id: 'date', desc: true }],
    },
  });

  return <ClouderaTable table={table} data-test-id="tasksTable" />;
};
