import classNames from 'classnames';
import React from 'react';
import FormCheck from 'react-bootstrap/FormCheck';
import FormControl from 'react-bootstrap/FormControl';
import FormGroup from 'react-bootstrap/FormGroup';
import FormLabel from 'react-bootstrap/FormLabel';
import { InputGroup } from 'react-bootstrap';

import { getGlobalArc } from 'utils/get-global-arc';
import { VizLink } from 'components/viz-link/VizLink';
import { useToggle } from 'hooks/use-toggle';

import classes from './LoginForm.module.scss';
import { UnauthorizedSubmitButton } from 'components/unauthorized-submit-button/UnauthorizedSubmitButton';

type LoginFormProps = {
  next: string;
  errorMessage: string;
  failureLimit: number;
  failureLimitRation: number;
  querystring: string;
  loginInstructions: string;

  toggleForgotPasswordModal: () => void;
};

const arc = getGlobalArc();

export const LoginForm = ({
  next,
  errorMessage,
  failureLimit,
  failureLimitRation,
  querystring,
  loginInstructions,
  toggleForgotPasswordModal,
}: LoginFormProps) => {
  const [isVisible, toggleVisibility] = useToggle();
  return (
    <form
      action={`${arc.frontend.globals.urls.loginUrl}?${querystring}`}
      method="post"
      noValidate
    >
      <input
        name="csrfmiddlewaretoken"
        type="hidden"
        value={arc.frontend.globals.authentication.csrfToken}
      />
      <input name="next" type="hidden" value={next} />

      <FormGroup>
        <FormLabel htmlFor="username">
          <span className={classes.label}>Username</span>
        </FormLabel>

        <FormControl
          id="username"
          name="username"
          autoFocus
          autoCapitalize="off"
          autoCorrect="off"
          spellCheck="false"
          data-test-id="username"
        />
      </FormGroup>

      <FormGroup>
        <FormLabel htmlFor="password">
          <span
            className={classNames(
              classes.label,
              errorMessage ? classes.error : ''
            )}
          >
            Password
          </span>
        </FormLabel>

        <InputGroup>
          <FormControl
            id="password"
            name="password"
            type={isVisible ? 'text' : 'password'}
            isInvalid={!!errorMessage || !!failureLimit}
            data-test-id="password"
          />

          <InputGroup.Append
            onClick={toggleVisibility}
            className={classes.clickable}
          >
            <InputGroup.Text id="show-password">
              <i className={`fa ${isVisible ? 'fa-eye' : 'fa-eye-slash'}`} />
            </InputGroup.Text>
          </InputGroup.Append>

          <FormControl.Feedback type="invalid">
            {errorMessage}

            {!errorMessage &&
              failureLimit &&
              `We locked your account because you made ${failureLimitRation} failed login attempts. You can try again after some time, or contact your administrator.`}
          </FormControl.Feedback>
        </InputGroup>
        <div className={classes.forgotPassword}>
          <VizLink
            onClick={toggleForgotPasswordModal}
            data-test-id="forgot-your-password"
          >
            Forgot your password?
          </VizLink>
        </div>
      </FormGroup>

      {loginInstructions && (
        <div className={classes.loginInstructions}>{loginInstructions}</div>
      )}

      <FormGroup className="my-3" controlId="remember-me">
        <FormCheck
          type="checkbox"
          label="Remember me on this computer"
          data-test-id="remember-me"
        />
      </FormGroup>

      <UnauthorizedSubmitButton type="submit" data-test-id="submit">
        LOGIN
      </UnauthorizedSubmitButton>
    </form>
  );
};
