import * as React from 'react';
import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  within,
} from '@testing-library/react';

import { mockAlias } from './utils/mocks';
import { getAllAliases } from 'apis/aliases/get-all-aliases';
import { ManageAliasesPage } from './ManageAliasesPage';
import { useDashboardOptions } from './hooks/use-fetch-dashboard-options';
import { useAppOptions } from './hooks/use-fetch-app-options';

configure({ testIdAttribute: 'data-test-id' });

const queryClient = new QueryClient({
  defaultOptions: {
    queries: {
      retry: false,
    },
  },
});

const mockAddNotification = jest.fn();

jest.mock('../../apis/aliases/get-all-aliases.ts', () => ({
  getAllAliases: jest.fn(),
}));

jest.mock('../../apis/get-reports.ts', () => ({
  getReports: jest.fn(),
}));

jest.mock('../../apis/get-apps.ts', () => ({
  getApps: jest.fn(),
}));

jest.mock('../../../src/stores/use-notification-store.ts', () => ({
  useNotificationStore: () => ({ addNotification: mockAddNotification }),
}));

jest.mock('arc/misc/misc', () => ({
  getBlankTargetUrl: jest.fn((url: string) => url),
  fromNow: jest.fn((dateString, customTxt = 'Invalid', alwaysInPast = true) => {
    if (!dateString) {
      return customTxt;
    }
    return 'a few seconds ago';
  }),
}));

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: {
          viewDashboardUrl: 'customUrl',
        },
        authentication: { isSuperUser: true },
      },
      featureFlipper: {
        getValueById: jest.fn(),
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

jest.mock('../../components/navigation-bar/NavigationBar', () => ({
  NavigationBar: () => null,
}));

jest.mock('../../components/settings-navigation/SettingsNavigation', () => ({
  SettingsNavigation: () => null,
}));

jest.mock(
  '../../pages/manage-aliases-page/hooks/use-fetch-dashboard-options.ts',
  () => ({
    useDashboardOptions: jest.fn(),
  })
);

jest.mock(
  '../../pages/manage-aliases-page/hooks/use-fetch-app-options.ts',
  () => ({
    useAppOptions: jest.fn(),
  })
);

beforeEach(() => {
  jest.mocked(getAllAliases).mockResolvedValue(mockAlias);

  jest.mocked(useDashboardOptions).mockReturnValue({
    dashboardOptions: [
      {
        id: 8,
        name: 'Taxi rides application',
        subgroups: [
          {
            id: 1,
            name: 'Sheet 1',
          },
        ],
      },
      {
        id: 12,
        name: 'Iris species w/ images',
        subgroups: [
          {
            id: 1,
            name: 'Sheet 1',
          },
        ],
      },
    ],
  });
  jest.mocked(useAppOptions).mockReturnValue({
    appOptions: [
      {
        id: 7,
        name: 'Sample App',
        subgroups: [
          {
            id: 8,
            name: 'Food Store Inspections - State View',
            apps: [93, 1],
          },
          {
            id: 9,
            name: 'Deficiency Details',
            apps: [94, 1],
          },
          {
            id: 10,
            name: 'Store Details',
            apps: [92, 1],
          },
        ],
      },
      {
        id: 15,
        name: 'test app',
        subgroups: [
          {
            id: 16,
            name: 'Cereal Comparisons',
            apps: [78, 1],
          },
        ],
      },
    ],
  });
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('Manage URL Alias component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ManageAliasesPage shouldShowModal={false} isAlreadySeen={false} />
      </QueryClientProvider>
    );
  });

  it('should display the manage url aliases table', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ManageAliasesPage shouldShowModal={false} isAlreadySeen={false} />
      </QueryClientProvider>
    );

    const URLAliasesColumnHeader = screen.getByRole('button', {
      name: 'URL Aliases',
    });
    expect(URLAliasesColumnHeader).toBeInTheDocument();

    const detailsColumnHeader = screen.getByRole('columnheader', {
      name: 'Details',
    });
    expect(detailsColumnHeader).toBeInTheDocument();

    const viewedColumnHeader = screen.getByRole('button', { name: 'Viewed' });
    expect(viewedColumnHeader).toBeInTheDocument();

    const lastViewedColumnHeader = screen.getByRole('button', {
      name: 'Last Viewed',
    });
    expect(lastViewedColumnHeader).toBeInTheDocument();

    const dashboardDataColumnHeader = screen.getByRole('button', {
      name: 'Dashboard/App (id, title)',
    });
    expect(dashboardDataColumnHeader).toBeInTheDocument();

    const numbersOfAliasesColumnHeader = screen.getByRole('button', {
      name: '# of Aliases',
    });
    expect(numbersOfAliasesColumnHeader).toBeInTheDocument();

    const createdColumnHeader = screen.getByRole('button', {
      name: 'Created',
    });
    expect(createdColumnHeader).toBeInTheDocument();

    const lastUpdatedColumnHeader = screen.getByRole('button', {
      name: 'Last Updated',
    });
    expect(lastUpdatedColumnHeader).toBeInTheDocument();

    const createdByColumnHeader = screen.getByRole('button', {
      name: 'Created By',
    });
    expect(createdByColumnHeader).toBeInTheDocument();
  });

  it('the first alias row should be disabled', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ManageAliasesPage shouldShowModal={false} isAlreadySeen={false} />
      </QueryClientProvider>
    );

    const rows = screen.getAllByRole('row');

    const disabledRow = rows[1];

    expect(disabledRow.classList).toContain('disabled');
  });

  it('should render the EditAliasModal on edit icon click', async () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ManageAliasesPage shouldShowModal={false} isAlreadySeen={false} />
      </QueryClientProvider>
    );

    const rows = screen.getAllByRole('row');

    const editIcon = within(rows[2]).getByTestId('edit-icon');

    fireEvent.click(editIcon);

    expect(
      await screen.findByText('Set the URL Alias to a dashboard or an app')
    ).toBeInTheDocument();
  });

  it('should open the delete alias confirmation modal', () => {
    render(
      <QueryClientProvider client={queryClient}>
        <ManageAliasesPage shouldShowModal={false} isAlreadySeen={false} />
      </QueryClientProvider>
    );

    const rows = screen.getAllByRole('row');

    const deleteIcon = within(rows[2]).getByTestId('delete-icon');

    fireEvent.click(deleteIcon);

    expect(screen.getByText('Delete URL Alias')).toBeInTheDocument();
  });
});
