import React, { useState } from 'react';
import classes from './ManageAliasesPage.module.scss';
import { Col, Container, Row } from 'react-bootstrap';
import { NavigationBar } from '../../components/navigation-bar/NavigationBar';
import { SettingsNavigation } from 'components/settings-navigation/SettingsNavigation';

import {
  getCoreRowModel,
  getPaginationRowModel,
  getSortedRowModel,
} from '@tanstack/react-table';
import { copyToClipboard } from 'arc/misc/misc';
import { ClouderaTable } from 'components/cloudera-table';
import { useQuery } from '@tanstack/react-query';
import {
  aliasesQueryToken,
  getAllAliases,
} from '../../apis/aliases/get-all-aliases';
import { useTable } from 'hooks/use-table';
import { EditAliasModal } from 'pages/manage-aliases-page/components/EditAliasModal';
import { AliasesColumnConfig } from './components/AliasesColumnConfig';

type UrlAliasesProps = {
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
};

export type Alias = {
  slug: string;
  parameters: {
    sheet?: string;
    initialapp?: string;
  };
  times_followed: number;
  last_viewed: string | null;
  dashboardId: string | number;
  alias_count: number;
  create_time: string;
  update_time: string;
  creator_name: string;
  visual_type: string;
  dashboard_name: string;
  is_created_in_email: boolean;
  path: string;
  fullUrl: string;
  isApp: boolean;
  isDashboard: boolean;
  isDisabled: boolean;
};

export const ManageAliasesPage = ({
  shouldShowModal,
  isAlreadySeen,
}: UrlAliasesProps): React.ReactElement => {
  const [isEditAliasModalOpen, setIsEditAliasModalOpen] = useState(false);
  const [selectedRowData, setSelectedRowData] = useState<Alias>();

  const { data: aliases, isLoading } = useQuery({
    queryKey: [aliasesQueryToken],
    queryFn: getAllAliases,
  });

  const handleCopy = (url: string) => {
    copyToClipboard(url).catch((error) => {
      console.error('Failed to copy to clipboard:', error);
    });
  };

  const handleEditAliasModal = (rowData: Alias) => {
    if (!rowData) return;
    setSelectedRowData(rowData);
    setIsEditAliasModalOpen(true);
  };

  const columns = AliasesColumnConfig(handleCopy, handleEditAliasModal);

  const table = useTable({
    data: aliases,
    columns,
    getCoreRowModel: getCoreRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    getSortedRowModel: getSortedRowModel(),
    state: {
      columnVisibility: {
        select: false,
        dashboard_name: false,
        visual_type: false,
        path: false,
        fullUrl: false,
        isApp: false,
        isDashboard: false,
        isDisabled: false,
        is_created_in_email: false,
      },
    },
    enableRowSelection: (row) => row.original.alias_count > 0,
  });

  return (
    <Container fluid>
      <NavigationBar
        isAlreadySeen={isAlreadySeen}
        shouldShowModal={shouldShowModal}
        activeTabs={{
          main: 'settings',
          sub: 'navbar-settings-menu-manage-url-aliases',
        }}
      />
      <SettingsNavigation activeTab="settings-aliases" color="white" />
      <form>
        <Row>
          <Col xs={12} md={12} className={classes.urlAliasesPage}>
            <div>
              <h3 className={classes.urlAliasesHeader}>
                Manage Dashboard URL Aliases
              </h3>
            </div>
            <ClouderaTable
              isLoading={isLoading}
              table={table}
              data-test-id="urlAliasesTable"
            />
          </Col>
        </Row>
      </form>
      {isEditAliasModalOpen && selectedRowData && (
        <EditAliasModal
          key={selectedRowData.slug}
          isShow={isEditAliasModalOpen}
          onClose={() => {
            setIsEditAliasModalOpen(false);
          }}
          rowData={selectedRowData}
        />
      )}
    </Container>
  );
};
