import React from 'react';
import { fromNow, getBlankTargetUrl } from 'arc/misc/misc';
import { createColumnHelper } from '@tanstack/react-table';

import { getGlobalArc } from 'utils/get-global-arc';
import { Alias } from '../ManageAliasesPage';
import classNames from 'classnames';
import {
  ClouderaDialog,
  ClouderaDialogTrigger,
} from 'components/cloudera-dialog/ClouderaDialog';
import { ClouderaLink } from 'components/cloudera-link/ClouderaLink';
import { Badge, Popover } from 'react-bootstrap';
import { DeleteAliasModal } from './DeleteAliasModal';
import { ParameterPopover } from './ParameterPopover';
import NewWindowIcon from '@cloudera-internal/cuix-core/icons/svg/ic-new-window.svg';
import EditIcon from '@cloudera-internal/cuix-core/icons/svg/ic-edit.svg';
import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';

import classes from '../ManageAliasesPage.module.scss';
import { VizTooltip } from 'components/VizTooltip';

const arc = getGlobalArc();

const columnHelper = createColumnHelper<Alias>();

export const AliasesColumnConfig = (
  handleCopy: (url: string) => void,
  handleEditAliasModal: (rowData: Alias) => void
) => {
  return [
    // those rows are invisible - need them for the data
    columnHelper.accessor('dashboard_name', {}),
    columnHelper.accessor('visual_type', {}),
    columnHelper.accessor('path', {}),
    columnHelper.accessor('fullUrl', {}),
    columnHelper.accessor('isApp', {}),
    columnHelper.accessor('isDashboard', {}),
    columnHelper.accessor('isDisabled', {}),
    columnHelper.accessor('is_created_in_email', {}),
    // need it for enableRowSelect - disable row
    columnHelper.display({ id: 'select' }),
    columnHelper.accessor('slug', {
      header: () => 'URL Aliases',
      cell: (info) => {
        const slug: string = info.getValue();
        const isDisabled: boolean = info.row.getValue('isDisabled');

        return (
          <div>
            <ClouderaLink
              aria-labelledby="visualIdLabel"
              href={getBlankTargetUrl(info.row.getValue('fullUrl'))}
              target="_blank"
              className="ml-2"
              disabled={isDisabled}
            >
              <span>{info.row.getValue('path')}</span>
              <span className="font-weight-bold current-alias-text">
                {slug}{' '}
              </span>
              <NewWindowIcon />
            </ClouderaLink>
          </div>
        );
      },
    }),
    columnHelper.accessor('parameters', {
      header: () => 'Details',
      cell: (info) => {
        const isCreatedInEmail: boolean = info.row.getValue(
          'is_created_in_email'
        );
        const isDisabled: boolean = info.row.getValue('isDisabled');
        const param = info.getValue();

        const parameterEntries = Object.keys(param)
          .filter((key) => key !== 'scope')
          .reduce<{ key: string; value: any; value_list: any[] }[]>(
            (cur, key) => {
              const value = param[key];
              cur.push({ key, value, value_list: [value] });
              return cur;
            },
            []
          );

        return (
          <div style={{ display: 'flex', alignItems: 'center', gap: '8px' }}>
            <ParameterPopover
              popOverTitle="Current Parameters"
              parameterEntries={parameterEntries}
            >
              <ClouderaLink
                aria-label="Filter parameters"
                disabled={isDisabled}
              >
                <i className="fa fa-filter" aria-hidden="true"></i>
              </ClouderaLink>
            </ParameterPopover>

            {isCreatedInEmail && (
              <ClouderaLink
                aria-label="Created in email"
                disabled={isDisabled}
                className={classes.emailIcon}
              >
                <VizTooltip
                  overlay={
                    <Popover id="emailTooltip">
                      <Popover.Content>
                        The alias has been created/used in an email.
                      </Popover.Content>
                    </Popover>
                  }
                >
                  <i className="fa fa-envelope-o" aria-hidden="true"></i>
                </VizTooltip>
              </ClouderaLink>
            )}
          </div>
        );
      },
      enableSorting: false,
    }),
    columnHelper.accessor('times_followed', {
      header: () => 'Viewed',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('last_viewed', {
      header: () => 'Last Viewed',
      cell: (info) => {
        const date = info.getValue();

        return date ? fromNow(date, 'Never', true) : 'Never';
      },
    }),
    columnHelper.accessor('dashboardId', {
      header: () => 'Dashboard/App (id, title)',
      cell: (info) => {
        const dashboardId: string | number = info.getValue();
        const dashboardName: string = info.row.getValue('dashboard_name');
        const aliasCount: number = info.row.getValue('alias_count');
        const isApp = info.row.getValue('isApp');
        const isDashboard = info.row.getValue('isDashboard');
        const isDisabled = info.row.getValue('isDisabled');

        let reportClassName = '';
        if (isApp) {
          reportClassName = classes.appIcon;
        } else if (isDashboard) {
          reportClassName = classes.dashboardIcon;
        } else {
          reportClassName = classes.visualIcon;
        }

        if (dashboardName !== '' && aliasCount > 0) {
          return (
            <div style={{ display: 'flex', alignItems: 'center', gap: '8px' }}>
              <div
                className={classNames(
                  classes.icon,
                  classes.reportIcon,
                  reportClassName
                )}
              />
              <span>
                <ClouderaLink
                  aria-labelledby="visualIdLabel"
                  href={getBlankTargetUrl(info.row.getValue('fullUrl'))}
                  target="_blank"
                  className="ml-2"
                >
                  {dashboardId} - {dashboardName}
                </ClouderaLink>
              </span>
            </div>
          );
        } else if (dashboardName !== '' && isDisabled) {
          return (
            <span>
              {dashboardId} - {dashboardName}
            </span>
          );
        } else if (dashboardName === '') {
          return (
            <div style={{ display: 'flex', alignItems: 'center', gap: '8px' }}>
              <div
                className={classNames(
                  classes.icon,
                  classes.reportIcon,
                  reportClassName
                )}
              />
              <span>
                <ClouderaLink
                  aria-labelledby="visualIdLabel"
                  href={getBlankTargetUrl(info.row.getValue('fullUrl'))}
                  target="_blank"
                  className="ml-2"
                >
                  {dashboardId} - Untitled {isApp ? 'app' : 'dashboard'}
                </ClouderaLink>
              </span>
            </div>
          );
        }
      },
    }),
    columnHelper.accessor('alias_count', {
      header: () => '# of Aliases',
      cell: (info) => {
        const numberOfAliasesPerDashboardOrApp = info.getValue() ?? 0;
        return (
          <Badge pill as="button" variant="secondary" className={classes.badge}>
            {numberOfAliasesPerDashboardOrApp}
          </Badge>
        );
      },
    }),
    columnHelper.accessor('create_time', {
      header: () => 'Created',
      cell: (info) => {
        const date = new Date(info.getValue());
        const day = date.getDate();
        const month = date.toLocaleString('default', { month: 'short' });
        const year = date.getFullYear();

        return `${month} ${day}, ${year}`;
      },
    }),
    columnHelper.accessor('update_time', {
      header: () => 'Last Updated',
      cell: (info) => {
        const date = info.getValue();

        return date ? fromNow(date, 'Never', true) : 'Never';
      },
    }),
    columnHelper.accessor('creator_name', {
      header: () => 'Created By',
      cell: (info) => info.getValue(),
    }),
    columnHelper.display({
      id: 'actions',
      header: () => ' ',
      cell: (info) => {
        const { path, slug, creator_name } = info.row.original;
        const isSuperUser = arc.frontend.globals.authentication.isSuperUser;
        const creatorUsername = arc.frontend.globals.authentication.username;
        const creatorOrSuperuser =
          isSuperUser || creator_name === creatorUsername;
        const isDisabled: boolean = info.row.getValue('isDisabled');

        return (
          <div style={{ display: 'flex', gap: '8px' }}>
            <ClouderaLink
              onClick={() => {
                handleEditAliasModal(info.row.original);
              }}
              disabled={isDisabled}
              testId="edit-icon"
            >
              <VizTooltip
                overlay={
                  <Popover id="editAliasTooltip">
                    <Popover.Content>Edit URL alias</Popover.Content>
                  </Popover>
                }
              >
                <EditIcon style={{ fontSize: '18px' }} />
              </VizTooltip>
            </ClouderaLink>

            <ClouderaLink
              onClick={() => handleCopy(`${path}${slug}`)}
              disabled={isDisabled}
            >
              <VizTooltip
                overlay={
                  <Popover id="copyToClipboardTooltip">
                    <Popover.Content>Copy to clipboard</Popover.Content>
                  </Popover>
                }
              >
                <i
                  className="fa fa-files-o existing url-copy-btn disabled"
                  aria-hidden="true"
                  data-toggle="tooltip"
                  data-placement="top"
                  title=""
                  data-original-title="Copy to clipboard"
                ></i>
              </VizTooltip>
            </ClouderaLink>

            <ClouderaDialog>
              <ClouderaDialogTrigger>
                <ClouderaLink
                  disabled={!creatorOrSuperuser || isDisabled}
                  testId="delete-icon"
                >
                  <VizTooltip
                    overlay={
                      <Popover id="deleteAliasTooltip">
                        <Popover.Content>Delete URL alias</Popover.Content>
                      </Popover>
                    }
                  >
                    <DeleteIcon />
                  </VizTooltip>
                </ClouderaLink>
              </ClouderaDialogTrigger>
              <DeleteAliasModal slug={slug} />
            </ClouderaDialog>
          </div>
        );
      },
    }),
  ];
};
