import React, { useCallback, useMemo } from 'react';
import Select from 'react-select';
import {
  Button,
  Form,
  InputGroup,
  Row,
  FormControl,
  Col,
  Modal,
  Container,
} from 'react-bootstrap';
import ParametersCard from './ParameterCard';
import { openInNewWindow } from 'arc/misc/misc';
import { Option, UrlAliasRow } from '../types/manage-aliases.type';
import * as yup from 'yup';
import { SubmitHandler, useForm, useWatch } from 'react-hook-form';
import { yupResolver } from '@hookform/resolvers/yup';
import { useDashboardOptions } from '../hooks/use-fetch-dashboard-options';
import { useAppOptions } from '../hooks/use-fetch-app-options';
import { useEditAliasMutation } from '../hooks/use-edit-alias-mutation';
import { setUrlValidationSchema } from '../utils/alias-schemas';
import { Alias } from '../ManageAliasesPage';
import { VizSpinner } from 'components/viz-spinner/VizSpinner';

export type SetUrlFormData = yup.InferType<typeof setUrlValidationSchema>;

type EditAliasModalProps = {
  isShow: boolean;
  onClose: () => void;
  rowData: Alias;
};

export const EditAliasModal = ({
  isShow,
  onClose,
  rowData,
}: EditAliasModalProps): JSX.Element => {
  const { dashboardId, visual_type, parameters } = rowData;

  const { dashboardOptions } = useDashboardOptions();
  const { appOptions } = useAppOptions();

  const options: Option[] =
    visual_type === 'dashboard' ? dashboardOptions : appOptions;

  const mainOption: Option | null = useMemo(() => {
    return (
      options.find(
        (option) =>
          (option && option.id === +dashboardId) ||
          option?.group_id === +dashboardId
      ) || null
    );
  }, [options, dashboardId]);

  const subOption: { id: number; name: string; apps?: number[] }[] | null =
    useMemo(() => {
      if (!mainOption) return null;

      return (
        mainOption?.subgroups?.filter((item) => {
          let itemId: number | number[];
          if (visual_type === 'dashboard') {
            if (!parameters.sheet) return false;
            itemId = +parameters?.sheet || 0;
            return item.id === itemId;
          } else {
            if (!parameters.initialapp) return false;
            itemId = +parameters?.initialapp;

            const [appId] = item.apps?.flat() || [];

            return appId === itemId;
          }
        }) || null
      );
    }, [mainOption, visual_type, parameters]);

  return (
    <Modal show={isShow} onHide={onClose} data-test-id="edit-alias-modal">
      <Modal.Header closeButton>
        <Modal.Title>Set the URL Alias to a dashboard or an app</Modal.Title>
      </Modal.Header>
      {options && mainOption && subOption ? (
        <AliasForm
          dashboardOptions={dashboardOptions}
          appOptions={appOptions}
          rowData={rowData}
          mainOption={mainOption}
          subOption={subOption}
          onClose={onClose}
        />
      ) : (
        <VizSpinner />
      )}
    </Modal>
  );
};

export type AliasFormProps = {
  dashboardOptions: Option[];
  appOptions: Option[];
  rowData: Alias;
  mainOption: Option | null;
  subOption: { id: number; name: string; apps?: number[] }[];
  onClose: () => void;
};

export const AliasForm = ({
  dashboardOptions,
  appOptions,
  rowData,
  mainOption,
  subOption,
  onClose,
}: AliasFormProps) => {
  const { visual_type, slug, path, is_created_in_email, parameters } = rowData;

  const updateType: string = 'managing';

  const oldAlias = rowData.slug;

  const { handleSubmit, register, formState, getValues, setValue, control } =
    useForm({
      resolver: yupResolver(setUrlValidationSchema),
      defaultValues: {
        type: visual_type === 'dashboard' ? 'dashboard' : 'apps',
        newAlias: slug,
        mainSelect: mainOption
          ? {
              id: mainOption.id || 0,
              name: mainOption.name || '',
              subgroups: mainOption.subgroups || [],
            }
          : undefined,
        subSelect: subOption ? subOption[0] : null,
        parameters: parameters,
        keepParameters: true,
      },
    });

  const type = useWatch({ control, name: 'type' });
  const mainSelect = useWatch({ control, name: 'mainSelect' });
  const subSelect = useWatch({ control, name: 'subSelect' });
  const newAlias = useWatch({ control, name: 'newAlias' });

  const handleTypeChange = (e: React.ChangeEvent<HTMLInputElement>) => {
    const { value } = e.target;
    setValue('type', value as 'dashboard' | 'apps', {
      shouldDirty: true,
      shouldValidate: true,
    });
    setValue('mainSelect', null);
    setValue('subSelect', null);
  };

  const createViewLink = () => {
    const linkToOpen =
      visual_type === 'dashboard'
        ? `/arc/apps/app/${mainSelect?.id}?sheet=${subSelect?.id}`
        : `/arc/apps/appgroup/${mainSelect?.id}?initialapp=${subSelect?.id}`;

    openInNewWindow(linkToOpen);
  };

  const getSubSelectValue = useCallback(
    (data: SetUrlFormData, type: 'apps' | 'dashboard'): string => {
      if (type === 'apps') {
        return data.subSelect?.apps?.[0]?.toString() || '0';
      }
      return data.subSelect?.id?.toString() || '0';
    },
    []
  );

  const updateParameters = useCallback(
    (
      url_data: {
        dashboardId: number;
        type: NonNullable<'apps' | 'dashboard' | undefined>;
        parameters: {};
        is_created_in_email: boolean | undefined;
      },
      restParameters: {},
      key: 'sheet' | 'initialapp',
      value: string
    ) => {
      url_data.parameters = {
        ...restParameters,
        [key]: value,
      };
    },
    []
  );

  const handleEditAlias = useEditAliasMutation();

  const onSubmit: SubmitHandler<SetUrlFormData> = (data) => {
    if (!data.mainSelect || !data.subSelect) {
      return;
    }

    const { type, keepParameters: shouldKeepParams } = data;
    const key: 'sheet' | 'initialapp' =
      type === 'dashboard' ? 'sheet' : 'initialapp';

    const url_data = {
      dashboardId: data.mainSelect.id,
      type,
      parameters: {},
      is_created_in_email: is_created_in_email,
    };

    const subSelectValue = getSubSelectValue(data, type);
    if (shouldKeepParams) {
      const parameters: { initialapp?: number; sheet?: number } =
        data.parameters || {};
      const { initialapp, sheet, ...restParameters } = parameters;

      updateParameters(url_data, restParameters, key, subSelectValue);
    } else {
      url_data.parameters = {
        [key]: subSelectValue,
      };
    }

    const payLoad: UrlAliasRow = {
      updateType: updateType,
      newAlias: oldAlias !== data.newAlias ? data.newAlias : '',
      originAlias: oldAlias,
      url_data: JSON.stringify(url_data),
    };

    handleEditAlias.mutate(payLoad);
    onClose();
  };

  return (
    <Form onSubmit={handleSubmit(onSubmit)} id="editAliasForm">
      <Modal.Body>
        <Container>
          <Row>
            <Form.Label>URL Alias</Form.Label>
            <InputGroup>
              <InputGroup.Prepend>
                <InputGroup.Text
                  className="short-url-origin"
                  data-initial-url-alias={newAlias}
                >
                  {path}
                </InputGroup.Text>
              </InputGroup.Prepend>
              <FormControl
                type="text"
                className="short-url-slug"
                isInvalid={!!formState.errors.newAlias}
                {...register('newAlias')}
              />
              {formState.errors.newAlias && (
                <Form.Control.Feedback type="invalid">
                  {formState.errors.newAlias.message}
                </Form.Control.Feedback>
              )}
            </InputGroup>
          </Row>
        </Container>
        <Row className="mt-3 mb-3">
          <Form.Label column md={2}>
            Type:
          </Form.Label>
          <Col md={3} className="align-items-center d-flex">
            <Form.Check
              id="dashboardOption"
              className="mr-1"
              type="radio"
              value="dashboard"
              checked={type === 'dashboard'}
              label="Dashboard"
              onChange={handleTypeChange}
            />
          </Col>
          <Col md={4} className="align-items-center d-flex">
            <Form.Check
              id="appOption"
              className="mr-1"
              type="radio"
              value="apps"
              checked={type === 'apps'}
              label="App"
              onChange={handleTypeChange}
            />
          </Col>
        </Row>
        <div className="d-flex gap-5">
          <Form.Group className="w-100">
            <Form.Label>
              Associated {type === 'dashboard' ? 'Dashboard' : 'App'}
            </Form.Label>
            <div
              data-test-id="setDashboardOrAppSelect"
              className="d-flex gap-3"
            >
              <Select
                inputId="dashboardOrAppSelect"
                className="w-100"
                classNamePrefix="form-select"
                isSearchable
                isClearable
                options={type === 'dashboard' ? dashboardOptions : appOptions}
                getOptionLabel={({ id, name }) => {
                  if (id && name) {
                    if (+id > 0) {
                      return `${name} (id: ${id})`;
                    }
                    return name;
                  } else if (id) {
                    return `Untitled ${
                      type === 'dashboard' ? 'dashboard' : 'app'
                    } (id: ${id})`;
                  }
                  return '';
                }}
                getOptionValue={({ id }) => `${id}`}
                placeholder="Please select an option"
                value={getValues('mainSelect')}
                {...register('mainSelect')}
                onChange={(data) => {
                  if (!data) {
                    setValue('mainSelect', null, {
                      shouldDirty: true,
                      shouldValidate: true,
                    });
                    return;
                  }

                  const isDashboard = type === 'dashboard';
                  const options = isDashboard ? dashboardOptions : appOptions;

                  const selectedOption = options.find(
                    (option) => option?.id === data.id
                  );

                  if (selectedOption) {
                    setValue('mainSelect', selectedOption, {
                      shouldDirty: true,
                      shouldValidate: true,
                    });
                  }
                }}
                required
              />
            </div>
            {formState.errors.mainSelect && (
              <div>
                <p className="text-danger mt-1">
                  {formState.errors.mainSelect.message}
                </p>
              </div>
            )}
          </Form.Group>
        </div>
        <div className="d-flex gap-5">
          <Form.Group className="w-100">
            <Form.Label>
              {type === 'dashboard' ? 'Sheet' : 'Menu Tab'}
            </Form.Label>
            <div
              data-test-id="setSheetOrMenuTabSelect"
              className="d-flex gap-3"
            >
              <Select
                name={type ? 'sheet' : 'subgroup'}
                inputId={type ? 'sheet' : 'subgroup'}
                className="w-100"
                classNamePrefix="form-select"
                isSearchable
                isClearable
                options={getValues('mainSelect')?.subgroups ?? []}
                getOptionLabel={({ id, name }) => {
                  if (+id > 0) {
                    return `${name} (Dashboard id: ${id})`;
                  }
                  return name;
                }}
                getOptionValue={({ id }) => `${id}`}
                value={getValues('subSelect')}
                onChange={(data) => {
                  setValue('subSelect', data, {
                    shouldDirty: true,
                    shouldValidate: true,
                  });
                }}
                placeholder="Please select an option"
                required
              />
            </div>
            {formState.errors.subSelect && (
              <div>
                <p className="text-danger mt-1">
                  {formState.errors.subSelect.message}
                </p>
              </div>
            )}
          </Form.Group>
        </div>
        <div className="d-flex gap-5">
          <Form.Group className="w-100">
            <Form.Label>
              Current {type === 'dashboard' ? 'Dashboard' : 'App'} Parameters
            </Form.Label>
            <div>
              {
                <>
                  <ParametersCard parameters={parameters} />
                  <Form.Check
                    className="mt-2 gap-5"
                    type="checkbox"
                    id="keepParametersCheckbox"
                    label="Keep the current parameters"
                    {...register('keepParameters')}
                    isInvalid={!!formState.errors.keepParameters}
                  />
                </>
              }
            </div>
          </Form.Group>
        </div>
        <div className="d-flex gap-5">
          <Button
            data-test-id="setDashboardOrAppView"
            variant="secondary"
            className="btn-sm"
            onClick={createViewLink}
          >
            <i className="fa fa-external-link"></i> View
          </Button>
        </div>
      </Modal.Body>
      <Modal.Footer>
        <div>
          <Button
            type="submit"
            disabled={!formState.isDirty || formState.isSubmitting}
          >
            Submit
          </Button>
        </div>
      </Modal.Footer>
    </Form>
  );
};
