import * as React from 'react';
import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import {
  configure,
  fireEvent,
  render,
  screen,
  waitFor,
} from '@testing-library/react';
import { getAllRoles } from 'pages/settings-page/apis/get-all-roles';
import { deleteRole } from 'pages/settings-page/apis/delete-role';
import { mockRoles } from './utils/mocks';
import { ManageRolesPage } from './ManageRolesPage';

configure({ testIdAttribute: 'data-test-id' });

const mockAddNotification = jest.fn();
jest.mock('../../../src/stores/use-notification-store.ts', () => ({
  useNotificationStore: () => ({ addNotification: mockAddNotification }),
}));

jest.mock('../../pages/settings-page/apis/get-all-roles.ts', () => ({
  getAllRoles: jest.fn(),
}));

jest.mock('../../pages/settings-page/apis/delete-role.ts', () => ({
  deleteRole: jest.fn(),
}));

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: {
          rbacRoles: 'customUrl',
        },
      },
      featureFlipper: {
        getValueById: jest.fn(),
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

jest.mock('../../components/navigation-bar/NavigationBar', () => ({
  NavigationBar: () => null,
}));

jest.mock('../../components/settings-navigation/SettingsNavigation', () => ({
  SettingsNavigation: () => null,
}));

beforeEach(() => {
  jest.mocked(getAllRoles).mockResolvedValue(mockRoles);
});

afterEach(() => {
  jest.resetAllMocks();
});

describe('Manage Roles page component', () => {
  it('renders without error', () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <ManageRolesPage shouldShowModal={false} isAlreadySeen={true} />
      </QueryClientProvider>
    );
  });

  it('should display the roles table', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <ManageRolesPage shouldShowModal={false} isAlreadySeen={true} />
      </QueryClientProvider>
    );

    const manageRolesPageHeader = await screen.findByText('Manage Roles');
    expect(manageRolesPageHeader).toBeInTheDocument();
    const roleColumnHeader = screen.getByRole('button', { name: 'Role' });
    expect(roleColumnHeader).toBeInTheDocument();
    const descColumnHeader = screen.getByRole('button', { name: 'Role' });
    expect(descColumnHeader).toBeInTheDocument();
    const compColumnHeader = screen.getByRole('button', { name: 'Role' });
    expect(compColumnHeader).toBeInTheDocument();
    const usersColumnHeader = screen.getByRole('button', { name: 'Role' });
    expect(usersColumnHeader).toBeInTheDocument();
    const groupsColumnHeader = screen.getByRole('button', { name: 'Role' });
    expect(groupsColumnHeader).toBeInTheDocument();
  });

  it('should open the Role details page on edit button', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <ManageRolesPage shouldShowModal={false} isAlreadySeen={true} />
      </QueryClientProvider>
    );

    Object.defineProperty(window, 'location', {
      value: {
        href: 'role_detail/',
      },
    });

    const editButton = await screen.findByTestId('editRole');
    fireEvent.click(editButton);

    expect(window.location.href).toContain('role_detail/');
  });

  it('should open delete dialog and delete role on confirm', async () => {
    render(
      <QueryClientProvider client={new QueryClient()}>
        <ManageRolesPage shouldShowModal={false} isAlreadySeen={true} />
      </QueryClientProvider>
    );

    const deleteButton = await screen.findByTestId('deleteRole');
    fireEvent.click(deleteButton);

    const input = await screen.findByRole('textbox', { name: /confirm/ });
    fireEvent.change(input, { target: { value: 'DELETE' } });

    fireEvent.click(screen.getByRole('button', { name: 'Delete' }));

    await waitFor(() => expect(deleteRole).toHaveBeenCalled());
    expect(mockAddNotification).toHaveBeenCalled();
  });
});
