import React, { useMemo, useState } from 'react';
import classNames from 'classnames';
import classes from './ManageRolesPage.module.scss';
import PlusIcon from '@cloudera-internal/cuix-core/icons/svg/ic-plus.svg';

import { NavigationBar } from '../../components/navigation-bar/NavigationBar';
import { SettingsNavigation } from 'components/settings-navigation/SettingsNavigation';
import { Badge, Col, Container, Row } from 'react-bootstrap';
import { useMutation, useQuery, useQueryClient } from '@tanstack/react-query';
import {
  PaginatedRoles,
  Privs,
  getAllRoles,
  rolesQueryToken,
} from 'pages/settings-page/apis/get-all-roles';
import { useNotificationStore } from 'stores/use-notification-store';
import {
  createColumnHelper,
  getCoreRowModel,
  PaginationState,
  SortingState,
} from '@tanstack/react-table';
import { ClouderaTable } from 'components/cloudera-table';
import { Role } from 'pages/settings-page/types/role-details.type';
import { VizButton } from 'components/viz-button/VizButton';
import { useToggle } from 'hooks/use-toggle';
import { deleteRole } from 'pages/settings-page/apis/delete-role';
import { DeleteDialog } from 'components/delete-dialog/DeleteDialog';
import { ClouderaButton } from '../../components/cloudera-button/ClouderaButton';
import DeleteIcon from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import { SearchBar } from '../visuals-page/components/search-bar/SearchBar';
import { useTable } from 'hooks/use-table';

type ManageRolesPageProps = {
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
};

const columnHelper = createColumnHelper<Role>();

const countPrivileges = (privs: Privs[]) => {
  const hasEmptySystemPriv = privs.some(
    (priv) => priv.ptype === 'system' && priv.perms.length === 0
  );

  return hasEmptySystemPriv ? privs.length - 1 : privs.length;
};

const loadNewOrExistingPage = (id?: number) => {
  if (id) {
    window.location.href = `role_detail/${id.toString()}`;
  } else {
    window.location.href = `role_detail`;
  }
};

export const ManageRolesPage: React.FC<ManageRolesPageProps> = ({
  shouldShowModal,
  isAlreadySeen,
}) => {
  const queryClient = useQueryClient();
  const [deleteDialogIsOpen, toggleDeleteDialog, closeDeleteDialog] =
    useToggle();

  const [selectedRowId, setSelectedRowId] = useState<number>(0);

  const [sorting, setSorting] = useState<SortingState>([
    { id: 'name', desc: false },
  ]);

  const [globalFilter, setGlobalFilter] = useState('');

  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const [pagination, setPagination] = useState<PaginationState>({
    pageIndex: 0,
    pageSize: 10,
  });

  const { data: paginatedRoles, isLoading: areRolesLoading } = useQuery({
    queryKey: [rolesQueryToken, pagination, sorting, globalFilter],
    queryFn: () =>
      getAllRoles({
        page: pagination.pageIndex + 1,
        pageSize: pagination.pageSize,
        sort: sorting[0]?.id,
        order: sorting[0]?.desc ? 'desc' : 'asc',
        searchTerm: table.getState().globalFilter,
      }),
    onError: () => {
      addNotification(
        {
          id: 'roles-danger',
          variant: 'danger',
          message: 'Error fetching roles.',
        },
        { delay: 3000 }
      );
    },
    select: (paginatedRoles) => paginatedRoles as PaginatedRoles,
  });

  const deleteRoleMutation = useMutation({
    mutationFn: (roleId: number) => deleteRole(roleId),
    onSuccess: () => {
      queryClient.invalidateQueries([rolesQueryToken]);
      table.resetRowSelection();
      closeDeleteDialog();

      addNotification(
        {
          variant: 'success',
          message: 'Deleted role',
        },
        { delay: 3000 }
      );
    },
    onError: () => {
      addNotification(
        {
          variant: 'danger',
          message: 'Failed to delete role',
        },
        { delay: 3000 }
      );
    },
  });

  const columns = useMemo(
    () => [
      columnHelper.accessor('name', {
        header: () => 'Role',
        cell: (info) => {
          const name: string = info.getValue();
          return (
            <div
              className={classNames(
                classes.manageRolesNoWrapRoleNames,
                'font-weight-bold'
              )}
            >
              {name}
            </div>
          );
        },
      }),
      columnHelper.accessor('desc', {
        header: () => 'Description',
        cell: (info) => {
          const desc: string = info.getValue();
          return (
            <div>
              <div>{desc}</div>
            </div>
          );
        },
        enableSorting: false,
      }),
      columnHelper.accessor('privs', {
        header: () => 'Components',
        cell: (info) => {
          const privs: Privs[] = info.getValue();

          return (
            <span>
              <Badge pill className={classes.manageRolesBadge}>
                {countPrivileges(privs)}
              </Badge>
            </span>
          );
        },
        enableSorting: false,
      }),
      columnHelper.accessor('users', {
        header: () => 'Users',
        cell: (info) => {
          const users: string[] = info.getValue();

          return (
            <span>
              <Badge pill className={classes.manageRolesBadge}>
                {users.length}
              </Badge>
            </span>
          );
        },
        enableSorting: false,
      }),
      columnHelper.accessor('groups', {
        header: () => 'Groups',
        cell: (info) => {
          const groups: string[] = info.getValue();
          const sentry_groups: string[] =
            info.row.getValue('sentry_groups') || [];

          return (
            <span>
              <Badge pill className={classes.manageRolesBadge}>
                {groups.length + sentry_groups.length}
              </Badge>
            </span>
          );
        },
        enableSorting: false,
      }),
      columnHelper.accessor('sentry_groups', {}),
      columnHelper.display({
        id: 'actions',
        header: () => '',
        cell: (info) => {
          const roleId = info.row.original.id;
          return (
            <div className="action-button-wrapper d-flex gap-3">
              <ClouderaButton
                icon={<span className="fa fa-pencil" />}
                onClick={() => loadNewOrExistingPage(roleId)}
                testId="editRole"
                variant="link"
              />
              <ClouderaButton
                icon={<DeleteIcon />}
                onClick={() => {
                  setSelectedRowId(roleId);
                  toggleDeleteDialog();
                }}
                testId="deleteRole"
                variant="link"
              />
            </div>
          );
        },
        enableSorting: false,
      }),
    ],
    [toggleDeleteDialog]
  );

  const data = useMemo(() => {
    return paginatedRoles?.roles || [];
  }, [paginatedRoles]);

  const table = useTable({
    data,
    columns,
    getCoreRowModel: getCoreRowModel(),
    rowCount: paginatedRoles?.total,
    manualPagination: true,
    manualSorting: true,
    onPaginationChange: setPagination,
    onSortingChange: setSorting,
    onGlobalFilterChange: setGlobalFilter,
    state: {
      pagination,
      sorting,
      globalFilter,
      columnVisibility: {
        sentry_groups: false,
      },
    },
    enableMultiRowSelection: false,
    getRowId: (row) => String(row.id),
  });

  return (
    <Container fluid>
      <NavigationBar
        isAlreadySeen={isAlreadySeen}
        shouldShowModal={shouldShowModal}
        activeTabs={{
          main: 'settings',
          sub: 'navbar-settings-menu-manage-roles',
        }}
      />
      <SettingsNavigation activeTab="settings-roles" color="white" />
      <Row>
        <Col xs={12} md={12} className={classes.manageRolesSummaryTable}>
          <div>
            <h3 className={classes.manageRolesHeader}>Manage Roles</h3>
          </div>
          <div className="d-flex justify-content-between">
            <VizButton
              color="secondary"
              size="small"
              onClick={() => loadNewOrExistingPage()}
              testId="newRole"
            >
              <span className="d-flex align-self-center align-items-center justify-content-center">
                <PlusIcon className="text-lg" />
                NEW ROLE
              </span>
            </VizButton>
            <SearchBar
              id="search-roles"
              placeholder="Search Role"
              testId="searchbox-roles"
              searchTerm={table.getState().globalFilter}
              setSearchTerm={table.setGlobalFilter}
              debounce={500}
            />
          </div>
          <ClouderaTable
            isLoading={areRolesLoading}
            table={table}
            className={classes.manageRolesTableStriped}
            data-test-id="manageRolesPage"
          />

          <DeleteDialog
            title="Delete Role"
            isOpen={deleteDialogIsOpen}
            close={closeDeleteDialog}
            onConfirm={() => deleteRoleMutation.mutateAsync(selectedRowId)}
            isLoading={deleteRoleMutation.isLoading}
          />
        </Col>
      </Row>
    </Container>
  );
};
