import React, { useEffect, useState, useCallback, useMemo } from 'react';
import { getGlobalArc } from 'utils/get-global-arc';
import { SetHomePageForm } from './components/set-home-page-form/SetHomePageForm';
import { NavigationBar } from 'components/navigation-bar/NavigationBar';
import { DashboardOption, FormType } from './types';
import classes from './SetHomePage.module.scss';
import classNames from 'classnames';
import { confirmAlert } from 'react-confirm-alert';
import 'react-confirm-alert/src/react-confirm-alert.css';
import { useNotificationStore } from 'stores/use-notification-store';
import { INITIAL_DASHBOARD_STATE } from './constants/initial-dashboard-option';
import { getReports, reportsQueryToken } from 'apis/get-reports';
import {
  submitUserDefinedHomePage,
  submitUserQueryToken,
} from './api/submit-user-defined-homepage';
import {
  submitAdminDefinedHomePage,
  submitAdminQueryToken,
} from './api/submit-admin-defined-homepage';
import {
  resetUserDefinedHomePage,
  resetUserQueryToken,
} from './api/reset-user-defined-homepage';
import {
  resetAdminDefinedHomePage,
  resetAdminQueryToken,
} from './api/reset-admin-defined-homepage';
import {
  overrideUserDefinedHomePage,
  overrideQueryToken,
} from './api/override-user-defined-homepage';
import { useMutation, useQuery } from '@tanstack/react-query';
import Modal from 'react-bootstrap/Modal';
import { VizButton } from 'components/viz-button/VizButton';

type SetHomepageProps = {
  userDefinedHomePage: DashboardOption;
  adminDefinedHomePage: DashboardOption;
  disableHomePageOverride: boolean;
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
};

const arc = getGlobalArc();

const hasSystemMiscPermission =
  arc.frontend.globals.permissions.system.includes('sys_misc');

const SetHomePage = ({
  userDefinedHomePage,
  adminDefinedHomePage,
  disableHomePageOverride,
  shouldShowModal,
  isAlreadySeen,
}: SetHomepageProps) => {
  const { data: dashboards, isFetching: areDashboardsFetching } = useQuery({
    queryKey: [reportsQueryToken],
    queryFn: () =>
      getReports({ reportType: 'dashboard' }).catch((error) => {
        console.error(error);
        return [];
      }),
  });
  const { mutateAsync: mutateSubmitUser } = useMutation(
    [submitUserQueryToken],
    submitUserDefinedHomePage
  );
  const { mutateAsync: mutateSubmitAdmin } = useMutation(
    [submitAdminQueryToken],
    submitAdminDefinedHomePage
  );
  const { mutateAsync: mutateResetUser } = useMutation(
    [resetUserQueryToken],
    resetUserDefinedHomePage
  );
  const { mutateAsync: mutateResetAdmin } = useMutation(
    [resetAdminQueryToken],
    resetAdminDefinedHomePage
  );
  const { mutateAsync: mutateOverride } = useMutation(
    [overrideQueryToken],
    overrideUserDefinedHomePage
  );
  const [userDefined, setUserDefined] = useState(INITIAL_DASHBOARD_STATE); // ui state
  const [adminDefined, setAdminDefined] = useState(INITIAL_DASHBOARD_STATE); // ui state
  const [userBackendState, setUserBackendState] = useState(
    INITIAL_DASHBOARD_STATE
  );
  const [adminBackendState, setAdminBackendState] = useState(
    INITIAL_DASHBOARD_STATE
  );
  const [disableOverride, setDisableOverride] = useState(false);
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  useEffect(() => {
    setDisableOverride(disableHomePageOverride);
  }, [disableHomePageOverride]);

  useEffect(() => {
    if (userDefinedHomePage.id !== INITIAL_DASHBOARD_STATE.id) {
      setUserDefined(userDefinedHomePage);
      setUserBackendState(userDefinedHomePage);
    }
  }, [userDefinedHomePage]);

  useEffect(() => {
    if (adminDefinedHomePage.id !== INITIAL_DASHBOARD_STATE.id) {
      setAdminDefined(adminDefinedHomePage);
      setAdminBackendState(adminDefinedHomePage);
    }
  }, [adminDefinedHomePage]);

  const memoizedHandleOverride = useCallback(
    (value: boolean) => {
      const submitOverride = async (value: boolean) => {
        try {
          await mutateOverride(value);

          setDisableOverride(value);
          addNotification(
            {
              id: 'override-success',
              variant: 'success',
              message: 'Setting has been saved',
            },
            { delay: 1500 }
          );
        } catch (error) {
          addNotification(
            {
              id: 'override-danger',
              variant: 'danger',
              message: `Failed to change setting. ${error}`,
            },
            { delay: 3000 }
          );
        }
      };

      confirmAlert({
        customUI: ({ onClose }) => {
          return (
            <Modal show={true} onHide={onClose} data-test-id="overrideModal">
              <Modal.Header closeButton>
                <Modal.Title>Please Confirm</Modal.Title>
              </Modal.Header>
              <Modal.Body>
                <p>{`Are you sure you want to ${
                  value ? 'prevent' : 'enable'
                } users to override the default homepage?`}</p>
              </Modal.Body>
              <Modal.Footer>
                <VizButton color="secondary" size="medium" onClick={onClose}>
                  Cancel
                </VizButton>
                <VizButton
                  color="primary"
                  size="medium"
                  testId="confirmOverride"
                  onClick={() => {
                    submitOverride(value);
                    onClose();
                  }}
                >
                  Yes
                </VizButton>
              </Modal.Footer>
            </Modal>
          );
        },
      });
    },
    [addNotification, mutateOverride]
  );

  const memoizedOnSubmit = useCallback(
    (type: FormType) => {
      const onSubmit = async (type: FormType) => {
        const isUserForm = type === 'user';

        const { backendState, dashboard, setBackendState } = isUserForm
          ? {
              backendState: userBackendState,
              dashboard: userDefined,
              setBackendState: setUserBackendState,
            }
          : {
              backendState: adminBackendState,
              dashboard: adminDefined,
              setBackendState: setAdminBackendState,
            };

        if (dashboard.id === backendState.id) {
          return;
        }

        try {
          await (isUserForm
            ? mutateSubmitUser(dashboard.id)
            : mutateSubmitAdmin(dashboard.id));

          setBackendState(dashboard);
          addNotification(
            {
              id: `submit-success-${dashboard.id}`,
              variant: 'success',
              message: isUserForm
                ? 'Homepage has been changed'
                : 'Default homepage has been changed',
            },
            { delay: 1500 }
          );
        } catch (error) {
          addNotification(
            {
              id: `submit-danger-${dashboard.id}`,
              variant: 'danger',
              message: `Failed to change the ${
                isUserForm ? '' : 'default'
              } homepage ${error}`,
            },
            { delay: 3000 }
          );
        }
      };

      onSubmit(type);
    },
    [
      addNotification,
      adminBackendState,
      adminDefined,
      mutateSubmitAdmin,
      mutateSubmitUser,
      userBackendState,
      userDefined,
    ]
  );

  const memoizedOnReset = useCallback(
    (type: FormType) => {
      const onReset = async (type: FormType) => {
        const isUserForm = type === 'user';
        const { backendState, newDashboard, setUIState, setBackendState } =
          isUserForm
            ? {
                backendState: userBackendState,
                newDashboard: adminBackendState,
                setUIState: setUserDefined,
                setBackendState: setUserBackendState,
              }
            : {
                backendState: adminBackendState,
                newDashboard: INITIAL_DASHBOARD_STATE,
                setUIState: setAdminDefined,
                setBackendState: setAdminBackendState,
              };

        if (
          backendState.id ===
          (isUserForm ? adminBackendState.id : INITIAL_DASHBOARD_STATE.id)
        ) {
          setUIState(newDashboard);
          return;
        }

        try {
          await (isUserForm
            ? mutateResetUser(adminBackendState.id)
            : mutateResetAdmin());

          setUIState(newDashboard);
          setBackendState(newDashboard);
          addNotification(
            {
              id: `reset-success-${newDashboard.id}`,
              variant: 'success',
              message: `Homepage was successfully set to ${
                isUserForm ? 'admin-defined' : ''
              } default`,
            },
            { delay: 1500 }
          );
        } catch (error) {
          addNotification(
            {
              id: `reset-danger-${newDashboard.id}`,
              variant: 'danger',
              message: `Failed to reset the homepage. ${error}`,
            },
            { delay: 3000 }
          );
        }
      };

      onReset(type);
    },
    [
      addNotification,
      adminBackendState,
      mutateResetAdmin,
      mutateResetUser,
      userBackendState,
    ]
  );

  const dashboardOptions = useMemo(
    () =>
      dashboards?.map(({ id, name }) => ({
        id: `${id}`,
        name,
      })) ?? [],
    [dashboards]
  );

  return (
    <section>
      <NavigationBar
        isAlreadySeen={isAlreadySeen}
        shouldShowModal={shouldShowModal}
        activeTabs={{
          main: 'settings',
          sub: 'navbar-settings-menu-set-home-page',
        }}
      />
      <div className="p-4">
        <h3
          data-test-id="pageHeader"
          className={classNames(classes.header, 'page-header mb-5 pb-4')}
        >
          Set Homepage
          <span className="d-block page-description">
            Set the Dashboard you would like as your homepage
          </span>
        </h3>
        <SetHomePageForm
          options={dashboardOptions}
          dashboard={userDefined}
          adminBackendState={adminBackendState}
          isAdmin={false}
          isDisabled={disableOverride}
          onDashboardChange={setUserDefined}
          onSubmit={memoizedOnSubmit}
          onReset={memoizedOnReset}
          areDashboardsFetching={areDashboardsFetching}
        />

        {(arc.frontend.globals.authentication.isSuperUser ||
          hasSystemMiscPermission) && (
          <>
            <div className="mb-4">
              <h4
                data-test-id="adminHeader"
                className={classNames(classes.adminHeader, 'pt-4')}
              >
                Admin Only
              </h4>
              <h5>Set the default homepage for all users</h5>
            </div>
            <SetHomePageForm
              options={dashboardOptions}
              dashboard={adminDefined}
              disableOverride={disableOverride}
              isAdmin={true}
              onDashboardChange={setAdminDefined}
              onOverrideChange={memoizedHandleOverride}
              onSubmit={memoizedOnSubmit}
              onReset={memoizedOnReset}
              areDashboardsFetching={areDashboardsFetching}
            />
          </>
        )}
      </div>
    </section>
  );
};

export { SetHomePage };
