/* eslint-disable testing-library/no-container */
/* eslint-disable testing-library/no-node-access */

import * as React from 'react';
import '@testing-library/jest-dom';
import { configure, render, fireEvent, screen } from '@testing-library/react';
import selectEvent from 'react-select-event';
import { INITIAL_DASHBOARD_STATE } from '../../constants/initial-dashboard-option';
import { SetHomePageForm } from './SetHomePageForm';
import { openInNewWindow } from 'arc/misc/misc';

const onDashboardChange = jest.fn();
const onSubmit = jest.fn();
const onReset = jest.fn();
const onOverrideChange = jest.fn();

configure({ testIdAttribute: 'data-test-id' });

jest.mock('components/VizTooltip', () => ({ VizTooltip: () => null }));
jest.mock('arc/misc/misc', () => ({
  openInNewWindow: jest.fn(),
}));

beforeEach(() => {
  jest.resetAllMocks();
});

describe('SetHomePageForm', () => {
  const options = [
    {
      id: '1',
      name: 'Option 1',
    },
    {
      id: '2',
      name: 'Option 2',
    },
    {
      id: '3',
      name: 'Option 3',
    },
    {
      id: '4',
      name: 'Option 4',
    },
  ];

  const dashboard = {
    id: '1',
    name: 'Option 1',
  };

  const validateCore = () => {
    const label = screen.getByTestId('setHomepageLabel');
    const select = screen.getByTestId('setHomepageSelect');
    const viewButton = screen.getByTestId('setHomepageView');
    const saveButton = screen.getByTestId('setHomepageSave');
    const resetButton = screen.getByTestId('setHomepageReset');

    expect(label).toBeInTheDocument();
    expect(label.textContent).toBe('Dashboard:');

    expect(select).toBeInTheDocument();

    expect(viewButton).toBeInTheDocument();
    expect(viewButton.textContent).toBe('View');

    expect(saveButton).toBeInTheDocument();
    expect(saveButton.textContent).toBe('Save');

    expect(resetButton).toBeInTheDocument();
  };

  it('renders the user form with initial values', () => {
    render(
      <SetHomePageForm
        options={options}
        dashboard={dashboard}
        onDashboardChange={onDashboardChange}
        onSubmit={onSubmit}
        onReset={onReset}
        isAdmin={false}
        areDashboardsFetching={false}
      />
    );

    validateCore();

    expect(screen.getByTestId('setHomepageReset').textContent).toBe(
      'Reset to admin-defined homepage'
    );
    expect(screen.queryByTestId('setHomepageDisable')).toBeNull();
  });

  it('renders the admin form with admin-only fields when isAdmin is true', () => {
    render(
      <SetHomePageForm
        options={options}
        dashboard={dashboard}
        onDashboardChange={onDashboardChange}
        onSubmit={onSubmit}
        onReset={onReset}
        onOverrideChange={onOverrideChange}
        isAdmin={true}
        areDashboardsFetching={false}
      />
    );

    validateCore();

    expect(screen.getByTestId('setHomepageReset').textContent).toBe(
      'Reset to default homepage'
    );

    expect(screen.getByTestId('setHomepageDisable')).toBeInTheDocument();
    expect(
      screen.getByLabelText('Disable user-override of homepage')
    ).toBeInTheDocument();
  });

  it('displays the selected dashboard and calls onDashboardChange when changed', async () => {
    render(
      <SetHomePageForm
        options={options}
        dashboard={dashboard}
        onDashboardChange={onDashboardChange}
        onSubmit={onSubmit}
        onReset={onReset}
        isAdmin={false}
        areDashboardsFetching={false}
      />
    );

    const form = screen.getByTestId('user-setHomepageForm');

    expect(form).toHaveFormValues({ dashboardSelect: '1' });

    await selectEvent.select(screen.getByRole('combobox'), 'Option 3 (id: 3)');

    expect(onDashboardChange).toHaveBeenCalledTimes(1);
    expect(onDashboardChange).toHaveBeenCalledWith(options[2]);
  });

  it('displays a disabled user form when isDisabled is true', () => {
    const { container } = render(
      <SetHomePageForm
        options={options}
        dashboard={dashboard}
        onDashboardChange={onDashboardChange}
        onSubmit={onSubmit}
        onReset={onReset}
        isAdmin={false}
        isDisabled={true}
        areDashboardsFetching={false}
      />
    );

    const select = container.querySelector('#dashboardSelect');
    const viewButton = screen.getByTestId('setHomepageView');
    const saveButton = screen.getByTestId('setHomepageSave');
    const resetButton = screen.getByTestId('setHomepageReset');

    expect(viewButton).toBeDisabled();
    expect(saveButton).toBeDisabled();
    expect(resetButton).toBeDisabled();
    expect(select).toBeDisabled();
  });

  it('displays a disabled View button if dashboard equals to INITIAL_STATE', () => {
    render(
      <SetHomePageForm
        options={options}
        dashboard={INITIAL_DASHBOARD_STATE}
        onDashboardChange={onDashboardChange}
        onSubmit={onSubmit}
        onReset={onReset}
        isAdmin={false}
        areDashboardsFetching={false}
      />
    );

    const viewButton = screen.getByTestId('setHomepageView');

    expect(viewButton).toBeDisabled();
  });

  it('calls openInNewWindow fn when View button is clicked', () => {
    render(
      <SetHomePageForm
        options={options}
        dashboard={dashboard}
        onDashboardChange={onDashboardChange}
        onSubmit={onSubmit}
        onReset={onReset}
        isAdmin={false}
        areDashboardsFetching={false}
      />
    );

    const viewButton = screen.getByTestId('setHomepageView');

    fireEvent.click(viewButton);

    expect(openInNewWindow).toHaveBeenCalledTimes(1);
    expect(openInNewWindow).toHaveBeenCalledWith('app/1');
  });

  it('calls onSubmit fn when Save button is clicked', () => {
    render(
      <SetHomePageForm
        options={options}
        dashboard={dashboard}
        onDashboardChange={onDashboardChange}
        onSubmit={onSubmit}
        onReset={onReset}
        isAdmin={false}
        areDashboardsFetching={false}
      />
    );

    const saveButton = screen.getByTestId('setHomepageSave');

    fireEvent.click(saveButton);

    expect(onSubmit).toHaveBeenCalledTimes(1);
    expect(onSubmit).toHaveBeenCalledWith('user');
  });

  it('calls onReset fn when Reset button is clicked', () => {
    render(
      <SetHomePageForm
        options={options}
        dashboard={dashboard}
        onDashboardChange={onDashboardChange}
        onSubmit={onSubmit}
        onReset={onReset}
        isAdmin={false}
        areDashboardsFetching={false}
      />
    );

    const resetButton = screen.getByTestId('setHomepageReset');

    fireEvent.click(resetButton);

    expect(onReset).toHaveBeenCalledTimes(1);
    expect(onReset).toHaveBeenCalledWith('user');
  });

  it('displays the override checkbox for admins and calls onOverrideChange when clicked', async () => {
    render(
      <SetHomePageForm
        options={options}
        dashboard={dashboard}
        onDashboardChange={onDashboardChange}
        disableOverride={true}
        onOverrideChange={onOverrideChange}
        onSubmit={onSubmit}
        onReset={onReset}
        isAdmin={true}
        areDashboardsFetching={false}
      />
    );

    const checkbox = screen.getByTestId(
      'setHomepageDisable'
    ) as HTMLInputElement;

    expect(checkbox.checked).toBe(true);

    fireEvent.click(checkbox);

    expect(onOverrideChange).toHaveBeenCalledTimes(1);
    expect(onOverrideChange).toHaveBeenCalledWith(false);
  });
});
