import React, { memo } from 'react';
import { Form, Button, FormCheck, Popover } from 'react-bootstrap';
import Select from 'react-select';
import classes from '../../SetHomePage.module.scss';
import { DashboardOption, FormType } from '../../types';
import classNames from 'classnames';
import { VizTooltip } from 'components/VizTooltip';
import { INITIAL_DASHBOARD_STATE } from '../../constants/initial-dashboard-option';
import { openInNewWindow } from 'arc/misc/misc';

type SetHomePageFormProps = {
  options: DashboardOption[];
  dashboard: DashboardOption;
  disableOverride?: boolean;
  adminBackendState?: DashboardOption;
  onDashboardChange: (val: DashboardOption) => void;
  onOverrideChange?: (val: boolean) => void;
  onSubmit: (type: FormType) => void;
  onReset: (type: FormType) => void;
  isDisabled?: boolean;
  isAdmin: boolean;
  areDashboardsFetching: boolean;
};

export const SetHomePageForm = memo(
  ({
    options,
    dashboard,
    disableOverride,
    adminBackendState,
    onDashboardChange,
    onOverrideChange,
    onSubmit,
    onReset,
    isDisabled,
    isAdmin,
    areDashboardsFetching,
  }: SetHomePageFormProps) => {
    const cannotBePreviewed =
      !dashboard.id || dashboard.id === INITIAL_DASHBOARD_STATE.id;

    const popover = (
      <Popover id="popover-basic">
        <Popover.Content>
          {isDisabled ? (
            <>
              <div>Changing the default homepage is disabled.</div>
              {adminBackendState?.id && (
                <div>{`Admin-defined homepage: ${adminBackendState.name} ${
                  +adminBackendState.id > 0 ? 'id: ' + adminBackendState.id : ''
                }`}</div>
              )}
            </>
          ) : (
            <>
              {cannotBePreviewed &&
                `${INITIAL_DASHBOARD_STATE.name} cannot be previewed.`}
            </>
          )}
        </Popover.Content>
      </Popover>
    );

    const formIdentifier = isAdmin ? 'admin' : 'user';

    return (
      <Form
        data-test-id={`${formIdentifier}-setHomepageForm`}
        className="ml-4 mb-5"
      >
        <div className={classNames('d-flex', classes.gap20)}>
          <Form.Label data-test-id="setHomepageLabel" className={classes.label}>
            Dashboard:
          </Form.Label>
          <div
            data-test-id="setHomepageSelect"
            className={classNames('d-flex', classes.gap15)}
          >
            <Select
              name="dashboardSelect"
              inputId="dashboardSelect"
              className={classNames(classes.select, 'select-container')}
              classNamePrefix="form-select"
              isSearchable
              isClearable
              options={[INITIAL_DASHBOARD_STATE, ...options]}
              isLoading={areDashboardsFetching}
              getOptionLabel={({ id, name }) => {
                if (+id > 0) {
                  return `${name} (id: ${id})`;
                }

                return name;
              }}
              getOptionValue={({ id }) => `${id}`}
              value={dashboard}
              isDisabled={isDisabled}
              onChange={(data) =>
                onDashboardChange(data ?? INITIAL_DASHBOARD_STATE)
              }
            />
            <div className={classNames('d-flex', classes.gap5)}>
              <Button
                data-test-id="setHomepageView"
                variant="secondary"
                disabled={isDisabled || cannotBePreviewed}
                onClick={() => openInNewWindow(`app/${dashboard.id}`)}
              >
                View
              </Button>
              <Button
                data-test-id="setHomepageSave"
                disabled={isDisabled}
                onClick={() => onSubmit(isAdmin ? 'admin' : 'user')}
              >
                Save
              </Button>

              {(isDisabled || cannotBePreviewed) && (
                <VizTooltip overlay={popover}>
                  <i className={classNames(classes.icon, 'info-icon')} />
                </VizTooltip>
              )}
            </div>
          </div>
        </div>
        <div className={classNames(classes.row, 'mt-2')}>
          <Button
            data-test-id="setHomepageReset"
            variant="link"
            className={classes.linkButton}
            disabled={isDisabled}
            onClick={() => onReset(isAdmin ? 'admin' : 'user')}
          >
            {`Reset to ${isAdmin ? 'default' : 'admin-defined'} homepage`}
          </Button>
        </div>
        {isAdmin && onOverrideChange && (
          <FormCheck
            data-test-id="setHomepageDisable"
            id="disableOverride"
            className={classNames(classes.checkbox, 'mt-4')}
            type="checkbox"
            checked={disableOverride}
            onChange={(e) => onOverrideChange(e.target.checked)}
            label={<span>Disable user-override of homepage</span>}
          />
        )}
      </Form>
    );
  }
);
