import React, { useEffect, useMemo, useState } from 'react';
import { getGlobalArc } from '../../utils/get-global-arc';
import { Col, Container, Form, Row } from 'react-bootstrap';
import { SubmitHandler, useForm } from 'react-hook-form';
import { useMutation, useQuery } from '@tanstack/react-query';
import * as yup from 'yup';
import { yupResolver } from '@hookform/resolvers/yup';

import classes from './RoleDetails.module.scss';
import classNames from 'classnames';
import Checkmark from '@cloudera-internal/cuix-core/icons/svg/ic-checkmark.svg';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import Import from '@cloudera-internal/cuix-core/icons/svg/ic-import.svg';
import User from '@cloudera-internal/cuix-core/icons/svg/ic-user.svg';

import { PermissionsSideNav } from './components/PermissionsSideNav';
import { SystemTable } from './components/SystemTable';
import { NavigationBar } from 'components/navigation-bar/NavigationBar';
import { SettingsNavigation } from '../../components/settings-navigation/SettingsNavigation';
import { VizButton } from 'components/viz-button/VizButton';

import {
  PERMISSION_DEPENDENCIES,
  PERMISSION_GROUPS,
} from './constants/permission-constants';

import { getRoleById, rolesByIdQueryToken } from './apis/get-role-by-id';
import { RolesTable } from './components/RolesTable';
import { ConnectionsTable } from './components/ConnectionsTable';
import { DatasetsTable } from './components/DatasetsTable';

import { getAllRoles, rolesQueryToken } from './apis/get-all-roles';
import {
  dataconnectionQueryToken,
  getAllDataConnections,
} from 'apis/get-all-data-connections';
import { datasetsQueryToken, getAllDatasets } from 'apis/get-all-datasets';
import { getUsers, usersQueryToken } from 'pages/visuals-page/apis/get-users';
import { getGroups, groupsQueryToken } from '../visuals-page/apis/get-groups';
import { submitRoles } from './apis/submit-roles';
import UsersTable from './components/UsersTable';
import { GroupsTable } from './components/GroupsTable';
import { useNotificationStore } from 'stores/use-notification-store';
import { transformFormData } from './utils/transform-formdata';
import {
  connectionsParsePermissions,
  datasetsParsePermissions,
  rolesParsePermissions,
  systemParsePermission,
} from './utils/parse-permissions';
import { Role } from 'pages/settings-page/types/role-details.type';
import { SentryGroupsTable } from './components/SentryGroupsTable';
import useBeforeUnload from 'hooks/use-before-unload';

const arc = getGlobalArc();

type RoleDetailsProps = {
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
  roleId?: string;
};

const rolesDetailsFormSchema = yup.object({
  roleId: yup.number(),
  name: yup.string().required(),
  description: yup.string().optional(),
  system: yup
    .object({
      sys_workspace: yup.boolean().required(),
      sys_viewroles: yup.boolean().required(),
      sys_editperm: yup.boolean().required(),
      sys_settings: yup.boolean().required(),
      sys_styles: yup.boolean().required(),
      sys_schedules: yup.boolean().required(),
      sys_viewlogs: yup.boolean().required(),
      sys_editconn: yup.boolean().required(),
      sys_misc: yup.boolean().required(),
    })
    .required(),
  roles: yup
    .array()
    .of(
      yup.object({
        role_ids: yup.array().of(yup.string().required()).required(),
        privileges: yup
          .object({
            rgds_manage: yup.boolean().required(),
            rgds_appedit: yup.boolean().required(),
            rgds_appview: yup.boolean().required(),
          })
          .required(),
      })
    )
    .required(),
  connections: yup
    .array()
    .of(
      yup.object({
        connection_ids: yup.array().of(yup.string().required()).required(),
        privileges: yup
          .object({
            dc_aviews: yup.boolean().required(),
            dc_upload: yup.boolean().required(),
            dc_explore: yup.boolean().required(),
          })
          .required(),
      })
    )
    .required(),
  datasets: yup
    .array()
    .of(
      yup.object({
        privileges: yup
          .object({
            ds_manage: yup.boolean().required(),
            ds_appedit: yup.boolean().required(),
            ds_appview: yup.boolean().required(),
          })
          .required(),
        dataset_ids: yup.array().of(yup.string().required()).required(),
        connection_id: yup.string().required(),
      })
    )
    .required(),
  users: yup.array().of(yup.string().required()).required(),
  groups: yup.array().of(yup.string().required()).required(),
  sentry_groups: yup.array().of(yup.string().required()).required(),
  require_all_groups: yup.boolean().required(),
});

const defaultPermissionValues: RoleDetailsFormData = {
  name: '',
  description: '',
  system: PERMISSION_GROUPS.system.privileges
    .map((privilege) => privilege.id)
    .reduce((obj, id) => {
      obj[id] = false;
      return obj;
    }, {} as Record<(typeof PERMISSION_GROUPS.system.privileges)[number]['id'], boolean>),
  roles: [],
  connections: [],
  datasets: [],
  users: [],
  groups: [],
  sentry_groups: [],
  require_all_groups: false,
};

export type RoleDetailsFormData = yup.InferType<typeof rolesDetailsFormSchema>;

export const RoleDetails = ({
  isAlreadySeen,
  shouldShowModal,
  roleId,
}: RoleDetailsProps): React.ReactElement => {
  const { mutateAsync: mutateRoles } = useMutation(submitRoles);
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));
  // get dataconnections
  const {
    data: { dataconnectionNamesById, dataconnectionOptions },
    isFetching: areDataconnectionsLoading,
  } = useQuery({
    queryKey: [dataconnectionQueryToken],
    queryFn: () => getAllDataConnections(),
    onError: () => {
      addNotification(
        {
          id: 'roles-danger',
          variant: 'danger',
          message: 'Error fetching dataconnections.',
        },
        { delay: 3000 }
      );
    },
    initialData: [],
    select: (dataconnections) => {
      const dataconnectionNamesById = new Map<number, string>([
        [-1, 'All connections'],
      ]);
      const dataconnectionOptions: { id: number; name: string }[] = [
        { id: -1, name: 'All connections' },
      ];

      dataconnections.forEach(({ id, name }) => {
        dataconnectionNamesById.set(id, name);
        dataconnectionOptions.push({ id, name });
      });

      return { dataconnectionNamesById, dataconnectionOptions };
    },
  });

  // get datasets
  const {
    data: { datasetNamesById, datasetOptions },
    isFetching: areDatasetsLoading,
  } = useQuery({
    queryKey: [datasetsQueryToken],
    queryFn: () => getAllDatasets({ isSummary: true }),
    onError: () => {
      addNotification(
        {
          id: 'roles-danger',
          variant: 'danger',
          message: 'Error fetching datasets.',
        },
        { delay: 3000 }
      );
    },
    initialData: [],
    select: (datasets) => {
      const datasetNamesById = new Map<number, string>([[-1, 'All datasets']]);
      const datasetOptions: {
        id: number;
        name: string;
        dc_id: number;
        dc_name: string;
      }[] = [
        { id: -1, name: 'All datasets', dc_name: 'All connections', dc_id: -1 },
      ];

      datasets.forEach(({ id, name, dc_id, dc_name }) => {
        datasetNamesById.set(id, name);
        datasetOptions.push({ id, name, dc_id, dc_name });
      });

      return { datasetNamesById, datasetOptions };
    },
  });

  // get roles
  const {
    data: { roleNamesById, roleOptions },
    isFetching: areRolesLoading,
  } = useQuery({
    queryKey: [rolesQueryToken],
    queryFn: getAllRoles,
    onError: () => {
      addNotification(
        {
          id: 'roles-danger',
          variant: 'danger',
          message: 'Error fetching roles.',
        },
        { delay: 3000 }
      );
    },
    initialData: [],
    select: (roles: Role[]) => {
      const roleNamesById = new Map<number, string>();
      const roleOptions: { id: number; name: string }[] = [];

      roles.forEach(({ id, name }) => {
        roleNamesById.set(id, name);

        roleOptions.push({ id, name });
      });

      return {
        roleNamesById,
        roleOptions,
      };
    },
  });

  // get groups
  const { data: groups } = useQuery({
    queryKey: [groupsQueryToken],
    queryFn: () =>
      getGroups({
        isSummary: true,
      }),
    onError: () => {
      addNotification(
        {
          id: 'roles-danger',
          variant: 'danger',
          message: 'Error fetching groups.',
        },
        { delay: 3000 }
      );
    },
  });

  // get users
  const { data: users } = useQuery({
    queryKey: [usersQueryToken],
    queryFn: () =>
      getUsers({
        isSummary: true,
      }),
    onError: () => {
      addNotification(
        {
          id: 'roles-danger',
          variant: 'danger',
          message: 'Error fetching users.',
        },
        { delay: 3000 }
      );
    },
  });

  const [hasInitialDataForEdit, setInitialDataForEdit] = useState(false);

  // reset will reset the useform once the roles coming back from the back
  const {
    handleSubmit,
    register,
    formState,
    watch,
    setValue,
    control,
    reset,
    getValues,
  } = useForm({
    resolver: yupResolver(rolesDetailsFormSchema),
    defaultValues: defaultPermissionValues,
  });
  const { data: currentPrivileges, isFetching: isCurrentPrivilegesFetching } =
    useQuery({
      queryKey: [rolesByIdQueryToken, roleId],
      enabled: !!roleId,
      queryFn: () => {
        if (roleId === undefined) {
          return undefined;
        }

        return getRoleById({ id: roleId });
      },
      onError: () => {
        addNotification(
          {
            id: 'roles-danger',
            variant: 'danger',
            message: 'Error fetching specific roles.',
          },
          { delay: 3000 }
        );
      },
    });

  useEffect(() => {
    if (
      currentPrivileges &&
      currentPrivileges.length > 0 &&
      !hasInitialDataForEdit
    ) {
      setInitialDataForEdit(true);
      const currentRoleData = currentPrivileges[0];

      const privs = currentRoleData.privs;

      const systemPermissions = privs.find(
        (priv) => priv.ptype === 'system'
      ) as { perms: string[]; ptype: string } | undefined;

      const connectionsPermissions = privs
        .filter((priv) => priv.ptype === 'dataconn')
        .map((connPriv) => ({
          ptype: connPriv.ptype,
          perms: connPriv.perms,
          dclist: connPriv.dclist ?? [],
        }));

      const datasetPermissions = privs
        .filter((priv) => priv.ptype === 'dataset')
        .map((dsPriv) => ({
          ptype: dsPriv.ptype,
          perms: dsPriv.perms,
          dslist: dsPriv.dslist ?? [],
          dcid: dsPriv.dcid ?? '',
        }));

      const rolePermissions = privs
        .filter((priv) => priv.ptype === 'role')
        .map((rolePriv) => ({
          perms: rolePriv.perms,
          ptype: rolePriv.ptype,
          rolelist: rolePriv.rolelist ?? [],
        }));

      const formData: RoleDetailsFormData = {
        name: currentPrivileges[0].name,
        description: currentPrivileges[0].desc,
        system: {
          sys_workspace: false,
          sys_viewroles: false,
          sys_editperm: false,
          sys_settings: false,
          sys_styles: false,
          sys_schedules: false,
          sys_viewlogs: false,
          sys_editconn: false,
          sys_misc: false,
        },
        connections: [],
        datasets: [],
        roles: [],
        users: currentRoleData.users,
        groups: currentRoleData.groups,
        sentry_groups: currentRoleData.sentry_groups,
        require_all_groups: currentRoleData.require_all_groups,
      };

      if (systemPermissions) {
        formData.system = systemParsePermission(systemPermissions);
      }

      if (connectionsPermissions.length > 0) {
        formData.connections = connectionsParsePermissions(
          connectionsPermissions
        );
      }

      if (datasetPermissions.length > 0) {
        formData.datasets = datasetsParsePermissions(datasetPermissions);
      }

      if (rolePermissions.length > 0) {
        formData.roles = rolesParsePermissions(rolePermissions);
      }

      if (roleId) {
        formData.roleId = +roleId;
      }

      reset(formData);
    }
  }, [currentPrivileges, hasInitialDataForEdit, reset, roleId]);

  useBeforeUnload(formState.isDirty);

  const name = watch('name');

  const onSubmit: SubmitHandler<RoleDetailsFormData> = async (
    preFormattedData
  ) => {
    const transformedData = transformFormData(
      preFormattedData,
      preFormattedData.roleId
    );

    try {
      const newRoles = await mutateRoles({
        data: JSON.stringify(transformedData),
      });

      addNotification(
        {
          id: 'roles-success',
          variant: 'success',
          message: 'Role has been saved',
        },
        { delay: 1500 }
      );

      reset({ ...preFormattedData, roleId: newRoles[0].id });
    } catch (error) {
      addNotification(
        {
          id: 'roles-danger',
          variant: 'danger',
          message: `Failed to change role. ${error}`,
        },
        { delay: 3000 }
      );
    }
  };

  const [selectedType, setSelectedType] = useState('privilege');

  const [activeTab, setActiveTab] = useState<keyof typeof PERMISSION_GROUPS>(
    () => {
      const isValidTab = (
        tab: string
      ): tab is keyof typeof PERMISSION_GROUPS => {
        const keys = Object.keys(PERMISSION_GROUPS);

        return keys.includes(tab);
      };

      const tabFromUrlHash = decodeURIComponent(
        window.location.hash.substring(1)
      );

      if (isValidTab(tabFromUrlHash)) {
        return tabFromUrlHash;
      }

      return 'system';
    }
  );

  const handleTabChange = (selectedTab: keyof typeof PERMISSION_GROUPS) => {
    setActiveTab(selectedTab);
    window.location.hash = selectedTab;
  };

  const formData = getValues();

  useEffect(() => {
    const systemPrivileges = formData['system'];
    let hasSystemChanges = false;
    const newSystemPrivileges = { ...systemPrivileges };

    Object.entries(systemPrivileges).forEach(([id, value]) => {
      if (!value) {
        return;
      }

      const dependencies = PERMISSION_DEPENDENCIES[id];
      if (!Array.isArray(dependencies)) {
        return;
      }

      dependencies.forEach((dependencyId) => {
        const isDependencyChecked = newSystemPrivileges[dependencyId];

        if (isDependencyChecked) {
          return;
        }

        newSystemPrivileges[dependencyId] = true;
        hasSystemChanges = true;
      });
    });

    if (hasSystemChanges) {
      setValue('system', newSystemPrivileges, {
        shouldDirty: true,
        shouldValidate: true,
      });
    }

    (['connections', 'roles', 'datasets'] as const).forEach((type) => {
      const arrayOfPrivileges = formData[type];
      let hasChanges = false;
      const newPrivileges = [...arrayOfPrivileges];

      arrayOfPrivileges.forEach(
        ({ privileges }: any, index: string | number) => {
          Object.entries(privileges).forEach(([id, value]) => {
            if (!value) {
              return;
            }

            const dependencies = PERMISSION_DEPENDENCIES[id];
            if (!Array.isArray(dependencies)) {
              return;
            }

            dependencies.forEach((dependencyId) => {
              const isDependencyChecked =
                newPrivileges[index].privileges[dependencyId];

              if (isDependencyChecked) {
                return;
              }

              newPrivileges[index].privileges[dependencyId] = true;
              hasChanges = true;
            });
          });

          if (hasChanges) {
            setValue(type, newPrivileges as any, {
              shouldDirty: true,
              shouldValidate: true,
            });
          }
        }
      );
    });
  }, [formData, setValue]);

  const system = watch('system');
  const datasets = watch('datasets');
  const connections = watch('connections');
  const roles = watch('roles');

  const numberOfPrivilegesByType = useMemo(() => {
    const map = {
      system: 0,
      roles: 0,
      datasets: 0,
      connections: 0,
    };

    if (Object.values(system).some((perm) => perm)) {
      map.system = 1;
    }

    datasets.forEach(({ privileges }) => {
      if (Object.values(privileges).some((perm) => perm)) {
        map.datasets += 1;
      }
    });

    connections.forEach(({ privileges }) => {
      if (Object.values(privileges).some((perm) => perm)) {
        map.connections += 1;
      }
    });

    roles.forEach(({ privileges }) => {
      if (Object.values(privileges).some((perm) => perm)) {
        map.roles += 1;
      }
    });

    return map;
  }, [connections, datasets, roles, system]);

  const numberOfPriviliges = useMemo(
    () =>
      Object.values(numberOfPrivilegesByType).reduce(
        (sum, length) => sum + length,
        0
      ),
    [numberOfPrivilegesByType]
  );

  const membersUsers = watch('users');
  const membersGroups = watch('groups');
  const membersSentryGroups = watch('sentry_groups');
  const sumMembers =
    membersUsers.length + membersGroups.length + membersSentryGroups.length;

  return (
    <>
      <Container fluid className={classes.roleDetailsContainer}>
        <NavigationBar
          isAlreadySeen={isAlreadySeen}
          shouldShowModal={shouldShowModal}
          activeTabs={{
            main: 'settings',
          }}
        />
        <SettingsNavigation activeTab="settings-roles" color="light" />
        <form onSubmit={handleSubmit(onSubmit)}>
          <Row>
            <Col xs={12} md={12} className={classes.roleDetailsPage}>
              <div>
                <Row>
                  <Col xs={12} md={12}>
                    <div className={classes.breadCrumbsWrapper}>
                      <ul className={classes.breadCrumbs}>
                        <li>
                          <a href={arc.frontend.globals.urls.rolesUrl}>Roles</a>
                        </li>
                        <li>Role Detail</li>
                      </ul>
                    </div>
                  </Col>
                </Row>

                <Row>
                  <Col xs={12} md={12}>
                    <div className={classes.roleNameWrapper}>
                      <div className={classes.roleName}>
                        <User />
                        <span>Role: {name}</span>
                      </div>
                    </div>
                  </Col>
                </Row>

                <Row className={'m-0'}>
                  <Col xs={12} md={12} className={classes.roleInputsWrapper}>
                    <Col md={6} className="p-0 pt-4 pb-4">
                      <Form.Group as={Col} className={'p-0'}>
                        <Form.Label htmlFor="name" className="mb-0">
                          Name
                        </Form.Label>
                        <Form.Control
                          data-test-id="role-name-input"
                          id="name"
                          type="text"
                          placeholder="Name"
                          isInvalid={!!formState.errors.name}
                          {...register('name')}
                        />
                        {formState.errors.name && (
                          <Form.Control.Feedback type="invalid">
                            {formState.errors.name.message}
                          </Form.Control.Feedback>
                        )}
                      </Form.Group>

                      <Form.Group as={Col} className={'p-0'}>
                        <Form.Label htmlFor="description" className="mb-0">
                          Description
                        </Form.Label>
                        <Form.Control
                          as="textarea"
                          data-test-id="role-desc-input"
                          id="description"
                          placeholder="Description"
                          isInvalid={!!formState.errors.description}
                          {...register('description')}
                        />
                        {formState.errors.description && (
                          <Form.Control.Feedback type="invalid">
                            {formState.errors.description.message}
                          </Form.Control.Feedback>
                        )}
                      </Form.Group>
                    </Col>
                  </Col>
                </Row>

                <div className={classes.roleDetailsTabs}>
                  <div
                    className={classNames(classes.roleDetailsTab, {
                      [classes.active]: selectedType === 'privilege',
                    })}
                    onClick={() => setSelectedType('privilege')}
                    data-test-id="privilege-view"
                  >
                    <span className="text-lg">
                      <Checkmark />
                    </span>
                    <span className="ml-1">
                      Privilege{' '}
                      <span className={classes.badgePill}>
                        {(!isCurrentPrivilegesFetching && numberOfPriviliges) ??
                          0}
                      </span>
                    </span>
                  </div>

                  <div
                    className={classNames(classes.roleDetailsTab, {
                      [classes.active]: selectedType === 'members',
                    })}
                    onClick={() => setSelectedType('members')}
                    data-test-id="members-view"
                  >
                    <span className="text-lg">
                      <User />
                    </span>
                    <span className="ml-1">
                      Members{' '}
                      <span className={classes.badgePill}>
                        {(!isCurrentPrivilegesFetching && sumMembers) ?? 0}
                      </span>
                    </span>
                  </div>
                </div>

                <div className={classes.privilegesAndMembersContainer}>
                  {selectedType === 'privilege' && (
                    <>
                      <PermissionsSideNav
                        activeTab={activeTab}
                        onClick={handleTabChange}
                        watch={watch}
                        areDataconnectionsLoading={areDataconnectionsLoading}
                        areDatasetsLoading={areDatasetsLoading}
                        areRolesLoading={areRolesLoading}
                        numberOfPrivilegesByType={numberOfPrivilegesByType}
                      />
                      <div
                        className={classes.privilegeTableWrapper}
                        data-test-id="role-priv-table"
                      >
                        <div>
                          {activeTab === 'system' && (
                            <SystemTable
                              register={register}
                              watch={watch}
                              setValue={setValue}
                            />
                          )}
                          {activeTab === 'roles' && (
                            <RolesTable
                              register={register}
                              control={control}
                              setValue={setValue}
                              roleNamesById={roleNamesById}
                              roleOptions={roleOptions}
                              areRolesLoading={areRolesLoading}
                            />
                          )}
                          {activeTab === 'connections' && (
                            <ConnectionsTable
                              register={register}
                              control={control}
                              setValue={setValue}
                              dataconnectionNamesById={dataconnectionNamesById}
                              dataconnectionOptions={dataconnectionOptions}
                              areDataconnectionsLoading={
                                areDataconnectionsLoading
                              }
                            />
                          )}
                          {activeTab === 'datasets' && (
                            <DatasetsTable
                              register={register}
                              control={control}
                              setValue={setValue}
                              dataconnectionNamesById={dataconnectionNamesById}
                              dataconnectionOptions={dataconnectionOptions}
                              datasetNamesById={datasetNamesById}
                              datasetOptions={datasetOptions}
                              areDatasetsLoading={areDatasetsLoading}
                              areDataconnectionsLoading={
                                areDataconnectionsLoading
                              }
                            />
                          )}
                        </div>
                      </div>
                    </>
                  )}

                  {selectedType === 'members' && (
                    <Container fluid className={classes.membersTab}>
                      <Row>
                        <UsersTable
                          users={users ?? []}
                          setValue={setValue}
                          watch={watch}
                          columnWidth={
                            formData.sentry_groups.length > 0 ? 4 : 6
                          }
                        />
                        <GroupsTable
                          groups={groups ?? []}
                          setValue={setValue}
                          watch={watch}
                          register={register}
                          columnWidth={
                            formData.sentry_groups.length > 0 ? 4 : 6
                          }
                        />

                        {formData.sentry_groups &&
                          formData.sentry_groups.length > 0 && (
                            <SentryGroupsTable
                              sentryGroups={formData.sentry_groups.map(
                                (groupName) => ({ name: groupName })
                              )}
                            />
                          )}
                      </Row>
                    </Container>
                  )}
                </div>

                <div className={classes.roleDetailsButtonWrapper}>
                  <VizButton
                    color="secondary"
                    size="medium"
                    testId="#undo-role-btn"
                    onClick={() => {
                      reset();
                    }}
                  >
                    <div className={classes.roleDetailsButtonInner}>
                      <div className={classes.roleDetailsButtonIcon}>
                        <Delete />
                      </div>
                      <div>Discard changes</div>
                    </div>
                  </VizButton>
                  <div className="mr-2"></div>
                  <VizButton
                    color="primary"
                    size="medium"
                    disabled={!formState.isDirty || formState.isSubmitting}
                    type="submit"
                    testId="save-role-btn"
                  >
                    <div className={classes.roleDetailsButtonInner}>
                      <div className={classes.roleDetailsButtonIcon}>
                        <Import />
                      </div>
                      <div>Apply changes</div>
                    </div>
                  </VizButton>
                </div>
              </div>
            </Col>
          </Row>
        </form>
      </Container>
    </>
  );
};
