import React, { useState } from 'react';
import { Form, Table } from 'react-bootstrap';
import classes from './../RoleDetails.module.scss';
import {
  Control,
  UseFormRegister,
  UseFormSetValue,
  useFieldArray,
} from 'react-hook-form';
import { RoleDetailsFormData } from '../RoleDetails';
import { PERMISSION_GROUPS } from '../constants/permission-constants';
import { VizButton } from 'components/viz-button/VizButton';
import CirclePlus from '@cloudera-internal/cuix-core/icons/svg/ic-plus-circle.svg';
import { useToggle } from 'hooks/use-toggle';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import { AddConnetionPrivilegeModal } from './AddConnetionPrivilegeModal';
import { Notification } from 'components/notification/Notification';
import { CircularProgressBar } from 'pages/visuals-page/components/common/CircularProgessBar';
import classNames from 'classnames';

type ConnectionsTableProps = {
  register: UseFormRegister<RoleDetailsFormData>;
  setValue: UseFormSetValue<RoleDetailsFormData>;
  control: Control<RoleDetailsFormData>;
  dataconnectionNamesById: Map<number, string>;
  dataconnectionOptions: { id: number; name: string }[];
  areDataconnectionsLoading: boolean;
};

const connectionsDependencyMessage = `
            The <strong>Create datasets, explore tables</strong> is mandatory if the privilege contains the <strong>Manage AVs/Extracts</strong> or <strong>Import data Connections</strong> permission and cannot be removed. 
            The checkboxes are filled in and fixed.<br><br>

            If <strong>Manage AVs/Extracts</strong> is deselected, <strong>Create datasets, explore tables</strong> becomes optional
            and can also be deselected.
            If <strong>Import data</strong> is deselected, <strong>Create datasets, explore tables</strong> becomes optional 
            and can also be deselected.`;

export const ConnectionsTable = ({
  register,
  control,
  dataconnectionNamesById,
  dataconnectionOptions,
  areDataconnectionsLoading,
}: ConnectionsTableProps): React.ReactElement => {
  const [didUserConfirmNotification, setDidUserConfirmNotification] =
    useState(false);
  const { privileges } = PERMISSION_GROUPS.connections;

  const { fields, append, remove, update } = useFieldArray({
    control,
    name: 'connections',
  });

  const [
    addConnectionPrivilegeModalIsOpen,
    toggleAddConnectionPrivilegeModal,
    closeAddConnectionPrivilegeModal,
  ] = useToggle();

  const handleNotificationClose = () => {
    setDidUserConfirmNotification(true);
  };

  return (
    <>
      <div className={classes.addPrivilegesButtonWrapper}>
        <VizButton
          color="secondary"
          size="medium"
          onClick={toggleAddConnectionPrivilegeModal}
          testId="add-privilege-btn"
        >
          <div className={classes.addPrivilegesButtonInner}>
            <div className={classes.addPrivilegesButtonIcon}>
              <CirclePlus />
            </div>
            <div>Add Connections</div>
          </div>
        </VizButton>
      </div>
      <div className={classes.dependencyContainer}>
        <Notification
          title="Privilege Dependency Information"
          useHtml
          message={connectionsDependencyMessage}
          variant="info"
          dismissible
          show={
            !didUserConfirmNotification &&
            fields.some(
              (field) =>
                field.privileges['dc_upload'] || field.privileges['dc_aviews']
            )
          }
          onClose={handleNotificationClose}
        />
      </div>
      <Table
        className={classes.privilegeContainer}
        data-test-id="connection-privilege-container"
      >
        <thead>
          <tr>
            <th className={classes.privilegeHeader}></th>
            <th className={classes.privilegeHeader}></th>
            {privileges.map(({ id, label, description }) => (
              <th className={classes.privilegeHeader} key={id}>
                <div className="font-medium">{label}</div>
                <div className="text-small font-normal">{description}</div>
              </th>
            ))}
            <th className={classes.privilegeHeader}></th>
          </tr>
        </thead>
        <tbody>
          {fields.map((field, index) => (
            <tr key={field.id}>
              <td className={classes.privilegeBody}>
                {areDataconnectionsLoading ? (
                  <CircularProgressBar />
                ) : (
                  field.connection_ids
                    ?.map((id) => dataconnectionNamesById.get(+id) || 'No Name')
                    .join(' / ') || 'No Name'
                )}
              </td>
              {(
                [
                  {
                    id: 'selectAll',
                    label: '',
                    description: '',
                  },
                  ...privileges,
                ] as const
              ).map((permission) => (
                <td
                  className={classNames(classes.privilegeBody, 'text-center')}
                  key={permission.id}
                >
                  <Form.Check
                    data-test-id={`${permission.id}`}
                    id={permission.id}
                    type="checkbox"
                    className={classes.privilegeCheckbox}
                    {...(permission.id === 'selectAll'
                      ? {
                          onChange: (e) => {
                            update(index, {
                              ...field,
                              privileges:
                                PERMISSION_GROUPS.connections.privileges
                                  .map((privilege) => privilege.id)
                                  .reduce((obj, id) => {
                                    obj[id] = e.target.checked;
                                    return obj;
                                  }, {} as Record<(typeof PERMISSION_GROUPS.connections.privileges)[number]['id'], boolean>),
                            });
                          },
                          checked: Object.values(field.privileges ?? {}).every(
                            (checked) => checked
                          ),
                        }
                      : register(
                          `connections.${index}.privileges.${permission.id}`
                        ))}
                  />
                </td>
              ))}
              <td
                className={classes.privilegeBody}
                key={`${field.id}-${index}`}
              >
                <Delete onClick={() => remove(index)} />
              </td>
            </tr>
          ))}
        </tbody>
      </Table>

      <AddConnetionPrivilegeModal
        isOpen={addConnectionPrivilegeModalIsOpen}
        dataconnectionOptions={dataconnectionOptions}
        close={closeAddConnectionPrivilegeModal}
        areDataconnectionsLoading={areDataconnectionsLoading}
        onSave={(dataconnections) => {
          append({
            privileges: PERMISSION_GROUPS.connections.privileges
              .map((privilege) => privilege.id)
              .reduce((obj, id) => {
                obj[id] = true;
                return obj;
              }, {} as Record<(typeof PERMISSION_GROUPS.connections.privileges)[number]['id'], boolean>),
            connection_ids: dataconnections.map(({ id }) => id.toString()),
          });
        }}
      />
    </>
  );
};
