import React, { useState } from 'react';
import { Form, Table } from 'react-bootstrap';
import classes from './../RoleDetails.module.scss';
import {
  Control,
  UseFormRegister,
  UseFormSetValue,
  useFieldArray,
} from 'react-hook-form';
import { RoleDetailsFormData } from '../RoleDetails';
import { PERMISSION_GROUPS } from '../constants/permission-constants';
import { VizButton } from 'components/viz-button/VizButton';
import CirclePlus from '@cloudera-internal/cuix-core/icons/svg/ic-plus-circle.svg';
import { useToggle } from 'hooks/use-toggle';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import { AddDatasetPrivilegeModal } from './AddDatasetPrivilegeModal';
import { Notification } from 'components/notification/Notification';
import { CircularProgressBar } from 'pages/visuals-page/components/common/CircularProgessBar';
import classNames from 'classnames';

type DatasetsTableProps = {
  register: UseFormRegister<RoleDetailsFormData>;
  setValue: UseFormSetValue<RoleDetailsFormData>;
  control: Control<RoleDetailsFormData>;
  dataconnectionNamesById: Map<number, string>;
  dataconnectionOptions: { id: number; name: string }[];
  datasetNamesById: Map<number, string>;
  datasetOptions: {
    id: number;
    name: string;
    dc_id: number;
    dc_name: string;
  }[];
  areDatasetsLoading: boolean;
  areDataconnectionsLoading: boolean;
};

const datasetsDependencyMessage = `
            The <strong>Manage dashboards</strong> and <strong>View dashboards</strong> are mandatory if the privilege contains the 
            <strong>Manage dataset</strong> permission and cannot be removed. The checkboxes are filled in and fixed.<br><br>

            If <strong>Manage dataset</strong> is deselected, <strong>Manage dashboards</strong> becomes optional and can also be deselected.<br><br>
            If <strong>Manage dashboards</strong> is deselected, <strong>View dashboards</strong> becomes optional and can also be deselected.`;

export const DatasetsTable = ({
  register,
  control,
  dataconnectionNamesById,
  dataconnectionOptions,
  datasetNamesById,
  datasetOptions,
  areDatasetsLoading,
  areDataconnectionsLoading,
}: DatasetsTableProps): React.ReactElement => {
  const [didUserConfirmNotification, setDidUserConfirmNotification] =
    useState(false);
  const { privileges } = PERMISSION_GROUPS.datasets;

  const { fields, append, remove, update } = useFieldArray({
    control,
    name: 'datasets',
  });

  const [
    addDatasetPrivilegeModalIsOpen,
    toggleAddDatasetPrivilegeModal,
    closeAddDatasetPrivilegeModal,
  ] = useToggle();

  const handleNotificationClose = () => {
    setDidUserConfirmNotification(true);
  };

  return (
    <>
      <div className={classes.addPrivilegesButtonWrapper}>
        <VizButton
          color="secondary"
          size="medium"
          onClick={toggleAddDatasetPrivilegeModal}
          testId="add-privilege-btn"
        >
          <div className={classes.addPrivilegesButtonInner}>
            <div className={classes.addPrivilegesButtonIcon}>
              <CirclePlus />
            </div>
            <div>Add Datasets</div>
          </div>
        </VizButton>
      </div>
      <div className={classes.dependencyContainer}>
        <Notification
          title="Privilege Dependency Information"
          useHtml
          message={datasetsDependencyMessage}
          variant="info"
          dismissible
          show={
            !didUserConfirmNotification &&
            fields.some(
              (field) =>
                field.privileges['ds_manage'] || field.privileges['ds_appedit']
            )
          }
          onClose={handleNotificationClose}
        />
      </div>

      <Table
        className={classes.privilegeContainer}
        data-test-id="dataset-privilege-container"
      >
        <thead>
          <tr>
            <th className={classes.privilegeHeader}></th>

            <th className={classes.privilegeHeader}></th>

            {privileges.map(({ id, label, description }) => (
              <th className={classes.privilegeHeader} key={id}>
                <div className="font-medium">{label}</div>
                <div className="text-small font-normal">{description}</div>
              </th>
            ))}

            <th className={classes.privilegeHeader}></th>
          </tr>
        </thead>
        <tbody>
          {fields.map((field, index) => {
            const connectionName = areDataconnectionsLoading ? (
              <CircularProgressBar />
            ) : (
              dataconnectionNamesById.get(+field.connection_id) || 'No name'
            );

            const datasetName = areDatasetsLoading ? (
              <CircularProgressBar />
            ) : (
              field.dataset_ids
                .map((id) => datasetNamesById.get(+id) || 'No name')
                .join(' / ')
            );

            return (
              <tr key={field.id}>
                <td className={classes.privilegeBody}>
                  <span className={classes.connectionName}>
                    {connectionName} /{' '}
                  </span>
                  {datasetName}
                </td>
                {(
                  [
                    {
                      id: 'selectAll',
                      label: '',
                      description: '',
                    },
                    ...privileges,
                  ] as const
                ).map((permission) => (
                  <td
                    className={classNames(classes.privilegeBody, 'text-center')}
                    key={permission.id}
                  >
                    <Form.Check
                      data-test-id={`${permission.id}`}
                      id={permission.id}
                      type="checkbox"
                      className={classes.privilegeCheckbox}
                      {...(permission.id === 'selectAll'
                        ? {
                            onChange: (e) => {
                              update(index, {
                                ...field,
                                privileges:
                                  PERMISSION_GROUPS.datasets.privileges
                                    .map((privilege) => privilege.id)
                                    .reduce((obj, id) => {
                                      obj[id] = e.target.checked;
                                      return obj;
                                    }, {} as Record<(typeof PERMISSION_GROUPS.datasets.privileges)[number]['id'], boolean>),
                              });
                            },
                            checked: Object.values(
                              field.privileges ?? {}
                            ).every((checked) => checked),
                          }
                        : register(
                            `datasets.${index}.privileges.${permission.id}`
                          ))}
                    />
                  </td>
                ))}

                <td
                  className={classes.privilegeBody}
                  key={`${field.id}-${index}`}
                >
                  <Delete onClick={() => remove(index)} />
                </td>
              </tr>
            );
          })}
        </tbody>
      </Table>

      <AddDatasetPrivilegeModal
        close={closeAddDatasetPrivilegeModal}
        isOpen={addDatasetPrivilegeModalIsOpen}
        areDatasetsLoading={areDatasetsLoading}
        areDataconnectionsLoading={areDataconnectionsLoading}
        onSave={(datasets, dataconnection) => {
          append({
            privileges: PERMISSION_GROUPS.datasets.privileges
              .map((privilege) => privilege.id)
              .reduce((obj, id) => {
                obj[id] = true;
                return obj;
              }, {} as Record<(typeof PERMISSION_GROUPS.datasets.privileges)[number]['id'], boolean>),
            dataset_ids: datasets.map(({ id }) => id.toString()),
            connection_id: dataconnection.id.toString(),
          });
        }}
        dataconnectionOptions={dataconnectionOptions}
        datasetOptions={datasetOptions}
      />
    </>
  );
};
