import React, { useMemo } from 'react';
import { Col, Form, Row } from 'react-bootstrap';
import { VizButton } from 'components/viz-button/VizButton';
import Groups from '@cloudera-internal/cuix-core/icons/svg/ic-groups.svg';
import classes from '../RoleDetails.module.scss';
import { useToggle } from 'hooks/use-toggle';
import {
  UseFormRegister,
  UseFormSetValue,
  UseFormWatch,
} from 'react-hook-form';
import { RoleDetailsFormData } from '../RoleDetails';
import type { Group } from '../types/role-details.type';
import {
  createColumnHelper,
  getCoreRowModel,
  getPaginationRowModel,
  getSortedRowModel,
  getFilteredRowModel,
} from '@tanstack/react-table';
import { ClouderaTable } from 'components/cloudera-table';
import { UsersOrGroupsModal } from './UsersOrGroupsModal';
import { useTable } from 'hooks/use-table';

const columnHelper = createColumnHelper<Group>();

type GroupsTableProps = {
  groups: Group[];
  setValue: UseFormSetValue<RoleDetailsFormData>;
  watch: UseFormWatch<RoleDetailsFormData>;
  register: UseFormRegister<RoleDetailsFormData>;
  columnWidth: number;
};

export const GroupsTable = ({
  groups,
  watch,
  setValue,
  register,
  columnWidth,
}: GroupsTableProps): React.ReactElement => {
  const groupNames = watch('groups');
  const requireAllGroups = watch('require_all_groups');

  const columns = useMemo(
    () => [
      columnHelper.accessor('name', {
        header: () => 'Group name',
        cell: (info) => info.getValue(),
      }),
    ],
    []
  );

  const currentGroups = useMemo(() => {
    return groupNames
      .map(
        (name) =>
          groups?.find((group) => group.name === name) ?? {
            id: name,
            name: name,
          }
      )
      .filter((group: Group | undefined): group is Group => Boolean(group));
  }, [groups, groupNames]);

  const table = useTable({
    data: currentGroups,
    columns,
    getCoreRowModel: getCoreRowModel(),
    getSortedRowModel: getSortedRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    getFilteredRowModel: getFilteredRowModel(),
    initialState: {
      pagination: {
        pageIndex: 0,
        pageSize: 10,
      },
    },
  });

  const handleSearchChange = (event: React.ChangeEvent<HTMLInputElement>) => {
    table.setGlobalFilter(event.target.value);
  };

  const [
    openUsersOrGroupsModal,
    toggleUsersOrGroupsModal,
    closeUsersOrGroupsModal,
  ] = useToggle();

  return (
    <Col sm md={columnWidth} className={classes.groupsColumn}>
      <div style={{ minHeight: '145px' }}>
        <Row className="d-flex justify-content-between align-items-center pt-2 pb-0 pl-3 pr-3">
          <div className="d-flex align-items-center">
            <Groups className={classes.groupIcon} />
            <h4>Groups</h4>
          </div>
        </Row>
        <Row className="d-flex justify-content-start  py-2 px-3">
          <div className="d-flex justify-content-start">
            <Form.Group controlId="require_all_groups" className="mb-0">
              <Form.Check
                type="checkbox"
                id={`require_all_groups`}
                checked={requireAllGroups}
                {...register('require_all_groups')}
                label="Require all groups"
              />
            </Form.Group>
          </div>
        </Row>

        <Row className="d-flex justify-content-between py-2 px-3">
          <div className="d-flex pr-0">
            <Form.Group
              as={Row}
              className="align-items-center mb-0 flex-nowrap mr-0 ml-0"
            >
              <Form.Label className="pr-1 mb-0">Search:</Form.Label>
              <Form.Control
                type="search"
                className="form-control form-control-sm"
                placeholder=""
                aria-controls="role-users"
                onChange={handleSearchChange}
              />
            </Form.Group>
          </div>
          <div>
            <VizButton
              color="secondary"
              size="small"
              onClick={toggleUsersOrGroupsModal}
            >
              Edit Group(s)
            </VizButton>
          </div>
        </Row>
      </div>

      <Row className="py-2 px-3">
        <ClouderaTable
          isLoading={!groups}
          table={table}
          data-test-id="roleGroupsTable"
        />
      </Row>

      {openUsersOrGroupsModal && (
        <UsersOrGroupsModal
          close={closeUsersOrGroupsModal}
          isOpen={openUsersOrGroupsModal}
          type="groups"
          allEntities={[
            ...(groups ?? []),
            ...currentGroups.filter(({ id }) => typeof id === 'string'),
          ]}
          currentEntities={currentGroups}
          createNewEntity={(name, id) => ({
            id,
            name,
          })}
          getName={(group) => group.name}
          title="Groups"
          onApply={(updatedGroups) => {
            setValue(
              'groups',
              updatedGroups.map((updatedGroup) => updatedGroup.name),
              { shouldDirty: true, shouldValidate: true }
            );
          }}
        />
      )}
    </Col>
  );
};
