import React, { useMemo } from 'react';
import classNames from 'classnames';
import classes from './../RoleDetails.module.scss';
import Clusters from '@cloudera-internal/cuix-core/icons/svg/ic-classic-clusters.svg';
import User from '@cloudera-internal/cuix-core/icons/svg/ic-user.svg';
import Link from '@cloudera-internal/cuix-core/icons/svg/ic-link.svg';
import Dataset from '@cloudera-internal/cuix-core/icons/svg/ic-data-sets.svg';
import PreviousIcon from '@cloudera-internal/cuix-core/icons/svg/ic-page-previous.svg';
import { PERMISSION_GROUPS } from '../constants/permission-constants';
import { UseFormWatch } from 'react-hook-form';
import { RoleDetailsFormData } from '../RoleDetails';
import { CircularProgressBar } from 'pages/visuals-page/components/common/CircularProgessBar';

type PermissionSideNavProps = {
  activeTab: keyof typeof PERMISSION_GROUPS;
  onClick: (tabId: keyof typeof PERMISSION_GROUPS) => void;
  watch: UseFormWatch<RoleDetailsFormData>;
  areDataconnectionsLoading: boolean;
  areDatasetsLoading: boolean;
  areRolesLoading: boolean;
  numberOfPrivilegesByType: {
    system: number;
    datasets: number;
    connections: number;
    roles: number;
  };
};

const ICONS_BY_ID = {
  system: <Clusters />,
  roles: <User />,
  connections: <Link />,
  datasets: <Dataset />,
} as const;

export const PermissionsSideNav = ({
  activeTab,
  onClick,
  watch,
  areDataconnectionsLoading,
  areDatasetsLoading,
  areRolesLoading,
  numberOfPrivilegesByType,
}: PermissionSideNavProps) => {
  const systemPrivileges = watch('system');
  const rolePrivileges = watch('roles');
  const datasetPrivileges = watch('datasets');
  const connectionPrivileges = watch('connections');

  const isActiveById = useMemo(
    () => ({
      system:
        Object.values(systemPrivileges).filter((value) => value).length !== 0,
      roles: rolePrivileges.length !== 0,
      datasets: datasetPrivileges.length !== 0,
      connections: connectionPrivileges.length !== 0,
    }),
    [
      connectionPrivileges.length,
      datasetPrivileges.length,
      rolePrivileges.length,
      systemPrivileges,
    ]
  );

  const isLoadingByKey: Record<
    keyof Omit<typeof PERMISSION_GROUPS, 'system'>,
    boolean
  > = {
    connections: areDataconnectionsLoading,
    datasets: areDatasetsLoading,
    roles: areRolesLoading,
  };

  const keys = Object.keys(
    PERMISSION_GROUPS
  ) as (keyof typeof PERMISSION_GROUPS)[];

  return (
    <div className={classes.permissionsSideNav}>
      {keys.map((key) => {
        const group = PERMISSION_GROUPS[key];

        return (
          <div
            className={classNames(
              { [classes.active]: activeTab === key },
              classes.permissionsTab
            )}
            key={key}
            onClick={(e) => {
              e.stopPropagation();
              onClick(key);
            }}
            data-test-id={`${group.tab.toLowerCase()}_tab`}
          >
            <div className={classNames(classes.permissionsLabel, 'col-6')}>
              {ICONS_BY_ID[key]}
              <span className="ml-1">{group.tab}</span>
            </div>
            <div className={classNames(classes.badgeWrapper, 'col-3 p-0')}>
              {isLoadingByKey[key] ? (
                <div className={classes.circularProgressBarWrapper}>
                  <CircularProgressBar />
                </div>
              ) : (
                <>
                  {isActiveById[key] && (
                    <span className={classes.badgePill}>
                      {numberOfPrivilegesByType[key]}
                    </span>
                  )}
                </>
              )}
            </div>
            <div className={classNames(classes.previousIcon, 'col-3')}>
              <PreviousIcon />
            </div>
          </div>
        );
      })}
    </div>
  );
};
