import React, { useState } from 'react';
import { Form, Table } from 'react-bootstrap';
import classes from './../RoleDetails.module.scss';
import {
  Control,
  UseFormRegister,
  UseFormSetValue,
  useFieldArray,
} from 'react-hook-form';
import { RoleDetailsFormData } from '../RoleDetails';
import { PERMISSION_GROUPS } from '../constants/permission-constants';
import { VizButton } from 'components/viz-button/VizButton';
import CirclePlus from '@cloudera-internal/cuix-core/icons/svg/ic-plus-circle.svg';
import { useToggle } from 'hooks/use-toggle';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import { AddRolePrivilegeModal } from './AddRolePrivilegeModal';
import { Notification } from 'components/notification/Notification';
import { CircularProgressBar } from 'pages/visuals-page/components/common/CircularProgessBar';
import classNames from 'classnames';

type RolesTableProps = {
  register: UseFormRegister<RoleDetailsFormData>;
  setValue: UseFormSetValue<RoleDetailsFormData>;
  control: Control<RoleDetailsFormData>;
  areRolesLoading: boolean;
  roleNamesById: Map<number, string>;
  roleOptions: { id: number; name: string }[];
};

const rolesDependencyMessage = `
            The <strong>Grant Manage dashboards</strong> and <strong>Grant View dashboards</strong> are mandatory if 
            the privilege contains the <strong>Grant Manage dataset</strong> permission 
            and cannot be removed. The checkboxes are filled in and fixed.<br><br>

            If <strong>Grant Manage dataset</strong> is deselected, 
            <strong>Grant Manage dashboards</strong> becomes optional and can also be deselected.<br><br>
            
            If <strong>Grant Manage dashboards</strong> is deselected, <strong>Grant View dashboards</strong> becomes optional and can also be deselected.`;

export const RolesTable = ({
  register,
  control,
  areRolesLoading,
  roleNamesById,
  roleOptions,
}: RolesTableProps): React.ReactElement => {
  const [didUserConfirmNotification, setDidUserConfirmNotification] =
    useState(false);
  const { privileges } = PERMISSION_GROUPS.roles;

  const { fields, append, remove, update } = useFieldArray({
    control,
    name: 'roles',
  });

  const [
    addRolePrivilegeModalIsOpen,
    toggleAddRolePrivilegeModal,
    closeAddRolePrivilegeModal,
  ] = useToggle();

  const handleNotificationClose = () => {
    setDidUserConfirmNotification(true);
  };

  return (
    <>
      <div className={classes.addPrivilegesButtonWrapper}>
        <VizButton
          color="secondary"
          size="medium"
          onClick={toggleAddRolePrivilegeModal}
          testId="add-privilege-btn"
        >
          <div className={classes.addPrivilegesButtonInner}>
            <div className={classes.addPrivilegesButtonIcon}>
              <CirclePlus />
            </div>
            <div>Add Roles</div>
          </div>
        </VizButton>
      </div>
      <div className={classes.dependencyContainer}>
        <Notification
          title="Privilege Dependency Information"
          useHtml
          message={rolesDependencyMessage}
          variant="info"
          dismissible
          show={
            !didUserConfirmNotification &&
            fields.some(
              (field) =>
                field.privileges['rgds_manage'] ||
                field.privileges['rgds_appedit']
            )
          }
          onClose={handleNotificationClose}
        />
      </div>

      <Table
        className={classes.privilegeContainer}
        data-test-id="role-privilege-container"
      >
        <thead>
          <tr>
            <th className={classes.privilegeHeader}></th>

            <th className={classes.privilegeHeader}></th>

            {privileges.map(({ id, label, description }) => (
              <th className={classes.privilegeHeader} key={id}>
                <div className="font-medium">{label}</div>
                <div className="text-small font-normal">{description}</div>
              </th>
            ))}

            <th className={classes.privilegeHeader}></th>
          </tr>
        </thead>

        <tbody>
          {fields.map((field, index) => (
            <tr key={field.id}>
              <td className={classes.privilegeBody}>
                {areRolesLoading ? (
                  <CircularProgressBar />
                ) : (
                  field.role_ids
                    ?.map((id) => roleNamesById.get(+id) || 'No Name')
                    .join(' / ') || 'No Name'
                )}
              </td>
              {(
                [
                  {
                    id: 'selectAll',
                    label: '',
                    description: '',
                  },
                  ...privileges,
                ] as const
              ).map((permission) => (
                <td
                  className={classNames(classes.privilegeBody, 'text-center')}
                  key={permission.id}
                >
                  <Form.Check
                    data-test-id={`${permission.id}`}
                    id={permission.id}
                    type="checkbox"
                    className={classes.privilegeCheckbox}
                    {...(permission.id === 'selectAll'
                      ? {
                          onChange: (e) => {
                            update(index, {
                              ...field,
                              privileges: PERMISSION_GROUPS.roles.privileges
                                .map((privilege) => privilege.id)
                                .reduce((obj, id) => {
                                  obj[id] = e.target.checked;
                                  return obj;
                                }, {} as Record<(typeof PERMISSION_GROUPS.roles.privileges)[number]['id'], boolean>),
                            });
                          },
                          checked: Object.values(field.privileges).every(
                            (checked) => checked
                          ),
                        }
                      : register(`roles.${index}.privileges.${permission.id}`))}
                  />
                </td>
              ))}

              <td
                className={classes.privilegeBody}
                key={`${field.id}-${index}`}
              >
                <Delete onClick={() => remove(index)} />
              </td>
            </tr>
          ))}
        </tbody>
      </Table>

      <AddRolePrivilegeModal
        isOpen={addRolePrivilegeModalIsOpen}
        roleOptions={roleOptions}
        close={closeAddRolePrivilegeModal}
        areRolesLoading={areRolesLoading}
        onSave={(roles) => {
          append({
            role_ids: roles.map(({ id }) => id.toString()),
            privileges: PERMISSION_GROUPS.roles.privileges
              .map((privilege) => privilege.id)
              .reduce((obj, id) => {
                obj[id] = true;
                return obj;
              }, {} as Record<(typeof PERMISSION_GROUPS.roles.privileges)[number]['id'], boolean>),
          });
        }}
      />
    </>
  );
};
