import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import '@testing-library/jest-dom';
import { configure, render, screen } from '@testing-library/react';
import * as React from 'react';
import { SentryGroupsTable } from './SentryGroupsTable';

const queryClient = new QueryClient();

const renderWithQueryClient = (ui: React.ReactElement) => {
  return render(
    <QueryClientProvider client={queryClient}>{ui}</QueryClientProvider>
  );
};

configure({ testIdAttribute: 'data-test-id' });

const mockAddNotification = jest.fn();
jest.mock('../../../../src/stores/use-notification-store.ts', () => ({
  useNotificationStore: () => ({ addNotification: mockAddNotification }),
}));

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        urls: {
          viewDashboardUrl: 'customUrl',
        },
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

afterEach(() => {
  jest.resetAllMocks();
});

const mockSentryGroups = [
  { name: 'mockSentryGroup1' },
  { name: 'mockSentryGroup2' },
];

describe('Sentry Groups components', () => {
  it('renders without error', () => {
    renderWithQueryClient(
      <SentryGroupsTable sentryGroups={mockSentryGroups} />
    );
  });

  it('shows the correct title and sentry groups', () => {
    renderWithQueryClient(
      <SentryGroupsTable sentryGroups={mockSentryGroups} />
    );

    expect(screen.getByText(/Sentry Groups/i)).toBeInTheDocument();
    expect(screen.getByText(/mockSentryGroup1/i)).toBeInTheDocument();
    expect(screen.getByText(/mockSentryGroup2/i)).toBeInTheDocument();
  });

  it('shows the correct description text', () => {
    renderWithQueryClient(
      <SentryGroupsTable sentryGroups={mockSentryGroups} />
    );

    expect(
      screen.getByText(
        /Membership for roles imported from Sentry can not be changed here and must be managed through the Sentry system/i
      )
    ).toBeInTheDocument();
  });

  it('handles empty sentry groups', () => {
    renderWithQueryClient(<SentryGroupsTable sentryGroups={[]} />);

    expect(screen.getByText(/Sentry Groups/i)).toBeInTheDocument();
    // Check that no group names are rendered
    expect(screen.queryByText(/mockSentryGroup1/i)).not.toBeInTheDocument();
    expect(screen.queryByText(/mockSentryGroup2/i)).not.toBeInTheDocument();
  });
});
