import React, { useState } from 'react';
import { Form, Table } from 'react-bootstrap';
import classes from './../RoleDetails.module.scss';
import {
  UseFormRegister,
  UseFormSetValue,
  UseFormWatch,
} from 'react-hook-form';
import { RoleDetailsFormData } from '../RoleDetails';
import { PERMISSION_GROUPS } from '../constants/permission-constants';
import { Notification } from 'components/notification/Notification';
import classNames from 'classnames';

type SystemTableCheckboxProps = {
  register: UseFormRegister<RoleDetailsFormData>;
  watch: UseFormWatch<RoleDetailsFormData>;
  setValue: UseFormSetValue<RoleDetailsFormData>;
};

const SOFT_FORCE_MAP = {
  sys_workspace: ['sys_viewroles'],
};

const systemDependencyMessage = `
            The <strong>View roles and users</strong> permission is mandatory if
            the privilege contains the
            <strong>Manage roles and users</strong> permission, and cannot be
            removed. The checkboxes are filled in and fixed.<br><br>
          
            If <strong>Manage roles and users</strong> is deselected, 
            <strong>View roles and users</strong> becomes optional and can also
            be deselected.`;

export const SystemTable = ({
  register,
  watch,
  setValue,
}: SystemTableCheckboxProps): React.ReactElement => {
  const [didUserConfirmNotification, setDidUserConfirmNotification] =
    useState(false);
  const { privileges } = PERMISSION_GROUPS.system;

  const formPrivileges = watch('system');

  const handleNotificationClose = () => {
    setDidUserConfirmNotification(true);
  };

  return (
    <>
      <div className={classes.dependencyContainer}>
        <Notification
          title="Privilege Dependency Information"
          useHtml
          message={systemDependencyMessage}
          variant="info"
          dismissible
          show={
            !didUserConfirmNotification &&
            formPrivileges['sys_viewroles'] === true
          }
          onClose={handleNotificationClose}
        />
      </div>
      <Table
        className={classes.privilegeContainer}
        data-test-id="system-privilege-container"
      >
        <thead>
          <tr>
            <th className={classes.privilegeHeader}></th>

            <th className={classes.privilegeHeader}></th>

            {privileges.map(({ id, label, description }) => (
              <th className={classes.privilegeHeader} key={id}>
                <div className="font-medium">{label}</div>
                <div className="text-small font-normal">{description}</div>
              </th>
            ))}
          </tr>
        </thead>
        <tbody>
          <tr>
            <td className={classes.privilegeBody}></td>

            {(
              [
                {
                  id: 'selectAll',
                  label: '',
                  description: '',
                },
                ...privileges,
              ] as const
            ).map((privilege) => (
              <td
                className={classNames(classes.privilegeBody, 'text-center')}
                key={privilege.id}
              >
                <Form.Check
                  data-test-id={`test-${privilege.id}`}
                  id={privilege.id}
                  type="checkbox"
                  className={classes.privilegeCheckbox}
                  checked={privilege?.[privilege.id]}
                  {...(privilege.id === 'selectAll'
                    ? {
                        onChange: (e) => {
                          setValue(
                            'system',
                            PERMISSION_GROUPS.system.privileges
                              .map((privilege) => privilege.id)
                              .reduce((obj, id) => {
                                obj[id] = e.target.checked;
                                return obj;
                              }, {} as Record<(typeof PERMISSION_GROUPS.system.privileges)[number]['id'], boolean>),
                            { shouldDirty: true, shouldValidate: true }
                          );
                        },
                        checked: Object.values(formPrivileges).every(
                          (value) => value
                        ),
                      }
                    : register(`system.${privilege.id}`, {
                        onChange: (event) => {
                          const newPrivileges = {
                            ...formPrivileges,
                          };

                          const softDependencies = SOFT_FORCE_MAP[privilege.id];

                          if (
                            event.target.checked &&
                            Array.isArray(softDependencies)
                          ) {
                            softDependencies.forEach((id) => {
                              newPrivileges[id] = true;
                            });
                          }

                          setValue('system', newPrivileges, {
                            shouldDirty: true,

                            shouldValidate: true,
                          });
                        },
                      }))}
                />
              </td>
            ))}
          </tr>
        </tbody>
      </Table>
    </>
  );
};
