import React, {
  ChangeEventHandler,
  useCallback,
  useMemo,
  useState,
} from 'react';
import type { Group, User } from '../types/role-details.type';
import {
  Col,
  Container,
  Form,
  FormControl,
  InputGroup,
  Modal,
  Row,
} from 'react-bootstrap';
import classes from './UsersOrGroupsModal.module.scss';
import classNames from 'classnames';
import { VizButton } from '../../../components/viz-button/VizButton';

type UsersOrGroupsModalProps<T extends User | Group> = {
  isOpen: boolean;
  close: () => void;
  type: 'users' | 'groups';
  allEntities: T[];
  currentEntities: T[];
  createNewEntity: (name: string, id: string) => T;
  getName: (entity: T) => string;
  title: string;
  onApply: (updatedEntities: T[]) => void;
};

export const UsersOrGroupsModal = <T extends User | Group>({
  isOpen,
  close,
  allEntities,
  currentEntities,
  type,
  createNewEntity,
  getName,
  title,
  onApply,
}: UsersOrGroupsModalProps<T>): React.ReactElement => {
  const [currentMembers, setCurrentMembers] = useState(currentEntities ?? []);
  const [createdEntities, setCreatedEntities] = useState<T[]>([]);
  const [newEntityName, setNewEntityName] = useState('');

  const [selectedEntities, setSelectedEntities] = useState<T[]>([]);

  const [searchQuery, setSearchQuery] = useState('');

  const [selectAllAvailable, setSelectAllAvailable] = useState(false);
  const [selectAllSelected, setSelectAllSelected] = useState(false);

  const [errorMessage, setErrorMessage] = useState<string | null>();

  const availableEntities = useMemo(() => {
    return (
      createdEntities
        .concat(allEntities ?? [])
        ?.filter(
          (entity) =>
            !currentMembers?.some(
              (currentEntity) => currentEntity.id === entity.id
            )
        ) ?? []
    );
  }, [allEntities, currentMembers, createdEntities]);

  const handleSelectAllAvailable = useCallback(() => {
    setSelectAllAvailable(!selectAllAvailable);
    setSelectedEntities(!selectAllAvailable ? availableEntities : []);
  }, [selectAllAvailable, availableEntities]);

  const handleSelectAllSelected = useCallback(() => {
    setSelectAllSelected(!selectAllSelected);
    setSelectedEntities(!selectAllSelected ? currentMembers : []);
  }, [selectAllSelected, currentMembers]);

  const filteredEntities = useMemo(() => {
    return availableEntities.filter((entity) =>
      getName(entity).toLowerCase().includes(searchQuery.toLowerCase())
    );
  }, [availableEntities, searchQuery, getName]);

  const handleAddNewItem = useCallback(() => {
    if (!newEntityName) return;

    const isDuplicate = availableEntities.some(
      (entity) => getName(entity).toLowerCase() === newEntityName.toLowerCase()
    );

    if (isDuplicate) {
      setErrorMessage('Duplicate name not allowed');
      return;
    }

    const newItem = createNewEntity(newEntityName, newEntityName);

    setCreatedEntities((current) => [...current, newItem]);
    setNewEntityName('');
    setErrorMessage('');
  }, [newEntityName, availableEntities, createNewEntity, getName]);

  const handleCheckboxChange: ChangeEventHandler<HTMLInputElement> = (
    event
  ) => {
    const itemId = Number.isNaN(+event.target.value)
      ? event.target.value
      : +event.target.value;
    const isChecked = event.target.checked;

    const selectedItem =
      allEntities?.find((entity) => entity.id === itemId) ||
      createdEntities?.find((entity) => entity.id === itemId);

    if (!selectedItem) return;

    if (isChecked) {
      setSelectedEntities((current) => [...current, selectedItem]);
    } else {
      setSelectedEntities((current) =>
        current.filter((item) => item.id !== selectedItem.id)
      );
    }
  };

  const handleAddSelected = useCallback(() => {
    setCurrentMembers((current) => [...current, ...selectedEntities]);
    setSelectedEntities([]);
    setSelectAllAvailable(false);
  }, [selectedEntities, setCurrentMembers]);

  const handleRemoveSelected = useCallback(() => {
    setCurrentMembers((current) =>
      current.filter(
        (user) =>
          !selectedEntities.some((selectedItem) => selectedItem.id === user.id)
      )
    );
    setSelectedEntities([]);
    setSelectAllSelected(false);
  }, [selectedEntities, setCurrentMembers]);

  const handleSearchChange: ChangeEventHandler<HTMLInputElement> = (event) => {
    setSearchQuery(event.target.value);
  };

  const handleApply = () => {
    onApply(currentMembers);
    setNewEntityName('');
    setErrorMessage('');
    close();
  };

  return (
    <Modal
      show={isOpen}
      onHide={close}
      className={classes.roleAssigmentModal}
      data-test-id="role-user-form"
    >
      <Modal.Header closeButton>
        <Modal.Title>Role Assigment</Modal.Title>
      </Modal.Header>

      <Modal.Body>
        <Form>
          <Container data-test-id="role-user-selection">
            <Row className="mb-2 mt-3">{title}</Row>

            <Row className="mb-2">
              <div
                className={classNames(
                  classes.seachBox,
                  classes.selectionSearchInput,
                  'mb-3'
                )}
              >
                <InputGroup>
                  <FormControl
                    placeholder="Search"
                    aria-describedby="inputGroup-search"
                    value={searchQuery}
                    onChange={handleSearchChange}
                  />
                  <InputGroup.Append>
                    <InputGroup.Text>
                      <i className="fa fa-search"></i>
                    </InputGroup.Text>
                  </InputGroup.Append>
                </InputGroup>
              </div>
            </Row>

            <Row data-test-id="selection-boxes">
              <Col
                xs={5}
                className={classes.selectionBox}
                data-test-id="available-selection-box"
              >
                <span className={classes.selectionBoxHeader}>
                  Available to select
                </span>

                <div className={classes.checkAllWrapper}>
                  <Form.Group controlId="select_all_available">
                    <Form.Check
                      type="checkbox"
                      id={`select_all_available`}
                      checked={selectAllAvailable}
                      onChange={handleSelectAllAvailable}
                      label="Select All"
                    />
                  </Form.Group>
                </div>

                <div>
                  <ul className={classes.selectionList}>
                    {filteredEntities?.map((item) => (
                      <li className="form-check" key={item.id}>
                        <input
                          type="checkbox"
                          name={type}
                          className="form-check-input"
                          value={item.id}
                          id={`input-label-id-${item.id}`}
                          onChange={handleCheckboxChange}
                          checked={selectedEntities.some(
                            (selectedItem) => selectedItem.id === item.id
                          )}
                        />
                        <label
                          className="form-check-label"
                          htmlFor={`input-label-id-${item.id}`}
                        >
                          {getName(item)}
                        </label>
                      </li>
                    ))}
                  </ul>
                </div>
              </Col>

              <Col xs={2}>
                <div className={classes.selectionBtnWrapper}>
                  <VizButton
                    color="secondary"
                    className={classes.addSelection}
                    size="small"
                    onClick={handleAddSelected}
                    disabled={filteredEntities.length === 0}
                    testId="add-selection"
                  >
                    Add{' '}
                    <i
                      className="fa fa-angle-double-right"
                      aria-hidden="true"
                    ></i>
                  </VizButton>

                  <VizButton
                    color="secondary"
                    className={classes.removeSelection}
                    size="small"
                    onClick={handleRemoveSelected}
                    disabled={currentMembers.length === 0}
                    testId="remove-selection"
                  >
                    <i className="fa fa-angle-double-left" aria-hidden="true" />
                  </VizButton>
                </div>
              </Col>

              <Col
                xs={5}
                className={classes.selectionBox}
                data-test-id="selected-selection-box"
              >
                <div className={classes.selectionBoxHeader}>Selected</div>

                <div className={classes.checkAllWrapper}>
                  <Form.Group controlId="select_all_current">
                    <Form.Check
                      type="checkbox"
                      id={`select_all_current`}
                      checked={selectAllSelected}
                      onChange={handleSelectAllSelected}
                      label="Select All"
                    />
                  </Form.Group>
                </div>

                <div>
                  <ul className={classes.selectionList}>
                    {currentMembers?.map((item) => (
                      <li className="form-check" key={item.id}>
                        <input
                          type="checkbox"
                          name={type}
                          className="form-check-input"
                          value={item.id}
                          id={`input-label-id-${item.id}`}
                          onChange={handleCheckboxChange}
                          checked={selectedEntities.some(
                            (selectedItem) => selectedItem.id === item.id
                          )}
                        />
                        <label
                          className="form-check-label"
                          htmlFor={`input-label-id-${item.id}`}
                        >
                          {getName(item)}
                        </label>
                      </li>
                    ))}
                  </ul>
                </div>
              </Col>
            </Row>

            <Row>
              <div className="mt-3">
                <InputGroup className={classes.newEntityInput}>
                  <FormControl
                    placeholder={`Add new ${
                      type === 'users' ? 'user' : 'group'
                    }`}
                    aria-describedby="inputGroup-new"
                    value={newEntityName}
                    onChange={(e) => setNewEntityName(e.target.value)}
                    isInvalid={!!errorMessage}
                    style={{ marginRight: '1px' }}
                  />
                  <InputGroup.Append>
                    <VizButton
                      color="secondary"
                      size="small"
                      onClick={handleAddNewItem}
                    >
                      <i className="fa fa-plus" aria-hidden="true"></i> Add
                    </VizButton>
                  </InputGroup.Append>
                  <Form.Control.Feedback type="invalid">
                    {errorMessage}
                  </Form.Control.Feedback>
                </InputGroup>
              </div>
            </Row>
          </Container>
        </Form>
      </Modal.Body>

      <Modal.Footer>
        <VizButton color="secondary" onClick={close} size="medium">
          Cancel
        </VizButton>

        <VizButton
          color="primary"
          size="medium"
          onClick={handleApply}
          testId="apply-btn"
        >
          Apply
        </VizButton>
      </Modal.Footer>
    </Modal>
  );
};
