import React, { useMemo } from 'react';
import { Col, Form, Row } from 'react-bootstrap';
import { VizButton } from 'components/viz-button/VizButton';
import UserIcon from '@cloudera-internal/cuix-core/icons/svg/ic-user.svg';
import classes from '../RoleDetails.module.scss';
import classNames from 'classnames';
import { useToggle } from 'hooks/use-toggle';
import { UsersOrGroupsModal } from './UsersOrGroupsModal';
import { UseFormSetValue, UseFormWatch } from 'react-hook-form';
import { RoleDetailsFormData } from '../RoleDetails';
import type { User } from '../types/role-details.type';
import {
  createColumnHelper,
  getCoreRowModel,
  getPaginationRowModel,
  getSortedRowModel,
  getFilteredRowModel,
} from '@tanstack/react-table';
import { ClouderaTable } from 'components/cloudera-table';
import { useTable } from 'hooks/use-table';

const columnHelper = createColumnHelper<User>();

type UsersTableProps = {
  users: User[];
  setValue: UseFormSetValue<RoleDetailsFormData>;
  watch: UseFormWatch<RoleDetailsFormData>;
  columnWidth: number;
};

const UsersTable = ({
  users,
  setValue,
  watch,
  columnWidth,
}: UsersTableProps): React.ReactElement => {
  const userNames = watch('users');

  const columns = useMemo(
    () => [
      columnHelper.accessor('username', {
        header: () => 'Username',
        cell: (info) => info.getValue(),
      }),
      columnHelper.accessor('first_name', {
        header: () => 'First Name',
        cell: (info) => info.getValue(),
      }),
      columnHelper.accessor('last_name', {
        header: () => 'Last Name',
        cell: (info) => info.getValue(),
      }),
    ],
    []
  );

  const currentUsers = useMemo<User[]>(() => {
    return userNames.map(
      (name) =>
        users?.find((user) => user.username === name) ?? {
          id: name,
          username: name,
          first_name: '',
          last_name: '',
          full_name: '',
          is_superuser: false,
        }
    );
  }, [users, userNames]);

  const table = useTable({
    data: currentUsers,
    columns,
    getCoreRowModel: getCoreRowModel(),
    getSortedRowModel: getSortedRowModel(),
    getPaginationRowModel: getPaginationRowModel(),
    getFilteredRowModel: getFilteredRowModel(),
    initialState: {
      pagination: {
        pageIndex: 0,
        pageSize: 10,
      },
    },
  });

  const handleSearchChange = (event: React.ChangeEvent<HTMLInputElement>) => {
    table.setGlobalFilter(event.target.value);
  };

  const [
    openUsersOrGroupsModal,
    toggleUsersOrGroupsModal,
    closeUsersOrGroupsModal,
  ] = useToggle();

  return (
    <Col sm md={columnWidth} className={classes.usersColumn}>
      <div style={{ minHeight: '145px' }}>
        <Row className="d-flex justify-content-between align-items-center pt-2 pb-0 pl-3 pr-3">
          <div className="d-flex align-items-center">
            <UserIcon className={classes.userIcon} />
            <h4>Users</h4>
          </div>
        </Row>
        <Row>
          <div
            className={classNames(
              classes.descriptionText,
              'pt-2 pb-2 pl-3 pr-3'
            )}
          >
            These users were explicitly added and are not necessarily part of
            the added groups
          </div>
        </Row>
        <Row className="d-flex justify-content-between pt-2 pb-2 pl-3 pr-3">
          <div className="d-flex pr-0">
            <Form.Group
              as={Row}
              className="align-items-center mb-0 flex-nowrap mr-0 ml-0"
            >
              <Form.Label className="pr-1 mb-0">Search:</Form.Label>
              <Form.Control
                type="search"
                className="form-control form-control-sm"
                placeholder=""
                aria-controls="role-users"
                onChange={handleSearchChange}
              />
            </Form.Group>
          </div>
          <div>
            <VizButton
              color="secondary"
              size="small"
              onClick={toggleUsersOrGroupsModal}
              testId="role-edit-users"
            >
              Edit User(s)
            </VizButton>
          </div>
        </Row>
      </div>
      <Row
        id="role-users-table"
        className={classNames('pt-2 pb-2 pl-3 pr-3', classes.roleUsersTable)}
      >
        <ClouderaTable
          isLoading={!users}
          table={table}
          data-test-id="roleUsersTable"
        />
      </Row>

      {openUsersOrGroupsModal && (
        <UsersOrGroupsModal
          close={closeUsersOrGroupsModal}
          isOpen={openUsersOrGroupsModal}
          type="users"
          allEntities={[
            ...(users ?? []),
            ...currentUsers.filter(({ id }) => typeof id === 'string'),
          ]}
          currentEntities={currentUsers}
          createNewEntity={(username, id) => ({
            id,
            is_superuser: false,
            username,
            first_name: '',
            last_name: '',
            full_name: '',
          })}
          getName={({ username }) => username}
          title="Users"
          onApply={(updatedUsers: User[]) => {
            setValue(
              'users',
              updatedUsers.map((updatedUser) => updatedUser.username),
              {
                shouldDirty: true,
                shouldValidate: true,
              }
            );
          }}
        />
      )}
    </Col>
  );
};

export default UsersTable;
