import React, {
  FormEvent,
  useCallback,
  useEffect,
  useReducer,
  useState,
} from 'react';
import { useMutation, useQuery } from '@tanstack/react-query';
import { NavigationBar } from 'components/navigation-bar/NavigationBar';
import { useNotificationStore } from 'stores/use-notification-store';
import { getSiteSettings, siteSettingsQueryToken } from './api/get-settings';
import classNames from 'classnames';
import { SideNav } from './components/SideNav';
import { Form, Modal } from 'react-bootstrap';
import classes from './SiteSettingsPage.module.scss';
import { SiteSettingsHeader } from './components/SiteSettingsHeader';
import { confirmAlert } from 'react-confirm-alert';
import { VizButton } from 'components/viz-button/VizButton';
import {
  submitSiteSettings,
  submitSiteSettingsQueryToken,
} from './api/submit-site-settings';
import {
  DEFAULT_SETTING_GROUPS,
  DEFAULT_SITE_SETTING_VALUES,
  FLIPPER_SETTINGS_MAP,
  SITE_SETTING_LABELS,
} from './constants/setting-constants';
import { useDebounce } from '../../hooks/use-debounce';
import { Settings } from './components/Settings';
import {
  DirtySetting,
  State,
  SettingValue,
  SiteSetting,
  SettingTab,
  SearchedSetting,
  SubmitData,
  SubmitFlipper,
  ConfirmObj,
} from './types/site-settings.type';
import { SiteSettingSearchResults } from './components/SiteSettingSearchResults';
import { findSettings } from './utils/setting-utils';
import { AiSettings } from './components/AiSettings';
import { AiTab } from './types/ai-settings.type';
import useBeforeUnload from 'hooks/use-before-unload';

type SiteSettingsPageProps = {
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
};

const getDefaultTab = () => {
  return decodeURIComponent(window.location.hash.substring(1)) || 'Data';
};

const INITIAL_STATE: State = {
  form: DEFAULT_SITE_SETTING_VALUES,
  defaultAfterApi: null,
  dirtyFields: null,
};
const INITIAL_SETTINGS_DATA = {
  tab: getDefaultTab(),
  settings: findSettings(getDefaultTab()),
};

function formReducer(state: State, action): State {
  switch (action.type) {
    case 'UPDATE_FIELD':
      const { field, value } = action.payload;
      const isChanged = state.form[field] !== value;

      return {
        ...state,
        form: { ...state.form, [field]: value },
        dirtyFields: isChanged
          ? { ...state.dirtyFields, [field]: value }
          : state.dirtyFields,
      };

    case 'UPDATE_MULTIPLE_FIELDS':
      const fieldsToUpdate: State['form'] = action.payload;
      let newForm = { ...state.form };
      let newDirtyFields = { ...state.dirtyFields };

      for (const [field, value] of Object.entries(fieldsToUpdate)) {
        newForm[field] = value;
        if (state.form[field] !== value) {
          newDirtyFields[field] = value;
        }
      }

      return {
        ...state,
        form: newForm,
        dirtyFields: newDirtyFields,
      };

    case 'SET_INITIAL_STATE_AFTER_API':
      return {
        ...state,
        form: { ...action.payload },
        defaultAfterApi: { ...action.payload },
      };

    case 'RESET_AFTER_SUBMIT':
      return {
        ...state,
        defaultAfterApi: { ...state.form },
        dirtyFields: null,
      };

    case 'RESET':
      return {
        ...state,
        form: state.defaultAfterApi as Exclude<
          typeof state.defaultAfterApi,
          null
        >,
        dirtyFields: null,
      };

    default:
      return state;
  }
}

const SiteSettingsPage = ({
  shouldShowModal,
  isAlreadySeen,
}: SiteSettingsPageProps) => {
  const { data: siteSettings } = useQuery({
    queryKey: [siteSettingsQueryToken],
    queryFn: getSiteSettings,
  });
  const { mutateAsync: mutateSubmitSettings, isLoading } = useMutation(
    [submitSiteSettingsQueryToken],
    submitSiteSettings,
    {
      onMutate: () => {
        addNotification({
          variant: 'info',
          message: 'Save in progress...',
        });
      },
    }
  );
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));
  const [activeTab, setActiveTab] = useState(INITIAL_SETTINGS_DATA.tab);
  const [isFormDirty, setIsFormDirty] = useState(false);
  const [searchTerm, setSearchTerm] = useState('');
  const [searchedSettings, setSearchedSettings] = useState<SearchedSetting>({});
  const [highlightedId, setHighlightedId] = useState<string>('');
  const [activeSettings, setActiveSettings] = useState<SiteSetting[]>(
    INITIAL_SETTINGS_DATA.settings as SiteSetting[]
  );
  const debouncedSearch = useDebounce(findSearchedSettings, 500);
  const [state, dispatch] = useReducer(formReducer, INITIAL_STATE);

  useBeforeUnload(isFormDirty);

  useEffect(() => {
    if (siteSettings && !state.defaultAfterApi) {
      const settingsAfterApi = {
        ...state.form,
        ...siteSettings,
        ...FLIPPER_SETTINGS_MAP,
      };

      dispatch({
        type: 'SET_INITIAL_STATE_AFTER_API',
        payload: settingsAfterApi,
      });
    }
  }, [siteSettings, state.defaultAfterApi, state.form]);

  const getDirtySettings = useCallback(() => {
    if (!state.defaultAfterApi) return [];

    const dirtySettings: DirtySetting[] = [];

    for (const key in state.dirtyFields) {
      if (
        state.dirtyFields[key] !== (state.defaultAfterApi as State['form'])[key]
      ) {
        dirtySettings.push({
          id: key,
          label: SITE_SETTING_LABELS[key as keyof typeof SITE_SETTING_LABELS],
          value: state.dirtyFields[key],
        });
      }
    }

    return dirtySettings;
  }, [state.defaultAfterApi, state.dirtyFields]);

  useEffect(() => {
    const isDirty = !!getDirtySettings().length;
    setIsFormDirty(isDirty);
  }, [state.dirtyFields, getDirtySettings]);

  function findSearchedSettings(term: string) {
    if (!term) return;

    const found: SearchedSetting = {};

    const find = (items) => {
      const regex = new RegExp(term, 'i');

      for (const group of items) {
        if (group.children) {
          find(group.children);
        } else {
          for (const setting of group.settings) {
            if (regex.test(setting.label)) {
              if (!found[group.tab]) {
                found[group.tab] = [];
              }

              found[group.tab].push(setting);
            }
          }
        }
      }
    };

    find(DEFAULT_SETTING_GROUPS);

    setSearchedSettings(found);
  }

  const handleSearch = (term: string) => {
    setSearchTerm(term);
    setActiveTab('');
    window.location.hash = '';

    debouncedSearch(term);
  };

  const renderDirtyLabels = () => {
    return (
      <ul>
        {getDirtySettings().map(({ id, label }) =>
          label ? <li key={id}>{label}</li> : null
        )}
      </ul>
    );
  };

  const onReset = () => {
    dispatch({ type: 'RESET' });
  };

  const onSubmit = (e: FormEvent) => {
    e.preventDefault();

    if (!isFormDirty) return;

    confirmAlert({
      customUI: ({ onClose }) => {
        return (
          <Modal
            show={true}
            onHide={onClose}
            data-test-id="siteSettingsSaveModal"
          >
            <Modal.Header closeButton>
              <Modal.Title>Confirm Changes</Modal.Title>
            </Modal.Header>
            <Modal.Body>
              <p>
                Any site settings changes you make will apply to the entire
                system and to all users.
              </p>
              <p>Changed settings:</p>
              {renderDirtyLabels()}
              <p>Are you sure you want to apply these changes?</p>
            </Modal.Body>
            <Modal.Footer>
              <VizButton color="secondary" size="medium" onClick={onClose}>
                Cancel
              </VizButton>
              <VizButton
                color="primary"
                size="medium"
                testId="confirmSiteSettingsSave"
                onClick={() => {
                  submitChanges(getDirtySettings());
                  onClose();
                }}
              >
                Save
              </VizButton>
            </Modal.Footer>
          </Modal>
        );
      },
    });
  };

  const showConfirmModal = (confirmObj: ConfirmObj) => {
    confirmAlert({
      customUI: ({ onClose }) => {
        return (
          <Modal show={true} onHide={onClose} data-test-id="confirmSetting">
            <Modal.Header closeButton>
              <Modal.Title>{confirmObj.title}</Modal.Title>
            </Modal.Header>
            <Modal.Body>{confirmObj.text}</Modal.Body>
            <Modal.Footer>
              <VizButton color="secondary" size="medium" onClick={onClose}>
                Cancel
              </VizButton>
              <VizButton
                color="primary"
                size="medium"
                testId="confirmSetting"
                onClick={() => {
                  dispatch({
                    type: confirmObj.dispatch.type,
                    payload: confirmObj.dispatch.payload,
                  });
                  onClose();
                }}
              >
                Accept
              </VizButton>
            </Modal.Footer>
          </Modal>
        );
      },
    });
  };

  const handleChanges = (
    id: string,
    value: SettingValue,
    confirmObj?: ConfirmObj
  ) => {
    if (confirmObj) {
      showConfirmModal(confirmObj);
    } else {
      dispatch({
        type: 'UPDATE_FIELD',
        payload: { field: id, value },
      });
    }
  };

  const handleTabChange = (
    tab: SettingTab['tab'],
    settings: SettingTab['settings'] | null
  ) => {
    let found = settings;

    if (!found) {
      found = findSettings(tab);
    }

    setActiveTab(tab);
    setActiveSettings(found as SettingTab['settings']);

    window.location.hash = tab;

    if (searchTerm) {
      setSearchTerm('');
      setSearchedSettings({});
    }
  };

  const changeToViewMode = (id: string, tab: string) => {
    setHighlightedId(id);
    handleTabChange(tab, null);

    setTimeout(() => {
      setHighlightedId('');
    }, 5000);
  };

  const submitChanges = async (settings: DirtySetting[]) => {
    const data: SubmitData = [];
    const flippers: SubmitFlipper = [];

    for (const { id, value } of settings) {
      const val = value.toString();

      if (FLIPPER_SETTINGS_MAP.hasOwnProperty(id)) {
        flippers.push({
          id,
          value: val,
        });
      } else {
        data.push({
          setting_key: id,
          setting_value: val,
          setting_read_admin_only: 0,
        });
      }
    }

    try {
      await mutateSubmitSettings({ settings: data, flippers });

      dispatch({ type: 'RESET_AFTER_SUBMIT' });

      addNotification(
        {
          id: 'site-settings-success',
          variant: 'success',
          message: 'Settings have been saved',
        },
        { delay: 1500 }
      );
    } catch (error) {
      addNotification(
        {
          id: 'site-settings-danger',
          variant: 'danger',
          message: `Failed to change settings. ${error}`,
        },
        { delay: 3000 }
      );
    }
  };

  const isAiTab = ['Embeddings', 'Completion'].includes(activeTab);

  const renderComponent = () => {
    if (!state.defaultAfterApi) return null;

    if (isAiTab) {
      return (
        <AiSettings
          tab={activeTab.toLowerCase() as AiTab}
          formState={state.form}
          onChange={(id, value, confirmObj) =>
            handleChanges(id, value, confirmObj)
          }
        />
      );
    }

    return searchTerm ? (
      <SiteSettingSearchResults
        searchedSettings={searchedSettings}
        formState={state.form}
        onChange={(id, value, confirmObj) =>
          handleChanges(id, value, confirmObj)
        }
        changeToViewMode={changeToViewMode}
      />
    ) : (
      <Settings
        settings={activeSettings}
        highlightedId={highlightedId}
        formState={state.form}
        onChange={(id, value, confirmObj) =>
          handleChanges(id, value, confirmObj)
        }
      />
    );
  };

  return (
    <section className="d-flex flex-column vh-100">
      <NavigationBar
        isAlreadySeen={isAlreadySeen}
        shouldShowModal={shouldShowModal}
        activeTabs={{
          main: 'settings',
          sub: 'navbar-settings-menu-site-setting',
        }}
      />
      <div className={classNames('d-flex', classes.bodyHeight)}>
        <SideNav
          items={DEFAULT_SETTING_GROUPS}
          activeTab={activeTab}
          onClick={(tab, settings) => handleTabChange(tab, settings)}
        />
        <Form className={classes.form} onSubmit={onSubmit} onReset={onReset}>
          <SiteSettingsHeader
            isUndoVisible={isFormDirty}
            isLoading={isLoading}
            doSearch={handleSearch}
            searchTerm={searchTerm}
            clear={() =>
              handleTabChange(
                INITIAL_SETTINGS_DATA.tab,
                INITIAL_SETTINGS_DATA.settings as SiteSetting[]
              )
            }
          />
          <div
            className={classNames('py-3 d-inline-flex flex-column', {
              [classes.mw400]: !searchTerm,
              [classes.fw]: searchTerm || isAiTab,
            })}
          >
            {renderComponent()}
          </div>
        </Form>
      </div>
    </section>
  );
};

export { SiteSettingsPage };
