import React from 'react';
import { render, screen, fireEvent, waitFor } from '@testing-library/react';
import '@testing-library/jest-dom';
import AiModal from './AiModal';
import { createAiSettings } from '../constants/setting-constants';
import { getDefaultSettingValues } from '../utils/setting-utils';
import { AiTab } from '../types/ai-settings.type';
import { SiteSetting } from '../types/site-settings.type';
import ShortUniqueId from 'short-unique-id';

jest.mock('../constants/setting-constants', () => ({
  createAiSettings: jest.fn(),
}));

jest.mock('../utils/setting-utils', () => ({
  getDefaultSettingValues: jest.fn(),
}));

jest.mock('short-unique-id', () => {
  return jest.fn().mockImplementation(() => {
    return {
      randomUUID: () => 'mocked-uuid',
    };
  });
});

window.arc.frontend = {
  globals: {
    settings: { isCdvModeCml: false },
  },
} as any;

const defaultProps = {
  show: true,
  handleClose: jest.fn(),
  handleSave: jest.fn(),
  profileLabels: [],
  tab: 'embeddings' as AiTab,
  profileID: null,
  sourceID: null,
  uuid: new ShortUniqueId(),
  settingObj: {
    activeSettings: [],
    propSettingState: {},
    aiEnabled: true,
  },
};

describe('AiModal', () => {
  beforeEach(() => {
    (createAiSettings as jest.Mock).mockReturnValue({
      open_ai: { embeddings: [], completion: [] },
      caii: { embeddings: [], completion: [] },
      azure_ai: { embeddings: [], completion: [] },
      llama: { embeddings: [], completion: [] },
      other: { embeddings: [], completion: [] },
    });
    (getDefaultSettingValues as jest.Mock).mockImplementation((settings) => {
      const state = {};
      settings.forEach(({ id }) => {
        state[id] = `default-${id}`;
      });
      return state;
    });
  });

  afterEach(() => {
    jest.clearAllMocks();
  });

  test('renders correctly', () => {
    render(<AiModal {...defaultProps} />);

    expect(screen.getByText('Save')).toBeInTheDocument();
    expect(screen.getByText('Name')).toBeInTheDocument();
    expect(screen.getByText('Create Profile')).toBeInTheDocument();

    const closeButton = screen
      .getAllByRole('button', { name: 'Close' })
      .find((button) => button.classList.contains('btn-secondary'));
    expect(closeButton).toBeInTheDocument();
  });

  test('profile name input updates state', () => {
    render(<AiModal {...defaultProps} />);
    const input = screen.getByLabelText('Name');
    fireEvent.change(input, { target: { value: 'New Profile' } });
    expect(input).toHaveValue('New Profile');
  });

  test('input is NOT disabled for default profile', () => {
    render(
      <AiModal
        {...defaultProps}
        profileID={{ value: 'default', label: 'Default' }}
      />
    );
    const input = screen.getByLabelText('Name');
    expect(input).not.toBeDisabled();
  });

  test('shows error message if profile name exists', async () => {
    render(<AiModal {...defaultProps} profileLabels={['existing profile']} />);
    const input = screen.getByLabelText('Name');
    fireEvent.change(input, { target: { value: 'Existing Profile' } });
    fireEvent.click(screen.getByText('Save'));
    await waitFor(() => {
      expect(
        screen.getByText('A profile with this name already exists.')
      ).toBeInTheDocument();
    });
  });

  test('handleSave is called with correct profile data', () => {
    const handleSave = jest.fn();
    render(<AiModal {...defaultProps} handleSave={handleSave} />);
    const input = screen.getByLabelText('Name');
    fireEvent.change(input, { target: { value: 'New Profile' } });
    fireEvent.click(screen.getByText('Save'));
    expect(handleSave).toHaveBeenCalledWith({
      prof: { value: 'mocked-uuid', label: 'New Profile' },
      settingValues: {},
    });
  });

  test('close button resets state and calls handleClose', () => {
    render(<AiModal {...defaultProps} />);

    const closeButton = screen
      .getAllByRole('button', { name: 'Close' })
      .find((button) => button.classList.contains('btn-secondary'));

    if (closeButton) {
      fireEvent.click(closeButton);
    }

    expect(defaultProps.handleClose).toHaveBeenCalled();
    expect(screen.getByLabelText('Name')).toHaveValue('');
  });

  test('changing settings updates settingState', () => {
    const mockSettings: SiteSetting = {
      id: 'test_setting',
      value: 'initial-value',
      label: 'Initial Value',
      type: 'input',
    };

    const mockFormState = {
      test_setting: 'initial-value',
    };

    const { rerender } = render(
      <AiModal
        {...defaultProps}
        settingObj={{
          ...defaultProps.settingObj,
          activeSettings: [mockSettings],
          propSettingState: mockFormState,
        }}
      />
    );

    let input = screen.getByDisplayValue('initial-value');
    expect(input).toBeInTheDocument();

    const newFormState = {
      test_setting: 'new-value',
    };

    rerender(
      <AiModal
        {...defaultProps}
        settingObj={{
          ...defaultProps.settingObj,
          activeSettings: [mockSettings],
          propSettingState: newFormState,
        }}
      />
    );

    input = screen.getByDisplayValue('new-value');
    expect(input).toBeInTheDocument();
  });
});
