import React, { useEffect, useMemo, useState } from 'react';
import { Modal, Button, Form, Alert } from 'react-bootstrap';
import { getGlobalArc } from 'utils/get-global-arc';
import { Settings } from './Settings';
import { getDefaultSettingValues } from '../utils/setting-utils';
import {
  AiEngine,
  AiTab,
  AlertMessage,
  ProfileModal,
  ProfileOption,
  AiCmlAuthMode,
} from '../types/ai-settings.type';
import { createAiSettings } from '../constants/setting-constants';
import { SettingValue, SiteSetting, State } from '../types/site-settings.type';
import ShortUniqueId from 'short-unique-id';
import { isAxiosError } from 'axios';
import { validateAiProfile } from '../api/validate-ai-profile';

const arc = getGlobalArc();

type AiModalProps = {
  show: boolean;
  handleClose: () => void;
  handleSave: (newProfile: ProfileModal) => void;
  profileLabels: string[];
  tab: AiTab;
  profileID: ProfileOption | null;
  sourceID: string | null;
  uuid: ShortUniqueId;
  settingObj: {
    activeSettings: SiteSetting[];
    propSettingState: State['form'];
    aiEnabled: boolean;
  };
};

const DEF_MSG: AlertMessage = {
  text: '',
  type: 'danger',
};

const AiModal = ({
  show,
  handleClose,
  handleSave,
  profileLabels,
  tab,
  profileID,
  sourceID,
  uuid,
  settingObj,
}: AiModalProps) => {
  const [profileName, setProfileName] = useState('');
  const [profileNameExists, setProfileNameExists] = useState(false);
  const [isLoading, setIsLoading] = useState(false);
  const [alertMessage, setAlertMessage] = useState<AlertMessage>(DEF_MSG);
  const [settingState, setSettingState] = useState<State['form']>({});
  const [activeSet, setActiveSet] = useState<SiteSetting[]>([]);
  const settings = useMemo(() => createAiSettings(tab), [tab]);
  const [shouldValidate, setShouldValidate] = useState(false);

  const { activeSettings, propSettingState, aiEnabled } = settingObj;

  useEffect(() => {
    setActiveSet(activeSettings);
    setSettingState(propSettingState);

    setProfileName(profileID ? profileID.label : '');
  }, [activeSettings, profileID, propSettingState]);

  const resetAlertMessage = () => {
    setAlertMessage(DEF_MSG);
    setProfileNameExists(false);
  };

  const handleChange = (id: string, value: SettingValue) => {
    let stateCpy = { ...settingState };
    stateCpy[id] = value;

    if (id === `${tab}_engine`) {
      setShouldValidate(false);
      const engineType = value as AiEngine;
      const engineSpecificSettings = settings[engineType][tab];
      const defState = getDefaultSettingValues(engineSpecificSettings);
      stateCpy = { ...defState, [id]: value };

      setActiveSet(engineSpecificSettings);
    }

    setSettingState(stateCpy);
    resetAlertMessage();
  };

  const isFormValid = (label: string): boolean => {
    const checkRequiredFields = (): boolean => {
      if (!label) {
        return false;
      }

      const engine = settingState[`${tab}_engine`] as AiEngine;

      const apiKey = settingState[`${tab}_api_key`];
      const isCdvModeCml = arc.frontend.globals.settings.isCdvModeCml;
      const isCmlAuthModeApiKey =
        (settingState[`${tab}_cml_auth_mode`] as AiCmlAuthMode) === 'api_key';

      const requiresApiKey =
        (isCdvModeCml && engine === 'caii' && isCmlAuthModeApiKey) ||
        (!isCdvModeCml && engine === 'caii') ||
        ['open_ai', 'azure_ai'].includes(engine);

      if (requiresApiKey && !apiKey) {
        return false;
      }

      if (engine === 'llama' && !settingState[`${tab}_api_key`]) {
        return false;
      }

      if (
        ['caii', 'azure_ai', 'llama', 'other'].includes(engine) &&
        !settingState[`${tab}_service_url`]
      ) {
        return false;
      }

      if (
        ['open_ai', 'caii'].includes(engine) &&
        !settingState[`${tab}_model`]
      ) {
        return false;
      }

      return true;
    };

    const checkExistingProfileName = () => {
      const lowerCaseLabel = label.toLocaleLowerCase();

      const profileNameExits =
        profileLabels.includes(lowerCaseLabel) &&
        (!profileID || (profileID && profileID.label !== label));

      setProfileNameExists(profileNameExits);

      return profileNameExits;
    };

    return checkRequiredFields() && !checkExistingProfileName();
  };

  const onSave = () => {
    setShouldValidate(true);
    const trimmedName = profileName.trim();

    if (!isFormValid(trimmedName)) {
      return;
    }

    const profile: ProfileModal = {
      prof: {
        value: profileID ? profileID.value : uuid.randomUUID(),
        label: trimmedName,
      },
      settingValues: settingState,
    };

    handleSave(profile);
    close();
  };

  const close = () => {
    resetAlertMessage();
    setShouldValidate(false);
    setActiveSet(activeSettings);
    setSettingState(propSettingState);
    handleClose();
  };

  const validateModel = async () => {
    setShouldValidate(true);
    if (!isFormValid(profileName.trim())) {
      return;
    }

    const id = profileID ? profileID.value : sourceID;
    let errorMessage = 'Profile Verification Failed!';

    setIsLoading(true);
    resetAlertMessage();

    try {
      const data = await validateAiProfile(id, tab, settingState);

      if (data.success) {
        setAlertMessage({ text: 'Profile Verified!', type: 'success' });
      } else {
        setAlertMessage({
          text: `${errorMessage} ${data.message}`,
          type: 'danger',
        });
      }
    } catch (error) {
      if (isAxiosError(error)) {
        const err = error.response?.data.error ?? '';
        errorMessage = `${errorMessage} ${err}`;
      } else {
        const err =
          (error instanceof Error ? error.message : String(error)) ?? '';
        errorMessage = `${errorMessage} ${err}`;
      }

      setAlertMessage({ text: errorMessage, type: 'danger' });
    } finally {
      setIsLoading(false);
    }
  };

  return (
    <Modal show={show} onHide={close} backdrop="static">
      <Modal.Header closeButton>
        <Modal.Title>{`${profileID ? 'Edit' : 'Create'} Profile`}</Modal.Title>
      </Modal.Header>
      <Modal.Body>
        <Form>
          <Form.Group className="mb-4">
            <Form.Label className="font-weight-bold" htmlFor="new-profile-name">
              Name
            </Form.Label>
            <Form.Control
              type="text"
              id="new-profile-name"
              value={profileName}
              onChange={(e) => {
                setProfileName(e.target.value);
                resetAlertMessage();
              }}
              isInvalid={shouldValidate && (!profileName || profileNameExists)}
            />
            <Form.Control.Feedback type="invalid">
              {!profileName
                ? 'This field is required.'
                : 'A profile with this name already exists.'}
            </Form.Control.Feedback>
          </Form.Group>
          <Settings
            settings={activeSet}
            formState={{
              ...settingState,
              enable_ai_assistant: aiEnabled,
            }}
            shouldValidate={shouldValidate}
            onChange={handleChange}
          />
          {isLoading && (
            <Alert variant="info">
              <b>Verifying Profile..</b>
            </Alert>
          )}
          {alertMessage.text && (
            <Alert variant={alertMessage.type}>
              <b>{alertMessage.text}</b>
            </Alert>
          )}
        </Form>
      </Modal.Body>
      <Modal.Footer className="justify-content-between">
        <Button
          variant="secondary"
          disabled={isLoading}
          onClick={validateModel}
        >
          Test
        </Button>
        <div className="d-flex gap-2">
          <Button variant="secondary" disabled={isLoading} onClick={close}>
            Close
          </Button>
          <Button variant="primary" disabled={isLoading} onClick={onSave}>
            Save
          </Button>
        </div>
      </Modal.Footer>
    </Modal>
  );
};

export default AiModal;
