import React, { useMemo } from 'react';
import { flexRender, getCoreRowModel } from '@tanstack/react-table';
import {
  AiTab,
  ProfileOption,
  ProfileSettings,
} from '../types/ai-settings.type';
import classes from './AiProfileTable.module.scss';
import classNames from 'classnames';
import {
  convertProfilesToArray,
  generateColumns,
} from '../utils/setting-utils';
import { useTable } from 'hooks/use-table';

type ProfileTableProps = {
  profileSettings: ProfileSettings;
  defaultProfValue?: string;
  tab: AiTab;
  disabled: boolean;
  handleDelete: (profile: ProfileOption) => void;
  handleClone: (profile: ProfileOption) => void;
  handleEdit: (profile: ProfileOption) => void;
  makeItDefault: (profile: ProfileOption) => void;
};

const ProfileTable = ({
  profileSettings,
  defaultProfValue,
  tab,
  disabled,
  handleDelete,
  handleEdit,
  handleClone,
  makeItDefault,
}: ProfileTableProps) => {
  const profilesArr = useMemo(
    () => convertProfilesToArray(profileSettings, tab),
    [profileSettings, tab]
  );

  const columns = useMemo(() => {
    return generateColumns(
      handleDelete,
      handleEdit,
      handleClone,
      makeItDefault,
      disabled,
      tab,
      defaultProfValue
    );
  }, [
    handleDelete,
    handleEdit,
    handleClone,
    makeItDefault,
    disabled,
    tab,
    defaultProfValue,
  ]);
  const table = useTable({
    data: profilesArr,
    columns,
    getCoreRowModel: getCoreRowModel(),
  });

  return (
    <table className={classes.profileTable}>
      <thead>
        {table.getHeaderGroups().map((headerGroup) => (
          <tr key={headerGroup.id}>
            {headerGroup.headers.map((header) => {
              return (
                <th
                  key={header.id}
                  className={classNames(classes.headerCell, {
                    [classes.fixedColumn]:
                      header.column.id === 'label' ||
                      header.column.id === 'engine',
                    [classes.flexColumn]: header.column.id === 'actions',
                  })}
                >
                  {flexRender(
                    header.column.columnDef.header,
                    header.getContext()
                  )}
                </th>
              );
            })}
          </tr>
        ))}
      </thead>
      <tbody>
        {table.getRowModel().rows.length > 0 ? (
          table.getRowModel().rows.map((row) => (
            <tr key={row.id} className={classes.dataRow}>
              {row.getVisibleCells().map((cell) => (
                <td
                  key={cell.id}
                  className={classNames(classes.dataCell, 'tableCell', {
                    [classes.fixedColumn]:
                      cell.column.id === 'label' || cell.column.id === 'engine',
                    [classes.flexColumn]: cell.column.id === 'actions',
                  })}
                >
                  {flexRender(cell.column.columnDef.cell, cell.getContext())}
                </td>
              ))}
            </tr>
          ))
        ) : (
          <tr>
            <td colSpan={columns.length} className={classes.noDataCell}>
              Create Profile to use the AI visual
            </td>
          </tr>
        )}
      </tbody>
    </table>
  );
};

export default ProfileTable;
