import React, { useEffect, useMemo, useState } from 'react';
import { Button, Modal } from 'react-bootstrap';
import { confirmAlert } from 'react-confirm-alert';
import { VizButton } from 'components/viz-button/VizButton';
import {
  AiSetting,
  AiSettingsProps,
  ProfileModal,
  ProfileOption,
  ProfileSetting,
  ProfilesConfig,
} from '../types/ai-settings.type';
import { SiteSetting, State } from '../types/site-settings.type';
import { createAiSettings } from '../constants/setting-constants';
import { getDefaultSettingValues } from '../utils/setting-utils';
import classes from '../SiteSettingsPage.module.scss';
import ProfileTable from './AiProfileTable';
import AiModal from './AiModal';
import ShortUniqueId from 'short-unique-id';

const defaultConfig: ProfilesConfig = {
  defaultProf: null,
  profileSettings: {},
};

const AiSettings = ({
  tab,
  formState,
  onChange,
}: AiSettingsProps): JSX.Element => {
  const [profileConfig, setProfileConfig] =
    useState<ProfilesConfig>(defaultConfig);
  const [selectedProfile, setSelectedProfile] = useState<ProfileOption | null>(
    null
  );
  const [activeSettings, setActiveSettings] = useState<SiteSetting[]>([]);
  const [currSettingState, setCurrSettingState] = useState<State['form']>({});
  const [isOpen, setIsOpen] = useState(false);
  const [sourceID, setSourceID] = useState<string | null>(null);

  const settings = useMemo(() => createAiSettings(tab), [tab]);
  const uuid = useMemo(() => new ShortUniqueId({ length: 10 }), []);

  useEffect(() => {
    const savedSettings = (formState[`ai_${tab}_settings`] as string) ?? '';

    setProfileConfig((prevConfig) => {
      if (savedSettings) {
        try {
          return JSON.parse(savedSettings);
        } catch (error) {
          console.error(error);
          return prevConfig;
        }
      }
      return defaultConfig;
    });
  }, [formState, tab]);

  const deleteProfile = (value: string) => {
    const updatedProfileSettings = { ...profileConfig.profileSettings };
    delete updatedProfileSettings[value];

    // Determine the new default profile
    let newDefaultProf = profileConfig.defaultProf;
    if (newDefaultProf?.value === value) {
      if (!Object.keys(updatedProfileSettings).length) {
        newDefaultProf = null;
      } else {
        const firstKey = Object.keys(updatedProfileSettings)[0];
        const firstLabel = updatedProfileSettings[firstKey].label;
        newDefaultProf = { label: firstLabel, value: firstKey };
      }
    }

    const newConfig = {
      ...profileConfig,
      profileSettings: updatedProfileSettings,
      defaultProf: newDefaultProf,
    };

    onChange(`ai_${tab}_settings`, JSON.stringify(newConfig));
  };

  const confirmDeleteProfile = (profile: ProfileOption) => {
    confirmAlert({
      customUI: ({ onClose }) => {
        return (
          <Modal show={true} onHide={onClose}>
            <Modal.Header closeButton>
              <Modal.Title>Confirm Deletion</Modal.Title>
            </Modal.Header>
            <Modal.Body>
              <p>{`Are you sure you wish to delete "${profile.label}"?`}</p>
            </Modal.Body>
            <Modal.Footer>
              <VizButton color="secondary" size="medium" onClick={onClose}>
                Cancel
              </VizButton>
              <VizButton
                color="primary"
                size="medium"
                onClick={() => {
                  deleteProfile(profile.value);
                  onClose();
                }}
              >
                Yes
              </VizButton>
            </Modal.Footer>
          </Modal>
        );
      },
    });
  };

  const handleEditProfile = (profile: ProfileOption) => {
    setSelectedProfile(profile);

    const currSettings = profileConfig.profileSettings[profile.value];
    const currState = currSettings.values;
    const activeEngine = currState[`${tab}_engine`] as string;

    setActiveSettings(settings[activeEngine][tab]);
    setCurrSettingState(currState);
    setIsOpen(true);
  };

  const handleCloneProfile = (profile: ProfileOption) => {
    const clonedProf = { ...profileConfig.profileSettings[profile.value] };
    setSourceID(profile.value);
    createProfile(clonedProf);
  };

  const makeItDefault = (profile: ProfileOption) => {
    const newConfig = {
      ...profileConfig,
      defaultProf: profile,
    };

    onChange(`ai_${tab}_settings`, JSON.stringify(newConfig));
  };

  const handleModalChange = ({ prof, settingValues }: ProfileModal) => {
    if (settingValues[`${tab}_engine`] === 'caii') {
      // don't save api_key for caii engine profile if cml_auth_mode is jwt
      const cmlAuthModeKey = `${tab}_cml_auth_mode`;
      const isCmlAuthModeJWT =
        cmlAuthModeKey in settingValues
          ? settingValues[cmlAuthModeKey] === 'jwt'
          : false;
      if (isCmlAuthModeJWT) {
        delete settingValues[`${tab}_api_key`];
      }
    }

    const newConfig = {
      ...profileConfig,
      profileSettings: {
        ...profileConfig.profileSettings,
        [prof.value]: {
          label: prof.label,
          values: settingValues,
          ...(sourceID ? { source: sourceID } : {}),
        },
      },
      defaultProf:
        !profileConfig.defaultProf ||
        profileConfig.defaultProf.value === selectedProfile?.value
          ? prof
          : profileConfig.defaultProf,
    };

    onChange(`ai_${tab}_settings`, JSON.stringify(newConfig));
  };

  const createProfile = (clonedProfile?: ProfileSetting) => {
    const engine = clonedProfile
      ? (clonedProfile.values[`${tab}_engine`] as string)
      : 'open_ai';
    const defaultSettings: AiSetting[] = settings[engine][tab];

    setSelectedProfile(null);
    setActiveSettings(defaultSettings);
    setCurrSettingState(
      clonedProfile
        ? clonedProfile.values
        : getDefaultSettingValues(defaultSettings)
    );
    setIsOpen(true);
  };

  const isDisabled = !formState['enable_ai_assistant'];

  return (
    <div className="d-flex flex-column">
      <div className="d-flex mb-4 w-100 justify-content-between">
        {isDisabled && (
          <span className={`${classes.warningMsg} align-self-start my-auto`}>
            Enable AI Assistant to create and change profiles
          </span>
        )}
        <Button
          className="ml-auto"
          variant="primary"
          disabled={isDisabled}
          onClick={() => createProfile()}
        >
          Create Profile
        </Button>
      </div>
      <ProfileTable
        profileSettings={profileConfig.profileSettings}
        defaultProfValue={profileConfig.defaultProf?.value}
        tab={tab}
        disabled={isDisabled}
        handleDelete={confirmDeleteProfile}
        handleEdit={handleEditProfile}
        handleClone={handleCloneProfile}
        makeItDefault={makeItDefault}
      />
      <AiModal
        show={isOpen}
        handleClose={() => {
          setSourceID(null);
          setIsOpen(false);
        }}
        handleSave={handleModalChange}
        profileLabels={Object.values(profileConfig.profileSettings).map(
          (profile) => profile.label.toLocaleLowerCase()
        )}
        profileID={selectedProfile}
        sourceID={sourceID}
        tab={tab}
        uuid={uuid}
        settingObj={{
          activeSettings,
          propSettingState: currSettingState,
          aiEnabled: formState['enable_ai_assistant'] as boolean,
        }}
      />
    </div>
  );
};

export { AiSettings };
