import React, { useState, useEffect, useMemo } from 'react';
import { Form, Popover } from 'react-bootstrap';
import classes from '../SiteSettingsPage.module.scss';
import { SettingValue, SiteSetting, State } from '../types/site-settings.type';
import { VizTooltip } from 'components/VizTooltip';

type SettingRadioProps = {
  setting: SiteSetting;
  formState: State['form'];
  onChange: (id: string, value: SettingValue) => void;
};

// This SettingRadio component supports adding a custom value through an input field. To enable this feature, include an extra option in options with id: "__custom__" and any label. When users select that “custom” option, an input field appears, and their input will be stored as the custom value. If no custom option is provided, the component will work as before with fixed options.

const SettingRadio = ({ setting, formState, onChange }: SettingRadioProps) => {
  const { id, label, popoverText, options, inline = true, disabled } = setting;
  const isDisabled = disabled?.(formState);

  const fixedOptionIds = useMemo(() => {
    return (
      options?.filter((opt) => opt.id !== '__custom__').map((opt) => opt.id) ||
      []
    );
  }, [options]);

  // Figure out if the current formState value is fixed or custom
  const initialRadioValue = fixedOptionIds.includes(formState[id] as string)
    ? (formState[id] as string)
    : '__custom__';

  const [radioValue, setRadioValue] = useState(initialRadioValue);

  // Keep local radio state in sync with parent's formState
  useEffect(() => {
    const newValue = fixedOptionIds.includes(formState[id] as string)
      ? (formState[id] as string)
      : '__custom__';
    setRadioValue(newValue);
  }, [fixedOptionIds, formState, id]);

  // Helper to determine if a radio option should be checked
  const isOptionChecked = (optionId: string): boolean => {
    if (optionId === '__custom__') {
      return !fixedOptionIds.includes(radioValue);
    }
    return optionId === radioValue;
  };

  return (
    <div className="d-flex flex-column">
      <Form.Label className="mb-1 font-weight-bold d-flex align-items-center gap-1">
        {label}
        {popoverText && (
          <VizTooltip
            placement="top"
            overlay={
              <Popover id={id}>
                {<span className="d-flex p-2">{popoverText}</span>}
              </Popover>
            }
          >
            <span className="info-icon" />
          </VizTooltip>
        )}
      </Form.Label>
      <div className="position-relative">
        {options?.map((option) => {
          const showLabel = !(
            option.id === '__custom__' && radioValue === '__custom__'
          );

          return (
            <Form.Check
              inline={inline}
              className={!inline ? 'mt-2 mb-1' : ''}
              name={id}
              id={option.id}
              key={option.id}
              label={showLabel ? option.label : ''}
              type="radio"
              disabled={isDisabled}
              onChange={(e) => {
                const value = e.target.id;
                setRadioValue(value);
                if (value === '__custom__') {
                  if (fixedOptionIds.includes(formState[id] as string)) {
                    onChange(id, '');
                  }
                } else {
                  onChange(id, value);
                }
              }}
              checked={isOptionChecked(option.id)}
            />
          );
        })}

        {radioValue === '__custom__' && (
          <Form.Control
            type="text"
            className={classes.customInput}
            placeholder="Enter custom model..."
            value={
              typeof formState[id] === 'string' &&
              !fixedOptionIds.includes(formState[id] as string)
                ? (formState[id] as string)
                : ''
            }
            onChange={(e) => {
              onChange(id, e.target.value);
            }}
            disabled={isDisabled}
          />
        )}
      </div>
    </div>
  );
};

export { SettingRadio };
