import React from 'react';
import {
  render,
  fireEvent,
  screen,
  queryByAttribute,
} from '@testing-library/react';
import '@testing-library/jest-dom';
import { SettingTextarea } from './SettingTextarea';
import { SiteSetting, State } from '../types/site-settings.type';

const mockOnChange = jest.fn();

const formState: State['form'] = {
  textareaSetting: '',
};

const baseSetting: SiteSetting = {
  id: 'textareaSetting',
  type: 'textarea',
  label: 'Textarea Setting',
  popoverText: 'This is a tooltip',
};

describe('SettingTextarea Component', () => {
  beforeEach(() => {
    mockOnChange.mockClear();
  });

  it('renders the textarea setting correctly', () => {
    render(
      <SettingTextarea
        setting={baseSetting}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    expect(screen.getByText('Textarea Setting')).toBeInTheDocument();
    expect(screen.getByRole('textbox')).toBeInTheDocument();
    expect(screen.getByRole('textbox')).toHaveValue('');
  });

  it('handles text input correctly', () => {
    render(
      <SettingTextarea
        setting={baseSetting}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    const textarea = screen.getByRole('textbox');
    fireEvent.change(textarea, { target: { value: 'New text' } });

    expect(mockOnChange).toHaveBeenCalledWith('textareaSetting', 'New text');
  });

  it('displays disabled state correctly', () => {
    const disabledSetting: SiteSetting = {
      ...baseSetting,
      disabled: () => true,
    };

    render(
      <SettingTextarea
        setting={disabledSetting}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    const textarea = screen.getByRole('textbox');
    expect(textarea).toBeDisabled();
  });

  it('handles multi-line input correctly', () => {
    render(
      <SettingTextarea
        setting={baseSetting}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    const textarea = screen.getByRole('textbox');
    fireEvent.change(textarea, {
      target: { value: 'Line 1\nLine 2\nLine 3' },
    });

    expect(mockOnChange).toHaveBeenCalledWith(
      'textareaSetting',
      'Line 1\nLine 2\nLine 3'
    );
  });

  it('displays info icon when popoverText is provided', () => {
    const { container } = render(
      <SettingTextarea
        setting={baseSetting}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    const infoIcon = queryByAttribute(
      'data-test-id',
      container,
      'info-icon-span'
    );

    expect(infoIcon).toBeInTheDocument();
  });

  it('does not display info icon when popoverText is not provided', () => {
    const settingWithoutPopover: SiteSetting = {
      ...baseSetting,
      popoverText: undefined,
    };

    const { container } = render(
      <SettingTextarea
        setting={settingWithoutPopover}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    const infoIcon = queryByAttribute(
      'data-test-id',
      container,
      'info-icon-span'
    );

    expect(infoIcon).not.toBeInTheDocument();
  });
});
