import React from 'react';
import {
  render,
  fireEvent,
  screen,
  queryByAttribute,
  waitFor,
} from '@testing-library/react';
import '@testing-library/jest-dom';
import { Settings } from './Settings';
import { SiteSetting } from '../types/site-settings.type';
import { testEmail } from '../api/test-email';
import { deleteUsageData } from '../api/delete-usage-data';

const mockAddNotification = jest.fn();
jest.mock('../../../stores/use-notification-store.ts', () => ({
  useNotificationStore: () => ({ addNotification: mockAddNotification }),
}));

jest.mock('../api/test-email', () => ({
  testEmail: jest.fn(),
}));

jest.mock('../api/delete-usage-data', () => ({
  deleteUsageData: jest.fn(),
}));

jest.mock('utils/get-global-arc', () => {
  const arcMock: any = {
    frontend: {
      globals: {
        settings: {
          thumbnailPilSupported: true,
          emailEncryptionSupported: true,
        },
      },
    },
  };

  return {
    getGlobalArc: () => arcMock,
  };
});

describe('Settings Component', () => {
  const mockOnChange = jest.fn();
  const mockChangeToViewMode = jest.fn();

  const formState = {
    email_host: 'smtp.gmail.com',
    email_host_password: 'password',
    email_port: '465',
    email_host_user: 'viztest00@gmail.com',
    email_protocol: 'EMAIL_USE_SSL',
    email_username: '',
    email_prefix: 'prefix',
  };

  const baseSettings: SiteSetting[] = [
    { id: 'checkboxSetting', type: 'checkbox', label: 'Checkbox Setting' },
    {
      id: 'dropdownSetting',
      type: 'dropdown',
      label: 'Dropdown Setting',
      options: [{ id: 'option1', label: 'Option 1' }],
    },
    { id: 'inputSetting', type: 'input', label: 'Input Setting' },
    { id: 'numericSetting', type: 'numeric', label: 'Numeric Setting' },
    {
      id: 'radioSetting',
      type: 'radio',
      label: 'Radio Setting',
      options: [{ id: 'option1', label: 'Option 1' }],
    },
    { id: 'textareaSetting', type: 'textarea', label: 'Textarea Setting' },
    { id: 'email_prefix', type: 'input', label: 'Email Prefix' },
    {
      id: 'usage_analytics_expiration_days',
      type: 'input',
      label: 'Usage Analytics Expiration Days',
    },
  ];

  it.each(baseSettings)('renders %s setting correctly', (setting) => {
    render(
      <Settings
        settings={[setting]}
        formState={formState}
        onChange={mockOnChange}
        changeToViewMode={mockChangeToViewMode}
      />
    );

    expect(screen.getByText(setting.label)).toBeInTheDocument();
  });

  it('triggers testEmail when Send test email button is clicked', async () => {
    (testEmail as jest.Mock).mockResolvedValue({});

    const settings: SiteSetting[] = [
      { id: 'email_prefix', type: 'input', label: 'Email Prefix' },
    ];
    render(
      <Settings
        settings={settings}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    const testEmailButton = screen.getByText('Send test email');
    fireEvent.click(testEmailButton);

    expect(testEmail).toHaveBeenCalled();

    await waitFor(() => {
      expect(mockAddNotification).toHaveBeenCalledWith(
        expect.objectContaining({
          id: 'site-settings-success',
          variant: 'success',
          message: 'Email Status: 1 sent',
        }),
        { delay: 1500 }
      );
    });
  });

  it('triggers deleteUsageData when Delete Usage Data button is clicked', async () => {
    const settings: SiteSetting[] = [
      {
        id: 'usage_analytics_expiration_days',
        type: 'input',
        label: 'Usage Analytics Expiration Days',
      },
    ];

    (deleteUsageData as jest.Mock).mockResolvedValue({});

    render(
      <Settings
        settings={settings}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    const deleteDataButton = screen.getByText('Delete Usage Data');
    fireEvent.click(deleteDataButton);

    const confirmDeleteButton = await screen.findByText('Delete');
    fireEvent.click(confirmDeleteButton);

    expect(deleteUsageData).toHaveBeenCalled();

    await waitFor(() => {
      expect(mockAddNotification).toHaveBeenCalledWith(
        expect.objectContaining({
          id: 'site-settings-success',
          variant: 'success',
          message: 'The usage data was successfully deleted',
        }),
        { delay: 1500 }
      );
    });
  });

  it('renders disabled setting', () => {
    const settings: SiteSetting[] = [
      {
        id: 'disabledSetting',
        type: 'input',
        label: 'Disabled Setting',
        disabled: () => true,
        disabledText: 'Disabled Info',
      },
    ];
    render(
      <Settings
        settings={settings}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    const triggerElement = screen.getByText('Disabled Setting');
    fireEvent.mouseOver(triggerElement);

    expect(screen.getByText('Disabled Setting')).toBeInTheDocument();
  });

  it('does not render hidden settings', () => {
    const settings: SiteSetting[] = [
      {
        id: 'hiddenSetting',
        type: 'input',
        label: 'Hidden Setting',
        hidden: () => true,
      },
    ];
    render(
      <Settings
        settings={settings}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    expect(screen.queryByText('Hidden Setting')).not.toBeInTheDocument();
  });

  it('renders setting with highlight', () => {
    const settings: SiteSetting[] = [
      { id: 'highlightedSetting', type: 'input', label: 'Highlighted Setting' },
    ];
    const highlightedId = 'highlightedSetting';
    const { container } = render(
      <Settings
        settings={settings}
        formState={formState}
        highlightedId={highlightedId}
        onChange={mockOnChange}
      />
    );

    const highlightedElement = queryByAttribute(
      'data-test-id',
      container,
      'highlighted-setting'
    );
    expect(highlightedElement).toBeInTheDocument();
  });

  it('triggers changeToViewMode when View in group button is clicked', () => {
    const settings: SiteSetting[] = [
      { id: 'viewModeSetting', type: 'input', label: 'View Mode Setting' },
    ];
    render(
      <Settings
        settings={settings}
        formState={formState}
        onChange={mockOnChange}
        changeToViewMode={mockChangeToViewMode}
        searchView={true}
      />
    );

    const viewModeButton = screen.getByText('View in group');
    fireEvent.click(viewModeButton);
    expect(mockChangeToViewMode).toHaveBeenCalledWith('viewModeSetting');
  });

  it('renders infoText when provided in a setting', () => {
    const settingsWithInfoText: SiteSetting[] = [
      {
        id: 'setting1',
        type: 'input',
        label: 'Test Setting',
        infoText: 'Additional information about this setting',
      },
    ];

    render(
      <Settings
        settings={settingsWithInfoText}
        formState={formState}
        onChange={mockOnChange}
      />
    );

    expect(
      screen.getByText('Additional information about this setting')
    ).toBeInTheDocument();
  });
});
