import React, { MouseEventHandler } from 'react';
import { Button, Form, Modal, Tooltip } from 'react-bootstrap';
import classNames from 'classnames';
import {
  State,
  SiteSetting,
  SettingValue,
  ConfirmObj,
} from '../types/site-settings.type';
import classes from '../SiteSettingsPage.module.scss';
import { SettingCheckbox } from './SettingCheckbox';
import { SettingInput } from './SettingInput';
import { SettingNumeric } from './SettingNumeric';
import { SettingRadio } from './SettingRadio';
import { SettingTextarea } from './SettingTextarea';
import { VizTooltip } from 'components/VizTooltip';
import { SettingRange } from './SettingRange';
import { getGlobalArc } from 'utils/get-global-arc';
import { useNotificationStore } from 'stores/use-notification-store';
import { testEmail } from '../api/test-email';
import { deleteUsageData } from '../api/delete-usage-data';
import axios, { AxiosError } from 'axios';
import { confirmAlert } from 'react-confirm-alert';
import { VizButton } from 'components/viz-button/VizButton';
import { SettingDropdown } from './SettingDropdown';

const arc = getGlobalArc();

type SettingsProps = {
  settings: SiteSetting[];
  formState: State['form'];
  highlightedId?: string;
  shouldValidate?: boolean;
  onChange: (id: string, value: SettingValue, confirmObj?: ConfirmObj) => void;
  changeToViewMode?: (id: string) => void;
  searchView?: boolean;
};

const Settings = ({
  settings,
  formState,
  highlightedId,
  shouldValidate = false,
  onChange,
  changeToViewMode,
  searchView,
}: SettingsProps): JSX.Element => {
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const renderInput = (setting: SiteSetting) => {
    switch (setting.type) {
      case 'checkbox':
        return (
          <SettingCheckbox
            setting={setting}
            formState={formState}
            onChange={onChange}
          />
        );
      case 'input':
        let element: JSX.Element | null = null;

        if (setting.id === 'email_prefix') {
          const handleTestEmail = async () => {
            const {
              email_host,
              email_port,
              email_host_user,
              email_host_password,
              email_protocol,
              email_username,
              email_prefix,
            } = formState;

            let errorText: string | null = null;

            if (!email_host || !email_port || !email_host_user) {
              errorText = 'Need to fill out SMTP Host, Port, Email Address';
            } else if (!arc.frontend.globals.settings.thumbnailPilSupported) {
              errorText =
                'Snapshots are not supported due to missing dependencies. Please see Documentation.';
            } else if (
              !arc.frontend.globals.settings.emailEncryptionSupported
            ) {
              errorText =
                'Encryption is not enabled due to missing dependencies. Please see Documentation.';
            }

            if (errorText) {
              addNotification(
                {
                  id: 'site-settings-danger',
                  variant: 'danger',
                  message: errorText,
                },
                { delay: 3000 }
              );

              return;
            }

            const params: {
              EMAIL_HOST: SettingValue;
              EMAIL_PORT: SettingValue;
              EMAIL_HOST_USER: SettingValue;
              EMAIL_PROTOCOL: SettingValue;
              email_username: SettingValue;
              email_prefix: SettingValue;
              EMAIL_HOST_PASSWORD?: SettingValue;
            } = {
              EMAIL_HOST: email_host,
              EMAIL_PORT: email_port,
              EMAIL_HOST_USER: email_host_user,
              EMAIL_PROTOCOL: email_protocol,
              email_username,
              email_prefix,
            };

            params['EMAIL_HOST_PASSWORD'] = email_host_password ?? '';

            try {
              await testEmail(params);

              addNotification(
                {
                  id: 'site-settings-success',
                  variant: 'success',
                  message: 'Email Status: 1 sent',
                },
                { delay: 1500 }
              );
            } catch (err: unknown | AxiosError) {
              addNotification(
                {
                  id: 'site-settings-danger',
                  variant: 'danger',
                  message: `${
                    axios.isAxiosError(err)
                      ? (err as AxiosError).response?.data
                      : 'Authentication failed'
                  }`,
                },
                { delay: 3000 }
              );
            }
          };

          element = (
            <Button
              variant="light"
              className="btn btn-secondary btn-sm mt-3"
              data-test-id="site_settings_test_email_btn"
              onClick={handleTestEmail}
            >
              Send test email
            </Button>
          );
        } else if (setting.id === 'usage_analytics_expiration_days') {
          const submitDeleteUsageData = async () => {
            try {
              await deleteUsageData();

              addNotification(
                {
                  id: 'site-settings-success',
                  variant: 'success',
                  message: 'The usage data was successfully deleted',
                },
                { delay: 1500 }
              );
            } catch (err: unknown | AxiosError) {
              addNotification(
                {
                  id: 'site-settings-danger',
                  variant: 'danger',
                  message: `${
                    axios.isAxiosError(err)
                      ? (err as AxiosError).response?.data
                      : 'Could not delete the usage data'
                  }`,
                },
                { delay: 3000 }
              );
            }
          };

          const deleteUsageDataModal: MouseEventHandler<HTMLDivElement> = (
            e
          ) => {
            e.preventDefault();

            confirmAlert({
              customUI: ({ onClose }) => {
                return (
                  <Modal
                    show={true}
                    onHide={onClose}
                    data-test-id="deleteUsageData"
                  >
                    <Modal.Header closeButton>
                      <Modal.Title>Delete Usage Data</Modal.Title>
                    </Modal.Header>
                    <Modal.Body>
                      <p>
                        Are you sure you want to delete all of the usage data?
                      </p>
                    </Modal.Body>
                    <Modal.Footer>
                      <VizButton
                        color="secondary"
                        size="medium"
                        onClick={onClose}
                      >
                        Cancel
                      </VizButton>
                      <VizButton
                        color="primary"
                        size="medium"
                        testId="confirmDeleteUsageData"
                        onClick={() => {
                          submitDeleteUsageData();
                          onClose();
                        }}
                      >
                        Delete
                      </VizButton>
                    </Modal.Footer>
                  </Modal>
                );
              },
            });
          };

          element = (
            <Button
              variant="light"
              className="btn btn-secondary btn-sm mt-3"
              data-test-id="delete-usage-data"
              onClick={deleteUsageDataModal}
            >
              Delete Usage Data
            </Button>
          );
        }

        return (
          <SettingInput
            setting={setting}
            formState={formState}
            shouldValidate={shouldValidate}
            onChange={onChange}
            additionalElement={element}
          />
        );
      case 'numeric':
        return (
          <SettingNumeric
            setting={setting}
            formState={formState}
            onChange={onChange}
          />
        );
      case 'radio':
        return (
          <SettingRadio
            setting={setting}
            formState={formState}
            onChange={onChange}
          />
        );
      case 'textarea':
        return (
          <SettingTextarea
            setting={setting}
            formState={formState}
            onChange={onChange}
          />
        );
      case 'range':
        return (
          <SettingRange
            setting={setting}
            formState={formState}
            onChange={onChange}
          />
        );
      case 'dropdown':
        return (
          <SettingDropdown
            setting={setting}
            formState={formState}
            onChange={onChange}
          />
        );
    }
  };

  return (
    <div>
      {settings.map((setting) => {
        const element = (
          <>
            {renderInput(setting)}
            {setting.infoText && (
              <p className="form-text mb-0 mt-2">{setting.infoText}</p>
            )}
          </>
        );

        return setting.hidden?.(formState) ? null : (
          <Form.Group
            className={classNames('mb-4 d-flex', classes.wrp, {
              [classes.highlight]: setting.id === highlightedId,
            })}
            data-test-id={
              setting.id === highlightedId ? 'highlighted-setting' : null
            }
            key={setting.id}
          >
            <div className={classNames('d-flex flex-column')}>
              {setting.disabled?.(formState) ? (
                <VizTooltip
                  placement="top"
                  overlay={
                    <Tooltip id={setting.id}>
                      {
                        <span className="d-flex p-2">
                          {setting.disabledText}
                        </span>
                      }
                    </Tooltip>
                  }
                >
                  <div>{element}</div>
                </VizTooltip>
              ) : (
                element
              )}
            </div>
            {searchView && (
              <Button
                variant="link"
                className={classes.linkBtn}
                data-test-id="site_settings_view_mode_btn"
                onClick={() => changeToViewMode?.(setting.id)}
              >
                View in group
              </Button>
            )}
          </Form.Group>
        );
      })}
    </div>
  );
};

export { Settings };
