import moment from 'moment';
import React from 'react';
import { getGlobalArc } from 'utils/get-global-arc';
import { SettingGroup, State } from '../types/site-settings.type';
import { getCmlAuthModeLabel, getEngineLabel } from '../utils/setting-utils';
import {
  AiSetting,
  AiTab,
  EngineSpecificSettings,
} from '../types/ai-settings.type';

const arc = getGlobalArc();

const modalPhrase = 'By clicking "accept" below';
const tooltipPhrase = 'By enabling this feature';
const apiKeyWarning =
  "You acknowledge that the API Key provided here is used to access the external third-party AI service provider's products and/or services.  You agree to treat this API Key as a sensitive credential and to ensure that it is not shared with unauthorized parties or exposed publicly.";
const bedrockPopoverText = (
  <p>
    Paste your AWS credentials in INI format here. This should include a
    [default] section with keys such as aws_access_key_id,
    aws_secret_access_key, aws_session_token, aws_security_token, etc. These
    credentials are used to authenticate with Amazon Bedrock services. If the
    field is left empty, your AWS credentials will be sourced from your AWS
    credentials file.
    <br />
    <br />
    You acknowledge that the credentials provided here are used to access Amazon
    Bedrock and related AWS services. You agree to treat these credentials as
    sensitive information and ensure they are not shared with unauthorized
    parties or exposed publicly.
  </p>
);
const bedrockCredentialsPlaceholder = `[default]
aws_access_key_id = ACCESS KEY
aws_secret_access_key = SECRET ACCESS KEY
aws_session_token = SESSION TOKEN
aws_security_token = SECURITY TOKEN
region = us-west-2
`;
const missingProfilesWarning =
  'Create at least one embeddings and completion profile to use the AI visual';

const bedrockEmbeddingModels = [
  {
    id: 'amazon.titan-embed-text-v1',
    label: 'amazon.titan-embed-text-v1',
  },
  {
    id: 'amazon.titan-embed-text-v2:0',
    label: 'amazon.titan-embed-text-v2:0',
  },
  {
    id: '__custom__',
    label: 'Custom',
  },
];

const bedrockCompletionModels = [
  {
    id: 'amazon.titan-text-lite-v1',
    label: 'amazon.titan-text-lite-v1',
  },
  {
    id: 'amazon.titan-text-express-v1',
    label: 'amazon.titan-text-express-v1',
  },
  {
    id: 'anthropic.claude-3-5-sonnet-20241022-v2:0',
    label: 'anthropic.claude-3-5-sonnet-20241022-v2:0',
  },
  {
    id: 'anthropic.claude-3-5-haiku-20241022-v1:0',
    label: 'anthropic.claude-3-5-haiku-20241022-v1:0',
  },
  {
    id: 'anthropic.claude-3-opus-20240229-v1:0',
    label: 'anthropic.claude-3-opus-20240229-v1:0',
  },
  {
    id: '__custom__',
    label: 'Custom',
  },
];

const AI_TERM = (phrase: string) => (
  <p>
    <strong>IMPORTANT: Please read the following before proceeding.</strong>{' '}
    <br />
    <br />
    {phrase}, you are agreeing to use the Artificial Intelligence (AI) visual
    subject to these terms of Cloudera and with the understanding that your data
    and metadata will be shared with the external third-party AI service
    provider (e.g., a vector database and/or a large language model service
    provider) of your own choosing at your own risk. For clarification, you
    choose which AI visual to configure and use, and Cloudera does not bundle
    any such AI visual on your behalf. The AI visual is provided “as is.”
    <br />
    <br />
    CLOUDERA DOES NOT MAKE OR GIVE ANY REPRESENTATION, WARRANTY, OR COVENANT OF
    ANY KIND, WHETHER EXPRESS OR IMPLIED, IN CONNECTION WITH THE AI VISUAL AND
    ANY THIRD-PARTY AI SERVICE PROVIDER. CLOUDERA IS NOT RESPONSIBLE FOR YOUR
    USE OF THE AI VISUAL AND ANY THIRD-PARTY AI SERVICE PROVIDER.
    <br />
    <br />
    If you are acting on behalf of an entity, you represent and warrant that you
    have the legal authority to accept these terms and to use the AI visual on
    behalf of that entity. You represent and warrant that you reviewed,
    understood, and accepted the terms and conditions of the third-party AI
    service provider, and that you will comply with said terms and conditions.
    If you do not accept these terms of Cloudera, then you must not use the AI
    visual.
  </p>
);

const { settings, documentation, product, authentication } =
  arc.frontend.globals;

export const DEFAULT_SITE_SETTING_VALUES = {
  site_enable_column_rename: false,
  enable_counting_dashboards_for_datasets: true,
  site_enable_extracts: false,
  site_enable_data_profiling: true,
  site_enable_download_from_url: true,
  site_hide_refresh_button_from_connection_explorer: false,
  site_enable_derived_data: true,
  site_enable_explore_interfaces: true,
  site_enable_visual_transitions: false,
  site_max_trellis_displayed: settings.siteMaxTrellisDisplayed,
  visual_max_dataresults: settings.visualMaxDataResults,
  csv_excel_cache_result: true,
  visual_maxrows: settings.visualMaxRows,
  maximum_canvas_width: settings.maximumCanvasWidth,
  site_enable_custom_styles: true,
  enable_clone_info_badge: false,
  enable_legacy_crosstab: false,
  site_enable_search_visual: false,
  site_csv_download_on: true,
  download_maxrows: settings.downloadMaxRows,
  site_thumbnails_enable: true,
  site_png_download_on: false,
  site_thumbnails_ip: settings.siteThumbnailsIp,
  dashboard_width: settings.dashboardWidth,
  dashboard_loading_message_in_viewmode:
    settings.dashboardLoadingMessageInViewMode,
  google_api_id: settings.googleApiId,
  google_client_id: settings.googleClientId,
  enable_ai_assistant: false,
  redact_ai_assistant_logs: true,
  embeddings_cml_auth_mode: settings.aiCmlAuthMode,
  completion_cml_auth_mode: settings.aiCmlAuthMode,
  embeddings_api_key: settings.aiApiKey,
  completion_api_key: settings.aiApiKey,
  embeddings_auth_key: settings.aiAuthKey,
  completion_auth_key: settings.aiAuthKey,
  embeddings_bedrock_credentials: settings.aiBedrockCredentials,
  completion_bedrock_credentials: settings.aiBedrockCredentials,
  embeddings_model: settings.embeddingsModel,
  completion_model: settings.completionModel,
  embeddings_bedrock_model: settings.embeddingsBedrockModel,
  completion_bedrock_model: settings.completionBedrockModel,
  embeddings_input_param: settings.embeddingInputParam,
  completion_input_param: settings.completionInputParam,
  embeddings_service_url: settings.aiServiceUrl,
  completion_service_url: settings.aiServiceUrl,
  embeddings_response_param: settings.embeddingResponseParam,
  completion_response_param: settings.completionResponseParam,
  embeddings_engine: settings.aiEngine,
  completion_engine: settings.aiEngine,
  completion_extra_arg: settings.completionExtraArg,
  completion_temperature: settings.completionTemperature,
  completion_stream: settings.completionStream,
  embeddings_custom_engine_settings: settings.defaultEmbeddingsEngineSettings,
  completion_custom_engine_settings: settings.defaultCompletionEngineSettings,
  embeddings_bedrock_custom_transformers:
    settings.embeddingsBedrockCustomTransformers,
  completion_bedrock_custom_transformers:
    settings.completionBedrockCustomTransformers,
  ai_term_agreed: settings.aiTermAgreed,
  ai_term_agreed_date: settings.aiTermAgreedDate,
  mapbox_access_token: settings.mapBoxAccessToken,
  local_mapbox_server: settings.localMapBoxServer,
  doc_link_text: documentation.label,
  doc_link_url: documentation.url,
  support_option: settings.supportOption,
  support_web_page_link_text: settings.supportWebPageLabel,
  support_web_page_link_url: settings.supportWebPageUrl,
  support_email: settings.supportEmail,
  enable_user_create: true,
  enable_remote_user_auth: false,
  min_pw_length: '5',
  pw_require_number: false,
  pw_require_symbol: false,
  default_group_membership_for_new_users:
    settings.defaultGroupMembershipForNewUsers,
  login_instructions: settings.loginInstructions,
  enable_speech_recognition: false,
  site_enable_search: false,
  site_enable_search_menu: false,
  scheduling_enabled: true,
  email_enabled: settings.emailEnabled,
  email_host_user: settings.emailHostUser,
  email_username: settings.emailUsername,
  email_host_password: settings.emailHostPassword,
  email_host: settings.emailHost,
  email_port: settings.emailPort,
  email_protocol: settings.emailProtocol,
  email_prefix: settings.emailPrefix,
  inspect_settings_enabled: false,
  jserrors_enabled: true,
  page_profile_enabled: true,
  page_profile_appview_enabled: true,
  bulk_max_entries: settings.bulkMaxEntries,
  bulk_max_interval: settings.bulkMaxInterval,
  bulk_max_retention: settings.bulkMaxRetention,
  notification_max_retention: '1',
  webhook_enabled: false,
  webhook_whitelist: settings.webhookWhitelist,
  remotedata_url_whitelist: settings.remoteDataUrlWhitelist,
  advanced_settings: settings.advancedSettings,
  jira_enabled: false,
  jira_server: settings.jiraServer,
  jira_user: settings.jiraUser,
  jira_password: settings.jiraPassword,
  jira_newissue_defaults: settings.jiraNewIssueDefaults,
  enable_usage_analytics: false,
  usage_analytics_expiration_days: settings.usageAnalyticsExpirationDays,
  enable_version_control: false,
  enable_dataset_version_control: false,
  unnamed_versions_deletion_interval: '7',
  unnamed_versions_max_retention: '30',
};

export const SITE_SETTING_LABELS = {
  site_enable_column_rename:
    'Enable automatic field renaming during dataset creation',
  enable_counting_dashboards_for_datasets:
    'Enable showing the count of associated dashboards of the datasets',
  site_enable_extracts: 'Enable Data Extracts',
  site_enable_data_profiling: 'Enable Data Profiling',
  site_enable_download_from_url: 'Enable Data Downloads from URL',
  site_hide_refresh_button_from_connection_explorer:
    'Hide "Refresh" button from Connection Explorer',
  site_enable_derived_data: 'Enable Derived Data',
  site_enable_explore_interfaces: 'Enable Explore',
  site_enable_visual_transitions: 'Enable Visual Transitions',
  site_max_trellis_displayed:
    'Max Count Of Trellises To Be Displayed Before Warning',
  visual_max_dataresults:
    'Maximum data results size for visuals and downloads (in MB)',
  csv_excel_cache_result: 'Cache results of CSV/Excel downloads',
  visual_maxrows: 'Maximum data rows for visuals',
  maximum_canvas_width: 'Maximum width of canvas visuals',
  site_enable_custom_styles: 'Enable Custom Styles',
  enable_clone_info_badge:
    'Show Descendant Dashboard information on the Visuals page',
  enable_legacy_crosstab: 'Enable legacy crosstab visual',
  site_enable_search_visual: 'Enable Search Visual',
  site_csv_download_on: 'Enable "Download as CSV/Excel" by default',
  download_maxrows: 'Maximum data rows for download requests',
  site_thumbnails_enable:
    'Enable snapshots (required for thumbnails, email, and image downloads)',
  site_png_download_on: 'Enable "Download as Image/PDF" by default',
  site_thumbnails_ip: 'IP/Hostname to use to connect back to the server',
  dashboard_width: 'Default Dashboard Width',
  dashboard_loading_message_in_viewmode:
    'Show "This dashboard is still loading data" in view mode',
  google_api_id: 'Google API Id',
  google_client_id: 'Google Client Id',
  enable_ai_assistant: 'Enable AI visual',
  redact_ai_assistant_logs: 'Redact AI visual logs',
  ai_cml_auth_mode: 'Authentication mode',
  ai_api_key: 'API Key',
  ai_access_key: 'Access Key',
  ai_auth_key: 'Authorization API Key',
  ai_bedrock_credentials: 'AWS Credentials',
  ai_service_url: 'Service URL',
  ai_input_param: 'Service input parameter name',
  ai_response_param: 'Service response parameter name',
  ai_engine: 'Engine',
  ai_model: 'Model',
  ai_bedrock_model: 'Model',
  ai_temperature: 'Temperature',
  ai_stream: 'Enable Streaming Responses',
  ai_custom_engine_settings: 'AI engine settings',
  ai_embeddings_settings: 'AI Embedding settings',
  ai_completion_settings: 'AI Completion settings',
  ai_bedrock_custom_transformers: 'Custom Request/Response transformers',
  ai_extra_arg: 'Extra arguments',
  mapbox_access_token: 'Mapbox Token',
  local_mapbox_server: 'IP/Hostname of Atlas Mapbox Server (optional)',
  doc_link_text: 'Text to display for documentation link',
  doc_link_url: 'URL for documentation link',
  support_option: 'Show Support',
  support_web_page_link_text: 'Text to display for Support web page link',
  support_web_page_link_url: 'URL for Support web page link',
  support_email: 'Email address for Support',
  enable_user_create: 'Enable Local User Creation',
  enable_remote_user_auth:
    'Permit authentication via the REMOTE_USER http header',
  min_pw_length: 'Minimum password length',
  pw_require_number: 'Require that passwords contain one or more numbers',
  pw_require_symbol: 'Require that passwords contain one or more symbols',
  default_group_membership_for_new_users:
    'Default group membership for new users',
  login_instructions: 'Instructions to display on login page',
  enable_speech_recognition: 'Enable Speech Detection on site (Tech Preview)',
  site_enable_search: 'Enable Search',
  site_enable_search_menu: 'Enable Search in Menu Bar',
  scheduling_enabled: 'Enable scheduled jobs',
  email_enabled: 'Permit emails',
  email_host_user: 'Email Address',
  email_username: 'Username',
  email_host_password: 'Password',
  email_host: 'SMTP Host',
  email_port: 'Port',
  email_protocol: 'Email protocol',
  email_prefix: 'Default email subject prefix',
  inspect_settings_enabled:
    'Enable the "Server Settings" tab within the "Activity Log" screen',
  jserrors_enabled:
    'Verbose Error Reporting. Display Javascript page errors on screen in app view/edit pages',
  page_profile_enabled:
    'Enable the "Show Usage Info" and Ctrl-Shift-G shortcut to view page profiles within all pages',
  page_profile_appview_enabled:
    'Additionally enable the Ctrl-Shift-G shortcut to view page profiles within dashboard/app view pages',
  bulk_max_entries: 'Maximum Entries',
  bulk_max_interval: 'Check Interval',
  bulk_max_retention: 'Maximum Retention Days',
  webhook_enabled: 'Allow users to interact with webhooks',
  webhook_whitelist: 'URLs that webhook actions are permitted to interact with',
  remotedata_url_whitelist: 'URLs that remote data columns can be fetched from',
  advanced_settings:
    'Advanced Site Settings (Application restart required after saving)',
  jira_enabled: 'Allow users to interact with a JIRA system',
  jira_server: 'JIRA Server URL',
  jira_user: 'JIRA User',
  jira_password: 'JIRA Password',
  jira_newissue_defaults: 'JSON string with default values for new issues',
  enable_usage_analytics: `Enable ${product.name} Usage Analytics (Tech Preview)`,
  usage_analytics_expiration_days: 'Maximum days for expiring usage data',
  mute_notifications: 'Mute Notifications',
  notification_max_retention: 'Notification maximum retention',
  enable_version_control: 'Enable Version Control (Tech Preview)',
  enable_dataset_version_control:
    'Enable Dataset Version Control (Tech Preview)',
  unnamed_versions_deletion_interval:
    'Unnamed Dataset Version Deletion Interval',
  unnamed_versions_max_retention:
    'Unnamed Dataset Version Maximum Retention Days',
};

const aiNotEnabledText = `This is currently disabled due to the "${SITE_SETTING_LABELS['enable_ai_assistant']}" setting`;
const isCdvModeCml = arc.frontend.globals.settings.isCdvModeCml;

export const createAiSettings = (tab: AiTab): EngineSpecificSettings => {
  const settings: { [key: string]: AiSetting } = {
    engine: {
      id: `${tab}_engine`,
      label: SITE_SETTING_LABELS['ai_engine'],
      type: 'dropdown',
      options: [
        {
          id: 'llama',
          label: getEngineLabel('llama', tab),
        },
        { id: 'caii', label: getEngineLabel('caii', tab) },
        { id: 'open_ai', label: getEngineLabel('open_ai', tab) },
        { id: 'azure_ai', label: getEngineLabel('azure_ai', tab) },
        { id: 'bedrock', label: getEngineLabel('bedrock', tab) },
        { id: 'other', label: getEngineLabel('other', tab) },
      ],
    },
    cml_auth_mode: {
      id: `${tab}_cml_auth_mode`,
      label: SITE_SETTING_LABELS['ai_cml_auth_mode'],
      type: 'radio',
      popoverText:
        'Please note for JWT authentication to work the Cloudera AI Inference model and Cloudera Data Visualization should be in the same environment.',
      options: [
        { id: 'jwt', label: getCmlAuthModeLabel('jwt') },
        { id: 'api_key', label: getCmlAuthModeLabel('api_key') },
      ],
    },
    api_key: {
      id: `${tab}_api_key`,
      label: SITE_SETTING_LABELS['ai_api_key'],
      type: 'input',
      popoverText: apiKeyWarning,
      hidden: (formState?: State['form']) => {
        if (!formState) return false;

        const authMode = formState[`${tab}_cml_auth_mode`];
        return isCdvModeCml && authMode === 'jwt';
      },
      required: true,
    },
    auth_key: {
      id: `${tab}_auth_key`,
      label: SITE_SETTING_LABELS['ai_auth_key'],
      type: 'input',
      popoverText: apiKeyWarning,
    },
    bedrock_credentials: {
      id: `${tab}_bedrock_credentials`,
      label: SITE_SETTING_LABELS['ai_bedrock_credentials'],
      type: 'textarea',
      placeholder: bedrockCredentialsPlaceholder,
      popoverText: bedrockPopoverText,
    },
    model: {
      id: `${tab}_model`,
      label: SITE_SETTING_LABELS['ai_model'],
      type: 'input',
      popoverText: `You can specify which model should be used for the ${tab}.`,
      required: true,
    },
    bedrock_model: {
      id: `${tab}_bedrock_model`,
      label: SITE_SETTING_LABELS['ai_model'],
      type: 'radio',
      inline: false,
      options:
        tab === 'embeddings' ? bedrockEmbeddingModels : bedrockCompletionModels,
      required: true,
    },
    bedrock_engine_settings: {
      id: `${tab}_bedrock_custom_transformers`,
      label: SITE_SETTING_LABELS['ai_bedrock_custom_transformers'],
      type: 'textarea',
      popoverText:
        'Please provide any necessary custom transformers in the textarea below. If this field is left empty, the default transformers (compatible with supported models) will be used.',
      hidden: (state) => {
        if (!state) return true;

        const models =
          tab === 'embeddings'
            ? bedrockEmbeddingModels
            : bedrockCompletionModels;

        const value = state[`${tab}_bedrock_model`];

        return models
          .filter((model) => model.id !== 'custom')
          .map((model) => model.id)
          .includes(value as string);
      },
    },
    temperature: {
      id: `${tab}_temperature`,
      label: SITE_SETTING_LABELS['ai_temperature'],
      type: 'range',
      min: 0,
      max: 2,
      step: 0.1,
      popoverText:
        'A number between 0 and 2. The temperature is used to control the randomness of the output. 0 means more deterministic completions 2 means the responses will vary wildly.',
    },
    stream: {
      id: `${tab}_stream`,
      label: SITE_SETTING_LABELS['ai_stream'],
      type: 'checkbox',
    },
    service_url: {
      id: `${tab}_service_url`,
      label: SITE_SETTING_LABELS['ai_service_url'],
      type: 'input',
      required: true,
    },
    input_param: {
      id: `${tab}_input_param`,
      label: SITE_SETTING_LABELS['ai_input_param'],
      type: 'input',
      required: true,
    },
    response_param: {
      id: `${tab}_response_param`,
      label: SITE_SETTING_LABELS['ai_response_param'],
      type: 'input',
      required: true,
    },
    extraArgs: {
      id: `${tab}_extra_arg`,
      label: SITE_SETTING_LABELS['ai_extra_arg'],
      type: 'textarea',
      popoverText: `Provide any extra arguments for your ${tab} query.`,
    },
    custom_engine_settings: {
      id: `${tab}_custom_engine_settings`,
      label: SITE_SETTING_LABELS['ai_custom_engine_settings'],
      type: 'textarea',
    },
  };

  const caii_embedding_model: AiSetting = {
    ...settings.model,
  };
  caii_embedding_model.popoverText +=
    ' Note: please add "-query" suffix to the model name if the model requires input_type parameter on NVIDIA documentation.';

  const commonOpenAiSettings = [
    settings.engine,
    settings.api_key,
    settings.model,
  ];

  const commonCAIISettings = isCdvModeCml
    ? [
        settings.engine,
        settings.cml_auth_mode,
        settings.api_key,
        settings.service_url,
      ]
    : [settings.engine, settings.api_key, settings.service_url];

  const commonAzureAiSettings = [
    settings.engine,
    settings.api_key,
    settings.service_url,
  ];

  const accessKey = { ...settings.api_key };
  accessKey.label = SITE_SETTING_LABELS['ai_access_key'];

  const commonLlamaSettings = [
    settings.engine,
    accessKey,
    settings.auth_key,
    settings.service_url,
    settings.input_param,
    settings.response_param,
  ];
  const commonBedrockSettings = [
    settings.engine,
    settings.bedrock_credentials,
    settings.bedrock_model,
    settings.bedrock_engine_settings,
  ];
  const commonOtherSettings = [
    settings.engine,
    settings.api_key,
    settings.service_url,
    settings.custom_engine_settings,
  ];

  const engineSpecificSettings: EngineSpecificSettings = {
    open_ai: {
      embeddings: commonOpenAiSettings,
      completion: [
        ...commonOpenAiSettings,
        settings.temperature,
        settings.stream,
      ],
    },
    caii: {
      embeddings: [...commonCAIISettings, caii_embedding_model],
      completion: [
        ...commonCAIISettings,
        settings.model,
        settings.temperature,
        settings.stream,
      ],
    },
    azure_ai: {
      embeddings: commonAzureAiSettings,
      completion: [...commonAzureAiSettings, settings.stream],
    },
    llama: {
      embeddings: commonLlamaSettings,
      completion: [...commonLlamaSettings, settings.extraArgs],
    },
    bedrock: {
      embeddings: commonBedrockSettings,
      completion: [...commonBedrockSettings, settings.stream],
    },
    other: {
      embeddings: commonOtherSettings,
      completion: commonOtherSettings,
    },
  };

  return engineSpecificSettings;
};

export const DEFAULT_SETTING_GROUPS: SettingGroup[] = [
  {
    tab: 'Data',
    settings: [
      {
        id: 'site_enable_column_rename',
        label: SITE_SETTING_LABELS['site_enable_column_rename'],
        type: 'checkbox',
      },
      {
        id: 'enable_counting_dashboards_for_datasets',
        label: SITE_SETTING_LABELS['enable_counting_dashboards_for_datasets'],
        type: 'checkbox',
        popoverText: 'Disable if Data page is loading slowly',
      },
      {
        id: 'site_enable_extracts',
        label: SITE_SETTING_LABELS['site_enable_extracts'],
        type: 'checkbox',
      },
      {
        id: 'site_enable_data_profiling',
        label: SITE_SETTING_LABELS['site_enable_data_profiling'],
        type: 'checkbox',
      },
      {
        id: 'site_hide_refresh_button_from_connection_explorer',
        label:
          SITE_SETTING_LABELS[
            'site_hide_refresh_button_from_connection_explorer'
          ],
        type: 'checkbox',
      },
      {
        id: 'site_enable_download_from_url',
        label: SITE_SETTING_LABELS['site_enable_download_from_url'],
        type: 'checkbox',
      },
      {
        id: 'site_enable_derived_data',
        label: SITE_SETTING_LABELS['site_enable_derived_data'],
        type: 'checkbox',
      },
    ],
  },
  {
    tab: 'Dashboards',
    children: [
      {
        tab: 'Visuals',
        settings: [
          {
            id: 'site_enable_explore_interfaces',
            label: SITE_SETTING_LABELS['site_enable_explore_interfaces'],
            type: 'checkbox',
          },
          {
            id: 'site_enable_visual_transitions',
            label: SITE_SETTING_LABELS['site_enable_visual_transitions'],
            type: 'checkbox',
          },
          {
            id: 'site_max_trellis_displayed',
            label: SITE_SETTING_LABELS['site_max_trellis_displayed'],
            type: 'numeric',
          },
          {
            id: 'visual_max_dataresults',
            label: SITE_SETTING_LABELS['visual_max_dataresults'],
            type: 'numeric',
            infoText:
              'Impala, Hive, SQLITE connection CSV/Excel downloads are not limited by this size',
          },
          {
            id: 'csv_excel_cache_result',
            label: SITE_SETTING_LABELS['csv_excel_cache_result'],
            type: 'checkbox',
            infoText:
              'If CSV/Excel downloads are large, caching them might use too much memory. Disable this option to reduce memory need.',
          },
          {
            id: 'visual_maxrows',
            label: SITE_SETTING_LABELS['visual_maxrows'],
            type: 'numeric',
          },
          {
            id: 'maximum_canvas_width',
            label: SITE_SETTING_LABELS['maximum_canvas_width'],
            type: 'numeric',
            popoverText:
              'Defines the maximum width a canvas visual can be. If a canvas is too wide it can cause browsers to crash',
          },
          {
            id: 'site_enable_custom_styles',
            label: SITE_SETTING_LABELS['site_enable_custom_styles'],
            type: 'checkbox',
          },
          {
            id: 'enable_clone_info_badge',
            label: SITE_SETTING_LABELS['enable_clone_info_badge'],
            type: 'checkbox',
          },
          {
            id: 'enable_legacy_crosstab',
            label: SITE_SETTING_LABELS['enable_legacy_crosstab'],
            type: 'checkbox',
          },
          {
            id: 'site_enable_search_visual',
            label: SITE_SETTING_LABELS['site_enable_search_visual'],
            type: 'checkbox',
            disabled: (formState: State['form']) =>
              !formState['site_enable_search'],
            disabledText: `This is currently disabled due to the "${SITE_SETTING_LABELS['site_enable_search']}" setting`,
          },
          {
            id: 'site_csv_download_on',
            label: SITE_SETTING_LABELS['site_csv_download_on'],
            type: 'checkbox',
          },
          {
            id: 'download_maxrows',
            label: SITE_SETTING_LABELS['download_maxrows'],
            type: 'numeric',
          },
        ],
      },
      {
        tab: 'Snapshots',
        settings: [
          {
            id: 'site_thumbnails_enable',
            label: SITE_SETTING_LABELS['site_thumbnails_enable'],
            type: 'checkbox',
          },
          {
            id: 'site_png_download_on',
            label: SITE_SETTING_LABELS['site_png_download_on'],
            type: 'checkbox',
          },
          {
            id: 'site_thumbnails_ip',
            label: SITE_SETTING_LABELS['site_thumbnails_ip'],
            type: 'input',
          },
        ],
      },
      {
        tab: 'Config',
        settings: [
          {
            id: 'dashboard_width',
            label: SITE_SETTING_LABELS['dashboard_width'],
            type: 'numeric',
            infoText: 'Clear field for full width.',
          },
          {
            id: 'dashboard_loading_message_in_viewmode',
            label: SITE_SETTING_LABELS['dashboard_loading_message_in_viewmode'],
            type: 'checkbox',
          },
        ],
      },
      {
        tab: 'Interactive Map',
        settings: [
          {
            id: 'google_api_id',
            label: SITE_SETTING_LABELS['google_api_id'],
            type: 'input',
          },
          {
            id: 'google_client_id',
            label: SITE_SETTING_LABELS['google_client_id'],
            type: 'input',
          },
          {
            id: 'mapbox_access_token',
            label: SITE_SETTING_LABELS['mapbox_access_token'],
            type: 'input',
          },
          {
            id: 'local_mapbox_server',
            label: SITE_SETTING_LABELS['local_mapbox_server'],
            type: 'input',
          },
        ],
      },
    ],
  },
  {
    tab: 'Email',
    settings: [
      {
        id: 'email_enabled',
        label: SITE_SETTING_LABELS['email_enabled'],
        type: 'checkbox',
        disabled: (formState: State['form']) =>
          !formState['email_host'] ||
          !formState['email_port'] ||
          !formState['email_host_user'],
        disabledText: 'Need to fill out SMTP Host, Port, Email Address',
        popoverText:
          'This setting is required to allow users to email visuals and receive jobs and extract refresh status messages.',
      },
      {
        id: 'email_host_user',
        label: SITE_SETTING_LABELS['email_host_user'],
        type: 'input',
      },
      {
        id: 'email_username',
        label: SITE_SETTING_LABELS['email_username'],
        type: 'input',
        infoText: 'Optional: Defaults to email address if empty',
      },
      {
        id: 'email_host_password',
        label: SITE_SETTING_LABELS['email_host_password'],
        type: 'input',
      },
      {
        id: 'email_host',
        label: SITE_SETTING_LABELS['email_host'],
        type: 'input',
        placeholder: 'eg: smtp.gmail.com',
      },
      {
        id: 'email_port',
        label: SITE_SETTING_LABELS['email_port'],
        type: 'input',
        placeholder: 'eg: 587',
      },
      {
        id: 'email_protocol',
        label: SITE_SETTING_LABELS['email_protocol'],
        type: 'radio',
        options: [
          { id: 'EMAIL_USE_SSL', label: 'SSL' },
          { id: 'EMAIL_USE_TLS', label: 'TLS' },
          { id: 'NEITHER', label: 'None' },
        ],
      },
      {
        id: 'email_prefix',
        label: SITE_SETTING_LABELS['email_prefix'],
        type: 'input',
      },
    ],
  },
  {
    tab: 'Site Configuration',
    settings: [
      {
        id: 'mute_notifications',
        label: SITE_SETTING_LABELS['mute_notifications'],
        type: 'checkbox',
      },
      {
        id: 'notification_max_retention',
        label: SITE_SETTING_LABELS['notification_max_retention'],
        type: 'numeric',
        min: 0.25,
        step: 0.25,
        infoText: 'Maximum days before deleting notifications.',
      },
      {
        id: 'enable_speech_recognition',
        label: SITE_SETTING_LABELS['enable_speech_recognition'],
        type: 'checkbox',
      },
      {
        id: 'site_enable_search',
        label: SITE_SETTING_LABELS['site_enable_search'],
        type: 'checkbox',
      },
      {
        id: 'site_enable_search_menu',
        label: SITE_SETTING_LABELS['site_enable_search_menu'],
        type: 'checkbox',
        disabled: (formState: State['form']) =>
          !formState['site_enable_search'],
        disabledText: `This is currently disabled due to the "${SITE_SETTING_LABELS['site_enable_search']}" setting`,
      },
      {
        id: 'scheduling_enabled',
        label: SITE_SETTING_LABELS['scheduling_enabled'],
        type: 'checkbox',
        popoverText:
          'Requires Visualization Server restart for changes to take effect.',
      },
      {
        id: 'enable_usage_analytics',
        label: SITE_SETTING_LABELS['enable_usage_analytics'],
        type: 'checkbox',
      },
      {
        id: 'usage_analytics_expiration_days',
        label: SITE_SETTING_LABELS['usage_analytics_expiration_days'],
        type: 'numeric',
        min: 0.25,
        step: 0.25,
      },
    ],
  },
  {
    tab: 'AI Settings',
    children: [
      {
        tab: 'General',
        settings: [
          {
            id: 'enable_ai_assistant',
            label: SITE_SETTING_LABELS['enable_ai_assistant'],
            type: 'checkbox',
            tooltipPlacement: 'bottom',
            popoverText: AI_TERM(tooltipPhrase),
            warningText: {
              visible: (formState: State['form']) => {
                if (!formState['enable_ai_assistant']) {
                  return false;
                }

                const embeddingsSettings = formState[
                  'ai_embeddings_settings'
                ] as string;
                const completionSettings = formState[
                  'ai_completion_settings'
                ] as string;

                if (!embeddingsSettings || !completionSettings) {
                  return true;
                }

                let embeddings;
                let completion;

                try {
                  embeddings = JSON.parse(embeddingsSettings);
                  completion = JSON.parse(completionSettings);

                  return (
                    !Object.keys(embeddings.profileSettings).length ||
                    !Object.keys(completion.profileSettings).length
                  );
                } catch (error) {
                  throw new Error('Error parsing AI settings');
                }
              },
              msg: missingProfilesWarning,
            },
            confirm: (formState: State['form']) =>
              !formState['enable_ai_assistant']
                ? {
                    title: 'Action Required',
                    text: AI_TERM(modalPhrase),
                    dispatch: {
                      type: 'UPDATE_MULTIPLE_FIELDS',
                      payload: {
                        enable_ai_assistant: true,
                        ai_term_agreed: true,
                        ai_term_agreed_date: moment().format(
                          'YYYY-MM-DD HH:mm:ss'
                        ),
                        ai_term_username: authentication.username,
                      },
                    },
                  }
                : undefined,
          },
          {
            id: 'redact_ai_assistant_logs',
            label: SITE_SETTING_LABELS['redact_ai_assistant_logs'],
            type: 'checkbox',
            disabled: (formState: State['form']) =>
              !formState['enable_ai_assistant'],
            disabledText: aiNotEnabledText,
            popoverText:
              'If enabled audit logs will be redacted to omit information sent over the network by the AI visual.',
          },
        ],
      },
      {
        tab: 'Embeddings',
        settings: [],
      },
      {
        tab: 'Completion',
        settings: [],
      },
    ],
  },
  {
    tab: 'Security and Authentication',
    settings: [
      {
        id: 'enable_user_create',
        label: SITE_SETTING_LABELS['enable_user_create'],
        type: 'checkbox',
      },
      {
        id: 'enable_remote_user_auth',
        label: SITE_SETTING_LABELS['enable_remote_user_auth'],
        type: 'checkbox',
      },
      {
        id: 'min_pw_length',
        label: SITE_SETTING_LABELS['min_pw_length'],
        type: 'numeric',
      },
      {
        id: 'pw_require_number',
        label: SITE_SETTING_LABELS['pw_require_number'],
        type: 'checkbox',
      },
      {
        id: 'pw_require_symbol',
        label: SITE_SETTING_LABELS['pw_require_symbol'],
        type: 'checkbox',
      },
      {
        id: 'default_group_membership_for_new_users',
        label: SITE_SETTING_LABELS['default_group_membership_for_new_users'],
        type: 'input',
        popoverText:
          'When users log into the system for the first time, they get added into the groups specified here. Multiple groups can be specified, separated by commas.',
      },
      {
        id: 'login_instructions',
        label: SITE_SETTING_LABELS['login_instructions'],
        type: 'textarea',
      },
    ],
  },
  {
    tab: 'Debugging and Logs',
    settings: [
      {
        id: 'inspect_settings_enabled',
        label: SITE_SETTING_LABELS['inspect_settings_enabled'],
        type: 'checkbox',
      },
      {
        id: 'jserrors_enabled',
        label: SITE_SETTING_LABELS['jserrors_enabled'],
        type: 'checkbox',
      },
      {
        id: 'page_profile_enabled',
        label: SITE_SETTING_LABELS['page_profile_enabled'],
        type: 'checkbox',
      },
      {
        id: 'page_profile_appview_enabled',
        label: SITE_SETTING_LABELS['page_profile_appview_enabled'],
        type: 'checkbox',
      },
      {
        id: 'bulk_max_entries',
        label: SITE_SETTING_LABELS['bulk_max_entries'],
        type: 'numeric',
        infoText: 'The number of job logs kept after bulk process.',
      },
      {
        id: 'bulk_max_interval',
        label: SITE_SETTING_LABELS['bulk_max_interval'],
        type: 'numeric',
        min: 0.25,
        step: 0.25,
        infoText: 'The number of days between log deletion checks.',
      },
      {
        id: 'bulk_max_retention',
        label: SITE_SETTING_LABELS['bulk_max_retention'],
        type: 'numeric',
        min: 0.25,
        step: 0.25,
        infoText: 'The maximum lifetime for the job logs in days.',
      },
    ],
  },
  {
    tab: 'Version Control',
    settings: [
      {
        id: 'enable_version_control',
        label: SITE_SETTING_LABELS['enable_version_control'],
        type: 'checkbox',
      },
      {
        id: 'enable_dataset_version_control',
        label: SITE_SETTING_LABELS['enable_dataset_version_control'],
        type: 'checkbox',
        disabled: (formState) => !formState['enable_version_control'],
        disabledText: `This is currently disabled due to the "${SITE_SETTING_LABELS['enable_version_control']}" setting`,
      },
      {
        id: 'unnamed_versions_deletion_interval',
        label: SITE_SETTING_LABELS['unnamed_versions_deletion_interval'],
        type: 'numeric',
        min: 0.25,
        step: 0.25,
        infoText:
          'The number of days between unnamed versions deletion checks.',
      },
      {
        id: 'unnamed_versions_max_retention',
        label: SITE_SETTING_LABELS['unnamed_versions_max_retention'],
        type: 'numeric',
        min: 0.25,
        step: 0.25,
        infoText:
          'The maximum lifetime for an unnamed dataset versions in days.',
      },
    ],
  },
  {
    tab: 'Help and Support',
    settings: [
      {
        id: 'doc_link_text',
        label: SITE_SETTING_LABELS['doc_link_text'],
        type: 'input',
      },
      {
        id: 'doc_link_url',
        label: SITE_SETTING_LABELS['doc_link_url'],
        type: 'input',
      },
      {
        id: 'support_option',
        label: SITE_SETTING_LABELS['support_option'],
        type: 'radio',
        options: [
          { id: 'web_page', label: 'web page link' },
          { id: 'email', label: 'email address' },
        ],
      },
      {
        id: 'support_web_page_link_text',
        label: SITE_SETTING_LABELS['support_web_page_link_text'],
        type: 'input',
      },
      {
        id: 'support_web_page_link_url',
        label: SITE_SETTING_LABELS['support_web_page_link_url'],
        type: 'input',
      },
      {
        id: 'support_email',
        label: SITE_SETTING_LABELS['support_email'],
        type: 'input',
      },
    ],
  },
  {
    tab: 'Advanced Settings',
    settings: [
      {
        id: 'jira_enabled',
        label: SITE_SETTING_LABELS['jira_enabled'],
        type: 'checkbox',
        hidden: () => !settings.enableJiraSettings,
      },
      {
        id: 'jira_server',
        label: SITE_SETTING_LABELS['jira_server'],
        type: 'input',
        hidden: () => !settings.enableJiraSettings,
      },
      {
        id: 'jira_user',
        label: SITE_SETTING_LABELS['jira_user'],
        type: 'input',
        hidden: () => !settings.enableJiraSettings,
      },
      {
        id: 'jira_password',
        label: SITE_SETTING_LABELS['jira_password'],
        type: 'input',
        hidden: () => !settings.enableJiraSettings,
      },
      {
        id: 'jira_newissue_defaults',
        label: SITE_SETTING_LABELS['jira_newissue_defaults'],
        type: 'textarea',
        hidden: () => !settings.enableJiraSettings,
      },
      {
        id: 'webhook_enabled',
        label: SITE_SETTING_LABELS['webhook_enabled'],
        type: 'checkbox',
        hidden: () => !settings.enableWebHook,
      },
      {
        id: 'webhook_whitelist',
        label: SITE_SETTING_LABELS['webhook_whitelist'],
        type: 'textarea',
        infoText: 'Multiple values should be separated by a new line.',
        hidden: () => !settings.enableWebHook,
      },
      {
        id: 'remotedata_url_whitelist',
        label: SITE_SETTING_LABELS['remotedata_url_whitelist'],
        type: 'textarea',
        hidden: () => !settings.enableVisualAction,
        infoText:
          'The url used in cviz_rest functions need be be added here. Multiple values should be separated by a new line.',
      },
      {
        id: 'advanced_settings',
        label: SITE_SETTING_LABELS['advanced_settings'],
        type: 'textarea',
        hidden: () => !settings.enableAdvanceSettings,
      },
    ],
  },
];

export const FLIPPER_SETTINGS_MAP: { [id: string]: boolean } = {};

arc.frontend.featureFlipper.getAll().forEach(({ id, value, level }) => {
  if (level === 'DB') {
    FLIPPER_SETTINGS_MAP[id] = value;
  }
});
