import React from 'react';
import {
  DEFAULT_SETTING_GROUPS,
  DEFAULT_SITE_SETTING_VALUES,
} from '../constants/setting-constants';
import { SettingGroup, SiteSetting, State } from '../types/site-settings.type';
import {
  AiCmlAuthMode,
  AiEngine,
  AiProfileCols,
  AiSetting,
  AiTab,
  ProfileOption,
  ProfileSettings,
} from '../types/ai-settings.type';
import classes from '../components/AiProfileTable.module.scss';
import { createColumnHelper } from '@tanstack/react-table';

export const findSettings = (tab: string): SiteSetting[] => {
  let found: SiteSetting[] = [];

  const find = (items: SettingGroup[], tab: string) => {
    for (const group of items) {
      if (found.length) return;

      if (group.children) {
        find(group.children, tab);
      } else if (group.settings && group.tab === tab) {
        found = group.settings;
      }
    }
  };

  find(DEFAULT_SETTING_GROUPS, tab);

  return found;
};

export const getDefaultSettingValues = (
  settings: AiSetting[]
): State['form'] => {
  const state: State['form'] = {};

  for (const { id } of settings) {
    state[id] = DEFAULT_SITE_SETTING_VALUES[id];
  }

  return state;
};

export const getEngineLabel = (value: AiEngine, tab: AiTab) => {
  switch (value) {
    case 'open_ai':
      return 'OpenAI';
    case 'caii':
      return 'Cloudera AI Inference';
    case 'azure_ai':
      return 'OpenAI Azure';
    case 'llama':
      return tab === 'embeddings'
        ? ' Cloudera AI Hosted MiniLM'
        : 'Cloudera AI Hosted Llama';
    case 'bedrock':
      return 'Amazon Bedrock';
    case 'other':
      return 'Other';
    default:
      return '';
  }
};

export const getCmlAuthModeLabel = (value: AiCmlAuthMode) => {
  switch (value) {
    case 'jwt':
      return 'JWT';
    case 'api_key':
      return 'API Key';
    default:
      return '';
  }
};

const columnHelper = createColumnHelper<AiProfileCols>();

export const generateColumns = (
  handleDelete: (profile: ProfileOption) => void,
  handleEdit: (profile: ProfileOption) => void,
  handleClone: (profile: ProfileOption) => void,
  makeItDefault: (profile: ProfileOption) => void,
  disabled: boolean,
  tab: AiTab,
  defaultProfValue?: string
) => {
  return [
    columnHelper.accessor('label', {
      header: () => 'Name',
      cell: (info) => info.getValue(),
    }),
    columnHelper.accessor('engine', {
      header: () => 'Engine',
      cell: (info) => getEngineLabel(info.getValue(), tab),
    }),
    columnHelper.display({
      id: 'actions',
      cell: (info) => {
        const { profile, label } = info.row.original;
        const profileOption = { value: profile, label };
        const defProf = defaultProfValue === profile;

        return (
          <div
            className={`${classes.actionsCell} ${
              disabled ? classes.disabled : ''
            }`}
          >
            <span
              className={`${classes.setDefaultAiProfile} ${classes.onHover} ${
                defProf ? classes.keepVisible : ''
              }`}
              onClick={defProf ? undefined : () => makeItDefault(profileOption)}
            >
              {defProf ? 'Default' : 'Set as default'}
            </span>
            <span
              className={`fa fa-clone ${classes.cloneProfile}`}
              onClick={disabled ? undefined : () => handleClone(profileOption)}
            ></span>
            <span
              className="fa fa-trash"
              onClick={disabled ? undefined : () => handleDelete(profileOption)}
            ></span>
            <span
              className="fa fa-pencil"
              onClick={disabled ? undefined : () => handleEdit(profileOption)}
            ></span>
          </div>
        );
      },
    }),
  ];
};

export const convertProfilesToArray = (
  settings: ProfileSettings,
  tab: AiTab
) => {
  return Object.keys(settings).map((key) => ({
    profile: key,
    label: settings[key].label,
    engine: settings[key].values[`${tab}_engine`] as AiEngine,
  }));
};
