import React, { useMemo, useState } from 'react';
import classes from './VisualsPage.module.scss';
import { NavigationBar } from 'components/navigation-bar/NavigationBar';
import { getGlobalArc } from 'utils/get-global-arc';
import { SideBar } from './components/side-bar/SideBar';
import { useReports } from './hooks/use-reports';
import { useReportCreationPermissions } from 'hooks/use-dataset-permissions';
import { useSelectedGroup } from './hooks/use-selected-group';
import { useApps } from './hooks/use-apps';
import { CombinedReport } from 'types/combined-report.type';
import {
  Workspace,
  getWorkspaces,
  workspacesQueryToken,
} from './apis/get-workspaces';
import { WorkspaceDialog } from './components/workspace-dialog/WorkspaceDialog';
import { DeleteWorkspaceDialog } from './components/delete-workspace-dialog/DeleteWorkspaceDialog';
import classNames from 'classnames';
import { GridView } from './components/grid-view/GridView';
import { ExportDialog } from './components/export-dialog/ExportDialog';
import { DeleteReportsDialog } from './components/delete-reports-dialog/DeleteReportsDialog';
import { MoveToWorkspaceDialog } from './components/move-to-workspace-dialog/MoveToWorkspaceDialog';
import { ListView } from './components/list-view/ListView';
import DashboardIcon from '@cloudera-internal/cuix-core/icons/svg/ic-grid-view.svg';
import AppIcon from '@cloudera-internal/cuix-core/icons/svg/ic-profiler-metrics.svg';
import { getDashboardSharedVisuals } from './utils/get-dashboard-shared-visuals';
import { useQuery } from '@tanstack/react-query';

import { useSelectedReportsStore } from './store/selected-reports.store';
import { useSelectedViewMode } from './hooks/use-selected-view-mode';
import { useSelectedReportType } from './hooks/use-selected-report-type';
import { getSearchFunction } from './utils/get-search-reports-function';
import { SelectedGroup } from './components/selected-group/SelectedGroup';
import { Toolbar } from './components/toolbar/Toolbar';

const arc = getGlobalArc();

const params = new URLSearchParams(window.location.search);

export type StaticSidebarItems = {
  all: {
    id: number;
    name: string;
    desc: string;
  };
  favorites: {
    id: number;
    name: string;
    desc: string;
  };
  samples?: {
    id: number;
    name: string;
    desc: string;
  };
  public?: {
    id: number;
    name: string;
    desc: string;
    editable: boolean;
    perms: string[];
    acl: [[type: 1 | 2, permission: 1 | 2 | 3, name: string]];
  };
  private: {
    id: number;
    name: string;
    desc: string;
    editable: boolean;
    perms: string[];
    acl: [[type: 1 | 2, permission: 1 | 2 | 3, name: string]];
  };
};

type VisualsPageProps = {
  shouldShowModal: boolean;
  isAlreadySeen: boolean;
  staticSideBarItems: StaticSidebarItems;
};

export const VisualsPage = ({
  isAlreadySeen,
  shouldShowModal,
  staticSideBarItems,
}: VisualsPageProps): JSX.Element => {
  const [reportsSearchTerm, setReportsSearchTerm] = useState(
    params.get('searchbox') ?? ''
  );

  const [publicWorkspace, setPublicWorkspace] = useState(
    staticSideBarItems.public
  );

  const { canCreateDashboard, canCreateApp } = useReportCreationPermissions();

  const { selectedViewMode, setSelectedViewMode } = useSelectedViewMode();

  const { selectedReportType, setSelectedReportType } = useSelectedReportType();

  const {
    groupSelectedReports,
    groupActionType,
    resetGroupAction,

    selectedReport,
    actionType,

    unselectGroup,

    unselect,
  } = useSelectedReportsStore();

  const {
    data: groupedWorkspaces,
    isLoading: areWorkspacesLoading,
    error: workspacesError,
  } = useQuery({
    queryKey: [workspacesQueryToken],
    queryFn: getWorkspaces,
    select: (workspaces) => {
      const shareds: typeof workspaces = [];
      const privates: typeof workspaces = [];

      workspaces.forEach((workspace) => {
        if (
          workspace.id === staticSideBarItems.private.id ||
          workspace.id === publicWorkspace?.id
        )
          return;

        if (workspace.editable) {
          shareds.push(workspace);
        } else {
          privates.push(workspace);
        }
      });

      return {
        sharedWorkspaces: shareds.sort((a, b) => a.name.localeCompare(b.name)),
        privateWorkspaces: privates.sort((a, b) =>
          a.name.localeCompare(b.name)
        ),
      };
    },
  });

  const { favoriteReportIds, appsAsCombinedReports, samplesReportIds } =
    useApps();

  const { selectedGroup, setSelectedGroup, toggleFavoriteGroup } =
    useSelectedGroup({
      staticSideBarItems,
      groupedWorkspaces,
      publicWorkspace,
    });

  const {
    areReportsLoading,
    relatedReportIdToReportMap,
    groupIdToReportTypeToCombinedReports,
    hasReports,
  } = useReports({
    selectedGroup,
    appsAsCombinedReports,
    favoriteReportIds,
    samplesReportIds,
    staticSideBarItems,
  });

  const visibleReports = useMemo(() => {
    if (selectedGroup === null) {
      return [];
    }

    let reports: CombinedReport[] = [];
    // Filter by group and report type
    if (selectedGroup.id === staticSideBarItems.all.id) {
      reports =
        groupIdToReportTypeToCombinedReports
          .get('all')
          ?.get(selectedReportType) ?? [];
    } else if (selectedGroup.id === staticSideBarItems.samples?.id) {
      reports =
        groupIdToReportTypeToCombinedReports
          .get('samples')
          ?.get(selectedReportType) ?? [];
    } else if (selectedGroup.id === staticSideBarItems.favorites.id) {
      reports =
        groupIdToReportTypeToCombinedReports
          .get('favorites')
          ?.get(selectedReportType) ?? [];

      reports = reports.filter(({ id, permissions }) => {
        const isFavoriteReport = favoriteReportIds.has(id);

        return isFavoriteReport && permissions.includes('ds_appview');
      });
    } else {
      reports =
        groupIdToReportTypeToCombinedReports
          .get(selectedGroup.id)
          ?.get(selectedReportType) ?? [];
    }

    // Filter by search
    if (reportsSearchTerm) {
      const searchFunction = getSearchFunction(reportsSearchTerm);

      reports = reports
        .map(searchFunction)
        .filter(
          (report): report is { obj: CombinedReport; priority: number } =>
            report !== null
        )
        .sort((a, b) => a.priority - b.priority)
        .map(({ obj }) => obj);
    }

    return reports;
  }, [
    favoriteReportIds,
    groupIdToReportTypeToCombinedReports,
    reportsSearchTerm,
    selectedGroup,
    selectedReportType,
    staticSideBarItems.all.id,
    staticSideBarItems.favorites.id,
    staticSideBarItems.samples?.id,
  ]);

  const [selectedWorkspaceToEdit, setSelectedWorkspaceToEdit] = useState<{
    id: number;
    name: string;
    desc: string;
    acl?: Workspace['acl'];
    perms?: string[];
  } | null>(null);

  const [selectedWorkspaceToDelete, setSelectedWorkspaceToDelete] = useState<{
    id: number;
    name: string;
  } | null>(null);

  const workspaces = useMemo(() => {
    if (!groupedWorkspaces) {
      return null;
    }

    const collectedWorkspaces: { id: number; name: string }[] = [
      ...groupedWorkspaces.privateWorkspaces,
      ...groupedWorkspaces.sharedWorkspaces,
      staticSideBarItems.private,
    ];

    if (publicWorkspace) {
      collectedWorkspaces.push(publicWorkspace);
    }

    return collectedWorkspaces;
  }, [groupedWorkspaces, staticSideBarItems.private, publicWorkspace]);

  return (
    <div className={classes.container}>
      <NavigationBar
        isAlreadySeen={isAlreadySeen}
        shouldShowModal={shouldShowModal}
        activeTabs={{ main: 'visuals' }}
      />

      <div className={classes.content}>
        {!arc.frontend.featureFlipper.getValueById('hide_workspaces') && (
          <SideBar
            staticSideBarItems={staticSideBarItems}
            selectedGroupId={selectedGroup?.id}
            setSelectedGroup={setSelectedGroup}
            groupedWorkspaces={groupedWorkspaces}
            workspaces={workspaces}
            areWorkspacesLoading={areWorkspacesLoading}
            hasWorkspacesError={!!workspacesError}
            setSelectedWorkspaceToEdit={setSelectedWorkspaceToEdit}
            publicWorkspace={publicWorkspace}
            setPublicWorkspace={setPublicWorkspace}
          />
        )}

        <div className={classes.main}>
          {!arc.frontend.featureFlipper.getValueById('hide_workspaces') && (
            <SelectedGroup
              areWorkspacesLoading={areWorkspacesLoading}
              selectedGroup={selectedGroup}
              staticSideBarItems={staticSideBarItems}
              hasReports={hasReports}
              setSelectedWorkspaceToDelete={setSelectedWorkspaceToDelete}
              setSelectedWorkspaceToEdit={setSelectedWorkspaceToEdit}
            />
          )}

          {arc.frontend.featureFlipper.getValueById('enable_explore_style') && (
            <span className={classes.title}>Visuals</span>
          )}

          <div className={classes.tabs}>
            <div
              className={classNames(classes.tab, {
                [classes.active]: selectedReportType === 'all',
              })}
              onClick={() => setSelectedReportType('all')}
              data-test-id="app-grid-all-btn"
            >
              All
            </div>

            <div
              className={classNames(classes.tab, {
                [classes.active]: selectedReportType === 'dashboards',
              })}
              onClick={() => setSelectedReportType('dashboards')}
              data-test-id="grid-view"
            >
              <span className="text-lg">
                <DashboardIcon />
              </span>
              <span className="ml-1">Dashboards</span>
            </div>

            <div
              className={classNames(classes.tab, {
                [classes.active]: selectedReportType === 'apps',
              })}
              onClick={() => setSelectedReportType('apps')}
              data-test-id="filter-for-apps-button"
            >
              <span className="text-lg">
                <AppIcon />
              </span>
              <span className="ml-1">Apps</span>
            </div>
          </div>

          <Toolbar
            reportsSearchTerm={reportsSearchTerm}
            setReportsSearchTerm={setReportsSearchTerm}
            canCreateDashboard={canCreateDashboard}
            canCreateApp={canCreateApp}
            selectedGroup={selectedGroup}
            favoritesGroupId={staticSideBarItems.favorites.id}
            toggleFavoriteGroup={toggleFavoriteGroup}
            selectedViewMode={selectedViewMode}
            setSelectedViewMode={setSelectedViewMode}
            areWorkspacesLoading={areWorkspacesLoading}
          />

          {selectedViewMode === 'grid' && (
            <GridView
              areReportsLoading={areReportsLoading}
              reports={visibleReports}
              selectedReportType={selectedReportType}
              searchTerm={reportsSearchTerm}
            />
          )}

          {selectedViewMode === 'list' && (
            <ListView
              relatedReportIdToReportMap={relatedReportIdToReportMap}
              areReportsLoading={areReportsLoading}
              reports={visibleReports}
              selectedReportType={selectedReportType}
              searchTerm={reportsSearchTerm}
            />
          )}
        </div>
      </div>

      {selectedWorkspaceToEdit?.perms?.includes('ws_manage') && workspaces && (
        <WorkspaceDialog
          isOpen
          close={() => setSelectedWorkspaceToEdit(null)}
          workspaces={workspaces}
          selectedWorkspace={selectedWorkspaceToEdit}
          publicWorkspace={publicWorkspace}
          setPublicWorkspace={setPublicWorkspace}
        />
      )}

      {selectedWorkspaceToDelete && (
        <DeleteWorkspaceDialog
          selected={selectedWorkspaceToDelete}
          close={() => setSelectedWorkspaceToDelete(null)}
        />
      )}

      {actionType === 'export' && selectedReport && (
        <ExportDialog
          close={unselect}
          isOpen={actionType === 'export' && !!selectedReport}
          selectedReports={[selectedReport]}
        />
      )}

      {groupActionType === 'export' && groupSelectedReports && (
        <ExportDialog
          close={resetGroupAction}
          isOpen={groupActionType === 'export' && !!groupSelectedReports}
          selectedReports={groupSelectedReports}
        />
      )}

      {actionType === 'delete' && selectedReport && (
        <DeleteReportsDialog
          selectedReports={[selectedReport]}
          onSuccess={() => {
            unselectGroup(selectedReport);
            unselect();
          }}
          close={unselect}
          isOpen={actionType === 'delete' && !!selectedReport}
          selectedGroup={selectedGroup}
          favoriteReportIds={favoriteReportIds}
        />
      )}

      {groupActionType === 'delete' && groupSelectedReports && (
        <DeleteReportsDialog
          selectedReports={groupSelectedReports}
          onSuccess={(removedIds) =>
            removedIds.forEach(({ id, type }) => unselectGroup({ id, type }))
          }
          close={resetGroupAction}
          isOpen={groupActionType === 'delete' && !!groupSelectedReports}
          selectedGroup={selectedGroup}
          favoriteReportIds={favoriteReportIds}
        />
      )}

      {!arc.frontend.featureFlipper.getValueById('hide_workspaces') && (
        <>
          {actionType === 'move' && selectedReport && (
            <MoveToWorkspaceDialog
              isOpen={actionType === 'move' && !!selectedReport}
              close={unselect}
              relatedReportIdToReportMap={relatedReportIdToReportMap}
              selectedGroup={selectedGroup}
              selectedLinkedVisuals={
                selectedReport.type === 'dashboard'
                  ? getDashboardSharedVisuals(
                      selectedReport.sharedVisualDashboardIds ?? [],
                      relatedReportIdToReportMap
                    )
                  : []
              }
              selectedReports={[selectedReport]}
              onSuccess={() => {
                unselectGroup(selectedReport);
                unselect();
              }}
              workspaces={workspaces}
            />
          )}

          {groupActionType === 'move' && groupSelectedReports && (
            <MoveToWorkspaceDialog
              isOpen={groupActionType === 'move' && !!groupSelectedReports}
              close={resetGroupAction}
              relatedReportIdToReportMap={relatedReportIdToReportMap}
              selectedGroup={selectedGroup}
              selectedLinkedVisuals={groupSelectedReports.flatMap(
                ({ type, sharedVisualDashboardIds }) =>
                  type === 'dashboard'
                    ? getDashboardSharedVisuals(
                        sharedVisualDashboardIds ?? [],
                        relatedReportIdToReportMap
                      )
                    : []
              )}
              selectedReports={groupSelectedReports}
              onSuccess={() =>
                groupSelectedReports.forEach((report) => unselectGroup(report))
              }
              workspaces={workspaces}
            />
          )}
        </>
      )}
    </div>
  );
};
