import { useMutation, useQueryClient } from '@tanstack/react-query';
import { deleteReport } from 'apis/delete-report';
import { getReportCountPerGroupQueryToken } from 'apis/get-report-count-per-group';
import { appsQueryToken } from 'apis/get-apps';
import { reportsQueryToken } from 'apis/get-reports';
import { AxiosError } from 'axios';
import { DeleteDialog } from 'components/delete-dialog/DeleteDialog';
import React, { useCallback } from 'react';
import { useNotificationStore } from 'stores/use-notification-store';
import { CombinedReport } from 'types/combined-report.type';
import classes from './DeleteReportsDialog.module.scss';
import { Table } from 'react-bootstrap';
import classNames from 'classnames';
import moment from 'moment';
import { ClouderaLink } from '../../../../components/cloudera-link/ClouderaLink';
import { getGlobalArc } from 'utils/get-global-arc';
import { getBlankTargetUrl } from 'arc/misc/misc';

const arc = getGlobalArc();

type DeleteReportsDialogProps = {
  selectedReports: CombinedReport[];
  isOpen: boolean;
  close: () => void;
  selectedGroup: { id: number; name: string } | null;
  favoriteReportIds: Set<number>;
  onSuccess: (removedReports: { id: number; type: string }[]) => void;
};

export const DeleteReportsDialog = ({
  selectedReports,
  selectedGroup,
  favoriteReportIds,
  onSuccess,
  isOpen,
  close,
}: DeleteReportsDialogProps): JSX.Element | null => {
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const queryClient = useQueryClient();

  const { mutateAsync, isLoading: isDeleteLoading } = useMutation({
    mutationFn: deleteReport,
    onMutate: async (report) => {
      type DashboardVisualOrApp = { id: number } | { group_id: number };

      const queryToken =
        report.type === 'app'
          ? [appsQueryToken]
          : [reportsQueryToken, selectedGroup?.id];

      await queryClient.cancelQueries({
        queryKey: queryToken,
      });

      await queryClient.cancelQueries({
        queryKey: [getReportCountPerGroupQueryToken],
      });

      const reports =
        queryClient.getQueryData<DashboardVisualOrApp[]>(queryToken);

      const reportCountPerGroup = await queryClient.getQueryData([
        getReportCountPerGroupQueryToken,
      ]);

      queryClient.setQueryData<DashboardVisualOrApp[]>(queryToken, (current) =>
        current?.filter((currentReport) => {
          if ('id' in currentReport) {
            return currentReport.id !== report.id;
          } else {
            return currentReport.group_id !== report.id;
          }
        })
      );

      queryClient.setQueryData<{
        favoritesReportCount: number | null;
        workspaceIdToReportCount: { [workspaceId: number]: number } | null;
        workspaceIdToAppCount: { [workspaceId: number]: number } | null;
      }>([getReportCountPerGroupQueryToken], (current) => {
        if (current) {
          const isApp = report.type === 'app';
          const isFavorite = favoriteReportIds.has(report.id);

          if (isApp) {
            const newWorkspaceIdToAppCount = {
              ...current.workspaceIdToAppCount,
            };

            newWorkspaceIdToAppCount[report.workspace.id] -= 1;

            return {
              favoritesReportCount:
                (current.favoritesReportCount ?? 0) - (isFavorite ? 1 : 0),
              workspaceIdToReportCount: current.workspaceIdToReportCount,
              workspaceIdToAppCount: newWorkspaceIdToAppCount,
            };
          } else {
            const newWorkspaceIdToReportCount = {
              ...current.workspaceIdToReportCount,
            };

            newWorkspaceIdToReportCount[report.workspace.id] -= 1;

            return {
              favoritesReportCount:
                (current.favoritesReportCount ?? 0) - (isFavorite ? 1 : 0),
              workspaceIdToReportCount: newWorkspaceIdToReportCount,
              workspaceIdToAppCount: current.workspaceIdToAppCount,
            };
          }
        }

        return current;
      });

      return { reports, reportCountPerGroup };
    },
    onError: (err: AxiosError, report, context) => {
      queryClient.setQueryData(
        report.type === 'app'
          ? [appsQueryToken]
          : [reportsQueryToken, selectedGroup?.id],
        context?.reports
      );

      queryClient.setQueryData(
        [getReportCountPerGroupQueryToken],
        context?.reportCountPerGroup
      );

      addNotification({
        message: `Failed to delete ${report.type}`,
        expandedMessage: JSON.stringify(
          err.response?.data ?? err.message,
          null,
          4
        ),
        variant: 'danger',
      });

      throw err;
    },
  });

  const onConfirm = useCallback(async () => {
    const results = await Promise.allSettled(
      selectedReports.map((report) => mutateAsync(report))
    );

    if (results.every((result) => result.status === 'fulfilled')) {
      close();

      addNotification(
        {
          message: `Successfully deleted reports with the following ids: ${selectedReports
            .map(({ id }) => id)
            .join(', ')}`,
          variant: 'success',
        },
        {
          delay: 2000,
        }
      );

      onSuccess(selectedReports.map(({ id, type }) => ({ id, type })));
    } else {
      const successfullRemoved: { id: number; type: string }[] = [];
      results.forEach((result) => {
        if (result.status === 'fulfilled') {
          successfullRemoved.push({
            id: result.value.id,
            type: result.value.type,
          });
        }
      });

      if (successfullRemoved.length) {
        addNotification(
          {
            message: `Successfully deleted reports with the following ids: ${successfullRemoved
              .map(({ id }) => id)
              .join(', ')}`,
            variant: 'success',
          },
          {
            delay: 2000,
          }
        );

        onSuccess(successfullRemoved);
      }
    }
  }, [addNotification, close, mutateAsync, onSuccess, selectedReports]);

  return (
    <DeleteDialog
      title="Delete Visual Artifacts"
      isOpen={isOpen}
      close={close}
      onConfirm={onConfirm}
      isLoading={isDeleteLoading}
      info={
        <div style={{ maxHeight: 400, overflow: 'scroll' }}>
          <Table
            bordered
            style={{
              borderSpacing: 0,
              borderCollapse: 'separate',
              borderRadius: 3,
              overflow: 'hidden',
            }}
          >
            <thead style={{ backgroundColor: '#fafbfc' }}>
              <tr>
                {[
                  'Name',
                  'Type',
                  'ID',
                  'Last updated',
                  'Modified by',
                  'Dataset',
                  'Workspace',
                ].map((label) => (
                  <th
                    key={label}
                    className={classNames(
                      'border-left-0 border-right-0 border-top-0',
                      classes.th
                    )}
                  >
                    {label}
                  </th>
                ))}
              </tr>
            </thead>

            <tbody>
              {selectedReports.map(
                ({ id, type, title, updated, datasetInfo, workspace }) => (
                  <tr key={`${id}-${type}`}>
                    <td
                      className={classNames(
                        'border-left-0 border-right-0 border-top-0 align-middle',
                        classes.td
                      )}
                    >
                      {title}
                    </td>

                    <td
                      className={classNames(
                        'border-left-0 border-right-0 border-top-0 align-middle',
                        classes.td
                      )}
                    >
                      {type !== 'app' && type !== 'dashboard' ? 'visual' : type}
                    </td>

                    <td
                      className={classNames(
                        'border-left-0 border-right-0 border-top-0 align-middle',
                        classes.td
                      )}
                    >
                      {id}
                    </td>

                    <td
                      className={classNames(
                        'border-left-0 border-right-0 border-top-0 align-middle',
                        classes.td
                      )}
                    >
                      {moment(updated.date).fromNow()}
                    </td>

                    <td
                      className={classNames(
                        'border-left-0 border-right-0 border-top-0 align-middle',
                        classes.td
                      )}
                    >
                      {updated.by}
                    </td>

                    <td
                      className={classNames(
                        'border-left-0 border-right-0 border-top-0 align-middle',
                        classes.td
                      )}
                    >
                      {datasetInfo ? (
                        <ClouderaLink
                          href={getBlankTargetUrl(
                            `${arc.frontend.globals.urls.datasetEditUrl}/${datasetInfo.id}`
                          )}
                          target="_blank"
                        >
                          {datasetInfo.name} - {datasetInfo.id}
                        </ClouderaLink>
                      ) : (
                        'N/A'
                      )}
                    </td>

                    <td
                      className={classNames(
                        'border-left-0 border-right-0 border-top-0 align-middle',
                        classes.td
                      )}
                    >
                      {workspace.name}
                    </td>
                  </tr>
                )
              )}
            </tbody>
          </Table>
        </div>
      }
    />
  );
};
