import { useMutation, useQueryClient } from '@tanstack/react-query';
import { AxiosError } from 'axios';
import { DeleteDialog } from 'components/delete-dialog/DeleteDialog';
import { deleteWorkspace } from 'pages/visuals-page/apis/delete-workspace';
import {
  Workspace,
  workspacesQueryToken,
} from 'pages/visuals-page/apis/get-workspaces';
import React, { useCallback } from 'react';
import { useNotificationStore } from 'stores/use-notification-store';
import classes from './DeleteWorkspaceDialog.module.scss';

type DeleteWorkspaceDialogProps = {
  selected: { id: number; name: string };
  close: () => void;
};

export const DeleteWorkspaceDialog = ({
  selected,
  close,
}: DeleteWorkspaceDialogProps): JSX.Element => {
  const queryClient = useQueryClient();

  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const { mutateAsync, isLoading: isDeleteLoading } = useMutation({
    mutationFn: deleteWorkspace,
    onMutate: async ({ id }) => {
      await queryClient.cancelQueries({
        queryKey: [workspacesQueryToken],
      });

      const workspaces = queryClient.getQueryData<Workspace[]>([
        workspacesQueryToken,
      ]);

      queryClient.setQueryData<Workspace[]>([workspacesQueryToken], (current) =>
        current?.filter((currentWorkspace) => {
          return currentWorkspace.id !== id;
        })
      );

      return { workspaces };
    },
    onError: (err: AxiosError, workspace, context) => {
      queryClient.setQueryData([workspacesQueryToken], context?.workspaces);

      addNotification({
        message: `Failed to delete ${workspace.name}\n${JSON.stringify(
          err.response?.data ?? err.message,
          null,
          4
        )}`,
        variant: 'danger',
      });
    },
  });

  const onConfirm = useCallback(async () => {
    if (selected) {
      await mutateAsync(selected);

      addNotification(
        {
          message: `Successfully deleted workspace "${selected.name}"`,
          variant: 'success',
        },
        {
          delay: 2000,
        }
      );

      close();
    }
  }, [addNotification, close, mutateAsync, selected]);

  return (
    <DeleteDialog
      title="Delete Workspace"
      isOpen={!!selected}
      close={close}
      onConfirm={onConfirm}
      isLoading={isDeleteLoading}
      info={
        <div className={classes.infoContainer}>
          <div>
            <p className={classes.label}>WORKSPACE NAME</p>
            <p className={classes.value}>{selected.name}</p>
          </div>

          <div>
            <p className={classes.label}>ID</p>
            <p className={classes.value}>{selected.id}</p>
          </div>
        </div>
      }
    />
  );
};
