import {
  ExportInfo,
  exportVisualArtifacts,
  getExportInfoQueryToken,
} from 'pages/visuals-page/apis/export-visual-artifacts';
import React, { useMemo, useState } from 'react';
import { Alert, Form, Modal, Table } from 'react-bootstrap';
import { getGlobalArc } from 'utils/get-global-arc';
import { ClouderaButton } from '../../../../components/cloudera-button/ClouderaButton';
import classNames from 'classnames';
import classes from './ExportDialog.module.scss';
import Close from '@cloudera-internal/cuix-core/icons/svg/ic-close.svg';
import Upload from '@cloudera-internal/cuix-core/icons/svg/ic-import.svg';
import { useQuery } from '@tanstack/react-query';
import { CombinedReport } from 'types/combined-report.type';
import { AxiosError } from 'axios';

const arc = getGlobalArc();

type ExportDialogProps = {
  isOpen: boolean;
  selectedReports: CombinedReport[];
  close: () => void;
};

const artifactKeyOrder: (keyof ExportInfo)[] = [
  'appgroups',
  'dashboards',
  'visuals',
  'datasets',
  'customcss',
  'colorpalette',
  'staticasset',
];

const iconAndLabelByType = {
  appgroups: { icon: 'sym-app', label: 'App' },
  dashboards: { icon: 'sym-dashboard', label: 'Dashboard' },
  visuals: { icon: 'sym-visual', label: 'Visual' },
  datasets: { icon: 'fa-cubes', label: 'Dataset' },
  customcss: { icon: 'fa-leaf', label: 'Custom Styles' },
  colorpalette: { icon: 'fa-paint-brush', label: 'Custom Color' },
  staticasset: { icon: 'fa-file-image-o', label: 'Static Asset' },
};

export const ExportDialog = ({
  isOpen,
  selectedReports,
  close,
}: ExportDialogProps): JSX.Element => {
  const [fileName, setFileName] = useState(`visuals_${new Date().getTime()}`);

  const groupedSelectedReportIds = useMemo(
    () =>
      selectedReports.reduce(
        (result, { id, type }) => {
          if (type === 'app') {
            result.appIds.push(id);
          } else if (type === 'dashboard') {
            result.dashboardIds.push(id);
          } else {
            result.visualIds.push(id);
          }

          return result;
        },
        {
          visualIds: [] as number[],
          dashboardIds: [] as number[],
          appIds: [] as number[],
        }
      ),
    [selectedReports]
  );

  const {
    data: exportInfo,
    isFetching,
    error,
  } = useQuery({
    queryKey: [getExportInfoQueryToken, groupedSelectedReportIds],
    queryFn: () =>
      exportVisualArtifacts({
        visualIds: groupedSelectedReportIds.visualIds,
        dashboardIds: groupedSelectedReportIds.dashboardIds,
        appIds: groupedSelectedReportIds.appIds,
        filename: 'temp',
        isDryRun: true,
      }),
    select: ({ visuals, ...data }) => ({
      ...data,
      visuals: visuals.filter(
        ({ type }) => !['picklist', 'daterange'].includes(type)
      ),
    }),
  });

  const onSubmit = () => {
    const url = `${
      arc.frontend.globals.urls.exportUrl
    }?dry_run=false&dashboards=[${groupedSelectedReportIds.dashboardIds.join(
      ','
    )}]&appgroups=[${groupedSelectedReportIds.appIds.join(
      ','
    )}]&visuals=[${groupedSelectedReportIds.visualIds.join(
      ','
    )}]&filename=${fileName}`;

    const link = document.createElement('a');

    link.download = `${fileName}.json`;

    link.href = url;

    link.dispatchEvent(new MouseEvent('click'));

    close();
  };

  return (
    <Modal show={isOpen} onHide={close}>
      <Modal.Header className="align-items-center border-bottom-0">
        <h3
          className="text-lg font-light text-dark m-0"
          style={{
            height: 28,
            lineHeight: 1.4,
          }}
        >
          Export Visual Artifact
        </h3>

        <button
          className="outline-none bg-transparent border-0 p-2 text-dark-gray text-lg"
          onClick={close}
        >
          <Close />
        </button>
      </Modal.Header>

      <Modal.Body>
        <Form.Group controlId="filename">
          <Form.Label>
            <span className="d-block text-right text-dark-gray">Filename</span>
          </Form.Label>

          <Form.Control
            placeholder="enter filename"
            value={fileName}
            onChange={(event) => {
              setFileName(event.target.value);
            }}
            data-test-id="export-filename"
          />
        </Form.Group>

        <div style={{ maxHeight: 400, overflow: 'scroll' }}>
          <Table
            bordered
            style={{
              borderSpacing: 0,
              borderCollapse: 'separate',
              borderRadius: 3,
              overflow: 'hidden',
            }}
          >
            <thead style={{ backgroundColor: '#fafbfc' }}>
              <tr>
                <th
                  className={classNames(
                    'border-left-0 border-right-0 border-top-0',
                    classes.th
                  )}
                >
                  Object Type
                </th>

                <th
                  className={classNames(
                    'border-left-0 border-right-0 border-top-0',
                    classes.th
                  )}
                >
                  Object Name
                </th>

                <th
                  className={classNames(
                    'border-left-0 border-right-0 border-top-0',
                    classes.th
                  )}
                >
                  ID
                </th>
              </tr>
            </thead>

            <tbody>
              {isFetching && (
                <tr>
                  <td
                    colSpan={3}
                    className={classNames(
                      'border-left-0 border-right-0 border-top-0 align-middle',
                      classes.td,
                      classes.loaderContainer
                    )}
                  >
                    <h4 className="d-flex justify-content-center m-0">
                      <div className={classes.loader}></div>
                    </h4>
                  </td>
                </tr>
              )}

              {exportInfo &&
                artifactKeyOrder.map((key) => {
                  const resources = exportInfo[key];

                  return resources.map(({ id, name }) => {
                    return (
                      <tr key={id}>
                        <td
                          className={classNames(
                            'border-left-0 border-right-0 border-top-0 align-middle',
                            classes.td
                          )}
                        >
                          <i
                            className={`fa ${iconAndLabelByType[key].icon} mr-2`}
                          />
                          {iconAndLabelByType[key].label}
                        </td>

                        <td
                          className={classNames(
                            'border-left-0 border-right-0 border-top-0 align-middle',
                            classes.td
                          )}
                        >
                          {name}
                        </td>

                        <td
                          className={classNames(
                            'border-left-0 border-right-0 border-top-0 align-middle',
                            classes.td
                          )}
                        >
                          {id}
                        </td>
                      </tr>
                    );
                  });
                })}
            </tbody>
          </Table>
        </div>

        {error instanceof AxiosError && !isFetching && (
          <Alert variant="danger" className="w-full">
            {error.message}
          </Alert>
        )}
      </Modal.Body>

      <Modal.Footer className="justify-content-end">
        <ClouderaButton onClick={close} testId="export-close">
          Cancel
        </ClouderaButton>

        <ClouderaButton
          color="primary"
          onClick={onSubmit}
          testId="export-submit"
          disabled={!fileName}
          icon={<Upload />}
        >
          Export
        </ClouderaButton>
      </Modal.Footer>
    </Modal>
  );
};
