import React from 'react';
import classNames from 'classnames';
import { InView } from 'react-intersection-observer';
import Upload from '@cloudera-internal/cuix-core/icons/svg/ic-import.svg';
import Move from '@cloudera-internal/cuix-core/icons/svg/ic-get-started.svg';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';

import { useSelectedReportsStore } from 'pages/visuals-page/store/selected-reports.store';
import { CombinedReport } from 'types/combined-report.type';
import { ReportFilterType } from 'types/report-filter-type.type';

import classes from './GridView.module.scss';
import { VizWrapperLink } from 'components/viz-wrapper-link/VizWrapperLink';
import { Dropdown } from 'react-bootstrap';
import { ReportTilePopover } from '../report-tile-popover/ReportTilePopover';
import { ReportTileImage } from 'components/report-tile-image/ReportTileImage';
import { getGlobalArc } from 'utils/get-global-arc';

const arc = getGlobalArc();

type GridViewProps = {
  reports: CombinedReport[];
  areReportsLoading: boolean;
  selectedReportType: ReportFilterType;
  searchTerm: string;
};

export const GridView = ({
  reports,
  areReportsLoading,
  selectedReportType,
  searchTerm,
}: GridViewProps): JSX.Element => {
  const {
    isReportSelected,
    getNumberOfGroupSelectedReports,

    toggleReport,

    selectToDelete,
    selectToExport,
    selectToMove,
  } = useSelectedReportsStore();

  return (
    <div
      className="d-flex align-items-center p-0 m-0 flex-wrap"
      style={{ gap: 15, overflowY: 'auto' }}
    >
      {!areReportsLoading && reports.length === 0 && (
        <div className="d-flex w-100 justify-content-center">
          <h4>
            There are no{' '}
            {selectedReportType === 'all'
              ? 'reports'
              : selectedReportType === 'dashboards'
              ? 'dashboards'
              : 'apps'}{' '}
            {searchTerm ? `matching "${searchTerm}" ` : ''}
            in this workspace
          </h4>
        </div>
      )}

      {areReportsLoading && (
        <div
          className={classNames(classes.tile)}
          data-test-id="reports-loading-indicator"
        >
          <div className={classes.loader}></div>
        </div>
      )}

      {reports
        .sort((a, b) => {
          if (a.updated.date > b.updated.date) {
            return -1;
          }

          if (a.updated.date < b.updated.date) {
            return 1;
          }

          return 0;
        })
        .map((report) => (
          <InView key={`${report.title}-${report.id}-${report.type}`}>
            {({ ref, inView }) => {
              const isApp = report.type === 'app';
              const isDashboard = report.type === 'dashboard';

              const hasEditPermission =
                (report.type === 'dashboard' &&
                  report.permissions.includes('ds_appedit')) ||
                (isApp && report.rw);

              const hasDsAppEditPermission =
                report.permissions.includes('ds_appedit');

              const hasRelatedReports =
                (report.sharedVisualDashboardIds?.length ?? 0) > 0;
              const hasDeletePermission =
                (hasDsAppEditPermission &&
                  (isDashboard || !hasRelatedReports)) ||
                (isApp && report.rw);

              let reportClassName = '';
              if (isApp) {
                reportClassName = classes.appIcon;
              } else if (isDashboard) {
                reportClassName = classes.dashboardIcon;
              } else {
                reportClassName = classes.visualIcon;
              }

              const isSelectMode = getNumberOfGroupSelectedReports() > 0;
              const isSelected = isReportSelected(report);

              return (
                <div
                  ref={ref}
                  className={classes.reportCard}
                  style={{
                    width: 182,
                    height: 196,
                  }}
                  data-test-report-id={`${report.type}-${report.id}`}
                >
                  {inView && (
                    <div
                      className={classes.relativeContainer}
                      data-test-id={report.title}
                    >
                      <VizWrapperLink
                        className="d-flex flex-column h-100 position-relative"
                        href={report.url}
                        target={isApp ? '_blank' : '_self'}
                        onClick={(event) => {
                          if (isSelectMode) {
                            event.preventDefault();
                            toggleReport(report);
                          }
                        }}
                        testId="grid-item-link"
                      >
                        {isApp && (
                          <div className={classes.launchApp}>
                            <span className="fa fa-external-link" /> LaunchApp
                          </div>
                        )}

                        <div className={classes.imageContainer}>
                          <ReportTileImage
                            img={report.type === 'app' ? undefined : report.img}
                            baseImg={report.baseImg}
                            type={report.type}
                            size="md"
                          />
                        </div>

                        <div
                          className="d-flex gap-2 justify-content-center"
                          style={{ padding: '0 10px' }}
                        >
                          <ReportTilePopover
                            report={{
                              ...report,
                              workspace: report.workspace.name,
                            }}
                            datasetName={report.datasetInfo?.name}
                            sharedVisualDashboardIds={
                              report.sharedVisualDashboardIds
                            }
                          >
                            <div
                              className={classNames(
                                classes.title,
                                'text-truncate'
                              )}
                            >
                              {report.title}
                            </div>
                          </ReportTilePopover>
                        </div>
                      </VizWrapperLink>

                      <div
                        className={classNames(classes.selectIcon, {
                          [classes.active]: isSelected,
                        })}
                        onClick={() => toggleReport(report)}
                        data-test-id={
                          isSelected ? 'btn-selected' : 'btn-unselected'
                        }
                      >
                        <div className={classes.circle}>
                          <div
                            className={classNames({
                              [classes.selected]: isSelected,
                            })}
                          />
                        </div>
                      </div>

                      <Dropdown
                        style={{
                          position: 'absolute',
                          top: 0,
                          right: 0,
                        }}
                      >
                        <Dropdown.Toggle
                          as="i"
                          className={classNames(
                            'fa fa-ellipsis-v',
                            classes.moreOptionsIcon
                          )}
                          bsPrefix="no"
                          data-test-id="report-tile-more-actions-icon"
                        />

                        <Dropdown.Menu>
                          <Dropdown.Item
                            className={classes.dropdownItem}
                            onClick={() => {
                              selectToExport(report);
                            }}
                          >
                            <Upload />

                            <p>Export</p>
                          </Dropdown.Item>
                          {!arc.frontend.featureFlipper.getValueById(
                            'hide_workspaces'
                          ) && (
                            <Dropdown.Item
                              className={classes.dropdownItem}
                              disabled={false}
                              onClick={() => {
                                selectToMove(report);
                              }}
                            >
                              <Move />

                              <p>Move to Workspace</p>
                            </Dropdown.Item>
                          )}
                          <Dropdown.Item
                            as="a"
                            className={classes.dropdownItem}
                            disabled={!hasDeletePermission}
                            onClick={() => {
                              selectToDelete(report);
                            }}
                            data-test-id={`delete-report-${report.id}`}
                          >
                            <Delete />

                            <p>Delete</p>
                          </Dropdown.Item>
                        </Dropdown.Menu>
                      </Dropdown>

                      <div
                        className={classNames(
                          classes.icon,
                          classes.reportIcon,
                          reportClassName
                        )}
                      />

                      {hasEditPermission && (
                        <VizWrapperLink
                          className={classNames(classes.icon, classes.editIcon)}
                          href={report.editUrl}
                          onClick={(event) => {
                            if (isSelectMode) {
                              event.preventDefault();
                              toggleReport(report);
                            }
                          }}
                          testId="edit-report-button"
                        />
                      )}
                    </div>
                  )}
                </div>
              );
            }}
          </InView>
        ))}
    </div>
  );
};
