import React, { Fragment, useEffect, useMemo, useState } from 'react';
import classNames from 'classnames';
import {
  ButtonGroup,
  Dropdown,
  FormCheck,
  Popover,
  Spinner,
  Table,
  Tooltip,
} from 'react-bootstrap';
import SortAscending from '@cloudera-internal/cuix-core/icons/svg/ic-sort-ascending.svg';
import SortDescending from '@cloudera-internal/cuix-core/icons/svg/ic-sort-descending.svg';
import PagePrevious from '@cloudera-internal/cuix-core/icons/svg/ic-page-previous.svg';
import PageNext from '@cloudera-internal/cuix-core/icons/svg/ic-page-next.svg';
import Edit from '@cloudera-internal/cuix-core/icons/svg/ic-edit.svg';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import moment from 'moment';
import { useQuery } from '@tanstack/react-query';

import { VizWrapperLink } from 'components/viz-wrapper-link/VizWrapperLink';
import { CombinedReport } from 'types/combined-report.type';
import { ClouderaButton } from '../../../../components/cloudera-button/ClouderaButton';
import { ReportTileImage } from 'components/report-tile-image/ReportTileImage';
import { ClouderaLink } from '../../../../components/cloudera-link/ClouderaLink';
import { VizTooltip } from 'components/VizTooltip';
import {
  getReportUsage,
  reportUsageQueryToken,
} from 'pages/visuals-page/apis/get-report-usage';
import { useSelectedReportsStore } from 'pages/visuals-page/store/selected-reports.store';
import { getGlobalArc } from 'utils/get-global-arc';
import { ReportFilterType } from 'types/report-filter-type.type';
import classes from './ListView.module.scss';

const arc = getGlobalArc();

const NUMBER_OF_ROWS_PER_PAGE = 50;

type ListViewProps = {
  reports: CombinedReport[];
  relatedReportIdToReportMap: Map<
    number,
    {
      id: number;
      title: string;
      sharedVisualDashboardIds: number[];
      workspace: { id: number; name: string };
    }
  >;
  areReportsLoading: boolean;
  searchTerm: string;
  selectedReportType: ReportFilterType;
};

export const ListView = ({
  reports,
  relatedReportIdToReportMap,
  searchTerm,
  areReportsLoading,
  selectedReportType,
}: ListViewProps): JSX.Element => {
  const { data: reportUsage, isLoading: isReportUsageLoading } = useQuery({
    queryKey: [reportUsageQueryToken],
    queryFn: getReportUsage,
    staleTime: Infinity,
  });

  const {
    isReportSelected,
    getNumberOfGroupSelectedReports,

    selectReport,
    toggleReport,

    selectToDelete,

    clear,
  } = useSelectedReportsStore();

  const [page, setPage] = useState(1);

  useEffect(() => {
    setPage(1);
  }, [searchTerm]);

  const isEverythingSelected =
    reports.length !== 0 &&
    getNumberOfGroupSelectedReports() !== 0 &&
    reports.length <= getNumberOfGroupSelectedReports();

  const [sortBy, setSortBy] = useState<{
    key: string;
    sorterFunction: (a: CombinedReport, b: CombinedReport) => number;
    order: 'asc' | 'desc';
  } | null>({
    key: 'last-updated',
    order: 'asc',
    sorterFunction: (a: CombinedReport, b: CombinedReport) =>
      moment(b.updated.date).diff(moment(a.updated.date)),
  });

  const sortedReports = useMemo(() => {
    if (sortBy) {
      return reports.sort(sortBy.sorterFunction);
    }

    return reports;
  }, [reports, sortBy]);

  const columns: {
    key: string;
    label: JSX.Element;
    sorterFunction?: (a: CombinedReport, b: CombinedReport) => number;
    renderCell: (report: CombinedReport) => JSX.Element;
    truncate: boolean;
  }[] = [
    {
      key: 'select-and-thumbnail',
      label: (
        <FormCheck
          checked={isEverythingSelected}
          onChange={() =>
            isEverythingSelected ? clear() : sortedReports.forEach(selectReport)
          }
        />
      ),
      renderCell: (report) => {
        const hasEditPermission =
          (report.type === 'dashboard' &&
            report.permissions.includes('ds_appedit')) ||
          (report.type === 'app' && report.rw);

        let reportClassName = '';

        if (report.type === 'app') {
          reportClassName = classes.appIcon;
        } else if (report.type === 'dashboard') {
          reportClassName = classes.dashboardIcon;
        } else {
          reportClassName = classes.visualIcon;
        }

        const isSelected = isReportSelected(report);

        return (
          <div className="d-flex align-items-center">
            <FormCheck
              className="mr-2"
              checked={isSelected}
              onChange={() => toggleReport(report)}
              data-test-id={isSelected ? 'btn-selected' : 'btn-unselected'}
            />

            <div className={classes.thumbnail}>
              <VizWrapperLink
                href={report.url}
                className={classes.imageContainer}
              >
                <ReportTileImage
                  baseImg={report.baseImg}
                  type={report.type}
                  img={report.img}
                  size="sm"
                />
              </VizWrapperLink>

              <div
                className={classNames(
                  classes.icon,
                  classes.reportIcon,
                  reportClassName
                )}
              />

              {hasEditPermission && (
                <VizWrapperLink
                  className={classNames(classes.icon, classes.editIcon)}
                  href={report.editUrl}
                />
              )}
            </div>
          </div>
        );
      },
      truncate: true,
    },
    {
      key: 'title',
      label: <span>Title</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) =>
        a.title.localeCompare(b.title),
      renderCell: (report) => (
        <ClouderaLink href={report.url} testId="grid-item-link">
          {report.title}
        </ClouderaLink>
      ),
      truncate: true,
    },
    {
      key: 'id',
      label: <span>ID</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) => a.id - b.id,
      renderCell: (report) => <>{report.id}</>,
      truncate: true,
    },
    {
      key: 'related-dashboards',
      label: <span>Related Dashboards / Linked Visuals</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) =>
        (a.sharedVisualDashboardIds?.length ?? 0) -
        (b.sharedVisualDashboardIds?.length ?? 0),
      renderCell: (report) => (
        <div className="d-flex align-items-center">
          <div
            className="mr-2"
            style={{
              width: 24,
              height: 24,
              backgroundColor: '#eaebec',
              borderRadius: '50%',
              display: 'grid',
              placeItems: 'center',
              fontSize: 12,
              lineHeight: 1.33,
              color: '#1b2329',
            }}
          >
            {report.sharedVisualDashboardIds?.length ?? 0}
          </div>

          {report.sharedVisualDashboardIds?.map((id, index) => (
            <Fragment key={id}>
              <VizTooltip
                trigger={['click', 'hover']}
                overlay={
                  <Tooltip id={id}>
                    {relatedReportIdToReportMap.get(+id)?.title}
                  </Tooltip>
                }
              >
                <ClouderaLink
                  href={`${arc.frontend.globals.urls.viewDashboardUrl}/${id}`}
                >
                  {id}
                </ClouderaLink>
              </VizTooltip>
              {index !== report.sharedVisualDashboardIds!.length - 1 && (
                <>,&nbsp;</>
              )}
            </Fragment>
          ))}
        </div>
      ),
      truncate: true,
    },
    {
      key: 'descendant-dashboards',
      label: <span>Descendant Dashboards</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) =>
        a.numberOfDescendants - b.numberOfDescendants,
      renderCell: (report) => (
        <div className="d-flex align-items-center" style={{ gap: 10 }}>
          <div
            style={{
              width: 24,
              height: 24,
              backgroundColor: '#eaebec',
              borderRadius: '50%',
              display: 'grid',
              placeItems: 'center',
              fontSize: 12,
              lineHeight: 1.33,
              color: '#1b2329',
            }}
          >
            {report.numberOfDescendants}
          </div>

          {report.isClone && (
            <div
              className="px-2 py-1"
              style={{
                backgroundColor: '#eaebec',
                borderRadius: 10,
                display: 'grid',
                placeItems: 'center',
                fontSize: 12,
                lineHeight: 1.33,
                color: '#1b2329',
              }}
            >
              Clone
            </div>
          )}
        </div>
      ),
      truncate: true,
    },
    {
      key: 'created',
      label: <span>Created</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) =>
        moment(b.created.date).diff(moment(a.created.date)),
      renderCell: (report) => (
        <>{moment(report.created.date).format('MMM DD, YYYY')}</>
      ),
      truncate: true,
    },
    {
      key: 'last-updated',
      label: <span>Last Updated</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) =>
        moment(b.updated.date).diff(moment(a.updated.date)),
      renderCell: (report) => <>{moment(report.updated.date).fromNow()}</>,
      truncate: true,
    },
    {
      key: 'modified-by',
      label: <span>Modified By</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) =>
        a.updated.by.localeCompare(b.updated.by),
      renderCell: (report) => <>{report.updated.by}</>,
      truncate: true,
    },
    {
      key: 'total-views',
      label: <span>Total Views</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) =>
        (reportUsage?.[a.id] ?? 0) - (reportUsage?.[b.id] ?? 0),
      renderCell: (report) => (
        <>
          {isReportUsageLoading ? (
            <Spinner animation="border" size="sm" variant="secondary" />
          ) : (
            reportUsage?.[report.id] ?? 0
          )}
        </>
      ),
      truncate: true,
    },
    {
      key: 'dataset',
      label: <span>Dataset</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) =>
        (a.datasetInfo?.name ?? '').localeCompare(b.datasetInfo?.name ?? ''),
      renderCell: (report) => (
        <>
          {report.datasetInfo && (
            <div className="d-flex flex-column">
              <ClouderaLink
                href={`${arc.frontend.globals.urls.datasetEditUrl}/${report.datasetInfo.id}`}
              >
                {report.datasetInfo.name}
              </ClouderaLink>

              <span className={classes.dataConnectionName}>
                {report.datasetInfo.dataConnectionName}
              </span>
            </div>
          )}
        </>
      ),
      truncate: true,
    },
    {
      key: 'workspace',
      label: <span>Workspace</span>,
      sorterFunction: (a: CombinedReport, b: CombinedReport) =>
        a.workspace.name.localeCompare(b.workspace.name),
      renderCell: (report) => (
        <ClouderaLink
          href={`${arc.frontend.globals.urls.appsUrl}?group=${report.workspace.name}`}
        >
          {report.workspace.name}
        </ClouderaLink>
      ),
      truncate: true,
    },
    {
      key: 'actions',
      label: <></>,
      renderCell: (report) => {
        const isApp = report.type === 'app';
        const isDashboard = report.type === 'dashboard';

        const hasDsAppEditPermission =
          report.permissions.includes('ds_appedit');

        const hasEditPermission =
          (isDashboard && hasDsAppEditPermission) || (isApp && report.rw);

        const hasRelatedReports =
          (report.sharedVisualDashboardIds?.length ?? 0) > 0;
        const hasDeletePermission =
          (hasDsAppEditPermission && (isDashboard || !hasRelatedReports)) ||
          (isApp && report.rw);

        return (
          <Dropdown>
            <Dropdown.Toggle
              as="i"
              className={classNames(
                'fa fa-ellipsis-v',
                classes.moreOptionsIcon
              )}
              bsPrefix="no"
            />

            <Dropdown.Menu align="right" popperConfig={{ strategy: 'fixed' }}>
              <VizTooltip
                disabled={hasEditPermission}
                overlay={
                  <Popover id={`${report.id}-${report.type}-edit`}>
                    <Popover.Content>
                      You don't have the necessary permissions
                    </Popover.Content>
                  </Popover>
                }
              >
                <span>
                  <Dropdown.Item
                    as="a"
                    className={classes.dropdownItem}
                    disabled={!hasEditPermission}
                    data-test-id="edit-report-button"
                    href={report.editUrl}
                  >
                    <Edit />

                    <p>Edit</p>
                  </Dropdown.Item>
                </span>
              </VizTooltip>

              <VizTooltip
                disabled={hasDeletePermission}
                overlay={
                  <Popover id={`${report.id}-${report.type}-delete`}>
                    <Popover.Content>
                      You don't have the necessary permissions
                    </Popover.Content>
                  </Popover>
                }
              >
                <span>
                  <Dropdown.Item
                    as="button"
                    className={classes.dropdownItem}
                    onClick={() => selectToDelete(report)}
                    disabled={!hasDeletePermission}
                  >
                    <Delete />

                    <p>Delete</p>
                  </Dropdown.Item>
                </span>
              </VizTooltip>
            </Dropdown.Menu>
          </Dropdown>
        );
      },
      truncate: false,
    },
  ];

  const visibleColumns = columns.filter(({ key }) => {
    if (
      !arc.frontend.featureFlipper.getValueById('enable_clone_info_badge') &&
      key === 'descendant-dashboards'
    ) {
      return false;
    }

    if (
      arc.frontend.featureFlipper.getValueById('hide_workspaces') &&
      key === 'workspace'
    ) {
      return false;
    }

    return true;
  });

  const numberOfPages = Math.ceil(reports.length / NUMBER_OF_ROWS_PER_PAGE);

  const paginationButtonNumbers: (number | null)[] = [];

  if (numberOfPages <= 3) {
    Array.from({ length: numberOfPages }, (_, index) =>
      paginationButtonNumbers.push(index + 1)
    );
  } else if (page <= 3) {
    paginationButtonNumbers.push(1, 2, 3, null, numberOfPages);
  } else if (page > numberOfPages - 3) {
    paginationButtonNumbers.push(
      1,
      null,
      numberOfPages - 2,
      numberOfPages - 1,
      numberOfPages
    );
  } else {
    paginationButtonNumbers.push(1, null, page, null, numberOfPages);
  }

  return (
    <>
      <div className={classes.tableContainer}>
        <Table
          hover
          style={{
            borderSpacing: 0,
            borderCollapse: 'separate',
            borderRadius: 3,
            overflow: 'hidden',
          }}
        >
          <thead style={{ backgroundColor: '#fafbfc' }}>
            <tr>
              {visibleColumns.map(({ key, sorterFunction, label }) => (
                <th
                  key={key}
                  className={classNames(
                    classes.headerCell,
                    'border-left-0 border-right-0 border-top-0',
                    {
                      [classes.clickable]: !!sorterFunction,
                    }
                  )}
                  onClick={() => {
                    if (sorterFunction) {
                      setSortBy((current) => {
                        if (current?.key === key) {
                          const order =
                            current.order === 'asc' ? 'desc' : 'asc';
                          return {
                            key,
                            order,
                            sorterFunction: (a, b) => {
                              if (order === 'asc') {
                                return sorterFunction(a, b);
                              } else {
                                return sorterFunction(b, a);
                              }
                            },
                          };
                        }
                        return {
                          key,
                          order: 'asc',
                          sorterFunction: (a, b) => sorterFunction(a, b),
                        };
                      });
                    }
                  }}
                >
                  <div className="d-flex justify-content-between">
                    {label}{' '}
                    {sorterFunction && (
                      <div className={classes.sortIconContainer}>
                        <div
                          className={classNames(classes.sortIcon, {
                            [classes.hide]: sortBy?.key !== key,
                          })}
                        >
                          {sortBy?.key === key ? (
                            sortBy.order === 'asc' ? (
                              <SortAscending />
                            ) : (
                              <SortDescending />
                            )
                          ) : (
                            <SortAscending />
                          )}
                        </div>
                      </div>
                    )}
                  </div>
                </th>
              ))}
            </tr>
          </thead>
          <tbody>
            {!areReportsLoading && sortedReports.length === 0 && (
              <tr style={{ backgroundColor: '#fff' }}>
                <td
                  colSpan={visibleColumns.length}
                  className={classNames(
                    'border-left-0 border-right-0 border-top-0 align-middle',
                    classes.td
                  )}
                >
                  <h4 className="d-flex justify-content-center m-0">
                    There are no{' '}
                    {selectedReportType === 'all'
                      ? 'reports'
                      : selectedReportType === 'dashboards'
                      ? 'dashboards'
                      : 'apps'}{' '}
                    {searchTerm ? `matching "${searchTerm}" ` : ''}in this
                    workspace
                  </h4>
                </td>
              </tr>
            )}

            {areReportsLoading && (
              <tr>
                <td
                  colSpan={visibleColumns.length}
                  className={classNames(
                    'border-left-0 border-right-0 border-top-0 align-middle',
                    classes.td,
                    classes.loaderContainer
                  )}
                >
                  <h4 className="d-flex justify-content-center m-0">
                    <div className={classes.loader}></div>
                  </h4>
                </td>
              </tr>
            )}

            {sortedReports
              .slice(
                (page - 1) * NUMBER_OF_ROWS_PER_PAGE,
                page * NUMBER_OF_ROWS_PER_PAGE
              )
              .map((report) => (
                <tr
                  key={`${report.id}-${report.type}`}
                  data-test-id={report.title}
                  data-test-report-id={report.id}
                  style={{ backgroundColor: '#fff' }}
                >
                  {visibleColumns.map(({ renderCell, key, truncate }) => (
                    <td
                      key={key}
                      className={classNames(
                        'border-left-0 border-right-0 border-top-0 align-middle',
                        classes.td
                      )}
                    >
                      <div
                        className={classNames({
                          [classes.columnContent]: truncate,
                        })}
                      >
                        {renderCell(report)}
                      </div>
                    </td>
                  ))}
                </tr>
              ))}
          </tbody>
        </Table>
      </div>

      <div className="d-flex justify-content-between mt-3">
        <span>
          Showing{' '}
          {Math.min((page - 1) * NUMBER_OF_ROWS_PER_PAGE, reports.length)} to{' '}
          {Math.min(page * NUMBER_OF_ROWS_PER_PAGE, reports.length)} of
          <span data-test-id="table-row-max-row-count"> {reports.length} </span>
          entries
        </span>

        <ButtonGroup className="gap-2">
          <span>
            <ClouderaButton
              onClick={() =>
                setPage((current) => (current - 1 <= 1 ? 1 : current - 1))
              }
              disabled={page === 1}
            >
              <PagePrevious />
            </ClouderaButton>
          </span>

          {paginationButtonNumbers.map((number, index) =>
            number === null ? (
              <div
                key={`${index}-${number}`}
                className={classes.paginationDots}
              >
                ...
              </div>
            ) : (
              <ClouderaButton
                key={number}
                color={page === number ? 'primary' : 'default'}
                onClick={() => {
                  setPage(number);
                }}
              >
                <div className="px-1">{number}</div>
              </ClouderaButton>
            )
          )}

          <span>
            <ClouderaButton
              disabled={page === numberOfPages || numberOfPages === 0}
              onClick={() =>
                setPage((current) =>
                  current + 1 >= numberOfPages ? numberOfPages : current + 1
                )
              }
            >
              <PageNext />
            </ClouderaButton>
          </span>
        </ButtonGroup>
      </div>
    </>
  );
};
