import React, { useEffect, useState } from 'react';
import { Form, Modal, Popover, Table } from 'react-bootstrap';
import { ClouderaButton } from '../../../../components/cloudera-button/ClouderaButton';
import classes from './MoveToWorkspaceDialog.module.scss';
import { CombinedReport } from 'types/combined-report.type';
import Select from 'react-select';
import { workspacesQueryToken } from 'pages/visuals-page/apis/get-workspaces';
import { useMutation, useQueryClient } from '@tanstack/react-query';
import {
  MoveReportsInfos,
  moveReportsToWorkspace,
} from 'pages/visuals-page/apis/move-reports-to-workspace';
import { reportsQueryToken } from 'apis/get-reports';
import { appsQueryToken } from 'apis/get-apps';
import { getReportCountPerGroupQueryToken } from 'apis/get-report-count-per-group';
import { useNotificationStore } from 'stores/use-notification-store';
import { AxiosError } from 'axios';
import Close from '@cloudera-internal/cuix-core/icons/svg/ic-close.svg';
import Move from '@cloudera-internal/cuix-core/icons/svg/ic-get-started.svg';
import Info from '@cloudera-internal/cuix-core/icons/svg/ic-info.svg';
import classNames from 'classnames';
import { VizTooltip } from 'components/VizTooltip';

const linkedVisualOptions = [
  { value: 'clone', label: 'Clone' },
  { value: 'move', label: 'Move' },
] as const;

type MoveToWorkspaceDialogProps = {
  isOpen: boolean;
  close: () => void;
  selectedReports: CombinedReport[];
  onSuccess: () => void;
  workspaces: { id: number; name: string }[] | null;
  selectedGroup: null | { id: number };
  relatedReportIdToReportMap: Map<
    number,
    {
      id: number;
      title: string;
      sharedVisualDashboardIds: number[];
      workspace: { id: number; name: string };
    }
  >;
  selectedLinkedVisuals: {
    id: number;
    type: 'visual';
    title: string;
    workspace: { id: number; name: string };
  }[];
};

export const MoveToWorkspaceDialog = ({
  isOpen,
  close,
  selectedReports,
  onSuccess,
  workspaces,
  selectedGroup,
  relatedReportIdToReportMap,
  selectedLinkedVisuals,
}: MoveToWorkspaceDialogProps): JSX.Element => {
  const queryClient = useQueryClient();

  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const [selectedWorkspaceId, setSelectedWorkspaceId] = useState<number | null>(
    null
  );
  const [linkedVisualActionByReportId, setLinkedVisualActionByReportId] =
    useState<Map<number, 'move' | 'clone'>>(new Map());

  useEffect(() => {
    setLinkedVisualActionByReportId(
      new Map(selectedLinkedVisuals.map(({ id }) => [id, 'clone']))
    );
  }, [selectedLinkedVisuals]);

  const { mutateAsync, isLoading } = useMutation({
    mutationFn: moveReportsToWorkspace,
    mutationKey: [
      workspacesQueryToken,
      reportsQueryToken,
      appsQueryToken,
      getReportCountPerGroupQueryToken,
    ],
    onSuccess: async (_, variables) => {
      queryClient.invalidateQueries({
        queryKey: [reportsQueryToken, selectedGroup?.id],
        refetchType: 'active',
      });
      queryClient.invalidateQueries({
        queryKey: [reportsQueryToken, variables.targetWorkspaceId],
        refetchType: 'active',
      });
      queryClient.invalidateQueries({
        queryKey: [appsQueryToken],
        refetchType: 'active',
      });
      queryClient.invalidateQueries({
        queryKey: [workspacesQueryToken],
        refetchType: 'active',
      });
      queryClient.invalidateQueries({
        queryKey: [getReportCountPerGroupQueryToken],
        refetchType: 'active',
      });

      addNotification(
        {
          message: 'Successfully moved reports',
          variant: 'success',
        },
        { delay: 2000 }
      );

      onSuccess();
    },
    onSettled: async (data) => {
      if (data?.some(({ error }) => !!error)) {
        const movedReportIds: string[] = [];
        const clonedReportIds: number[] = [];
        const failedReportMessages: string[] = [];

        data.forEach(
          ({ reports_to_move, reports_to_clone, reports_failed }) => {
            movedReportIds.push(...Object.keys(reports_to_move));
            clonedReportIds.push(
              ...reports_to_clone.map(({ old_id }) => old_id)
            );
            failedReportMessages.push(...reports_failed);
          }
        );

        let moveStatusInfo = '';

        if (failedReportMessages.length) {
          moveStatusInfo += `Failed to move the following reports: ${failedReportMessages.join(
            ', '
          )}.`;
        }

        if (movedReportIds.length) {
          moveStatusInfo += ` Successfully moved #${movedReportIds.join(', ')}`;
        }

        if (clonedReportIds.length) {
          moveStatusInfo += ` Successfully cloned #${clonedReportIds.join(
            ', '
          )}`;
        }

        addNotification({
          message: `Move failed ${moveStatusInfo}`,
          variant: 'danger',
        });

        close();
      }
    },
    onError: (error: AxiosError) => {
      addNotification({
        message: `Failed to move dashboards: ${
          error.response?.data ?? error.message
        }`,
        variant: 'danger',
      });
    },
  });

  const move = async () => {
    if (selectedWorkspaceId === null) {
      return;
    }

    let dashboardIdToVisualIdsToCloneMap: Map<number, Set<string>> = new Map();

    const moveReportsInfos: MoveReportsInfos = selectedReports.map((report) => {
      if (report.type === 'app') {
        return {
          id: report.id,
          type: 'app' as const,
        };
      }

      if (report.type === 'dashboard') {
        // Side effect :(
        report.sharedVisualDashboardIds?.forEach((linkedVisualId) => {
          const action = linkedVisualActionByReportId.get(+linkedVisualId);
          if (action === 'clone') {
            const set =
              dashboardIdToVisualIdsToCloneMap.get(report.id) ??
              new Set<string>();

            set.add(linkedVisualId);

            dashboardIdToVisualIdsToCloneMap.set(report.id, set);
          }
        });

        return {
          id: report.id,
          type: 'dashboard' as const,
          dashboardName: report.title,
          sharedVisualIds: report.sharedVisualDashboardIds ?? [],
        };
      }

      return {
        id: report.id,
        type: 'visual' as const,
      };
    });

    await mutateAsync({
      moveReportsInfos,
      targetWorkspaceId: selectedWorkspaceId,
      dashboardIdToVisualIdsToCloneMap:
        dashboardIdToVisualIdsToCloneMap.size > 0
          ? dashboardIdToVisualIdsToCloneMap
          : undefined,
    });

    close();
  };

  return (
    <Modal show={isOpen} onHide={close}>
      <Modal.Header className="align-items-center border-bottom-0">
        <h3
          className="text-lg font-light text-dark m-0"
          style={{
            height: 28,
            lineHeight: 1.4,
          }}
        >
          Move To Workspace
        </h3>

        <button
          className="outline-none bg-transparent border-0 p-2 text-dark-gray text-lg"
          onClick={close}
        >
          <Close />
        </button>
      </Modal.Header>

      <Modal.Body>
        <div style={{ maxHeight: 400, overflow: 'scroll' }}>
          <Table
            bordered
            style={{
              borderSpacing: 0,
              borderCollapse: 'separate',
              borderRadius: 3,
              overflow: 'hidden',
            }}
          >
            <thead style={{ backgroundColor: '#fafbfc' }}>
              <tr>
                {['Name', 'Type', 'ID', 'Current workspace', 'Actions'].map(
                  (label) => (
                    <th
                      key={label}
                      className={classNames(
                        'border-left-0 border-right-0 border-top-0',
                        classes.th
                      )}
                    >
                      {label}
                    </th>
                  )
                )}
              </tr>
            </thead>

            <tbody>
              {[
                ...selectedLinkedVisuals,
                ...selectedReports.filter(({ id, type }) =>
                  type !== 'dashboard' && type !== 'app'
                    ? selectedLinkedVisuals.every(
                        (linkedVisual) => id !== linkedVisual.id
                      )
                    : true
                ),
              ].map(({ id, type, title, workspace }) => (
                <tr key={id}>
                  <td
                    className={classNames(
                      'border-left-0 border-right-0 border-top-0 align-middle',
                      classes.td
                    )}
                  >
                    {title}
                  </td>

                  <td
                    className={classNames(
                      'border-left-0 border-right-0 border-top-0 align-middle',
                      classes.td
                    )}
                  >
                    {type !== 'app' && type !== 'dashboard' ? 'visual' : type}
                  </td>

                  <td
                    className={classNames(
                      'border-left-0 border-right-0 border-top-0 align-middle',
                      classes.td
                    )}
                  >
                    {id}
                  </td>

                  <td
                    className={classNames(
                      'border-left-0 border-right-0 border-top-0 align-middle',
                      classes.td
                    )}
                  >
                    {workspace.name}
                  </td>

                  <td
                    className={classNames(
                      'border-left-0 border-right-0 border-top-0 align-middle',
                      classes.td
                    )}
                  >
                    {type !== 'app' &&
                      type !== 'dashboard' &&
                      selectedLinkedVisuals.some(
                        (linkedVisual) => linkedVisual.id === id
                      ) && (
                        <div className="d-flex gap-1">
                          <VizTooltip
                            overlay={
                              <Popover id={`${id}-linked`}>
                                <Popover.Content>
                                  This visual is used in the following
                                  dashboards:{' '}
                                  {relatedReportIdToReportMap
                                    .get(id)
                                    ?.sharedVisualDashboardIds.join(', ') ||
                                    'None'}
                                </Popover.Content>
                              </Popover>
                            }
                          >
                            <div className="d-flex align-items-center text-base">
                              <Info />
                            </div>
                          </VizTooltip>

                          <Select
                            // https://github.com/JedWatson/react-select/issues/3263
                            menuPortalTarget={document.body}
                            styles={{
                              menuPortal: (base) => ({ ...base, zIndex: 9999 }),
                            }}
                            options={linkedVisualOptions}
                            value={
                              linkedVisualOptions.find(
                                ({ value }) =>
                                  linkedVisualActionByReportId.get(id) === value
                              ) ?? linkedVisualOptions[0]
                            }
                            onChange={(event) => {
                              const value = event?.value;

                              if (value && ['clone', 'move'].includes(value)) {
                                setLinkedVisualActionByReportId((current) => {
                                  const newMap = new Map(current);

                                  newMap.set(id, value as 'clone' | 'move');

                                  return newMap;
                                });
                              }
                            }}
                          />
                        </div>
                      )}
                  </td>
                </tr>
              ))}
            </tbody>
          </Table>
        </div>

        <Form.Group>
          <Form.Label htmlFor="workspace-select">Target workspace</Form.Label>

          <Select
            name="workspace-select"
            inputId="workspace-select"
            isSearchable
            options={workspaces ?? []}
            getOptionLabel={({ name }) => name}
            getOptionValue={({ id }) => `${id}`}
            value={workspaces?.find(({ id }) => id === selectedWorkspaceId)}
            onChange={(data) => {
              if (data) {
                setSelectedWorkspaceId(data.id);
              }
            }}
          />
        </Form.Group>
      </Modal.Body>

      <Modal.Footer className="justify-content-end">
        <ClouderaButton onClick={close}>Cancel</ClouderaButton>

        <ClouderaButton
          color="primary"
          disabled={selectedWorkspaceId === null || isLoading}
          onClick={move}
          testId="submit-move-workspace"
          icon={<Move />}
        >
          Move
        </ClouderaButton>
      </Modal.Footer>
    </Modal>
  );
};
