import React, { useMemo } from 'react';
import { Dropdown, Popover } from 'react-bootstrap';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import Edit from '@cloudera-internal/cuix-core/icons/svg/ic-edit.svg';
import ArrowDown from '@cloudera-internal/cuix-core/icons/svg/ic-dropdown.svg';
import User from '@cloudera-internal/cuix-core/icons/svg/ic-user.svg';
import Placeholder from '@cloudera-internal/cuix-core/icons/svg/ic-placeholder-fpo.svg';
import Globe from '@cloudera-internal/cuix-core/icons/svg/ic-globe.svg';
import Lightbulb from '@cloudera-internal/cuix-core/icons/svg/ic-lightbulb.svg';
import Favorite from '@cloudera-internal/cuix-core/icons/svg/ic-star-outlined.svg';
import Key from '@cloudera-internal/cuix-core/icons/svg/ic-key.svg';
import Groups from '@cloudera-internal/cuix-core/icons/svg/ic-groups.svg';

import { ClouderaButton } from '../../../../components/cloudera-button/ClouderaButton';
import { VizTooltip } from 'components/VizTooltip';
import classes from './SelectedGroup.module.scss';
import { Group } from 'pages/visuals-page/hooks/use-selected-group';
import { StaticSidebarItems } from 'pages/visuals-page/VisualsPage';
import { Workspace } from 'pages/visuals-page/apis/get-workspaces';

type SelectedGroupProps = {
  selectedGroup: Group | null;
  areWorkspacesLoading: boolean;
  staticSideBarItems: StaticSidebarItems;
  hasReports: (groupId: number) => boolean;
  setSelectedWorkspaceToDelete: (
    workspace: {
      id: number;
      name: string;
    } | null
  ) => void;
  setSelectedWorkspaceToEdit: (
    workspace: {
      id: number;
      name: string;
      desc: string;
      acl?: Workspace['acl'];
      perms?: string[];
    } | null
  ) => void;
};

export const SelectedGroup = ({
  areWorkspacesLoading,
  selectedGroup,
  staticSideBarItems,
  hasReports,
  setSelectedWorkspaceToDelete,
  setSelectedWorkspaceToEdit,
}: SelectedGroupProps): JSX.Element => {
  const tooltips = useMemo<{
    actions: string | false;
    edit: string | false;
    delete: string | false;
  }>(() => {
    if (areWorkspacesLoading) {
      return {
        actions: 'Workspaces are loading',
        edit: 'Workspaces are loading',
        delete: 'Workspaces are loading',
      };
    }

    if (selectedGroup === null) {
      return {
        actions: 'Nothing is selected',
        edit: 'Nothing is selected',
        delete: 'Nothing is selected',
      };
    }

    const staticGroups = [
      staticSideBarItems.all,
      staticSideBarItems.favorites,
      staticSideBarItems.private,
    ];

    if (staticSideBarItems.samples) {
      staticGroups.push(staticSideBarItems.samples);
    }

    const selectedStaticGroup = staticGroups.find(
      ({ id }) => selectedGroup.id === id
    );

    if (selectedStaticGroup) {
      return {
        actions: `"${selectedStaticGroup.name}" can't be managed`,
        edit: `"${selectedStaticGroup.name}" can't be managed`,
        delete: `"${selectedStaticGroup.name}" can't be managed`,
      };
    }

    const isPublicWorkspace =
      selectedGroup && selectedGroup.id === staticSideBarItems.public?.id;

    const canEdit =
      selectedGroup.type === 'workspace'
        ? (selectedGroup.editable || isPublicWorkspace) &&
          selectedGroup.perms.includes('ws_manage')
        : false;

    const hasManagePermission =
      selectedGroup.type === 'workspace'
        ? selectedGroup.perms.includes('ws_manage')
        : false;

    const workspaceHasReports = hasReports(selectedGroup.id);
    const workspaceIsEmpty = !workspaceHasReports;
    const canDelete =
      !isPublicWorkspace && workspaceIsEmpty && hasManagePermission;

    if (!canEdit && !canDelete) {
      return {
        actions: "You don't have the necessary permissions",
        edit: "You don't have the necessary permissions",
        delete: "You don't have the necessary permissions",
      };
    }

    return {
      actions: false,
      edit: !canEdit && "You don't have the necessary permissions",
      delete: isPublicWorkspace
        ? "Public workspace can't be deleted"
        : !hasManagePermission
        ? "You don't have the necessary permissions"
        : workspaceHasReports
        ? "Can't delete non empty workspace"
        : false,
    };
  }, [
    areWorkspacesLoading,
    hasReports,
    selectedGroup,
    staticSideBarItems.all,
    staticSideBarItems.favorites,
    staticSideBarItems.private,
    staticSideBarItems.public?.id,
    staticSideBarItems.samples,
  ]);

  let title = <></>;
  let desc = <></>;

  if (!areWorkspacesLoading) {
    if (selectedGroup !== null) {
      let icon = <User />;

      if (selectedGroup.id === staticSideBarItems.all.id) {
        icon = <Globe />;
      } else if (selectedGroup.id === staticSideBarItems.favorites.id) {
        icon = <Favorite />;
      } else if (selectedGroup.id === staticSideBarItems.samples?.id) {
        icon = <Lightbulb />;
      } else if (selectedGroup.id === staticSideBarItems.public?.id) {
        icon = <Groups />;
      } else if (
        selectedGroup.type === 'workspace' &&
        !selectedGroup.editable
      ) {
        icon = <Key />;
      }

      title = (
        <>
          <span className="text-2xl">{icon}</span>
          <h3 className={classes.name}>{selectedGroup.name}</h3>
        </>
      );
      desc = <p className={classes.desc}>{selectedGroup.desc}</p>;
    }
  } else {
    title = (
      <>
        <span className="text-xl text-dark-gray mr-2">
          <Placeholder />
        </span>
        <div className={classes.stripe} />
      </>
    );
    desc = (
      <>
        <div className={classes.stripe} />
        <div className={classes.stripe} />
      </>
    );
  }

  return (
    <div className={classes.selectedGroup}>
      <div className="d-flex justify-content-between align-items-center">
        <div
          className="d-flex align-items-center grow overflow-hidden gap-2"
          style={{ height: 28 }}
        >
          {title}
        </div>

        <div className="ml-2 shrink-0">
          <Dropdown>
            <VizTooltip
              overlay={
                <Popover id="selected-group-actions-tooltip">
                  <Popover.Content>{tooltips.actions}</Popover.Content>
                </Popover>
              }
              disabled={!tooltips.actions}
            >
              <div>
                <Dropdown.Toggle
                  as={ClouderaButton}
                  icon={<ArrowDown />}
                  iconPosition="right"
                  disabled={!!tooltips.actions}
                  id="selected-group-actions"
                  size="sm"
                >
                  Actions
                </Dropdown.Toggle>
              </div>
            </VizTooltip>

            <Dropdown.Menu>
              <VizTooltip
                overlay={
                  <Popover id="edit-group-tooltip">
                    <Popover.Content>{tooltips.edit}</Popover.Content>
                  </Popover>
                }
                disabled={!tooltips.edit}
              >
                <Dropdown.Item
                  as="div"
                  className={classes.dropdownItem}
                  onClick={() => {
                    if (!tooltips.edit) {
                      setSelectedWorkspaceToEdit(selectedGroup);
                    }
                  }}
                  disabled={!!tooltips.edit}
                >
                  <Edit />
                  <p>Edit Workspace</p>
                </Dropdown.Item>
              </VizTooltip>

              <VizTooltip
                overlay={
                  <Popover id="delete-group-tooltip">
                    <Popover.Content>{tooltips.delete}</Popover.Content>
                  </Popover>
                }
                disabled={!tooltips.delete}
              >
                <div>
                  <Dropdown.Item
                    as="div"
                    className={classes.dropdownItem}
                    onClick={() => {
                      if (!tooltips.delete) {
                        setSelectedWorkspaceToDelete(selectedGroup);
                      }
                    }}
                    disabled={!!tooltips.delete}
                  >
                    <Delete />
                    <p>Delete Workspace</p>
                  </Dropdown.Item>
                </div>
              </VizTooltip>
            </Dropdown.Menu>
          </Dropdown>
        </div>
      </div>
      {desc}
    </div>
  );
};
