import React, { DragEventHandler, PropsWithChildren, useState } from 'react';

import Groups from '@cloudera-internal/cuix-core/icons/svg/ic-groups.svg';
import Key from '@cloudera-internal/cuix-core/icons/svg/ic-key.svg';
import Plus from '@cloudera-internal/cuix-core/icons/svg/ic-plus-circle.svg';
import User from '@cloudera-internal/cuix-core/icons/svg/ic-user.svg';
import Globe from '@cloudera-internal/cuix-core/icons/svg/ic-globe.svg';
import Lightbulb from '@cloudera-internal/cuix-core/icons/svg/ic-lightbulb.svg';
import Favorite from '@cloudera-internal/cuix-core/icons/svg/ic-star-outlined.svg';
import StatusError from '@cloudera-internal/cuix-core/icons/svg/ic-status-error.svg';
import Edit from '@cloudera-internal/cuix-core/icons/svg/ic-edit.svg';
import ArrowDown from '@cloudera-internal/cuix-core/icons/svg/ic-dropdown.svg';
import ArrowRight from '@cloudera-internal/cuix-core/icons/svg/ic-dropright.svg';

import classes from './SideBar.module.scss';
import { SearchBar } from '../search-bar/SearchBar';
import classNames from 'classnames';
import { ClouderaButton } from '../../../../components/cloudera-button/ClouderaButton';
import { useToggle } from 'hooks/use-toggle';
import { getGlobalArc } from 'utils/get-global-arc';
import { StaticSidebarItems } from 'pages/visuals-page/VisualsPage';
import { WorkspaceDialog } from '../workspace-dialog/WorkspaceDialog';
import { Workspace } from 'pages/visuals-page/apis/get-workspaces';
import { vizCookie } from 'services/viz-cookie';
import { Group } from 'pages/visuals-page/hooks/use-selected-group';
import {
  getReportCountPerGroupQueryToken,
  getReportCountPerGroup,
} from 'apis/get-report-count-per-group';
import { useQuery } from '@tanstack/react-query';
import { CircularProgressBar } from '../common/CircularProgessBar';
import { VizTooltip } from 'components/VizTooltip';
import { Collapse, Popover } from 'react-bootstrap';
import { ClouderaLink } from '../../../../components/cloudera-link/ClouderaLink';

const arc = getGlobalArc();

const limitSideBarWidth = (width?: number) => {
  if (width === undefined || width <= 240) {
    return 240;
  }

  if (width >= 600) {
    return 600;
  }

  return width;
};

type SideBarProps = {
  staticSideBarItems: StaticSidebarItems;

  selectedGroupId: number | undefined;
  setSelectedGroup: (group: Group) => void;

  groupedWorkspaces:
    | { sharedWorkspaces: Workspace[]; privateWorkspaces: Workspace[] }
    | undefined;
  workspaces: { id: number; name: string }[] | null;

  areWorkspacesLoading: boolean;
  hasWorkspacesError: boolean;

  setSelectedWorkspaceToEdit: (
    workspace: {
      id: number;
      name: string;
      desc: string;
      acl?: Workspace['acl'];
      perms?: string[];
    } | null
  ) => void;

  publicWorkspace:
    | {
        id: number;
        name: string;
        desc: string;
        editable: boolean;
        perms: string[];
        acl: [[type: 1 | 2, permission: 1 | 2 | 3, name: string]];
      }
    | undefined;
  setPublicWorkspace: (publicWorkspace: {
    id: number;
    name: string;
    desc: string;
    editable: boolean;
    perms: string[];
    acl: [[type: 1 | 2, permission: 1 | 2 | 3, name: string]];
  }) => void;
};

export const SideBar = ({
  staticSideBarItems,
  selectedGroupId,
  setSelectedGroup,
  groupedWorkspaces,
  workspaces,
  areWorkspacesLoading,
  hasWorkspacesError,
  setSelectedWorkspaceToEdit,
  publicWorkspace,
  setPublicWorkspace,
}: SideBarProps): JSX.Element => {
  const [sideBarWidth, setSidebarWidth] = useState(
    limitSideBarWidth(vizCookie.getVisualTabSettingByKey('sideBarStyle')?.width)
  );

  const { data, error, isFetching } = useQuery({
    queryKey: [getReportCountPerGroupQueryToken],
    queryFn: getReportCountPerGroup,
  });

  const [workspaceSearchTerm, setWorkspaceSearchTerm] = useState('');

  const [
    isNewWorkspaceDialogOpen,
    toggleNewWorkspaceDialog,
    closeNewWorkspaceDialog,
  ] = useToggle();

  const [showSharedWorkspaces, toggleShowSharedWorkspaces] = useToggle(true);
  const [showPrivateWorkspaces, toggleShowPrivateWorkspaces] = useToggle();

  const visibleSharedWorkspaces = groupedWorkspaces?.sharedWorkspaces.filter(
    ({ name }) => name.toLowerCase().includes(workspaceSearchTerm.toLowerCase())
  );

  const visiblePrivateWorkspaces = groupedWorkspaces?.privateWorkspaces.filter(
    ({ name }) => name.toLowerCase().includes(workspaceSearchTerm.toLowerCase())
  );

  const onResize: DragEventHandler = (event) => {
    if (event.pageX) {
      const limitedWith = limitSideBarWidth(event.pageX);

      setSidebarWidth(limitedWith);

      vizCookie.setVisualTabSettingByKey('sideBarStyle', {
        width: limitedWith,
      });
    }
  };

  return (
    <>
      <div
        className={classes.workspaceContainer}
        style={{ width: sideBarWidth }}
      >
        <div className={classes.title} data-test-id="workspace-title">
          Workspaces
        </div>

        <div className={classes.toolbar}>
          <div className={classNames('grow', classes.searchbarContainer)}>
            <SearchBar
              id="search-workspaces"
              placeholder="Search Workspace"
              searchTerm={workspaceSearchTerm}
              setSearchTerm={setWorkspaceSearchTerm}
            />
          </div>

          {arc.frontend.globals.permissions.system.includes(
            'sys_workspace'
          ) && (
            <ClouderaButton
              onClick={toggleNewWorkspaceDialog}
              icon={<Plus />}
              testId="create-workspace-icon"
            />
          )}
        </div>

        <div className={classes.workspaces}>
          <GroupSelector
            id={staticSideBarItems.all.id}
            isActive={selectedGroupId === staticSideBarItems.all.id}
            icon={<Globe />}
            reportCount={{
              hasError: !!error,
              isLoading: isFetching,
              count:
                data?.workspaceIdToReportCount && data?.workspaceIdToAppCount
                  ? Object.values(data.workspaceIdToReportCount)
                      .concat(Object.values(data.workspaceIdToAppCount))
                      .reduce((sum, count) => sum + count, 0)
                  : undefined,
            }}
            testId="app-group-all"
            onClick={() =>
              setSelectedGroup({ ...staticSideBarItems.all, type: 'all' })
            }
          >
            {staticSideBarItems.all.name.toUpperCase()}
          </GroupSelector>

          <GroupSelector
            id={staticSideBarItems.favorites.id}
            isActive={selectedGroupId === staticSideBarItems.favorites.id}
            icon={<Favorite />}
            reportCount={{
              hasError: !!error,
              isLoading: isFetching,
              count: data?.favoritesReportCount,
            }}
            testId="favorites-app-group"
            onClick={() =>
              setSelectedGroup({
                ...staticSideBarItems.favorites,
                type: 'app-group',
              })
            }
            tooltip={
              <div className="pb-2">
                <span className="d-block w-100 text-left">
                  Name: {staticSideBarItems.favorites.name}
                </span>
                <span className="d-block w-100 text-left">
                  Description: {staticSideBarItems.favorites.desc}
                </span>
              </div>
            }
          >
            {staticSideBarItems.favorites.name.toUpperCase()}
          </GroupSelector>

          {staticSideBarItems.samples && (
            <GroupSelector
              id={staticSideBarItems.samples.id}
              isActive={selectedGroupId === staticSideBarItems.samples.id}
              icon={<Lightbulb />}
              reportCount={{
                hasError: !!error || data?.samplesReportCount === null,
                isLoading: isFetching,
                count: data?.samplesReportCount ?? 0,
              }}
              testId="app-group-samples"
              onClick={() => {
                if (staticSideBarItems.samples) {
                  setSelectedGroup({
                    ...staticSideBarItems.samples,
                    type: 'app-group',
                  });
                }
              }}
              tooltip={
                <div className="pb-2">
                  <span className="d-block w-100 text-left">
                    Name: {staticSideBarItems.samples.name}
                  </span>
                  <span className="d-block w-100 text-left">
                    Description: {staticSideBarItems.samples.desc}
                  </span>
                </div>
              }
            >
              {staticSideBarItems.samples.name.toUpperCase()}
            </GroupSelector>
          )}

          {publicWorkspace && (
            <GroupSelector
              id={publicWorkspace.id}
              isActive={selectedGroupId === publicWorkspace.id}
              icon={<Groups />}
              reportCount={{
                hasError: !!error,
                isLoading: isFetching,
                count:
                  (data?.workspaceIdToAppCount?.[publicWorkspace.id] ?? 0) +
                  (data?.workspaceIdToReportCount?.[publicWorkspace.id] ?? 0),
              }}
              testId={publicWorkspace.name}
              tooltip={
                <div className="pb-2">
                  <span className="d-block w-100 text-left">
                    Name: {publicWorkspace.name}
                  </span>
                  <span className="d-block w-100 text-left">
                    Description: {publicWorkspace.desc}
                  </span>
                </div>
              }
              onClick={() => {
                if (publicWorkspace) {
                  setSelectedGroup({
                    ...publicWorkspace,
                    type: 'workspace',
                  });
                }
              }}
              onEdit={
                publicWorkspace.perms.includes('ws_manage')
                  ? () => {
                      if (publicWorkspace) {
                        setSelectedWorkspaceToEdit(publicWorkspace);
                      }
                    }
                  : undefined
              }
            >
              PUBLIC
            </GroupSelector>
          )}

          <GroupSelector
            id={staticSideBarItems.private.id}
            isActive={selectedGroupId === staticSideBarItems.private.id}
            icon={<Key />}
            reportCount={{
              hasError: !!error,
              isLoading: isFetching,
              count:
                (data?.workspaceIdToAppCount?.[staticSideBarItems.private.id] ??
                  0) +
                (data?.workspaceIdToReportCount?.[
                  staticSideBarItems.private.id
                ] ?? 0),
            }}
            testId={staticSideBarItems.private.name}
            tooltip={
              <div className="pb-2">
                <span className="d-block w-100 text-left">
                  Name: {staticSideBarItems.private.name}
                </span>
                <span className="d-block w-100 text-left">
                  Description: {staticSideBarItems.private.desc}
                </span>
              </div>
            }
            onClick={() =>
              setSelectedGroup({
                ...staticSideBarItems.private,
                type: 'workspace',
              })
            }
          >
            PRIVATE
          </GroupSelector>

          <ClouderaLink
            className={classes.link}
            onClick={toggleShowSharedWorkspaces}
          >
            <p className={classes.groupTitle}>
              <span className={classes.dropIcon}>
                {showSharedWorkspaces ? <ArrowDown /> : <ArrowRight />}
              </span>
              <span className="ml-2">Shared Workspaces</span>
            </p>
          </ClouderaLink>

          <Collapse in={showSharedWorkspaces}>
            <div>
              {areWorkspacesLoading && (
                <div
                  className={classNames(
                    'd-flex align-items-center',
                    classes.container
                  )}
                >
                  <i
                    className={classNames('fa fa-picture-o mr-2', classes.icon)}
                  />

                  <div className={classes.stripe} />

                  <div className={classes.badge}>
                    <CircularProgressBar />
                  </div>
                </div>
              )}

              {!areWorkspacesLoading && hasWorkspacesError && (
                <div className="d-flex align-items-center flex-column">
                  <div className="d-flex align-items-center mb-1">
                    <span className="text-base">
                      <StatusError />
                    </span>

                    <h3 className={classes.errorTitle}>
                      Error loading workspaces
                    </h3>
                  </div>
                  <p className={classes.errorDescription}>
                    Please refresh the page or try again later
                  </p>
                </div>
              )}

              {!areWorkspacesLoading &&
                visibleSharedWorkspaces?.length === 0 &&
                workspaceSearchTerm !== '' && (
                  <p className="text-center">There are no matches</p>
                )}

              {visibleSharedWorkspaces?.map((workspace) => (
                <GroupSelector
                  key={workspace.id}
                  id={workspace.id}
                  isActive={selectedGroupId === workspace.id}
                  icon={<User />}
                  reportCount={{
                    hasError: !!error,
                    isLoading: isFetching,
                    count:
                      (data?.workspaceIdToAppCount?.[workspace.id] ?? 0) +
                      (data?.workspaceIdToReportCount?.[workspace.id] ?? 0),
                  }}
                  testId={workspace.name}
                  tooltip={
                    <div className="pb-2">
                      <span className="d-block w-100 text-left">
                        Name: {workspace.name}
                      </span>
                      <span className="d-block w-100 text-left">
                        Description: {workspace.desc}
                      </span>
                    </div>
                  }
                  onClick={() =>
                    setSelectedGroup({ ...workspace, type: 'workspace' })
                  }
                  onEdit={
                    workspace.editable && workspace.perms.includes('ws_manage')
                      ? () => setSelectedWorkspaceToEdit(workspace)
                      : undefined
                  }
                >
                  {workspace.name}
                </GroupSelector>
              ))}
            </div>
          </Collapse>

          {(groupedWorkspaces?.privateWorkspaces.length ?? 0) > 0 && (
            <>
              <ClouderaLink
                className={classes.link}
                onClick={toggleShowPrivateWorkspaces}
              >
                <p className={classes.groupTitle}>
                  <span className={classes.dropIcon}>
                    {showPrivateWorkspaces ? <ArrowDown /> : <ArrowRight />}
                  </span>
                  <span className="ml-2">Other Private Workspaces</span>
                </p>
              </ClouderaLink>

              <Collapse in={showPrivateWorkspaces}>
                <div>
                  {areWorkspacesLoading && (
                    <div
                      className={classNames(
                        'd-flex align-items-center',
                        classes.container
                      )}
                    >
                      <i
                        className={classNames(
                          'fa fa-picture-o mr-2',
                          classes.icon
                        )}
                      />

                      <div className={classes.stripe} />

                      <div className={classes.badge}>
                        <CircularProgressBar />
                      </div>
                    </div>
                  )}

                  {!areWorkspacesLoading && hasWorkspacesError && (
                    <div className="d-flex align-items-center flex-column">
                      <div className="d-flex align-items-center mb-1">
                        <span className="text-base">
                          <StatusError />
                        </span>

                        <h3 className={classes.errorTitle}>
                          Error loading workspaces
                        </h3>
                      </div>
                      <p className={classes.errorDescription}>
                        Please refresh the page or try again later
                      </p>
                    </div>
                  )}

                  {!areWorkspacesLoading &&
                    visiblePrivateWorkspaces?.length === 0 &&
                    workspaceSearchTerm !== '' && (
                      <p className="text-center">There are no matches</p>
                    )}

                  {visiblePrivateWorkspaces?.map((workspace) => (
                    <GroupSelector
                      key={workspace.id}
                      id={workspace.id}
                      isActive={selectedGroupId === workspace.id}
                      icon={<Key />}
                      reportCount={{
                        hasError: !!error,
                        isLoading: isFetching,
                        count:
                          (data?.workspaceIdToAppCount?.[workspace.id] ?? 0) +
                          (data?.workspaceIdToReportCount?.[workspace.id] ?? 0),
                      }}
                      testId={workspace.name}
                      tooltip={
                        <div className="pb-2">
                          <span className="d-block w-100 text-left">
                            Name: {workspace.name}
                          </span>
                          <span className="d-block w-100 text-left">
                            Description: {workspace.desc}
                          </span>
                        </div>
                      }
                      onClick={() =>
                        setSelectedGroup({ ...workspace, type: 'workspace' })
                      }
                    >
                      {workspace.name}
                    </GroupSelector>
                  ))}
                </div>
              </Collapse>
            </>
          )}
        </div>

        <div className={classes.resizer} draggable onDrag={onResize} />
      </div>

      {arc.frontend.globals.permissions.system.includes('sys_workspace') &&
        isNewWorkspaceDialogOpen &&
        workspaces && (
          <WorkspaceDialog
            isOpen
            close={closeNewWorkspaceDialog}
            workspaces={workspaces}
            publicWorkspace={publicWorkspace}
            setPublicWorkspace={setPublicWorkspace}
            selectedWorkspace={null}
          />
        )}
    </>
  );
};

type GroupSelectorProps = PropsWithChildren<{
  id: number;
  isActive: boolean;
  icon: JSX.Element;
  reportCount: {
    hasError: boolean;
    isLoading: boolean;
    count: number | undefined;
  };
  testId: string;
  tooltip?: JSX.Element;
  onClick: () => void;
  onEdit?: () => void;
}>;

export const GroupSelector = ({
  id,
  isActive,
  icon,
  reportCount,
  testId,
  tooltip,
  onClick,
  onEdit,
  children,
}: GroupSelectorProps): JSX.Element => {
  const button = (
    <button
      className={classNames(classes.groupSelector, {
        [classes.active]: isActive,
      })}
      onClick={onClick}
    >
      <div className={classes.text}>
        <span className="mr-2 text-base">{icon}</span>
        <span data-test-id={testId} data-test-active={isActive}>
          {children}
        </span>
      </div>

      <div className="d-flex gap-4">
        {onEdit && (
          <span
            className={classNames('p-1 ml-2 text-base', classes.editIcon)}
            onClick={(event) => {
              event.stopPropagation();
              onEdit();
            }}
            data-test-id="edit-workspace"
          >
            <Edit />
          </span>
        )}

        {reportCount.hasError ? (
          <VizTooltip
            overlay={
              <Popover id={`${id}-report-count-error`}>
                <Popover.Content>
                  Error while loading the report count
                </Popover.Content>
              </Popover>
            }
          >
            <div className="text-lg">
              <StatusError />
            </div>
          </VizTooltip>
        ) : (
          <div
            className={classes.badge}
            data-test-id={`${testId}-report-count`}
          >
            {reportCount.isLoading ? (
              <CircularProgressBar />
            ) : (
              reportCount.count
            )}
          </div>
        )}
      </div>
    </button>
  );

  if (tooltip) {
    return (
      <VizTooltip
        overlay={
          <Popover id={id.toString()}>
            <Popover.Content>{tooltip}</Popover.Content>
          </Popover>
        }
        placement="right"
      >
        {button}
      </VizTooltip>
    );
  }

  return button;
};
