import React from 'react';
import { Dropdown, Popover } from 'react-bootstrap';
import classNames from 'classnames';
import MoveToIcon from '@cloudera-internal/cuix-core/icons/svg/ic-get-started.svg';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import Export from '@cloudera-internal/cuix-core/icons/svg/ic-import.svg';
import ArrowDown from '@cloudera-internal/cuix-core/icons/svg/ic-dropdown.svg';
import ListIcon from '@cloudera-internal/cuix-core/icons/svg/ic-list-view.svg';
import Favorite from '@cloudera-internal/cuix-core/icons/svg/ic-star-outlined.svg';
import Plus from '@cloudera-internal/cuix-core/icons/svg/ic-plus-circle.svg';
import Accelerate from '@cloudera-internal/cuix-core/icons/svg/ic-expand-nav.svg';
import DashboardIcon from '@cloudera-internal/cuix-core/icons/svg/ic-grid-view.svg';
import AppIcon from '@cloudera-internal/cuix-core/icons/svg/ic-profiler-metrics.svg';

import { VizTooltip } from 'components/VizTooltip';
import { getGlobalArc } from 'utils/get-global-arc';
import { ClouderaButton } from '../../../../components/cloudera-button/ClouderaButton';
import { SearchBar } from '../search-bar/SearchBar';
import { useNotificationStore } from 'stores/use-notification-store';
import { useSelectedReportsStore } from 'pages/visuals-page/store/selected-reports.store';
import { Group } from 'pages/visuals-page/hooks/use-selected-group';
import classes from './Toolbar.module.scss';
import { ReportViewType } from 'pages/visuals-page/hooks/use-selected-view-mode';

const arc = getGlobalArc();

type ToolbarProps = {
  reportsSearchTerm: string;
  setReportsSearchTerm: (searchTerm: string) => void;
  canCreateDashboard: boolean;
  canCreateApp: boolean;
  selectedGroup: Group | null;
  favoritesGroupId: number;
  toggleFavoriteGroup: () => void;
  selectedViewMode: ReportViewType;
  setSelectedViewMode: (viewMode: ReportViewType) => void;
  areWorkspacesLoading: boolean;
};

export const Toolbar = ({
  reportsSearchTerm,
  setReportsSearchTerm,
  canCreateDashboard,
  canCreateApp,
  selectedGroup,
  favoritesGroupId,
  toggleFavoriteGroup,
  selectedViewMode,
  setSelectedViewMode,
  areWorkspacesLoading,
}: ToolbarProps): JSX.Element => {
  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const {
    groupSelectedReports,
    setGroupActionToDelete,
    setGroupActionToMove,
    setGroupActionToExport,
    getNumberOfGroupSelectedReports,
  } = useSelectedReportsStore();

  return (
    <div className="d-flex justify-content-between align-items-end mb-4">
      <div className="d-flex gap-2">
        <SearchBar
          id="search-reports"
          placeholder="Search Report"
          searchTerm={reportsSearchTerm}
          setSearchTerm={setReportsSearchTerm}
          testId="searchbox"
        />

        {arc.frontend.featureFlipper.getValueById('enable_explore_style') && (
          <VizTooltip
            overlay={
              <Popover id="favorites">
                <Popover.Content>Show Favorites</Popover.Content>
              </Popover>
            }
          >
            <ClouderaButton
              color={
                selectedGroup?.id === favoritesGroupId ? 'primary' : 'default'
              }
              onClick={toggleFavoriteGroup}
              icon={<Favorite />}
            >
              <span className={classes.hideOnSmallScreen}>Favorites</span>
            </ClouderaButton>
          </VizTooltip>
        )}
      </div>

      <div className="d-flex gap-1 mt-2">
        {arc.frontend.globals.settings.isArcengineDataconnectionEnable && (
          <VizTooltip
            overlay={
              <Popover id="accelerate-reports-toolbar">
                <Popover.Content>
                  {getNumberOfGroupSelectedReports() === 0
                    ? 'Select reports below'
                    : 'Accelerate'}
                </Popover.Content>
              </Popover>
            }
          >
            <ClouderaButton
              disabled={getNumberOfGroupSelectedReports() === 0}
              onClick={() => {
                const ids = groupSelectedReports
                  ?.filter(({ type }) => type !== 'app')
                  .map(({ id }) => id);

                if (!ids) {
                  addNotification(
                    {
                      message: 'Please select dashboard(s) below',
                      variant: 'info',
                    },
                    {
                      delay: 2000,
                    }
                  );

                  return;
                }

                document.location.href = `${
                  arc.frontend.globals.urls.accelerateUrl
                }?appids=${ids.join('_')}`;
              }}
              testId="accelerate"
              icon={<Accelerate />}
            >
              <span className={classes.hideOnSmallScreen}>Accelerate</span>
            </ClouderaButton>
          </VizTooltip>
        )}

        <VizTooltip
          overlay={
            <Popover id="export-report-toolbar">
              <Popover.Content>
                {getNumberOfGroupSelectedReports() === 0
                  ? 'Select reports below'
                  : 'Export'}
              </Popover.Content>
            </Popover>
          }
        >
          <ClouderaButton
            disabled={getNumberOfGroupSelectedReports() === 0}
            onClick={setGroupActionToExport}
            testId="export-reports"
            icon={<Export />}
          >
            <span className={classes.hideOnSmallScreen}>Export</span>
          </ClouderaButton>
        </VizTooltip>

        <VizTooltip
          overlay={
            <Popover id="delete-reports-toolbar">
              <Popover.Content>
                {getNumberOfGroupSelectedReports() === 0
                  ? 'Select reports below'
                  : 'Delete'}
              </Popover.Content>
            </Popover>
          }
        >
          <ClouderaButton
            disabled={getNumberOfGroupSelectedReports() === 0}
            onClick={setGroupActionToDelete}
            icon={<Delete />}
            testId="delete-apps-btn"
          >
            <span className={classes.hideOnSmallScreen}>Delete</span>
          </ClouderaButton>
        </VizTooltip>

        {!arc.frontend.featureFlipper.getValueById('hide_workspaces') && (
          <VizTooltip
            overlay={
              <Popover id="move-to-workspace-toolbar">
                <Popover.Content>
                  {getNumberOfGroupSelectedReports() === 0
                    ? 'Select reports below'
                    : 'Move To Workspace'}
                </Popover.Content>
              </Popover>
            }
          >
            <ClouderaButton
              disabled={getNumberOfGroupSelectedReports() === 0}
              onClick={setGroupActionToMove}
              icon={<MoveToIcon />}
            >
              <span className={classes.hideOnSmallScreen}>
                Move To Workspace
              </span>
            </ClouderaButton>
          </VizTooltip>
        )}

        <div className={classNames('d-flex', classes.buttonGroup)}>
          <VizTooltip
            overlay={
              <Popover id="grid-view">
                <Popover.Content>Grid view</Popover.Content>
              </Popover>
            }
          >
            <button
              className={classNames(classes.filterButton, {
                [classes.activeButton]: selectedViewMode === 'grid',
              })}
              onClick={() => setSelectedViewMode('grid')}
              data-test-id="app-grid-dashboards-btn"
            >
              <DashboardIcon />
            </button>
          </VizTooltip>

          <VizTooltip
            overlay={
              <Popover id="list-view">
                <Popover.Content>List view</Popover.Content>
              </Popover>
            }
          >
            <button
              className={classNames(classes.filterButton, {
                [classes.activeButton]: selectedViewMode === 'list',
              })}
              onClick={() => setSelectedViewMode('list')}
              data-test-id="list-view"
            >
              <ListIcon />
            </button>
          </VizTooltip>
        </div>

        <Dropdown>
          <VizTooltip
            overlay={
              <Popover id="create-new-button-popover">
                <Popover.Content>
                  {areWorkspacesLoading
                    ? 'Loading...'
                    : !canCreateApp && !canCreateDashboard
                    ? "You don't have the necessary permissions"
                    : 'Create new report'}
                </Popover.Content>
              </Popover>
            }
          >
            <Dropdown.Toggle
              as={ClouderaButton}
              color="primary"
              icon={<ArrowDown />}
              iconPosition="right"
              disabled={
                areWorkspacesLoading || (!canCreateApp && !canCreateDashboard)
              }
              id="selected-app-group-actions"
              testId="create-new-button"
            >
              <div className="d-flex align-items-center">
                <span className="text-lg">
                  <Plus />
                </span>
                <span className={classNames(classes.hideOnSmallScreen, 'ml-1')}>
                  Create New
                </span>
              </div>
            </Dropdown.Toggle>
          </VizTooltip>

          <Dropdown.Menu>
            <VizTooltip
              overlay={
                <Popover id="create-new-dashboard-button-popover">
                  <Popover.Content>
                    {!canCreateDashboard &&
                      "You don't have the necessary permissions"}
                  </Popover.Content>
                </Popover>
              }
              disabled={canCreateDashboard}
            >
              <div>
                <Dropdown.Item
                  as="a"
                  className={classes.dropdownItem}
                  disabled={!canCreateDashboard}
                  href={`${arc.frontend.globals.urls.reportBuilderUrl}/0?report_type=dashboard`}
                  data-test-id="home-new-dashboard"
                >
                  <DashboardIcon />
                  <p>Dashboard</p>
                </Dropdown.Item>
              </div>
            </VizTooltip>

            <VizTooltip
              overlay={
                <Popover id="create-new-app-button-popover">
                  <Popover.Content>
                    {!canCreateApp &&
                      "You don't have the necessary permissions"}
                  </Popover.Content>
                </Popover>
              }
              disabled={canCreateApp}
            >
              <div>
                <Dropdown.Item
                  as="a"
                  className={classes.dropdownItem}
                  href={arc.frontend.globals.urls.appGroupEditUrl}
                  disabled={!canCreateApp}
                  data-test-id="new-app-btn"
                >
                  <AppIcon />
                  <p>App</p>
                </Dropdown.Item>
              </div>
            </VizTooltip>
          </Dropdown.Menu>
        </Dropdown>
      </div>
    </div>
  );
};
