import { useMutation, useQuery, useQueryClient } from '@tanstack/react-query';
import { Alert, Form, Table } from 'react-bootstrap';
import {
  getGroups,
  groupsQueryToken,
} from 'pages/visuals-page/apis/get-groups';
import { getUsers, usersQueryToken } from 'pages/visuals-page/apis/get-users';
import {
  Workspace,
  workspacesQueryToken,
} from 'pages/visuals-page/apis/get-workspaces';
import React, { useState } from 'react';
import { Modal } from 'react-bootstrap';
import Creatable from 'react-select/creatable';
import { upsertWorkspace } from 'pages/visuals-page/apis/upsert-workspace';
import { isAxiosError } from 'axios';
import classes from './WorkspaceDialog.module.scss';
import classNames from 'classnames';
import { getGlobalArc } from 'utils/get-global-arc';
import { useNotificationStore } from 'stores/use-notification-store';
import { components } from 'react-select';
import { ClouderaButton } from '../../../../components/cloudera-button/ClouderaButton';
import Close from '@cloudera-internal/cuix-core/icons/svg/ic-close.svg';
import Search from '@cloudera-internal/cuix-core/icons/svg/ic-search.svg';
import Delete from '@cloudera-internal/cuix-core/icons/svg/ic-delete.svg';
import Plus from '@cloudera-internal/cuix-core/icons/svg/ic-plus-circle.svg';
import User from '@cloudera-internal/cuix-core/icons/svg/ic-user.svg';
import Group from '@cloudera-internal/cuix-core/icons/svg/ic-groups.svg';
import Select from 'react-select';

const arc = getGlobalArc();

// Update from WSACL_Member and WSACL_Level class in workspace/model.py
const WS_ACL = {
  member: {
    valuesByLabel: {
      user: 1,
      group: 2,
    },
    labelsByValue: {
      1: 'user',
      2: 'group',
    } as const,
  },
  level: {
    valuesByLabel: {
      view: 1,
      edit: 2,
      manage: 3,
    },
    labelsByValue: {
      1: 'view',
      2: 'edit',
      3: 'manage',
    } as const,
  },
};

const permissionOptions = [
  { value: 'view', label: 'View Only' },
  { value: 'edit', label: 'Edit' },
  { value: 'manage', label: 'Manage' },
] as const;

type WorkspaceDialogProps = {
  isOpen: boolean;
  close: () => void;
  workspaces: { id: number; name: string }[];
  selectedWorkspace: {
    id: number;
    name: string;
    desc: string;
    acl?: Workspace['acl'];
  } | null;
  publicWorkspace:
    | {
        id: number;
        name: string;
        desc: string;
      }
    | undefined;
  setPublicWorkspace: (publicWorkspace: {
    id: number;
    name: string;
    desc: string;
    editable: boolean;
    perms: string[];
    acl: [[type: 1 | 2, permission: 1 | 2 | 3, name: string]];
  }) => void;
};

export const WorkspaceDialog = ({
  isOpen,
  close,
  selectedWorkspace,
  workspaces,
  publicWorkspace,
  setPublicWorkspace,
}: WorkspaceDialogProps): JSX.Element => {
  const queryClient = useQueryClient();

  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const {
    data: users,
    isLoading: areUsersLoading,
    error: usersError,
  } = useQuery({
    queryKey: [usersQueryToken],
    queryFn: () =>
      getUsers({
        isSummary: true,
      }),
    staleTime: Infinity,
  });

  const {
    data: groups,
    isLoading: areGroupsLoading,
    error: groupsError,
  } = useQuery({
    queryKey: [groupsQueryToken],
    queryFn: () =>
      getGroups({
        isSummary: true,
      }),
    staleTime: Infinity,
  });

  const isPublic =
    publicWorkspace && selectedWorkspace?.id === publicWorkspace.id;

  const [name, setName] = useState(selectedWorkspace?.name ?? '');
  const [description, setDescription] = useState(selectedWorkspace?.desc ?? '');

  const [selectedUsers, setSelectedUsers] = useState<
    { value: string; label: string }[]
  >([]);
  const [selectedGroups, setSelectedGroups] = useState<
    { value: string; label: string }[]
  >([]);
  const [selectedPermission, setSelectedPermission] =
    useState<keyof (typeof WS_ACL)['level']['valuesByLabel']>('view');

  const [addedUsersAndGroups, setAddedUsersAndGroups] = useState<
    {
      permission: keyof (typeof WS_ACL)['level']['valuesByLabel'];
      type: keyof (typeof WS_ACL)['member']['valuesByLabel'];
      name: string;
    }[]
  >(
    selectedWorkspace?.acl
      ? selectedWorkspace.acl.map(([type, permission, name]) => ({
          name,
          type: WS_ACL.member.labelsByValue[type],
          permission: WS_ACL.level.labelsByValue[permission],
        }))
      : [
          {
            name: arc.frontend.globals.authentication.username,
            type: 'user',
            permission: 'manage',
          },
        ]
  );

  const [errorMessage, setErrorMessage] = useState<string | null>();

  const { mutateAsync, isLoading } = useMutation({
    mutationFn: upsertWorkspace,
    onMutate: async () => {
      await queryClient.cancelQueries({
        queryKey: [workspacesQueryToken],
      });
    },
    onSuccess: ([workspace]) => {
      const workspaces = queryClient.getQueryData<Workspace[]>([
        workspacesQueryToken,
      ]);

      let newWorkspaces = workspaces ? [...workspaces] : [];

      const isEditMode = !!selectedWorkspace;
      if (isEditMode) {
        if (publicWorkspace?.id === selectedWorkspace.id) {
          setPublicWorkspace(workspace);
        } else {
          newWorkspaces = newWorkspaces.map((currentWorkspace) =>
            workspace.id === currentWorkspace.id ? workspace : currentWorkspace
          );
        }
      } else {
        newWorkspaces.push(workspace);
      }

      queryClient.setQueryData([workspacesQueryToken], newWorkspaces);

      addNotification(
        {
          message: selectedWorkspace
            ? 'Successfully edited workspace.'
            : 'Successfully created workspace',
          variant: 'success',
        },
        {
          delay: 2000,
        }
      );
    },
    onError: () => {
      addNotification({
        message: 'Something went wrong.',
        variant: 'danger',
      });
    },
  });

  const onSubmit = async () => {
    let error: string | null = null;

    if (!name) {
      error = 'No workspace name specified';
    }

    const isNameDuplicate = workspaces.some((workspace) => {
      const hasSameId =
        selectedWorkspace && workspace.id === selectedWorkspace.id;
      const hasSameName = workspace.name === name;

      return !hasSameId && hasSameName;
    });

    if (isNameDuplicate) {
      error = 'Duplicate workspace name not allowed';
    }

    if (error) {
      setErrorMessage(error);
      return;
    }

    const acl: [number, number, string][] = addedUsersAndGroups.map(
      ({ name, permission, type }) => [
        WS_ACL.member.valuesByLabel[type],
        WS_ACL.level.valuesByLabel[permission],
        name,
      ]
    );

    try {
      const data: Parameters<typeof upsertWorkspace>[0] = {
        name: isPublic ? publicWorkspace.name : name,
        desc: isPublic ? publicWorkspace.desc : description,
        acl,
      };

      if (selectedWorkspace) {
        data.id = selectedWorkspace.id;
      }

      await mutateAsync(data);

      close();
    } catch (err) {
      if (isAxiosError(err)) {
        error = err.message;
      }
    }
  };

  return (
    <Modal show={isOpen} onHide={close} size="lg">
      <Modal.Header className="align-items-center border-bottom-0">
        <h3
          className="text-lg font-light text-dark m-0"
          style={{
            height: 28,
            lineHeight: 1.4,
          }}
        >
          {selectedWorkspace ? 'Edit' : 'Add'} Workspace
        </h3>

        <button
          className="outline-none bg-transparent border-0 p-2 text-dark-gray text-lg"
          onClick={close}
        >
          <Close />
        </button>
      </Modal.Header>

      <Modal.Body>
        {!!groupsError && (
          <Alert variant="warning">
            Failed to load groups for autocomplete and dropdown selection; but,
            you can still add groups by name
          </Alert>
        )}

        {!!usersError && (
          <Alert variant="warning">
            Failed to load users for autocomplete and dropdown selection; but,
            you can still add users by name
          </Alert>
        )}

        <Form.Group controlId="name">
          <Form.Label>
            <span className="d-block text-right text-dark-gray">Name</span>
          </Form.Label>

          <Form.Control
            placeholder="enter workspace name"
            value={name}
            onChange={(event) => {
              setName(event.target.value);
            }}
            isInvalid={!!errorMessage}
            disabled={isPublic}
            data-test-id="workspace-name-input"
          />

          <Form.Control.Feedback type="invalid">
            {errorMessage}
          </Form.Control.Feedback>
        </Form.Group>

        <Form.Group controlId="description">
          <Form.Label>
            <span className="d-block text-right text-dark-gray">
              Description
            </span>
          </Form.Label>

          <Form.Control
            as="textarea"
            placeholder="enter workspace description"
            value={description}
            onChange={(event) => {
              setDescription(event.target.value);
            }}
            disabled={isPublic}
            data-test-id="workspace-desc-input"
          />
        </Form.Group>

        <h4 className="text-base font-medium text-dark">
          Add User(s) or Group(s)
        </h4>

        <div className="d-flex justify-content-between align-items-end">
          <div className="d-flex gap-3 flex-wrap">
            <Form.Group controlId="users-search-field">
              <Form.Label>
                <span className="d-block text-right text-dark-gray">
                  Search for User(s)
                </span>
              </Form.Label>

              <Creatable
                styles={{
                  container: (baseStyles) => ({ ...baseStyles, width: 200 }),
                }}
                isMulti
                name="users"
                options={
                  users?.map(({ id, username }) => ({
                    value: id.toString(),
                    label: username,
                  })) ?? []
                }
                isClearable={false}
                placeholder={
                  <>
                    <span className="text-base">
                      <Search />{' '}
                    </span>
                    Search
                  </>
                }
                value={selectedUsers}
                onChange={(newValue) => {
                  if (Array.isArray(newValue)) {
                    setSelectedUsers(newValue);
                  }
                }}
                isLoading={areUsersLoading}
                components={{
                  Input: (props) => (
                    <components.Input
                      {...props}
                      data-test-id="users-search-field"
                    />
                  ),
                }}
              />
            </Form.Group>

            <Form.Group controlId="groups-search-field">
              <Form.Label>
                <span className="d-block text-right text-dark-gray">
                  Search for Group(s)
                </span>
              </Form.Label>

              <Creatable
                styles={{
                  container: (baseStyles) => ({ ...baseStyles, width: 200 }),
                }}
                isMulti
                name="groups"
                options={[
                  { value: 'Everyone', label: 'Everyone' },
                  ...(groups?.map(({ id, name }) => ({
                    value: id.toString(),
                    label: name,
                  })) ?? []),
                ]}
                isClearable={false}
                placeholder={
                  <>
                    <span className="text-base">
                      <Search />{' '}
                    </span>
                    Search
                  </>
                }
                value={selectedGroups}
                onChange={(newValue) => {
                  if (Array.isArray(newValue)) {
                    setSelectedGroups(newValue);
                  }
                }}
                isLoading={areGroupsLoading}
                components={{
                  Input: (props) => (
                    <components.Input
                      {...props}
                      data-test-id="groups-search-field"
                    />
                  ),
                }}
              />
            </Form.Group>

            <Form.Group controlId="permission">
              <Form.Label>
                <span className="d-block text-right text-dark-gray">
                  Permission
                </span>
              </Form.Label>

              <Select
                options={permissionOptions}
                value={permissionOptions.find(
                  ({ value }) => value === selectedPermission
                )}
                className={classNames(classes.select)}
                data-test-id="permission"
                onChange={(event) => {
                  const value = event?.value;

                  if (value) {
                    setSelectedPermission(value);
                  }
                }}
              />
            </Form.Group>
          </div>

          <div style={{ marginBottom: '1rem' }}>
            <ClouderaButton
              onClick={() => {
                setAddedUsersAndGroups((current) => {
                  const newUsersAndGroups = [...current];

                  selectedUsers.forEach(({ label }) => {
                    const isAlreadyAdded = newUsersAndGroups.some(
                      ({ name }) => name === label
                    );

                    if (isAlreadyAdded) {
                      return;
                    }

                    newUsersAndGroups.push({
                      name: label,
                      permission: selectedPermission,
                      type: 'user',
                    });
                  });

                  selectedGroups.forEach(({ label }) => {
                    const isAlreadyAdded = newUsersAndGroups.some(
                      ({ name }) => name === label
                    );

                    if (isAlreadyAdded) {
                      return;
                    }

                    newUsersAndGroups.push({
                      name: label,
                      permission: selectedPermission,
                      type: 'group',
                    });
                  });

                  return newUsersAndGroups;
                });
                setSelectedGroups([]);
                setSelectedUsers([]);
              }}
              testId="add-users-and-groups-button"
              icon={<Plus />}
            >
              Add
            </ClouderaButton>
          </div>
        </div>

        <Table
          bordered
          style={{
            borderSpacing: 0,
            borderCollapse: 'separate',
            borderRadius: 3,
          }}
        >
          <thead style={{ backgroundColor: '#fafbfc' }}>
            <tr>
              <th
                className={classNames(
                  'border-left-0 border-right-0 border-top-0',
                  classes.th
                )}
              >
                Name
              </th>
              <th
                className={classNames(
                  'border-left-0 border-right-0 border-top-0',
                  classes.th
                )}
              >
                Type
              </th>
              <th
                className={classNames(
                  'border-left-0 border-right-0 border-top-0',
                  classes.th
                )}
              >
                Permission
              </th>
              <th
                className={classNames(
                  'border-left-0 border-right-0 border-top-0',
                  classes.th
                )}
              />
            </tr>
          </thead>

          <tbody>
            {addedUsersAndGroups.map(({ name, permission, type }, index) => (
              <tr key={name}>
                <td
                  className={classNames(
                    'border-left-0 border-right-0 border-top-0 align-middle',
                    classes.td
                  )}
                >
                  {name}
                </td>

                <td
                  className={classNames(
                    'border-left-0 border-right-0 border-top-0 align-middle',
                    classes.td
                  )}
                >
                  <span className="text-base">
                    {type === 'user' && <User />}
                    {type === 'group' && <Group />}
                  </span>{' '}
                  {type}
                </td>

                <td className="border-left-0 border-right-0 border-top-0">
                  <Select
                    options={permissionOptions}
                    value={permissionOptions.find(
                      ({ value }) => value === permission
                    )}
                    className={classes.select}
                    onChange={(event) => {
                      const newPermission = event?.value;

                      if (newPermission) {
                        setAddedUsersAndGroups((current) =>
                          current.map((value, i) =>
                            i === index
                              ? { ...value, permission: newPermission }
                              : value
                          )
                        );
                      }
                    }}
                  />
                </td>

                <td
                  className={classNames(
                    'border-left-0 border-right-0 border-top-0 align-middle',
                    classes.td
                  )}
                >
                  <span
                    className={classNames('p-2', classes.removeIcon)}
                    onClick={() => {
                      setAddedUsersAndGroups((current) =>
                        current.filter((_, i) => i !== index)
                      );
                    }}
                  >
                    <Delete />
                  </span>
                </td>
              </tr>
            ))}
          </tbody>
        </Table>
      </Modal.Body>

      <Modal.Footer className="justify-content-end">
        <ClouderaButton onClick={close} testId="close-workspace-dialog-button">
          Cancel
        </ClouderaButton>

        <ClouderaButton
          color="primary"
          disabled={isLoading}
          onClick={onSubmit}
          testId="save-workspace-button"
        >
          Save
        </ClouderaButton>
      </Modal.Footer>
    </Modal>
  );
};
