import { useQuery } from '@tanstack/react-query';
import { appsQueryToken, getApps } from 'apis/get-apps';
import { useMemo } from 'react';
import { getGlobalArc } from 'utils/get-global-arc';
import { getReports, reportsQueryToken } from 'apis/get-reports';
import { CombinedReport } from 'types/combined-report.type';

const arc = getGlobalArc();

export const useApps = () => {
  const { data: apps } = useQuery({
    queryKey: [appsQueryToken],
    queryFn: getApps,
  });

  const firstDashboardIdOfAppToAppId = useMemo(() => {
    const map = new Map<number, number>();

    if (apps) {
      apps.forEach((app) => {
        if (app.group_type !== 'default') {
          return;
        }

        const sortedSubgroups =
          app.subgroups?.sort((a, b) => {
            const aKey = a.subgroup_order ?? 0;
            const bKey = b.subgroup_order ?? 0;

            if (aKey > bKey) {
              return 1;
            }

            if (aKey < bKey) {
              return -1;
            }

            return 0;
          }) ?? [];

        const subgroupApps = sortedSubgroups.reduce(
          (total, current) => total.concat(current.apps),
          [] as [number, number][]
        );

        const dashboardIds = Array.from(
          new Set(app.apps.concat(subgroupApps).map(([id]) => id))
        );

        const [firstDashboardIdOfApp] = dashboardIds;

        map.set(firstDashboardIdOfApp, app.group_id);
      });
    }

    return map;
  }, [apps]);

  const dashboardIds = useMemo(
    () => [...firstDashboardIdOfAppToAppId.keys()].filter(Boolean),
    [firstDashboardIdOfAppToAppId]
  );

  const { data: dashboards } = useQuery({
    queryKey: [reportsQueryToken, 'dashobard-ids', ...dashboardIds],
    enabled: dashboardIds.length > 0,
    queryFn: () =>
      getReports({
        reportType: 'dashboard',
        reportIds: dashboardIds,
      }),
  });

  const appIdToCurrentDashboardBaseImg = useMemo(() => {
    const map = new Map<number, string>();

    if (dashboards) {
      dashboards.forEach(({ id, baseImg }) => {
        const appId = firstDashboardIdOfAppToAppId.get(id);

        if (appId) {
          map.set(appId, baseImg);
        }
      });
    }

    return map;
  }, [firstDashboardIdOfAppToAppId, dashboards]);

  const samplesReportIds = useMemo(() => {
    const samplesIds = {
      apps: new Set<number>(),
      dashboards: new Set<number>(),
    };

    const sampleAppAppGroup = apps?.find(
      ({ group_type }) => group_type === 'sample-app'
    );

    if (sampleAppAppGroup) {
      samplesIds.apps.add(sampleAppAppGroup.group_id);
    }

    const samplesAppGroup = apps?.find(
      ({ group_type }) => group_type === 'samples'
    );

    if (samplesAppGroup) {
      samplesAppGroup.apps.forEach(([id]) => {
        samplesIds.dashboards.add(id);
      });

      samplesAppGroup.subgroups?.forEach(({ apps }) =>
        apps.forEach(([id]) => samplesIds.dashboards.add(id))
      );
    }

    return samplesIds;
  }, [apps]);

  const favoriteReportIds = useMemo(() => {
    const favoriteIdSet = new Set<number>();

    const favoriteAppGroup = apps?.find(
      ({ group_type }) => group_type === 'starred'
    );

    if (favoriteAppGroup) {
      favoriteAppGroup.apps.forEach(([id]) => {
        favoriteIdSet.add(id);
      });

      favoriteAppGroup.subgroups?.forEach(({ apps }) =>
        apps.forEach(([id]) => favoriteIdSet.add(id))
      );
    }

    return favoriteIdSet;
  }, [apps]);

  const appsAsCombinedReports = useMemo(() => {
    const result: CombinedReport[] = [];

    apps?.forEach(
      ({
        group_id,
        name,
        baseImg,
        perm,
        updated,
        update_user,
        workspace_id,
        workspace_name,
        rw,
        created,
        group_type,
        create_user,
      }) => {
        if (group_type !== 'default' && group_type !== 'sample-app') {
          return;
        }

        const currentDashboadBaseImg =
          appIdToCurrentDashboardBaseImg.get(group_id);

        result.push({
          id: group_id,
          baseImg,
          editUrl: `${arc.frontend.globals.urls.editAppUrl}/${group_id}`,
          img: currentDashboadBaseImg ?? baseImg,
          permissions: perm,
          title: name,
          type: 'app',
          updated: {
            date: updated,
            by: update_user,
          },
          isClone: false,
          numberOfDescendants: 0,
          url: `${arc.frontend.globals.urls.viewAppUrl}/${group_id}`,
          workspace: {
            id: workspace_id,
            name: workspace_name,
          },
          rw,
          created: {
            date: created,
            by: create_user,
          },
        });
      }
    );

    return result;
  }, [appIdToCurrentDashboardBaseImg, apps]);

  return {
    samplesReportIds,
    favoriteReportIds,
    appsAsCombinedReports,
  };
};
