import { useMemo } from 'react';
import { getGlobalArc } from 'utils/get-global-arc';
import { useQuery } from '@tanstack/react-query';

import { getReports, reportsQueryToken } from 'apis/get-reports';
import { CombinedReport } from 'types/combined-report.type';
import { ReportFilterType } from 'types/report-filter-type.type';
import { useNotificationStore } from 'stores/use-notification-store';
import { StaticSidebarItems } from '../VisualsPage';

const arc = getGlobalArc();

type GroupId = number | keyof Omit<StaticSidebarItems, 'private' | 'public'>;

const groupProcessedReport = ({
  combinedReport,
  relatedReportIds,
  groupIdToReportTypeToCombinedReports,
  favoriteReportIds,
  samplesReportIds,
}: {
  combinedReport: CombinedReport;
  relatedReportIds: Set<string>;
  samplesReportIds: { apps: Set<number>; dashboards: Set<number> };
  favoriteReportIds: Set<number>;
  groupIdToReportTypeToCombinedReports: Map<
    GroupId,
    Map<ReportFilterType, CombinedReport[]>
  >;
}) => {
  combinedReport.sharedVisualDashboardIds?.forEach((id) =>
    relatedReportIds.add(id)
  );

  let reportType: ReportFilterType | undefined;

  if (combinedReport.type === 'app') {
    reportType = 'apps';
  } else if (combinedReport.type === 'dashboard') {
    reportType = 'dashboards';
  }

  // All app group
  if (!groupIdToReportTypeToCombinedReports.has('all')) {
    groupIdToReportTypeToCombinedReports.set(
      'all',
      new Map<ReportFilterType, CombinedReport[]>([
        ['all', []],
        ['dashboards', []],
        ['apps', []],
      ])
    );
  }

  const allAppGroup = groupIdToReportTypeToCombinedReports.get('all');
  if (allAppGroup) {
    //   all type
    const allAppGroupAllType = allAppGroup.get('all');
    allAppGroupAllType?.push(combinedReport);

    //   report type
    if (reportType) {
      const allAppGroupReportType = allAppGroup.get(reportType);

      allAppGroupReportType?.push(combinedReport);
    }
  }

  // Workspace app group
  if (!groupIdToReportTypeToCombinedReports.has(combinedReport.workspace.id)) {
    groupIdToReportTypeToCombinedReports.set(
      combinedReport.workspace.id,
      new Map<ReportFilterType, CombinedReport[]>([
        ['all', []],
        ['dashboards', []],
        ['apps', []],
      ])
    );
  }
  const workspaceAppGroup = groupIdToReportTypeToCombinedReports.get(
    combinedReport.workspace.id
  );
  if (workspaceAppGroup) {
    //   all type
    const workspaceAppGroupAllType = workspaceAppGroup.get('all');
    workspaceAppGroupAllType?.push(combinedReport);

    //   report type
    if (reportType) {
      const workspaceAppGroupReportType = workspaceAppGroup.get(reportType);

      workspaceAppGroupReportType?.push(combinedReport);
    }
  }

  // Samples app group
  if (
    combinedReport.type === 'app'
      ? samplesReportIds.apps.has(combinedReport.id)
      : samplesReportIds.dashboards.has(combinedReport.id)
  ) {
    if (!groupIdToReportTypeToCombinedReports.has('samples')) {
      groupIdToReportTypeToCombinedReports.set(
        'samples',
        new Map<ReportFilterType, CombinedReport[]>([
          ['all', []],
          ['dashboards', []],
          ['apps', []],
        ])
      );
    }

    const samplesAppGroup = groupIdToReportTypeToCombinedReports.get('samples');
    if (samplesAppGroup) {
      //   all type
      const samplesAppGroupAllType = samplesAppGroup.get('all');
      samplesAppGroupAllType?.push(combinedReport);

      //   report type
      if (reportType) {
        const samplesAppGroupReportType = samplesAppGroup.get(reportType);

        samplesAppGroupReportType?.push(combinedReport);
      }
    }
  }

  // Favorites app group
  if (
    combinedReport.type !== 'app' &&
    favoriteReportIds.has(combinedReport.id)
  ) {
    if (!groupIdToReportTypeToCombinedReports.has('favorites')) {
      groupIdToReportTypeToCombinedReports.set(
        'favorites',
        new Map<ReportFilterType, CombinedReport[]>([
          ['all', []],
          ['dashboards', []],
          ['apps', []],
        ])
      );
    }
    const favoritesAppGroup =
      groupIdToReportTypeToCombinedReports.get('favorites');
    if (favoritesAppGroup) {
      //   all type
      const favoritesAllType = favoritesAppGroup.get('all');
      favoritesAllType?.push(combinedReport);

      //   report type
      if (reportType) {
        const favoritesReportType = favoritesAppGroup.get(reportType);

        favoritesReportType?.push(combinedReport);
      }
    }
  }
};

export const useReports = ({
  selectedGroup,
  appsAsCombinedReports,
  samplesReportIds,
  favoriteReportIds,
  staticSideBarItems,
}: {
  selectedGroup: { id: number } | null;
  appsAsCombinedReports: CombinedReport[];
  samplesReportIds: { apps: Set<number>; dashboards: Set<number> };
  favoriteReportIds: Set<number>;
  staticSideBarItems: Pick<StaticSidebarItems, 'all' | 'samples' | 'favorites'>;
}) => {
  const addNotification = useNotificationStore(
    ({ addNotification }) => addNotification
  );

  const { data: reports, fetchStatus } = useQuery({
    queryKey: [reportsQueryToken, selectedGroup?.id],
    queryFn: async () => {
      let workspaceId: number | undefined = undefined;

      if (selectedGroup) {
        if (selectedGroup.id !== staticSideBarItems.all.id) {
          workspaceId = selectedGroup.id;
        }
      } else {
        return [];
      }

      return getReports({
        shouldGetCloneInfo: arc.frontend.featureFlipper.getValueById(
          'enable_clone_info_badge'
        ),
        workspaceId,
      }).catch((error) => {
        console.error(error);

        addNotification({
          message: 'Something went wrong',
          variant: 'danger',
        });
      });
    },
    keepPreviousData: true,
    staleTime: Infinity,
  });

  const { groupIdToReportTypeToCombinedReports, relatedReportIds } =
    useMemo(() => {
      const relatedReportIds = new Set<string>();

      const groupIdToReportTypeToCombinedReports = new Map<
        GroupId,
        Map<ReportFilterType, CombinedReport[]>
      >();

      appsAsCombinedReports.forEach((combinedReport) =>
        groupProcessedReport({
          combinedReport,
          relatedReportIds,
          favoriteReportIds,
          samplesReportIds,
          groupIdToReportTypeToCombinedReports,
        })
      );

      reports?.forEach(
        ({
          id,
          baseImg,
          edit_url,
          img,
          title,
          report_type,
          perm,
          url,
          workspace_id,
          workspace_name,
          updated,
          updated_by,
          dataset_id,
          dataconn_name,
          dataset_name,
          shared_visual_dashboards,
          is_clone,
          descendants_count,
          created,
          created_by,
          report_tags,
          dataset_detail,
        }) => {
          const areVisualsEnabled =
            !arc.frontend.featureFlipper.getValueById(
              'disable_visuals_for_workflows'
            ) && shared_visual_dashboards;

          if (report_type !== 'dashboard' && !areVisualsEnabled) {
            return;
          }

          let sharedVisualDashboardIds: string[] = [];

          if (shared_visual_dashboards) {
            try {
              const dashboardIds = JSON.parse(shared_visual_dashboards);

              if (
                Array.isArray(dashboardIds) &&
                dashboardIds.every((id) => typeof id === 'number')
              ) {
                sharedVisualDashboardIds = dashboardIds;
              }
            } catch (error) {
              console.error(error);
            }
          }

          const combinedReport: CombinedReport = {
            id,
            baseImg,
            editUrl: edit_url,
            img,
            permissions: perm,
            title,
            type: report_type,
            updated: {
              date: updated,
              by: updated_by,
            },
            url,
            workspace: {
              id: workspace_id,
              name: workspace_name,
            },
            isClone: is_clone,
            numberOfDescendants:
              descendants_count === 'n/a' ? 0 : Number(descendants_count || 0),
            datasetInfo: {
              id: dataset_id,
              name: dataset_name,
              dataConnectionName: dataconn_name,
              detail: dataset_detail,
            },
            sharedVisualDashboardIds,
            created: {
              date: created,
              by: created_by,
            },
            tags: report_tags,
          };

          groupProcessedReport({
            combinedReport,
            relatedReportIds,
            favoriteReportIds,
            samplesReportIds,
            groupIdToReportTypeToCombinedReports,
          });
        }
      );

      return {
        relatedReportIds,
        groupIdToReportTypeToCombinedReports,
      };
    }, [favoriteReportIds, appsAsCombinedReports, reports, samplesReportIds]);

  const { data: relatedReports } = useQuery({
    queryKey: [reportsQueryToken, 'related-reports', ...relatedReportIds],
    queryFn: () =>
      getReports({
        reportIds: [...relatedReportIds],
      }),
    enabled: relatedReportIds.size !== 0,
    keepPreviousData: true,
    staleTime: Infinity,
  });

  const relatedReportIdToReportMap = useMemo(() => {
    const map = new Map<
      number,
      {
        id: number;
        title: string;
        sharedVisualDashboardIds: number[];
        workspace: { id: number; name: string };
      }
    >();

    relatedReports?.forEach((report) => {
      let sharedVisualDashboardIds: number[] = [];

      if (report.shared_visual_dashboards) {
        try {
          const dashboardIds = JSON.parse(report.shared_visual_dashboards);

          if (
            Array.isArray(dashboardIds) &&
            dashboardIds.every((id) => typeof id === 'number')
          ) {
            sharedVisualDashboardIds = dashboardIds;
          }
        } catch (error) {
          console.error(error);
        }
      }

      map.set(report.id, {
        id: report.id,
        title: report.title,
        sharedVisualDashboardIds,
        workspace: { id: report.workspace_id, name: report.workspace_name },
      });
    });

    return map;
  }, [relatedReports]);

  return {
    areReportsLoading: fetchStatus === 'fetching',
    relatedReportIdToReportMap,
    groupIdToReportTypeToCombinedReports,
    hasReports: (workspaceId: number) =>
      !!groupIdToReportTypeToCombinedReports.get(workspaceId)?.get('all')
        ?.length,
  };
};
