import { useCallback, useEffect, useMemo, useState } from 'react';
import { vizCookie } from 'services/viz-cookie';
import { getGlobalArc } from 'utils/get-global-arc';
import { Workspace } from '../apis/get-workspaces';
import { StaticSidebarItems } from '../VisualsPage';

type AllGroup = {
  type: 'all';
  id: number;
  name: string;
  desc: string;
};

type AppGroup = {
  type: 'app-group';
  id: number;
  name: string;
  desc: string;
};

type WorkspaceGroup = {
  type: 'workspace';
  id: number;
  name: string;
  desc: string;
  editable: boolean;
  perms: string[];
  acl: [[type: 1 | 2, permission: 1 | 2 | 3, name: string]];
};

export type Group = AllGroup | AppGroup | WorkspaceGroup;

const arc = getGlobalArc();

const params = new URLSearchParams(window.location.search);

const cookieName = 'appgroupid' as const;

export const useSelectedGroup = ({
  staticSideBarItems,
  groupedWorkspaces,
  publicWorkspace,
}: {
  staticSideBarItems: StaticSidebarItems;
  groupedWorkspaces:
    | {
        sharedWorkspaces: Workspace[];
        privateWorkspaces: Workspace[];
      }
    | undefined;
  publicWorkspace: StaticSidebarItems['public'];
}) => {
  const [selectedGroup, setSelectedGroup] = useState<Group | null>(null);

  const selectedGroupSetter = useCallback((group: Group) => {
    setSelectedGroup(group);

    vizCookie.setVisualTabSettingByKey(cookieName, group.id);
  }, []);

  const toggleFavoriteGroup = useCallback(() => {
    setSelectedGroup((current) => {
      if (current?.id === staticSideBarItems.favorites.id) {
        vizCookie.setVisualTabSettingByKey(
          cookieName,
          staticSideBarItems.all.id
        );
        return { ...staticSideBarItems.all, type: 'all' };
      }

      vizCookie.setVisualTabSettingByKey(
        cookieName,
        staticSideBarItems.favorites.id
      );

      return { ...staticSideBarItems.favorites, type: 'app-group' };
    });
  }, [staticSideBarItems.all, staticSideBarItems.favorites]);

  const privateWorkspace = useMemo(
    () => ({
      ...staticSideBarItems.private,
      type: 'workspace' as const,
    }),
    [staticSideBarItems.private]
  );

  const workspaces = useMemo(() => {
    const combinedWorkspaces = [
      privateWorkspace,
      ...(groupedWorkspaces?.sharedWorkspaces ?? []),
      ...(groupedWorkspaces?.privateWorkspaces ?? []),
    ].map((workspace) => ({
      ...workspace,
      type: 'workspace' as const,
    }));

    if (publicWorkspace) {
      combinedWorkspaces.push({
        ...publicWorkspace,
        type: 'workspace' as const,
      });
    }

    return combinedWorkspaces;
  }, [
    groupedWorkspaces?.privateWorkspaces,
    groupedWorkspaces?.sharedWorkspaces,
    privateWorkspace,
    publicWorkspace,
  ]);

  useEffect(() => {
    if (selectedGroup === null && groupedWorkspaces) {
      if (arc.frontend.featureFlipper.getValueById('hide_workspaces')) {
        setSelectedGroup({ ...staticSideBarItems.all, type: 'all' });
        return;
      }

      let group: Group | null = null;

      const groupNameFromUrl = params.get('group');
      const groupIdFromCookies = vizCookie.getVisualTabSettingByKey(cookieName);

      if (groupNameFromUrl) {
        const groups: Group[] = [
          { ...staticSideBarItems.all, type: 'all' },
          { ...staticSideBarItems.favorites, type: 'app-group' },
          ...workspaces,
        ];

        if (staticSideBarItems.samples) {
          groups.push({ ...staticSideBarItems.samples, type: 'app-group' });
        }

        const groupFromUrl = groups.find(
          ({ name }) => name.toLowerCase() === groupNameFromUrl.toLowerCase()
        );

        if (groupFromUrl) {
          group = groupFromUrl;
        }
      }

      if (group === null && groupIdFromCookies) {
        if (groupIdFromCookies === staticSideBarItems.all.id) {
          group = { ...staticSideBarItems.all, type: 'all' };
        } else if (groupIdFromCookies === staticSideBarItems.favorites.id) {
          group = { ...staticSideBarItems.favorites, type: 'app-group' };
        } else if (
          staticSideBarItems.samples &&
          groupIdFromCookies === staticSideBarItems.samples.id
        ) {
          group = { ...staticSideBarItems.samples, type: 'app-group' };
        } else {
          const workspace = workspaces.find(
            ({ id }) => id === groupIdFromCookies
          );

          if (workspace) {
            group = {
              id: workspace.id,
              name: workspace.name,
              desc: workspace.desc,
              editable: workspace.editable ?? false,
              perms: workspace.perms,
              acl: workspace.acl,
              type: workspace.type,
            };
          }
        }
      }

      if (group === null) {
        group = privateWorkspace;
      }

      setSelectedGroup(group);
    }
  }, [
    groupedWorkspaces,
    privateWorkspace,
    selectedGroup,
    staticSideBarItems.all,
    staticSideBarItems.favorites,
    staticSideBarItems.private,
    publicWorkspace,
    staticSideBarItems.samples,
    workspaces,
  ]);

  useEffect(() => {
    if (groupedWorkspaces) {
      setSelectedGroup((current) => {
        if (current?.type !== 'workspace') {
          return current;
        }

        return workspaces.find(({ id }) => current.id === id) ?? null;
      });
    }
  }, [groupedWorkspaces, workspaces]);

  return {
    selectedGroup,
    setSelectedGroup: selectedGroupSetter,
    toggleFavoriteGroup,
  };
};
