import { createStore } from 'stores/create-store';
import { CombinedReport } from 'types/combined-report.type';

type SelectedReports = {
  groupSelectedReports: CombinedReport[] | null;
  groupActionType: 'selection' | 'export' | 'delete' | 'move';

  selectedReport: CombinedReport | null;
  actionType: 'export' | 'delete' | 'move' | null;

  selectReport: (report: CombinedReport) => void;
  toggleReport: (report: CombinedReport) => void;
  unselectGroup: (report: { id: number; type: CombinedReport['type'] }) => void;
  setGroupActionToDelete: () => void;
  setGroupActionToMove: () => void;
  setGroupActionToExport: () => void;
  resetGroupAction: () => void;
  getNumberOfGroupSelectedReports: () => number;
  isReportSelected: (report: {
    id: number;
    type: CombinedReport['type'];
  }) => boolean;

  selectToDelete: (report: CombinedReport) => void;
  selectToMove: (report: CombinedReport) => void;
  selectToExport: (report: CombinedReport) => void;
  unselect: () => void;

  clear: () => void;
};

const getAreSameReports = (
  a: {
    id: number;
    type: CombinedReport['type'];
  },
  b: {
    id: number;
    type: CombinedReport['type'];
  }
) => a.type === b.type && a.id === b.id;

export const useSelectedReportsStore = createStore<SelectedReports>(
  'visuals-page-selected-reports',
  (set, get) => ({
    groupSelectedReports: null,
    groupActionType: 'selection',

    selectedReport: null,
    actionType: null,

    selectReport(report: CombinedReport) {
      set((state) => ({
        ...state,
        groupSelectedReports: state.groupSelectedReports
          ? [...state.groupSelectedReports, report]
          : [report],
      }));
    },
    toggleReport(report: CombinedReport) {
      const { groupSelectedReports } = get();

      if (!groupSelectedReports) {
        set({ groupSelectedReports: [report] });
        return;
      }

      if (
        groupSelectedReports.some((currentReport) =>
          getAreSameReports(report, currentReport)
        )
      ) {
        set({
          groupSelectedReports: groupSelectedReports.filter(
            (currentReport) => !getAreSameReports(report, currentReport)
          ),
        });
      } else {
        set({ groupSelectedReports: [...groupSelectedReports, report] });
      }
    },
    unselectGroup(report: { id: number; type: CombinedReport['type'] }) {
      set((state) => ({
        ...state,
        groupSelectedReports: state.groupSelectedReports
          ? [
              ...state.groupSelectedReports.filter(
                (currentReport) => !getAreSameReports(report, currentReport)
              ),
            ]
          : null,
      }));
    },
    setGroupActionToExport() {
      set({
        groupActionType: 'export',
      });
    },
    setGroupActionToDelete() {
      set({
        groupActionType: 'delete',
      });
    },
    setGroupActionToMove() {
      set({
        groupActionType: 'move',
      });
    },
    resetGroupAction() {
      set({ groupActionType: 'selection' });
    },
    isReportSelected(report) {
      return !!get().groupSelectedReports?.some((currentReport) =>
        getAreSameReports(report, currentReport)
      );
    },
    getNumberOfGroupSelectedReports() {
      return get().groupSelectedReports?.length ?? 0;
    },

    selectToDelete(report: CombinedReport) {
      set({
        selectedReport: report,
        actionType: 'delete',
      });
    },
    selectToMove(report: CombinedReport) {
      set({
        selectedReport: report,
        actionType: 'move',
      });
    },
    selectToExport(report: CombinedReport) {
      set({
        selectedReport: report,
        actionType: 'export',
      });
    },
    unselect() {
      set((state) => ({
        ...state,
        selectedReport: null,
        actionType: null,
      }));
    },

    clear() {
      set({
        groupSelectedReports: null,
        groupActionType: 'selection',

        selectedReport: null,
        actionType: null,
      });
    },
  })
);
