import { get } from 'lodash-es';
import { CombinedReport } from 'types/combined-report.type';

const searchKeys = [
  'id',
  'type',
  'title',
  'sharedVisualDashboardIds',
  'tags',
  'datasetInfo.name',
  'datasetInfo.detail',
  'updated.by',
  'created.by',
];

const tokenizeSearchTerm = (searchTerm: string) => {
  const include: string[] = [];
  const exclude: string[] = [];

  let word = '';
  let quoteEnclosed = false;
  let quoteType: null | '"' | "'" = null;
  let negated = false;

  const reset = () => {
    (negated ? exclude : include).push(word);
    word = '';
    negated = false;
    quoteEnclosed = false;
    quoteType = null;
  };

  for (let idx = 0; idx < searchTerm.length; idx++) {
    const [prev, curr, next] = [-1, 0, 1].map((n) => searchTerm[idx + n] || '');

    if (!word) {
      if (curr === ' ') {
        continue;
      }

      if (curr === '"' || curr === "'") {
        quoteEnclosed = true;
        quoteType = curr;
        continue;
      }

      if (curr === '!' && !quoteEnclosed && !negated) {
        negated = true;
        continue;
      }
    }

    if (word) {
      if (curr === '\\' && quoteEnclosed && next === quoteType) {
        continue;
      }

      if (curr === quoteType && prev !== '\\') {
        reset();
        continue;
      }

      if (curr === ' ' && !quoteEnclosed) {
        reset();
        continue;
      }
    }
    word += curr;

    if (!next) {
      reset();
    }
  }

  return { exclude, include };
};

export const getSearchFunction = (query: string) => {
  const { include, exclude } = tokenizeSearchTerm(query);

  if ([...include, ...exclude].length === 0) {
    return (obj: CombinedReport) => ({ obj, priority: 1 });
  }

  const flags = 'i';

  const excludeRegex = exclude.length
    ? new RegExp(exclude.join('|'), flags)
    : null;
  const includeRegex = RegExp(include.join('|'), flags);

  return (obj: CombinedReport) => {
    let found = false;
    let priority = 1;
    for (let key of searchKeys) {
      const value = String(get(obj, key));

      if (excludeRegex && excludeRegex.test(value)) {
        return null;
      }

      if (!found) {
        found = includeRegex.test(value);

        if (found && !excludeRegex) {
          return { obj, priority };
        }

        priority++;
      }
    }

    return found ? { obj, priority } : null;
  };
};
