import { exposeToGlobal } from 'utils/expose-to-global';
import { getGlobalArc } from 'utils/get-global-arc';
import { vizHttp } from 'services/viz-http';
import { logger } from 'services/logger';

const arc = getGlobalArc();
const isCMLUsageEventsEnabled = arc.frontend.featureFlipper.getValueById(
  'enable_cml_usage_events'
);

export interface LoadTimeData {
  url: string;
  loadTime: number;
}

let jsErrors = 0;
let visualErrors = 0;
logger.logs$.subscribe(() => {
  incrementErrorCount(false);
});

export function incrementErrorCount(isVisual: boolean) {
  if (isVisual) {
    visualErrors += 1;
  } else {
    jsErrors += 1;
  }
}

const halfAnHour =
  arc.frontend.globals.settings.eventsTrackerFrontendDataToUsageDB;

if (isCMLUsageEventsEnabled) {
  setInterval(() => {
    const isUsageUploadInProgress = localStorage.getItem(
      'usageUploadInProgress'
    );
    if (isUsageUploadInProgress === 'true') {
      return;
    }
    localStorage.setItem('usageUploadInProgress', 'true');

    try {
      const loadTimeData = getLoadTimeData();
      if (loadTimeData.length > 0) {
        // the logger.logs$ and base.alertMessage is collecting
        // the console.errors which both jsErrors and visualErrors
        // so substrack the visualErrors from the jsErrors
        // returning the jsErrors which are not visualErrors
        const finalJsErrors = jsErrors - visualErrors;
        sendFrontendEventsTrackerData(
          loadTimeData,
          finalJsErrors,
          visualErrors
        );
      }
    } catch (error) {
      console.error('Error processing frontend events tracker data:', error);
    }
  }, halfAnHour);
}

export const getLoadTimeData = (): LoadTimeData[] => {
  const loadTimesJSON = localStorage.getItem('loadTimes');
  const loadTimes = loadTimesJSON ? JSON.parse(loadTimesJSON) : [];
  return loadTimes;
};

export const clearLoadTimeData = () => {
  localStorage.removeItem('loadTimes');
  localStorage.removeItem('usageUploadInProgress');
};

export const sendFrontendEventsTrackerData = async (
  data: LoadTimeData[],
  jsErrors: number,
  visualErrors: number
) => {
  try {
    await vizHttp.post({
      url: arc.frontend.globals.urls.saveFrontendEventsTrackerData,
      headers: {
        'X-CSRFTOKEN': arc.frontend.globals.authentication.csrfToken,
      },
      isOldEndpoint: true,
      data: {
        data: JSON.stringify(data),
        jsErrors,
        visualErrors,
      },
    });

    clearLoadTimeData();
    jsErrors = 0;
    visualErrors = 0;
  } catch (error) {
    console.error('Error sending load time data:', error);
  }
};

export const usageTracker = {
  incrementErrorCount,
};

exposeToGlobal({ usageTracker }, 'frontend');
