import { exposeToGlobal } from 'utils/expose-to-global';
import { getGlobalArc } from 'utils/get-global-arc';

import Cookies from 'js-cookie';
import { ReportFilterType } from 'types/report-filter-type.type';

const arc = getGlobalArc();

type ReportType = 'dashboard' | 'visual';

const get = (key: string): string | undefined => Cookies.get(key);

const getJson = <T extends Object = Object>(key: string): T | undefined => {
  const data = get(key);

  if (data) {
    try {
      return JSON.parse(data);
    } catch (error) {
      console.error(error);
    }
  }

  return undefined;
};

const set = (
  key: string,
  value: string,
  options: Cookies.CookieAttributes = {}
) => {
  /**
   * It's needed in order for backward compatibility.
   * By doing this, we can prevent js-cookie from assigning `/`
   * to path by default, because earlier we expected all the cookies
   * set under the document's location path.
   *
   * If path is undefined, js-cookie will override it with `/`.
   * By setting an empty string, we can make sure, every cookie
   * will be set under the document's location path. That's the
   * browsers` default behaviour.
   */
  Cookies.set(key, value, {
    path: options.path ?? '',
    secure:
      arc.frontend.globals.settings.isSessionCookieSecure || options.secure,
    sameSite: arc.frontend.globals.settings.sessionCookieSameSite ?? 'Lax',
  });
};

const setJson = (key: string, value: unknown, expireDays?: number) => {
  const options = {
    expires: expireDays || 365 * 10,
    path: '/arc/apps',
  };

  set(key, JSON.stringify(value), options);
};

const getArcCookieKey = (key: string, isUserSpecific?: boolean): string => {
  const arcCookieKey = `arc_${key}`;

  if (isUserSpecific) {
    return `${arcCookieKey}_${arc.frontend.globals.authentication.username}`;
  }

  return arcCookieKey;
};

const getLatestVisualsKey = (type: ReportType) =>
  getArcCookieKey(
    type === 'dashboard' ? 'latest_viewed_dashboards' : 'latest_viewed_data',
    true
  );

const getLatestVisuals = (type: ReportType) => {
  const key = getLatestVisualsKey(type);

  const value = getJson<{ latestVisuals?: string[] }>(key);

  return value?.latestVisuals ?? [];
};

const setLatestVisuals = (reportId: string, type: ReportType) => {
  const key = getLatestVisualsKey(type);

  let latestVisuals = getLatestVisuals(type);

  const isVisualAlreadyAdded = latestVisuals.some((id) => id === reportId);

  if (isVisualAlreadyAdded) {
    latestVisuals = latestVisuals.filter((id) => id !== reportId);
  }

  latestVisuals.push(reportId);

  if (latestVisuals.length >= 15) {
    latestVisuals.shift();
  }

  setJson(key, { latestVisuals });
};

const getAppBuilderConnectionDatasetKey = () =>
  getArcCookieKey('appbuilder_conn_ds', true);

const getAppBuilderConnectionDataset = () => {
  const appBuilderKey = getAppBuilderConnectionDatasetKey();

  return getJson(appBuilderKey);
};

const setAppBuilderConnectionDataset = (key: string, value: unknown) => {
  const appBuilderKey = getAppBuilderConnectionDatasetKey();
  const appBuilder = getJson(appBuilderKey);

  setJson(appBuilderKey, {
    ...(appBuilder ?? {}),
    [key]: value,
  });
};

const getDiplayFormatKey = () => getArcCookieKey('dataset_displayformat', true);

const getDisplayFormat = () => {
  const key = getDiplayFormatKey();
  return getJson(key) !== false;
};

const setDisplayFormat = (value: unknown) => {
  const key = getDiplayFormatKey();
  setJson(key, value);
};

const getVisualTabSettingsKey = () =>
  getArcCookieKey('visual_tab_layout', true);

type VisualTabLayout = {
  appgroupid: number;
  selectiontype: 'appgroups';
  sideBarStyle: {
    width: number;
  };
  displaystyle: boolean;
  displaymode: ReportFilterType;
};

const getVisualTabSettingByKey = <Key extends keyof VisualTabLayout>(
  settingKey: Key
): VisualTabLayout[Key] | undefined => {
  const key = getVisualTabSettingsKey();
  const value = getJson<VisualTabLayout>(key);

  return value?.[settingKey];
};

const setVisualTabSettingByKey = <Key extends keyof VisualTabLayout>(
  settingKey: Key,
  newValue: VisualTabLayout[Key]
) => {
  const key = getVisualTabSettingsKey();
  const value = getJson(key);

  setJson(key, { ...(value ?? {}), [settingKey]: newValue });
};

const removeCookie = (key: string, options: Cookies.CookieAttributes) =>
  Cookies.remove(key, options);

export const vizCookie = {
  get,
  getJson,
  set,
  setJson,
  getArcCookieKey,
  getLatestVisuals,
  setLatestVisuals,
  getAppBuilderConnectionDataset,
  setAppBuilderConnectionDataset,
  getDisplayFormat,
  setDisplayFormat,
  getVisualTabSettingByKey,
  setVisualTabSettingByKey,
  removeCookie,
};

exposeToGlobal({ vizCookie }, 'frontend');
