import { AlertVariant } from 'types/alert-variant.type';
import { createStore } from './create-store';

type Notification = {
  id: number | string;
  variant: AlertVariant;
  message: string;
  title?: string;
  useHtml?: boolean;
  onClose?: () => void;
  expandedMessage?: string;
  timeoutRef?: NodeJS.Timeout;
};

type NotificationState = {
  notifications: Notification[];

  /**
   * Displays a new popup notification. For inline notifications use the Notification component directly
   * @param notification.id Unique identifier of the notification. If not provided, a new id will be generated.
   * @param notification.variant Possible types: `danger` | `warning` | `info` | `success`
   * @param notification.message The message to display
   * @param notification.title The alert title: defaults to 'Error', 'Warning', 'Info', or 'Success'
   * @param notification.useHtml Indicates wether the message contains HTML tags
   * @param notification.onClose Callback that is invoked when the Close button is pressed on the notification
   * @param expandedMessage Additional message (ex. error details) that is displayed when pressing the More button
   *
   * @param options.delay Specifies how long the notification should be displayed. If 0 or undefined, it will not be dismissed automatically
   * @param stackMessages Specifies wether the notification can be displayed together with other notifications
   */
  addNotification: (
    notification: Omit<Notification, 'id' | 'timeoutRef'> & {
      id?: number | string;
    },
    options?: {
      delay?: number;
      stackMessages?: boolean;
    }
  ) => void;
  removeNotification: (id: number | string) => void;
  removeAllNotifications: () => void;
};

let generatedId = 0;

export const useNotificationStore = createStore<NotificationState>(
  'notifications',
  (set) => ({
    notifications: [],

    addNotification: (notification, options) =>
      set(
        (state) => {
          let id = notification.id;

          // if there is no explicit id, a numeric one will be generated
          if (id === undefined) {
            generatedId += 1;
            id = generatedId;
          }

          // hide the notification after the specified delay
          let timeoutRef: NodeJS.Timeout | undefined;
          if (options?.delay) {
            timeoutRef = setTimeout(() => {
              set(
                (state) => ({
                  notifications: state.notifications.filter(
                    (notification) => notification.id !== id
                  ),
                }),
                false,
                'remove notification after delay'
              );
            }, options.delay);
          }

          const newNotification = { ...notification, id, timeoutRef };

          if (options?.stackMessages) {
            // check if the same notification was already added to the list
            const existingNotification = state.notifications.find(
              (item) =>
                item.id === id ||
                (item.message === notification.message &&
                  item.title === notification.title)
            );
            clearTimeout(existingNotification?.timeoutRef);

            return {
              notifications: [
                ...state.notifications.filter(
                  (notification) => notification.id !== existingNotification?.id
                ),
                newNotification,
              ],
            };
          } else {
            // if the notification is not stackable, the notification container has to be cleared
            state.notifications.forEach((item) =>
              clearTimeout(item.timeoutRef)
            );
            return {
              notifications: [newNotification],
            };
          }
        },
        false,
        'add notification'
      ),

    removeNotification: (id) =>
      set(
        (state) => {
          const notification = state.notifications.find(
            (notification) => notification.id === id
          );
          clearTimeout(notification?.timeoutRef);

          return {
            notifications: state.notifications.filter(
              (notification) => notification.id !== id
            ),
          };
        },
        false,
        'remove notification'
      ),

    removeAllNotifications: () =>
      set(
        (state) => {
          state.notifications.forEach((notification) =>
            clearTimeout(notification?.timeoutRef)
          );

          return { notifications: [] };
        },
        false,
        'remove all notifications'
      ),
  })
);
