import { Token } from 'types/api.type';
import {
  AddParamArg,
  ParamSelection,
  ParamSource,
} from 'types/chatbot-message.type';

const citationRegex = /@@@Citations:\s*(.*?)@@@/i;

function extractCitations(
  llmResponse: string,
  numberOfRows: number
): number[] | null {
  const match = llmResponse.match(citationRegex);
  if (!match) {
    return null;
  }

  const inside = match[1].trim();

  // 3) Split by commas, parse them as integers, validate range
  const strNumbers = inside.split(',').map((n) => n.trim());
  const rowIndices: number[] = [];

  for (const sn of strNumbers) {
    const rowNum = parseInt(sn, 10) - 1; // Convert to zero-based index
    if (!isNaN(rowNum) && rowNum >= 0 && rowNum < numberOfRows) {
      rowIndices.push(rowNum);
    }
  }

  return rowIndices.length > 0 ? rowIndices : null;
}

function finalizeAnswer({
  answerText,
  relevantRows,
  settings,
  addParams,
  convertParam,
}: {
  answerText: string;
  relevantRows: string[][];
  settings: {
    paramSelection: ParamSelection;
    paramSource: ParamSource;
  };
  addParams: (p: AddParamArg[]) => void;
  convertParam: (updatedRows: string[][]) => AddParamArg[];
}): { finalText: string; updatedRows: string[][] } {
  let finalText = answerText;
  let updatedRows = relevantRows;

  if (
    settings.paramSource !== 'none' &&
    settings.paramSelection === 'used-by-llm'
  ) {
    // Extract citations
    const citedRowIndices = extractCitations(answerText, updatedRows.length);
    if (citedRowIndices?.length) {
      updatedRows = citedRowIndices.map((i) => updatedRows[i]);
    }
    finalText = answerText.replace(citationRegex, '').trim();

    // Now that relevantRows is filtered, addParams
    addParams(convertParam(updatedRows));
  }

  return { finalText, updatedRows };
}

// Helper function to tokenize the text
function tokenize(text: string, isMarkdown: boolean): Token[] {
  if (!isMarkdown) {
    // Plain text: split on whitespace/newlines
    return splitIntoTextTokens(text);
  }

  // For Markdown:
  // 1. Extract code blocks as immediate tokens.
  // 2. For the text outside code blocks, tokenize like normal text.
  const tokens: Token[] = [];
  let lastIndex = 0;
  const codeBlockRegex = /```[\s\S]*?```/g;
  let match: RegExpExecArray | null;

  while ((match = codeBlockRegex.exec(text)) !== null) {
    const start = match.index!;
    const end = start + match[0].length;

    // Tokens before this code block
    if (start > lastIndex) {
      const beforeText = text.slice(lastIndex, start);
      tokens.push(...splitIntoTextTokens(beforeText));
    }

    // Code block token (immediate)
    tokens.push({
      content: match[0],
      isImmediate: true,
    });
    lastIndex = end;
  }

  // Remaining text after last code block
  if (lastIndex < text.length) {
    const remainingText = text.slice(lastIndex);
    tokens.push(...splitIntoTextTokens(remainingText));
  }

  return tokens;
}

function splitIntoTextTokens(text: string): Token[] {
  // Split on any non-empty sequence or newlines
  const splitTokens = text.match(/\S+|\n+/g) || [];
  return splitTokens.map((content) => ({
    content,
    // all non-code text tokens are animated
    isImmediate: false,
  }));
}

export { finalizeAnswer, tokenize };
