type SortDirection = 'asc' | 'desc';

type SortableField<T> = keyof T;

type TableData<T> = {
  data: T[];
  column?: SortableField<T>;
  direction?: SortDirection;
};

export function dataSorter<T>(tableData: TableData<T>): T[] {
  const { data, column, direction } = tableData;

  if (!column) {
    // If column is not provided, apply default sorting
    return data.slice();
  }

  return data.slice().sort((a, b) => {
    const valueA = a[column];
    const valueB = b[column];

    if (typeof valueA === 'string' && typeof valueB === 'string') {
      // Extract numeric parts from strings (if any)
      const numPartA = parseFloat(valueA.match(/\d+/)?.[0] || '');
      const numPartB = parseFloat(valueB.match(/\d+/)?.[0] || '');

      // Compare the numeric parts first
      if (!isNaN(numPartA) && !isNaN(numPartB)) {
        if (numPartA !== numPartB) {
          return direction === 'asc'
            ? numPartA - numPartB
            : numPartB - numPartA;
        }
      }

      // If the numeric parts are equal or not applicable, compare the strings as usual
      return direction === 'asc'
        ? valueA.localeCompare(valueB)
        : valueB.localeCompare(valueA);
    } else if (typeof valueA === 'number' && typeof valueB === 'number') {
      return direction === 'asc' ? valueA - valueB : valueB - valueA;
    }

    // Default case, no sorting applied
    return 0;
  });
}
