// It loads js or css by injecting script or link tags into the DOM
// If the script or link tag already exists in the DOM
// 1. And loadResources is called BEFORE the already existing script is loaded
//   It will attach a callback function onto the element.onload and will track
//   when the script is loaded
// 2. And loadResources is called AFTER the already existing script is loaded
//   onEverythingLoaded callback will never be called cause it can't track if
//   the script is loaded or not.
//   You have to mark the script or link tag with an data-loaded="true" attribute
//   If the tag is marked it will not wait for the already loaded script.
export const loadResources = (
  type: 'script' | 'link',
  filePaths: string[],
  classes: string[],
  onEverythingLoaded: () => void
) => {
  if (type !== 'script' && type !== 'link') {
    throw new Error('Invalid type ' + type);
  }

  const alreadyExistingElements: (HTMLScriptElement | HTMLLinkElement)[] = [];
  const elementsToAppend: (HTMLScriptElement | HTMLLinkElement)[] = [];

  filePaths.forEach((path) => {
    const attribute = type === 'script' ? 'src' : 'href';
    // script[src="$[path"] || head link[href="${path}"]
    const selector = `${
      type === 'script' ? 'script' : 'head link'
    }[${attribute}="${path}"]`;
    const $selection = $<HTMLScriptElement | HTMLLinkElement>(selector);

    const isTagAlreadyExisting = $selection.length !== 0;

    if (isTagAlreadyExisting) {
      const $element = $selection.first();

      $element.addClass(classes);

      const element = $element.get(0);
      if (element && $element.data('loaded') !== true) {
        alreadyExistingElements.push(element);
      }
    } else {
      let element: HTMLScriptElement | HTMLLinkElement;

      if (type === 'script') {
        element = document.createElement('script');
        element.type = 'text/javascript';
        element.async = false;
        element.src = path;
      } else {
        element = document.createElement('link');
        element.type = 'text/css';
        element.rel = 'stylesheet';
        element.href = path;
      }

      if (classes) {
        element.classList.add(...classes);
      }

      elementsToAppend.push(element);
    }
  });

  let numberOfLoadedElements = 0;

  const onLoadHandler = (element: HTMLScriptElement | HTMLLinkElement) => {
    $(element).attr('data-loaded', 'true');

    numberOfLoadedElements += 1;

    const isEverythingLoaded =
      numberOfLoadedElements ===
      alreadyExistingElements.length + elementsToAppend.length;

    if (isEverythingLoaded) {
      if (onEverythingLoaded) {
        onEverythingLoaded();
      }
    }
  };

  alreadyExistingElements.forEach((element) => {
    const initialOnLoadHandler = element.onload as () => void;

    element.onload = function () {
      onLoadHandler(element);
      if (initialOnLoadHandler) {
        initialOnLoadHandler();
      }
    };
  });

  elementsToAppend.forEach((element) => {
    element.onload = function () {
      onLoadHandler(element);
    };

    document.head.appendChild(element);
  });
};
