import { Configuration, BannerPlugin } from 'webpack';
import TsconfigPathsPlugin from 'tsconfig-paths-webpack-plugin';
import MiniCssExtractPlugin from 'mini-css-extract-plugin';
import CssMinimizerPlugin from 'css-minimizer-webpack-plugin';
import TerserPlugin from 'terser-webpack-plugin';
import fs from 'fs';
import path from 'path';

const copyright = fs
  .readFileSync(path.resolve(__dirname, '../../copyright.txt'))
  .toString();

export const commonWebpackConfiguration: Configuration = {
  entry: {
    // Vendor
    'expose-shared-vendors-to-arc':
      './webpack/entries/expose-shared-vendors-to-arc.ts',
    // Code needed ASAP
    navbar: './webpack/entries/navbar.ts',
    init: './webpack/entries/init.ts',
    // Everything that the old code base depends on
    'arc-dependency': './webpack/entries/arc-dependency.ts',
    // Pages
    'login-page': './src/pages/login-page/index.ts',
    'home-page': './src/pages/home/index.ts',
    'set-home-page': './src/pages/set-home-page/index.ts',
    'site-settings-page': './src/pages/site-settings-page/index.ts',
    'change-password-page': './src/pages/change-password-page/index.ts',
    'visuals-page': './src/pages/visuals-page/index.ts',
    'job-page': './src/pages/job-page/index.ts',
    'job-overview-page': './src/pages/job-overview-page/index.ts',
    'error-page': './src/pages/error-page/index.ts',
    'role-details-page': './src/pages/settings-page/index.ts',
    'manage-roles-page': './src/pages/manage-roles-page/index.ts',
    'manage-aliases-page': './src/pages/manage-aliases-page/index.ts',
  },
  output: {
    clean: true,
    filename: '[name].js',
  },
  module: {
    rules: [
      {
        test: /\.(ts|tsx)?$/,
        loader: 'ts-loader',
        exclude: /node_modules/,
      },
      {
        test: /\.svg$/i,
        issuer: /\.tsx?$/,
        use: [
          {
            loader: '@svgr/webpack',
            options: {
              typescript: true,
              ext: 'tsx',
            },
          },
        ],
      },
      {
        test: /\.(sa|sc|c)ss$/,
        exclude: /\.module\.(sa|sc|c)ss$/,
        use: [
          MiniCssExtractPlugin.loader,
          {
            loader: 'css-loader',
            options: {
              modules: 'global',
              importLoaders: 2,
            },
          },
          {
            loader: 'postcss-loader',
            options: {
              postcssOptions: {
                plugins: [['postcss-preset-env']],
              },
            },
          },
          'sass-loader',
        ],
      },
      {
        test: /\.module\.(sa|sc|c)ss$/,
        use: [
          MiniCssExtractPlugin.loader,
          {
            loader: 'css-loader',
            options: {
              modules: {
                mode: 'local',
                exportLocalsConvention: 'camelCaseOnly',
              },
              importLoaders: 2,
            },
          },
          {
            loader: 'postcss-loader',
            options: {
              postcssOptions: {
                plugins: [['postcss-preset-env']],
              },
            },
          },
          'sass-loader',
        ],
      },
    ],
  },
  resolve: {
    extensions: ['.tsx', '.ts', '.js'],
    plugins: [
      new TsconfigPathsPlugin({
        configFile: './tsconfig.json',
      }),
    ],
  },
  optimization: {
    splitChunks: {
      cacheGroups: {
        vendor: {
          test: /[\\/]node_modules[\\/]/,
          name: 'node-modules',
          chunks: 'all',
        },
      },
    },
    runtimeChunk: {
      name: 'webpack-runtime',
    },
    minimizer: [
      new TerserPlugin({ extractComments: false }),
      new CssMinimizerPlugin(),
    ],
  },
  plugins: [
    new MiniCssExtractPlugin(),
    new BannerPlugin({
      banner: copyright,
      exclude: [
        'node-modules',
        'expose-shared-vendors-to-arc',
        'webpack-runtime',
      ],
    }),
  ],
};
