# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.

import os
import string
import subprocess
import sys


DIR = os.path.dirname(os.path.realpath(__file__))


class ENV():
  def __getattr__(self, name):
    return os.environ.get(name)

env = ENV()

VIZ_INSTANCE = """\
[program:viz{index}]
environment = PERSISTENT_PROCESS_ID="{index}"
command = {DIR}/arcviz runcpserver --host=127.0.0.1 --port={port} --minthreads={HTTP_THREADS} --maxthreads={HTTP_THREADS}
"""

SUPERVISOR_CONF_CONTENT = """\
[supervisord]
environment = DJANGO_SETTINGS_MODULE='settings_local', PYTHONPATH={env.PYTHONPATH_STRIPPED}
nodaemon = true
childlogdir = {env.BASEDIR}

{viz_instances_text}
[program:viz_master]
command = {DIR}/caddy -conf {CADDY_CONF} -log {env.BASEDIR}/caddy.log
stderr_logfile = {env.BASEDIR}/caddy.err
stdout_logfile = {env.BASEDIR}/caddy.out
"""

def get_conf_dir():
  return env.VIZ_MULTIPROC_CONF or f'{env.BASEDIR}/multiprocess'

def get_worker_start_port():
  # Add one to the base Viz Port number so that workers don't conflict with proxy listen port (i.e. 8100)
  VIZ_WORKER_START_PORT = int(env.VIZ_PORT or env.CDSW_APP_PORT) + 1
  VIZ_WORKER_START_PORT = int(env.VIZ_WEBSERVER_PORT or VIZ_WORKER_START_PORT)
  return VIZ_WORKER_START_PORT

def generate_supervisord_conf():

  VIZ_WORKER_START_PORT = get_worker_start_port()
  VIZ_MULTIPROC_NUM = int(env.MULTIPROC_WORKER_NUM or 1)

  VIZ_MULTIPROC_CONF = get_conf_dir()
  SUPERVISOR_CONF = f'{VIZ_MULTIPROC_CONF}/supervisord.conf'
  CADDY_CONF = f'{VIZ_MULTIPROC_CONF}/Caddyfile'

  HTTP_THREADS = env.HTTP_THREADS or 10

  viz_instances = []

  for i in range(0, VIZ_MULTIPROC_NUM):
    index = i + 1
    port = VIZ_WORKER_START_PORT + i
    viz_instances.append(VIZ_INSTANCE.format(**(globals() | locals())))

  viz_instances_text = '\n'.join(viz_instances)
  content = SUPERVISOR_CONF_CONTENT.format(**(globals() | locals()))

  print(f'writing {SUPERVISOR_CONF}')
  os.makedirs(VIZ_MULTIPROC_CONF, exist_ok=True)
  with open(SUPERVISOR_CONF, 'wt') as f:
    f.write(content)

# for Caddyfile we use string.Template instead of an f-string in order
# to avoid the need of escaping curly braces
CADDY_CONF_CONTENT = string.Template("""\
${HTTP_PROTOCOL}://*:${VIZ_PORT}
bind ${VIZ_HOST}
tls ${TLS_SETTING}
gzip
errors ${LOG_DIR}/webserver_error.txt
log / ${LOG_DIR}/access_logs.txt "${ACCESS_LOG_FORMAT}"
root ${ROOT_STATIC_FILE_PATH}
proxy / ${UPSTREAM_HTTP_PROTOCOL}://127.0.0.1:${VIZ_WORKER_START_PORT} {
  # transparent

  header_upstream Host {host}
  header_upstream X-Real-IP {remote}
  # header_upstream +X-Forwarded-For {remote}
  header_upstream X-Forwarded-Proto {scheme}

  insecure_skip_verify # needed because of missing ca_certificates support
  except ${MASTER_EXCEPT} /arc/static /static/
}

rewrite /arc/static {
  r  (.*)
  to /static/{1}
}
expires {
  match /static/.* 1h
}
""")

def upstream_http_protocol():
  # we assume the upstream (the web server) has the same SSL setting as the reverse proxy
  # if this is wrong, one can use the UPSTREAM_HTTP_PROTOCOL environment variable to provide the correct value

  env_setting = env.UPSTREAM_HTTP_PROTOCOL
  if env_setting:
    return env_setting

  return 'https' if ssl_enabled() else 'http'

def generate_Caddyfile():

  VIZ_HOST = env.VIZ_HOST or '0.0.0.0'
  VIZ_WORKER_START_PORT = get_worker_start_port()

  MASTER_EXCEPT = ''  # TODO
  HTTP_PROTOCOL = 'http'
  UPSTREAM_HTTP_PROTOCOL = upstream_http_protocol()
  TLS_SETTING = 'off'

  if ssl_enabled():
    HTTP_PROTOCOL = 'https'
    TLS_SETTING = tls_setting()

  ACCESS_LOG_FORMAT = env.ACCESS_LOG_FORMAT or "{upstream} {common}"

  VIZ_MULTIPROC_CONF = get_conf_dir()
  CADDY_CONF = f'{VIZ_MULTIPROC_CONF}/Caddyfile'
  LOG_DIR = env.LOG_DIR or env.BASEDIR

  content = CADDY_CONF_CONTENT.substitute(**(os.environ | globals() | locals()))

  print(f'writing {CADDY_CONF}')
  os.makedirs(VIZ_MULTIPROC_CONF, exist_ok=True)
  with open(CADDY_CONF, 'wt') as f:
    f.write(content)


def ssl_enabled():
  return env.SSL_ENABLED == 'true'


def tls_setting():
  # must create a passwordless key:
  dataviz_key_file = f'{env.BASEDIR}/dataviz-key.pem'
  command = [
    'openssl',
    'rsa',
    '-inform',
    'PEM',
    '-outform',
    'PEM',
    '-in',
    f'{env.SSL_PRIVATE_KEY}',
    '-passin',
    f'pass:{env.SSL_PRIVATE_KEY_PASSWORD}',
    '-out',
    dataviz_key_file
  ]

  print('Generating passwordless key:')
  print(f'  openssl command: {" ".join(command)}')

  openssl = subprocess.Popen(
    args=command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  stdout, stderr = openssl.communicate()
  print(f'  command returncode {openssl.returncode}')
  print(f'  command stdout {stdout}')
  print(f'  command stderr {stderr}')

  if openssl.returncode != 0:
    print(f'ERROR executing openssl command. Aborting.')
    sys.exit(1)

  # tls is configured by providing the certificate and key files:
  return f'{env.SSL_SERVER_CERTIFICATE} {dataviz_key_file}'

def main():
  generate_supervisord_conf()
  generate_Caddyfile()

if __name__ == '__main__':
  main()