#!/usr/bin/env bash

set -x
set -e

DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" >/dev/null 2>&1 && pwd )"

echo "${PYTHONPATH}"
export PYTHONPATH_STRIPPED=${PYTHONPATH/:\${PYTHONPATH\}:/:} # replace :${PYTHONPATH}
echo "${PYTHONPATH_STRIPPED}"

# attempt a python-based config generation
$VENV_BASE_PATH/bin/python $DIR/gen_multiproc_conf.py
exit 0

# Add one to the base Viz Port number so that workers don't conflict with proxy listen port (i.e. 8100)
let VIZ_WORKER_START_PORT=${VIZ_PORT:-${CDSW_APP_PORT}}+1
VIZ_WORKER_START_PORT=${VIZ_WEBSERVER_PORT:-${VIZ_WORKER_START_PORT}}
VIZ_HOST=${VIZ_HOST:-0.0.0.0}
VIZ_MULTIPROC_NUM=${VIZ_HOST}

VIZ_MULTIPROC_NUM=${MULTIPROC_WORKER_NUM}
VIZ_MULTIPROC_CONF=${VIZ_MULTIPROC_CONF:-${BASEDIR}/multiprocess}
mkdir -p $VIZ_MULTIPROC_CONF
SUPERVISOR_CONF=$VIZ_MULTIPROC_CONF/supervisord.conf
CADDY_CONF=$VIZ_MULTIPROC_CONF/Caddyfile

HTTP_THREADS=${HTTP_THREADS:-10}

# Generate Caddyfile and supervisord.conf for multiprocess arcviz

# Generate supervisord.conf file
echo "[supervisord]" > $SUPERVISOR_CONF
echo "environment = DJANGO_SETTINGS_MODULE='settings_local', PYTHONPATH=${PYTHONPATH_STRIPPED}" >> $SUPERVISOR_CONF
echo "nodaemon = true" >> $SUPERVISOR_CONF
echo "childlogdir = ${BASEDIR}" >> $SUPERVISOR_CONF
for ((i=1;i<=VIZ_MULTIPROC_NUM;i+=1))
do
  let VIZ_WORKER_PORT=${VIZ_WORKER_START_PORT}+$i-1
  echo "[program:viz$i]" >> $SUPERVISOR_CONF
  echo "environment = PERSISTENT_PROCESS_ID=\"$i\"" >> $SUPERVISOR_CONF
  echo "command = "${DIR}/arcviz runcpserver --host=127.0.0.1 --port=$VIZ_WORKER_PORT --minthreads=$HTTP_THREADS --maxthreads=$HTTP_THREADS >> $SUPERVISOR_CONF
done
echo [program:viz_master] >> $SUPERVISOR_CONF
echo command = ${DIR}/caddy -conf $VIZ_MULTIPROC_CONF/Caddyfile -log ${BASEDIR}/caddy.log >> $SUPERVISOR_CONF
echo stderr_logfile = ${BASEDIR}/caddy.err >> $SUPERVISOR_CONF
echo stdout_logfile = ${BASEDIR}/caddy.out >> $SUPERVISOR_CONF


# Generate Caddyfile file

echo "http://*:${VIZ_PORT}" > $CADDY_CONF
# If testing locally with docker, we don't want to bind to localhost only
if [[ "${MULTIPROC_TESTING}" != "true" ]]; then
  echo "bind ${VIZ_HOST}" >> $CADDY_CONF
fi
echo "tls off" >> $CADDY_CONF

# Generate SSL settings (Deprecated)

# NOTE: This is deprecated in CDV form factors because SSL is handled upstream through the different CML and CDW proxies,
#  but this might be handy to have around in the future in case we need to also enable ssl when running arcviz processes.

#  if [[ ${ARCVIZ_SSL_ENABLED} == "true" ]];
#  then
#    echo "https://*:${ARCADIA_UI_PORT}" > $CADDY_CONF
#    echo "tls ${ARCVIZ_SSL_CERT} ${ARCVIZ_SSL_KEY} {" >> $CADDY_CONF
#    if [[ ! -z ${ARCVIZ_SSL_CA_CERT} ]];
#    then
#      echo "  ca ${ARCVIZ_SSL_CA_CERT}" >> $CADDY_CONF
#    fi
#    if [[ ! -z ${ARCVIZ_SSL_KEY} ]];
#      then
#        caddy_password_script=`grep password.script arcviz.properties | cut -d'=' -f2`
#        if [[ ! -z ${caddy_password_script} ]];
#        then
#          echo "  key_pass_script ${caddy_password_script}" >> $CADDY_CONF
#        fi
#      fi
#    echo "}" >> $CADDY_CONF

echo "gzip" >> $CADDY_CONF

echo "errors ${BASEDIR}/webserver_error.txt " >> $CADDY_CONF

# there are two predefined Caddy log formats: {common} and {combined}: https://caddy.its-em.ma/v1/docs/log#log-format
# additional placeholders can be added: https://caddy.its-em.ma/v1/docs/placeholders
# example: ACCESS_LOG_FORMAT="{upstream} {combined} [{>X-Forwarded-For}] {latency_ms}"
ACCESS_LOG_FORMAT=${ACCESS_LOG_FORMAT:-"{upstream} {common}"}

echo "log / ${BASEDIR}/access_logs.txt \"${ACCESS_LOG_FORMAT}\"" >> $CADDY_CONF
echo "root ${ROOT_STATIC_FILE_PATH}" >> $CADDY_CONF

# List of urls which should be load-balanced accross workers
declare -a bal=(
  "/arc/sqlrun/jsonselect"
  "/arc/sqlrun/jsonselect_parallel"
  "/arc/reports/customjsraw"
  "/arc/reports/customcssraw"
  "/arc/apps/apps"
  "/arc/workspace/workspace"
  "/arc/reports/report_usage"
  "/arc/datasets/dataset"
  "/arc/reports/appgroup"
  "/arc/reports/report"
  "/arc/apps/app/"
  "/arc/apps/login"
  "/arc/apps/logout"
  "/arc/apps/snapshot"
  "/arc/apps/home"
  "arc/apps/homepage_statistics"
  "arc/apps/homepage_number_of_recently_created_resources"
  "arc/apps/homepage_recently_created_dashboards_by_user"
  "arc/apps/homepage_last_viewed_reports_by_user"
  "arc/apps/homepage_most_popular_reports"
  "arc/apps/homepage_samples_dashboards"
  "arc/apps/homepage_favorite_dashboards"
  "arc/apps/homepage_datasets"
)

# List of urls which do not play nice with master except list
declare -a mas=(
  "/arc/apps/appgroup"
  "/arc/apps/app_accel"
  "/arc/apps/appedit"
  "/arc/sqlrun/jsonselect_parallel_async"
)

# Generate master routed proxy rules
echo "proxy / 127.0.0.1:${VIZ_WORKER_START_PORT} {" >> $CADDY_CONF
# In caddy 0.10.10 the transparent directive breaks the x-forwarded-for header.
# workaround: https://github.com/caddyserver/caddy/issues/1960#issuecomment-346138997
echo "  # transparent" >> $CADDY_CONF
echo "
  header_upstream Host {host}
  header_upstream X-Real-IP {remote}
  # header_upstream +X-Forwarded-For {remote}
  header_upstream X-Forwarded-Proto {scheme}
  header_upstream X-Forwarded-Host {host}
" >> $CADDY_CONF

VIZ_MASTER_EXCEPT="except "
if [[ ${VIZ_MULTIPROC_NUM} -gt 1 ]];
then
  for i in "${bal[@]}"
  do
    VIZ_MASTER_EXCEPT+=$i
    VIZ_MASTER_EXCEPT+=" "
  done

  for ((j=1;j<=VIZ_MULTIPROC_NUM;j+=1))
  do
    VIZ_MASTER_EXCEPT+="/arc/sqlrun/p$j "
#    VIZ_MASTER_EXCEPT+="/arc/sqlrun/p$j/asyncstatus_async "
  done
fi
VIZ_MASTER_EXCEPT+="/arc/static /static/"
echo "$VIZ_MASTER_EXCEPT" >> $CADDY_CONF
echo "}" >> $CADDY_CONF

if [[ ${VIZ_MULTIPROC_NUM} -gt 1 ]];
then
  for i in "${mas[@]}"
  do
    MAS_PROXY_LIST="proxy $i "
    MAS_PROXY_LIST+="127.0.0.1:${VIZ_WORKER_START_PORT}"
    echo "$MAS_PROXY_LIST {" >> $CADDY_CONF
    echo "  transparent" >> $CADDY_CONF
    echo "}" >> $CADDY_CONF
  done

  # Generate load balanced proxy rules (non-master)
  for i in "${bal[@]}"
  do
    PROXY_LIST="proxy $i "
    for ((k=2;k<=VIZ_MULTIPROC_NUM;k+=1))
    do
      PROXY_LIST+="127.0.0.1:$(( $k+${VIZ_WORKER_START_PORT}-1 )) "
    done
    echo "$PROXY_LIST {" >> $CADDY_CONF
    echo "  transparent" >> $CADDY_CONF
    echo "}" >> $CADDY_CONF
  done

  # Generate specified process ID proxy rules
  for ((i=1;i<=VIZ_MULTIPROC_NUM;i+=1))
  do
    let VIZ_WORKER_PORT=${VIZ_WORKER_START_PORT}+$i-1
    echo "" >> $CADDY_CONF
    echo "proxy /arc/sqlrun/p$i 127.0.0.1:$VIZ_WORKER_PORT {" >> $CADDY_CONF
    echo "  transparent" >> $CADDY_CONF
    echo "}" >> $CADDY_CONF
  done
fi
# Rewrite static urls for caddy serving
echo "" >> $CADDY_CONF
echo "rewrite /arc/static {" >> $CADDY_CONF
echo "  r  (.*)" >> $CADDY_CONF
echo "  to /static/{1}" >> $CADDY_CONF
echo "}" >> $CADDY_CONF

echo "expires {" >> $CADDY_CONF
echo "  match /static/.* 1h" >> $CADDY_CONF
echo "}" >> $CADDY_CONF
