# Copyright (c) 2013 - 2025 Cloudera, Inc.  All Rights Reserved.
#
# Unless otherwise agreed to in writing by Cloudera, use of this file is exclusively governed by the Cloudera Standard License (the “License”), a copy of which may be found at https://www.cloudera.com/legal/commercial-terms-and-conditions.html.
#
# This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the License for the specific permissions and limitations governing your use of the file.


import unittest
from unittest.mock import mock_open, patch, Mock

import gen_multiproc_conf

EXPECTED_SUPERVISORD = f"""\
[supervisord]
environment = DJANGO_SETTINGS_MODULE='settings_local', PYTHONPATH=MOCK_PYTHONPATH
nodaemon = true
childlogdir = MOCK_BASEDIR

[program:viz1]
environment = PERSISTENT_PROCESS_ID="1"
command = {gen_multiproc_conf.DIR}/arcviz runcpserver --host=127.0.0.1 --port=8101 --minthreads=10 --maxthreads=10

[program:viz_master]
command = {gen_multiproc_conf.DIR}/caddy -conf MOCK_BASEDIR/multiprocess/Caddyfile -log MOCK_BASEDIR/caddy.log
stderr_logfile = MOCK_BASEDIR/caddy.err
stdout_logfile = MOCK_BASEDIR/caddy.out
"""

EXPECTED_SUPERVISORD_WITH_20_THREADS = f"""\
[supervisord]
environment = DJANGO_SETTINGS_MODULE='settings_local', PYTHONPATH=MOCK_PYTHONPATH
nodaemon = true
childlogdir = MOCK_BASEDIR

[program:viz1]
environment = PERSISTENT_PROCESS_ID="1"
command = {gen_multiproc_conf.DIR}/arcviz runcpserver --host=127.0.0.1 --port=8101 --minthreads=20 --maxthreads=20

[program:viz_master]
command = {gen_multiproc_conf.DIR}/caddy -conf MOCK_BASEDIR/multiprocess/Caddyfile -log MOCK_BASEDIR/caddy.log
stderr_logfile = MOCK_BASEDIR/caddy.err
stdout_logfile = MOCK_BASEDIR/caddy.out
"""


class TestSupervisordConfGen(unittest.TestCase):
  def setUp(self):
    self.patches = []

    os_patch = patch.object(gen_multiproc_conf, 'os')
    self.patches.append(os_patch)
    self.os_mock = os_patch.start()

    open_patch = patch.object(
      gen_multiproc_conf, 'open', new_callable=mock_open)
    self.patches.append(open_patch)
    self.open_mock = open_patch.start()

  def tearDown(self):
    for patch in self.patches:
      patch.stop()

  def base_environ(self):
    return {
      'VIZ_PORT': '8100',
      'BASEDIR': 'MOCK_BASEDIR',
      'PYTHONPATH_STRIPPED': 'MOCK_PYTHONPATH'
    }

  def test_basic(self):

    self.os_mock.environ = self.base_environ()

    gen_multiproc_conf.generate_supervisord_conf()

    self.os_mock.makedirs.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess', exist_ok=True)
    self.open_mock.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess/supervisord.conf', 'wt')

    handle: Mock = self.open_mock()
    written_text = handle.write.mock_calls[0].args[0]
    self.assertEqual(written_text, EXPECTED_SUPERVISORD)

  def test_with_threads(self):

    self.os_mock.environ = self.base_environ()
    self.os_mock.environ['HTTP_THREADS'] = '20'

    gen_multiproc_conf.generate_supervisord_conf()

    self.os_mock.makedirs.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess', exist_ok=True)
    self.open_mock.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess/supervisord.conf', 'wt')

    handle: Mock = self.open_mock()
    written_text = handle.write.mock_calls[0].args[0]
    self.assertEqual(written_text, EXPECTED_SUPERVISORD_WITH_20_THREADS)


EXPECTED_CADDYCONF_BASIC = """\
http://*:8100
bind 0.0.0.0
tls off
gzip
errors MOCK_BASEDIR/webserver_error.txt
log / MOCK_BASEDIR/access_logs.txt "{upstream} {common}"
root MOCK_ROOT_STATIC_FILE_PATH
proxy / http://127.0.0.1:8101 {
  # transparent

  header_upstream Host {host}
  header_upstream X-Real-IP {remote}
  # header_upstream +X-Forwarded-For {remote}
  header_upstream X-Forwarded-Proto {scheme}

  insecure_skip_verify # needed because of missing ca_certificates support
  except  /arc/static /static/
}

rewrite /arc/static {
  r  (.*)
  to /static/{1}
}
expires {
  match /static/.* 1h
}
"""

EXPECTED_CADDYCONF_BASIC_SSL = """\
https://*:8100
bind 0.0.0.0
tls MOCK_SSL_SERVER_CERTIFICATE MOCK_BASEDIR/dataviz-key.pem
gzip
errors MOCK_BASEDIR/webserver_error.txt
log / MOCK_BASEDIR/access_logs.txt "{upstream} {common}"
root MOCK_ROOT_STATIC_FILE_PATH
proxy / https://127.0.0.1:8101 {
  # transparent

  header_upstream Host {host}
  header_upstream X-Real-IP {remote}
  # header_upstream +X-Forwarded-For {remote}
  header_upstream X-Forwarded-Proto {scheme}

  insecure_skip_verify # needed because of missing ca_certificates support
  except  /arc/static /static/
}

rewrite /arc/static {
  r  (.*)
  to /static/{1}
}
expires {
  match /static/.* 1h
}
"""

EXPECTED_CADDYCONF_CUSTOM = """\
http://*:8100
bind 0.0.0.0
tls off
gzip
errors MOCK_LOG_DIR/webserver_error.txt
log / MOCK_LOG_DIR/access_logs.txt "{upstream} {combined} [{>X-Forwarded-For}] {latency_ms}"
root MOCK_ROOT_STATIC_FILE_PATH
proxy / https://127.0.0.1:8200 {
  # transparent

  header_upstream Host {host}
  header_upstream X-Real-IP {remote}
  # header_upstream +X-Forwarded-For {remote}
  header_upstream X-Forwarded-Proto {scheme}

  insecure_skip_verify # needed because of missing ca_certificates support
  except  /arc/static /static/
}

rewrite /arc/static {
  r  (.*)
  to /static/{1}
}
expires {
  match /static/.* 1h
}
"""


class TestCaddyConfGen(unittest.TestCase):
  def setUp(self):
    self.maxDiff = None
    self.patches = []

    os_patch = patch.object(gen_multiproc_conf, 'os')
    self.patches.append(os_patch)
    self.os_mock = os_patch.start()

    subprocess_patch = patch.object(gen_multiproc_conf, 'subprocess')
    self.patches.append(subprocess_patch)
    self.subprocess_mock = subprocess_patch.start()

    open_patch = patch.object(
      gen_multiproc_conf, 'open', new_callable=mock_open)
    self.patches.append(open_patch)
    self.open_mock = open_patch.start()

  def tearDown(self):
    for patch in self.patches:
      patch.stop()

  def base_environ(self):
    return {
      'VIZ_PORT': '8100',
      'BASEDIR': 'MOCK_BASEDIR',
      'ROOT_STATIC_FILE_PATH': 'MOCK_ROOT_STATIC_FILE_PATH'
    }

  def test_basic(self):

    self.os_mock.environ = self.base_environ()

    gen_multiproc_conf.generate_Caddyfile()

    self.os_mock.makedirs.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess', exist_ok=True)
    self.open_mock.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess/Caddyfile', 'wt')

    # no SSL is configured, so no openssl calls:
    self.subprocess_mock.Popen.assert_not_called()

    handle: Mock = self.open_mock()
    written_text = handle.write.mock_calls[0].args[0]
    self.assertEqual(written_text, EXPECTED_CADDYCONF_BASIC)

  def test_basic_ssl(self):

    self.os_mock.environ = self.base_environ()
    self.os_mock.environ.update({
      'SSL_ENABLED': 'true',
      'SSL_PRIVATE_KEY': 'MOCK_SSL_PRIVATE_KEY',
      'SSL_PRIVATE_KEY_PASSWORD': 'MOCK_SSL_PRIVATE_KEY_PASSWORD',
      'SSL_SERVER_CERTIFICATE': 'MOCK_SSL_SERVER_CERTIFICATE',
    })

    self.subprocess_mock.Popen.return_value.returncode = 0
    self.subprocess_mock.Popen.return_value.communicate.return_value = (
      'MOCK_OPENSSL_STDOUT', 'MOCK_OPENSSL_STDERR')

    gen_multiproc_conf.generate_Caddyfile()

    self.os_mock.makedirs.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess', exist_ok=True)
    self.open_mock.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess/Caddyfile', 'wt')

    # with SSL we need to invoke openssl
    self.subprocess_mock.Popen.assert_called_once()
    self.assertIn(
      'openssl', self.subprocess_mock.Popen.call_args.kwargs['args'])

    handle: Mock = self.open_mock()
    written_text = handle.write.mock_calls[0].args[0]
    self.assertEqual(written_text, EXPECTED_CADDYCONF_BASIC_SSL)

  def test_overrides(self):

    self.os_mock.environ = self.base_environ()
    self.os_mock.environ.update({
      'UPSTREAM_HTTP_PROTOCOL': 'https',
      'VIZ_WEBSERVER_PORT': '8200',
      'LOG_DIR': 'MOCK_LOG_DIR',
      # Caddy log format customization
      'ACCESS_LOG_FORMAT': '{upstream} {combined} [{>X-Forwarded-For}] {latency_ms}'
    })

    self.subprocess_mock.Popen.return_value.returncode = 0
    self.subprocess_mock.Popen.return_value.communicate.return_value = (
      'MOCK_OPENSSL_STDOUT', 'MOCK_OPENSSL_STDERR')

    gen_multiproc_conf.generate_Caddyfile()

    self.os_mock.makedirs.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess', exist_ok=True)
    self.open_mock.assert_called_once_with(
      'MOCK_BASEDIR/multiprocess/Caddyfile', 'wt')

    # no SSL configured on the reverse proxy:
    self.subprocess_mock.Popen.assert_not_called()

    handle: Mock = self.open_mock()
    written_text = handle.write.mock_calls[0].args[0]
    self.assertEqual(written_text, EXPECTED_CADDYCONF_CUSTOM)


"""
Running:

PYTHONPATH=`pwd`/../../.. python test_gen_multiproc_conf.py
"""
if __name__ == '__main__':
  unittest.main()
