import arc from 'arc';
import { ParamEnums, Param, CustomParam } from './paramClass';
import { RangeValueParam, DateRangeParam } from './rangeParamClass';
import { ColParam, PivotParam } from './colParamClass';

/*
                       Param
                         |
      ----------------------------------------
      |                  |                    |
  Col Param         Custom Param      Range Value Param
      |                                       |
  Pivot Param                         Date Range Param

*/
function convertOldParamArrToNewParamObjects(
  oldParams,
  lib = null,
  scopeConfig = null,
  arcapi = null,
  isFromURL = false
) {
  const paramObjects = [];
  const paramObjectHash = {};
  const keyToOldParamMap = {};
  const oldParamArr = oldParams || [];

  oldParamArr.forEach((param) => {
    keyToOldParamMap[param.key] = param;
  });

  oldParamArr.forEach(_convertOneOldParamToNewParamObject);

  function _convertOneOldParamToNewParamObject(oldParam) {
    const keyArr = arc.params.splitParams(oldParam.key);
    const rootName = keyArr[0];
    const paramExtension = keyArr[1];
    let paramObject = paramObjectHash[rootName];
    if (paramObject) {
      return;
    }

    let paramConfig = {
      arcapi,
      outputParameter: rootName,
    };

    if (isFromURL) {
      paramConfig.entityType = ParamEnums.EntityTypeEnum.URL;
    }

    const ParamObjectClass = _identifyClassForOldParam(
      keyToOldParamMap, rootName, paramExtension, oldParam
    );
    if (!ParamObjectClass) {
      return;
    }

    const configProperties = {
      ParamObjectClass,
      keyToOldParamMap,
      oldParam,
      rootName,
      paramExtension,
      scopeConfig,
      paramConfig
    };
    paramConfig = _populatePropertiesForParamConfig(configProperties);
    paramObject = new ParamObjectClass(lib, paramConfig);
    paramObjects.push(paramObject);
    paramObjectHash[rootName] = paramObject;
  }

  return paramObjects;
}

function _identifyClassForOldParam(keyToOldParamMap, rootName, paramExtension, oldParam) {
  const aliasParam = keyToOldParamMap[`${rootName}.alias`];
  let ParamObjectClassName;
  if (aliasParam) {
    return CustomParam;
  }

  if (paramExtension === 'start') {
    const number = Number(oldParam.value);

    return Number.isNaN(number) ? DateRangeParam : RangeValueParam;
  }

  switch (paramExtension) {
    case 'end':
    case 'config':
      return;
    case 'data':
    case 'include_nulls':
      ParamObjectClassName = Param;
      break;
    case 'ignore':
      if (oldParam.value === true) {
        ParamObjectClassName = Param;
      } else {
        return;
      }

      break;
    case 'exclude':
      return;
    case 'pivot':
      ParamObjectClassName = PivotParam;
      break;
    default:
      if (oldParam.custom_expr || oldParam.expr || oldParam.filtername ||
        oldParam.fn_cast || oldParam.fn_name) {
        ParamObjectClassName = ColParam;
      } else {
        ParamObjectClassName = Param;
      }

      break;
  }

  return ParamObjectClassName;
}

function _populatePropertiesForParamConfig(configProperties) {
  const {
    ParamObjectClass,
    keyToOldParamMap,
    oldParam,
    rootName,
    scopeConfig,
  } = configProperties;
  const identifiedParamExtension = configProperties.paramExtension;
  const currentParamConfig = configProperties.paramConfig;
  let valueParam = oldParam;
  const paramConfig = $.extend({}, currentParamConfig);
  let paramExtension = identifiedParamExtension;
  if (ParamObjectClass === CustomParam) {
    const aliasParam = keyToOldParamMap[`${rootName}.alias`];
    if (aliasParam) {
      paramConfig.alias = aliasParam.value_list ||
        arc.params.parse_list(aliasParam.value) ||
        [];
    }

    if (keyToOldParamMap[`${rootName}.data`]) {
      valueParam = keyToOldParamMap[`${rootName}.data`];
      paramExtension = 'data';
    } else if (keyToOldParamMap[`${rootName}`]) {
      valueParam = keyToOldParamMap[`${rootName}`];
      paramExtension = '';
    }
  }

  if (paramExtension === 'start') {
    const start = keyToOldParamMap[`${rootName}.start`].value;
    const end = keyToOldParamMap[`${rootName}.end`].value;
    const isDateRangeParam = ParamObjectClass === DateRangeParam;

    paramConfig.value = [{
      start,
      end
    }];

    if (isDateRangeParam) {
      if (keyToOldParamMap[`${rootName}.config`]) {
        // set dataType or hour/day format
        // set include_nulls
        // set relative/fixed/quick etc.
        const localConfig = JSON.parse(keyToOldParamMap[`${rootName}.config`].value);
        paramConfig.datetimeFormat = localConfig.datetimeFormat;
        paramConfig.relativeDate = localConfig[arc.misc.relativeFilterSuffix];
        paramConfig.quickFilters = localConfig.btn;
        paramConfig.datetimeFormat =
          (paramConfig.relativeDate || paramConfig.quickFilters)?.datetimeFormat;
      } else {
        paramConfig.datetimeFormat = _getDateTimeFormat(start);
      }
    }
  } else if (paramExtension === 'data') {
    paramConfig.valueType = ParamEnums.ValueTypeEnum.MULTI;

    let valueArray = keyToOldParamMap[`${rootName}.data`].value_list;
    if (!valueArray) {
      // For Multi value param, if value is string and not an array then parse it to an array.
      valueArray = arc.params.parse_list(keyToOldParamMap[`${rootName}.data`].value);
    }

    paramConfig.value = valueArray;
    paramConfig.colDataType = valueParam.valuetype ||
      (typeof valueArray[0] === 'number' ? 'NUMERIC' : 'STRING');
    const excludeParam = keyToOldParamMap[`${rootName}.exclude`];
    paramConfig.isExcluded = excludeParam && excludeParam.value === 'not in';
  } else if (paramExtension === 'pivot') {
    const pivotParam = keyToOldParamMap[`${rootName}.pivot`];
    paramConfig.value = pivotParam && pivotParam.value;
  } else if (paramExtension !== 'include_nulls') {
    paramConfig.value = valueParam.value_list || [valueParam.value];
  }

  const isIgnoreObject = keyToOldParamMap[`${rootName}.ignore`];
  if (isIgnoreObject && isIgnoreObject.value === true) {
    paramConfig.isIgnore = true;
  }

  if ([ColParam, PivotParam].includes(ParamObjectClass)) {
    paramConfig.pivot = {
      sourceShelf: valueParam.source_shelf,
      customExpr: valueParam.custom_expr,
      expr: valueParam.expr,
      fnName: valueParam.fn_name,
      fnCast: valueParam.fn_cast,
    };
    paramConfig.baseField = valueParam.filtername;
  }

  if (scopeConfig) {
    if (scopeConfig.dataset[rootName] !== undefined) {
      paramConfig.scopeType = ParamEnums.ScopeTypeEnum.DATASET;
      paramConfig.datasetId = scopeConfig.dataset[rootName].datasetId;
      paramConfig.entityId = scopeConfig.dataset[rootName].filterId;
    } else {
      paramConfig.scopeType = ParamEnums.ScopeTypeEnum.GLOBAL;
      paramConfig.entityId = scopeConfig.global[rootName]?.filterId;
      paramConfig.entityType = arc.params.ParamEnums.EntityTypeEnum.FILTER;
    }
  }

  if (keyToOldParamMap[`${rootName}.include_nulls`]) {
    paramConfig.includeNulls = keyToOldParamMap[`${rootName}.include_nulls`].value === 'true';
  }

  return paramConfig;
}

function _getDateTimeFormat(value) {
  const strVal = String(value);
  if (strVal.indexOf(':') < 0) {
    return ParamEnums.DateEnums.DateTimeFormats.DATE[0];
  }

  if (strVal.indexOf('[') < 0) {
    return ParamEnums.DateEnums.DateTimeFormats.DATETIME[0];
  }

  return ParamEnums.DateEnums.DateTimeFormats.DATETIME[1];
}

function convertOldParamHashToNewParamObjects(
  oldParamHash,
  lib = null,
  scopeConfig = null,
  arcapi = null
) {
  const paramHash = oldParamHash || {};

  const oldParamArr = Object.entries(paramHash).map(([key, value]) => ({
    key,
    value_list: arc.params.parse_list(value),
    value,
  }));

  return convertOldParamArrToNewParamObjects(oldParamArr, lib, scopeConfig, arcapi);
}

// Converts params to hash object.
function convertToOldFormatHash(paramObjects = []) {
  const oldParamFormatHash = {};

  const oldParamFormatArr = convertToOldFormatArr(paramObjects);
  oldParamFormatArr.forEach((paramOldFormat) => {
    const keyExtension = arc.params.splitParams(paramOldFormat.key)[1];
    if (keyExtension === 'data') {
      oldParamFormatHash[paramOldFormat.key] = arc.params.format_list(
        paramOldFormat.value_list, paramOldFormat.valueType
      ).join(', ');
    } else {
      oldParamFormatHash[paramOldFormat.key] = paramOldFormat.value;
    }
  });

  return oldParamFormatHash;
}

// Converts params to old format {key, value} list.
function convertToOldFormatArr(paramObjects = []) {
  const oldParamFormatArr = [];

  paramObjects.forEach(function _convertOneParamObjectToOldFormatParams(param) {
    const paramOldFormat = param.toOldParamFormat();
    if (param instanceof ColParam) {
      const oldFormatObject =
        updateOldParamWithExtraProperties(param, paramOldFormat);
      if (!(param instanceof PivotParam)) {
        // PivotParams already go through this code being ColParams
        oldFormatObject.value = oldFormatObject.value[0];
      }

      oldParamFormatArr.push(oldFormatObject);

      const key = `${oldFormatObject.key}.include_nulls`;

      if (oldFormatObject[key]) {
        oldParamFormatArr.push({
          key,
          value: oldFormatObject[key][0],
          value_list: oldFormatObject[key],
        });
      }

      return;
    }

    Object.keys(paramOldFormat).forEach(function _populateOldFormatObject(key) {
      const paramExtension = arc.params.splitParams(key)[1];
      const isSingleValueParam =
        param.meta.valueType !== ParamEnums.ValueTypeEnum.MULTI ||
        ['ignore'].includes(paramExtension);

      if (isSingleValueParam ||
        param.shouldHideExtension() ||
        paramExtension === 'exclude' || paramExtension === 'include_nulls') {
        const paramValue = paramOldFormat[key];
        if (paramValue.length) {
          let oldFormatObject = {
            key,
            value: paramValue[0],
            value_list: paramValue,
          };
          oldFormatObject = updateOldParamWithExtraProperties(param, oldFormatObject);
          oldParamFormatArr.push(oldFormatObject);
        }

        return;
      }

      let valueType = param.dataType;
      if (![
        ParamEnums.DataTypeEnum.STRING, ParamEnums.DataTypeEnum.NUMERIC,
      ].includes(valueType)) {
        valueType = ParamEnums.DataTypeEnum.STRING;
      }

      let oldFormatObject = {
        key,
        value: arc.params.format_list(
          paramOldFormat[key], valueType, false,
          param.source.connectionType || 'url', true
        ).join(', '),
        value_list: paramOldFormat[key],
      };
      oldFormatObject = updateOldParamWithExtraProperties(param, oldFormatObject);
      oldParamFormatArr.push(oldFormatObject);
    });
  });

  function updateOldParamWithExtraProperties(paramObject, oldFormatObject) {
    let retObj = arc.misc.extend({}, oldFormatObject);
    const paramSource = paramObject.source;

    const valueType = oldFormatObject.key === `${paramObject.outputParameter}.include_nulls` ?
      ParamEnums.DataTypeEnum.STRING :
      (paramSource.colDataType || ParamEnums.DataTypeEnum.STRING);

    retObj = arc.misc.extend(retObj, {
      keytype: paramObject.meta.keyType,
      isDatasetSegment: paramObject.isDatasetSegment,
      source_visual_id: paramSource.entityId,
      source_dataset_id: paramSource.datasetId,
      valuetype: valueType,
    });

    const paramScope = paramObject.scope || {};
    const isDatasetScope = paramScope.scopeType === ParamEnums.ScopeTypeEnum.DATASET;
    const isGlobalScope = paramScope.scopeType === ParamEnums.ScopeTypeEnum.GLOBAL;
    const scopeObject = paramScope; // intentionally pointing to same object

    if (isGlobalScope || isDatasetScope || paramScope.isAppScope) {
      scopeObject.filterName = paramObject.outputParameter;

      if (paramSource.entityType === ParamEnums.EntityTypeEnum.FILTER ||
        paramSource.entityType === ParamEnums.EntityTypeEnum.URL) {
        scopeObject.filterId = paramSource.entityId;
      }

      if (isDatasetScope) {
        scopeObject.datasetId = paramSource.datasetId;
      }

      if (paramScope.isAppScope) {
        scopeObject.appId = paramSource.appId;
        retObj.appId = paramSource.appId;
      }

      retObj.scope = scopeObject;
    }

    return retObj;
  }

  return oldParamFormatArr;
}

function stringifyParamObjects(paramObjects = []) {
  return JSON.stringify(paramObjects);
}

function parseParamObjects(stringifiedObjects = '[]') {
  const parsedObjects = JSON.parse(stringifiedObjects);
  return parsedObjects.map((jsonifiedObject) => {
    const _className = jsonifiedObject._className || 'Param';
    const ParamObjectClass = arc.params[
      ParamEnums.getKeyForValue(_className, ParamEnums.ClassTypeEnum)
    ];
    const object = new ParamObjectClass();
    object.importFromJSON(jsonifiedObject);
    delete object.name;
    return object;
  });
}

arc.params = arc.params || {};
arc.params.Param = Param;
arc.params.ParamEnums = ParamEnums;
arc.params.CustomParam = CustomParam;
arc.params.RangeValueParam = RangeValueParam;
arc.params.DateRangeParam = DateRangeParam;
arc.params.ColParam = ColParam;
arc.params.PivotParam = PivotParam;
arc.params.convertOldParamArrToNewParamObjects = convertOldParamArrToNewParamObjects;
arc.params.convertOldParamHashToNewParamObjects = convertOldParamHashToNewParamObjects;
arc.params.convertToOldFormatHash = convertToOldFormatHash;
arc.params.convertToOldFormatArr = convertToOldFormatArr;
arc.params.stringifyParamObjects = stringifyParamObjects;
arc.params.parseParamObjects = parseParamObjects;
