import { getGlobalArc } from 'utils/get-global-arc';
import { vizHttp } from 'services/viz-http';
import {
  JobThresholdTriggerCondition,
  JobThresholdTriggerType,
} from 'types/job.type';

const arc = getGlobalArc();

type CreateEmailJobParams = {
  isThreshold: boolean;
  name?: string;
  scheduleGroupId?: number;
  scheduleGroup?: {
    name: string;
    interval: string;
  } | null;
  report: {
    url: string;
    dbWidth: number;
    dbHeight: number;
  };
  emailTarget: {
    subject: string;
    message?: string;
    to: string[];
    cc?: string[];
    errorTo?: string[];
    replyTo?: string;
    templateId?: number;
    urlParams?: string;
    alias?: string;
    pdfOption?: string;
    includeEmbeddedImage: boolean;
    includePdf: boolean;
    includePng: boolean;
    includeCSV: boolean;
    includeXLSX: boolean;
    csvDownloadRows?: number;
  };
  thresholdTarget?: {
    triggerId: number;
    triggerType: JobThresholdTriggerType;
    triggerConditions: JobThresholdTriggerCondition[];
    minimumTimeInterval?: number;
  } | null;
};

const getFormType = (params: CreateEmailJobParams) => {
  if (params.isThreshold) {
    return 'trigger';
  }

  return params.scheduleGroupId || params.scheduleGroup
    ? 'schedule'
    : 'emailnow';
};

export const createEmailJob = async (
  reportId: number,
  params: CreateEmailJobParams
) => {
  const result = await vizHttp.post<{ schedule_id: number }>({
    url: `${arc.frontend.globals.urls.emailUrl}/${reportId}`,
    headers: {
      'X-CSRFTOKEN': arc.frontend.globals.authentication.csrfToken,
    },
    isOldEndpoint: true,
    data: {
      formType: getFormType(params),
      emailFormData: JSON.stringify({
        saveType: 'screenshot',
        paramString: params.emailTarget.urlParams, // Same as url_params
        csv_download_rows: params.emailTarget.csvDownloadRows,
        dbWidth: params.report.dbWidth,
        dbHeight: params.report.dbHeight,
        location_pathname: params.report.url,
        ...(params.emailTarget.alias && {
          visual_url: `${window.location.origin}${arc.frontend.globals.urls.urlShortenerUrl}${params.emailTarget.alias}`,
        }),
        image_type: [
          params.emailTarget.includeEmbeddedImage && 'E',
          params.emailTarget.includePdf && 'D',
          params.emailTarget.includePng && 'A',
          params.emailTarget.includeXLSX && 'X',
          params.emailTarget.includeCSV && 'C',
        ]
          .filter(Boolean)
          .join(','),
        job_name: params.name,
        schedule_group: params.scheduleGroupId,
        schedule_group_name: params.scheduleGroup?.name,
        run_interval: params.scheduleGroup?.interval,

        // Email target params
        email_subject: params.emailTarget.subject,
        message: params.emailTarget.message,
        email_template: params.emailTarget.templateId,
        to_addresses: params.emailTarget.to,
        cc_addresses: params.emailTarget.cc,
        error_to_addresses: params.emailTarget.errorTo,
        reply_to_address: params.emailTarget.replyTo,
        url_params: params.emailTarget.urlParams,
        pdf_option: params.emailTarget.pdfOption,

        // Threshold target params
        trigger_id: params.thresholdTarget?.triggerId,
        trigger_type: params.thresholdTarget?.triggerType,
        trigger_conditions: params.thresholdTarget?.triggerConditions.map(
          (condition, index) => ({
            [index]: {
              'alert-equality': condition.equality,
              'alert-trigger-value': condition.value,
            },
          })
        ),
        minimum_time_interval: params.thresholdTarget?.minimumTimeInterval,
      }),
    },
  });

  return result.data.schedule_id;
};
