import {
  DataCopyTarget,
  EmailTarget,
  JobThresholdTriggerType,
  PdfUploadFormatType,
  RefreshTarget,
  SnapshotTarget,
  ThresholdTarget,
} from 'types/job.type';
import { BackendJob } from './parse-job';

type JobTargets = BackendJob['targets'];

export type BackendTargetByTargetType = {
  D: {
    extractId: number;
    username: string;
    srcdataset_url: string;
    jobs_url: string;
  };
  E: {
    app_id: number;
    visual_url?: string;
    email_subject: string;
    message: string;
    email_template?: number | '';
    image_type?: string;
    url_params: string;
    to_addresses?: string[];
    cc_addresses?: string[];
    error_to_addresses?: string[];
    reply_to_address: string;
  };
  S: {
    app_id: number;
    include_csv: number;
    include_xlsx: number;
    csv_download_rows: number;
    dbWidth: number;
    dbHeight: number;
    email_subject: string;
    error_to_addresses?: string[];
    reply_to_address?: string[];
    paramString: string;
    uploadFormat?: PdfUploadFormatType | 'unknown';
  };
  T: {
    trigger_id: number;
    trigger_type: JobThresholdTriggerType;
    trigger_conditions?: {
      [key: string]: {
        'alert-quality': string;
        'alert-trigger-value': string;
      };
    }[];
  };
  R: {
    dcid: number;
    dataconn_name: string;
    view: string;
  };
};

const getBackendTarget = <Key extends keyof BackendTargetByTargetType>(
  targets: JobTargets,
  targetType: Key
):
  | null
  | (BackendTargetByTargetType[Key] & {
      minimum_time_interval?: number;
    }) => {
  const target = targets?.find(
    ({ fields }) => fields.target_type === targetType
  );

  if (!target || !target.fields.target_data) {
    return null;
  }

  try {
    const data = JSON.parse(
      target.fields.target_data
    ) as BackendTargetByTargetType[typeof targetType];

    return {
      ...data,
      minimum_time_interval: target.fields.minimum_time_interval,
    };
  } catch (error) {
    console.error('Failed to parse job target', error);
    return null;
  }
};

export const parseSnapshotTarget = (
  targets: JobTargets
): SnapshotTarget | undefined => {
  const snapshot = getBackendTarget(targets, 'S');

  if (!snapshot) {
    return undefined;
  }

  return {
    id: snapshot.app_id,
    includePageParams: [
      ...new URLSearchParams(snapshot.paramString).keys(),
    ].some((param) => !['embed', 'sheet'].includes(param)),
    includeCSV: Boolean(snapshot.include_csv),
    includeXLSX: Boolean(snapshot.include_xlsx),
    csvDownloadRows: snapshot.csv_download_rows,
    dbWidth: snapshot.dbWidth,
    dbHeight: snapshot.dbHeight,
    subject: snapshot.email_subject,
    errorToAddresses: snapshot.error_to_addresses || [],
    ccAddresses: snapshot.reply_to_address || [],
    paramString: snapshot.paramString,
    uploadFormat:
      snapshot.uploadFormat === 'unknown' ? undefined : snapshot.uploadFormat,
  };
};

export const parseDataCopyTarget = (
  targets: JobTargets
): DataCopyTarget | undefined => {
  const dataCopy = getBackendTarget(targets, 'D');

  if (!dataCopy) {
    return undefined;
  }

  return {
    id: dataCopy.extractId,
    username: dataCopy.username,
    scrDatasetURL: dataCopy.srcdataset_url,
    jobsUrl: dataCopy.jobs_url,
  };
};

export const parseEmailTarget = (
  targets: JobTargets
): EmailTarget | undefined => {
  const email = getBackendTarget(targets, 'E');

  if (!email) {
    return undefined;
  }

  return {
    id: email.app_id,
    visualURL: email.visual_url || '',
    useUrlAlias: email.visual_url?.includes('/short/') || false,
    includeEmbeddedImage: email.image_type?.includes('E') || false,
    includePng: email.image_type?.includes('A') || false,
    includePdf: email.image_type?.includes('D') || false,
    subject: email.email_subject,
    message: email.message,
    emailTemplateId:
      typeof email.email_template === 'number'
        ? email.email_template
        : undefined,
    urlParams: email.url_params,
    toAddresses: email.to_addresses || [],
    ccAddresses: email.cc_addresses || [],
    errorToAddresses: email.error_to_addresses || [],
    replyToAddress: email.reply_to_address,
  };
};

export const parseThresholdTarget = (
  targets: JobTargets
): ThresholdTarget | undefined => {
  const threshold = getBackendTarget(targets, 'T');

  if (!threshold) {
    return undefined;
  }

  return {
    triggerId: threshold.trigger_id,
    triggerType: threshold.trigger_type,
    triggerConditions:
      threshold.trigger_conditions?.map((condition, index) => {
        const value = condition[index];
        return {
          equality: value['alert-equality'],
          value: Number(value['alert-trigger-value']),
        };
      }) || [],
    minimumTimeInterval: threshold.minimum_time_interval || 0,
  };
};

export const parseRefreshTarget = (
  targets: JobTargets
): RefreshTarget | undefined => {
  const refresh = getBackendTarget(targets, 'R');

  if (!refresh) {
    return undefined;
  }

  return {
    dataConnection: {
      id: refresh.dcid,
      name: refresh.dataconn_name,
    },
    view: refresh.view,
    minimumTimeInterval: refresh.minimum_time_interval,
  };
};
