import ErrorIcon from '@cloudera-internal/cuix-core/icons/svg/ic-error.svg';
import { yupResolver } from '@hookform/resolvers/yup';
import { useMutation, useQueryClient } from '@tanstack/react-query';
import { createAlias } from 'apis/aliases/create-alias';
import { editAlias } from 'apis/aliases/edit-alias';
import { useGetAliasesQuery } from 'apis/aliases/get-aliases-by-dashboard-id';
import { editJob } from 'apis/jobs/edit-job';
import { createEmailJob } from 'apis/jobs/email/create-email-job';
import { QueryKeyBase } from 'apis/query-keys';
import { AxiosError } from 'axios';
import { ClouderaButton } from 'components/cloudera-button/ClouderaButton';
import { useClouderaDialogContext } from 'components/cloudera-dialog/ClouderaDialog';
import { Notification } from 'components/notification/Notification';
import React, { useCallback, useState } from 'react';
import { Form, Modal, Tab, Tabs } from 'react-bootstrap';
import { Controller, useForm } from 'react-hook-form';
import { useNotificationStore } from 'stores/use-notification-store';
import { EmailJob, JobReport } from 'types/job.type';
import { getGlobalArc } from 'utils/get-global-arc';
import { EmailJobEmailTab } from './components/EmailJobEmailTab';
import { EmailJobScheduleTab } from './components/EmailJobScheduleTab';
import { EmailJobThresholdTab } from './components/EmailJobThresholdTab';
import {
  EmailJobFormSchema,
  emailJobFormSchema,
} from './email-job-form-schema';
import classes from './EmailJobModal.module.scss';
import { cdlRed500 } from '@cloudera-internal/cuix-core/variables';

const arc = getGlobalArc();

type EditJobModalProps =
  | { job: EmailJob }
  | { isCreate: true; report: JobReport; onEmailSent: (jobId: number) => void };

/**
 * Combined modal for creating or editing email jobs
 * @param job The email job to be edited
 * @param isCreate Indicates that the form should be set up for creating a new email job
 * @param report Object that contains the information passed down from the dashboard about the report and the dashboard itself
 */
export const EmailJobModal = (props: EditJobModalProps): JSX.Element => {
  const isCreate = 'isCreate' in props;

  const [activeTabKey, setActiveTabKey] = useState<string | null>('email');
  const { reset: resetDialog } = useClouderaDialogContext();
  const queryClient = useQueryClient();

  const { addNotification } = useNotificationStore(({ addNotification }) => ({
    addNotification,
  }));

  const { data: aliases } = useGetAliasesQuery(
    isCreate ? props.report.id : NaN,
    { enabled: isCreate }
  );

  const form = useForm({
    resolver: yupResolver(emailJobFormSchema),
    defaultValues: isCreate
      ? {
          isCreate: true,
          isThreshold: false,
          isScheduled: false,
          emailTarget: {
            message: '',
            subject:
              `[Cloudera Data Visualization] ${props.report.title}`.trim(),
            cc: [],
            to: [],
            errorTo: [],
            replyTo: arc.frontend.globals.settings.emailHostUser,
            urlParams: props.report.paramString,
            useUrlAlias: false,
            includeEmbeddedImage: true,
            includePdf: false,
            includePng: false,
            includePageParams: true,
            includeXLSX: false,
            includeCSV: false,
          },
        }
      : {
          isCreate: false,
          isThreshold: !!props.job.thresholdTarget,
          isScheduled: !!props.job.scheduleGroup,
          name: props.job.name,
          scheduleGroupId: props.job.scheduleGroup?.id,
          emailTarget: {
            message: props.job.emailTarget.message,
            subject: props.job.emailTarget.subject ?? '',
            templateId: props.job.emailTarget.emailTemplateId,
            cc: props.job.emailTarget.ccAddresses ?? [],
            to: props.job.emailTarget.toAddresses ?? [],
            errorTo: props.job.emailTarget.errorToAddresses ?? [],
            replyTo: props.job.emailTarget.replyToAddress,
            urlParams: props.job.emailTarget.urlParams,
            useUrlAlias: props.job.emailTarget.useUrlAlias,
            includeEmbeddedImage: props.job.emailTarget.includeEmbeddedImage,
            includePdf: props.job.emailTarget.includePdf,
            includePng: props.job.emailTarget.includePng,
            includePageParams:
              props.job.snapshotTarget?.includePageParams ?? false,
            includeXLSX: props.job.snapshotTarget?.includeXLSX ?? false,
            includeCSV: props.job.snapshotTarget?.includeCSV ?? false,
            csvDownloadRows: props.job.snapshotTarget?.csvDownloadRows,
          },
          thresholdTarget: props.job.thresholdTarget,
        },
  });

  const {
    register,
    handleSubmit,
    control,
    clearErrors,
    formState,
    getValues,
    setValue,
    watch,
  } = form;
  const isScheduled = watch('isScheduled');
  const isThreshold = watch('isThreshold');

  const {
    mutateAsync: createMutation,
    reset: resetCreate,
    error: createError,
  } = useMutation<number | undefined, AxiosError<string>, EmailJobFormSchema>({
    mutationFn: async (formFields) => {
      if (!isCreate) {
        return;
      }

      if (formFields.emailTarget.useUrlAlias) {
        if (
          aliases?.some((alias) => alias.slug === formFields.emailTarget.alias)
        ) {
          await editAlias({
            slug: formFields.emailTarget.alias!,
            updateType: 'email',
          });
        } else {
          await createAlias({
            dashboardId: props.report.dashboardId,
            slug: formFields.emailTarget.alias!,
            parameters: formFields.emailTarget.urlParams,
            isCreatedInEmail: true,
            type: 'dashboard',
          });
        }
      }

      return createEmailJob(props.report.id, {
        ...formFields,
        report: props.report,
      });
    },
    onSuccess: (jobId, formFields) => {
      queryClient.invalidateQueries([QueryKeyBase.Jobs]);

      if (
        formFields.scheduleGroup?.name &&
        formFields.scheduleGroup?.interval
      ) {
        queryClient.invalidateQueries([QueryKeyBase.ScheduleGroups]);
      }
      resetDialog();

      if (!jobId || !isCreate) {
        return;
      }

      if (formFields.isThreshold) {
        addNotification(
          {
            variant: 'success',
            message: `Threshold email successfully created with job ID ${jobId}`,
          },
          { delay: 3000 }
        );
      } else if (formFields.isScheduled) {
        addNotification(
          {
            variant: 'success',
            message: `Scheduled email successfully created with job ID ${jobId}`,
          },
          { delay: 3000 }
        );
      } else {
        props.onEmailSent(jobId);
      }
    },
  });

  const {
    mutateAsync: editMutation,
    reset: resetEdit,
    error: editError,
  } = useMutation<void, AxiosError<string>, EmailJobFormSchema>({
    mutationFn: async (formFields) => {
      if (isCreate) {
        return;
      }

      return editJob(props.job.id, formFields);
    },
    onSuccess: (_, formFields) => {
      queryClient.invalidateQueries([QueryKeyBase.Jobs]);

      if (
        formFields.scheduleGroup?.name &&
        formFields.scheduleGroup?.interval
      ) {
        queryClient.invalidateQueries([QueryKeyBase.ScheduleGroups]);
      }
      resetDialog();
    },
  });

  const resetErrors = useCallback(() => {
    resetCreate();
    resetEdit();
    clearErrors('root');
  }, [clearErrors, resetCreate, resetEdit]);

  const onSubmit = useCallback(
    async (formFields: EmailJobFormSchema) => {
      try {
        isCreate
          ? await createMutation(formFields)
          : await editMutation(formFields);
      } catch {
        // error is handled inside useMutation
      }
    },
    [createMutation, editMutation, isCreate]
  );

  const onThresholdChange = useCallback(
    (event: React.ChangeEvent<HTMLInputElement>) => {
      if (!event.target.checked && activeTabKey === 'threshold') {
        setActiveTabKey('email');
      } else if (
        event.target.checked &&
        !getValues('thresholdTarget.triggerConditions')?.length
      ) {
        setValue('thresholdTarget.triggerConditions', [
          { value: 0, equality: '>=' },
        ]);
      }
    },
    [activeTabKey, getValues, setValue]
  );

  return (
    <>
      <Modal.Header closeButton>
        <Modal.Title>
          {isCreate ? `Send ${props.report.type}` : 'Edit Job Details'}
        </Modal.Title>
      </Modal.Header>

      <Modal.Body>
        <Notification
          title="Error"
          message={
            createError?.response?.data ??
            editError?.response?.data ??
            formState.errors.root?.message ??
            ''
          }
          variant="danger"
          show={!!createError || !!editError || !!formState.errors.root}
          dismissible
          onClose={resetErrors}
        />

        <Form role="form" onSubmit={handleSubmit(onSubmit)} id="emailModalForm">
          {isCreate && (
            <Form.Group controlId="emailingMethod">
              <Form.Label>Emailing Method</Form.Label>

              <div>
                <Controller
                  name="isScheduled"
                  control={control}
                  render={({ field: { value, onChange, ...field } }) => (
                    <Form.Check
                      inline
                      type="radio"
                      label="Email now"
                      id="isScheduledFalse"
                      checked={!value}
                      onChange={() => {
                        setActiveTabKey('email');
                        onChange(false);
                      }}
                      {...field}
                    />
                  )}
                />
                <Controller
                  name="isScheduled"
                  control={control}
                  render={({ field: { value, onChange, ...field } }) => (
                    <Form.Check
                      inline
                      type="radio"
                      label="Schedule email"
                      id="isScheduledTrue"
                      checked={value}
                      onChange={() => onChange(true)}
                      {...field}
                    />
                  )}
                />
              </div>
            </Form.Group>
          )}
          {isCreate && isScheduled && (
            <Form.Group controlId="isThreshold">
              <Form.Label>Threshold</Form.Label>
              <Form.Switch
                label="Enable"
                {...register('isThreshold', {
                  shouldUnregister: true,
                  onChange: onThresholdChange,
                })}
              />
            </Form.Group>
          )}

          {!isCreate && (
            <Form.Group>
              <Form.Label>ID</Form.Label>
              <Form.Control value={props.job.id} disabled />
            </Form.Group>
          )}

          <Tabs
            activeKey={activeTabKey || 'email'}
            onSelect={setActiveTabKey}
            className={classes.editModalTabs}
          >
            <Tab
              eventKey="email"
              title={
                <span className="d-flex align-items-center gap-1">
                  Email Details
                  {!!formState.errors.emailTarget && (
                    <ErrorIcon color={cdlRed500} />
                  )}
                </span>
              }
            >
              <EmailJobEmailTab
                reportId={isCreate ? props.report.id : props.job.emailTarget.id}
                reportType={isCreate ? props.report.type : 'sheet'}
                reportUrl={
                  isCreate ? props.report.url : props.job.emailTarget.visualURL
                }
                dashboardId={isCreate ? props.report.dashboardId : undefined}
                disableXlsxAndCsv={isCreate && props.report.disableXlsxAndCsv}
                form={form}
              />
            </Tab>
            {isScheduled && (
              <Tab
                eventKey="schedule"
                title={
                  <span className="d-flex align-items-center gap-1">
                    Schedule Details
                    {!!formState.errors.name && <ErrorIcon color={cdlRed500} />}
                  </span>
                }
              >
                <EmailJobScheduleTab form={form} />
              </Tab>
            )}
            {isThreshold && (
              <Tab
                eventKey="threshold"
                title={
                  <span className="d-flex align-items-center gap-1">
                    Threshold Details
                    {!!formState.errors.thresholdTarget && (
                      <ErrorIcon color={cdlRed500} />
                    )}
                  </span>
                }
              >
                <EmailJobThresholdTab
                  form={form}
                  report={isCreate ? props.report : undefined}
                />
              </Tab>
            )}
          </Tabs>
        </Form>
      </Modal.Body>
      <Modal.Footer>
        <ClouderaButton
          type="button"
          disabled={formState.isSubmitting}
          onClick={resetDialog}
        >
          Cancel
        </ClouderaButton>
        <ClouderaButton
          disabled={
            formState.isSubmitting || !!Object.keys(formState.errors).length
          }
          color="primary"
          type="submit"
          form="emailModalForm"
          testId="submitEmailModalForm"
        >
          Save
        </ClouderaButton>
      </Modal.Footer>
    </>
  );
};
